package de.mannheim.uni.index.io;

import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.stream.IntStream;

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.index.DirectoryReader;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.IndexWriterConfig;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;
import org.apache.lucene.store.MMapDirectory;
import org.apache.lucene.util.Version;

import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.index.IIndex;

public class DefaultIndex implements IIndex {

	private IndexWriter indexWriter = null;
	IndexSearcher indexSearcher = null;
	private String indexDir;
	private IndexReader indexReader = null;
	private Directory dir;
	
	private Map<Integer, Document> map = new ConcurrentHashMap<>();
	
	public DefaultIndex(String indexdir)
	{
		indexDir = indexdir;
	}
	
	public String getIndexDir() {
		return indexDir;
	}
	
	@Override
	public Document getDocument(Integer i ) {
		return map.get(i);
	}
	
	public IndexSearcher getIndexSearcher() {
		if (indexSearcher == null) {

			try {
//				File indexDirFile = new File(this.indexDir);
				 Path path = Paths.get(this.indexDir);
				   MMapDirectory mDir = new MMapDirectory(path);
				   mDir.setPreload(true);
				dir = mDir;
 
				 
				//Directory dir = MMapDirectory.open(indexDirFile); 
				indexReader = DirectoryReader.open(dir);
				indexSearcher = new IndexSearcher(indexReader);
				
				ConcurrentLinkedQueue<Integer> queue = new ConcurrentLinkedQueue<>();
				
				IntStream.range(0, indexReader.maxDoc()).parallel().forEach((e) -> {
					queue.add(e);
				});
				 
				 PersistentParallel6.runAndFinish(queue, (i) -> {
					 try {
						  Document doc = indexReader.document(i);
						    map.put(i, doc);
					 }
					 catch(IOException e) {
						 throw new RuntimeException(e);
					 }
				 }, this.getClass());
						 

			} catch (IOException ioe) {
				ioe.printStackTrace();
			}
		}
		return this.indexSearcher;
	}

	public IndexWriter getIndexWriter() {
		if (indexWriter == null) {
			try {
				File indexDirFile = new File(this.indexDir);

				Path path = Paths.get(this.indexDir);
				 MMapDirectory dir = new MMapDirectory(path);
				 dir.setPreload(true);
				 this.dir = dir;
//				Directory dir = FSDirectory.open(indexDirFile);
				Analyzer analyzer = new StandardAnalyzer( );
				IndexWriterConfig iwc = new IndexWriterConfig(  analyzer);

				iwc.setOpenMode(IndexWriterConfig.OpenMode.CREATE_OR_APPEND);

				indexWriter = new IndexWriter(dir, iwc);
				// IndexWriterConfig conf = new IndexWriterConfig(
				// Version.LUCENE_46, analyzer);
				indexWriter.getConfig().setRAMBufferSizeMB(1024);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return indexWriter;
	}

	public void closeIndexWriter() {
		if (indexWriter != null) {
			try {
				indexWriter.commit();
				indexWriter.close();
				indexWriter = null;
			} catch (Exception e) {
				e.printStackTrace();
			}

		}
	}
	
	public void closeIndexReader() {
		try {
			if (indexReader != null) {
				indexReader.close();
				indexReader = null;
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public int getNmDocs() {
		return getIndexSearcher().getIndexReader().numDocs();
	}

	@Override
	public void closeDir() throws IOException {
		this.dir.close();
		
	}
	@Override
	public Directory getDir() {
		return dir;
	}

	public IndexReader getIndexReader() {
		return indexReader;
	}

 
	
	
}
