package de.mannheim.uni.index.io;

import java.io.IOException;

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.index.DirectoryReader;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.IndexWriterConfig;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;
import org.apache.lucene.store.IOContext;
import org.apache.lucene.store.RAMDirectory;
import org.apache.lucene.util.Version;

import de.mannheim.uni.index.IIndex;

public class InMemoryIndex implements IIndex {

	private IndexWriter indexWriter = null;
	volatile IndexSearcher indexSearcher = null;
	private IndexReader indexReader = null;
	Directory directory = null;
	
	public InMemoryIndex()
	{
		directory = new RAMDirectory();
	}
	
	public InMemoryIndex(FSDirectory dir)
	{
		try {
			directory = new RAMDirectory(dir, IOContext.DEFAULT);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	public IndexSearcher getIndexSearcher() {
		if (indexSearcher == null) {

			try {
				indexReader = DirectoryReader.open(directory);
				indexSearcher = new IndexSearcher(indexReader);

			} catch (IOException ioe) {
				ioe.printStackTrace();
			}
		}
		return this.indexSearcher;
	}
	
	public IndexSearcher reopenIndexSearcher() {
		try {
			indexReader = DirectoryReader.open(directory);
			indexSearcher = new IndexSearcher(indexReader);
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
		return this.indexSearcher;
	}

	public  IndexWriter getIndexWriter() {
		if (indexWriter == null) {
			try {
				Analyzer analyzer = new StandardAnalyzer( );
				IndexWriterConfig iwc = new IndexWriterConfig(
						  analyzer);

				iwc.setOpenMode(IndexWriterConfig.OpenMode.CREATE_OR_APPEND);

				indexWriter = new IndexWriter(directory, iwc);
				// IndexWriterConfig conf = new IndexWriterConfig(
				// Version.LUCENE_46, analyzer);
				indexWriter.getConfig().setRAMBufferSizeMB(1024);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return indexWriter;
	}

	public void closeIndexWriter() {
		if (indexWriter != null) {
			try {
				indexWriter.commit();
				indexWriter.close();
				indexWriter = null;
			} catch (Exception e) {
				e.printStackTrace();
			}

		}
	}
	
	public void closeIndexReader() {
		try {
			if (indexReader != null)
			{
				indexReader.close();
				indexReader = null;
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public int getNmDocs() {
		return getIndexSearcher().getIndexReader().numDocs();
	}

	@Override
	public void closeDir() throws IOException {
		directory.close();
		
	}

	@Override
	public Directory getDir() {
		return directory;
	}

}
