package de.uni_mannheim.informatik.dws.t2k.normalisation;

import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern; 

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.TokenStream;
import org.apache.lucene.analysis.core.WhitespaceTokenizer;
import org.apache.lucene.analysis.core.LowerCaseFilter;
import org.apache.lucene.analysis.core.StopFilter;
import org.apache.lucene.analysis.en.PorterStemFilter;
import org.apache.lucene.analysis.miscellaneous.WordDelimiterFilterFactory;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.analysis.tokenattributes.CharTermAttribute;
import org.apache.lucene.util.Version;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;

import de.mannheim.uni.expansion.utils.ExpansionStringUnescape;

import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.commons.lang3.StringUtils;

public class StringNormalizer {

    //determines which string/character corresponds to the null value
    public static String nullValue = "NULL";
    //remove customized stopwords
    public static List<String> stopWords = new ArrayList<>();

    /**
     * Use normaliseValue and normaliseHeader instead
     *
     * @param columnName
     * @return
     */
    public static String normaliseHeader(String columnName) {
        columnName = StringEscapeUtils.unescapeJava(columnName);
        columnName = columnName.replace("\"", "");
        columnName = columnName.replace("|", " ");
        columnName = columnName.replace(",", "");
        columnName = columnName.replace("{", "");
        columnName = columnName.replace("}", "");
        columnName = columnName.replaceAll("\n", "");

        columnName = columnName.replace("&nbsp;", " ");
        columnName = columnName.replace("&nbsp", " ");
        columnName = columnName.replace("nbsp", " ");
        columnName = columnName.replaceAll("<.*>", "");


        columnName = columnName.toLowerCase();
        columnName = columnName.trim();

        columnName = columnName.replaceAll("\\.", "");
        columnName = columnName.replaceAll("\\$", "");
        // clean the values from additional strings
        if (columnName.contains("/")) {
            columnName = columnName.substring(0, columnName.indexOf("/"));
        }

        if (columnName.contains("\\")) {
            columnName = columnName.substring(0, columnName.indexOf("\\"));
        }
        if (columnName.equals("")) {
            columnName = nullValue;
        }
        return columnName;
    }
    private static final Pattern bracketsPattern = Pattern.compile("\\(.*\\)");
     
    
 
    public static String cleanValue (String value ) {
        try {
            value = value.replaceAll("\n", "");
            value = value.replace("&nbsp;", " ");
            value = value.replace("&nbsp", " ");
            value = value.replaceAll("[&\\?]#[0-9]{1,3};", "");
            value = value.replace("nbsp", " ");
            value = value.replaceAll("<.*>", "");
            value = value.toLowerCase();
            value = value.trim();
            if (value.equals("")) {
                value = nullValue;
            }
       
        } catch (Exception e) {
        }
        return value;
    }
    
    private final static Pattern p1 = Pattern.compile("\n");
    private final static Pattern p2 = Pattern.compile("&nbsp;".toString(), Pattern.LITERAL);
    private final static Pattern p3 = Pattern.compile("&nbsp".toString(), Pattern.LITERAL);
    private final static Pattern p4 = Pattern.compile(".".toString(), Pattern.LITERAL);
    private final static Pattern p5 = Pattern.compile("[&\\?]#[0-9]{1,3};");
    private final static Pattern p6 = Pattern.compile("nbsp".toString(), Pattern.LITERAL);
    private final static Pattern p7 = Pattern.compile("&quot;".toString(), Pattern.LITERAL);
    private final static Pattern p8 = Pattern.compile("–".toString(), Pattern.LITERAL);
    private final static Pattern p9 = Pattern.compile("<.*>");
    // 
    
    public static String normaliseValue (String value, boolean removeContentInBrackets) {
        try {
 
            value = p1.matcher(value).replaceAll("");
            value = p2.matcher(value).replaceAll(Matcher.quoteReplacement(" ".toString()));
            value = p3.matcher(value).replaceAll(Matcher.quoteReplacement(" ".toString())); 
            value = p4.matcher(value).replaceAll(Matcher.quoteReplacement("".toString()));  
            value = p5.matcher(value).replaceAll("");
            value = p6.matcher(value).replaceAll(Matcher.quoteReplacement(" ".toString()));  
            value = p7.matcher(value).replaceAll(Matcher.quoteReplacement(" ".toString()));  
            value = p8.matcher(value).replaceAll(Matcher.quoteReplacement("-".toString()));   
            value = p9.matcher(value).replaceAll(""); 
            value = value.toLowerCase();
            value = value.trim();
            if (value.equals("")) {
                value = nullValue;
            }
            if (removeContentInBrackets) {
                value = bracketsPattern.matcher(value).replaceAll("");
            }
        } catch (Exception e) {
        }
        return value;
    }
    
//    {
//    	  value = value.replaceAll("\n", "");
//          value = value.replace("&nbsp;", " ");
//          value = value.replace("&nbsp", " ");
//          value = value.replace(".", "");
//          value = value.replaceAll("[&\\?]#[0-9]{1,3};", "");
//          value = value.replace("nbsp", " ");
//          value = value.replace("&quot;", " ");
//          value = value.replace("–", "-");
//          value = value.replaceAll("<.*>", "");
//          value = value.toLowerCase();
//          value = value.trim();
//          if (value.equals("")) {
//              value = nullValue;
//          }
//          if (removeContentInBrackets) {
//              value = bracketsPattern.matcher(value).replaceAll("");
//          }
//    }
    
    public static String removeBrackets(String value) {
    	
    	 return bracketsPattern.matcher(value).replaceAll("");
    }
    
    public static String removeAfterComma(String value) {
    	
    	int commaIndexOf =  value.indexOf(',');
    	if(commaIndexOf >= 0) {
    		return value.substring(0, commaIndexOf);
    	}
    	else {
    		return value;
    	}
   
   }
    public static 	 Set<String> removeAfterCommaAndRemoveBrackets(Set<String> labels, boolean normalize) {
		
    	Set<String> ret = new HashSet<>();
    	
    	for(String label : labels) {
    		if(normalize) {
    			label = StringNormalizer.normaliseValue(label, false);
    		}
    		ret.add(StringNormalizer.removeAfterCommaAndRemoveBrackets(label));
    	}
		 
		return ret;
	}
    
    public static 	 Set<String> removeAfterCommaAndRemoveBrackets(Set<String> labels) {
		
    	return removeAfterCommaAndRemoveBrackets(labels, false);
	}
    public static String removeAfterCommaAndRemoveBrackets(String value) {
    	return removeBrackets(removeAfterComma(value));
    }

    /**
     * splits the string into tokens and concatenates it again, inserting white
     * spaces between all tokens
     *
     * @param s
     * @return
     */
    public static String normalise(String s, boolean useStemmer) {

        return StringUtils.join(tokenise(s, useStemmer), " ");

    }
    public static List<String> tokenise(String s, boolean useStemmer) {
    	return tokenise(s, useStemmer, true);
    }
    
    public static String sortAndJoin(List<String> tokens) {
    	Collections.sort(tokens);
    	return StringUtils.join(tokens, " ");
    }
    public static List<String> tokenise(String s, boolean useStemmer, boolean removeStopWords) {
        Analyzer analyzer = new StandardAnalyzer();
        List<String> result = new ArrayList<String>();

        try {

            Map<String, String> args = new HashMap<String, String>();
            args.put("generateWordParts", "1");
            args.put("generateNumberParts", "1");
            args.put("catenateNumbers", "0");
            args.put("splitOnCaseChange", "1");
            WordDelimiterFilterFactory fact = new WordDelimiterFilterFactory(args);

            try {
            // resolve non unicode chars
            s = ExpansionStringUnescape.unescapeJava(s);
            }
            catch(Exception e) {
            	System.err.println("WE IGNORED THIS ERROR:");
            	
            	e.printStackTrace();
            	System.err.println("String is: " + s);
            }

            // remove brackets (but keep content)
            s = s.replaceAll("[\\(\\)]", " ");

            WhitespaceTokenizer wst = new WhitespaceTokenizer();
                    wst.setReader(new StringReader(s));
            // tokenise
            TokenStream stream = fact.create(wst);
            stream.reset();

            if (useStemmer) {
                // use stemmer if requested
                stream = new PorterStemFilter(stream);
            }

            // lower case all tokens
            stream = new LowerCaseFilter(stream);

            if(removeStopWords) {
            // remove stop words
            stream = new StopFilter( stream,
                    ((org.apache.lucene.analysis.StopwordAnalyzerBase) analyzer).getStopwordSet());
            }

            // enumerate tokens
            while (stream.incrementToken()) {
                result.add(stream.getAttribute(CharTermAttribute.class)
                        .toString());
            }
            stream.close();
        } catch (IOException e) {
            // not thrown b/c we're using a string reader...
        }

        analyzer.close();

        return result;
    }
}
