package de.mannheim.uni;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class FootInchParser {

	
	public static void main (String[] args) {
		
		
		Map<String, Double> testMap = new LinkedHashMap<>();
		

		testMap.put("6/2", 1.8796);
		testMap.put("6-2", 1.8796);
		testMap.put("6-0", 1.8288);
		
		testMap.put("6'2", 1.8796);
		testMap.put("10-2", 3.0988);
		testMap.put("8-11", 2.7178);
		testMap.put("7.10", 2.3876);
		testMap.put("6.9", 2.0574);
		testMap.put("6.09", 2.0574);
		testMap.put("6.90", null);
		testMap.put("6-12", null);
	 
		for(Entry<String, Double> entry : testMap.entrySet()) {
			
			Optional<Double> parsed = parseMetre(entry.getKey());
			if(entry.getValue() == null) {
				if(!parsed.isPresent()) {
					System.out.println("Correct: " + entry.getKey() + ", NOT PARSED");
				}
				else {
					System.out.println("Wrong:   " + entry.getKey() + ", " + parsed.get());
				}
			}
			else {
				if(parsed.isPresent()) {
					if(parsed.get() - entry.getValue() < 0.001) {
						System.out.println("Correct: " + entry.getKey() + ", " + parsed.get());
					}
					else {
						System.out.println("Wrong:   " + entry.getKey() + ", returned : " + parsed.get() + ", expected: " + entry.getValue());
					}
				}
				else {
					System.out.println("Wrong:   " + entry.getKey() + ", NOT PARSED");
				}
			}
		}
		
		
		
		
	}
	
	
	
	private static final Pattern pattern = Pattern.compile("((?:[0-9])*[0-9])[-'.-/]([0-9]|0[0-9]|1[01])(?![0-9])");
	private static final double footToMeter = 0.3048;
	private static final double inchToMeter = 0.0254;
	
	public static Optional<Double> parseMetre(String parse) {
		Matcher matches = pattern.matcher(parse);
		if(matches.find()) {
			int feet = Integer.parseInt(matches.group(1));
			int inches = Integer.parseInt(matches.group(2));
			double metres = feet * footToMeter + inches * inchToMeter;
			return Optional.of(metres);
		}
		else {
			return Optional.empty();
		}
		
	}
}
