package de.mannheim.uni.tableprocessor.model;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;

import de.mannheim.uni.util.Variables;
import de.mannheim.uni.utils.data.Pair;

import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.google.common.util.concurrent.SettableFuture;
import java.util.ArrayList;
import java.util.List;

/**
 * Represents a column of a Table. Holds the references to all the values of the column.
 * 
 * No longer compatible to the original SearchJoins Project, but now uses less than one third of the memory previously needed.
 * @author Oliver
 *
 */
public class TableColumn
        implements java.io.Serializable, Comparable<TableColumn> {

    /**
     *
     */
    private static final long serialVersionUID = 7927975071509939076L;

    /**
     * @return the columnStatistic
     */
    public Statistic getColumnStatistic() {
        return columnStatistic;
    }

    /**
     * @param columnStatistic the columnStatistic to set
     */
    public void setColumnStatistic(Statistic columnStatistic) {
        this.columnStatistic = columnStatistic;
    }

    /**
     * @return the headerList
     */
    public List getHeaderList() {
        return headerList;
    }

    /**
     * @param headerList the headerList to set
     */
    public void setHeaderList(List headerList) {
        this.headerList = headerList;
    }

    public Object getHashCodeObject() {
    	return hashCodeObject;
    	
    }
    public static enum ColumnDataType {
        numeric, string, coordinate, date, link, bool, unknown, unit, list, reference
    };

    /*
     * Column meta data
     */
    private String header;
    private ColumnDataType dataType;
    private String dataSource;
    private Table table;
    private String URI;
    private boolean isKey;
    private int numRows;
    private Statistic columnStatistic;
    private transient List<String> headerList = new ArrayList<>();
    
    /*
     * Column values
     */
    Map<Integer, Object> values;
	private TableColumnPair hashCodeObject;
    
    
    /**
     * @return the table that contains this column
     */
    public Table getTable() {
        return table;
    }

    public String getDataSource() {
        return dataSource;
    }

    public void setDataSource(String dataSource) {
        this.dataSource = dataSource;
    }

    public boolean isKey() {
        return isKey;
    }

    public void setKey(boolean isKey) {
        this.isKey = isKey;
    }

    public Object getHeader() {
        return header;
    }

    public void setHeader(Object header) {
        this.header = header.toString();
    }

    public String getURI() {
        return URI;
    }
    
    protected void setURI(String URI) {
        this.URI = URI;
    }
    
    public ColumnDataType getDataType() {
        return dataType;
    }

    public void setDataType(ColumnDataType dataType) {
        this.dataType = dataType;
    }

    public Map<Integer, Object> getValues() {
        return values;
    }
    
    public int getNumRows() {
        return numRows;
    }
    public void setNumRows(int numRows) {
        this.numRows = numRows;
    }

    public TableColumn(Table table, int columnIdx) {
    	this.hashCodeObject = new TableColumnPair(table.getIdObject(), columnIdx);
        values = new TreeMap<Integer, Object>();
        header = "";
        dataType = ColumnDataType.unknown;
        this.table = table; 
        this.columnStatistic = new Statistic();
    }

    /**
     * The uniqueness rank of a column is the fraction of unique values minus
     * the fraction of null values.
     *
     * @return
     */
    public double getColumnUniqnessRank() {
        //TODO check if this calculation is still correct!
        int uniqueValues = getNumberOfUniqueValues();
        int totalValues = getValues().size();
        
//        double rank = (double) ((double) (uniqueValues) / (double) totalValues);
//        
//        int numberOfNullValues = totalValues - numRows;
//        
//        rank = rank - ((double) ((double) numberOfNullValues / (double) totalValues));
//
//        return rank;
        
        //the whole time it was like this:
//        double rank1 = (double)uniqueValues / (double)numRows;
//        int numNulls = numRows - totalValues;
//        double rank2 = (double)numNulls / (double)numRows;
        
        double rank1 = (double)uniqueValues / (double)getTable().getTotalNumOfRows();
        int numNulls = getTable().getTotalNumOfRows() - totalValues;
        double rank2 = (double)numNulls / (double)getTable().getTotalNumOfRows();        
        return rank1 - rank2;
    }

    public int getNumberOfUniqueValues() {
        
        HashSet<Object> values = new HashSet<>();
        
        for(Object o : getValues().values()) {
            values.add(o);
        }
        
        return values.size();
    }
    
    @Override
    public String toString() {
        return header.toString();
    }

    @Override
    public int compareTo(TableColumn o) {
//        int comp = getTable().getHeader().compareTo(o.getTable().getHeader());
//        
//        if(comp==0) {
//            comp = Integer.compare(getTable().getColumns().indexOf(this), o.getTable().getColumns().indexOf(o));
//            
//            if(comp==0 && getURI()!=null && o.getURI()!=null) {
//                comp = getURI().compareTo(o.getURI());
//            }
//        }
//        
//        return comp;
        String me = getTable().getHeader() + getTable().getColumns().indexOf(this) + getURI()+ "";
        String other = o.getTable().getHeader() + o.getTable().getColumns().indexOf(o) + o.getURI()+  "";
        return me.compareTo(other);
    }

	public void setValues(Map<Integer, Object> values) {
		this.values = values;
	}
    
	
    
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((hashCodeObject == null) ? 0 : hashCodeObject.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		TableColumn other = (TableColumn) obj;
		if (hashCodeObject == null) {
			if (other.hashCodeObject != null)
				return false;
		} else if (!hashCodeObject.equals(other.hashCodeObject))
			return false;
		return true;
	}



	public static class TableColumnPair {
		
		private final Object tableHashObject;
		private final int columnNumber;
		public TableColumnPair(Object tableHashObject, int columnNumber) {
			super();
			this.tableHashObject = tableHashObject;
			this.columnNumber = columnNumber;
		}
		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + columnNumber;
			result = prime * result + ((tableHashObject == null) ? 0 : tableHashObject.hashCode());
			return result;
		}
		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			TableColumnPair other = (TableColumnPair) obj;
			if (columnNumber != other.columnNumber)
				return false;
			if (tableHashObject == null) {
				if (other.tableHashObject != null)
					return false;
			} else if (!tableHashObject.equals(other.tableHashObject))
				return false;
			return true;
		}
	 
		
		
		
	}
    
//  too slow...
//    @Override
//    public int hashCode() {
//        return new HashCodeBuilder().append(dataSource.hashCode()).append(header.hashCode()).append(table.hashCode()).append(table.getColumns().indexOf(this)).toHashCode();
//    }
//    
//    @Override
//    public boolean equals(Object obj) {
//        return hashCode()==obj.hashCode();
//    }
}
