package de.mannheim.uni.expansion.tablefusion.adapters.columnmatches;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.Map;
import java.util.Map.Entry;
import java.util.zip.GZIPOutputStream;

import org.apache.commons.compress.archivers.tar.TarArchiveEntry;
import org.apache.commons.compress.archivers.tar.TarArchiveOutputStream;
import org.apache.commons.io.FileUtils;

import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.CellIdx;

public class CellMatchesWriter {


	public void write(String fileName, Map<AnnotatedTableId, Map<CellIdx, InstanceId>> matches) throws FileNotFoundException, IOException {

		TarArchiveOutputStream tarFile = new TarArchiveOutputStream(new GZIPOutputStream(new FileOutputStream(fileName)));

		for (Entry<AnnotatedTableId, Map<CellIdx, InstanceId>> e : matches.entrySet()) {

			byte[] bytes = getCSVFile(e.getValue());

			String name = e.getKey().getTableName();
			if (name.getBytes().length > 90) {
				System.out.println("Skipped table because name too long: " + name);
				continue;
			}

			TarArchiveEntry entry = new TarArchiveEntry(name + ".csv");
			entry.setSize(bytes.length);

			tarFile.putArchiveEntry(entry);

			tarFile.write(bytes);

			tarFile.closeArchiveEntry();

		}

		tarFile.finish();

		tarFile.close();

	}

	public void writeSingle(String fileName, String tableName, Map<CellIdx, InstanceId> matches) throws FileNotFoundException, IOException {

		byte[] bytes = getCSVFile(matches);

		String name = tableName;
		if (name.getBytes().length > 90) {
			System.out.println("Skipped table because name too long: " + name);
		}

		FileUtils.writeByteArrayToFile(new File(fileName), bytes);

	}

	public byte[] getCSVFile(Map<CellIdx, InstanceId> matches) throws IOException {
		byte[] bytes = null;
		ByteArrayOutputStream baos = new ByteArrayOutputStream();

		CSVWriter writer = new CSVWriter(new OutputStreamWriter(baos), ';', '"');

		for (Entry<CellIdx, InstanceId> e : matches.entrySet()) {
			String[] line = new String[3];

			CellIdx vid = e.getKey();
			line[0] = String.valueOf(vid.getColumnIdx().getColumnIdx());
			line[1] = String.valueOf(vid.getRowIdx().getRowIdx());
			line[2] = e.getValue().getInstanceId();

			writer.writeNext(line);

		}

		writer.close();

		bytes = baos.toByteArray();

		return bytes;
	}
}
