package de.mannheim.uni.expansion.tablefusion.adapters.columnmatches;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.CellIdx;
import de.mannheim.uni.expansion.data.tables.reader.TableConverter;
import de.mannheim.uni.expansion.tablefusion.core.matching.impl.reference.ColumnMatcher;
import de.mannheim.uni.expansion.tablefusion.core.matching.impl.reference.MatchedColumnData;
import de.mannheim.uni.matching.dbpedia.model.adapters.TableToRowHierarchyAdapter;
import de.mannheim.uni.tableprocessor.model.Table;
import de.mannheim.uni.tableprocessor.model.TableColumn.ColumnDataType;

public class ColumnMatchesForTable {

	private final OntologyRepository ontologyRepository;
	private final ColumnMatcher columnMatcher;
	private final TableConverter tableConverter;

	@Inject
	public ColumnMatchesForTable(OntologyRepository ontologyRepository, ColumnMatcher columnMatcher, TableConverter tableConverter) {

		this.ontologyRepository = ontologyRepository;
		this.columnMatcher = columnMatcher;
		this.tableConverter = tableConverter;

	}

	public Map<CellIdx, InstanceId> run(AnnotatedTablesReaderData data) throws IOException {
		TableToRowHierarchyAdapter rowAdapter = new TableToRowHierarchyAdapter();

		Table t = tableConverter.getTable(data);
		AnnotatedTableMetaData metaData = data.getMetaData();

		rowAdapter.getParts(t);

		Map<Integer, ColumnDataType> allPropertyColumns = AnnotatedTable.getDBPediaPropertyColumns(t, metaData);

		Map<Integer, InstanceId> rowInstances = metaData.getInstances();
		// int numHeaderRow = metaData.getNumHeaderRows();

		Map<CellIdx, InstanceId> matchedCells = new HashMap<CellIdx, InstanceId>();

		for (Entry<Integer, ColumnDataType> column : allPropertyColumns.entrySet()) {
			PropertyId propertyId = metaData.getProperties().get(column.getKey());
			if (column.getValue() == ColumnDataType.string && ontologyRepository.isReferenceField(propertyId)) {

				MatchedColumnData mcd = columnMatcher.getMatchedColumnData(data, column.getKey());

				for (Entry<Integer, InstanceId> match : mcd.getRowInstanceMatches().entrySet()) {
					int instanceIndex = match.getKey();
					InstanceId instanceId = rowInstances.get(instanceIndex);
					CellIdx cellId = new CellIdx(column.getKey(), instanceIndex);
					matchedCells.put(cellId, match.getValue());

				}

			}

		}

		return matchedCells;
	}

}
