package de.mannheim.uni.expansion.tablefusion.adapters.columnmatches;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.CellIdx;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.utils.CHMHelper;

public class ColumnMatchesRepositoryImpl implements ColumnMatchesRepository {

	private final Map<AnnotatedTableId, Map<CellIdx, InstanceId>> matches;
	private final Map<AnnotatedTableId, Map<RowIdx, Set<ColumnIdx>>> columnsPerRowIdx;
	private final Map<AnnotatedTableId, Map<ColumnIdx, Set<RowIdx>>> rowsPerColumnIdx;

	public ColumnMatchesRepositoryImpl(Map<AnnotatedTableId, Map<CellIdx, InstanceId>> matches) {
		this.matches = matches;

		this.columnsPerRowIdx = new HashMap<AnnotatedTableId, Map<RowIdx, Set<ColumnIdx>>>();
		this.rowsPerColumnIdx = new HashMap<AnnotatedTableId, Map<ColumnIdx, Set<RowIdx>>>();

		for (Entry<AnnotatedTableId, Map<CellIdx, InstanceId>> e : matches.entrySet()) {
			Map<RowIdx, Set<ColumnIdx>> columnsPerRowMap = new HashMap<>();
			Map<ColumnIdx, Set<RowIdx>> rowsPerColumnMap = new HashMap<>();

			for (CellIdx vid : e.getValue().keySet()) {
				
				Set<ColumnIdx> porperties = CHMHelper.getFromMap(columnsPerRowMap, vid.getRowIdx(), () -> new HashSet<>());

				porperties.add(vid.getColumnIdx());
				

				Set<RowIdx> instances = CHMHelper.getFromMap(rowsPerColumnMap, vid.getColumnIdx(), () -> new HashSet<>());

				instances.add(vid.getRowIdx());

			

			}

			columnsPerRowIdx.put(e.getKey(), columnsPerRowMap);
			rowsPerColumnIdx.put(e.getKey(), rowsPerColumnMap);
		}

	}

	@Override
	public Map<CellIdx, InstanceId> getMatches(AnnotatedTableId id) {
		return Collections.unmodifiableMap(matches.get(id));
	}

	@Override
	public InstanceId getMatch(CellId id) {
		Map<CellIdx, InstanceId> table = matches.get(id.getTableId());
		if (table != null)
			return table.get(new CellIdx(id.getColumnIdx(), id.getRowIdx()));
		return null;
	}


	@Override
	public Set<RowIdx> getInstancesWithMatch(AnnotatedTableId id, ColumnIdx columnIdx) {
		if (rowsPerColumnIdx.containsKey(id)) {
			Set<RowIdx> set = rowsPerColumnIdx.get(id).get(columnIdx);
			if (set != null) {
				return Collections.unmodifiableSet(set);
			}

		}

		return new HashSet<RowIdx>();
	}

	@Override
	public Set<ColumnIdx> getPropertiesWithMatch(AnnotatedTableId id, RowIdx rowIdx) {
		if (columnsPerRowIdx.containsKey(id)) {
			Set<ColumnIdx> set = columnsPerRowIdx.get(id).get(rowIdx);
			if (set != null) {
				return Collections.unmodifiableSet(set);
			}
		}

		return new HashSet<ColumnIdx>();
	}

}
