package de.mannheim.uni.expansion.tablefusion.adapters.columnmatches;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;
import javax.inject.Named;

import de.mannheim.uni.expansion.columnmatching.CMProperties;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.CellIdx;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTablesKryoReader;
import de.mannheim.uni.expansion.data.tables.reader.TableConverter;
import de.mannheim.uni.expansion.tablefusion.core.matching.impl.reference.ColumnMatcher;
import de.mannheim.uni.expansion.tablefusion.core.matching.impl.reference.MatchedColumnData;
import de.mannheim.uni.matching.dbpedia.model.adapters.TableToRowHierarchyAdapter;
import de.mannheim.uni.tableprocessor.model.Table;
import de.mannheim.uni.tableprocessor.model.TableColumn.ColumnDataType;

public class ColumnMatchesWriter {

	private final OntologyRepository ontologyRepository;
	private final ColumnMatcher columnMatcher;
	private final String columnMappingsFolder;

	private final AnnotatedTablesKryoReader kryoReader;

	private final CellMatchesWriter matchesWriter = new CellMatchesWriter();

	private final TableConverter tableConverter;

	@Inject
	public ColumnMatchesWriter(OntologyRepository ontologyRepository, ColumnMatcher columnMatcher,
			@Named(CMProperties.COLUMN_MATCHES_FOLDER) String columnMappingsFolder, AnnotatedTablesKryoReader kryoReader, TableConverter tableConverter) {

		this.ontologyRepository = ontologyRepository;
		this.columnMatcher = columnMatcher;
		this.columnMappingsFolder = columnMappingsFolder;

		this.kryoReader = kryoReader;
		this.tableConverter = tableConverter;

	}

	public void run() throws IOException {
		
		PropertyId recordLabelProperty = new PropertyId("http://dbpedia.org/ontology/recordLabel");
		
		TableToRowHierarchyAdapter rowAdapter = new TableToRowHierarchyAdapter();
		ConcurrentHashMap<AnnotatedTableId, Map<CellIdx, InstanceId>> cellMatches = new ConcurrentHashMap<>();

		kryoReader.runForEachTable((data) -> {
			// if
			// (!data.getId().getTableName().equals("1438042988962.66_20150728002308-00090-ip-10-236-191-2_433710852_4"))
			// {
			// return;
			// }
			Table t = tableConverter.getTable(data);
			AnnotatedTableMetaData metaData = data.getMetaData();


			rowAdapter.getParts(t);

			Map<Integer, ColumnDataType> allPropertyColumns = AnnotatedTable.getDBPediaPropertyColumns(t, metaData);

			Map<Integer, InstanceId> rowInstances = metaData.getInstances();
			// int numHeaderRow = metaData.getNumHeaderRows();

			Map<CellIdx, InstanceId> matchedCells = new HashMap<CellIdx, InstanceId>();

			for (Entry<Integer, ColumnDataType> column : allPropertyColumns.entrySet()) {
				PropertyId propertyId = metaData.getProperties().get(column.getKey());
				if(propertyId.equals(recordLabelProperty)) {
					int sdf = 123;
				}
				if (column.getValue() == ColumnDataType.string && ontologyRepository.isReferenceField(propertyId)) {

					MatchedColumnData mcd = columnMatcher.getMatchedColumnData(data, column.getKey());

					if(propertyId.equals(recordLabelProperty)) {
						int sdf = 123;
					}
					
					for (Entry<Integer, InstanceId> match : mcd.getRowInstanceMatches().entrySet()) {
						int instanceIndex = match.getKey();
						InstanceId instanceId = rowInstances.get(instanceIndex);
						CellIdx cellId = new CellIdx(column.getKey(), instanceIndex);
						matchedCells.put(cellId, match.getValue());

					}

				}

			}

			if (matchedCells.size() > 0) {
				cellMatches.put(data.getId(), matchedCells);
			}
		} , 1, true);

		try {
			matchesWriter.write(this.columnMappingsFolder + "/cell_matches.tar.gz", cellMatches);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}


}
