package de.mannheim.uni.expansion.tablefusion.core.matching.impl.reference;

import javax.inject.Inject;
import javax.inject.Named;

import de.mannheim.uni.expansion.columnmatching.CMProperties;
import de.mannheim.uni.matching.dbpedia.algorithm.CandidateSelectionComponent;
import de.mannheim.uni.matching.dbpedia.algorithm.KeyIndex;
import de.mannheim.uni.matching.dbpedia.algorithm.Matchers;
import de.mannheim.uni.matching.dbpedia.logging.MatchingLogger;
import de.mannheim.uni.matching.dbpedia.model.EvaluationParameters;
import de.mannheim.uni.matching.dbpedia.model.GoldStandard;
import de.mannheim.uni.matching.dbpedia.model.MatchingData;
import de.mannheim.uni.matching.dbpedia.model.MatchingParameters;
import de.mannheim.uni.matching.dbpedia.model.Similarities;
import de.mannheim.uni.matching.process.Configuration;
import de.mannheim.uni.matching.process.Parameter;
import de.mannheim.uni.utils.timer.Timer;

public class CandidateSelectionRunner {

	private final MatchingParameters matchingParameters;
	private final KeyIndex keyIndex;

	private final Timer parentTimer;

	private final Configuration config;
	private String emptyFile;

	@Inject
	public CandidateSelectionRunner(MatchingParameters matchingParameters, KeyIndex keyIndex, Timer parentTimer, Configuration config,
			@Named(CMProperties.EMPTY_FILE) String emptyFile) {
		this.matchingParameters = matchingParameters;
		this.keyIndex = keyIndex;
		this.parentTimer = parentTimer;
		this.config = config;
		this.emptyFile = emptyFile;
	}

	public static final Parameter PAR_WEBTABLE = new Parameter("Process.WebTable");
	public static final Parameter PAR_RUN = new Parameter("Process.Run");
	public static final Parameter PAR_MAPPED_RATIO_FILTER = new Parameter("Process.MappedRatioFilter", 0.5);

	public MatchedColumnData run(MatchingData data, String webTableName) {

		keyIndex.setLuceneBlocking(null);
		// Configuration config = new Configuration(new HashMap<Parameter,
		// Object>(), new HashMap<String, Object>());
		// config.getConfig().put(PAR_WEBTABLE, webTableName);

		// config.getConfig().put(PAR_WEBTABLE, webTableName);

		CandidateSelectionComponent candidateSelection = new CandidateSelectionComponent();
		candidateSelection.setData(data);
		candidateSelection.setKeyIndex(keyIndex);
		candidateSelection.setMatchingParameters(matchingParameters);

		Similarities similarities = new Similarities();
		Timer timer = Timer.getNamed("matchWebTable", parentTimer);

		MatchingLogger logger = new MatchingLogger();
		logger.prepareLog();

		EvaluationParameters eval = new EvaluationParameters();
		eval.setClassGoldStandardLocation(emptyFile);
		eval.setClassHierarchyLocation(emptyFile);
		eval.setCorrectedInstancesLocation(emptyFile);
		eval.setEquivalentPropertiesLocation(emptyFile);
		eval.setInstanceGoldStandardLocation(emptyFile);
		eval.setPropertyGoldStandardLocation(emptyFile);
		eval.setPropertyRangeGoldstandardLocation(emptyFile);
		eval.setPropertyRangesLocation(emptyFile);

		GoldStandard gs = new GoldStandard();
		// gs.initialise(webTableName, data.getWebtable(), eval);

		candidateSelection.setMatchers(new Matchers(similarities, matchingParameters, timer, gs, logger));

		candidateSelection.getMatchers().getCandidateSelectionMatcher().setKeyIndex(keyIndex);
		candidateSelection.getMatchers().getCandidateRefinementMatcher().setKeyIndex(keyIndex);

		candidateSelection.run(config);

		return new MatchedColumnData(candidateSelection.getInitialInstanceMappings());
	}
}
