package de.mannheim.uni.expansion.tablefusion.core.matching.impl.reference;

import java.util.List;

import javax.inject.Inject;

import com.fasterxml.uuid.Generators;

import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.reader.TableConverter;
import de.mannheim.uni.matching.dbpedia.model.MatchingData;
import de.mannheim.uni.matching.dbpedia.model.TableRowCache;
import de.mannheim.uni.tableprocessor.model.Table;
import de.mannheim.uni.tableprocessor.model.TableColumn;

public class ColumnMatcher {

	private final MatchingDataBuilder matchingDataBuilder;
	private final CandidateSelectionRunner candidateSelectionRunner;
	private final TableConverter tableConverter;

	@Inject
	public ColumnMatcher(MatchingDataBuilder matchingDataBuilder, CandidateSelectionRunner candidateSelectionRunner, TableConverter tableConverter) {
		this.matchingDataBuilder = matchingDataBuilder;
		this.candidateSelectionRunner = candidateSelectionRunner;
		this.tableConverter = tableConverter;
	}

//	public Map<TripleId, MatchedValues> matchColumns(AnnotatedTablesReaderData readerData, Set<Integer> columns) {
//
//		int oldKey = readerData.getTableData().getKeyIndex();
//		readerData.getTableData().getColumns().get(oldKey).setKey(false);
//
//		Map<Integer, MatchedColumnData> map = new HashMap<Integer, MatchedColumnData>();
//
//		for (int i : columns) {
//
//			readerData.getTableData().getColumns().get(i).setKey(true);
//
//			MatchingData data = matchingDataBuilder.get(readerData.getTableData(), readerData.getMetaData().getProperties().get(i), i);
//
//			map.put(i, candidateSelectionRunner.run(data, readerData.getId().getTableName()));
//
//			readerData.getTableData().getColumns().get(i).setKey(false);
//		}
//
//		readerData.getTableData().getColumns().get(oldKey).setKey(true);
//
//		AnnotatedTableMetaData metaData = readerData.getMetaData();
//
//		Map<TripleId, MatchedValues> groupedMap = new HashMap<TripleId, MatchedValues>();
//
//		for (Entry<Integer, MatchedColumnData> columnMatches : map.entrySet()) {
//
//			ClassId dBpediaClass = metaData.getDbPediaClass();
//			PropertyId dBpediaProperty = metaData.getProperties().get(columnMatches.getKey());
//			Map<Integer, InstanceId> rowInstances = metaData.getInstances();
//			// int numHeaderRow = metaData.getNumHeaderRows();
//
//			for (Entry<Integer, InstanceId> match : columnMatches.getValue().getRowInstanceMatches().entrySet()) {
//				int instanceIndex = match.getKey();
//				if (rowInstances.get(instanceIndex) != null) {
//					ColumnDataType cdt = readerData.getTableData().getColumns().get(metaData.getCol(dBpediaProperty)).getDataType();
//					TripleId mid = new TripleId(dBpediaClass, dBpediaProperty, rowInstances.get(instanceIndex));
//					MatchedValues mv = new MatchedValues(readerData.getId(), match.getValue(), ColumnDataType.reference);
//
//					groupedMap.put(mid, mv);
//				}
//
//			}
//
//		}
//
//		return groupedMap;
//
//	}
//
//	public Map<TripleId, MatchedValues> matchColumns(AnnotatedTablesReaderData readerData, int columnIndex) {
//
//		MatchedColumnData mcd = getMatchedColumnData(readerData, columnIndex);
//
//		AnnotatedTableMetaData metaData = readerData.getMetaData();
//
//		Map<TripleId, MatchedValues> groupedMap = new HashMap<TripleId, MatchedValues>();
//
//		ClassId dBpediaClass = metaData.getDbPediaClass();
//		PropertyId dBpediaProperty = metaData.getProperties().get(columnIndex);
//		Map<Integer, InstanceId> rowInstances = metaData.getInstances();
//		// int numHeaderRow = metaData.getNumHeaderRows();
//
//		for (Entry<Integer, InstanceId> match : mcd.getRowInstanceMatches().entrySet()) {
//			int instanceIndex = match.getKey();
//			if (rowInstances.get(instanceIndex) != null) {
//				ColumnDataType cdt = readerData.getTableData().getColumns().get(metaData.getCol(dBpediaProperty)).getDataType();
//
//				TripleId mid = new TripleId(dBpediaClass, dBpediaProperty, rowInstances.get(instanceIndex));
//				MatchedValues mv = new MatchedValues(readerData.getId(), match.getValue(), ColumnDataType.reference);
//
//				groupedMap.put(mid, mv);
//			}
//
//		}
// 
//		return groupedMap;
//
//	}

	public MatchedColumnData getMatchedColumnData(AnnotatedTablesReaderData readerData, int columnIndex) {
		Table table = tableConverter.getTable(readerData);
		table.setIdObject(Generators.timeBasedGenerator().generate());

		int oldKey = table.getKeyIndex();

		List<TableColumn> columns = table.getColumns();

		columns.get(oldKey).setKey(false);
		columns.get(columnIndex).setKey(true);
		table.setKey(columns.get(columnIndex), columnIndex);
		TableRowCache.get().removeTable(table);
		MatchingData data = matchingDataBuilder.getRanged(table, readerData.getMetaData().getDbPediaClass(),
				readerData.getMetaData().getProperties().get(columnIndex));

		MatchedColumnData mcd = candidateSelectionRunner.run(data, readerData.getId().getTableName());

		// table.setKeyIndex(oldKey);
		// columns.get(columnIndex).setKey(false);
		// columns.get(oldKey).setKey(true);

		return mcd;
	}
}
