package de.mannheim.uni.expansion.data.dbpedia;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.PropertyIndexComposite;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueContainer;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;

public class PropertiesDBpediaTableFilter implements AnnotatedTableFilter, KBValueCreator {

	private final Set<PropertyId> filteredProperties = new HashSet<>();

	public PropertiesDBpediaTableFilter() {
		filteredProperties.add(new PropertyId("http://dbpedia.org/ontology/wikiPageID"));
		filteredProperties.add(new PropertyId("http://dbpedia.org/ontology/wikiPageRevisionID"));
		filteredProperties.add(new PropertyId("http://www.w3.org/2002/07/owl#sameAs"));

		filteredProperties.add(new PropertyId("http://purl.org/dc/elements/1.1/description"));
		filteredProperties.add(new PropertyId("http://dbpedia.org/ontology/thumbnail"));
		filteredProperties.add(new PropertyId("http://dbpedia.org/ontology/type"));
		filteredProperties.add(new PropertyId("http://www.w3.org/1999/02/22-rdf-syntax-ns#type"));
		filteredProperties.add(new PropertyId("http://www.w3.org/2000/01/rdf-schema#comment"));

		filteredProperties.add(new PropertyId("http://dbpedia.org/ontology/abstract"));

		filteredProperties.add(new PropertyId("http://dbpedia.org/ontology/abstract"));

		filteredProperties.add(new PropertyId("URI"));
		filteredProperties.add(new PropertyId("http://purl.org/dc/elements/1.1/description"));
		filteredProperties.add(new PropertyId("http://www.w3.org/1999/02/22-rdf-syntax-ns#type"));
		filteredProperties.add(new PropertyId("http://dublincore.org/2012/06/14/dcterms#description"));
		filteredProperties.add(new PropertyId("https://www.w3.org/1999/02/22-rdf-syntax-ns#type"));
		filteredProperties.add(new PropertyId("http://dbpedia.org/ontology/thumbnail"));
		filteredProperties.add(new PropertyId("http://dbpedia.org/ontology/wikiPageID"));
		filteredProperties.add(new PropertyId("http://www.w3.org/2000/01/rdf-schema#comment"));
		filteredProperties.add(new PropertyId("http://www.w3.org/2002/07/owl#sameAs"));
		filteredProperties.add(new PropertyId("http://dbpedia.org/ontology/wikiPageRevisionID"));
		filteredProperties.add(new PropertyId("http://www.w3.org/2000/01/rdf-schema#label"));

	}

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {

		AnnotatedTableMetaData md = data.getMetaData();

		HashMap<PropertyIndexComposite, Double> newScoresMap = new HashMap<>(md.getAllPropertyScores());
		Map<Integer, List<PropertyId>> newProperties = new HashMap<>();

		for (Entry<Integer, List<PropertyId>> e : md.getAllProperties().entrySet()) {

			List<PropertyId> newList = new ArrayList<>();

			for (PropertyId propertyId : e.getValue()) {

				if (!filteredProperties.contains(propertyId)) {
					newList.add(propertyId);
				} else {

					newScoresMap.remove(new PropertyIndexComposite(e.getKey(), propertyId));
				}

			}

			newProperties.put(e.getKey(), newList);

		}

		md = new AnnotatedTableMetaData(md.getSource(), md.getDbPediaClass(), md.getClassConf(), md.getNumHeaderRows(), md.getNumberOfRows(), newProperties,
				newScoresMap, md.getAllInstances(), md.getInstancesScores(), md.getKeyColumn(), md.getYearsBeforeText(), md.getYearsAfterText(),
				md.getYearsInContextBefore(), md.getYearsInTableTitle(), md.getYearsInPageTitle(), md.getPropertiesYears(), md.getPageTitle(),
				md.getTableTitle(), md.getTableContextTimeStampBeforeTable(), md.getTextBeforeTable(), md.getTextAfterTable());

		return Optional.of(new AnnotatedTablesReaderData(data.getId(), md, data.getTableData()));

	}

	@Override
	public KBValueContainer getMatchedValues(KnowledgeBase kb) {

		Set<ValueId> removeValues = Collections.newSetFromMap(new ConcurrentHashMap<>());

		PersistentParallel6.runAndFinish(kb.getClasses(), (setId) -> {
			for (PropertyId propertyId : kb.getProperties(setId).keySet()) {
				if (filteredProperties.contains(propertyId)) {
					PersistentParallel6.runAndFinish(kb.getInstances(setId, propertyId), (instanceId) -> {
						ValueId vid = new ValueId(instanceId, propertyId);

						removeValues.add(vid);

					}, this.getClass());
				}
			}
		}, this.getClass());

		return KBValueContainer.defaultConfig().rebuilder().setRemoveValues(removeValues).build();
	}

}
