package de.mannheim.uni.expansion.data.dbpedia.adapter;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import com.google.common.collect.ImmutableBiMap;
import com.google.common.collect.ImmutableBiMap.Builder;
import com.google.common.collect.ImmutableSet;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedAbstractObjectHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.Qualifier;

public class KnowledgeBaseLoadCache {
	Map<ClassId, ClassId> classCache = new ConcurrentHashMap<ClassId, ClassId>();
	ConcurrentHashMap<Set<ClassId>, ImmutableSet<ClassId>> classSetCache = new ConcurrentHashMap<>();
	Map<InstanceId, InstanceId> instanceCache = new ConcurrentHashMap<InstanceId, InstanceId>();
	Map<PropertyId, PropertyId> propertyCache = new ConcurrentHashMap<PropertyId, PropertyId>();
	Map<Object, Object> valueCache = new ConcurrentHashMap<Object, Object>();

	public ClassId getFromCache(ClassId dBpediaClass) {
		ClassId cached = classCache.get(dBpediaClass);
		if (cached == null) {
			ClassId ret = classCache.putIfAbsent(dBpediaClass, dBpediaClass);
			if (ret == null)
				return dBpediaClass;
			else
				return ret;
		} else {
			return cached;
		}
	}

	public ImmutableSet<ClassId> getImmutableSetFromCache(Set<ClassId> setIdSet) {
		final Set<ClassId> newSet = new HashSet<>();
		for (ClassId setId : setIdSet) {
			newSet.add(getFromCache(setId));
		}

		return CHMHelper.getFromCHM(classSetCache, newSet, () -> ImmutableSet.copyOf(newSet));

	}

	// public Set<ClassId> getFromCache(Set<ClassId> setIdSet) {
	//// final Set<ClassId> newSet = new HashSet<>();
	//// for (ClassId setId : setIdSet) {
	//// newSet.add(getFromCache(setId));
	//// }
	//
	// return CHMHelper.getFromCHM(classSetCache, setIdSet, () -> setIdSet);
	//
	// }

	public PropertyId getFromCache(PropertyId propertyId) {
		PropertyId cached = propertyCache.get(propertyId);
		if (cached == null) {
			PropertyId ret = propertyCache.putIfAbsent(propertyId, propertyId);
			if (ret == null)
				return propertyId;
			else
				return ret;
		} else {
			return cached;
		}
	}

	public InstanceId getFromCache(InstanceId instanceId) {
		InstanceId cached = instanceCache.get(instanceId);
		if (cached == null) {
			InstanceId ret = instanceCache.putIfAbsent(instanceId, instanceId);
			if (ret == null)
				return instanceId;
			else
				return ret;
		} else {
			return cached;
		}
	}

	public ValueId getFromCache(ValueId valueId) {

		return new ValueId(getFromCache(valueId.getInstanceId()), getFromCache(valueId.getPropertyId()));
	}

	public ClassPropertyId getFromCache(ClassPropertyId cpid) {

		return new ClassPropertyId(getFromCache(cpid.getClassId()), getFromCache(cpid.getPropertyId()));
	}

	public Object getFromCache(Object object) {
		Object cached = valueCache.get(object);
		if (cached == null) {
			Object ret = valueCache.putIfAbsent(object, object);
			if (ret == null)
				return object;
			else
				return ret;
		} else {
			return cached;
		}
	}

	public ImmutableBiMap<PropertyId, Integer> getImmutableBiMap(Map<PropertyId, Integer> value) {
		Builder<PropertyId, Integer> builder = new ImmutableBiMap.Builder<>();
		for (Entry<PropertyId, Integer> entry : value.entrySet()) {
			builder.put(getFromCache(entry.getKey()), entry.getValue());
		}

		return builder.build();
	}

	public ImmutableBiMap<InstanceId, Integer> getImmutableBiMapInstance(Map<InstanceId, Integer> value) {
		Builder<InstanceId, Integer> builder = new ImmutableBiMap.Builder<>();
		for (Entry<InstanceId, Integer> entry : value.entrySet()) {
			builder.put(getFromCache(entry.getKey()), entry.getValue());
		}

		return builder.build();
	}

	// public Map<ValueId, QualifiedValueList>
	// getImmutableMapValueId(Map<ValueId, QualifiedValueList> values) {
	// com.google.common.collect.ImmutableMap.Builder<ValueId,
	// QualifiedValueList> builder = new ImmutableMap.Builder<>();
	// for (Entry<ValueId, QualifiedValueList> entry : values.entrySet()) {
	// builder.put(getFromCache(entry.getKey()),
	// transformQVL(entry.getValue()));
	// }
	//
	// return builder.build();
	// }
	//
	public Map<ValueId, QualifiedValueList> getImmutableMapValueId(Map<ValueId, QualifiedValueList> values) {
		ConcurrentHashMap<ValueId, QualifiedValueList> map = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(values.entrySet(), (entry) -> {
			map.put(getFromCache(entry.getKey()), transformQVL(entry.getValue()));
		}, this.getClass());

		return Collections.unmodifiableMap(map);
	}

	private QualifiedValueList transformQVL(QualifiedValueList value) {
		List<QualifiedValueHolder> claims = new ArrayList<>();

		for (int i = 0; i < value.getClaims().size(); i++) {
			QualifiedValueHolder claim = value.getClaims().get(i);
			AbstractObject ao = claim.getValue();

			if (ao instanceof InstanceId) {
				ao = getFromCache((InstanceId) ao);
			}

			if (claim.getQualifiers() != null) {
				List<Qualifier> qualifiers = new ArrayList<>();

				for (Qualifier qualifier : claim.getQualifiers()) {
					qualifiers.add(new Qualifier(getFromCache(qualifier.getQualifier()), qualifier.getQualifierValue()));
				}

				claims.add(new QualifiedAbstractObjectHolder(ao, qualifiers));
			} else {
				claims.add(new QualifiedAbstractObjectHolder(ao));
			}
		}

		return new QualifiedValueList(claims);
	}

	public Set<InstanceId> getImmutableSetWithoutCache(Set<InstanceId> set) {
		final Set<InstanceId> newSet = new HashSet<>();
		for (InstanceId instanceId : set) {
			newSet.add(getFromCache(instanceId));
		}

		return ImmutableSet.copyOf(newSet);
	}

	// public Set<ClassId> getImmutableClassIdSetWithCache(Set<ClassId> value) {
	// final Set<ClassId> newSet = new HashSet<>();
	// for (ClassId setId : setIdSet) {
	// newSet.add(getFromCache(setId));
	// }
	//
	// return CHMHelper.getFromCHM(classSetCache, newSet, () ->
	// ImmutableSet.copyOf(newSet));
	// }

}