package de.mannheim.uni.expansion.data.dbpedia.kbtables;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.mannheim.uni.expansion.data.dbpedia.kbtables.RepsitoryTableRow.TableRowId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.matching.dbpedia.model.TableRow;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.ValueConverter;

public class RepositoryCandidateMap implements Map<String, List<TableRow>> {

	private final KnowledgeBase repository;
	private final OntologyRepository ontologyRepository;
	private final ValueConverter valueConverter;

	private final Set<ClassId> allowedClasses;

	private final int size;

	public RepositoryCandidateMap(KnowledgeBase dbPediaRepository, OntologyRepository ontologyRepository, ValueConverter valueConverter) {
		this.repository = dbPediaRepository;
		this.ontologyRepository = ontologyRepository;
		this.valueConverter = valueConverter;
		allowedClasses = null;
		this.size = repository.getNumberOfInstances();
	}

	public RepositoryCandidateMap(KnowledgeBase dbPediaRepository, OntologyRepository ontologyRepository, ValueConverter valueConverter,
			Set<ClassId> allowedClasses) {
		this.repository = dbPediaRepository;
		this.ontologyRepository = ontologyRepository;
		this.valueConverter = valueConverter;
		this.allowedClasses = new HashSet<ClassId>(allowedClasses);

		int size = 0;
		for (ClassId classId : allowedClasses) {
			size += dbPediaRepository.getNumberOfInstances(classId);
		}
		this.size = size;
	}

	@Override
	public void clear() {
		throw new UnsupportedOperationException();

	}

	@Override
	public boolean containsKey(Object key) {
		InstanceId instanceId = new InstanceId((String) key, (String) key);
		Set<ClassId> classes = repository.getClasses(instanceId);

		if (allowedClasses == null) {
			if (classes.size() > 0) {
				return true;
			} else {
				return false;
			}
		} else {

			for (ClassId classId : classes) {
				if (allowedClasses.contains(classId)) {
					return true;
				}
			}

			return false;
		}

	}

	@Override
	public boolean containsValue(Object value) {
		throw new UnsupportedOperationException();
	}

	@Override
	public Set<java.util.Map.Entry<String, List<TableRow>>> entrySet() {
		throw new UnsupportedOperationException();
	}

	@Override
	public List<TableRow> get(Object key) {

		InstanceId instanceId = new InstanceId((String) key, (String) key);

		Set<ClassId> classes = repository.getClasses(instanceId);

		List<TableRow> list = new ArrayList<TableRow>();

		if (allowedClasses == null) {
			if (classes != null) {
				for (ClassId classId : classes) {

					int rowIdx = repository.getInstanceIndex(classId, instanceId) + 1;
					RepsitoryTableRow tr = RepsitoryTableRow.tableRowCacheGetter.getRaw(new TableRowId(classId, rowIdx));
					if (tr == null) {
						tr = RepsitoryTableRow.tableRowCacheGetter
								.get(new RepsitoryTableRow(repository, ontologyRepository, valueConverter, classId, instanceId, rowIdx));
					}

					list.add(tr);
				}
			}
		} else if (classes != null) {

			for (ClassId classId : classes) {
				if (allowedClasses.contains(classId)) {

					int rowIdx = repository.getInstanceIndex(classId, instanceId) + 1;
					RepsitoryTableRow tr = RepsitoryTableRow.tableRowCacheGetter.getRaw(new TableRowId(classId, rowIdx));
					if (tr == null) {
						tr = RepsitoryTableRow.tableRowCacheGetter
								.get(new RepsitoryTableRow(repository, ontologyRepository, valueConverter, classId, instanceId, rowIdx));
					}

					list.add(tr);
				}
			}

		}

		if (list.size() == 0)
			return null;
		else
			return list;

		// ClassId classId = repository.getClass(instanceId);
		//
		// if (classId == null)
		// return null;
		// else if (allowedClasses != null && !allowedClasses.contains(classId))
		// {
		// return null;
		// } else
		//
		// return new RepsitoryTableRow(repository, classId, instanceId,
		// repository.getInstanceIndex(instanceId, classId));
	}

	@Override
	public boolean isEmpty() {
		if (repository.getNumberOfInstances() > 0)
			return false;
		else
			return true;
	}

	@Override
	public Set<String> keySet() {
		throw new UnsupportedOperationException();
	}

	@Override
	public List<TableRow> put(String key, List<TableRow> value) {
		throw new UnsupportedOperationException();
	}

	@Override
	public void putAll(Map<? extends String, ? extends List<TableRow>> m) {
		throw new UnsupportedOperationException();

	}

	@Override
	public List<TableRow> remove(Object key) {
		throw new UnsupportedOperationException();
	}

	@Override
	public int size() {
		return size;
	}

	@Override
	public Collection<List<TableRow>> values() {
		throw new UnsupportedOperationException();
	}

}
