package de.mannheim.uni.expansion.data.dbpedia.kbtables;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.utils.MapSorter;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.matching.dbpedia.model.adapters.TableColumnToCellHierarchyAdapter;
import de.mannheim.uni.tableprocessor.model.TableColumn;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.ValueConverter;

public class RepositoryColSet implements Collection<TableColumn> {

	private final KnowledgeBase dbPediaRepository;
	private final OntologyRepository ontologyRepository;
	private final ValueConverter valueConverter;
	private final int size;

	private final Set<ClassId> allowedClasses;

	public RepositoryColSet(KnowledgeBase dbPediaRepository, OntologyRepository ontologyRepository, ValueConverter valueConverter) {
		this.dbPediaRepository = dbPediaRepository;
		size = dbPediaRepository.getTotalNumberOfColumns();
		this.allowedClasses = null;
		this.ontologyRepository = ontologyRepository;
		this.valueConverter = valueConverter;
	}

	public RepositoryColSet(KnowledgeBase dbPediaRepository, OntologyRepository ontologyRepository, ValueConverter valueConverter,
			Set<ClassId> allowedClasses) {
		this.dbPediaRepository = dbPediaRepository;
		this.ontologyRepository = ontologyRepository;
		this.valueConverter = valueConverter;

		int size = 0;
		for (ClassId classId : allowedClasses) {
			size += dbPediaRepository.getNumberOfColumns(classId) + 1;
		}

		this.size = size;
		this.allowedClasses = new HashSet<ClassId>(allowedClasses);
	}

	private boolean isKeyColumn(PropertyId propertyId) {
		return RepositoryColumnList.isKeyColum(propertyId);
	}

	@Override
	public boolean add(TableColumn arg0) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean addAll(Collection<? extends TableColumn> arg0) {
		throw new UnsupportedOperationException();
	}

	@Override
	public void clear() {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean contains(Object arg0) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean containsAll(Collection<?> arg0) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean isEmpty() {
		throw new UnsupportedOperationException();
	}

	private static SimpleLockableConcurrentCache<ClassId, Set<PropertyId>> cache = new SimpleLockableConcurrentCache<>();

	@Override
	public Iterator<TableColumn> iterator() {
		final Iterator<ClassId> classes;

		if (allowedClasses != null)
			classes = allowedClasses.iterator();
		else
			classes = dbPediaRepository.getClasses().iterator();

		return new Iterator<TableColumn>() {

			private ClassId currentTable;
			private boolean idColumnDone = false;
			private Iterator<PropertyId> columns;

			@Override
			public TableColumn next() {
				if ((currentTable == null || !columns.hasNext()) && classes.hasNext()) {

					currentTable = classes.next();
					// LinkedHashMap<PropertyId, Integer> sortedProperties = ;
					columns = cache.getOrCreate(currentTable, () -> {
						return MapSorter.sortHashMapByValuesAscending(dbPediaRepository.getPropertiesWithQualifiers(currentTable)).keySet();
					}).iterator();
					// columns = sortedProperties.keySet().iterator();

					idColumnDone = true;

					TableColumn tc = TableColumnToCellHierarchyAdapter.tableColumnCacheGetter.getRaw(new TableColumn.TableColumnPair(currentTable, 0));
					if (tc == null) {
						return TableColumnToCellHierarchyAdapter.tableColumnCacheGetter
								.get(new RepositoryTableColumn(dbPediaRepository, ontologyRepository, valueConverter, currentTable, null, false, true));
					} else {
						return tc;
					}

				} else if (currentTable != null && (!idColumnDone || columns.hasNext())) {
					PropertyId propertyId = columns.next();
					TableColumn tc = TableColumnToCellHierarchyAdapter.tableColumnCacheGetter
							.getRaw(new TableColumn.TableColumnPair(currentTable, dbPediaRepository.getPropertyIndex(currentTable, propertyId) + 1));
					if (tc == null) {
						return TableColumnToCellHierarchyAdapter.tableColumnCacheGetter.get(new RepositoryTableColumn(dbPediaRepository, ontologyRepository,
								valueConverter, currentTable, propertyId, isKeyColumn(propertyId), false));
					} else {
						return tc;
					}

					// return new RepositoryTableColumn(dbPediaRepository,
					// ontologyRepository, valueConverter, currentTable,
					// propertyId, isKeyColumn(propertyId),
					// false);
				}

				return null;
			}

			@Override
			public boolean hasNext() {
				if ((currentTable == null || !columns.hasNext()) && classes.hasNext()) {

					// currentTable = classes.next();
					//
					// LinkedHashMap<PropertyId, Integer> sortedProperties =
					// MapSorter
					// .sortHashMapByValuesAscending(dbPediaRepository.getPropertiesWithQualifiers(currentTable));
					// columns = sortedProperties.keySet().iterator();
					// KeyColumn !!!
					return true;
				} else if (currentTable != null && (!idColumnDone || columns.hasNext())) {
					return true;
				}

				return false;
			}
		};
	}

	@Override
	public boolean remove(Object arg0) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean removeAll(Collection<?> arg0) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean retainAll(Collection<?> arg0) {
		throw new UnsupportedOperationException();
	}

	@Override
	public int size() {
		return size;
	}

	@Override
	public Object[] toArray() {
		throw new UnsupportedOperationException();
	}

	@Override
	public <T> T[] toArray(T[] arg0) {
		throw new UnsupportedOperationException();
	}

}
