package de.mannheim.uni.expansion.data.dbpedia.kbtables;

import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.TreeSet;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.matching.dbpedia.model.adapters.TableColumnToCellHierarchyAdapter;
import de.mannheim.uni.tableprocessor.model.TableColumn;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.ValueConverter;

public class RepositoryColumnList implements List<TableColumn> {

	private final KnowledgeBase dbPediaRepository;
	private final OntologyRepository ontologyRepository;
	private final ValueConverter valueConverter;
	private final ClassId classId;

	public RepositoryColumnList(KnowledgeBase dbPediaRepository, OntologyRepository ontologyRepository, ValueConverter valueConverter, ClassId dbPediaClass) {
		this.dbPediaRepository = dbPediaRepository;
		this.classId = dbPediaClass;
		this.ontologyRepository = ontologyRepository;
		this.valueConverter = valueConverter;

	}

	@Override
	public boolean add(TableColumn e) {
		throw new UnsupportedOperationException();
	}

	@Override
	public void add(int index, TableColumn element) {
		throw new UnsupportedOperationException();

	}

	@Override
	public boolean addAll(Collection<? extends TableColumn> c) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean addAll(int index, Collection<? extends TableColumn> c) {
		throw new UnsupportedOperationException();
	}

	@Override
	public void clear() {
		throw new UnsupportedOperationException();

	}

	@Override
	public boolean contains(Object o) {
		if (o instanceof RepositoryTableColumn) {
			RepositoryTableColumn column = (RepositoryTableColumn) o;

			if (column.getDbPediaClass().equals(this.classId)) {
				return true;
			} else {
				return false;
			}
		}

		else
			return false;
	}

	@Override
	public boolean containsAll(Collection<?> c) {
		throw new UnsupportedOperationException();
	}

	private boolean isKeyColumn(PropertyId propertyId) {
		return RepositoryColumnList.isKeyColum(propertyId);
	}

	static boolean isKeyColum(PropertyId propertyId) {
		boolean keyColumn = false;
		if (propertyId.equals(PropertiesExpansionData.LABEL_PROPERTY)) {
			keyColumn = true;
		}
		return keyColumn;
	}

	@Override
	public TableColumn get(int index) {
		if (index >= size()) {
			throw new IndexOutOfBoundsException();
		} else {

			TableColumn tc = TableColumnToCellHierarchyAdapter.tableColumnCacheGetter.getRaw(new TableColumn.TableColumnPair(classId, 0));
			if (tc != null) {
				return tc;
			} else {

				if (index == 0) {

					return TableColumnToCellHierarchyAdapter.tableColumnCacheGetter
							.get(new RepositoryTableColumn(dbPediaRepository, ontologyRepository, valueConverter, classId, null, false, true));
				} else {
					PropertyId propertyId = dbPediaRepository.getPropertiesWithQualifiers(classId).inverse().get(index - 1);

					return TableColumnToCellHierarchyAdapter.tableColumnCacheGetter.get(new RepositoryTableColumn(dbPediaRepository, ontologyRepository,
							valueConverter, classId, propertyId, isKeyColumn(propertyId), false));

				}
			}
		}

	}

	@Override
	public int indexOf(Object o) {
		if (o instanceof RepositoryTableColumn) {
			RepositoryTableColumn column = (RepositoryTableColumn) o;

			if (column.getDbPediaClass().equals(this.classId)) {
				if (column.isIdColumn())
					return 0;
				return this.dbPediaRepository.getPropertiesWithQualifiers(classId).get(column.getdBpediaProperty()) + 1;
			} else {
				return -1;
			}
		}

		else
			return -1;
	}

	@Override
	public boolean isEmpty() {
		throw new UnsupportedOperationException();
	}

	@Override
	public Iterator<TableColumn> iterator() {

		final Iterator<Integer> indexIterator = new TreeSet<>(dbPediaRepository.getPropertiesWithQualifiers(classId).values()).iterator();

		return new Iterator<TableColumn>() {
			private boolean idDone = false;

			@Override
			public TableColumn next() {
				if (!idDone) {
					idDone = true;
					return new RepositoryTableColumn(dbPediaRepository, ontologyRepository, valueConverter, classId, null, false, true);

				} else {
					PropertyId propertyId = dbPediaRepository.getPropertiesWithQualifiers(classId).inverse().get(indexIterator.next());
					return new RepositoryTableColumn(dbPediaRepository, ontologyRepository, valueConverter, classId, propertyId, isKeyColumn(propertyId),
							false);
				}
			}

			@Override
			public boolean hasNext() {
				if (!idDone) {
					return true;
				} else {
					return indexIterator.hasNext();
				}
			}
		};
	}

	@Override
	public int lastIndexOf(Object o) {
		throw new UnsupportedOperationException();
	}

	@Override
	public ListIterator<TableColumn> listIterator() {
		throw new UnsupportedOperationException();
	}

	@Override
	public ListIterator<TableColumn> listIterator(int index) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean remove(Object o) {
		throw new UnsupportedOperationException();
	}

	@Override
	public TableColumn remove(int index) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean removeAll(Collection<?> c) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean retainAll(Collection<?> c) {
		throw new UnsupportedOperationException();
	}

	@Override
	public TableColumn set(int index, TableColumn element) {
		throw new UnsupportedOperationException();
	}

	@Override
	public int size() {
		return dbPediaRepository.getPropertiesWithQualifiers(classId).size() + 1;
	}

	@Override
	public List<TableColumn> subList(int fromIndex, int toIndex) {
		throw new UnsupportedOperationException();
	}

	@Override
	public Object[] toArray() {
		throw new UnsupportedOperationException();
	}

	@Override
	public <T> T[] toArray(T[] a) {
		throw new UnsupportedOperationException();
	}

}
