package de.mannheim.uni.expansion.data.dbpedia.kbtables;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.NoSuchElementException;
import java.util.Set;

import de.mannheim.uni.expansion.data.dbpedia.kbtables.RepsitoryTableRow.TableRowId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.matching.dbpedia.model.TableRow;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.ValueConverter;

public class RepositoryRowSet implements Collection<TableRow> {

	private final KnowledgeBase dbPediaRepository;
	private final OntologyRepository ontologyRepository;
	private final ValueConverter valueConverter;
	private final int size;

	private final Set<ClassId> allowedClasses;

	public RepositoryRowSet(KnowledgeBase dbPediaRepository, OntologyRepository ontologyRepository, ValueConverter valueConverter,
			Set<ClassId> allowedClasses) {
		this.dbPediaRepository = dbPediaRepository;
		this.ontologyRepository = ontologyRepository;
		this.valueConverter = valueConverter;

		int size = 0;
		for (ClassId classId : allowedClasses) {
			size += dbPediaRepository.getNumberOfInstances(classId);
		}

		this.size = size;
		this.allowedClasses = new HashSet<ClassId>(allowedClasses);
	}

	public RepositoryRowSet(KnowledgeBase dbPediaRepository, OntologyRepository ontologyRepository, ValueConverter valueConverter) {
		this.dbPediaRepository = dbPediaRepository;
		this.ontologyRepository = ontologyRepository;
		this.valueConverter = valueConverter;

		size = dbPediaRepository.getTotalNumberOfRows();
		allowedClasses = null;
	}

	@Override
	public boolean add(TableRow e) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean addAll(Collection<? extends TableRow> c) {
		throw new UnsupportedOperationException();
	}

	@Override
	public void clear() {

		throw new UnsupportedOperationException();

	}

	@Override
	public boolean contains(Object o) {

		throw new UnsupportedOperationException();
	}

	@Override
	public boolean containsAll(Collection<?> c) {

		throw new UnsupportedOperationException();
	}

	@Override
	public boolean isEmpty() {

		throw new UnsupportedOperationException();
	}

	@Override
	public Iterator<TableRow> iterator() {

		final Iterator<ClassId> classes;

		if (allowedClasses != null)
			classes = allowedClasses.iterator();
		else
			classes = dbPediaRepository.getClasses().iterator();

		return new Iterator<TableRow>() {

			private ClassId currentTable;
			private Iterator<Entry<InstanceId, Integer>> instancesPerClass;

			@Override
			public TableRow next() {
				if ((currentTable == null || !instancesPerClass.hasNext()) && classes.hasNext()) {

					currentTable = classes.next();
					instancesPerClass = dbPediaRepository.getInstances(currentTable).entrySet().iterator();

					if (instancesPerClass.hasNext()) {
						Entry<InstanceId, Integer> nextEntry = instancesPerClass.next();
						RepsitoryTableRow tr = RepsitoryTableRow.tableRowCacheGetter.getRaw(new TableRowId(currentTable, nextEntry.getValue()));
						if (tr == null)
							tr = RepsitoryTableRow.tableRowCacheGetter.get(new RepsitoryTableRow(dbPediaRepository, ontologyRepository, valueConverter,
									currentTable, nextEntry.getKey(), nextEntry.getValue()));
						return tr;
					}

				} else if (currentTable != null && instancesPerClass.hasNext()) {
					Entry<InstanceId, Integer> nextEntry = instancesPerClass.next();

					RepsitoryTableRow tr = RepsitoryTableRow.tableRowCacheGetter.getRaw(new TableRowId(currentTable, nextEntry.getValue()));
					if (tr == null)
						tr = RepsitoryTableRow.tableRowCacheGetter.get(new RepsitoryTableRow(dbPediaRepository, ontologyRepository, valueConverter,
								currentTable, nextEntry.getKey(), nextEntry.getValue()));
					return tr;
				}

				throw new NoSuchElementException();
			}

			@Override
			public boolean hasNext() {
				if ((currentTable == null || !instancesPerClass.hasNext()) && classes.hasNext()) {

					currentTable = classes.next();
					instancesPerClass = dbPediaRepository.getInstances(currentTable).entrySet().iterator();

					if (instancesPerClass.hasNext()) {
						return true;
					}

				} else if (currentTable != null && instancesPerClass.hasNext()) {
					return true;
				}

				return false;
			}
		};
	}

	@Override
	public boolean remove(Object o) {

		throw new UnsupportedOperationException();
	}

	@Override
	public boolean removeAll(Collection<?> c) {

		throw new UnsupportedOperationException();
	}

	@Override
	public boolean retainAll(Collection<?> c) {

		throw new UnsupportedOperationException();
	}

	@Override
	public int size() {
		return size;
	}

	@Override
	public Object[] toArray() {

		throw new UnsupportedOperationException();
	}

	@Override
	public <T> T[] toArray(T[] a) {

		throw new UnsupportedOperationException();
	}

}
