package de.mannheim.uni.expansion.data.dbpedia.kbtables;

import java.util.List;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.tableprocessor.model.Table;
import de.mannheim.uni.tableprocessor.model.TableColumn;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.ValueConverter;

public class RepositoryTable extends Table {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	private final KnowledgeBase repository;
	private final ClassId classId;
	private final OntologyRepository ontologyRepository;
	private final ValueConverter valueConverter;

	public RepositoryTable(KnowledgeBase dbPediaRepository, OntologyRepository ontologyRepository, ValueConverter valueConverter, ClassId classId) {
		this.setIdObject(classId);
		this.setNumHeaderRows(1);
		this.setTotalNumOfRows(dbPediaRepository.getNumberOfInstances(classId) + 1);
		this.repository = dbPediaRepository;
		this.ontologyRepository = ontologyRepository;
		this.valueConverter = valueConverter;
		this.classId = classId;
		this.setFullPath(classId.toString());
	}

	public List<TableColumn> getColumns() {
		return new RepositoryColumnList(repository, ontologyRepository, valueConverter, classId);
	}

	public TableColumn getKey() {

		PropertyId propertyId = PropertiesExpansionData.LABEL_PROPERTY;

		return new RepositoryTableColumn(repository, ontologyRepository, valueConverter, classId, propertyId, true, false);
	}

	public int getKeyIndex() {

		PropertyId propertyId = PropertiesExpansionData.LABEL_PROPERTY;

		return repository.getPropertiesWithQualifiers(classId).get(propertyId);

		// Integer keyIndex = repository.getPropertyIndex(classId,
		// repository.getClassKey(classId));
		// if (keyIndex == null) {
		// return -1;
		// } else {
		// return keyIndex;
		// }
	}

	public String getHeader() {
		return classId.getClassId();
	}

	public Object getIdObject() {
		return classId;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = super.hashCode();
		result = prime * result + ((classId == null) ? 0 : classId.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		RepositoryTable other = (RepositoryTable) obj;
		if (classId == null) {
			if (other.classId != null)
				return false;
		} else if (!classId.equals(other.classId))
			return false;
		return true;
	}

}
