package de.mannheim.uni.expansion.data.dbpedia.kbtables;

import java.util.Map;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.tableprocessor.model.TableColumn;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.ValueConverter;

public class RepositoryTableColumn extends TableColumn {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private final KnowledgeBase dbPediaRepository;
	private final ClassId dbPediaClass;
	private final PropertyId dBpediaProperty;
	private final OntologyRepository ontologyRepository;
	private final ValueConverter valueConverter;
	private final boolean keyColumn;
	private final boolean idColumn;

	public RepositoryTableColumn(KnowledgeBase dbPediaRepository, OntologyRepository ontologyRepository, ValueConverter valueConverter, ClassId dbPediaClass,
			PropertyId dbPediaProperty, boolean keyColumn, boolean idColumn) {
		super(new RepositoryTable(dbPediaRepository, ontologyRepository, valueConverter, dbPediaClass),
				idColumn ? 0 : dbPediaRepository.getPropertyIndex(dbPediaClass, dbPediaProperty) + 1);
		this.setNumRows(1);
		this.dbPediaRepository = dbPediaRepository;
		this.ontologyRepository = ontologyRepository;
		this.dbPediaClass = dbPediaClass;
		this.dBpediaProperty = dbPediaProperty;
		this.valueConverter = valueConverter;
		this.keyColumn = keyColumn;
		this.idColumn = idColumn;
		if (this.dBpediaProperty != null) {
			this.setHeader(dBpediaProperty.getFinalPart());
		} else if (idColumn) {
			this.setHeader("ID");
		}
	}

	public Map<Integer, Object> getValues() {
		return new RepositoryTableColumnValues(dbPediaRepository, dbPediaClass, dBpediaProperty, keyColumn, idColumn, valueConverter, ontologyRepository);
	}

	@Override
	public ColumnDataType getDataType() {
		// FIXME: THIS IS NOT AN APPROAPRIATE FIX
		if (isKey())
			return ColumnDataType.string;
		else if (idColumn) {
			return ColumnDataType.string;
		}
		ColumnDataType cdt = valueConverter.getDataType(ontologyRepository.getDataType(dBpediaProperty));
		if (cdt == null) {
			System.out.println("EVEN IN RepositoryTableColumn WE NEED ONTOLOGY");
		}
		return cdt;
	}

	@Override
	public boolean isKey() {
		return keyColumn;
	}

	public boolean isIdColumn() {
		return idColumn;
	}

	public KnowledgeBase getDbPediaRepository() {
		return dbPediaRepository;
	}

	public ClassId getDbPediaClass() {
		return dbPediaClass;
	}

	public PropertyId getdBpediaProperty() {
		return dBpediaProperty;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((dBpediaProperty == null) ? 0 : dBpediaProperty.hashCode());
		result = prime * result + ((dbPediaClass == null) ? 0 : dbPediaClass.hashCode());
		result = prime * result + (idColumn ? 1231 : 1237);
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		RepositoryTableColumn other = (RepositoryTableColumn) obj;
		if (dBpediaProperty == null) {
			if (other.dBpediaProperty != null)
				return false;
		} else if (!dBpediaProperty.equals(other.dBpediaProperty))
			return false;
		if (dbPediaClass == null) {
			if (other.dbPediaClass != null)
				return false;
		} else if (!dbPediaClass.equals(other.dbPediaClass))
			return false;
		if (idColumn != other.idColumn)
			return false;
		return true;
	}

	public String getURI() {
		if (idColumn)
			return "ID";
		else
			return dBpediaProperty.toCombinedString();
	}

}
