package de.mannheim.uni.expansion.data.dbpedia.kbtables;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.ValueConverter;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class RepositoryTableColumnValues implements Map<Integer, Object> {

	private final KnowledgeBase repository;
	private final ClassId classId;
	private final PropertyId propertyId;
	private final boolean isKey;
	private final boolean idColumn;
	private final ValueConverter valueConverter;
	private final OntologyRepository ontologyRepository;

	public RepositoryTableColumnValues(KnowledgeBase repository, ClassId classId, PropertyId propertyId, boolean isKey, boolean idColumn,
			ValueConverter valueConverter, OntologyRepository ontologyRepository) {
		super();
		this.repository = repository;
		this.classId = classId;
		this.propertyId = propertyId;
		// if (repository.isKey(classId, propertyId)) {
		// isKey = true;
		// } else {
		this.isKey = isKey;
		this.valueConverter = valueConverter;
		this.idColumn = idColumn;
		this.ontologyRepository = ontologyRepository;

		// }
		// TODO: FIX THIS !!
	}

	@Override
	public void clear() {
		throw new UnsupportedOperationException();

	}

	@Override
	public boolean containsKey(Object key) {
		if (key instanceof Integer && key != null) {

			int intKey = (int) key;

			if (intKey == 0)
				return true;
			else if (this.idColumn) {
				return true;
			} else {

				return repository.hasValue(new ValueId(repository.getInstance(classId, intKey - 1), propertyId));
			}
		} else {
			return false;
		}

	}

	@Override
	public boolean containsValue(Object value) {
		throw new UnsupportedOperationException();
	}

	@Override
	public Set<java.util.Map.Entry<Integer, Object>> entrySet() {
		return new Set<Map.Entry<Integer, Object>>() {

			@Override
			public <T> T[] toArray(T[] a) {
				throw new UnsupportedOperationException();
			}

			@Override
			public Object[] toArray() {
				throw new UnsupportedOperationException();
			}

			@Override
			public int size() {
				return RepositoryTableColumnValues.this.size();
			}

			@Override
			public boolean retainAll(Collection<?> c) {

				throw new UnsupportedOperationException();
			}

			@Override
			public boolean removeAll(Collection<?> c) {

				throw new UnsupportedOperationException();
			}

			@Override
			public boolean remove(Object o) {

				throw new UnsupportedOperationException();
			}

			@Override
			public Iterator<Entry<Integer, Object>> iterator() {

				final Iterator<InstanceId> rowIterator;
				if (idColumn) {

					rowIterator = repository.getInstances(classId).keySet().iterator();
				} else {
					rowIterator = repository.getInstances(classId, propertyId).iterator();

				}

				return new Iterator<Map.Entry<Integer, Object>>() {
					private boolean header = true;

					@Override
					public Entry<Integer, Object> next() {
						Integer key;
						Object value;
						if (header) {
							key = 0;
							value = propertyId.getReadableString();
							header = false;
						} else {
							InstanceId instanceId = rowIterator.next();
							if (instanceId == null)
								return null;
							key = repository.getInstanceIndex(classId, instanceId) + 1;

							value = get(key);

						}

						return new Entry<Integer, Object>() {

							@Override
							public Object setValue(Object value) {

								throw new UnsupportedOperationException();
							}

							@Override
							public Object getValue() {
								return value;
							}

							@Override
							public Integer getKey() {
								return key;
							}
						};

					}

					@Override
					public boolean hasNext() {
						if (header) {
							return true;
						} else {
							return rowIterator.hasNext();
						}
					}
				};
			}

			@Override
			public boolean isEmpty() {
				return isEmpty();
			}

			@Override
			public boolean containsAll(Collection<?> c) {

				throw new UnsupportedOperationException();
			}

			@Override
			public boolean contains(Object o) {

				throw new UnsupportedOperationException();
			}

			@Override
			public void clear() {

				throw new UnsupportedOperationException();

			}

			@Override
			public boolean addAll(Collection<? extends Entry<Integer, Object>> c) {

				throw new UnsupportedOperationException();
			}

			@Override
			public boolean add(Entry<Integer, Object> e) {

				throw new UnsupportedOperationException();
			}
		};
	}

	@Override
	public Object get(Object key) {
		if (key instanceof Integer && key != null) {
			int intKey = (int) key;
			if (intKey == 0) {
				if (idColumn)
					return "ID";
				else
					return propertyId.getReadableString();
			} else {
				InstanceId instance = repository.getInstance(classId, intKey - 1);
				if (idColumn) {
					List<AbstractObject> label = repository.getValuesAsList(instance, PropertiesExpansionData.LABEL_PROPERTY);

					if (label != null && label.size() > 0 && label.get(0) instanceof StringObject) {
						instance = new InstanceId(instance.getInstanceId(), label.get(0).readableString());
					}

					return instance.toCombinedString();
				} else if (isKey || propertyId.equals(PropertiesExpansionData.LABEL_PROPERTY)) {
					LinkedHashSet<String> linkedHashSet = new LinkedHashSet<>();

					List<AbstractObject> label = repository.getValuesAsList(instance, PropertiesExpansionData.LABEL_PROPERTY);
					if (label != null) {
						for (AbstractObject so : label) {
							if (so instanceof StringObject) {
								String string = ((StringObject) so).getString();
								if (!linkedHashSet.contains(string)) {
									linkedHashSet.add(string);
								}
							}
						}
					}

					label = repository.getValuesAsList(instance, PropertiesExpansionData.TRANSLATED_LABELS_PROPERTY);
					if (label != null) {
						for (AbstractObject so : label) {
							if (so instanceof StringObject) {
								String string = ((StringObject) so).getString();
								if (!linkedHashSet.contains(string)) {
									linkedHashSet.add(string);
								}
							}
						}
					}

					label = repository.getValuesAsList(instance, PropertiesExpansionData.TRANSLATED_ALIASES_PROPERTY);
					if (label != null) {
						for (AbstractObject so : label) {
							if (so instanceof StringObject) {
								String string = ((StringObject) so).getString();
								if (!linkedHashSet.contains(string)) {
									linkedHashSet.add(string);
								}
							}
						}
					}

					List<String> list = new ArrayList<>(linkedHashSet);
					if (list.size() == 1) {
						return list.get(0);
					} else {
						return list;
					}

				} else {
					List<AbstractObject> list = repository.getValuesAsList(instance, propertyId);
					if (list != null) {
						if (list.size() == 1) {
							return fixStupidMistake(valueConverter.getObject(list.get(0)));
						} else {
							return fixStupidMistake(valueConverter.convertList(list));
						}
					} else
						return null;
				}
			}

		} else {
			throw new IllegalStateException("key must be integer");
		}

	}

	public Object fixStupidMistake(Object input) {
		if (this.propertyId.equals(PropertiesExpansionData.LABEL_PROPERTY)) {
			return input;
		} else {
			ValueType targetType = ontologyRepository.getDataType(propertyId);
			if (targetType == ValueType.QUANTITY) {
				if (input instanceof List<?>) {
					List<Object> ret = new ArrayList<>();
					for (Object o : (List<?>) input) {
						try {
							if (o instanceof String) {
								try {
									ret.add(Double.parseDouble((String) o));
								}

								catch (NumberFormatException e) {

								}
							} else if (o instanceof Double) {
								ret.add(o);
							} else {
								// Pech gehabt
							}
						}

						catch (NumberFormatException e) {
							// Pech gehabt

						}
					}
				} else {
					if (input instanceof String) {
						try {
							return Double.parseDouble((String) input);

						}

						catch (NumberFormatException e) {
							// Pech gehabt
							return 0.0;
						}
					} else if (input instanceof Double) {
						return input;
					}
				}
			}
			return input;
		}
	}

	@Override
	public boolean isEmpty() {
		if (this.size() == 0)
			return true;
		else
			return false;
	}

	@Override
	public Set<Integer> keySet() {
		return new Set<Integer>() {

			@Override
			public <T> T[] toArray(T[] a) {

				throw new UnsupportedOperationException();
			}

			@Override
			public Object[] toArray() {

				throw new UnsupportedOperationException();
			}

			@Override
			public int size() {

				return RepositoryTableColumnValues.this.size();
			}

			@Override
			public boolean retainAll(Collection<?> c) {
				throw new UnsupportedOperationException();
			}

			@Override
			public boolean removeAll(Collection<?> c) {
				throw new UnsupportedOperationException();
			}

			@Override
			public boolean remove(Object o) {
				throw new UnsupportedOperationException();
			}

			@Override
			public Iterator<Integer> iterator() {
				final Iterator<InstanceId> rowIterator;
				if (idColumn) {

					rowIterator = repository.getInstances(classId).keySet().iterator();
				} else {
					rowIterator = repository.getInstances(classId, propertyId).iterator();

				}

				return new Iterator<Integer>() {

					private boolean header = true;

					@Override
					public Integer next() {
						if (header) {
							header = false;
							return 0;
						} else {
							int index = repository.getInstanceIndex(classId, rowIterator.next()) + 1;

							return index;
						}

					}

					@Override
					public boolean hasNext() {
						if (header)
							return true;
						else
							return rowIterator.hasNext();
					}
				};

			}

			@Override
			public boolean isEmpty() {
				return isEmpty();
			}

			@Override
			public boolean containsAll(Collection<?> c) {

				for (Object oc : c) {
					if (!contains(oc)) {
						return false;
					}
				}

				return false;
			}

			@Override
			public boolean contains(Object o) {
				return containsKey(o);
			}

			@Override
			public void clear() {
				throw new UnsupportedOperationException();

			}

			@Override
			public boolean addAll(Collection<? extends Integer> c) {
				throw new UnsupportedOperationException();
			}

			@Override
			public boolean add(Integer e) {
				throw new UnsupportedOperationException();
			}
		};
	}

	@Override
	public Object put(Integer key, Object value) {
		throw new UnsupportedOperationException();
	}

	@Override
	public void putAll(Map<? extends Integer, ? extends Object> m) {
		throw new UnsupportedOperationException();

	}

	@Override
	public Object remove(Object key) {
		throw new UnsupportedOperationException();
	}

	@Override
	public int size() {
		if (idColumn) {

			return repository.getInstances(classId).size() + 1;
		} else {
			return repository.getInstances(classId, propertyId).size() + 1;

		}
	}

	@Override
	public Collection<Object> values() {

		throw new UnsupportedOperationException();
	}
}
