package de.mannheim.uni.expansion.data.dbpedia.kbtables;

import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.matching.dbpedia.model.TableRow;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.ValueConverter;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;

public class RepsitoryTableRow extends TableRow {

	private final KnowledgeBase dbPediaRepository;
	private final OntologyRepository ontologyRepository;
	private final ValueConverter valueConverter;
	private final ClassId classId;
	private final InstanceId instanceId;
	private final TableRowId tableRowId;

	public InstanceId getInstanceId() {
		return instanceId;
	}

	public RepsitoryTableRow(KnowledgeBase dbPediaRepository, OntologyRepository ontologyRepository, ValueConverter valueConverter, ClassId classId,
			InstanceId instanceId, int rowIndex) {
		super(new RepositoryTable(dbPediaRepository, ontologyRepository, valueConverter, classId), rowIndex, rowIndex);
		this.ontologyRepository = ontologyRepository;
		this.valueConverter = valueConverter;
		this.dbPediaRepository = dbPediaRepository;
		this.classId = classId;
		List<AbstractObject> label = dbPediaRepository.getValuesAsList(instanceId, PropertiesExpansionData.LABEL_PROPERTY);
		if (label != null && label.size() > 0 && label.get(0) instanceof StringObject) {
			this.instanceId = new InstanceId(instanceId.getInstanceId(), label.get(0).readableString());
		} else {
			this.instanceId = instanceId;
		}
		this.tableRowId = new TableRowId(getTable().getIdObject(), rowIndex);
	}

	public RepsitoryTableRow(KnowledgeBase dbPediaRepository, OntologyRepository ontologyRepository, ValueConverter valueConverter, ClassId classId,
			InstanceId instanceId) {
		this(dbPediaRepository, ontologyRepository, valueConverter, classId, instanceId, dbPediaRepository.getInstanceIndex(classId, instanceId) + 1);
	}

	// public Object getKey() {
	// return instanceId.getReadableString();
	//
	// }

	public Object getURI() {

		return instanceId.toCombinedString();

	}

	public TableRowId getTableRowId() {
		return tableRowId;
	}

	public static class TableRowId {
		private final Object tableIdentifier;
		private final int rowIdx;

		public TableRowId(Object tableIdentifier, int rowIdx) {
			super();
			this.tableIdentifier = tableIdentifier;
			this.rowIdx = rowIdx;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + rowIdx;
			result = prime * result + ((tableIdentifier == null) ? 0 : tableIdentifier.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (!(obj instanceof TableRowId))
				return false;
			TableRowId other = (TableRowId) obj;
			if (rowIdx != other.rowIdx)
				return false;
			if (tableIdentifier == null) {
				if (other.tableIdentifier != null)
					return false;
			} else if (!tableIdentifier.equals(other.tableIdentifier))
				return false;
			return true;
		}

	}

	public static class TableRowCacheGetter {
		private final ConcurrentHashMap<TableRowId, RepsitoryTableRow> tableColumnMap = new ConcurrentHashMap<>();

		RepsitoryTableRow get(RepsitoryTableRow tr) {
			return CHMHelper.getFromCHM(tableColumnMap, tr.getTableRowId(), () -> tr);
		}

		RepsitoryTableRow getRaw(TableRowId tr) {
			return tableColumnMap.get(tr);
		}

	}

	public static TableRowCacheGetter tableRowCacheGetter = new TableRowCacheGetter();

}
