package de.mannheim.uni.expansion.data.ontology.model;

import java.util.HashMap;
import java.util.Map;

public class ResourceURI {

	private final String prefix;
	private final String URI;

	public ResourceURI(String fullURI) {

		OntologyUriPrefix foundPrefix = null;

		for (OntologyUriPrefix o : OntologyUriPrefix.values()) {
			if (fullURI.startsWith(o.getFullPrefix())) {

				foundPrefix = o;
				break;
			}

		}

		if (foundPrefix == null) {
			prefix = null;
			URI = fullURI;
		} else {
			this.prefix = foundPrefix.getCompressedPrefix();
			this.URI = fullURI.substring(0, foundPrefix.getFullPrefix().length());
		}

	}

	public String getURI() {
		if (prefix == null) {
			return URI;
		} else {
			return OntologyUriPrefix.getFromCompressed(prefix).getFullPrefix() + this.URI;
		}
	}

	public String toString() {
		if (prefix == null) {
			return URI;
		} else {
			return prefix + ":" + this.URI;
		}
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((URI == null) ? 0 : URI.hashCode());
		result = prime * result + ((prefix == null) ? 0 : prefix.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		ResourceURI other = (ResourceURI) obj;
		if (URI == null) {
			if (other.URI != null)
				return false;
		} else if (!URI.equals(other.URI))
			return false;
		if (prefix == null) {
			if (other.prefix != null)
				return false;
		} else if (!prefix.equals(other.prefix))
			return false;
		return true;
	}

	private enum OntologyUriPrefix {
		DBPEDIA_ONTOLOGY("http://dbpedia.org/ontology/", "dbo"),

		DBPEDIA_MAPPINGS_ONTOLOGY_PROPERTY("http://mappings.dbpedia.org/index.php/OntologyProperty:", "mdop"), DBPEDIA_MAPPINGS_ONTOLOGY_CLASS(
				"http://mappings.dbpedia.org/index.php/OntologyClass:", "mdoc"), DESIGN_PATTERNS("http://www.ontologydesignpatterns.org/ont/dul/DUL.owl#",
				"dul"), W3("http://www.w3.org/", "w3")

		;

		private final String fullPrefix;
		private final String compressedPrefix;

		private OntologyUriPrefix(String fullPrefix, String compressedPrefix) {
			this.fullPrefix = fullPrefix;
			this.compressedPrefix = compressedPrefix;
		}

		private static final Map<String, OntologyUriPrefix> fullMap = new HashMap<String, ResourceURI.OntologyUriPrefix>();
		private static final Map<String, OntologyUriPrefix> compressedMap = new HashMap<String, ResourceURI.OntologyUriPrefix>();

		static {
			for (OntologyUriPrefix o : OntologyUriPrefix.values()) {
				fullMap.put(o.getFullPrefix(), o);
				compressedMap.put(o.getCompressedPrefix(), o);
			}

		}

		private static OntologyUriPrefix getFromCompressed(String compressedPrefix) {
			return compressedMap.get(compressedPrefix);

		}

		@SuppressWarnings("unused")
		private static OntologyUriPrefix getFromFull(String fullPrefix) {
			return fullMap.get(fullPrefix);

		}

		private String getCompressedPrefix() {
			return compressedPrefix;

		}

		private String getFullPrefix() {
			return fullPrefix;

		}
	}

}
