package de.mannheim.uni.expansion.data.ontology.model.reader;

import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.UserDataHandler;
import org.xml.sax.SAXException;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassOntology;
import de.mannheim.uni.expansion.data.ontology.model.properties.DataTypePropertyOntology;
import de.mannheim.uni.expansion.data.ontology.model.properties.DataTypePropertyOntology.Type;
import de.mannheim.uni.expansion.data.ontology.model.properties.ObjectPropertyOntology;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyOntology;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepositoryImplMemory;
import de.mannheim.uni.expansion.utils.ConsumerHelper;

public class XMLOntologyReader {

	@Inject
	public XMLOntologyReader() {

	}

	public OntologyRepositoryImplMemory run(String xmlFile) {

		ConcurrentHashMap<ClassId, ClassOntology> classes = new ConcurrentHashMap<ClassId, ClassOntology>();
		ConcurrentHashMap<PropertyId, PropertyOntology> properties = new ConcurrentHashMap<PropertyId, PropertyOntology>();

		ConsumerHelper<Node> helper = ConsumerHelper.withLinkedQueue("OntologyReader", (node) -> {
			String name = node.getNodeName();

			if (name.equals("owl:Class")) {
				ClassOntology co = getClassOntology((Element) node);
				classes.put(co.getId(), co);
			} else if (name.equals("owl:ObjectProperty")) {
				PropertyOntology po = getObjectPropertyOntology((Element) node);
				properties.put(po.getId(), po);
			} else if (name.equals("owl:DatatypeProperty")) {
				PropertyOntology po = getDatatypePropertyOntology((Element) node);
				properties.put(po.getId(), po);
			}
		}, POISEN);
		helper.start();

		try {
			File fXmlFile = new File(xmlFile);
			DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
			Document doc = dBuilder.parse(fXmlFile);

			NodeList nodes = doc.getDocumentElement().getChildNodes();

			for (int i = 0; i < nodes.getLength(); i++) {
				final Node node = nodes.item(i);

				helper.add(node);

			}

		} catch (IOException e) {
			e.printStackTrace();
		} catch (SAXException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (ParserConfigurationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		try {
			helper.finishCloseAndAwait();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		{
			PropertyId propertyId = new PropertyId("http://www.w3.org/2003/01/geo/wgs84_pos#lat");
			properties.put(propertyId, new DataTypePropertyOntology(propertyId, null, Type.FLOAT, new HashSet<>(), null));
		}
		{
			PropertyId propertyId = new PropertyId("http://www.w3.org/2003/01/geo/wgs84_pos#long");
			properties.put(propertyId, new DataTypePropertyOntology(propertyId, null, Type.FLOAT, new HashSet<>(), null));
		}
		return new OntologyRepositoryImplMemory(classes, properties);

	}

	private PropertyOntology getObjectPropertyOntology(Element e) {
		PropertyId id = null;
		ClassId domain = null;
		ClassId range = null;
		PropertyId wasDerivedFrom = null;
		Set<PropertyId> subPropertyOf;

		// ID
		id = new PropertyId(e.getAttribute("rdf:about"));

		// domain
		NodeList domainList = e.getElementsByTagName("rdfs:domain");
		if (domainList.getLength() > 0) {
			Element domainTag = (Element) domainList.item(0);
			if (domainTag != null)
				domain = new ClassId(domainTag.getAttribute("rdf:resource"));
		}

		// range
		NodeList rangeList = e.getElementsByTagName("rdfs:range");
		if (rangeList.getLength() > 0) {
			Element rangeTag = (Element) rangeList.item(0);
			if (rangeTag != null)
				range = new ClassId(rangeTag.getAttribute("rdf:resource"));
		}

		// wasDerviedFrom
		NodeList wasDerviedFromList = e.getElementsByTagName("prov:wasDerivedFrom");
		if (wasDerviedFromList.getLength() > 0) {
			Element wasDerivedFromTag = (Element) wasDerviedFromList.item(0);
			if (wasDerivedFromTag != null)
				wasDerivedFrom = new PropertyId(wasDerivedFromTag.getAttribute("rdf:resource"));
		}

		// SubClassOf
		subPropertyOf = new HashSet<PropertyId>();
		NodeList subPropertyOfNodes = e.getElementsByTagName("rdfs:subPropertyOf");
		for (int i = 0; i < subPropertyOfNodes.getLength(); i++) {
			Element subPropertyOfTag = (Element) subPropertyOfNodes.item(i);
			PropertyId subPropertyOfId = new PropertyId(subPropertyOfTag.getAttribute("rdf:resource"));
			subPropertyOf.add(subPropertyOfId);
		}

		return new ObjectPropertyOntology(id, domain, range, subPropertyOf, wasDerivedFrom);
	}

	private PropertyOntology getDatatypePropertyOntology(Element e) {

		PropertyId id = null;
		ClassId domain = null;
		DataTypePropertyOntology.Type range = null;
		PropertyId wasDerivedFrom = null;
		Set<PropertyId> subPropertyOf;

		// ID
		id = new PropertyId(e.getAttribute("rdf:about"));

		// domain
		NodeList domainList = e.getElementsByTagName("rdfs:domain");
		if (domainList.getLength() > 0) {
			Element domainTag = (Element) domainList.item(0);
			if (domainTag != null)
				domain = new ClassId(domainTag.getAttribute("rdf:resource"));
		}

		// range
		NodeList rangeList = e.getElementsByTagName("rdfs:range");
		if (rangeList.getLength() > 0) {
			Element rangeTag = (Element) rangeList.item(0);
			if (rangeTag != null)
				range = getDatatypePropertyType(rangeTag.getAttribute("rdf:resource"));
		}

		// wasDerviedFrom
		NodeList wasDerviedFromList = e.getElementsByTagName("prov:wasDerivedFrom");
		if (wasDerviedFromList.getLength() > 0) {
			Element wasDerivedFromTag = (Element) wasDerviedFromList.item(0);
			if (wasDerivedFromTag != null)
				wasDerivedFrom = new PropertyId(wasDerivedFromTag.getAttribute("rdf:resource"));
		}

		// SubClassOf
		subPropertyOf = new HashSet<PropertyId>();
		NodeList subPropertyOfNodes = e.getElementsByTagName("rdfs:subClassOf");
		for (int i = 0; i < subPropertyOfNodes.getLength(); i++) {
			Element subPropertyOfTag = (Element) subPropertyOfNodes.item(i);
			PropertyId subPropertyOfId = new PropertyId(subPropertyOfTag.getAttribute("rdf:resource"));
			subPropertyOf.add(subPropertyOfId);
		}

		return new DataTypePropertyOntology(id, domain, range, subPropertyOf, wasDerivedFrom);

	}

	private Type getDatatypePropertyType(String attribute) {

		Type t = Type.getTypeFromString(attribute);
		if (t == null)
			System.out.println(attribute);

		return t;
	}

	private ClassOntology getClassOntology(Element e) {

		ClassId id = null;
		ClassId wasDerivedFrom = null;
		Set<ClassId> subClassOf;
		Set<ClassId> equivalentClass;

		// ID
		id = new ClassId(e.getAttribute("rdf:about"));

		// wasDerviedFrom
		NodeList wasDerviedFromList = e.getElementsByTagName("prov:wasDerivedFrom");
		if (wasDerviedFromList.getLength() > 0) {
			Element wasDerivedFromTag = (Element) wasDerviedFromList.item(0);
			if (wasDerivedFromTag != null)
				wasDerivedFrom = new ClassId(wasDerivedFromTag.getAttribute("rdf:resource"));
		}

		// Element wasDerivedFromTag = (Element)
		// e.getElementsByTagName("prov:wasDerivedFrom").item(0);
		// wasDerivedFrom = new
		// ClassId(wasDerivedFromTag.getAttribute("rdf:resource"));

		// SubClassOf
		subClassOf = new HashSet<ClassId>();
		NodeList subClassOfNodes = e.getElementsByTagName("rdfs:subClassOf");
		for (int i = 0; i < subClassOfNodes.getLength(); i++) {
			Element subClassOfTag = (Element) subClassOfNodes.item(i);
			ClassId subClassOfId = new ClassId(subClassOfTag.getAttribute("rdf:resource"));
			subClassOf.add(subClassOfId);
		}

		// EquivalentClass
		equivalentClass = new HashSet<ClassId>();
		NodeList equivalentClassNodes = e.getElementsByTagName("owl:equivalentClass");
		for (int i = 0; i < equivalentClassNodes.getLength(); i++) {
			Element equivalentClassTag = (Element) equivalentClassNodes.item(i);

			ClassId equivalentClassId = new ClassId(equivalentClassTag.getAttribute("rdf:resource"));
			equivalentClass.add(equivalentClassId);

		}

		return new ClassOntology(id, subClassOf, equivalentClass, wasDerivedFrom);

	}

	Node POISEN = new Node() {

		@Override
		public Object setUserData(String key, Object data, UserDataHandler handler) {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public void setTextContent(String textContent) throws DOMException {
			// TODO Auto-generated method stub

		}

		@Override
		public void setPrefix(String prefix) throws DOMException {
			// TODO Auto-generated method stub

		}

		@Override
		public void setNodeValue(String nodeValue) throws DOMException {
			// TODO Auto-generated method stub

		}

		@Override
		public Node replaceChild(Node newChild, Node oldChild) throws DOMException {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public Node removeChild(Node oldChild) throws DOMException {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public void normalize() {
			// TODO Auto-generated method stub

		}

		@Override
		public String lookupPrefix(String namespaceURI) {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public String lookupNamespaceURI(String prefix) {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public boolean isSupported(String feature, String version) {
			// TODO Auto-generated method stub
			return false;
		}

		@Override
		public boolean isSameNode(Node other) {
			// TODO Auto-generated method stub
			return false;
		}

		@Override
		public boolean isEqualNode(Node arg) {
			// TODO Auto-generated method stub
			return false;
		}

		@Override
		public boolean isDefaultNamespace(String namespaceURI) {
			// TODO Auto-generated method stub
			return false;
		}

		@Override
		public Node insertBefore(Node newChild, Node refChild) throws DOMException {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public boolean hasChildNodes() {
			// TODO Auto-generated method stub
			return false;
		}

		@Override
		public boolean hasAttributes() {
			// TODO Auto-generated method stub
			return false;
		}

		@Override
		public Object getUserData(String key) {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public String getTextContent() throws DOMException {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public Node getPreviousSibling() {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public String getPrefix() {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public Node getParentNode() {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public Document getOwnerDocument() {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public String getNodeValue() throws DOMException {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public short getNodeType() {
			// TODO Auto-generated method stub
			return 0;
		}

		@Override
		public String getNodeName() {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public Node getNextSibling() {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public String getNamespaceURI() {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public String getLocalName() {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public Node getLastChild() {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public Node getFirstChild() {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public Object getFeature(String feature, String version) {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public NodeList getChildNodes() {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public String getBaseURI() {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public NamedNodeMap getAttributes() {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public short compareDocumentPosition(Node other) throws DOMException {
			// TODO Auto-generated method stub
			return 0;
		}

		@Override
		public Node cloneNode(boolean deep) {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public Node appendChild(Node newChild) throws DOMException {
			// TODO Auto-generated method stub
			return null;
		}
	};
}
