package de.mannheim.uni.expansion.data.ontology.model.repository;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.DataTypePropertyOntology;
import de.mannheim.uni.expansion.data.ontology.model.properties.DataTypePropertyOntology.Type;
import de.mannheim.uni.expansion.data.ontology.model.properties.ObjectPropertyOntology;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyOntology;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.tableprocessor.model.TableColumn.ColumnDataType;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.ValueConverter;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class OntologyConverter {
	private final ValueConverter valueConverter;

	@Inject
	public OntologyConverter(ValueConverter valueConverter) {
		this.valueConverter = valueConverter;
	}

	public OntologyRepositoryImpl convert(KnowledgeBase knowledgeBase, OntologyRepositoryImplMemory old, Map<ClassPropertyId, ColumnDataType> columnDataTypes) {

		Map<PropertyId, Set<ClassId>> reverseIndex = knowledgeBase.getReversePropertyIndex();

		final Map<PropertyId, ValueType> propertyDataType = new HashMap<>();
		final Set<PropertyId> propertyTimeDependent = new HashSet<>();
		final Map<ClassPropertyId, Set<ClassId>> propertyRange = new HashMap<>();
		final Map<ClassPropertyId, Set<ClassId>> propertyDomain = new HashMap<>();
		Map<ClassId, Set<ClassId>> children = old.getSubClasses();

		Map<PropertyId, Set<PropertyId>> propertyParents = new HashMap<>();

		Map<ClassId, Set<ClassId>> parent = new HashMap<>();

		for (Entry<ClassId, Set<ClassId>> childrenEntry : children.entrySet()) {

			for (ClassId child : childrenEntry.getValue()) {
				CHMHelper.getFromMap(parent, child, () -> new HashSet<>()).add(childrenEntry.getKey());
			}

		}

		for (Entry<PropertyId, PropertyOntology> poEntry : old.getProperties().entrySet()) {
			if (poEntry.getValue() instanceof ObjectPropertyOntology && reverseIndex.containsKey(poEntry.getKey())) {

				propertyDataType.put(poEntry.getKey(), ValueType.INSTANCE_REFERENCE);

				ObjectPropertyOntology opo = (ObjectPropertyOntology) poEntry.getValue();
				ClassId rangeClass = opo.getRange();
				if (rangeClass != null) {

					Set<ClassId> range = getWithAllSubClases(old, rangeClass);
					for (ClassId setId : reverseIndex.get(poEntry.getKey())) {
						propertyRange.put(new ClassPropertyId(setId, poEntry.getKey()), range);
					}
				}

				ClassId domainClass = opo.getDomain();
				if (domainClass != null) {

					Set<ClassId> domain = getWithAllSubClases(old, domainClass);
					for (ClassId setId : reverseIndex.get(poEntry.getKey())) {
						propertyDomain.put(new ClassPropertyId(setId, poEntry.getKey()), domain);
					}
				}

				propertyParents.put(poEntry.getKey(), opo.getSubPropertyOf());

			} else if (poEntry.getValue() instanceof DataTypePropertyOntology) {
				Type range = ((DataTypePropertyOntology) poEntry.getValue()).getRange();

				if (range != null) {
					ColumnDataType cdt = range.getColumnDataType();
					ValueType valueType = valueConverter.getDataType(cdt);
					propertyDataType.put(poEntry.getKey(), valueType);
				}
			}
		}

		for (Entry<ClassPropertyId, ColumnDataType> cdtEntry : columnDataTypes.entrySet()) {
			if (!propertyDataType.containsKey(cdtEntry.getKey().getPropertyId())) {
				ValueType valueType = valueConverter.getDataType(cdtEntry.getValue());
				propertyDataType.put(cdtEntry.getKey().getPropertyId(), valueType);
			}

		}

		return new OntologyRepositoryImpl(propertyDataType, propertyTimeDependent, propertyDomain, propertyRange, parent, children, propertyParents);

	}

	// private Set<ClassId> getAllSubClases(ClassId classId) {
	//
	// Set<ClassId> classes = new HashSet<ClassId>();
	// classes.add(classId);
	//
	// Set<ClassId> subClasses = ontologyRepository.getSubClasses(classId);
	// if (subClasses != null) {
	// for (ClassId subClassId : subClasses) {
	// classes.addAll(getAllSubClases(subClassId));
	// }
	// }
	//
	// return classes;
	// }

	Set<ClassId> getWithAllSubClases(OntologyRepositoryImplMemory ontologyRepository, ClassId classId) {

		Set<ClassId> classes = new HashSet<ClassId>();
		classes.add(classId);

		Set<ClassId> subClasses = ontologyRepository.getSubClasses(classId);
		if (subClasses != null) {
			for (ClassId subClassId : subClasses) {
				classes.addAll(getWithAllSubClases(ontologyRepository, subClassId));
			}
		}

		return classes;
	}
}
