package de.mannheim.uni.expansion.data.tables.filters;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

import javax.inject.Inject;

import com.google.common.collect.ImmutableTable;
import com.google.common.collect.ImmutableTable.Builder;

import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.PropertyIndexComposite;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.measurement.UnitConverter;
import de.uni_mannheim.dws.expansion.data.ontology.model.measurement.UnitOfMeasurement;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.QuantityObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class TablePropertyAndUnitNormalizer implements AnnotatedTableFilter {

	private final Map<PropertyId, UnitOfMeasurement> units;
	private final Map<PropertyId, PropertyId> mapping;
	private final UnitConverter unitConverter;

	@Inject
	public TablePropertyAndUnitNormalizer(KnowledgeBase kb, UnitConverter unitConverter) {
		super();
		this.units = kb.getKbUnits();
		this.mapping = kb.getNormalizationIndex();
		this.unitConverter = unitConverter;
	}

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {

		boolean changed = false;

		Builder<RowIdx, ColumnIdx, AbstractObject> builder = null;

		Map<Integer, PropertyId> propertyMapping = null;
		Map<PropertyIndexComposite, Double> scores = null;

		for (Entry<Integer, PropertyId> properties : data.getMetaData().getProperties().entrySet()) {
			if (mapping.containsKey(properties.getValue())) {

				if (builder == null) {
					builder = ImmutableTable.builder();
					builder.putAll(data.getTableData().getTable());

					propertyMapping = new HashMap<>(data.getMetaData().getProperties());
					scores = new HashMap<>(data.getMetaData().getAllPropertyScores());

				}

				UnitOfMeasurement currentUnit = units.get(properties.getValue());
				PropertyId targetProperty = mapping.get(properties.getValue());
				UnitOfMeasurement targetUnit = units.get(targetProperty);

				propertyMapping.put(properties.getKey(), targetProperty);

				Double score = scores.remove(new PropertyIndexComposite(properties.getKey(), properties.getValue()));
				scores.put(new PropertyIndexComposite(properties.getKey(), targetProperty), score);

				ColumnIdx columnIdx = new ColumnIdx(properties.getKey());

				for (Entry<RowIdx, AbstractObject> row : data.getTableData().getColumn(columnIdx).entrySet()) {

					if (row.getValue().getType().equals(ValueType.QUANTITY)) {
						double newValue = unitConverter.convert(((QuantityObject) row.getValue()).getValue(), currentUnit, targetUnit);

						builder.put(row.getKey(), columnIdx, QuantityObject.fromDouble(newValue));
					} else {
						int sdf = 123;
					}
				}

			}
		}

		if (changed) {
			AnnotatedTable oldData = data.getTableData();
			AnnotatedTableMetaData newMd = data.getMetaData().rebuilder().setProperties(propertyMapping).setAllPropertyScores(scores).build();
			AnnotatedTable newData = new AnnotatedTable(builder.build(), oldData.getStringTable(), oldData.getDataTypes(), oldData.getKeyColumnIndex(),
					oldData.getColumnHeaders());
			return Optional.of(new AnnotatedTablesReaderData(data.getId(), newMd, newData));
		} else
			return Optional.of(data);
	}

}
