package de.mannheim.uni.expansion.data.tables.model;

import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.BiConsumer;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier.CascadedType;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier.SimpleType;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier.TimestampType;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.ConfigInterface;
import de.mannheim.uni.expansion.utils.CHMHelper;

public class AnnotatedTableMetaData implements Serializable, ConfigInterface<AnnotatedTableMetaData> {

	/**
	 * 
	 */
	private static final long serialVersionUID = 2L;
	private final DomainSource source;
	private final HostId hostId;
	private ClassId dbPediaClass;
	private final double classConf;
	private final int numHeaderRows;

	// FIXME: MOVE TO ROWIDX AND COLIDX
	private Map<Integer, List<PropertyId>> properties;
	// private final Map<Integer, Double> propertiesScores;
	private Map<Integer, List<InstanceId>> instances;
	// private final Map<Integer, Double> instancesScores;
	private final int keyColumn;

	private final Map<InstanceId, List<Integer>> reverseInstances;
	private final Map<PropertyId, List<Integer>> reverseProperties;

	public Map<PropertyIndexComposite, Double> propertiesScores;
	public Map<InstanceIndexComposite, Double> instancesScores;

	private final Map<Integer, PropertyId> highestRatedProperties;
	private final Map<Integer, InstanceId> highestRatedInstances;

	private final List<Integer> yearsBeforeText;
	private final Integer maxYearBeforeText;
	private final List<Integer> yearsAfterText;
	private final Integer maxYearAfterText;
	private final List<Integer> yearsInContextBefore;
	private final Integer maxYearInContextBefore;
	private final List<Integer> yearsInTableTitle;
	private final Integer maxYearInTableTitle;
	private final List<Integer> yearsInPageTitle;
	private final Integer maxYearInPageTitle;

	private final Map<Integer, List<Integer>> propertiesYears;
	private final Map<Integer, Integer> propertiesMaxYear;

	private final String pageTitle;
	private final String tableTitle;
	private final String tableContextTimeStampBeforeTable;
	private final String textBeforeTable;
	private final String textAfterTable;
	private int numberOfRows;

	public AnnotatedTableMetaData(DomainSource source, ClassId dbPediaClass, double classConf, int numHeaderRows, int numberOfRows,
			Map<Integer, List<PropertyId>> properties, Map<PropertyIndexComposite, Double> propertiesScores, Map<Integer, List<InstanceId>> instances,
			Map<InstanceIndexComposite, Double> instancesScores, int keyColumn) {
		this(source, dbPediaClass, classConf, numHeaderRows, numberOfRows, properties, propertiesScores, instances, instancesScores, keyColumn, null, null,
				null, null, null, null, null, null, null, null, null);
	}

	public AnnotatedTableMetaData(DomainSource source, ClassId dbPediaClass, double classConf, int numHeaderRows, int numberOfRows,
			Map<Integer, List<PropertyId>> properties, Map<PropertyIndexComposite, Double> propertiesScores, Map<Integer, List<InstanceId>> instances,
			Map<InstanceIndexComposite, Double> instancesScores, int keyColumn, List<Integer> beforeTextYears, List<Integer> afterTextYears,
			List<Integer> contextBeforeYears, List<Integer> tableTitleYears, List<Integer> pageTitleYears, Map<Integer, List<Integer>> propertiesYears,
			String pageTitle, String tableTitle, String tableContextTimeStampBeforeTable, String textBeforeTable, String textAfterTable) {
		super();
		this.source = source;
		this.hostId = source.getHostId();
		this.dbPediaClass = dbPediaClass;
		this.classConf = classConf;
		this.numHeaderRows = numHeaderRows;
		this.propertiesScores = propertiesScores;
		this.instancesScores = instancesScores;
		this.keyColumn = keyColumn;
		this.numberOfRows = numberOfRows;

		// WORKING WITH INSTANCES
		Comparator<InstanceIndexComposite> instanceComparator = (i1, i2) -> {
			return instancesScores.get(i1).compareTo(instancesScores.get(i2));

		};

		HashMap<Integer, List<InstanceId>> tempInstances = new HashMap<Integer, List<InstanceId>>();
		for (Entry<Integer, List<InstanceId>> e : instances.entrySet()) {
			final Integer index = e.getKey();

			List<InstanceId> sortedList = e.getValue();
			if (sortedList != null) {
				sortedList = e.getValue().stream().map((p) -> {
					return new InstanceIndexComposite(index, p);
				}).sorted(instanceComparator).map((p) -> {
					return p.getInstanceId();
				}).collect(Collectors.toList());
			}

			tempInstances.put(e.getKey(), sortedList);

		}
		this.instances = tempInstances;

		highestRatedInstances = this.instances.entrySet().stream().filter((e) -> {
			return (e.getValue() != null && !e.getValue().isEmpty());
		}).collect(Collectors.toMap(Entry::getKey, (e) -> {

			return e.getValue().get(0);
		}));

		HashMap<InstanceId, List<Integer>> reverseInstancesTemp = new HashMap<InstanceId, List<Integer>>();

		for (Entry<Integer, List<InstanceId>> e : instances.entrySet()) {

			if (e.getValue() != null) {
				for (InstanceId id : e.getValue()) {
					List<Integer> indeces = reverseInstancesTemp.get(id);
					if (indeces == null) {
						indeces = new ArrayList<Integer>();
						reverseInstancesTemp.put(id, indeces);
					}
					indeces.add(e.getKey());
				}
			}

		}

		for (Entry<InstanceId, List<Integer>> e : reverseInstancesTemp.entrySet()) {
			final InstanceId instanceId = e.getKey();

			List<Integer> sortedList = e.getValue().stream().map((p) -> {
				return new InstanceIndexComposite(p, instanceId);
			}).sorted(instanceComparator).map((p) -> {
				return p.getIndex();
			}).collect(Collectors.toList());

			reverseInstancesTemp.put(e.getKey(), sortedList);

		}

		reverseInstances = reverseInstancesTemp;

		// WORKING WITH PROPERTIES
		Comparator<PropertyIndexComposite> propertyComparator = (p1, p2) -> {
			return propertiesScores.get(p1).compareTo(propertiesScores.get(p2));

		};

		HashMap<Integer, List<PropertyId>> tempProperties = new HashMap<Integer, List<PropertyId>>();
		for (Entry<Integer, List<PropertyId>> e : properties.entrySet()) {
			final Integer index = e.getKey();

			List<PropertyId> sortedList = e.getValue().stream().map((p) -> {
				return new PropertyIndexComposite(index, p);
			}).sorted(propertyComparator).map((p) -> {
				return p.getPropertyId();
			}).collect(Collectors.toList());

			tempProperties.put(e.getKey(), sortedList);

		}
		this.properties = tempProperties;

		highestRatedProperties = this.properties.entrySet().stream().filter((e) -> {
			return e.getValue() != null && !e.getValue().isEmpty();
		}).collect(Collectors.toMap(Entry::getKey, (e) -> {
			return e.getValue().get(0);
		}));

		HashMap<PropertyId, List<Integer>> reversePropertiesTemp = new HashMap<PropertyId, List<Integer>>();

		for (Entry<Integer, List<PropertyId>> e : properties.entrySet())

		{

			for (PropertyId id : e.getValue()) {
				List<Integer> indeces = reversePropertiesTemp.get(id);
				if (indeces == null) {
					indeces = new ArrayList<Integer>();
					reversePropertiesTemp.put(id, indeces);
				}
				indeces.add(e.getKey());
			}

		}

		for (

		Entry<PropertyId, List<Integer>> e : reversePropertiesTemp.entrySet())

		{
			final PropertyId propertyId = e.getKey();

			List<Integer> sortedList = e.getValue().stream().map((p) -> {
				return new PropertyIndexComposite(p, propertyId);
			}).sorted(propertyComparator).map((p) -> {
				return p.getIndex();
			}).collect(Collectors.toList());

			reversePropertiesTemp.put(e.getKey(), sortedList);

		}

		reverseProperties = reversePropertiesTemp;

		if (beforeTextYears != null && !beforeTextYears.isEmpty()) {
			this.yearsBeforeText = new ArrayList<>(beforeTextYears);
			this.maxYearBeforeText = Collections.max(beforeTextYears);
		} else {
			this.yearsBeforeText = new ArrayList<>();
			this.maxYearBeforeText = null;
		}

		if (afterTextYears != null && !afterTextYears.isEmpty()) {
			this.yearsAfterText = new ArrayList<>(afterTextYears);
			this.maxYearAfterText = Collections.max(afterTextYears);
		} else {
			this.yearsAfterText = new ArrayList<>();
			this.maxYearAfterText = null;
		}

		if (contextBeforeYears != null && !contextBeforeYears.isEmpty()) {
			this.yearsInContextBefore = new ArrayList<>(contextBeforeYears);
			this.maxYearInContextBefore = Collections.max(contextBeforeYears);
		} else {
			this.yearsInContextBefore = new ArrayList<>();
			this.maxYearInContextBefore = null;
		}

		if (tableTitleYears != null && !tableTitleYears.isEmpty()) {
			this.yearsInTableTitle = new ArrayList<>(tableTitleYears);
			this.maxYearInTableTitle = Collections.max(tableTitleYears);
		} else {
			this.yearsInTableTitle = new ArrayList<>();
			this.maxYearInTableTitle = null;
		}

		if (pageTitleYears != null && !pageTitleYears.isEmpty()) {
			this.yearsInPageTitle = new ArrayList<>(pageTitleYears);
			this.maxYearInPageTitle = Collections.max(pageTitleYears);
		} else {
			this.yearsInPageTitle = new ArrayList<>();
			this.maxYearInPageTitle = null;
		}

		if (propertiesYears != null && !propertiesYears.isEmpty()) {
			Map<Integer, Integer> propertiesMaxYearTemp = new HashMap<>();
			Map<Integer, List<Integer>> propertiesYearsTemp = new HashMap<>();
			for (Entry<Integer, List<Integer>> e : propertiesYears.entrySet()) {
				if (e.getValue() != null) {
					propertiesMaxYearTemp.put(e.getKey(), Collections.max(e.getValue()));
					propertiesYearsTemp.put(e.getKey(), Collections.unmodifiableList(new ArrayList<>(e.getValue())));
				}
			}

			this.propertiesMaxYear = Collections.unmodifiableMap(propertiesMaxYearTemp);
			this.propertiesYears = Collections.unmodifiableMap(propertiesYearsTemp);

		} else {
			this.propertiesMaxYear = null;
			this.propertiesYears = null;
		}

		this.pageTitle = pageTitle;
		this.tableTitle = tableTitle;
		this.tableContextTimeStampBeforeTable = tableContextTimeStampBeforeTable;
		this.textBeforeTable = textBeforeTable;
		this.textAfterTable = textAfterTable;
	}

	public List<Integer> getYearsBeforeText() {
		return yearsBeforeText;
	}

	public Integer getMaxYearBeforeText() {
		return maxYearBeforeText;
	}

	public List<Integer> getYearsAfterText() {
		return yearsAfterText;
	}

	public Integer getMaxYearAfterText() {
		return maxYearAfterText;
	}

	public List<Integer> getYearsInContextBefore() {
		return yearsInContextBefore;
	}

	public Integer getMaxYearInContextBefore() {
		return maxYearInContextBefore;
	}

	public List<Integer> getYearsInTableTitle() {
		return yearsInTableTitle;
	}

	public Integer getMaxYearInTableTitle() {
		return maxYearInTableTitle;
	}

	public List<Integer> getYearsInPageTitle() {
		return yearsInPageTitle;
	}

	public Integer getMaxYearInPageTitle() {
		return maxYearInPageTitle;
	}

	public Map<Integer, List<Integer>> getPropertiesYears() {
		if (propertiesYears == null)
			return null;
		else {
			return propertiesYears;
		}
	}

	public Map<Integer, Integer> getPropertiesMaxYear() {
		if (propertiesMaxYear == null)
			return null;
		else
			return propertiesMaxYear;
	}

	public String getPageTitle() {
		return pageTitle;
	}

	public String getTableTitle() {
		return tableTitle;
	}

	public String getTableContextTimeStampBeforeTable() {
		return tableContextTimeStampBeforeTable;
	}

	public String getTextBeforeTable() {
		return textBeforeTable;
	}

	public String getTextAfterTable() {
		return textAfterTable;
	}

	public DomainSource getSource() {
		return source;
	}

	public ClassId getDbPediaClass() {
		return dbPediaClass;
	}

	public double getClassConf() {
		return classConf;
	}

	public int getNumHeaderRows() {
		return numHeaderRows;
	}

	public Map<Integer, PropertyId> getProperties() {

		return highestRatedProperties;
	}

	public int getNumberOfRows() {
		return numberOfRows;
	}

	public Map<Integer, List<InstanceId>> getAllInstances() {
		return instances;
	}

	public Map<Integer, InstanceId> getInstances() {
		return highestRatedInstances;
	}

	public int getKeyColumn() {
		return keyColumn;
	}

	public int getNumOfColums() {
		return this.properties.size();
	}

	public int getRow(InstanceId instance) {
		return reverseInstances.get(instance).get(0);
	}

	public Optional<RowIdx> getRowIdx(InstanceId instance) {
		if (reverseInstances.containsKey(instance)) {
			return Optional.of(new RowIdx(reverseInstances.get(instance).get(0)));
		}
		return Optional.empty();

	}

	public int getCol(PropertyId property) {
		return reverseProperties.get(property).get(0);
	}

	public Optional<List<Integer>> getCols(PropertyId property) {
		if (reverseProperties.containsKey(property))
			return Optional.of(reverseProperties.get(property));
		else
			return Optional.empty();
	}

	public HostId getHostId() {
		return hostId;
	}

	public InstanceId getInstance(int index) {
		return highestRatedInstances.get(index);
	}

	public PropertyId getProperty(int index) {
		return highestRatedProperties.get(index);
	}

	public Map<Integer, List<PropertyId>> getAllProperties() {
		return properties;
	}

	public Map<PropertyIndexComposite, Double> getAllPropertyScores() {
		return propertiesScores;
	}

	private transient ConcurrentHashMap<InstanceId, Double> maxScorePerEntity = new ConcurrentHashMap<>();

	public double getScore(InstanceId instanceId) {
		Double score = CHMHelper.getFromCHM(maxScorePerEntity, instanceId, () -> {
			double max = 0.0;
			for (Integer i : this.reverseInstances.get(instanceId)) {
				double tempScore = instancesScores.get(new InstanceIndexComposite(i, instanceId));
				if (tempScore > max)
					max = tempScore;
			}
			return max;
		});
		if (score != null) {
			return score;
		} else {
			return 0.0;
		}

	}

	public Map<InstanceIndexComposite, Double> getInstancesScores() {
		return this.instancesScores;
	}

	public Double getInstanceScore(InstanceIndexComposite instanceIndexComposite) {
		return this.instancesScores.get(instanceIndexComposite);
	}

	public Double getPropertyScore(PropertyIndexComposite propertyIndexComposite) {
		return this.propertiesScores.get(propertyIndexComposite);
	}

	private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {

		in.defaultReadObject();
		this.maxScorePerEntity = new ConcurrentHashMap<>();
	}

	public AnnotatedTableMetaData filterOut(Set<PropertyId> removeProperties, Set<InstanceId> removeInstances) {

		Map<Integer, List<PropertyId>> properties = new HashMap<>();
		Map<Integer, List<InstanceId>> instances = new HashMap<>();
		Map<PropertyIndexComposite, Double> propertiesScores = new HashMap<>(this.propertiesScores);
		Map<InstanceIndexComposite, Double> instancesScores = new HashMap<>(this.instancesScores);

		for (Entry<Integer, List<PropertyId>> e : this.properties.entrySet()) {
			for (PropertyId propertyId : e.getValue()) {
				if (!removeProperties.contains(propertyId)) {
					CHMHelper.getFromMap(properties, e.getKey(), () -> new ArrayList<>()).add(propertyId);
				} else {
					propertiesScores.remove(new PropertyIndexComposite(e.getKey(), propertyId));
				}
			}
		}

		for (Entry<Integer, List<InstanceId>> e : this.instances.entrySet()) {
			for (InstanceId instanceId : e.getValue()) {
				if (!removeProperties.contains(instanceId)) {
					CHMHelper.getFromMap(instances, e.getKey(), () -> new ArrayList<>()).add(instanceId);
				} else {
					propertiesScores.remove(new InstanceIndexComposite(e.getKey(), instanceId));
				}
			}
		}

		return new AnnotatedTableMetaData(getSource(), getDbPediaClass(), getClassConf(), getNumHeaderRows(), getNumberOfRows(), properties, propertiesScores,
				instances, instancesScores, getKeyColumn(), getYearsBeforeText(), getYearsAfterText(), getYearsInContextBefore(), getYearsInTableTitle(),
				getYearsInPageTitle(), getPropertiesYears(), getPageTitle(), getTableTitle(), getTableContextTimeStampBeforeTable(), getTextBeforeTable(),
				getTextAfterTable());
	}

	public Map<TimestampIdentifier, Integer> getTableYears() {
		Map<TimestampIdentifier, Integer> maps = new HashMap<>();

		BiConsumer<SimpleType, List<Integer>> simpleIdentifier = (simpleType, list) -> {
			for (int i = 0; i < list.size(); i++) {
				maps.put(TimestampIdentifier.createSimpleIdentifier(simpleType, i), list.get(i));
			}
		};

		BiConsumer<CascadedType, Map<Integer, List<Integer>>> cascadedIdentifier = (cascadedType, map) -> {
			if (map != null) {
				for (Entry<Integer, List<Integer>> e : map.entrySet()) {
					List<Integer> list = e.getValue();
					for (int i = 0; i < list.size(); i++) {
						maps.put(TimestampIdentifier.createCascadedIdentifier(cascadedType, e.getKey(), i), list.get(i));
					}
				}
			}
		};

		simpleIdentifier.accept(SimpleType.PAGE_TITLE, getYearsInPageTitle());
		simpleIdentifier.accept(SimpleType.TABLE_TITLE, getYearsInTableTitle());
		simpleIdentifier.accept(SimpleType.BEFORE, getYearsBeforeText());
		simpleIdentifier.accept(SimpleType.AFTER, getYearsAfterText());
		simpleIdentifier.accept(SimpleType.CONTEXT_BEFORE, getYearsInContextBefore());

		cascadedIdentifier.accept(CascadedType.COLUMN_HEADER, getPropertiesYears());

		return maps;
	}

	public Integer getTimeStamp(TimestampIdentifier tsi) {
		if (tsi.getType() == TimestampType.SIMPLE) {
			switch (tsi.getSimpleType()) {
			case AFTER:
				return this.yearsAfterText.get(tsi.getTimeStampPosition());

			case BEFORE:
				return this.yearsBeforeText.get(tsi.getTimeStampPosition());

			case CONTEXT_BEFORE:
				return this.yearsInContextBefore.get(tsi.getTimeStampPosition());

			case PAGE_TITLE:

				return this.yearsInPageTitle.get(tsi.getTimeStampPosition());

			case TABLE_TITLE:
				return this.yearsInTableTitle.get(tsi.getTimeStampPosition());

			}
		} else if (tsi.getCascadedType() == CascadedType.COLUMN_HEADER) {
			return this.getPropertiesYears().get(tsi.getCascadedTypePosition()).get(tsi.getTimeStampPosition());
		}

		throw new IllegalStateException();
	}

	@Override
	public AnnotatedTableMetaData copy() {
		return new AnnotatedTableMetaData(source, dbPediaClass, classConf, numHeaderRows, numberOfRows, properties, propertiesScores, instances,
				instancesScores, keyColumn, this.yearsBeforeText, this.yearsAfterText, this.yearsInContextBefore, this.yearsInTableTitle, this.yearsInPageTitle,
				propertiesYears, this.pageTitle, tableTitle, this.tableContextTimeStampBeforeTable, this.textBeforeTable, this.textAfterTable);

	}

	@Override
	public AnnotatedTableMetaDataBuilder rebuilder() {
		return new AnnotatedTableMetaDataBuilder(this);
	}

	public static class AnnotatedTableMetaDataBuilder implements ConfigBuilderInterface<AnnotatedTableMetaData> {

		AnnotatedTableMetaData md;

		public AnnotatedTableMetaDataBuilder(AnnotatedTableMetaData annotatedTableMetaData) {
			this.md = annotatedTableMetaData.copy();
		}

		@Override
		public AnnotatedTableMetaData build() {
			return md.copy();
		}

		public AnnotatedTableMetaDataBuilder setClassId(ClassId classId) {
			md.dbPediaClass = classId;
			return this;
		}

		public AnnotatedTableMetaDataBuilder setNumberOfRows(int size) {
			md.numberOfRows = size;
			return this;
		}

		public AnnotatedTableMetaDataBuilder setInstanceScores(Map<InstanceIndexComposite, Double> instanceScores) {
			md.instancesScores = new HashMap<>(instanceScores);
			return this;
		}

		public AnnotatedTableMetaDataBuilder setPropertiesWithScores(Map<Integer, List<PropertyId>> properties, Map<PropertyIndexComposite, Double> scores) {

			md.properties = new HashMap<>(properties);
			md.propertiesScores = new HashMap<>(scores);
			return this;
		}

		public AnnotatedTableMetaDataBuilder setInstancesWithScores(Map<Integer, List<InstanceId>> instances, Map<InstanceIndexComposite, Double> scores) {

			md.instances = new HashMap<>(instances);
			md.instancesScores = new HashMap<>(scores);
			return this;
		}

		public AnnotatedTableMetaDataBuilder setProperties(Map<Integer, PropertyId> highestRatedProperties) {
			Map<Integer, List<PropertyId>> properties = highestRatedProperties.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				List<PropertyId> list = new ArrayList<>();
				list.add(e.getValue());
				return list;
			}));
			md.properties = properties;
			return this;
		}

		public AnnotatedTableMetaDataBuilder setAllPropertyScores(Map<PropertyIndexComposite, Double> scores) {
			md.propertiesScores = new HashMap<>(scores);
			return this;
		}

	}

}
