package de.mannheim.uni.expansion.data.tables.model;

public class CellId {
	private final AnnotatedTableId tableId;
	private final ColumnIdx columnIdx;
	private final RowIdx rowIdx;

	public CellId(AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {
		super();
		this.tableId = tableId;
		this.columnIdx = columnIdx;
		this.rowIdx = rowIdx;
	}

	public CellId(AnnotatedTableId tableId, Integer columnIdx, Integer rowIdx) {
		super();
		this.tableId = tableId;
		this.columnIdx = new ColumnIdx(columnIdx);
		this.rowIdx = new RowIdx(rowIdx);
	}

	public AnnotatedTableId getTableId() {
		return tableId;
	}

	public TableColumnId getTableColumnId() {
		return new TableColumnId(tableId, columnIdx);
	}

	public TableRowId getTableRowId() {
		return new TableRowId(tableId, rowIdx);
	}

	public ColumnIdx getColumnIdx() {
		return columnIdx;
	}

	public RowIdx getRowIdx() {
		return rowIdx;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((columnIdx == null) ? 0 : columnIdx.hashCode());
		result = prime * result + ((rowIdx == null) ? 0 : rowIdx.hashCode());
		result = prime * result + ((tableId == null) ? 0 : tableId.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		CellId other = (CellId) obj;
		if (columnIdx == null) {
			if (other.columnIdx != null)
				return false;
		} else if (!columnIdx.equals(other.columnIdx))
			return false;
		if (rowIdx == null) {
			if (other.rowIdx != null)
				return false;
		} else if (!rowIdx.equals(other.rowIdx))
			return false;
		if (tableId == null) {
			if (other.tableId != null)
				return false;
		} else if (!tableId.equals(other.tableId))
			return false;
		return true;
	}

	@Override
	public String toString() {
		return "CellId [tableId=" + tableId + ", columnIdx=" + columnIdx + ", rowIdx=" + rowIdx + "]";
	}

}
