package de.mannheim.uni.expansion.data.tables.model;

public class TableColumnId {
	private final AnnotatedTableId tableId;
	private final ColumnIdx columnIdx;
	private transient Integer hashCode;

	public TableColumnId(AnnotatedTableId tableId, Integer columnIdx) {
		super();
		this.tableId = tableId;
		this.columnIdx = new ColumnIdx(columnIdx);
	}

	public TableColumnId(AnnotatedTableId tableId, ColumnIdx columnIdx) {
		super();
		this.tableId = tableId;
		this.columnIdx = columnIdx;
	}

	public TableColumnId(CellId cellId) {
		this.tableId = cellId.getTableId();
		this.columnIdx = cellId.getColumnIdx();
	}

	public AnnotatedTableId getTableId() {
		return tableId;
	}

	public ColumnIdx getColumnIdx() {
		return columnIdx;
	}

	@Override
	public int hashCode() {
		if (hashCode == null) {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((columnIdx == null) ? 0 : columnIdx.hashCode());
			result = prime * result + ((tableId == null) ? 0 : tableId.hashCode());
			hashCode = result;
		}
		return hashCode;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		if (this.hashCode() != obj.hashCode())
			return false;
		TableColumnId other = (TableColumnId) obj;
		if (columnIdx == null) {
			if (other.columnIdx != null)
				return false;
		} else if (!columnIdx.equals(other.columnIdx))
			return false;
		if (tableId == null) {
			if (other.tableId != null)
				return false;
		} else if (!tableId.equals(other.tableId))
			return false;
		return true;
	}

	@Override
	public String toString() {
		return "TableColumnId [tableId=" + tableId + ", columnIdx=" + columnIdx + "]";
	}

}
