package de.mannheim.uni.expansion.data.tables.model;

import java.util.Comparator;

public class TableIdPair {
	private final AnnotatedTableId id1;
	private final AnnotatedTableId id2;
	private transient Integer hashCode;

	private final static Comparator<AnnotatedTableId> comp = (t1, t2) -> {
		return AnnotatedTableId.compare(t1, t2);

	};

	public TableIdPair(AnnotatedTableId id1, AnnotatedTableId id2) {

		int compare = comp.compare(id1, id2);
		if (compare == 0) {
			throw new IllegalStateException("A Table ID Pair can not be between the same entry");
		} else if (compare < 0) {
			this.id1 = id1;
			this.id2 = id2;
		} else {
			this.id1 = id2;
			this.id2 = id1;
		}
	}

	@Override
	public int hashCode() {
		if (hashCode == null) {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((id1 == null) ? 0 : id1.hashCode());
			result = prime * result + ((id2 == null) ? 0 : id2.hashCode());
			hashCode = result;
		}
		return hashCode;

	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		if (this.hashCode() != obj.hashCode())
			return false;
		TableIdPair other = (TableIdPair) obj;
		if (id1 == null) {
			if (other.id1 != null)
				return false;
		} else if (!id1.equals(other.id1))
			return false;
		if (id2 == null) {
			if (other.id2 != null)
				return false;
		} else if (!id2.equals(other.id2))
			return false;
		return true;
	}

	public AnnotatedTableId getId1() {
		return id1;
	}

	public AnnotatedTableId getId2() {
		return id2;
	}

}
