package de.mannheim.uni.expansion.data.tables.model;

public class TableRowId {
	private final AnnotatedTableId tableId;
	private final RowIdx rowIdx;
	private transient int hashCode = 0;

	public TableRowId(AnnotatedTableId tableId, RowIdx rowIdx) {
		super();
		this.tableId = tableId;
		this.rowIdx = rowIdx;
	}

	public static int compare(TableRowId t1, TableRowId t2) {
		int c1 = t1.getTableId().getTableName().compareTo(t2.getTableId().getTableName());

		if (c1 == 0) {
			c1 = t1.getRowIdx().getRowIdx().compareTo(t2.getRowIdx().getRowIdx());
		}

		return c1;
	}

	public TableRowId(String tableId, Integer rowIdx) {
		super();
		this.tableId = new AnnotatedTableId(tableId);
		this.rowIdx = new RowIdx(rowIdx);
	}

	public TableRowId(AnnotatedTableId tableId, Integer rowIdx) {
		super();
		this.tableId = tableId;
		this.rowIdx = new RowIdx(rowIdx);
	}

	public TableRowId(CellId cellId) {
		this.tableId = cellId.getTableId();
		this.rowIdx = cellId.getRowIdx();
	}

	public AnnotatedTableId getTableId() {
		return tableId;
	}

	public RowIdx getRowIdx() {
		return rowIdx;
	}

	@Override
	public int hashCode() {
		if (hashCode == 0) {
			int prime = 31;
			final int rowIdxHash = (rowIdx == null) ? 0 : rowIdx.hashCode();
			final int tableIdHash = (tableId == null) ? 0 : tableId.hashCode();
			int result = 1;
			result = prime * result + (rowIdxHash);
			result = prime * result + (tableIdHash);
			if (result == 0) {
				prime = 92821;
				result = 1;
				result = prime * result + (rowIdxHash);
				result = prime * result + (tableIdHash);
			}
			if (result == 0) {
				result = 1;
			}
			hashCode = result;

		}
		return hashCode;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		if (this.hashCode() != obj.hashCode())
			return false;
		TableRowId other = (TableRowId) obj;
		if (rowIdx == null) {
			if (other.rowIdx != null)
				return false;
		} else if (!rowIdx.equals(other.rowIdx))
			return false;
		if (tableId == null) {
			if (other.tableId != null)
				return false;
		} else if (!tableId.equals(other.tableId))
			return false;
		return true;
	}

	@Override
	public String toString() {
		return "TableRowId [tableId=" + tableId + ", rowIdx=" + rowIdx + "]";
	}

	public String getCombinedString() {
		return tableId.toString() + "|^~" + rowIdx.getRowIdx();
	}

	public String getCombinedString2() {
		return tableId.toString() + "," + rowIdx.getRowIdx();
	}

	public static TableRowId fromCombinedString2(String string) {
		String[] split = string.split(",");
		return new TableRowId(split[0], Integer.parseInt(split[1]));
	}

}
