package de.mannheim.uni.expansion.data.tables.model.timestamp;

public class TimestampIdentifier {
	public enum TimestampType {
		SIMPLE, CASCADED, PROPAGATED;
	}

	public enum CascadedType {
		COLUMN_HEADER, COLUMN_VALUE;
	}

	public enum SimpleType {
		TABLE_TITLE, PAGE_TITLE, BEFORE, AFTER, CONTEXT_BEFORE;

	}

	private final TimestampType type;

	private final SimpleType simpleType;
	private final CascadedType cascadedType;

	private final Integer cascadedTypePosition;
	private final Integer timeStampPosition;
	private final Integer propogatedYear;

	private TimestampIdentifier(TimestampType type, SimpleType simpleType, CascadedType cascadedType, Integer cascadedTypePosition, Integer timeStampPosition,
			Integer propagatedYear) {
		super();
		this.type = type;
		this.simpleType = simpleType;
		this.cascadedType = cascadedType;
		this.cascadedTypePosition = cascadedTypePosition;
		this.timeStampPosition = timeStampPosition;
		this.propogatedYear = propagatedYear;
	}

	public static TimestampIdentifier createSimpleIdentifier(SimpleType simpleType, int timeStampPosition) {
		return new TimestampIdentifier(TimestampType.SIMPLE, simpleType, null, null, timeStampPosition, null);

	}

	public static TimestampIdentifier createCascadedIdentifier(CascadedType cascadedType, int cascadedTypePosition, int timeStampPosition) {
		return new TimestampIdentifier(TimestampType.CASCADED, null, cascadedType, cascadedTypePosition, timeStampPosition, null);

	}

	public static TimestampIdentifier createPropagatedIdentifier(Integer propagatedYear) {
		return new TimestampIdentifier(TimestampType.PROPAGATED, null, null, null, null, propagatedYear);

	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((cascadedType == null) ? 0 : cascadedType.hashCode());
		result = prime * result + ((cascadedTypePosition == null) ? 0 : cascadedTypePosition.hashCode());
		result = prime * result + ((propogatedYear == null) ? 0 : propogatedYear.hashCode());
		result = prime * result + ((simpleType == null) ? 0 : simpleType.hashCode());
		result = prime * result + ((timeStampPosition == null) ? 0 : timeStampPosition.hashCode());
		result = prime * result + ((type == null) ? 0 : type.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		TimestampIdentifier other = (TimestampIdentifier) obj;
		if (cascadedType != other.cascadedType)
			return false;
		if (cascadedTypePosition == null) {
			if (other.cascadedTypePosition != null)
				return false;
		} else if (!cascadedTypePosition.equals(other.cascadedTypePosition))
			return false;
		if (propogatedYear == null) {
			if (other.propogatedYear != null)
				return false;
		} else if (!propogatedYear.equals(other.propogatedYear))
			return false;
		if (simpleType != other.simpleType)
			return false;
		if (timeStampPosition == null) {
			if (other.timeStampPosition != null)
				return false;
		} else if (!timeStampPosition.equals(other.timeStampPosition))
			return false;
		if (type != other.type)
			return false;
		return true;
	}

	public Integer getPropogatedYear() {
		return propogatedYear;
	}

	public TimestampType getType() {
		return type;
	}

	public SimpleType getSimpleType() {
		return simpleType;
	}

	public CascadedType getCascadedType() {
		return cascadedType;
	}

	public Integer getCascadedTypePosition() {
		return cascadedTypePosition;
	}

	public Integer getTimeStampPosition() {
		return timeStampPosition;
	}

	@Override
	public String toString() {
		return "TimestampIdentifier [type=" + type + ", simpleType=" + simpleType + ", cascadedType=" + cascadedType + ", cascadedTypePosition="
				+ cascadedTypePosition + ", timeStampPosition=" + timeStampPosition + ", propogatedYear=" + propogatedYear + "]";
	}

}
