package de.mannheim.uni.expansion.data.tables.reader;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.ConcurrentLinkedQueue;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.data.tables.repository.memory.AnnotatedTablesRepositoryImplMemory;
import de.mannheim.uni.expansion.utils.PersistentParallel6;

public class AnnotatedTableFilterRunner {

	@Inject
	public AnnotatedTableFilterRunner() {

	}

	public List<AnnotatedTablesReaderData> asList(Collection<AnnotatedTablesReaderData> list, List<AnnotatedTableFilter> filters) {

		ConcurrentLinkedQueue<AnnotatedTablesReaderData> queue = new ConcurrentLinkedQueue<>();
		PersistentParallel6.runAndFinish(list, (table) -> {

			for (AnnotatedTableFilter filter : filters) {
				Optional<AnnotatedTablesReaderData> opt = filter.filter(table);
				if (opt.isPresent()) {
					table = opt.get();
				} else {
					return;
				}
			}

			queue.add(table);

		}, this.getClass());

		return new ArrayList<>(queue);
	}

	public List<AnnotatedTablesReaderData> asList(Collection<AnnotatedTablesReaderData> list, AnnotatedTableFilter filter) {
		List<AnnotatedTableFilter> filters = new ArrayList<>();
		filters.add(filter);
		return asList(list, filters);
	}

	public List<AnnotatedTablesReaderData> asList(AnnotatedTablesRepository repository, List<AnnotatedTableFilter> filters) {
		return asList(repository.getTableReaderData(), filters);
	}

	public List<AnnotatedTablesReaderData> asList(AnnotatedTablesRepository repository, AnnotatedTableFilter filter) {
		return asList(repository.getTableReaderData(), filter);
	}

	public AnnotatedTablesRepository asRepository(Collection<AnnotatedTablesReaderData> list, List<AnnotatedTableFilter> filters) {
		return new AnnotatedTablesRepositoryImplMemory(asList(list, filters));
	}

	public AnnotatedTablesRepository asRepository(Collection<AnnotatedTablesReaderData> list, AnnotatedTableFilter filter) {
		return new AnnotatedTablesRepositoryImplMemory(asList(list, filter));
	}

	public AnnotatedTablesRepository asRepository(AnnotatedTablesRepository repository, List<AnnotatedTableFilter> filters) {
		return new AnnotatedTablesRepositoryImplMemory(asList(repository.getTableReaderData(), filters));
	}

	public AnnotatedTablesRepository asRepository(AnnotatedTablesRepository repository, AnnotatedTableFilter filter) {
		return new AnnotatedTablesRepositoryImplMemory(asList(repository.getTableReaderData(), filter));
	}

}
