package de.mannheim.uni.expansion.data.tables.reader;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;

import javax.inject.Inject;

import com.google.common.collect.HashBasedTable;
import com.google.common.io.LineReader;

import au.com.bytecode.opencsv.CSVReader;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId.FileEnding;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.DomainSource;
import de.mannheim.uni.expansion.data.tables.model.InstanceIndexComposite;
import de.mannheim.uni.expansion.data.tables.model.PropertyIndexComposite;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableFactory;
import de.mannheim.uni.expansion.data.tables.model.TableFactory.TableType;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.tableprocessor.model.Table;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;

public class AnnotatedTableParserCSV implements AnnotatedTableParser {
	private final TableFactory factory;

	private final KnowledgeBase dbPediaRepository;
	private final TableConverter tableConverter;

	// FIXME: Convert to nonblocking using threadlocal
	private final AtomicInteger ai = new AtomicInteger(0);

	@Inject
	public AnnotatedTableParserCSV(TableFactory factory, KnowledgeBase dbPediaRepository, TableConverter tableConverter) {
		this.factory = factory;
		this.dbPediaRepository = dbPediaRepository;
		this.tableConverter = tableConverter;
	}

	// public AnnotatedTablesReaderData get(File file, FileEnding fileEnding,
	// TableType tableType) throws IOException, InvalidTableException {
	// return get(new FileReader(file), new FileReader(file), file.getName(),
	// fileEnding, tableType);
	//
	// }

	public AnnotatedTablesReaderData get(byte[] content, String name, FileEnding fileEnding) throws IOException, InvalidTableException {

		if (name.equals("1346876860567_1346912847079_425.arc6182731458050535848#94689835_0_3430466309574287599.csv")) {
			int sdf = 123;
		}
		InputStreamReader reader2 = new InputStreamReader(new ByteArrayInputStream(content));
		InputStreamReader reader3 = new InputStreamReader(new ByteArrayInputStream(content));
		try {

			String s = new String(content);

			Pair<AnnotatedTableMetaData, Integer> mdPair = readMetaData(content, name, fileEnding);
			AnnotatedTableMetaData metaData = mdPair.getFirst();
			Integer headerLength = mdPair.getSecond();
			AnnotatedTableId id = new AnnotatedTableId(name.substring(0, name.length() - 4));

			com.google.common.collect.Table<RowIdx, ColumnIdx, String> stringTable = HashBasedTable.create();
			CSVReader csvReaderString = new CSVReader(reader3, ',', '"', headerLength);
			for (ListEntry<String[]> line : ListIterable.iterable(csvReaderString.readAll())) {
				for (int i = 0; i < line.getValue().length; i++) {
					stringTable.put(new RowIdx(line.getIndex()), new ColumnIdx(i), line.getValue()[i]);
				}
			}
			csvReaderString.close();

			if (name.equals("1346876860567_1346912847079_425.arc6182731458050535848#94689835_0_3430466309574287599.csv")) {
				int sdf = 123;
			}

			CSVReader csvReader = new CSVReader(reader2, ',', '"', headerLength);
			Table table = factory.readTable(TableType.WEB_TABLE, id, metaData, csvReader.readAll());

			csvReader.close();

			if (name.equals("1346876860567_1346912847079_425.arc6182731458050535848#94689835_0_3430466309574287599.csv")) {
				int sdf = 123;
			}

			AnnotatedTable annotatedTable = tableConverter.getTable(table, stringTable);

			if (name.equals("1346876860567_1346912847079_425.arc6182731458050535848#94689835_0_3430466309574287599.csv")) {
				int sdf = 123;
			}
			return new AnnotatedTablesReaderData(id, metaData, annotatedTable);
		} finally {
			reader2.close();
		}

	}

	private Pair<AnnotatedTableMetaData, Integer> readMetaData(byte[] content, String name, FileEnding fileEnding) throws IOException, InvalidTableException {
		InputStreamReader reader = new InputStreamReader(new ByteArrayInputStream(content));
		try {
			LineReader lr = new LineReader(reader);

			DomainSource source = null;
			ClassId dbPediaClass = null;
			Double classConf = null;
			Integer numHeaderRows = null;
			Map<Integer, List<PropertyId>> properties = null;
			Map<PropertyIndexComposite, Double> propertiesScores = null;
			Map<Integer, List<InstanceId>> instances = null;
			Map<InstanceIndexComposite, Double> instancesScores = null;
			Integer keyColumn = null;

			String[] instancesArrayTemp = null;
			int lineCount = 0;
			for (String line = lr.readLine(); line != null; line = lr.readLine()) {
				if (line.startsWith("#")) {
					lineCount++;

					String[] parts = line.substring(1, line.length()).split("=", 2);

					switch (parts[0]) {
					case "source":
						source = new DomainSource(parts[1]);
						break;
					case "class":
						dbPediaClass = ClassId.fromString(parts[1], fileEnding);
						break;
					case "classConf":
						classConf = new Double(parts[1]);
						break;
					case "numHeaderRows":
						numHeaderRows = new Integer(parts[1]);
						break;
					case "properties": {
						String[] propertiesArray = parts[1].substring(1, parts[1].length() - 1).split("\\|", -1);

						properties = new HashMap<Integer, List<PropertyId>>();
						for (int i = 0; i < propertiesArray.length; i++) {

							if (propertiesArray[i].equals("")) {
								properties.put(i, new ArrayList<PropertyId>());
							} else {
								ai.incrementAndGet();
								List<PropertyId> listOfProperties = new ArrayList<PropertyId>();

								if (dbPediaRepository == null) {
									listOfProperties.add(new PropertyId(propertiesArray[i]));
									properties.put(i, listOfProperties);
								} else {
									PropertyId propertyId = new PropertyId(propertiesArray[i]);
									listOfProperties.add(propertyId);
									// if
									// (dbPediaRepository.hasProperty(dbPediaClass,
									// propertyId)) {
									//
									// } else {
									// // TODO: Remove this workaround
									// String[] split =
									// propertiesArray[i].split("/");
									// propertyId =
									// dbPediaRepository.getPropertyFromShortName(dbPediaClass,
									// split[split.length - 1]);
									//
									// if (propertyId != null) {
									// listOfProperties.add(propertyId);
									// } else {
									// //
									// System.out.println(dbPediaClass.toString()
									// // + "::" + propertiesArray[i]);
									// }
									//
									// }

									properties.put(i, listOfProperties);
								}

							}

						}

					}

						break;
					case "propertyScores": {
						String[] propertiesScoresArray = parts[1].substring(1, parts[1].length() - 1).split("\\|", -1);

						propertiesScores = new HashMap<PropertyIndexComposite, Double>();
						for (int i = 0; i < propertiesScoresArray.length; i++) {

							if (propertiesScoresArray[i].trim().equals("") || properties.get(i).size() == 0) {
								// propertiesScores.put(i, null);
							} else {
								PropertyId propertyId = properties.get(i).get(0);
								PropertyIndexComposite composite = new PropertyIndexComposite(i, propertyId);
								try {
									propertiesScores.put(composite, new Double(propertiesScoresArray[i]));
								} catch (NumberFormatException e) {
									propertiesScores.put(composite, 0.0);
								}
							}

						}

					}

						break;
					case "instances": {
						String[] instancesArray = parts[1].substring(1, parts[1].length() - 1).split("\\|", -1);
						instancesArrayTemp = instancesArray;
						instances = new HashMap<Integer, List<InstanceId>>();
						for (int i = 0; i < (instancesArray.length); i++) {

							if (instancesArray[i].equals("")) {
								instances.put(i + numHeaderRows, new ArrayList<InstanceId>());
							} else {
								List<InstanceId> listOfInstances = new ArrayList<InstanceId>();
								listOfInstances.add(new InstanceId(instancesArray[i]));
								instances.put(i + numHeaderRows, listOfInstances);
							}

						}
					}
						break;
					case "instanceScores": {
						String[] instancesScoresArray = parts[1].substring(1, parts[1].length() - 1).split("\\|", -1);

						instancesScores = new HashMap<InstanceIndexComposite, Double>();
						for (int i = 0; i < (instancesScoresArray.length); i++) {

							if (instancesScoresArray[i].trim().equals("")) {
								// instancesScores.put(i + numHeaderRows, null);
							} else {
								List<InstanceId> list = instances.get(i + numHeaderRows);
								try {

									InstanceId instanceId = list.get(0);
									InstanceIndexComposite composite = new InstanceIndexComposite(i + numHeaderRows, instanceId);
									instancesScores.put(composite, new Double(instancesScoresArray[i]));
								} catch (Exception e) {
									e.printStackTrace();
								}
							}

						}

					}
						break;
					case "keyColumn":
						keyColumn = new Integer(parts[1]);
						break;
					}
				} else {
					break;
				}
			}

			if (source == null || dbPediaClass == null || classConf == null || numHeaderRows == null || properties == null || propertiesScores == null
					|| instances == null || instancesScores == null || keyColumn == null) {
				throw new InvalidTableException(name);
			}

			return new Pair<AnnotatedTableMetaData, Integer>(new AnnotatedTableMetaData(source, dbPediaClass, classConf, numHeaderRows, instances.size(),
					properties, propertiesScores, instances, instancesScores, keyColumn), lineCount);
		} finally {
			reader.close();
		}

	}

	public int getColumnCount() {
		return this.ai.get();
	}

	@Override
	public AnnotatedTableMetaData getMetaData(byte[] content, String name, FileEnding fileEnding) throws IOException, InvalidTableException {
		return readMetaData(content, name, fileEnding).getFirst();
	}

}
