package de.mannheim.uni.expansion.data.tables.reader;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.atomic.AtomicInteger;

import javax.inject.Inject;

import com.google.common.collect.HashBasedTable;
import com.google.gson.Gson;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId.FileEnding;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.DomainSource;
import de.mannheim.uni.expansion.data.tables.model.InstanceIndexComposite;
import de.mannheim.uni.expansion.data.tables.model.PropertyIndexComposite;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.timestamps.HeidelTimeWrapper;
import de.mannheim.uni.tableprocessor.IO.JsonTableParser;
import de.mannheim.uni.tableprocessor.model.Table;
import de.mannheim.uni.tableprocessor.model.json.AnnotatedTable;
import de.mannheim.uni.tableprocessor.model.json.TableData.TableOrientation;
import de.mannheim.uni.utils.data.Pair;

public class AnnotatedTableParserJSON implements AnnotatedTableParser {

	private final JsonTableParser jsonTableParser = new JsonTableParser();
	private final Gson gson = new Gson();
	private final HeidelTimeWrapper heidelTimeWrapper;

	// FIXME: Convert to nonblocking using threadlocal
	private final AtomicInteger ai = new AtomicInteger(0);
	private final TableConverter tableConverter;

	@Inject
	public AnnotatedTableParserJSON(HeidelTimeWrapper heidelTimeWrapper, TableConverter tableConverter) {
		this.heidelTimeWrapper = heidelTimeWrapper;
		this.tableConverter = tableConverter;
	}

	@Override
	public AnnotatedTablesReaderData get(byte[] content, String name, FileEnding fileEnding) throws IOException, InvalidTableException {
		InputStreamReader reader = new InputStreamReader(new ByteArrayInputStream(content));
		try {
			AnnotatedTable data = gson.fromJson(reader, AnnotatedTable.class);
			if (!TableOrientation.HORIZONTAL.equals(data.getTable().getTableOrientation())) {
				return null;
			}
			if (!data.getMapping().getMappedProperties().isEmpty()) {
				int sdf = 123;
			}
			AnnotatedTableId id = new AnnotatedTableId(name.substring(0, name.length() - 5));

			com.google.common.collect.Table<RowIdx, ColumnIdx, String> stringTable = HashBasedTable.create();

			String[][] relation = data.getTable().getRelation();
			for (int c = 0; c < relation.length; c++) {
				for (int r = 0; r < relation[c].length; r++) {
					stringTable.put(new RowIdx(r), new ColumnIdx(c), relation[c][r]);
				}
			}

			Table table = jsonTableParser.parseJson(data.getTable(), name);

			if (table == null) {
				return null;
			}

			// table.setFullPath(name);
			AnnotatedTableMetaData md = getMetaData(data);

			return new AnnotatedTablesReaderData(id, md, tableConverter.getTable(table, stringTable));
		} catch (NullPointerException e) {
			e.printStackTrace();
			throw e;
		} finally {
			reader.close();
		}

	}

	// public Table parseJson(TableData data, TableMapping mapping, String
	// fileName) {
	//// UnitParser.readInUnits();
	//
	// if (data.getTableType() != TableType.RELATION) {
	// return null;
	// }
	//
	//
	//
	// // create a new table
	// Table t = new Table();
	//
	// t.setSource(data.getUrl());
	// t.setContextTimestamp(data.getTableContextTimeStampBeforeTable());
	// t.setContextBeforeTable(data.getTextBeforeTable());
	// t.setContextAfterTable(data.getTextAfterTable());
	// t.setHeader(fileName);
	// t.setPageTitle(data.getPageTitle());
	// t.setTableTitle(data.getTitle());
	//
	// // determine the total number of rows
	//// int ttlNumRows = 0;
	//// for (int col = 0; col < data.getRelation().length; col++) {
	//// ttlNumRows = Math.max(ttlNumRows, data.getRelation()[col].length);
	//// }
	// t.setTotalNumOfRows(mapping.getNumHeaderRows());
	// final int numRowsToSkip = data.getNumberOfHeaderRows();
	// //final int numRowsToSkip = 1;
	// t.setNumHeaderRows(numRowsToSkip);
	//
	// List<TableColumn> tableColumns = new ArrayList<>();
	//
	//
	// // create the table columns
	// List<Pair<TableColumnBuilder, String[]>> columns = getColumnData(data,
	// t);
	//
	//
	//
	// // fill the columns with values
	// new Parallel<Pair<TableColumnBuilder, String[]>>(1).tryForeach(columns,
	// new Consumer<Pair<TableColumnBuilder, String[]>>() {
	// @Override
	// public void execute(Pair<TableColumnBuilder, String[]> parameter) {
	//
	// List<Integer> intStream = new ArrayList<>();
	// for (int row = numRowsToSkip; row < parameter.getSecond().length; row++)
	// {
	// intStream.add(row);
	// }
	// PersistentParallel6.runAndFinish(intStream, (row) -> {
	// String value = parameter.getSecond()[row];
	//
	// processValue(parameter.getFirst(), value, row);
	// }, this.getClass(), intStream.size() < 500);
	//
	//
	// }
	// });
	//
	// // complete the column construction (decide final data type, etc.)
	// for (Pair<TableColumnBuilder, String[]> p : columns) {
	// p.getFirst().buildColumn();
	// }
	//
	// TableKeyIdentifier keyIdentifier = new TableKeyIdentifier();
	// keyIdentifier.identifyKeys(t);
	//
	//
	//
	// return t;
	// }

	private AnnotatedTableMetaData getMetaData(AnnotatedTable data) {

		DomainSource source = new DomainSource(data.getTable().getUrl());
		ClassId classId = new ClassId(data.getMapping().getMappedClass().getFirst());
		double classConf = data.getMapping().getMappedClass().getSecond();
		int numHeaderRows = data.getTable().getNumberOfHeaderRows();

		int numberOfRows = 0;

		String[][] relation = data.getTable().getRelation();
		for (int c = 0; c < relation.length; c++) {
			int rowLength = relation[c].length - numHeaderRows;
			if (rowLength > numberOfRows)
				numberOfRows = rowLength;
		}

		Map<Integer, List<PropertyId>> properties = new HashMap<>();
		Map<PropertyIndexComposite, Double> propertiesScores = new HashMap<>();

		for (Entry<Integer, Pair<String, Double>> e : data.getMapping().getMappedProperties().entrySet()) {
			PropertyId propertyId = PropertyId.fromCombinedString(e.getValue().getFirst());
			List<PropertyId> mappingList = CHMHelper.getFromMap(properties, e.getKey(), () -> new ArrayList<>());
			mappingList.add(propertyId);

			propertiesScores.put(new PropertyIndexComposite(e.getKey(), propertyId), e.getValue().getSecond());
		}

		Map<Integer, List<InstanceId>> instances = new HashMap<>();
		Map<InstanceIndexComposite, Double> instancesScores = new HashMap<>();

		for (Entry<Integer, Pair<String, Double>> e : data.getMapping().getMappedInstances().entrySet()) {
			InstanceId instanceId = InstanceId.fromCombinedString(e.getValue().getFirst());
			List<InstanceId> mappingList = CHMHelper.getFromMap(instances, e.getKey(), () -> new ArrayList<>());
			mappingList.add(instanceId);

			instancesScores.put(new InstanceIndexComposite(e.getKey(), instanceId), e.getValue().getSecond());
		}

		int keyColumn = data.getTable().getKeyColumnIndex();

		List<Integer> beforeTextYears = new ArrayList<>();
		List<Integer> afterTextYears = new ArrayList<>();
		List<Integer> contextBeforeYears = new ArrayList<>();
		List<Integer> tableTitleYears = new ArrayList<>();
		List<Integer> pageTitleYears = new ArrayList<>();

		Map<Integer, List<Integer>> propertiesYears = new HashMap<>();
		//
		// if (heidelTimeWrapper != null) {
		// beforeTextYears =
		// heidelTimeWrapper.getYears(data.getTable().getTextBeforeTable());
		// afterTextYears =
		// heidelTimeWrapper.getYears(data.getTable().getTextAfterTable());
		// contextBeforeYears =
		// heidelTimeWrapper.getYears(data.getTable().getTableContextTimeStampBeforeTable());
		// tableTitleYears =
		// heidelTimeWrapper.getYears(data.getTable().getTitle());
		// pageTitleYears =
		// heidelTimeWrapper.getYears(data.getTable().getPageTitle());
		//
		// String[] columHeaders = data.getTable().getColumnHeaders();
		// for (int i = 0; i < columHeaders.length; i++) {
		// List<Integer> columnYear =
		// heidelTimeWrapper.getYears(columHeaders[i]);
		// if (columnYear != null) {
		// propertiesYears.put(i, columnYear);
		// }
		// }
		// }
		return new AnnotatedTableMetaData(source, classId, classConf, numHeaderRows, numberOfRows, properties, propertiesScores, instances, instancesScores,
				keyColumn, beforeTextYears, afterTextYears, contextBeforeYears, tableTitleYears, pageTitleYears, propertiesYears,
				data.getTable().getPageTitle(), data.getTable().getTitle(), data.getTable().getTableContextTimeStampBeforeTable(),
				data.getTable().getTextBeforeTable(), data.getTable().getTextAfterTable());

	}

	@Override
	public AnnotatedTableMetaData getMetaData(byte[] content, String name, FileEnding fileEnding) throws IOException, InvalidTableException {
		InputStreamReader reader = new InputStreamReader(new ByteArrayInputStream(content));
		try {
			AnnotatedTable data = gson.fromJson(reader, AnnotatedTable.class);
			return getMetaData(data);
		} finally {
			reader.close();
		}
	}

	@Override
	public int getColumnCount() {
		return this.ai.get();
	}

}
