package de.mannheim.uni.expansion.data.tables.reader;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Consumer;

import javax.inject.Named;

import org.apache.commons.compress.archivers.tar.TarArchiveEntry;
import org.apache.commons.compress.archivers.tar.TarArchiveInputStream;
import org.apache.commons.compress.compressors.gzip.GzipCompressorInputStream;

import com.esotericsoftware.kryo.Kryo;
import com.esotericsoftware.kryo.Registration;
import com.esotericsoftware.kryo.io.Input;

import de.mannheim.uni.expansion.data.DataNamedProperties;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderDataOld;
import de.mannheim.uni.expansion.utils.BatchConsumerProducer;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeNoReferenceKryo;

public class AnnotatedTablesKryoReader implements AnnotatedTablesReader {

	private final String folder;

	private final ThreadSafeNoReferenceKryo threadSafeKryo;

	// private final PersistentParallel3<byte[]> kryoPersistentP;

	// private final PersistentParallel3<AnnotatedTablesReaderData>
	// matchingPersistentP = new PersistentParallel3<>("RUN COLUMN MATCHING",
	// 40);

	private final boolean references;

	private final TableConverter tableConverter;

	public AnnotatedTablesKryoReader(@Named(DataNamedProperties.CACHE_INPUT_FOLDER) String cacheInputFolder, TableConverter tableConverter, boolean references,
			ThreadSafeNoReferenceKryo threadSafeKryo) {
		this.folder = cacheInputFolder;
		this.tableConverter = tableConverter;
		this.threadSafeKryo = threadSafeKryo;

		this.references = references;

		// if (references) {
		// kryoPersistentP = new PersistentParallel3<>("READ GZ TABLE ARCHIVE",
		// 400);
		// } else {
		// kryoPersistentP = new PersistentParallel3<>("READ GZ TABLE ARCHIVE",
		// 400);
		// }

	}

	private List<byte[]> readGzFile(File file) throws IOException {
		List<byte[]> contents = new ArrayList<>();

		TarArchiveInputStream tarStream = new TarArchiveInputStream(new GzipCompressorInputStream(new FileInputStream(file)));

		TarArchiveEntry entry;

		while ((entry = tarStream.getNextTarEntry()) != null) {

			byte[] content = new byte[(int) entry.getSize()];

			tarStream.read(content, 0, content.length);

			contents.add(content);
			// String name = entry.getName().split("/")[1];

		}

		tarStream.close();

		return contents;
	}

	@Override
	public void runForEachTable(List<AnnotatedTableFilter> filters, Consumer<AnnotatedTablesReaderData> fullConsumer) {
		runForEachTable(filters, fullConsumer, 40);
	}

	@Override
	public void runForEachTable(Consumer<AnnotatedTablesReaderData> fullConsumer) {
		runForEachTable(null, fullConsumer, 40);

	}

	public void runForEachTable(Consumer<AnnotatedTablesReaderData> fullConsumer, int batchSize, boolean reportProgress) {
		runForEachTable(null, fullConsumer, batchSize, reportProgress);

	}

	public void runForEachTable(List<AnnotatedTableFilter> filters, Consumer<AnnotatedTablesReaderData> fullConsumer, int batchSize, boolean reportProgress) {

		final BatchConsumerProducer<byte[]> batchConsumerProducer;
		batchConsumerProducer = BatchConsumerProducer.create("READ KRYO", (content) -> {
			Input input = new Input(new ByteArrayInputStream(content));
			AnnotatedTablesReaderData dataNew = null;

			// Registration registration = getKryo().readClass(input);
			Kryo kryo = threadSafeKryo.getKryo();
			dataNew = kryo.readObject(input, AnnotatedTablesReaderData.class);
			if (dataNew == null) {
				Registration registration = kryo.readClass(input);

				boolean old = false;
				if (registration != null) {
					Class type = registration.getType();
					if (type.isAssignableFrom(AnnotatedTablesReaderData.class)) {
						dataNew = kryo.readObject(input, AnnotatedTablesReaderData.class);

					} else if (type.isAssignableFrom(AnnotatedTablesReaderDataOld.class)) {
						old = true;

					}
				} else {
					old = true;
				}

				if (dataNew == null) {

					// if (old) {
					input.close();
					input = new Input(new ByteArrayInputStream(content));
					AnnotatedTablesReaderDataOld data = null;
					try {

						int sadf = 123;
					} catch (Exception e) {
						e.printStackTrace();
					}
					input.close();

					dataNew = new AnnotatedTablesReaderData(data.getId(), data.getMetaData(), tableConverter.getTable(data.getTableData(), null));

					// } else {
					// throw new IllegalStateException(
					// "This is an unexpected state for this table reader: if
					// the
					// class is written it must be AnnotateddTablesReaderData");
					// }
				}
			}

			if (dataNew.getMetaData().getNumberOfRows() == 0) {
				int rowSizeFromMD = dataNew.getMetaData().getAllInstances().size();
				AnnotatedTableMetaData metaData = dataNew.getMetaData().rebuilder().setNumberOfRows(rowSizeFromMD).build();
				// int rowSizeFromTable =
				// dataNew.getTableData().getTable().rowKeySet().size();
				// if (rowSizeFromTable != metaData.getNumberOfRows()) {
				// throw new IllegalStateException("Row Size from Table differs
				// from dervies Row Size");
				// }
				dataNew = new AnnotatedTablesReaderData(dataNew.getId(), metaData, dataNew.getTableData());
			}

			// AnnotatedTablesReaderDataOld data = getKryo().readObject(input,
			// AnnotatedTablesReaderDataOld.class);
			// AnnotatedTablesReaderData dataNew = new
			// AnnotatedTablesReaderData(data.getId(), data.getMetaData(),
			// tableConverter.getTable(data.getTableData()));
			input.close();

			if (fullConsumer != null) {
				if (filters != null) {
					for (AnnotatedTableFilter filter : filters) {
						Optional<AnnotatedTablesReaderData> oTable = filter.filter(dataNew);
						if (oTable.isPresent()) {
							dataNew = oTable.get();
						} else {
							return;
						}
					}
				}

				if (!dataNew.getMetaData().getAllProperties().isEmpty()) {
					int sdf = 123;
				}

				fullConsumer.accept(dataNew);
			}

		}, 2000, 2);

		// PersistentParallel3<File> reader = new PersistentParallel3<>("READ
		// KRYO GZ FILE", 1, 2, false);

		File[] files = new File(folder).listFiles();

		try {
			AtomicInteger ai = new AtomicInteger(0);
			PersistentParallel6.runAndFinish(Arrays.asList(files), (file) -> {
				System.out.println("Now reading table " + ai.incrementAndGet() + " out of " + files.length);

				try {

					batchConsumerProducer.addBatch(readGzFile(file));
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}, "READ KRYO GZ FILE", 2);

			batchConsumerProducer.finish();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		// try {
		// PersistentParallel.forLoop(0, files.length, new Consumer<Integer>() {
		//
		// @Override
		// public void accept(Integer i) {
		//
		// try {
		//
		// TarArchiveInputStream tarStream = new TarArchiveInputStream(new
		// GzipCompressorInputStream(new FileInputStream(files[i])));
		//
		// TarArchiveEntry entry = tarStream.getNextTarEntry();
		//
		// while ((entry = tarStream.getNextTarEntry()) != null) {
		//
		// byte[] content = new byte[(int) entry.getSize()];
		//
		// tarStream.read(content, 0, content.length);
		//
		// // String name = entry.getName().split("/")[1];
		//
		// Input input = new Input(new ByteArrayInputStream(content));
		// AnnotatedTablesReaderData data = getKryo().readObject(input,
		// AnnotatedTablesReaderData.class);
		// input.close();
		//
		// // Table td = magnitudeFixer.get(data.getMetaData(),
		// // data.getTableData());
		//
		// data = new AnnotatedTablesReaderData(data.getId(),
		// data.getMetaData(), data.getTableData());
		// if (fullConsumer != null) {
		// fullConsumer.accept(data);
		// }
		//
		// }
		//
		// tarStream.close();
		// } catch (IOException e) {
		// // TODO Auto-generated catch block
		// e.printStackTrace();
		// }
		// }

		// }, "READ TABLES KRYO");
		// } catch (Exception e) {
		// e.printStackTrace();
		// e.getCause().printStackTrace();
		// }

	}

	public void runForEachTable(List<AnnotatedTableFilter> filters, Consumer<AnnotatedTablesReaderData> fullConsumer, int batchSize) {
		runForEachTable(filters, fullConsumer, batchSize, false);

	}

}
