package de.mannheim.uni.expansion.data.tables.reader;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.List;
import java.util.function.Consumer;

import org.apache.commons.io.IOUtils;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId.FileEnding;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.utils.PersistentParallel;

public class AnnotatedTablesSimpleReader implements AnnotatedTablesReader {

	private final String folder;
	private final File[] files;
	private final AnnotatedTableParser parser;
	private final FileEnding fileEnding;

	public AnnotatedTablesSimpleReader(String folder, FileEnding fileEnding, AnnotatedTableParser parser) {
		this.fileEnding = fileEnding;
		this.folder = folder;

		files = new File(folder).listFiles();

		this.parser = parser;

	}

	public AnnotatedTablesReaderData get(int i) throws IOException, InvalidTableException {
		byte[] content = IOUtils.toByteArray(new FileReader(files[i]));

		return parser.get(content, files[i].getName(), fileEnding);
	}

	public int length() {
		return files.length;
	}

	public String getFolder() {
		return folder;
	}

	public void runForEachTable(Consumer<AnnotatedTablesReaderData> consumer) {

		try {
			PersistentParallel.forLoop(0, this.length(), new Consumer<Integer>() {

				public void accept(Integer index) {
					try {
						AnnotatedTablesReaderData data;

						data = get(index);

						consumer.accept(data);
					} catch (IOException e) {
						e.printStackTrace();
						throw new RuntimeException(e);
					} catch (InvalidTableException e) {
						System.out.println("WARNING: Invalid Header in File: " + e.getTable());
					}

				}

			}, "Annotated Simple Reader");
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	@Override
	public void runForEachTable(List<AnnotatedTableFilter> filters, Consumer<AnnotatedTablesReaderData> fullConsumer) {
		throw new UnsupportedOperationException();

	}

}
