package de.mannheim.uni.expansion.data.tables.reader;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.zip.GZIPOutputStream;

import javax.inject.Inject;
import javax.inject.Named;

import org.apache.commons.compress.archivers.tar.TarArchiveEntry;
import org.apache.commons.compress.archivers.tar.TarArchiveOutputStream;

import com.esotericsoftware.kryo.io.Input;
import com.esotericsoftware.kryo.io.Output;

import de.mannheim.uni.expansion.data.DataNamedProperties;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.utils.ConsumerHelper;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeNoReferenceKryo;

public class AnnotatedTablesWriter {

	private int fileCount = 0;
	private int entryCount = 0;
	private int groupNum = -1;
	private int perGroup = 20000;
	private TarArchiveOutputStream tarFile;
	private String folder;
	private final ThreadSafeNoReferenceKryo threadSafeKryo;

	private ConsumerHelper<AnnotatedTablesReaderData> consumerHelper;

	@Inject
	public AnnotatedTablesWriter(@Named(DataNamedProperties.CACHE_OUTPUT_FOLDER) String folder, ThreadSafeNoReferenceKryo threadSafeKryo) {
		this.threadSafeKryo = threadSafeKryo;
		this.folder = folder;

	}

	public synchronized void setGroupNum(int groupNum, int numberOfFiles) {
		this.groupNum = groupNum;
		this.perGroup = numberOfFiles / groupNum;
	}

	public synchronized void writeEntry(byte[] bytes) throws IOException {

		if (entryCount == perGroup && fileCount != (groupNum - 1)) {
			finishAndClose();
			fileCount++;
			tarFile = new TarArchiveOutputStream(new GZIPOutputStream(new FileOutputStream(folder + "/tables" + fileCount + ".tar.gz")));
			entryCount = 0;
		}

		TarArchiveEntry entry = new TarArchiveEntry(String.valueOf(entryCount));
		entry.setSize(bytes.length);

		tarFile.putArchiveEntry(entry);

		tarFile.write(bytes);

		tarFile.closeArchiveEntry();

		entryCount++;
	}

	public synchronized void finishAndClose() throws IOException {

		tarFile.finish();

		tarFile.close();
	}

	public void add(AnnotatedTablesReaderData data) {
		consumerHelper.add(data);
	}

	public void finishCloseAndAwait() throws InterruptedException, IOException {
		consumerHelper.finishCloseAndAwait();
		finishAndClose();

	}

	public void writeAll(AnnotatedTablesRepository annotatedTablesRepository) {

		try {
			tarFile = new TarArchiveOutputStream(new GZIPOutputStream(new FileOutputStream(folder + "/tables" + fileCount + ".tar.gz")));

			consumerHelper = ConsumerHelper.withLinkedQueue("TABLE STORER", 1, (data) -> {

				ByteArrayOutputStream baos = new ByteArrayOutputStream();
				Output output = new Output(baos);
				this.threadSafeKryo.getKryo().writeClassAndObject(output, data);
				output.close();

				Input input = new Input(new ByteArrayInputStream(baos.toByteArray()));

				// // Registration registration = getKryo().readClass(input);
				// Kryo kryo = threadSafeKryo.getKryo();
				// Registration registration = kryo.readClass(input);
				//
				// AnnotatedTablesReaderData dataNew = null;
				//
				// boolean old = false;
				// if (registration != null) {
				// Class type = registration.getType();
				// if (type.isAssignableFrom(AnnotatedTablesReaderData.class)) {
				// dataNew = kryo.readObject(input,
				// AnnotatedTablesReaderData.class);
				//
				// } else if
				// (type.isAssignableFrom(AnnotatedTablesReaderDataOld.class)) {
				// old = true;
				//
				// }
				// } else {
				// old = true;
				// }
				//
				// if (!data.getMetaData().getProperties().isEmpty()) {
				// int sdf = 123;
				// }
				try {
					writeEntry(baos.toByteArray());
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}, new AnnotatedTablesReaderData(null, null, null));

			consumerHelper.start();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}

		this.setGroupNum(32, annotatedTablesRepository.getAllTableIds().size());

		for (AnnotatedTableId id : annotatedTablesRepository.getAllTableIds()) {
			AnnotatedTablesReaderData data = new AnnotatedTablesReaderData(id, annotatedTablesRepository.getMetaData(id),
					annotatedTablesRepository.getTable(id));
			add(data);
		}

		try {
			finishCloseAndAwait();
		} catch (InterruptedException | IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

}
