package de.mannheim.uni.expansion.data.tables.reader;

import java.io.File;
import java.util.Collection;
import java.util.concurrent.ConcurrentLinkedQueue;

import javax.inject.Inject;
import javax.inject.Named;

import de.mannheim.uni.expansion.data.DataNamedProperties;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ConcurrentKryoWriter;

public class AnnotatedTablesWriterConcurrent {

	private final ConcurrentKryoWriter writer;

	private String folder;

	@Inject
	public AnnotatedTablesWriterConcurrent(@Named(DataNamedProperties.CACHE_OUTPUT_FOLDER) String folder, ConcurrentKryoWriter writer) {
		this.writer = writer;
		this.folder = folder;

	}

	public boolean cacheExists() {
		File dirFile = new File(folder);
		if (dirFile.isDirectory()) {
			String[] dirFileList = dirFile.list();
			if (dirFileList.length > 0) {
				for (String f : dirFileList) {
					if (f.startsWith("tables")) {
						return true;
					}
				}

			}
		}
		return false;
	}

	public void writeAll(AnnotatedTablesRepository annotatedTablesRepository) {

		ConcurrentLinkedQueue<AnnotatedTablesReaderData> list = new ConcurrentLinkedQueue<>();
		for (AnnotatedTableId id : annotatedTablesRepository.getAllTableIds()) {
			AnnotatedTablesReaderData data = new AnnotatedTablesReaderData(id, annotatedTablesRepository.getMetaData(id),
					annotatedTablesRepository.getTable(id));
			list.add(data);
		}

		writer.writeAll(new AnnotatedTablesReaderData(null, null, null), list, annotatedTablesRepository.getAllTableIds().size(), folder, "tables", 32);
	}

	public void writeAll(Collection<AnnotatedTablesReaderData> data) {

		writer.writeAll(new AnnotatedTablesReaderData(null, null, null), data, data.size(), folder, "tables", 32);
	}

}
