package de.mannheim.uni.expansion.data.tables.reader;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.inject.Inject;

import com.google.common.collect.HashBasedTable;
import com.google.common.collect.ImmutableMap;

import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.utils.SparseArrayMap;
import de.mannheim.uni.tableprocessor.model.Table;
import de.mannheim.uni.tableprocessor.model.TableColumn;
import de.uni_mannheim.dws.expansion.data.ontology.model.ValueConverter;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class TableConverter {

	private ValueConverter valueConverter;

	@Inject
	public TableConverter(ValueConverter valueConverter) {
		this.valueConverter = valueConverter;
	}

	public AnnotatedTable getTable(Table old, com.google.common.collect.Table<RowIdx, ColumnIdx, String> stringTable) {
		com.google.common.collect.Table<RowIdx, ColumnIdx, AbstractObject> table = HashBasedTable.create();

		Map<ColumnIdx, ValueType> dataTypes = new HashMap<>();
		Map<ColumnIdx, AbstractObject> columnHeaders = new HashMap<>();

		for (int c = 0; c < old.getColumns().size(); c++) {
			ColumnIdx columnKey = new ColumnIdx(c);

			for (Entry<Integer, Object> e : old.getColumns().get(c).getValues().entrySet()) {
				table.put(new RowIdx(e.getKey()), columnKey, valueConverter.getAbstractObject(e.getValue()));
			}
			dataTypes.put(columnKey, valueConverter.getDataType(old.getColumns().get(c).getDataType()));

			columnHeaders.put(columnKey, valueConverter.getAbstractObject(old.getColumns().get(c).getHeader()));

		}

		ColumnIdx keyColumn = new ColumnIdx(old.getKeyIndex());

		return new AnnotatedTable(table, stringTable, dataTypes, keyColumn, columnHeaders);

	}

	public Table getTable(AnnotatedTablesReaderData old) {
		Table table = new Table();
		List<TableColumn> columns = new ArrayList<>();

		for (ColumnIdx columnIdx : old.getTableData().getColumns()) {

			ImmutableMap<RowIdx, AbstractObject> column = old.getTableData().getColumn(columnIdx);
			SparseArrayMap<Object> values = new SparseArrayMap<>();

			TableColumn tc = new TableColumn(table, columnIdx.getColumnIdx());
			tc.setDataSource(old.getMetaData().getSource().getSource());
			tc.setDataType(valueConverter.getDataType(old.getTableData().getDataType(columnIdx)));
			tc.setHeader(valueConverter.getObject(old.getTableData().getColumnHeader(columnIdx)));
			tc.setKey(columnIdx.equals(old.getTableData().getKeyColumnIndex()));

			if (column == null) {
				tc.setNumRows(0);

			} else {
				tc.setNumRows(column.size());

				for (Entry<RowIdx, AbstractObject> value : column.entrySet()) {

					values.put(value.getKey().getRowIdx(), valueConverter.getObject(value.getValue()));

				}
			}

			tc.setValues(values);

			columns.add(tc);
		}

		table.setColumns(columns);

		table.setContextAfterTable(old.getMetaData().getTextAfterTable());
		table.setContextBeforeTable(old.getMetaData().getTextBeforeTable());
		table.setContextTimestamp(old.getMetaData().getTableContextTimeStampBeforeTable());
		table.setFullPath(old.getId().getTableName());
		table.setHasKey(true);
		table.setHeader(old.getId().getTableName());
		try {
			table.setKey(columns.get(old.getMetaData().getKeyColumn()), old.getMetaData().getKeyColumn());
		} catch (Exception e) {
			e.printStackTrace();
		}
		// table.setNmNulls(nmNulls);
		table.setNumHeaderRows(old.getMetaData().getNumHeaderRows());
		table.setPageTitle(old.getMetaData().getPageTitle());
		table.setSource(old.getMetaData().getSource().getSource());
		table.setTableTitle(old.getMetaData().getTableTitle());
		table.setTotalNumOfRows(old.getMetaData().getNumberOfRows() + old.getMetaData().getNumHeaderRows());

		int sdf = 123;

		return table;

	}
}
