package de.mannheim.uni.expansion.data.tables.repository;

import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.HostId;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;

public interface AnnotatedTablesRepository {

	Set<AnnotatedTableId> ofClass(ClassId dbpediaClass);

	Set<AnnotatedTableId> hasProperty(PropertyId property);

	Set<AnnotatedTableId> hasProperties(Set<PropertyId> properties);

	Set<AnnotatedTableId> getTables(InstanceId instances);

	Set<TableRowId> getRows(InstanceId instance);

	int getNumOfMatchedColumns(InstanceId instance);

	boolean hasInstance(InstanceId instance);

	Set<AnnotatedTableId> hasInstances(Set<InstanceId> instances);

	AnnotatedTableMetaData getMetaData(AnnotatedTableId id);

	HostId getHostId(AnnotatedTableId id);

	// Map<AnnotatedTableId, String> getValues(InstanceId instance, PropertyId
	// property);

	AnnotatedTable getTable(AnnotatedTableId id);

	Set<InstanceId> getInstances();

	Set<AnnotatedTableId> getAllTableIds();

	Set<AnnotatedTableId> getAllTableIds(ClassId classId);

	Set<AnnotatedTableId> getAllTableIdsWithMatches(ClassId classId);

	Map<RowIdx, AbstractObject> getColumn(AnnotatedTableId id, ColumnIdx columnIndex);

	AbstractObject getValue(AnnotatedTableId id, RowIdx rowIdx, ColumnIdx colIdx);

	AbstractObject getKeyValue(AnnotatedTableId id, RowIdx rowIdx);

	void printStatistics();

	Integer getTimeStamp(CellId cellId, TimestampIdentifier tsi);

	Integer getTimeStamp(TableColumnId tableColumnId, TimestampIdentifier tsi);

	Set<TableRowId> getUnmatchedRowIds(Set<InstanceId> value);

	Set<TableRowId> getUnmatchedRowIds(Set<InstanceId> value, ClassId classId);

	Set<TableRowId> getUnmatchedRowIds(ClassId classId);

	Set<TableRowId> getUnmatchedRowIdsForTables(Set<AnnotatedTableId> tableId);

	Set<TableRowId> getMatchedRowsIds(Set<InstanceId> value, ClassId classId);

	// AnnotatedTableRow getTableRow(TableRowId tableRowId);

	List<String> getNeighborHood(TableRowId tableRowId);

	List<String> getLabels(AnnotatedTableId tableId);

	StringObject getLabel(TableRowId tableRowId);

	String getLabelRaw(TableRowId tableRowId);

	ClassId getClassId(AnnotatedTableId tableId);

	Set<TableRowId> getAllRowIds(ClassId wikidatafootball);

	Set<TableRowId> getAllRowIds(Set<AnnotatedTableId> tableIds);

	Optional<InstanceId> getMatchedInstanceId(TableRowId rowId);

	int getNumOfUnmatchedRowIdsForTable(AnnotatedTableId tableId);

	Set<TableRowId> getMatchedRowIdsForTables(Set<AnnotatedTableId> tableIdsSet);

	int getNumOfMatchedRowIdsForTable(AnnotatedTableId tableId);

	List<TableRowId> getRowIds(AnnotatedTableId tableId);

	Map<PropertyId, List<CellId>> getMatches(InstanceId instanceId);

	AbstractObject getValue(CellId id);

	List<AnnotatedTablesReaderData> getTableReaderData();

	List<AnnotatedTablesReaderData> getTableReaderData(Set<AnnotatedTableId> tableIds);

}
