package de.mannheim.uni.expansion.data.tables.repository;

import java.rmi.RemoteException;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import de.mannheim.uni.expansion.data.RemoteInterfaceWithCache;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.HostId;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier;
import de.mannheim.uni.expansion.data.tables.repository.remote.AnnotatedTablesRepositoryRemote;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;

public class AnnotatedTablesRepositoryImpl extends RemoteInterfaceWithCache implements AnnotatedTablesRepository {

	AnnotatedTablesRepositoryRemote remote;

	public AnnotatedTablesRepositoryImpl(AnnotatedTablesRepositoryRemote remote) {
		this.remote = remote;

	}

	@Override
	public Set<AnnotatedTableId> ofClass(ClassId dbpediaClass) {
		try {
			return remote.ofClass(dbpediaClass);
		} catch (RemoteException e) {
			e.printStackTrace();
		}
		return null;
	}

	@Override
	public Set<AnnotatedTableId> hasProperty(PropertyId property) {
		try {
			return remote.hasProperty(property);
		} catch (RemoteException e) {
			e.printStackTrace();
		}
		return null;
	}

	@Override
	public Set<AnnotatedTableId> hasProperties(Set<PropertyId> properties) {
		try {
			return remote.hasProperties(properties);
		} catch (RemoteException e) {
			e.printStackTrace();
		}
		return null;
	}

	@Override
	public Set<AnnotatedTableId> getTables(InstanceId instances) {
		try {
			return remote.getTables(instances);
		} catch (RemoteException e) {
			e.printStackTrace();
		}
		return null;
	}

	@Override
	public Set<AnnotatedTableId> hasInstances(Set<InstanceId> instances) {
		try {
			return remote.hasInstances(instances);
		} catch (RemoteException e) {
			e.printStackTrace();
		}
		return null;
	}

	ConcurrentHashMap<AnnotatedTableId, Optional<AnnotatedTableMetaData>> metaData = new ConcurrentHashMap<>();

	@Override
	public AnnotatedTableMetaData getMetaData(AnnotatedTableId id) {
		return cacheGetter(metaData, id, (key) -> {
			return remote.getMetaData(key);
		});
	}

	@Override
	public AnnotatedTable getTable(AnnotatedTableId id) {
		try {
			return remote.getTable(id);
		} catch (RemoteException e) {
			e.printStackTrace();
		}
		return null;
	}

	@Override
	public Set<InstanceId> getInstances() {
		try {
			return remote.getInstances();
		} catch (RemoteException e) {
			e.printStackTrace();
		}
		return null;
	}

	Set<AnnotatedTableId> allTableIds;

	@Override
	public Set<AnnotatedTableId> getAllTableIds() {
		try {
			if (allTableIds == null)
				allTableIds = remote.getAllTableIds();
			return allTableIds;
		} catch (RemoteException e) {
			e.printStackTrace();
		}
		return null;
	}

	@Override
	public Map<RowIdx, AbstractObject> getColumn(AnnotatedTableId id, ColumnIdx columnIndex) {
		try {
			return remote.getColumn(id, columnIndex);
		} catch (RemoteException e) {
			e.printStackTrace();
		}
		return null;
	}

	@Override
	public AbstractObject getValue(AnnotatedTableId id, RowIdx rowIdx, ColumnIdx colIdx) {
		try {
			return remote.getValue(id, rowIdx, colIdx);
		} catch (RemoteException e) {
			e.printStackTrace();
		}
		return null;
	}

	@Override
	public AbstractObject getKeyValue(AnnotatedTableId id, RowIdx rowIdx) {
		try {
			return remote.getKeyValue(id, rowIdx);
		} catch (RemoteException e) {
			e.printStackTrace();
		}
		return null;
	}

	ConcurrentHashMap<AnnotatedTableId, Optional<HostId>> hostIds = new ConcurrentHashMap<>();

	private ValueGetter<AnnotatedTableId, HostId> hostIdGetter = (key) -> {
		return remote.getHostId(key);
	};

	@Override
	public HostId getHostId(AnnotatedTableId id) {
		return cacheGetter(hostIds, id, hostIdGetter);
	}

	@Override
	public void printStatistics() {
		throw new UnsupportedOperationException();

	}

	@Override
	public Integer getTimeStamp(CellId cellId, TimestampIdentifier tsi) {
		try {
			return remote.getTimeStamp(cellId, tsi);
		} catch (RemoteException e) {
			e.printStackTrace();
		}
		return null;
	}

	ConcurrentHashMap<InstanceId, Optional<Set<TableRowId>>> rowsCache = new ConcurrentHashMap<>();

	@Override
	public Set<TableRowId> getRows(InstanceId instance) {
		return cacheGetter(rowsCache, instance, (k) -> {
			return remote.getRows(k);
		});
	}

	ConcurrentHashMap<InstanceId, Optional<Integer>> numOfMatchedColumnsCache = new ConcurrentHashMap<>();

	@Override
	public int getNumOfMatchedColumns(InstanceId instance) {
		// TODO Auto-generated method stub
		return 0;
	}

	ConcurrentHashMap<InstanceId, Optional<Boolean>> hasInstanceCache = new ConcurrentHashMap<>();

	@Override
	public boolean hasInstance(InstanceId instance) {
		// TODO Auto-generated method stub
		return false;
	}

	ConcurrentHashMap<ClassId, Optional<Set<AnnotatedTableId>>> allTableIdsCache = new ConcurrentHashMap<>();

	@Override
	public Set<AnnotatedTableId> getAllTableIds(ClassId classId) {
		// TODO Auto-generated method stub
		return null;
	}

	ConcurrentHashMap<ClassId, Optional<Set<AnnotatedTableId>>> allTableIdsWithMatchesCache = new ConcurrentHashMap<>();

	@Override
	public Set<AnnotatedTableId> getAllTableIdsWithMatches(ClassId classId) {
		// TODO Auto-generated method stub
		return null;
	}

	ConcurrentHashMap<Pair<TableColumnId, TimestampIdentifier>, Optional<Integer>> timeStampCache = new ConcurrentHashMap<>();

	@Override
	public Integer getTimeStamp(TableColumnId tableColumnId, TimestampIdentifier tsi) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public Set<TableRowId> getUnmatchedRowIds(Set<InstanceId> value) {
		try {
			return remote.getUnmatchedRowIds(value);
		} catch (RemoteException e) {
			e.printStackTrace();

		}
		return null;
	}

	@Override
	public Set<TableRowId> getUnmatchedRowIds(Set<InstanceId> value, ClassId classId) {
		try {
			return remote.getUnmatchedRowIds(value, classId);
		} catch (RemoteException e) {
			e.printStackTrace();

		}
		return null;
	}

	ConcurrentHashMap<ClassId, Optional<Set<TableRowId>>> unmatchedRowIds = new ConcurrentHashMap<>();

	@Override
	public Set<TableRowId> getUnmatchedRowIds(ClassId classId) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public Set<TableRowId> getUnmatchedRowIdsForTables(Set<AnnotatedTableId> tableId) {
		try {
			return remote.getUnmatchedRowIdsForTables(tableId);
		} catch (RemoteException e) {
			e.printStackTrace();

		}
		return null;
	}

	@Override
	public Set<TableRowId> getMatchedRowsIds(Set<InstanceId> value, ClassId classId) {
		try {
			return remote.getMatchedRowsIds(value, classId);
		} catch (RemoteException e) {
			e.printStackTrace();

		}
		return null;
	}

	ConcurrentHashMap<TableRowId, Optional<List<String>>> neighborHoodCache = new ConcurrentHashMap<>();

	@Override
	public List<String> getNeighborHood(TableRowId tableRowId) {
		// TODO Auto-generated method stub
		return null;
	}

	ConcurrentHashMap<AnnotatedTableId, Optional<List<String>>> labelsCache = new ConcurrentHashMap<>();

	@Override
	public List<String> getLabels(AnnotatedTableId tableId) {
		// TODO Auto-generated method stub
		return null;
	}

	ConcurrentHashMap<AnnotatedTableId, Optional<ClassId>> classIdsOfTablesCache = new ConcurrentHashMap<>();

	@Override
	public ClassId getClassId(AnnotatedTableId tableId) {
		// TODO Auto-generated method stub
		return null;
	}

	ConcurrentHashMap<ClassId, Optional<Set<TableRowId>>> rowIdsPerClasscache = new ConcurrentHashMap<>();

	@Override
	public Set<TableRowId> getAllRowIds(ClassId wikidatafootball) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public Set<TableRowId> getAllRowIds(Set<AnnotatedTableId> tableIds) {
		try {
			return remote.getAllRowIds(tableIds);
		} catch (RemoteException e) {
			e.printStackTrace();

		}
		return null;
	}

	ConcurrentHashMap<TableRowId, Optional<Optional<InstanceId>>> matchedInstances = new ConcurrentHashMap<>();

	@Override
	public Optional<InstanceId> getMatchedInstanceId(TableRowId rowId) {
		// TODO Auto-generated method stub
		return null;
	}

	ConcurrentHashMap<AnnotatedTableId, Optional<Integer>> numOfUnmatchedRowIdsPerTableCache = new ConcurrentHashMap<>();

	@Override
	public int getNumOfUnmatchedRowIdsForTable(AnnotatedTableId tableId) {
		// TODO Auto-generated method stub
		return 0;
	}

	@Override
	public Set<TableRowId> getMatchedRowIdsForTables(Set<AnnotatedTableId> tableIdsSet) {
		try {
			return remote.getMatchedRowIdsForTables(tableIdsSet);
		} catch (RemoteException e) {
			e.printStackTrace();

		}
		return null;
	}

	ConcurrentHashMap<AnnotatedTableId, Optional<Integer>> numOfmatchedRowIdsPerTableCache = new ConcurrentHashMap<>();

	@Override
	public int getNumOfMatchedRowIdsForTable(AnnotatedTableId tableId) {
		// TODO Auto-generated method stub
		return 0;
	}

	ConcurrentHashMap<AnnotatedTableId, Optional<List<TableRowId>>> rowIdsPerTable = new ConcurrentHashMap<>();

	@Override
	public List<TableRowId> getRowIds(AnnotatedTableId tableId) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public Map<PropertyId, List<CellId>> getMatches(InstanceId instanceId) {
		try {
			return remote.getMatches(instanceId);
		} catch (RemoteException e) {
			e.printStackTrace();

		}
		return null;
	}

	@Override
	public AbstractObject getValue(CellId id) {
		try {
			return remote.getValue(id);
		} catch (RemoteException e) {
			e.printStackTrace();

		}
		return null;
	}

	volatile List<AnnotatedTablesReaderData> tableReaderData = null;

	@Override
	public List<AnnotatedTablesReaderData> getTableReaderData() {
		try {
			if (tableReaderData == null) {
				tableReaderData = remote.getTableReaderData();
			}
		} catch (RemoteException e) {

		}

		return tableReaderData;
	}

	@Override
	public List<AnnotatedTablesReaderData> getTableReaderData(Set<AnnotatedTableId> tableIds) {
		try {
			return remote.getTableReaderData(tableIds);
		} catch (RemoteException e) {
			e.printStackTrace();

		}
		return null;
	}

	@Override
	public StringObject getLabel(TableRowId tableRowId) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public String getLabelRaw(TableRowId tableRowId) {
		// TODO Auto-generated method stub
		return null;
	}

}
