package de.mannheim.uni.expansion.data.tables.repository.memory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import org.apache.commons.lang.mutable.MutableInt;

import com.google.common.collect.ImmutableMap;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.HostId;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier.CascadedType;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier.TimestampType;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.data.tables.repository.remote.AnnotatedTablesRepositoryRemote;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;

public class AnnotatedTablesRepositoryImplMemory implements AnnotatedTablesRepositoryRemote, AnnotatedTablesRepository {

	private final ConcurrentHashMap<ClassId, Set<AnnotatedTableId>> classIndex;
	private final ConcurrentHashMap<PropertyId, Set<AnnotatedTableId>> propertyIndex;
	private final ConcurrentHashMap<InstanceId, Set<AnnotatedTableId>> instanceIndex;

	private final ConcurrentHashMap<AnnotatedTableId, AnnotatedTable> tableData;
	private final ConcurrentHashMap<AnnotatedTableId, AnnotatedTableMetaData> tableMetaData;

	public AnnotatedTablesRepositoryImplMemory(Collection<AnnotatedTablesReaderData> data) {

		this.classIndex = new ConcurrentHashMap<>();
		this.propertyIndex = new ConcurrentHashMap<>();
		this.instanceIndex = new ConcurrentHashMap<>();
		this.tableData = new ConcurrentHashMap<>();
		this.tableMetaData = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(data, (table) -> {
			tableMetaData.put(table.getId(), table.getMetaData());
			tableData.put(table.getId(), table.getTableData());

			CHMHelper.getFromCHM(classIndex, table.getMetaData().getDbPediaClass(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>()))
					.add(table.getId());

			for (PropertyId propertyId : table.getMetaData().getProperties().values()) {
				CHMHelper.getFromCHM(propertyIndex, propertyId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(table.getId());
			}

			for (InstanceId instanceId : table.getMetaData().getInstances().values()) {
				CHMHelper.getFromCHM(instanceIndex, instanceId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(table.getId());
			}

		}, this.getClass());

	}

	AnnotatedTablesRepositoryImplMemory(ConcurrentHashMap<ClassId, Set<AnnotatedTableId>> classIndex,
			ConcurrentHashMap<PropertyId, Set<AnnotatedTableId>> propertyIndex, ConcurrentHashMap<InstanceId, Set<AnnotatedTableId>> instanceIndex,
			ConcurrentHashMap<AnnotatedTableId, AnnotatedTable> tableData, ConcurrentHashMap<AnnotatedTableId, AnnotatedTableMetaData> tableMetaData) {
		this.classIndex = classIndex;
		this.propertyIndex = propertyIndex;
		this.instanceIndex = instanceIndex;
		this.tableData = tableData;
		this.tableMetaData = tableMetaData;

		Map<ClassId, MutableInt> countMap = new HashMap<>();
		int i = 0;
		for (AnnotatedTableMetaData md : tableMetaData.values()) {
			MutableInt mi = CHMHelper.getFromMap(countMap, md.getDbPediaClass(), () -> new MutableInt());
			for (List<PropertyId> list : md.getAllProperties().values()) {
				i += list.size();
				mi.increment();
			}
		}
		System.out.println("Number of mapped properties: " + i);
		System.out.println("Number of mapped properties by Class:");
		for (Entry<ClassId, MutableInt> entry : countMap.entrySet()) {
			System.out.println(entry.getKey().getClassId() + ";" + entry.getValue().toString());
		}
	}

	public Set<AnnotatedTableId> ofClass(ClassId dbpediaClass) {
		Set<AnnotatedTableId> set = classIndex.get(dbpediaClass);
		if (set == null)
			return null;
		return Collections.unmodifiableSet(set);
	}

	public Set<AnnotatedTableId> hasProperty(PropertyId property) {
		Set<AnnotatedTableId> set = propertyIndex.get(property);
		if (set == null)
			return null;
		return Collections.unmodifiableSet(set);
	}

	public Set<AnnotatedTableId> hasProperties(Set<PropertyId> properties) {
		// TODO Auto-generated method stub
		throw new IllegalStateException("NOT YET IMPLEMENTED");
	}

	public Set<AnnotatedTableId> getTables(InstanceId instances) {

		Set<AnnotatedTableId> set = instanceIndex.get(instances);
		if (set == null)
			return null;
		return Collections.unmodifiableSet(set);
	}

	public boolean hasInstance(InstanceId instances) {

		Set<AnnotatedTableId> set = instanceIndex.get(instances);
		if (set == null)
			return false;
		return true;
	}

	public Set<AnnotatedTableId> hasInstances(Set<InstanceId> instances) {
		// TODO Auto-generated method stub
		throw new IllegalStateException("NOT YET IMPLEMENTED");
	}

	public AnnotatedTableMetaData getMetaData(AnnotatedTableId id) {
		return this.tableMetaData.get(id);
	}

	public AnnotatedTableData getData(AnnotatedTableId id, int row) {
		// TODO Auto-generated method stub
		throw new IllegalStateException("NOT YET IMPLEMENTED");
	}

	// public Map<AnnotatedTableId, String> getValues(InstanceId instance,
	// PropertyId property) {
	// Map<AnnotatedTableId, String> values = new HashMap<AnnotatedTableId,
	// String>();
	//
	// Set<AnnotatedTableId> tables = new HashSet<AnnotatedTableId>();
	//
	// Set<AnnotatedTableId> instanceTables = this.instanceIndex.get(instance);
	// Set<AnnotatedTableId> propertyTables = this.propertyIndex.get(property);
	//
	// if (instanceTables == null | propertyTables == null)
	// return values;
	//
	// for (AnnotatedTableId id : instanceTables) {
	// if (propertyTables.contains(id)) {
	// tables.add(id);
	// }
	// }
	//
	// for (AnnotatedTableId id : tables) {
	//
	// AnnotatedTableMetaData metaData = this.tableMetaData.get(id);
	// int row = metaData.getRow(instance);
	// int col = metaData.getCol(property);
	//
	// Table t = this.tableData.get(id);
	// Object o = t.getColumns().get(col).getValues().get(row);
	// if (o != null) {
	//
	// values.put(id, o.toString());
	// }
	// }
	//
	// return values;
	// }

	@Override
	public Set<InstanceId> getInstances() {
		return Collections.unmodifiableSet(instanceIndex.keySet());
	}

	@Override
	public Set<AnnotatedTableId> getAllTableIds() {
		return Collections.unmodifiableSet(tableMetaData.keySet());
	}

	@Override
	public Set<AnnotatedTableId> getAllTableIds(ClassId setId) {
		return Collections.unmodifiableSet(tableMetaData.entrySet().stream().filter((e) -> {
			return e.getValue().getDbPediaClass().equals(setId);
		}).map(Entry::getKey).collect(Collectors.toSet()));
	}

	@Override
	public Map<RowIdx, AbstractObject> getColumn(AnnotatedTableId id, ColumnIdx columnIndex) {
		return tableData.get(id).getColumn(columnIndex);
	}

	@Override
	public AbstractObject getValue(AnnotatedTableId id, RowIdx rowIdx, ColumnIdx colIdx) {

		try {
			return this.tableData.get(id).getValue(rowIdx, colIdx);
		} catch (NullPointerException e) {
			return null;
		}
	}

	@Override
	public AbstractObject getKeyValue(AnnotatedTableId id, RowIdx rowIdx) {

		try {
			return this.tableData.get(id).getKeyValue(rowIdx);
		} catch (NullPointerException e) {
			return null;
		}
	}

	@Override
	public HostId getHostId(AnnotatedTableId id) {
		return this.getMetaData(id).getHostId();
	}

	@Override
	public void printStatistics() {
		AtomicInteger totalValues = new AtomicInteger(0);
		AtomicInteger mappedValues = new AtomicInteger(0);
		AtomicInteger allTables = new AtomicInteger(0);
		AtomicInteger mappedColumns = new AtomicInteger(0);

		// ConcurrentHashMap<AnnotatedTableId, Boolean> tableIds = new
		// ConcurrentHashMap<AnnotatedTableId, Boolean>();

		try {
			new PersistentParallel<AnnotatedTableId>("Print Table Statistics").foreach(getAllTableIds(), (id) -> {
				allTables.incrementAndGet();

				int totalValuesTemp = 0;
				int mappedValuesTemp = 0;
				int mappedColumnsTemp = 0;

				AnnotatedTableMetaData md = this.getMetaData(id);
				AnnotatedTable td = this.tableData.get(id);

				Map<Integer, List<PropertyId>> columnMapings = md.getAllProperties();
				Map<Integer, InstanceId> rowMappings = md.getInstances();

				int columnIdx = 0;
				for (ColumnIdx colIdx : td.getColumns()) {
					boolean mapped = false;
					if (columnMapings.get(columnIdx).size() > 0) {
						mapped = true;
						mappedColumnsTemp++;
					}
					for (Entry<RowIdx, AbstractObject> v : td.getColumn(colIdx).entrySet()) {

						if (v.getValue() != null) {
							totalValuesTemp++;
							if (mapped && rowMappings.containsKey(v.getKey().getRowIdx())) {
								mappedValuesTemp++;
							}
						}
					}

					columnIdx++;
				}

				totalValues.addAndGet(totalValuesTemp);
				mappedValues.addAndGet(mappedValuesTemp);
				mappedColumns.addAndGet(mappedColumnsTemp);
			});
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		System.out.println("Number of all Tables: " + allTables.get());
		System.out.println("Number of mapped Properties: " + mappedColumns.get());
		System.out.println("Number of total Values: " + totalValues.get());
		System.out.println("Number of mapped Values: " + mappedValues.get());

	}

	@Override
	public AnnotatedTable getTable(AnnotatedTableId id) {
		return this.tableData.get(id);
	}

	@Override
	public Integer getTimeStamp(CellId cellId, TimestampIdentifier tsi) {

		if (tsi.getType() == TimestampType.CASCADED && tsi.getCascadedType() == CascadedType.COLUMN_VALUE) {
			return TimeObject.getYear(this.tableData.get(cellId.getTableId()).getValue(cellId.getRowIdx(), new ColumnIdx(tsi.getCascadedTypePosition())));
		} else {
			return this.tableMetaData.get(cellId.getTableId()).getTimeStamp(tsi);
		}

	}

	@Override
	public Integer getTimeStamp(TableColumnId tableColumnId, TimestampIdentifier tsi) {

		if (tsi.getType() == TimestampType.CASCADED && tsi.getCascadedType() == CascadedType.COLUMN_VALUE) {
			return null;
		} else if (TimestampType.PROPAGATED == tsi.getType()) {
			return tsi.getPropogatedYear();
		} else {
			return this.tableMetaData.get(tableColumnId.getTableId()).getTimeStamp(tsi);
		}

	}

	@Override
	public Set<TableRowId> getUnmatchedRowIds(Set<InstanceId> matchedInstances) {

		Set<TableRowId> rows = Collections.newSetFromMap(new ConcurrentHashMap<>());

		PersistentParallel6.runAndFinish(this.tableMetaData.entrySet(), (table) -> {

			AnnotatedTableId tableId = table.getKey();
			int start = table.getValue().getNumHeaderRows();
			int end = table.getValue().getNumberOfRows() + table.getValue().getNumHeaderRows();

			for (int i = start; i < end; i++) {
				InstanceId instanceId = table.getValue().getInstance(i);
				if (instanceId == null || !matchedInstances.contains(instanceId)) {
					// AnnotatedTable fulTable = this.getTable(tableId);
					// ImmutableMap<RowIdx, AbstractObject> keyColumn =
					// fulTable.getKeyColumn();
					// AbstractObject ao = keyColumn.get(new RowIdx(i));

					rows.add(new TableRowId(tableId, i));
				}

			}

		}, this.getClass());

		return rows;
	}

	@Override
	public Set<TableRowId> getUnmatchedRowIds(Set<InstanceId> matchedInstances, ClassId classId) {
		Set<TableRowId> rows = Collections.newSetFromMap(new ConcurrentHashMap<>());

		PersistentParallel6.runAndFinish(this.tableMetaData.entrySet(), (table) -> {
			if (table.getValue().getDbPediaClass().equals(classId)) {
				AnnotatedTableId tableId = table.getKey();
				int start = table.getValue().getNumHeaderRows();
				int end = table.getValue().getNumberOfRows() + table.getValue().getNumHeaderRows();

				for (int i = start; i < end; i++) {
					InstanceId instanceId = table.getValue().getInstance(i);
					if (instanceId == null || !matchedInstances.contains(instanceId)) {
						// AnnotatedTable fulTable = this.getTable(tableId);
						// ImmutableMap<RowIdx, AbstractObject> keyColumn =
						// fulTable.getKeyColumn();
						// AbstractObject ao = keyColumn.get(new RowIdx(i));

						rows.add(new TableRowId(tableId, i));
					}

				}
			}

		}, this.getClass());

		return rows;
	}

	@Override
	public Set<TableRowId> getUnmatchedRowIds(ClassId classId) {
		Set<TableRowId> rows = Collections.newSetFromMap(new ConcurrentHashMap<>());

		PersistentParallel6.runAndFinish(this.tableMetaData.entrySet(), (table) -> {
			if (table.getValue().getDbPediaClass().equals(classId)) {
				AnnotatedTableId tableId = table.getKey();
				int start = table.getValue().getNumHeaderRows();
				int end = table.getValue().getNumberOfRows() + table.getValue().getNumHeaderRows();

				for (int i = start; i < end; i++) {
					InstanceId instanceId = table.getValue().getInstance(i);
					if (instanceId == null) {
						// AnnotatedTable fulTable = this.getTable(tableId);
						// ImmutableMap<RowIdx, AbstractObject> keyColumn =
						// fulTable.getKeyColumn();
						// AbstractObject ao = keyColumn.get(new RowIdx(i));

						rows.add(new TableRowId(tableId, i));
					}

				}
			}

		}, this.getClass());

		return rows;
	}

	@Override
	public Set<TableRowId> getUnmatchedRowIdsForTables(Set<AnnotatedTableId> tableIdsSet) {
		Set<TableRowId> rows = Collections.newSetFromMap(new ConcurrentHashMap<>());

		PersistentParallel6.runAndFinish(tableIdsSet, (tableId) -> {
			AnnotatedTableMetaData table = this.tableMetaData.get(tableId);

			int start = table.getNumHeaderRows();
			int end = table.getNumberOfRows() + table.getNumHeaderRows();

			for (int i = start; i < end; i++) {
				InstanceId instanceId = table.getInstance(i);
				if (instanceId == null) {
					// AnnotatedTable fulTable = this.getTable(tableId);
					// ImmutableMap<RowIdx, AbstractObject> keyColumn =
					// fulTable.getKeyColumn();
					// AbstractObject ao = keyColumn.get(new RowIdx(i));

					rows.add(new TableRowId(tableId, i));
				}

			}
		}, this.getClass());

		return rows;

	}

	@Override
	public Set<TableRowId> getMatchedRowIdsForTables(Set<AnnotatedTableId> tableIdsSet) {
		Set<TableRowId> rows = Collections.newSetFromMap(new ConcurrentHashMap<>());

		PersistentParallel6.runAndFinish(tableIdsSet, (tableId) -> {
			AnnotatedTableMetaData table = this.tableMetaData.get(tableId);

			int start = table.getNumHeaderRows();
			int end = table.getNumberOfRows() + table.getNumHeaderRows();

			for (int i = start; i < end; i++) {
				InstanceId instanceId = table.getInstance(i);
				if (instanceId != null) {
					// AnnotatedTable fulTable = this.getTable(tableId);
					// ImmutableMap<RowIdx, AbstractObject> keyColumn =
					// fulTable.getKeyColumn();
					// AbstractObject ao = keyColumn.get(new RowIdx(i));

					rows.add(new TableRowId(tableId, i));
				}

			}
		}, this.getClass());

		return rows;

	}

	@Override
	public int getNumOfUnmatchedRowIdsForTable(AnnotatedTableId tableId) {
		int ret = 0;

		AnnotatedTableMetaData table = this.tableMetaData.get(tableId);

		int start = table.getNumHeaderRows();
		int end = table.getNumberOfRows() + table.getNumHeaderRows();

		for (int i = start; i < end; i++) {
			InstanceId instanceId = table.getInstance(i);
			if (instanceId == null) {
				// AnnotatedTable fulTable = this.getTable(tableId);
				// ImmutableMap<RowIdx, AbstractObject> keyColumn =
				// fulTable.getKeyColumn();
				// AbstractObject ao = keyColumn.get(new RowIdx(i));

				ret++;
			}

		}

		return ret;

	}

	@Override
	public int getNumOfMatchedRowIdsForTable(AnnotatedTableId tableId) {
		int ret = 0;

		AnnotatedTableMetaData table = this.tableMetaData.get(tableId);

		int start = table.getNumHeaderRows();
		int end = table.getNumberOfRows() + table.getNumHeaderRows();

		for (int i = start; i < end; i++) {
			InstanceId instanceId = table.getInstance(i);
			if (instanceId != null) {
				// AnnotatedTable fulTable = this.getTable(tableId);
				// ImmutableMap<RowIdx, AbstractObject> keyColumn =
				// fulTable.getKeyColumn();
				// AbstractObject ao = keyColumn.get(new RowIdx(i));

				ret++;
			}

		}

		return ret;

	}

	@Override
	public Set<TableRowId> getMatchedRowsIds(Set<InstanceId> matchedInstances, ClassId classId) {
		Set<TableRowId> rows = Collections.newSetFromMap(new ConcurrentHashMap<>());

		PersistentParallel6.runAndFinish(this.tableMetaData.entrySet(), (table) -> {
			if (table.getValue().getDbPediaClass().equals(classId)) {
				AnnotatedTableId tableId = table.getKey();
				int start = table.getValue().getNumHeaderRows();
				int end = table.getValue().getNumberOfRows() + table.getValue().getNumHeaderRows();

				for (int i = start; i < end; i++) {
					InstanceId instanceId = table.getValue().getInstance(i);
					if (instanceId != null && matchedInstances.contains(instanceId)) {
						// AnnotatedTable fulTable = this.getTable(tableId);
						// ImmutableMap<RowIdx, AbstractObject> keyColumn =
						// fulTable.getKeyColumn();
						// AbstractObject ao = keyColumn.get(new RowIdx(i));

						rows.add(new TableRowId(tableId, i));
					}

				}
			}

		}, this.getClass());

		return rows;
	}

	// @Override
	// public AnnotatedTableRow getTableRow(TableRowId tableRowId) {
	// String label = null;
	// Map<PropertyId, AbstractObject> values = new HashMap<>();
	// AnnotatedTable table = this.tableData.get(tableRowId.getTableId());
	// AbstractObject keyValue = table.getKeyValue(tableRowId.getRowIdx());
	// if (keyValue != null) {
	// if (keyValue instanceof StringObject) {
	// label = ((StringObject) keyValue).getString();
	// } else {
	// label = keyValue.toString();
	// }
	// }
	//
	// for (Entry<Integer, PropertyId> property :
	// this.tableMetaData.get(tableRowId.getTableId()).getProperties().entrySet())
	// {
	// AbstractObject ao = table.getValue(tableRowId.getRowIdx(), new
	// ColumnIdx(property.getKey()));
	// if (ao != null) {
	// values.put(property.getValue(), ao);
	// }
	// }
	//
	// return new AnnotatedTableRow(label, values);
	// }

	@Override
	public List<String> getNeighborHood(TableRowId tableRowId) {
		List<String> ret = new ArrayList<>();
		ImmutableMap<RowIdx, AbstractObject> keyColumn = this.getTable(tableRowId.getTableId()).getKeyColumn();

		for (Entry<RowIdx, AbstractObject> entry : keyColumn.entrySet()) {
			if (!entry.getKey().equals(tableRowId.getRowIdx())) {
				if (entry.getValue() instanceof StringObject) {
					StringObject so = (StringObject) entry.getValue();
					ret.add(so.getString());

				}
			}
		}
		return ret;
	}

	@Override
	public List<String> getLabels(AnnotatedTableId tableId) {
		List<String> ret = new ArrayList<>();
		ImmutableMap<RowIdx, AbstractObject> keyColumn = this.getTable(tableId).getKeyColumn();

		for (Entry<RowIdx, AbstractObject> entry : keyColumn.entrySet()) {
			if (entry.getValue() instanceof StringObject) {
				StringObject so = (StringObject) entry.getValue();
				ret.add(so.getString());

			}

		}
		return ret;
	}

	@Override
	public String getLabelRaw(TableRowId tableRowId) {
		AnnotatedTable table = getTable(tableRowId.getTableId());
		ColumnIdx keyColumnIdx = table.getKeyColumnIndex();

		String string = table.getStringTable().get(tableRowId.getRowIdx(), keyColumnIdx);

		return string;
	}

	@Override
	public StringObject getLabel(TableRowId tableRowId) {
		AnnotatedTable table = getTable(tableRowId.getTableId());
		ImmutableMap<RowIdx, AbstractObject> keyColumn = table.getKeyColumn();
		AbstractObject ao = keyColumn.get(tableRowId.getRowIdx());

		if (ao instanceof StringObject) {
			StringObject labelSO = (StringObject) ao;
			return labelSO;
		}

		return null;
	}

	@Override
	public ClassId getClassId(AnnotatedTableId tableId) {
		AnnotatedTableMetaData md = this.getMetaData(tableId);
		if (md != null) {
			return md.getDbPediaClass();
		}
		return null;
	}

	@Override
	public Set<TableRowId> getAllRowIds(ClassId classId) {

		Set<TableRowId> rows = Collections.newSetFromMap(new ConcurrentHashMap<>());

		PersistentParallel6.runAndFinish(this.tableMetaData.entrySet(), (table) -> {

			if (table.getValue().getDbPediaClass().equals(classId)) {
				AnnotatedTableId tableId = table.getKey();
				int start = table.getValue().getNumHeaderRows();
				int end = table.getValue().getNumberOfRows() + table.getValue().getNumHeaderRows();

				for (int i = start; i < end; i++) {
					rows.add(new TableRowId(tableId, i));
				}
			}

		}, this.getClass());

		return rows;
	}

	@Override
	public Set<TableRowId> getAllRowIds(Set<AnnotatedTableId> tableIds) {
		Set<TableRowId> rows = Collections.newSetFromMap(new ConcurrentHashMap<>());

		PersistentParallel6.runAndFinish(tableIds, (tableId) -> {
			AnnotatedTableMetaData table = tableMetaData.get(tableId);

			int start = table.getNumHeaderRows();
			int end = table.getNumberOfRows() + table.getNumHeaderRows();

			for (int i = start; i < end; i++) {
				rows.add(new TableRowId(tableId, i));
			}

		}, this.getClass());

		return rows;
	}

	@Override
	public List<TableRowId> getRowIds(AnnotatedTableId tableId) {
		List<TableRowId> rows = new ArrayList<>();
		AnnotatedTableMetaData table = tableMetaData.get(tableId);

		int start = table.getNumHeaderRows();
		int end = table.getNumberOfRows() + table.getNumHeaderRows();

		for (int i = start; i < end; i++) {
			rows.add(new TableRowId(tableId, i));
		}

		return rows;
	}

	@Override
	public Set<AnnotatedTableId> getAllTableIdsWithMatches(ClassId setId) {

		return Collections.unmodifiableSet(tableMetaData.entrySet().stream().filter((e) -> {
			return e.getValue().getDbPediaClass().equals(setId) && e.getValue().getAllProperties().size() > 0;
		}).map(Entry::getKey).collect(Collectors.toSet()));
	}

	@Override
	public Optional<InstanceId> getMatchedInstanceId(TableRowId rowId) {
		AnnotatedTableMetaData md = this.tableMetaData.get(rowId.getTableId());
		if (md != null) {
			InstanceId instanceId = md.getInstance(rowId.getRowIdx().getRowIdx());
			if (instanceId != null)
				return Optional.of(instanceId);
		}
		return Optional.empty();
	}

	@Override
	public Set<TableRowId> getRows(InstanceId instance) {
		Set<TableRowId> rowIds = new HashSet<>();
		for (AnnotatedTableId tableId : this.getTables(instance)) {
			for (Entry<Integer, InstanceId> e : tableMetaData.get(tableId).getInstances().entrySet()) {
				if (e.getValue().equals(instance)) {
					rowIds.add(new TableRowId(tableId, e.getKey()));
				}
			}
		}
		return rowIds;
	}

	@Override
	public int getNumOfMatchedColumns(InstanceId instance) {
		int count = 0;
		for (AnnotatedTableId tableId : this.getTables(instance)) {
			count += tableMetaData.get(tableId).getProperties().size();

		}
		return count;
	}

	@Override
	public Map<PropertyId, List<CellId>> getMatches(InstanceId instanceId) {
		Map<PropertyId, List<CellId>> ret = new HashMap<>();
		for (AnnotatedTableId tableId : this.instanceIndex.get(instanceId)) {

			for (Entry<Integer, InstanceId> instanceEntry : this.tableMetaData.get(tableId).getInstances().entrySet()) {
				if (instanceEntry.getValue().equals(instanceId)) {
					for (Entry<Integer, PropertyId> propertyEntry : this.tableMetaData.get(tableId).getProperties().entrySet()) {
						RowIdx rowIdx = new RowIdx(instanceEntry.getKey());
						ColumnIdx columnIdx = new ColumnIdx(propertyEntry.getKey());
						if (getValue(tableId, rowIdx, columnIdx) != null) {
							CHMHelper.getFromMap(ret, propertyEntry.getValue(), () -> new ArrayList<>()).add(new CellId(tableId, columnIdx, rowIdx));
						}
					}
				}
			}

		}

		return ret;
	}

	@Override
	public AbstractObject getValue(CellId id) {
		return getValue(id.getTableId(), id.getRowIdx(), id.getColumnIdx());
	}

	@Override
	public List<AnnotatedTablesReaderData> getTableReaderData() {

		List<AnnotatedTablesReaderData> readerData = new ArrayList<>();
		for (Entry<AnnotatedTableId, AnnotatedTableMetaData> md : tableMetaData.entrySet()) {
			readerData.add(new AnnotatedTablesReaderData(md.getKey(), md.getValue(), tableData.get(md.getKey())));
		}

		return readerData;

	}

	@Override
	public List<AnnotatedTablesReaderData> getTableReaderData(Set<AnnotatedTableId> tableIds) {
		List<AnnotatedTablesReaderData> readerData = new ArrayList<>();
		for (AnnotatedTableId tableId : tableIds) {
			readerData.add(new AnnotatedTablesReaderData(tableId, tableMetaData.get(tableId), tableData.get(tableId)));
		}

		return readerData;
	}

	// public Integer getCellYear(TripleId tid, CellId aid) {
	//
	// PropertyId activeYearsStartYear = new
	// PropertyId("http://dbpedia.org/ontology/activeYearsStartYear");
	// AnnotatedTableMetaData md =
	// annotatedTablesRepository.getMetaData(aid.getTableId());
	// AnnotatedTable td = annotatedTablesRepository.getTable(aid.getTableId());
	//
	// Integer year = null;
	// for (int i = 0; i < td.getColumns().size(); i++) {
	// if (td.getDataType(new ColumnIdx(i)) == ValueType.TIME) {
	// PropertyId propertyId = md.getProperty(i);
	// if (propertyId == null || propertyId.equals(activeYearsStartYear)) {
	// Object oYear = td.getValue(aid.getRowIdx(), new ColumnIdx(i));
	// if (oYear instanceof Date) {
	// calendar.getCalendar().setTime((Date) oYear);
	// int iYear = calendar.getCalendar().get(Calendar.YEAR);
	//
	// if (year == null || iYear > year) {
	// year = iYear;
	// }
	// }
	// }
	// }
	//
	// }
	//
	// return year;
	// }

}
