package de.mannheim.uni.expansion.data.tables.repository.memory;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Consumer;

import com.google.common.collect.TreeBasedTable;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.InstanceIndexComposite;
import de.mannheim.uni.expansion.data.tables.model.PropertyIndexComposite;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTablesReader;
import de.mannheim.uni.expansion.data.tables.repository.MagnitudeFixer;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SparseArrayMap;
import de.mannheim.uni.tableprocessor.model.Table;
import de.mannheim.uni.tableprocessor.model.TableColumn;
import de.mannheim.uni.tableprocessor.model.TableColumn.ColumnDataType;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class AnnotatedTablesRepositoryImplMemoryFactory {

	private final ConcurrentHashMap<ClassId, Set<AnnotatedTableId>> indexClass = new ConcurrentHashMap<ClassId, Set<AnnotatedTableId>>();
	private final ConcurrentHashMap<PropertyId, Set<AnnotatedTableId>> indexProperty = new ConcurrentHashMap<PropertyId, Set<AnnotatedTableId>>();
	private final ConcurrentHashMap<InstanceId, Set<AnnotatedTableId>> indexInstances = new ConcurrentHashMap<InstanceId, Set<AnnotatedTableId>>();

	private final ConcurrentHashMap<AnnotatedTableId, AnnotatedTable> tableData = new ConcurrentHashMap<AnnotatedTableId, AnnotatedTable>();
	private final ConcurrentHashMap<AnnotatedTableId, AnnotatedTableMetaData> tableMetaData = new ConcurrentHashMap<AnnotatedTableId, AnnotatedTableMetaData>();

	private void addToClassIndex(ClassId dbPediaClass, AnnotatedTableId id) {
		Set<AnnotatedTableId> set = null;
		set = new HashSet<AnnotatedTableId>();
		if (indexClass.putIfAbsent(dbPediaClass, set) != null) {
			set = indexClass.get(dbPediaClass);
		}
		set.add(id);
	}

	private void addToPropertyIndex(PropertyId dbPediaProperty, AnnotatedTableId id) {
		Set<AnnotatedTableId> set = null;
		set = new HashSet<AnnotatedTableId>();
		if (indexProperty.putIfAbsent(dbPediaProperty, set) != null) {
			set = indexProperty.get(dbPediaProperty);
		}
		set.add(id);
	}

	private void addToInstanceIndex(InstanceId dbPediaInstance, AnnotatedTableId id) {
		Set<AnnotatedTableId> set = null;
		set = new HashSet<AnnotatedTableId>();
		if (indexInstances.putIfAbsent(dbPediaInstance, set) != null) {
			set = indexInstances.get(dbPediaInstance);
		}
		set.add(id);
	}

	public AnnotatedTablesRepositoryImplMemoryFactory(final AnnotatedTablesReader reader, boolean filter, KnowledgeBase classFilter,
			MagnitudeFixer magnitudefixer, List<AnnotatedTableFilter> filters) {

		Set<PropertyId> filteredProperties = new HashSet<>();
		filteredProperties.add(new PropertyId("http://dbpedia.org/ontology/wikiPageID"));
		filteredProperties.add(new PropertyId("http://dbpedia.org/ontology/wikiPageRevisionID"));
		filteredProperties.add(new PropertyId("http://www.w3.org/2002/07/owl#sameAs"));

		filteredProperties.add(new PropertyId("http://purl.org/dc/elements/1.1/description"));
		filteredProperties.add(new PropertyId("http://dbpedia.org/ontology/thumbnail"));
		filteredProperties.add(new PropertyId("http://dbpedia.org/ontology/type"));
		filteredProperties.add(new PropertyId("http://www.w3.org/1999/02/22-rdf-syntax-ns#type"));
		filteredProperties.add(new PropertyId("http://www.w3.org/2000/01/rdf-schema#comment"));

		// try {
		// AnnotatedTablesWriter writer;
		// if (outputFolder != null) {
		// writer = new AnnotatedTablesWriter(outputFolder);
		// } else {
		// writer = null;
		// }
		long startTime = System.nanoTime();

		AtomicInteger duplicateTables = new AtomicInteger(0);
		AtomicInteger allTables = new AtomicInteger(0);

		AtomicInteger ai = new AtomicInteger(0);

		reader.runForEachTable(filters, new Consumer<AnnotatedTablesReaderData>() {

			public void accept(AnnotatedTablesReaderData data) {

				// if (writer != null) {
				// writer.add(data);
				// }

				if (classFilter != null) {
					ClassId classId = data.getMetaData().getDbPediaClass();
					if (!classFilter.getClasses().contains(classId)) {
						return;
					}

					// else {
					// BidiMap<PropertyId, Integer> properties =
					// classFilter.getProperties(classId);
					// BidiMap<InstanceId, Integer> instances =
					// classFilter.getInstances(classId);
					// Set<PropertyId> removeProperties = new HashSet<>();
					// Set<InstanceId> removeInstances = new HashSet<>();
					//
					// for (Entry<Integer, PropertyId> e :
					// data.getMetaData().getProperties().entrySet()) {
					// if (!properties.containsKey(e.getValue())) {
					// removeProperties.add(e.getValue());
					// }
					// }
					//
					// for (Entry<Integer, InstanceId> e :
					// data.getMetaData().getInstances().entrySet()) {
					// if (!instances.containsKey(e.getValue())) {
					// removeInstances.add(e.getValue());
					// }
					// }
					//
					// if (!removeProperties.isEmpty() ||
					// !removeInstances.isEmpty()) {
					// data = new AnnotatedTablesReaderData(data.getId(),
					// data.getMetaData().filterOut(removeProperties,
					// removeInstances),
					// data.getTableData());
					// }
					//
					// }
				}

				allTables.incrementAndGet();

				AnnotatedTableId id = data.getId();
				if (!tableMetaData.containsKey(id)) {

					// AnnotatedTable td = data.getTableData();

					if (filter) {
						// data = new AnnotatedTablesReaderData(data.getId(),
						// filterOutProperties(data.getMetaData(),
						// filteredProperties), data.getTableData());
						//
						// data = new AnnotatedTablesReaderData(data.getId(),
						// replaceNFLProperties(data.getMetaData()),
						// data.getTableData());
						//
						// data = new AnnotatedTablesReaderData(data.getId(),
						// data.getMetaData(),
						// filterOutUnmatchedData(data.getTableData(),
						// data.getMetaData()));
						//
						// data = new AnnotatedTablesReaderData(data.getId(),
						// filterOutTimestampsSources(data.getMetaData()),
						// data.getTableData());

					}

					// if (filter && classFilter != null) {
					// data = new AnnotatedTablesReaderData(data.getId(),
					// filterOutUnavailableMatches(data.getMetaData(),
					// data.getTableData()),
					// data.getTableData(), classFilter);
					//
					// }

					// td = compressTableColumns(td);

					// if (data == null) {
					// return;
					// }

					// int size1 = md.getNumberOfRows();
					// int size2 = td.getKey().getNumRows();
					// int size2 = td.getKey().getValues().size();

					// Adding entry to class index
					addToClassIndex(data.getMetaData().getDbPediaClass(), id);

					// Adding entries to property index
					for (PropertyId p : data.getMetaData().getProperties().values()) {
						if (p != null)
							addToPropertyIndex(p, id);
					}

					// Adding entries to instance index
					for (InstanceId i : data.getMetaData().getInstances().values()) {
						if (i != null)
							addToInstanceIndex(i, id);
					}

					if (!data.getMetaData().getProperties().isEmpty()) {
						int sdf = 130;
					}
					// Storing the Data
					tableData.put(id, data.getTableData());
					tableMetaData.put(id, data.getMetaData());

					// if (size1 != size2) {
					// // System.out.println(id.getTableName());
					// ai.incrementAndGet();
					// }
				} else {
					duplicateTables.incrementAndGet();
				}

			}

		});

		if (classFilter != null) {
			countData(classFilter, tableData, tableMetaData);
		}

		// try {
		// writer.finishCloseAndAwait();
		// } catch (InterruptedException e) {
		// // TODO Auto-generated catch block
		// e.printStackTrace();
		// }

		System.out.println("Number of all Tables: " + allTables.get());
		System.out.println("Number of duplicate tables: " + duplicateTables.get());
		System.out.println("TABLED with HEADER BUG: " + ai.intValue());

		System.out.println("Time taken to read annotated Tables:" + TimeUnit.SECONDS.convert(System.nanoTime() - startTime, TimeUnit.NANOSECONDS));
		// } catch (IOException e1) {
		// // TODO Auto-generated catch block
		// e1.printStackTrace();
		// }

	}

	protected void countData(KnowledgeBase dbpediaRepository, ConcurrentHashMap<AnnotatedTableId, AnnotatedTable> tableData2,
			ConcurrentHashMap<AnnotatedTableId, AnnotatedTableMetaData> tableMetaData2) {

		ConcurrentHashMap<ClassPropertyId, AtomicInteger> countMap = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(tableData2.entrySet(), (entry) -> {
			AnnotatedTable td = entry.getValue();
			AnnotatedTableMetaData md = tableMetaData2.get(entry.getKey());

			for (Entry<Integer, List<PropertyId>> e : md.getAllProperties().entrySet()) {
				if (e.getValue() != null) {
					for (PropertyId propertyId : e.getValue()) {
						if (dbpediaRepository.hasProperty(md.getDbPediaClass(), propertyId)) {
							for (Entry<Integer, List<InstanceId>> ie : md.getAllInstances().entrySet()) {
								if (ie.getValue() != null) {
									List<InstanceId> newList = new ArrayList<>();
									for (InstanceId instanceId : ie.getValue()) {
										if (dbpediaRepository.hasInstance(md.getDbPediaClass(), instanceId)) {

											if (td.getValue(new RowIdx(ie.getKey()), new ColumnIdx(e.getKey())) != null) {
												CHMHelper.getFromCHM(countMap, new ClassPropertyId(md.getDbPediaClass(), propertyId), () -> new AtomicInteger())
														.incrementAndGet();
											}

										}
									}

								}
							}

						}
					}

				}
			}
		}, "Count Matches");
		System.out.println(countMap.toString());

	}

	protected AnnotatedTableMetaData replaceCountryProperties(AnnotatedTableMetaData md) {
		Map<Integer, List<PropertyId>> newProperties = new HashMap<>();
		Map<PropertyIndexComposite, Double> newPropertyScores = new HashMap<>();
		for (Entry<Integer, List<PropertyId>> e : md.getAllProperties().entrySet()) {
			if (e.getValue() != null) {
				List<PropertyId> newList = new ArrayList<>();
				for (PropertyId propertyId : e.getValue()) {

					if (propertyId.getAsString().equals("http://dbpedia.org/ontology/PopulatedPlace/populationDensity")) {
						PropertyId newProperty = new PropertyId("http://dbpedia.org/ontology/populationDensity");
						newList.add(newProperty);
						PropertyIndexComposite composite = new PropertyIndexComposite(e.getKey(), propertyId);
						PropertyIndexComposite newComposite = new PropertyIndexComposite(e.getKey(), newProperty);
						newPropertyScores.put(newComposite, md.getPropertyScore(composite));
					} else {
						newList.add(propertyId);
						PropertyIndexComposite composite = new PropertyIndexComposite(e.getKey(), propertyId);
						newPropertyScores.put(composite, md.getPropertyScore(composite));
					}

				}

				if (!newList.isEmpty()) {
					newProperties.put(e.getKey(), newList);
				}
			}
		}

		return new AnnotatedTableMetaData(md.getSource(), md.getDbPediaClass(), md.getClassConf(), md.getNumHeaderRows(), md.getNumberOfRows(), newProperties,
				newPropertyScores, md.getAllInstances(), md.getInstancesScores(), md.getKeyColumn(), md.getYearsBeforeText(), md.getYearsAfterText(),
				md.getYearsInContextBefore(), md.getYearsInTableTitle(), md.getYearsInPageTitle(), md.getPropertiesYears(), md.getPageTitle(),
				md.getTableTitle(), md.getTableContextTimeStampBeforeTable(), md.getTextBeforeTable(), md.getTextAfterTable());
	}

	protected AnnotatedTableMetaData replaceNFLProperties(AnnotatedTableMetaData md) {
		Map<Integer, List<PropertyId>> newProperties = new HashMap<>();
		Map<PropertyIndexComposite, Double> newPropertyScores = new HashMap<>();
		for (Entry<Integer, List<PropertyId>> e : md.getAllProperties().entrySet()) {
			if (e.getValue() != null) {
				List<PropertyId> newList = new ArrayList<>();
				for (PropertyId propertyId : e.getValue()) {

					if (propertyId.getAsString().equals("http://dbpedia.org/ontology/formerTeam")) {
						PropertyId newProperty = new PropertyId("http://dbpedia.org/ontology/team");
						newList.add(newProperty);
						PropertyIndexComposite composite = new PropertyIndexComposite(e.getKey(), propertyId);
						PropertyIndexComposite newComposite = new PropertyIndexComposite(e.getKey(), newProperty);
						newPropertyScores.put(newComposite, md.getPropertyScore(composite));
					} else {
						newList.add(propertyId);
						PropertyIndexComposite composite = new PropertyIndexComposite(e.getKey(), propertyId);
						newPropertyScores.put(composite, md.getPropertyScore(composite));
					}

				}

				if (!newList.isEmpty()) {
					newProperties.put(e.getKey(), newList);
				}
			}
		}

		return new AnnotatedTableMetaData(md.getSource(), md.getDbPediaClass(), md.getClassConf(), md.getNumHeaderRows(), md.getNumberOfRows(), newProperties,
				newPropertyScores, md.getAllInstances(), md.getInstancesScores(), md.getKeyColumn(), md.getYearsBeforeText(), md.getYearsAfterText(),
				md.getYearsInContextBefore(), md.getYearsInTableTitle(), md.getYearsInPageTitle(), md.getPropertiesYears(), md.getPageTitle(),
				md.getTableTitle(), md.getTableContextTimeStampBeforeTable(), md.getTextBeforeTable(), md.getTextAfterTable());
	}

	protected AnnotatedTableMetaData filterOutUnavailableMatches(AnnotatedTableMetaData md, Table td, KnowledgeBase dbpediaRepository) {
		Map<Integer, List<InstanceId>> newInstances = new HashMap<>();
		Map<InstanceIndexComposite, Double> newInstanceScores = new HashMap<>();
		for (Entry<Integer, List<InstanceId>> e : md.getAllInstances().entrySet()) {
			if (e.getValue() != null) {
				List<InstanceId> newList = new ArrayList<>();
				for (InstanceId instanceId : e.getValue()) {
					if (dbpediaRepository.hasInstance(md.getDbPediaClass(), instanceId)) {
						newList.add(instanceId);
						InstanceIndexComposite composite = new InstanceIndexComposite(e.getKey(), instanceId);
						newInstanceScores.put(composite, md.getInstanceScore(composite));
					}
				}

				if (!newList.isEmpty()) {
					newInstances.put(e.getKey(), newList);
				}
			}
		}

		Map<Integer, List<PropertyId>> newProperties = new HashMap<>();
		Map<PropertyIndexComposite, Double> newPropertyScores = new HashMap<>();
		for (Entry<Integer, List<PropertyId>> e : md.getAllProperties().entrySet()) {
			if (e.getValue() != null) {
				List<PropertyId> newList = new ArrayList<>();
				for (PropertyId propertyId : e.getValue()) {
					if (dbpediaRepository.hasProperty(md.getDbPediaClass(), propertyId)
							|| td.getColumns().get(e.getKey()).getDataType().equals(ColumnDataType.date)) {
						newList.add(propertyId);
						PropertyIndexComposite composite = new PropertyIndexComposite(e.getKey(), propertyId);
						newPropertyScores.put(composite, md.getPropertyScore(composite));
					}
				}

				if (!newList.isEmpty()) {
					newProperties.put(e.getKey(), newList);
				}
			}
		}

		return new AnnotatedTableMetaData(md.getSource(), md.getDbPediaClass(), md.getClassConf(), md.getNumHeaderRows(), md.getNumberOfRows(), newProperties,
				newPropertyScores, newInstances, newInstanceScores, md.getKeyColumn(), md.getYearsBeforeText(), md.getYearsAfterText(),
				md.getYearsInContextBefore(), md.getYearsInTableTitle(), md.getYearsInPageTitle(), md.getPropertiesYears(), md.getPageTitle(),
				md.getTableTitle(), md.getTableContextTimeStampBeforeTable(), md.getTextBeforeTable(), md.getTextAfterTable());

	}

	@Deprecated
	private Table compressTableColumns(Table td) {
		if (td == null) {
			int fdf = 123;
			return null;
		}
		try {
			for (Iterator<TableColumn> it = td.getColumns().iterator(); it.hasNext();) {
				TableColumn tc = it.next();

				Map<Integer, Object> values = tc.getValues();
				if (!(values instanceof SparseArrayMap)) {

					// int capacity = 0;
					// if (values instanceof TreeMap && !values.isEmpty()) {
					// capacity = ((int) ((TreeMap) values).lastKey()) + 1;
					// } else {
					// for (Entry<Integer, Object> e : values.entrySet()) {
					// if (e.getKey() > capacity) {
					// capacity = e.getKey();
					// }
					// }
					// capacity++;
					// }
					Map<Integer, Object> arrayMap = new SparseArrayMap<Object>();
					arrayMap.putAll(values);

					throw new UnsupportedOperationException();
					// tc.setValues(arrayMap);
				}
			}

			return td;
		} catch (NullPointerException e) {
			throw e;
		}
	}

	private AnnotatedTable filterOutUnmatchedData(AnnotatedTable td, AnnotatedTableMetaData md) {

		TreeBasedTable<RowIdx, ColumnIdx, AbstractObject> table = TreeBasedTable.create();
		table.putAll(td.getTable());

		Map<ColumnIdx, ValueType> dataTypes = new HashMap<>(td.getDataTypes());

		// WE ITERATE OVER ORIGINAL BUT MODIFY NEW TO PREVENT:
		// CONCURRENTMODIFICATIONEXCEPTION
		for (ColumnIdx colIdx : td.getColumns()) {

			if (td.getDataType(colIdx) != ValueType.TIME && !colIdx.equals(td.getKeyColumnIndex())) {
				if (md.getProperties().containsKey(colIdx.getColumnIdx())) {
					Map<RowIdx, AbstractObject> values = new HashMap<>(td.getColumn(colIdx));

					for (Entry<RowIdx, AbstractObject> e : values.entrySet()) {

						if (!md.getInstances().containsKey(e.getKey())) {
							table.remove(e.getKey(), colIdx);
						}
					}
				} else {
					table.column(colIdx).clear();
					dataTypes.remove(colIdx);
				}
			}
		}

		return new AnnotatedTable(table, td.getStringTable(), dataTypes, td.getKeyColumnIndex(), td.getColumnHeaders());
	}

	// FIXME: THIS IS NOT SUPPOSED TO BE HERE AT ALLLL
	private AnnotatedTableMetaData filterOutProperties(AnnotatedTableMetaData md, Set<PropertyId> filteredProperties) {

		HashMap<PropertyIndexComposite, Double> newScoresMap = new HashMap<>(md.getAllPropertyScores());
		Map<Integer, List<PropertyId>> newProperties = new HashMap<>();

		for (Entry<Integer, List<PropertyId>> e : md.getAllProperties().entrySet()) {

			List<PropertyId> newList = new ArrayList<>();

			for (PropertyId propertyId : e.getValue()) {

				if (!filteredProperties.contains(propertyId)) {
					newList.add(propertyId);
				} else {

					newScoresMap.remove(new PropertyIndexComposite(e.getKey(), propertyId));
				}

			}

			newProperties.put(e.getKey(), newList);

		}

		return new AnnotatedTableMetaData(md.getSource(), md.getDbPediaClass(), md.getClassConf(), md.getNumHeaderRows(), md.getNumberOfRows(), newProperties,
				newScoresMap, md.getAllInstances(), md.getInstancesScores(), md.getKeyColumn(), md.getYearsBeforeText(), md.getYearsAfterText(),
				md.getYearsInContextBefore(), md.getYearsInTableTitle(), md.getYearsInPageTitle(), md.getPropertiesYears(), md.getPageTitle(),
				md.getTableTitle(), md.getTableContextTimeStampBeforeTable(), md.getTextBeforeTable(), md.getTextAfterTable());

	}

	public AnnotatedTablesRepositoryImplMemory getRepository() {
		return new AnnotatedTablesRepositoryImplMemory(indexClass, indexProperty, indexInstances, tableData, tableMetaData);
	}

	public AnnotatedTableMetaData filterOutTimestampsSources(AnnotatedTableMetaData md) {
		return new AnnotatedTableMetaData(md.getSource(), md.getDbPediaClass(), md.getClassConf(), md.getNumHeaderRows(), md.getNumberOfRows(),
				md.getAllProperties(), md.getAllPropertyScores(), md.getAllInstances(), md.getInstancesScores(), md.getKeyColumn(), md.getYearsBeforeText(),
				md.getYearsAfterText(), md.getYearsInContextBefore(), md.getYearsInTableTitle(), md.getYearsInPageTitle(), md.getPropertiesYears(), null, null,
				null, null, null);
	}
}
