package de.mannheim.uni.expansion.data.tables.repository.remote;

import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.HostId;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;

public interface AnnotatedTablesRepositoryRemote extends Remote {

	Set<AnnotatedTableId> ofClass(ClassId dbpediaClass) throws RemoteException;

	Set<AnnotatedTableId> hasProperty(PropertyId property) throws RemoteException;

	Set<AnnotatedTableId> hasProperties(Set<PropertyId> properties) throws RemoteException;

	Set<AnnotatedTableId> getTables(InstanceId instances) throws RemoteException;

	boolean hasInstance(InstanceId instance);

	StringObject getLabel(TableRowId tableRowId);

	Set<AnnotatedTableId> hasInstances(Set<InstanceId> instances) throws RemoteException;

	AnnotatedTableMetaData getMetaData(AnnotatedTableId id) throws RemoteException;

	HostId getHostId(AnnotatedTableId id) throws RemoteException;

	AnnotatedTable getTable(AnnotatedTableId id) throws RemoteException;

	Set<InstanceId> getInstances() throws RemoteException;

	Set<AnnotatedTableId> getAllTableIds() throws RemoteException;

	Map<RowIdx, AbstractObject> getColumn(AnnotatedTableId id, ColumnIdx columnIndex) throws RemoteException;

	AbstractObject getValue(AnnotatedTableId id, RowIdx rowIdx, ColumnIdx colIdx) throws RemoteException;

	AbstractObject getKeyValue(AnnotatedTableId id, RowIdx rowIdx) throws RemoteException;

	void printStatistics() throws RemoteException;

	Integer getTimeStamp(CellId cellId, TimestampIdentifier tsi) throws RemoteException;

	List<String> getNeighborHood(TableRowId tableRowId) throws RemoteException;

	List<String> getLabels(AnnotatedTableId tableId) throws RemoteException;

	Set<AnnotatedTableId> getAllTableIds(ClassId setId) throws RemoteException;

	Integer getTimeStamp(TableColumnId tableColumnId, TimestampIdentifier tsi) throws RemoteException;

	Set<TableRowId> getUnmatchedRowIds(Set<InstanceId> matchedInstances) throws RemoteException;

	Set<TableRowId> getMatchedRowsIds(Set<InstanceId> matchedInstances, ClassId classId) throws RemoteException;

	List<AnnotatedTablesReaderData> getTableReaderData(Set<AnnotatedTableId> tableIds) throws RemoteException;

	int getNumOfMatchedColumns(InstanceId instance) throws RemoteException;

	ClassId getClassId(AnnotatedTableId tableId) throws RemoteException;

	Set<TableRowId> getUnmatchedRowIds(Set<InstanceId> matchedInstances, ClassId classId) throws RemoteException;

	Set<TableRowId> getUnmatchedRowIds(ClassId classId) throws RemoteException;

	Set<TableRowId> getUnmatchedRowIdsForTables(Set<AnnotatedTableId> tableIdsSet) throws RemoteException;

	Set<TableRowId> getMatchedRowIdsForTables(Set<AnnotatedTableId> tableIdsSet) throws RemoteException;

	int getNumOfUnmatchedRowIdsForTable(AnnotatedTableId tableId) throws RemoteException;

	int getNumOfMatchedRowIdsForTable(AnnotatedTableId tableId) throws RemoteException;

	Set<TableRowId> getAllRowIds(ClassId classId) throws RemoteException;

	Set<TableRowId> getAllRowIds(Set<AnnotatedTableId> tableIds) throws RemoteException;

	List<TableRowId> getRowIds(AnnotatedTableId tableId) throws RemoteException;

	Set<AnnotatedTableId> getAllTableIdsWithMatches(ClassId setId) throws RemoteException;

	Optional<InstanceId> getMatchedInstanceId(TableRowId rowId) throws RemoteException;

	Set<TableRowId> getRows(InstanceId instance) throws RemoteException;

	Map<PropertyId, List<CellId>> getMatches(InstanceId instanceId) throws RemoteException;

	AbstractObject getValue(CellId id) throws RemoteException;

	List<AnnotatedTablesReaderData> getTableReaderData() throws RemoteException;

}
