package de.uni_mannheim.dws.expansion.data.kb.abstracts;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.GZIPInputStream;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.BatchConsumerProducer;
import de.mannheim.uni.expansion.utils.ExpansionStringUnescape;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.utils.kryo.CHMKryoAdapter;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ConcurrentKryoAdapter;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ConcurrentKryoReader;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ConcurrentKryoWriter;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeKryo;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeNoReferenceKryo;

public class AbstractsAdapter {
	private final static String PREFIX = "ABSTRACTS_";

	private final CHMKryoAdapter kryoAdapter;

	@Inject
	public AbstractsAdapter(CHMKryoAdapter kryoAdapter) {
		this.kryoAdapter = kryoAdapter;
	}

	public boolean cacheExists(String cacheFolder) {
		return kryoAdapter.cacheExists(cacheFolder, PREFIX);
	}

	public AbstractsRepositoryImpl readFromCache(String cacheFolder) {
		Map<InstanceId, String> map = kryoAdapter.read(InstanceId.class, String.class, cacheFolder, PREFIX);

		return new AbstractsRepositoryImpl(map);
	}

	public void writeToCache(AbstractsRepositoryImpl abstractsAdapter, String cacheFolder) {
		kryoAdapter.write(cacheFolder, PREFIX, abstractsAdapter.getAbstractsMap());
	}

	public AbstractsRepositoryImpl readFromSource(String source, Set<InstanceId> instanceIds) {

		try {
			ConcurrentHashMap<InstanceId, String> map = new ConcurrentHashMap<>();

			Pattern pattern = Pattern.compile("<(.*?)> <(?:.*?)> \"(.*?)\"@en .");

			final BatchConsumerProducer<String> batchConsumerProducer = BatchConsumerProducer.create("Read Abstracts", (e) -> {
				Matcher matcher = pattern.matcher(e);

				if (matcher.find()) {
					InstanceId instanceId = new InstanceId(matcher.group(1));
					String string;
					try {
						string = ExpansionStringUnescape.unescapeJava(matcher.group(2));
						if (instanceIds == null || instanceIds.contains(instanceId)) {
							map.put(instanceId, string);
						}
					} catch (Exception e1) {
						// TODO Auto-generated catch block
						e1.printStackTrace();
					}

				}

			}, 5000, 1);

			BufferedReader br = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(source))));
			List<String> list = new ArrayList<>();
			String nextLine = null;
			int sum = 0;
			while ((nextLine = br.readLine()) != null) {
				if (!nextLine.startsWith("#")) {
					list.add(nextLine);
					// if (list.size() == 10000) {
					// batchConsumerProducer.addBatch(list);
					// break;
					// }
					if (list.size() == 200000) {
						sum += 200000;
						batchConsumerProducer.addBatch(list);
						list = new ArrayList<>();
						// break;
					}
					if (sum == 400000) {
						sum = 0;
						System.out.println("read 1000000");
						// break;
					}
				}

			}

			if (!list.isEmpty()) {
				batchConsumerProducer.addBatch(list);
			}

			batchConsumerProducer.finish();

			br.close();

			return new AbstractsRepositoryImpl(map);
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
	}

	public static void main(String[] args) {
		ThreadSafeKryo threadSafeKryo = new ThreadSafeKryo();
		ThreadSafeNoReferenceKryo threadSafeNoReferencesKryo = new ThreadSafeNoReferenceKryo();
		AbstractsAdapter adapter = new AbstractsAdapter(new CHMKryoAdapter(new ConcurrentKryoAdapter(
				new ConcurrentKryoReader(threadSafeKryo, threadSafeNoReferencesKryo), new ConcurrentKryoWriter(threadSafeKryo, threadSafeNoReferencesKryo))));
		AbstractsRepositoryImpl ret;
		if (false) {
			ret = adapter.readFromCache("D:/stuff/abstractscache/");
		} else {
			ret = adapter.readFromSource("D:/stuff/long_abstracts_en.nt.gz", null);
			adapter.writeToCache(ret, "D:/stuff/abstractscache/");
		}

		int sdf = 123;

		for (InstanceId instanceId : ret.getAbstractsMap().keySet()) {
			Map<String, Integer> map = ret.getBowMap(instanceId);
			int sasdfsddf = 13;
		}

		try {
			PersistentParallel6.close();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

}
