package de.uni_mannheim.dws.expansion.data.kb.abstracts;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.stream.Collectors;

import org.apache.commons.lang.mutable.MutableInt;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ImmuMaker;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;

public class AbstractsRepositoryImpl implements AbstractsRepository {
	private final Map<InstanceId, String> abstractsMap;

	public static final SimpleLockableConcurrentCache<InstanceId, Map<String, Integer>> cache = new SimpleLockableConcurrentCache<>();

	public AbstractsRepositoryImpl(Map<InstanceId, String> abstractsMap) {
		this.abstractsMap = ImmuMaker.map(abstractsMap);
	}

	@Override
	public String getAbstract(InstanceId instanceId) {

		return abstractsMap.get(instanceId);
	}

	@Override
	public Map<String, Integer> getBowMap(InstanceId instanceId) {
		return cache.getOrCreate(instanceId, () -> {
			Map<String, MutableInt> ret = new HashMap<>();

			String value = abstractsMap.get(instanceId);

			if (value != null) {
				value = StringNormalizer.normaliseValue(value, false);
				List<String> tokens = StringNormalizer.tokenise(value, false, false);
				for (String token : tokens) {
					CHMHelper.getFromMap(ret, token, () -> new MutableInt()).increment();
				}
			}

			return ret.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().intValue()));
		});
	}

	public Map<InstanceId, String> getAbstractsMap() {
		return abstractsMap;
	}

}
