package de.uni_mannheim.dws.expansion.data.kb.degree;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.ImmuMaker;

public class DegreeRepositoryImpl implements DegreeRepository {

	private final Map<InstanceId, Integer> degreeMap;

	public DegreeRepositoryImpl(Map<InstanceId, Integer> degreeMap) {
		this.degreeMap = ImmuMaker.map(degreeMap);
	}

	@Override
	public int getNumIndegree(InstanceId instanceId) {
		Integer degree = degreeMap.get(instanceId);
		if (degree == null)
			return 0;
		return degree;
	}

	@Override
	public Map<InstanceId, Integer> getNumIndegree(Set<InstanceId> instanceIds) {
		Map<InstanceId, Integer> ret = new HashMap<>();
		for (InstanceId instanceId : instanceIds) {
			ret.put(instanceId, getNumIndegree(instanceId));
		}

		return ret;
	}

	public Map<InstanceId, Integer> getDegreeMap() {
		return degreeMap;
	}

	public void printStatistics() {
		Map<Integer, Set<InstanceId>> map = degreeMap.entrySet().stream()
				.collect(Collectors.groupingBy((e) -> e.getValue(), Collectors.mapping((e) -> e.getKey(), Collectors.toSet())));

		List<Entry<Integer, Set<InstanceId>>> mm = map.entrySet().stream().sorted((e1, e2) -> Integer.compare(e2.getKey(), e1.getKey()))
				.collect(Collectors.toList());

		int sdf = 123;
	}

}
