package de.uni_mannheim.dws.expansion.data.kb.repository;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.ConfigInterface;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;

public class KBValueContainer implements ConfigInterface<KBValueContainer> {
	private Map<ValueId, QualifiedValueList> replaceValues;
	private Set<ValueId> removeValues;
	private Set<InstanceId> removeInstances;
	private Set<Pair<InstanceId, ClassId>> removeSetMappings;
	private Set<Pair<InstanceId, ClassId>> newSetMappings;

	public KBValueContainer(Set<Pair<InstanceId, ClassId>> newSetMappings) {
		this.replaceValues = ImmutableMap.copyOf(new HashMap<>());
		this.removeValues = ImmutableSet.copyOf(new HashSet<>());
		this.removeInstances = ImmutableSet.copyOf(new HashSet<>());
		this.newSetMappings = ImmutableSet.copyOf(newSetMappings);
		this.removeSetMappings = ImmutableSet.copyOf(new HashSet<>());
	}

	private KBValueContainer() {
		this.replaceValues = ImmutableMap.copyOf(new HashMap<>());
		this.removeValues = ImmutableSet.copyOf(new HashSet<>());
		this.removeInstances = ImmutableSet.copyOf(new HashSet<>());
		this.newSetMappings = ImmutableSet.copyOf(new HashSet<>());
		this.removeSetMappings = ImmutableSet.copyOf(new HashSet<>());
	}

	public KBValueContainer(Set<Pair<InstanceId, ClassId>> newSetMappings, Map<ValueId, QualifiedValueList> replaceValues) {
		this.replaceValues = ImmutableMap.copyOf(replaceValues);
		this.removeValues = ImmutableSet.copyOf(new HashSet<>());
		this.removeInstances = ImmutableSet.copyOf(new HashSet<>());
		this.newSetMappings = ImmutableSet.copyOf(newSetMappings);
		this.removeSetMappings = ImmutableSet.copyOf(new HashSet<>());
	}

	public KBValueContainer(Map<ValueId, QualifiedValueList> replaceValues, Set<ValueId> removeValues) {
		super();
		this.replaceValues = ImmutableMap.copyOf(replaceValues);
		this.removeValues = ImmutableSet.copyOf(removeValues);
		this.removeInstances = ImmutableSet.copyOf(new HashSet<>());
		this.newSetMappings = ImmutableSet.copyOf(new HashSet<>());
		this.removeSetMappings = ImmutableSet.copyOf(new HashSet<>());
	}

	public KBValueContainer(Set<Pair<InstanceId, ClassId>> newSetMappings, Map<ValueId, QualifiedValueList> replaceValues, Set<InstanceId> removeInstances,
			Set<ValueId> removeValues) {
		super();
		this.replaceValues = ImmutableMap.copyOf(replaceValues);
		this.removeValues = ImmutableSet.copyOf(removeValues);
		this.removeInstances = ImmutableSet.copyOf(removeInstances);
		this.newSetMappings = ImmutableSet.copyOf(newSetMappings);
		this.removeSetMappings = ImmutableSet.copyOf(new HashSet<>());
	}

	public KBValueContainer(Set<Pair<InstanceId, ClassId>> removeSetMappings, Set<Pair<InstanceId, ClassId>> newSetMappings,
			Map<ValueId, QualifiedValueList> replaceValues, Set<InstanceId> removeInstances, Set<ValueId> removeValues) {
		super();
		this.replaceValues = ImmutableMap.copyOf(replaceValues);
		this.removeValues = ImmutableSet.copyOf(removeValues);
		this.removeInstances = ImmutableSet.copyOf(removeInstances);
		this.newSetMappings = ImmutableSet.copyOf(newSetMappings);
		this.removeSetMappings = ImmutableSet.copyOf(removeSetMappings);
	}

	public KBValueContainer(Map<ValueId, QualifiedValueList> replaceValues, Set<InstanceId> removeInstances, Set<ValueId> removeValues) {
		super();
		this.replaceValues = ImmutableMap.copyOf(replaceValues);
		this.removeValues = ImmutableSet.copyOf(removeValues);
		this.removeInstances = ImmutableSet.copyOf(removeInstances);
		this.newSetMappings = ImmutableSet.copyOf(new HashSet<>());
		this.removeSetMappings = ImmutableSet.copyOf(new HashSet<>());
	}

	public Map<ValueId, QualifiedValueList> getReplaceValues() {
		return replaceValues;
	}

	public Set<ValueId> getRemoveValues() {
		return removeValues;
	}

	public Set<InstanceId> getRemoveInstances() {
		return removeInstances;
	}

	public Set<Pair<InstanceId, ClassId>> getNewSetMappings() {
		return newSetMappings;
	}

	public Set<Pair<InstanceId, ClassId>> getRemoveSetMappings() {
		return removeSetMappings;
	}

	@Override
	public KBValueContainer copy() {
		KBValueContainer container = new KBValueContainer();
		container.replaceValues = replaceValues;
		container.removeValues = removeValues;
		container.removeInstances = removeInstances;
		container.newSetMappings = newSetMappings;
		container.removeSetMappings = removeSetMappings;
		return container;
	}

	@Override
	public KBValueContainerBuilder rebuilder() {
		return new KBValueContainerBuilder(this);
	}

	public static KBValueContainer defaultConfig() {
		return new KBValueContainer();
	}

	public static KBValueContainerBuilder builder() {
		return new KBValueContainerBuilder(defaultConfig());
	}

	public static class KBValueContainerBuilder implements ConfigBuilderInterface<KBValueContainer> {

		KBValueContainer container;

		public KBValueContainerBuilder(KBValueContainer container) {
			this.container = container.copy();
		}

		public KBValueContainerBuilder setReplaceValues(Map<ValueId, QualifiedValueList> replaceValues) {
			this.container.replaceValues = replaceValues;
			return this;
		}

		public KBValueContainerBuilder setRemoveValues(Set<ValueId> removeValues) {
			this.container.removeValues = removeValues;
			return this;
		}

		public KBValueContainerBuilder setRemoveInstances(Set<InstanceId> removeInstances) {
			this.container.removeInstances = removeInstances;
			return this;
		}

		public KBValueContainerBuilder setNewSetMappings(Set<Pair<InstanceId, ClassId>> newSetMappings) {
			this.container.newSetMappings = newSetMappings;
			return this;
		}

		public KBValueContainerBuilder setRemoveSetMappings(Set<Pair<InstanceId, ClassId>> removeSetMappings) {
			this.container.removeSetMappings = removeSetMappings;
			return this;
		}

		@Override
		public KBValueContainer build() {
			return container.copy();
		}

	}

}
