package de.uni_mannheim.dws.expansion.data.kb.repository;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;

public class KBValueCreatorsRunner {
	@Inject
	public KBValueCreatorsRunner() {
	}

	public KnowledgeBaseImpl runReplacers(KnowledgeBaseImpl kb, List<KBValueCreator> replacers) {
		if (replacers != null) {
			Map<ValueId, QualifiedValueList> kbValues = new HashMap<>(kb.getValues());
			Map<InstanceId, Set<ClassId>> setMappings = new HashMap<>();
			for (Entry<InstanceId, Set<ClassId>> setMapping : kb.getSetMappings().entrySet()) {
				setMappings.put(setMapping.getKey(), new HashSet<>(setMapping.getValue()));
			}

			for (KBValueCreator replacer : replacers) {
				KBValueContainer v = replacer.getMatchedValues(kb);
				for (Pair<InstanceId, ClassId> setMapping : v.getRemoveSetMappings()) {

					CHMHelper.getFromMap(setMappings, setMapping.getFirst(), () -> new HashSet<>()).remove(setMapping.getSecond());
				}
				for (Pair<InstanceId, ClassId> setMapping : v.getNewSetMappings()) {

					CHMHelper.getFromMap(setMappings, setMapping.getFirst(), () -> new HashSet<>()).add(setMapping.getSecond());
				}
				kbValues.putAll(v.getReplaceValues());
				for (ValueId valueId : v.getRemoveValues()) {
					kbValues.remove(valueId);
				}

				for (InstanceId instanceId : v.getRemoveInstances()) {
					Set<ClassId> sets = setMappings.get(instanceId);
					setMappings.remove(instanceId);

					for (ClassId setId : sets) {
						Set<PropertyId> propertyIds = kb.getProperties(setId).keySet();
						for (PropertyId propertyId : propertyIds) {
							kbValues.remove(new ValueId(instanceId, propertyId));
						}
					}

				}

			}

			kb = KnowledgeBaseImpl.create(kbValues, setMappings, kb.getEqualProperties(), kb.getKbUnits(), kb.getNormalizationIndex());
		}

		return kb;
	}
}
