package de.uni_mannheim.dws.expansion.data.kb.repository;

import java.util.List;
import java.util.Map;
import java.util.Set;

import com.google.common.base.Optional;
import com.google.common.collect.BiMap;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.uni_mannheim.dws.expansion.data.ontology.model.measurement.UnitOfMeasurement;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;

public interface KnowledgeBase {

	Optional<Set<String>> getSurfaceForms(String string);

	Optional<Set<String>> getCandidatesSurfaceForms(String string);

	Set<ClassId> getClasses();

	Set<ClassId> getClasses(InstanceId instanceId);

	InstanceId getInstance(ClassId dbPediaClass, int index);

	int getInstanceIndex(ClassId setId, InstanceId instanceId);

	BiMap<InstanceId, Integer> getInstances(ClassId currentTable);

	Set<InstanceId> getInstances(ClassId classId, PropertyId propertyId);

	int getInstancesSize(ClassId classId, PropertyId propertyId);

	int getNumberOfColumns(ClassId classId);

	int getNumberOfInstances();

	int getNumberOfInstances(ClassId classId);

	BiMap<PropertyId, Integer> getProperties(ClassId currentTable);

	BiMap<PropertyId, Integer> getPropertiesWithQualifiers(ClassId currentTable);

	PropertyId getProperty(ClassId classId, int index);

	Integer getPropertyIndex(ClassId classId, PropertyId propertyId);

	int getTotalNumberOfClasses();

	int getTotalNumberOfColumns();

	int getTotalNumberOfRows();

	QualifiedValueList getQualifiedValueList(InstanceId instanceId, PropertyId propertyId);

	QualifiedValueList getQualifiedValueList(ValueId valueId);

	boolean hasClass(ClassId classId);

	boolean hasInstance(ClassId classId, InstanceId instanceId);

	boolean hasProperty(ClassId classId, PropertyId propertyId);

	boolean hasValue(InstanceId instanceId, PropertyId propertyId);

	boolean hasValue(ValueId valueId);

	Set<ClassPropertyId> getClassPropertyIds();

	List<AbstractObject> getValuesAsList(InstanceId instanceId, PropertyId propertyId);

	Set<String> getLabels(InstanceId instanceId);

	Map<PropertyId, Set<ClassId>> getReversePropertyIndex();

	Set<InstanceId> getAllInstances();

	Set<ValueId> getValueIds();

	Map<InstanceId, Set<ClassId>> getSetMappings();

	QualifiedValueList getQualifiedValueListWithEqualProperties(ValueId valueId);

	List<AbstractObject> getValuesAsListWithEqualProperties(InstanceId instanceId, PropertyId propertyId);

	Map<ClassPropertyId, Set<ClassPropertyId>> getEqualProperties();

	Map<PropertyId, UnitOfMeasurement> getKbUnits();

	Map<PropertyId, PropertyId> getNormalizationIndex();

}
