package de.uni_mannheim.dws.expansion.data.kb.repository;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.uni_mannheim.dws.expansion.data.ontology.model.measurement.UnitOfMeasurement;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.utils.kryo.CHMKryoAdapter;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeKryo;

public class KnowledgeBaseCryoAdapter {

	public static final String VALUES_FILE = "values";
	public static final String INSTANCES_TO_CLASS_FILE = "instancesToClass";
	public static final String INSTANCES_PER_CLASS_PROPERTY_FILE = "instancesPerClassProperty";

	public static final String CLASSES_FILE = "classes";
	public static final String PROPERTY_INDICES_FILES = "propertyIndices";
	public static final String PROPERTY_INDICES_WITH_QUALIFIERS_FILES = "propertyIndicesWithQualifiers";
	public static final String INSTANCE_INDICES_FILES = "instanceIndices";

	private final ThreadSafeKryo kryo;
	private final CHMKryoAdapter chmKryo;

	@Inject
	public KnowledgeBaseCryoAdapter(ThreadSafeKryo kryo, CHMKryoAdapter chmKryo) {

		this.kryo = kryo;
		this.chmKryo = chmKryo;
	}

	public boolean cacheExists(String cacheFolder) {
		File dirFile = new File(cacheFolder);
		if (dirFile.isDirectory()) {
			String[] dirFileList = dirFile.list();
			if (dirFileList.length > 0) {
				for (String f : dirFileList) {
					if (f.startsWith(VALUES_FILE)) {
						return true;
					}
				}

			}
		}

		return false;
	}

	@SuppressWarnings("unchecked")
	public KnowledgeBaseImpl read(String cacheFolder, Map<ClassPropertyId, Set<ClassPropertyId>> equalProperties, Map<PropertyId, UnitOfMeasurement> kbUnits,
			Map<PropertyId, PropertyId> normalizationIndex) throws IOException {

		Map<InstanceId, Set<ClassId>> instancesToClassMappings = chmKryo.readWithSet(InstanceId.class, ClassId.class, cacheFolder, INSTANCES_TO_CLASS_FILE);
		Map<ValueId, QualifiedValueList> values = chmKryo.read(ValueId.class, QualifiedValueList.class, cacheFolder, VALUES_FILE);

		if (false) {
			Set<ClassId> classes = (Set<ClassId>) kryo.read(cacheFolder + "/" + CLASSES_FILE);
			Map<ClassId, Map<PropertyId, Integer>> propertyIndices = (Map<ClassId, Map<PropertyId, Integer>>) kryo
					.read(cacheFolder + "/" + PROPERTY_INDICES_FILES);
			Map<ClassId, Map<PropertyId, Integer>> propertyIndicesWithQualifiers = (Map<ClassId, Map<PropertyId, Integer>>) kryo
					.read(cacheFolder + "/" + PROPERTY_INDICES_WITH_QUALIFIERS_FILES);

			Map<ClassId, Map<InstanceId, Integer>> instanceIndices = (Map<ClassId, Map<InstanceId, Integer>>) kryo
					.read(cacheFolder + "/" + INSTANCE_INDICES_FILES);

			Map<ClassPropertyId, Set<InstanceId>> instancesPerClassProperty = chmKryo.readWithSet(ClassPropertyId.class, InstanceId.class, cacheFolder,
					INSTANCES_PER_CLASS_PROPERTY_FILE);

			return new KnowledgeBaseImpl(classes, propertyIndices, propertyIndicesWithQualifiers, instanceIndices, values, instancesPerClassProperty,
					instancesToClassMappings, equalProperties, kbUnits, normalizationIndex);

		}

		return KnowledgeBaseImpl.create(values, instancesToClassMappings, equalProperties, kbUnits, normalizationIndex);

	}

	public void write(String cacheFolder, KnowledgeBaseImpl kb) throws IOException {
		if (false) {
			kryo.write(kb.sets, cacheFolder + "/" + CLASSES_FILE);
			kryo.write(kb.propertyIndices.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				return new HashMap<>(e.getValue());
			})), cacheFolder + "/" + PROPERTY_INDICES_FILES);
			kryo.write(kb.instanceIndices.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				return new HashMap<>(e.getValue());
			})), cacheFolder + "/" + INSTANCE_INDICES_FILES);

			kryo.write(kb.propertyIndicesWithQualifiers.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				return new HashMap<>(e.getValue());
			})), cacheFolder + "/" + PROPERTY_INDICES_WITH_QUALIFIERS_FILES);

			this.chmKryo.writeWithSet(cacheFolder, INSTANCES_PER_CLASS_PROPERTY_FILE, kb.instancesPerClassProperty);
		}

		this.chmKryo.write(cacheFolder, VALUES_FILE, kb.values);

		this.chmKryo.writeWithSet(cacheFolder, INSTANCES_TO_CLASS_FILE, kb.setMappings);

	}

}
