package de.uni_mannheim.dws.expansion.data.kb.repository.remote;

import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.google.common.base.Optional;
import com.google.common.collect.BiMap;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.uni_mannheim.dws.expansion.data.ontology.model.measurement.UnitOfMeasurement;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;

public interface KnowledgeBaseRemote extends Remote {

	com.google.common.base.Optional<Set<String>> getSurfaceForms(String string) throws RemoteException;

	Optional<Set<String>> getCandidatesSurfaceForms(String string) throws RemoteException;

	Set<ClassId> getClasses() throws RemoteException;

	Set<ClassId> getClasses(InstanceId instanceId) throws RemoteException;

	InstanceId getInstance(ClassId dbPediaClass, int index) throws RemoteException;

	int getInstanceIndex(ClassId setId, InstanceId instanceId) throws RemoteException;

	BiMap<InstanceId, Integer> getInstances(ClassId currentTable) throws RemoteException;

	Set<InstanceId> getInstances(ClassId classId, PropertyId propertyId) throws RemoteException;

	int getInstancesSize(ClassId classId, PropertyId propertyId) throws RemoteException;

	int getNumberOfColumns(ClassId classId) throws RemoteException;

	int getNumberOfInstances() throws RemoteException;

	int getNumberOfInstances(ClassId classId) throws RemoteException;

	BiMap<PropertyId, Integer> getProperties(ClassId currentTable) throws RemoteException;

	BiMap<PropertyId, Integer> getPropertiesWithQualifiers(ClassId currentTable) throws RemoteException;

	PropertyId getProperty(ClassId classId, int index) throws RemoteException;

	Integer getPropertyIndex(ClassId classId, PropertyId propertyId) throws RemoteException;

	int getTotalNumberOfClasses() throws RemoteException;

	int getTotalNumberOfColumns() throws RemoteException;

	int getTotalNumberOfRows() throws RemoteException;

	QualifiedValueList getQualifiedValueList(InstanceId instanceId, PropertyId propertyId) throws RemoteException;

	QualifiedValueList getQualifiedValueList(ValueId valueId) throws RemoteException;

	boolean hasClass(ClassId classId) throws RemoteException;

	boolean hasInstance(ClassId classId, InstanceId instanceId) throws RemoteException;

	boolean hasProperty(ClassId classId, PropertyId propertyId) throws RemoteException;

	boolean hasValue(InstanceId instanceId, PropertyId propertyId) throws RemoteException;

	boolean hasValue(ValueId valueId) throws RemoteException;

	Set<ClassPropertyId> getClassPropertyIds() throws RemoteException;

	List<AbstractObject> getValuesAsList(InstanceId instance, PropertyId propertyId) throws RemoteException;

	Set<ValueId> getValueIds() throws RemoteException;

	QualifiedValueList getQualifiedValueListWithEqualProperties(ValueId valueId) throws RemoteException;

	Map<InstanceId, Set<ClassId>> getSetMappings() throws RemoteException;

	List<AbstractObject> getValuesAsListWithEqualProperties(InstanceId instanceId, PropertyId propertyId) throws RemoteException;

	Set<String> getLabels(InstanceId instanceId) throws RemoteException;

	Map<PropertyId, Set<ClassId>> getReversePropertyIndex() throws RemoteException;

	Set<InstanceId> getAllInstances() throws RemoteException;

	Map<ClassPropertyId, Set<ClassPropertyId>> getEqualProperties() throws RemoteException;

	Map<PropertyId, UnitOfMeasurement> getKbUnits() throws RemoteException;

	Map<PropertyId, PropertyId> getNormalizationIndex() throws RemoteException;

}
