package de.uni_mannheim.dws.expansion.data.kb.repository.remote;

import java.rmi.RemoteException;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import com.google.common.collect.BiMap;

import de.mannheim.uni.expansion.data.RemoteInterfaceWithCache;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.measurement.UnitOfMeasurement;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;

public class KnowledgeBaseRemoteImpl extends RemoteInterfaceWithCache implements KnowledgeBase {

	private KnowledgeBaseRemote remote;

	public KnowledgeBaseRemoteImpl(KnowledgeBaseRemote remote) {
		this.remote = remote;
	}

	Set<ClassId> classes;

	@Override
	public Set<ClassId> getClasses() {
		try {
			if (classes == null)
				classes = remote.getClasses();
			return classes;
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return null;
	}

	@Override
	public BiMap<InstanceId, Integer> getInstances(ClassId currentTable) {
		try {
			return remote.getInstances(currentTable);
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return null;
	}

	@Override
	public int getTotalNumberOfColumns() {
		try {
			return remote.getTotalNumberOfColumns();
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return 0;
	}

	@Override
	public BiMap<PropertyId, Integer> getProperties(ClassId currentTable) {
		try {
			return remote.getProperties(currentTable);
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return null;
	}

	@Override
	public int getTotalNumberOfClasses() {
		try {
			return remote.getTotalNumberOfClasses();
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return 0;
	}

	@Override
	public int getNumberOfColumns(ClassId classId) {
		try {
			return remote.getNumberOfColumns(classId);
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return 0;
	}

	@Override
	public InstanceId getInstance(ClassId dbPediaClass, int index) {
		try {
			return remote.getInstance(dbPediaClass, index);
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return null;
	}

	@Override
	public int getInstanceIndex(ClassId classId, InstanceId instanceId) {
		try {
			return remote.getInstanceIndex(classId, instanceId);
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return 0;
	}

	@Override
	public int getNumberOfInstances() {
		try {
			return remote.getNumberOfInstances();
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return 0;
	}

	@Override
	public int getNumberOfInstances(ClassId classId) {
		try {
			return remote.getNumberOfInstances(classId);
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return 0;
	}

	ConcurrentHashMap<InstanceId, Optional<Set<ClassId>>> classesCache = new ConcurrentHashMap<>();

	@Override
	public Set<ClassId> getClasses(InstanceId instanceId) {

		return cacheGetter(classesCache, instanceId, (k) -> {
			return remote.getClasses(k);
		});

	}

	@Override
	public PropertyId getProperty(ClassId classId, int index) {
		try {
			return remote.getProperty(classId, index);
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return null;
	}

	@Override
	public boolean hasValue(ValueId valueId) {
		try {
			return remote.hasValue(valueId);
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return false;
	}

	@Override
	public boolean hasValue(InstanceId instanceId, PropertyId propertyId) {
		try {
			return remote.hasValue(instanceId, propertyId);
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return false;
	}

	@Override
	public Set<InstanceId> getInstances(ClassId classId, PropertyId propertyId) {
		try {
			return remote.getInstances(classId, propertyId);
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return null;
	}

	@Override
	public int getInstancesSize(ClassId classId, PropertyId propertyId) {
		try {
			return remote.getInstancesSize(classId, propertyId);
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return 0;
	}

	@Override
	public int getTotalNumberOfRows() {
		try {
			return remote.getTotalNumberOfRows();
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return 0;
	}

	ConcurrentHashMap<ValueId, Optional<QualifiedValueList>> values = new ConcurrentHashMap<>();

	@Override
	public QualifiedValueList getQualifiedValueList(ValueId key) {
		return cacheGetter(values, key, (k) -> {
			return remote.getQualifiedValueList(k);
		});
	}

	@Override
	public boolean hasClass(ClassId classId) {
		try {
			return remote.hasClass(classId);
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return false;
	}

	@Override
	public BiMap<PropertyId, Integer> getPropertiesWithQualifiers(ClassId currentTable) {
		try {
			return remote.getPropertiesWithQualifiers(currentTable);
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return null;
	}

	@Override
	public Integer getPropertyIndex(ClassId classId, PropertyId propertyId) {
		try {
			return remote.getPropertyIndex(classId, propertyId);
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return null;
	}

	@Override
	public QualifiedValueList getQualifiedValueList(InstanceId instanceId, PropertyId propertyId) {
		return this.getQualifiedValueList(new ValueId(instanceId, propertyId));
	}

	@Override
	public boolean hasInstance(ClassId classId, InstanceId instanceId) {
		try {
			return remote.hasInstance(classId, instanceId);
		} catch (RemoteException e) {

			e.printStackTrace();
		}

		return false;
	}

	@Override
	public boolean hasProperty(ClassId classId, PropertyId propertyId) {
		try {
			return remote.hasProperty(classId, propertyId);
		} catch (RemoteException e) {

			e.printStackTrace();
		}
		return false;
	}

	@Override
	public Set<ClassPropertyId> getClassPropertyIds() {
		try {
			return remote.getClassPropertyIds();
		} catch (RemoteException e) {

			e.printStackTrace();
		}

		return null;
	}

	ConcurrentHashMap<ValueId, Optional<List<AbstractObject>>> valuesAsList = new ConcurrentHashMap<>();

	@Override
	public List<AbstractObject> getValuesAsList(InstanceId instanceId, PropertyId propertyId) {
		return cacheGetter(valuesAsList, new ValueId(instanceId, propertyId), (k) -> {
			return remote.getValuesAsList(k.getInstanceId(), k.getPropertyId());
		});

	}

	ConcurrentHashMap<InstanceId, Optional<Set<String>>> labels = new ConcurrentHashMap<>();

	@Override
	public Set<String> getLabels(InstanceId instanceId) {
		return cacheGetter(labels, instanceId, (k) -> {
			return remote.getLabels(k);
		});
	}

	@Override
	public Map<PropertyId, Set<ClassId>> getReversePropertyIndex() {
		try {
			return remote.getReversePropertyIndex();
		} catch (RemoteException e) {

			e.printStackTrace();
		}

		return null;
	}

	@Override
	public Set<InstanceId> getAllInstances() {
		try {
			return remote.getAllInstances();
		} catch (RemoteException e) {

			e.printStackTrace();
		}

		return null;
	}

	@Override
	public Set<ValueId> getValueIds() {
		try {
			return remote.getValueIds();
		} catch (RemoteException e) {

			e.printStackTrace();
		}

		return null;
	}

	@Override
	public Map<InstanceId, Set<ClassId>> getSetMappings() {
		try {
			return remote.getSetMappings();
		} catch (RemoteException e) {

			e.printStackTrace();
		}

		return null;
	}

	ConcurrentHashMap<ValueId, Optional<QualifiedValueList>> valuesWithEqualProperties = new ConcurrentHashMap<>();

	@Override
	public QualifiedValueList getQualifiedValueListWithEqualProperties(ValueId valueId) {
		return cacheGetter(valuesWithEqualProperties, valueId, (k) -> {
			return remote.getQualifiedValueListWithEqualProperties(k);
		});
	}

	ConcurrentHashMap<ValueId, Optional<List<AbstractObject>>> valuesAsListWithEqualProperties = new ConcurrentHashMap<>();

	@Override
	public List<AbstractObject> getValuesAsListWithEqualProperties(InstanceId instanceId, PropertyId propertyId) {
		return cacheGetter(valuesAsListWithEqualProperties, new ValueId(instanceId, propertyId), (k) -> {
			return remote.getValuesAsListWithEqualProperties(k.getInstanceId(), k.getPropertyId());
		});
	}

	volatile Map<ClassPropertyId, Set<ClassPropertyId>> equalProperties;

	@Override
	public Map<ClassPropertyId, Set<ClassPropertyId>> getEqualProperties() {
		if (equalProperties == null) {
			try {
				equalProperties = remote.getEqualProperties();
			} catch (RemoteException e) {

				e.printStackTrace();
			}
		}

		return equalProperties;
	}

	volatile Map<PropertyId, UnitOfMeasurement> kbUnits;

	@Override
	public Map<PropertyId, UnitOfMeasurement> getKbUnits() {

		if (kbUnits == null) {
			try {
				kbUnits = remote.getKbUnits();
			} catch (RemoteException e) {

				e.printStackTrace();
			}
		}

		return kbUnits;
	}

	volatile Map<PropertyId, PropertyId> normalizationIndex;

	@Override
	public Map<PropertyId, PropertyId> getNormalizationIndex() {
		if (normalizationIndex == null) {
			try {
				normalizationIndex = remote.getNormalizationIndex();
			} catch (RemoteException e) {

				e.printStackTrace();
			}
		}

		return normalizationIndex;
	}

	ConcurrentHashMap<String, Optional<com.google.common.base.Optional<Set<String>>>> surfaceForms = new ConcurrentHashMap<>();

	@Override
	public com.google.common.base.Optional<Set<String>> getSurfaceForms(String string) {
		return cacheGetter(surfaceForms, string, (k) -> {
			return remote.getSurfaceForms(k);
		});
	}

	ConcurrentHashMap<String, Optional<com.google.common.base.Optional<Set<String>>>> candidatesSurfaceForms = new ConcurrentHashMap<>();

	@Override
	public com.google.common.base.Optional<Set<String>> getCandidatesSurfaceForms(String string) {
		return cacheGetter(candidatesSurfaceForms, string, (k) -> {
			return remote.getCandidatesSurfaceForms(k);
		});
	}

	// ConcurrentHashMap<ClassPropertyId, Object> propertyExists = new
	// ConcurrentHashMap<ClassPropertyId, Object>();
	//
	// private ValueGetter<ClassPropertyId> propertyExistsGetter = (key) -> {
	// return (Boolean) remote.propertyExists(key.getClassId(),
	// key.getPropertyId());
	// };
	//
	// @Override
	// public boolean propertyExists(ClassId classId, PropertyId propertyId) {
	// return (boolean) cacheGetter(propertyExists, new ClassPropertyId(classId,
	// propertyId), propertyExistsGetter);
	// }

}
