package de.uni_mannheim.dws.expansion.data.utils.kryo;

import java.io.File;
import java.lang.reflect.Array;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;

import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Quad;
import de.mannheim.uni.expansion.utils.containers.Tri;

public class CHMKryoAdapter {
	private final ConcurrentKryoAdapter kryoAdapter;

	@Inject
	public CHMKryoAdapter(ConcurrentKryoAdapter kryoAdapter) {
		this.kryoAdapter = kryoAdapter;
	}

	public boolean cacheExists(String cacheFolder, String PREFIX) {
		File dirFile = new File(cacheFolder);
		if (dirFile.isDirectory()) {
			String[] dirFileList = dirFile.list();
			if (dirFileList.length > 0) {
				for (String f : dirFileList) {
					if (f.startsWith(PREFIX)) {
						return true;
					}
				}

			}
		}
		return false;
	}

	public <K, V> Map<K, V> read(Class<K> keyClass, Class<V> valueClass, String folder, String prefix) {
		ConcurrentHashMap<K, V> chm = new ConcurrentHashMap<>();

		kryoAdapter.runForEachTable(Pair.class, (e) -> {
			@SuppressWarnings("unchecked")
			Pair<K, V> pair = e;
			chm.put(pair.getFirst(), pair.getSecond());
		}, folder, prefix);

		return chm;
	}

	@Deprecated
	public <K, V> Map<K, V> readWithReferences(Class<K> keyClass, Class<V> valueClass, String folder, String prefix) {
		ConcurrentHashMap<K, V> chm = new ConcurrentHashMap<>();

		kryoAdapter.runForEachTableWithReferences(Pair.class, (e) -> {
			@SuppressWarnings("unchecked")
			Pair<K, V> pair = e;
			chm.put(pair.getFirst(), pair.getSecond());
		}, folder, prefix);

		return chm;
	}

	public <K, V> void write(String folder, String prefix, Map<K, V> values) {
		kryoAdapter.writeAll(new Pair<K, V>(null, null), new MapPairIterable<K, V>(values), values.size(), folder, prefix, 32);

	}

	public <K> void write(String folder, String prefix, Set<K> values, K poison) {
		kryoAdapter.writeAll(poison, values, values.size(), folder, prefix, 32);

	}

	@Deprecated
	public <K> void writeWithRefrences(String folder, String prefix, Set<K> values, K poison) {
		kryoAdapter.writeAllWithReferences(poison, values, values.size(), folder, prefix, 32);

	}

	public <K> Set<K> read(Class<K> keyClass, String folder, String prefix) {
		Set<K> set = Collections.newSetFromMap(new ConcurrentHashMap<>());
		kryoAdapter.runForEachTable(keyClass, (e) -> {
			set.add(e);
		}, folder, prefix);

		return set;
	}

	@Deprecated
	public <K> Set<K> readWithReferences(Class<K> keyClass, String folder, String prefix) {
		Set<K> set = Collections.newSetFromMap(new ConcurrentHashMap<>());
		kryoAdapter.runForEachTableWithReferences(keyClass, (e) -> {
			set.add(e);
		}, folder, prefix);

		return set;
	}

	public <K, V> Map<K, Set<V>> readWithSet(Class<K> keyClass, Class<V> valueClass, String folder, String prefix) {
		ConcurrentHashMap<K, Set<V>> chm = new ConcurrentHashMap<>();

		kryoAdapter.runForEachTable(Pair.class, (e) -> {
			@SuppressWarnings("unchecked")
			Pair<K, V> pair = e;

			CHMHelper.getFromCHM(chm, pair.getFirst(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(pair.getSecond());

		}, folder, prefix);

		return chm;

	}

	public <K, V> void writeWithSet(String folder, String prefix, Map<K, Set<V>> values) {
		int size = values.entrySet().stream().mapToInt((e) -> {
			return e.getValue().size();
		}).sum();
		kryoAdapter.writeAll(new Pair<K, V>(null, null), new MapSetPairIterable<K, V>(values), size, folder, prefix, 32);

	}

	public <K, I, V> Map<K, Map<I, V>> readWithMap(Class<K> keyClass, Class<I> indexClass, Class<V> valueClass, String folder, String prefix) {
		ConcurrentHashMap<K, Map<I, V>> chm = new ConcurrentHashMap<>();

		kryoAdapter.runForEachTable(Tri.class, (e) -> {
			@SuppressWarnings("unchecked")
			Tri<K, I, V> tri = e;

			CHMHelper.getFromCHM(chm, tri.getA(), () -> new ConcurrentHashMap<>()).put(tri.getB(), tri.getC());

		}, folder, prefix);

		return chm;

	}

	@Deprecated
	public <K, I, V> Map<K, Map<I, V>> readWithMapWithReferences(Class<K> keyClass, Class<I> indexClass, Class<V> valueClass, String folder, String prefix) {
		ConcurrentHashMap<K, Map<I, V>> chm = new ConcurrentHashMap<>();

		kryoAdapter.runForEachTableWithReferences(Tri.class, (e) -> {
			@SuppressWarnings("unchecked")
			Tri<K, I, V> tri = e;

			CHMHelper.getFromCHM(chm, tri.getA(), () -> new ConcurrentHashMap<>()).put(tri.getB(), tri.getC());

		}, folder, prefix);

		return chm;

	}

	public <K, I, V> void writeWithMap(String folder, String prefix, Map<K, Map<I, V>> values) {
		int size = values.entrySet().stream().mapToInt((e) -> {
			return e.getValue().size();
		}).sum();
		kryoAdapter.writeAll(new Tri<K, I, V>(null, null, null), new MapMapPairIterable<K, I, V>(values), size, folder, prefix, 32);

	}

	@Deprecated
	public <K, I, V> void writeWithMapWithReferences(String folder, String prefix, Map<K, Map<I, V>> values) {
		int size = values.entrySet().stream().mapToInt((e) -> {
			return e.getValue().size();
		}).sum();
		kryoAdapter.writeAllWithReferences(new Tri<K, I, V>(null, null, null), new MapMapPairIterable<K, I, V>(values), size, folder, prefix, 32);

	}

	@SuppressWarnings("unchecked")
	public <K, V> Map<K, List<V>> readWithList(Class<K> keyClass, Class<V> valueClass, String folder, String prefix) {
		ConcurrentHashMap<K, V[]> chm = new ConcurrentHashMap<>();

		kryoAdapter.runForEachTable(Quad.class, (e) -> {
			Quad<K, V, Integer, Integer> pair = e;

			CHMHelper.getFromCHM(chm, pair.getA(), () -> (V[]) Array.newInstance(valueClass, pair.getD()))[pair.getC()] = pair.getB();

		}, folder, prefix);

		ConcurrentHashMap<K, List<V>> ret = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(chm.entrySet(), (e) -> {
			ret.put(e.getKey(), Arrays.asList(e.getValue()));

		}, this.getClass());

		return ret;

	}

	@Deprecated
	@SuppressWarnings("unchecked")
	public <K, V> Map<K, List<V>> readWithListWithReferences(Class<K> keyClass, Class<V> valueClass, String folder, String prefix) {
		ConcurrentHashMap<K, V[]> chm = new ConcurrentHashMap<>();

		kryoAdapter.runForEachTableWithReferences(Quad.class, (e) -> {
			Quad<K, V, Integer, Integer> pair = e;

			CHMHelper.getFromCHM(chm, pair.getA(), () -> (V[]) Array.newInstance(valueClass, pair.getD()))[pair.getC()] = pair.getB();

		}, folder, prefix);

		ConcurrentHashMap<K, List<V>> ret = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(chm.entrySet(), (e) -> {
			ret.put(e.getKey(), Arrays.asList(e.getValue()));

		}, this.getClass());

		return ret;

	}

	@Deprecated
	public <K, V> void writeWithListWithReferences(String folder, String prefix, Map<K, List<V>> values) {
		int size = values.entrySet().stream().mapToInt((e) -> {
			return e.getValue().size();
		}).sum();
		kryoAdapter.writeAllWithReferences(new Quad<K, V, Integer, Integer>(null, null, null, null), new MapListPairIterable<K, V>(values), size, folder,
				prefix, 32);

	}

	public <K, V> void writeWithList(String folder, String prefix, Map<K, List<V>> values) {
		int size = values.entrySet().stream().mapToInt((e) -> {
			return e.getValue().size();
		}).sum();
		kryoAdapter.writeAll(new Quad<K, V, Integer, Integer>(null, null, null, null), new MapListPairIterable<K, V>(values), size, folder, prefix, 32);

	}

	private static class MapPairIterable<K, V> implements Iterable<Pair<K, V>> {

		private final Map<K, V> map;

		public MapPairIterable(Map<K, V> map) {
			this.map = map;
		}

		@Override
		public Iterator<Pair<K, V>> iterator() {
			return new MapPairIterator();
		}

		private class MapPairIterator implements Iterator<Pair<K, V>> {
			Iterator<Entry<K, V>> it = map.entrySet().iterator();

			@Override
			public boolean hasNext() {
				return it.hasNext();
			}

			@Override
			public Pair<K, V> next() {
				Entry<K, V> next = it.next();
				if (next != null) {
					return new Pair<K, V>(next.getKey(), next.getValue());
				}
				return null;
			}

		}

	}

	private static class MapSetPairIterable<K, V> implements Iterable<Pair<K, V>> {

		private final Map<K, Set<V>> map;

		public MapSetPairIterable(Map<K, Set<V>> map) {
			this.map = map;
		}

		@Override
		public Iterator<Pair<K, V>> iterator() {
			return new MapPairIterator();
		}

		private class MapPairIterator implements Iterator<Pair<K, V>> {
			Iterator<Entry<K, Set<V>>> it = map.entrySet().iterator();

			Iterator<V> it2 = null;
			K k2 = null;

			@Override
			public boolean hasNext() {

				if (it2 != null && it2.hasNext()) {
					return true;
				} else {
					it2 = null;
					k2 = null;

					if (it.hasNext()) {
						it2 = null;

						return true;
					}
					return false;
				}

			}

			@Override
			public Pair<K, V> next() {
				if (it2 == null) {
					Entry<K, Set<V>> n = it.next();
					if (n != null && n.getValue() != null) {
						it2 = n.getValue().iterator();
						k2 = n.getKey();
					}
				}

				if (it2 != null && k2 != null) {
					return new Pair<K, V>(k2, it2.next());
				} else {
					return null;

				}

			}

		}

	}

	private static class MapMapPairIterable<K, I, V> implements Iterable<Tri<K, I, V>> {

		private final Map<K, Map<I, V>> map;

		public MapMapPairIterable(Map<K, Map<I, V>> map) {
			this.map = map;
		}

		@Override
		public Iterator<Tri<K, I, V>> iterator() {
			return new MapPairIterator();
		}

		private class MapPairIterator implements Iterator<Tri<K, I, V>> {
			Iterator<Entry<K, Map<I, V>>> it = map.entrySet().iterator();

			Iterator<Entry<I, V>> it2 = null;
			K k2 = null;

			@Override
			public boolean hasNext() {

				if (it2 != null && it2.hasNext()) {
					return true;
				} else {
					it2 = null;
					k2 = null;

					if (it.hasNext()) {
						it2 = null;

						return true;
					}
					return false;
				}

			}

			@Override
			public Tri<K, I, V> next() {
				if (it2 == null) {
					Entry<K, Map<I, V>> n = it.next();
					if (n != null && n.getValue() != null) {
						it2 = n.getValue().entrySet().iterator();
						k2 = n.getKey();
					}
				}

				if (it2 != null && k2 != null) {
					Entry<I, V> next = it2.next();
					return new Tri<K, I, V>(k2, next.getKey(), next.getValue());
				} else {
					return null;

				}

			}

		}

	}

	private static class MapListPairIterable<K, V> implements Iterable<Quad<K, V, Integer, Integer>> {

		private final Map<K, List<V>> map;

		public MapListPairIterable(Map<K, List<V>> map) {
			this.map = map;
		}

		@Override
		public Iterator<Quad<K, V, Integer, Integer>> iterator() {
			return new MapPairIterator();
		}

		private class MapPairIterator implements Iterator<Quad<K, V, Integer, Integer>> {
			Iterator<Entry<K, List<V>>> it = map.entrySet().iterator();

			Iterator<ListEntry<V>> it2 = null;
			K k2 = null;
			Integer l2Size = null;

			@Override
			public boolean hasNext() {

				if (it2 != null && it2.hasNext()) {
					return true;
				} else {
					it2 = null;
					k2 = null;
					l2Size = null;

					if (it.hasNext()) {
						it2 = null;

						return true;
					}
					return false;
				}

			}

			@Override
			public Quad<K, V, Integer, Integer> next() {
				if (it2 == null) {
					Entry<K, List<V>> n = it.next();
					if (n != null && n.getValue() != null) {
						l2Size = n.getValue().size();
						it2 = ListIterable.iterable(n.getValue()).iterator();
						k2 = n.getKey();

					}
				}

				if (it2 != null && k2 != null && l2Size != null) {
					ListEntry<V> entry = it2.next();
					return new Quad<K, V, Integer, Integer>(k2, entry.getValue(), entry.getIndex(), l2Size);
				} else {
					return null;

				}

			}

		}

	}
}
