package de.uni_mannheim.dws.expansion.data.utils.kryo;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Consumer;

import javax.inject.Inject;

import org.apache.commons.compress.archivers.tar.TarArchiveEntry;
import org.apache.commons.compress.archivers.tar.TarArchiveInputStream;
import org.apache.commons.compress.compressors.gzip.GzipCompressorInputStream;

import com.esotericsoftware.kryo.Kryo;
import com.esotericsoftware.kryo.io.Input;

import de.mannheim.uni.expansion.utils.BatchConsumerProducer;
import de.mannheim.uni.expansion.utils.PersistentParallel6;

public class ConcurrentKryoReader {

	private final ThreadSafeKryo threadSafeWithReferenceKryo;
	private final ThreadSafeNoReferenceKryo threadSafeKryo;

	@Inject
	public ConcurrentKryoReader(ThreadSafeKryo threadSafeWithReferenceKryo, ThreadSafeNoReferenceKryo threadSafeKryo) {
		this.threadSafeKryo = threadSafeKryo;
		this.threadSafeWithReferenceKryo = threadSafeWithReferenceKryo;
	}

	private List<byte[]> readGzFile(File file) throws IOException {
		List<byte[]> contents = new ArrayList<>();

		TarArchiveInputStream tarStream = new TarArchiveInputStream(new GzipCompressorInputStream(new FileInputStream(file)));

		TarArchiveEntry entry;

		while ((entry = tarStream.getNextTarEntry()) != null) {

			byte[] content = new byte[(int) entry.getSize()];

			tarStream.read(content, 0, content.length);

			contents.add(content);
			// String name = entry.getName().split("/")[1];

		}

		tarStream.close();

		return contents;
	}

	public <T> void runForEachTable(Class<T> clazz, Consumer<T> fullConsumer, String folder, String prefix) {
		runForEachTable(clazz, fullConsumer, folder, prefix, 40);
	}

	public <T> void runForEachTable(Class<T> clazz, Consumer<T> fullConsumer, String folder, String prefix, int batchSize, boolean reportProgress) {
		runForEachTable(clazz, fullConsumer, folder, prefix, batchSize, reportProgress, false);
	}

	@Deprecated
	public <T> void runForEachTableWithReferences(Class<T> clazz, Consumer<T> fullConsumer, String folder, String prefix) {
		runForEachTable(clazz, fullConsumer, folder, prefix, 40, false, true);
	}

	private <T> void runForEachTable(Class<T> clazz, Consumer<T> fullConsumer, String folder, String prefix, int batchSize, boolean reportProgress,
			boolean withReferences) {

		final BatchConsumerProducer<byte[]> batchConsumerProducer;
		batchConsumerProducer = BatchConsumerProducer.create("READ KRYO", (content) -> {
			Input input = new Input(new ByteArrayInputStream(content));

			Kryo kryo;
			if (withReferences) {
				kryo = threadSafeWithReferenceKryo.getKryo();
			} else {
				kryo = threadSafeKryo.getKryo();
			}

			T data = kryo.readObject(input, clazz);

			input.close();

			fullConsumer.accept(data);

		}, 2000, 2);

		File[] files = new File(folder).listFiles((dir, name) -> {
			return name.startsWith(prefix);
		});

		try {
			AtomicInteger ai = new AtomicInteger(0);
			PersistentParallel6.runAndFinish(Arrays.asList(files), (file) -> {
				System.err.println("Now reading " + prefix + " " + ai.incrementAndGet() + " out of " + files.length);

				try {

					batchConsumerProducer.addBatch(readGzFile(file));
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}, this.getClass(), 12);

			batchConsumerProducer.finish();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public <T> void runForEachTable(Class<T> clazz, Consumer<T> fullConsumer, String folder, String prefix, int batchSize) {
		runForEachTable(clazz, fullConsumer, folder, prefix, batchSize, false);

	}
}
