package de.uni_mannheim.dws.expansion.data.utils.kryo;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import javax.inject.Inject;

import org.objenesis.strategy.StdInstantiatorStrategy;

import com.esotericsoftware.kryo.Kryo;
import com.esotericsoftware.kryo.Kryo.DefaultInstantiatorStrategy;
import com.esotericsoftware.kryo.io.Input;
import com.esotericsoftware.kryo.io.Output;
import com.esotericsoftware.kryo.serializers.CompatibleFieldSerializer;
import com.esotericsoftware.kryo.util.MapReferenceResolver;

import de.javakaffee.kryoserializers.UnmodifiableCollectionsSerializer;
import de.javakaffee.kryoserializers.guava.ImmutableListSerializer;
import de.javakaffee.kryoserializers.guava.ImmutableMapSerializer;
import de.javakaffee.kryoserializers.guava.ImmutableSetSerializer;
import de.mannheim.uni.expansion.data.tables.reader.FixedClassResolver;

public class ThreadSafeKryo {

	private final static ThreadLocal<Kryo> kryos = new ThreadLocal<Kryo>();

	@Inject
	public ThreadSafeKryo() {

	}

	public Kryo getKryo() {
		Kryo kryo = kryos.get();

		if (kryo == null) {
			kryo = new Kryo(new FixedClassResolver(), new MapReferenceResolver());
			kryo.setInstantiatorStrategy(new DefaultInstantiatorStrategy(new StdInstantiatorStrategy()));
			kryo.setDefaultSerializer(CompatibleFieldSerializer.class);
			// kryo.addDefaultSerializer(AnnotatedTablesReaderData.class, new
			// AnnotatedTableReaderDataSerializer());

			ImmutableSetSerializer.registerSerializers(kryo);
			ImmutableMapSerializer.registerSerializers(kryo);
			ImmutableListSerializer.registerSerializers(kryo);

			UnmodifiableCollectionsSerializer.registerSerializers(kryo);
			// kryo.setReferences(false);
			kryos.set(kryo);
		}
		// kryo.reset();
		return kryo;

	}

	public <T> void write(T object, String file) throws IOException {
		FileOutputStream fos = new FileOutputStream(file);
		Output output = new Output(fos);
		getKryo().writeClassAndObject(output, object);
		output.close();
		fos.close();
	}

	public <T> void writeWithoutClass(T object, String file) throws IOException {
		FileOutputStream fos = new FileOutputStream(file);
		Output output = new Output(fos);
		getKryo().writeObject(output, object);
		output.close();
		fos.close();
	}

	public <T> T read(Class<T> clazz, String file) throws IOException {
		FileInputStream fis = new FileInputStream(file);
		Input input = new Input(fis);
		T o = getKryo().readObject(input, clazz);
		input.close();
		fis.close();
		return o;
	}

	public Object read(String file) throws IOException {
		FileInputStream fis = new FileInputStream(file);
		Input input = new Input(fis);
		Object o = getKryo().readClassAndObject(input);
		input.close();
		fis.close();
		return o;
	}
}
