package de.uni_mannheim.informatik.dws.expansion.evaluation;

import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleIdContainer;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository; 
import de.mannheim.uni.expansion.tablefusion.core.similarity.QualifiedSimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.similarity.Similarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.Similarity.SimilarityType;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.similarity.list.OverlapSimilarity; 
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.evaluation.timestamps.TimestampQualifiersAdapter;

public class ValueEvaluator {

	private final KnowledgeBase dbPediaRepository;

	private final OntologyRepository ontologyRepository;

	private final SimilarityCalculator similarityCalculator;

	// private final PersistentParallel<Entry<TripleId, Object>> parallel;

	private final OverlapSimilarity overlapSimilarity;

	private final QualifiedSimilarityCalculator qualifiedSimilarityCalculator;

	private final TimestampQualifiersAdapter timestampQualifiersAdapter;

	@Inject
	public ValueEvaluator(KnowledgeBase dbPediaRepository, SimilarityCalculator similarityCalculator, OntologyRepository ontologyRepository,
			OverlapSimilarity overlapSimilarity, QualifiedSimilarityCalculator qualifiedSimilarityCalculator,
			TimestampQualifiersAdapter timestampQualifiersAdapter) {
		this.dbPediaRepository = dbPediaRepository;
		this.similarityCalculator = similarityCalculator;
		// this.parallel = new PersistentParallel<>("AccuracyCalculator");
		this.ontologyRepository = ontologyRepository;
		this.overlapSimilarity = overlapSimilarity;
		this.qualifiedSimilarityCalculator = qualifiedSimilarityCalculator;
		this.timestampQualifiersAdapter = timestampQualifiersAdapter;
	}

	public Similarity getSimilarity(AbstractObject correctValue, AbstractObject fusedValue, ValueType cdt, SimilarityType type) {
		return new Similarity(this.similarityCalculator.getSimilarity(correctValue, fusedValue), type, cdt);
	}

	// public Similarity getSimilarity(TripleId tripleId, Object fusedValue,
	// boolean fuseLists) {
	//
	// boolean isList =
	// ontologyRepository.isListType(tripleId.getClassPropertyId().getPropertyId());
	// SimilarityType type = SimilarityType.DIRECT;
	// if (fuseLists && isList) {
	// type = SimilarityType.LIST_OVERLAP;
	// }
	// QualifiedValueList correctValue =
	// dbPediaRepository.getQualifiedValueList(tripleId.getValueId());
	// if (correctValue == null)
	// return null;
	// ValueType cdt = ontologyRepository.getDataType(tripleId.getPropertyId());
	// if (cdt == null) {
	// System.out.println("B: CDT == null, please check!");
	// return null;
	// }
	// return new Similarity(this.similarityCalculator.getSimilarity(cdt,
	// correctValue, fusedValue), type, cdt);
	// }

	// public boolean correct(Similarity similarity, ColumnDataType cdt) {
	//
	// if (similarity.getSim() >
	// similarityCalculator.getEquivalenceThreshold(cdt)) {
	// return true;
	// } else {
	// return false;
	// }
	//
	// }

	// public boolean correct(Similarity similarity, TripleId key) {
	// ColumnDataType cdt = ontologyRepository.getDataType(key.getClassId(),
	// key.getPropertyId());
	// return correct(similarity, cdt);
	// }

	// public boolean correct(TripleId tripleId, Object correctValue, Object
	// fusedValue, ColumnDataType cdt, Set<Object> filteredValues) {
	//
	// if (correctValue == null)
	// return false;
	// else if (cdt == null) {
	// System.out.println("B: CDT == null, please check!");
	// return false;
	// }
	//
	// else if (filteredValues != null && fusedValue instanceof List<?>) {
	//
	// List<Object> fusedList = new ArrayList<>((List<Object>) fusedValue);
	//
	// List<Object> correctList = new ArrayList<>(filteredValues);
	//
	// double sim = similarityCalculator.getSimilarity(cdt, fusedList,
	// correctList);
	// if (sim > 0.95) {
	// return true;
	// } else {
	// return false;
	// }
	//
	// } else {
	// return correct(correctValue, fusedValue, cdt);
	// }
	// // double sim = this.similarityCalculator.getSimilarity(cdt,
	// // correctValue, fusedValue);
	// // double equivlanceThreshold =
	// // this.similarityCalculator.getEquivalenceThreshold(cdt);
	// //
	// // if (sim > equivlanceThreshold)
	// // return true;
	// // else
	// // return false;
	// }

	public <E extends TripleIdContainer> Optional<Boolean> correct(E tripleId, FusedValue fusedValue) {
		QualifiedValueList correctValue = this.dbPediaRepository.getQualifiedValueListWithEqualProperties(tripleId.tripleId().getValueId());
		if (correctValue == null) {
			return Optional.empty();
		} else {
			if (tripleId instanceof TripleId) {
				return correctIntern((TripleId) tripleId, correctValue, fusedValue);
			}

			return correctIntern((QualifiedTripleId) tripleId, correctValue, fusedValue);
		}
	}

	public <E> Optional<Boolean> correct(E tripleId, QualifiedValueList correctValue, FusedValue fusedValue, ValueType cdt) {
		if (tripleId instanceof TripleId) {
			return correctIntern((TripleId) tripleId, correctValue, fusedValue);
		}

		return correctIntern((QualifiedTripleId) tripleId, correctValue, fusedValue);
	}

	public <E> Optional<Boolean> correct(E tripleId, QualifiedValueList correctValue, FusedValue fusedValue, ValueType cdt, FilteredValues filteredValues) {
		if (tripleId instanceof TripleId) {
			return correctIntern((TripleId) tripleId, correctValue, fusedValue);
		}

		return correctIntern((QualifiedTripleId) tripleId, correctValue, fusedValue);
	}

	public <E extends TripleIdContainer> Optional<Double> score(E tripleId, FusedValue fusedValue) {
		QualifiedValueList correctValue = this.dbPediaRepository.getQualifiedValueListWithEqualProperties(tripleId.tripleId().getValueId());
		if (correctValue == null) {
			return Optional.empty();
		} else { 
				return Optional.of(scoreIntern(correctValue, fusedValue.getSingle()));
			 
		}
	}
	
	public <E extends TripleIdContainer> Optional<Double> score(ValueId vid, AbstractObject fusedValue) {
		QualifiedValueList correctValue = this.dbPediaRepository.getQualifiedValueListWithEqualProperties(vid);
		if (correctValue == null) {
			return Optional.empty();
		} else { 
				return Optional.of(scoreIntern( correctValue, fusedValue )); 
		}
	}

	private double scoreIntern(QualifiedTripleId tripleId, QualifiedValueList correctValue, AbstractObject fusedValue) {

		Set<AbstractObject> correctValues = timestampQualifiersAdapter.getForQualifiers(correctValue, tripleId.getQualifiers());

		if (correctValues.isEmpty()) {
			return 0.0;
		} else {
			Double maxSim = null;
			for (AbstractObject cao : correctValues) {
				double sim = this.similarityCalculator.getSimilarity(cao, fusedValue);
				if (maxSim == null || sim > maxSim) {
					maxSim = sim;
				}
			}
			return maxSim;
		}
	}

	private double scoreIntern( QualifiedValueList correctValue, AbstractObject fusedValue) {

		List<AbstractObject> correctValues = correctValue.getClaims().stream().map((e) -> {
			return e.getValue();
		}).collect(Collectors.toList());

		if (correctValues.isEmpty()) {
			return 0.0;
		} else {
			Double maxSim = null;
			for (AbstractObject cao : correctValues) {
				double sim = this.similarityCalculator.getSimilarity(cao, fusedValue );
				if (maxSim == null || sim > maxSim) {
					maxSim = sim;
				}
			}
			return maxSim;
		}

	}

	public Optional<Boolean> correctIntern(QualifiedTripleId tripleId, QualifiedValueList correctValue, FusedValue fusedValue) {

		Set<AbstractObject> correctValues = timestampQualifiersAdapter.getForQualifiers(correctValue, tripleId.getQualifiers());

		if (correctValues.isEmpty()) {
			return Optional.empty();
		} else {
			int fdfdf = 123;
			for (AbstractObject cao : correctValues) {
				if (this.similarityCalculator.isEqual(cao, fusedValue.getSingle())) {
					return Optional.of(Boolean.TRUE);
				}
			}
			return Optional.of(Boolean.FALSE);
		}
	}

	private Optional<Boolean> correctIntern(TripleId tripleId, QualifiedValueList correctValue, FusedValue fusedValue) {

		List<AbstractObject> correctValues = correctValue.getClaims().stream().map((e) -> {
			return e.getValue();
		}).collect(Collectors.toList());

		for (AbstractObject cao : correctValues) {
			if (this.similarityCalculator.isEqual(cao, fusedValue.getSingle())) {
				return Optional.of(Boolean.TRUE);
			}
		}
		return Optional.of(Boolean.FALSE);

	}

	// @SuppressWarnings("unchecked")
	// public boolean correctIntern(TripleId tripleId, QualifiedValueList
	// correctValue, FusedValue fusedValue, ValueType cdt, FilteredValues
	// filteredValues) {
	//
	// if (correctValue == null)
	// return false;
	// else if (cdt == null) {
	// System.out.println("B: CDT == null, please check!");
	// return false;
	// } else if
	// (ontologyRepository.isListType(tripleId.getClassPropertyId().getPropertyId()))
	// {
	// if (filteredValues != null) {
	// if (!(fusedValue.isList()))
	// throw new IllegalStateException("No List provided even though we are in
	// the ListExpansion use case");
	// else {
	// Set<AbstractObject> lookingFor =
	// filteredValues.getFilteredValues(tripleId);
	// if (lookingFor != null && lookingFor.size() > 0) {
	// List<AbstractObject> correctList = new ArrayList<>(lookingFor);
	//
	// // Set<Object> valueToEliminate = new
	// // HashSet<>((List<Object>) correctValue);
	// // valueToEliminate.removeAll(correctList);
	// // for (Iterator<Object> it = fusedList.iterator();
	// // it.hasNext();) {
	// // if (valueToEliminate.contains(it.next())) {
	// // it.remove();
	// // }
	// //
	// // }
	//
	// double sim = overlapSimilarity.getMAP(correctList,
	// fusedValue.getRankedList());
	//
	// if (sim > 0.9)
	// return true;
	// else
	// return false;
	//
	// }
	//
	// else {
	// return false;
	// }
	// }
	// } else {
	// if (correctValue instanceof List<?> && fusedValue.isList()) {
	//
	// List<AbstractObject> correctList = (List<AbstractObject>) correctValue;
	//
	// double sim = overlapSimilarity.getMAP(correctList,
	// fusedValue.getRankedList());
	// if (sim > 0.9) {
	// return true;
	// } else {
	// return false;
	// }
	//
	// } else {
	// List<AbstractObject> correctList;
	//
	// correctList = (List<AbstractObject>) correctValue;
	//
	// List<AbstractObject> fusedValueList;
	// if (!(fusedValue.isList())) {
	// fusedValueList = new ArrayList<>();
	// fusedValueList.add(fusedValue.getSingle());
	// } else {
	// fusedValueList = fusedValue.getSimpleList();
	// }
	//
	// int overlap = overlapSimilarity.getOverlap(fusedValueList, correctList);
	//
	// if (overlap == 1) {
	// return true;
	// } else {
	// return false;
	// }
	// }
	// }
	//
	// } else {
	//
	// return this.qualifiedSimilarityCalculator.isCorrect(correctValue,
	// fusedValue.getSingle());
	//
	// }
	//
	// }

	public Optional<Boolean> correct(ValueId valueId, QualifiedValueHolder fusedValue) {

		QualifiedValueList correctValue = dbPediaRepository.getQualifiedValueListWithEqualProperties(valueId);
		if (correctValue == null) {
			return Optional.empty();
		}
		return qualifiedSimilarityCalculator.isContainedAndCorrect(correctValue, fusedValue);

	}

	public Optional<Boolean> correct(ValueId valueId, FusedValue fusedValue) {
//		valueId = new ValueId(valueId.getInstanceId(), PropertyId.fromString("weightPounds"));
		QualifiedValueList correctValue = dbPediaRepository.getQualifiedValueListWithEqualProperties(valueId);
		if (correctValue == null) {
			return Optional.empty();
		}

		return Optional.of(qualifiedSimilarityCalculator.isCorrect(correctValue, fusedValue.getSingle()));

	}
	
	public Optional<Boolean> correctWithoutEquivalent(ValueId valueId, FusedValue fusedValue) {
//		valueId = new ValueId(valueId.getInstanceId(), PropertyId.fromString("weightPounds"));
		QualifiedValueList correctValue = dbPediaRepository.getQualifiedValueList(valueId);
		if (correctValue == null) {
			return Optional.empty();
		}

		return Optional.of(qualifiedSimilarityCalculator.isCorrect(correctValue, fusedValue.getSingle()));

	}

	// public boolean maxCorrect(TripleId tripleId, List<AbstractObject> values,
	// boolean fuseLists, FilteredValues filteredValues) {
	//
	// ClassPropertyId classPropertyId = tripleId.getClassPropertyId();
	// QualifiedValueList correctValue =
	// dbPediaRepository.getQualifiedValueList(tripleId.getValueId());
	// if (correctValue == null)
	// return false;
	// ValueType cdt = ontologyRepository.getDataType(tripleId.getPropertyId());
	// if (cdt == null) {
	// System.out.println("B: CDT == null, please check!");
	// return false;
	// }
	//
	// else if (fuseLists &&
	// ontologyRepository.isListType(classPropertyId.getPropertyId())) {
	//
	// Set<AbstractObject> groupedSet = new HashSet<>();
	// for (AbstractObject v : values) {
	// if (v instanceof List<?>) {
	// @SuppressWarnings("unchecked")
	// List<AbstractObject> valueList = (List<AbstractObject>) v;
	//
	// groupedSet.addAll(valueList);
	// } else {
	// groupedSet.add(v);
	// }
	//
	// }
	//
	// if (groupedSet.size() == 0)
	// return false;
	//
	// if (filteredValues != null) {
	// Set<AbstractObject> lookingFor =
	// filteredValues.getFilteredValues(tripleId);
	// if (lookingFor != null && lookingFor.size() > 0) {
	//
	// if (overlapSimilarity.hasOverlap(new ArrayList<>(lookingFor), new
	// ArrayList<>(groupedSet), lookingFor.size())) {
	// return true;
	// } else {
	// return false;
	// }
	//
	// } else {
	// return false;
	// }
	//
	// } else {
	//
	// List<AbstractObject> groupedList = new ArrayList<>(groupedSet);
	//
	// // List<AbstractObject> correctList = new ArrayList<>();
	// // correctList.add(correctValue);
	//
	// // int overlap = overlapSimilarity.getOverlap(correctList,
	// // groupedList, similarityCalculator.getMeasure(cdt));
	// //
	// // if (overlap == 1) {
	// // return true;
	// // } else {
	// // return false;
	// // }
	// return qualifiedSimilarityCalculator.isCorrect(correctValue,
	// groupedList);
	//
	// }
	//
	// } else {
	// return this.qualifiedSimilarityCalculator.isCorrect(correctValue,
	// values);
	// }
	// }

	public boolean maxCorrect(QualifiedTripleId tid, List<AbstractObject> values, boolean fuseLists, FilteredValues filteredValues) {
		QualifiedValueList correctValue = dbPediaRepository.getQualifiedValueListWithEqualProperties(tid.getTripleId().getValueId());

		if (correctValue != null) {
			Set<AbstractObject> correctValues = this.timestampQualifiersAdapter.getForQualifiers(correctValue, tid.getQualifiers());

			for (AbstractObject cao : correctValues) {
				for (AbstractObject ao : values) {
					if (this.similarityCalculator.isEqual(cao, ao)) {
						return true;
					}
				}
			}
		}
		return false;
	}

	public boolean maxCorrect(TripleId tid, List<AbstractObject> values, boolean fuseLists, FilteredValues filteredValues) {
		QualifiedValueList correctValue = dbPediaRepository.getQualifiedValueListWithEqualProperties(tid.getValueId());

		if (correctValue != null) {
			List<AbstractObject> correctValues = correctValue.getClaims().stream().map((e) -> {
				return e.getValue();
			}).collect(Collectors.toList());

			for (AbstractObject cao : correctValues) {
				for (AbstractObject ao : values) {
					if (this.similarityCalculator.isEqual(cao, ao)) {
						return true;
					}
				}
			}
		}
		return false;
	}

}
