package de.uni_mannheim.informatik.dws.expansion.evaluation.timestamps;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.qualifiers.TimestampPointInTime;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats.Stats;

public class TimestampQualifiersAdapter {

	@Inject
	public TimestampQualifiersAdapter() {
	}

	public Integer getFromQualifiers(Set<AbstractQualifier> qualifiers) {
		Integer year = null;
		for (AbstractQualifier aq : qualifiers) {
			if (aq instanceof TimestampPointInTime) {
				year = ((TimestampPointInTime) aq).getYear();
				break;
			}
		}
		return year;
	}

	public Set<AbstractObject> getForQualifiers(QualifiedValueList a, Set<AbstractQualifier> qualifiers) {

		return new HashSet<>(getForQualifiersMap(a, qualifiers).values());
	}

	public Map<Integer, AbstractObject> getForQualifiersMap(QualifiedValueList a, Set<AbstractQualifier> qualifiers) {
		HashMap<Integer, AbstractObject> returnSet = new HashMap<>();

		Integer year = null;
		for (AbstractQualifier q : qualifiers) {
			if (q instanceof TimestampPointInTime) {
				TimestampPointInTime tpit = (TimestampPointInTime) q;
				year = tpit.getYear();
				break;
			}

		}

		if (year == null) {
			int index = 0;
			for (QualifiedValueHolder claim : a.getClaims()) {
				returnSet.put(index, claim.getValue());
				index++;
			}

		} else {
			int index = 0;
			for (QualifiedValueHolder claim : a.getClaims()) {
				Set<PropertyId> qualifierIds = claim.getQualifierIds();

				if (qualifierIds.contains(Stats.pointInTime)) {
					AbstractObject o = claim.getQualifierValue(Stats.pointInTime).getQualifierValue();
					Integer claimYear = TimeObject.getYear(o);
					if (claimYear != null && claimYear.equals(year)) {
						returnSet.put(index, claim.getValue());
					}
				} else if (qualifierIds.contains(Stats.startTimeQualifier)) {
					AbstractObject o = claim.getQualifierValue(Stats.startTimeQualifier).getQualifierValue();
					Integer claimYear = TimeObject.getYear(o);
					if (claimYear != null && year >= claimYear) {

						if (qualifierIds.contains(Stats.endTimeQualifier)) {
							AbstractObject endO = claim.getQualifierValue(Stats.endTimeQualifier).getQualifierValue();
							Integer endYear = TimeObject.getYear(endO);
							if (endYear != null && year <= endYear) {
								returnSet.put(index, claim.getValue());
							}
						} else {
							returnSet.put(index, claim.getValue());
						}

					}

				}

				index++;
			}
		}

		return returnSet;
	}

	public Set<TimestampPointInTime> getQualifierFromQVH(QualifiedValueHolder claim) {
		Set<TimestampPointInTime> timeStampPoints = new HashSet<>();

		Set<PropertyId> qualifierIds = claim.getQualifierIds();

		if (qualifierIds.contains(Stats.pointInTime)) {
			AbstractObject o = claim.getQualifierValue(Stats.pointInTime).getQualifierValue();
			Integer claimYear = TimeObject.getYear(o);
			if (claimYear != null) {
				timeStampPoints.add(new TimestampPointInTime(claimYear));
				// CHMHelper.getFromMap(ret, new
				// TimestampPointInTime(claimYear), () -> new
				// ArrayList<Integer>()).add(i);

			}

		} else if (qualifierIds.contains(Stats.startTimeQualifier)) {
			AbstractObject o = claim.getQualifierValue(Stats.startTimeQualifier).getQualifierValue();
			Integer claimYear = TimeObject.getYear(o);
			if (claimYear != null) {
				if (qualifierIds.contains(Stats.endTimeQualifier)) {
					AbstractObject endO = claim.getQualifierValue(Stats.endTimeQualifier).getQualifierValue();
					Integer endYear = TimeObject.getYear(endO);
					if (endYear != null) {
						for (int cYear = claimYear; cYear <= endYear; cYear++) {
							timeStampPoints.add(new TimestampPointInTime(cYear));
							// CHMHelper.getFromMap(ret, new
							// TimestampPointInTime(cYear), () -> new
							// ArrayList<Integer>()).add(i);
						}

					}
				} else {
					for (int cYear = claimYear; cYear <= 2016; cYear++) {
						timeStampPoints.add(new TimestampPointInTime(cYear));
						// CHMHelper.getFromMap(ret, new
						// TimestampPointInTime(cYear), () -> new
						// ArrayList<Integer>()).add(i);
					}
				}
			}

		}

		return timeStampPoints;
	}

	public Map<TimestampPointInTime, List<Integer>> getQualifierMapFromQVL(QualifiedValueList a) {
		Map<TimestampPointInTime, List<Integer>> ret = new HashMap<>();
		List<QualifiedValueHolder> claims = a.getClaims();
		for (int i = 0; i < claims.size(); i++) {

			QualifiedValueHolder claim = claims.get(i);

			Set<TimestampPointInTime> timestamps = getQualifierFromQVH(claim);

			for (TimestampPointInTime tpit : timestamps) {
				CHMHelper.getFromMap(ret, tpit, () -> new ArrayList<Integer>()).add(i);
			}

		}

		return ret;
	}

	public List<Integer> getNonQualifierMapFromQVL(QualifiedValueList valueList) {
		List<Integer> ret = new ArrayList<>();
		List<QualifiedValueHolder> claims = valueList.getClaims();
		for (int i = 0; i < claims.size(); i++) {
			QualifiedValueHolder claim = claims.get(i);
			Set<PropertyId> qualifierIds = claim.getQualifierIds();
			if (!qualifierIds.contains(Stats.pointInTime) && !qualifierIds.contains(Stats.startTimeQualifier)) {
				ret.add(i);
			}
		}
		return ret;
	}
}
