package de.mannheim.uni.expansion.tablefusion.core.fusing;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.inject.Inject;

import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityMeasure;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;

public class EquivalenceClusterer {

	@Inject
	public EquivalenceClusterer() {

	}

	public void addValue(List<Cluster> clusters, SimilarityMeasure measure, BaseFuser baseFuser, AbstractObject value) {

		boolean added = false;
		for (int i = 0; i < clusters.size(); i++) {
			if (measure.isEqual(clusters.get(i).getValue(), value)) {
				List<AbstractObject> valueList = new ArrayList<>(clusters.get(i).getValues());
				valueList.add(value);

				clusters.set(i, new Cluster(baseFuser.simpleFuse(valueList).getSingle(), valueList));

				added = true;
				break;
			}
		}
		if (!added) {
			List<AbstractObject> valueList = new ArrayList<>();
			valueList.add(value);

			clusters.add(new Cluster(value, valueList));
		}

	}

	public Cluster getHighestRatedCluster(List<Cluster> clusters) {
		clusters = new ArrayList<>(clusters);
		Collections.sort(clusters);
		return clusters.get(clusters.size() - 1);
	}

	public class Cluster implements Comparable<Cluster> {
		private final AbstractObject fusedValue;
		private final List<AbstractObject> values;
		private final int count;

		public Cluster(AbstractObject fusedValue, List<AbstractObject> values) {
			super();
			this.fusedValue = fusedValue;
			this.values = Collections.unmodifiableList(values);
			this.count = values.size();
		}

		public List<AbstractObject> getValues() {
			return values;
		}

		@Override
		public int compareTo(Cluster o) {
			return Integer.compare(count, o.getCount());
		}

		public AbstractObject getValue() {
			return fusedValue;
		}

		public int getCount() {
			return count;
		}

	}
}
