package de.mannheim.uni.expansion.tablefusion.core.fusing;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.OptionalDouble;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityMeasure;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;

public class ScoredEquivalenceClusterer {

	@Inject
	public ScoredEquivalenceClusterer() {

	}

	public List<Cluster> addValue(List<Cluster> clusters, SimilarityMeasure measure, BaseFuser baseFuser, AbstractObject value, double score) {
		clusters = new ArrayList<>(clusters);
		boolean added = false;
		for (int i = 0; i < clusters.size(); i++) {
			if (measure.isEqual(clusters.get(i).getValue(), value)) {
				List<Pair<AbstractObject, Double>> valueList = new ArrayList<>(clusters.get(i).getValues());
				valueList.add(new Pair<>(value, score));

				clusters.set(i, new Cluster(baseFuser.simpleFuse(valueList.stream().map(Pair::getFirst).collect(Collectors.toList())).getSingle(), valueList));

				added = true;
				break;
			}
		}
		if (!added) {
			List<Pair<AbstractObject, Double>> valueList = new ArrayList<>();
			valueList.add(new Pair<>(value, score));

			clusters.add(new Cluster(value, valueList));
		}

		return clusters;
	}

	public Cluster getHighestRatedCluster(List<Cluster> clusters) {
		clusters = new ArrayList<>(clusters);
		Collections.sort(clusters);
		return clusters.get(clusters.size() - 1);
	}

	public class Cluster implements Comparable<Cluster> {
		private final AbstractObject fusedValue;
		private final List<Pair<AbstractObject, Double>> values;

		public Cluster(AbstractObject fusedValue, List<Pair<AbstractObject, Double>> values) {
			super();
			this.fusedValue = fusedValue;
			this.values = Collections.unmodifiableList(values);
		}

		public List<Pair<AbstractObject, Double>> getValues() {
			return values;
		}

		@Override
		public int compareTo(Cluster o) {
			return Double.compare(getScore(), o.getScore());
		}

		public AbstractObject getValue() {
			return fusedValue;
		}

		public double getScore() {
			OptionalDouble od = this.values.stream().mapToDouble(Pair::getSecond).average();
			if (od.isPresent()) {
				return od.getAsDouble();
			} else {
				return 0.0;
			}

		}

	}
}
