package de.mannheim.uni.expansion.tablefusion.core.fusing.impl.base;

import java.util.Collections;

import javax.inject.Inject;

import de.mannheim.uni.expansion.tablefusion.core.fusing.impl.generic.GenericMajorityMedianFuser;
import de.mannheim.uni.expansion.tablefusion.core.fusing.impl.generic.GenericMedianFuser;
import de.mannheim.uni.expansion.tablefusion.core.fusing.impl.generic.GenericMedianFuser.Comparer;
import de.mannheim.uni.expansion.tablefusion.core.fusing.impl.generic.GenericMedianFuser.ListSorter;
import de.mannheim.uni.expansion.tablefusion.core.fusing.impl.generic.GenericMedianFuser.Medianizer;
import de.mannheim.uni.expansion.tablefusion.core.fusing.impl.generic.GenericMedianFuser.TypeCaster;
import de.mannheim.uni.expansion.tablefusion.core.fusing.impl.generic.GenericMedianFuser.TypeChecker;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityMeasure;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.CoordinateObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class BaseCoordinateFuser extends GenericMajorityMedianFuser<CoordinateObject> {
	private final static TypeChecker<CoordinateObject> typeChecker = (o) -> {

		if (o instanceof CoordinateObject) {
			return true;
		}

		return false;
	};

	private final static TypeCaster<CoordinateObject> typeCaster = (o) -> {

		if (o instanceof CoordinateObject) {
			return (CoordinateObject) o;
		} else {
			return null;
		}

	};

	private final static ListSorter<CoordinateObject> listSorter = (list) -> {
		Collections.sort(list);
	};

	private final static Comparer<CoordinateObject> comparer = (object, other) -> {
		return object.compareTo(other);
	};

	private final static Medianizer<CoordinateObject> medianizer = (sortedSet, threshold) -> {
		return GenericMedianFuser.defaultMedianizer(sortedSet, threshold, CoordinateObject::averager);
	};

	@Inject
	public BaseCoordinateFuser(SimilarityCalculator sc) {
		this(sc.getMeasure(ValueType.COORDINATE));

	}

	public BaseCoordinateFuser(SimilarityMeasure measure) {
		super(new GenericMedianFuser<CoordinateObject>(typeChecker, typeCaster, listSorter, CoordinateObject::averager, comparer, medianizer), measure);

	}

}
