package de.mannheim.uni.expansion.tablefusion.core.fusing.impl.base;

import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.tablefusion.core.fusing.impl.generic.GenericMajorityMedianFuser;
import de.mannheim.uni.expansion.tablefusion.core.fusing.impl.generic.GenericMedianFuser;
import de.mannheim.uni.expansion.tablefusion.core.fusing.impl.generic.GenericMedianFuser.TypeCaster;
import de.mannheim.uni.expansion.tablefusion.core.fusing.impl.generic.GenericMedianFuser.TypeChecker;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject.Precision;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.dws.expansion.data.ontology.model.valuepairs.ValueSortedGenericPair;

public class BaseDateFuser extends GenericMajorityMedianFuser<TimeObject> {

	public final static TypeChecker<TimeObject> typeChecker = (o) -> {
		return (o instanceof TimeObject);
	};

	public final static TypeCaster<TimeObject> typeCaster = (o) -> {
		if (o instanceof TimeObject)
			return (TimeObject) o;
		else
			return null;
	};

	private static FusedValue medianizeDirect(List<ValueSortedGenericPair<TimeObject>> values) {
		return GenericMedianFuser.defaultMedianizer(values, null, TimeObject::averager);
	}

	public static FusedValue medianize(List<ValueSortedGenericPair<TimeObject>> values, Double threshold) {

		if (values.size() > 0) {
			if (threshold == null || values.stream().mapToDouble(ValueSortedGenericPair::getScore).sum() > threshold) {
				Precision highestPrecision = TimeObject.getHighestPrecision(values.stream().map(ValueSortedGenericPair::getValue).collect(Collectors.toList()));

				return BaseDateFuser.medianizeForPrecision(values, threshold, highestPrecision);
			}

		}

		return null;
		// double sum =
		// values.stream().mapToDouble(ValueSortedGenericPair::getScore).sum();
		//
		// double averageScore = sum / values.size();
		//
		//
		//
		// double median = sum / 2.0;
		//
		// double current = 0.0;
		//
		// TimeObject value = null;
		//
		// getMedianValueForHighestPrecision() {
		//
		// }
		//
		// ValueSortedGenericPair<TimeObject> lastPair = null;
		// for (ValueSortedGenericPair<TimeObject> p : values) {
		// current += Math.abs(p.getScore());
		//
		// if (median > current) {
		// continue;
		// } else {
		// if (lastPair == null) {
		// value = p.getValue();
		// break;
		// } else {
		// // double diff = p.getScore();
		// double s1 = (current) - median;
		// double s2 = median - (current - Math.abs(p.getScore()));
		//
		// value = averager.get(p.getValue(), s1, lastPair.getValue(), s2);
		// }
		// }
		//
		// lastPair = p;
		// }
		//
		// if (value == null)
		// return null;
		// return new FusedValue(value, sum);

	}

	// private static Supplier<TreeSet<ValueSortedGenericPair<TimeObject>>>
	// supplier = () -> new TreeSet<>();;

	private static FusedValue medianizeForPrecision(List<ValueSortedGenericPair<TimeObject>> values, Double threshold, Precision precision) {

		List<ValueSortedGenericPair<TimeObject>> filteredList = values.stream().filter((e) -> {
			return e.getValue().getPrecision().getPosition() >= precision.getPosition();
		}).collect(Collectors.toList());

		double sum = filteredList.stream().mapToDouble(ValueSortedGenericPair::getScore).sum();

		if ((threshold != null && sum > threshold) || sum >= values.stream().mapToDouble(ValueSortedGenericPair::getScore).average().getAsDouble()) {
			return medianizeDirect(filteredList);
		} else if (precision.getPosition() > values.stream().mapToInt((e) -> {
			return e.getValue().getPrecision().getPosition();
		}).min().getAsInt()) {
			return medianizeForPrecision(values, threshold, Precision.getPrecision(precision.getPosition() - 1));
		} else {
			return null;
		}

	}

	@Inject
	public BaseDateFuser(SimilarityCalculator sc) {
		super(new GenericMedianFuser<>(typeChecker, typeCaster, Collections::sort, TimeObject::averager, TimeObject::compare, BaseDateFuser::medianize),
				sc.getMeasure(ValueType.TIME));
	}
}
