package de.mannheim.uni.expansion.tablefusion.core.fusing.impl.base;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.tablefusion.core.fusing.BaseFuser;
import de.mannheim.uni.expansion.tablefusion.core.fusing.FusionScoreThresholdAdapter;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valuepairs.AbstractClassPair;

public class BaseMajorityFuser implements BaseFuser {

	@Inject
	public BaseMajorityFuser() {

	}

	@Override
	public FusedValue simpleFuse(List<AbstractObject> values) {
		Map<AbstractObject, Integer> countMap = new HashMap<AbstractObject, Integer>();

		for (AbstractObject o : values) {

			Integer count = countMap.get(o);
			if (count == null) {
				count = 0;
			}
			count++;
			countMap.put(o, count);

		}

		Comparator<Entry<AbstractObject, Integer>> comparator = (e1, e2) -> {
			if (e1.getValue() < e2.getValue())
				return -1;
			else if (e1.getValue() > e2.getValue()) {
				return 1;
			} else {
				return 0;
			}
		};

		// Find the max count
		Optional<Entry<AbstractObject, Integer>> o = countMap.entrySet().stream().max(comparator);

		if (o.isPresent())
			return new FusedValue(o.get().getKey(), o.get().getValue().doubleValue(), o.get().getValue().intValue());
		else
			return null;
	}

	@Override
	public FusedValue fuse(List<? extends AbstractClassPair> values, Double threshold, FusionScoreThresholdAdapter adapter) {
		if (values.size() == 0) {
			return null;
		}
		Map<AbstractObject, List<Double>> countMap = new HashMap<AbstractObject, List<Double>>();

		for (AbstractClassPair pair : values) {
			Double probability = pair.getScore();
			if (probability > threshold) {
				AbstractObject o = pair.getValue();
				List<Double> count = countMap.get(o);
				if (count == null) {
					count = new ArrayList<Double>();
					countMap.put(o, count);
				}
				count.add(probability);
			}

		}

		Function<List<Double>, Double> sumListOfDoubles = (listOfDouble) -> {
			return listOfDouble.stream().mapToDouble(Double::doubleValue).sum();
		};

		Map<AbstractObject, Double> summedMap = countMap.entrySet().stream().collect(Collectors.toMap(Entry::getKey, e -> {
			double sum = sumListOfDoubles.apply(e.getValue());
			int count = e.getValue().size();

			if (adapter != null && threshold != null) {
				return adapter.adapt(sum, count, threshold);
			} else {
				return sum;
			}

		}));

		Comparator<Entry<AbstractObject, Double>> comparator = (e1, e2) -> {
			if (e1.getValue() < e2.getValue())
				return -1;
			else if (e1.getValue() > e2.getValue()) {
				return 1;
			} else {
				return 0;
			}
		};

		Optional<Entry<AbstractObject, Double>> o = summedMap.entrySet().stream().max(comparator);

		if (o.isPresent()) {
			// if (threshold == null || o.get().getValue() > threshold)
			return new FusedValue(o.get().getKey(), o.get().getValue(), countMap.get(o.get().getKey()).size());
			// else
			// return null;
		} else
			return null;
	}

	@Override
	public boolean checkType(AbstractObject value) {
		return true;
	}

	@Override
	public FusedValue clusteredFuse(List<AbstractObject> values) {
		return simpleFuse(values);
	}
}
