package de.mannheim.uni.expansion.tablefusion.core.fusing.impl.generic;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.tablefusion.core.fusing.BaseFuser;
import de.mannheim.uni.expansion.tablefusion.core.fusing.FusionScoreThresholdAdapter;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityMeasure;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valuepairs.AbstractClassPair;
import de.uni_mannheim.dws.expansion.data.ontology.model.valuepairs.ScoreSortedAbstractPair;

public class GenericMajorityMedianFuser<T extends AbstractObject> implements BaseFuser {
	private final BaseFuser medianFuser;

	private final SimilarityMeasure measure;

	public GenericMajorityMedianFuser(BaseFuser medianFuser, SimilarityMeasure measure) {
		super();
		this.medianFuser = medianFuser;
		this.measure = measure;
	}

	@Override
	public FusedValue simpleFuse(List<AbstractObject> values) {
		List<AbstractClassPair> vpList = values.stream().map((o) -> new ScoreSortedAbstractPair(o, 1.0)).collect(Collectors.toList());

		return fuse(vpList, 0.0, null);

	}

	@Override
	public FusedValue fuse(List<? extends AbstractClassPair> values, Double threshold, FusionScoreThresholdAdapter adapter) {
		ConcurrentLinkedQueue<Pair<AbstractObject, ConcurrentLinkedQueue<AbstractClassPair>>> clusters = new ConcurrentLinkedQueue<>();
		// List<Cluster> countMap = new ArrayList<Cluster>();

		PersistentParallel6.runAndFinish(values, (vp) -> {
			boolean added = false;
			if (vp.getScore() > threshold) {
				for (Pair<AbstractObject, ConcurrentLinkedQueue<AbstractClassPair>> cluster : clusters) {
					if (measure.isEqual(cluster.getFirst(), vp.getValue())) {
						cluster.getSecond().add(vp);

						// List<AbstractClassPair> valueProbabilityList = new
						// ArrayList<>(countMap.get(i).getValues());
						// valueProbabilityList.add(new
						// ScoreSortedAbstractPair(vp.getValue(),
						// vp.getScore()));
						//
						// FusedValue fv =
						// FusedValue.fromObject(countMap.get(i).getValue());
						// if (fv == null) {
						// int fsdf = 123;
						// }
						//
						// countMap.set(i, new Cluster(fv.getSingle(),
						// valueProbabilityList));

						added = true;
						break;
					}
				}
				if (!added) {
					ConcurrentLinkedQueue<AbstractClassPair> list = new ConcurrentLinkedQueue<>();
					list.add(vp);
					clusters.add(new Pair<>(vp.getValue(), list));
					// List<AbstractClassPair> valueProbabilityList = new
					// ArrayList<>();
					// valueProbabilityList.add(new
					// ScoreSortedAbstractPair(vp.getValue(), vp.getScore()));
					//
					// countMap.add(new Cluster(vp.getValue(),
					// valueProbabilityList));
				}
			}

		}, this.getClass(), values.size() < 1000000);

		List<Pair<AbstractObject, List<AbstractClassPair>>> finalClusters = new ArrayList<>();
		{
			Pair<AbstractObject, ConcurrentLinkedQueue<AbstractClassPair>> cluster = null;
			while ((cluster = clusters.poll()) != null) {
				boolean added = false;
				for (Pair<AbstractObject, List<AbstractClassPair>> newCluster : finalClusters) {
					if (measure.isEqual(cluster.getFirst(), cluster.getFirst())) {
						newCluster.getSecond().addAll(cluster.getSecond());

						added = true;
						break;
					}
				}
				if (!added) {
					List<AbstractClassPair> list = new ArrayList<>();
					list.addAll(cluster.getSecond());
					finalClusters.add(new Pair<>(cluster.getFirst(), list));
					// List<AbstractClassPair> valueProbabilityList = new
					// ArrayList<>();
					// valueProbabilityList.add(new
					// ScoreSortedAbstractPair(vp.getValue(), vp.getScore()));
					//
					// countMap.add(new Cluster(vp.getValue(),
					// valueProbabilityList));
				}
			}
		}

		if (finalClusters.size() > 0) {
			double maxScore = -0.1;
			int index = -1;
			for (int i = 0; i < finalClusters.size(); i++) {
				Pair<AbstractObject, List<AbstractClassPair>> cluster = finalClusters.get(i);
				double sum = cluster.getSecond().stream().mapToDouble(AbstractClassPair::getScore).sum();

				int count = cluster.getSecond().size();

				double finalSum;
				if (adapter != null && threshold != null) {
					finalSum = adapter.adapt(sum, count, threshold);
				} else {
					finalSum = sum;
				}

				if (finalSum > maxScore) {
					maxScore = finalSum;
					index = i;
				}
			}
			if (index > -1) {
				FusedValue fv = medianFuser.fuse(finalClusters.get(index).getSecond(), null, null);
				return fv;

				// TripleId tid = new TripleId(new ClassId("Country"),
				// new PropertyId("P2132", "NominalGDPPerCapita"),
				//
				// new InstanceId("Q800", "Costa_Rica"));
			} else {
				return null;
			}

		} else {
			return null;
		}
	}

	@Override
	public boolean checkType(AbstractObject value) {
		return medianFuser.checkType(value);
	}

	private class Cluster implements AbstractClassPair {
		private final AbstractObject fusedValue;
		private final List<AbstractClassPair> values;
		private final double score;

		public Cluster(AbstractObject fusedValue, List<AbstractClassPair> values) {
			super();
			this.fusedValue = fusedValue;
			this.values = Collections.unmodifiableList(values);
			this.score = values.stream().mapToDouble(AbstractClassPair::getScore).sum();
		}

		public List<AbstractClassPair> getValues() {
			return values;
		}

		@Override
		public int compareTo(AbstractClassPair o) {
			return Double.compare(score, o.getScore());
		}

		@Override
		public AbstractObject getValue() {
			return fusedValue;
		}

		@Override
		public Double getScore() {
			return score;
		}

	}

	@Override
	public FusedValue clusteredFuse(List<AbstractObject> values) {
		return medianFuser.simpleFuse(values);
	}

}
