package expansion.model;

import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.ThreadLocalRandom;

import de.mannheim.uni.expansion.tablefusion.core.fusing.impl.base.BaseCoordinateFuser;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.CoordinateSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.NumericSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThresholdImpl;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.CoordinateObject;

public class CoordinateObjectTest {

	public static void main(String[] args) throws IOException {

		String filePath = null;
		if (args.length > 0) {
			filePath = args[0];
		}

		NumericSimilarity ns = new NumericSimilarity(new SimilarityThresholdImpl(0.98));
		CoordinateSimilarity cs = new CoordinateSimilarity(new SimilarityThresholdImpl(0.98), ns);
		BaseCoordinateFuser fuser = new BaseCoordinateFuser(cs);

		List<CoordinateObject> objects = new ArrayList<>();
		ThreadLocalRandom random = ThreadLocalRandom.current();
		double cLatitude = random.nextDouble(-100.0, 100.0);
		double cLongitude = random.nextDouble(-100.0, 100.0);
		for (int i = 0; i < 50; i++) {
			double latitude = random.nextDouble(Math.max(-100.0, cLatitude - 2.0), Math.min(100.0, cLatitude + 2.0));
			double longitude = random.nextDouble(Math.max(-100.0, cLongitude - 2.0), Math.min(100.0, cLongitude + 2.0));

			objects.add(new CoordinateObject(latitude, longitude, 1.0));
		}

		double dLatitude = random.nextDouble(Math.max(-100.0, cLatitude - 30.0), Math.min(100.0, cLatitude + 30.0));
		double dLongitude = random.nextDouble(Math.max(-100.0, cLongitude - 30.0), Math.min(100.0, cLongitude + 30.0));

		for (int i = 0; i < 30; i++) {
			double latitude = random.nextDouble(Math.max(-100.0, dLatitude - 2.0), Math.min(100.0, dLatitude + 2.0));
			double longitude = random.nextDouble(Math.max(-100.0, dLongitude - 2.0), Math.min(100.0, dLongitude + 2.0));

			objects.add(new CoordinateObject(latitude, longitude, 1.0));
		}

		double eLatitude = random.nextDouble(Math.max(-100.0, cLatitude - 30.0), Math.min(100.0, cLatitude + 30.0));
		double eLongitude = random.nextDouble(Math.max(-100.0, cLongitude - 30.0), Math.min(100.0, cLongitude + 30.0));

		for (int i = 0; i < 30; i++) {
			double latitude = random.nextDouble(Math.max(-100.0, eLatitude - 2.0), Math.min(100.0, eLatitude + 2.0));
			double longitude = random.nextDouble(Math.max(-100.0, eLongitude - 2.0), Math.min(100.0, eLongitude + 2.0));

			objects.add(new CoordinateObject(latitude, longitude, 1.0));
		}

		for (int i = 0; i < 30; i++) {
			double latitude = random.nextDouble(-100.0, 100.0);
			double longitude = random.nextDouble(-100.0, 100.0);

			objects.add(new CoordinateObject(latitude, longitude, 1.0));
		}

		Collections.sort(objects);

		StringBuilder css = new StringBuilder();
		StringBuilder body = new StringBuilder();
		body.append("<div id=\"container\">");

		for (ListEntry<CoordinateObject> coordinate : ListIterable.iterable(objects)) {
			css.append(String.format("#o%s {left:%s;bottom:%s;position: absolute;}", String.valueOf(coordinate.getIndex()),
					String.valueOf(coordinate.getValue().getLatitude() * 5 + 500), String.valueOf(coordinate.getValue().getLongitude() * 5 + 500)));
			body.append(String.format("<div class=\"point\" id=\"o%s\"></div>", String.valueOf(coordinate.getIndex())));

		}

		FusedValue fv = fuser.simpleFuse(new ArrayList<AbstractObject>(objects));

		CoordinateObject co = (CoordinateObject) fv.getSingle();
		css.append(String.format("#fv {left:%s;bottom:%s;position: absolute;}", String.valueOf(co.getLatitude() * 5 + 500),
				String.valueOf(co.getLongitude() * 5 + 500)));
		body.append("<div class=\"point\" id=\"fv\"></div>");

		css.append(String.format("#ov {left:%s;bottom:%s;position: absolute;}", String.valueOf(cLatitude * 5 + 500), String.valueOf(cLongitude * 5 + 500)));
		body.append("<div class=\"point\" id=\"ov\"></div>");

		css.append(".point { border-radius: 5px; width: 10px; height: 10px; background-color: blue;}");
		css.append("#fv {   background-color: yellow;}");
		css.append("#ov {   background-color: green;}");
		css.append("#container { width: 1000px; height: 1000px; border: 1px solid black;}");

		body.append("</div>");
		StringBuilder html = new StringBuilder();
		html.append("<html>");
		html.append("<head>");
		html.append("<style>");
		html.append(css.toString());
		html.append("</style>");
		html.append("</head>");
		html.append("<body>");
		html.append(body.toString());
		html.append("</body>");
		html.append("</html>");

		System.out.println("Diff Latitude:" + (Math.abs(co.getLatitude() - cLatitude)));
		System.out.println("Diff Longitude:" + (Math.abs(co.getLongitude() - cLongitude)));

		if (filePath != null) {
			FileWriter fw = new FileWriter(filePath);
			fw.write(html.toString());
			fw.close();
		} else {
			System.out.println(html.toString());
		}

		try {
			PersistentParallel6.close();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
}
