package de.mannheim.uni.expansion.tablefusion.core.matching.impl.reference;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.dbpedia.kbtables.RepositoryCandidateMap;
import de.mannheim.uni.expansion.data.dbpedia.kbtables.RepositoryColSet;
import de.mannheim.uni.expansion.data.dbpedia.kbtables.RepositoryRowSet;
import de.mannheim.uni.expansion.data.dbpedia.kbtables.RepositoryTableList;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.matching.dbpedia.model.MatchingData;
import de.mannheim.uni.matching.dbpedia.model.TableRow;
import de.mannheim.uni.matching.dbpedia.model.adapters.TableToRowHierarchyAdapter;
import de.mannheim.uni.tableprocessor.model.Table;
import de.mannheim.uni.tableprocessor.model.TableColumn;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.ValueConverter;

public class MatchingDataBuilder {

	private final OntologyRepository ontologyRepository;

	private final KnowledgeBase knowledgeBase;
	private final ValueConverter valueConverter;

	@Inject
	public MatchingDataBuilder(KnowledgeBase knowledgeBase, OntologyRepository ontologyRepository, ValueConverter valueConverter) {
		this.knowledgeBase = knowledgeBase;
		this.ontologyRepository = ontologyRepository;
		this.valueConverter = valueConverter;

	}

	public MatchingData get(Table webTable) {

		// ObjectPropertyOntology opo = (ObjectPropertyOntology)
		// ontologyRepository.get(propertyId);

		MatchingData data = getMatchingData();

		TableToRowHierarchyAdapter rowAdapter = new TableToRowHierarchyAdapter();
		Collection<TableRow> webtableRowSet = rowAdapter.getParts(webTable);
		data.setWebtableRowSet(webtableRowSet);
		Collection<Table> webTables = new ArrayList<Table>();
		data.setWebTables(webTables);
		webTables.add(webTable);
		data.setWebtable(webTable);

		return data;
	}

	public MatchingData getRanged(Table webTable, ClassId classId, PropertyId propertyId) {

		// ObjectPropertyOntology opo = (ObjectPropertyOntology)
		// ontologyRepository.get(propertyId);

		Set<ClassId> range = ontologyRepository.getRange(new ClassPropertyId(classId, propertyId));
		MatchingData data;

		if (range != null && !range.isEmpty()) {
			// System.out.println(propertyId.toString() + "::" +
			// range.toString());
			Set<ClassId> allowedClasses = new HashSet<>();
			allowedClasses.addAll(range);
			for (ClassId setId : range) {
				allowedClasses.addAll(getAllSubClases(setId));
				// allowedClasses.addAll(getAllParentClases(setId));
			}

			data = getMatchingDataRanged(allowedClasses);

		} else {
			data = getMatchingData();
		}
		TableToRowHierarchyAdapter rowAdapter = new TableToRowHierarchyAdapter();
		Collection<TableRow> webtableRowSet = rowAdapter.getParts(webTable);
		data.setWebtableRowSet(webtableRowSet);
		Collection<Table> webTables = new ArrayList<Table>();
		data.setWebTables(webTables);
		webTables.add(webTable);
		data.setWebtable(webTable);

		return data;
	}

	private Set<ClassId> getAllSubClases(ClassId classId) {

		Set<ClassId> classes = new HashSet<ClassId>();
		classes.add(classId);

		Set<ClassId> subClasses = ontologyRepository.getSubClasses(classId);
		if (subClasses != null) {
			for (ClassId subClassId : subClasses) {
				classes.addAll(getAllSubClases(subClassId));
			}
		}

		return classes;
	}

	private Set<ClassId> getAllParentClases(ClassId classId) {

		Set<ClassId> classes = new HashSet<ClassId>();
		classes.add(classId);

		Set<ClassId> parentClasses = ontologyRepository.getParentClasses(classId);

		if (parentClasses != null) {
			for (ClassId subClassId : parentClasses) {
				classes.addAll(getAllParentClases(subClassId));
			}
		}

		return classes;
	}

	public MatchingData getMatchingData() {

		Map<String, List<TableRow>> candidateMap = new RepositoryCandidateMap(knowledgeBase, ontologyRepository, valueConverter);
		Collection<TableRow> dbpediaRowSet = new RepositoryRowSet(knowledgeBase, ontologyRepository, valueConverter);
		Collection<TableColumn> dbpediaColSet = new RepositoryColSet(knowledgeBase, ontologyRepository, valueConverter);
		Collection<Table> dbpediaTables = new RepositoryTableList(knowledgeBase, ontologyRepository, valueConverter);

		MatchingData d = new MatchingData();
		d.setCandidateMap(candidateMap);
		d.setDbpediaColSet(dbpediaColSet);
		d.setDbpediaPropUris(new HashSet<String>());
		d.setDbpediaRowSet(dbpediaRowSet);
		d.setDbpediaTables(dbpediaTables);
		return d;

	}

	public MatchingData getMatchingDataRanged(Set<ClassId> allowedClasses) {

		Set<ClassId> checkedAllowedClasses = new HashSet<ClassId>();

		for (Iterator<ClassId> it = allowedClasses.iterator(); it.hasNext();) {
			ClassId next = it.next();
			if (knowledgeBase.hasClass(next)) {
				checkedAllowedClasses.add(next);
			}
		}

		Map<String, List<TableRow>> candidateMap = new RepositoryCandidateMap(knowledgeBase, ontologyRepository, valueConverter, checkedAllowedClasses);
		Collection<TableRow> dbpediaRowSet = new RepositoryRowSet(knowledgeBase, ontologyRepository, valueConverter, checkedAllowedClasses);
		Collection<TableColumn> dbpediaColSet = new RepositoryColSet(knowledgeBase, ontologyRepository, valueConverter, checkedAllowedClasses);
		Collection<Table> dbpediaTables = new RepositoryTableList(knowledgeBase, ontologyRepository, valueConverter, checkedAllowedClasses);

		MatchingData d = new MatchingData();
		d.setCandidateMap(candidateMap);
		d.setDbpediaColSet(dbpediaColSet);
		d.setDbpediaPropUris(new HashSet<String>());
		d.setDbpediaRowSet(dbpediaRowSet);
		d.setDbpediaTables(dbpediaTables);
		return d;
	}

}
