package de.uni_mannheim.informatik.dws.attribute;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.PropertyIndexComposite;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.uni_mannheim.informatik.dws.attribute.model.AttributeMappings;
import de.uni_mannheim.informatik.dws.attribute.model.PropertyMapping;
import de.uni_mannheim.informatik.dws.attribute.model.TableAttributeMappings;

public class AttributeMatchingAdapter {

	private final AnnotatedTablesRepository tableRepository;

	@Inject
	public AttributeMatchingAdapter(AnnotatedTablesRepository tableRepository) {
		this.tableRepository = tableRepository;
	}

	public AttributeMappings getAttributeMappings(Set<AnnotatedTableId> tableIds) {
		Map<AnnotatedTableId, TableAttributeMappings> ret = new HashMap<>();
		Map<AnnotatedTableId, ClassId> classIds = new HashMap<>();
		for (AnnotatedTableId tableId : tableIds) {
			Map<ColumnIdx, PropertyMapping> innerMap = new HashMap<>();
			AnnotatedTableMetaData md = tableRepository.getMetaData(tableId);
			if (md != null) {
				for (Entry<Integer, PropertyId> property : md.getProperties().entrySet()) {
					innerMap.put(new ColumnIdx(property.getKey()),
							new PropertyMapping(property.getValue(), md.getPropertyScore(new PropertyIndexComposite(property.getKey(), property.getValue()))));
				}

				classIds.put(tableId, md.getDbPediaClass());

				ret.put(tableId, new TableAttributeMappings(innerMap));
			}
		}

		return new AttributeMappings(ret, classIds);
	}
}
