package de.uni_mannheim.informatik.dws.attribute.matcher;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.inject.Inject;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableMap.Builder;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.PropertyInstanceId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.InstanceIndexComposite;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.similarity.numericRange.NumericRangeRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ImmuMaker;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.tableprocessor.model.TableColumnBuilder;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.CategoricalIntObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.QuantityObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject.Precision;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.attribute.matcher.AttributeMatchingSingleRun.ParameteredMatcherScore;
import de.uni_mannheim.informatik.dws.attribute.model.AttributeMappings;
import de.uni_mannheim.informatik.dws.attribute.model.PropertyMapping;
import de.uni_mannheim.informatik.dws.attribute.model.TableAttributeMappings;
import de.uni_mannheim.informatik.dws.attribute.valueAdaption.CategoricalIntTableConverter;
import de.uni_mannheim.informatik.dws.attribute.valueAdaption.StringPropertyConverter;
import de.uni_mannheim.informatik.dws.attribute.valueAdaption.TableColumnTimeToQuantityConverter;
import de.uni_mannheim.informatik.dws.attribute.valueAdaption.WeightInPoundsTableConverter;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.random.WeightsRandomParallelOptimizer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepositoryImpl;
import de.uni_mannheim.informatik.dws.matching.LabelMatcher;
import de.uni_mannheim.informatik.dws.matching.LabelMatcher.MatchStrictness;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;

public class AttributeMatcher {

	private final KnowledgeBase knowledgeBase;
	private final OntologyRepository ontologyRepository;
	private final ValueEvaluator valueEvaluator;
	private final Map<ClassPropertyId, ClassPropertyId> propertyConverter;
	private final LabelMatcher labelMatcher;
	@Inject
	protected SimilarityCalculator sc;
	@Inject
	protected AttributeMatcherEquation equation;

	@Inject
	protected NumericRangeRepository numericRangeRepository;

	@Inject
	public AttributeMatcher(KnowledgeBase knowledgeBase, OntologyRepository ontologyRepository, ValueEvaluator valueEvaluator, LabelMatcher labelMatcher) {
		this.knowledgeBase = knowledgeBase;
		this.ontologyRepository = ontologyRepository;
		this.valueEvaluator = valueEvaluator;
		Map<ClassPropertyId, ClassPropertyId> propertyConverter = new HashMap<>();
		for (Entry<ClassPropertyId, Set<ClassPropertyId>> entry : knowledgeBase.getEqualProperties().entrySet()) {
			for (ClassPropertyId inner : entry.getValue()) {
				propertyConverter.put(inner, entry.getKey());
			}
		}
		this.propertyConverter = ImmuMaker.map(propertyConverter);
		this.labelMatcher = labelMatcher;
	}

	// public AttributeMappings match(List<AnnotatedTablesReaderData> tables,
	// ThresholdContainer thresholdContainer,
	// Map<ClassId, Set<PropertyId>> nonTailProperties) {
	//
	// Set<AnnotatedTableId> tableIds = new HashSet<>();
	// tableIds.add(new
	// AnnotatedTableId("1346876860779_1346961089435_3900.arc5504794985066264630#30219675_1_433888948023861627"));
	// tableIds.add(new
	// AnnotatedTableId("1346876860609_1346965067128_728.arc4941513350675067041#53635965_0_4100477248612278954"));
	// tableIds.add(new
	// AnnotatedTableId("1346876860565_1346912219601_1704.arc8689279114261751938#3512341_5_5919837545333214767"));
	// tableIds.add(new
	// AnnotatedTableId("1346876860779_1346961276986_5601.arc3719795019286941883#45451800_0_7520172109909715831"));
	// tableIds.add(new
	// AnnotatedTableId("1346823846150_1346831707124_98.arc8289082661436581445#58505000_0_3971736499393838515"));
	// tableIds.add(new
	// AnnotatedTableId("1346876860791_1346881958401_2145.arc8468166189656932701#63206027_0_264701290058443984"));
	// tableIds.add(new
	// AnnotatedTableId("1346876860767_1346930413885_4219.arc6133765969811430556#8067855_0_7255939946700078827"));
	// tableIds.add(new
	// AnnotatedTableId("1346876860807_1346935022722_2014.arc8333932477257942082#17751131_0_8206891990416770602"));
	// tableIds.add(new
	// AnnotatedTableId("1346823846125_1346831228806_4601.arc19611991799741214#5774269_4_6373690942998186890"));
	// tableIds.add(new
	// AnnotatedTableId("1346876860777_1346911201139_5302.arc4950307784433668143#99781539_0_145784639298511497"));
	// tableIds.add(new
	// AnnotatedTableId("1346876860467_1346894614144_1467.arc8229090527093920145#26816383_0_5184733204356271199"));
	// tableIds.add(new
	// AnnotatedTableId("1346876860774_1346976279740_1167.arc5193714840123267708#23770978_0_8843872857851427183"));
	// tableIds.add(new
	// AnnotatedTableId("1346876860828_1346952444486_2789.arc4355746145008370050#37881601_0_5165264125271974655"));
	// tableIds.add(new
	// AnnotatedTableId("1346823846125_1346825813463_6.arc2362183109663985770#35523234_0_194649281147111494"));
	// tableIds.add(new
	// AnnotatedTableId("1350433107027_1350488164308_684.arc2045104594958977797#30633610_0_7633630560803713625"));
	//
	// Map<AnnotatedTableId, TableAttributeMappings> mappings = new
	// ConcurrentHashMap<>();
	// Map<AnnotatedTableId, ClassId> classIds = new ConcurrentHashMap<>();
	//
	// PersistentParallel6.runAndFinish(tables, (table) -> {
	//
	// int beginIndex = table.getMetaData().getNumHeaderRows();
	// int endIndex = table.getMetaData().getNumberOfRows() +
	// table.getMetaData().getNumHeaderRows();
	//
	// Map<ColumnIdx, PropertyId> propertyMap = new HashMap<>();
	//
	// for (ColumnIdx columnIdx : table.getTableData().getColumns()) {
	// if (!columnIdx.equals(table.getTableData().getKeyColumnIndex())) {
	//
	// ImmutableMap<RowIdx, AbstractObject> column =
	// table.getTableData().getColumn(columnIdx);
	// ImmutableMap<RowIdx, String> stringColumn =
	// table.getTableData().getStringTable().column(columnIdx);
	//
	// Optional<PropertyId> highestMatch =
	// getMatch(table.getMetaData().getDbPediaClass(), new
	// TableColumnId(table.getId(), columnIdx),
	// nonTailProperties.get(table.getMetaData().getDbPediaClass()),
	// table.getMetaData().getInstances(),
	// table.getMetaData().getInstancesScores(), column, stringColumn,
	// table.getTableData().getDataType(columnIdx), thresholdContainer,
	// beginIndex, endIndex);
	//
	// if (highestMatch.isPresent()) {
	// propertyMap.put(columnIdx, highestMatch.get());
	// }
	//
	// }
	// }
	//
	// if (tableIds.contains(table.getId())) {
	// int sdf = 123;
	// }
	// if
	// (table.getMetaData().getDbPediaClass().getFinalPart().equals("Single")) {
	// int sdf = 123;
	// }
	//
	// mappings.put(table.getId(), new TableAttributeMappings(propertyMap));
	// classIds.put(table.getId(), table.getMetaData().getDbPediaClass());
	// }, this.getClass());
	// return new AttributeMappings(mappings, classIds);
	// }

	public AttributeMappings match(List<AnnotatedTablesReaderData> tables, Map<ClassId, WeightsRepositoryImpl> weights,
			Map<ClassId, Set<PropertyId>> nonTailProperties, OptimizableClass optmizableClazz, OptimizableClass optmizableClazzSecondary,
			Set<AnnotatedTableId> tableIdsInput, boolean secondary, boolean baseline) {
		AttributeMappings firstMappings = match(tables, weights, nonTailProperties, optmizableClazz, tableIdsInput, null, baseline);

		if (secondary) {
			AttributeMappings secondaryMappings = match(tables, weights, nonTailProperties, optmizableClazzSecondary, tableIdsInput, firstMappings, baseline);

			return secondaryMappings;
		} else {
			return firstMappings;
		}
	}

	public AttributeMappings match(List<AnnotatedTablesReaderData> tables, Map<ClassId, WeightsRepositoryImpl> weights,
			Map<ClassId, Set<PropertyId>> nonTailProperties, OptimizableClass optmizableClazz, Set<AnnotatedTableId> tableIdsInput,
			AttributeMappings mappingsForSecondary, boolean baseline) {

		Map<ClassId, Map<String, Pair<Integer, Map<PropertyId, Set<TableColumnId>>>>> otherMatchLabelScores;
		Map<ClassId, Map<PropertyInstanceId, List<Pair<AbstractObject, TableColumnId>>>> otherTableMatches;
		if (mappingsForSecondary != null) {
			otherMatchLabelScores = getLabelScores(tables, mappingsForSecondary);
			otherTableMatches = getOtherTableMatch(tables, mappingsForSecondary);
		} else {
			otherMatchLabelScores = null;
			otherTableMatches = null;
		}

		Map<ClassId, Map<PropertyId, Set<String>>> labelForPropertyIdMap = new HashMap<>();
		for (Entry<ClassId, Set<PropertyId>> nonTailPropertiesByClass : nonTailProperties.entrySet()) {
			labelForPropertyIdMap.put(nonTailPropertiesByClass.getKey(),
					getLabelsForPropertyIdSet(nonTailPropertiesByClass.getKey(), nonTailPropertiesByClass.getValue()));
		}

		Set<AnnotatedTableId> tableIds = new HashSet<>();
		tableIds.add(new AnnotatedTableId("1346876860779_1346961089435_3900.arc5504794985066264630#30219675_1_433888948023861627"));
		tableIds.add(new AnnotatedTableId("1346876860609_1346965067128_728.arc4941513350675067041#53635965_0_4100477248612278954"));
		tableIds.add(new AnnotatedTableId("1346876860565_1346912219601_1704.arc8689279114261751938#3512341_5_5919837545333214767"));
		tableIds.add(new AnnotatedTableId("1346876860779_1346961276986_5601.arc3719795019286941883#45451800_0_7520172109909715831"));
		tableIds.add(new AnnotatedTableId("1346823846150_1346831707124_98.arc8289082661436581445#58505000_0_3971736499393838515"));
		tableIds.add(new AnnotatedTableId("1346876860791_1346881958401_2145.arc8468166189656932701#63206027_0_264701290058443984"));
		tableIds.add(new AnnotatedTableId("1346876860767_1346930413885_4219.arc6133765969811430556#8067855_0_7255939946700078827"));
		tableIds.add(new AnnotatedTableId("1346876860807_1346935022722_2014.arc8333932477257942082#17751131_0_8206891990416770602"));
		tableIds.add(new AnnotatedTableId("1346823846125_1346831228806_4601.arc19611991799741214#5774269_4_6373690942998186890"));
		tableIds.add(new AnnotatedTableId("1346876860777_1346911201139_5302.arc4950307784433668143#99781539_0_145784639298511497"));
		tableIds.add(new AnnotatedTableId("1346876860467_1346894614144_1467.arc8229090527093920145#26816383_0_5184733204356271199"));
		tableIds.add(new AnnotatedTableId("1346876860774_1346976279740_1167.arc5193714840123267708#23770978_0_8843872857851427183"));
		tableIds.add(new AnnotatedTableId("1346876860828_1346952444486_2789.arc4355746145008370050#37881601_0_5165264125271974655"));
		tableIds.add(new AnnotatedTableId("1346823846125_1346825813463_6.arc2362183109663985770#35523234_0_194649281147111494"));
		tableIds.add(new AnnotatedTableId("1350433107027_1350488164308_684.arc2045104594958977797#30633610_0_7633630560803713625"));
		tableIds.add(new AnnotatedTableId("1346823846039_1346871392535_1142.arc5700028916776158073#84758490_0_8043330651070690587"));
		tableIds.add(new AnnotatedTableId("1346876860877_1346943510155_2114.arc6735973191580564331#75677792_0_1815660811182748280"));

		Map<AnnotatedTableId, TableAttributeMappings> mappings = new ConcurrentHashMap<>();
		Map<AnnotatedTableId, ClassId> classIds = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(tables, (table) -> {

			int beginIndex = table.getMetaData().getNumHeaderRows();
			int endIndex = table.getMetaData().getNumberOfRows() + table.getMetaData().getNumHeaderRows();

			if (!tableIdsInput.contains(table.getId())) {
				return;
			}
			Map<ColumnIdx, PropertyMapping> propertyMap = new HashMap<>();

			for (ColumnIdx columnIdx : table.getTableData().getColumns()) {
				if (!columnIdx.equals(table.getTableData().getKeyColumnIndex())) {

					ImmutableMap<RowIdx, AbstractObject> column = table.getTableData().getColumn(columnIdx);
					ImmutableMap<RowIdx, String> stringColumn = table.getTableData().getStringTable().column(columnIdx);

					String columnHeader = null;
					AbstractObject columnHeaderSO = table.getTableData().getColumnHeader(columnIdx);
					if (columnHeaderSO != null && columnHeaderSO.getType() == ValueType.STRING) {
						columnHeader = ((StringObject) columnHeaderSO).getString();
					}

					ClassId classId = table.getMetaData().getDbPediaClass();
					Optional<PropertyMapping> highestMatch = getWeightedMatch(classId, new TableColumnId(table.getId(), columnIdx),
							nonTailProperties.get(table.getMetaData().getDbPediaClass()), table.getMetaData().getInstances(),
							table.getMetaData().getInstancesScores(), column, stringColumn, table.getTableData().getDataType(columnIdx), weights.get(classId),
							optmizableClazz, beginIndex, endIndex, columnHeader, labelForPropertyIdMap, otherMatchLabelScores, otherTableMatches, baseline);

					if (highestMatch.isPresent()) {
						propertyMap.put(columnIdx, highestMatch.get());
					}

				}
			}

			if (tableIds.contains(table.getId())) {
				int sdf = 123;
			}
			if (table.getMetaData().getDbPediaClass().getFinalPart().equals("Single")) {
				int sdf = 123;
			}

			if (propertyMap.containsValue(PropertyId.fromString("elevation"))) {
				int sdf = 123;
			}

			mappings.put(table.getId(), new TableAttributeMappings(propertyMap));
			classIds.put(table.getId(), table.getMetaData().getDbPediaClass());
		}, this.getClass());
		return new AttributeMappings(mappings, classIds);
	}

	// public Optional<PropertyId> getMatch(ClassId classId, TableColumnId
	// tableColumnId, Set<PropertyId> nonTailProperties, Map<Integer,
	// InstanceId> instances,
	// Map<InstanceIndexComposite, Double> instanceScores, ImmutableMap<RowIdx,
	// AbstractObject> column, ImmutableMap<RowIdx, String> stringColumn,
	// ValueType dataType, ThresholdContainer thresholdContainer, int
	// beginIndex, int endIndex) {
	// Map<PropertyId, Double> matches = getHighestMatches(classId,
	// tableColumnId, nonTailProperties, instances, instanceScores, column,
	// stringColumn,
	// dataType, beginIndex, endIndex);
	//
	// PropertyId highestMatch = null;
	// double score = 0.0;
	//
	// for (Entry<PropertyId, Double> entry : matches.entrySet()) {
	// if (entry.getValue() >= thresholdContainer.getScore(classId,
	// entry.getKey(), null)) {
	// if (highestMatch == null || entry.getValue() > score) {
	// highestMatch = entry.getKey();
	// score = entry.getValue();
	// }
	// }
	// }
	//
	// if (highestMatch != null) {
	// return Optional.of(highestMatch);
	// } else {
	// return Optional.empty();
	// }
	// }

	// public Optional<Pair<PropertyId, Double>> getMatchWithScore(ClassId
	// classId, TableColumnId tableColumnId, Set<PropertyId> nonTailProperties,
	// Map<Integer, InstanceId> instances, Map<InstanceIndexComposite, Double>
	// instanceScores, ImmutableMap<RowIdx, AbstractObject> column,
	// ImmutableMap<RowIdx, String> stringColumn, ValueType dataType, int
	// beginIndex, int endIndex) {
	// Map<PropertyId, Double> matches = getHighestMatches(classId,
	// tableColumnId, nonTailProperties, instances, instanceScores, column,
	// stringColumn,
	// dataType, beginIndex, endIndex);
	//
	// PropertyId highestMatch = null;
	// double score = 0.0;
	//
	// for (Entry<PropertyId, Double> entry : matches.entrySet()) {
	// if (highestMatch == null || entry.getValue() > score) {
	// highestMatch = entry.getKey();
	// score = entry.getValue();
	// }
	// }
	//
	// if (highestMatch != null) {
	// // ClassPropertyId replace = this.propertyConverter.get(new
	// // ClassPropertyId(classId, highestMatch));
	// // if (replace != null)
	// // return Optional.of(new Pair<>(replace.getPropertyId(), score));
	// // else
	// return Optional.of(new Pair<>(highestMatch, score));
	// } else {
	// return Optional.empty();
	// }
	// }

	private final ConcurrentHashMap<ClassPropertyId, Set<TimeObject>> allTimeObjectMap = new ConcurrentHashMap<>();
	private final ConcurrentHashMap<ClassPropertyId, Set<CategoricalIntObject>> allCIObjectMap = new ConcurrentHashMap<>();
	private final ConcurrentHashMap<ClassPropertyId, Set<StringObject>> allStringObjectMap = new ConcurrentHashMap<>();

	public Map<PropertyId, Double> getHighestMatchesOld(ClassId classId, Set<PropertyId> nonTailProperties, Map<Integer, InstanceId> instances,
			Map<InstanceIndexComposite, Double> instanceScores, ImmutableMap<RowIdx, AbstractObject> column, ValueType dataType) {
		// Map<PropertyId, MutableDouble> correctMap = new HashMap<>();
		//
		// Map<PropertyId, MutableDouble> totalMap = new HashMap<>();

		Map<PropertyId, Double> ret = new HashMap<>();

		Set<PropertyId> properties = nonTailProperties;
		for (PropertyId propertyId : properties) {
			if (!propertyId.equals(PropertiesExpansionData.LABEL_PROPERTY) && (ontologyRepository.getDataType(propertyId) == dataType
					|| (ontologyRepository.getDataType(propertyId) == ValueType.INSTANCE_REFERENCE && dataType == ValueType.STRING))) {

				double correctSum = 0.0;
				double totalSum = 0.0;

				int sdf = 123;
				for (Entry<Integer, InstanceId> instanceEntry : instances.entrySet()) {

					RowIdx rowIdx = new RowIdx(instanceEntry.getKey());
					AbstractObject ao = column.get(rowIdx);
					if (ao != null) {
						ValueId valueId = new ValueId(instanceEntry.getValue(), propertyId);
						Optional<Boolean> correct = valueEvaluator.correct(valueId, FusedValue.fromObject(ao));
						if (correct.isPresent()) {
							double instanceScore = instanceScores.get(new InstanceIndexComposite(instanceEntry.getKey(), instanceEntry.getValue()));
							totalSum += instanceScore;

							if (correct.get()) {
								correctSum += instanceScore;
							}

						}

					}

				}
				double firstScore = 0.0;
				if (correctSum > 0.00001) {
					firstScore = correctSum / totalSum;

					ValueType vt = ontologyRepository.getDataType(propertyId);
					ClassPropertyId cpid = new ClassPropertyId(classId, propertyId);

					if (vt == ValueType.TIME || (vt == ValueType.INSTANCE_REFERENCE && ontologyRepository.getRange(cpid) != null)) {
						int secondaryCorrect = 0;
						double secondaryTotal = column.size();

						Set<TimeObject> vtMap = new HashSet<>();
						if (vt == ValueType.TIME) {
							vtMap = CHMHelper.getFromCHM(this.allTimeObjectMap, cpid, () -> {
								Set<InstanceId> cpidInstances = knowledgeBase.getInstances(classId, propertyId);
								Set<TimeObject> set = new HashSet<>();

								for (InstanceId instanceId : cpidInstances) {
									List<AbstractObject> valueList = knowledgeBase.getValuesAsList(instanceId, propertyId);
									if (valueList != null) {
										for (AbstractObject innerAo : valueList) {
											if (innerAo.getType().equals(ValueType.TIME)) {
												set.add((TimeObject) innerAo);
											}
										}
									}

								}

								return set;
							});
						}
						for (AbstractObject ao : column.values()) {
							if (vt == ValueType.INSTANCE_REFERENCE) {
								if (ao.getType() == ValueType.STRING) {
									StringObject so = (StringObject) ao;
									Set<String> labels = new HashSet<>();
									labels.add(so.getString());
									if (labelMatcher.getMatches(labels, ontologyRepository.getRange(cpid), MatchStrictness.STRICT, false).isPresent()) {
										secondaryCorrect++;
									}
								}
							} else if (vt == ValueType.TIME) {
								if (ao.getType() == ValueType.TIME) {
									if (vtMap.contains((TimeObject) ao)) {
										secondaryCorrect++;
									}
								}
							}

						}

						double secondScore = secondaryCorrect / secondaryTotal;

						if (firstScore > 0.00001 | secondScore > 0.00001) {
							int sdfsdf = 123;
						}
						firstScore = Math.min(1.0, firstScore * (1.0 + secondScore));

					}
				}
				if (firstScore > 0.00001)
					ret.put(propertyId, firstScore);
			}
		}

		Map<PropertyId, Double> ret2 = new HashMap<>();
		for (Entry<PropertyId, Double> entry : ret.entrySet()) {
			ClassPropertyId replace = this.propertyConverter.get(new ClassPropertyId(classId, entry.getKey()));
			PropertyId propertyId = entry.getKey();
			if (replace != null) {
				propertyId = replace.getPropertyId();
			}

			if (!ret2.containsKey(propertyId) || entry.getValue() > ret2.get(propertyId)) {
				ret2.put(propertyId, entry.getValue());
			}
		}

		return ret2;
	}

	public Map<PropertyId, Map<String, ParameteredMatcherScore>> getFullMatches(ClassId classId, TableColumnId tableColumnId, Set<PropertyId> nonTailProperties,
			Map<Integer, InstanceId> instances, Map<InstanceIndexComposite, Double> instanceScores, final ImmutableMap<RowIdx, AbstractObject> rawColumn,
			final ImmutableMap<RowIdx, String> stringColumn, ValueType dataType, int beginIndex, int endIndex, String columnHeader,
			Map<ClassId, Map<PropertyId, Set<String>>> labelForPropertyIdMap,
			Map<ClassId, Map<String, Pair<Integer, Map<PropertyId, Set<TableColumnId>>>>> otherMatchLabelScores,
			Map<ClassId, Map<PropertyInstanceId, List<Pair<AbstractObject, TableColumnId>>>> otherTableMatches) {
		// Map<PropertyId, MutableDouble> correctMap = new HashMap<>();
		//
		// Map<PropertyId, MutableDouble> totalMap = new HashMap<>();

		Map<PropertyId, Map<String, ParameteredMatcherScore>> ret = new HashMap<>();

		ImmutableMap<RowIdx, AbstractObject> categoricalIntColumn = null;
		ImmutableMap<RowIdx, AbstractObject> numericColumn = null;
		ImmutableMap<RowIdx, AbstractObject> stringAOColumn = null;

		if (dataType == ValueType.TIME) {
			numericColumn = convertTimeToQuantity(stringColumn, beginIndex, endIndex);

		}

		Set<PropertyId> properties = nonTailProperties;
		for (PropertyId propertyId : properties) {
			ValueType propertyValueType = ontologyRepository.getDataType(propertyId);
			if (!propertyId.equals(PropertiesExpansionData.LABEL_PROPERTY)
					&& (ontologyRepository.getDataType(propertyId) == dataType
							|| (ontologyRepository.getDataType(propertyId) == ValueType.INSTANCE_REFERENCE && dataType == ValueType.STRING)
							|| (ontologyRepository.getDataType(propertyId) == ValueType.CATEGORICAL_INT && dataType == ValueType.QUANTITY)
							|| (ontologyRepository.getDataType(propertyId) == ValueType.CATEGORICAL_INT && dataType == ValueType.STRING)
							|| (ontologyRepository.getDataType(propertyId) == ValueType.QUANTITY
									&& (propertyId.getPropertyId().equals("http://www.w3.org/2003/01/geo/wgs84_pos#lat")
											|| propertyId.getPropertyId().equals("http://www.w3.org/2003/01/geo/wgs84_pos#long"))
									&& dataType == ValueType.COORDINATE)

							|| (ontologyRepository.getDataType(propertyId) == ValueType.STRING)
							|| (propertyValueType == ValueType.QUANTITY && dataType == ValueType.TIME))) {

				Map<String, ParameteredMatcherScore> value = new HashMap<>();

				double correctSum = 0.0;
				double totalSum = 0.0;

				ImmutableMap<RowIdx, AbstractObject> column = rawColumn;
				if (propertyValueType == ValueType.QUANTITY && dataType == ValueType.TIME) {
					column = numericColumn;
				}
				if (ontologyRepository.getDataType(propertyId) == ValueType.CATEGORICAL_INT) {
					if (categoricalIntColumn == null) {
						categoricalIntColumn = createCategoricalIntColumn(stringColumn, beginIndex, endIndex);
						// if (dataType == ValueType.QUANTITY) {
						// categoricalIntColumn =
						// convertQuantityToCategoricalInt(column);
						// } else if (dataType == ValueType.STRING) {
						// Optional<ImmutableMap<RowIdx, AbstractObject>>
						// convertedColumn =
						// convertStringToCategoricalInt(column);
						// if (convertedColumn.isPresent()) {
						// categoricalIntColumn = convertedColumn.get();
						// } else {
						// continue;
						// }
						// }
					}

					column = categoricalIntColumn;
				}

				if (propertyValueType == ValueType.STRING && dataType != ValueType.STRING) {
					if (stringAOColumn == null) {
						stringAOColumn = createStringColumn(stringColumn, beginIndex, endIndex);

					}

					column = stringAOColumn;
				}

				Map<Integer, WeightedScore> duplicateMethodScores = new HashMap<>();

				int sdf = 123;
				for (Entry<Integer, InstanceId> instanceEntry : instances.entrySet()) {

					RowIdx rowIdx = new RowIdx(instanceEntry.getKey());
					AbstractObject ao = column.get(rowIdx);

					if (ao != null) {

						ValueId valueId = new ValueId(instanceEntry.getValue(), propertyId);
						Optional<Boolean> correct = valueEvaluator.correct(valueId, FusedValue.fromObject(ao));
						if (correct.isPresent()) {
							double instanceScore = instanceScores.get(new InstanceIndexComposite(instanceEntry.getKey(), instanceEntry.getValue()));
							// TESTING WITH INSTANCE SCORE SHOWED THAT IT WAS
							// WORSE!!

							totalSum += instanceScore;

							if (correct.get()) {
								correctSum += instanceScore;
							} else {
								int sdffff = 123;
							}

						}

					}

				}

				for (int i = 0; i <= WeightsRandomParallelOptimizer.STEPS; i++) {
					Integer parameter = AttributeMatchingSingleRun.convertParameter(i * WeightsRandomParallelOptimizer.STEP);
					double minWeight = (WeightsRandomParallelOptimizer.STEP * i) * column.size();
					if (correctSum > 0.01) {
						double firstScore = 0.0;
						if (totalSum >= minWeight) {
							firstScore = correctSum / totalSum; // totalSum;
						} else {
							firstScore = correctSum / minWeight; // totalSum;
						}

						duplicateMethodScores.put(parameter, new WeightedScore(firstScore, 1.0));
					} else {
						duplicateMethodScores.put(parameter, new WeightedScore(0.0, 1.0));
					}
				}

				value.put(AttributeMatchingSingleRun.getSuffixed(AttributeMatchingSingleRun.DUPLICATE_METHOD, propertyId), new ParameteredMatcherScore(
						AttributeMatchingSingleRun.getSuffixed(AttributeMatchingSingleRun.DUPLICATE_MIN_WEIGHT, propertyId), duplicateMethodScores));

				ValueType vt = ontologyRepository.getDataType(propertyId);
				ClassPropertyId cpid = new ClassPropertyId(classId, propertyId);
				Set<ClassId> range = ontologyRepository.getRange(cpid);
				if (range == null) {
					range = new HashSet<>();
				} else {
					range = new HashSet<>(range);
				}
				Set<ClassPropertyId> equalProperties = knowledgeBase.getEqualProperties().get(cpid);
				if (equalProperties != null) {
					for (ClassPropertyId equalCpid : equalProperties) {
						Set<ClassId> equalRange = ontologyRepository.getRange(equalCpid);
						if (equalRange != null)
							range.addAll(equalRange);
					}
				}
				if (range.isEmpty()) {
					range = null;
				}
				if (propertyId.getFinalPart().equals("isPartOf")) {
					range = new HashSet<>();
					range.add(new ClassId("http://dbpedia.org/ontology/PopulatedPlace"));
				}

				int maxRowOverlapCount = (int) (0.25 * column.size());

				Map<Pair<Integer, Integer>, WeightedScore> overlapMethodScores = new HashMap<>();

				if (column.isEmpty()) {

					for (int i = 0; i <= WeightsRandomParallelOptimizer.STEPS; i++) {
						Integer firstParameter = AttributeMatchingSingleRun.convertParameter(i * WeightsRandomParallelOptimizer.STEP);

						for (int j = 0; j <= WeightsRandomParallelOptimizer.STEPS; j++) {
							Integer secondParameter = AttributeMatchingSingleRun.convertParameter(j * WeightsRandomParallelOptimizer.STEP);

							overlapMethodScores.put(new Pair<>(firstParameter, secondParameter), new WeightedScore(0.0, 1.0));
						}
					}

				} else {

					Map<AbstractObject, Integer> countColumnValues = column.values().stream()
							.collect(Collectors.groupingBy(Function.identity(), Collectors.reducing(0, e -> 1, Integer::sum)));

					/*
					 * .entrySet().stream()
					 * .collect(Collectors.toMap(Entry::getKey, (e) -> { if
					 * (e.getValue() > maxRowOverlapCount) { return
					 * maxRowOverlapCount; } else { return e.getValue(); } }));(
					 */

					if (vt == ValueType.CATEGORICAL_INT || vt == ValueType.TIME || (vt == ValueType.INSTANCE_REFERENCE && range != null)
							|| vt == ValueType.QUANTITY || vt == ValueType.STRING) {

						Set<TimeObject> vtMap = new HashSet<>();
						Set<StringObject> stringObjectMap = new HashSet<>();
						Set<CategoricalIntObject> ciMAP = new HashSet<>();
						if (vt == ValueType.TIME) {
							vtMap = CHMHelper.getFromCHM(this.allTimeObjectMap, cpid, () -> {
								Set<InstanceId> cpidInstances = knowledgeBase.getInstances(classId, propertyId);
								Set<TimeObject> set = new HashSet<>();

								for (InstanceId instanceId : cpidInstances) {
									List<AbstractObject> valueList = knowledgeBase.getValuesAsList(instanceId, propertyId);
									if (valueList != null) {
										for (AbstractObject innerAo : valueList) {
											if (innerAo.getType().equals(ValueType.TIME)) {
												TimeObject to = ((TimeObject) innerAo);
												set.add(to);
												if (to.getPrecision() == Precision.T11_days || to.getPrecision() == Precision.T10_months) {
													TimeObject yearTo = new TimeObject(to.getYear(), 0, 0, 0, 0, 0, 0, Precision.T09_years);

													set.add(yearTo);
												}
												if (to.getPrecision() == Precision.T11_days) {
													TimeObject monthTo = new TimeObject(to.getYear(), to.getMonth(), 0, 0, 0, 0, 0, Precision.T10_months);

													set.add(monthTo);
												}

											}
										}
									}

								}

								return set;
							});
						}

						if (vt == ValueType.STRING) {
							stringObjectMap = CHMHelper.getFromCHM(this.allStringObjectMap, cpid, () -> {
								Set<InstanceId> cpidInstances = knowledgeBase.getInstances(classId, propertyId);
								Set<StringObject> set = new HashSet<>();

								for (InstanceId instanceId : cpidInstances) {
									List<AbstractObject> valueList = knowledgeBase.getValuesAsList(instanceId, propertyId);
									if (valueList != null) {
										for (AbstractObject innerAo : valueList) {
											if (innerAo.getType().equals(ValueType.STRING)) {
												StringObject to = ((StringObject) innerAo);
												set.add(to);

											}
										}
									}

								}

								return set;
							});
						}

						if (vt == ValueType.CATEGORICAL_INT) {
							ciMAP = CHMHelper.getFromCHM(this.allCIObjectMap, cpid, () -> {
								Set<InstanceId> cpidInstances = knowledgeBase.getInstances(classId, propertyId);
								Set<CategoricalIntObject> set = new HashSet<>();

								for (InstanceId instanceId : cpidInstances) {
									List<AbstractObject> valueList = knowledgeBase.getValuesAsList(instanceId, propertyId);
									if (valueList != null) {
										for (AbstractObject innerAo : valueList) {
											if (innerAo.getType().equals(ValueType.CATEGORICAL_INT)) {
												CategoricalIntObject coObject = ((CategoricalIntObject) innerAo);
												set.add(coObject);

											}
										}
									}

								}

								return set;
							});
						}

						Optional<Pair<Double, Double>> minMax = null;
						Optional<Pair<Double, Double>> extendedMinMax = null;
						if (vt == ValueType.QUANTITY) {

							minMax = numericRangeRepository.getMinMax(cpid);

							extendedMinMax = numericRangeRepository.getExtendedMinMax(cpid);
						}

						List<Integer> correctCount = new ArrayList<>();
						List<Integer> extendedMaxCorrectCount = new ArrayList<>();
						for (Entry<AbstractObject, Integer> aoEntry : countColumnValues.entrySet()) {
							if (vt == ValueType.INSTANCE_REFERENCE) {
								if (aoEntry.getKey().getType() == ValueType.STRING) {
									StringObject so = (StringObject) aoEntry.getKey();
									Set<String> labels = new HashSet<>();
									labels.add(so.getString());
									if (labelMatcher.getMatches(labels, range, MatchStrictness.STRICT, false).isPresent()) {
										correctCount.add(aoEntry.getValue());
									}
								}
							} else if (vt == ValueType.TIME) {
								if (aoEntry.getKey().getType() == ValueType.TIME) {
									correctCount.add(aoEntry.getValue());
									// if (vtMap.contains((TimeObject)
									// aoEntry.getKey())) {
									// correctCount.add(aoEntry.getValue());
									// }
									// else {
									// correctCount.add(aoEntry.getValue());
									// }
								}
							} else if (vt == ValueType.CATEGORICAL_INT) {
								if (aoEntry.getKey().getType() == ValueType.CATEGORICAL_INT) {
									if (ciMAP.contains((CategoricalIntObject) aoEntry.getKey())) {
										correctCount.add(aoEntry.getValue());
									}
								}
							} else if (vt == ValueType.QUANTITY) {

								if (minMax == null || !minMax.isPresent()) {

								} else {
									if (aoEntry.getKey().getType() == ValueType.QUANTITY) {
										double numericValue = ((QuantityObject) aoEntry.getKey()).getValue();

										if (numericValue > minMax.get().getFirst() && numericValue < minMax.get().getSecond()) {
											correctCount.add(aoEntry.getValue());
										} else if (numericValue > extendedMinMax.get().getFirst() && numericValue < extendedMinMax.get().getSecond()) {
											extendedMaxCorrectCount.add(aoEntry.getValue());

										}
									}

									// else if (aoEntry.getKey().getType() ==
									// ValueType.CATEGORICAL_INT) {
									// long numericValue =
									// ((CategoricalIntObject)
									// aoEntry.getKey()).getValue();
									//
									// if (numericValue >
									// minMax.get().getFirst() && numericValue <
									// minMax.get().getSecond()) {
									// secondaryCorrect += aoEntry.getValue();
									// } else if (numericValue >
									// extendedMinMax.get().getFirst() &&
									// numericValue <
									// extendedMinMax.get().getSecond()) {
									// secondaryCorrect += aoEntry.getValue() *
									// 0.5;
									// }
									//
									// } else {
									// throw new IllegalStateException();
									// }

								}

							}

						}

						for (int i = 0; i <= WeightsRandomParallelOptimizer.STEPS; i++) {
							Integer firstParameter = AttributeMatchingSingleRun.convertParameter(i * WeightsRandomParallelOptimizer.STEP);

							for (int j = 0; j <= WeightsRandomParallelOptimizer.STEPS; j++) {
								Integer secondParameter = AttributeMatchingSingleRun.convertParameter(j * WeightsRandomParallelOptimizer.STEP);

								double sum = 0.0;

								double maxOverlap = i * WeightsRandomParallelOptimizer.STEP * column.size();
								double extendedMaxPenalizier = j * WeightsRandomParallelOptimizer.STEP;

								for (Integer c : correctCount) {
									if (c > maxOverlap) {
										sum += maxOverlap;
									} else {
										sum += c;
									}
								}

								for (Integer c : extendedMaxCorrectCount) {
									if (c > maxOverlap) {
										sum += maxOverlap * extendedMaxPenalizier;
									} else {
										sum += c * extendedMaxPenalizier;
									}
								}
								overlapMethodScores.put(new Pair<>(firstParameter, secondParameter), new WeightedScore(sum / column.size(), 1.0));
							}
						}

						// firstScore = Math.min(1.0, firstScore * (1.0 +
						// secondScore));

					} else {
						for (int i = 0; i <= WeightsRandomParallelOptimizer.STEPS; i++) {
							Integer firstParameter = AttributeMatchingSingleRun.convertParameter(i * WeightsRandomParallelOptimizer.STEP);

							for (int j = 0; j <= WeightsRandomParallelOptimizer.STEPS; j++) {
								Integer secondParameter = AttributeMatchingSingleRun.convertParameter(j * WeightsRandomParallelOptimizer.STEP);

								overlapMethodScores.put(new Pair<>(firstParameter, secondParameter), new WeightedScore(0.0, 1.0));
							}
						}
					}

				}
				value.put(AttributeMatchingSingleRun.getSuffixed(AttributeMatchingSingleRun.OVERLAP_METHOD, propertyId),
						new ParameteredMatcherScore(AttributeMatchingSingleRun.getSuffixed(AttributeMatchingSingleRun.MAX_OVERLAP_WEIGHT_PER_VALUE, propertyId),
								AttributeMatchingSingleRun.getSuffixed(AttributeMatchingSingleRun.EXTENDED_MAX_SCORE, propertyId), overlapMethodScores));

				if (labelForPropertyIdMap != null) {
					double labelScore = 0.0;
					if (columnHeader != null) {
						if (labelForPropertyIdMap.get(classId).containsKey(propertyId)
								&& labelForPropertyIdMap.get(classId).get(propertyId).contains(columnHeader)) {

							labelScore = 1.0;
						}
					}
					value.put(AttributeMatchingSingleRun.getSuffixed(AttributeMatchingSingleRun.KB_LABEL, propertyId),
							new ParameteredMatcherScore(new WeightedScore(labelScore, 1.0)));

				}

				if (otherMatchLabelScores != null) {
					double labelScore = 0.0;
					if (columnHeader != null) {
						if (otherMatchLabelScores.get(classId).containsKey(columnHeader)
								&& otherMatchLabelScores.get(classId).get(columnHeader).getSecond().containsKey(propertyId)) {
							Pair<Integer, Map<PropertyId, Set<TableColumnId>>> pair = otherMatchLabelScores.get(classId).get(columnHeader);
							double size = pair.getSecond().get(propertyId).size();
							if (pair.getSecond().get(propertyId).contains(tableColumnId)) {
								size--;
							}
							labelScore = size / pair.getFirst();
						}
					}
					value.put(AttributeMatchingSingleRun.getSuffixed(AttributeMatchingSingleRun.SECONDARY_LABEL, propertyId),
							new ParameteredMatcherScore(new WeightedScore(labelScore, 1.0)));

				}

				if (otherTableMatches != null) {
					Map<Integer, WeightedScore> secondaryDuplicateMethodScores = new HashMap<>();
					Map<PropertyInstanceId, List<Pair<AbstractObject, TableColumnId>>> matchesPerClass = otherTableMatches.get(classId);

					if (!column.isEmpty()) {
						double score = 0.0;
						double secondaryTotalSum = 0.0;
						for (Entry<Integer, InstanceId> instanceEntry : instances.entrySet()) {
							PropertyInstanceId propertyInstanceId = new PropertyInstanceId(propertyId, instanceEntry.getValue());
							if (matchesPerClass.containsKey(propertyInstanceId)) {

								RowIdx rowIdx = new RowIdx(instanceEntry.getKey());
								AbstractObject ao = column.get(rowIdx);

								if (ao != null) {

									List<Pair<AbstractObject, TableColumnId>> matches = matchesPerClass.get(propertyInstanceId);
									if (!matches.isEmpty()) {
										totalSum++;
										for (Pair<AbstractObject, TableColumnId> candidate : matches) {
											if (!candidate.getSecond().equals(tableColumnId) && sc.isEqual(ao, candidate.getFirst())) {
												score++;
												break;
											}
										}
									}
								}

							}
						}

						if (score > 0.0) {
							for (int i = 0; i <= WeightsRandomParallelOptimizer.STEPS; i++) {
								Integer parameter = AttributeMatchingSingleRun.convertParameter(i * WeightsRandomParallelOptimizer.STEP);
								double minWeight = (WeightsRandomParallelOptimizer.STEP * i) * column.size();

								if (secondaryTotalSum > minWeight) {
									secondaryDuplicateMethodScores.put(parameter, new WeightedScore(score / secondaryTotalSum, 1.0));
								} else {

									secondaryDuplicateMethodScores.put(parameter, new WeightedScore(score / minWeight, 1.0));
								}

							}

						} else {
							for (int i = 0; i <= WeightsRandomParallelOptimizer.STEPS; i++) {
								Integer parameter = AttributeMatchingSingleRun.convertParameter(i * WeightsRandomParallelOptimizer.STEP);
								double minWeight = (WeightsRandomParallelOptimizer.STEP * i) * column.size();

								secondaryDuplicateMethodScores.put(parameter, new WeightedScore(0.0, 1.0));

							}
						}

					} else {

						for (int i = 0; i <= WeightsRandomParallelOptimizer.STEPS; i++) {
							Integer parameter = AttributeMatchingSingleRun.convertParameter(i * WeightsRandomParallelOptimizer.STEP);
							double minWeight = (WeightsRandomParallelOptimizer.STEP * i) * column.size();

							secondaryDuplicateMethodScores.put(parameter, new WeightedScore(0.0, 1.0));

						}

					}

					value.put(AttributeMatchingSingleRun.getSuffixed(AttributeMatchingSingleRun.SECONDARY_DUPLICATE, propertyId),
							new ParameteredMatcherScore(
									AttributeMatchingSingleRun.getSuffixed(AttributeMatchingSingleRun.SECONDARY_DUPLICATE_MIN_WEIGHT, propertyId),
									secondaryDuplicateMethodScores));

				}

				// Map<String, ParameteredMatcherScore> valuePrefixed = new
				// HashMap<>();
				// for (Entry<String, ParameteredMatcherScore> unprefixed :
				// value.entrySet()) {
				// valuePrefixed.put(unprefixed.getKey() + "_" +
				// propertyId.getPropertyId(), unprefixed.getValue());
				// }
				ret.put(propertyId, value);

			}
		}

		return ret;

	}

	public static ImmutableMap<RowIdx, AbstractObject> createStringColumn(ImmutableMap<RowIdx, String> stringColumn, int beginIndex, int endIndex) {
		Builder<RowIdx, AbstractObject> builder = ImmutableMap.builder();

		for (int i = beginIndex; i < endIndex; i++) {
			RowIdx rowIdx = new RowIdx(i);
			String sValue = stringColumn.get(rowIdx);
			if (sValue != null) {
				builder.put(rowIdx, new StringObject(AttributeMatcher.normaliseValueLenient(sValue)));
			}

		}

		return builder.build();

	}

	public static ImmutableMap<RowIdx, AbstractObject> createCategoricalIntColumn(ImmutableMap<RowIdx, String> stringColumn, int beginIndex, int endIndex) {

		Builder<RowIdx, AbstractObject> builder = ImmutableMap.builder();

		for (int i = beginIndex; i < endIndex; i++) {
			RowIdx rowIdx = new RowIdx(i);
			String sValue = stringColumn.get(rowIdx);
			if (sValue != null) {
				String stringValue = normaliseValueLenient(sValue);

				if (AttributeMatcher.isInteger(stringValue, 10)) {
					int value = Integer.parseInt(stringValue);
					builder.put(rowIdx, new CategoricalIntObject(value));
				} else {
					try {
						Double typed = TableColumnBuilder.typeNumeric(stringValue, null);
						QuantityObject qo = QuantityObject.fromDouble(typed);

						long longValue = Math.round(qo.getValue());

						builder.put(rowIdx, new CategoricalIntObject(longValue));
					} catch (java.text.ParseException e) {
						int sdf = 123;
					}
				}

			}

		}
		ImmutableMap<RowIdx, AbstractObject> newMap = builder.build();
		return newMap;

	}

	public static ImmutableMap<RowIdx, AbstractObject> convertTimeToQuantity(ImmutableMap<RowIdx, String> stringColumn, int beginIndex, int endIndex) {

		Builder<RowIdx, AbstractObject> builder = ImmutableMap.builder();

		for (int i = beginIndex; i < endIndex; i++) {
			RowIdx rowIdx = new RowIdx(i);
			String sValue = stringColumn.get(rowIdx);
			if (sValue != null) {
				String value = normaliseValueLenient(sValue);
				try {
					Double typed = TableColumnBuilder.typeNumeric(value, null);
					QuantityObject qo = QuantityObject.fromDouble(typed);

					builder.put(rowIdx, qo);
				} catch (java.text.ParseException e) {
					int sdf = 123;
				}
			}

		}
		ImmutableMap<RowIdx, AbstractObject> newMap = builder.build();
		return newMap;

	}

	public static Optional<ImmutableMap<RowIdx, AbstractObject>> convertStringToCategoricalInt(ImmutableMap<RowIdx, AbstractObject> column) {
		Builder<RowIdx, AbstractObject> builder = ImmutableMap.builder();

		for (Entry<RowIdx, AbstractObject> entry : column.entrySet()) {

			if (entry.getValue().getType() == ValueType.STRING) {
				StringObject so = (StringObject) entry.getValue();

				if (AttributeMatcher.isInteger(so.getString(), 10)) {
					int value = Integer.parseInt(so.getString());
					builder.put(entry.getKey(), new CategoricalIntObject(value));
				}
			}

		}
		ImmutableMap<RowIdx, AbstractObject> newMap = builder.build();
		if (newMap.size() > (0.5 * column.size())) {
			return Optional.of(newMap);
		} else {
			return Optional.empty();
		}

	}

	public static ImmutableMap<RowIdx, AbstractObject> convertQuantityToCategoricalInt(ImmutableMap<RowIdx, AbstractObject> column) {
		Builder<RowIdx, AbstractObject> builder = ImmutableMap.builder();

		for (Entry<RowIdx, AbstractObject> entry : column.entrySet()) {

			if (entry.getValue().getType() == ValueType.QUANTITY) {
				QuantityObject qo = (QuantityObject) entry.getValue();

				long longValue = Math.round(qo.getValue());

				builder.put(entry.getKey(), new CategoricalIntObject(longValue));

			}

		}
		ImmutableMap<RowIdx, AbstractObject> newMap = builder.build();
		return newMap;
	}

	public Optional<PropertyMapping> getWeightedMatch(ClassId classId, TableColumnId tableColumnId, Set<PropertyId> nonTailProperties,
			Map<Integer, InstanceId> instances, Map<InstanceIndexComposite, Double> instanceScores, ImmutableMap<RowIdx, AbstractObject> column,
			ImmutableMap<RowIdx, String> stringColumn, ValueType dataType, WeightsRepository weightsRepository, OptimizableClass optmizableClazz,
			int beginIndex, int endIndex, String columnHeader, Map<ClassId, Map<PropertyId, Set<String>>> labelForPropertyIdMap,
			Map<ClassId, Map<String, Pair<Integer, Map<PropertyId, Set<TableColumnId>>>>> otherMatchLabelScores,
			Map<ClassId, Map<PropertyInstanceId, List<Pair<AbstractObject, TableColumnId>>>> otherTableMatches, boolean baseline) {
		// Map<PropertyId, MutableDouble> correctMap = new HashMap<>();
		//
		// Map<PropertyId, MutableDouble> totalMap = new HashMap<>();

		Map<PropertyId, Map<String, ParameteredMatcherScore>> ret = getFullMatches(classId, tableColumnId, nonTailProperties, instances, instanceScores, column,
				stringColumn, dataType, beginIndex, endIndex, columnHeader, labelForPropertyIdMap, otherMatchLabelScores, otherTableMatches);

		if (baseline) {
			ret = filterForBaseline(ret);
		}
		// double maxScore = 0.0;
		// PropertyId maxPropertyId = null;

		return equation.get(optmizableClazz, classId, new AttributeMatchingSingleRun(ret), weightsRepository);

		// for (Entry<PropertyId, Map<String, Double>> entry : ret.entrySet()) {
		// ClassPropertyId replace = this.propertyConverter.get(new
		// ClassPropertyId(classId, entry.getKey()));
		//
		//
		//
		// PropertyId propertyId = entry.getKey();
		// double overlap =
		// entry.getValue().get(AttributeMatchingSingleRun.OVERLAP_METHOD);
		// double overlapCount =
		// weightsRepository.getOrDefault(optmizableClazz.getWeight(AttributeMatchingSingleRun.OVERLAP_METHOD,
		// 0.0));
		//
		// double duplicate =
		// entry.getValue().get(AttributeMatchingSingleRun.DUPLICATE_METHOD);
		// double duplicateCount =
		// weightsRepository.getOrDefault(optmizableClazz.getWeight(AttributeMatchingSingleRun.DUPLICATE_METHOD,
		// 0.0));
		//
		// double score = (overlap * overlapCount) + (duplicate *
		// duplicateCount);
		// double count = overlapCount + duplicateCount;
		//
		// if (replace != null) {
		// propertyId = replace.getPropertyId();
		// }
		//
		// if (count > 0.001) {
		// score = score / count;
		//
		// if (maxPropertyId == null || score > maxScore) {
		// maxPropertyId = propertyId;
		// score = maxScore;
		// }
		// }
		//
		// }
		//
		// if (maxPropertyId != null
		// && maxScore >
		// weightsRepository.getOrDefault(optmizableClazz.getWeight(AttributeMatchingSingleRun.getThreshold(maxPropertyId),
		// 0.0))) {
		// return Optional.of(maxPropertyId);
		// }
		// return Optional.empty();
	}
	//
	// public Map<PropertyId, Double> getHighestMatches(ClassId classId,
	// TableColumnId tableColumnId, Set<PropertyId> nonTailProperties,
	// Map<Integer, InstanceId> instances, Map<InstanceIndexComposite, Double>
	// instanceScores, ImmutableMap<RowIdx, AbstractObject> column,
	// ImmutableMap<RowIdx, String> stringColumn, ValueType dataType, int
	// beginIndex, int endIndex) {
	// // Map<PropertyId, MutableDouble> correctMap = new HashMap<>();
	// //
	// // Map<PropertyId, MutableDouble> totalMap = new HashMap<>();
	//
	// Map<PropertyId, Map<String, ParameteredMatcherScore>> ret =
	// getFullMatches(classId, tableColumnId, nonTailProperties, instances,
	// instanceScores, column,
	// stringColumn, dataType, beginIndex, endIndex, null, null, null, null);
	//
	// Map<PropertyId, Double> ret2 = new HashMap<>();
	// for (Entry<PropertyId, Map<String, ParameteredMatcherScore>> entry :
	// ret.entrySet()) {
	// ClassPropertyId replace = this.propertyConverter.get(new
	// ClassPropertyId(classId, entry.getKey()));
	// PropertyId propertyId = entry.getKey();
	// WeightedScore overlap =
	// entry.getValue().get(AttributeMatchingSingleRun.OVERLAP_METHOD);
	// WeightedScore duplicate =
	// entry.getValue().get(AttributeMatchingSingleRun.DUPLICATE_METHOD);
	//
	// double score = Math.min(1.0, overlap.getScore() * (1.0 +
	// duplicate.getScore()));
	// ;
	// if (replace != null) {
	// propertyId = replace.getPropertyId();
	// }
	//
	// if (!ret2.containsKey(propertyId) || score > ret2.get(propertyId)) {
	// ret2.put(propertyId, score);
	// }
	// }
	//
	// return ret2;
	// }

	public AttributeMatchingFullRuns getFullMatches(List<AnnotatedTablesReaderData> tables, Map<ClassId, Set<PropertyId>> nonTailProperties,
			AttributeMappings mappingsForSecondary, boolean baseline) {

		Map<ClassId, Map<String, Pair<Integer, Map<PropertyId, Set<TableColumnId>>>>> otherMatchLabelScores;
		Map<ClassId, Map<PropertyInstanceId, List<Pair<AbstractObject, TableColumnId>>>> otherTableMatches;
		if (mappingsForSecondary != null) {
			otherMatchLabelScores = getLabelScores(tables, mappingsForSecondary);
			otherTableMatches = getOtherTableMatch(tables, mappingsForSecondary);
		} else {
			otherMatchLabelScores = null;
			otherTableMatches = null;
		}

		Map<ClassId, Map<PropertyId, Set<String>>> labelForPropertyIdMap = new HashMap<>();
		for (Entry<ClassId, Set<PropertyId>> nonTailPropertiesByClass : nonTailProperties.entrySet()) {
			labelForPropertyIdMap.put(nonTailPropertiesByClass.getKey(),
					getLabelsForPropertyIdSet(nonTailPropertiesByClass.getKey(), nonTailPropertiesByClass.getValue()));
		}

		Set<TableColumnId> tableIds = new HashSet<>();
		tableIds.add(new TableColumnId(new AnnotatedTableId("1346823846176_1346846758903_2725.arc2481406962457227330#22326655_2_1346037750385205130"), 0));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1346876860779_1346961089435_3900.arc5504794985066264630#30219675_1_433888948023861627"), 4));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1346876860609_1346965067128_728.arc4941513350675067041#53635965_0_4100477248612278954"), 2));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1346876860565_1346912219601_1704.arc8689279114261751938#3512341_5_5919837545333214767"), 5));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1346876860779_1346961276986_5601.arc3719795019286941883#45451800_0_7520172109909715831"), 2));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1346823846150_1346831707124_98.arc8289082661436581445#58505000_0_3971736499393838515"), 3));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1346876860791_1346881958401_2145.arc8468166189656932701#63206027_0_264701290058443984"), 0));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1346876860767_1346930413885_4219.arc6133765969811430556#8067855_0_7255939946700078827"), 5));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1346981172258_1346986443836_2250.arc3298195217085472013#17077555_0_2384416577531919178"), 3));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1346876860807_1346935022722_2014.arc8333932477257942082#17751131_0_8206891990416770602"), 2));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1346823846125_1346831228806_4601.arc19611991799741214#5774269_4_6373690942998186890"), 1));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1346876860777_1346911201139_5302.arc4950307784433668143#99781539_0_145784639298511497"), 1));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1346876860467_1346894614144_1467.arc8229090527093920145#26816383_0_5184733204356271199"), 3));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1346876860774_1346976279740_1167.arc5193714840123267708#23770978_0_8843872857851427183"), 3));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1346876860774_1346976279740_1167.arc5193714840123267708#23770978_0_8843872857851427183"), 4));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1346876860828_1346952444486_2789.arc4355746145008370050#37881601_0_5165264125271974655"), 1));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1346823846125_1346825813463_6.arc2362183109663985770#35523234_0_194649281147111494"), 6));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1350433107027_1350488164308_684.arc2045104594958977797#30633610_0_7633630560803713625"), 0));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1350433107027_1350488164308_684.arc2045104594958977797#30633610_0_7633630560803713625"), 1));
		tableIds.add(new TableColumnId(new AnnotatedTableId("1350433107027_1350488164308_684.arc2045104594958977797#30633610_0_7633630560803713625"), 4));

		Map<TableColumnId, AnnotatedTablesReaderData> tableColumnIdsSet = new HashMap<>();
		Map<AnnotatedTableId, ClassId> classIds = new ConcurrentHashMap<>();
		for (AnnotatedTablesReaderData table : tables) {

			for (ColumnIdx columnIdx : table.getTableData().getColumns()) {
				if (!columnIdx.equals(table.getTableData().getKeyColumnIndex())) {
					tableColumnIdsSet.put(new TableColumnId(table.getId(), columnIdx), table);
				}
			}

			classIds.put(table.getId(), table.getMetaData().getDbPediaClass());
		}

		Map<TableColumnId, AttributeMatchingSingleRun> mappings = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(tableColumnIdsSet.entrySet(), (entry) -> {
			if (tableIds.contains(entry.getKey())) {
				int sdf = 123;
			}
			AnnotatedTablesReaderData table = entry.getValue();

			int beginIndex = table.getMetaData().getNumHeaderRows();
			int endIndex = table.getMetaData().getNumberOfRows() + table.getMetaData().getNumHeaderRows();

			ColumnIdx columnIdx = entry.getKey().getColumnIdx();

			ImmutableMap<RowIdx, AbstractObject> column = table.getTableData().getColumn(columnIdx);
			ImmutableMap<RowIdx, String> stringColumn = table.getTableData().getStringTable().column(columnIdx);

			String columnHeader = null;
			AbstractObject columnHeaderSO = table.getTableData().getColumnHeader(columnIdx);
			if (columnHeaderSO != null && columnHeaderSO.getType() == ValueType.STRING) {
				columnHeader = ((StringObject) columnHeaderSO).getString();
			}

			Map<PropertyId, Map<String, ParameteredMatcherScore>> fullMatches = getFullMatches(table.getMetaData().getDbPediaClass(),
					new TableColumnId(table.getId(), columnIdx), nonTailProperties.get(table.getMetaData().getDbPediaClass()),
					table.getMetaData().getInstances(), table.getMetaData().getInstancesScores(), column, stringColumn,
					table.getTableData().getDataType(columnIdx), beginIndex, endIndex, columnHeader, labelForPropertyIdMap, otherMatchLabelScores,
					otherTableMatches);

			if (baseline) {
				fullMatches = filterForBaseline(fullMatches);
			}

			mappings.put(new TableColumnId(table.getId(), columnIdx), new AttributeMatchingSingleRun(fullMatches));
			if (tableIds.contains(entry.getKey())) {
				int sdf = 123;
			}
		}, this.getClass());

		return new AttributeMatchingFullRuns(mappings, classIds);

	}

	private Map<PropertyId, Map<String, ParameteredMatcherScore>> filterForBaseline(Map<PropertyId, Map<String, ParameteredMatcherScore>> fullMatches) {

		Map<PropertyId, Map<String, ParameteredMatcherScore>> newRet = new HashMap<>();
		for (Entry<PropertyId, Map<String, ParameteredMatcherScore>> entry : fullMatches.entrySet()) {
			Map<String, ParameteredMatcherScore> innerMap = entry.getValue().entrySet()
					.stream().filter((e) -> e.getKey().equals(AttributeMatchingSingleRun.KB_LABEL)
							|| e.getKey().equals(AttributeMatchingSingleRun.SECONDARY_LABEL) || e.getKey().equals(AttributeMatchingSingleRun.OVERLAP_METHOD))
					.collect(Collectors.toMap(Entry::getKey, Entry::getValue));
			if (!innerMap.isEmpty()) {
				newRet.put(entry.getKey(), innerMap);
			}
		}

		return newRet;
	}

	public static boolean isInteger(String s, int radix) {
		if (s.isEmpty())
			return false;
		for (int i = 0; i < s.length(); i++) {
			if (i == 0 && s.charAt(i) == '-') {
				if (s.length() == 1)
					return false;
				else
					continue;
			}
			if (Character.digit(s.charAt(i), radix) < 0)
				return false;
		}
		return true;
	}

	public static String normaliseValueLenient(String value) {
		try {
			value = value.replaceAll("\n", "");
			value = value.replace("&nbsp;", " ");
			value = value.replace("&nbsp", " ");

			value = value.replace("nbsp", " ");
			value = value.replace("&quot;", " ");
			value = value.replace("", "-");
			value = value.replaceAll("<.*>", "");

			value = value.toLowerCase();
			value = value.trim();

			if (value.equals("")) {
				value = StringNormalizer.nullValue;
			}

		} catch (Exception e) {
		}
		return value;
	}

	public static final PropertyId noProperty = PropertyId.fromString("NO_PROPERTY");

	Map<ClassId, Map<String, Pair<Integer, Map<PropertyId, Set<TableColumnId>>>>> getLabelScores(List<AnnotatedTablesReaderData> tables,
			AttributeMappings mappings) {
		Map<AnnotatedTableId, TableAttributeMappings> mappingsMap = mappings.getMappings();
		ConcurrentHashMap<ClassId, ConcurrentHashMap<String, ConcurrentHashMap<PropertyId, Set<TableColumnId>>>> collector = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(tables, (table) -> {
			TableAttributeMappings tableMappings = mappingsMap.get(table.getId());
			for (ColumnIdx columnIdx : table.getTableData().getColumns()) {
				if (!table.getTableData().getKeyColumnIndex().equals(columnIdx)) {

					StringObject header = (StringObject) table.getTableData().getColumnHeader(columnIdx);

					ConcurrentHashMap<String, ConcurrentHashMap<PropertyId, Set<TableColumnId>>> innerMap = CHMHelper.getFromCHM(collector,
							table.getMetaData().getDbPediaClass(), () -> new ConcurrentHashMap<>());
					ConcurrentHashMap<PropertyId, Set<TableColumnId>> innerQueue = CHMHelper.getFromCHM(innerMap, header.getString(),
							() -> new ConcurrentHashMap<>());

					if (tableMappings != null && !tableMappings.getMappings().isEmpty()) {
						PropertyMapping currentMapping = tableMappings.getMappings().get(columnIdx);
						if (currentMapping != null) {
							CHMHelper.getFromCHM(innerQueue, currentMapping.getPropertyId(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>()))
									.add(new TableColumnId(table.getId(), columnIdx));

						} else {
							CHMHelper.getFromCHM(innerQueue, noProperty, () -> Collections.newSetFromMap(new ConcurrentHashMap<>()))
									.add(new TableColumnId(table.getId(), columnIdx));

						}

					} else {

						CHMHelper.getFromCHM(innerQueue, noProperty, () -> Collections.newSetFromMap(new ConcurrentHashMap<>()))
								.add(new TableColumnId(table.getId(), columnIdx));

					}
				}

			}
		}, this.getClass());

		Map<ClassId, Map<String, Pair<Integer, Map<PropertyId, Set<TableColumnId>>>>> ret = new HashMap<>();

		for (Entry<ClassId, ConcurrentHashMap<String, ConcurrentHashMap<PropertyId, Set<TableColumnId>>>> entry : collector.entrySet()) {
			Map<String, Pair<Integer, Map<PropertyId, Set<TableColumnId>>>> classMap = new HashMap<>();

			for (Entry<String, ConcurrentHashMap<PropertyId, Set<TableColumnId>>> stringEntry : entry.getValue().entrySet()) {
				int size = stringEntry.getValue().values().stream().mapToInt((e) -> e.size()).sum();
				classMap.put(stringEntry.getKey(), new Pair<>(size, ImmuMaker.mapWithSet(stringEntry.getValue())));
			}

			ret.put(entry.getKey(), Collections.unmodifiableMap(classMap));
		}

		return Collections.unmodifiableMap(ret);
	}

	Map<ClassId, Map<PropertyInstanceId, List<Pair<AbstractObject, TableColumnId>>>> getOtherTableMatch(List<AnnotatedTablesReaderData> tables,
			AttributeMappings mappings) {
		Map<AnnotatedTableId, TableAttributeMappings> mappingsMap = mappings.getMappings();
		ConcurrentHashMap<ClassId, ConcurrentHashMap<PropertyInstanceId, ConcurrentLinkedQueue<Pair<AbstractObject, TableColumnId>>>> collector = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(tables, (table) -> {
			TableAttributeMappings tableMappings = mappingsMap.get(table.getId());
			if (tableMappings != null && !tableMappings.getMappings().isEmpty() && !table.getMetaData().getInstances().isEmpty()) {
				ConcurrentHashMap<PropertyInstanceId, ConcurrentLinkedQueue<Pair<AbstractObject, TableColumnId>>> innerMap = CHMHelper.getFromCHM(collector,
						table.getMetaData().getDbPediaClass(), () -> new ConcurrentHashMap<>());

				for (Entry<ColumnIdx, PropertyMapping> propertyEntry : tableMappings.getMappings().entrySet()) {

					for (Entry<Integer, InstanceId> instanceEntry : table.getMetaData().getInstances().entrySet()) {
						AbstractObject ao = table.getTableData().getValue(new RowIdx(instanceEntry.getKey()), propertyEntry.getKey());

						if (ao != null) {
							CHMHelper.getFromCHM(innerMap, new PropertyInstanceId(propertyEntry.getValue().getPropertyId(), instanceEntry.getValue()),
									() -> new ConcurrentLinkedQueue<>()).add(new Pair<>(ao, new TableColumnId(table.getId(), propertyEntry.getKey())));
						}
					}

				}
			}
		}, this.getClass());

		Function<Entry<ClassId, ConcurrentHashMap<PropertyInstanceId, ConcurrentLinkedQueue<Pair<AbstractObject, TableColumnId>>>>, Map<PropertyInstanceId, List<Pair<AbstractObject, TableColumnId>>>> tansformMap = (
				entry) -> {

			return Collections.unmodifiableMap(entry.getValue().entrySet().stream()
					.collect(Collectors.toMap(Entry::getKey, (e2) -> Collections.unmodifiableList(new ArrayList<>(e2.getValue())))));

		};

		return Collections.unmodifiableMap(collector.entrySet().stream().collect(Collectors.toMap(Entry::getKey, tansformMap)));

	}

	public Map<PropertyId, Set<String>> getLabelsForPropertyIdSet(ClassId classId, Set<PropertyId> propertyIds) {
		Map<PropertyId, Set<String>> ret = new HashMap<>();
		for (PropertyId propertyId : propertyIds) {

			Set<String> labels = new HashSet<>();
			labels.addAll(getLabels(propertyId));

			PropertyId normalized = knowledgeBase.getNormalizationIndex().get(propertyId);
			Optional<PropertyId> normalizedFrom = knowledgeBase.getNormalizationIndex().entrySet().stream().filter((e) -> e.getValue().equals(propertyId))
					.map(Entry::getKey).findFirst();

			if (normalized != null) {
				labels.addAll(getLabels(normalized));
			}

			if (normalizedFrom.isPresent()) {
				labels.addAll(getLabels(normalizedFrom.get()));
			}

			Set<ClassPropertyId> equalProperties = knowledgeBase.getEqualProperties().get(new ClassPropertyId(classId, propertyId));
			if (equalProperties != null) {
				for (ClassPropertyId equalPropertyId : equalProperties) {
					labels.addAll(getLabels(equalPropertyId.getPropertyId()));
				}
			}

			ret.put(propertyId, labels);
		}
		return ret;
	}

	public Set<String> getLabels(PropertyId propertyId) {
		Set<String> labels = new HashSet<>();
		String finalPart = propertyId.getFinalPart();
		labels.add(finalPart);

		int lastPoint = 0;
		String combined = "";
		for (int i = 0; i < finalPart.length(); i++) {
			if (Character.isUpperCase(i)) {
				String part = finalPart.substring(lastPoint, i + 1);
				if (!combined.equals("")) {
					combined += " ";
				}
				combined += part;
				lastPoint = i + 1;
			}
		}

		if (!combined.equals("")) {
			labels.add(combined);
		}

		return labels;
	}

	// public static final String BEFORE_VALUE_ADAPTION =
	// "BEFORE_VALUE_ADAPTION";

	public static List<AnnotatedTableFilter> getValueAdaptionFilters(OntologyRepository ontologyRepository) {
		List<AnnotatedTableFilter> filters = new ArrayList<>();
		filters.add(new TableColumnTimeToQuantityConverter(ontologyRepository));
		filters.add(new WeightInPoundsTableConverter());
		filters.add(new CategoricalIntTableConverter(ontologyRepository));
		filters.add(new StringPropertyConverter(ontologyRepository));

		return filters;
	}
}
