package de.uni_mannheim.informatik.dws.attribute.matcher;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.ImmuMaker;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.informatik.dws.attribute.matcher.AttributeMatchingSingleRun.ParameteredMatcherScore;
import de.uni_mannheim.informatik.dws.attribute.model.PropertyMapping;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeight;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;

public class AttributeMatcherEquation {

	private final Map<ClassPropertyId, ClassPropertyId> propertyConverter;

	@Inject
	public AttributeMatcherEquation(KnowledgeBase knowledgeBase) {
		Map<ClassPropertyId, ClassPropertyId> propertyConverter = new HashMap<>();
		for (Entry<ClassPropertyId, Set<ClassPropertyId>> entry : knowledgeBase.getEqualProperties().entrySet()) {
			for (ClassPropertyId inner : entry.getValue()) {
				propertyConverter.put(inner, entry.getKey());
			}
		}
		this.propertyConverter = ImmuMaker.map(propertyConverter);
	}

	public Optional<PropertyMapping> get(OptimizableClass optmizableClazz, ClassId classId, AttributeMatchingSingleRun attributeMatchingSingleRun,
			WeightsRepository weightRepository) {
		double maxScore = 0.0;
		PropertyId maxPropertyId = null;

		for (Entry<PropertyId, Map<String, ParameteredMatcherScore>> propertyEntry : attributeMatchingSingleRun.getScores().entrySet()) {
			double sum = 0.0;
			double count = 0.0;

			// double threshold =
			// weightRepository.getOrDefault(optmizableClazz.getWeight(AttributeMatchingSingleRun.getThreshold(propertyEntry.getKey()),
			// 0.0));
			double threshold = weightRepository.getOrDefault(optmizableClazz.getWeight(AttributeMatchingSingleRun.getThreshold(propertyEntry.getKey()), 0.5));

			for (Entry<String, ParameteredMatcherScore> scoreEntry : propertyEntry.getValue().entrySet()) {
				OptimizableWeight weightKey = optmizableClazz.getWeight(scoreEntry.getKey(), 1.0);
				double weight = weightRepository.getOrDefault(weightKey);
				WeightedScore score = null;

				Optional<Set<String>> parameters = scoreEntry.getValue().getParameters();
				if (parameters.isPresent()) {
					Map<String, Double> parameterKeys = new HashMap<>();
					for (String paramter : parameters.get()) {
						parameterKeys.put(paramter, weightRepository.getOrDefault(optmizableClazz.getWeight(paramter, 1.0)));
					}
					score = scoreEntry.getValue().getScore(parameterKeys);
				} else {
					score = scoreEntry.getValue().getScore();
				}

				sum += score.getScore() * weight * score.getWeight();
				count += weight * score.getWeight();

			}
			double score = sum / count;
			// if (score > threshold) {
			if (maxPropertyId == null || score > maxScore) {
				maxScore = score;
				maxPropertyId = propertyEntry.getKey();
			}
			// }
		}

		if (maxPropertyId != null) {
			double threshold = weightRepository.getOrDefault(optmizableClazz.getWeight(AttributeMatchingSingleRun.getThreshold(maxPropertyId), 0.5));

			ClassPropertyId replace = this.propertyConverter.get(new ClassPropertyId(classId, maxPropertyId));
			if (replace != null) {
				maxPropertyId = replace.getPropertyId();
			}

			// double threshold =
			// weightRepository.getOrDefault(optmizableClazz.getWeight("THRESHOLD",
			// 0.5));

			if (maxScore > threshold) {
				if (maxPropertyId.getFinalPart().equals("height")) {
					int sdf = 123;
				}
				return Optional.of(new PropertyMapping(maxPropertyId, maxScore));
			}

			// return Optional.of(maxPropertyId);
		}

		return Optional.empty();
	}
}
