package de.uni_mannheim.informatik.dws.attribute.matcher;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ImmuMaker;

public class AttributeMatchingFullRuns {

	public final Map<TableColumnId, AttributeMatchingSingleRun> allRuns;
	private final Map<AnnotatedTableId, ClassId> classIds;

	public AttributeMatchingFullRuns(Map<TableColumnId, AttributeMatchingSingleRun> allRuns, Map<AnnotatedTableId, ClassId> classIds) {
		super();
		this.allRuns = Collections.unmodifiableMap(new HashMap<>(allRuns));
		this.classIds = ImmuMaker.map(classIds);
	}

	public Map<TableColumnId, AttributeMatchingSingleRun> getAllRuns() {
		return allRuns;
	}

	public Map<TableColumnId, AttributeMatchingSingleRun> getAllRuns(Set<AnnotatedTableId> set) {
		return allRuns.entrySet().stream().filter((e) -> set.contains(e.getKey().getTableId())).collect(Collectors.toMap(Entry::getKey, Entry::getValue));
	}

	public Map<AnnotatedTableId, ClassId> getClassIds() {
		return classIds;
	}

	public ClassId getClassId(AnnotatedTableId tableId) {
		return classIds.get(tableId);
	}

	public Map<ClassId, AttributeMatchingFullRuns> getByClass() {
		Map<ClassId, Map<TableColumnId, AttributeMatchingSingleRun>> ret = new HashMap<>();
		Map<ClassId, Map<AnnotatedTableId, ClassId>> classIds = new HashMap<>();

		for (Entry<TableColumnId, AttributeMatchingSingleRun> entry : allRuns.entrySet()) {
			ClassId classId = this.classIds.get(entry.getKey().getTableId());
			CHMHelper.getFromMap(ret, classId, () -> new HashMap<>()).put(entry.getKey(), entry.getValue());

		}

		for (Entry<AnnotatedTableId, ClassId> classIdEntry : this.classIds.entrySet()) {

			CHMHelper.getFromMap(classIds, classIdEntry.getValue(), () -> new HashMap<>()).put(classIdEntry.getKey(), classIdEntry.getValue());
		}

		return ret.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> new AttributeMatchingFullRuns(e.getValue(), classIds.get(e.getKey()))));
	}

}
