package de.uni_mannheim.informatik.dws.attribute.model;

import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.utils.ImmuMaker;

public class AttributeMappings {
	private final Map<AnnotatedTableId, TableAttributeMappings> mappings;
	private final Map<AnnotatedTableId, ClassId> classIds;

	public AttributeMappings(Map<AnnotatedTableId, TableAttributeMappings> mappings, Map<AnnotatedTableId, ClassId> classIds) {
		this.mappings = ImmuMaker.map(mappings);
		this.classIds = ImmuMaker.map(classIds);
	}

	public Map<AnnotatedTableId, TableAttributeMappings> getMappings() {
		return mappings;
	}

	public Set<AnnotatedTableId> tableSet() {
		return mappings.keySet();
	}

	public Set<TableColumnId> columnSet() {
		Set<TableColumnId> ret = new HashSet<>();

		for (Entry<AnnotatedTableId, TableAttributeMappings> entry : mappings.entrySet()) {
			for (Entry<ColumnIdx, PropertyMapping> innerEntry : entry.getValue().getMappings().entrySet()) {
				ret.add(new TableColumnId(entry.getKey(), innerEntry.getKey()));
			}
		}

		return Collections.unmodifiableSet(ret);
	}

	public Optional<PropertyMapping> getMapping(AnnotatedTableId tableId, ColumnIdx colIdx) {
		TableAttributeMappings innerMappings = mappings.get(tableId);
		if (innerMappings != null) {
			PropertyMapping propertyId = innerMappings.getMappings().get(colIdx);
			if (propertyId != null) {
				return Optional.of(propertyId);
			}
		}
		return Optional.empty();
	}

	public Optional<PropertyMapping> getMapping(TableColumnId columnId) {
		return getMapping(columnId.getTableId(), columnId.getColumnIdx());
	}

	public Map<AnnotatedTableId, ClassId> getClassIds() {
		return classIds;
	}

	public ClassId getClassId(AnnotatedTableId tableId) {
		return classIds.get(tableId);
	}

}
