package de.uni_mannheim.informatik.dws.attribute.valueAdaption;

import java.util.Map.Entry;
import java.util.Optional;
import java.util.SortedMap;
import java.util.TreeMap;

import com.google.common.collect.HashBasedTable;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Table;

import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.attribute.matcher.AttributeMatcher;

public class CategoricalIntTableConverter implements AnnotatedTableFilter {

	private final OntologyRepository ontology;

	public CategoricalIntTableConverter(OntologyRepository ontology) {
		this.ontology = ontology;
	}

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {
		if (data.getId().getTableName().equals("1346981172231_1347011264447_3050.arc9167679534671747539#2935814_0_7628926464279511553")) {
			int sdf = 123;
		}

		Table<RowIdx, ColumnIdx, AbstractObject> replace = null;
		SortedMap<ColumnIdx, ValueType> newDataTypes = null;

		int beginIndex = data.getMetaData().getNumHeaderRows();
		int endIndex = data.getMetaData().getNumberOfRows() + data.getMetaData().getNumHeaderRows();

		for (Entry<ColumnIdx, ValueType> dataType : data.getTableData().getDataTypes().entrySet()) {
			PropertyId propertyId = data.getMetaData().getProperties().get(dataType.getKey().getColumnIdx());
			if (propertyId != null && ontology.getDataType(propertyId) == ValueType.CATEGORICAL_INT && dataType.getValue() != ValueType.CATEGORICAL_INT) {

				if (replace == null) {
					replace = HashBasedTable.create(data.getTableData().getTable());
					newDataTypes = new TreeMap<>(data.getTableData().getDataTypes());
				}

				newDataTypes.put(dataType.getKey(), ValueType.CATEGORICAL_INT);

				ImmutableMap<RowIdx, AbstractObject> newColumn = AttributeMatcher
						.createCategoricalIntColumn(data.getTableData().getStringTable().column(dataType.getKey()), beginIndex, endIndex);

				for (int i = beginIndex; i < endIndex; i++) {
					RowIdx rowIdx = new RowIdx(i);
					if (!newColumn.containsKey(rowIdx)) {
						replace.remove(rowIdx, dataType.getKey());
					} else {
						replace.put(rowIdx, dataType.getKey(), newColumn.get(rowIdx));
					}

				}

				// ImmutableMap<RowIdx, AbstractObject> newColumn;
				// if (dataType.getValue() == ValueType.STRING) {
				// Optional<ImmutableMap<RowIdx, AbstractObject>> ret =
				// AttributeMatcher
				// .convertStringToCategoricalInt(data.getTableData().getColumn(dataType.getKey()));
				// if (ret.isPresent()) {
				// newColumn = ret.get();
				// } else {
				// throw new IllegalStateException();
				// }
				// } else if (dataType.getValue() == ValueType.QUANTITY) {
				// newColumn =
				// AttributeMatcher.convertQuantityToCategoricalInt(data.getTableData().getColumn(dataType.getKey()));
				//
				// } else {
				// throw new IllegalStateException();
				// }
				//
				// for (Entry<RowIdx, AbstractObject> row :
				// data.getTableData().getColumn(dataType.getKey()).entrySet())
				// {
				//
				// if (newColumn.containsKey(row.getKey())) {
				// AbstractObject removed = replace.remove(row.getKey(),
				// dataType.getKey());
				// if (removed == null)
				// throw new IllegalStateException();
				// } else {
				// replace.put(row.getKey(), dataType.getKey(),
				// newColumn.get(row.getKey()));
				// }
				//
				// }
			}
		}

		if (replace == null) {
			return Optional.of(data);
		} else {
			return Optional.of(new AnnotatedTablesReaderData(data.getId(), data.getMetaData(), new AnnotatedTable(replace, data.getTableData().getStringTable(),
					newDataTypes, data.getTableData().getKeyColumnIndex(), data.getTableData().getColumnHeaders())));
		}
	}

}
