package de.uni_mannheim.informatik.dws.attribute.valueAdaption;

import java.util.Map.Entry;
import java.util.Optional;
import java.util.SortedMap;
import java.util.TreeMap;

import com.google.common.collect.HashBasedTable;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Table;

import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.attribute.matcher.AttributeMatcher;

public class StringPropertyConverter implements AnnotatedTableFilter {

	private final OntologyRepository ontology;

	public StringPropertyConverter(OntologyRepository ontology) {
		this.ontology = ontology;
	}

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {
		Table<RowIdx, ColumnIdx, AbstractObject> replace = null;
		SortedMap<ColumnIdx, ValueType> newDataTypes = null;

		int beginIndex = data.getMetaData().getNumHeaderRows();
		int endIndex = data.getMetaData().getNumberOfRows() + data.getMetaData().getNumHeaderRows();

		for (Entry<ColumnIdx, ValueType> dataType : data.getTableData().getDataTypes().entrySet()) {
			PropertyId propertyId = data.getMetaData().getProperties().get(dataType.getKey().getColumnIdx());
			if (propertyId != null && ontology.getDataType(propertyId) == ValueType.STRING && dataType.getValue() != ValueType.STRING) {

				if (replace == null) {
					replace = HashBasedTable.create(data.getTableData().getTable());
					newDataTypes = new TreeMap<>(data.getTableData().getDataTypes());
				}

				newDataTypes.put(dataType.getKey(), ValueType.STRING);

				ImmutableMap<RowIdx, AbstractObject> newColumn = AttributeMatcher
						.createStringColumn(data.getTableData().getStringTable().column(dataType.getKey()), beginIndex, endIndex);

				for (int i = beginIndex; i < endIndex; i++) {
					RowIdx rowIdx = new RowIdx(i);
					if (!newColumn.containsKey(rowIdx)) {
						replace.remove(rowIdx, dataType.getKey());
					} else {
						replace.put(rowIdx, dataType.getKey(), newColumn.get(rowIdx));
					}

				}

			}
		}

		if (replace == null) {
			return Optional.of(data);
		} else {
			return Optional.of(new AnnotatedTablesReaderData(data.getId(), data.getMetaData(), new AnnotatedTable(replace, data.getTableData().getStringTable(),
					newDataTypes, data.getTableData().getKeyColumnIndex(), data.getTableData().getColumnHeaders())));
		}
	}

}
