package de.uni_mannheim.informatik.dws.attribute.valueAdaption;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

import com.google.common.collect.HashBasedTable;

import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.PropertyIndexComposite;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.uni_mannheim.dws.expansion.data.ontology.model.measurement.UnitConverter;
import de.uni_mannheim.dws.expansion.data.ontology.model.measurement.UnitOfMeasurement;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.QuantityObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class WeightInPoundsTableConverter implements AnnotatedTableFilter {
	private final UnitConverter unitConverter = new UnitConverter();

	public static final PropertyId weight = new PropertyId("http://dbpedia.org/ontology/Person/weight");
	public static final PropertyId weightPounds = new PropertyId("http://dbpedia.org/ontology/Person/weightPounds");

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {

		if (data.getMetaData().getProperties().values().contains(WeightInPoundsTableConverter.weightPounds)) {

			Map<Integer, PropertyId> properties = data.getMetaData().getProperties();
			Map<Integer, PropertyId> newProperties = new HashMap<>();

			Map<PropertyIndexComposite, Double> propertiesScores = data.getMetaData().getAllPropertyScores();
			Map<PropertyIndexComposite, Double> newPropertiesScores = new HashMap<>();

			HashBasedTable<RowIdx, ColumnIdx, AbstractObject> table = HashBasedTable.create(data.getTableData().getTable());

			for (Entry<Integer, PropertyId> entry : properties.entrySet()) {
				PropertyIndexComposite composite = new PropertyIndexComposite(entry.getKey(), entry.getValue());
				if (entry.getValue().equals(WeightInPoundsTableConverter.weightPounds)) {
					ColumnIdx colIdx = new ColumnIdx(entry.getKey());

					for (Entry<RowIdx, AbstractObject> rowEntry : table.column(colIdx).entrySet()) {
						if (rowEntry.getValue().getType() == ValueType.QUANTITY) {
							double value = ((QuantityObject) rowEntry.getValue()).getValue();

							double convertedValue = unitConverter.convert(value, UnitOfMeasurement.WEIGHT_POUND, UnitOfMeasurement.WEIGHT_KG);

							rowEntry.setValue(QuantityObject.fromDouble(convertedValue));
						} else {
							int sdf = 13;
							throw new IllegalStateException();
						}
					}

					newProperties.put(entry.getKey(), WeightInPoundsTableConverter.weight);
					if (propertiesScores.containsKey(composite)) {
						PropertyIndexComposite compositeNew = new PropertyIndexComposite(entry.getKey(), WeightInPoundsTableConverter.weight);
						newPropertiesScores.put(compositeNew, propertiesScores.get(composite));
					}

				} else {
					newProperties.put(entry.getKey(), entry.getValue());
					if (propertiesScores.containsKey(composite)) {
						newPropertiesScores.put(composite, propertiesScores.get(composite));
					}
				}
			}

			AnnotatedTableMetaData newMetaData = data.getMetaData().rebuilder().setAllPropertyScores(newPropertiesScores).setProperties(newProperties).build();
			AnnotatedTable newTableData = new AnnotatedTable(table, data.getTableData().getStringTable(), data.getTableData().getDataTypes(),
					data.getTableData().getKeyColumnIndex(), data.getTableData().getColumnHeaders());
			return Optional.of(new AnnotatedTablesReaderData(data.getId(), newMetaData, newTableData));
		} else {
			return Optional.of(data);
		}

	}

}
