package de.uni_mannheim.informatik.dws.expansion.matching;

import java.io.IOException;
import java.text.Normalizer;
import java.util.HashSet;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.regex.Pattern;

import javax.inject.Inject;

import org.apache.lucene.document.Document;
import org.apache.lucene.index.IndexWriter;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.index.dbpedia.DBpediaIndexEntry;
import de.mannheim.uni.index.io.DefaultIndex;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;
import net.sf.junidecode.Junidecode;

public class IndexBuilder {
	@Inject
	public IndexBuilder() {

	}

	public DefaultIndex create(String indexdir, KnowledgeBase knowledgeBase) throws IOException {

		final DefaultIndex idx = new DefaultIndex(indexdir);

		IndexWriter writer = idx.getIndexWriter();

		for (ClassId setId : knowledgeBase.getClasses()) {
			for (Entry<InstanceId, Integer> instanceEntry : knowledgeBase.getInstances(setId).entrySet()) {
				DBpediaIndexEntry entry = new DBpediaIndexEntry();
				entry.setClass_label(setId.getClassId());
				entry.setUri(instanceEntry.getKey().getInstanceId());

				Set<String> labelsFinal = getLabels(instanceEntry.getKey(), knowledgeBase);
				int sdf = 123;
				for (String label : labelsFinal) {

					entry.setLabel(label);
					Document doc = entry.createDocument();
					writer.addDocument(doc);

				}

			}
		}

		idx.closeIndexWriter();

		return idx;
	}

	private Set<String> getLabels(InstanceId instanceId, KnowledgeBase knowledgeBase) {
		Set<String> labelsFinal = new HashSet<>();

		for (String label : getLabels(knowledgeBase, instanceId)) {

			String labelClean = StringNormalizer.normaliseValue(label, false);
			labelsFinal.add(labelClean);

			Optional<String> concatenated = IndexBuilder.concatenate(labelClean);
			if (concatenated.isPresent()) {
				labelsFinal.add(concatenated.get());

				Optional<String> deAccented = IndexBuilder.deAccent(concatenated.get());
				if (deAccented.isPresent()) {
					labelsFinal.add(deAccented.get());
				}
			}

			Optional<String> deAccented = IndexBuilder.deAccent(labelClean);
			if (deAccented.isPresent()) {
				labelsFinal.add(deAccented.get());
			}

			Optional<String> withCombinedLetters = lettersCombiner(labelClean);
			if (withCombinedLetters.isPresent()) {
				labelsFinal.add(withCombinedLetters.get());

				deAccented = IndexBuilder.deAccent(withCombinedLetters.get());
				if (deAccented.isPresent()) {
					labelsFinal.add(deAccented.get());
				}

			}

		}

		return labelsFinal;
	}

	public static Optional<String> deAccentOld(String str) {
		String nfdNormalizedString = Normalizer.normalize(str, Normalizer.Form.NFD);
		Pattern pattern = Pattern.compile("\\p{InCombiningDiacriticalMarks}+");
		String ret = pattern.matcher(nfdNormalizedString).replaceAll("");
		if (!str.equals(ret)) {
			return Optional.of(ret);
		} else
			return Optional.empty();
	}

	public static Optional<String> concatenate(String str) {
		String concatenated = str.replace(" ", "");
		if (str.equals(concatenated)) {
			return Optional.empty();
		} else {
			return Optional.of(concatenated);
		}

	}

	public static Optional<String> deAccent(String str) {
		String ret = Junidecode.unidecode(str);
		if (!str.equals(ret)) {
			return Optional.of(ret);
		} else
			return Optional.empty();
	}

	public static Optional<String> lettersCombiner(String labelClean) {

		if (labelClean.equals("")) {
			return Optional.empty();
		}

		StringBuilder sb = new StringBuilder();

		boolean previousIsSpace = true;
		int previousIsSpaceExtra = 0;
		int start = -1;
		for (int i = 0; i < labelClean.length(); i++) {
			char c = labelClean.charAt(i);
			if (start == -1 && c != ' ' && previousIsSpace) {
				// we append previous, as it was not part of letterCombination
				if (i > 0)
					sb.append(labelClean.charAt(i - 1));
				start = i;
				previousIsSpace = false;
			} else if (start != -1 && c != ' ' && previousIsSpace) {
				// we do not append previous, because it was a space and this is
				// a candidate, so we continue
				previousIsSpace = false;
			} else if (start != -1 && c == ' ' && !previousIsSpace) {
				for (int s = start; s <= i; s++) {
					char c2 = labelClean.charAt(s);
					if (c2 != ' ')
						sb.append(c2);
				}
				start = i;
				previousIsSpace = true;
			} else if (start != -1 && c == ' ' && previousIsSpace) {

			} else if (start != -1 && c != ' ' && !previousIsSpace) {
				for (int s = start; s <= i; s++) {
					char c2 = labelClean.charAt(s);
					sb.append(c2);
				}
				start = -1;
			} else if (start == -1 && c != ' ' && !previousIsSpace) {
				sb.append(labelClean.charAt(i));
			} else if (start == -1 && c == ' ' && !previousIsSpace) {
				// sb.append(labelClean.charAt(i));
				previousIsSpace = true;
			} else if (start == -1 && c == ' ' && previousIsSpace) {
				if (i > 0)
					sb.append(' ');
				previousIsSpaceExtra++;
			} else {
				int sdf = 123;
			}
			// Process char
		}
		if (start != -1) {
			for (int s = start; s < labelClean.length(); s++) {
				char c = labelClean.charAt(s);
				if (c != ' ' || s >= labelClean.length() - 1)
					sb.append(c);
			}
		} else if (previousIsSpace && labelClean.length() > 0) {
			sb.append(labelClean.charAt(labelClean.length() - 1));
		}

		String s = sb.toString();
		if (s.equals(labelClean)) {
			return Optional.empty();
		} else {
			return Optional.of(sb.toString());
		}

	}

	private Set<String> getLabels(KnowledgeBase knowledgeBase, InstanceId instanceId) {
		QualifiedValueList qvl = knowledgeBase.getQualifiedValueList(new ValueId(instanceId, PropertiesExpansionData.LABEL_PROPERTY));
		Set<String> ret = new HashSet<>();
		if (qvl != null) {
			for (QualifiedValueHolder claim : qvl.getClaims()) {
				if (claim.getValue() instanceof StringObject) {
					ret.add(((StringObject) claim.getValue()).getString());
				}
			}
		}

		qvl = knowledgeBase.getQualifiedValueList(new ValueId(instanceId, PropertiesExpansionData.TRANSLATED_LABELS_PROPERTY));

		if (qvl != null) {
			for (QualifiedValueHolder claim : qvl.getClaims()) {
				if (claim.getValue() instanceof StringObject) {
					ret.add(((StringObject) claim.getValue()).getString());
				}
			}
		}

		qvl = knowledgeBase.getQualifiedValueList(new ValueId(instanceId, PropertiesExpansionData.TRANSLATED_ALIASES_PROPERTY));

		if (qvl != null) {
			for (QualifiedValueHolder claim : qvl.getClaims()) {
				if (claim.getValue() instanceof StringObject) {
					ret.add(((StringObject) claim.getValue()).getString());
				}
			}
		}
		return ret;
	}

}
