package de.uni_mannheim.informatik.dws.expansion.matching;

import java.util.List;
import java.util.Set;

import dagger.ObjectGraph;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.impl.reference.MatchingDataBuilder;
import de.mannheim.uni.matching.dbpedia.algorithm.KeyIndex;
import de.mannheim.uni.matching.dbpedia.model.MatchingData;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;

public class KnowledgeBaseAdapter {

	private final ObjectGraph objectGraph;

	public KnowledgeBaseAdapter(String pathToKnowledgeBase, String pathToOntology, String indexCachePath) {
		this(pathToKnowledgeBase, pathToOntology, indexCachePath, null);

	}

	public KnowledgeBaseAdapter(String pathToKnowledgeBase, String pathToOntology, String indexCachePath, List<KBValueCreator> valueCreatorList) {
		KnowledgeBaseModule knowledgeBaseModule = new KnowledgeBaseModule(pathToKnowledgeBase, pathToOntology, indexCachePath);
		if (valueCreatorList != null) {
			knowledgeBaseModule.setValueCreatorList(valueCreatorList);
		}
		objectGraph = ObjectGraph.create(knowledgeBaseModule);
	}

	public MatchingData getMatchingData() {

		MatchingData matchingData = objectGraph.get(MatchingDataBuilder.class).getMatchingData();

		return matchingData;

	}

	public MatchingData getMatchingData(Set<ClassId> allowedClasses) {

		MatchingData matchingData = objectGraph.get(MatchingDataBuilder.class).getMatchingDataRanged(allowedClasses);

		return matchingData;

	}

	public KeyIndex getLuceneIndex() {

		KeyIndex keyIndex = objectGraph.get(KeyIndex.class);

		return keyIndex;

	}

	public KnowledgeBase getKnowledgeBase() {
		return objectGraph.get(KnowledgeBase.class);
	}

	public OntologyRepository getOntologyRepository() {
		return objectGraph.get(OntologyRepository.class);
	}

}
