package de.uni_mannheim.informatik.dws.expansion.matching.testing;

import java.io.File;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.UUID;

import dagger.ObjectGraph;
import de.mannheim.uni.expansion.data.dbpedia.PropertiesDBpediaTableFilter;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.informatik.dws.expansion.matching.KnowledgeBaseModule;
import de.uni_mannheim.informatik.dws.expansion.matching.testing.cases.DefaultTestCases;

public class CreateTestingTables {
	public static void main(String[] args) {

		String dataHome = null;

		if (args.length > 0 && args[0] != null && !args[0].equals("")) {
			dataHome = args[0];
		}

		String kbName = null;

		if (args.length > 1 && args[1] != null && !args[1].equals("")) {
			kbName = args[1];
		}

		String testName = null;

		if (args.length > 2 && args[2] != null && !args[2].equals("")) {
			testName = args[2];
		}

		String kbHome = dataHome + "/kbs/" + kbName + "/";
		String kbPath = kbHome + "/kb/";
		String ontologyRepositoryFile = kbHome + "/ontology.kryo";
		String luceneIndexCache = kbHome + "/lucene/";

		String outputFolder = dataHome + "/tests/";

		KnowledgeBaseModule kbModule = new KnowledgeBaseModule(kbPath, ontologyRepositoryFile, luceneIndexCache);

		List<KBValueCreator> valueCreatorList = new ArrayList<>();
		valueCreatorList.add(new PropertiesDBpediaTableFilter());
		kbModule.setValueCreatorList(valueCreatorList);

		ObjectGraph graph = ObjectGraph.create(kbModule);

		KnowledgeBase kb = graph.get(KnowledgeBase.class);

		OntologyRepository ontologyRepository = graph.get(OntologyRepository.class);

		if (testName == null) {
			testName = UUID.randomUUID().toString();
		}
		run(testName, outputFolder, kb, ontologyRepository);

		try {
			PersistentParallel4.close();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public static void run(String uuid, String outputFolder, KnowledgeBase kb, OntologyRepository ontologyRepository) {
		TestingFramework testingFramework = new TestingFramework(DefaultTestCases.getDefaultCases(), kb, ontologyRepository);

		MKDIR.createDirBoolean(new File(outputFolder));
		MKDIR.createDirBoolean(new File(outputFolder + uuid + "/"));
		MKDIR.createDirBoolean(new File(outputFolder + uuid + "/" + "input/"));
		MKDIR.createDirBoolean(new File(outputFolder + uuid + "/" + "output/"));

		Set<ClassId> sets = new HashSet<>();
		// sets.remove(new ClassId("OfficeHolder"));
		sets.add(new ClassId("http://dbpedia.org/ontology/Settlement"));

		testingFramework.createTestCases(sets, outputFolder + uuid + "/" + "input/");

		System.out.println();
	}

}
