package de.uni_mannheim.informatik.dws.expansion.matching.testing.cases.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ThreadLocalRandom;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.tableprocessor.model.TableMapping;
import de.mannheim.uni.tableprocessor.model.json.AnnotatedTable;
import de.mannheim.uni.tableprocessor.model.json.TableData;
import de.mannheim.uni.tableprocessor.model.json.TableData.HeaderPosition;
import de.mannheim.uni.tableprocessor.model.json.TableData.TableOrientation;
import de.mannheim.uni.tableprocessor.model.json.TableData.TableType;
import de.mannheim.uni.utils.data.Pair;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.informatik.dws.expansion.matching.testing.cases.MatchingTestCaseInterface;

public class MatchingTestCaseDefault implements MatchingTestCaseInterface {

	@Override
	public List<AnnotatedTable> create(KnowledgeBase knowledgeBase, OntologyRepository ontologyRepository, ClassId setId) {

		List<AnnotatedTable> ret = new ArrayList<>();

		ret.addAll(createClassNonTimeDependentTables(knowledgeBase, ontologyRepository, setId));
		return ret;
	}

	public List<AnnotatedTable> createClassNonTimeDependentTables(KnowledgeBase knowledgeBase, OntologyRepository ontologyRepository, ClassId setId) {

		List<AnnotatedTable> ret = new ArrayList<>();

		List<PropertyId> properties = knowledgeBase.getProperties(setId).keySet().stream().filter((e) -> {
			return !ontologyRepository.isTimeDependent(e);
		}).collect(Collectors.toList());

		int maxSize = knowledgeBase.getNumberOfInstances(setId);
		if (maxSize >= 5)
			ret.addAll(createTableOfDifferentSizes(knowledgeBase, setId, properties, 5, 5));
		if (maxSize >= 10)
			ret.addAll(createTableOfDifferentSizes(knowledgeBase, setId, properties, 10, 5));
		if (maxSize >= 50)
			ret.addAll(createTableOfDifferentSizes(knowledgeBase, setId, properties, 50, 5));
		if (maxSize >= 100)
			ret.addAll(createTableOfDifferentSizes(knowledgeBase, setId, properties, 100, 5));

		return ret;
	}

	private List<AnnotatedTable> createTableOfDifferentSizes(KnowledgeBase knowledgeBase, ClassId setId, List<PropertyId> propertiesOriginal, int size,
			int count) {
		List<AnnotatedTable> ret = new ArrayList<>();

		for (int c = 0; c < count; c++) {

			List<PropertyId> properties = new ArrayList<>(propertiesOriginal);
			properties.remove(PropertiesExpansionData.LABEL_PROPERTY);
			Collections.shuffle(properties);

			properties = properties.subList(0, Math.min(properties.size(), ThreadLocalRandom.current().nextInt(3, 6 + 1)));

			properties.add(PropertiesExpansionData.LABEL_PROPERTY);
			Collections.shuffle(properties);

			AnnotatedTable annotatedTable = new AnnotatedTable();
			TableData tableData = new TableData();
			TableMapping tableMapping = new TableMapping();
			HashMap<Integer, Pair<String, Double>> mappedProperties = new HashMap<>();
			HashMap<Integer, Pair<String, Double>> mappingInstances = new HashMap<>();

			tableData.setHasHeader(true);
			tableData.setTableType(TableType.RELATION);
			tableData.setHeaderPosition(HeaderPosition.FIRST_ROW);
			tableData.setTableOrientation(TableOrientation.HORIZONTAL);
			tableData.setUrl("http://dws.informatik.uni-mannheim.de/" + setId.getClassId() + "_" + size + "_" + c);
			tableData.setTitle(setId.getClassId() + "_" + size + "_" + c);
			tableData.setPageTitle(setId.getClassId());
			tableData.setTableNum(0);
			tableData.setRecordEndOffset(0);
			tableData.setRecordOffset(0);
			tableData.setHasKeyColumn(true);
			tableData.setHeaderRowIndex(0);
			tableData.setTextAfterTable("");
			tableData.setTextBeforeTable("");

			List<String[]> relationColumns = new ArrayList<>();

			// int columnSize = properties.size();

			List<InstanceId> instanceIds = new ArrayList<>();

			for (InstanceId instanceId : knowledgeBase.getInstances(setId).keySet()) {
				boolean add = true;
				for (PropertyId propertyId : properties) {
					if (!knowledgeBase.hasValue(new ValueId(instanceId, propertyId))) {
						add = false;
						break;
					}
				}
				if (add) {
					instanceIds.add(instanceId);
				}
			}
			if (instanceIds.isEmpty()) {
				c--;
				continue;
			}
			Collections.shuffle(instanceIds);
			instanceIds = instanceIds.subList(0, Math.min(size, instanceIds.size()));

			int rowIdx = 1;
			for (InstanceId instanceId : instanceIds) {
				mappingInstances.put(rowIdx, new Pair<>(instanceId.toCombinedString(), 1.0));
				rowIdx++;
			}

			{
				int columnIndex = 0;
				for (PropertyId propertyId : properties) {
					mappedProperties.put(columnIndex, new Pair<>(propertyId.toCombinedString(), 1.0));
					if (PropertiesExpansionData.LABEL_PROPERTY.equals(propertyId)) {
						tableData.setKeyColumnIndex(columnIndex);
					}
					String[] column = new String[instanceIds.size() + 1];
					column[0] = propertyId.getFinalPart();

					int i = 1;
					for (InstanceId instanceId : instanceIds) {

						List<AbstractObject> list = knowledgeBase.getValuesAsList(instanceId, propertyId);
						if (list != null) {
							if (list.size() == 1) {

								column[i] = list.get(0).readableString();
							} else if (list.size() > 1) {
								System.out.println("THIS A NON-TIME-DEPENDENT PROPERTY WITH MULTIPLE VALUES");
								column[i] = list.get(0).readableString();
							} else {
								column[i] = "";
							}
						} else {
							column[i] = "";
						}
						if (column[i] == null) {
							int sdf = 123;
						}

						i++;
					}

					relationColumns.add(column);
					columnIndex++;
				}

			}
			tableMapping.setMappedInstances(mappingInstances);
			tableMapping.setMappedProperties(mappedProperties);
			tableMapping.setMappedClass(new Pair<>(setId.getClassId(), 1.0));
			tableData.setRelation(relationColumns.toArray(new String[relationColumns.size()][]));

			annotatedTable.setMapping(tableMapping);
			annotatedTable.setTable(tableData);

			ret.add(annotatedTable);

		}
		return ret;
	}

	@Override
	public Map<String, List<Double>> evaluate(KnowledgeBase knowledgeBase, OntologyRepository ontologyRepository,
			List<de.mannheim.uni.expansion.utils.containers.Pair<AnnotatedTable, AnnotatedTable>> testAndGeneratedTables) {

		Map<String, List<Double>> ret = new HashMap<>();
		int size = testAndGeneratedTables.size();

		for (int i = 0; i < testAndGeneratedTables.size(); i++) {

			AnnotatedTable testTable = testAndGeneratedTables.get(i).getFirst();
			AnnotatedTable generatedTable = testAndGeneratedTables.get(i).getSecond();

			{

				List<Double> retListClass = CHMHelper.getFromMap(ret, "CLASS::" + testTable.getMapping().getMappedClass().getFirst(), () -> createList(size));
				if (testTable.getMapping().getMappedClass().getFirst().equals(generatedTable.getMapping().getMappedClass().getFirst())) {
					retListClass.set(i, 1.0);
				}

			}

			for (Entry<Integer, de.mannheim.uni.utils.data.Pair<String, Double>> propertyEntry : testTable.getMapping().getMappedProperties().entrySet()) {
				List<Double> retList = CHMHelper.getFromMap(ret, "PROPERTY::" + propertyEntry.getValue().getFirst(), () -> createList(size));
				if (generatedTable.getMapping().getMappedProperties().containsKey(propertyEntry.getKey()) && propertyEntry.getValue().getFirst()
						.equals(generatedTable.getMapping().getMappedProperties().get(propertyEntry.getKey()).getFirst())) {
					retList.set(i, 1.0);
				} else {
					retList.set(i, 0.0);
				}

			}

			double total = 0.0;
			double correct = 0.0;
			for (Entry<Integer, de.mannheim.uni.utils.data.Pair<String, Double>> instanceEntry : testTable.getMapping().getMappedInstances().entrySet()) {
				total++;

				if (generatedTable.getMapping().getMappedInstances().containsKey(instanceEntry.getKey()) && instanceEntry.getValue().getFirst()
						.equals(generatedTable.getMapping().getMappedInstances().get(instanceEntry.getKey()).getFirst())) {
					correct++;
				}
			}

			CHMHelper.getFromMap(ret, "INSTANCES", () -> createList(size)).set(i, correct / total);
		}

		return ret;
	}

	private List<Double> createList(int size) {
		List<Double> ret = new ArrayList<>();
		for (int i = 0; i < size; i++) {
			ret.add(null);
		}
		return ret;
	}

}
