package de.uni_mannheim.informatik.dws.matching;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Consumer;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import org.apache.commons.compress.archivers.tar.TarArchiveEntry;
import org.apache.commons.compress.archivers.tar.TarArchiveInputStream;
import org.apache.commons.compress.archivers.tar.TarArchiveOutputStream;
import org.apache.commons.compress.compressors.gzip.GzipCompressorInputStream;
import org.apache.commons.io.IOUtils;

import com.google.gson.Gson;

import de.mannheim.uni.expansion.data.dbpedia.PropertiesDBpediaTableFilter;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.utils.BatchConsumerProducer;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.index.io.DefaultIndex;
import de.mannheim.uni.matching.dbpedia.Preprocessing;
import de.mannheim.uni.matching.dbpedia.algorithm.InstanceMatchingTask;
import de.mannheim.uni.matching.dbpedia.algorithm.KeyIndex;
import de.mannheim.uni.matching.dbpedia.algorithm.WebTableToDBpediaComponentProcess;
import de.mannheim.uni.matching.dbpedia.algorithm.WebtableToDBpediaMatchingProcess;
import de.mannheim.uni.matching.dbpedia.model.EvaluationParameters;
import de.mannheim.uni.matching.dbpedia.model.MatchingData;
import de.mannheim.uni.matching.dbpedia.model.MatchingParameters;
import de.mannheim.uni.matching.dbpedia.model.MatchingResult;
import de.mannheim.uni.matching.dbpedia.model.TableRow;
import de.mannheim.uni.matching.dbpedia.model.adapters.TableColumnToCellHierarchyAdapter;
import de.mannheim.uni.matching.dbpedia.model.adapters.TableRowToCellHierarchyAdapter;
import de.mannheim.uni.matching.dbpedia.model.adapters.TableToRowHierarchyAdapter;
import de.mannheim.uni.matching.process.Configuration;
import de.mannheim.uni.similarity.functions.QuadraticSimilarityDecorator;
import de.mannheim.uni.similarity.functions.date.WeightedDatePartSimilarity;
import de.mannheim.uni.similarity.functions.numeric.DeviationSimilarity;
import de.mannheim.uni.similarity.functions.numeric.NormalisedNumericSimilarity;
import de.mannheim.uni.similarity.functions.string.GeneralisedStringJaccard;
import de.mannheim.uni.similarity.functions.string.JaccardOnNGramsSimilarity;
import de.mannheim.uni.similarity.functions.string.LevenshteinSimilarity;
import de.mannheim.uni.tableprocessor.IO.TableReader;
import de.mannheim.uni.tableprocessor.model.Table;
import de.mannheim.uni.tableprocessor.model.TableColumn;
import de.mannheim.uni.tableprocessor.model.json.AnnotatedTable;
import de.mannheim.uni.tableprocessor.model.json.TableData;
import de.mannheim.uni.units.UnitParser;
import de.mannheim.uni.utils.concurrent.Parallel;
import de.mannheim.uni.utils.timer.Timer;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.informatik.dws.expansion.matching.KnowledgeBaseAdapter;

/**
 *
 * @author domi
 */
public class MatchFull {
	public static enum JsonType {
		AnnotatedTable, TableData;

	}

	public static void main(String[] args) throws IOException {
		TableCacheSetter.setCaches();

		String dataHomeTemp = null;

		if (args.length > 0 && args[0] != null && !args[0].equals("")) {
			dataHomeTemp = args[0];
		}

		String kbName = null;

		if (args.length > 1 && args[1] != null && !args[1].equals("")) {
			kbName = args[1];
		}

		String mappingsName = null;

		if (args.length > 2 && args[2] != null && !args[2].equals("")) {
			mappingsName = args[2];
		}
		String dataHome = dataHomeTemp;

		String mappingsTargetName = null;

		if (args.length > 3 && args[3] != null && !args[3].equals("")) {
			mappingsTargetName = args[3];
		}

		String kbHome = dataHome + "/kbs/" + kbName + "/";
		String kbPath = kbHome + "/kb/";
		String ontologyRepositoryFile = kbHome + "/ontology.kryo";
		String luceneIndexCache = kbHome + "/lucene/";

		String tablePath = dataHome + "/tables/" + mappingsName + "/mappings/";
		String outputFolder = dataHome + "/tables/" + mappingsTargetName + "/mappings/";
		String config = dataHome + "/conf.config";
		String emptyFile = dataHome + "/empty.txt";
		// String unitPath = dataHome + "/empty.txt";

		UnitParser.unitPath = dataHome + "/Units/Convertible";

		UnitParser.readInUnits();

		JsonType jsonType = JsonType.AnnotatedTable;

		List<KBValueCreator> valueCreatorList = new ArrayList<>();
		valueCreatorList.add(new PropertiesDBpediaTableFilter());

		KnowledgeBaseAdapter adapter = new KnowledgeBaseAdapter(kbPath, ontologyRepositoryFile, luceneIndexCache, valueCreatorList);

		// KnowledgeBaseAdapter adapter = new KnowledgeBaseAdapter(kbPath,
		// ontologyRepositoryFile, luceneIndexCache);

		Set<ClassId> allowedClasses = new HashSet<>();
		KnowledgeBase kb = adapter.getKnowledgeBase();
		OntologyRepository ontologyRepository = adapter.getOntologyRepository();

		for (ClassId setId : kb.getClasses()) {
			if (kb.getProperties(setId).size() > 1) {
				allowedClasses.add(setId);
			}
		}

		// final ClassId placeClass = new
		// ClassId("http://dbpedia.org/ontology/Settlement");
		//
		// allowedClasses = new HashSet<>();
		// allowedClasses.add(placeClass);
		// allowedClasses.addAll(ontologyRepository.getSubClasses(placeClass));

		MatchingData matchingData = adapter.getMatchingData(allowedClasses);

		matchingData.getDbpediaRowSet();
		matchingData.getDbpediaColSet();

		TableToRowHierarchyAdapter tableToRow = new TableToRowHierarchyAdapter();
		TableColumnToCellHierarchyAdapter columnToCell = new TableColumnToCellHierarchyAdapter();
		TableRowToCellHierarchyAdapter rowToCell = new TableRowToCellHierarchyAdapter();

		PersistentParallel6.runAndFinish(matchingData.getDbpediaTables(), (table) -> {
			tableToRow.getParts(table);
		}, MatchFull.class);

		PersistentParallel6.runAndFinish(matchingData.getDbpediaRowSet(), (row) -> {
			rowToCell.getParts(row);
		}, MatchFull.class);

		PersistentParallel6.runAndFinish(matchingData.getDbpediaColSet(), (column) -> {
			columnToCell.getParts(column);
		}, MatchFull.class);

		KeyIndex index = adapter.getLuceneIndex();
		MatchFull full = new MatchFull(matchingData, index, luceneIndexCache, config, emptyFile, jsonType);
		full.run(tablePath, outputFolder);

		try {
			PersistentParallel6.close();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		// System.exit(0);

	}

	TableReader r;
	TableToRowHierarchyAdapter rowAdapter;
	Configuration conf;
	final InstanceMatchingTask m;
	MatchingData matchingData;
	KeyIndex index;
	MatchingParameters params;
	EvaluationParameters eval;
	MatchingParameters loadDbpParams = new MatchingParameters();
	Timer mainTimer = new Timer("Matching");
	String version = "";
	JsonType jsonType;

	TableToRowHierarchyAdapter tableToRow = new TableToRowHierarchyAdapter();
	TableColumnToCellHierarchyAdapter columnToCell = new TableColumnToCellHierarchyAdapter();
	TableRowToCellHierarchyAdapter rowToCell = new TableRowToCellHierarchyAdapter();

	public MatchFull(MatchingData matchingData, KeyIndex index, String luceneIndexCache, String config, String emptyFile, JsonType jsonType)
			throws FileNotFoundException, IOException {
		this.matchingData = matchingData;
		this.index = index;
		this.jsonType = jsonType;

		m = new InstanceMatchingTask();

		params = new MatchingParameters();
		eval = new EvaluationParameters();

		m.setMatchingParameters(loadDbpParams);
		m.setData(matchingData.shallowCopy());

		if (new File(luceneIndexCache).exists()) {
			m.setLuceneIndex(new DefaultIndex(luceneIndexCache));
		}
		Preprocessing.loadSurfaceForms(emptyFile);

		Preprocessing.loadRedirects(emptyFile);
		m.getData().setEquivalentProperties(new HashMap<>());

		m.setMatchingParameters(params);

		Map<String, Object> aliases = new HashMap<>();
		aliases.put("QuadraticDeviation", new QuadraticSimilarityDecorator<Double>(new DeviationSimilarity()));
		aliases.put("QuadraticNormalized", new QuadraticSimilarityDecorator<Double>(new NormalisedNumericSimilarity()));
		aliases.put("GeneralizedLevenshtein", new GeneralisedStringJaccard(new LevenshteinSimilarity(), 0.5, 0.5));
		aliases.put("Jaccard2Grams", new JaccardOnNGramsSimilarity(2));
		aliases.put("Jaccard3Grams", new JaccardOnNGramsSimilarity(3));
		aliases.put("Jaccard4Grams", new JaccardOnNGramsSimilarity(4));
		aliases.put("Jaccard5Grams", new JaccardOnNGramsSimilarity(5));
		aliases.put("WeightedDate", new WeightedDatePartSimilarity(1, 3, 5));
		aliases.put("lodtable", WebtableToDBpediaMatchingProcess.tableType.lodtable);
		aliases.put("webtable", WebtableToDBpediaMatchingProcess.tableType.webtable);
		aliases.put("jsonWebTable", WebtableToDBpediaMatchingProcess.tableType.jsonWebTable);

		try {
			conf = Configuration.readConfiguration(config, m.getParameters(), aliases);
		} catch (ClassNotFoundException | InstantiationException | IllegalAccessException e1) {
			e1.printStackTrace();
			return;
		}

		if (((double) conf.getValue(InstanceMatchingTask.PAR_ABSTRACT_WEIGHT) > 0.0)) {
			m.getData().initializeAbstracts();
		}

		System.out.println("wikidata classes: " + m.getData().getDbpediaTables().size());
		System.out.println("wikidata cols: " + m.getData().getDbpediaColSet().size());
		System.out.println("wikidata rows: " + m.getData().getDbpediaRowSet().size());

		Integer numThreads = (Integer) conf.getValue(WebtableToDBpediaMatchingProcess.PAR_MAX_PARALLEL);
		if (numThreads > 0) {
			Parallel.SetDefaultNumProcessors(numThreads);
		}

		loadDbpParams.setMaxSpanningCells((Integer) conf.getValue(WebTableToDBpediaComponentProcess.PAR_SPANNING_CELL_THRESHOLD));
		loadDbpParams.setTableType((de.mannheim.uni.matching.dbpedia.algorithm.WebtableToDBpediaMatchingProcess.tableType) conf
				.getValue(WebTableToDBpediaComponentProcess.PAR_TABLE_TYPE));

		r = new TableReader();
		r.setUseUnitDetection(loadDbpParams.isUseUnitDetection());
		r.setSpanningCellThreshold(loadDbpParams.getMaxSpanningCells());

		rowAdapter = new TableToRowHierarchyAdapter();

	}

	public void run(String tablesPath, String outputFolder) {
		try {

			setSystemOut();

			BatchConsumerProducer<Pair<String, List<Pair<byte[], String>>>> batch = BatchConsumerProducer.create("BatchMatching", (pair) -> {
				try {
					runGzFile(pair.getSecond(), outputFolder, pair.getFirst());
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			});

			File folder = new File(tablesPath);
			File[] listOfFiles = folder.listFiles();

			PersistentParallel6.runAndFinish(Arrays.asList(listOfFiles), (file) -> {
				String name = file.getName();
				List<Pair<byte[], String>> files = getFiles(file);
				List<Pair<String, List<Pair<byte[], String>>>> taskList = new ArrayList<>();
				taskList.add(new Pair<>(name, files));
				batch.addBatch(taskList);
			}, this.getClass(), 6);

			batch.finish();

			// @SuppressWarnings("unchecked")
			// List<Pair<byte[], String>>[] loadedStuff = new
			// List[listOfFiles.length];
			//
			// CountDownLatch latch = loadStuff(0, loadedStuff, listOfFiles);
			//
			// for (int i = 0; i < listOfFiles.length; i++) {
			//
			// latch.await();
			// if ((i + 1) < listOfFiles.length)
			// latch = loadStuff(i + 1, loadedStuff, listOfFiles);
			//
			//
			//
			// runGzFile(files, outputFolder, name);
			//
			// }

		} catch (Exception e) {
			e.printStackTrace();

		}

		int ddfdf = 123;
	}

	public void runOld(String tablesPath, String outputFolder) {
		try {

			setSystemOut();

			File folder = new File(tablesPath);
			File[] listOfFiles = folder.listFiles();
			@SuppressWarnings("unchecked")
			List<Pair<byte[], String>>[] loadedStuff = new List[listOfFiles.length];

			CountDownLatch latch = loadStuff(0, loadedStuff, listOfFiles);

			for (int i = 0; i < listOfFiles.length; i++) {

				latch.await();
				if ((i + 1) < listOfFiles.length)
					latch = loadStuff(i + 1, loadedStuff, listOfFiles);

				String name = listOfFiles[i].getName();
				List<Pair<byte[], String>> files = loadedStuff[i];

				runGzFile(files, outputFolder, name);

			}

		} catch (Exception e) {
			e.printStackTrace();

		}

		int ddfdf = 123;
	}

	static void setSystemOut() {
		System.setOut(new DummyPrintStream());

	}

	private CountDownLatch loadStuff(int i, List<Pair<byte[], String>>[] loadedStuff, File[] listOfFiles) {
		CountDownLatch latch = new CountDownLatch(1);

		Runnable r = () -> {
			loadedStuff[i] = getFiles(listOfFiles[i]);
			latch.countDown();
		};

		Thread t = new Thread(r);
		t.start();

		return latch;
	}

	private void runGzFile(List<Pair<byte[], String>> files, String outputFolder, String name) throws IOException {
		ConcurrentLinkedQueue<Pair<byte[], String>> resultQueue = new ConcurrentLinkedQueue<>();

		final Configuration configFinal = conf;
		final AtomicInteger count = new AtomicInteger();
		try {
			// new
			// Parallel<File>().foreach(Arrays.asList(new
			// File(web).listFiles()), new Consumer<File>()
			// {
			PersistentParallel6.runAndFinish(files, new Consumer<Pair<byte[], String>>() {
				@Override
				public void accept(Pair<byte[], String> file) {
					// if(!matchedTables.contains(parameter.getName()))
					// {
					try {
						WebTableToDBpediaComponentProcess p = new WebTableToDBpediaComponentProcess();
						p.setMatchingParameters(loadDbpParams);
						p.setData(m.getData().shallowCopy());
						MatchingData data = p.getData();
						// String json = IOUtils.toString(file.getFirst());
						InputStreamReader reader = new InputStreamReader(new ByteArrayInputStream(file.getFirst()));

						// AnnotatedTable annotatedTable = gson.fromJson(reader,
						// AnnotatedTable.class);
						TableData tableData = null;
						if (jsonType == JsonType.AnnotatedTable) {
							AnnotatedTable annotatedTable = gson.fromJson(reader, AnnotatedTable.class);

							tableData = annotatedTable.getTable();

						} else {
							tableData = gson.fromJson(reader, TableData.class);
						}
						reader.close();
						// TableData tableData = annotatedTable.getTable();
						Table table = r.readWebTableFromJson(tableData, file.getSecond());

						Collection<Table> webTables = new ArrayList<Table>();
						webTables.add(table);
						p.setWebtableName(file.getSecond());
						data.setWebTables(webTables);
						data.setWebtable(table);
						Collection<TableRow> webTableRowSet = rowAdapter.getParts(table);

						for (TableColumn tc : table.getColumns()) {
							columnToCell.getParts(tc);
						}

						for (TableRow row : webTableRowSet) {
							rowToCell.getParts(row);
						}

						data.setWebtableRowSet(webTableRowSet);
						p.setMatchingParameters(params);
						p.setEvaluationParameters(eval);
						p.setLuceneIndex(index.getLuceneIndex());

						Configuration c = configFinal.clone();
						c.getConfig().put(WebTableToDBpediaComponentProcess.PAR_WEBTABLE, WebTableToDBpediaComponentProcess.MANUAL_TABLE_LOAD);
						p.run(c);
						MatchingResult r = p.getResult();
						if (r.getClassMappings() != null && !r.getClassMappings().isEmpty()) {
							AnnotatedTable result = r.getJsonString(tableData);

							ByteArrayOutputStream baos = new ByteArrayOutputStream();
							OutputStreamWriter writer = new OutputStreamWriter(baos);
							writer.write(gson.toJson(result));
							writer.close();
							resultQueue.add(new Pair<>(baos.toByteArray(), file.getSecond()));
						}
						// r.write(WebtableToDBpediaMatchingProcess.tableType.jsonWebTable,
						// outputFolder);

						int currentCount = count.incrementAndGet();
						if (currentCount % 5000 == 0) {
							System.err.println(
									"Finished " + count + " out of " + files.size() + " (" + ((100 * currentCount) / files.size()) + " % ) of File: " + name);
						}

					} catch (Exception e) {
						System.err.println(String.format("Matching Table %s failed!", file.getSecond()));
						e.printStackTrace();
						int sdf = 123;
					}
				}
			}, "MatchFull");
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		writeFiles(resultQueue, outputFolder, name);

	}

	public static Gson gson = new Gson();

	public static byte[] decompress(byte[] contentBytes) {
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		try {
			IOUtils.copy(new GZIPInputStream(new ByteArrayInputStream(contentBytes)), out);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
		return out.toByteArray();
	}

	private static List<Pair<byte[], String>> getFiles(File gzFile) {

		TarArchiveInputStream tarStream = null;
		try {
			List<Pair<byte[], String>> entryList = new ArrayList<>();
			tarStream = new TarArchiveInputStream(new GzipCompressorInputStream(new FileInputStream(gzFile)));

			TarArchiveEntry entry;

			while ((entry = tarStream.getNextTarEntry()) != null) {

				if (!entry.isDirectory()) {

					byte[] content = new byte[(int) entry.getSize()];

					tarStream.read(content, 0, content.length);

					String a = new String(content);
					//
					// int numLines = countLines(content) - 9;

					String[] nameParts = entry.getName().split("/");

					String name = nameParts[nameParts.length - 1];
					if (name.endsWith(".gz")) {
						BufferedReader reader = new BufferedReader(new InputStreamReader(new GZIPInputStream(new ByteArrayInputStream(content))));
						try {
							int i = 0;
							String line;
							while ((line = reader.readLine()) != null) {

								entryList.add(new Pair<byte[], String>(line.getBytes(), name + "." + i));
								i++;
							}
						} finally {
							reader.close();
						}

					} else {
						entryList.add(new Pair<byte[], String>(content, name));
					}

					// System.out.println(name);

				}
			}

			return entryList;
			// Table[]
			// Array<File> ret = new ConcurrentLinkedQueue();
			// PersistentParallel4.runAndFinish(entryList, body, "JSONIZING");
			//
			// t.setFullPath(STRING);
			//
			// consumer.addBatch(entryList);

		} catch (Exception e) {
			e.printStackTrace();
			return null;
		} finally {
			if (tarStream != null)
				try {
					tarStream.close();
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
		}
	}

	void writeFiles(Collection<Pair<byte[], String>> files, String outputFolder, String gzFileName) throws IOException {

		Runnable r = () -> {
			try {
				TarArchiveOutputStream tarFile = new TarArchiveOutputStream(new GZIPOutputStream(new FileOutputStream(outputFolder + "/" + gzFileName)));

				// TarArchiveEntry mappingsFolder = new TarArchiveEntry
				// tarFile.putArchiveEntry(mappingsFolder);
				// tarFile.closeArchiveEntry();

				for (Pair<byte[], String> file : files) {
					TarArchiveEntry entry = new TarArchiveEntry("mappings/" + file.getSecond());
					entry.setSize(file.getFirst().length);

					tarFile.putArchiveEntry(entry);

					tarFile.write(file.getFirst());

					tarFile.closeArchiveEntry();
				}

				tarFile.finish();

				tarFile.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		};

		Thread t = new Thread(r);
		t.start();

	}

}
