package de.uni_mannheim.informatik.dws.expansion.model.utils.performance;

import java.util.Collections;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.qualifiers.TimestampPointInTime;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.Accumulateable;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentAccumulator;

public class FusionPerformance<E> {
	private E total;
	private Map<String, E> perSet;
	private Map<Boolean, E> perTimeDependence;
	private Map<ValueType, E> perDataType;
	private Map<ClassId, E> perClassId; 
	private Map<ClassPropertyId, E> perProperty; 

	private Map<TimestampPointInTime, E> perYear;

	private Map<Pair<ClassPropertyId, TimestampPointInTime>, E> perClassPropertyAndYear;

	// private Map<Pair<ClassId, String>, E> perClassAndSet;

	@Deprecated
	public FusionPerformance(E total, Map<ValueType, E> perDataType, Map<ClassId, E> perClassId, Map<ClassPropertyId, E> perProperty) {
		this.total = total;
		Map<ValueType, E> perDataTypeTemp = new EnumMap<>(ValueType.class);
		perDataTypeTemp.putAll(perDataType);
		this.perTimeDependence = Collections.unmodifiableMap(new HashMap<>());
		this.perDataType = Collections.unmodifiableMap(perDataTypeTemp);
		this.perClassId = Collections.unmodifiableMap(new HashMap<>(perClassId));
		this.perProperty = Collections.unmodifiableMap(new HashMap<>(perProperty));
		this.perYear = Collections.unmodifiableMap(new HashMap<>());
		this.perClassPropertyAndYear = Collections.unmodifiableMap(new HashMap<>());
		// this.perClassAndSet = Collections.unmodifiableMap(new HashMap<>());

	}

	public FusionPerformance(E total, Map<Boolean, E> perTimeDependence, Map<ValueType, E> perDataType, Map<ClassId, E> perClassId,
			Map<ClassPropertyId, E> perProperty, Map<TimestampPointInTime, E> perYear,
			Map<Pair<ClassPropertyId, TimestampPointInTime>, E> perClassPropertyAndYear) {
		this.total = total;
		Map<ValueType, E> perDataTypeTemp = new EnumMap<>(ValueType.class);
		perDataTypeTemp.putAll(perDataType);
		this.perTimeDependence = Collections.unmodifiableMap(new HashMap<>(perTimeDependence));
		this.perDataType = Collections.unmodifiableMap(perDataTypeTemp);
		this.perClassId = Collections.unmodifiableMap(new HashMap<>(perClassId));
		this.perProperty = Collections.unmodifiableMap(new HashMap<>(perProperty));
		this.perYear = Collections.unmodifiableMap(new HashMap<>(perYear));
		this.perClassPropertyAndYear = Collections.unmodifiableMap(perClassPropertyAndYear);
		// this.perClassAndSet = Collections.unmodifiableMap(perClassAndSet);

	}

	@Deprecated
	public FusionPerformance(E total, Map<Boolean, E> perTimeDependence, Map<ValueType, E> perDataType, Map<ClassId, E> perClassId,
			Map<ClassPropertyId, E> perProperty, Map<Pair<ClassId, String>, E> perClassAndSet) {
		this.total = total;
		Map<ValueType, E> perDataTypeTemp = new EnumMap<>(ValueType.class);
		perDataTypeTemp.putAll(perDataType);
		this.perTimeDependence = Collections.unmodifiableMap(new HashMap<>(perTimeDependence));
		this.perDataType = Collections.unmodifiableMap(perDataTypeTemp);
		this.perClassId = Collections.unmodifiableMap(new HashMap<>(perClassId));
		this.perProperty = Collections.unmodifiableMap(new HashMap<>(perProperty));
		this.perYear = Collections.unmodifiableMap(new HashMap<>());
		this.perClassPropertyAndYear = Collections.unmodifiableMap(new HashMap<>());
		// this.perClassAndSet = Collections.unmodifiableMap(perClassAndSet);
	}

	private FusionPerformance(E total) {
		this.total = total;
		this.perTimeDependence = new ConcurrentHashMap<>();
		this.perDataType = new ConcurrentHashMap<>();
		this.perClassId = new ConcurrentHashMap<>();
		this.perProperty = new ConcurrentHashMap<>();
		this.perYear = new ConcurrentHashMap<>();
		this.perClassPropertyAndYear = new ConcurrentHashMap<>();
	}

	private FusionPerformance() {
		this.total = null;
		this.perTimeDependence = new HashMap<>();
		this.perDataType = new HashMap<>();
		this.perClassId = new HashMap<>();
		this.perProperty = new HashMap<>();
		this.perYear = new HashMap<>();
		this.perClassPropertyAndYear = new HashMap<>();
	}

	private FusionPerformance<E> copy() {
		FusionPerformance<E> fusionPerformance = new FusionPerformance<>();
		fusionPerformance.total = this.total;
		fusionPerformance.perTimeDependence = Collections.unmodifiableMap(new HashMap<>(this.perTimeDependence));
		fusionPerformance.perDataType = Collections.unmodifiableMap(new HashMap<>(this.perDataType));
		fusionPerformance.perClassId = Collections.unmodifiableMap(new HashMap<>(this.perClassId));
		fusionPerformance.perProperty = Collections.unmodifiableMap(new HashMap<>(this.perProperty));
		fusionPerformance.perYear = Collections.unmodifiableMap(new HashMap<>(this.perYear));
		fusionPerformance.perClassPropertyAndYear = Collections.unmodifiableMap(new HashMap<>(this.perClassPropertyAndYear));

		return fusionPerformance;
	}

	public FusionPerformanceBuilder<E> rebuilder() {
		return new FusionPerformanceBuilder<>(this);
	}

	public E getTotal() {
		return total;
	}

	public Map<Boolean, E> getPerTimeDependence() {
		return perTimeDependence;
	}

	public Map<ValueType, E> getPerDataType() {
		return perDataType;
	}

	public Map<ClassId, E> getPerClassId() {
		return perClassId;
	}

	public Map<ClassPropertyId, E> getPerProperty() {
		return perProperty;
	}

	public Map<TimestampPointInTime, E> getPerYear() {
		return perYear;
	}

	public Map<Pair<ClassPropertyId, TimestampPointInTime>, E> getPerClassPropertyAndYear() {
		return perClassPropertyAndYear;
	}

	public static <E> FusionPerformanceBuilder<E> builder() {
		return new FusionPerformanceBuilder<>();
	}

	public static <E extends Accumulateable<E>> FusionPerformanceCreator<E> accumulator(E empty) {
		return new FusionPerformanceCreator<>(empty);
	}

	public static class FusionPerformanceCreator<F extends Accumulateable<F>> {
		private final FusionPerformance<ConcurrentAccumulator<F>> fusionPerformance;
		private final F empty;

		private FusionPerformanceCreator(F empty) {
			this.fusionPerformance = new FusionPerformance<ConcurrentAccumulator<F>>(new ConcurrentAccumulator<>(empty));
			this.empty = empty;
		}

		public void addTotal(F val) {
			fusionPerformance.total.add(val);
		}
		
		public void adddPerSet(String set, F val) {
			CHMHelper.getFromCHM((ConcurrentHashMap<String, ConcurrentAccumulator<F>>) fusionPerformance.perSet,
					set, () -> new ConcurrentAccumulator<>(empty)).add(val);
		}

		public void addPerTimeDependence(boolean timeDepedence, F val) {
			CHMHelper.getFromCHM((ConcurrentHashMap<Boolean, ConcurrentAccumulator<F>>) fusionPerformance.perTimeDependence,
					timeDepedence ? Boolean.TRUE : Boolean.FALSE, () -> new ConcurrentAccumulator<>(empty)).add(val);
		}

		public void addPerDataType(ValueType key, F val) {
			CHMHelper.getFromCHM((ConcurrentHashMap<ValueType, ConcurrentAccumulator<F>>) fusionPerformance.perDataType, key,
					() -> new ConcurrentAccumulator<>(empty)).add(val);
		}

		public void addPerClassId(ClassId key, F val) {
			CHMHelper.getFromCHM((ConcurrentHashMap<ClassId, ConcurrentAccumulator<F>>) fusionPerformance.perClassId, key,
					() -> new ConcurrentAccumulator<>(empty)).add(val);
		}
		
 

		// public void addPerClassIdAndSet(ClassId classId, String set, F val) {
		// CHMHelper.getFromCHM((ConcurrentHashMap<Pair<ClassId, String>,
		// ConcurrentAccumulator<F>>) fusionPerformance.perClassAndSet,
		// new Pair<>(classId, set), () -> new
		// ConcurrentAccumulator<>(empty)).add(val.copy());
		// }

		public void addPerProperty(ClassPropertyId key, F val) {
			CHMHelper.getFromCHM((ConcurrentHashMap<ClassPropertyId, ConcurrentAccumulator<F>>) fusionPerformance.perProperty, key,
					() -> new ConcurrentAccumulator<>(empty)).add(val);
		}

		public void addPerYear(TimestampPointInTime tpit, F val) {
			CHMHelper.getFromCHM((ConcurrentHashMap<TimestampPointInTime, ConcurrentAccumulator<F>>) fusionPerformance.perYear, tpit,
					() -> new ConcurrentAccumulator<>(empty)).add(val);
		}

		public void addPerClassPropertyIdAndYear(ClassPropertyId classPropertyId, TimestampPointInTime tpit, F val) {
			Pair<ClassPropertyId, TimestampPointInTime> key = new Pair<>(classPropertyId, tpit);
			CHMHelper.getFromCHM(
					(ConcurrentHashMap<Pair<ClassPropertyId, TimestampPointInTime>, ConcurrentAccumulator<F>>) fusionPerformance.perClassPropertyAndYear, key,
					() -> new ConcurrentAccumulator<>(empty)).add(val);
		}

		// public void addForAll(ClassPropertyId cpid, OntologyRepository
		// ontologyRepository, F val, String set) {
		// addTotal(val);
		// addPerTimeDependence(ontologyRepository.isTimeDependent(cpid.getPropertyId()),
		// val);
		// addPerDataType(ontologyRepository.getDataType(cpid.getPropertyId()),
		// val);
		// addPerClassId(cpid.getClassId(), val);
		// addPerProperty(cpid, val);
		// }
		//
		// public void addForAll(ClassPropertyId cpid, OntologyRepository
		// ontologyRepository, F val) {
		// addTotal(val);
		// addPerTimeDependence(ontologyRepository.isTimeDependent(cpid.getPropertyId()),
		// val);
		// addPerDataType(ontologyRepository.getDataType(cpid.getPropertyId()),
		// val);
		// addPerClassId(cpid.getClassId(), val);
		// addPerProperty(cpid, val);
		// }

		public void addWithYear(ClassId classId, QualifiedPropertyId qualifiedPropertyId, boolean timeDependent, ValueType valueType, F val) {
			ClassPropertyId cpid = new ClassPropertyId(classId, qualifiedPropertyId.getPropertyId());

			TimestampPointInTime tipt = null;
			for (AbstractQualifier qualifier : qualifiedPropertyId.getQualifiers()) {
				if (qualifier instanceof TimestampPointInTime) {
					tipt = (TimestampPointInTime) qualifier;
					break;
				}
			}
			if (tipt != null) {
				addPerYear(tipt, val);
				addPerClassPropertyIdAndYear(cpid, tipt, val);

			}

			addTotal(val);
			addPerTimeDependence(timeDependent, val);
			addPerDataType(valueType, val);
			addPerClassId(cpid.getClassId(), val);
			addPerProperty(cpid, val);
		}

		public <X> FusionPerformance<X> get(Function<F, X> mapper) {
			X total = mapper.apply(fusionPerformance.total.get());
			Map<Boolean, X> perTimeDependence = accumulateMap(fusionPerformance.perTimeDependence, mapper);
			Map<ValueType, X> perDataType = accumulateMap(fusionPerformance.perDataType, mapper);
			Map<ClassId, X> perClassId = accumulateMap(fusionPerformance.perClassId, mapper);
			Map<ClassPropertyId, X> perProperty = accumulateMap(fusionPerformance.perProperty, mapper);
			Map<TimestampPointInTime, X> perYear = accumulateMap(fusionPerformance.perYear, mapper);
			Map<Pair<ClassPropertyId, TimestampPointInTime>, X> perClassPropertyAndYear = accumulateMap(fusionPerformance.perClassPropertyAndYear, mapper);

			return new FusionPerformance<>(total, perTimeDependence, perDataType, perClassId, perProperty, perYear, perClassPropertyAndYear);
		}

		public FusionPerformance<F> get() {
			return get((e) -> {
				return e;
			});
		}

		private <K, X> Map<K, X> accumulateMap(Map<K, ConcurrentAccumulator<F>> original, Function<F, X> mapper) {
			return original.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				return mapper.apply(e.getValue().get());
			}));
		}

	}

	public static class FusionPerformanceBuilder<F> {
		FusionPerformance<F> fusionPerformance;

		private FusionPerformanceBuilder(FusionPerformance<F> fusionPerformance) {
			this.fusionPerformance = fusionPerformance.copy();
		}

		private FusionPerformanceBuilder() {
			this.fusionPerformance = new FusionPerformance<F>();
		}

		public FusionPerformance<F> get() {
			return this.fusionPerformance.copy();
		}

		public FusionPerformanceBuilder<F> setTotal(F total) {
			fusionPerformance.total = total;
			return this;
		}

		public FusionPerformanceBuilder<F> setPerTimeDependence(Map<Boolean, F> perTimeDependence) {
			fusionPerformance.perTimeDependence = perTimeDependence;
			return this;
		}

		public FusionPerformanceBuilder<F> setPerDataType(Map<ValueType, F> perDataType) {
			fusionPerformance.perDataType = perDataType;
			return this;
		}

		public FusionPerformanceBuilder<F> setPerClassId(Map<ClassId, F> perClassId) {
			fusionPerformance.perClassId = perClassId;
			return this;
		}

		public FusionPerformanceBuilder<F> setPerProperty(Map<ClassPropertyId, F> perProperty) {
			fusionPerformance.perProperty = perProperty;
			return this;
		}

		public FusionPerformanceBuilder<F> setPerYear(Map<TimestampPointInTime, F> perYear) {
			fusionPerformance.perYear = perYear;
			return this;
		}

		public FusionPerformanceBuilder<F> setPerClassPropertyAndYear(Map<Pair<ClassPropertyId, TimestampPointInTime>, F> perClassPropertyAndYear) {
			fusionPerformance.perClassPropertyAndYear = perClassPropertyAndYear;
			return this;
		}
	}

	@Override
	public String toString() {
		return "FusionPerformance [total=" + total + ", perTimeDependence=" + perTimeDependence + ", perDataType=" + perDataType + ", perClassId=" + perClassId
				+ ", perProperty=" + perProperty + ", perYear=" + perYear + ", perClassPropertyAndYear=" + perClassPropertyAndYear + "]";
	}

}
