package de.mannheim.uni.expansion.data.ontology.model;

import java.util.Set;

import com.google.common.collect.ImmutableSet;

import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;

public class QualifiedPropertyId implements Comparable<QualifiedPropertyId>, PropertyIdContainer {
	private final PropertyId propertyId;
	private final Set<AbstractQualifier> qualifiers;

	public QualifiedPropertyId(PropertyId propertyId, Set<AbstractQualifier> qualifiers) {
		super();
		this.propertyId = propertyId;
		this.qualifiers = ImmutableSet.copyOf(qualifiers);
	}

	public PropertyId getPropertyId() {
		return propertyId;
	}

	public Set<AbstractQualifier> getQualifiers() {
		return qualifiers;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((propertyId == null) ? 0 : propertyId.hashCode());
		result = prime * result + ((qualifiers == null) ? 0 : qualifiers.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		QualifiedPropertyId other = (QualifiedPropertyId) obj;
		if (propertyId == null) {
			if (other.propertyId != null)
				return false;
		} else if (!propertyId.equals(other.propertyId))
			return false;
		if (qualifiers == null) {
			if (other.qualifiers != null)
				return false;
		} else if (!qualifiers.equals(other.qualifiers))
			return false;
		return true;
	}

	@Override
	public String toString() {
		return "QualifiedPropertyId [propertyId=" + propertyId + ", qualifiers=" + qualifiers + "]";
	}

	@Override
	public int compareTo(QualifiedPropertyId o) {
		return this.toString().compareTo(o.toString());
	}

	@Override
	public PropertyId propertyId() {
		return this.propertyId;
	}
}
