package de.mannheim.uni.expansion.data.ontology.model;

import java.util.Set;

import com.google.common.collect.ImmutableSet;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;

public class QualifiedTripleId implements Comparable<QualifiedTripleId>, TripleIdContainer {

	private final TripleId tripleId;
	private final Set<AbstractQualifier> qualifiers;

	public QualifiedTripleId(ClassId classId, InstanceId instanceId, QualifiedPropertyId qualifiedPropertyId) {
		super();
		this.tripleId = new TripleId(classId, qualifiedPropertyId.getPropertyId(), instanceId);
		this.qualifiers = ImmutableSet.copyOf(qualifiedPropertyId.getQualifiers());
	}

	public QualifiedTripleId(TripleId tripleId, Set<AbstractQualifier> qualifiers) {
		super();
		this.tripleId = tripleId;
		this.qualifiers = ImmutableSet.copyOf(qualifiers);
	}

	// private Map<PropertyId, Integer> getQualifierNameIndex(List<Qualifier>
	// qualifiers) {
	// Map<PropertyId, Integer> index = new HashMap<>();
	//
	// for (int i = 0; i < qualifiers.size(); i++) {
	// Qualifier q = qualifiers.get(i);
	// if (index.containsKey(i)) {
	// throw new IllegalStateException("This value has the same qualifier
	// twice");
	// }
	// index.put(q.getQualifier(), i);
	// }
	// return index;
	//
	// }

	public TripleId getTripleId() {
		return tripleId;
	}

	public Set<AbstractQualifier> getQualifiers() {
		return qualifiers;
	}

	public QualifiedPropertyId getQualifiedPropertyId() {
		return new QualifiedPropertyId(this.getTripleId().getPropertyId(), qualifiers);
	}

	private transient Integer hashCode;

	@Override
	public int hashCode() {
		if (hashCode == null) {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((qualifiers == null) ? 0 : qualifiers.hashCode());
			result = prime * result + ((tripleId == null) ? 0 : tripleId.hashCode());
			hashCode = result;
		}
		return hashCode;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		if (this.hashCode() != obj.hashCode())
			return false;
		QualifiedTripleId other = (QualifiedTripleId) obj;
		if (qualifiers == null) {
			if (other.qualifiers != null)
				return false;
		} else if (!qualifiers.equals(other.qualifiers))
			return false;
		if (tripleId == null) {
			if (other.tripleId != null)
				return false;
		} else if (!tripleId.equals(other.tripleId))
			return false;
		return true;
	}

	@Override
	public String toString() {
		return "QualifiedTripleId [tripleId=" + tripleId + ", qualifiers=" + qualifiers + "]";
	}

	@Override
	public int compareTo(QualifiedTripleId o) {
		return this.toString().compareTo(o.toString());
	}

	@Override
	public TripleId tripleId() {
		return tripleId;
	}

	// public Qualifier getQualifierValue(PropertyId propertyId) {
	// Integer index = this.qualifierNameIndex.get(propertyId);
	// if (index == null)
	// return null;
	// else
	// return qualifiers.get(index);
	// }

}
