package de.mannheim.uni.expansion.data.ontology.model;

import java.io.Serializable;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;

public class TripleId implements Serializable, Comparable<TripleId>, TripleIdContainer {

	/**
	 * 
	 */
	private static final long serialVersionUID = 2L;

	private final ClassId classId;

	private final InstanceId instanceId;

	private final PropertyId propertyId;

	public TripleId(ClassId classId, PropertyId propertyId, InstanceId instanceId) {
		super();
		this.classId = classId;
		this.instanceId = instanceId;
		this.propertyId = propertyId;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((propertyId == null) ? 0 : propertyId.hashCode());
		result = prime * result + ((classId == null) ? 0 : classId.hashCode());
		result = prime * result + ((instanceId == null) ? 0 : instanceId.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		TripleId other = (TripleId) obj;
		if (propertyId == null) {
			if (other.propertyId != null)
				return false;
		} else if (!propertyId.equals(other.propertyId))
			return false;
		if (classId == null) {
			if (other.classId != null)
				return false;
		} else if (!classId.equals(other.classId))
			return false;
		if (instanceId == null) {
			if (other.instanceId != null)
				return false;
		} else if (!instanceId.equals(other.instanceId))
			return false;
		return true;
	}

	public ClassId getClassId() {
		return classId;
	}

	public InstanceId getInstanceId() {
		return instanceId;
	}

	public PropertyId getPropertyId() {
		return propertyId;
	}

	public ClassPropertyId getClassPropertyId() {
		return new ClassPropertyId(this);
	}

	@Override
	public String toString() {
		return "[class=" + classId + ", instance=" + instanceId + ", property=" + propertyId + "]";
	}

	public ValueId getValueId() {
		return new ValueId(this.instanceId, this.propertyId);
	}

	@Override
	public TripleId tripleId() {
		return this;
	}

	@Override
	public int compareTo(TripleId o) {
		return this.toString().compareTo(o.toString());
	}

}
