package de.mannheim.uni.expansion.data.ontology.model.classes;

import java.io.Serializable;

public class ClassId implements Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	final private String classString;

	public enum ClassIdStyle {
		DBPEDIA_URI, NONE;
	}

	@Override
	public String toString() {
		return classString;
	}

	public ClassId(String clazz) {
		this.classString = clazz;
	}

	public String getClassId() {
		return classString;
	}

	public static ClassId fromString(String s, FileEnding ending) {

		if (ending == FileEnding.CSV) {
			s = s.substring(0, s.length() - 4);
		} else if (ending == FileEnding.CSV_GZ) {
			s = s.substring(0, s.length() - 7);
		} else if (ending == FileEnding.NO_URI_GZ) {
			s = s.substring(0, s.length() - 3);
		}

		if (ending == FileEnding.NO_URI_GZ || ending == FileEnding.NO_URL) {
			return new ClassId(s);
		} else {

			return new ClassId("http://dbpedia.org/ontology/" + (s).replace(" ", "_"));
		}
	}

	public static ClassId fromString(String s, ClassIdStyle classIdStyle) {

		switch (classIdStyle) {
		case DBPEDIA_URI:
			return new ClassId("http://dbpedia.org/ontology/" + (s).replace(" ", "_"));
		case NONE:
			return new ClassId(s);
		default:
			throw new IllegalStateException();

		}

	}

	private transient Integer hashCode;

	@Override
	public int hashCode() {
		if (hashCode == null) {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((classString == null) ? 0 : classString.hashCode());
			hashCode = result;
		}
		return hashCode;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		if (this.hashCode() != obj.hashCode())
			return false;
		ClassId other = (ClassId) obj;
		if (classString == null) {
			if (other.classString != null)
				return false;
		} else if (!classString.equals(other.classString))
			return false;
		return true;
	}

	public static enum FileEnding {
		NONE, CSV, CSV_GZ, NO_URI_GZ, NO_URL;
	}

	public String getFinalPart() {
		String[] splits = classString.split("/");
		return splits[splits.length - 1];
	}

	public String getReadableString(ClassIdStyle classIdStyle) {
		switch (classIdStyle) {
		case DBPEDIA_URI:
			String[] splits = classString.split("/");
			return splits[splits.length - 1];
		case NONE:
			return classString;
		default:
			throw new IllegalStateException();

		}
	}

}
