package de.mannheim.uni.expansion.data.ontology.model.classes;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

public class ClassOntology {

	private final ClassId id;

	private final Set<ClassId> subClassOf;

	private final Set<ClassId> equivalentClass;

	private final ClassId wasDerviedFrom;

	public ClassOntology(ClassId id, Set<ClassId> subClassOf, Set<ClassId> equivalentClass, ClassId wasDerviedFrom) {
		super();
		this.id = id;
		this.subClassOf = new HashSet<ClassId>(subClassOf);
		this.equivalentClass = new HashSet<ClassId>(equivalentClass);
		this.wasDerviedFrom = wasDerviedFrom;
	}

	public ClassId getId() {
		return id;
	}

	public Set<ClassId> getSubClassOf() {
		return Collections.unmodifiableSet(subClassOf);
	}

	public Set<ClassId> getEquivalentClass() {
		return Collections.unmodifiableSet(equivalentClass);
	}

	public ClassId getWasDerviedFrom() {
		return wasDerviedFrom;
	}

	@Override
	public String toString() {
		return id.toString();
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((id == null) ? 0 : id.hashCode());
		result = prime * result + ((equivalentClass == null) ? 0 : equivalentClass.hashCode());
		result = prime * result + ((subClassOf == null) ? 0 : subClassOf.hashCode());
		result = prime * result + ((wasDerviedFrom == null) ? 0 : wasDerviedFrom.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		ClassOntology other = (ClassOntology) obj;
		if (id == null) {
			if (other.id != null)
				return false;
		} else if (!id.equals(other.id))
			return false;
		if (equivalentClass == null) {
			if (other.equivalentClass != null)
				return false;
		} else if (!equivalentClass.equals(other.equivalentClass))
			return false;
		if (subClassOf == null) {
			if (other.subClassOf != null)
				return false;
		} else if (!subClassOf.equals(other.subClassOf))
			return false;
		if (wasDerviedFrom == null) {
			if (other.wasDerviedFrom != null)
				return false;
		} else if (!wasDerviedFrom.equals(other.wasDerviedFrom))
			return false;
		return true;
	}

}
