package de.mannheim.uni.expansion.data.ontology.model.instances;

import java.io.Serializable;
import java.util.EnumMap;
import java.util.Map;

import de.mannheim.uni.expansion.utils.CHMHelper;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class InstanceId extends AbstractObject implements Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	public enum InstanceIdStyle {
		DBPEDIA_URI, NONE;
	}

	final private String instance;

	final private String readableString;

	public InstanceId(String instance) {
		this.instance = instance;
		this.readableString = getReadableString(instance);
	}

	public InstanceId(String instance, String readableString) {
		this.instance = instance;
		this.readableString = readableString;
	}

	public static InstanceId simple(String s) {
		return new InstanceId(s, s);
	}

	@Deprecated
	public static InstanceId fromString(String s) {
		return new InstanceId("http://dbpedia.org/resource/" + (s).replace(" ", "_"), s);
	}

	public InstanceId(String instance, InstanceIdStyle instanceIdStyle) {
		String readableString;
		switch (instanceIdStyle) {
		case DBPEDIA_URI:
			readableString = getReadableString(instance);
			break;
		case NONE:
			readableString = instance;
			break;
		default:
			throw new IllegalStateException();

		}

		this.instance = instance;
		this.readableString = readableString;

	}

	public static InstanceId fromReversibleString(String s, InstanceIdStyle instanceIdStyle) {

		switch (instanceIdStyle) {
		case DBPEDIA_URI:
			return new InstanceId("http://dbpedia.org/resource/" + (s).replace(" ", "_"), s);

		case NONE:
			return InstanceId.fromCombinedString(s);
		default:
			throw new IllegalStateException();

		}

	}

	@Override
	public String toString() {
		return "[" + instance + ", " + readableString + "]";
	}

	public String getInstanceId() {
		return instance;
	}

	private static int beginIndex = "http://dbpedia.org/resource/".length();

	public String getReadableString(String instanceString) {
		String ret = instanceString.substring(beginIndex).replace("_", " ");
		return ret;
	}

	public String getReadableString() {
		return readableString;
	}

	private transient String combinedString;

	public String toCombinedString() {
		if (combinedString == null)
			combinedString = getReadableString().replace(" ", "_").replace("__", "_") + "__" + getInstanceId();
		return combinedString;
	}

	private transient Map<InstanceIdStyle, String> reversibleReadbleStringMap = new EnumMap<>(InstanceIdStyle.class);

	public String toReversibleReadableString(InstanceIdStyle instanceIdStyle) {

		return CHMHelper.getFromMap(reversibleReadbleStringMap, instanceIdStyle, () -> createReversibleReadableString(instanceIdStyle));

	}

	private String createReversibleReadableString(InstanceIdStyle instanceIdStyle) {
		switch (instanceIdStyle) {
		case DBPEDIA_URI:
			return readableString;
		case NONE:
			return toCombinedString();
		default:
			throw new IllegalStateException();

		}
	}

	public static InstanceId fromCombinedString(String combinedString) {
		try {
			String[] parts = combinedString.split("__");
			if (parts.length == 1) {
				return InstanceId.simple(combinedString);
			} else {
				return new InstanceId(parts[1], parts[0]);
			}
		} catch (Exception e) {
			throw e;
		}
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((instance == null) ? 0 : instance.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		InstanceId other = (InstanceId) obj;
		if (instance == null) {
			if (other.instance != null)
				return false;
		} else if (!instance.equals(other.instance))
			return false;
		return true;
	}

	@Override
	public String readableString() {
		return readableString;
	}

	@Override
	public ValueType getType() {
		return ValueType.INSTANCE_REFERENCE;
	}

	public static int compare(InstanceId second, InstanceId second2) {
		return second.instance.compareTo(second2.instance);
	}

}
