package de.mannheim.uni.expansion.data.ontology.model.properties;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.tableprocessor.model.TableColumn.ColumnDataType;

public class DataTypePropertyOntology extends PropertyOntology {

	public enum Type {
		NON_NEGATIVE_INTEGER(ColumnDataType.numeric),
		POSITIVE_INTEGER(ColumnDataType.numeric),
		INTEGER(ColumnDataType.numeric),
		STRING(ColumnDataType.string),
		G_YEAR(ColumnDataType.date),
		DOUBLE(ColumnDataType.numeric),
		BOOLEAN(ColumnDataType.bool),
		DATE(ColumnDataType.date),
		G_YEAR_MONTH(ColumnDataType.date),
		FLOAT(ColumnDataType.numeric);

		private final ColumnDataType columnDataType;

		private Type(ColumnDataType columnDataType) {
			this.columnDataType = columnDataType;
		}

		public ColumnDataType getColumnDataType() {
			return columnDataType;
		}

		private static Map<String, Type> reverseMap = new HashMap<String, Type>();

		static {
			reverseMap.put("http://www.w3.org/2001/XMLSchema#nonNegativeInteger", NON_NEGATIVE_INTEGER);
			reverseMap.put("http://www.w3.org/2001/XMLSchema#positiveInteger", POSITIVE_INTEGER);
			reverseMap.put("http://www.w3.org/2001/XMLSchema#integer", INTEGER);
			reverseMap.put("http://www.w3.org/1999/02/22-rdf-syntax-ns#langString", STRING);
			reverseMap.put("http://www.w3.org/2001/XMLSchema#string", STRING);
			reverseMap.put("http://www.w3.org/2001/XMLSchema#gYear", G_YEAR);
			reverseMap.put("http://www.w3.org/2001/XMLSchema#double", DOUBLE);
			reverseMap.put("http://www.w3.org/2001/XMLSchema#boolean", BOOLEAN);
			reverseMap.put("http://www.w3.org/2001/XMLSchema#date", DATE);

			reverseMap.put("http://www.w3.org/2001/XMLSchema#gYearMonth", G_YEAR_MONTH);
			reverseMap.put("http://www.w3.org/2001/XMLSchema#float", FLOAT);

			// http://www.w3.org/2001/XMLSchema#float
			// http://www.w3.org/2001/XMLSchema#dateTime
			// http://dbpedia.org/datatype/millimetre
			// http://dbpedia.org/datatype/millimetre
			// http://dbpedia.org/datatype/kilogram
			// http://dbpedia.org/datatype/squareKilometre
			// http://dbpedia.org/datatype/kilogram
			// http://dbpedia.org/datatype/cubicKilometre
			// http://dbpedia.org/datatype/cubicMetre
			// http://dbpedia.org/datatype/millimetre
			// http://dbpedia.org/datatype/kilogram
			// http://dbpedia.org/datatype/millimetre
			// http://dbpedia.org/datatype/kilogram
			// http://dbpedia.org/datatype/kilometre
			// http://dbpedia.org/datatype/day
			// http://dbpedia.org/datatype/kilometre
			// http://dbpedia.org/datatype/millimetre
			// http://dbpedia.org/datatype/metre
			// http://dbpedia.org/datatype/hour
			// http://dbpedia.org/datatype/kilogram
			// http://dbpedia.org/datatype/cubicMetrePerSecond
			// http://dbpedia.org/datatype/kilometre
			// http://dbpedia.org/datatype/kelvin
			// http://dbpedia.org/datatype/kilogram
			// http://dbpedia.org/datatype/millimetre
			// http://dbpedia.org/datatype/kilometre
			// http://dbpedia.org/datatype/squareKilometre
			// http://dbpedia.org/datatype/kilogram
			// http://dbpedia.org/datatype/metre
			// http://dbpedia.org/datatype/kelvin
			// http://dbpedia.org/datatype/squareKilometre
			// http://dbpedia.org/datatype/minute
			// http://dbpedia.org/datatype/cubicMetrePerSecond
			// http://dbpedia.org/datatype/millimetre
			// http://dbpedia.org/datatype/day
			// http://dbpedia.org/datatype/kilogram
			// http://dbpedia.org/datatype/inhabitantsPerSquareKilometre
			// http://dbpedia.org/datatype/millimetre
			// http://dbpedia.org/datatype/kilogram
			// http://dbpedia.org/datatype/kilometre
			// http://dbpedia.org/datatype/inhabitantsPerSquareKilometre
			// http://dbpedia.org/datatype/millimetre
			// http://dbpedia.org/datatype/cubicMetrePerSecond
			// http://dbpedia.org/datatype/kelvin
			// http://dbpedia.org/datatype/kilometre
			// http://dbpedia.org/datatype/millimetre
			// http://dbpedia.org/datatype/kilowatt
			// http://dbpedia.org/datatype/kilogram
			// http://dbpedia.org/datatype/day
			// http://dbpedia.org/datatype/day
			// http://dbpedia.org/datatype/squareKilometre
			// http://dbpedia.org/datatype/cubicMetrePerSecond
			// http://dbpedia.org/datatype/hour
			// http://dbpedia.org/datatype/kelvin
			// http://dbpedia.org/datatype/metre
			// http://dbpedia.org/datatype/kilogramPerCubicMetre
			// http://dbpedia.org/datatype/litre
			// http://dbpedia.org/datatype/kelvin
			// http://dbpedia.org/datatype/inhabitantsPerSquareKilometre
			// http://dbpedia.org/datatype/millimetre
			// http://dbpedia.org/datatype/hour
			// http://dbpedia.org/datatype/kilogramPerCubicMetre
			// http://dbpedia.org/datatype/minute
			// http://dbpedia.org/datatype/kilogram
			// http://dbpedia.org/datatype/millimetre
			// http://dbpedia.org/datatype/kilometre
			// http://dbpedia.org/datatype/kelvin
			// http://dbpedia.org/datatype/kilogram
			// http://dbpedia.org/datatype/kilometre
			// http://dbpedia.org/datatype/cubicMetre
			// http://dbpedia.org/datatype/hour
			// http://dbpedia.org/datatype/inhabitantsPerSquareKilometre
			// http://dbpedia.org/datatype/day
			// http://dbpedia.org/datatype/kilometrePerHour
			// http://dbpedia.org/datatype/kelvin
			// http://dbpedia.org/datatype/metre
			// http://dbpedia.org/datatype/squareKilometre
			// http://dbpedia.org/datatype/kelvin
			// http://dbpedia.org/datatype/metre
			// http://dbpedia.org/datatype/gramPerKilometre
			// http://dbpedia.org/datatype/newtonMetre
			// http://dbpedia.org/datatype/kelvin
			// http://dbpedia.org/datatype/kilogram
			// http://dbpedia.org/datatype/hour
			// http://dbpedia.org/datatype/cubicCentimetre
			// http://dbpedia.org/datatype/day
			// http://dbpedia.org/datatype/kelvin
			// http://dbpedia.org/datatype/kelvin
			// http://dbpedia.org/datatype/squareKilometre
			// http://dbpedia.org/datatype/kilometre
			// http://dbpedia.org/datatype/millimetre
			// http://dbpedia.org/datatype/kilometre
			// http://dbpedia.org/datatype/day
			// http://dbpedia.org/datatype/kilogram
			// http://dbpedia.org/datatype/kilometre
			// http://dbpedia.org/datatype/kilogram
			// http://dbpedia.org/datatype/second
			// http://dbpedia.org/datatype/squareKilometre
			// http://dbpedia.org/datatype/kilogramPerCubicMetre
			// http://dbpedia.org/datatype/kilogram
			// http://dbpedia.org/datatype/megabyte
			// http://dbpedia.org/datatype/kilometre
			// http://dbpedia.org/datatype/kilometre
			// http://dbpedia.org/datatype/squareKilometre
			// http://dbpedia.org/datatype/squareMetre
			// http://dbpedia.org/datatype/kilometrePerSecond
			// http://dbpedia.org/datatype/squareKilometre
			// http://dbpedia.org/datatype/kilogram
			// http://dbpedia.org/datatype/millimetre
			// http://dbpedia.org/datatype/kilometrePerSecond
			// http://dbpedia.org/datatype/day
			// http://dbpedia.org/datatype/squareKilometre
			// http://dbpedia.org/datatype/kilometre
			// http://dbpedia.org/datatype/kelvin
			// http://dbpedia.org/datatype/cubicKilometre
			// http://dbpedia.org/datatype/hour
			// http://dbpedia.org/datatype/kilometre
			// http://dbpedia.org/datatype/centimetre

			// reverseMap.put("", );

		}

		public static Type getTypeFromString(String s) {
			return reverseMap.get(s);
		}

	}

	private final Type range;

	public DataTypePropertyOntology(PropertyId id, ClassId domain, Type range, Set<PropertyId> subPropertyOf, PropertyId wasDerivedFrom) {
		super(id, domain, subPropertyOf, wasDerivedFrom);
		this.range = range;

	}

	public Type getRange() {
		return range;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = super.hashCode();
		result = prime * result + ((range == null) ? 0 : range.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (!super.equals(obj))
			return false;
		if (getClass() != obj.getClass())
			return false;
		DataTypePropertyOntology other = (DataTypePropertyOntology) obj;
		if (range != other.range)
			return false;
		return true;
	}

}
