package de.mannheim.uni.expansion.data.ontology.model.properties;

import java.io.Serializable;

import de.mannheim.uni.expansion.data.ontology.model.PropertyIdContainer;
import de.mannheim.uni.expansion.data.ontology.model.instances.QualifierId;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class PropertyId extends AbstractObject implements Serializable, PropertyIdContainer {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	private final String property;
	private final String readableString;

	public PropertyId(String property) {
		this.property = property;
		this.readableString = property.trim();
	}

	public PropertyId(String property, String readableString) {
		this.property = property;
		this.readableString = readableString.trim();
	}

	@Deprecated
	public static PropertyId fromString(String s) {
		return new PropertyId("http://dbpedia.org/ontology/" + (s).replace(" ", "_"));
	}

	@Override
	public String toString() {
		if (readableString != null)
			return readableString;
		return property;
	}

	public String getReadableString() {
		return readableString;
	}

	public String getAsString() {
		if (readableString != null)
			return readableString;
		return property;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((property == null) ? 0 : property.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		PropertyId other = (PropertyId) obj;
		if (property == null) {
			if (other.property != null)
				return false;
		} else if (!property.equals(other.property))
			return false;
		return true;
	}

	private transient String finalPart;

	public String getFinalPart() {
		if (finalPart == null) {
			String[] splits = property.split("/");
			finalPart = splits[splits.length - 1];
		}
		return finalPart;
	}

	public static String dbpediaPrefix = "http://dbpedia.org/ontology/";

	public String getFinalPart2() {

		return property.substring(dbpediaPrefix.length());
	}

	public String getPropertyId() {
		return property;
	}

	public String toCombinedString() {
		return getReadableString().trim().replace(" ", "_").replace("__", "_") + "__" + getPropertyId();
	}

	public static PropertyId fromCombinedString(String combinedString) {

		String[] qualifierPart = combinedString.split("::");
		if (qualifierPart.length == 2) {
			return new QualifierId(fromCombinedPropertyId(qualifierPart[0]), fromCombinedPropertyId(qualifierPart[1]));
		} else {
			return fromCombinedPropertyId(combinedString);
		}

	}

	private static PropertyId fromCombinedPropertyId(String combinedString) {
		String[] parts = combinedString.split("__");
		if (parts.length < 2) {
			return new PropertyId(combinedString, combinedString);
		} else {
			return new PropertyId(parts[1], parts[0]);
		}
	}

	@Override
	public PropertyId propertyId() {
		return this;
	}

	@Override
	public String readableString() {
		return getReadableString();
	}

	@Override
	public ValueType getType() {
		return ValueType.PROPERTY_REFERENCE;
	}

}
