package de.mannheim.uni.expansion.data.ontology.model.properties;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;

public abstract class PropertyOntology {

	private final PropertyId id;

	private final ClassId domain;

	private final Set<PropertyId> subPropertyOf;

	private final PropertyId wasDerivedFrom;

	public PropertyOntology(PropertyId id, ClassId domain, Set<PropertyId> subPropertyOf, PropertyId wasDerivedFrom) {
		this.id = id;
		this.domain = domain;
		this.subPropertyOf = new HashSet<PropertyId>(subPropertyOf);
		this.wasDerivedFrom = wasDerivedFrom;
	}

	public PropertyId getId() {
		return id;
	}

	public ClassId getDomain() {
		return domain;
	}

	public Set<PropertyId> getSubPropertyOf() {
		return Collections.unmodifiableSet(subPropertyOf);
	}

	public PropertyId getWasDerivedFrom() {
		return wasDerivedFrom;
	}

	@Override
	public String toString() {
		return id.toString();
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((domain == null) ? 0 : domain.hashCode());
		result = prime * result + ((id == null) ? 0 : id.hashCode());
		result = prime * result + ((subPropertyOf == null) ? 0 : subPropertyOf.hashCode());
		result = prime * result + ((wasDerivedFrom == null) ? 0 : wasDerivedFrom.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		PropertyOntology other = (PropertyOntology) obj;
		if (domain == null) {
			if (other.domain != null)
				return false;
		} else if (!domain.equals(other.domain))
			return false;
		if (id == null) {
			if (other.id != null)
				return false;
		} else if (!id.equals(other.id))
			return false;
		if (subPropertyOf == null) {
			if (other.subPropertyOf != null)
				return false;
		} else if (!subPropertyOf.equals(other.subPropertyOf))
			return false;
		if (wasDerivedFrom == null) {
			if (other.wasDerivedFrom != null)
				return false;
		} else if (!wasDerivedFrom.equals(other.wasDerivedFrom))
			return false;
		return true;
	}

}
