package de.uni_mannheim.dws.expansion.data.ontology.model;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.tableprocessor.model.TableColumn.ColumnDataType;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.BooleanObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.CoordinateObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.QuantityObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.UrlObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class ValueConverter {

	@Inject
	public ValueConverter() {
	}

	public QualifiedValueList getQVL(Object value, boolean isReference) {
		if (value instanceof ArrayList<?>) {
			List<AbstractObject> aoList = new ArrayList<>();
			ArrayList<?> list = (ArrayList<?>) value;

			for (Object o : list) {
				aoList.add(getAbstractObject(o, isReference));
			}

			return QualifiedValueList.fromList(aoList);
		} else {
			return new QualifiedValueList(getAbstractObject(value, isReference));
		}
	}

	public QualifiedValueList getQVL(Object value) {
		return getQVL(value, false);
	}

	public AbstractObject getAbstractObject(Object value, boolean isReference) {

		if (isReference) {
			if (value instanceof String) {
				return InstanceId.fromString((String) value);
			} else {
				throw new IllegalStateException();
			}
		} else if (value instanceof String) {
			return new StringObject((String) value);
		} else if (value instanceof Double) {
			return new QuantityObject((Double) value, 0.0, 0.0, "");
		} else if (value instanceof Date) {
			return TimeObject.fromDate((Date) value);
		} else if (value instanceof Boolean) {
			return new BooleanObject((Boolean) value);
		} else if (value instanceof ArrayList<?>) {
			// System.out.println("FIX LISTS IN TABLES AND STUFF" +
			// this.getClass());
			return getAbstractObject(((List<?>) value).get(0));
			// FIXME: FIX LISTS IN TABLES AND STUFF

		}
		throw new IllegalStateException();
	}

	public AbstractObject getAbstractObject(Object value) {
		return getAbstractObject(value, false);
	}

	public Object getObject(AbstractObject value) {
		if (value instanceof StringObject) {
			return ((StringObject) value).getString();
		} else if (value instanceof QuantityObject) {
			return ((QuantityObject) value).getValue();

		} else if (value instanceof TimeObject) {
			return ((TimeObject) value).getDate();
		} else if (value instanceof InstanceId) {
			return ((InstanceId) value).readableString();
		} else if (value instanceof CoordinateObject) {
			CoordinateObject co = (CoordinateObject) value;
			List<Object> ret = new ArrayList<>();
			ret.add(QuantityObject.fmt(co.getLatitude()));
			ret.add(QuantityObject.fmt(co.getLongitude()));
			return ret;
		} else if (value instanceof UrlObject) {
			return ((UrlObject) value).getUrl();
		} else if (value instanceof BooleanObject) {
			return ((BooleanObject) value).getValue();
		}

		throw new IllegalStateException();

	}

	public List<Object> convertList(List<AbstractObject> input) {
		List<Object> ret = new ArrayList<>();
		for (AbstractObject ao : input) {
			Object newObject = this.getObject(ao);
			if (newObject instanceof List<?>) {
				ret.addAll((List<? extends Object>) newObject);
			} else {
				ret.add(newObject);
			}

		}

		return ret;
	}

	public ColumnDataType getDataType(ValueType value) {
		if (value == null) {
			throw new IllegalStateException("ValueType is null. It can't be.");
		}
		switch (value) {
		case BOOLEAN:
			return ColumnDataType.bool;

		case COORDINATE:
			return ColumnDataType.coordinate;

		case EXTERNAL_ID:
			break;
		case INSTANCE_REFERENCE:
			return ColumnDataType.string;

		case PROPERTY_REFERENCE:
			break;
		case QUANTITY:
			return ColumnDataType.numeric;

		case STRING:
			return ColumnDataType.string;

		case TIME:
			return ColumnDataType.date;

		case URL:
			return ColumnDataType.link;

		default:
			break;

		}
		throw new IllegalStateException();
	}

	public ValueType getDataType(ColumnDataType value) {
		switch (value) {
		case bool:
			return ValueType.BOOLEAN;
		case coordinate:
			return ValueType.COORDINATE;
		case date:
			return ValueType.TIME;
		case link:
			return ValueType.URL;
		case list:
		case numeric:
			return ValueType.QUANTITY;
		case reference:
			return ValueType.INSTANCE_REFERENCE;
		case string:
			return ValueType.STRING;
		case unit:
			break;
		case unknown:
			return ValueType.STRING;

		default:
			break;

		}
		throw new IllegalStateException();
	}

}
