package de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.NavigableSet;
import java.util.Set;
import java.util.TreeSet;
import java.util.stream.Collectors;

import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valuepairs.AbstractClassPair;
import de.uni_mannheim.dws.expansion.data.ontology.model.valuepairs.ScoreSortedAbstractPair;

public class FusedValue {

	private final boolean isList;
	private final AbstractObject singleValue;
	private final Double singleScore;
	private final Integer singleContributors;
	private final NavigableSet<AbstractClassPair> list;

	public FusedValue(Set<? extends AbstractClassPair> list) {
		this.list = new TreeSet<>(list);
		this.isList = true;
		this.singleValue = null;
		this.singleScore = null;
		this.singleContributors = null;
	}

	public FusedValue(AbstractObject singleValue, Double singleScore, int contributors) {
		this.singleValue = singleValue;
		this.singleScore = singleScore;
		this.list = null;
		this.isList = false;
		this.singleContributors = contributors;
	}

	public static FusedValue fromObject(AbstractObject value) {

		return new FusedValue(value, 1.0, 1);
	}

	public static FusedValue fromObject(List<AbstractObject> values) {

		Set<AbstractClassPair> set = values.stream().map((e) -> {
			return new ScoreSortedAbstractPair(e, 1.0);
		}).collect(Collectors.toSet());
		return new FusedValue(set);

	}

	public boolean isList() {
		return isList;
	}

	public AbstractObject getSingle() {
		return singleValue;
	}

	public Double getSingleScore() {
		return singleScore;
	}

	private final Double delta = 0.00001;

	public List<Set<AbstractObject>> getRankedList() {
		List<Set<AbstractObject>> rankedList = new ArrayList<>();

		Double lastSim = null;
		Set<AbstractObject> currentSet = null;
		for (AbstractClassPair e : this.list.descendingSet()) {

			if (currentSet == null) {
				currentSet = new HashSet<>();
				lastSim = e.getScore();

				currentSet.add(e.getValue());
			} else if (equal(lastSim, e.getScore(), delta)) {

				currentSet.add(e.getValue());
			} else {
				rankedList.add(currentSet);
				currentSet = new HashSet<>();
				lastSim = e.getScore();

				currentSet.add(e.getValue());
			}

		}

		rankedList.add(currentSet);

		return rankedList;
	}

	public NavigableSet<AbstractClassPair> getScoredList() {
		return Collections.unmodifiableNavigableSet(this.list);
	}

	private boolean equal(Double d1, Double d2, Double delta) {
		if (Math.abs(d1 - d2) < delta) {
			return true;
		} else {
			return false;
		}
	}

	public List<AbstractObject> getSimpleList() {
		List<AbstractObject> simpleList = new ArrayList<>();

		for (AbstractClassPair e : this.list.descendingSet()) {
			simpleList.add(e.getValue());

		}

		return simpleList;
	}

	public int listSize() {
		if (list == null)
			return 0;
		return list.size();

	}

	@Override
	public String toString() {
		return "FusedValue [isList=" + isList + ", singleValue=" + singleValue + ", singleScore=" + singleScore + ", singleContributors=" + singleContributors
				+ ", list=" + list + ", delta=" + delta + "]";
	}

	public Integer getSingleContributors() {
		return singleContributors;
	}

}
