package de.uni_mannheim.dws.expansion.data.ontology.model.objects;

import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class CoordinateObject extends AbstractObject implements Comparable<CoordinateObject> {

	private final double latitude;
	private final double longitude;
	private final double precision;

	public CoordinateObject(double latitude, double longitude, double precision) {
		this.latitude = latitude;
		this.longitude = longitude;
		this.precision = precision;
	}

	public double getLatitude() {
		return latitude;
	}

	public double getLongitude() {
		return longitude;
	}

	public double getPrecision() {
		return precision;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		long temp;
		temp = Double.doubleToLongBits(latitude);
		result = prime * result + (int) (temp ^ (temp >>> 32));
		temp = Double.doubleToLongBits(longitude);
		result = prime * result + (int) (temp ^ (temp >>> 32));
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		CoordinateObject other = (CoordinateObject) obj;
		if (Double.doubleToLongBits(latitude) != Double.doubleToLongBits(other.latitude))
			return false;
		if (Double.doubleToLongBits(longitude) != Double.doubleToLongBits(other.longitude))
			return false;
		return true;
	}

	@Override
	public String readableString() {
		return "" + QuantityObject.fmt(latitude) + "�, " + QuantityObject.fmt(longitude) + "�";
	}

	@Override
	public String toString() {
		return readableString();
	}

	@Override
	public ValueType getType() {
		return ValueType.COORDINATE;
	}

	@Override
	public int compareTo(CoordinateObject o) {
		double avgThis = (this.getLatitude() + this.getLongitude()) / 2;
		double avgO = (o.getLatitude() + o.getLongitude()) / 2;

		return Double.compare(avgThis, avgO);
	}

	public int compareToOld(CoordinateObject o) {

		if (this.getLatitude() > o.getLatitude()) {

			if (this.getLongitude() >= o.getLongitude()) {
				return 1;
			} else {
				double diff1 = this.getLatitude() - o.getLatitude();
				double diff2 = o.getLongitude() - this.getLongitude();
				if (diff1 > diff2) {
					return 1;
				} else {
					return -1;
				}

			}
		} else if (this.getLatitude() < o.getLongitude()) {
			if (this.getLongitude() <= o.getLongitude()) {
				return -1;
			} else {
				double diff1 = o.getLatitude() - this.getLatitude();
				double diff2 = this.getLongitude() - o.getLongitude();
				if (diff1 > diff2) {
					return -1;
				} else {
					return 1;
				}
			}
		} else {
			if (this.getLongitude() > o.getLongitude()) {
				return 1;
			} else if (this.getLongitude() < o.getLongitude()) {
				return -1;
			} else {
				return 0;
			}

		}
	}

	public static CoordinateObject averager(CoordinateObject value1, double score1, CoordinateObject value2, double score2) {
		double precision = Math.min(value1.getPrecision(), value2.getPrecision());
		double averageLatitude = ((value1.getLatitude() * score1) + (value2.getLatitude() * score2)) / (score1 + score2);
		double averageLongitude = ((value1.getLongitude() * score1) + (value2.getLongitude() * score2)) / (score1 + score2);

		return new CoordinateObject(averageLatitude, averageLongitude, precision);

	}

}
