package de.uni_mannheim.dws.expansion.data.ontology.model.objects;

import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class QuantityObject extends AbstractObject implements Comparable<QuantityObject> {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private final Integer exactInt;
	private final Double amount;
	private final Double upperBound;
	private final Double lowerBound;
	private final String unit;

	public QuantityObject(int exactInt, String unit) {
		this.exactInt = exactInt;
		this.amount = null;
		this.upperBound = null;
		this.lowerBound = null;
		this.unit = unit;
	}

	public QuantityObject(double amount, Double upperBound, Double lowerBound, String unit) {
		this.exactInt = null;
		this.amount = amount;
		this.upperBound = upperBound;
		this.lowerBound = lowerBound;
		this.unit = unit;
	}

	public boolean isExactInt() {
		if (exactInt != null)
			return true;
		return false;
	}

	public static QuantityObject fromExactInt(int exactInt) {
		return new QuantityObject(exactInt, null);
	}

	public static QuantityObject fromDouble(double amount) {
		return new QuantityObject(amount, null, null, null);
	}

	public double getValue() {
		if (isExactInt())
			return exactInt;
		return amount;
	}

	public int getExactInt() {
		return exactInt;
	}

	public double getAmount() {
		return amount;
	}

	public Double getUpperBound() {
		return upperBound;
	}

	public Double getLowerBound() {
		return lowerBound;
	}

	public String getUnit() {
		return unit;
	}

	@Override
	public String readableString() {
		if (isExactInt())
			return String.valueOf(exactInt);
		else
			return QuantityObject.fmt(amount);
	}

	@Override
	public int compareTo(QuantityObject o) {
		return Double.compare(this.getValue(), o.getValue());
	}

	public static QuantityObject averager(QuantityObject value1, double score1, QuantityObject value2, double score2) {
		double upperBound = Math.max(value1.upperBound, value2.upperBound);
		double lowerBound = Math.min(value1.lowerBound, value2.lowerBound);

		double average = ((value1.amount * score1) + (value2.amount * score2)) / (score1 + score2);

		if (value1.unit == value2.unit) {
			return new QuantityObject(average, upperBound, lowerBound, value1.unit);
		}
		throw new IllegalStateException("QUANTITIES HAVE DIFFERENT UNITS. WHAT DO WE DO?");

	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((amount == null) ? 0 : amount.hashCode());
		result = prime * result + ((exactInt == null) ? 0 : exactInt.hashCode());
		result = prime * result + ((lowerBound == null) ? 0 : lowerBound.hashCode());
		result = prime * result + ((unit == null) ? 0 : unit.hashCode());
		result = prime * result + ((upperBound == null) ? 0 : upperBound.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (!(obj instanceof QuantityObject))
			return false;
		QuantityObject other = (QuantityObject) obj;
		if (amount == null) {
			if (other.amount != null)
				return false;
		} else if (!amount.equals(other.amount))
			return false;
		if (exactInt == null) {
			if (other.exactInt != null)
				return false;
		} else if (!exactInt.equals(other.exactInt))
			return false;
		if (lowerBound == null) {
			if (other.lowerBound != null)
				return false;
		} else if (!lowerBound.equals(other.lowerBound))
			return false;
		if (unit == null) {
			if (other.unit != null)
				return false;
		} else if (!unit.equals(other.unit))
			return false;
		if (upperBound == null) {
			if (other.upperBound != null)
				return false;
		} else if (!upperBound.equals(other.upperBound))
			return false;
		return true;
	}

	@Override
	public String toString() {
		return this.readableString();
	}

	public static String fmt(double d) {
		if (d == (long) d)
			return String.format("%d", (long) d);
		else {
			String s = String.format("%f", d);
			s = s.indexOf(".") < 0 ? s : s.replaceAll("0*$", "").replaceAll("\\.$", "");
			return s;
		}
	}

	@Override
	public ValueType getType() {
		return ValueType.QUANTITY;
	}

}
