package de.uni_mannheim.dws.expansion.data.ontology.model.types;

import java.util.Collections;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableMap.Builder;
import com.google.common.collect.ImmutableSet;

import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;

public abstract class WikidataParsedObject {
	private final Map<PropertyId, QualifiedValueList> valuesMap;

	private final String label;
	private final Map<String, String> labels;
	private final Map<String, Set<String>> aliases;

	protected WikidataParsedObject(Map<PropertyId, QualifiedValueList> valuesMap, String label, Map<String, String> labels, Map<String, Set<String>> aliases) {

		this.valuesMap = Collections.unmodifiableMap(valuesMap);
		this.label = label;
		this.labels = ImmutableMap.copyOf(labels);
		Builder<String, Set<String>> bulder = ImmutableMap.builder();
		for (Entry<String, Set<String>> entry : aliases.entrySet()) {
			bulder.put(entry.getKey(), ImmutableSet.copyOf(entry.getValue()));
		}
		this.aliases = bulder.build();
	}

	public Map<PropertyId, QualifiedValueList> getValuesMap() {
		return valuesMap;
	}

	public QualifiedValueList getValues(PropertyId propertyId) {

		QualifiedValueList list = valuesMap.get(propertyId);
		if (list == null) {
			list = new QualifiedValueList();
		}
		return list;
	}

	public boolean hasValues(PropertyId propertyId) {
		return valuesMap.containsKey(propertyId);
	}

	public String getLabel() {
		return label;
	}

	public Map<String, String> getLabels() {
		return labels;
	}

	public Map<String, Set<String>> getAliases() {
		return aliases;
	}

}
