package de.uni_mannheim.dws.expansion.data.ontology.model.valueholders;

import java.io.Serializable;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;

import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;

public abstract class QualifiedValueHolder implements Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	public abstract AbstractObject getValue();

	private final List<Qualifier> qualifiers;
	private final Map<PropertyId, Integer> qualifierNameIndex;

	public QualifiedValueHolder(List<Qualifier> qualifiers) {

		this.qualifiers = ImmutableList.copyOf(qualifiers);
		this.qualifierNameIndex = ImmutableMap.copyOf(getQualifierNameIndex(this.qualifiers));
	}

	public QualifiedValueHolder() {

		this.qualifiers = null;
		this.qualifierNameIndex = null;
	}

	private Map<PropertyId, Integer> getQualifierNameIndex(List<Qualifier> qualifiers) {
		Map<PropertyId, Integer> index = new HashMap<>();

		for (int i = 0; i < qualifiers.size(); i++) {
			Qualifier q = qualifiers.get(i);
			if (index.containsKey(i)) {
				throw new IllegalStateException("This value has the same qualifier twice");
			}
			index.put(q.getQualifier(), i);
		}
		return index;

	}

	public boolean hasQualifier(PropertyId qualifier) {
		if (qualifierNameIndex != null)
			return qualifierNameIndex.containsKey(qualifier);
		else
			return false;
	}

	public List<Qualifier> getQualifiers() {
		return qualifiers;
	}

	public Set<PropertyId> getQualifierIds() {
		if (this.qualifierNameIndex != null)
			return this.qualifierNameIndex.keySet();
		else
			return new HashSet<>();
	}

	public Qualifier getQualifierValue(PropertyId propertyId) {
		if (this.qualifierNameIndex != null) {
			Integer index = this.qualifierNameIndex.get(propertyId);
			if (index == null)
				return null;
			else
				return qualifiers.get(index);
		}
		return null;
	}

	@Override
	public String toString() {
		return "QualifiedValueHolder [qualifiers=" + qualifiers + ", qualifierNameIndex=" + qualifierNameIndex + ", value=" + getValue() + "]";
	}

	public boolean hasQualifiers() {
		return this.qualifiers != null;
	}

}
