package de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables;

import java.util.Iterator;
import java.util.List;
import java.util.function.BinaryOperator;

public class AccumAverage implements Accumulateable<AccumAverage> {

	private final AccumLinkedList<Double> values;

	public AccumAverage(AccumAverage a, AccumAverage b) {
		this.values = new AccumLinkedList<>(a.values, b.values);

	}

	public AccumAverage() {
		this.values = new AccumLinkedList<Double>();
	}

	public AccumAverage(Double value) {
		this.values = new AccumLinkedList<Double>(value);
	}

	// public AccumAverage(AccumLinkedList<Double> values) {
	// this.values = values.copy();
	// }

	@Override
	public BinaryOperator<AccumAverage> getAccumulator() {
		return AccumAverage::accum;
	}

	public static AccumAverage accum(AccumAverage a, AccumAverage b) {
		return new AccumAverage(a, b);
	}

	@Override
	public AccumAverage accumulate(AccumAverage other) {
		if (other instanceof AccumAverage) {
			AccumAverage oCast = (AccumAverage) other;
			return new AccumAverage(this, oCast);
		} else {
			throw new IllegalStateException("This is not the same Accumateable Class");
		}
	}

	public Double get() {
		return getValue();
	}

	public Double getValue() {

		List<Double> valuesCopy = values.getList();

		double sum = 0.0;
		int count = 0;
		for (Iterator<Double> $iter = valuesCopy.iterator(); $iter.hasNext();) {
			Double d = $iter.next();
			sum += d;
			count++;
		}

		if (count == 0) {
			return Double.NaN;
		} else {
			double average = sum / count;
			return average;
		}

	}

	// @Override
	// public AccumAverage copy() {
	// return new AccumAverage(values);
	// }

	@Override
	public Class<AccumAverage> getAcumClass() {
		return AccumAverage.class;
	}

	public static AccumAverage empty() {
		return new AccumAverage();
	}

}
