package de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables;

import java.util.Collections;
import java.util.List;
import java.util.function.BinaryOperator;

public class AccumMedian implements Accumulateable<AccumMedian> {

	private final AccumLinkedList<Double> values;

	public AccumMedian(AccumMedian a, AccumMedian b) {
		values = new AccumLinkedList<>(a.values, b.values);
	}

	public AccumMedian() {
		values = new AccumLinkedList<>();
	}

	public AccumMedian(Double value) {
		values = new AccumLinkedList<Double>(value);
	}

	// private AccumMedian(AccumLinkedList<Double> values) {
	// this.values = values;
	// }

	@Override
	public BinaryOperator<AccumMedian> getAccumulator() {
		return AccumMedian::accum;
	}

	public static AccumMedian accum(AccumMedian a, AccumMedian b) {
		return new AccumMedian(a, b);
	}

	@Override
	public AccumMedian accumulate(AccumMedian other) {
		if (other instanceof AccumMedian) {
			AccumMedian oCast = (AccumMedian) other;
			return new AccumMedian(this, oCast);
		} else {
			throw new IllegalStateException("This is not the same Accumateable Class");
		}
	}

	public Double getValue() {

		List<Double> valuesCopy = values.getList();

		if (valuesCopy.size() == 0)
			return Double.NaN;

		Collections.sort(valuesCopy);
		int size = valuesCopy.size();
		if (size % 2 == 0) {
			int half = size / 2;
			int secondHalf = half - 1;

			return valuesCopy.get(half) + valuesCopy.get(secondHalf) / 2.0;
		} else {
			int half = size / 2;
			return valuesCopy.get(half);
		}
	}

	// @Override
	// public AccumMedian copy() {
	// return new AccumMedian(this.values);
	// }

	@Override
	public Class<AccumMedian> getAcumClass() {
		return AccumMedian.class;
	}

	public static AccumMedian empty() {
		return new AccumMedian();
	}

	public Double get() {
		return getValue();
	}

}
