package de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.UUID;

import com.fasterxml.uuid.Generators;

import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.Accumulateable;

public class AccumulatorListImmutable implements AccumulatorList {

	private final UUID uuid = Generators.timeBasedGenerator().generate();

	private final List<Class<? extends Accumulateable<?>>> accumulators;
	final private List<Accumulateable<?>> emptyAccumulators;

	public AccumulatorListImmutable(List<Class<? extends Accumulateable<?>>> accumulators, List<Accumulateable<?>> emptyAccumulators) {
		this.accumulators = Collections.unmodifiableList(new ArrayList<>(accumulators));
		this.emptyAccumulators = Collections.unmodifiableList(new ArrayList<>(emptyAccumulators));
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((uuid == null) ? 0 : uuid.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		AccumulatorListImmutable other = (AccumulatorListImmutable) obj;
		if (uuid == null) {
			if (other.uuid != null)
				return false;
		} else if (!uuid.equals(other.uuid))
			return false;
		return true;
	}

	public static ImmutableAccumulatorListBuild builder() {
		return new ImmutableAccumulatorListBuild();
	}

	public int getAccumulatorsSize() {
		return accumulators.size();
	}

	public Class<? extends Accumulateable<?>> getType(int key) {
		return accumulators.get(key);
	}

	public Accumulateable<?> getEmpty(int key) {
		return emptyAccumulators.get(key);
	}

	public static class ImmutableAccumulatorListBuild {

		private ImmutableAccumulatorListBuild() {

		}

		private final List<Class<? extends Accumulateable<?>>> accumulators = new ArrayList<>();
		private final List<Accumulateable<?>> emptyAccumulators = new ArrayList<>();

		public <E extends Accumulateable<E>> AcumReference<E> addAcummulator(Class<E> type, E empty) {
			accumulators.add(type);
			emptyAccumulators.add(empty);
			return new AcumReference<>(accumulators.size() - 1);

		}

		public <E extends Accumulateable<E>> AcumReference<E> add(E empty) {
			accumulators.add(empty.getAcumClass());
			emptyAccumulators.add(empty);
			return new AcumReference<>(accumulators.size() - 1);

		}

		public AccumulatorListImmutable getResultsType() {
			return new AccumulatorListImmutable(accumulators, emptyAccumulators);
		}
	}

}