package de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.UUID;

import com.fasterxml.uuid.Generators;

import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.Accumulateable;

public class AccumulatorListSimple implements AccumulatorList {

	private final UUID uuid = Generators.timeBasedGenerator().generate();

	private final List<Class<? extends Accumulateable<?>>> accumulators;
	final private List<Accumulateable<?>> emptyAccumulators;

	public AccumulatorListSimple() {
		this.accumulators = Collections.synchronizedList(new ArrayList<>());
		this.emptyAccumulators = Collections.synchronizedList(new ArrayList<>());
	}

	public int getAccumulatorsSize() {
		return Math.min(accumulators.size(), emptyAccumulators.size());
	}

	public Class<? extends Accumulateable<?>> getType(int key) {
		return accumulators.get(key);
	}

	public Accumulateable<?> getEmpty(int key) {
		return emptyAccumulators.get(key);
	}

	public <E extends Accumulateable<E>> AcumReference<E> addAcummulator(Class<E> type, E empty) {
		accumulators.add(type);
		emptyAccumulators.add(empty);
		return new AcumReference<>(accumulators.size() - 1);

	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((uuid == null) ? 0 : uuid.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		AccumulatorListSimple other = (AccumulatorListSimple) obj;
		if (uuid == null) {
			if (other.uuid != null)
				return false;
		} else if (!uuid.equals(other.uuid))
			return false;
		return true;
	}

}
