package de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.UUID;

import com.fasterxml.uuid.Generators;

import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.Accumulateable;

public class ColumnList<X> {

	private final UUID uuid = Generators.timeBasedGenerator().generate();

	private final List<String> columnHeaders;
	private final List<ColumnList.ColumnAcumValueGeneric<X>> columnValueGetters;

	public ColumnList(List<String> columnHeaders, List<ColumnList.ColumnAcumValueGeneric<X>> columnValueGetters) {
		this.columnHeaders = Collections.unmodifiableList(new ArrayList<>(columnHeaders));
		this.columnValueGetters = Collections.unmodifiableList(new ArrayList<>(columnValueGetters));
	}

	public int getColumnSize() {
		return columnHeaders.size();
	}

	public List<String> getColumnHeaders() {
		return columnHeaders;
	}

	public List<ColumnAcumValueGeneric<X>> getColumnValueGetters() {
		return columnValueGetters;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((uuid == null) ? 0 : uuid.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (!(obj instanceof ColumnList))
			return false;
		ColumnList other = (ColumnList) obj;
		if (uuid == null) {
			if (other.uuid != null)
				return false;
		} else if (!uuid.equals(other.uuid))
			return false;
		return true;
	}

	public static class ColumnListBuilder extends ColumnListBuilderGeneric<AccumRepository> {

	}

	public static class ColumnListBuilderGeneric<F> {

		private ColumnListBuilderGeneric() {

		}

		private final List<String> columnHeaders = new ArrayList<>();
		private final List<ColumnList.ColumnAcumValueGeneric<F>> columnValueGetters = new ArrayList<>();

		public <E extends Accumulateable<E>> void addColumn(String columnHeader, ColumnList.ColumnAcumValueGeneric<F> valueGetter) {
			columnHeaders.add(columnHeader);
			columnValueGetters.add(valueGetter);

		}

		public ColumnList<F> build() {
			return new ColumnList<>(columnHeaders, columnValueGetters);
		}
	}

	@FunctionalInterface
	public static interface ColumnAcumValue extends ColumnAcumValueGeneric<AccumRepository> {
	}

	@FunctionalInterface
	public static interface ColumnAcumValueGeneric<E> {
		String getValue(E repository);
	}

	public static ColumnListBuilder builderAccumRepository() {
		return new ColumnListBuilder();
	}

	public static <E> ColumnListBuilderGeneric<E> builder() {
		return new ColumnListBuilderGeneric<>();
	}
}
