package de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators;

import java.util.List;

import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList.ColumnAcumValueGeneric;

public class ColumnListDirectOutputter<X> {

	private final ColumnList<X> list;
	public ColumnListDirectOutputter(ColumnList<X> list) {
		this.list = list;
	}
	
	
	protected String output(X content, List<Pair<String, String>> prefix, boolean outputHeader) {
		StringBuilder sb = new StringBuilder();
		
		int size = list.getColumnSize();
		int prefixSize = 0;
		if(prefix != null)
			prefixSize = prefix.size();
		
		if(outputHeader) {
			if(prefix != null) {
				for(ListEntry<Pair<String, String>> header : ListIterable.iterable(prefix)) {
					sb.append(header.getValue().getFirst()); 
					sb.append(";"); 
				}
			}
			
			
			for(ListEntry<String> header : ListIterable.iterable(list.getColumnHeaders())) {
				sb.append(header.getValue());
				if(header.getIndex() < (size - 1)) {
					sb.append(";");
				}
			}
		}
		
		sb.append("\n");
		
		if(prefix != null) {
			for(ListEntry<Pair<String, String>> prefixEntry : ListIterable.iterable(prefix)) {
				sb.append(prefixEntry.getValue().getSecond()); 
				sb.append(";"); 
			}
		}
		
		for(ListEntry<ColumnAcumValueGeneric<X>> getter : ListIterable.iterable(list.getColumnValueGetters())) {
			sb.append(getter.getValue().getValue(content));
			if(getter.getIndex() < (size - 1)) {
				sb.append(";");
			}
		}
		
		return sb.toString();
	}
	
	public String outputWithHeader(X content) {
		return output(content, null, true);
	}
	
	public String outputWithoutHeader(X content) {
		return output(content, null, false);
	}
}
