package de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators;

import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.function.BinaryOperator;

import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.Accumulateable;
import de.uni_mannheim.informatik.dws.expansion.utils.MutableReference;

public class ConcurrentAccumulator<E extends Accumulateable<E>> {

	private final ThreadLocal<MutableReference<E>> threadLocalVal;

	private final BinaryOperator<E> accumulator;

	private final ConcurrentLinkedQueue<MutableReference<E>> references;
	private final E empty;

	public ConcurrentAccumulator(E empty) {
		this.empty = empty;
		threadLocalVal = new ThreadLocal<>();
		this.accumulator = empty.getAccumulator();
		this.references = new ConcurrentLinkedQueue<>();
	}

	private MutableReference<E> getReference() {
		MutableReference<E> reference = threadLocalVal.get();
		if (reference == null) {
			reference = new MutableReference<E>(empty);
			references.add(reference);
			threadLocalVal.set(reference);
		}
		return reference;
	}

	public void add(E val) {
		getReference().accumulate(val, accumulator);
	}

	public E get() {
		MutableReference<E> ret = new MutableReference<E>(empty);
		for (MutableReference<E> ref : references) {
			ret.accumulate(ref.get(), accumulator);
		}
		return ret.get();
	}

	@Override
	public String toString() {
		return get().toString();
	}

}
