package de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators;

import java.util.HashMap;
import java.util.Map;
import java.util.function.BinaryOperator;

import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.Accumulateable;

public class GenericAccumaltableResult implements Accumulateable<GenericAccumaltableResult> {

	private final Map<Integer, Accumulateable<?>> result = new HashMap<>();
	private final AccumulatorList accumulatorList;
	
	@SuppressWarnings("unchecked")
	private <T extends Accumulateable<T>> void captureAccumulateInner(GenericAccumaltableResult a, GenericAccumaltableResult b,  int index, Class<T> type) {
		T aResult =   (T) a.result.get(index);
		T bResult =   (T) b.result.get(index);
		 
		if (aResult != null) {
			if (bResult != null) {
				result.put(index, aResult.accumulate(bResult));
			} else {
				result.put(index, aResult);
			}
		} else if (bResult != null) {
			result.put(index, bResult);
		} else {
			result.put(index, this.accumulatorList.getEmpty(index));
		}
	}
	
	 @SuppressWarnings("unchecked")
	private <T extends Accumulateable<T>> void captureAccumulate(GenericAccumaltableResult a, GenericAccumaltableResult b,  int index, Class<? extends Accumulateable<?>> type) {
		 captureAccumulateInner(a,b,index,   (Class<T>) type);
		
	}

	private GenericAccumaltableResult(GenericAccumaltableResult a, GenericAccumaltableResult b) {
		if (a.accumulatorList.equals(b.accumulatorList)) {
			this.accumulatorList = a.accumulatorList;

			for (int i = 0; i < this.accumulatorList.getAccumulatorsSize(); i++) {
				Class<? extends Accumulateable<?>> type = accumulatorList.getType(i);
 
				captureAccumulate(a,b,i,type);
				 

			}

		} else {
			throw new IllegalStateException("These Results have differing ResultTypes");
		}

	}
	


	<E extends Accumulateable<E>> void cast(Class<E> clazz) {
		
	}

	private GenericAccumaltableResult(AccumulatorList resultsType, Map<Integer, Accumulateable<?>> result) {
		this.accumulatorList = resultsType;
		this.result.putAll(result);
	}

	public GenericAccumaltableResult(AccumulatorList resultsType) {
		this.accumulatorList = resultsType;
	}

	public String getValue(Integer key) {
		Accumulateable<?> accum = result.get(key);
		if (accum == null) {
			return "";
		} else {
			return accum.toString();
		}
	}

	@Override
	public BinaryOperator<GenericAccumaltableResult> getAccumulator() {
		return accumulator;
	}

	private static final BinaryOperator<GenericAccumaltableResult> accumulator = (a, b) -> {
		return new GenericAccumaltableResult(a, b);
	};

	public static ResultBuilder builder(AccumulatorList resultsType) {
		return new ResultBuilder(resultsType);
	}

	public static class ResultBuilder {
		private Map<Integer, Accumulateable<?>> result = new HashMap<>();
		private final AccumulatorList resultsType;

		public ResultBuilder(AccumulatorList resultsType) {
			this.resultsType = resultsType;
		}

		public <E extends Accumulateable<E>> void setValue(AcumReference<E> key, E value) {
			if (value.getClass().equals(resultsType.getType(key.getKey()))) {
				result.put(key.getKey(), value);
			} else {
				throw new IllegalStateException("This column has a different type of Accumulateable");
			}

		}

		public GenericAccumaltableResult build() {
			return new GenericAccumaltableResult(resultsType, this.result);
		}
	}

	public static GenericAccumaltableResult empty(AccumulatorList resultsType) {
		return new GenericAccumaltableResult(resultsType);
	}

	@Override
	public GenericAccumaltableResult accumulate(GenericAccumaltableResult other) {
		if (other instanceof GenericAccumaltableResult) {
			GenericAccumaltableResult oCast = (GenericAccumaltableResult) other;
			return new GenericAccumaltableResult(this, oCast);
		} else {
			throw new IllegalStateException("This is not the same Accumateable Class");
		}
	}

	public AccumRepository getAcummulatorRepository() {
		return new GenericAccumRepository(this.result, this.accumulatorList);
	}

	// @Override
	// public GenericAccumaltableResult copy() {
	// Map<Integer, Accumulateable<?>> resultsCopy = new HashMap<>();
	// for (Entry<Integer, Accumulateable<?>> r : result.entrySet()) {
	// resultsCopy.put(r.getKey(), (Accumulateable<?>) r.getValue().copy());
	// }
	// return new GenericAccumaltableResult(accumulatorList, resultsCopy);
	// }

	public class GenericAccumRepository implements AccumRepository {
		private final Map<Integer, Accumulateable<?>> values;
		private final AccumulatorList resultsType;

		public GenericAccumRepository(Map<Integer, Accumulateable<?>> values, AccumulatorList resultsType) {
			this.values = values;
			this.resultsType = resultsType;
		}

		@SuppressWarnings("unchecked")
		public <E extends Accumulateable<E>> E get(AcumReference<E> reference) {
			if (this.values.containsKey(reference.getKey())) {
				return (E) this.values.get(reference.getKey());
			} else {
				return (E) resultsType.getEmpty(reference.getKey());
			}
		}
	}

	@Override
	public Class<GenericAccumaltableResult> getAcumClass() {
		return GenericAccumaltableResult.class;
	}
}
