package de.uni_mannheim.informatik.dws.expansion.samples.dbpedia;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;

import org.apache.commons.collections4.BidiMap;
import org.apache.commons.collections4.bidimap.DualHashBidiMap;

import dagger.Module;
import dagger.ObjectGraph;
import de.mannheim.uni.expansion.data.dbpedia.adapter.csv.DBpediaRepositoryCSVAdapter;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId.FileEnding;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.reader.XMLOntologyReader;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyAdapterRunner;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyConverter;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepositoryImpl;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepositoryImplMemory;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.matching.dbpedia.algorithm.KeyIndex;
import de.mannheim.uni.tableprocessor.IO.TableReader;
import de.mannheim.uni.tableprocessor.model.Table;
import de.mannheim.uni.tableprocessor.model.TableColumn;
import de.mannheim.uni.tableprocessor.model.TableColumn.ColumnDataType;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreatorsRunner;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBaseCryoAdapter;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBaseImpl;
import de.uni_mannheim.dws.expansion.data.ontology.model.ValueConverter;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeKryo;
import de.uni_mannheim.informatik.dws.expansion.matching.KnowledgeBaseModule;

public class ConvertDBpedia {

	@Module(injects = { ConvertDBpedia.class }, library = true, complete = false)
	public static class ConvertModule {

	}

	public static void main(String[] args) throws ClassNotFoundException, IOException, InterruptedException {
		String rawHome = null;
		String ontologyRaw = null;
		String targetHome = null;
		String labelsFolder = null;

		if (args.length > 0 && args[0] != null && !args[0].equals("")) {
			rawHome = args[0] + "/";
		}

		if (args.length > 1 && args[1] != null && !args[1].equals("")) {
			ontologyRaw = args[1];
		}

		if (args.length > 2 && args[2] != null && !args[2].equals("")) {
			targetHome = args[2] + "/";
		}

		if (args.length > 3 && args[3] != null && !args[3].equals("")) {
			labelsFolder = args[3] + "/";
		}

		if (rawHome == null) {
			throw new IllegalStateException("Folder with Raw Data is Empty");
		}

		if (ontologyRaw == null) {
			throw new IllegalStateException("Ontolgy XML file path is missing");
		}

		if (targetHome == null) {
			throw new IllegalStateException("Folder for target Data is Empty");
		}

		ObjectGraph graph = ObjectGraph.create(ConvertModule.class);

		ConvertDBpedia converter = graph.get(ConvertDBpedia.class);

		converter.run(rawHome, ontologyRaw, targetHome, labelsFolder);

		PersistentParallel6.close();

	}

	@Inject
	protected KnowledgeBaseCryoAdapter kryoAdapter;

	@Inject
	protected DBpediaRepositoryCSVAdapter csvAdapter;

	@Inject
	protected XMLOntologyReader ontologyReader;

	@Inject
	protected ThreadSafeKryo threadSafeKryo;

	@Inject
	protected OntologyConverter ontologyConverter;

	@Inject
	protected KBValueCreatorsRunner kbValueCreatorsRunner;

	@Inject
	protected OntologyAdapterRunner ontologyAdapterRunner;

	public ConvertDBpedia() {

	}

	public void run(String csvDBpedia, String ontologyOwlFile, String dbPediaFolder, String labelsFolder)
			throws ClassNotFoundException, IOException, InterruptedException {
		MKDIR.createDirUnchecked(new File(dbPediaFolder));
		MKDIR.createDirUnchecked(new File(dbPediaFolder + "/kb"));
		KnowledgeBaseImpl kb = (KnowledgeBaseImpl) csvAdapter.read(csvDBpedia, true);

		if (false) {
			KbLabelRenamer kbLabelRenamer;
			if (labelsFolder != null) {
				Set<String> files = new HashSet<>();
				files.add(labelsFolder + "mappingbased_properties_en.nt.gz");
				files.add(labelsFolder + "labels_en.nt.gz");
				Set<String> redirectFiles = new HashSet<>();
				redirectFiles.add(labelsFolder + "redirects_en.nt.gz");
				Set<String> surfaceFormFiles = new HashSet<>();
				surfaceFormFiles.add(labelsFolder + "SurfaceForms_LRD-CC_filter_thr3.80.tsv.gz");
				surfaceFormFiles.add(labelsFolder + "SurfaceForms_LRD-WAT_filter_thr2.60.tsv.gz");
				kbLabelRenamer = new KbLabelRenamer(files, redirectFiles, surfaceFormFiles);
			} else {
				kbLabelRenamer = new KbLabelRenamer();
			}
			List<KBValueCreator> valueCreatorList = new ArrayList<>();
			valueCreatorList.add(kbLabelRenamer);
			kb = kbValueCreatorsRunner.runReplacers(kb, valueCreatorList);
		}

		Map<ClassPropertyId, ColumnDataType> columnDataTypes = ConvertOntology.readColumnDataTypes(csvDBpedia + "/");

		OntologyRepositoryImpl ontologyRepository = ontologyConverter.convert(kb, ontologyReader.run(ontologyOwlFile), columnDataTypes);

		// List<OntologyAdapter> ontologyAdapters = new ArrayList<>();
		// ontologyAdapters.add(new MusicalBandOntologyAdapter());
		// ontologyRepository = ontologyAdapterRunner.run(ontologyRepository,
		// ontologyAdapters);

		// valueCreatorList = new ArrayList<>();
		// valueCreatorList.add(new
		// KbAbstractObjectInstanceConverter(ontologyRepository));
		// valueCreatorList.add(new PropertiesDBpediaTableFilter());
		// kb = kbValueCreatorsRunner.runReplacers(kb, valueCreatorList);

		System.out.println(kb.getClassPropertyIds().toString());

		kryoAdapter.write(dbPediaFolder + "kb", kb);

		threadSafeKryo.write(ontologyRepository, dbPediaFolder + "ontology.kryo");

		String kbPath = dbPediaFolder + "/kb/";
		String ontologyRepositoryFile = dbPediaFolder + "/ontology.kryo";
		String luceneIndexCache = dbPediaFolder + "/lucene/";

		KnowledgeBaseModule kbModule = new KnowledgeBaseModule(kbPath, ontologyRepositoryFile, luceneIndexCache);

		ObjectGraph graph = ObjectGraph.create(kbModule);

		KeyIndex keyIndex = graph.get(KeyIndex.class);

	}

	public static Pair<KnowledgeBase, Map<ClassPropertyId, ColumnDataType>> loadFromCSVfiles(String dbpediaAsTables,
			OntologyRepositoryImplMemory ontologyPreliminary) {
		final FileEnding fileEnding = FileEnding.CSV_GZ;
		final ValueConverter valueConverter = new ValueConverter();

		TableReader r = new TableReader();
		r.setUseUnitDetection(false);

		ArrayList<Table> tables = new ArrayList<Table>();

		List<File> files = new LinkedList<File>();

		if (new File(dbpediaAsTables).isDirectory()) {
			for (File f : new File(dbpediaAsTables).listFiles()) {
				files.add(f);
			}
		} else {
			files.add(new File(dbpediaAsTables));
		}

		for (File f : files) {
			try {
				System.out.println(f.getName());
				Table t = r.readLODTable(f.getAbsolutePath());
				tables.add(t);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		//

		final ConcurrentHashMap<InstanceId, Set<ClassId>> setMappings = new ConcurrentHashMap<>();

		final Map<ClassId, Integer> tableIndex = new ConcurrentHashMap<ClassId, Integer>();

		final Map<ClassId, BidiMap<PropertyId, Integer>> propertyIndices = new ConcurrentHashMap<ClassId, BidiMap<PropertyId, Integer>>();

		final Map<ClassId, BidiMap<InstanceId, Integer>> instanceIndices = new ConcurrentHashMap<ClassId, BidiMap<InstanceId, Integer>>();

		final ConcurrentHashMap<ValueId, QualifiedValueList> values = new ConcurrentHashMap<>();

		final Map<ClassPropertyId, ColumnDataType> columnDataTypes = new ConcurrentHashMap<ClassPropertyId, ColumnDataType>();
		final Map<ClassId, PropertyId> classKeys = new ConcurrentHashMap<ClassId, PropertyId>();

		int tCounter = 0;

		for (Table t : tables) {

			try {

				final int tCounterFinal = tCounter;

				TableColumn key = null;

				for (TableColumn tc : t.getColumns()) {
					if (tc.getHeader().equals("URI")) {
						key = tc;
						break;
					}
				}

				if (key != null) {

					ClassId dbPediaClass = ClassId.fromString(t.getHeader(), fileEnding);

					tableIndex.put(dbPediaClass, tCounter);

					BidiMap<PropertyId, Integer> columns = new DualHashBidiMap<PropertyId, Integer>();
					int c = 0;
					for (TableColumn column : t.getColumns()) {

						PropertyId dbPediaProperty = new PropertyId(column.getURI());
						if (dbPediaProperty.getPropertyId().equals("http://www.w3.org/2000/01/rdf-schema#label")) {
							dbPediaProperty = PropertiesExpansionData.LABEL_PROPERTY;
						}
						columns.put(dbPediaProperty, c);

						if (column.getHeader().equals("URI")) {
							classKeys.put(dbPediaClass, dbPediaProperty);
						}

						c++;
					}

					propertyIndices.put(dbPediaClass, columns);

					BidiMap<InstanceId, Integer> instances = new DualHashBidiMap<InstanceId, Integer>();

					for (Entry<Integer, Object> instance : key.getValues().entrySet()) {
						InstanceId dbPediaInstance = new InstanceId(instance.getValue().toString());

						ValueId valueId = new ValueId(dbPediaInstance, PropertiesExpansionData.LABEL_PROPERTY);
						values.put(valueId, valueConverter.getQVL(dbPediaInstance.getReadableString(), false));

						instances.put(dbPediaInstance, instance.getKey());
						CHMHelper.getFromCHM(setMappings, dbPediaInstance, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(dbPediaClass);
					}

					instanceIndices.put(dbPediaClass, instances);

					int columnSize = t.getColumns().size();

					PersistentParallel6.runAndFinish(t.getColumns(), (column) -> {
						ColumnDataType type = column.getDataType();
						PropertyId propertyId = new PropertyId(column.getURI());

						boolean isReference = ontologyPreliminary.isReference(propertyId);
						if (isReference && type == ColumnDataType.link) {
							type = ColumnDataType.reference;

						} else if (isReference) {

							if (column.toString().endsWith("_label")) {
								return;
							} else {
								throw new IllegalStateException();
							}
						}

						if (propertyId.getPropertyId().equals("http://www.w3.org/2000/01/rdf-schema#label")) {
							propertyId = PropertiesExpansionData.LABEL_PROPERTY;
						}
						columnDataTypes.put(new ClassPropertyId(dbPediaClass, propertyId), column.getDataType());

						for (Entry<Integer, Object> e : column.getValues().entrySet()) {

							int instanceCounterFinal = e.getKey();

							InstanceId dbPediaInstance = instances.getKey(instanceCounterFinal);

							Object value = e.getValue();

							if (value != null && dbPediaInstance != null) {

								ValueId valueId = new ValueId(dbPediaInstance, propertyId);

								values.put(valueId, valueConverter.getQVL(value, type == ColumnDataType.reference));

								if (column.getDataType() == ColumnDataType.list) {
									column.setDataType(ColumnDataType.string);
								}

								// valueExists.add(new
								// TripleId(dbPediaClass, propertyId,
								// dbPediaInstance));
							}

						}
					}, ConvertDBpedia.class);

				} else {
					System.out.println("WARNING: Table without key");
				}

				tCounter++;

			} catch (Exception e) {
				e.printStackTrace();
			}

		}

		System.gc();

		System.out.println("Loading DBpedia ... done");
		System.err.println("Loading DBpedia ... done");
		KnowledgeBase kb = KnowledgeBaseImpl.create(values, setMappings, new HashMap<>(), new HashMap<>(), new HashMap<>());

		return new Pair<>(kb, columnDataTypes);

	}

}
