package de.uni_mannheim.informatik.dws.expansion.samples.dbpedia;

import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import javax.inject.Inject;

import com.opencsv.CSVReader;

import dagger.Module;
import dagger.ObjectGraph;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.reader.XMLOntologyReader;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyConverter;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepositoryImpl;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.tableprocessor.model.TableColumn.ColumnDataType;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeKryo;
import de.uni_mannheim.informatik.dws.expansion.matching.KnowledgeBaseModule;

public class ConvertOntology {

	@Module(injects = { ConvertOntology.class }, library = true, complete = false)
	public static class ConvertModule {

	}

	public static void main(String[] args) throws ClassNotFoundException, IOException, InterruptedException {
		String rawHome = null;
		String ontologyRaw = null;
		String targetHome = null;

		if (args.length > 0 && args[0] != null && !args[0].equals("")) {
			rawHome = args[0] + "/";
		}

		if (args.length > 1 && args[1] != null && !args[1].equals("")) {
			ontologyRaw = args[1];
		}

		if (args.length > 2 && args[2] != null && !args[2].equals("")) {
			targetHome = args[2] + "/";
		}

		if (rawHome == null) {
			throw new IllegalStateException("Folder with Raw Data is Empty");
		}

		if (ontologyRaw == null) {
			throw new IllegalStateException("Ontolgy XML file path is missing");
		}

		if (targetHome == null) {
			throw new IllegalStateException("Folder for target Data is Empty");
		}

		ObjectGraph graph = ObjectGraph.create(ConvertModule.class, new KnowledgeBaseModule(targetHome + "/kb", null, null));

		ConvertOntology converter = graph.get(ConvertOntology.class);

		converter.run(rawHome, ontologyRaw, targetHome);

		PersistentParallel6.close();

	}

	@Inject
	protected XMLOntologyReader ontologyReader;

	@Inject
	protected ThreadSafeKryo threadSafeKryo;

	@Inject
	protected OntologyConverter ontologyConverter;

	@Inject
	protected KnowledgeBase kb;

	public ConvertOntology() {

	}

	public void run(String csvDBpedia, String ontologyOwlFile, String dbPediaFolder) throws ClassNotFoundException, IOException, InterruptedException {

		Map<ClassPropertyId, ColumnDataType> columnDataTypes = readColumnDataTypes(csvDBpedia + "/");

		OntologyRepositoryImpl ontologyRepository = ontologyConverter.convert(kb, ontologyReader.run(ontologyOwlFile), columnDataTypes);

		threadSafeKryo.write(ontologyRepository, dbPediaFolder + "ontology.kryo");

		System.out.println(ontologyRepository.getPropertyParents());

	}

	static Map<ClassPropertyId, ColumnDataType> readColumnDataTypes(String folder) throws IOException {

		Map<String, ColumnDataType> columnDataTypesMapping = new HashMap<String, ColumnDataType>();

		for (ColumnDataType e : ColumnDataType.values()) {
			columnDataTypesMapping.put(e.name(), e);
		}

		Map<ClassPropertyId, ColumnDataType> columnDataTypes = new HashMap<>();

		CSVReader reader = new CSVReader(new FileReader(folder + "columnDataTypes"), ';', '"');
		for (String[] line = reader.readNext(); line != null; line = reader.readNext()) {
			ClassId classId = new ClassId(line[0]);
			PropertyId propertyId = new PropertyId(line[1]);
			ClassPropertyId classPropertyId = new ClassPropertyId(classId, propertyId);

			ColumnDataType columnDataType = columnDataTypesMapping.get(line[2]);

			columnDataTypes.put(classPropertyId, columnDataType);
		}
		reader.close();

		return columnDataTypes;

	}

}
