package de.uni_mannheim.informatik.dws.expansion.samples.dbpedia;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.inject.Inject;

import dagger.Module;
import dagger.ObjectGraph;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTablesWriter;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.modules.HeidelTimeModule;
import de.mannheim.uni.expansion.tablefusion.modules.MatchingModule;
import de.mannheim.uni.expansion.tablefusion.modules.SimilarityModule;
import de.mannheim.uni.expansion.tablefusion.modules.TablesModule;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.units.UnitParser;
import de.uni_mannheim.informatik.dws.expansion.matching.KnowledgeBaseModule;
import de.uni_mannheim.informatik.dws.expansion.samples.setExtension.ClassConstants;

public class FilterCachedTables {

	@Module(injects = { FilterCachedTables.class }, library = true, complete = false)
	public static class TablesConverterModule {

	}

	public static void main(String[] args) throws ClassNotFoundException, IOException, InterruptedException {
		String dataHomeTemp = null;

		if (args.length > 0 && args[0] != null && !args[0].equals("")) {
			dataHomeTemp = args[0];
		}

		String kbName = null;

		if (args.length > 1 && args[1] != null && !args[1].equals("")) {
			kbName = args[1];
		}

		String mappingsName = null;

		if (args.length > 2 && args[2] != null && !args[2].equals("")) {
			mappingsName = args[2];
		}
		String dataHome = dataHomeTemp;

		String mappingsTargetName = null;

		if (args.length > 3 && args[3] != null && !args[3].equals("")) {
			mappingsTargetName = args[3];
		}

		mappingsName = "/tables/" + mappingsName;

		mappingsTargetName = "/tables/" + mappingsTargetName;

		String kbPath = dataHome + "/kbs/" + kbName + "/kb/";
		String ontologyRepositoryFile = dataHome + "/kbs/" + kbName + "/ontology.kryo";
		String luceneIndexCache = dataHome + "/kbs/" + kbName + "/lucene/";

		String mappingsFolder = dataHome + "/" + mappingsName + "/mappings/";
		String cacheInputFolder = dataHome + "/" + mappingsName + "/cache/";
		String cacheOutputFolder = dataHome + "/" + mappingsTargetName + "/cache/";
		String resultsFolder = dataHome + "/" + mappingsName + "/results/";
		String problemsFolder = dataHome + "/" + mappingsName + "/problems/";

		String optimizingFolder = dataHome + "/" + mappingsName + "/optimizing/";

		String columnMappingsFolder = dataHome + "/" + mappingsTargetName + "/columnMappings/";
		String heidelTimeConfigPath = dataHome + "/heidelTime/config.props";
		String emptyFile = dataHome + "/empty.txt";
		String candidateSelectionConfig = dataHome + "/config.conf";
		String matchingConfig = dataHome + "/config.conf";
		UnitParser.unitPath = dataHome + "/Units/Convertible";

		SimilarityModule similarityModule = new SimilarityModule();
		MatchingModule matchingModule = new MatchingModule(true, columnMappingsFolder, candidateSelectionConfig, matchingConfig, emptyFile);
		HeidelTimeModule heidelTimeModule = new HeidelTimeModule(heidelTimeConfigPath);
		TablesModule tableModule = TablesModule.withCache(mappingsFolder, cacheInputFolder, cacheOutputFolder, true);

		KnowledgeBaseModule kbModule = new KnowledgeBaseModule(kbPath, ontologyRepositoryFile, luceneIndexCache);

		ObjectGraph graph = ObjectGraph.create(new TablesConverterModule(), tableModule, kbModule, tableModule, matchingModule, heidelTimeModule,
				similarityModule);

		OntologyRepository ontologyRepository = graph.get(OntologyRepository.class);

		Set<ClassId> classes = new HashSet<>();
		// classes.add(musicalWork);
		//
		// classes.addAll(ontologyRepository.getSubClasses(musicalWork));

		classes.add(ClassConstants.musicalWorkClass);

		classes.addAll(ontologyRepository.getSubClasses(ClassConstants.musicalWorkClass));

		// tableModule.setClassFilter(graph.get(KnowledgeBase.class));
		List<AnnotatedTableFilter> filters = new ArrayList<>();
		filters.add(new TableClassFilter(classes));

		tableModule.setFilters(filters);

		FilterCachedTables converter = graph.get(FilterCachedTables.class);

		converter.run();

		PersistentParallel6.close();

	}

	public static final ClassId musicalWork = new ClassId("http://dbpedia.org/ontology/MusicalWork");

	public static final ClassId placeClass = new ClassId("http://dbpedia.org/ontology/Settlement");

	@Inject
	protected AnnotatedTablesWriter annotatedTablesWriter;

	@Inject
	protected AnnotatedTablesRepository tablesRepository;

	public FilterCachedTables() {

	}

	public void run() throws ClassNotFoundException, IOException, InterruptedException {

		annotatedTablesWriter.writeAll(tablesRepository);

	}

}
