package de.uni_mannheim.informatik.dws.expansion.samples.dbpedia;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueContainer;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedAbstractObjectHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;

public class KbAbstractObjectInstanceConverter implements KBValueCreator {

	private final OntologyRepository ontologyRepository;

	@Inject
	public KbAbstractObjectInstanceConverter(OntologyRepository ontologyRepository) {
		this.ontologyRepository = ontologyRepository;
	}

	@Override
	public KBValueContainer getMatchedValues(KnowledgeBase kb) {

		ConcurrentHashMap<ValueId, QualifiedValueList> values = new ConcurrentHashMap<>();

		Set<ValueId> removeValues = Collections.newSetFromMap(new ConcurrentHashMap<>());

		PersistentParallel6.runAndFinish(kb.getClasses(), (setId) -> {
			for (PropertyId propertyId : kb.getProperties(setId).keySet()) {
				if (ontologyRepository.isReferenceField(propertyId)) {
					PersistentParallel6.runAndFinish(kb.getInstances(setId, propertyId), (instanceId) -> {
						ValueId vid = new ValueId(instanceId, propertyId);

						if (!values.containsKey(vid) && !removeValues.contains(vid)) {
							QualifiedValueList qvl = kb.getQualifiedValueList(vid);

							List<QualifiedValueHolder> claims = new ArrayList<>();

							for (QualifiedValueHolder c : qvl.getClaims()) {
								StringObject so = (StringObject) c.getValue();
								if (so.getString().equals("")) {
									removeValues.add(vid);
									return;
								} else {
									InstanceId ao = InstanceId.fromString(so.getString());

									if (c.getQualifiers() != null) {
										claims.add(new QualifiedAbstractObjectHolder(ao, c.getQualifiers()));
									} else {
										claims.add(new QualifiedAbstractObjectHolder(ao));
									}
								}
							}

							values.putIfAbsent(vid, new QualifiedValueList(claims));

						}
					}, this.getClass());
				}
			}
		}, this.getClass());

		return KBValueContainer.defaultConfig().rebuilder().setReplaceValues(values).setRemoveValues(removeValues).build();
	}

}
