package de.uni_mannheim.informatik.dws.expansion.samples.dbpedia;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.zip.GZIPInputStream;

import org.apache.commons.compress.compressors.CompressorException;
import org.apache.commons.lang.StringEscapeUtils;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId.InstanceIdStyle;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.BatchConsumerProducer;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueContainer;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedAbstractObjectHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;

public class KbLabelRenamer implements KBValueCreator {

	private final Set<String> files = new HashSet<>();
	private final Set<String> redirectFiles = new HashSet<>();
	private final Set<String> surfaceFormFiles = new HashSet<>();

	public KbLabelRenamer() {

	}

	public KbLabelRenamer(Set<String> files, Set<String> redirectFiles, Set<String> surfaceFormFiles) {
		this.files.addAll(files);
		this.redirectFiles.addAll(redirectFiles);
		this.surfaceFormFiles.addAll(surfaceFormFiles);
	}

	public static void main(String[] args) {

		Set<String> files = new HashSet<>();
		// files.add("D:/stuff/labels_en.nt.gz2");

		Set<String> redirectFiles = new HashSet<>();
		// redirectFiles.add("D:/stuff/redirects_en.nt.gz");

		Set<String> surfaceFormFiles = new HashSet<>();
		surfaceFormFiles.add("D:/stuff/surfaceforms/SurfaceForms_LRD-CC_filter_thr3.80.tsv.gz");
		surfaceFormFiles.add("D:/stuff/surfaceforms/SurfaceForms_LRD-WAT_filter_thr2.60.tsv.gz");

		KbLabelRenamer renamer = new KbLabelRenamer(files, redirectFiles, surfaceFormFiles);

		ConcurrentHashMap<InstanceId, Set<String>> ret = new ConcurrentHashMap<>();

		try {
			for (String file : surfaceFormFiles) {
				renamer.readSurfaceFormsFile(ret, file);
			}
			// renamer.readFile(ret, "D:/stuff/labels_en.nt.gz");
			renamer.readFile(ret, "D:/stuff/mappingbased_properties_en.nt.gz");

			// renamer.readRedirectFile(ret, "D:/stuff/redirects_en.nt.gz");
		} catch (IOException | CompressorException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		int sdf = 123;

		// Pattern pattern = Pattern.compile("<(.*?)> <(.*?)> \"(.*)\"@(.*) .");
		//
		// Matcher matcher = pattern
		// .matcher("<http://dbpedia.org/resource/Alfonso_Cuar%C3%B3n>
		// <http://dbpedia.org/ontology/birthName> \"Alfonso Cuar\u00F3n
		// Orozco\"@en .");
		//
		// while (matcher.find()) {
		// System.out.println(matcher.group(1));
		// System.out.println(matcher.group(2));
		//
		// System.out.println(org.apache.commons.lang.StringEscapeUtils.unescapeJava(matcher.group(3)));
		//
		// }
		//
		// matcher = pattern.matcher(
		// "<http://dbpedia.org/resource/Ashoka>
		// <http://dbpedia.org/ontology/activeYearsStartYear>
		// \"0268\"^^<http://www.w3.org/2001/XMLSchema#gYear> .");
		//
		// while (matcher.find()) {
		// System.out.println(matcher.group(1));
		// System.out.println(matcher.group(2));
		// System.out.println(matcher.group(3));
		// }
	}

	private final static Set<PropertyId> propertyIds = new HashSet<>();

	{

		propertyIds.add(new PropertyId("http://xmlns.com/foaf/0.1/name"));
		propertyIds.add(new PropertyId("http://xmlns.com/foaf/0.1/nick"));
		propertyIds.add(new PropertyId("http://dbpedia.org/ontology/formerName"));
		propertyIds.add(new PropertyId("http://dbpedia.org/ontology/longName"));
		propertyIds.add(new PropertyId("http://dbpedia.org/ontology/birthName"));
		propertyIds.add(new PropertyId("http://dbpedia.org/ontology/alias"));
		propertyIds.add(new PropertyId("http://www.w3.org/2000/01/rdf-schema#label"));
	}

	void readSurfaceFormsFile(ConcurrentHashMap<InstanceId, Set<String>> ret, String file) throws IOException, CompressorException {

		Pattern pattern = Pattern.compile("(.*?)\t(.*?)\t(.*)");

		final BatchConsumerProducer<String> batchConsumerProducer = BatchConsumerProducer.create("Read Wikidata", (e) -> {
			Matcher matcher = pattern.matcher(e);

			if (matcher.find()) {

				InstanceId instanceId = InstanceId.fromReversibleString(matcher.group(1), InstanceIdStyle.DBPEDIA_URI);

				String str = matcher.group(2);
				try {
					// String strNew = java.net.URLDecoder.decode(str, "UTF-8");

					CHMHelper.getFromCHM(ret, instanceId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(str);
				} catch (Exception e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
			}

		}, 5000, 1);

		BufferedReader br = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(file)), "UTF-8"));
		List<String> list = new ArrayList<>();
		String nextLine = null;
		int sum = 0;
		while ((nextLine = br.readLine()) != null) {
			if (!nextLine.startsWith("#")) {
				list.add(nextLine);
				// if (list.size() == 10000) {
				// batchConsumerProducer.addBatch(list);
				// break;
				// }
				if (list.size() == 200000) {
					sum += 200000;
					batchConsumerProducer.addBatch(list);
					list = new ArrayList<>();
					// break;
				}
				if (sum == 400000) {
					sum = 0;
					System.out.println("read 1000000");
					// break;
				}
			}

		}

		if (!list.isEmpty()) {
			batchConsumerProducer.addBatch(list);
		}

		batchConsumerProducer.finish();

		br.close();

	}

	void readRedirectFile(ConcurrentHashMap<InstanceId, Set<String>> ret, String file) throws IOException, CompressorException {

		Pattern pattern = Pattern.compile("<(.*?)> <http:\\/\\/dbpedia\\.org\\/ontology\\/wikiPageRedirects> <(.*?)> .");

		final BatchConsumerProducer<String> batchConsumerProducer = BatchConsumerProducer.create("Read Wikidata", (e) -> {
			Matcher matcher = pattern.matcher(e);

			if (!matcher.find()) {
				int sdf = 123;
			}
			InstanceId instanceId = new InstanceId(matcher.group(1));

			String str = matcher.group(2);
			try {
				String strNew = java.net.URLDecoder.decode(str, "UTF-8");

				CHMHelper.getFromCHM(ret, instanceId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(strNew);
			} catch (Exception e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}

		}, 5000, 1);

		BufferedReader br = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(file))));
		List<String> list = new ArrayList<>();
		String nextLine = null;
		int sum = 0;
		while ((nextLine = br.readLine()) != null) {
			if (!nextLine.startsWith("#")) {
				list.add(nextLine);
				// if (list.size() == 10000) {
				// batchConsumerProducer.addBatch(list);
				// break;
				// }
				if (list.size() == 200000) {
					sum += 200000;
					batchConsumerProducer.addBatch(list);
					list = new ArrayList<>();
					// break;
				}
				if (sum == 400000) {
					sum = 0;
					System.out.println("read 1000000");
					// break;
				}
			}

		}

		if (!list.isEmpty()) {
			batchConsumerProducer.addBatch(list);
		}

		batchConsumerProducer.finish();

		br.close();

	}

	void readFile(ConcurrentHashMap<InstanceId, Set<String>> ret, String file) throws IOException, CompressorException {

		Pattern pattern = Pattern.compile("<(.*?)> <(.*?)> \"(.*)\"@(.*) .");

		final BatchConsumerProducer<String> batchConsumerProducer = BatchConsumerProducer.create("Read Wikidata", (e) -> {
			Matcher matcher = pattern.matcher(e);

			if (matcher.find()) {
				InstanceId instanceId = new InstanceId(matcher.group(1));
				PropertyId propertyId = new PropertyId(matcher.group(2));
				if (propertyIds.contains(propertyId)) {
					String str = matcher.group(3);
					str = StringEscapeUtils.unescapeJava(str);

					CHMHelper.getFromCHM(ret, instanceId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(str);
				}
			}

		}, 5000, 1);

		BufferedReader br = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(file))));
		List<String> list = new ArrayList<>();
		String nextLine = null;
		int sum = 0;
		while ((nextLine = br.readLine()) != null) {
			if (!nextLine.startsWith("#")) {
				list.add(nextLine);
				// if (list.size() == 10000) {
				// batchConsumerProducer.addBatch(list);
				// break;
				// }
				if (list.size() == 200000) {
					sum += 200000;
					batchConsumerProducer.addBatch(list);
					list = new ArrayList<>();
					// break;
				}
				if (sum == 400000) {
					sum = 0;
					System.out.println("read 1000000");
					// break;
				}
			}

		}

		if (!list.isEmpty()) {
			batchConsumerProducer.addBatch(list);
		}

		batchConsumerProducer.finish();

		br.close();

	}

	@Override
	public KBValueContainer getMatchedValues(KnowledgeBase kb) {

		ConcurrentHashMap<InstanceId, Set<String>> ret = new ConcurrentHashMap<>();
		for (String file : files) {
			try {
				readFile(ret, file);
			} catch (IOException | CompressorException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		for (String file : redirectFiles) {
			try {
				readRedirectFile(ret, file);
			} catch (IOException | CompressorException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		for (String file : surfaceFormFiles) {
			try {
				readSurfaceFormsFile(ret, file);
			} catch (IOException | CompressorException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		PropertyId oldLabelProperty = new PropertyId("http://www.w3.org/2000/01/rdf-schema#label");

		Map<ValueId, QualifiedValueList> replaceValues = new ConcurrentHashMap<>();
		Set<ValueId> removeValues = Collections.newSetFromMap(new ConcurrentHashMap<>());

		PersistentParallel6.runAndFinish(kb.getAllInstances(), (instanceId) -> {
			ValueId oldValueId = new ValueId(instanceId, oldLabelProperty);
			ValueId newValueId = new ValueId(instanceId, PropertiesExpansionData.LABEL_PROPERTY);
			QualifiedValueList qvl = kb.getQualifiedValueList(oldValueId);

			// else {

			LinkedHashSet<String> labels = new LinkedHashSet<>();

			if (qvl != null) {
				removeValues.add(oldValueId);
				// replaceValues.put(newValueId, qvl);

				for (QualifiedValueHolder claim : qvl.getClaims()) {
					if (claim.getValue() instanceof StringObject) {
						labels.add(((StringObject) claim.getValue()).getString());
					}
				}
			}

			Set<String> newValuesSet = ret.get(instanceId);
			if (newValuesSet != null) {
				labels.addAll(newValuesSet);
			}
			try {
				labels.add(java.net.URLDecoder.decode(instanceId.getReadableString(), "UTF-8"));
			} catch (Exception e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
			;

			List<QualifiedValueHolder> claims = labels.stream().map((e) -> new QualifiedAbstractObjectHolder(new StringObject(e))).collect(Collectors.toList());

			qvl = new QualifiedValueList(claims);
			replaceValues.put(newValueId, qvl);
			// }

		}, this.getClass());

		return KBValueContainer.defaultConfig().rebuilder().setRemoveValues(removeValues).setReplaceValues(replaceValues).build();
	}

}
