package de.uni_mannheim.informatik.dws.expansion.samples.dbpedia;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.PropertyIndexComposite;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;

public class NFLPropertiesDensity implements AnnotatedTableFilter {

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {

		AnnotatedTableMetaData md = data.getMetaData();

		Map<Integer, List<PropertyId>> newProperties = new HashMap<>();
		Map<PropertyIndexComposite, Double> newPropertyScores = new HashMap<>();
		for (Entry<Integer, List<PropertyId>> e : md.getAllProperties().entrySet()) {
			if (e.getValue() != null) {
				List<PropertyId> newList = new ArrayList<>();
				for (PropertyId propertyId : e.getValue()) {

					if (propertyId.getAsString().equals("http://dbpedia.org/ontology/formerTeam")
							|| propertyId.getAsString().equals("http://dbpedia.org/ontology/debutTeam")) {
						PropertyId newProperty = new PropertyId("http://dbpedia.org/ontology/team");
						newList.add(newProperty);
						PropertyIndexComposite composite = new PropertyIndexComposite(e.getKey(), propertyId);
						PropertyIndexComposite newComposite = new PropertyIndexComposite(e.getKey(), newProperty);
						newPropertyScores.put(newComposite, md.getPropertyScore(composite));
					} else {
						newList.add(propertyId);
						PropertyIndexComposite composite = new PropertyIndexComposite(e.getKey(), propertyId);
						newPropertyScores.put(composite, md.getPropertyScore(composite));
					}

				}

				if (!newList.isEmpty()) {
					newProperties.put(e.getKey(), newList);
				}
			}
		}

		md = new AnnotatedTableMetaData(md.getSource(), md.getDbPediaClass(), md.getClassConf(), md.getNumHeaderRows(), md.getNumberOfRows(), newProperties,
				newPropertyScores, md.getAllInstances(), md.getInstancesScores(), md.getKeyColumn(), md.getYearsBeforeText(), md.getYearsAfterText(),
				md.getYearsInContextBefore(), md.getYearsInTableTitle(), md.getYearsInPageTitle(), md.getPropertiesYears(), md.getPageTitle(),
				md.getTableTitle(), md.getTableContextTimeStampBeforeTable(), md.getTextBeforeTable(), md.getTextAfterTable());

		return Optional.of(new AnnotatedTablesReaderData(data.getId(), md, data.getTableData()));
	}

}
