package de.uni_mannheim.informatik.dws.expansion.samples.dbpedia;

import java.io.IOException;

import javax.inject.Inject;

import dagger.Module;
import dagger.ObjectGraph;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTablesSimpleReader;
import de.mannheim.uni.expansion.tablefusion.modules.HeidelTimeModule;
import de.mannheim.uni.expansion.tablefusion.modules.MatchingModule;
import de.mannheim.uni.expansion.tablefusion.modules.SimilarityModule;
import de.mannheim.uni.expansion.tablefusion.modules.TablesModule;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.units.UnitParser;
import de.uni_mannheim.informatik.dws.expansion.matching.KnowledgeBaseModule;
import de.uni_mannheim.informatik.dws.expansion.utils.remote.RepositoryLoading;

public class ProcessCSVSingleTable {

	@Module(injects = { ProcessCSVSingleTable.class }, library = true, complete = false)
	public static class TablesConverterModule {

	}

	public static void main(String[] args) throws ClassNotFoundException, IOException, InterruptedException {
		String dataHomeTemp = null;

		if (args.length > 0 && args[0] != null && !args[0].equals("")) {
			dataHomeTemp = args[0];
		}

		String kbName = null;

		if (args.length > 1 && args[1] != null && !args[1].equals("")) {
			kbName = args[1];
		}

		String mappingsName = null;

		if (args.length > 2 && args[2] != null && !args[2].equals("")) {
			mappingsName = args[2];
		}
		String dataHome = dataHomeTemp;

		mappingsName = "/tables/" + mappingsName;

		String kbPath = dataHome + "/kbs/" + kbName + "/kb/";
		String ontologyRepositoryFile = dataHome + "/kbs/" + kbName + "/ontology.kryo";
		String luceneIndexCache = dataHome + "/kbs/" + kbName + "/lucene/";

		String mappingsFolder = dataHome + "/" + mappingsName + "/single/";
		String cacheFolder = dataHome + "/" + mappingsName + "/cache/";
		String resultsFolder = dataHome + "/" + mappingsName + "/results/";
		String problemsFolder = dataHome + "/" + mappingsName + "/problems/";

		String optimizingFolder = dataHome + "/" + mappingsName + "/optimizing/";

		String columnMappingsFolder = dataHome + "/" + mappingsName + "/columnMappings/";
		String heidelTimeConfigPath = dataHome + "/heidelTime/config.props";
		String emptyFile = dataHome + "/empty.txt";
		String candidateSelectionConfig = dataHome + "/config.conf";
		String matchingConfig = dataHome + "/config.conf";
		UnitParser.unitPath = dataHome + "/Units/Convertible";

		SimilarityModule similarityModule = new SimilarityModule();
		MatchingModule matchingModule = new MatchingModule(true, columnMappingsFolder, candidateSelectionConfig, matchingConfig, emptyFile);
		HeidelTimeModule heidelTimeModule = new HeidelTimeModule(heidelTimeConfigPath);
		TablesModule tableModule = TablesModule.csvWithCache(mappingsFolder, cacheFolder, false);

		UnitParser.unitPath = dataHome + "/Units/Convertible";

		UnitParser.readInUnits();

		KnowledgeBaseModule kbModule = new KnowledgeBaseModule(kbPath, ontologyRepositoryFile, luceneIndexCache, RepositoryLoading.REMOTE_LOAD_REMOTELY);

		ObjectGraph graph = ObjectGraph.create(new TablesConverterModule(), tableModule, kbModule, tableModule, matchingModule, heidelTimeModule,
				similarityModule);

		ProcessCSVSingleTable processor = graph.get(ProcessCSVSingleTable.class);

		processor.run(cacheFolder);

		PersistentParallel6.close();

	}

	@Inject
	protected AnnotatedTablesSimpleReader reader;

	// @Inject
	// protected ColumnMatchesWriter columnMatchesWriter;

	public ProcessCSVSingleTable() {

	}

	public void run(String cacheFolder) throws ClassNotFoundException, IOException, InterruptedException {

		reader.runForEachTable((data) -> {
			int sdf = 123;
		});

	}

}
