package de.uni_mannheim.informatik.dws.expansion.samples.dbpedia;

import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.containers.Pair;

public class SmallTableFilter implements AnnotatedTableFilter {

	private final ConcurrentHashMap<Integer, Pair<AtomicInteger, AtomicInteger>> count = new ConcurrentHashMap<>();
	private final AtomicInteger total = new AtomicInteger(0);
	private final AtomicInteger removed = new AtomicInteger(0);

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {
		Pair<AtomicInteger, AtomicInteger> pair = CHMHelper.getFromCHM(count, data.getMetaData().getNumberOfRows(),
				() -> new Pair<>(new AtomicInteger(), new AtomicInteger()));
		pair.getFirst().incrementAndGet();
		int matchedCount = data.getMetaData().getProperties().size() * data.getMetaData().getInstances().size();

		pair.getSecond().addAndGet(matchedCount);
		total.incrementAndGet();
		if (data.getMetaData().getInstances().size() < 8) {
			removed.incrementAndGet();
			return Optional.empty();
		}
		return Optional.of(data);
	}

	public ConcurrentHashMap<Integer, Pair<AtomicInteger, AtomicInteger>> getCount() {
		return count;
	}

	public int getTotal() {
		return total.intValue();
	}

	public int getRemoved() {
		return removed.intValue();
	}

}
