package de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.musicalWork;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.GZIPInputStream;

import org.apache.commons.compress.compressors.CompressorException;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.BatchConsumerProducer;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueContainer;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject.Precision;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedAbstractObjectHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.samples.setExtension.ClassConstants;
import de.uni_mannheim.informatik.dws.expansion.samples.setExtension.PropertyConstants;

public class DraftYearValueCreator implements KBValueCreator {

	private final String file;

	public DraftYearValueCreator(String file) {
		this.file = file;
	}

	public static void main(String[] args) {

		DraftYearValueCreator renamer = new DraftYearValueCreator("D:/stuff/mappingbased_properties_en.nt.gz");

		ConcurrentHashMap<InstanceId, List<TimeObject>> ret = new ConcurrentHashMap<>();
		Set<InstanceId> instances = new HashSet<>();
		instances.add(new InstanceId("http://dbpedia.org/resource/Gaines_Adams"));
		instances.add(new InstanceId("http://dbpedia.org/resource/Bert_Bell"));
		instances.add(new InstanceId("http://dbpedia.org/resource/Johnny_Unitas"));
		instances.add(new InstanceId("http://dbpedia.org/resource/Kareem_Abdul-Jabbar"));
		instances.add(new InstanceId("http://dbpedia.org/resource/Knute_Rockne"));
		instances.add(new InstanceId("http://dbpedia.org/resource/Vince_Lombardi"));
		instances.add(new InstanceId("http://dbpedia.org/resource/Byron_White"));
		instances.add(new InstanceId("http://dbpedia.org/resource/Jim_Thorpe"));
		try {
			// renamer.readFile(ret, "D:/stuff/labels_en.nt.gz");
			renamer.readDraftYearFile(ret, instances);

			// renamer.readRedirectFile(ret, "D:/stuff/redirects_en.nt.gz");
		} catch (IOException | CompressorException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		int sdf = 123;

		// Pattern pattern = Pattern.compile("<(.*?)> <(.*?)> \"(.*)\"@(.*) .");
		//
		// Matcher matcher = pattern
		// .matcher("<http://dbpedia.org/resource/Alfonso_Cuar%C3%B3n>
		// <http://dbpedia.org/ontology/birthName> \"Alfonso Cuar\u00F3n
		// Orozco\"@en .");
		//
		// while (matcher.find()) {
		// System.out.println(matcher.group(1));
		// System.out.println(matcher.group(2));
		//
		// System.out.println(org.apache.commons.lang.StringEscapeUtils.unescapeJava(matcher.group(3)));
		//
		// }
		//
		// matcher = pattern.matcher(
		// "<http://dbpedia.org/resource/Ashoka>
		// <http://dbpedia.org/ontology/activeYearsStartYear>
		// \"0268\"^^<http://www.w3.org/2001/XMLSchema#gYear> .");
		//
		// while (matcher.find()) {
		// System.out.println(matcher.group(1));
		// System.out.println(matcher.group(2));
		// System.out.println(matcher.group(3));
		// }
	}

	@Override
	public KBValueContainer getMatchedValues(KnowledgeBase kb) {
		try {
			Set<InstanceId> instances = kb.getInstances(ClassConstants.gridIronFootballPlayer).keySet();
			ConcurrentHashMap<InstanceId, List<TimeObject>> map = new ConcurrentHashMap<>();
			readDraftYearFile(map, instances);

			ConcurrentHashMap<ValueId, QualifiedValueList> replaceValues = new ConcurrentHashMap<>();

			// Set<InstanceId> instancesWithValue =
			// kb.getInstances(ClassConstants.gridIronFootballPlayer,
			// PropertyConstants.college);

			PersistentParallel6.runAndFinish(map.entrySet(), (value) -> {
				ValueId valueId = new ValueId(value.getKey(), PropertyConstants.draftYear);
				List<QualifiedValueHolder> claims = new ArrayList<>();
				if (kb.hasValue(valueId)) {
					claims.addAll(kb.getQualifiedValueList(valueId).getClaims());
				}

				for (TimeObject college : value.getValue()) {
					claims.add(new QualifiedAbstractObjectHolder(college));
				}

				replaceValues.put(valueId, new QualifiedValueList(claims));
			}, this.getClass());

			return KBValueContainer.builder().setReplaceValues(replaceValues).build();
		} catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException(e);
		}
	}

	void readDraftYearFile(ConcurrentHashMap<InstanceId, List<TimeObject>> ret, Set<InstanceId> filter) throws IOException, CompressorException {

		Pattern pattern = Pattern
				.compile("<(.*?)> <http:\\/\\/dbpedia\\.org\\/ontology\\/draftYear> \"(.*?)\"\\^\\^<http:\\/\\/www\\.w3\\.org\\/2001\\/XMLSchema#gYear> .");

		final BatchConsumerProducer<String> batchConsumerProducer = BatchConsumerProducer.create("Read Wikidata", (e) -> {
			Matcher matcher = pattern.matcher(e);

			if (matcher.find()) {
				InstanceId player = new InstanceId(matcher.group(1));
				String year = matcher.group(2);
				Integer yearInt = Integer.parseInt(year);

				TimeObject to = new TimeObject(yearInt, 0, 0, 0, 0, 0, 0, Precision.T09_years);
				if (filter.contains(player)) {

					CHMHelper.getFromCHM(ret, player, () -> Collections.synchronizedList(new ArrayList<>())).add(to);
				}
			}

		}, 5000, 1);

		BufferedReader br = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(file))));
		List<String> list = new ArrayList<>();
		String nextLine = null;
		int sum = 0;
		while ((nextLine = br.readLine()) != null) {
			if (!nextLine.startsWith("#")) {
				list.add(nextLine);
				// if (list.size() == 10000) {
				// batchConsumerProducer.addBatch(list);
				// break;
				// }
				if (list.size() == 200000) {
					sum += 200000;
					batchConsumerProducer.addBatch(list);
					list = new ArrayList<>();
					// break;
				}
				if (sum == 400000) {
					sum = 0;
					System.out.println("read 1000000");
					// break;
				}
			}

		}

		if (!list.isEmpty()) {
			batchConsumerProducer.addBatch(list);
		}

		batchConsumerProducer.finish();

		br.close();

	}
}
