package de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.musicalWork;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import com.google.common.collect.BiMap;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueContainer;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.samples.setExtension.ClassConstants;
import de.uni_mannheim.informatik.dws.expansion.samples.setExtension.PropertyConstants;

public class IsPartOfFixer implements KBValueCreator {

	private final OntologyRepository ontology;

	public IsPartOfFixer(OntologyRepository ontology) {
		this.ontology = ontology;
	}

	@Override
	public KBValueContainer getMatchedValues(KnowledgeBase kb) {

		BiMap<PropertyId, Integer> properties = kb.getProperties(ClassConstants.placeClass);

		Set<ClassId> populatedPlaceSubClasses = new HashSet<>(ontology.getSubClasses(ClassConstants.populatedPlace));
		populatedPlaceSubClasses.add(ClassConstants.populatedPlace);

		ConcurrentHashMap<ValueId, QualifiedValueList> replaceValues = new ConcurrentHashMap<>();
		for (PropertyId propertyId : properties.keySet()) {

			if (ontology.getAllPropertyParent(propertyId).contains(PropertyConstants.isPartOfDUL)) {
				ClassPropertyId cpid = new ClassPropertyId(ClassConstants.placeClass, propertyId);
				Set<ClassId> range = ontology.getRange(cpid);
				Set<ClassId> domain = ontology.getDomain(cpid);
				if ((range != null && atLeastOneOverlap(range, populatedPlaceSubClasses))) {

					Set<InstanceId> instances = kb.getInstances(ClassConstants.placeClass, propertyId);

					PersistentParallel6.runAndFinish(instances, (instanceId) -> {

						QualifiedValueList qvl = kb.getQualifiedValueList(new ValueId(instanceId, propertyId));
						if (qvl == null) {
							int sdf = 123;
							return;
						}

						ValueId isPartOfVid = new ValueId(instanceId, PropertyConstants.isPartOf);
						QualifiedValueList isPartOfQvl = kb.getQualifiedValueList(isPartOfVid);

						List<QualifiedValueHolder> claims;

						if (isPartOfQvl != null) {
							claims = isPartOfQvl.getClaims();
						} else {
							claims = new ArrayList<>();

						}
						List<QualifiedValueHolder> newClaims = new ArrayList<>();
						for (QualifiedValueHolder claim : qvl.getClaims()) {
							boolean foundClaim = false;
							for (QualifiedValueHolder candidate : claims) {
								if (claim.getValue().equals(candidate.getValue())) {
									foundClaim = true;
									break;
								}
							}
							if (!foundClaim) {
								newClaims.add(claim);
							}
						}

						newClaims.addAll(claims);

						replaceValues.put(isPartOfVid, new QualifiedValueList(newClaims));
					}, this.getClass());
				}
			}
		}

		return KBValueContainer.builder().setReplaceValues(replaceValues).build();
	}

	private boolean atLeastOneOverlap(Set<ClassId> c1Set, Set<ClassId> c2Set) {
		for (ClassId c1 : c1Set) {
			if (c2Set.contains(c1)) {
				return true;
			}
		}
		return false;
	}

}
