package de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.musicalWork;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.PropertyIndexComposite;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.uni_mannheim.informatik.dws.expansion.samples.setExtension.ClassConstants;
import de.uni_mannheim.informatik.dws.expansion.samples.setExtension.PropertyConstants;

public class MusicalArtistPropertyFix implements AnnotatedTableFilter {

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {

		AnnotatedTableMetaData md = data.getMetaData();

		if (!md.getDbPediaClass().equals(ClassConstants.musicalWorkClass) && !md.getDbPediaClass().equals(ClassConstants.single)
				&& !md.getDbPediaClass().equals(ClassConstants.song) && !md.getDbPediaClass().equals(ClassConstants.album)) {
			return Optional.of(data);
		}

		Map<Integer, List<PropertyId>> newProperties = new HashMap<>();
		Map<PropertyIndexComposite, Double> newPropertyScores = new HashMap<>();
		for (Entry<Integer, List<PropertyId>> e : md.getAllProperties().entrySet()) {
			if (e.getValue() != null) {
				List<PropertyId> newList = new ArrayList<>();
				for (PropertyId propertyId : e.getValue()) {

					if (propertyId.equals(PropertyConstants.musicalBand) || propertyId.equals(PropertyConstants.artist)) {
						newList.add(PropertyConstants.musicalArtist);
						PropertyIndexComposite composite = new PropertyIndexComposite(e.getKey(), PropertyConstants.musicalArtist);
						newPropertyScores.put(composite, md.getPropertyScore(new PropertyIndexComposite(e.getKey(), propertyId)));
					} else {
						newList.add(propertyId);
						PropertyIndexComposite composite = new PropertyIndexComposite(e.getKey(), propertyId);
						newPropertyScores.put(composite, md.getPropertyScore(composite));
					}

				}

				if (!newList.isEmpty()) {
					newProperties.put(e.getKey(), newList);
				}
			}
		}

		md = new AnnotatedTableMetaData(md.getSource(), md.getDbPediaClass(), md.getClassConf(), md.getNumHeaderRows(), md.getNumberOfRows(), newProperties,
				newPropertyScores, md.getAllInstances(), md.getInstancesScores(), md.getKeyColumn(), md.getYearsBeforeText(), md.getYearsAfterText(),
				md.getYearsInContextBefore(), md.getYearsInTableTitle(), md.getYearsInPageTitle(), md.getPropertiesYears(), md.getPageTitle(),
				md.getTableTitle(), md.getTableContextTimeStampBeforeTable(), md.getTextBeforeTable(), md.getTextAfterTable());

		return Optional.of(new AnnotatedTablesReaderData(data.getId(), md, data.getTableData()));

	}

}
