package de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.musicalWork;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyAdapter;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepositoryImpl;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueContainer;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueContainer.KBValueContainerBuilder;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.CategoricalIntObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.QuantityObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedAbstractObjectHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.attribute.matcher.AttributeMatcher;
import de.uni_mannheim.informatik.dws.expansion.samples.setExtension.ClassConstants;
import de.uni_mannheim.informatik.dws.expansion.samples.setExtension.PropertyConstants;

public class OntologicalFixes implements KBValueCreator, OntologyAdapter {

	public OntologicalFixes(OntologyRepository ontology) {
	}

	@Override
	public OntologyRepositoryImpl adapt(OntologyRepositoryImpl input) {

		Map<ClassPropertyId, Set<ClassId>> propertyRange = fixRange(input.getPropertyRange());

		Map<PropertyId, ValueType> dataTypes = fixDataType(input.getPropertyDataType());

		return new OntologyRepositoryImpl(dataTypes, input.getPropertyTimeDependent(), input.getPropertyDomain(), propertyRange, input.getParents(),
				input.getChildren(), input.getPropertyParents());
	}

	@Override
	public KBValueContainer getMatchedValues(KnowledgeBase kb) {
		Set<ValueId> valueIds = new HashSet<>();
		for (ClassPropertyId cpid : kb.getClassPropertyIds()) {
			Set<InstanceId> instances = kb.getInstances(cpid.getClassId(), cpid.getPropertyId());

			for (InstanceId instanceId : instances) {
				valueIds.add(new ValueId(instanceId, cpid.getPropertyId()));
			}
		}

		ConcurrentHashMap<ValueId, QualifiedValueList> replaceValues = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(valueIds, (valueId) -> {
			if (kb.hasValue(valueId)) {
				QualifiedValueList qvl = kb.getQualifiedValueList(valueId);
				if (valueId.getPropertyId().equals(PropertyConstants.draftPick) || valueId.getPropertyId().equals(PropertyConstants.draftRound)
						|| valueId.getPropertyId().equals(PropertyConstants.number)) {
					replaceValues.put(valueId, convertFromStringToInt(qvl));
				}
			}
		}, this.getClass());

		// Set<Pair<InstanceId, ClassId>> newSetMappings = new HashSet<>();
		// for (InstanceId instanceId :
		// kb.getInstances(ClassConstants.song).keySet()) {
		// newSetMappings.add(new Pair<>(instanceId, ClassConstants.single));
		// }

		KBValueContainerBuilder builder = KBValueContainer.builder();
		builder.setReplaceValues(replaceValues);
		return builder.build();
	}

	private QualifiedValueList convertFromStringToInt(QualifiedValueList qvl) {
		List<QualifiedValueHolder> claims = new ArrayList<>();

		for (QualifiedValueHolder claim : qvl.getClaims()) {
			if (claim.getValue() instanceof StringObject) {
				StringObject so = (StringObject) claim.getValue();
				if (AttributeMatcher.isInteger(so.getString(), 10)) {
					int value = Integer.parseInt(so.getString());
					claims.add(new QualifiedAbstractObjectHolder(new CategoricalIntObject(value)));
				}
			} else if (claim.getValue() instanceof QuantityObject) {

				QuantityObject qo = (QuantityObject) claim.getValue();
				long longValue = Math.round(qo.getValue());

				claims.add(new QualifiedAbstractObjectHolder(new CategoricalIntObject(longValue)));

			}
		}

		return new QualifiedValueList(claims);
	}

	Map<ClassPropertyId, Set<ClassId>> fixRange(Map<ClassPropertyId, Set<ClassId>> propertyRangeInput) {
		Map<ClassPropertyId, Set<ClassId>> propertyRange = propertyRangeInput.entrySet().stream()
				.collect(Collectors.toMap(Entry::getKey, (e) -> new HashSet<>(e.getValue())));

		CHMHelper.getFromMap(propertyRange, new ClassPropertyId(ClassConstants.musicalWorkClass, PropertyConstants.musicalArtist), () -> new HashSet<>())
				.add(ClassConstants.band);

		CHMHelper.getFromMap(propertyRange, new ClassPropertyId(ClassConstants.gridIronFootballPlayer, PropertyConstants.position), () -> new HashSet<>())
				.add(ClassConstants.gridironFootballPosition);

		CHMHelper.getFromMap(propertyRange, new ClassPropertyId(ClassConstants.gridIronFootballPlayer, PropertyConstants.college), () -> new HashSet<>())
				.add(ClassConstants.collegeFootballTeam);

		CHMHelper.getFromMap(propertyRange, new ClassPropertyId(ClassConstants.placeClass, PropertyConstants.isPartOf), () -> new HashSet<>())
				.add(ClassConstants.populatedPlace);

		return propertyRange;
	}

	private Map<PropertyId, ValueType> fixDataType(Map<PropertyId, ValueType> propertyDataTypeInput) {
		Map<PropertyId, ValueType> propertyDataType = new HashMap<>(propertyDataTypeInput);
		propertyDataType.put(PropertyConstants.draftPick, ValueType.CATEGORICAL_INT);
		propertyDataType.put(PropertyConstants.number, ValueType.CATEGORICAL_INT);
		propertyDataType.put(PropertyConstants.draftRound, ValueType.CATEGORICAL_INT);
		propertyDataType.put(PropertyConstants.weightPounds, ValueType.QUANTITY);

		return propertyDataType;
	}

}
