package de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.musicalWork;

import java.util.Map.Entry;
import java.util.Optional;
import java.util.SortedMap;
import java.util.TreeMap;

import com.google.common.collect.HashBasedTable;
import com.google.common.collect.Table;

import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.QuantityObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class TableColumnCoordinateToQuantityConverter implements AnnotatedTableFilter {

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {
		Table<RowIdx, ColumnIdx, AbstractObject> replace = null;
		SortedMap<ColumnIdx, ValueType> newDataTypes = null;
		for (Entry<ColumnIdx, ValueType> dataType : data.getTableData().getDataTypes().entrySet()) {
			if (dataType.getValue() == ValueType.COORDINATE) {
				if (replace == null) {
					replace = HashBasedTable.create(data.getTableData().getTable());
					newDataTypes = new TreeMap<>(data.getTableData().getDataTypes());
				}

				newDataTypes.put(dataType.getKey(), ValueType.QUANTITY);

				for (Entry<RowIdx, AbstractObject> row : replace.column(dataType.getKey()).entrySet()) {
					if (row.getValue().getType() == ValueType.QUANTITY) {
						throw new IllegalStateException();
					}
					double doubleValue = Double.parseDouble(row.getValue().toString());
					row.setValue(QuantityObject.fromDouble(doubleValue));

				}
			}
		}

		if (replace == null) {
			return Optional.of(data);
		} else {
			return Optional.of(new AnnotatedTablesReaderData(data.getId(), data.getMetaData(), new AnnotatedTable(replace, data.getTableData().getStringTable(),
					newDataTypes, data.getTableData().getKeyColumnIndex(), data.getTableData().getColumnHeaders())));
		}
	}

}
