package de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.musicalWork;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueContainer;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.measurement.UnitConverter;
import de.uni_mannheim.dws.expansion.data.ontology.model.measurement.UnitOfMeasurement;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.QuantityObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedAbstractObjectHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.samples.setExtension.ClassConstants;
import de.uni_mannheim.informatik.dws.expansion.samples.setExtension.PropertyConstants;

public class WeightInPoundsValueCreator implements KBValueCreator {

	private final UnitConverter unitConverter = new UnitConverter();

	@Override
	public KBValueContainer getMatchedValues(KnowledgeBase kb) {
		try {

			Set<InstanceId> instances = kb.getInstances(ClassConstants.gridIronFootballPlayer, PropertyConstants.weight);
			ConcurrentHashMap<ValueId, QualifiedValueList> replaceValues = new ConcurrentHashMap<>();

			PersistentParallel6.runAndFinish(instances, (instanceId) -> {

				ValueId valueId = new ValueId(instanceId, PropertyConstants.weight);

				List<QualifiedValueHolder> claims = new ArrayList<>();

				for (QualifiedValueHolder qvh : kb.getQualifiedValueList(valueId).getClaims()) {
					if (qvh.getValue().getType().equals(ValueType.QUANTITY)) {
						double convertedValue = unitConverter.convert(((QuantityObject) qvh.getValue()).getValue(), UnitOfMeasurement.WEIGHT_KG,
								UnitOfMeasurement.WEIGHT_POUND);
						if (qvh.getQualifiers() != null)
							claims.add(new QualifiedAbstractObjectHolder(new QuantityObject(convertedValue, null, null, null), qvh.getQualifiers()));
						else
							claims.add(new QualifiedAbstractObjectHolder(new QuantityObject(convertedValue, null, null, null)));
					} else {
						claims.add(qvh);
					}
				}

				replaceValues.put(new ValueId(instanceId, PropertyConstants.weightPounds), new QualifiedValueList(claims));
			}, this.getClass());

			return KBValueContainer.builder().setReplaceValues(replaceValues).build();
		} catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException(e);
		}
	}

}
