package de.uni_mannheim.informatik.dws.expansion.samples.postMatching;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

import dagger.ObjectGraph;
import de.mannheim.uni.expansion.data.ontology.model.instances.QualifierId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.PropertyIndexComposite;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTablesWriter;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.modules.HeidelTimeModule;
import de.mannheim.uni.expansion.tablefusion.modules.MatchingModule;
import de.mannheim.uni.expansion.tablefusion.modules.SimilarityModule;
import de.mannheim.uni.expansion.tablefusion.modules.TableFusionModule;
import de.mannheim.uni.expansion.tablefusion.modules.TablesModule;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.units.UnitParser;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.informatik.dws.expansion.matching.KnowledgeBaseModule;

public class CorrectTablesQualifierId_and_Continent {

	public static void main(String[] args) throws IOException {
		String dataHome = null;

		if (args.length > 0 && args[0] != null && !args[0].equals("")) {
			dataHome = args[0];
		}

		String mappingsName = null;

		if (args.length > 0 && args[1] != null && !args[1].equals("")) {
			mappingsName = args[1];
		}

		mappingsName = "/tables/" + mappingsName;

		String kbPath = dataHome + "/kb/";
		String ontologyRepositoryFile = dataHome + "/ontology.kryo";
		String luceneIndexCache = dataHome + "/lucene/";

		String mappingsFolder = dataHome + "/" + mappingsName + "/mappings/";
		String cacheFolder = dataHome + "/" + mappingsName + "/cache/";
		String resultsFolder = dataHome + "/" + mappingsName + "/results/";
		String problemsFolder = dataHome + "/" + mappingsName + "/problems/";

		String optimizingFolder = dataHome + "/" + mappingsName + "/optimizing/";

		String columnMappingsFolder = dataHome + "/" + mappingsName + "/columnMappings/";

		KnowledgeBaseModule kbModule = new KnowledgeBaseModule(kbPath, ontologyRepositoryFile, luceneIndexCache);

		TablesModule tableModule = TablesModule.withCache(mappingsFolder, cacheFolder, true);

		List<AnnotatedTableFilter> filters = new ArrayList<>();
		filters.add(new AnnotatedTableFilter() {

			@Override
			public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {

				AnnotatedTableMetaData md = data.getMetaData();
				Map<Integer, List<PropertyId>> properties = new HashMap<>(md.getAllProperties());
				Map<PropertyIndexComposite, Double> oldPropertiesScores = md.getAllPropertyScores();
				Map<PropertyIndexComposite, Double> newPropertiesScores = md.getAllPropertyScores();
				for (Entry<Integer, List<PropertyId>> e : properties.entrySet()) {
					List<PropertyId> newList = new ArrayList<>();
					for (PropertyId propertyId : e.getValue()) {
						PropertyId newPropertyId;
						String[] parts = propertyId.getPropertyId().split("::");
						if (parts.length == 2) {
							newPropertyId = new QualifierId(new PropertyId(parts[0]), new PropertyId(parts[1]));
						} else {
							if (propertyId.getPropertyId().equals("_P30")) {
								newPropertyId = new PropertyId("P30", "Continent");
							} else {
								newPropertyId = propertyId;
							}
						}

						newList.add(newPropertyId);
						newPropertiesScores.put(new PropertyIndexComposite(e.getKey(), newPropertyId),
								oldPropertiesScores.get(new PropertyIndexComposite(e.getKey(), propertyId)));
					}
					e.setValue(newList);

				}
				md = new AnnotatedTableMetaData(md.getSource(), md.getDbPediaClass(), md.getClassConf(), md.getNumHeaderRows(), md.getNumberOfRows(),
						properties, newPropertiesScores, md.getAllInstances(), md.getInstancesScores(), md.getKeyColumn(), md.getYearsBeforeText(),
						md.getYearsAfterText(), md.getYearsInContextBefore(), md.getYearsInTableTitle(), md.getYearsInPageTitle(), md.getPropertiesYears(),
						md.getPageTitle(), md.getTableTitle(), md.getTableContextTimeStampBeforeTable(), md.getTextBeforeTable(), md.getTextAfterTable());

				return Optional.of(new AnnotatedTablesReaderData(data.getId(), md, data.getTableData()));
			}
		});
		tableModule.setFilters(filters);

		String heidelTimeConfigPath = dataHome + "/heidelTime/config.props";
		String emptyFile = dataHome + "/empty.txt";
		String candidateSelectionConfig = dataHome + "/config.conf";
		String matchingConfig = dataHome + "/config.conf";
		UnitParser.unitPath = dataHome + "/Units/Convertible";

		SimilarityModule similarityModule = new SimilarityModule();
		MatchingModule matchingModule = new MatchingModule(true, columnMappingsFolder, candidateSelectionConfig, matchingConfig, emptyFile);
		HeidelTimeModule heidelTimeModule = new HeidelTimeModule(heidelTimeConfigPath);
		TableFusionModule tableFusionModule = new TableFusionModule(resultsFolder, problemsFolder, null, null, optimizingFolder, null);
		ObjectGraph graph = ObjectGraph.create(tableFusionModule, kbModule, tableModule, matchingModule, similarityModule, heidelTimeModule);

		tableModule.setClassFilter(graph.get(KnowledgeBase.class));

		AnnotatedTablesRepository tablesRepository = graph.get(AnnotatedTablesRepository.class);

		AnnotatedTablesWriter writer = graph.get(AnnotatedTablesWriter.class);

		writer.writeAll(tablesRepository);

		try {
			PersistentParallel6.close();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		// System.exit(0);

	}

}
