package de.uni_mannheim.informatik.dws.expansion.samples.postMatching;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.mutable.MutableDouble;
import org.json.JSONObject;

import com.google.common.collect.ImmutableMap;

import dagger.Module;
import dagger.ObjectGraph;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.PerformancePR;
import de.mannheim.uni.expansion.tablefusion.core.fusing.FusionScoreThresholdAdapter;
import de.mannheim.uni.expansion.tablefusion.core.scoring.TimestampScorerCleanedUp;
import de.mannheim.uni.expansion.tablefusion.core.splitting.NonDuplicateSplitterFactory;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.AccuracyAnalyzer;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig.FusionConfigBuilder;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig.StrategiesConfigBuilder;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesFilter;
import de.mannheim.uni.expansion.tablefusion.modules.HeidelTimeModule;
import de.mannheim.uni.expansion.tablefusion.modules.MatchingModule;
import de.mannheim.uni.expansion.tablefusion.modules.RankingModule;
import de.mannheim.uni.expansion.tablefusion.modules.SimilarityModule;
import de.mannheim.uni.expansion.tablefusion.modules.TableFusionModule;
import de.mannheim.uni.expansion.tablefusion.modules.TablesModule;
import de.mannheim.uni.expansion.tablefusion.services.AnnotatedTableOtsuFilter;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategy;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategy.FusionStrategyId;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategyRepository;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.MaxDiffStrategy;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.ScorerRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.mannheim.uni.units.UnitParser;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.informatik.dws.expansion.matching.KnowledgeBaseModule;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;
import de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.worldbank.CountryListFixer;
import edu.stanford.nlp.util.MutableInteger;

public class Fusion {

	@Module(injects = { Fusion.class, AnnotatedTableOtsuFilter.class }, library = true, complete = false)
	public static class FusionModule {

	}

	public static void main(String[] args) throws IOException {
		String dataHome = null;

		if (args.length > 0 && args[0] != null && !args[0].equals("")) {
			dataHome = args[0];
		}

		String mappingsName = null;

		if (args.length > 0 && args[1] != null && !args[1].equals("")) {
			mappingsName = args[1];
		}

		mappingsName = "/tables/" + mappingsName;

		String kbPath = dataHome + "/kb/";
		String ontologyRepositoryFile = dataHome + "/ontology.kryo";
		String luceneIndexCache = dataHome + "/lucene/";

		String mappingsFolder = dataHome + "/" + mappingsName + "/mappings/";
		String cacheFolder = dataHome + "/" + mappingsName + "/cache/";
		String resultsFolder = dataHome + "/" + mappingsName + "/results/";
		String problemsFolder = dataHome + "/" + mappingsName + "/problems/";

		String optimizingFolder = dataHome + "/" + mappingsName + "/optimizing/";

		String columnMappingsFolder = dataHome + "/" + mappingsName + "/columnMappings/";
		String heidelTimeConfigPath = dataHome + "/heidelTime/config.props";
		String emptyFile = dataHome + "/empty.txt";
		String candidateSelectionConfig = dataHome + "/config.conf";
		String matchingConfig = dataHome + "/config.conf";
		UnitParser.unitPath = dataHome + "/Units/Convertible";
		String maxDiffsFolder = dataHome + "/" + "maxDiffs/";
		String maxDiffsFileQKBTMapped = dataHome + "/" + "maxDiffs_qkbt_mapped.json";
		String maxDiffsFileNewTS = dataHome + "/" + "maxDiffs_new_ts.json";
		String maxDiffsFileNewTSPropgated = dataHome + "/" + "maxDiffs_new_ts_propagated.json";

		String rankingsFile = dataHome + "/" + mappingsName + "/rankings/filtered.tsv.gz";

		KnowledgeBaseModule kbModule = new KnowledgeBaseModule(kbPath, ontologyRepositoryFile, luceneIndexCache);

		SimilarityModule similarityModule = new SimilarityModule();
		MatchingModule matchingModule = new MatchingModule(true, columnMappingsFolder, candidateSelectionConfig, matchingConfig, emptyFile);
		TablesModule tableModule = TablesModule.withCache(mappingsFolder, cacheFolder, true);
		HeidelTimeModule heidelTimeModule = new HeidelTimeModule(heidelTimeConfigPath);
		TableFusionModule tableFusionModule = new TableFusionModule(resultsFolder, problemsFolder, null, null, optimizingFolder, null);
		RankingModule rankingModule = new RankingModule(emptyFile, rankingsFile);

		List<KBValueCreator> valueCreatorList = new ArrayList<>();
		valueCreatorList.add(new CountryListFixer(dataHome + "/" + "countrycodes.txt"));
		kbModule.setValueCreatorList(valueCreatorList);
		// RUN WITHOUT OTSU FILTERING

		{
			ObjectGraph graph = ObjectGraph.create(tableFusionModule, kbModule, tableModule, heidelTimeModule, matchingModule, similarityModule, rankingModule,
					new FusionModule());

			KnowledgeBase kb = graph.get(KnowledgeBase.class);

			List<AnnotatedTableFilter> filters = new ArrayList<>();
			AnnotatedTableOtsuFilter tableOtsuFilter = graph.get(AnnotatedTableOtsuFilter.class);
			tableOtsuFilter.setMin(0.1);
			filters.add(tableOtsuFilter);

			filters.add(new GoodPropertiesFilter());
			filters.add(new KbInstanceFilter(kb));
			// filters.add(new SmallTableFilter());

			SourcesFilter sourcesFilter = new SourcesFilter();
			filters.add(sourcesFilter);
			tableModule.setFilters(filters);

			tableModule.setClassFilter(graph.get(KnowledgeBase.class));
			Fusion fusion = graph.get(Fusion.class);
			FusionConfig fusionConfig = FusionConfig.defaultConfig();
			String prefix = "NEW_TS_NO_HARMONIC_MEAN_TOTAL_OVERLAP_AVERAGE_MAX_SCORES_APPROXIMATE_WITHOUT_KBT_ADJUSTMENT";

			TimestampScorerCleanedUp.kbtAsBackup = false;

			FusionConfigBuilder rebuilder;

			// fusion.runFull(resultsFolder, "NORMAL", fusionConfig);
			//
			// {
			// // fusion.createMaxDiffsFile(fusionConfig,
			// // FusionStrategyRepository.NEW_TS_QKBT, maxDiffsFileNewTS);
			// // fusion.createMaxDiffsFile(fusionConfig,
			// // FusionStrategyRepository.NEW_TS_QKBT_PROPAGATED,
			// // maxDiffsFileNewTSPropgated);
			// // fusion.createMaxDiffsFile(fusionConfig,
			// // FusionStrategyRepository.QKBT_Mapped,
			// // maxDiffsFileQKBTMapped);
			//
			// int sdf = 123;
			// }
			//
			// fusion.setMaxDiffFromFile(FusionStrategyRepository.NEW_TS_QKBT,
			// maxDiffsFileNewTS);
			//
			// fusion.setMaxDiffFromFile(FusionStrategyRepository.NEW_TS_QKBT_PROPAGATED,
			// maxDiffsFileNewTSPropgated);
			//
			// fusion.setMaxDiffFromFile(FusionStrategyRepository.QKBT_Mapped,
			// maxDiffsFileQKBTMapped);
			MKDIR.createDirBoolean(new File(maxDiffsFolder));

			//
			rebuilder = fusionConfig.rebuilder();
			rebuilder.setMinSource(2);
			rebuilder.setAdjustedRecall(true);
			fusionConfig = rebuilder.build();
			prefix = "PROPAGETED_TEST";

			// fusion.learnAndStoreMaxDiff(fusionConfig, "6", maxDiffsFolder,
			// false, false);
			// fusion.learnAndStoreMaxDiff(fusionConfig, "6", maxDiffsFolder,
			// false, true);

			if (true && false) {
				fusion.learnAndStoreMaxDiff(fusionConfig, "7", maxDiffsFolder, true, true);
			}

			if (true && false) {
				fusion.learnAndStoreMaxDiff(fusionConfig, "6", maxDiffsFolder, true, false);

				System.out.println("Running with MaxDiff = 0 for normal and adjusted");
				fusion.setMaxDiffFromInteger(0);
				fusion.runFull(resultsFolder, "No_MaxDiff_normal", fusionConfig.rebuilder().setAdjustedRecall(false).build());
				fusion.runFull(resultsFolder, "No_MaxDiff_adjusted", fusionConfig.rebuilder().setAdjustedRecall(true).build());
			}

			if (true && false) {
				System.out.println("Running with MaxDiff per Property Non-Adjusted");
				fusion.loadMaxDiffs("6", maxDiffsFolder, true, false);
				fusion.runFull(resultsFolder, "MAXDIFF_6_normal", fusionConfig.rebuilder().setAdjustedRecall(false).build());
			}

			if (true && false) {
				System.out.println("Running with MaxDiff per Property Adjusted");
				fusion.loadMaxDiffs("7", maxDiffsFolder, true, true);
				fusion.runFull(resultsFolder, "MAXDIFF_7_adjusted", fusionConfig.rebuilder().setAdjustedRecall(true).build());
			}
			if (true && false) {
				fusion.setMaxDiffFromInteger(0);
				fusion.runFull(resultsFolder, "NO_MAX_DIFF_ADJUSTED_CORRECTED", fusionConfig.rebuilder().setAdjustedRecall(true).build());
			}
			if (true && false) {
				fusion.setMaxDiffFromInteger(0);
				fusion.runFull(resultsFolder, "NO_MAX_DIFF_NORMAL", fusionConfig.rebuilder().setAdjustedRecall(false).build());
			}

			System.out.println("Running with MaxDiff per Property Adjusted");
			fusion.loadMaxDiffs("7", maxDiffsFolder, true, true);
			fusion.runFull(resultsFolder, "RUNNING_FOT_TS_USE_DELETE", fusionConfig.rebuilder().setAdjustedRecall(true).build());
			// fusion.runFull(resultsFolder,
			// "Timed-KBT-NO-MaxDiff-F025-corrected",
			// fusionConfig.rebuilder().setAdjustedRecall(true).build());

			// fusion.learnAndStoreMaxDiff(fusionConfig, "5", maxDiffsFolder,
			// true, false);
			// fusion.loadMaxDiffs("5", maxDiffsFolder, true, false);
			//
			// fusion.runFull(resultsFolder, "FULL_VARIANCE_RUN_2",
			// fusionConfig);

			// sourcesFilter.printCount();
			// fusion.runReplicateSimple(resultsFolder, prefix, fusionConfig);
			// fusion.runReplicate(resultsFolder, prefix, fusionConfig);
			// fusion.runSplit(resultsFolder, prefix, fusionConfig);

			// fusion.runFull(resultsFolder, "MAXDIFF_NEW", rebuilder.build(),
			// maxDiffsFolder, true);
			// fusion.runFull(resultsFolder, "MAXDIFF", fusionConfig,
			// maxDiffsFolder, true);

			// prefix = "MAXDIFF_NEW_2";
			// fusion.learnMaxDiff(resultsFolder, prefix, fusionConfig,
			// maxDiffsFolder, false);
			// fusion.learnMaxDiff(resultsFolder, prefix, fusionConfig,
			// maxDiffsFolder, true);
			//
			// prefix = "MAXDIFF_NEW_2_ADJUSTED";
			// rebuilder = fusionConfig.rebuilder();
			// rebuilder.setAdjustedRecall(true);
			// fusionConfig = rebuilder.build();
			// fusion.learnMaxDiff(resultsFolder, prefix, fusionConfig,
			// maxDiffsFolder, false);
			// fusion.learnMaxDiff(resultsFolder, prefix, fusionConfig,
			// maxDiffsFolder, true);

			// rebuilder = fusionConfig.rebuilder();
			// rebuilder.setAdjustedRecall(true);
			// fusionConfig = rebuilder.build();
			//
			// fusion.setMaxDiffFromFile(FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL,
			// maxDiffsFolder + "/maxDiff_new_ts_mapping_" +
			// "MAXDIFF_NEW_ADJUSTED" + "_" + "perProperty" + ".json");
			//
			// //
			// fusion.setMaxDiffFromFile(FusionStrategyRepository.QKBT_Mapped,
			// maxDiffsFolder + "/maxDiff_new_timed_kbt_" +
			// "MAXDIFF_NEW_ADJUSTED" + "_" + "perProperty" + ".json");
			//
			// prefix = "YEAR_FULL_RUN";
			// fusion.runFull(resultsFolder, prefix, fusionConfig,
			// maxDiffsFolder, false);

			// fusion.runReplicateSimple(resultsFolder, prefix, fusionConfig);
			// fusion.runReplicate(resultsFolder, prefix, fusionConfig);
			// fusion.runSplit(resultsFolder, prefix, fusionConfig);

			// fusion.runFull(resultsFolder, "MAXDIFF_NEW_ADJUSTED",
			// rebuilder.build(), maxDiffsFolder, true);
			// fusion.runFull(resultsFolder, "MAXDIFF_ADJUSTED_RECALL",
			// rebuilder.build(), maxDiffsFolder, true);

			// rebuilder = fusionConfig.rebuilder();
			// rebuilder.setReSplitTestingSet(true);
			// fusion.runFull(resultsFolder, "MAXDIFF_RESPLIT",
			// rebuilder.build(), maxDiffsFolder, false);
			// // fusion.runFull(resultsFolder, "MAXDIFF_RESPLIT",
			// // rebuilder.build(), maxDiffsFolder, true);
			//
			// rebuilder = fusionConfig.rebuilder();
			// rebuilder.setAdjustedRecall(true);
			// rebuilder.setReSplitTestingSet(true);
			// fusion.runFull(resultsFolder, "MAXDIFF_ADJUSTED_RECALL_RESPLIT",
			// rebuilder.build(), maxDiffsFolder, false);
			// fusion.runFull(resultsFolder, "MAXDIFF_ADJUSTED_RECALL_RESPLIT",
			// rebuilder.build(), maxDiffsFolder, true);

			// fusion.runReplicateSimple(resultsFolder, prefix, fusionConfig);
			// fusion.runReplicate(resultsFolder, prefix, fusionConfig);
			// fusion.runSplit(resultsFolder, prefix, fusionConfig);
			// fusion.runFull(resultsFolder, prefix, fusionConfig);

			// fusion.runReplicate("_NEW_THRESHOLD", thresholdAdapter);

		}

		// // RUN WITH OTSU FILTERING
		// {
		// List<AnnotatedTableFilter> filters = new ArrayList<>();
		// filters.add();
		// tableModule.setFilters(filters);
		// ObjectGraph graph = ObjectGraph.create(tableFusionModule, kbModule,
		// tableModule, heidelTimeModule, matchingModule, similarityModule,
		// new FusionModule());
		// Fusion fusion = graph.get(Fusion.class);
		// fusion.run();
		// }

		// kbModule.setOntologyReplacer((ontologyRepository) -> {
		// return new
		// ProcessGzToKryoWithTimestampsAndColumnMatching.OntologyRepositoryWrapperReference(ontologyRepository);
		// });
		//
		// List<KBValueCreator> valueCreatorList = new ArrayList<>();
		//
		// valueCreatorList.add(new
		// ProcessGzToKryoWithTimestampsAndColumnMatching.CreateReferencePropertiesRange(graph.get(OntologyRepository.class)));
		// kbModule.setValueCreatorList(valueCreatorList);
		// testSerialize(new ThreadSafeNoReferenceKryo(),
		// graph.get(AnnotatedTableParserJSON.class), dataHome + "/" +
		// mappingsName + "/");

		try {
			PersistentParallel4.close();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		// System.exit(0);

	}

	private void loadMaxDiffs(String maxDiffPrefix, String maxDiffFolder, boolean maxDiffPerProperty, boolean adjusted) throws IOException {
		this.setMaxDiffFromFile(FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL,
				maxDiffFolder + "/" + getMaxDiffFileName(maxDiffPrefix, FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL, maxDiffPerProperty, adjusted));

		//
		this.setMaxDiffFromFile(FusionStrategyRepository.QKBT_Mapped,
				maxDiffFolder + "/" + getMaxDiffFileName(maxDiffPrefix, FusionStrategyRepository.QKBT_Mapped, maxDiffPerProperty, adjusted));

		// this.setMaxDiffFromFile(FusionStrategyRepository.NEW_TS_QKBT_WITHOUR_ROW_CELL_PROPAGATED,
		// maxDiffFolder + "/"
		// + getMaxDiffFileName(maxDiffPrefix,
		// FusionStrategyRepository.NEW_TS_QKBT_WITHOUR_ROW_CELL_PROPAGATED,
		// maxDiffPerProperty, adjusted));

	}

	private void setMaxDiffFromInteger(Integer maxDiff) throws IOException {
		this.setMaxDiff(FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL, maxDiff);

		//
		this.setMaxDiff(FusionStrategyRepository.QKBT_Mapped, maxDiff);

		// this.setMaxDiff(FusionStrategyRepository.NEW_TS_QKBT_WITHOUR_ROW_CELL_PROPAGATED,
		// maxDiff);
	}

	private void setMaxDiffFromFile(FusionStrategy fusionStrategy, String maxDiffsFile) throws IOException {
		String maxDiffsJson = FileUtils.readFileToString(new File(maxDiffsFile));

		Map<ClassPropertyId, Integer> maxDiffs = jsonToMaxDiff(maxDiffsJson);

		if (maxDiffs.containsKey(this.allClasProperty)) {
			setMaxDiff(fusionStrategy, maxDiffs.get(this.allClasProperty));
		} else {
			setMaxDiff(fusionStrategy, maxDiffs);
		}

	}

	private void setMaxDiff(FusionStrategy strategy, Integer maxDiff) {
		Scorer scorer = this.scorerRepository.get(strategy.getScoringMethod());
		if (scorer instanceof MaxDiffStrategy) {
			MaxDiffStrategy scorerTyped = (MaxDiffStrategy) scorer;
			scorerTyped.setMaxDiffsContainer((classPropertyId) -> {
				return Optional.of(maxDiff);

			});
		}

	}

	private void createMaxDiffsFile(FusionConfig fusionConfig, FusionStrategy fusionStrategy, String maxDiffsFile, boolean maxDiffPerProperty)
			throws IOException {
		Map<ClassPropertyId, Integer> maxDiffMap;
		if (maxDiffPerProperty) {
			maxDiffMap = findMaxDiff(fusionStrategy, fusionConfig);
		} else {
			maxDiffMap = findMaxDiffGeneral(fusionStrategy, fusionConfig);
		}

		System.out.println(fusionStrategy.getName());
		System.out.println(maxDiffMap.toString());
		String json = maxDiffToJson(maxDiffMap);

		FileUtils.writeStringToFile(new File(maxDiffsFile), json);
		System.out.println(json);
		Map<ClassPropertyId, Integer> maxDiffMapNew = jsonToMaxDiff(json);

		if (maxDiffMap.equals(maxDiffMapNew)) {
			System.out.println("EVERYTHING GOOD");
		} else {

			System.out.println("EVERYTHING BAD");
		}

	}

	// private void storeMappings(String file) {
	// Scorer scorer =
	// this.scorerRepository.get(FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL.getScoringMethod());
	//// FusionStrategyId fusionStrategyId =
	// FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL.getFusionStrategyId();
	//
	// }
	private void setMaxDiff(FusionStrategy strategy, Map<ClassPropertyId, Integer> maxDiffs) {
		Scorer scorer = this.scorerRepository.get(strategy.getScoringMethod());
		if (scorer instanceof MaxDiffStrategy) {
			MaxDiffStrategy scorerTyped = (MaxDiffStrategy) scorer;
			final Map<ClassPropertyId, Integer> maxDiffsFinal = ImmutableMap.copyOf(maxDiffs);
			scorerTyped.setMaxDiffsContainer((classPropertyId) -> {
				Integer maxDiff = maxDiffsFinal.get(classPropertyId);
				if (maxDiff != null) {
					return Optional.of(maxDiff);
				}
				return Optional.empty();

			});
		}

	}

	private Map<ClassPropertyId, Integer> findMaxDiff(FusionStrategy fusionStrategy, FusionConfig fusionConfig) {
		try {
			Scorer scorer = this.scorerRepository.get(fusionStrategy.getScoringMethod());
			FusionStrategyId fusionStrategyId = fusionStrategy.getFusionStrategyId();
			if (scorer instanceof MaxDiffStrategy) {
				MaxDiffStrategy scorerTyped = (MaxDiffStrategy) scorer;

				Map<ClassPropertyId, MutableInteger> maxScoreMaxDiff = new HashMap<>();

				// Map<ClassPropertyId, MutableDouble> maxPrecision = new
				// HashMap<>();
				Map<ClassPropertyId, MutableDouble> maxFmeasure = new HashMap<>();
				// Map<ClassPropertyId, MutableDouble> maxRecall = new
				// HashMap<>();

				for (int i = 0; i < 5; i++) {
					final Optional<Integer> maxDiff = Optional.of(i);
					scorerTyped.setMaxDiffsContainer((classPropertyId) -> {
						return maxDiff;
					});

					StrategiesConfigBuilder strategiesConfigBuilder = StrategiesConfig.builder();
					strategiesConfigBuilder.addStrategy(fusionStrategy);

					try {
						Map<FusionStrategyId, FusionPerformance<PerformancePR>> scores = accuracyAnalyzer.getPerformanceReturnCrossValidation(fusionConfig,
								strategiesConfigBuilder.build());

						FusionPerformance<PerformancePR> performance = scores.get(fusionStrategyId);
						for (Entry<ClassPropertyId, PerformancePR> entry : performance.getPerProperty().entrySet()) {
							// MutableDouble currentMaxPrecision =
							// CHMHelper.getFromMap(maxPrecision,
							// entry.getKey(), () -> new MutableDouble(0.0));
							MutableDouble currentMaxFmeasure = CHMHelper.getFromMap(maxFmeasure, entry.getKey(), () -> new MutableDouble(0.0));
							// MutableDouble currentMaxRecall =
							// CHMHelper.getFromMap(maxRecall, entry.getKey(),
							// () -> new MutableDouble(0.0));
							MutableInteger currentMaxDiff = CHMHelper.getFromMap(maxScoreMaxDiff, entry.getKey(), () -> new MutableInteger(0));

							if (fusionConfig.isAdjustedRecall()) {
								if (entry.getValue().getaFMeasure() > currentMaxFmeasure.doubleValue()) {
									currentMaxFmeasure.setValue(entry.getValue().getaFMeasure());
									// currentMaxPrecision.setValue(entry.getValue().getPrecision());
									// currentMaxRecall.setValue(entry.getValue().getaRecall());
									currentMaxDiff.set(i);
								}
							} else {
								if (entry.getValue().getfMeasure() > currentMaxFmeasure.doubleValue()) {
									currentMaxFmeasure.setValue(entry.getValue().getfMeasure());
									// currentMaxPrecision.setValue(entry.getValue().getPrecision());
									// currentMaxRecall.setValue(entry.getValue().getRecall());
									currentMaxDiff.set(i);
								}
							}
						}

					} catch (IOException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}

				System.out.println(maxScoreMaxDiff.toString());

				Map<ClassPropertyId, Integer> genericMap = maxScoreMaxDiff.entrySet().stream().collect(Collectors.toMap((e) -> {
					return e.getKey();
				} , (e) -> {
					return e.getValue().intValue();
				}));

				return genericMap;

			} else {
				throw new IllegalStateException();
			}
		} catch (Exception e) {
			e.printStackTrace();

		}

		int sdf = 123;
		return null;
	}

	private static final ClassPropertyId allClasProperty = new ClassPropertyId(new ClassId("ALL"), new PropertyId("ALL"));

	private Map<ClassPropertyId, Integer> findMaxDiffGeneral(FusionStrategy fusionStrategy, FusionConfig fusionConfig) {
		try {
			Scorer scorer = this.scorerRepository.get(fusionStrategy.getScoringMethod());
			FusionStrategyId fusionStrategyId = fusionStrategy.getFusionStrategyId();
			if (scorer instanceof MaxDiffStrategy) {
				MaxDiffStrategy scorerTyped = (MaxDiffStrategy) scorer;

				Map<ClassPropertyId, MutableInteger> maxScoreMaxDiff = new HashMap<>();

				// Map<ClassPropertyId, MutableDouble> maxPrecision = new
				// HashMap<>();
				Map<ClassPropertyId, MutableDouble> maxFmeasure = new HashMap<>();
				// Map<ClassPropertyId, MutableDouble> maxRecall = new
				// HashMap<>();

				// MutableDouble currentMaxPrecision =
				// CHMHelper.getFromMap(maxPrecision, allClasProperty, () -> new
				// MutableDouble(0.0));
				MutableDouble currentMaxFmeasure = CHMHelper.getFromMap(maxFmeasure, allClasProperty, () -> new MutableDouble(0.0));
				// MutableDouble currentMaxRecall =
				// CHMHelper.getFromMap(maxRecall, allClasProperty, () -> new
				// MutableDouble(0.0));
				MutableInteger currentMaxDiff = CHMHelper.getFromMap(maxScoreMaxDiff, allClasProperty, () -> new MutableInteger(0));

				for (int i = 0; i < 5; i++) {
					final Optional<Integer> maxDiff = Optional.of(i);
					scorerTyped.setMaxDiffsContainer((classPropertyId) -> {
						return maxDiff;
					});

					StrategiesConfigBuilder strategiesConfigBuilder = StrategiesConfig.builder();
					strategiesConfigBuilder.addStrategy(fusionStrategy);

					try {
						Map<FusionStrategyId, FusionPerformance<PerformancePR>> scores = accuracyAnalyzer.getPerformanceReturnCrossValidation(fusionConfig,
								strategiesConfigBuilder.build());

						FusionPerformance<PerformancePR> performanceContainer = scores.get(fusionStrategyId);
						PerformancePR performance = performanceContainer.getTotal();

						if (fusionConfig.isAdjustedRecall()) {
							if (performance.getaFMeasure() > currentMaxFmeasure.doubleValue()) {
								currentMaxFmeasure.setValue(performance.getaFMeasure());
								// currentMaxPrecision.setValue(performance.getPrecision());
								// currentMaxRecall.setValue(performance.getaRecall());
								currentMaxDiff.set(i);
							}
						} else {
							if (performance.getfMeasure() > currentMaxFmeasure.doubleValue()) {
								currentMaxFmeasure.setValue(performance.getfMeasure());
								// currentMaxPrecision.setValue(performance.getPrecision());
								// currentMaxRecall.setValue(performance.getRecall());
								currentMaxDiff.set(i);
							}
						}

					} catch (IOException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}

				System.out.println(maxScoreMaxDiff.toString());

				Map<ClassPropertyId, Integer> genericMap = maxScoreMaxDiff.entrySet().stream().collect(Collectors.toMap((e) -> {
					return e.getKey();
				} , (e) -> {
					return e.getValue().intValue();
				}));

				return genericMap;

			} else {
				throw new IllegalStateException();
			}
		} catch (Exception e) {
			e.printStackTrace();

		}

		int sdf = 123;
		return null;
	}

	private String maxDiffToJson(Map<ClassPropertyId, Integer> genericMap) {
		Map<ClassId, JSONObject> classObjects = new HashMap<>();
		for (Entry<ClassPropertyId, Integer> e : genericMap.entrySet()) {
			CHMHelper.getFromMap(classObjects, e.getKey().getClassId(), () -> new JSONObject()).put(e.getKey().getPropertyId().toCombinedString(),
					e.getValue());
		}

		JSONObject finalObject = new JSONObject();
		for (Entry<ClassId, JSONObject> e : classObjects.entrySet()) {
			finalObject.put(e.getKey().getClassId(), e.getValue());
		}

		return finalObject.toString();
	}

	private Map<ClassPropertyId, Integer> jsonToMaxDiff(String json) {
		Map<ClassPropertyId, Integer> ret = new HashMap<>();

		JSONObject completeObject = new JSONObject(json);

		for (Iterator<String> keys = completeObject.keys(); keys.hasNext();) {
			String key = keys.next();
			ClassId classId = new ClassId(key);
			JSONObject classObject = completeObject.getJSONObject(key);

			for (Iterator<String> innerKeys = classObject.keys(); innerKeys.hasNext();) {
				String propertyKey = innerKeys.next();
				PropertyId propertyId = PropertyId.fromCombinedString(propertyKey);
				ret.put(new ClassPropertyId(classId, propertyId), classObject.getInt(propertyKey));
			}

		}

		return ret;
	}

	private final AccuracyAnalyzer accuracyAnalyzer;
	private final ScorerRepository scorerRepository;

	@Inject
	public Fusion(AccuracyAnalyzer accuracyAnalyzer, ScorerRepository scorerRepository) {
		this.accuracyAnalyzer = accuracyAnalyzer;
		this.scorerRepository = scorerRepository;
	}

	private static final FusionScoreThresholdAdapter thresholdAdapter = (score, count, threshold) -> {
		double ret = Math.min(score / (count * threshold) + (score - count * threshold), score);
		return ret;
	};

	public void runReplicate(String resultsFolder, String prefix, FusionConfig fusionConfig) throws IOException {

		// List<Integer> permittedYears = new ArrayList<>();
		// permittedYears.add(2008);
		// permittedYears.add(2010);
		// permittedYears.add(2012);
		// permittedYears.add(2014);
		// SplitGroupedValuesFilter splitGroupedValuesFilter = new
		// SplitGroupedValuesFilter((e) -> {
		// return permittedYears.contains(e.getYear());
		// });

		FusionConfigBuilder builder = fusionConfig.rebuilder();
		// builder.setSplitGroupedValuesFilter(splitGroupedValuesFilter);
		builder.setSplitterFactory(NonDuplicateSplitterFactory.class);

		accuracyAnalyzer.getPerformance(prefix + "_Replication", builder.build(), resultsFolder);

	}

	public void runReplicateSimple(String resultsFolder, String prefix, FusionConfig fusionConfig) throws IOException {

		List<Integer> permittedYears = new ArrayList<>();
		permittedYears.add(2008);
		permittedYears.add(2010);
		permittedYears.add(2012);
		permittedYears.add(2014);
		SplitGroupedValuesFilter splitGroupedValuesFilter = new SplitGroupedValuesFilter((e) -> {
			return permittedYears.contains(e.getYear());
		});

		FusionConfigBuilder builder = fusionConfig.rebuilder();
		builder.setSplitGroupedValuesFilter(splitGroupedValuesFilter);
		// builder.setSplitterFactory(NonDuplicateSplitterFactory.class);

		accuracyAnalyzer.getPerformance(prefix + "_ReplicationSimple", builder.build(), resultsFolder);

	}

	public String getMaxDiffFileName(String maxDiffPrefix, FusionStrategy fusionStrategy, boolean maxDiffPerProperty, boolean adjusted) {
		String strategy = null;
		if (fusionStrategy.getFusionStrategyId().equals(FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL.getFusionStrategyId())) {
			strategy = "ts_mapping";
		} else if (fusionStrategy.getFusionStrategyId().equals(FusionStrategyRepository.NEW_TS_QKBT_WITHOUR_ROW_CELL_PROPAGATED.getFusionStrategyId())) {
			strategy = "ts_mapping_propagated";
		} else if (fusionStrategy.getFusionStrategyId().equals(FusionStrategyRepository.QKBT_Mapped.getFusionStrategyId())) {
			strategy = "timed_kbt";
		}

		if (strategy != null) {
			return "maxDiff_" + maxDiffPrefix + "_" + strategy + "_" + (adjusted ? "adjusted" : "default") + "_"
					+ (maxDiffPerProperty ? "perProperty" : "general") + ".json";
		} else {
			return null;
		}

	}

	public void learnAndStoreMaxDiff(FusionConfig fusionConfig, String maxDiffPrefix, String maxDiffFolder, boolean maxDiffPerProperty, boolean adjusted)
			throws IOException {

		fusionConfig = fusionConfig.rebuilder().setAdjustedRecall(adjusted).build();

		createMaxDiffsFile(fusionConfig, FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL,
				maxDiffFolder + "/" + getMaxDiffFileName(maxDiffPrefix, FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL, maxDiffPerProperty, adjusted),
				maxDiffPerProperty);
		// createMaxDiffsFile(fusionConfig,
		// FusionStrategyRepository.NEW_TS_QKBT_WITHOUR_ROW_CELL_PROPAGATED,
		// maxDiffFolder + "/"
		// + getMaxDiffFileName(maxDiffPrefix,
		// FusionStrategyRepository.NEW_TS_QKBT_WITHOUR_ROW_CELL_PROPAGATED,
		// maxDiffPerProperty, adjusted),
		// maxDiffPerProperty);
		createMaxDiffsFile(fusionConfig, FusionStrategyRepository.QKBT_Mapped,
				maxDiffFolder + "/" + getMaxDiffFileName(maxDiffPrefix, FusionStrategyRepository.QKBT_Mapped, maxDiffPerProperty, adjusted),
				maxDiffPerProperty);
				//
				// setMaxDiffFromFile(FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL,
				// maxDiffFolder + "/maxDiff_new_ts_mapping_" + prefix + "_" +
				// (maxDiffPerProperty ? "perProperty" : "general") + ".json");

		// setMaxDiffFromFile(FusionStrategyRepository.NEW_TS_QKBT_WITHOUR_ROW_CELL_PROPAGATED,
		// maxDiffFolder + "/maxDiff_ts_mapping_propagated_" + prefix + "_" +
		// (maxDiffPerProperty ? "perProperty" : "general") + ".json");

		// setMaxDiffFromFile(FusionStrategyRepository.QKBT_Mapped,
		// maxDiffFolder + "/maxDiff_new_timed_kbt_" + prefix + "_" +
		// (maxDiffPerProperty ? "perProperty" : "general") + ".json");

		// TimestampScorerCleanedUp.kbtAsBackup = false;
		// setMaxDiff(FusionStrategyRepository.QKBT_Mapped, 0);
		// setMaxDiff(FusionStrategyRepository.NEW_TS_QKBT, 0);
		// setMaxDiff(FusionStrategyRepository.NEW_TS_QKBT_PROPAGATED, 0);
		// accuracyAnalyzer.getPerformance(prefix + "_" + (maxDiffPerProperty ?
		// "perProperty" : "general") + "_Full_NO_MAX_DIFF", fusionConfig,
		// resultsFolder);

		// TimestampScorerCleanedUp.kbtAsBackup = false;
		// setMaxDiff(FusionStrategyRepository.QKBT_Mapped, 0);
		// setMaxDiff(FusionStrategyRepository.NEW_TS_QKBT, 0);
		// setMaxDiff(FusionStrategyRepository.NEW_TS_QKBT_PROPAGATED, 0);
		// accuracyAnalyzer.getPerformance(prefix + "_Full_NO_KBT_BACKUP",
		// fusionConfig, resultsFolder);

	}

	public void runFull(String resultsFolder, String prefix, FusionConfig fusionConfig) throws IOException {

		// createMaxDiffsFile(fusionConfig,
		// FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL,
		// maxDiffFolder + "/maxDiff_new_ts_mapping_" + prefix + "_" +
		// (maxDiffPerProperty ? "perProperty" : "general") + ".json",
		// maxDiffPerProperty);
		// createMaxDiffsFile(fusionConfig,
		// FusionStrategyRepository.NEW_TS_QKBT_PROPAGATED,
		// maxDiffFolder + "/maxDiff_ts_mapping_propagated_" + prefix + "_" +
		// (maxDiffPerProperty ? "perProperty" : "general") + ".json",
		// maxDiffPerProperty);
		// createMaxDiffsFile(fusionConfig,
		// FusionStrategyRepository.QKBT_Mapped,
		// maxDiffFolder + "/maxDiff_new_timed_kbt_" + prefix + "_" +
		// (maxDiffPerProperty ? "perProperty" : "general") + ".json",
		// maxDiffPerProperty);
		//
		// setMaxDiffFromFile(FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL,
		// maxDiffFolder + "/maxDiff_new_ts_mapping_" + prefix + "_" +
		// (maxDiffPerProperty ? "perProperty" : "general") + ".json");

		// setMaxDiffFromFile(FusionStrategyRepository.NEW_TS_QKBT_PROPAGATED,
		// maxDiffFolder + "/maxDiff_ts_mapping_propagated_" + prefix + "_" +
		// (maxDiffPerProperty ? "perProperty" : "general") + ".json");

		// setMaxDiffFromFile(FusionStrategyRepository.QKBT_Mapped,
		// maxDiffFolder + "/maxDiff_new_timed_kbt_" + prefix + "_" +
		// (maxDiffPerProperty ? "perProperty" : "general") + ".json");

		// setMaxDiff(FusionStrategyRepository.QKBT_Mapped, 0);
		// setMaxDiff(FusionStrategyRepository.NEW_TS_QKBT, 0);
		// setMaxDiff(FusionStrategyRepository.NEW_TS_QKBT_PROPAGATED, 0);
		accuracyAnalyzer.getPerformance(prefix + "_Full", fusionConfig, resultsFolder);

		// TimestampScorerCleanedUp.kbtAsBackup = false;
		// setMaxDiff(FusionStrategyRepository.QKBT_Mapped, 0);
		// setMaxDiff(FusionStrategyRepository.NEW_TS_QKBT, 0);
		// setMaxDiff(FusionStrategyRepository.NEW_TS_QKBT_PROPAGATED, 0);
		// accuracyAnalyzer.getPerformance(prefix + "_Full_NO_KBT_BACKUP",
		// fusionConfig, resultsFolder);

	}

	public void runSplit(String resultsFolder, String prefix, FusionConfig fusionConfig) throws IOException {

		SplitGroupedValuesFilter splitGroupedValuesFilter = new SplitGroupedValuesFilter((e) -> {
			return e.getYear() < 2000;
		});

		FusionConfigBuilder builder = fusionConfig.rebuilder();
		builder.setSplitGroupedValuesFilter(splitGroupedValuesFilter);

		accuracyAnalyzer.getPerformance(prefix + "_SplitBefore2000", builder.build(), resultsFolder);

		splitGroupedValuesFilter = new SplitGroupedValuesFilter((e) -> {
			return e.getYear() >= 2000;
		});

		builder.setSplitGroupedValuesFilter(splitGroupedValuesFilter);

		accuracyAnalyzer.getPerformance(prefix + "_SplitAfter2000", builder.build(), resultsFolder);

	}

}
