package de.uni_mannheim.informatik.dws.expansion.samples.postMatching;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.uni_mannheim.informatik.dws.expansion.wikidata.WikidataClasses;
import de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats.Stats;

public class GoodPropertiesFilter implements AnnotatedTableFilter {

	static Set<ClassId> classes = new HashSet<>();

	static Set<PropertyId> properties = new HashSet<>();

	static Set<ClassPropertyId> classProperties = new HashSet<>();

	static {

		classes.add(WikidataClasses.wikidataCountry);
		classes.add(WikidataClasses.wikidataFootball);
		classes.add(WikidataClasses.wikidataSoccer);
		classes.add(WikidataClasses.wikidataBasketball);
		classes.add(WikidataClasses.wikidataCity);
		// classes.add(WikidataClasses.wikidataTradedCompany);

		properties.add(Stats.population);
		properties.add(Stats.populationDensity);
		properties.add(Stats.teams);
		properties.add(Stats.sportNumber);
		properties.add(Stats.nominalGDP);
		properties.add(Stats.nominalGDPperCapita);
		// properties.add(Stats.netIncome);

		classProperties.add(new ClassPropertyId(WikidataClasses.wikidataCountry, Stats.population));
		classProperties.add(new ClassPropertyId(WikidataClasses.wikidataCountry, Stats.populationDensity));
		classProperties.add(new ClassPropertyId(WikidataClasses.wikidataCountry, Stats.nominalGDP));
		classProperties.add(new ClassPropertyId(WikidataClasses.wikidataCountry, Stats.nominalGDPperCapita));
		classProperties.add(new ClassPropertyId(WikidataClasses.wikidataCity, Stats.population));
		classProperties.add(new ClassPropertyId(WikidataClasses.wikidataFootball, Stats.teams));
		classProperties.add(new ClassPropertyId(WikidataClasses.wikidataFootball, Stats.sportNumber));
		classProperties.add(new ClassPropertyId(WikidataClasses.wikidataSoccer, Stats.teams));
		classProperties.add(new ClassPropertyId(WikidataClasses.wikidataBasketball, Stats.teams));
	}

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {
		if (classes.contains(data.getMetaData().getDbPediaClass())) {
			Map<Integer, List<PropertyId>> oldProperties = data.getMetaData().getAllProperties();
			Map<Integer, List<PropertyId>> newProperties = new HashMap<>();
			for (Entry<Integer, List<PropertyId>> e : oldProperties.entrySet()) {
				for (PropertyId propertyId : e.getValue()) {
					if (classProperties.contains(new ClassPropertyId(data.getMetaData().getDbPediaClass(), propertyId))) {
						CHMHelper.getFromMap(newProperties, e.getKey(), () -> new ArrayList<>()).add(propertyId);
					}
				}
			}

			if (newProperties.size() == 0) {
				return Optional.empty();
			} else {
				AnnotatedTableMetaData md = data.getMetaData();
				md = new AnnotatedTableMetaData(md.getSource(), md.getDbPediaClass(), md.getClassConf(), md.getNumHeaderRows(), md.getNumberOfRows(),
						newProperties, md.getAllPropertyScores(), md.getAllInstances(), md.getInstancesScores(), md.getKeyColumn(), md.getYearsBeforeText(),
						md.getYearsAfterText(), md.getYearsInContextBefore(), md.getYearsInTableTitle(), md.getYearsInPageTitle(), md.getPropertiesYears(),
						md.getPageTitle(), md.getTableTitle(), md.getTableContextTimeStampBeforeTable(), md.getTextBeforeTable(), md.getTextAfterTable());

				return Optional.of(new AnnotatedTablesReaderData(data.getId(), md, data.getTableData()));
			}

		} else {
			return Optional.empty();
		}
	}

}
