package de.uni_mannheim.informatik.dws.expansion.samples.postMatching;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;

import au.com.bytecode.opencsv.CSVReader;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.mannheim.uni.expansion.utils.CHMHelper;

public class GoodPropertiesFilterAll implements AnnotatedTableFilter {

	private final Set<ClassId> classes;
	private final Set<ClassPropertyId> classProperties;

	public GoodPropertiesFilterAll(String goodPropertiesFile) throws IOException {
		CSVReader reader = new CSVReader(new FileReader(new File(goodPropertiesFile)));
		try {
			classes = new HashSet<>();
			classProperties = new HashSet<>();

			String[] line = null;
			while ((line = reader.readNext()) != null) {
				ClassId classId = new ClassId(line[0]);
				PropertyId propertyId = new PropertyId(line[1]);

				classes.add(classId);
				classProperties.add(new ClassPropertyId(classId, propertyId));

			}

		} finally {
			reader.close();
		}
	}

	public GoodPropertiesFilterAll(Map<ClassId, Set<PropertyId>> input) throws IOException {

		classes = new HashSet<>();
		classProperties = new HashSet<>();
		for (Entry<ClassId, Set<PropertyId>> classEntry : input.entrySet()) {
			classes.add(classEntry.getKey());
			for (PropertyId propertyId : classEntry.getValue()) {
				classProperties.add(new ClassPropertyId(classEntry.getKey(), propertyId));
			}
		}

	}

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {
		if (classes.contains(data.getMetaData().getDbPediaClass())) {
			Map<Integer, List<PropertyId>> oldProperties = data.getMetaData().getAllProperties();
			Map<Integer, List<PropertyId>> newProperties = new HashMap<>();
			for (Entry<Integer, List<PropertyId>> e : oldProperties.entrySet()) {
				for (PropertyId propertyId : e.getValue()) {
					if (classProperties.contains(new ClassPropertyId(data.getMetaData().getDbPediaClass(), propertyId))) {
						CHMHelper.getFromMap(newProperties, e.getKey(), () -> new ArrayList<>()).add(propertyId);
					}
				}
			}

			// if (newProperties.size() == 0) {
			// return Optional.empty();
			// } else {
			AnnotatedTableMetaData md = data.getMetaData();
			md = new AnnotatedTableMetaData(md.getSource(), md.getDbPediaClass(), md.getClassConf(), md.getNumHeaderRows(), md.getNumberOfRows(), newProperties,
					md.getAllPropertyScores(), md.getAllInstances(), md.getInstancesScores(), md.getKeyColumn(), md.getYearsBeforeText(),
					md.getYearsAfterText(), md.getYearsInContextBefore(), md.getYearsInTableTitle(), md.getYearsInPageTitle(), md.getPropertiesYears(),
					md.getPageTitle(), md.getTableTitle(), md.getTableContextTimeStampBeforeTable(), md.getTextBeforeTable(), md.getTextAfterTable());

			return Optional.of(new AnnotatedTablesReaderData(data.getId(), md, data.getTableData()));
			// }

		} else {
			return Optional.empty();
		}
	}

}
