package de.uni_mannheim.informatik.dws.expansion.samples.postMatching;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;

public class KbInstanceFilter implements AnnotatedTableFilter {

	private final KnowledgeBase kb;

	public KbInstanceFilter(KnowledgeBase kb) {
		this.kb = kb;
	}

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {
		ClassId classId = data.getMetaData().getDbPediaClass();
		Map<Integer, List<InstanceId>> oldInstances = data.getMetaData().getAllInstances();
		Map<Integer, List<InstanceId>> newInstances = new HashMap<>();
		for (Entry<Integer, List<InstanceId>> e : oldInstances.entrySet()) {
			for (InstanceId instanceId : e.getValue()) {
				if (kb.hasInstance(classId, instanceId)) {
					CHMHelper.getFromMap(newInstances, e.getKey(), () -> new ArrayList<>()).add(instanceId);
				}
			}
		}

		if (newInstances.size() == 0) {
			return Optional.empty();
		} else {
			AnnotatedTableMetaData md = data.getMetaData();
			md = new AnnotatedTableMetaData(md.getSource(), md.getDbPediaClass(), md.getClassConf(), md.getNumHeaderRows(), md.getNumberOfRows(),
					md.getAllProperties(), md.getAllPropertyScores(), newInstances, md.getInstancesScores(), md.getKeyColumn(), md.getYearsBeforeText(),
					md.getYearsAfterText(), md.getYearsInContextBefore(), md.getYearsInTableTitle(), md.getYearsInPageTitle(), md.getPropertiesYears(),
					md.getPageTitle(), md.getTableTitle(), md.getTableContextTimeStampBeforeTable(), md.getTextBeforeTable(), md.getTextAfterTable());

			return Optional.of(new AnnotatedTablesReaderData(data.getId(), md, data.getTableData()));
		}
	}

}
