package de.uni_mannheim.informatik.dws.expansion.samples.postMatching;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import com.google.common.collect.BiMap;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueContainer;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.informatik.dws.expansion.wikidata.WikidataClasses;
import de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats.Stats;

public class OfficeHolderFixer implements KBValueCreator {

	@Override
	public KBValueContainer getMatchedValues(KnowledgeBase kb) {
		Set<ValueId> removeValues = new HashSet<>();
		ClassId classId = WikidataClasses.wikidataOfficeHolder;
		BiMap<PropertyId, Integer> properties = kb.getProperties(classId);

		Set<PropertyId> badProperties = new HashSet<>();
		badProperties.add(Stats.ebit);
		badProperties.add(Stats.netIncome);
		badProperties.add(Stats.totalRevenue);
		badProperties.add(Stats.totalAssets);
		badProperties.add(Stats.totalEquity);
		for (PropertyId propertyId : properties.keySet()) {
			if (badProperties.contains(propertyId)) {
				// System.out.println("==============================================");
				// System.out.println(propertyId.toCombinedString());
				Set<InstanceId> instances = kb.getInstances(classId, propertyId);
				for (InstanceId instance : instances) {
					removeValues.add(new ValueId(instance, propertyId));
				}
			}
		}

		return new KBValueContainer(new HashMap<>(), removeValues);
	}

}
