package de.uni_mannheim.informatik.dws.expansion.samples.postMatching;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.utils.DomainUtils;
import de.uni_mannheim.informatik.dws.expansion.wikidata.WikidataClasses;

public class SourcesFilter implements AnnotatedTableFilter {

	public ConcurrentHashMap<ClassId, AtomicInteger> count = new ConcurrentHashMap<>();

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {

		String source = data.getMetaData().getSource().getSource();
		ClassId classId = data.getMetaData().getDbPediaClass();
		String PLD = DomainUtils.getPayLevelDomainFromWholeURL(source);

		Map<ClassId, String> filter = new HashMap<>();
		filter.put(WikidataClasses.wikidataCountry, "worldbank.org");
		filter.put(WikidataClasses.wikidataCountry, "cia.gov");
		filter.put(WikidataClasses.wikidataBasketball, "basketball-reference.com");
		filter.put(WikidataClasses.wikidataFootball, "footballdb.com");
		filter.put(WikidataClasses.wikidataTradedCompany, "stockrow.com");

		if (filter.containsKey(classId) && filter.get(classId).equals(PLD)) {
			CHMHelper.getFromCHM(count, classId, () -> new AtomicInteger()).incrementAndGet();
			return Optional.empty();
		}

		return Optional.of(data);
	}

	public void printCount() {
		System.out.println("Class;NumFiltered");
		for (Entry<ClassId, AtomicInteger> entry : count.entrySet()) {
			System.out.println(entry.getKey() + ";" + entry.getValue().intValue());
		}

	}

}
