package de.uni_mannheim.informatik.dws.expansion.samples.postMatching;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.mannheim.uni.expansion.utils.CHMHelper;

public class TimeDependentPropertiesFilter implements AnnotatedTableFilter {
	private final OntologyRepository ontologyRepository;

	public TimeDependentPropertiesFilter(OntologyRepository ontologyRepository) {
		this.ontologyRepository = ontologyRepository;
	}

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {

		Map<Integer, List<PropertyId>> oldProperties = data.getMetaData().getAllProperties();
		Map<Integer, List<PropertyId>> newProperties = new HashMap<>();
		for (Entry<Integer, List<PropertyId>> e : oldProperties.entrySet()) {
			for (PropertyId propertyId : e.getValue()) {
				if (ontologyRepository.isTimeDependent(propertyId)) {
					CHMHelper.getFromMap(newProperties, e.getKey(), () -> new ArrayList<>()).add(propertyId);
				}
			}
		}

		if (newProperties.size() == 0) {
			return Optional.empty();
		} else {
			AnnotatedTableMetaData md = data.getMetaData();
			md = new AnnotatedTableMetaData(md.getSource(), md.getDbPediaClass(), md.getClassConf(), md.getNumHeaderRows(), md.getNumberOfRows(), newProperties,
					md.getAllPropertyScores(), md.getAllInstances(), md.getInstancesScores(), md.getKeyColumn(), md.getYearsBeforeText(),
					md.getYearsAfterText(), md.getYearsInContextBefore(), md.getYearsInTableTitle(), md.getYearsInPageTitle(), md.getPropertiesYears(),
					md.getPageTitle(), md.getTableTitle(), md.getTableContextTimeStampBeforeTable(), md.getTextBeforeTable(), md.getTextAfterTable());

			return Optional.of(new AnnotatedTablesReaderData(data.getId(), md, data.getTableData()));
		}

	}

}
