package de.uni_mannheim.informatik.dws.expansion.samples.setExtension;

import java.io.File;
import java.io.IOException;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;
import java.util.SortedMap;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import com.opencsv.CSVWriter;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId.InstanceIdStyle;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilterRunner;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig.TimeDependence;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategyRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.DecimalFormatter;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.informatik.dws.attribute.AttributeMatchingAdapter;
import de.uni_mannheim.informatik.dws.attribute.evaluation.AttributeMappingsEvaluator;
import de.uni_mannheim.informatik.dws.attribute.evaluation.AttributeMappingsPerformance;
import de.uni_mannheim.informatik.dws.attribute.matcher.AttributeMatcher;
import de.uni_mannheim.informatik.dws.attribute.model.AttributeMappings;
import de.uni_mannheim.informatik.dws.attribute.model.PropertyMapping;
import de.uni_mannheim.informatik.dws.attribute.model.TableAttributeMappings;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;
import de.uni_mannheim.informatik.dws.expansion.utils.remote.RepositoryLoading;
import expansion.set.attributeMatching.AttributeMatchingPipeline;
import expansion.set.attributeMatching.AttributeMatchingWeightLearner;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.evaluation.SetExpansionStrategiesConfig.SetExpansionStrategiesConfigBuilder;
import expansion.set.gscreation.GsOutputReader;
import expansion.set.optimizing.WeightsAdapter;
import expansion.set.strategies.rowclusterers.strategies.association.AssociationSimilarities;

public class ColumnReMatcherRun {
	public static void main(String[] args) throws IOException {

		Map<ClassId, Set<ClassId>> topClasses = ExtractTablesForGoldStandardRun.getClassesWithHierarchy();
		// topClasses.remove(ClassConstants.single);
		// topClasses.remove(ClassConstants.placeClass);
		// topClasses.add(ClassConstants.placeClass);
		// topClasses.add(ClassConstants.person);
		// topClasses.add(ClassConstants.populatedPlace);

		int currentRun = 0;
		if (args.length > 5) {
			currentRun = Integer.parseInt(args[5]);
		}

		boolean secondary = true;
		if (args.length > 6 && args[6].equals("false")) {
			secondary = false;
		}

		boolean alwaysDisableBlocking = true;
		if (args.length > 7) {
			alwaysDisableBlocking = Boolean.parseBoolean(args[7]);
		}

		SetExtension setExtension = new SetExtension(args, topClasses.keySet(), ExtractTablesForGoldStandardRun.getMinPropertyDensity(),
				RepositoryLoading.REMOTE_LOAD_REMOTELY, currentRun, alwaysDisableBlocking);

		String clusteringsWeightsFolderPrefix = setExtension.getMappingsCachesFolder() + "/models/clusteringModelsFull";

		String instanceMatchingWeightsFolderPrefix = setExtension.getMappingsCachesFolder() + "/models/instanceMatchingModelsFull";

		ColumnReMatcherRun run = setExtension.getGraph().get(ColumnReMatcherRun.class);

		boolean baseline = true;

		run.run(setExtension.getMappingsFolder(), setExtension.getNonTailProperties(), topClasses, clusteringsWeightsFolderPrefix,
				instanceMatchingWeightsFolderPrefix, setExtension.getMappingsCachesFolder(), currentRun, secondary, baseline, alwaysDisableBlocking);

		try {
			PersistentParallel4.close();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	@Inject
	protected WeightsAdapter weightsAdapter;

	// @Inject
	// protected AttributeThresholdLearner thresholdLearner;

	@Inject
	protected AttributeMappingsEvaluator evaluator;

	@Inject
	protected AttributeMatchingAdapter adapter;

	@Inject
	protected AttributeMatcher matcher;

	@Inject
	protected KnowledgeBase kb;

	@Inject
	protected AnnotatedTablesRepository repository;

	@Inject
	protected AnnotatedTableFilterRunner filterRunner;

	@Inject
	protected AttributeMatchingPipeline pipeline;

	@Inject
	protected AttributeMatchingWeightLearner weightLearner;

	public ColumnReMatcherRun() {

	}

	public void run(String mappingsFolder, Map<ClassId, Set<PropertyId>> nonTailProperties, Map<ClassId, Set<ClassId>> allParents,
			String clusteringsWeightsFolderPrefix, String instanceMatchingWeightsFolderPrefix, String cachesFolder, int totalRuns, boolean secondary,
			boolean baseline, boolean alwaysDisableBlocking) throws IOException {
		GsOutputReader reader = new GsOutputReader(InstanceIdStyle.DBPEDIA_URI);
		String gsOutputFolder = mappingsFolder + "/goldStandard/";

		String tableClassesFile = gsOutputFolder + "/classes_GS.csv";
		String processedTablesFile = gsOutputFolder + "/processedTables.csv";
		String forLearningTablesFile = gsOutputFolder + "/forLearningTables.csv";
		Set<AnnotatedTableId> processedTables = reader.readTableList(processedTablesFile);
		Set<AnnotatedTableId> forLearningTables = reader.readTableList(forLearningTablesFile);
		Map<AnnotatedTableId, ClassId> tableClasses = reader.readTableClasses(tableClassesFile);

		Set<AnnotatedTableId> allTableIds = repository.getAllTableIds();
		Map<ClassId, Set<AnnotatedTableId>> tableSet = new HashMap<>();
		for (Entry<AnnotatedTableId, ClassId> tableEntry : tableClasses.entrySet()) {
			if (allTableIds.contains(tableEntry.getKey())) {
				CHMHelper.getFromMap(tableSet, tableEntry.getValue(), () -> new HashSet<>()).add(tableEntry.getKey());
			}
		}
		AssociationSimilarities.tableSet = tableSet;

		AttributeMappings gsMappings = reader.getAttributeMappings(gsOutputFolder + "/attributes/", processedTables, tableClasses, nonTailProperties);
		AttributeMappings forLearningMappingsNonFinal = reader.getAttributeMappings(gsOutputFolder + "/attributes/", forLearningTables, tableClasses,
				nonTailProperties);
		AttributeMappings forLearningMappings;
		AttributeMappings forTestingMappings;

		{
			List<AnnotatedTableId> tableIdsList = new ArrayList<>(gsMappings.tableSet());
			Collections.sort(tableIdsList, (e1, e2) -> {
				return e1.getTableName().compareTo(e2.getTableName());
			});
			Collections.shuffle(tableIdsList, new Random("SHUFFLE".hashCode()));

			Pair<AttributeMappings, AttributeMappings> mappingsPair = splitTables(tableIdsList, gsMappings);

			forLearningMappings = mappingsPair.getFirst();
			forTestingMappings = mappingsPair.getSecond();

		}

		AttributeMappings currentMappings = adapter.getAttributeMappings(gsMappings.tableSet());

		FusionPerformance<AttributeMappingsPerformance> gsResult = evaluator.evaluate(gsMappings, gsMappings, kb.getNormalizationIndex());
		FusionPerformance<AttributeMappingsPerformance> t2kResult = evaluator.evaluate(gsMappings, currentMappings, kb.getNormalizationIndex());

		// AttributeMappings yaserMappings =
		// matcher.match(repository.getTableReaderData(gsMappings.tableSet()),
		// new ThresholdConstant(0.5), nonTailProperties);
		//
		// FusionPerformance<AttributeMappingsPerformance> yaserResult =
		// evaluator.evaluate(gsMappings, yaserMappings,
		// kb.getNormalizationIndex());
		//

		//
		// ThresholdContainer classThresholds = new
		// ThresholdContainerWithFallback(
		// thresholdLearner.findThresholdClassId(repository.getTableReaderData(gsMappings.tableSet()),
		// nonTailProperties, gsMappings, 4, (tableId) -> {
		// return gsMappings.getClassId(tableId);
		// }), new ThresholdConstant(0.25));
		// //
		// AttributeMappings yaserClassThresholdMappings =
		// matcher.match(repository.getTableReaderData(gsMappings.tableSet()),
		// classThresholds, nonTailProperties);
		// //
		// FusionPerformance<AttributeMappingsPerformance>
		// yaserClassThresholdResult = evaluator.evaluate(gsMappings,
		// yaserClassThresholdMappings,
		// kb.getNormalizationIndex());
		//
		// AnnotatedTablesRepository newTableRepository =
		// filterRunner.asRepository(repository.getTableReaderData(gsMappings.tableSet()),
		// new AttributeMappingsFilter(yaserThresholdMappings));
		// AttributeMatchingAdapter adapterNew = new
		// AttributeMatchingAdapter(newTableRepository);
		// AttributeMappings newMappings =
		// adapterNew.getAttributeMappings(gsMappings.tableSet());
		// FusionPerformance<AttributeMappingsPerformance> newMappingsResult =
		// evaluator.evaluate(gsMappings, newMappings,
		// kb.getNormalizationIndex());

		FusionConfig fusionConfig = SetExtension.getFusionConfig();

		SetExpansionStrategiesConfigBuilder builder = SetExtension.getSetExpansionStrategiesConfig().rebuilder();

		builder.clearFusionSetups().addFusionSetup(FusionStrategyRepository.NORMAL, PropertiesConfig.builder().setTimeDependence(TimeDependence.NO).build());
		SetExpansionStrategiesConfig setExpansionStrategiesConfig = builder.build();

		StrategiesConfig strategiesConfig = StrategiesConfig.defaultConfig();

		String matchingPipelineCachesFolder = cachesFolder + "matchingPipeline";
		if (secondary) {
			matchingPipelineCachesFolder += "Secondary";

		}
		MKDIR.createDirUnchecked(new File(matchingPipelineCachesFolder));

		String matchingPipelineWeightsCachesFolder = cachesFolder + "matchingPipelineWeightsTesting";
		if (baseline) {
			matchingPipelineWeightsCachesFolder += "Baseline";
		}
		MKDIR.createDirUnchecked(new File(matchingPipelineWeightsCachesFolder));

		SortedMap<Integer, AttributeMappings> pipelineMappings = pipeline.runPipeline(setExpansionStrategiesConfig, fusionConfig, strategiesConfig,
				gsMappings.tableSet(), allParents, nonTailProperties, 0, totalRuns, forLearningMappings, forTestingMappings, clusteringsWeightsFolderPrefix,
				instanceMatchingWeightsFolderPrefix, matchingPipelineCachesFolder, matchingPipelineWeightsCachesFolder, secondary, baseline,
				alwaysDisableBlocking);

		// FusionPerformance<AttributeMappingsPerformance> pipelinePerformance1
		// = evaluator.evaluate(gsMappings, pipelineMappings.get(1),
		// kb.getNormalizationIndex());
		// FusionPerformance<AttributeMappingsPerformance> pipelinePerformance1
		// = evaluator.evaluate(forTestingMappings, pipelineMappings.get(1),
		// kb.getNormalizationIndex());

		Set<AnnotatedTableId> fullSet = new HashSet<>(gsMappings.tableSet());

		fullSet.addAll(forLearningMappings.tableSet());

		// FusionPerformance<AttributeMappingsPerformance> yaserThresholdResult;
		// if (false) {
		//
		// ThresholdContainer thresholds = new
		// ThresholdContainerWithFallback(thresholdLearner
		// .findThreshold(repository.getTableReaderData(forLearningMappings.tableSet()),
		// nonTailProperties, forLearningMappings, 4, (tableId) -> {
		// return forLearningMappings.getClassId(tableId);
		// }), new ThresholdConstant(0.25));
		//
		// System.out.println(thresholds.toString());
		//
		// AttributeMappings yaserThresholdMappings =
		// matcher.match(repository.getTableReaderData(fullSet), thresholds,
		// nonTailProperties);
		//
		// yaserThresholdResult = evaluator.evaluate(gsMappings,
		// yaserThresholdMappings, kb.getNormalizationIndex());
		// } else if (false) {
		// OptimizableClass optimizableClazz = new
		// OptimizableClass("NON_PIPELINE_CP");
		// String nonPipelineWeightsFolder = cachesFolder +
		// "nonPipelineMatchingWeights";
		// MKDIR.createDirUnchecked(new File(nonPipelineWeightsFolder));
		//
		// Map<ClassId, WeightsRepositoryImpl> weights =
		// weightsAdapter.readWeights(allParents.keySet(),
		// nonPipelineWeightsFolder);
		//
		// boolean containsWeights = true;
		// for (ClassId classId : allParents.keySet()) {
		// if (!weights.get(classId).containsWeightsForClass(optimizableClazz))
		// {
		// containsWeights = false;
		// break;
		// }
		// }
		//
		// AttributeMappings mappings;
		// if (containsWeights) {
		//
		// mappings = matcher.match(repository.getTableReaderData(fullSet),
		// weights, nonTailProperties, optimizableClazz, null,
		// gsMappings.tableSet());
		// yaserThresholdResult = evaluator.evaluate(gsMappings, mappings,
		// kb.getNormalizationIndex());
		//
		// } else {
		//
		// AttributeMatchingFullRuns fullRuns =
		// matcher.getFullMatches(repository.getTableReaderData(fullSet),
		// nonTailProperties, null);
		//
		// Map<ClassId, WeightsRepositoryImpl> newWeights =
		// weightLearner.learn(optimizableClazz, fullRuns.getByClass(),
		// forLearningMappings,
		// kb.getNormalizationIndex());
		//
		// for (Entry<ClassId, WeightsRepositoryImpl> entry :
		// newWeights.entrySet()) {
		// weights.put(entry.getKey(),
		// weights.get(entry.getKey()).combineWith(entry.getValue()));
		//
		// }
		// mappings = weightLearner.applyWeights(optimizableClazz,
		// fullRuns.getByClass(), new HashMap<>(weights),
		// forLearningMappings.tableSet());
		//
		// weightsAdapter.storeWeights(nonPipelineWeightsFolder, weights);
		// yaserThresholdResult = evaluator.evaluate(gsMappings, mappings,
		// kb.getNormalizationIndex());
		//
		// }
		//
		// }

		// print("GS", gsResult);
		// print("T2K", t2kResult);
		// print("YASER", yaserResult);
		// print("Y_CPTH", yaserThresholdResult);
		// print("Y_CLASS", yaserClassThresholdResult);

		for (int i = 0; i <= totalRuns; i++) {
			FusionPerformance<AttributeMappingsPerformance> pipelinePerformance = evaluator.evaluate(forTestingMappings, pipelineMappings.get(i),
					kb.getNormalizationIndex());

			print("PIPELINE_" + i, pipelinePerformance);
		}

		// print("TEST", newMappingsResult);

		int sdf = 123;
	}

	private Pair<AttributeMappings, AttributeMappings> splitTables(List<AnnotatedTableId> tableIdsListInput, AttributeMappings gsMappings) {
		Map<AnnotatedTableId, ClassId> learningClassMap = new HashMap<>();
		Map<AnnotatedTableId, TableAttributeMappings> learningMappings = new HashMap<>();
		Map<AnnotatedTableId, ClassId> testingClassMap = new HashMap<>();
		Map<AnnotatedTableId, TableAttributeMappings> testingMappings = new HashMap<>();

		Map<ClassId, List<AnnotatedTableId>> byClass = new HashMap<>();
		for (AnnotatedTableId tableId : tableIdsListInput) {
			CHMHelper.getFromMap(byClass, gsMappings.getClassId(tableId), () -> new ArrayList<>()).add(tableId);

		}

		for (Entry<ClassId, List<AnnotatedTableId>> entry : byClass.entrySet()) {
			List<AnnotatedTableId> tableIdsList = entry.getValue();

			Map<PropertyId, MutableInt> count = new HashMap<>();
			Map<AnnotatedTableId, Set<PropertyId>> mapping = new HashMap<>();

			int forLearning = 0;
			for (AnnotatedTableId tableId : tableIdsList) {
				Set<PropertyId> set = new HashSet<>();
				for (Entry<ColumnIdx, PropertyMapping> property : gsMappings.getMappings().get(tableId).getMappings().entrySet()) {
					CHMHelper.getFromMap(count, property.getValue().getPropertyId(), () -> new MutableInt()).increment();
					set.add(property.getValue().getPropertyId());
				}
				mapping.put(tableId, set);
			}

			Map<PropertyId, Integer> remaining = count.entrySet().stream()
					.collect(Collectors.toMap(Entry::getKey, (e) -> Math.round(0.66F * e.getValue().intValue())));
			Map<PropertyId, Integer> skipped = new HashMap<>();
			Map<PropertyId, Integer> total = count.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().intValue()));
			List<PropertyId> propertyIds = count.entrySet().stream().sorted((e1, e2) -> Integer.compare(e1.getValue().intValue(), e2.getValue().intValue()))
					.map((e) -> e.getKey()).collect(Collectors.toList());
			double size = propertyIds.size();
			for (AnnotatedTableId tableId : tableIdsList) {
				double sum = 0.0;
				for (ListEntry<PropertyId> propertyId : ListIterable.iterable(propertyIds)) {
					if (mapping.get(tableId).contains(propertyId.getValue())) {
						double weight = (size - propertyId.getIndex()) / size;
						if (remaining.get(propertyId.getValue()) > 0) {
							sum += weight;
						} else {
							sum -= weight;
						}
					}

				}

				if (sum > 0.0) {
					learningClassMap.put(tableId, gsMappings.getClassId(tableId));
					learningMappings.put(tableId, gsMappings.getMappings().get(tableId));
					for (PropertyId propertyId : mapping.get(tableId)) {
						int r = remaining.get(propertyId);
						r--;
						remaining.put(propertyId, r);
					}
				} else {
					testingClassMap.put(tableId, gsMappings.getClassId(tableId));
					testingMappings.put(tableId, gsMappings.getMappings().get(tableId));
				}

			}
		}
		Map<ClassId, Map<PropertyId, Pair<MutableInt, MutableInt>>> statisticsMap = new HashMap<>();

		for (Entry<AnnotatedTableId, TableAttributeMappings> entry : learningMappings.entrySet()) {
			Map<PropertyId, Pair<MutableInt, MutableInt>> innerMap = CHMHelper.getFromMap(statisticsMap, gsMappings.getClassId(entry.getKey()),
					() -> new HashMap<>());
			for (Entry<ColumnIdx, PropertyMapping> mappings : entry.getValue().getMappings().entrySet()) {
				CHMHelper.getFromMap(innerMap, mappings.getValue().getPropertyId(), () -> new Pair<>(new MutableInt(), new MutableInt())).getFirst()
						.increment();
			}
		}

		for (Entry<AnnotatedTableId, TableAttributeMappings> entry : testingMappings.entrySet()) {
			Map<PropertyId, Pair<MutableInt, MutableInt>> innerMap = CHMHelper.getFromMap(statisticsMap, gsMappings.getClassId(entry.getKey()),
					() -> new HashMap<>());
			for (Entry<ColumnIdx, PropertyMapping> mappings : entry.getValue().getMappings().entrySet()) {
				CHMHelper.getFromMap(innerMap, mappings.getValue().getPropertyId(), () -> new Pair<>(new MutableInt(), new MutableInt())).getSecond()
						.increment();
			}
		}

		StringWriter sw = new StringWriter();
		CSVWriter csv = new CSVWriter(sw);

		{
			int i = 0;
			String[] line = new String[5];
			line[i++] = "Class";
			line[i++] = "Property";
			line[i++] = "Total";
			line[i++] = "Learning";
			line[i++] = "Testing";
			csv.writeNext(line);
		}

		for (Entry<ClassId, Map<PropertyId, Pair<MutableInt, MutableInt>>> entry : statisticsMap.entrySet()) {
			for (Entry<PropertyId, Pair<MutableInt, MutableInt>> inner : entry.getValue().entrySet()) {
				int i = 0;
				String[] line = new String[5];
				line[i++] = entry.getKey().getFinalPart();
				line[i++] = inner.getKey().getFinalPart();
				line[i++] = String.valueOf(inner.getValue().getFirst().intValue() + inner.getValue().getSecond().intValue());
				line[i++] = String.valueOf(inner.getValue().getFirst().intValue());
				line[i++] = String.valueOf(inner.getValue().getSecond().intValue());
				csv.writeNext(line);
			}
		}

		try {
			csv.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		System.out.println(sw.toString());
		return new Pair<>(new AttributeMappings(learningMappings, learningClassMap), new AttributeMappings(testingMappings, testingClassMap));
	}

	private Pair<AttributeMappings, AttributeMappings> splitTablesOld(List<AnnotatedTableId> tableIdsList, AttributeMappings gsMappings) {
		Map<AnnotatedTableId, ClassId> learningClassMap = new HashMap<>();
		Map<AnnotatedTableId, TableAttributeMappings> learningMappings = new HashMap<>();
		Map<AnnotatedTableId, ClassId> testingClassMap = new HashMap<>();
		Map<AnnotatedTableId, TableAttributeMappings> testingMappings = new HashMap<>();

		int forLearning = 0;
		for (AnnotatedTableId tableId : tableIdsList) {
			if (forLearning < 2) {
				learningClassMap.put(tableId, gsMappings.getClassId(tableId));
				learningMappings.put(tableId, gsMappings.getMappings().get(tableId));
				forLearning++;
			} else {

				testingClassMap.put(tableId, gsMappings.getClassId(tableId));
				testingMappings.put(tableId, gsMappings.getMappings().get(tableId));
				forLearning = 0;
			}
		}

		return new Pair<>(new AttributeMappings(learningMappings, learningClassMap), new AttributeMappings(testingMappings, testingClassMap));
	}

	public static void print(String prefix, FusionPerformance<AttributeMappingsPerformance> result) {
		System.out.println();
		System.out.println("Run,Correct,Wrong,MaxCorrect,Precision,Recall,F-Measure");
		printLine(prefix, result.getTotal());

		System.out.println("Run,Class,Correct,Wrong,MaxCorrect,Precision,Recall,F-Measure");
		for (Entry<ClassId, AttributeMappingsPerformance> entry : result.getPerClassId().entrySet()) {
			printLine(prefix + "," + entry.getKey().getClassId(), entry.getValue());

		}

		System.out.println("Run,Class,Property,Correct,Wrong,MaxCorrect,Precision,Recall,F-Measure");
		for (Entry<ClassPropertyId, AttributeMappingsPerformance> entry : result.getPerProperty().entrySet()) {
			printLine(prefix + "," + entry.getKey().getClassId() + "," + entry.getKey().getPropertyId().getFinalPart2(), entry.getValue());

		}
		System.out.println();

	}

	public static void printLine(String prefix, AttributeMappingsPerformance performance) {
		System.out.println(prefix + "," + performance.getCorrect() + "," + performance.getWrong() + "," + performance.getMaxCorrect() + ","
				+ DecimalFormatter.format(performance.getPrecision()) + "," + DecimalFormatter.format(performance.getRecall()) + ","
				+ DecimalFormatter.format(performance.getfMeasure()));

	}
}
