package de.uni_mannheim.informatik.dws.expansion.samples.setExtension;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.BiConsumer;
import java.util.stream.Collectors;

import com.google.common.collect.BiMap;

import au.com.bytecode.opencsv.CSVReader;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId.ClassIdStyle;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ImmuMaker;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueContainer;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedAbstractObjectHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class ExtendedKnowledgeBaseValueCreator {

	private final String mappingsFolder;

	Map<ClassId, Map<InstanceId, Set<String>>> newEntitiesWithClassAndLabel;
	Map<InstanceId, Set<String>> collegeFootballWithLabels;
	Map<InstanceId, Set<String>> positionFootballWithLabels;
	Map<String, InstanceId> positionIndexNormalized;

	public ExtendedKnowledgeBaseValueCreator(String mappingsFolder) {
		super();
		this.mappingsFolder = mappingsFolder;

		String positionFile = mappingsFolder + "/football_position.csv";
		String collegeFile = mappingsFolder + "/college_football.csv";
		String newEntititiesFile = mappingsFolder + "/new_entities.csv";

		newEntitiesWithClassAndLabel = getNewEntities(newEntititiesFile);
		collegeFootballWithLabels = getCollegeFootballs(collegeFile);
		positionFootballWithLabels = getPositionFootballs(positionFile);

		Map<String, InstanceId> positionIndexNormalized = new HashMap<>();
		for (Entry<InstanceId, Set<String>> entry : positionFootballWithLabels.entrySet()) {
			positionIndexNormalized.put(entry.getKey().getInstanceId().substring("http://dbpedia.org/resource/".length()).toLowerCase(), entry.getKey());
			for (String label : entry.getValue()) {
				positionIndexNormalized.put(label.toLowerCase().trim(), entry.getKey());
			}
		}

		this.positionIndexNormalized = ImmuMaker.map(positionIndexNormalized);
	}

	public Map<String, InstanceId> getPositionIndexNormalized() {
		return positionIndexNormalized;
	}

	public void setPositionIndexNormalized(Map<String, InstanceId> positionIndexNormalized) {
		this.positionIndexNormalized = positionIndexNormalized;
	}

	public KBValueCreator getCreator() {

		KBValueCreator kbValueCreator = new KBValueCreator() {

			@Override
			public KBValueContainer getMatchedValues(KnowledgeBase kb) {

				Set<Pair<InstanceId, ClassId>> newSetMappings = Collections.newSetFromMap(new ConcurrentHashMap<>());

				Map<ValueId, QualifiedValueList> replaceValues = new ConcurrentHashMap<>();

				BiMap<InstanceId, Integer> songInstances = kb.getInstances(ClassConstants.song);
				for (InstanceId songInstanceId : songInstances.keySet()) {
					newSetMappings.add(new Pair<>(songInstanceId, ClassConstants.single));
				}

				BiConsumer<Map<InstanceId, Set<String>>, ClassId> addInstances = (instances, instancesClass) -> {
					for (Entry<InstanceId, Set<String>> position : instances.entrySet()) {
						newSetMappings.add(new Pair<>(position.getKey(), instancesClass));
						ValueId labelValueId = new ValueId(position.getKey(), PropertiesExpansionData.LABEL_PROPERTY);

						Set<String> labels = new HashSet<>(position.getValue());
						QualifiedValueList qvl = kb.getQualifiedValueList(labelValueId);
						if (qvl != null) {
							for (QualifiedValueHolder claim : qvl.getClaims()) {
								if (claim.getValue().getType() == ValueType.STRING) {
									labels.add(((StringObject) claim.getValue()).getString());
								}
							}

						}
						List<QualifiedValueHolder> claims = labels.stream().map((e) -> new QualifiedAbstractObjectHolder(new StringObject(e)))
								.collect(Collectors.toList());
						qvl = new QualifiedValueList(claims);
						replaceValues.put(labelValueId, qvl);
					}
				};

				addInstances.accept(positionFootballWithLabels, ClassConstants.gridironFootballPosition);
				addInstances.accept(collegeFootballWithLabels, ClassConstants.collegeFootballTeam);

				for (Entry<ClassId, Map<InstanceId, Set<String>>> newEntitiesEntry : newEntitiesWithClassAndLabel.entrySet()) {

					addInstances.accept(newEntitiesEntry.getValue(), newEntitiesEntry.getKey());
				}

				return KBValueContainer.builder().setReplaceValues(replaceValues).setNewSetMappings(newSetMappings).build();

			}

		};

		return kbValueCreator;
	}

	private static Map<InstanceId, Set<String>> getPositionFootballs(String positionFile) {
		Map<InstanceId, Set<String>> ret = new HashMap<>();
		try {

			CSVReader reader = new CSVReader(new FileReader(new File(positionFile)), ',', '"');

			for (String[] entry : reader.readAll()) {
				Set<String> labels = new HashSet<>();
				labels.add(entry[1]);
				labels.add(entry[2]);
				InstanceId instanceId = InstanceId.fromString(entry[0]);
				CHMHelper.getFromMap(ret, instanceId, () -> new HashSet<>()).addAll(labels);
			}

			reader.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return ret;
	}

	private static Map<InstanceId, Set<String>> getCollegeFootballs(String collegeFile) {
		Map<InstanceId, Set<String>> ret = new HashMap<>();
		try {

			CSVReader reader = new CSVReader(new FileReader(new File(collegeFile)), ',', '"');

			for (String[] entry : reader.readAll()) {
				Set<String> labels = new HashSet<>();
				labels.add(entry[1]);
				labels.add(entry[2]);
				labels.add(entry[3]);
				if (entry[3].toLowerCase().endsWith("state")) {
					String withoutState = entry[3].substring(0, entry[3].length() - "state".length());
					labels.add(withoutState + "st");
					labels.add(withoutState + "st.");
					int sdf = 123;
				}
				InstanceId instanceId = InstanceId.fromString(entry[0]);
				CHMHelper.getFromMap(ret, instanceId, () -> new HashSet<>()).addAll(labels);
			}

			reader.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return ret;
	}

	private static Map<ClassId, Map<InstanceId, Set<String>>> getNewEntities(String newEntititiesFile) {
		Map<ClassId, Map<InstanceId, Set<String>>> ret = new HashMap<>();
		try {

			CSVReader reader = new CSVReader(new FileReader(new File(newEntititiesFile)), ',', '"');

			for (String[] entry : reader.readAll()) {
				ClassId classId = ClassId.fromString(entry[2], ClassIdStyle.DBPEDIA_URI);
				Set<String> labels = new HashSet<>();
				labels.add(entry[1]);
				Map<InstanceId, Set<String>> classMap = CHMHelper.getFromMap(ret, classId, () -> new HashMap<>());
				InstanceId instanceId = InstanceId.fromString(entry[0]);
				CHMHelper.getFromMap(classMap, instanceId, () -> new HashSet<>()).addAll(labels);

			}

			reader.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return ret;
	}
}
