package de.uni_mannheim.informatik.dws.expansion.samples.setExtension;

import java.io.File;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.commons.lang.RandomStringUtils;

import com.google.common.collect.ImmutableSet;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId.InstanceIdStyle;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.informatik.dws.expansion.utils.remote.RepositoryLoading;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.gscreation.ExtractTablesForGoldStandard3;
import expansion.set.gscreation.GsOutputReader;

public class ExtractHomonymTablesForGoldStandardRun {

	public static Map<ClassId, Set<ClassId>> getClassesWithHierarchy() {
		Map<ClassId, Set<ClassId>> topClasses = new HashMap<>();
		topClasses.put(ClassConstants.single, ImmutableSet.of(ClassConstants.musicalWorkClass, ClassConstants.song));
		topClasses.put(ClassConstants.gridIronFootballPlayer, ImmutableSet.of(ClassConstants.person, ClassConstants.athlete));
		topClasses.put(ClassConstants.placeClass, ImmutableSet.of(ClassConstants.populatedPlace));
		// topClasses.add(ClassConstants.animal);

		return topClasses;
	}

	public static double getMinPropertyDensity() {
		return 0.3;
	}

	public static void main(String[] args) {
		try {
			Map<ClassId, Set<ClassId>> topClasses = getClassesWithHierarchy();
			topClasses.remove(ClassConstants.single);
			// topClasses.remove(ClassConstants.placeClass);

			SetExtension setExtension = new SetExtension(args, topClasses.keySet(), getMinPropertyDensity(), RepositoryLoading.REMOTE_LOAD_REMOTELY);

			String mappingsFolder = setExtension.getMappingsFolder();

			GsOutputReader reader = new GsOutputReader(InstanceIdStyle.DBPEDIA_URI);
			String gsOutputFolder = mappingsFolder + "/gsOutput/";

			String tableClassesFile = gsOutputFolder + "/full_table_set.csv";
			Map<AnnotatedTableId, ClassId> tableClasses = reader.readTableClasses(tableClassesFile);

			Map<ClassId, Set<AnnotatedTableId>> tableClassesReverse = tableClasses.entrySet().stream()
					.collect(Collectors.groupingBy(Entry::getValue, Collectors.mapping(Map.Entry::getKey, Collectors.toSet())));

			ExtractTablesForGoldStandard3 runner = setExtension.getGraph().get(ExtractTablesForGoldStandard3.class);

			FusionConfig fusionConfig = setExtension.getFusionConfig();

			SetExpansionStrategiesConfig setExpansionStrategiesConfig = setExtension.getSetExpansionStrategiesConfig();
			String thresholdsFolder = setExtension.getThresholdsFolder();
			StrategiesConfig strategiesConfig = StrategiesConfig.defaultConfig();

			String clusterCacheFolder = setExtension.getMappingsCachesFolder();
			MKDIR.createDirUnchecked(new File(clusterCacheFolder));

			String gsFolder = setExtension.getTablesFolder() + "/gsExtra/";
			MKDIR.createDirUnchecked(new File(gsFolder));
			String currentGS = new SimpleDateFormat("yyyyMMddHHmmss").format(new Date()) + "-" + RandomStringUtils.randomAlphanumeric(4);
			String currentGSfolder = gsFolder + "/" + currentGS;
			MKDIR.createDirUnchecked(new File(currentGSfolder));

			runner.run(topClasses, tableClassesReverse, clusterCacheFolder, currentGSfolder + "/", fusionConfig, setExpansionStrategiesConfig, thresholdsFolder,
					strategiesConfig);

			PersistentParallel6.close();

		} catch (IOException | InterruptedException e) {
			e.printStackTrace();
		}
	}
}
