package de.uni_mannheim.informatik.dws.expansion.samples.setExtension;

import java.io.File;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.commons.lang.RandomStringUtils;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId.InstanceIdStyle;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.informatik.dws.expansion.utils.remote.RepositoryLoading;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.gscreation.ExtractTablesForGoldStandard4;
import expansion.set.gscreation.GsOutputReader;

public class ExtractNewTablesForGoldStandard {

	public static void main(String[] args) {
		try {
			Map<ClassId, Set<ClassId>> topClasses = ExtractHomonymTablesForGoldStandardRun.getClassesWithHierarchy();
			topClasses.remove(ClassConstants.single);
			topClasses.remove(ClassConstants.gridIronFootballPlayer);
			// topClasses.remove(ClassConstants.placeClass);

			SetExtension setExtension = new SetExtension(args, topClasses.keySet(), ExtractHomonymTablesForGoldStandardRun.getMinPropertyDensity(),
					RepositoryLoading.REMOTE_LOAD_REMOTELY);

			String mappingsFolder = setExtension.getMappingsFolder();

			GsOutputReader reader = new GsOutputReader(InstanceIdStyle.DBPEDIA_URI);
			String gsOutputFolder = mappingsFolder + "/gsOutput/";

			String tableClassesFile = gsOutputFolder + "/classes_GS.csv";
			Map<AnnotatedTableId, ClassId> tableClasses = reader.readTableClasses(tableClassesFile);

			Map<ClassId, Set<AnnotatedTableId>> tableClassesReverse = tableClasses.entrySet().stream()
					.collect(Collectors.groupingBy(Entry::getValue, Collectors.mapping(Map.Entry::getKey, Collectors.toSet())));

			ExtractTablesForGoldStandard4 runner = setExtension.getGraph().get(ExtractTablesForGoldStandard4.class);

			FusionConfig fusionConfig = setExtension.getFusionConfig();

			SetExpansionStrategiesConfig setExpansionStrategiesConfig = setExtension.getSetExpansionStrategiesConfig();
			String thresholdsFolder = setExtension.getThresholdsFolder();
			StrategiesConfig strategiesConfig = StrategiesConfig.defaultConfig();

			String clusterCacheFolder = setExtension.getMappingsCachesFolder();
			MKDIR.createDirUnchecked(new File(clusterCacheFolder));

			String gsFolder = setExtension.getTablesFolder() + "/gsNew/";
			MKDIR.createDirUnchecked(new File(gsFolder));
			String currentGS = new SimpleDateFormat("yyyyMMddHHmmss").format(new Date()) + "-" + RandomStringUtils.randomAlphanumeric(4);
			String currentGSfolder = gsFolder + "/" + currentGS;
			MKDIR.createDirUnchecked(new File(currentGSfolder));

			String newCacheFolder = clusterCacheFolder + "/newTablesCache/";
			MKDIR.createDirUnchecked(new File(newCacheFolder));
			runner.run(topClasses, tableClassesReverse, newCacheFolder, currentGSfolder + "/", fusionConfig, setExpansionStrategiesConfig, thresholdsFolder,
					strategiesConfig);

			PersistentParallel6.close();

		} catch (IOException | InterruptedException e) {
			e.printStackTrace();
		}
	}
}
