package de.uni_mannheim.informatik.dws.expansion.samples.setExtension;

import java.io.File;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang.RandomStringUtils;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.informatik.dws.expansion.utils.remote.RepositoryLoading;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.gscreation.ExtractTablesForGoldStandard2;

public class ExtractTablesForGoldStandardRun {

	public static Map<ClassId, Set<ClassId>> getClassesWithHierarchy() {
		Map<ClassId, Set<ClassId>> topClasses = new HashMap<>();
		topClasses.put(ClassConstants.single, ImmutableSet.of(ClassConstants.work, ClassConstants.musicalWorkClass));
		topClasses.put(ClassConstants.gridIronFootballPlayer, ImmutableSet.of(ClassConstants.person, ClassConstants.athlete));
		topClasses.put(ClassConstants.placeClass, ImmutableSet.of(ClassConstants.place, ClassConstants.populatedPlace));
		// topClasses.add(ClassConstants.animal);

		return topClasses;
	}

	public static Map<ClassId, Set<ClassId>> getClassesWithHierarchyFilteredOrNot() {
		Map<ClassId, Set<ClassId>> topClasses = getClassesWithHierarchy();

		InetAddress ip;
		String hostname;
		try {
			ip = InetAddress.getLocalHost();
			hostname = ip.getHostName();

			switch (hostname) {

			case "wifo5-13":

				break;
			case "dws-06":
				topClasses.remove(ClassConstants.single);
				topClasses.remove(ClassConstants.placeClass);
				break;

			}

		} catch (UnknownHostException e) {

			e.printStackTrace();
		}

		return topClasses;
	}

	public static double getMinPropertyDensity() {
		return 0.3;
	}

	public static void main(String[] args) {
		try {
			Map<ClassId, Set<ClassId>> topClasses = getClassesWithHierarchy();
			// topClasses.remove(ClassConstants.single);
			// topClasses.remove(ClassConstants.gridIronFootballPlayer);

			SetExtension setExtension = new SetExtension(args, topClasses.keySet(), getMinPropertyDensity(), RepositoryLoading.REMOTE_LOAD_REMOTELY);

			ExtractTablesForGoldStandard2 runner = setExtension.getGraph().get(ExtractTablesForGoldStandard2.class);

			FusionConfig fusionConfig = setExtension.getFusionConfig();
			// SetExpansionStrategy setExpansionStrategy =
			// SetExpansionStrategiesRepository.LABEL_BLOCKED_SIMILARITY;
			SetExpansionStrategiesConfig setExpansionStrategiesConfig = setExtension.getSetExpansionStrategiesConfig();
			String thresholdsFolder = setExtension.getThresholdsFolder();
			StrategiesConfig strategiesConfig = StrategiesConfig.defaultConfig();

			Map<ClassId, Integer> maxIndegree = ImmutableMap.<ClassId, Integer> builder().put(ClassConstants.single, 2181).build();
			String clusterCacheFolder = setExtension.getMappingsCachesFolder();
			MKDIR.createDirUnchecked(new File(clusterCacheFolder));

			String gsFolder = setExtension.getTablesFolder() + "/gs/";
			MKDIR.createDirUnchecked(new File(gsFolder));
			String currentGS = new SimpleDateFormat("yyyyMMddHHmmss").format(new Date()) + "-" + RandomStringUtils.randomAlphanumeric(4);
			String currentGSfolder = gsFolder + "/" + currentGS;
			MKDIR.createDirUnchecked(new File(currentGSfolder));

			runner.run(topClasses, maxIndegree, clusterCacheFolder, currentGSfolder + "/", fusionConfig, setExpansionStrategiesConfig, thresholdsFolder,
					strategiesConfig);

			PersistentParallel6.close();

		} catch (IOException | InterruptedException e) {
			e.printStackTrace();
		}
	}
}
