package de.uni_mannheim.informatik.dws.expansion.samples.setExtension;

import java.io.File;
import java.io.IOException;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import javax.inject.Inject;
import javax.inject.Named;

import org.apache.commons.lang.mutable.MutableInt;

import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdsSet;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig.TimeDependence;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.tablefusion.learning.RowValuesGetter;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;
import de.mannheim.uni.expansion.tablefusion.learning.SplitRowValues;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategyRepository;
import de.mannheim.uni.expansion.tablefusion.services.grouping.GroupedValuesRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.DecimalFormatter;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.utils.kryo.CHMKryoAdapter;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdContainer;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumInteger;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentAccumulator;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.LearnedModelRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.remote.RepositoryLoading;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;
import expansion.set.SetExpansionProperties;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusterer;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.clustering.model.similarities.RowClusterConstantSimilarities;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.creation.Entity;
import expansion.set.creation.EntityCreator;
import expansion.set.entitymatching.EntityMatcherWeightedResults;
import expansion.set.entitymatching.EntityMatchingRunner;
import expansion.set.entitymatching.candidateselection.CandidateSelectionConfig;
import expansion.set.entitymatching.candidateselection.strategies.CandidateSelection_Label;
import expansion.set.entitymatching.matchers.EntityMatcherConfig;
import expansion.set.entitymatching.modelLearning.EM_LearnedModelRepository;
import expansion.set.entitymatching.run.EntityMatchingRunRepository;
import expansion.set.evaluation.CaseCreator;
import expansion.set.evaluation.SetExpansionRunner;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.evaluation.SetExpansionStrategiesConfig.SetExpansionStrategiesConfigBuilder;
import expansion.set.evaluation.matching.ScoredEntity;
import expansion.set.exportimport.ClusterPairsExporter;
import expansion.set.exportimport.OptionalClusterCache;
import expansion.set.implicitAttributes.ImplicitAttributeFinder;
import expansion.set.optimizing.LearnedModelAdapter;
import expansion.set.optimizing.WeightsAdapter;
import expansion.set.strategies.SetExpansionStrategiesRepository;
import expansion.set.strategies.SetExpansionStrategy;
import expansion.set.strategies.rowclusterers.GreedyClusteringParallelSplitter;
import expansion.set.strategies.rowclusterers.RowClustererRepository;
import expansion.set.strategies.rowclusterers.label.LabelRowSimilaritiesCreator;
import expansion.set.strategies.rowclusterers.strategies.association.AssociationSimilarities;
import expansion.set.tablerow.AnnotatedQualifiedTableRow;
import expansion.set.tablerow.TableRowCreator;

public class FullRun {

	public static void main(String[] args) throws IOException {

		Map<ClassId, Set<ClassId>> topClasses = ExtractTablesForGoldStandardRun.getClassesWithHierarchy();

		int setId = -1;
		int currentRun = 0;
		if (args.length > 5) {
			currentRun = Integer.parseInt(args[5]);
		}

		boolean alwaysDisableBlocking = true;
		if (args.length > 6) {
			alwaysDisableBlocking = Boolean.parseBoolean(args[6]);
		}

		if (args.length > 7) {
			int classToRun;
			if (args[7].contains("-")) {
				String[] split = args[7].split("-");
				classToRun = Integer.parseInt(split[0]);
				setId = Integer.parseInt(split[1]);

			} else {
				classToRun = Integer.parseInt(args[7]);
			}

			switch (classToRun) {
			case 0:
				topClasses.remove(ClassConstants.single);
				topClasses.remove(ClassConstants.placeClass);
				break;
			case 1:
				topClasses.remove(ClassConstants.gridIronFootballPlayer);
				topClasses.remove(ClassConstants.placeClass);
				break;
			case 2:
				topClasses.remove(ClassConstants.gridIronFootballPlayer);
				topClasses.remove(ClassConstants.single);
				break;
			}
		}

		SetExtension setExtension = new SetExtension(args, topClasses.keySet(), ExtractTablesForGoldStandardRun.getMinPropertyDensity(),
				RepositoryLoading.REMOTE_LOAD_REMOTELY, false, true, currentRun, alwaysDisableBlocking);

		FusionConfig fusionConfig = SetExtension.getFusionConfig();

		SetExpansionStrategiesConfigBuilder builder = SetExtension.getSetExpansionStrategiesConfig().rebuilder();

		builder.clearFusionSetups().addFusionSetup(FusionStrategyRepository.NORMAL, PropertiesConfig.builder().setTimeDependence(TimeDependence.NO).build());
		SetExpansionStrategiesConfig setExpansionStrategiesConfig = builder.build();

		StrategiesConfig strategiesConfig = StrategiesConfig.defaultConfig();

		FullRun run = setExtension.getGraph().get(FullRun.class);

		String clusteringWeightsFolder = setExtension.getMappingsCachesFolder() + "/models/clusteringModelsFull" + currentRun + "/";

		MKDIR.createDirUnchecked(new File(clusteringWeightsFolder));

		String instanceMatchingWeightsFolder = setExtension.getMappingsCachesFolder() + "/models/instanceMatchingModelsFull" + currentRun + "/";
		MKDIR.createDirUnchecked(new File(instanceMatchingWeightsFolder));

		String currentCaseFolder = setExtension.getCurrentManualCaseFolder();

		run.run(topClasses, fusionConfig, setExpansionStrategiesConfig, strategiesConfig, setExtension.getNonTailProperties(), clusteringWeightsFolder,
				instanceMatchingWeightsFolder, currentCaseFolder, alwaysDisableBlocking, setId);
	}

	@Inject
	@Named(SetExpansionProperties.tablesCachesFolder)
	protected String cacheFolder;

	@Inject
	protected AnnotatedTablesRepository tablesRepository;

	@Inject
	protected GroupedValuesRepository groupedValuesRepository;

	@Inject
	protected SetExpansionRunner setExpansionRunner;

	@Inject
	protected AnnotatedTablesRepository repository;

	@Inject
	protected RowValuesGetter rowValuesGetter;

	@Inject
	protected GreedyClusteringParallelSplitter greedyClusteringParallelSplitter;

	@Inject
	protected EntityMatchingRunner entityMatchingRunner;

	@Inject
	protected RowClustererRepository rowClustererRepository;

	@Inject
	protected EntityCreator entityCreator;

	@Inject
	protected WeightsAdapter weightsAdapter;

	@Inject
	protected LearnedModelAdapter LearnedModelAdapter;

	@Inject
	protected CaseCreator caseCreator;

	@Inject
	protected CHMKryoAdapter chmKryoAdapter;

	@Inject
	protected ClusterPairsExporter ClusterPairsExporter;

	@Inject
	protected ImplicitAttributeFinder implicitAttributeFinder;

	@Inject
	protected CandidateSelection_Label csLabel;

	@Inject
	protected TableRowCreator tableRowCreator;

	@Inject
	protected ValueEvaluator valueEvaluator;

	public void run(Map<ClassId, Set<ClassId>> allParents, FusionConfig fusionConfig, SetExpansionStrategiesConfig setExpansionStrategiesConfig,
			StrategiesConfig strategiesConfig, Map<ClassId, Set<PropertyId>> nonTailPropertiesMap, String clusteringWeightsFolder,
			String instanceMatchingWeightsFolder, String currentCaseFolder, boolean alwaysDisableBlocking, int setId) throws IOException {
		if (false) {

			CandidateSelectionConfig config = new CandidateSelectionConfig(allParents);
			ClassId classId = null;
			for (ClassId classIdIt : allParents.keySet()) {
				classId = classIdIt;
			}

			Set<TableRowId> rowIds = new HashSet<>();

			rowIds.add(new TableRowId("1346981172239_1346996367380_1095.arc4045566334203770707#18339854_1_477012871080658822", 2));
			rowIds.add(new TableRowId("1346981172258_1346987054498_3667.arc1293561988245259067#83551231_0_2024899370140227261", 8));
			rowIds.add(new TableRowId("1346981172258_1346986187274_1347.arc6720568582813124125#46685496_0_6990327112563130929", 19));
			rowIds.add(new TableRowId("1346981172155_1347002801083_2615.arc4798755754814711921#12088900_1_171084248064070967", 164));
			SplitRowValues srv = new SplitRowValues(new HashMap<>());

			for (TableRowId rowId : rowIds) {
				List<TableRowId> rows = new ArrayList<>();
				rows.add(rowId);
				RowCluster rowCluster = new RowCluster(classId, rows, new RowClusterConstantSimilarities(1.0));

				ConcurrentHashMap<String, AtomicInteger> labelMap = new ConcurrentHashMap<>();
				Map<QualifiedPropertyId, FusedValue> values = new HashMap<>();

				AnnotatedQualifiedTableRow annotatedTableRow = tableRowCreator.getTableRow(rowId, srv);
				if (annotatedTableRow.getLabel() != null) {
					CHMHelper.getFromCHM(labelMap, StringNormalizer.normaliseValue(annotatedTableRow.getLabel().trim(), false), () -> new AtomicInteger())
							.getAndIncrement();
				}

				Entity entity = new Entity(rowCluster.getUuid(), classId,
						labelMap.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().intValue())), values, rowCluster);

				csLabel.run(entity, config);
			}

		}

		int numberOfTables = repository.getAllTableIds().size();

		Set<AnnotatedTableId> filteredTableIds = Collections.newSetFromMap(new ConcurrentHashMap<>());

		if (false) {

			PersistentParallel6.runAndFinish(repository.getAllTableIds(), (tableId) -> {
				boolean done = false;
				for (String s : repository.getTable(tableId).getStringTable().values()) {
					for (int i = 0; i < s.length(); i++) {
						if ((int) s.charAt(i) == 65533) {
							filteredTableIds.add(tableId);
							done = true;
							break;
						}
					}
					if (done)
						break;
				}
			}, this.getClass());

			int filteredSize = filteredTableIds.size();
		}
		MKDIR.createDirUnchecked(new File(getCacheFolder()));
		MKDIR.createDirUnchecked(new File(getPrecacheFolder()));

		// Map<ClassId, WeightsRepositoryImpl> clusteringWeightsRepository =
		// weightsAdapter.readWeights(allParents.keySet(),
		// clusteringWeightsFolder);
		Map<ClassId, LearnedModelRepository> clusteringWeightsRepository = LearnedModelAdapter.readModels((modelMap) -> new LearnedModelRepository(modelMap),
				allParents.keySet(), clusteringWeightsFolder);
		Map<ClassId, EM_LearnedModelRepository> instanceMatchingWeightsRepository = LearnedModelAdapter
				.readModels((modelMap) -> new EM_LearnedModelRepository(modelMap), allParents.keySet(), instanceMatchingWeightsFolder);
		// null; // weightsAdapter.readWeights(allParents.keySet(),
		// // instanceMatchingWeightsFolder)

		Map<ClassId, TableIdsSet> tableIdsSetMap = new HashMap<>();
		Map<ClassId, Set<AnnotatedTableId>> tableIdsRawSetMap = new HashMap<>();

		for (ClassId classId : allParents.keySet()) {
			TableIdsSet tableIdsSet = new TableIdsSet(this.tablesRepository.getAllTableIds(classId), new UuidStringId(classId.getFinalPart()));
			tableIdsSetMap.put(classId, tableIdsSet);
			tableIdsRawSetMap.put(classId, tableIdsSet.getTableIds());
		}

		AssociationSimilarities.tableSet = tableIdsRawSetMap;
		long startTime = System.nanoTime();

		Set<TableRowId> rowIdsAllClasses = new HashSet<>();
		for (ClassId classId : allParents.keySet()) {
			TableIdsSet tableIdsSet = tableIdsSetMap.get(classId);
			Set<TableRowId> rowIds = this.tablesRepository.getAllRowIds(tableIdsSet.getTableIds());
			rowIdsAllClasses.addAll(rowIds);
		}
		TableRowIdsSet rowIdsSetAllClasses = new TableRowIdsSet(rowIdsAllClasses, new UuidStringId("ALL"));

		for (ClassId classId : allParents.keySet()) {

			TableIdsSet tableIdsSet = tableIdsSetMap.get(classId);
			Set<TableRowId> rowIds = this.tablesRepository.getAllRowIds(tableIdsSet.getTableIds());
			System.out.println("FULL ROW SIZE FOR CLASS " + classId.getFinalPart() + " : " + rowIds.size());

			boolean setsCheck = false;
			if (classId.equals(ClassConstants.single)) {
				LabelRowSimilaritiesCreator.numDocs = 3;
				// setsCheck = true;

				// rowIds =
				// rowIds.stream().limit(60000).collect(Collectors.toSet());
			}
			// if (classId.equals(ClassConstants.single)) {
			// // rowIds =
			// // rowIds.stream().limit(80000).collect(Collectors.toSet());
			// }
			TableRowIdsSet rowIdsSet = new TableRowIdsSet(rowIds, new UuidStringId(classId.getFinalPart()));

			GroupedValues groupedValues = groupedValuesRepository.getGroupedValues(tablesRepository, tableIdsSet.getTableIds(), fusionConfig);
			SplitGroupedValues sgv = setExpansionRunner.getSplitGroupedValues(groupedValues, fusionConfig);
			SplitGroupedValuesSets splitGroupedValuesSets = new SplitGroupedValuesSets(sgv, sgv, sgv);

			SplitRowValues splitRowValues = rowValuesGetter.getSplitRowValues(tablesRepository, tableIdsSet);

			ClustererInput clustererInput = new ClustererInput(splitGroupedValuesSets, rowIdsSet.getRowIds(), rowIdsSet, tableIdsSet, null, classId,
					splitRowValues, allParents, nonTailPropertiesMap, rowIdsSetAllClasses);
			List<ThresholdContainer> thresholdContainers = setExpansionRunner.getThresholdContainers(setExpansionStrategiesConfig, null, classId, fusionConfig,
					strategiesConfig, sgv);

			SetExpansionStrategy strategy = SetExpansionStrategiesRepository.COMBINED_RF_BASE;

			RowClusteringConfig rowClusteringConfig = new RowClusteringConfig(setExpansionStrategiesConfig, thresholdContainers,
					clusteringWeightsRepository.get(classId), false, fusionConfig);

			implicitAttributeFinder.prepare(nonTailPropertiesMap);
			rowClustererRepository.init();

			ClusterList clusters;
			if (cacheExists(classId)) {
				clusters = readClusterCache(classId);
			} else {

				int sets = -1;
				if (setsCheck) {
					sets = 12;
				}

				RowClusterer clusterer = rowClustererRepository.get(strategy.getClusterer());

				RowClusterSimilarities sim = clusterer.clusterRows(rowClusteringConfig, clustererInput);

				if (!clusteringOutputCacheExists(classId) || ((setsCheck && setId != -1))) {
					if (setId != -1) {
						if (setId == 0) {
							ClusterPairsExporter.writeFull(getOutputCacheFile(classId), sim, clustererInput, sets, 0);
							ClusterPairsExporter.writeFull(getOutputCacheFile(classId), sim, clustererInput, sets, 3);
							ClusterPairsExporter.writeFull(getOutputCacheFile(classId), sim, clustererInput, sets, 6);
							ClusterPairsExporter.writeFull(getOutputCacheFile(classId), sim, clustererInput, sets, 9);
						} else if (setId == 1) {
							ClusterPairsExporter.writeFull(getOutputCacheFile(classId), sim, clustererInput, sets, 1);
							ClusterPairsExporter.writeFull(getOutputCacheFile(classId), sim, clustererInput, sets, 4);
							ClusterPairsExporter.writeFull(getOutputCacheFile(classId), sim, clustererInput, sets, 7);
							ClusterPairsExporter.writeFull(getOutputCacheFile(classId), sim, clustererInput, sets, 10);
						} else if (setId == 2) {
							ClusterPairsExporter.writeFull(getOutputCacheFile(classId), sim, clustererInput, sets, 2);
							ClusterPairsExporter.writeFull(getOutputCacheFile(classId), sim, clustererInput, sets, 5);
							ClusterPairsExporter.writeFull(getOutputCacheFile(classId), sim, clustererInput, sets, 8);
							ClusterPairsExporter.writeFull(getOutputCacheFile(classId), sim, clustererInput, sets, 11);
						}
						return;
					} else {
						ClusterPairsExporter.writeFull(getOutputCacheFile(classId), sim, clustererInput, sets, setId);
					}

				}

				if (!setsCheck || setId != -1) {
					if (true) {

						OptionalClusterCache cache = ClusterPairsExporter.getFullCache(getOutputCacheFile(classId), sim, clustererInput, sets);
						// Map<TableRowId, Integer> mappings =
						// ClusterPairsExporter.getMapping(getOutputCacheFile(classId)
						// + "_MAPPING.gz");
						// OptionalClusterCache cache = new
						// OptionalClusterCache(mappings, new HashMap<>(), new
						// OptionalClusterCacheMap[mappings.size()]);
						clusters = greedyClusteringParallelSplitter.clusterRows(sim, rowClusteringConfig, clustererInput, strategy.getOptimizableClassCluster(),
								alwaysDisableBlocking, cache, new Pair<>(getPrecacheFolder(), getPrefix(classId)));
					} else {
						clusters = greedyClusteringParallelSplitter.clusterRows(sim, rowClusteringConfig, clustererInput, alwaysDisableBlocking);

					}

					writeClustersCache(classId, clusters);
				} else {
					clusters = null;
				}
			}

			if (!setsCheck || setId != -1) {

				if (false && classId.equals(ClassConstants.placeClass)) {
					List<RowCluster> newList = new ArrayList<>();
					for (RowCluster rc : clusters.getClusterList()) {
						List<TableRowId> newRowList = rc.getRows().stream().filter((e) -> !filteredTableIds.contains(e.getTableId()))
								.collect(Collectors.toList());
						if (!newRowList.isEmpty()) {
							newList.add(new RowCluster(classId, newRowList, rc.getSimilarities()));
						}
					}

					clusters = new ClusterList(newList);
				}

				EntityMatcherConfig emConfig = new EntityMatcherConfig(clustererInput, nonTailPropertiesMap.get(classId), allParents);

				List<Entity> entities = entityCreator.createEntites(clustererInput.getGroupedValuesSets(), rowClusteringConfig.getThresholdContainers(),
						fusionConfig, strategiesConfig, rowClusteringConfig.getSetExpansionStrategiesConfig(), clusters.getClusterList(), clustererInput);

				EntityMatcherWeightedResults matcherReturn = entityMatchingRunner.weightedRun(emConfig, EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL,
						entities, allParents, instanceMatchingWeightsRepository.get(classId));

				System.out.println("Clsas: " + classId.getFinalPart());
				System.out.println("Total Number of unique existing instances matched to:" + matcherReturn.getExistingEntities().size());
				System.out
						.println("Total Number of existing entities:" + matcherReturn.getExistingEntities().values().stream().mapToInt((e) -> e.size()).sum());

				System.out.println("WITH AT LEAST ONE FACT: Total Number of unique existing instances matched to:"
						+ matcherReturn.getExistingEntities().values().stream().filter((e) -> {
							boolean found = false;
							for (ScoredEntity ent : e) {
								if (ent.getEntity().getValues().size() > 0) {
									found = true;
									break;
								}
							}
							return found;
						}).count());
				System.out.println(
						"WITH AT LEAST ONE FACT: Total Number of existing entities:" + matcherReturn.getExistingEntities().values().stream().mapToInt((e) -> {
							int c = 0;
							for (ScoredEntity ent : e) {
								if (ent.getEntity().getValues().size() > 0) {
									c++;
								}
							}
							return c;
						}).sum());

				printStatistics(entities.size(), matcherReturn.getNewEntities());

				MKDIR.createDirUnchecked(new File(currentCaseFolder));

				caseCreator.createForNewEntitiesByValuesSizeKperSize(currentCaseFolder + "/", classId.getFinalPart(), matcherReturn.getNewEntities(), 80);

				caseCreator.createForNewEntitiesByValuesSizeKperSize(currentCaseFolder + "/", classId.getFinalPart() + "_EXISTING",

						matcherReturn.getExistingEntities().values().stream().flatMap((e) -> e.stream()).filter((e) -> {
							return e.getEntity().getValues().size() > 0 && e.getEntity().getLabels().size() > 0;
						}).map((e) -> e).collect(Collectors.toList())

						, 15);

				ConcurrentHashMap<PropertyId, ConcurrentAccumulator<AccumInteger>> wrong = new ConcurrentHashMap<>();
				ConcurrentHashMap<PropertyId, ConcurrentAccumulator<AccumInteger>> correct = new ConcurrentHashMap<>();
				ConcurrentHashMap<PropertyId, ConcurrentAccumulator<AccumInteger>> newMap = new ConcurrentHashMap<>();

				PersistentParallel6.runAndFinish(matcherReturn.getExistingEntities().entrySet(), (entry) -> {
					for (ScoredEntity se : entry.getValue()) {
						for (Entry<QualifiedPropertyId, FusedValue> valueEntry : se.getEntity().getValues().entrySet()) {
							Optional<Boolean> correctValue = valueEvaluator.correct(new ValueId(entry.getKey(), valueEntry.getKey().getPropertyId()),
									valueEntry.getValue());
							ConcurrentAccumulator<AccumInteger> wrongCounter = CHMHelper.getFromCHM(wrong, valueEntry.getKey().getPropertyId(),
									() -> new ConcurrentAccumulator<>(AccumInteger.empty()));
							ConcurrentAccumulator<AccumInteger> correctCounter = CHMHelper.getFromCHM(correct, valueEntry.getKey().getPropertyId(),
									() -> new ConcurrentAccumulator<>(AccumInteger.empty()));
							ConcurrentAccumulator<AccumInteger> newCounter = CHMHelper.getFromCHM(newMap, valueEntry.getKey().getPropertyId(),
									() -> new ConcurrentAccumulator<>(AccumInteger.empty()));

							if (correctValue.isPresent()) {
								if (correctValue.get()) {
									wrongCounter.add(new AccumInteger(1));
								} else {
									correctCounter.add(new AccumInteger(1));
								}
							} else {
								newCounter.add(new AccumInteger(1));
							}
						}
					}
				}, this.getClass());

				{
					StringWriter sw = new StringWriter();
					CSVWriter csvWriter = new CSVWriter(sw);
					int columnCount = 7;
					{
						int i = 0;
						String[] headerLine = new String[columnCount];
						headerLine[i++] = "Class";
						headerLine[i++] = "Property";
						headerLine[i++] = "Correct";
						headerLine[i++] = "Wrong";
						headerLine[i++] = "TotalOverlapping";
						headerLine[i++] = "ExpectedAccuracy";
						headerLine[i++] = "New";
						csvWriter.writeNext(headerLine);
					}

					for (PropertyId propertyId : newMap.keySet()) {
						int correctCount = correct.get(propertyId).get().get();
						int wrongCount = wrong.get(propertyId).get().get();
						int total = correctCount + wrongCount;
						int newCount = newMap.get(propertyId).get().get();
						double accuracy = ((double) correctCount) / ((double) total);
						int i = 0;
						String[] headerLine = new String[columnCount];
						headerLine[i++] = classId.getFinalPart();
						headerLine[i++] = propertyId.getFinalPart();
						headerLine[i++] = String.valueOf(correctCount);
						headerLine[i++] = String.valueOf(wrongCount);
						headerLine[i++] = String.valueOf(total);
						headerLine[i++] = DecimalFormatter.format3(accuracy);
						headerLine[i++] = String.valueOf(newCount);
						csvWriter.writeNext(headerLine);
					}

					csvWriter.close();

					System.out.println(sw.toString());

				}

			}
		}

		long estimatedTime = System.nanoTime() - startTime;

		System.out.println("ThisRunTook: " + GSRun.formatInterval(estimatedTime));
		int sdfgag = 123;
		try {
			PersistentParallel6.close();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		int sdf = 123;
	}

	private void writeClustersCache(ClassId classId, ClusterList clusters) {
		Map<RowClusterUUID, List<TableRowId>> clusterMap = new HashMap<>();
		for (RowCluster rc : clusters.getClusterList()) {
			clusterMap.put(rc.getUuid(), rc.getRows());
		}
		chmKryoAdapter.writeWithList(getCacheFolder(), getPrefix(classId), clusterMap);

	}

	private String getPrefix(ClassId classId) {
		return classId.getFinalPart() + "_";
	}

	private String getCacheFolder() {
		return cacheFolder + "/FullRowClustering3_Done/";
	}

	private String getPrecacheFolder() {
		return cacheFolder + "/FullRowClustering2_BeforeKLJ/";
	}

	private String getOutputCacheFile(ClassId classId) {
		MKDIR.createDirUnchecked(new File(cacheFolder + "/FullRowClustering1_SimCache/"));
		return cacheFolder + "/FullRowClustering1_SimCache/" + classId.getFinalPart();
	}

	private ClusterList readClusterCache(ClassId classId) {
		Map<RowClusterUUID, List<TableRowId>> map = chmKryoAdapter.readWithList(RowClusterUUID.class, TableRowId.class, getCacheFolder(), getPrefix(classId));
		List<RowCluster> rowClusterList = new ArrayList<>();
		for (Entry<RowClusterUUID, List<TableRowId>> entry : map.entrySet()) {
			rowClusterList.add(new RowCluster(entry.getKey(), classId, entry.getValue(), new RowClusterConstantSimilarities(1.0)));
		}

		return new ClusterList(rowClusterList);

	}

	private boolean clusteringOutputCacheExists(ClassId classId) {
		return new File(getOutputCacheFile(classId) + ".gz").exists();
	}

	private boolean cacheExists(ClassId classId) {
		return chmKryoAdapter.cacheExists(getCacheFolder(), getPrefix(classId));
	}

	public static void printStatistics(int totalEntitiesSize, Collection<ScoredEntity> entities) {

		Collection<Integer> fusedValuesDistribution = new ArrayList<>();
		Collection<Integer> rowSizeDistribution = new ArrayList<>();
		int entitiesSize = 0;
		int factsSize = 0;

		Map<ClassId, Map<PropertyId, MutableInt>> propertyCount = new HashMap<>();
		for (ScoredEntity scoredEntity : entities) {
			Entity entity = scoredEntity.getEntity();
			if (entity.getValues().size() > 0 && entity.getLabels().size() > 0) {
				entitiesSize++;
				fusedValuesDistribution.add(entity.getValues().size());
				factsSize += entity.getValues().size();
				rowSizeDistribution.add(entity.getRowCluster().getRows().size());
				Map<PropertyId, MutableInt> classMap = CHMHelper.getFromMap(propertyCount, entity.getClassId(), () -> new HashMap<>());

				for (QualifiedPropertyId propertyId : entity.getValues().keySet()) {
					CHMHelper.getFromMap(classMap, propertyId.getPropertyId(), () -> new MutableInt(0)).increment();
				}
			}
		}

		System.out.println("Number of total entities: " + totalEntitiesSize);
		System.out.println("Number of new entities: " + entitiesSize);
		System.out.println("Number of total new fused facts: " + factsSize);

		System.out.println("Fused Value Distribution");
		SetExpansionRunner.printDistribution(fusedValuesDistribution);

		System.out.println("Row Size Distribution");
		SetExpansionRunner.printDistribution(rowSizeDistribution);

		StringWriter sw = new StringWriter();
		CSVWriter csvWriter = new CSVWriter(sw);
		int columnCount = 3;
		{
			int i = 0;
			String[] headerLine = new String[columnCount];
			headerLine[i++] = "Class";
			headerLine[i++] = "Property";
			headerLine[i++] = "Facts";
			csvWriter.writeNext(headerLine);
		}
		for (Entry<ClassId, Map<PropertyId, MutableInt>> entry : propertyCount.entrySet()) {

			for (Entry<PropertyId, MutableInt> innerEntry : entry.getValue().entrySet()) {
				int i = 0;
				String[] line = new String[columnCount];
				line[i++] = entry.getKey().getFinalPart();
				line[i++] = innerEntry.getKey().getPropertyId();
				line[i++] = String.valueOf(innerEntry.getValue().intValue());

				csvWriter.writeNext(line);

			}

			{
				int i = 0;
				String[] line = new String[columnCount];
				line[i++] = "";
				line[i++] = "";
				line[i++] = "";

				csvWriter.writeNext(line);
			}
		}

		try {
			csvWriter.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		System.out.println(sw.toString());
	}
}
