package de.uni_mannheim.informatik.dws.expansion.samples.setExtension;

import java.io.File;
import java.io.IOException;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Random;
import java.util.Set;
import java.util.TreeMap;
import java.util.UUID;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

import javax.inject.Inject;

import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import com.opencsv.CSVWriter;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId.ClassIdStyle;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId.InstanceIdStyle;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdsSet;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.ResultsCategory;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.generic.GenericResultsWriter;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig.TimeDependence;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategyRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.DecimalFormatter;
import de.mannheim.uni.expansion.utils.DoubleIteraterable;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.NumOfCpusGetter;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.CategoricalIntObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.QuantityObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject.Precision;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedAbstractObjectHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.attribute.AttributeMatchingAdapter;
import de.uni_mannheim.informatik.dws.attribute.evaluation.AttributeMappingsEvaluator;
import de.uni_mannheim.informatik.dws.attribute.evaluation.AttributeMappingsPerformance;
import de.uni_mannheim.informatik.dws.attribute.model.AttributeMappings;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentAccumulator;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.LearnedModelRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepositoryImpl;
import de.uni_mannheim.informatik.dws.expansion.utils.remote.RepositoryLoading;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities;
import expansion.set.distant_supervision.DS_Container;
import expansion.set.distant_supervision.DistantSupervisionInput;
import expansion.set.distant_supervision.baseline.average.ND_Average_Creator;
import expansion.set.distant_supervision.baseline.average.RC_Average_Creator;
import expansion.set.distant_supervision.baseline.average.RC_Average_Creator.ND_Average_Strategy;
import expansion.set.distant_supervision.bootsrap.GenericStrategyCreator;
import expansion.set.distant_supervision.bootsrap.WS_RC_No_Clustering_Strategy_Creator;
import expansion.set.distant_supervision.bootsrap.WS_RC_No_Clustering_Strategy_Creator.WS_METHOD;
import expansion.set.distant_supervision.bootsrap.WS_RC_Strategy_Creator;
import expansion.set.distant_supervision.matching_rules.MatchingRulesWrapperCreator;
import expansion.set.distant_supervision.matchingmethods.ND_MatchingRules_Creator;
import expansion.set.distant_supervision.matchingmethods.RC_MatchingRules_Creator;
import expansion.set.distant_supervision.runners.DS_ClusteringRunnerCreator;
import expansion.set.distant_supervision.runners.DS_ClusteringRunnerCreator.ClusteringRunner;
import expansion.set.entitymatching.EntityMatcherFullResults;
import expansion.set.entitymatching.EntityMatchingModelLearner;
import expansion.set.entitymatching.matchers.EntityMatcherStrategies;
import expansion.set.entitymatching.modelLearning.EM_LearnedModelRepository;
import expansion.set.entitymatching.modelLearning.weka.EntityMatcherWekaLearner;
import expansion.set.entitymatching.run.EntityMatchingRun;
import expansion.set.entitymatching.run.EntityMatchingRunRepository;
import expansion.set.evaluation.SetExpansionRunner;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.evaluation.SetExpansionStrategiesConfig.SetExpansionStrategiesConfigBuilder;
import expansion.set.evaluation.goldstandard.GoldStandardClusters;
import expansion.set.evaluation.goldstandard.GoldStandardClustersSet;
import expansion.set.evaluation.goldstandard.GoldStandardClustersSetPerClass;
import expansion.set.evaluation.matching.NewDetectionPerformance;
import expansion.set.evaluation.pipelineresults.PipelinePerformance;
import expansion.set.gscreation.GsOutputReader;
import expansion.set.hyperparameter.HyperparameterOptimizer;
import expansion.set.hyperparameter.HyperparameterOptimizer.ValidationRun;
import expansion.set.hyperparameter.RowClusteringLearnableSimCache;
import expansion.set.learning.ClusteringModelLearnerRunner;
import expansion.set.learning.GoldStandardSplitter;
import expansion.set.learning.MT_WekaLearnedModel;
import expansion.set.optimizing.LearnedModelAdapter;
import expansion.set.optimizing.WeightsAdapter;
import expansion.set.strategies.SetExpansionStrategiesRepository;
import expansion.set.strategies.SetExpansionStrategy;
import expansion.set.strategies.combined.CombinedWekaGeneric;

public class GSRun {

	public static final String LEARN_MODELS = "LEARN_MODELS";
	public static final String WEAK_SUPERVISION = "WEAK_SUPERVISION";
	public static final String EVALUATE_METHOD = "EVALUATE_METHOD";

	public static void main(String[] args) throws IOException {

		for (int FOLDS = 3; FOLDS > 2; FOLDS--) {

			// REGEX FOR COLLEGE TEAMS
			// https://regex101.com/r/9X4YiJ/1

			Map<ClassId, Set<ClassId>> topClasses = ExtractTablesForGoldStandardRun.getClassesWithHierarchyFilteredOrNot();
			// topClasses.remove(ClassConstants.single);

			// topClasses.remove(ClassConstants.gridIronFootballPlayer);
			// topClasses.remove(ClassConstants.placeClass);

			// topClasses.add(ClassConstants.placeClass);
			// topClasses.add(ClassConstants.person);
			// topClasses.add(ClassConstants.populatedPlace);
			// MODES = EVALUATE_METHOD and WEAK_SUPERVISION
			String RUN_MODE = EVALUATE_METHOD;

			if (args.length > 7) {
				RUN_MODE = args[7];
			}

			boolean alwaysDisableBlocking = true;
			if (args.length > 6) {
				alwaysDisableBlocking = Boolean.parseBoolean(args[6]);
			}

			int currentRun = 1;
			if (args.length > 5) {
				currentRun = Integer.parseInt(args[5]);
			}

			SetExtension setExtension = new SetExtension(args, topClasses.keySet(), ExtractHomonymTablesForGoldStandardRun.getMinPropertyDensity(),
					RepositoryLoading.REMOTE_LOAD_REMOTELY, false, true, currentRun, alwaysDisableBlocking);

			String mappingsFolder = setExtension.getMappingsFolder();

			GsOutputReader reader = new GsOutputReader(InstanceIdStyle.DBPEDIA_URI);
			String gsOutputFolder = mappingsFolder + "/goldStandard/";

			ExtendedKnowledgeBaseValueCreator extendedKnowledgeBaseValueCreator = new ExtendedKnowledgeBaseValueCreator(
					setExtension.getDataHome() + "/configs/");

			Map<String, InstanceId> positionIndexNormalized = extendedKnowledgeBaseValueCreator.getPositionIndexNormalized();

			// String luceneIndexCache = mappingsFolder + "lucene/";
			// MKDIR.createDirUnchecked(new File(luceneIndexCache));

			String instanceMappingsFolder = gsOutputFolder + "/instances/";
			String newInstancesFile = gsOutputFolder + "/newInstances.csv";
			String processedInstancesFile = gsOutputFolder + "/processed.csv";
			String processedTablesFile = gsOutputFolder + "/processedTables.csv";
			String tableClassesFile = gsOutputFolder + "/classes_GS.csv";
			String factsFile = gsOutputFolder + "/facts";

			Map<AnnotatedTableId, ClassId> tableClasses = reader.readTableClasses(tableClassesFile);
			Set<AnnotatedTableId> processedTables = reader.readListOfTables(processedTablesFile);
			Set<InstanceId> processedInstancesNonFinal = reader.readInstancesList(processedInstancesFile);
			GoldStandardClusters mappings = reader.getMapping(instanceMappingsFolder, processedTables);
			GoldStandardClusters fullMappings = reader.getMapping(instanceMappingsFolder, null);
			Set<InstanceId> processedInstances = processedInstancesNonFinal.stream().filter((e) -> mappings.getCluster(e) != null).collect(Collectors.toSet());
			Set<InstanceId> missingInstances = processedInstancesNonFinal.stream().filter((e) -> mappings.getCluster(e) == null).collect(Collectors.toSet());
			Set<InstanceId> newInstances = reader.readInstancesList(newInstancesFile);

			Set<InstanceId> extraExisting = new HashSet<>();

			extraExisting.add(new InstanceId("http://dbpedia.org/resource/When_Love_Comes_to_Town"));
			extraExisting.add(new InstanceId("http://dbpedia.org/resource/Enter_Sandman"));
			extraExisting.add(new InstanceId("http://dbpedia.org/resource/Only_Girl_(in_the_World)"));
			extraExisting.add(new InstanceId("http://dbpedia.org/resource/Teenage_Dream_(Katy_Perry_song)"));
			extraExisting.add(new InstanceId("http://dbpedia.org/resource/Sunday_Bloody_Sunday"));

			extraExisting.add(new InstanceId("http://dbpedia.org/resource/Hey_There_Delilah"));
			extraExisting.add(new InstanceId("http://dbpedia.org/resource/Apologize_(song)"));
			extraExisting.add(new InstanceId("http://dbpedia.org/resource/New_Soul"));
			extraExisting.add(new InstanceId("http://dbpedia.org/resource/Master_of_Puppets_(song)"));
			extraExisting.add(new InstanceId("http://dbpedia.org/resource/Get_On_Your_Boots"));

			extraExisting.add(new InstanceId("http://dbpedia.org/resource/Praise_You"));
			extraExisting.add(new InstanceId("http://dbpedia.org/resource/Wind_of_Change_(Scorpions_song)"));
			extraExisting.add(new InstanceId("http://dbpedia.org/resource/Bubbly"));
			extraExisting.add(new InstanceId("http://dbpedia.org/resource/Beautiful_Day"));
			extraExisting.add(new InstanceId("http://dbpedia.org/resource/Down_Down"));

			Map<InstanceId, Set<TableRowId>> additionalExisting = fullMappings.getTestingSetMappingClusters().entrySet().stream()
					.filter((e) -> extraExisting.contains(e.getKey())).collect(Collectors.toMap(Entry::getKey, Entry::getValue));

			Map<InstanceId, ClassId> instancesToClassMappings = fullMappings.getTestingSetMappingClusters().entrySet().stream()
					.collect(Collectors.toMap(Entry::getKey, (e) -> {
						ClassId classId = null;
						for (TableRowId rowId : e.getValue()) {
							ClassId classIdOfRow = tableClasses.get(rowId.getTableId());
							if (classId == null) {
								classId = classIdOfRow;
							} else if (!classId.equals(classIdOfRow)) {
								throw new IllegalStateException();
							}
						}
						return classId;
					}));

			Map<Pair<InstanceId, PropertyId>, Tri<String, Boolean, Boolean>> facts = reader.getFacts(factsFile, instancesToClassMappings,
					setExtension.getNonTailProperties());

			Map<PropertyId, ValueType> typeMappings = new HashMap<>();
			typeMappings.put(PropertyId.fromString("birthDate"), ValueType.TIME);
			typeMappings.put(PropertyId.fromString("deathDate"), ValueType.TIME);
			typeMappings.put(PropertyId.fromString("position"), ValueType.INSTANCE_REFERENCE);
			typeMappings.put(PropertyId.fromString("birthPlace"), ValueType.INSTANCE_REFERENCE);
			typeMappings.put(PropertyId.fromString("college"), ValueType.INSTANCE_REFERENCE);
			typeMappings.put(PropertyId.fromString("draftYear"), ValueType.TIME);
			typeMappings.put(PropertyId.fromString("team"), ValueType.INSTANCE_REFERENCE);
			typeMappings.put(PropertyId.fromString("number"), ValueType.CATEGORICAL_INT);
			typeMappings.put(PropertyId.fromString("Person/weight"), ValueType.QUANTITY);
			typeMappings.put(PropertyId.fromString("height"), ValueType.QUANTITY);
			typeMappings.put(PropertyId.fromString("draftPick"), ValueType.CATEGORICAL_INT);
			typeMappings.put(PropertyId.fromString("draftRound"), ValueType.CATEGORICAL_INT);
			typeMappings.put(PropertyId.fromString("highschool"), ValueType.INSTANCE_REFERENCE);
			typeMappings.put(PropertyId.fromString("country"), ValueType.INSTANCE_REFERENCE);
			typeMappings.put(PropertyId.fromString("populationTotal"), ValueType.QUANTITY);
			typeMappings.put(PropertyId.fromString("continent"), ValueType.INSTANCE_REFERENCE);
			typeMappings.put(PropertyId.fromString("elevation"), ValueType.QUANTITY);
			typeMappings.put(PropertyId.fromString("isPartOf"), ValueType.INSTANCE_REFERENCE);
			typeMappings.put(PropertyId.fromString("postalCode"), ValueType.STRING);
			typeMappings.put(PropertyId.fromString("utcOffset"), ValueType.QUANTITY);
			typeMappings.put(PropertyId.fromString("populationMetro"), ValueType.QUANTITY);
			typeMappings.put(PropertyId.fromString("populationDensity"), ValueType.QUANTITY);
			typeMappings.put(PropertyId.fromString("postalCode"), ValueType.STRING);

			typeMappings.put(new PropertyId("http://www.w3.org/2003/01/geo/wgs84_pos#lat"), ValueType.QUANTITY);
			typeMappings.put(new PropertyId("http://www.w3.org/2003/01/geo/wgs84_pos#long"), ValueType.QUANTITY);

			typeMappings.put(PropertyId.fromString("releaseDate"), ValueType.TIME);
			typeMappings.put(PropertyId.fromString("musicalArtist"), ValueType.INSTANCE_REFERENCE);
			typeMappings.put(PropertyId.fromString("recordLabel"), ValueType.INSTANCE_REFERENCE);
			typeMappings.put(PropertyId.fromString("album"), ValueType.INSTANCE_REFERENCE);
			typeMappings.put(PropertyId.fromString("genre"), ValueType.INSTANCE_REFERENCE);

			typeMappings.put(PropertyId.fromString("runtime"), ValueType.QUANTITY);
			typeMappings.put(PropertyId.fromString("producer"), ValueType.INSTANCE_REFERENCE);
			typeMappings.put(PropertyId.fromString("writer"), ValueType.INSTANCE_REFERENCE);
			typeMappings.put(PropertyId.fromString("bSide"), ValueType.INSTANCE_REFERENCE);
			typeMappings.put(PropertyId.fromString("area"), ValueType.QUANTITY);

			Map<ClassId, Set<PropertyId>> classPropertyIds = new HashMap<>();
			classPropertyIds.put(ClassConstants.gridIronFootballPlayer, new HashSet<>(typeMappings.keySet()));

			Map<InstanceId, ClassId> processedInstanceToClassMapping = mappings.getTestingSetMappingClusters().entrySet().stream()
					.filter((e) -> processedInstances.contains(e.getKey())).collect(Collectors.toMap(Entry::getKey, (e) -> {
						ClassId classId = null;
						for (TableRowId rowId : e.getValue()) {
							ClassId classIdOfRow = tableClasses.get(rowId.getTableId());
							if (classId == null) {
								classId = classIdOfRow;
							} else if (!classId.equals(classIdOfRow)) {
								throw new IllegalStateException();
							}
						}
						return classId;
					}));

			Pair<Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>>, Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>>> factsMapPair = getFactsMap(
					facts, processedInstances, processedInstanceToClassMapping, typeMappings, positionIndexNormalized);

			Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>> factsMap = factsMapPair.getFirst();
			Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>> fullCorrectMap = factsMapPair.getSecond();

			Map<ClassId, Set<TableRowId>> input = new HashMap<>();

			for (InstanceId instanceId : processedInstances) {
				CHMHelper.getFromMap(input, processedInstanceToClassMapping.get(instanceId), () -> new HashSet<>()).addAll(mappings.getCluster(instanceId));
			}

			for (Iterator<ClassId> it = input.keySet().iterator(); it.hasNext();) {
				if (!topClasses.containsKey(it.next())) {
					it.remove();
				}
			}

			for (ClassId classId : topClasses.keySet()) {
				String instanceIdString = classId.getReadableString(ClassIdStyle.DBPEDIA_URI) + "_BAD_ROWS";
				InstanceId instanceId = new InstanceId(instanceIdString, instanceIdString);
				if (!mappings.getTestingSetMappingClusters().containsKey(instanceId)) {
					int sdf = 123;
				}
				Set<TableRowId> badCluster = mappings.getCluster(instanceId);
				// CHMHelper.getFromMap(input, classId, () -> new
				// HashSet<>()).addAll(badCluster);

			}

			Map<ClassId, Set<InstanceId>> newInstancesProcessed = new HashMap<>();
			for (InstanceId instanceId : processedInstances) {
				if (newInstances.contains(instanceId)) {
					CHMHelper.getFromMap(newInstancesProcessed, processedInstanceToClassMapping.get(instanceId), () -> new HashSet<>()).add(instanceId);

				}
			}

			for (ClassId classId : topClasses.keySet()) {
				if (!newInstancesProcessed.containsKey(classId)) {
					newInstancesProcessed.put(classId, new HashSet<>());
				}
			}

			Map<ClassId, GoldStandardClusters> goldStandardClusters = splitAndFilterGoldStandard(mappings, processedInstanceToClassMapping);

			for (Iterator<ClassId> it = goldStandardClusters.keySet().iterator(); it.hasNext();) {
				if (!topClasses.containsKey(it.next())) {
					it.remove();
				}
			}

			// KBValueCreator kbValueCreator =
			// extendedKnowledgeBaseValueCreator.getCreator();
			//
			// // SetExtension setExtension = new SetExtension(args,
			// // topClasses.keySet(),
			// // ExtractTablesForGoldStandardRun.getMinPropertyDensity());
			//
			// List<KBValueCreator> valueCreatorList = new ArrayList<>();
			// valueCreatorList.add(kbValueCreator);

			// MaxDiffSetter maxDiffSetter =
			// setExtension.getGraph().get(MaxDiffSetter.class);

			String thresholdsFolder = setExtension.getThresholdsFolder();
			String resultsFolder = setExtension.getResultsFolder();

			// kb = setExtension.getGraph().get(KnowledgeBase.class);
			// OntologyRepository ontology =
			// setExtension.getGraph().get(OntologyRepository.class);

			GSRun fusion = setExtension.getGraph().get(GSRun.class);

			KnowledgeBase kb = setExtension.getGraph().get(KnowledgeBase.class);
			for (Entry<InstanceId, ClassId> instanceEntry : processedInstanceToClassMapping.entrySet()) {
				if (!newInstances.contains(instanceEntry.getKey())) {
					if (!kb.hasInstance(instanceEntry.getValue(), instanceEntry.getKey())) {
						System.out.println("Instance missing in KB: " + instanceEntry.getKey() + ", ClassId: " + instanceEntry.getValue());
					}
				}
			}

			fusion.checkTables(tableClasses);

			// maxDiffSetter.run(setExtension.getDataHome());

			// fusion.runFull(resultsFolder, "NORMAL", fusionConfig);
			//
			// {
			// // fusion.createMaxDiffsFile(fusionConfig,
			// // FusionStrategyRepository.NEW_TS_QKBT, maxDiffsFileNewTS);
			// // fusion.createMaxDiffsFile(fusionConfig,
			// // FusionStrategyRepository.NEW_TS_QKBT_PROPAGATED,
			// // maxDiffsFileNewTSPropgated);
			// // fusion.createMaxDiffsFile(fusionConfig,
			// // FusionStrategyRepository.QKBT_Mapped,
			// // maxDiffsFileQKBTMapped);
			//
			// int sdf = 123;
			// }
			//
			// fusion.setMaxDiffFromFile(FusionStrategyRepository.NEW_TS_QKBT,
			// maxDiffsFileNewTS);
			//
			// fusion.setMaxDiffFromFile(FusionStrategyRepository.NEW_TS_QKBT_PROPAGATED,
			// maxDiffsFileNewTSPropgated);
			//
			// fusion.setMaxDiffFromFile(FusionStrategyRepository.QKBT_Mapped,
			// maxDiffsFileQKBTMapped);
			AttributeMappings gsMappings = reader.getAttributeMappings(gsOutputFolder + "/attributes/", processedTables, tableClasses,
					setExtension.getNonTailProperties());

			AttributeMappings currentMappings = fusion.adapter.getAttributeMappings(gsMappings.tableSet());

			FusionPerformance<AttributeMappingsPerformance> t2kResult = fusion.evaluator.evaluate(gsMappings, currentMappings, kb.getNormalizationIndex());
			ColumnReMatcherRun.print("Current_Mappings", t2kResult);

			//
			FusionConfig fusionConfig = setExtension.getFusionConfig();

			SetExpansionStrategiesConfigBuilder builder = setExtension.getSetExpansionStrategiesConfig().rebuilder();
			// builder.addFusionSetup(FusionStrategyRepository.QKBT_Mapped,
			// PropertiesConfig.builder().setTimeDependence(TimeDependence.YES).build());

			// builder.addStrategy(SetExpansionStrategiesRepository.LABEL_BLOCKED_SIMILARITY);

			// builder.addStrategy(SetExpansionStrategiesRepository.INSTANCE_BASED_CLUSTERER);
			//
			// builder.addStrategy(SetExpansionStrategiesRepository.GS_CLUSTERER);

			// builder.addStrategy(SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY);
			// builder.addStrategy(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_DIRECT);
			// builder.addStrategy(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_KBT_SCORED);
			// //
			// builder.addStrategy(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_KBT_THRESHOLDED);
			// builder.addStrategy(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_KBT_THRESHOLDED_AND_SCORED);

			if (false) {
				builder.addStrategy(SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY);
				// builder.addStrategy(SetExpansionStrategiesRepository.BAG_OF_WORDS_TABLE_AND_LABEL_WITHOUT_TFIDF);
				// builder.addStrategy(SetExpansionStrategiesRepository.ALL_NO_AA);
				// builder.addStrategy(SetExpansionStrategiesRepository.NEW_COMBINED);

			}

			if (false) {

				// builder.addStrategy(SetExpansionStrategiesRepository.NEW_COMBINED);
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_LinearRegression);
				//
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_LogisticRegression);
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_J48);
				//
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_REPTree);
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_M5P);
				// //
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_SimpleCART);
				//
				// //
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest_BackFitting);
				//
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_LMT);
				//
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomTree);
				//
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest_Classifier);
				//
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest);
				//
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_LMT_Options);
				//
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest_Classifier);
				//
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BOW_NO_LABEL);
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BOW_NO_LABEL_SAME_TABLE_SIM);
				// builder.addStrategy(SetExpansionStrategiesRepository.NEW_COMBINED);
				//
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest_OPTIMIZED);
				//
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_LMT);
				//
				// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_LMT_MT);
			}

			// if (false) {
			// builder.addStrategy(SetExpansionStrategiesRepository.NEW_COMBINED);
			//
			// //
			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest);
			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BOW_NO_LABEL);
			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BOW_NO_LABEL_SAME_TABLE_SIM);
			// //
			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest_Classifier_MT);
			// builder.addStrategy(SetExpansionStrategiesRepository.Combined_RandomForest_Classifier_MultiTree_New_BOW);
			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest_Classifier_MT_NEW_BOW_TBLE_SIM);
			//
			// }
			// //
			// builder.addStrategy(SetExpansionStrategiesRepository.NEW_COMBINED);
			//
			// //
			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_LinearRegression);
			// //
			// //
			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_LogisticRegression);
			// // if (alwaysDisableBlocking) {
			// if (false) {
			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest);
			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomTree_BackFitting);
			// }
			//
			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest);

			// builder.addStrategy(SetExpansionStrategiesRepository.EXTERNAL);
			// }
			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BOW_NO_LABEL);

			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_LMT);

			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BOW_NO_LABEL);
			// builder.addStrategy(SetExpansionStrategiesRepository.Combined_RandomForest_Classifier_MultiTree_New_BOW);
			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest_Classifier_MT_NEW_BOW_TBLE_SIM);

			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BOW_NO_LABEL);
			//
			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest_Classifier_MT_NEW_BOW_TBLE_SIM);

			//

			//

			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomTree_BackFitting);
			// builder.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomTree_BackFitting);
			// builder.addStrategy(SetExpansionStrategiesRepository.ALL_WTIH_IMPLICIT);

			if (false) {
				// builder.addStrategy(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_DIRECT_FULL_INDIVIDUAL_OPTIMIZED);
				//
				// //
				// builder.addStrategy(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_DIRECT_FULL);
				// builder.addStrategy(SetExpansionStrategiesRepository.ASSOCIATION_LABEL_COMBINED);
			}
			////
			// builder.addStrategy(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_KBT_POSITIVE_SCORED);
			// builder.addStrategy(SetExpansionStrategiesRepository.FULL_LABEL_ATTRIBUTE_KBT_JACCARD_LABEL_NORMALIZED);

			// builder.addStrategy(SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY);
			// builder.addStrategy(SetExpansionStrategiesRepository.FULL_LABEL_ATTRIBUTE_KBT_JACCARD_LABEL_NORMALIZED);

			// builder.addStrategy(SetExpansionStrategiesRepository.LABEL_SIMILARITY_NORMALIZED);

			// builder.addStrategy(SetExpansionStrategiesRepository.ATTRIBUTE_KBT);
			// builder.addStrategy(SetExpansionStrategiesRepository.ATTRIBUTE_DIRECT_LABEL_ABSMAX);
			// builder.addStrategy(SetExpansionStrategiesRepository.TABLE_JACCARD_INSTANCE_NORMALIZED);

			builder.clearFusionSetups().addFusionSetup(FusionStrategyRepository.COLUMN_SCORES,
					PropertiesConfig.builder().setTimeDependence(TimeDependence.NO).build());
			SetExpansionStrategiesConfig setExpansionStrategiesConfig = builder.build();

			StrategiesConfig strategiesConfig = StrategiesConfig.defaultConfig();

			// fusion.learnAndStoreMaxDiff(fusionConfig, "6", maxDiffsFolder,
			// false, false);
			// fusion.learnAndStoreMaxDiff(fusionConfig, "6", maxDiffsFolder,
			// false, true);

			// fusion.runner.run("SetExtension", fusionConfig, strategiesConfig,
			// setExpansionStrategiesConfig, resultsFolder, thresholdsFolder);

			Map<ClassId, Set<PropertyId>> classesAndProperties = classPropertyIds;
			List<Pair<ClassId, ?>> hierarchy = topClasses.keySet().stream().map((e) -> new Pair<>(e, null)).collect(Collectors.toList());
			Map<ClassId, Set<ClassId>> allParents = topClasses;

			System.out.println(classesAndProperties);

			String currentCase = setExtension.getCurrentCase();

			String currentCaseFolder = setExtension.getCurrentManualCaseFolder();

			String currentResultsFolder = resultsFolder + "/" + currentCase;
			MKDIR.createDirUnchecked(new File(currentResultsFolder));

			// ... the code being measured ...

			GoldStandardClustersSetPerClass goldStandardClustersSet = fusion.goldStandardSplitter.split(goldStandardClusters, newInstancesProcessed, factsMap,
					FOLDS);

			System.out.println("Positive-Pairs: ");
			for (Entry<ClassId, GoldStandardClusters> entry : goldStandardClusters.entrySet()) {
				int sum = 0;
				for (Set<TableRowId> set : entry.getValue().getTestingSetMappingClusters().values()) {
					if (set.size() > 1) {
						sum += DoubleIteraterable.getResultingSize(set.size());
					}
				}
				System.out.println("ClassId : " + entry.getKey().getFinalPart() + ", size: " + sum);

			}

			System.out.println("Positive-Cluster-Annotations: ");
			for (Entry<ClassId, GoldStandardClusters> entry : goldStandardClusters.entrySet()) {
				int newSum = 0;
				int existingSum = 0;

				for (InstanceId instanceId : entry.getValue().getTestingSetMappingClusters().keySet()) {
					if (newInstances.contains(instanceId)) {
						newSum++;
					} else {
						existingSum++;
					}
				}
				System.out.println("ClassId : " + entry.getKey().getFinalPart() + ", existing: " + existingSum + ", new: " + newSum);

			}

			Map<ClassId, TableRowIdsSet> fullInput = goldStandardClustersSet.getFullInput();

			Map<ClassId, Set<TableRowId>> fullInputRaw = goldStandardClustersSet.getFullInputRaw();

			String gsSplitStatistics = GoldStandardClustersSet.statistics(goldStandardClustersSet);

			Map<ClassId, Map<InstanceId, Integer>> conflictingSets = goldStandardClustersSet.getConflictingSets();

			if (RUN_MODE.equals(LEARN_MODELS)) {

				if (true) {
					String weightsFolder = setExtension.getMappingsCachesFolder() + "/models/clusteringModelsFull" + currentRun + "/";
					MKDIR.createDirUnchecked(new File(weightsFolder));

					// // FULL WEIGHTS LEARNING
					// Map<ClassId, WeightsRepositoryImpl> weightsRepository =
					// goldStandardClusters.keySet().stream()
					// .collect(Collectors.toMap(java.util.function.Function.identity(),
					// (e) -> new WeightsRepositoryImpl()));
					//
					// weightsRepository =
					// fusion.weightsAdapter.readWeights(goldStandardClusters.keySet(),
					// weightsFolder);
					SetExpansionStrategiesConfig setExpansionStrategiesConfigFull = setExpansionStrategiesConfig.rebuilder().clearStrategies()
							.addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE).build();
					//
					// Map<ClassId, WeightsRepositoryImpl> newWeightsRepository
					// =
					// fusion.weightsLearnerRunner.getAllWeights(weightsRepository,
					// fullInputRaw, fullInput,
					// fusionConfig, setExpansionStrategiesConfigFull,
					// strategiesConfig,
					// thresholdsFolder, goldStandardClusters,
					// newInstancesProcessed,
					// conflictingSets, allParents,
					// setExtension.getNonTailProperties());

					Map<ClassId, LearnedModelRepository> learnedModels = fusion.clusteringLearnerRunner.getAllWeights(new HashMap<>(), fullInputRaw, fullInput,
							fusionConfig, setExpansionStrategiesConfigFull, strategiesConfig, thresholdsFolder, goldStandardClusters, newInstancesProcessed,
							conflictingSets, allParents, setExtension.getNonTailProperties(), null);

					fusion.learnedModelAdapter.storeModels(weightsFolder, learnedModels);

				}

				if (true) {

					String instanceMatchingWeightsFolder = setExtension.getMappingsCachesFolder() + "/models/instanceMatchingModelsFull" + currentRun + "/";
					MKDIR.createDirUnchecked(new File(instanceMatchingWeightsFolder));

					List<EntityMatchingRun> runs = Arrays.asList(EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL,
							EntityMatchingRunRepository.CS_LABEL_EM_LABEL);
					Map<ClassId, Pair<EM_LearnedModelRepository, Map<EntityMatchingRun, NewDetectionPerformance>>> entityMatchingWeights = fusion.entityMatchingGS_ClustererWeightsAdapterAndLearner
							.getLearnedModel(runs, goldStandardClusters, goldStandardClusters, fusionConfig, setExpansionStrategiesConfig, strategiesConfig,
									fullInput, fullInputRaw, fullInputRaw, thresholdsFolder, allParents, newInstancesProcessed, newInstancesProcessed,
									setExtension.getNonTailProperties(), instanceMatchingWeightsFolder, null, additionalExisting, "");
				}

			}

			// String weightsFolder = setExtension.getMappingsCachesFolder() +
			// "/clusteringWeights" + currentRun + "/";
			// MKDIR.createDirUnchecked(new File(weightsFolder));
			//
			// String instanceMatchingWeightsFolder =
			// setExtension.getMappingsCachesFolder() +
			// "/instanceMatchingWeights" + currentRun + "/";
			// MKDIR.createDirUnchecked(new
			// File(instanceMatchingWeightsFolder));

			Map<ClassId, WeightsRepositoryImpl> weightsRepository = goldStandardClusters.keySet().stream()
					.collect(Collectors.toMap(java.util.function.Function.identity(), (e) -> new WeightsRepositoryImpl()));

			// ==== RUN TESTING WITHOUT WEIGHTS

			// if (false) {
			// fusion.runner.gsRunNew(testingInput, fullInput, fusionConfig,
			// thresholdsFolder, setExpansionStrategiesConfig, strategiesConfig,
			// testingGoldStandard,
			// testingNewEntityIds, allParents, testingFactsMap, fullCorrectMap,
			// null, new HashMap<>(), new HashMap<>(weightsRepository),
			// setExtension.getNonTailProperties(), alwaysDisableBlocking);
			// }

			if (RUN_MODE.equals(WEAK_SUPERVISION)) {

				List<Integer> tableSizes = new ArrayList<>();

				if (false) {
					tableSizes.add(250);
					tableSizes.add(500);
					tableSizes.add(1000);
					tableSizes.add(2000);
				} else {

					tableSizes.add(1000);
				}

				for (Integer tableSize : tableSizes) {
					System.out.println("TABLE SIZE: " + tableSize);

					if (true) {
						for (int i = 0; i < 1; i++) {
							SimpleLockableConcurrentCache.clearAllCaches();
							Map<String, TreeMap<String, List<String>>> rett = fusion.runForWeakSupervision(goldStandardClustersSet, fusionConfig,
									setExpansionStrategiesConfig, strategiesConfig, fullInput, thresholdsFolder, allParents, setExtension, null,
									conflictingSets, currentCaseFolder, fullCorrectMap, alwaysDisableBlocking, classesAndProperties, additionalExisting,
									tableSize);

							printout(rett);
						}
					}

					if (false) {
						System.out.println("COUNTRY_ATT_NO_MATCH");
						MatchingRulesWrapperCreator.COUNTRY_ATT_NO_MATCH = true;
						fusion.runForWeakSupervision(goldStandardClustersSet, fusionConfig, setExpansionStrategiesConfig, strategiesConfig, fullInput,
								thresholdsFolder, allParents, setExtension, null, conflictingSets, currentCaseFolder, fullCorrectMap, alwaysDisableBlocking,
								classesAndProperties, additionalExisting, tableSize);

						MatchingRulesWrapperCreator.COUNTRY_ATT_NO_MATCH = false;
						MatchingRulesWrapperCreator.COUNTRY_IMPL_NO_MATCH = false;
						MatchingRulesWrapperCreator.POSTAL_CODE_COUNTRY_MATCH = false;
						MatchingRulesWrapperCreator.LABEL_ISPARTOF_MATCH = false;
						MatchingRulesWrapperCreator.LABEL_POSTAL_CODE_MATCH = false;

						System.out.println("POSTAL_CODE_COUNTRY_MATCH");
						MatchingRulesWrapperCreator.POSTAL_CODE_COUNTRY_MATCH = true;

						SimpleLockableConcurrentCache.clearAllCaches();
						fusion.runForWeakSupervision(goldStandardClustersSet, fusionConfig, setExpansionStrategiesConfig, strategiesConfig, fullInput,
								thresholdsFolder, allParents, setExtension, null, conflictingSets, currentCaseFolder, fullCorrectMap, alwaysDisableBlocking,
								classesAndProperties, additionalExisting, tableSize);
						MatchingRulesWrapperCreator.COUNTRY_ATT_NO_MATCH = false;
						MatchingRulesWrapperCreator.COUNTRY_IMPL_NO_MATCH = false;
						MatchingRulesWrapperCreator.POSTAL_CODE_COUNTRY_MATCH = false;
						MatchingRulesWrapperCreator.LABEL_ISPARTOF_MATCH = false;
						MatchingRulesWrapperCreator.LABEL_POSTAL_CODE_MATCH = false;

						System.out.println("LABEL_ISPARTOF_MATCH");
						MatchingRulesWrapperCreator.LABEL_ISPARTOF_MATCH = true;

						SimpleLockableConcurrentCache.clearAllCaches();
						fusion.runForWeakSupervision(goldStandardClustersSet, fusionConfig, setExpansionStrategiesConfig, strategiesConfig, fullInput,
								thresholdsFolder, allParents, setExtension, null, conflictingSets, currentCaseFolder, fullCorrectMap, alwaysDisableBlocking,
								classesAndProperties, additionalExisting, tableSize);
						MatchingRulesWrapperCreator.COUNTRY_ATT_NO_MATCH = false;
						MatchingRulesWrapperCreator.COUNTRY_IMPL_NO_MATCH = false;
						MatchingRulesWrapperCreator.POSTAL_CODE_COUNTRY_MATCH = false;
						MatchingRulesWrapperCreator.LABEL_ISPARTOF_MATCH = false;
						MatchingRulesWrapperCreator.LABEL_POSTAL_CODE_MATCH = false;

						System.out.println("LABEL_POSTAL_CODE_MATCH");
						MatchingRulesWrapperCreator.LABEL_POSTAL_CODE_MATCH = true;

						SimpleLockableConcurrentCache.clearAllCaches();
						fusion.runForWeakSupervision(goldStandardClustersSet, fusionConfig, setExpansionStrategiesConfig, strategiesConfig, fullInput,
								thresholdsFolder, allParents, setExtension, null, conflictingSets, currentCaseFolder, fullCorrectMap, alwaysDisableBlocking,
								classesAndProperties, additionalExisting, tableSize);
					}
					// }
				}

			}

			if (RUN_MODE.equals(EVALUATE_METHOD)) {

				Map<String, TreeMap<String, List<String>>> rett = fusion.runEvaluateMethod(goldStandardClustersSet, fusionConfig, setExpansionStrategiesConfig,
						strategiesConfig, fullInput, thresholdsFolder, allParents, setExtension, null, conflictingSets, currentCaseFolder, fullCorrectMap,
						alwaysDisableBlocking, classesAndProperties, additionalExisting);

				printout(rett);

				// weightsRepository =
				// fusion.weightsAdapter.readWeights(goldStandardClusters.keySet(),
				// weightsFolder);
				//
				// Map<ClassId, WeightsRepositoryImpl> newWeightsRepository =
				// fusion.weightsLearnerRunner.getAllWeights(weightsRepository,
				// learningInput, fullInput,
				// fusionConfig, setExpansionStrategiesConfig, strategiesConfig,
				// thresholdsFolder, learningGoldStandard, learningNewEntityIds,
				// conflictingSets,
				// allParents, setExtension.getNonTailProperties());
				//
				// if (!weightsRepository.equals(newWeightsRepository)) {
				// fusion.weightsAdapter.storeWeights(weightsFolder,
				// newWeightsRepository);
				//
				// }
				//
				// // ==== RUN LEARNING WITH WEIGHTS
				//
				// if (true) {
				// fusion.runner.gsRunNew(learningInput, fullInput,
				// fusionConfig,
				// thresholdsFolder, setExpansionStrategiesConfig,
				// strategiesConfig,
				// learningGoldStandard, learningNewEntityIds, allParents,
				// learningFactsMap, fullCorrectMap, null, new HashMap<>(),
				// new HashMap<>(entityMatchingWeights),
				// setExtension.getNonTailProperties(), alwaysDisableBlocking);
				//
				// }
				//
				// // ==== RUN TESTING WITH WEIGHTS
				// if (true) {
				//
				// Map<String, SetExpansionStrategiesConfig> fusionStrategies =
				// new
				// HashMap<>();
				//
				// fusionStrategies.put("ColumnScores",
				// setExpansionStrategiesConfig.rebuilder().clearFusionSetups()
				// .addFusionSetup(FusionStrategyRepository.COLUMN_SCORES,
				// PropertiesConfig.builder().setTimeDependence(TimeDependence.NO).build())
				// .build());
				//
				// if (false) {
				// fusionStrategies.put("VOTING",
				// setExpansionStrategiesConfig.rebuilder().clearFusionSetups()
				// .addFusionSetup(FusionStrategyRepository.NORMAL,
				// PropertiesConfig.builder().setTimeDependence(TimeDependence.NO).build()).build());
				//
				// fusionStrategies.put("KBT",
				// setExpansionStrategiesConfig.rebuilder().clearFusionSetups()
				// .addFusionSetup(FusionStrategyRepository.KBT_NO_THRESHOLD,
				// PropertiesConfig.builder().setTimeDependence(TimeDependence.NO).build())
				// .build());
				// }
				//
				// for (Entry<String, SetExpansionStrategiesConfig>
				// fusionRunEntry :
				// fusionStrategies.entrySet()) {
				//
				// System.out.println("FUSION_RUN: " + fusionRunEntry.getKey());
				// long startTime = System.nanoTime();
				//
				// fusion.runner.gsRunNew(testingInput, fullInput, fusionConfig,
				// thresholdsFolder, fusionRunEntry.getValue(),
				// strategiesConfig,
				// testingGoldStandard, testingNewEntityIds, allParents,
				// testingFactsMap, fullCorrectMap, currentCaseFolder, new
				// HashMap<>(),
				// new HashMap<>(entityMatchingWeights),
				// setExtension.getNonTailProperties(), alwaysDisableBlocking);
				//
				// long estimatedTime = System.nanoTime() - startTime;
				//
				// System.out.println("ThisRunTook: " +
				// formatInterval(estimatedTime));
				// }
				//
				// }
			}

			// ("SetExtension", fusionConfig, strategiesConfig,
			// setExpansionStrategiesConfig, currentResultsFolder + "/",
			// thresholdsFolder,
			// TestingMode.ALL, classesAndProperties, allParents, hierarchy,
			// currentCaseFolder + "/");

			// fusion.runFull(resultsFolder,
			// "Timed-KBT-NO-MaxDiff-F025-corrected",
			// fusionConfig.rebuilder().setAdjustedRecall(true).build());

			// fusion.learnAndStoreMaxDiff(fusionConfig, "5", maxDiffsFolder,
			// true, false);
			// fusion.loadMaxDiffs("5", maxDiffsFolder, true, false);
			//
			// fusion.runFull(resultsFolder, "FULL_VARIANCE_RUN_2",
			// fusionConfig);

			// sourcesFilter.printCount();
			// fusion.runReplicateSimple(resultsFolder, prefix, fusionConfig);
			// fusion.runReplicate(resultsFolder, prefix, fusionConfig);
			// fusion.runSplit(resultsFolder, prefix, fusionConfig);

			// fusion.runFull(resultsFolder, "MAXDIFF_NEW", rebuilder.build(),
			// maxDiffsFolder, true);
			// fusion.runFull(resultsFolder, "MAXDIFF", fusionConfig,
			// maxDiffsFolder, true);

			// prefix = "MAXDIFF_NEW_2";
			// fusion.learnMaxDiff(resultsFolder, prefix, fusionConfig,
			// maxDiffsFolder, false);
			// fusion.learnMaxDiff(resultsFolder, prefix, fusionConfig,
			// maxDiffsFolder, true);
			//
			// prefix = "MAXDIFF_NEW_2_ADJUSTED";
			// rebuilder = fusionConfig.rebuilder();
			// rebuilder.setAdjustedRecall(true);
			// fusionConfig = rebuilder.build();
			// fusion.learnMaxDiff(resultsFolder, prefix, fusionConfig,
			// maxDiffsFolder, false);
			// fusion.learnMaxDiff(resultsFolder, prefix, fusionConfig,
			// maxDiffsFolder, true);

			// rebuilder = fusionConfig.rebuilder();
			// rebuilder.setAdjustedRecall(true);
			// fusionConfig = rebuilder.build();
			//
			// fusion.setMaxDiffFromFile(FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL,
			// maxDiffsFolder + "/maxDiff_new_ts_mapping_" +
			// "MAXDIFF_NEW_ADJUSTED" + "_" + "perProperty" + ".json");
			//
			// //
			// fusion.setMaxDiffFromFile(FusionStrategyRepository.QKBT_Mapped,
			// maxDiffsFolder + "/maxDiff_new_timed_kbt_" +
			// "MAXDIFF_NEW_ADJUSTED" + "_" + "perProperty" + ".json");
			//
			// prefix = "YEAR_FULL_RUN";
			// fusion.runFull(resultsFolder, prefix, fusionConfig,
			// maxDiffsFolder, false);

			// fusion.runReplicateSimple(resultsFolder, prefix, fusionConfig);
			// fusion.runReplicate(resultsFolder, prefix, fusionConfig);
			// fusion.runSplit(resultsFolder, prefix, fusionConfig);

			// fusion.runFull(resultsFolder, "MAXDIFF_NEW_ADJUSTED",
			// rebuilder.build(), maxDiffsFolder, true);
			// fusion.runFull(resultsFolder, "MAXDIFF_ADJUSTED_RECALL",
			// rebuilder.build(), maxDiffsFolder, true);

			// rebuilder = fusionConfig.rebuilder();
			// rebuilder.setReSplitTestingSet(true);
			// fusion.runFull(resultsFolder, "MAXDIFF_RESPLIT",
			// rebuilder.build(), maxDiffsFolder, false);
			// // fusion.runFull(resultsFolder, "MAXDIFF_RESPLIT",
			// // rebuilder.build(), maxDiffsFolder, true);
			//
			// rebuilder = fusionConfig.rebuilder();
			// rebuilder.setAdjustedRecall(true);
			// rebuilder.setReSplitTestingSet(true);
			// fusion.runFull(resultsFolder, "MAXDIFF_ADJUSTED_RECALL_RESPLIT",
			// rebuilder.build(), maxDiffsFolder, false);
			// fusion.runFull(resultsFolder, "MAXDIFF_ADJUSTED_RECALL_RESPLIT",
			// rebuilder.build(), maxDiffsFolder, true);

			// fusion.runReplicateSimple(resultsFolder, prefix, fusionConfig);
			// fusion.runReplicate(resultsFolder, prefix, fusionConfig);
			// fusion.runSplit(resultsFolder, prefix, fusionConfig);
			// fusion.runFull(resultsFolder, prefix, fusionConfig);

			// fusion.runReplicate("_NEW_THRESHOLD", thresholdAdapter);

			try {
				PersistentParallel6.close();
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}

			// PersistentParallel6.restart();
			// SimpleLockableConcurrentCache.clearAllCaches();

			int sdf = 123;

		}
	}

	private static void printout(Map<String, TreeMap<String, List<String>>> rett) {
		for (Entry<String, TreeMap<String, List<String>>> entry : rett.entrySet()) {

			System.out.println(entry.getKey());
			StringWriter sw = new StringWriter();
			CSVWriter writer = new CSVWriter(sw);

			for (Entry<String, List<String>> innerEntry : entry.getValue().entrySet()) {
				String[] line = new String[innerEntry.getValue().size() + 1];
				line[0] = innerEntry.getKey();

				int i = 1;
				for (String s : innerEntry.getValue()) {

					line[i] = s;

					i++;
				}
				writer.writeNext(line);
			}

			try {
				writer.close();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}

			System.out.println(sw.toString());
		}

	}

	public Set<TableRowId> rowsWithLabelOnly(Set<TableRowId> rowIds) {
		return rowIds.stream().filter((e) -> {
			String labelRaw = this.tableRepository.getLabelRaw(e);
			if (labelRaw != null && !labelRaw.equals(""))
				return true;
			else
				return false;
		}).collect(Collectors.toSet());
	}

	private Map<String, TreeMap<String, List<String>>> runForWeakSupervision(GoldStandardClustersSetPerClass goldStandardClustersSet, FusionConfig fusionConfig,
			SetExpansionStrategiesConfig setExpansionStrategiesConfig, StrategiesConfig strategiesConfig, Map<ClassId, TableRowIdsSet> fullInput,
			String thresholdsFolder, Map<ClassId, Set<ClassId>> allParents, SetExtension setExtension, String instanceMatchingWeightsFolder,
			Map<ClassId, Map<InstanceId, Integer>> conflictingSets, String currentCaseFolder,
			Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>> fullCorrectMap, boolean alwaysDisableBlocking,
			Map<ClassId, Set<PropertyId>> classesAndProperties, Map<InstanceId, Set<TableRowId>> additionalExisting, int tableSize) throws IOException {
		SetExpansionStrategy METHOD_WHEN_TESTING_ST = SetExpansionStrategiesRepository.COMBINED_RandomForest_ET;// NEW;
		Map<String, TreeMap<String, List<String>>> optimizationRetMap = new HashMap<String, TreeMap<String, List<String>>>();
		int cpus = NumOfCpusGetter.getProcessors();
		boolean GS_RUN = false;
		boolean GS_RUN_ND = false;
		boolean run_HyperParameter = false;

		boolean RUN_CLUSTERING_FLAG = false;
		boolean RUN_CLUSTERING_NON_CLUSTERING = false;
		boolean RUN_CLUSTERING_NON_CLUSTERING_CORRECT = false;
		boolean RUN_CLUSTERING = false;
		boolean RUN_ND_MR = false;
		boolean RUN_CLUSTERING_MR = false;
		boolean RUN_CLUSTERING_NO_DS = false;
		boolean RUN_CLUSTERING_DS_ONLY = false;
		boolean RUN_CLUSTERING_DS_TOP_ONLY = false;
		boolean RUN_ND = false;
		boolean RUN_ND_with_clustering = false;
		boolean RUN_ND_iterative = false;
		boolean RUN_CLUSTERING_WITH_ND_WITH_CLUSTERING = false;
		boolean RUN_CLUSTERING_MR_ONLY = false;
		boolean RUN_ND_WITH_MR_CLUSTERING = false;

		// GS_RUN
		if (false) {
			GS_RUN = true;
			GS_RUN_ND = true;
			RUN_CLUSTERING_FLAG = true;
		}

		// CLUSTERING FIRST ITERATION
		if (false) {
			RUN_CLUSTERING_FLAG = true;
			RUN_CLUSTERING = true;
			RUN_CLUSTERING_MR = true;
			RUN_CLUSTERING_MR_ONLY = false;
			RUN_CLUSTERING_NO_DS = true;
		}

		// ND FIRST ITERATION
		if (false) {
			RUN_ND = true;
			RUN_ND_MR = true;
			RUN_CLUSTERING_MR = true;
			RUN_ND_WITH_MR_CLUSTERING = true;
		}

		// RUN END-to-END-NEW
		if (true) {
			RUN_CLUSTERING_FLAG = true;
			RUN_CLUSTERING = true;
			RUN_ND = true;
			RUN_ND_MR = true;
			RUN_CLUSTERING_DS_ONLY = false;
			RUN_ND_with_clustering = false;
			RUN_CLUSTERING_WITH_ND_WITH_CLUSTERING = false;
			RUN_CLUSTERING_NO_DS = true;
			RUN_ND_WITH_MR_CLUSTERING = true;
		}

		// RUN END-to-END
		if (false) {
			RUN_CLUSTERING_FLAG = true;
			RUN_CLUSTERING = true;
			RUN_CLUSTERING_DS_ONLY = true;
			RUN_ND_with_clustering = true;
			RUN_CLUSTERING_WITH_ND_WITH_CLUSTERING = true;
		}

		List<ValidationRun> validationRuns = new ArrayList<>();

		GSRun fusion = this;

		// setExpansionStrategiesConfig =
		// setExpansionStrategiesConfig.rebuilder().clearStrategies().addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE)
		// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RFDES_BASE).build();

		// setExpansionStrategiesConfig =
		// setExpansionStrategiesConfig.rebuilder().clearStrategies()
		// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest).addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest_NEW)
		// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE).build();

		// builder.addStrategy(SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY);
		setExpansionStrategiesConfig = setExpansionStrategiesConfig.rebuilder().clearStrategies()

				.addStrategy(SetExpansionStrategiesRepository.FUSION_TESTER).addStrategy(SetExpansionStrategiesRepository.GS_CLUSTERER)
				.addStrategy(SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY)
				// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE_Label_SameTable)
				// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE_Label_SameTable_Bow)
				// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE_Label_SameTable_Bow_Phi)
				// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE_Label_SameTable_Bow_Phi_Attribute)
				// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE_Label_SameTable_Bow_Phi_Attribute_Implicit)
				.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest_NEW).addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE).build();

		setExpansionStrategiesConfig = setExpansionStrategiesConfig.rebuilder().clearStrategies()
				.addStrategy(SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY)
				// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE_Label_SameTable_Bow)
				// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE_Label_SameTable_Bow_Phi)
				// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE_Label_SameTable_Bow_Phi_Attribute)
				// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE_Label_SameTable_Bow_Phi_Attribute_Implicit)
				.addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE).addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest_NEW)
				.addStrategy(SetExpansionStrategiesRepository.COMBINED_BASE).build();

		setExpansionStrategiesConfig = setExpansionStrategiesConfig.rebuilder().clearStrategies()
				.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest_NEW).build();

		//// setExpansionStrategiesConfig =
		//// setExpansionStrategiesConfig.rebuilder().clearStrategies()
		//// .addStrategy(SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY).addStrategy(SetExpansionStrategiesRepository.NEW_COMBINED)
		// //
		//// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE_Label_SameTable_Bow)
		// //
		//// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE_Label_SameTable_Bow_Phi)
		// //
		//// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE_Label_SameTable_Bow_Phi_Attribute)
		// //
		//// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE_Label_SameTable_Bow_Phi_Attribute_Implicit)
		// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE).addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest_NEW)
		// .addStrategy(SetExpansionStrategiesRepository.COMBINED_BASE).build();

		setExpansionStrategiesConfig = setExpansionStrategiesConfig.rebuilder().clearStrategies()

				.addStrategy(SetExpansionStrategiesRepository.FUSION_TESTER).addStrategy(SetExpansionStrategiesRepository.GS_CLUSTERER).build();

		setExpansionStrategiesConfig = setExpansionStrategiesConfig.rebuilder().clearStrategies()

				.addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE).addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest_NEW)
				.addStrategy(SetExpansionStrategiesRepository.COMBINED_BASE).build();

		setExpansionStrategiesConfig = setExpansionStrategiesConfig.rebuilder().clearStrategies().addStrategy(SetExpansionStrategiesRepository.FUSION_TESTER)
				.addStrategy(SetExpansionStrategiesRepository.GS_CLUSTERER).addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE).build();

		// setExpansionStrategiesConfig =
		// setExpansionStrategiesConfig.rebuilder().clearStrategies()
		// //
		// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest_ET).addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest_NEW)
		// // .addStrategy(SetExpansionStrategiesRepository.COMBINED_BASE)
		// .addStrategy(SetExpansionStrategiesRepository.COMBINED_Logistic_NEW).addStrategy(SetExpansionStrategiesRepository.COMBINED_Logistic_NEW_ET)
		// .build();

		setExpansionStrategiesConfig = setExpansionStrategiesConfig.rebuilder().clearStrategies()

				.addStrategy(SetExpansionStrategiesRepository.COMBINED_RandomForest_ET).build();
		// setExpansionStrategiesConfig =
		// setExpansionStrategiesConfig.rebuilder().clearStrategies().addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE)
		// .build();

		// setExpansionStrategiesConfig =
		// setExpansionStrategiesConfig.rebuilder().clearStrategies()
		// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE_Label_SameTable_Bow_Phi_Attribute_Implicit).build();

		List<EntityMatchingRun> runs = Arrays.asList(EntityMatchingRunRepository.CS_LABEL_EM_LABEL, EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE,
				EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE_BOW, EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE_INDEGREE_BOW,
				EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE_INDEGREE_BOW_ATTRIBUTE, EntityMatchingRunRepository.CS_LABEL_EM_ALL);

		runs = Arrays.asList(EntityMatchingRunRepository.LIN_CS_LABEL_EM_ALL, EntityMatchingRunRepository.LOG_CS_LABEL_EM_ALL,
				EntityMatchingRunRepository.CS_LABEL_EM_ALL, EntityMatchingRunRepository.RF_CS_LABEL_EM_ALL);

		runs = Arrays.asList(EntityMatchingRunRepository.CS_LABEL_EM_ALL, EntityMatchingRunRepository.RF_CS_LABEL_EM_LABEL,
				EntityMatchingRunRepository.RF_CS_LABEL_EM_LABEL_TYPE, EntityMatchingRunRepository.RF_CS_LABEL_EM_LABEL_TYPE_BOW,
				EntityMatchingRunRepository.RF_CS_LABEL_EM_LABEL_TYPE_BOW_INDEGREE, EntityMatchingRunRepository.RF_CS_LABEL_EM_LABEL_TYPE_BOW_INDEGREE_ATT,
				EntityMatchingRunRepository.RF_CS_LABEL_EM_ALL);

		runs = Arrays.asList(EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE_INDEGREE_BOW_ATTRIBUTE, EntityMatchingRunRepository.CS_LABEL_EM_ALL,
				EntityMatchingRunRepository.RF_CS_LABEL_EM_LABEL_TYPE_BOW_INDEGREE_ATT, EntityMatchingRunRepository.RF_CS_LABEL_EM_ALL);

		runs = Arrays.asList(EntityMatchingRunRepository.CS_LABEL_EM_ALL, EntityMatchingRunRepository.SW_CS_LABEL_EM_ALL,
				EntityMatchingRunRepository.RF_CS_LABEL_EM_ALL, EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL);

		runs = Arrays.asList(EntityMatchingRunRepository.CS_LABEL_EM_ALL, EntityMatchingRunRepository.RF_CS_LABEL_EM_ALL,
				EntityMatchingRunRepository.SW_CS_LABEL_EM_ALL, EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL);

		runs = Arrays.asList(EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL, EntityMatchingRunRepository.RFDES_SW_CS_LABEL_EM_ALL);

		runs = Arrays.asList(EntityMatchingRunRepository.CS_LABEL_EM_ALL, EntityMatchingRunRepository.RF_CS_LABEL_EM_ALL,
				EntityMatchingRunRepository.RF_NEW_CS_LABEL_EM_ALL);

		runs = Arrays.asList(EntityMatchingRunRepository.CS_LABEL_EM_ALL, EntityMatchingRunRepository.SW_CS_LABEL_EM_ALL,
				EntityMatchingRunRepository.RF_NEW_CS_LABEL_EM_ALL, EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL);

		runs = Arrays.asList(EntityMatchingRunRepository.CS_LABEL_EM_LABEL, EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE,
				EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE_BOW, EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE_INDEGREE_BOW_ATTRIBUTE,
				EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL, EntityMatchingRunRepository.RF_NEW_CS_LABEL_EM_ALL,
				EntityMatchingRunRepository.SW_CS_LABEL_EM_ALL);

		runs = Arrays.asList(EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL, EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE_INDEGREE_BOW_ATTRIBUTE,
				EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE_INDEGREE_BOW_ATTRIBUTE_IMPLICIT);

		runs = Arrays.asList(EntityMatchingRunRepository.CS_LABEL_EM_LABEL, EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE,
				EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE_BOW, EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE_INDEGREE_BOW_ATTRIBUTE,
				EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE_INDEGREE_BOW_ATTRIBUTE_IMPLICIT, EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL,
				EntityMatchingRunRepository.RF_NEW_CS_LABEL_EM_ALL, EntityMatchingRunRepository.SW_CS_LABEL_EM_ALL);

		runs = Arrays.asList(EntityMatchingRunRepository.RF_ET_CS_LABEL_EM_ALL);

		Map<String, SetExpansionStrategiesConfig> fusionStrategies = new HashMap<>();

		fusionStrategies.put("ColumnScores", setExpansionStrategiesConfig.rebuilder().clearFusionSetups()
				.addFusionSetup(FusionStrategyRepository.COLUMN_SCORES, PropertiesConfig.builder().setTimeDependence(TimeDependence.NO).build()).build());

		if (false) {
			fusionStrategies.put("VOTING", setExpansionStrategiesConfig.rebuilder().clearFusionSetups()
					.addFusionSetup(FusionStrategyRepository.NORMAL, PropertiesConfig.builder().setTimeDependence(TimeDependence.NO).build()).build());

			fusionStrategies.put("KBT",
					setExpansionStrategiesConfig.rebuilder().clearFusionSetups()
							.addFusionSetup(FusionStrategyRepository.KBT_NO_THRESHOLD, PropertiesConfig.builder().setTimeDependence(TimeDependence.NO).build())
							.build());

		}

		Map<ClassId, TableIdsSet> tableIds = new HashMap<>();
		Map<ClassId, TableRowIdsSet> fullRowIdsSet = new HashMap<>();
		Map<ClassId, Set<TableRowId>> wsRowIdsSet = new HashMap<>();
		Map<ClassId, Set<TableRowId>> wsRowIdsSetSecond = new HashMap<>();

		for (ClassId classId : allParents.keySet()) {

			Random rnd = new Random(classId.getFinalPart().hashCode());
			List<AnnotatedTableId> list = new ArrayList<>(tableRepository.getAllTableIds(classId));
			Collections.sort(list, (e1, e2) -> {
				return e1.getTableName().compareTo(e2.getTableName());
			});

			int fiftenPercent = (int) (0.15 * tableSize);
			fiftenPercent = (RUN_ND ? fiftenPercent : 0);
			fiftenPercent = 0;
			Collections.shuffle(list, rnd);
			Set<AnnotatedTableId> tableIdsSetRaw = new HashSet<>(list.subList(0, tableSize + fiftenPercent));

			Set<TableRowId> wsRowIds = rowsWithLabelOnly(this.tableRepository.getAllRowIds(tableIdsSetRaw));

			Set<AnnotatedTableId> tableIdsSetRawSecond = new HashSet<>(list.subList(tableSize + fiftenPercent, tableSize + fiftenPercent + tableSize));
			Set<TableRowId> wsRowIdsSecond = rowsWithLabelOnly(this.tableRepository.getAllRowIds(tableIdsSetRawSecond));

			tableIdsSetRaw.addAll(goldStandardClustersSet.getPerClass().get(classId).getFull().getClusters().getCandidates().stream().map((e) -> e.getTableId())
					.collect(Collectors.toSet()));
			tableIdsSetRaw.addAll(tableIdsSetRawSecond);

			TableIdsSet tableIdsSet = new TableIdsSet(tableIdsSetRaw, new UuidStringId("DS_" + classId.getFinalPart()));
			tableIds.put(classId, tableIdsSet);

			Set<TableRowId> fullRowIds = this.tableRepository.getAllRowIds(tableIdsSet.getTableIds());

			fullRowIdsSet.put(classId, new TableRowIdsSet(fullRowIds, new UuidStringId("DS_" + classId.getFinalPart())));
			wsRowIdsSet.put(classId, wsRowIds);
			wsRowIdsSetSecond.put(classId, wsRowIdsSecond);
		}

		String defaultNDModel = "AverageLow";

		Map<String, List<EntityMatchingRun>> ndStrategies = new HashMap<>();

		Map<String, List<SetExpansionStrategy>> clusteringStrategies = new HashMap<>();

		if (GS_RUN) {
			clusteringStrategies.put("GS_RUN", setExpansionStrategiesConfig.getStrategies());
		}

		if (GS_RUN_ND) {
			ndStrategies.put("GS_RUN", runs);
		}

		DistantSupervisionInput distantSupervisionInput = DistantSupervisionInput.builder().setClassesWithParents(allParents)
				.setSetExpansionStrategiesConfig(setExpansionStrategiesConfig).setEntityMatchingRuns(runs).setTableIdsPerClass(tableIds)
				.setNonTailProperties(setExtension.getNonTailProperties()).setFusionConfig(fusionConfig).setThresholdsFolder(thresholdsFolder)
				.setStrategiesConfig(strategiesConfig).setFullTableRowIdSet(fullRowIdsSet).setWsRowIds(wsRowIdsSet)
				.setGoldStandardClustersSet(goldStandardClustersSet).build();

		ClusteringRunner clusteringRunner = DS_ClusteringRunnerCreator.getRunner(distantSupervisionInput);

		DistantSupervisionInput distantSupervisionInputSecond = DistantSupervisionInput.builder().setClassesWithParents(allParents)
				.setSetExpansionStrategiesConfig(setExpansionStrategiesConfig).setEntityMatchingRuns(runs).setTableIdsPerClass(tableIds)
				.setNonTailProperties(setExtension.getNonTailProperties()).setFusionConfig(fusionConfig).setThresholdsFolder(thresholdsFolder)
				.setStrategiesConfig(strategiesConfig).setFullTableRowIdSet(fullRowIdsSet).setWsRowIds(wsRowIdsSetSecond)
				.setGoldStandardClustersSet(goldStandardClustersSet).build();

		ClusteringRunner clusteringRunnerSecond = DS_ClusteringRunnerCreator.getRunner(distantSupervisionInputSecond);

		// Map<String, DS_Container> gsStrategies =
		// genericStrategyCreator.get().get(distantSupervisionInput,
		// clusteringRunner);

		// Map<String, DS_Container> bsStreategies = bootStrapCreator.get(0.5,
		// 0.5).get(distantSupervisionInput, clusteringRunner);

		Map<String, Map<ClassId, LearnedModelRepository>> dsClusteringStrategies = new LinkedHashMap<>();
		ND_Average_Strategy zerofiveCreator = rcAverageCreator.getStrategy(0.50);
		Map<ClassId, LearnedModelRepository> zeroFiveModels = zerofiveCreator.get(distantSupervisionInput);
		dsClusteringStrategies.put("AverageLow", zeroFiveModels);
		// dsClusteringStrategies.put("AverageHigh",
		// rcAverageCreator.getStrategy(0.95).get(distantSupervisionInput));

		clusteringStrategies.put("AverageLow", zerofiveCreator.getStrategies());

		Map<String, Map<ClassId, EM_LearnedModelRepository>> dsNdStrategies = new LinkedHashMap<>();
		expansion.set.distant_supervision.baseline.average.ND_Average_Creator.ND_Average_Strategy zerfiveNDStrategy = ndAverageCreator.getStrategy(0.50);
		Map<ClassId, EM_LearnedModelRepository> zeroFiveNDMOdels = zerfiveNDStrategy.get(distantSupervisionInput);
		dsNdStrategies.put("AverageLow", zeroFiveNDMOdels);
		ndStrategies.put("AverageLow", zerfiveNDStrategy.getStrategies());

		// dsNdStrategies.put("AverageHigh",
		// ndAverageCreator.getStrategy(0.95).get(distantSupervisionInput));
		//
		// for (Entry<String, DS_Container> entry : bsStreategies.entrySet()) {
		// dsClusteringStrategies.put(entry.getKey(),
		// entry.getValue().getRowClustering());
		// dsNdStrategies.put(entry.getKey(),
		// entry.getValue().getNewDetection());
		// }

		// for (Entry<String, DS_Container> entry : gsStrategies.entrySet()) {
		// dsClusteringStrategies.put(entry.getKey(),
		// entry.getValue().getRowClustering());
		// clusteringStrategies.put(entry.getKey(),
		// entry.getValue().getClusteringStrategies());
		// if (entry.getValue().hasNewDetection()) {
		// dsNdStrategies.put(entry.getKey(),
		// entry.getValue().getNewDetection());
		// }
		// }

		Map<ClassId, LearnedModelRepository> RC_AL_MR_WS_DS = null;

		Map<ClassId, LearnedModelRepository> RC_AL_WS_DS = null;

		if (RUN_CLUSTERING_NON_CLUSTERING) {

			for (int i = 0; i < 2; i++) {
				boolean negativeRandom = false;
				if (i == 0) {
					negativeRandom = true;
				}
				if (false) {
					Map<String, DS_Container> wsStrategies = WS_RC_No_Clustering_Strategy_Creator.get("AL+NC+Weighted" + (negativeRandom ? "+NR" : ""),
							SetExpansionStrategiesRepository.COMBINED_RF_BASE, zeroFiveModels, WS_METHOD.WEIGHTED, negativeRandom)
							.get(distantSupervisionInput, clusteringRunner);
					for (Entry<String, DS_Container> entry : wsStrategies.entrySet()) {
						dsClusteringStrategies.put(entry.getKey(), entry.getValue().getRowClustering());
						clusteringStrategies.put(entry.getKey(), entry.getValue().getClusteringStrategies());
						if (entry.getValue().hasNewDetection()) {
							dsNdStrategies.put(entry.getKey(), entry.getValue().getNewDetection());
						}

						if (false) {
							wsStrategies = WS_RC_No_Clustering_Strategy_Creator
									.get("AL+WSNC+2", SetExpansionStrategiesRepository.COMBINED_RandomForest_ET, entry.getValue().getRowClustering())
									.get(distantSupervisionInput, clusteringRunner);

							for (Entry<String, DS_Container> entry2 : wsStrategies.entrySet()) {
								dsClusteringStrategies.put(entry2.getKey(), entry2.getValue().getRowClustering());
								clusteringStrategies.put(entry2.getKey(), entry2.getValue().getClusteringStrategies());
								if (entry2.getValue().hasNewDetection()) {
									dsNdStrategies.put(entry2.getKey(), entry2.getValue().getNewDetection());
								}
							}
						}
					}
				}

				{
					Map<String, DS_Container> wsStrategies = WS_RC_No_Clustering_Strategy_Creator.get("AL+NC+AlwaysOne" + (negativeRandom ? "+NR" : ""),
							SetExpansionStrategiesRepository.COMBINED_RF_BASE, zeroFiveModels, WS_METHOD.ALWAYS_ONE, negativeRandom)
							.get(distantSupervisionInput, clusteringRunner);
					for (Entry<String, DS_Container> entry : wsStrategies.entrySet()) {
						dsClusteringStrategies.put(entry.getKey(), entry.getValue().getRowClustering());
						clusteringStrategies.put(entry.getKey(), entry.getValue().getClusteringStrategies());
						if (entry.getValue().hasNewDetection()) {
							dsNdStrategies.put(entry.getKey(), entry.getValue().getNewDetection());
						}

					}
				}

				if (false) {

					Map<String, DS_Container> wsStrategies = WS_RC_No_Clustering_Strategy_Creator.get("AL+NC+Rounded" + (negativeRandom ? "+NR" : ""),
							SetExpansionStrategiesRepository.COMBINED_RF_BASE, zeroFiveModels, WS_METHOD.ROUNDED, negativeRandom)
							.get(distantSupervisionInput, clusteringRunner);
					for (Entry<String, DS_Container> entry : wsStrategies.entrySet()) {
						dsClusteringStrategies.put(entry.getKey(), entry.getValue().getRowClustering());
						clusteringStrategies.put(entry.getKey(), entry.getValue().getClusteringStrategies());
						if (entry.getValue().hasNewDetection()) {
							dsNdStrategies.put(entry.getKey(), entry.getValue().getNewDetection());
						}

					}
				}

			}

		}

		Map<ClassId, LearnedModelRepository> MR_MODEL_WITHOUT_WS = null;
		Map<ClassId, LearnedModelRepository> mr_WS_Models = null;
		Map<ClassId, LearnedModelRepository> al_WS_Models = null;
		if (RUN_CLUSTERING || RUN_CLUSTERING_MR) {
			Map<String, DS_Container> wsStrategies = RC_MatchingRules_Creator.get("MR", SetExpansionStrategiesRepository.COMBINED_RF_BASE, zeroFiveModels)
					.get(distantSupervisionInput, clusteringRunner);
			for (Entry<String, DS_Container> entry : wsStrategies.entrySet()) {
				dsClusteringStrategies.put(entry.getKey(), entry.getValue().getRowClustering());
				clusteringStrategies.put(entry.getKey(), entry.getValue().getClusteringStrategies());
				if ((RUN_CLUSTERING && (!RUN_CLUSTERING_DS_ONLY || !RUN_CLUSTERING_DS_TOP_ONLY)) || RUN_CLUSTERING_MR) {
					Map<String, DS_Container> ws2Strategies = WS_RC_Strategy_Creator.get("MR+WS", SetExpansionStrategiesRepository.Container_Clusterer,
							entry.getValue().getRowClustering(), SetExpansionStrategiesRepository.COMBINED_RandomForest_ET)
							.get(distantSupervisionInput, clusteringRunner);

					for (Entry<String, DS_Container> entry2 : ws2Strategies.entrySet()) {
						dsClusteringStrategies.put(entry2.getKey(), entry2.getValue().getRowClustering());
						clusteringStrategies.put(entry2.getKey(), entry2.getValue().getClusteringStrategies());
						mr_WS_Models = entry2.getValue().getRowClustering();
						// if (entry.getValue().hasNewDetection()) {
						// dsNdStrategies.put(entry.getKey(),
						// entry.getValue().getNewDetection());
						// }

					}
				}

				MR_MODEL_WITHOUT_WS = entry.getValue().getRowClustering();
			}
		}

		Map<String, DS_Container> mrStrategies = ND_MatchingRules_Creator.get("ND_AL+MR", EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL, zeroFiveNDMOdels)
				.get(distantSupervisionInput, clusteringRunner);

		if (RUN_CLUSTERING) {

			if (!RUN_CLUSTERING_DS_ONLY && !RUN_CLUSTERING_MR_ONLY) {
				{
					Map<String, DS_Container> wsStrategies = WS_RC_Strategy_Creator.get("AL+WS", SetExpansionStrategiesRepository.COMBINED_RF_BASE,
							zeroFiveModels, SetExpansionStrategiesRepository.COMBINED_RandomForest_ET).get(distantSupervisionInput, clusteringRunner);

					for (Entry<String, DS_Container> entry : wsStrategies.entrySet()) {
						al_WS_Models = entry.getValue().getRowClustering();

						dsClusteringStrategies.put(entry.getKey(), entry.getValue().getRowClustering());
						clusteringStrategies.put(entry.getKey(), entry.getValue().getClusteringStrategies());
						if (entry.getValue().hasNewDetection()) {
							dsNdStrategies.put(entry.getKey(), entry.getValue().getNewDetection());
						}

					}
				}

				if (false) {
					Map<String, DS_Container> wsStrategies = WS_RC_Strategy_Creator.get("AL+WS_ST", SetExpansionStrategiesRepository.COMBINED_RF_BASE,
							zeroFiveModels, SetExpansionStrategiesRepository.COMBINED_RandomForest_NEW).get(distantSupervisionInput, clusteringRunner);

					for (Entry<String, DS_Container> entry : wsStrategies.entrySet()) {
						dsClusteringStrategies.put(entry.getKey(), entry.getValue().getRowClustering());
						clusteringStrategies.put(entry.getKey(), entry.getValue().getClusteringStrategies());
						if (entry.getValue().hasNewDetection()) {
							dsNdStrategies.put(entry.getKey(), entry.getValue().getNewDetection());
						}

					}
				}
			}

			if (true && !RUN_CLUSTERING_DS_TOP_ONLY && !RUN_CLUSTERING_MR_ONLY && !RUN_CLUSTERING_NO_DS) {

				Map<String, DS_Container> ws2Strategies = WS_RC_Strategy_Creator
						.getWithDistantSupervision("WS+DS", SetExpansionStrategiesRepository.COMBINED_RF_BASE, zeroFiveModels,
								SetExpansionStrategiesRepository.COMBINED_RandomForest_ET, EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL, zeroFiveNDMOdels)
						.get(distantSupervisionInput, clusteringRunner);

				for (Entry<String, DS_Container> entry2 : ws2Strategies.entrySet()) {
					dsClusteringStrategies.put(entry2.getKey(), entry2.getValue().getRowClustering());
					clusteringStrategies.put(entry2.getKey(), entry2.getValue().getClusteringStrategies());
					RC_AL_WS_DS = entry2.getValue().getRowClustering();
					// mr_WS_Models = entry2.getValue().getRowClustering();
					// if (entry.getValue().hasNewDetection()) {
					// dsNdStrategies.put(entry.getKey(),
					// entry.getValue().getNewDetection());
					// }

				}

			}

			if (true && !RUN_CLUSTERING_NO_DS) {

				if (true && MR_MODEL_WITHOUT_WS != null) {

					for (Entry<String, DS_Container> entry : mrStrategies.entrySet()) {

						Map<String, DS_Container> ws2Strategies = WS_RC_Strategy_Creator
								.getWithDistantSupervision("MR+WS+DS", SetExpansionStrategiesRepository.Container_Clusterer, MR_MODEL_WITHOUT_WS,
										SetExpansionStrategiesRepository.COMBINED_RandomForest_ET, EntityMatchingRunRepository.CONTAINER_ND_ALL,
										entry.getValue().getNewDetection())
								.get(distantSupervisionInput, clusteringRunner);

						for (Entry<String, DS_Container> entry2 : ws2Strategies.entrySet()) {
							dsClusteringStrategies.put(entry2.getKey(), entry2.getValue().getRowClustering());
							clusteringStrategies.put(entry2.getKey(), entry2.getValue().getClusteringStrategies());
							RC_AL_MR_WS_DS = entry2.getValue().getRowClustering();
							// if (entry.getValue().hasNewDetection()) {
							// dsNdStrategies.put(entry.getKey(),
							// entry.getValue().getNewDetection());
							// }

						}

					}

				}

				// if (entry.getValue().hasNewDetection()) {
				// dsNdStrategies.put(entry.getKey(),
				// entry.getValue().getNewDetection());
				// }

			}

			if (false) {
				Map<String, DS_Container> wsStrategies = WS_RC_Strategy_Creator.get("AL+WS", SetExpansionStrategiesRepository.COMBINED_RF_BASE, zeroFiveModels,
						SetExpansionStrategiesRepository.COMBINED_RandomForest_ET).get(distantSupervisionInput, clusteringRunner);

				for (Entry<String, DS_Container> entry : wsStrategies.entrySet()) {
					dsClusteringStrategies.put(entry.getKey(), entry.getValue().getRowClustering());
					clusteringStrategies.put(entry.getKey(), entry.getValue().getClusteringStrategies());
					if (entry.getValue().hasNewDetection()) {
						dsNdStrategies.put(entry.getKey(), entry.getValue().getNewDetection());
					}

					if (true) {
						wsStrategies = WS_RC_Strategy_Creator.get("AL+WS+2", SetExpansionStrategiesRepository.COMBINED_RandomForest_ET,
								entry.getValue().getRowClustering(), SetExpansionStrategiesRepository.COMBINED_RandomForest_ET)
								.get(distantSupervisionInputSecond, clusteringRunnerSecond);

						for (Entry<String, DS_Container> entry2 : wsStrategies.entrySet()) {
							dsClusteringStrategies.put(entry2.getKey(), entry2.getValue().getRowClustering());
							clusteringStrategies.put(entry2.getKey(), entry2.getValue().getClusteringStrategies());
							if (entry2.getValue().hasNewDetection()) {
								dsNdStrategies.put(entry2.getKey(), entry2.getValue().getNewDetection());
							}
						}
					}
				}
			}

		}

		if (RUN_CLUSTERING_NON_CLUSTERING_CORRECT) {

			boolean negativeRandom = true;
			if (true) {
				{
					Map<String, DS_Container> wsStrategies = WS_RC_No_Clustering_Strategy_Creator.get("AL+NC+Rounded" + (negativeRandom ? "+NR" : ""),
							SetExpansionStrategiesRepository.COMBINED_RF_BASE, zeroFiveModels, WS_METHOD.ROUNDED, negativeRandom)
							.get(distantSupervisionInput, clusteringRunner);
					for (Entry<String, DS_Container> entry : wsStrategies.entrySet()) {
						dsClusteringStrategies.put(entry.getKey(), entry.getValue().getRowClustering());
						clusteringStrategies.put(entry.getKey(), entry.getValue().getClusteringStrategies());
						if (entry.getValue().hasNewDetection()) {
							dsNdStrategies.put(entry.getKey(), entry.getValue().getNewDetection());
						}

					}
				}

				Map<String, DS_Container> wsStrategies = WS_RC_No_Clustering_Strategy_Creator.get("AL+MR+NC+Rounded" + (negativeRandom ? "+NR" : ""),
						SetExpansionStrategiesRepository.Container_Clusterer, MR_MODEL_WITHOUT_WS, WS_METHOD.ROUNDED, negativeRandom)
						.get(distantSupervisionInput, clusteringRunner);
				for (Entry<String, DS_Container> entry : wsStrategies.entrySet()) {
					dsClusteringStrategies.put(entry.getKey(), entry.getValue().getRowClustering());
					clusteringStrategies.put(entry.getKey(), entry.getValue().getClusteringStrategies());
					if (entry.getValue().hasNewDetection()) {
						dsNdStrategies.put(entry.getKey(), entry.getValue().getNewDetection());
					}

				}
			}

		}

		if (RUN_ND_MR) {
			for (Entry<String, DS_Container> entry : mrStrategies.entrySet()) {
				ndStrategies.put(entry.getKey(), entry.getValue().getNewDetectionRuns());
				dsNdStrategies.put(entry.getKey(), entry.getValue().getNewDetection());
			}
		}
		if (RUN_ND) {

			if (true) {

				// .get(distantSupervisionInputSecond, clusteringRunnerSecond);

				for (Entry<String, DS_Container> entry : mrStrategies.entrySet()) {
					// ndStrategies.put(entry.getKey(),
					// entry.getValue().getNewDetectionRuns());
					// dsNdStrategies.put(entry.getKey(),
					// entry.getValue().getNewDetection());

					if (true) {
						Map<String, DS_Container> wsStrategies = WS_ND_Strategy_Creator.get("ND_AL_MR+WS", EntityMatchingRunRepository.CONTAINER_ND_ALL,
								entry.getValue().getNewDetection(), SetExpansionStrategiesRepository.COMBINED_RF_BASE, zeroFiveModels)
								.get(distantSupervisionInput, clusteringRunner);
						// .get(distantSupervisionInputSecond,
						// clusteringRunnerSecond);

						for (Entry<String, DS_Container> innerEntry : wsStrategies.entrySet()) {
							ndStrategies.put(innerEntry.getKey(), innerEntry.getValue().getNewDetectionRuns());
							dsNdStrategies.put(innerEntry.getKey(), innerEntry.getValue().getNewDetection());

							// if (true) {
							// wsStrategies =
							// WS_RC_Strategy_Creator.get("AL+WS+2",
							// SetExpansionStrategiesRepository.COMBINED_RandomForest_ET,
							// entry.getValue().getRowClustering(),
							// SetExpansionStrategiesRepository.COMBINED_RandomForest_ET)
							// .get(distantSupervisionInputSecond,
							// clusteringRunnerSecond);
							//
							// for (Entry<String, DS_Container> entry2 :
							// wsStrategies.entrySet()) {
							// ndStrategies.putAll(entry2.getKey(),
							// entry2.getValue().getNewDetectionRuns());
							// dsNdStrategies.put(entry2.getKey(),
							// entry2.getValue().getNewDetection());
							//
							// }
							// }
						}
					}

					if (RUN_ND_WITH_MR_CLUSTERING) {
						Map<String, DS_Container> wsStrategies = WS_ND_Strategy_Creator.get("ND_MR+WS_MR+WS", EntityMatchingRunRepository.CONTAINER_ND_ALL,
								entry.getValue().getNewDetection(), METHOD_WHEN_TESTING_ST, mr_WS_Models).get(distantSupervisionInput, clusteringRunner);
						// .get(distantSupervisionInputSecond,
						// clusteringRunnerSecond);

						for (Entry<String, DS_Container> innerEntry : wsStrategies.entrySet()) {
							ndStrategies.put(innerEntry.getKey(), innerEntry.getValue().getNewDetectionRuns());
							dsNdStrategies.put(innerEntry.getKey(), innerEntry.getValue().getNewDetection());
						}
					}

				}
			}

			if (true) {
				{
					Map<String, DS_Container> wsStrategies = WS_ND_Strategy_Creator.get("ND_AL_AL+WS", EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL,
							zeroFiveNDMOdels, SetExpansionStrategiesRepository.COMBINED_RF_BASE, zeroFiveModels).get(distantSupervisionInput, clusteringRunner);

					for (Entry<String, DS_Container> entry : wsStrategies.entrySet()) {
						ndStrategies.put(entry.getKey(), entry.getValue().getNewDetectionRuns());
						dsNdStrategies.put(entry.getKey(), entry.getValue().getNewDetection());

						// if (true) {
						// wsStrategies = WS_RC_Strategy_Creator.get("AL+WS+2",
						// SetExpansionStrategiesRepository.COMBINED_RandomForest_ET,
						// entry.getValue().getRowClustering(),
						// SetExpansionStrategiesRepository.COMBINED_RandomForest_ET)
						// .get(distantSupervisionInputSecond,
						// clusteringRunnerSecond);
						//
						// for (Entry<String, DS_Container> entry2 :
						// wsStrategies.entrySet()) {
						// ndStrategies.putAll(entry2.getKey(),
						// entry2.getValue().getNewDetectionRuns());
						// dsNdStrategies.put(entry2.getKey(),
						// entry2.getValue().getNewDetection());
						//
						// }
						// }
					}
				}
				Map<String, DS_Container> wsStrategies = WS_ND_Strategy_Creator
						.get("ND_AL+WS_AL+WS", EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL, zeroFiveNDMOdels, METHOD_WHEN_TESTING_ST, al_WS_Models)
						.get(distantSupervisionInput, clusteringRunner);

				for (Entry<String, DS_Container> entry : wsStrategies.entrySet()) {
					ndStrategies.put(entry.getKey(), entry.getValue().getNewDetectionRuns());
					dsNdStrategies.put(entry.getKey(), entry.getValue().getNewDetection());
				}
			}

		}

		// Map<ClassId, EM_LearnedModelRepository> ND_WITH_RCMRDS_MR_AS = null;

		if (RUN_ND_with_clustering) {

			Map<ClassId, EM_LearnedModelRepository> BESTC_BESTND_WITHOUT_MR = null;

			if (RC_AL_WS_DS != null) {
				Map<String, DS_Container> wsStrategies = WS_ND_Strategy_Creator.get("BESTC_AL+WS", EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL,
						zeroFiveNDMOdels, SetExpansionStrategiesRepository.COMBINED_RandomForest_ET, RC_AL_WS_DS)
						.get(distantSupervisionInput, clusteringRunner);

				for (Entry<String, DS_Container> entry : wsStrategies.entrySet()) {
					BESTC_BESTND_WITHOUT_MR = entry.getValue().getNewDetection();
					ndStrategies.put(entry.getKey(), entry.getValue().getNewDetectionRuns());
					dsNdStrategies.put(entry.getKey(), entry.getValue().getNewDetection());

					if (RUN_ND_iterative) {
						Map<String, DS_Container> wsStrategies2 = WS_ND_Strategy_Creator
								.get("BESTC_AL+WS+ITEARTIVE", EntityMatchingRunRepository.RF_ET_CS_LABEL_EM_ALL, entry.getValue().getNewDetection(),
										SetExpansionStrategiesRepository.COMBINED_RandomForest_ET, RC_AL_WS_DS)
								.get(distantSupervisionInputSecond, clusteringRunnerSecond);

						for (Entry<String, DS_Container> entry2 : wsStrategies2.entrySet()) {
							ndStrategies.put(entry2.getKey(), entry2.getValue().getNewDetectionRuns());
							dsNdStrategies.put(entry2.getKey(), entry2.getValue().getNewDetection());

						}
					}

				}

			}

			Map<ClassId, EM_LearnedModelRepository> BESTC_BESTND = null;

			if (RC_AL_MR_WS_DS != null) {

				for (Entry<String, DS_Container> entry : mrStrategies.entrySet()) {

					Map<String, DS_Container> wsStrategies = WS_ND_Strategy_Creator.get("BESTC_AL+MR+WS", EntityMatchingRunRepository.CONTAINER_ND_ALL,
							entry.getValue().getNewDetection(), SetExpansionStrategiesRepository.COMBINED_RandomForest_ET, RC_AL_MR_WS_DS)
							.get(distantSupervisionInput, clusteringRunner);
					// .get(distantSupervisionInputSecond,
					// clusteringRunnerSecond);

					for (Entry<String, DS_Container> innerEntry : wsStrategies.entrySet()) {
						BESTC_BESTND = innerEntry.getValue().getNewDetection();
						ndStrategies.put(innerEntry.getKey(), innerEntry.getValue().getNewDetectionRuns());
						dsNdStrategies.put(innerEntry.getKey(), innerEntry.getValue().getNewDetection());
						// ND_WITH_RCMRDS_MR_AS.putAll(innerEntry.getValue().getNewDetection());

						if (RUN_ND_iterative) {
							Map<String, DS_Container> wsStrategies2 = WS_ND_Strategy_Creator
									.get("BESTC_AL+MR+WS+ITEARTIVE", EntityMatchingRunRepository.RF_ET_CS_LABEL_EM_ALL, innerEntry.getValue().getNewDetection(),
											SetExpansionStrategiesRepository.COMBINED_RandomForest_ET, RC_AL_MR_WS_DS)
									.get(distantSupervisionInputSecond, clusteringRunnerSecond);

							for (Entry<String, DS_Container> entry2 : wsStrategies2.entrySet()) {
								ndStrategies.put(entry2.getKey(), entry2.getValue().getNewDetectionRuns());
								dsNdStrategies.put(entry2.getKey(), entry2.getValue().getNewDetection());

							}
						}

						// if (true) {
						// wsStrategies = WS_RC_Strategy_Creator.get("AL+WS+2",
						// SetExpansionStrategiesRepository.COMBINED_RandomForest_ET,
						// entry.getValue().getRowClustering(),
						// SetExpansionStrategiesRepository.COMBINED_RandomForest_ET)
						// .get(distantSupervisionInputSecond,
						// clusteringRunnerSecond);
						//
						// for (Entry<String, DS_Container> entry2 :
						// wsStrategies.entrySet()) {
						// ndStrategies.putAll(entry2.getKey(),
						// entry2.getValue().getNewDetectionRuns());
						// dsNdStrategies.put(entry2.getKey(),
						// entry2.getValue().getNewDetection());
						//
						// }
						// }
					}

				}
			}

			if (RUN_CLUSTERING_WITH_ND_WITH_CLUSTERING && MR_MODEL_WITHOUT_WS != null) {

				if (BESTC_BESTND_WITHOUT_MR != null) {

					Map<String, DS_Container> ws2Strategies = WS_RC_Strategy_Creator.getWithDistantSupervision("BESTC+BESTND+WS+DS",
							SetExpansionStrategiesRepository.COMBINED_RF_BASE, zeroFiveModels, SetExpansionStrategiesRepository.COMBINED_RandomForest_ET,
							EntityMatchingRunRepository.RF_ET_CS_LABEL_EM_ALL, BESTC_BESTND_WITHOUT_MR).get(distantSupervisionInput, clusteringRunner);

					for (Entry<String, DS_Container> entry2 : ws2Strategies.entrySet()) {
						dsClusteringStrategies.put(entry2.getKey(), entry2.getValue().getRowClustering());
						clusteringStrategies.put(entry2.getKey(), entry2.getValue().getClusteringStrategies());
						// mr_WS_Models = entry2.getValue().getRowClustering();
						// if (entry.getValue().hasNewDetection()) {
						// dsNdStrategies.put(entry.getKey(),
						// entry.getValue().getNewDetection());
						// }

					}

				}

				Map<String, DS_Container> ws2Strategies = WS_RC_Strategy_Creator
						.getWithDistantSupervision("BESTC+BESTND+MR+WS+DS", SetExpansionStrategiesRepository.Container_Clusterer, MR_MODEL_WITHOUT_WS,
								SetExpansionStrategiesRepository.COMBINED_RandomForest_ET, EntityMatchingRunRepository.RF_ET_CS_LABEL_EM_ALL, BESTC_BESTND)
						.get(distantSupervisionInput, clusteringRunner);

				for (Entry<String, DS_Container> entry2 : ws2Strategies.entrySet()) {
					dsClusteringStrategies.put(entry2.getKey(), entry2.getValue().getRowClustering());
					clusteringStrategies.put(entry2.getKey(), entry2.getValue().getClusteringStrategies());
					RC_AL_MR_WS_DS = entry2.getValue().getRowClustering();
					// if (entry.getValue().hasNewDetection()) {
					// dsNdStrategies.put(entry.getKey(),
					// entry.getValue().getNewDetection());
					// }

				}
			}
		}

		Map<Integer, Map<ClassId, Pair<EM_LearnedModelRepository, Map<EntityMatchingRun, NewDetectionPerformance>>>> entityWeightsMap = new HashMap<>();
		Map<Integer, Map<ClassId, LearnedModelRepository>> learnedModelsMap = new HashMap<>();

		Set<TableRowId> rowIdsAllClasses = new HashSet<>();
		for (ClassId classId : allParents.keySet()) {
			Set<TableRowId> rowIds = fullInput.get(classId).getRowIds();
			rowIdsAllClasses.addAll(rowIds);
		}
		TableRowIdsSet rowIdsSetAllClasses = new TableRowIdsSet(rowIdsAllClasses, new UuidStringId("ALL"));

		// BESTC+BESTND+WS+DS
		// BESTC_AL+WS
		//
		// BESTC+BESTND+MR+WS+DS
		// BESTC_AL+MR+WS

		if (false) {

			if (dsClusteringStrategies.containsKey("BESTC+BESTND+WS+DS") && dsNdStrategies.containsKey("BESTC_AL+WS")) {
				dsNdStrategies.put("BESTC+BESTND+WS+DS", dsNdStrategies.get("BESTC_AL+WS"));
				CHMHelper.getFromMap(ndStrategies, "BESTC+BESTND+WS+DS", () -> new ArrayList<>()).add(EntityMatchingRunRepository.RF_ET_CS_LABEL_EM_ALL);
			} else {
				System.out.println("BEST METHOD NOT FOUND (WITHOUT MR)");
			}
			if (dsClusteringStrategies.containsKey("BESTC+BESTND+MR+WS+DS") && dsNdStrategies.containsKey("BESTC_AL+MR+WS")) {
				dsNdStrategies.put("BESTC+BESTND+MR+WS+DS", dsNdStrategies.get("BESTC_AL+MR+WS"));
				CHMHelper.getFromMap(ndStrategies, "BESTC+BESTND+MR+WS+DS", () -> new ArrayList<>()).add(EntityMatchingRunRepository.RF_ET_CS_LABEL_EM_ALL);
			} else {
				System.out.println("BEST METHOD NOT FOUND (WITH MR)");
			}
		}

		if (dsClusteringStrategies.containsKey("AL+WS") && dsNdStrategies.containsKey("ND_AL+WS_AL+WS")) {
			dsNdStrategies.put("AL+WS", dsNdStrategies.get("ND_AL+WS_AL+WS"));
			CHMHelper.getFromMap(ndStrategies, "AL+WS", () -> new ArrayList<>()).add(EntityMatchingRunRepository.RF_ET_CS_LABEL_EM_ALL);
		} else {
			System.out.println("Method not found: AL+WS");
		}
		if (dsClusteringStrategies.containsKey("MR+WS") && dsNdStrategies.containsKey("ND_MR+WS_MR+WS")) {
			dsNdStrategies.put("MR+WS", dsNdStrategies.get("ND_MR+WS_MR+WS"));
			CHMHelper.getFromMap(ndStrategies, "MR+WS", () -> new ArrayList<>()).add(EntityMatchingRunRepository.RF_ET_CS_LABEL_EM_ALL);
		} else {
			System.out.println("Method not found: MR+WS");
		}

		if (dsClusteringStrategies.containsKey("MR") && dsNdStrategies.containsKey("ND_AL+MR")) {
			dsNdStrategies.put("MR", dsNdStrategies.get("ND_AL+MR"));
			CHMHelper.getFromMap(ndStrategies, "MR", () -> new ArrayList<>()).add(EntityMatchingRunRepository.CONTAINER_ND_ALL);
		} else {
			System.out.println("Method not found: MR");
		}

		Map<Integer, Map<String, Map<ClassId, EM_LearnedModelRepository>>> ndExtraPerFold = new HashMap<>();
		Map<Integer, Map<String, Map<ClassId, LearnedModelRepository>>> dsExtraPerFold = new HashMap<>();

		if (GS_RUN_ND || GS_RUN) {
			for (int i = 0; i < goldStandardClustersSet.getSetCount(); i++) {
				CHMHelper.getFromMap(ndExtraPerFold, i, () -> new HashMap<>());

				String currentRunResultsFolder = currentCaseFolder + "/" + i;
				MKDIR.createDirUnchecked(new File(currentRunResultsFolder));

				Map<ClassId, GoldStandardClusters> learningGoldStandard = goldStandardClustersSet.getClustersOfOtherThan(i);
				Map<ClassId, GoldStandardClusters> testingGoldStandard = goldStandardClustersSet.getClustersOf(i);

				Map<ClassId, Set<InstanceId>> learningNewEntityIds = goldStandardClustersSet.getNewInstanceIdsOfOtherThan(i);

				Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>> learningFactsMap = goldStandardClustersSet.getFactsMapOfOtherThan(i);

				Map<ClassId, Set<TableRowId>> learningInput = goldStandardClustersSet.getInputOfOtherThan(i);

				Map<ClassId, Set<InstanceId>> testingNewEntityIds = goldStandardClustersSet.getNewInstanceIdsOf(i);

				Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>> testingFactsMap = goldStandardClustersSet.getFactsMapOf(i);
				Map<ClassId, Set<TableRowId>> testingInput = goldStandardClustersSet.getInputOf(i);

				for (ClassId classId : learningGoldStandard.keySet()) {
					validationRuns.add(new ValidationRun(learningGoldStandard.get(classId), learningInput.get(classId), learningNewEntityIds.get(classId),
							classId, fullInput.get(classId), rowIdsSetAllClasses));
				}

				Map<ClassId, Pair<EM_LearnedModelRepository, Map<EntityMatchingRun, NewDetectionPerformance>>> entityMatchingWeights = null;

				if (GS_RUN_ND) {

					entityMatchingWeights = fusion.entityMatchingGS_ClustererWeightsAdapterAndLearner.getLearnedModel(ndStrategies.get("GS_RUN"),
							learningGoldStandard, testingGoldStandard, fusionConfig, setExpansionStrategiesConfig, strategiesConfig, fullRowIdsSet,
							learningInput, testingInput, thresholdsFolder, allParents, learningNewEntityIds, testingNewEntityIds,
							setExtension.getNonTailProperties(), null, currentRunResultsFolder, additionalExisting, "GS_RUN");

					entityWeightsMap.put(i, entityMatchingWeights);

				}

				Map<ClassId, LearnedModelRepository> learnedModels = null;
				if (GS_RUN) {

					learnedModels = fusion.clusteringLearnerRunner.getAllWeights(new HashMap<>(), learningInput, fullRowIdsSet, fusionConfig,
							setExpansionStrategiesConfig, strategiesConfig, thresholdsFolder, learningGoldStandard, learningNewEntityIds, conflictingSets,
							allParents, setExtension.getNonTailProperties(), null);
					learnedModelsMap.put(i, learnedModels);

				}

				if (GS_RUN || GS_RUN_ND) {

					Map<ClassId, EM_LearnedModelRepository> GS_TRAINED = entityMatchingWeights.entrySet().stream()
							.collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().getFirst()));

					// Map<String, DS_Container> GS_WITH_MR =
					// ND_MatchingRules_Creator
					// .get("GS_RUN+MR",
					// EntityMatchingRunRepository.RF_ET_CS_LABEL_EM_ALL,
					// GS_TRAINED).get(distantSupervisionInput,
					// clusteringRunner);
					//
					// for (Entry<String, DS_Container> entry :
					// GS_WITH_MR.entrySet()) {
					// ndStrategies.put(entry.getKey(),
					// entry.getValue().getNewDetectionRuns());
					// CHMHelper.getFromMap(ndExtraPerFold, i, () -> new
					// HashMap<>()).put(entry.getKey(),
					// entry.getValue().getNewDetection());
					//
					// Map<String, DS_Container> wsStrategies =
					// WS_ND_Strategy_Creator
					// .get("GS_RUN+MR+WS",
					// EntityMatchingRunRepository.CONTAINER_ND_ALL,
					// entry.getValue().getNewDetection(),
					// SetExpansionStrategiesRepository.COMBINED_RandomForest_ET,
					// learnedModels)
					// .get(distantSupervisionInput, clusteringRunner);
					//
					// for (Entry<String, DS_Container> entry2 :
					// wsStrategies.entrySet()) {
					// ndStrategies.put(entry2.getKey(),
					// entry2.getValue().getNewDetectionRuns());
					// CHMHelper.getFromMap(ndExtraPerFold, i, () -> new
					// HashMap<>()).put(entry2.getKey(),
					// entry2.getValue().getNewDetection());
					// }
					//
					// }

					{
						Map<String, DS_Container> wsStrategies = WS_RC_Strategy_Creator
								.get("GS_RUN+WS", SetExpansionStrategiesRepository.COMBINED_RandomForest_ET, learnedModels,
										SetExpansionStrategiesRepository.COMBINED_RandomForest_ET)
								.get(distantSupervisionInput, clusteringRunner);

						for (Entry<String, DS_Container> entry : wsStrategies.entrySet()) {

							CHMHelper.getFromMap(dsExtraPerFold, i, () -> new HashMap<>()).put(entry.getKey(), entry.getValue().getRowClustering());

							dsClusteringStrategies.put(entry.getKey(), entry.getValue().getRowClustering());
							clusteringStrategies.put(entry.getKey(), entry.getValue().getClusteringStrategies());
							// if (entry.getValue().hasNewDetection()) {
							// dsNdStrategies.put(entry.getKey(),
							// entry.getValue().getNewDetection());
							// }

							Map<String, DS_Container> ws2Strategies = WS_ND_Strategy_Creator
									.get("GS_RUN+WS", EntityMatchingRunRepository.RF_ET_CS_LABEL_EM_ALL, GS_TRAINED,
											SetExpansionStrategiesRepository.COMBINED_RandomForest_ET, entry.getValue().getRowClustering())
									.get(distantSupervisionInput, clusteringRunner);

							for (Entry<String, DS_Container> entry2 : ws2Strategies.entrySet()) {
								ndStrategies.put(entry2.getKey(), entry2.getValue().getNewDetectionRuns());
								CHMHelper.getFromMap(ndExtraPerFold, i, () -> new HashMap<>()).put(entry2.getKey(), entry2.getValue().getNewDetection());
							}

						}
					}

				}
			}
		}

		for (Entry<String, SetExpansionStrategiesConfig> fusionRunEntry : fusionStrategies.entrySet()) {

			System.out.println("FUSION_RUN: " + fusionRunEntry.getKey());

			Map<ClassId, Map<SetExpansionStrategy, Map<String, ConcurrentAccumulator<PipelinePerformance>>>> performanceClassIdMap = new LinkedHashMap<>();

			Map<ClassId, Map<EntityMatchingRun, Map<String, ConcurrentAccumulator<NewDetectionPerformance>>>> newDetectionPerformanceClassIdMap = new LinkedHashMap<>();

			long estimatedTime = 0;

			for (int i = 0; i < goldStandardClustersSet.getSetCount(); i++) {

				Map<ClassId, Pair<EM_LearnedModelRepository, Map<EntityMatchingRun, NewDetectionPerformance>>> entityMatchingWeights = entityWeightsMap.get(i);
				Map<ClassId, LearnedModelRepository> learnedModels = learnedModelsMap.get(i);

				String currentRunResultsFolder = currentCaseFolder + "/" + i;
				MKDIR.createDirUnchecked(new File(currentRunResultsFolder));

				Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>> testingFactsMap = goldStandardClustersSet.getFactsMapOf(i);
				Map<ClassId, GoldStandardClusters> testingGoldStandard = goldStandardClustersSet.getClustersOf(i);
				Map<ClassId, Set<TableRowId>> testingInput = goldStandardClustersSet.getInputOf(i);
				Map<ClassId, Set<InstanceId>> testingNewEntityIds = goldStandardClustersSet.getNewInstanceIdsOf(i);

				if (true) {

					Map<String, Map<ClassId, Pair<EM_LearnedModelRepository, Map<EntityMatchingRun, NewDetectionPerformance>>>> ndModelsLearnedRun = new LinkedHashMap<>();
					Map<String, Map<ClassId, EM_LearnedModelRepository>> dsNdStrategiesPerFold = new HashMap<>(dsNdStrategies);
					if (GS_RUN_ND && entityMatchingWeights == null) {
						throw new IllegalStateException();

						// ndModelsLearnedRun.put("GS_RUN",
						// entityMatchingWeights);
					} else if (GS_RUN_ND) {

						ndModelsLearnedRun.put("GS_RUN", entityMatchingWeights);
						dsNdStrategiesPerFold.put("GS_RUN",
								entityMatchingWeights.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().getFirst())));
						for (Entry<String, Map<ClassId, EM_LearnedModelRepository>> extraPerFold : ndExtraPerFold.get(i).entrySet()) {
							dsNdStrategiesPerFold.put(extraPerFold.getKey(), extraPerFold.getValue());

						}
					}

					if (GS_RUN_ND) {
						for (String key : dsClusteringStrategies.keySet()) {
							// ndModelsLearnedRun.put(key,
							// entityMatchingWeights);
						}
					}

					for (Entry<String, Map<ClassId, EM_LearnedModelRepository>> entry : dsNdStrategiesPerFold.entrySet()) {

						Map<ClassId, EM_LearnedModelRepository> matchingWeightsMap = entry.getValue();
						Map<ClassId, Map<EntityMatchingRun, NewDetectionPerformance>> resultsMap = new HashMap<>();
						fusion.entityMatchingGS_ClustererWeightsAdapterAndLearner.getWeightsNewDetection(ndStrategies.get(entry.getKey()), testingGoldStandard,
								fusionConfig, setExpansionStrategiesConfig, strategiesConfig, fullInput, testingInput, thresholdsFolder, allParents,
								testingNewEntityIds, matchingWeightsMap, resultsMap, setExtension.getNonTailProperties(), currentRunResultsFolder,
								new HashMap<>(), entry.getKey());
						Map<ClassId, Pair<EM_LearnedModelRepository, Map<EntityMatchingRun, NewDetectionPerformance>>> ret = new HashMap<>();
						for (Entry<ClassId, EM_LearnedModelRepository> iEntry : entry.getValue().entrySet()) {
							ret.put(iEntry.getKey(), new Pair<>(iEntry.getValue(), resultsMap.get(iEntry.getKey())));
						}
						ndModelsLearnedRun.put(entry.getKey(), ret);
					}

					for (Entry<String, Map<ClassId, Pair<EM_LearnedModelRepository, Map<EntityMatchingRun, NewDetectionPerformance>>>> entry : ndModelsLearnedRun
							.entrySet()) {

						for (Entry<ClassId, Pair<EM_LearnedModelRepository, Map<EntityMatchingRun, NewDetectionPerformance>>> m : entry.getValue().entrySet()) {

							for (Entry<EntityMatchingRun, NewDetectionPerformance> strategyEntry : m.getValue().getSecond().entrySet()) {
								Map<EntityMatchingRun, Map<String, ConcurrentAccumulator<NewDetectionPerformance>>> innerMap = CHMHelper
										.getFromMap(newDetectionPerformanceClassIdMap, m.getKey(), () -> new LinkedHashMap<>());
								Map<String, ConcurrentAccumulator<NewDetectionPerformance>> innerInnerMap = CHMHelper.getFromMap(innerMap,
										strategyEntry.getKey(), () -> new LinkedHashMap<>());

								CHMHelper.getFromMap(innerInnerMap, entry.getKey(), () -> new ConcurrentAccumulator<>(new NewDetectionPerformance()))
										.add(strategyEntry.getValue());

							}
						}
					}

					MKDIR.createDirUnchecked(new File(currentCaseFolder));
					if (RUN_CLUSTERING_FLAG) {
						if (GS_RUN && learnedModels == null) {
							throw new IllegalStateException();
						}

						// fusion.runner.gsRunNew(learningInput, fullInput,
						// fusionConfig, thresholdsFolder,
						// setExpansionStrategiesConfig,
						// strategiesConfig,
						// learningGoldStandard, learningNewEntityIds,
						// allParents,
						// learningFactsMap, fullCorrectMap, null, new
						// HashMap<>(learnedModels),
						// new HashMap<>(entityMatchingWeights),
						// setExtension.getNonTailProperties(),
						// alwaysDisableBlocking);

						Map<String, Map<ClassId, LearnedModelRepository>> clusterindModelsLearnedRun = new LinkedHashMap<>();
						if (GS_RUN) {
							clusterindModelsLearnedRun.put("GS_RUN", learnedModels);

							for (Entry<String, Map<ClassId, LearnedModelRepository>> extraPerFold : dsExtraPerFold.get(i).entrySet()) {
								clusterindModelsLearnedRun.put(extraPerFold.getKey(), extraPerFold.getValue());

							}
						}

						clusterindModelsLearnedRun.putAll(dsClusteringStrategies);

						for (Entry<String, Map<ClassId, LearnedModelRepository>> entry : clusterindModelsLearnedRun.entrySet()) {
							System.out.println("LEARNING-METHOD: " + entry.getKey());
							long startTime = System.nanoTime();

							SetExpansionStrategiesConfig adaptedSetExpansionStrategiesConfig = setExpansionStrategiesConfig.rebuilder().clearStrategies()
									.addAllStrategy(clusteringStrategies.get(entry.getKey())).build();
							String ndModel = entry.getKey();
							if (!dsNdStrategiesPerFold.containsKey(ndModel) && defaultNDModel != null) {
								ndModel = defaultNDModel;
							}
							System.out.println("WITH CLUSTERING METHOD " + entry.getKey() + ", running ND Method " + ndModel + " and model "
									+ ndStrategies.get(ndModel).get(0).getName());
							Map<SetExpansionStrategy, FusionPerformance<PipelinePerformance>> ret = fusion.runner.gsRunNew(testingInput, fullInput,
									fusionConfig, thresholdsFolder, adaptedSetExpansionStrategiesConfig, strategiesConfig, testingGoldStandard,
									testingNewEntityIds, allParents, testingFactsMap, fullCorrectMap, currentRunResultsFolder, new HashMap<>(entry.getValue()),

									dsNdStrategiesPerFold.get(ndModel), setExtension.getNonTailProperties(), alwaysDisableBlocking, i,
									ndStrategies.get(ndModel).get(0));

							estimatedTime += System.nanoTime() - startTime;

							for (Entry<SetExpansionStrategy, FusionPerformance<PipelinePerformance>> m : ret.entrySet()) {

								for (Entry<ClassId, PipelinePerformance> perClassId : m.getValue().getPerClassId().entrySet()) {
									Map<SetExpansionStrategy, Map<String, ConcurrentAccumulator<PipelinePerformance>>> innerMap = CHMHelper
											.getFromMap(performanceClassIdMap, perClassId.getKey(), () -> new LinkedHashMap<>());
									Map<String, ConcurrentAccumulator<PipelinePerformance>> innerInnerMap = CHMHelper.getFromMap(innerMap, m.getKey(),
											() -> new LinkedHashMap<>());
									CHMHelper.getFromMap(innerInnerMap, entry.getKey(), () -> new ConcurrentAccumulator<>(new PipelinePerformance()))
											.add(perClassId.getValue());

								}
							}
						}

					} else {
						for (SetExpansionStrategy strategy : setExpansionStrategiesConfig.getStrategies()) {

							for (ClassId classId : classesAndProperties.keySet()) {
								Map<SetExpansionStrategy, Map<String, ConcurrentAccumulator<PipelinePerformance>>> innerMap = CHMHelper
										.getFromMap(performanceClassIdMap, classId, () -> new LinkedHashMap<>());
								Map<String, ConcurrentAccumulator<PipelinePerformance>> innerInnerMap = CHMHelper.getFromMap(innerMap, strategy,
										() -> new LinkedHashMap<>());

								CHMHelper.getFromMap(innerInnerMap, "EMPTY", () -> new ConcurrentAccumulator<>(new PipelinePerformance()))
										.add(new PipelinePerformance());

							}
						}
					}

				} else {

					for (EntityMatchingRun run : runs) {

						for (ClassId classId : classesAndProperties.keySet()) {
							Map<EntityMatchingRun, Map<String, ConcurrentAccumulator<NewDetectionPerformance>>> innerMap = CHMHelper
									.getFromMap(newDetectionPerformanceClassIdMap, classId, () -> new LinkedHashMap<>());

							Map<String, ConcurrentAccumulator<NewDetectionPerformance>> innerInnerMap = CHMHelper.getFromMap(innerMap, run,
									() -> new LinkedHashMap<>());

							CHMHelper.getFromMap(innerInnerMap, "EMPTY", () -> new ConcurrentAccumulator<>(new NewDetectionPerformance()));

						}
					}

					for (SetExpansionStrategy strategy : setExpansionStrategiesConfig.getStrategies()) {

						for (ClassId classId : classesAndProperties.keySet()) {
							Map<SetExpansionStrategy, Map<String, ConcurrentAccumulator<PipelinePerformance>>> innerMap = CHMHelper
									.getFromMap(performanceClassIdMap, classId, () -> new LinkedHashMap<>());
							Map<String, ConcurrentAccumulator<PipelinePerformance>> innerInnerMap = CHMHelper.getFromMap(innerMap, strategy,
									() -> new LinkedHashMap<>());
							CHMHelper.getFromMap(innerInnerMap, "EMPTY", () -> new ConcurrentAccumulator<>(new PipelinePerformance()))
									.add(new PipelinePerformance());

						}
					}

				}

			}

			Map<Pair<EntityMatchingRun, String>, ConcurrentAccumulator<NewDetectionPerformance>> ndPerformanceMap = new LinkedHashMap<>();

			GenericResultsWriter<NewDetectionPerformance> ndWriter = new GenericResultsWriter<>(
					fusion.entityMatchingGS_ClustererWeightsAdapterAndLearner.getColumnListforPipelinePerformance());

			for (Entry<ClassId, Map<EntityMatchingRun, Map<String, ConcurrentAccumulator<NewDetectionPerformance>>>> entry : newDetectionPerformanceClassIdMap
					.entrySet()) {

				for (Entry<EntityMatchingRun, Map<String, ConcurrentAccumulator<NewDetectionPerformance>>> innerEntry : entry.getValue().entrySet()) {

					for (Entry<String, ConcurrentAccumulator<NewDetectionPerformance>> innerInnerEntry : innerEntry.getValue().entrySet()) {

						NewDetectionPerformance pp = innerInnerEntry.getValue().get();
						ndWriter.writeClassIdWithLearner(innerEntry.getKey().getName(), innerInnerEntry.getKey(), entry.getKey(), pp);

						CHMHelper.getFromMap(ndPerformanceMap, new Pair<>(innerEntry.getKey(), innerInnerEntry.getKey()),
								() -> new ConcurrentAccumulator<>(new NewDetectionPerformance())).add(pp);

						TreeMap<String, List<String>> treeMap = CHMHelper.getFromMap(optimizationRetMap, "ND", () -> new TreeMap<>());
						List<String> list = CHMHelper.getFromMap(treeMap, innerEntry.getKey().getName() + "_" + innerInnerEntry.getKey(),
								() -> new ArrayList<>());
						list.add(DecimalFormatter.format(pp.getTotalAccuracy()));
						list.add(DecimalFormatter.format(pp.getExistingFmeasure()));
						list.add(DecimalFormatter.format(pp.getNewFmeasure()));
						list.add("");
					}
				}

				// for (EntityMatchingRun run : runs) {
				//
				// NewDetectionPerformance pp = entry.getValue().get(run).get();
				// ndWriter.writeClassId(run.getName(), entry.getKey(), pp);
				//
				// CHMHelper.getFromMap(ndPerformanceMap, run, () -> new
				// ConcurrentAccumulator<>(new
				// NewDetectionPerformance())).add(pp);
				// }

			}

			for (Entry<Pair<EntityMatchingRun, String>, ConcurrentAccumulator<NewDetectionPerformance>> entry : ndPerformanceMap.entrySet()) {
				NewDetectionPerformance pp = entry.getValue().get();
				ndWriter.writeTotalWithLearner(entry.getKey().getFirst().getName(), entry.getKey().getSecond(), pp);

				TreeMap<String, List<String>> treeMap = CHMHelper.getFromMap(optimizationRetMap, "ND", () -> new TreeMap<>());
				List<String> list = CHMHelper.getFromMap(treeMap, entry.getKey().getFirst().getName() + "_" + entry.getKey().getSecond(),
						() -> new ArrayList<>());
				list.add(DecimalFormatter.format(pp.getTotalAccuracy()));
				list.add(DecimalFormatter.format(pp.getExistingFmeasure()));
				list.add(DecimalFormatter.format(pp.getNewFmeasure()));
			}

			System.out.println("AGGREGATED:");
			System.out.println(ndWriter.getAll(ResultsCategory.TOTAL_WITH_LEARNER));
			System.out.println();
			System.out.println(ndWriter.getAll(ResultsCategory.CLASS_ID_WITH_LEARNER));

			System.out.println("ThisRunTook: " + formatInterval(estimatedTime));

			Map<Pair<SetExpansionStrategy, String>, ConcurrentAccumulator<PipelinePerformance>> performanceMap = new LinkedHashMap<>();

			GenericResultsWriter<PipelinePerformance> writer = new GenericResultsWriter<>(fusion.runner.getColumnListforPipelinePerformance());

			for (Entry<ClassId, Map<SetExpansionStrategy, Map<String, ConcurrentAccumulator<PipelinePerformance>>>> entry : performanceClassIdMap.entrySet()) {

				for (Entry<SetExpansionStrategy, Map<String, ConcurrentAccumulator<PipelinePerformance>>> innerEntry : entry.getValue().entrySet()) {
					for (Entry<String, ConcurrentAccumulator<PipelinePerformance>> innerInnerEntry : innerEntry.getValue().entrySet()) {
						ConcurrentAccumulator<PipelinePerformance> value = innerInnerEntry.getValue();
						PipelinePerformance pp = value.get();
						writer.writeClassIdWithLearner(innerEntry.getKey().getName(), innerInnerEntry.getKey(), entry.getKey(), pp);

						CHMHelper.getFromMap(performanceMap, new Pair<>(innerEntry.getKey(), innerInnerEntry.getKey()),
								() -> new ConcurrentAccumulator<>(new PipelinePerformance())).add(pp);

						TreeMap<String, List<String>> treeMap = CHMHelper.getFromMap(optimizationRetMap, "RC", () -> new TreeMap<>());
						List<String> list = CHMHelper.getFromMap(treeMap, innerEntry.getKey().getName() + "_" + innerInnerEntry.getKey(),
								() -> new ArrayList<>());
						list.add(DecimalFormatter.format(pp.getClusteringAccuracy().getFMeasureMacro()));

					}
				}

			}

			for (Entry<Pair<SetExpansionStrategy, String>, ConcurrentAccumulator<PipelinePerformance>> strategy : performanceMap.entrySet()) {

				PipelinePerformance pp = strategy.getValue().get();
				writer.writeTotalWithLearner(strategy.getKey().getFirst().getName(), strategy.getKey().getSecond(), pp);
				TreeMap<String, List<String>> treeMap = CHMHelper.getFromMap(optimizationRetMap, "RC", () -> new TreeMap<>());
				List<String> list = CHMHelper.getFromMap(treeMap, strategy.getKey().getFirst().getName() + "_" + strategy.getKey().getSecond(),
						() -> new ArrayList<>());
				list.add("");
				list.add(DecimalFormatter.format(pp.getClusteringAccuracy().getFMeasureMacro()));
			}

			System.out.println("AGGREGATED:, blocking: " + !alwaysDisableBlocking);
			System.out.println(writer.getAll(ResultsCategory.TOTAL_WITH_LEARNER));
			System.out.println();
			System.out.println(writer.getAll(ResultsCategory.CLASS_ID_WITH_LEARNER));

			System.out.println("ThisRunTook: " + formatInterval(estimatedTime));

		}

		if (run_HyperParameter) {
			SetExpansionStrategiesConfig finalsetExpansionStrategiesConfig = setExpansionStrategiesConfig;
			if (true) {
				Map<UUID, Pair<EntityMatcherFullResults, Map<RowClusterUUID, Optional<InstanceId>>>> em_cache_map = new HashMap<>();

				Map<ClassId, Set<PropertyId>> nonTailProperties = setExtension.getNonTailProperties();
				for (ValidationRun validationRun : validationRuns) {

					em_cache_map.put(validationRun.getUuid(),
							fusion.entityMatchingGS_ClustererWeightsAdapterAndLearner.getFullResults(EntityMatchingRunRepository.CS_LABEL_EM_ALL,
									validationRun.getGs(), fusionConfig, finalsetExpansionStrategiesConfig, strategiesConfig, validationRun.getFullInput(),
									validationRun.getCandidates(), thresholdsFolder, validationRun.getClassId(), allParents, validationRun.getNewEntities(),
									nonTailProperties, validationRun.getFullInputAllClasses()));
				}

				HyperparameterOptimizer.RandomForestRun run = (validationRun, parameters) -> {

					EntityMatcherWekaLearner wekaLearner = new EntityMatcherWekaLearner(
							Sets.newHashSet(EntityMatcherStrategies.ATTRIBUTE, EntityMatcherStrategies.IMPLICIT_ATTRIBUTE), "RandomForest",
							parameters.getParameters(), parameters.isNumeric(), null, parameters.isDescretize(), true, new WeightsRepositoryImpl(), true);
					Pair<EntityMatcherFullResults, Map<RowClusterUUID, Optional<InstanceId>>> pair = em_cache_map.get(validationRun.getUuid());

					double error = fusion.entityMatchingGS_ClustererWeightsAdapterAndLearner.errorRun(pair.getFirst(), pair.getSecond(),
							EntityMatchingRunRepository.CS_LABEL_EM_ALL, wekaLearner, validationRun.getNewEntities());
					return error;

				};

				System.out.println("EntityMatchingHyperParameters: ");
				hyperparameterOptimizer.findHyperParameters_EM_2(validationRuns, run, false);

				// System.out.println("EntityMatchingHyperParameters: ");
				// hyperparameterOptimizer.findHyperParameters(validationRuns,
				// run, false);

			}

			if (true) {
				Map<UUID, RowClusteringLearnableSimCache> clusteringCache = new HashMap<>();

				for (ValidationRun validationRun : validationRuns) {
					RowClusteringLearnableSimCache learnCache = new RowClusteringLearnableSimCache();

					SetExpansionStrategy strategy = new SetExpansionStrategy("WEKA_GENERIC", learnCache, true);
					SetExpansionStrategiesConfig newSetExpansionStrategiesConfig = finalsetExpansionStrategiesConfig.rebuilder().clearStrategies()
							.addStrategy(strategy).addStrategy(SetExpansionStrategiesRepository.COMBINED_BASE).build();
					Map<ClassId, Set<TableRowId>> learningInput = new HashMap<>();
					Map<ClassId, GoldStandardClusters> learningGoldStandard = new HashMap<>();
					Map<ClassId, TableRowIdsSet> fullInputInner = new HashMap<>();
					Map<ClassId, Set<InstanceId>> learningNewEntityIds = new HashMap<>();
					learningInput.put(validationRun.getClassId(), validationRun.getCandidates());
					learningGoldStandard.put(validationRun.getClassId(), validationRun.getGs());
					fullInputInner.put(validationRun.getClassId(), validationRun.getFullInput());
					learningNewEntityIds.put(validationRun.getClassId(), validationRun.getNewEntities());
					fusion.clusteringLearnerRunner.getAllWeights(new HashMap<>(), learningInput, fullInputInner, fusionConfig, newSetExpansionStrategiesConfig,
							strategiesConfig, thresholdsFolder, learningGoldStandard, learningNewEntityIds, conflictingSets, allParents,
							setExtension.getNonTailProperties(), validationRun.getFullInputAllClasses());

					clusteringCache.put(validationRun.getUuid(), learnCache);
				}

				HyperparameterOptimizer.RandomForestRun clusterRun = (validationRun, parameters) -> {

					RowClusteringLearnableSimCache simCache = clusteringCache.get(validationRun.getUuid());
					MT_WekaLearnedModel model = (MT_WekaLearnedModel) ((LearnableRowClusterSimilarities) wekaGeneric.get(parameters, "WEKA_GENERIC")
							.clusterRows(simCache.getConfig(), simCache.getInput())).learn(simCache.getPairs(), simCache.getClassId(),
									simCache.getEvaluationMethod());

					// SetExpansionStrategy strategy = new
					// SetExpansionStrategy("WEKA_GENERIC", , true);
					// SetExpansionStrategiesConfig
					// newSetExpansionStrategiesConfig
					// =
					// finalsetExpansionStrategiesConfig.rebuilder().clearStrategies()
					// .addStrategy(strategy).build();
					// Map<ClassId, Set<TableRowId>> learningInput = new
					// HashMap<>();
					// Map<ClassId, GoldStandardClusters> learningGoldStandard =
					// new
					// HashMap<>();
					// Map<ClassId, TableRowIdsSet> fullInputInner = new
					// HashMap<>();
					// Map<ClassId, Set<InstanceId>> learningNewEntityIds = new
					// HashMap<>();
					// learningInput.put(validationRun.getClassId(),
					// validationRun.getCandidates());
					// learningGoldStandard.put(validationRun.getClassId(),
					// validationRun.getGs());
					// fullInputInner.put(validationRun.getClassId(),
					// validationRun.getFullInput());
					// learningNewEntityIds.put(validationRun.getClassId(),
					// validationRun.getNewEntities());
					//
					// Map<ClassId, LearnedModelRepository> learnedModels =
					// fusion.clusteringLearnerRunner.getAllWeights(new
					// HashMap<>(),
					// learningInput,
					// fullInputInner, fusionConfig,
					// newSetExpansionStrategiesConfig,
					// strategiesConfig, thresholdsFolder, learningGoldStandard,
					// learningNewEntityIds, conflictingSets, allParents,
					// setExtension.getNonTailProperties());
					//
					// double error = ((MT_WekaLearnedModel)
					// learnedModels.get(validationRun.getClassId()).get(strategy.getOptimizableClassCluster())).getError();
					double error = model.getError();
					return error;

				};

				System.out.println("ClusterHyperParameters: ");
				hyperparameterOptimizer.findHyperParameters_cluster(validationRuns, clusterRun);
			}

		}

		return optimizationRetMap;

	}

	private Map<String, TreeMap<String, List<String>>> runEvaluateMethod(GoldStandardClustersSetPerClass goldStandardClustersSet, FusionConfig fusionConfig,
			SetExpansionStrategiesConfig setExpansionStrategiesConfig, StrategiesConfig strategiesConfig, Map<ClassId, TableRowIdsSet> fullInput,
			String thresholdsFolder, Map<ClassId, Set<ClassId>> allParents, SetExtension setExtension, String instanceMatchingWeightsFolder,
			Map<ClassId, Map<InstanceId, Integer>> conflictingSets, String currentCaseFolder,
			Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>> fullCorrectMap, boolean alwaysDisableBlocking,
			Map<ClassId, Set<PropertyId>> classesAndProperties, Map<InstanceId, Set<TableRowId>> additionalExisting) throws IOException {
		EntityMatchingRun defaultNDModel = EntityMatchingRunRepository.CS_LABEL_EM_LABEL;

		Map<String, TreeMap<String, List<String>>> optimizationRetMap = new HashMap<String, TreeMap<String, List<String>>>();
		int cpus = NumOfCpusGetter.getProcessors();
		boolean GS_RUN = true;
		boolean RUN_CLUSTERING_FLAG = true;
		boolean GS_RUN_ND = true;
		boolean run_HyperParameter = false;
		boolean HyperParameter_NewDetection = false;
		boolean HyperParameter_Clustering = false;

		List<ValidationRun> validationRuns = new ArrayList<>();

		GSRun fusion = this;

		setExpansionStrategiesConfig = setExpansionStrategiesConfig.rebuilder().clearStrategies()
				.addAllStrategy(Lists.newArrayList(SetExpansionStrategiesRepository.FUSION_TESTER,
						// SetExpansionStrategiesRepository.GS_CLUSTERER_WA,
						// SetExpansionStrategiesRepository.GS_CLUSTERER_RF,
						SetExpansionStrategiesRepository.GS_CLUSTERER_ET, SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY, //

						// SetExpansionStrategiesRepository.RF_LABEL_LABELSF_LABELOVERLAP,

						SetExpansionStrategiesRepository.RF_LABEL_BOW, SetExpansionStrategiesRepository.RF_LABEL_BOW_PHI,

						SetExpansionStrategiesRepository.RF_LABEL_BOW_PHI_ATTRIBUTE, SetExpansionStrategiesRepository.RF_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT,

						SetExpansionStrategiesRepository.RF_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT_SAMETABLE,
						SetExpansionStrategiesRepository.RF_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT_SAMETABLE_LABELOVERLAP,

						SetExpansionStrategiesRepository.WA_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT_SAMETABLE_LABELOVERLAP,
						//
						// SetExpansionStrategiesRepository.WA_LABEL_LABELSF_LABELOVERLAP,
						//
						SetExpansionStrategiesRepository.WA_LABEL_BOW, SetExpansionStrategiesRepository.WA_LABEL_BOW_PHI,

						SetExpansionStrategiesRepository.WA_LABEL_BOW_PHI_ATTRIBUTE, SetExpansionStrategiesRepository.WA_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT,

						SetExpansionStrategiesRepository.WA_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT_SAMETABLE,
						// // // ET
						// //
						// //
						// SetExpansionStrategiesRepository.ET_LABEL_LABELSF_LABELOVERLAP,
						SetExpansionStrategiesRepository.ET_LABEL_BOW,

						SetExpansionStrategiesRepository.ET_LABEL_BOW_PHI,

						SetExpansionStrategiesRepository.ET_LABEL_BOW_PHI_ATTRIBUTE, SetExpansionStrategiesRepository.ET_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT,

						SetExpansionStrategiesRepository.ET_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT_SAMETABLE,
						SetExpansionStrategiesRepository.ET_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT_SAMETABLE_LABELOVERLAP
				// SetExpansionStrategiesRepository.COMBINED_RandomForest_ET
				)).build();

		// setExpansionStrategiesConfig =
		// setExpansionStrategiesConfig.rebuilder().clearStrategies().addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE)
		// .build();

		// setExpansionStrategiesConfig =
		// setExpansionStrategiesConfig.rebuilder().clearStrategies()
		// .addStrategy(SetExpansionStrategiesRepository.COMBINED_RF_BASE_Label_SameTable_Bow_Phi_Attribute_Implicit).build();

		// List<EntityMatchingRun> runs =
		// Arrays.asList(EntityMatchingRunRepository.CS_LABEL_EM_LABEL,
		// EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE,
		// EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE_BOW,
		// EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE_INDEGREE_BOW_ATTRIBUTE,
		// EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE_INDEGREE_BOW_ATTRIBUTE_IMPLICIT,
		// EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL,
		// EntityMatchingRunRepository.RF_NEW_CS_LABEL_EM_ALL,
		// EntityMatchingRunRepository.SW_CS_LABEL_EM_ALL);

		List<EntityMatchingRun> runs = Arrays.asList(EntityMatchingRunRepository.CS_LABEL_EM_LABEL,
				// EntityMatchingRunRepository.CS_LABEL_EM_PRIME_SURFACE_LABEL,
				// EntityMatchingRunRepository.CS_LABEL_EM_LABEL_TYPE,//
				EntityMatchingRunRepository.WA_LABEL_LABELSF, //

				EntityMatchingRunRepository.WA_LABEL_LABELSF_TYPE, //
				EntityMatchingRunRepository.WA_LABEL_LABELSF_TYPE_BOW, //
				EntityMatchingRunRepository.WA_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE,
				//
				EntityMatchingRunRepository.WA_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT,
				//
				EntityMatchingRunRepository.WA_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREE,
				//
				//
				EntityMatchingRunRepository.WA_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREENEW,
				// // // //
				// //
				EntityMatchingRunRepository.WA_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREE_INDEGREENEW,
				//
				//
				EntityMatchingRunRepository.RF_LABEL_LABELSF, //

				EntityMatchingRunRepository.RF_LABEL_LABELSF_TYPE, //
				EntityMatchingRunRepository.RF_LABEL_LABELSF_TYPE_BOW, //
				EntityMatchingRunRepository.RF_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE,
				//
				EntityMatchingRunRepository.RF_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT,
				//
				EntityMatchingRunRepository.RF_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREE,
				//
				//
				EntityMatchingRunRepository.RF_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREENEW,
				// //
				//
				EntityMatchingRunRepository.RF_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREE_INDEGREENEW,
				// // //
				//
				EntityMatchingRunRepository.ET_LABEL_LABELSF, EntityMatchingRunRepository.ET_LABEL_LABELSF_TYPE, //
				EntityMatchingRunRepository.ET_LABEL_LABELSF_TYPE_BOW, //
				EntityMatchingRunRepository.ET_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE, //
				EntityMatchingRunRepository.ET_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT,

				EntityMatchingRunRepository.ET_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREE,
				//
				EntityMatchingRunRepository.ET_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREENEW,
				EntityMatchingRunRepository.ET_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREE_INDEGREENEW
		// EntityMatchingRunRepository.ENSEMBLED_LABEL_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREENEW
		// // //
		// //

		);

		Map<String, SetExpansionStrategiesConfig> fusionStrategies = new HashMap<>();

		fusionStrategies.put("ColumnScores", setExpansionStrategiesConfig.rebuilder().clearFusionSetups()
				.addFusionSetup(FusionStrategyRepository.COLUMN_SCORES, PropertiesConfig.builder().setTimeDependence(TimeDependence.NO).build()).build());

		if (false) {
			fusionStrategies.put("VOTING", setExpansionStrategiesConfig.rebuilder().clearFusionSetups()
					.addFusionSetup(FusionStrategyRepository.NORMAL, PropertiesConfig.builder().setTimeDependence(TimeDependence.NO).build()).build());

			fusionStrategies.put("KBT",
					setExpansionStrategiesConfig.rebuilder().clearFusionSetups()
							.addFusionSetup(FusionStrategyRepository.KBT_NO_THRESHOLD, PropertiesConfig.builder().setTimeDependence(TimeDependence.NO).build())
							.build());

		}

		Map<SetExpansionStrategy, EntityMatchingRun> ndStrategiesPerCluster = new HashMap<>();
		ndStrategiesPerCluster.put(SetExpansionStrategiesRepository.GS_CLUSTERER_WA,
				EntityMatchingRunRepository.WA_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREENEW);
		ndStrategiesPerCluster.put(SetExpansionStrategiesRepository.GS_CLUSTERER_RF,
				EntityMatchingRunRepository.RF_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREENEW);
		ndStrategiesPerCluster.put(SetExpansionStrategiesRepository.GS_CLUSTERER_ET,
				EntityMatchingRunRepository.ET_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREENEW);

		ndStrategiesPerCluster.put(SetExpansionStrategiesRepository.WA_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT_SAMETABLE_LABELOVERLAP,
				EntityMatchingRunRepository.WA_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREENEW);
		ndStrategiesPerCluster.put(SetExpansionStrategiesRepository.RF_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT_SAMETABLE_LABELOVERLAP,
				EntityMatchingRunRepository.RF_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREENEW);
		ndStrategiesPerCluster.put(SetExpansionStrategiesRepository.ET_LABEL_BOW_PHI_ATTRIBUTE_IMPLICIT_SAMETABLE_LABELOVERLAP,
				EntityMatchingRunRepository.ET_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREENEW);

		Map<Integer, Map<ClassId, Pair<EM_LearnedModelRepository, Map<EntityMatchingRun, NewDetectionPerformance>>>> entityWeightsMap = new HashMap<>();
		Map<Integer, Map<ClassId, LearnedModelRepository>> learnedModelsMap = new HashMap<>();

		Set<TableRowId> rowIdsAllClasses = new HashSet<>();
		for (ClassId classId : allParents.keySet()) {
			Set<TableRowId> rowIds = fullInput.get(classId).getRowIds();
			rowIdsAllClasses.addAll(rowIds);
		}
		TableRowIdsSet rowIdsSetAllClasses = new TableRowIdsSet(rowIdsAllClasses, new UuidStringId("ALL"));

		// BESTC+BESTND+WS+DS
		// BESTC_AL+WS
		//
		// BESTC+BESTND+MR+WS+DS
		// BESTC_AL+MR+WS

		Map<Integer, Map<String, Map<ClassId, EM_LearnedModelRepository>>> ndExtraPerFold = new HashMap<>();
		Map<Integer, Map<String, Map<ClassId, LearnedModelRepository>>> dsExtraPerFold = new HashMap<>();

		if (GS_RUN_ND || GS_RUN || run_HyperParameter) {
			for (int i = 0; i < goldStandardClustersSet.getSetCount(); i++) {
				CHMHelper.getFromMap(ndExtraPerFold, i, () -> new HashMap<>());

				String currentRunResultsFolder = currentCaseFolder + "/" + i;
				MKDIR.createDirUnchecked(new File(currentRunResultsFolder));

				Map<ClassId, TableRowIdsSet> fullRowIdsSet = goldStandardClustersSet.getFullInput();

				Map<ClassId, GoldStandardClusters> learningGoldStandard = goldStandardClustersSet.getClustersOfOtherThan(i);
				Map<ClassId, GoldStandardClusters> testingGoldStandard = goldStandardClustersSet.getClustersOf(i);

				Map<ClassId, Set<InstanceId>> learningNewEntityIds = goldStandardClustersSet.getNewInstanceIdsOfOtherThan(i);

				Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>> learningFactsMap = goldStandardClustersSet.getFactsMapOfOtherThan(i);

				Map<ClassId, Set<TableRowId>> learningInput = goldStandardClustersSet.getInputOfOtherThan(i);

				Map<ClassId, Set<InstanceId>> testingNewEntityIds = goldStandardClustersSet.getNewInstanceIdsOf(i);

				Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>> testingFactsMap = goldStandardClustersSet.getFactsMapOf(i);
				Map<ClassId, Set<TableRowId>> testingInput = goldStandardClustersSet.getInputOf(i);

				for (ClassId classId : learningGoldStandard.keySet()) {
					validationRuns.add(new ValidationRun(learningGoldStandard.get(classId), learningInput.get(classId), learningNewEntityIds.get(classId),
							classId, fullInput.get(classId), rowIdsSetAllClasses));
				}

				Map<ClassId, Pair<EM_LearnedModelRepository, Map<EntityMatchingRun, NewDetectionPerformance>>> entityMatchingWeights = null;

				if (GS_RUN_ND) {

					entityMatchingWeights = fusion.entityMatchingGS_ClustererWeightsAdapterAndLearner.getLearnedModel(runs, learningGoldStandard,
							testingGoldStandard, fusionConfig, setExpansionStrategiesConfig, strategiesConfig, fullRowIdsSet, learningInput, testingInput,
							thresholdsFolder, allParents, learningNewEntityIds, testingNewEntityIds, setExtension.getNonTailProperties(), null,
							currentRunResultsFolder, additionalExisting, "GS_RUN");

					entityWeightsMap.put(i, entityMatchingWeights);

				}

				Map<ClassId, LearnedModelRepository> learnedModels = null;
				if (GS_RUN && RUN_CLUSTERING_FLAG) {

					learnedModels = fusion.clusteringLearnerRunner.getAllWeights(new HashMap<>(), learningInput, fullRowIdsSet, fusionConfig,
							setExpansionStrategiesConfig, strategiesConfig, thresholdsFolder, learningGoldStandard, learningNewEntityIds, conflictingSets,
							allParents, setExtension.getNonTailProperties(), null);
					learnedModelsMap.put(i, learnedModels);

				}

				if (GS_RUN || GS_RUN_ND) {

					Map<ClassId, EM_LearnedModelRepository> GS_TRAINED = entityMatchingWeights.entrySet().stream()
							.collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().getFirst()));

					// Map<String, DS_Container> GS_WITH_MR =
					// ND_MatchingRules_Creator
					// .get("GS_RUN+MR",
					// EntityMatchingRunRepository.RF_ET_CS_LABEL_EM_ALL,
					// GS_TRAINED).get(distantSupervisionInput,
					// clusteringRunner);
					//
					// for (Entry<String, DS_Container> entry :
					// GS_WITH_MR.entrySet()) {
					// ndStrategies.put(entry.getKey(),
					// entry.getValue().getNewDetectionRuns());
					// CHMHelper.getFromMap(ndExtraPerFold, i, () -> new
					// HashMap<>()).put(entry.getKey(),
					// entry.getValue().getNewDetection());
					//
					// Map<String, DS_Container> wsStrategies =
					// WS_ND_Strategy_Creator
					// .get("GS_RUN+MR+WS",
					// EntityMatchingRunRepository.CONTAINER_ND_ALL,
					// entry.getValue().getNewDetection(),
					// SetExpansionStrategiesRepository.COMBINED_RandomForest_ET,
					// learnedModels)
					// .get(distantSupervisionInput, clusteringRunner);
					//
					// for (Entry<String, DS_Container> entry2 :
					// wsStrategies.entrySet()) {
					// ndStrategies.put(entry2.getKey(),
					// entry2.getValue().getNewDetectionRuns());
					// CHMHelper.getFromMap(ndExtraPerFold, i, () -> new
					// HashMap<>()).put(entry2.getKey(),
					// entry2.getValue().getNewDetection());
					// }
					//
					// }

				}
			}
		}

		for (Entry<String, SetExpansionStrategiesConfig> fusionRunEntry : fusionStrategies.entrySet()) {

			System.out.println("FUSION_RUN: " + fusionRunEntry.getKey());

			Map<ClassId, Map<SetExpansionStrategy, Map<String, ConcurrentAccumulator<PipelinePerformance>>>> performanceClassIdMap = new LinkedHashMap<>();

			Map<ClassId, Map<EntityMatchingRun, Map<String, ConcurrentAccumulator<NewDetectionPerformance>>>> newDetectionPerformanceClassIdMap = new LinkedHashMap<>();

			long estimatedTime = 0;

			for (int i = 0; i < goldStandardClustersSet.getSetCount(); i++) {

				Map<ClassId, Pair<EM_LearnedModelRepository, Map<EntityMatchingRun, NewDetectionPerformance>>> entityMatchingWeights = entityWeightsMap.get(i);
				Map<ClassId, LearnedModelRepository> learnedModels = learnedModelsMap.get(i);

				String currentRunResultsFolder = currentCaseFolder + "/" + i;
				MKDIR.createDirUnchecked(new File(currentRunResultsFolder));

				Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>> testingFactsMap = goldStandardClustersSet.getFactsMapOf(i);
				Map<ClassId, GoldStandardClusters> testingGoldStandard = goldStandardClustersSet.getClustersOf(i);
				Map<ClassId, Set<TableRowId>> testingInput = goldStandardClustersSet.getInputOf(i);
				Map<ClassId, Set<InstanceId>> testingNewEntityIds = goldStandardClustersSet.getNewInstanceIdsOf(i);

				if (true) {

					Map<String, Map<ClassId, Pair<EM_LearnedModelRepository, Map<EntityMatchingRun, NewDetectionPerformance>>>> ndModelsLearnedRun = new LinkedHashMap<>();
					Map<String, Map<ClassId, EM_LearnedModelRepository>> dsNdStrategiesPerFold = new HashMap<>();
					if (GS_RUN_ND && entityMatchingWeights == null) {
						throw new IllegalStateException();

						// ndModelsLearnedRun.put("GS_RUN",
						// entityMatchingWeights);
					} else if (GS_RUN_ND) {

						ndModelsLearnedRun.put("GS_RUN", entityMatchingWeights);
						dsNdStrategiesPerFold.put("GS_RUN",
								entityMatchingWeights.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().getFirst())));
						for (Entry<String, Map<ClassId, EM_LearnedModelRepository>> extraPerFold : ndExtraPerFold.get(i).entrySet()) {
							dsNdStrategiesPerFold.put(extraPerFold.getKey(), extraPerFold.getValue());

						}
					}

					for (Entry<String, Map<ClassId, EM_LearnedModelRepository>> entry : dsNdStrategiesPerFold.entrySet()) {

						Map<ClassId, EM_LearnedModelRepository> matchingWeightsMap = entry.getValue();
						Map<ClassId, Map<EntityMatchingRun, NewDetectionPerformance>> resultsMap = new HashMap<>();
						fusion.entityMatchingGS_ClustererWeightsAdapterAndLearner.getWeightsNewDetection(runs, testingGoldStandard, fusionConfig,
								setExpansionStrategiesConfig, strategiesConfig, fullInput, testingInput, thresholdsFolder, allParents, testingNewEntityIds,
								matchingWeightsMap, resultsMap, setExtension.getNonTailProperties(), currentRunResultsFolder, new HashMap<>(), entry.getKey());
						Map<ClassId, Pair<EM_LearnedModelRepository, Map<EntityMatchingRun, NewDetectionPerformance>>> ret = new HashMap<>();
						for (Entry<ClassId, EM_LearnedModelRepository> iEntry : entry.getValue().entrySet()) {
							ret.put(iEntry.getKey(), new Pair<>(iEntry.getValue(), resultsMap.get(iEntry.getKey())));
						}
						ndModelsLearnedRun.put(entry.getKey(), ret);
					}

					for (Entry<String, Map<ClassId, Pair<EM_LearnedModelRepository, Map<EntityMatchingRun, NewDetectionPerformance>>>> entry : ndModelsLearnedRun
							.entrySet()) {

						for (Entry<ClassId, Pair<EM_LearnedModelRepository, Map<EntityMatchingRun, NewDetectionPerformance>>> m : entry.getValue().entrySet()) {

							for (Entry<EntityMatchingRun, NewDetectionPerformance> strategyEntry : m.getValue().getSecond().entrySet()) {
								Map<EntityMatchingRun, Map<String, ConcurrentAccumulator<NewDetectionPerformance>>> innerMap = CHMHelper
										.getFromMap(newDetectionPerformanceClassIdMap, m.getKey(), () -> new LinkedHashMap<>());
								Map<String, ConcurrentAccumulator<NewDetectionPerformance>> innerInnerMap = CHMHelper.getFromMap(innerMap,
										strategyEntry.getKey(), () -> new LinkedHashMap<>());

								CHMHelper.getFromMap(innerInnerMap, entry.getKey(), () -> new ConcurrentAccumulator<>(new NewDetectionPerformance()))
										.add(strategyEntry.getValue());

							}
						}
					}

					MKDIR.createDirUnchecked(new File(currentCaseFolder));
					if (RUN_CLUSTERING_FLAG) {
						if (GS_RUN && learnedModels == null) {
							throw new IllegalStateException();
						}

						// fusion.runner.gsRunNew(learningInput, fullInput,
						// fusionConfig, thresholdsFolder,
						// setExpansionStrategiesConfig,
						// strategiesConfig,
						// learningGoldStandard, learningNewEntityIds,
						// allParents,
						// learningFactsMap, fullCorrectMap, null, new
						// HashMap<>(learnedModels),
						// new HashMap<>(entityMatchingWeights),
						// setExtension.getNonTailProperties(),
						// alwaysDisableBlocking);

						Map<String, Map<ClassId, LearnedModelRepository>> clusterindModelsLearnedRun = new LinkedHashMap<>();
						if (GS_RUN) {
							clusterindModelsLearnedRun.put("GS_RUN", learnedModels);

							// for (Entry<String, Map<ClassId,
							// LearnedModelRepository>> extraPerFold :
							// dsExtraPerFold.get(i).entrySet()) {
							// clusterindModelsLearnedRun.put(extraPerFold.getKey(),
							// extraPerFold.getValue());
							//
							// }
						}

						for (Entry<String, Map<ClassId, LearnedModelRepository>> entry : clusterindModelsLearnedRun.entrySet()) {
							System.out.println("LEARNING-METHOD: " + entry.getKey());
							long startTime = System.nanoTime();

							Map<EntityMatchingRun, List<SetExpansionStrategy>> strategies = new HashMap<>();

							for (SetExpansionStrategy strategy : setExpansionStrategiesConfig.getStrategies()) {
								EntityMatchingRun ndModel;
								if (ndStrategiesPerCluster.containsKey(strategy) && runs.contains(ndStrategiesPerCluster.get(strategy))) {
									ndModel = ndStrategiesPerCluster.get(strategy);
								} else {
									ndModel = defaultNDModel;
								}
								CHMHelper.getFromMap(strategies, ndModel, () -> new ArrayList<>()).add(strategy);
							}

							// SetExpansionStrategiesConfig
							// adaptedSetExpansionStrategiesConfig =
							// setExpansionStrategiesConfig.rebuilder().clearStrategies()
							// .addAllStrategy(setExpansionStrategiesConfig.getStrategies()).build();

							for (Entry<EntityMatchingRun, List<SetExpansionStrategy>> ndModelsEntry : strategies.entrySet()) {
								SetExpansionStrategiesConfig setExpansionStrategiesTemp = setExpansionStrategiesConfig.rebuilder().clearStrategies()
										.addAllStrategy(ndModelsEntry.getValue()).build();
								Map<SetExpansionStrategy, FusionPerformance<PipelinePerformance>> ret = fusion.runner.gsRunNew(testingInput, fullInput,
										fusionConfig, thresholdsFolder, setExpansionStrategiesTemp, strategiesConfig, testingGoldStandard, testingNewEntityIds,
										allParents, testingFactsMap, fullCorrectMap, currentRunResultsFolder, new HashMap<>(entry.getValue()),

										ndModelsLearnedRun.get("GS_RUN").entrySet().stream()
												.collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().getFirst())),
										setExtension.getNonTailProperties(), alwaysDisableBlocking, i, ndModelsEntry.getKey());

								estimatedTime += System.nanoTime() - startTime;

								for (Entry<SetExpansionStrategy, FusionPerformance<PipelinePerformance>> m : ret.entrySet()) {

									for (Entry<ClassId, PipelinePerformance> perClassId : m.getValue().getPerClassId().entrySet()) {
										Map<SetExpansionStrategy, Map<String, ConcurrentAccumulator<PipelinePerformance>>> innerMap = CHMHelper
												.getFromMap(performanceClassIdMap, perClassId.getKey(), () -> new LinkedHashMap<>());
										Map<String, ConcurrentAccumulator<PipelinePerformance>> innerInnerMap = CHMHelper.getFromMap(innerMap, m.getKey(),
												() -> new LinkedHashMap<>());
										CHMHelper.getFromMap(innerInnerMap, "GS_RUN", () -> new ConcurrentAccumulator<>(new PipelinePerformance()))
												.add(perClassId.getValue());

									}
								}
							}

						}

					} else {
						for (SetExpansionStrategy strategy : setExpansionStrategiesConfig.getStrategies()) {

							for (ClassId classId : classesAndProperties.keySet()) {
								Map<SetExpansionStrategy, Map<String, ConcurrentAccumulator<PipelinePerformance>>> innerMap = CHMHelper
										.getFromMap(performanceClassIdMap, classId, () -> new LinkedHashMap<>());
								Map<String, ConcurrentAccumulator<PipelinePerformance>> innerInnerMap = CHMHelper.getFromMap(innerMap, strategy,
										() -> new LinkedHashMap<>());

								CHMHelper.getFromMap(innerInnerMap, "EMPTY", () -> new ConcurrentAccumulator<>(new PipelinePerformance()))
										.add(new PipelinePerformance());

							}
						}
					}

				} else {

					for (EntityMatchingRun run : runs) {

						for (ClassId classId : classesAndProperties.keySet()) {
							Map<EntityMatchingRun, Map<String, ConcurrentAccumulator<NewDetectionPerformance>>> innerMap = CHMHelper
									.getFromMap(newDetectionPerformanceClassIdMap, classId, () -> new LinkedHashMap<>());

							Map<String, ConcurrentAccumulator<NewDetectionPerformance>> innerInnerMap = CHMHelper.getFromMap(innerMap, run,
									() -> new LinkedHashMap<>());

							CHMHelper.getFromMap(innerInnerMap, "EMPTY", () -> new ConcurrentAccumulator<>(new NewDetectionPerformance()));

						}
					}

					for (SetExpansionStrategy strategy : setExpansionStrategiesConfig.getStrategies()) {

						for (ClassId classId : classesAndProperties.keySet()) {
							Map<SetExpansionStrategy, Map<String, ConcurrentAccumulator<PipelinePerformance>>> innerMap = CHMHelper
									.getFromMap(performanceClassIdMap, classId, () -> new LinkedHashMap<>());
							Map<String, ConcurrentAccumulator<PipelinePerformance>> innerInnerMap = CHMHelper.getFromMap(innerMap, strategy,
									() -> new LinkedHashMap<>());
							CHMHelper.getFromMap(innerInnerMap, "EMPTY", () -> new ConcurrentAccumulator<>(new PipelinePerformance()))
									.add(new PipelinePerformance());

						}
					}

				}

			}

			Map<Pair<EntityMatchingRun, String>, ConcurrentAccumulator<NewDetectionPerformance>> ndPerformanceMap = new LinkedHashMap<>();

			GenericResultsWriter<NewDetectionPerformance> ndWriter = new GenericResultsWriter<>(
					fusion.entityMatchingGS_ClustererWeightsAdapterAndLearner.getColumnListforPipelinePerformance());

			for (Entry<ClassId, Map<EntityMatchingRun, Map<String, ConcurrentAccumulator<NewDetectionPerformance>>>> entry : newDetectionPerformanceClassIdMap
					.entrySet()) {

				for (Entry<EntityMatchingRun, Map<String, ConcurrentAccumulator<NewDetectionPerformance>>> innerEntry : entry.getValue().entrySet()) {

					for (Entry<String, ConcurrentAccumulator<NewDetectionPerformance>> innerInnerEntry : innerEntry.getValue().entrySet()) {

						NewDetectionPerformance pp = innerInnerEntry.getValue().get();
						ndWriter.writeClassIdWithLearner(innerEntry.getKey().getName(), innerInnerEntry.getKey(), entry.getKey(), pp);

						CHMHelper.getFromMap(ndPerformanceMap, new Pair<>(innerEntry.getKey(), innerInnerEntry.getKey()),
								() -> new ConcurrentAccumulator<>(new NewDetectionPerformance())).add(pp);

						TreeMap<String, List<String>> treeMap = CHMHelper.getFromMap(optimizationRetMap, "ND", () -> new TreeMap<>());
						List<String> list = CHMHelper.getFromMap(treeMap, innerEntry.getKey().getName() + "+" + innerInnerEntry.getKey(),
								() -> new ArrayList<>());
						list.add(DecimalFormatter.format3(pp.getTotalAccuracy()));
						list.add(DecimalFormatter.format3(pp.getExistingFmeasure()));
						list.add(DecimalFormatter.format3(pp.getNewFmeasure()));
						list.add("");
					}
				}

				// for (EntityMatchingRun run : runs) {
				//
				// NewDetectionPerformance pp = entry.getValue().get(run).get();
				// ndWriter.writeClassId(run.getName(), entry.getKey(), pp);
				//
				// CHMHelper.getFromMap(ndPerformanceMap, run, () -> new
				// ConcurrentAccumulator<>(new
				// NewDetectionPerformance())).add(pp);
				// }

			}

			for (Entry<Pair<EntityMatchingRun, String>, ConcurrentAccumulator<NewDetectionPerformance>> entry : ndPerformanceMap.entrySet()) {
				NewDetectionPerformance pp = entry.getValue().get();
				ndWriter.writeTotalWithLearner(entry.getKey().getFirst().getName(), entry.getKey().getSecond(), pp);

				TreeMap<String, List<String>> treeMap = CHMHelper.getFromMap(optimizationRetMap, "ND", () -> new TreeMap<>());
				List<String> list = CHMHelper.getFromMap(treeMap, entry.getKey().getFirst().getName() + "+" + entry.getKey().getSecond(),
						() -> new ArrayList<>());
				list.add(DecimalFormatter.format3(pp.getTotalAccuracy()));
				list.add(DecimalFormatter.format3(pp.getExistingFmeasure()));
				list.add(DecimalFormatter.format3(pp.getNewFmeasure()));
			}

			System.out.println("AGGREGATED:");
			System.out.println(ndWriter.getAll(ResultsCategory.TOTAL_WITH_LEARNER));
			System.out.println();
			System.out.println(ndWriter.getAll(ResultsCategory.CLASS_ID_WITH_LEARNER));

			System.out.println("ThisRunTook: " + formatInterval(estimatedTime));

			Map<Pair<SetExpansionStrategy, String>, ConcurrentAccumulator<PipelinePerformance>> performanceMap = new LinkedHashMap<>();

			GenericResultsWriter<PipelinePerformance> writer = new GenericResultsWriter<>(fusion.runner.getColumnListforPipelinePerformance());

			for (Entry<ClassId, Map<SetExpansionStrategy, Map<String, ConcurrentAccumulator<PipelinePerformance>>>> entry : performanceClassIdMap.entrySet()) {

				for (Entry<SetExpansionStrategy, Map<String, ConcurrentAccumulator<PipelinePerformance>>> innerEntry : entry.getValue().entrySet()) {
					for (Entry<String, ConcurrentAccumulator<PipelinePerformance>> innerInnerEntry : innerEntry.getValue().entrySet()) {
						ConcurrentAccumulator<PipelinePerformance> value = innerInnerEntry.getValue();
						PipelinePerformance pp = value.get();
						writer.writeClassIdWithLearner(innerEntry.getKey().getName(), innerInnerEntry.getKey(), entry.getKey(), pp);

						CHMHelper.getFromMap(performanceMap, new Pair<>(innerEntry.getKey(), innerInnerEntry.getKey()),
								() -> new ConcurrentAccumulator<>(new PipelinePerformance())).add(pp);

						TreeMap<String, List<String>> treeMap = CHMHelper.getFromMap(optimizationRetMap, "RC", () -> new TreeMap<>());
						List<String> list = CHMHelper.getFromMap(treeMap, innerEntry.getKey().getName() + "+" + innerInnerEntry.getKey(),
								() -> new ArrayList<>());
						list.add(DecimalFormatter.format3(pp.getClusteringAccuracy().getFMeasureMacro()));

					}
				}

			}

			for (Entry<Pair<SetExpansionStrategy, String>, ConcurrentAccumulator<PipelinePerformance>> strategy : performanceMap.entrySet()) {

				PipelinePerformance pp = strategy.getValue().get();
				writer.writeTotalWithLearner(strategy.getKey().getFirst().getName(), strategy.getKey().getSecond(), pp);
				TreeMap<String, List<String>> treeMap = CHMHelper.getFromMap(optimizationRetMap, "RC", () -> new TreeMap<>());
				List<String> list = CHMHelper.getFromMap(treeMap, strategy.getKey().getFirst().getName() + "+" + strategy.getKey().getSecond(),
						() -> new ArrayList<>());
				list.add("");
				list.add(DecimalFormatter.format3(pp.getClusteringAccuracy().getFMeasureMacro()));
			}

			System.out.println("AGGREGATED:, blocking: " + !alwaysDisableBlocking);
			System.out.println(writer.getAll(ResultsCategory.TOTAL_WITH_LEARNER));
			System.out.println();
			System.out.println(writer.getAll(ResultsCategory.CLASS_ID_WITH_LEARNER));

			System.out.println("ThisRunTook: " + formatInterval(estimatedTime));

		}

		if (run_HyperParameter) {
			SetExpansionStrategiesConfig finalsetExpansionStrategiesConfig = setExpansionStrategiesConfig;
			if (HyperParameter_NewDetection) {
				Map<UUID, Pair<EntityMatcherFullResults, Map<RowClusterUUID, Optional<InstanceId>>>> em_cache_map = new HashMap<>();

				Map<ClassId, Set<PropertyId>> nonTailProperties = setExtension.getNonTailProperties();
				for (ValidationRun validationRun : validationRuns) {

					em_cache_map.put(validationRun.getUuid(), fusion.entityMatchingGS_ClustererWeightsAdapterAndLearner.getFullResults(
							EntityMatchingRunRepository.ET_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREENEW, validationRun.getGs(), fusionConfig,
							finalsetExpansionStrategiesConfig, strategiesConfig, validationRun.getFullInput(), validationRun.getCandidates(), thresholdsFolder,
							validationRun.getClassId(), allParents, validationRun.getNewEntities(), nonTailProperties, validationRun.getFullInputAllClasses()));
				}

				HyperparameterOptimizer.RandomForestRun run = (validationRun, parameters) -> {

					EntityMatcherWekaLearner wekaLearner = new EntityMatcherWekaLearner(
							Sets.newHashSet(EntityMatcherStrategies.ATTRIBUTE, EntityMatcherStrategies.IMPLICIT_ATTRIBUTE), "RandomForest",
							parameters.getParameters(), parameters.isNumeric(), null, parameters.isDescretize(), false, new WeightsRepositoryImpl(), true);
					Pair<EntityMatcherFullResults, Map<RowClusterUUID, Optional<InstanceId>>> pair = em_cache_map.get(validationRun.getUuid());

					double error = fusion.entityMatchingGS_ClustererWeightsAdapterAndLearner.errorRun(pair.getFirst(), pair.getSecond(),
							EntityMatchingRunRepository.ET_LABEL_LABELSF_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREENEW, wekaLearner, validationRun.getNewEntities());
					return error;

				};

				System.out.println("EntityMatchingHyperParameters: ");
				hyperparameterOptimizer.findHyperParameters_EM_2(validationRuns, run, false);

				// System.out.println("EntityMatchingHyperParameters: ");
				// hyperparameterOptimizer.findHyperParameters(validationRuns,
				// run, false);

			}

			if (HyperParameter_Clustering) {
				Map<UUID, RowClusteringLearnableSimCache> clusteringCache = new HashMap<>();

				for (ValidationRun validationRun : validationRuns) {
					RowClusteringLearnableSimCache learnCache = new RowClusteringLearnableSimCache();

					SetExpansionStrategy strategy = new SetExpansionStrategy("WEKA_GENERIC", learnCache, true);
					SetExpansionStrategiesConfig newSetExpansionStrategiesConfig = finalsetExpansionStrategiesConfig.rebuilder().clearStrategies()
							.addStrategy(strategy).addStrategy(SetExpansionStrategiesRepository.COMBINED_BASE).build();
					Map<ClassId, Set<TableRowId>> learningInput = new HashMap<>();
					Map<ClassId, GoldStandardClusters> learningGoldStandard = new HashMap<>();
					Map<ClassId, TableRowIdsSet> fullInputInner = new HashMap<>();
					Map<ClassId, Set<InstanceId>> learningNewEntityIds = new HashMap<>();
					learningInput.put(validationRun.getClassId(), validationRun.getCandidates());
					learningGoldStandard.put(validationRun.getClassId(), validationRun.getGs());
					fullInputInner.put(validationRun.getClassId(), validationRun.getFullInput());
					learningNewEntityIds.put(validationRun.getClassId(), validationRun.getNewEntities());
					fusion.clusteringLearnerRunner.getAllWeights(new HashMap<>(), learningInput, fullInputInner, fusionConfig, newSetExpansionStrategiesConfig,
							strategiesConfig, thresholdsFolder, learningGoldStandard, learningNewEntityIds, conflictingSets, allParents,
							setExtension.getNonTailProperties(), validationRun.getFullInputAllClasses());

					clusteringCache.put(validationRun.getUuid(), learnCache);
				}

				HyperparameterOptimizer.RandomForestRun clusterRun = (validationRun, parameters) -> {

					RowClusteringLearnableSimCache simCache = clusteringCache.get(validationRun.getUuid());
					MT_WekaLearnedModel model = (MT_WekaLearnedModel) ((LearnableRowClusterSimilarities) wekaGeneric.get(parameters, "WEKA_GENERIC")
							.clusterRows(simCache.getConfig(), simCache.getInput())).learn(simCache.getPairs(), simCache.getClassId(),
									simCache.getEvaluationMethod());

					// SetExpansionStrategy strategy = new
					// SetExpansionStrategy("WEKA_GENERIC", , true);
					// SetExpansionStrategiesConfig
					// newSetExpansionStrategiesConfig
					// =
					// finalsetExpansionStrategiesConfig.rebuilder().clearStrategies()
					// .addStrategy(strategy).build();
					// Map<ClassId, Set<TableRowId>> learningInput = new
					// HashMap<>();
					// Map<ClassId, GoldStandardClusters> learningGoldStandard =
					// new
					// HashMap<>();
					// Map<ClassId, TableRowIdsSet> fullInputInner = new
					// HashMap<>();
					// Map<ClassId, Set<InstanceId>> learningNewEntityIds = new
					// HashMap<>();
					// learningInput.put(validationRun.getClassId(),
					// validationRun.getCandidates());
					// learningGoldStandard.put(validationRun.getClassId(),
					// validationRun.getGs());
					// fullInputInner.put(validationRun.getClassId(),
					// validationRun.getFullInput());
					// learningNewEntityIds.put(validationRun.getClassId(),
					// validationRun.getNewEntities());
					//
					// Map<ClassId, LearnedModelRepository> learnedModels =
					// fusion.clusteringLearnerRunner.getAllWeights(new
					// HashMap<>(),
					// learningInput,
					// fullInputInner, fusionConfig,
					// newSetExpansionStrategiesConfig,
					// strategiesConfig, thresholdsFolder, learningGoldStandard,
					// learningNewEntityIds, conflictingSets, allParents,
					// setExtension.getNonTailProperties());
					//
					// double error = ((MT_WekaLearnedModel)
					// learnedModels.get(validationRun.getClassId()).get(strategy.getOptimizableClassCluster())).getError();
					double error = model.getError();
					return error;

				};

				System.out.println("ClusterHyperParameters: ");
				hyperparameterOptimizer.findHyperParameters_cluster(validationRuns, clusterRun);
			}

		}

		return optimizationRetMap;

	}

	private void checkTables(Map<AnnotatedTableId, ClassId> tableClasses) {
		for (Entry<AnnotatedTableId, ClassId> tableEntry : tableClasses.entrySet()) {
			AnnotatedTable table = this.tableRepository.getTable(tableEntry.getKey());
			AnnotatedTableMetaData md = this.tableRepository.getMetaData(tableEntry.getKey());

			int sdf = 123;
		}

	}

	private static Map<ClassId, GoldStandardClusters> splitAndFilterGoldStandard(GoldStandardClusters mappings,
			Map<InstanceId, ClassId> instanceToClassMapping) {

		Map<ClassId, Map<TableRowId, InstanceId>> ret = new HashMap<>();

		for (Entry<TableRowId, InstanceId> entry : mappings.getIndex().entrySet()) {
			if (instanceToClassMapping.containsKey(entry.getValue())) {
				CHMHelper.getFromMap(ret, instanceToClassMapping.get(entry.getValue()), () -> new HashMap<>()).put(entry.getKey(), entry.getValue());
			}
		}

		return ret.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> new GoldStandardClusters(e.getValue())));
	}

	private static Pair<Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>>, Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>>> getFactsMap(
			Map<Pair<InstanceId, PropertyId>, Tri<String, Boolean, Boolean>> facts, Set<InstanceId> processed, Map<InstanceId, ClassId> instanceToClassMapping,
			Map<PropertyId, ValueType> typeMappings, Map<String, InstanceId> positionIndexNormalized) {

		Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>> ret = new HashMap<>();
		Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>> retFull = new HashMap<>();

		for (Entry<Pair<InstanceId, PropertyId>, Tri<String, Boolean, Boolean>> entry : facts.entrySet()) {
			if (entry.getKey().getSecond().equals(PropertiesExpansionData.LABEL_PROPERTY)) {
				continue;
			}

			InstanceId instanceId = entry.getKey().getFirst();
			ClassId classId = instanceToClassMapping.get(instanceId);
			PropertyId propertyId = entry.getKey().getSecond();
			ValueId valueId = new ValueId(instanceId, propertyId);
			QualifiedTripleId tripleId = new QualifiedTripleId(new TripleId(classId, propertyId, instanceId), new HashSet<>());

			List<QualifiedValueHolder> claims = new ArrayList<>();

			String[] values = entry.getValue().getA().split("\\|");

			for (String value : values) {
				ValueType vt = typeMappings.get(propertyId);

				switch (vt) {
				case BOOLEAN:
					break;
				case COORDINATE:
					break;
				case EXTERNAL_ID:
					break;
				case INSTANCE_REFERENCE:
					if (propertyId.getFinalPart().equals("position")) {
						InstanceId ao = positionIndexNormalized.get(value.toLowerCase());
						if (ao == null) {
							System.out.println(value);
							System.out.println(instanceId.getInstanceId());
							throw new IllegalStateException();
						}
						claims.add(new QualifiedAbstractObjectHolder(ao));
					} else {

						InstanceId ao = InstanceId.fromString(value);
						claims.add(new QualifiedAbstractObjectHolder(ao));
					}
					break;
				case PROPERTY_REFERENCE:
					break;
				case QUANTITY: {
					if (propertyId.equals(PropertyId.fromString("runtime"))) {
						String[] split = value.split(":");
						int sum = 0;
						if (split.length == 2) {
							sum = Integer.parseInt(split[0]) * 60 + Integer.parseInt(split[1]);

						} else if (split.length == 3) {
							sum = Integer.parseInt(split[0]) * 60 * 60 + Integer.parseInt(split[1]) * 60 + Integer.parseInt(split[2]);
						} else {
							throw new IllegalStateException();
						}
						QuantityObject ao = QuantityObject.fromDouble(sum);
						claims.add(new QualifiedAbstractObjectHolder(ao));
					} else if (propertyId.equals(PropertyId.fromString("height"))) {
						QuantityObject ao = QuantityObject.fromDouble(Double.parseDouble(value) / 100.0);
						claims.add(new QualifiedAbstractObjectHolder(ao));
					} else {
						QuantityObject ao = QuantityObject.fromDouble(Double.parseDouble(value));
						claims.add(new QualifiedAbstractObjectHolder(ao));
					}
				}
					break;
				case CATEGORICAL_INT: {
					CategoricalIntObject ao = new CategoricalIntObject(Integer.parseInt(value));
					claims.add(new QualifiedAbstractObjectHolder(ao));
				}

					break;
				case STRING: {
					StringObject ao = new StringObject(value);
					claims.add(new QualifiedAbstractObjectHolder(ao));

				}
					break;
				case TIME: {
					TimeObject ao = transformTimeObject(value);
					claims.add(new QualifiedAbstractObjectHolder(ao));
				}
					break;
				case URL:
					break;
				default:
					throw new IllegalStateException();

				}
			}

			if (!processed.contains(entry.getKey().getFirst()) || !entry.getValue().getB() || entry.getValue().getC()) {
				Map<QualifiedTripleId, QualifiedValueList> replaceValues = CHMHelper.getFromMap(retFull, classId, () -> new HashMap<>());
				if (!claims.isEmpty()) {

					QualifiedValueList qvl = new QualifiedValueList(claims);
					replaceValues.put(tripleId, qvl);

				}
			} else {
				Map<QualifiedTripleId, QualifiedValueList> replaceValues = CHMHelper.getFromMap(ret, classId, () -> new HashMap<>());
				if (!claims.isEmpty()) {

					QualifiedValueList qvl = new QualifiedValueList(claims);
					replaceValues.put(tripleId, qvl);

				}
			}

		}

		return new Pair<>(ret, retFull);

	}

	private static TimeObject transformTimeObject(String value) {
		if (value.length() == 4) {
			return new TimeObject(Long.parseLong(value), 0, 0, 0, 0, 0, 0, Precision.T09_years);
		} else if (value.length() == 10) {
			long year = Long.parseLong(value.substring(0, 4));
			int month = Integer.parseInt(value.substring(5, 7));
			int day = Integer.parseInt(value.substring(8, 10));
			return new TimeObject(year, month, day, 0, 0, 0, 0, Precision.T11_days);
		}
		throw new IllegalStateException();
	}

	public static String formatInterval(final long nano) {
		final long l = TimeUnit.NANOSECONDS.toMillis(nano);

		final long hr = TimeUnit.MILLISECONDS.toHours(l);
		final long min = TimeUnit.MILLISECONDS.toMinutes(l - TimeUnit.HOURS.toMillis(hr));
		final long sec = TimeUnit.MILLISECONDS.toSeconds(l - TimeUnit.HOURS.toMillis(hr) - TimeUnit.MINUTES.toMillis(min));
		final long ms = TimeUnit.MILLISECONDS.toMillis(l - TimeUnit.HOURS.toMillis(hr) - TimeUnit.MINUTES.toMillis(min) - TimeUnit.SECONDS.toMillis(sec));
		return String.format("%02d:%02d:%02d.%03d", hr, min, sec, ms);
	}

	@Inject
	protected LearnedModelAdapter learnedModelAdapter;

	@Inject
	protected SetExpansionRunner runner;

	@Inject
	protected expansion.set.exportimport.ClusterPairsExporter ClusterPairsExporter;
	@Inject
	protected GoldStandardSplitter goldStandardSplitter;

	@Inject
	protected AnnotatedTablesRepository tableRepository;

	@Inject
	protected WeightsAdapter weightsAdapter;

	@Inject
	protected HyperparameterOptimizer hyperparameterOptimizer;
	// @Inject
	// protected LearnerRunner weightsLearnerRunner;

	@Inject
	protected ClusteringModelLearnerRunner clusteringLearnerRunner;

	@Inject
	protected AttributeMappingsEvaluator evaluator;

	@Inject
	protected EntityMatchingModelLearner entityMatchingGS_ClustererWeightsAdapterAndLearner;

	// @Inject
	// protected EntityMatchingGS_ClustererWeightsAdapterAndLearner
	// oldEntityMatchingWeightsLearner;

	@Inject
	protected AttributeMatchingAdapter adapter;

	@Inject
	protected CombinedWekaGeneric wekaGeneric;

	@Inject
	protected RC_Average_Creator rcAverageCreator;

	@Inject
	protected ND_Average_Creator ndAverageCreator;

	// @Inject
	// protected BootStrapStrategyCreator bootStrapCreator;

	@Inject
	protected DS_ClusteringRunnerCreator DS_ClusteringRunnerCreator;

	@Inject
	protected GenericStrategyCreator genericStrategyCreator;

	@Inject
	protected WS_RC_Strategy_Creator WS_RC_Strategy_Creator;

	@Inject
	protected WS_RC_No_Clustering_Strategy_Creator WS_RC_No_Clustering_Strategy_Creator;

	@Inject
	protected expansion.set.distant_supervision.bootsrap.WS_ND_Strategy_Creator WS_ND_Strategy_Creator;

	@Inject
	protected RC_MatchingRules_Creator RC_MatchingRules_Creator;

	@Inject
	protected ND_MatchingRules_Creator ND_MatchingRules_Creator;

	public GSRun() {

	}

}
