package de.uni_mannheim.informatik.dws.expansion.samples.setExtension;

import java.io.File;
import java.io.IOException;
import java.util.Map;
import java.util.Set;

import javax.inject.Inject;

import dagger.ObjectGraph;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.reader.XMLOntologyReader;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyConverter;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepositoryImpl;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepositoryImplMemory;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.matching.dbpedia.algorithm.KeyIndex;
import de.mannheim.uni.tableprocessor.model.TableColumn.ColumnDataType;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBaseCryoAdapter;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBaseImpl;
import de.uni_mannheim.dws.expansion.data.ontology.model.ValueConverter;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeKryo;
import de.uni_mannheim.informatik.dws.expansion.matching.KnowledgeBaseModule;
import de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.ConvertDBpedia;
import de.uni_mannheim.informatik.dws.expansion.utils.remote.RepositoryLoading;

public class KbOntologyAndTablesAdapter {
	public static void main(String[] args) throws IOException {

		Map<ClassId, Set<ClassId>> topClasses = ExtractTablesForGoldStandardRun.getClassesWithHierarchyFilteredOrNot();
		// topClasses.remove(ClassConstants.single);
		// topClasses.remove(ClassConstants.placeClass);
		// topClasses.add(ClassConstants.placeClass);
		// topClasses.add(ClassConstants.person);
		// topClasses.add(ClassConstants.populatedPlace);

		if (!args[3].equals("true")) {

			String dataHome = null;

			if (args.length > 0 && args[0] != null && !args[0].equals("")) {
				dataHome = args[0];
			}

			String csvDBpedia = dataHome + "/rawDBpediaDatasets/" + "DBpediaAsTables/";
			String ontologyOwlFile = dataHome + "/rawDBpediaDatasets/" + "dbpedia_2014.owl";
			XMLOntologyReader ontologyReader = new XMLOntologyReader();
			OntologyConverter ontologyConverter = new OntologyConverter(new ValueConverter());

			OntologyRepositoryImplMemory ontologyPreliminary = ontologyReader.run(ontologyOwlFile);

			Pair<KnowledgeBase, Map<ClassPropertyId, ColumnDataType>> pair = ConvertDBpedia.loadFromCSVfiles(csvDBpedia, ontologyPreliminary);
			KnowledgeBase kb = pair.getFirst();
			Map<ClassPropertyId, ColumnDataType> columnDataTypes = pair.getSecond();

			OntologyRepositoryImpl ontologyRepository = ontologyConverter.convert(kb, ontologyPreliminary, columnDataTypes);

			SetExtension setExtension = new SetExtension(args, topClasses.keySet(), kb, ontologyRepository,
					ExtractTablesForGoldStandardRun.getMinPropertyDensity(), RepositoryLoading.LOCAL_ONLY);

			KbAdapter kbAdapter = setExtension.getGraph().get(KbAdapter.class);

			String adaptedKbName = setExtension.getKbName() + "Adapted";
			kbAdapter.run(setExtension.getKbsHome() + "/" + adaptedKbName + "/");

			args[1] = adaptedKbName;
			args[3] = "true";
			if (args.length > 5 && args[5].equals("true")) {

				try {
					PersistentParallel4.close();
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				return;

			}
		}

		// int currentRun = 0;
		// if (args.length > 6) {
		// currentRun = Integer.parseInt(args[6]);
		// }
		//
		// boolean alwaysDisableBlocking = false;
		// SetExtension setExtension;
		// if (args.length > 4 && args[4].equals("true")) {
		// setExtension = new SetExtension(args, topClasses.keySet(),
		// ExtractTablesForGoldStandardRun.getMinPropertyDensity(),
		// RepositoryLoading.REMOTE_LOAD_REMOTELY, currentRun,
		// alwaysDisableBlocking);
		// } else {
		// setExtension = new SetExtension(args, topClasses.keySet(),
		// ExtractTablesForGoldStandardRun.getMinPropertyDensity(),
		// RepositoryLoading.REMOTE_LOAD_REMOTELY);
		// }
		//
		// String adaptedCacheFolder = setExtension.getMappingsFolder() +
		// "/cacheAdapted" + (currentRun + 1) + "/";
		//
		// AnnotatedTablesWriterConcurrent annotatedTablesWriter = new
		// AnnotatedTablesWriterConcurrent(adaptedCacheFolder, null);
		// if (!annotatedTablesWriter.cacheExists()) {
		// AnnotatedTablesRepository tableRepository =
		// setExtension.getGraph().get(AnnotatedTablesRepository.class);
		//
		// TablesAdapter tableAdapter =
		// setExtension.getGraph().get(TablesAdapter.class);
		//
		// String clusteringsWeightsFolderPrefix =
		// setExtension.getMappingsCachesFolder() +
		// "/models/clusteringModelsFull";
		//
		// String instanceMatchingWeightsFolderPrefix =
		// setExtension.getMappingsCachesFolder() +
		// "/models/instanceMatchingModelsFull";
		//
		// Map<ClassId, Set<PropertyId>> nonTailProperties = new
		// HashMap<>(setExtension.getNonTailProperties());
		//
		// boolean secondary = true;
		// if (args.length > 7 && args[7].equals("false")) {
		// secondary = false;
		// }
		//
		// boolean allTablesMatch = false;
		// if (args.length > 8 && args[8].equals("true")) {
		// allTablesMatch = true;
		// }
		//
		// boolean store = true;
		// if (args.length > 9 && args[9].equals("false")) {
		// store = false;
		// }
		//
		// if (args.length > 10) {
		//
		// int classToRun = Integer.parseInt(args[10]);
		// switch (classToRun) {
		// case 0:
		// topClasses.remove(ClassConstants.single);
		// topClasses.remove(ClassConstants.placeClass);
		// nonTailProperties.remove(ClassConstants.single);
		// nonTailProperties.remove(ClassConstants.placeClass);
		// break;
		// case 1:
		// topClasses.remove(ClassConstants.gridIronFootballPlayer);
		// topClasses.remove(ClassConstants.placeClass);
		// nonTailProperties.remove(ClassConstants.gridIronFootballPlayer);
		// nonTailProperties.remove(ClassConstants.placeClass);
		// break;
		// case 2:
		// topClasses.remove(ClassConstants.gridIronFootballPlayer);
		// topClasses.remove(ClassConstants.single);
		// nonTailProperties.remove(ClassConstants.gridIronFootballPlayer);
		// nonTailProperties.remove(ClassConstants.single);
		// break;
		// }
		// }
		//
		// tableAdapter.run(tableRepository, setExtension.getMappingsFolder(),
		// adaptedCacheFolder, nonTailProperties, topClasses,
		// clusteringsWeightsFolderPrefix, instanceMatchingWeightsFolderPrefix,
		// setExtension.getMappingsCachesFolder(), currentRun, secondary,
		// allTablesMatch, store);
		//
		// }

		try {
			PersistentParallel4.close();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public static class KbAdapter {

		@Inject
		protected ThreadSafeKryo threadSafeKryo;

		@Inject
		protected KnowledgeBase kb;

		@Inject
		protected OntologyRepository ontologyRepository;

		@Inject
		protected KnowledgeBaseCryoAdapter kryoAdapter;

		public KbAdapter() {
			// TODO Auto-generated constructor stub
		}

		void run(String adaptedKBFolder) throws IOException {

			MKDIR.createDirUnchecked(new File(adaptedKBFolder));
			String kbPath = adaptedKBFolder + "/kb/";
			String ontologyRepositoryFile = adaptedKBFolder + "/ontology.kryo";
			String luceneIndexCache = adaptedKBFolder + "/lucene/";

			MKDIR.createDirUnchecked(new File(kbPath));

			threadSafeKryo.write(ontologyRepository, ontologyRepositoryFile);

			MKDIR.createDirUnchecked(new File(luceneIndexCache));

			kryoAdapter.write(kbPath, (KnowledgeBaseImpl) kb);

			KnowledgeBaseModule kbModule = new KnowledgeBaseModule(kbPath, ontologyRepositoryFile, luceneIndexCache);

			ObjectGraph graph = ObjectGraph.create(kbModule);

			KeyIndex keyIndex = graph.get(KeyIndex.class);
		}
	}

	// public static class TablesAdapter {
	//
	// @Inject
	// protected WeightsAdapter weightsAdapter;
	//
	// @Inject
	// protected AttributeThresholdLearner thresholdLearner;
	//
	// @Inject
	// protected AttributeMatcher matcher;
	//
	// @Inject
	// protected ConcurrentKryoWriter concurrentKryoWriter;
	//
	// @Inject
	// protected AnnotatedTableFilterRunner filterRunner;
	//
	// @Inject
	// protected AttributeMatchingPipeline pipeline;
	//
	// @Inject
	// protected KnowledgeBase kb;
	//
	// @Inject
	// protected AttributeMappingsEvaluator evaluator;
	//
	// @Inject
	// protected CHMKryoAdapter chmKryoAdapter;
	//
	// public TablesAdapter() {
	// // TODO Auto-generated constructor stub
	// }
	//
	// public void run(AnnotatedTablesRepository repository, String
	// mappingsFolder, String adaptedTablesCacheFolder,
	// Map<ClassId, Set<PropertyId>> nonTailProperties, Map<ClassId,
	// Set<ClassId>> allParents, String clusteringWeightsFolderPrefix,
	// String instanceMatchingWeightsFolderPrefix, String cachesFolder, int
	// currentRun, boolean secondary, boolean allTablesMatch, boolean store)
	// throws IOException {
	//
	// AnnotatedTablesWriterConcurrent annotatedTablesWriter = new
	// AnnotatedTablesWriterConcurrent(adaptedTablesCacheFolder,
	// concurrentKryoWriter);
	// if (!annotatedTablesWriter.cacheExists()) {
	//
	// if (true) {
	//
	// String attributeMappingsCacheFolder = cachesFolder +
	// "/AttributeMappingsCacheFolder/";
	// MKDIR.createDirUnchecked(new File(attributeMappingsCacheFolder));
	//
	// boolean hasCache = false;
	// for (ClassId classId : allParents.keySet()) {
	// if (chmKryoAdapter.cacheExists(attributeMappingsCacheFolder,
	// getPrefix(classId))) {
	// hasCache = true;
	// }
	// }
	//
	// AttributeMappings newMappings;
	//
	// GsOutputReader reader = new GsOutputReader(InstanceIdStyle.DBPEDIA_URI);
	// String gsOutputFolder = mappingsFolder + "/gsOutput/";
	//
	// String tableClassesFile = gsOutputFolder + "/classes_GS.csv";
	// String processedTablesFile = gsOutputFolder + "/processedTables.csv";
	// Set<AnnotatedTableId> processedTables =
	// reader.readTableList(processedTablesFile);
	// Map<AnnotatedTableId, ClassId> tableClasses =
	// reader.readTableClasses(tableClassesFile);
	//
	// processedTables = processedTables.stream().filter((e) ->
	// allParents.keySet().contains(tableClasses.get(e))).collect(Collectors.toSet());
	//
	// AttributeMappings gsMappings = reader.getAttributeMappings(gsOutputFolder
	// + "/attributes/", processedTables, tableClasses,
	// nonTailProperties);
	//
	// List<AnnotatedTablesReaderData> reMatch = new ArrayList<>();
	// Set<AnnotatedTableId> reMatchIds = new HashSet<>();
	// List<AnnotatedTablesReaderData> keep = new ArrayList<>();
	//
	// for (AnnotatedTablesReaderData data : repository.getTableReaderData()) {
	// // if
	// //
	// (data.getMetaData().getDbPediaClass().equals(ClassConstants.placeClass))
	// // {
	// // keep.add(data);
	// // } else
	// if
	// (data.getId().getTableName().equals("1346981172231_1347011264447_3050.arc9167679534671747539#2935814_0_7628926464279511553"))
	// {
	// int sdf = 123;
	// if (processedTables.contains(data.getId())) {
	// int sggggdf = 123;
	// }
	// }
	//
	// if (allParents.keySet().contains(data.getMetaData().getDbPediaClass())) {
	//
	// if (allTablesMatch) {
	// reMatch.add(data);
	// reMatchIds.add(data.getId());
	// } else {
	// if (!gsMappings.tableSet().contains(data.getId())) {
	// keep.add(data);
	// } else {
	// reMatch.add(data);
	// reMatchIds.add(data.getId());
	// }
	// }
	// }
	// }
	//
	// if (!hasCache) {
	//
	// String forLearningTablesFile = gsOutputFolder + "/forLearningTables.csv";
	// Set<AnnotatedTableId> forLearningTables =
	// reader.readTableList(forLearningTablesFile);
	//
	// AttributeMappings forLearningMappingsNonFinal =
	// reader.getAttributeMappings(gsOutputFolder + "/attributes/",
	// forLearningTables,
	// tableClasses, nonTailProperties);
	// AttributeMappings forLearningMappings = gsMappings;
	//
	// if (currentRun >= 0) {
	//
	// FusionConfig fusionConfig = SetExtension.getFusionConfig();
	//
	// SetExpansionStrategiesConfigBuilder builder =
	// SetExtension.getSetExpansionStrategiesConfig().rebuilder();
	//
	// builder.clearFusionSetups().addFusionSetup(FusionStrategyRepository.COLUMN_SCORES,
	// PropertiesConfig.builder().setTimeDependence(TimeDependence.NO).build());
	// SetExpansionStrategiesConfig setExpansionStrategiesConfig =
	// builder.build();
	//
	// StrategiesConfig strategiesConfig = StrategiesConfig.defaultConfig();
	//
	// String matchingPipelineCachesFolder = cachesFolder +
	// "matchingPipelineSecondary";
	// MKDIR.createDirUnchecked(new File(matchingPipelineCachesFolder));
	//
	// String matchingPipelineWeightsCachesFolder = cachesFolder +
	// "matchingPipelineWeights";
	// MKDIR.createDirUnchecked(new File(matchingPipelineWeightsCachesFolder));
	//
	// SortedMap<Integer, AttributeMappings> pipelineMappings =
	// pipeline.runPipeline(setExpansionStrategiesConfig, fusionConfig,
	// strategiesConfig, reMatchIds, allParents, nonTailProperties, currentRun,
	// forLearningMappings, gsMappings,
	// clusteringWeightsFolderPrefix, instanceMatchingWeightsFolderPrefix,
	// matchingPipelineCachesFolder,
	// matchingPipelineWeightsCachesFolder, secondary, false, false);
	//
	// newMappings = pipelineMappings.get(currentRun);
	// FusionPerformance<AttributeMappingsPerformance> newMappingsPeformance =
	// evaluator.evaluate(gsMappings, newMappings,
	// kb.getNormalizationIndex());
	//
	// // FusionPerformance<AttributeMappingsPerformance>
	// // pipelinePerformance1 =
	// // evaluator.evaluate(gsMappings,
	// // pipelineMappings.get(1),
	// // kb.getNormalizationIndex());
	//
	// ColumnReMatcherRun.print("PIPELINE_" + currentRun,
	// newMappingsPeformance);
	// // ColumnReMatcherRun.print("PIPELINE_1",
	// // pipelinePerformance1);
	//
	// } else {
	// newMappings = new AttributeMappings(
	// reMatchIds.stream().collect(Collectors.toMap(Function.identity(), (e) ->
	// new TableAttributeMappings(new HashMap<>()))),
	// gsMappings.getClassIds().entrySet().stream().filter((e) ->
	// reMatchIds.contains(e.getKey()))
	// .collect(Collectors.toMap(Entry::getKey, Entry::getValue)));
	// }
	//
	// } else {
	// newMappings = readCache(attributeMappingsCacheFolder,
	// allParents.keySet());
	//
	// FusionPerformance<AttributeMappingsPerformance> newMappingsPeformance =
	// evaluator.evaluate(gsMappings, newMappings,
	// kb.getNormalizationIndex());
	//
	// // FusionPerformance<AttributeMappingsPerformance>
	// // pipelinePerformance1 =
	// // evaluator.evaluate(gsMappings,
	// // pipelineMappings.get(1),
	// // kb.getNormalizationIndex());
	//
	// ColumnReMatcherRun.print("PIPELINE_" + currentRun,
	// newMappingsPeformance);
	// }
	//
	// if (store) {
	// List<AnnotatedTablesReaderData> finalList = new
	// ArrayList<>(filterRunner.asList(reMatch, new
	// AttributeMappingsFilter(newMappings)));
	// finalList.addAll(keep);
	//
	// MKDIR.createDirUnchecked(new File(adaptedTablesCacheFolder));
	//
	// annotatedTablesWriter.writeAll(finalList);
	// }
	//
	// // ThresholdContainer thresholds = new
	// // ThresholdContainerWithFallback(
	// //
	// thresholdLearner.findThreshold(repository.getTableReaderData(gsMappings.tableSet()),
	// // nonTailProperties, gsMappings, 4, (tableId) -> {
	// // return gsMappings.getClassId(tableId);
	// // }), new ThresholdConstant(0.25));
	// //
	// // System.out.println(thresholds.toString());
	// //
	// // AttributeMappings yaserThresholdMappings =
	// // matcher.match(reMatch,
	// // thresholds, nonTailProperties);
	// if (!hasCache && !store) {
	//
	// for (ClassId classId : allParents.keySet()) {
	// Map<AnnotatedTableId, TableAttributeMappings> forClass =
	// AttributeMatchingWeightLearner.getForClass(newMappings, classId);
	// writeCache(attributeMappingsCacheFolder, classId, forClass);
	//
	// }
	// }
	//
	// } else {
	// annotatedTablesWriter.writeAll(repository.getTableReaderData());
	//
	// }
	// }
	//
	// }
	//
	// private void writeCache(String attributeMappingsCacheFolder, ClassId
	// classId, Map<AnnotatedTableId, TableAttributeMappings> forClass) {
	// Map<AnnotatedTableId, Map<ColumnIdx, PropertyMapping>> remapped =
	// forClass.entrySet().stream().filter((e) ->
	// !e.getValue().getMappings().isEmpty())
	// .collect(Collectors.toMap(Entry::getKey, (e) ->
	// e.getValue().getMappings()));
	// chmKryoAdapter.writeWithMap(attributeMappingsCacheFolder,
	// getPrefix(classId), remapped);
	//
	// }
	//
	// private AttributeMappings readCache(String attributeMappingsCacheFolder,
	// Set<ClassId> classIds) {
	// Map<AnnotatedTableId, TableAttributeMappings> map = new HashMap<>();
	// Map<AnnotatedTableId, ClassId> classMap = new HashMap<>();
	// for (ClassId classId : classIds) {
	// Map<AnnotatedTableId, Map<ColumnIdx, PropertyMapping>> read =
	// chmKryoAdapter.readWithMap(AnnotatedTableId.class, ColumnIdx.class,
	// PropertyMapping.class, attributeMappingsCacheFolder, getPrefix(classId));
	//
	// for (Entry<AnnotatedTableId, Map<ColumnIdx, PropertyMapping>> entry :
	// read.entrySet()) {
	// map.put(entry.getKey(), new TableAttributeMappings(entry.getValue()));
	// classMap.put(entry.getKey(), classId);
	// }
	// }
	// return new AttributeMappings(map, classMap);
	// }
	//
	// private String getPrefix(ClassId classId) {
	// return classId.getFinalPart() + "_";
	// }
	// }

}
