package de.uni_mannheim.informatik.dws.expansion.samples.setExtension;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

import javax.inject.Inject;

import org.apache.commons.io.FileUtils;
import org.json.JSONObject;

import com.google.common.collect.ImmutableMap;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategy;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategyRepository;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.MaxDiffStrategy;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.ScorerRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.uni_mannheim.informatik.dws.expansion.wikidata.WikidataClasses;

public class MaxDiffSetter {

	private final ScorerRepository scorerRepository;

	@Inject
	public MaxDiffSetter(ScorerRepository scorerRepository) {
		this.scorerRepository = scorerRepository;

	}

	public void run(String dataHome) throws IOException {
		String maxDiffsFolder = dataHome + "/" + "maxDiffs/";
		String maxDiffsFileQKBTMapped = dataHome + "/" + "maxDiffs_qkbt_mapped.json";
		String maxDiffsFileNewTS = dataHome + "/" + "maxDiffs_new_ts.json";
		String maxDiffsFileNewTSPropgated = dataHome + "/" + "maxDiffs_new_ts_propagated.json";

		MKDIR.createDirBoolean(new File(maxDiffsFolder));

		if (true && false) {
			System.out.println("Running with MaxDiff per Property Non-Adjusted");
			this.loadMaxDiffs("7", maxDiffsFolder, true, false);
			// fusion.runner.run(runName, fusionConfig, strategiesConfig,
			// setExpansionStrategiesConfig, folderPath);

		}

		System.out.println("Running with MaxDiff per Property Adjusted");
		this.loadMaxDiffs("7", maxDiffsFolder, true, true);
	}

	private void loadMaxDiffs(String maxDiffPrefix, String maxDiffFolder, boolean maxDiffPerProperty, boolean adjusted) throws IOException {
		this.setMaxDiffFromFile(FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL,
				maxDiffFolder + "/" + getMaxDiffFileName(maxDiffPrefix, FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL, maxDiffPerProperty, adjusted));

		//
		this.setMaxDiffFromFile(FusionStrategyRepository.QKBT_Mapped,
				maxDiffFolder + "/" + getMaxDiffFileName(maxDiffPrefix, FusionStrategyRepository.QKBT_Mapped, maxDiffPerProperty, adjusted));

		// this.setMaxDiffFromFile(FusionStrategyRepository.NEW_TS_QKBT_WITHOUR_ROW_CELL_PROPAGATED,
		// maxDiffFolder + "/"
		// + getMaxDiffFileName(maxDiffPrefix,
		// FusionStrategyRepository.NEW_TS_QKBT_WITHOUR_ROW_CELL_PROPAGATED,
		// maxDiffPerProperty, adjusted));

	}

	private void setMaxDiffFromInteger(Integer maxDiff) throws IOException {
		this.setMaxDiff(FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL, maxDiff);

		//
		this.setMaxDiff(FusionStrategyRepository.QKBT_Mapped, maxDiff);

		// this.setMaxDiff(FusionStrategyRepository.NEW_TS_QKBT_WITHOUR_ROW_CELL_PROPAGATED,
		// maxDiff);
	}

	private void setMaxDiffFromFile(FusionStrategy fusionStrategy, String maxDiffsFile) throws IOException {
		String maxDiffsJson = FileUtils.readFileToString(new File(maxDiffsFile));

		Map<ClassPropertyId, Integer> maxDiffs = jsonToMaxDiff(maxDiffsJson);

		if (maxDiffs.containsKey(this.allClasProperty)) {
			setMaxDiff(fusionStrategy, maxDiffs.get(this.allClasProperty));
		} else {
			setMaxDiff(fusionStrategy, maxDiffs);
		}

	}

	private void setMaxDiff(FusionStrategy strategy, Integer maxDiff) {
		Scorer scorer = this.scorerRepository.get(strategy.getScoringMethod());
		if (scorer instanceof MaxDiffStrategy) {
			MaxDiffStrategy scorerTyped = (MaxDiffStrategy) scorer;
			scorerTyped.setMaxDiffsContainer((classPropertyId) -> {
				return Optional.of(maxDiff);

			});
		}

	}

	// private void storeMappings(String file) {
	// Scorer scorer =
	// this.scorerRepository.get(FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL.getScoringMethod());
	//// FusionStrategyId fusionStrategyId =
	// FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL.getFusionStrategyId();
	//
	// }
	private void setMaxDiff(FusionStrategy strategy, Map<ClassPropertyId, Integer> maxDiffs) {
		Scorer scorer = this.scorerRepository.get(strategy.getScoringMethod());
		if (scorer instanceof MaxDiffStrategy) {
			MaxDiffStrategy scorerTyped = (MaxDiffStrategy) scorer;
			final Map<ClassPropertyId, Integer> maxDiffsFinal = ImmutableMap.copyOf(maxDiffs);
			scorerTyped.setMaxDiffsContainer((classPropertyId) -> {
				Integer maxDiff = maxDiffsFinal.get(classPropertyId);
				if (maxDiff != null) {
					return Optional.of(maxDiff);
				}
				return Optional.empty();

			});
		}

	}

	private static final ClassPropertyId allClasProperty = new ClassPropertyId(WikidataClasses.wikidataAll, new PropertyId("ALL"));

	private String maxDiffToJson(Map<ClassPropertyId, Integer> genericMap) {
		Map<ClassId, JSONObject> classObjects = new HashMap<>();
		for (Entry<ClassPropertyId, Integer> e : genericMap.entrySet()) {
			CHMHelper.getFromMap(classObjects, e.getKey().getClassId(), () -> new JSONObject()).put(e.getKey().getPropertyId().toCombinedString(),
					e.getValue());
		}

		JSONObject finalObject = new JSONObject();
		for (Entry<ClassId, JSONObject> e : classObjects.entrySet()) {
			finalObject.put(e.getKey().getClassId(), e.getValue());
		}

		return finalObject.toString();
	}

	private Map<ClassPropertyId, Integer> jsonToMaxDiff(String json) {
		Map<ClassPropertyId, Integer> ret = new HashMap<>();

		JSONObject completeObject = new JSONObject(json);

		for (Iterator<String> keys = completeObject.keys(); keys.hasNext();) {
			String key = keys.next();
			ClassId classId = new ClassId(key);
			JSONObject classObject = completeObject.getJSONObject(key);

			for (Iterator<String> innerKeys = classObject.keys(); innerKeys.hasNext();) {
				String propertyKey = innerKeys.next();
				PropertyId propertyId = PropertyId.fromCombinedString(propertyKey);
				ret.put(new ClassPropertyId(classId, propertyId), classObject.getInt(propertyKey));
			}

		}

		return ret;
	}

	public String getMaxDiffFileName(String maxDiffPrefix, FusionStrategy fusionStrategy, boolean maxDiffPerProperty, boolean adjusted) {
		String strategy = null;
		if (fusionStrategy.getFusionStrategyId().equals(FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL.getFusionStrategyId())) {
			strategy = "ts_mapping";
		} else if (fusionStrategy.getFusionStrategyId().equals(FusionStrategyRepository.NEW_TS_QKBT_WITHOUR_ROW_CELL_PROPAGATED.getFusionStrategyId())) {
			strategy = "ts_mapping_propagated";
		} else if (fusionStrategy.getFusionStrategyId().equals(FusionStrategyRepository.QKBT_Mapped.getFusionStrategyId())) {
			strategy = "timed_kbt";
		}

		if (strategy != null) {
			return "maxDiff_" + maxDiffPrefix + "_" + strategy + "_" + (adjusted ? "adjusted" : "default") + "_"
					+ (maxDiffPerProperty ? "perProperty" : "general") + ".json";
		} else {
			return null;
		}

	}
}
