package de.uni_mannheim.informatik.dws.expansion.samples.setExtension;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;

import javax.inject.Inject;

import com.google.common.collect.ImmutableMap;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.StringSimilarityJaccard;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.QuantityObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;

public class NumericKeyFilter implements AnnotatedTableFilter {

	private final KnowledgeBase knowledgeBase;
	private final SimilarityCalculator similarityCalculator;

	private final StringSimilarityJaccard sim = new StringSimilarityJaccard();

	@Inject
	public NumericKeyFilter(KnowledgeBase kb, SimilarityCalculator similarityCalculator) {
		this.knowledgeBase = kb;
		this.similarityCalculator = similarityCalculator;
	}

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {
		ImmutableMap<RowIdx, AbstractObject> keyColumn = data.getTableData().getKeyColumn();

		ValueType dataType = data.getTableData().getDataType(data.getTableData().getKeyColumnIndex());
		if (dataType == ValueType.QUANTITY) {
			int sdf = 123;
		}
		double count = 0.0;
		int subtract = 0;
		int total = 0;

		List<Pair<String, InstanceId>> list = new ArrayList<>();

		Map<Integer, InstanceId> instances = data.getMetaData().getInstances();

		for (Entry<RowIdx, AbstractObject> entry : keyColumn.entrySet()) {
			if (entry.getValue() instanceof QuantityObject) {
				int sdf = 123;
			}
			if (entry.getValue() instanceof StringObject) {

				total++;

				String s = ((StringObject) entry.getValue()).getString();
				String sNormalized = StringNormalizer.normaliseValue(s, false);
				String alphanum = sNormalized.replaceAll("[^a-zA-Z0-9]", "");
				String num = sNormalized.replaceAll("[^0-9]", "");
				if (((double) num.length()) / alphanum.length() > 0.65) {
					count++;
					if (instances.containsKey(entry.getKey().getRowIdx())) {
						InstanceId instanceId = instances.get(entry.getKey().getRowIdx());
						list.add(new Pair<>(s, instanceId));

						Set<String> labels = knowledgeBase.getLabels(instanceId);
						if (labels.size() > 0) {

							for (String lbl : labels) {
								if (sim.getLevenstein(((StringObject) entry.getValue()).getString(), lbl) > 0.7) {
									subtract++;
									break;
								}
							}

						}

					}
				}

			}

		}

		double prop = (count - subtract) / total;

		if (prop > 0.7) {
			return Optional.empty();
		} else {

			Set<InstanceId> badSet = new HashSet<>();
			badSet.add(new InstanceId("http://dbpedia.org/resource/Live_1.0"));
			badSet.add(new InstanceId("http://dbpedia.org/resource/3.2.1."));
			badSet.add(new InstanceId("http://dbpedia.org/resource/Space_Age_1.0"));
			for (Entry<Integer, InstanceId> instanceEntry : instances.entrySet()) {
				if (badSet.contains(instanceEntry.getValue())) {
					int sdfsdgagh = 1;
					break;
				}

			}

			return Optional.of(data);
		}
	}

}
