package de.uni_mannheim.informatik.dws.expansion.samples.setExtension;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentSkipListSet;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.inject.Named;
import javax.inject.Singleton;

import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.RandomStringUtils;

import com.google.common.collect.ImmutableSet;

import dagger.Module;
import dagger.ObjectGraph;
import dagger.Provides;
import de.mannheim.uni.expansion.data.dbpedia.PropertiesDBpediaTableFilter;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyAdapter;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator.LabelsGetter;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig.TimeDependence;
import de.mannheim.uni.expansion.tablefusion.modules.AbstractsModule;
import de.mannheim.uni.expansion.tablefusion.modules.DegreeModule;
import de.mannheim.uni.expansion.tablefusion.modules.HeidelTimeModule;
import de.mannheim.uni.expansion.tablefusion.modules.MatchingModule;
import de.mannheim.uni.expansion.tablefusion.modules.SimilarityModule;
import de.mannheim.uni.expansion.tablefusion.modules.TableFusionModule;
import de.mannheim.uni.expansion.tablefusion.modules.TablesModule;
import de.mannheim.uni.expansion.tablefusion.services.AnnotatedTableInstanceRescorer;
import de.mannheim.uni.expansion.tablefusion.services.AnnotatedTableOtsuFilter;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategyRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.units.UnitParser;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBaseImpl;
import de.uni_mannheim.dws.expansion.data.ontology.model.measurement.UnitOfMeasurement;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeKryo;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeNoReferenceKryo;
import de.uni_mannheim.informatik.dws.attribute.matcher.AttributeMatcher;
import de.uni_mannheim.informatik.dws.expansion.matching.KnowledgeBaseModule;
import de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.KbLabelRenamer;
import de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.NFLPropertiesDensity;
import de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.PopulationDensityFixer;
import de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.ProcessCSVTables;
import de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.SmallTableFilter;
import de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.musicalWork.DraftYearValueCreator;
import de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.musicalWork.FootBallCollegeValueCreator;
import de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.musicalWork.IsPartOfFixer;
import de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.musicalWork.MusicalArtistPropertyFix;
import de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.musicalWork.OntologicalFixes;
import de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.musicalWork.RuntimePropertyFixer;
import de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.musicalWork.TableColumnMapingsClearerer;
import de.uni_mannheim.informatik.dws.expansion.samples.dbpedia.musicalWork.WeightInPoundsValueCreator;
import de.uni_mannheim.informatik.dws.expansion.samples.postMatching.GoodPropertiesFilterAll;
import de.uni_mannheim.informatik.dws.expansion.samples.postMatching.KbInstanceFilter;
import de.uni_mannheim.informatik.dws.expansion.samples.postMatching.SourcesFilter;
import de.uni_mannheim.informatik.dws.expansion.samples.setExtension.KbOntologyAndTablesAdapter.KbAdapter;
import de.uni_mannheim.informatik.dws.expansion.samples.setExtension.TablesAdapterRun.TablesAdapter;
import de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.worldbank.CountryListFixer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.WeightsOptimizer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.genetic.WeightsGeneticParallelOptimizer;
import de.uni_mannheim.informatik.dws.expansion.utils.remote.RepositoryLoading;
import de.uni_mannheim.informatik.dws.matching.LabelMatcher;
import expansion.set.SetExpansionProperties;
import expansion.set.entitymatching.candidateselection.strategies.KnowledgeBase_Attribute_Indeces;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.evaluation.SetExpansionStrategiesConfig.SetExpansionStrategiesConfigBuilder;
import expansion.set.gscreation.ExtractTablesForGoldStandard2;
import expansion.set.gscreation.ExtractTablesForGoldStandard3;
import expansion.set.gscreation.ExtractTablesForGoldStandard4;
import expansion.set.profiling.PossibleConflictProfiler;
import expansion.set.profiling.SetExpansionSuitabilityProfiler;
import expansion.set.strategies.clustersplitters.unaryattributes.UnaryAttributeModule;
import expansion.set.strategies.rowclusterers.GreedyClusteringSplitter;
//import expansion.set.strategies.rowclusterers.cooccurence.InstanceSetScoresCreator;
//import expansion.set.strategies.rowclusterers.cooccurence.SetScoresCreator;

public class SetExtension {

	public static enum KbType {
		DEFAULT(), WIKIDATA("wikidata"), DBPEDIA();

		private static final Map<String, KbType> map = new HashMap<>();

		private final Set<String> namesSet = new HashSet<>();

		private KbType(String... names) {
			for (String name : names) {
				namesSet.add(name);
			}
		}

		static {
			for (KbType kbType : KbType.values()) {
				for (String name : kbType.namesSet) {
					map.put(name, kbType);
				}
			}
		}

		public static KbType getType(String name) {
			if (map.containsKey(name))
				return map.get(name);
			else
				return DEFAULT;
		}
	}

	@Module(injects = { ExtractTablesForGoldStandard4.class, ExtractTablesForGoldStandard3.class, TablesAdapter.class, KbAdapter.class,
			KbOntologyAndTablesAdapter.class, AnnotatedTableInstanceRescorer.class, GsRunRemoteKb.class, ColumnReMatcherRun.class, GSRun.class,
			SetExtensionProfiling.class, SetExtensionSecondProfiling.class, AnnotatedTableOtsuFilter.class, SimilarityCalculator.class, MaxDiffSetter.class,
			SetExtensionFindClasses.class, SetExpansionSuitabilityProfiler.class, PossibleConflictProfiler.class, ThreadSafeKryo.class,
			ThreadSafeNoReferenceKryo.class, ExtractTablesForGoldStandard2.class, FullRun.class, LabelMatcher.class,
			ProcessCSVTables.class }, library = true, complete = false)
	public static class FusionModule {

		private final String setScoresCreatorCacheFolder;
		private final String mappingsCachesFolder;
		private final boolean alwaysDisableBlocking;
		private KnowledgeBase_Attribute_Indeces knowledgeBase_Attribute_Indeces;

		public FusionModule(String setScoresCreatorCacheFolder, String mappingsCachesFolder, boolean alwaysDisableBlocking) {
			super();
			this.setScoresCreatorCacheFolder = setScoresCreatorCacheFolder;
			this.mappingsCachesFolder = mappingsCachesFolder;
			this.alwaysDisableBlocking = alwaysDisableBlocking;
		}

		@Provides
		@Singleton
		WeightsOptimizer getOptimizer(WeightsGeneticParallelOptimizer optimizer) {
			return optimizer;
		}

		// @Provides
		// @Named(SetScoresCreator.PROPERTY_CACHE_FOLDER)
		// String getSetCreatorsCacheFolder() {
		// return setScoresCreatorCacheFolder;
		// }
		//
		// @Provides
		// @Named(InstanceSetScoresCreator.PROPERTY_CACHE_FOLDER)
		// String getInstanceSetCreatorsCacheFolder() {
		// return setScoresCreatorCacheFolder;
		// }

		@Provides
		@Named(SetExpansionProperties.tablesCachesFolder)
		String getTablesCachesFolder() {
			return mappingsCachesFolder;
		}

		@Provides
		@Named(GreedyClusteringSplitter.DISABLE_BLOCKING)
		boolean getAlwaysDisableBlocking() {
			return alwaysDisableBlocking;
		}

		@Provides
		KnowledgeBase_Attribute_Indeces getKnowledgeBaseAttributeIndeces() {
			return knowledgeBase_Attribute_Indeces;
		}

		public void setKnowledgeBaseAttributeIndeces(KnowledgeBase_Attribute_Indeces knowledgeBase_Attribute_Indeces) {
			this.knowledgeBase_Attribute_Indeces = knowledgeBase_Attribute_Indeces;

		}
	}

	private final ObjectGraph graph;
	private final String dataHome;
	private final String resultsFolder;
	private Set<ClassId> topClases;
	private final String thresholdsFolder;
	private final String classAndPropertyFilterFolder;
	private Map<ClassId, Set<ClassId>> directChildrenMap;
	private Map<ClassId, Set<ClassId>> directParentsMap;
	private OntologyRepository ontology;
	private final String caseFolder;
	private final SimilarityModule similarityModule;
	private final String conflictDataFolter;
	private final String mappingsCachesFolder;
	private final String tablesFolder;
	private final String mappingsFolder;
	private final String kbName;
	private final String kbsHome;
	private final String kbPath;
	private Map<ClassId, Set<PropertyId>> classesAndProperties = null;

	private final String currentCase;

	public Map<ClassId, Set<PropertyId>> getNonTailProperties() {
		return Collections.unmodifiableMap(classesAndProperties);
	}

	public String getKbName() {
		return kbName;
	}

	public String getKbsHome() {
		return kbsHome;
	}

	String getMappingsCachesFolder() {
		return mappingsCachesFolder;
	}

	String getTablesFolder() {
		return tablesFolder;
	}

	public String getMappingsFolder() {
		return mappingsFolder;
	}

	public SetExtension(String[] args, boolean loadGraph, RepositoryLoading repositoryLoading) throws IOException {
		this(args, false, null, false, null, null, loadGraph, null, null, null, repositoryLoading);
	}

	public SetExtension(String[] args, boolean loadGraph) throws IOException {
		this(args, false, null, false, null, null, loadGraph, null, null, null, RepositoryLoading.LOCAL_ONLY);
	}

	public SetExtension(String[] args) throws IOException {
		this(args, false, null, false, null, null, null, null);
	}

	public SetExtension(String[] args, Set<ClassId> topClasses) throws IOException {
		this(args, true, topClasses, false, null, null, null, null);
	}

	public SetExtension(String[] args, Set<ClassId> topClasses, List<KBValueCreator> valueCreatorList) throws IOException {
		this(args, true, topClasses, false, null, null, valueCreatorList, null);
	}

	public SetExtension(String[] args, Set<ClassId> topClasses, RepositoryLoading repositoryLoading) throws IOException {
		this(args, true, topClasses, false, null, null, true, null, null, null, repositoryLoading);
	}

	public SetExtension(String[] args, Set<ClassId> topClasses, double propertyThreshold, RepositoryLoading repositoryLoading) throws IOException {
		this(args, true, topClasses, false, null, null, true, Double.valueOf(propertyThreshold), null, null, repositoryLoading);
	}

	public SetExtension(String[] args, Set<ClassId> topClasses, KnowledgeBase inputKB, OntologyRepository ontology, double propertyThreshold,
			RepositoryLoading repositoryLoading) throws IOException {
		this(args, true, topClasses, false, inputKB, ontology, true, Double.valueOf(propertyThreshold), null, null, repositoryLoading);
	}

	public SetExtension(String[] args, Set<ClassId> topClasses, double propertyThreshold, RepositoryLoading repositoryLoading, int currentAdaptedTablesRun)
			throws IOException {
		this(args, true, topClasses, false, null, null, true, Double.valueOf(propertyThreshold), null, null, repositoryLoading, false, false,
				currentAdaptedTablesRun, true);
	}

	public SetExtension(String[] args, Set<ClassId> topClasses, double propertyThreshold, RepositoryLoading repositoryLoading, int currentAdaptedTablesRun,
			boolean alwaysDisableBlocking) throws IOException {
		this(args, true, topClasses, false, null, null, true, Double.valueOf(propertyThreshold), null, null, repositoryLoading, false, false,
				currentAdaptedTablesRun, alwaysDisableBlocking);
	}

	public SetExtension(String[] args, Set<ClassId> topClasses, double propertyThreshold, RepositoryLoading repositoryLoading, boolean clearAttributeMappings,
			boolean adaptBasedOnAttributeMappings, int currentAdaptedTablesRun) throws IOException {
		this(args, true, topClasses, false, null, null, true, Double.valueOf(propertyThreshold), null, null, repositoryLoading, clearAttributeMappings,
				adaptBasedOnAttributeMappings, currentAdaptedTablesRun, true);
	}

	public SetExtension(String[] args, Set<ClassId> topClasses, double propertyThreshold, RepositoryLoading repositoryLoading, boolean clearAttributeMappings,
			boolean adaptBasedOnAttributeMappings, int currentAdaptedTablesRun, boolean alwaysDisableBlocking) throws IOException {
		this(args, true, topClasses, false, null, null, true, Double.valueOf(propertyThreshold), null, null, repositoryLoading, clearAttributeMappings,
				adaptBasedOnAttributeMappings, currentAdaptedTablesRun, alwaysDisableBlocking);
	}

	public SetExtension(String[] args, Set<ClassId> topClasses, double propertyThreshold, RepositoryLoading repositoryLoading, boolean clearAttributeMappings,
			boolean adaptBasedOnAttributeMappings) throws IOException {
		this(args, true, topClasses, false, null, null, true, Double.valueOf(propertyThreshold), null, null, repositoryLoading, clearAttributeMappings,
				adaptBasedOnAttributeMappings, 1, true);
	}

	public SetExtension(String[] args, Set<ClassId> topClasses, double propertyThreshold) throws IOException {
		this(args, true, topClasses, false, null, null, true, Double.valueOf(propertyThreshold), null, null, RepositoryLoading.LOCAL_ONLY);
	}

	public SetExtension(String[] args, Set<ClassId> topClasses, double propertyThreshold, String luceneIndexCache, RepositoryLoading repositoryLoading)
			throws IOException {
		this(args, true, topClasses, false, null, null, true, Double.valueOf(propertyThreshold), null, luceneIndexCache, repositoryLoading);
	}

	public SetExtension(String[] args, Set<ClassId> topClasses, double propertyThreshold, List<KBValueCreator> valueCreatorList, String luceneIndexCache,
			RepositoryLoading repositoryLoading) throws IOException {
		this(args, true, topClasses, false, null, null, true, Double.valueOf(propertyThreshold), valueCreatorList, luceneIndexCache, repositoryLoading);
	}

	public SetExtension(String[] args, Set<ClassId> topClasses, double propertyThreshold, List<KBValueCreator> valueCreatorList, String luceneIndexCache)
			throws IOException {
		this(args, true, topClasses, false, null, null, true, Double.valueOf(propertyThreshold), valueCreatorList, luceneIndexCache,
				RepositoryLoading.LOCAL_ONLY);
	}

	public SetExtension(String[] args, Set<ClassId> topClasses, boolean loadFromFilter, RepositoryLoading repositoryLoading) throws IOException {
		this(args, true, topClasses, loadFromFilter, null, null, null, null, repositoryLoading);
	}

	public SetExtension(String[] args, Set<ClassId> topClasses, boolean loadFromFilter) throws IOException {
		this(args, true, topClasses, loadFromFilter, null, null, null, null);
	}

	public SetExtension(String[] args, boolean mapTopClasses, Set<ClassId> topClasses, boolean loadFromFilter, KnowledgeBase kb, OntologyRepository ontology,
			List<KBValueCreator> valueCreatorList, String luceneIndexCache, RepositoryLoading repositoryLoading) throws IOException {
		this(args, mapTopClasses, topClasses, loadFromFilter, kb, ontology, true, null, valueCreatorList, luceneIndexCache, repositoryLoading);
	}

	public SetExtension(String[] args, boolean mapTopClasses, Set<ClassId> topClasses, boolean loadFromFilter, KnowledgeBase kb, OntologyRepository ontology,
			List<KBValueCreator> valueCreatorList, String luceneIndexCache) throws IOException {
		this(args, mapTopClasses, topClasses, loadFromFilter, kb, ontology, true, null, valueCreatorList, luceneIndexCache, RepositoryLoading.LOCAL_ONLY);
	}

	public SetExtension(String[] args, boolean mapTopClasses, Set<ClassId> topClasses, boolean loadFromFilter, KnowledgeBase inputKb,
			OntologyRepository ontology, boolean loadGraph, Double propertyDensityThreshold, List<KBValueCreator> kbValueCreatorList, String luceneIndexCache,
			RepositoryLoading repositoryLoading) throws IOException {
		this(args, mapTopClasses, topClasses, loadFromFilter, inputKb, ontology, loadGraph, propertyDensityThreshold, kbValueCreatorList, luceneIndexCache,
				repositoryLoading, false);
	}

	public SetExtension(String[] args, boolean mapTopClasses, Set<ClassId> topClasses, boolean loadFromFilter, KnowledgeBase inputKb,
			OntologyRepository ontology, boolean loadGraph, Double propertyDensityThreshold, List<KBValueCreator> kbValueCreatorList, String luceneIndexCache,
			RepositoryLoading repositoryLoading, boolean clearAttributeMappings) throws IOException {
		this(args, mapTopClasses, topClasses, loadFromFilter, inputKb, ontology, loadGraph, propertyDensityThreshold, kbValueCreatorList, luceneIndexCache,
				repositoryLoading, false, false, 1, true);
	}

	public SetExtension(String[] args, boolean mapTopClasses, Set<ClassId> topClasses, boolean loadFromFilter, KnowledgeBase inputKb,
			OntologyRepository ontology, boolean loadGraph, Double propertyDensityThreshold, List<KBValueCreator> kbValueCreatorList, String luceneIndexCache,
			RepositoryLoading repositoryLoading, boolean clearAttributeMappings, boolean adaptBasedOnAttributeMappings, int currentAdaptedTablesRun,
			boolean alwaysDisableBlocking) throws IOException {

		currentCase = new SimpleDateFormat("yyyyMMddHHmmss").format(new Date()) + "-" + RandomStringUtils.randomAlphanumeric(4);

		System.out.println("adfasdf");
		String dataHomeTemp = null;

		if (args.length > 0 && args[0] != null && !args[0].equals("")) {
			dataHomeTemp = args[0];
		}

		String kbName = null;

		if (args.length > 1 && args[1] != null && !args[1].equals("")) {
			kbName = args[1];
		}

		this.kbName = kbName;

		KbType kbType = KbType.getType(kbName);

		String mappingsName = null;

		if (args.length > 2 && args[2] != null && !args[2].equals("")) {
			if (args[2].equals("none")) {
				mappingsName = null;
			} else {
				mappingsName = args[2];
			}
		}
		this.dataHome = dataHomeTemp;

		boolean kbAndTablesAlreadyAdapted = false;

		if (args.length > 3 && args[3] != null && !args[3].equals("")) {
			if (args[3].equalsIgnoreCase("true")) {
				kbAndTablesAlreadyAdapted = true;
			} else if (args[3].equalsIgnoreCase("false")) {
				kbAndTablesAlreadyAdapted = false;
			} else {
				throw new IllegalStateException("Please add true or false for adapted or not adpated to your args");
			}

		} else {
			throw new IllegalStateException("Please add true or false for adapted or not adpated to your args");
		}

		boolean adaptedTable = true;
		if (args.length > 4 && args[4] != null && !args[4].equals("")) {
			if (args[4].equalsIgnoreCase("false")) {
				adaptedTable = false;
			}

		}

		if (mappingsName != null) {
			mappingsName = "/tables/" + mappingsName;

			this.mappingsFolder = dataHome + mappingsName + "/";
		} else {
			mappingsName = "";

			this.mappingsFolder = dataHome + "/";
		}

		this.kbsHome = dataHome + "/caches/kbs/";
		this.kbPath = dataHome + "/caches/kbs/" + kbName + "/kb/";
		String ontologyRepositoryFile = dataHome + "/caches/kbs/" + kbName + "/ontology.kryo";
		if (luceneIndexCache == null) {
			// if (!kbAndTablesAlreadyAdapted) {
			// luceneIndexCache = this.mappingsFolder + "/adaptedLucene/";
			// MKDIR.createDirUnchecked(new File(luceneIndexCache));
			// } else
			luceneIndexCache = dataHome + "/caches/kbs/" + kbName + "/lucene/";
		}
		this.tablesFolder = dataHome + "/" + mappingsName + "/";
		String rawMappingsFolder = this.mappingsFolder + "/rawTables/";
		String cacheFolder = dataHome + "/" + mappingsName + "/cache/";

		cacheFolder = dataHome + "/" + mappingsName + "/caches/tables_" + currentAdaptedTablesRun + "/";

		if (currentAdaptedTablesRun == -1) {

			cacheFolder = dataHome + "/" + mappingsName + "/caches/tables_0/";
		}

		MKDIR.createDirUnchecked(new File(cacheFolder));

		this.resultsFolder = dataHome + "/" + mappingsName + "/results/";
		this.thresholdsFolder = dataHome + "/" + mappingsName + "/thresholds/";
		this.classAndPropertyFilterFolder = dataHome + "/" + mappingsName + "/filter/";
		this.caseFolder = dataHome + "/" + mappingsName + "/cases/";
		this.conflictDataFolter = dataHome + "/" + mappingsName + "/conflictData/";

		MKDIR.createDirUnchecked(new File(this.caseFolder));

		String problemsFolder = dataHome + "/" + mappingsName + "/problems/";

		String optimizingFolder = dataHome + "/" + mappingsName + "/optimizing/";

		String columnMappingsFolder = dataHome + "/" + mappingsName + "/columnMappings/";
		String heidelTimeConfigPath = dataHome + "/configs/heidelTime/config.props";
		String emptyFile = dataHome + "/configs/empty.txt";
		String candidateSelectionConfig = dataHome + "/configs/config.conf";
		String matchingConfig = dataHome + "/configs/config.conf";
		UnitParser.unitPath = dataHome + "/configs/Units/Convertible";

		String rankingsFile = dataHome + "/" + mappingsName + "/rankings/filtered.tsv.gz";
		String unaryFile = dataHome + "/unary";

		String setScoresCacheFolder = dataHome + "/" + mappingsName + "/hierarchyCache/";

		this.mappingsCachesFolder = dataHome + "/" + mappingsName + "/caches/";

		String degreeFile = dataHome + "/rawDBpediaDatasets/others/page_links_en.nt.gz";
		String degreeCache = dataHome + "/caches/indegree/";

		String abstractsFile = dataHome + "/rawDBpediaDatasets/others/long_abstracts_en.nt.gz";
		String abstractsCache = dataHome + "/caches/abstracts/";

		String labelsFolder = dataHome + "/rawDBpediaDatasets/others/";

		MKDIR.createDirUnchecked(new File(mappingsCachesFolder));

		if (!loadGraph) {
			graph = null;
			similarityModule = null;
			return;
		}

		similarityModule = new SimilarityModule(dataHome + "/configs/stopwords.txt");

		KnowledgeBaseModule kbModule = new KnowledgeBaseModule(kbPath, ontologyRepositoryFile, luceneIndexCache, repositoryLoading);
		Map<PropertyId, UnitOfMeasurement> kbUnits = new HashMap<>();
		kbUnits.put(PropertyConstants.weight, UnitOfMeasurement.WEIGHT_KG);
		kbUnits.put(PropertyConstants.weightPounds, UnitOfMeasurement.WEIGHT_POUND);

		Map<PropertyId, PropertyId> normalizationIndex = new HashMap<>();
		normalizationIndex.put(PropertyConstants.weightPounds, PropertyConstants.weight);

		Map<ClassPropertyId, Set<ClassPropertyId>> equalProperties = new HashMap<>();

		equalProperties.put(ClassPropertyConstants.gridIronFootballPlayer_team,
				ImmutableSet.of(ClassPropertyConstants.gridIronFootballPlayer_formerTeam, ClassPropertyConstants.gridIronFootballPlayer_debutTeam));

		equalProperties.put(ClassPropertyConstants.single_musicalArtist,
				ImmutableSet.of(ClassPropertyConstants.single_artist, ClassPropertyConstants.single_musicalBand));
		kbModule.setEqualProperties(equalProperties);
		kbModule.setKbUnits(kbUnits);
		kbModule.setNormalizationIndex(normalizationIndex);

		if (inputKb != null) {
			kbModule.setKb((KnowledgeBaseImpl) inputKb);
		}
		if (ontology != null) {
			kbModule.setOntologyRepository(ontology);
		}
		MatchingModule matchingModule = new MatchingModule(true, columnMappingsFolder, candidateSelectionConfig, matchingConfig, emptyFile);

		TablesModule tableModule;
		if (currentAdaptedTablesRun == -1) {
			tableModule = TablesModule.csvWithCache(rawMappingsFolder, cacheFolder, false);
		} else {
			tableModule = TablesModule.withCache(rawMappingsFolder, cacheFolder, true);
		}
		HeidelTimeModule heidelTimeModule = new HeidelTimeModule(heidelTimeConfigPath);
		TableFusionModule tableFusionModule = new TableFusionModule(resultsFolder, problemsFolder, null, null, optimizingFolder, null);
		DegreeModule degreeModule = new DegreeModule(degreeFile, degreeCache);
		AbstractsModule abstractsModule = new AbstractsModule(abstractsFile, abstractsCache);

		FusionModule fusionModule = new FusionModule(setScoresCacheFolder, mappingsCachesFolder, alwaysDisableBlocking);
		this.graph = ObjectGraph.create(abstractsModule, degreeModule, tableFusionModule, kbModule, tableModule, heidelTimeModule, matchingModule,
				similarityModule, fusionModule, new UnaryAttributeModule(unaryFile));

		OntologicalFixes ontologicalFixes = new OntologicalFixes(ontology);

		if (!kbAndTablesAlreadyAdapted) {
			List<OntologyAdapter> ontologyAdaptersList = new ArrayList<>();
			ontologyAdaptersList.add(ontologicalFixes);
			kbModule.setOntologyAdaptersList(ontologyAdaptersList);
		}

		this.ontology = graph.get(OntologyRepository.class);
		ontology = this.ontology;

		// RankingModule rankingModule = new RankingModule(emptyFile,
		// rankingsFile);

		if (kbValueCreatorList == null) {
			kbValueCreatorList = new ArrayList<>();
		}
		if (kbType == KbType.WIKIDATA) {
			// List<KBValueCreator> valueCreatorList = new ArrayList<>();
			kbValueCreatorList.add(new CountryListFixer(dataHome + "/" + "countrycodes.txt"));

		}

		if (!kbAndTablesAlreadyAdapted) {
			ExtendedKnowledgeBaseValueCreator extendedKnowledgeBaseValueCreator = new ExtendedKnowledgeBaseValueCreator(dataHome + "/configs/");
			KBValueCreator kbValueCreator = extendedKnowledgeBaseValueCreator.getCreator();
			kbValueCreatorList.add(kbValueCreator);

			// kbValueCreatorList.add(new
			// KbAbstractObjectInstanceConverter(ontology));
			kbValueCreatorList.add(new PropertiesDBpediaTableFilter());

			kbValueCreatorList.add(new FootBallCollegeValueCreator(labelsFolder + "/infobox_properties_en.nt.gz"));
			kbValueCreatorList.add(new DraftYearValueCreator(labelsFolder + "/mappingbased_properties_en.nt.gz"));
			kbValueCreatorList.add(new WeightInPoundsValueCreator());
			kbValueCreatorList.add(new IsPartOfFixer(ontology));

			kbValueCreatorList.add(ontologicalFixes);

			// if (!kbValueCreatorList.isEmpty()) {
			// kbModule.setValueCreatorList(kbValueCreatorList);
			//
			// }

			if (labelsFolder != null) {
				KbLabelRenamer kbLabelRenamer;

				Set<String> files = new HashSet<>();
				files.add(labelsFolder + "mappingbased_properties_en.nt.gz");
				files.add(labelsFolder + "labels_en.nt.gz");
				Set<String> redirectFiles = new HashSet<>();
				redirectFiles.add(labelsFolder + "redirects_en.nt.gz");
				Set<String> surfaceFormFiles = new HashSet<>();
				surfaceFormFiles.add(labelsFolder + "SurfaceForms_LRD-CC_filter_thr3.80.tsv.gz");
				surfaceFormFiles.add(labelsFolder + "SurfaceForms_LRD-WAT_filter_thr2.60.tsv.gz");
				kbLabelRenamer = new KbLabelRenamer(files, redirectFiles, surfaceFormFiles);

				kbValueCreatorList.add(kbLabelRenamer);

			}
		}

		if (!kbValueCreatorList.isEmpty()) {
			kbModule.setValueCreatorList(kbValueCreatorList);

		}
		// RUN WITHOUT OTSU FILTERING

		{

			// if (kbType != KbType.WIKIDATA) {
			// List<KBValueCreator> valueCreatorList = new ArrayList<>();
			// KbAbstractObjectInstanceConverter converter = new
			// KbAbstractObjectInstanceConverter(graph.get(OntologyRepository.class));
			// valueCreatorList.add(converter);
			// kbModule.setValueCreatorList(valueCreatorList);
			// }

			KnowledgeBase kb = graph.get(KnowledgeBase.class);

			abstractsModule.setInstanceIds(kb.getAllInstances());
			setLabelGetter(kb::getLabels);

			Set<ClassPropertyId> numericRangeProperties = new HashSet<>();
			numericRangeProperties.add(ClassPropertyConstants.gridIronFootballPlayer_height);
			numericRangeProperties.add(ClassPropertyConstants.gridIronFootballPlayer_weight);
			numericRangeProperties.add(ClassPropertyConstants.gridIronFootballPlayer_weightPounds);
			numericRangeProperties.add(ClassPropertyConstants.gridIronFootballPlayer_draftRound);
			numericRangeProperties.add(ClassPropertyConstants.gridIronFootballPlayer_draftPick);
			numericRangeProperties.add(ClassPropertyConstants.gridIronFootballPlayer_number);

			if (topClasses.contains(ClassConstants.single)) {
				numericRangeProperties.add(ClassPropertyConstants.single_runtime);
			}

			if (kbAndTablesAlreadyAdapted && !similarityModule.numericRangeCacheExists()) {
				Map<ClassPropertyId, Collection<AbstractObject>> numericValues = new ConcurrentHashMap<>();
				for (ClassPropertyId cpid : numericRangeProperties) {
					if (topClasses.contains(cpid.getClassId())) {
						Collection<AbstractObject> list = new ConcurrentSkipListSet<>();
						Set<InstanceId> instances = kb.getInstances(cpid.getClassId(), cpid.getPropertyId());
						PersistentParallel6.runAndFinish(instances, (instanceId) -> {

							QualifiedValueList qvl = kb.getQualifiedValueList(instanceId, cpid.getPropertyId());
							if (qvl != null) {
								for (QualifiedValueHolder claim : qvl.getClaims()) {
									if (claim.getValue().getType() == ValueType.QUANTITY || claim.getValue().getType() == ValueType.CATEGORICAL_INT) {
										list.add(claim.getValue());
									}
								}
							}

						}, this.getClass());
						numericValues.put(cpid, list);
					}
				}

				this.similarityModule.setNumericRangesValues(numericValues);
			}
			this.similarityModule.setSurfaceFormsGetter(kb::getSurfaceForms);
			this.similarityModule.setCandidatesSurfaceFormsGetter(kb::getCandidatesSurfaceForms);
			// {
			// System.out.println("Now before filter: ");
			// System.out.println(kb.getClassPropertyIds());
			// for (ClassId classId : kb.getClasses()) {
			// for (PropertyId propertyId : kb.getProperties(classId).keySet())
			// {
			// System.out.println(kb.getInstancesSize(classId, propertyId));
			// }
			// }
			//
			// KBValueCreatorsRunner runner = new KBValueCreatorsRunner();
			//
			// List<KBValueCreator> replacers = new ArrayList<>();
			// replacers.add(new PropertiesDBpediaTableFilter());
			// KnowledgeBaseImpl newKb = runner.runReplacers((KnowledgeBaseImpl)
			// kb, replacers);
			// System.out.println("Now after filter: ");
			// System.out.println(newKb.getClassPropertyIds());
			// for (ClassId classId : newKb.getClasses()) {
			// for (PropertyId propertyId :
			// newKb.getProperties(classId).keySet()) {
			// System.out.println(newKb.getInstancesSize(classId, propertyId));
			// }
			// }
			//
			// int sdf = 123;
			//
			// }
			KnowledgeBase finalKb = kb;

			SimilarityCalculator sc = graph.get(SimilarityCalculator.class);

			if (loadFromFilter) {
				topClasses = this.getClassFilter();
			}

			if (kbAndTablesAlreadyAdapted && propertyDensityThreshold != null) {
				this.classesAndProperties = getNonTailPropeties(topClasses, kb, ontology, propertyDensityThreshold, graph.get(ThreadSafeNoReferenceKryo.class));
			}

			if (kbAndTablesAlreadyAdapted && repositoryLoading != RepositoryLoading.REMOTE_LOAD_LOCALLY) {
				fusionModule.setKnowledgeBaseAttributeIndeces(
						new KnowledgeBase_Attribute_Indeces(this.ontology, kb, this.classesAndProperties, graph.get(LabelMatcher.class)));
			}

			List<AnnotatedTableFilter> filters = new ArrayList<>();

			if (kbAndTablesAlreadyAdapted && !adaptedTable) {

				if (kbType == KbType.WIKIDATA) {
					// AnnotatedTableOtsuFilter tableOtsuFilter =
					// graph.get(AnnotatedTableOtsuFilter.class);
					// tableOtsuFilter.setMin(0.1);
					// filters.add(tableOtsuFilter);
					filters.add(new SmallTableFilter());

					filters.add(new GoodPropertiesFilterAll(dataHome + "/" + "goodProperties.csv"));
					filters.add(new KbInstanceFilter(kb));
					// filters.add(new SmallTableFilter());

					SourcesFilter sourcesFilter = new SourcesFilter();
					filters.add(sourcesFilter);
				}

				Map<ClassId, Set<ClassId>> parents = new HashMap<>();
				Map<ClassId, Set<ClassId>> children = new HashMap<>();
				Set<ClassId> classes = kb.getClasses();
				for (ClassId classId : classes) {
					Set<ClassId> parentsSet = new HashSet<>();
					for (ClassId parentClass : ontology.getDirectParentClasses(classId)) {
						if (classes.contains(parentClass)) {
							parentsSet.add(parentClass);
							CHMHelper.getFromMap(children, parentClass, () -> new HashSet<>()).add(classId);
						}
					}
					parents.put(classId, parentsSet);
				}

				this.directParentsMap = parents;
				this.directChildrenMap = children;

				if (mapTopClasses && topClasses == null) {

					// classes = classes.stream().filter((e) -> {
					// try {
					// return finalKb.getInstances(e,
					// PropertiesExpansionData.LABEL_PROPERTY).size() > 0;
					// } catch (Exception exc) {
					// return false;
					// }
					// }).collect(Collectors.toSet());

					Set<ClassId> levelZero = parents.entrySet().stream().filter((e) -> e.getValue().size() == 0).map(Entry::getKey).collect(Collectors.toSet());

					System.out.println("Level Zero: ");
					for (ClassId classId : levelZero) {
						System.out.println(classId.getClassId());
					}

					Set<ClassId> levelOneNew = getNextLevel(levelZero);

					System.out.println("Level-Zero-Classes: " + levelZero.size());
					System.out.println("Level-One-Classes: " + levelOneNew.size());
					System.out.println("Total-Classes: " + classes.size());

					topClasses = new HashSet<>();
					// int i = 0;
					for (ClassId classId : levelZero) {

						topClasses.add(classId);

						// if (i == 5) {
						// break;
						// }
						// i++;
					}

				}
				// topClasses.add(ClassConstants.musicalWorkClass);
				BadDecodedTableFilter badDecodedTableFilter = new BadDecodedTableFilter();
				filters.add(badDecodedTableFilter);
				System.out.println("TablesFilteredOutDueToBadDecoding: " + badDecodedTableFilter.getCount());

				filters.add(new PropertiesDBpediaTableFilter());
				filters.add(new PopulationDensityFixer());
				filters.add(new RuntimePropertyFixer());

				filters.add(new NFLPropertiesDensity());
				filters.add(new MusicalArtistPropertyFix());
				filters.add(new NumericKeyFilter(kb, sc));

				filters.add(new SettlementTableFixer(kb, ontology));

				// AnnotatedTableOtsuFilter otstuFilter =
				// graph.get(AnnotatedTableOtsuFilter.class);
				// otstuFilter.setMin(0.3);

				// filters.add(otstuFilter);

				if (mapTopClasses && topClasses != null) {

					OntologyRepository ontologyRepository = graph.get(OntologyRepository.class);

					Map<ClassId, ClassId> alreadyAdded = new HashMap<>();

					Map<ClassId, Set<ClassId>> topClassesMap = new HashMap<>();

					for (ClassId topClass : topClasses) {
						Set<ClassId> subClasses = new HashSet<>(ontologyRepository.getSubClasses(topClass));
						for (Iterator<ClassId> it = subClasses.iterator(); it.hasNext();) {
							ClassId next = it.next();
							if (alreadyAdded.containsKey(next)) {
								it.remove();
								System.out.println("Removed class " + next.getClassId() + " from TopClass " + topClass.getClassId()
										+ " because it already exists in TopClass: " + alreadyAdded.get(next).getClassId());
							} else {
								alreadyAdded.put(next, topClass);
							}

						}
						CHMHelper.getFromMap(topClassesMap, topClass, () -> new HashSet<>()).addAll(subClasses);
					}

					filters.add(new TableClassMapper(topClassesMap));
				}

				if (propertyDensityThreshold != null) {

					GoodPropertiesFilterAll propertiesFilter = new GoodPropertiesFilterAll(classesAndProperties);
					filters.add(propertiesFilter);
				}

				// AnnotatedTableInstanceRescorer rescorer =
				// graph.get(AnnotatedTableInstanceRescorer.class);
				// filters.add(rescorer);

				// filters.add(new TableColumnCoordinateToQuantityConverter());

				tableModule.setClassFilter(graph.get(KnowledgeBase.class));
			}

			// filters.add(new TableColumnCoordinateToQuantityConverter());
			// duplicateFilter = new
			// DuplicateCurrentStateOfTableRepositoryFilter()
			// filters.add();
			//
			// tableModule.setBeforeValueAdaptionRepository(beforeValueAdaptionRepository);
			if (adaptBasedOnAttributeMappings) {
				filters.addAll(AttributeMatcher.getValueAdaptionFilters(ontology));
			}

			if (clearAttributeMappings) {
				filters.add(new TableColumnMapingsClearerer());
			}

			if (!filters.isEmpty()) {
				tableModule.setFilters(filters);
			}

			this.topClases = topClasses;

		}

		// // RUN WITH OTSU FILTERING
		// {
		// List<AnnotatedTableFilter> filters = new ArrayList<>();
		// filters.add();
		// tableModule.setFilters(filters);
		// ObjectGraph graph = ObjectGraph.create(tableFusionModule, kbModule,
		// tableModule, heidelTimeModule, matchingModule, similarityModule,
		// new FusionModule());
		// Fusion fusion = graph.get(Fusion.class);
		// fusion.run();
		// }

		// kbModule.setOntologyReplacer((ontologyRepository) -> {
		// return new
		// ProcessGzToKryoWithTimestampsAndColumnMatching.OntologyRepositoryWrapperReference(ontologyRepository);
		// });
		//
		// List<KBValueCreator> valueCreatorList = new ArrayList<>();
		//
		// valueCreatorList.add(new
		// ProcessGzToKryoWithTimestampsAndColumnMatching.CreateReferencePropertiesRange(graph.get(OntologyRepository.class)));
		// kbModule.setValueCreatorList(valueCreatorList);
		// testSerialize(new ThreadSafeNoReferenceKryo(),
		// graph.get(AnnotatedTableParserJSON.class), dataHome + "/" +
		// mappingsName + "/");

		// System.exit(0);

	}

	private Set<ClassId> getNextLevel(Set<ClassId> levelZero) {
		Set<ClassId> levelOne = levelZero.stream().flatMap((classId) -> {
			if (this.directChildrenMap.containsKey(classId)) {
				return this.directChildrenMap.get(classId).stream();
			} else {
				return Stream.of(classId);
			}

		}).collect(Collectors.toSet());

		Set<ClassId> levelOneNew = filterOutIfParentsExist(levelOne);

		System.out.println("Level One: ");
		for (ClassId classId : levelOneNew) {
			System.out.println(classId.getClassId());
		}

		return levelOneNew;
	}

	public Set<ClassId> filterOutIfParentsExist(Set<ClassId> set) {
		Set<ClassId> levelOneNew = set.stream().filter((e) -> {

			Set<ClassId> allParents = ontology.getParentClasses(e);
			for (ClassId parent : allParents) {
				if (set.contains(parent)) {
					return false;
				}
			}
			return true;
		}).collect(Collectors.toSet());

		return levelOneNew;
	}

	public ObjectGraph getGraph() {
		return graph;
	}

	public String getDataHome() {
		return dataHome;
	}

	public String getThresholdsFolder() {
		return thresholdsFolder;
	}

	public String getResultsFolder() {
		return resultsFolder;
	}

	public Set<ClassId> getTopClases() {
		return topClases;
	}

	public Set<ClassId> getDirectChildren(ClassId classId) {
		Set<ClassId> set = this.directChildrenMap.get(classId);
		if (set == null)
			set = new HashSet<>();
		return Collections.unmodifiableSet(set);
	}

	public Set<ClassId> getDirectParents(ClassId classId) {
		return Collections.unmodifiableSet(this.directParentsMap.get(classId));
	}

	String getClassAndPropertyFilterFolder() {
		return classAndPropertyFilterFolder;
	}

	public Map<ClassId, Set<PropertyId>> getNonTailPropeties(Set<ClassId> topClasses, KnowledgeBase kb, OntologyRepository ontologyRepository,
			ThreadSafeNoReferenceKryo kryoAdapter) {
		return getNonTailPropeties(topClasses, kb, ontologyRepository, ExtractTablesForGoldStandardRun.getMinPropertyDensity(), kryoAdapter, "tailProperties");
	}

	public Map<ClassId, Set<PropertyId>> getNonTailPropeties(Set<ClassId> topClasses, KnowledgeBase kb, OntologyRepository ontologyRepository,
			double propertyThreshold, ThreadSafeNoReferenceKryo kryoAdapter) {
		return getNonTailPropeties(topClasses, kb, ontologyRepository, propertyThreshold, kryoAdapter, "tailProperties");

	}

	public Map<ClassId, Set<PropertyId>> getNonTailPropeties(Set<ClassId> topClasses, KnowledgeBase kb, OntologyRepository ontologyRepository,
			ThreadSafeNoReferenceKryo kryoAdapter, String cacheFolderName) {
		return getNonTailPropeties(topClasses, kb, ontologyRepository, ExtractTablesForGoldStandardRun.getMinPropertyDensity(), kryoAdapter, cacheFolderName);
	}

	public Map<ClassId, Set<PropertyId>> getNonTailPropeties(Set<ClassId> topClasses, KnowledgeBase kb, OntologyRepository ontologyRepository,
			double propertyThreshold, ThreadSafeNoReferenceKryo kryoAdapter, String cacheFolderName) {

		try {

			String fullCachePath = this.kbPath + "/" + cacheFolderName;
			MKDIR.createDirUnchecked(new File(this.kbPath));
			Map<ClassId, Set<PropertyId>> ret;

			if (Files.exists(Paths.get(fullCachePath))) {
				ret = new HashMap<>((Map<ClassId, Set<PropertyId>>) kryoAdapter.read(fullCachePath));

				for (Entry<ClassPropertyId, Set<ClassPropertyId>> entry : kb.getEqualProperties().entrySet()) {
					if (topClasses.contains(entry.getKey().getClassId())) {
						Set<PropertyId> remove = entry.getValue().stream().map((e) -> e.getPropertyId()).collect(Collectors.toSet());
						Set<PropertyId> set = new HashSet<>(ret.get(entry.getKey().getClassId()));
						set.removeAll(remove);
						ret.put(entry.getKey().getClassId(), set);
					}
				}
			} else {

				ConcurrentHashMap<ClassId, Set<PropertyId>> classesAndProperties = new ConcurrentHashMap<>();

				for (ClassId classId : topClasses) {

					Set<ClassId> allClassIds = new HashSet<>();
					allClassIds.add(classId);
					allClassIds.addAll(ontologyRepository.getSubClasses(classId));

					Set<InstanceId> allInstanceIds = new HashSet<>();
					Set<PropertyId> allPropertyIds = new HashSet<>();
					for (ClassId subClassId : allClassIds) {
						if (kb.hasClass(subClassId)) {

							allInstanceIds.addAll(kb.getInstances(subClassId).keySet());
							allPropertyIds.addAll(kb.getProperties(subClassId).keySet());
						}
					}

					double perClass = allInstanceIds.size();
					PersistentParallel6.runAndFinish(allPropertyIds, (propertyId) -> {
						if (!PropertiesExpansionData.LABEL_PROPERTY.equals(propertyId)) {
							int perClassProperty = 0;

							for (InstanceId instanceId : allInstanceIds) {
								if (kb.hasValue(new ValueId(instanceId, propertyId))) {
									perClassProperty++;
								}
							}

							double proportion = perClassProperty / perClass;
							if (proportion >= propertyThreshold) {
								CHMHelper.getFromCHM(classesAndProperties, classId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(propertyId);
							}

						}
					}, this.getClass());
				}
				System.out.println(classesAndProperties.toString());
				ret = new HashMap<>(classesAndProperties.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> new HashSet<>(e.getValue()))));
				kryoAdapter.write(ret, fullCachePath);

			}

			Map<ClassId, Set<PropertyId>> newRet = new HashMap<>();
			for (Entry<ClassId, Set<PropertyId>> entrySet : ret.entrySet()) {
				Set<PropertyId> newSet = new HashSet<>(entrySet.getValue());
				if (entrySet.getKey().equals(ClassConstants.single)) {
					newSet.remove(PropertyConstants.workRuntime);
				}
				if (entrySet.getKey().equals(ClassConstants.placeClass)) {
					newSet.remove(PropertyConstants.utcOffset);
				}

				if (entrySet.getKey().equals(ClassConstants.placeClass)) {
					newSet.remove(PropertyConstants.latitude);
				}

				if (entrySet.getKey().equals(ClassConstants.placeClass)) {
					newSet.remove(PropertyConstants.longitude);
				}

				if (entrySet.getKey().equals(ClassConstants.placeClass)) {
					newSet.remove(PropertyConstants.point);
				}

				if (entrySet.getKey().equals(ClassConstants.placeClass)) {
					newSet.remove(PropertyConstants.timeZone);
				}

				if (entrySet.getKey().equals(ClassConstants.single)) {
					newSet.remove(PropertyConstants.bSide);
				}

				newRet.put(entrySet.getKey(), newSet);
			}
			return newRet;

		}

		catch (IOException e) {
			e.printStackTrace();
			throw new RuntimeException();
		}
	}

	public Set<ClassId> getClassFilter() {
		Set<ClassId> ret = new HashSet<>();
		try {
			@SuppressWarnings("unchecked")
			List<String> lines = FileUtils.readLines(new File(this.getClassFilterFile()));
			for (String line : lines) {
				ret.add(new ClassId(line));
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();

		}
		System.out.println(ret.toString());
		return ret;
	}

	public String getClassFilterFile() {
		return this.classAndPropertyFilterFolder + "/classes";

	}

	public List<Pair<ClassId, ?>> getClassFilterHierarchy() {
		try {
			@SuppressWarnings("unchecked")
			List<String> lines = FileUtils.readLines(new File(this.getClassFilterHierarchyFile()));

			return getClassFilterHierarchy(lines);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return new ArrayList<>();

	}

	public static List<Pair<ClassId, ?>> getClassFilterHierarchy(List<String> lines) {

		List<Pair<ClassId, ?>> ret = new ArrayList<>();

		LinkedList<String> linkedLines = new LinkedList<>(lines);

		fillHierarchy(ret, "", linkedLines);

		System.out.println(ret.toString());
		return ret;
	}

	private static void fillHierarchy(List<Pair<ClassId, ?>> ret, String prefix, LinkedList<String> lines) {
		while (!lines.isEmpty()) {
			String peek = lines.peek();
			if (peek.length() > prefix.length() && peek.substring(0, prefix.length()).equals(prefix)) {
				String line = lines.poll().substring(prefix.length());
				ClassId classId = new ClassId(line);
				if (!lines.isEmpty() && lines.peek().startsWith(prefix + " ")) {
					List<Pair<ClassId, ?>> newRet = new ArrayList<Pair<ClassId, ?>>();
					ret.add(new Pair<>(classId, newRet));
					fillHierarchy(newRet, prefix + " ", lines);
				} else {
					ret.add(new Pair<>(classId, null));
				}
			} else {
				break;
			}
		}

	}

	public void writeClassFilter(Set<ClassId> finalClasses) {

		try {
			String classesFile = this.getClassFilterFile();

			MKDIR.createDirUnchecked(new File(this.getClassAndPropertyFilterFolder()));

			StringBuilder sb = new StringBuilder();
			for (Iterator<ClassId> it = finalClasses.iterator(); it.hasNext();) {

				sb.append(it.next().getClassId());
				if (it.hasNext()) {
					sb.append("\n");
				}

			}

			FileUtils.writeStringToFile(new File(classesFile), sb.toString());
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public void writeClassFilterHiearchy(Map<ClassId, Set<ClassId>> hierarchy) {

		try {
			MKDIR.createDirUnchecked(new File(this.getClassAndPropertyFilterFolder()));
			String classesFile = this.getClassFilterHierarchyFile();
			String string = writeClassFilterHiearchyToString(hierarchy);
			FileUtils.writeStringToFile(new File(classesFile), string);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public static String writeClassFilterHiearchyToString(Map<ClassId, Set<ClassId>> hierarchy) {
		List<ClassId> classesThatAreNotChildren = hierarchy.keySet().stream().filter((classId) -> {
			for (Set<ClassId> children : hierarchy.values()) {
				if (children.contains(classId))
					return false;
			}
			return true;
		}).sorted((e1, e2) -> e1.getClassId().compareTo(e2.getClassId())).collect(Collectors.toList());

		StringBuilder sb = new StringBuilder();
		for (Iterator<ClassId> it = classesThatAreNotChildren.iterator(); it.hasNext();) {

			appendWithChildren("", sb, it.next(), hierarchy);

			if (it.hasNext()) {
				sb.append("\n");
			}
		}
		return sb.toString();
	}

	private String getClassFilterHierarchyFile() {
		return this.classAndPropertyFilterFolder + "/hierarchy";
	}

	private static void appendWithChildren(String prefix, StringBuilder sb, ClassId currentClass, Map<ClassId, Set<ClassId>> hierarchy) {
		sb.append(prefix + currentClass.getClassId());

		if (hierarchy.containsKey(currentClass) && !hierarchy.get(currentClass).isEmpty()) {
			sb.append("\n");
			List<ClassId> children = hierarchy.get(currentClass).stream().sorted((e1, e2) -> e1.getClassId().compareTo(e2.getClassId()))
					.collect(Collectors.toList());
			for (Iterator<ClassId> it = children.iterator(); it.hasNext();) {
				appendWithChildren(prefix + " ", sb, it.next(), hierarchy);
				if (it.hasNext()) {
					sb.append("\n");
				}
			}
		}

	}

	String getConflictDataFolter() {
		return conflictDataFolter;
	}

	public Map<ClassId, Set<ClassId>> getAllParents(List<Pair<ClassId, ?>> hierarchy) {
		Map<ClassId, Set<ClassId>> allParents = new HashMap<>();

		addParent(allParents, new HashSet<>(), hierarchy);

		return allParents;
	}

	private void addParent(Map<ClassId, Set<ClassId>> ret, Set<ClassId> currentParents, List<Pair<ClassId, ?>> hierarchy) {
		for (Pair<ClassId, ?> pair : hierarchy) {
			if (pair.getSecond() == null) {
				ret.put(pair.getFirst(), new HashSet<>(currentParents));
			} else {
				Set<ClassId> newParents = new HashSet<>(currentParents);
				newParents.add(pair.getFirst());
				addParent(ret, newParents, (List<Pair<ClassId, ?>>) pair.getSecond());
			}
		}

	}

	String getCaseFolder() {
		return caseFolder;
	}

	String getCurrentManualCaseFolder() {
		String manualEvaluationFolder = getCaseFolder() + "/manual";
		MKDIR.createDirUnchecked(new File(manualEvaluationFolder));
		String currentCaseFolder = manualEvaluationFolder + "/" + currentCase;
		return currentCaseFolder;
	}

	void setLabelGetter(LabelsGetter labelsGetter) {
		this.similarityModule.setLabelsGetter(labelsGetter);
	}

	public static FusionConfig getFusionConfig() {
		return FusionConfig.builder().setMinSource(0).setAdjustedRecall(false).build();
	}

	public static SetExpansionStrategiesConfig getSetExpansionStrategiesConfig() {
		SetExpansionStrategiesConfigBuilder builder = SetExpansionStrategiesConfig.builder();
		builder.addFusionSetup(FusionStrategyRepository.KBT_ClassPropertyId_Threshold, PropertiesConfig.builder().setTimeDependence(TimeDependence.NO).build());
		return builder.build();
	}

	public String getCurrentCase() {
		return currentCase;
	}

}
