package de.uni_mannheim.informatik.dws.expansion.samples.setExtension;

import java.io.File;
import java.io.IOException;
import java.io.StringWriter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.inject.Inject;

import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.RandomStringUtils;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.PerformanceMacroPR;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig.TimeDependence;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategyRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.DecimalFormatter;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeKryo;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ThreadSafeNoReferenceKryo;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance.FusionPerformanceCreator;
import expansion.set.evaluation.ClusterPerformance;
import expansion.set.evaluation.SetExpansionRunner;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.evaluation.SetExpansionStrategiesConfig.SetExpansionStrategiesConfigBuilder;
import expansion.set.profiling.PossibleConflictProfiler;
import expansion.set.profiling.PossibleConflictProfiler.ConflictData;
import expansion.set.profiling.SetExpansionSuitabilityProfiler;
import expansion.set.strategies.SetExpansionStrategiesRepository;
import expansion.set.strategies.SetExpansionStrategy;

public class SetExtensionFindClasses {

	public static final String CONFLICT_FILE_NAME = "CONFLICT_FILE";

	public static void main(String[] args) throws IOException {
		// ProfilingResult profilingResults;
		Map<ClassId, ConflictData> conflictResults = new HashMap<>();
		if (true) {
			SetExtension setExtensionProfiling = new SetExtension(args, false);
			ThreadSafeKryo kryo = new ThreadSafeKryo();
			MKDIR.createDirUnchecked(new File(setExtensionProfiling.getConflictDataFolter()));

			String file = setExtensionProfiling.getConflictDataFolter() + "/" + CONFLICT_FILE_NAME;
			Path path = Paths.get(file);
			if (Files.exists(path)) {
				System.out.println("Reading Conflict From Chache");
				conflictResults = (Map<ClassId, ConflictData>) kryo.read(file);
				System.out.println(conflictResults.toString());
			} else {
				setExtensionProfiling = new SetExtension(args, true);
				PossibleConflictProfiler conflictProfiler = setExtensionProfiling.getGraph().get(PossibleConflictProfiler.class);
				conflictResults = conflictProfiler.run(null);
				kryo.write(conflictResults, file);
			}
			setExtensionProfiling = null;

			// SetExpansionSuitabilityProfiler profiler =
			// setExtensionProfiling.getGraph().get(SetExpansionSuitabilityProfiler.class);
			// profilingResults = profiler.measureAndStoreSuitability(null,
			// true, null, null);

		}

		Set<ClassId> topClassesStart = new HashSet<>();
		topClassesStart.add(ClassConstants.placeClass);

		SetExtension setExtension = new SetExtension(args, topClassesStart);
		String dataHome = setExtension.getDataHome();

		String resultsFolder = setExtension.getResultsFolder();
		String thresholdsFolder = setExtension.getThresholdsFolder();

		MaxDiffSetter maxDiffSetter = setExtension.getGraph().get(MaxDiffSetter.class);
		KnowledgeBase kb = setExtension.getGraph().get(KnowledgeBase.class);
		OntologyRepository ontology = setExtension.getGraph().get(OntologyRepository.class);
		maxDiffSetter.run(dataHome);
		//
		FusionConfig fusionConfig = FusionConfig.builder().setMinSource(3).setAdjustedRecall(false).build();

		if (true) {
			throw new IllegalStateException();
			// IT WAS BEFORE:
			// LABEL_BLOCKED_SIMILARITY
		}
		SetExpansionStrategy mainStrategy = SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY;
		int minNewValues = 50;
		int minNewEntities = 50;
		double minOfParent = 0.00;

		SetExpansionStrategiesConfigBuilder builder = SetExpansionStrategiesConfig.builder();
		builder.addFusionSetup(FusionStrategyRepository.QKBT_Mapped, PropertiesConfig.builder().setTimeDependence(TimeDependence.YES).build());
		builder.addFusionSetup(FusionStrategyRepository.KBT_ClassPropertyId_Threshold, PropertiesConfig.builder().setTimeDependence(TimeDependence.NO).build());

		builder.addStrategy(mainStrategy);

		SetExpansionStrategiesConfig setExpansionStrategiesConfig = builder.build();

		StrategiesConfig strategiesConfig = StrategiesConfig.defaultConfig();

		Set<ClassId> topClasses = setExtension.getTopClases();
		Set<ClassId> newTopClasses = new HashSet<>();
		Map<ClassId, Integer> kbValuesCount = new HashMap<>();
		Map<ClassId, Integer> newValuesCount = new HashMap<>();
		Map<ClassId, Integer> kbEntitiesCount = new HashMap<>();
		Map<ClassId, Integer> newEntitesCount = new HashMap<>();

		Map<ClassId, Set<ClassId>> includedChildren = new HashMap<>();
		Map<ClassId, Set<ClassId>> nextRunWithParent = new HashMap<>();

		do {

			SetExtensionFindClasses fusion = setExtension.getGraph().get(SetExtensionFindClasses.class);

			Map<ClassId, Set<PropertyId>> classesAndProperties = setExtension.getNonTailPropeties(setExtension.getTopClases(), kb, ontology,
					new ThreadSafeNoReferenceKryo());

			if (true) {
				throw new IllegalStateException();
			}
			Pair<Map<SetExpansionStrategy, FusionPerformanceCreator<PerformanceMacroPR>>, Map<SetExpansionStrategy, FusionPerformanceCreator<ClusterPerformance>>> result = null;
			// fusion.runner
			// .unmatchedRun("SetExtension", fusionConfig, strategiesConfig,
			// setExpansionStrategiesConfig, null, thresholdsFolder,
			// TestingMode.NORMAL,
			// classesAndProperties, null, null, null);

			FusionPerformance<ClusterPerformance> clusterPerformance = result.getSecond().get(mainStrategy).get();
			Map<ClassId, ClusterPerformance> clusterPerformancePerClassId = clusterPerformance.getPerClassId();
			FusionPerformance<PerformanceMacroPR> valuesPerformance = result.getFirst().get(mainStrategy).get();
			Map<ClassId, PerformanceMacroPR> valuesPerformancePerClassId = valuesPerformance.getPerClassId();

			Map<ClassId, Set<ClassId>> nextRunWithParentNew = new HashMap<>();

			for (ClassId classId : setExtension.getTopClases()) {
				if (clusterPerformancePerClassId.containsKey(classId) && valuesPerformancePerClassId.containsKey(classId)
						&& clusterPerformancePerClassId.get(classId).getInstancesCreated() > minNewEntities
						&& valuesPerformancePerClassId.get(classId).getNewValues() > minNewValues) {

					if (!topClasses.contains(classId)) {
						for (ClassId parent : nextRunWithParent.get(classId)) {
							CHMHelper.getFromMap(includedChildren, parent, () -> new HashSet<>()).add(classId);
						}
					} else {
						newTopClasses.add(classId);
					}

					newValuesCount.put(classId, valuesPerformancePerClassId.get(classId).getNewValues());
					newEntitesCount.put(classId, clusterPerformancePerClassId.get(classId).getInstancesCreated());

					kbValuesCount.put(classId, valuesPerformancePerClassId.get(classId).getValuesInKB());
					kbEntitiesCount.put(classId, clusterPerformancePerClassId.get(classId).getInstancesInKB());

					for (ClassId subClassId : setExtension.getDirectChildren(classId)) {
						CHMHelper.getFromMap(nextRunWithParentNew, subClassId, () -> new HashSet<>()).add(classId);
					}

				}
			}

			Set<ClassId> resultantSet = setExtension.filterOutIfParentsExist(nextRunWithParentNew.keySet());

			nextRunWithParent = new HashMap<>();
			if (!resultantSet.isEmpty()) {
				for (ClassId newClassId : resultantSet) {
					nextRunWithParent.put(newClassId, nextRunWithParentNew.get(newClassId));

				}

				SimpleLockableConcurrentCache.clearAllCaches();
				if (true) {
					throw new IllegalStateException();
				}
				setExtension = null;// new SetExtension(args, true,
									// resultantSet, false, kb, ontology);
			}

		} while (!nextRunWithParent.isEmpty());

		StringWriter sw = new StringWriter();
		Pair<Set<ClassId>, Map<ClassId, Set<ClassId>>> result = writeOutput(sw, newTopClasses, kbValuesCount, kbEntitiesCount, newValuesCount, newEntitesCount,
				includedChildren, minOfParent, minOfParent, conflictResults);

		String currentCase = new SimpleDateFormat("yyyyMMddHHmmss").format(new Date()) + "-" + RandomStringUtils.randomAlphanumeric(4);
		String resultsFile = setExtension.getResultsFolder() + "/FIND_CLASSES_" + currentCase;

		FileUtils.writeStringToFile(new File(resultsFile), sw.toString());

		System.out.println(sw.toString());

		Set<ClassId> finalClasses = result.getFirst();

		Map<ClassId, Set<ClassId>> hierarchy = result.getSecond();

		System.out.println("Final Set Of Classes: " + finalClasses);
		System.out.println("Hierarchy:" + hierarchy.toString());

		if (false) {
			setExtension.writeClassFilter(finalClasses);
			setExtension.writeClassFilterHiearchy(hierarchy);
		}

		try {
			PersistentParallel6.close();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private final SetExpansionRunner runner;

	private final SetExpansionSuitabilityProfiler profiler;

	@Inject
	public SetExtensionFindClasses(SetExpansionRunner runner, SetExpansionSuitabilityProfiler profiler) {
		this.runner = runner;
		this.profiler = profiler;
	}

	public static Pair<Set<ClassId>, Map<ClassId, Set<ClassId>>> writeOutput(StringWriter sw, Set<ClassId> topClasses, Map<ClassId, Integer> kbValuesCount,
			Map<ClassId, Integer> kbEntitiesCount, Map<ClassId, Integer> newValuesCount, Map<ClassId, Integer> newEntitesCount,
			Map<ClassId, Set<ClassId>> includedChildren, double minValuesOfParent, double minEntitiesOfParent, Map<ClassId, ConflictData> conflictResults) {
		Set<ClassId> ret = new HashSet<>();
		Map<ClassId, Set<ClassId>> hierarchy = new HashMap<>();
		sw.append("Class;KbValues;NewValues;ValuesCoveredByChildren;KbEntities;NewEntities;EntitiesCoveredByChildren;ConflictKB;ConflictWT" + "\n");
		for (ClassId topClass : topClasses) {
			Set<ClassId> innerRet = new HashSet<>();
			hierarchy.put(topClass, new HashSet<>());
			printClassAndSubClasses(sw, innerRet, hierarchy, "-", topClass, kbValuesCount, kbEntitiesCount, newValuesCount, newEntitesCount, includedChildren,
					minValuesOfParent, minEntitiesOfParent, conflictResults);

			ret.addAll(innerRet);

		}

		return new Pair<>(ret, hierarchy);
	}

	public static void printClassAndSubClasses(StringWriter sw, Set<ClassId> ret, Map<ClassId, Set<ClassId>> hierarchy, String prefix, ClassId classId,
			Map<ClassId, Integer> kbValuesCount, Map<ClassId, Integer> kbEntitiesCount, Map<ClassId, Integer> newValuesCount,
			Map<ClassId, Integer> newEntitesCount, Map<ClassId, Set<ClassId>> includedChildren, double minValuesOfParent, double minEntitiesOfParent,
			Map<ClassId, ConflictData> conflictResults) {

		if (ret.contains(classId))
			return;

		double sumOfValues = 0.0;
		double sumOfEntities = 0.0;
		if (includedChildren.containsKey(classId)) {
			for (ClassId childClassId : includedChildren.get(classId)) {
				sumOfValues += newValuesCount.get(childClassId);
				sumOfEntities += newEntitesCount.get(childClassId);
			}
		}

		double valuesCoveredByChildren = sumOfValues / newValuesCount.get(classId);
		double entitiesCoveredByChildren = sumOfEntities / newEntitesCount.get(classId);

		// GenericAccumaltableResult profiling =
		// profilingResults.results.getPerClassId().get(classId);
		//
		// AccumRepository repository = profiling.getAcummulatorRepository();

		String conflictKB = "";
		String conflictMatch = "";

		ConflictData conflictData = conflictResults.get(classId);

		if (conflictData == null) {
			System.out.println("CONFLICT DATA MISSING FOR CLASS " + classId.getClassId());
		} else {
			conflictKB = DecimalFormatter.format(conflictData.getPotential());
			conflictMatch = DecimalFormatter.format(conflictData.getPotentialMatch());
		}

		sw.append(prefix + " " + classId.getFinalPart() + ";" + kbValuesCount.get(classId) + ";" + newValuesCount.get(classId) + ";"
				+ DecimalFormatter.format(valuesCoveredByChildren) + ";" + kbEntitiesCount.get(classId) + ";" + newEntitesCount.get(classId) + ";"
				+ DecimalFormatter.format(entitiesCoveredByChildren) + ";" + conflictKB + ";" + conflictMatch + "\n");

		if (valuesCoveredByChildren > minValuesOfParent && entitiesCoveredByChildren > minEntitiesOfParent) {
			Set<ClassId> innerRet = new HashSet<>();
			for (ClassId childClassId : includedChildren.get(classId)) {
				printClassAndSubClasses(sw, ret, hierarchy, prefix + "-", childClassId, kbValuesCount, kbEntitiesCount, newValuesCount, newEntitesCount,
						includedChildren, minValuesOfParent, minEntitiesOfParent, conflictResults);
			}
			hierarchy.put(classId, new HashSet<>(includedChildren.get(classId)));
			ret.addAll(innerRet);

		} else {
			ret.add(classId);
		}
	}
}
