package de.uni_mannheim.informatik.dws.expansion.samples.setExtension;

import java.util.HashSet;
import java.util.Optional;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;

public class SettlementTableFixer implements AnnotatedTableFilter {

	private final KnowledgeBase kb;
	private final OntologyRepository ontologyRepository;
	private final Set<ClassId> classes;

	public SettlementTableFixer(KnowledgeBase kb, OntologyRepository ontologyRepository) {
		this.kb = kb;
		this.ontologyRepository = ontologyRepository;
		this.classes = new HashSet<>();
		this.classes.add(ClassConstants.place);
		this.classes.addAll(ontologyRepository.getSubClasses(ClassConstants.place));
	}

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {
		AnnotatedTableMetaData md = data.getMetaData();
		if (this.classes.contains(md.getDbPediaClass())) {
			double total = 0.0;
			int count = 0;
			for (InstanceId instanceId : md.getInstances().values()) {
				total++;
				Set<ClassId> kbClasses = kb.getClasses(instanceId);
				if (kbClasses != null && kbClasses.contains(ClassConstants.placeClass)) {
					count++;
				}
			}
			double proportion = count / total;
			if (proportion > 0.70) {
				md = md.rebuilder().setClassId(ClassConstants.placeClass).build();
				return Optional.of(new AnnotatedTablesReaderData(data.getId(), md, data.getTableData()));

			} else if (proportion > 0.50) {
				return Optional.of(data);
			} else {
				return Optional.of(data);
			}

		} else {
			return Optional.of(data);
		}
	}

}
