package de.uni_mannheim.informatik.dws.expansion.samples.setExtension;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;

public class TableClassMapper implements AnnotatedTableFilter {

	private Map<ClassId, ClassId> invertedMap = new HashMap<>();

	public TableClassMapper(Map<ClassId, Set<ClassId>> topClassesMap) {

		Map<ClassId, Set<ClassId>> topClassesMapCopy = new HashMap<>();
		for (Entry<ClassId, Set<ClassId>> entry : topClassesMap.entrySet()) {
			Set<ClassId> newSet = new HashSet<>(entry.getValue());
			newSet.add(entry.getKey());
			topClassesMapCopy.put(entry.getKey(), newSet);
		}
		for (Entry<ClassId, Set<ClassId>> entry : topClassesMapCopy.entrySet()) {

			for (ClassId innerClass : entry.getValue()) {
				if (invertedMap.containsKey(innerClass)) {
					IllegalStateException e = new IllegalStateException("Class " + innerClass.getClassId() + " already in index as part of top class "
							+ invertedMap.get(innerClass).getClassId() + ", when trying to add for top class " + entry.getKey().getClassId());
					e.printStackTrace();
				} else {
					invertedMap.put(innerClass, entry.getKey());
				}
			}
		}

	}

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {

		ClassId oldClassId = data.getMetaData().getDbPediaClass();

		if (invertedMap.containsKey(oldClassId)) {

			AnnotatedTableMetaData md = data.getMetaData().rebuilder().setClassId(invertedMap.get(oldClassId)).build();

			return Optional.of(new AnnotatedTablesReaderData(data.getId(), md, data.getTableData()));
		} else {
			// System.out.println("Attention: This table is of a class that is
			// not part of any top Class. Class: " + oldClassId.getClassId());
			return Optional.empty();
		}
	}

}
