package de.uni_mannheim.informatik.dws.expansion.samples.setExtension;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.SortedMap;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId.InstanceIdStyle;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilterRunner;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTablesWriterConcurrent;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig.TimeDependence;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.tablefusion.modules.AbstractsModule;
import de.mannheim.uni.expansion.tablefusion.modules.SimilarityModule;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategyRepository;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.utils.kryo.CHMKryoAdapter;
import de.uni_mannheim.dws.expansion.data.utils.kryo.ConcurrentKryoWriter;
import de.uni_mannheim.informatik.dws.attribute.AttributeMappingsFilter;
import de.uni_mannheim.informatik.dws.attribute.evaluation.AttributeMappingsEvaluator;
import de.uni_mannheim.informatik.dws.attribute.evaluation.AttributeMappingsPerformance;
import de.uni_mannheim.informatik.dws.attribute.matcher.AttributeMatcher;
import de.uni_mannheim.informatik.dws.attribute.model.AttributeMappings;
import de.uni_mannheim.informatik.dws.attribute.model.PropertyMapping;
import de.uni_mannheim.informatik.dws.attribute.model.TableAttributeMappings;
import de.uni_mannheim.informatik.dws.expansion.matching.KnowledgeBaseModule;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;
import de.uni_mannheim.informatik.dws.expansion.utils.remote.RepositoryLoading;
import expansion.set.attributeMatching.AttributeMatchingPipeline;
import expansion.set.attributeMatching.AttributeMatchingWeightLearner;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.evaluation.SetExpansionStrategiesConfig.SetExpansionStrategiesConfigBuilder;
import expansion.set.gscreation.GsOutputReader;
import expansion.set.optimizing.WeightsAdapter;

public class TablesAdapterRun {
	public static void main(String[] args) throws IOException {

		Map<ClassId, Set<ClassId>> topClasses = ExtractTablesForGoldStandardRun.getClassesWithHierarchyFilteredOrNot();
		// topClasses.remove(ClassConstants.single);
		// topClasses.remove(ClassConstants.placeClass);
		// topClasses.add(ClassConstants.placeClass);
		// topClasses.add(ClassConstants.person);
		// topClasses.add(ClassConstants.populatedPlace);

		KnowledgeBaseModule.CACHE_KEY_INDEX_STATICALLY = true;
		AbstractsModule.CACHE_ABSTRACTS_STATICALLY = true;
		SimilarityModule.CACHE_NUMERIC_RANGE_STATICALLY = true;

		int currentRun = 0;
		if (args.length > 5) {
			currentRun = Integer.parseInt(args[5]);
		}

		// this learns the models for the current run
		GSRun.main(new String[] { args[0], args[1], args[2], args[3], args[4], args[5], "true", GSRun.LEARN_MODELS });

		PersistentParallel6.restart();

		boolean alwaysDisableBlocking = false;
		SetExtension setExtension;
		// if (args.length > 4 && args[4].equals("true")) {
		setExtension = new SetExtension(args, topClasses.keySet(), ExtractTablesForGoldStandardRun.getMinPropertyDensity(),
				RepositoryLoading.REMOTE_LOAD_REMOTELY, currentRun, alwaysDisableBlocking);
		// } else {
		// setExtension = new SetExtension(args, topClasses.keySet(),
		// ExtractTablesForGoldStandardRun.getMinPropertyDensity(),
		// RepositoryLoading.REMOTE_LOAD_REMOTELY);
		// }

		String adaptedCacheFolder = setExtension.getMappingsFolder() + "/caches/tables_" + (currentRun + 1) + "/";

		AnnotatedTablesWriterConcurrent annotatedTablesWriter = new AnnotatedTablesWriterConcurrent(adaptedCacheFolder, null);
		if (!annotatedTablesWriter.cacheExists()) {
			AnnotatedTablesRepository tableRepository = setExtension.getGraph().get(AnnotatedTablesRepository.class);

			TablesAdapter tableAdapter = setExtension.getGraph().get(TablesAdapter.class);

			String clusteringsWeightsFolderPrefix = setExtension.getMappingsCachesFolder() + "/models/clusteringModelsFull";

			String instanceMatchingWeightsFolderPrefix = setExtension.getMappingsCachesFolder() + "/models/instanceMatchingModelsFull";

			Map<ClassId, Set<PropertyId>> nonTailProperties = new HashMap<>(setExtension.getNonTailProperties());

			boolean secondary = true;
			// if (args.length > 7 && args[7].equals("false")) {
			// secondary = false;
			// }

			boolean allTablesMatch = false;
			// if (args.length > 8 && args[8].equals("true")) {
			// allTablesMatch = true;
			// }

			boolean store = true;
			// if (args.length > 9 && args[9].equals("false")) {
			// store = false;
			// }

			// if (args.length > 10) {
			//
			// int classToRun = Integer.parseInt(args[10]);
			// switch (classToRun) {
			// case 0:
			// topClasses.remove(ClassConstants.single);
			// topClasses.remove(ClassConstants.placeClass);
			// nonTailProperties.remove(ClassConstants.single);
			// nonTailProperties.remove(ClassConstants.placeClass);
			// break;
			// case 1:
			// topClasses.remove(ClassConstants.gridIronFootballPlayer);
			// topClasses.remove(ClassConstants.placeClass);
			// nonTailProperties.remove(ClassConstants.gridIronFootballPlayer);
			// nonTailProperties.remove(ClassConstants.placeClass);
			// break;
			// case 2:
			// topClasses.remove(ClassConstants.gridIronFootballPlayer);
			// topClasses.remove(ClassConstants.single);
			// nonTailProperties.remove(ClassConstants.gridIronFootballPlayer);
			// nonTailProperties.remove(ClassConstants.single);
			// break;
			// }
			// }

			tableAdapter.run(tableRepository, setExtension.getMappingsFolder(), adaptedCacheFolder, nonTailProperties, topClasses,
					clusteringsWeightsFolderPrefix, instanceMatchingWeightsFolderPrefix, setExtension.getMappingsCachesFolder(), currentRun + 1, secondary,
					allTablesMatch, store);

		}

		try {
			PersistentParallel4.close();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public static class TablesAdapter {

		@Inject
		protected WeightsAdapter weightsAdapter;

		// @Inject
		// protected AttributeThresholdLearner thresholdLearner;

		@Inject
		protected AttributeMatcher matcher;

		@Inject
		protected ConcurrentKryoWriter concurrentKryoWriter;

		@Inject
		protected AnnotatedTableFilterRunner filterRunner;

		@Inject
		protected AttributeMatchingPipeline pipeline;

		@Inject
		protected KnowledgeBase kb;

		@Inject
		protected AttributeMappingsEvaluator evaluator;

		@Inject
		protected CHMKryoAdapter chmKryoAdapter;

		public TablesAdapter() {
			// TODO Auto-generated constructor stub
		}

		public void run(AnnotatedTablesRepository repository, String mappingsFolder, String adaptedTablesCacheFolder,
				Map<ClassId, Set<PropertyId>> nonTailProperties, Map<ClassId, Set<ClassId>> allParents, String clusteringWeightsFolderPrefix,
				String instanceMatchingWeightsFolderPrefix, String cachesFolder, int currentRun, boolean secondary, boolean allTablesMatch, boolean store)
				throws IOException {

			AnnotatedTablesWriterConcurrent annotatedTablesWriter = new AnnotatedTablesWriterConcurrent(adaptedTablesCacheFolder, concurrentKryoWriter);
			if (!annotatedTablesWriter.cacheExists()) {

				if (true) {

					String attributeMappingsCacheFolder = cachesFolder + "/AttributeMappingsCacheFolder/";
					MKDIR.createDirUnchecked(new File(attributeMappingsCacheFolder));

					boolean hasCache = false;
					for (ClassId classId : allParents.keySet()) {
						if (chmKryoAdapter.cacheExists(attributeMappingsCacheFolder, getPrefix(classId))) {
							hasCache = true;
						}
					}

					AttributeMappings newMappings;

					GsOutputReader reader = new GsOutputReader(InstanceIdStyle.DBPEDIA_URI);
					String gsOutputFolder = mappingsFolder + "/goldStandard/";

					String tableClassesFile = gsOutputFolder + "/classes_GS.csv";
					String processedTablesFile = gsOutputFolder + "/processedTables.csv";
					Set<AnnotatedTableId> processedTables = reader.readTableList(processedTablesFile);
					Map<AnnotatedTableId, ClassId> tableClasses = reader.readTableClasses(tableClassesFile);

					processedTables = processedTables.stream().filter((e) -> allParents.keySet().contains(tableClasses.get(e))).collect(Collectors.toSet());

					AttributeMappings gsMappings = reader.getAttributeMappings(gsOutputFolder + "/attributes/", processedTables, tableClasses,
							nonTailProperties);

					List<AnnotatedTablesReaderData> reMatch = new ArrayList<>();
					Set<AnnotatedTableId> reMatchIds = new HashSet<>();
					List<AnnotatedTablesReaderData> keep = new ArrayList<>();

					for (AnnotatedTablesReaderData data : repository.getTableReaderData()) {
						// if
						// (data.getMetaData().getDbPediaClass().equals(ClassConstants.placeClass))
						// {
						// keep.add(data);
						// } else
						if (data.getId().getTableName().equals("1346981172231_1347011264447_3050.arc9167679534671747539#2935814_0_7628926464279511553")) {
							int sdf = 123;
							if (processedTables.contains(data.getId())) {
								int sggggdf = 123;
							}
						}

						if (allParents.keySet().contains(data.getMetaData().getDbPediaClass())) {

							if (allTablesMatch) {
								reMatch.add(data);
								reMatchIds.add(data.getId());
							} else {
								if (!gsMappings.tableSet().contains(data.getId())) {
									keep.add(data);
								} else {
									reMatch.add(data);
									reMatchIds.add(data.getId());
								}
							}
						}
					}

					if (!hasCache) {

						String forLearningTablesFile = gsOutputFolder + "/forLearningTables.csv";
						Set<AnnotatedTableId> forLearningTables = reader.readTableList(forLearningTablesFile);

						AttributeMappings forLearningMappingsNonFinal = reader.getAttributeMappings(gsOutputFolder + "/attributes/", forLearningTables,
								tableClasses, nonTailProperties);
						AttributeMappings forLearningMappings = gsMappings;

						if (currentRun >= 0) {

							FusionConfig fusionConfig = SetExtension.getFusionConfig();

							SetExpansionStrategiesConfigBuilder builder = SetExtension.getSetExpansionStrategiesConfig().rebuilder();

							builder.clearFusionSetups().addFusionSetup(FusionStrategyRepository.COLUMN_SCORES,
									PropertiesConfig.builder().setTimeDependence(TimeDependence.NO).build());
							SetExpansionStrategiesConfig setExpansionStrategiesConfig = builder.build();

							StrategiesConfig strategiesConfig = StrategiesConfig.defaultConfig();

							String matchingPipelineCachesFolder = cachesFolder + "/models/matchingPipelineSecondary";
							MKDIR.createDirUnchecked(new File(matchingPipelineCachesFolder));

							String matchingPipelineWeightsCachesFolder = cachesFolder + "/models/matchingPipelineWeights";
							MKDIR.createDirUnchecked(new File(matchingPipelineWeightsCachesFolder));

							SortedMap<Integer, AttributeMappings> pipelineMappings = pipeline.runPipeline(setExpansionStrategiesConfig, fusionConfig,
									strategiesConfig, reMatchIds, allParents, nonTailProperties, currentRun, currentRun, forLearningMappings, gsMappings,
									clusteringWeightsFolderPrefix, instanceMatchingWeightsFolderPrefix, matchingPipelineCachesFolder,
									matchingPipelineWeightsCachesFolder, secondary, false, false);

							newMappings = pipelineMappings.get(currentRun);
							FusionPerformance<AttributeMappingsPerformance> newMappingsPeformance = evaluator.evaluate(gsMappings, newMappings,
									kb.getNormalizationIndex());

							// FusionPerformance<AttributeMappingsPerformance>
							// pipelinePerformance1 =
							// evaluator.evaluate(gsMappings,
							// pipelineMappings.get(1),
							// kb.getNormalizationIndex());

							ColumnReMatcherRun.print("PIPELINE_" + currentRun, newMappingsPeformance);
							// ColumnReMatcherRun.print("PIPELINE_1",
							// pipelinePerformance1);

						} else {
							newMappings = new AttributeMappings(
									reMatchIds.stream().collect(Collectors.toMap(Function.identity(), (e) -> new TableAttributeMappings(new HashMap<>()))),
									gsMappings.getClassIds().entrySet().stream().filter((e) -> reMatchIds.contains(e.getKey()))
											.collect(Collectors.toMap(Entry::getKey, Entry::getValue)));
						}

					} else {
						newMappings = readCache(attributeMappingsCacheFolder, allParents.keySet());

						FusionPerformance<AttributeMappingsPerformance> newMappingsPeformance = evaluator.evaluate(gsMappings, newMappings,
								kb.getNormalizationIndex());

						// FusionPerformance<AttributeMappingsPerformance>
						// pipelinePerformance1 =
						// evaluator.evaluate(gsMappings,
						// pipelineMappings.get(1),
						// kb.getNormalizationIndex());

						ColumnReMatcherRun.print("PIPELINE_" + currentRun, newMappingsPeformance);
					}

					if (store) {
						List<AnnotatedTablesReaderData> finalList = new ArrayList<>(filterRunner.asList(reMatch, new AttributeMappingsFilter(newMappings)));
						finalList.addAll(keep);

						MKDIR.createDirUnchecked(new File(adaptedTablesCacheFolder));

						annotatedTablesWriter.writeAll(finalList);
					}

					// ThresholdContainer thresholds = new
					// ThresholdContainerWithFallback(
					// thresholdLearner.findThreshold(repository.getTableReaderData(gsMappings.tableSet()),
					// nonTailProperties, gsMappings, 4, (tableId) -> {
					// return gsMappings.getClassId(tableId);
					// }), new ThresholdConstant(0.25));
					//
					// System.out.println(thresholds.toString());
					//
					// AttributeMappings yaserThresholdMappings =
					// matcher.match(reMatch,
					// thresholds, nonTailProperties);
					if (!hasCache && !store) {

						for (ClassId classId : allParents.keySet()) {
							Map<AnnotatedTableId, TableAttributeMappings> forClass = AttributeMatchingWeightLearner.getForClass(newMappings, classId);
							writeCache(attributeMappingsCacheFolder, classId, forClass);

						}
					}

				} else {
					annotatedTablesWriter.writeAll(repository.getTableReaderData());

				}
			}

		}

		private void writeCache(String attributeMappingsCacheFolder, ClassId classId, Map<AnnotatedTableId, TableAttributeMappings> forClass) {
			Map<AnnotatedTableId, Map<ColumnIdx, PropertyMapping>> remapped = forClass.entrySet().stream().filter((e) -> !e.getValue().getMappings().isEmpty())
					.collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().getMappings()));
			chmKryoAdapter.writeWithMap(attributeMappingsCacheFolder, getPrefix(classId), remapped);

		}

		private AttributeMappings readCache(String attributeMappingsCacheFolder, Set<ClassId> classIds) {
			Map<AnnotatedTableId, TableAttributeMappings> map = new HashMap<>();
			Map<AnnotatedTableId, ClassId> classMap = new HashMap<>();
			for (ClassId classId : classIds) {
				Map<AnnotatedTableId, Map<ColumnIdx, PropertyMapping>> read = chmKryoAdapter.readWithMap(AnnotatedTableId.class, ColumnIdx.class,
						PropertyMapping.class, attributeMappingsCacheFolder, getPrefix(classId));

				for (Entry<AnnotatedTableId, Map<ColumnIdx, PropertyMapping>> entry : read.entrySet()) {
					map.put(entry.getKey(), new TableAttributeMappings(entry.getValue()));
					classMap.put(entry.getKey(), classId);
				}
			}
			return new AttributeMappings(map, classMap);
		}

		private String getPrefix(ClassId classId) {
			return classId.getFinalPart() + "_";
		}
	}

}
