package expansion.samples;

import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.informatik.dws.expansion.samples.setExtension.SetExtension;

public class TestHierarchyStorer {
	public static void main(String[] args) {
		ClassId A = new ClassId("A");
		ClassId AA = new ClassId("AA");
		ClassId AAA = new ClassId("AAA");
		ClassId AAAA = new ClassId("AAAA");
		ClassId AB = new ClassId("AB");
		ClassId AC = new ClassId("AC");

		ClassId B = new ClassId("B");
		ClassId BA = new ClassId("BA");
		ClassId BAA = new ClassId("BAA");
		ClassId BAB = new ClassId("BAB");
		ClassId BB = new ClassId("BB");
		ClassId BC = new ClassId("BC");

		ClassId C = new ClassId("C");

		ClassId D = new ClassId("D");
		ClassId DA = new ClassId("DA");
		ClassId DAA = new ClassId("DAA");
		ClassId DAAA = new ClassId("DAAA");

		Map<ClassId, Set<ClassId>> hierarchy = new HashMap<>();
		hierarchy.put(A, new HashSet<>(Arrays.asList(AA, AB, AC)));
		hierarchy.put(AA, new HashSet<>(Arrays.asList(AAA)));
		hierarchy.put(AAA, new HashSet<>(Arrays.asList(AAAA)));

		hierarchy.put(B, new HashSet<>(Arrays.asList(BA, BB, BC)));
		hierarchy.put(BA, new HashSet<>(Arrays.asList(BAA, BAB)));

		hierarchy.put(C, new HashSet<>());

		hierarchy.put(D, new HashSet<>(Arrays.asList(DA)));
		hierarchy.put(DA, new HashSet<>(Arrays.asList(DAA)));
		hierarchy.put(DAA, new HashSet<>(Arrays.asList(DAAA)));

		String s = SetExtension.writeClassFilterHiearchyToString(hierarchy);
		System.out.println(s);

		List<Pair<ClassId, ?>> hierarchyNew = SetExtension.getClassFilterHierarchy(Arrays.asList(s.split("\n")));

		boolean correct = checkCorrect(hierarchy, hierarchyNew);
		if (correct) {
			System.out.println("CORRECT");
		} else {
			System.out.println("BAD");
		}

	}

	public static boolean checkCorrect(Map<ClassId, Set<ClassId>> hierarchy, List<Pair<ClassId, ?>> hierarchyNew) {
		for (Pair<ClassId, ?> h : hierarchyNew) {
			if (h.getSecond() instanceof List<?>) {

				List<Pair<ClassId, ?>> list = (List<Pair<ClassId, ?>>) h.getSecond();
				boolean inner = checkCorrectInner(hierarchy, h.getFirst(), list);
				if (!inner) {
					return false;
				}
			} else {
				if (!hierarchy.containsKey(h.getFirst())) {
					return false;
				}
			}
		}

		return true;
	}

	private static boolean checkCorrectInner(Map<ClassId, Set<ClassId>> hierarchy, ClassId classId, List<Pair<ClassId, ?>> list) {
		Set<ClassId> innerSet = new HashSet<>();
		for (Pair<ClassId, ?> h : list) {
			innerSet.add(h.getFirst());

			if (h.getSecond() instanceof List<?>) {

				List<Pair<ClassId, ?>> newList = (List<Pair<ClassId, ?>>) h.getSecond();
				boolean inner = checkCorrectInner(hierarchy, h.getFirst(), newList);
				if (!inner) {
					return false;
				}
			}

		}

		if (hierarchy.containsKey(classId) && innerSet.equals(hierarchy.get(classId))) {
			return true;
		} else {
			return false;
		}

	}
}
