package expansion.set.attributeMatching;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.InstanceIndexComposite;
import de.mannheim.uni.expansion.data.tables.model.PropertyIndexComposite;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdsSet;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilterRunner;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.data.tables.repository.memory.AnnotatedTablesRepositoryImplMemory;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.tablefusion.learning.RowValuesGetter;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;
import de.mannheim.uni.expansion.tablefusion.learning.SplitRowValues;
import de.mannheim.uni.expansion.tablefusion.services.AnnotatedTableInstanceScorer;
import de.mannheim.uni.expansion.tablefusion.services.grouping.GroupedValuesRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.utils.kryo.CHMKryoAdapter;
import de.uni_mannheim.informatik.dws.attribute.matcher.AttributeMatcher;
import de.uni_mannheim.informatik.dws.attribute.matcher.AttributeMatchingFullRuns;
import de.uni_mannheim.informatik.dws.attribute.model.AttributeMappings;
import de.uni_mannheim.informatik.dws.attribute.model.PropertyMapping;
import de.uni_mannheim.informatik.dws.attribute.model.TableAttributeMappings;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdContainer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.LearnedModelRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepositoryImpl;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusterer;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.creation.Entity;
import expansion.set.creation.EntityCreator;
import expansion.set.entitymatching.EntityMatcherWeightedResults;
import expansion.set.entitymatching.EntityMatchingRunner;
import expansion.set.entitymatching.matchers.EntityMatcherConfig;
import expansion.set.entitymatching.modelLearning.EM_LearnedModelRepository;
import expansion.set.entitymatching.run.EntityMatchingRunRepository;
import expansion.set.evaluation.SetExpansionRunner;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.evaluation.crossvalidation.EntitySetGroupedValuesSplitter;
import expansion.set.evaluation.matching.EntityMatcher;
import expansion.set.evaluation.matching.ScoredEntity;
import expansion.set.implicitAttributes.ImplicitAttributeFinder;
import expansion.set.optimizing.LearnedModelAdapter;
import expansion.set.optimizing.WeightsAdapter;
import expansion.set.strategies.SetExpansionStrategiesRepository;
import expansion.set.strategies.SetExpansionStrategy;
import expansion.set.strategies.rowclusterers.GreedyClusteringParallelSplitter;
import expansion.set.strategies.rowclusterers.RowClustererRepository;
import expansion.set.strategies.rowclusterers.label.LabelRowSimilaritiesCreator;


public class AttributeMatchingPipeline {

	@Inject
	protected AnnotatedTableFilterRunner filterRunner;
	@Inject
	protected RowValuesGetter rowValuesGetter;

	@Inject
	protected AttributeMatcher attributeMatcher;

	@Inject
	protected AnnotatedTablesRepository tablesRepository;

	@Inject
	protected RowClustererRepository rowClustererRepository;

	@Inject
	protected GreedyClusteringParallelSplitter greedyClusteringSplitter;

	@Inject
	protected EntityMatcher entityMatcher;

	@Inject
	protected SetExpansionRunner setExpansionRunner;

	@Inject
	protected EntityCreator entityCreator;

	@Inject
	protected EntitySetGroupedValuesSplitter splitter;

	// @Inject
	// protected AttributeThresholdLearner thresholdLearner;


	@Inject
	protected GroupedValuesRepository groupedValuesRepository;

	@Inject
	protected AnnotatedTableInstanceScorer instanceScorer;

	@Inject
	protected CHMKryoAdapter chmKryoAdapter;

	@Inject
	protected AttributeMatchingWeightLearner weightLearner;

	@Inject
	protected LearnedModelAdapter learnedModelAdapter;

	@Inject
	protected WeightsAdapter weightsAdapter;

	@Inject
	protected EntityMatchingRunner entityMatchingRuner;

	@Inject
	protected KnowledgeBase kb;

	// @Inject
	// protected EntityMatchingGS_ClustererWeightsAdapterAndLearner
	// instanceMatchingWeightsLearner;

	@Inject
	protected OntologyRepository ontologyRepository;

	@Inject
	protected LabelRowSimilaritiesCreator labelQuick;
	public static final String PREFIX = "MATCHING_PIPELINE";

	@Inject
	protected ImplicitAttributeFinder implicitAttributeFinder;

	public AttributeMatchingPipeline() {
	}

	public SortedMap<Integer, AttributeMappings> runPipeline(SetExpansionStrategiesConfig setExpansionStrategiesConfig,
			FusionConfig fusionConfig, StrategiesConfig strategiesConfig, Set<AnnotatedTableId> tableIds,
			Map<ClassId, Set<ClassId>> allParents,
			Map<ClassId, Set<PropertyId>> nonTailProperties, int begin, int repeats, AttributeMappings forLearningMappings,
			AttributeMappings forTestingMappings,
			String clusteringWeightsFolderPrefix, String instanceMatchingWeightsFolderPrefix,
			String cacheFolder, String weightsFolder, boolean secondary, boolean baseline, boolean alwaysDisableBlocking)
			throws IOException {

		Set<AnnotatedTableId> forLearningTableIds = forLearningMappings.getMappings().keySet();

		Set<AnnotatedTableId> fullTableIds = new HashSet<>(tableIds);
		fullTableIds.addAll(forLearningMappings.tableSet());
		fullTableIds.addAll(forTestingMappings.tableSet());

		Map<AnnotatedTableId, ClassId> tableToClassIndex = new HashMap<>();
		List<AnnotatedTablesReaderData> tableData = new ArrayList<>();

		for (AnnotatedTableId tableId : fullTableIds) {
			AnnotatedTable table = tablesRepository.getTable(tableId);
			AnnotatedTableMetaData md = tablesRepository.getMetaData(tableId);
			md = md.rebuilder().setPropertiesWithScores(new HashMap<>(), new HashMap<>()).setInstancesWithScores(new HashMap<>(), new HashMap<>()).build();
			tableData.add(new AnnotatedTablesReaderData(tableId, md, table));
			tableToClassIndex.put(tableId, md.getDbPediaClass());
		}

		SortedMap<Integer, AttributeMappings> ret = new TreeMap<>();

		Map<ClassId, WeightsRepositoryImpl> weights = weightsAdapter.readWeights(allParents.keySet(),
				weightsFolder);
		boolean learnedNewWeights = false;

		for (int i = 0; i <= repeats; i++) {

			if (i == 0)
				baseline = true;
			else
				baseline = false;
			// if (i == 0) {
			// GreedyClustering.minBlocks = 1;
			// }

			OptimizableClass optimizableClazz = new OptimizableClass("PIPELINE_MATCHING_" + i);
			OptimizableClass optimizableClazzSecondary = new OptimizableClass("PIPELINE_MATCHING_" + i + "_SECONDARY");

			if (!baseline) {
				Map<TableRowId, InstanceScorePair> instanceMappings;
				boolean cacheExists = false;
				for (ClassId classId : allParents.keySet()) {
					String prefix = getPrefix(i, classId);
					if (chmKryoAdapter.cacheExists(cacheFolder, prefix)) {
						cacheExists = true;
					} else if (cacheExists) {
						throw new IllegalStateException("SOME CACHE EXISTS, SOME DONT");
					}
				}
				if (cacheExists) {
					instanceMappings = new HashMap<>();
					for (ClassId classId : allParents.keySet()) {
						String prefix = getPrefix(i, classId);
						instanceMappings.putAll(chmKryoAdapter.read(TableRowId.class, InstanceScorePair.class, cacheFolder, prefix));
					}
				} else {
					String clusteringWeightsFolder = clusteringWeightsFolderPrefix + (i - 1) + "/";

					Map<ClassId, LearnedModelRepository> clusteringWeightsRepository = learnedModelAdapter
							.readModels((modelMap) -> new LearnedModelRepository(modelMap), allParents.keySet(), clusteringWeightsFolder);

					String instanceMatchingWeightsFolder = instanceMatchingWeightsFolderPrefix + (i - 1) + "/";
					// MKDIR.createDirUnchecked(new
					// File(instanceMatchingWeightsFolder));
					Map<ClassId, EM_LearnedModelRepository> instanceMatchingWeightsRepository = learnedModelAdapter.readModels(
							(modelMap) -> new EM_LearnedModelRepository(modelMap), allParents.keySet(),
							instanceMatchingWeightsFolder);
					List<AnnotatedTablesReaderData> valueAdaptedTableData = filterRunner.asList(tableData,
							AttributeMatcher.getValueAdaptionFilters(ontologyRepository));
					boolean quickRun = false;
					if (i == 1)
						quickRun = true;
					instanceMappings = run(valueAdaptedTableData, setExpansionStrategiesConfig, fusionConfig, strategiesConfig, allParents,
							clusteringWeightsRepository, instanceMatchingWeightsRepository, nonTailProperties, quickRun, alwaysDisableBlocking);

					Map<ClassId, Map<TableRowId, InstanceScorePair>> cacheMap = new HashMap<>();
					for (Entry<TableRowId, InstanceScorePair> entry : instanceMappings.entrySet()) {
						CHMHelper.getFromMap(cacheMap, tableToClassIndex.get(entry.getKey().getTableId()), () -> new HashMap<>()).put(entry.getKey(),
								entry.getValue());
					}
					for (Entry<ClassId, Map<TableRowId, InstanceScorePair>> entry : cacheMap.entrySet()) {
						chmKryoAdapter.write(cacheFolder, getPrefix(i, entry.getKey()), entry.getValue());

					}
				}

				tableData = recreateWithInstanceMappings(tableData, instanceMappings);
			}

			// ThresholdContainer thresholds = new
			// ThresholdContainerWithFallback(
			// thresholdLearner.findThreshold(tableData, nonTailProperties,
			// gsMappings, 4, (tableId) -> {
			// return gsMappings.getClassId(tableId);
			// }), new ThresholdConstant(0.25));

			// AttributeMappings mappings = attributeMatcher.match(tableData,
			// thresholds, nonTailProperties);

			// HERE ADD SPLIT, STORAGE, AND QUICKAPPLY
			boolean containsWeights = true;
			for (ClassId classId : allParents.keySet()) {
				if (!weights.get(classId).containsWeightsForClass(optimizableClazz) || !weights.get(classId).containsWeightsForClass(optimizableClazzSecondary)

				) {
					containsWeights = false;
					break;
				}
			}

			AttributeMappings mappings;
			if (containsWeights) {

				mappings = attributeMatcher.match(tableData, weights, nonTailProperties, optimizableClazz, optimizableClazzSecondary, tableIds, secondary,
						baseline);
				ret.put(i, mappings);

			} else {
				learnedNewWeights = true;

				List<AnnotatedTablesReaderData> forLearningTableData = tableData.stream().filter((e) -> forLearningTableIds.contains(e.getId()))
						.collect(Collectors.toList());

				AttributeMatchingFullRuns fullRuns = attributeMatcher.getFullMatches(forLearningTableData, nonTailProperties, null, baseline);

				Map<ClassId, WeightsRepositoryImpl> newWeights = weightLearner.learn(optimizableClazz, fullRuns.getByClass(), forLearningMappings,
						kb.getNormalizationIndex());

				for (Entry<ClassId, WeightsRepositoryImpl> entry : newWeights.entrySet()) {
					weights.put(entry.getKey(), weights.get(entry.getKey()).combineWith(entry.getValue()));

				}
				mappings = attributeMatcher.match(tableData, weights, nonTailProperties, optimizableClazz, optimizableClazzSecondary, tableIds, false,
						baseline);

				if (secondary) {
					AttributeMatchingFullRuns secondaryFullRuns = attributeMatcher.getFullMatches(forLearningTableData, nonTailProperties, mappings, baseline);

					Map<ClassId, WeightsRepositoryImpl> secondaryWeights = weightLearner.learn(optimizableClazzSecondary, secondaryFullRuns.getByClass(),
							forLearningMappings, kb.getNormalizationIndex());

					for (Entry<ClassId, WeightsRepositoryImpl> entry : secondaryWeights.entrySet()) {
						weights.put(entry.getKey(), weights.get(entry.getKey()).combineWith(entry.getValue()));

					}

					AttributeMappings secondaryMappings = attributeMatcher.match(tableData, weights, nonTailProperties, optimizableClazzSecondary, tableIds,
							mappings, baseline);

					// AttributeMappings secondaryMappings =
					// weightLearner.applyWeights(optimizableClazzSecondary,
					// secondaryFullRuns.getByClass(),
					// new HashMap<>(weights), tableIds);

					ret.put(i, secondaryMappings);
				}
				else {
					ret.put(i, mappings);
				}
			}


			if (i < repeats) {

				// AttributeMappings nonThresholdedMappings =
				// attributeMatcher.match(tableData, new
				// ThresholdConstant(0.25), nonTailProperties);
				tableData = recreateWithColumnMappings(tableData, mappings);
			}

			// if (i == 0) {
			// GreedyClustering.minBlocks = 2;
			// }
		}

		if (learnedNewWeights) {
			weightsAdapter.storeWeights(weightsFolder, weights);
		}
		return ret;
	}



	private String getPrefix(int i, ClassId classId) {
		return PREFIX + "_" + i + "_" + classId.getFinalPart() + "_";

	}

	public static List<AnnotatedTablesReaderData> recreateWithColumnMappings(List<AnnotatedTablesReaderData> tableData, AttributeMappings mappings) {

		Map<AnnotatedTableId, Map<Integer, List<PropertyId>>> propertyMappingsGrouped = new HashMap<>();
		Map<AnnotatedTableId, Map<PropertyIndexComposite, Double>> propertyMappingsScore = new HashMap<>();

		for (Entry<AnnotatedTableId, TableAttributeMappings> entry : mappings.getMappings().entrySet()) {
			Map<Integer, List<PropertyId>> mappingsGrouped = new HashMap<>();
			Map<PropertyIndexComposite, Double> scores = new HashMap<>();
			for (Entry<ColumnIdx, PropertyMapping> e : entry.getValue().getMappings().entrySet()) {
				List<PropertyId> propertyList = new ArrayList<>();
				propertyList.add(e.getValue().getPropertyId());
				mappingsGrouped.put(e.getKey().getColumnIdx(), propertyList);
				scores.put(new PropertyIndexComposite(e.getKey().getColumnIdx(), e.getValue().getPropertyId()), e.getValue().getScore());
			}
			propertyMappingsScore.put(entry.getKey(), scores);

			propertyMappingsGrouped.put(entry.getKey(), mappingsGrouped);
		}

		List<AnnotatedTablesReaderData> ret = new ArrayList<>();

		for (AnnotatedTablesReaderData tableReaderData : tableData) {
			AnnotatedTableMetaData md = tableReaderData.getMetaData();
			if (propertyMappingsGrouped.containsKey(tableReaderData.getId())) {
				Map<Integer, List<PropertyId>> properties = propertyMappingsGrouped.get(tableReaderData.getId());
				Map<PropertyIndexComposite, Double> scores = propertyMappingsScore.get(tableReaderData.getId());
				md = md.rebuilder().setPropertiesWithScores(properties, scores).build();

			} else {

				md = md.rebuilder().setPropertiesWithScores(new HashMap<>(), new HashMap<>()).build();
			}

			ret.add(new AnnotatedTablesReaderData(tableReaderData.getId(), md, tableReaderData.getTableData()));
		}

		return ret;
	}

	private List<AnnotatedTablesReaderData> recreateWithInstanceMappings(List<AnnotatedTablesReaderData> tableData,
			Map<TableRowId, InstanceScorePair> instanceMappings) {

		Map<AnnotatedTableId, Map<Integer, InstanceScorePair>> instanceMappingsGrouped = new HashMap<>();

		for (Entry<TableRowId, InstanceScorePair> entry : instanceMappings.entrySet()) {
			// List<InstanceId> instanceIds = new ArrayList<>();
			// instanceIds.add(entry.getValue());
			CHMHelper.getFromMap(instanceMappingsGrouped, entry.getKey().getTableId(), () -> new HashMap<>()).put(entry.getKey().getRowIdx().getRowIdx(),
					entry.getValue());
		}


		List<AnnotatedTablesReaderData> ret1 = new ArrayList<>();

		for (AnnotatedTablesReaderData tableReaderData : tableData) {
			AnnotatedTableMetaData md = tableReaderData.getMetaData();
			if (instanceMappingsGrouped.containsKey(tableReaderData.getId())) {
				Map<Integer, InstanceScorePair> instances = instanceMappingsGrouped.get(tableReaderData.getId());
				Map<Integer, List<InstanceId>> instancesLists = new HashMap<>();
				Map<InstanceIndexComposite, Double> scores = new HashMap<>();
				for (Entry<Integer, InstanceScorePair> entry : instances.entrySet()) {
					List<InstanceId> instanceIdList = new ArrayList<>();
					instanceIdList.add(entry.getValue().getFirst());
					instancesLists.put(entry.getKey(), instanceIdList);
					scores.put(new InstanceIndexComposite(entry.getKey(), entry.getValue().getFirst()), entry.getValue().getSecond());

				}
				md = md.rebuilder().setInstancesWithScores(instancesLists, scores).build();

			} else {

				md = md.rebuilder().setInstancesWithScores(new HashMap<>(), new HashMap<>()).build();
			}

			ret1.add(new AnnotatedTablesReaderData(tableReaderData.getId(), md, tableReaderData.getTableData()));
		}

		ConcurrentLinkedQueue<AnnotatedTablesReaderData> ret = new ConcurrentLinkedQueue<>();

		PersistentParallel6.runAndFinish(ret1, (data) -> {
			Map<InstanceIndexComposite, Double> instanceScores = instanceScorer.getScores(data);

			AnnotatedTableMetaData md = data.getMetaData().rebuilder().setInstanceScores(instanceScores).build();

			ret.add(new AnnotatedTablesReaderData(data.getId(), md, data.getTableData()));
		}, this.getClass());

		return new ArrayList<>(ret);
	}

	private Map<TableRowId, InstanceScorePair> run(List<AnnotatedTablesReaderData> tableData, SetExpansionStrategiesConfig setExpansionStrategiesConfig,
			FusionConfig fusionConfig, StrategiesConfig strategiesConfig, Map<ClassId, Set<ClassId>> allParents,
			Map<ClassId, LearnedModelRepository> weightsRepository, Map<ClassId, EM_LearnedModelRepository> instanceMatchingWeightsRepository,
			Map<ClassId, Set<PropertyId>> nonTailPropertiesMap, boolean quickRun, boolean alwaysDisableBlocking) throws IOException {

		implicitAttributeFinder.prepare(nonTailPropertiesMap);
		rowClustererRepository.init();

		AnnotatedTablesRepository tableRepository = new AnnotatedTablesRepositoryImplMemory(tableData);

		
		Map<TableRowId, InstanceScorePair> ret = new HashMap<>();
		Set<ClassId> classIds = new HashSet<>();

		Map<ClassId, Set<AnnotatedTableId>> tableIds = new HashMap<>();
		Map<ClassId, Set<TableRowId>> tableRowIds = new HashMap<>();

		for (AnnotatedTablesReaderData data : tableData) {
			ClassId classId = data.getMetaData().getDbPediaClass();
			classIds.add(classId);
			CHMHelper.getFromMap(tableIds, classId, () -> new HashSet<>()).add(data.getId());

			Set<TableRowId> candidates = CHMHelper.getFromMap(tableRowIds, classId, () -> new HashSet<>());


			int start = data.getMetaData().getNumHeaderRows();
			int end = data.getMetaData().getNumberOfRows() + data.getMetaData().getNumHeaderRows();

			for (int i = start; i < end; i++) {
				TableRowId rowId = new TableRowId(data.getId(), i);
				StringObject label = tablesRepository.getLabel(rowId);
				if (label != null)
					candidates.add(rowId);
			}

		}

		Set<TableRowId> rowIdsAllClasses = new HashSet<>();
		for (ClassId classId : allParents.keySet()) {
			Set<TableRowId> rowIds = tableRowIds.get(classId);
			rowIdsAllClasses.addAll(rowIds);
		}
		TableRowIdsSet rowIdsSetAllClasses = new TableRowIdsSet(rowIdsAllClasses, new UuidStringId("ALL"));

		for (ClassId classId : classIds) {

			TableIdsSet tableIdsSet = new TableIdsSet(this.tablesRepository.getAllTableIds(classId), new UuidStringId(classId.getFinalPart()));

			TableRowIdsSet rowIdsSet = new TableRowIdsSet(tableRowIds.get(classId), new UuidStringId(classId.getFinalPart()));
			TableIdsSet filteredTableIdsSet = new TableIdsSet(tableIds.get(classId), new UuidStringId(classId.getFinalPart()));

			GroupedValues groupedValues = groupedValuesRepository.getGroupedValues(tableRepository, tableRepository.getAllTableIds(), fusionConfig);
			SplitGroupedValues sgv = setExpansionRunner.getSplitGroupedValues(groupedValues, fusionConfig);
			SplitGroupedValuesSets splitGroupedValuesSets = new SplitGroupedValuesSets(sgv, sgv, sgv);
			
			SplitRowValues splitRowValues = rowValuesGetter.getSplitRowValues(tableRepository, filteredTableIdsSet);

			ClustererInput clustererInput = new ClustererInput(splitGroupedValuesSets, rowIdsSet.getRowIds(), rowIdsSet, tableIdsSet, null, classId,
					splitRowValues, allParents, nonTailPropertiesMap, rowIdsSetAllClasses);
			List<ThresholdContainer> thresholdContainers = setExpansionRunner.getThresholdContainers(setExpansionStrategiesConfig, null, classId, fusionConfig,
					strategiesConfig, sgv);

			ret.putAll(run(setExpansionStrategiesConfig, thresholdContainers, fusionConfig, strategiesConfig, allParents, clustererInput,
					weightsRepository.get(classId), instanceMatchingWeightsRepository.get(classId), nonTailPropertiesMap.get(classId), quickRun,
					alwaysDisableBlocking));

		}

		return ret;
	}

	private Map<TableRowId, InstanceScorePair> run(SetExpansionStrategiesConfig setExpansionStrategiesConfig,
			List<ThresholdContainer> thresholdContainers,
			FusionConfig fusionConfig, StrategiesConfig strategiesConfig, Map<ClassId, Set<ClassId>> allParents, ClustererInput clustererInput,
			LearnedModelRepository weightsRepository, EM_LearnedModelRepository instanceMatchingWeightsRepository, Set<PropertyId> nonTailProperties,
			boolean quickRun, boolean alwaysDisableBlocking) {

		SetExpansionStrategy strategy = SetExpansionStrategiesRepository.COMBINED_RF_BASE;

		RowClusteringConfig rowClusteringConfig = new RowClusteringConfig(setExpansionStrategiesConfig, thresholdContainers, weightsRepository, true, false,
				fusionConfig);


		RowClusterer clusterer = rowClustererRepository.get(strategy.getClusterer());



		RowClusterSimilarities sim = clusterer.clusterRows(rowClusteringConfig, clustererInput);

		ClusterList clusters;
		if (quickRun) {
			clusters = labelQuick.getQuickClusterList(clustererInput.getClassId(), clustererInput.getCandidates());
		} else {
			clusters = greedyClusteringSplitter.clusterRows(sim, rowClusteringConfig, clustererInput, alwaysDisableBlocking);
		}


		List<Entity> entities = entityCreator.createEntites(clustererInput.getGroupedValuesSets(), rowClusteringConfig.getThresholdContainers(), fusionConfig,
				strategiesConfig, rowClusteringConfig.getSetExpansionStrategiesConfig(), clusters.getClusterList(), clustererInput);

		EntityMatcherConfig emConfig = new EntityMatcherConfig(clustererInput, nonTailProperties, allParents);

		EntityMatcherWeightedResults matcherReturn = entityMatchingRuner.weightedRun(emConfig, EntityMatchingRunRepository.CS_LABEL_EM_ALL, entities,
				allParents, instanceMatchingWeightsRepository);

		// Tri<Map<InstanceId, List<ScoredEntity>>, List<Entity>, List<Entity>>
		// matcherReturn = entityMatcher.matchMultipleAndUnsureAndNew(entities,
		// allParents);


		Map<TableRowId, InstanceScorePair> ret = new HashMap<>();
		for (Entry<InstanceId, List<ScoredEntity>> entry : matcherReturn.getExistingEntities().entrySet()) {
			for (ScoredEntity entity : entry.getValue()) {
				for (TableRowId tableRowId : entity.getEntity().getRowCluster().getRows()) {
					ret.put(tableRowId, new InstanceScorePair(entry.getKey(), entity.getScore()));
				}
			}
		}

		for (ScoredEntity scoredEntity : matcherReturn.getNewEntities()) {
			Entity entity = scoredEntity.getEntity();
			String mainLabel = entity.getMainLabel();
			InstanceId instanceId;
			if (mainLabel != null) {
				instanceId = InstanceId.fromString(mainLabel + "-" + entity.getRowClusterUUID().toString());
			} else {
				instanceId = InstanceId.fromString(entity.getRowClusterUUID().toString());
			}

			for (TableRowId tableRowId : entity.getRowCluster().getRows()) {
				ret.put(tableRowId, new InstanceScorePair(instanceId, 0.0));
			}
		}

		return ret;

	}

	public static class InstanceScorePair extends Pair<InstanceId, Double> {

		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;

		public InstanceScorePair(InstanceId first, Double second) {
			super(first, second);
		}

	}
}
