package expansion.set.attributeMatching;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.informatik.dws.attribute.matcher.AttributeMatcherEquation;
import de.uni_mannheim.informatik.dws.attribute.matcher.AttributeMatchingFullRuns;
import de.uni_mannheim.informatik.dws.attribute.matcher.AttributeMatchingSingleRun;
import de.uni_mannheim.informatik.dws.attribute.matcher.AttributeMatchingSingleRun.ParameteredMatcherScore;
import de.uni_mannheim.informatik.dws.attribute.model.AttributeMappings;
import de.uni_mannheim.informatik.dws.attribute.model.PropertyMapping;
import de.uni_mannheim.informatik.dws.attribute.model.TableAttributeMappings;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.WeightsOptimizer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepositoryImpl;

public class AttributeMatchingWeightLearner {

	@Inject
	protected WeightsOptimizer weightsParallelOptimizer;

	@Inject
	protected AttributeMatcherEquation equation;

	public AttributeMatchingWeightLearner() {
	}

	public AttributeMappings applyWeights(OptimizableClass optimizableClazz, Map<ClassId, AttributeMatchingFullRuns> runs,
			Map<ClassId, WeightsRepository> weights, Set<AnnotatedTableId> tableIds) {
		ConcurrentHashMap<AnnotatedTableId, Map<ColumnIdx, PropertyMapping>> mappings = new ConcurrentHashMap<>();
		Map<AnnotatedTableId, ClassId> classIds = new HashMap<>();

		for (Entry<ClassId, AttributeMatchingFullRuns> entry : runs.entrySet()) {

			classIds.putAll(entry.getValue().getClassIds());

			ClassId classId = entry.getKey();

			WeightsRepository weightRepository = weights.get(entry.getKey());

			PersistentParallel6.runAndFinish(entry.getValue().getAllRuns(tableIds).entrySet(), (e) -> {
				Optional<PropertyMapping> maxPropertyId = equation.get(optimizableClazz, classId, e.getValue(), weightRepository);
				if (maxPropertyId.isPresent()) {
					CHMHelper.getFromCHM(mappings, e.getKey().getTableId(), () -> new ConcurrentHashMap<>()).put(e.getKey().getColumnIdx(),
							maxPropertyId.get());
				}
			}, this.getClass());
		}

		Map<AnnotatedTableId, TableAttributeMappings> mappingsFormed = mappings.entrySet().stream()
				.collect(Collectors.toMap(Entry::getKey, (e) -> new TableAttributeMappings(e.getValue())));
		return new AttributeMappings(mappingsFormed, classIds);
	}

	public Map<ClassId, WeightsRepositoryImpl> learn(OptimizableClass optimizableClazz,
			Map<ClassId, AttributeMatchingFullRuns> fullRuns,
			AttributeMappings gsMappings, Map<PropertyId, PropertyId> normalization) {
		
		Map<ClassId, WeightsRepositoryImpl> ret = new HashMap<>();

		for(Entry<ClassId, AttributeMatchingFullRuns> inputEntry : fullRuns.entrySet()) {

			ClassId classId = inputEntry.getKey();

			 Map<AnnotatedTableId, TableAttributeMappings> gsForClass = getForClass(gsMappings, classId);
			 
			Set<PropertyId> propertyIdSet = gsForClass.values().stream().flatMap((e) -> e.getMappings().values().stream()).map((e) -> e.getPropertyId())
					.collect(Collectors.toSet());
			propertyIdSet
					.addAll(inputEntry.getValue().getAllRuns().values().stream().flatMap((e) -> e.getScores().keySet().stream()).collect(Collectors.toSet()));

			WeightsRepositoryImpl repository = new WeightsRepositoryImpl();

			if (false) {
			for (PropertyId propertyId : propertyIdSet) {
					Set<PropertyId> innerSet = new HashSet<>();
					innerSet.add(propertyId);

					repository = repository.combineWith(learnWeights(optimizableClazz, classId, inputEntry.getValue(), gsForClass, normalization, innerSet));
				}
			} else {
				repository = repository.combineWith(learnWeights(optimizableClazz, classId, inputEntry.getValue(), gsForClass, normalization, propertyIdSet));

			}
			ret.put(classId, new WeightsRepositoryImpl(repository.getWeights()));
		}
		
		
		return ret;
	}

	private WeightsRepository learnWeights(OptimizableClass optmizableClazz, ClassId classId, AttributeMatchingFullRuns value,
			Map<AnnotatedTableId, TableAttributeMappings> gsForClass, Map<PropertyId, PropertyId> normalization, Set<PropertyId> propertyIdSetInput) {
		final Set<PropertyId> propertyIdSet = new HashSet<>(propertyIdSetInput);
		for (PropertyId propertyId : propertyIdSet) {
			if (normalization.containsKey(propertyId)) {
				propertyIdSet.add(normalization.get(propertyId));
			}
		}
		WeightsRepository weightsRepository = weightsParallelOptimizer.optimize((weightsRepositoryInput) -> {

			
			Map<PropertyId, Integer> byProperty = gsForClass.values().stream().flatMap((e) -> e.getMappings().values().stream())
					.collect(Collectors.groupingBy((e) -> e.getPropertyId(), Collectors.reducing(0, e -> 1, Integer::sum)));

			double correct = 0;
			double wrong = 0;
			// double maxCorrectOriginal =
			// gsForClass.values().stream().mapToInt(
			//
			// (e) -> e.getMappings().values().stream().filter((e2) -> {
			// if (propertyIdSet.contains(e2.getPropertyId()))
			// return true;
			// else
			// return false;
			// }).collect(Collectors.toList()).size()
			//
			// ).sum();
			double maxSize = byProperty.values().stream().mapToInt(Integer::intValue).max().getAsInt();
			
			Map<PropertyId, Double> upsampler = byProperty.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> maxSize / e.getValue()));
			double maxCorrect = gsForClass.values().stream().mapToDouble(

					(e) -> e.getMappings().values().stream().filter((e2) -> {
						if (propertyIdSet.contains(e2.getPropertyId()))
							return true;
						else
							return false;
					}).collect(Collectors.toList()).stream().mapToDouble((e3) -> upsampler.get(e3.getPropertyId())).sum()

			).sum();

			for (Entry<TableColumnId, AttributeMatchingSingleRun> entry : value.getAllRuns(gsForClass.keySet()).entrySet()) {
				if (gsForClass.containsKey(entry.getKey().getTableId())) {

					PropertyMapping propertyIdMapping = gsForClass.get(entry.getKey().getTableId()).getMappings().get(entry.getKey().getColumnIdx());
					PropertyId propertyId = null;
					if (propertyIdMapping != null) {
						propertyId = propertyIdMapping.getPropertyId();
					}

					for (PropertyId currentPropertyIdTest : propertyIdSetInput) {

						if (entry.getValue().getScores().containsKey(currentPropertyIdTest)) {

							Map<PropertyId, Map<String, ParameteredMatcherScore>> map = new HashMap<>();
							map.put(currentPropertyIdTest, entry.getValue().getScores().get(currentPropertyIdTest));
							AttributeMatchingSingleRun filteredRun = new AttributeMatchingSingleRun(map);
							Optional<PropertyMapping> maxPropertyId = equation.get(optmizableClazz, classId, filteredRun, weightsRepositoryInput);

							if (maxPropertyId.isPresent()) {
								PropertyId maxPropertyIdObject = maxPropertyId.get().getPropertyId();
								if (normalization.containsKey(maxPropertyIdObject))
									maxPropertyIdObject = normalization.get(maxPropertyIdObject);

								double upsamplingNumber = upsampler.getOrDefault(maxPropertyIdObject, 1.0);
								if (maxPropertyIdObject.equals(propertyId)) {
									correct += upsamplingNumber;
								} else {
									wrong += upsamplingNumber;
								}
							}
						}
					}

				}

			}
			
			double correctDouble = correct;
			double precision = 0.0;
			if (correct > 0 || wrong > 0)
				precision = correctDouble / (wrong + correctDouble);

			double recall = correctDouble / (maxCorrect);

			double fmeasure = 0.0;
			if (precision > 0.001 && recall > 0.001)
				fmeasure = (2 * precision * recall) / (precision + recall);

			return fmeasure;
		});

		return weightsRepository;
	}

	public static Map<AnnotatedTableId, TableAttributeMappings> getForClass(AttributeMappings gsMappings, ClassId classId) {
		Map<AnnotatedTableId, TableAttributeMappings> ret = new HashMap<>();

		for (Entry<AnnotatedTableId, TableAttributeMappings> entry : gsMappings.getMappings().entrySet()) {
			if (gsMappings.getClassId(entry.getKey()).equals(classId)) {
				ret.put(entry.getKey(), entry.getValue());
			}
		}
		return ret;
	}
}
