package expansion.set.clustering;

import java.util.Map;
import java.util.Set;
import java.util.UUID;

import com.fasterxml.uuid.Generators;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdsSet;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;
import de.mannheim.uni.expansion.tablefusion.learning.SplitRowValues;
import expansion.set.evaluation.goldstandard.GoldStandardClusters;

public class ClustererInput {

	private final UUID uuid = Generators.timeBasedGenerator().generate();

	private final SplitGroupedValuesSets groupedValuesSets;
	private final Set<TableRowId> candidates;
	private final TableRowIdsSet fullCandidates;

	private final TableRowIdsSet fullCandidatesAllClasses;
	private final TableIdsSet tableIds;
	private final TableIdsSet fullTableIds;
	private final GoldStandardClusters goldStandardClusters;
	private final ClassId classId;
	private final SplitRowValues splitRowValues;

	private final Map<ClassId, Set<ClassId>> allParents;

	private final Map<ClassId, Set<PropertyId>> nonTailProperties;

	@Deprecated
	public ClustererInput(SplitGroupedValuesSets groupedValuesSets, Set<TableRowId> candidates, TableRowIdsSet fullCandidates, TableIdsSet tableIds,
			GoldStandardClusters goldStandardClusters, ClassId classId, SplitRowValues splitRowValues, TableRowIdsSet fullCandidatesAllClasses) {
		super();
		this.groupedValuesSets = groupedValuesSets;
		this.candidates = candidates;
		this.fullCandidates = fullCandidates;
		this.tableIds = tableIds;
		this.goldStandardClusters = goldStandardClusters;
		this.classId = classId;
		this.splitRowValues = splitRowValues;
		this.fullTableIds = null;
		this.allParents = null;
		this.nonTailProperties = null;
		this.fullCandidatesAllClasses = fullCandidatesAllClasses;

	}

	@Deprecated
	public ClustererInput(SplitGroupedValuesSets groupedValuesSets, Set<TableRowId> candidates, TableRowIdsSet fullCandidates, TableIdsSet tableIds,
			GoldStandardClusters goldStandardClusters, ClassId classId, SplitRowValues splitRowValues, TableIdsSet fullTableIds) {
		super();
		this.groupedValuesSets = groupedValuesSets;
		this.candidates = candidates;
		this.fullCandidates = fullCandidates;
		this.tableIds = tableIds;
		this.goldStandardClusters = goldStandardClusters;
		this.classId = classId;
		this.splitRowValues = splitRowValues;
		this.fullTableIds = fullTableIds;
		this.allParents = null;
		this.nonTailProperties = null;
		this.fullCandidatesAllClasses = null;

	}

	@Deprecated
	public ClustererInput(SplitGroupedValuesSets groupedValuesSets, Set<TableRowId> candidates, TableRowIdsSet fullCandidates, TableIdsSet tableIds,
			GoldStandardClusters goldStandardClusters, ClassId classId, TableRowIdsSet fullCandidatesAllClasses) {
		super();
		this.groupedValuesSets = groupedValuesSets;
		this.candidates = candidates;
		this.fullCandidates = fullCandidates;
		this.tableIds = tableIds;
		this.goldStandardClusters = goldStandardClusters;
		this.classId = classId;
		this.splitRowValues = null;
		this.fullTableIds = null;
		this.allParents = null;
		this.nonTailProperties = null;
		this.fullCandidatesAllClasses = fullCandidatesAllClasses;

	}
	@Deprecated
	public ClustererInput(SplitGroupedValuesSets groupedValuesSets, Set<TableRowId> candidates, TableIdsSet tableIds, GoldStandardClusters goldStandardClusters, ClassId classId) {
		super();
		this.groupedValuesSets = groupedValuesSets;
		this.candidates = candidates;
		this.fullCandidates = null;
		this.tableIds = tableIds;
		this.goldStandardClusters = goldStandardClusters;
		this.classId = classId;
		this.splitRowValues = null;
		this.fullTableIds = null;
		this.allParents = null;
		this.nonTailProperties = null;
		this.fullCandidatesAllClasses = null;
		
		
		
	}

	public ClustererInput(SplitGroupedValuesSets groupedValuesSets, Set<TableRowId> candidates, TableRowIdsSet fullCandidates, TableIdsSet tableIds,
			GoldStandardClusters goldStandardClusters, ClassId classId, SplitRowValues splitRowValues, 
			Map<ClassId, Set<ClassId>> allParents, Map<ClassId, Set<PropertyId>> nonTailProperties, TableRowIdsSet fullCandidatesAllClasses) {
		super();
		this.groupedValuesSets = groupedValuesSets;
		this.candidates = candidates;
		this.fullCandidates = fullCandidates;
		this.tableIds = tableIds;
		this.goldStandardClusters = goldStandardClusters;
		this.classId = classId;
		this.splitRowValues = splitRowValues;
		this.fullTableIds = null;
		this.allParents = allParents;
		this.nonTailProperties = nonTailProperties;
		this.fullCandidatesAllClasses = fullCandidatesAllClasses;

	}
	
	@Deprecated
	public ClustererInput(SplitGroupedValuesSets groupedValuesSets, Set<TableRowId> candidates, TableRowIdsSet fullCandidates, TableIdsSet tableIds,
			GoldStandardClusters goldStandardClusters, ClassId classId, SplitRowValues splitRowValues, TableIdsSet fullTableIds,
			Map<ClassId, Set<ClassId>> allParents, Map<ClassId, Set<PropertyId>> nonTailProperties) {
		super();
		this.groupedValuesSets = groupedValuesSets;
		this.candidates = candidates;
		this.fullCandidates = fullCandidates;
		this.tableIds = tableIds;
		this.goldStandardClusters = goldStandardClusters;
		this.classId = classId;
		this.splitRowValues = splitRowValues;
		this.fullTableIds = fullTableIds;
		this.allParents = allParents;
		this.nonTailProperties = nonTailProperties;
		this.fullCandidatesAllClasses = null;

	}

	public UUID getUuid() {
		return uuid;
	}

	public SplitGroupedValuesSets getGroupedValuesSets() {
		return groupedValuesSets;
	}

	public Set<TableRowId> getCandidates() {
		return candidates;
	}


	public TableRowIdsSet getFullCandidates() {
		return fullCandidates;
	}

	public TableIdsSet getTableIds() {
		return tableIds;
	}
	
	

 

	public ClassId getClassId() {
		return classId;
	}

	public GoldStandardClusters getGoldStandardClusters() {
		return goldStandardClusters;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((uuid == null) ? 0 : uuid.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		ClustererInput other = (ClustererInput) obj;
		if (uuid == null) {
			if (other.uuid != null)
				return false;
		} else if (!uuid.equals(other.uuid))
			return false;
		return true;
	}

	public SplitRowValues getRowValuesSets() {
		return splitRowValues;
	}

	public TableIdsSet getFullTableIds() {
		return fullTableIds;
	}

	public Map<ClassId, Set<ClassId>> getAllParents() {
		return allParents;
	}

	public Map<ClassId, Set<PropertyId>> getNonTailProperties() {
		return nonTailProperties;
	}

	public TableRowIdsSet getFullCandidatesAllClasses() {
		return fullCandidatesAllClasses;
	}

}
