package expansion.set.clustering;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.function.Function;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import expansion.set.clustering.model.RowCluster;
import expansion.set.strategies.blockers.Mappings;
import expansion.set.strategies.rowclusterers.CreateClustersFromMappings;

public interface RowClusterSplitter {
	public ClusterList clusterRows(RowClusteringConfig config, ClustererInput input,
 ClusterList rowCluster);
	
	 
	

	public List<RowCluster> cluster(ClassId classId, List<TableRowId> list);

	public default ClusterList clusterInput(ClustererInput input) {
		return new ClusterList(this.cluster(input.getClassId(), new ArrayList<>(input.getCandidates())));
	}
	
	public  default ClusterList clusterRows( Function<AnnotatedTableId, ClassId> getClassId,    Mappings... mappingsArray) {
		Map<Integer, Set<TableRowId>> clustersMap = new HashMap<>();
		Map<Integer, ClassId> clusterClassId = new HashMap<>();
		Map<TableRowId, Integer> clustersIndex = new HashMap<>();
	 
		ConcurrentLinkedQueue<RowCluster> list = new ConcurrentLinkedQueue<>();
		Map<TableRowId,Set<TableRowId>> mappings = CreateClustersFromMappings.combineMappings(mappingsArray);
		CreateClustersFromMappings.addCluster(mappings, clustersMap, clustersIndex, clusterClassId, getClassId, 
				(rowId) -> {
					return null;
				});
		
		List<Integer> sizeList = clustersMap.entrySet().stream().map((e) -> e.getValue().size()).sorted().collect(Collectors.toList());
		System.out.println("Max Sizes: ");
		for(int i = 1; i <= 10; i++) {
			System.out.println(sizeList.get(sizeList.size() - i));
		}
		
		System.out.println("Number of blocks: " + clustersMap.size());
		
		PersistentParallel6.runAndFinish(clustersMap.entrySet(), (e) -> {
 

			list.addAll(this.cluster(clusterClassId.get(e.getKey()), new ArrayList<>(e.getValue())));
			
		}, this.getClass());
		
		return new ClusterList(list);
	}
	
 
}
