package expansion.set.clustering.model;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.BatchConsumerProducer;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.dws.expansion.data.utils.kryo.CHMKryoAdapter;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.exportimport.OptionalClusterCache;
import expansion.set.strategies.rowclusterers.ClusterSimilarities;

public class ClusterSimilaritiesImplCacheWrapper implements ClusterSimilarities {

	private final ClusterSimilarities similarities;
	private final OptionalClusterCache cache;
	private final RowCluster rowCluster;

	private final CHMKryoAdapter chmKryoAdapter;
	private final String folder;
	private final String prefix;

	public ClusterSimilaritiesImplCacheWrapper(ClusterSimilarities similarities, RowCluster rowCluster, OptionalClusterCache clusterCache,
			CHMKryoAdapter chmKryoAdapter, String folder, String prefix) {
		this.similarities = similarities;
		this.cache = clusterCache;
		this.rowCluster = rowCluster;
		this.chmKryoAdapter = chmKryoAdapter;
		this.folder = folder;
		this.prefix = prefix;
	}

	public void writeUpdates() {
		cache.writeUpdates();
	}

	@Override
	public double get(ClusterEntryKey key) {
		TableRowId rowId1 = rowCluster.getRows().get(key.getE1());
		TableRowId rowId2 = rowCluster.getRows().get(key.getE2());
		Double fromCache = cache.get(rowId1, rowId2);
		if (fromCache != null) {
			return fromCache;
		} else {
			Double value = similarities.get(key);
			cache.add(rowId1, rowId2, value);
			return value;
		}


	}

	@Override
	public Set<Integer> regionQuery(Integer point, double eps) {
		throw new UnsupportedOperationException();

	}

	@Override
	public List<Pair<Set<Integer>, Integer>> getSplitBlocksWithMin(Integer point) {
		return similarities.getSplitBlocksWithMin(point);
	}

	@Override
	public List<Tri<Set<Integer>, Integer, Double>> getSplitBlocksWithScores(Integer point) {
		return similarities.getSplitBlocksWithScores(point);

	}

	@Override
	public List<Set<Integer>> getSplitBlocks(Integer point) {
		return similarities.getSplitBlocks(point);
	}

	@Override
	public Set<Integer> getBlocks(Integer point) {
		return similarities.getBlocks(point);
	}

	public boolean supportsBlocking() {
		return similarities.supportsBlocking();
	}

	@Override
	public boolean supportsCaching() {
		return true;
	}

	@Override
	public boolean cacheExists() {
		return chmKryoAdapter.cacheExists(folder, prefix);
	}

	@Override
	public void storeClusters(List<Set<Integer>> clusters) {
		Map<Integer, Set<TableRowId>> map = new HashMap<>();
		for (Set<Integer> cluster : clusters) {
			map.put(map.size(), cluster.stream().map((e) -> rowCluster.getRows().get(e)).collect(Collectors.toSet()));
		}

		chmKryoAdapter.writeWithSet(folder, prefix, map);
	}

	@Override
	public List<Set<Integer>> readClusters() {

		Map<TableRowId, Integer> index = new HashMap<>();
		for (ListEntry<TableRowId> entry : ListIterable.iterable(rowCluster.getRows())) {
			index.put(entry.getValue(), entry.getIndex());
		}

		Map<Integer, Set<TableRowId>> map = chmKryoAdapter.readWithSet(Integer.class, TableRowId.class, folder, prefix);

		List<Set<Integer>> ret = new ArrayList<>();
		for (Set<TableRowId> cluster : map.values()) {
			Set<Integer> set = new HashSet<>();
			for (TableRowId rowId : cluster) {
				set.add(index.get(rowId));
			}
			ret.add(set);
		}

		return ret;
	}

	public void precompute(ConcurrentHashMap<Integer, Set<Integer>> precompute) {

		BatchConsumerProducer<ClusterEntryKey> consumerProducer = BatchConsumerProducer.createMultiRun(this.getClass() + ":PRECOMPUTE", (pair) -> {
			get(pair);
		}, 3);

		int min = precompute.entrySet().stream().mapToInt((e) -> e.getValue().size()).min().getAsInt();
		int batchSize = Math.max(4500000, min);
		List<ClusterEntryKey> keyList = new ArrayList<>(batchSize);
		// int i = 0;
		for (Entry<Integer, Set<Integer>> pairEntry : precompute.entrySet()) {
			if (keyList.size() + pairEntry.getValue().size() > batchSize) {
				consumerProducer.addBatch(keyList);
				// i++;
				// if (i == 40) {
				// break;
				// }
				keyList = new ArrayList<>(batchSize);
			}
			for (Integer second : pairEntry.getValue()) {
				keyList.add(new ClusterEntryKey(pairEntry.getKey(), second));
			}
			precompute.remove(pairEntry.getKey());

		}
		if (keyList.size() > 0) {
			consumerProducer.addBatch(keyList);
			keyList = null;
		}

		precompute.clear();

		consumerProducer.finish();

		this.writeUpdates();

	}
}
