package expansion.set.clustering.model;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.strategies.rowclusterers.ClusterSimilarities;

public class ClusterSimilaritiesImplNormalized implements ClusterSimilarities {

	private final RowClusterSimilarities similarities;
	private final RowCluster rowCluster;
	private final Map<TableRowId, Integer> index;
	private final boolean alwaysDisableBlocking;

	private final double threshold;

	public ClusterSimilaritiesImplNormalized(RowCluster rowCluster, RowClusterSimilarities similarities, boolean alwaysDisableBlocking) {
		this(rowCluster, similarities, alwaysDisableBlocking, 0.5);
	}

	public ClusterSimilaritiesImplNormalized(RowCluster rowCluster, RowClusterSimilarities similarities, boolean alwaysDisableBlocking,
			double threshold) {
		super();
		this.similarities = similarities;
		this.rowCluster = rowCluster;
		this.alwaysDisableBlocking = alwaysDisableBlocking;

		this.threshold = threshold;
		this.index = new HashMap<>();
		for (ListEntry<TableRowId> rowId : ListIterable.iterable(rowCluster.getRows())) {
			index.put(rowId.getValue(), rowId.getIndex());
		}
	}

	public double get(TableRowIdPair rowIdPair) {
		return get(new ClusterEntryKey(index.get(rowIdPair.getId1()), index.get(rowIdPair.getId2())));
	}

	@Override
	public double get(ClusterEntryKey key) {
		// Set<TableRowId> rowIds = new HashSet<>();
		//
		// rowIds.add(new
		// TableRowId("1346823846150_1346835704683_3109.arc8646851352477934348#60834880_0_4803626601892752074",
		// 57));
		// rowIds.add(new
		// TableRowId("1350433107082_1350482259427_736.arc784927006338575789#20499084_0_4976499811886839004",
		// 16));
		//
		// rowIds.add(new
		// TableRowId("1346876860804_1346920194056_1250.arc4757528133752403897#13435487_0_9118807868790417018",
		// 3));
		// rowIds.add(new
		// TableRowId("1346876860782_1346909948375_4041.arc6066155847369688043#71748033_0_170995977879854048",
		// 23));

		double score = similarities.get(rowCluster.getRows(), rowCluster.getClassId(), key);
		if(score < 0.0) {
			int sfd= 123;
		}
		if (threshold > -0.01 && threshold < 0.01) {
			return score;
		}

		double transformedScore;

		if (score > threshold) {
			transformedScore = (score - threshold) / (1.0 - threshold);
		} else {
			transformedScore = (score - threshold) / threshold;
		}

		// if (rowIds.contains(rowCluster.getRows().get(key.getE1())) &&
		// rowIds.contains(rowCluster.getRows().get(key.getE2()))) {
		// TableRowId rowId1 = rowCluster.getRows().get(key.getE1());
		// TableRowId rowId2 = rowCluster.getRows().get(key.getE2());
		// int sfd = 123;
		// }

		return transformedScore;

	}

	@Override
	public Set<Integer> regionQuery(Integer point, double eps) {
		TableRowId tableRowId = rowCluster.getRows().get(point);
		Set<TableRowId> candidates = similarities.regionQuery(rowCluster, tableRowId, eps);
		return candidates.stream().map((e) -> rowCluster.getIndex(e)).collect(Collectors.toSet());

	}

	@Override
	public List<Pair<Set<Integer>, Integer>> getSplitBlocksWithMin(Integer point) {
		return similarities.getSplitBlocksWithMin(rowCluster.getRows().get(point));
	}

	@Override
	public List<Tri<Set<Integer>, Integer, Double>> getSplitBlocksWithScores(Integer point) {
		
		Set<TableRowId> rowIds = new HashSet<>();
		rowIds.add(new TableRowId("1346981172186_1346994903857_1044.arc4636266937237321018#46166619_0_591080969917940788", 28));
		rowIds.add(new TableRowId("1350433107054_1350486965107_799.arc7111710114070000786#21907331_0_5050878971533094232", 11)); 
		if (rowIds.contains(rowCluster.getRows().get(point))) {
			int sdf= 123;
		}
		
		return similarities.getSplitBlocksWithScores(rowCluster.getRows().get(point));

	}

	@Override
	public List<Set<Integer>> getSplitBlocks(Integer point) {
		return similarities.getSplitBlocks(rowCluster.getRows().get(point));
	}

	@Override
	public Set<Integer> getBlocks(Integer point) {
		return similarities.getBlocks(rowCluster.getRows().get(point));
	}

	public boolean supportsBlocking() {
		return !alwaysDisableBlocking && similarities.supportsBlocking();
	}

}
