package expansion.set.clustering.model.similarities;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import expansion.set.clustering.model.Block;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;

public interface RowClusterSimilarities {

	public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key);
	

	public default List<WeightedScore> getList(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
		List<WeightedScore> list = new ArrayList<>();
		list.add(new WeightedScore(get(rows, classId, key), 1.0));
		return list;
	}

	public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps);
	
	public String getName();

	default Set<Integer> getBlocks(TableRowId rowId) {
		return new HashSet<>();
	}

	public default boolean supportsBlocking() {
		return false;
	}

	public default List<Set<Integer>> getSplitBlocks(TableRowId tableRowId) {
		List<Set<Integer>> list = new ArrayList<>();
		Set<Integer> blocks = getBlocks(tableRowId);
		if (blocks == null)
			blocks = new HashSet<>();
		list.add(blocks);
		return list;
	}

	public default List<Pair<Set<Integer>, Integer>> getSplitBlocksWithMin(TableRowId tableRowId) {
		List<Pair<Set<Integer>, Integer>> list = new ArrayList<>();
		Set<Integer> blocks = getBlocks(tableRowId);
		if (blocks == null)
			blocks = new HashSet<>();
		list.add(new Pair<>(blocks, getBlockMin(tableRowId)));
		return list;
	}

	public default List<Tri<Set<Integer>, Integer, Double>> getSplitBlocksWithScores(TableRowId tableRowId) {
		List<Tri<Set<Integer>, Integer, Double>> list = new ArrayList<>();
		Set<Integer> blocks = getBlocks(tableRowId);
		if (blocks == null)
			blocks = new HashSet<>();
		list.add(new Tri<>(blocks, getBlockMin(tableRowId), 1.0));
		return list;
	}

	public default List<Pair<Set<Block>, Integer>> getSplitBlocksWithScoresAndMinWithMin(TableRowId tableRowId) {
		List<Pair<Set<Block>, Integer>> list = new ArrayList<>();
		list.add(new Pair<>(getBlocks(tableRowId).stream().map((e) -> new Block(e, 1.0)).collect(Collectors.toSet()), 1));
		return list;
	}
	default int getBlockMin(TableRowId tableRowId) {
		return 1;
	}

	default public int getBlockerSizes() {
		if (supportsBlocking())
			return 1;
		else
			return 0;
	}
}
