package expansion.set.creation;

import java.util.Collections;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowClusterUUID;

public class Entity {
	private final RowClusterUUID rowClusterUUID;
	private final ClassId classId;
	private final Map<String, Integer> labels;
	private final Map<QualifiedPropertyId, FusedValue> values;
	private final RowCluster rowCluster;

	public Entity(RowClusterUUID rowClusterUuid, ClassId classId, Map<String, Integer> labels, Map<QualifiedPropertyId, FusedValue> values, RowCluster rowCluster) {
		super();
		this.rowClusterUUID = rowClusterUuid;
		this.classId = classId;
		this.labels = Collections.unmodifiableMap(labels);
		this.values = Collections.unmodifiableMap(values);
		this.rowCluster = rowCluster;
		
	}

	public Map<String, Integer> getLabels() {
		return labels;
	}

	public Map<QualifiedPropertyId, FusedValue> getValues() {
		return values;
	}

	public ClassId getClassId() {
		return classId;
	}

	public RowClusterUUID getRowClusterUUID() {
		return rowClusterUUID;
	}

	public RowCluster getRowCluster() {
		return rowCluster;
	}
	
	public String getMainLabel() {
		Optional<Entry<String, Integer>> mainLabelOptional = getLabels().entrySet().stream().sorted((e1, e2) -> { 
			int compare = Integer.compare(e2.getValue(), e1.getValue());
			if(compare != 0)
				return compare;
			else
				return e1.getKey().compareTo(e2.getKey());
		}).findFirst();
		if (mainLabelOptional.isPresent()) {
			String mainLabel = mainLabelOptional.get().getKey();
			return mainLabel;
		} else {
			return null;
		}
	}
	
	
}
