package expansion.set.distant_supervision.baseline.average;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.uni_mannheim.informatik.dws.expansion.utils.learning.weka.WekaLearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.learner.WeightsLearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeightKey;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepositoryImpl;
import expansion.set.distant_supervision.DS_NewDetection_Strategy;
import expansion.set.distant_supervision.DistantSupervisionInput;
import expansion.set.entitymatching.modelLearning.EM_LearnedModelRepository;
import expansion.set.entitymatching.modelLearning.EntityMatchingModel;
import expansion.set.entitymatching.modelLearning.EntityMatchingModelLearnerRepository;
import expansion.set.entitymatching.modelLearning.simpleWeights.SimpleEntityMatcherEquation;
import expansion.set.entitymatching.modelLearning.weightsGeneric.EM_WeightsGenericModel;
import expansion.set.entitymatching.modelLearning.weka.EntityMatcherWekaLearner;
import expansion.set.entitymatching.run.EntityMatchingRun;
import expansion.set.entitymatching.run.EntityMatchingRunRepository;

public class ND_Average_Creator {

	@Inject
	protected EntityMatchingModelLearnerRepository repository;

	@Inject

	public ND_Average_Creator() {

	}

	public ND_Average_Strategy getStrategy(double threshold) {
		return new ND_Average_Strategy(threshold);
	}
	public class ND_Average_Strategy implements DS_NewDetection_Strategy {
		private final double threshold;

		private ND_Average_Strategy(double threshold) {
			this.threshold = threshold;
		}

		@Override
		public Map<ClassId, EM_LearnedModelRepository> get(DistantSupervisionInput input) {

			Map<ClassId, EM_LearnedModelRepository> ret = new HashMap<>();

			for (ClassId classId : input.getClassesWithParents().keySet()) {
				Map<OptimizableClass, EntityMatchingModel> modelMap = new HashMap<>();
				
				EntityMatchingRun run = EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL;

				// for (EntityMatchingRun run : input.getEntityMatchingRuns()) {
				// if
				// (run.getModel().equals(EntityMatchingWekaRFWithSWLearner.class))
				// {

						OptimizableClass optimiableClass = run.getOptimizableClass();

						Map<OptimizableWeightKey, Double> weightsMap = new HashMap<>();
						weightsMap.put(optimiableClass.getKey(WeightsLearnedModel.THRESHOLD_FEATURE), threshold);

				weightsMap.put(optimiableClass.getKey("THRESHOLD_NEW"), threshold);

				weightsMap.put(optimiableClass.getKey("THRESHOLD_MATCH"), threshold);

						weightsMap.put(optimiableClass.getKey("WEKA_WEIGHT"), 0.0);
						weightsMap.put(optimiableClass.getKey("NON_WEKA_NORMALIZE"), 0.5);

				weightsMap.put(optimiableClass.getKey("ATTRIBUTE_WEIGHT"), 0.75);
				weightsMap.put(optimiableClass.getKey("BOW_WEIGHT"), 0.75);
				weightsMap.put(optimiableClass.getKey("IMPLICIT_ATTRIBUTE_WEIGHT"), 0.50);
				weightsMap.put(optimiableClass.getKey("INDEGREE_WEIGHT"), 0.25);
						weightsMap.put(optimiableClass.getKey("LABEL_WEIGHT"), 1.0);
				weightsMap.put(optimiableClass.getKey("TYPE_WEIGHT"), 0.50);

						EntityMatcherWekaLearner learner = (EntityMatcherWekaLearner) repository.get(run.getModel());

						WekaLearnedModel model = new WekaLearnedModel(true);

						EntityMatchingModel customModel = learner.getCustomModel(run, model, new WeightsRepositoryImpl(weightsMap),
						new EM_WeightsGenericModel(run, new WeightsRepositoryImpl(weightsMap), new SimpleEntityMatcherEquation()),
						input.getNonTailProperties().get(classId));

						modelMap.put(optimiableClass, customModel);

				// }
				// }
				
				ret.put(classId, new EM_LearnedModelRepository(modelMap));
				

			}

			return ret;
		}

		public List<EntityMatchingRun> getStrategies() {
			List<EntityMatchingRun> ret = new ArrayList<>();
			ret.add(EntityMatchingRunRepository.RF_SW_CS_LABEL_EM_ALL);
			return ret;
		}

	}
}

