package expansion.set.distant_supervision.baseline.average;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.learner.WeightsLearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.LearnedModelRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeightKey;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepositoryImpl;
import expansion.set.distant_supervision.DS_Clustering_Strategy;
import expansion.set.distant_supervision.DistantSupervisionInput;
import expansion.set.strategies.SetExpansionStrategiesRepository;
import expansion.set.strategies.SetExpansionStrategy;
import expansion.set.strategies.clustersplitters.minScoreSplitter.CombinedModel;
import expansion.set.strategies.rowclusterers.emptyLearnable.EmptyLearnableModel;

public class RC_Average_Creator {

	@Inject

	public RC_Average_Creator() {

	}

	public ND_Average_Strategy getStrategy(double threshold) {
		return new ND_Average_Strategy(threshold);
	}

	public class ND_Average_Strategy implements DS_Clustering_Strategy {
		private final double threshold;

		private ND_Average_Strategy(double threshold) {
			this.threshold = threshold;
		}

		@Override
		public Map<ClassId, LearnedModelRepository> get(DistantSupervisionInput input) {
			Map<ClassId, LearnedModelRepository> ret = new HashMap<>();

			for (ClassId classId : input.getClassesWithParents().keySet()) {
				Map<OptimizableClass, LearnedModel> modelMap = new HashMap<>();

				// for (SetExpansionStrategy strategy :
				// input.getSetExpansionStrategiesConfig().getStrategies()) {

				OptimizableClass optimiableClass = SetExpansionStrategiesRepository.COMBINED_RF_BASE.getOptimizableClassCluster();

						Map<OptimizableWeightKey, Double> weightsMap = new HashMap<>();
						weightsMap.put(optimiableClass.getKey(WeightsLearnedModel.THRESHOLD_FEATURE), threshold);

						weightsMap.put(optimiableClass.getKey("AA_LABEL_PHI"), 0.25);
				weightsMap.put(optimiableClass.getKey("ATTRIBUTE"), 0.5);
				weightsMap.put(optimiableClass.getKey("BOW"), 0.5);
				weightsMap.put(optimiableClass.getKey("IMPLICIT_ATTRIBUTES"), 0.25);
				weightsMap.put(optimiableClass.getKey("LABEL"), 1.0);
						weightsMap.put(optimiableClass.getKey("SAME_TABLE_ZERO"), 0.25);

						WeightsLearnedModel weightsModel = new WeightsLearnedModel(optimiableClass, new WeightsRepositoryImpl(weightsMap));


						Map<OptimizableWeightKey, Double> combinedWeightsMap = new HashMap<>();
						combinedWeightsMap.put(optimiableClass.getKey("aWeight"), 0.0);

						CombinedModel model = new CombinedModel(new EmptyLearnableModel(), weightsModel, new WeightsRepositoryImpl(combinedWeightsMap));

						modelMap.put(optimiableClass, model);

				// }

				ret.put(classId, new LearnedModelRepository(modelMap));
			}

			return ret;
		}

		@Override
		public List<SetExpansionStrategy> getStrategies() {
			List<SetExpansionStrategy> ret = new ArrayList<>();
			ret.add(SetExpansionStrategiesRepository.COMBINED_RF_BASE);
			return ret;
		}

	}
}

