package expansion.set.distant_supervision.bootsrap;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import javax.inject.Inject;

import com.google.common.collect.Lists;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.utils.DoubleIteraterable;
import de.mannheim.uni.expansion.utils.DoubleIteraterable.DoubleEntry;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.LearnedModelRepository;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities.PairLearningData;
import expansion.set.distant_supervision.DS_Container;
import expansion.set.distant_supervision.DS_Multi_Strategy;
import expansion.set.distant_supervision.DistantSupervisionInput;
import expansion.set.distant_supervision.baseline.average.ND_Average_Creator;
import expansion.set.distant_supervision.runners.DS_ClusteringRunnerCreator.ClusteringRunner;
import expansion.set.entitymatching.modelLearning.EM_LearnedModelRepository;
import expansion.set.evaluation.goldstandard.GoldStandardClustersSet;
import expansion.set.strategies.SetExpansionStrategiesRepository;

public class GenericStrategyCreator {

	@Inject
	protected ND_Average_Creator ndAverageCreator;

	@Inject
	public GenericStrategyCreator() {

	}

	public GenericStrategy get() {
		return new GenericStrategy();
	}

	public class GenericStrategy implements DS_Multi_Strategy {

		@Override
		public Map<String, DS_Container> get(DistantSupervisionInput input, ClusteringRunner runner) {
 
			Map<ClassId, LearnedModelRepository> rowClusteringRF = new HashMap<>();
			Map<ClassId, LearnedModelRepository> rowClusteringLOG = new HashMap<>();
			Map<ClassId, LearnedModelRepository> rowClusteringBASE = new HashMap<>();
			Map<ClassId, EM_LearnedModelRepository> newDetection = new HashMap<>();

			Map<ClassId, EM_LearnedModelRepository> ndStrategies = ndAverageCreator.getStrategy(0.5).get(input);
					
			for (ClassId classId : input.getClassesWithParents().keySet()) {
				Set<ClassId> otherClasses = new HashSet<>(input.getClassesWithParents().keySet());
				otherClasses.remove(classId);
				// ClassId otherClass = otherClasses.iterator().next();
				// otherClasses.clear();
				// otherClasses.add(otherClass);
				// System.out.println("OtherClass: " + otherClass.getClassId());

				Set<InstanceId> newEntities = new HashSet<>();

				int positiveMaxCount = 0;
				int negativeMaxCount = 0;

				Map<ClassId, List<TableRowIdPair>> positivePairMap = new HashMap<>();
				Map<ClassId, List<TableRowIdPair>> negativePairMap = new HashMap<>();
				Set<TableRowId> allRowIds = new HashSet<>();
				for (ClassId oClassId : otherClasses) {

					Set<TableRowIdPair> positiveRowIdPairs = new HashSet<>();

					Set<TableRowIdPair> negativeRowIdPairs = new HashSet<>();

					GoldStandardClustersSet data = input.getGoldStandardClustersSet().getPerClass().get(oClassId);

					allRowIds.addAll(data.getFull().getClusters().getIndex().keySet());

					for (Set<TableRowId> set : data.getFull().getClusters().getTestingSetMappingClusters().values()) {
						for (DoubleEntry<TableRowId> doubleEntry : DoubleIteraterable.iterable(set)) {
							positiveRowIdPairs.add(new TableRowIdPair(doubleEntry.getFirst(), doubleEntry.getSecond()));
						}
					}

					// for (DoubleEntry<TableRowId> doubleEntry :
					// DoubleIteraterable.iterable(data.getFull().getClusters().getIndex().keySet()))
					// {
					// TableRowIdPair pair = new
					// TableRowIdPair(doubleEntry.getFirst(),
					// doubleEntry.getSecond());
					// if (!positiveRowIdPairs.contains(pair)) {
					// negativeRowIdPairs.add(pair);
					// }
					// }

					newEntities.addAll(data.getFull().getNewInstances());

					if (positiveRowIdPairs.size() > positiveMaxCount) {
						positiveMaxCount = positiveRowIdPairs.size();
					}
					// if (negativeRowIdPairs.size() > negativeMaxCount) {
					// negativeMaxCount = negativeRowIdPairs.size();
					// }

					List<TableRowIdPair> positiveRowIdPairsList = new ArrayList<>(positiveRowIdPairs);
					Collections.sort(positiveRowIdPairsList, TableRowIdPair::compare);
					// List<TableRowIdPair> negativeRowIdPairsList = new
					// ArrayList<>(negativeRowIdPairs);
					// Collections.sort(negativeRowIdPairsList,
					// TableRowIdPair::compare);

					positivePairMap.put(oClassId, positiveRowIdPairsList);
					// negativePairMap.put(oClassId, negativeRowIdPairsList);
				}

				Map<TableRowIdPair, PairLearningData> rowIdPairs = new HashMap<>();


				for (ClassId oClassId : otherClasses) {
					Random rnd = new Random(classId.getFinalPart().hashCode());

					// double positiveWeight = positiveMaxCount /
					// positivePairMap.get(oClassId).size();
					// double negativeWeight = negativeMaxCount /
					// negativePairMap.get(oClassId).size();
					List<TableRowIdPair> postList = positivePairMap.get(oClassId);
					// List<TableRowIdPair> negList =
					// negativePairMap.get(oClassId);
					int positiveDif = positiveMaxCount - postList.size();
					// int negativeDif = negativeMaxCount - negList.size();

					for (TableRowIdPair pair : postList) {
						rowIdPairs.put(pair, new PairLearningData(true, 1.0));
					}

					// for (TableRowIdPair pair : negList) {
					// rowIdPairs.put(pair, new PairLearningData(false, 1.0));
					// }

					for (int i = 0; i < positiveDif; i++) {

						rowIdPairs.put(postList.get(rnd.nextInt(postList.size())), new PairLearningData(true, 1.0));
					}

					// for (int i = 0; i < negativeDif; i++) {
					// rowIdPairs.put(negList.get(rnd.nextInt(negList.size())),
					// new PairLearningData(false, 1.0));
					//
					// }

					for (DoubleEntry<TableRowId> doubleEntry : DoubleIteraterable.iterable(allRowIds)) {
						TableRowIdPair pair = new TableRowIdPair(doubleEntry.getFirst(), doubleEntry.getSecond());
						if (!rowIdPairs.containsKey(pair)) {
							// negativeRowIdPairs.add(pair);
							rowIdPairs.put(pair, new PairLearningData(false, 1.0));

						}
					}
				}

				LearnedModelRepository rfLearnedModel = runner.learn(classId, SetExpansionStrategiesRepository.COMBINED_BASE, rowIdPairs,
						newEntities);

				System.out.println(classId);
				System.out.println(rfLearnedModel.getMap().toString());

				rowClusteringRF.put(classId, rfLearnedModel);

				// LearnedModelRepository logLearnedModel =
				// runner.learn(classId,
				// SetExpansionStrategiesRepository.COMBINED_Logistic_NEW,
				// rowIdPairs, newEntities);
				//
				// rowClusteringLOG.put(classId, logLearnedModel);
				//
				// LearnedModelRepository baseLearnedModel =
				// runner.learn(classId,
				// SetExpansionStrategiesRepository.COMBINED_BASE, rowIdPairs,
				// newEntities);
				//
				// rowClusteringBASE.put(classId, baseLearnedModel);

				newDetection.put(classId, ndStrategies.get(classId));

				// rowClustering2.put(fClassId, sameClassLearned);

				// LearnedModelRepository sameClassLearned =
				// runner.learn(otherClass,
				// SetExpansionStrategiesRepository.COMBINED_Logistic_NEW,
				// rowIdPairs,
				// newEntities);

				// System.out.println("DifferentClass: ");
				// System.out.println(differentClassLearned.getMap().toString());
				// System.out.println("SameClassLearned: ");
				// System.out.println(sameClassLearned.getMap().toString());

				// for (ClassId fClassId :
				// input.getClassesWithParents().keySet()) {


				// }


			}
			
			Map<String, DS_Container> ret = new HashMap<>();
			DS_Container dsContainerRF = new DS_Container(rowClusteringRF, Lists.newArrayList(SetExpansionStrategiesRepository.COMBINED_BASE),
					newDetection);
			ret.put("NCSTD+GRF", dsContainerRF);
			// DS_Container dsContainerLOG = new DS_Container(rowClusteringLOG,
			// Lists.newArrayList(SetExpansionStrategiesRepository.COMBINED_Logistic_NEW),
			// newDetection);
			// ret.put("NCSTD+GLOG", dsContainerLOG);
			// DS_Container dsContainerBASE = new
			// DS_Container(rowClusteringBASE,
			// Lists.newArrayList(SetExpansionStrategiesRepository.COMBINED_BASE),
			// newDetection);
			// ret.put("NCSTD+BASE", dsContainerBASE);

			// DS_Container dsContainer2 = new DS_Container(rowClustering2,
			// Lists.newArrayList(SetExpansionStrategiesRepository.COMBINED_Logistic_NEW),
			// newDetection);
			// ret.put("NCSTD+GRF_2", dsContainer2);
			return ret;
		}

	}
}
