package expansion.set.distant_supervision.bootsrap;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.TreeMap;

import cern.colt.Arrays;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.containers.Pair;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities.PairLearningData;

public class HyperParamter_RC {

	
	
	public <X> String[] findOptimumParamters(long seed, Map<X, PairLearningData> data, GetPerformance<X> getPerformance, Comparator<X> comparator) {
		int count = 3;
		TreeMap<Integer, Map<X, PairLearningData>> splitMap = splitMap(seed, data, count, comparator);
		
		return findOptimumParamtersInner(seed, splitMap, getPerformance, comparator).getFirst();

	}

	public <X> Pair<String[], Double> findOptimumParamtersInner(long seed, TreeMap<Integer, Map<X, PairLearningData>> splitMap,
			GetPerformance<X> getPerformance,
			Comparator<X> comparator) {
		int count = 3;

		Map<Integer, String[]> runs = getRuns();
		
		Map<Integer, Map<Integer, Double>> result = new HashMap<>();
		
		double total = count * runs.size();
		int current = 0;
		for (Entry<Integer, Map<X, PairLearningData>> split : splitMap.entrySet()) {
			Map<X, PairLearningData> trainingData = new HashMap<>();
			for (Entry<Integer, Map<X, PairLearningData>> eSplit : splitMap.entrySet()) {
				 if(!eSplit.getKey().equals(split.getKey())) {
					 trainingData.putAll(eSplit.getValue());
				 }
			 }

			 for(Entry<Integer, String[]> run : runs.entrySet()) {
				double performance = getPerformance.performance(split.getKey(), trainingData, split.getValue(), run.getValue().clone());
				CHMHelper.getFromMap(result, run.getKey(), () -> new HashMap<>()).put(split.getKey(), performance);
				current++;
				if (current % 10 == 0) {
					System.out.println(current / total);
				}
			} 
		}
		
		List<Pair<Integer, Double>> aggregatedResult = new ArrayList<>();
		
		
		for(Entry<Integer, Map<Integer, Double>> entry : result.entrySet()) {
			double sum = 0.0;
			for(Double d : entry.getValue().values()) {
				sum += d;
			}
			aggregatedResult.add(new Pair<>(entry.getKey(), sum / count));
		}
		
		Collections.sort(aggregatedResult, (e1, e2) -> {
			return Double.compare(e2.getSecond(), e1.getSecond());
		});
		
		
		for(Pair<Integer, Double> res : aggregatedResult) {
			System.out.println(Arrays.toString(runs.get(res.getFirst())));
			System.out.println(res.getSecond());
			System.out.println();
		}
		
		return new Pair<>(runs.get(aggregatedResult.get(0).getFirst()), aggregatedResult.get(0).getSecond());
		
	}

	private Map<Integer, String[]> getRuns() {
		Map<Integer, String[]> ret = new HashMap<>();

		List<Integer> oobRates = new ArrayList<>();
		oobRates.add(33);
		oobRates.add(66);
		oobRates.add(85);
		oobRates.add(100);

		List<Integer> minVars = new ArrayList<>();
		minVars.add(2);
		minVars.add(3);
		minVars.add(4);

		List<Integer> ks = new ArrayList<>();
		ks.add(0);
		ks.add(2);
		ks.add(10);
		// ks.add(20);

		for (Integer minVar : minVars) {
			for (Integer k : ks) {
				for (Integer oobRate : oobRates) {
					String[] parameters = new String[11];

					parameters[0] = "-U";
					parameters[1] = "-K";
					parameters[2] = String.valueOf(k);
					parameters[3] = "-V";
					parameters[4] = "1e-" + minVar;
					parameters[5] = "-P";
					parameters[6] = String.valueOf(oobRate);
					parameters[7] = "-I";
					parameters[8] = "150";
					parameters[9] = "-num-slots";
					parameters[10] = "80";

					ret.put(ret.size(), parameters);
				}
			}
		}

		return ret;
	}

	private <X> TreeMap<Integer, Map<X, PairLearningData>> splitMap(long seed, Map<X, PairLearningData> data, int folds, Comparator<X> comparator) {
		Random r = new Random(seed);
		List<Map<X, PairLearningData>> ret = new ArrayList<>();

		for (int i = 0; i < folds; i++) {
			ret.add(new HashMap<>());
		}
		List<X> positive = new ArrayList<>();
		List<X> negative = new ArrayList<>();

		for (Entry<X, PairLearningData> entry : data.entrySet()) {
			if (entry.getValue().isMatch()) {
				positive.add(entry.getKey());
			} else {
				negative.add(entry.getKey());
			}
		}

		Collections.sort(positive, comparator);
		Collections.sort(negative, comparator);

		Collections.shuffle(positive, r);
		Collections.shuffle(negative, r);

		int i = 0;
		for (X pair : positive) {

			ret.get(i).put(pair, data.get(pair));

			i++;
			if (i == folds) {
				i = 0;
			}
		}

		i = 0;
		for (X pair : negative) {

			ret.get(i).put(pair, data.get(pair));

			i++;
			if (i == folds) {
				i = 0;
			}
		}

		TreeMap<Integer, Map<X, PairLearningData>> retMap = new TreeMap<>();

		for (int c = 0; c < folds; c++) {
			retMap.put(c, ret.get(c));
		}

		return retMap;
	}
	
	@FunctionalInterface
	public static interface GetPerformance<X> {
		
		double performance(int testingFold, Map<X, PairLearningData> trainingData, Map<X, PairLearningData> testingData, String[] parameters);
	}
}
