package expansion.set.distant_supervision.bootsrap;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import javax.inject.Inject;
import javax.inject.Provider;

import com.beust.jcommander.internal.Lists;
import com.google.common.collect.Sets;

import cern.colt.Arrays;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Quad;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumDouble;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentAccumulator;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.LearnedModelRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities.PairLearningData;
import expansion.set.distant_supervision.DS_Clustering_Strategy_With_Runner;
import expansion.set.distant_supervision.DS_Container;
import expansion.set.distant_supervision.DistantSupervisionInput;
import expansion.set.distant_supervision.runners.DS_ClusteringRunnerCreator.ClusteringRunner;
import expansion.set.entitymatching.EntityMatcherFullResults;
import expansion.set.entitymatching.matchers.EntityMatcherStrategies;
import expansion.set.entitymatching.modelLearning.EM_LearnedModelRepository;
import expansion.set.entitymatching.modelLearning.EntityMatchingModelLearnerRepository;
import expansion.set.entitymatching.modelLearning.simpleWeights.SimpleEntityMatchingWeightLearner;
import expansion.set.entitymatching.modelLearning.weightsGeneric.EntityMatchingGenericWeightsOptimizier;
import expansion.set.entitymatching.modelLearning.weka.EntityMatcherWekaLearner;
import expansion.set.entitymatching.run.EntityMatchingRun;
import expansion.set.entitymatching.run.EntityMatchingRunRepository;
import expansion.set.evaluation.matching.NewDetectionEvaluator;
import expansion.set.evaluation.matching.NewDetectionPerformance;
import expansion.set.strategies.SetExpansionStrategy;

public class WS_ND_Strategy_Creator {

	// @Inject
	// protected RowClustererRepository rowClustererRepository;

	// @Inject
	// protected ND_Average_Creator ndAverageCreator;

	// @Inject
	// protected RC_Average_Creator rcAverageCreator;

	// @Inject
	// protected EntityMatchingRunner entityMatchingRuner;

	// @Inject
	// protected SetExpansionRunner setExpansionRunner;

	// @Inject
	// protected GreedyClusteringSplitter greedyClusteringSplitter;

	// @Inject
	// protected AnnotatedTablesRepository tableRepository;

	// @Inject
	// protected RowValuesGetter rowValuesGetter;

	// @Inject
	// protected EntitySetGroupedValuesSplitter entitySetGroupedValuesSplitter;

	// @Inject
	// protected EntityCreator entityCreator;

	// @Inject
	// protected ClusteringModelLearnerRunner clusteringLearner;

	@Inject
	protected NewDetectionEvaluator newDetectionEvaluator;

	@Inject
	protected SimpleEntityMatchingWeightLearner simpleLearner;

	@Inject
	protected EntityMatchingGenericWeightsOptimizier generic;

	@Inject
	protected Provider<EntityMatchingModelLearnerRepository> repository;

	@Inject
	public WS_ND_Strategy_Creator() {

	}

	public WS_ND_Strategy get(String name, EntityMatchingRun strategy, Map<ClassId, EM_LearnedModelRepository> models, SetExpansionStrategy clusteringStrategy,
			Map<ClassId, LearnedModelRepository> clusteringModels) {
		Map<String, Pair<EntityMatchingRun, Map<ClassId, EM_LearnedModelRepository>>> map = new HashMap<>();
		map.put(name, new Pair<>(strategy, models));

		return new WS_ND_Strategy(map, clusteringStrategy, clusteringModels);
	}

	public WS_ND_Strategy get(Map<String, Pair<EntityMatchingRun, Map<ClassId, EM_LearnedModelRepository>>> models, SetExpansionStrategy clusteringStrategy,
			Map<ClassId, LearnedModelRepository> clusteringModels) {
		return new WS_ND_Strategy(models, clusteringStrategy, clusteringModels);
	}

	public class WS_ND_Strategy implements DS_Clustering_Strategy_With_Runner {
		private final Map<String, Pair<EntityMatchingRun, Map<ClassId, EM_LearnedModelRepository>>> models;
		private final SetExpansionStrategy strategy;
		private final Map<ClassId, LearnedModelRepository> clusteringModels;

		public WS_ND_Strategy(Map<String, Pair<EntityMatchingRun, Map<ClassId, EM_LearnedModelRepository>>> models, SetExpansionStrategy clusteringStrategy,
				Map<ClassId, LearnedModelRepository> clusteringModels) {
			super();
			this.models = models;
			this.strategy = clusteringStrategy;
			this.clusteringModels = clusteringModels;
		}

		@Override
		public Map<String, DS_Container> get(DistantSupervisionInput input, ClusteringRunner runner) {
			EntityMatchingRun targetRun = EntityMatchingRunRepository.RF_ET_CS_LABEL_EM_ALL;

			HyperParamter_RC hyper = new HyperParamter_RC();

			// try {

			// targetRun = EntityMatchingRunRepository.RF_ET_CS_LABEL_EM_ALL;
			Map<String, DS_Container> ret = new HashMap<>();
			// Map<ClassId, EM_LearnedModelRepository> retND_WITH_WEIGHTS =
			// new HashMap<>();

			for (Entry<String, Pair<EntityMatchingRun, Map<ClassId, EM_LearnedModelRepository>>> entry : this.models.entrySet()) {

				Map<ClassId, EM_LearnedModelRepository> newDetection = new HashMap<>();

				Map<String, Map<ClassId, EM_LearnedModelRepository>> newDetectionMap = new HashMap<>();

				for (Entry<ClassId, EM_LearnedModelRepository> innerEntry : entry.getValue().getSecond().entrySet()) {

					ClassId classId = innerEntry.getKey();

					// ClusterSimilaritiesImplNormalized cs =
					// runner.getNormalizedCS(classId, strategy,
					// clusteringModels.get(classId));

					// RowClusterSimilarities sim = runner.getSim(classId,
					// strategy, clusteringModels.get(classId));

					Set<TableRowId> fullRowIds = input.getWsRowIdsSet(classId);

					Quad<ClusterList, Map<Pair<RowClusterUUID, InstanceId>, PairLearningData>, Set<InstanceId>, Map<RowClusterUUID, Optional<InstanceId>>> gsData = getGsData(
							classId, fullRowIds, runner, entry.getValue().getFirst(), innerEntry.getValue());

					if (true) {
					EM_LearnedModelRepository rfLearnedModel;

						if (false) {
							int folds = 3;

							TreeMap<Integer, Pair<Set<AnnotatedTableId>, Set<TableRowId>>> splitTables = WS_RC_Strategy_Creator.splitTables(folds, classId,
									fullRowIds.stream().map((e) -> e.getTableId()).collect(Collectors.toSet()), fullRowIds);

							final String[] parametersRet;

							Map<Integer, Quad<ClusterList, Map<Pair<RowClusterUUID, InstanceId>, PairLearningData>, Set<InstanceId>, Map<RowClusterUUID, Optional<InstanceId>>>> trainingMap = new HashMap<>();

							Map<Integer, Quad<ClusterList, Map<Pair<RowClusterUUID, InstanceId>, PairLearningData>, Set<InstanceId>, Map<RowClusterUUID, Optional<InstanceId>>>> testingMap = new HashMap<>();

							TreeMap<Integer, Map<Pair<RowClusterUUID, InstanceId>, PairLearningData>> splitMap = new TreeMap<>();

							for (int fold = 0; fold < folds; fold++) {

								testingMap.put(fold,
										getGsData(classId, splitTables.get(fold).getSecond(), runner, entry.getValue().getFirst(), innerEntry.getValue()));

								Set<TableRowId> trainingRowId = new HashSet<>();

								for (int innerFold = 0; innerFold < folds; innerFold++) {
									if (innerFold != fold) {
										trainingRowId.addAll(splitTables.get(innerFold).getSecond());
									}
								}
								trainingMap.put(fold, getGsData(classId, trainingRowId, runner, entry.getValue().getFirst(), innerEntry.getValue()));

								splitMap.put(fold, new HashMap<>());
							}

							Pair<String[], Double> paramWithScores = hyper.findOptimumParamtersInner((long) classId.getFinalPart().hashCode(), splitMap,
									(testingFold, trainingData, testingData, parameters) -> {

										EntityMatcherWekaLearner wekaLearner = getWekaLearner(parameters, targetRun.getOptimizableClass());

										Quad<ClusterList, Map<Pair<RowClusterUUID, InstanceId>, PairLearningData>, Set<InstanceId>, Map<RowClusterUUID, Optional<InstanceId>>> trainingGSData = trainingMap
												.get(testingFold);
										EM_LearnedModelRepository rfLearnedModelMM = runner.learnEM(classId, targetRun, wekaLearner,
												trainingGSData.getA().getClusterList(), trainingGSData.getB(), trainingGSData.getC(), trainingGSData.getD());

										Quad<ClusterList, Map<Pair<RowClusterUUID, InstanceId>, PairLearningData>, Set<InstanceId>, Map<RowClusterUUID, Optional<InstanceId>>> testingGSData = testingMap
												.get(testingFold);

										Quad<ClusterList, Map<Pair<RowClusterUUID, InstanceId>, PairLearningData>, Set<InstanceId>, Map<RowClusterUUID, Optional<InstanceId>>> outputGSData = getGsData(
												testingGSData.getA(), classId, runner, targetRun, rfLearnedModelMM);

										Pair<Map<RowClusterUUID, InstanceId>, Set<RowClusterUUID>> output = getClusters(outputGSData.getD(),
												outputGSData.getC());

										NewDetectionPerformance newDetectionPerformance = newDetectionEvaluator.evaluateAccuracy(testingGSData.getC(),
												testingGSData.getD(), output.getFirst(), output.getSecond());

										return (newDetectionPerformance.getNewRecall() + newDetectionPerformance.getExistingRecall()) / 2.0;
									}, (x1, x2) -> {
										int compare = RowClusterUUID.compare(x1.getFirst(), x2.getFirst());

										if (compare == 0) {
											compare = InstanceId.compare(x1.getSecond(), x2.getSecond());
										}
										return compare;
									});

							parametersRet = paramWithScores.getFirst();

							EntityMatcherWekaLearner wekaLearner = getWekaLearner(parametersRet, targetRun.getOptimizableClass());
							rfLearnedModel = runner.learnEM(classId, targetRun, wekaLearner, gsData.getA().getClusterList(), gsData.getB(), gsData.getC(),
									gsData.getD());

						} else {

							String[] defaultParameters = new String[12];

							defaultParameters[0] = "-U";
							defaultParameters[1] = "-K";
							defaultParameters[2] = String.valueOf(2);
							defaultParameters[3] = "-V";
							defaultParameters[4] = "1e-" + 3;
							defaultParameters[5] = "-P";
							defaultParameters[6] = String.valueOf(66);
							defaultParameters[7] = "-I";
							defaultParameters[8] = "2000";
							defaultParameters[9] = "-num-slots";
							defaultParameters[10] = "80";
							defaultParameters[11] = "-attribute-importance";

							EntityMatcherWekaLearner wekaLearner = getWekaLearner(defaultParameters, targetRun.getOptimizableClass());

							rfLearnedModel = runner.learnEM(classId, targetRun, wekaLearner, gsData.getA().getClusterList(), gsData.getB(), gsData.getC(),
									gsData.getD());
						}

						// rfLearnedModel =
						// rfLearnedModel.combineWith(runner.learnEM(classId,
						// EntityMatchingRunRepository.RF_NEW_CS_LABEL_EM_ALL,
						// clusters.getClusterList(), pairsFull,
						// gsData.getFirst(), gsData.getSecond()));

						System.out.println(classId.getFinalPart());
						System.out.println(rfLearnedModel.getMap());

						newDetection.put(classId, rfLearnedModel);
					} else {

						String[] defaultParameters = new String[11];

						defaultParameters[0] = "-U";
						defaultParameters[1] = "-K";
						defaultParameters[2] = String.valueOf(2);
						defaultParameters[3] = "-V";
						defaultParameters[4] = "1e-" + 3;
						defaultParameters[5] = "-P";
						defaultParameters[6] = String.valueOf(40);
						defaultParameters[7] = "-I";
						defaultParameters[8] = "1000";
						defaultParameters[9] = "-num-slots";
						defaultParameters[10] = "80";

						for (int i = 1; i <= 9; i++) {

							String[] parameters = defaultParameters.clone();

							switch (i) {
							case 1:
								parameters[2] = String.valueOf(1);
								break;
							case 2:
								parameters[2] = String.valueOf(2);
								break;
							case 3:
								parameters[2] = String.valueOf(0);
								break;
							case 4:
								parameters[2] = String.valueOf(200);
								break;
							case 5:
								parameters[6] = String.valueOf(20);
								break;
							case 6:
								parameters[6] = String.valueOf(40);
								break;
							case 7:
								parameters[6] = String.valueOf(60);
								break;
							case 8:
								parameters[6] = String.valueOf(80);
								break;
							case 9:
								parameters[6] = String.valueOf(100);
								break;

							}



							System.out.println(Arrays.toString(parameters));

							EntityMatcherWekaLearner wekaLearner = getWekaLearner(parameters, targetRun.getOptimizableClass());
							EM_LearnedModelRepository rfLearnedModel = runner.learnEM(classId, targetRun, wekaLearner, gsData.getA().getClusterList(),
									gsData.getB(), gsData.getC(), gsData.getD());

							CHMHelper.getFromMap(newDetectionMap, entry.getKey() + "_" + i, () -> new HashMap<>()).put(classId, rfLearnedModel);

						}

						;

					}
				}

				if (!newDetection.isEmpty()) {
					ret.put(entry.getKey(), new DS_Container(null, null, newDetection, Lists.newArrayList(targetRun)));
				}

				if (!newDetectionMap.isEmpty()) {
					for (Entry<String, Map<ClassId, EM_LearnedModelRepository>> nEntry : newDetectionMap.entrySet()) {
						ret.put(nEntry.getKey(), new DS_Container(null, null, nEntry.getValue(), Lists.newArrayList(targetRun)));
					}

				}



				// EntityMatchingRunRepository.RF_ET_CS_LABEL_EM_ALL,
			}

			//
			// container.put("KB", new DS_Container(retRC_KB,
			// retND_WITHOUT_WEIGHTS));
			// container.put("BS_WITHOUT_WEIGHTS", new
			// DS_Container(retRC_WITHOUT_WEIGHTS, retND_WITHOUT_WEIGHTS));
			//
			// container.put("BS_WITH_WEIGHTS", new
			// DS_Container(retRC_WITH_WEIGHTS, retND_WITHOUT_WEIGHTS));
			return ret;

			// RowClusterSimilarities sim = runner.getSim(classId, strategy,
			// null);
			//
			// TableRowIdsSet rowIdsSet = input.getFullRowIdsSet(classId);
			// Map<TableRowIdPair, Double> positivePairsWithScores = new
			// HashMap<>();
			//
			// input.getFullRowIdsSet(classId);
			// for (RowCluster rc : clusters.getClusterList()) {
			// if (rc.getRows().size() > 1) {
			// long resultingSize =
			// DoubleIteraterable.getResultingSize(rc.getRows().size());
			// PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(rc.getRows()),
			// (doubleEntry) -> {
			// List<TableRowId> rows = new ArrayList<>();
			// rows.add(doubleEntry.getFirst());
			// rows.add(doubleEntry.getSecond());
			// double score = sim.get(rows, classId, new ClusterEntryKey(0, 1));
			// // double score = sim.get(rows, classId, new
			// // ClusterEntryKey(0, 1));
			// positivePairsWithScores.put(new
			// TableRowIdPair(doubleEntry.getFirst(), doubleEntry.getSecond()),
			// score);
			// }, resultingSize, this.getClass(), rc.getRows().size() < 600L);
			// }
			//
			// }
			//
			// SortedSet<TableRowIdPairScorable> sortedSet = new TreeSet<>();
			// for (Entry<TableRowIdPair, Double> entry :
			// positivePairsWithScores.entrySet()) {
			// if (entry.getValue() > 0.0) {
			// sortedSet.add(new TableRowIdPairScorable(entry.getValue(),
			// entry.getKey()));
			// }
			// }
			//
			// OtsuThresholderRounded thresholder = new
			// OtsuThresholderRounded();
			//
			// List<TableRowIdPairScorable> positivePairs =
			// thresholder.thresholdDefault(new ArrayList<>(sortedSet));
			// positivePairs = sortedSet.stream().filter((e) -> e.getScore() >
			// 0.5).collect(Collectors.toList());
			//
			// Set<TableRowIdPair> negativePairs = new HashSet<>();
			//
			// List<TableRowId> sortedList = new
			// ArrayList<>(rowIdsSet.getRowIds());
			// Collections.sort(sortedList, TableRowId::compare);
			//
			// Random rnd = new Random(classId.getFinalPart().hashCode());
			//
			// for (int i = 0; i < 1000; i++) {
			// int a = rnd.nextInt(sortedList.size());
			// int b = rnd.nextInt(sortedList.size());
			// TableRowIdPair rowIdPair = new TableRowIdPair(sortedList.get(a),
			// sortedList.get(b));
			// if (!positivePairsWithScores.containsKey(rowIdPair)) {
			// negativePairs.add(rowIdPair);
			// }
			// }
			//
			// Map<TableRowIdPair, PairLearningData> rowIdPairs = new
			// HashMap<>();
			// Map<TableRowIdPair, PairLearningData> rowIdPairsWithWeights = new
			// HashMap<>();
			// for (TableRowIdPairScorable pair : positivePairs) {
			// rowIdPairs.put(pair.getPair(), new PairLearningData(true, 1.0));
			// rowIdPairsWithWeights.put(pair.getPair(), new
			// PairLearningData(true, pair.getScore()));
			// }
			//
			// for (TableRowIdPair pair : negativePairs) {
			// rowIdPairs.put(pair, new PairLearningData(false, 1.0));
			// List<TableRowId> rows = new ArrayList<>();
			// rows.add(pair.getId1());
			// rows.add(pair.getId2());
			// rowIdPairsWithWeights.put(pair, new PairLearningData(false,
			// 1.0));
			// }
			//
			// Set<InstanceId> newEntities = new HashSet<>();
			//
			//
			// if (true) {
			// retRC_WITH_WEIGHTS.put(classId, runner.learn(classId, strategy,
			// rowIdPairsWithWeights, newEntities));
			// retRC_WITHOUT_WEIGHTS.put(classId, runner.learn(classId,
			// strategy, rowIdPairs, newEntities));
			// }
			//
			// {
			//
			// EntityMatcherWeightedResults matches =
			// runner.runMatching(classId, clusters.getClusterList(),
			// input.getEntityMatchingRuns().get(0),
			// ndStrategies.get(classId));
			//
			// SortedSet<RowClusterUUIDScorable> positiveWithScores = new
			// TreeSet<>();
			// for (List<ScoredEntity> set :
			// matches.getExistingEntities().values()) {
			// for (ScoredEntity se : set) {
			// positiveWithScores.add(new RowClusterUUIDScorable(se.getScore(),
			// se.getEntity().getRowClusterUUID()));
			// }
			//
			// }
			//
			// Set<RowClusterUUID> keepPositive =
			// thresholder.thresholdDefault(new
			// ArrayList<>(positiveWithScores)).stream()
			// .map((e) -> e.getUUID())
			// .collect(Collectors.toSet());
			//
			// // .filter((e) -> e.getScore() > 0.75)
			// keepPositive = positiveWithScores.stream().map((e) ->
			// e.getUUID()).collect(Collectors.toSet());
			//
			// SortedSet<RowClusterUUIDScorable> negativeWithScores = new
			// TreeSet<>();
			// for (ScoredEntity se : matches.getNewEntities()) {
			//
			// negativeWithScores.add(new RowClusterUUIDScorable(se.getScore(),
			// se.getEntity().getRowClusterUUID()));
			//
			// }
			//
			// Set<RowClusterUUID> keepNegative =
			// thresholder.thresholdDefault(new
			// ArrayList<>(negativeWithScores)).stream()
			// .map((e) -> e.getUUID())
			// .collect(Collectors.toSet());
			//
			// // .filter((e) -> e.getScore() < -0.75)
			// keepNegative = negativeWithScores.stream().map((e) ->
			// e.getUUID()).collect(Collectors.toSet());
			//
			// Map<TableRowIdPair, PairLearningData> rowIdPairs_EM = new
			// HashMap<>();
			//
			// for (List<ScoredEntity> set :
			// matches.getExistingEntities().values()) {
			// Set<TableRowId> rowIds = new HashSet<>();
			// for (ScoredEntity se : set) {
			// if (keepPositive.contains(se.getEntity().getRowClusterUUID())) {
			// rowIds.addAll(se.getEntity().getRowCluster().getRows());
			// }
			// }
			// if (rowIds.size() > 1) {
			// for (DoubleEntry<TableRowId> doubleEntry :
			// DoubleIteraterable.iterable(rowIds)) {
			// rowIdPairs_EM.put(new TableRowIdPair(doubleEntry.getFirst(),
			// doubleEntry.getSecond()), new PairLearningData(true, 1.0));
			// }
			// }
			// }
			//
			// for (ScoredEntity se : matches.getNewEntities()) {
			//
			// if (keepNegative.contains(se.getEntity().getRowClusterUUID()) &&
			// se.getEntity().getRowCluster().getRows().size() > 1) {
			// for (DoubleEntry<TableRowId> doubleEntry :
			// DoubleIteraterable.iterable(se.getEntity().getRowCluster().getRows()))
			// {
			// rowIdPairs_EM.put(new TableRowIdPair(doubleEntry.getFirst(),
			// doubleEntry.getSecond()), new PairLearningData(true, 1.0));
			// }
			// }
			//
			// }
			//
			// List<RowCluster> singleClusters = new ArrayList<>();
			// for (TableRowId rowId : rowIdsSet.getRowIds()) {
			// singleClusters.add(new RowCluster(classId,
			// Lists.newArrayList(rowId), sim));
			// }
			// EntityMatcherWeightedResults singleMatches =
			// runner.runMatching(classId, singleClusters,
			// input.getEntityMatchingRuns().get(0),
			// ndStrategies.get(classId));
			//
			// Map<TableRowId, Optional<InstanceId>> normalized = new
			// HashMap<>();
			// for (ScoredEntity se : singleMatches.getNewEntities()) {
			// for (TableRowId rowId : se.getEntity().getRowCluster().getRows())
			// {
			// normalized.put(rowId, Optional.empty());
			// }
			// }
			//
			// for (Entry<InstanceId, List<ScoredEntity>> entry :
			// singleMatches.getExistingEntities().entrySet()) {
			// for (ScoredEntity se : entry.getValue()) {
			// for (TableRowId rowId : se.getEntity().getRowCluster().getRows())
			// {
			// normalized.put(rowId, Optional.of(entry.getKey()));
			// }
			// }
			// }
			//
			// Set<TableRowIdPair> negativeFromMatch = new HashSet<>();
			//
			// for (TableRowIdPair pair : positivePairsWithScores.keySet()) {
			// Optional<InstanceId> n1 = normalized.get(pair.getId1());
			// Optional<InstanceId> n2 = normalized.get(pair.getId2());
			// if (n1 != null && n2 != null) {
			// if (n1.isPresent() && !n2.isPresent()) {
			// negativeFromMatch.add(pair);
			// // rowIdPairs_EM.put(pair, new
			// // PairLearningData(false, 1.0));
			// } else if (!n1.isPresent() && n2.isPresent()) {
			// negativeFromMatch.add(pair);
			// // rowIdPairs_EM.put(pair, new
			// // PairLearningData(false, 1.0));
			// } else if (n1.isPresent() && n2.isPresent()) {
			// if (!n1.get().equals(n2.get())) {
			// negativeFromMatch.add(pair);
			// // rowIdPairs_EM.put(pair, new
			// // PairLearningData(false, 1.0));
			// }
			// }
			// }
			// }
			//
			// Set<TableRowIdPair> negativeFromHomonym = new HashSet<>();
			//
			// ClusterList homonymClusters = runner.getHomonymClusters(classId,
			// rowIdsSet.getRowIds());
			//
			// for (RowCluster rc : homonymClusters.getClusterList()) {
			// if (rc.getRows().size() > 1) {
			// for (DoubleEntry<TableRowId> doubleEntry :
			// DoubleIteraterable.iterable(rc.getRows())) {
			// TableRowIdPair pair = new TableRowIdPair(doubleEntry.getFirst(),
			// doubleEntry.getSecond());
			//
			// if (!positivePairs.contains(pair)) {
			// negativeFromHomonym.add(pair);
			// }
			//
			// }
			// }
			// }
			//
			// double negativePart = 0.3 * (negativePairs.size() +
			// negativeFromHomonym.size() + negativeFromMatch.size());
			// double matchUpSample = negativePart / negativeFromMatch.size();
			//
			// double homonymUpsample = negativePart /
			// negativeFromHomonym.size();
			//
			// double randomUpsample = negativePart / negativePairs.size();
			//
			// for (TableRowIdPair pair : negativeFromMatch) {
			// rowIdPairs_EM.put(pair, new PairLearningData(false,
			// matchUpSample));
			// }
			//
			// for (TableRowIdPair pair : negativeFromHomonym) {
			// rowIdPairs_EM.put(pair, new PairLearningData(false,
			// homonymUpsample));
			// }
			//
			// for (TableRowIdPair pair : negativePairs) {
			// rowIdPairs_EM.put(pair, new PairLearningData(false,
			// randomUpsample));
			// }
			//
			//
			// System.out.println("matchUpSample: " + matchUpSample);
			// System.out.println("homonymUpsample: " + homonymUpsample);
			// System.out.println("randomUpsample: " + randomUpsample);
			//
			// retRC_KB.put(classId, runner.learn(classId, strategy,
			// rowIdPairs_EM, newEntities));
			// }
			//
			//
			// if (false) {
			//
			// } else {
			// // retND_WITH_WEIGHTS.put(classId,
			// // ndStrategies.get(classId));
			// retND_WITHOUT_WEIGHTS.put(classId, ndStrategies.get(classId));
			// }
			//
			//
			//
			// }
			//
			// Map<String, DS_Container> container = new HashMap<>();
			//
			// container.put("KB", new DS_Container(retRC_KB,
			// retND_WITHOUT_WEIGHTS));
			// container.put("BS_WITHOUT_WEIGHTS", new
			// DS_Container(retRC_WITHOUT_WEIGHTS, retND_WITHOUT_WEIGHTS));
			//
			// container.put("BS_WITH_WEIGHTS", new
			// DS_Container(retRC_WITH_WEIGHTS, retND_WITHOUT_WEIGHTS));
			// return container;

			// } catch (Exception e) {
			// throw new RuntimeException(e);
			// }
		}

		private Pair<Map<RowClusterUUID, InstanceId>, Set<RowClusterUUID>> getClusters(Map<RowClusterUUID, Optional<InstanceId>> mapping,
				Set<InstanceId> newInstances) {

			Set<RowClusterUUID> newSet = new HashSet<>();
			Map<RowClusterUUID, InstanceId> existingMap = new HashMap<>();

			for (Entry<RowClusterUUID, Optional<InstanceId>> d : mapping.entrySet()) {
				if (newInstances.contains(d.getValue().get())) {
					newSet.add(d.getKey());
				} else {
					existingMap.put(d.getKey(), d.getValue().get());
				}
			}

			return new Pair<>(existingMap, newSet);
		}

		private TreeMap<Integer, Map<Pair<RowClusterUUID, InstanceId>, PairLearningData>> getSplitMap() {
			// TODO Auto-generated method stub
			return null;
		}

		private ClusterList getClusters(ClassId classId, Set<TableRowId> rowIds, ClusteringRunner runner) {
			ClusterList clusters = runner.getList(classId, strategy, clusteringModels.get(classId), rowIds);
			return clusters;
		}

		private Quad<ClusterList, Map<Pair<RowClusterUUID, InstanceId>, PairLearningData>, Set<InstanceId>, Map<RowClusterUUID, Optional<InstanceId>>> getGsData(
				ClassId classId, Set<TableRowId> rowIds, ClusteringRunner runner, EntityMatchingRun run, EM_LearnedModelRepository models) {
			ClusterList clusters = getClusters(classId, rowIds, runner);

			return getGsData(clusters, classId, runner, run, models);
		}

		private Quad<ClusterList, Map<Pair<RowClusterUUID, InstanceId>, PairLearningData>, Set<InstanceId>, Map<RowClusterUUID, Optional<InstanceId>>> getGsData(
				ClusterList clusters, ClassId classId, ClusteringRunner runner, EntityMatchingRun run, EM_LearnedModelRepository models) {

			EntityMatcherFullResults fullRun = runner.getFullRun(classId, clusters.getClusterList(), run);

			System.out.println("For Class: " + classId.getFinalPart() + ", For Method: " + run.getName());
			Map<Pair<RowClusterUUID, InstanceId>, Pair<Boolean, Double>> matched = runner.getFullNonAbsoluteEMresults(classId, run, models, fullRun);

			ConcurrentHashMap<Pair<RowClusterUUID, InstanceId>, PairLearningData> pairsFull = new ConcurrentHashMap<>();

			ConcurrentAccumulator<AccumDouble> matchWeight = new ConcurrentAccumulator<AccumDouble>(AccumDouble.empty());
			ConcurrentAccumulator<AccumDouble> nonMatchWeight = new ConcurrentAccumulator<AccumDouble>(AccumDouble.empty());

			PersistentParallel6.runAndFinish(matched.entrySet(), (e) -> {
				// if (!e.getValue().getFirst() && e.getValue().getSecond() >
				// 0.0) {
				// // throw new IllegalStateException();
				// } else if (e.getValue().getFirst() &&
				// e.getValue().getSecond() < 0.0) {
				// // throw new IllegalStateException();
				// }

				if (e.getValue().getSecond() == null) {
					pairsFull.put(e.getKey(), new PairLearningData(e.getValue().getFirst(), -1.0));
				} else {
					pairsFull.put(e.getKey(), new PairLearningData(e.getValue().getFirst(), Math.abs(e.getValue().getSecond())));
				}


			}, this.getClass());

			Pair<Set<InstanceId>, Map<RowClusterUUID, Optional<InstanceId>>> gsData = runner.getGsEmData(pairsFull);

			System.out.println("Existing-Positive: " + (gsData.getSecond().size() - gsData.getFirst().size()));
			System.out.println("New-Classification: " + gsData.getFirst().size());

			Set<InstanceId> secondLearningNewEntities = new HashSet<>();
			Map<RowClusterUUID, Optional<InstanceId>> secondLearningMapping = new HashMap<>();
			ConcurrentHashMap<Pair<RowClusterUUID, InstanceId>, PairLearningData> firstLearningPairs = new ConcurrentHashMap<>();

			double percentage = 0.15;

			int secondEntities = (int) (percentage * Math.min(gsData.getFirst().size(), gsData.getSecond().size() - gsData.getFirst().size()));

			{
				int i = 0;
				for (InstanceId instanceId : gsData.getFirst()) {
					secondLearningNewEntities.add(instanceId);
					i++;
					if (i == secondEntities) {
						break;
					}
				}

				for (Entry<RowClusterUUID, Optional<InstanceId>> entry : gsData.getSecond().entrySet()) {
					if (secondLearningNewEntities.contains(entry.getValue().get())) {
						secondLearningMapping.put(entry.getKey(), entry.getValue());
					}
				}

				int j = 0;
				for (Entry<RowClusterUUID, Optional<InstanceId>> entry : gsData.getSecond().entrySet()) {

					if (!gsData.getFirst().contains(entry.getValue().get())) {
						secondLearningMapping.put(entry.getKey(), entry.getValue());
						j++;
						if (j == secondEntities) {
							break;
						}
					}

				}

			}

			for (Entry<Pair<RowClusterUUID, InstanceId>, PairLearningData> pair : pairsFull.entrySet()) {
				if (!secondLearningMapping.containsKey(pair.getKey().getFirst()) && pair.getValue().getWeight() >= 0.0) {
					firstLearningPairs.put(pair.getKey(), pair.getValue());
				}
			}

			if (false) {
				return new Quad<>(clusters, pairsFull, gsData.getFirst(), gsData.getSecond());
			} else {
				return new Quad<>(clusters, firstLearningPairs, secondLearningNewEntities, secondLearningMapping);
			}

		}

		private EntityMatcherWekaLearner getWekaLearner(String[] parameters, OptimizableClass optimiableClass) {

			// Map<OptimizableWeightKey, Double> weightsMap = new HashMap<>();
			// weightsMap.put(optimiableClass.getKey("THRESHOLD_NEW"), 0.5);
			// weightsMap.put(optimiableClass.getKey("THRESHOLD_MATCH"), 0.5);
			//
			// WeightsRepositoryImpl fixed = new
			// WeightsRepositoryImpl(weightsMap);

			return new EntityMatcherWekaLearner(Sets.newHashSet(EntityMatcherStrategies.ATTRIBUTE, EntityMatcherStrategies.IMPLICIT_ATTRIBUTE), "RandomForest",
					parameters, true, null, false, false, null, true, simpleLearner, generic, repository);
		}


	}
}
