package expansion.set.distant_supervision.bootsrap;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;

import com.beust.jcommander.internal.Lists;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.LearnedModelRepository;
import expansion.set.clustering.model.ClusterSimilaritiesImplNormalized;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities.PairLearningData;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.distant_supervision.DS_Clustering_Strategy_With_Runner;
import expansion.set.distant_supervision.DS_Container;
import expansion.set.distant_supervision.DistantSupervisionInput;
import expansion.set.distant_supervision.runners.DS_ClusteringRunnerCreator.ClusteringRunner;
import expansion.set.exportimport.ClusterPairsExporter;
import expansion.set.strategies.SetExpansionStrategiesRepository;
import expansion.set.strategies.SetExpansionStrategy;
import expansion.set.strategies.ea.ParamterClusterer;

public class WS_RC_No_Clustering_Strategy_Creator {

	// @Inject
	// protected RowClustererRepository rowClustererRepository;

	// @Inject
	// protected ND_Average_Creator ndAverageCreator;

	// @Inject
	// protected RC_Average_Creator rcAverageCreator;

	// @Inject
	// protected EntityMatchingRunner entityMatchingRuner;

	// @Inject
	// protected SetExpansionRunner setExpansionRunner;

	// @Inject
	// protected GreedyClusteringSplitter greedyClusteringSplitter;

	// @Inject
	// protected AnnotatedTablesRepository tableRepository;

	// @Inject
	// protected RowValuesGetter rowValuesGetter;

	// @Inject
	// protected EntitySetGroupedValuesSplitter entitySetGroupedValuesSplitter;

	// @Inject
	// protected EntityCreator entityCreator;

	// @Inject
	// protected ClusteringModelLearnerRunner clusteringLearner;

	@Inject
	public WS_RC_No_Clustering_Strategy_Creator() {

	}

	public WS_RC_Strategy get(String name, SetExpansionStrategy strategy, Map<ClassId, LearnedModelRepository> models) {
		Map<String, Pair<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>>> map = new HashMap<>();
		map.put(name, new Pair<>(strategy, models));

		return new WS_RC_Strategy(map);
	}

	public static enum WS_METHOD {
		WEIGHTED, ROUNDED, ALWAYS_ONE;
	}

	public WS_RC_Strategy get(String name, SetExpansionStrategy strategy, Map<ClassId, LearnedModelRepository> models, WS_METHOD ws_method,
			boolean negativeRandom) {
		Map<String, Pair<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>>> map = new HashMap<>();
		map.put(name, new Pair<>(strategy, models));

		return new WS_RC_Strategy(map, ws_method, negativeRandom);
	}

	public WS_RC_Strategy get(Map<String, Pair<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>>> models) {
		return new WS_RC_Strategy(models);
	}

	public class WS_RC_Strategy implements DS_Clustering_Strategy_With_Runner {
		private final Map<String, Pair<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>>> models;
		private final WS_METHOD ws_method;
		private final boolean negativeRandom;
		public WS_RC_Strategy(Map<String, Pair<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>>> models) {
			super();
			this.models = models;
			this.ws_method = WS_METHOD.ROUNDED;

			this.negativeRandom = false;
		}

		public WS_RC_Strategy(Map<String, Pair<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>>> models, WS_METHOD ws_method,
				boolean negativeRandom) {
			super();
			this.models = models;
			this.ws_method = ws_method;
			this.negativeRandom = negativeRandom;
		}

		@Override
		public Map<String, DS_Container> get(DistantSupervisionInput input, ClusteringRunner runner) {

			// try {

			Map<String, DS_Container> ret = new HashMap<>();
			// Map<ClassId, EM_LearnedModelRepository> retND_WITH_WEIGHTS =
			// new HashMap<>();

			for (Entry<String, Pair<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>>> entry : this.models.entrySet()) {

				Map<ClassId, LearnedModelRepository> rowClustering = new HashMap<>();

				for (ClassId classId : input.getClassesWithParents().keySet()) {

					SetExpansionStrategy strategy = entry.getValue().getFirst();
					ClusterSimilaritiesImplNormalized cs = runner.getNormalizedCS(classId, strategy, entry.getValue().getSecond().get(classId));

					RowClusterSimilarities sim = runner.getSim(classId, strategy, entry.getValue().getSecond().get(classId));
					//
					// ClusterList clusters = runner.getList(classId, strategy,
					// entry.getValue().getSecond().get(classId));
					//
					// Set<TableRowIdPair> positivePairs = new HashSet<>();
					//
					// Set<TableRowIdPair> negativePairs =
					// Collections.newSetFromMap(new ConcurrentHashMap<>());
					//
					// for (RowCluster rc : clusters.getClusterList()) {
					// if (rc.getRows().size() > 1) {
					// long resultingSize =
					// DoubleIteraterable.getResultingSize(rc.getRows().size());
					// PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(rc.getRows()),
					// (doubleEntry) -> {
					// positivePairs.add(new
					// TableRowIdPair(doubleEntry.getFirst(),
					// doubleEntry.getSecond()));
					// }, resultingSize, this.getClass(), rc.getRows().size() <
					// 600L);
					// }
					//
					// }

					Set<TableRowId> rowIds = input.getWsRowIdsSet(classId);

					Random rnd = new Random(classId.getFinalPart().hashCode());
					List<TableRowId> sortedList = new ArrayList<>(rowIds);
					Collections.sort(sortedList, TableRowId::compare);



					Map<TableRowId, Integer> mapping = new HashMap<>();
					Map<Integer, TableRowId> mappingIndex = new HashMap<>();

					for (int i = 0; i < sortedList.size(); i++) {
						mapping.put(sortedList.get(i), i);
						mappingIndex.put(i, sortedList.get(i));
					}


					Map<Integer, Set<Integer>> pairs = ClusterPairsExporter.getPairs(mapping, sim);
					List<TableRowIdPair> blockedSortedList = new ArrayList<>();

					for (Entry<Integer, Set<Integer>> pair : pairs.entrySet()) {
						for (Integer p2 : pair.getValue()) {
							blockedSortedList.add(new TableRowIdPair(mappingIndex.get(pair.getKey()), mappingIndex.get(p2)));
						}
					}

					ConcurrentHashMap<TableRowIdPair, Double> positivePairs = new ConcurrentHashMap<>();
					ConcurrentHashMap<TableRowIdPair, Double> negativePairs = new ConcurrentHashMap<>();

					PersistentParallel6.runAndFinish(blockedSortedList, (tableRowIdPair) -> {
						double score = cs.get(tableRowIdPair);
						if (score > 0.0) {
							positivePairs.put(tableRowIdPair, score);
						} else if (score < 0.0) {
							score = Math.abs(score);
							negativePairs.put(tableRowIdPair, score);
						}
					}, this.getClass());

					ConcurrentHashMap<TableRowIdPair, PairLearningData> rowIdPairs = new ConcurrentHashMap<>();

					if (negativeRandom) {

						Set<TableRowIdPair> additionalPairs = new HashSet<>();
						int collect = 1 * positivePairs.size();
						System.out.println("For ClassId " + classId.getFinalPart() + " positive " + positivePairs.size() + " homonym pairs "
								+ negativePairs.size() + " we collect " + collect);
						// if(collect < rowIdsSet.getRowIds().size() -
						// positivePairs.iterator()) {
						// collect = rowIdsSet.getRowIds().size() -
						// positivePairs.iterator();
						// }
						int i = 0;
						while (i < collect) {
							int a = rnd.nextInt(sortedList.size());
							int b = rnd.nextInt(sortedList.size());
							if (a != b) {
								TableRowIdPair rowIdPair = new TableRowIdPair(sortedList.get(a), sortedList.get(b));

									additionalPairs.add(rowIdPair);
									i++;

							}
						}

						PersistentParallel6.runAndFinish(additionalPairs, (rowIdPair) -> {
							// if (!positivePairs.contains(rowIdPair) &&
							// !negativePairs.contains(rowIdPair)) {
							double score = cs.get(rowIdPair);
							if (score > 0.0) {
								// rowIdPairs.put(rowIdPair, new
								// PairLearningData(true, score));
							} else if (score < 0.0) {
								score = Math.abs(score);

								switch (ws_method) {
								case ALWAYS_ONE:
									rowIdPairs.put(rowIdPair, new PairLearningData(false, 1.0));
									break;
								case ROUNDED:
									rowIdPairs.put(rowIdPair, new PairLearningData(false, Math.round(score)));
									break;
								case WEIGHTED:
									rowIdPairs.put(rowIdPair, new PairLearningData(false, score));
									break;
								default:
									break;

								}

							}
							// }
						}, this.getClass());

					}



					for (Entry<TableRowIdPair, Double> pair : positivePairs.entrySet()) {

						switch (ws_method) {
						case ALWAYS_ONE:
							rowIdPairs.put(pair.getKey(), new PairLearningData(true, 1.0));
							break;
						case ROUNDED:
							rowIdPairs.put(pair.getKey(), new PairLearningData(true, Math.round(pair.getValue())));
							break;
						case WEIGHTED:
							rowIdPairs.put(pair.getKey(), new PairLearningData(true, pair.getValue()));
							break;
						default:
							break;

						}

					}

					for (Entry<TableRowIdPair, Double> pair : negativePairs.entrySet()) {

						switch (ws_method) {
						case ALWAYS_ONE:
							rowIdPairs.put(pair.getKey(), new PairLearningData(false, 1.0));
							break;
						case ROUNDED:
							rowIdPairs.put(pair.getKey(), new PairLearningData(false, Math.round(pair.getValue())));
							break;
						case WEIGHTED:
							rowIdPairs.put(pair.getKey(), new PairLearningData(false, pair.getValue()));
							break;
						default:
							break;

						}

					}

					String[] parameters = new String[11];

					parameters[0] = "-U";
					parameters[1] = "-K";
					parameters[2] = String.valueOf(2);
					parameters[3] = "-V";
					parameters[4] = "1e-" + 3;
					parameters[5] = "-P";
					parameters[6] = String.valueOf(100);
					parameters[7] = "-I";
					parameters[8] = "1000";
					parameters[9] = "-num-slots";
					parameters[10] = "80";

					ParamterClusterer parameterClusterer = runner.getParamterClusterer(classId, SetExpansionStrategiesRepository.COMBINED_RandomForest_ET,
							new HashSet<>());

					LearnedModelRepository rfLearnedModel = runner.learnFromParamter(classId, SetExpansionStrategiesRepository.COMBINED_RandomForest_ET,
							parameterClusterer, parameters.clone(), rowIdPairs);

					rowClustering.put(classId, rfLearnedModel);
				}

				ret.put(entry.getKey(),
						new DS_Container(rowClustering, Lists.newArrayList(SetExpansionStrategiesRepository.COMBINED_RandomForest_ET), null));
			}

			//
			// container.put("KB", new DS_Container(retRC_KB,
			// retND_WITHOUT_WEIGHTS));
			// container.put("BS_WITHOUT_WEIGHTS", new
			// DS_Container(retRC_WITHOUT_WEIGHTS, retND_WITHOUT_WEIGHTS));
			//
			// container.put("BS_WITH_WEIGHTS", new
			// DS_Container(retRC_WITH_WEIGHTS, retND_WITHOUT_WEIGHTS));
			return ret;

			// RowClusterSimilarities sim = runner.getSim(classId, strategy,
			// null);
			//
			// TableRowIdsSet rowIdsSet = input.getFullRowIdsSet(classId);
			// Map<TableRowIdPair, Double> positivePairsWithScores = new
			// HashMap<>();
			//
			// input.getFullRowIdsSet(classId);
			// for (RowCluster rc : clusters.getClusterList()) {
			// if (rc.getRows().size() > 1) {
			// long resultingSize =
			// DoubleIteraterable.getResultingSize(rc.getRows().size());
			// PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(rc.getRows()),
			// (doubleEntry) -> {
			// List<TableRowId> rows = new ArrayList<>();
			// rows.add(doubleEntry.getFirst());
			// rows.add(doubleEntry.getSecond());
			// double score = sim.get(rows, classId, new ClusterEntryKey(0, 1));
			// // double score = sim.get(rows, classId, new
			// // ClusterEntryKey(0, 1));
			// positivePairsWithScores.put(new
			// TableRowIdPair(doubleEntry.getFirst(), doubleEntry.getSecond()),
			// score);
			// }, resultingSize, this.getClass(), rc.getRows().size() < 600L);
			// }
			//
			// }
			//
			// SortedSet<TableRowIdPairScorable> sortedSet = new TreeSet<>();
			// for (Entry<TableRowIdPair, Double> entry :
			// positivePairsWithScores.entrySet()) {
			// if (entry.getValue() > 0.0) {
			// sortedSet.add(new TableRowIdPairScorable(entry.getValue(),
			// entry.getKey()));
			// }
			// }
			//
			// OtsuThresholderRounded thresholder = new
			// OtsuThresholderRounded();
			//
			// List<TableRowIdPairScorable> positivePairs =
			// thresholder.thresholdDefault(new ArrayList<>(sortedSet));
			// positivePairs = sortedSet.stream().filter((e) -> e.getScore() >
			// 0.5).collect(Collectors.toList());
			//
			// Set<TableRowIdPair> negativePairs = new HashSet<>();
			//
			// List<TableRowId> sortedList = new
			// ArrayList<>(rowIdsSet.getRowIds());
			// Collections.sort(sortedList, TableRowId::compare);
			//
			// Random rnd = new Random(classId.getFinalPart().hashCode());
			//
			// for (int i = 0; i < 1000; i++) {
			// int a = rnd.nextInt(sortedList.size());
			// int b = rnd.nextInt(sortedList.size());
			// TableRowIdPair rowIdPair = new TableRowIdPair(sortedList.get(a),
			// sortedList.get(b));
			// if (!positivePairsWithScores.containsKey(rowIdPair)) {
			// negativePairs.add(rowIdPair);
			// }
			// }
			//
			// Map<TableRowIdPair, PairLearningData> rowIdPairs = new
			// HashMap<>();
			// Map<TableRowIdPair, PairLearningData> rowIdPairsWithWeights = new
			// HashMap<>();
			// for (TableRowIdPairScorable pair : positivePairs) {
			// rowIdPairs.put(pair.getPair(), new PairLearningData(true, 1.0));
			// rowIdPairsWithWeights.put(pair.getPair(), new
			// PairLearningData(true, pair.getScore()));
			// }
			//
			// for (TableRowIdPair pair : negativePairs) {
			// rowIdPairs.put(pair, new PairLearningData(false, 1.0));
			// List<TableRowId> rows = new ArrayList<>();
			// rows.add(pair.getId1());
			// rows.add(pair.getId2());
			// rowIdPairsWithWeights.put(pair, new PairLearningData(false,
			// 1.0));
			// }
			//
			// Set<InstanceId> newEntities = new HashSet<>();
			//
			//
			// if (true) {
			// retRC_WITH_WEIGHTS.put(classId, runner.learn(classId, strategy,
			// rowIdPairsWithWeights, newEntities));
			// retRC_WITHOUT_WEIGHTS.put(classId, runner.learn(classId,
			// strategy, rowIdPairs, newEntities));
			// }
			//
			// {
			//
			// EntityMatcherWeightedResults matches =
			// runner.runMatching(classId, clusters.getClusterList(),
			// input.getEntityMatchingRuns().get(0),
			// ndStrategies.get(classId));
			//
			// SortedSet<RowClusterUUIDScorable> positiveWithScores = new
			// TreeSet<>();
			// for (List<ScoredEntity> set :
			// matches.getExistingEntities().values()) {
			// for (ScoredEntity se : set) {
			// positiveWithScores.add(new RowClusterUUIDScorable(se.getScore(),
			// se.getEntity().getRowClusterUUID()));
			// }
			//
			// }
			//
			// Set<RowClusterUUID> keepPositive =
			// thresholder.thresholdDefault(new
			// ArrayList<>(positiveWithScores)).stream()
			// .map((e) -> e.getUUID())
			// .collect(Collectors.toSet());
			//
			// // .filter((e) -> e.getScore() > 0.75)
			// keepPositive = positiveWithScores.stream().map((e) ->
			// e.getUUID()).collect(Collectors.toSet());
			//
			// SortedSet<RowClusterUUIDScorable> negativeWithScores = new
			// TreeSet<>();
			// for (ScoredEntity se : matches.getNewEntities()) {
			//
			// negativeWithScores.add(new RowClusterUUIDScorable(se.getScore(),
			// se.getEntity().getRowClusterUUID()));
			//
			// }
			//
			// Set<RowClusterUUID> keepNegative =
			// thresholder.thresholdDefault(new
			// ArrayList<>(negativeWithScores)).stream()
			// .map((e) -> e.getUUID())
			// .collect(Collectors.toSet());
			//
			// // .filter((e) -> e.getScore() < -0.75)
			// keepNegative = negativeWithScores.stream().map((e) ->
			// e.getUUID()).collect(Collectors.toSet());
			//
			// Map<TableRowIdPair, PairLearningData> rowIdPairs_EM = new
			// HashMap<>();
			//
			// for (List<ScoredEntity> set :
			// matches.getExistingEntities().values()) {
			// Set<TableRowId> rowIds = new HashSet<>();
			// for (ScoredEntity se : set) {
			// if (keepPositive.contains(se.getEntity().getRowClusterUUID())) {
			// rowIds.addAll(se.getEntity().getRowCluster().getRows());
			// }
			// }
			// if (rowIds.size() > 1) {
			// for (DoubleEntry<TableRowId> doubleEntry :
			// DoubleIteraterable.iterable(rowIds)) {
			// rowIdPairs_EM.put(new TableRowIdPair(doubleEntry.getFirst(),
			// doubleEntry.getSecond()), new PairLearningData(true, 1.0));
			// }
			// }
			// }
			//
			// for (ScoredEntity se : matches.getNewEntities()) {
			//
			// if (keepNegative.contains(se.getEntity().getRowClusterUUID()) &&
			// se.getEntity().getRowCluster().getRows().size() > 1) {
			// for (DoubleEntry<TableRowId> doubleEntry :
			// DoubleIteraterable.iterable(se.getEntity().getRowCluster().getRows()))
			// {
			// rowIdPairs_EM.put(new TableRowIdPair(doubleEntry.getFirst(),
			// doubleEntry.getSecond()), new PairLearningData(true, 1.0));
			// }
			// }
			//
			// }
			//
			// List<RowCluster> singleClusters = new ArrayList<>();
			// for (TableRowId rowId : rowIdsSet.getRowIds()) {
			// singleClusters.add(new RowCluster(classId,
			// Lists.newArrayList(rowId), sim));
			// }
			// EntityMatcherWeightedResults singleMatches =
			// runner.runMatching(classId, singleClusters,
			// input.getEntityMatchingRuns().get(0),
			// ndStrategies.get(classId));
			//
			// Map<TableRowId, Optional<InstanceId>> normalized = new
			// HashMap<>();
			// for (ScoredEntity se : singleMatches.getNewEntities()) {
			// for (TableRowId rowId : se.getEntity().getRowCluster().getRows())
			// {
			// normalized.put(rowId, Optional.empty());
			// }
			// }
			//
			// for (Entry<InstanceId, List<ScoredEntity>> entry :
			// singleMatches.getExistingEntities().entrySet()) {
			// for (ScoredEntity se : entry.getValue()) {
			// for (TableRowId rowId : se.getEntity().getRowCluster().getRows())
			// {
			// normalized.put(rowId, Optional.of(entry.getKey()));
			// }
			// }
			// }
			//
			// Set<TableRowIdPair> negativeFromMatch = new HashSet<>();
			//
			// for (TableRowIdPair pair : positivePairsWithScores.keySet()) {
			// Optional<InstanceId> n1 = normalized.get(pair.getId1());
			// Optional<InstanceId> n2 = normalized.get(pair.getId2());
			// if (n1 != null && n2 != null) {
			// if (n1.isPresent() && !n2.isPresent()) {
			// negativeFromMatch.add(pair);
			// // rowIdPairs_EM.put(pair, new
			// // PairLearningData(false, 1.0));
			// } else if (!n1.isPresent() && n2.isPresent()) {
			// negativeFromMatch.add(pair);
			// // rowIdPairs_EM.put(pair, new
			// // PairLearningData(false, 1.0));
			// } else if (n1.isPresent() && n2.isPresent()) {
			// if (!n1.get().equals(n2.get())) {
			// negativeFromMatch.add(pair);
			// // rowIdPairs_EM.put(pair, new
			// // PairLearningData(false, 1.0));
			// }
			// }
			// }
			// }
			//
			// Set<TableRowIdPair> negativeFromHomonym = new HashSet<>();
			//
			// ClusterList homonymClusters = runner.getHomonymClusters(classId,
			// rowIdsSet.getRowIds());
			//
			// for (RowCluster rc : homonymClusters.getClusterList()) {
			// if (rc.getRows().size() > 1) {
			// for (DoubleEntry<TableRowId> doubleEntry :
			// DoubleIteraterable.iterable(rc.getRows())) {
			// TableRowIdPair pair = new TableRowIdPair(doubleEntry.getFirst(),
			// doubleEntry.getSecond());
			//
			// if (!positivePairs.contains(pair)) {
			// negativeFromHomonym.add(pair);
			// }
			//
			// }
			// }
			// }
			//
			// double negativePart = 0.3 * (negativePairs.size() +
			// negativeFromHomonym.size() + negativeFromMatch.size());
			// double matchUpSample = negativePart / negativeFromMatch.size();
			//
			// double homonymUpsample = negativePart /
			// negativeFromHomonym.size();
			//
			// double randomUpsample = negativePart / negativePairs.size();
			//
			// for (TableRowIdPair pair : negativeFromMatch) {
			// rowIdPairs_EM.put(pair, new PairLearningData(false,
			// matchUpSample));
			// }
			//
			// for (TableRowIdPair pair : negativeFromHomonym) {
			// rowIdPairs_EM.put(pair, new PairLearningData(false,
			// homonymUpsample));
			// }
			//
			// for (TableRowIdPair pair : negativePairs) {
			// rowIdPairs_EM.put(pair, new PairLearningData(false,
			// randomUpsample));
			// }
			//
			//
			// System.out.println("matchUpSample: " + matchUpSample);
			// System.out.println("homonymUpsample: " + homonymUpsample);
			// System.out.println("randomUpsample: " + randomUpsample);
			//
			// retRC_KB.put(classId, runner.learn(classId, strategy,
			// rowIdPairs_EM, newEntities));
			// }
			//
			//
			// if (false) {
			//
			// } else {
			// // retND_WITH_WEIGHTS.put(classId,
			// // ndStrategies.get(classId));
			// retND_WITHOUT_WEIGHTS.put(classId, ndStrategies.get(classId));
			// }
			//
			//
			//
			// }
			//
			// Map<String, DS_Container> container = new HashMap<>();
			//
			// container.put("KB", new DS_Container(retRC_KB,
			// retND_WITHOUT_WEIGHTS));
			// container.put("BS_WITHOUT_WEIGHTS", new
			// DS_Container(retRC_WITHOUT_WEIGHTS, retND_WITHOUT_WEIGHTS));
			//
			// container.put("BS_WITH_WEIGHTS", new
			// DS_Container(retRC_WITH_WEIGHTS, retND_WITHOUT_WEIGHTS));
			// return container;

			// } catch (Exception e) {
			// throw new RuntimeException(e);
			// }
		}



	}
}
