package expansion.set.distant_supervision.bootsrap;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Random;
import java.util.Set;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import com.beust.jcommander.internal.Lists;

import cern.colt.Arrays;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.DoubleIteraterable;
import de.mannheim.uni.expansion.utils.MapSorter;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.LearnedModelRepository;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.model.ClusterSimilaritiesImplNormalized;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities.PairLearningData;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.creation.Entity;
import expansion.set.distant_supervision.DS_Clustering_Strategy_With_Runner;
import expansion.set.distant_supervision.DS_Container;
import expansion.set.distant_supervision.DistantSupervisionInput;
import expansion.set.distant_supervision.matching_rules.MatchingRulesWrapperCreator;
import expansion.set.distant_supervision.matching_rules.MatchingRulesWrapperCreator.MR_ClusterSimilarities;
import expansion.set.distant_supervision.runners.DS_ClusteringRunnerCreator.ClusteringRunner;
import expansion.set.entitymatching.EntityMatcherWeightedResults;
import expansion.set.entitymatching.modelLearning.EM_LearnedModelRepository;
import expansion.set.entitymatching.run.EntityMatchingRun;
import expansion.set.evaluation.goldstandard.oktie.OktieClusterEvaluator;
import expansion.set.evaluation.goldstandard.oktie.OktiePerformance;
import expansion.set.evaluation.matching.ScoredEntity;
import expansion.set.exportimport.ClusterPairsExporter;
import expansion.set.strategies.SetExpansionStrategy;
import expansion.set.strategies.clustersplitters.minScoreSplitter.WekaLearnableRowClusterSimilarities;
import expansion.set.strategies.ea.ParamterClusterer;

public class WS_RC_Strategy_Creator {

	// @Inject
	// protected RowClustererRepository rowClustererRepository;

	// @Inject
	// protected ND_Average_Creator ndAverageCreator;

	// @Inject
	// protected RC_Average_Creator rcAverageCreator;

	// @Inject
	// protected EntityMatchingRunner entityMatchingRuner;

	// @Inject
	// protected SetExpansionRunner setExpansionRunner;

	// @Inject
	// protected GreedyClusteringSplitter greedyClusteringSplitter;

	// @Inject
	// protected AnnotatedTablesRepository tableRepository;

	// @Inject
	// protected RowValuesGetter rowValuesGetter;

	// @Inject
	// protected EntitySetGroupedValuesSplitter entitySetGroupedValuesSplitter;

	// @Inject
	// protected EntityCreator entityCreator;

	// @Inject
	// protected ClusteringModelLearnerRunner clusteringLearner;

	@Inject
	protected OktieClusterEvaluator oktieClusterEvaluator;

	@Inject
	protected MatchingRulesWrapperCreator matchingRulesWrapperCreator;

	@Inject
	public WS_RC_Strategy_Creator() {

	}

	public WS_RC_Strategy get(String name, SetExpansionStrategy strategy, Map<ClassId, LearnedModelRepository> models, SetExpansionStrategy targetStrategy) {
		Map<String, Tri<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>, SetExpansionStrategy>> map = new HashMap<>();
		map.put(name, new Tri<>(strategy, models, targetStrategy));// SetExpansionStrategiesRepository.COMBINED_RandomForest_NEW));

		return new WS_RC_Strategy(map, false);
	}

	public WS_RC_Strategy getWithDistantSupervision(String name, SetExpansionStrategy strategy, Map<ClassId, LearnedModelRepository> models,
			SetExpansionStrategy targetStrategy, EntityMatchingRun ndStrategy, Map<ClassId, EM_LearnedModelRepository> newDetectionStrategies) {
		Map<String, Tri<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>, SetExpansionStrategy>> map = new HashMap<>();
		map.put(name, new Tri<>(strategy, models, targetStrategy));

		return new WS_RC_Strategy(map, true, ndStrategy, newDetectionStrategies);
	}


	public WS_RC_Strategy get(Map<String, Tri<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>, SetExpansionStrategy>> models) {
		return new WS_RC_Strategy(models, false);
	}

	public class WS_RC_Strategy implements DS_Clustering_Strategy_With_Runner {
		private final Map<String, Tri<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>, SetExpansionStrategy>> models;
		private final EntityMatchingRun dsNdStrategy;
		private final Map<ClassId, EM_LearnedModelRepository> dsNewDetectionStrategies;

		public WS_RC_Strategy(Map<String, Tri<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>, SetExpansionStrategy>> models,
				boolean withDistantSupervision) {
			super();
			this.models = models;
			this.dsNdStrategy = null;
			this.dsNewDetectionStrategies = null;
		}

		public WS_RC_Strategy(Map<String, Tri<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>, SetExpansionStrategy>> models,
				boolean withDistantSupervision, EntityMatchingRun ndStrategy, Map<ClassId, EM_LearnedModelRepository> newDetectionStrategies) {
			super();
			this.models = models;
			this.dsNdStrategy = ndStrategy;
			this.dsNewDetectionStrategies = newDetectionStrategies;
		}

		@Override
		public Map<String, DS_Container> get(DistantSupervisionInput input, ClusteringRunner runner) {

			HyperParamter_RC hyper = new HyperParamter_RC();
			// try {

			Map<String, DS_Container> ret = new HashMap<>();
			// Map<ClassId, EM_LearnedModelRepository> retND_WITH_WEIGHTS =
			// new HashMap<>();

			for (Entry<String, Tri<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>, SetExpansionStrategy>> entry : this.models.entrySet()) {

				Map<ClassId, LearnedModelRepository> rowClustering = new HashMap<>();
				Map<String, Map<ClassId, LearnedModelRepository>> rowClusteringNew = new HashMap<>();

				for (ClassId classId : input.getClassesWithParents().keySet()) {

					SetExpansionStrategy strategy = entry.getValue().getA();
					ClusterSimilaritiesImplNormalized cs = runner.getNormalizedCS(classId, strategy, entry.getValue().getB().get(classId));

					RowClusterSimilarities sim = runner.getSim(classId, strategy, entry.getValue().getB().get(classId));

					if (false) {
						WekaLearnableRowClusterSimilarities fullSim = (WekaLearnableRowClusterSimilarities) runner.getSim(classId, entry.getValue().getC(),
								null);
						double threshold = entry.getValue().getB().get(classId).get(strategy.getOptimizableClassCluster()).getNormalizingThreshold().get();

						sim = matchingRulesWrapperCreator.getSim(sim, fullSim);
					}

					if (sim instanceof MR_ClusterSimilarities) {
						((MR_ClusterSimilarities) sim).clean();
					}

					ClusterList clusters = runner.getList(classId, strategy, entry.getValue().getB().get(classId));

					Set<TableRowId> rowIds = input.getWsRowIdsSet(classId);
					if (false) {
						Set<TableRowIdPair> positivePairs = Collections.newSetFromMap(new ConcurrentHashMap<>());

						Set<TableRowIdPair> negativePairs = Collections.newSetFromMap(new ConcurrentHashMap<>());

						for (RowCluster rc : clusters.getClusterList()) {
							if (rc.getRows().size() > 1) {
								long resultingSize = DoubleIteraterable.getResultingSize(rc.getRows().size());
								PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(rc.getRows()), (doubleEntry) -> {
									positivePairs.add(new TableRowIdPair(doubleEntry.getFirst(), doubleEntry.getSecond()));
								}, resultingSize, this.getClass(), rc.getRows().size() < 600L);
							}

						}



						Random rnd = new Random(classId.getFinalPart().hashCode());
						List<TableRowId> sortedList = new ArrayList<>(rowIds);
						Collections.sort(sortedList, TableRowId::compare);



						Map<TableRowId, Integer> mapping = new HashMap<>();
						Map<Integer, TableRowId> mappingIndex = new HashMap<>();

						for (int i = 0; i < sortedList.size(); i++) {
							mapping.put(sortedList.get(i), i);
							mappingIndex.put(i, sortedList.get(i));
						}


						Map<Integer, Set<Integer>> pairs = ClusterPairsExporter.getPairs(mapping, sim);
						List<TableRowIdPair> blockedSortedList = new ArrayList<>();

						for (Entry<Integer, Set<Integer>> pair : pairs.entrySet()) {
							for (Integer p2 : pair.getValue()) {
								blockedSortedList.add(new TableRowIdPair(mappingIndex.get(pair.getKey()), mappingIndex.get(p2)));
							}
						}

						Collections.sort(blockedSortedList, TableRowIdPair::compare);
						{
							// int collect = positivePairs.size() * 8;
							// if (true || collect > blockedSortedList.size() -
							// positivePairs.size() - negativePairs.size()) {
							// System.out.println("Changed collect from " +
							// collect
							// + " to "
							// + (blockedSortedList.size() -
							// positivePairs.size() -
							// negativePairs.size()) + " classId: " +
							// classId.getFinalPart());
							// collect = blockedSortedList.size() -
							// positivePairs.size() - negativePairs.size();
							//
							// }
							Collections.shuffle(blockedSortedList, rnd);

							for (int i = 0; i < blockedSortedList.size(); i++) {
								TableRowIdPair rowIdPair = blockedSortedList.get(i);
								if (!positivePairs.contains(rowIdPair) && !negativePairs.contains(rowIdPair)) {
									negativePairs.add(rowIdPair);
								}
							}

						}

						if (true) {
							int collect = 4 * positivePairs.size();
							System.out.println("For ClassId " + classId.getFinalPart() + " positive " + positivePairs.size() + " homonym pairs "
									+ negativePairs.size() + " we collect " + collect);
							// if(collect < rowIdsSet.getRowIds().size() -
							// positivePairs.iterator()) {
							// collect = rowIdsSet.getRowIds().size() -
							// positivePairs.iterator();
							// }
							int i = 0;
							while (i < collect) {
								int a = rnd.nextInt(sortedList.size());
								int b = rnd.nextInt(sortedList.size());
								if (a != b) {
									TableRowIdPair rowIdPair = new TableRowIdPair(sortedList.get(a), sortedList.get(b));
									if (!positivePairs.contains(rowIdPair) && !negativePairs.contains(rowIdPair)) {
										negativePairs.add(rowIdPair);
										i++;
									}
								}
							}
						}

						// long resultingSize =
						// DoubleIteraterable.getResultingSize(rowIdsSet.getRowIds().size());
						//
						// PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(rowIdsSet.getRowIds()),
						// (doubleEntry) -> {
						// TableRowIdPair pair = new
						// TableRowIdPair(doubleEntry.getFirst(),
						// doubleEntry.getSecond());
						// if (!positivePairs.contains(pair)) {
						// negativePairs.add(pair);
						// }
						// }, resultingSize, this.getClass());

						ConcurrentHashMap<TableRowIdPair, PairLearningData> rowIdPairs = new ConcurrentHashMap<>();
						for (TableRowIdPair pair : positivePairs) {
							double score = cs.get(pair);

							if (score < 0.0) {
								score = 1.0;
							} else {
								score = Math.abs(score);
								// score = 1.0;
								rowIdPairs.put(pair, new PairLearningData(true, score));
							}

						}

						PersistentParallel6.runAndFinish(negativePairs, (pair) -> {
							double score = cs.get(pair);

							if (score > 0.0) {
								score = 1.0;
							} else {
								score = Math.abs(score);
								// score = 1.0;
								rowIdPairs.put(pair, new PairLearningData(false, score));
							}

						}, this.getClass());
					}

					int negativeMultipliyer = 2;
					int negativeRuns = 4;

					LearnedModelRepository learnedModelRepository = entry.getValue().getB().get(classId);

					Pair<Map<TableRowIdPair, PairLearningData>, List<Map<TableRowIdPair, PairLearningData>>> fullTrainingMap = getTrainingMap(classId, rowIds,
							runner, strategy, learnedModelRepository, sim, cs, negativeMultipliyer, negativeRuns, dsNdStrategy, dsNewDetectionStrategies);


					LearnedModelRepository rfLearnedModel;

					if (false) {

						int defaultRandom = 0;

						String[] defaultParameters = new String[11];

						defaultParameters[0] = "-U";
						defaultParameters[1] = "-K";
						defaultParameters[2] = String.valueOf(2);
						defaultParameters[3] = "-V";
						defaultParameters[4] = "1e-" + 3;
						defaultParameters[5] = "-P";
						defaultParameters[6] = String.valueOf(40);
						defaultParameters[7] = "-I";
						defaultParameters[8] = "500";
						defaultParameters[9] = "-num-slots";
						defaultParameters[10] = "80";

						ParamterClusterer parameterClusterer = runner.getParamterClusterer(classId, entry.getValue().getC(), new HashSet<>());

						for (int i = 1; i <= 27; i++) {

							String[] parameters = defaultParameters.clone();
							int random = defaultRandom;


							switch (i) {
							case 1:
								parameters[2] = String.valueOf(1);
								parameters[6] = String.valueOf(33);
								random = 0;
								break;
							case 2:
								parameters[2] = String.valueOf(1);
								parameters[6] = String.valueOf(33);
								random = 1;
								break;
							case 3:
								parameters[2] = String.valueOf(1);
								parameters[6] = String.valueOf(33);
								random = 2;
								break;
							case 4:
								parameters[2] = String.valueOf(1);
								parameters[6] = String.valueOf(66);
								random = 0;
								break;
							case 5:
								parameters[2] = String.valueOf(1);
								parameters[6] = String.valueOf(66);
								random = 1;
								break;
							case 6:
								parameters[2] = String.valueOf(1);
								parameters[6] = String.valueOf(66);
								random = 2;

								break;
							case 7:
								parameters[2] = String.valueOf(1);
								parameters[6] = String.valueOf(100);
								random = 0;
								break;
							case 8:
								parameters[2] = String.valueOf(1);
								parameters[6] = String.valueOf(100);
								random = 1;
								break;
							case 9:
								parameters[2] = String.valueOf(1);
								parameters[6] = String.valueOf(100);
								random = 2;
								break;
							case 10:
								parameters[2] = String.valueOf(2);
								parameters[6] = String.valueOf(33);
								random = 0;
								break;
							case 11:
								parameters[2] = String.valueOf(2);
								parameters[6] = String.valueOf(33);
								random = 1;
								break;

							case 12:
								parameters[2] = String.valueOf(2);
								parameters[6] = String.valueOf(33);
								random = 2;
								break;

							case 13:
								parameters[2] = String.valueOf(2);
								parameters[6] = String.valueOf(66);
								random = 0;
								break;
							case 14:
								parameters[2] = String.valueOf(2);
								parameters[6] = String.valueOf(66);
								random = 1;
								break;
							case 15:
								parameters[2] = String.valueOf(2);
								parameters[6] = String.valueOf(66);
								random = 2;
								break;
							case 16:
								parameters[2] = String.valueOf(2);
								parameters[6] = String.valueOf(100);
								random = 0;
								break;
							case 17:
								parameters[2] = String.valueOf(2);
								parameters[6] = String.valueOf(100);
								random = 1;
								break;
							case 18:
								parameters[2] = String.valueOf(2);
								parameters[6] = String.valueOf(100);
								random = 2;
								break;
							case 19:
								parameters[2] = String.valueOf(0);
								parameters[6] = String.valueOf(33);
								random = 0;
								break;
							case 20:
								parameters[2] = String.valueOf(0);
								parameters[6] = String.valueOf(33);
								random = 1;
								break;
							case 21:
								parameters[2] = String.valueOf(0);
								parameters[6] = String.valueOf(33);
								random = 2;
								break;
							case 22:
								parameters[2] = String.valueOf(0);
								parameters[6] = String.valueOf(66);
								random = 0;
								break;
							case 23:
								parameters[2] = String.valueOf(0);
								parameters[6] = String.valueOf(66);
								random = 1;
								break;
							case 24:
								parameters[2] = String.valueOf(0);
								parameters[6] = String.valueOf(66);
								random = 2;
								break;
							case 25:
								parameters[2] = String.valueOf(0);
								parameters[6] = String.valueOf(100);
								random = 0;
								break;
							case 26:
								parameters[2] = String.valueOf(0);
								parameters[6] = String.valueOf(100);
								random = 1;
								break;
							case 27:
								parameters[2] = String.valueOf(0);
								parameters[6] = String.valueOf(100);
								random = 2;
								break;
							// case 17:
							// parameters[6] = String.valueOf(100);
							// break;
							}

							// switch (i) {
							// case 1:
							// parameters[2] = String.valueOf(1);
							// break;
							// case 2:
							// parameters[2] = String.valueOf(2);
							// break;
							// case 3:
							// parameters[2] = String.valueOf(0);
							// break;
							// case 4:
							// parameters[2] = String.valueOf(200);
							// break;
							// case 5:
							// parameters[6] = String.valueOf(20);
							// break;
							// case 6:
							// parameters[6] = String.valueOf(40);
							// continue;
							// // break;
							// case 7:
							// parameters[6] = String.valueOf(60);
							// break;
							// case 8:
							// parameters[6] = String.valueOf(80);
							// break;
							// case 9:
							// parameters[6] = String.valueOf(100);
							// break;
							// case 10:
							// random = 0;
							// continue;
							// // break;
							//
							// case 11:
							// random = 1;
							// break;
							//
							// case 12:
							// random = 2;
							// break;
							//
							// case 13:
							// random = 3;
							// break;
							// case 14:
							// parameters[8] = "250";
							// break;
							// case 15:
							// parameters[8] = "500";
							// break;
							// case 16:
							// parameters[8] = "1000";
							// continue;
							// case 17:
							// parameters[8] = "2000";
							// break;
							// case 18:
							// parameters[4] = "1e-" + 1;
							// break;
							// case 19:
							// parameters[4] = "1e-" + 2;
							// break;
							// // break;
							// case 20:
							// parameters[4] = "1e-" + 3;
							// continue;
							// case 21:
							// parameters[4] = "1e-" + 4;
							// break;
							// // case 17:
							// // parameters[6] = String.valueOf(100);
							// // break;
							// }

							Map<TableRowIdPair, PairLearningData> rowIdPairs = new HashMap<>(fullTrainingMap.getFirst());

							for (int r = 1; r <= random; r++) {
								rowIdPairs.putAll(fullTrainingMap.getSecond().get(r - 1));
							}

							System.out.println(Arrays.toString(parameters));

							rfLearnedModel = runner.learnFromParamter(classId, entry.getValue().getC(), parameterClusterer, parameters, rowIdPairs);

							CHMHelper.getFromMap(rowClusteringNew, entry.getKey() + "_" + i, () -> new HashMap<>()).put(classId, rfLearnedModel);

						}

					} else {
					if (true) {

						final String[] parameters;
						final Integer random;

						ParamterClusterer parameterClusterer = runner.getParamterClusterer(classId, entry.getValue().getC(), new HashSet<>());



						if (false) {



							if (true) {
								 
								int folds = 3;


								TreeMap<Integer, Pair<Set<AnnotatedTableId>, Set<TableRowId>>> splitTables = splitTables(folds, classId,
										rowIds.stream().map((e) -> e.getTableId()).collect(Collectors.toSet()), rowIds);
								
								Map<Integer, Pair<Map<TableRowId, InstanceId>, Map<InstanceId, Set<TableRowId>>>> testingDataMap = getTestingMap(classId,
										runner, strategy,
										splitTables.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().getSecond())),
										learnedModelRepository);
								Map<Integer, Pair<Map<TableRowIdPair, PairLearningData>, List<Map<TableRowIdPair, PairLearningData>>>> trainingDataMap = getTrainingMap(
										classId, runner, strategy, learnedModelRepository, sim, cs,

										splitTables.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().getSecond())),
										negativeMultipliyer, negativeRuns);

								
							 
								String[] insideParameters = null;
								Double max = null;
								Integer maxRandom = null;

								for (int r = 0; r < negativeRuns; r++) {
									System.out.println("Random X Positive: " + r * negativeMultipliyer);
									
									TreeMap<Integer, Map<TableRowIdPair, PairLearningData>> splitMap = new TreeMap<>();
									
										for(int fold = 0; fold < folds; fold++) {
											Map<TableRowIdPair, PairLearningData> pairInner = new HashMap<>();
										pairInner.putAll(trainingDataMap.get(fold).getFirst());
												if(r > 0) {
											pairInner.putAll(trainingDataMap.get(fold).getSecond().get(0));
												}
												if(r > 1) {
											pairInner.putAll(trainingDataMap.get(fold).getSecond().get(1));

												}

										splitMap.put(fold, pairInner);
											}

									
									
									Pair<String[], Double> parametersForRun = hyper.findOptimumParamtersInner(classId.getFinalPart().hashCode(), splitMap,
											(fold, trainingData, testingData, pp) -> {


										LearnedModelRepository learnedModel = runner.learnFromParamter(classId, entry.getValue().getC(), parameterClusterer, pp,
												trainingData);

												ClusterList clusters2 = runner.getList(classId, entry.getValue().getC(), learnedModel,
														splitTables.get(fold).getSecond());
												List<Set<TableRowId>> clusters2s = clusters2.getClusterList().stream().map((e) -> new HashSet<>(e.getRows()))
														.collect(Collectors.toList());

												OktiePerformance clusterPerformance = oktieClusterEvaluator.oktieEvaluate(testingDataMap.get(fold).getSecond(),
														testingDataMap.get(fold).getFirst(), clusters2s);

												return clusterPerformance.getFMeasureMacro();


									}, TableRowIdPair::compare);
									
									if (max == null || parametersForRun.getSecond() > max) {
										max = parametersForRun.getSecond();
										insideParameters = parametersForRun.getFirst();
										maxRandom = r * negativeMultipliyer;
									}

								}
								 
								
								System.out.println("Random X Positive for this class: " + maxRandom);
								parameters = insideParameters;
								random = maxRandom;

							} else {

								// parameters =
								// hyper.findOptimumParamters(classId.getFinalPart().hashCode(),
								// rowIdPairs,
								// (fold, trainingData, testingData, pp) -> {
								//
								// LearnedModelRepository learnedModel =
								// runner.learnFromParamter(classId,
								// entry.getValue().getC(), parameterClusterer,
								// pp,
								// trainingData);
								// ClusterSimilaritiesImplNormalized csFold =
								// runner.getNormalizedCS(classId,
								// entry.getValue().getC(), learnedModel);
								// return runner.getPerformance(csFold,
								// testingData);
								// }, TableRowIdPair::compare);

							}
						} else {

								random = 1;

								parameters = new String[12];

							parameters[0] = "-U";
							parameters[1] = "-K";
								parameters[2] = String.valueOf(2);
							parameters[3] = "-V";
								parameters[4] = "1e-" + 3;
							parameters[5] = "-P";
								parameters[6] = String.valueOf(66);
							parameters[7] = "-I";
								parameters[8] = "2000";
							parameters[9] = "-num-slots";
							parameters[10] = "80";
								parameters[11] = "-attribute-importance";
						}


						Map<TableRowIdPair, PairLearningData> rowIdPairs = new HashMap<>(fullTrainingMap.getFirst());

							for (int i = 1; i <= random; i++) {
								rowIdPairs.putAll(fullTrainingMap.getSecond().get(i - 1));
						}

						System.out.println(Arrays.toString(parameters));


							int positivePairsSize = 0;
							int negativePairsSize = 0;

							for (PairLearningData v : rowIdPairs.values()) {
								if (v.isMatch()) {
									positivePairsSize++;
								} else {
									negativePairsSize++;
								}
							}

							if (sim instanceof MR_ClusterSimilarities) {
								System.out.print("For Class: " + classId.getFinalPart() + ", ForMethod:" + entry.getKey() + " Positive Pairs "
										+ positivePairsSize + ", Negative Pairs: "
										+ negativePairsSize + ", positive rules fired: " + ((MR_ClusterSimilarities) sim).getPositivePairsSize()
										+ ", negative rules fired: " + ((MR_ClusterSimilarities) sim).getNegativePairsSize());
							} else {
								System.out.print("For Class: " + classId.getFinalPart() + 
										"ForMethod:" + entry.getKey() + " Positive Pairs " + positivePairsSize + ", Negative Pairs: " + negativePairsSize);
							}

							rfLearnedModel = runner.learnFromParamter(classId, entry.getValue().getC(), parameterClusterer, parameters.clone(), rowIdPairs);

					} else {
						Map<TableRowIdPair, PairLearningData> rowIdPairs = new HashMap<>(fullTrainingMap.getFirst());

						for (int i = 1; i < negativeRuns; i++) {
								rowIdPairs.putAll(fullTrainingMap.getSecond().get(i - 1));
						}

						rfLearnedModel = runner.learn(classId, entry.getValue().getC(), rowIdPairs, new HashSet<>());
					}


					rowClustering.put(classId, rfLearnedModel);
					}
				}

				if (!rowClustering.isEmpty()) {
					ret.put(entry.getKey(), new DS_Container(rowClustering, Lists.newArrayList(entry.getValue().getC()), null));
				}

				if (!rowClusteringNew.isEmpty()) {
					for (Entry<String, Map<ClassId, LearnedModelRepository>> rcEntry : rowClusteringNew.entrySet()) {
						ret.put(rcEntry.getKey(), new DS_Container(rcEntry.getValue(), Lists.newArrayList(entry.getValue().getC()), null));
					}
				}

			}

			//
			// container.put("KB", new DS_Container(retRC_KB,
			// retND_WITHOUT_WEIGHTS));
			// container.put("BS_WITHOUT_WEIGHTS", new
			// DS_Container(retRC_WITHOUT_WEIGHTS, retND_WITHOUT_WEIGHTS));
			//
			// container.put("BS_WITH_WEIGHTS", new
			// DS_Container(retRC_WITH_WEIGHTS, retND_WITHOUT_WEIGHTS));
			return ret;

			// RowClusterSimilarities sim = runner.getSim(classId, strategy,
			// null);
			//
			// TableRowIdsSet rowIdsSet = input.getFullRowIdsSet(classId);
			// Map<TableRowIdPair, Double> positivePairsWithScores = new
			// HashMap<>();
			//
			// input.getFullRowIdsSet(classId);
			// for (RowCluster rc : clusters.getClusterList()) {
			// if (rc.getRows().size() > 1) {
			// long resultingSize =
			// DoubleIteraterable.getResultingSize(rc.getRows().size());
			// PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(rc.getRows()),
			// (doubleEntry) -> {
			// List<TableRowId> rows = new ArrayList<>();
			// rows.add(doubleEntry.getFirst());
			// rows.add(doubleEntry.getSecond());
			// double score = sim.get(rows, classId, new ClusterEntryKey(0, 1));
			// // double score = sim.get(rows, classId, new
			// // ClusterEntryKey(0, 1));
			// positivePairsWithScores.put(new
			// TableRowIdPair(doubleEntry.getFirst(), doubleEntry.getSecond()),
			// score);
			// }, resultingSize, this.getClass(), rc.getRows().size() < 600L);
			// }
			//
			// }
			//
			// SortedSet<TableRowIdPairScorable> sortedSet = new TreeSet<>();
			// for (Entry<TableRowIdPair, Double> entry :
			// positivePairsWithScores.entrySet()) {
			// if (entry.getValue() > 0.0) {
			// sortedSet.add(new TableRowIdPairScorable(entry.getValue(),
			// entry.getKey()));
			// }
			// }
			//
			// OtsuThresholderRounded thresholder = new
			// OtsuThresholderRounded();
			//
			// List<TableRowIdPairScorable> positivePairs =
			// thresholder.thresholdDefault(new ArrayList<>(sortedSet));
			// positivePairs = sortedSet.stream().filter((e) -> e.getScore() >
			// 0.5).collect(Collectors.toList());
			//
			// Set<TableRowIdPair> negativePairs = new HashSet<>();
			//
			// List<TableRowId> sortedList = new
			// ArrayList<>(rowIdsSet.getRowIds());
			// Collections.sort(sortedList, TableRowId::compare);
			//
			// Random rnd = new Random(classId.getFinalPart().hashCode());
			//
			// for (int i = 0; i < 1000; i++) {
			// int a = rnd.nextInt(sortedList.size());
			// int b = rnd.nextInt(sortedList.size());
			// TableRowIdPair rowIdPair = new TableRowIdPair(sortedList.get(a),
			// sortedList.get(b));
			// if (!positivePairsWithScores.containsKey(rowIdPair)) {
			// negativePairs.add(rowIdPair);
			// }
			// }
			//
			// Map<TableRowIdPair, PairLearningData> rowIdPairs = new
			// HashMap<>();
			// Map<TableRowIdPair, PairLearningData> rowIdPairsWithWeights = new
			// HashMap<>();
			// for (TableRowIdPairScorable pair : positivePairs) {
			// rowIdPairs.put(pair.getPair(), new PairLearningData(true, 1.0));
			// rowIdPairsWithWeights.put(pair.getPair(), new
			// PairLearningData(true, pair.getScore()));
			// }
			//
			// for (TableRowIdPair pair : negativePairs) {
			// rowIdPairs.put(pair, new PairLearningData(false, 1.0));
			// List<TableRowId> rows = new ArrayList<>();
			// rows.add(pair.getId1());
			// rows.add(pair.getId2());
			// rowIdPairsWithWeights.put(pair, new PairLearningData(false,
			// 1.0));
			// }
			//
			// Set<InstanceId> newEntities = new HashSet<>();
			//
			//
			// if (true) {
			// retRC_WITH_WEIGHTS.put(classId, runner.learn(classId, strategy,
			// rowIdPairsWithWeights, newEntities));
			// retRC_WITHOUT_WEIGHTS.put(classId, runner.learn(classId,
			// strategy, rowIdPairs, newEntities));
			// }
			//
			// {
			//
			// EntityMatcherWeightedResults matches =
			// runner.runMatching(classId, clusters.getClusterList(),
			// input.getEntityMatchingRuns().get(0),
			// ndStrategies.get(classId));
			//
			// SortedSet<RowClusterUUIDScorable> positiveWithScores = new
			// TreeSet<>();
			// for (List<ScoredEntity> set :
			// matches.getExistingEntities().values()) {
			// for (ScoredEntity se : set) {
			// positiveWithScores.add(new RowClusterUUIDScorable(se.getScore(),
			// se.getEntity().getRowClusterUUID()));
			// }
			//
			// }
			//
			// Set<RowClusterUUID> keepPositive =
			// thresholder.thresholdDefault(new
			// ArrayList<>(positiveWithScores)).stream()
			// .map((e) -> e.getUUID())
			// .collect(Collectors.toSet());
			//
			// // .filter((e) -> e.getScore() > 0.75)
			// keepPositive = positiveWithScores.stream().map((e) ->
			// e.getUUID()).collect(Collectors.toSet());
			//
			// SortedSet<RowClusterUUIDScorable> negativeWithScores = new
			// TreeSet<>();
			// for (ScoredEntity se : matches.getNewEntities()) {
			//
			// negativeWithScores.add(new RowClusterUUIDScorable(se.getScore(),
			// se.getEntity().getRowClusterUUID()));
			//
			// }
			//
			// Set<RowClusterUUID> keepNegative =
			// thresholder.thresholdDefault(new
			// ArrayList<>(negativeWithScores)).stream()
			// .map((e) -> e.getUUID())
			// .collect(Collectors.toSet());
			//
			// // .filter((e) -> e.getScore() < -0.75)
			// keepNegative = negativeWithScores.stream().map((e) ->
			// e.getUUID()).collect(Collectors.toSet());
			//
			// Map<TableRowIdPair, PairLearningData> rowIdPairs_EM = new
			// HashMap<>();
			//
			// for (List<ScoredEntity> set :
			// matches.getExistingEntities().values()) {
			// Set<TableRowId> rowIds = new HashSet<>();
			// for (ScoredEntity se : set) {
			// if (keepPositive.contains(se.getEntity().getRowClusterUUID())) {
			// rowIds.addAll(se.getEntity().getRowCluster().getRows());
			// }
			// }
			// if (rowIds.size() > 1) {
			// for (DoubleEntry<TableRowId> doubleEntry :
			// DoubleIteraterable.iterable(rowIds)) {
			// rowIdPairs_EM.put(new TableRowIdPair(doubleEntry.getFirst(),
			// doubleEntry.getSecond()), new PairLearningData(true, 1.0));
			// }
			// }
			// }
			//
			// for (ScoredEntity se : matches.getNewEntities()) {
			//
			// if (keepNegative.contains(se.getEntity().getRowClusterUUID()) &&
			// se.getEntity().getRowCluster().getRows().size() > 1) {
			// for (DoubleEntry<TableRowId> doubleEntry :
			// DoubleIteraterable.iterable(se.getEntity().getRowCluster().getRows()))
			// {
			// rowIdPairs_EM.put(new TableRowIdPair(doubleEntry.getFirst(),
			// doubleEntry.getSecond()), new PairLearningData(true, 1.0));
			// }
			// }
			//
			// }
			//
			// List<RowCluster> singleClusters = new ArrayList<>();
			// for (TableRowId rowId : rowIdsSet.getRowIds()) {
			// singleClusters.add(new RowCluster(classId,
			// Lists.newArrayList(rowId), sim));
			// }
			// EntityMatcherWeightedResults singleMatches =
			// runner.runMatching(classId, singleClusters,
			// input.getEntityMatchingRuns().get(0),
			// ndStrategies.get(classId));
			//
			// Map<TableRowId, Optional<InstanceId>> normalized = new
			// HashMap<>();
			// for (ScoredEntity se : singleMatches.getNewEntities()) {
			// for (TableRowId rowId : se.getEntity().getRowCluster().getRows())
			// {
			// normalized.put(rowId, Optional.empty());
			// }
			// }
			//
			// for (Entry<InstanceId, List<ScoredEntity>> entry :
			// singleMatches.getExistingEntities().entrySet()) {
			// for (ScoredEntity se : entry.getValue()) {
			// for (TableRowId rowId : se.getEntity().getRowCluster().getRows())
			// {
			// normalized.put(rowId, Optional.of(entry.getKey()));
			// }
			// }
			// }
			//
			// Set<TableRowIdPair> negativeFromMatch = new HashSet<>();
			//
			// for (TableRowIdPair pair : positivePairsWithScores.keySet()) {
			// Optional<InstanceId> n1 = normalized.get(pair.getId1());
			// Optional<InstanceId> n2 = normalized.get(pair.getId2());
			// if (n1 != null && n2 != null) {
			// if (n1.isPresent() && !n2.isPresent()) {
			// negativeFromMatch.add(pair);
			// // rowIdPairs_EM.put(pair, new
			// // PairLearningData(false, 1.0));
			// } else if (!n1.isPresent() && n2.isPresent()) {
			// negativeFromMatch.add(pair);
			// // rowIdPairs_EM.put(pair, new
			// // PairLearningData(false, 1.0));
			// } else if (n1.isPresent() && n2.isPresent()) {
			// if (!n1.get().equals(n2.get())) {
			// negativeFromMatch.add(pair);
			// // rowIdPairs_EM.put(pair, new
			// // PairLearningData(false, 1.0));
			// }
			// }
			// }
			// }
			//
			// Set<TableRowIdPair> negativeFromHomonym = new HashSet<>();
			//
			// ClusterList homonymClusters = runner.getHomonymClusters(classId,
			// rowIdsSet.getRowIds());
			//
			// for (RowCluster rc : homonymClusters.getClusterList()) {
			// if (rc.getRows().size() > 1) {
			// for (DoubleEntry<TableRowId> doubleEntry :
			// DoubleIteraterable.iterable(rc.getRows())) {
			// TableRowIdPair pair = new TableRowIdPair(doubleEntry.getFirst(),
			// doubleEntry.getSecond());
			//
			// if (!positivePairs.contains(pair)) {
			// negativeFromHomonym.add(pair);
			// }
			//
			// }
			// }
			// }
			//
			// double negativePart = 0.3 * (negativePairs.size() +
			// negativeFromHomonym.size() + negativeFromMatch.size());
			// double matchUpSample = negativePart / negativeFromMatch.size();
			//
			// double homonymUpsample = negativePart /
			// negativeFromHomonym.size();
			//
			// double randomUpsample = negativePart / negativePairs.size();
			//
			// for (TableRowIdPair pair : negativeFromMatch) {
			// rowIdPairs_EM.put(pair, new PairLearningData(false,
			// matchUpSample));
			// }
			//
			// for (TableRowIdPair pair : negativeFromHomonym) {
			// rowIdPairs_EM.put(pair, new PairLearningData(false,
			// homonymUpsample));
			// }
			//
			// for (TableRowIdPair pair : negativePairs) {
			// rowIdPairs_EM.put(pair, new PairLearningData(false,
			// randomUpsample));
			// }
			//
			//
			// System.out.println("matchUpSample: " + matchUpSample);
			// System.out.println("homonymUpsample: " + homonymUpsample);
			// System.out.println("randomUpsample: " + randomUpsample);
			//
			// retRC_KB.put(classId, runner.learn(classId, strategy,
			// rowIdPairs_EM, newEntities));
			// }
			//
			//
			// if (false) {
			//
			// } else {
			// // retND_WITH_WEIGHTS.put(classId,
			// // ndStrategies.get(classId));
			// retND_WITHOUT_WEIGHTS.put(classId, ndStrategies.get(classId));
			// }
			//
			//
			//
			// }
			//
			// Map<String, DS_Container> container = new HashMap<>();
			//
			// container.put("KB", new DS_Container(retRC_KB,
			// retND_WITHOUT_WEIGHTS));
			// container.put("BS_WITHOUT_WEIGHTS", new
			// DS_Container(retRC_WITHOUT_WEIGHTS, retND_WITHOUT_WEIGHTS));
			//
			// container.put("BS_WITH_WEIGHTS", new
			// DS_Container(retRC_WITH_WEIGHTS, retND_WITHOUT_WEIGHTS));
			// return container;

			// } catch (Exception e) {
			// throw new RuntimeException(e);
			// }
		}

		private Map<Integer, Pair<Map<TableRowIdPair, PairLearningData>, List<Map<TableRowIdPair, PairLearningData>>>> getTrainingMap(ClassId classId,
				ClusteringRunner runner, SetExpansionStrategy strategy, LearnedModelRepository models, RowClusterSimilarities sim,
				ClusterSimilaritiesImplNormalized cs, Map<Integer, Set<TableRowId>> foldInput, int negativeMultipliyer, int negativeRuns) {

			Map<Integer, Pair<Map<TableRowIdPair, PairLearningData>, List<Map<TableRowIdPair, PairLearningData>>>> ret = new HashMap<>();

			for (int fold = 0; fold < foldInput.size(); fold++) {

				Set<TableRowId> rowIds = new HashSet<>();
				for (int innerFold = 0; innerFold < foldInput.size(); innerFold++) {
					if (innerFold != fold) {
						rowIds.addAll(foldInput.get(innerFold));
					}
				}

				Pair<Map<TableRowIdPair, PairLearningData>, List<Map<TableRowIdPair, PairLearningData>>> trainingMap = getTrainingMap(classId, rowIds, runner,
						strategy, models, sim, cs, negativeMultipliyer, negativeRuns, null, null);

				ret.put(fold, trainingMap);

			}

			return ret;

		}

		public double SCORE_OF_CONFLICTING_PAIR = 1.0;
		public double THRESHOLD_POSITIVE = 0.0;
		public double THRESHOLD_NEGATIVE = 0.0;
		public boolean INCLUDE_CONFLICTING_PAIR = false;
		private final boolean ALWAYS_SCORE_ONE = false;
		private final boolean ROUND = false;
		private final boolean ROUND_2 = true;

		private Pair<Map<TableRowIdPair, PairLearningData>, List<Map<TableRowIdPair, PairLearningData>>> getTrainingMap(ClassId classId, Set<TableRowId> rowIds,
				ClusteringRunner runner, SetExpansionStrategy strategy, LearnedModelRepository models, RowClusterSimilarities sim,
				ClusterSimilaritiesImplNormalized cs, int negativeMultipliyer, int negativeRuns, EntityMatchingRun dsNdRun,
				Map<ClassId, EM_LearnedModelRepository> dsNdModels) {

			Random rnd = new Random(classId.getFinalPart().hashCode());

			Set<TableRowIdPair> positivePairs = Collections.newSetFromMap(new ConcurrentHashMap<>());
			Set<TableRowIdPair> negativePairs = Collections.newSetFromMap(new ConcurrentHashMap<>());

			List<TableRowId> sortedList = new ArrayList<>(rowIds);
			Collections.sort(sortedList, TableRowId::compare);

			final Set<TableRowId> conflictingDS;
			ClusterList clusters = runner.getList(classId, strategy, models, rowIds);
			
			if (dsNdRun != null) {
				Pair<ClusterList, Set<TableRowId>> dsRun = applyDistantSupervision(runner, classId, sim, rowIds, clusters, dsNdRun, dsNdModels);
				clusters = dsRun.getFirst();
				conflictingDS = dsRun.getSecond();
			} else {
				conflictingDS = new HashSet<>();
			}

			for (RowCluster rc : clusters.getClusterList()) {
				if (rc.getRows().size() > 1) {
					long resultingSize = DoubleIteraterable.getResultingSize(rc.getRows().size());
					PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(rc.getRows()), (doubleEntry) -> {
						positivePairs.add(new TableRowIdPair(doubleEntry.getFirst(), doubleEntry.getSecond()));
					}, resultingSize, this.getClass(), rc.getRows().size() < 600L);
				}

			}

			Map<TableRowId, Integer> mapping = new HashMap<>();
			Map<Integer, TableRowId> mappingIndex = new HashMap<>();

			for (int i = 0; i < sortedList.size(); i++) {
				mapping.put(sortedList.get(i), i);
				mappingIndex.put(i, sortedList.get(i));
			}

			Map<Integer, Set<Integer>> pairs = ClusterPairsExporter.getPairs(mapping, sim);
			List<TableRowIdPair> blockedSortedList = new ArrayList<>();

			for (Entry<Integer, Set<Integer>> pair : pairs.entrySet()) {
				for (Integer p2 : pair.getValue()) {
					blockedSortedList.add(new TableRowIdPair(mappingIndex.get(pair.getKey()), mappingIndex.get(p2)));
				}
			}



			ConcurrentHashMap<TableRowIdPair, PairLearningData> rowIdPairs = new ConcurrentHashMap<>();
			for (TableRowIdPair pair : positivePairs) {

				double score = cs.get(pair);

				if (score < THRESHOLD_POSITIVE) {
					score = SCORE_OF_CONFLICTING_PAIR;
					if (ADD_CONFLICTING_DS && (conflictingDS.contains(pair.getId1()) || conflictingDS.contains(pair.getId2()))) {
						rowIdPairs.put(pair, new PairLearningData(true, score));
					} else if (INCLUDE_CONFLICTING_PAIR) {
						rowIdPairs.put(pair, new PairLearningData(true, score));
					}
				} else {
					score = Math.abs(score);
					if (ALWAYS_SCORE_ONE) {
						score = 1.0;
						rowIdPairs.put(pair, new PairLearningData(true, 1.0));
					} else if (ROUND) {
						if (score >= 0.5) {
							rowIdPairs.put(pair, new PairLearningData(true, 1.0));
						}
					} else if (ROUND_2) {
						rowIdPairs.put(pair, new PairLearningData(true, score));
					}
				}

			}

			int posPairSize = rowIdPairs.size();

			Collections.sort(blockedSortedList, TableRowIdPair::compare);
			{
				int collect = posPairSize;
				;
				// int collect = positivePairs.size() * 8;
				// if (true || collect > blockedSortedList.size() -
				// positivePairs.size() - negativePairs.size()) {
				// System.out.println("Changed collect from " + collect
				// + " to "
				// + (blockedSortedList.size() - positivePairs.size() -
				// negativePairs.size()) + " classId: " +
				// classId.getFinalPart());
				// collect = blockedSortedList.size() -
				// positivePairs.size() - negativePairs.size();
				//
				// }
				Collections.shuffle(blockedSortedList, rnd);
				int added = 0;
				for (int i = 0; i < blockedSortedList.size(); i++) {
					TableRowIdPair rowIdPair = blockedSortedList.get(i);
					if (!positivePairs.contains(rowIdPair) && !negativePairs.contains(rowIdPair)) {
						negativePairs.add(rowIdPair);
						added++;
						// if (added >= collect) {
						// break;
						// }
					}
				}

			}

			Set<TableRowIdPair> addedNegativeIdPairs = new HashSet<>();

			List<Map<TableRowIdPair, PairLearningData>> additionalNegativePairs = new ArrayList<>();

			for (int f = 1; f < negativeRuns; f++) {

				int collect = negativeMultipliyer * posPairSize;
				if (f == 1) {
					collect = posPairSize;
				}
				System.out.println("For ClassId " + classId.getFinalPart() + " positive " + posPairSize + " homonym pairs " + negativePairs.size()
						+ " we collect " + collect);
				// if(collect < rowIdsSet.getRowIds().size() -
				// positivePairs.iterator()) {
				// collect = rowIdsSet.getRowIds().size() -
				// positivePairs.iterator();
				// }

				Set<TableRowIdPair> innerNegativePairs = new HashSet<>();

				ConcurrentHashMap<TableRowIdPair, PairLearningData> innerNegativePairMap = new ConcurrentHashMap<>();

				int i = 0;
				while (i < collect) {
					int a = rnd.nextInt(sortedList.size());
					int b = rnd.nextInt(sortedList.size());
					if (a != b) {
						TableRowIdPair rowIdPair = new TableRowIdPair(sortedList.get(a), sortedList.get(b));
						if (!positivePairs.contains(rowIdPair) && !negativePairs.contains(rowIdPair) && !addedNegativeIdPairs.contains(rowIdPair)) {
							addedNegativeIdPairs.add(rowIdPair);
							innerNegativePairs.add(rowIdPair);
							i++;
					}
					}
				}

				PersistentParallel6.runAndFinish(innerNegativePairs, (pair) -> {

					double score = cs.get(pair);

					if (score > THRESHOLD_NEGATIVE) {
						score =  SCORE_OF_CONFLICTING_PAIR;
						if (ADD_CONFLICTING_DS && (conflictingDS.contains(pair.getId1()) || conflictingDS.contains(pair.getId2()))) {
							innerNegativePairMap.put(pair, new PairLearningData(false, score));
						} else if (INCLUDE_CONFLICTING_PAIR) {
							innerNegativePairMap.put(pair, new PairLearningData(false, score));						
						}
					} else {
						score = Math.abs(score);
						if (ALWAYS_SCORE_ONE) {
							score = 1.0;
							innerNegativePairMap.put(pair, new PairLearningData(false, 1.0));
						}
						else if (ROUND) {
							if (score >= 0.5) {
								innerNegativePairMap.put(pair, new PairLearningData(false, 1.0));
							}

						} else if (ROUND_2) {
							innerNegativePairMap.put(pair, new PairLearningData(false, score));
						}

					}
				}, this.getClass());

				additionalNegativePairs.add(innerNegativePairMap);
			}



			PersistentParallel6.runAndFinish(negativePairs, (pair) -> {

				double score = cs.get(pair);

				if (score > THRESHOLD_NEGATIVE) {
					score =  SCORE_OF_CONFLICTING_PAIR;
					if (ADD_CONFLICTING_DS && (conflictingDS.contains(pair.getId1()) || conflictingDS.contains(pair.getId2()))) {
						rowIdPairs.put(pair, new PairLearningData(false, score));
					} else if (INCLUDE_CONFLICTING_PAIR) {
						rowIdPairs.put(pair, new PairLearningData(false, score));
					}
				} else {
					score = Math.abs(score);
					if (ALWAYS_SCORE_ONE) {
						rowIdPairs.put(pair, new PairLearningData(false, 1.0));
					}
					else if (ROUND) {
						if (score >= 0.5) {
							rowIdPairs.put(pair, new PairLearningData(false, 1.0));
						}
					} else if (ROUND_2) {
						rowIdPairs.put(pair, new PairLearningData(false, score));
					}

				}
			}, this.getClass());

			int homonymPairs = rowIdPairs.size() - posPairSize;
			System.out.print("Positive Pairs actually: " + posPairSize + ", Homonym Pairs actually: " + homonymPairs);

			for (Map<TableRowIdPair, PairLearningData> entry : additionalNegativePairs) {
				System.out.print(", Random Set: " + entry.size());
			}

			System.out.println();

			return new Pair<>(rowIdPairs, additionalNegativePairs);
		}

		private boolean ADD_CONFLICTING_DS = false;

		private Pair<ClusterList, Set<TableRowId>> applyDistantSupervision(ClusteringRunner runner, ClassId classId, RowClusterSimilarities sim,
				Set<TableRowId> rowIds, ClusterList clusters, EntityMatchingRun dsNdRun, Map<ClassId, EM_LearnedModelRepository> dsNdModels) {
			Set<TableRowId> conflictingDS = new HashSet<>();
			List<RowCluster> singleClusters = new ArrayList<>();
			for (TableRowId rowId : rowIds) {
				singleClusters.add(new RowCluster(classId, Lists.newArrayList(rowId), sim));
			}

			EntityMatcherWeightedResults singleMatches = runner.runMatching(classId, singleClusters, dsNdRun, dsNdModels.get(classId));

			Map<TableRowId, Optional<InstanceId>> normalized = new HashMap<>();
			for (ScoredEntity se : singleMatches.getNewEntities()) {
				for (TableRowId rowId : se.getEntity().getRowCluster().getRows()) {
					normalized.put(rowId, Optional.empty());
				}
			}

			for (Entry<InstanceId, List<ScoredEntity>> entry : singleMatches.getExistingEntities().entrySet()) {
				for (ScoredEntity se : entry.getValue()) {
					for (TableRowId rowId : se.getEntity().getRowCluster().getRows()) {
						normalized.put(rowId, Optional.of(entry.getKey()));
					}
				}
			}

			EntityMatcherWeightedResults matches = runner.runMatching(classId, clusters.getClusterList(), dsNdRun, dsNdModels.get(classId));

			Map<RowClusterUUID, Optional<InstanceId>> mappingClusters = new HashMap<>();

			for (Entry<InstanceId, List<Entity>> existing : matches.getExistingEntitiesUnscored().entrySet()) {

				for (Entity inner : existing.getValue()) {
					mappingClusters.put(inner.getRowClusterUUID(), Optional.of(existing.getKey()));
				}
			}

			for (ScoredEntity newEntity : matches.getNewEntities()) {
				mappingClusters.put(newEntity.getEntity().getRowClusterUUID(), Optional.empty());
			}

			// Map<TableRowId, Optional<InstanceId>> newMappingForRows = new
			// HashMap<>();

			Map<InstanceId, List<TableRowId>> newExistingClusters = new HashMap<>();
			Map<InstanceId, MutableInt> newExistingClustersCount = new HashMap<>();


			List<RowCluster> clusterList = new ArrayList<>();

			for (RowCluster rc : clusters.getClusterList()) {

				Optional<InstanceId> match = mappingClusters.get(rc.getUuid());
				List<TableRowId> thisRowIds = new ArrayList<>();
				for (TableRowId rowId : rc.getRows()) {
					boolean added = false;
					if (normalized.containsKey(rowId)) {
						Optional<InstanceId> rowMatch = normalized.get(rowId);
						if ((rowMatch.isPresent()) && (match == null || !match.isPresent() || !match.equals(rowMatch))) {
							CHMHelper.getFromMap(newExistingClusters, rowMatch.get(), () -> new ArrayList<>()).add(rowId);
							conflictingDS.add(rowId);
							added = true;
						}
					}
					if (!added) {
						thisRowIds.add(rowId);
					}
				}

				if (match != null && match.isPresent()) {

					CHMHelper.getFromMap(newExistingClusters, match.get(), () -> new ArrayList<>()).addAll(thisRowIds);
					CHMHelper.getFromMap(newExistingClustersCount, match.get(), () -> new MutableInt()).increment();

				} else {
					clusterList.add(new RowCluster(classId, thisRowIds, sim));
				}

			}

			for (Entry<InstanceId, List<TableRowId>> entry : newExistingClusters.entrySet()) {
				if (CHMHelper.getFromMap(newExistingClustersCount, entry.getKey(), () -> new MutableInt()).intValue() > 0) {
					for (TableRowId rowId : entry.getValue()) {
						conflictingDS.add(rowId);
					}
				}
				clusterList.add(new RowCluster(classId, entry.getValue(), sim));

			}

			return new Pair<>(new ClusterList(clusterList), conflictingDS);
		}

		private Map<Integer, Pair<Map<TableRowId, InstanceId>, Map<InstanceId, Set<TableRowId>>>> getTestingMap( 
				ClassId classId, 
				ClusteringRunner runner, SetExpansionStrategy strategy, Map<Integer, Set<TableRowId>> foldInput, LearnedModelRepository models) {
			
			Map<Integer, Pair<Map<TableRowId, InstanceId>, Map<InstanceId, Set<TableRowId>>>> ret = new HashMap<>();
			for (Entry<Integer, Set<TableRowId>> splitEntry : foldInput.entrySet()) {
				
				
				ClusterList clusters = runner.getList(classId, strategy, models, splitEntry.getValue());
				 

				Map<TableRowId, InstanceId> eIndex = new HashMap<>();
				Map<InstanceId, Set<TableRowId>> eGS = new HashMap<>();

				for(RowCluster rc : clusters.getClusterList()) {
					InstanceId instanceId = new InstanceId(rc.getUuid().toString(), rc.getUuid().toString());

					for (TableRowId rowId : rc.getRows()) {
							eIndex.put(rowId, instanceId);
					}
				 
					eGS.put(instanceId, new HashSet<>(rc.getRows()));

				}
					

				ret.put(splitEntry.getKey(), new Pair<>(eIndex, eGS));
			}

			return ret;

		}



		private TreeMap<Integer, Tri<Set<RowClusterUUID>, Set<TableRowId>, Map<TableRowIdPair, PairLearningData>>> split(ClassId classId, int folds,
				ConcurrentHashMap<Integer, Set<RowClusterUUID>> clusterMap, Map<RowClusterUUID, Set<TableRowId>> clusterIndex,
				ConcurrentHashMap<TableRowIdPair, PairLearningData> rowIdPairsWithScores, ClusterSimilaritiesImplNormalized cs,
				List<TableRowIdPair> blockedSortedList) {

			// int totalClusters = clusterIndex.size();

			Map<Integer, Integer> sizesPerSet = clusterMap.entrySet().stream().collect(Collectors.toMap((e) -> e.getKey(), (e) -> e.getValue().size()));
			LinkedHashMap<Integer, Integer> sorted = MapSorter.sortHashMapByValuesD(sizesPerSet);

			// double perFold = totalClusters / ((double) folds);

			Map<Integer, Set<RowClusterUUID>> foldMap = new HashMap<>();

			{
				int i = 0;
				for (Entry<Integer, Integer> s : sorted.entrySet()) {

					CHMHelper.getFromMap(foldMap, i, () -> new HashSet<>()).addAll(clusterMap.get(s.getKey()));

					i++;
					if (i < folds) {
						i = 0;
					}
				}
			}

			Map<Integer, Set<TableRowId>> foldRowIdMap = new HashMap<>();

			for (Entry<Integer, Set<RowClusterUUID>> entry : foldMap.entrySet()) {
				Set<TableRowId> rowIds = new HashSet<>();
				for (RowClusterUUID uuid : entry.getValue()) {
					rowIds.addAll(clusterIndex.get(uuid));
				}
				foldRowIdMap.put(entry.getKey(), rowIds);
			}

			Random rnd = new Random(classId.getFinalPart().hashCode());

			Map<Integer, Map<TableRowIdPair, PairLearningData>> pairDataMap = new HashMap<>();

			for (Entry<Integer, Set<RowClusterUUID>> entry : foldMap.entrySet()) {

				// ConcurrentHashMap<TableRowIdPair, PairLearningData> pairData
				// = new ConcurrentHashMap<>();

				Set<TableRowId> allRowIds = foldRowIdMap.get(entry.getKey());

				Set<TableRowIdPair> positivePairs = Collections.newSetFromMap(new ConcurrentHashMap<>());
				Set<TableRowIdPair> negativePairs = Collections.newSetFromMap(new ConcurrentHashMap<>());

				List<TableRowId> sortedList = new ArrayList<>(allRowIds);
				Collections.sort(sortedList, TableRowId::compare);

				for (RowClusterUUID uuid : entry.getValue()) {
					Set<TableRowId> rowIds = clusterIndex.get(uuid);
					if (rowIds.size() > 1) {
						long resultingSize = DoubleIteraterable.getResultingSize(rowIds.size());
						PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(rowIds), (doubleEntry) -> {
							positivePairs.add(new TableRowIdPair(doubleEntry.getFirst(), doubleEntry.getSecond()));
						}, resultingSize, this.getClass(), rowIds.size() < 600L);
					}

				}

				for (int i = 0; i < blockedSortedList.size(); i++) {
					TableRowIdPair rowIdPair = blockedSortedList.get(i);
					if (!positivePairs.contains(rowIdPair)) {
						negativePairs.add(rowIdPair);
					}
				}

				if (true) {
					int collect = 4 * positivePairs.size();
					System.out.println("For ClassId " + classId.getFinalPart() + " positive " + positivePairs.size() + " homonym pairs " + negativePairs.size()
							+ " we collect " + collect);
					// if(collect < rowIdsSet.getRowIds().size() -
					// positivePairs.iterator()) {
					// collect = rowIdsSet.getRowIds().size() -
					// positivePairs.iterator();
					// }
					int i = 0;
					while (i < collect) {
						int a = rnd.nextInt(sortedList.size());
						int b = rnd.nextInt(sortedList.size());
						if (a != b) {
							TableRowIdPair rowIdPair = new TableRowIdPair(sortedList.get(a), sortedList.get(b));
							if (!positivePairs.contains(rowIdPair) && !negativePairs.contains(rowIdPair)) {
								negativePairs.add(rowIdPair);
								i++;
							}
						}
					}
				}

				ConcurrentHashMap<TableRowIdPair, PairLearningData> rowIdPairs = new ConcurrentHashMap<>();
				for (TableRowIdPair pair : positivePairs) {

					if (rowIdPairsWithScores.containsKey(pair)) {
						rowIdPairs.put(pair, rowIdPairsWithScores.get(pair));
					} else {

						double score = cs.get(pair);

						if (score < 0.0) {
							score = SCORE_OF_CONFLICTING_PAIR;
							if (INCLUDE_CONFLICTING_PAIR) {
								rowIdPairs.put(pair, new PairLearningData(true, score));
							}
						} else {
							score = Math.abs(score);
							// score = 1.0;
							rowIdPairs.put(pair, new PairLearningData(true, score));
						}
					}

				}

				PersistentParallel6.runAndFinish(negativePairs, (pair) -> {
					if (rowIdPairsWithScores.containsKey(pair)) {
						rowIdPairs.put(pair, rowIdPairsWithScores.get(pair));
					} else {
						double score = cs.get(pair);

						if (score > 0.0) {
							score = 1.0;
						} else {
							score = Math.abs(score);
							// score = 1.0;
							rowIdPairs.put(pair, new PairLearningData(false, score));
						}
					}
				}, this.getClass());

				pairDataMap.put(entry.getKey(), rowIdPairs);

			}

			TreeMap<Integer, Tri<Set<RowClusterUUID>, Set<TableRowId>, Map<TableRowIdPair, PairLearningData>>> ret = new TreeMap<>();

			for (Entry<Integer, Set<RowClusterUUID>> entry : foldMap.entrySet()) {

				ret.put(entry.getKey(), new Tri<>(entry.getValue(), foldRowIdMap.get(entry.getKey()), pairDataMap.get(entry.getKey())));
			}

			return ret;
		}

	// final String[] parameters;
	//
	// ParamterClusterer parameterClusterer =
	// runner.getParamterClusterer(classId, entry.getValue().getC(), rowIdPairs,
	// new HashSet<>());
	//
	//
	//
	// if (true) {
	//
	//
	//
	// if (true) {
	// ClusterCombiner<TableRowId> cc = new ClusterCombiner<>();
	//
	// Set<TableRowId> rowIdsRemaining = Collections.newSetFromMap(new
	// ConcurrentHashMap<>());
	// rowIdsRemaining.addAll(rowIds);
	//
	// ConcurrentLinkedQueue<Set<TableRowId>> ccList = new
	// ConcurrentLinkedQueue<>();
	// PersistentParallel6.runAndFinish(blockedSortedList, (pair) -> {
	// Set<TableRowId> set = new HashSet<>();
	// set.add(pair.getId1());
	// set.add(pair.getId2());
	// ccList.add(set);
	// rowIdsRemaining.remove(pair.getId1());
	// rowIdsRemaining.remove(pair.getId2());
	// }, this.getClass());
	//
	// for (TableRowId rowId : rowIdsRemaining) {
	// ccList.add(Sets.newHashSet(rowId));
	// }
	// cc.add(ccList);
	//
	// Collection<Set<TableRowId>> clusterIngOfRowIds = cc.get();
	//
	// Map<TableRowId, Integer> index = new HashMap<>();
	//
	// int i = 0;
	// for (Set<TableRowId> rowIdsIterator : clusterIngOfRowIds) {
	// for (TableRowId rowId : rowIdsIterator) {
	// index.put(rowId, i);
	// }
	// i++;
	// }
	//
	// ConcurrentHashMap<Integer, Set<RowClusterUUID>> clusterList = new
	// ConcurrentHashMap<>();
	//
	// PersistentParallel6.runAndFinish(clusters.getClusterList(), (cluster) ->
	// {
	// Integer indexEntry = null;
	// for (TableRowId rowId : cluster.getRows()) {
	// Integer ii = index.get(rowId);
	// if (indexEntry == null) {
	// indexEntry = ii;
	// } else if (!indexEntry.equals(ii)) {
	// throw new IllegalStateException();
	// }
	// }
	//
	// CHMHelper.getFromCHM(clusterList, indexEntry, () ->
	// Collections.newSetFromMap(new ConcurrentHashMap<>()))
	// .add(cluster.getUuid());
	// }, this.getClass());
	//
	// Map<RowClusterUUID, Set<TableRowId>> clusterIndex =
	// clusters.getClusterList().stream()
	// .collect(Collectors.toMap((e) -> e.getUuid(), (e) -> new
	// HashSet<>(e.getRows())));
	//
	// TreeMap<Integer, Tri<Set<RowClusterUUID>, Set<TableRowId>,
	// Map<TableRowIdPair, PairLearningData>>> splitData = split(classId, 3,
	// clusterList,
	// clusterIndex, rowIdPairs,
	// cs, blockedSortedList);
	//
	//
	//
	// TreeMap<Integer, Map<TableRowIdPair, PairLearningData>> splitData2 = new
	// TreeMap<>();
	//
	// Map<Integer, Map<TableRowId, InstanceId>> evaluationIndex = new
	// HashMap<>();
	// Map<Integer, Map<InstanceId, Set<TableRowId>>> evaluationGS = new
	// HashMap<>();
	//
	// for (Entry<Integer, Tri<Set<RowClusterUUID>, Set<TableRowId>,
	// Map<TableRowIdPair, PairLearningData>>> splitEntry : splitData
	// .entrySet()) {
	// splitData2.put(splitEntry.getKey(), splitEntry.getValue().getC());
	//
	// Map<TableRowId, InstanceId> eIndex =
	// CHMHelper.getFromMap(evaluationIndex, splitEntry.getKey(), () -> new
	// HashMap<>());
	// Map<InstanceId, Set<TableRowId>> eGS = CHMHelper.getFromMap(evaluationGS,
	// splitEntry.getKey(), () -> new HashMap<>());
	//
	// for (RowClusterUUID ee : splitEntry.getValue().getA()) {
	// InstanceId instanceId = new InstanceId(ee.toString(), ee.toString());
	//
	// for (TableRowId rowId : clusterIndex.get(ee)) {
	// eIndex.put(rowId, instanceId);
	//
	// }
	// eGS.put(instanceId, clusterIndex.get(ee));
	//
	// }
	//
	// }
	//
	// parameters =
	// hyper.findOptimumParamtersInner(classId.getFinalPart().hashCode(),
	// splitData2,
	// (fold, trainingData, testingData, pp) -> {
	//
	//
	// LearnedModelRepository learnedModel = runner.learnFromParamter(classId,
	// entry.getValue().getC(), parameterClusterer, pp,
	// trainingData);
	//
	// ClusterList clusters2 = runner.getList(classId, entry.getValue().getC(),
	// learnedModel, splitData.get(fold).getB());
	// List<Set<TableRowId>> clusters2s =
	// clusters2.getClusterList().stream().map((e) -> new
	// HashSet<>(e.getRows()))
	// .collect(Collectors.toList());
	//
	// OktiePerformance clusterPerformance =
	// oktieClusterEvaluator.oktieEvaluate(evaluationGS.get(fold),
	// evaluationIndex.get(fold), clusters2s);
	//
	// return clusterPerformance.getFMeasureMacro();
	//
	//
	// }, TableRowIdPair::compare);
	//
	// } else {
	//
	// parameters =
	// hyper.findOptimumParamters(classId.getFinalPart().hashCode(), rowIdPairs,
	// (fold, trainingData, testingData, pp) -> {
	//
	// LearnedModelRepository learnedModel = runner.learnFromParamter(classId,
	// entry.getValue().getC(), parameterClusterer, pp,
	// trainingData);
	// ClusterSimilaritiesImplNormalized csFold =
	// runner.getNormalizedCS(classId, entry.getValue().getC(), learnedModel);
	// return runner.getPerformance(csFold, testingData);
	// }, TableRowIdPair::compare);
	//
	// }
	// } else {
	//
	// parameters = new String[11];
	//
	// parameters[0] = "-U";
	// parameters[1] = "-K";
	// parameters[2] = String.valueOf(0);
	// parameters[3] = "-V";
	// parameters[4] = "1e-" + 3;
	// parameters[5] = "-P";
	// parameters[6] = String.valueOf(85);
	// parameters[7] = "-I";
	// parameters[8] = "600";
	// parameters[9] = "-num-slots";
	// parameters[10] = "80";
	// }
	//
	//
	//
	// System.out.println(Arrays.toString(parameters));
	//
	// rfLearnedModel = runner.learnFromParamter(classId,
	// entry.getValue().getC(), parameterClusterer, parameters, rowIdPairs);

	}

	public static TreeMap<Integer, Pair<Set<AnnotatedTableId>, Set<TableRowId>>> splitTables(int folds, ClassId classIds, Set<AnnotatedTableId> tableIds,
			Set<TableRowId> rowIds) {

		Random random = new Random(classIds.getFinalPart().hashCode());
		List<AnnotatedTableId> tableList = new ArrayList<>(tableIds);
		Collections.sort(tableList, AnnotatedTableId::compare);

		Collections.shuffle(tableList, random);

		Map<Integer, Set<AnnotatedTableId>> tableIdsMap = new HashMap<>();
		Map<AnnotatedTableId, Integer> tableIndex = new HashMap<>();
		{
			int fold = 0;
			for (AnnotatedTableId tableId : tableList) {

				CHMHelper.getFromMap(tableIdsMap, fold, () -> new HashSet<>()).add(tableId);
				tableIndex.put(tableId, fold);

				fold++;
				if (fold == folds) {
					fold = 0;
				}
			}
		}

		Map<Integer, Set<TableRowId>> tableRowIdMap = new HashMap<>();

		for (TableRowId rowId : rowIds) {
			Integer fold = tableIndex.get(rowId.getTableId());
			CHMHelper.getFromMap(tableRowIdMap, fold, () -> new HashSet<>()).add(rowId);
		}

		TreeMap<Integer, Pair<Set<AnnotatedTableId>, Set<TableRowId>>> ret = new TreeMap<>();

		for (int i = 0; i < folds; i++) {
			ret.put(i, new Pair<>(tableIdsMap.get(i), tableRowIdMap.get(i)));
		}

		return ret;

	}
}
