package expansion.set.distant_supervision.matchingmethods;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import javax.inject.Inject;

import com.beust.jcommander.internal.Lists;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import expansion.set.distant_supervision.DS_Clustering_Strategy_With_Runner;
import expansion.set.distant_supervision.DS_Container;
import expansion.set.distant_supervision.DistantSupervisionInput;
import expansion.set.distant_supervision.matching_rules.MatchingRulesWrapperCreator;
import expansion.set.distant_supervision.runners.DS_ClusteringRunnerCreator.ClusteringRunner;
import expansion.set.entitymatching.modelLearning.EM_LearnedModelRepository;
import expansion.set.entitymatching.modelLearning.EntityMatchingModel;
import expansion.set.entitymatching.modelLearning.weka.EntityMatchingWekaModel;
import expansion.set.entitymatching.run.EntityMatchingRun;
import expansion.set.entitymatching.run.EntityMatchingRunRepository;

public class ND_MatchingRules_Creator {

	public WS_MR_ND_Strategy get(String name, EntityMatchingRun run, Map<ClassId, EM_LearnedModelRepository> learnedModels) {
		Map<String, Pair<EntityMatchingRun, Map<ClassId, EM_LearnedModelRepository>>> map = new HashMap<>();
		map.put(name, new Pair<>(run, learnedModels));

		return new WS_MR_ND_Strategy(map);
	}
	@Inject
	protected MatchingRulesWrapperCreator matchingRulesWrapperCreator;

	public class WS_MR_ND_Strategy implements DS_Clustering_Strategy_With_Runner {

		private final Map<String, Pair<EntityMatchingRun, Map<ClassId, EM_LearnedModelRepository>>> map;

		public WS_MR_ND_Strategy(Map<String, Pair<EntityMatchingRun, Map<ClassId, EM_LearnedModelRepository>>> map) {
			this.map = map;
		}

		@Override
		public Map<String, DS_Container> get(DistantSupervisionInput input, ClusteringRunner runner) {
			Map<String, DS_Container> ret = new HashMap<>();

			for (Entry<String, Pair<EntityMatchingRun, Map<ClassId, EM_LearnedModelRepository>>> entry : map.entrySet()) {

				Map<ClassId, EM_LearnedModelRepository> rowClustering = new HashMap<>();

				for (Entry<ClassId, EM_LearnedModelRepository> innerEntry : entry.getValue().getSecond().entrySet()) {

					Map<OptimizableClass, EntityMatchingModel> repMap = new HashMap<>();

					EntityMatchingWekaModel unAdaptedModel = (EntityMatchingWekaModel) innerEntry.getValue()
							.get(entry.getValue().getFirst().getOptimizableClass());

					repMap.put(EntityMatchingRunRepository.CONTAINER_ND_ALL.getOptimizableClass(),
							new EntityMatchingWekaModel(unAdaptedModel, matchingRulesWrapperCreator.getNdScoreGetter(innerEntry.getKey())));

					EM_LearnedModelRepository rep = new EM_LearnedModelRepository(repMap);
					rowClustering.put(innerEntry.getKey(), rep);
				}

				ret.put(entry.getKey(), new DS_Container(null, null, rowClustering, Lists.newArrayList(EntityMatchingRunRepository.CONTAINER_ND_ALL)));
			}

			return ret;
		}

	}
}
