package expansion.set.distant_supervision.matchingmethods;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import javax.inject.Inject;

import com.beust.jcommander.internal.Lists;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.LearnedModelRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.distant_supervision.DS_Clustering_Strategy_With_Runner;
import expansion.set.distant_supervision.DS_Container;
import expansion.set.distant_supervision.DistantSupervisionInput;
import expansion.set.distant_supervision.matching_rules.MatchingRulesWrapperCreator;
import expansion.set.distant_supervision.runners.DS_ClusteringRunnerCreator.ClusteringRunner;
import expansion.set.strategies.SetExpansionStrategiesRepository;
import expansion.set.strategies.SetExpansionStrategy;
import expansion.set.strategies.clustersplitters.minScoreSplitter.WekaLearnableRowClusterSimilarities;
import expansion.set.strategies.ea.ContainerLearnedModel;

public class RC_MatchingRules_Creator {
	@Inject
	protected MatchingRulesWrapperCreator matchingRulesWrapperCreator;

	@Inject
	public RC_MatchingRules_Creator() {

	}


	public RC_MatchingRules_Creator_Strategy get(String name, SetExpansionStrategy strategy, Map<ClassId, LearnedModelRepository> models) {
		Map<String, Pair<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>>> map = new HashMap<>();
		map.put(name, new Pair<>(strategy, models));

		return new RC_MatchingRules_Creator_Strategy(map);
	}

	// public RC_MatchingRules_Creator_Strategy get(Map<String,
	// Tri<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>,
	// SetExpansionStrategy>> models) {
	// return new RC_MatchingRules_Creator_Strategy(models);
	// }

	public class RC_MatchingRules_Creator_Strategy implements DS_Clustering_Strategy_With_Runner {
		private final Map<String, Pair<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>>> map;

		public RC_MatchingRules_Creator_Strategy(Map<String, Pair<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>>> map) {
			this.map = map;
		}

		@Override
		public Map<String, DS_Container> get(DistantSupervisionInput input, ClusteringRunner runner) {

			Map<String, DS_Container> ret = new HashMap<>();

			for (Entry<String, Pair<SetExpansionStrategy, Map<ClassId, LearnedModelRepository>>> entry : map.entrySet()) {

				Map<ClassId, LearnedModelRepository> rowClustering = new HashMap<>();

				for (Entry<ClassId, LearnedModelRepository> innerEntry : entry.getValue().getSecond().entrySet()) {

					Map<OptimizableClass, LearnedModel> repMap = new HashMap<>();

					RowClusterSimilarities sim = runner.getSim(innerEntry.getKey(), entry.getValue().getFirst(), innerEntry.getValue());

					WekaLearnableRowClusterSimilarities fullSim = (WekaLearnableRowClusterSimilarities) runner.getSim(innerEntry.getKey(),
							SetExpansionStrategiesRepository.COMBINED_RandomForest_ET, null);

					repMap.put(SetExpansionStrategiesRepository.Container_Clusterer.getOptimizableClassCluster(),
							new ContainerLearnedModel(matchingRulesWrapperCreator.getSim(sim, fullSim)));

					LearnedModelRepository rep = new LearnedModelRepository(repMap);
					rowClustering.put(innerEntry.getKey(), rep);
				}

				ret.put(entry.getKey(), new DS_Container(rowClustering, Lists.newArrayList(SetExpansionStrategiesRepository.Container_Clusterer), null));
			}

			return ret;
		}

	}
}
