package expansion.set.distant_supervision.runners;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import javax.inject.Inject;

import com.google.common.collect.Lists;
import com.google.common.util.concurrent.AtomicDouble;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdsSet;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.tablefusion.learning.RowValuesGetter;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;
import de.mannheim.uni.expansion.tablefusion.learning.SplitRowValues;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ImmuMaker;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationMethod;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdContainer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.learner.WeightsLearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.LearnedModelRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeightKey;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepositoryImpl;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusterer;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.clustering.model.ClusterSimilaritiesImplNormalized;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities.PairLearningData;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.creation.Entity;
import expansion.set.creation.EntityCreator;
import expansion.set.distant_supervision.DistantSupervisionInput;
import expansion.set.distant_supervision.baseline.average.ND_Average_Creator;
import expansion.set.distant_supervision.baseline.average.RC_Average_Creator;
import expansion.set.distant_supervision.matching_rules.MatchingRulesWrapperCreator.ND_MR_SCORE_ADAPTER;
import expansion.set.distant_supervision.runners.DS_PreparedRun.DS_PreparedRunBuilder;
import expansion.set.entitymatching.EntityMatcherFullResults;
import expansion.set.entitymatching.EntityMatcherWeightedResults;
import expansion.set.entitymatching.EntityMatchingModelLearner;
import expansion.set.entitymatching.EntityMatchingRunner;
import expansion.set.entitymatching.candidateselection.CandidateSelectionStrategies;
import expansion.set.entitymatching.matchers.EntityMatcherConfig;
import expansion.set.entitymatching.modelLearning.EM_LearnedModelRepository;
import expansion.set.entitymatching.modelLearning.EntityMatchingModel;
import expansion.set.entitymatching.modelLearning.EntityMatchingModelLearnerRepository;
import expansion.set.entitymatching.modelLearning.EntityMatchingModelLearnerRunner;
import expansion.set.entitymatching.modelLearning.weka.EntityMatchingWekaModel;
import expansion.set.entitymatching.run.EntityMatchingRun;
import expansion.set.evaluation.SetExpansionRunner;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.evaluation.crossvalidation.EntitySetGroupedValuesSplitter;
import expansion.set.evaluation.crossvalidation.EntitySets;
import expansion.set.evaluation.goldstandard.oktie.OktieClusterEvaluator.WeightGetter;
import expansion.set.learning.ClusteringModelLearnerRunner;
import expansion.set.strategies.SetExpansionStrategiesRepository;
import expansion.set.strategies.SetExpansionStrategy;
import expansion.set.strategies.ea.ParameterLearnableClusterer;
import expansion.set.strategies.ea.ParamterClusterer;
import expansion.set.strategies.rowclusterers.GreedyClusteringSplitter;
import expansion.set.strategies.rowclusterers.RowClustererRepository;
import expansion.set.strategies.rowclusterers.label.LabelRowSimilaritiesCreator;

public class DS_ClusteringRunnerCreator {

	@Inject
	protected LabelRowSimilaritiesCreator labelSimCreator;

	@Inject
	protected RowClustererRepository rowClustererRepository;

	@Inject
	protected ND_Average_Creator ndAverageCreator;

	@Inject
	protected RC_Average_Creator rcAverageCreator;

	@Inject
	protected EntityMatchingRunner entityMatchingRuner;

	@Inject
	protected SetExpansionRunner setExpansionRunner;

	@Inject
	protected GreedyClusteringSplitter greedyClusteringSplitter;

	@Inject
	protected AnnotatedTablesRepository tableRepository;

	@Inject
	protected RowValuesGetter rowValuesGetter;

	@Inject
	protected EntitySetGroupedValuesSplitter entitySetGroupedValuesSplitter;

	@Inject
	protected EntityCreator entityCreator;

	@Inject
	protected ClusteringModelLearnerRunner clusteringLearner;
	
	@Inject
	protected EntityMatchingModelLearner entityMatchingModelLearner;
	
	@Inject
	protected EntityMatchingModelLearnerRepository learnerRepository;

	@Inject
	public DS_ClusteringRunnerCreator() {

	}

	public ClusteringRunner getRunner(DistantSupervisionInput input) {
		try {
			Map<ClassId, DS_PreparedRun> preapredRuns = new HashMap<>();

			for (ClassId classId : input.getClassesWithParents().keySet()) {
				TableIdsSet tableIdsSet = input.getTableIdsSet(classId);



				SplitGroupedValues splitGroupedValues = setExpansionRunner.getSplitGroupedValues(input.getFusionConfig(), tableIdsSet);
				Map<ClassId, Set<InstanceId>> instancesWithMatches = setExpansionRunner.getInstancesWithMatches(classId, tableIdsSet.getTableIds(), 1);
				EntitySets entitySets = new EntitySets(instancesWithMatches.get(classId), instancesWithMatches.get(classId));

				SplitGroupedValuesSets splitGroupedValuesSets = entitySetGroupedValuesSplitter.split(splitGroupedValues, entitySets);

				SplitRowValues splitRowValues = rowValuesGetter.getSplitRowValues(tableRepository, tableIdsSet);

				List<ThresholdContainer> thresholdContainers = setExpansionRunner.getThresholdContainers(input.getSetExpansionStrategiesConfig(),
						input.getThresholdsFolder(), classId, input.getFusionConfig(), input.getStrategiesConfig(), splitGroupedValues);

				DS_PreparedRunBuilder builder = DS_PreparedRun.builder();
				builder.setSplitGroupedValues(splitGroupedValues);
				builder.setInstancesWithMatches(instancesWithMatches);
				builder.setEntitySets(entitySets);
				builder.setSplitGroupedValuesSets(splitGroupedValuesSets);
				builder.setSplitRowValues(splitRowValues);
				builder.setThresholdContainers(thresholdContainers);

				preapredRuns.put(classId, builder.build());

			}

			return new ClusteringRunner(input, preapredRuns);
		} catch (Exception e) {
			throw new IllegalStateException(e);
		}
	}

	public class ClusteringRunner {
		private final DistantSupervisionInput input;
		private final Map<ClassId, DS_PreparedRun> preapredRuns;
		private final EvaluationMethod evaluationMethod = clusteringLearner.getEvaluationMethod();
		private final WeightGetter<InstanceId> weightsGetter = (i) -> {
			return 1.0;
		};

		public ClusteringRunner(DistantSupervisionInput input, Map<ClassId, DS_PreparedRun> preapredRuns) {
			super();
			this.input = input;
			this.preapredRuns = ImmuMaker.map(preapredRuns);
		}

		public ClusterList getList(ClassId classId, SetExpansionStrategy strategy, LearnedModelRepository strategies) {



			RowClusteringConfig rowClusteringConfig = getRowClusteringConfig(classId, strategies);
			ClustererInput clustererInput = getClustererInput(classId);

			RowClusterer clusterer = rowClustererRepository.get(strategy.getClusterer());
			RowClusterSimilarities sim = clusterer.clusterRows(rowClusteringConfig, clustererInput);

			ClusterList clusters = greedyClusteringSplitter.clusterRows(sim, rowClusteringConfig, clustererInput, strategy.getOptimizableClassCluster(), false,
					false);

			return clusters;
		}

		public ClusterList getList(ClassId classId, SetExpansionStrategy strategy, LearnedModelRepository strategies, RowClusterSimilarities sim) {

			RowClusteringConfig rowClusteringConfig = getRowClusteringConfig(classId, strategies);
			ClustererInput clustererInput = getClustererInput(classId);

			ClusterList clusters = greedyClusteringSplitter.clusterRows(sim, rowClusteringConfig, clustererInput, strategy.getOptimizableClassCluster(), false,
					false);

			return clusters;
		}

		public ClusterList getList(ClassId classId, SetExpansionStrategy strategy, LearnedModelRepository strategies, Set<TableRowId> candidates) {

			RowClusteringConfig rowClusteringConfig = getRowClusteringConfig(classId, strategies);
			ClustererInput clustererInput = getClustererInput(classId, candidates);

			RowClusterer clusterer = rowClustererRepository.get(strategy.getClusterer());
			RowClusterSimilarities sim = clusterer.clusterRows(rowClusteringConfig, clustererInput);

			ClusterList clusters = greedyClusteringSplitter.clusterRows(sim, rowClusteringConfig, clustererInput, strategy.getOptimizableClassCluster(), false,
					false);

			return clusters;
		}

		public ClustererInput getClustererInput(ClassId classId, Set<TableRowId> candidates) {
			TableIdsSet tableIdsSet = input.getTableIdsSet(classId);
			TableRowIdsSet rowIdsSet = input.getFullRowIdsSet(classId);
			DS_PreparedRun preapredRun = preapredRuns.get(classId);
			return new ClustererInput(preapredRun.getSplitGroupedValuesSets(), candidates, rowIdsSet, tableIdsSet, null, classId,
					preapredRun.getSplitRowValues(), input.getClassesWithParents(), input.getNonTailProperties(), input.getFullRowIdsSetAllClasses());
		}

		public ClustererInput getClustererInput(ClassId classId) {
			return getClustererInput(classId, input.getWsRowIdsSet(classId));
			// TableIdsSet tableIdsSet = input.getTableIdsSet(classId);
			// TableRowIdsSet rowIdsSet = input.getFullRowIdsSet(classId);
			// DS_PreparedRun preapredRun = preapredRuns.get(classId);
			// return new
			// ClustererInput(preapredRun.getSplitGroupedValuesSets(),
			// input.getWsRowIdsSet(classId), rowIdsSet, tableIdsSet, null,
			// classId,
			// preapredRun.getSplitRowValues(), input.getClassesWithParents(),
			// input.getNonTailProperties(),
			// input.getFullRowIdsSetAllClasses());
		}

		public RowClusteringConfig getRowClusteringConfig(ClassId classId, LearnedModelRepository strategies) {
			return new RowClusteringConfig(input.getSetExpansionStrategiesConfig(), preapredRuns.get(classId).getThresholdContainers(), strategies, true,
					false, input.getFusionConfig());
		}

		public ParamterClusterer getParamterClusterer(ClassId classId, SetExpansionStrategy strategy, 
				Set<InstanceId> newEntities) {

			RowClusteringConfig rowClusteringConfigLearning = getRowClusteringConfig(classId, null);
			ClustererInput clustererInputLearning = getClustererInput(classId);
			// DS_PreparedRun preapredRun = preapredRuns.get(classId);
			RowClusterer clusterer = rowClustererRepository.get(strategy.getClusterer());

			if (clusterer instanceof ParameterLearnableClusterer) {
				return ((ParameterLearnableClusterer) clusterer).getCreator(rowClusteringConfigLearning, clustererInputLearning);
			}
			else {
				throw new IllegalStateException();
			}

		}

		public LearnedModelRepository learn(ClassId classId, SetExpansionStrategy strategy, Map<TableRowIdPair, PairLearningData> rowIdPairs,
				Set<InstanceId> newEntities) {

			RowClusteringConfig rowClusteringConfigLearning = getRowClusteringConfig(classId, null);
			ClustererInput clustererInputLearning = getClustererInput(classId);
			DS_PreparedRun preapredRun = preapredRuns.get(classId);
			RowClusterer clusterer = rowClustererRepository.get(strategy.getClusterer());

			LearnedModel learnedModel = clusteringLearner.learn(classId,
					(LearnableRowClusterSimilarities) clusterer.clusterRows(rowClusteringConfigLearning, clustererInputLearning), strategy, rowIdPairs,
					evaluationMethod, clusterer, input.getSetExpansionStrategiesConfig(), preapredRun.getThresholdContainers(), clustererInputLearning, null,
					newEntities,
					weightsGetter, input.getFusionConfig());

			Map<OptimizableClass, LearnedModel> repository = new HashMap<>();

			repository.put(strategy.getOptimizableClassCluster(), learnedModel);

			return new LearnedModelRepository(repository);
		}

		// public EM_LearnedModelRepository learnEMWithIndex(ClassId classId,
		// EntityMatchingRun run, Map<TableRowId, InstanceId> index,
		// Map<ClassId, Set<InstanceId>> newEntitiesLearning) {
		//
		// return learnEM(classId, run, new GoldStandardClusters(index),
		// newEntitiesLearning);
		// }
		//
		// public EM_LearnedModelRepository learnEMWithClusters(ClassId classId,
		// EntityMatchingRun run, Map<InstanceId, Set<TableRowId>> index,
		// Map<ClassId, Set<InstanceId>> newEntitiesLearning) {
		//
		// return learnEM(classId, run, new GoldStandardClusters(index, true),
		// newEntitiesLearning);
		// }

		public EM_LearnedModelRepository learnEM(ClassId classId, EntityMatchingRun run, List<RowCluster> clusterList,
				Map<Pair<RowClusterUUID, InstanceId>, PairLearningData> pairs, Set<InstanceId> newEntities,
				Map<RowClusterUUID, Optional<InstanceId>> gsMapped) {
			EntityMatchingModelLearnerRunner learner = learnerRepository.get(run.getModel());

			return learnEM(classId, run, learner, clusterList, pairs, newEntities, gsMapped);

		}

		public EM_LearnedModelRepository learnEM(ClassId classId, EntityMatchingRun run, EntityMatchingModelLearnerRunner learner, List<RowCluster> clusterList,
				Map<Pair<RowClusterUUID, InstanceId>, PairLearningData> pairs, Set<InstanceId> newEntities,
				Map<RowClusterUUID, Optional<InstanceId>> gsMapped) {
			EntityMatcherFullResults fullRun = getFullRun(classId, clusterList, run);


			EntityMatchingModel model = learner.learnModel(run, pairs, fullRun, newEntities, gsMapped);

			Map<OptimizableClass, EntityMatchingModel> retWeights = new HashMap<>();
			retWeights.put(run.getOptimizableClass(), model);

			return new EM_LearnedModelRepository(retWeights);

		}


		public EM_LearnedModelRepository learnEM(ClassId classId, EntityMatchingRun run, Map<Pair<RowClusterUUID, InstanceId>, PairLearningData> pairs,
				EntityMatcherFullResults fullRun, Set<InstanceId> newEntities, Map<RowClusterUUID, Optional<InstanceId>> gsMapped) {


			EntityMatchingModelLearnerRunner learner = learnerRepository.get(run.getModel());
			EntityMatchingModel model = learner.learnModel(run, pairs, fullRun, newEntities, gsMapped);

			Map<OptimizableClass, EntityMatchingModel> retWeights = new HashMap<>();
			retWeights.put(run.getOptimizableClass(), model);

			return new EM_LearnedModelRepository(retWeights);

		}

		public EntityMatcherWeightedResults runMatching(ClassId classId, List<RowCluster> clusterList, EntityMatchingRun run,
				EM_LearnedModelRepository learnedModels) {
			DS_PreparedRun preapredRun = preapredRuns.get(classId);
			ClustererInput clustererInput = getClustererInput(classId);

			List<Entity> entities = entityCreator.createEntites(preapredRun.getSplitGroupedValuesSets(), preapredRun.getThresholdContainers(),
					input.getFusionConfig(), input.getStrategiesConfig(), input.getSetExpansionStrategiesConfig(), clusterList, clustererInput);

			EntityMatcherConfig emConfig = new EntityMatcherConfig(clustererInput, input.getNonTailProperties().get(classId), input.getClassesWithParents());

			EntityMatcherWeightedResults matcherReturn = entityMatchingRuner.weightedRun(emConfig, run, entities, input.getClassesWithParents(),
					learnedModels);

			return matcherReturn;
		}

		public EntityMatcherWeightedResults runMatching(ClassId classId, EntityMatchingRun run, EM_LearnedModelRepository learnedModels,
				EntityMatcherFullResults fullRun) {

			EntityMatcherWeightedResults matcherReturn = entityMatchingRuner.weightedRun(run, learnedModels, fullRun);

			return matcherReturn;
		}

		public Map<Pair<RowClusterUUID, InstanceId>, Pair<Boolean, Double>> getFullNonAbsoluteEMresults(ClassId classId, EntityMatchingRun run,
				EM_LearnedModelRepository models, EntityMatcherFullResults fullRun) {

			EntityMatchingModel model = models.get(run.getOptimizableClass());

			if (model instanceof EntityMatchingWekaModel) {
				EntityMatchingWekaModel wekaModel = (EntityMatchingWekaModel) model;
				ND_MR_SCORE_ADAPTER scoreAdapter = wekaModel.getScoreAdapter();
				if (scoreAdapter != null) {
					scoreAdapter.clean();
				}
				Map<Pair<RowClusterUUID, InstanceId>, Pair<Boolean, Double>> fullResults = model.getFullResults(fullRun);

				int positivePairsSize = 0;
				int negativePairsSize = 0;

				for (Entry<Pair<RowClusterUUID, InstanceId>, Pair<Boolean, Double>> v : fullResults.entrySet()) {
					if (v.getValue().getFirst()) {
						positivePairsSize++;
					} else {
						negativePairsSize++;
					}
				}

				if (scoreAdapter != null) {
					System.out.println("For Class: " + fullRun.getEmConfig().getClustererInput().getClassId().getFinalPart() + ", Positive Pairs "
							+ positivePairsSize + ", Negative Pairs: " + negativePairsSize + ", positive rules fired: " + scoreAdapter.getPositivePairsSize()
							+ ", negative rules fired: " + scoreAdapter.getNegativePairsSize());
				} else {
					System.out.println("For Class: " + fullRun.getEmConfig().getClustererInput().getClassId().getFinalPart() + " Positive Pairs "
							+ positivePairsSize + ", Negative Pairs: " + negativePairsSize);
				}

				return fullResults;
			} else {
				return model.getFullResults(fullRun);
			}

		}

		public EntityMatcherFullResults getFullRun(ClassId classId, List<RowCluster> clusterList, EntityMatchingRun run) {
			DS_PreparedRun preapredRun = preapredRuns.get(classId);
			ClustererInput clustererInput = getClustererInput(classId);

			List<Entity> entities = entityCreator.createEntites(preapredRun.getSplitGroupedValuesSets(), preapredRun.getThresholdContainers(),
					input.getFusionConfig(), input.getStrategiesConfig(), input.getSetExpansionStrategiesConfig(), clusterList, clustererInput);

			EntityMatcherConfig emConfig = new EntityMatcherConfig(clustererInput, input.getNonTailProperties().get(classId), input.getClassesWithParents());

			EntityMatcherFullResults fullRun = entityMatchingRuner.fullRun(emConfig, run, entities, input.getClassesWithParents()
					, Lists.newArrayList(CandidateSelectionStrategies.RANDOM, CandidateSelectionStrategies.ATTRIBUTE)
			);

			return fullRun;
		}

		public EntityMatcherWeightedResults runMatching(EntityMatchingRun run, EM_LearnedModelRepository learnedModels, EntityMatcherFullResults fullRun) {

			EntityMatcherWeightedResults matcherReturn = entityMatchingRuner.weightedRun(run, learnedModels, fullRun);

			return matcherReturn;
		}

		public RowClusterSimilarities getSim(ClassId classId, SetExpansionStrategy strategy, LearnedModelRepository strategies) {
			RowClusteringConfig rowClusteringConfig = getRowClusteringConfig(classId, strategies);
			ClustererInput clustererInput = getClustererInput(classId);

			RowClusterer clusterer = rowClustererRepository.get(strategy.getClusterer());
			RowClusterSimilarities sim = clusterer.clusterRows(rowClusteringConfig, clustererInput);
			return sim;
		}

		public ClusterList getHomonymClusters(ClassId classId, Set<TableRowId> list) {

			RowClusterer clusterer = rowClustererRepository.get(SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY.getClusterer());
			OptimizableClass optimizableClassClustering = SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY.getOptimizableClassCluster();
			OptimizableClass optimizableClassClusteringSimilarity = SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY
					.getOptimizableClassSimilarityThreshold();

			SetExpansionStrategiesConfig setExpansionStrategiesConfig = SetExpansionStrategiesConfig.defaultConfig();

			Map<OptimizableWeightKey, Double> weights = new HashMap<>();
			weights.put(optimizableClassClustering.getKey("LABEL"), 1.0);
			weights.put(optimizableClassClustering.getKey("ZERO"), 0.0);
			weights.put(optimizableClassClusteringSimilarity.getKey("SIMILARITY_THRESHOLD"), 0.40);

			Map<OptimizableClass, LearnedModel> map = new HashMap<>();
			map.put(optimizableClassClusteringSimilarity, new WeightsLearnedModel(optimizableClassClusteringSimilarity, new WeightsRepositoryImpl(weights)));

			RowClusteringConfig rowClusteringConfig = new RowClusteringConfig(setExpansionStrategiesConfig, new ArrayList<>(), new LearnedModelRepository(map),
					true, false, input.getFusionConfig());

			ClustererInput clustererInput = new ClustererInput(
					new SplitGroupedValuesSets(new SplitGroupedValues(new HashMap<>()), new SplitGroupedValues(new HashMap<>()),
							new SplitGroupedValues(new HashMap<>())),
					list, input.getFullRowIdsSet(classId), input.getTableIdsSet(classId), null, classId, input.getFullRowIdsSetAllClasses());
			RowClusterSimilarities sim = clusterer.clusterRows(rowClusteringConfig, clustererInput);

			sim = labelSimCreator.createForGs(clustererInput.getFullCandidates());

			ClusterList clusters = greedyClusteringSplitter.clusterRows(sim, rowClusteringConfig, clustererInput, optimizableClassClusteringSimilarity, false,
					false);

			return clusters;
		}

		public LearnedModelRepository learnFromParamter(ClassId classId, SetExpansionStrategy strategy, ParamterClusterer clusterer, String[] parameters,
				Map<TableRowIdPair, PairLearningData> rowIdPairs) {
			LearnedModel learnedModel = clusterer.clusterRows(parameters).learn(rowIdPairs, classId, evaluationMethod);

			System.out.println(learnedModel.toString());
			Map<OptimizableClass, LearnedModel> repository = new HashMap<>();

			repository.put(strategy.getOptimizableClassCluster(), learnedModel);

			return new LearnedModelRepository(repository);

		}
		public ClusterSimilaritiesImplNormalized getNormalizedCS(ClassId classId, SetExpansionStrategy strategy, LearnedModelRepository strategies) {
			RowClusteringConfig rowClusteringConfig = getRowClusteringConfig(classId, strategies);
			ClustererInput clustererInput = getClustererInput(classId);
			RowClusterer clusterer = rowClustererRepository.get(strategy.getClusterer());
			RowClusterSimilarities sim = clusterer.clusterRows(rowClusteringConfig, clustererInput);
			return (ClusterSimilaritiesImplNormalized) greedyClusteringSplitter.getCs(sim, rowClusteringConfig, clustererInput,
					strategy.getOptimizableClassCluster());
		}

		public double getPerformance(ClusterSimilaritiesImplNormalized cs, Map<TableRowIdPair, PairLearningData> rowIdPairs) {
			AtomicDouble correct = new AtomicDouble();
			AtomicDouble wrong = new AtomicDouble();
			AtomicDouble maxCorrect = new AtomicDouble();
			PersistentParallel6.runAndFinish(rowIdPairs.entrySet(), (e) -> {
				double score = cs.get(e.getKey());
				if (score > 0.0) {
					if (e.getValue().isMatch()) {
						correct.addAndGet(1.0);
					} else {
						wrong.addAndGet(1.0);
					}
				}

				if (e.getValue().isMatch()) {
					maxCorrect.addAndGet(1.0);
				}
			}, this.getClass());

			double precision = correct.get() / (correct.get() + wrong.get());
			double recall = correct.get() / maxCorrect.get();

			if (Double.isNaN(precision)) {
				precision = 0.0;
			}
			if (Double.isNaN(recall)) {
				recall = 0.0;
			}

			double fmeasure = 2 * precision * recall / (precision + recall);

			if (Double.isNaN(fmeasure)) {
				fmeasure = 0.0;
			}

			return fmeasure;
		}

		public Pair<Set<InstanceId>, Map<RowClusterUUID, Optional<InstanceId>>> getGsEmData(Map<Pair<RowClusterUUID, InstanceId>, PairLearningData> pairsFull) {

			Set<RowClusterUUID> rowClusters = pairsFull.keySet().stream().map((e) -> e.getFirst()).collect(Collectors.toSet());

			ConcurrentHashMap<RowClusterUUID, Map<InstanceId, Double>> matches = new ConcurrentHashMap<>();
			Set<InstanceId> newEntities = new HashSet<>();
			Map<RowClusterUUID, Optional<InstanceId>> gsMapped = new HashMap<>();

			PersistentParallel6.runAndFinish(pairsFull.entrySet(), (pair) -> {

				if (pair.getValue().isMatch()) {
					CHMHelper.getFromCHM(matches, pair.getKey().getFirst(), () -> new ConcurrentHashMap<>()).put(pair.getKey().getSecond(),
							pair.getValue().getWeight());

				}

			}, this.getClass());

			for (RowClusterUUID uuid : rowClusters) {
				if (!matches.containsKey(uuid)) {
					InstanceId instanceId = new InstanceId(uuid.toString());

					newEntities.add(instanceId);
					gsMapped.put(uuid, Optional.of(instanceId));
				}
			}

			for (Entry<RowClusterUUID, Map<InstanceId, Double>> entry : matches.entrySet()) {
				InstanceId instanceId = null;
				double d = 0.0;
				for (Entry<InstanceId, Double> innerEntry : entry.getValue().entrySet()) {

					if (innerEntry.getValue() >= 0.0) {
						if (instanceId == null || innerEntry.getValue() > d) {
							instanceId = innerEntry.getKey();
							d = innerEntry.getValue();
						}
					}
				}
				if (instanceId != null) {
					gsMapped.put(entry.getKey(), Optional.of(instanceId));
				}
			}

			return new Pair<>(newEntities, gsMapped);
		}

		// public double getPerformance(EntityMatchingModel model,
		// EntityMatcherFullResults fullRun, Map<Pair<RowClusterUUID,
		// InstanceId>, PairLearningData> rowIdPairs) {
		// AtomicDouble correct = new AtomicDouble();
		// AtomicDouble wrong = new AtomicDouble();
		// AtomicDouble maxCorrect = new AtomicDouble();
		// EntityMatcherWeightedResults singleResults = model.apply(fullRun);
		//
		// Set<Pair<RowClusterUUID, InstanceId>> pairMatches =
		// singleResults.getPairMatches();
		//
		//
		// PersistentParallel6.runAndFinish(rowIdPairs.entrySet(), (e) -> {
		// double score = model..get(e.getKey());
		// if (score > 0.0) {
		// if (e.getValue().isMatch()) {
		// correct.addAndGet(1.0);
		// } else {
		// wrong.addAndGet(1.0);
		// }
		// }
		//
		// if (e.getValue().isMatch()) {
		// maxCorrect.addAndGet(1.0);
		// }
		// }, this.getClass());
		//
		// double precision = correct.get() / (correct.get() + wrong.get());
		// double recall = correct.get() / maxCorrect.get();
		//
		// if (Double.isNaN(precision)) {
		// precision = 0.0;
		// }
		// if (Double.isNaN(recall)) {
		// recall = 0.0;
		// }
		//
		// double fmeasure = 2 * precision * recall / (precision + recall);
		//
		// if (Double.isNaN(fmeasure)) {
		// fmeasure = 0.0;
		// }
		//
		// return fmeasure;
		// }

	}
	
}
