package expansion.set.distant_supervision.runners;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.ConfigInterface;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;
import de.mannheim.uni.expansion.tablefusion.learning.SplitRowValues;
import de.mannheim.uni.expansion.utils.ImmuMaker;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdContainer;
import expansion.set.evaluation.crossvalidation.EntitySets;

public class DS_PreparedRun implements ConfigInterface<DS_PreparedRun> {
	private SplitGroupedValues splitGroupedValues;
	private Map<ClassId, Set<InstanceId>> instancesWithMatches;
	private EntitySets entitySets;
	private SplitGroupedValuesSets splitGroupedValuesSets;
	private SplitRowValues splitRowValues;
	private List<ThresholdContainer> thresholdContainers;

	private DS_PreparedRun() {
	}

	@Override
	public DS_PreparedRun copy() {
		DS_PreparedRun copy = new DS_PreparedRun();
		copy.splitGroupedValues = splitGroupedValues;
		copy.instancesWithMatches = instancesWithMatches;
		copy.entitySets = entitySets;
		copy.splitGroupedValuesSets = splitGroupedValuesSets;
		copy.splitRowValues = splitRowValues;
		copy.thresholdContainers = thresholdContainers;
		return copy;
	}

	@Override
	public DS_PreparedRunBuilder rebuilder() {
		return new DS_PreparedRunBuilder(this);
	}

	public static DS_PreparedRunBuilder builder() {
		return new DS_PreparedRunBuilder();
	}


	public SplitGroupedValues getSplitGroupedValues() {
		return splitGroupedValues;
	}

	public Map<ClassId, Set<InstanceId>> getInstancesWithMatches() {
		return instancesWithMatches;
	}

	public EntitySets getEntitySets() {
		return entitySets;
	}


	public SplitGroupedValuesSets getSplitGroupedValuesSets() {
		return splitGroupedValuesSets;
	}

	public SplitRowValues getSplitRowValues() {
		return splitRowValues;
	}

	public List<ThresholdContainer> getThresholdContainers() {
		return thresholdContainers;
	}

	public static class DS_PreparedRunBuilder implements ConfigBuilderInterface<DS_PreparedRun> {
		private final DS_PreparedRun run;

		public DS_PreparedRunBuilder(DS_PreparedRun run) {
			this.run = run.copy();
		}

		public DS_PreparedRunBuilder() {
			run = new DS_PreparedRun();
		}

		public DS_PreparedRunBuilder setSplitGroupedValues(SplitGroupedValues splitGroupedValues) {
			run.splitGroupedValues = splitGroupedValues;
			return this;
		}

		public DS_PreparedRunBuilder setInstancesWithMatches(Map<ClassId, Set<InstanceId>> instancesWithMatches) {
			run.instancesWithMatches = ImmuMaker.mapWithSet(instancesWithMatches);
			return this;
		}

		public DS_PreparedRunBuilder setEntitySets(EntitySets entitySets) {
			run.entitySets = entitySets;
			return this;
		}

		public DS_PreparedRunBuilder setSplitGroupedValuesSets(SplitGroupedValuesSets splitGroupedValuesSets) {
			run.splitGroupedValuesSets = splitGroupedValuesSets;
			return this;
		}

		public DS_PreparedRunBuilder setSplitRowValues(SplitRowValues splitRowValues) {
			run.splitRowValues = splitRowValues;
			return this;
		}

		public void setThresholdContainers(List<ThresholdContainer> thresholdContainers) {
			run.thresholdContainers = Collections.unmodifiableList(new ArrayList<>(thresholdContainers));
		}

		@Override
		public DS_PreparedRun build() {
			return run.copy();
		}

	}


}
