package expansion.set.entitymatching;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.ImmuMaker;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.creation.Entity;
import expansion.set.entitymatching.matchers.EntityMatcherConfig;

public class EntityMatcherFullResults {
	private final EntityMatcherConfig emConfig;
	private final Map<RowClusterUUID, Entity> entities;
	private final Map<RowClusterUUID, Map<InstanceId, EntityMatcherSingleResults>> results;

	public EntityMatcherFullResults(EntityMatcherConfig emConfig, Map<RowClusterUUID, Entity> entities,
			Map<RowClusterUUID, Map<InstanceId, EntityMatcherSingleResults>> results) {
		super();
		this.entities = ImmuMaker.map(entities);
		this.results = ImmuMaker.mapWithMap(results);
		this.emConfig = emConfig;
	}

	public EntityMatcherFullResults(EntityMatcherConfig emConfig, List<Entity> entities,
			Map<RowClusterUUID, Map<InstanceId, EntityMatcherSingleResults>> results) {
		super();
		this.entities = Collections.unmodifiableMap(entities.stream().collect(Collectors.toMap(Entity::getRowClusterUUID, Function.identity())));

		this.results = ImmuMaker.mapWithMap(results);
		this.emConfig = emConfig;
	}

	public Map<RowClusterUUID, Map<InstanceId, EntityMatcherSingleResults>> getResults() {
		return results;
	}

	public Map<RowClusterUUID, Entity> getEntities() {
		return entities;
	}

	public EntityMatcherConfig getEmConfig() {
		return emConfig;
	}

}
