package expansion.set.entitymatching;

import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.ImmuMaker;
import de.mannheim.uni.expansion.utils.WeightedScore;
import expansion.set.entitymatching.matchers.EntityMatcherStrategy;

public class EntityMatcherSingleResults {

	private final Map<EntityMatcherStrategy, List<WeightedScore>> scores;

	public EntityMatcherSingleResults(Map<EntityMatcherStrategy, List<WeightedScore>> scores) {
		super();
		this.scores = ImmuMaker.withList(scores);
	}

	public Map<EntityMatcherStrategy, List<WeightedScore>> getScores() {
		return scores;
	}

	public LinkedHashMap<String, Double> getScoreMap(Set<PropertyId> propertyIds, Set<EntityMatcherStrategy> withConfidence) {
		LinkedHashMap<String, Double> map = new LinkedHashMap<>();
		
		for(Entry<EntityMatcherStrategy, List<WeightedScore>> entry : this.scores.entrySet()) {
			
			if (withConfidence.contains(entry.getKey())) {
				for (PropertyId propertyId : propertyIds) {
					map.put(entry.getKey().getName() + "::" + propertyId.readableString() + "::MATCH", 0.0);
					map.put(entry.getKey().getName() + "::" + propertyId.readableString() + "::NO_MATCH", 0.0);
				}
			}

			double sum = 0.0;
			double count = 0.0;
			for(WeightedScore score : entry.getValue()) {
				sum += score.getScore() * score.getWeight();
				count += score.getWeight();

				if (withConfidence.contains(entry.getKey())) {
					if (score.isMatch()) {
						map.put(entry.getKey().getName() + "::" + score.getSuffix().get() + "::MATCH", score.getWeight());
					} else {
						map.put(entry.getKey().getName() + "::" + score.getSuffix().get() + "::NO_MATCH", score.getWeight());
					}
				}
			}
			
			if(count > 0.0) {
				double score = sum / count;
				
				map.put(entry.getKey().getName(), score);
				if (withConfidence.contains(entry.getKey())) {
					map.put(entry.getKey().getName() + "::CONFIDENCE", count);
				}
			}
			else {

				map.put(entry.getKey().getName(), 0.0);
				if (withConfidence.contains(entry.getKey())) {
					map.put(entry.getKey().getName() + "::CONFIDENCE", 0.0);
				}
			}
			
			
		
		}
		
		
		return map;
	}

}
