package expansion.set.entitymatching;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import javax.inject.Inject;

import com.beust.jcommander.internal.Lists;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdsSet;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.ResultsCategory;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.generic.GenericResultsWriter;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.tablefusion.learning.RowValuesGetter;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;
import de.mannheim.uni.expansion.tablefusion.learning.SplitRowValues;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.DecimalFormatter;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdContainer;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList.ColumnListBuilderGeneric;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentAccumulator;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.creation.Entity;
import expansion.set.creation.EntityCreator;
import expansion.set.entitymatching.candidateselection.CandidateSelectionStrategies;
import expansion.set.entitymatching.matchers.EntityMatcherConfig;
import expansion.set.entitymatching.modelLearning.EM_LearnedModelRepository;
import expansion.set.entitymatching.modelLearning.EntityMatchingModel;
import expansion.set.entitymatching.modelLearning.EntityMatchingModelLearnerRepository;
import expansion.set.entitymatching.modelLearning.EntityMatchingModelLearnerRunner;
import expansion.set.entitymatching.modelLearning.weka.EntityMatcherWekaLearner;
import expansion.set.entitymatching.modelLearning.weka.EntityMatchingWekaModel;
import expansion.set.entitymatching.run.EntityMatchingRun;
import expansion.set.evaluation.ND_CaseCreator;
import expansion.set.evaluation.SetExpansionRunner;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.evaluation.crossvalidation.EntitySetGroupedValuesSplitter;
import expansion.set.evaluation.crossvalidation.EntitySets;
import expansion.set.evaluation.goldstandard.GoldStandardClusters;
import expansion.set.evaluation.matching.NewDetectionEvaluator;
import expansion.set.evaluation.matching.NewDetectionPerformance;
import expansion.set.optimizing.LearnedModelAdapter;
import expansion.set.optimizing.WeightsAdapter;

public class EntityMatchingModelLearner {

	@Inject
	protected ND_CaseCreator caseCreator;

	@Inject
	protected NewDetectionEvaluator newDetectionEvaluator;

	@Inject
	protected EntityMatchingModelLearnerRepository learnerRepository;

	@Inject
	protected EntityMatchingRunner entityMatchingRunner;

	@Inject
	protected EntitySetGroupedValuesSplitter entitySetGroupedValuesSplitter;

	@Inject
	protected RowValuesGetter rowValuesGetter;

	// @Inject
	// protected EntityMatchingGS_ClustererWeightsAdapterAndLearner oldLearner;

	@Inject
	protected SetExpansionRunner setExpansionRunner;

	@Inject
	protected AnnotatedTablesRepository tablesRepository;

	@Inject
	protected EntityCreator entityCreator;

	@Inject
	protected WeightsAdapter weightsAdapter;

	@Inject
	protected LearnedModelAdapter learnedModelAdapter;

	public Map<ClassId, Pair<EM_LearnedModelRepository, Map<EntityMatchingRun, NewDetectionPerformance>>> getLearnedModel(List<EntityMatchingRun> runs,
			Map<ClassId, GoldStandardClusters> learningGoldStandard,
			Map<ClassId, GoldStandardClusters> testingGoldStandard, FusionConfig fusionConfig, SetExpansionStrategiesConfig setExpansionStrategiesConfig,
			StrategiesConfig strategiesConfig, Map<ClassId, TableRowIdsSet> fullInput, Map<ClassId, Set<TableRowId>> learningInput,
			Map<ClassId, Set<TableRowId>> testingInput, String thresholdsFolder,
			Map<ClassId, Set<ClassId>> allParents, Map<ClassId, Set<InstanceId>> newEntitiesLearning, Map<ClassId, Set<InstanceId>> newEntitiesTesting,
			Map<ClassId, Set<PropertyId>> nonTailProperties, String instanceMatcherWeightsFolder, String caseFolder,
			Map<InstanceId, Set<TableRowId>> additionalExisting, String suffix) throws IOException {

		Map<ClassId, EM_LearnedModelRepository> weights = null;
		boolean hasWeights = false;
		if (false && instanceMatcherWeightsFolder != null) {
		weights = learnedModelAdapter.readModels((modelMap) ->  new EM_LearnedModelRepository(modelMap), learningGoldStandard.keySet(), instanceMatcherWeightsFolder);
		
		// if (false) {
//		 weightsAdapter.readWeights(learningGoldStandard.keySet(),
//		 instanceMatcherWeightsFolder);
		
		 hasWeights = true;
		 for (ClassId classId : learningGoldStandard.keySet()) {
			if (!weights.containsKey(classId) || weights.get(classId).getMap().isEmpty()) {
				hasWeights = false;
			}
		}
		}
		//
		// }

		if (!hasWeights) {
			Map<ClassId, EM_LearnedModelRepository> matchingWeightsMap = getWeightsNewDetection(runs, learningGoldStandard, fusionConfig,
					setExpansionStrategiesConfig, strategiesConfig, fullInput, learningInput, thresholdsFolder, allParents, newEntitiesLearning,
					null, null, nonTailProperties, null, additionalExisting, suffix);
			weights = matchingWeightsMap;
			if (instanceMatcherWeightsFolder != null)
				learnedModelAdapter.storeModels(instanceMatcherWeightsFolder, matchingWeightsMap);

			// weightsAdapter.storeWeights(instanceMatcherWeightsFolder,
			// weights);
		}
		Map<ClassId, Map<EntityMatchingRun, NewDetectionPerformance>> resultsMap = new
		HashMap<>();
		getWeightsNewDetection(runs, testingGoldStandard, fusionConfig, setExpansionStrategiesConfig, strategiesConfig, fullInput, testingInput,
				thresholdsFolder, allParents, newEntitiesTesting, new HashMap<>(weights), resultsMap, nonTailProperties, caseFolder, new HashMap<>(), suffix);

		Map<ClassId, Pair<EM_LearnedModelRepository, Map<EntityMatchingRun, NewDetectionPerformance>>> ret = new HashMap<>();

		for (ClassId classId : weights.keySet()) {
			ret.put(classId, new Pair<>(weights.get(classId), resultsMap.get(classId)));
		 }

		 
		return ret;

	}

	public Map<ClassId, EM_LearnedModelRepository> getWeightsNewDetection(List<EntityMatchingRun> runs, Map<ClassId, GoldStandardClusters> goldStandard,
			FusionConfig fusionConfig, SetExpansionStrategiesConfig setExpansionStrategiesConfig, StrategiesConfig strategiesConfig,
			Map<ClassId, TableRowIdsSet> fullInput, Map<ClassId, Set<TableRowId>> input, String thresholdsFolder,
			Map<ClassId, Set<ClassId>> allParents, Map<ClassId, Set<InstanceId>> newEntitiesMap,
			Map<ClassId, EM_LearnedModelRepository> matchingWeightsMap, Map<ClassId, Map<EntityMatchingRun, NewDetectionPerformance>> resultsMap,
			Map<ClassId, Set<PropertyId>> nonTailProperties, String caseFolder, Map<InstanceId, Set<TableRowId>> additionalExisting, String suffix)
			throws IOException {
		Map<ClassId, Map<OptimizableClass, EntityMatchingModel>> ret = new HashMap<>();

		Set<TableRowId> rowIdsAllClasses = new HashSet<>();
		for (ClassId classId : goldStandard.keySet()) {
			Set<TableRowId> rowIds = fullInput.get(classId).getRowIds();
			rowIdsAllClasses.addAll(rowIds);
		}
		TableRowIdsSet rowIdsSetAllClasses = new TableRowIdsSet(rowIdsAllClasses, new UuidStringId("ALL"));

		for (Entry<ClassId, GoldStandardClusters> classEntry : goldStandard.entrySet()) {
			ClassId classId = classEntry.getKey();
			Set<InstanceId> newEntities = newEntitiesMap.get(classId);
			Set<TableRowId> learningCandidates = new HashSet<>(input.get(classId));
			if (classId.getFinalPart().equals("Single")) {
				for (Entry<InstanceId, Set<TableRowId>> cluster : additionalExisting.entrySet()) {
					learningCandidates.addAll(cluster.getValue());

				}
			}


			TableRowIdsSet fullCandidates = fullInput.get(classId);

			TableIdsSet tableIdsSet = new TableIdsSet(this.tablesRepository.getAllTableIds(classId), new UuidStringId(classId.getFinalPart()));

			Map<ClassId, Set<InstanceId>> instancesWithMatches = setExpansionRunner.getInstancesWithMatches(classId, tableIdsSet.getTableIds(), 1);
			if (instancesWithMatches.get(classId).isEmpty()) {
				System.out.println("No Instances with MinSource found for: " + classId.getClassId());
				throw new IllegalStateException();
			}

			EntitySets entitySets = new EntitySets(instancesWithMatches.get(classId), instancesWithMatches.get(classId));
			SplitGroupedValues splitGroupedValues = setExpansionRunner.getSplitGroupedValues(fusionConfig, tableIdsSet);
			SplitGroupedValuesSets splitGroupedValuesSets = entitySetGroupedValuesSplitter.split(splitGroupedValues, entitySets);
			SplitRowValues splitRowValues = rowValuesGetter.getSplitRowValues(tablesRepository, tableIdsSet);
			ClustererInput clustererInput = new ClustererInput(splitGroupedValuesSets, learningCandidates, fullCandidates, tableIdsSet, null, classId,
					splitRowValues, allParents, nonTailProperties, rowIdsSetAllClasses);

			List<RowCluster> clusterList = new ArrayList<>();
			Map<RowClusterUUID, Optional<InstanceId>> gsMapped = new HashMap<>();

			if (true) {
				for (Entry<InstanceId, Set<TableRowId>> cluster : goldStandard.get(classId).getTestingSetMappingClusters().entrySet()) {
					RowCluster rc = new RowCluster(classId, new ArrayList<>(cluster.getValue()), null);
					clusterList.add(rc);

					gsMapped.put(rc.getUuid(), Optional.of(cluster.getKey()));

				}

				if (classId.getFinalPart().equals("Single")) {

					for (Entry<InstanceId, Set<TableRowId>> cluster : additionalExisting.entrySet()) {
						RowCluster rc = new RowCluster(classId, new ArrayList<>(cluster.getValue()), null);
						clusterList.add(rc);

						gsMapped.put(rc.getUuid(), Optional.of(cluster.getKey()));

					}

				}
			}
			List<ThresholdContainer> thresholdContainers = setExpansionRunner.getThresholdContainers(setExpansionStrategiesConfig, thresholdsFolder, classId,
					fusionConfig, strategiesConfig, splitGroupedValues);



			ClusterList clusters = new ClusterList(clusterList);

			List<Entity> entities = entityCreator.createEntites(splitGroupedValuesSets, thresholdContainers, fusionConfig, strategiesConfig,
					setExpansionStrategiesConfig, clusters.getClusterList(), clustererInput);// .stream().filter((e)


			EntityMatcherConfig emConfig = new EntityMatcherConfig(clustererInput, nonTailProperties.get(classId), allParents);
			for (EntityMatchingRun run : runs) {

				Map<OptimizableClass, EntityMatchingModel> retWeights = CHMHelper.getFromMap(ret, classId, () -> new HashMap<>());
				EntityMatchingModel matchingWeights;
				if (matchingWeightsMap != null)
					matchingWeights = matchingWeightsMap.get(classId).get(run.getOptimizableClass());
				else {
					EntityMatcherFullResults fullRun = entityMatchingRunner.fullRun(emConfig, run, entities, allParents,
							Lists.newArrayList(CandidateSelectionStrategies.ATTRIBUTE));

					EntityMatchingModelLearnerRunner learner = learnerRepository.get(run.getModel());

					matchingWeights = learner.learnModel(run, newEntities, gsMapped, fullRun);

					// if (run ==
					// EntityMatchingRunRepository.WA_LABEL_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREE)
					// System.out.println(matchingWeights.toString());
					// if (run ==
					// EntityMatchingRunRepository.RF_LABEL_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREE)
					// System.out.println(matchingWeights.toString());
					// if (run ==
					// EntityMatchingRunRepository.ET_LABEL_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREE)
					// System.out.println(matchingWeights.toString());
					// if (run ==
					// EntityMatchingRunRepository.ENSEMBLED_LABEL_TYPE_BOW_ATTRIBUTE_IMPLICIT_INDEGREENEW)
					// System.out.println(matchingWeights.toString());

					retWeights.put(run.getOptimizableClass(), matchingWeights);

				}

				if (resultsMap != null) {
					EntityMatcherFullResults fullRun = entityMatchingRunner.fullRun(emConfig, run, entities, allParents);

					EntityMatcherWeightedResults results = matchingWeights.apply(fullRun);

					NewDetectionPerformance accuracy = newDetectionEvaluator.evaluateAccuracy(newEntities, gsMapped, results.getExistingEntities(),
							results.getNewEntities(), classEntry.getValue());

					CHMHelper.getFromMap(resultsMap, classId, () -> new HashMap<>()).put(run, accuracy);

					if (caseFolder != null) {
						String evaluation = caseFolder + "/ndEvaluation";

						MKDIR.createDir(new File(evaluation));

						String strategyFolder = evaluation + "/" + run.getName() + "_" + suffix;
						MKDIR.createDir(new File(strategyFolder));

						caseCreator.createWithErrors(matchingWeights, strategyFolder, classId.getFinalPart(), newEntities, gsMapped, fullRun, results);
					}

				}

			}

		}

		if (resultsMap != null) {
			ColumnList<NewDetectionPerformance> columnList = getColumnListforPipelinePerformance();
			Map<EntityMatchingRun, ConcurrentAccumulator<NewDetectionPerformance>> totalAccum = new HashMap<>();
			GenericResultsWriter<NewDetectionPerformance> writer = new GenericResultsWriter<>(columnList);
			for (Entry<ClassId, Map<EntityMatchingRun, NewDetectionPerformance>> entry : resultsMap.entrySet()) {

				for (Entry<EntityMatchingRun, NewDetectionPerformance> innerEntry : entry.getValue().entrySet()) {
					CHMHelper.getFromMap(totalAccum, innerEntry.getKey(),
							() -> new ConcurrentAccumulator<NewDetectionPerformance>(new NewDetectionPerformance())).add(innerEntry.getValue());

					writer.writeClassId(innerEntry.getKey().getName(), entry.getKey(), innerEntry.getValue());
				}

			}

			for (Entry<EntityMatchingRun, ConcurrentAccumulator<NewDetectionPerformance>> entry : totalAccum.entrySet()) {
				writer.writeTotal(entry.getKey().getName(), entry.getValue().get());

			}

			System.out.println();
			System.out.println(writer.getAll(ResultsCategory.TOTAL));
			System.out.println();
			System.out.println(writer.getAll(ResultsCategory.CLASS_ID));
			int sdf = 123;

		}
		return ret.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> new EM_LearnedModelRepository(e.getValue())));

	}

	public Pair<EntityMatcherFullResults, Map<RowClusterUUID, Optional<InstanceId>>> getFullResults(EntityMatchingRun run, GoldStandardClusters goldStandard,
			FusionConfig fusionConfig,
			SetExpansionStrategiesConfig setExpansionStrategiesConfig, StrategiesConfig strategiesConfig, TableRowIdsSet fullCandidates,
			Set<TableRowId> learningCandidates, String thresholdsFolder, ClassId classId, Map<ClassId, Set<ClassId>> allParents, Set<InstanceId> newEntities,

			Map<ClassId, Set<PropertyId>> nonTailProperties, TableRowIdsSet fullInputAllClasses) {
		try {
		Map<ClassId, Map<OptimizableClass, EntityMatchingModel>> ret = new HashMap<>();

		TableIdsSet tableIdsSet = new TableIdsSet(this.tablesRepository.getAllTableIds(classId), new UuidStringId(classId.getFinalPart()));

		Map<ClassId, Set<InstanceId>> instancesWithMatches = setExpansionRunner.getInstancesWithMatches(classId, tableIdsSet.getTableIds(), 1);
		if (instancesWithMatches.get(classId).isEmpty()) {
			System.out.println("No Instances with MinSource found for: " + classId.getClassId());
			throw new IllegalStateException();
		}

		EntitySets entitySets = new EntitySets(instancesWithMatches.get(classId), instancesWithMatches.get(classId));
		SplitGroupedValues splitGroupedValues = setExpansionRunner.getSplitGroupedValues(fusionConfig, tableIdsSet);
		SplitGroupedValuesSets splitGroupedValuesSets = entitySetGroupedValuesSplitter.split(splitGroupedValues, entitySets);
		SplitRowValues splitRowValues = rowValuesGetter.getSplitRowValues(tablesRepository, tableIdsSet);
		ClustererInput clustererInput = new ClustererInput(splitGroupedValuesSets, learningCandidates, fullCandidates, tableIdsSet, null, classId,
					splitRowValues, allParents, nonTailProperties, fullInputAllClasses);

		List<RowCluster> clusterList = new ArrayList<>();
		Map<RowClusterUUID, Optional<InstanceId>> gsMapped = new HashMap<>();

		if (true) {
			for (Entry<InstanceId, Set<TableRowId>> cluster : goldStandard.getTestingSetMappingClusters().entrySet()) {
				RowCluster rc = new RowCluster(classId, new ArrayList<>(cluster.getValue()), null);
				clusterList.add(rc);

				gsMapped.put(rc.getUuid(), Optional.of(cluster.getKey()));

			}
		}

		List<ThresholdContainer> thresholdContainers = setExpansionRunner.getThresholdContainers(setExpansionStrategiesConfig, thresholdsFolder, classId,
				fusionConfig, strategiesConfig, splitGroupedValues);

		ClusterList clusters = new ClusterList(clusterList);

		List<Entity> entities = entityCreator.createEntites(splitGroupedValuesSets, thresholdContainers, fusionConfig, strategiesConfig,
				setExpansionStrategiesConfig, clusters.getClusterList(), clustererInput);// .stream().filter((e)

		EntityMatcherConfig emConfig = new EntityMatcherConfig(clustererInput, nonTailProperties.get(classId), allParents);

			EntityMatcherFullResults fullRun = entityMatchingRunner.fullRun(emConfig, run, entities, allParents,
					Lists.newArrayList(CandidateSelectionStrategies.ATTRIBUTE));
			return new Pair<>(fullRun, gsMapped);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	public double errorRun(EntityMatcherFullResults fullRun, Map<RowClusterUUID, Optional<InstanceId>> gsMapped, EntityMatchingRun run,
			EntityMatcherWekaLearner wekaLearner, Set<InstanceId> newEntities) {

		EntityMatchingModelLearnerRunner learner = wekaLearner;
		EntityMatchingModel matchingWeights = learner.learnModel(run, newEntities, gsMapped, fullRun);

			return ((EntityMatchingWekaModel) matchingWeights).getError();


	}

	public ColumnList<NewDetectionPerformance> getColumnListforPipelinePerformance() {

		ColumnListBuilderGeneric<NewDetectionPerformance> builder = ColumnList.builder();

		// ClusteringPerformance,GS-Size,Clusters,PCPrecision,Recall,FMeasure

		builder.addColumn("TotalDetectionAccuracy", (p) -> {
			return DecimalFormatter.format(p.getTotalAccuracy());
		});

		builder.addColumn("", (p) -> {
			return "";
		});

		builder.addColumn("ExistingClusters", (p) -> {
			return String.valueOf(p.getMappedExisting());
		});

		builder.addColumn("ReturnedExisting", (p) -> {
			return String.valueOf(p.getReturnedExisting());
		});

		builder.addColumn("ExistingPrecision", (p) -> {
			return DecimalFormatter.format(p.getExistingPrecision());
		});
		builder.addColumn("ExistingRecall", (p) -> {
			return DecimalFormatter.format(p.getExistingRecall());
		});

		builder.addColumn("ExistingF1", (p) -> {
			return DecimalFormatter.format(p.getExistingFmeasure());
		});

		builder.addColumn("", (p) -> {
			return "";
		});

		builder.addColumn("NewClusters", (p) -> {
			return String.valueOf(p.getMappedNew());
		});

		builder.addColumn("ReturnedNew", (p) -> {
			return String.valueOf(p.getReturnedNew());
		});

		builder.addColumn("NewPrecision", (p) -> {
			return DecimalFormatter.format(p.getNewPrecision());
		});
		builder.addColumn("NewRecall", (p) -> {
			return DecimalFormatter.format(p.getNewRecall());
		});
		builder.addColumn("NewF1", (p) -> {
			return DecimalFormatter.format(p.getNewFmeasure());
		});

		builder.addColumn("", (p) -> {
			return "";
		});

		builder.addColumn("IrPrecision", (p) -> {
			return DecimalFormatter.format(p.getIRPrecision());
		});
		builder.addColumn("IrRecall", (p) -> {
			return DecimalFormatter.format(p.getIRRecall());
		});
		builder.addColumn("IrF1", (p) -> {
			return DecimalFormatter.format(p.getIRFmeasure());
		});
		// builder.addColumn("BadClusters", (p) -> {
		// return String.valueOf(p.getMappedBad());
		// });

		//
		// builder.addColumn("BadPrecision", (p) -> {
		// return DecimalFormatter.format(p.getBadPrecision());
		// });
		// builder.addColumn("BadRecall", (p) -> {
		// return DecimalFormatter.format(p.getBadRecall());
		// });
		//
		//
		//
		// builder.addColumn("AverageFMeasure", (p) -> {
		// return DecimalFormatter.format(p.getAverageFmeasure());
		// });

		return builder.build();

	}
}
