package expansion.set.entitymatching.candidateselection;

import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.Map;

import javax.inject.Inject;

import expansion.set.entitymatching.candidateselection.strategies.CandidateSelection_Attribute;
import expansion.set.entitymatching.candidateselection.strategies.CandidateSelection_Label;
import expansion.set.entitymatching.candidateselection.strategies.CandidateSelection_Random;

public class CandidateSelectionRepository {

	@Inject
	protected CandidateSelection_Label CandidateSelection_Label;
	@Inject
	protected CandidateSelection_Random CandidateSelection_Random;

	@Inject
	protected CandidateSelection_Attribute CandidateSelection_Attribute;

	private volatile Map<Class<? extends CandidateSelection>, CandidateSelection> repository;

	@Inject
	public CandidateSelectionRepository() {

	}

	@SuppressWarnings("unchecked")
	public synchronized void init() {
		if (repository == null) {

			this.repository = new HashMap<>();

			try {
				for (Field field : this.getClass().getDeclaredFields()) {
					field.setAccessible(true); // You might want to set modifier
												// to public first.
					Class<?> clazz = field.getType();
					if (CandidateSelection.class.isAssignableFrom(clazz)) {
						CandidateSelection value = (CandidateSelection) field.get(this);
						this.repository.put((Class<? extends CandidateSelection>) clazz, value);

					}

				}
			} catch (IllegalArgumentException | IllegalAccessException e) {
				throw new RuntimeException(e);
			}

		}

	}

	public CandidateSelection get(Class<? extends CandidateSelection> clazz) {
		if (repository == null)
			init();

		if (clazz == null) {
			throw new IllegalStateException("Asking for Candidate Selection of class null is not allowed!");
		} else {
			CandidateSelection candidateSelection = this.repository.get(clazz);
			if (candidateSelection == null) {
				throw new IllegalStateException("No Candidate Selection was registered for Class: " + clazz.getSimpleName());
			} else {
				return candidateSelection;
			}
		}
	}
}
