package expansion.set.entitymatching.candidateselection;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.ConfigInterface;

public class CandidateSelectionResults implements ConfigInterface<CandidateSelectionResults> {

	private final Map<CandidateSelectionResultKey<?>, Object> results;

	private CandidateSelectionResults(Map<CandidateSelectionResultKey<?>, Object> results) {
		this.results = Collections.unmodifiableMap(new HashMap<>(results));
	}

	private CandidateSelectionResults(Map<CandidateSelectionResultKey<?>, Object> results, boolean flag) {
		this.results = results;
	}

	@Override
	public CandidateSelectionResults copy() {
		return new CandidateSelectionResults(results);
	}

	@Override
	public CandidateSelectionResultsBuilder rebuilder() {
		return new CandidateSelectionResultsBuilder(this);
	}

	public static CandidateSelectionResultsBuilder builder() {
		return new CandidateSelectionResultsBuilder();
	}

	@SuppressWarnings("unchecked")
	public <E> Optional<E> get(CandidateSelectionResultKey<E> key) {
		if (results.containsKey(key)) {
			return Optional.of((E) results.get(key));
		}
		return Optional.empty();
	}

	public static class CandidateSelectionResultsBuilder implements ConfigBuilderInterface<CandidateSelectionResults> {

		private final CandidateSelectionResults instance;

		public CandidateSelectionResultsBuilder(CandidateSelectionResults candidateSelectionResults) {
			instance = new CandidateSelectionResults(candidateSelectionResults.results, true);
		}

		public CandidateSelectionResultsBuilder() {
			instance = new CandidateSelectionResults(new HashMap<>(), true);
		}

		public CandidateSelectionResultsBuilder clear() {
			instance.results.clear();
			return this;
		}

		public <E> CandidateSelectionResultsBuilder put(CandidateSelectionResultKey<E> key, E object) {
			instance.results.put(key, object);
			return this;
		}

		@Override
		public CandidateSelectionResults build() {
			return instance.copy();
		}

	}


	public static class CandidateSelectionResultKey<E> {
		private final String name;
		private final Class<E> clazz;

		public CandidateSelectionResultKey(String name, Class<E> clazz) {
			this.name = name;
			this.clazz = clazz;
		}

		public String getName() {
			return name;
		}

		public Class<E> getClazz() {
			return clazz;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((clazz == null) ? 0 : clazz.hashCode());
			result = prime * result + ((name == null) ? 0 : name.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (!(obj instanceof CandidateSelectionResultKey))
				return false;
			@SuppressWarnings("rawtypes")
			CandidateSelectionResultKey other = (CandidateSelectionResultKey) obj;
			if (clazz == null) {
				if (other.clazz != null)
					return false;
			} else if (!clazz.equals(other.clazz))
				return false;
			if (name == null) {
				if (other.name != null)
					return false;
			} else if (!name.equals(other.name))
				return false;
			return true;
		}

	}

}
