package expansion.set.entitymatching.candidateselection.strategies;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.informatik.dws.matching.LabelMatcher;
import de.uni_mannheim.informatik.dws.matching.LabelMatcher.MatchStrictness;
import expansion.set.creation.Entity;
import expansion.set.entitymatching.candidateselection.CandidateSelection;
import expansion.set.entitymatching.candidateselection.CandidateSelectionConfig;
import expansion.set.entitymatching.candidateselection.CandidateSelectionResults;

public class CandidateSelection_Attribute implements CandidateSelection {

	@Inject
	protected KnowledgeBase_Attribute_Indeces indeces;

	@Inject
	protected LabelMatcher labelMatcher;

	@Inject
	public CandidateSelection_Attribute() {
	}



	@Override
	public Optional<Map<InstanceId, CandidateSelectionResults>> run(Entity entity, CandidateSelectionConfig config) {
		Set<ClassId> classIds = new HashSet<>(config.getAllParents().get(entity.getClassId()));
		Map<InstanceId, CandidateSelectionResults> ret = new HashMap<>();
		for (Entry<QualifiedPropertyId, FusedValue> entry : entity.getValues().entrySet()) {
			PropertyId propertyId = entry.getKey().getPropertyId();
			Optional<Set<InstanceId>> instanceIds = indeces.getRandomByAttribute(entity.getMainLabel().hashCode(), 1, entity.getClassId(), propertyId,
					entry.getValue().getSingle());
			if (instanceIds.isPresent()) {
				for (InstanceId instanceId : instanceIds.get()) {

					CandidateSelectionResults results = CandidateSelectionResults.builder()
							.put(CandidateSelection_Label.CS_LABEL_SCORE,
									labelMatcher.getLabelSim(MatchStrictness.NON_STRICT, entity.getLabels().keySet(), instanceId, true, false))
							.put(CandidateSelection_Label.CS_TYPE_SCORE, labelMatcher.getClassSimiliarity(instanceId, classIds))
							.put(CandidateSelection_Label.CS_PRIME_LABEL_SCORE,
									labelMatcher.getLabelSim(MatchStrictness.NON_STRICT, entity.getLabels().keySet(), instanceId, true, true))
							.build();

					ret.put(instanceId, results);

				}
			}
		}

		if (!ret.isEmpty()) {
			return Optional.of(ret);
		} else {
			return Optional.empty();
		}
	}

}
