package expansion.set.entitymatching.candidateselection.strategies;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Optional;
import java.util.Random;
import java.util.Set;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.informatik.dws.matching.LabelMatcher;
import de.uni_mannheim.informatik.dws.matching.LabelMatcher.MatchStrictness;
import expansion.set.creation.Entity;
import expansion.set.entitymatching.candidateselection.CandidateSelection;
import expansion.set.entitymatching.candidateselection.CandidateSelectionConfig;
import expansion.set.entitymatching.candidateselection.CandidateSelectionResults;
import expansion.set.entitymatching.candidateselection.CandidateSelectionResults.CandidateSelectionResultKey;

public class CandidateSelection_Random implements CandidateSelection {

	@Inject
	protected LabelMatcher labelMatcher;

	@Inject
	protected KnowledgeBase kb;

	@Inject
	public CandidateSelection_Random() {
	}

	public static CandidateSelectionResultKey<Double> CS_LABEL_SCORE = new CandidateSelectionResultKey<>("CS_LABEL_SCORE", Double.class);
	public static CandidateSelectionResultKey<Double> CS_TYPE_SCORE = new CandidateSelectionResultKey<>("CS_TYPE_SCORE", Double.class);
	public static CandidateSelectionResultKey<Double> CS_PRIME_LABEL_SCORE = new CandidateSelectionResultKey<>("CS_PRIME_LABEL_SCORE", Double.class);


	@Override
	public Optional<Map<InstanceId, CandidateSelectionResults>> run(Entity entity, CandidateSelectionConfig config) {
		Set<ClassId> classIds = new HashSet<>(config.getAllParents().get(entity.getClassId()));
		Set<ClassId> withoutMainClass = new HashSet<>(config.getAllParents().get(entity.getClassId()));
		withoutMainClass.remove(entity.getClassId());

		int hashCode = entity.getRowClusterUUID().hashCode();
		if (entity.getMainLabel() != null) {
			hashCode = entity.getMainLabel().hashCode();
		}

		Random random = new Random(hashCode);
		int instanceSize = kb.getNumberOfInstances(entity.getClassId());
		Set<InstanceId> instanceIds = new HashSet<>();
		for (int i = 0; i < 6; i++) {
			int instanceNum = random.nextInt(instanceSize);
			instanceIds.add(kb.getInstance(entity.getClassId(), instanceNum));
		}

		for (ClassId classId : withoutMainClass) {
			int instanceSizeOtherClass = kb.getNumberOfInstances(classId);
			for (int i = 0; i < 2; i++) {
				int instanceNum = random.nextInt(instanceSizeOtherClass);
				instanceIds.add(kb.getInstance(classId, instanceNum));
			}
		}

		Map<InstanceId, CandidateSelectionResults> ret = new HashMap<>();
		for (InstanceId instanceId : instanceIds) {

			CandidateSelectionResults results = CandidateSelectionResults.builder()
					.put(CS_LABEL_SCORE, labelMatcher.getLabelSim(MatchStrictness.NON_STRICT, entity.getLabels().keySet(), instanceId, true, false))
					.put(CS_TYPE_SCORE, labelMatcher.getClassSimiliarity(instanceId, classIds))
					.put(CS_PRIME_LABEL_SCORE, labelMatcher.getLabelSim(MatchStrictness.NON_STRICT, entity.getLabels().keySet(), instanceId, true, true))
					.build();

			ret.put(instanceId, results);

		}

		if (ret.isEmpty()) {
			throw new IllegalStateException();
		} else {

			return Optional.of(ret);
		}
	}

}
