package expansion.set.entitymatching.candidateselection.strategies;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NavigableMap;
import java.util.Optional;
import java.util.Random;
import java.util.Set;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentHashMap;

import com.google.common.collect.Sets;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.QuantityObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject.Precision;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.matching.LabelMatcher;
import de.uni_mannheim.informatik.dws.matching.LabelMatcher.MatchStrictness;

public class KnowledgeBase_Attribute_Indeces {

	private final Map<ClassPropertyId, PercentileGetter> percentileGetter;
	private final LabelMatcher labelMatcher;

	public KnowledgeBase_Attribute_Indeces(OntologyRepository ontology, KnowledgeBase kb, Map<ClassId, Set<PropertyId>> propertyIds,
			LabelMatcher labelMatcher) {
		Map<ClassPropertyId, PercentileGetter> percentileGetter = new HashMap<>();
		for (Entry<ClassId, Set<PropertyId>> entry : propertyIds.entrySet()) {
			ClassId classId = entry.getKey();
			for (PropertyId propertyId : entry.getValue()) {
				ValueType valueType = ontology.getDataType(propertyId);
				Set<InstanceId> instances = kb.getInstances(classId, propertyId);

				ConcurrentHashMap<InstanceId, List<AbstractObject>> values = new ConcurrentHashMap<>();
				PersistentParallel6.runAndFinish(instances, (instanceId) -> {
					List<AbstractObject> ao = kb.getValuesAsListWithEqualProperties(instanceId, propertyId);
					if (ao != null) {
						values.put(instanceId, ao);
					}
				}, this.getClass());
				ClassPropertyId cpid = new ClassPropertyId(classId, propertyId);
				Set<ClassId> range = ontology.getRange(cpid);
				percentileGetter.put(cpid, new PercentileGetter(valueType, values, range));
			}

		}

		this.percentileGetter = Collections.unmodifiableMap(percentileGetter);
		this.labelMatcher = labelMatcher;
	}

	public Optional<Set<InstanceId>> getRandomByAttribute(long seed, int max, ClassId classId, PropertyId propertyId, AbstractObject value) {
		Random random = new Random(seed);
		List<InstanceId> instanceIds = percentileGetter.get(new ClassPropertyId(classId, propertyId)).get(value);

		if (instanceIds == null) {
			return Optional.empty();
		} else if (instanceIds.size() <= max) {
			return Optional.of(new HashSet<>(instanceIds));
		} else {
			Set<Integer> indices = new HashSet<>();
			while (indices.size() < max) {
				int nextInt = random.nextInt(instanceIds.size());
				if (!indices.contains(nextInt)) {
					indices.add(nextInt);
				}
			}

			Set<InstanceId> ret = new HashSet<>();
			for (Integer index : indices) {
				ret.add(instanceIds.get(index));
			}

			return Optional.of(ret);
		}
	};

	private class PercentileGetter {
		private final Set<ClassId> targetClassIds;

		private final NavigableMap<Double, List<InstanceId>> percentileMap;

		private final Map<AbstractObject, List<InstanceId>> map;
		private final ValueType valueType;

		public PercentileGetter(ValueType type, Map<InstanceId, List<AbstractObject>> values, Set<ClassId> targetClassIds) {
			this.targetClassIds = targetClassIds;
			this.valueType = type;
			if (type == ValueType.QUANTITY) {

				map = null;
				int percentiles = values.size() / 15;
				Double max = null;
				Double min = null;
				for (List<AbstractObject> aoList : values.values()) {
					for (AbstractObject ao : aoList) {
						QuantityObject qo = (QuantityObject) ao;
						if (max == null || qo.getAmount() > max) {
							max = qo.getAmount();
						}

						if (min == null || qo.getAmount() < min) {
							min = qo.getAmount();
						}
					}
				}

				ConcurrentHashMap<Double, Set<InstanceId>> percentileMap = new ConcurrentHashMap<>();

				PersistentParallel6.runAndFinish(values.entrySet(), (entry) -> {
					for (AbstractObject ao : entry.getValue()) {
						QuantityObject qo = (QuantityObject) ao;
						CHMHelper.getFromCHM(percentileMap, qo.getAmount(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(entry.getKey());
					}
				}, this.getClass());

				TreeMap<Double, Set<InstanceId>> sortedPercentileMap = new TreeMap<>(percentileMap);
				TreeMap<Double, Set<InstanceId>> sortedCombinedPercentileMap = new TreeMap<>();

				{
					Set<InstanceId> instanceIds = new HashSet<>();
					Double d = null;
					for (Entry<Double, Set<InstanceId>> entry : sortedPercentileMap.entrySet()) {
						if (d == null) {
							d = entry.getKey();
						}
						if (instanceIds.size() + entry.getValue().size() > 50) {
							instanceIds.addAll(entry.getValue());
							sortedCombinedPercentileMap.put(d, instanceIds);
							instanceIds = new HashSet<>();
							d = null;
						} else {
							instanceIds.addAll(entry.getValue());
						}
					}
				}

				// double step = (max - min) / percentiles;

				// for (int i = -1; i <= percentiles; i++) {
				// percentileMap.put(min + (i * step),
				// Collections.newSetFromMap(new ConcurrentHashMap<>()));
				// }
				//
				// PersistentParallel6.runAndFinish(values.entrySet(), (entry)
				// -> {
				// for (AbstractObject ao : entry.getValue()) {
				// QuantityObject qo = (QuantityObject) ao;
				// percentileMap.floorEntry(qo.getAmount()).getValue().add(entry.getKey());
				// }
				// }, this.getClass());

				TreeMap<Double, List<InstanceId>> percentileMapFinal = new TreeMap<>();

				for (Entry<Double, Set<InstanceId>> entry : sortedCombinedPercentileMap.entrySet()) {
					List<InstanceId> list = new ArrayList<>(entry.getValue());
					Collections.sort(list, (i1, i2) -> {
						return i1.getInstanceId().compareTo(i2.getInstanceId());
					});
					percentileMapFinal.put(entry.getKey(), list);
				}

				this.percentileMap = Collections.unmodifiableNavigableMap(percentileMapFinal);
			} else {

				this.percentileMap = null;

				ConcurrentHashMap<AbstractObject, Set<InstanceId>> map = new ConcurrentHashMap<>();

				if (valueType == ValueType.TIME) {
					PersistentParallel6.runAndFinish(values.entrySet(), (entry) -> {
						for (AbstractObject ao : entry.getValue()) {
							CHMHelper.getFromCHM(map, ao, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(entry.getKey());

							TimeObject to = (TimeObject) ao;
							if (to.getPrecision() == Precision.T11_days) {
								TimeObject toDay = new TimeObject(to.getYear(), to.getMonth(), to.getDay(), 0, 0, 0, 0, Precision.T11_days);
								TimeObject toMonth = new TimeObject(to.getYear(), to.getMonth(), 0, 0, 0, 0, 0, Precision.T10_months);
								TimeObject toYear = new TimeObject(to.getYear(), 0, 0, 0, 0, 0, 0, Precision.T09_years);

								CHMHelper.getFromCHM(map, toDay, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(entry.getKey());
								CHMHelper.getFromCHM(map, toMonth, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(entry.getKey());
								CHMHelper.getFromCHM(map, toYear, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(entry.getKey());

							}
							if (to.getPrecision() == Precision.T10_months) {
								TimeObject toMonth = new TimeObject(to.getYear(), to.getMonth(), 0, 0, 0, 0, 0, Precision.T10_months);
								TimeObject toYear = new TimeObject(to.getYear(), 0, 0, 0, 0, 0, 0, Precision.T09_years);
								CHMHelper.getFromCHM(map, toMonth, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(entry.getKey());
								CHMHelper.getFromCHM(map, toYear, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(entry.getKey());
							}

							if (to.getPrecision() == Precision.T09_years) {
								TimeObject toYear = new TimeObject(to.getYear(), 0, 0, 0, 0, 0, 0, Precision.T09_years);
								CHMHelper.getFromCHM(map, toYear, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(entry.getKey());
							}
						}

					}, this.getClass());
				} else {
					PersistentParallel6.runAndFinish(values.entrySet(), (entry) -> {
						for (AbstractObject ao : entry.getValue()) {
							CHMHelper.getFromCHM(map, ao, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(entry.getKey());
						}
					}, this.getClass());
				}

				ConcurrentHashMap<AbstractObject, List<InstanceId>> mapFinal = new ConcurrentHashMap<>();

				PersistentParallel6.runAndFinish(map.entrySet(), (entry) -> {
					List<InstanceId> list = new ArrayList<>(entry.getValue());
					Collections.sort(list, (i1, i2) -> {
						return i1.getInstanceId().compareTo(i2.getInstanceId());
					});
					mapFinal.put(entry.getKey(), list);
				}, this.getClass());

				this.map = Collections.unmodifiableMap(mapFinal);
			}
			// ValueType.TIME;
			// ValueType.INSTANCE_REFERENCE;
			// ValueType.CATEGORICAL_INT;
			// ValueType.QUANTITY;
			// ValueType.STRING;
		}

		public List<InstanceId> get(AbstractObject ao) {

			if (valueType == ValueType.QUANTITY) {
				QuantityObject qo = (QuantityObject) ao;
				Entry<Double, List<InstanceId>> entry = percentileMap.floorEntry(qo.getAmount());
				if (entry != null) {
					List<InstanceId> list = entry.getValue();
					return list;
				} else {
					return null;
				}
			} else if (valueType == ValueType.INSTANCE_REFERENCE && ao.getType() == ValueType.STRING) {
				Optional<Map<InstanceId, Double>> matches = labelMatcher.getMatches(Sets.newHashSet(((StringObject) ao).getString()), targetClassIds,
						MatchStrictness.STRICT);
				if (matches.isPresent()) {
					List<InstanceId> list = new ArrayList<>();
					for (InstanceId instanceId : matches.get().keySet()) {
						List<InstanceId> innerList = map.get(instanceId);
						if (innerList != null) {
							list.addAll(innerList);
						}

					}
					Collections.sort(list, (i1, i2) -> {
						return i1.getInstanceId().compareTo(i2.getInstanceId());
					});
					return list;
				} else {
					return null;
				}
			} else if (valueType == ValueType.TIME) {

				TimeObject to = (TimeObject) ao;

				List<InstanceId> list = new ArrayList<>();

				if (to.getPrecision() == Precision.T11_days) {
					TimeObject toDay = new TimeObject(to.getYear(), to.getMonth(), to.getDay(), 0, 0, 0, 0, Precision.T11_days);
					TimeObject toMonth = new TimeObject(to.getYear(), to.getMonth(), 0, 0, 0, 0, 0, Precision.T10_months);
					TimeObject toYear = new TimeObject(to.getYear(), 0, 0, 0, 0, 0, 0, Precision.T09_years);

					List<InstanceId> toDayList = map.get(toDay);
					if (toDayList != null) {
						list.addAll(toDayList);
					}

					List<InstanceId> toMonthList = map.get(toMonth);
					if (toMonthList != null) {
						list.addAll(toMonthList);
					}

					List<InstanceId> toYearList = map.get(toYear);
					if (toYearList != null) {
						list.addAll(toYearList);
					}

				} else if (to.getPrecision() == Precision.T10_months) {
					TimeObject toMonth = new TimeObject(to.getYear(), to.getMonth(), 0, 0, 0, 0, 0, Precision.T10_months);
					TimeObject toYear = new TimeObject(to.getYear(), 0, 0, 0, 0, 0, 0, Precision.T09_years);
					List<InstanceId> toMonthList = map.get(toMonth);
					if (toMonthList != null) {
						list.addAll(toMonthList);
					}

					List<InstanceId> toYearList = map.get(toYear);
					if (toYearList != null) {
						list.addAll(toYearList);
					}

				} else if (to.getPrecision() == Precision.T09_years) {
					TimeObject toYear = new TimeObject(to.getYear(), 0, 0, 0, 0, 0, 0, Precision.T09_years);
					List<InstanceId> toYearList = map.get(toYear);
					if (toYearList != null) {
						list.addAll(toYearList);
					}
				}
				return list;
			} else {
				List<InstanceId> list = map.get(ao);
				return list;
			}

		}
	}
}
