package expansion.set.entitymatching.matchers;

import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.Map;

import javax.inject.Inject;

import expansion.set.entitymatching.matchers.strategies.EntityMatcher_Attribute;
import expansion.set.entitymatching.matchers.strategies.EntityMatcher_BOW;
import expansion.set.entitymatching.matchers.strategies.EntityMatcher_ImplicitAttribute;
import expansion.set.entitymatching.matchers.strategies.EntityMatcher_Indegree;
import expansion.set.entitymatching.matchers.strategies.EntityMatcher_Indegree_New;
import expansion.set.entitymatching.matchers.strategies.EntityMatcher_Label;
import expansion.set.entitymatching.matchers.strategies.EntityMatcher_PrimeLabel;
import expansion.set.entitymatching.matchers.strategies.EntityMatcher_Type;

public class EntityMatcherRepository {

	@Inject
	protected EntityMatcher_Label entityMatcher_Label;

	@Inject
	protected EntityMatcher_PrimeLabel entityMatcher_PrimeLabel;

	@Inject
	protected EntityMatcher_Attribute entityMatcher_Attribute;

	@Inject
	protected EntityMatcher_Type entityMatcher_Type;

	@Inject
	protected EntityMatcher_BOW entityMatcher_BOW;

	@Inject
	protected EntityMatcher_Indegree entityMatcher_Indegree;
	
	@Inject
	protected EntityMatcher_Indegree_New entityMatcher_Indegree_new;

	@Inject
	protected EntityMatcher_ImplicitAttribute entityMatcher_ImplicitAttribute;

	private volatile Map<Class<? extends EntityMatcher>, EntityMatcher> repository;

	@Inject
	public EntityMatcherRepository() {

	}

	@SuppressWarnings("unchecked")
	public synchronized void init() {
		if (repository == null) {

			this.repository = new HashMap<>();

			try {
				for (Field field : this.getClass().getDeclaredFields()) {
					field.setAccessible(true); // You might want to set modifier
												// to public first.
					Class<?> clazz = field.getType();
					if (EntityMatcher.class.isAssignableFrom(clazz)) {
						EntityMatcher value = (EntityMatcher) field.get(this);
						this.repository.put((Class<? extends EntityMatcher>) clazz, value);

					}

				}
			} catch (IllegalArgumentException | IllegalAccessException e) {
				throw new RuntimeException(e);
			}

		}

	}

	public EntityMatcher get(Class<? extends EntityMatcher> clazz) {
		if (repository == null)
			init();

		if (clazz == null) {
			throw new IllegalStateException("Asking for Entity Matcher of class null is not allowed!");
		} else {
			EntityMatcher matcher = this.repository.get(clazz);
			if (matcher == null) {
				throw new IllegalStateException("No Entity Matcher was registered for Class: " + clazz.getSimpleName());
			} else {
				return matcher;
			}
		}
	}
}
