package expansion.set.entitymatching.matchers.strategies;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.QuantityObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;
import expansion.set.creation.Entity;
import expansion.set.entitymatching.candidateselection.CandidateSelectionResults;
import expansion.set.entitymatching.candidateselection.CandidateSelectionStrategy;
import expansion.set.entitymatching.matchers.EntityMatcher;
import expansion.set.entitymatching.matchers.EntityMatcherConfig;

public class EntityMatcher_Attribute implements EntityMatcher {

	@Inject
	protected expansion.set.evaluation.matching.EntityMatcher entityMatcher;

	@Inject
	protected ValueEvaluator valueEvaluator;

	@Inject
	public EntityMatcher_Attribute() {
	}

	@Override
	public List<WeightedScore> run(EntityMatcherConfig config, Entity entity, InstanceId instanceId,
			Map<CandidateSelectionStrategy, CandidateSelectionResults> csResults) {

		List<WeightedScore> ret = new ArrayList<>();

		for (Entry<QualifiedPropertyId, FusedValue> value : entity.getValues().entrySet()) {

			Double singleScore = value.getValue().getSingleScore() / value.getValue().getSingleContributors();

			singleScore = Math.min(1.0, singleScore);

			Double minScore = null;
			String propertyIdString = value.getKey().getPropertyId().getPropertyId();
			String readableString = value.getKey().getPropertyId().readableString();
			List<FusedValue> fvList = new ArrayList<>();

			if (value.getValue().getSingle() instanceof StringObject) {
				String str = ((StringObject) value.getValue().getSingle()).getString();
				if (str.length() < 3 || str.trim().equals("") || str.trim().equals("??") || entityMatcher.isOnlyOneChar(str)) {
					minScore = 0.0;
				} else if (value.getValue().getSingleScore() != null) {
					String[] result = ((StringObject) value.getValue().getSingle()).getString().split(";");
					if (result.length > 1) {
						for (int i = 0; i < result.length; i++) {
							fvList.add(new FusedValue(new StringObject(result[i].trim()), value.getValue().getSingleScore(),
									value.getValue().getSingleContributors()));
						}
					}
				}
			}

			if (value.getValue().getSingle() instanceof QuantityObject) {
				QuantityObject qo = ((QuantityObject) value.getValue().getSingle());
				if (Math.abs(qo.getValue()) < 0.01) {
					minScore = 0.0;
				}
			}

			Double maxInnerScore = null;
			double propertyWeight = 1.0;
			// double propertyWeight = entityMatcher.getScoreFromCache(new
			// ClassPropertyId(entity.getClassId(),
			// value.getKey().getPropertyId()));

			if (!fvList.isEmpty()) {

				Double maxScore = null;
				boolean isCorrectBoolean = false;
				// Double correctWeight = null;
				Double maxFvScore = null;
				Double maxCorrectWeight = null;

				for (FusedValue fv : fvList) {
					double fvSingleScore = fv.getSingleScore() / fv.getSingleContributors();

					fvSingleScore = Math.min(1.0, fvSingleScore);

					Optional<Boolean> isCorrect = valueEvaluator.correct(new QualifiedTripleId(entity.getClassId(), instanceId, value.getKey()),
							fv);
					Optional<Double> scoreOptional = valueEvaluator.score(new QualifiedTripleId(entity.getClassId(), instanceId, value.getKey()), fv);

					if (isCorrect.isPresent()) {
						double score = 0.0;
						Double correctWeight = null;
						if (isCorrect.get()) {
							isCorrectBoolean = true;
							score = 1.0;
							if (fv.getSingle().getType() == ValueType.TIME) {
								if (correctWeight == null || scoreOptional.get() > correctWeight) {
									correctWeight = scoreOptional.get();
								}
							}

							// if(value.getKey().get)
						}

						// if(singleScore != null)
						// score = score;
						double fvScore;
						fvScore = score * singleScore;

						if (maxScore == null || (fvScore > maxScore && (maxFvScore == null || fvSingleScore > maxFvScore))) {
							maxScore = fvScore;
							maxFvScore = fvSingleScore;
							maxCorrectWeight = correctWeight;
						}
					}

					// if (scoreOptional.isPresent()) {
					// double score = scoreOptional.get();
					// // if(singleScore != null)
					// // score = score;
					// double fvScore;
					// fvScore = score * singleScore;
					//
					// if (maxScore == null || fvScore > maxScore) {
					// maxScore = fvScore;
					// }
					// }
				}

				if (maxScore != null) {
					if (minScore == null || maxScore > 0.99) {
						if (isCorrectBoolean) {
							double weight = propertyWeight * maxFvScore;

							if (maxCorrectWeight != null) {
								weight = propertyWeight * maxCorrectWeight * maxFvScore;
							}
							ret.add(new WeightedScore(maxScore, weight, readableString, true));
						} else {
							ret.add(new WeightedScore(maxScore, propertyWeight * maxFvScore, readableString, true));
						}
					}
				}
			} else {
				Optional<Boolean> isCorrect = valueEvaluator.correct(new QualifiedTripleId(entity.getClassId(), instanceId, value.getKey()), value.getValue());
				Optional<Double> scoreOptional = valueEvaluator.score(new QualifiedTripleId(entity.getClassId(), instanceId, value.getKey()), value.getValue());

				if (isCorrect.isPresent()) {
					double score = scoreOptional.get();

					double fvScore = score * singleScore;

					if (isCorrect.get()) {

						double weight = propertyWeight * singleScore;

						if (value.getValue().getSingle().getType() == ValueType.TIME) {
							weight = propertyWeight * scoreOptional.get() * singleScore;
						}

						ret.add(new WeightedScore(1.0, weight, readableString, true));
					} else {
						if (minScore == null) {
							ret.add(new WeightedScore(0.0, propertyWeight * singleScore, readableString, false));
						} else {
							// IGNORE AS MINSCORE IS SET, WHICH MEANS BAD VALUE
						}
					}

					// if (minScore != null ) {
					// ret.add(new WeightedScore(fvScore,
					// Math.min(Math.max(fvScore, minScore), propertyWeight),
					// propertyIdString));
					// } else {
					// ret.add(new WeightedScore(fvScore, propertyWeight,
					// propertyIdString));
					// }
				}

			}

		}

		return ret;
	}

	@Override
	public void prepare(EntityMatcherConfig config, List<Entity> entities, Set<InstanceId> candidatesSet) {
		// TODO Auto-generated method stub

	}


}
