package expansion.set.entitymatching.matchers.strategies;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.uni_mannheim.dws.expansion.data.kb.abstracts.AbstractsRepository;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.creation.Entity;
import expansion.set.entitymatching.candidateselection.CandidateSelectionResults;
import expansion.set.entitymatching.candidateselection.CandidateSelectionStrategy;
import expansion.set.entitymatching.matchers.EntityMatcher;
import expansion.set.entitymatching.matchers.EntityMatcherConfig;
import expansion.set.strategies.newstrategies.BagOfWords_Component;

public class EntityMatcher_BOW implements EntityMatcher {

	@Inject
	protected BagOfWords_Component bowComponent;

	@Inject
	protected AbstractsRepository abstractsRepository;

	@Inject
	protected KnowledgeBase knowledgeBase;

	public static SimpleLockableConcurrentCache<RowClusterUUID, Map<String, Integer>> bowMapCache = new SimpleLockableConcurrentCache<>();

	public EntityMatcher_BOW() {

	}

	@Override
	public List<WeightedScore> run(EntityMatcherConfig config, Entity entity, InstanceId instanceId,
			Map<CandidateSelectionStrategy, CandidateSelectionResults> csResults) {
		// Map<TableRowId, Map<String, Integer>> bowMap =
		// bowMapCache.get(config.getClustererInput().getUuid()).get(entity.getRowClusterUUID());

		Map<String, Integer> rowBowMap = bowMapCache.get(entity.getRowClusterUUID());

		Map<String, Integer> instanceBowMap = instanceBowRowMap.get(instanceId);
		if (instanceBowMap == null) {
			instanceBowMap = getInstanceBowMap(instanceId, config.getNonTailProperties());
			instanceBowRowMap.add(instanceId, instanceBowMap);
		}
		if (rowBowMap.size() == 0 || instanceBowMap.size() == 0) {
			List<WeightedScore> ret = new ArrayList<>();
			ret.add(new WeightedScore(0.0, 1.0));
			return ret;
		} else {
			List<WeightedScore> ret = new ArrayList<>();
			double score = score(rowBowMap, instanceBowMap);
			ret.add(new WeightedScore(score, 1.0));
			if (score > 0.5) {
				int sdf = 123;
			}
			return ret;
		}

	}

	@Override
	public void prepare(EntityMatcherConfig config, List<Entity> entities, Set<InstanceId> candidatesSet) {

		ConcurrentHashMap<RowClusterUUID, Map<TableRowId, Map<String, Integer>>> rowBowCache = new ConcurrentHashMap<>();

		ConcurrentHashMap<TableRowId, RowClusterUUID> firstStep = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(entities, (e) -> {
			for (TableRowId rowId : e.getRowCluster().getRows()) {
				firstStep.put(rowId, e.getRowClusterUUID());
			}
		}, this.getClass());

		PersistentParallel6.runAndFinish(firstStep.entrySet(), (e) -> {
			CHMHelper.getFromCHM(rowBowCache, e.getValue(), () -> new ConcurrentHashMap<>()).put(e.getKey(),
					bowComponent.getBagOfWords(e.getKey(), false, true));
		}, this.getClass());

		// ConcurrentHashMap<RowClusterUUID, Map<String, Integer>> rowFinalCache
		// = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(rowBowCache.entrySet(), (e) -> {
			// rowFinalCache.put(e.getKey(), getRowBowMap(e.getValue()));
			bowMapCache.add(e.getKey(), getRowBowMap(e.getValue()));
		}, this.getClass());



		ConcurrentHashMap<InstanceId, Map<String, Integer>> instanceCache = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(candidatesSet, (instanceId) -> {
			if (!instanceBowRowMap.hasCacheFor(instanceId)) {
				instanceBowRowMap.add(instanceId, getInstanceBowMap(instanceId, config.getNonTailProperties()));
			}

		}, this.getClass());



	}

	public void clean() {
		bowMapCache.clear();
	}

	public static SimpleLockableConcurrentCache<InstanceId, Map<String, Integer>> instanceBowRowMap = new SimpleLockableConcurrentCache<>();

	private void addToBowMap(Map<String, Integer> map, String s, int count) {
		if(map.containsKey(s)) {
			Integer c= map.get(s);
			c = c + count;
			map.put(s, c);
		}
		else {
			map.put(s, count);
		}
	}

	public double score(Map<String, Integer> row, Map<String, Integer> instance) {
		Map<String, Integer> smallerSet = row;
		Map<String, Integer> largerSet = instance;
		if (instance.size() < row.size()) {
			smallerSet = instance;
			largerSet = row;
		}
		double intersection = 0.0;

		for (String key : smallerSet.keySet()) {
			if (largerSet.containsKey(key)) {
				intersection++;
			}
		}

		double score = intersection / smallerSet.size();
		return score;
	}

	private Map<String, Integer> getInstanceBowMap(InstanceId instanceId, Set<PropertyId> nonTailProperties) {


		Map<String, Integer> abstractBowMap = abstractsRepository.getBowMap(instanceId);
		Map<String, Integer> bowMap = new HashMap<>(abstractBowMap);

			{
				Set<String> labels = knowledgeBase.getLabels(instanceId);
				if (labels == null) {
					labels = new HashSet<>();
				} else {
					labels = new HashSet<>(labels);
				}
				labels.add(instanceId.getReadableString());

				for (String s : labels) {
					List<String> tokens = StringNormalizer.tokenise(s, false, false);
					for (String token : tokens) {
						addToBowMap(bowMap, token, 1);

					}
				}
			}

		for(PropertyId propertyId : nonTailProperties) {
			if(knowledgeBase.hasValue(instanceId, propertyId)) {
				QualifiedValueList qvl = knowledgeBase.getQualifiedValueList(instanceId, propertyId);
				for(QualifiedValueHolder claim : qvl.getClaims()) {
					AbstractObject ao = claim.getValue();
					switch(ao.getType()) {
					 
					case COORDINATE:
						int sdf = 123;
						break;
					 
					case INSTANCE_REFERENCE:
						InstanceId instanceIdQVL = (InstanceId) ao;
						Set<String> labels = knowledgeBase.getLabels(instanceIdQVL);
						if(labels == null) {
							labels = new HashSet<>();
						}
						else {
							labels = new HashSet<>(labels);
						}
						labels.add(instanceIdQVL.getReadableString());
						
							for(String s : labels) {
								List<String> tokens = StringNormalizer.tokenise(s, false, false);
								for (String token : tokens) {
									addToBowMap(bowMap, token, 1);

								}
						}
						break;
				 
					case QUANTITY:
						addToBowMap(bowMap, ao.toString(), 1);
						break;
						case CATEGORICAL_INT:

							addToBowMap(bowMap, ao.toString(), 1);
							break;
					case STRING:
						addToBowMap(bowMap, ao.toString(), 1);
						break;
					case TIME:
						addToBowMap(bowMap, ao.toString(), 1);
							addToBowMap(bowMap, String.valueOf(TimeObject.getYear(ao)), 1);
						break;
					 
					default:
						break;
					
					}
				}
			}
		}
		
		return bowMap;


	}

	// private Map<String, Integer> getInstanceBowMap(InstanceId instanceId) {
	// return abstractsRepository.getBowMap(instanceId);
	// }

	private Map<String, Integer> getRowBowMap(Map<TableRowId, Map<String, Integer>> bowMap) {

		Map<String, MutableInt> ret = new HashMap<>();

		for (TableRowId rowId : bowMap.keySet()) {
			Map<String, Integer> bow = bowMap.get(rowId);
			for (Entry<String, Integer> bowEntry : bow.entrySet()) {
				CHMHelper.getFromMap(ret, bowEntry.getKey(), () -> new MutableInt()).add(bowEntry.getValue());
				;
			}
		}

		return ret.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().intValue()));
	}


}
