package expansion.set.entitymatching.matchers.strategies;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;

import com.google.common.collect.BiMap;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.uni_mannheim.dws.expansion.data.kb.degree.DegreeRepository;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import expansion.set.creation.Entity;
import expansion.set.entitymatching.candidateselection.CandidateSelectionResults;
import expansion.set.entitymatching.candidateselection.CandidateSelectionStrategy;
import expansion.set.entitymatching.matchers.EntityMatcher;
import expansion.set.entitymatching.matchers.EntityMatcherConfig;

public class EntityMatcher_Indegree_New implements EntityMatcher {

	@Inject
	protected DegreeRepository degreeRepository;

	@Inject
	protected KnowledgeBase knowledgeBase;

	private static final ConcurrentHashMap<ClassId, Integer> maxMap = new ConcurrentHashMap<>();

	private static final ConcurrentHashMap<ClassId, Double> logMap = new ConcurrentHashMap<>();

	@Override
	public List<WeightedScore> run(EntityMatcherConfig config, Entity entity, InstanceId instanceId,
			Map<CandidateSelectionStrategy, CandidateSelectionResults> csResults) {
		List<WeightedScore> ret = new ArrayList<>();
		ret.add(new WeightedScore(compute(degreeRepository.getNumIndegree(instanceId), entity.getClassId()), 1.0));
		return ret;
	}

	private double compute(int numIndegree, ClassId classId) {
		double deg = (Math.log(numIndegree) + 1) / logMap.get(classId);
		if (numIndegree == 0) {
			deg = 0.0;
		}
		return deg;

	}

	@Override
	public void prepare(EntityMatcherConfig config, List<Entity> entities, Set<InstanceId> candidatesSet) {

		for (Entry<ClassId, Set<ClassId>> entry : config.getAllParents().entrySet()) {
			if (!logMap.containsKey(entry.getKey())) {
				int max = getMax(entry.getKey());
				for (ClassId classId : entry.getValue()) {
					int innerMax = getMax(classId);
					if (innerMax > max) {
						max = innerMax;
					}
				}

				maxMap.put(entry.getKey(), max);
				logMap.put(entry.getKey(), Math.log(max) + 1.0);
			}
		}

	}

	private int getMax(ClassId key) {
		BiMap<InstanceId, Integer> instances = knowledgeBase.getInstances(key);
		return degreeRepository.getNumIndegree(instances.keySet()).values().stream().mapToInt(Integer::intValue).max().getAsInt();
	}

}
