package expansion.set.entitymatching.modelLearning;

import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.Map;

import javax.inject.Inject;

import expansion.set.entitymatching.modelLearning.simpleWeights.SimpleEntityMatchingWeightLearner;
import expansion.set.entitymatching.modelLearning.weights.EntityMatchingWeightLearner;
import expansion.set.entitymatching.modelLearning.weka.randomforest.EntityMatchingWekaRFDescretizedWithSWLearner;
import expansion.set.entitymatching.modelLearning.weka.randomforest.EntityMatchingWekaRFLearner;
import expansion.set.entitymatching.modelLearning.weka.randomforest.EntityMatchingWekaRFNewLearner;
import expansion.set.entitymatching.modelLearning.weka.randomforest.EntityMatchingWekaRFNew_ET_Learner;
import expansion.set.entitymatching.modelLearning.weka.randomforest.EntityMatchingWekaRFWithSWLearner;
import expansion.set.entitymatching.modelLearning.weka.regression.EntityMatchingWekaLinearLearner;
import expansion.set.entitymatching.modelLearning.weka.regression.EntityMatchingWekaLogisticLearner;

public class EntityMatchingModelLearnerRepository {

	@Inject
	EntityMatchingWeightLearner EntityMatcherWeightLearner;

	@Inject
	EntityMatchingWekaRFLearner EntityMatcherWekaRFLearner;

	@Inject
	EntityMatchingWekaRFWithSWLearner EntityMatchingWekaRFWithSWLearner;

	@Inject
	EntityMatchingWekaLinearLearner EntityMatchingWekaLinearLearner;

	@Inject
	SimpleEntityMatchingWeightLearner SimpleEntityMatchingWeightLearner;

	@Inject
	EntityMatchingWekaRFDescretizedWithSWLearner EntityMatchingWekaRFDescretizedWithSWLearner;

	@Inject
	EntityMatchingWekaRFNewLearner EntityMatchingWekaRFNewLearner;

	@Inject
	EntityMatchingWekaRFNew_ET_Learner EntityMatchingWekaRFNew_ET_Learner;

	@Inject
	EntityMatchingWekaLogisticLearner EntityMatchingWekaLogisticLearner;
	private volatile Map<Class<? extends EntityMatchingModelLearnerRunner>, EntityMatchingModelLearnerRunner> repository;

	public EntityMatchingModelLearnerRepository() {

	}

	@SuppressWarnings("unchecked")
	public synchronized void init() {
		if (repository == null) {

			this.repository = new HashMap<>();

			try {
				for (Field field : this.getClass().getDeclaredFields()) {
					field.setAccessible(true); // You might want to set modifier
												// to public first.
					Class<?> clazz = field.getType();
					if (EntityMatchingModelLearnerRunner.class.isAssignableFrom(clazz)) {
						EntityMatchingModelLearnerRunner value = (EntityMatchingModelLearnerRunner) field.get(this);
						this.repository.put((Class<? extends EntityMatchingModelLearnerRunner>) clazz, value);

					}

				}
			} catch (IllegalArgumentException | IllegalAccessException e) {
				throw new RuntimeException(e);
			}

		}

	}

	public EntityMatchingModelLearnerRunner get(Class<? extends EntityMatchingModelLearnerRunner> clazz) {
		if (repository == null)
			init();

		if (clazz == null) {
			throw new IllegalStateException("Asking for Clusterer of class null is not allowed!");
		} else {
			EntityMatchingModelLearnerRunner learner = this.repository.get(clazz);
			if (learner == null) {
				throw new IllegalStateException("No Clusterer was registered for Class: " + clazz.getSimpleName());
			} else {
				return learner;
			}
		}
	}
}
