package expansion.set.entitymatching.modelLearning;

import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import expansion.set.clustering.model.RowClusterUUID;

public class EntityMatchingUpsamplingHelper {

	private final double upsampleNew;
	private final double upsampleExisting;
	private final int newCount;
	private final int existingCount;

	public EntityMatchingUpsamplingHelper(Set<InstanceId> newEntities, Map<RowClusterUUID, Optional<InstanceId>> gsMapped) {

		int existingCount = 0;
		int newCount = 0;
		for (Entry<RowClusterUUID, Optional<InstanceId>> entityEntry : gsMapped.entrySet()) {
			if (entityEntry.getValue().isPresent()) {
				if (newEntities.contains(entityEntry.getValue().get())) {
					newCount++;
				}
				else {
					existingCount++;
				}
			} else {
				throw new IllegalStateException();
			}
		}

		double max = 0.0;

		max = newCount;
		if (existingCount > max) {
			max = existingCount;
		}

		double upsampleNew = 1.0;
		double upsampleExisting = 1.0;

		if (newCount != 0) {
			upsampleNew = max / newCount;
		}

		if (existingCount != 0) {
			upsampleExisting = max / upsampleExisting;
		}

		this.upsampleExisting = upsampleExisting;
		this.upsampleNew = upsampleNew;
		this.newCount = newCount;
		this.existingCount = existingCount;

	}

	public double getUpsampledAccuracy(int correctNew, int correctExisting) {
		return ((correctNew * upsampleNew) + (correctExisting * upsampleExisting)) / ((newCount * upsampleNew) + (existingCount * upsampleExisting));
	}

	public double getUpsampleNew() {
		return upsampleNew;
	}

	public double getUpsampleExisting() {
		return upsampleExisting;
	}

}
