package expansion.set.entitymatching.modelLearning.weights;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;
import expansion.set.entitymatching.EntityMatcherEquationResult;
import expansion.set.entitymatching.EntityMatcherSingleResults;
import expansion.set.entitymatching.matchers.EntityMatcherStrategy;
import expansion.set.entitymatching.modelLearning.weightsGeneric.EM_WeightsGenericEquation;

public class EntityMatcherEquation implements EM_WeightsGenericEquation {
	
	@Inject
	public EntityMatcherEquation() {
	}


	public Tri<EntityMatcherEquationResult, InstanceId, Double> runEquation(Map<InstanceId, EntityMatcherSingleResults> candidates,
			WeightsRepository weightsRepository, OptimizableClass optimizableClazz, List<EntityMatcherStrategy> emStrategies) {
		if (candidates == null || candidates.isEmpty()) {
			return new Tri<>(EntityMatcherEquationResult.NEW, null, null);
		} else {
			Pair<InstanceId, Double> match = getUnThresholdedMatch(weightsRepository, optimizableClazz, emStrategies, candidates).getSecond();

			double newThreshold = weightsRepository.getOrDefault(optimizableClazz.getWeight("THRESHOLD_NEW", 0.5));

			double matchThreshold = newThreshold + ((1.0 - newThreshold) * weightsRepository.getOrDefault(optimizableClazz.getWeight("THRESHOLD_MATCH", 0.75)));

			if (match.getSecond() <= newThreshold) {
				return new Tri<>(EntityMatcherEquationResult.NEW, null, null);
			} else if (match.getSecond() < matchThreshold) {
				return new Tri<>(EntityMatcherEquationResult.BAD, match.getFirst(), match.getSecond());
			} else {

				return new Tri<>(EntityMatcherEquationResult.EXISTING, match.getFirst(), match.getSecond());

			}
		}
	}

	private Pair<Map<InstanceId, Double>, Pair<InstanceId, Double>> getUnThresholdedMatch(WeightsRepository weightsRepository,
			OptimizableClass optimizableClazz,
			List<EntityMatcherStrategy> emStrategies, Map<InstanceId, EntityMatcherSingleResults> value) {

		Map<EntityMatcherStrategy, Double> diffRankMap = new HashMap<>();
		Map<EntityMatcherStrategy, Double> minRankMap = new HashMap<>();

		for (EntityMatcherStrategy emStrategy : emStrategies) {

			if (emStrategy.isRanked()) {
				Double maxRank = null;
				Double minRank = null;
				for (Entry<InstanceId, EntityMatcherSingleResults> instanceEntry : value.entrySet()) {
					for (WeightedScore ws : instanceEntry.getValue().getScores().get(emStrategy)) {
						if (maxRank == null || ws.getScore() > maxRank) {
							maxRank = ws.getScore();
						}
						if (minRank == null || ws.getScore() < minRank) {
							minRank = ws.getScore();
						}
					}
				}
				if (maxRank != null && minRank != null) {
					double diff = maxRank - minRank;
					if (diff > 0.01) {
						diffRankMap.put(emStrategy, diff);
						minRankMap.put(emStrategy, minRank);
					}

				}
			}
		}

		InstanceId maxMatch = null;
		double maxScore = 0.0;

		Map<InstanceId, Double> scoreMap = new HashMap<>();

		for (Entry<InstanceId, EntityMatcherSingleResults> instanceEntry : value.entrySet()) {
			double sum = 0.0;
			double count = 0.0;

			for (EntityMatcherStrategy emStrategy : emStrategies) {
				List<WeightedScore> values = instanceEntry.getValue().getScores().get(emStrategy);

				double innerSum = 0.0;
				double innerCount = 0.0;

				if (values == null || values.size() == 0) {
					// double missingScore =
					// weightsRepository.getOrDefault(optimizableClazz.getWeight(emStrategy.getName()
					// + "_MISSING", 0.5));
					// innerSum += missingScore;
					// innerCount += 1.0;
				} else {
					for (WeightedScore score : values) {
						double suffixWeight = 1.0;
						if (score.getSuffix().isPresent()) {
							suffixWeight = weightsRepository
									.getOrDefault(optimizableClazz.getWeight(emStrategy.getName() + "_INDIVIDUAL", 0.5)); // +
																															// score.getSuffix().get()
						}
						double innerInnerScore = score.getScore();
						if (emStrategy.isRanked()) {
							if (diffRankMap.containsKey(emStrategy)) {
								innerInnerScore = (innerInnerScore - minRankMap.get(emStrategy)) / diffRankMap.get(emStrategy);
								int sdf = 123;
							} else {
								innerInnerScore = 1.0;
							}
						}

						innerSum += innerInnerScore * score.getWeight() * suffixWeight;
						innerCount += score.getWeight() * suffixWeight;

					}
				}

				// double innerScore = 0.0;
				// if (innerCount > 0.001) {
				// innerScore = innerSum / innerCount;
				// }

				double weight = weightsRepository.getOrDefault(optimizableClazz.getWeight(emStrategy.getName() + "_WEIGHT", 0.5));

				if (innerCount < 1.0) {
					double emptyScore = weightsRepository.getOrDefault(optimizableClazz.getWeight(emStrategy.getName() + "_EMPTY", 1.0));
					innerSum += (1.0 - innerCount) * emptyScore;
				}
				if (innerSum >= 1.0) {
					sum += 1.0 * weight;
				} else {
					sum += innerSum * weight;
				}

				// if (false) {
				// double minimumScore =
				// weightsRepository.getOrDefault(optimizableClazz.getWeight(emStrategy.getName()
				// + "_MINIMUM", 0.5));
				//
				// if (innerScore > minimumScore) {
				// innerScore = (innerScore - minimumScore) / (1.0 -
				// minimumScore);
				// } else {
				// innerScore = 0.0;
				// }
				// }

				// sum += innerScore * weight;
				count += weight;

			}

			double score = 0.0;
			if (count > 0.0001) {
				score = sum / count;
			}

			scoreMap.put(instanceEntry.getKey(), score);
			if (maxMatch == null || score > maxScore) {
				maxMatch = instanceEntry.getKey();
				maxScore = score;
			}
		}

		return new Pair<>(scoreMap, new Pair<>(maxMatch, maxScore));

	}

	@Override
	public Map<InstanceId, Double> getNormalized(Map<InstanceId, EntityMatcherSingleResults> candidates, WeightsRepository weightsRepository,
			OptimizableClass optimizableClazz, List<EntityMatcherStrategy> emStrategies) {
		if (candidates == null || candidates.isEmpty()) {
			return null;
		} else {

			double newThreshold = weightsRepository.getOrDefault(optimizableClazz.getWeight("THRESHOLD_NEW", 0.5));

			double matchThreshold = newThreshold + ((1.0 - newThreshold) * weightsRepository.getOrDefault(optimizableClazz.getWeight("THRESHOLD_MATCH", 0.75)));

			Map<InstanceId, Double> match = getUnThresholdedMatch(weightsRepository, optimizableClazz, emStrategies, candidates).getFirst();

			for (Entry<InstanceId, Double> entry : match.entrySet()) {

				if (entry.getValue() <= newThreshold) {
					double ret = -1.0 * ((entry.getValue()) / (newThreshold));
					entry.setValue(Double.isNaN(ret) ? -1.0 : ret);
				} else if (entry.getValue() < matchThreshold) {
					entry.setValue(0.0);
				} else {
					double ret = (entry.getValue() - matchThreshold) / (1.0 - matchThreshold);

					entry.setValue(Double.isNaN(ret) ? 1.0 : ret);

				}
			}

			return match;

		}
	}
}
