package expansion.set.entitymatching.modelLearning.weightsGeneric;

import java.util.Collection;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.WeightsOptimizer;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities.PairLearningData;
import expansion.set.entitymatching.EntityMatcherEquationResult;
import expansion.set.entitymatching.EntityMatcherFullResults;
import expansion.set.entitymatching.EntityMatcherSingleResults;
import expansion.set.entitymatching.modelLearning.EntityMatchingModel;
import expansion.set.entitymatching.modelLearning.EntityMatchingModelLearnerRunner;
import expansion.set.entitymatching.run.EntityMatchingRun;
import expansion.set.evaluation.matching.NewDetectionEvaluator;
import expansion.set.evaluation.matching.ScoredEntity;

public abstract class EM_WeightsGenericLearner

		implements EntityMatchingModelLearnerRunner {

	@Inject
	protected NewDetectionEvaluator newDetectionEvaluator;

	@Inject
	protected WeightsOptimizer weightsParallelOptimizer;

	@Inject
	protected EntityMatchingGenericWeightsOptimizier generic;



	protected abstract EM_WeightsGenericEquation getEquation();

	@Override
	public EntityMatchingModel learnModel(EntityMatchingRun run, Set<InstanceId> newEntities, Map<RowClusterUUID, Optional<InstanceId>> gsMapped,
			EntityMatcherFullResults fullRun) {

		EntityMatchingGenericWeightsOptimizier.EntityMatchingWeightsOptimizationRun optimizableRun = (weightsRepository) -> {

			ConcurrentHashMap<InstanceId, Collection<ScoredEntity>> mappingForExistingClusters = new ConcurrentHashMap<>();

			ConcurrentLinkedQueue<ScoredEntity> newClusters = new ConcurrentLinkedQueue<>();

			PersistentParallel6.runAndFinish(fullRun.getEntities().entrySet(), (entityEntry) -> {
				Map<InstanceId, EntityMatcherSingleResults> candidates = fullRun.getResults().get(entityEntry.getKey());

				Tri<EntityMatcherEquationResult, InstanceId, Double> eqResult = getEquation().runEquation(candidates, weightsRepository,
						run.getOptimizableClass(),
						run.getEntityMatcher());

				switch (eqResult.getA()) {
				case BAD:
					break;
				case EXISTING:
					CHMHelper.getFromCHM(mappingForExistingClusters, eqResult.getB(), () -> new ConcurrentLinkedQueue<>())
							.add(new ScoredEntity(entityEntry.getValue(), eqResult.getC()));
					break;
				case NEW:
					newClusters.add(new ScoredEntity(entityEntry.getValue(), eqResult.getC()));
					break;
				default:
					throw new IllegalStateException();

				}
			}, this.getClass());

			return new Pair<>(mappingForExistingClusters, newClusters);

		};

		EntityMatchingGenericWeightsOptimizier.EntityMatchingPenalizationRun penalization = (weightsRepository) -> {
			return (1.0 - weightsRepository.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_MATCH", 0.75))) / 100.0
					+ (weightsRepository.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_NEW", 0.50))) / 100.0;
		};

		// EntityMatchingUpsamplingHelper upsamplingHelper = new
		// EntityMatchingUpsamplingHelper(gsMapped);

		return new EM_WeightsGenericModel(run, generic.run(gsMapped, newEntities, optimizableRun, penalization), getEquation());
	}

	public EntityMatchingModel learnModel(EntityMatchingRun run, Map<Pair<RowClusterUUID, InstanceId>, PairLearningData> pairs,
			EntityMatcherFullResults fullRun) {
		throw new UnsupportedOperationException();
	}
}
