package expansion.set.entitymatching.modelLearning.weightsGeneric;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.WeightsOptimizer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.entitymatching.modelLearning.EntityMatchingUpsamplingHelper;
import expansion.set.evaluation.matching.NewDetectionEvaluator;
import expansion.set.evaluation.matching.NewDetectionPerformance;
import expansion.set.evaluation.matching.ScoredEntity;

public class EntityMatchingGenericWeightsOptimizier {

	@Inject
	protected NewDetectionEvaluator newDetectionEvaluator;

	@Inject
	protected WeightsOptimizer weightsParallelOptimizer;

	@Inject
	public EntityMatchingGenericWeightsOptimizier() {
		// TODO Auto-generated constructor stub
	}

	@FunctionalInterface
	public static interface EntityMatchingWeightsOptimizationRun {
		Pair<Map<InstanceId, Collection<ScoredEntity>>, Collection<ScoredEntity>> get(WeightsRepository weights);
	}

	@FunctionalInterface
	public static interface EntityMatchingPenalizationRun {
		Double get(WeightsRepository weights);
	}

	public WeightsRepository run(Map<RowClusterUUID, Optional<InstanceId>> gsMapped, Set<InstanceId> newEntities, EntityMatchingWeightsOptimizationRun rr,
			EntityMatchingPenalizationRun penaliztion) {
		return run(gsMapped, newEntities, rr, penaliztion, true);
	}

	public WeightsRepository run(Map<RowClusterUUID, Optional<InstanceId>> gsMapped, Set<InstanceId> newEntities, EntityMatchingWeightsOptimizationRun rr,
			EntityMatchingPenalizationRun penaliztion, boolean useAccuracy) {
		EntityMatchingUpsamplingHelper upsamplingHelper = new EntityMatchingUpsamplingHelper(newEntities, gsMapped);
		return weightsParallelOptimizer.optimize((weightsRepository) -> {

			

			Pair<Map<InstanceId, Collection<ScoredEntity>>, Collection<ScoredEntity>> pair = rr.get(weightsRepository);
			Map<InstanceId, Collection<ScoredEntity>> mappingForExistingClusters = pair.getFirst();
			Collection<ScoredEntity> newClusters = pair.getSecond();


			double total = 0.0;
			double totalCorrect = 0.0;

			int countNew = 0;
			int countExisting = 0;
			int countBad = 0;

			int correctNew = 0;
			int correctExisting = 0;
			int correctBad = 0;

			Map<RowClusterUUID, InstanceId> existingMap = mappingForExistingClusters.entrySet().stream()
					.flatMap((e) -> e.getValue().stream().map((e2) -> new Pair<>(e2.getEntity().getRowClusterUUID(), e.getKey())))
					.collect(Collectors.toMap(Pair::getFirst, Pair::getSecond));
			Set<RowClusterUUID> newSet = newClusters.stream().map((e) -> e.getEntity().getRowClusterUUID()).collect(Collectors.toSet());

			for (Entry<RowClusterUUID, Optional<InstanceId>> entry : gsMapped.entrySet()) {
				if (entry.getValue().isPresent()) {
					if (newEntities.contains(entry.getValue().get())) {
						countNew++;
						if (newSet.contains(entry.getKey())) {
							correctNew++;
						}
					} else {
						countExisting++;
						if (entry.getValue().get().equals(existingMap.get(entry.getKey()))) {
							correctExisting++;
						}
					}
				} else {
					countBad++;
					if (!newSet.contains(entry.getKey()) && !existingMap.containsKey(entry.getKey())) {
						correctBad++;
					}
				}
			}

			double max = 0.0;

			max = countNew;
			if (countExisting > max) {
				max = countExisting;
			}
			if (countBad > max) {
				max = countBad;
			}

			double upsampleNew = 1.0;
			double upsampleExisting = 1.0;
			double upsampleBad = 1.0;

			if (countNew != 0) {
				upsampleNew = max / countNew;
			}

			if (countExisting != 0) {
				upsampleExisting = max / countExisting;
			}

			if (countBad != 0) {
				upsampleBad = max / countBad;
			}

			total += upsampleBad * countBad;
			total += upsampleExisting * countExisting;
			total += upsampleNew * countNew;

			totalCorrect += upsampleBad * correctBad;
			totalCorrect += upsampleExisting * correctExisting;
			totalCorrect += upsampleNew * correctNew;

			double upscaledAccuracy = totalCorrect / total;
			double nonUpscaledAccuracy = (correctBad + correctExisting + correctNew) / ((double) (countBad + countExisting + countNew));
			NewDetectionPerformance accuracy = newDetectionEvaluator.evaluateAccuracy(newEntities, gsMapped,
					mappingForExistingClusters.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> new ArrayList<>(e.getValue()))),
					new ArrayList<>(newClusters));

			// double originalAccuracy = accuracy.getTotalAccuracy();
			// upscaledAccuracy =
			// upsamplingHelper.getUpsampledAccuracy(correctNew,
			// correctExisting);

			double badThreshold = penaliztion.get(weightsRepository);

			double reducedUpscaledAccuracy = upscaledAccuracy - badThreshold;

			if (useAccuracy) {

				// double existingF1 = accuracy.getExistingFmeasure();
				// double newF1 = accuracy.getNewFmeasure();
				//
				// double harmonicF1 = (2 * existingF1 * newF1) / (existingF1 +
				// newF1);
				// return harmonicF1;

				return reducedUpscaledAccuracy;
			} else {
				return accuracy.getAverageFmeasure() - badThreshold;
			}

		});
	}
}
