package expansion.set.entitymatching.modelLearning.weka;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.function.Function;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.DataSetData;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.RecordPrediction;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordFeature;
import de.uni_mannheim.informatik.dws.expansion.utils.learning.weka.WekaLearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepositoryImpl;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities;
import expansion.set.distant_supervision.matching_rules.MatchingRulesWrapperCreator.ND_MR_SCORE_ADAPTER;
import expansion.set.entitymatching.EntityMatcherFullResults;
import expansion.set.entitymatching.EntityMatcherSingleResults;
import expansion.set.entitymatching.EntityMatcherWeightedResults;
import expansion.set.entitymatching.matchers.EntityMatcherStrategy;
import expansion.set.entitymatching.modelLearning.EntityMatchingModel;
import expansion.set.entitymatching.modelLearning.weightsGeneric.EM_WeightsGenericModel;
import expansion.set.entitymatching.run.EntityMatchingRun;
import expansion.set.evaluation.matching.ScoredEntity;
import weka.core.SerializationHelper;

public class EntityMatchingWekaModel extends EntityMatchingModel implements Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	private final WekaLearnedModel model;
	private final DataSetData datasetData;
	private final Set<EntityMatcherStrategy> withConfidence;
	private final EntityMatchingRun run;
	private final boolean numeric;
	private final WeightsRepositoryImpl weights;
	private final EntityMatchingModel simpleWeightsModel;
	private final boolean rangeDescretization;
	private final Set<PropertyId> propertyIds;

	private transient final ND_MR_SCORE_ADAPTER scoreAdapter;


	public ND_MR_SCORE_ADAPTER getScoreAdapter() {
		return scoreAdapter;
	}

	// private final static SimpleEntityMatcherEquation equation = new
	// SimpleEntityMatcherEquation();
	public EntityMatchingWekaModel(byte[] input) {
		ByteArrayInputStream stream = new ByteArrayInputStream(input);
		try {
			EntityMatchingWekaModel readModel = (EntityMatchingWekaModel) weka.core.SerializationHelper.read(stream);
			model = readModel.model;
			datasetData = readModel.datasetData;
			withConfidence = readModel.withConfidence;
			run = readModel.run;
			numeric = readModel.numeric;
			weights = readModel.weights;
			simpleWeightsModel = readModel.simpleWeightsModel;
			if (simpleWeightsModel instanceof EM_WeightsGenericModel) {
				((EM_WeightsGenericModel) simpleWeightsModel).startup();
			}
			rangeDescretization = readModel.rangeDescretization;
			propertyIds = readModel.propertyIds;
			scoreAdapter = null;
		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
	}

	public EntityMatchingWekaModel(EntityMatchingWekaModel readModel, ND_MR_SCORE_ADAPTER adapter) {
		model = readModel.model;
		datasetData = readModel.datasetData;
		withConfidence = readModel.withConfidence;
		run = readModel.run;
		numeric = readModel.numeric;
		weights = readModel.weights;
		simpleWeightsModel = readModel.simpleWeightsModel;
		if (simpleWeightsModel instanceof EM_WeightsGenericModel) {
			((EM_WeightsGenericModel) simpleWeightsModel).startup();
		}
		rangeDescretization = readModel.rangeDescretization;
		propertyIds = readModel.propertyIds;
		scoreAdapter = adapter;

	}

	public double getError() {
		return model.getError();
	}

	@Override
	public byte[] outputModelFile() {
		ByteArrayOutputStream stream = new ByteArrayOutputStream();

		try {

			SerializationHelper.write(stream, this);

			return stream.toByteArray();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
	}

	@Override
	public String toString() {
		return "EntityMatchingWekaModel [model=" + model + ", weights=" + weights + ", simpleWeightsModel=" + simpleWeightsModel + "]";
	}

	public EntityMatchingWekaModel(EntityMatchingRun run, DataSetData datasetData, Set<EntityMatcherStrategy> withConfidence, boolean numeric,
			WekaLearnedModel model, WeightsRepository weights, EntityMatchingModel entityMatchingModel, boolean rangeDescretization,
			Set<PropertyId> propertyIds) {
		this.run = run;
		this.model = model;
		this.datasetData = datasetData;
		this.withConfidence = Collections.unmodifiableSet(new HashSet<>(withConfidence));
		this.numeric = numeric;
		if (weights != null)
			this.weights = new WeightsRepositoryImpl(weights);
		else
			this.weights = null;

		if (entityMatchingModel != null)
			this.simpleWeightsModel = entityMatchingModel;
		else
			this.simpleWeightsModel = null;

		this.rangeDescretization = rangeDescretization;
		this.propertyIds = propertyIds;
		this.scoreAdapter = null;
	}

	public EntityMatchingWekaModel(EntityMatchingRun run, DataSetData datasetData, Set<EntityMatcherStrategy> withConfidence, boolean numeric,
			WekaLearnedModel model, WeightsRepository weights, EntityMatchingModel entityMatchingModel, boolean rangeDescretization,
			Set<PropertyId> propertyIds, ND_MR_SCORE_ADAPTER scoreAdapter) {
		this.run = run;
		this.model = model;
		this.datasetData = datasetData;
		this.withConfidence = Collections.unmodifiableSet(new HashSet<>(withConfidence));
		this.numeric = numeric;
		if (weights != null)
			this.weights = new WeightsRepositoryImpl(weights);
		else
			this.weights = null;

		if (entityMatchingModel != null)
			this.simpleWeightsModel = entityMatchingModel;
		else
			this.simpleWeightsModel = null;

		this.rangeDescretization = rangeDescretization;
		this.propertyIds = propertyIds;
		this.scoreAdapter = scoreAdapter;
	}

	private final transient SimpleLockableConcurrentCache<Map<String, Double>, Double> cache = new SimpleLockableConcurrentCache<>();


	@Override
	public EntityMatcherWeightedResults apply(EntityMatcherFullResults fullRun) {
		Tri<Map<InstanceId, Collection<ScoredEntity>>, Collection<ScoredEntity>, Map<InstanceId, Collection<ScoredEntity>>> tri = getInner(weights, fullRun)
				.getFirst();

		Map<InstanceId, Collection<ScoredEntity>> mappingForExistingClusters = tri.getA();
		Map<InstanceId, Collection<ScoredEntity>> unsure = tri.getC();
		Collection<ScoredEntity> newEntities = tri.getB();

		return new EntityMatcherWeightedResults(
				mappingForExistingClusters.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> new ArrayList<>(e.getValue()))),
				unsure.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> new ArrayList<>(e.getValue()))), new ArrayList<>(newEntities));

	}

	public Pair<Map<InstanceId, Collection<ScoredEntity>>, Collection<ScoredEntity>> get(WeightsRepository weights, EntityMatcherFullResults fullRun) {
		
		Tri<Map<InstanceId, Collection<ScoredEntity>>, Collection<ScoredEntity>, Map<InstanceId, Collection<ScoredEntity>>> tri = getInner(weights, fullRun)
				.getFirst();
		return new Pair<>(tri.getA(), tri.getB());

	}

	public Map<Pair<RowClusterUUID, InstanceId>, Pair<Boolean, Double>> getFullResults(EntityMatcherFullResults fullRun) {
		return getInner(weights, fullRun).getSecond();

	}

	private Pair<Tri<Map<InstanceId, Collection<ScoredEntity>>, Collection<ScoredEntity>, Map<InstanceId, Collection<ScoredEntity>>>, Map<Pair<RowClusterUUID, InstanceId>, Pair<Boolean, Double>>> getInner(
			WeightsRepository weights,
			EntityMatcherFullResults fullRun) {
		ConcurrentHashMap<InstanceId, Collection<ScoredEntity>> mappingForExistingClusters = new ConcurrentHashMap<>();
		ConcurrentHashMap<InstanceId, Collection<ScoredEntity>> unsure = new ConcurrentHashMap<>();
		ConcurrentLinkedQueue<ScoredEntity> newEntities = new ConcurrentLinkedQueue<>();
		
		ConcurrentHashMap<Pair<RowClusterUUID, InstanceId>, Pair<Boolean, Double>> nonAbsoluteScores = new ConcurrentHashMap<>();

		double newThreshold = (weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_NEW", 0.5)));

		// double matchThreshold =
		// weights.getOrDefault(run.getOptimizableClass().getWeight("THRESHOLD_MATCH",
		// 0.5));
		double matchThreshold = 0.0;

		double wekaWeight;
		double nonWekaWeight;
		double nonWekaNormalizer;
		if (simpleWeightsModel != null) {
			wekaWeight = weights.getOrDefault(run.getOptimizableClass().getWeight("WEKA_WEIGHT", 0.5));
			nonWekaNormalizer = weights.getOrDefault(run.getOptimizableClass().getWeight("NON_WEKA_NORMALIZE", 0.5));
			nonWekaWeight = 1.0 - wekaWeight;
		} else {
			wekaWeight = 1.0;
			nonWekaWeight = 0.0;
			nonWekaNormalizer = 0.0;
		}

		double newThresholdFinal = newThreshold * 2.0 - 1.0;

		// double matchThresholdFinal = (newThreshold + ((1.0 - newThreshold) *
		// matchThreshold)) * 2.0 - 1.0;
		double matchThresholdFinal = matchThreshold * 2.0 - 1.0;

		// double newThresholdFinal = -1.0 * newThreshold;
		// double matchThresholdFinal = matchThreshold;
		// double matchThresholdFinal = matchThreshold;



		PersistentParallel6.runAndFinish(fullRun.getEntities().entrySet(), (entityEntry) -> {
			Map<InstanceId, EntityMatcherSingleResults> candidates = fullRun.getResults().get(entityEntry.getKey());
			double finalOutput = -1.0;
			boolean isNew = true;
			boolean isExisting = false;
			InstanceId match = null;

			if (candidates != null) {

				Function<InstanceId, Double> otherModelSupplier = null;
				if (simpleWeightsModel != null) {
					otherModelSupplier = simpleWeightsModel.getOtherModelSupplier(candidates);
				}

				Map<InstanceId, RecordFeature> featureVectors = EntityMatcherWekaLearner.getFeatureVectors(run.getEntityMatcher(), candidates, datasetData,
						withConfidence, rangeDescretization, this.propertyIds);

				List<Tri<RowClusterUUID, InstanceId, Double>> nonMatches = new ArrayList<>();

				for (Entry<InstanceId, EntityMatcherSingleResults> innerEntry : candidates.entrySet()) {

					RecordFeature rf = featureVectors.get(innerEntry.getKey());
					double output;
					if (numeric) {

						// output = model.applyNumeric(rf);

						output = cache.getOrCreateNoBlocking(rf.getFeatures(), () -> {
							double v = model.applyNumeric(rf);
							return v;
						});

					} else {
						output = cache.getOrCreateNoBlocking(rf.getFeatures(), () -> {
							RecordPrediction prediction = model.apply(rf);
							double v;
							if (prediction.getLabel().equals(LearnableRowClusterSimilarities.getLabel(true))) {
								v = prediction.getConfidence();
							} else {
								v = prediction.getConfidence() * -1.0;
							}
							return v;
						});
						// output = v;
					}

					if (simpleWeightsModel != null) {

						double d = otherModelSupplier.apply(innerEntry.getKey());// this.simpleWeightsModel.apply(innerEntry.get);;//equation.normalizedScore(candidates,
																					// simpleWeightsModel,
																					// run.getOptimizableClass(),
																					// run.getEntityMatcher());
						double t;
						if (d > nonWekaNormalizer) {
							t = ((d - nonWekaNormalizer) / (1.0 - nonWekaNormalizer));
						} else {
							t = -1.0 * (1.0 - (d) / (nonWekaNormalizer));
						}
						output = (wekaWeight * output) + (nonWekaWeight * t);
					}

					if (scoreAdapter != null) {
						output = scoreAdapter.get(output, rf);
					}

					double adaptedOutput = Math.abs(1.0 - output) / 2.0;

					if (output < newThresholdFinal) {
						adaptedOutput = -1.0 * (output - newThresholdFinal) / (-1.0 - newThresholdFinal);
					} else {
						adaptedOutput = (output - newThresholdFinal) / (1.0 - newThresholdFinal);
					}

					// if (Math.abs(adaptedOutput) > 1.0) {
					// int sdf = 123;
					// }
					// adaptedOutput = output;
					nonMatches.add(new Tri<>(entityEntry.getKey(), innerEntry.getKey(), adaptedOutput));

					if (output < newThresholdFinal) {
						
						
						// nonAbsoluteScores.put(new
						// Pair<>(entityEntry.getKey(), innerEntry.getKey()),
						// new Pair<>(false, output ));
						
						if (output > matchThresholdFinal) {
							int sdf = 123;
						}
						if (match == null || output > finalOutput) {
							finalOutput = output;
							match = innerEntry.getKey();
						}

					}

					else if (output > matchThresholdFinal) {
						
						// nonAbsoluteScores.put(new
						// Pair<>(entityEntry.getKey(), innerEntry.getKey()),
						// new Pair<>(true, output ));

						
						if ((match == null || output > finalOutput)) {
							finalOutput = output;
							match = innerEntry.getKey();
							isNew = false;
							isExisting = true;
						}

					} else {
						if (match == null || output > finalOutput) {
							finalOutput = output;
							match = innerEntry.getKey();
							isNew = false;
							isExisting = false;
						}

					}

				}

				Map<Pair<RowClusterUUID, InstanceId>, Pair<Boolean, Double>> nonAbsoluteScoresLocal = new HashMap<>();

				boolean add = false;

				if (match != null && isExisting == true) {
					double finalAdaptedOutput;
					if (finalOutput < newThresholdFinal) {
						finalAdaptedOutput = -1.0 * (finalOutput - newThresholdFinal) / (-1.0 - newThresholdFinal);
					} else {
						finalAdaptedOutput = (finalOutput - newThresholdFinal) / (1.0 - newThresholdFinal);
					}
					if (finalAdaptedOutput >= 0.0) {
						if (FILTER_LOW_SCORE && finalOutput < 0.10) {
							nonAbsoluteScoresLocal.put(new Pair<>(entityEntry.getKey(), match), new Pair<>(true, 0.0));

						} else {
							nonAbsoluteScoresLocal.put(new Pair<>(entityEntry.getKey(), match), new Pair<>(true, (double) Math.abs(finalAdaptedOutput)));
						}

						// if (finalOutput >= 0.5) {
						// nonAbsoluteScoresLocal.put(new
						// Pair<>(entityEntry.getKey(), match), new Pair<>(true,
						// 1.0));
						// add = true;
						// } else {
						// // add = false;
						// // nonAbsoluteScores.put(new
						// // Pair<>(entityEntry.getKey(), match), new
						// // Pair<>(true, null));
						// }
						int sdf = 123;
					}
					else {
						// add = false;
						// nonAbsoluteScores.put(new
						// Pair<>(entityEntry.getKey(), match), new Pair<>(true,
						// null));
					}
				}
				else {
					// add = true;
				}

				double highestScore = -1.0;

				for (Tri<RowClusterUUID, InstanceId, Double> nonMatch : nonMatches) {
					if (match == null || isExisting == false || !nonMatch.getB().equals(match)) {
						if (nonMatch.getC() < 0.0) {
							if (nonMatch.getC() > highestScore) {
								highestScore = nonMatch.getC();
							}
							if (FILTER_LOW_SCORE && Math.abs(nonMatch.getC()) < 0.10) {
								nonAbsoluteScoresLocal.put(new Pair<>(nonMatch.getA(), nonMatch.getB()), new Pair<>(false, 0.0));
							} else {
								nonAbsoluteScoresLocal.put(new Pair<>(nonMatch.getA(), nonMatch.getB()), new Pair<>(false, (double) Math.abs(nonMatch.getC())));
							}

						}

						// if (nonMatch.getC() < 0.0) {
						// if (Math.abs(nonMatch.getC()) >= 0.5) {
						// nonAbsoluteScores.put(new Pair<>(nonMatch.getA(),
						// nonMatch.getB()), new Pair<>(false, 1.0));
						// }
						//
						// int hhh = 123;
						// }
						// else if (false) {
						// nonAbsoluteScores.put(new Pair<>(nonMatch.getA(),
						// nonMatch.getB()),
						// new Pair<>(false, (double) Math.round(Math.abs(1.0 -
						// nonMatch.getC()))));
						// }

					}
				}

//				if (add || highestScore <= -0.5) {
//					nonAbsoluteScores.putAll(nonAbsoluteScoresLocal);
				// }
				
				nonAbsoluteScores.putAll(nonAbsoluteScoresLocal);

				if (highestScore <= -0.5) {
					int sfd = 123;
				}

			}

			if (isNew) {
				newEntities.add(new ScoredEntity(entityEntry.getValue(), finalOutput));
			} else if (isExisting) {
				CHMHelper.getFromCHM(mappingForExistingClusters, match, () -> new ConcurrentLinkedQueue<>())
						.add(new ScoredEntity(entityEntry.getValue(), finalOutput));
			}
			else {
				CHMHelper.getFromCHM(unsure, match, () -> new ConcurrentLinkedQueue<>()).add(new ScoredEntity(entityEntry.getValue(), finalOutput));
			}

		}, EntityMatcherWekaLearner.class);



		return new Pair<>(new Tri<>(mappingForExistingClusters, newEntities, unsure), nonAbsoluteScores);
	}

	public static boolean FILTER_LOW_SCORE = false;
	@Override
	public Function<InstanceId, Double> getOtherModelSupplier(Map<InstanceId, EntityMatcherSingleResults> candidates) {

		double wekaWeight;
		double nonWekaWeight;
		double nonWekaNormalizer;
		if (simpleWeightsModel != null) {
			wekaWeight = weights.getOrDefault(run.getOptimizableClass().getWeight("WEKA_WEIGHT", 0.5));
			nonWekaNormalizer = weights.getOrDefault(run.getOptimizableClass().getWeight("NON_WEKA_NORMALIZE", 0.5));
			nonWekaWeight = 1.0 - wekaWeight;
		} else {
			wekaWeight = 1.0;
			nonWekaWeight = 0.0;
			nonWekaNormalizer = 0.0;
		}

		Function<InstanceId, Double> otherModelSupplier;
		if (simpleWeightsModel != null) {
			otherModelSupplier = simpleWeightsModel.getOtherModelSupplier(candidates);
		} else {
			otherModelSupplier = null;
		}

		Map<InstanceId, RecordFeature> featureVectors = EntityMatcherWekaLearner.getFeatureVectors(run.getEntityMatcher(), candidates, datasetData,
				withConfidence, rangeDescretization, this.propertyIds);

		return (intanceId) -> {
			RecordFeature rf = featureVectors.get(intanceId);
			double output;
			if (numeric) {

				// output = model.applyNumeric(rf);

				output = cache.getOrCreateNoBlocking(rf.getFeatures(), () -> {
					double v = model.applyNumeric(rf);
					return v;
				});

			} else {
				output = cache.getOrCreateNoBlocking(rf.getFeatures(), () -> {
					RecordPrediction prediction = model.apply(rf);
					double v;
					if (prediction.getLabel().equals(LearnableRowClusterSimilarities.getLabel(true))) {
						v = prediction.getConfidence();
					} else {
						v = prediction.getConfidence() * -1.0;
					}
					return v;
				});

			}

			if (simpleWeightsModel != null) {

				double d = otherModelSupplier.apply(intanceId);// this.simpleWeightsModel.apply(innerEntry.get);;//equation.normalizedScore(candidates,
																// simpleWeightsModel,
																// run.getOptimizableClass(),
																// run.getEntityMatcher());

				double t;
				if (d > nonWekaNormalizer) {
					t = ((d - nonWekaNormalizer) / (1.0 - nonWekaNormalizer));
				} else {
					t = -1.0 * (1.0 - (d) / (nonWekaNormalizer));
				}
				output = (wekaWeight * output) + (nonWekaWeight * t);
			}

			if (scoreAdapter != null) {
				output = scoreAdapter.get(output, rf);
			}

			return output;
		};
	}

}
