package expansion.set.entitymatching.run;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.ConfigInterface;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import expansion.set.entitymatching.candidateselection.CandidateSelectionStrategy;
import expansion.set.entitymatching.matchers.EntityMatcherStrategy;
import expansion.set.entitymatching.modelLearning.EntityMatchingModelLearnerRunner;

public class EntityMatchingRun implements ConfigInterface<EntityMatchingRun>, Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	private final String name;

	private final Class<? extends EntityMatchingModelLearnerRunner> model;
	private final List<CandidateSelectionStrategy> candidateSelection;
	private final List<EntityMatcherStrategy> entityMatcher;

	public EntityMatchingRun(String name, Class<? extends EntityMatchingModelLearnerRunner> model, List<CandidateSelectionStrategy> candidateSelection,
			List<EntityMatcherStrategy> entityMatcher) {
		super();
		this.name = name;
		this.model = model;
		this.candidateSelection = Collections.unmodifiableList(new ArrayList<>(candidateSelection));
		this.entityMatcher = Collections.unmodifiableList(new ArrayList<>(entityMatcher));

	}

	private EntityMatchingRun(String name, Class<? extends EntityMatchingModelLearnerRunner> model) {
		this.name = name;
		this.model = model;
		this.candidateSelection = new ArrayList<>();
		this.entityMatcher = new ArrayList<>();
	}

	public String getName() {
		return name;
	}

	public List<CandidateSelectionStrategy> getCandidateSelection() {
		return candidateSelection;
	}

	public Class<? extends EntityMatchingModelLearnerRunner> getModel() {
		return model;
	}

	public List<EntityMatcherStrategy> getEntityMatcher() {
		return entityMatcher;
	}

	public static class EntityMatchingRunBuilder implements ConfigBuilderInterface<EntityMatchingRun> {

		private final EntityMatchingRun run;

		public EntityMatchingRunBuilder(EntityMatchingRun entityMatchingRun) {
			run = new EntityMatchingRun(entityMatchingRun.name, entityMatchingRun.model);
			run.candidateSelection.addAll(entityMatchingRun.candidateSelection);
			run.entityMatcher.addAll(entityMatchingRun.entityMatcher);
		}

		public EntityMatchingRunBuilder(String name, Class<? extends EntityMatchingModelLearnerRunner> model) {
			run = new EntityMatchingRun(name, model);
		}

		public EntityMatchingRunBuilder clearCandidateSelection() {
			run.candidateSelection.clear();
			return this;
		}

		public EntityMatchingRunBuilder addCandidateSelection(CandidateSelectionStrategy candidateSelection) {
			run.candidateSelection.add(candidateSelection);
			return this;
		}

		public EntityMatchingRunBuilder clearEntityMatcher() {
			run.entityMatcher.clear();
			return this;
		}

		public EntityMatchingRunBuilder addEntityMatcher(EntityMatcherStrategy entityMatcher) {
			run.entityMatcher.add(entityMatcher);
			return this;
		}

		@Override
		public EntityMatchingRun build() {
			return run.copy();
		}

	}

	@Override
	public EntityMatchingRun copy() {
		return new EntityMatchingRun(name, model, candidateSelection, entityMatcher);
	}

	@Override
	public EntityMatchingRunBuilder rebuilder() {
		return new EntityMatchingRunBuilder(this);
	}

	public static EntityMatchingRunBuilder builder(String name, Class<? extends EntityMatchingModelLearnerRunner> model) {
		return new EntityMatchingRunBuilder(name, model);
	}

	public OptimizableClass getOptimizableClass() {
		return new OptimizableClass(name);
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((candidateSelection == null) ? 0 : candidateSelection.hashCode());
		result = prime * result + ((entityMatcher == null) ? 0 : entityMatcher.hashCode());
		result = prime * result + ((model == null) ? 0 : model.getName().hashCode());
		result = prime * result + ((name == null) ? 0 : name.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (!(obj instanceof EntityMatchingRun))
			return false;
		EntityMatchingRun other = (EntityMatchingRun) obj;
		if (candidateSelection == null) {
			if (other.candidateSelection != null)
				return false;
		} else if (!candidateSelection.equals(other.candidateSelection))
			return false;
		if (entityMatcher == null) {
			if (other.entityMatcher != null)
				return false;
		} else if (!entityMatcher.equals(other.entityMatcher))
			return false;
		if (model == null) {
			if (other.model != null)
				return false;
		} else if (!model.getName().equals(other.model.getName()))
			return false;
		if (name == null) {
			if (other.name != null)
				return false;
		} else if (!name.equals(other.name))
			return false;
		return true;
	}


}
