package expansion.set.evaluation;

import java.util.function.BinaryOperator;

import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumAverage;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumMedian;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.Accumulateable;

	public class ClusterPerformance implements Accumulateable<ClusterPerformance> {
	private final int instancesCreated;
	private final int instancesInKB;
	private final double precisionSum;
	private final double recallSum;
//	private final double fMeasureSum;
	private final int precisionCount;
	private final int recallCount;
	private final AccumMedian medianPerCreatedCluster;
	private final AccumAverage averagePerCreatedCluster;

	private final AccumMedian medianPerKBInstance;
	private final AccumAverage averagePerKBInstance;
	

	private final int mappings;
	private final int mappingsWithAttributes;
	private final int totalRows;
	private final int mappedRows;
	private final int unmappedRows;

	public static ClusterPerformance kbInstanceSize(double size) {
		return new ClusterPerformance(0, 0, 0.0, 0.0, 0, 0, new AccumMedian(), new AccumAverage(), new AccumMedian(size), new AccumAverage(size), 0, 0, 0, 0,
				0);
	}
	public static ClusterPerformance clusterCreatedSize(double size) {
		return new ClusterPerformance(0, 0, 0.0, 0.0, 0, 0, new AccumMedian(size), new AccumAverage(size), new AccumMedian(), new AccumAverage(), 0, 0, 0, 0,
				0);

	}
	
	public static ClusterPerformance withCorpusMappinsAndInstances(int mappings, int mappingsWithAttributes, int totalRows, int mappedRows, int unmappedRows) {
		return new ClusterPerformance(mappings, mappingsWithAttributes, totalRows, mappedRows, unmappedRows, true);
	
	}
	
	private ClusterPerformance(int instancesCreated,int instancesInKB, double precision, double recall, int precisionCount, int recallCount, AccumMedian medianPerCreatedCluster,
			AccumAverage averagePerCreatedCluster,
			AccumMedian medianPerKBInstance,
			AccumAverage averagePerKBInstance, int mappings, int mappingsWithAttributes, int totalRows, int mappedRows, int unmappedRows) {
		super();
		this.instancesCreated = instancesCreated;
		this.instancesInKB = instancesInKB;
		this.precisionSum = precision;
		this.recallSum = recall;
		this.precisionCount = precisionCount;
		this.recallCount = recallCount;
		this.medianPerCreatedCluster = medianPerCreatedCluster;
		this.averagePerCreatedCluster = averagePerCreatedCluster;
		this.medianPerKBInstance = medianPerKBInstance;
		this.averagePerKBInstance = averagePerKBInstance;
		this.mappings = mappings;
		this.totalRows = totalRows;
		this.mappedRows = mappedRows;
		this.unmappedRows = unmappedRows;
		this.mappingsWithAttributes = mappingsWithAttributes;
	}
	
	public ClusterPerformance( double precision, double recall, int precisionCount, int recallCount ) {
		super();
		this.instancesCreated = 0;
		this.instancesInKB = 0;
		this.precisionSum = precision;
		this.recallSum = recall; 
		this.precisionCount = precisionCount;
		this.recallCount = recallCount;
		this.medianPerCreatedCluster = new AccumMedian();
		this.averagePerCreatedCluster = new AccumAverage();
		this.medianPerKBInstance = new AccumMedian();
		this.averagePerKBInstance = new AccumAverage();
		this.mappings = 0;
		this.totalRows = 0;
		this.mappedRows = 0;
		this.unmappedRows = 0;
		this.mappingsWithAttributes = 0;
	}

	public ClusterPerformance( ) {
		super();
		this.instancesCreated = 0;
		this.instancesInKB = 0;
		this.precisionSum = 0.0;
		this.recallSum = 0.0;
		this.precisionCount = 0;
		this.recallCount = 0;
		this.medianPerCreatedCluster = new AccumMedian();
		this.averagePerCreatedCluster = new AccumAverage();
		this.medianPerKBInstance = new AccumMedian();
		this.averagePerKBInstance = new AccumAverage();
		this.mappings = 0;
		this.totalRows = 0;
		this.mappedRows = 0;
		this.unmappedRows = 0;
		this.mappingsWithAttributes = 0;
	}
	public ClusterPerformance(int instancesCreated,int instancesInKB) {
		super();
		this.instancesCreated = instancesCreated;
		this.instancesInKB = instancesInKB;
		this.precisionSum = 0.0;
		this.recallSum = 0.0; 
		this.precisionCount = 0;
		this.recallCount = 0;
		this.medianPerCreatedCluster = new AccumMedian();
		this.averagePerCreatedCluster = new AccumAverage();
		this.medianPerKBInstance = new AccumMedian();
		this.averagePerKBInstance = new AccumAverage();
		this.mappings = 0;
		this.totalRows = 0;
		this.mappedRows = 0;
		this.unmappedRows = 0;
		this.mappingsWithAttributes = 0;
	}
	
	private ClusterPerformance(int mappings, int mappingsWithAttributes, int totalRows, int mappedRows, int unmappedRows, boolean mappingAndInstances) {
		super();
		this.instancesCreated = 0;
		this.instancesInKB = 0;
		this.precisionSum = 0.0;
		this.recallSum = 0.0; 
		this.precisionCount = 0;
		this.recallCount = 0;
		this.medianPerCreatedCluster = new AccumMedian();
		this.averagePerCreatedCluster = new AccumAverage();
		this.medianPerKBInstance = new AccumMedian();
		this.averagePerKBInstance = new AccumAverage();
		this.mappings = mappings;
		this.totalRows = totalRows;
		this.mappedRows = mappedRows;
		this.unmappedRows = unmappedRows;
		this.mappingsWithAttributes = mappingsWithAttributes;
	}
	
	

	
	public double getIncrease() {
		return ((double) instancesCreated) / instancesInKB;
	}
	
	public int getInstancesInKB() {
		return instancesInKB;
	}

	public int getInstancesCreated() {
		return instancesCreated;
	}


	public double getPurity() {
		if(precisionCount == 0)
			return 0.0;
		return precisionSum / precisionCount;
	}


	public double getRecall() {
		if(recallCount == 0)
			return 0.0;
		return recallSum / recallCount;
	}


	public double getfMeasure() {
		double precision = getPurity();
		double recall = getRecall();
		return (2 * precision * recall) / (precision + recall);
	}
	
	
	

	public Double getMedianPerCreatedCluster() {
		return medianPerCreatedCluster.get();
	}
	public Double getAveragePerCreatedCluster() {
		return averagePerCreatedCluster.get();
	}
	public Double getMedianPerKBInstance() {
		return medianPerKBInstance.get();
	}
	public Double getAveragePerKBInstance() {
		return averagePerKBInstance.get();
	}
	
	
	
	
	int getMappings() {
		return mappings;
	}
	 
	public int getMappingsWithAttributes() {
		return mappingsWithAttributes;
	}

	int getTotalRows() {
		return totalRows;
	}
	int getMappedRows() {
		return mappedRows;
	}
	int getUnmappedRows() {
		return unmappedRows;
	}
	public static ClusterPerformance accumulater(ClusterPerformance a, ClusterPerformance b) {
		return new ClusterPerformance(a.instancesCreated + b.instancesCreated, a.instancesInKB + b.instancesInKB, a.precisionSum + b.precisionSum, a.recallSum + b.recallSum, a.precisionCount + b.precisionCount,a.recallCount + b.recallCount,
				AccumMedian.accum(a.medianPerCreatedCluster, b.medianPerCreatedCluster), AccumAverage.accum(a.averagePerCreatedCluster, b.averagePerCreatedCluster), AccumMedian.accum(a.medianPerKBInstance,b.medianPerKBInstance), AccumAverage.accum(a.averagePerKBInstance,b.averagePerKBInstance),
				a.mappings + b.mappings, a.mappingsWithAttributes + b.mappingsWithAttributes, a.totalRows + b.totalRows, a.mappedRows + b.mappedRows,
				a.unmappedRows + b.unmappedRows);
	}

	@Override
	public BinaryOperator<ClusterPerformance> getAccumulator() {
		return ClusterPerformance::accumulater;
	}



//	@Override
//	public ClusterPerformance copy() {
//		return new ClusterPerformance(instancesCreated, instancesInKB, precisionSum, recallSum, fMeasureSum,count, medianPerCreatedCluster.copy(), averagePerCreatedCluster.copy(), medianPerKBInstance.copy(), averagePerCreatedCluster.copy());
//	}

	@Override
	public Class<ClusterPerformance> getAcumClass() {
		return ClusterPerformance.class;
	}

	public static ClusterPerformance empty() {
		return new ClusterPerformance();
	}

	@Override
	public ClusterPerformance accumulate(ClusterPerformance other) {

		return accumulater(this, other);

	}

}
