package expansion.set.evaluation;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.io.FileUtils;

import com.google.common.collect.Sets;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.utils.DecimalFormatter;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;
import de.uni_mannheim.informatik.dws.expansion.evaluation.timestamps.TimestampQualifiersAdapter;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.entitymatching.EntityMatcherFullResults;
import expansion.set.entitymatching.EntityMatcherSingleResults;
import expansion.set.entitymatching.EntityMatcherWeightedResults;
import expansion.set.entitymatching.matchers.EntityMatcherStrategies;
import expansion.set.entitymatching.modelLearning.EntityMatchingModel;

public class ND_CaseCreator {

	private final ValueEvaluator valueEvaluator;
	private final KnowledgeBase knowledgeBase;
	private final TimestampQualifiersAdapter adapter;
	private final AnnotatedTablesRepository tableRepository;
	
	
	 
	@Inject
	public ND_CaseCreator(ValueEvaluator valueEvaluator, KnowledgeBase knowledgeBase, TimestampQualifiersAdapter adapter,
			AnnotatedTablesRepository tableRepository) {
		super();
		this.valueEvaluator = valueEvaluator;
		this.knowledgeBase = knowledgeBase;
		this.adapter = adapter;
		this.tableRepository = tableRepository;
	}
 
	
	public void createWithErrors(EntityMatchingModel matchingWeights, String caseFolder, String fileName, Set<InstanceId> newEntities,
			Map<RowClusterUUID, Optional<InstanceId>> gsMapped,
			EntityMatcherFullResults fullRun, EntityMatcherWeightedResults results) {
		List<Case> cases = new ArrayList<>();

		Set<RowClusterUUID> newFound = results.getNewEntities().stream().map((e) -> e.getEntity().getRowClusterUUID()).collect(Collectors.toSet());
		
		Map<RowClusterUUID, InstanceId> mappedFound = results.getExistingEntities().entrySet().stream().flatMap(
				(e) -> { 
					return e.getValue().stream().map((e2) -> new Pair<>(e.getKey(), e2.getEntity().getRowClusterUUID()));
				
				}
		).collect(Collectors.toMap((f) -> f.getSecond(), (f) -> f.getFirst()));
		
		Map<RowClusterUUID, InstanceId> badFound = results.getBadEntitiesMap().entrySet().stream().flatMap((e) -> {
			return e.getValue().stream().map((e2) -> new Pair<>(e.getKey(), e2.getEntity().getRowClusterUUID()));

		}).collect(Collectors.toMap((f) -> f.getSecond(), (f) -> f.getFirst()));
		
		// Set<RowClusterUUID> added = new HashSet<>();
		for(Entry<RowClusterUUID, Optional<InstanceId>> entry : gsMapped.entrySet()) {

			// if
			// (entry.getValue().get().getInstanceId().equals("http://dbpedia.org/resource/Buisson"))
			// {
			// int sdf = 123;
			// }
			//
			// if (entry.getValue().get().getReadableString().equals("Seek
			// Bromance")) {
			// int sdf = 123;
			// }

			String mainLabel = entry.getValue().get().getReadableString();
			boolean isNew = newEntities.contains(entry.getValue().get());
			InstanceId instanceId = entry.getValue().get();
			
			Map<String, Integer> labels = new HashMap<>();
			labels.put(mainLabel, 1);
			Case caseObject = new Case(instanceId, labels, isNew);

			Map<InstanceId, EntityMatcherSingleResults> full = fullRun.getResults().get(entry.getKey());

			Function<InstanceId, Double> modelSupplier = matchingWeights.getOtherModelSupplier(full);
			
			if (full == null) {
				int sf = 123;
			}
			Function<InstanceId, Map<String, Double>> scoresGetter = (sInstanceId) -> {
				EntityMatcherSingleResults single = full.get(sInstanceId);
				if (single == null) {
					return null;
				} else {
					LinkedHashMap<String, Double> m = single.getScoreMap(fullRun.getEmConfig().getNonTailProperties(),
							Sets.newHashSet(EntityMatcherStrategies.ATTRIBUTE, EntityMatcherStrategies.IMPLICIT_ATTRIBUTE));
					m.put("COMBINED", modelSupplier.apply(sInstanceId));
					return m;
				}
			};

			caseObject.addScoreGetter(scoresGetter);

			if (isNew) {
				
				if(newFound.contains(entry.getKey())) {
					caseObject.addCorrectRow(entry.getValue().get());
				}
				else if (mappedFound.containsKey(entry.getKey())) {
					caseObject.addBadRow(mappedFound.get(entry.getKey()));
				}
				else {
					caseObject.addUnsureBadRow(badFound.get(entry.getKey()));

				}
				
			}
			else {
				if(newFound.contains(entry.getKey())) {
					caseObject.addMissingRow(entry.getValue().get());
				}
				else if (mappedFound.containsKey(entry.getKey())) {
					if(mappedFound.get(entry.getKey()).equals(entry.getValue().get())) {
						caseObject.addCorrectRow(entry.getValue().get());
					}
					else {

						// if
						// (mappedFound.get(entry.getKey()).getReadableString().equals("Seek
						// Bromance")) {
						// int sdf = 123;
						// }
						// if
						// (mappedFound.get(entry.getKey()).getInstanceId().equals("http://dbpedia.org/resource/Buisson"))
						// {
						// int sdf = 123;
						// }

						caseObject.addBadRow(mappedFound.get(entry.getKey()));
						caseObject.addMissingRow(entry.getValue().get());
					}
				}
				else {
					caseObject.addBadRow(badFound.get(entry.getKey()));
					caseObject.addMissingRow(entry.getValue().get());
				}
			}
			

			cases.add(caseObject);
		}

		try {
			Case.writeToFolder(fileName, caseFolder, cases);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}



	 

	// public void create(String caseFolder, Map<InstanceId, List<ScoredEntity>>
	// mappedEntities, ClusterList clusters, FusionPerformance<Integer>
	// maxCorrect, SplitGroupedValues originalTestingSet) {
	//// ConcurrentHashMap<QualifiedTripleId, FusedValue> fusedValues = new
	// ConcurrentHashMap<>();
	//
	// ConcurrentHashMap<RowClusterUUID, RowCluster> clusterMap = new
	// ConcurrentHashMap<>();
	//
	// PersistentParallel6.runAndFinish(clusters.getClusterList(), (cluster) ->
	// {
	// clusterMap.put(cluster.getUuid(), cluster);
	// }, this.getClass());
	//
	//// FusionPerformanceCreator<AccumInteger> accumulator =
	// FusionPerformance.accumulator(new AccumInteger(0));
	//
	// PersistentParallel6.runAndFinish(mappedEntities.entrySet(), (mappings) ->
	// {
	//
	// Integer maxIndex = null;
	// double maxScore = 0.0;
	//
//			for (ListEntry<ScoredEntity> entry : ListIterable.iterable(mappings.getValue())) {
	// if (maxIndex == null || entry.getValue().getScore() > maxScore) {
	// maxIndex = entry.getIndex();
	// maxScore = entry.getValue().getScore();
//				}
//			}
	//
	//
	// ScoredEntity max = mappings.getValue().get(maxIndex);
	//
	// writeCase(caseFolder, mappings.getKey(), max,originalTestingSet,
	// clusterMap.get(max.getEntity().getRowClusterUUID()));
	//
	//// for (ListEntry<ScoredEntity> entry :
	// ListIterable.iterable(mappings.getValue())) {
	//// if (entry.getIndex() != maxIndex) {
	//// for (Entry<QualifiedPropertyId, FusedValue> values :
	// entry.getValue().getEntity().getValues().entrySet()) {
	//// QualifiedTripleId qtid = new QualifiedTripleId(
	//// new TripleId(max.getEntity().getClassId(),
	// values.getKey().getPropertyId(), mappings.getKey()),
	//// values.getKey().getQualifiers());
	//// if (originalTestingSet.containsKey(qtid)) {
	//// accumulator.addWithYear(entry.getValue().getEntity().getClassId(),
	// values.getKey(), ontologyRepository, new AccumInteger(1));
	//// }
	//// }
	//// }
	//// }
	//
	// } , this.getClass());
	//
	// }
	
	
	// private void writeCase(String baseFolder, InstanceId instanceId,
	// ScoredEntity max, SplitGroupedValues originalTestingSet, RowCluster
	// rowCluster ) {
	// boolean include = false;
	//
	//// if(false || true)
	//// return;
	//
	// Set<String> labels = knowledgeBase.getLabels(instanceId);
	//
	// Case caseObject = new Case( instanceId, max.getEntity().getLabels(),
	// max.getEntity().getRowClusterUUID().toString());
	//
	//// List<Pair<QualifiedTripleId, FusedValue>> correctValues = new
	// ArrayList<>();
	//// List<Pair<QualifiedTripleId, FusedValue>> incorrectValues = new
	// ArrayList<>();
	//
	// for(Entry<QualifiedPropertyId, FusedValue> values :
	// max.getEntity().getValues().entrySet()) {
	// QualifiedTripleId qtid = new QualifiedTripleId(new
	// TripleId(max.getEntity().getClassId(), values.getKey().getPropertyId(),
	// instanceId),
	// values.getKey().getQualifiers());
	// if (originalTestingSet.containsKey(qtid)) {
	//
	//
	//
	// Optional<Boolean> correct = valueEvaluator.correct(qtid,
	// values.getValue());
	//
	// if(correct.isPresent()) {
	//
	//
	// if(!correct.get()) {
	// include = true;
	//
	// }
	//
	//
	// QualifiedValueList qvl =
	// knowledgeBase.getQualifiedValueList(qtid.getTripleId().getValueId());
	// Set<AbstractObject> correctValues = adapter.getForQualifiers(qvl,
	// qtid.getQualifiers());
	//
	// Integer contributors = values.getValue().getSingleContributors();
	// if(contributors == null)
	// contributors = 0;
	//
	// caseObject.addValue(Optional.of(correct.get()), qtid, values.getValue(),
	// correctValues, contributors);
	//
	// }
	// }
	// else {
	// int adsd = 122;
	// }
	// }
	//
	// for(TableRowId rowId : rowCluster.getRows()) {
	//
	// caseObject.addTables(rowId.getTableId(),
	// tableRepository.getTable(rowId.getTableId()), rowId.getRowIdx());
	//
	// }
	//
	//
	// try {
	// if(include) {
	// caseObject.writeToFolder(baseFolder);
	//
	// int df = 123;
	// }
	// } catch (IOException e) {
	// // TODO Auto-generated catch block
	// e.printStackTrace();
	// }
	//
	// int sdf = 13;
	//
	// }
	
	
	public static class Case {
		
//		private final Set<String> labels;
//		private final InstanceId instanceId;
		
		private String header = "";
		// private String values = "";
		// private String tables = "";
		// private final String uuid;
		// private Set<TableRowId> rowIds = new HashSet<>();
		private Set<InstanceId> correctRowIds = new HashSet<>();
		private Set<InstanceId> missingRowIds = new HashSet<>();
		private Set<InstanceId> badRowIds = new HashSet<>();
		private Set<InstanceId> unsureBadRowIds = new HashSet<>();

		// private Map<AnnotatedTableId, AnnotatedTable> tablesMap = new
		// HashMap<>();
		private Map<AnnotatedTableId, Set<RowIdx>> rowIdx = new HashMap<>();
		private Function<InstanceId, Map<String, Double>> scoreGetter;
		// private Function<TableRowId, Set<TableRowId>>
		// blockingCandidatesGetter;
		
		// public Case(Entity entity, String uuid, boolean isNew) {
		//
		// List<Pair<String, Integer>> list =
		// entity.getLabels().entrySet().stream().sorted((e1, e2) ->
		// Integer.compare(e2.getValue(), e1.getValue()))
		// .map((e) -> new Pair<>(e.getKey(),
		// e.getValue())).collect(Collectors.toList());
		//
		// Collections.sort(list, (e1, e2) -> Integer.compare(e2.getSecond(),
		// e1.getSecond()));
		// header += "<hr />";
		// header += "<h1>" + entity.getMainLabel() + "</h1>";
		//
		// header += "<h2>Labels</h2><ul>";
		//
		// for (Pair<String, Integer> label : list) {
		// header += "<li>" + label.getFirst() + " - " + label.getSecond() +
		// "</li>";
		// }
		// header += "</ul>";
		//
		// // values += "<h2>Values</h2>";
		//
		// this.uuid = uuid;
		//
		// // this.instanceId = instanceId;
		// // this.labels = labels;
		// }

		public Case(InstanceId instanceId, Map<String, Integer> map, boolean isNew) {
			
			List<Pair<String, Integer>> list = map.entrySet().stream().sorted(  (e1, e2) -> Integer.compare(e2.getValue(), e1.getValue())).map((e) -> new Pair<>(e.getKey(), e.getValue())).collect(Collectors.toList());
 

			Collections.sort(list, (e1, e2) -> Integer.compare(e2.getSecond(), e1.getSecond()));
			header += "<hr />";
			header += "<h1>" + (isNew ? "NEW: " : "EXISTING: ") + instanceId.getReadableString() + " - " + instanceId.getInstanceId() + "</h1>";
			
			header += "<h2>Labels</h2><ul>";
			
			for(Pair<String, Integer> label : list) {
				header += "<li>" + label.getFirst() + " - " + label.getSecond() + "</li>";
			}
			header += "</ul>";
			
//			values += "<h2>Values</h2>";
			
			// this.uuid = uuid;
			 
//			this.instanceId = instanceId;
//			this.labels = labels;
		}
		

		public void addScoreGetter(Function<InstanceId, Map<String, Double>> scoreGetter) {
			this.scoreGetter = scoreGetter;

		}


		
		public void addCorrectRow(InstanceId instanceId) {

			this.correctRowIds.add(instanceId);

		}

		public void addMissingRow(InstanceId rowId) {
			this.missingRowIds.add(rowId);

		}

		public void addBadRow(InstanceId rowId) {

			this.badRowIds.add(rowId);

		}

		
		public void addUnsureBadRow(InstanceId rowId) {

			this.unsureBadRowIds.add(rowId);

		}

		
		private String createRows() {
			String ret = "";
			// if (!rowIds.isEmpty()) {
			// ret += "<h2>Rows</h2>";
			//
			// ret += "<ul>";
			//
			// for (TableRowId rowId : rowIds) {
			// ret += "<li>" + rowId.getTableId().getTableName() + ", " +
			// rowId.getRowIdx().getRowIdx() + "</li>";
			// }
			//
			// ret += "</ul>";
			// }

			if (!correctRowIds.isEmpty()) {
				ret += "<h2>Correct Match</h2>";

				ret += "<ul>";

				for (InstanceId rowId : correctRowIds) {
					ret += "<li style=\"color: green;\">" + rowId.getReadableString() + "<br />";
					if (scoreGetter != null) {
						ret += comparisonTable(rowId, "green");
					}


					ret += "</li>";

				}

				ret += "</ul>";
			}

			if (!badRowIds.isEmpty()) {
				ret += "<h2>Bad Match</h2>";

				ret += "<ul>";

				for (InstanceId rowId : badRowIds) {
					ret += "<li style=\"color: red;\">" + rowId.getReadableString() + "<br />";

					if (scoreGetter != null) {
						ret += comparisonTable(rowId, "red");
					}

					ret += "</li>";
				}

				ret += "</ul>";
			}

			if (!unsureBadRowIds.isEmpty()) {
				ret += "<h2>Unsure Bad Match</h2>";

				ret += "<ul>";

				for (InstanceId rowId : unsureBadRowIds) {
					ret += "<li style=\"color: purple;\">" + rowId.getReadableString() + "<br />";

					if (scoreGetter != null) {
						ret += comparisonTable(rowId, "purple");
					}

					ret += "</li>";
				}

				ret += "</ul>";
			}

			if (!missingRowIds.isEmpty()) {
				ret += "<h2>Missing Match</h2>";

				ret += "<ul>";

				for (InstanceId rowId : missingRowIds) {
					ret += "<li style=\"color: orange;\">" + rowId.getReadableString() + "<br />";

					if (scoreGetter != null) {
						ret += comparisonTable(rowId, "orange");
					}

					ret += "</li>";
				}

				ret += "</ul>";
			}

			return ret;

		}

		private String comparisonTable(InstanceId rowId, String color) {

			Map<String, Double> score = scoreGetter.apply(rowId);
			String ret;
			if (score == null) {
				ret = "NOT FOUND IN CANDIDATE SELECTION: " + rowId.getReadableString();
			} else {
				ret = "";

				ret += "<table cellspacing=\"0\" border=\"2\" cellpadding=\"5\">";
				ret += "<tr style=\"color: " + color + ";\">";
				ret += "<th>Row</th>";
				for (String string : score.keySet()) {
					if (string.equals("AA_LABEL_PHI"))
						string = "PHI";
					ret += "<th>" + string + "</th>";

				}

				ret += "<tr style=\"color: " + color + ";\">";
				ret += "<td>" + rowId.getInstanceId() + "</td>";
				for (Double d : score.values()) {

					ret += "<td>" + DecimalFormatter.format(d) + "</td>";

				}
				ret += "</tr>";

				ret += "</table>";

			}

			return ret;
		}

		// private String createTables() {
		// String ret = "<h2>Tables</h2>";
		//
		// for( Entry<AnnotatedTableId, AnnotatedTable> tableEntry :
		// this.tablesMap.entrySet()) {
		//
		// Set<RowIdx> rowIds = this.rowIdx.get(tableEntry.getKey());
		//
		// int columnSize = tableEntry.getValue().getColumns().size();
		//
		// int rowSize = tableEntry.getValue().getValues().size() / columnSize;
		//
		//
		// ret += "<h3>" + tableEntry.getKey().getTableName() + "</h3>";
		// ret += "<table>";
		//
		// for(int i = 0; i < rowSize; i++) {
		//
		//
		// RowIdx rowIdx = new RowIdx(i);
		//
		// if(rowIds.contains(rowIdx)) {
		//
		//
		// ret += "<tr style=\"background-color: #eeeeee;\">";
		// }
		// else {
		// ret += "<tr>";
		// }
		//
		//
		//
		// for(int c = 0; c < columnSize; c++) {
		// AbstractObject v = tableEntry.getValue().getValue(rowIdx, new
		// ColumnIdx(c));
		// String str= "";
		// if(v != null)
		// str = v.toString();
		// if(i == 0) {
		// ret += "<th>" + str + "</th>";
		// }
		// else {
		// ret += "<td>" + str + "</td>";
		// }
		//
		// }
		//
		// ret += "</tr>";
		// }
		//
		// ret += "</table>";
		// }
		//
		// return ret;
		// }
		
		// private String getHtml() {
		// StringBuilder sb = new StringBuilder( "<html><head><title>" +
		// this.uuid + "</title></head><body>");
		//
		// sb.append(header);
		//
		// sb.append(createValues());
		// sb.append(createTables());
		// sb.append(createRows());
		// sb.append("</body></html>");
		//
		// return sb.toString();
		// }
		//
		private String getInnerHtml() {
			StringBuilder sb = new StringBuilder(  );
			
			sb.append(header);
			

			sb.append(createRows());
			
			return sb.toString();
		}
		
		// public void writeToFolder(String folder) throws IOException {
		// String html = getHtml();
		//
		// FileUtils.writeStringToFile(new File(folder + "/" + uuid.toString() +
		// ".html"), html);
		//
		// }
		//
		public static void writeToFolder(String uuid, String folder, Collection<Case> cases) throws IOException {
			StringBuilder sb = new StringBuilder( "<html><head><title>" + uuid + "</title></head><body>");
			
			for(Case caze : cases) {
				sb.append(caze.getInnerHtml());
			}
			
			
			sb.append("</body></html>");
			
		 
			
			FileUtils.writeStringToFile(new File(folder + "/" + uuid + ".html"), sb.toString());
		}


	}


 


}
