package expansion.set.evaluation;

import java.io.File;
import java.io.IOException;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.mutable.MutableInt;

import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdsSet;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.AccuracyAverager;
import de.mannheim.uni.expansion.tablefusion.analysis.AccuracyCalculator;
import de.mannheim.uni.expansion.tablefusion.analysis.MaxCorrectFinder;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.FusionPerformanceSynchronizedCreator;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.ResultsCategory;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.generic.GenericResultsWriter;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.generic.GenericResultsWriterAccumRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.similarity.Similarity.SimilarityType;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfigFilter;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.tablefusion.learning.LearningSetCreatorStratified;
import de.mannheim.uni.expansion.tablefusion.learning.RowValuesGetter;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;
import de.mannheim.uni.expansion.tablefusion.learning.SplitRowValues;
import de.mannheim.uni.expansion.tablefusion.optimization.ScoreFusionThresholdFinderFullQualifiedSimple;
import de.mannheim.uni.expansion.tablefusion.optimization.ThresholdKryoCache;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionMethod;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategy;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.ScorerRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.DecimalFormatter;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.consumers.FourConsumer;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordFeature;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.ColumnListExtendedOutputterGenerator;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.ColumnListExtendedOutputterGenerator.ColumnListExtendedOutputter;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance.FusionPerformanceCreator;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdConstant;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdContainer;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumAverage;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AccumulatorListImmutable;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AccumulatorListImmutable.ImmutableAccumulatorListBuild;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AcumReference;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList.ColumnListBuilder;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList.ColumnListBuilderGeneric;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentAccumulator;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.GenericAccumaltableResult;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.GenericAccumaltableResult.ResultBuilder;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.LearnedModelRepository;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusterer;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.creation.Entity;
import expansion.set.creation.EntityCreator;
import expansion.set.entitymatching.EntityMatcherWeightedResults;
import expansion.set.entitymatching.EntityMatchingRunner;
import expansion.set.entitymatching.matchers.EntityMatcherConfig;
import expansion.set.entitymatching.modelLearning.EM_LearnedModelRepository;
import expansion.set.entitymatching.run.EntityMatchingRun;
import expansion.set.evaluation.crossvalidation.EntitySetGroupedValuesSplitter;
import expansion.set.evaluation.crossvalidation.EntitySets;
import expansion.set.evaluation.crossvalidation.EntitySetsCreator;
import expansion.set.evaluation.crossvalidation.EntitySetsCreatorHard;
import expansion.set.evaluation.entities.EntitiesPerformance;
import expansion.set.evaluation.entities.EntitiesPerformanceEvaluator;
import expansion.set.evaluation.fusion.FusionSetCreator;
import expansion.set.evaluation.fusion.FusionSetEvaluator;
import expansion.set.evaluation.fusion.FusionSetPerformance;
import expansion.set.evaluation.goldstandard.ClusterEvaluator;
import expansion.set.evaluation.goldstandard.ClusterToGoldStandardMapper;
import expansion.set.evaluation.goldstandard.GoldStandardClusters;
import expansion.set.evaluation.goldstandard.oktie.OktieClusterEvaluator;
import expansion.set.evaluation.goldstandard.oktie.OktieClusterEvaluator.OktieMapping;
import expansion.set.evaluation.goldstandard.oktie.OktiePerformance;
import expansion.set.evaluation.matching.EntityMatcher;
import expansion.set.evaluation.matching.NewDetectionEvaluator;
import expansion.set.evaluation.matching.NewDetectionPerformance;
import expansion.set.evaluation.matching.ScoredEntity;
import expansion.set.evaluation.pipelineresults.GsStatistics;
import expansion.set.evaluation.pipelineresults.PipelinePerformance;
import expansion.set.evaluation.rankedEvaluation.RankedEvaluationPerformance;
import expansion.set.evaluation.rankedEvaluation.RankingEvalutator;
import expansion.set.evaluation.sim.SimEvaluator;
import expansion.set.evaluation.sim.SimPerformance;
import expansion.set.exportimport.ExternalClusteringImporter;
import expansion.set.implicitAttributes.ImplicitAttributeFinder;
import expansion.set.profiling.ClusterListEvaluator;
import expansion.set.profiling.TableRowSetProfiler;
import expansion.set.profiling.TableSetProfiler;
import expansion.set.profiling.TableSetProfiler.TableProfiling;
import expansion.set.strategies.SetExpansionStrategiesRepository;
import expansion.set.strategies.SetExpansionStrategy;
import expansion.set.strategies.clustersplitters.minScoreSplitter.CombinedRowClusterSimilaritiesCreator.CombinedRowClusterSimilarities;
import expansion.set.strategies.clustersplitters.minScoreSplitter.WeightedAverageScoreSplitterFactory.AverageScoresSplitterSimilarities;
import expansion.set.strategies.clustersplitters.minScoreSplitter.WekaLearnableRowClusterSimilarities;
import expansion.set.strategies.rowclusterers.GreedyClusteringSplitter;
import expansion.set.strategies.rowclusterers.RowClustererRepository;
import expansion.set.strategies.rowclusterers.label.TfIdfBlockCreator;

public class SetExpansionRunner {

	@Inject
	protected EntityMatchingRunner entityMatchingRuner;
	private final EntitySetsCreator entitySetsCreator;
	private final EntitySetGroupedValuesSplitter entitySetGroupedValuesSplitter;
	private final AnnotatedTablesRepository tablesRepository;
	private final RowClustererRepository rowClustererRepository;
	private final EntityCreator entityCreator;
	private final EntityMatcher entityMatcher;
	private final FusionSetCreator fusionSetCreator;
	private final MaxCorrectFinder maxCorrectFinder;
	private final AccuracyCalculator accuracyCalculator;
	private final OntologyRepository ontologyRepository;
	private final ScoreFusionThresholdFinderFullQualifiedSimple scoreFusionThresholdFinderFullQualifiedSimple;
	private final ThresholdKryoCache thresholdKryoCache;
	private final AccuracyAverager averager;
	private final EntitySetsEvaluator entitySetsEvaluator;
	private final CaseCreator caseCreator;
	private final EntitySetsCreatorHard entitySetsCreatorHard;
	private final KnowledgeBase knowledgeBase;
	private final ClusterToGoldStandardMapper clusterToGoldStandardMapper;
	private final ClusterEvaluator clusterEvaluator;
	private final AnnotatedTablesRepository tableRepository; 

	@Inject
	protected expansion.set.exportimport.ClusterPairsExporter ClusterPairsExporter;

	@Inject
	protected RowValuesGetter rowValuesGetter;
	
	@Inject
	protected ColumnListExtendedOutputterGenerator columnListExtendedOutputterGenerator;
	@Inject
	protected GreedyClusteringSplitter greedyClusteringSplitter;
	
	@Inject
	protected SimEvaluator simEvaluator;
	
	@Inject
	protected NewDetectionEvaluator newDetectionEvaluator;
	
	@Inject
	protected FusionSetEvaluator fusionSetEvaluator;
	
	
	@Inject
	protected OktieClusterEvaluator oktieClusterEvaluator;
	
	
	@Inject
	protected TableRowSetProfiler tableRowSetProfiler;
	
	@Inject
	protected TableSetProfiler tableSetProfiler;
	
	@Inject
	protected SplitValuesGetter splitValuesGetter;

	@Inject
	protected ClusterListEvaluator clusterListEvaluator;
	
	@Inject
	protected ValueEvaluator valueEvaluator;
	
	@Inject
	protected EntitiesPerformanceEvaluator entitiesPerformanceEvaluator;
	
	@Inject
	protected TfIdfBlockCreator TfIdfBlockCreator;
	
	@Inject
	protected LearningSetCreatorStratified learningSetCreatorStratified;

	@Inject
	protected ExternalClusteringImporter externalClusteringImporter;

	@Inject
	protected PropertiesConfigFilter propertiesConfigFilter;
	
	@Inject 
	protected  ScorerRepository scorerRepository;
	
	@Inject
	protected RankingEvalutator rankingEvalutator;

	@Inject
	protected ImplicitAttributeFinder implicitAttributeFinder;

	@Inject
	public SetExpansionRunner(
			EntitySetsCreator entitySetsCreator, EntitySetGroupedValuesSplitter entitySetGroupedValuesSplitter, AnnotatedTablesRepository tablesRepository,
			RowClustererRepository rowClustererRepository, EntityCreator entityCreator, EntityMatcher entityMatcher, FusionSetCreator setExpansionFuser,
			MaxCorrectFinder maxCorrectFinder, AccuracyCalculator accuracyCalculator, OntologyRepository ontologyRepository,
			ScoreFusionThresholdFinderFullQualifiedSimple scoreFusionThresholdFinderFullQualifiedSimple, ThresholdKryoCache thresholdKryoCache,
			AccuracyAverager averager,CaseCreator caseCreator, EntitySetsEvaluator entitySetsEvaluator, EntitySetsCreatorHard entitySetsCreatorHard,KnowledgeBase knowledgeBase,
			ClusterToGoldStandardMapper clusterToGoldStandardMapper,
 ClusterEvaluator clusterEvaluator,AnnotatedTablesRepository tableRepository) {
		super();
		this.entitySetsCreator = entitySetsCreator;
		this.entitySetGroupedValuesSplitter = entitySetGroupedValuesSplitter;
		this.tablesRepository = tablesRepository;
		this.rowClustererRepository = rowClustererRepository; 
		this.entityCreator = entityCreator;
		this.entityMatcher = entityMatcher;
		this.fusionSetCreator = setExpansionFuser;
		this.maxCorrectFinder = maxCorrectFinder;
		this.accuracyCalculator = accuracyCalculator;
		this.ontologyRepository = ontologyRepository;
		this.scoreFusionThresholdFinderFullQualifiedSimple = scoreFusionThresholdFinderFullQualifiedSimple;
		this.thresholdKryoCache = thresholdKryoCache;
		this.averager = averager;
		this.caseCreator = caseCreator;
		this.entitySetsEvaluator  = entitySetsEvaluator;
		this.entitySetsCreatorHard = entitySetsCreatorHard;
		this.knowledgeBase = knowledgeBase;
		this.clusterToGoldStandardMapper = clusterToGoldStandardMapper;
		this.clusterEvaluator = clusterEvaluator;
		this.tableRepository = tableRepository;
	}
	
	public static enum TestingMode {
		NORMAL, EASY, HARD, ALL;
	}
	
	
	public static final ClassId settlementClass = new ClassId("http://dbpedia.org/ontology/Settlement");
	public static final ClassId musicalWorkClass = new ClassId("http://dbpedia.org/ontology/MusicalWork");
	public static final ClassId populatedPlace = new ClassId("http://dbpedia.org/ontology/PopulatedPlace");
	

	public ColumnList<PipelinePerformance> getColumnListforPipelinePerformance() {
		
		ColumnListBuilderGeneric<PipelinePerformance> builder = ColumnList.builder();
		
//		ClusteringPerformance,GS-Size,Clusters,PCPrecision,Recall,FMeasure
		
		
		
		builder.addColumn("Tables", (p) -> {
			return String.valueOf(p.getGsStatistics().getTables());
		});
		
		builder.addColumn("Rows", (p) -> {
			return String.valueOf(p.getGsStatistics().getRows());
		});
		
		builder.addColumn("GS-Size", (p) -> {
			return String.valueOf(p.getGsStatistics().getClusters());
		});
		
		builder.addColumn("Existing-Clusters", (p) -> {
			return String.valueOf(p.getGsStatistics().getExistingClusters());
		});
		
		builder.addColumn("Existing-Facts", (p) -> {
			return String.valueOf(p.getGsStatistics().getExistingClustersFacts());
		});
		
		builder.addColumn("New-Clusters", (p) -> {
			return String.valueOf(p.getGsStatistics().getNewClusters());
		});
		

		builder.addColumn("New-Facts", (p) -> {
			return String.valueOf(p.getGsStatistics().getNewClustersFacts());
		});
		
		builder.addColumn("", (p) -> {
			return "";
		});
		
		
		builder.addColumn("SimCorrect", (p) -> {
			return String.valueOf(p.getSimAccuracy().getPerformance().getCorrect());
		});

		builder.addColumn("SimWrong", (p) -> {
			return String.valueOf(p.getSimAccuracy().getPerformance().getWrong());
		});

		builder.addColumn("SimMaxCorrect", (p) -> {
			return String.valueOf(p.getSimAccuracy().getPerformance().getMaxCorrect());
		});
		
		builder.addColumn("SimPrecision", (p) -> {
			return DecimalFormatter.format(p.getSimAccuracy().getPerformance().getPrecision());
		});
		
		builder.addColumn("SimRecall", (p) -> {
			return DecimalFormatter.format(p.getSimAccuracy().getPerformance().getRecall());
		});
		
		builder.addColumn("SimFmeasure", (p) -> {
			return DecimalFormatter.format(p.getSimAccuracy().getPerformance().getfMeasure());
		});
		
		builder.addColumn("RMSE", (p) -> {
			return DecimalFormatter.format(p.getSimAccuracy().getRmse());
		});
		
		builder.addColumn("", (p) -> {
			return "";
		});

		builder.addColumn("NC_Clusters", (p) -> {
			return String.valueOf(p.getNonClusteringClusteringPerformance().getClusters());
		});

		builder.addColumn("NC_WithExistingMapping", (p) -> {
			return String.valueOf(p.getNonClusteringClusteringPerformance().getExistingClustersWithMapping());
		});

		builder.addColumn("NC_WithNewMapping", (p) -> {
			return String.valueOf(p.getNonClusteringClusteringPerformance().getNewClustersWithMapping());
		});

		builder.addColumn("NC_NoMapping", (p) -> {
			return String.valueOf(
					p.getNonClusteringClusteringPerformance().getClusters() - p.getNonClusteringClusteringPerformance().getExistingClustersWithMapping()
							- p.getNonClusteringClusteringPerformance().getNewClustersWithMapping());
		});

		builder.addColumn("NC_PCPrecision", (p) -> {
			return DecimalFormatter.format(p.getNonClusteringClusteringPerformance().getPenalizedClusteringPrecisionMacro());
		});

		builder.addColumn("NC_CPrecision", (p) -> {
			return DecimalFormatter.format(p.getNonClusteringClusteringPerformance().getClusteringPrecisionMacro());
		});

		builder.addColumn("NC_Recall", (p) -> {
			return DecimalFormatter.format(p.getNonClusteringClusteringPerformance().getRecallMacro());
		});

		builder.addColumn("NC_FMeasure", (p) -> {
			return DecimalFormatter.format(p.getNonClusteringClusteringPerformance().getFMeasureMacro());
		});

		builder.addColumn("NC_FMeasureNP", (p) -> {
			return DecimalFormatter.format(p.getNonClusteringClusteringPerformance().getFMeasureNonPenalizedMacro());
		});
 
		

		builder.addColumn("", (p) -> {
			return "";
		});

		builder.addColumn("Clusters", (p) -> {
			return String.valueOf(p.getClusteringAccuracy().getClusters());
		});
		
		builder.addColumn("WithExistingMapping", (p) -> {
			return String.valueOf(p.getClusteringAccuracy().getExistingClustersWithMapping());
		});
		
		builder.addColumn("WithNewMapping", (p) -> {
			return String.valueOf(p.getClusteringAccuracy().getNewClustersWithMapping());
		});
		
		builder.addColumn("NoMapping", (p) -> {
			return String.valueOf(p.getClusteringAccuracy().getClusters() - p.getClusteringAccuracy().getExistingClustersWithMapping() - p.getClusteringAccuracy().getNewClustersWithMapping());
		});
		
		builder.addColumn("PCPrecision", (p) -> {
			return DecimalFormatter.format(p.getClusteringAccuracy().getPenalizedClusteringPrecisionMacro());
		});

		builder.addColumn("CPrecision", (p) -> {
			return DecimalFormatter.format(p.getClusteringAccuracy().getClusteringPrecisionMacro());
		});
		
		builder.addColumn("Recall", (p) -> {
			return DecimalFormatter.format(p.getClusteringAccuracy().getRecallMacro());
		});
		
		builder.addColumn("FMeasure", (p) -> {
			return DecimalFormatter.format(p.getClusteringAccuracy().getFMeasureMacro());
		});
		
		builder.addColumn("FMeasureNP", (p) -> {
			return DecimalFormatter.format(p.getClusteringAccuracy().getFMeasureNonPenalizedMacro());
		});
		
		builder.addColumn("", (p) -> {
			return "";
		});
		
		
 
		builder.addColumn("ExistingClusters", (p) -> {
			return String.valueOf(p.getNewDetectionPerformance().getMappedExisting());
		});
		
		builder.addColumn("NewClusters", (p) -> {
			return String.valueOf(p.getNewDetectionPerformance().getMappedNew());
		});
		
		builder.addColumn("BadClusters", (p) -> {
			return String.valueOf(p.getNewDetectionPerformance().getMappedBad());
		});

		builder.addColumn("ExistingPrecision", (p) -> {
			return DecimalFormatter.format(p.getNewDetectionPerformance().getExistingPrecision());
		});

		builder.addColumn("ExistingRecall", (p) -> {
			return DecimalFormatter.format(p.getNewDetectionPerformance().getExistingRecall());
		});

		builder.addColumn("NewPrecision", (p) -> {
			return DecimalFormatter.format(p.getNewDetectionPerformance().getNewPrecision());
		});

		builder.addColumn("NewRecall", (p) -> {
			return DecimalFormatter.format(p.getNewDetectionPerformance().getNewRecall());
		});
		
		builder.addColumn("BadPrecision", (p) -> {
			return DecimalFormatter.format(p.getNewDetectionPerformance().getBadPrecision());
		});
		
		builder.addColumn("BadRecall", (p) -> {
			return DecimalFormatter.format(p.getNewDetectionPerformance().getBadRecall());
		});
		
		builder.addColumn("TotalDetectionAccuracy", (p) -> {
			return DecimalFormatter.format(p.getNewDetectionPerformance().getTotalAccuracy());
		});
 
		builder.addColumn("", (p) -> {
			return "";
		});
		
		builder.addColumn("MAP", (p) -> {
			return String.valueOf(p.getRankedEvaluationPerformance().getAveragePrecisionSum());
		});
		builder.addColumn("P@5", (p) -> {
			return String.valueOf(p.getRankedEvaluationPerformance().getpAt5());
		});
		builder.addColumn("P@20", (p) -> {
			return String.valueOf(p.getRankedEvaluationPerformance().getpAt20());
		});
		builder.addColumn("P@50", (p) -> {
			return String.valueOf(p.getRankedEvaluationPerformance().getpAt50());
		});

		builder.addColumn("P@100", (p) -> {
			return String.valueOf(p.getRankedEvaluationPerformance().getpAt100());
		});

		builder.addColumn("", (p) -> {
			return "";
		});

		builder.addColumn("IrPrecision", (p) -> {
			return DecimalFormatter.format(p.getNewDetectionPerformance().getIRPrecision());
		});
		builder.addColumn("IrRecall", (p) -> {
			return DecimalFormatter.format(p.getNewDetectionPerformance().getIRRecall());
		});
		builder.addColumn("IrF1", (p) -> {
			return DecimalFormatter.format(p.getNewDetectionPerformance().getIRFmeasure());
		});
		
		builder.addColumn("", (p) -> {
			return "";
		});
		

		builder.addColumn("ExistingFound", (p) -> {
			return String.valueOf(p.getEntitiesPerformance().getExistingEntitiesPerformance().getFoundClusters());
		});
		
		builder.addColumn("WithCollisions", (p) -> {
			return String.valueOf(p.getEntitiesPerformance().getExistingEntitiesPerformance().getClustersWithCollision());
		});
		
		builder.addColumn("TotalCollisions", (p) -> {
			return String.valueOf(p.getEntitiesPerformance().getExistingEntitiesPerformance().getTotalCollisions());
		});
		
		builder.addColumn("Precision", (p) -> {
			return DecimalFormatter.format(p.getEntitiesPerformance().getExistingEntitiesPerformance().getPrecision());
		});
		
		builder.addColumn("Recall", (p) -> {
			return DecimalFormatter.format(p.getEntitiesPerformance().getExistingEntitiesPerformance().getRecall());
		});
		
		builder.addColumn("Fmeasure", (p) -> {
			return DecimalFormatter.format(p.getEntitiesPerformance().getExistingEntitiesPerformance().getFmeasure());
		});
		
		
		
		builder.addColumn("", (p) -> {
			return "";
		});
		
		

		builder.addColumn("NewFound", (p) -> {
			return String.valueOf(p.getEntitiesPerformance().getNewEntitiesPerformance().getFoundClusters());
		});
		
		builder.addColumn("WithCollisions", (p) -> {
			return String.valueOf(p.getEntitiesPerformance().getNewEntitiesPerformance().getClustersWithCollision());
		});
		
		builder.addColumn("TotalCollisions", (p) -> {
			return String.valueOf(p.getEntitiesPerformance().getNewEntitiesPerformance().getTotalCollisions());
		});
		
		builder.addColumn("Precision", (p) -> {
			return DecimalFormatter.format(p.getEntitiesPerformance().getNewEntitiesPerformance().getPrecision());
		});
		
		builder.addColumn("Recall", (p) -> {
			return DecimalFormatter.format(p.getEntitiesPerformance().getNewEntitiesPerformance().getRecall());
		});
		
		builder.addColumn("Fmeasure", (p) -> {
			return DecimalFormatter.format(p.getEntitiesPerformance().getNewEntitiesPerformance().getFmeasure());
		});
		
		
		
		builder.addColumn("", (p) -> {
			return "";
		});
		
		builder.addColumn("E-Correct-Facts", (p) -> {
			return String.valueOf(p.getFusionSetPerformance().getExistingClusters().getCorrect());
		});
		
		builder.addColumn("E-Wrong-Facts", (p) -> {
			return String.valueOf(p.getFusionSetPerformance().getExistingClusters().getWrong());
		});
		
		builder.addColumn("E-Missing", (p) -> {
			return String.valueOf(p.getFusionSetPerformance().getExistingClusters().getMaxCorrect() - p.getFusionSetPerformance().getExistingClusters().getCorrect());
		});
		
		builder.addColumn("E-NotInGS", (p) -> {
			return String.valueOf(p.getFusionSetPerformance().getExistingClusters().getNotInGs());
		});

		builder.addColumn("", (p) -> {
			return "";
		});
		
		
		builder.addColumn("N-Correct-Facts", (p) -> {
			return String.valueOf(p.getFusionSetPerformance().getNewClusters().getCorrect());
		});
		
		builder.addColumn("N-Wrong-Facts", (p) -> {
			return String.valueOf(p.getFusionSetPerformance().getNewClusters().getWrong());
		});
		
		builder.addColumn("N-Missing", (p) -> {
			return String.valueOf(p.getFusionSetPerformance().getNewClusters().getMaxCorrect() - p.getFusionSetPerformance().getNewClusters().getCorrect());
		});
		
		builder.addColumn("N-NotInGS", (p) -> {
			return String.valueOf(p.getFusionSetPerformance().getNewClusters().getNotInGs());
		});

		builder.addColumn("", (p) -> {
			return "";
		});

		builder.addColumn("N-Facts-Precision", (p) -> {
			return DecimalFormatter.format(p.getFusionSetPerformance().getNewClusters().getPrecisionMicro());
		});

		builder.addColumn("N-Facts-Recall", (p) -> {
			return DecimalFormatter.format(p.getFusionSetPerformance().getNewClusters().getRecallMicro());
		});

		builder.addColumn("N-Facts-FMeasure", (p) -> {
			return DecimalFormatter.format(p.getFusionSetPerformance().getNewClusters().getFMeasureMicro());
		});

		builder.addColumn("", (p) -> {
			return "";
		});
		
		
		builder.addColumn("Facts-Precision", (p) -> {
			return DecimalFormatter.format(p.getFusionSetPerformance().getAllClusters().getPrecisionMicro());
		});
		
		builder.addColumn("Facts-Recall", (p) -> {
			return DecimalFormatter.format(p.getFusionSetPerformance().getAllClusters().getRecallMicro());
		});
		
		builder.addColumn("Facts-FMeasure", (p) -> {
			return DecimalFormatter.format(p.getFusionSetPerformance().getAllClusters().getFMeasureMicro());
		});
		
		
		return builder.build();
		
	}

	public Map<SetExpansionStrategy, FusionPerformance<PipelinePerformance>> gsRunNew(Map<ClassId, Set<TableRowId>> input,
			Map<ClassId, TableRowIdsSet> fullInput, FusionConfig fusionConfig, String thresholdsFolder,
			SetExpansionStrategiesConfig setExpansionStrategiesConfig, StrategiesConfig strategiesConfig, Map<ClassId, GoldStandardClusters> goldStandard,
			Map<ClassId, Set<InstanceId>> newEntities, Map<ClassId, Set<ClassId>> allParents,
			Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>> correctMap, Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>> fullCorrectMap,
			String caseFolder, Map<ClassId, LearnedModelRepository> weightsRepository,
			Map<ClassId, EM_LearnedModelRepository> instanceMatchingWeightsRepository,
			Map<ClassId, Set<PropertyId>> nonTailPropertiesMap, boolean withCacheAndDisabledBlocking, int run, EntityMatchingRun entityMatchingRun)
			throws IOException {

		String clusteringOutputFolder = caseFolder + "/clusterOutput/";

		MKDIR.createDirUnchecked(new File(clusteringOutputFolder));

		String fusionErrorFolder = null;
		if (caseFolder != null) {
			fusionErrorFolder = caseFolder + "/fusionError/";
			MKDIR.createDirUnchecked(new File(fusionErrorFolder));
		}

		implicitAttributeFinder.prepare(nonTailPropertiesMap);
		rowClustererRepository.init();

		List<SetExpansionStrategy> strategies = new ArrayList<>();
		// strategies.add(SetExpansionStrategiesRepository.FUSION_TESTER);
		// strategies.add(SetExpansionStrategiesRepository.GS_CLUSTERER);
		if (false) {

		strategies.add(SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY);
			// strategies.add(SetExpansionStrategiesRepository.BAG_OF_WORDS_TABLE_AND_LABEL_WITHOUT_TFIDF);
			// strategies.add(SetExpansionStrategiesRepository.ASSOCIATION_LABEL_COMBINED);
			//
			// strategies.add(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_DIRECT_FULL_INDIVIDUAL_OPTIMIZED);
			//
			// strategies.add(SetExpansionStrategiesRepository.NEW_COMBINED);
			// strategies.add(SetExpansionStrategiesRepository.ALL_NO_AA);

		}
		//
		// ////


		strategies.addAll(setExpansionStrategiesConfig.getStrategies());

		// strategies.add(SetExpansionStrategiesRepository.BAG_OF_WORDS_TABLE_AND_LABEL);
		//
		// strategies.add(SetExpansionStrategiesRepository.BAG_OF_WORDS);
		// strategies.add(SetExpansionStrategiesRepository.BAG_OF_WORDS_WITH_TABLE);
		// strategies.add(SetExpansionStrategiesRepository.BAG_OF_WORDS_TABLE_DEVIATION);
		// strategies.add(SetExpansionStrategiesRepository.BAG_OF_WORDS_TABLE_AND_LABEL);
		// strategies.add(SetExpansionStrategiesRepository.BAG_OF_WORDS_TABLE_AND_LABEL_COVARIANCE);
		// strategies.add(SetExpansionStrategiesRepository.BAG_OF_WORDS_TABLE_AND_LABEL_DEVIATION);
		//
		//
		// strategies.add(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_DIRECT_FULL);
		// strategies.add(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_DIRECT_FULL_CORRELATION);
		// strategies.add(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_DIRECT_FULL_DEVIATION);
		//
		//
		// strategies.add(SetExpansionStrategiesRepository.ASSOCIATION_LABEL_PHI);
		// strategies.add(SetExpansionStrategiesRepository.ASSOCIATION_LABEL_PMI);
		// strategies.add(SetExpansionStrategiesRepository.ASSOCIATION_LABEL_TFIDF);
		// strategies.add(SetExpansionStrategiesRepository.ASSOCIATION_LABEL_COMBINED);
		//
		// strategies.add(SetExpansionStrategiesRepository.ASSOCIATION_LABEL_PHI_DEVIATION);
		// strategies.add(SetExpansionStrategiesRepository.ASSOCIATION_LABEL_PMI_DEVIATION);
		// strategies.add(SetExpansionStrategiesRepository.ASSOCIATION_LABEL_TFIDF_DEVIATION);
		// strategies.add(SetExpansionStrategiesRepository.ASSOCIATION_LABEL_COMBINED_DEVIATION);

		// strategies.add(SetExpansionStrategiesRepository.NEW_COMBINED_PREDICTION);
		// strategies.add(SetExpansionStrategiesRepository.NEW_COMBINED_DEVIATION);

		// strategies.add(SetExpansionStrategiesRepository.BAG_OF_WORDS_TABLE_STOPWORDS);
		// strategies.add(SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY_WITH_SURFACE_FORMS);
		// strategies.add(SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY_WITH_SURFACE_FORMS_AND_STOP_WORDS);
		// strategies.add(SetExpansionStrategiesRepository.FULL_LABEL_COSINE_WITH_NEGATIVE);
		// strategies.add(SetExpansionStrategiesRepository.FULL_LABEL_COSINE_WITH_TF_IDF);

		// strategies.add(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_DIRECT);
		// strategies.add(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_DIRECT_WITH_SF);
		// strategies.add(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_DIRECT_WITH_NR);
		// strategies.add(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_DIRECT_FULL);
		// strategies.add(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_DIRECT_FULL_CORRELATION);
		//
		//
		// strategies.add(SetExpansionStrategiesRepository.NEW_COMBINED);
		// strategies.add(SetExpansionStrategiesRepository.NEW_COMBINED_PREDICTION);

		// builder.addStrategy();
		// builder.addStrategy(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_DIRECT);
		// builder.addStrategy(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_KBT_SCORED);
		// //
		// builder.addStrategy(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_KBT_THRESHOLDED);
		// builder.addStrategy(SetExpansionStrategiesRepository.NEW_ATTRIBUTE_KBT_THRESHOLDED_AND_SCORED);
		// builder.addStrategy(SetExpansionStrategiesRepository.BAG_OF_WORDS);

		ColumnList<PipelinePerformance> columnList = getColumnListforPipelinePerformance();

		ColumnListExtendedOutputter<PipelinePerformance> outputter = columnListExtendedOutputterGenerator.generate(columnList);

		LinkedHashMap<SetExpansionStrategy, ConcurrentAccumulator<PipelinePerformance>> totalMap = new LinkedHashMap<>();

		Map<ClassId, LinkedHashMap<SetExpansionStrategy, ConcurrentAccumulator<PipelinePerformance>>> classIdMap = new LinkedHashMap<>();

		Map<ClassPropertyId, LinkedHashMap<SetExpansionStrategy, ConcurrentAccumulator<PipelinePerformance>>> classPropertyIdMap = new LinkedHashMap<>();

		Set<TableRowId> rowIdsAllClasses = new HashSet<>();
		for (TableRowIdsSet rowIds : fullInput.values()) {
			rowIdsAllClasses.addAll(rowIds.getRowIds());
		}
		TableRowIdsSet rowIdsSetAllClasses = new TableRowIdsSet(rowIdsAllClasses, new UuidStringId("ALL"));

		for (Entry<ClassId, Set<TableRowId>> entry : input.entrySet()) {


			ClassId classId = entry.getKey();
			System.out.println("ClassId: " + classId.getFinalPart());

			Set<PropertyId> nonTailProperties = nonTailPropertiesMap.get(classId);

			Set<TableRowId> candidates = entry.getValue();
			TableRowIdsSet fullCandidates = fullInput.get(classId);

			TableIdsSet tableIdsSet = new TableIdsSet(this.tablesRepository.getAllTableIds(classId), new UuidStringId(classId.getFinalPart()));
			if (tableIdsSet.getTableIds().isEmpty()) {
				System.out.println("No tables found for: " + classId.getClassId());
				throw new IllegalStateException();
			}

			SplitGroupedValues splitGroupedValues = getSplitGroupedValues(fusionConfig, tableIdsSet);

			Map<ClassId, Set<InstanceId>> instancesWithMatches = getInstancesWithMatches(classId, tableIdsSet.getTableIds(), 1);
			if (instancesWithMatches.get(classId).isEmpty()) {
				System.out.println("No Instances with MinSource found for: " + classId.getClassId());
				throw new IllegalStateException();
			}

			EntitySets entitySets = new EntitySets(instancesWithMatches.get(classId), instancesWithMatches.get(classId));

			List<ThresholdContainer> thresholdContainers = getThresholdContainers(setExpansionStrategiesConfig, thresholdsFolder, classId, fusionConfig,
					strategiesConfig, splitGroupedValues);

			RowClusteringConfig rowClusteringConfig = new RowClusteringConfig(setExpansionStrategiesConfig, thresholdContainers,
					weightsRepository.get(classId), withCacheAndDisabledBlocking, fusionConfig);

			SplitGroupedValuesSets splitGroupedValuesSets = entitySetGroupedValuesSplitter.split(splitGroupedValues, entitySets);

			SplitRowValues splitRowValues = rowValuesGetter.getSplitRowValues(tableRepository, tableIdsSet);

			ClustererInput clustererInput = new ClustererInput(splitGroupedValuesSets, candidates, fullCandidates, tableIdsSet, null, classId, splitRowValues,
					allParents, nonTailPropertiesMap, rowIdsSetAllClasses);



			for (SetExpansionStrategy strategy : strategies) {

				SimPerformance simPerformance = null;
				OktiePerformance clusteringPerformance = null;
				NewDetectionPerformance newDetectionPerformance = null;
				Map<PropertyId, FusionSetPerformance> fusionPerformance = null;

				System.out.println("Strategy: " + strategy.getName());
				ClusterList clusters;

				Function<TableRowIdPair, Map<String, Double>> scoreGetter = (pair) -> {
					return new HashMap<>();
				};


				List<Set<TableRowId>> nonClustersAsSet = null;

				List<Entity> ncEntities = null;

				RowClusterSimilarities clusterSim = null;

				OktiePerformance nonClusteringClusteringPerformance = new OktiePerformance();

				if (strategy.getName().startsWith("GS_CLUSTERER") || strategy == SetExpansionStrategiesRepository.FUSION_TESTER) {
					List<RowCluster> clusterList = new ArrayList<>();
					for (Set<TableRowId> cluster : goldStandard.get(classId).getTestingSetMappingClusters().values()) {
						clusterList.add(new RowCluster(classId, new ArrayList<>(cluster), null));
					}
					clusters = new ClusterList(clusterList);
					simPerformance = new SimPerformance();

				} else if (strategy == SetExpansionStrategiesRepository.EXTERNAL) {
					List<RowCluster> clusterList = new ArrayList<>();
					for (Set<TableRowId> cluster : externalClusteringImporter.get(classId, "COMBINED_RandomForest", run)) {
						clusterList.add(new RowCluster(classId, new ArrayList<>(cluster), null));
					}
					clusters = new ClusterList(clusterList);
					simPerformance = new SimPerformance();
				} else {
					RowClusterer clusterer = rowClustererRepository.get(strategy.getClusterer());
					RowClusterSimilarities sim = clusterer.clusterRows(rowClusteringConfig, clustererInput);

					if (false) {
					ClusterPairsExporter.writePairList(strategy.getName(), sim, classId, input.get(classId), run,
							goldStandard.get(classId).getTestingSetMappingClusters());
					}

					{
						clusterSim = sim;
					}
					if (sim instanceof AverageScoresSplitterSimilarities) {
						AverageScoresSplitterSimilarities castedSim = (AverageScoresSplitterSimilarities) sim;
						scoreGetter = (pair) -> {
							return castedSim.getIndividualScores(classId, pair.getId1(), pair.getId2());
						};
					}
					else if (sim instanceof WekaLearnableRowClusterSimilarities) {
						WekaLearnableRowClusterSimilarities castedSim = (WekaLearnableRowClusterSimilarities) sim;
						scoreGetter = (pair) -> {
							RecordFeature fv = castedSim.getFeatureVector(pair.getId1(), pair.getId2(), classId);
							Map<String, Double> ret = new LinkedHashMap<>();
							for (String feature : fv.getDatasetData().getFeatures().getFeatures().values()) {
								ret.put(feature, fv.getFeature(feature));
							}
							ret.put("OUT", castedSim.get(pair.getId1(), pair.getId2(), classId));

							return ret;
						};
					}
					else if (sim instanceof CombinedRowClusterSimilarities) {
						if (((CombinedRowClusterSimilarities) sim).getASim() instanceof WekaLearnableRowClusterSimilarities) {
							WekaLearnableRowClusterSimilarities castedSim = (WekaLearnableRowClusterSimilarities) ((CombinedRowClusterSimilarities) sim)
									.getASim();
							scoreGetter = (pair) -> {
								RecordFeature fv = castedSim.getFeatureVector(pair.getId1(), pair.getId2(), classId);
								Map<String, Double> ret = new LinkedHashMap<>();
								for (String feature : fv.getDatasetData().getFeatures().getFeatures().values()) {
									ret.put(feature, fv.getFeature(feature));
								}
								ret.put("OUT", castedSim.get(pair.getId1(), pair.getId2(), classId));

								return ret;
							};
						}
					}

					simPerformance = simEvaluator.evaluateSimNew(sim, rowClusteringConfig, clustererInput, goldStandard.get(classId),
							strategy.getOptimizableClassCluster());

					clusters = greedyClusteringSplitter.clusterRows(sim, rowClusteringConfig, clustererInput,
							strategy.getOptimizableClassCluster(), false, withCacheAndDisabledBlocking);

					if (false) {
						externalClusteringImporter.writeYaserSol(strategy.getName(), classId, clusters, run);
					}
					ClusterList nonClusteringClusters = simEvaluator.getClusterList(classId, sim, rowClusteringConfig, clustererInput,
							goldStandard.get(classId), strategy.getOptimizableClassCluster(), withCacheAndDisabledBlocking);

					ncEntities = entityCreator.createEntites(splitGroupedValuesSets, rowClusteringConfig.getThresholdContainers(), fusionConfig,
							strategiesConfig, rowClusteringConfig.getSetExpansionStrategiesConfig(), nonClusteringClusters.getClusterList(), clustererInput);

					nonClustersAsSet = nonClusteringClusters.getClusterList().stream().map((e) -> new HashSet<>(e.getRows()))
							.collect(Collectors.toList());

					nonClusteringClusteringPerformance = oktieClusterEvaluator.oktieEvaluateWithGsMap(goldStandard.get(classId).getTestingSetMappingClusters(),
							nonClustersAsSet, newEntities.get(classId));

				}

				GsStatistics gsStatistics = new GsStatistics(goldStandard.get(classId), correctMap.get(classId), newEntities.get(classId));

				List<Entity> entities = entityCreator.createEntites(splitGroupedValuesSets, rowClusteringConfig.getThresholdContainers(), fusionConfig,
						strategiesConfig, rowClusteringConfig.getSetExpansionStrategiesConfig(), clusters.getClusterList(), clustererInput);// .stream().filter((e)
																															// ->
																															// e.getValues().size()
																															// >
																															// 0).collect(Collectors.toList());

				List<Set<TableRowId>> clustersAsSet = entities.stream().map((e) -> new HashSet<>(e.getRowCluster().getRows())).collect(Collectors.toList());

				clusteringPerformance = oktieClusterEvaluator.oktieEvaluateWithGsMap(goldStandard.get(classId).getTestingSetMappingClusters(), clustersAsSet,
						newEntities.get(classId));

				if (caseFolder != null && strategy != SetExpansionStrategiesRepository.EXTERNAL && strategy != SetExpansionStrategiesRepository.FUSION_TESTER
						&& !strategy.getName().startsWith("GS_CLUSTERER")) {
					try {

						FourConsumer<String, List<Set<TableRowId>>, Function<TableRowIdPair, Map<String, Double>>, List<Entity>> consumer = (suffix, cc, sc,
								ent) -> {

						};

						outputErrorCases(caseFolder, strategy, goldStandard.get(classId), classId, "", clustersAsSet, scoreGetter, entities);

						if (nonClustersAsSet != null) {
							outputErrorCases(caseFolder, strategy, goldStandard.get(classId), classId, "NC", nonClustersAsSet, scoreGetter, ncEntities);
						}

						String blocking = caseFolder + "/blockingEvaluation";

						MKDIR.createDir(new File(blocking));

						String blockingStrategyFolder = blocking + "/" + strategy.getName();

						caseCreator.createForBlockers(blockingStrategyFolder, classId.getFinalPart(), classId,
								goldStandard.get(classId).getTestingSetMappingClusters(), clusterSim);

					} catch (Exception e) {
						e.printStackTrace();
						int sdf = 123;
					}
				}


				Map<Integer, RowClusterUUID> indexToCluster = ListIterable.iterable(entities).stream().collect(Collectors.toMap(ListEntry::getIndex, (e) -> e.getValue().getRowClusterUUID()));
				
				Map<Integer, Optional<InstanceId>> simpleMajorityMapping = oktieClusterEvaluator.simpleMajorityMap(goldStandard.get(classId).getIndex(), clustersAsSet);
				
				Map<RowClusterUUID, Optional<InstanceId>> remapped = simpleMajorityMapping.entrySet().stream()
						.collect(Collectors.toMap((e) -> indexToCluster.get(e.getKey()), Entry::getValue));
				

				Map<InstanceId, List<ScoredEntity>> mappedEntities;
				List<ScoredEntity> remainingClusters;

				if (strategy == SetExpansionStrategiesRepository.FUSION_TESTER) {

					newDetectionPerformance = new NewDetectionPerformance();

					mappedEntities = new HashMap<>();
					remainingClusters = new ArrayList<>();

					for (Entity e : entities) {
						Optional<InstanceId> instanceId = remapped.get(e.getRowClusterUUID());
						if (!instanceId.isPresent()) {
							int sdf = 123;
						} else {
							if (newEntities.get(classId).contains(instanceId.get())) {
								remainingClusters.add(new ScoredEntity(e, 0.0));
							} else {
								CHMHelper.getFromMap(mappedEntities, instanceId.get(), () -> new ArrayList<>()).add(new ScoredEntity(e, 1.0));
							}
						}
					}


				}
				else {

					remainingClusters = new ArrayList<>();
					List<Entity> unsureClusters = new ArrayList<>();

					if (strategy == SetExpansionStrategiesRepository.INSTANCE_BASED_CLUSTERER) {
						Map<InstanceId, Entity> mappingInner = clusterListEvaluator.maxMajorityMappings(entities);

						mappedEntities = mappingInner.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
							List<ScoredEntity> list = new ArrayList<>();
							list.add(new ScoredEntity(e.getValue(), 1.0));
							return list;
						}));

					} else {
						EntityMatcherConfig emConfig = new EntityMatcherConfig(clustererInput, nonTailProperties, allParents);
						EntityMatcherWeightedResults matcherReturn = entityMatchingRuner.weightedRun(emConfig,
								entityMatchingRun,
								entities, allParents, instanceMatchingWeightsRepository.get(classId));

						// Tri<Map<InstanceId, List<ScoredEntity>>,
						// List<Entity>, List<Entity>> matcherReturn =
						// entityMatcher
						// .matchMultipleAndUnsureAndNew(entities, allParents);

						mappedEntities = matcherReturn.getExistingEntities();
						remainingClusters = matcherReturn.getNewEntities();
						unsureClusters = matcherReturn.getBadEntities();
		

					}

					// Pair<OktiePerformance, Map<InstanceId, Set<TableRowId>> >
					// newDetectionPair =
					// this.newDetectionEvaluator.evaluate(goldStandard.get(classId),
					// entities, mappingForExistingClusters,
					// newEntities.get(classId));

					newDetectionPerformance = this.newDetectionEvaluator.evaluateAccuracy(newEntities.get(classId), remapped, mappedEntities,
							remainingClusters, goldStandard.get(classId));

				}
				
				Map<InstanceId, List<Entity>> mappingForExistingClusters;
				mappingForExistingClusters = mappedEntities.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
					return e.getValue().stream().map((e2) -> e2.getEntity()).collect(Collectors.toList());
				}));

				List<Entity> remainingClustersUnscored = remainingClusters.stream().map((e) -> e.getEntity()).collect(Collectors.toList());
				Map<InstanceId, Set<TableRowId>> clusterChoice = this.newDetectionEvaluator.chooseClustersForFusion(goldStandard.get(classId),
						remainingClustersUnscored, mappingForExistingClusters, newEntities.get(classId), remapped);

				Map<InstanceId, Set<TableRowId>> entitiesFoundClusterChoice = this.newDetectionEvaluator.chooseClustersForEntitiesFound(
						goldStandard.get(classId), remainingClustersUnscored, mappingForExistingClusters, newEntities.get(classId), remapped);

				
				EntitiesPerformance entitiesPerformance = this.entitiesPerformanceEvaluator.evaluate(goldStandard.get(classId), newEntities.get(classId),
						entitiesFoundClusterChoice.keySet(), mappingForExistingClusters, remainingClustersUnscored);

				RankedEvaluationPerformance rePerformance = rankingEvalutator.chooseClustersForEntitiesFound(goldStandard.get(classId), remainingClusters,
						newEntities.get(classId), remapped);
				
				List<RowCluster> rowClusters = new ArrayList<>();
				Map<RowClusterUUID, InstanceId> mapping = new HashMap<>();

				for (Entry<InstanceId, Set<TableRowId>> newMappingEntry : clusterChoice.entrySet()) {

					RowCluster rc = new RowCluster(classId, new ArrayList<>(newMappingEntry.getValue()), null);
					rowClusters.add(rc);
					mapping.put(rc.getUuid(), newMappingEntry.getKey());
				}

				List<Entity> entitiesAfterDetection = entityCreator.createEntites(splitGroupedValuesSets, thresholdContainers, fusionConfig, strategiesConfig,
						setExpansionStrategiesConfig, rowClusters, clustererInput);

				Map<InstanceId, List<Entity>> entitiesAfterDetectionMap = entitiesAfterDetection.stream()
						.collect(Collectors.toMap((e) -> mapping.get(e.getRowClusterUUID()), (e) -> {
							List<Entity> retList = new ArrayList<>();
							retList.add(e);
							return retList;
						}));

				Map<QualifiedTripleId, List<FusedValue>> fusionSet = fusionSetCreator.fuse4(entitiesAfterDetectionMap);

				if (strategy == SetExpansionStrategiesRepository.FUSION_TESTER) {
					if (fusionErrorFolder != null) {
						String performanceOutput = fusionSetEvaluator.printPerformance(clusterChoice, goldStandard.get(classId), newEntities.get(classId),
								fusionSet, correctMap.get(classId), fullCorrectMap.get(classId));

						FileUtils.writeStringToFile(new File(fusionErrorFolder + "/" + classId.getFinalPart() + ".html"), performanceOutput, "UTF-8");

					}
				}

				fusionPerformance = fusionSetEvaluator.evaluate(fusionSet, correctMap.get(classId), newEntities.get(classId), fullCorrectMap.get(classId));

				{
					PipelinePerformance pipelinePerformance = new PipelinePerformance(gsStatistics, simPerformance, nonClusteringClusteringPerformance,
							clusteringPerformance,
							newDetectionPerformance, entitiesPerformance, rePerformance, new FusionSetPerformance());
		
						CHMHelper.getFromMap(totalMap, strategy, () -> new ConcurrentAccumulator<>(new PipelinePerformance())).add(pipelinePerformance);
		
						LinkedHashMap<SetExpansionStrategy, ConcurrentAccumulator<PipelinePerformance>> classIdInnerMap = CHMHelper.getFromMap(classIdMap, classId,
								() -> new LinkedHashMap<>());
						CHMHelper.getFromMap(classIdInnerMap, strategy, () -> new ConcurrentAccumulator<>(new PipelinePerformance())).add(pipelinePerformance);

				}
				
				
				for (Entry<PropertyId, FusionSetPerformance> fusionEntry : fusionPerformance.entrySet()) {

					// FusionPerformanceSynchronizedCreator<PipelinePerformance>
					// performance =
					// CHMHelper.getFromMap(pipelinePerformanceCreator,
					// strategy, () -> new
					// FusionPerformanceSynchronizedCreator<PipelinePerformance>(new
					// PipelinePerformance()));

					PipelinePerformance pipelinePerformance = new PipelinePerformance(new GsStatistics(), new SimPerformance(),
							new OktiePerformance(), new OktiePerformance(),
							new NewDetectionPerformance(), new EntitiesPerformance(), rePerformance, fusionEntry.getValue());

					CHMHelper.getFromMap(totalMap, strategy, () -> new ConcurrentAccumulator<>(new PipelinePerformance())).add(pipelinePerformance);

					LinkedHashMap<SetExpansionStrategy, ConcurrentAccumulator<PipelinePerformance>> classIdInnerMap = CHMHelper.getFromMap(classIdMap, classId,
							() -> new LinkedHashMap<>());
					CHMHelper.getFromMap(classIdInnerMap, strategy, () -> new ConcurrentAccumulator<>(new PipelinePerformance())).add(pipelinePerformance);

					LinkedHashMap<SetExpansionStrategy, ConcurrentAccumulator<PipelinePerformance>> classPropertyIdInnerMap = CHMHelper
							.getFromMap(classPropertyIdMap, new ClassPropertyId(classId, fusionEntry.getKey()), () -> new LinkedHashMap<>());
					CHMHelper.getFromMap(classPropertyIdInnerMap, strategy, () -> new ConcurrentAccumulator<>(new PipelinePerformance()))
							.add(pipelinePerformance);

				}

			}
		}

		GenericResultsWriter<PipelinePerformance> writer = new GenericResultsWriter<>(columnList);
		Map<SetExpansionStrategy, FusionPerformanceCreator<PipelinePerformance>> accumulatorMap = new HashMap<>();

		for (Entry<SetExpansionStrategy, ConcurrentAccumulator<PipelinePerformance>> entry : totalMap.entrySet()) {
			PipelinePerformance pp = entry.getValue().get();
			writer.writeTotal(entry.getKey().getName(), pp);
			CHMHelper.getFromMap(accumulatorMap, entry.getKey(), () -> FusionPerformance.accumulator(new PipelinePerformance())).addTotal(pp);

		}

		for (Entry<ClassId, LinkedHashMap<SetExpansionStrategy, ConcurrentAccumulator<PipelinePerformance>>> classIdEntry : classIdMap.entrySet()) {

			for (Entry<SetExpansionStrategy, ConcurrentAccumulator<PipelinePerformance>> entry : classIdEntry.getValue().entrySet()) {
				PipelinePerformance pp = entry.getValue().get();
				writer.writeClassId(entry.getKey().getName(), classIdEntry.getKey(), pp);
				CHMHelper.getFromMap(accumulatorMap, entry.getKey(), () -> FusionPerformance.accumulator(new PipelinePerformance()))
						.addPerClassId(classIdEntry.getKey(), pp);

			}
		}

		for (Entry<ClassPropertyId, LinkedHashMap<SetExpansionStrategy, ConcurrentAccumulator<PipelinePerformance>>> classIdEntry : classPropertyIdMap
				.entrySet()) {

			for (Entry<SetExpansionStrategy, ConcurrentAccumulator<PipelinePerformance>> entry : classIdEntry.getValue().entrySet()) {

				writer.writeClassPropertyId(entry.getKey().getName(), classIdEntry.getKey().getClassId(), classIdEntry.getKey().getPropertyId(),
						entry.getValue().get());

			}
		}

		System.out.println(writer.getAll(ResultsCategory.TOTAL));
		System.out.println();
		System.out.println(writer.getAll(ResultsCategory.CLASS_ID));

		// writer.getAll(ResultsCategory.CLASS_PROPERTY_ID_WITH_DATASET)
		//
		// System.out.println(simWriter.getAll(ResultsCategory.CLASS_ID));
		// System.out.println(allWriter.getAll(ResultsCategory.CLASS_ID));
		// System.out.println(existingWriter.getAll(ResultsCategory.CLASS_ID));
		// System.out.println(newWriter.getAll(ResultsCategory.CLASS_ID));
		int sdf = 123;

		return accumulatorMap.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().get()));

	}
	
	private void outputErrorCases(String caseFolder, SetExpansionStrategy strategy, GoldStandardClusters goldStandard, ClassId classId, String suffix,
			List<Set<TableRowId>> clustersAsSet, Function<TableRowIdPair, Map<String, Double>> scoreGetter, List<Entity> entities) {
		Map<InstanceId, OktieMapping> mapping = oktieClusterEvaluator.oktieMapping(oktieClusterEvaluator.getIndex(goldStandard.getTestingSetMappingClusters()),
				clustersAsSet);

		Map<InstanceId, Integer> clusterOverlapMap = new HashMap<>();
		for (Entry<InstanceId, OktieMapping> mappingEntry : mapping.entrySet()) {
			clusterOverlapMap.put(mappingEntry.getKey(), mappingEntry.getValue().getCluster());
		}
		String evaluation = caseFolder + "/simEvaluation" + suffix;

		MKDIR.createDir(new File(evaluation));

		String strategyFolder = evaluation + "/" + strategy.getName();
		MKDIR.createDir(new File(strategyFolder));
		if (DO_WRITE_CASES) {
			caseCreator.createWithErrors(strategyFolder, classId.getFinalPart(), entities, clusterOverlapMap, goldStandard.getTestingSetMappingClusters(),
					scoreGetter);
		}

	}

	public static boolean DO_WRITE_CASES = true;
	
	
	
	private <F> int getIntersection(Set<F> gs, Set<F> cluster) {
		int intersection = 0;
		if(gs.size() < cluster.size()) {
			for(F point : gs) {
				if(cluster.contains(point)) {
					intersection++;
				}
			}
			
		}
		else {
			for(F point : cluster) {
				if(gs.contains(point)) {
					intersection++;
				}
			}
		}

		return intersection;
	}
	
	
	public List<ThresholdContainer> getThresholdContainers(SetExpansionStrategiesConfig setExpansionStrategiesConfig, String thresholdsFolder, ClassId classId, FusionConfig fusionConfig, StrategiesConfig strategiesConfig, SplitGroupedValues splitGroupedValues) throws IOException {
		List<ThresholdContainer> thresholdContainers = new ArrayList<>();
		for (int i = 0; i < setExpansionStrategiesConfig.getFusionSetups().size(); i++) {
			
			if(false) {

				thresholdContainers.add(new ThresholdConstant(-0.1));
			}
			else {

				Optional<ThresholdContainer> thresholdContainerOptional = Optional.empty();

				Pair<FusionStrategy, PropertiesConfig> strategy = setExpansionStrategiesConfig.getFusionSetups().get(i);
				String fileName = classId.getFinalPart() + "_" + strategy.getFirst().getName() + "_" + fusionConfig.isAdjustedRecall();

				if (strategy.getFirst().getFusionMethod() == FusionMethod.WEIGHTED_SCORED || strategy.getFirst().getFusionMethod() == FusionMethod.SCORED) {
					

					if (thresholdsFolder != null) {
						MKDIR.createDirUnchecked(new File(thresholdsFolder));

						thresholdContainerOptional = this.thresholdKryoCache.readThresholdContainer(thresholdsFolder, fileName);
					}

					if (thresholdContainerOptional.isPresent()) {
						thresholdContainers.add(thresholdContainerOptional.get());
						System.out.println(thresholdContainerOptional.get());
					} else {
						System.out.println("Threshold" + strategy.getFirst().getName() + "running");

						Map<Integer, SplitGroupedValuesSets> splitGroupedValuesSets = this.learningSetCreatorStratified.createByTableProperty(4,
								splitGroupedValues);

						ThresholdContainer thresholdContainer = scoreFusionThresholdFinderFullQualifiedSimple.findOptimumThresholdCrossValidation(
								splitGroupedValuesSets, strategy.getFirst(), strategiesConfig, fusionConfig, strategy.getSecond());

						if (thresholdsFolder != null) {
							this.thresholdKryoCache.writeThresholdContainer(thresholdContainer, thresholdsFolder, fileName);
						}

						thresholdContainers.add(thresholdContainer);
						System.out.println("Threshold" + strategy.getFirst().getName() + "done");
						System.out.println(thresholdContainer);

					}
			 }
				else {
					thresholdContainers.add(new ThresholdConstant(-0.1));
				}
			}
		}
		return thresholdContainers;
	}

	boolean atLeastOneOverlap(Set<AnnotatedTableId> set1, Set<AnnotatedTableId> set2) {
		if(set1.size() < set2.size()) {
			for(AnnotatedTableId tid : set1) {
				if(set2.contains(tid)) {
					return true;
				}
			}
		}
		else {
			for(AnnotatedTableId tid : set2) {
				if(set1.contains(tid)) {
					return true;
				}
			}
		}
		return false;
	}
	
	boolean atLeastOverlap(Set<AnnotatedTableId> set1, Set<AnnotatedTableId> set2,int minOverlap) {
		int overlap = 0;
		if(set1.size() < set2.size()) {
			for(AnnotatedTableId tid : set1) {
				if(set2.contains(tid)) {
					overlap++;
					if(overlap == minOverlap) {
						return true;
					}
				}
			}
		}
		else {
			for(AnnotatedTableId tid : set2) {
				if(set1.contains(tid)) {
					overlap++;
					if(overlap == minOverlap) {
						return true;
					}
				}
			}
		}
		if(overlap >= minOverlap) {
			return true;
		}
		return false;
	}
	
	public Map<ClassId, Set<InstanceId>> getInstancesWithMatches(ClassId setId, Set<AnnotatedTableId> set, int minSource) {
		Map<ClassId, Set<InstanceId>> ret = new HashMap<>();
		ret.put(setId, 	knowledgeBase.getInstances(setId).keySet().stream().filter((e) ->    tableRepository.hasInstance(e) && 
				atLeastOverlap(set, tableRepository.getTables(e), minSource)
				 ).collect(Collectors.toSet())); 
		return ret;
	}
	
	
	public Map<ClassId, Set<TableRowId>> getUnmatchedRows(ClassId setId, Set<AnnotatedTableId> set ) {
		
		Map<ClassId, Set<TableRowId>> ret = new HashMap<>();
		ret.put(setId, 	tableRepository.getUnmatchedRowIdsForTables(set)); 
		return ret;
	}
	
	private   String getDistribution( List<Pair<String, Collection<Integer>>>  collection) {
		
		List<TreeMap<Integer, MutableInt>> list = new ArrayList<>();
		
		TreeSet<Integer> columns = new TreeSet<>();
		
		for(int k = 0; k < collection.size(); k++) {
			TreeMap<Integer, MutableInt> sortedMap = new TreeMap<>();
			
			for(Integer i : collection.get(k).getSecond()) {
				MutableInt mi = CHMHelper.getFromMap(sortedMap, i, () -> new MutableInt());
				mi.increment();
			}
			
			list.add(sortedMap);
			columns.addAll(sortedMap.keySet());
		}
		
		StringBuilder sb = new StringBuilder();
		
		sb.append("ClassId;");
		for(Iterator<Integer> it = columns.iterator(); it.hasNext();) {
			sb.append("" + it.next() + (it.hasNext() ? ";" : ""));
		}
		
		sb.append("\n");
		
		for(int i = 0; i < collection.size(); i++) {
			Pair<String, Collection<Integer>> pair = collection.get(i);
			sb.append(pair.getFirst() + ";");
			
			for(Iterator<Integer> it = columns.iterator(); it.hasNext();) {
				MutableInt mi = list.get(i).get(it.next());
				sb.append("" + ( mi != null ? mi.intValue() : "") + (it.hasNext() ? ";" : ""));
			}
			
			if(i < collection.size() - 1) {
				sb.append("\n");
			}
			 
			
		}
		
		return sb.toString();
		
	}

	public static void printDistribution(Collection<Integer> collection) {
		TreeMap<Integer, MutableInt> sortedMap = new TreeMap<>();
		
		for(Integer i : collection) {
			MutableInt mi = CHMHelper.getFromMap(sortedMap, i, () -> new MutableInt());
			mi.increment();
		}
		
		StringBuilder firstLine = new StringBuilder();
		StringBuilder secondLine = new StringBuilder();
		boolean first = true;
		for(Entry<Integer, MutableInt> entry : sortedMap.entrySet()) {
			String suffix = ";";
			if(first) {
				suffix = "";
				first = false;
			}
			
			firstLine.append(suffix + entry.getKey());
			secondLine.append(suffix + entry.getValue().intValue());
			
			 
		}
		
		System.out.println(firstLine.toString());
		System.out.println(secondLine.toString());
		 
	}

	
	
	private void addSlotFillingPerformanceWithList(ClassId classId, FusionPerformanceCreator<SlotFillingPerformance> slotFillingPerformance,
			Map<InstanceId, List<ScoredEntity>> mappedEntities) {
		
		 
		
		PersistentParallel6.runAndFinish(mappedEntities.entrySet(), (e) -> {
			int clusters = e.getValue().size();
			int values = e.getValue().stream().mapToInt((e2) -> e2.getEntity().getValues().size()).sum();
			int newValues = 0;
			
			for(ScoredEntity se : e.getValue()) {
				for(QualifiedPropertyId qpid : se.getEntity().getValues().keySet()) {
					if(!knowledgeBase.hasValue(new ValueId(e.getKey(), qpid.getPropertyId())))
						newValues++;
				}
				
			}
			
			slotFillingPerformance.addPerClassId(classId, new SlotFillingPerformance(1, clusters, values, newValues));
			slotFillingPerformance.addTotal(new SlotFillingPerformance(1, clusters, values, newValues));
		}, this.getClass());
		
	}

	private void addSlotFillingPerformance(ClassId classId, FusionPerformanceCreator<SlotFillingPerformance> slotFillingPerformance, Map<InstanceId, Entity> mapping) {
		addSlotFillingPerformanceWithList(classId, slotFillingPerformance, mapping.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			List<ScoredEntity> list = new ArrayList<>();
			list.add(new ScoredEntity(e.getValue(), 1.0));
			return list;
		})));
		
	}


	public static enum PerformanceMeasure {
		Name(0),
		ClutersCreated(4), AverageClusterPurity(6), TotalInstancesWithAtLeastOneCluster(7), TotalInstancesWithExactlyOneCluster(8),ClusterFmeasure(9), AverageDistinctInstanceInCluster(10), MedianDistinctInstanceInCluster(11),TriplePresent(14),PresentWithMultiple(15),PresentWithNoCorrect(16),PresentWithCorrect(17), PresentWithCorrectAndOneOnly(18),
		PropPersent(20),PropPresentWithMultiple(21),PropPresentWithNoCorrect(22),AvgPropOfCorrectPerTriple(23), PropPresentWithCorrectAndOneOnly(24),PropPresentWithCorrect(25),ValuesFmeasure(26);
		
		
		private static final Map<Integer, PerformanceMeasure> map = new HashMap<>();
		private final int index;

		private PerformanceMeasure(int index) {
		 this.index =index;
		
		}
		
		static {
			for(PerformanceMeasure pm : PerformanceMeasure.values()) {
				 map.put(pm.index, pm);
			}
		}
		
		public  static String writeCsv(int totalInstancesInTestingSet, int totalBlocks, int maxCorrect, List<Map<PerformanceMeasure, String>> content) {
		 
				StringWriter sw = new StringWriter();
				CSVWriter writer = new CSVWriter(sw);
				int size = map.size() + 8;
				String[] headerLine = new String[size];
				headerLine[0] = "Name";
				headerLine[1] = "";
				
				headerLine[2] = "TotalInstancesInTestingSet";
				headerLine[3] = "TotalBlocks";
				headerLine[4] = "ClustersCreated";
				headerLine[5] = "";
				
				headerLine[6] = "AverageClusterPurity (Precision)";
				headerLine[7] = "ProportionOfInstancesWithAtLeastOneCluster";
				headerLine[8] = "ProportionOfInstancesWithExactlyOneCluster (Recall)";
				headerLine[9] = "FMeasure";
				
//				headerLine[9] = "";
				
				headerLine[10] = "Average Distinct Instance in Cluster";
				headerLine[11] = "Median Distinct Instance in Cluster";
				
				headerLine[12] = "";
				
				headerLine[13] = "MaxCorrect";
				headerLine[14] = "TriplePresent";
				headerLine[15] = "PresentWithMultiple";
				headerLine[16] = "PresentWithNoCorrect";
				headerLine[17] = "PresentWithCorrect";
				headerLine[18] = "PresentWithCorrectAndOneOnly";
				
				headerLine[19] = "";
				
				headerLine[20] = "PropPresent";
				headerLine[21] = "PropPresentWithMultiple";
				headerLine[22] = "PropPresentWithNoCorrect";
				headerLine[23] = "AvgPropOfCorrectPerTriple(Precision)";
				headerLine[24] = "PropPresentWithCorrectAndOnlyOne(Recall)";
				headerLine[25] = "PropPresentWithCorrect";
				headerLine[26] = "FMeasure";
				
				writer.writeNext(headerLine);
				for(Map<PerformanceMeasure, String> contentMap : content) {
					String[] line = new String[size];
					
					for(int i = 0; i < size;i++) {
						line[i] = contentMap.getOrDefault( map.getOrDefault(i, null), "");
						
					}
					line[2] = "" + totalInstancesInTestingSet;
					line[3] = "" + totalBlocks;
					line[13] = "" + maxCorrect;
					
					writer.writeNext(line);
					
					
				}
				
				try {
					writer.close();
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				
				String ret = sw.toString();
				
				return ret; 
				
	 
		}
		
		
	}

	private void log(String string) {
		System.out.println(string);

	}

	public void runStupid(String runName, FusionConfig fusionConfig, StrategiesConfig strategiesConfig,
			SetExpansionStrategiesConfig setExpansionStrategiesConfig,
			Map<String, Pair<ClassId, List<InstanceId>>> targetSets, String folder, String fileNamePrefix) throws IOException {

		// SetExpansionResult result = new SetExpansionResult();
		//
		// SplitGroupedValues splitGroupedValues =
		// getSplitGroupedValues(fusionConfig);
		//
		// Map<ClassId, Map<String, List<FusionPerformance<PerformancePR>>>>
		// performanceMap = new HashMap<>();
		//
		// for (Entry<String, Pair<ClassId, List<InstanceId>>> targetSetEntry :
		// targetSets.entrySet()) {
		// Map<Integer, EntitySets> entitySets =
		// entitySetsCreator.split(targetSetEntry.getValue().getSecond(), 4);
		// Map<Integer, SplitGroupedValuesSets> splitGroupedValuesSets =
		// entitySetGroupedValuesSplitter.split(splitGroupedValues, entitySets);
		//
		//
		//
		// for (Entry<Integer, EntitySets> entry : entitySets.entrySet()) {
		// Set<TableRowId> rowIds =
		// tablesRepository.getUnmatchedRowIds(entry.getValue().getLearningSet(),
		// targetSetEntry.getValue().getFirst());
		//
		// for (SetExpansionStrategy strategy :
		// setExpansionStrategiesConfig.getStrategies()) {
		// RowClusterer clusterer =
		// rowClustererRepository.get(strategy.getClusterer());
		// List<Entity> entities =
		// entityCreator.createEntites(splitGroupedValuesSets.get(entry.getKey()),
		// fusionConfig, strategiesConfig, clusters);
		// Map<InstanceId, List<Entity>> mappedEntities =
		// entityMatcher.matchMultiple(entities);
		//
		// // result.addValue(classId, set, precision, recall,
		// // fMeasure);
		// }
		// }
		// }
		//
		// result.write(runName, folder, fileNamePrefix);

	}


	
	public SplitGroupedValues getSplitGroupedValues(FusionConfig fusionConfig,   TableIdsSet tableIdsSet) {

		return splitValuesGetter.getSplitGroupedValues(fusionConfig, tableIdsSet);

		// GroupedValues groupedValues =
		// groupedValuesRepository.getGroupedValues(fusionConfig, tableIdsSet );
		//
		// return getSplitGroupedValues(groupedValues, fusionConfig);
	}

	public SplitGroupedValues getSplitGroupedValues(GroupedValues groupedValues, FusionConfig fusionConfig) {

		return splitValuesGetter.getSplitGroupedValues(groupedValues, fusionConfig);

		// SplitterFactory splitterFactory =
		// this.splitterFactoriesRepository.get(fusionConfig.getSplitterFactory());
		// Splitter splitter = splitterFactory.getSplitter(groupedValues,
		// groupedValues, groupedValues);
		// SplitGroupedValues splitGroupedValues =
		// splitter.split(groupedValues);
		//
		// return splitGroupedValues;
	}
	


	public void printTableStatistics(TableIdsSet tableIdsSet, SplitGroupedValues splitGroupedValues) {
		 System.out.println("Number of tables: " + tableIdsSet.getTableIds().size());
		 TableProfiling tableStatistics = tableSetProfiler.getStatistics(tableIdsSet.getTableIds());
		 
		 
		 System.out.println("Number of average columns per table: " + tableStatistics.getAverageColumns());
		 System.out.println("Number of average rows per table: " + tableStatistics.getAverageRows());
		 System.out.println("Number of average mapped columns per table: " + tableStatistics.getAverageMappedColumns());
		 System.out.println("Number of average mapped instances per table: " + tableStatistics.getAverageMappedInstances());
		 
		 System.out.println("Total mapped columns: " + tableStatistics.getTotalMappedColumns());
		 System.out.println("Total mapped rows: " + tableStatistics.getTotalMappedInstances());
		 System.out.println("Total mapped values: " + tableStatistics.getTotalMappedValues());
		 

		 System.out.println("Total mapped unique instances: " + tableStatistics.getTotalUniqueInstancesMapped());
		 
		 double proportionOfUnmappedRows = tableStatistics.getTotalUnmappedRows() / ( (double) (tableStatistics.getTotalUnmappedRows() +  tableStatistics.getTotalMappedInstances() ));

		 System.out.println("Total unmapped rows: " + tableStatistics.getTotalUnmappedRows()  + "(" + ( DecimalFormatter.format(proportionOfUnmappedRows) )+ ")");
		 double proportionOfUnmappedValues = tableStatistics.getTotalUnmappedValues() / ( (double) (tableStatistics.getTotalUnmappedValues() +  tableStatistics.getTotalMappedValues() ));
		 System.out.println("Total unmapped values: " + tableStatistics.getTotalUnmappedValues()   + "(" + ( DecimalFormatter.format(proportionOfUnmappedValues) )+ ")"); 
		 
		 
		 
		 
		 
		 System.out.println("Number of mapped values in tables with correspondence in KB: " +splitGroupedValues.getValues().values().stream().mapToInt((e) -> e.getValues().size()).sum());

		 
		 System.out.println("Distribution of mapped values per Instances: ");
		 printDistribution(tableStatistics.getDistributionOfMappedValuesPerInstance());
		 

		 System.out.println("Distribution of mapped rows per Instances: ");
		 printDistribution(tableStatistics.getDistributionOfMappedRowsPerInstance());
		 
	}



	private static class SetExpansionResult {
		private AcumReference<AccumAverage> precision;
		private AcumReference<AccumAverage> recall;
		private AcumReference<AccumAverage> fMeasure;
		private AccumulatorListImmutable resultsType;
		private ColumnList columnLilst;
		FusionPerformanceSynchronizedCreator<GenericAccumaltableResult> creator;

		public SetExpansionResult() {
			ImmutableAccumulatorListBuild builder = AccumulatorListImmutable.builder();
			precision = builder.addAcummulator(AccumAverage.class, new AccumAverage());
			recall = builder.addAcummulator(AccumAverage.class, new AccumAverage());
			fMeasure = builder.addAcummulator(AccumAverage.class, new AccumAverage());

			resultsType = builder.getResultsType();
			
			ColumnListBuilder columnListBuilder = ColumnList.builderAccumRepository();
			columnListBuilder.addColumn("Precision", (e) -> {
				return String.valueOf(e.get(precision).getValue());
			});
			columnListBuilder.addColumn("Recall", (e) -> {
				return String.valueOf(e.get(recall).getValue());
			});
			columnListBuilder.addColumn("FMeasure", (e) -> {
				return String.valueOf(e.get(fMeasure).getValue());
			});
			
			columnLilst = columnListBuilder.build();

			
			creator = new FusionPerformanceSynchronizedCreator<>(
					GenericAccumaltableResult.empty(resultsType));


		}

		void addValue(ClassId classId, String set, double precision, double recall, double fMeasure) {
			ResultBuilder builder = GenericAccumaltableResult.builder(resultsType);
			builder.setValue(this.precision, new AccumAverage(precision));
			builder.setValue(this.precision, new AccumAverage(recall));
			builder.setValue(this.precision, new AccumAverage(fMeasure));
			GenericAccumaltableResult result = builder.build();

			creator.addPerClassId(classId, result);
			// creator.addPerClassIdAndSet(classId, set, result);
		}

		void write(String run, String folder, String fileNamePrefix) throws IOException {
			GenericResultsWriterAccumRepository writer = new GenericResultsWriterAccumRepository(columnLilst);
			writer.writeTotalAccumRepository(run, creator.get());
			writer.writeClassIdAccumRepository(run, creator.get());
			// writer.writeClassIdAndSet(run, creator.get());
			writer.writeToFolder(ResultsCategory.TOTAL, folder, fileNamePrefix);
			writer.writeToFolder(ResultsCategory.CLASS_ID, folder, fileNamePrefix);
			writer.writeToFolder(ResultsCategory.CLASS_ID_AND_SET, folder, fileNamePrefix);
		}
	}
	
	boolean included(SplitGroupedValuesSets sets, List<ThresholdContainer> thresholdContainers,  
			 SetExpansionStrategiesConfig setExpansionStrategiesConfig, ClassPropertyId classPropertyId,
			TableColumnId tableColumnId) {
		
		Pair<FusionStrategy, PropertiesConfig> chosenFusionSetup = null;
		ThresholdContainer chosenThresholdContainer = null;
		for (int i = 0; i < setExpansionStrategiesConfig.getFusionSetups().size(); i++) {
			Pair<FusionStrategy, PropertiesConfig> fusionSetup = setExpansionStrategiesConfig.getFusionSetups().get(i);
  
			if( propertiesConfigFilter.include(fusionSetup.getSecond(),classPropertyId.getPropertyId())) {
				chosenFusionSetup = fusionSetup;
				chosenThresholdContainer = thresholdContainers.get(i);
			}
			 

		}
		
		if(chosenFusionSetup == null) {
			int sdf= 123;
		}
		
		
		Scorer scorer = scorerRepository.get(chosenFusionSetup.getFirst().getScoringMethod());
		ScoresContainer probabilities = scorer.getScores(sets);
		
		Probability score = probabilities.getScore(new TripleId(classPropertyId.getClassId(), classPropertyId.getPropertyId(), null), new CellId(tableColumnId.getTableId(), tableColumnId.getColumnIdx(), null), new HashSet<>());
		if(score == null) {
			return false;
		}
		else {
//		Probability score = probabilities.getScore(classPropertyId.getClassId(), classPropertyId.getPropertyId(), null, tableColumnId.getTableId(), tableColumnId.getColumnIdx(), null);
		
			Double threshold = chosenThresholdContainer.getScore(classPropertyId.getClassId(), classPropertyId.getPropertyId(), null);
			if(threshold == null)
				return false;
			return score.get(SimilarityType.DIRECT).getAccuracy() > threshold;
		}
	}
}
