package expansion.set.evaluation;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.ConfigInterface;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategy;
import de.mannheim.uni.expansion.utils.containers.Pair;
import expansion.set.strategies.SetExpansionStrategy;

public class SetExpansionStrategiesConfig implements ConfigInterface<SetExpansionStrategiesConfig> {

	List<SetExpansionStrategy> strategies = new ArrayList<>();
	List<Pair<FusionStrategy, PropertiesConfig>> fusionSetups = new ArrayList<>();

	private SetExpansionStrategiesConfig() { 
	}

	public static SetExpansionStrategiesConfig defaultConfig() {
		return new SetExpansionStrategiesConfig();
	}

	@Override
	public SetExpansionStrategiesConfig copy() {
		SetExpansionStrategiesConfig config = new SetExpansionStrategiesConfig();
		config.strategies = new ArrayList<>(this.strategies);
		config.fusionSetups = new ArrayList<>(this.fusionSetups);
		return config;
	}

	@Override
	public SetExpansionStrategiesConfigBuilder rebuilder() {
		return new SetExpansionStrategiesConfigBuilder(this);
	}

	public static SetExpansionStrategiesConfigBuilder builder() {
		return new SetExpansionStrategiesConfigBuilder(defaultConfig());

	}

	public List<SetExpansionStrategy> getStrategies() {
		return Collections.unmodifiableList(strategies);
	}

	public List<Pair<FusionStrategy, PropertiesConfig>> getFusionSetups() {
		return Collections.unmodifiableList(fusionSetups);
	}

	public static class SetExpansionStrategiesConfigBuilder implements ConfigBuilderInterface<SetExpansionStrategiesConfig> {

		SetExpansionStrategiesConfig config;

		public SetExpansionStrategiesConfigBuilder(SetExpansionStrategiesConfig config) {
			this.config = config.copy();
		}

		@Override
		public SetExpansionStrategiesConfig build() {
			return config.copy();
		}

		public SetExpansionStrategiesConfigBuilder addStrategy(SetExpansionStrategy strategy) {
			config.strategies.add(strategy);
			return this;
		}

		public SetExpansionStrategiesConfigBuilder addAllStrategy(List<SetExpansionStrategy> strategies) {
			config.strategies.addAll(strategies);
			return this;
		}

		public SetExpansionStrategiesConfigBuilder clearStrategies() {
			config.strategies.clear();
			return this;
		}

		public SetExpansionStrategiesConfigBuilder addFusionSetup(FusionStrategy fusionStrategy, PropertiesConfig propertiesConfig) {
			config.fusionSetups.add(new Pair<>(fusionStrategy, propertiesConfig));
			return this;
		}

		public SetExpansionStrategiesConfigBuilder clearFusionSetups() {
			config.fusionSetups.clear();
			return this;
		}
	}


}
