package expansion.set.evaluation;

import java.util.function.BinaryOperator;

import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.Accumulateable;

public class ValuesTrustPerformance implements Accumulateable<ValuesTrustPerformance> {

	private final int _0To20;
	private final int _20To40;
	private final int _40To60;
	private final int _60To80;
	private final int _80To100;
	
	private ValuesTrustPerformance(int _0To20, int _20To40, int _40To60, int _60To80, int _80To100) {
		super();
		this._0To20 = _0To20;
		this._20To40 = _20To40;
		this._40To60 = _40To60;
		this._60To80 = _60To80;
		this._80To100 = _80To100;
	}
	
	
	public static ValuesTrustPerformance create(double performance) {
		if(performance < 20.0) {
			return new ValuesTrustPerformance(1, 0, 0, 0, 0);
		}
		else if(performance < 40.0) {
			return new ValuesTrustPerformance(0,1, 0, 0, 0);
			
		}
		else if (performance < 60.0) {
			return new ValuesTrustPerformance(0,0, 1, 0, 0);
		}

		else if (performance < 80.0) {
			return new ValuesTrustPerformance(0,0, 0, 1, 0);
		}
		else {
			return new ValuesTrustPerformance(0,0, 0, 0, 1);
		}
	}


	int get_0To20() {
		return _0To20;
	}


	int get_20To40() {
		return _20To40;
	}


	int get_40To60() {
		return _40To60;
	}


	int get_60To80() {
		return _60To80;
	}


	int get_80To100() {
		return _80To100;
	}

	public static ValuesTrustPerformance accumulater(ValuesTrustPerformance a, ValuesTrustPerformance b) {
		return new ValuesTrustPerformance(a._0To20 + b._0To20, a._20To40+b._20To40, a._40To60 + b._40To60, a._60To80 + b._60To80, a._80To100 + b._80To100);
	}

	@Override
	public BinaryOperator<ValuesTrustPerformance> getAccumulator() {
		return ValuesTrustPerformance::accumulater;
	}


	@Override
	public ValuesTrustPerformance accumulate(ValuesTrustPerformance other) {
		if (other instanceof ValuesTrustPerformance) {
			ValuesTrustPerformance oCast = (ValuesTrustPerformance) other;

			return accumulater(this, oCast);
		} else {
			throw new IllegalStateException("This is not the same Accumateable Class");
		}
	}


	@Override
	public Class<ValuesTrustPerformance> getAcumClass() {
		return ValuesTrustPerformance.class;
	}
	
 

	public static ValuesTrustPerformance empty() {
		return new ValuesTrustPerformance(0,0,0,0,0);
	}
	
	
	
}
