package expansion.set.evaluation.crossvalidation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;

public class EntitySetsCreator {
	private final AnnotatedTablesRepository tableRepository;
	@Inject
	public EntitySetsCreator(AnnotatedTablesRepository tableRepository) {
		this.tableRepository = tableRepository;
	}

	private final Comparator<InstanceId> byTimestamp = Comparator.comparing(InstanceId::getInstanceId);


	public Map<Integer, EntitySets> split(Map<ClassId, Set<InstanceId>> instanceIdSet, int setSize) {

		Comparator<InstanceId> byTimestamp = Comparator.comparing(InstanceId::getInstanceId);

		Supplier<TreeSet<InstanceId>> supplier = () -> new TreeSet<InstanceId>(byTimestamp);

//		Map<ClassId, TreeSet<InstanceId>> classes = groupedValues.entrySet().stream().collect(Collectors.groupingBy((e) -> {
//			return e.getKey().getTripleId().getClassId();
//		} , Collectors.mapping((e) -> {
//			return e.getKey().getTripleId().getInstanceId();
//		} , Collectors.toCollection(supplier))));
//		
		Map<ClassId, TreeSet<InstanceId>> classes = instanceIdSet.entrySet().stream().collect(Collectors.toMap((e) -> {
			return e.getKey();
		} , (e) -> {
			TreeSet<InstanceId> ts = new TreeSet<>(byTimestamp);
			ts.addAll(e.getValue());
			return  ts;
		}));

 
		Map<Integer, Set<InstanceId>> learning = new HashMap<>();
		Map<Integer, Set<InstanceId>> testing = new HashMap<>();

		for(Entry<ClassId, TreeSet<InstanceId>> entry : classes.entrySet()) {
			Map<Integer, EntitySets> splitted = splitInner(entry.getValue(), setSize);
			for (Entry<Integer, EntitySets> splittedEntry : splitted.entrySet()) {
				CHMHelper.getFromMap(learning, splittedEntry.getKey(), () -> new HashSet<>()).addAll(splittedEntry.getValue().getLearningSet());
				CHMHelper.getFromMap(testing, splittedEntry.getKey(), () -> new HashSet<>()).addAll(splittedEntry.getValue().getTestingSet());
			}

		}

		Map<Integer, EntitySets> ret = new HashMap<>();
		for (Integer key : learning.keySet()) {
			ret.put(key, new EntitySets(learning.get(key), testing.get(key)));
		}
		
		return ret;

	}

	public Map<Integer, EntitySets> split(Collection<InstanceId> entities, int setSize) {

		TreeSet<InstanceId> entityIds = new TreeSet<>(byTimestamp);
		entityIds.addAll(entities);

		return splitInner(entityIds, setSize);
	}

	private Map<Integer, EntitySets> splitInner(TreeSet<InstanceId> entityIds, int setSize) {

		Map<Integer, List<InstanceId>> learningSet = new HashMap<>();
		Map<Integer, List<InstanceId>> testingSet = new HashMap<>();

		for (int i = 0; i < setSize; i++) {
			learningSet.put(i, new ArrayList<>());
			testingSet.put(i, new ArrayList<>());
		}


		int i = 0;
		for (InstanceId instanceId : entityIds) {
			if(this.tableRepository.hasInstance(instanceId)) {
				for (int set = 0; set < setSize; set++) {				
					if (set == i) {
						testingSet.get(set).add(instanceId);
					} else {
						learningSet.get(set).add(instanceId);
					}
				}
	
				i++;
				if (i == setSize) {
					i = 0;
				}
			}
		}

		Map<Integer, EntitySets> ret = new HashMap<>();

		for (int s = 0; s < setSize; s++) {
			ret.put(s, new EntitySets(learningSet.get(s), testingSet.get(s)));
		}

		return ret;
	}

}
