package expansion.set.evaluation.entities;

import java.util.function.BinaryOperator;

import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.Accumulateable;

public class EntitiesPerformanceSingle implements Accumulateable<EntitiesPerformanceSingle>{

	
	private final double precisionSum ;
	private final double recallSum;
	private final double fmeasureSum;
	private final int count;
	private final int foundClusters;
	private final int clustersWithCollision;
	private final int totalCollisions;
	
	
	
	public EntitiesPerformanceSingle() {
		this.precisionSum = 0.0;
		this.recallSum = 0.0;
		this.fmeasureSum = 0.0;
		this.count = 0;
		this.foundClusters = 0;
		this.clustersWithCollision = 0;
		this.totalCollisions = 0;
	}


	private EntitiesPerformanceSingle(double precisionSum, double recallSum, double fmeasureSum, int count, int foundClusters, int clustersWithCollision,
			int totalCollisions) {
		super();
		this.precisionSum = precisionSum;
		this.recallSum = recallSum;
		this.fmeasureSum = fmeasureSum;
		this.count = count;
		this.foundClusters = foundClusters;
		this.clustersWithCollision = clustersWithCollision;
		this.totalCollisions = totalCollisions;
	}

	
	public EntitiesPerformanceSingle(double precisionSum, double recallSum,    int foundClusters, int clustersWithCollision,
			int totalCollisions) {
		super();
		this.precisionSum = precisionSum;
		this.recallSum = recallSum;
		this.fmeasureSum = (2* precisionSum * recallSum) / (precisionSum + recallSum);
		this.count = 1;
		this.foundClusters = foundClusters;
		this.clustersWithCollision = clustersWithCollision;
		this.totalCollisions = totalCollisions;
	}
	
	
	

	public double getPrecision() {
		if (count == 0)
			return 0.0;
		return precisionSum / count;
	}


	public double getRecall() {
		if (count == 0)
			return 0.0;
		return recallSum / count;
	}


	public double getFmeasure() {
		if (count == 0)
			return 0.0;
		return fmeasureSum / count;
	}


	public int getFoundClusters() {
		return foundClusters;
	}


	public int getClustersWithCollision() {
		return clustersWithCollision;
	}


	public int getTotalCollisions() {
		return totalCollisions;
	}


	public static EntitiesPerformanceSingle accumulator(EntitiesPerformanceSingle a, EntitiesPerformanceSingle b) {
		return new EntitiesPerformanceSingle(a.precisionSum + b.precisionSum, a.recallSum + b.recallSum, a.fmeasureSum + b.fmeasureSum, a.count + b.count, a.foundClusters + b.foundClusters, a.clustersWithCollision + b.clustersWithCollision, a.totalCollisions + b.totalCollisions);
	}
	
	
	@Override
	public BinaryOperator<EntitiesPerformanceSingle> getAccumulator() {
		return EntitiesPerformanceSingle::accumulator;
	}

	@Override
	public EntitiesPerformanceSingle accumulate(EntitiesPerformanceSingle other) {
		return EntitiesPerformanceSingle.accumulator(this, other);
	}

	@Override
	public Class<EntitiesPerformanceSingle> getAcumClass() {
		return EntitiesPerformanceSingle.class;
	}

}
