package expansion.set.evaluation.fusion;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance.FusionPerformanceCreator;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumInteger;
import expansion.set.creation.Entity;
import expansion.set.evaluation.matching.ScoredEntity;

public class FusionSetCreator {

	private final OntologyRepository ontologyRepository;

	@Inject
	public FusionSetCreator(OntologyRepository ontologyRepository) {

		this.ontologyRepository = ontologyRepository;
	}

	
	public Map<QualifiedTripleId, FusedValue> fuse2(Map<InstanceId, Entity> mapping , SplitGroupedValues originalTestingSet) {
		
		ConcurrentHashMap<QualifiedTripleId, FusedValue> fusedValues = new ConcurrentHashMap<>();
 

		PersistentParallel6.runAndFinish(mapping.entrySet(), (entry) -> {
			
			Entity entity = entry.getValue();
			
			for(Entry<QualifiedPropertyId, FusedValue> values : entity.getValues().entrySet()) {
				QualifiedTripleId qtid = new QualifiedTripleId(new TripleId(entity.getClassId(), values.getKey().getPropertyId(), entry.getKey()),
						values.getKey().getQualifiers());
				
				if (originalTestingSet.containsKey(qtid)) {
					fusedValues.put(qtid, values.getValue());
				}
 
			}
			
		}, this.getClass());
		
 
		return fusedValues;
	}
 
	public Map<QualifiedTripleId, List<FusedValue>> fuse3(Map<InstanceId, List<Entity>> mapping , SplitGroupedValues originalTestingSet) {
		
		ConcurrentHashMap<QualifiedTripleId,  List<FusedValue>> fusedValues = new ConcurrentHashMap<>();
 

		PersistentParallel6.runAndFinish(mapping.entrySet(), (entry) -> {
			
			 
			
			for(Entity entity : entry.getValue()) {
				for(Entry<QualifiedPropertyId, FusedValue> values : entity.getValues().entrySet()) {
					QualifiedTripleId qtid = new QualifiedTripleId(new TripleId(entity.getClassId(), values.getKey().getPropertyId(), entry.getKey()),
							values.getKey().getQualifiers());
					
					if (originalTestingSet.containsKey(qtid)) {
						CHMHelper.getFromCHM(fusedValues, qtid, () -> Collections.synchronizedList(new ArrayList<>())).add(values.getValue());
					}
	 
				}
			}
			 
			
		
			
		}, this.getClass());
		
 
		return fusedValues;
	}
	
	
	public Map<QualifiedTripleId, List<FusedValue>> fuse4(Map<InstanceId, List<Entity>> mapping ) {
		
		ConcurrentHashMap<QualifiedTripleId,  List<FusedValue>> fusedValues = new ConcurrentHashMap<>();
 

		PersistentParallel6.runAndFinish(mapping.entrySet(), (entry) -> {
			
			 
			
			for(Entity entity : entry.getValue()) {
				for(Entry<QualifiedPropertyId, FusedValue> values : entity.getValues().entrySet()) {
					QualifiedTripleId qtid = new QualifiedTripleId(new TripleId(entity.getClassId(), values.getKey().getPropertyId(), entry.getKey()),
							values.getKey().getQualifiers());
					
//					if (originalTestingSet.containsKey(qtid)) {
						CHMHelper.getFromCHM(fusedValues, qtid, () -> Collections.synchronizedList(new ArrayList<>())).add(values.getValue());
//					}
	 
				}
			}
			 
			
		
			
		}, this.getClass());
		
 
		return fusedValues;
	}
 
	public FusionSet fuse(Map<InstanceId, List<ScoredEntity>> mappedEntities, SplitGroupedValues originalTestingSet) {

		ConcurrentHashMap<QualifiedTripleId, FusedValue> fusedValues = new ConcurrentHashMap<>();

		FusionPerformanceCreator<AccumInteger> accumulator = FusionPerformance.accumulator(new AccumInteger(0));

		PersistentParallel6.runAndFinish(mappedEntities.entrySet(), (mappings) -> {

			Integer maxIndex = null;
			double maxScore = 0.0;

			for (ListEntry<ScoredEntity> entry : ListIterable.iterable(mappings.getValue())) {
				if (maxIndex == null || entry.getValue().getScore() > maxScore) {
					maxIndex = entry.getIndex();
					maxScore = entry.getValue().getScore();
				}
			}


			ScoredEntity max = mappings.getValue().get(maxIndex);
			
			for(Entry<QualifiedPropertyId, FusedValue> values : max.getEntity().getValues().entrySet()) {
				QualifiedTripleId qtid = new QualifiedTripleId(new TripleId(max.getEntity().getClassId(), values.getKey().getPropertyId(), mappings.getKey()),
						values.getKey().getQualifiers());
				if (originalTestingSet.containsKey(qtid)) {
					fusedValues.put(qtid, values.getValue());
				}
 else {
					int adsd = 122;
				}
			}

			for (ListEntry<ScoredEntity> entry : ListIterable.iterable(mappings.getValue())) {
				if (entry.getIndex() != maxIndex) {
					for (Entry<QualifiedPropertyId, FusedValue> values : entry.getValue().getEntity().getValues().entrySet()) {
						QualifiedTripleId qtid = new QualifiedTripleId(
								new TripleId(max.getEntity().getClassId(), values.getKey().getPropertyId(), mappings.getKey()),
								values.getKey().getQualifiers());
						if (originalTestingSet.containsKey(qtid)) {
							accumulator.addWithYear(entry.getValue().getEntity().getClassId(), values.getKey(), ontologyRepository.isTimeDependent(qtid.getQualifiedPropertyId().getPropertyId()), ontologyRepository.getDataType(qtid.getQualifiedPropertyId().getPropertyId()), new AccumInteger(1));
						}
					}
				}
			}

		} , this.getClass());

		return new FusionSet(fusedValues, accumulator.get((e) -> {
			return e.get();
		}));
	}

	public static class FusionSet {
		private final Map<QualifiedTripleId, FusedValue> fusedValues;
		private final FusionPerformance<Integer> wrongCount;

		private FusionSet(Map<QualifiedTripleId, FusedValue> fusedValues, FusionPerformance<Integer> wrongCount) {
			super();
			this.fusedValues = fusedValues;
			this.wrongCount = wrongCount;
		}

		public Map<QualifiedTripleId, FusedValue> getFusedValues() {
			return fusedValues;
		}

		public FusionPerformance<Integer> getWrongCount() {
			return wrongCount;
		}

	}


}
