package expansion.set.evaluation.fusion;

import static j2html.TagCreator.body;
import static j2html.TagCreator.div;
import static j2html.TagCreator.h1;
import static j2html.TagCreator.h3;
import static j2html.TagCreator.head;
import static j2html.TagCreator.hr;
import static j2html.TagCreator.html;
import static j2html.TagCreator.li;
import static j2html.TagCreator.link;
import static j2html.TagCreator.p;
import static j2html.TagCreator.span;
import static j2html.TagCreator.strong;
import static j2html.TagCreator.table;
import static j2html.TagCreator.td;
import static j2html.TagCreator.th;
import static j2html.TagCreator.tr;
import static j2html.TagCreator.ul;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import javax.inject.Inject;

import com.google.common.collect.ImmutableTable;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.PerformanceMacroPR;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;
import expansion.set.evaluation.goldstandard.ClusterEvaluator;
import expansion.set.evaluation.goldstandard.GoldStandardClusters;
import j2html.tags.ContainerTag;

public class FusionSetEvaluator {

	@Inject
	protected ClusterEvaluator clusterEvaluator;
	
	@Inject
	protected ValueEvaluator valueEvaluator;


	@Inject
	protected AnnotatedTablesRepository tableRepository;

	public FusionSetEvaluator() {
		 
	}
	
	public Map<PropertyId, FusionSetPerformance> evaluate(Map<QualifiedTripleId, List<FusedValue>> fusionSet,
			Map<QualifiedTripleId, QualifiedValueList> correctMap, Set<InstanceId> newEntities, Map<QualifiedTripleId, QualifiedValueList> fullCorrectMap) {
		  Map<QualifiedTripleId, QualifiedValueList> correctMapExisting = new HashMap<>();
		   Map<QualifiedTripleId, QualifiedValueList> correctMapNew = new HashMap<>();
	 
		   correctMapExisting = correctMap.entrySet().stream().filter((e) -> !newEntities.contains(e.getKey().getTripleId().getInstanceId())).collect(Collectors.toMap(Entry::getKey, Entry::getValue));
		   correctMapNew = correctMap.entrySet().stream().filter((e) -> newEntities.contains(e.getKey().getTripleId().getInstanceId())).collect(Collectors.toMap(Entry::getKey, Entry::getValue));
 

		   Map<PropertyId, Map<QualifiedTripleId, QualifiedValueList>> correctMapMap = new HashMap<>();
		   for(Entry<QualifiedTripleId, QualifiedValueList> entry : correctMap.entrySet()) {
			   CHMHelper.getFromMap(correctMapMap, entry.getKey().getTripleId().getPropertyId(), () -> new HashMap<>()).put(entry.getKey(), entry.getValue());
		   }
		   
		   Map<PropertyId, Map<QualifiedTripleId, QualifiedValueList>> correctMapExistingMap = new HashMap<>();
		   for(Entry<QualifiedTripleId, QualifiedValueList> entry : correctMapExisting.entrySet()) {
			   CHMHelper.getFromMap(correctMapExistingMap, entry.getKey().getTripleId().getPropertyId(), () -> new HashMap<>()).put(entry.getKey(), entry.getValue());
		   }
		   

		   Map<PropertyId, Map<QualifiedTripleId, QualifiedValueList>> correctMapNewMap = new HashMap<>();
		   for(Entry<QualifiedTripleId, QualifiedValueList> entry : correctMapNew.entrySet()) {
			   CHMHelper.getFromMap(correctMapNewMap, entry.getKey().getTripleId().getPropertyId(), () -> new HashMap<>()).put(entry.getKey(), entry.getValue());
		   }
			
		   
		   Map<PropertyId, Map<QualifiedTripleId, List<FusedValue>>> fusionSetMap = new HashMap<>();

		   
		   for(Entry<QualifiedTripleId, List<FusedValue>> entry : fusionSet.entrySet()) {
			   CHMHelper.getFromMap(fusionSetMap, entry.getKey().getTripleId().getPropertyId(), () -> new HashMap<>()).put(entry.getKey(), entry.getValue());

		   }
		   
		   
		   Set<PropertyId> propertyIdSet = new HashSet<>();
		   propertyIdSet.addAll(correctMapMap.keySet());
		   propertyIdSet.addAll(correctMapExistingMap.keySet());
		   propertyIdSet.addAll(correctMapNewMap.keySet());
		   propertyIdSet.addAll(fusionSetMap.keySet());
		   
			Map<PropertyId, FusionSetPerformance> performance = new HashMap<>();
			
			
			for(PropertyId propertyId : propertyIdSet) {
				Map<QualifiedTripleId, List<FusedValue>> fusionSetInner = CHMHelper.getFromMap(fusionSetMap, propertyId, () -> new HashMap<>());
				
			PerformanceMacroPR performanceAll = clusterEvaluator.evaluteFusionSet(CHMHelper.getFromMap(correctMapMap, propertyId, () -> new HashMap<>()),
					fusionSetInner, fullCorrectMap);
				
				Map<QualifiedTripleId, List<FusedValue>> fusionSetInnerExisting = fusionSetInner.entrySet().stream().filter((e) -> !newEntities.contains(e.getKey().getTripleId().getInstanceId())).collect(Collectors.toMap(Entry::getKey, Entry::getValue));
				
			PerformanceMacroPR performanceExisting = clusterEvaluator
					.evaluteFusionSet(CHMHelper.getFromMap(correctMapExistingMap, propertyId, () -> new HashMap<>()), fusionSetInnerExisting, fullCorrectMap);
				
				Map<QualifiedTripleId, List<FusedValue>> fusionSetInnerNew = fusionSetInner.entrySet().stream().filter((e) ->  newEntities.contains(e.getKey().getTripleId().getInstanceId())).collect(Collectors.toMap(Entry::getKey, Entry::getValue));

				
			PerformanceMacroPR performanceNew = clusterEvaluator.evaluteFusionSet(CHMHelper.getFromMap(correctMapNewMap, propertyId, () -> new HashMap<>()),
					fusionSetInnerNew, fullCorrectMap);
				
				 performance.put(propertyId, new FusionSetPerformance(performanceAll, performanceExisting, performanceNew));
			}
			
			return performance;
		
	}

	public String printPerformance(Map<InstanceId, Set<TableRowId>> clusters, GoldStandardClusters gs, Set<InstanceId> newEntities,
			Map<QualifiedTripleId, List<FusedValue>> fusionSet, Map<QualifiedTripleId, QualifiedValueList> correctMap,
			Map<QualifiedTripleId, QualifiedValueList> fullCorrectMap) {

		Set<InstanceId> instanceIds = new HashSet<>();

		Map<QualifiedTripleId, List<AbstractObject>> fullCorrectMapTransformed = fullCorrectMap.entrySet().stream()
				.collect(Collectors.toMap(Entry::getKey, (e) -> {
					List<AbstractObject> ret = new ArrayList<>();
					for (QualifiedValueHolder claim : e.getValue().getClaims()) {
						ret.add(claim.getValue());
					}
					return ret;
				}));
		Map<InstanceId, Map<QualifiedTripleId, List<FusedValue>>> correctCount = new HashMap<>();
		Map<InstanceId, Map<QualifiedTripleId, List<FusedValue>>> wrongCount = new HashMap<>();
		Map<InstanceId, Map<QualifiedTripleId, QualifiedValueList>> missingCount = new HashMap<>();
		Map<InstanceId, Map<QualifiedTripleId, List<FusedValue>>> ignoredCount = new HashMap<>();
		Map<InstanceId, Map<QualifiedTripleId, List<FusedValue>>> ignoredCorrectCount = new HashMap<>();

		Map<InstanceId, Map<QualifiedTripleId, QualifiedValueList>> allCorrectCount = new HashMap<>();

		for (Entry<QualifiedTripleId, List<FusedValue>> entry : fusionSet.entrySet()) {
			InstanceId instanceId = entry.getKey().getTripleId().getInstanceId();
			instanceIds.add(instanceId);

			if (correctMap.containsKey(entry.getKey())) {
				boolean isCorrect = isCorrect(entry.getKey(), correctMap.get(entry.getKey()), entry.getValue());
				if (isCorrect) {
					CHMHelper.getFromMap(correctCount, instanceId, () -> new HashMap<>()).put(entry.getKey(), entry.getValue());
				} else {
					CHMHelper.getFromMap(wrongCount, instanceId, () -> new HashMap<>()).put(entry.getKey(), entry.getValue());
				}
			} else if (fullCorrectMap.containsKey(entry.getKey())) {
				boolean isCorrect = isCorrect(entry.getKey(), fullCorrectMap.get(entry.getKey()), entry.getValue());
				if (isCorrect) {
					CHMHelper.getFromMap(ignoredCorrectCount, instanceId, () -> new HashMap<>()).put(entry.getKey(), entry.getValue());
				} else {
					CHMHelper.getFromMap(wrongCount, instanceId, () -> new HashMap<>()).put(entry.getKey(), entry.getValue());
				}
			} else {
				CHMHelper.getFromMap(ignoredCount, instanceId, () -> new HashMap<>()).put(entry.getKey(), entry.getValue());
			}

		}

		for (Entry<QualifiedTripleId, QualifiedValueList> entry : correctMap.entrySet()) {
			InstanceId instanceId = entry.getKey().getTripleId().getInstanceId();
			instanceIds.add(instanceId);

			CHMHelper.getFromMap(allCorrectCount, instanceId, () -> new HashMap<>()).put(entry.getKey(), entry.getValue());

			if (!fusionSet.containsKey(entry.getKey())) {
				CHMHelper.getFromMap(missingCount, instanceId, () -> new HashMap<>()).put(entry.getKey(), entry.getValue());

			}
		}

		List<InstanceId> sortedWrong = instanceIds.stream().sorted((e1, e2) -> {
			int e1Score = CHMHelper.getFromMap(wrongCount, e1, () -> new HashMap<>()).size();
			int e2Score = CHMHelper.getFromMap(wrongCount, e2, () -> new HashMap<>()).size();
			int compare = Integer.compare(e2Score, e1Score);
			if (compare != 0) {
				return compare;
			}
			compare = Integer.compare(e1.hashCode(), e2.hashCode());
			if (compare != 0) {
				return compare;
			}
			compare = e1.getInstanceId().compareTo(e2.getInstanceId());
			return compare;

		}).collect(Collectors.toList());

		List<InstanceId> sortedMissing = instanceIds.stream().sorted((e1, e2) -> {
			int e1Score = CHMHelper.getFromMap(missingCount, e1, () -> new HashMap<>()).size();
			int e2Score = CHMHelper.getFromMap(missingCount, e2, () -> new HashMap<>()).size();
			int compare = Integer.compare(e2Score, e1Score);
			if (compare != 0) {
				return compare;
			}
			compare = Integer.compare(e1.hashCode(), e2.hashCode());
			if (compare != 0) {
				return compare;
			}
			compare = e1.getInstanceId().compareTo(e2.getInstanceId());
			return compare;

		}).collect(Collectors.toList());

		List<InstanceId> sortedNotInGS = instanceIds.stream().sorted((e1, e2) -> {
			int e1Score = CHMHelper.getFromMap(ignoredCount, e1, () -> new HashMap<>()).size();
			int e2Score = CHMHelper.getFromMap(ignoredCount, e2, () -> new HashMap<>()).size();
			int compare = Integer.compare(e2Score, e1Score);
			if (compare != 0) {
				return compare;
			}
			compare = Integer.compare(e1.hashCode(), e2.hashCode());
			if (compare != 0) {
				return compare;
			}
			compare = e1.getInstanceId().compareTo(e2.getInstanceId());
			return compare;

		}).collect(Collectors.toList());

		ContainerTag contentDiv = div().withClass("row");

		Set<InstanceId> sortedInstanceids = new HashSet<>();
		int i = 0;
		for (InstanceId instanceId : sortedMissing) {
			sortedInstanceids.add(instanceId);
			i++;
			if (i == 10)
				break;
		}

		i = 0;
		for (InstanceId instanceId : sortedWrong) {
			sortedInstanceids.add(instanceId);
			i++;
			if (i == 10)
				break;
		}

		i = 0;
		for (InstanceId instanceId : sortedNotInGS) {
			sortedInstanceids.add(instanceId);
			i++;
			if (i == 10)
				break;
		}

		i = 0;
		for (InstanceId instanceId : sortedInstanceids) {

			Set<TableRowId> gsCluster = gs.getCluster(instanceId);

			Set<TableRowId> cluster;
			if (clusters.containsKey(instanceId)) {
				cluster = clusters.get(instanceId);
			} else {
				cluster = new HashSet<>();
			}

			Set<TableRowId> correctRows = cluster.stream().filter((e) -> gsCluster.contains(e)).collect(Collectors.toSet());
			Set<TableRowId> incorrectRows = cluster.stream().filter((e) -> !gsCluster.contains(e)).collect(Collectors.toSet());
			Set<TableRowId> missingRows = gsCluster.stream().filter((e) -> !cluster.contains(e)).collect(Collectors.toSet());

			ContainerTag correctRowsTag = getRowsContainer(correctRows);
			ContainerTag incorrectRowsTag = getRowsContainer(incorrectRows);
			ContainerTag missingRowsTag = getRowsContainer(missingRows);

			Map<QualifiedTripleId, List<AbstractObject>> maxCorrect = CHMHelper.getFromMap(allCorrectCount, instanceId, () -> new HashMap<>()).entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				List<AbstractObject> ret = new ArrayList<>();
				for(QualifiedValueHolder claim : e.getValue().getClaims()) {
					ret.add(claim.getValue());
				}
				return ret; 
					}));

			Map<QualifiedTripleId, List<FusedValue>> correct = CHMHelper.getFromMap(correctCount, instanceId, () -> new HashMap<>());
			Map<QualifiedTripleId, List<FusedValue>> wrong = CHMHelper.getFromMap(wrongCount, instanceId, () -> new HashMap<>());
			Map<QualifiedTripleId, List<AbstractObject>> missing = CHMHelper.getFromMap(missingCount, instanceId, () -> new HashMap<>()).entrySet().stream()
					.collect(Collectors.toMap(Entry::getKey, (e) -> {
						List<AbstractObject> ret = new ArrayList<>();
						for (QualifiedValueHolder claim : e.getValue().getClaims()) {
							ret.add(claim.getValue());
						}
						return ret;
					}));

			Map<QualifiedTripleId, List<FusedValue>> ignored = CHMHelper.getFromMap(ignoredCount, instanceId, () -> new HashMap<>());


			Map<QualifiedTripleId, List<FusedValue>> ignoredCorrect = CHMHelper.getFromMap(ignoredCorrectCount, instanceId, () -> new HashMap<>());

			ContainerTag correctfactsUL = ul();
			
			for(Entry<QualifiedTripleId, List<FusedValue>> factEntry: correct.entrySet()) {
				ContainerTag innerUl = ul();
				for (FusedValue fv : factEntry.getValue()) {
					innerUl.with(li(fv.getSingle().toString()));
				}
				correctfactsUL.with(
						li(factEntry.getKey().getTripleId().getPropertyId().getReadableString() + "(" + maxCorrect.get(factEntry.getKey()).toString() + ")")
								.with(innerUl));
			}
			
			ContainerTag wrongfactsUL = ul();

			for (Entry<QualifiedTripleId, List<FusedValue>> factEntry : wrong.entrySet()) {
				ContainerTag innerUl = ul();
				for (FusedValue fv : factEntry.getValue()) {
					innerUl.with(li(fv.getSingle().toString()));
				}

				if (maxCorrect.containsKey(factEntry.getKey())) {
					wrongfactsUL.with(
							li(factEntry.getKey().getTripleId().getPropertyId().getReadableString() + "(" + maxCorrect.get(factEntry.getKey()).toString() + ")")
									.with(innerUl));
				} else {
					wrongfactsUL.with(li(factEntry.getKey().getTripleId().getPropertyId().getReadableString() + "("
							+ fullCorrectMapTransformed.get(factEntry.getKey()).toString() + ")").with(innerUl));
				}

			}

			ContainerTag missingfactsUL = ul();

			for (Entry<QualifiedTripleId, List<AbstractObject>> factEntry : missing.entrySet()) {
				ContainerTag innerUl = ul();
				for (AbstractObject fv : factEntry.getValue()) {
					innerUl.with(li(fv.toString()));
				}
				missingfactsUL.with(li(factEntry.getKey().getTripleId().getPropertyId().getReadableString()).with(innerUl));
			}

			ContainerTag notInGSfactsUL = ul();

			for (Entry<QualifiedTripleId, List<FusedValue>> factEntry : ignored.entrySet()) {
				ContainerTag innerUl = ul();
				for (FusedValue fv : factEntry.getValue()) {
					innerUl.with(li(fv.getSingle().toString()));
				}
				notInGSfactsUL.with(li(factEntry.getKey().getTripleId().getPropertyId().getReadableString()).with(innerUl));
			}

			ContainerTag ignoredCorrectUL = ul();

			for (Entry<QualifiedTripleId, List<FusedValue>> factEntry : ignoredCorrect.entrySet()) {
				ContainerTag innerUl = ul();
				for (FusedValue fv : factEntry.getValue()) {
					innerUl.with(li(fv.getSingle().toString()));
				}
				ignoredCorrectUL.with(li(factEntry.getKey().getTripleId().getPropertyId().getReadableString()).with(innerUl));
			}

			contentDiv.with(div().withClass("col-xs-12").with(div().withClass("card border-primary").with(

					div().withClass("card-body").with( //
							h1(instanceId.readableString()) //
					), //

					div().withClass("card-body").with(h3("Facts") //
					), //

					div().withClass("card-body").with(strong("Correct"), //
							correctfactsUL, //
							strong("Wrong"), //
							wrongfactsUL, //
							strong("Missing"), //
							missingfactsUL, //

							strong("Not in Gold Standard"), //
							notInGSfactsUL, //

							strong("Ignored Correct"), //
							ignoredCorrectUL //
					),

					div().withClass("card-body").with(h3("Rows") //
					),

					div().withClass("card-body").with(strong("CorrectRows"), //
							correctRowsTag //
					),

					div().withClass("card-body").with(strong("IncorrectRows"), //
							incorrectRowsTag //
					), //

					div().withClass("card-body").with(strong("MissingRows"), //
							missingRowsTag //
					) //

			)));


		}

		ContainerTag html = html(head(link().attr("rel", "stylesheet").attr("href", "https://maxcdn.bootstrapcdn.com/bootstrap/4.0.0/css/bootstrap.min.css")),
				body().withStyle("padding: 50px;").with(div().withClass("container-fluid").with(div().withClass("row").with(contentDiv))));
		return html.render();

	}

	public boolean isCorrect(QualifiedTripleId tripleId, QualifiedValueList qvl, List<FusedValue> fact) {

		boolean foundWrong = false;
		for (FusedValue factEntry : fact) {
			Optional<Boolean> correctResult = valueEvaluator.correctIntern(tripleId, qvl, factEntry);
			if (correctResult.isPresent() && !correctResult.get()) {
				foundWrong = true;
				break;
			}
		}
		if (foundWrong) {
			return false;
		} else {
			return true;
		}

	}

	private ContainerTag getRowsContainer(Set<TableRowId> rows) {

		ContainerTag div = div();

		for (TableRowId rowId : rows) {
			div.with(p(strong("Table: "), span(rowId.getTableId().getTableName())), p(strong("RowIdx: "), span(String.valueOf(rowId.getRowIdx().getRowIdx()))));



			ContainerTag mappingRowTag = tr();
			ContainerTag headerRowTag = tr();
			ContainerTag rowTag = tr();


			AnnotatedTable table = this.tableRepository.getTable(rowId.getTableId());
			int colSize = table.getColumns().size();

			AnnotatedTableMetaData md = tableRepository.getMetaData(rowId.getTableId());
			Map<Integer, PropertyId> properties = md.getProperties();
			ImmutableTable<RowIdx, ColumnIdx, String> stringTable = table.getStringTable();
			RowIdx headerRowIdx = new RowIdx(0);
			ColumnIdx keyColumnIdx = table.getKeyColumnIndex();
			RowIdx rowIdx = rowId.getRowIdx();

			for (int i = 0; i < colSize; i++) {
				if (properties.containsKey(i)) {
					mappingRowTag.with(th(properties.get(i).getReadableString()));
				} else {
					mappingRowTag.with(th(""));
				}
				ColumnIdx columnIdx = new ColumnIdx(i);

				headerRowTag.with(th(stringTable.get(headerRowIdx, columnIdx)));

				if (columnIdx.equals(keyColumnIdx)) {
					rowTag.with(th(stringTable.get(rowIdx, columnIdx)));
				} else {
					rowTag.with(td(stringTable.get(rowIdx, columnIdx)));
				}

			}

			div.with(table(mappingRowTag, headerRowTag, rowTag).withClass("table"));
			div.with(hr());
		}
		return div;
	}

}
