package expansion.set.evaluation.goldstandard;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.PerformanceMacroPR;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.DoubleIteraterable;
import de.mannheim.uni.expansion.utils.DoubleIteraterable.DoubleEntry;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumInteger;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentAccumulator;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.creation.Entity;
 
public class ClusterEvaluator {
	
	private final ValueEvaluator valueEvaluator;
	@Inject
	public ClusterEvaluator(ValueEvaluator valueEvaluator) { 
		this.valueEvaluator = valueEvaluator;
	}
	
	private static List<TableRowIdPair> createList() {
		return new ArrayList<>();
	}
	
//	public <E> GoldStandardPerformance evaluate(Map<E, Set<TableRowId>>  goldStandardClusters,   List<List<Entity>> entities) {
//		List<Set<TableRowId>> clusters =  entities.stream().map((e) -> { Set<TableRowId>ret = new HashSet<>(e.stream().getRowCluster().getRows())).collect(Collectors.toList());
//		
//		return evaluateRaw(goldStandardClusters, clusters);
//	}
	public GoldStandardPerformance evaluate(GoldStandardClusters goldStandardClusters,   List<Entity> entities) {
		List<Set<TableRowId>> clusters =  entities.stream().map((e) -> new HashSet<>(e.getRowCluster().getRows())).collect(Collectors.toList());
		
		return evaluateRaw(goldStandardClusters.getTestingSetMappingClusters(), clusters);
	}
	public <E> GoldStandardPerformance evaluateRaw(Map<E, Set<TableRowId>>  goldStandardClusters,  List<Set<TableRowId>> clusters) {
		 
		
		
		
		Map<TableRowId, Integer> resultMap = new ConcurrentHashMap<>();
		Map<TableRowId, E> gsMap = new ConcurrentHashMap<>();
		Map<Integer, Set<TableRowId>> entitiesMap = new ConcurrentHashMap<>();
	 
		
		PersistentParallel6.runAndFinish(ListIterable.iterable(clusters), (e) -> {
			
			for(TableRowId rowId : e.getValue()) {
				resultMap.put(rowId, e.getIndex());
			}
			
			entitiesMap.put(e.getIndex(), e.getValue());
			
		}, this.getClass());
		
		
		PersistentParallel6.runAndFinish(goldStandardClusters.entrySet(), (e) -> {
			for(TableRowId rowId : e.getValue()) {
				gsMap.put(rowId, e.getKey());
			}
			
		}, this.getClass());
		
		
//		  final ThreadLocalList<List<TableRowIdPair>>  list = new ThreadLocalList<>();
		  
		ConcurrentLinkedQueue<TableRowIdPair> queue = new ConcurrentLinkedQueue<>();
		
		PersistentParallel6.runAndFinish(entitiesMap.entrySet(), (e) -> {
			PersistentParallel6.runAndFinishWithDefaultBatchSize(DoubleIteraterable.iterable(e.getValue()), (d) -> {
				queue.add(new TableRowIdPair(d.getFirst(), d.getSecond()));
			}, this.getClass(), e.getValue().size() < 100);
		}, this.getClass());
		
		Set<TableRowIdPair> set = new HashSet<>(queue);
		
		PersistentParallel6.runAndFinish(goldStandardClusters.entrySet(), (e) -> {
			PersistentParallel6.runAndFinishWithDefaultBatchSize(DoubleIteraterable.iterable(e.getValue()), (d) -> {
				TableRowIdPair pair = new TableRowIdPair(d.getFirst(), d.getSecond());
				if(!set.contains(pair)) {
					queue.add(pair);
				}
			}, this.getClass(), e.getValue().size() < 100);
		}, this.getClass());
		
//		PersistentParallel6.runAndFinish(goldStandardClusters.getCandidates(), (rowId) -> {
//			List<TableRowIdPair> currentList = list.getOrCreate(ClusterEvaluator::createList);
//			for(TableRowId second : entitiesMap.get(resultMap.get(rowId)).getRowCluster().getRows() ) {
//				if(!second.equals(rowId)) {
//					currentList.add(new TableRowIdPair(rowId, second));
//				}
//			}
//			
//			for(TableRowId second : goldStandardClusters.getTestingSetMappingClusters().get(gsMap.get(rowId))  ) {
//				if(!second.equals(rowId)) {
//					currentList.add(new TableRowIdPair(rowId, second));
//				}
//			}
//		}, this.getClass());
//		
//		for(List<TableRowIdPair> cList : list.getCurrentList()) {
//			set.addAll(cList);
//		}
//		 
		ConcurrentAccumulator<AccumInteger> truePositive = new ConcurrentAccumulator<>(AccumInteger.empty());
		ConcurrentAccumulator<AccumInteger> falsePositive = new ConcurrentAccumulator<>(AccumInteger.empty());
//		ConcurrentAccumulator<AccumInteger> trueNegative = new ConcurrentAccumulator<>(AccumInteger.empty());
		ConcurrentAccumulator<AccumInteger> falseNegative = new ConcurrentAccumulator<>(AccumInteger.empty());
		 
		
		PersistentParallel6.runAndFinish(queue, (e) -> {
			Integer res1 = resultMap.get(e.getId1());
			Integer res2 = resultMap.get(e.getId2());
			boolean equalResult = false;
			// FIXME
			if(res1 != null && res2 != null) {
				equalResult =res1.equals(res2);
			} 
//			boolean equalResult = resultMap.get(e.getId1()).equals(resultMap.get(e.getId2()));
			boolean equalGs = gsMap.get(e.getId1()).equals(gsMap.get(e.getId2()));
			
			if(equalGs) {	
				if(equalResult) {
					truePositive.add(new AccumInteger(1));
				}
				else {
					falseNegative.add(new AccumInteger(1));
				}
			}
			else {
				if(equalResult) {
					falsePositive.add(new AccumInteger(1));
				}
				else {
					//trueNegative.add(new AccumInteger(1));
					// skipped because it is dumb
				}
			}
		}, this.getClass());
		
		
		 
		return new GoldStandardPerformance(truePositive.get().get(), falsePositive.get().get(),   falseNegative.get().get());
	}

	
	public GoldStandardPerformance evaluate2(GoldStandardClusters goldStandardClusters, GoldStandardMapping gsMapping, List<Entity> entities) {
		
		Map<TableRowId, RowClusterUUID> resultMap = new ConcurrentHashMap<>();
		Map<TableRowId, InstanceId> gsMap = new ConcurrentHashMap<>();
		Map<RowClusterUUID, Entity> entitiesMap = new ConcurrentHashMap<>();
	 
		
		PersistentParallel6.runAndFinish(entities, (e) -> {
			
			for(TableRowId rowId : e.getRowCluster().getRows()) {
				resultMap.put(rowId, e.getRowClusterUUID());
			}
			
			entitiesMap.put(e.getRowClusterUUID(), e);
			
		}, this.getClass());
		
		
		PersistentParallel6.runAndFinish(goldStandardClusters.getTestingSetMappingClusters().entrySet(), (e) -> {
			for(TableRowId rowId : e.getValue()) {
				gsMap.put(rowId, e.getKey());
			}
			
		}, this.getClass());
		
		Set<TableRowIdPair> set = Collections.newSetFromMap(new ConcurrentHashMap<>());
		
		
		PersistentParallel6.runAndFinish(goldStandardClusters.getCandidates(), (rowId) -> {
			for(TableRowId second : entitiesMap.get(resultMap.get(rowId)).getRowCluster().getRows() ) {
				if(!second.equals(rowId)) {
					set.add(new TableRowIdPair(rowId, second));
				}
			}
			
			for(TableRowId second : goldStandardClusters.getTestingSetMappingClusters().get(gsMap.get(rowId))  ) {
				if(!second.equals(rowId)) {
					set.add(new TableRowIdPair(rowId, second));
				}
			}
		}, this.getClass());
		
		
		 
		ConcurrentAccumulator<AccumInteger> truePositive = new ConcurrentAccumulator<>(AccumInteger.empty());
		ConcurrentAccumulator<AccumInteger> falsePositive = new ConcurrentAccumulator<>(AccumInteger.empty());
//		ConcurrentAccumulator<AccumInteger> trueNegative = new ConcurrentAccumulator<>(AccumInteger.empty());
		ConcurrentAccumulator<AccumInteger> falseNegative = new ConcurrentAccumulator<>(AccumInteger.empty());
		 
		
		PersistentParallel6.runAndFinish(set, (e) -> {
			boolean equalResult = resultMap.get(e.getId1()).equals(resultMap.get(e.getId2()));
			boolean equalGs = gsMap.get(e.getId1()).equals(gsMap.get(e.getId2()));
			
			if(equalGs) {	
				if(equalResult) {
					truePositive.add(new AccumInteger(1));
				}
				else {
					falseNegative.add(new AccumInteger(1));
				}
			}
			else {
				if(equalResult) {
					falsePositive.add(new AccumInteger(1));
				}
				else {
					//trueNegative.add(new AccumInteger(1));
					// skipped because it is dumb
				}
			}
		}, this.getClass());
		
		
		 
		return new GoldStandardPerformance(truePositive.get().get(), falsePositive.get().get(),   falseNegative.get().get());
	}
	
	
	public GoldStandardPerformance directPerformance(Map<InstanceId, Set<TableRowId>> goldStandard, Map<InstanceId, Set<TableRowId>> result) {
		int truePositive = 0;
		int falsePositive = 0;
//		ConcurrentAccumulator<AccumInteger> trueNegative = new ConcurrentAccumulator<>(AccumInteger.empty());
		int falseNegative = 0;
		
		
		Set<TableRowIdPair> fullSet = new HashSet<>();
		
		for(Entry<InstanceId, Set<TableRowId>> gs : goldStandard.entrySet()) {
			
			
			Set<TableRowIdPair> set = new HashSet<>();
			
			Set<TableRowId> gsSet = gs.getValue();
			for(DoubleEntry<TableRowId> entry : DoubleIteraterable.iterable(gsSet)) {
				set.add(new TableRowIdPair(entry.getFirst(), entry.getSecond()));	
			}
			
			if(result.containsKey(gs.getKey())) {
				
				
				Set<TableRowId> res = result.get(gs.getKey());
				 
				for(DoubleEntry<TableRowId> entry : DoubleIteraterable.iterable(res)) {
					TableRowIdPair pair = new TableRowIdPair(entry.getFirst(), entry.getSecond());
					if(!set.contains(pair))
						set.add(pair);	
				}
				
				
				for(TableRowIdPair e : set) {
					fullSet.add(e);
					
					boolean res1 = res.contains(e.getId1());
					boolean res2 = res.contains(e.getId2());
					boolean equalResult = false;
					// FIXME
					if(res1   && res2  ) {
						equalResult = true;
					} 
//					boolean equalResult = resultMap.get(e.getId1()).equals(resultMap.get(e.getId2()));
					boolean equalGs = gsSet.contains(e.getId1()) && gsSet.contains(e.getId2());
					
					if(equalGs) {	
						if(equalResult) {
							truePositive++;
						}
						else {
							falseNegative++;
						}
					}
					else {
						if(equalResult) {
							falsePositive++;
						}
						else {
							//trueNegative.add(new AccumInteger(1));
							// skipped because it is dumb
						}
					}
 
				}
				
				 
				
			}
			else {
				falseNegative += set.size();
			} 
		}
		
		for(Entry<InstanceId, Set<TableRowId>> resEntry : result.entrySet()) {
			
			
			if(!goldStandard.containsKey(resEntry.getKey())) {
				
				for(DoubleEntry<TableRowId> entry : DoubleIteraterable.iterable(resEntry.getValue())) {
					if(!fullSet.contains(new TableRowIdPair(entry.getFirst(), entry.getSecond()))) {
						falsePositive += resEntry.getValue().size();
					}
					
				}
				
				
			}
		}
		
		
		return new GoldStandardPerformance(truePositive, falsePositive,   falseNegative);

	}
	
	public GoldStandardPerformance directPerformanceNonPair(Map<InstanceId, Set<TableRowId>> goldStandard, Map<InstanceId, List<Set<TableRowId>>> result) {
		int truePositive = 0;
		int falsePositive = 0;
//		ConcurrentAccumulator<AccumInteger> trueNegative = new ConcurrentAccumulator<>(AccumInteger.empty());
		int falseNegative = 0;
		
		
		Set<TableRowIdPair> fullSet = new HashSet<>();
		
		for(Entry<InstanceId, Set<TableRowId>> gs : goldStandard.entrySet()) {
			
			
			Set<TableRowIdPair> set = new HashSet<>();
			
			Set<TableRowId> gsSet = gs.getValue();
			for(DoubleEntry<TableRowId> entry : DoubleIteraterable.iterable(gsSet)) {
				set.add(new TableRowIdPair(entry.getFirst(), entry.getSecond()));	
			}
			
			if(result.containsKey(gs.getKey())) {
				
				int maxTruePositive = 0;
				int maxFalsePositive = 0;
				int maxFalseNegative = 0;
				Integer maxIndex = null;
				
				 
				
				for(ListEntry<Set<TableRowId>> cluster : ListIterable.iterable(result.get(gs.getKey()))) {
					int currentTruePositive = 0;
					int currentFalseNegative = 0;
					int currentFalsePositive = 0;
					
					for(TableRowId gsRowId : gs.getValue()) {
						if(cluster.getValue().contains(gsRowId)) {
							currentTruePositive++;
						}
						else {
							currentFalseNegative++;
						}
					}
					
					for(TableRowId rowId : cluster.getValue()) {
						if(!gs.getValue().contains(rowId)) {
							currentFalsePositive++;
						}
					}
					
					if(maxIndex == null || currentTruePositive > maxTruePositive) {
						maxTruePositive = currentTruePositive;
						maxFalsePositive= currentFalsePositive;
						maxFalseNegative= currentFalseNegative;
						maxIndex = cluster.getIndex();
					}
//					else if() {
////						truePositive += maxTruePositive;
//						falsePositive += maxFalsePositive;
////						falseNegative += maxFalseNegative;
//
//						maxTruePositive = currentTruePositive;
//						maxFalsePositive= currentFalsePositive;
//						maxFalseNegative= currentFalseNegative;
//						maxIndex = cluster.getIndex();
//					}
//					else {
////						truePositive += currentTruePositive;
//						falsePositive += currentFalsePositive;
////						falseNegative += currentFalseNegative;
//					}
					
				}
				
				
				truePositive += maxTruePositive;
				falseNegative += maxFalseNegative;
				falsePositive += maxFalsePositive;
				
				for(ListEntry<Set<TableRowId>> cluster : ListIterable.iterable(result.get(gs.getKey()))) {
					if(cluster.getIndex() != maxIndex) {
						falsePositive += cluster.getValue().size();
					}
				}
				
			}
			else {
				falseNegative += gs.getValue().size();
			}
				
			  
				
				 
				 
		}
		
	 
		
		
		return new GoldStandardPerformance(truePositive, falsePositive,   falseNegative);

	}

	public Map<InstanceId, List<Entity>> majorityMapping(Map<InstanceId, Set<TableRowId>> onlyNewGoldStandard, List<Entity> remainingClusters) {
		
		Map<TableRowId, InstanceId> index = new HashMap<>();
		for(Entry<InstanceId, Set<TableRowId>> entry : onlyNewGoldStandard.entrySet()) {
			for(TableRowId rowId : entry.getValue()) {
				index.put(rowId, entry.getKey());
			}
		}
		
		
		Map<InstanceId, List<Entity>> ret = new HashMap<>();
		
		for(Entity entity : remainingClusters) { 
			Map<InstanceId, MutableInt> count = new HashMap<>();
			for(TableRowId rowId : entity.getRowCluster().getRows()) {
				if(index.containsKey(rowId)) { 
					CHMHelper.getFromMap(count, index.get(rowId), () -> new MutableInt()).increment();
				}
			}
			
			if(!count.isEmpty()) {
				int maxCount = 0;
				InstanceId max = null;
				for(Entry<InstanceId, MutableInt> countEntry : count.entrySet()) {
					if(max == null || countEntry.getValue().intValue() > maxCount) {
						maxCount = countEntry.getValue().intValue();
						max = countEntry.getKey();
					}
				}
				
				CHMHelper.getFromMap(ret, max, () -> new ArrayList<>()).add(entity);
			}
			else {
				CHMHelper.getFromMap(ret, InstanceId.fromString(UUID.randomUUID().toString()), () -> new ArrayList<>()).add(entity);
			}
		}
		 
		return ret;
	}

	
	public GoldStandardPerformance directPerformanceWithEntities(Map<InstanceId, Set<TableRowId>> goldStandard,
			Map<InstanceId, List<Entity>> resultRaw) {
		Map<InstanceId, List<Set<TableRowId>>> result = resultRaw.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			List<Set<TableRowId>> set = new ArrayList<>();
			for(Entity entity : e.getValue()) {
				set.add(new HashSet<>(entity.getRowCluster().getRows()));
			}
			return set;
		}));
		return directPerformanceNonPair(goldStandard, result);
	}
	
	public GoldStandardPerformance directPerformanceWithEntitiesWithPairs(Map<InstanceId, Set<TableRowId>> goldStandard,
			Map<InstanceId, List<Entity>> resultRaw) {
		Map<InstanceId, Set<TableRowId>> result = resultRaw.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			Set<TableRowId> set = new HashSet<>();
			for(Entity entity : e.getValue()) {
				set.addAll(entity.getRowCluster().getRows());
			}
			return set;
		}));
		return directPerformance(goldStandard, result);
	}

	public PerformanceMacroPR evaluteFusionSet(Map<QualifiedTripleId, QualifiedValueList> map, Map<QualifiedTripleId, List<FusedValue>> fusionSet,
			Map<QualifiedTripleId, QualifiedValueList> fullCorrectMap) {
		
		int correct = 0;
		int wrong = 0;
		int maxCorrect = map.size();
		int notInGs = 0;
		 
		
		for(Entry<QualifiedTripleId, List<FusedValue>> fact : fusionSet.entrySet()) {
			if(fact.getValue().isEmpty()) {
				throw new IllegalStateException();
			}
			if(map.containsKey(fact.getKey())) {
				QualifiedValueList qvl = map.get(fact.getKey());
				boolean foundWrong = false;
				for(FusedValue factEntry : fact.getValue()) {
					Optional<Boolean> correctResult = valueEvaluator.correctIntern(fact.getKey(), qvl, factEntry);
					if(correctResult.isPresent() && !correctResult.get()) {
						foundWrong = true;
						break;
					}
				}
				if(foundWrong) {
					wrong++;
				}
				else {
					correct++;
				}
			}
			else {
				if (fullCorrectMap.containsKey(fact.getKey())) {
					QualifiedValueList qvl = fullCorrectMap.get(fact.getKey());
					boolean foundWrong = false;
					for (FusedValue factEntry : fact.getValue()) {
						Optional<Boolean> correctResult = valueEvaluator.correctIntern(fact.getKey(), qvl, factEntry);
						if (correctResult.isPresent() && !correctResult.get()) {
							foundWrong = true;
							break;
						}
					}
					if (foundWrong) {
						wrong++;
					} else {

					}
				} else {
					wrong++;
					notInGs++;
				}
			}
		}
		return new PerformanceMacroPR(correct, wrong, maxCorrect, notInGs);
	}
	
	
	

 
	
}
