package expansion.set.evaluation.goldstandard;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ImmuMaker;

public class GoldStandardClusters {
	
	private final Map<InstanceId, Set<TableRowId>> clusters;
	private final  Map<TableRowId, InstanceId > index;
	

	public GoldStandardClusters(Set<TableRowId> rowIds, Map<InstanceId, Set<TableRowId>> testingSetMappingClusters) {
		this(testingSetMappingClusters, true);
	}

	public GoldStandardClusters(Map<InstanceId, Set<TableRowId>> testingSetMappingClusters, boolean flag) {
		 Map<InstanceId, Set<TableRowId>> testingSetMappingClustersTemp = new HashMap<>();
		 for(Entry<InstanceId, Set<TableRowId>> entry : testingSetMappingClusters.entrySet()) {
			 testingSetMappingClustersTemp.put(entry.getKey(),Collections.unmodifiableSet(new HashSet<>(entry.getValue())));
		 }
		 this.clusters = Collections.unmodifiableMap(testingSetMappingClustersTemp);
		 
		 Map<TableRowId, InstanceId> index = new HashMap<>();
		 for(Entry<InstanceId, Set<TableRowId>> clusters : testingSetMappingClusters.entrySet()) {
			 for(TableRowId row : clusters.getValue()) {
				 index.put(row, clusters.getKey());
			 }
		 }
		 this.index = Collections.unmodifiableMap(index);
	}


	public GoldStandardClusters(Map<TableRowId, InstanceId> index) {
		this.index = Collections.unmodifiableMap(new HashMap<>(index));
		 
		 Map<InstanceId, Set<TableRowId>> clusters = new HashMap<>();
		 for(Entry<TableRowId, InstanceId> entry : index.entrySet()) {
			 CHMHelper.getFromMap(clusters, entry.getValue(), () -> new HashSet<>()).add(entry.getKey());
		 }
 
		 this.clusters = ImmuMaker.mapWithSet(clusters);
		 
	}


	public Map<InstanceId, Set<TableRowId>> getTestingSetMappingClusters() {
		return clusters;
	}
	
	public  Set<TableRowId>  getCluster(InstanceId instanceId) {
		return clusters.get(instanceId);
	}


	public Set<TableRowId> getCandidates() {
		return index.keySet();
	}
	
	public InstanceId getMapping(TableRowId rowId) {
		return index.get(rowId);
	}


	public Map<TableRowId, InstanceId> getIndex() {
		return index;
	}
	
	
	
	

}
