package expansion.set.evaluation.goldstandard;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdsSet;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;

public class GoldStandardClustersSetPerClass {

	private final Map<ClassId, GoldStandardClustersSet> goldStandardClustersSet;
	public GoldStandardClustersSetPerClass(Map<ClassId, GoldStandardClustersSet> ret) {
		this.goldStandardClustersSet = Collections.unmodifiableMap(new HashMap<>(ret));
	}


	public Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>> getFactsMapOf(int i) {
		return goldStandardClustersSet.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().getSet(i).getFactsMap()));
	}

	public Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>> getFactsMapOfOtherThan(int i) {
		return goldStandardClustersSet.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			Map<QualifiedTripleId, QualifiedValueList> map = new HashMap<>();

			Set<Integer> keys = e.getValue().getSetsKeys();
			for (Integer key : keys) {
				if (!key.equals(i)) {
					map.putAll(e.getValue().getSet(key).getFactsMap());
				}
			}

			return map;
		}));
	}

	public Map<ClassId, Set<TableRowId>> getInputOf(int i) {
		return goldStandardClustersSet.entrySet().stream()
				.collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().getSet(i).getClusters().getIndex().keySet()));
	}

	public Map<ClassId, Set<TableRowId>> getInputOfOtherThan(int i) {
		return goldStandardClustersSet.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			Set<TableRowId> map = new HashSet<>();

			Set<Integer> keys = e.getValue().getSetsKeys();
			for (Integer key : keys) {
				if (!key.equals(i)) {
					map.addAll(e.getValue().getSet(key).getClusters().getIndex().keySet());
				}
			}

			return map;
		}));
	}

	public Map<ClassId, Set<InstanceId>> getNewInstanceIdsOf(int i) {
		return goldStandardClustersSet.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().getSet(i).getNewInstances()));
	}

	public Map<ClassId, Set<InstanceId>> getNewInstanceIdsOfOtherThan(int i) {
		return goldStandardClustersSet.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			Set<InstanceId> map = new HashSet<>();

			Set<Integer> keys = e.getValue().getSetsKeys();
			for (Integer key : keys) {
				if (!key.equals(i)) {
					map.addAll(e.getValue().getSet(key).getNewInstances());
				}
			}

			return map;
		}));
	}

	public Map<ClassId, GoldStandardClusters> getClustersOf(int i) {
		return goldStandardClustersSet.entrySet().stream()
				.collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().getSet(i).getClusters()));
	}

	public Map<ClassId, GoldStandardClusters> getClustersOfOtherThan(int i) {
		return goldStandardClustersSet.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			Map<InstanceId, Set<TableRowId>> map = new HashMap<>();

			Set<Integer> keys = e.getValue().getSetsKeys();
			for (Integer key : keys) {
				if (!key.equals(i)) {
					map.putAll(e.getValue().getSet(key).getClusters().getTestingSetMappingClusters());
				}
			}

			return new GoldStandardClusters(map, true);
		}));
	}

	public Map<ClassId, TableRowIdsSet> getFullInput() {
		return goldStandardClustersSet.entrySet().stream().collect(Collectors.toMap(Entry::getKey,
				(e) -> new TableRowIdsSet(e.getValue().getFull().getClusters().getIndex().keySet(), new UuidStringId(e.getKey().getFinalPart()))));

	}

	public Map<ClassId, Set<TableRowId>> getFullInputRaw() {
		return goldStandardClustersSet.entrySet().stream()
				.collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().getFull().getClusters().getIndex().keySet()));

	}

	public Map<ClassId, Map<InstanceId, Integer>> getConflictingSets() {
		return goldStandardClustersSet.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().getConflictingClustersIndex()));
	}

	public int getSetCount() {
		Integer setCount = null;
		for (Entry<ClassId, GoldStandardClustersSet> entry : goldStandardClustersSet.entrySet()) {
			int inner = entry.getValue().getSetCount();
			if (setCount == null)
				setCount = inner;
			else if (!setCount.equals(inner)) {
				throw new IllegalStateException();
			}
		}
		return setCount;
	}

	public Set<ClassId> getClasses() {
		return goldStandardClustersSet.keySet();
	}

	public Map<ClassId, GoldStandardClustersSet> getPerClass() {
		return goldStandardClustersSet;
	}
}
