package expansion.set.evaluation.goldstandard.oktie;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import com.google.common.primitives.Ints;

import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.DoubleIteraterable;
import de.mannheim.uni.expansion.utils.DoubleIteraterable.DoubleEntry;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.containers.Pair;

public class OktieClusterEvaluator {
	@Inject	
	public OktieClusterEvaluator() {
		
	}

	public <E,F> OktiePerformance oktieEvaluateWithGsMap(Map<E, Set<F>> gsMap, List<Set<F>> clusters ) {
		Map<F, E> gsIndex = new HashMap<>();
		for(Entry<E, Set<F>> entry : gsMap.entrySet()){
			for(F point : entry.getValue()) {
				gsIndex.put(point, entry.getKey());
			}
		}

		return oktieEvaluate(gsMap, gsIndex, clusters, null );
	}

	public <E, F> OktiePerformance oktieEvaluateWithGsMap(Map<E, Set<F>> gsMap, List<Set<F>> clusters, Set<E> newEntities, WeightGetter<E> weightGetter) {
		Map<F, E> gsIndex = new HashMap<>();
		for (Entry<E, Set<F>> entry : gsMap.entrySet()) {
			for (F point : entry.getValue()) {
				gsIndex.put(point, entry.getKey());
			}
		}

		return oktieEvaluate(gsMap, gsIndex, clusters, newEntities, weightGetter);
	}

	public <E, F> Map<F, E> getIndex(Map<E, Set<F>> gsMap) {
		Map<F, E> gsIndex = new HashMap<>();
		for(Entry<E, Set<F>> entry : gsMap.entrySet()){
			for(F point : entry.getValue()) {
				gsIndex.put(point, entry.getKey());
			}
		}
		return gsIndex;
	}

	public <E, F> OktiePerformance oktieEvaluateWithGsMap(Map<E, Set<F>> gsMap, List<Set<F>> clusters, Set<E> newEntities) {
		Map<F, E> gsIndex = getIndex(gsMap);

		return oktieEvaluate(gsMap, gsIndex, clusters, newEntities );
	}
	
	public <E,F> OktiePerformance oktieEvaluateWithGSIndex(Map<F, E> gsIndex, List<Set<F>> clusters) {
		Map<E, Set<F>> gsMap = new HashMap<>();
		
		for(Entry<F, E> entry : gsIndex.entrySet()) {
			CHMHelper.getFromMap(gsMap, entry.getValue(), () -> new HashSet<>()).add(entry.getKey());
		}
		
		return oktieEvaluate(gsMap, gsIndex, clusters);
	}
	
	
	private <E,F> OktiePerformance oktieEvaluate(Map<E, Set<F>> gs, Map<F, E> gsIndex, List<Set<F>> clusters, Set<E> newEntities) {
		Map<E, OktieMapping> mapping = oktieMapping(gsIndex, clusters);
		 
		return oktieEvaluate(gs, gsIndex, clusters, mapping, newEntities, (f) -> 1.0);
	}

	private <E, F> OktiePerformance oktieEvaluate(Map<E, Set<F>> gs, Map<F, E> gsIndex, List<Set<F>> clusters, Set<E> newEntities,
			WeightGetter<E> weightGetter) {
		Map<E, OktieMapping> mapping = oktieMapping(gsIndex, clusters);

		return oktieEvaluate(gs, gsIndex, clusters, mapping, newEntities, weightGetter);
	}
	

	public <E, F> OktiePerformance oktieEvaluate(Map<E, Set<F>> gs, Map<F, E> gsIndex, List<Set<F>> clusters) {
		return oktieEvaluate(gs, gsIndex, clusters,null);
	}
	
	
	public <E, F> OktiePerformance oktieEvaluate(Map<E, Set<F>> gs, Map<F, E> gsIndex, List<Set<F>> clusters, Map<E, OktieMapping> mapping, Set<E> newEntities,
			WeightGetter<E> weightGetter) {
		double precisionSum = 0.0;
		double recallSum = 0.0;
		double precisionWeightSum = 0.0;
		double recallWeightSum = 0.0;
		
		
		// COMPUTING PRECISION AND RECALL
		for(Entry<E, Set<F>> entry : gs.entrySet()) {
			double weight = (entry.getValue().size() / ((double) gsIndex.size()));
			double inputWeight = weightGetter.getWeights(entry.getKey());
			weight = inputWeight * weight;
			recallWeightSum += (weight);
			OktieMapping oktieMapping = mapping.get(entry.getKey());
			if(oktieMapping != null) {
				
				double precision = oktieMapping.getIntersecion() / ((double) clusters.get(oktieMapping.getCluster()).size());
				double recall = oktieMapping.getIntersecion() / ((double) entry.getValue().size());
				
				
				precisionSum += (weight * precision);
				recallSum += (weight * recall);
				
				precisionWeightSum += weight;
				
			}
			else {
				int sdf = 13;
			}
		}
		
		// compute clustering precision using pairs
		double clusteringPrecisionSum = 0.0;
		double clusteringPrecisionCount = 0.0;
		for(Set<F> cluster : clusters) {
			double maxWeight = -1.0;
			for (F point : cluster) {
				E gs1 = gsIndex.get(point);
				if (gs1 != null) {
					double score = weightGetter.getWeights(gs1);
					if (score > maxWeight) {
						maxWeight = score;
					}
				}
			}

			if (maxWeight < 0.0) {
				throw new IllegalStateException();
			}

			maxWeight = 1.0;

			if(cluster.size() > 1) {
				double pairsCount = 0.0;
				double pairsCorrect = 0;
				for(DoubleEntry<F> doubleEntry : DoubleIteraterable.iterable(cluster)) {

					E gs1 = gsIndex.get(doubleEntry.getFirst());
					E gs2 = gsIndex.get(doubleEntry.getSecond());

					double score = 1.0;
					// if (gs1 != null)
					// score += weightGetter.getWeights(gs1);
					//
					// if (gs2 != null)
					// score += weightGetter.getWeights(gs2);

					if(gs1 != null || gs2 != null) {

						pairsCount += score;
					}
					if(gs1 != null && gs2 != null && gs1.equals(gs2)) {
						pairsCorrect += score;
					}
				}
				if(pairsCount > 0.1) {
					clusteringPrecisionSum += (pairsCorrect / pairsCount) * maxWeight;
					clusteringPrecisionCount += maxWeight;
				}
			}

		}
		
		if (clusteringPrecisionCount < 0.0001) {
			clusteringPrecisionSum = 1.0;
			clusteringPrecisionCount = 1.0;
		}
		
		// average precisions and clustering percisions
		double precision = precisionSum / precisionWeightSum;
		double recall = recallSum / recallWeightSum;
		double clusteringPrecision = clusteringPrecisionSum / clusteringPrecisionCount;
		
		if(Double.isNaN(clusteringPrecision)){
			int sdf = 13;
		}
			
		// compute penalized clustering precision
		double mappingPenalizedClusteringPrecision;
		if (true) {

			int mappingPenalizationSize = mapping.size();
			if (gs.size() < mappingPenalizationSize) {
				mappingPenalizedClusteringPrecision = (((double) gs.size()) / mappingPenalizationSize) * clusteringPrecision;
			} else if (gs.size() > mappingPenalizationSize) {

				mappingPenalizedClusteringPrecision = (mappingPenalizationSize / ((double) gs.size())) * clusteringPrecision;
			} else {
				mappingPenalizedClusteringPrecision = clusteringPrecision;
			}
		}

		double allPenalizedClusteringPrecision;
		if (true) {
		// compute penalized clustering precision

		int allPenalizationSize = clusters.size();
		if(gs.size() < allPenalizationSize) {
			allPenalizedClusteringPrecision = ( ((double) gs.size()) / allPenalizationSize) * clusteringPrecision;
		}
		else if(gs.size() > allPenalizationSize){

			allPenalizedClusteringPrecision = ( allPenalizationSize / ((double) gs.size())) * clusteringPrecision;
		}
		else {
			allPenalizedClusteringPrecision = clusteringPrecision;
		}
		}
		
		double gsSize = gs.size();
		double noMapping = clusters.size() - mapping.size();
		double missingMapping = gs.size() - mapping.size();
		double penalizedClusteringPrecision;
		if (false) {
			penalizedClusteringPrecision = clusteringPrecision * (1.0 - ((noMapping + missingMapping) / gsSize));
		} else {
			if (allPenalizedClusteringPrecision < mappingPenalizedClusteringPrecision) {
				penalizedClusteringPrecision = allPenalizedClusteringPrecision;
			} else {
				penalizedClusteringPrecision = mappingPenalizedClusteringPrecision;
			}
		}
		 
		int newMappedClusters = 0;
		if(newEntities != null) {
			newMappedClusters = Ints.checkedCast(mapping.keySet().stream().filter((e) -> newEntities.contains(e)).count());
		}

		if (clusters.size() - mapping.size() < 0) {
			int sfd = 123;
		}

		return new OktiePerformance(clusters.size(), mapping.size() - newMappedClusters, newMappedClusters, precision, recall, clusteringPrecision, penalizedClusteringPrecision);
	}

	public <E,F>  Map<E, OktieMapping> oktieMapping(Map<F, E> gsIndex, List<Set<F>> clusters) {
		Map<E, Pair<Double, OktieMapping>> highest = new HashMap<>();
		for(ListEntry<Set<F>> cluster : ListIterable.iterable(clusters)) {
			double size = cluster.getValue().size();
			Map<E, MutableInt> count = new HashMap<>();
			for(F point : cluster.getValue()) {
				CHMHelper.getFromMap(count, gsIndex.get(point), () -> new MutableInt()).increment();
			}
			

			
			E maxCluster = null;
			int maxIntersection = 0;
			
			for(Entry<E, MutableInt>  countEntry : count.entrySet()) {
				if (maxCluster == null || countEntry.getValue().intValue() > maxIntersection) {
					maxCluster = countEntry.getKey();
					maxIntersection = countEntry.getValue().intValue();
					
				}
			}
			
			if(maxCluster != null) {
				double score = maxIntersection / size;
				if (score > 0.5) {
					if (highest.containsKey(maxCluster)) {

						Pair<Double, OktieMapping> current = highest.get(maxCluster);
						// double largerSize = size;
						// double otherClusterSize =
						// clusters.get(current.getSecond().getCluster()).size();
						// if (otherClusterSize > largerSize) {
						// largerSize = otherClusterSize;
						// }
						//
						// double thisScore = maxIntersection / largerSize;
						// double otherScore =
						// current.getSecond().getIntersecion()
						// / largerSize;

						if ((maxIntersection > current.getSecond().getIntersecion())) {

							highest.put(maxCluster, new Pair<>(score, new OktieMapping(cluster.getIndex(), maxIntersection)));
						}
					} else {
						highest.put(maxCluster, new Pair<>(score, new OktieMapping(cluster.getIndex(), maxIntersection)));
					}
				}
			}
			
			
		}
		
		return  highest.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().getSecond())); 
	}
	
	
	public <E,F>  Map<Integer, Optional<E>> simpleMajorityMap(Map<F, E> gsIndex, List<Set<F>> clusters) {
		Map<Integer, Optional<E>> ret = new HashMap<>();
		for(ListEntry<Set<F>> cluster : ListIterable.iterable(clusters)) {
			double size = cluster.getValue().size();
			Map<E, MutableInt> count = new HashMap<>();
			for(F point : cluster.getValue()) {
				CHMHelper.getFromMap(count, gsIndex.get(point), () -> new MutableInt()).increment();
			}
			
			
			
			E maxCluster = null;
			int maxIntersection = 0;
			
			for(Entry<E, MutableInt>  countEntry : count.entrySet()) {
				if(maxCluster == null && countEntry.getValue().intValue() > maxIntersection) {
					maxCluster = countEntry.getKey();
					maxIntersection = countEntry.getValue().intValue();
					
				}
			}
			
			if(maxCluster != null) {
				double score = maxIntersection / size;
				
				if(score > 0.5) {
					ret.put(cluster.getIndex(), Optional.of(maxCluster));
					
					 
				} 
				else {

					ret.put(cluster.getIndex(), Optional.empty());
				}
			}
			else {
				ret.put(cluster.getIndex(), Optional.empty());
			}
			
			
		}
		
		return  ret; 
	}
	
	@FunctionalInterface
	public static interface WeightGetter<E> {
		double getWeights(E e);
	}
	public static class OktieMapping {
		private final Integer cluster;
		private final int intersecion;
		public OktieMapping(Integer cluster, int intersecion) {
			super();
			this.cluster = cluster;
			this.intersecion = intersecion;
		}
		public Integer getCluster() {
			return cluster;
		}
		public int getIntersecion() {
			return intersecion;
		}
		
	}
}
