package expansion.set.evaluation.goldstandard.oktie;

import java.util.function.BinaryOperator;

import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.Accumulateable;

public class OktiePerformance  implements Accumulateable<OktiePerformance> {
	private final int clusters;
	private final int existingClustersWithMapping;
	private final int newClustersWithMapping;
	private final double precision;
	private final double recall;
	private final double clusteringPrecision;
	private final double penalizedClusteringPrecision;
	private final double fmeasure;
	private final double fmeasureNonPenzlized;
	private final int count;
	 
	public OktiePerformance( ) {
		super();
		this.clusters = 0;
		this.existingClustersWithMapping = 0;
		this.newClustersWithMapping = 0;
		this.precision = 0.0;
		this.recall = 0.0;
		this.clusteringPrecision = 0.0;
		this.penalizedClusteringPrecision = 0.0;
		this.fmeasure = 0.0;
		this.count = 0;
		this.fmeasureNonPenzlized = 0.0;
	}
	public OktiePerformance(int clusters, int existingClustersWithMapping, int newClustersWithMapping, double precision, double recall, double clusteringPrecision, double penalizedClusteringPrecision) {
		super();
		this.clusters = clusters;
		this.existingClustersWithMapping = existingClustersWithMapping;
		this.newClustersWithMapping = newClustersWithMapping;
		this.precision = precision;
		this.recall = recall;
		this.clusteringPrecision = clusteringPrecision;
		this.penalizedClusteringPrecision = penalizedClusteringPrecision;
		Double fmeasure = (2.0 * this.penalizedClusteringPrecision * this.recall) / (this.penalizedClusteringPrecision + this.recall);
		if(fmeasure.isNaN()) {
			this.fmeasure = 0.0;
		}
		else {
			this.fmeasure = fmeasure;
		}

		Double fmeasureNonPenzlized = (2.0 * this.clusteringPrecision * this.recall) / (this.clusteringPrecision + this.recall);
		if (fmeasureNonPenzlized.isNaN()) {
			this.fmeasureNonPenzlized = 0.0;
		} else {
			this.fmeasureNonPenzlized = fmeasureNonPenzlized;
		}

		this.count = 1;
	}
	
	private OktiePerformance(int clusters, int existingClustersWithMapping, int newClustersWithMapping, double precision, double recall,
			double clusteringPrecision, double penalizedClusteringPrecision, double fmeasure, double fmeasureNonPenzlized, int count) {
		super();
		this.clusters = clusters;
		this.existingClustersWithMapping = existingClustersWithMapping;
		this.newClustersWithMapping = newClustersWithMapping;
		this.precision = precision;
		this.recall = recall;
		this.clusteringPrecision = clusteringPrecision;
		this.penalizedClusteringPrecision = penalizedClusteringPrecision;
		this.fmeasure = fmeasure;
		this.count = count;
		this.fmeasureNonPenzlized = fmeasureNonPenzlized;
	}
	
	public double getPrecisionMacro() {
		return precision / count;
	}
	public double getRecallMacro() {
		return recall / count;
	}
	public double getClusteringPrecisionMacro() {
		return clusteringPrecision / count;
	}
	public double getPenalizedClusteringPrecisionMacro() {
		return penalizedClusteringPrecision / count;
	}
	public double getFMeasureMacro() {
		return fmeasure / count;
	}
	
	public double getFMeasureNonPenalizedMacro() {
		return fmeasureNonPenzlized / count;
	}

	public int getClusters() {
		return clusters;
	}
	 
	public int getExistingClustersWithMapping() {
		return existingClustersWithMapping;
	}
	public int getNewClustersWithMapping() {
		return newClustersWithMapping;
	}
	public static   OktiePerformance accumulator(OktiePerformance a, OktiePerformance b) {
		
		return new OktiePerformance(a.clusters + b.clusters, a.existingClustersWithMapping + b.existingClustersWithMapping,
				a.newClustersWithMapping + b.newClustersWithMapping, a.precision + b.precision, a.recall + b.recall,
				a.clusteringPrecision + b.clusteringPrecision, a.penalizedClusteringPrecision + b.penalizedClusteringPrecision, a.fmeasure + b.fmeasure,
				a.fmeasureNonPenzlized + b.fmeasureNonPenzlized, a.count + b.count);
	}
	
	@Override
	public BinaryOperator<OktiePerformance> getAccumulator() { 
		return OktiePerformance::accumulator;
	}
	@Override
	public OktiePerformance accumulate(OktiePerformance other) {
		return OktiePerformance.accumulator(this, other);
	}
	@Override
	public Class<OktiePerformance> getAcumClass() {
		return OktiePerformance.class;
	}
	
}
