package expansion.set.evaluation.matching;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableBoolean;

import com.google.common.collect.BiMap;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.QuantityObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;
import de.uni_mannheim.informatik.dws.matching.LabelMatcher;
import de.uni_mannheim.informatik.dws.matching.LabelMatcher.MatchStrictness;
import expansion.set.creation.Entity;

public class EntityMatcher {

	private final LabelMatcher labelMatcher;
	private final ValueEvaluator valueEvaluator;
	private final OntologyRepository ontology;
	private final KnowledgeBase knowledgeBase;
//	private final Map<ClassPropertyId, Double> propertyScores; 

	@Inject
	public EntityMatcher(LabelMatcher labelMatcher, ValueEvaluator valueEvaluator,OntologyRepository ontology, KnowledgeBase knowledgeBase) {

		this.labelMatcher = labelMatcher;
		this.valueEvaluator = valueEvaluator;
		this.ontology =   ontology;
		this.knowledgeBase = knowledgeBase;
//		ConcurrentHashMap<ClassPropertyId, Double> score = new ConcurrentHashMap<>();
//		for(ClassPropertyId cpid : this.knowledgeBase.getClassPropertyIds()) {
//			if(ontology.getDataType(cpid.getPropertyId()) == ValueType.QUANTITY) {
//				score.put(cpid, 0.5);
//			}
//			else {
//			
//			BiMap<InstanceId, Integer> instances = this.knowledgeBase.getInstances(cpid.getClassId());
// 
//			
//			AtomicInteger totalCount = new AtomicInteger();
//			Set<AbstractObject> abstractCount = Collections.newSetFromMap(new ConcurrentHashMap<>());
//			
//			PersistentParallel6.runAndFinish(instances.entrySet(), (instanceEntry) -> {
//				Optional<List<AbstractObject>> list = this.knowledgeBase.getValuesAsList(instanceEntry.getKey(), cpid.getPropertyId());
//				if(list.isPresent()) {
//					totalCount.incrementAndGet();
//					abstractCount.addAll(list.get());
//				}
//
//				
//			}, this.getClass());
//			
//			double total= totalCount.get();
//			double valueCount = Math.min(abstractCount.size(), totalCount.get());
//			double normalizer = Math.log((total/total) / (1/total)) ;
//			double pmi = Math.log((valueCount/total) / (1/total)) ;
//			double npmi = pmi / normalizer;
//			double scoreDouble = 0.5;
//			if(!Double.isNaN(npmi))
//				scoreDouble = npmi;
//			score.put(cpid, scoreDouble);
//					System.out.println(cpid.toString() + " : " + scoreDouble);
//			}
//		}
//		
//		this.propertyScores = ImmuMaker.map(score);

	}
	private SimpleLockableConcurrentCache<ClassPropertyId, Double> cpidScoreCache = new SimpleLockableConcurrentCache<>();
	
	public double getScoreFromCache(ClassPropertyId cpid) {
		return cpidScoreCache.getOrCreate(cpid, () -> {
			return getScore(cpid);
		});
	}
	
	public double getScore(ClassPropertyId cpid) {
		double score = 0.5;
		if(ontology.getDataType(cpid.getPropertyId()) == ValueType.QUANTITY) {
//			score = 0.5;
		}
		else {
		
		BiMap<InstanceId, Integer> instances = this.knowledgeBase.getInstances(cpid.getClassId());

		
		AtomicInteger totalCount = new AtomicInteger();
		Set<AbstractObject> abstractCount = Collections.newSetFromMap(new ConcurrentHashMap<>());
		
		PersistentParallel6.runAndFinish(instances.entrySet(), (instanceEntry) -> {
			 List<AbstractObject>  list = this.knowledgeBase.getValuesAsListWithEqualProperties(instanceEntry.getKey(), cpid.getPropertyId());
			if(list != null) {
				totalCount.incrementAndGet();
				abstractCount.addAll(list);
			}

			
		}, this.getClass());
		
		double total= totalCount.get();
		double valueCount = Math.min(abstractCount.size(), totalCount.get());
		double normalizer = Math.log((total/total) / (1/total)) ;
		double pmi = Math.log((valueCount/total) / (1/total)) ;
		double npmi = pmi / normalizer; 
		if(!Double.isNaN(npmi))
			score = npmi; 
		}
		return score;
	}

	public InstanceId match(Entity entity) {
		return null;
	}
	
	public static Pattern pattern = Pattern.compile("^(.)\\1+$");
	
	public static boolean isOnlyOneChar(String str) {
		Matcher matcher = pattern.matcher(str);
		if(matcher.find())
			return true;
		else
			return false;
	}

	public Map<InstanceId, List<ScoredEntity>> matchMultiple(List<Entity> entities ) {
		
		return matchMultiple(entities, null);
	}
	

	public Map<InstanceId, List<ScoredEntity>> matchMultiple(List<Entity> entities, Map<ClassId, Set<ClassId>> allParents) {
		return matchMultipleAndUnsure(entities, allParents).getFirst();
		
	}
	
	public Pair<Map<InstanceId, List<ScoredEntity>>, List<Entity>> matchMultipleAndUnsure(List<Entity> entities, Map<ClassId, Set<ClassId>> allParents) {
		
		Tri<Map<InstanceId, List<ScoredEntity>>, List<Entity>, List<ScoredEntity>> ret = matchMultipleAndUnsureAndNew(entities, allParents);
		return new Pair<>(ret.getA(), ret.getB());
	}

	public Tri<Map<InstanceId, List<ScoredEntity>>, List<Entity>, List<ScoredEntity>> matchMultipleAndUnsureAndNew(List<Entity> entities,
			Map<ClassId, Set<ClassId>> allParents) {
		 
		Set<String> uuidStrings = new HashSet<>();
//		uuidStrings.add("106034d6-3935-11e8-99cd-7d77a4539a6c");
//		uuidStrings.add("106081fe-3935-11e8-99cd-693bad3825ae");
//		uuidStrings.add("105d7599-3935-11e8-99cd-2f719b6a3c11");
//		uuidStrings.add("105e3868-3935-11e8-99cd-f946a08ddacf");
//		uuidStrings.add("105eae29-3935-11e8-99cd-c5e96badde7e");
//		uuidStrings.add("1f7a6ddc-4721-11e8-8cee-dbc1f0e63187");
//		uuidStrings.add("f7b76750-471f-11e8-8cee-21f88777e52d");
//		uuidStrings.add("1f7a6ddc-4721-11e8-8cee-dbc1f0e63187");
//		uuidStrings.add("1f75b2d7-4721-11e8-8cee-dba4518e7226");
//		uuidStrings.add("6ac777fd-4721-11e8-8cee-93001ab36dc0");
//		uuidStrings.add("6ac6b467-4721-11e8-8cee-2570fe11d774");
//		 
//		uuidStrings.add("6ac68dfd-4721-11e8-8cee-1319bdad95af");
//		uuidStrings.add("6ac7c581-4721-11e8-8cee-17a15ceac8aa");
//		uuidStrings.add("6ac6b454-4721-11e8-8cee-7505a72d09c6");
		uuidStrings.add("8b16931d-4f15-11e8-888b-8de17f96dcdb");
		uuidStrings.add("b1871a38-4f0d-11e8-bde9-f3f632161caa");
		uuidStrings.add("096880d2-4f0e-11e8-bde9-e5083e9778eb");
		uuidStrings.add("b1832214-4f0d-11e8-bde9-a7436fdf95f9");
		uuidStrings.add("");
		uuidStrings.add("");
		
		
		
		
		
		
		Comparator<Pair<Integer, Double>> comp = (p1, p2) -> {
			return Double.compare(p2.getSecond(), p1.getSecond());
		};
		Comparator<Pair<InstanceId, Double>> comp2 = (p1, p2) -> {
			return Double.compare(p2.getSecond(), p1.getSecond());
		};

		ConcurrentHashMap<InstanceId, TreeSet<Pair<Integer, Double>>> candidatesMap = new ConcurrentHashMap<>();
		ConcurrentHashMap<Integer, TreeSet<Pair<InstanceId, Double>>> matchedMap = new ConcurrentHashMap<>();
		ConcurrentLinkedQueue<Entity> unsure = new ConcurrentLinkedQueue<>();
		ConcurrentLinkedQueue<ScoredEntity> newEntities = new ConcurrentLinkedQueue<>();
		
		PersistentParallel6.runAndFinish(ListIterable.iterable(entities), (entityEntry) -> {

			Integer index = entityEntry.getIndex();
			Entity entity = entityEntry.getValue();
			Set<ClassId> classIds = new HashSet<>();
			classIds.add(entity.getClassId());
//			classIds.addAll(getAllSubClases(entity.getClassId()));
			if(allParents != null) {
				classIds.addAll(allParents.get(entity.getClassId()));
				// classIds.addAll(getAllParentClases(entity.getClassId()));
			}
			boolean addToUnsure = false;
		
			
			Optional<Map<InstanceId, Double>> matches = labelMatcher.getMatches(entity.getLabels().keySet(),classIds ,
					MatchStrictness.NON_STRICT);

			if(entity.getLabels().keySet().contains("five children and it")) {
				int sdf = 123 ;
			}
			
			if(entity.getLabels().keySet().contains("brazil")) {
				int sdf = 123 ;
			}
			
			// for (QualifiedPropertyId property :
			// entityEntry.getValue().getValues().keySet()) {
			// System.out.println(property.toString());
			// } 
			if (matches.isPresent()) {
				
//				matches.get().clear();
//				matches.get().put(new InstanceId("http://dbpedia.org/resource/Angel_Stadium_of_Anaheim", "http://dbpedia.org/resource/Angel_Stadium_of_Anaheim"), 1.0);

				if(!entity.getValues().isEmpty()) {
					int sdf= 123;
				}
				
				InstanceId maxMatch = null;
				double maxScore = 0.0;

				for (Entry<InstanceId, Double> match : matches.get().entrySet()) {

					if(match.getKey().getInstanceId().equals("http://dbpedia.org/resource/Helsinki")) {
						int sdf= 123;
					}
				  
					double correctValues = 0.0;
					correctValues += Math.max(0.0, (match.getValue() - 1.3) / 1.4);
					int totalValues = 0;

					for (Entry<QualifiedPropertyId, FusedValue> value : entity.getValues().entrySet()) {
						

						Double singleScore = value.getValue().getSingleScore() / value.getValue().getSingleContributors();
						Double minScore = null;
						
						List<FusedValue> fvList = new ArrayList<>(); 
						
						
						if(value.getValue().getSingle() instanceof StringObject) {
							String str = ((StringObject) value.getValue().getSingle()).getString();
							if(str.length() < 3 || str.trim().equals("") || str.trim().equals("??") || isOnlyOneChar(str)) {
								minScore = 0.0;
							}
							else if(value.getValue().getSingleScore() != null) {
								String[] result = ((StringObject) value.getValue().getSingle()).getString().split(";");
								if(result.length > 1) {
									for(int i = 0; i < result.length; i++) {
										fvList.add(new FusedValue(new StringObject(result[i].trim()), value.getValue().getSingleScore(), value.getValue().getSingleContributors()));
									}
								}
							}
						}
						
						if(value.getValue().getSingle() instanceof QuantityObject) {
							QuantityObject qo = ((QuantityObject) value.getValue().getSingle());
							if(Math.abs(qo.getValue()) < 0.01) {
								minScore = 0.0;
							}
						}
						
					
						
						Double maxInnerScore = null;
						
						double propertyWeight = getScoreFromCache(new ClassPropertyId(entity.getClassId(), value.getKey().getPropertyId()));;
						
						for(FusedValue fv : fvList) {
							Optional<Double> scoreOptional = valueEvaluator.score(new QualifiedTripleId(entity.getClassId(), match.getKey(), value.getKey()),
									fv);
							
							if(scoreOptional.isPresent()) {
								double score = scoreOptional.get(); 
//								if(singleScore != null) 
//									score = score;
								double fvScore;
								fvScore = score;
								fvScore *= propertyWeight;
								fvScore -= (0.50 * singleScore * propertyWeight);
								
								if(maxInnerScore == null)
									maxInnerScore = 0.0;
								maxInnerScore += fvScore;
							}
						}
						
						Optional<Double> scoreOptional = valueEvaluator.score(new QualifiedTripleId(entity.getClassId(), match.getKey(), value.getKey()),
								value.getValue());
						
						if(scoreOptional.isPresent()) {
							double score = scoreOptional.get();
//							if(singleScore != null) 
//								score = score;
							double fvScore;
							fvScore = score;
							fvScore *= propertyWeight;
							fvScore -= (0.50 * singleScore * propertyWeight);
							if(maxInnerScore == null || fvScore > maxInnerScore) {
								maxInnerScore = fvScore;
							}
						}
						if(maxInnerScore != null) {
							if(minScore != null)
								correctValues += Math.max(maxInnerScore, minScore);
							else
	
								correctValues +=  maxInnerScore ;
						}
						 
					}

					
					 
					
					
					double score = (correctValues);// / (1.0 + totalValues);
					
					
					if(score  >= 0.49999 ) {
						if( (maxMatch == null || score > maxScore)) {
							maxMatch = match.getKey();
							maxScore = score;
						}
					}
					else if (score > 0.40) {
						addToUnsure = true;
						
					}
					else {
						int sdf= 123;
					}
				
				}
				
				if(uuidStrings.contains(entity.getRowClusterUUID().toString())) {
					int sdf= 123;
				}
				
				if(maxMatch == null) {
					if(addToUnsure) {
						unsure.add(entity);
					}
					else {
						newEntities.add(new ScoredEntity(entity, maxScore));
					}
				}
				else {
				
					TreeSet<Pair<Integer, Double>> treeSet = CHMHelper.getFromCHM(candidatesMap, maxMatch, () -> new TreeSet<>(comp));
	
					synchronized (treeSet) {
						treeSet.add(new Pair<>(index, maxScore));
					}
	
					TreeSet<Pair<InstanceId, Double>> treeSet2 = CHMHelper.getFromCHM(matchedMap, index, () -> new TreeSet<>(comp2));
	
					synchronized (treeSet2) {
						treeSet2.add(new Pair<>(maxMatch, maxScore));
					}
				}

			}
			else {
				
				if(uuidStrings.contains(entity.getRowClusterUUID().toString())) {
					int sdf= 123;
				}
				int sdf = 123;
			}

	
		} , this.getClass(), "MainMatchingLoop");

		// if (matches.isPresent()) {
		// double maxSimilarity = maxMatchPair.get().getSecond();
		// InstanceId maxMatch = maxMatchPair.get().getFirst();
		//
		// if (maxMatch != null) {
		// List<ScoredEntity> list = CHMHelper.getFromCHM(ret, maxMatch, () ->
		// new ArrayList<>());
		// synchronized (list) {
		// list.add(new ScoredEntity(entity, maxSimilarity));
		// }
		// }
		// }
		if(true)
		{

		ConcurrentHashMap<InstanceId, List<ScoredEntity>> ret = new ConcurrentHashMap<>(); 

		PersistentParallel6.runAndFinish(candidatesMap.entrySet(), (entry) -> {

			TreeSet<Pair<Integer, Double>> treeSet = entry.getValue();

		 
			List<ScoredEntity> list = CHMHelper.getFromCHM(ret, entry.getKey(), () -> new ArrayList<>());
			for (Pair<Integer, Double> pair : treeSet) {
				list.add(new ScoredEntity(entities.get(pair.getFirst()), pair.getSecond()));
							
 
			}

 

		} , this.getClass(), "RetLoop");
		
		return new Tri<>(ret, new ArrayList<>(unsure), new ArrayList<>(newEntities));
		
		}
		else {
		ConcurrentHashMap<InstanceId, List<ScoredEntity>> ret = new ConcurrentHashMap<>();
		ConcurrentHashMap<Integer, MutableBoolean> added = new ConcurrentHashMap<>();

		Set<Integer> secondRound = Collections.newSetFromMap(new ConcurrentHashMap<>());

		PersistentParallel6.runAndFinish(candidatesMap.entrySet(), (entry) -> {

			TreeSet<Pair<Integer, Double>> treeSet = entry.getValue();

			Set<Integer> secondRoundInner = treeSet.stream().map(Pair::getFirst).collect(Collectors.toSet());
			Integer addedInner = null;

			for (Pair<Integer, Double> pair : treeSet) {
				MutableBoolean isAdded = CHMHelper.getFromCHM(added, pair.getFirst(), () -> new MutableBoolean(false));
				boolean breakFor = false;
				synchronized (isAdded) {
					if (!isAdded.booleanValue()) {
						TreeSet<Pair<InstanceId, Double>> matched = matchedMap.get(pair.getFirst());

						if (matched.first().getFirst().equals(entry.getKey())) {
							List<ScoredEntity> list = CHMHelper.getFromCHM(ret, entry.getKey(), () -> new ArrayList<>());

							synchronized (list) {
								list.add(new ScoredEntity(entities.get(pair.getFirst()), pair.getSecond()));
							}
							breakFor = true;
							isAdded.setValue(true);
						}
					}

				}
				if (breakFor) {
					addedInner = pair.getFirst();
					break;
				}

			}

			secondRoundInner.remove(addedInner);

			secondRound.addAll(secondRoundInner);

		} , this.getClass(), "FirstRoundLoop");

		// secondRound.removeAll(added.keySet());
		System.out.println("FirstRound: " + added.size());
		System.out.println("SecondRound: " + secondRound.size());

		PersistentParallel6.runAndFinish(secondRound, (index) -> {
			boolean breakFor = false;

			TreeSet<Pair<InstanceId, Double>> matches = matchedMap.get(index);
			InstanceId addedInner = null;

			for (Pair<InstanceId, Double> pair : matches) {
				if (!ret.containsKey(pair.getFirst())) {
					List<ScoredEntity> list = CHMHelper.getFromCHM(ret, pair.getFirst(), () -> new ArrayList<>());
					synchronized (list) {
						if (list.isEmpty()) {
							list.add(new ScoredEntity(entities.get(index), pair.getSecond()));
							breakFor = true;
						}
					}
				}
				if (breakFor) {
					addedInner = pair.getFirst();
					break;
				}

			}

			if (addedInner == null) {
				Pair<InstanceId, Double> firstMatch = matches.first();
				List<ScoredEntity> list = CHMHelper.getFromCHM(ret, firstMatch.getFirst(), () -> new ArrayList<>());
				synchronized (list) {
						list.add(new ScoredEntity(entities.get(index), firstMatch.getSecond()));
				}
			}
		} , this.getClass(), "SecondRoundLoop");

		return new Tri<>(ret, new ArrayList<>(unsure), new ArrayList<>(newEntities));
		}
	}
	
	
	private Set<ClassId> getAllSubClases(ClassId classId) {

		Set<ClassId> classes = new HashSet<ClassId>();
		classes.add(classId);

		Set<ClassId> subClasses = ontology.getSubClasses(classId);
		if (subClasses != null) {
			for (ClassId subClassId : subClasses) {
				classes.addAll(getAllSubClases(subClassId));
			}
		}

		return classes;
	}

	private Set<ClassId> getAllParentClases(ClassId classId) {

		Set<ClassId> classes = new HashSet<ClassId>();
		classes.add(classId);

		Set<ClassId> parentClasses = ontology.getParentClasses(classId);

		if (parentClasses != null) {
			for (ClassId subClassId : parentClasses) {
				classes.addAll(getAllParentClases(subClassId));
			}
		}

		return classes;
	}

	public Map<InstanceId, Map<InstanceId, Double>> matchConflict(Map<InstanceId, Entity> entitiesRemainingInstanceIdWithEntity, Map<ClassId, Set<ClassId>> allParents) {
		ConcurrentHashMap<InstanceId, Map<InstanceId, Double>> ret = new ConcurrentHashMap<>(); 

		PersistentParallel6.runAndFinish( entitiesRemainingInstanceIdWithEntity.entrySet(), (entityEntry) -> {

			Entity entity = entityEntry.getValue();
			Set<ClassId> classIds = new HashSet<>();
			classIds.add(entity.getClassId());
//			classIds.addAll(getAllSubClases(entity.getClassId()));
			if(allParents != null) {
				classIds.addAll(allParents.get(entity.getClassId()));
				// classIds.addAll(getAllParentClases(entity.getClassId()));
			}
			
		
			
			Optional<Map<InstanceId, Double>> matches = labelMatcher.getMatches(entity.getLabels().keySet(),classIds ,
					MatchStrictness.NON_STRICT);
			
			if(matches.isPresent()) {
				
				
				for(Entry<InstanceId, Double> entry : matches.get().entrySet()) {
					if(entry.getValue() >= 1.0) {
						CHMHelper.getFromCHM(ret, entry.getKey(), () -> new HashMap<>()).put(entityEntry.getKey(), entry.getValue());
					}
				}
			}
		}, this.getClass());
		
		return ret;
	}

}
