package expansion.set.evaluation.pipelineresults;

import java.util.Map;
import java.util.Set;
import java.util.function.BinaryOperator;

import com.google.common.primitives.Ints;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.Accumulateable;
import expansion.set.evaluation.goldstandard.GoldStandardClusters;

public class GsStatistics implements Accumulateable<GsStatistics>{
	private final Integer tables ;
	private final Integer rows ;
	private final Integer clusters;

	private final Integer existingClusters;
	private final Integer existingClustersFacts ;
	

	private final Integer newClusters;
	private final Integer newClustersFacts;
	 
	
	
	public GsStatistics(GoldStandardClusters clusters, Map<QualifiedTripleId, QualifiedValueList> correctMap, Set<InstanceId> newEntities) {
		
		this.tables = Ints.checkedCast(clusters.getIndex().keySet().stream().map((e) -> e.getTableId()).distinct().count());
		this.rows = clusters.getIndex().size();
		this.clusters = clusters.getTestingSetMappingClusters().size();
		
		this.existingClusters = Ints.checkedCast(clusters.getTestingSetMappingClusters().keySet().stream().filter((e) -> !newEntities.contains(e)).count());
		this.existingClustersFacts = Ints.checkedCast(correctMap.keySet().stream().filter((e) -> !newEntities.contains(e.getTripleId().getInstanceId())).count());
		
		this.newClusters = Ints.checkedCast(clusters.getTestingSetMappingClusters().keySet().stream().filter((e) -> newEntities.contains(e)).count());
		this.newClustersFacts = Ints.checkedCast(correctMap.keySet().stream().filter((e) -> newEntities.contains(e.getTripleId().getInstanceId())).count());
		
	}
	
	private GsStatistics(Integer tables, Integer rows, Integer clusters, Integer existingClusters, Integer existingClustersFacts, Integer newClusters,
			Integer newClustersFacts) {
		super();
		this.tables = tables;
		this.rows = rows;
		this.clusters = clusters;
		this.existingClusters = existingClusters;
		this.existingClustersFacts = existingClustersFacts;
		this.newClusters = newClusters;
		this.newClustersFacts = newClustersFacts;
	}
	
	public GsStatistics() {
		super();
		this.tables = 0;
		this.rows = 0;
		this.clusters = 0;
		this.existingClusters = 0;
		this.existingClustersFacts = 0;
		this.newClusters = 0;
		this.newClustersFacts = 0;
	}
	
	public Integer getTables() {
		return tables;
	}

	public Integer getRows() {
		return rows;
	}
	public Integer getClusters() {
		return clusters;
	}
	public Integer getExistingClusters() {
		return existingClusters;
	}
	public Integer getExistingClustersFacts() {
		return existingClustersFacts;
	}
	public Integer getNewClusters() {
		return newClusters;
	}
	public Integer getNewClustersFacts() {
		return newClustersFacts;
	}
	
	
	public static GsStatistics accumulator(GsStatistics a, GsStatistics b) {
		return new GsStatistics(a.tables + b.tables, a.rows + b.rows, a.clusters + b.clusters, a.existingClusters + b.existingClusters, a.existingClustersFacts + b.existingClustersFacts, a.newClusters + b.newClusters, a.newClustersFacts + b.newClustersFacts);
	}

	@Override
	public BinaryOperator<GsStatistics> getAccumulator() {
		return GsStatistics::accumulator;
	}

	@Override
	public GsStatistics accumulate(GsStatistics other) {
		return GsStatistics.accumulator(this, other);
	}

	@Override
	public Class<GsStatistics> getAcumClass() {
		return GsStatistics.class;
	}
	
	
	
}
