package expansion.set.evaluation.pipelineresults;

import java.util.function.BinaryOperator;

import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.Accumulateable;
import expansion.set.evaluation.entities.EntitiesPerformance;
import expansion.set.evaluation.fusion.FusionSetPerformance;
import expansion.set.evaluation.goldstandard.oktie.OktiePerformance;
import expansion.set.evaluation.matching.NewDetectionPerformance;
import expansion.set.evaluation.rankedEvaluation.RankedEvaluationPerformance;
import expansion.set.evaluation.sim.SimPerformance;

public class PipelinePerformance  implements Accumulateable<PipelinePerformance > {
	

	private final GsStatistics gsStatistics;
	

	private final SimPerformance simAccuracy;

	private final OktiePerformance clusteringAccuracy; 
	
	private final OktiePerformance nonClusteringClusteringPerformance;

	private final NewDetectionPerformance newDetectionPerformance;
	
	private final EntitiesPerformance entitiesPerformance;
	
	private final FusionSetPerformance fusionSetPerformance;
	private final RankedEvaluationPerformance rankedEvaluationPerformance;

	 public PipelinePerformance( ) {
		 simAccuracy = new SimPerformance();
		 clusteringAccuracy = new OktiePerformance();
		nonClusteringClusteringPerformance = new OktiePerformance();
		 gsStatistics = new GsStatistics();
		 newDetectionPerformance = new NewDetectionPerformance();
		 entitiesPerformance = new EntitiesPerformance();
		 fusionSetPerformance = new FusionSetPerformance();
		rankedEvaluationPerformance = new RankedEvaluationPerformance();
		 
	}
	
	 
	 
	public PipelinePerformance(GsStatistics gsStatistics, SimPerformance simAccuracy, OktiePerformance nonClusteringClusteringPerformance,
			OktiePerformance clusteringAccuracy,
			NewDetectionPerformance newDetectionPerformance, EntitiesPerformance entitiesPerformance, RankedEvaluationPerformance rankedEvaluationPerformance,
			FusionSetPerformance fusionSetPerformance) {
		super();
		this.gsStatistics = gsStatistics;
		this.simAccuracy = simAccuracy;
		this.nonClusteringClusteringPerformance = nonClusteringClusteringPerformance;
		this.clusteringAccuracy = clusteringAccuracy;
		this.newDetectionPerformance = newDetectionPerformance;
		this.entitiesPerformance = entitiesPerformance;
		this.fusionSetPerformance = fusionSetPerformance;
		this.rankedEvaluationPerformance = rankedEvaluationPerformance;
	}


	
	

	public GsStatistics getGsStatistics() {
		return gsStatistics;
	}



	public SimPerformance getSimAccuracy() {
		return simAccuracy;
	}



	public OktiePerformance getClusteringAccuracy() {
		return clusteringAccuracy;
	}



	public NewDetectionPerformance getNewDetectionPerformance() {
		return newDetectionPerformance;
	}



	public FusionSetPerformance getFusionSetPerformance() {
		return fusionSetPerformance;
	}



	public EntitiesPerformance getEntitiesPerformance() {
		return entitiesPerformance;
	}



	public static   PipelinePerformance  accumulator(PipelinePerformance  a, PipelinePerformance  b) {
		 
		return new PipelinePerformance(a.getGsStatistics().accumulate(b.getGsStatistics()),
				a.getSimAccuracy().accumulate(b.getSimAccuracy()),
				a.getNonClusteringClusteringPerformance().accumulate(b.getNonClusteringClusteringPerformance()),
				a.getClusteringAccuracy().accumulate(b.getClusteringAccuracy()),
				a.getNewDetectionPerformance().accumulate(b.getNewDetectionPerformance()),
				a.getEntitiesPerformance().accumulate(b.getEntitiesPerformance()),
				a.getRankedEvaluationPerformance().accumulate(b.getRankedEvaluationPerformance()),
				a.getFusionSetPerformance().accumulate(b.getFusionSetPerformance()));
		
		
	}
	@Override
	public BinaryOperator<PipelinePerformance > getAccumulator() {
		return PipelinePerformance::accumulator;
	}

	@Override
	public PipelinePerformance accumulate(PipelinePerformance other) { 
			return PipelinePerformance.accumulator(this, other);
	 
		
	} 
	@Override
	public Class<PipelinePerformance > getAcumClass() {
		return PipelinePerformance.class;
	}

	public RankedEvaluationPerformance getRankedEvaluationPerformance() {
		return rankedEvaluationPerformance;
	}

	public OktiePerformance getNonClusteringClusteringPerformance() {
		return nonClusteringClusteringPerformance;
	}
 

}
