package expansion.set.evaluation.rankedEvaluation;

import java.util.function.BinaryOperator;

import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.Accumulateable;

public class RankedEvaluationPerformance implements Accumulateable<RankedEvaluationPerformance> {

	
	private final double averagePrecisionSum;
	private final double pAt5;
	private final double pAt20;
	private final double pAt50;
	private final double pAt100;
	private final int count;
	

	public RankedEvaluationPerformance() {
		super();
		this.averagePrecisionSum = 0.0;
		this.pAt5 = 0.0;
		this.pAt20 = 0.0;
		this.pAt50 = 0.0;
		this.pAt100 = 0.0;
		this.count = 0;
	}
	
	public RankedEvaluationPerformance(double averagePrecisionSum, double pAt5, double pAt20, double pAt50, double pAt100) {
		super();
		this.averagePrecisionSum = averagePrecisionSum;
		this.pAt5 = pAt5;
		this.pAt20 = pAt20;
		this.pAt50 = pAt50;
		this.pAt100 = pAt100;
		this.count = 1;
	}

	private RankedEvaluationPerformance(double averagePrecisionSum, double pAt5, double pAt20, double pAt50, double pAt100, int count) {
		super();
		this.averagePrecisionSum = averagePrecisionSum;
		this.pAt5 = pAt5;
		this.pAt20 = pAt20;
		this.pAt50 = pAt50;
		this.pAt100 = pAt100;
		this.count = count;
	}

	public static RankedEvaluationPerformance accumulator(RankedEvaluationPerformance a, RankedEvaluationPerformance b) {
		return new RankedEvaluationPerformance(a.averagePrecisionSum + b.averagePrecisionSum, a.pAt5 + b.pAt5, a.pAt20 + b.pAt20, a.pAt50 + b.pAt50,
				a.pAt100 + b.pAt100, a.count + b.count);
	}
	@Override
	public BinaryOperator<RankedEvaluationPerformance> getAccumulator() {
		return RankedEvaluationPerformance::accumulate;
	}

	@Override
	public RankedEvaluationPerformance accumulate(RankedEvaluationPerformance other) {
		return accumulator(this, other);
	}

	@Override
	public Class<RankedEvaluationPerformance> getAcumClass() {
		return RankedEvaluationPerformance.class;
	}

	public double getAveragePrecisionSum() {
		return averagePrecisionSum / count;
	}

	public double getpAt5() {
		return pAt5 / count;
	}

	public double getpAt20() {
		return pAt20 / count;
	}

	public double getpAt50() {
		return pAt50 / count;
	}

	public double getpAt100() {
		return pAt100 / count;
	}

}
