package expansion.set.evaluation.rankedEvaluation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.containers.Pair;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.evaluation.goldstandard.GoldStandardClusters;
import expansion.set.evaluation.matching.ScoredEntity;

public class RankingEvalutator {

	@Inject
	public RankingEvalutator() {

	}

	public RankedEvaluationPerformance chooseClustersForEntitiesFound(GoldStandardClusters goldStandardClusters, List<ScoredEntity> newDetected,
			Set<InstanceId> newEntities, Map<RowClusterUUID, Optional<InstanceId>> remapped) {

		List<Boolean> rankedList;

		List<Pair<ScoredEntity, Boolean>> correct = new ArrayList<>();

		// Collection the new ones by instance
		Map<InstanceId, List<ScoredEntity>> newMap = new HashMap<>();

		for (ScoredEntity e : newDetected) {
			Optional<InstanceId> mapping = remapped.get(e.getEntity().getRowClusterUUID());
			if (mapping.isPresent() && newEntities.contains(mapping.get())) {
				CHMHelper.getFromMap(newMap, mapping.get(), () -> new ArrayList<>()).add(e);
			}
			else {
				correct.add(new Pair<>(e, false));
			}
		}

		// Finding the one with max intersection
		for (Entry<InstanceId, List<ScoredEntity>> entry : newMap.entrySet()) {
			Set<TableRowId> cluster = goldStandardClusters.getCluster(entry.getKey());
			int maxIntersection = 0;
			ScoredEntity maxEntity = null;
			for (ScoredEntity e : entry.getValue()) {
				int intersection = 0;
				for (TableRowId rowId : e.getEntity().getRowCluster().getRows()) {
					if (cluster.contains(rowId)) {
						intersection++;
					}
				}
				if (intersection > maxIntersection) {
					maxIntersection = intersection;
					maxEntity = e;
				}
			}
			if (maxIntersection / ((double) cluster.size()) > 0.5) {
				correct.add(new Pair<>(maxEntity, true));

			}

			for (ScoredEntity se : entry.getValue()) {
				if (se != maxEntity) {
					correct.add(new Pair<>(maxEntity, false));

				}
			}
		}

		rankedList = correct.stream().sorted((e1, e2) -> {
			int compare = Double.compare(e1.getFirst().getScore(), e2.getFirst().getScore());
			if (compare == 0) {
				return e1.getFirst().getEntity().getRowClusterUUID().toString().compareTo(e2.getFirst().getEntity().getRowClusterUUID().toString());
			}
			return compare;
		}).map((e) -> e.getSecond()).collect(Collectors.toList());

		return new RankedEvaluationPerformance(getAP(rankedList), getPrecision(rankedList, 5), getPrecision(rankedList, 20), getPrecision(rankedList, 50),
				getPrecision(rankedList, 100));
	}

	double getPrecision(List<Boolean> rankedList, int limit) {
		List<Boolean> shortenedList = rankedList.stream().limit(limit).collect(Collectors.toList());

		double correct = shortenedList.stream().filter((e) -> e).count();
		double total = shortenedList.size();

		return correct / total;
	}

	double getAP(List<Boolean> rankedList) {
		double sum = 0.0;
		int i = 1;
		double correct = 0;

		for (Boolean b : rankedList) {
			if (b) {
				correct++;
				sum += correct / i;
			}
			i++;
		}

		return sum / correct;
	}
}
