package expansion.set.evaluation.sim;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Random;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.stream.Collectors;

import javax.inject.Inject;

import com.beust.jcommander.internal.Lists;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.PerformanceMacroPR;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.DoubleIteraterable;
import de.mannheim.uni.expansion.utils.DoubleIteraterable.DoubleEntry;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumAverage;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumDouble;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumInteger;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentAccumulator;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.clustering.model.Block;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.evaluation.goldstandard.GoldStandardClusters;
import expansion.set.evaluation.goldstandard.oktie.OktieClusterEvaluator.WeightGetter;
import expansion.set.strategies.rowclusterers.unary.ClusterCombiner;

public class SimEvaluator {

	@Inject
	public SimEvaluator() {

	}

	
	public <E> ClusterList getClusterList(ClassId classId, RowClusterSimilarities sim, RowClusteringConfig config, ClustererInput input,
			GoldStandardClusters gs, OptimizableClass optimizingClass, boolean disableBlocking) {

		List<RowCluster> clusterList = new ArrayList<>();

		final double threshold;

		Optional<Double> thresholdOpt = config.getWeightsRepository().get(optimizingClass).getNormalizingThreshold();
		if (thresholdOpt.isPresent()) {
			threshold = thresholdOpt.get();
		} else {
			threshold = 0.0;
		}
		List<TableRowId> rowIds = new ArrayList<>(input.getCandidates());

		ConcurrentLinkedQueue<Set<TableRowId>> positivePairs = new ConcurrentLinkedQueue<>();

		long resultingSize = DoubleIteraterable.getResultingSize(rowIds.size());

		PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(ListIterable.iterable(rowIds)), (doubleEntry) -> {

			boolean compare = disableBlocking;
			if (!disableBlocking) {
				List<Pair<Set<Block>, Integer>> first = sim.getSplitBlocksWithScoresAndMinWithMin(doubleEntry.getFirst().getValue());
				List<Pair<Set<Block>, Integer>> second = sim.getSplitBlocksWithScoresAndMinWithMin(doubleEntry.getSecond().getValue());

				for (int i = 0; i < first.size(); i++) {

					Set<Block> firstSet = first.get(i).getFirst();
					Set<Block> secondSet = second.get(i).getFirst();
					if (!Collections.disjoint(firstSet, secondSet)) {
						compare = true;
						break;
					}

				}

			}
			if (compare) {
				double score = sim.get(rowIds, input.getClassId(), new ClusterEntryKey(doubleEntry.getFirst().getIndex(), doubleEntry.getSecond().getIndex()));

				if (score > threshold) {
					Set<TableRowId> pairs = new HashSet<>();
					pairs.add(doubleEntry.getFirst().getValue());
					pairs.add(doubleEntry.getSecond().getValue());
					positivePairs.add(pairs);
				}
			}

		}, resultingSize, this.getClass());

		ClusterCombiner<TableRowId> cc = new ClusterCombiner<>();
		cc.add(positivePairs);
		clusterList = cc.get().stream().map((e) -> new RowCluster(classId, new ArrayList<>(e), sim)).collect(Collectors.toList());
		Set<TableRowId> addedRowIds = new HashSet<>();
		for (RowCluster rc : clusterList) {
			boolean changed = addedRowIds.addAll(rc.getRows());
			if (!changed) {
				int sdf = 13;
			}
		}

		for (TableRowId rowId : input.getCandidates()) {

			if (!addedRowIds.contains(rowId)) {
				List<TableRowId> list = new ArrayList<>();
				list.add(rowId);
				clusterList.add(new RowCluster(classId, list, sim));
			}

		}
		return new ClusterList(clusterList);

	}

	public <E> SimPerformance evaluateSimNew( RowClusterSimilarities sim,
			RowClusteringConfig config, ClustererInput input, GoldStandardClusters gs, OptimizableClass optimizingClass) {
		
		final double threshold;
		final double defaultExpected;
		final double normalization;
		Optional<Double> thresholdOpt = config.getWeightsRepository().get(optimizingClass).getNormalizingThreshold();
		 if(thresholdOpt.isPresent()) {
			 threshold = thresholdOpt.get();
			normalization = 1.0;
			defaultExpected = 0.0;
		 }
		else {
			threshold = 0.0;
			normalization = 2.0;
			defaultExpected = -1.0;
		}
		
		
		ConcurrentAccumulator<AccumInteger> accumMaxCorrect = new ConcurrentAccumulator<>(AccumInteger.empty());

		List<TableRowId> rowIds = new ArrayList<>(input.getCandidates());
		
		ConcurrentAccumulator<AccumInteger> accumCorrect = new ConcurrentAccumulator<>(AccumInteger.empty());
		ConcurrentAccumulator<AccumInteger> accumWrong = new ConcurrentAccumulator<>(AccumInteger.empty());

		ConcurrentAccumulator<AccumAverage> meanSquaredError = new ConcurrentAccumulator<>(AccumAverage.empty());
		long resultingSize = DoubleIteraterable.getResultingSize(rowIds.size());

		
		PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(ListIterable.iterable(rowIds)), (doubleEntry) -> {
			double score = sim.get(rowIds, input.getClassId(), new ClusterEntryKey(doubleEntry.getFirst().getIndex(), doubleEntry.getSecond().getIndex()));
			boolean positive = false;
			double expected = defaultExpected;
			if (gs.getIndex().get(doubleEntry.getFirst().getValue()).equals(gs.getIndex().get(doubleEntry.getSecond().getValue()))) {
				positive = true;
				accumMaxCorrect.add(new AccumInteger(1));
				expected = 1.0;


			}
			meanSquaredError.add(new AccumAverage(Math.pow(score - expected, 2.0)));

			if (score > threshold) {
				if (positive) {

					accumCorrect.add(new AccumInteger(1));

				} else {

					accumWrong.add(new AccumInteger(1));
				}
			}

		}, resultingSize, this.getClass());

		int maxCorrect = accumMaxCorrect.get().get();
		int correct = accumCorrect.get().get();
		int wrong = accumWrong.get().get();

		PerformanceMacroPR pr = new PerformanceMacroPR(correct, wrong, maxCorrect);

		double rmse = Math.sqrt(meanSquaredError.get().get()) / normalization;
		if (Double.isNaN(rmse))
			rmse = 0.0;
		return new SimPerformance(rmse, pr);
		
	}
	public <E> SimPerformance evaluateSim(WeightsRepository weightsRepository, OptimizableClass optimizableClass, RowClusterSimilarities sim,
			RowClusteringConfig config, ClustererInput input, GoldStandardClusters gs,
			Random random) {


		ConcurrentAccumulator<AccumInteger> accumMaxCorrect = new ConcurrentAccumulator<>(AccumInteger.empty());


		ConcurrentAccumulator<AccumInteger> accumCorrect = new ConcurrentAccumulator<>(AccumInteger.empty());
		ConcurrentAccumulator<AccumInteger> accumWrong = new ConcurrentAccumulator<>(AccumInteger.empty());

		ConcurrentAccumulator<AccumAverage> precisionAcc = new ConcurrentAccumulator<>(AccumAverage.empty());

		ConcurrentAccumulator<AccumAverage> recallAcc = new ConcurrentAccumulator<>(AccumAverage.empty());


		List<TableRowId> rowIds = new ArrayList<>(input.getCandidates());
		long resultingSize = DoubleIteraterable.getResultingSize(rowIds.size());
		Map<TableRowId, Integer> rowIdsIndex = ListIterable.iterable(rowIds).stream().collect(Collectors.toMap(ListEntry::getValue, ListEntry::getIndex));



		double threshold = weightsRepository.getOrDefault(optimizableClass.getWeight("SIMILARITY_THRESHOLD", 0.50));

		PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(ListIterable.iterable(rowIds)), (doubleEntry) -> {
			double score = sim.get(rowIds, input.getClassId(), new ClusterEntryKey(doubleEntry.getFirst().getIndex(), doubleEntry.getSecond().getIndex()));
			boolean positive = false;
			double expected = 0.0;
			if (gs.getIndex().get(doubleEntry.getFirst().getValue()).equals(gs.getIndex().get(doubleEntry.getSecond().getValue()))) {
				positive = true;
				accumMaxCorrect.add(new AccumInteger(1));
				expected = 1.0;

				
			}
			double recall;
			double precision;

			if (positive) {
				precisionAcc.add(new AccumAverage(score));
				recallAcc.add(new AccumAverage(score));
			} else {
				precisionAcc.add(new AccumAverage(1.0 - score));

			}

			if (score > threshold) {
				if (positive) {

					accumCorrect.add(new AccumInteger(1));

				} else {

					accumWrong.add(new AccumInteger(1));
				}
			}

		}, resultingSize, this.getClass());

		int maxCorrect = accumMaxCorrect.get().get();
		int correct = accumCorrect.get().get();
		int wrong = accumWrong.get().get();

		
		
		PerformanceMacroPR pr = new PerformanceMacroPR(correct, wrong, maxCorrect);
		
		boolean testing = true;

		// ConcurrentAccumulator<AccumAverage> accumulator = new
		// ConcurrentAccumulator<>(AccumAverage.empty());

		List<Pair<TableRowId, TableRowId>> correctList = new ArrayList<>();

		int correctPairsSize = 0;

		for (Entry<InstanceId, Set<TableRowId>> entry : gs.getTestingSetMappingClusters().entrySet()) {

			for (DoubleEntry<TableRowId> doubleEntry : DoubleIteraterable.iterable(entry.getValue())) {

				double score = sim.get(rowIds, input.getClassId(),
						new ClusterEntryKey(rowIdsIndex.get(doubleEntry.getFirst()), rowIdsIndex.get(doubleEntry.getSecond())));
				correctPairsSize++;

				double expected = 1.0;
				double performance = Math.abs(score - expected);

				if (TableRowId.compare(doubleEntry.getFirst(), doubleEntry.getSecond()) == -1) {
					correctList.add(new Pair<>(doubleEntry.getFirst(), doubleEntry.getSecond()));
				} else {
					correctList.add(new Pair<>(doubleEntry.getSecond(), doubleEntry.getFirst()));
				}

				if (performance < 0.0) {
					int sdf = 123;
				}
				if (!testing) {
					// accumulator.add(new AccumAverage(Math.pow(performance,
					// 2.0)));
				}
			}
		}

		Collections.sort(correctList, (e1, e2) -> {
			int compare = TableRowId.compare(e1.getFirst(), e2.getFirst());
			if (compare == 0) {
				compare = TableRowId.compare(e1.getSecond(), e2.getSecond());
			}

			return compare;
		});

		Collections.shuffle(correctList, random);

		Map<Integer, List<Pair<TableRowId, TableRowId>>> sets = new HashMap<>();

		int skip = 0;
		for (Pair<TableRowId, TableRowId> pair : correctList) {
			for (int i = 0; i < 1; i++) {

				CHMHelper.getFromMap(sets, i, () -> new ArrayList<>()).add(pair);

			}

			skip++;
			if (skip == 1) {
				skip = 0;
			}
		}

		List<InstanceId> clusterList = new ArrayList<>(gs.getTestingSetMappingClusters().keySet());
		Collections.sort(clusterList, (e1, e2) -> {
			return e1.getInstanceId().compareTo(e2.getInstanceId());
		});
		Collections.shuffle(clusterList, random);

		double sum = 0.0;
		int count = 0;
		for (List<Pair<TableRowId, TableRowId>> entry : sets.values()) {

			ConcurrentAccumulator<AccumAverage> accumulator = new ConcurrentAccumulator<>(AccumAverage.empty());

			PersistentParallel6.runAndFinish(entry, (pair) -> {
				double score = sim.get(rowIds, input.getClassId(), new ClusterEntryKey(rowIdsIndex.get(pair.getFirst()), rowIdsIndex.get(pair.getSecond())));

				double expected = 1.0;
				double performance = Math.abs(score - expected);

				accumulator.add(new AccumAverage(Math.pow(performance, 2.0)));

			}, this.getClass(), true);

			Set<Pair<TableRowId, TableRowId>> pairSet = new HashSet<>();

			while (pairSet.size() < entry.size()) {
				addToSet(pairSet, gs, entry.size(), clusterList, random);
			}

			PersistentParallel6.runAndFinish(pairSet, (pair) -> {
				double score = sim.get(rowIds, input.getClassId(), new ClusterEntryKey(rowIdsIndex.get(pair.getFirst()), rowIdsIndex.get(pair.getSecond())));

				double expected = 0.0;
				double performance = Math.abs(score - expected);

				accumulator.add(new AccumAverage(Math.pow(performance, 2.0)));

			}, this.getClass(), true);

			double rmse = Math.sqrt(accumulator.get().get());
			sum += rmse;
			count++;
		}

		double precision = precisionAcc.get().get();
		double recall = recallAcc.get().get();
		double rmse = sum / count;
		return new SimPerformance(rmse, pr);
	}

	private void addToSet(Set<Pair<TableRowId, TableRowId>> pairSet, GoldStandardClusters gs, int maxCorrect, List<InstanceId> clusterList, Random random) {
		for (DoubleEntry<InstanceId> doubleEntry : DoubleIteraterable.iterable(clusterList)) {
			if (pairSet.size() < maxCorrect) {

				List<TableRowId> cluster1 = new ArrayList<>(gs.getCluster(doubleEntry.getFirst()));
				Collections.sort(cluster1, TableRowId::compare);
				Collections.shuffle(cluster1, random);

				List<TableRowId> cluster2 = new ArrayList<>(gs.getCluster(doubleEntry.getSecond()));
				Collections.sort(cluster2, TableRowId::compare);
				Collections.shuffle(cluster2, random);

				if (TableRowId.compare(cluster1.get(0), cluster2.get(0)) == -1) {
					pairSet.add(new Pair<>(cluster1.get(0), cluster2.get(0)));
				} else {
					pairSet.add(new Pair<>(cluster2.get(0), cluster1.get(0)));
				}

			}
		}

	}

	private static enum PairType {
		CORRECT_NORMAL, CORRECT_CONFLICT, CORRECT_NEW, CORRECT_NEW_CONFLICT, WRONG_NORMAL, WRONG_NEW, WRONG_NEW_NORMAL, WRONG_CONFLICT;
	}

	public List<Tri<TableRowIdPair, Boolean, Double>> getUpsampledRMSESet(ClustererInput clustererInput, GoldStandardClusters goldStandard,
			Set<InstanceId> newEntities, Map<InstanceId, Integer> conflictingSets, Random random) {

		int countNormal = 0;
		int countNew = 0;
		int countConflict = 0;

		Map<TableRowIdPair, PairType> pairs = new HashMap<>();
		for (Entry<InstanceId, Set<TableRowId>> entry : goldStandard.getTestingSetMappingClusters().entrySet()) {
			PairType type = PairType.CORRECT_NORMAL;
			if (conflictingSets.containsKey(entry.getKey())) {
				if (newEntities.contains(entry.getKey())) {
					type = PairType.CORRECT_NEW_CONFLICT;
				} else {
					type = PairType.CORRECT_CONFLICT;
				}
			} else if (newEntities.contains(entry.getKey())) {
				type = PairType.CORRECT_NEW;
			}

			int size = 0;
			for (DoubleEntry<TableRowId> dEntry : DoubleIteraterable.iterable(entry.getValue())) {
				pairs.put(new TableRowIdPair(dEntry.getFirst(), dEntry.getSecond()), type);
				size++;
			}

			if (conflictingSets.containsKey(entry.getKey())) {
				countConflict += size;
			}
			if (newEntities.contains(entry.getKey())) {
				countNew += size;
			}

			countNormal += size;
		}

		int wrongNormal = 0;
		int wrongConflict = 0;

		for (DoubleEntry<Entry<InstanceId, Set<TableRowId>>> doubleEntry : DoubleIteraterable
				.iterable(goldStandard.getTestingSetMappingClusters().entrySet())) {

			PairType pairType = PairType.WRONG_NORMAL;
			boolean inConflict = false;
			if (conflictingSets.get(doubleEntry.getFirst().getKey()) != null
					&& conflictingSets.get(doubleEntry.getFirst().getKey()).equals(conflictingSets.get(doubleEntry.getSecond().getKey()))) {
				inConflict = true;
				pairType = PairType.WRONG_CONFLICT;
			}

			for (TableRowId first : doubleEntry.getFirst().getValue()) {
				for (TableRowId second : doubleEntry.getSecond().getValue()) {
					pairs.put(new TableRowIdPair(first, second), pairType);
					wrongNormal++;
					if (inConflict) {
						wrongConflict++;
					}
				}
			}

		}

		double maxClass = (double) Lists.newArrayList(countNormal, countConflict, countNew, wrongNormal, wrongConflict).stream().mapToInt(Integer::intValue)
				.max().getAsInt();

		double mNormal = maxClass / countNormal;
		double mConflict = maxClass / countConflict;
		double mNew = maxClass / countNew;
		double mWrongNormal = maxClass / wrongNormal;
		double mWrongConflict = maxClass / wrongConflict;

		List<Tri<TableRowIdPair, Boolean, Double>> ret = new ArrayList<>();

		for (Entry<TableRowIdPair, PairType> pairEntry : pairs.entrySet()) {
			boolean correct = true;
			double score = 0.0;
			switch (pairEntry.getValue()) {
			case CORRECT_CONFLICT:
				score = mNormal + mConflict;
				break;
			case CORRECT_NEW:
				score = mNormal + mNew;
				break;
			case CORRECT_NEW_CONFLICT:
				score = mNormal + mNew + mConflict;
				break;
			case CORRECT_NORMAL:
				score = mNormal;
				break;
			case WRONG_CONFLICT:
				correct = false;
				score = mWrongNormal + mWrongConflict;
				break;
			case WRONG_NORMAL:
				correct = false;
				score = mWrongNormal;
				break;

			}
			ret.add(new Tri<>(pairEntry.getKey(), correct, score));
		}

		return ret;
	}

	public Map<TableRowIdPair, Pair<Boolean, Double>> getUpsampledRowPairSet(ClustererInput input, GoldStandardClusters gs,
			Set<InstanceId> newEntities, Map<InstanceId, Integer> conflictingSets ) {

		ConcurrentAccumulator<AccumInteger> correctExisting  = new ConcurrentAccumulator<AccumInteger>(new AccumInteger(0));
		ConcurrentAccumulator<AccumInteger> correctNew  = new ConcurrentAccumulator<AccumInteger>(new AccumInteger(0));
		
		
		ConcurrentAccumulator<AccumInteger> wrongExisting  = new ConcurrentAccumulator<AccumInteger>(new AccumInteger(0));
		ConcurrentAccumulator<AccumInteger> wrongNew  = new ConcurrentAccumulator<AccumInteger>(new AccumInteger(0));
		ConcurrentAccumulator<AccumInteger> wrongExistingNew  = new ConcurrentAccumulator<AccumInteger>(new AccumInteger(0));
		
		ConcurrentAccumulator<AccumInteger> wrongWithConflict = new ConcurrentAccumulator<AccumInteger>(new AccumInteger(0));
		
		List<TableRowId> rowIds = new ArrayList<>(input.getCandidates());
		
		long resultingSize = DoubleIteraterable.getResultingSize(rowIds.size());
		
		Map<TableRowIdPair, Pair<Boolean, PairType>> map = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(ListIterable.iterable(rowIds)), (doubleEntry) -> {
 
			InstanceId mapping1 = gs.getMapping(doubleEntry.getFirst().getValue());
			InstanceId mapping2 = gs.getMapping(doubleEntry.getSecond().getValue());

			TableRowIdPair key = new TableRowIdPair(doubleEntry.getFirst().getValue(), doubleEntry.getSecond().getValue());
			boolean positive = false;
			boolean inConflict = false;
			if (gs.getIndex().get(doubleEntry.getFirst().getValue()).equals(gs.getIndex().get(doubleEntry.getSecond().getValue()))) {
				positive = true; 

			}
			else if (conflictingSets.containsKey(mapping1) && conflictingSets.containsKey(mapping2)
					&& conflictingSets.get(mapping1).equals(conflictingSets.get(mapping2))) {
				inConflict = true;
			}
			
			if (false && inConflict) {
				wrongWithConflict.add(new AccumInteger(1));
				map.put(key, new Pair<>(positive, PairType.WRONG_CONFLICT));
			}
			else {

				if (newEntities.contains(mapping1) && newEntities.contains(mapping2)) {
					if (positive) {
						correctNew.add(new AccumInteger(1));
						map.put(key, new Pair<>(positive, PairType.CORRECT_NEW));
					} else {
						wrongNew.add(new AccumInteger(1));
						map.put(key, new Pair<>(positive, PairType.WRONG_NEW));
					}
				}
				if (newEntities.contains(mapping1) || newEntities.contains(mapping2)) {
					wrongExistingNew.add(new AccumInteger(1));
					map.put(key, new Pair<>(positive, PairType.WRONG_NEW_NORMAL));
				} else {
					if (positive) {
						correctExisting.add(new AccumInteger(1));
						map.put(key, new Pair<>(positive, PairType.CORRECT_NORMAL));
					} else {
						wrongExisting.add(new AccumInteger(1));
						map.put(key, new Pair<>(positive, PairType.WRONG_NORMAL));
					}
				}
			}


		},resultingSize,this.getClass());


		int correctExistingInt = correctExisting.get().get();
		int correctNewInt = correctNew.get().get();
		int wrongExistingInt = wrongExisting.get().get();
		int wrongNewInt = wrongNew.get().get();
		int wrongExistingNewInt = wrongExistingNew.get().get();
		int wrongWithConflictInt = wrongWithConflict.get().get();

		final double upsampleCorrect;
		final double upsampleWrong;
		final double upsampleWrongWithConflict;

		final double upSampleCorrectExisting;
		final double upSampleCorrectNew;

		final double upSampleWrongExisting;
		final double upSampleWrongNew;
		final double upSampleWrongExistingNew;

		if (false) {

		double maxCorrectClass = (double) Lists.newArrayList(correctExistingInt, correctNewInt).stream().mapToInt(Integer::intValue).max().getAsInt();

			upSampleCorrectExisting = maxCorrectClass / correctExistingInt;
			upSampleCorrectNew = maxCorrectClass / correctNewInt;



		double maxWrongClass = (double) Lists.newArrayList(wrongExistingInt, wrongNewInt, wrongExistingNewInt).stream().mapToInt(Integer::intValue).max()
				.getAsInt();

			upSampleWrongExisting = maxWrongClass / wrongExistingInt;
			upSampleWrongNew = maxWrongClass / wrongNewInt;
			upSampleWrongExistingNew = maxWrongClass / wrongExistingNewInt;

		double totalCorrect = correctExistingInt * upSampleCorrectExisting + correctNewInt * upSampleCorrectNew;
		double totalWrong = wrongExistingInt * upSampleWrongExisting + wrongNewInt * upSampleWrongNew + wrongExistingNewInt * upSampleWrongExistingNew;

		double maxClass = Lists.newArrayList(totalCorrect, totalWrong).stream().mapToDouble(Double::doubleValue).max().getAsDouble();

			upsampleCorrect = maxClass / totalCorrect;
			upsampleWrong = maxClass / totalWrong;

		} else {
			double maxClassAll = (double) Lists
					.newArrayList(correctExistingInt, correctNewInt, wrongExistingInt, wrongNewInt, wrongExistingNewInt, wrongWithConflictInt).stream()
					.mapToInt(Integer::intValue).max().getAsInt();

			upsampleCorrect = 1.0;
			upsampleWrong = 1.0;
			upsampleWrongWithConflict = maxClassAll / wrongWithConflictInt;

			upSampleCorrectExisting = maxClassAll / correctExistingInt;
			upSampleCorrectNew = maxClassAll / correctNewInt;

			upSampleWrongExisting = maxClassAll / wrongExistingInt;
			upSampleWrongNew = maxClassAll / wrongNewInt;
			upSampleWrongExistingNew = maxClassAll / wrongExistingNewInt;
		}

		Map<TableRowIdPair, Pair<Boolean, Double>> pairs = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(map.entrySet(), (e) -> {

			switch (e.getValue().getSecond()) {

			case CORRECT_NEW:
				pairs.put(e.getKey(), new Pair<>(e.getValue().getFirst(), upSampleCorrectNew * upsampleCorrect));
				break;

			case CORRECT_NORMAL:
				pairs.put(e.getKey(), new Pair<>(e.getValue().getFirst(), upSampleCorrectExisting * upsampleCorrect));
				break;

			case WRONG_NEW:
				pairs.put(e.getKey(), new Pair<>(e.getValue().getFirst(), upSampleWrongNew * upsampleWrong));
				break;
			case WRONG_NEW_NORMAL:
				pairs.put(e.getKey(), new Pair<>(e.getValue().getFirst(), upSampleWrongExistingNew * upsampleWrong));
				break;
			case WRONG_NORMAL:
				pairs.put(e.getKey(), new Pair<>(e.getValue().getFirst(), upSampleWrongExisting * upsampleWrong));
				break;
			case WRONG_CONFLICT:
				pairs.put(e.getKey(), new Pair<>(e.getValue().getFirst(), upsampleWrongWithConflict));

				break;
			default:
				throw new IllegalStateException();

			}
		}, this.getClass());

		return pairs;
	}

	public Map<InstanceId, Double> getUpsampledClusterSet(ClustererInput clustererInput, GoldStandardClusters goldStandard, Set<InstanceId> newEntities,
			Map<InstanceId, Integer> conflictingSets) {

		int countNormal = 0;
		int countNew = 0;
		int countConflict = 0;

		Map<InstanceId, PairType> pairs = new HashMap<>();
		for (Entry<InstanceId, Set<TableRowId>> entry : goldStandard.getTestingSetMappingClusters().entrySet()) {
			PairType type = PairType.CORRECT_NORMAL;
			if (conflictingSets.containsKey(entry.getKey())) {
				if (newEntities.contains(entry.getKey())) {
					type = PairType.CORRECT_NEW_CONFLICT;
				} else {
					type = PairType.CORRECT_CONFLICT;
				}
			} else if (newEntities.contains(entry.getKey())) {
				type = PairType.CORRECT_NEW;
			}

			int size = 1;

			pairs.put(entry.getKey(), type);

			if (conflictingSets.containsKey(entry.getKey())) {
				countConflict += size;

				if (newEntities.contains(entry.getKey())) {
					countNew += size;
				}
			}
			else if (newEntities.contains(entry.getKey())) {
				countNew += size;
			}
			else {
				countNormal += size;
			}
		}

		double maxClass = (double) Lists.newArrayList(countNormal, countNew).stream().mapToInt(Integer::intValue).max().getAsInt();

		double mNormal = maxClass / countNormal;
		double mConflict = maxClass / countConflict;
		double mNew = maxClass / countNew;

		Map<InstanceId, Double> ret = new HashMap<>();

		for (Entry<InstanceId, PairType> pairEntry : pairs.entrySet()) {

			double score = 0.0;
			switch (pairEntry.getValue()) {
			case CORRECT_CONFLICT:
				score = mNormal;
				break;
			case CORRECT_NEW:
				score = mNew;
				break;
			case CORRECT_NEW_CONFLICT:
				score = mNew;
				break;
			case CORRECT_NORMAL:
				score = mNormal;
				break;
			}

			ret.put(pairEntry.getKey(), score);
		}

		return ret;
	}

	public double evaluateRMSE(ClustererInput input, RowClusterSimilarities sim, List<Tri<TableRowIdPair, Boolean, Double>> evaluationSet) {

		List<TableRowId> rowIds = new ArrayList<>(input.getCandidates());

		Map<TableRowId, Integer> rowIdsIndex = ListIterable.iterable(rowIds).stream().collect(Collectors.toMap(ListEntry::getValue, ListEntry::getIndex));

		ConcurrentAccumulator<AccumDouble> accumulator = new ConcurrentAccumulator<>(AccumDouble.empty());
		ConcurrentAccumulator<AccumDouble> accumulatorDivider = new ConcurrentAccumulator<>(AccumDouble.empty());

		PersistentParallel6.runAndFinish(evaluationSet, (e) -> {
			double score = sim.get(rowIds, input.getClassId(), new ClusterEntryKey(rowIdsIndex.get(e.getA().getId1()), rowIdsIndex.get(e.getA().getId2())));

			double expectedScore = 1.0;
			if (!e.getB()) {
				expectedScore = 0.0;
			}

			double performance = expectedScore - score;
			accumulator.add(new AccumDouble(e.getC() * Math.pow(performance, 2.0)));
			accumulatorDivider.add((new AccumDouble(e.getC())));

		}, this.getClass(), true);

		double rmse = Math.sqrt(accumulator.get().get() / accumulatorDivider.get().get());

		return rmse;
	}

	public double evaluateFMeasure(WeightsRepository weightsRepository, OptimizableClass optimizableClass, ClustererInput input, RowClusterSimilarities sim,
			Map<TableRowIdPair, Pair<Boolean, Double>> pairs, GoldStandardClusters gs) {

		ConcurrentAccumulator<AccumDouble> accumMaxCorrect = new ConcurrentAccumulator<>(AccumDouble.empty());

		ConcurrentAccumulator<AccumDouble> accumCorrect = new ConcurrentAccumulator<>(AccumDouble.empty());
		ConcurrentAccumulator<AccumDouble> accumTotal = new ConcurrentAccumulator<>(AccumDouble.empty());

		List<TableRowId> rowIds = new ArrayList<>(input.getCandidates());
		Map<TableRowId, Integer> rowIdsIndex = ListIterable.iterable(rowIds).stream().collect(Collectors.toMap(ListEntry::getValue, ListEntry::getIndex));

		double threshold = weightsRepository.getOrDefault(optimizableClass.getWeight("SIMILARITY_THRESHOLD", 0.50));

		PersistentParallel6.runAndFinish(pairs.entrySet(), (pair) -> {

			double score = sim.get(rowIds, input.getClassId(),
					new ClusterEntryKey(rowIdsIndex.get(pair.getKey().getId1()), rowIdsIndex.get(pair.getKey().getId2())));

			double weight = pair.getValue().getSecond();
			boolean positive = pair.getValue().getFirst();

			if (positive) {
				accumMaxCorrect.add(new AccumDouble(weight));

			}

			if (score > threshold) {

				accumTotal.add(new AccumDouble(weight));
				if (positive) {
					accumCorrect.add(new AccumDouble(weight));

				} else {

				}
			}

		}, this.getClass());

		double correct = accumCorrect.get().get();

		double total = accumTotal.get().get();
		double maxCorrect = accumMaxCorrect.get().get();

		double precision = correct / (total);
		if (Double.isNaN(precision)) {
			precision = 0.0;
		}
		double recall = correct / maxCorrect;

		double fmeasure = 2 * precision * recall / (precision + recall);

		if (Double.isNaN(fmeasure)) {
			fmeasure = 0.0;
		}
		return fmeasure;
	}

	public double evaluateFMeasure(WeightsRepository weightsRepository, OptimizableClass optimizableClass, ClustererInput input, RowClusterSimilarities sim,
			WeightGetter<InstanceId> weights, GoldStandardClusters gs) {

		ConcurrentAccumulator<AccumDouble> accumMaxCorrect = new ConcurrentAccumulator<>(AccumDouble.empty());

		ConcurrentAccumulator<AccumDouble> accumCorrectRecall = new ConcurrentAccumulator<>(AccumDouble.empty());
		ConcurrentAccumulator<AccumDouble> accumCorrect = new ConcurrentAccumulator<>(AccumDouble.empty());
		ConcurrentAccumulator<AccumDouble> accumTotal = new ConcurrentAccumulator<>(AccumDouble.empty());
		ConcurrentAccumulator<AccumDouble> accumWrong = new ConcurrentAccumulator<>(AccumDouble.empty());

		List<TableRowId> rowIds = new ArrayList<>(input.getCandidates());
		long resultingSize = DoubleIteraterable.getResultingSize(rowIds.size());

		double threshold = weightsRepository.getOrDefault(optimizableClass.getWeight("SIMILARITY_THRESHOLD", 0.50));

		PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(ListIterable.iterable(rowIds)), (doubleEntry) -> {
			double score = sim.get(rowIds, input.getClassId(), new ClusterEntryKey(doubleEntry.getFirst().getIndex(), doubleEntry.getSecond().getIndex()));

			InstanceId mapping1 = gs.getMapping(doubleEntry.getFirst().getValue());
			InstanceId mapping2 = gs.getMapping(doubleEntry.getSecond().getValue());
			// double weight = (weights.getWeights(mapping1) +
			// weights.getWeights(mapping2)) / 2.0;
			double weight = 1.0;
			double recallWeight = weights.getWeights(mapping1);
			// weight = 1.0;
			boolean positive = false;
			if (gs.getIndex().get(doubleEntry.getFirst().getValue()).equals(gs.getIndex().get(doubleEntry.getSecond().getValue()))) {
				positive = true;
				accumMaxCorrect.add(new AccumDouble((1.0) * recallWeight));

			}

			if (score > threshold) {

				// accumTotal.add(new AccumDouble((1.0 - threshold) * weight));

				if (positive) {
					accumCorrect.add(new AccumDouble((1.0) * weight));
					accumCorrectRecall.add(new AccumDouble((1.0) * recallWeight));

				} else {
					accumWrong.add(new AccumDouble((1.0) * weight));

				}
			}
			else {
				if (positive) {
					// accumWrong.add(new AccumDouble((threshold - score) *
					// weight));

				} else {
					// accumCorrect.add(new AccumDouble((threshold - score) *
					// weight));

				}

			}
		}, resultingSize, this.getClass(), true);

		double correctRecall = accumCorrectRecall.get().get();
		double correct = accumCorrect.get().get();
		double wrong = accumWrong.get().get();
		double total = accumTotal.get().get();
		double maxCorrect = accumMaxCorrect.get().get();
		double precision = correct / (correct + wrong);
		if (Double.isNaN(precision)) {
			precision = 0.0;
		}
		double recall = correctRecall / maxCorrect;

		double fmeasure = 2 * precision * recall / (precision + recall);

		if (Double.isNaN(fmeasure)) {
			fmeasure = 0.0;
		}
		return fmeasure;
	}

	// {
	// ConcurrentAccumulator<AccumAverage> accumulator = new
	// ConcurrentAccumulator<AccumAverage>(AccumAverage.empty());
	// ConcurrentAccumulator<AccumAverage> positive = new
	// ConcurrentAccumulator<AccumAverage>(AccumAverage.empty());
	// ConcurrentAccumulator<AccumAverage> negative = new
	// ConcurrentAccumulator<AccumAverage>(AccumAverage.empty());
	// List<TableRowId> rowIds = new ArrayList<>(input.getCandidates() );
	//
	// long resultingSize = DoubleIteraterable.getResultingSize(rowIds.size());
	//
	// PersistentParallel6.runAndFinishWithSize(
	// DoubleIteraterable.iterable(ListIterable.iterable(rowIds)) ,
	// (doubleEntry) -> {
	// double score = sim.get(rowIds, input.getClassId(), new
	// ClusterEntryKey(doubleEntry.getFirst().getIndex(),
	// doubleEntry.getSecond().getIndex()));
	// double expected = -1.0;
	// if(gsIndex.get(doubleEntry.getFirst().getValue()).equals(gsIndex.get(doubleEntry.getSecond().getValue())))
	// {
	// expected = 1.0;
	// }
	// double performance = 1.0 - ( Math.abs(score - expected) / 2.0 );
	// if(performance < 0.0) {
	// int sdf= 123;
	// }
	// accumulator.add(new AccumAverage(performance));
	//
	// if(expected > 0.0) {
	// positive.add(new AccumAverage(performance));
	// }
	// else {
	// negative.add(new AccumAverage(performance));
	// }
	// }, resultingSize,this.getClass());
	//
	//
	// return new SimPerformance(accumulator.get().get(), positive.get().get(),
	// negative.get().get());
	// }
}
