package expansion.set.evaluation.sim;

import java.util.function.BinaryOperator;

import de.mannheim.uni.expansion.tablefusion.analysis.stores.PerformanceMacroPR;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.Accumulateable;

public class SimPerformance  implements Accumulateable<SimPerformance>{

	
	private final PerformanceMacroPR performance; 
	private final double rmse;
	private final int count;
	
	
	public SimPerformance() {
		this.rmse = 0.0;
		this.count = 0;
		this.performance = new PerformanceMacroPR(); 
	}
	
	public SimPerformance(double rmse, PerformanceMacroPR performance ) {
		super();
		this.rmse = rmse;
		this.count = 1;
		this.performance = performance;
	}
	
	
	private SimPerformance(double rmse, int count, PerformanceMacroPR performance ) {
		super();
		this.rmse = rmse;
		this.count = count;
		this.performance = performance;
	}
	
	 
	
	public PerformanceMacroPR getPerformance() {
		return performance;
	}
	
	

	public double getRmse() {
		if(count == 0)
			return 0.0;
		else
			return rmse / count;
	}

	public static SimPerformance accumulator(SimPerformance a, SimPerformance b) {
		return new SimPerformance(a.rmse + b.rmse, a.count + b.count, a.performance.accumulate(b.performance));
	}
	
	
	@Override
	public BinaryOperator<SimPerformance> getAccumulator() {
		return SimPerformance::accumulator;				
	}

	@Override
	public SimPerformance accumulate(SimPerformance other) {
		return accumulator(this, other);
	}

	@Override
	public Class<SimPerformance> getAcumClass() {
		return SimPerformance.class;
	}

	@Override
	public String toString() {
		return "SimPerformance [performance=" + performance + "]";
	}
 
	
	
	
 
}
