package expansion.set.exportimport;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Queue;
import java.util.Set;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Consumer;
import java.util.stream.Collectors;

import javax.inject.Inject;
import javax.inject.Named;

import org.apache.commons.compress.compressors.gzip.GzipCompressorInputStream;
import org.apache.commons.compress.compressors.gzip.GzipCompressorOutputStream;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.utils.BatchConsumerProducer;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.DoubleIteraterable;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Tri;
import expansion.set.SetExpansionProperties;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.exportimport.OptionalClusterCache.OptionalClusterCacheMap;

public class ClusterPairsExporter {

	@Inject
	public ClusterPairsExporter() {
		// TODO Auto-generated constructor stub
	}

	@Inject
	@Named(SetExpansionProperties.tablesCachesFolder)
	protected String cacheFolder;

	@Inject
	protected ExternalClusteringImporter externalClusteringImporter;

	// List<Tri<TableRowId, TableRowId, Double>> pairsList() {
	//
	// }

	public void writePairList(String name, RowClusterSimilarities sim, ClassId classId, Set<TableRowId> testingInput, int run,
			Map<InstanceId, Set<TableRowId>> gsMap) {

		// String classFolder = clusteringOutputFolder + "/" +
		// classId.getFinalPart() + "/";
		// MKDIR.createDirUnchecked(new File(classFolder));

		String fileName = cacheFolder + "/" + ExternalClusteringImporter.EXTERNAL_FOLDER_NAME + "/" + classId.getFinalPart() + "_" + name;

		Map<TableRowIdPair, Double> finalMap = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(testingInput), (entry) -> {
			List<TableRowId> rows = new ArrayList<>();
			rows.add(entry.getFirst());
			rows.add(entry.getSecond());

			finalMap.put(new TableRowIdPair(entry.getFirst(), entry.getSecond()), sim.get(rows, classId, new ClusterEntryKey(0, 1)));
		}, DoubleIteraterable.getResultingSize(testingInput.size()), this.getClass());

		Map<TableRowId, Integer> gsIndex = new HashMap<>();
		int i = 0;
		for (Set<TableRowId> rows : gsMap.values()) {
			for (TableRowId rowId : rows) {
				gsIndex.put(rowId, i);
			}
			i++;
		}

		write(fileName, testingInput, finalMap, run, gsIndex);
	}

	public Map<TableRowId, Integer> getMapping(String file) {
		try {
			Map<TableRowId, Integer> mapping = new HashMap<>();
			try (BufferedReader br = new BufferedReader(
					new InputStreamReader(new GzipCompressorInputStream(new FileInputStream(new File(file))), StandardCharsets.UTF_8))) {
				String line;
				while ((line = br.readLine()) != null) {
					String[] split = line.split(" ");
					mapping.put(TableRowId.fromCombinedString2(split[1]), Integer.parseInt(split[0]));
				}
			}
			return mapping;
		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
	}

	public OptionalClusterCache getFullCache(String file, RowClusterSimilarities sim, ClustererInput clustererInput, int sets) {
		Map<TableRowId, Integer> mapping = getMapping(file + "_MAPPING.gz");

		Map<Integer, Set<Integer>> pairs = getPairs(mapping, sim);

		OptionalClusterCache cache = getCache(mapping, pairs, file, sets);

		return cache; // OptionalClusterCache(mapping, pairs, cache);
	}

	public OptionalClusterCache getCache(Map<TableRowId, Integer> mapping, Map<Integer, Set<Integer>> pairs, String file, int sets) {
		OptionalClusterCacheMap[] cacheArray = new OptionalClusterCacheMap[mapping.size()];
		OptionalClusterCache cache = new OptionalClusterCache(mapping, pairs, cacheArray, file);
		BlockingQueue<Runnable> linkedBlockingQueue = new LinkedBlockingQueue<Runnable>(15);
		ExecutorService executorService = new ThreadPoolExecutor(1, 1, 30, TimeUnit.SECONDS, linkedBlockingQueue, new ThreadPoolExecutor.AbortPolicy());

		// return new ThreadPoolExecutor(nThreads, nThreads,
		// 0L, TimeUnit.MILLISECONDS,
		// new LinkedBlockingQueue<Runnable>());

		try {

			Consumer<List<String>> batch = (list) -> {
				ConcurrentLinkedQueue<Tri<Integer, Integer, Double>> queue = new ConcurrentLinkedQueue<>();
				PersistentParallel6.runAndFinish(list, (line) -> {

					String[] split = line.split(" ");
					if (split.length != 3) {
						int sdf = 123;
					} else {
						Integer id1 = Integer.parseInt(split[0]);
						Integer id2 = Integer.parseInt(split[1]);
						Double value = Double.parseDouble(split[2]);
						cache.addPublic(id1, id2, value);

						queue.add(new Tri<>(id1, id2, value));
					}

				}, this.getClass());

				cache.writeUpdates(false);
			};

			Consumer<File> readFile = (ffFile) -> {
				try {
					try (BufferedReader br = new BufferedReader(
							new InputStreamReader(new GzipCompressorInputStream(new FileInputStream(ffFile)), StandardCharsets.UTF_8))) {
						String line;
						long i = 0;
						int numberInOneQueue = 700000000;
						List<String> list = new ArrayList<>(numberInOneQueue);
						while ((line = br.readLine()) != null) {
							list.add(line);
							if (list.size() == numberInOneQueue) {

								boolean added = false;
								final List<String> finalList = list;
								while (!added) {
									try {
										executorService.submit(() -> batch.accept(finalList));
										added = true;
									} catch (RejectedExecutionException e) {
										try {
											TimeUnit.SECONDS.sleep(4);
										} catch (Exception e1) {
											// TODO Auto-generated catch block
											e1.printStackTrace();
										}
									}
								}
								list = new ArrayList<>(numberInOneQueue);
							}
							i++;
						}
						if (!list.isEmpty()) {

							boolean added = false;
							final List<String> finalList = list;
							while (!added) {
								try {
									executorService.submit(() -> batch.accept(finalList));
									added = true;
								} catch (RejectedExecutionException e) {
									try {
										TimeUnit.SECONDS.sleep(4);
									} catch (Exception e1) {
										// TODO Auto-generated catch block
										e1.printStackTrace();
									}
								}
							}
						}
					}
				} catch (IOException e) {
					e.printStackTrace();
				}
			};

			boolean withSets = false;
			if (sets != -1) {
				withSets = true;
			} else {
				sets = 1;
			}
			for (int currentSet = 0; currentSet < sets; currentSet++) {
				String suffix = "";
				if (withSets) {
					suffix = "_" + currentSet;
				}

				readFile.accept(new File(file + suffix + ".gz"));

			}

			File dir = new File(file);
			MKDIR.createDirUnchecked(dir);

			File[] listOfFiles = (dir).listFiles();

			for (File oFile : listOfFiles) {
				readFile.accept(oFile);
			}

			executorService.shutdown();
			boolean terminated = false;
			while (!terminated) {
				terminated = executorService.awaitTermination(Long.MAX_VALUE, TimeUnit.NANOSECONDS);
			}

		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
		return cache;
	}

	public static Map<Integer, Set<Integer>> getPairs(Map<TableRowId, Integer> mapping, RowClusterSimilarities sim) {

		ConcurrentHashMap<Integer, Set<Integer>> blockToNode = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(mapping.entrySet(), (entry) -> {
			List<Tri<Set<Integer>, Integer, Double>> rowBlocks = sim.getSplitBlocksWithScores(entry.getKey());
			for (Tri<Set<Integer>, Integer, Double> blockData : rowBlocks) {
				if (blockData.getC() >= 1.3) {
					for (Integer block : blockData.getA()) {
						CHMHelper.getFromCHM(blockToNode, block, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(entry.getValue());
					}
				}
			}
		}, ClusterPairsExporter.class + "::GetBlockToNodes");

		ConcurrentHashMap<Integer, Set<Integer>> pairs = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(blockToNode.entrySet(), (entry) -> {

			for (DoubleIteraterable.DoubleEntry<Integer> de : DoubleIteraterable.iterable(entry.getValue())) {
				if (de.getFirst() < de.getSecond()) {
					CHMHelper.getFromCHM(pairs, de.getFirst(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(de.getSecond());
				} else {
					CHMHelper.getFromCHM(pairs, de.getSecond(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(de.getFirst());
				}
			}

			blockToNode.remove(entry.getKey());
		}, ClusterPairsExporter.class + "::FromBlockToPairs");
		blockToNode.clear();

		return pairs;
	}

	public void writeFull(String outputCacheFile, RowClusterSimilarities sim, ClustererInput clustererInput, int sets, int setId) {
		// WE ASSUME
		boolean MAPPING_EXISTS = Files.exists(Paths.get(outputCacheFile + "_MAPPING.gz"));
		Map<TableRowId, Integer> mapping;
		List<TableRowId> list;
		if (sets != -1 || MAPPING_EXISTS) {
			mapping = getMapping(outputCacheFile + "_MAPPING.gz");
			TableRowId[] array = new TableRowId[mapping.size()];

			for (Entry<TableRowId, Integer> entry : mapping.entrySet()) {
				array[entry.getValue()] = entry.getKey();
			}

			list = Arrays.asList(array);

		} else {
			mapping = new HashMap<>();
			list = new ArrayList<>();
			int i = 0;
			for (TableRowId rowId : clustererInput.getCandidates()) {
				mapping.put(rowId, i);
				list.add(rowId);
				i++;
			}
		}

		Map<Integer, Set<Integer>> pairs = getPairs(mapping, sim);

		// PersistentParallel6.runAndFinish(mapping.entrySet(), (entry) -> {
		//
		// List<Tri<Set<Integer>, Integer, Double>> rowBlocks =
		// sim.getSplitBlocksWithScores(entry.getKey());
		// for (Tri<Set<Integer>, Integer, Double> blockData : rowBlocks) {
		// if (blockData.getC() >= 1.3) {
		// for (Integer block : blockData.getA()) {
		// Set<Integer> rowIds = blockToNode.get(block);
		// for (Integer rowId : rowIds) {
		// if (!rowId.equals(entry.getValue())) {
		// if (rowId < entry.getValue()) {
		// CHMHelper.getFromCHM(pairs, rowId, () ->
		// Collections.newSetFromMap(new
		// ConcurrentHashMap<>())).add(entry.getValue());
		// } else {
		// CHMHelper.getFromCHM(pairs, entry.getValue(), () ->
		// Collections.newSetFromMap(new ConcurrentHashMap<>())).add(rowId);
		// }
		// }
		// }
		// }
		// }
		// }
		// }, this.getClass());

		String suffix;
		if (sets != -1) {
			suffix = "_" + setId;
		} else {
			suffix = "";
		}

		ConcurrentLinkedQueue<Tri<Integer, Integer, Double>> queueFinal = new ConcurrentLinkedQueue<>();

		AtomicInteger doneComputing = new AtomicInteger(1);

		CountDownLatch doneWritingFile = new CountDownLatch(1);
		String file = outputCacheFile;
		try {

			PersistentParallel6.LIMIT_OBSERVER_RUN = true;

			if (!(sets != -1 || MAPPING_EXISTS)) {
				BufferedWriter out = new BufferedWriter(
						new OutputStreamWriter(new GzipCompressorOutputStream(new FileOutputStream(new File(file + "_MAPPING.gz"))), StandardCharsets.UTF_8),
						32768);

				for (Entry<TableRowId, Integer> entry : mapping.entrySet()) {
					out.write(String.valueOf(entry.getValue()) + " " + entry.getKey().getCombinedString2() + "\n");

				}
				out.close();

			}

			@SuppressWarnings("resource")
			BufferedWriter outPairs = new BufferedWriter(
					new OutputStreamWriter(new GzipCompressorOutputStream(new FileOutputStream(new File(file + suffix + ".gz"))), StandardCharsets.UTF_8));

			Runnable runA = new Runnable() {
				public void run() {
					try {
						Tri<Integer, Integer, Double> entry;
						boolean lastRun = false;
						while (true) {
							entry = queueFinal.poll();
							if (entry == null) {
								if (lastRun) {
									break;
								} else {
									java.util.concurrent.TimeUnit.SECONDS.sleep(5);
									if (doneComputing.intValue() == 0) {
										lastRun = true;
									}
								}
							} else {

								outPairs.write(String.valueOf(entry.getA()) + " " + String.valueOf(entry.getB()) + " " + formatter.format(entry.getC()) + "\n");

							}

						}
						outPairs.close();
						doneWritingFile.countDown();
					} catch (Exception e) {
						e.printStackTrace();
						throw new IllegalStateException(e);
					}
				}

			};

			Thread threadA = new Thread(runA, "threadA");
			threadA.start();

		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalStateException(e);
		}

		BatchConsumerProducer<ClusterEntryKey> consumerProducer = BatchConsumerProducer.createMultiRun("ClusterPairBatchConsumerProducer", (pair) -> {
			Tri<Integer, Integer, Double> tri = new Tri<>(pair.getE1(), pair.getE2(), sim.get(list, clustererInput.getClassId(), pair));
			if (tri.getC() > -0.99) {
				queueFinal.add(tri);
			}
		}, 4, 85000);

		int begin = 0;
		int endExclusive = mapping.size();

		if (sets != -1) {

			int perSet = mapping.size() / sets;
			begin = setId * perSet;
			endExclusive = begin + perSet;
			if ((setId + 1) == sets) {
				endExclusive = mapping.size();
			}

		}

		int min = pairs.entrySet().stream().mapToInt((e) -> e.getValue().size()).min().getAsInt();
		int batchSize = Math.max(3500000, min);
		List<ClusterEntryKey> keyList = new ArrayList<>(batchSize);
		for (int i = begin; i < endExclusive; i++) {
			// for (Entry<Integer, Set<Integer>> pairEntry : pairs.entrySet()) {
			if (pairs.containsKey(i)) {
				Set<Integer> value = pairs.get(i);

				if (keyList.size() + value.size() > batchSize) {
					consumerProducer.addBatch(keyList);
					keyList = new ArrayList<>(batchSize);
				}
				for (Integer second : value) {
					keyList.add(new ClusterEntryKey(i, second));
				}
				pairs.remove(i);
			}
		}
		if (keyList.size() > 0) {
			consumerProducer.addBatch(keyList);
			keyList = null;
		}

		pairs.clear();

		consumerProducer.finish();

		doneComputing.decrementAndGet();

		// PersistentParallel6.runAndFinish(pairs.entrySet(), (pairEntry) -> {
		// List<ClusterEntryKey> keyList = new ArrayList<>();
		// for (Integer second : pairEntry.getValue()) {
		// keyList.add(new ClusterEntryKey(pairEntry.getKey(), second));
		// }
		// pairs.remove(pairEntry.getKey());
		//
		// consumerProducer.addBatch(keyList);
		// }, this.getClass() + "::FromPairsToQueue");
		//
		// ConcurrentLinkedQueue<ClusterEntryKey> queue = new
		// ConcurrentLinkedQueue<>();

		try {
			doneWritingFile.await();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		// write(outputCacheFile, mapping, queueFinal);
		PersistentParallel6.LIMIT_OBSERVER_RUN = false;
	}

	public static NumberFormat formatter = new DecimalFormat("#0.0000");

	void write(String file, Set<TableRowId> candidates, Map<TableRowIdPair, Double> map, int run, Map<TableRowId, Integer> gsIndex) {
		try {

			String mappingsFile = file + "_MAPPING" + "_" + run;
			Map<TableRowId, Integer> mapping;
			Map<Integer, TableRowId> mappingReverse;
			if (Files.exists(Paths.get(file))) {
				mappingReverse = externalClusteringImporter.getMapping(mappingsFile);
				mapping = mappingReverse.entrySet().stream().collect(Collectors.toMap(Entry::getValue, Entry::getKey));
			} else {
				mappingReverse = new HashMap<>();
				mapping = new HashMap<>();

				int i = 0;
				for (TableRowId rowId : candidates) {
					mappingReverse.put(i, rowId);
					mapping.put(rowId, i);
					i++;
				}

				BufferedWriter out = new BufferedWriter(new FileWriter(mappingsFile), 32768);

				for (Entry<TableRowId, Integer> entry : mapping.entrySet()) {
					out.write(String.valueOf(entry.getValue()) + " " + entry.getKey().getCombinedString2() + "\n");

				}
				out.close();

			}


			BufferedWriter gsOut = new BufferedWriter(new FileWriter(file + "_GS" + "_" + run), 32768);

			for (int i = 0; i < mappingReverse.size(); i++) {
				gsOut.write(Integer.valueOf(gsIndex.get(mappingReverse.get(i))) + "\n");

			}

			gsOut.close();

			BufferedWriter out = new BufferedWriter(new FileWriter(file + "_" + run), 32768);

			for (Entry<TableRowIdPair, Double> entry : map.entrySet()) {
				out.write(mapping.get(entry.getKey().getId1()) + " " + mapping.get(entry.getKey().getId2()) + " "
						+ formatter.format(entry.getValue()) + "\n");

			}
			out.close();
		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
	}

	void write(String file, Map<TableRowId, Integer> mapping, Collection<Tri<Integer, Integer, Double>> pairs) {
		try {

			BufferedWriter out = new BufferedWriter(new FileWriter(file + "_MAPPING"), 32768);

			for (Entry<TableRowId, Integer> entry : mapping.entrySet()) {
				out.write(String.valueOf(entry.getValue()) + " " + entry.getKey().getCombinedString2() + "\n");

			}
			out.close();

			out = new BufferedWriter(new FileWriter(file), 32768);

			for (Tri<Integer, Integer, Double> entry : pairs) {
				out.write(String.valueOf(entry.getA()) + " " + String.valueOf(entry.getB()) + " " + formatter.format(entry.getC()) + "\n");

			}
			out.close();
		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
	}

	class IterableThatRemovesReturnedValues<T> implements Iterable<T> {
		Queue<T> queue;

		IterableThatRemovesReturnedValues(Queue<T> q) {
			queue = q;
			iterator = new IteratorThatRemovesReturnedValues();
		}

		boolean hasNext() {
			return queue.peek() != null;
		}

		T next() {
			return queue.poll();
		}

		Iterator<T> iterator;

		@Override
		public Iterator<T> iterator() {
			return iterator;
		}

		class IteratorThatRemovesReturnedValues implements Iterator<T> {
			public boolean hasNext() {
				return queue.peek() != null;
			}

			public T next() {
				return queue.poll();
			}
		}
	}


}
