package expansion.set.exportimport;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.inject.Inject;
import javax.inject.Named;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import expansion.set.SetExpansionProperties;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.model.RowCluster;

public class ExternalClusteringImporter {

	@Inject
	@Named(SetExpansionProperties.tablesCachesFolder)
	protected String cacheFolder;

	public static final String EXTERNAL_FOLDER_NAME = "externalInputFolder";

	@Inject
	public ExternalClusteringImporter() {

	}

	public Map<Integer, TableRowId> getMapping(String file) {
		try {
			Map<Integer, TableRowId> mapping = new HashMap<>();
			try (BufferedReader br = new BufferedReader(new InputStreamReader(
					new FileInputStream(new File(file)),
					StandardCharsets.UTF_8))) {
				String line;
				while ((line = br.readLine()) != null) {
					String[] split = line.split(" ");
					mapping.put(Integer.parseInt(split[0]), TableRowId.fromCombinedString2(split[1]));
				}
			}
			return mapping;
		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
	}

	public Collection<Set<TableRowId>> get(ClassId classId, String name, int run) {
		try {
			Map<Integer, TableRowId> mapping = getMapping(
					cacheFolder + "/" + EXTERNAL_FOLDER_NAME + "/" + classId.getFinalPart() + "_" + name + "_MAPPING_" + run);
			Map<Integer, Set<TableRowId>> ret = new HashMap<>();

			try (BufferedReader br = new BufferedReader(new InputStreamReader(
					new FileInputStream(new File(cacheFolder + "/externalInputFolder/" + classId.getFinalPart() + "_" + name + "_" + run + "_sol")),
					StandardCharsets.UTF_8))) {
				String line;
				int i = 0;
				while ((line = br.readLine()) != null) {
					TableRowId rowId = mapping.get(i);
					CHMHelper.getFromMap(ret, Integer.parseInt(line), () -> new HashSet<>()).add(rowId);
					i++;
				}
			}

			return ret.values();
		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
	}

	public void writeYaserSol(String name, ClassId classId, ClusterList clusters, int run) {
		try {
			Map<Integer, TableRowId> mapping = getMapping(
					cacheFolder + "/" + EXTERNAL_FOLDER_NAME + "/" + classId.getFinalPart() + "_" + name + "_MAPPING_" + run);

			Map<TableRowId, Integer> clustering = new HashMap<>();
			int c = 0;
			for (RowCluster rc : clusters.getClusterList()) {

				for (TableRowId rowId : rc.getRows()) {
					clustering.put(rowId, c);
				}
				c++;
			}

			BufferedWriter gsOut = new BufferedWriter(
					new FileWriter(cacheFolder + "/" + EXTERNAL_FOLDER_NAME + "/" + classId.getFinalPart() + "_" + name + "_" + run + "_yasersol"), 32768);

			for (int i = 0; i < mapping.size(); i++) {
				gsOut.write(Integer.valueOf(clustering.get(mapping.get(i))) + "\n");

			}

			gsOut.close();
		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
	}
}
