package expansion.set.gscreation;

import java.io.File;
import java.io.IOException;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NavigableMap;
import java.util.NavigableSet;
import java.util.Optional;
import java.util.Random;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.mutable.MutableInt;

import com.google.gson.Gson;

import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId.ClassIdStyle;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.DecimalFormatter;
import de.mannheim.uni.expansion.utils.ImmuMaker;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.QuartileData;
import de.mannheim.uni.expansion.utils.QuartileData.Quartile;
import de.mannheim.uni.expansion.utils.QuartileFilter;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.dws.expansion.data.kb.degree.DegreeRepository;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.utils.kryo.CHMKryoAdapter;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdContainer;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumAverage;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumMedian;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentAccumulator;
import de.uni_mannheim.informatik.dws.expansion.utils.MutableReference;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.clustering.model.RowClusterUUID;
import expansion.set.creation.Entity;
import expansion.set.creation.EntityCreator;
import expansion.set.evaluation.SetExpansionRunner;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.evaluation.crossvalidation.EntitySetGroupedValuesSplitter;
import expansion.set.evaluation.crossvalidation.EntitySets;
import expansion.set.evaluation.matching.EntityMatcher;
import expansion.set.evaluation.matching.ScoredEntity;
import expansion.set.strategies.rowclusterers.RowClustererRepository;
import expansion.set.strategies.rowclusterers.label.LabelSimilarityClustererClassBlocked;

public class ExtractTablesForGoldStandard2 {
 
	@Inject
	protected LabelSimilarityClustererClassBlocked labelSimilarityClusterer;
	

	public static final int MIN_COLUMNS = 1;
	
	public static final int MAX_ROWS_PER_TABLE = 50;
	
	public static final int MIN_ROWS_PER_CLUSTER_INSTANCE = 4;
	

	public static final int INSTANCES_PER_QUARTILE = 3;
	
	public static final int MIN_ROWS_PER_INSTANCE = 5;
	public static final int MAX_ROWS_PER_INSTANCE = 10; 
	public static final int MAX_ROWS_PER_ADDITIONAL_INSTANCE = 5; 

	@Inject
	protected RowClustererRepository rowClustererRepository;
	
	@Inject
	protected GoldStandardWriter goldStandardWriter;

	@Inject
	protected KnowledgeBase knowledgeBase;
	
	@Inject
	protected AnnotatedTablesRepository tableRepository;
	

	@Inject
	protected SetExpansionRunner setExpansionRunner;
	

	@Inject
	protected DegreeRepository degreeRepository;
	

	@Inject
	protected EntitySetGroupedValuesSplitter entitySetGroupedValuesSplitter;
	

	@Inject
	protected   EntityCreator entityCreator;
	
	@Inject
	protected EntityMatcher entityMatcher;
	

	@Inject
	protected CHMKryoAdapter chmKryoAdapter;
	 
	
	public ExtractTablesForGoldStandard2() {
	
	}
	
	
	public void run(Map<ClassId, Set<ClassId>> topClassesAndParents, Map<ClassId, Integer> maxIndegree,String clusterCacheFolder, String gsPath, FusionConfig fusionConfig,  SetExpansionStrategiesConfig setExpansionStrategiesConfig, String thresholdsFolder, StrategiesConfig strategiesConfig) {
		
		Map<ClassId, Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>>> finalGs = new HashMap<>();
		Map<ClassId, Set<InstanceId>> kbInstances = new HashMap<>();
		Map<ClassId, Set<InstanceId>> newInstances = new HashMap<>();
		Map<ClassId, Set<InstanceId>> additionalInstances = new HashMap<>();
		Map<ClassId, TableRowInstanceBiMap> kbMappingsMap = new HashMap<>();
		Map<ClassId, TableRowInstanceBiMap> clusterMappingsMap = new HashMap<>();
		

		Map<ClassId, Set<InstanceId>> quartileKbInstances = new HashMap<>();
		Map<ClassId, Set<InstanceId>> quartileNewInstances = new HashMap<>();
		Map<ClassId, TableRowInstanceBiMap> quartileKbMappingsMap = new HashMap<>();
		Map<ClassId, TableRowInstanceBiMap> quartileClusterMappingsMap = new HashMap<>();
		
		Map<AnnotatedTableId, ClassId> fullTableSet = new HashMap<>();

		Map<ClassId, Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>>> quartileGs = new LinkedHashMap<>();
		
		Map<ClassId, Map<String, InstanceData>> instanceJson = new HashMap<>();
		
		for(Entry<ClassId, Set<ClassId>> classEntry : topClassesAndParents.entrySet()) {
			
			ClassId classId = classEntry.getKey();
			
			Map<String, InstanceData> instanceDataList = new HashMap<>();
			
			ConcurrentHashMap<PropertyId, AtomicInteger> propertyCountMap = new ConcurrentHashMap<>();
			
			TableIdsSet tableIdsSet = new TableIdsSet(this.tableRepository.getAllTableIds(classId).stream().filter((e) -> {
				AnnotatedTableMetaData md = tableRepository.getMetaData(e);
				
				AnnotatedTable data = tableRepository.getTable(e);
				double matched = md.getInstances().size();
				int total = md.getNumberOfRows();
				
				double minProportion = -0.1;
				double maxRows = MAX_ROWS_PER_TABLE;
				
//				if(classId.getClassId().equals("http://dbpedia.org/ontology/Settlement")) {
//					minProportion = 0.29;
//					maxRows = 100;
//				}
				
				double matchedProportion = matched / total;
				boolean matchProportionBoolean = matchedProportion > minProportion && matchedProportion < 1.1;
				boolean keep =  md.getNumberOfRows() <=  maxRows && md.getProperties().size() >= MIN_COLUMNS  && matchProportionBoolean;
				double score = 0.0;
				
				 for(PropertyId propertyId : md.getProperties().values()) {
					 CHMHelper.getFromCHM(propertyCountMap,propertyId, () -> new AtomicInteger()).incrementAndGet();
				 }
				if( md.getProperties().size() >= MIN_COLUMNS && keep == false) {
					int sdf= 123;
				}
				if(keep) {
					int sdf= 123;
				}
				return keep;
				}
			).collect(Collectors.toSet()), new UuidStringId(classId.getFinalPart()));
			
			
			Random random = new Random( classId.getClassId().hashCode());
			random = new Random(random.nextLong());
			
			if(classId.getFinalPart().equals("Settlement")  ) {
			
				System.out.println("===============");
				System.out.println("Before FILTER");
				System.out.println();
				
				for(Entry<PropertyId, AtomicInteger> countEntry : propertyCountMap.entrySet()) {
					System.out.println(countEntry.getKey().getPropertyId()+ ": " + countEntry.getValue().intValue());
				}
				
				Map<AnnotatedTableId, Double> scores = new HashMap<>();
				
				Map<PropertyId, AtomicInteger> newPropertyCountMap = new HashMap<>();
				
				
				
				for(AnnotatedTableId tableId : tableIdsSet.getTableIds()) {
					Set<PropertyId> set = new HashSet<>(this.tableRepository.getMetaData(tableId).getProperties().values());
					
					double countryOffsetScore = 0.66;
					if(random.nextDouble() > 0.8) {
						  countryOffsetScore = 1.50;
					}
					double score = 0.0;
					if(set.contains(new PropertyId("http://dbpedia.org/ontology/postalCode"))) {
						score += 2.0;
					}
					if(set.contains(new PropertyId("http://dbpedia.org/ontology/isPartOf"))) {
						score += 1.35;
					}
					if(set.contains(new PropertyId("http://dbpedia.org/ontology/populationTotal"))) {
						score += 2.0;
					}
					if(set.contains(new PropertyId("http://dbpedia.org/ontology/elevation"))) {
						score += 2.0;
					}
					if(set.contains(new PropertyId("http://www.w3.org/2003/01/geo/wgs84_pos#lat")) && set.contains(new PropertyId("http://www.w3.org/2003/01/geo/wgs84_pos#long"))) {
						score += 2.0;
					}
					
					if(set.contains(new PropertyId("http://dbpedia.org/ontology/country"))) {
						score += countryOffsetScore;
						if(set.contains(new PropertyId("http://dbpedia.org/ontology/utcOffset"))) {
							score -= countryOffsetScore;
						}
					}
					else {
						if(set.contains(new PropertyId("http://dbpedia.org/ontology/utcOffset"))) {
							score -= 0.66;
						}
					}
					if(score > 1.4) {
						scores.put(tableId, score);
						
						for(PropertyId propertyId : set) {
							CHMHelper.getFromMap(newPropertyCountMap, propertyId, () -> new AtomicInteger()).incrementAndGet();
						}
						
					}
					else {
						int sdf= 123;
					}
				}
				
				System.out.println();
				System.out.println("===============");
				System.out.println("AFTER FILTER");
				System.out.println();
				
				
				for(Entry<PropertyId, AtomicInteger> countEntry : newPropertyCountMap.entrySet()) {
					System.out.println(countEntry.getKey().getPropertyId()+ ": " + countEntry.getValue().intValue());
				}
			
				tableIdsSet = new TableIdsSet(scores.keySet(),  new UuidStringId(classId.getFinalPart()));
			
			
			}
			
			
			for(AnnotatedTableId tableId : tableIdsSet.getTableIds()) {
				fullTableSet.put(tableId, classId);
			}
			
			if(false) {
			Tri<TableRowInstanceBiMap, TableRowInstanceBiMap, Map<InstanceId, Map<InstanceId, Double>>> clusters = getRowClusters(clusterCacheFolder, classId, classEntry.getValue(), tableIdsSet, fusionConfig, setExpansionStrategiesConfig, thresholdsFolder, strategiesConfig);
			
			Tri<TreeMap<InstanceId, Set<TableRowId>>, Map<InstanceId, Quartile>, TableRowInstanceBiMap> mapping = getMappings(classId, random, tableIdsSet, maxIndegree.getOrDefault(classId, Integer.MAX_VALUE), clusters.getC().keySet());
			kbInstances.put(classId, new HashSet<>(mapping.getA().keySet()));
			 
			kbMappingsMap.put(classId, mapping.getC());
			clusterMappingsMap.put(classId, clusters.getB());
			Map<InstanceId, Quartile> instanceMap = mapping.getB();
			Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>> tableAndMappings = new HashMap<>();
			NavigableMap<ClassId, Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>>> quartileMap = new TreeMap<>(
					(e1,e2) -> {
						return e1.getClassId().compareTo(e2.getClassId());
					});
			
			Set<ClassId> quartileClassIds = new HashSet<>();
			Set<InstanceId> newInstancesSetSet = new HashSet<>();
			for(Entry<InstanceId, Set<TableRowId>> e : mapping.getA().entrySet()) {
				
				InstanceData instanceData = new InstanceData();

				MutableReference<InstanceId> newInstancesSet = new MutableReference<>();
				
//				Set<TableRowId> tableRowIdsWithLabelConflicts = new HashSet<>();
//				tableRowIdsWithLabelConflicts.addAll(e.getValue());
				
				Set<AnnotatedTableId> allTables = e.getValue().stream().map((row) -> row.getTableId()).collect(Collectors.toSet());
				
				Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>> tables = findBestTables(e.getKey(), e.getValue(), random, clusters.getA(),clusters.getB(), newInstancesSet, newInstancesSetSet);
				
				
				Map<InstanceId, Pair<Double, Set<TableRowId>>> labelConflictFilteredTables = clusters.getC().get(e.getKey()).entrySet().stream().filter( (c) -> {
					return clusters.getB().getCluster(c.getKey()).stream().map((row) -> row.getTableId()).filter((tid) -> !allTables.contains(tid)).distinct().count() > 1; 
				}).collect(Collectors.toMap(Entry::getKey, (c) -> {
					return new  Pair<>(c.getValue(), clusters.getB().getCluster(c.getKey()).stream().filter((tid) -> !allTables.contains(tid.getTableId())).collect(Collectors.toSet())); 

				}));
				List< InstanceId > labelConflictChoices =labelConflictFilteredTables.entrySet().stream().sorted((e1,e2) -> {
					int compare = Double.compare(e2.getValue().getFirst(), e1.getValue().getFirst());
					if(compare != 0) {
						return compare;
					}
					else {
						compare = Integer.compare(e2.getValue().getSecond().size(), e1.getValue().getSecond().size());
						if(compare != 0)
							return compare;
						else
							return e1.getKey().getInstanceId().compareTo(e2.getKey().getInstanceId());
					}
				}).map((c) ->c.getKey()).collect(Collectors.toList());
				
				int choice = 0;
				int choiceLimit = 2;
				Set<InstanceId> additionalInstancesSet = new HashSet<>();
				for( InstanceId  candidate : labelConflictChoices) {
					if(choice < choiceLimit) {
						
//						tableRowIdsWithLabelConflicts.addAll(clusters.getB().getCluster(candidate.getKey()));
						Set<AnnotatedTableId> additionalTables = findBestTables(labelConflictFilteredTables.get(candidate).getSecond().stream().map((f) -> f.getTableId()).collect(Collectors.toSet()), random, clusters.getB(), MAX_ROWS_PER_ADDITIONAL_INSTANCE);
						for(AnnotatedTableId tableId : additionalTables) {
							CHMHelper.getFromMap(tables, tableId, () -> new ArrayList<>());
						}
						choice++;
					}
					else {
						break;
					}
					
				}
				
				instanceData.setQuartile(instanceMap.get(e.getKey()).name());
				instanceData.setTotalTables(tables.size());
				Set<AnnotatedTableId> mainInstanceTableCount = new HashSet<>();
				for(TableRowId rowId : e.getValue()) {
					if(tables.containsKey(rowId.getTableId())) {
						mainInstanceTableCount.add(rowId.getTableId());
					}
					
				}
				instanceData.setMainInstance(new InstancePair(e.getKey().getInstanceId(), mainInstanceTableCount.size()));
				
				if(newInstancesSet.isPresent()) {
					Set<AnnotatedTableId> newInstanceTableCount = new HashSet<>();
					for(TableRowId rowId : clusters.getA().getCluster(newInstancesSet.get())) {
						if(tables.containsKey(rowId.getTableId())) {
							newInstanceTableCount.add(rowId.getTableId());
						}
					}
					instanceData.setNewInstanceAdded(new InstancePair(newInstancesSet.get().getInstanceId(), newInstanceTableCount.size()));

				}
				
				List<InstancePair> conflictInstances = new ArrayList<>();
				  choice = 0;
				for( InstanceId  candidate : labelConflictChoices) {
					if(choice < choiceLimit) {
//						Set<AnnotatedTableId> conflictInstanceTableCount = new HashSet<>();
						Map<InstanceId, MutableInt> count = new HashMap<>();
						for(TableRowId rowId : labelConflictFilteredTables.get(candidate).getSecond()) {
							if(tables.containsKey(rowId.getTableId())) {
								Optional<InstanceId> match = tableRepository.getMatchedInstanceId(rowId);
								if(match.isPresent()) {
									CHMHelper.getFromMap(count, match.get(), () -> new MutableInt()).increment();
								}
								else {
									CHMHelper.getFromMap(count, candidate, () -> new MutableInt()).increment();
								} 
							}
						}
						for(Entry<InstanceId, MutableInt> entry : count.entrySet()) {
							if(entry.getValue().intValue() > 0) {
								additionalInstancesSet.add(entry.getKey());
								conflictInstances.add(new InstancePair(entry.getKey().getInstanceId(), entry.getValue().intValue()));
							}
						}
		 
						choice++;
					}
					else {
						break;
					}
				}
				
				instanceData.setHomonymsAdded(conflictInstances);
				
				ClassId classAndQuartile =   ClassId.fromString(classId.getReadableString(ClassIdStyle.DBPEDIA_URI) + "_" + instanceMap.get(e.getKey()),ClassIdStyle.DBPEDIA_URI );
				quartileClassIds.add(classAndQuartile);
				if(newInstancesSet.isPresent()) {
					CHMHelper.getFromMap(newInstances, classId, () -> new HashSet<>()).add(newInstancesSet.get());
				}
				CHMHelper.getFromMap(additionalInstances, classId, () -> new HashSet<>()).addAll(additionalInstancesSet );
				CHMHelper.getFromMap(quartileNewInstances, classAndQuartile, () -> new HashSet<>()).add(newInstancesSet.get());
				CHMHelper.getFromMap(quartileKbInstances, classAndQuartile, () -> new HashSet<>()).add(e.getKey());

				
				for(Entry<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>> table : tables.entrySet()) {
					CHMHelper.getFromMap(tableAndMappings, table.getKey(), () -> new ArrayList<>()).addAll(table.getValue());
					Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>> innerMap = CHMHelper.getFromMap(quartileMap, classAndQuartile, () -> new HashMap<>());
					
					CHMHelper.getFromMap(innerMap, table.getKey(), () -> new ArrayList<>()).addAll(table.getValue());

				}
				instanceDataList.put(new InstanceId(e.getKey().getInstanceId()).getReadableString(), instanceData);
			}
			 
			for(ClassId quartileClassId : quartileClassIds) {
				quartileKbMappingsMap.put(quartileClassId, mapping.getC());
				quartileClusterMappingsMap.put(quartileClassId, clusters.getA());
			}
			

			finalGs.put(classId, tableAndMappings);
			
			quartileGs.putAll(quartileMap);
			
			instanceJson.put(classId, instanceDataList);
			
			}
		}
		
		if(false) {

		printStatisticsForTable(finalGs,kbInstances,newInstances,kbMappingsMap,clusterMappingsMap);
		printStatisticsForTable(quartileGs,quartileKbInstances,quartileNewInstances,quartileKbMappingsMap,quartileClusterMappingsMap);

		printStatisticsForAttributes(finalGs.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().keySet())),kbInstances,newInstances,kbMappingsMap,clusterMappingsMap);
		{
			
			for(Entry<ClassId, Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>>> gsEntry : finalGs.entrySet()) {
				
				
				ConcurrentHashMap<PropertyId, Set<InstanceId>> propertyCountMap = new ConcurrentHashMap<>();
				Set<InstanceId> instanceIds = new HashSet<>();
				ConcurrentHashMap<InstanceId, Set<PropertyId>> instanceCountMap = new ConcurrentHashMap<>();
				for(AnnotatedTableId tableId : gsEntry.getValue().keySet()) {
					
					List<TableRowId> rowIds = tableRepository.getRowIds(tableId);
					Set<PropertyId> propertyIds = new HashSet<>(tableRepository.getMetaData(tableId).getProperties().values());
					for(TableRowId tableRowId : rowIds) {
						InstanceId instanceId = kbMappingsMap.get(gsEntry.getKey()).getInstanceId(tableRowId);
						if(instanceId == null) {
							instanceId = clusterMappingsMap.get(gsEntry.getKey()).getInstanceId(tableRowId);
							
						}
						
						if(instanceId != null) {
							instanceIds.add(instanceId);
							for(PropertyId propertyId : propertyIds) {
								CHMHelper.getFromCHM(propertyCountMap, propertyId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(instanceId);
								CHMHelper.getFromCHM(instanceCountMap, instanceId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(propertyId);

							}
						}
					}
				}
				
				ConcurrentAccumulator<AccumAverage> averagePropertyPerInstance = new ConcurrentAccumulator<>(AccumAverage.empty());
				ConcurrentAccumulator<AccumMedian> medianPropertyPerInstance = new ConcurrentAccumulator<>(AccumMedian.empty());

				for(Set<PropertyId> set : instanceCountMap.values()) {
					averagePropertyPerInstance.add(new AccumAverage((double) set.size()));
					medianPropertyPerInstance.add(new AccumMedian((double) set.size()));
				}
				
				Map<Integer, Long> groups =new TreeMap<>( instanceCountMap.entrySet().stream().collect(Collectors.groupingBy((e) -> e.getValue().size(),Collectors.counting())));
				double average = averagePropertyPerInstance.get().get();
				double median = medianPropertyPerInstance.get().get();
				
				Map<PropertyId, Integer> propertyCountCountMap = propertyCountMap.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().size()));
				
				int fsdf = 123;
			
			}
		
			
		}
		goldStandardWriter.writeGoldStandard(finalGs, kbInstances, newInstances, kbMappingsMap, clusterMappingsMap, kbInstances, newInstances, additionalInstances,gsPath);
		
		}
		
		
		goldStandardWriter.writeClassesFile(fullTableSet, gsPath + "/full_table_set.csv");
		
		try {
		   Gson gson = new Gson();
		   FileUtils.writeStringToFile(new File(gsPath + "/instances.json"), gson.toJson(instanceJson));
		}
		catch(IOException e) {
			e.printStackTrace();
		}

	}
	
	private <E> Map<Quartile, NavigableSet<E>> selectEquidistantInstances(int topK, QuartileData<E> qd, QuartileFilter<E> filter) {
		
		Map<Quartile, NavigableSet<E>> ret = new HashMap<>();
		
		for(  Entry<Quartile, NavigableSet<E>> e : qd.getSelectedPerQuartile().entrySet()) {
			
			NavigableSet<E> filteredSet;
			
			if(filter != null) {
				filteredSet = filter.filter(e.getValue());
			}
			else {
				filteredSet  = e.getValue();
			}
			
			{
				NavigableSet<E> newSet = new TreeSet<>(qd.getComparator());
				int i = 0; 
				int size = filteredSet.size();
				int distance = (size - topK) / (topK + 1);
				int moved = 0;
				for(E instanceId : filteredSet.descendingSet()) {
					if(i == topK)
						break;
					
					else if(moved == distance) {
						newSet.add(instanceId);
						moved = 0;
						i++;
					}
					else {
						moved++;
					}
				}
				ret.put(e.getKey(), newSet);
			}
		}

		return ret;
	}
	
	


	
	public boolean cacheExists(String cacheFolder, String PREFIX) {
		File dirFile = new File(cacheFolder);
		if (dirFile.isDirectory()) {
			String[] dirFileList = dirFile.list();
			if (dirFileList.length > 0) {
				for (String f : dirFileList) {
					if (f.startsWith(PREFIX)) {
						return true;
					}
				}

			}
		}
		return false;
	}
	
	public InstanceId getInstanceIdObject(Entity e) {
		String mainLabel = e.getMainLabel();
		String instanceId;
		if(mainLabel == null) {
			mainLabel = e.getRowClusterUUID().toString();
			instanceId = "http://dbpedia.org/resource/NO_LABEL-" + e.getRowClusterUUID().toString();
		}
		else {
			instanceId = "http://dbpedia.org/resource/" + mainLabel.replace(" ", "_") + "-" + e.getRowClusterUUID().toString();
		}
			 
		InstanceId instanceIdObject = new InstanceId(instanceId, mainLabel);
		return instanceIdObject;
	}
	
	
	public  Tri<TableRowInstanceBiMap, TableRowInstanceBiMap,Map<InstanceId, Map<InstanceId, Double>>> getRowClusters(String cacheFolder, ClassId classId,  Set<ClassId> parents, TableIdsSet tableIdsSet, FusionConfig fusionConfig ,  SetExpansionStrategiesConfig setExpansionStrategiesConfig, String thresholdsFolder, StrategiesConfig strategiesConfig)   {
		if (true) {
			throw new IllegalStateException();
		}
		
		 try {
			 String clusterCacheFolder = cacheFolder + "/cluster_cache/";
			 String remainingEntitiesCacheFolder = cacheFolder + "/remaining_entities/";
			 String labelConflictCacheFolder = cacheFolder + "/label_conflict_entities/";
			 
			 MKDIR.createDirUnchecked(new File(clusterCacheFolder));
			 MKDIR.createDirUnchecked(new File(remainingEntitiesCacheFolder));
			 MKDIR.createDirUnchecked(new File(labelConflictCacheFolder));
			 
		Map<RowClusterUUID, List<TableRowId>> rowCluster;
		 String PREFIX = "CLUSTER_CACHE_" + classId.getFinalPart() + "_";
		 String ENTITIES_PREFIX = "ENTITIES_" + classId.getFinalPart() + "_";
		 
		 SplitGroupedValues splitGroupedValues = setExpansionRunner.getSplitGroupedValues(fusionConfig, tableIdsSet);
		 
			Map<ClassId, Set<InstanceId>> instancesWithMatches = setExpansionRunner.getInstancesWithMatches(classId, tableIdsSet.getTableIds(), 1);

			
		 SplitGroupedValuesSets splitGroupedValuesSets = entitySetGroupedValuesSplitter.split(splitGroupedValues,
					new EntitySets(new HashSet<>(instancesWithMatches.get(classId)), new HashSet<>()));
		 
		 List<ThresholdContainer> thresholdContainers = setExpansionRunner.getThresholdContainers(setExpansionStrategiesConfig, thresholdsFolder, classId, fusionConfig, strategiesConfig, splitGroupedValues);

		 
		 if(cacheExists(clusterCacheFolder, PREFIX)) {
			 rowCluster = chmKryoAdapter.readWithListWithReferences(RowClusterUUID.class, TableRowId.class, clusterCacheFolder, PREFIX);
		 }
		 else {
			

			 
			 
				Set<TableRowId> allRows = tableRepository.getAllRowIds(tableIdsSet.getTableIds());//.getUnmatchedRowIdsForTables(tableIdsSet.getTableIds());

			 
				if(instancesWithMatches.get(classId).isEmpty()) {
					System.out.println("No Instances with MinSource found for: " + classId.getClassId());
					return null;
				}
				
				
		

				
				ClustererInput clustererInput = new ClustererInput(splitGroupedValuesSets, allRows, tableIdsSet, null, classId);
				 
				RowClusteringConfig rowClusteringConfig = null;// new
																// RowClusteringConfig(setExpansionStrategiesConfig,
																// thresholdContainers);


//					RowClusterer clusterer = rowClustererRepository.get(strategy.getClusterer());
					
//					RowClusterSimilarities sim = clusterer.clusterRows(rowClusteringConfig, clustererInput);
					 
				ClusterList clusters = labelSimilarityClusterer.getDirect(clustererInput.getUuid(), clustererInput.getCandidates());
				 rowCluster = clusters.getClusterList().stream().collect(Collectors.toMap((e) -> e.getUuid(), (e) -> e.getRows() ));
					
					 
					
					chmKryoAdapter.writeWithListWithReferences(clusterCacheFolder, PREFIX , rowCluster);
					

		 }
		 
		 Map<RowClusterUUID, List<TableRowId>> rowClusterNew = new HashMap<>();
		 for(Entry<RowClusterUUID, List<TableRowId>> rowClusterEntry : rowCluster.entrySet()) {
			 List<TableRowId> newList = rowClusterEntry.getValue().stream().filter((e) -> tableIdsSet.getTableIds().contains(e.getTableId())).collect(Collectors.toList());
			 if(newList.size() > 0) {
				 rowClusterNew.put(rowClusterEntry.getKey(), newList);
			 }
		 }
		 rowCluster = rowClusterNew;
		 
			List<Entity> entities = null; // entityCreator.createEntites(splitGroupedValuesSets,
											// thresholdContainers,
											// fusionConfig,
			// strategiesConfig,
			// setExpansionStrategiesConfig,
			// rowCluster.entrySet().stream().map((e) -> new
			// RowCluster(e.getKey(), classId, e.getValue(),
			// null)).collect(Collectors.toList()));
			
			entities = entities.stream().filter((e) -> e.getMainLabel().trim().length() > 1).collect(Collectors.toList());

			Map<RowClusterUUID, Entity> entitiesMap = entities.stream().collect(Collectors.toMap(Entity::getRowClusterUUID, Function.identity()));
			
			Map<ClassId, Set<ClassId>> classWithParents = new HashMap<>();
			classWithParents.put(classId,parents);
			
			List<Entity> entitiesUnmatchedOnly = null;// entityCreator.createEntites(splitGroupedValuesSets,
														// thresholdContainers,
														// fusionConfig,
			// strategiesConfig,
			// setExpansionStrategiesConfig,
			// rowCluster.entrySet().stream().map((e) -> { return new
			// RowCluster(e.getKey(), classId,
			//
			//
			// e.getValue().stream().filter((rowId) ->
			// !tableRepository.getMatchedInstanceId(rowId).isPresent()).collect(Collectors.toList())
			// , null);
			// }).filter((e) -> {return
			// !e.getRows().isEmpty();}).collect(Collectors.toList())).stream().filter((e)
			// -> e.getMainLabel().trim().length() >
			// 1).collect(Collectors.toList());
			
			entitiesUnmatchedOnly = entities;
			
			final Set<RowClusterUUID> entitiesRemaining;
			if(cacheExists(remainingEntitiesCacheFolder, ENTITIES_PREFIX)){
				entitiesRemaining = chmKryoAdapter.readWithReferences(RowClusterUUID.class, remainingEntitiesCacheFolder, ENTITIES_PREFIX);
			}
			else {
//				entities.stream().map((e) -> new Entity(e.getRowClusterUUID(), e.getClassId(), e.get, values, rowCluster))
				
				
				
				
				
				
				Pair<Map<InstanceId, List<ScoredEntity>>, List<Entity>> mappedEntities = entityMatcher.matchMultipleAndUnsure(entitiesUnmatchedOnly, classWithParents);
				
				 entitiesRemaining = entitiesUnmatchedOnly.stream().map(Entity::getRowClusterUUID).collect(Collectors.toSet());
	//			Map<InstanceId, Entity> mapping = clusterListEvaluator.maxMajorityMappings(entities); 
				
				for(Entry<InstanceId, List<ScoredEntity>> entry : mappedEntities.getFirst().entrySet()) {
					for(ScoredEntity se : entry.getValue()) {
						entitiesRemaining.remove(se.getEntity().getRowClusterUUID());
					}
				}
				
				for(Entity se : mappedEntities.getSecond()) {
					entitiesRemaining.remove(se.getRowClusterUUID());
				}
				
				chmKryoAdapter.writeWithRefrences(remainingEntitiesCacheFolder, ENTITIES_PREFIX, entitiesRemaining, new RowClusterUUID());
			}
			
			Map<RowClusterUUID, Entity> entitiesRemainingMap =entitiesUnmatchedOnly.stream().filter((e) -> entitiesRemaining.contains(e.getRowClusterUUID())).collect(Collectors.toMap((e) -> e.getRowClusterUUID(), Function.identity()));
			
			Map<InstanceId, Integer> entitiesFactCount = new HashMap<>();
			
			Map<RowClusterUUID, InstanceId  > entitiesKey = entities.stream().collect(Collectors.toMap((e) -> e.getRowClusterUUID(), (e) -> {
				InstanceId instanceIdObject = getInstanceIdObject(e);
				entitiesFactCount.put(instanceIdObject, e.getValues().size());
				return instanceIdObject;
			}));
				
			for(Entry<RowClusterUUID, Entity> entry : entitiesRemainingMap.entrySet()) {
				if(!entitiesKey.containsKey(entry.getKey())) {
					entitiesKey.put(entry.getKey(), getInstanceIdObject(entry.getValue()));
				}
			}
			
			
			Map<InstanceId, RowClusterUUID  > entitiesKeyReverse = entitiesKey.entrySet().stream().collect(Collectors.toMap(Entry::getValue, Entry::getKey)); 
			
//			entities = new ArrayList<>(entitiesRemaining.values());
			

			Map<InstanceId, Integer> entitiesRemainingFactCount = entitiesRemainingMap.entrySet().stream().collect(Collectors.toMap((e)-> {
				 InstanceId key = entitiesKey.get(e.getKey());
				 if(key == null) {
					 int sdf = 123;
				 }
				return key;
				}, (e) -> {
					
				return e.getValue().getValues().size();
			}));

			
			for(Iterator<RowClusterUUID> it = entitiesRemaining.iterator(); it.hasNext();) {
				RowClusterUUID next = it.next();
				if(!entitiesKey.containsKey(next)) {
					it.remove();
					continue;
				}
				if(!entitiesRemainingMap.containsKey(next)) {
					it.remove();
					continue;
				}
			}
			
			int sizeOfMap = entitiesRemaining.size();
			
			Map<InstanceId,  Entity > entitiesRemainingInstanceIdWithEntity = entitiesRemaining.stream().collect(Collectors.toMap((e) -> {
				return entitiesKey.get(e);
			}, (e) -> entitiesRemainingMap.get(e)));
			
			Map<InstanceId, Set<TableRowId>> entitiesRemainingInstanceId = entitiesRemainingInstanceIdWithEntity.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				return new HashSet<>(e.getValue().getRowCluster().getRows());
			}));
			
			
			Map<TableRowId, InstanceId> clusterMapping = new HashMap<>();
			Map<InstanceId, Set<TableRowId>> clusterMappingReverse = entities.stream().collect(Collectors.toMap((e) -> entitiesKey.get(e.getRowClusterUUID()), (e) -> new HashSet<>(e.getRowCluster().getRows())));
			Map<InstanceId, Integer> clusterMappingFactCount = entitiesFactCount;
			for(Entry<InstanceId, Set<TableRowId>> clusterMappingEntry : clusterMappingReverse.entrySet()) {
				for(TableRowId rowId : clusterMappingEntry.getValue()) { 
					clusterMapping.put(rowId,  clusterMappingEntry.getKey());
				}
			}
 
			final Map<InstanceId, Map<InstanceId, Double>> labelConflictEntities;
		 
			if(cacheExists(labelConflictCacheFolder, ENTITIES_PREFIX)){
				labelConflictEntities = chmKryoAdapter.readWithMapWithReferences(InstanceId.class, InstanceId.class, Double.class, labelConflictCacheFolder, ENTITIES_PREFIX);
			}
			else {
			 labelConflictEntities = entityMatcher.matchConflict(entities.stream().collect(Collectors.toMap((e) -> entitiesKey.get(e.getRowClusterUUID()), Function.identity())), classWithParents);
			 
			 chmKryoAdapter.writeWithMapWithReferences(labelConflictCacheFolder, ENTITIES_PREFIX, labelConflictEntities);
				Map<InstanceId, Map<InstanceId, Double>> test = chmKryoAdapter.readWithMapWithReferences(InstanceId.class, InstanceId.class, Double.class, labelConflictCacheFolder, ENTITIES_PREFIX);

			 if(test.equals(labelConflictEntities)) {
				 System.out.println("TEST SUCCESFUL");
			 }
			}
			
			
			Map<TableRowId, InstanceId> remainingInstanceMapping = new HashMap<>(); 
			for( Entry<InstanceId, Set<TableRowId>> entry : entitiesRemainingInstanceId.entrySet()) {

				for(TableRowId rowId : entry.getValue()) {
					remainingInstanceMapping.put(rowId,  entry.getKey());
					
				}
			}
		 
		 return new Tri<>( new TableRowInstanceBiMap(remainingInstanceMapping,entitiesRemainingInstanceId, entitiesRemainingFactCount),new TableRowInstanceBiMap(clusterMapping, clusterMappingReverse, clusterMappingFactCount), labelConflictEntities);
		 }
		 catch(IOException e) {
			 e.printStackTrace();
			 return null;
		 }
	}
	
	
	private Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>> findBestTables(InstanceId instanceId, Set<TableRowId> rows, Random random, TableRowInstanceBiMap tableRowInstanceBiMap, TableRowInstanceBiMap tableRowInstanceBiMapFull, MutableReference<InstanceId> subInstances,
			Set<InstanceId> newInstancesSetSet) {
		Set<AnnotatedTableId> candidates = new HashSet<>();
		Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>> unfilteredMappings = new HashMap<>();
		
		for(TableRowId rowEntry : rows) {
			candidates.add(rowEntry.getTableId());
			CHMHelper.getFromMap(unfilteredMappings, rowEntry.getTableId(), () -> new ArrayList<>()).add(new Pair<>(rowEntry, instanceId));
		}
		
		Set<AnnotatedTableId> selectedTables = findBestTables(candidates,   random, tableRowInstanceBiMapFull,MAX_ROWS_PER_INSTANCE);
		
		
		Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>> filteredMappings = unfilteredMappings.entrySet().stream().filter((e) -> selectedTables.contains(e.getKey())).collect(Collectors.toMap(Entry::getKey, Entry::getValue));

		
		

		Map<InstanceId, Set<AnnotatedTableId>> mostCommonClusterInstanceId = new HashMap<>();
		
		
		
		for(AnnotatedTableId tableId : selectedTables) {
			
			List<TableRowId> rowIds = tableRepository.getRowIds(tableId);
			for(TableRowId rowId : rowIds) {
				Optional<InstanceId> matchedInstances = tableRepository.getMatchedInstanceId(rowId);
				if(!matchedInstances.isPresent()) { 
					InstanceId clusterInstanceId = tableRowInstanceBiMap.getInstanceId(rowId);

					if(clusterInstanceId != null && !newInstancesSetSet.contains(clusterInstanceId)) {
						CHMHelper.getFromMap(mostCommonClusterInstanceId, clusterInstanceId, () -> new HashSet<>()).add(tableId);
						 
					}
					 
				}
	 
			}

			
		}
		
		if(mostCommonClusterInstanceId.isEmpty()) {
			int sdf= 123;
		}
		else {
 
		double maxFactCount = mostCommonClusterInstanceId.entrySet().stream().mapToDouble((e) -> Double.valueOf( tableRowInstanceBiMap.getFactCount(e.getKey()))).max().getAsDouble();
		double maxRowCount = mostCommonClusterInstanceId.entrySet().stream().mapToDouble((e) -> Double.valueOf( e.getValue().size())).max().getAsDouble();
		
		double maxClusterSize = mostCommonClusterInstanceId.entrySet().stream().mapToDouble((e) -> Double.valueOf(tableRowInstanceBiMap.getCluster(e.getKey()).size())).max().getAsDouble();
		
		List<InstanceId> secondarySorting = mostCommonClusterInstanceId.keySet().stream().sorted((e1,e2) -> e1.getInstanceId().compareTo(e2.getInstanceId())).collect(Collectors.toList());
		Collections.shuffle(secondarySorting, random);
		
		Map<InstanceId, Integer> secondarySortingMap = new HashMap<>();
		for(ListEntry<InstanceId> entry : ListIterable.iterable(secondarySorting)) {
			secondarySortingMap.put(entry.getValue(), entry.getIndex());
		}
		
		Optional<Entry<InstanceId, Set<AnnotatedTableId>>> mostCommonInstanceId = mostCommonClusterInstanceId.entrySet().stream().sorted((e1, e2) -> {
			double e1Score =  (  ( tableRowInstanceBiMap.getFactCount(e1.getKey()) / maxFactCount) + (e1.getValue().size() / maxRowCount)  ) / 2.0;
			double e2Score =  (  ( tableRowInstanceBiMap.getFactCount(e2.getKey()) / maxFactCount) + (e2.getValue().size() / maxRowCount)  ) / 2.0;
			
//			e1Score = e1.getValue().size() / maxRowCount;
//			e2Score = e2.getValue().size() / maxRowCount;
			int compare = Double.compare(e2Score, e1Score);
			if(compare != 0)
				return compare;
			else  {
 
//				  e1Score =  (  ( tableRowInstanceBiMap.getFactCount(e1.getKey()) / maxFactCount) + (tableRowInstanceBiMap.getCluster(e1.getKey()).size() / maxRowCount)  ) / 2.0;
//				  e2Score =  (  ( tableRowInstanceBiMap.getFactCount(e2.getKey()) / maxFactCount) + (tableRowInstanceBiMap.getCluster(e2.getKey()).size() / maxRowCount)  ) / 2.0;
				
//				e1Score = tableRowInstanceBiMap.getCluster(e1.getKey()).size();
//				e2Score = tableRowInstanceBiMap.getCluster(e2.getKey()).size();
//				compare = Double.compare(e2Score, e1Score);
				if(compare != 0)
					return compare;
				else
					return Integer.compare(secondarySortingMap.get(e1.getKey()), secondarySortingMap.get(e2.getKey()));
//					return e1.getKey().getInstanceId().compareTo(e2.getKey().getInstanceId());
	 
			}
		}
		).findFirst();
		
		if(mostCommonInstanceId.isPresent()) {
			newInstancesSetSet.add(mostCommonInstanceId.get().getKey());
			subInstances.set(mostCommonInstanceId.get().getKey());
			Set<AnnotatedTableId> candidatesNew = tableRowInstanceBiMap.getCluster(mostCommonInstanceId.get().getKey()).stream().map(TableRowId::getTableId).filter((e) -> !selectedTables.contains(e)).collect(Collectors.toSet());
			
			Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>> cluterFiltered = new HashMap<>();
			
			for(TableRowId rowId : tableRowInstanceBiMap.getCluster(mostCommonInstanceId.get().getKey())) {
				CHMHelper.getFromMap(cluterFiltered, rowId.getTableId(), () -> new ArrayList<>()).add(new Pair<>(rowId, mostCommonInstanceId.get().getKey()));

			}
			
			Set<AnnotatedTableId> clusterTables = findBestTables(candidatesNew,random, tableRowInstanceBiMapFull, MAX_ROWS_PER_INSTANCE);
			for(AnnotatedTableId clusterTable : clusterTables) {
				
				
				
				//CHMHelper.getFromMap(filteredMappings, clusterTable, () -> new ArrayList<>()).addAll(cluterFiltered.get(clusterTable));
				CHMHelper.getFromMap(filteredMappings, clusterTable, () -> new ArrayList<>());
			}
		}
		
		else {
			System.out.println("THERE IS REALLY NO NEW ROWS?");
		}
		}
		
		return filteredMappings;
	}


	private Set<AnnotatedTableId> findBestTables(Set<AnnotatedTableId> candidates,  Random random, TableRowInstanceBiMap tableRowInstanceBiMap, int maxRowsPerInstance) {
		 		
		
		Map<AnnotatedTableId, Set<PropertyId>> tableToPropertiesMutable = new HashMap<>();
		Map<PropertyId, Set<AnnotatedTableId>> propertiesToTable = new HashMap<>();
		
				
		Map<String, Set<AnnotatedTableId>> uniquePLD = new HashMap<>(); 
		Map<AnnotatedTableId, String> reversePLD = new HashMap<>();
		for(AnnotatedTableId candidate : candidates) {
			Set<PropertyId> propertyIds = new HashSet<>(tableRepository.getMetaData(candidate).getProperties().values());
			tableToPropertiesMutable.put(candidate, propertyIds);
			
			for(PropertyId propertyId : propertyIds) {
				CHMHelper.getFromMap(propertiesToTable, propertyId, () -> new HashSet<>()).add(candidate);
			}
			
			String PLD = tableRepository.getMetaData(candidate).getSource().getPLD();
			CHMHelper.getFromMap(uniquePLD, PLD, () -> new HashSet<>()).add(candidate);
			reversePLD.put(candidate, PLD);
		}
		

		MutableInt remainingPerInstance = new MutableInt(Math.min(Math.min(candidates.size(), Math.max(uniquePLD.size(), propertiesToTable.size())), maxRowsPerInstance));
		
		List<AnnotatedTableId> secondarySorting = candidates.stream().sorted((e1,e2) -> e1.getTableName().compareTo(e2.getTableName())).collect(Collectors.toList());
		Collections.shuffle(secondarySorting, random);
		Map<AnnotatedTableId, Integer> tableIdsSortingIndex = new HashMap<>();
		for(ListEntry<AnnotatedTableId> entry : ListIterable.iterable(secondarySorting)) {
			tableIdsSortingIndex.put(entry.getValue(), entry.getIndex());
		}
		
		

		Map<InstanceId, MutableInt> instanceScores = new HashMap<>();
		for(AnnotatedTableId tableId : candidates) {
	 
			
			List<TableRowId> rowIds = tableRepository.getRowIds(tableId);
			for(TableRowId rowId : rowIds) {
				Optional<InstanceId> matchedInstances = tableRepository.getMatchedInstanceId(rowId);
				if(matchedInstances.isPresent()) {
					CHMHelper.getFromMap(instanceScores, matchedInstances.get(), () -> new MutableInt()).increment();
				}
				else  {
					InstanceId clusterInstanceId = tableRowInstanceBiMap.getInstanceId(rowId);
					if (tableRowInstanceBiMap != null) {
						CHMHelper.getFromMap(instanceScores, clusterInstanceId, () -> new MutableInt()).increment();
	 
					}
				}
			}
 
				
		}
		
		Map<InstanceId, Set<AnnotatedTableId>> instanceIdReverseIndex = new HashMap<>();
		Map<AnnotatedTableId, Double> tableScores = new HashMap<>();
		
		
		for(AnnotatedTableId tableId : candidates) {
			double kbScore = 0.0;
			
			double clusterScore = 0.0;
			
			List<TableRowId> rowIds = tableRepository.getRowIds(tableId);
			for(TableRowId rowId : rowIds) {
				Optional<InstanceId> matchedInstances = tableRepository.getMatchedInstanceId(rowId);
				if(matchedInstances.isPresent()) {
					kbScore += instanceScores.get(matchedInstances.get()).doubleValue();
					CHMHelper.getFromMap(instanceIdReverseIndex, matchedInstances.get(), () -> new HashSet<>()).add(tableId);

				}
				else  {
					InstanceId clusterInstanceId = tableRowInstanceBiMap.getInstanceId(rowId);
					if(clusterInstanceId != null) {
						clusterScore += instanceScores.get(clusterInstanceId).doubleValue();
						CHMHelper.getFromMap(instanceIdReverseIndex, clusterInstanceId, () -> new HashSet<>()).add(tableId);
					}
				}
	 
			}
			
			double proportionKB = kbScore / (kbScore + clusterScore);
			double proportionCluster = clusterScore / (kbScore + clusterScore);
			
			double weightKB = 1.0 + (0.6 - proportionKB);
			double weightCluster = 1.0 + (0.4 - proportionCluster);
			
			double score = ((kbScore * weightKB) + (clusterScore * weightCluster) ) / (weightKB + weightCluster);
			
//			double score = (1.0 - Math.abs(proportionKB - 0.7)) * kbScore + (1.0 - Math.abs(proportionCluster - 0.3)) *clusterScore; 
//			double score = (kbScore * finalClusterScore + clusterScore * finalKbScore) / rowIds.size(); 
//			double score = ((kbScore + clusterScore) / 2.0) /rowIds.size();
			score = score / rowIds.size();
			tableScores.put(tableId, score);
				
		}
		
		


		NavigableSet<AnnotatedTableId> sortedTables = new TreeSet<>((t1, t2) -> {
			int compare = Integer.compare( tableToPropertiesMutable.get(t1).size(), tableToPropertiesMutable.get(t2).size());
			if(compare != 0) {
				return compare;
			}
			else {
				compare = Double.compare(tableScores.get(t1), tableScores.get(t2));
				if(compare != 0)
					return compare;
				else
					return Integer.compare(tableIdsSortingIndex.get(t1), tableIdsSortingIndex.get(t2));
			}
		});
		
		sortedTables.addAll(tableScores.keySet());
		
		Set<AnnotatedTableId> selectedTables = new HashSet<>();
		

		
		while(remainingPerInstance.intValue() > 0) {
			AnnotatedTableId tableId = sortedTables.descendingSet().first();
			selectedTables.add(tableId);
			
			for(PropertyId propertyId : tableToPropertiesMutable.get(tableId)) {
				for(AnnotatedTableId innerTableId : propertiesToTable.get(propertyId)) {
					if(!innerTableId.equals(tableId)) {
						tableToPropertiesMutable.get(innerTableId).remove(propertyId);
					}
				}
				
			}
			List<TableRowId> rowIds = tableRepository.getRowIds(tableId);
			for(TableRowId rowId : rowIds) {
				Optional<InstanceId> matchedInstances = tableRepository.getMatchedInstanceId(rowId);
				if(matchedInstances.isPresent()) {
					for(AnnotatedTableId innerTableId : instanceIdReverseIndex.get(matchedInstances.get())) {
						Double d = tableScores.get(innerTableId);
//						d += 1.0;
						tableScores.put(innerTableId, d);
					}
				}
				else {
					InstanceId clusterInstanceId = tableRowInstanceBiMap.getInstanceId(rowId);

					if(clusterInstanceId != null) { 
						for(AnnotatedTableId innerTableId : instanceIdReverseIndex.get(clusterInstanceId)) {
							Double d = tableScores.get(innerTableId);
//							d += 1.0;
							tableScores.put(innerTableId, d);
						}
					}
					 
				}
				
				
	 
			}
			
			for(AnnotatedTableId innerTableId : uniquePLD.get(reversePLD.get(tableId))) {
				
				tableScores.put(innerTableId, -1.0);
			}
			
			sortedTables.clear();
			sortedTables.addAll(tableScores.keySet());
			sortedTables.removeAll(selectedTables);
			
			remainingPerInstance.decrement();
			
		}
		
	
		
 
		 return selectedTables;
		
	}
	
	
 


		// BOTH INCLUSIVE
		public static int randInt(Random random, int min, int max) {

		    // BOTH INCLUSIVE
		    int randomNum = random.nextInt((max - min) + 1) + min;

		    return randomNum;
		}

	public TableRowInstanceBiMap getKBBiMap(TableIdsSet tableIdSet, Set<InstanceId> instancesFilter) {
		Set<TableRowId> rowIds = tableRepository.getMatchedRowIdsForTables(tableIdSet.getTableIds());

		ConcurrentHashMap<InstanceId, Set<TableRowId>> instanceToRowMapping = new ConcurrentHashMap<>();


		ConcurrentHashMap<TableRowId, InstanceId> rowToInstanceMapping = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(rowIds, (rowId) -> {
			Optional<InstanceId> matchedInstance = tableRepository.getMatchedInstanceId(rowId);
			if (matchedInstance.isPresent() && (instancesFilter == null || instancesFilter.contains(matchedInstance.get()))) {
				CHMHelper.getFromCHM(instanceToRowMapping, matchedInstance.get(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(rowId);
			}
			rowToInstanceMapping.put(rowId, matchedInstance.get());

		}, this.getClass());

		// degreeMapping =
		// instanceToRowMapping.entrySet().stream().collect(Collectors.toMap(Entry::getKey,
		// (e) -> e.getValue().size()));

		Map<InstanceId, Integer> factCount = instanceToRowMapping.keySet().stream()
				.collect(Collectors.toMap(Function.identity(), (e) -> tableRepository.getMatches(e).size()));

		return new TableRowInstanceBiMap(rowToInstanceMapping, instanceToRowMapping, factCount);

	}

	public Tri<TreeMap<InstanceId, Set<TableRowId>>, Map<InstanceId, Quartile>, TableRowInstanceBiMap> getMappings(ClassId classId, Random random,
			TableIdsSet tableIdSet, int maxIndegree, Set<InstanceId> instancesFilter) {
		Set<TableRowId> rowIds = tableRepository.getMatchedRowIdsForTables(tableIdSet.getTableIds());
		
		TableRowInstanceBiMap biMap = getKBBiMap(tableIdSet, instancesFilter);
		
		Map<InstanceId, Integer> degreeMapping = biMap.getInstanceToCluster().entrySet().stream()
				.filter((e) -> e.getValue().stream().map((e2) -> tableRepository.getMetaData(e2.getTableId()).getSource().getPLD()).collect(Collectors.toSet())
						.size() >= MIN_ROWS_PER_CLUSTER_INSTANCE && degreeRepository.getNumIndegree(e.getKey()) < maxIndegree
						&& this.knowledgeBase.getClasses(e.getKey()).contains(classId))
				.collect(Collectors.toMap(Entry::getKey, (e) -> degreeRepository.getNumIndegree(e.getKey())));

		
		
		QuartileFilter<InstanceId> filter = (input) -> {
			double average = input.stream().mapToInt((e) -> biMap.getFactCountMap().get(e)).average().getAsDouble();
			TreeSet<InstanceId> newTreeSet = new TreeSet<>(input.comparator());
			for(InstanceId instanceId : input) {
				if (biMap.getFactCount(instanceId) > average) {
					newTreeSet.add(instanceId);
				}
			}
			return newTreeSet;
		};
		
		Pair<TreeMap<InstanceId, Set<TableRowId>>, Map<InstanceId, Quartile>> selections = getMap(degreeMapping, random, classId, "KB",
				(e1, e2) -> e1.getInstanceId().compareTo(e2.getInstanceId()), (instanceId) -> {
					return instanceId.getReadableString();
				}, (instanceId) -> {
					return biMap.getCluster(instanceId);
				}, filter);
		return new Tri<>( 
				selections.getFirst(), selections.getSecond(),
				biMap);
		
	}
	
	<E, F extends Number & Comparable<F>> Pair<TreeMap<E, Set<TableRowId>>, Map<E, Quartile>> getMap(Map<E, F> degreeMapping, Random random, ClassId classId, String suffix, Comparator<E> comparator, Function<E, String> getReadableString, Function<E, Set<TableRowId>> rowsGetter, QuartileFilter<E> filter) { 
		
		QuartileData<E> qd = getQuartileData(degreeMapping, random, comparator);
	 
		printQuartileStatistics(classId.getReadableString(ClassIdStyle.DBPEDIA_URI) + "_" + suffix, qd, rowsGetter);
		Map<Quartile, NavigableSet<E>> selectedInstances = selectEquidistantInstances(INSTANCES_PER_QUARTILE, qd,filter);
	 
		printSelectedInstancesPerQuartile(classId.getReadableString(ClassIdStyle.DBPEDIA_URI) + "_" + suffix, selectedInstances, degreeMapping, getReadableString, rowsGetter);
		
			
		TreeMap<E, Set<TableRowId>> mappings = new TreeMap<>(qd.getComparator());

		Map<E, Quartile> instanceToQuartileMapping = new HashMap<>();
		
		for(Entry<Quartile, NavigableSet<E>> sets : selectedInstances.entrySet()) {
			for(E instanceId : sets.getValue()) {
				mappings.put(instanceId, rowsGetter.apply(instanceId));
				instanceToQuartileMapping.put(instanceId, sets.getKey());
			}
		}
		
		return new Pair<>(mappings, instanceToQuartileMapping);
	}
	
	
	public <E, F extends Number & Comparable<F>> QuartileData<E> getQuartileData(Map<E, F> degreeMapping, Random random, Comparator<E> comparator) {
		List<E> instanceIds = degreeMapping.keySet().stream().sorted(comparator).collect(Collectors.toList());
		Collections.shuffle(instanceIds, random);
		Map<E, Integer> instanceIdsSortingIndex = new HashMap<>();
		for(ListEntry<E> entry : ListIterable.iterable(instanceIds)) {
			instanceIdsSortingIndex.put(entry.getValue(), entry.getIndex());
		}
		 
		
		List<E> sortedList = degreeMapping.entrySet().stream().sorted((e1, e2) -> {
			int compare =  e1.getValue().compareTo(e2.getValue());
			if(compare != 0)
				return compare;
			else{
				return Integer.compare(instanceIdsSortingIndex.get(e1.getKey()), instanceIdsSortingIndex.get(e2.getKey()));
			}
				
		}).map(Entry::getKey).collect(Collectors.toList());
		
		
		
		
	
		QuartileData<E> qd = new QuartileData<>(sortedList, new HashMap<>( degreeMapping));
		return qd;
	}
	
	private <E> Map<Quartile, NavigableSet<E>> selectTopInstances(int topK, QuartileData<E> qd) {
		
		Map<Quartile, NavigableSet<E>> ret = new HashMap<>();
		
		for(  Entry<Quartile, NavigableSet<E>> e : qd.getSelectedPerQuartile().entrySet()) {
			NavigableSet<E> newSet = new TreeSet<>(qd.getComparator());
			int i = 0; 
			for(E instanceId : e.getValue().descendingSet()) {
				if(i == topK)
					break;
				else {
					newSet.add(instanceId);
					i++;
				}
			}
			ret.put(e.getKey(), newSet);
		}

		return ret;
	}

	



	
	
	
	
	
	

	private void printStatisticsForAttributes(Map<ClassId,Set<AnnotatedTableId>> selectedTables, Map<ClassId,Set<InstanceId>> kbInstances, Map<ClassId,Set<InstanceId>> newInstances, Map<ClassId,TableRowInstanceBiMap> kbMappings, Map<ClassId,TableRowInstanceBiMap> clusterMappings ) {
		try {
		StringWriter sw = new StringWriter(); 
		 
			CSVWriter writer = new CSVWriter(sw);
			
			int columns = 14;
			{
				int i = 0;
				String[] line = new String[columns];
				line[i++] = "Class";
				line[i++] = "Attribute";
				line[i++] = "";
				line[i++] = "KbInstances";
				line[i++] = "PotentialFacts"; 
				line[i++] = "TotalValues";
				line[i++] = "";
				line[i++] = "NewInstances";
				line[i++] = "PotentialFacts"; 
				line[i++] = "TotalValues";
				line[i++] = "";
				line[i++] = "TotalInstances";
				line[i++] = "PotentialFacts"; 
				line[i++] = "TotalValues";
				
				writer.writeNext(line);
			}
		
		
		for(ClassId classId : selectedTables.keySet()) {
			
			Set<AnnotatedTableId> tables = selectedTables.get(classId);
			Set<InstanceId> kbInstcs = kbInstances.get(classId);
			Set<InstanceId> newInstcs = newInstances.get(classId);
			TableRowInstanceBiMap kbMap = kbMappings.get(classId);
			TableRowInstanceBiMap clusterMap = clusterMappings.get(classId);
			
			Map<PropertyId, Map<InstanceId,MutableInt>> counting = new HashMap<>();
			
			Set<InstanceId> instanceCounter = new HashSet<>();
			
			for(AnnotatedTableId tableId : tables) {
				Collection<PropertyId> properties = tableRepository.getMetaData(tableId).getProperties().values();
				for(TableRowId rowId : tableRepository.getRowIds(tableId)) {
					InstanceId instanceId = null;
					Optional<InstanceId> kbMatch = tableRepository.getMatchedInstanceId(rowId);
					if(kbMatch.isPresent()) {
						instanceId = kbMatch.get();
					}
					else {
						instanceId = clusterMap.getInstanceId(rowId);
					}
					
					if(instanceId != null) {
						instanceCounter.add(instanceId);
						
						for(PropertyId propertyId : properties) {
							Map<InstanceId, MutableInt> innerMap = CHMHelper.getFromMap(counting,propertyId , () -> new HashMap<>());
							
							CHMHelper.getFromMap(innerMap, instanceId, () -> new MutableInt()).increment();;
							
							
						}
					}
				}
			}
			
			
			for(Entry<PropertyId, Map<InstanceId, MutableInt>> propertyEntry : counting.entrySet()) {
				
				Map<InstanceId, MutableInt> kbFiltered = propertyEntry.getValue().entrySet().stream().filter((e) -> kbInstcs.contains(e.getKey())).collect(Collectors.toMap(Entry::getKey, Entry::getValue));
				Map<InstanceId, MutableInt> newFiltered = propertyEntry.getValue().entrySet().stream().filter((e) -> newInstcs.contains(e.getKey())).collect(Collectors.toMap(Entry::getKey, Entry::getValue));
				
				int i = 0;
				String[] line = new String[columns];
				line[i++] = classId.getReadableString(ClassIdStyle.DBPEDIA_URI);
				line[i++] = propertyEntry.getKey().getPropertyId();
				line[i++] = "";
				line[i++] = String.valueOf(kbInstcs.size());
				line[i++] = String.valueOf(kbFiltered.size()); 
				line[i++] = String.valueOf(kbFiltered.values().stream().mapToInt(MutableInt::intValue).sum());
				line[i++] = "";
				line[i++] = String.valueOf(newInstcs.size());
				line[i++] = String.valueOf(newFiltered.size()); 
				line[i++] = String.valueOf(newFiltered.values().stream().mapToInt(MutableInt::intValue).sum());
				line[i++] = "";
				line[i++] = String.valueOf(instanceCounter.size());
				line[i++] = String.valueOf(propertyEntry.getValue().size()); 
				line[i++] = String.valueOf(propertyEntry.getValue().values().stream().mapToInt(MutableInt::intValue).sum());
				
				writer.writeNext(line);
			}
			
		}
		
		writer.close();
		
		System.out.println(sw.toString());
		}
		catch(IOException e) {
			e.printStackTrace();
		}
	}

	private <E, F extends Number> void printSelectedInstancesPerQuartile(String title, Map<Quartile, NavigableSet<E>> selectedInstances, Map<E, F> degreeMapping, Function<E, String> readableStringGetter, Function<E, Set<TableRowId>> rowsGetter) {
		try {
		StringWriter sw = new StringWriter();
		 sw.write(title + "\n"); 
		 
			CSVWriter writer = new CSVWriter(sw);
			
			int columns = 5;
			{
				String[] line = new String[columns];
				line[0] = "Quartile";
				line[1] = "Entity Name";
				line[2] = "Mapped Rows";
				line[3] = "Mapped Values"; 
				line[4] = "Ranking";
				
				writer.writeNext(line);
			}
			
			
		for(Quartile q : Quartile.values()) { 
			for(E instanceId : selectedInstances.get(q)) {
				Set<TableRowId> tableRows = rowsGetter.apply(instanceId);
				
				String[] line = new String[columns];
				line[0] =  q.name();
				line[1] =  readableStringGetter.apply(instanceId);
				line[2] = String.valueOf(tableRows.size());
				line[3] = String.valueOf(getNumOfMatchedColumns(tableRows));
				line[4] = String.valueOf(degreeMapping.get(instanceId));
				
				writer.writeNext(line);
				 
			}
			
		}
		
		writer.close();
		System.out.println(sw.toString());
		}
		catch(IOException e) {
			e.printStackTrace();
		}
		
	}
	
	public int getNumOfMatchedColumns(Set<TableRowId> tableRowIds) {
		int i = 0;
		for(TableRowId tableRowId : tableRowIds) {
			i += tableRepository.getMetaData(tableRowId.getTableId()).getProperties().size();
		}
		 
		return i; 
	}
	
	private <E>  void printQuartileStatistics(String readableString, QuartileData<E> qd, Function<E, Set<TableRowId>> rowsGetter  )   {
		try {
		StringWriter sw = new StringWriter();
		 sw.write(readableString + "\n"); 
		 
		CSVWriter writer = new CSVWriter(sw);
		
		int columns = 8;
		{
			String[] line = new String[columns];
			line[0] = "Quartile";
			line[1] = "Size";
			line[2] = "Begin";
			line[3] = "End";
			line[4] = "Average Mapped Rows";
			line[5] = "Median Mapped Rows";
			line[6] = "Average Mapped Values";
			line[7] = "Median Mapped Values";
			
			writer.writeNext(line);
		}
		
		for(Quartile q : Quartile.values()) {
			QuartileStatistics statistics = getQuartileStatistics(qd.getSelectedPerQuartile().get(q), rowsGetter);
			String[] line = new String[columns];
			line[0] = q.name();
			line[1] = String.valueOf(statistics.getSize());
			line[2] = DecimalFormatter.format(qd.getBeginning(q));
			line[3] = DecimalFormatter.format(qd.getEnd(q));
			line[4] = DecimalFormatter.format(statistics.getAverageRows());
			line[5] =  DecimalFormatter.format(statistics.getMedianRows());
			line[6] =  DecimalFormatter.format(statistics.getAverageColumnMappings());
			line[7] =  DecimalFormatter.format(statistics.getMedianColumnMappings());
			
			writer.writeNext(line);
		}
		
		writer.close();
		
		System.out.println(sw.toString());
		}
		catch(IOException e) {
			e.printStackTrace();
		}
		
	}
	

	public <E> QuartileStatistics getQuartileStatistics(Set<E> instanceIds, Function<E, Set<TableRowId>> rowsGetter) {
		 
		ConcurrentAccumulator<AccumAverage> averageRows = new ConcurrentAccumulator<>(new AccumAverage());
		ConcurrentAccumulator<AccumAverage> averageColumnMappings = new ConcurrentAccumulator<>(new AccumAverage());

		ConcurrentAccumulator<AccumMedian> medianRows = new ConcurrentAccumulator<>(new AccumMedian());
		ConcurrentAccumulator<AccumMedian> medianColumnMappings = new ConcurrentAccumulator<>(new AccumMedian());
		AtomicInteger totalRows = new AtomicInteger();
		PersistentParallel6.runAndFinish(instanceIds, (instanceId) -> {
			Set<TableRowId> tableRows = rowsGetter.apply(instanceId);
			int rows =tableRows.size();
			totalRows.addAndGet(rows);
			int columns = getNumOfMatchedColumns(tableRows);
			
			averageRows.add(new AccumAverage((double) rows));
			averageColumnMappings.add(new AccumAverage((double) columns));
			
			
			medianRows.add(new AccumMedian((double) rows));
			medianColumnMappings.add(new AccumMedian((double) columns));
		}, this.getClass());
		
		return new QuartileStatistics(instanceIds.size(),totalRows.get(), averageRows.get().get(), averageColumnMappings.get().get(), medianRows.get().get(), medianColumnMappings.get().get());
	}
	
	void printStatisticsForTable(Map<ClassId, Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>>> finalGs, Map<ClassId, Set<InstanceId>> kbInstances, Map<ClassId, Set<InstanceId>> newInstances, Map<ClassId, TableRowInstanceBiMap> kbMappingsMap, Map<ClassId, TableRowInstanceBiMap> clusterMappingsMap) {
		try {
			StringWriter sw = new StringWriter();
			CSVWriter writer = new CSVWriter(sw);
			int columnCount = 37;	
			{
				int i = 0;
				String[] line = new String[columnCount];
				line[i++] = "ClassId";
				line[i++] = "Number of chosen KB Instances";
				line[i++] = "Total rows for KB Instances";
				line[i++] = "Average rows per Instance";
				line[i++] = "Median rows per Instance";
				line[i++] = "Average values per Instance";
				line[i++] = "Median values per Instance"; 
				line[i++] = "Number of chosen Label Clusters";
				line[i++] = "Total rows for Label Clusters";
				line[i++] = "Average rows per Instance";
				line[i++] = "Median rows per Instance";
				line[i++] = "Average values per Instance";
				line[i++] = "Median values per Instance";
				line[i++] = "Number of Tables";
				line[i++] = "Total Rows";
				line[i++] = "Average Rows";
				line[i++] = "Median Rows";
				line[i++] = "Rows with 1 KB-Match";
				line[i++] = "Rows with 1 KB-Match (%)";
				line[i++] = "Rows with 2 KB-Match";
				line[i++] = "Rows with 2 KB-Match (%)";
				line[i++] = "Rows with 3 KB-Match";
				line[i++] = "Rows with 3 KB-Match (%)";
				line[i++] = "Rows with 5 KB-Match"; 
				line[i++] = "Rows with 5 KB-Match (%)"; 
				line[i++] = "Rows with 10 KB-Match";
				line[i++] = "Rows with 10 KB-Match (%)";
				
				

				line[i++] = "Rows with 1 Any-Match";
				line[i++] = "Rows with 1 Any-Match (%)";
				line[i++] = "Rows with 2 Any-Match";
				line[i++] = "Rows with 2 Any-Match (%)";
				line[i++] = "Rows with 3 Any-Match";
				line[i++] = "Rows with 3 Any-Match (%)";
				line[i++] = "Rows with 5 Any-Match";
				line[i++] = "Rows with 5 Any-Match (%)";
				line[i++] = "Rows with 10 Any-Match";
				line[i++] = "Rows with 10 Any-Match (%)";
				 
				
				writer.writeNext(line);
				
			}
			
			
			for(Entry<ClassId, Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>>> entry : finalGs.entrySet()) {
				
				ConcurrentAccumulator<AccumMedian> medianRows = new ConcurrentAccumulator<AccumMedian>(AccumMedian.empty());
				ConcurrentAccumulator<AccumAverage> averageRows = new ConcurrentAccumulator<AccumAverage>(AccumAverage.empty());
				
				
				Set<InstanceId> kbInstancesPerClass = kbInstances.get(entry.getKey());
				Set<InstanceId> newInstancesPerClass = newInstances.get(entry.getKey());
				TableRowInstanceBiMap kbMap = kbMappingsMap.get(entry.getKey());
				TableRowInstanceBiMap clusterMap = clusterMappingsMap.get(entry.getKey());
				
				Map<InstanceId, Set<AnnotatedTableId>> kbInstancesData = new HashMap<>();
				Map<InstanceId, Set<AnnotatedTableId>> newInstancesData = new HashMap<>();
				
				int totalRows = 0;
				int randomTables = 0;
				Map<TableRowId, Set<TableRowId>> kbMatch = new HashMap<>();
				Map<TableRowId, Set<TableRowId>>  anyMatch = new HashMap<>();
				
				for(Entry<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>> table : entry.getValue().entrySet()) {
					
					if(table.getValue().size() == 0) {
						randomTables++;
					}
					
					AnnotatedTableMetaData metaData = tableRepository.getMetaData(table.getKey());
					
					int start = metaData.getNumHeaderRows();
					int end = metaData.getNumberOfRows() + metaData.getNumHeaderRows();

					for (int i = start; i < end; i++) {
						InstanceId instanceId = metaData.getInstance(i);
						if (instanceId != null) {
							TableRowId tableRowId = new TableRowId(table.getKey(), new RowIdx(i));
							// AnnotatedTable fulTable = this.getTable(tableId);
							// ImmutableMap<RowIdx, AbstractObject> keyColumn =
							// fulTable.getKeyColumn();
							// AbstractObject ao = keyColumn.get(new RowIdx(i));
							 
	 
							Set<TableRowId> set = kbMap.getCluster(instanceId);
							if(set == null) {
								int sdf = 123;
							}
							if(entry.getValue() == null) {
								int sdf= 123;
							}
							set = set.stream().filter((e) -> entry.getValue().containsKey(e.getTableId())).collect(Collectors.toSet());
							set.remove(tableRowId);
	 
							CHMHelper.getFromMap(kbMatch, tableRowId, () -> new HashSet<>()).addAll(set);
							CHMHelper.getFromMap(anyMatch, tableRowId, () -> new HashSet<>()).addAll(set);
							
						}
						else   {
							TableRowId rowId = new TableRowId(table.getKey(), i);
							instanceId = clusterMap.getInstanceId(rowId);
							if(instanceId != null) { 
								Set<TableRowId> set = clusterMap.getCluster(instanceId).stream().filter((e) -> entry.getValue().containsKey(e.getTableId())).collect(Collectors.toSet());
								set.remove(rowId);
								CHMHelper.getFromMap(anyMatch, rowId, () -> new HashSet<>()).addAll(set);
							}
						}

					}
					
					totalRows += metaData.getNumberOfRows();
					
					medianRows.add(new AccumMedian((double) metaData.getNumberOfRows()));
					averageRows.add(new AccumAverage((double) metaData.getNumberOfRows()));
				 
					
					for(Pair<TableRowId, InstanceId> mapping : table.getValue()) {
						
						if(kbInstancesPerClass.contains(mapping.getSecond())) { 
							CHMHelper.getFromMap(kbInstancesData, mapping.getSecond(), () -> new HashSet<>()).add(mapping.getFirst().getTableId());
						}
						else if(newInstancesPerClass.contains(mapping.getSecond())) { 
							CHMHelper.getFromMap(newInstancesData, mapping.getSecond(), () -> new HashSet<>()).add(mapping.getFirst().getTableId());
						}
						else {
							throw new IllegalStateException("Instance needs to be either from KB or NEW");
						}
					}
				}
				
				QuartileStatistics kbInstancesStatistics = getStatistics(kbInstancesData);
				QuartileStatistics newInstancesStatistics = getStatistics(newInstancesData);
				
				int i = 0;
				String[] line = new String[columnCount];
				line[i++] = entry.getKey().getFinalPart();
				line[i++] = String.valueOf(kbInstancesStatistics.getSize());
				line[i++] = String.valueOf(kbInstancesStatistics.getTotalSize());
				line[i++] = DecimalFormatter.format(kbInstancesStatistics.getAverageRows());
				line[i++] = DecimalFormatter.format(kbInstancesStatistics.getMedianRows());
				line[i++] = DecimalFormatter.format(kbInstancesStatistics.getAverageColumnMappings());
				line[i++] = DecimalFormatter.format(kbInstancesStatistics.getMedianColumnMappings());
				line[i++] = String.valueOf(newInstancesStatistics.getSize());
				line[i++] = String.valueOf(newInstancesStatistics.getTotalSize());
				line[i++] = DecimalFormatter.format(newInstancesStatistics.getAverageRows());
				line[i++] = DecimalFormatter.format(newInstancesStatistics.getMedianRows());
				line[i++] = DecimalFormatter.format(newInstancesStatistics.getAverageColumnMappings());
				line[i++] = DecimalFormatter.format(newInstancesStatistics.getMedianColumnMappings());


				line[i++] = String.valueOf(entry.getValue().size());
				line[i++] = String.valueOf(totalRows);
				
				line[i++] = DecimalFormatter.format(averageRows.get().get());
				line[i++] = DecimalFormatter.format(medianRows.get().get());
				
				line[i++] = String.valueOf(kbMatch.entrySet().stream().filter((e) -> e.getValue().size() > 1).mapToInt((e) -> e.getValue().size()).count());
				line[i++] = DecimalFormatter.format(((double) kbMatch.entrySet().stream().filter((e) -> e.getValue().size() > 1).mapToInt((e) -> e.getValue().size()).count()) / totalRows);
				line[i++] = String.valueOf(kbMatch.entrySet().stream().filter((e) -> e.getValue().size() > 2).mapToInt((e) -> e.getValue().size()).count());
				line[i++] = DecimalFormatter.format(  ((double) kbMatch.entrySet().stream().filter((e) -> e.getValue().size() > 2).mapToInt((e) -> e.getValue().size()).count()) / totalRows);
				line[i++] = String.valueOf(kbMatch.entrySet().stream().filter((e) -> e.getValue().size() > 3).mapToInt((e) -> e.getValue().size()).count());
				line[i++] = DecimalFormatter.format( ((double) kbMatch.entrySet().stream().filter((e) -> e.getValue().size() > 3).mapToInt((e) -> e.getValue().size()).count()) / totalRows);
				line[i++] = String.valueOf(kbMatch.entrySet().stream().filter((e) -> e.getValue().size() > 5).mapToInt((e) -> e.getValue().size()).count());
				line[i++] = DecimalFormatter.format( ((double) kbMatch.entrySet().stream().filter((e) -> e.getValue().size() > 5).mapToInt((e) -> e.getValue().size()).count()) / totalRows);
				line[i++] = String.valueOf(kbMatch.entrySet().stream().filter((e) -> e.getValue().size() > 10).mapToInt((e) -> e.getValue().size()).count());
				line[i++] = DecimalFormatter.format(  ((double) kbMatch.entrySet().stream().filter((e) -> e.getValue().size() > 10).mapToInt((e) -> e.getValue().size()).count()) / totalRows);
				 

				line[i++] = String.valueOf(anyMatch.entrySet().stream().filter((e) -> e.getValue().size() > 1).mapToInt((e) -> e.getValue().size()).count());
				line[i++] = DecimalFormatter.format(  ((double) anyMatch.entrySet().stream().filter((e) -> e.getValue().size() > 1).mapToInt((e) -> e.getValue().size()).count()) / totalRows);
				line[i++] = String.valueOf(anyMatch.entrySet().stream().filter((e) -> e.getValue().size() > 2).mapToInt((e) -> e.getValue().size()).count());
				line[i++] = DecimalFormatter.format(  ((double) anyMatch.entrySet().stream().filter((e) -> e.getValue().size() > 2).mapToInt((e) -> e.getValue().size()).count()) / totalRows);
				line[i++] = String.valueOf(anyMatch.entrySet().stream().filter((e) -> e.getValue().size() > 3).mapToInt((e) -> e.getValue().size()).count());
				line[i++] = DecimalFormatter.format( ((double) anyMatch.entrySet().stream().filter((e) -> e.getValue().size() > 3).mapToInt((e) -> e.getValue().size()).count()) / totalRows);
				line[i++] = String.valueOf(anyMatch.entrySet().stream().filter((e) -> e.getValue().size() > 5).mapToInt((e) -> e.getValue().size()).count());
				line[i++] = DecimalFormatter.format( ((double) anyMatch.entrySet().stream().filter((e) -> e.getValue().size() > 5).mapToInt((e) -> e.getValue().size()).count()) / totalRows);
				line[i++] = String.valueOf(anyMatch.entrySet().stream().filter((e) -> e.getValue().size() > 10).mapToInt((e) -> e.getValue().size()).count());
				line[i++] = DecimalFormatter.format(  ((double) anyMatch.entrySet().stream().filter((e) -> e.getValue().size() > 10).mapToInt((e) -> e.getValue().size()).count()) / totalRows);
				
				writer.writeNext(line);
				
				
				
			}
			
			writer.close();
			
			System.out.println(sw.toString());
		}
		catch(IOException e) {
			e.printStackTrace();
		}
			
		}
	
	private QuartileStatistics getStatistics(Map<InstanceId, Set<AnnotatedTableId>> instances) {
		ConcurrentAccumulator<AccumAverage> averageRows = new ConcurrentAccumulator<>(new AccumAverage());
		ConcurrentAccumulator<AccumAverage> averageColumnMappings = new ConcurrentAccumulator<>(new AccumAverage());

		ConcurrentAccumulator<AccumMedian> medianRows = new ConcurrentAccumulator<>(new AccumMedian());
		ConcurrentAccumulator<AccumMedian> medianColumnMappings = new ConcurrentAccumulator<>(new AccumMedian());
		AtomicInteger totalRows = new AtomicInteger();
		
		PersistentParallel6.runAndFinish(instances.entrySet(), (instanceEntry) -> {
			int rows = instanceEntry.getValue().size();
			totalRows.addAndGet(rows);
			int columns = 0;
			
			for(AnnotatedTableId tableId : instanceEntry.getValue()) {
				columns += tableRepository.getMetaData(tableId).getProperties().size();
			}
			
			averageRows.add(new AccumAverage((double) rows));
			averageColumnMappings.add(new AccumAverage((double) columns));
			
			
			medianRows.add(new AccumMedian((double) rows));
			medianColumnMappings.add(new AccumMedian((double) columns));
		}, this.getClass());
		
		return new QuartileStatistics(instances.size(), totalRows.get(), averageRows.get().get(), averageColumnMappings.get().get(), medianRows.get().get(), medianColumnMappings.get().get());

	}
	
	
	public static class QuartileStatistics {
		private final int size;
		private final int totalSize;
		private final double averageRows;
		private final double averageColumnMappings;
		private final double medianRows;
		private final double medianColumnMappings;
		public QuartileStatistics(int size, int totalSize, double averageRows, double averageColumnMappings, double medianRows, double medianColumnMappings) {
			super();
			this.size = size;
			this.totalSize = totalSize;
			this.averageRows = averageRows;
			this.averageColumnMappings = averageColumnMappings;
			this.medianRows = medianRows;
			this.medianColumnMappings = medianColumnMappings;
		}
		
		
		int getTotalSize() {
			return totalSize;
		}


		int getSize() {
			return size;
		}


		double getAverageRows() {
			return averageRows;
		}
		double getAverageColumnMappings() {
			return averageColumnMappings;
		}
		double getMedianRows() {
			return medianRows;
		}
		double getMedianColumnMappings() {
			return medianColumnMappings;
		}
	 
		
		
	}
	
	public static class TableRowInstanceBiMap {
		private final Map<TableRowId, InstanceId> rowToInstance;
		private final Map<InstanceId, Set<TableRowId>> instanceToCluster;
		private final Map<InstanceId, Integer> factCount;
		
		public TableRowInstanceBiMap(Map<TableRowId, InstanceId> rowToInstance, Map<InstanceId, Set<TableRowId>> instanceToCluster,
				Map<InstanceId, Integer> factCount) {
			this.rowToInstance = ImmuMaker.map(rowToInstance);
			this.instanceToCluster = ImmuMaker.mapWithSet(instanceToCluster);
			this.factCount = ImmuMaker.map(factCount);
		}

		public Map<InstanceId, Integer> getFactCountMap() {
			return factCount;
		}
		public InstanceId getInstanceId(TableRowId tableRowId) {
			return rowToInstance.get(tableRowId);
		}
		
		public Integer getFactCount(InstanceId instanceId) {
			return factCount.get(instanceId);
		}
		
		public Set<TableRowId> getCluster(InstanceId instanceId) {
Set<TableRowId> set = instanceToCluster.get(instanceId);	
if(set == null) {
	int sdf = 123;
}
			return set;
		}
		Map<TableRowId, InstanceId> getRowToInstance() {
			return rowToInstance;
		}
		Map<InstanceId, Set<TableRowId>> getInstanceToCluster() {
			return instanceToCluster;
		}
		
		
	}
	
	public static class InstanceData {
		private int totalTables;
		private InstancePair mainInstance;
		private InstancePair newInstanceAdded;
		private List<InstancePair > homonymsAdded;
		private String quartile;

		
		public int getTotalTables() {
			return totalTables;
		}


		public void setTotalTables(int totalTables) {
			this.totalTables = totalTables;
		}

		public InstancePair getMainInstance() {
			return mainInstance;
		}
		

		public void setMainInstance(InstancePair mainInstance) {
			this.mainInstance = mainInstance;
		}
		public InstancePair getNewInstanceAdded() {
			return newInstanceAdded;
		}
		public void setNewInstanceAdded(InstancePair newInstanceAdded) {
			this.newInstanceAdded = newInstanceAdded;
		}
		public List<InstancePair> getHomonymsAdded() {
			return homonymsAdded;
		}
		public void setHomonymsAdded(List<InstancePair> homonymsAdded) {
			this.homonymsAdded = homonymsAdded;
		}
		
		public void setQuartile(String quartile) {
			this.quartile = quartile;
		}


		public String getQuartile() {
			return quartile;
		}
		
		
		
		
	}
	
	public static class InstancePair {
		private final String instanceId;
		private final int tableCount;
		public InstancePair(String instanceId, int tableCount) {
			super();
			this.instanceId = instanceId;
			this.tableCount = tableCount;
		}
		public String getInstanceId() {
			return instanceId;
		}
		public int getTableCount() {
			return tableCount;
		}
		
		
	}
}
