package expansion.set.gscreation;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Random;
import java.util.Set;
import java.util.TreeMap;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.io.FileUtils;

import com.google.common.collect.ImmutableTable;
import com.google.gson.Gson;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.QuartileData.Quartile;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.gscreation.ExtractTablesForGoldStandard2.InstanceData;
import expansion.set.gscreation.ExtractTablesForGoldStandard2.InstancePair;
import expansion.set.gscreation.ExtractTablesForGoldStandard2.TableRowInstanceBiMap;

public class ExtractTablesForGoldStandard3 {

	@Inject
	protected ExtractTablesForGoldStandard2 extract2;
	
	@Inject
	protected AnnotatedTablesRepository tablesRepository;
	
	@Inject
	protected GoldStandardWriter goldStandardWriter;
 

	public void run( Map<ClassId, Set<ClassId>> topClassesAndParents, 
			Map<ClassId, Set<AnnotatedTableId>> tableIdsSetMap, String cacheFolder, String gsPath,
			FusionConfig fusionConfig,  SetExpansionStrategiesConfig setExpansionStrategiesConfig,
			String thresholdsFolder, StrategiesConfig strategiesConfig) {
		
		
		Map<ClassId, Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>>> finalGs = new HashMap<>();
		Map<ClassId, Set<InstanceId>> kbInstances = new HashMap<>();
		Map<ClassId, Set<InstanceId>> newInstances = new HashMap<>();
		Map<ClassId, TableRowInstanceBiMap> kbMappingsMap = new HashMap<>();
		Map<ClassId, TableRowInstanceBiMap> clusterMappingsMap = new HashMap<>();
		Map<ClassId, Map<String, InstanceData>> instanceJson = new HashMap<>();
		
		Map<ClassId, Set<InstanceId>> selectedAdditionalInstances = new HashMap<>();
		
		for(Entry<ClassId, Set<ClassId>> classIdEntry : topClassesAndParents.entrySet()) {
			ClassId classId = classIdEntry.getKey();
			TableIdsSet tableIdsSet = new TableIdsSet( tableIdsSetMap.get(classIdEntry.getKey()),  new UuidStringId(classId.getFinalPart()));
			Tri<TableRowInstanceBiMap, TableRowInstanceBiMap, Map<InstanceId, Map<InstanceId, Double>>> rc = extract2.getRowClusters(cacheFolder, classId, classIdEntry.getValue(), tableIdsSet, fusionConfig,setExpansionStrategiesConfig, thresholdsFolder, strategiesConfig);
			Set<InstanceId> kbInstancesPerClass = tableIdsSet.getTableIds().stream().flatMap((e) -> tablesRepository.getMetaData(e).getInstances().values().stream()).collect(Collectors.toSet());
			Map<InstanceId, Map<InstanceId, Double>> labelConflictEntitiesUnFiltered = rc.getC();
			
			Map<InstanceId, Map<InstanceId, Double>> labelConflictEntities =  labelConflictEntitiesUnFiltered.entrySet().stream().filter((e) -> kbInstancesPerClass.contains(e.getKey())).collect(Collectors.toMap(Entry::getKey, Entry::getValue));
			
			Map<InstanceId, Map<InstanceId, Double>> labelConflictEntitiesFiltered = new HashMap<>();
			for(Entry<InstanceId, Map<InstanceId, Double>> entry : labelConflictEntities.entrySet()) {
				 Map<InstanceId, Double> inner = new HashMap<>();
				 
				 for(Entry<InstanceId, Double> innerEntry : entry.getValue().entrySet()) {
					 if(innerEntry.getValue() > 1.95) {
						 inner.put(innerEntry.getKey(), innerEntry.getValue());
					 }
				 }
				 if(!inner.isEmpty()) {
					 labelConflictEntitiesFiltered.put(entry.getKey(), inner);
				 }
			}
			Random random = new Random( classId.getClassId().hashCode());
			
			List<InstanceId> sortList = labelConflictEntitiesFiltered.keySet().stream().sorted((e1,e2) -> e1.getInstanceId().compareTo(e2.getInstanceId())).collect(Collectors.toList());
			Collections.shuffle(sortList, random);
			
			Map<InstanceId, Integer> sortMap = new HashMap<>();
			for(ListEntry<InstanceId> listEntry : ListIterable.iterable(sortList)) {
				sortMap.put(listEntry.getValue(), listEntry.getIndex());
			}
			
			
			Set<InstanceId> chosenInstances = labelConflictEntitiesFiltered.entrySet().stream().sorted((e1,e2) -> {
			 
					Set<AnnotatedTableId> tablesE1 = new HashSet<>(); 
					for(Entry<InstanceId, Double> e1Entry : e1.getValue().entrySet()) {
					Set<TableRowId> rows = rc.getB().getCluster(e1Entry.getKey());
						for(TableRowId rowId : rows) {
							Optional<InstanceId> matchedInstance = tablesRepository.getMatchedInstanceId(rowId);
							if(matchedInstance.isPresent()) {
								if(!matchedInstance.get().equals(e1.getKey())) {
									tablesE1.add(rowId.getTableId());
								}
							}
							else {
								tablesE1.add(rowId.getTableId());
							}
							
						}
					 
					}
					Set<AnnotatedTableId> tablesE2 = new HashSet<>(); 
					for(Entry<InstanceId, Double> e2Entry : e2.getValue().entrySet()) {
						
						Set<TableRowId> rows = rc.getB().getCluster(e2Entry.getKey());
						for(TableRowId rowId : rows) {
							Optional<InstanceId> matchedInstance = tablesRepository.getMatchedInstanceId(rowId);
						if (matchedInstance.isPresent()) {
							if (!matchedInstance.get().equals(e2.getKey())) {
									tablesE2.add(rowId.getTableId());
								}
							}
							else {
								tablesE2.add(rowId.getTableId());
							}
							
						}
					}
					int compare = Integer.compare(tablesE2.size(), tablesE1.size());
				
					if(compare != 0) {
						return compare;
					}
					else {
						return Integer.compare(sortMap.get(e1.getKey()), sortMap.get(e2.getKey()));
					}
				} 
			).limit(12).map(Entry::getKey).collect(Collectors.toSet());
			
			
		 
			
			Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>> finalGsInner = new HashMap<>();
			Set<InstanceId> kbInstancesInner = new HashSet<>();

			Set<InstanceId> newInstancesInner = new HashSet<>();
			Set<InstanceId> selectedAdditionalInstancesInner = new HashSet<>();
			
			Map<String, InstanceData> instanceDataList = new HashMap<>();

			
			for(InstanceId instanceId : chosenInstances) {
				InstanceData instanceData = new InstanceData();
instanceData.setQuartile("ZE");
				
				System.out.println("==================================");
				System.out.println(instanceId.getReadableString());
				System.out.println();
				instanceData.setMainInstance(new InstancePair(instanceId.getInstanceId(), 0));
				{
					kbInstancesInner.add(instanceId);
					Set<InstanceId> set = new HashSet<>();
					set.add(instanceId);
//					selectedAdditionalInstancesInner.add(instanceId);
					int i = 0;
					for(TableRowId rowId : tablesRepository.getMatchedRowsIds(set, classId)) {
						
						System.out.println(getRow(rowId));
						CHMHelper.getFromMap(finalGsInner, rowId.getTableId(), () -> new ArrayList<>()).add(new Pair<>(rowId, instanceId));
						
						
						i++;
						if(i == 3)
							break;
					}
				}
				
				List<InstancePair> homonymsAdded = new ArrayList<>();
				for( Entry<InstanceId, Double>  homonym : labelConflictEntitiesFiltered.get(instanceId).entrySet()) {
					System.out.println("==");
					System.out.println(homonym.getKey().getReadableString());
					Set<InstanceId> set = new HashSet<>();
					set.add(homonym.getKey());
					Set<TableRowId> rowSet = rc.getB().getCluster(homonym.getKey());
					if(rowSet != null) {
						int i = 0;
						for(TableRowId rowId : rowSet) {
							
							Optional<InstanceId> matchedInstance = tablesRepository.getMatchedInstanceId(rowId);
							System.out.println(getRow(rowId)); 

							if(matchedInstance.isPresent()) {
								if(!matchedInstance.get().equals(instanceId)) {
									System.out.println(matchedInstance.get().toString());
									CHMHelper.getFromMap(finalGsInner, rowId.getTableId(), () -> new ArrayList<>()).add(new Pair<>(rowId, matchedInstance.get()));

									selectedAdditionalInstancesInner.add(matchedInstance.get());
									homonymsAdded.add(new InstancePair(matchedInstance.get().getInstanceId(), 0));
									set.add(matchedInstance.get());
									i++;
								}
								
							}
							else {

								selectedAdditionalInstancesInner.add(homonym.getKey());
								homonymsAdded.add(new InstancePair(homonym.getKey().getInstanceId(), 0));
								CHMHelper.getFromMap(finalGsInner, rowId.getTableId(), () -> new ArrayList<>()).add(new Pair<>(rowId, homonym.getKey()));
								i++;
							}
							
//							newInstancesInner.add(homonym.getKey()); 
							
							if(i == 6)
								break;
						} 
					}
					
					int i = 0;
					for(TableRowId rowId : tablesRepository.getMatchedRowsIds(set, classId)) {
						System.out.println(getRow(rowId)); 
						// selectedAdditionalInstancesInner.add(homonym.getKey());
						// homonymsAdded.add(new
						// InstancePair(homonym.getKey().getInstanceId(), 0));

						Optional<InstanceId> matchedInstance = tablesRepository.getMatchedInstanceId(rowId);
						System.out.println(getRow(rowId));

						if (matchedInstance.isPresent()) {
							if (!matchedInstance.get().equals(instanceId)) {
								CHMHelper.getFromMap(finalGsInner, rowId.getTableId(), () -> new ArrayList<>()).add(new Pair<>(rowId, matchedInstance.get()));
							}
						}
//						kbInstancesInner.add(homonym.getKey());
						
						i++;
						if(i == 3)
							break;
					}
					
				}
				
				instanceData.setHomonymsAdded(homonymsAdded);
				instanceDataList.put( instanceId.getReadableString(), instanceData);

			}
			
			for(AnnotatedTableId tableId : tableIdsSet.getTableIds()) {
				finalGs.remove(tableId);
			}
			Tri<TreeMap<InstanceId, Set<TableRowId>>, Map<InstanceId, Quartile>, TableRowInstanceBiMap> mappingsT = extract2.getMappings(classId, random, tableIdsSet, 800, null);
 			
			
			finalGs.put(classId, finalGsInner);
			 
			
			
			kbInstances.put(classId, kbInstancesInner);
			newInstances.put(classId, newInstancesInner);
			
			kbMappingsMap.put(classId, mappingsT.getC());
			clusterMappingsMap.put(classId, rc.getB());
			
			 
			selectedAdditionalInstances.put(classId, selectedAdditionalInstancesInner);
			
			instanceJson.put(classId, instanceDataList);
		} 
		
		
		goldStandardWriter.writeGoldStandard(finalGs, kbInstances, newInstances, kbMappingsMap, clusterMappingsMap, new HashMap<>(), new HashMap<>(), new HashMap<>(), gsPath);
		
		goldStandardWriter.writeSelectedInstances(selectedAdditionalInstances, gsPath + "/extraHomonyms.csv");
		goldStandardWriter.writeSelectedInstances(kbInstances, gsPath + "/extraKB.csv");
		
		try {
		   Gson gson = new Gson();
		   FileUtils.writeStringToFile(new File(gsPath + "/instancesExtra.json"), gson.toJson(instanceJson));
		}
		catch(IOException e) {
			e.printStackTrace();
		}
		int sfd= 123;
		
	}
	
 
	
	public String getRow(TableRowId tableRowId) {
		String ret = "";
		AnnotatedTable table = this.tablesRepository.getTable(tableRowId.getTableId());
		ImmutableTable<RowIdx, ColumnIdx, String> stringTable = table.getStringTable();
		for( ColumnIdx colIdx : table.getColumns()) {
			String value = stringTable.get(tableRowId.getRowIdx(), colIdx);
			if(!ret.equals("")) {
				ret += "|";
			}
			if(value != null) {
				 ret += value;
			}
			else {
				ret += "";
			}
		}
		return ret;
	}
}
