package expansion.set.gscreation;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Random;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.QuartileData;
import de.mannheim.uni.expansion.utils.QuartileData.Quartile;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.dws.expansion.data.kb.degree.DegreeRepository;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.gscreation.ExtractTablesForGoldStandard2.TableRowInstanceBiMap;

public class ExtractTablesForGoldStandard4 {

	@Inject
	protected AnnotatedTablesRepository tableRepository;

	@Inject
	protected ExtractTablesForGoldStandard2 extract2;


	@Inject
	protected GoldStandardWriter goldStandardWriter;

	@Inject
	protected DegreeRepository degreeRepository;

	public void run(Map<ClassId, Set<ClassId>> topClassesAndParents, Map<ClassId, Set<AnnotatedTableId>> tableFilterMap,
			String cacheFolder, String gsPath, FusionConfig fusionConfig, SetExpansionStrategiesConfig setExpansionStrategiesConfig, String thresholdsFolder,
			StrategiesConfig strategiesConfig) {

		Map<ClassId, Set<InstanceId>> kbInstances = new HashMap<>();
		Map<ClassId, Set<InstanceId>> newInstances = new HashMap<>();
		Map<ClassId, TableRowInstanceBiMap> kbMappingsMap = new HashMap<>();
		Map<ClassId, TableRowInstanceBiMap> clusterMappingsMap = new HashMap<>();
		Map<ClassId, Set<InstanceId>> selectedAdditionalInstances = new HashMap<>();
		Map<ClassId, Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>>> finalGs = new HashMap<>();

		for (Entry<ClassId, Set<ClassId>> classEntry : topClassesAndParents.entrySet()) {
			ClassId classId = classEntry.getKey();
			Set<ClassId> parents = classEntry.getValue();
			Set<AnnotatedTableId> tableFilter = tableFilterMap.get(classId);

			TableIdsSet tableIdsSet = new TableIdsSet(this.tableRepository.getAllTableIds(classId).stream().filter((e) -> {
				AnnotatedTableMetaData md = tableRepository.getMetaData(e);

				AnnotatedTable data = tableRepository.getTable(e);
				double matched = md.getInstances().size();
				int total = md.getNumberOfRows();

				// if(classId.getClassId().equals("http://dbpedia.org/ontology/Settlement"))
				// {
				// minProportion = 0.29;
				// maxRows = 100;
				// }


				return !tableFilter.contains(e) && md.getProperties().size() >= extract2.MIN_COLUMNS;

			}).collect(Collectors.toSet()), new UuidStringId(classId.getFinalPart()));

			Random random = new Random(classId.hashCode());

			Map<InstanceId, Integer> degreeMapping = new HashMap<>();
			ConcurrentHashMap<InstanceId, Set<AnnotatedTableId>> tableIndex = new ConcurrentHashMap<>();
			ConcurrentHashMap<AnnotatedTableId, Set<InstanceId>> instanceIndex = new ConcurrentHashMap<>();
			
			Set<InstanceId> matchedInstances = Collections.newSetFromMap(new ConcurrentHashMap<>());
			PersistentParallel6.runAndFinish(tableIdsSet.getTableIds(), (tableId) -> {
				for (TableRowId rowId : tableRepository.getRowIds(tableId)) {
					Optional<InstanceId> matchedInstance = tableRepository.getMatchedInstanceId(rowId);
					if (matchedInstance.isPresent()) {
						matchedInstances.add(matchedInstance.get());
						CHMHelper.getFromCHM(tableIndex, matchedInstance.get(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableId);
						CHMHelper.getFromCHM(instanceIndex, tableId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(matchedInstance.get());

					}
				}
			}, this.getClass());

			for (InstanceId instanceId : matchedInstances) {
				degreeMapping.put(instanceId, degreeRepository.getNumIndegree(instanceId));
			}

			QuartileData<InstanceId> quartileData = extract2.getQuartileData(degreeMapping, random, (e1, e2) -> {
				return e1.getInstanceId().compareTo(e2.getInstanceId());
			});

			Set<InstanceId> selectedInstances = new HashSet<>(quartileData.getSelectedPerQuartile().get(Quartile.Q1));

			Set<AnnotatedTableId> tableIdsSetNew = Collections.newSetFromMap(new ConcurrentHashMap<>());
			PersistentParallel6.runAndFinish(tableIdsSet.getTableIds(), (tableId) -> {
				Set<InstanceId> instances = instanceIndex.get(tableId);
				double size = instances.size();
				int countLowQuartile = 0;
				for (InstanceId instance : instances) {
					if (selectedInstances.contains(instance)) {
						countLowQuartile++;
					}
				}

				if (countLowQuartile / size >= 0.3) {
					tableIdsSetNew.add(tableId);
				}

			}, this.getClass());

			TableIdsSet newTableIdsSet = new TableIdsSet(new HashSet<>(tableIdsSetNew));

			Tri<TableRowInstanceBiMap, TableRowInstanceBiMap, Map<InstanceId, Map<InstanceId, Double>>> rc = extract2.getRowClusters(cacheFolder, classId,
					parents, newTableIdsSet, fusionConfig, setExpansionStrategiesConfig, thresholdsFolder, strategiesConfig);

			TableRowInstanceBiMap kbBiMap = extract2.getKBBiMap(newTableIdsSet, null);
			kbMappingsMap.put(classId, kbBiMap);

			clusterMappingsMap.put(classId, rc.getB());

			Set<InstanceId> kbInstancesInner = new HashSet<>(kbBiMap.getInstanceToCluster().keySet());

			Set<InstanceId> newInstancesInner = new HashSet<>();

			newInstancesInner.addAll(rc.getA().getInstanceToCluster().keySet());

			kbInstances.put(classId, kbInstancesInner);
			newInstances.put(classId, newInstancesInner);

			selectedAdditionalInstances.put(classId, new HashSet<>());

			Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>> gs = new ConcurrentHashMap<>();
			for (AnnotatedTableId tableId : newTableIdsSet.getTableIds()) {
				gs.put(tableId, new ArrayList<>());
			}

			finalGs.put(classId, gs);
		}

		goldStandardWriter.writeGoldStandard(finalGs, kbInstances, newInstances, kbMappingsMap, clusterMappingsMap, new HashMap<>(), new HashMap<>(),
				new HashMap<>(), gsPath);

		int asdf = 123;

	}
}
