package expansion.set.gscreation;

import java.io.File;
import java.io.IOException;
import java.io.StringWriter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.inject.Inject;

import org.apache.commons.io.FileUtils;
import org.json.JSONArray;
import org.json.JSONObject;

import com.opencsv.CSVWriter;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.ValueConverter;
import expansion.set.gscreation.ExtractTablesForGoldStandard2.TableRowInstanceBiMap;

public class GoldStandardWriter {
	
	
	@Inject
	protected ValueConverter valueConverter;
	
	
	@Inject
	protected AnnotatedTablesRepository tableRepository;
	
	
	public GoldStandardWriter() {
		// TODO Auto-generated constructor stub
	}
 
	
	public void writeGoldStandard(Map<ClassId, Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>>> finalGs, Map<ClassId, Set<InstanceId>> kbInstances, Map<ClassId, Set<InstanceId>> newInstances, Map<ClassId, TableRowInstanceBiMap> kbMappingsMap, Map<ClassId, TableRowInstanceBiMap> clusterMappingsMap, 
			Map<ClassId, Set<InstanceId>> selectedKbInstances,Map<ClassId,  Set<InstanceId>> selectedNewInstances, Map<ClassId, Set<InstanceId>> selectedAdditionalInstances, String gsPath) {
		MKDIR.createDirUnchecked(new File(gsPath + "/tables/"));
		MKDIR.createDirUnchecked(new File(gsPath + "/instance/"));
		MKDIR.createDirUnchecked(new File(gsPath + "/attributes/"));
		Map<AnnotatedTableId, ClassId> classMappings = new HashMap<>();
		Set<InstanceId> nonKbInstances = new HashSet<>();
		for(Entry<ClassId, Map<AnnotatedTableId, List<Pair<TableRowId, InstanceId>>>> classEntry : finalGs.entrySet()) {
			for(AnnotatedTableId tableId : classEntry.getValue().keySet()) {
				writeTable(  tableId, gsPath + "/tables/");
				writePropertyColumns( tableId, gsPath + "/attributes/");
				nonKbInstances.addAll( writeInstances(tableId, gsPath + "/instance/", kbMappingsMap.get(classEntry.getKey()), clusterMappingsMap.get(classEntry.getKey())));
				classMappings.put(tableId, classEntry.getKey());
			}
		}
		writeClassesFile(classMappings, gsPath + "/classes_GS.csv");
		writeNonKbInstances(nonKbInstances, gsPath + "/nonKBInstances.csv");
		writeSelectedInstances(selectedKbInstances, gsPath + "/selectedKbInstances.csv");
		writeSelectedInstances(selectedNewInstances, gsPath + "/selectedClusterInstances.csv");
		writeSelectedInstances(selectedAdditionalInstances, gsPath + "/selectedAdditionalInstances.csv");
		
	}
	
	void writeSelectedInstances(Map<ClassId, Set<InstanceId>> selectedInstances, String file) {
		try {
		StringWriter sw = new StringWriter();
		CSVWriter writer = new CSVWriter(sw, ',');
		
		for(Entry<ClassId, Set<InstanceId>> classEntry : selectedInstances.entrySet()) {
			
			for(InstanceId  instanceId : classEntry.getValue()) {
				 
				 
				String[] line = new String[3];
				line[0] = classEntry.getKey().getClassId();
				line[1] = instanceId.getInstanceId();
				line[2] = instanceId.getReadableString(); 
				writer.writeNext(line);
			}
		}
		

		
		writer.close();
		
			FileUtils.writeStringToFile(new File(file), sw.toString(), "UTF-8");
	 
		}
		catch(IOException e) {
			e.printStackTrace(); 
		}
		
	}


	 

	private void writeNonKbInstances(Set<InstanceId> nonKbInstances, String file) {
		try {
		  
			StringWriter sw = new StringWriter();
			CSVWriter writer = new CSVWriter(sw, ',');
			
			for(InstanceId  instanceId : nonKbInstances) {
				 
				 
				String[] line = new String[2];
				line[0] = instanceId.getInstanceId();
				line[1] = instanceId.getReadableString(); 
				writer.writeNext(line);
			}
			
			writer.close();
			
			FileUtils.writeStringToFile(new File(file), sw.toString(), "UTF-8");
		 
			}
			catch(IOException e) {
				e.printStackTrace(); 
			}
		
	}


	Set<InstanceId> writeInstances(AnnotatedTableId tableId, String folder, TableRowInstanceBiMap kbMap,
			TableRowInstanceBiMap clusterMap) {
		try {
			Set<InstanceId> nonKbInstances = new HashSet<>();
			
			List<TableRowId> rowIds = tableRepository.getRowIds(tableId);
			StringWriter sw = new StringWriter();
			CSVWriter writer = new CSVWriter(sw, ',' );
			
			for(TableRowId tableRowId : rowIds) {
				InstanceId instanceId = kbMap.getInstanceId(tableRowId);
				if(instanceId == null) {
					instanceId = clusterMap.getInstanceId(tableRowId);
					if(instanceId == null)
						continue;
					else
						nonKbInstances.add(instanceId);
				}
				String[] line = new String[3];
				line[0] = instanceId.getInstanceId();
				line[1] = instanceId.getReadableString();
				line[2] = String.valueOf(tableRowId.getRowIdx().getRowIdx());
				writer.writeNext(line);
			}
			
			writer.close();
			
			FileUtils.writeStringToFile(new File(folder + "/" + tableId.getTableName()  + ".csv"), sw.toString());
			
				return nonKbInstances;
			}
			catch(IOException e) {
				e.printStackTrace();
				return null;
			}
		
	}


	public void writeClassesFile(Map<AnnotatedTableId, ClassId> classMappings, String file) {
		try {
		
		StringWriter sw = new StringWriter();
		CSVWriter writer = new CSVWriter(sw, ',', CSVWriter.NO_QUOTE_CHARACTER);
		
		for(Entry<AnnotatedTableId, ClassId> entry : classMappings.entrySet()) {
			String[] line = new String[3];
			line[0] = entry.getKey().getTableName() + ".tar.gz";
			line[1] = entry.getValue().getFinalPart();
			line[2] = entry.getValue().getClassId();
			writer.writeNext(line);
		}
		
		writer.close();
		
		FileUtils.writeStringToFile(new File(file), sw.toString());
		
		}
		catch(IOException e) {
			e.printStackTrace();
		}
		
		
	}

	void writePropertyColumns(AnnotatedTableId tableId, String attributesFolder) {
		try {
			Map<Integer, PropertyId> propertyMap = tableRepository.getMetaData(tableId).getProperties();

			StringWriter sw = new StringWriter();
			CSVWriter writer = new CSVWriter(sw, ',');

			for (Entry<Integer, PropertyId> entry : propertyMap.entrySet()) {

				String[] line = new String[3];
				line[0] = String.valueOf(entry.getKey());
				line[1] = entry.getValue().toCombinedString();

				writer.writeNext(line);
			}

			writer.close();

			FileUtils.writeStringToFile(new File(attributesFolder + "/" + tableId.getTableName() + ".csv"), sw.toString());

		} catch (IOException e) {
			e.printStackTrace();
		}

	}
	
	
	void writeTable(AnnotatedTableId tableId, String tableFolder) {
		try {
			AnnotatedTableMetaData md = tableRepository.getMetaData(tableId);
		AnnotatedTable data = tableRepository.getTable(tableId);
			
		JSONObject json = new JSONObject();
		json.put("hasKeyColumn", true);
		json.put("keyColumnIndex", md.getKeyColumn());
		json.put("headerRowIndex", 0);
		

		json.put("pageTitle", md.getPageTitle());
		json.put("title", md.getTableTitle());
		json.put("pld", md.getSource().getPLD());
		json.put("url", md.getSource().getSource());
		
		

		JSONArray columns = new JSONArray();
		int maxRowIdx = data.getStringTable().rowKeySet().stream().mapToInt(RowIdx::getRowIdx).max().getAsInt();
		
		for(Entry<ColumnIdx, Map<RowIdx, String>> col : data.getStringTable().columnMap().entrySet()) {
			JSONArray column = new JSONArray();
			
			for(int i = 0; i <= maxRowIdx; i++) {
				RowIdx rowIdx = new RowIdx(i);
				if(col.getValue().containsKey(rowIdx)) {
					column.put(col.getValue().get(rowIdx));
				}
				else {
					column.put("");
				}
				
			}
	 
			
			columns.put(column);
		}
		
		json.put("relation", columns);
		
			FileUtils.writeStringToFile(new File(tableFolder + "/" + tableId.getTableName() + ".json"), json.toString(), "UTF-8");
		}
		catch(IOException e) {
			e.printStackTrace();
		}
		
		
	}
}
