package expansion.set.gscreation;

import java.io.File;
import java.io.IOException;
import java.io.StringReader;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.filefilter.TrueFileFilter;

import au.com.bytecode.opencsv.CSVReader;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId.InstanceIdStyle;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.informatik.dws.attribute.model.AttributeMappings;
import de.uni_mannheim.informatik.dws.attribute.model.PropertyMapping;
import de.uni_mannheim.informatik.dws.attribute.model.TableAttributeMappings;
import expansion.set.evaluation.goldstandard.GoldStandardClusters;

public class GsOutputReader {


	private final int csvLength = ".csv".length();
	private final int tarGzLength = ".tar.gz".length();
	 
private final 	InstanceIdStyle instanceIdStyle;
	
	@Inject
	public GsOutputReader(InstanceIdStyle instanceIdStyle) {
		super();
		this.instanceIdStyle = instanceIdStyle;
	}


	public Set<AnnotatedTableId> readListOfTables(String file) {
		Set<AnnotatedTableId> ret = new HashSet<>();
		try {

			StringReader stringReader = new StringReader(FileUtils.readFileToString(new File(file), "UTF-8"));

			CSVReader reader = new CSVReader(stringReader, ',');

			for (String[] entry : reader.readAll()) {
				ret.add(new AnnotatedTableId(entry[0]));

			}

			reader.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return ret;
	}

	public Set<InstanceId> readInstancesList(String file) {
		Set<InstanceId> ret = new HashSet<>();
		try {

			StringReader stringReader = new StringReader(FileUtils.readFileToString(new File(file), "UTF-8"));

			CSVReader reader = new CSVReader(stringReader, ',');

			for (String[] entry : reader.readAll()) {
				if(!entry[0].startsWith("http://")) {
					int sdf= 123;
				}
				else {
					InstanceId entityId = new InstanceId(entry[0], instanceIdStyle);
					ret.add(entityId);
				}
				
			}

			reader.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return ret;
	}
	
	
	
	
	public Map<Pair<InstanceId, PropertyId>, Tri<String, Boolean, Boolean>> getFacts(String factFile, Map<InstanceId, ClassId> instancesToClassMappings,
			Map<ClassId, Set<PropertyId>> nonTailProperties) {

		Map<Pair<InstanceId, PropertyId>, Tri<String, Boolean, Boolean>> ret = new HashMap<>();

		try {

			StringReader stringReader = new StringReader(FileUtils.readFileToString(new File(factFile), "UTF-8"));

			CSVReader reader = new CSVReader(stringReader, ',', '"');

			for (String[] entry : reader.readAll()) {

				InstanceId entityIdPojo =  new InstanceId(entry[0], instanceIdStyle) ;


				PropertyId propertyId = PropertyId.fromCombinedString(entry[1]);

				ClassId classId = instancesToClassMappings.get(entityIdPojo);
				if (nonTailProperties.containsKey(classId) && nonTailProperties.get(classId).contains(propertyId)) {
			 
					String string = entry[2];

					boolean presence = false;
					if (entry[3].equals("true")) {
						presence = true;
					}

					boolean implicit = false;
					if (entry[4].equals("true")) {
						implicit = true;
					}

					ret.put(new Pair<>(entityIdPojo, propertyId), new Tri<>(string, presence, implicit));
				}

			}

			reader.close();

		} catch (Exception e) {
			e.printStackTrace();
		}

		return ret;
	}
	
	public GoldStandardClusters getMapping(String mappingsFolder, Set<AnnotatedTableId> processedTables) {
		@SuppressWarnings("unchecked")
		List<File> files = (List<File>) FileUtils.listFiles(new File(mappingsFolder), TrueFileFilter.INSTANCE, null);

		ConcurrentHashMap<TableRowId, InstanceId> ret = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(files, (tableFile) -> {
			try {


				AnnotatedTableId tableId = new AnnotatedTableId(tableFile.getName().substring(0, tableFile.getName().length() - csvLength));

				if (processedTables == null || processedTables.contains(tableId)) {

					StringReader stringReader = new StringReader(FileUtils.readFileToString(tableFile, "UTF-8"));

					CSVReader reader = new CSVReader(stringReader, ',', '"');

					for (String[] entry : reader.readAll()) {
						if (!entry[0].startsWith("http://")) {
							InstanceId entityIdPojo = new InstanceId(entry[0], entry[0]);

							int rowIdx = Integer.parseInt(entry[2]);
							TableRowId rowId = new TableRowId(tableId, rowIdx);
							ret.put(rowId, entityIdPojo);
						} else {
							InstanceId entityIdPojo = new InstanceId(entry[0], instanceIdStyle);
							// if(processedInstances.contains(entityIdPojo)) {
							int rowIdx = Integer.parseInt(entry[2]);
							TableRowId rowId = new TableRowId(tableId, rowIdx);
							ret.put(rowId, entityIdPojo);
							// }
						}

					}

					reader.close();

				}
			} catch (Exception e) {
				e.printStackTrace();
			}

		}, this.getClass());

		return new GoldStandardClusters(ret);
	}

	public Set<AnnotatedTableId> readTableList(String file) {
		Set<AnnotatedTableId> ret = new HashSet<>();
		try {

			StringReader stringReader = new StringReader(FileUtils.readFileToString(new File(file), "UTF-8"));

			CSVReader reader = new CSVReader(stringReader, ',');

			for (String[] entry : reader.readAll()) {
			 
					AnnotatedTableId tableId = new AnnotatedTableId(entry[0]);
					ret.add(tableId);
		 
				
			}

			reader.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return ret;
	}

	public Map<AnnotatedTableId, ClassId> readTableClasses(String tableClassesFile) {
		Map<AnnotatedTableId, ClassId> ret = new HashMap<>();
		try {

			StringReader stringReader = new StringReader(FileUtils.readFileToString(new File(tableClassesFile), "UTF-8"));


			CSVReader reader = new CSVReader(stringReader, ',', '"');

			for (String[] entry : reader.readAll()) {
				ret.put(new AnnotatedTableId(entry[0].substring(0, entry[0].length() - tarGzLength)), new ClassId(entry[2]));
			}

			reader.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return ret;
	}


	public AttributeMappings getAttributeMappings(String mappingsFolder, Set<AnnotatedTableId> processedTable, Map<AnnotatedTableId, ClassId> tableClasses,
			Map<ClassId, Set<PropertyId>> nonTailProperties) {
		@SuppressWarnings("unchecked")
		List<File> files = (List<File>) FileUtils.listFiles(new File(mappingsFolder), TrueFileFilter.INSTANCE, null);

		ConcurrentHashMap<AnnotatedTableId, TableAttributeMappings> ret = new ConcurrentHashMap<>();
		ConcurrentHashMap<AnnotatedTableId, ClassId> classIds = new ConcurrentHashMap<>();
		 
		PersistentParallel6.runAndFinish(files, (tableFile) -> {
			try {
				AnnotatedTableId tableId = new AnnotatedTableId(tableFile.getName().substring(0, tableFile.getName().length() - csvLength));

				
				if(processedTable.contains(tableId)) {
					ClassId classId = tableClasses.get(tableId);
					Set<PropertyId> nonTailProperty = nonTailProperties.get(classId);
					if (nonTailProperty != null) {

						Map<ColumnIdx, PropertyMapping> innerMap = new HashMap<>();

						StringReader stringReader = new StringReader(FileUtils.readFileToString(tableFile, "UTF-8"));
	
						CSVReader reader = new CSVReader(stringReader, ',', '"');

						for (String[] entry : reader.readAll()) {

							ColumnIdx columnIdx = new ColumnIdx(Integer.parseInt(entry[0]));

							PropertyId propertyId = PropertyId.fromCombinedString(entry[1]);

							if (nonTailProperty.contains(propertyId))
								innerMap.put(columnIdx, new PropertyMapping(propertyId, 1.0));

						}

						reader.close();

						if (!innerMap.isEmpty()) {
							classIds.put(tableId, classId);
							ret.put(tableId, new TableAttributeMappings(innerMap));
						}
					}
				}
				
			} catch (Exception e) {
				e.printStackTrace();
			}

		}, this.getClass(), true);
		
		
		for (AnnotatedTableId tableId : processedTable) {
			if (!ret.containsKey(tableId)) {
				ClassId classId = tableClasses.get(tableId);
				ret.put(tableId, new TableAttributeMappings(new HashMap<>()));
				classIds.put(tableId, classId);
			}
		}
		 

		return new AttributeMappings(ret, classIds);
	}
}
