package expansion.set.hyperparameter;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentLinkedQueue;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdsSet;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import expansion.set.evaluation.goldstandard.GoldStandardClusters;

public class HyperparameterOptimizer {

	@Inject
	public HyperparameterOptimizer() {
	}

	public void findHyperParametersEM(List<ValidationRun> validationRuns, RandomForestRun run) {

		// Parameters [parameters=[, , -K, 3, -depth, 20, -N, 0, -M, 10, -V,
		// 1e-1, -P, 85, -O, -store-out-of-bag-predictions, -I, 1000],
		// numeric=true]

		// 0.19069189762683014
		// Parameters [parameters=[, , -K, 3, -depth, 15, , , -M, 15, -V, 1e-1,
		// -P, 85, -O, -store-out-of-bag-predictions, -I, 1000], numeric=true]

		// 0.015500396962012334
		// Parameters [parameters=[, , -K, 4, -depth, 0, -N, 3, -M, 0, -V, 1e-2,
		// -P, 95, -O, -store-out-of-bag-predictions, -I, 150], numeric=false]

		int k = 3;

		List<Parameters> parameterList = new ArrayList<>();

		List<Integer> minInstancesPerLeaf = new ArrayList<>();
		minInstancesPerLeaf.add(0);
		minInstancesPerLeaf.add(5);
		minInstancesPerLeaf.add(10);
		minInstancesPerLeaf.add(15);
		minInstancesPerLeaf.add(20);
		minInstancesPerLeaf.add(30);

		List<Integer> depths = new ArrayList<>();
		depths.add(0);
		depths.add(15);
		depths.add(20);
		depths.add(25);
		depths.add(35);

		List<Integer> oobRates = new ArrayList<>();
		oobRates.add(50);
		oobRates.add(75);
		oobRates.add(85);
		oobRates.add(90);
		oobRates.add(95);

		int minVar = 1;
		// boolean numeric = true;
		// for (boolean numeric = true; numeric == true; numeric = false) {

		// for (Integer depth : depths) {
		//
		// for (Integer minPerLeaf : minInstancesPerLeaf) {

		int depth = 15;
		int minPerLeaf = 15;

		// for (Integer oobRate : oobRates) {
		List<Integer> folds = new ArrayList<>();
		folds.add(0);
		folds.add(3);
		folds.add(4);
		folds.add(5);
		folds.add(10);
		folds.add(20);
		Boolean numeric = null;
		while (numeric != false) {
			if (numeric == null) {
				numeric = true;
			} else {
				numeric = false;
			}

			for (Integer n : folds) {
				String[] parameters = new String[18];

				parameters[0] = "";
				parameters[1] = "";
				parameters[2] = "-K";
				parameters[3] = String.valueOf(k);
				parameters[4] = "-depth";
				parameters[5] = String.valueOf(depth);
				parameters[6] = "-N";
				parameters[7] = String.valueOf(n);
				parameters[8] = "-M";
				parameters[9] = String.valueOf(minPerLeaf);
				parameters[10] = "-V";
				parameters[11] = "1e-" + minVar;
				parameters[12] = "-P";
				parameters[13] = String.valueOf(95);
				parameters[14] = "-O";
				parameters[15] = "-store-out-of-bag-predictions";
				parameters[16] = "-I";
				parameters[17] = "1000";

				parameterList.add(new Parameters(parameters, numeric, false));
			}
		}

		// }
		// }
		//
		// }

		// }

		ConcurrentLinkedQueue<Pair<Double, Parameters>> queue = new ConcurrentLinkedQueue<>();

		PersistentParallel6.runAndFinish(parameterList, (input) -> {
			double sum = 0.0;
			int count = 0;

			for (ValidationRun validationRun : validationRuns) {
				count++;

				sum += run.run(validationRun, input);
			}

			queue.add(new Pair<>(sum / count, input));
		}, this.getClass());

		for (Pair<Double, Parameters> pair : queue) {
			System.out.println(pair.getFirst());
			System.out.println(pair.getSecond().toString());
		}

	}

	public void findHyperParameters(List<ValidationRun> validationRuns, RandomForestRun run, boolean descretizeOn) {

		List<Integer> minInstancesPerLeaf = new ArrayList<>();
		minInstancesPerLeaf.add(0);
		minInstancesPerLeaf.add(10);
		minInstancesPerLeaf.add(50);
		minInstancesPerLeaf.add(150);

		List<Integer> depths = new ArrayList<>();
		depths.add(0);
		depths.add(5);
		depths.add(10);
		depths.add(20);

		List<Integer> folds = new ArrayList<>();
		folds.add(0);
		folds.add(3);
		folds.add(4);
		folds.add(5);
		folds.add(10);
		folds.add(20);

		List<Integer> oobRates = new ArrayList<>();
		oobRates.add(85);
		oobRates.add(90);
		oobRates.add(95);
		oobRates.add(75);
		oobRates.add(50);

		for (Integer oobRate : oobRates) {
			List<Parameters> parameterList = new ArrayList<>();

			Boolean descretize = null;
			if (!descretizeOn) {
				descretize = true;
			}

			while (descretize == null || descretize == true) {
				if (descretize == null) {
					descretize = true;
				} else {
					descretize = false;
				}

				Boolean numeric = null;

				numeric = true;

				// while (numeric == null || numeric == true) {
				// if (numeric == null) {
				// numeric = true;
				// } else {
				// numeric = false;
				// }

				for (int minVar = 1; minVar < 5; minVar++) {
					for (int k = 0; k < 10; k++) {

						for (Integer n : folds) {
							for (Integer depth : depths) {

								for (Integer minPerLeaf : minInstancesPerLeaf) {
									String[] parameters = new String[18];

									parameters[0] = "";
									parameters[1] = "-U";
									parameters[2] = "-K";
									parameters[3] = String.valueOf(k);
									parameters[4] = "-depth";
									parameters[5] = String.valueOf(depth);
									parameters[6] = "-N";
									parameters[7] = String.valueOf(n);
									parameters[8] = "-M";
									parameters[9] = String.valueOf(minPerLeaf);
									parameters[10] = "-V";
									parameters[11] = "1e-" + minVar;
									parameters[12] = "-P";
									parameters[13] = String.valueOf(oobRate);
									parameters[14] = "-O";
									parameters[15] = "-store-out-of-bag-predictions";
									parameters[16] = "-I";
									parameters[17] = "300";

									parameterList.add(new Parameters(parameters, numeric, descretize));

								}

							}
						}
						// }

					}

				}
			}

			ConcurrentLinkedQueue<Pair<Double, Parameters>> queue = new ConcurrentLinkedQueue<>();

			PersistentParallel6.runAndFinish(parameterList, (input) -> {
				double sum = 0.0;
				int count = 0;

				for (ValidationRun validationRun : validationRuns) {
					count++;

					sum += run.run(validationRun, input);
				}

				queue.add(new Pair<>(sum / count, input));
			}, this.getClass());

			double minScore = 0.0;
			Parameters min = null;
			for (Pair<Double, Parameters> pair : queue) {
				if (min == null || pair.getFirst() < minScore) {
					minScore = pair.getFirst();
					min = pair.getSecond();
				}
			}
			System.out.println(minScore);
			System.out.println(min.toString());

		}
	};

	public void findHyperParameters_EM_2(List<ValidationRun> validationRuns, RandomForestRun run, boolean descretizeOn) {

		List<Integer> minInstancesPerLeaf = new ArrayList<>();
		minInstancesPerLeaf.add(0);
		// minInstancesPerLeaf.add(2);
		minInstancesPerLeaf.add(5);
		// minInstancesPerLeaf.add(6);
		// minInstancesPerLeaf.add(7);
		// minInstancesPerLeaf.add(8);
		// minInstancesPerLeaf.add(9);
		minInstancesPerLeaf.add(10);
		// minInstancesPerLeaf.add(12);
		minInstancesPerLeaf.add(15);

		List<Integer> depths = new ArrayList<>();
		depths.add(0);
		depths.add(3);
		// depths.add(4);
		// depths.add(5);
		depths.add(6);
		// depths.add(7);
		// depths.add(8);
		depths.add(10);
		depths.add(15);

		// List<Integer> folds = new ArrayList<>();
		// folds.add(0);
		// folds.add(3);
		// folds.add(4);
		// folds.add(5);
		// folds.add(10);
		// folds.add(20);

		List<Integer> ks = new ArrayList<>();
		ks.add(0);
		ks.add(2);
		ks.add(10);
		ks.add(50);

		List<Integer> oobRates = new ArrayList<>();
		// oobRates.add(85);
		// oobRates.add(90);
		// oobRates.add(95);
		// oobRates.add(75);
		// oobRates.add(50);

		// oobRates.add(10);
		// oobRates.add(20);
		// oobRates.add(30);
		// oobRates.add(40);
		// oobRates.add(50);
		oobRates.add(60);
		// oobRates.add(70);
		oobRates.add(80);
		oobRates.add(90);
		// oobRates.add(100);

		Boolean numeric = null;

		numeric = null;

		while (numeric == null || numeric == true) {
			if (numeric == null) {
				numeric = true;
			} else {
				numeric = false;
			}

		for (Integer oobRate : oobRates) {

			List<Parameters> parameterList = new ArrayList<>();

			Boolean descretize = null;
			if (!descretizeOn) {
				descretize = true;
			}

			while (descretize == null || descretize == true) {
				if (descretize == null) {
					descretize = true;
				} else {
					descretize = false;
				}



				for (int minVar = 2; minVar < 4; minVar++) {
					// int minVar = 2;
					for (Integer k : ks) {
						// int k = 2;
						// for (Integer n : folds) {
						int n = 0;
						// int depth = 0;
						// int minPerLeaf = 0;
						for (Integer depth : depths) {

							for (Integer minPerLeaf : minInstancesPerLeaf) {

								String[] parameters = new String[19];

								parameters[0] = "-B";
								parameters[1] = "-U";
								parameters[2] = "-K";
								parameters[3] = String.valueOf(k);
								parameters[4] = "-depth";
								parameters[5] = String.valueOf(depth);
								parameters[6] = "-N";
								parameters[7] = String.valueOf(n);
								parameters[8] = "-M";
								parameters[9] = String.valueOf(minPerLeaf);
								parameters[10] = "-V";
								parameters[11] = "1e-" + minVar;
								parameters[12] = "-P";
								parameters[13] = String.valueOf(oobRate);
								parameters[14] = "-O";
								parameters[15] = "-store-out-of-bag-predictions";
								parameters[16] = "-I";
									parameters[17] = "800";
								parameters[18] = "-attribute-importance";

								parameterList.add(new Parameters(parameters, numeric, descretize));

							}

						}
					}
				}

				}

				// }


			ConcurrentLinkedQueue<Pair<Double, Parameters>> queue = new ConcurrentLinkedQueue<>();

			PersistentParallel6.runAndFinish(parameterList, (input) -> {
				double sum = 0.0;
				int count = 0;

				for (ValidationRun validationRun : validationRuns) {
					count++;

					sum += run.run(validationRun, input);
				}

				queue.add(new Pair<>(sum / count, input));
			}, this.getClass());

			double minScore = 0.0;
			Parameters min = null;
			for (Pair<Double, Parameters> pair : queue) {
				if (min == null || pair.getFirst() < minScore) {
					minScore = pair.getFirst();
					min = pair.getSecond();
				}
			}
				System.out.println();
				System.out.println();
				System.out.println("RESULT ========================");
				System.out.println("oobRate: " + oobRate);
				System.out.println("numeric: " + numeric);

			System.out.println(minScore);
			System.out.println(min.toString());

		}
		}
	};

	public void findHyperParameters_cluster(List<ValidationRun> validationRuns, RandomForestRun run) {

		// ClusterHyperParameters:
		// 0.06479908867160382
		// Parameters [parameters=[, , -K, 0, -depth, 0, -N, 0, -M, 0, -V, 1e-4,
		// -P, 85, -O, -store-out-of-bag-predictions, -I, 150], numeric=true]

		List<Integer> oobRates = new ArrayList<>();
		// oobRates.add(50);
		// oobRates.add(75);
		// oobRates.add(85);
		// oobRates.add(90);
		// oobRates.add(95);

		oobRates.add(10);
		oobRates.add(20);
		oobRates.add(30);
		oobRates.add(40);
		oobRates.add(50);
		oobRates.add(60);
		oobRates.add(70);
		oobRates.add(80);
		oobRates.add(90);
		oobRates.add(100);

		List<Parameters> parameterList = new ArrayList<>();

		// int minVar = 4;
		// boolean numeric = true;
		// int k = 3;

		Boolean numeric = null;

		while (numeric == null || numeric == true) {
			if (numeric == null) {
				numeric = true;
			} else {
				numeric = false;
			}
			int minVar = 3;
			// for (int minVar = 1; minVar < 5; minVar++) {

			// for (int k = 0; k < 10; k++) {
			int k = 2;
			for (Integer oobRate : oobRates) {
				String[] parameters = new String[18];

				parameters[0] = "";
				parameters[1] = "-U";
				parameters[2] = "-K";
				parameters[3] = String.valueOf(k);
				parameters[4] = "";
				parameters[5] = "";
				parameters[6] = "";
				parameters[7] = "";
				parameters[8] = "";
				parameters[9] = "";
				parameters[10] = "-V";
				parameters[11] = "1e-" + minVar;
				parameters[12] = "-P";
				parameters[13] = String.valueOf(oobRate);
				parameters[14] = "-O";
				parameters[15] = "-store-out-of-bag-predictions";
				parameters[16] = "-I";
				parameters[17] = "350";

				parameterList.add(new Parameters(parameters, numeric, false));
			}
			// }

			// }

		}

		ConcurrentLinkedQueue<Pair<Double, Parameters>> queue = new ConcurrentLinkedQueue<>();

		PersistentParallel6.runAndFinish(parameterList, (input) -> {
			double sum = 0.0;
			int count = 0;

			for (ValidationRun validationRun : validationRuns) {
				count++;

				sum += run.run(validationRun, input);
			}

			queue.add(new Pair<>(sum / count, input));
		}, this.getClass(), 80);

		for (Pair<Double, Parameters> pair : queue) {
			System.out.println(pair.getFirst());
			System.out.println(pair.getSecond().toString());
		}

		double minScore = 0.0;
		Parameters min = null;
		for (Pair<Double, Parameters> pair : queue) {
			if (min == null || pair.getFirst() < minScore) {
				minScore = pair.getFirst();
				min = pair.getSecond();
			}
		}
		System.out.println(minScore);
		System.out.println(min.toString());

	};

	public static class ValidationRun {

		private final UUID uuid = UUID.randomUUID();
		private final GoldStandardClusters gs;
		private final Set<TableRowId> candidates;
		private final Set<InstanceId> newEntities;
		private final ClassId classId;
		private final TableRowIdsSet fullInput;
		private final TableRowIdsSet fullInputAllClasses;

		public ValidationRun(GoldStandardClusters gs, Set<TableRowId> candidates, Set<InstanceId> newEntities, ClassId classId, TableRowIdsSet fullInput,
				TableRowIdsSet fullInputAllClasses) {
			super();
			this.gs = gs;
			this.candidates = candidates;
			this.newEntities = newEntities;
			this.classId = classId;
			this.fullInput = fullInput;
			this.fullInputAllClasses = fullInputAllClasses;
		}

		public TableRowIdsSet getFullInputAllClasses() {
			return fullInputAllClasses;
		}

		public GoldStandardClusters getGs() {
			return gs;
		}

		public Set<TableRowId> getCandidates() {
			return candidates;
		}

		public Set<InstanceId> getNewEntities() {
			return newEntities;
		}

		public ClassId getClassId() {
			return classId;
		}

		public TableRowIdsSet getFullInput() {
			return fullInput;
		}

		public UUID getUuid() {
			return uuid;
		}

	}

	public static class Parameters {

		private final String[] parameters;
		private final boolean numeric;
		private final boolean descretize;

		public Parameters(String[] parameters, boolean numeric, boolean descretize) {
			super();
			this.parameters = parameters;
			this.numeric = numeric;
			this.descretize = descretize;
		}

		public String[] getParameters() {
			return parameters;
		}

		public boolean isNumeric() {
			return numeric;
		}

		public boolean isDescretize() {
			return descretize;
		}

		@Override
		public String toString() {
			return "Parameters [parameters=" + Arrays.toString(parameters) + ", numeric=" + numeric + ", descretize=" + descretize + "]";
		}

	}

	@FunctionalInterface
	public static interface RandomForestRun {
		double run(ValidationRun validationRun, Parameters parameters);
	}
}
