package expansion.set.implicitAttributes;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.PropertyValuePair;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.matching.LabelMatcher;
import de.uni_mannheim.informatik.dws.matching.LabelMatcher.MatchStrictness;
import expansion.set.entitymatching.candidateselection.strategies.CandidateSelection_Label;

public class ImplicitAttributeFinder {

	@Inject
	protected CandidateSelection_Label candidateSelection;

	@Inject
	protected AnnotatedTablesRepository repository;

	@Inject
	protected LabelMatcher labelMatcher;

	@Inject
	protected KnowledgeBase knowledgeBase;

	@Inject
	public ImplicitAttributeFinder() {

	}

	public static SimpleLockableConcurrentCache<AnnotatedTableId, ImplicitAttributes> cache = new SimpleLockableConcurrentCache<>();
	public static SimpleLockableConcurrentCache<AnnotatedTableId, ImplicitAttributes> cachePMI = new SimpleLockableConcurrentCache<>();

	private static Map<ClassId, Map<PropertyValuePair, Double>> pmiScores;

	public void prepare(Map<ClassId, Set<PropertyId>> nonTailProperties) {
		if (false && pmiScores == null) {
			Map<ClassId, Map<PropertyValuePair, Double>> pmiScoresMap = new HashMap<>();

			for (Entry<ClassId, Set<PropertyId>> classEntry : nonTailProperties.entrySet()) {



				ConcurrentHashMap<PropertyValuePair, AtomicInteger> countMap = new ConcurrentHashMap<>();
				// double size =
				// knowledgeBase.getNumberOfInstances(classEntry.getKey());

				PersistentParallel6.runAndFinish(knowledgeBase.getInstances(classEntry.getKey()).keySet(), (instanceId) -> {
					for (PropertyId propertyId : classEntry.getValue()) {
						QualifiedValueList qvl = knowledgeBase.getQualifiedValueList(instanceId, propertyId);
						if (qvl != null) {
							for (QualifiedValueHolder qvh : qvl.getClaims()) {
								CHMHelper.getFromCHM(countMap, new PropertyValuePair(propertyId, qvh.getValue()), () -> new AtomicInteger()).incrementAndGet();
							}
						}
					}
				}, this.getClass());

				// Collectors.averagingInt(mapper)

				Map<PropertyId, Double> propertySizes = countMap.entrySet().stream().collect(Collectors.groupingBy((e) -> e.getKey().getPropertyId()))
						.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
							return (double) e.getValue().stream().mapToInt((e2) -> e2.getValue().intValue()).max().getAsInt();
						}));

				// classEntry.getValue().stream().collect(Collectors.toMap(Function.identity(),
				// (e) -> {
				// return (double)
				// knowledgeBase.getInstancesSize(classEntry.getKey(), e);
				// }));

				Map<PropertyId, Double> propertyThresholds = propertySizes.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
					return Math.ceil((1.0 - 0.95) * e.getValue());
				}));

				double maxSize = knowledgeBase.getNumberOfInstances(classEntry.getKey());
				double threshold = (1.0 - 0.85) * maxSize;
				//
				Map<PropertyValuePair, Double> pmiScores = Collections
						.unmodifiableMap(
								countMap.entrySet().stream().filter((e) -> e.getValue().intValue() > threshold)
										.collect(Collectors.toMap(Entry::getKey, (e) -> {
											return Math.max(0.0, 1.0 - (e.getValue().intValue() / maxSize) - 0.85);
						})));

				pmiScoresMap.put(classEntry.getKey(), pmiScores);
			}

			this.pmiScores = Collections.unmodifiableMap(pmiScoresMap);

		}
	}

	public ImplicitAttributes get(AnnotatedTableId tableId, Map<ClassId, Set<ClassId>> allParents,  Set<PropertyId>  nonTailProperties ) {
		return cache.getOrCreate(tableId, () -> {
			return find(tableId, allParents, nonTailProperties, false, false);
		});
	}

	public ImplicitAttributes getPMI(AnnotatedTableId tableId, Map<ClassId, Set<ClassId>> allParents, Set<PropertyId> nonTailProperties) {
		return cachePMI.getOrCreate(tableId, () -> {
			return find(tableId, allParents, nonTailProperties, false, true);
		});
	}

	public ImplicitAttributes find(AnnotatedTableId tableId, Map<ClassId, Set<ClassId>> allParents, Set<PropertyId> nonTailProperties, boolean sequential) {
		return find(tableId, allParents, nonTailProperties, sequential, false);
	}

	public ImplicitAttributes find(AnnotatedTableId tableId, Map<ClassId, Set<ClassId>> allParents, Set<PropertyId> nonTailProperties, boolean sequential,
			boolean pmi) {
		List<TableRowId> rowIds = repository.getRowIds(tableId);
		ClassId classId = repository.getMetaData(tableId).getDbPediaClass(); 

		ConcurrentHashMap<PropertyValuePair, Set<RowIdx>> countMap = new ConcurrentHashMap<>();
		AtomicInteger ai = new AtomicInteger();
		PersistentParallel6.runAndFinish(rowIds, (rowId) -> {
			StringObject label = repository.getLabel(rowId);
			if (label != null) {
			Set<String> labels = new HashSet<>();
			labels.add(label.getString());
				Optional<Map<InstanceId, Tri<Double, Double, Double>>> matches = labelMatcher.getMatchesFull(labels, allParents.get(classId),
						MatchStrictness.NON_STRICT);

				if (matches.isPresent()) {
					ai.incrementAndGet();
					for (Entry<InstanceId, Tri<Double, Double, Double>> match : matches.get().entrySet()) {
						for (PropertyId propertyId : nonTailProperties) {
							ValueId vid = new ValueId(match.getKey(), propertyId);
							if (knowledgeBase.hasValue(vid)) {
								for (QualifiedValueHolder claim : knowledgeBase.getQualifiedValueList(vid).getClaims()) {
									CHMHelper.getFromCHM(countMap, new PropertyValuePair(propertyId, claim.getValue()),
											() -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(rowId.getRowIdx());

								}
							}
						}
					}
				}
			}
		}, this.getClass(), rowIds.size() < 1500 || sequential);
		
		// int rowIdsSizeWithMatch = ai.intValue();
		int rowIdsSizeWithMatch = rowIds.size();

		double filter = 0.5 * rowIds.size();

		Map<PropertyValuePair, Double> retMap = countMap.entrySet().stream().filter((e) -> {
			return e.getValue().size() > filter;
		}).collect(Collectors.toMap(Entry::getKey, (e) -> {

			return ((double) e.getValue().size()) / rowIdsSizeWithMatch;
		}));
		
		if (false && pmi) {
			Map<PropertyValuePair, Double> retMap2 = retMap.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				Double d = pmiScores.get(classId).get(e.getKey());
				if (d == null) {
					return e.getValue();
				} else {
					return e.getValue() * d;
				}
			}));

			return new ImplicitAttributes(retMap2);
		} else {

			return new ImplicitAttributes(retMap);
		}
	}
}

