package expansion.set.learning;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationMethod;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.RecordPrediction;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordData;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordFeature;
import de.uni_mannheim.informatik.dws.expansion.utils.learning.weka.WekaLearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;

public class ET_WekaLearnedModel implements LearnedModel{
	
	private final Map<List<Integer>, WekaLearnedModel> learnedModels;
	private final WeightsRepository weights;
	private final OptimizableClass optimizableClass;

	public ET_WekaLearnedModel(Map<List<Integer>, WekaLearnedModel> learnedModels, WeightsRepository weights, OptimizableClass optimizableClass) {
		this.learnedModels = Collections.unmodifiableMap(new HashMap<>(learnedModels));
		this.weights = weights;
		this.optimizableClass = optimizableClass;
	}

	public ET_WekaLearnedModel(Map<List<Integer>, WekaLearnedModel> learnedModels, OptimizableClass optimizableClass) {
		this.learnedModels = Collections.unmodifiableMap(new HashMap<>(learnedModels));
		this.weights = null;
		this.optimizableClass = optimizableClass;
	}

	public ET_WekaLearnedModel(ET_WekaLearnedModel casted, WeightsRepository weights) {
		this(casted.learnedModels, weights, casted.optimizableClass);
	}

	public WekaLearnedModel get(List<Integer> list) {
		return learnedModels.get(list);
	}

	@Override
	public RecordPrediction apply(RecordFeature d) {
		throw new UnsupportedOperationException();
	}

	@Override
	public byte[] outputModelFile() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public double runAndEvaluate(Collection<RecordData> data, EvaluationMethod evaluator) {
		throw new UnsupportedOperationException();
	}

	@Override
	public Optional<Double> getNormalizingThreshold() {
		return Optional.empty();
	}

	public Optional<Double> getPositiveConfidenceThreshold() {
		if(weights == null) {
			return Optional.empty();
		} else {
			return weights.get(optimizableClass.getWeight("POSITIVE_CONFIDENCE_THRESHOLD", 0.0));
		}
	}

	public Optional<Double> getNegativeConfidenceThreshold() {
		if (weights == null) {
			return Optional.empty();
		} else {
			return weights.get(optimizableClass.getWeight("NEGATIVE_CONFIDENCE_THRESHOLD", 0.0));
		}
	}

	public Optional<Double> getClusteringThreshold() {
		// return Optional.empty();
		if (weights == null) {
			return Optional.empty();
		} else {
			return weights.get(optimizableClass.getWeight("CLUSTERING_THRESHOLD", 0.0));
		}
	}

	public Optional<Double> getVeryLowPairThreshold() {
		// return Optional.empty();
		if (weights == null) {
			return Optional.empty();
		} else {
			return weights.get(optimizableClass.getWeight("VERY_LOW_PAIR_THRESHOLD", 0.0));
		}
	}

}
