package expansion.set.learning;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;
import java.util.TreeMap;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.inject.Inject;

import com.google.common.primitives.Ints;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdsSet;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.learner.WeightsLearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.LearnedModelRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeightKey;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepositoryImpl;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusterer;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.evaluation.SetExpansionStrategiesConfig;
import expansion.set.evaluation.goldstandard.GoldStandardClusters;
import expansion.set.evaluation.goldstandard.GoldStandardClustersSet;
import expansion.set.evaluation.goldstandard.GoldStandardClustersSet.GoldStandardData;
import expansion.set.evaluation.goldstandard.GoldStandardClustersSetPerClass;
import expansion.set.strategies.SetExpansionStrategiesRepository;
import expansion.set.strategies.rowclusterers.GreedyClusteringSplitter;
import expansion.set.strategies.rowclusterers.RowClustererRepository;
import expansion.set.strategies.rowclusterers.label.LabelRowSimilaritiesCreator;
import expansion.set.strategies.rowclusterers.label.LabelSimilarityClustererClassBlocked;

public class GoldStandardSplitter {

	@Inject
	protected LabelSimilarityClustererClassBlocked clusterer;
	
	@Inject
	protected EqualSetSplitter equalSetSplitter;

	@Inject
	protected LabelRowSimilaritiesCreator labelSimCreator;

	public GoldStandardSplitter() {
	}
	
	public GoldStandardClustersSetPerClass split(Map<ClassId, GoldStandardClusters> goldStandardClusters, Map<ClassId, Set<InstanceId>> newEntityIds,
			Map<ClassId, Map<QualifiedTripleId, QualifiedValueList>> factsMap,
			int denominator) {

		Set<TableRowId> rowIdsAllClasses = new HashSet<>();
		for (Entry<ClassId, GoldStandardClusters> classId : goldStandardClusters.entrySet()) {

			rowIdsAllClasses.addAll(classId.getValue().getCandidates());
		}
		TableRowIdsSet rowIdsSetAllClasses = new TableRowIdsSet(rowIdsAllClasses, new UuidStringId("ALL"));

		for (Entry<ClassId, GoldStandardClusters> entry : goldStandardClusters.entrySet()) {

		}

		goldStandardClusters.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> new GoldStandardClusters(e.getValue().getIndex())));
		Map<ClassId, GoldStandardClustersSet> ret = new HashMap<>();
		for (Entry<ClassId, GoldStandardClusters> entry : goldStandardClusters.entrySet()) {

			if (denominator > 5) {
				System.out.println("CROSS VALIDATION WITH k LARGER THAN 5, I AM IGNORING CONFLICTING SETS");

				GoldStandardClustersSet set = splitEntitiesOnly(entry.getKey(), entry.getValue(), newEntityIds.get(entry.getKey()),
						factsMap.get(entry.getKey()), denominator, rowIdsSetAllClasses);

				ret.put(entry.getKey(), set);
			} else {
				GoldStandardClustersSet set = split(entry.getKey(), entry.getValue(), newEntityIds.get(entry.getKey()), factsMap.get(entry.getKey()),
						denominator, rowIdsSetAllClasses);

				ret.put(entry.getKey(), set);
			}

		}

		for (Entry<ClassId, GoldStandardClustersSet> entry : ret.entrySet()) {
			System.out.println();
			System.out.println();
			System.out.println();
			System.out.println("===========================================");
			System.out.println("===========================================");
			System.out.println(entry.getKey().getFinalPart());

			for (int i = 0; i < entry.getValue().getSetCount(); i++) {
				System.out.println();
				System.out.println();
				System.out.println("===========================================");
				System.out.println("Fold: " + i);

				for (InstanceId instanceId : entry.getValue().getSet(i).getClusters().getTestingSetMappingClusters().keySet()) {
					System.out.println(instanceId.getInstanceId());
				}

				System.out.println();
				System.out.println();
			}

		}

		GoldStandardClustersSetPerClass retret = new GoldStandardClustersSetPerClass(ret);

		try {
			GoldStandardClustersSet.statistics(retret);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return retret;
	}

	private GoldStandardClustersSet splitEntitiesOnly(ClassId classId, GoldStandardClusters full, Set<InstanceId> newEntityIds,
			Map<QualifiedTripleId, QualifiedValueList> factsMap, int denominator, TableRowIdsSet fullInputAllClasses) {

		Random random = new Random(classId.getFinalPart().hashCode());
		List<InstanceId> list = full.getTestingSetMappingClusters().keySet().stream().sorted((e1, e2) -> {
			return e1.getInstanceId().compareTo(e2.getInstanceId());
		}).collect(Collectors.toList());
		Collections.shuffle(list, random);

		List<Set<InstanceId>> conflictingSets = new ArrayList<>(getConflictingSets(full, classId, fullInputAllClasses));

		Map<Integer, Set<InstanceId>> setMap = new HashMap<>();

		int totalNew = newEntityIds.size();
		int totalExisting = full.getTestingSetMappingClusters().keySet().size() - totalNew;

		int newPerFold = Math.round(((float) totalNew) / denominator);
		int existingPerFold = Math.round(((float) totalExisting) / denominator);

		int currentNewFold = 0;
		int currentExistingFold = 0;

		int sumInCurrentNewFold = 0;
		int sumInCurrentExistingFold = 0;

		for (InstanceId instanceId : list) {
			if (newEntityIds.contains(instanceId)) {
				CHMHelper.getFromMap(setMap, currentNewFold, () -> new HashSet<>()).add(instanceId);
				sumInCurrentNewFold++;
				if (sumInCurrentNewFold >= newPerFold) {
					currentNewFold++;
					sumInCurrentNewFold = 0;
				}
			} else {
				CHMHelper.getFromMap(setMap, currentExistingFold, () -> new HashSet<>()).add(instanceId);
				sumInCurrentExistingFold++;
				if (sumInCurrentExistingFold >= existingPerFold) {
					currentExistingFold++;
					sumInCurrentExistingFold = 0;
				}
			}
		}

		int redistributeFold = 0;
		for (InstanceId instanceId : setMap.get(denominator)) {
			CHMHelper.getFromMap(setMap, redistributeFold, () -> new HashSet<>()).add(instanceId);
			redistributeFold++;

		}

		setMap.remove(denominator);

		List<GoldStandardData> data = new ArrayList<>();
		Map<InstanceId, Set<TableRowId>> fullGsMap = full.getTestingSetMappingClusters();

		for (Set<InstanceId> instanceIds : setMap.values()) {

			Map<InstanceId, Set<TableRowId>> gsMap = instanceIds.stream().collect(Collectors.toMap(Function.identity(), (e) -> fullGsMap.get(e)));
			Set<InstanceId> setNewEntityIds = newEntityIds.stream().filter((e) -> instanceIds.contains(e)).collect(Collectors.toSet());

			Map<QualifiedTripleId, QualifiedValueList> setFactsMap = factsMap.entrySet().stream()
					.filter((e) -> instanceIds.contains(e.getKey().getTripleId().getInstanceId())).collect(Collectors.toMap(Entry::getKey, Entry::getValue));

			data.add(new GoldStandardData(new GoldStandardClusters(gsMap, true), setNewEntityIds, setFactsMap));

		}

		GoldStandardData fullData = new GoldStandardData(full, newEntityIds, factsMap);

		return new GoldStandardClustersSet(fullData, data, conflictingSets);

	}

	public GoldStandardClustersSet split(ClassId classId, GoldStandardClusters full, Set<InstanceId> newEntityIds,
			Map<QualifiedTripleId, QualifiedValueList> factsMap, int denominator, TableRowIdsSet fullInputAllClasses) {
		
		Random random = new Random(classId.getFinalPart().hashCode());
		
		List<InstanceId> list = full.getTestingSetMappingClusters().keySet().stream().sorted((e1, e2) -> {
			return e1.getInstanceId().compareTo(e2.getInstanceId());
		}).collect(Collectors.toList());
		Collections.shuffle(list, random);
		List<Set<InstanceId>> conflictingSets;

		conflictingSets = new ArrayList<>(getConflictingSets(full, classId, fullInputAllClasses));
		
		conflictingSets = conflictingSets.stream().map((e) -> list.stream().filter((e2) -> e.contains(e2)).collect(Collectors.toCollection(LinkedHashSet::new)))
				.collect(Collectors.toList());

		Collections.sort(conflictingSets, (e1, e2) -> {
			int compare = Integer.compare(e1.stream().mapToInt((e) -> e.hashCode()).sum(), e2.stream().mapToInt((e) -> e.hashCode()).sum());
			if (compare == 0) {
				compare = e1.iterator().next().getInstanceId().compareTo(e2.iterator().next().getInstanceId());
			}
			return compare;
		});
	
		
		
		Map<Integer, Set<Integer>> distribution = equalSetSplitter.get(conflictingSets.stream().map((e) -> e.size()).collect(Collectors.toList()),
				conflictingSets.stream().map((e) -> e.stream().mapToInt((e2) -> full.getCluster(e2).size()).sum()).collect(Collectors.toList()),
				 denominator);
		
		
		int conflictHashCode = conflictingSets.hashCode();

		double part = (1.0 / denominator);
		int minClustersInLearning = Ints.checkedCast(Math.round(full.getTestingSetMappingClusters().keySet().size() * part));
		int minNewClustersInLearning = Ints.checkedCast(Math.round(newEntityIds.size() * part));
		int minConflictInLearning = Ints.checkedCast(Math.round(conflictingSets.stream().mapToInt(Collection::size).sum() * part));
		
		



		
	 Collections.sort(conflictingSets, (e1, e2) -> {
		 int compare =  Integer.compare(e1.size(), e2.size());
		 if(compare != 0) {
			 return compare;
		 }
		 compare = Integer.compare(e1.hashCode(), e2.hashCode());
		 if(compare != 0) {
			 return compare;
		 }
		 throw new IllegalStateException("I CAN NOT CREATE AN ORDERING AS THERE ARE TWO SETS WITH EQUAL SIZE AND HASHCODE. GO FIX THE STUPID CODE.");
	 });
	 
	 
		Map<InstanceId, Integer> conflictingSetsIndex = new HashMap<>();
		for(ListEntry<Set<InstanceId>> entry : ListIterable.iterable(conflictingSets)) {
			for (InstanceId instanceId : entry.getValue()) {
				conflictingSetsIndex.put(instanceId, entry.getIndex());
			}
		}
 
		// int currentSet = 0;

		Map<Integer, Tri<Set<InstanceId>, Set<InstanceId>, Set<InstanceId>>> setMap = new TreeMap<>();
		Set<InstanceId> added = new HashSet<>();

		for (Entry<Integer, Set<Integer>> entry : distribution.entrySet()) {
			Tri<Set<InstanceId>, Set<InstanceId>, Set<InstanceId>> addedTri = CHMHelper.getFromMap(setMap, entry.getKey(), () -> {
				return new Tri<>(new HashSet<>(), new HashSet<>(), new HashSet<>());
			});

			Set<InstanceId> conflictAdded = addedTri.getA();
			Set<InstanceId> newAdded = addedTri.getB();
			Set<InstanceId> totalAdded = addedTri.getC();

			for (Integer i : entry.getValue()) {
				Set<InstanceId> set = conflictingSets.get(i);

				for (InstanceId instanceId : set) {
					boolean isNew = newEntityIds.contains(instanceId);

					added.add(instanceId);
					totalAdded.add(instanceId);
					if (isNew) {
						newAdded.add(instanceId);
					}

					conflictAdded.add(instanceId);

				}
			}

		}

		List<InstanceId> remainingNewClusters = list.stream().filter((e) -> newEntityIds.contains(e) && !added.contains(e)).collect(Collectors.toList());

		List<Integer> newClustersSizes = ListIterable.iterable(remainingNewClusters).stream().map((e) -> full.getCluster(e.getValue()).size())
				.collect(Collectors.toList());

		Map<Integer, Pair<Integer, Integer>> existingCountAndSize = setMap.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			int count = 0;
			int size = 0;
			for (InstanceId instanceId : e.getValue().getB()) {
				count++;
				size = full.getCluster(instanceId).size();
			}
			return new Pair<>(count, size);
		}));

		Map<Integer, Set<Integer>> newDistribution = equalSetSplitter.getSetNumberPriority(newClustersSizes, denominator, existingCountAndSize);

		for (Entry<Integer, Set<Integer>> entry : newDistribution.entrySet()) {
			Tri<Set<InstanceId>, Set<InstanceId>, Set<InstanceId>> addedTri = CHMHelper.getFromMap(setMap, entry.getKey(), () -> {
				return new Tri<>(new HashSet<>(), new HashSet<>(), new HashSet<>());
			});

			Set<InstanceId> conflictAdded = addedTri.getA();
			Set<InstanceId> newAdded = addedTri.getB();
			Set<InstanceId> totalAdded = addedTri.getC();

			for (Integer i : entry.getValue()) {
				InstanceId instanceId = remainingNewClusters.get(i);

					added.add(instanceId);
					totalAdded.add(instanceId);
					newAdded.add(instanceId);

			}

		}

		List<InstanceId> remainingClusters = list.stream().filter((e) -> !added.contains(e)).collect(Collectors.toList());
		List<Integer> remainingClustersSizes = ListIterable.iterable(remainingClusters).stream().map((e) -> full.getCluster(e.getValue()).size())
				.collect(Collectors.toList());

		existingCountAndSize = setMap.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			int count = 0;
			int size = 0;
			for (InstanceId instanceId : e.getValue().getB()) {
				count++;
				size = full.getCluster(instanceId).size();
			}
			return new Pair<>(count, size);
		}));

		Map<Integer, Set<Integer>> remainingDistribution = equalSetSplitter.getSetNumberPriority(remainingClustersSizes, denominator, existingCountAndSize);

		for (Entry<Integer, Set<Integer>> entry : remainingDistribution.entrySet()) {
			Tri<Set<InstanceId>, Set<InstanceId>, Set<InstanceId>> addedTri = CHMHelper.getFromMap(setMap, entry.getKey(), () -> {
					return new Tri<>(new HashSet<>(), new HashSet<>(), new HashSet<>());
				});

				Set<InstanceId> conflictAdded = addedTri.getA();
				Set<InstanceId> newAdded = addedTri.getB();
				Set<InstanceId> totalAdded = addedTri.getC();

			for (Integer i : entry.getValue()) {
				InstanceId instanceId = remainingClusters.get(i);


					added.add(instanceId);
					totalAdded.add(instanceId);



				}

			}

		// int currentSet = 0;
		// for (InstanceId instanceId : list) {
		// int currentTry = 0;
		// while (!added.contains(instanceId)) {
		//
		// Tri<Set<InstanceId>, Set<InstanceId>, Set<InstanceId>> addedTri =
		// CHMHelper.getFromMap(setMap, currentSet, () -> {
		// return new Tri<>(new HashSet<>(), new HashSet<>(), new HashSet<>());
		// });
		//
		// Set<InstanceId> conflictAdded = addedTri.getA();
		// Set<InstanceId> newAdded = addedTri.getB();
		// Set<InstanceId> totalAdded = addedTri.getC();
		//
		// boolean isNew = newEntityIds.contains(instanceId);
		// boolean isConflict = conflictingSetsIndex.containsKey(instanceId);
		// Set<InstanceId> conflictSet = null;
		// if(isConflict)
		// conflictSet =
		// conflictingSets.get(conflictingSetsIndex.get(instanceId));
		//
		// boolean add = false;
		// if (currentTry == denominator) {
		// add = true;
		// }
		// if (totalAdded.size() < minClustersInLearning) {
		// if (!isNew || (newAdded.size() < minNewClustersInLearning)) {
		// if (!isConflict || ( (conflictAdded.size() + conflictSet.size()) <=
		// minConflictInLearning)) {
		// add = true;
		// }
		// }
		// }
		// if (isNew && newAdded.size() < minNewClustersInLearning) {
		// if (!isConflict || ( (conflictAdded.size() + conflictSet.size()) <=
		// minConflictInLearning)) {
		// add = true;
		//
		// }
		// }
		// if (isConflict && ((conflictAdded.size() + conflictSet.size()) <=
		// minConflictInLearning)) {
		// add = true;
		// }
		//
		// if (add) {
		// added.add(instanceId);
		// totalAdded.add(instanceId);
		// if (isNew) {
		// newAdded.add(instanceId);
		// }
		// if (isConflict) {
		// int conflictIndex = conflictingSetsIndex.get(instanceId);
		// // totalAdded.add(instanceId);
		// conflictAdded.add(instanceId);
		// for (InstanceId additionalInstanceId :
		// conflictingSets.get(conflictIndex)) {
		// added.add(additionalInstanceId);
		// totalAdded.add(additionalInstanceId);
		// conflictAdded.add(additionalInstanceId);
		// if (newEntityIds.contains(additionalInstanceId)) {
		// newAdded.add(additionalInstanceId);
		// }
		// }
		// }
		// }
		//
		// currentSet++;
		// if (currentSet == denominator) {
		// currentSet = 0;
		// }
		// currentTry++;
		// }
		//
		// }

		// Map<Integer, GoldStandardClusters> setsGsMap = new HashMap<>();
		//
		// Map<Integer, GoldStandardClusters> setsNewEntityMap = new
		// HashMap<>();
		// Map<Integer, GoldStandardClusters> setsFactsMap = new HashMap<>();
		//
		// Map<InstanceId, Set<TableRowId>> testingMap = new HashMap<>();
		// Map<InstanceId, Set<TableRowId>> learningMap = new HashMap<>();

		List<GoldStandardData> data = new ArrayList<>();
		Map<InstanceId, Set<TableRowId>> fullGsMap = full.getTestingSetMappingClusters();

		for (Entry<Integer, Tri<Set<InstanceId>, Set<InstanceId>, Set<InstanceId>>> entry : setMap.entrySet()) {

			Set<InstanceId> instanceIds = entry.getValue().getC();

			Map<InstanceId, Set<TableRowId>> gsMap = instanceIds.stream().collect(Collectors.toMap(Function.identity(), (e) -> fullGsMap.get(e)));
			Set<InstanceId> setNewEntityIds = newEntityIds.stream().filter((e) -> instanceIds.contains(e)).collect(Collectors.toSet());

			Map<QualifiedTripleId, QualifiedValueList> setFactsMap = factsMap.entrySet().stream()
					.filter((e) -> instanceIds.contains(e.getKey().getTripleId().getInstanceId())).collect(Collectors.toMap(Entry::getKey, Entry::getValue));

			data.add(new GoldStandardData(new GoldStandardClusters(gsMap, true), setNewEntityIds, setFactsMap));

		}

		// for (Entry<InstanceId, Set<TableRowId>> entry :
		// full.getTestingSetMappingClusters().entrySet()) {
		// if (totalAdded.contains(entry.getKey())) {
		// learningMap.put(entry.getKey(), entry.getValue());
		// } else {
		// testingMap.put(entry.getKey(), entry.getValue());
		// }
		// }
		//
		// Set<InstanceId> learningNewEntityIds =
		// newEntityIds.stream().filter((e) ->
		// learningMap.containsKey(e)).collect(Collectors.toSet());
		// Set<InstanceId> testingNewEntityIds =
		// newEntityIds.stream().filter((e) ->
		// testingMap.containsKey(e)).collect(Collectors.toSet());
		//
		// Map<QualifiedTripleId, QualifiedValueList> learningFactsMap =
		// factsMap.entrySet().stream()
		// .filter((e) ->
		// learningMap.containsKey(e.getKey().getTripleId().getInstanceId())).collect(Collectors.toMap(Entry::getKey,
		// Entry::getValue));
		// Map<QualifiedTripleId, QualifiedValueList> testingFactsMap =
		// factsMap.entrySet().stream()
		// .filter((e) ->
		// testingMap.containsKey(e.getKey().getTripleId().getInstanceId())).collect(Collectors.toMap(Entry::getKey,
		// Entry::getValue));

		GoldStandardData fullData = new GoldStandardData(full, newEntityIds, factsMap);

		return new GoldStandardClustersSet(fullData, data, conflictingSets);
		// return new GoldStandardClustersSet(full, new
		// GoldStandardClusters(learningMap, true), new
		// GoldStandardClusters(testingMap, true), newEntityIds,
		// learningNewEntityIds, testingNewEntityIds, factsMap,
		// learningFactsMap, testingFactsMap, conflictingSets);

	}
	
	@Inject
	protected RowClustererRepository rowClustererRepository;

	@Inject
	protected GreedyClusteringSplitter greedyClusteringSplitter;
	
	List<Set<InstanceId>> getConflictingSets(GoldStandardClusters full, ClassId classId, TableRowIdsSet fullInputAllClasses) {
		List<Set<InstanceId>> ret = new ArrayList<>();
		 
		// Random random = new Random(classId.getFinalPart().hashCode());
		
		Set<TableRowId> list = full.getCandidates().stream().sorted((e1, e2) -> {
			return e1.getCombinedString().compareTo(e2.getCombinedString());
		}).collect(Collectors.toCollection(LinkedHashSet::new));
		
		RowClusterer clusterer = rowClustererRepository.get(SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY.getClusterer());
		OptimizableClass optimizableClassClustering = SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY.getOptimizableClassCluster();
		OptimizableClass optimizableClassClusteringSimilarity = SetExpansionStrategiesRepository.FULL_LABEL_SIMILARITY.getOptimizableClassSimilarityThreshold();

		SetExpansionStrategiesConfig setExpansionStrategiesConfig =SetExpansionStrategiesConfig.defaultConfig();
		
		Map<OptimizableWeightKey, Double> weights = new HashMap<>();
		weights.put(optimizableClassClustering.getKey("LABEL"), 1.0);
		weights.put(optimizableClassClustering.getKey("ZERO"), 0.0);
		weights.put(optimizableClassClusteringSimilarity.getKey("SIMILARITY_THRESHOLD"), 0.40);
		
		Map<OptimizableClass, LearnedModel> map = new HashMap<>();
		map.put(optimizableClassClusteringSimilarity, new WeightsLearnedModel(optimizableClassClusteringSimilarity, new WeightsRepositoryImpl(weights)));

		RowClusteringConfig rowClusteringConfig = new RowClusteringConfig(setExpansionStrategiesConfig, new ArrayList<>(), new LearnedModelRepository(map),
				null);

		ClustererInput clustererInput = new ClustererInput(
				new SplitGroupedValuesSets(new SplitGroupedValues(new HashMap<>()), new SplitGroupedValues(new HashMap<>()),
						new SplitGroupedValues(new HashMap<>())),
				list, new TableRowIdsSet(list, new UuidStringId(classId.getFinalPart())),
				new TableIdsSet(full.getCandidates().stream().map((e) -> e.getTableId()).collect(Collectors.toSet())), null, classId, fullInputAllClasses);
		RowClusterSimilarities sim = clusterer.clusterRows(rowClusteringConfig, clustererInput);

		sim = labelSimCreator.createForGs(clustererInput.getFullCandidates());

		ClusterList clusters = greedyClusteringSplitter.clusterRows(sim, rowClusteringConfig, clustererInput, optimizableClassClusteringSimilarity, true, true);
		
		
//		ClusterList clusters = clusterer.getDirect(UUID.randomUUID(), full.getCandidates());
		
		for( RowCluster cluster : clusters.getClusterList()) {
			Set<InstanceId> conflict = new HashSet<>();
			for(TableRowId row : cluster.getRows()) {
				conflict.add(full.getMapping(row));
			}
			if (conflict.size() > 1)
				ret.add(conflict);
		}
		
		return ret; 
		
	}

}
