package expansion.set.learning;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationMethod;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.RecordPrediction;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordData;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordFeature;
import de.uni_mannheim.informatik.dws.expansion.utils.learning.weka.WekaLearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;
import weka.core.SerializationHelper;

public class MT_WekaLearnedModel implements LearnedModel{
	
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private final Map<List<Boolean>, WekaLearnedModel> learnedModels;
	private final WeightsRepository weights;
	private final OptimizableClass optimizableClass;

	public MT_WekaLearnedModel(byte[] string) {
		ByteArrayInputStream stream = new ByteArrayInputStream(string);
		try {
			this.learnedModels = (Map<List<Boolean>, WekaLearnedModel>) weka.core.SerializationHelper.read(stream);
		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
		this.weights = null;
		this.optimizableClass = null;

	}
	public MT_WekaLearnedModel(Map<List<Boolean>, WekaLearnedModel> learnedModels, WeightsRepository weights, OptimizableClass optimizableClass) {
		this.learnedModels = Collections.unmodifiableMap(new HashMap<>(learnedModels));
		this.weights = weights;
		this.optimizableClass = optimizableClass;
	}

	public MT_WekaLearnedModel(Map<List<Boolean>, WekaLearnedModel> learnedModels, OptimizableClass optimizableClass) {
		this.learnedModels = Collections.unmodifiableMap(new HashMap<>(learnedModels));
		this.weights = null;
		this.optimizableClass = optimizableClass;
	}

	public MT_WekaLearnedModel(MT_WekaLearnedModel casted, WeightsRepository weights) {
		this(casted.learnedModels, weights, casted.optimizableClass);
	}

	public WekaLearnedModel get(List<Boolean> list) {
		return learnedModels.get(list);
	}

	@Override
	public RecordPrediction apply(RecordFeature d) {
		throw new UnsupportedOperationException();
	}

	@Override
	public byte[] outputModelFile() {
		Map<List<Boolean>, WekaLearnedModel> serializeMap = new HashMap<>();
		for (Entry<List<Boolean>, WekaLearnedModel> entry : learnedModels.entrySet()) {
			serializeMap.put(new ArrayList<>(entry.getKey()), entry.getValue());
		}

		ByteArrayOutputStream stream = new ByteArrayOutputStream();

		try {

			SerializationHelper.write(stream, serializeMap);

			return stream.toByteArray();

		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			throw new IllegalStateException(e);
		}

	}

	@Override
	public double runAndEvaluate(Collection<RecordData> data, EvaluationMethod evaluator) {
		throw new UnsupportedOperationException();
	}

	@Override
	public Optional<Double> getNormalizingThreshold() {
		return Optional.empty();
	}

	public Optional<Double> getPositiveConfidenceThreshold() {
		if(weights == null) {
			return Optional.empty();
		} else {
			return weights.get(optimizableClass.getWeight("POSITIVE_CONFIDENCE_THRESHOLD", 0.0));
		}
	}

	public Optional<Double> getNegativeConfidenceThreshold() {
		if (weights == null) {
			return Optional.empty();
		} else {
			return weights.get(optimizableClass.getWeight("NEGATIVE_CONFIDENCE_THRESHOLD", 0.0));
		}
	}

	public Optional<Double> getClusteringThreshold() {
		// return Optional.empty();
		if (weights == null || true) {
			return Optional.empty();
		} else {
			return weights.get(optimizableClass.getWeight("CLUSTERING_THRESHOLD", 0.0));
		}
	}

	public Optional<Double> getVeryLowPairThreshold() {
		// return Optional.empty();
		if (weights == null || true) {
			return Optional.empty();
		} else {
			return weights.get(optimizableClass.getWeight("VERY_LOW_PAIR_THRESHOLD", 0.0));
		}
	}

	@Override
	public String toString() {
		return "MT_WekaLearnedModel [learnedModels=" + learnedModels + "]";
	}

	public double getError() {
		return learnedModels.values().iterator().next().getError();
	}

}
