package expansion.set.optimizing;

import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;

import javax.inject.Inject;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.filefilter.DirectoryFileFilter;
import org.apache.commons.io.filefilter.FileFileFilter;
import org.apache.commons.lang.RandomStringUtils;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId.ClassIdStyle;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.OutputableModel;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.GenericLearnedModelRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;

public class LearnedModelAdapter {
	@Inject
	public LearnedModelAdapter() {
		// TODO Auto-generated constructor stub
	}

	public <F extends OutputableModel, E extends GenericLearnedModelRepository<F, E>> void storeModels(String weightsFolder, Map<ClassId, E> repository)
			throws IOException {

		String currentCase = new SimpleDateFormat("yyyyMMddHHmmss").format(new Date()) + "-" + RandomStringUtils.randomAlphanumeric(4);

		MKDIR.createDirUnchecked(new File(weightsFolder + "/" + currentCase + "/"));

		for (Entry<ClassId, E> entry : repository.entrySet()) {
			MKDIR.createDirUnchecked(new File(weightsFolder + "/" + currentCase + "/" + entry.getKey().getFinalPart() + "/"));

			for (Entry<OptimizableClass, F> inner : entry.getValue().getMap().entrySet()) {

				MKDIR.createDirUnchecked(
						new File(weightsFolder + "/" + currentCase + "/" + entry.getKey().getFinalPart() + "/" + inner.getValue().getClass().getName() + "/"));

				FileUtils.writeByteArrayToFile(new File(weightsFolder + "/" + currentCase + "/" + entry.getKey().getFinalPart() + "/"
						+ inner.getValue().getClass().getName() + "/" + inner.getKey().getName()), inner.getValue().outputModelFile());

			}

		}

	}

	private Optional<String> getLastFolder(String weightsFolder) {
		File directory = new File(weightsFolder);
		List<File> files = Arrays.asList(directory.listFiles((FileFilter) DirectoryFileFilter.DIRECTORY));

		java.util.Optional<String> present = files.stream().map((e) -> e.getName()).sorted((e1, e2) -> e2.compareTo(e1)).findFirst();
		return present;
	}

	public <F extends OutputableModel, E extends GenericLearnedModelRepository<F, E>> Map<ClassId, E> readModels(Function<Map<OptimizableClass, F>, E> creator,
			Set<ClassId> classIds, String weightsFolder) {
		try {
			Optional<String> lastFile = getLastFolder(weightsFolder);

			Map<ClassId, E> ret = new HashMap<>();

		if (lastFile.isPresent()) {
				File caseDirectory = new File(weightsFolder + "/" + lastFile.get());
				List<File> classes = Arrays.asList(caseDirectory.listFiles((FileFilter) DirectoryFileFilter.DIRECTORY));

				for (File classDirecotry : classes) {
					ClassId classId = ClassId.fromString(classDirecotry.getName(), ClassIdStyle.DBPEDIA_URI);

					if (classIds.contains(classId)) {
						Map<OptimizableClass, F> modelMap = new HashMap<>();
						List<File> methods = Arrays.asList(classDirecotry.listFiles((FileFilter) DirectoryFileFilter.DIRECTORY));

						for (File methodDirectory : methods) {
							List<File> models = Arrays.asList(methodDirectory.listFiles((FileFilter) FileFileFilter.FILE));

							for (File model : models) {
								Class<? extends F> cl = (Class<? extends F>) Class.forName(methodDirectory.getName());
								Constructor<? extends F> cons = cl.getConstructor(byte[].class);
								byte[] string = FileUtils.readFileToByteArray(model);
								;
								modelMap.put(new OptimizableClass(model.getName()), cons.newInstance(string));
							}
						}

						ret.put(classId, creator.apply(modelMap));
					} else {
						int fsdf = 123;
					}

				}



		} else {
			ret = new HashMap<>();
			for (ClassId classId : classIds) {
					ret.put(classId, creator.apply(new HashMap<>()));
			}
		}
		return ret;
		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
	}
}
