package expansion.set.optimizing;

import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.TreeMap;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.filefilter.FileFileFilter;
import org.apache.commons.lang.RandomStringUtils;

import au.com.bytecode.opencsv.CSVReader;
import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId.ClassIdStyle;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeightKey;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepositoryImpl;

public class WeightsAdapter {
	@Inject
	public WeightsAdapter() {

	}

	public String outputWeights(Map<ClassId, WeightsRepositoryImpl> repository) {
		try {
		StringWriter sw = new StringWriter();


		CSVWriter csvWriter = new CSVWriter(sw);

		int columns = 4;
		{
			int i = 0;
			String[] headerLine = new String[columns];
			headerLine[i++] = "ClassId";
			headerLine[i++] = "WeightSet";
			headerLine[i++] = "Weight";
			headerLine[i++] = "Value";

			csvWriter.writeNext(headerLine);
		}

		for (Entry<ClassId, WeightsRepositoryImpl> entry : repository.entrySet()) {

			TreeMap<OptimizableWeightKey, Double> map = new TreeMap<>(entry.getValue().getWeights());

			for (Entry<OptimizableWeightKey, Double> innerEntry : map.entrySet()) {
				{
					int i = 0;
					String[] line = new String[columns];
					line[i++] = entry.getKey().getFinalPart();
					line[i++] = innerEntry.getKey().getOptimizableClazz().getName();
					line[i++] = innerEntry.getKey().getName();
					line[i++] = String.valueOf(innerEntry.getValue());

					csvWriter.writeNext(line);
				}
			}

		}

		csvWriter.close();

			return sw.toString();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	public void storeWeights(String weightsFolder, Map<ClassId, WeightsRepositoryImpl> repository) throws IOException {

		String currentCase = new SimpleDateFormat("yyyyMMddHHmmss").format(new Date()) + "-" + RandomStringUtils.randomAlphanumeric(4);

		String output = outputWeights(repository);

		FileUtils.writeStringToFile(new File(weightsFolder + "/" + currentCase + ".csv"), output, "UTF-8");
	}


	private Optional<String> getLastFile(String weightsFolder) {
		File directory = new File(weightsFolder);
		List<File> files = Arrays.asList(directory.listFiles((FileFilter) FileFileFilter.FILE));

		java.util.Optional<String> present = files.stream().map((e) -> e.getName()).sorted((e1, e2) -> e2.compareTo(e1)).findFirst();
		return present;
	}

	public Map<ClassId, WeightsRepositoryImpl> parseOutput(String output) {
		try {
			Map<ClassId, Map<OptimizableWeightKey, Double>> weights = new HashMap<>();

			// for (ClassId classId : classIds) {
			// weights.put(classId, new HashMap<>());
			// }

			CSVReader reader = new CSVReader(new StringReader(output));

			List<String[]> lines = reader.readAll();

			reader.close();

			for (int i = 1; i < lines.size(); i++) {
				String[] line = lines.get(i);
				ClassId classId = ClassId.fromString(line[0], ClassIdStyle.DBPEDIA_URI);
				CHMHelper.getFromMap(weights, classId, () -> new HashMap<>()).put(new OptimizableWeightKey(new OptimizableClass(line[1]), line[2]),
						Double.parseDouble(line[3]));
			}

			return weights.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> new WeightsRepositoryImpl(e.getValue())));
		} catch (IOException e) {
			throw new RuntimeException(e);
		}

	}

	public Map<ClassId, WeightsRepositoryImpl> readWeights(Set<ClassId> classIds, String weightsFolder) throws IOException {
		Optional<String> lastFile = getLastFile(weightsFolder);

		Map<ClassId, WeightsRepositoryImpl> ret;

		if (lastFile.isPresent()) {

			String output = FileUtils.readFileToString(new File(weightsFolder + "/" + lastFile.get()), "UTF-8");

			ret = parseOutput(output);

		} else {
			ret = new HashMap<>();
			for (ClassId classId : classIds) {
				ret.put(classId, new WeightsRepositoryImpl());
			}
		}
		return ret;
	}
}
