package expansion.set.profiling;

import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.document.StoredField;
import org.apache.lucene.document.TextField;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.queryparser.classic.ParseException;
import org.apache.lucene.queryparser.classic.QueryParser;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.TopDocs;

import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.index.io.InMemoryIndex;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;
import expansion.set.strategies.rowclusterers.label.LabelRowIndexClusterMapCreator;

public class GenericLabelRowIndex { 
	
	

		public static final String STRING_AS_ID = "stringAsId";
		public static final String LABEL = "label";
//		private static final String SEPERATOR_BIG = "::B::";
//
//		private static final String SEPERATOR_SMALL = "::S::";

		private final InMemoryIndex idx = new InMemoryIndex();

		private final ThreadLocal<QueryParser> queryParserLocal = new ThreadLocal<>();

		private QueryParser getQueryParser() {
			QueryParser qp = queryParserLocal.get();
			if (qp == null) {
				qp = new QueryParser(  LABEL, new StandardAnalyzer( ));
				queryParserLocal.set(qp);
			} 
			return qp;
		}

		private final   ConcurrentHashMap<String, Set<IndexId>> index = new ConcurrentHashMap<>();
		private final int maxEditDistance;
		
		public GenericLabelRowIndex(Map<String, Set<IndexId>> preIndex) throws IOException {
			this(preIndex, 2);
		}
		
		
		public GenericLabelRowIndex(Map<String, Set<IndexId>> preIndex, int maxEditDistance) throws IOException {

			this.maxEditDistance = maxEditDistance;
			IndexWriter w = idx.getIndexWriter();

			index.clear();

			PersistentParallel6.runAndFinish(preIndex.entrySet(), (entry) -> {
				String label = entry.getKey();
//				if (labelSO != null) {
//					String label = labelSO.getString();

					String labelClean = StringNormalizer.normaliseValue(label, false);
					// normalises the value to improve lookup results
					label = LabelRowIndexClusterMapCreator.normalise(labelClean, true, true);
					CHMHelper.getFromCHM(index, label, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).addAll(entry.getValue());

//				}

			} , this.getClass());

			for (String indexEntry : index.keySet()) {
				Document doc = createDocument(indexEntry);

				w.addDocument(doc);
			}

			// for (TableRowId tableRowId : candidates) {
			//
			// StringObject labelSO = getLabel(tableRepository, tableRowId);
			// if (labelSO != null) {
			// String label = labelSO.getString();
			//
			// String labelClean = StringNormalizer.normaliseValue(label,
			// false);
			// // normalises the value to improve lookup results
			// label = LabelSimilarityRowClusterer.normalise(labelClean, true);
			//
			// Document doc = createDocument(label, tableRowId);
			//
			// w.addDocument(doc);
			// }
			// }

			idx.closeIndexWriter();
			
			idx.getIndexSearcher();
		}

		private SimpleLockableConcurrentCache<String, Map<String, Set<IndexId>>> cache = new SimpleLockableConcurrentCache<>();
		
		
		public Map<String, Set<IndexId>> getMatches(String label) throws IOException, ParseException {
			try {
	
				String labelClean = StringNormalizer.normaliseValue(label, false);
				// normalises the value to improve lookup results
				label = LabelRowIndexClusterMapCreator.normalise(labelClean, true, true);
				label = QueryParser.escape(label);
				
				
	                
	                
	                
				if (label.length() > 0) {
					
					 List<String> tokens = StringNormalizer.tokenise(label, false);
		                
		                StringBuilder sb = new StringBuilder();
		                boolean notFirst = false;
		                for(String token : tokens) {
		                	if(notFirst)
//		                		 sb.append(" AND ");
		                		sb.append(" ");
		                	
//		                	sb.append("\"");
		                    sb.append(token);
		                    
		                    if(maxEditDistance>0) {
		                    	if(token.length() < 3) {
		                    		//sb.append("~");
				                    //sb.append(maxEditDistance);
		                    	}
		                    	else if(token.length() < 5){
		                    		sb.append("~");
				                    sb.append(1);
		                    	}
		                    	else  {
		                    		 sb.append("~");
				                     sb.append(maxEditDistance);
		                    	}
		                       
		                    }
//		                    sb.append("\"");
		                   
		                    
		                    notFirst = true;
		                }
		                
					String finalLabel = sb.toString();
					return cache.getOrCreate(finalLabel, () -> {
						try {
							Query q = getQueryParser().parse(finalLabel);
							IndexSearcher indexSearcher = this.idx.getIndexSearcher();
							// new TermQuery(new Term(LABEL, search));
							TopDocs results = indexSearcher.search(q, 30);
							Map<String, Set<IndexId>> ret = new HashMap<>();
	
							for (ScoreDoc doc : results.scoreDocs) {
								String key = stringFromDocument(indexSearcher.doc(doc.doc));
	
								Set<IndexId> rows = this.index.get(key);
								if (rows == null) {
									throw new IllegalStateException();
								}
	
								ret.put(key, rows);
								// Set<IndexId> tableRowId =
								// fromDocument(indexSearcher.doc(doc.doc));
								// if (tableRowId != null)
								// instanceIds.addAll(tableRowId);
							}
							return ret;
						} catch (Exception e) {
							e.printStackTrace();
							System.out.println(e.getMessage());
							System.out.println("CAUGHT EXCEPTION");
							return new HashMap<>();
	
						}
	
					});
	
				}
	
			} catch (Exception e) {
				e.printStackTrace();
				System.out.println(e.getMessage());
				System.out.println("CAUGHT EXCEPTION");
			}
	
			return new HashMap<>();
	
		}

		public Set<IndexId> fromDocument(Document doc) {
			// Set<TableRowId> rowId = new HashSet<>();
			String s = doc.getField(STRING_AS_ID).stringValue();

			Set<IndexId> rows = this.index.get(s);
			if (rows == null) {
				throw new IllegalStateException();
			}

			return rows;
			//
			// String[] bigSplit = s.split(SEPERATOR_BIG);
			// for (String inner : bigSplit) {
			// String[] split = inner.split(SEPERATOR_SMALL);
			// if (split.length == 2) {
			// rowId.add(new TableRowId(new AnnotatedTableId(split[0]),
			// Integer.valueOf(split[1])));
			//
			// } else {
			// throw new IllegalStateException();
			// }
			// }

			// return rowId;

		}
		
		public String stringFromDocument(Document doc) {
			// Set<TableRowId> rowId = new HashSet<>();
			String s = doc.getField(STRING_AS_ID).stringValue();

//			Set<IndexId> rows = this.index.get(s);
//			if (rows == null) {
//				throw new IllegalStateException();
//			}

			return s;
			//
			// String[] bigSplit = s.split(SEPERATOR_BIG);
			// for (String inner : bigSplit) {
			// String[] split = inner.split(SEPERATOR_SMALL);
			// if (split.length == 2) {
			// rowId.add(new TableRowId(new AnnotatedTableId(split[0]),
			// Integer.valueOf(split[1])));
			//
			// } else {
			// throw new IllegalStateException();
			// }
			// }

			// return rowId;

		}

		public Document createDocument(String keyAndLabel) {
			Document doc = new Document();
			// String tableRowIdString = null;
			// for (TableRowId tableRowId : rowIds) {
			// if (tableRowIdString == null) {
			// tableRowIdString = "";
			// } else {
			// tableRowIdString += SEPERATOR_BIG;
			// }
			//
			// tableRowIdString += tableRowId.getTableId().getTableName() +
			// SEPERATOR_SMALL + tableRowId.getRowIdx().getRowIdx();
			// }

			// String tableRowIdString = tableRowId.getTableId().getTableName()
			// + SEPERATOR_BIG + tableRowId.getRowIdx().getRowIdx();
			doc.add(new StoredField(STRING_AS_ID, keyAndLabel));
			doc.add(new TextField(LABEL, keyAndLabel, Field.Store.YES));

			return doc;
		} 
		
		
		public static class IndexId {
			private final String indexId;

			public IndexId(String indexId) {
				super();
				this.indexId = indexId;
			}

			public String getIndexId() {
				return indexId;
			}

			@Override
			public int hashCode() {
				final int prime = 31;
				int result = 1;
				result = prime * result + ((indexId == null) ? 0 : indexId.hashCode());
				return result;
			}

			@Override
			public boolean equals(Object obj) {
				if (this == obj)
					return true;
				if (obj == null)
					return false;
				if (getClass() != obj.getClass())
					return false;
				IndexId other = (IndexId) obj;
				if (indexId == null) {
					if (other.indexId != null)
						return false;
				} else if (!indexId.equals(other.indexId))
					return false;
				return true;
			}
			
			
		}
}
