package expansion.set.profiling;

import java.io.IOException;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import org.apache.commons.lang.mutable.MutableFloat;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.document.StoredField;
import org.apache.lucene.document.TextField;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.queryparser.classic.ParseException;
import org.apache.lucene.queryparser.classic.QueryParser;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.TopDocs;

import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.index.io.InMemoryIndex;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;
import expansion.set.strategies.rowclusterers.label.LabelRowIndexClusterMapCreator;

public class GenericLabelRowIndexGeneric<E> { 
	
	

		public static final String STRING_AS_ID = "stringAsId";
		public static final String LABEL = "label";
		private final boolean useStemmer = false;
//		private static final String SEPERATOR_BIG = "::B::";
//
//		private static final String SEPERATOR_SMALL = "::S::";

		private final InMemoryIndex idx = new InMemoryIndex();
	private final int searchNum;

		private final ThreadLocal<QueryParser> queryParserLocal = new ThreadLocal<>();

		private QueryParser getQueryParser() {
			QueryParser qp = queryParserLocal.get();
			if (qp == null) {
				qp = new QueryParser(  LABEL, new StandardAnalyzer( ));
				queryParserLocal.set(qp);
			} 
			return qp;
		}

		private final   ConcurrentHashMap<String, Map<E,MutableFloat>> index = new ConcurrentHashMap<>();
		private final int maxEditDistance;
		
		public GenericLabelRowIndexGeneric(Map<String, Map<E,Number>> preIndex) throws IOException {
			this(preIndex, 2);
		}
		
		public void remove(String label, E key) {
			String labelClean = StringNormalizer.normaliseValue(label, false);
			// normalises the value to improve lookup results
			label  = LabelRowIndexClusterMapCreator.normalise(labelClean, true, useStemmer);
			ConcurrentHashMap<E, MutableFloat> innerMap = (ConcurrentHashMap<E, MutableFloat>) CHMHelper.getFromCHM(index, label, () -> new ConcurrentHashMap<>());
			innerMap.remove(key);
 

			  cache.remove(label);
			
		}
		
		public void add(String label, E key, MutableFloat score) throws IOException {
			String labelClean = StringNormalizer.normaliseValue(label, false);
			// normalises the value to improve lookup results
			label  = LabelRowIndexClusterMapCreator.normalise(labelClean, true, useStemmer);
			
			if(!index.containsKey(label)) {
				IndexWriter w = idx.getIndexWriter();
				
				Document doc = createDocument(label);

				w.addDocument(doc);
				
				idx.closeIndexWriter();
				
				 idx.reopenIndexSearcher();
			}
			
			
			ConcurrentHashMap<E, MutableFloat> innerMap = (ConcurrentHashMap<E, MutableFloat>) CHMHelper.getFromCHM(index, label, () -> new ConcurrentHashMap<>());
	 
 
			MutableFloat mf = CHMHelper.getFromCHM(innerMap, key, () -> new MutableFloat(0.0));
			  synchronized(mf) {
				  if(score.floatValue() > mf.floatValue()) {
					  mf.setValue(score.floatValue());
				  }
			  }
			  

			  cache.remove(label);
		}
		public GenericLabelRowIndexGeneric(Map<String, Map<E,Number>> preIndex, int maxEditDistance) throws IOException {
		this(preIndex, maxEditDistance, 30, true);
	}

	public GenericLabelRowIndexGeneric(Map<String, Map<E, Number>> preIndex, int maxEditDistance, int searchNum, boolean withCache) throws IOException {
		this.searchNum = searchNum;
		this.withCache = withCache;
			this.maxEditDistance = maxEditDistance;
			IndexWriter w = idx.getIndexWriter();

			index.clear();

			PersistentParallel6.runAndFinish(preIndex.entrySet(), (entry) -> {
				String label = entry.getKey();
//				if (labelSO != null) {
//					String label = labelSO.getString();

					String labelClean = StringNormalizer.normaliseValue(label, false);
					// normalises the value to improve lookup results
					label = LabelRowIndexClusterMapCreator.normalise(labelClean, true, useStemmer);
					  ConcurrentHashMap<E, MutableFloat> innerMap = (ConcurrentHashMap<E, MutableFloat>) CHMHelper.getFromCHM(index, label, () -> new ConcurrentHashMap<>());
					  
					  for(Entry<E,Number> innerEntry : entry.getValue().entrySet()) {
						  MutableFloat mf = CHMHelper.getFromCHM(innerMap, innerEntry.getKey(), () -> new MutableFloat(0.0));
						  synchronized(mf) {
							  if(innerEntry.getValue().floatValue() > mf.floatValue()) {
								  mf.setValue(innerEntry.getValue().floatValue());
							  }
						  }
						  
					  }

//				}

			} , this.getClass());

			for (String indexEntry : index.keySet()) {
				Document doc = createDocument(indexEntry);

				w.addDocument(doc);
			}

			// for (TableRowId tableRowId : candidates) {
			//
			// StringObject labelSO = getLabel(tableRepository, tableRowId);
			// if (labelSO != null) {
			// String label = labelSO.getString();
			//
			// String labelClean = StringNormalizer.normaliseValue(label,
			// false);
			// // normalises the value to improve lookup results
			// label = LabelSimilarityRowClusterer.normalise(labelClean, true);
			//
			// Document doc = createDocument(label, tableRowId);
			//
			// w.addDocument(doc);
			// }
			// }

			idx.closeIndexWriter();
			
			idx.getIndexSearcher();
		}

		private SimpleLockableConcurrentCache<String, Map<E, Float>> cache = new SimpleLockableConcurrentCache<>();
	private final boolean withCache;
		
		
		public Map<E, Float> getMatches(String label) throws IOException, ParseException {
			try {
				String labelClean = StringNormalizer.normaliseValue(label, false);
				// normalises the value to improve lookup results
				label = LabelRowIndexClusterMapCreator.normalise(labelClean, true, useStemmer);
				label = QueryParser.escape(label);
				

	                
	                
	                
				if (label.length() > 0) {
					
					 List<String> tokens = StringNormalizer.tokenise(label, false);
		                
		                StringBuilder sb = new StringBuilder();
		                
		                boolean notFirst = false;
		                for(String token : tokens) {
		                	if(notFirst)
//		                		 sb.append(" AND ");
		                		sb.append(" ");
		                	
//		                	sb.append("\"");
		                    sb.append(token);
		                    
		                    if(maxEditDistance>0) {
		                    	if(token.length() < 3) {
		                    		//sb.append("~");
				                    //sb.append(maxEditDistance);
		                    	}
		                    	else if(token.length() < 5){
		                    		sb.append("~");
				                    sb.append(1);
		                    	}
		                    	else  {
		                    		 sb.append("~");
				                     sb.append(maxEditDistance);
		                    	}
		                       
		                    }
//		                    sb.append("\"");
		                   
		                    
		                    notFirst = true;
		                }
		                
					String finalLabel = sb.toString();

				Supplier<Map<E, Float>> supplier = () -> {
						try {
							Query q = getQueryParser().parse(finalLabel);
							IndexSearcher indexSearcher = this.idx.getIndexSearcher();
							// new TermQuery(new Term(LABEL, search));
						TopDocs results = indexSearcher.search(q, searchNum);
							Map<E, MutableFloat> ret = new HashMap<>();
	
							for (ScoreDoc doc : results.scoreDocs) {
								String key = stringFromDocument(indexSearcher.doc(doc.doc));
								float score = doc.score;
								Map<E, MutableFloat> rows = this.index.get(key);
								for(Entry<E, MutableFloat> row : rows.entrySet()) {
									MutableFloat md = CHMHelper.getFromMap(ret, row.getKey(),() -> new MutableFloat(0.0F));
									if(  score > md.floatValue() ){
										md.setValue(  score);;
									}
								}
								if (rows == null) {
									throw new IllegalStateException();
								}
	
								 
								// Set<IndexId> tableRowId =
								// fromDocument(indexSearcher.doc(doc.doc));
								// if (tableRowId != null)
								// instanceIds.addAll(tableRowId);
							}

						// StringBuilder sbOutput = new StringBuilder();
						// sbOutput.append(finalLabel + ": ");
						// for (ScoreDoc sd : results.scoreDocs) {
						// sbOutput.append(stringFromDocument(indexSearcher.doc(sd.doc))
						// + "(" + sd.score + ") ");
						// }
						// System.out.println(sbOutput.toString());
							return ret.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().floatValue()));
						} catch (Exception e) {
							e.printStackTrace();
							System.out.println(e.getMessage());
							System.out.println("CAUGHT EXCEPTION");
							return new HashMap<>();
	
						}
	
				};
				if (withCache) {
					return cache.getOrCreate(finalLabel, supplier);
				} else {
					return supplier.get();
				}
	
				}
	
			} catch (Exception e) {
				e.printStackTrace();
				System.out.println(e.getMessage());
				System.out.println("CAUGHT EXCEPTION");
			}
	
			return new HashMap<>();
	
		}

		public Collection<E> fromDocument(Document doc) {
			// Set<TableRowId> rowId = new HashSet<>();
			String s = doc.getField(STRING_AS_ID).stringValue();

			Map<E, MutableFloat> rows = this.index.get(s);
			if (rows == null) {
				throw new IllegalStateException();
			}

			return rows.keySet();
			//
			// String[] bigSplit = s.split(SEPERATOR_BIG);
			// for (String inner : bigSplit) {
			// String[] split = inner.split(SEPERATOR_SMALL);
			// if (split.length == 2) {
			// rowId.add(new TableRowId(new AnnotatedTableId(split[0]),
			// Integer.valueOf(split[1])));
			//
			// } else {
			// throw new IllegalStateException();
			// }
			// }

			// return rowId;

		}
		
		public String stringFromDocument(Document doc) {
			// Set<TableRowId> rowId = new HashSet<>();
			String s = doc.getField(STRING_AS_ID).stringValue();

//			Set<IndexId> rows = this.index.get(s);
//			if (rows == null) {
//				throw new IllegalStateException();
//			}

			return s;
			//
			// String[] bigSplit = s.split(SEPERATOR_BIG);
			// for (String inner : bigSplit) {
			// String[] split = inner.split(SEPERATOR_SMALL);
			// if (split.length == 2) {
			// rowId.add(new TableRowId(new AnnotatedTableId(split[0]),
			// Integer.valueOf(split[1])));
			//
			// } else {
			// throw new IllegalStateException();
			// }
			// }

			// return rowId;

		}

		public Document createDocument(String keyAndLabel) {
			Document doc = new Document();
			// String tableRowIdString = null;
			// for (TableRowId tableRowId : rowIds) {
			// if (tableRowIdString == null) {
			// tableRowIdString = "";
			// } else {
			// tableRowIdString += SEPERATOR_BIG;
			// }
			//
			// tableRowIdString += tableRowId.getTableId().getTableName() +
			// SEPERATOR_SMALL + tableRowId.getRowIdx().getRowIdx();
			// }

			// String tableRowIdString = tableRowId.getTableId().getTableName()
			// + SEPERATOR_BIG + tableRowId.getRowIdx().getRowIdx();
			doc.add(new StoredField(STRING_AS_ID, keyAndLabel));
			doc.add(new TextField(LABEL, keyAndLabel, Field.Store.YES));

			return doc;
		} 
		
		
		 
}
