package expansion.set.profiling;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Serializable;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import javax.inject.Inject;

import com.opencsv.CSVWriter;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.DecimalFormatter;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;
import expansion.set.profiling.GenericLabelRowIndex.IndexId;

public class PossibleConflictProfiler {
	private final KnowledgeBase knowledgeBase;
	private final SimilarityCalculator similarityCalculator;
	private final AnnotatedTablesRepository tableRepository;
	private final OntologyRepository ontologyRepositoryl;
	@Inject
	PossibleConflictProfiler(KnowledgeBase knowledgeBase,SimilarityCalculator similarityCalculator, AnnotatedTablesRepository tableRepository, OntologyRepository ontologyRepository) {
		this.knowledgeBase = knowledgeBase;
		this.similarityCalculator = similarityCalculator;
		this.tableRepository = tableRepository;
		this.ontologyRepositoryl = ontologyRepository;
	}

	public static final ClassId placeClass = new ClassId("http://dbpedia.org/ontology/Settlement");
	public static final ClassId officeHolderClass = new ClassId("http://dbpedia.org/ontology/OfficeHolder");
	
	
	
	public Map<ClassId, ConflictData> run(String resultsFolder) throws IOException {
		
		Map<ClassId, ConflictData> conflictDataMap = new HashMap<>();
		
		Set<ClassId> classes = new HashSet<>();

//		classes.add(placeClass);
		
		
//		classes.add(WikidataClasses.wikidataCountry);
//		classes.add(WikidataClasses.wikidataFootball);
//		classes.add(WikidataClasses.wikidataSoccer);
//		classes.add(WikidataClasses.wikidataBasketball);
//		classes.add(WikidataClasses.wikidataCity);
//		classes.add(WikidataClasses.wikidataOfficeHolder);
//		classes.add(WikidataClasses.wikidataTradedCompany);
			classes = knowledgeBase.getClasses();
			int total = this.knowledgeBase.getClasses().size();
			int i = 0;
		for(ClassId setId : classes) {
			
			
			ConflictData conflictData = evaluateSetId(setId, true);
			
			conflictDataMap.put(setId, conflictData);
			i++;
			
			System.out.println("Finished " + i + " out of " + total + ".");
			
			
		}
		
		

//		ConflictData conflictDataAll = evaluateSetId(this.knowledgeBase.getClasses());
//		
//		conflictDataMap.put(WikidataClasses.wikidataAll, conflictDataAll);
		
		if(resultsFolder != null) {
			MKDIR.createDirUnchecked(new File(resultsFolder));
			 
			CSVWriter writer = new CSVWriter( new FileWriter(new File(resultsFolder + "/conflict.csv" )));
			
			int length = 7;
			{
				String[] line = new String[length];
				line[0] = "Class";
				line[1] = "Total";
				line[2] = "Conflict";
				line[3] = "Percent";
				line[4] = "MatchTotal";
				line[5] = "MatchConflict";
				line[6] = "MatchPercent";
				writer.writeNext(line);
			}
			
			for(Entry<ClassId, ConflictData> entry : conflictDataMap.entrySet()) {
				String[] line = new String[length];
				line[0] = entry.getKey().getClassId();
				line[1] = String.valueOf(entry.getValue().getTotal());
				line[2] = String.valueOf(entry.getValue().getConflict());
				line[3] = DecimalFormatter.format(entry.getValue().getPotential());
				line[4] = String.valueOf(entry.getValue().getTotalMatch());
				line[5] = String.valueOf(entry.getValue().getConflictMatch());
				line[6] = DecimalFormatter.format(entry.getValue().getPotentialMatch());
				writer.writeNext(line);
			}
			
			writer.close();
		}
		
		int ghg =12;
		
		return conflictDataMap;
		 
	}
	
	
	private ConflictData evaluateSetId(Set<ClassId> originalClasses, boolean removeAfterCommaAndRemoveBrackets) throws IOException {
		
		Set<InstanceId> instances = new HashSet<>();
		
		Set<ClassId> classes = new HashSet<>();
		
		for(ClassId parentSetId : classes) {
			
			classes.add(parentSetId);
			classes.addAll(this.ontologyRepositoryl.getSubClasses(parentSetId));

			 
		}
		
		for(ClassId setId : classes) {
			if(knowledgeBase.hasClass(setId)) {
				instances.addAll(knowledgeBase.getInstances(setId).keySet());
			}
		}
		
		return evaluateInstances(classes, instances, removeAfterCommaAndRemoveBrackets);
		 
	}

	private ConflictData evaluateSetId(ClassId parentSetId, boolean removeAfterCommaAndRemoveBrackets) throws IOException {
		
	
		
		
		Set<ClassId> classes = new HashSet<>(this.ontologyRepositoryl.getSubClasses(parentSetId));
		classes.add(parentSetId);
		Set<InstanceId> instances = new HashSet<>();
		for(ClassId setId : classes) {
			if(knowledgeBase.hasClass(setId)) {
				instances.addAll(knowledgeBase.getInstances(setId).keySet());
			}
		}
		
		 
		
		return evaluateInstances(classes, instances, removeAfterCommaAndRemoveBrackets);
	
	}
	
	
	private ConflictData evaluateInstances(Set<ClassId> classes, Set<InstanceId> instances, boolean removeAfterCommaAndRemoveBrackets ) throws IOException {
		
		AtomicInteger totalCount = new AtomicInteger(0);
		AtomicInteger conflictCount = new AtomicInteger(0);
		AtomicInteger totalMatchCount = new AtomicInteger(0);
		AtomicInteger conflictMatchCount = new AtomicInteger(0);

		ConcurrentHashMap<String, Set<IndexId>> preIndex = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(instances, (instance) -> {
			
			
			Set<String> labels = knowledgeBase.getLabels(instance);
			
			if(removeAfterCommaAndRemoveBrackets) {
				labels = removeAfterCommaAndRemoveBrackets(labels);
			}
			for(String string : labels) {
				CHMHelper.getFromCHM(preIndex, string, () -> new HashSet<>()).add(new IndexId(instance.toCombinedString()));
			}
		
		}, this.getClass());
		
		GenericLabelRowIndex index = new GenericLabelRowIndex(preIndex);
		
		PersistentParallel6.runAndFinish(instances, (instance) -> {
			try {
				boolean hasMatch = false;
				
				if(tableRepository.hasInstance(instance)) {
					for(AnnotatedTableId id : tableRepository.getTables(instance)) {
						if(classes.contains(tableRepository.getMetaData(id).getDbPediaClass())) {
							hasMatch = true;
							break;
						}
					}
				}
				boolean conflict = false;
				Set<String> labels = knowledgeBase.getLabels(instance);
				if(removeAfterCommaAndRemoveBrackets) {
					labels = removeAfterCommaAndRemoveBrackets(labels);
				}
				for(String string : labels) {
					
					Map<String, Set<IndexId>> matches = index.getMatches(string);
				 
					StringObject target = new StringObject(string);
					for(Entry<String, Set<IndexId>> matchEntry : matches.entrySet()) {
						 if(matchEntry.getValue().size() > 1 || !instance.equals(InstanceId.fromCombinedString(matchEntry.getValue().iterator().next().getIndexId()))) {
							if(similarityCalculator.getLabelBlockingSimilarity().isEqual(target, new StringObject(matchEntry.getKey()))) {
								conflict = true;
								break;
							}
						 }
					}
					
					if(conflict) {
						break;
					}
					
					
				}
				
				totalCount.incrementAndGet();
				if(hasMatch) {
					totalMatchCount.incrementAndGet();
				}
				if(conflict) {
					conflictCount.incrementAndGet();
					if(hasMatch) {
						conflictMatchCount.incrementAndGet();
					}
				}
				
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
		}, this.getClass());
	 
		return new ConflictData(totalCount.get(), conflictCount.get(), totalMatchCount.get(), conflictMatchCount.get());
	}


	private Set<String> removeAfterCommaAndRemoveBrackets(Set<String> labels) {
		
		
		return labels.stream().map(StringNormalizer::removeAfterCommaAndRemoveBrackets).collect(Collectors.toSet());
  
	}


	public static class ConflictData implements Serializable{
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;
		
		private final int total;
		private final int conflict;

		private final int totalMatch;
		private final int conflictMatch;
		
		
		public ConflictData(int total, int conflict,int totalMatch, int conflictMatch) {
			super();
			this.total = total;
			this.conflict = conflict;
			this.totalMatch = totalMatch;
			this.conflictMatch = conflictMatch;
		}



		public double getPotential() {
			return ((double) conflict) / total;
		}



		public int getTotal() {
			return total;
		}



		public int getConflict() {
			return conflict;
		}



		public int getTotalMatch() {
			return totalMatch;
		}



		public int getConflictMatch() {
			return conflictMatch;
		}
		
		public double getPotentialMatch() {
			return ((double) conflictMatch) / totalMatch;
		}



		@Override
		public String toString() {
			return "ConflictData [total=" + total + ", conflict=" + conflict + ", totalMatch=" + totalMatch + ", conflictMatch=" + conflictMatch
					+ ", getPotential()=" + getPotential() + ", getPotentialMatch()=" + getPotentialMatch() + "]";
		}



		 

	}
}
