package expansion.set.profiling;

import java.util.Collection;
import java.util.Set;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumInteger;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentAccumulator;

public class TableRowSetProfiler {
	
	@Inject
	AnnotatedTablesRepository tablesRepository;

	public TableRowSetProfiler() {
	}
	
	public int getNumberOfMappedValues(Collection<TableRowId> rowIds) {

		ConcurrentAccumulator<AccumInteger> totalMappedValues = new ConcurrentAccumulator<>(AccumInteger.empty());
		
		PersistentParallel6.runAndFinish(rowIds, (rowId) -> {
			AnnotatedTableMetaData md = tablesRepository.getMetaData(rowId.getTableId());
			totalMappedValues.add( new AccumInteger(md.getProperties().size()));
		}, this.getClass());
		
		return totalMappedValues.get().get();		
	}

	public int getNumberOfMappedValues(Set<TableRowId> rowIds, ClassId classId, PropertyId propertyId) {
	ConcurrentAccumulator<AccumInteger> totalMappedValues = new ConcurrentAccumulator<>(AccumInteger.empty());
		
		PersistentParallel6.runAndFinish(rowIds, (rowId) -> {
			AnnotatedTableMetaData md = tablesRepository.getMetaData(rowId.getTableId());
			if(md.getDbPediaClass().equals(classId)) {
				
				if(md.getProperties().values().contains(propertyId)) {
					totalMappedValues.add( new AccumInteger(1));
				}
				
				
			}
		}, this.getClass());
		
		return totalMappedValues.get().get();	
	}
}
