package expansion.set.strategies;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import expansion.set.clustering.RowClusterer;
import expansion.set.strategies.evaluatemethod.CLUSTERING_SIMS_AGGREGATOR;
import expansion.set.strategies.evaluatemethod.CLUSTERING_SIMS_LIST;
import expansion.set.strategies.evaluatemethod.CLUSTERING_STRATEGY_DEFINITION_WORKAROUND;

public class SetExpansionStrategy {
	private final String name;
	private final Class<? extends RowClusterer> clusterer;
	private final List<Set<String>> optimizingPriors;
	private final RowClusterer clustererImpl;

	public static final Map<Class<? extends RowClusterer>, CLUSTERING_STRATEGY_DEFINITION_WORKAROUND> strategyDefinitionWorkaroundMap = new ConcurrentHashMap<>();

	private final static AtomicInteger indexCounter = new AtomicInteger();

	private final static List<Class<? extends RowClusterer>> workAroundClassesKeys = new ArrayList<>();


	public SetExpansionStrategy(String name, CLUSTERING_SIMS_LIST simsList, CLUSTERING_SIMS_AGGREGATOR aggregator) {
		this.name = name;
		this.clustererImpl = null;
		this.optimizingPriors = null;
		CLUSTERING_STRATEGY_DEFINITION_WORKAROUND clustererInstance = new CLUSTERING_STRATEGY_DEFINITION_WORKAROUND() {

			@Override
			public OptimizableClass getOptimizableClass() {
				return new OptimizableClass("CLUSTERING::" + name);

			}

			@Override
			public CLUSTERING_SIMS_LIST getSimsList() {
				return simsList;
			}

			@Override
			public CLUSTERING_SIMS_AGGREGATOR getAggregator() {
				return aggregator;
			}
		};

		int index = SetExpansionStrategy.indexCounter.getAndIncrement();
		if (index >= SetExpansionStrategy.workAroundClassesKeys.size()) {
			throw new IllegalStateException("We exhausted all mock classes, you need to add more. sorry. ");
		}

		this.clusterer = SetExpansionStrategy.workAroundClassesKeys.get(index);

		SetExpansionStrategy.strategyDefinitionWorkaroundMap.put(this.clusterer, clustererInstance);
	}
	public SetExpansionStrategy(String name, RowClusterer clustererImpl, boolean flag) {
		super();
		this.clustererImpl = clustererImpl;
		this.name = name;
		this.clusterer = null;
		this.optimizingPriors = null;
	}

	public SetExpansionStrategy(String name, Class<? extends RowClusterer> clusterer) {
		super();
		this.name = name;
		this.clusterer = clusterer;
		this.optimizingPriors = null;
		this.clustererImpl = null;
	}

	public SetExpansionStrategy(String name, Class<? extends RowClusterer> clusterer, List<Set<String>> optimizingPriors) {
		super();
		this.name = name;
		this.clusterer = clusterer;

		if (optimizingPriors != null)
			this.optimizingPriors = optimizingPriors.stream().map((e) -> new HashSet<>(e)).collect(Collectors.toList());
		else
			this.optimizingPriors = null;

		this.clustererImpl = null;
	}

	public Class<? extends RowClusterer> getClusterer() {
		return clusterer;
	}

	public String getName() {
		return name;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((name == null) ? 0 : name.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		SetExpansionStrategy other = (SetExpansionStrategy) obj;
		if (name == null) {
			if (other.name != null)
				return false;
		} else if (!name.equals(other.name))
			return false;
		return true;
	}

	@Override
	public String toString() {
		return name;
	}

	public OptimizableClass getOptimizableClassCluster() {
		return new OptimizableClass("CLUSTERING::" + name);
	}

	public OptimizableClass getOptimizableClassSimilarityThreshold() {
		return new OptimizableClass("CLUSTERING::" + name);
	}

	public List<Set<String>> getOptimizingPriors() {
		return optimizingPriors;
	}

	public RowClusterer getClustererImpl() {
		return clustererImpl;
	}

	private static class WorkAroundKey extends CLUSTERING_STRATEGY_DEFINITION_WORKAROUND {

		@Override
		public CLUSTERING_SIMS_LIST getSimsList() {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public CLUSTERING_SIMS_AGGREGATOR getAggregator() {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public OptimizableClass getOptimizableClass() {
			// TODO Auto-generated method stub
			return null;
		}
	};

	static {

		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());
		workAroundClassesKeys.add(new WorkAroundKey() {
		}.getClass());

	}

}
