package expansion.set.strategies.blockers;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.strategies.rowclusterers.CreateClustersFromMappings;

public class BlockedSimilarities implements RowClusterSimilarities {

	
	private final Map<TableRowId, Integer> equalBlocks;
	private final Map<TableRowId, Integer> computeBlocks;
	private final RowClusterSimilarities innerSimilarity;
	
	
	public BlockedSimilarities(Mappings equalBlocks, Mappings computeBlocks, RowClusterSimilarities innerSimilarity) {
		
		
		Map<TableRowId, Integer> equalBlocksMap = CreateClustersFromMappings.createClusters(null, equalBlocks).getClustersIndex();
		Map<TableRowId, Integer> computeBlocksMap = CreateClustersFromMappings.createClusters(null, computeBlocks).getClustersIndex();
		
		
		float loadFactor = 0.35F;
		
		this.equalBlocks = new HashMap<>( 5 + (int) (equalBlocksMap.size() / loadFactor), loadFactor);
		this.equalBlocks.putAll(equalBlocksMap);
		
		this.computeBlocks = new HashMap<>( 5 + (int) (computeBlocksMap.size() / loadFactor), loadFactor);
		this.computeBlocks.putAll(computeBlocksMap);
		this.innerSimilarity = innerSimilarity;
		
	}
	
	
	@Override
	public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
		TableRowId row1 = rows.get(key.getE1());
		TableRowId row2 = rows.get(key.getE2());
		
		Integer equal1 = equalBlocks.get(row1);
		if(equal1 != null && equal1.equals(equalBlocks.get(row2))) {
			return 1.0;
		}
		else {
			Integer compute1 = computeBlocks.get(row1);
			if(compute1 != null && compute1.equals(computeBlocks.get(row2))) {
				return innerSimilarity.get(rows, classId, key);
			}
			else {
				return 0.0;
			} 
		}
		
	}
	@Override
	public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
		throw new UnsupportedOperationException();
	}

	@Override
	public String getName() {
		// TODO Auto-generated method stub
		return null;
	}
	
	
	
}
