package expansion.set.strategies.blockers;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.utils.ImmuMaker;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.similarities.RowClusterConstantSimilarities;
import expansion.set.strategies.rowclusterers.label.LabelRowIndexClusterMapCreator;

public class Blocks {
	private final Map<Integer, Set<TableRowId>> clustersMap;
	private final Map<Integer, ClassId> clusterClassId;
	public Blocks(Map<Integer, Set<TableRowId>> clustersMap, Map<Integer, ClassId> clusterClassId) {
		this.clusterClassId = ImmuMaker.map(clusterClassId);
		this.clustersMap = ImmuMaker.mapWithSet(clustersMap);
	}

	public ClusterList getClusterList(AnnotatedTablesRepository tableRepository) {
		
		Map<Integer, List<TableRowId>> clustersMapWithList = clustersMap.entrySet().parallelStream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			return new ArrayList<>(e.getValue());
		}));

	 

		ConcurrentLinkedQueue<RowCluster> retList = new ConcurrentLinkedQueue<>();
	 
		AtomicInteger removed = new AtomicInteger(0);
		PersistentParallel6.runAndFinish(clustersMapWithList.entrySet(), (entry) -> {
			if(entry.getValue().size() > 5000) {
				System.out.println("it exists " + entry.getValue().size());
				List<String> labels = entry.getValue().stream().map((e) -> {
					StringObject so = LabelRowIndexClusterMapCreator.getLabel(tableRepository, e);
					if (so != null) {
						return so.getString();
					} else {
						return "";
					}
					}).collect(Collectors.toList());
//				System.out.println(labels.toString());
				int sdf = 123;
			}
//			 System.out.println(entry.getValue().size());
//			 if (entry.getValue().size() < 5000)
				 retList.add(new RowCluster(clusterClassId.get(entry.getKey()), entry.getValue(), new RowClusterConstantSimilarities(1.0)));
//			 else
//				 removed.incrementAndGet();

		} , this.getClass(), "CreateClusters");

		System.out.println("Clusters removed: " + removed.get());

		ArrayList<RowCluster> retListList = new ArrayList<>(retList);

		return new ClusterList(retListList);
	}

	public Map<Integer, Set<TableRowId>> getClustersMap() {
		return clustersMap;
	}

	public Map<Integer, ClassId> getClusterClassId() {
		return clusterClassId;
	}

	
}
