package expansion.set.strategies.clustersplitters;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicInteger;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusterSplitter;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.clustering.model.ClusterSimilaritiesImpl;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.strategies.rowclusterers.ClusterSimilarities;
import expansion.set.strategies.rowclusterers.GreedyClustering;
import expansion.set.tablerow.TableRowCreator;

public class SimpleRowClusterSplitterFactory {
	private final GreedyClustering clusterer;
	private final TableRowCreator rowCreator;

	@Inject
	public SimpleRowClusterSplitterFactory(GreedyClustering clusterer,TableRowCreator rowCreator) {
		super();
		this.clusterer = clusterer;
		this.rowCreator = rowCreator;
	}

	public RowClusterSplitter getSplitter(RowClusterSimilarities scores) {
		return new SimpleSplitter(scores);
	}

	private class SimpleSplitter implements RowClusterSplitter {
		private final RowClusterSimilarities sim;

		private SimpleSplitter(RowClusterSimilarities sim) {
			super();
			this.sim = sim;
		}

		@Override
		public ClusterList clusterRows(RowClusteringConfig config, ClustererInput input, ClusterList rowCluster) {
			AtomicInteger skipped = new AtomicInteger();
			ConcurrentLinkedQueue<RowCluster> list = new ConcurrentLinkedQueue<>();
			PersistentParallel6.runAndFinish(rowCluster.getClusterList(), (oldRc) -> {
				List<expansion.set.tablerow.AnnotatedTableRow> rowList = new ArrayList<>();
				
				for(TableRowId rowId : oldRc.getRows()) {
					expansion.set.tablerow.AnnotatedTableRow row = rowCreator.getTableRow(rowId);
					rowList.add(row);
				}
				if(rowList.size() > 3000) {
					skipped.incrementAndGet();
					return;
				}
				ClusterSimilarities cs = new ClusterSimilaritiesImpl(oldRc, sim);

				list.addAll(clusterer.rowSplitter(oldRc, cs));
			} , this.getClass());

			System.out.println("skipped over very large clusters: " + skipped.get());
			return new ClusterList(list);

		}

	 

		@Override
		public List<RowCluster> cluster(ClassId classId, List<TableRowId> list) {
			return clusterer.rowSplitter(classId, list, sim);
		}
		
		

	}
}
