package expansion.set.strategies.clustersplitters.attributes;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class AttributeScoreGetter {
	@Inject
	protected SimilarityCalculator similarityCalculator;

	@Inject
	protected OntologyRepository ontologyRepository;

	@Inject
	public AttributeScoreGetter() {
	}

	public List<WeightedScore> getScores(Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>> firstValues,
			Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>> secondValues, boolean withSf, boolean withNr, ClassId classId) {
		return getScores(firstValues, secondValues, withSf, withNr, classId, null, null);
	}

	public static final boolean SCORING_MATCHING_SCORES = true;
	public List<WeightedScore> getScores(Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>> firstValues,
			Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>> secondValues, boolean withSf, boolean withNr, ClassId classId,
			Map<QualifiedPropertyId, Map<TableColumnId, Double>> scores1, Map<QualifiedPropertyId, Map<TableColumnId, Double>> scores2) {

		if (firstValues == null || secondValues == null) {
			return null;
		}

		List<WeightedScore> weightedScore = new ArrayList<>();

		// int total = 0;
		// double correct = 0;
		for (Entry<QualifiedPropertyId, Map<TableColumnId, AbstractObject>> firstEntry : firstValues.entrySet()) {

			ValueType vt = ontologyRepository.getDataType(firstEntry.getKey().getPropertyId());
			ClassPropertyId cpid = new ClassPropertyId(classId, firstEntry.getKey().getPropertyId());

			// Double maxSim = null;
			for (Entry<TableColumnId, AbstractObject> innerFirstEntry : firstEntry.getValue().entrySet()) {
				if (secondValues.containsKey(firstEntry.getKey())) {

					for (Entry<TableColumnId, AbstractObject> innerSecondValue : secondValues.get(firstEntry.getKey()).entrySet()) {

						double weight = 1.0;
						if (SCORING_MATCHING_SCORES && scores1 != null && scores2 != null) {
							double weight1 = scores1.get(firstEntry.getKey()).get(innerFirstEntry.getKey());
							double weight2 = scores2.get(firstEntry.getKey()).get(innerSecondValue.getKey());
							weight = weight1 * weight2;
						}
						double sim = 0.0;
						if (withSf && !withNr) {
							sim = similarityCalculator.getSimilarity(vt, innerFirstEntry.getValue(), innerSecondValue.getValue());

						} else if (!withSf && withNr) {
							sim = similarityCalculator.getSimilarity(cpid, innerFirstEntry.getValue(), innerSecondValue.getValue());
						} else if (withSf && withNr) {
							sim = similarityCalculator.getSimilarity(cpid, vt, innerFirstEntry.getValue(), innerSecondValue.getValue());

						} else {
							sim = similarityCalculator.getSimilarity(innerFirstEntry.getValue(), innerSecondValue.getValue());
						}

						if (vt.equals(ValueType.QUANTITY)) {
							double et = similarityCalculator.getEquivalenceThreshold(vt);
							if (sim > et) {
								double n2 = (sim - et) / (1.0 - et);
								weightedScore.add(new WeightedScore(n2, weight, firstEntry.getKey().getPropertyId().readableString(), true));

							} else {
								weightedScore.add(new WeightedScore(0, weight, firstEntry.getKey().getPropertyId().readableString(), false));
								// weightedScore.add(new
								// WeightedScore(-1.0,1.0,firstEntry.getKey().getPropertyId().readableString()));
							}
						} else if (vt.equals(ValueType.TIME)) {
							if (sim > similarityCalculator.getEquivalenceThreshold(vt)) {
								weightedScore.add(new WeightedScore(sim, sim * weight, firstEntry.getKey().getPropertyId().readableString(), true));
							} else {
								weightedScore.add(new WeightedScore(sim, weight, firstEntry.getKey().getPropertyId().readableString(), false));
							}

						} else {
							weightedScore.add(new WeightedScore(sim, weight, firstEntry.getKey().getPropertyId().readableString(),
									sim > similarityCalculator.getEquivalenceThreshold(vt)));
						}
					}

				}
			}

		}

		if (!weightedScore.isEmpty())
			return weightedScore;
		else
			return null;
	}

	public List<WeightedScore> getScores2(Map<PropertyId, List<Pair<AbstractObject, Double>>> firstValues,
			Map<PropertyId, List<Pair<AbstractObject, Double>>> secondValues, boolean withSf, boolean withNr, ClassId classId) {

		if (firstValues == null || secondValues == null) {
			return null;
		}

		Map<PropertyId, List<Pair<AbstractObject, Double>>> smallerMap;
		Map<PropertyId, List<Pair<AbstractObject, Double>>> largerMap;
		if (firstValues.size() < secondValues.size()) {
			smallerMap = firstValues;
			largerMap = secondValues;
		} else {
			smallerMap = secondValues;
			largerMap = firstValues;
		}

		List<WeightedScore> weightedScore = new ArrayList<>();

		// int total = 0;
		// double correct = 0;
		for (Entry<PropertyId, List<Pair<AbstractObject, Double>>> firstEntry : smallerMap.entrySet()) {

			ValueType vt = ontologyRepository.getDataType(firstEntry.getKey());
			ClassPropertyId cpid = new ClassPropertyId(classId, firstEntry.getKey());

			// Double maxSim = null;
			if (largerMap.containsKey(firstEntry.getKey())) {
				for (Pair<AbstractObject, Double> innerFirstEntry : firstEntry.getValue()) {

					for (Pair<AbstractObject, Double> innerSecondValue : largerMap.get(firstEntry.getKey())) {

						double sim = 0.0;
						if (withSf && !withNr) {
							sim = similarityCalculator.getSimilarity(vt, innerFirstEntry.getFirst(), innerSecondValue.getFirst());

						} else if (!withSf && withNr) {
							sim = similarityCalculator.getSimilarity(cpid, innerFirstEntry.getFirst(), innerSecondValue.getFirst());
						} else if (withSf && withNr) {
							sim = similarityCalculator.getSimilarity(cpid, vt, innerFirstEntry.getFirst(), innerSecondValue.getFirst());

						} else {
							sim = similarityCalculator.getSimilarity(innerFirstEntry.getFirst(), innerSecondValue.getFirst());
						}
						double weight = innerFirstEntry.getSecond() * innerSecondValue.getSecond();
						if (vt.equals(ValueType.QUANTITY)) {
							double et = similarityCalculator.getEquivalenceThreshold(vt);

							if (sim > et) {
								double n2 = (sim - et) / (1.0 - et);
								weightedScore.add(new WeightedScore(n2, weight, firstEntry.getKey().readableString(), true));

							} else {
								weightedScore.add(new WeightedScore(0, weight, firstEntry.getKey().readableString(), false));
								// weightedScore.add(new
								// WeightedScore(-1.0,1.0,firstEntry.getKey().getPropertyId().readableString()));
							}
						} else {
							weightedScore.add(new WeightedScore(sim, weight, firstEntry.getKey().readableString(),
									sim > similarityCalculator.getEquivalenceThreshold(vt)));
						}
					}

				}
			}

		}

		if (!weightedScore.isEmpty())
			return weightedScore;
		else
			return null;
	}

}
