package expansion.set.strategies.clustersplitters.attributes;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdsSet;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.Accuracy;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.FixedScore;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.similarity.Similarity.SimilarityType;
import de.mannheim.uni.expansion.tablefusion.core.splitting.NonTimeDependentSplitter;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.splitting.Splitter;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfigFilter;
import de.mannheim.uni.expansion.tablefusion.learning.RowValuesGetter;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;
import de.mannheim.uni.expansion.tablefusion.learning.SplitRowValues;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategy;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.ScorerRepository;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.SplitterProvider;
import de.mannheim.uni.expansion.tablefusion.services.grouping.GroupedValuesRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdClassPropertyId;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdContainer;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.evaluation.SplitValuesGetter;

public class AttributeValuesCreator {

	// private final TableRowCreator tableRowCreator;
	private final PropertiesConfigFilter propertiesConfigFilter;
	private final ScorerRepository scorerRepository;
	private final NonTimeDependentSplitter nonTimeDependentSplitter;

	@Inject
	protected AnnotatedTablesRepository tablesRepository;

	@Inject
	protected GroupedValuesRepository groupedValuesRepository;

	@Inject
	protected SplitValuesGetter splitValuesGetter;

	@Inject
	protected RowValuesGetter rowValuesGetter;

	// private final AnnotatedTablesRepository tableRepository;

	@Inject
	public AttributeValuesCreator(PropertiesConfigFilter propertiesConfigFilter, ScorerRepository scorerRepository,
			NonTimeDependentSplitter nonTimeDependentSplitter) {
		super();
		// this.tableRowCreator = tableRowCreator;
		this.propertiesConfigFilter = propertiesConfigFilter;
		this.scorerRepository = scorerRepository;
		this.nonTimeDependentSplitter = nonTimeDependentSplitter;
		// this.tableRepository = tableRepository;
	}

	private static SimpleLockableConcurrentCache<UuidStringId, AttributeValueContainer> cache = new SimpleLockableConcurrentCache<>();

	public AttributeValueContainer get(RowClusteringConfig config, ClustererInput input) {

		TableIdsSet tableIdsSet = new TableIdsSet(
				input.getFullCandidatesAllClasses().getRowIds().stream().map((e) -> e.getTableId()).collect(Collectors.toSet()),
				input.getFullCandidatesAllClasses().getId());
		
		GroupedValues groupedValues = groupedValuesRepository.getGroupedValues(tablesRepository, tableIdsSet.getTableIds(), config.getFusionConfig());
		SplitGroupedValues sgv = splitValuesGetter.getSplitGroupedValues(groupedValues, config.getFusionConfig());
		SplitGroupedValuesSets splitGroupedValuesSets = new SplitGroupedValuesSets(sgv, sgv, sgv);

		SplitRowValues splitRowValues = rowValuesGetter.getSplitRowValues(tablesRepository, tableIdsSet);
		
		
		return cache.getOrCreate(input.getFullCandidatesAllClasses().getId(), () -> {
			return getInner(config, splitGroupedValuesSets, splitRowValues, input.getFullCandidatesAllClasses());
		});
	}

	private AttributeValueContainer getInner(RowClusteringConfig config, SplitGroupedValuesSets splitGroupedValuesSets, SplitRowValues splitRowValues,
			TableRowIdsSet fullRowIds) {
		// Map<ClassId, Map<TableRowIdPair, Double>> scoresMap = new
		// ConcurrentHashMap<>();
		// SplitGroupedValuesSets splitGroupedValuesSets =
		// input.getGroupedValuesSets();
		// ConcurrentHashMap<TableRowId, Map<TableRowId, AttributeScores>> ret =
		// new ConcurrentHashMap<>();

		List<ScoresContainer> scoresContainer = new ArrayList<>();
		for (Pair<FusionStrategy, PropertiesConfig> setup : config.getSetExpansionStrategiesConfig().getFusionSetups()) {
			Scorer scorer = scorerRepository.get(setup.getFirst().getScoringMethod());
			ScoresContainer scores = scorer.getScores(splitGroupedValuesSets);
			scoresContainer.add(scores);
		}

		// Set<QualifiedPropertyId> splitted = null;

		List<Splitter> splitters = new ArrayList<>();

		for (Pair<FusionStrategy, PropertiesConfig> setup : config.getSetExpansionStrategiesConfig().getFusionSetups()) {
			FusionStrategy strategy = setup.getFirst();

			Scorer scoresGetter = scorerRepository.get(strategy.getScoringMethod());

			if (scoresGetter instanceof SplitterProvider) {
				splitters.add(((SplitterProvider) scoresGetter).getSplitter(splitGroupedValuesSets));

			} else {
				Splitter splitter = nonTimeDependentSplitter;
				splitters.add(splitter);
			}
		}

		List<ThresholdContainer> thresholdContainers = config.getThresholdContainers();

		List<Pair<FusionStrategy, PropertiesConfig>> fusionSetups = config.getSetExpansionStrategiesConfig().getFusionSetups();

		ConcurrentHashMap<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId,AbstractObject>>> kbtValues = new ConcurrentHashMap<>();
		ConcurrentHashMap<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId, Double>>> kbtScores = new ConcurrentHashMap<>();

		ConcurrentHashMap<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId,AbstractObject>>> directValues = new ConcurrentHashMap<>();

		ScoresContainer scoresContainerConstant = new FixedScore(new Probability(new Accuracy(1.0, 1)));

		System.out.println("If Threshold does not exist, we assuming a threshold of 1.1, check it out ");
		
		Map<TableRowId, Map<QualifiedPropertyId, MatchedValues>> valuesByTableRowProperty = splitRowValues.getValues();

		ConcurrentHashMap<ClassPropertyId, Double> thresholds = new ConcurrentHashMap<>();
		
		PersistentParallel6.runAndFinish(fullRowIds.getRowIds(), (tableRowId) -> {
			ClassId classId = tablesRepository.getClassId(tableRowId.getTableId());// input.getClassId();

		 
			// Map<PropertyId, MatchedValues> rowValues =
			// valuesByTableRowProperty.get(tableRowId);
			//
			// Map<QualifiedPropertyId, MatchedValues> splittedValues =
			// split(fusionSetups, splitters, classId, row.getValues());

			Map<QualifiedPropertyId, MatchedValues> splittedValues = valuesByTableRowProperty.get(tableRowId);
			if (splittedValues != null) {

				for (Entry<QualifiedPropertyId, MatchedValues> firstEntry : splittedValues.entrySet()) {
					MatchedValues firstValue = firstEntry.getValue();

					int scoresIndex = -1;
					for (ListEntry<Pair<FusionStrategy, PropertiesConfig>> setupListEntry : ListIterable.iterable(fusionSetups)) {
						if (propertiesConfigFilter.include(setupListEntry.getValue().getSecond(), firstEntry.getKey().getPropertyId())) {
							scoresIndex = setupListEntry.getIndex();
						}
					}

					if (scoresIndex == -1) {
						throw new IllegalStateException();
					}

					int finalScoresIndex = scoresIndex;
					Double threshold = CHMHelper.getFromCHM(thresholds, new ClassPropertyId(classId, firstEntry.getKey().getPropertyId()), () -> {
						Double tt = thresholdContainers.get(finalScoresIndex).getScore(classId, firstEntry.getKey().getPropertyId(), null);

						if (tt == null) {
							tt = 1.1;
						}

						return tt;
					});

					TripleId tripleId = new TripleId(classId, firstEntry.getKey().getPropertyId(), null);
					// Optional<Pair<AbstractObject, Double>> kbtObject =
					// getValue(tripleId, firstEntry.getKey().getQualifiers(),
					// firstValue, scoresContainer.get(scoresIndex), threshold);

					// if (kbtObject.isPresent()) {
					//
					//
					// CHMHelper.getFromCHM(kbtValues, tableRowId, () -> new
					// ConcurrentHashMap<>()).put(firstEntry.getKey(),
					// kbtObject.get().getFirst());
					//
					// double score = 0.0;
					// if(kbtObject.get().getSecond() != null) {
					// score = kbtObject.get().getSecond();
					// }
					//
					// CHMHelper.getFromCHM(kbtScores, tableRowId, () -> new
					// ConcurrentHashMap<>()).put(firstEntry.getKey(), score);
					// }
					//
					// Optional<Pair<AbstractObject, Double>> direct =
					// getValue(tripleId, firstEntry.getKey().getQualifiers(),
					// firstValue, scoresContainerConstant, -0.01);
					//
					// if (direct.isPresent()) {
					//
					// CHMHelper.getFromCHM(directValues, tableRowId, () -> new
					// ConcurrentHashMap<>()).put(firstEntry.getKey(),
					// direct.get().getFirst());
					//
					// }

					for (SingleValue sv : firstValue.getValues()) {

						double score = 0.0;
						Probability scoreProbability = scoresContainer.get(scoresIndex).getScore(tripleId, sv.getId(), firstEntry.getKey().getQualifiers());
						if (scoreProbability != null) {
							score = scoreProbability.get(SimilarityType.DIRECT).getAccuracy();
						}

						ConcurrentHashMap<QualifiedPropertyId, Map<TableColumnId, Double>> scoresInnerMap = (ConcurrentHashMap<QualifiedPropertyId, Map<TableColumnId, Double>>) CHMHelper
								.getFromCHM(kbtScores, tableRowId, () -> new ConcurrentHashMap<>());
						CHMHelper.getFromCHM(scoresInnerMap, firstEntry.getKey(), () -> new ConcurrentHashMap<>()).put(sv.getId().getTableColumnId(), score);

						if (score > threshold) {
							ConcurrentHashMap<QualifiedPropertyId, Map<TableColumnId, AbstractObject>> kbtInnerMap = (ConcurrentHashMap<QualifiedPropertyId, Map<TableColumnId, AbstractObject>>) CHMHelper
									.getFromCHM(kbtValues, tableRowId, () -> new ConcurrentHashMap<>());
							CHMHelper.getFromCHM(kbtInnerMap, firstEntry.getKey(), () -> new ConcurrentHashMap<>()).put(sv.getId().getTableColumnId(),
									sv.getValue());

						}

						ConcurrentHashMap<QualifiedPropertyId, Map<TableColumnId, AbstractObject>> directInnerMap = (ConcurrentHashMap<QualifiedPropertyId, Map<TableColumnId, AbstractObject>>) CHMHelper
								.getFromCHM(directValues, tableRowId, () -> new ConcurrentHashMap<>());
						CHMHelper.getFromCHM(directInnerMap, firstEntry.getKey(), () -> new ConcurrentHashMap<>()).put(sv.getId().getTableColumnId(),
								sv.getValue());
					}

					}
					
				}

		} , this.getClass());

		return new AttributeValueContainer(kbtValues, kbtScores, directValues, new ThresholdClassPropertyId(thresholds));
	}

	Map<QualifiedPropertyId, MatchedValues> split(List<Pair<FusionStrategy, PropertiesConfig>> fusionSetups, List<Splitter> splitters, ClassId classId,
			Map<PropertyId, MatchedValues> unqualified) {
		Map<QualifiedPropertyId, MatchedValues> retMap = new HashMap<>();
		for (Entry<PropertyId, MatchedValues> entry : unqualified.entrySet()) {

			for (SingleValue sv : entry.getValue().getValues()) {
				// FusionStrategy strategy = null;
				// for (Pair<FusionStrategy, PropertiesConfig> setup :
				// fusionSetups) {
				// if (propertiesConfigFilter.include(setup.getSecond(),
				// entry.getKey())) {
				// strategy = setup.getFirst();
				// break;
				// }
				// }

				int scoresIndex = -1;
				for (ListEntry<Pair<FusionStrategy, PropertiesConfig>> setupListEntry : ListIterable.iterable(fusionSetups)) {
					if (propertiesConfigFilter.include(setupListEntry.getValue().getSecond(), entry.getKey())) {
						scoresIndex = setupListEntry.getIndex();
					}
				}

				if (scoresIndex == -1) {
					throw new IllegalStateException();
				}

				Set<QualifiedPropertyId> splitted = splitters.get(scoresIndex).split(classId, entry.getKey(), sv.getId());

				for (QualifiedPropertyId qpid : splitted) {
					retMap.put(qpid, entry.getValue());
				}
			}

		}
		return retMap;
	}

	private Optional<Pair<AbstractObject, Double>> getValue(TripleId tripleId, Set<AbstractQualifier> qualifiers, MatchedValues mv, ScoresContainer scoresContainer, double threshold) {
		Double firstMaxScore = null;
		AbstractObject value = null;
		for (SingleValue sv : mv.getValues()) {

			Probability scoreProbability = scoresContainer.getScore(tripleId, sv.getId(), qualifiers);
			if (scoreProbability != null) {
				double score = scoreProbability.get(SimilarityType.DIRECT).getAccuracy();

				if (score > threshold && (firstMaxScore == null || score > firstMaxScore)) {
					value = sv.getValue();
					firstMaxScore = score;
				}
			}
		}
		if(value == null)
			return Optional.empty();
		else
			return Optional.of(new Pair<>(value, firstMaxScore));
	}

	public static class AttributeValueContainer {

		private final Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>>> kbtValues;
		private final Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId, Double>>> kbtScores;
		private final Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>>> directValues;
		private final ThresholdContainer thresholdContainer;

		public AttributeValueContainer(Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId,AbstractObject>>> kbtValues,
				ConcurrentHashMap<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId,Double>>> kbtScores, Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId,AbstractObject>>> directValues,
				ThresholdContainer thresholdContainer) {
			super();
			this.kbtValues = Collections.unmodifiableMap(kbtValues.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				return Collections.unmodifiableMap(e.getValue());
			})));
			this.kbtScores = Collections.unmodifiableMap(kbtScores.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				return Collections.unmodifiableMap(e.getValue());
			})));
			this.directValues = Collections.unmodifiableMap(directValues.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				return Collections.unmodifiableMap(e.getValue());
			})));
			this.thresholdContainer = thresholdContainer;
		}

		public Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>>> getKbtValues() {
			return kbtValues;
		}

		public Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId, AbstractObject>>> getDirectValues() {
			return directValues;
		}

		public Map<TableRowId, Map<QualifiedPropertyId, Map<TableColumnId, Double>>> getKbtScores() {
			return kbtScores;
		}

		public ThresholdContainer getThresholdContainer() {
			return thresholdContainer;
		}


	}
}
