package expansion.set.strategies.clustersplitters.minScoreSplitter;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import expansion.set.clustering.model.Block;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;

public abstract class AbstractAggregatedSimlitarities implements RowClusterSimilarities {

	private final List<RowClusterSimilarities> sims;

	private final boolean supportsBlocking;
	private final int totalBlockers;

	public AbstractAggregatedSimlitarities(List<RowClusterSimilarities> sims) {
		this.sims = Collections.unmodifiableList(new ArrayList<>(sims));

		boolean supportsBlocking = false;
		int totalBlockers = 0;
		for (RowClusterSimilarities sim : sims) {

			this.blockMap.add(new ConcurrentHashMap<>());
			List<ConcurrentHashMap<Integer, Integer>> innerBlockMap = new ArrayList<>();

			int blockerSizes = sim.getBlockerSizes();
			for (int i = 0; i < blockerSizes; i++) {
				innerBlockMap.add(new ConcurrentHashMap<>());
				totalBlockers++;
			}
			this.blockMapDouble.add(innerBlockMap);

			if (sim.supportsBlocking()) {

				supportsBlocking = true;
			}
		}
		this.totalBlockers = totalBlockers;
		this.supportsBlocking = supportsBlocking;

	}

	List<ConcurrentHashMap<Integer, Integer>> blockMap = new ArrayList<>();
	List<List<ConcurrentHashMap<Integer, Integer>>> blockMapDouble = new ArrayList<>();

	List<List<ConcurrentHashMap<Integer, Block>>> blockMapDoubleBlock = new ArrayList<>();

	private final AtomicInteger blockCounter = new AtomicInteger(0);

	public Set<Integer> getBlocksNew(TableRowId rowId) {
		Set<Integer> ret = new HashSet<>();
		for (ListEntry<RowClusterSimilarities> listEntry : ListIterable.iterable(sims)) {
			if (listEntry.getValue().supportsBlocking()) {

				for (ListEntry<Set<Integer>> innerEntry : ListIterable.iterable(listEntry.getValue().getSplitBlocks(rowId))) {

					for (Integer simBlock : innerEntry.getValue()) {
						int block = CHMHelper.getFromCHM(blockMapDouble.get(listEntry.getIndex()).get(innerEntry.getIndex()), simBlock,
								() -> blockCounter.getAndIncrement());
						ret.add(block);
					}

				}

			}

		}
		return ret;
	}

	@Override
	public Set<Integer> getBlocks(TableRowId rowId) {
		Set<Integer> ret = new HashSet<>();
		for (ListEntry<RowClusterSimilarities> listEntry : ListIterable.iterable(sims)) {
			RowClusterSimilarities sim = listEntry.getValue();
			if (sim.supportsBlocking()) {

				for (ListEntry<Tri<Set<Integer>, Integer, Double>> innerEntry : ListIterable.iterable(listEntry.getValue().getSplitBlocksWithScores(rowId))) {
					Set<Integer> simAdaptedBlocks = new HashSet<>();
					Tri<Set<Integer>, Integer, Double> simBlocks = innerEntry.getValue();

					for (Integer simBlock : simBlocks.getA()) {
						List<ConcurrentHashMap<Integer, Integer>> map = blockMapDouble.get(listEntry.getIndex());
						if (innerEntry.getIndex() == 3) {
							int sdf = 123;
						}
						ConcurrentHashMap<Integer, Integer> map2 = map.get(innerEntry.getIndex());
						int block = CHMHelper.getFromCHM(map2, simBlock, () -> blockCounter.getAndIncrement());
						simAdaptedBlocks.add(block);
					}

					ret.addAll(simAdaptedBlocks);
				}
			}

		}
		return ret;
	}

	public Set<Integer> getBlocksOld(TableRowId rowId) {
		Set<Integer> ret = new HashSet<>();
		for (ListEntry<RowClusterSimilarities> listEntry : ListIterable.iterable(sims)) {
			if (listEntry.getValue().supportsBlocking()) {

				Set<Integer> simBlocks = listEntry.getValue().getBlocks(rowId);
				if (simBlocks != null) {
					for (Integer simBlock : simBlocks) {
						int block = CHMHelper.getFromCHM(blockMap.get(listEntry.getIndex()), simBlock, () -> blockCounter.getAndIncrement());
						ret.add(block);
					}
				}
			}

		}
		return ret;
	}

	@Override
	public List<Set<Integer>> getSplitBlocks(TableRowId rowId) {
		List<Set<Integer>> ret = new ArrayList<>();
		for (ListEntry<RowClusterSimilarities> listEntry : ListIterable.iterable(sims)) {
			if (listEntry.getValue().supportsBlocking()) {
				Set<Integer> simAdaptedBlocks = new HashSet<>();
				Set<Integer> simBlocks = listEntry.getValue().getBlocks(rowId);
				if (simBlocks != null) {
					for (Integer simBlock : simBlocks) {
						int block = CHMHelper.getFromCHM(blockMap.get(listEntry.getIndex()), simBlock, () -> blockCounter.getAndIncrement());
						simAdaptedBlocks.add(block);
					}
				}
				ret.add(simAdaptedBlocks);
			}

		}
		return ret;
	}

	public List<Set<Integer>> getSplitBlocksNew(TableRowId rowId) {
		List<Set<Integer>> ret = new ArrayList<>();
		for (ListEntry<RowClusterSimilarities> listEntry : ListIterable.iterable(sims)) {
			if (listEntry.getValue().supportsBlocking()) {
				for (ListEntry<Set<Integer>> innerEntry : ListIterable.iterable(listEntry.getValue().getSplitBlocks(rowId))) {

					Set<Integer> simAdaptedBlocks = new HashSet<>();
					Set<Integer> simBlocks = innerEntry.getValue();

					for (Integer simBlock : simBlocks) {
						int block = CHMHelper.getFromCHM(blockMapDouble.get(listEntry.getIndex()).get(innerEntry.getIndex()), simBlock,
								() -> blockCounter.getAndIncrement());
						simAdaptedBlocks.add(block);
					}

					ret.add(simAdaptedBlocks);
				}
			}

		}
		return ret;
	}

	@Override
	public List<Pair<Set<Integer>, Integer>> getSplitBlocksWithMin(TableRowId rowId) {
		List<Pair<Set<Integer>, Integer>> ret = new ArrayList<>();
		for (ListEntry<RowClusterSimilarities> listEntry : ListIterable.iterable(sims)) {
			RowClusterSimilarities sim = listEntry.getValue();
			if (sim.supportsBlocking()) {
				Set<Integer> simAdaptedBlocks = new HashSet<>();
				Set<Integer> simBlocks = sim.getBlocks(rowId);
				if (simBlocks != null) {
					for (Integer simBlock : simBlocks) {
						int block = CHMHelper.getFromCHM(blockMap.get(listEntry.getIndex()), simBlock, () -> blockCounter.getAndIncrement());
						simAdaptedBlocks.add(block);
					}
				}
				ret.add(new Pair<>(simAdaptedBlocks, sim.getBlockMin(rowId)));
			}

		}
		return ret;
	}

	public List<Pair<Set<Integer>, Integer>> getSplitBlocksWithMinNew(TableRowId rowId) {
		List<Pair<Set<Integer>, Integer>> ret = new ArrayList<>();
		for (ListEntry<RowClusterSimilarities> listEntry : ListIterable.iterable(sims)) {
			RowClusterSimilarities sim = listEntry.getValue();
			if (sim.supportsBlocking()) {

				for (ListEntry<Set<Integer>> innerEntry : ListIterable.iterable(listEntry.getValue().getSplitBlocks(rowId))) {
					Set<Integer> simAdaptedBlocks = new HashSet<>();
					Set<Integer> simBlocks = innerEntry.getValue();

					for (Integer simBlock : simBlocks) {
						int block = CHMHelper.getFromCHM(blockMapDouble.get(listEntry.getIndex()).get(innerEntry.getIndex()), simBlock,
								() -> blockCounter.getAndIncrement());
						simAdaptedBlocks.add(block);
					}

					ret.add(new Pair<>(simAdaptedBlocks, sim.getBlockMin(rowId)));
				}
			}

		}
		return ret;
	}

	public int getBlockerSizes() {
		return totalBlockers;
	}


	public List<Tri<Set<Integer>, Integer, Double>> getSplitBlocksWithScores(TableRowId rowId) {
		List<Tri<Set<Integer>, Integer, Double>> ret = new ArrayList<>();
		for (ListEntry<RowClusterSimilarities> listEntry : ListIterable.iterable(sims)) {
			RowClusterSimilarities sim = listEntry.getValue();
			if (sim.supportsBlocking()) {

				for (ListEntry<Tri<Set<Integer>, Integer, Double>> innerEntry : ListIterable.iterable(listEntry.getValue().getSplitBlocksWithScores(rowId))) {
					Set<Integer> simAdaptedBlocks = new HashSet<>();
					Tri<Set<Integer>, Integer, Double> simBlocks = innerEntry.getValue();

					for (Integer simBlock : simBlocks.getA()) {
						List<ConcurrentHashMap<Integer, Integer>> map = blockMapDouble.get(listEntry.getIndex());
						if (innerEntry.getIndex() == 3) {
							int sdf = 123;
						}
						ConcurrentHashMap<Integer, Integer> map2 = map.get(innerEntry.getIndex());
						int block = CHMHelper.getFromCHM(map2, simBlock, () -> blockCounter.getAndIncrement());
						simAdaptedBlocks.add(block);
					}

					ret.add(new Tri<>(simAdaptedBlocks, innerEntry.getValue().getB(), innerEntry.getValue().getC()));
				}
			}

		}
		return ret;
	}

	@Override
	public boolean supportsBlocking() {
		return supportsBlocking;
	}
}
