package expansion.set.strategies.clustersplitters.minScoreSplitter;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.Serializable;
import java.util.Collection;
import java.util.Optional;

import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationMethod;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.RecordPrediction;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordData;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordFeature;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepositoryImpl;
import weka.core.SerializationHelper;

public class CombinedModel implements LearnedModel, Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	final LearnedModel aModel;
	final LearnedModel bModel;
	final WeightsRepository weights;
	transient final SimpleLockableConcurrentCache<TableRowIdPair, Double> aCache;
	transient final SimpleLockableConcurrentCache<TableRowIdPair, Double> bCache;

	public CombinedModel(byte[] input) {
		ByteArrayInputStream stream = new ByteArrayInputStream(input);
		try {
			CombinedModel readModel = (CombinedModel) weka.core.SerializationHelper.read(stream);
			aModel = readModel.aModel;
			bModel = readModel.bModel;
			weights = readModel.weights;
			aCache = new SimpleLockableConcurrentCache<>();
			bCache = new SimpleLockableConcurrentCache<>();

		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
	}
	public CombinedModel(LearnedModel aModel, LearnedModel bModel, WeightsRepository weights) {
		super();
		this.aModel = aModel;
		this.bModel = bModel;
		aCache = new SimpleLockableConcurrentCache<>();
		bCache = new SimpleLockableConcurrentCache<>();
		this.weights = new WeightsRepositoryImpl(weights);
	}

	public CombinedModel(CombinedModel casted, WeightsRepository weights) {
		this.aModel = casted.aModel;
		this.bModel = casted.bModel;
		this.aCache = casted.aCache;
		this.bCache = casted.bCache;
		this.weights = weights;
	}
	@Override
	public RecordPrediction apply(RecordFeature d) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public byte[] outputModelFile() {
		ByteArrayOutputStream stream = new ByteArrayOutputStream();

		try {

			SerializationHelper.write(stream, this);

			return stream.toByteArray();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
	}

	@Override
	public double runAndEvaluate(Collection<RecordData> data, EvaluationMethod evaluator) {
		throw new UnsupportedOperationException();
	}

	@Override
	public Optional<Double> getNormalizingThreshold() {
		return Optional.empty();
	}

	@Override
	public String toString() {
		return "aModel \n" + aModel + " \n\n\n bModel \n" + bModel + " \n\n\n weights \n" + weights;
	}

	public SimpleLockableConcurrentCache<TableRowIdPair, Double> getACache() {
		return aCache;
	}

	public SimpleLockableConcurrentCache<TableRowIdPair, Double> getBCache() {
		return bCache;
	}


}