package expansion.set.strategies.clustersplitters.minScoreSplitter;

import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Function;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationMethod;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.learner.WeightsModelLearner;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.genetic.WeightsGeneticParallelOptimizer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepositoryImpl;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;

public class CombinedRowClusterSimilaritiesCreator {

	@Inject
	protected WeightsGeneticParallelOptimizer genetic;

	@Inject
	protected WeightsModelLearner learner;

	@Inject
	public CombinedRowClusterSimilaritiesCreator() {
	}

	public CombinedRowClusterSimilarities get(String name, OptimizableClass optimizableClass, List<RowClusterSimilarities> sims,
			Function<LearnedModel, LearnableRowClusterSimilarities> a, Function<LearnedModel, LearnableRowClusterSimilarities> b, LearnedModel model) {
		return new CombinedRowClusterSimilarities(name, optimizableClass, sims, a, b, model, false);
	}

	public CombinedRowClusterSimilarities get(String name, OptimizableClass optimizableClass, List<RowClusterSimilarities> sims,
			Function<LearnedModel, LearnableRowClusterSimilarities> a, Function<LearnedModel, LearnableRowClusterSimilarities> b, LearnedModel model,
			boolean withCache) {
		return new CombinedRowClusterSimilarities(name, optimizableClass, sims, a, b, model, withCache);
	}
	public class CombinedRowClusterSimilarities extends AbstractAggregatedSimlitarities implements LearnableRowClusterSimilarities {

		private final LearnableRowClusterSimilarities a;
		private final LearnableRowClusterSimilarities b;
		private final OptimizableClass optimizableClass;
		private final String name;
		private final WeightsRepository weights;
		private final boolean withCache;

		private final Function<LearnedModel, LearnableRowClusterSimilarities> aFunction;
		private final Function<LearnedModel, LearnableRowClusterSimilarities> bFunction;

		private final SimpleLockableConcurrentCache<TableRowIdPair, Double> aCache;
		private final SimpleLockableConcurrentCache<TableRowIdPair, Double> bCache;
		private final Double bNormalizer;
		public LearnableRowClusterSimilarities getASim() {
			return a;
		}

		public CombinedRowClusterSimilarities(String name, OptimizableClass optimizableClass, List<RowClusterSimilarities> sims,
				Function<LearnedModel, LearnableRowClusterSimilarities> a, Function<LearnedModel, LearnableRowClusterSimilarities> b, LearnedModel model,
				boolean withCache) {
			super(sims);

			this.withCache = withCache;
			this.name = name;
			this.optimizableClass = optimizableClass;

			this.aFunction = a;
			this.bFunction = b;

			if (model != null) {
				CombinedModel castModel;
				if (model instanceof CombinedModel) {
					castModel = (CombinedModel) model;

					this.a = a.apply(castModel.aModel);
					this.b = b.apply(castModel.bModel);
					this.bNormalizer = castModel.bModel.getNormalizingThreshold().get();
					this.weights = castModel.weights;
				} else {
					throw new IllegalStateException();
				}
				this.aCache = ((CombinedModel) model).getACache();
				this.bCache = ((CombinedModel) model).getBCache();
			} else {
				this.a = null;
				this.b = null;
				this.weights = null;
				this.aCache = null;
				this.bCache = null;
				this.bNormalizer = null;
			}
		}

		@Override
		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {

			return get(rows, classId, key, a, b, weights);
		}

		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key, LearnableRowClusterSimilarities a, LearnableRowClusterSimilarities b,
				WeightsRepository weights) {
			double bNormalizerUpper = this.bNormalizer;
			double bNormalizerLower = bNormalizerUpper;

			double aScore;
			double bScore;
			if (withCache) {
				TableRowIdPair pair = new TableRowIdPair(rows.get(key.getE1()), rows.get(key.getE2()));
				aScore = aCache.getOrCreateNoBlocking(pair, () -> {
					return a.get(rows, classId, key);
				});
				bScore = bCache.getOrCreateNoBlocking(pair, () -> {
					return b.get(rows, classId, key);
				});

			} else {
				aScore = a.get(rows, classId, key);
				bScore = b.get(rows, classId, key);
			}

			double t;
			if (bScore > bNormalizerUpper) {
				t = ((bScore - bNormalizerUpper) / (1.0 - bNormalizerUpper));
			} else if (bScore < bNormalizerLower) {
				t = -1.0 * (1.0 - (bScore) / (bNormalizerLower));
			} else {
				t = 0.0;
			}

			if (true) {
				double aWeight = weights.getOrDefault(optimizableClass.getWeight("aWeight", 0.50));

				double bWeight = 1.0 - aWeight;

				return (aWeight * aScore) + (bWeight * t);
			} else {
				if (aScore > 0.0 & t > 0.0) {
					return aScore * t;
				} else if (aScore < 0.0 & t < 0.0) {
					return -1.0 * aScore * t;
				} else {
					return 0.0;
				}
			}

		}

		@Override
		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getName() {
			return "COMBINED_" + name;
		}

		@Override
		public LearnedModel learn(Map<TableRowIdPair, PairLearningData> pairs, ClassId classId, EvaluationMethod evaluationMethod) {
			LearnedModel aLearned = aFunction.apply(null).learn(pairs, classId, evaluationMethod);

			LearnedModel bLearned = bFunction.apply(null).learn(pairs, classId, evaluationMethod);

			LearnableRowClusterSimilarities a = aFunction.apply(aLearned);
			LearnableRowClusterSimilarities b = bFunction.apply(bLearned);



			// WeightsRepository finalWeights = genetic.optimize((weights) -> {
			//
			//
			//
			//
			// List<EvaluationPrediction> predictions = new ArrayList<>();
			// for (Entry<TableRowIdPair, PairLearningData> entry :
			// pairs.entrySet()) {
			//
			// List<TableRowId> list = new ArrayList<>();
			// list.add(entry.getKey().getId1());
			// list.add(entry.getKey().getId2());
			//
			// double score = get(list, classId, new ClusterEntryKey(0, 1), a,
			// b, weights);
			//
			// if (score > 0.0) {
			// predictions.add(new
			// EvaluationPrediction(entry.getValue().getLabel(),
			// entry.getValue().getWeight(),
			// new
			// RecordPrediction(LearnableRowClusterSimilarities.getLabel(true),
			// 1.0)));
			// } else {
			// predictions.add(new
			// EvaluationPrediction(entry.getValue().getLabel(),
			// entry.getValue().getWeight(),
			// new
			// RecordPrediction(LearnableRowClusterSimilarities.getLabel(false),
			// 1.0)));
			// }
			//
			// }
			//
			// return evaluationMethod.performance(predictions);
			// });

			return new CombinedModel(aLearned, bLearned, new WeightsRepositoryImpl());
		}

	}
}
