package expansion.set.strategies.clustersplitters.minScoreSplitter;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;

import javax.inject.Inject;

import com.google.common.collect.Sets;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationMethod;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationPrediction;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.RecordPrediction;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.learner.WeightsLearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.learner.WeightsModelLearner;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.LearnedModelRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepositoryImpl;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;

public class WeightedAverageScoreSplitterFactory {
 
	
//	private final AnnotatedTablesRepository repository;

	@Inject
	protected WeightsModelLearner learner;

	@Inject
	public WeightedAverageScoreSplitterFactory(  ) {
		super(); 
//		this.repository = repository;
	}

	public RowClusterSimilarities getScores(List<RowClusterSimilarities> scores) {
		List<Pair<RowClusterSimilarities, Double>> list = new ArrayList<>();
		for(RowClusterSimilarities score : scores) {
			list.add(new Pair<>(score, 1.0));
		}
		return new AverageScoresSplitterSimilarities(list, true);
	}
	
	
	public RowClusterSimilarities getSplitterWithScores(List<Pair<RowClusterSimilarities, Double>> scores) {
 
		return new AverageScoresSplitterSimilarities(scores, true);
	}


	public RowClusterSimilarities getSplitterWithScores(List<Pair<RowClusterSimilarities, Double>> scores, WeightsRepository weights,
			OptimizableClass optimizableClass) {

		return new AverageScoresSplitterSimilarities(scores, weights, optimizableClass, learner, true);
	}


	public RowClusterSimilarities getSplitter(List<RowClusterSimilarities> scores, LearnedModelRepository weights, OptimizableClass optimizableClass) {
		return getSplitter(scores, weights, optimizableClass, true);
	}

	public AverageScoresSplitterSimilarities getSplitter(List<RowClusterSimilarities> scores, LearnedModelRepository weights, OptimizableClass optimizableClass,
			boolean sophisticatedMethod) {
		List<Pair<RowClusterSimilarities, Double>> list = new ArrayList<>();
		for (RowClusterSimilarities score : scores) {
			list.add(new Pair<>(score, 1.0));
		}
		if (weights == null) {
			return new AverageScoresSplitterSimilarities(list, null, optimizableClass, learner, sophisticatedMethod);
		}

		else {
			LearnedModel model = weights.get(optimizableClass);
		if (model == null) {
				return new AverageScoresSplitterSimilarities(list, null, optimizableClass, learner, sophisticatedMethod);
			} else if (model instanceof WeightsLearnedModel) {
				WeightsRepository rep = ((WeightsLearnedModel) model).getWeights();
				return new AverageScoresSplitterSimilarities(list, rep, optimizableClass, learner, sophisticatedMethod);
		} else {
			throw new IllegalStateException();
		}
		}

	}

	public AverageScoresSplitterSimilarities getSplitter(List<RowClusterSimilarities> scores, LearnedModel model, OptimizableClass optimizableClass,
			boolean sophisticatedMethod) {
		List<Pair<RowClusterSimilarities, Double>> list = new ArrayList<>();
		for (RowClusterSimilarities score : scores) {
			list.add(new Pair<>(score, 1.0));
		}

		if (model == null) {
			return new AverageScoresSplitterSimilarities(list, null, optimizableClass, learner, sophisticatedMethod);
		} else {
			if (model instanceof WeightsLearnedModel) {
				WeightsRepository rep = ((WeightsLearnedModel) model).getWeights();
				return new AverageScoresSplitterSimilarities(list, rep, optimizableClass, learner, sophisticatedMethod);
			} else {
				throw new IllegalStateException();
			}
		}

	}

//	private class AverageScoresSplitter implements RowClusterSplitter {
//
//		private final AverageScoresSplitterSimilarities sim;
//
//		private AverageScoresSplitter(List<Pair<RowClusterSimilarities, Double>> sims) {
//			super();
//			this.sim = new AverageScoresSplitterSimilarities(sims);
//		}
//
//		@Override
//		public ClusterList clusterRows(RowClusteringConfig config, ClustererInput input, ClusterList rowCluster) {
//
//			ConcurrentLinkedQueue<RowCluster> list = new ConcurrentLinkedQueue<>();
//			PersistentParallel6.runAndFinish(rowCluster.getClusterList(), (oldRc) -> {
////				if(oldRc.getRows().size() > 3000) {
////					return;
////				}
//				ClusterSimilarities cs = new ClusterSimilaritiesImpl(oldRc, sim);
//				list.addAll(clusterer.rowSplitter(oldRc, cs));
//			} , this.getClass());
//
//			return new ClusterList(list);
//
//		}
//		
//		@Override
//		public List<RowCluster> cluster(ClassId classId, List<TableRowId> list) {
//			return clusterer.rowSplitter(classId, list, sim);
//		}
//
//	}

	public class AverageScoresSplitterSimilarities extends AbstractAggregatedSimlitarities implements LearnableRowClusterSimilarities {
		private final List<Pair<RowClusterSimilarities, Double>> sims;
		private final boolean normalized;
		private final WeightsRepository weights;
		private final OptimizableClass optimizableClass;

		private final WeightsModelLearner learner;
		private final boolean sophisticatedMethod;

		// private final Double normalizer;
		private AverageScoresSplitterSimilarities(List<Pair<RowClusterSimilarities, Double>> sims, boolean sophisticatedMethod) {
			super(sims.stream().map((e) -> e.getFirst()).collect(Collectors.toList()));
			this.sims = sims;
			this.normalized = true;
			this.weights = null;
			this.optimizableClass = null;

			this.learner = null;
			this.sophisticatedMethod = sophisticatedMethod;
			// this.normalizer = null;
		}


		public AverageScoresSplitterSimilarities(List<Pair<RowClusterSimilarities, Double>> sims, WeightsRepository weights,
				OptimizableClass optimizableClass, WeightsModelLearner learner, boolean sophisticatedMethod) {
			super(sims.stream().map((e) -> e.getFirst()).collect(Collectors.toList()));
			this.sims = sims;
			this.normalized = false;
			this.optimizableClass = optimizableClass;
			this.weights = weights;
			this.sophisticatedMethod = sophisticatedMethod;
			// this.normalizer = null;
			// Map<String, Double> allWeights =
			// weights.getWeights(optimizableClass);
			// if (allWeights.isEmpty()) {
			// this.normalizer = null;
			// } else {
			// this.normalizer =
			// allWeights.values().stream().mapToDouble(Double::doubleValue).sum();
			// }
			this.learner = learner;

		}

		public Map<String, Double> getIndividualScores(ClassId classId, TableRowId rowId1, TableRowId rowId2) {
			Map<String, Double> ret = new LinkedHashMap<>();

			List<TableRowId> rowIds = new ArrayList<>();
			rowIds.add(rowId1);
			rowIds.add(rowId2);

			ClusterEntryKey key = new ClusterEntryKey(0, 1);
			for (Pair<RowClusterSimilarities, Double> sim : sims) {
				List<WeightedScore> similarity = sim.getFirst().getList(rowIds, classId, key);

				double innerSum = 0.0;
				double innerCount = 0.0;

				double weight = sim.getSecond();

				if (weights != null) {
					weight = weight * weights.getOrDefault(optimizableClass.getWeight(sim.getFirst().getName(), 1.0));
				}

				for (WeightedScore ws : similarity) {
					double suffixWeight = 1.0;
					if (sophisticatedMethod && weights != null) {
						if (ws.getSuffix().isPresent()) {
							suffixWeight = weights.getOrDefault(optimizableClass.getWeight(sim.getFirst().getName() + "_INDIVIDUAL", 0.5)) / 2.0;
						}
					}

					double finalWeight = ws.getWeight() * suffixWeight;

					// if (normalizer != null)
					// finalWeight = finalWeight / normalizer;
					if (ws.isMatch()) {
						innerSum += ws.getScore() * finalWeight;
					}
					innerCount += finalWeight;
				}

				// if (similarity.isEmpty()) {
				// double noScorePenalty = 1.0;
				// if (weights != null) {
				// noScorePenalty =
				// weights.getOrDefault(optimizableClass.getWeight(sim.getFirst().getName()
				// + "_NO_SCORE_PENALTY", 1.0));
				// }
				// innerCount = noScorePenalty * sim.getSecond();
				// }
				//
				if (sophisticatedMethod) {
					if (innerCount < 1.0) {
						double emptyScore = weights.getOrDefault(optimizableClass.getWeight(sim.getFirst().getName() + "_EMPTY", 1.0)) / 4.0;
						innerSum += (1.0 - innerCount) * emptyScore;
					}
					if (innerSum >= 1.0) {
						ret.put(sim.getFirst().getName(), 1.0);
					} else {
						ret.put(sim.getFirst().getName(), innerSum);
					}
				}
				else {
					if (innerCount > 0.0) {
						ret.put(sim.getFirst().getName(), innerSum / innerCount);
					} else {
						ret.put(sim.getFirst().getName(), 0.0);
					}
				}
				// if (innerCount > 0.001) {
				// sum += (innerSum / innerCount) * weight;
				//
				// }
				// count += weight;

				//
				// if(normalized) {
				//
				// sum += innerSum * sim.getSecond();
				// count += innerCount * sim.getSecond();
				// }
				// else {
				// sum += innerSum * sim.getSecond();
				// }
				//

			}

			ret.put("COMBINED", get(rowIds, classId, key));
			return ret;

		}

		// public static final String PREFIX = "NWA_";


		@Override
		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			return get(rows, classId, key, weights);
		}

		private final Set<String> weightOnlySet = Sets.newHashSet("ATTRIBUTE", "IMPLICIT_ATTRIBUTES");

		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key, WeightsRepository weights) {
			 
			Double sum = 0.0;
			Double count = 0.0;
			
			TableRowId row1 = rows.get(key.getE1());
			TableRowId row2 = rows.get(key.getE2()); 
			
			TableRowId r1 = new TableRowId(new AnnotatedTableId("1346981172155_1347001390512_1198.arc7550936969698106119#5262523_3_3250531893659458759"), 9);
			
			TableRowId r2 = new TableRowId(new AnnotatedTableId("1346876860779_1346959977349_969.arc1922261997516901431#71226425_0_6720751649015718857"), 7);
			
//			StringObject label1 = (StringObject) repository.getKeyValue(row1.getTableId(), row1.getRowIdx());
//			StringObject label2 = (StringObject)  repository.getKeyValue(row2.getTableId(), row2.getRowIdx());

			for (Pair<RowClusterSimilarities, Double> sim : sims) {
				List<WeightedScore> similarity = sim.getFirst().getList(rows, classId, key);

				double innerSum = 0.0;
				double innerCount = 0.0;

				double weight = sim.getSecond();

				boolean weightOnly = weightOnlySet.contains(sim.getFirst().getName());

				if (weights != null) {
					weight = weight * weights.getOrDefault(optimizableClass.getWeight(sim.getFirst().getName(), 1.0));
				}

				for(WeightedScore ws : similarity) {
					double suffixWeight = 1.0;
					double finalWeight = ws.getWeight() * suffixWeight;
					if (sophisticatedMethod && weights != null) {
						if (ws.getSuffix().isPresent()) {
							suffixWeight = weights.getOrDefault(optimizableClass.getWeight(sim.getFirst().getName() + "_INDIVIDUAL", 0.5)) / 4.0;
							finalWeight = ws.getWeight() * suffixWeight;
						}
					}



					// if (normalizer != null)
					// finalWeight = finalWeight / normalizer;
					if (weightOnly) {
						if (ws.isMatch()) {
							innerSum += finalWeight * ws.getScore();
														// *
						}
					} else {
						innerSum += ws.getScore() * finalWeight; // ws.getScore()
					}

					innerCount += finalWeight;
				}
				
				// if (similarity.isEmpty()) {
				// double noScorePenalty = 1.0;
				// if (weights != null) {
				// noScorePenalty =
				// weights.getOrDefault(optimizableClass.getWeight(sim.getFirst().getName()
				// + "_NO_SCORE_PENALTY", 1.0));
				// }
				// innerCount = noScorePenalty * sim.getSecond();
				// }
				//
				if (sophisticatedMethod) {
					if (sophisticatedMethod && innerCount < 1.0) {
						double emptyScore = weights.getOrDefault(optimizableClass.getWeight(sim.getFirst().getName() + "_EMPTY", 1.0)) / 2.0;
						innerSum += (1.0 - innerCount) * emptyScore;
					}
					if (innerSum >= 1.0 || (!sophisticatedMethod && innerCount > 0.0001)) {
						sum += (innerSum / innerCount) * weight;
					} else if (!sophisticatedMethod) {
						sum += innerSum * weight;
					}
				}
				else {
					if (innerCount > 0.0) {
						sum += (innerSum / innerCount) * weight;
					}
				}
				// if (innerCount > 0.001) {
				// sum += (innerSum / innerCount) * weight;
				//
				// }
				count += weight;
				
//		 
//					if(normalized) {
//					
//						sum += innerSum * sim.getSecond();
//						count +=  innerCount * sim.getSecond();
//					}
//					else {
//						sum += innerSum * sim.getSecond();
//					}
//			 
				

				

			}
			
			double finalSim = sum ;
//			if((label1 != null && label1.getString().contains("misi") ) && (label2 != null && label2.getString().contains("misi") )) {
//				int sdf= 123;
//			}
//			if((label1 != null && label1.getString().contains("mike") ) || (label2 != null && label2.getString().contains("mike") )) {
//				int sdf= 123;
//			}
//			if((label1 != null && label1.getString().contains("mike") ) && (label2 != null && label2.getString().contains("mike") )) {
//				int sdf= 123;
//			}

			// if (finalSim / count > 1.0) {
			// int sdf = 123;
			// }
			if (normalized || true) {
				 if(count > 0.0001)
					 return finalSim / count;
				 else
					 return 0.0;
			}
			else {
				return finalSim;
			}
				
		 }

		public Set<TableRowId> getLegalCandidates(RowCluster cluster, TableRowId point) {
			throw new UnsupportedOperationException();
		}

		@Override
		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getName() {
			return "WEIGHTED_SCORES";
		}



		@Override
		public LearnedModel learn(Map<TableRowIdPair, PairLearningData> pairs, ClassId classId, EvaluationMethod evaluationMethod) {
			PersistentParallel6.runAndFinish(pairs.entrySet(), (entry) -> {
				List<TableRowId> list = new ArrayList<>();
				list.add(entry.getKey().getId1());
				list.add(entry.getKey().getId2());

				double score = get(list, classId, new ClusterEntryKey(0, 1), new WeightsRepositoryImpl());
			}, this.getClass());

			return learner.learn((weights) -> {
				double threshold = weights.getOrDefault(optimizableClass.getWeight("SIMILARITY_THRESHOLD", 0.50));
				
				double penalty = (1.0 - threshold) * 0.05;

				List<EvaluationPrediction> predictions = new ArrayList<>();
				for(Entry<TableRowIdPair, PairLearningData> entry : pairs.entrySet()) {
					List<TableRowId> list = new ArrayList<>();
					list.add(entry.getKey().getId1());
					list.add(entry.getKey().getId2());

					double score = get(list, classId, new ClusterEntryKey(0, 1), weights);
					
					if (score > threshold) {
						predictions.add(
								new EvaluationPrediction(entry.getValue().getLabel(), entry.getValue().getWeight(),
										new RecordPrediction(LearnableRowClusterSimilarities.getLabel(true), 1.0)));
					}
					else {
						predictions.add(new EvaluationPrediction(entry.getValue().getLabel(), entry.getValue().getWeight(),
								new RecordPrediction(LearnableRowClusterSimilarities.getLabel(false), 1.0)));
					}

				}

				return evaluationMethod.performance(predictions) - penalty;
			}, optimizableClass);
			 
		}
	}
}
