package expansion.set.strategies.clustersplitters.minScoreSplitter;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentLinkedQueue;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableDouble;

import com.google.common.primitives.Ints;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.DataSetData;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationMethod;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearningFeatures;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.RecordPrediction;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordData;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordFeature;
import de.uni_mannheim.informatik.dws.expansion.utils.learning.weka.WekaLearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.learning.weka.WekaLearnerCreator;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;

public class WekaClassifierSimilaritiesCreator {

	@Inject
	public WekaClassifierSimilaritiesCreator() {
		super(); 
//		this.repository = repository;
	}

	public RowClusterSimilarities getSplitter(String classifier, String[] parameters, List<RowClusterSimilarities> sims, LearnedModel model) {

		return ClassifierBasedSimilarities.build(classifier, parameters, sims, model, true);

	}

	public RowClusterSimilarities getSplitter(String classifier, String[] parameters, List<RowClusterSimilarities> sims, LearnedModel model, boolean numeric) {

		return ClassifierBasedSimilarities.build(classifier, parameters, sims, model, numeric);

	}


	public static class ClassifierBasedSimilarities implements WekaLearnableRowClusterSimilarities {
		private final String classifierName;
		private final String[] paramters;
		private final DataSetData datasetData;
		private final WekaLearnedModel model;
		private final List<Pair<RowClusterSimilarities, Boolean>> sims;
		private final boolean numeric;
		private final boolean withConfidence = true;
		private final Set<PropertyId> propertyIds;
		private final boolean weighted;

		public static ClassifierBasedSimilarities build(String classifier, String[] parameters, List<RowClusterSimilarities> sims, LearnedModel model,
				boolean numeric) {
			List<Pair<RowClusterSimilarities, Boolean>> transformedSims = new ArrayList<>();

			for (RowClusterSimilarities sim : sims) {
				transformedSims.add(new Pair<>(sim, false));
			}

			return new ClassifierBasedSimilarities(classifier, parameters, transformedSims, model, numeric, null, false);
		}

		public ClassifierBasedSimilarities(String classifier, String[] parameters, List<Pair<RowClusterSimilarities, Boolean>> sims, LearnedModel model,
				boolean numeric, Set<PropertyId> propertyIds, boolean weighted) {
			this.weighted = weighted;
			if (propertyIds != null) {
				this.propertyIds = Collections.unmodifiableSet(new HashSet<>(propertyIds));
			} else {
				this.propertyIds = null;
			}


			List<String> featuresList = new ArrayList<>();
			for (Pair<RowClusterSimilarities, Boolean> sim : sims) {
				if (withConfidence && sim.getSecond().booleanValue()) {
					if (propertyIds != null) {
						for (PropertyId propertyId : propertyIds) {
							featuresList.add(sim.getFirst().getName() + "::" + propertyId.readableString() + "::MATCH");
							featuresList.add(sim.getFirst().getName() + "::" + propertyId.readableString() + "::NO_MATCH");
						}
					} else {
						featuresList.add(sim.getFirst().getName());
						featuresList.add(sim.getFirst().getName() + "::CONFIDENCE");
					}
				}
				else {

					featuresList.add(sim.getFirst().getName());
				}
			}

			if (numeric) {
				Map<String, Double> labels = new HashMap<>();
				labels.put(LearnableRowClusterSimilarities.getLabel(true), 1.0);
				labels.put(LearnableRowClusterSimilarities.getLabel(false), -1.0);
				this.datasetData = new DataSetData("TEST", labels, new LearningFeatures(featuresList));
			} else {
				List<String> classLabels = new ArrayList<>();
				classLabels.add(LearnableRowClusterSimilarities.getLabel(true));
				classLabels.add(LearnableRowClusterSimilarities.getLabel(false));
				this.datasetData = new DataSetData("TEST", classLabels, new LearningFeatures(featuresList));
			}

			this.classifierName = classifier;
			this.paramters = parameters;
			this.model = (WekaLearnedModel) model;
			this.sims = sims;
			this.numeric = numeric;

		}

		private final SimpleLockableConcurrentCache<Map<String, Double>, Double> cache = new SimpleLockableConcurrentCache<>();

		@Override
		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			if(model == null) {
				throw new IllegalStateException();
			}
			else {

				RecordFeature features = getFeatureVector(rows.get(key.getE1()), rows.get(key.getE2()), classId);

				if (numeric) {
					
					return cache.getOrCreateNoBlocking(features.getFeatures(), () -> {
						double v = model.applyNumeric(features);
						return v;
					});
				  
				

				} else {
					return cache.getOrCreateNoBlocking(features.getFeatures(), () -> {
						RecordPrediction prediction = model.apply(getFeatureVector(rows.get(key.getE1()), rows.get(key.getE2()), classId));
						if (prediction.getLabel().equals(LearnableRowClusterSimilarities.getLabel(true))) {
							return prediction.getConfidence();
						} else {
							return prediction.getConfidence() * -1.0;
						}
					});
				}
			} 
		}

		@Override
		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public String getName() {
			return classifierName;
		}

		@Override
		public LearnedModel learn(Map<TableRowIdPair, PairLearningData> pairs, ClassId classId, EvaluationMethod evaluationMethod) {
			WekaLearnerCreator creator = new WekaLearnerCreator(classifierName, paramters.clone());
			ConcurrentLinkedQueue<RecordData> recordData = new ConcurrentLinkedQueue<>();
			Map<String, MutableDouble> classCount = new HashMap<>();

			for (Entry<TableRowIdPair, PairLearningData> pair : pairs.entrySet()) {
				CHMHelper.getFromMap(classCount, pair.getValue().getLabel(), () -> new MutableDouble()).add(pair.getValue().getWeight());
			}

			String maxClass = null;
			double max = -1;
			for (Entry<String, MutableDouble> entry : classCount.entrySet()) {
				if (maxClass == null || entry.getValue().intValue() > max) {
					max = entry.getValue().doubleValue();
					maxClass = entry.getKey();
				}
			}
			double maxDouble = (double) max;

			Map<String, Double> replicationMap = new HashMap<>();
			for (Entry<String, MutableDouble> entry : classCount.entrySet()) {
				double weight = maxDouble / entry.getValue().doubleValue();
				if (!weighted) {
					weight = 1.0;
				}
				replicationMap.put(entry.getKey(), weight);
			}

			PersistentParallel6.runAndFinish(pairs.entrySet(), (pair) -> {
				RecordFeature fv = getFeatureVector(pair.getKey().getId1(), pair.getKey().getId2(), classId);
				double recordWeight = pair.getValue().getWeight();
				double classWeight = replicationMap.get(pair.getValue().getLabel());
				
				if (creator.supportsWeights()) {
					RecordData rd = new RecordData(pair.getValue().getLabel(), classWeight * recordWeight, fv);
						recordData.add(rd);
				} else {
					int replication = Ints.checkedCast(Math.round(classWeight * recordWeight));
					for (int i = 0; i < replication; i++) {
						RecordData rd = new RecordData(pair.getValue().getLabel(), fv);
						recordData.add(rd);
					}
				}
			}, this.getClass(), PersistentParallel6.AlreadyParalell.isCurrentThreadParallel());

			return creator.learn(new TrainingData(datasetData, recordData), evaluationMethod);
		}



		@Override
		public RecordFeature getFeatureVector(TableRowId id1, TableRowId id2, ClassId classId) {

			Map<String, Double> map = new HashMap<>();
			List<TableRowId> list = new ArrayList<>();
			list.add(id1);
			list.add(id2);

			ClusterEntryKey key = new ClusterEntryKey(0, 1);
			for (Pair<RowClusterSimilarities, Boolean> sim : sims) {
				List<WeightedScore> simList = sim.getFirst().getList(list, classId, key);

				if (sim.getSecond() && propertyIds != null) {
						if(false) {
							for (WeightedScore ws : simList) {
								if (ws.isMatch()) {
									map.put(sim.getFirst().getName() + "::" + ws.getSuffix().get() + "::MATCH", ws.getWeight());
								} else {
									map.put(sim.getFirst().getName() + "::" + ws.getSuffix().get() + "::NO_MATCH", ws.getWeight());
								}
							}
						}
						else {
											Map<String, MutableDouble> innerSum = new HashMap<>();
											Map<String, MutableDouble> innerCount = new HashMap<>();
						
											for (WeightedScore ws : simList) {
												String innerKey;
												if (ws.isMatch()) {
													innerKey = sim.getFirst().getName() + "::" + ws.getSuffix().get() + "::MATCH";
													// map.put(sim.getName() + "::" +
													// ws.getSuffix().get() + "::MATCH",
													// ws.getWeight());
												} else {
													innerKey = sim.getFirst().getName() + "::" + ws.getSuffix().get() + "::NO_MATCH";
													// map.put(sim.getName() + "::" +
													// ws.getSuffix().get() + "::NO_MATCH",
													// ws.getWeight());
												}
						
												CHMHelper.getFromMap(innerSum, innerKey, () -> new MutableDouble()).add(ws.getWeight());
												CHMHelper.getFromMap(innerCount, innerKey, () -> new MutableDouble()).add(1.0);
						
											}
						
											for (Entry<String, MutableDouble> innerEntry : innerSum.entrySet()) {
												map.put(innerEntry.getKey(), innerEntry.getValue().doubleValue() / innerCount.get(innerEntry.getKey()).doubleValue());
											}
											//
											// for (WeightedScore ws : simList) {
											// if (ws.isMatch()) {
											// map.put(sim.getFirst().getName() + "::" +
											// ws.getSuffix().get() + "::MATCH", ws.getWeight());
											// } else {
											// map.put(sim.getFirst().getName() + "::" +
											// ws.getSuffix().get() + "::NO_MATCH", ws.getWeight());
											// }
											// }
						}
				}
				else {
					double confidence = 0.0;
					double sum = 0.0;
					double count = 0.0;
					for (WeightedScore ws : simList) {
						confidence += ws.getWeight();
						count++;
						sum += ws.getScore();
					}
					// if (confidence > 0.0) {
					if (count > 0.0) {
						map.put(sim.getFirst().getName(), sum / count);
						if (withConfidence && sim.getSecond().booleanValue()) {
							map.put(sim.getFirst().getName() + "::CONFIDENCE", confidence);
						}
					} else {
						map.put(sim.getFirst().getName(), 0.0);
						if (withConfidence && sim.getSecond().booleanValue()) {
							map.put(sim.getFirst().getName() + "::CONFIDENCE", 0.0);
						}
						}

					// } else {
					// map.put(sim.getFirst().getName(), 0.0);
					// if (withConfidence && sim.getSecond().booleanValue()) {
					// map.put(sim.getFirst().getName() + "::CONFIDENCE", 0.0);
					// }
					// }
				}

			}

			for (String feature : this.datasetData.getFeatures().getFeatures().values()) {
				if (!map.containsKey(feature)) {
					map.put(feature, 0.0);
				}
			}
			return new RecordFeature(map, datasetData);
		}

	}

}
