package expansion.set.strategies.clustersplitters.minScoreSplitter;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableDouble;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.DataSetData;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationMethod;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearningFeatures;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordFeature;
import de.uni_mannheim.informatik.dws.expansion.utils.learning.weka.WekaLearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.learning.MT_WekaLearnedModel;
import expansion.set.strategies.clustersplitters.minScoreSplitter.WekaClassifierSimilaritiesCreator.ClassifierBasedSimilarities;

public class WekaClassifierSimilaritiesCreator_MultiTree {

	@Inject
	public WekaClassifierSimilaritiesCreator_MultiTree() {
		super();
		// this.repository = repository;
	}

	public RowClusterSimilarities getSplitter(String classifier, String[] parameters, List<Tri<RowClusterSimilarities, Boolean, Boolean>> sims,
			LearnedModel model,
			OptimizableClass optimizableClass) {

		return new MT_ClassifierBasedSimilarities(classifier, parameters, sims, model, optimizableClass, true, false, null, null, false);

	}

	public MT_ClassifierBasedSimilarities getSplitter(String classifier, String[] parameters, List<Tri<RowClusterSimilarities, Boolean, Boolean>> sims,
			LearnedModel model,
			OptimizableClass optimizableClass,
			boolean numeric) {
		return getSplitter(classifier, parameters, sims, model, optimizableClass, numeric, false, null);
	}

	public MT_ClassifierBasedSimilarities getSplitter(String classifier, String[] parameters, List<Tri<RowClusterSimilarities, Boolean, Boolean>> sims,
			LearnedModel model, OptimizableClass optimizableClass, boolean numeric, Double confidenceThreshold) {
		return getSplitter(classifier, parameters, sims, model, optimizableClass, numeric, false, confidenceThreshold);
	}

	public MT_ClassifierBasedSimilarities getSplitter(String classifier, String[] parameters, List<Tri<RowClusterSimilarities, Boolean, Boolean>> sims,
			LearnedModel model, OptimizableClass optimizableClass, boolean numeric, Double confidenceThreshold, Set<PropertyId> propertyIds) {
		return getSplitter(classifier, parameters, sims, model, optimizableClass, numeric, false, confidenceThreshold, propertyIds);
	}

	public MT_ClassifierBasedSimilarities getSplitter(String classifier, String[] parameters, List<Tri<RowClusterSimilarities, Boolean, Boolean>> sims,
			LearnedModel model, OptimizableClass optimizableClass, boolean numeric, Double confidenceThreshold, Set<PropertyId> propertyIds, boolean weighted) {
		return getSplitter(classifier, parameters, sims, model, optimizableClass, numeric, false, confidenceThreshold, propertyIds, weighted);
	}

	public MT_ClassifierBasedSimilarities getSplitter(String classifier, String[] parameters, List<Tri<RowClusterSimilarities, Boolean, Boolean>> sims,
			LearnedModel model, OptimizableClass optimizableClass, boolean numeric, boolean rangeDescretize) {

		return new MT_ClassifierBasedSimilarities(classifier, parameters, sims, model, optimizableClass, numeric, rangeDescretize, null, null, false);

	}

	public MT_ClassifierBasedSimilarities getSplitter(String classifier, String[] parameters, List<Tri<RowClusterSimilarities, Boolean, Boolean>> sims,
			LearnedModel model, OptimizableClass optimizableClass, boolean numeric, boolean rangeDescretize, Double confidenceThreshold) {

		return new MT_ClassifierBasedSimilarities(classifier, parameters, sims, model, optimizableClass, numeric, rangeDescretize, confidenceThreshold, null,
				false);

	}

	public MT_ClassifierBasedSimilarities getSplitter(String classifier, String[] parameters, List<Tri<RowClusterSimilarities, Boolean, Boolean>> sims,
			LearnedModel model, OptimizableClass optimizableClass, boolean numeric, boolean rangeDescretize, Double confidenceThreshold,
			Set<PropertyId> propertyIds) {

		return new MT_ClassifierBasedSimilarities(classifier, parameters, sims, model, optimizableClass, numeric, rangeDescretize, confidenceThreshold,
				propertyIds, false);

	}

	public MT_ClassifierBasedSimilarities getSplitter(String classifier, String[] parameters, List<Tri<RowClusterSimilarities, Boolean, Boolean>> sims,
			LearnedModel model, OptimizableClass optimizableClass, boolean numeric, boolean rangeDescretize, Double confidenceThreshold,
			Set<PropertyId> propertyIds, boolean weighted) {

		return new MT_ClassifierBasedSimilarities(classifier, parameters, sims, model, optimizableClass, numeric, rangeDescretize, confidenceThreshold,
				propertyIds, weighted);

	}

	public class MT_ClassifierBasedSimilarities extends AbstractAggregatedSimlitarities implements WekaLearnableRowClusterSimilarities {
		private final String classifierName;
		private final DataSetData datasetData;
		private final List<Tri<RowClusterSimilarities, Boolean, Boolean>> sims;
		private final MT_WekaLearnedModel castModel;
		private final OptimizableClass optimizableClass;
		private final boolean rangeDescretized;
		private final Double confidenceThreshold;

		private final Map<List<Boolean>, ClassifierBasedSimilarities> models = new HashMap<>();
		private final Set<PropertyId> propertyIds;
		private final boolean weighted;

		public MT_ClassifierBasedSimilarities(String classifier, String[] parameters, List<Tri<RowClusterSimilarities, Boolean, Boolean>> sims,
				LearnedModel model,
				OptimizableClass optimizableClass,
				boolean numeric, boolean rangeDescretized, Double confidenceThreshold, Set<PropertyId> propertyIds, boolean weighted) {
			super(sims.stream().map((e) -> e.getA()).collect(Collectors.toList()));

			this.weighted = weighted;

			if (propertyIds != null) {
				this.propertyIds = Collections.unmodifiableSet(new HashSet<>(propertyIds));
			} else {
				this.propertyIds = null;
			}
			this.optimizableClass = optimizableClass;
			this.sims = Collections.unmodifiableList(new ArrayList<>(sims));
			this.confidenceThreshold = confidenceThreshold;

			List<String> classLabels = new ArrayList<>();
			classLabels.add(LearnableRowClusterSimilarities.getLabel(true));
			classLabels.add(LearnableRowClusterSimilarities.getLabel(false));

			List<String> featuresList = new ArrayList<>();
			for (Tri<RowClusterSimilarities, Boolean, Boolean> sim : sims) {

				if (sim.getC()) {
					if (propertyIds != null) {
						for (PropertyId propertyId : propertyIds) {
							featuresList.add(sim.getA().getName() + "::" + propertyId.readableString() + "::MATCH");
							featuresList.add(sim.getA().getName() + "::" + propertyId.readableString() + "::NO_MATCH");
						}
					} else {
						featuresList.add(sim.getA().getName());
						featuresList.add(sim.getA().getName() + "::CONFIDENCE");
					}
				}
				else {

					featuresList.add(sim.getA().getName());
				}

			}

			if (numeric) {
				Map<String, Double> labels = new HashMap<>();
				labels.put(LearnableRowClusterSimilarities.getLabel(true), 1.0);
				labels.put(LearnableRowClusterSimilarities.getLabel(false), -1.0);
				this.datasetData = new DataSetData("TEST", labels, new LearningFeatures(featuresList));
			} else {
				this.datasetData = new DataSetData("TEST", classLabels, new LearningFeatures(featuresList));
			}

			this.classifierName = classifier;

			List<List<Boolean>> lists = new ArrayList<>();

			List<Boolean> falseList = new ArrayList<>();
			for (Tri<RowClusterSimilarities, Boolean, Boolean> sim : sims) {
				falseList.add(false);
			}

			lists.add(falseList);

			for (ListEntry<Tri<RowClusterSimilarities, Boolean, Boolean>> sim : ListIterable.iterable(sims)) {
				List<List<Boolean>> additionalList = new ArrayList<>();
				if (sim.getValue().getB().booleanValue()) {

					for (List<Boolean> l : lists) {
						List<Boolean> newList = new ArrayList<>(l);
						newList.set(sim.getIndex(), true);
						additionalList.add(newList);
					}

				}
				if (!additionalList.isEmpty()) {
					lists.addAll(additionalList);
				}
			}

			if (model != null) {
				if (model instanceof MT_WekaLearnedModel) {
					castModel = (MT_WekaLearnedModel) model;
				} else {
					throw new IllegalStateException();
				}
			}
			else {
				castModel = null;
			}

			for (List<Boolean> list : lists) {
				List<Pair<RowClusterSimilarities, Boolean>> innerSims = new ArrayList<>();
				for (ListEntry<Boolean> simEntry : ListIterable.iterable(list)) {
					if (!simEntry.getValue().booleanValue()) {
						innerSims.add(new Pair<>(sims.get(simEntry.getIndex()).getA(), sims.get(simEntry.getIndex()).getC()));
					}
				}
				if (castModel != null) {
					WekaLearnedModel innerModel = castModel.get(list);
					if (innerModel == null) {
						throw new IllegalStateException();
					}
					models.put(list, new ClassifierBasedSimilarities(classifier, parameters.clone(), innerSims, innerModel, numeric, propertyIds, weighted));
				} else {
					models.put(list, new ClassifierBasedSimilarities(classifier, parameters.clone(), innerSims, null, numeric, propertyIds, weighted));
				}
			}

			int sdf = 123;
			this.rangeDescretized = rangeDescretized;
		}

		@Override
		public double get(List<TableRowId> list, ClassId classId, ClusterEntryKey key) {
			if (castModel == null) {
				throw new IllegalStateException();
			} else {
				double total = 0;
				int totalConfidence = 0;
				List<Boolean> listKey = new ArrayList<>();
				for (Tri<RowClusterSimilarities, Boolean, Boolean> simPair : sims) {
					if (!simPair.getC()) {
						total++;
					}
					if (!simPair.getB().booleanValue()) {
						listKey.add(false);
						if (!simPair.getC()) {
							totalConfidence++;
						}
					} else {
						List<WeightedScore> simList = simPair.getA().getList(list, classId, key);
						double confidence = 0.0;
						for (WeightedScore ws : simList) {
							confidence += ws.getWeight();
						}

						if (confidence > 0.0) {
							if (!simPair.getC()) {
								totalConfidence++;
							}
							listKey.add(false);
						} else {
							listKey.add(true);
						}

					}

				}

				double weight = 1.0; // totalConfidence / total;
				double score = weight * this.models.get(listKey).get(list, classId, key);

				Optional<Double> positiveConfidenceThreshold = castModel.getPositiveConfidenceThreshold();

				if (positiveConfidenceThreshold.isPresent() && positiveConfidenceThreshold.get() < 1.0 && score > 0.0) {
					if (score < positiveConfidenceThreshold.get()) {
						score = 0.0;
					}
				}

				Optional<Double> negativeConfidenceThreshold = castModel.getNegativeConfidenceThreshold();

				if (negativeConfidenceThreshold.isPresent() && negativeConfidenceThreshold.get() < 1.0 && score < 0.0) {
					double negativeThreshold = -1.0 * negativeConfidenceThreshold.get();
					if (score > negativeThreshold) {
						score = 0.0;
					}
				}

				if (confidenceThreshold != null) {

					if (Math.abs(score) < confidenceThreshold) {
						score = 0.0;
					}

				}
				if (score > 0.0) {
					int dsf = 123;
				}
				return score;

			}
		}

		@Override
		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getName() {
			return "MT_" + classifierName;
		}

		@Override
		public LearnedModel learn(Map<TableRowIdPair, PairLearningData> pairs, ClassId classId, EvaluationMethod evaluationMethod) {
			Map<List<Boolean>, WekaLearnedModel> learnedModels = new ConcurrentHashMap<>();

			PersistentParallel6.runAndFinish(this.models.entrySet(), (b) -> {
				WekaLearnedModel lm = (WekaLearnedModel) b.getValue().learn(pairs, classId, evaluationMethod);
				learnedModels.put(b.getKey(), lm);
			}, this.getClass());
			return new MT_WekaLearnedModel(learnedModels, optimizableClass);
		}

		@Override
		public RecordFeature getFeatureVector(TableRowId id1, TableRowId id2, ClassId classId) {
			Map<String, Double> map = new HashMap<>();
			List<TableRowId> list = new ArrayList<>();
			list.add(id1);
			list.add(id2);

			ClusterEntryKey key = new ClusterEntryKey(0, 1);
			for (Tri<RowClusterSimilarities, Boolean, Boolean> simPair : sims) {
				RowClusterSimilarities sim = simPair.getA();

				List<WeightedScore> simList = sim.getList(list, classId, key);

				if (simPair.getC() && propertyIds != null) {
					
					if(false) {
						for (WeightedScore ws : simList) {
							if (ws.isMatch()) {
								map.put(sim.getName() + "::" + ws.getSuffix().get() + "::MATCH", ws.getWeight());
							} else {
								map.put(sim.getName() + "::" + ws.getSuffix().get() + "::NO_MATCH", ws.getWeight());
							}
						}
					}
					else {
						Map<String, MutableDouble> innerSum = new HashMap<>();
						Map<String, MutableDouble> innerCount = new HashMap<>();
	
						for (WeightedScore ws : simList) {
							String innerKey;
							if (ws.isMatch()) {
								innerKey = sim.getName() + "::" + ws.getSuffix().get() + "::MATCH";
								// map.put(sim.getName() + "::" +
								// ws.getSuffix().get() + "::MATCH",
								// ws.getWeight());
							} else {
								innerKey = sim.getName() + "::" + ws.getSuffix().get() + "::NO_MATCH";
								// map.put(sim.getName() + "::" +
								// ws.getSuffix().get() + "::NO_MATCH",
								// ws.getWeight());
							}
	
							CHMHelper.getFromMap(innerSum, innerKey, () -> new MutableDouble()).add(ws.getWeight());
							CHMHelper.getFromMap(innerCount, innerKey, () -> new MutableDouble()).add(1.0);
	
						}
	
						for (Entry<String, MutableDouble> innerEntry : innerSum.entrySet()) {
							map.put(innerEntry.getKey(), innerEntry.getValue().doubleValue() / innerCount.get(innerEntry.getKey()).doubleValue());
						}
					
					}
				}
				else {
					double confidence = 0.0;
					double sum = 0.0;
					double count = 0.0;
					for (WeightedScore ws : simList) {
						confidence += ws.getWeight();
						count++;
						sum += ws.getScore();
					}
					if (confidence > 0.0) {
						if (count > 0.0) {
							map.put(sim.getName(), sum / count);
							if (simPair.getC()) {
								map.put(sim.getName() + "::CONFIDENCE", confidence);
							}
						} else {
							map.put(sim.getName(), 0.0);

							if (simPair.getC()) {
								map.put(sim.getName() + "::CONFIDENCE", 0.0);
							}
						}

					} else {
						map.put(sim.getName(), 0.0);
						if (simPair.getC()) {
							map.put(sim.getName() + "::CONFIDENCE", 0.0);
						}
					}
				}

			}
			if (rangeDescretized)
				map = RecordFeature.descretizeMap(map, 8);

			for (String feature : this.datasetData.getFeatures().getFeatures().values()) {
				if (!map.containsKey(feature)) {
					map.put(feature, 0.0);
				}
			}
			return new RecordFeature(map, datasetData);
		}
	}

}
