package expansion.set.strategies.clustersplitters.unaryattributes;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;

public class UnaryAttribute {
	private final ClassId classId;
	private final Set<PropertyId> properties;

	public UnaryAttribute(ClassId classId, PropertyId... properties) {
		super();
		this.classId = classId;
		this.properties = Collections.unmodifiableSet(new HashSet<>(Arrays.asList(properties)));
	}

	public UnaryAttribute(ClassId classId, Set<PropertyId> properties) {
		super();
		this.classId = classId;
		this.properties = Collections.unmodifiableSet(new HashSet<>(properties));
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((classId == null) ? 0 : classId.hashCode());
		result = prime * result + ((properties == null) ? 0 : properties.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		UnaryAttribute other = (UnaryAttribute) obj;
		if (classId == null) {
			if (other.classId != null)
				return false;
		} else if (!classId.equals(other.classId))
			return false;
		if (properties == null) {
			if (other.properties != null)
				return false;
		} else if (!properties.equals(other.properties))
			return false;
		return true;
	}

	public ClassId getClassId() {
		return classId;
	}

	public Set<PropertyId> getProperties() {
		return properties;
	}

	@Override
	public String toString() {
		return "UnaryAttribute [classId=" + classId + ", properties=" + properties + "]";
	}

}
