package expansion.set.strategies.combined;

import java.util.ArrayList;
import java.util.List;

import javax.inject.Inject;

import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusterer;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.clustering.model.similarities.ZeroRowClusterSimilarities;
import expansion.set.strategies.SetExpansionStrategiesRepository;
import expansion.set.strategies.clustersplitters.attributes.AttributeBasedSplitter;
import expansion.set.strategies.clustersplitters.attributes.AttributeBasedSplitter.AttributeMethod;
import expansion.set.strategies.clustersplitters.minScoreSplitter.WeightedAverageScoreSplitterFactory;
import expansion.set.strategies.implicit.ImplicitClustererSimilarityCreator;
import expansion.set.strategies.newstrategies.BagOfWords_Component;
import expansion.set.strategies.rowclusterers.SameTableZeroStrategy;
import expansion.set.strategies.rowclusterers.label.LabelRowSimilaritiesCreator;
import expansion.set.strategies.rowclusterers.strategies.association.AssociationSimilarities;
import expansion.set.strategies.rowclusterers.strategies.association.AssociationSimilarities.AssociationType;
import expansion.set.strategies.rowclusterers.strategies.overlap.RowToIndexMapper.MappingType;

public class Combined_BASE implements RowClusterer{

	private final OptimizableClass optimizingClazz = SetExpansionStrategiesRepository.COMBINED_BASE.getOptimizableClassCluster();




	@Inject
	protected   WeightedAverageScoreSplitterFactory splitterFactory;
	@Inject
	protected   LabelRowSimilaritiesCreator labelRowSimilaritiesCreator; 
	 

	@Inject
	protected   AttributeBasedSplitter attributeSplitter;
	 
	
	@Inject
	BagOfWords_Component component;
	
	@Inject
	protected ImplicitClustererSimilarityCreator implicitCreator;
	@Inject
	protected AssociationSimilarities association;
 
	public Combined_BASE(  ) {
	 
	}
 
	public RowClusterSimilarities clusterRows(RowClusteringConfig config, ClustererInput input ) {
//		ClusterList clusters = clusterer.clusterRows(config, input);
		 
		RowClusterSimilarities attributeScores = attributeSplitter.get(config, input, AttributeMethod.DIRECT_FULL);
		RowClusterSimilarities labelScores = labelRowSimilaritiesCreator.createLucene(input.getFullCandidates(), config.isWithCache());
		RowClusterSimilarities bowScores = component.getBagOfWords_for_RowWithoutTFIDF(input, false, config.isWithCache());
		// RowClusterSimilarities tableBowScores =
		// component.getBagOfWords_for_TableWithoutTFIDF(input, false);
		 
		RowClusterSimilarities phiScores = association.getScores(config, input, MappingType.LABEL_ONLY, AssociationType.PHI);

		RowClusterSimilarities empty = new ZeroRowClusterSimilarities();
		RowClusterSimilarities implicit = implicitCreator.create(input, config);

		RowClusterSimilarities sameTableSim = new SameTableZeroStrategy();
		// RowClusterSimilarities pmiScores = association.getScores(config,
		// input, MappingType.LABEL_ONLY, AssociationType.PMI);
		// RowClusterSimilarities tfidfScores = association.getScores(config,
		// input, MappingType.LABEL_ONLY, AssociationType.TF_IDF);
		
		// double phiWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("PHI",
		// 1.0));
		// double pmiWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("PMI",
		// 1.0));
		// double labelWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("LABEL",
		// 1.0));
		// double attributeWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("ATTRIBUTE",
		// 1.0));
		// double bowWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("BOW",
		// 1.0));
		// double tableBowWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("TABLE_BOW",
		// 1.0));



		List<RowClusterSimilarities> sims = new ArrayList<>();
		sims.add(attributeScores);
		sims.add(labelScores);
		sims.add(bowScores);
		// sims.add(tableBowScores);
		sims.add(implicit);
		sims.add(phiScores);
		sims.add(sameTableSim);
		
		
		return splitterFactory.getSplitter(sims, config.getWeightsRepository(), optimizingClazz, false);
	}
	
 
}
