package expansion.set.strategies.connectors;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.similarities.RowClusterConstantSimilarities;
import expansion.set.strategies.rowclusterers.unary.ClusterCombiner;

public class ClusterConnector {

	private final AnnotatedTablesRepository tableRepository;

	@Inject
	public ClusterConnector(AnnotatedTablesRepository tableRepository) {
		this.tableRepository = tableRepository;
	}

	public ClusterList connect(ClusterList a, ClusterList b) {
		ClusterCombiner<TableRowId> cc = new ClusterCombiner<>();

		List<Set<TableRowId>> aList = a.getClusterList().stream().map((e) -> new HashSet<>(e.getRows())).collect(Collectors.toList());
		List<Set<TableRowId>> bList = b.getClusterList().stream().map((e) -> new HashSet<>(e.getRows())).collect(Collectors.toList());

		cc.add(aList);
		cc.add(bList);
		int size = 0;
		ArrayList<RowCluster> ret = new ArrayList<>();
		for (Set<TableRowId> cluster : cc.get()) {
			ClassId classId = null;
			for (TableRowId rowId : cluster) {
				classId = tableRepository.getClassId(rowId.getTableId());
				break;
			}
			ret.add(new RowCluster(classId, new ArrayList<>(cluster), new RowClusterConstantSimilarities(1.0)));
			size += cluster.size();
		}

		System.out.println("Total number of rows in cluster: " + size);
		return new ClusterList(ret);
	}
}
