package expansion.set.strategies.evaluatemethod;

import java.util.ArrayList;
import java.util.List;
import java.util.function.Function;

import javax.inject.Inject;

import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusterer;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.strategies.clustersplitters.attributes.AttributeBasedSplitter;
import expansion.set.strategies.clustersplitters.attributes.AttributeBasedSplitter.AttributeMethod;
import expansion.set.strategies.clustersplitters.minScoreSplitter.CombinedRowClusterSimilaritiesCreator;
import expansion.set.strategies.clustersplitters.minScoreSplitter.WeightedAverageScoreSplitterFactory;
import expansion.set.strategies.clustersplitters.minScoreSplitter.WekaClassifierSimilaritiesCreator_MultiTree;
import expansion.set.strategies.ea.ParamterClusterer;
import expansion.set.strategies.implicit.ImplicitClustererSimilarityCreator;
import expansion.set.strategies.newstrategies.BagOfWords_Component;
import expansion.set.strategies.rowclusterers.SameTableZeroStrategy;
import expansion.set.strategies.rowclusterers.label.LabelRowSimilaritiesCreator;
import expansion.set.strategies.rowclusterers.strategies.association.AssociationSimilarities;
import expansion.set.strategies.rowclusterers.strategies.association.AssociationSimilarities.AssociationType;
import expansion.set.strategies.rowclusterers.strategies.overlap.RowToIndexMapper.MappingType;

public class CLUSTERING_METHOD_CREATOR {

	public final static String[] parameters = new String[] { "-U", "-attribute-importance",
			"-store-out-of-bag-predictions", "-I", "4000", "-num-slots", "60", "-K", "2", "-V", "1e-3", "-P", "85" };//

	// public final static String[] parameters = new String[] { "-U",
	// "-attribute-importance", "-store-out-of-bag-predictions", "-I", "2000",
	// "-num-slots", "60",
	// "-K", "2", "-V", "1e-3", "-P", "66" };//
	//
	// private final OptimizableClass optimizingClazz =
	// SetExpansionStrategiesRepository.COMBINED_RandomForest_ET.getOptimizableClassCluster();

	@Inject
	protected WekaClassifierSimilaritiesCreator_MultiTree splitterFactory;
	@Inject
	protected LabelRowSimilaritiesCreator labelRowSimilaritiesCreator;

	@Inject
	protected AttributeBasedSplitter attributeSplitter;

	@Inject
	BagOfWords_Component component;

	@Inject
	protected AssociationSimilarities association;

	@Inject
	protected ImplicitClustererSimilarityCreator implicitCreator;

	@Inject
	protected WeightedAverageScoreSplitterFactory weightedAverageSplitterFactory;

	@Inject
	protected CombinedRowClusterSimilaritiesCreator combinedSplitter;

	public CLUSTERING_METHOD_CREATOR() {

	}

	public RowClusterSimilarities clusterRows(RowClusteringConfig config, ClustererInput input, OptimizableClass optimizingClazz,
			CLUSTERING_SIMS_AGGREGATOR aggregator, CLUSTERING_SIMS_LIST simList, String[] parameters) {
		// ClusterList clusters = clusterer.clusterRows(config, input);

		// RowClusterSimilarities attributeScores =
		// attributeSplitter.get(config, input, AttributeMethod.DIRECT_FULL);
		RowClusterSimilarities attributeScores = attributeSplitter.get(config, input, AttributeMethod.DIRECT_FULL);

		RowClusterSimilarities labelScores = labelRowSimilaritiesCreator.createLucene(input.getFullCandidates(), config.isWithCache(), true);
		RowClusterSimilarities overlapLabel = labelRowSimilaritiesCreator.createOverlap(input.getFullCandidates(), config.isWithCache());
		RowClusterSimilarities surfaceForms = labelRowSimilaritiesCreator.createSurfaceForms(input.getFullCandidates(), config.isWithCache());

		RowClusterSimilarities bowScores = component.getBagOfWords_for_RowWithoutTFIDF(input, false, config.isWithCache());
		// RowClusterSimilarities tableBowScores =
		// component.getBagOfWords_for_TableWithoutTFIDF(input, false);

		RowClusterSimilarities phiScores = association.getScores(config, input, MappingType.LABEL_ONLY, AssociationType.PHI);

		RowClusterSimilarities implicit = implicitCreator.create(input, config);

		RowClusterSimilarities implicitWithDirect = implicitCreator.createWithDirectCompared(input, config);

		RowClusterSimilarities sameTableSim = new SameTableZeroStrategy();
		// RowClusterSimilarities pmiScores = association.getScores(config,
		// input, MappingType.LABEL_ONLY, AssociationType.PMI);
		// RowClusterSimilarities tfidfScores = association.getScores(config,
		// input, MappingType.LABEL_ONLY, AssociationType.TF_IDF);

		// double phiWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("PHI",
		// 1.0));
		// double pmiWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("PMI",
		// 1.0));
		// double labelWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("LABEL",
		// 1.0));
		// double attributeWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("ATTRIBUTE",
		// 1.0));
		// double bowWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("BOW",
		// 1.0));
		// double tableBowWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("TABLE_BOW",
		// 1.0));

		List<RowClusterSimilarities> simsSimple = new ArrayList<>();
		List<Tri<RowClusterSimilarities, Boolean, Boolean>> sims = new ArrayList<>();

		for (CLUSTERING_SIM sim : simList.getSims()) {

			switch (sim) {
			case ATTRIBUTE:
				simsSimple.add(attributeScores);
				sims.add(new Tri<>(attributeScores, false, true));
				break;
			case BOW:
				simsSimple.add(bowScores);
				sims.add(new Tri<>(bowScores, false, false));
				break;
			case IMPLICIT:
				simsSimple.add(implicitWithDirect);
				sims.add(new Tri<>(implicitWithDirect, false, true));
				break;
			case LABEL:
				simsSimple.add(labelScores);
				sims.add(new Tri<>(labelScores, false, false));
				break;
			case LABELOVERLAP:
				simsSimple.add(overlapLabel);
				sims.add(new Tri<>(overlapLabel, false, false));
				break;
			case LABELSF:
				simsSimple.add(surfaceForms);
				sims.add(new Tri<>(surfaceForms, false, false));
				break;
			case PHI:
				simsSimple.add(phiScores);
				sims.add(new Tri<>(phiScores, false, false));

				break;
			case SAME_TABLE:
				simsSimple.add(sameTableSim);
				sims.add(new Tri<>(sameTableSim, false, false));

				break;
			default:
				throw new IllegalStateException();

			}

		}
		// sims.add(tableBowScores);

		// sims.add(new Tri<>(implicitWithDirect, false, true));
		// sims.add(tableBowScores);
		// sims.add(sameTableSim);
		// sims.add(empty);
		// sims.add(pmiScores);
		// sims.add( new Pair<>(tfidfScores, 0.0) );

		LearnedModel learnedModel = null;
		if (config.getWeightsRepository() != null) {
			learnedModel = config.getWeightsRepository().get(optimizingClazz);
		}

		switch (aggregator) {
		case AGNOSTIC_TREE:
			return splitterFactory.getSplitter("RandomForest", parameters, sims, learnedModel, optimizingClazz, true, null, null);// input.getNonTailProperties().get(input.getClassId()));

		case EXTENDED_TREE:
			return splitterFactory.getSplitter("RandomForest", parameters, sims, learnedModel, optimizingClazz, true, null,
					input.getNonTailProperties().get(input.getClassId()));

		case WEIGHTED_AVERAGE:
			return weightedAverageSplitterFactory.getSplitter(simsSimple, config.getWeightsRepository(), optimizingClazz, false);

		case EXTENDED_TREE_AND_WEIGHTED_AVERAGE:

			Function<LearnedModel, LearnableRowClusterSimilarities> rfFunction = (model) -> {
				return splitterFactory.getSplitter("RandomForest", parameters, sims, model, optimizingClazz, true, null,
						input.getNonTailProperties().get(input.getClassId()));
			};

			Function<LearnedModel, LearnableRowClusterSimilarities> baseFunction = (model) -> {
				return weightedAverageSplitterFactory.getSplitter(simsSimple, model, optimizingClazz, true);
			};

			return combinedSplitter.get(optimizingClazz.getName(), optimizingClazz, simsSimple, rfFunction, baseFunction, learnedModel);

		default:
			throw new IllegalStateException();

		}

		// if (KEEP_ET) {
		// return splitterFactory.getSplitter("RandomForest", parameters, sims,
		// learnedModel, optimizingClazz, true, null,
		// input.getNonTailProperties().get(input.getClassId()));
		// } else {
		// return splitterFactory.getSplitter("RandomForest", parameters, sims,
		// learnedModel, optimizingClazz, true, null, null);//
		// input.getNonTailProperties().get(input.getClassId()));
		// }
	}

	public Method_Creator getCreator(RowClusteringConfig config, ClustererInput input, OptimizableClass optimizingClazz, CLUSTERING_SIMS_AGGREGATOR aggregator,
			CLUSTERING_SIMS_LIST simList) {
		return new Method_Creator(config, input, optimizingClazz, aggregator, simList);
	}

	public class Method_Creator implements ParamterClusterer {
		private final RowClusteringConfig config;
		private final ClustererInput input;
		private final CLUSTERING_SIMS_AGGREGATOR aggregator;
		private final CLUSTERING_SIMS_LIST simList;
		private final OptimizableClass optimizingClazz;

		public Method_Creator(RowClusteringConfig config, ClustererInput input, OptimizableClass optimizingClazz, CLUSTERING_SIMS_AGGREGATOR aggregator,
				CLUSTERING_SIMS_LIST simList) {
			this.config = config;
			this.input = input;
			this.aggregator = aggregator;
			this.simList = simList;
			this.optimizingClazz = optimizingClazz;

		}

		@Override
		public LearnableRowClusterSimilarities clusterRows(String[] parameters) {
			return ((LearnableRowClusterSimilarities) CLUSTERING_METHOD_CREATOR.this.clusterRows(config, input, optimizingClazz, aggregator, simList,
					parameters));
		}

	}

	public RowClusterer getRowClusterer(OptimizableClass clazz, CLUSTERING_SIMS_LIST simsList,
			CLUSTERING_SIMS_AGGREGATOR aggregator) {
		return new RowClusterer() {

			@Override
			public RowClusterSimilarities clusterRows(RowClusteringConfig config, ClustererInput input) {
				return CLUSTERING_METHOD_CREATOR.this.clusterRows(config, input, clazz, aggregator, simsList,
						parameters);
			}
		};
	}
}
