package expansion.set.strategies.newstrategies;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import com.google.common.collect.ImmutableTable;
import com.google.common.collect.Table.Cell;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableIdPair;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdsSet;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.similarity.stopwords.StopWordsRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.WeightedScore;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
 
public class BagOfWords_Component {

	@Inject
	AnnotatedTablesRepository tableRepository;	
	
	@Inject
	protected StopWordsRepository stopWordsRepository;
	
	public <E> Map<String, Double> getTfIdf(Map<E, Map<String, Integer>> bagOfWords) {
		double N = bagOfWords.size();
		
		Map<String, Set<E>> dfs = new HashMap<>();
		
		for(Entry<E, Map<String, Integer>> entry : bagOfWords.entrySet()) {
			for(String s : entry.getValue().keySet()) {
				CHMHelper.getFromMap(dfs, s, () -> new HashSet<>()).add(entry.getKey());
			}
		}
		Map<String, Double> idf = dfs.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			return Math.log( N / e.getValue().size());
		}));

		return idf;
	}
	
	
	public <E> Map<String, Double> getSimple(Map<E, Map<String, Integer>> bagOfWords) {
	 
		
		Set<String > dfs = new HashSet<>();
		
		for(Entry<E, Map<String, Integer>> entry : bagOfWords.entrySet()) {
			for(String s : entry.getValue().keySet()) {
				dfs.add(s);
			}
		}
		Map<String, Double> idf = dfs.stream().collect(Collectors.toMap(Function.identity(), (e) -> {
			return 1.0;
		}));

		return idf;
	}
	
	public RowClusterSimilarities getBagOfWords_for_TableWithoutTFIDF(ClustererInput input, boolean removeStopWords) {
		return getBagOfWords_for_Table(input, false, removeStopWords, "BOW_TABLE" + (removeStopWords ? "_WITH_STOPWORDS_REMOVAL" : ""));
	}

	public RowClusterSimilarities getBagOfWords_for_Table(ClustererInput input, boolean removeStopWords) {
		return getBagOfWords_for_Table(input, true, removeStopWords, "BOW_TABLE_WITH_TF_IDF" + (removeStopWords ? "_WITH_STOPWORDS_REMOVAL" : ""));
	}
	
	
	public Map<AnnotatedTableId, Map<String, Integer>> getBagOfWordsByTable(TableRowIdsSet input, boolean removeStopWords) {
		
		Set<AnnotatedTableId> tableIds = new HashSet<>();
		
		for (TableRowId rowId : input.getRowIds()) {
			tableIds.add(rowId.getTableId());
		}
		
		ConcurrentHashMap<AnnotatedTableId, Map<String, Integer>> ret = new ConcurrentHashMap<>();
		 
		PersistentParallel6.runAndFinish(tableIds, (tableId) -> {
			Map<String, MutableInt>bag = new HashMap<>();
			AnnotatedTable table = tableRepository.getTable(tableId);
			
			ImmutableTable<RowIdx, ColumnIdx, String> stringTable = table.getStringTable();
			for (Cell<RowIdx, ColumnIdx, String> cell : stringTable.cellSet()) {
				String value = cell.getValue();
				if(value != null) {
					value = StringNormalizer.normaliseValue(value, false);
					List<String> tokens = StringNormalizer.tokenise(value, false, false);
					for(String token : tokens) {
						if(!removeStopWords || !stopWordsRepository.isStopWord(token)) {
							CHMHelper.getFromMap(bag, token, () -> new MutableInt()).increment();
						}
					}
				}
			}
			
			ret.put(tableId, bag.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().intValue())));
		}, this.getClass());
		return ret;
	}
	
	public RowClusterSimilarities getBagOfWords_for_Row(ClustererInput input, boolean removeStopWords) {
		
		return getBagOfWords_for_Row(input, true, removeStopWords, "BOW_WITH_TF_IDF" + (removeStopWords ? "_WITH_STOPWORDS_REMOVAL" : ""));
	}

	public RowClusterSimilarities getBagOfWords_for_RowWithoutTFIDF(ClustererInput input, boolean removeStopWords, boolean isWithCache) {
		
		return getBagOfWords_for_Row(input, false, removeStopWords, true, "BOW" + (removeStopWords ? "_WITH_STOPWORDS_REMOVAL" : ""), isWithCache);
	}

	public RowClusterSimilarities getBagOfWords_for_RowWithoutTFIDF_Without_Label(ClustererInput input, boolean removeStopWords, boolean isWithCache) {

		return getBagOfWords_for_Row(input, false, removeStopWords, false, "BOW_NO_LABEL" + (removeStopWords ? "_WITH_STOPWORDS_REMOVAL" : ""), isWithCache);
	}

	public Map<TableRowId, Map<String, Integer>> getBagOfWords(TableRowIdsSet input, boolean removeStopWords, boolean withLabel) {
		return getBagOfWords(input, removeStopWords, withLabel, false);
	}

	public Map<String, Integer> getBagOfWords(TableRowId tableRowId, boolean removeStopWords, boolean withLabel) {
		Map<String, MutableInt> bag = new HashMap<>();
		AnnotatedTable table = tableRepository.getTable(tableRowId.getTableId());
		ImmutableTable<RowIdx, ColumnIdx, String> stringTable = table.getStringTable();
		for (ColumnIdx colIdx : table.getColumns()) {
			if (!table.getKeyColumnIndex().equals(colIdx) || withLabel) {
				String value = stringTable.get(tableRowId.getRowIdx(), colIdx);
				if (value != null) {
					value = StringNormalizer.normaliseValue(value, false);
					List<String> tokens = StringNormalizer.tokenise(value, false, false);
					for (String token : tokens) {
						if (!removeStopWords || !stopWordsRepository.isStopWord(token)) {
							if (!token.equals("null") && token.length() > 1) {
								CHMHelper.getFromMap(bag, token, () -> new MutableInt()).increment();
							}
						}
					}
				}
			}

		}

		return bag.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().intValue()));
	}

	public Map<TableRowId, Map<String, Integer>> getBagOfWords(TableRowIdsSet input, boolean removeStopWords, boolean withLabel, boolean sequential) {
		ConcurrentHashMap<TableRowId, Map<String, Integer>> ret = new ConcurrentHashMap<>();
		 
		// Set<TableRowId> rowIds = new HashSet<>();
		// rowIds.add(new
		// TableRowId("1346876860791_1346879917084_1215.arc8961897601177170488#95474902_0_2342902186035127911",
		// 27));
		// rowIds.add(new
		// TableRowId("1346981172186_1346994903857_1044.arc4636266937237321018#46166619_0_591080969917940788",
		// 28));

		PersistentParallel6.runAndFinish(input.getRowIds(), (tableRowId) -> {
			// if (rowIds.contains(tableRowId)) {
			// int sfd = 123;
			// }
			ret.put(tableRowId, getBagOfWords(tableRowId, removeStopWords, withLabel));
		}, this.getClass(), sequential);
		return ret;
	}

	private double getScore(String string, Map<String, Double> idfScores, Double defaultScore) {
		if(idfScores != null) {
			Double score = idfScores.get(string); 
			if(score == null  ) {
				return defaultScore;
			}
			else {
				return score;
			}
				
		}
		else {
			return defaultScore;
		}
	}

	public double score(Map<String, Integer> ba1, Map<String, Integer> ba2, Double defaultScore, Map<String, Double> idfScores, boolean negativeScores) {
		return scoreFull(ba1, ba2, defaultScore, idfScores, negativeScores, false).getScore();

	}

	public WeightedScore scoreFull(Map<String, Integer> ba1, Map<String, Integer> ba2, Double defaultScore, Map<String, Double> idfScores,
			boolean negativeScores, boolean confidenceBased) {
		if(idfScores == null && defaultScore == null) {
			throw new IllegalStateException();
		}
		// Set<String> keys = new HashSet<>();
		// keys.addAll(ba1.keySet());
		// keys.addAll(ba2.keySet());

		Set<String> bigger;
		Set<String> smaller;
		if (ba1.size() > ba2.size()) {
			bigger = ba1.keySet();
			smaller = ba2.keySet();
		} else {
			smaller = ba1.keySet();
			bigger = ba2.keySet();
		}
		
		int overlap = 0;
		double ret = 0.0;
		double denominator = 0.0;
		
		for (String token : bigger) {
			double idfScore = getScore(token, idfScores, defaultScore);
			double s1 = 0.0;
			double s2 = 0.0;
			if(negativeScores) {
				s1 = -1.0;
				s2 = -1.0;
			}
			boolean firstHasToken = false;
			if(ba1.containsKey(token)) {
				s1 = ba1.get(token) * idfScore;
				firstHasToken = true;
			}
			if(ba2.containsKey(token)) {
				s2 = ba2.get(token) * idfScore;
				if (firstHasToken) {
					overlap++;
				}
			}
			ret += s1  * s2 ;
			

		}
		
		for (String token : smaller) {
			if (!bigger.contains(token)) {
				double idfScore = getScore(token, idfScores, defaultScore);
				double s1 = 0.0;
				double s2 = 0.0;
				if (negativeScores) {
					s1 = -1.0;
					s2 = -1.0;
				}
				if (ba1.containsKey(token)) {
					s1 = ba1.get(token) * idfScore;
				}
				if (ba2.containsKey(token)) {
					s2 = ba2.get(token) * idfScore;
				}
				ret += s1 * s2;
			}
		}

		
		if (false && confidenceBased) {
			if (overlap == 0) {
				return new WeightedScore(0.0, 1.0);
			} else {
				return new WeightedScore(Math.min(1.0, overlap * 0.20), 1.0);
			}
		} else {
			double e1 = 0.0;
			for (Entry<String, Integer> entry : ba1.entrySet()) {
				e1 += Math.pow(entry.getValue() * getScore(entry.getKey(), idfScores, defaultScore), 2.0);
			}
			double e2 = 0.0;
			for (Entry<String, Integer> entry : ba2.entrySet()) {
				e2 += Math.pow(entry.getValue() * getScore(entry.getKey(), idfScores, defaultScore), 2.0);
			}
			denominator = Math.sqrt(e1) * Math.sqrt(e2);

			if (denominator < 0.001 && denominator > -0.001) {
				return new WeightedScore(0.0, 0.0);
			}

			double score = ret / denominator;

			if (score < 0.0) {
				int sdf = 123;
			}

			return new WeightedScore(score, 1.0);
		}
		

	}

	public static SimpleLockableConcurrentCache<Pair<UuidStringId, String>, RowClusterSimilarities> bagOfWordsCache = new SimpleLockableConcurrentCache<>();

	public Map<TableRowId, Map<String, Integer>> getBowMap(ClustererInput input, boolean tfidf, boolean removeStopWords) {
		return getBagOfWords(input.getFullCandidates(), removeStopWords, true);

	}
	private RowClusterSimilarities getBagOfWords_for_Row(ClustererInput input, boolean tfidf, boolean removeStopWords, String name) {
		return getBagOfWords_for_Row(input, tfidf, removeStopWords, true, name, true);
	}

	private RowClusterSimilarities getBagOfWords_for_Row(ClustererInput input, boolean tfidf, boolean removeStopWords, boolean withLabel, String name,
			boolean withCache) {
		
		// UuidStringId stringId;
		//
		// if (false) {
		// stringId = input.getFullCandidates().getId();
		//
		// } else {
		// stringId = new UuidStringId("ALL");
		// }
		return bagOfWordsCache.getOrCreate(new Pair<>(input.getFullCandidatesAllClasses().getId(), name), () -> {

			// TableRowIdsSet rowIdsSet = input.getFullCandidates();
			// Set<TableRowId> rowIds = new HashSet<>();
			//
			// for (ClassId classId : input.getAllParents().keySet()) {
			// rowIds.addAll(tableRepository.getAllRowIds(classId));
			//
			// }

			// rowIdsSet = new TableRowIdsSet(rowIds, stringId);

			Map<TableRowId, Map<String, Integer>> bagOfWords = getBagOfWords(input.getFullCandidatesAllClasses(), removeStopWords, withLabel);
			Map<String, Double> idfScores;

			if (tfidf) {
				idfScores = getTfIdf(bagOfWords);
			} else {
				idfScores = getSimple(bagOfWords);
			}
			if (withCache)
				return new BagOfWords_Similarities(bagOfWords, idfScores, name, !withLabel);
			else
				return new BagOfWords_Similarities_no_Cache(bagOfWords, idfScores, name, !withLabel);
		});

	}
	
	
	public static SimpleLockableConcurrentCache<UuidStringId, RowClusterSimilarities> bagOfWordsTableCache = new SimpleLockableConcurrentCache<>();

	public RowClusterSimilarities getBagOfWords_for_Table(ClustererInput input, boolean tfidf, boolean removeStopWords, String name) {

		return bagOfWordsTableCache.getOrCreate(input.getFullCandidates().getId(), () -> {
			Map<AnnotatedTableId, Map<String, Integer>> bagOfWords = getBagOfWordsByTable(input.getFullCandidates(), removeStopWords);
			Map<String, Double> idfScores;

			if (tfidf) {
				idfScores = getTfIdf(bagOfWords);
			} else {
				idfScores = getSimple(bagOfWords);
			}

			return new BagOfWords_Table_Similarities(bagOfWords, idfScores, name);
		});
	}

	private class BagOfWords_Similarities_no_Cache implements RowClusterSimilarities {

		private final Map<TableRowId, Map<String, Integer>> bagOfWords;
		private final Map<String, Double> idfScores;
		private final SimpleLockableConcurrentCache<TableRowIdPair, Double> cache = new SimpleLockableConcurrentCache<>();
		private final String name;
		private final boolean confidenceBased;

		public BagOfWords_Similarities_no_Cache(Map<TableRowId, Map<String, Integer>> bagOfWords, Map<String, Double> idfScores, String name,
				boolean confidenceBased) {
			super();
			this.bagOfWords = bagOfWords;
			this.idfScores = idfScores;
			this.name = name;
			this.confidenceBased = confidenceBased;
		}

		@Override
		public List<WeightedScore> getList(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			List<WeightedScore> list = new ArrayList<>();

			Map<String, Integer> ba1 = bagOfWords.get(rows.get(key.getE1()));
			Map<String, Integer> ba2 = bagOfWords.get(rows.get(key.getE2()));

			list.add(scoreFull(ba1, ba2, null, idfScores, false, confidenceBased));

			return list;
		}

		@Override
		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {

			TableRowIdPair pair = new TableRowIdPair(rows.get(key.getE1()), rows.get(key.getE2()));

			Map<String, Integer> ba1 = bagOfWords.get(rows.get(key.getE1()));
			Map<String, Integer> ba2 = bagOfWords.get(rows.get(key.getE2()));

			return scoreFull(ba1, ba2, null, idfScores, false, confidenceBased).getScore();

		}

		@Override
		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getName() {
			return name;
		}

	}

	   private class BagOfWords_Similarities implements RowClusterSimilarities {
		
		private final Map<TableRowId, Map<String, Integer>> bagOfWords;
		private final Map<String, Double> idfScores;
		private final SimpleLockableConcurrentCache<TableRowIdPair, WeightedScore> cache = new SimpleLockableConcurrentCache<>();
		private final String name;
		private final boolean confidenceBased;

		public BagOfWords_Similarities(Map<TableRowId, Map<String, Integer>> bagOfWords, Map<String, Double> idfScores, String name, boolean confidenceBased) {
			super();
			this.bagOfWords = bagOfWords;
			this.idfScores = idfScores;
			this.name = name;
			this.confidenceBased = confidenceBased;
		}

		@Override
		public List<WeightedScore> getList(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			List<WeightedScore> list = new ArrayList<>();

			TableRowIdPair pair = new TableRowIdPair(rows.get(key.getE1()), rows.get(key.getE2()));

			list.add(cache.getOrCreateNoBlocking(pair, () -> {

				if (confidenceBased) {
					if (pair.getId1().equals(new TableRowId("1346876860789_1346944484018_327.arc2286104877941280553#57060121_1_7502732949961164628", 26))
							&& pair.getId2()
									.equals(new TableRowId("1350433107058_1350501668478_684.arc3249480116582433297#17146756_1_8797054154240939108", 10))) {
						int sdf = 13;
					}

					if (pair.getId2().equals(new TableRowId("1346876860789_1346944484018_327.arc2286104877941280553#57060121_1_7502732949961164628", 26))
							&& pair.getId1()
									.equals(new TableRowId("1350433107058_1350501668478_684.arc3249480116582433297#17146756_1_8797054154240939108", 10))) {
						int sdf = 13;
					}
				}

				Map<String, Integer> ba1 = bagOfWords.get(rows.get(key.getE1()));
				Map<String, Integer> ba2 = bagOfWords.get(rows.get(key.getE2()));

				return scoreFull(ba1, ba2, null, idfScores, false, confidenceBased);
			}));

			return list;
		}

		@Override
		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			
			TableRowIdPair pair = new TableRowIdPair(rows.get(key.getE1()), rows.get(key.getE2()));

			return cache.getOrCreateNoBlocking(pair, () -> {

				if (confidenceBased) {
					if (pair.getId1().equals(new TableRowId("1346876860789_1346944484018_327.arc2286104877941280553#57060121_1_7502732949961164628", 26))
							&& pair.getId2()
									.equals(new TableRowId("1350433107058_1350501668478_684.arc3249480116582433297#17146756_1_8797054154240939108", 10))) {
						int sdf = 13;
					}

					if (pair.getId2().equals(new TableRowId("1346876860789_1346944484018_327.arc2286104877941280553#57060121_1_7502732949961164628", 26))
							&& pair.getId1()
									.equals(new TableRowId("1350433107058_1350501668478_684.arc3249480116582433297#17146756_1_8797054154240939108", 10))) {
						int sdf = 13;
					}
				}

				Map<String, Integer> ba1 = bagOfWords.get(rows.get(key.getE1()));
				Map<String, Integer> ba2 = bagOfWords.get(rows.get(key.getE2()));

				return scoreFull(ba1, ba2, null, idfScores, false, confidenceBased);
			}).getScore();

		}

		@Override
		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getName() {
			return name;
		}
		
	}
	   
	   
	   private   class BagOfWords_Table_Similarities implements RowClusterSimilarities {
			
			private final Map<AnnotatedTableId, Map<String, Integer>> bagOfWords;
			private final Map<String, Double> idfScores;
		private final SimpleLockableConcurrentCache<TableIdPair, Double> cache = new SimpleLockableConcurrentCache<>();
		private final String name;


		public BagOfWords_Table_Similarities(Map<AnnotatedTableId, Map<String, Integer>> bagOfWords, Map<String, Double> idfScores, String name) {
				super();
				this.bagOfWords = bagOfWords;
				this.idfScores = idfScores;
			this.name = name;
			}

			@Override
			public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
				AnnotatedTableId tableId1 = rows.get(key.getE1()).getTableId();
				AnnotatedTableId tableId2 = rows.get(key.getE2()).getTableId();
				if (tableId1.equals(tableId2))
					return 0.0;

			TableIdPair pair = new TableIdPair(tableId1, tableId2);

			return cache.getOrCreateNoBlocking(pair, () -> {

				Map<String, Integer> ba1 = bagOfWords.get(tableId1);
				Map<String, Integer> ba2 = bagOfWords.get(tableId2);

				double sim = score(ba1, ba2, null, idfScores, false);
				if (sim > 0) {
					int sdf = 123;
				}
				return sim;
			});

			}

			@Override
			public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
				throw new UnsupportedOperationException();
			}
			
		@Override
		public String getName() {
			return name;
		}
		}



}


