package expansion.set.strategies.rowclusterers;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.mutable.MutableInt;

import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;

public class ComponentCreator {

	 @Inject
	public ComponentCreator() {
	
	}
	
	public Map<String, Set<String>> create(String a, String b) {
		
		LinkedHashMap<String, Integer> aSet = new LinkedHashMap<>();
		
		List<String> listA = Arrays.asList( StringNormalizer.normaliseValue(a, false).split(" "));
 
		
		for(ListEntry<String> e : ListIterable.iterable(listA)) {
			aSet.put(e.getValue(), e.getIndex());
		}
		

		LinkedHashMap<String, Integer> bSet = new LinkedHashMap<>();
		
		List<String> listB = Arrays.asList( StringNormalizer.normaliseValue(b, false).split(" "));
		for(ListEntry<String> e : ListIterable.iterable(listB)) {
			bSet.put(e.getValue(), e.getIndex());
		}
 
		
		
		
		if(aSet.size() < bSet.size()) {
			return getComponents(bSet, aSet);
		}
		else {
			return getComponents(aSet, bSet);
		}
		
		
	}

	private Map<String,Set<String>> getComponents(LinkedHashMap<String, Integer> largerSet, LinkedHashMap<String, Integer> smallerSet) {
		Map<String, Set<String>> ret = new HashMap<>();
		List<String> currentComponent = new ArrayList<>();
		int currentIndex = -1;
		for(Entry<String, Integer> s :  largerSet.entrySet() ) {		
			if(smallerSet.containsKey(s.getKey())) { 
				int index = smallerSet.get(s.getKey());
				if(currentIndex == -1 || index - 1 == currentIndex) {
					currentComponent.add(s.getKey());
					currentIndex = index;
				} 
				else {

					addCurrentComponent(ret, currentComponent);
					currentComponent = new ArrayList<>();
					currentComponent.add(s.getKey());
					currentIndex = index;
				}
				
				 
			}
			else {
				
				

				addCurrentComponent(ret, currentComponent);
				currentComponent = new ArrayList<>();
				currentIndex = -1;
			}
		}
		
	
		addCurrentComponent(ret, currentComponent); 
		
		return ret;
	}
	
//	private void addCurrentComponent(Map<String, Set<String>> ret, List<String> currentComponent) {
//		for(int i = 0; i < currentComponent.size(); i++) {
//			Set<String> children = new   HashSet<>();
//			for(int j = i + 1; j <= currentComponent.size(); j++) {
//				String s = StringUtils.join(currentComponent.subList(i, j), " ");
//				ret.put(s, new HashSet<>(children));
//				
//				children.add(s);
//			}
//		}
//	}
	
	
	public static Set<String> addCurrentComponent(Map<String, Set<String>> components, List<String> currentComponent) {
		
		String s = StringUtils.join(currentComponent.subList(0, currentComponent.size()), " ");
		Set<String> children = new HashSet<>();
		
		int size = currentComponent.size() - 1;
		if(size > 0) {
 
			for(int i = 0; i < currentComponent.size(); i++) {
				if(i < size && currentComponent.size() - i >= size) {
					children.addAll(addCurrentComponent(components, currentComponent.subList(i, i + size)));
				}
				else if (i >= size) {
					children.addAll(addCurrentComponent(components, currentComponent.subList(currentComponent.size() - i, currentComponent.size() - i + size)));
				}
			 
			}
			
		}
		components.put(s, new HashSet<>(children));
		children.add(s);
		return   children;
	}

	public <E> Map<E, Set<Integer>> createIndex(List<E> list) {
		Map<E, Set<Integer>> aSet = new HashMap<>();
		 
 
		
		for(ListEntry<E> e : ListIterable.iterable(list)) {
			CHMHelper.getFromMap(aSet, e.getValue(), () -> new HashSet<>()).add(e.getIndex());
			
		}
		
		return aSet;
	}
	public <E> List<Pair<List<E>, ComponentExist>> create(List<E> componentsA, List<E> componentsB, boolean includeNegativeTokens, boolean includeNegativeComponents, boolean includeAllSubComponents) {
		
 

		Map<E, Set<Integer>> aSet = createIndex(componentsA);
		

		Map<E,  Set<Integer>> bSet = createIndex(componentsB);
 
		
		return create(componentsA, componentsB, aSet, bSet, includeNegativeTokens, includeNegativeComponents, includeAllSubComponents);
		
		
		
		
 
	}
	
	public <E> List<Pair<List<E>, ComponentExist>> create(List<E> componentsA, List<E> componentsB, Map<E, Set<Integer>> aSet, Map<E,  Set<Integer>> bSet  , boolean includeNegativeTokens, boolean includeNegativeComponents, boolean includeAllSubComponents) {
		
		 
 
 
		
		List<Pair<List<E>, ComponentExist>> ret;
		if(aSet.size() < bSet.size()) {
			ret = getComponentsInt(componentsB, componentsA, bSet, aSet, ComponentExist.ONLY_B, ComponentExist.ONLY_A, includeNegativeTokens, includeNegativeComponents, includeAllSubComponents);
		}
		else {
			ret =  getComponentsInt(componentsA, componentsB, aSet, bSet, ComponentExist.ONLY_A, ComponentExist.ONLY_B, includeNegativeTokens, includeNegativeComponents, includeAllSubComponents);
		}
		
//		ret.addAll(getNonOverlapLargestComponent(ret, componentsA, componentsB, aSet, bSet));
		
		return ret; 
		
		
		
		
 
	}

	
//	private List< Pair<List<Integer>, ComponentExist>> getNonOverlapLargestComponent(List<Pair<List<Integer>, ComponentExist>> overlap, List<Integer> aList, List<Integer> bList, Map<Integer, Set<Integer>> aSet,
//			Map<Integer, Set<Integer>> bSet) {
//		
//		List<Pair<List<Integer>, ComponentExist>> nonOverlap = new ArrayList<>();
//		
//		List<Integer> currentList = new ArrayList<>();
//		
//		for(Entry<Integer, Integer> aEntry : aSet.entrySet()) {
//			if(!bSet.containsKey(aEntry.getKey())) {
//				currentList.add(aEntry.getKey());
//			}
//			else {
//				if(!currentList.isEmpty()) {
//					nonOverlap.add(new Pair<>(currentList, ComponentExist.ONLY_A));
//					currentList = new ArrayList<>();
//				}
//			}
//		}
//		
//		if(!currentList.isEmpty()) {
//			nonOverlap.add(new Pair<>(currentList, ComponentExist.ONLY_A));
//			currentList = new ArrayList<>();
//		}
//		
//		for(Entry<Integer, Integer> aEntry : bSet.entrySet()) {
//			if(!aSet.containsKey(aEntry.getKey())) {
//				currentList.add(aEntry.getKey());
//			}
//			else {
//				if(!currentList.isEmpty()) {
//					nonOverlap.add(new Pair<>(currentList, ComponentExist.ONLY_B));
//					currentList = new ArrayList<>();
//				}
//			}
//		}
//		
//		if(!currentList.isEmpty()) {
//			nonOverlap.add(new Pair<>(currentList, ComponentExist.ONLY_B));
//			currentList = new ArrayList<>();
//		}
//		
//		
//		
//		return nonOverlap;
//	}

	private <E> List<Pair<List<E>, ComponentExist>> getComponentsInt(
			List<E> largerList, List<E> smallerList,
			Map<E, Set<Integer>> largerSet, Map<E, Set<Integer>>  smallerSet,
			ComponentExist largerSetCE, ComponentExist smallerSetCE, boolean includeNegativeTokens, boolean includeNegativeComponents, boolean includeAllSubComponents) {
		List<List<Pair<E, Pair<Integer,Integer>>>> components = new ArrayList<>();
		int mapKey = 0;
//		Map<String, Set<String>> ret = new HashMap<>();
//		List<Integer> currentComponent = new ArrayList<>();
		Map<E, MutableInt> overlapRemaining = new HashMap<>();
//		int currentIndex = -1;
		
	 
		
		Map<Integer, Integer> currentIndex = new HashMap<>();
		// First Component, Second is indeces
		// indeces needed for non-overlap components
		
		Map<Integer, List<Pair<E, Pair<Integer, Integer>>>> currentComponent = new HashMap<>();
		for(  ListEntry<E> s :  ListIterable.iterable(largerList) ) {		
			if(smallerSet.containsKey(s.getValue())) { 
				int overlapSize = largerSet.get(s.getValue()).size();
				
				Set<Integer> index = smallerSet.get(s.getValue());
				if(index.size() > overlapSize)
					overlapSize = index.size();
				overlapRemaining.put(s.getValue(), new MutableInt(overlapSize));
			 
				 
					for(Iterator<Entry<Integer, Integer>> it = currentIndex.entrySet().iterator(); it.hasNext();) {
						Entry<Integer, Integer> entry = it.next();
						Integer cur = entry.getKey();
						int currIndex = entry.getValue();
						List<Pair<E, Pair<Integer, Integer>>> currComponent = currentComponent.get(cur);
						
						if(index.contains(currIndex + 1)) {
							currComponent.add( new Pair<>(s.getValue(), new Pair<>(s.getIndex(), currIndex + 1)));
							entry.setValue(currIndex + 1); 
						}
						else { 
							addCurrentComponentWithIndex(components, currComponent);
							it.remove();
						}
					}
 
				if(currentIndex.isEmpty()) {
					for(Integer i : index)  {
						Integer cur = mapKey++;
						currentIndex.put(cur, i);
						List<Pair<E, Pair<Integer, Integer>>> currComponent = new ArrayList<>();
						currComponent.add(new Pair<>(s.getValue(), new Pair<>(s.getIndex(), i)));
						currentComponent.put(cur, currComponent);
					}
				}
//				if(currentIndex == -1 || index.contains(currentIndex + 1) ) {
//					currentComponent.add(s.getValue());
//					currentIndex = currentIndex + 1;
//				} 
//				else {
//					components.add(currentComponent);
//					
//					addCurrentComponent(components, currentComponent);
//					currentComponent = new ArrayList<>();
//					currentComponent.add(s.getKey());
//					currentIndex = index;
//				}
				
				 
			}
			else {
				 
				
				for(Iterator<Entry<Integer, Integer>> it = currentIndex.entrySet().iterator(); it.hasNext();) {
					Entry<Integer, Integer> entry = it.next();
					Integer cur = entry.getKey(); 
					List<Pair<E, Pair<Integer, Integer>>> currComponent = currentComponent.get(cur);
					
			 
					addCurrentComponentWithIndex(components, currComponent);
						it.remove();
			 
				}
				
				
			}
		}
		
	
		List<Pair<List<E>, ComponentExist>> ret = new ArrayList<>();
		
		
		for(Iterator<Entry<Integer, Integer>> it = currentIndex.entrySet().iterator(); it.hasNext();) {
			Entry<Integer, Integer> entry = it.next();
			Integer cur = entry.getKey(); 
			List<Pair<E, Pair<Integer, Integer>>> currComponent = currentComponent.get(cur);
			
	 
			addCurrentComponentWithIndex(components, currComponent);
				it.remove();
	 
		}
		
		
		Collections.sort(components, (a,b) -> Integer.compare(b.size(), a.size()));
 
		Set<Integer> remaingingIndecesInLargerSet = new LinkedHashSet<>();
		Set<Integer> remaingingIndecesInSmallerSet = new LinkedHashSet<>();
		for(int i = 0; i < largerList.size();i++) {
			remaingingIndecesInLargerSet.add(i);
		}
		for(int i = 0; i < smallerList.size();i++) {
			remaingingIndecesInSmallerSet.add(i);
		}
		
		for(List<Pair<E, Pair<Integer, Integer>>> component : components) {
			boolean add= true;
			for(Pair<E, Pair<Integer, Integer>> c : component) {
				if(!remaingingIndecesInLargerSet.contains(c.getSecond().getFirst()) || !remaingingIndecesInSmallerSet.contains(c.getSecond().getSecond()) ) {
					add=false;
					
				}
			}
			if(add) {
				
				for(Pair<E, Pair<Integer, Integer>> c : component) {
					remaingingIndecesInLargerSet.remove(c.getSecond().getFirst());
					remaingingIndecesInSmallerSet.remove(c.getSecond().getSecond());
					MutableInt mi = overlapRemaining.get(c.getFirst());
					mi.decrement();
					if(mi.intValue() == 0) {
						overlapRemaining.remove(c);
					}
				}
				if(includeAllSubComponents) {
					addCurrentComponent(ret, component.stream().map((e) -> e.getFirst()).collect(Collectors.toList()), ComponentExist.BOTH);
				}
				else {
					ret.add(new Pair<>(component.stream().map((e) -> e.getFirst()).collect(Collectors.toList()), ComponentExist.BOTH));
				}
				
				
			}
//			if(overlapRemaining.isEmpty()) {
//				break;
//			}
		}
		
		
		if(includeNegativeTokens) {
			for(Integer i : remaingingIndecesInLargerSet) {
				List<E> list = new ArrayList<>();
				list.add(largerList.get(i));
				
				
				ret.add(new Pair<>(list, largerSetCE));
			}
			
			for(Integer i : remaingingIndecesInSmallerSet) {
				List<E> list = new ArrayList<>();
				list.add(smallerList.get(i));
				
				
				ret.add(new Pair<>(list, smallerSetCE));
			}
		}
//		boolean addConnectedNonOverlappingComponents = includeAll;
	if(includeNegativeComponents) {
		List<E> currentList = new ArrayList<>();
		int currentIndexNonOverlap = -1;
		for(Integer i : remaingingIndecesInLargerSet) {
			if(currentIndexNonOverlap == -1 || i == currentIndexNonOverlap + 1) {
				currentList.add(largerList.get(i));
				currentIndexNonOverlap = i;
			}
			else {
				if(includeAllSubComponents) {
					addCurrentComponent(ret, currentList, largerSetCE);
				}
				else {
					ret.add(new Pair<>(currentList, largerSetCE));
				}
//				
				
				currentList = new ArrayList<>();
				currentList.add(largerList.get(i));
				currentIndexNonOverlap = i;
			}
		}
		
		if(!currentList.isEmpty()) {
			if(includeAllSubComponents) {
				addCurrentComponent(ret, currentList, largerSetCE);
			}
			else {
				ret.add(new Pair<>(currentList, largerSetCE));
			}
		}
		currentList = new ArrayList<>();
		currentIndexNonOverlap = -1;
		
 
		for(Integer i : remaingingIndecesInSmallerSet) {
			if(currentIndexNonOverlap == -1 || i == currentIndexNonOverlap + 1) {
				currentList.add(smallerList.get(i));
				currentIndexNonOverlap = i;
			}
			else {
				if(includeAllSubComponents) {
					addCurrentComponent(ret, currentList, smallerSetCE);
				}
				else {
					ret.add(new Pair<>(currentList, smallerSetCE));
				}
				currentList = new ArrayList<>();
				currentList.add(smallerList.get(i));
				currentIndexNonOverlap = i;
			}
		}
		
		if(!currentList.isEmpty()) {
			if(includeAllSubComponents) {
				addCurrentComponent(ret, currentList, smallerSetCE);
			}
			else {			
				ret.add(new Pair<>(currentList, smallerSetCE));
			}
		}
	}
		
		boolean addIndividualTokens = false;
		
		if(addIndividualTokens) {
			List<Pair<List<E>, ComponentExist> > ret2 = new ArrayList<>();
			
			for(Pair<List<E>, ComponentExist> r : ret) {
				if(r.getFirst().size() > 1) {
					for(E component : r.getFirst()) {
						List<E> list = new ArrayList<>();
						list.add(component);
						ret2.add(new Pair<>(list, r.getSecond()));
					}
				}
			}
			ret.addAll(ret2);
			
		}
		
		
		return ret;
	}

 

	private <E> void addCurrentComponent(List<Pair<List<E>, ComponentExist>> ret, List<E> currentList, ComponentExist setCe) {
		for(int i = 0; i < currentList.size(); i++) { 
			for(int j = i + 1; j <= currentList.size(); j++) {
				ret.add(  new Pair<>( currentList.subList(i, j), setCe));
				 
			}
		}
		
	}

	private void addCurrentComponent(List<List<Integer>> ret, List<Integer> currentComponent) {
		for(int i = 0; i < currentComponent.size(); i++) { 
			for(int j = i + 1; j <= currentComponent.size(); j++) {
				ret.add(  new ArrayList<>( currentComponent.subList(i, j)));
				 
			}
		}
	}
	
	
	private <E> void addCurrentComponentWithIndex(List<List<Pair<E, Pair<Integer,Integer>>>> ret, List<Pair<E, Pair<Integer,Integer>>> currentComponent) {
	 
				ret.add(   currentComponent); 
	}
	
	private <E> void addCurrentComponentWithIndexAndSubComponents(List<List<Pair<E, Pair<Integer,Integer>>>> ret, List<Pair<E, Pair<Integer,Integer>>> currentComponent) {
		for(int i = 0; i < currentComponent.size(); i++) { 
			for(int j = i + 1; j <= currentComponent.size(); j++) {
				ret.add(  new ArrayList<>( currentComponent.subList(i, j)));
				 
			}
		}
	}


	public static enum ComponentExist {
		BOTH,ONLY_A,ONLY_B;
	}
}
