package expansion.set.strategies.rowclusterers;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ImmuMaker;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import expansion.set.strategies.blockers.Mappings;
import expansion.set.strategies.rowclusterers.label.LabelRowIndexClusterMapCreator;

public class CreateClustersFromMappings {
	public static void addCluster(Map<TableRowId, Set<TableRowId>> candidates, Map<Integer, Set<TableRowId>> clustersMap, Map<TableRowId, Integer> clustersIndex,
			Map<Integer, ClassId> clusterClassId, Function<AnnotatedTableId, ClassId> getClassId, Function<TableRowId,String> getLabel) {
		int indexCounter = 0;
		
		clustersMap.clear();
		clustersIndex.clear();
		clusterClassId.clear();

		for (Entry<TableRowId, Set<TableRowId>> entry : candidates.entrySet()) {

			if (!clustersIndex.containsKey(entry.getKey())) {
				Set<TableRowId> rowIds = getAllRowIds(candidates, entry.getKey());

				Set<Integer> existingClusters = searchCluster(clustersIndex, rowIds);

				int index = -1;

				if (existingClusters.size() == 1) {
					index = existingClusters.iterator().next();
				} else if (existingClusters.size() > 1) {
					Iterator<Integer> it = existingClusters.iterator();
					index = it.next();
					it.remove();

					for (Integer otherIndeces : existingClusters) {
						Set<TableRowId> insideCluster = clustersMap.get(otherIndeces);

						CHMHelper.getFromMap(clustersMap, index, () -> new HashSet<>()).addAll(insideCluster);
						
//						if(clustersMap.get(index).size() > 5000) {
//							List<String> lbs = new ArrayList<>();
//							
//							for(TableRowId match : clustersMap.get(index)) {
//							lbs.add(getLabel.apply(match));
//							}
//					
//							int sdf = 123;
//						}

						for (TableRowId inRowId : insideCluster) {
							clustersIndex.put(inRowId, index);
						}

						clustersMap.remove(otherIndeces);
					}
				} else {
					index = indexCounter++;
					if(getClassId != null)
						clusterClassId.put(index, getClassId.apply(entry.getKey().getTableId()));

				}

				for (TableRowId rowId : rowIds) {

					if(rowId.equals(entry.getKey())) {
						int sdf = 123;
					}
					if (clustersIndex.containsKey(rowId)) {
						if (!clustersIndex.get(rowId).equals(index)) {
							throw new IllegalStateException();
						}
					} else {
						CHMHelper.getFromMap(clustersMap, index, () -> new HashSet<>()).add(rowId);
						clustersIndex.put(rowId, index);
						
						

//						if(clustersMap.get(index).size() > 5000) {
//							List<String> lbs = new ArrayList<>();
//							
//							for(TableRowId match : clustersMap.get(index)) {
//							lbs.add(getLabel.apply(match));
//							}
//					
//							int sdf = 123;
//						}

					}

				}

			}



			// Set<TableRowId> tableRowIdSet = new HashSet<>();
			// tableRowIdSet.add(entry.getKey());
			// tableRowIdSet.addAll(entry.getValue());

		}

		// boolean added = false;
		//
		// for (TableRowId tableRowId : tableRowIdSet) {
		// Integer index = clustersIndex.get(entry.getKey());
		// clustersMap.get(index).addAll(entry.getValue());
		// for (TableRowId tableRowId : entry.getValue()) {
		// clustersIndex.put(tableRowId, index);
		// }
		// }
		// TODO Auto-generated method stub

	}



	private static Set<Integer> searchCluster(Map<TableRowId, Integer> clustersIndex, Set<TableRowId> rowIds) {

		Set<Integer> index = Collections.newSetFromMap(new ConcurrentHashMap<>());

		PersistentParallel6.runAndFinish(rowIds, (rowId) -> {
			if (clustersIndex.containsKey(rowId)) {

				index.add(clustersIndex.get(rowId));
			}
		} , LabelRowIndexClusterMapCreator.class, rowIds.size() < 1000);

		return index;
	}

	private  static Set<TableRowId> getAllRowIds(Map<TableRowId, Set<TableRowId>> candidates, TableRowId key) {

		Set<TableRowId> tableRowIdSet = new HashSet<>();

		getAllRowIds(candidates, tableRowIdSet, key);

		return tableRowIdSet;

	}

	private static void getAllRowIds(Map<TableRowId, Set<TableRowId>> candidates, Set<TableRowId> tableRowIdSet, TableRowId key) {

		tableRowIdSet.add(key);

		Set<TableRowId> cand = candidates.get(key);

		LinkedList<TableRowId> stack = new LinkedList<>();

		stack.addAll(cand);

		TableRowId rowId = null;
		while ((rowId = stack.poll()) != null) {
			if (!tableRowIdSet.contains(rowId)) {
				tableRowIdSet.add(rowId);
				if(candidates.containsKey(rowId)) {
					for (TableRowId innerRowId : candidates.get(rowId)) {
						if (!tableRowIdSet.contains(innerRowId)) {
							stack.add(innerRowId);

						}
					}
				}
				
			}
		}
		// for (TableRowId rowId : candidates.get(key)) {
		//
		// if (!tableRowIdSet.contains(rowId)) {
		// getAllRowIds(candidates, tableRowIdSet, rowId);
		//
		// }
		// }

	}



	public static Map<TableRowId, Set<TableRowId>> combineMappings(Mappings[] mappingsArray) {
		
		ConcurrentHashMap<TableRowId, Set<TableRowId>> ret = new ConcurrentHashMap<>();
		
		for(Mappings entry : mappingsArray) {
			
			PersistentParallel6.runAndFinish(entry.getMappings().entrySet(), (e) -> {
				Set<TableRowId> set = CHMHelper.getFromCHM(ret, e.getKey(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>()));
				set.addAll(e.getValue());
				set.add(e.getKey());
			}, CreateClustersFromMappings.class);
		}
		
		return ret; 
		 
	}
	
	
	public static ClustersFromMappings createClusters(Function<AnnotatedTableId, ClassId> getClassId,    Mappings... mappingsArray) {
		Map<Integer, Set<TableRowId>> clustersMap = new HashMap<>();
		Map<Integer, ClassId> clusterClassId = new HashMap<>();
		Map<TableRowId, Integer> clustersIndex = new HashMap<>();
	  
		Map<TableRowId,Set<TableRowId>> mappings = CreateClustersFromMappings.combineMappings(mappingsArray);
		CreateClustersFromMappings.addCluster(mappings, clustersMap, clustersIndex, clusterClassId, getClassId, 
				(rowId) -> {
					return null;
				});
		
		return new ClustersFromMappings(clustersMap, clusterClassId, clustersIndex);
	}
	public static class ClustersFromMappings{
		private final Map<Integer, Set<TableRowId>> clustersMap;
		private final Map<Integer, ClassId> clusterClassId;
		private final Map<TableRowId, Integer> clustersIndex;
		public ClustersFromMappings(Map<Integer, Set<TableRowId>> clustersMap, Map<Integer, ClassId> clusterClassId, Map<TableRowId, Integer> clustersIndex) {
			super();
			this.clustersMap = ImmuMaker.mapWithSet(clustersMap);
			this.clusterClassId = ImmuMaker.map(clusterClassId);
			this.clustersIndex = ImmuMaker.map(clustersIndex);
		}
		public Map<Integer, Set<TableRowId>> getClustersMap() {
			return clustersMap;
		}
		public Map<Integer, ClassId> getClusterClassId() {
			return clusterClassId;
		}
		public Map<TableRowId, Integer> getClustersIndex() {
			return clustersIndex;
		}
	 
		
		 
		
	}
}
