package expansion.set.strategies.rowclusterers;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.OtsuThresholderRounded;
import de.mannheim.uni.expansion.utils.OtsuThresholderRounded.DefaultIntegerScorable.DoubleScorable;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Tri;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;

public class DBScanSliptter {

	OtsuThresholderRounded thresholder = new OtsuThresholderRounded();

	@Inject
	public DBScanSliptter(OtsuThresholderRounded thresholder) {
		this.thresholder = thresholder;
	}

	public List<RowCluster> rowSplitter(RowCluster rowCluster, ClusterSimilarities sim, boolean duplicateUnassigned) {
		if (rowCluster.getRows().size() == 1) {
			List<RowCluster> ret = new ArrayList<>();
			ret.add(rowCluster);
			return ret;
		}

		Set<Integer> points = new HashSet<>();
		for (int i = 0; i < rowCluster.getRows().size(); i++) {
			points.add(i);
		}

		int minPts = 1;
		double legalEps = 0.80;
		double eps = 0.05;
		double candidateEps = 0.55;

		List<Set<Integer>> result = DBScan2(sim, points, eps, legalEps, candidateEps, minPts, true);

		// if (true && result.size() > 1) {
		// int maxSize = 0;
		// Set<Integer> maxCluster = null;
		// for (Set<Integer> c : result) {
		// if (c.size() > maxSize || maxCluster == null) {
		// maxCluster = c;
		// maxSize = c.size();
		// }
		// }
		// if (maxCluster != null) {
		// result = new ArrayList<>();
		// result.add(maxCluster);
		// }
		// }

		List<RowCluster> newRowClusters = getRowClusters(rowCluster, result);

		return newRowClusters;

	}

	public List<RowCluster> rowSplitter2(RowCluster rowCluster, ClusterSimilarities sim, boolean duplicateUnassigned) {
		if (rowCluster.getRows().size() == 1) {
			List<RowCluster> ret = new ArrayList<>();
			ret.add(rowCluster);
			return ret;
		}

		Set<Integer> points = new HashSet<>();
		for (int i = 0; i < rowCluster.getRows().size(); i++) {
			points.add(i);
		}
		int minPts = 2;
		double eps = 0.05;
		boolean finished = false;
		List<Set<Integer>> result = null;
		while (!finished) {
			result = DBScan(sim, points, eps, minPts);
			finished = true;
			// if (result.isEmpty()) {
			// eps = eps + 0.1;
			// } else {
			// finished = true;
			// }
		}

		if (result.size() > 1) {
			List<Set<Integer>> newResult = new ArrayList<>();

			double maxSize = result.stream().mapToDouble((e) -> e.size()).sum();
			List<DoubleScorable> scorable = new ArrayList<>();

			for (Set<Integer> res : result) {
				scorable.add(new DoubleScorable(res.size() / maxSize));
			}

			double threshold = thresholder.getThresholdDefault(scorable);

			for (Set<Integer> res : result) {
				double score = res.size() / maxSize;
				if (score > threshold) {
					newResult.add(res);
				}
			}

			result = newResult;

		}

		if (duplicateUnassigned) {
			Set<Integer> assigned = result.stream().flatMap(coll -> coll.stream()).collect(Collectors.toSet());
			Set<Integer> unassigned = new HashSet<Integer>();
			for (int i = 0; i < rowCluster.getRows().size(); i++) {
				if (!assigned.contains(i))
					unassigned.add(i);
			}
			List<Set<Integer>> newResult = new ArrayList<>();
			for (Set<Integer> res : result) {
				Set<Integer> newRes = new HashSet<>();
				newRes.addAll(res);
				newRes.addAll(unassigned);
				newResult.add(newRes);
			}

			result = newResult;
		}

		if (true) {
			int maxSize = 0;
			Set<Integer> maxCluster = null;
			for (Set<Integer> c : result) {
				if (c.size() > maxSize || maxCluster == null) {
					maxCluster = c;
					maxSize = c.size();
				}
			}
			if (maxCluster != null) {
				result = new ArrayList<>();
				result.add(maxCluster);
			}
		}
		List<RowCluster> newRowClusters = getRowClusters(rowCluster, result);
		if (result.size() > 1) {
			int sdf = 123;
		}

		return newRowClusters;
	}

	private List<RowCluster> getRowClusters(RowCluster rowCluster, List<Set<Integer>> optimum) {
		List<RowCluster> ret = new ArrayList<>();
		List<TableRowId> originalRows = rowCluster.getRows();

		for (Set<Integer> cluster : optimum) {
			ArrayList<Integer> clusterList = new ArrayList<>(cluster);
			List<TableRowId> rows = new ArrayList<>();
			// Map<ClusterEntryKey, Double> similarities = new HashMap<>();
			for (ListEntry<Integer> endEntry1 : ListIterable.iterable(clusterList)) {
				rows.add(originalRows.get(endEntry1.getValue()));
				for (ListEntry<Integer> endEntry2 : ListIterable.iterable(clusterList)) {
					if (endEntry1.getIndex() >= endEntry2.getIndex()) {
						continue;
					}
					// similarities.put(new
					// ClusterEntryKey(endEntry1.getIndex(),
					// endEntry2.getIndex()),
					// originalSimilarities.get(new
					// ClusterEntryKey(endEntry1.getValue(),
					// endEntry2.getValue())));

				}
			}

			ret.add(new RowCluster(rowCluster.getClassId(), rows, rowCluster.getSimilarities()));

		}

		return ret;
	}

	Map<Integer, Set<Integer>> createRegionQueryParallel(ClusterSimilarities clusterSimilarities, Set<Integer> points, double eps) {
		ConcurrentHashMap<Integer, Set<Integer>> retMap = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(points, (point) -> {
			Set<Integer> neighboringPoints = regionQuery(point, eps, clusterSimilarities, points);
			retMap.put(point, neighboringPoints);
		} , this.getClass(), points.size() < 1000);

		return retMap;
	}

	public List<Set<Integer>> DBScan(ClusterSimilarities clusterSimilarities, Set<Integer> points, double eps, int minPts) {
		
		Map<Integer, Set<Integer>> regionsQueries = createRegionQueryParallel(clusterSimilarities, points, eps);
		
		List<Set<Integer>> clusters = new ArrayList<>();
//		Set<Integer> visited = new HashSet<>();
		Set<Integer> noise = new HashSet<>();
		Set<Integer> added = new HashSet<>();

		for (Integer point : points) {
			if (added.contains(point) || noise.contains(point)) {
				continue;
			}
//			visited.add(point);
			LinkedHashSet<Integer> neighboringPoints = new LinkedHashSet<>(regionsQueries.get(point));// regionQuery(point,
																		// eps,
																		// clusterSimilarities,
																		// points);
			
//			neighboringPoints.removeAll(added);
//			neighboringPoints.removeAll(noise);
			if (neighboringPoints.size() < minPts) {
				noise.add(point);
			} else {
				Set<Integer> newCluster = new HashSet<>();
				newCluster.add(point);
				expandCluster(point, newCluster, neighboringPoints, clusterSimilarities, points, eps, minPts,   added, noise,regionsQueries);
				clusters.add(newCluster);
			}
		}

		return clusters;
	}

	private void expandCluster(Integer point, Set<Integer> newCluster, LinkedHashSet<Integer> neighboringPoints, ClusterSimilarities clusterSimilarities,
			Set<Integer> points, double eps, int minPts,   Set<Integer> added, Set<Integer> noise, Map<Integer, Set<Integer>> regionsQueries) {
		newCluster.add(point);
		while(!neighboringPoints.isEmpty()) {
			Integer nPoint = neighboringPoints.iterator().next();
			if (!added.contains(nPoint)) {
				newCluster.add(nPoint);
				added.add(nPoint);
				Set<Integer> nNeighboringPoints = regionsQueries.get(nPoint);// regionQuery(nPoint,
																				// eps,
																				// clusterSimilarities,
																				// points);
				
				 
				
				if (nNeighboringPoints.size() >= minPts) {
					neighboringPoints.addAll(nNeighboringPoints);
//					if (!added.contains(nPoint)) {
						
//					}
				}
			}
			
			
			
			neighboringPoints.remove(nPoint);
		}


	}

	private Set<Integer> regionQuery(Integer point, double eps, ClusterSimilarities clusterSimilarities, Set<Integer> points) {
		return clusterSimilarities.regionQuery(point, eps);

	}

	public List<Set<Integer>> DBScan2(ClusterSimilarities clusterSimilarities, Set<Integer> points, double eps, double legalEps, double candidateEps, int minPts,
			boolean addLegalFields) {


		Map<Integer, Set<Integer>> regionsQueries = createRegionQueryParallel(clusterSimilarities, points, eps);

		// minPts = Math.min(minPts,
		// regionsQueries.values().stream().mapToInt(Set::size).max().orElse(1));

		Map<Integer, Set<Integer>> legalFields = createRegionQueryParallel(clusterSimilarities, points, legalEps);

		Map<Integer, Set<Integer>> illegalFields = new HashMap<>();
		for (Entry<Integer, Set<Integer>> e : legalFields.entrySet()) {
			Set<Integer> illegalFieldsSet = new HashSet<>(points);
			illegalFieldsSet.removeAll(e.getValue());
			illegalFieldsSet.remove(e.getKey());
			illegalFields.put(e.getKey(), illegalFieldsSet);
		}

		// List<Entry<Integer, Set<Integer>>> list = new
		// ArrayList<>(illegalFields.entrySet());
		// Collections.sort(list, (e1,e2) -> {
		// int c1 = Integer.compare(x, y)
		// return Integer.compare(e1.getValue().size(), e2.getValue().size());
		// });
		//
		// Collections.reverse(list);

		Map<Integer, Set<Integer>> candidateFields = createRegionQueryParallel(clusterSimilarities, points, candidateEps);
		List<Integer> sorted = new ArrayList<>();
		sorted.addAll(points);
		Collections.sort(sorted, (e1, e2) -> {
			int c1 = Integer.compare(regionsQueries.get(e2).size(), regionsQueries.get(e1).size());
			if (c1 != 0)
				return c1;
			return Integer.compare(illegalFields.get(e1).size(), illegalFields.get(e2).size());
		});
		// Collections.reverse(sorted);

		TreeSet<Integer> sortedPoints = new TreeSet<>((e1, e2) -> {
			return Integer.compare(regionsQueries.get(e1).size(), regionsQueries.get(e2).size());
		});
		sortedPoints.addAll(points);

		List<Tri<Set<Integer>, Set<Integer>, Set<Integer>>> clustersWithLegalPoints = new ArrayList<>();
		Set<Integer> noise = new HashSet<>();
		Set<Integer> added = new HashSet<>();
		Set<Integer> visited = new HashSet<>();
		

		Set<Integer> addedLegalPoints = new HashSet<>();

		for (Integer point : sorted) {
			// System.out.println("NeighborhoodSize: " +
			// regionsQueries.get(point).size() + ", IllegalSize: " +
			// illegalFields.get(point).size());


			if (visited.contains(point)) {
				continue;
			}
			visited.add(point);

			Set<Integer> neighboringPoints = new HashSet<>(regionsQueries.get(point));// regionQuery(point,


			// eps,
			// clusterSimilarities,
			// points);
			// if (neighboringPoints.size() < minPts) {

			if (false) {
				noise.add(point);
			} else {
				Set<Integer> illegalPoints = new HashSet<>(illegalFields.get(point));
				// neighboringPoints.removeAll(visited);
				// legalPoints.removeAll(visited);
				// legalPoints.removeAll(neighboringPoints);

				Set<Integer> newCluster = new HashSet<>();
				// newCluster.addAll(neighboringPoints);
				newCluster.add(point);
				// newCluster.addAll(legalFields.get(point));
				// visited.addAll(newCluster);
				
				HashSet<Integer> newVisted = new HashSet<>(visited);

				Set<Integer> expandedLegalPoints = expandCluster2(point, newCluster, neighboringPoints, illegalPoints, clusterSimilarities, points, eps, minPts,
						newVisted, added,
						regionsQueries,
 illegalFields, candidateFields);

				Set<Integer> legalPoints = candidateFields.get(point);

				legalPoints.addAll(expandedLegalPoints);
				
				for (Integer lPoint : legalPoints) {
					illegalPoints.addAll(illegalFields.get(lPoint));
				}

				visited = newVisted;
				legalPoints.removeAll(illegalPoints);
				legalPoints.removeAll(visited);
				legalPoints.removeAll(addedLegalPoints);
				newCluster.addAll(legalPoints);

				visited.addAll(legalPoints);
				addedLegalPoints.addAll(legalPoints);

				// if (newCluster.size() > minPts) {
				if (newCluster.size() > minPts) {


					clustersWithLegalPoints.add(new Tri<>(newCluster, legalPoints, illegalPoints));
				} else {
					int sdf = 123;
				}
			}

		}

		// if (clustersWithLegalPoints.size() > 1) {
		// List<Tri<Set<Integer>, Set<Integer>, Set<Integer>>> newResult = new
		// ArrayList<>();
		//
		// double maxSize = clustersWithLegalPoints.stream().mapToDouble((e) ->
		// e.getA().size()).sum();
		// List<DoubleScorable> scorable = new ArrayList<>();
		//
		// for (Tri<Set<Integer>, Set<Integer>, Set<Integer>> res :
		// clustersWithLegalPoints) {
		// scorable.add(new DoubleScorable(res.getA().size() / maxSize));
		// }
		//
		// double threshold = thresholder.getThresholdDefault(scorable);
		//
		// for (Tri<Set<Integer>, Set<Integer>, Set<Integer>> res :
		// clustersWithLegalPoints) {
		// double score = res.getA().size() / maxSize;
		// if (score > threshold) {
		// newResult.add(res);
		// }
		// }
		//
		// clustersWithLegalPoints = newResult;
		//
		// }

		// if (true && clustersWithLegalPoints.size() > 1) {
		// int maxSize = 0;
		// Tri<Set<Integer>, Set<Integer>, Set<Integer>> maxCluster = null;
		// for (Tri<Set<Integer>, Set<Integer>, Set<Integer>> c :
		// clustersWithLegalPoints) {
		// if (c.getA().size() > maxSize || maxCluster == null) {
		// maxCluster = c;
		// maxSize = c.getA().size();
		// }
		// }
		//
		// List<Tri<Set<Integer>, Set<Integer>, Set<Integer>>> newResult = new
		// ArrayList<>();
		//
		// for (Tri<Set<Integer>, Set<Integer>, Set<Integer>> c :
		// clustersWithLegalPoints) {
		// if (c.getA().size() >= maxSize) {
		// newResult.add(c);
		// }
		// }
		//
		// // if (maxCluster != null) {
		// clustersWithLegalPoints = newResult;
		// // }
		// }

		List<Set<Integer>> clusters = new ArrayList<>();

		for (Tri<Set<Integer>, Set<Integer>, Set<Integer>> e : clustersWithLegalPoints) {
			Set<Integer> cluster = e.getA();
			if (addLegalFields) {
				Set<Integer> legalPoints = e.getB();
				// legalPoints.removeAll(visited);
				Set<Integer> illegalPoints = e.getC();
				legalPoints.removeAll(visited);
				legalPoints.removeAll(illegalPoints);

				cluster.addAll(legalPoints);
			}
			clusters.add(cluster);
		}

//		if (clusters.size() > 1) {
//			List<Set<Integer>> newResult = new ArrayList<>();
//
//			double maxSize = clusters.stream().mapToDouble((e) -> e.size()).sum();
//			List<DoubleScorable> scorable = new ArrayList<>();
//
//			for (Set<Integer> res : clusters) {
//				scorable.add(new DoubleScorable(res.size() / maxSize));
//			}
//
//			double threshold = thresholder.getThresholdDefault(scorable);
//
//			for (Set<Integer> res : clusters) {
//				double score = res.size() / maxSize;
//				if (score > threshold) {
//					newResult.add(res);
//				}
//			}
//
//			clusters = newResult;
//
//		}

		// if (true && clusters.size() > 1) {
		// int maxSize = 0;
		// Set<Integer> maxCluster = null;
		// for (Set<Integer> c : clusters) {
		// if (c.size() > maxSize || maxCluster == null) {
		// maxCluster = c;
		// maxSize = c.size();
		// }
		// }
		// if (maxCluster != null) {
		// clusters = new ArrayList<>();
		// clusters.add(maxCluster);
		// }
		// }

		// Set<Integer> unassignedPoints = new HashSet<>(points);
		// for (Set<Integer> c : clusters) {
		// unassignedPoints.removeAll(c);
		// }
		//
		// for (Set<Integer> p : clusters) {
		// p.addAll(unassignedPoints);
		// }

		if (clusters.size() == 0) {
			// clusters.add(points);
		}

		return clusters;
	}

	private Set<Integer> expandCluster2(Integer point, Set<Integer> newCluster, Set<Integer> neighboringPoints, Set<Integer> illegalPoints,
			ClusterSimilarities clusterSimilarities, Set<Integer> points, double eps, int minPts, Set<Integer> visited, Set<Integer> added,
			Map<Integer, Set<Integer>> regionsQueries, Map<Integer, Set<Integer>> illegalFields, Map<Integer, Set<Integer>> candidateFields) {

		Set<Integer> legalPoints = new HashSet<>();
		// newCluster.add(point);
		while (!neighboringPoints.isEmpty()) {
			Integer nPoint = neighboringPoints.iterator().next();
			if (!visited.contains(nPoint)) {
				visited.add(nPoint);


				Set<Integer> nNeighboringPoints = regionsQueries.get(nPoint);// regionQuery(nPoint,
																				// eps,
																				// clusterSimilarities,
																				// points);
				Set<Integer> nIllegalPoints = new HashSet<>(illegalFields.get(nPoint));

				boolean contains = false;
				for (Integer illegalPoint : illegalPoints) {
					if (nNeighboringPoints.contains(illegalPoint)) {
						contains = true;
						break;
					}
				}

				for (Integer illegalPoint : nIllegalPoints) {
					if (newCluster.contains(illegalPoint)) {
						contains = true;
						break;
					}
				}

				if (contains) {

				} else {

					newCluster.add(nPoint);

					if (nNeighboringPoints.size() >= minPts) {

						Set<Integer> nLegalPoints = candidateFields.get(nPoint);
						legalPoints.addAll(nLegalPoints);
						// newCluster.addAll(nNeighboringPoints);

						// nIllegalPoints.removeAll(newCluster);

						illegalPoints.addAll(nIllegalPoints);
						// newCluster.addAll(legalFields.get(point));

						nNeighboringPoints.removeAll(visited);
						neighboringPoints.addAll(nNeighboringPoints);
					}
 else {
						int sf = 123;
					}
				}

				// nNeighboringPoints.removeAll(illegalPoints);

				// neighboringPoints.addAll(nNeighboringPoints);


			}

			// if (!added.contains(nPoint)) {
			// newCluster.add(nPoint);
			// }

			neighboringPoints.remove(nPoint);
		}

		return legalPoints;

	}

	private Set<Integer> regionQueryOld(Integer point, double eps, ClusterSimilarities clusterSimilarities, Set<Integer> points) {
		Set<Integer> neighborHood = new HashSet<>();
		for (Integer nPoint : points) {
			if (nPoint == point) {
				neighborHood.add(nPoint);
			} else {
				double sim = clusterSimilarities.get(new ClusterEntryKey(point, nPoint));
				if ((1 - sim) < eps) {
					neighborHood.add(nPoint);
				}
			}
		}

		return neighborHood;
	}

}
