package expansion.set.strategies.rowclusterers;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ConcurrentSkipListSet;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableDouble;
import org.apache.commons.lang.mutable.MutableInt;

import com.google.common.util.concurrent.AtomicDouble;

import de.mannheim.uni.expansion.ThreadLocalList;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.DecimalFormatter;
import de.mannheim.uni.expansion.utils.DoubleIteraterable;
import de.mannheim.uni.expansion.utils.DoubleIteraterable.DoubleEntry;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.OtsuThresholderRounded;
import de.mannheim.uni.expansion.utils.OtsuThresholderRounded.DefaultIntegerScorable.DoubleScorable;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.informatik.dws.expansion.utils.MutableReference;
import expansion.set.clustering.model.ClusterSimilaritiesImpl;
import expansion.set.clustering.model.ClusterSimilaritiesImplCacheWrapper;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;

public class GreedyClustering {
 
	
	private final OtsuThresholderRounded otsu = new OtsuThresholderRounded(); 
	public static int numOfThreads = 98;
	
	public static double threshold = 0.0;
	public static double veryLowPairThreshold = 0.0;

	@Inject
	public GreedyClustering( ) {
	 
	}
	
	public static boolean SEQUENTIAL = false;

	public static boolean NO_THRESHOLD_ADJUSTMENT = true;

	public List<RowCluster> rowSplitter(ClassId classId, List<TableRowId> list, RowClusterSimilarities sim) {
		RowCluster rc = new RowCluster(classId, list, sim);
		return rowSplitter(rc, new ClusterSimilaritiesImpl(rc, sim)); 
	}

	public List<RowCluster> rowSplitter(RowCluster rowCluster, ClusterSimilarities sim) {
		return rowSplitter(rowCluster, sim, SEQUENTIAL);
	}
	// public List<RowCluster> rowSplitter(RowCluster rowCluster,
	// ClusterSimilarities sim, boolean sequential) {
	// return rowSplitter(rowCluster, sim, sequential, null);
	// }

	public List<RowCluster> rowSplitter(RowCluster rowCluster, ClusterSimilarities sim, boolean sequential) {
		return rowSplitter(rowCluster, sim, sequential, threshold, veryLowPairThreshold);
	}

	public List<RowCluster> rowSplitter(RowCluster rowCluster, ClusterSimilarities sim, boolean sequential, double threshold) {
		return rowSplitter(rowCluster, sim, sequential, threshold, veryLowPairThreshold);
	}

	public List<RowCluster> rowSplitter(RowCluster rowCluster, ClusterSimilarities sim, boolean sequential, double threshold, double veryLowPairThreshold) {
//		if (rowCluster.getRows().size() == 1) {
//			List<RowCluster> ret = new ArrayList<>();
////			ret.add(rowCluster);
//			return ret;
//		}
// 
		Set<Integer> points = new HashSet<>();
		for (int i = 0; i < rowCluster.getRows().size(); i++) {
			points.add(i);
		}

 

		List<Set<Integer>> result = run(points, sim, true, false, false, sequential, threshold, veryLowPairThreshold);

		// if (true && result.size() > 1) {
		// int maxSize = 0;
		// Set<Integer> maxCluster = null;
		// for (Set<Integer> c : result) {
		// if (c.size() > maxSize || maxCluster == null) {
		// maxCluster = c;
		// maxSize = c.size();
		// }
		// }
		// if (maxCluster != null) {
		// result = new ArrayList<>();
		// result.add(maxCluster);
		// }
		// }

		List<RowCluster> newRowClusters = getRowClusters(rowCluster, result);
 
		int sdf = 123;
		return newRowClusters;

	}
 
	private List<RowCluster> getRowClusters(RowCluster rowCluster, List<Set<Integer>> optimum) {
		List<RowCluster> ret = new ArrayList<>();
		List<TableRowId> originalRows = rowCluster.getRows();

		for (Set<Integer> cluster : optimum) {
			ArrayList<Integer> clusterList = new ArrayList<>(cluster);
			List<TableRowId> rows = new ArrayList<>();
			// Map<ClusterEntryKey, Double> similarities = new HashMap<>();
			for (ListEntry<Integer> endEntry1 : ListIterable.iterable(clusterList)) {
				rows.add(originalRows.get(endEntry1.getValue()));
				for (ListEntry<Integer> endEntry2 : ListIterable.iterable(clusterList)) {
					if (endEntry1.getIndex() >= endEntry2.getIndex()) {
						continue;
					}
					// similarities.put(new
					// ClusterEntryKey(endEntry1.getIndex(),
					// endEntry2.getIndex()),
					// originalSimilarities.get(new
					// ClusterEntryKey(endEntry1.getValue(),
					// endEntry2.getValue())));

				}
			}

			ret.add(new RowCluster(rowCluster.getClassId(), rows, rowCluster.getSimilarities()));

		}
		
		if(ret.size() > 1) {
			int sdf = 123;
		}

		return ret;
	}

	public List<Set<Integer>> run(Set<Integer> keySet, ClusterSimilarities similarities, boolean reassignTest, boolean filterSizeOne, boolean silhoutteRun,
			boolean sequential) {
		return run(keySet, similarities, reassignTest, filterSizeOne, silhoutteRun, sequential, threshold, veryLowPairThreshold);
	}
	
	public List<Cluster> mainRun(List<Cluster> clusters, Set<Integer> keySet, ClusterSimilarities similarities, boolean sequential, double threshold,
			double veryLowPairThreshold) {
		

		
		ConcurrentHashMap<UUID, Cluster> blockingClusterMap;
		ConcurrentHashMap<Integer, Set<UUID>> blockingBlockToClusters;
		ConcurrentLinkedQueue<Cluster> clusterQueue;

		if (clusters == null || clusters.isEmpty()) {
			blockingClusterMap = new ConcurrentHashMap<>();
			blockingBlockToClusters = new ConcurrentHashMap<>();
			clusterQueue = new ConcurrentLinkedQueue<>();
		} else {
			blockingBlockToClusters = getBlockToClusterIndex(clusters, similarities);
			blockingClusterMap = getCHMFromClusterCollection(clusters);
			clusterQueue = new ConcurrentLinkedQueue<>(clusters);
		}

		PersistentParallel6.runAndFinish(keySet, (i) -> {
			Optional<Pair<Double, Cluster>> max = getMax(similarities, i, clusterQueue, blockingClusterMap, blockingBlockToClusters, null,
					Optional.of(Boolean.TRUE), threshold, veryLowPairThreshold);
			
			if(max.isPresent() && max.get().getFirst() > 0) {
				max.get().getSecond().addPoint(i, similarities, null);
				if (similarities.supportsBlocking()) {

					for (Integer block : similarities.getBlocks(i)) {
						CHMHelper.getFromCHM(blockingBlockToClusters, block, () -> Collections.newSetFromMap(new ConcurrentHashMap<>()))
								.add(max.get().getSecond().getKey());
					}
				}

			}
			else {
				Cluster newCluster = new Cluster(i,WITH_CENTROIDS);
//				newCluster.add(i);

				if (similarities.supportsBlocking()) {
					blockingClusterMap.put(newCluster.getKey(), newCluster);
					Set<Integer> rowBlocks = similarities.getBlocks(i);
					if (rowBlocks != null) {
						for (Integer block : rowBlocks) {
							CHMHelper.getFromCHM(blockingBlockToClusters, block, () -> Collections.newSetFromMap(new ConcurrentHashMap<>()))
									.add(newCluster.getKey());
						}
					}
				} else {
					clusterQueue.add(newCluster);
				}
			}
			
//			done++; if(done % 2000 == 0) {System.out.println("Finished " + DecimalFormatter.format(done / totalSize) + " %");}
		}, this.getClass(), "MAIN", sequential || keySet.size() < POINTS_CONCURRENT_FLOOR);

		if (similarities.supportsBlocking()) {
			clusters = new ArrayList<>(blockingClusterMap.values());
		} else {
			clusters = new ArrayList<>(clusterQueue);
		}

		return clusters;

	}

	private static boolean SUBTRACT_AVERAGE_DIFFERENCE = false;

	private static boolean AGREEMENTS_AND_DISAGREEMENTS = true;

	private static boolean MINIMIZE_DISAGREEMENTS = false;

	private static boolean OLD_ADDITION = false;

	public List<Set<Integer>> run(Set<Integer> keySet, ClusterSimilarities similarities, boolean reassignTest, boolean filterSizeOne, boolean silhoutteRun,
			boolean sequential, double threshold, double veryLowPairThreshold) {

		// if (reassignTest) {
		// OLD_ADDITION = false;
		// }
		//
		// if (silhoutteRun) {
		// // POWER_THE_SIM = true;
		// OLD_ADDITION = true;
		// reassignTest = true;
		// }
		// int done = 0;
		// double totalSize = keySet.size();

		// final ExecutorService jobQueue =
		// Executors.newFixedThreadPool(numOfThreads);
		boolean needToRun = true;

		List<Cluster> clusters;

		if (keySet.size() < 100000 || true) {
			if (similarities.supportsCaching() && similarities.cacheExists()) {
				clusters = similarities.readClusters().stream().map((e) -> new Cluster(e, similarities, WITH_CENTROIDS)).collect(Collectors.toList());
			} else {
				clusters = mainRun(null, keySet, similarities, sequential, threshold, veryLowPairThreshold);

				if (similarities.supportsCaching()) {
					similarities.storeClusters(clusters.stream().map((e) -> e.getPoints()).collect(Collectors.toList()));
				}
			}

			if (reassignTest) {

				if (similarities instanceof ClusterSimilaritiesImplCacheWrapper) {
					((ClusterSimilaritiesImplCacheWrapper) similarities).writeUpdates();
				}

				// System.out.println("Size before: " + clusters.size());
				// List<Set<Integer>> showInside = clusters.stream().map((e) ->
				// e.getPoints()).collect(Collectors.toList());
				int beforeKLj = clusters.size();
				clusters = kljReassign(similarities, clusters, sequential, threshold, veryLowPairThreshold, keySet.size() > 100000);
				if (keySet.size() > 100000) {
					System.out.println("Before KLj: " + beforeKLj);
					System.out.println("After KLj: " + clusters.size());
				}

				int sumSize = clusters.stream().mapToInt((e) -> e.size()).sum();

				if (sumSize != keySet.size()) {
					throw new IllegalStateException();
				}
				// clusters = reassignReassignableClusters(keySet,similarities,
				// clusters, sequential);
				// clusters = reassignReassignableClusters(keySet,similarities,
				// clusters, sequential);
				// System.out.println("Size after: " + clusters.size());
			}
		} else {

			clusters = new ArrayList<>();
			Iterator<Integer> it = keySet.iterator();

			Set<Integer> next100K = new HashSet<>();
			while (it.hasNext()) {
				needToRun = true;
				next100K.add(it.next());
				if (next100K.size() == 200000) {
					if (!clusters.isEmpty()) {
						if (similarities instanceof ClusterSimilaritiesImplCacheWrapper) {
							((ClusterSimilaritiesImplCacheWrapper) similarities).writeUpdates();
						}
					}
					clusters = mainRun(clusters, next100K, similarities, sequential, threshold, veryLowPairThreshold);
					if (similarities instanceof ClusterSimilaritiesImplCacheWrapper) {
						((ClusterSimilaritiesImplCacheWrapper) similarities).writeUpdates();
					}

					int beforeKLj = clusters.size();
					clusters = kljReassign(similarities, clusters, sequential, threshold, veryLowPairThreshold, true);
					System.out.println("Before KLj: " + beforeKLj);
					System.out.println("After KLj: " + clusters.size());
					next100K = new HashSet<>();
					needToRun = false;

				}

			}

			if (needToRun) {
				if (similarities instanceof ClusterSimilaritiesImplCacheWrapper) {
					((ClusterSimilaritiesImplCacheWrapper) similarities).writeUpdates();
				}
				clusters.addAll(mainRun(clusters, next100K, similarities, sequential, threshold, veryLowPairThreshold));
				int beforeKLj = clusters.size();
				clusters = kljReassign(similarities, clusters, sequential, threshold, veryLowPairThreshold, true);
				System.out.println("Before KLj: " + beforeKLj);
				System.out.println("After KLj: " + clusters.size());
			}
		}

		// jobQueue.shutdown();
		// try {
		// jobQueue.awaitTermination(Long.MAX_VALUE, TimeUnit.NANOSECONDS);
		// } catch (InterruptedException e) {
		// e.printStackTrace();
		// }

		int sumSize = clusters.stream().mapToInt((e) -> e.size()).sum();

		if (sumSize != keySet.size()) {
			throw new IllegalStateException();
		}
//		for(Integer i : keySet) {
//			
//		}
		if (silhoutteRun) {
			clusters = reassignReassignableClusters(keySet, similarities, clusters, sequential, threshold, veryLowPairThreshold);
			clusters = reassignReassignableClusters(keySet, similarities, clusters, sequential, threshold, veryLowPairThreshold);
			clusters = reassignReassignableClusters(keySet, similarities, clusters, sequential, threshold, veryLowPairThreshold);
			clusters = reassignReassignableClusters(keySet, similarities, clusters, sequential, threshold, veryLowPairThreshold);

		}
		

		if(filterSizeOne) {
		clusters = clusters.stream().filter((c) -> c.size() > 1).collect(Collectors.toList());
		}
		
		if (false && silhoutteRun) {
			Set<Integer> reassign = new HashSet<>();
			Map<Integer, Cluster> newClusters = new HashMap<>();
			Map<Integer, Double> sils = new HashMap<>();
			List<DoubleScorable> silList = new ArrayList<>();
				for(int c= 0; c<clusters.size();c++) {
					double silhoutte = 0.0;
					int count = 0;
					Cluster clusterSet = clusters.get(c);
					for(Integer i: clusterSet.getPoints()) { 
						Optional<Double> aOptional = averageDissimilarity(similarities, i, clusterSet.getPoints());
						
						double a = 0.5;
						if(aOptional.isPresent()) {
							a = aOptional.get();
						}
						else {
							int sdf = 123;
						}
						
						double b = 0.5;
						for(int c2= 0; c2<clusters.size();c2++) {
							if(c != c2) {
								
								Optional<Double> bTemp =  averageDissimilarity(similarities, i, clusters.get(c2).getPoints());
								if(bTemp.isPresent() && bTemp.get() < b) {
									b = bTemp.get();
								}
								
							}
						}
						silhoutte += (b - a) / Math.max(a, b);
						count++;
						
					}
					if(count == 0) {
						int df = 123;
					}
					double sil =  silhoutte / count;
//					System.out.println("Silhoutte: " + sil);
					
					sils.put(c, sil);
					silList.add( new DoubleScorable(sil ));
//					if(sil < 0) {
//						reassign.addAll(clusterSet);
//					}
//					else {
//						
//					}
					
					
					
				}
				
				double otsuThreshold = otsu.getThresholdNoExtra(silList);
				
				for(Entry<Integer, Double> silEntry : sils.entrySet()) {
					if(silEntry.getValue() < otsuThreshold && otsuThreshold - silEntry.getValue() > 0.2 ) {
						reassign.addAll(clusters.get(silEntry.getKey()).getPoints());
					}
					else {
						newClusters.put(silEntry.getKey(), clusters.get(silEntry.getKey()));
					}
					
				}
			
				
				clusters = new ArrayList<>();
			for(Entry<Integer, Cluster>entry : newClusters.entrySet()) {
				clusters.add(entry.getValue());
			}
			
			
			
			
			for(Integer i : reassign) {
				Double maxQuality = null;
				int maxCluster =-1;
				for(int c = 0; c < clusters.size(); c++) {
					
//					double quality = 0.0;
					Pair<Double, Double> qualityAndLowest = clusters.get(c).getQualityAndLowestScore(i, similarities);
					double quality = qualityAndLowest.getFirst();
//					for(Integer j : clusters.get(c)) {
//						if(j != i) {
//							Double sim = getScore(similarities, i, j);
//							if(sim != null)
//								quality += sim;
//						}
//					}
					
					if(maxQuality == null || quality > maxQuality) {
						maxQuality = quality;
						maxCluster = c;
					}
					
				}
				
				if(maxQuality == null || maxQuality <= 0) {
//					Set<Integer> newCluster = new HashSet<>();
//					newCluster.add(i);
//					clusters.add(newCluster);
				}
				else { 
					Cluster cluster = clusters.get(maxCluster);
					 
						
					cluster.addPoint(i, similarities, null);
				}
			}
			
			
		}
		

		
		return clusters.stream().map(Cluster::getPoints).collect(Collectors.toList());
	}
	
	// private Optional<Pair<Double, Integer>> getMaxOld(ClusterSimilarities
	// similarities,Integer i, List<Cluster> clusters, Set<Integer> except) {
	// Double maxQuality = null;
	// int maxCluster =-1;
	// for(int c = 0; c < clusters.size(); c++) {
	//
	// if(except == null || !except.contains(c)) {
	// double quality = clusters.get(c).getQuality(i, similarities);
	//
	// if(maxQuality == null || quality > maxQuality) {
	// maxQuality = quality;
	// maxCluster = c;
	// }
	// }
	//
	// }
	//
	// if(maxQuality == null)
	// return Optional.empty();
	// return Optional.of(new Pair<>(maxQuality, maxCluster));
	// }
	
	private List<Cluster> kljReassign(ClusterSimilarities similarities, List<Cluster> clusters, boolean sequential, double threshold,
			double veryLowPairThreshold, boolean verbose) {

		ConcurrentHashMap<UUID, Cluster> blockingClusterMap = getCHMFromClusterCollection(clusters);

		Set<UUID> changed = Collections
				.newSetFromMap(new ConcurrentHashMap<>());
		changed.addAll(blockingClusterMap.keySet());

		// int iteration = 0;
		while (!changed.isEmpty()) {
			final ConcurrentHashMap<Integer, Set<UUID>> blockingBlockToClusters = getBlockToClusterIndex(blockingClusterMap.values(), similarities);
			final ConcurrentHashMap<UUID, Set<Integer>> clusterToBlocks = getClusterToBlocks(blockingBlockToClusters);

			Set<UUID> newChanged = Collections.newSetFromMap(new ConcurrentHashMap<>());

			// ConcurrentHashMap<UUID, Set<UUID>> needToCompareTo = new
			// ConcurrentHashMap<>();
			ConcurrentHashMap<Integer, Set<Integer>> precompute;
			if (similarities instanceof ClusterSimilaritiesImplCacheWrapper) {
				precompute = new ConcurrentHashMap<>();
			} else {
				precompute = null;
			}

			final Set<UUID> finalChanged = changed;
			// ConcurrentHashMap<UUID, Set<UUID>> needToCompareTo = new
			// ConcurrentHashMap<>();
			Set<Pair<UUID, UUID>> needToCompareTo = Collections.newSetFromMap(new ConcurrentHashMap<>());
			PersistentParallel6.runAndFinish(changed, (uuid) -> {
				Cluster cluster1 = blockingClusterMap.get(uuid);

				if (cluster1 != null) {
					if (precompute != null) {

						for (Integer i1 : cluster1.getPoints()) {
							for (Integer i2 : cluster1.getPoints()) {
								if (i1 < i2) {
									CHMHelper.getFromCHM(precompute, i1, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(i2);
								} else if (i2 > i2) {
									CHMHelper.getFromCHM(precompute, i2, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(i1);
								}
							}
						}
					}

					if (similarities.supportsBlocking()) {
						Set<Integer> blocks = clusterToBlocks.get(uuid);
						if (uuid.toString().equals("bab9fb23-29af-44c4-a91d-91ac360f7bdb")) {
							int sdfsgsg = 123;
						}
						if (blocks != null) {
							Set<UUID> otherClusters = new HashSet<>();
							for (Integer block : blocks) {
								otherClusters.addAll(blockingBlockToClusters.get(block));
							}

							{

								for (UUID c2 : otherClusters) {
									if (c2.equals(uuid)) {
										// continue;
									} else if (!finalChanged.contains(c2) || uuid.compareTo(c2) == -1) {
										if (uuid.compareTo(c2) == -1) {
											// CHMHelper.getFromCHM(needToCompareTo,
											// uuid, () ->
											// Collections.newSetFromMap(new
											// ConcurrentHashMap<>())).add(c2);
											needToCompareTo.add(new Pair<>(uuid, c2));

										} else {
											// CHMHelper.getFromCHM(needToCompareTo,
											// c2,
											// () ->
											// Collections.newSetFromMap(new
											// ConcurrentHashMap<>())).add(uuid);
											needToCompareTo.add(new Pair<>(c2, uuid));
										}

										if (precompute != null) {

											Cluster cluster2 = blockingClusterMap.get(c2);

											for (Integer i1 : cluster1.getPoints()) {
												for (Integer i2 : cluster2.getPoints()) {
													if (i1 < i2) {
														CHMHelper.getFromCHM(precompute, i1, () -> Collections.newSetFromMap(new ConcurrentHashMap<>()))
																.add(i2);
													} else {
														CHMHelper.getFromCHM(precompute, i2, () -> Collections.newSetFromMap(new ConcurrentHashMap<>()))
																.add(i1);
													}
												}
											}
										}

									}
								}
							}
						}
					} else {
						for (Cluster c2 : clusters) {
							if (!finalChanged.contains(c2.getKey()) || uuid.compareTo(c2.getKey()) == -1) {
								if (uuid.compareTo(c2.getKey()) == -1) {
									// CHMHelper.getFromCHM(needToCompareTo,
									// uuid, () ->
									// Collections.newSetFromMap(new
									// ConcurrentHashMap<>())).add(c2.getKey());
									needToCompareTo.add(new Pair<>(uuid, c2.getKey()));
								} else {
									// CHMHelper.getFromCHM(needToCompareTo,
									// c2.getKey(), () ->
									// Collections.newSetFromMap(new
									// ConcurrentHashMap<>())).add(uuid);
									needToCompareTo.add(new Pair<>(c2.getKey(), uuid));
								}
							}
						}
					}
				}

			}, this.getClass() + "::FINDING_PAIRS", changed.size() < 50);

			if (similarities instanceof ClusterSimilaritiesImplCacheWrapper) {
				((ClusterSimilaritiesImplCacheWrapper) similarities).precompute(precompute);
			}

			ConcurrentHashMap<Integer, ConcurrentLinkedQueue<Pair<UUID, UUID>>> queueMap = new ConcurrentHashMap<>();
			ConcurrentHashMap<Integer, ConcurrentHashMap<UUID, Object>> containsCheckMap = new ConcurrentHashMap<>();
			
			PersistentParallel6.runAndFinish(needToCompareTo, (pair) -> {
				Object o = new Object();
				boolean found = false;
				int i = 0;
				while (!found) {
					ConcurrentHashMap<UUID, Object> checkMap = CHMHelper.getFromCHM(containsCheckMap, i, () -> new ConcurrentHashMap<>());
					Object retFirst = CHMHelper.getFromCHM(checkMap, pair.getFirst(), () -> o);
					if (retFirst == o) {
						Object retSecond = CHMHelper.getFromCHM(checkMap, pair.getSecond(), () -> o);
						if (retSecond == o) {
							found = true;
							CHMHelper.getFromCHM(queueMap, i, () -> new ConcurrentLinkedQueue<>()).add(pair);
						}
						else {
							checkMap.remove(pair.getFirst());
						}
					}

					i++;
				}
			}, this.getClass());
			double totalSize = needToCompareTo.size();
			int done = 0;
			double nextSize = 0.1;
			needToCompareTo.clear();
			for (ConcurrentLinkedQueue<Pair<UUID, UUID>> queue : queueMap.values()) {
				int queueSize = queue.size();
				int largePairSize = 0;
				for (Pair<UUID, UUID> pair : queue) {
					UUID uuid1 = pair.getFirst();
					UUID uuid2 = pair.getSecond();
					Cluster c1 = blockingClusterMap.get(uuid1);
					Cluster c2 = blockingClusterMap.get(uuid2);
					if (c1 != null && c2 != null) {
						if (c1.size() > 15 && c2.size() > 15) {
							largePairSize++;
						}
					}
				}
				PersistentParallel6.runAndFinish(queue, (pair) -> {

					UUID uuid1 = pair.getFirst();
					UUID uuid2 = pair.getSecond();
					Cluster c1 = blockingClusterMap.get(uuid1);
					Cluster c2 = blockingClusterMap.get(uuid2);
					if (c1 != null && c2 != null) {
						Tri<Cluster, Cluster, Boolean> returnedClusters = updateClusters(similarities, c1, c2);
						if (returnedClusters.getA() != null && returnedClusters.getB() != null
								&& returnedClusters.getA().getPoints().equals(returnedClusters.getB().getPoints())) {
							int sdf = 123;
						}
						if (returnedClusters.getC()) {
							if (returnedClusters.getA() != null) {
								blockingClusterMap.put(returnedClusters.getA().getKey(), returnedClusters.getA());
								newChanged.add(returnedClusters.getA().getKey());
							} else {
								blockingClusterMap.remove(uuid1);
							}

							if (returnedClusters.getB() != null) {
								blockingClusterMap.put(returnedClusters.getB().getKey(), returnedClusters.getB());
								newChanged.add(returnedClusters.getB().getKey());
							} else {
								blockingClusterMap.remove(uuid2);
							}
						}
					}

				}, this.getClass() + "::RUNNING_PAIRS", queueSize < 3000 && largePairSize < 2);
			
				done += queue.size();
				if (verbose) {
					double percent = done / totalSize;
					if (percent > nextSize) {
						System.err.println("KLJ-Comparison-Done: " + DecimalFormatter.format(done / totalSize));
						nextSize += 0.1;
					}
				}


			}
//			ConcurrentHashMap<UUID, Lock> locks = new ConcurrentHashMap<>();

//			PersistentParallel6.runAndFinish(needToCompareTo, (pair) -> {
//				Lock firstLock = CHMHelper.getFromCHM(locks, pair.getFirst(), () -> new ReentrantLock());
//				Lock secondLock = CHMHelper.getFromCHM(locks, pair.getSecond(), () -> new ReentrantLock());
//				boolean notDone = true;
//				while (notDone) {
//					try {
//						boolean firstLockSucces = firstLock.tryLock(ThreadLocalRandom.current().nextInt(30, 100), TimeUnit.MILLISECONDS);
//						if (firstLockSucces) {
//							boolean secondLockSuccess = secondLock.tryLock(ThreadLocalRandom.current().nextInt(30, 100), TimeUnit.MILLISECONDS);
//							if (secondLockSuccess) {
//								Cluster c1 = blockingClusterMap.get(pair.getFirst());
//								Cluster c2 = blockingClusterMap.get(pair.getSecond());
//								if (c1 != null && c2 != null) {
//									Tri<Cluster, Cluster, Boolean> returnedClusters = updateClusters(similarities, c1, c2);
//									if (returnedClusters.getC()) {
//										if (returnedClusters.getA() != null) {
//											blockingClusterMap.put(returnedClusters.getA().getKey(), returnedClusters.getA());
//											newChanged.add(returnedClusters.getA().getKey());
//										} else {
//											blockingClusterMap.remove(pair.getFirst());
//										}
//
//										if (returnedClusters.getB() != null) {
//											blockingClusterMap.put(returnedClusters.getB().getKey(), returnedClusters.getB());
//											newChanged.add(returnedClusters.getB().getKey());
//										} else {
//											blockingClusterMap.remove(pair.getSecond());
//										}
//									}
//								}
//								secondLock.unlock();
//								notDone = false;
//							}
//							firstLock.unlock();
//						}
//					} catch (InterruptedException e) {
//
//					}
//
//				}
//
//			}, this.getClass() + "::RUNNING_PAIRS", needToCompareTo.size() < 100);

			// PersistentParallel6.runAndFinish(needToCompareTo, (pair) -> {
			// boolean notDone = true;
			// while (notDone) {
			//
			//
			// CountDownLatch latch = new CountDownLatch(1);
			// CountDownLatch returnedLatch1 = CHMHelper.getFromCHM(locks,
			// pair.getFirst(), () -> latch);
			// CountDownLatch returnedLatch2 = CHMHelper.getFromCHM(locks,
			// pair.getSecond(), () -> latch);
			// if (latch == returnedLatch1 && latch == returnedLatch2) {
			// Cluster c1 = blockingClusterMap.get(pair.getFirst());
			// Cluster c2 = blockingClusterMap.get(pair.getSecond());
			// if (c1 != null && c2 != null) {
			// Tri<Cluster, Cluster, Boolean> returnedClusters =
			// updateClusters(similarities, c1, c2);
			// if (returnedClusters.getC()) {
			// if (returnedClusters.getA() != null) {
			// blockingClusterMap.put(returnedClusters.getA().getKey(),
			// returnedClusters.getA());
			// newChanged.add(returnedClusters.getA().getKey());
			// } else {
			// blockingClusterMap.remove(pair.getFirst());
			// }
			//
			// if (returnedClusters.getB() != null) {
			// blockingClusterMap.put(returnedClusters.getB().getKey(),
			// returnedClusters.getB());
			// newChanged.add(returnedClusters.getB().getKey());
			// } else {
			// blockingClusterMap.remove(pair.getSecond());
			// }
			// }
			// }
			// latch.countDown();
			//
			// notDone = false;
			//
			// locks.remove(pair.getFirst());
			// locks.remove(pair.getSecond());
			// } else {
			// latch.countDown();
			// if (latch == returnedLatch1) {
			// locks.remove(pair.getFirst());
			// }
			// if (latch == returnedLatch2) {
			// locks.remove(pair.getSecond());
			// }
			//
			// try {
			// returnedLatch1.await();
			// returnedLatch2.await();
			// TimeUnit.SECONDS.sleep(1);
			// } catch (Exception e1) {
			// // TODO Auto-generated catch block
			// e1.printStackTrace();
			// }
			// }
			//
			// }
			// }, this.getClass());

			Set<UUID> emptySetChanged = new HashSet<>();
			emptySetChanged.addAll(changed);
			emptySetChanged.addAll(newChanged);

			while (!emptySetChanged.isEmpty()) {

				Set<UUID> newEmptySetChanged = Collections.newSetFromMap(new ConcurrentHashMap<>());
				Set<UUID> finalEmptySetChanged = emptySetChanged;
				PersistentParallel6.runAndFinish(finalEmptySetChanged, (uuid) -> {
					Cluster c = blockingClusterMap.get(uuid);
					if (c != null && c.size() > 1) {
						Tri<Cluster, Cluster, Boolean> returnedClusters = updateClusters(similarities, c, null);
						if (returnedClusters.getC()) {
							if (returnedClusters.getA() != null) {
								blockingClusterMap.put(returnedClusters.getA().getKey(), returnedClusters.getA());
								newChanged.add(returnedClusters.getA().getKey());
								newEmptySetChanged.add(returnedClusters.getA().getKey());
							} else {
								throw new IllegalStateException();
							}

							if (returnedClusters.getB() != null) {
								blockingClusterMap.put(returnedClusters.getB().getKey(), returnedClusters.getB());
								newChanged.add(returnedClusters.getB().getKey());
								newEmptySetChanged.add(returnedClusters.getB().getKey());
							} else {
								throw new IllegalStateException();
							}
						}
					}
				}, this.getClass() + "::RUNNING_EMPTY_SET");

				emptySetChanged = newEmptySetChanged;

			}
			changed = newChanged;

			for (Iterator<UUID> it = changed.iterator(); it.hasNext();) {
				Cluster cluster = blockingClusterMap.get(it.next());
				if (cluster == null) {
					it.remove();
				}
			}
			if (verbose) {
				System.out.println("KLJ-CHANGED-Size: " + changed.size());
			}
			
			// iteration++;
			//
			// if(iteration % 100000 == 0) {
			// int sfd = 123;
			// }

		}

		return new ArrayList<>(blockingClusterMap.values());
	}

	public static class PriorityQueueType implements Comparable<PriorityQueueType> {

		private final SimType type;
		private final double score;
		private final Integer pointA;
		private final Integer pointBorCluster;

		public PriorityQueueType(SimType type, double score, Integer pointA, Integer pointBorCluster) {
			super();
			this.type = type;
			this.score = score;
			this.pointA = pointA;
			this.pointBorCluster = pointBorCluster;
		}

		public SimType getType() {
			return type;
		}

		public double getScore() {
			return score;
		}

		public Integer getPointA() {
			return pointA;
		}

		public Integer getPointBorCluster() {
			return pointBorCluster;
		}

		@Override
		public int compareTo(PriorityQueueType o) {
			return Double.compare(o.score, score);
		}

	}

	public static enum SimType {
		POINT_TO_POINT, POINT_TO_CLUSTER;
	}

	double getQuality(Integer point, Set<Integer> cluster, Map<ClusterEntryKey, Double> scores) {
		double sum = 0.0;
		double highest = -1.0;
		double lowest = 1.0;
		double sumNeg = 0.0;
		int countNeg = 0;
		double sumPos = 0.0;
		int countPos = 0;
		int count = 0;
		for (Integer otherPoint : cluster) {
			if (!point.equals(otherPoint)) {
				double sim = scores.get(new ClusterEntryKey(point, otherPoint));
				;
				sum += sim;
				count++;
				if (sim < 0.0) {
					sumNeg += sim;
					countNeg++;
				}
				if (sim > 0.0) {
					sumPos += sim;
					countPos++;
				}
				if (sim < lowest) {
					lowest = sim;
				}
				if (sim > highest) {
					highest = sim;
				}

			}
		}

		if (count == 0) {
			return 0.0;
		}

		if (true) {
			return sum;
		} else {

			double averagePos = countPos > 0 ? sumPos / countPos : 0.0;
			double averageNeg = countNeg > 0 ? -1.0 * (sumNeg / countNeg) : 0.0;

			double averageSum = sum / (count);

			if (averageSum > 0.0) {

				if (averagePos > averageNeg) {
					return sum;
				} else if ((averageNeg - averagePos) < averageSum) {
					return sum;
				} else {
					return 0.01;
				}

			}
			else if (averageSum < 0.0) {
				if (averageNeg > averagePos) {
					return sum;
				} else if ((averagePos - averageNeg) < averageSum) {
					return sum;
				} else {
					return 0.01;
				}
			}
			else {
				return 0.0;
			}
		}
		// if (lowest < 0.0) {
		// lowest = -1.0 * lowest;
		// } else {
		// lowest = 0.0;
		// }
		//
		// if (highest < 0.0) {
		// highest = 0.0;
		// }
		//
		// if (sum < 0.0) {
		// if (averagePos > averageNeg) {
		// if (highest > lowest) {
		// return 0.01;
		// } else {
		// return 0.0;
		// }
		// } else {
		// return sum;
		// }
		// } else if (sum > 0.0) {
		// if (averageNeg > averagePos) {
		// if (lowest > highest) {
		// return -0.01;
		// } else {
		// return 0.0;
		// }
		// } else {
		// return sum;
		// }
		// } else {
		//
		// return sum;
		// }

		// return sum;
	}

	public static class ClusterScore {

		private double sum = 0.0;
		private double sumPos = 0.0;
		private int countPos = 0;
		private double sumNeg = 0.0;
		private int countNeg = 0;

		public void add(double score) {

			if (score > 0.0) {
				sumPos += score;
				countPos++;
				if (!MINIMIZE_DISAGREEMENTS) {
					sum += score;
				}
			} else {
				sumNeg += Math.abs(score);
				countNeg++;
				if (AGREEMENTS_AND_DISAGREEMENTS) {
					sum += score;
				}
				if (MINIMIZE_DISAGREEMENTS) {
					sum -= score;
				}

			}

		}

		public void subtract(double score) {
			if (score < 0.0) {
				sumPos += Math.abs(score);
				countPos++;
				if (!MINIMIZE_DISAGREEMENTS) {
					sum -= score;
				}
			} else {
				sumNeg += score;
				countNeg++;
				if (AGREEMENTS_AND_DISAGREEMENTS) {
					sum -= score;
				}
				if (MINIMIZE_DISAGREEMENTS) {
					sum += score;
				}
			}
		}

		public double getScore() {

			if (SUBTRACT_AVERAGE_DIFFERENCE) {
				double ret = sum;

				if (countPos > 0) {
					ret += (sumPos / countPos);

				}
				if (countNeg > 0) {
					ret -= (sumNeg / countNeg);
				}
				return ret;
			} else {
				return sum;
			}

		}
	}

	public static double MIN_IMPACT = 0.001;
	private Tri<Cluster, Cluster, Boolean> updateClusters(ClusterSimilarities similarities, Cluster cluster1, Cluster cluster2) {

		Set<Integer> key1 = new HashSet<>(cluster1.getPoints());
		Set<Integer> key2;
		if (cluster2 != null) {
			key2 = new HashSet<>(cluster2.getPoints());
		} else {
			key2 = new HashSet<>();
		}

		Set<Integer> keySet = new HashSet<>(cluster1.getPoints());
		if (cluster2 != null) {
			keySet.addAll(cluster2.getPoints());
		}

		Map<ClusterEntryKey, Double> map = new HashMap<>();

		long resultingSize = DoubleIteraterable.getResultingSize(keySet.size());
		// PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(keySet),
		// (doubleEntry) -> {
		boolean sequential = resultingSize < 500000L;
		
		for (DoubleEntry<Integer> doubleEntry : DoubleIteraterable.iterable(keySet)) {
			ClusterEntryKey key = new ClusterEntryKey(doubleEntry.getFirst(), doubleEntry.getSecond());
			map.put(key, similarities.get(key));
		}
		// }, resultingSize, this.getClass() + "::COMPUTE_SIM_UPDATE_CLUSTER",
		// resultingSize < 50000);
		
		

		boolean changed = false;
		boolean keepMoving = true;

		Double beforeChanges = null;
		Double impactOfJoin = null;
		MutableDouble current = new MutableDouble();
		final ConcurrentHashMap<Integer, AtomicDouble> moveImpact = new ConcurrentHashMap<>();
		if (cluster2 != null) {
			AtomicDouble before = new AtomicDouble();
			AtomicDouble after = new AtomicDouble();
			
			
			PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(keySet), (doubleEntry) -> {		
//		 
				ClusterEntryKey key = new ClusterEntryKey(doubleEntry.getFirst(), doubleEntry.getSecond());
				double value = similarities.get(key);
			
			
//			for (Entry<ClusterEntryKey, Double> entry : map.entrySet()) {
				after.addAndGet(value);
				if (key1.contains(key.getE1())) {
					if (key1.contains(key.getE2())) {
						before.addAndGet(value);
						CHMHelper.getFromCHM(moveImpact, key.getE1(), () -> new AtomicDouble()).addAndGet(-1.0 * value);
						CHMHelper.getFromCHM(moveImpact, key.getE2(), () -> new AtomicDouble()).addAndGet(-1.0 * value);
					} else {
						before.addAndGet(-1.0 * value);
						CHMHelper.getFromCHM(moveImpact, key.getE1(), () -> new AtomicDouble()).addAndGet(value);
						CHMHelper.getFromCHM(moveImpact, key.getE2(), () -> new AtomicDouble()).addAndGet(value);
					}
				} else {
					if (key2.contains(key.getE2())) {
						before.addAndGet(value);
						CHMHelper.getFromCHM(moveImpact, key.getE1(), () -> new AtomicDouble()).addAndGet(-1.0 * value);
						CHMHelper.getFromCHM(moveImpact, key.getE2(), () -> new AtomicDouble()).addAndGet(-1.0 * value);
					} else {
						before.addAndGet(-1.0 * value);
						CHMHelper.getFromCHM(moveImpact, key.getE1(), () -> new AtomicDouble()).addAndGet(value);
						CHMHelper.getFromCHM(moveImpact, key.getE2(), () -> new AtomicDouble()).addAndGet(value);
					}
				}

//			}
			
			}, resultingSize, this.getClass() + "::WITH_BEFORE", sequential);
			
			double beforeScore = before.doubleValue();
			double impact;
			if (MINIMIZE_DISAGREEMENTS) {
				impact = beforeScore - after.doubleValue();
			} else {
				impact = after.doubleValue() - beforeScore;
			}
			if (impact > MIN_IMPACT) {
				impactOfJoin = impact;
				beforeChanges = beforeScore;
			}
			current.setValue(beforeScore);
		} else {
			AtomicDouble before = new AtomicDouble();

			PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(keySet), (doubleEntry) -> {
				//
				ClusterEntryKey key = new ClusterEntryKey(doubleEntry.getFirst(), doubleEntry.getSecond());
				double value = similarities.get(key);
				// for (Entry<ClusterEntryKey, Double> entry : map.entrySet()) {

				if (key1.contains(key.getE1())) {
					if (key1.contains(key.getE2())) {
						before.addAndGet(value);
						CHMHelper.getFromCHM(moveImpact, key.getE1(), () -> new AtomicDouble()).addAndGet(-1.0 * value);
						CHMHelper.getFromCHM(moveImpact, key.getE2(), () -> new AtomicDouble()).addAndGet(-1.0 * value);
					} else {
						before.addAndGet(-1.0 * value);
						CHMHelper.getFromCHM(moveImpact, key.getE1(), () -> new AtomicDouble()).addAndGet(value);
						CHMHelper.getFromCHM(moveImpact, key.getE2(), () -> new AtomicDouble()).addAndGet(value);
					}
				} else {
					if (key2.contains(key.getE2())) {
						before.addAndGet(value);
						CHMHelper.getFromCHM(moveImpact, key.getE1(), () -> new AtomicDouble()).addAndGet(-1.0 * value);
						CHMHelper.getFromCHM(moveImpact, key.getE2(), () -> new AtomicDouble()).addAndGet(-1.0 * value);
					} else {
						before.addAndGet(-1.0 * value);
						CHMHelper.getFromCHM(moveImpact, key.getE1(), () -> new AtomicDouble()).addAndGet(value);
						CHMHelper.getFromCHM(moveImpact, key.getE2(), () -> new AtomicDouble()).addAndGet(value);
					}
				}

				// }
			}, resultingSize, this.getClass() + "::WITHOUT_BEFORE", sequential);

			current.setValue(before.doubleValue());

		}

		for (AtomicDouble md : moveImpact.values()) {
			md.addAndGet(current.doubleValue());
		}

		// Map<Integer, List<Integer>> recursionMap = new HashMap<>();

		// int iteration = 0;
		while (keepMoving) {


			if (!OLD_ADDITION) {
				if (moveImpact.isEmpty()) {

					PersistentParallel6.runAndFinishWithSize(DoubleIteraterable.iterable(keySet), (doubleEntry) -> {
						//
						ClusterEntryKey key = new ClusterEntryKey(doubleEntry.getFirst(), doubleEntry.getSecond());
						double value = similarities.get(key);

						// for (Entry<ClusterEntryKey, Double> entry :
						// map.entrySet()) {
						if (key1.contains(key.getE1())) {
							if (key1.contains(key.getE2())) {
								CHMHelper.getFromCHM(moveImpact, key.getE1(), () -> new AtomicDouble()).addAndGet(-1.0 * value);
								CHMHelper.getFromCHM(moveImpact, key.getE2(), () -> new AtomicDouble()).addAndGet(-1.0 * value);
							} else {
								CHMHelper.getFromCHM(moveImpact, key.getE1(), () -> new AtomicDouble()).addAndGet(value);
								CHMHelper.getFromCHM(moveImpact, key.getE2(), () -> new AtomicDouble()).addAndGet(value);
							}
						} else {
							if (key2.contains(key.getE2())) {
								CHMHelper.getFromCHM(moveImpact, key.getE1(), () -> new AtomicDouble()).addAndGet(-1.0 * value);
								CHMHelper.getFromCHM(moveImpact, key.getE2(), () -> new AtomicDouble()).addAndGet(-1.0 * value);
							} else {
								CHMHelper.getFromCHM(moveImpact, key.getE1(), () -> new AtomicDouble()).addAndGet(value);
								CHMHelper.getFromCHM(moveImpact, key.getE2(), () -> new AtomicDouble()).addAndGet(value);
							}
						}

						// }
					}, resultingSize, this.getClass() + "::WITHOUT_BEFORE", sequential);

					for (AtomicDouble md : moveImpact.values()) {
						md.addAndGet(current.doubleValue());
					}
				}

				double maxEffect = current.doubleValue();
				Integer move = null;

				for (Entry<Integer, AtomicDouble> pair : moveImpact.entrySet()) {
					if (pair.getValue().doubleValue() - MIN_IMPACT > maxEffect) {
						maxEffect = pair.getValue().doubleValue();
						move = pair.getKey();
					}
				}

				if (move == null) {
					keepMoving = false;
				} else {
					changed = true;
					if (key1.contains(move)) {
						key1.remove(move);
						key2.add(move);
					} else {
						key1.add(move);
						key2.remove(move);
					}
					current.setValue(maxEffect);
					moveImpact.clear();
				}

			} else {

				ConcurrentLinkedQueue<Pair<Integer, Double>> list = new ConcurrentLinkedQueue<>();

				PersistentParallel6.runAndFinish(keySet, (point) -> {
					Set<Integer> other;
					// Set<Integer> own;
					if (key1.contains(point)) {
						// own = key1;
						other = key2;
					} else {
						// own = key2;
						other = key1;

					}

					double afterMoveScore;

					// if (!OLD_ADDITION) {
					// afterMoveScore = current.doubleValue();
					// for (Integer otherPoint : keySet) {
					// if (!otherPoint.equals(point)) {
					// ClusterEntryKey key = new ClusterEntryKey(point,
					// otherPoint);
					//
					// if (own.contains(otherPoint)) {
					// afterMoveScore -= map.get(key);
					// } else {
					//
					// afterMoveScore += map.get(key);
					// }
					// }
					// }
					// } else {
						ClusterScore afterMove = new ClusterScore();

						for (Entry<ClusterEntryKey, Double> entry : map.entrySet()) {

							if (entry.getKey().getE1() == point.intValue()) {
								if (other.contains(entry.getKey().getE2())) {
									afterMove.add(entry.getValue());
								} else {
									afterMove.subtract(entry.getValue());
								}
							} else if (entry.getKey().getE2() == point.intValue()) {
								if (other.contains(entry.getKey().getE1())) {
									afterMove.add(entry.getValue());
								} else {
									afterMove.subtract(entry.getValue());
								}
							}
							else {

								if (key1.contains(entry.getKey().getE1())) {
									if (key1.contains(entry.getKey().getE2())) {
										afterMove.add(entry.getValue());
									} else {
										afterMove.subtract(entry.getValue());
									}
								} else {
									if (key2.contains(entry.getKey().getE2())) {
										afterMove.add(entry.getValue());
									} else {
										afterMove.subtract(entry.getValue());
									}
								}
							}

						}
						afterMoveScore = afterMove.getScore();
					// }

					double effect;
					if (MINIMIZE_DISAGREEMENTS) {
						effect = current.doubleValue() - afterMoveScore;
					} else {
						effect = afterMoveScore - current.doubleValue();
					}
					// double effect = getQuality(point, other, map) -
					// getQuality(point, own, map);
					if (effect > MIN_IMPACT) {
						list.add(new Pair<>(point, afterMoveScore));
						// MutableReference<Pair<Integer, Double>> reference =
						// list.getOrCreate(() -> new MutableReference<>());
						// Pair<Integer, Double> maxPair = reference.get();
						// if (maxPair == null || effect > maxPair.getSecond())
						// {
						// reference.set();
						// }
					}

				}, this.getClass() + "::COMPUTE_MAX_EFFECT", keySet.size() < 30000);

				double maxEffect = 0.0;
				Integer move = null;

				for (Pair<Integer, Double> pair : list) {
					if (move == null || pair.getSecond() > maxEffect) {
						maxEffect = pair.getSecond();
						move = pair.getFirst();
					}
				}

				if (move == null) {
					keepMoving = false;
				} else {
					changed = true;
					if (key1.contains(move)) {
						key1.remove(move);
						key2.add(move);
					} else {
						key1.add(move);
						key2.remove(move);
					}
					current.setValue(maxEffect);
				}
			}


			// iteration++;

			// if (iteration == 10000) {
			// System.out.println("THIS IS ITERATION 10,000");
			// }
		}

		if (beforeChanges != null) {
			if (changed) {
				// ClusterScore afterMoves = new ClusterScore();
				// for (Entry<ClusterEntryKey, Double> entry : map.entrySet()) {
				//
				// if (key1.contains(entry.getKey().getE1())) {
				// if (key1.contains(entry.getKey().getE2())) {
				// afterMoves.add(entry.getValue());
				// } else {
				// afterMoves.subtract(entry.getValue());
				// }
				// } else {
				// if (key2.contains(entry.getKey().getE2())) {
				// afterMoves.add(entry.getValue());
				// } else {
				// afterMoves.subtract(entry.getValue());
				// }
				// }
				//
				// }
				//
				// double impact = afterMoves.getScore() - beforeChanges;

				double impact;

				if (MINIMIZE_DISAGREEMENTS) {
					impact = beforeChanges - current.doubleValue();
				} else {
					impact = current.doubleValue() - beforeChanges;
				}
				if (impactOfJoin - impact > MIN_IMPACT) {
						key1.addAll(key2);
						key2.clear();
					}


			} else {
				key1.addAll(key2);
				key2.clear();
				changed = true;
			}
		}

		if (changed) {
			if (cluster2 == null) {
				return new Tri<>(new Cluster(cluster1.getKey(), key1, similarities, WITH_CENTROIDS), new Cluster(key2, similarities, WITH_CENTROIDS), true);
			} else {
				if (key1.isEmpty()) {
					return new Tri<>(null, new Cluster(cluster2.getKey(), key2, similarities, WITH_CENTROIDS), true);
				} else if (key2.isEmpty()) {
					return new Tri<>(new Cluster(cluster1.getKey(), key1, similarities, WITH_CENTROIDS), null, true);
				}
				else {
					return new Tri<>(new Cluster(cluster1.getKey(), key1, similarities, WITH_CENTROIDS),
							new Cluster(cluster2.getKey(), key2, similarities, WITH_CENTROIDS), true);
				}

			}
		} else {
			return new Tri<>(null, null, false);

		}
	}


	private ConcurrentHashMap<UUID, Set<Integer>> getClusterToBlocks(ConcurrentHashMap<Integer, Set<UUID>> blockingBlockToClusters) {
		ConcurrentHashMap<UUID, Set<Integer>> ret = new ConcurrentHashMap<>();

		for (Entry<Integer, Set<UUID>> entry : blockingBlockToClusters.entrySet()) {
			for (UUID uuid : entry.getValue()) {
				CHMHelper.getFromCHM(ret, uuid, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(entry.getKey());
			}
		}
		return ret;
	}

	private Optional<Pair<Double, Cluster>> getMax(ClusterSimilarities similarities, Integer i, Collection<Cluster> clusters,
			ConcurrentHashMap<UUID, Cluster> blockingClusterMap, ConcurrentHashMap<Integer, Set<UUID>> blockingBlockToClusters, Set<UUID> except,
			Optional<Boolean> sequential, double threshold, double veryLowPairThreshold) {
		
//		ConcurrentLinkedQueue<Pair<Double,Integer>> done = new ConcurrentLinkedQueue<>();
		
		List<MutableReference<Pair<Double, Cluster>>> references =Collections.synchronizedList( new ArrayList<>());
		ThreadLocal<MutableReference<Pair<Double, Cluster>>> threadLocal = ThreadLocal.withInitial(() -> {
			MutableReference<Pair<Double, Cluster>> mr = new MutableReference<>(null);
			references.add(mr);
			return mr;
		});
		 
		Collection<Cluster> clustersVar;
		if (similarities.supportsBlocking()) {
			clustersVar = getClustersFromBlockMap(i, similarities, blockingBlockToClusters, blockingClusterMap);
		} else {
			clustersVar = clusters;
		}
		
		PersistentParallel6.runAndFinish(clustersVar, (cluster) -> {
//			int c= cluster.getIndex();
			
			if(except == null || !except.contains(cluster.key)) {
				Pair<Double, Double> quality = cluster.getQualityAndLowestScore(i, similarities);
//				double quality = 0.0;
//				for(Integer j : cluster.getValue()) {
//					Double sim = 1.0;
//					if(i != j) {
//						  sim = getScore(similarities, i, j);
//					}
//						if(sim != null)
//							quality += sim;
//					
//				}
				double adjustedThreshold = threshold * cluster.size();
				if (NO_THRESHOLD_ADJUSTMENT)
					adjustedThreshold = threshold;
				if (quality.getFirst() > adjustedThreshold || (quality.getSecond() >= veryLowPairThreshold)) {
					double adjustedQuality = quality.getFirst() - adjustedThreshold;
					if (NO_THRESHOLD_ADJUSTMENT)
						adjustedQuality = quality.getFirst();
					MutableReference<Pair<Double, Cluster>> mr = threadLocal.get();
					if (mr.get() == null || adjustedQuality > mr.get().getFirst()) {
						mr.set(new Pair<>(adjustedQuality, cluster));
					}
				}
				
				
				
			}
		}, this.getClass(), sequential.isPresent() ? sequential.get() : clusters.size() < 10000);
		
		
		Double maxQuality = null;
		Cluster maxCluster =null;
		
		for(MutableReference<Pair<Double, Cluster>> c : references) {
			if(maxQuality == null || c.get().getFirst() > maxQuality) {
				maxQuality = c.get().getFirst();
				maxCluster = c.get().getSecond();
			}
		}
		
		
		
//		Double[][] sims = new Double[clusters.size()][];
//		ConcurrentLinkedQueue<Pair<Integer,Integer>> queue = new ConcurrentLinkedQueue<>();
//
//		for(int c = 0; c < clusters.size(); c++) {
//			sims[c] = new Double[clusters.get(c).size()];
// 
//			for(Integer j : clusters.get(c)) {
//				queue.add(new Pair<>(c,j ));
//			}
//		}
		 
		
//		Double maxQuality = null;
//		int maxCluster =-1;
//		for(int c = 0; c < clusters.size(); c++) {
//			
//			if(except == null || !except.contains(c)) {
//				double quality = 0.0;
//				for(Integer j : clusters.get(c)) {
//					Double sim = 1.0;
//					if(i != j) {
//						  sim = getScore(similarities, i, j);
//					}
//						if(sim != null)
//							quality += sim;
//					
//				}
//				
//				if(maxQuality == null || quality > maxQuality) {
//					maxQuality = quality;
//					maxCluster = c;
//				}
//			}
//			
//		}
		
		if(maxQuality == null)
			return Optional.empty();
		return Optional.of(new Pair<>(maxQuality, maxCluster));
	}

	public static final int POINTS_CONCURRENT_FLOOR = 20;

	public static final int GET_MAX_CONCURRENT_FLOOR = POINTS_CONCURRENT_FLOOR;
	
	public static boolean WITH_CENTROIDS = false;
	
	
	private Pair<Set<UUID>, Map<Integer, List<Pair<UUID, Double>>>> getReassignedClustersNew(List<Cluster> reassignableClusters, List<Cluster> clusters,
			ClusterSimilarities similarities,
			boolean sequential, double threshold, double veryLowPairThreshold) {
		
		Set<UUID> reassignedCluster = new HashSet<>();
		 
		ThreadLocalList<List<Pair<Integer, UUID>>> list = new ThreadLocalList<>(); 
		PersistentParallel6.runAndFinish(reassignableClusters, (c) ->  {
			list.getOrCreate(() -> new ArrayList<>()).addAll(c.getPoints().stream().map((e) -> new Pair<>(e, c.getKey())).collect(Collectors.toList()));
		}, this.getClass(), sequential);
		
		List<Pair<Integer, UUID>> finalList = list.getCurrentList().stream().flatMap((e) -> e.stream()).collect(Collectors.toList());
		
		ConcurrentHashMap<Integer, List<Pair<UUID, Double>>> index = new ConcurrentHashMap<>();
		ConcurrentHashMap<UUID, Set<Integer>> reverseIndex = new ConcurrentHashMap<>();
		
		
		ConcurrentHashMap<UUID, Cluster> blockingClusterMap = getCHMFromClusterCollection(clusters);
		ConcurrentHashMap<Integer, Set<UUID>> blockingBlockToClusters = getBlockToClusterIndex(clusters, similarities);

		Set<UUID> notReassignable = Collections.newSetFromMap(new ConcurrentHashMap<>());

		PersistentParallel6.runAndFinish(finalList, (i) -> {

			if (!notReassignable.contains(i.getSecond())) {
				List<Pair<UUID, Double>> clusterSet = new ArrayList<>();
				Collection<Cluster> clustersVar;
				if (similarities.supportsBlocking()) {
					clustersVar = getClustersFromBlockMap(i.getFirst(), similarities, blockingBlockToClusters, blockingClusterMap);
				} else {
					clustersVar = clusters;

				}

				for (Cluster cluster : clustersVar) {

					if (cluster.getKey().equals(i.getSecond()))
						continue;
					// int c= cluster.getIndex();

					Pair<Double, Double> quality = cluster.getQualityAndLowestScore(i.getFirst(), similarities);

					double adjustedThreshold = threshold * cluster.size();
					if (NO_THRESHOLD_ADJUSTMENT)
						adjustedThreshold = threshold;
					if (quality.getFirst() > adjustedThreshold || ((quality.getSecond() >= veryLowPairThreshold))) {
						double adjustedQuality = quality.getFirst() - (threshold * cluster.size());
						if (NO_THRESHOLD_ADJUSTMENT)
							adjustedQuality = quality.getFirst();
						clusterSet.add(new Pair<>(cluster.key, adjustedQuality));
						CHMHelper.getFromCHM(reverseIndex, cluster.key, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(i.getFirst());
					}

					}
				if (!clusterSet.isEmpty()) {
					index.put(i.getFirst(), clusterSet);
				} else {
					notReassignable.add(i.getSecond());
				}
			}
		}, this.getClass(), "REASSIGN_LOOP_NEW", sequential);
		
		
		int reassignablePoints = index.size();
		
		Map<Integer, UUID> reassignment = new ConcurrentHashMap<>();
		ConcurrentHashMap<Integer, Set<UUID>> index2 = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(index.entrySet(), (entry) -> {

			Set<UUID> set = entry.getValue().stream().map((e) -> e.getFirst()).collect(Collectors.toSet());

			index2.put(entry.getKey(), set);
		}, this.getClass());

		for(Cluster cluster : reassignableClusters) {
			if (index2.isEmpty())
				break;
			
			boolean reassign = true;
			for(Integer point : cluster.getPoints()) {
				if (!index2.containsKey(point)) {
					reassign = false;
					break;
				}
			}
			
			if(reassign) {
				reassignedCluster.add(cluster.key);
				if(reverseIndex.containsKey(cluster.key)) {
					for(Integer point : reverseIndex.get(cluster.key)) {
						Set<UUID> set = index2.get(point);
						if(set != null) {
							if(set.size() > 1) 
								set.remove(cluster.key);
							else if(set.size() == 1) 
								index2.remove(point);
							
						}
					}
				}
				
			}
		}
		 
		return new Pair<>(reassignedCluster, index);
	}
	
	public static <E> Collection<E> getClustersFromBlockMap(Integer point, ClusterSimilarities similarities,
			Map<Integer, Set<UUID>> blockingBlockToClusters, Map<UUID, E> blockingClusterMap) {
		if (similarities.supportsBlocking()) {
			Map<UUID, MutableDouble> uuids = new HashMap<>();
			List<Tri<Set<Integer>, Integer, Double>> rowBlocks = similarities.getSplitBlocksWithScores(point);
			if (rowBlocks != null) {

				for (Tri<Set<Integer>, Integer, Double> blockPair : rowBlocks) {

					Map<UUID, MutableInt> innerUUIDs = new HashMap<>();

					for (Integer block : blockPair.getA()) {
						Set<UUID> blockSet = blockingBlockToClusters.get(block);
						if (blockSet != null) {
							for (UUID uuid : blockSet) {
								CHMHelper.getFromMap(innerUUIDs, uuid, () -> new MutableInt(0)).increment();
							}

						}
					}

					innerUUIDs.entrySet().stream().filter((e) -> e.getValue().intValue() >= blockPair.getB()).forEach((e) -> {
						CHMHelper.getFromMap(uuids, e.getKey(), () -> new MutableDouble(0.0))
								.add(blockPair.getC() * Math.min(1.0, e.getValue().doubleValue() / blockPair.getB()));

					});

				}
			}

			List<E> ret = uuids.entrySet().stream().filter((e) -> e.getValue().doubleValue() >= 1.3).map((e) -> blockingClusterMap.get(e.getKey()))
					.collect(Collectors.toList());

			return ret;
		} else {
			return blockingClusterMap.values();
		}
	}

	public static int minBlocks = 2;

	private Set<UUID> getReassignedClusters(List<Cluster> reassignableClusters, List<Cluster> clusters, ClusterSimilarities similarities, double threshold,
			double veryLowPairThreshold) {
		

		Map<Integer, Pair<Double, Cluster>> reassignments = new HashMap<>();
		
		Set<UUID> reassignedCluster = new HashSet<>();
		
		Collections.sort(reassignableClusters, (a,b) -> { return Integer.compare(a.size(), b.size());});
		double totalClusterSize = reassignableClusters.size();
		int clustersDone = 0;
		
		Optional<Boolean> innerSequential;
	
		boolean reassignAlwaysSequential = false;
		if(reassignableClusters.size() >= GET_MAX_CONCURRENT_FLOOR) {
			innerSequential = Optional.of(Boolean.FALSE);
			reassignAlwaysSequential = true;
		}
		else {
			innerSequential = Optional.empty();
		}
		for(ListEntry<Cluster> c : ListIterable.iterable(reassignableClusters)) {
			
			boolean reassignSequential = reassignAlwaysSequential; // || c.getValue().getPoints().size() < POINTS_CONCURRENT_FLOOR;
			
			Map<Integer, Pair<Double, Cluster>> reassignable = new ConcurrentHashMap<>();
			
			PersistentParallel6.runAndFinish( c.getValue().getPoints(), (i) -> {
				Set<UUID> except = new HashSet<>(reassignedCluster);
				except.add(c.getValue().getKey());
				Optional<Pair<Double, Cluster>> max = getMax(similarities, i, reassignableClusters, null, null, except, innerSequential, threshold,
						veryLowPairThreshold);
				
				if(max.isPresent() && max.get().getFirst() > 0) {
					reassignable.put(i, max.get());
				}
				
			}, "GreedyCluster-Reassign-Big-Loop", reassignSequential );
			
		 
			
			if(reassignable.size() == c.getValue().size()) {
				reassignedCluster.add(c.getValue().getKey());
				reassignments.putAll(reassignable);
			}
			
			clustersDone++; if(clustersDone % 10 == 0) {System.out.println("Finished " + DecimalFormatter.format(clustersDone / totalClusterSize) + " %");}

		}
		
		return reassignedCluster;
		
	}

	private List<Cluster> reassignReassignableClusters(Set<Integer> keySet, ClusterSimilarities similarities, List<Cluster> clusters, boolean sequential,
			double threshold, double veryLowPairThreshold) {

		int size = clusters.size(); 
		
//		clusters  = new ArrayList<>(clusters);
		double totalSize = keySet.size();
		List<Cluster> reassignableClusters = clusters.stream().filter((e) ->   e.size() / totalSize < 0.10 ).sorted((e1,e2) -> Integer.compare(e1.size(), e2.size())).collect(Collectors.toList());
		
		Pair<Set<UUID>, Map<Integer, List<Pair<UUID, Double>>>> reassignedClusterPair = getReassignedClustersNew(reassignableClusters, clusters, similarities,
				sequential, threshold, veryLowPairThreshold);
		Set<UUID> reassignedCluster = reassignedClusterPair.getFirst();
		
		Map<Integer, UUID> index = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(clusters, (cluster) -> {
			for (Integer point : cluster.getPoints()) {
				index.put(point, cluster.getKey());
			}
		}, this.getClass());
		
		Map<Integer, Cluster> newClusters = new HashMap<>();
		for(ListEntry<Cluster> c : ListIterable.iterable(clusters)) { 
			if(!reassignedCluster.contains(c.getValue().key)) {
				newClusters.put(c.getIndex(), c.getValue());
			}
			else {
				int sdf= 123;
			}
		}
		
		List<Cluster> 	newclustersList = new ArrayList<>(newClusters.values());
		
		double totalKeySetSize = keySet.size(); int done = 0; 
		ConcurrentHashMap<UUID, Set<Integer>> newClustersMap = new ConcurrentHashMap<>();
		
		ConcurrentHashMap<UUID, Cluster> blockingClusterMap = getCHMFromClusterCollection(newclustersList);
		ConcurrentHashMap<Integer, Set<UUID>> blockingBlockToClusters = getBlockToClusterIndex(newclustersList, similarities);

		Map<Integer, List<Pair<UUID, Double>>> reassignScores = reassignedClusterPair.getSecond();
		PersistentParallel6.runAndFinish(index.entrySet(), (iEntry) -> {
			Integer i = iEntry.getKey();

			if (false && !reassignedCluster.contains(iEntry.getValue())) {
				CHMHelper.getFromCHM(newClustersMap, iEntry.getValue(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(i);

			}
			else {
				UUID maxCluster = null;
				double maxScore = -1.0;

				List<Pair<UUID, Double>> reassignedScores = reassignScores.get(iEntry.getKey());
				if (reassignedScores != null) {
					for (Pair<UUID, Double> pair : reassignedScores) {
						if (pair.getSecond() > maxScore || maxCluster == null) {
							maxCluster = pair.getFirst();
							maxScore = pair.getSecond();
						}
					}
				}

				if (maxCluster == null) {
					Optional<Pair<Double, Cluster>> max = getMax(similarities, i, newclustersList, blockingClusterMap, blockingBlockToClusters, null,
							Optional.of(Boolean.TRUE), threshold, veryLowPairThreshold);
					if (max.isPresent()) {
						maxCluster = max.get().getSecond().getKey();
					} else {
						maxCluster = iEntry.getValue();
						// CHMHelper.getFromCHM(newClustersMap,
						// iEntry.getValue(), () ->
						// Collections.newSetFromMap(new
						// ConcurrentHashMap<>())).add(i);
						// throw new IllegalStateException();
					}
				}

				CHMHelper.getFromCHM(newClustersMap, maxCluster, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(i);

				if (false) {
					Optional<Pair<Double, Cluster>> max = getMax(similarities, i, newclustersList, blockingClusterMap, blockingBlockToClusters, null,
							Optional.of(Boolean.TRUE), threshold, veryLowPairThreshold);
					if (i.equals(199)) {
						int sdf = 234;
					}
					if (max.isPresent()) {
						CHMHelper.getFromCHM(newClustersMap, max.get().getSecond().getKey(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(i);
					} else {
						int fsdf = 123;
					}
				}
			}

//			done++; if(done % 2000 == 0) {System.out.println("Finished " + DecimalFormatter.format(done / totalKeySetSize) + " %");}
		}, this.getClass(), sequential || keySet.size() < POINTS_CONCURRENT_FLOOR);
		
		ConcurrentLinkedQueue<Cluster> queue = new ConcurrentLinkedQueue<>();
		
		PersistentParallel6.runAndFinish(newClustersMap.values(), (e) -> {
			queue.add(new Cluster(e, similarities, WITH_CENTROIDS));
		}, this.getClass(), "CreateClusters", sequential || newClustersMap.size() < 500);
		 
		
//		List<Cluster> finalClusterList =  newClustersMap.values().stream().map((e) -> new Cluster(e, similarities)).collect(Collectors.toList());
		List<Cluster> finalClusterList =  new ArrayList<>(queue);
		if(clusters.size() > 1) {
			int fsdf = 123;
		}
		int finalSum = finalClusterList.stream().mapToInt((e) -> e.size()).sum();
		return finalClusterList;
	}



	private ConcurrentHashMap<Integer, Set<UUID>> getBlockToClusterIndex(Collection<Cluster> newclustersList, ClusterSimilarities similarities) {
		ConcurrentHashMap<Integer, Set<UUID>> ret = new ConcurrentHashMap<>();

		if (similarities.supportsBlocking()) {
			PersistentParallel6.runAndFinish(newclustersList, (c) -> {
				Set<Integer> allBlocks = new HashSet<>();
				for (Integer point : c.getPoints()) {
					Set<Integer> blocks = similarities.getBlocks(point);
					if (point.equals(51)) {
						int df = 123;
					}
					if (blocks != null)
						allBlocks.addAll(blocks);
				}

				for (Integer block : allBlocks) {
					CHMHelper.getFromCHM(ret, block, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(c.getKey());
				}

			}, this.getClass(), newclustersList.size() < GET_MAX_CONCURRENT_FLOOR);

		}
		return ret;
	}

	private ConcurrentHashMap<UUID, Cluster> getCHMFromClusterCollection(List<Cluster> newclustersList) {
		return new ConcurrentHashMap<>(newclustersList.stream().collect(Collectors.toMap(Cluster::getKey, Function.identity())));
	}

	private Optional<Double> averageDissimilarity(ClusterSimilarities similarities, int i, Set<Integer> cluster) {
		double quality = 0.0;
		int count = 0;
		int countDis = 0;
		for(Integer j : cluster) {
			if(j != i) {
				double sim = getScore(similarities, i, j);
				 
					if(sim < 0) {
						quality +=  -1 * sim;
						countDis++;
					}
					count++;
				 
			}
		}
		
		if(countDis == 0 || count == 0) 
			return Optional.empty();
		else
			return Optional.of(quality / count);
	}
	
	private double getScore(ClusterSimilarities similarities, int a, int b) {
		 double ret = similarities.get(new ClusterEntryKey(a, b));

		 return ret;
	}
	
	
//	private volatile ExecutorService jobQueue;
	
	private   class Cluster {
		
		private final UUID key;
		
		ConcurrentHashMap<Integer, AtomicDouble> points = new ConcurrentHashMap<>();
		
		public final boolean withCentroids;
		public static final int maxInTopK = 20;
	 
		final Set<Integer> sortedSet;
		

		volatile Set<Integer> topK;
		
		private AtomicInteger changes = new AtomicInteger();
		
//		PriorityQueue<Integer> queue;
		
//		public Cluster(Collection<Integer> keys, ClusterSimilarities similarities) {
//			this(keys, similarities, true);
//		}
		public Cluster(UUID key, Collection<Integer> keys, ClusterSimilarities similarities, boolean withCentroids) {
			this.key = key;
			this.withCentroids = withCentroids;
			if(!withCentroids) {
				sortedSet = new ConcurrentSkipListSet<>();
				sortedSet.addAll(keys);
			}
			else {
				
				for(Integer i : keys) {
					  CHMHelper.getFromCHM(points, i, () -> new AtomicDouble()).addAndGet(1.0);
//					 points.put(i, 1.0);
				}
				for(DoubleEntry<Integer> doubleEntry : DoubleIteraterable.iterable(keys)) {
					double sim = 0.0;
					 Integer i = doubleEntry.getFirst();
					 Integer j = doubleEntry.getSecond();
						  sim = getScore(similarities, i, j);
						  CHMHelper.getFromCHM(points, i, () -> new AtomicDouble()).addAndGet(sim);
						  CHMHelper.getFromCHM(points, j, () -> new AtomicDouble()).addAndGet(sim); 
						  
				}
			 
		 
				sortedSet = createSet();
			sortedSet.addAll(points.keySet());
			
			this.topK = new HashSet<>(sortedSet);
			
//			sortedSet = new TreeSet<>((k1, k2) -> {
//				int compare = Double.compare(points.get(k1), points.get(k2));
//				
//				if(compare == 0) {
//					compare = Integer.compare(k1, k2);
//				}
//				
//				return compare;
//			});
//			
//			sortedSet = new HashSet<>();
//			
//			sortedSet.add(key);
			}

		}

		public Cluster(Collection<Integer> keys, ClusterSimilarities similarities, boolean withCentroids) {
			this(UUID.randomUUID(), keys, similarities, withCentroids);
		}
//		public Cluster(Integer key) {
//			this(key, true);
//		}
		public Cluster(Integer key, boolean withCentroids) {
			this.key = UUID.randomUUID();
			this.withCentroids = withCentroids;
			
			if(!withCentroids) {
				sortedSet = new ConcurrentSkipListSet<>();
				sortedSet.add(key);
			}
			else {
				
			
			

			  CHMHelper.getFromCHM(points, key, () -> new AtomicDouble(0.0)).addAndGet(1.0);
			sortedSet = createSet();
			sortedSet.add(key);
			 
			
			 Set<Integer> topKnew = new HashSet<>();
			 int f = 0;
			 for(Integer j : this.sortedSet ) {
					if(f >= maxInTopK) {
						break;
					}
				topKnew.add(j);
				f++;
			 }
			 this.topK = topKnew;
			
 
			} 
			
			
		}
		
		
		public int compare(Integer k1, Integer k2) {
			int compare = Double.compare(points.get(k2).doubleValue(), points.get(k1).doubleValue());
			
			if(compare == 0) {
				compare = Integer.compare(k2, k1);
			}
			
			return compare;
		}
		 
		
		UUID getKey() {
			return key;
		}

		
		public Set<Integer> createSet() {
			return new ConcurrentSkipListSet<>(this::compare);
		}


//		private void recreateSet() {
//			
////			sortedSet = new ConcurrentSkipListSet<>((k1, k2) -> {
////				int compare = Double.compare(points.get(k2).doubleValue(), points.get(k1).doubleValue());
////				
////				if(compare == 0) {
////					compare = Integer.compare(k2, k1);
////				}
////				
////				return compare;
////			});
//				
////			sortedSet.addAll(points.keySet());
//		}

		public Cluster(boolean withCentroids) {
			this.key = UUID.randomUUID();
			this.withCentroids = withCentroids;
			
			if(!withCentroids) {
				sortedSet = new HashSet<>();
			}
			else {
				sortedSet = createSet();
				
//				sortedSet = new TreeSet<>((k1, k2) -> {
//					int compare = Double.compare(points.get(k1), points.get(k2));
//					
//					if(compare == 0) {
//						compare = Integer.compare(k1, k2);
//					}
//					
//					return compare;
//				});
			}
			
//			queue = new PriorityQueue<>((k1, k2) -> {
//				int compare = Double.compare(points.get(k1), points.get(k2));
//				
//				if(compare == 0) {
//					compare = Integer.compare(k1, k2);
//				}
//				
//				if(compare == -1) {
//					return 1;
//				}
//				else if (compare == 1) {
//					return -1;
//				}
//				
//				return compare;
//			});
			
//			sortedSet = new HashSet<>();
		}

		public int size() {
			if(!withCentroids) {
			return sortedSet.size();
			}
			else {
				return points.size();
			}
		}

		public Set<Integer> getPoints() {
		
			if(!withCentroids) {
				return sortedSet;
				}
				else {
					return points.keySet();
				}
		}

		public Pair<Double, Double> getQualityAndLowestScore(int i, ClusterSimilarities similarities) {
			
			if(!withCentroids) {
				
				
				// ConcurrentAccumulator<AccumDouble> accumDouble = new
				// ConcurrentAccumulator< >(AccumDouble.empty());
				ConcurrentLinkedQueue<Double> list = new ConcurrentLinkedQueue<>();
				
//				double sum = 0.0;
				PersistentParallel6.runAndFinish(sortedSet, (j) -> {
//					for(Integer j : this.sortedSet ) {
//						if(f >= 20) {
//							break;
//						}
						
						double sim = 1.0;
						if(i != j) {
							  sim = getScore(similarities, i, j);
						list.add(sim);
						}
//						sum += sim;
					// accumDouble.add(new AccumDouble(sim));


						
//						f++;
//					}
				}, this.getClass(), sortedSet.size() < 40000);

				double sum = 0.0;
				double highest = -1.0;
				double lowest = 1.0;
				double sumNeg = 0.0;
				int countNeg = 0;
				double sumPos = 0.0;
				int countPos = 0;
				int count = 0;

				for (double sim : list) {
					sum += sim;
					count++;
					if (sim < 0.0) {
						sumNeg += sim;
						countNeg++;
					}
					if (sim > 0.0) {
						sumPos += sim;
						countPos++;
					}
					if (sim < lowest) {
						lowest = sim;
					}
					if (sim > highest) {
						highest = sim;
					}
				}

				if (count == 0) {
					return new Pair<>(0.0, lowest);
				}

				double averagePos = countPos > 0 ? sumPos / countPos : 0.0;
				double averageNeg = countNeg > 0 ? -1.0 * (sumNeg / countNeg) : 0.0;

				if (SUBTRACT_AVERAGE_DIFFERENCE) {
					sum += averagePos;
					sum -= averageNeg;
				}
				if (true) {
					return new Pair<>(sum, lowest);
				} else {
					// double averagePos = countPos > 0 ? sumPos / countPos :
					// 0.0;
					// double averageNeg = countNeg > 0 ? -1.0 * (sumNeg /
					// countNeg) : 0.0;

					double averageSum = sum / (count);

					if (averageSum > 0.0) {

						if (averagePos > averageNeg) {
							return new Pair<>(sum, lowest);
						} else if ((averageNeg - averagePos) < averageSum) {
							return new Pair<>(sum, lowest);
						} else {
							return new Pair<>(0.0, lowest);
						}

					} else if (averageSum < 0.0) {
						if (averageNeg > averagePos) {
							return new Pair<>(sum, lowest);
						} else if ((averagePos - averageNeg) < averageSum) {
							return new Pair<>(sum, lowest);
						} else {
							return new Pair<>(0.0, lowest);
						}
					}
					else {
						return new Pair<>(0.0, lowest);
					}
				}

				// if (sum < 0.0) {
				// if (averagePos > averageNeg) {
				// if (highest > lowest) {
				// return new Pair<>(0.01, lowest);
				// } else {
				// return new Pair<>(0.0, lowest);
				// }
				// } else {
				// return new Pair<>(sum, lowest);
				// }
				// } else if (sum > 0.0) {
				// if (averageNeg > averagePos) {
				// if (lowest > highest) {
				// return new Pair<>(-0.01, lowest);
				// } else {
				// return new Pair<>(0.0, lowest);
				// }
				// } else {
				// return new Pair<>(sum, lowest);
				// }
				// } else {
				//
				// return new Pair<>(sum, lowest);
				// }
			}
			else {
				int f = 0;
				double highest = -1.0;
				double lowest = 1.0;
				double sum = 0.0;
				double sumPos = 0.0;
				double sumNeg = 0.0;
				int countPos = 0;
				int countNeg = 0;
//				PersistentParallel6.runAndFinish(sortedSet, (j) -> {
					for(Integer j : this.topK ) {
//						if(f >= 1000) {
//							break;
//						}
						
						double sim = 1.0;
						if(i != j) {
							  sim = getScore(similarities, i, j);
						}
					if (sim < 0.0) {
						sumNeg += sim;
						countNeg++;
					}
					if (sim > 0.0) {
						sumPos += sim;
						countPos++;
					}
					if (sim < lowest) {
						lowest = sim;
					}
					if (sim > highest) {
						highest = sim;
					}
						sum += sim;
//						accumDouble.add(new AccumDouble(sim));
						
//						f++;
					}
//				}, this.getClass(), sortedSet.size() < 500);
				
				double averagePos = countPos > 0 ? sumPos / countPos : 0.0;
				double averageNeg = countNeg > 0 ? -1.0 * (sumNeg / countNeg) : 0.0;

				if (lowest < 0.0) {
					lowest = -1.0 * lowest;
				} else {
					lowest = 0.0;
				}

				if (highest < 0.0) {
					highest = 0.0;
				}

				if (sum < 0.0) {
					if (averagePos > averageNeg) {
						if (highest > lowest) {
							return new Pair<>(0.01, lowest);
						} else {
							return new Pair<>(0.0, lowest);
						}
					} else {
						return new Pair<>(sum, lowest);
					}
				} else if (sum > 0.0) {
					if (averageNeg > averagePos) {
						if (lowest > highest) {
							return new Pair<>(-0.01, lowest);
						} else {
							return new Pair<>(0.0, lowest);
						}
					} else {
						return new Pair<>(sum, lowest);
					}
				} else {

					return new Pair<>(sum, lowest);
				}
				// , highest, () , ( )

				// return new Quint<>(sum, lowest );
			}
//		
		}
		
		
		public synchronized  void recreateSet() {
			
			Set<Integer> copy = new HashSet<>(sortedSet);
			for(Integer key : copy) {
				sortedSet.remove(key);
				sortedSet.add(key);
			}
		 
			 Set<Integer> topKnew = new HashSet<>();
			 int f = 0;
			 for(Integer j : this.sortedSet ) {
					if(f >= maxInTopK) {
						break;
					}
				topKnew.add(j);
				f++;
			 }
			 this.topK = topKnew;
		}
		
		
		public   void addPoint(int i, ClusterSimilarities similarities, ExecutorService jobQueue) {
			if(!withCentroids) {
			sortedSet.add(i);
			}
			else {
				if (true)
					throw new IllegalStateException();
				 CHMHelper.getFromCHM(points, i, () -> new AtomicDouble(0.0)).addAndGet(1.0);
					double sum =0.0;
					for(Iterator<Integer> it = points.keySet().iterator(); it.hasNext(); ) {
						int j = it.next();
						
						
						
						Double sim = 1.0;
						if(i != j) {
							  sim = getScore(similarities, i, j);
							  

							  CHMHelper.getFromCHM(points, j, () -> new AtomicDouble()).addAndGet(sim);
							  
							  
							  sum += sim;
						}
						
						
						
//						entry.setValue(entry.getValue()+sim);
						
//						it.remove();
//						queue.add(j);
						
					}
					
					 
					 
					 CHMHelper.getFromCHM(points, i, () -> new AtomicDouble(0.0)).addAndGet(sum);
				
					 sortedSet.add(i);
					 
				 if(points.size() > maxInTopK) {
					 
					 int currentChange = changes.incrementAndGet();
					 if(currentChange > 0.10 * points.size()) {
						 currentChange = changes.getAndSet(0);
						 
						 if(currentChange > 0.10 * points.size()) {
							 jobQueue.submit(() -> {
								 recreateSet();
							 });
						 }
					 	
					 }
					 else {
						
						
					 }
					 
					 
					
					
				 }
				 else {
					 recreateSet();
				 }
				
			}
//			
//			changes++;
//			 points.put(i, sum);
//			 if(points.size() > 50) {
//				
//				 if(changes > 0.10 * points.size()) {
//					 changes = 0;
//				 	createSet();
//				 }
//			 }
//			 else {
//
//				 createSet();
//			 }

		}
		@Override
		public String toString() {
			return sortedSet.toString();
		}
		
		
		
	}
	
}
