package expansion.set.strategies.rowclusterers;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

import javax.inject.Inject;

import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.utils.kryo.CHMKryoAdapter;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusterSplitterNew;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.clustering.model.ClusterSimilaritiesImplCacheWrapper;
import expansion.set.clustering.model.ClusterSimilaritiesImplNormalized;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.exportimport.OptionalClusterCache;

public class GreedyClusteringSplitter implements RowClusterSplitterNew {

	public static final String DISABLE_BLOCKING = "DISABLE_BLOCKING";
	@Inject
	protected GreedyClustering clusterer;

	// @Inject
	// @Named(DISABLE_BLOCKING)
	// protected boolean disableBlocking;

	@Inject
	protected CHMKryoAdapter chmKryoAdapter;

	@Inject
	public GreedyClusteringSplitter() {
	}

	public ClusterSimilarities getCs(RowClusterSimilarities sim, RowClusteringConfig config, ClustererInput input, OptimizableClass optimizingClass) {
		RowCluster oldRc = new RowCluster(input.getClassId(), new ArrayList<>(input.getCandidates()), sim);

		return getCs(sim, config, input, optimizingClass, oldRc);
	}

	public ClusterSimilarities getCs(RowClusterSimilarities sim, RowClusteringConfig config, ClustererInput input, OptimizableClass optimizingClass,
			RowCluster oldRc) {
		ClusterSimilarities cs;
		if (optimizingClass == null) {
			cs = new ClusterSimilaritiesImplNormalized(oldRc, sim, config.isDisableBlocking(), 0.0);
		} else {
			Optional<Double> threshold = config.getWeightsRepository().get(optimizingClass).getNormalizingThreshold();

			if (threshold.isPresent()) {
				cs = new ClusterSimilaritiesImplNormalized(oldRc, sim, config.isDisableBlocking(), threshold.get());

			} else {
				cs = new ClusterSimilaritiesImplNormalized(oldRc, sim, config.isDisableBlocking(), 0.0);
			}
			// .getOrDefault(optimizingClass.getWeight("SIMILARITY_THRESHOLD",
			// 0.5));
			// if (threshold - 0.0 < 0.0001) {
			// List<RowCluster> list = new ArrayList<>();
			// list.add(oldRc);
			// return new ClusterList(list);
			// }

		}
		return cs;
	}

	public ClusterList clusterRows(RowClusterSimilarities sim, RowClusteringConfig config, ClustererInput input, OptimizableClass optimizingClass,
			boolean sequential) {
		return clusterRows(sim, config, input, optimizingClass, sequential, true);
	}


	public ClusterList clusterRows(RowClusterSimilarities sim, RowClusteringConfig config, ClustererInput input, OptimizableClass optimizingClass,
			boolean sequential, boolean disableBlocking) {
		return clusterRows(sim, config, input, optimizingClass, sequential, disableBlocking, null, null);
	}

	public ClusterList clusterRows(RowClusterSimilarities sim, RowClusteringConfig config, ClustererInput input, OptimizableClass optimizingClass,
			boolean sequential, boolean disableBlocking, OptionalClusterCache optionalClusterCache, Pair<String, String> preCacheFolderAndPrefix) {
		RowCluster oldRc = new RowCluster(input.getClassId(), new ArrayList<>(input.getCandidates()), sim);

		ClusterSimilarities cs = getCs(sim, config, input, optimizingClass, oldRc);
		if (optionalClusterCache != null) {
			cs = new ClusterSimilaritiesImplCacheWrapper(cs, oldRc, optionalClusterCache, chmKryoAdapter, preCacheFolderAndPrefix.getFirst(),
					preCacheFolderAndPrefix.getSecond());
		}
		List<RowCluster> list;
		if (optimizingClass == null) {
			list = clusterer.rowSplitter(oldRc, cs, sequential);
			;
		} else {
			Optional<Double> threshold = config.getWeightsRepository().get(optimizingClass).getClusteringThreshold();
			Optional<Double> veryLowPairThreshold = config.getWeightsRepository().get(optimizingClass).getVeryLowPairThreshold();
			if (threshold.isPresent()) {

				if (veryLowPairThreshold.isPresent()) {
					list = clusterer.rowSplitter(oldRc, cs, sequential, threshold.get(), -1.0 * veryLowPairThreshold.get());
				} else {
					list = clusterer.rowSplitter(oldRc, cs, sequential, threshold.get());
				}
			} else {

				list = clusterer.rowSplitter(oldRc, cs, sequential);

			}
		}


		return new ClusterList(list);
	}

	// public ClusterList clusterRows(RowClusterSimilarities sim,
	// RowClusteringConfig config, ClustererInput input, OptimizableClass
	// optimizingClass) {
	//
	// return clusterRows(sim, config, input, optimizingClass, true, true);
	//
	// }


	// public ClusterList clusterRows(RowClusterSimilarities sim,
	// RowClusteringConfig config, ClustererInput input) {
	// return clusterRows(sim, config, input, null);
	// }

}
