package expansion.set.strategies.rowclusterers;

import java.util.ArrayList;
import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusterer;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.similarities.RowClusterConstantSimilarities;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;

public class InstanceBasedClusterer implements RowClusterer {

	private final AnnotatedTablesRepository tablesRepository;
	
	
	@Inject
	public InstanceBasedClusterer(AnnotatedTablesRepository tablesRepository) {
		super();
		this.tablesRepository = tablesRepository;
	}

	public ClusterList clusterRowsInstanceBased(RowClusteringConfig config, ClustererInput input) {
		
		ConcurrentHashMap<InstanceId, ConcurrentLinkedQueue<TableRowId>> map = new ConcurrentHashMap<>();
		
		AtomicInteger ai = new AtomicInteger();
		
		PersistentParallel6.runAndFinish(input.getCandidates(), (tableRowId) -> {
			Optional<InstanceId> instanceId = tablesRepository.getMatchedInstanceId(tableRowId);
			
			if(instanceId.isPresent()) {
				CHMHelper.getFromCHM(map, instanceId.get(), () -> new ConcurrentLinkedQueue<>()).add(tableRowId);
			}
			else {
//				CHMHelper.getFromCHM(map, new InstanceId("INSTANCE_BASED_CLUSTER_CREATED_"+ai.getAndIncrement()), () -> new ConcurrentLinkedQueue<>()).add(tableRowId);
			}
		}, this.getClass());
	 
		return new ClusterList(map.values().stream().map((e) -> new RowCluster( getClassId(e)	, new ArrayList<>(e), new RowClusterConstantSimilarities(1.0))).collect(Collectors.toList()));
	}
	
	ClassId getClassId(ConcurrentLinkedQueue<TableRowId> e) {
		
		TableRowId tableRowId = e.peek();
		
		return tablesRepository.getClassId(tableRowId.getTableId());
		
	}

	@Override
	public RowClusterSimilarities clusterRows(RowClusteringConfig config, ClustererInput input) {
		// TODO Auto-generated method stub
		return null;
	}

}
