package expansion.set.strategies.rowclusterers.label;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;

import org.apache.commons.lang3.StringUtils;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.document.StoredField;
import org.apache.lucene.document.TextField;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.queryparser.classic.ParseException;
import org.apache.lucene.queryparser.classic.QueryParser;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.TopDocs;

import com.google.common.collect.ImmutableMap;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityMeasure;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.index.IIndex;
import de.mannheim.uni.index.io.DefaultIndex;
import de.mannheim.uni.index.io.InMemoryIndex;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.informatik.dws.expansion.matching.IndexBuilder;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.strategies.blockers.Blocks;
import expansion.set.strategies.blockers.Mappings;
import expansion.set.strategies.rowclusterers.ClusterSimilarities;
import expansion.set.strategies.rowclusterers.CreateClustersFromMappings;

public class LabelRowIndexClusterMapCreator {

	private final AnnotatedTablesRepository tableRepository;
	
//	private final AnnotatedTablesRepository tableRepository;
	private final SimilarityCalculator similarityCalculator;
	
	@Inject
	public LabelRowIndexClusterMapCreator(AnnotatedTablesRepository tablesRepository, SimilarityCalculator similarityCalculator) {
		this.tableRepository = tablesRepository;
		this.similarityCalculator = similarityCalculator;
	}
	
	private final static SimpleLockableConcurrentCache<UUID, Pair<Mappings, Mappings>> cache = new SimpleLockableConcurrentCache<>();
	
	public Mappings getClustersMappings(ClustererInput clustererInput, boolean blockByClass)   {
		return getClustersMappingsInner(clustererInput, blockByClass).getFirst();
		
	}
	
	public Pair<Mappings, Mappings> getClustersMappingsInner(ClustererInput clustererInput, boolean blockByClass)   {
		return cache.getOrCreate(clustererInput.getUuid(), () -> {
			try {
				return  getClustersMappings(clustererInput.getCandidates(), blockByClass);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
				return null;
			}
		});
		
	}
	
	public Pair<Mappings, Mappings>  getClustersMappings(Set<TableRowId> candidates, boolean blockByClass) throws IOException {
		
		
		
		 
		// MutableReference<IOException> exception = new
		// MutableReference<IOException>(null);
		LabelRowIndex labelRowIndex = new LabelRowIndex(tableRepository, candidates, 2);
		// LabelRowIndex labelRowIndex =
		// cacheIndex.getOrCreate(input.getUuid(), () -> {
		// try {
		// return new LabelRowIndex(tableRepository, candidates);
		// } catch (IOException e1) {
		// exception.set(e1);
		// return null;
		// }
		// });
		//
		// if (labelRowIndex == null) {
		// throw exception.get();
		// }

		SimilarityMeasure smSimilairty = similarityCalculator.stringSimilarityMixed;
//		SimilarityMeasure smClustering = similarityCalculator.getLabelBlockingSimilarity();
		SimilarityMeasure smClustering = similarityCalculator.getLabelBlockingSimilarity();
		ConcurrentHashMap<TableRowId, Set<TableRowId>> mappings = new ConcurrentHashMap<>();
		ConcurrentHashMap<TableRowId, TableRowId> oneToOneMapping = new ConcurrentHashMap<>();
		
//		Set<String> done = Collections.newSetFromMap(new ConcurrentHashMap<>());
		
		Map<String, TableRowId> done = new ConcurrentHashMap<>();
		

		ConcurrentHashMap<String, Set<TableRowId>> oneToOneBlocking = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(candidates, (tableRowId) -> {
			StringObject labelSO = getLabel(tableRepository, tableRowId); 
			
			//			TableRowId tableRowIdCheck = new TableRowId(new AnnotatedTableId("1346876860789_1346946768085_4249.arc4550171461300486933#66673628_0_6604262406652303152"), 1);
//			if(tableRowId.equals(tableRowIdCheck)) {
//				int sdf= 123;
//			}
//			if (tableRepository.getMetaData(tableRowId.getTableId()).getProperties().size() == 0)
//				return;
			if (labelSO != null) {
				
				String normalized = LabelRowIndex.normalise(labelSO.getString());
				AnnotatedTableMetaData md = tableRepository.getMetaData(tableRowId.getTableId());
				AnnotatedTable data = tableRepository.getTable(tableRowId.getTableId());
				ClassId classId = tableRepository.getMetaData(tableRowId.getTableId()).getDbPediaClass();


				CHMHelper.getFromCHM(oneToOneBlocking, normalized, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowId);
				
				if(done.containsKey(normalized)) {
					oneToOneMapping.put(tableRowId, done.get(normalized));
					return;
				}
				
				try {
//					List<String> finalMatchesForDebug = new ArrayList<>();
					
					 Set<TableRowId> finalRowIds = new HashSet<>();
					Map<StringObject, Set<TableRowId>> matches = labelRowIndex.getMatches(labelSO.getString());
					
					if(labelRowIndex.index.containsKey(normalized)) {
						finalRowIds.addAll(labelRowIndex.index.get(normalized));
//						matches.put(labelSO, labelRowIndex.index.get(normalized));
					}
					else {
						int sfffdf = 123;
					}
					int sdf = 123;
					for (Iterator<Entry<StringObject, Set<TableRowId>>> it = matches.entrySet().iterator(); it.hasNext();) {
						Entry<StringObject, Set<TableRowId>> next = it.next();
						
						StringObject combinedSO = null;
						StringObject deAccentedSO = null;
						StringObject combinedAndDeAccentedSO = null;
						
						Optional<String> otherLabel = IndexBuilder.lettersCombiner(labelSO.getString());
						if(otherLabel.isPresent()) {
							combinedSO = new StringObject(otherLabel.get());
							
							Optional<String> combinedAndDeAccented = IndexBuilder.deAccent(otherLabel.get());
							if(combinedAndDeAccented.isPresent() ) {
								combinedAndDeAccentedSO = new StringObject(combinedAndDeAccented.get());
							}
						}
						
						Optional<String> deAccented = IndexBuilder.deAccent(labelSO.getString());
						if(deAccented.isPresent() ) {
							deAccentedSO = new StringObject(deAccented.get());
						}
						
						if (!smClustering.isEqual(labelSO, next.getKey())) {
							
							if(combinedSO != null) {
								if (!smClustering.isEqual(combinedSO, next.getKey())) {
									it.remove();
								}
							}
							else if(deAccentedSO != null) {
								if (!smClustering.isEqual(deAccentedSO, next.getKey())) {
									it.remove();
								}
							}
							else if (combinedAndDeAccentedSO != null) {
								if (!smClustering.isEqual(combinedAndDeAccentedSO, next.getKey())) {
									it.remove();
								}
							}
							else {
								it.remove();
							}

							if (labelSO.getString().equals("compareLabel")) {
								int fasdf = 123;
							}
						}
						else {
							
							if(blockByClass) {
								if(next.getValue() == null) {
									int dfs = 123;
								}
								for(TableRowId candidate : next.getValue()) {
									 
									if( classId.equals(tableRepository.getMetaData(candidate.getTableId()).getDbPediaClass())) {
										finalRowIds.add(candidate);
									}
									else {
										int ssssdf =123;
									}
					  
								}
							}
							else {
								finalRowIds.addAll(next.getValue());
							}
							
							
						}
						
//						StringObject compareLabel = getLabel(tableRepository, next);
//						if (blockByClass && !classId.equals(tableRepository.getMetaData(next.getTableId()).getDbPediaClass())) {
//							it.remove();
//						} else if (!smClustering.isEqual(labelSO, compareLabel)) {
//							it.remove();
//
//							if (labelSO.getString().equals("compareLabel")) {
//								int fasdf = 123;
//							}
//						}
//						else {
//							finalMatchesForDebug.add(compareLabel.getString());
//						}

					}
					
					if(matches.isEmpty()) {
						int sdffff= 123;
					}
					
					if(finalRowIds.isEmpty()) {
						Set<TableRowId> concurrentSet = Collections.newSetFromMap(new ConcurrentHashMap<>());
						concurrentSet.add(tableRowId);
						mappings.put(tableRowId,concurrentSet );
					}
					else {
						List<String> lbs = new ArrayList<>();
						
						for(TableRowId match : finalRowIds) {
							StringObject so = getLabel(tableRepository, match);
							if(so != null)
								lbs.add(so.getString());
						}
						
						if(lbs.size() > 2000) {
							int sdgggg = 123;
						}
						boolean added = false;
//						for(TableRowId match : finalRowIds) {
//							if(mappings.containsKey(match)) {
//								Set<TableRowId> set = mappings.get(match);
//								set.addAll(finalRowIds);
//								set.add(tableRowId);
//								added = true;
//								break;
//							}
//						}
						
						if(!added) {
							Set<TableRowId> concurrentSet = Collections.newSetFromMap(new ConcurrentHashMap<>());
							concurrentSet.addAll(finalRowIds);
							mappings.put(tableRowId,concurrentSet );
						}
						
					}
					

					done.put(normalized, tableRowId);
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
//			else {
////				int sdfsf = 123;
////				System.out.println(tableRowId.toString());
//			}

		} , this.getClass(), "FindMappings");
		
		int maxSize =0;
//		for(Entry<TableRowId, Set<TableRowId>> mapEntry : mappings.entrySet()) {
//			if(mapEntry.getValue().size() > 1500) {
//				ClassId classId = tableRepository.getMetaData(mapEntry.getKey().getTableId()).getDbPediaClass();
//				
//				StringObject target = getLabel(tableRepository, mapEntry.getKey());
//				
//				List<String> labels = mapEntry.getValue().stream().map((e) -> {
//					StringObject so = getLabel(tableRepository, e);
//					if (so != null) {
//						return so.getString();
//					} else {
//						return "";
//					}
//					}).collect(Collectors.toList());
//				
//				maxSize = mapEntry.getValue().size();
//			}
//		}
		
		if(true && false) {
		
//		GreedyClustering gc = new GreedyClustering();
//		
//		ConcurrentHashMap<ClassId, Set<TableRowId>> classMap = new ConcurrentHashMap<>();
//		
//		BinarySimilarities bs = new BinarySimilarities(mappings, oneToOneMapping);
//		
//		PersistentParallel6.runAndFinish(candidates, (tableRowId) -> {
//			ClassId classId = 	tableRepository.getMetaData(tableRowId.getTableId()).getDbPediaClass(); ;
//			CHMHelper.getFromCHM(classMap, classId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowId);
//		}, this.getClass());
//		
//		
//		
//		Map<Integer, Set<TableRowId>> clustersMap = new HashMap<>();
//		Map<Integer, ClassId> clusterClassId = new HashMap<>();
//		
//		AtomicInteger ai = new AtomicInteger();
//		for(Entry<ClassId, Set<TableRowId>> entry : classMap.entrySet()) {
//			List<TableRowId> rowList = new ArrayList<>(entry.getValue());
//			RowCluster rc = new RowCluster(entry.getKey(), rowList, bs);
//			
//			
//			List<RowCluster> ret = gc.rowSplitter(rc, new BinaryClusterSimilarities(rowList, bs, entry.getKey()));
//			
//			for(RowCluster rcEntry : ret) {
//				int index = ai.getAndIncrement();
//				clustersMap.put(index, new HashSet<>(rcEntry.getRows()));
//				clusterClassId.put(index, entry.getKey());
//			}
//		}
//		return new Blocks(clustersMap, clusterClassId);
			return null;
		
	}
		
		else {
			
			
		Mappings oneToOne = new Mappings(oneToOneBlocking.values());
		Mappings blocking = new Mappings(mappings);
		return new Pair<>(blocking, oneToOne);
//		Map<Integer, Set<TableRowId>> clustersMap = new HashMap<>();
//		Map<Integer, ClassId> clusterClassId = new HashMap<>();
//		Map<TableRowId, Integer> clustersIndex = new HashMap<>();
//
//		CreateClustersFromMappings.addCluster(mappings, clustersMap, clustersIndex, clusterClassId, (tid) -> {
//			return tableRepository.getMetaData(tid).getDbPediaClass(); 
//		}, 
//				(rowId) -> {
//					StringObject so = getLabel(tableRepository, rowId);
//					if(so != null) {
//						return so.getString();
//					}
//					else {
//						return null;
//					}
//				});
//		
//		return new Blocks(clustersMap, clusterClassId);
		}
		
//		TableRowId tableRowIdCheck = new TableRowId(new AnnotatedTableId("1346876860789_1346946768085_4249.arc4550171461300486933#66673628_0_6604262406652303152"), 1);
//		Integer index = clustersIndex.get(tableRowIdCheck);
		
		
	}
	
	

	public Blocks  getClustersMap(Set<TableRowId> candidates, boolean blockByClass) throws IOException {
		Map<TableRowId, Set<TableRowId>> mappings = getClustersMappings(candidates, blockByClass).getFirst().getMappings();
		Map<Integer, Set<TableRowId>> clustersMap = new HashMap<>();
		Map<Integer, ClassId> clusterClassId = new HashMap<>();
		Map<TableRowId, Integer> clustersIndex = new HashMap<>();

		CreateClustersFromMappings.addCluster(mappings, clustersMap, clustersIndex, clusterClassId, (tid) -> {
			return tableRepository.getMetaData(tid).getDbPediaClass(); 
		}, 
				(rowId) -> {
					StringObject so = getLabel(tableRepository, rowId);
					if(so != null) {
						return so.getString();
					}
					else {
						return null;
					}
				});
		
		return new Blocks(clustersMap, clusterClassId);
	}
	
	
	public static StringObject getLabel(AnnotatedTablesRepository tablesRepository, TableRowId tableRowId) {
		AnnotatedTable table = tablesRepository.getTable(tableRowId.getTableId());
		ImmutableMap<RowIdx, AbstractObject> keyColumn = table.getKeyColumn();
		AbstractObject ao = keyColumn.get(tableRowId.getRowIdx());

		if (ao instanceof StringObject) {
			StringObject labelSO = (StringObject) ao;
			return labelSO;
		}

		String so = table.getStringTable().get(tableRowId.getRowIdx(), table.getKeyColumnIndex());
		if (so == null)
			return null;
		return new StringObject(so);
	}

	public static String normalise(String s, boolean order, boolean useStemmer) {
		List<String> list = StringNormalizer.tokenise(s, useStemmer);
		if(order)
			Collections.sort(list);
		return StringUtils.join(list, " ");

	}
	
	public static class LabelRowIndex {

		public static final String LABEL_STRING = "tableRowId";
		public static final String LABEL = "label";
		private static final String SEPERATOR_BIG = "::B::";

		private static final String SEPERATOR_SMALL = "::S::";

		private final IIndex idx;

		private final ThreadLocal<QueryParser> queryParserLocal = new ThreadLocal<>();

		private QueryParser getQueryParser() {
			QueryParser qp = queryParserLocal.get();
			if (qp == null) {
				qp = new QueryParser( LABEL, new StandardAnalyzer());
				queryParserLocal.set(qp);
			}
			return qp;
		}

		final ConcurrentHashMap<String, Set<TableRowId>> index = new ConcurrentHashMap<>();
		private final int maxEditDistance;
		private IndexSearcher indexSearcher;
		
		public static final String normalise(String string) {
			  string = StringNormalizer.normaliseValue(string, false);
			// normalises the value to improve lookup results
			string = LabelRowIndexClusterMapCreator.normalise(string, false, false);
			
			return string;
		}

		public LabelRowIndex( AnnotatedTablesRepository tableRepository, Set<TableRowId> candidates, int maxEditDistance) throws IOException {
			this(null, tableRepository, candidates, maxEditDistance);
		}
		private LabelRowIndex(String dir, AnnotatedTablesRepository tableRepository, Set<TableRowId> candidates, int maxEditDistance) throws IOException {
			if(dir != null) {
				this.idx = new DefaultIndex(dir);
			}
			else {
				this.idx = new InMemoryIndex();
			}
			this.maxEditDistance = maxEditDistance;
			IndexWriter w = idx.getIndexWriter();

			index.clear();

			PersistentParallel6.runAndFinish(candidates, (tableRowId) -> {
				StringObject labelSO = getLabel(tableRepository, tableRowId);
				if (labelSO != null) {
					String label = labelSO.getString();
label = normalise(label);

				
					CHMHelper.getFromCHM(index, label, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowId);
					
//					Optional<String> concatenatedLabel = IndexBuilder.concatenate(label);;
//					if(concatenatedLabel.isPresent()) {
//						CHMHelper.getFromCHM(index, concatenatedLabel.get(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowId);
//
//						Optional<String> deAccentedLabel = IndexBuilder.deAccent(concatenatedLabel.get());;
//						if(deAccentedLabel.isPresent()) {
//							CHMHelper.getFromCHM(index, deAccentedLabel.get(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowId);
//
//						}
//					}
					
					Optional<String> combinedLabel = IndexBuilder.lettersCombiner(label);;
					if(combinedLabel.isPresent()) {
						CHMHelper.getFromCHM(index, combinedLabel.get(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowId);

						Optional<String> deAccentedLabel = IndexBuilder.deAccent(combinedLabel.get());;
						if(deAccentedLabel.isPresent()) {
							CHMHelper.getFromCHM(index, deAccentedLabel.get(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowId);

						}
					}
					
					Optional<String> deAccentedLabel = IndexBuilder.deAccent(label);;
					if(deAccentedLabel.isPresent()) {
						CHMHelper.getFromCHM(index, deAccentedLabel.get(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowId);

					}
					

				}

			} , this.getClass());

			for (Entry<String, Set<TableRowId>> indexEntry : index.entrySet()) {
				Document doc = createDocument(indexEntry.getKey());

				w.addDocument(doc);
			}

			// for (TableRowId tableRowId : candidates) {
			//
			// StringObject labelSO = getLabel(tableRepository, tableRowId);
			// if (labelSO != null) {
			// String label = labelSO.getString();
			//
			// String labelClean = StringNormalizer.normaliseValue(label,
			// false);
			// // normalises the value to improve lookup results
			// label = LabelSimilarityRowClusterer.normalise(labelClean, true);
			//
			// Document doc = createDocument(label, tableRowId);
			//
			// w.addDocument(doc);
			// }
			// }

			idx.closeIndexWriter();
			
			this.indexSearcher = this.idx.getIndexSearcher();
		}

		private LabelRowIndex(String dir, int maxEditDistance) {
			this.maxEditDistance = maxEditDistance;
			this.idx = new DefaultIndex(dir);
			this.indexSearcher = this.idx.getIndexSearcher();
		}

		Map<StringObject, Set<TableRowId>> getMatches(String label) throws IOException, ParseException {
			try {

				Map<StringObject, Set<TableRowId>> instanceIds = new HashMap<>();
				String labelClean = StringNormalizer.normaliseValue(label, false);
				// normalises the value to improve lookup results
				label = LabelRowIndexClusterMapCreator.normalise(labelClean,true,  true);
				label = QueryParser.escape(label);
				if (label.length() > 0) {
					
					
					 List<String> tokens = StringNormalizer.tokenise(label, false);
		                
		                StringBuilder sb = new StringBuilder();
		                
		                for(String token : tokens) {
		                    sb.append(token);
		                    
		                    if(maxEditDistance>0) {
		                        sb.append("~");
		                        sb.append(maxEditDistance);
		                    }
		                    sb.append(" ");
		                }
		                
		                
		                label = sb.toString();
		                
		                
					Query q = getQueryParser().parse(label);
//					IndexSearcher indexSearcher = this.idx.getIndexSearcher();
					// new TermQuery(new Term(LABEL, search));
					TopDocs results = indexSearcher.search(q, 30);
					for (ScoreDoc doc : results.scoreDocs) {
						Document document = indexSearcher.doc(doc.doc);
						Pair<StringObject, Set<TableRowId>> ret = fromDocument(document);
						instanceIds.put(ret.getFirst(), ret.getSecond());
					}

					return instanceIds;
				}

			} catch (Exception e) {
				e.printStackTrace();
				System.out.println(e.getMessage());
				System.out.println("CAUGHT EXCEPTION");
			}

			return new HashMap<>();

		}

		public Pair<StringObject, Set<TableRowId>> fromDocument(Document doc) {
			// Set<TableRowId> rowId = new HashSet<>();
			String s = doc.getField(LABEL_STRING).stringValue();

			Set<TableRowId> rows = this.index.get(s);
			if (rows == null) {
				throw new IllegalStateException();
			}

			return new Pair<>(new StringObject(s), rows);
			//
			// String[] bigSplit = s.split(SEPERATOR_BIG);
			// for (String inner : bigSplit) {
			// String[] split = inner.split(SEPERATOR_SMALL);
			// if (split.length == 2) {
			// rowId.add(new TableRowId(new AnnotatedTableId(split[0]),
			// Integer.valueOf(split[1])));
			//
			// } else {
			// throw new IllegalStateException();
			// }
			// }

			// return rowId;

		}

		public Document createDocument(String label ) {
			Document doc = new Document();
			// String tableRowIdString = null;
			// for (TableRowId tableRowId : rowIds) {
			// if (tableRowIdString == null) {
			// tableRowIdString = "";
			// } else {
			// tableRowIdString += SEPERATOR_BIG;
			// }
			//
			// tableRowIdString += tableRowId.getTableId().getTableName() +
			// SEPERATOR_SMALL + tableRowId.getRowIdx().getRowIdx();
			// }

			// String tableRowIdString = tableRowId.getTableId().getTableName()
			// + SEPERATOR_BIG + tableRowId.getRowIdx().getRowIdx();
			doc.add(new StoredField(LABEL_STRING, label));
			doc.add(new TextField(LABEL, label, Field.Store.YES));

			return doc;
		}
	}
	
	
	public static class BinaryClusterSimilarities implements ClusterSimilarities {

		private final List<TableRowId> rows;
		private final BinarySimilarities bs;
		
		private final ClassId classId;
		
		public BinaryClusterSimilarities(List<TableRowId> rows, BinarySimilarities bs, ClassId classId) {
			super();
			this.rows = rows;
			this.bs = bs;
			this.classId = classId;
		}

		@Override
		public double get(ClusterEntryKey key) {
			return bs.get(rows, classId, key);
		}

		@Override
		public Set<Integer> regionQuery(Integer point, double eps) {
			// TODO Auto-generated method stub
			return null;
		}
		
	}
	public static class BinarySimilarities implements RowClusterSimilarities {

		Map<TableRowId, Set<TableRowId>> mapping;

		Map<TableRowId, TableRowId> oneToOneMapping;
		
		public BinarySimilarities(Map<TableRowId, Set<TableRowId>> mapping, Map<TableRowId, TableRowId> oneToOneMapping) {
			this.mapping = mapping;
			this.oneToOneMapping = oneToOneMapping;
		}
		@Override
		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			TableRowId row1 = rows.get(key.getE1());
			TableRowId row2 = rows.get(key.getE2());
			TableRowId o1 = oneToOneMapping.get(row1);
			TableRowId o2 = oneToOneMapping.get(row2);
			
			
			Set<TableRowId> m1 = mapping.get(row1);
			if(m1 != null && (m1.contains(row2) || (o2 != null && m1.contains(o2)))) {
				return 1.0;
			}
			 
			Set<TableRowId> m2 = mapping.get(row2);
			if(m2 != null && ( m2.contains(row1) || (o1 != null && m2.contains(o1)))) {
				return 1.0;
			}
			
			if(o1 != null) {
				  m1 = mapping.get(o1);
				if(m1 != null && (m1.contains(row2) || (o2 != null && m1.contains(o2)))) {
					return 1.0;
				}
			}
			 
			if(o2 != null) {
			m2 = mapping.get(o2);
			if(m2 != null && ( m2.contains(row1) || (o1 != null && m2.contains(o1)))) {
				return 1.0;
			}
			}
			 
			 
			return -0.25;
		}

		@Override
		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public String getName() {
			// TODO Auto-generated method stub
			return null;
		}
		
	}
}
