package expansion.set.strategies.rowclusterers.label;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicReference;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;
import org.apache.lucene.queryparser.classic.ParseException;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdsSet;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityMeasure;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.matching.LevelSteinSimilarityWithCache;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.matching.MatchingStringSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.stopwords.StopWordsRepository;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityMeasures;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThreshold;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThresholdImpl;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThresholdsRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.CHMHelper.Creator;
import de.mannheim.uni.expansion.utils.DoubleIteraterable;
import de.mannheim.uni.expansion.utils.DoubleIteraterable.DoubleEntry;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.mannheim.uni.index.io.StringNormaliser;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.matching.IndexBuilder;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;
import edu.stanford.nlp.util.StringUtils;
import expansion.set.clustering.ClusterList;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.similarities.RowClusterConstantSimilarities;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.profiling.GenericLabelRowIndexGeneric;
import expansion.set.strategies.newstrategies.BagOfWords_Component;
import expansion.set.strategies.rowclusterers.unary.ClusterCombiner;

public class LabelRowSimilaritiesCreator {

	
	private final AnnotatedTablesRepository tablesRepository;
	private final SimilarityCalculator similarityCalculator;
	
	@Inject
	StopWordsRepository stopWordsRepository;
	
	@Inject
	BagOfWords_Component BagOfWords_Component;
	
	@Inject
	public LabelRowSimilaritiesCreator(AnnotatedTablesRepository tablesRepository, SimilarityCalculator similarityCalculator) {
		super();
		this.tablesRepository = tablesRepository;
		this.similarityCalculator = similarityCalculator;
	}

	private final static AtomicReference<RowClusterSimilarities> normal = new AtomicReference<>();

	private final static SimpleLockableConcurrentCache<UuidStringId, RowClusterSimilarities> normalCache = new SimpleLockableConcurrentCache<>();

	private final static SimpleLockableConcurrentCache<UuidStringId, RowClusterSimilarities> normalWithoutCache = new SimpleLockableConcurrentCache<>();

	private final static SimpleLockableConcurrentCache<UuidStringId, RowClusterSimilarities> overlapCache = new SimpleLockableConcurrentCache<>();

	private final static SimpleLockableConcurrentCache<UuidStringId, RowClusterSimilarities> overlapWithoutCache = new SimpleLockableConcurrentCache<>();

	private final static SimpleLockableConcurrentCache<UuidStringId, RowClusterSimilarities> luceneCache = new SimpleLockableConcurrentCache<>();

	private final static SimpleLockableConcurrentCache<UuidStringId, RowClusterSimilarities> luceneWithoutCache = new SimpleLockableConcurrentCache<>();

	private final static SimpleLockableConcurrentCache<UuidStringId, RowClusterSimilarities> sfCache = new SimpleLockableConcurrentCache<>();

	private final static SimpleLockableConcurrentCache<UuidStringId, RowClusterSimilarities> sfWithoutCache = new SimpleLockableConcurrentCache<>();

	public RowClusterSimilarities create(TableRowIdsSet tableRowIdsSet, boolean withCache) {
		UuidStringId uuid = new UuidStringId("NULL");
		if (tableRowIdsSet != null) {
			uuid = tableRowIdsSet.getId();
		}

		if (withCache) {
		return normalCache.getOrCreateNoBlocking(uuid, () -> {
			return new CacheWrapper(new LabelSimilarities(tableRowIdsSet));
		});
		} else {
			return normalWithoutCache.getOrCreateNoBlocking(uuid, () -> {
				return new LabelSimilarities(tableRowIdsSet);
			});
		}

	}
	
	private final static SimpleLockableConcurrentCache<UuidStringId, RowClusterSimilarities> gsCache = new SimpleLockableConcurrentCache<>();

	public RowClusterSimilarities createForGs(TableRowIdsSet tableRowIdsSet) {
		UuidStringId uuid = new UuidStringId("NULL");
		if (tableRowIdsSet != null) {
			uuid = tableRowIdsSet.getId();
		}
		return gsCache.getOrCreateNoBlocking(uuid, () -> {
			return new CacheWrapper(new LabelSimilaritiesGs(tableRowIdsSet));
		});
	}

	public RowClusterSimilarities createLucene(TableRowIdsSet tableRowIdsSet, boolean withCache) {
		return createLucene(tableRowIdsSet, withCache, false);

	}

	public RowClusterSimilarities createLucene(TableRowIdsSet tableRowIdsSet, boolean withCache, boolean withSurfaceForms) {
		UuidStringId uuid = new UuidStringId("NULL");
		if (tableRowIdsSet != null) {
			uuid = tableRowIdsSet.getId();
		}

		if (withCache) {
			return luceneCache.getOrCreateNoBlocking(uuid, () -> {
				return new CacheWrapper(new LabelSimilaritiesLucene(tableRowIdsSet, withSurfaceForms));
			});
		} else {
			return luceneWithoutCache.getOrCreateNoBlocking(uuid, () -> {
				return new LabelSimilaritiesLucene(tableRowIdsSet, withSurfaceForms);
			});
		}

	}

	public RowClusterSimilarities createSurfaceForms(TableRowIdsSet tableRowIdsSet, boolean withCache) {
		UuidStringId uuid = new UuidStringId("NULL");
		if (tableRowIdsSet != null) {
			uuid = tableRowIdsSet.getId();
		}

		if (withCache) {
			return sfCache.getOrCreateNoBlocking(uuid, () -> {
				return new CacheWrapper(new LabelSimilaritiesRaw(true, false, false));
			});
		} else {
			return sfWithoutCache.getOrCreateNoBlocking(uuid, () -> {
				return new LabelSimilaritiesRaw(true, false, false);
			});
		}

	}

	public RowClusterSimilarities createOverlap(TableRowIdsSet tableRowIdsSet, boolean withCache) {
		UuidStringId uuid = new UuidStringId("NULL");
		if (tableRowIdsSet != null) {
			uuid = tableRowIdsSet.getId();
		}

		if (withCache) {
			return overlapCache.getOrCreateNoBlocking(uuid, () -> {
				return new CacheWrapper(new OverlapLabelSimilarities());
			});
		} else {
			return overlapWithoutCache.getOrCreateNoBlocking(uuid, () -> {
				return new OverlapLabelSimilarities();
			});
		}

	}

	public RowClusterSimilarities create(TableRowIdsSet tableRowIdsSet, boolean withSurfaceForms, boolean withStopWordsRemoval, boolean withTfIdf) {
		return new LabelSimilarities(tableRowIdsSet, withSurfaceForms, withStopWordsRemoval, withTfIdf);
	}
	
	private class CacheWrapper implements RowClusterSimilarities {

		private final RowClusterSimilarities inside;

		ConcurrentHashMap<TableRowIdPair, Double> cache = new ConcurrentHashMap<TableRowIdPair, Double>();

		public CacheWrapper(RowClusterSimilarities inside) {
			this.inside = inside;

		}

		@Override
		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			TableRowIdPair rowIdPair = new TableRowIdPair(rows.get(key.getE1()), rows.get(key.getE2()));

			return CHMHelper.getFromCHM(cache, rowIdPair, () -> inside.get(rows, classId, key));
		}

		@Override
		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getName() {
			return inside.getName();
		}

		@Override
		public boolean supportsBlocking() {
			return inside.supportsBlocking();
		}

		public int getBlockerSizes() {
			return inside.getBlockerSizes();
		}

		public List<Tri<Set<Integer>, Integer, Double>> getSplitBlocksWithScores(TableRowId tableRowId) {
			return inside.getSplitBlocksWithScores(tableRowId);
		}

		@Override
		public Set<Integer> getBlocks(TableRowId rowId) {

			return inside.getBlocks(rowId);
		}

		// @Override
		// public Set<Integer> getBlocks(TableRowId rowId) {
		// return inside.getBlocks(rowId);
		// }
		//
		// @Override
		// public boolean supportsBlocking() {
		// return inside.supportsBlocking();
		// }

	}

	public ClusterList getQuickClusterList(ClassId classId, Set<TableRowId> rowIds) {
		RowClusterConstantSimilarities sim = new RowClusterConstantSimilarities(1.0, "EXACT_LABEL");

		ConcurrentLinkedQueue<RowCluster> ret = new ConcurrentLinkedQueue<>();

		ConcurrentHashMap<String, Set<TableRowId>> blocks = new ConcurrentHashMap<>();

		AtomicInteger blockCounter = new AtomicInteger(0);
		PersistentParallel6.runAndFinish(rowIds, (rowId) -> {
			StringObject label = LabelRowIndexClusterMapCreator.getLabel(tablesRepository, rowId);
			if (label != null) {
				List<String> tokensList = StringNormaliser.tokenise(label.getString(), false);

				tokensList = tokensList.stream().map((e) -> StringNormalizer.normaliseValue(e, false)).collect(Collectors.toList());

				String joined = org.apache.commons.lang.StringUtils.join(tokensList, " ");
				if (joined.equals("")) {
					List<TableRowId> list = new ArrayList<>();
					ret.add(new RowCluster(classId, list, sim));
				} else {
					CHMHelper.getFromCHM(blocks, joined, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(rowId);

				}
			} else {
				List<TableRowId> list = new ArrayList<>();
				ret.add(new RowCluster(classId, list, sim));
			}
		}, this.getClass());

		PersistentParallel6.runAndFinish(blocks.entrySet(), (e) -> {
			List<TableRowId> list = new ArrayList<>(e.getValue());
			ret.add(new RowCluster(classId, list, sim));
		}, this.getClass());

		return new ClusterList(ret);
	}

	private class OverlapLabelSimilarities implements RowClusterSimilarities {

		@Override
		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			StringObject label1 = LabelRowIndexClusterMapCreator.getLabel(tablesRepository, rows.get(key.getE1()));
			StringObject label2 = LabelRowIndexClusterMapCreator.getLabel(tablesRepository, rows.get(key.getE2()));

			SimilarityMeasure sm;

			sm = similarityCalculator.stringSimilarity;

			if (label1 != null && label2 != null) {
				return sm.get(label1, label2);
			}
			return 0.0;
		}

		@Override
		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getName() {
			return "LABEL_OVERLAP";
		}

	}

	public static int numDocs = 5;

	private class LabelSimilaritiesLucene implements RowClusterSimilarities {

		private GenericLabelRowIndexGeneric<Integer> index;
		private Map<TableRowId, String> stringIndex;

		private final ConcurrentHashMap<String, List<Tri<Set<Integer>, Integer, Double>>> cache = new ConcurrentHashMap<>();;

		private final LabelSimilaritiesRaw raw;

		private final Map<Integer, String> intToStringMap;
		private final Map<Integer, Set<TableRowId>> blockToRowMap;

		public LabelSimilaritiesLucene(TableRowIdsSet tableRowIdsSet, boolean withSurfaceForms) {
			this.raw = new LabelSimilaritiesRaw(withSurfaceForms, false, false);

			ConcurrentHashMap<String, Map<Integer, Number>> index = new ConcurrentHashMap<>();

			ConcurrentHashMap<String, Integer> blockCounter = new ConcurrentHashMap<>();

			ConcurrentHashMap<TableRowId, String> rowIdToString = new ConcurrentHashMap<>();

			ConcurrentHashMap<Integer, Set<TableRowId>> blockToRowMap = new ConcurrentHashMap<>();
			AtomicInteger ai = new AtomicInteger(0);
			PersistentParallel6.runAndFinish(tableRowIdsSet.getRowIds(), (tableRowId) -> {

				StringObject label = LabelRowIndexClusterMapCreator.getLabel(tablesRepository, tableRowId);
				if (label == null || label.getString() == null) {
					int sdf = 123;
				} else {
					if (tableRowId.equals(new TableRowId("1346876860798_1346941853400_2953.arc2527404313287902461#59379225_0_6616355908335718299", 23))) {
						int sdf = 123;
					}
					Integer block = CHMHelper.getFromCHM(blockCounter, label.getString(), () -> ai.getAndIncrement()).intValue();
					CHMHelper.getFromCHM(blockToRowMap, block, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowId);
					rowIdToString.put(tableRowId, label.getString());
				}
			}, this.getClass());

			PersistentParallel6.runAndFinish(blockCounter.entrySet(), (entry) -> {
				CHMHelper.getFromCHM(index, entry.getKey(), () -> new ConcurrentHashMap<>()).put(entry.getValue(), 1.0);
			}, this.getClass());

			try {
				this.stringIndex = rowIdToString;
				this.intToStringMap = blockCounter.entrySet().stream().collect(Collectors.toMap(Entry::getValue, Entry::getKey));
				this.blockToRowMap = blockToRowMap;
				this.index = new GenericLabelRowIndexGeneric<>(index, 2, numDocs, true);

			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
				throw new RuntimeException(e);
			}
		}

		@Override
		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			return raw.get(rows, classId, key);
		}

		@Override
		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getName() {
			return "LABEL";
		}

		@Override
		public boolean supportsBlocking() {
			return true;
		}

		public int getBlockerSizes() {
			return 1;
		}

		@Override
		public Set<Integer> getBlocks(TableRowId rowId) {

			return getSplitBlocksWithScores(rowId).get(0).getA();
		}

		public List<Tri<Set<Integer>, Integer, Double>> getSplitBlocksWithScores(TableRowId tableRowId) {

			String string = stringIndex.get(tableRowId);

			Creator<List<Tri<Set<Integer>, Integer, Double>>> supplier = () -> {
				try {

					List<Tri<Set<Integer>, Integer, Double>> ret = new ArrayList<>();

					if (string == null) {
						ret.add(new Tri<>(new HashSet<>(), 1, 1.5));
					} else {
						Map<Integer, Float> rows = index.getMatches(string);

						Set<Integer> set = new HashSet<>();

						for (Entry<Integer, Float> row : rows.entrySet()) {
							set.add(row.getKey());
						}
						ret.add(new Tri<>(set, 1, 1.5));
					}
					return ret;

				} catch (IOException | ParseException e) {
					return null;
				}
			};
			if (true) {
				List<Tri<Set<Integer>, Integer, Double>> ret = supplier.create();

				// Set<TableRowId> rowIdSet = new HashSet<>();
				// rowIdSet.add(new
				// TableRowId("1346876860840_1346953393351_1858.arc4944083447630949323#2865484_1_8078897309408309233",
				// 12));
				// rowIdSet.add(new
				// TableRowId("1346823845675_1346865484759_551.arc5679908987767700891#21791343_2_1487834260780300373",
				// 5));
				// rowIdSet.add(new
				// TableRowId("1346876860445_1346962229057_555.arc690512194785518291#28168330_0_3483009587671943125",
				// 3));
				// rowIdSet.add(new
				// TableRowId("1346981172239_1346996411405_686.arc5220620847597988200#80178618_0_9094416607926131718",
				// 11));
				// rowIdSet.add(new
				// TableRowId("1346876860454_1346971822823_987.arc2109482148997547632#85409501_0_5286798261832614426",
				// 3));
				// rowIdSet.add(new
				// TableRowId("1346823846150_1346831690408_45.arc5994122906536653366#78012523_2_6448012290109416123",
				// 5));
				// rowIdSet.add(new
				// TableRowId("1346876860779_1346958288804_595.arc7823972158741128416#56229781_0_2382996291996445910",
				// 4));
				// rowIdSet.add(new
				// TableRowId("1346876860828_1346952706663_3429.arc5921241954962074354#20517249_1_9167235913169124029",
				// 12));
				// rowIdSet.add(new
				// TableRowId("1346981172155_1347001440797_1392.arc9090411764390287749#66457071_11_4481519565885296706",
				// 3));
				// rowIdSet.add(new
				// TableRowId("1346981172264_1347018377126_3653.arc5782822234765809404#30042358_0_8283684810574430634",
				// 10));
				// rowIdSet.add(new
				// TableRowId("1346981172234_1347006187111_1817.arc3336533571957179335#77829756_0_5604638202298041456",
				// 142));
				// rowIdSet.add(new
				// TableRowId("1346981172264_1347018690685_2692.arc5694695746453878202#5401079_0_644130583662511127",
				// 5));
				// rowIdSet.add(new
				// TableRowId("1346981172142_1347012114824_828.arc6735235393062592958#85134424_0_1746425791170541054",
				// 1));
				// rowIdSet.add(new
				// TableRowId("1346981172250_1346985391828_2366.arc2516789212534356496#82069893_0_103809824742711773",
				// 6));
				// rowIdSet.add(new
				// TableRowId("1346876860798_1346941853400_2953.arc2527404313287902461#59379225_0_6616355908335718299",
				// 23));
				//
				// Set<String> returnedStrings =
				// ret.get(0).getA().stream().map((e) ->
				// this.intToStringMap.get(e)).collect(Collectors.toSet());
				// Set<TableRowId> returnedRowids =
				// ret.get(0).getA().stream().flatMap((e) ->
				// this.blockToRowMap.get(e).stream()).collect(Collectors.toSet());
				//
				// if (rowIdSet.contains(tableRowId)) {
				// int sdf = 123;
				// }

				return ret;
			} else {
				return CHMHelper.getFromCHM(cache, string, supplier);
			}

		}

	}
	private class LabelSimilaritiesRaw implements RowClusterSimilarities {
		private final boolean withSurfaceForms;
		private final boolean withStopWordsRemoval;
		private final boolean withTfIdf;
		
		public LabelSimilaritiesRaw() {
			this.withSurfaceForms = false;
			this.withStopWordsRemoval = false;
			;
			this.withTfIdf = false;
		}

		public LabelSimilaritiesRaw(boolean withSurfaceForms, boolean withStopWordsRemoval, boolean withTfIdf) {
			super();
			this.withSurfaceForms = withSurfaceForms;
			this.withStopWordsRemoval = withStopWordsRemoval;
			this.withTfIdf = withTfIdf;
		}

		@Override
		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			StringObject label1 = LabelRowIndexClusterMapCreator.getLabel(tablesRepository, rows.get(key.getE1()));
			StringObject label2 = LabelRowIndexClusterMapCreator.getLabel(tablesRepository, rows.get(key.getE2()));

			Set<TableRowId> rowIds = new HashSet<>();

			rowIds.add(new TableRowId("1346876860877_1346943559923_2215.arc4004439957116847367#57748220_0_2612366228613617350", 4));
			rowIds.add(new TableRowId("1350433107002_1350494550274_543.arc4603560520287513930#55129959_0_5449848692245788620", 9));

			if (rowIds.contains(rows.get(key.getE1())) && rowIds.contains(rows.get(key.getE2()))) {
				int sdf = 13;
			}


			if (withStopWordsRemoval) {
				if (label1 != null)
					label1 = new StringObject(stopWordsRepository.removeStopWords(label1.getString()));
				if (label2 != null)
					label2 = new StringObject(stopWordsRepository.removeStopWords(label2.getString()));
			}

			SimilarityMeasure sm;
			if (withSurfaceForms) {
				sm = similarityCalculator.matchingStringSimilarityWithSurfaceFormsSimple;
			} else {
				sm = similarityCalculator.getMeasure(ValueType.STRING);
			}

			if (label1 != null && label2 != null) {
				return sm.get(label1, label2);
			}
			return 0.0;
		}

		@Override
		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getName() {
			return "LABEL_RAW" + (withSurfaceForms ? "_WITH_SURFACE_FORMS" : "");
		}

	}

	private class LabelSimilaritiesGs extends LabelSimilaritiesLucene {

		private final MatchingStringSimilarity sim;

		public LabelSimilaritiesGs(TableRowIdsSet tableRowIdsSet) {
			super(tableRowIdsSet, false);
			this.sim = new MatchingStringSimilarity(new SimilarityThresholdsRepository() {

				@Override
				public SimilarityThreshold get(SimilarityMeasures sm) {
					return new SimilarityThresholdImpl(0.0);
				}
			});
		}

		@Override
		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			StringObject label1 = LabelRowIndexClusterMapCreator.getLabel(tablesRepository, rows.get(key.getE1()));
			StringObject label2 = LabelRowIndexClusterMapCreator.getLabel(tablesRepository, rows.get(key.getE2()));

			// Set<TableRowId> rowIds = new HashSet<>();
			//
			// rowIds.add(new
			// TableRowId("1346876860445_1346962354404_1053.arc3561445202962273515#50446515_26_5175382634929694534",
			// 15));
			//
			// rowIds.add(new
			// TableRowId("1346876860789_1346944484018_327.arc2286104877941280553#57060121_1_7502732949961164628",
			// 26));
			//
			// if (rowIds.contains(rows.get(key.getE1())) &&
			// rowIds.contains(rows.get(key.getE2()))) {
			// int asdf = 123;
			// }
			if (label1 != null && label2 != null) {
				return sim.get(label1, label2);
			}
			return 0.0;
		}

		@Override
		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getName() {
			return "LABEL_GS";
		}

	}

	// private class LabelSimilarities implements RowClusterSimilarities {
	//
	// private final Map<TableRowId, Set<Integer>> rowToBlocks;
	// private final Map<TableRowId, Set<Integer>> rowToBlocksFull;
	//
	// private final LabelSimilaritiesRaw labelSimilaritiesRaw;
	//
	// public LabelSimilarities(TableRowIdsSet tableRowIdsSet) {
	// this.labelSimilaritiesRaw = new LabelSimilaritiesRaw();
	//
	// if (tableRowIdsSet == null) {
	// this.rowToBlocks = null;
	// this.rowToBlocksFull = null;
	// } else {
	// this.rowToBlocks = getBlocksNew(tableRowIdsSet.getRowIds());
	// this.rowToBlocksFull = getFullBlocks(tableRowIdsSet.getRowIds());
	// }
	// }
	//
	// public LabelSimilarities(TableRowIdsSet tableRowIdsSet, boolean
	// withSurfaceForms, boolean withStopWordsRemoval, boolean withTfIdf) {
	// this.labelSimilaritiesRaw = new LabelSimilaritiesRaw(withSurfaceForms,
	// withStopWordsRemoval, withTfIdf);
	//
	// this.rowToBlocks = getBlocksNew(tableRowIdsSet.getRowIds());
	// this.rowToBlocksFull = getFullBlocks(tableRowIdsSet.getRowIds());
	// }
	//
	// public int getBlockerSizes() {
	// return 2;
	// }
	//
	// private final AtomicInteger blockCounter = new AtomicInteger(0);
	// private ConcurrentHashMap<TableRowId, Set<Integer>> rowToBlocksForSizes;
	//
	// private Map<TableRowId, Set<Integer>> getFullBlocks(Set<TableRowId>
	// rowIds) {
	// ConcurrentHashMap<String, Integer> blocksToInteger = new
	// ConcurrentHashMap<>();
	// ConcurrentHashMap<TableRowId, Set<Integer>> rowToBlocks = new
	// ConcurrentHashMap<>();
	//
	//
	//
	// PersistentParallel6.runAndFinish(rowIds, (rowId) -> {
	// StringObject label =
	// LabelRowIndexClusterMapCreator.getLabel(tablesRepository, rowId);
	// if (label != null) {
	// Set<Integer> blocks = new HashSet<>();
	// List<String> tokensList = StringNormaliser.tokenise(label.getString(),
	// false);
	//
	// String normalized = StringUtils
	// .join(tokensList.stream().map((e) -> StringNormalizer.normaliseValue(e,
	// false)).sorted().collect(Collectors.toList()),
	// " ");
	//
	// Integer block = CHMHelper.getFromCHM(blocksToInteger, normalized, () ->
	// blockCounter.getAndIncrement());
	// blocks.add(block);
	//
	// rowToBlocks.put(rowId, Collections.unmodifiableSet(blocks));
	// }
	// }, this.getClass());
	//
	// return rowToBlocks;
	// }
	//
	// Map<TableRowId, Set<Integer>> getBlocks(Set<TableRowId> rowIds) {
	//
	// ConcurrentHashMap<String, Integer> blocksToInteger = new
	// ConcurrentHashMap<>();
	//
	// ConcurrentHashMap<TableRowId, Set<Integer>> rowToBlocks = new
	// ConcurrentHashMap<>();
	//
	// ConcurrentHashMap<Integer, AtomicInteger> blockCount = new
	// ConcurrentHashMap<>();
	//
	// AtomicInteger blockCounter = new AtomicInteger(0);
	// PersistentParallel6.runAndFinish(rowIds, (rowId) -> {
	// StringObject label =
	// LabelRowIndexClusterMapCreator.getLabel(tablesRepository, rowId);
	// if (label != null) {
	// Set<Integer> blocks = new HashSet<>();
	// List<String> tokensList = StringNormaliser.tokenise(label.getString(),
	// false);
	//
	// tokensList = tokensList.stream().map((e) ->
	// StringNormalizer.normaliseValue(e, false)).collect(Collectors.toList());
	//
	// Collections.sort(tokensList);
	//
	// if (tokensList.size() == 1) {
	// String normalized = tokensList.get(0);
	//
	// Integer block = CHMHelper.getFromCHM(blocksToInteger, normalized, () ->
	// blockCounter.getAndIncrement());
	// CHMHelper.getFromCHM(blockCount, block, () -> new
	// AtomicInteger()).incrementAndGet();
	// blocks.add(block);
	// }
	// else if (tokensList.size() > 1) {
	//
	// for (DoubleEntry<String> doubleEntry :
	// DoubleIteraterable.iterable(tokensList)) {
	//
	// String normalized = doubleEntry.getFirst() + " " +
	// doubleEntry.getSecond();
	//
	// Integer block = CHMHelper.getFromCHM(blocksToInteger, normalized, () ->
	// blockCounter.getAndIncrement());
	// CHMHelper.getFromCHM(blockCount, block, () -> new
	// AtomicInteger()).incrementAndGet();
	// blocks.add(block);
	// }
	// }
	//
	// rowToBlocks.put(rowId, Collections.unmodifiableSet(blocks));
	// }
	// }, this.getClass());
	//
	// ConcurrentHashMap<Integer, String> integerToBlock = new
	// ConcurrentHashMap<>();
	//
	// for (Entry<String, Integer> entry : blocksToInteger.entrySet()) {
	// integerToBlock.put(entry.getValue(), entry.getKey());
	// }
	//
	// List<Pair<String, Integer>> debugMap =
	// integerToBlock.keySet().stream().sorted((e1, e2) -> {
	// int compare = Integer.compare(blockCount.get(e2).intValue(),
	// blockCount.get(e1).intValue());
	// if (compare != 0) {
	// return compare;
	// }
	// return Integer.compare(e1, e2);
	// }).map((e) -> new Pair<>(integerToBlock.get(e),
	// blockCount.get(e).intValue())).collect(Collectors.toList());
	// return rowToBlocks;
	// }
	//
	//
	// Map<TableRowId, Set<Integer>> getBlocksNew(Set<TableRowId> rowIds) {
	//
	// ConcurrentHashMap<String, Integer> blocksToInteger = new
	// ConcurrentHashMap<>();
	// ConcurrentHashMap<String, Integer> longStringBlocksToInteger = new
	// ConcurrentHashMap<>();
	//
	// ConcurrentHashMap<TableRowId, Set<Integer>> rowToBlocks = new
	// ConcurrentHashMap<>();
	// ConcurrentHashMap<TableRowId, Set<Integer>> longStringBlocks = new
	// ConcurrentHashMap<>();
	//
	//
	//
	// PersistentParallel6.runAndFinish(rowIds, (rowId) -> {
	// StringObject label =
	// LabelRowIndexClusterMapCreator.getLabel(tablesRepository, rowId);
	// if (label != null) {
	// Set<Integer> blocks = new HashSet<>();
	// Set<Integer> longStringBlock = new HashSet<>();
	//
	// List<String> tokensList = StringNormaliser.tokenise(label.getString(),
	// false);
	//
	// tokensList = tokensList.stream().map((e) ->
	// StringNormalizer.normaliseValue(e, false)).collect(Collectors.toList());
	//
	// Collections.sort(tokensList);
	//
	// if (tokensList.size() == 1) {
	// String normalized = tokensList.get(0);
	//
	//
	//
	// if(normalized.length() > 4) {
	//
	// Integer block = CHMHelper.getFromCHM(longStringBlocksToInteger,
	// normalized, () -> blockCounter.getAndIncrement());
	//
	// longStringBlock.add(block);
	// }
	// else {
	// Integer block = CHMHelper.getFromCHM(blocksToInteger, normalized, () ->
	// blockCounter.getAndIncrement());
	//
	// blocks.add(block);
	// }
	// }
	// else if (tokensList.size() > 1) {
	//
	// for (String entry : tokensList) {
	//
	// String normalized = entry;
	//
	// if(normalized.length() > 4) {
	// Integer block = CHMHelper.getFromCHM(longStringBlocksToInteger,
	// normalized, () -> blockCounter.getAndIncrement());
	//
	// longStringBlock.add(block);
	// }
	// else {
	// Integer block = CHMHelper.getFromCHM(blocksToInteger, normalized, () ->
	// blockCounter.getAndIncrement());
	//
	// blocks.add(block);
	// }
	// }
	// }
	// if (!blocks.isEmpty()) {
	// rowToBlocks.put(rowId, Collections.unmodifiableSet(blocks));
	// }
	// if (!longStringBlock.isEmpty()) {
	// longStringBlocks.put(rowId,
	// Collections.unmodifiableSet(longStringBlock));
	// }
	//
	// }
	// }, this.getClass());
	//
	// Map<String, Map<Integer, Number>> inputMap =
	// longStringBlocksToInteger.entrySet().stream().collect(Collectors.toMap(Entry::getKey,
	// (e) -> {
	// Map<Integer, Number> map = new HashMap<>();
	// map.put(e.getValue(), 1.0);
	// return map;
	// }));
	//
	//
	// Map<Integer, Integer> mapping = getClustersMappings(inputMap,
	// longStringBlocksToInteger);
	//
	// PersistentParallel6.runAndFinish(longStringBlocks.entrySet(), (entry) ->
	// {
	// Set<Integer> newBlocks = new HashSet<>();
	// for (Integer block : entry.getValue()) {
	// newBlocks.add(mapping.get(block));
	// }
	// Set<Integer> set = CHMHelper.getFromCHM(rowToBlocks, entry.getKey(), ()
	// -> new HashSet<>());
	// set = new HashSet<>(set);
	// set.addAll(newBlocks);
	// if (set.contains(null)) {
	// int sdf = 123;
	// }
	// rowToBlocks.put(entry.getKey(), Collections.unmodifiableSet(set));
	//
	// }, this.getClass());
	//
	// ConcurrentHashMap<Integer, AtomicInteger> countMap = new
	// ConcurrentHashMap<>();
	//
	// PersistentParallel6.runAndFinish((rowToBlocks.entrySet()), (e) -> {
	// for (Integer i : e.getValue()) {
	// CHMHelper.getFromCHM(countMap, i, () -> new
	// AtomicInteger()).incrementAndGet();
	// }
	// }, this.getClass());
	//
	// Map<Integer, String> longIndex =
	// longStringBlocksToInteger.entrySet().stream().collect(Collectors.toMap(Entry::getValue,
	// Entry::getKey));
	// Map<Integer, String> shortIndex =
	// blocksToInteger.entrySet().stream().collect(Collectors.toMap(Entry::getValue,
	// Entry::getKey));
	//
	// List<Pair<String, Integer>> debugMap =
	// countMap.entrySet().stream().sorted((e1, e2) -> {
	// int compare = Integer.compare(e2.getValue().intValue(),
	// e1.getValue().intValue());
	// if (compare != 0) {
	// return compare;
	// }
	// return Integer.compare(e1.getKey(), e2.getKey());
	// }).map((e) -> {
	// String s = longIndex.get(e.getKey());
	// if (s == null)
	// s = shortIndex.get(e.getKey());
	// return new Pair<>(s, e.getValue().intValue());
	// }).collect(Collectors.toList());
	//
	// this.rowToBlocksForSizes = new ConcurrentHashMap<>(rowToBlocks);
	//
	// Set<Integer> problemBlocks = new HashSet<>();
	// double rowSize = rowIds.size();
	// for (Pair<String, Integer> debugPair : debugMap) {
	// if (debugPair.getSecond() / rowSize > 0.009) {
	// Integer block = longStringBlocksToInteger.get(debugPair.getFirst());
	// if (block == null) {
	// block = blocksToInteger.get(debugPair.getFirst());
	// }
	// problemBlocks.add(block);
	// }
	// }
	//
	// ConcurrentHashMap<Pair<Integer, Integer>, Integer> index = new
	// ConcurrentHashMap<>();
	//
	//
	// PersistentParallel6.runAndFinish(rowIds, (rowId) -> {
	//
	// Set<Integer> blocks = rowToBlocks.get(rowId);
	//
	// if (blocks != null) {
	// Set<Integer> problemBlock = new HashSet<>();
	// Set<Integer> nonProblemBlock = new HashSet<>();
	//
	// for (Integer block : blocks) {
	// if (problemBlocks.contains(block)) {
	// problemBlock.add(block);
	// } else {
	// nonProblemBlock.add(block);
	// }
	// }
	//
	// if (!problemBlock.isEmpty()) {
	// Set<Integer> newBlocks = new HashSet<>();
	//
	// for (Integer block : problemBlocks) {
	// for (Integer npBlock : nonProblemBlock) {
	// if (!npBlock.equals(block)) {
	// Integer newBlock = CHMHelper.getFromCHM(index, new Pair<>(block,
	// npBlock), () -> blockCounter.getAndIncrement());
	// newBlocks.add(newBlock);
	// }
	// }
	// }
	//
	// newBlocks.addAll(nonProblemBlock);
	// rowToBlocks.put(rowId, Collections.unmodifiableSet(newBlocks));
	//
	// }
	//
	// }
	// }, this.getClass());
	//
	// ConcurrentHashMap<Integer, Pair<Integer, Integer>> reverseIndex = new
	// ConcurrentHashMap<>();
	// for (Entry<Pair<Integer, Integer>, Integer> e : index.entrySet()) {
	// reverseIndex.put(e.getValue(), e.getKey());
	// }
	//
	// ConcurrentHashMap<Integer, AtomicInteger> countMap2 = new
	// ConcurrentHashMap<>();
	//
	// PersistentParallel6.runAndFinish((rowToBlocks.entrySet()), (e) -> {
	// for (Integer i : e.getValue()) {
	// CHMHelper.getFromCHM(countMap2, i, () -> new
	// AtomicInteger()).incrementAndGet();
	// }
	// }, this.getClass());
	//
	// List<Pair<String, Integer>> debugMap2 =
	// countMap2.entrySet().stream().sorted((e1, e2) -> {
	// int compare = Integer.compare(e2.getValue().intValue(),
	// e1.getValue().intValue());
	// if (compare != 0) {
	// return compare;
	// }
	// return Integer.compare(e1.getKey(), e2.getKey());
	// }).map((e) -> {
	// String s = longIndex.get(e.getKey());
	// if (s == null)
	// s = shortIndex.get(e.getKey());
	// if (s == null) {
	// Pair<Integer, Integer> combined = reverseIndex.get(e.getKey());
	// String s1 = longIndex.get(combined.getFirst());
	// if (s1 == null)
	// s1 = shortIndex.get(combined.getFirst());
	// String s2 = longIndex.get(combined.getSecond());
	// if (s2 == null)
	// s2 = shortIndex.get(combined.getSecond());
	// s = s1 + " " + s2;
	// }
	// return new Pair<>(s, e.getValue().intValue());
	// }).collect(Collectors.toList());
	// return rowToBlocks;
	// }
	//
	// public Map<Integer, Integer> getClustersMappings(Map<String, Map<Integer,
	// Number>> input, Map<String, Integer> map) {
	//
	// LevelSteinSimilarityWithCache sim = new LevelSteinSimilarityWithCache();
	// double threshold = 0.9;
	// Map<Integer, String> index =
	// map.entrySet().stream().collect(Collectors.toMap(Entry::getValue,
	// Entry::getKey));
	// GenericLabelRowIndexGeneric<Integer> labelRowIndex;
	// try {
	// labelRowIndex = new GenericLabelRowIndexGeneric<>(input, 2);
	//
	// // MutableReference<IOException> exception = new
	// // MutableReference<IOException>(null);
	// // LabelRowIndex labelRowIndex =
	// // cacheIndex.getOrCreate(input.getUuid(), () -> {
	// // try {
	// // return new LabelRowIndex(tableRepository, candidates);
	// // } catch (IOException e1) {
	// // exception.set(e1);
	// // return null;
	// // }
	// // });
	// //
	// // if (labelRowIndex == null) {
	// // throw exception.get();
	// // }
	//
	// SimilarityMeasure smSimilairty =
	// similarityCalculator.stringSimilarityMixed;
	// // SimilarityMeasure smClustering =
	// similarityCalculator.getLabelBlockingSimilarity();
	// SimilarityMeasure smClustering =
	// similarityCalculator.getLabelBlockingSimilarity();
	// ConcurrentHashMap<TableRowId, Set<TableRowId>> mappings = new
	// ConcurrentHashMap<>();
	// ConcurrentHashMap<TableRowId, TableRowId> oneToOneMapping = new
	// ConcurrentHashMap<>();
	//
	// // Set<String> done = Collections.newSetFromMap(new
	// // ConcurrentHashMap<>());
	//
	// Map<String, TableRowId> done = new ConcurrentHashMap<>();
	//
	// ConcurrentHashMap<Integer, Set<Integer>> mainMapping = new
	// ConcurrentHashMap<>();
	//
	// PersistentParallel6.runAndFinish(input.keySet(), (string) -> {
	//
	// // TableRowId tableRowIdCheck = new TableRowId(new
	// //
	// AnnotatedTableId("1346876860789_1346946768085_4249.arc4550171461300486933#66673628_0_6604262406652303152"),
	// // 1);
	// // if(tableRowId.equals(tableRowIdCheck)) {
	// // int sdf= 123;
	// // }
	// // if
	// //
	// (tableRepository.getMetaData(tableRowId.getTableId()).getProperties().size()
	// // == 0)
	// // return;
	// // if (labelSO != null) {
	//
	// StringObject labelSO = new StringObject(string);
	// Optional<String> deAccented = IndexBuilder.deAccent(labelSO.getString());
	// StringObject deAccentedSO = null;
	//
	// if (deAccented.isPresent()) {
	// deAccentedSO = new StringObject(deAccented.get());
	// }
	//
	// try {
	// // List<String> finalMatchesForDebug = new ArrayList<>();
	//
	// Map<Integer, Float> matches = labelRowIndex.getMatches(string);
	//
	// for (Iterator<Entry<Integer, Float>> it = matches.entrySet().iterator();
	// it.hasNext();) {
	// Entry<Integer, Float> next = it.next();
	//
	// String candidateString = index.get(next.getKey());
	//
	// StringObject candidateSO = new StringObject(candidateString);
	// StringObject deAccentedCS = null;
	// deAccented = IndexBuilder.deAccent(candidateString);
	// if (deAccented.isPresent()) {
	// deAccentedCS = new StringObject(deAccented.get());
	// }
	//
	// if (sim.calculate(candidateSO.getString(), labelSO.getString()) <
	// threshold) {
	//
	// if (deAccentedSO != null) {
	// if (deAccentedCS != null) {
	// if (sim.calculate(deAccentedSO.getString(), deAccentedCS.getString()) <
	// threshold) {
	// it.remove();
	// }
	// } else {
	// if (sim.calculate(deAccentedSO.getString(), candidateSO.getString()) <
	// threshold) {
	// it.remove();
	// }
	// }
	// } else if (deAccentedCS != null) {
	// if (sim.calculate(labelSO.getString(), deAccentedCS.getString()) <
	// threshold) {
	// it.remove();
	// }
	// }
	// else {
	// it.remove();
	// }
	//
	// } else {
	//
	//
	// }
	//
	// // StringObject compareLabel = getLabel(tableRepository,
	// // next);
	// // if (blockByClass &&
	// //
	// !classId.equals(tableRepository.getMetaData(next.getTableId()).getDbPediaClass()))
	// // {
	// // it.remove();
	// // } else if (!smClustering.isEqual(labelSO,
	// // compareLabel)) {
	// // it.remove();
	// //
	// // if (labelSO.getString().equals("compareLabel")) {
	// // int fasdf = 123;
	// // }
	// // }
	// // else {
	// // finalMatchesForDebug.add(compareLabel.getString());
	// // }
	//
	// }
	//
	// if (matches.isEmpty()) {
	// int sdffff = 123;
	// Set<Integer> equalBlocks = new HashSet<>();
	// equalBlocks.add(map.get(string));
	// mainMapping.put(map.get(string), equalBlocks);
	// } else {
	//
	// Set<String> debugSet = matches.keySet().stream().map((e) ->
	// index.get(e)).collect(Collectors.toSet());
	// Set<Integer> equalBlocks = new HashSet<>(matches.keySet());
	// equalBlocks.add(map.get(string));
	// if (debugSet.size() > 1) {
	// int df = 12;
	// }
	// mainMapping.put(map.get(string), equalBlocks);
	// }
	//
	// } catch (Exception e) {
	// // TODO Auto-generated catch block
	// e.printStackTrace();
	// }
	// // }
	// // else {
	// //// int sdfsf = 123;
	// //// System.out.println(tableRowId.toString());
	// // }
	//
	// }, this.getClass(), "FindMappings");
	//
	// ClusterCombiner<Integer> cc = new ClusterCombiner<>();
	//
	// Map<Integer, Integer> finalMapping = new HashMap<>();
	//
	// List<Set<Integer>> list = new ArrayList<>();
	// int largestSize = 0;
	// for (Entry<Integer, Set<Integer>> entry : mainMapping.entrySet()) {
	// Set<Integer> set = new HashSet<>();
	// set.add(entry.getKey());
	// set.addAll(entry.getValue());
	// list.add(set);
	// if (set.size() > largestSize) {
	// largestSize = set.size();
	// }
	// }
	//
	// cc.add(list);
	//
	// // Map<Integer, MutableInt> countMap =new HashMap<>();
	// for (Set<Integer> set : cc.get()) {
	// if (set.size() >= 1) {
	// Integer s = set.iterator().next();
	// for (Integer i : set) {
	// finalMapping.put(i, s);
	// }
	// if (set.size() > 5) {
	// List<String> debugList = set.stream().map((e) ->
	// index.get(e)).collect(Collectors.toList());
	//
	// int sdf = 123;
	// }
	// } else {
	// throw new IllegalStateException();
	// }
	// }
	//
	// return finalMapping;
	//
	// } catch (IOException e1) {
	// // TODO Auto-generated catch block
	// e1.printStackTrace();
	// return null;
	// }
	// }
	//
	// public List<Tri<Set<Integer>, Integer, Double>>
	// getSplitBlocksWithScores(TableRowId tableRowId) {
	// List<Tri<Set<Integer>, Integer, Double>> list = new ArrayList<>();
	// Set<Integer> blocksApproximate = rowToBlocks.get(tableRowId);
	// Set<Integer> blocksFull = rowToBlocksFull.get(tableRowId);
	// if (blocksApproximate == null)
	// blocksApproximate = new HashSet<>();
	// if (blocksFull == null)
	// blocksFull = new HashSet<>();
	// list.add(new Tri<>(blocksFull, 1, 1.55));
	// list.add(new Tri<>(blocksApproximate, getBlockMin(tableRowId), 1.55));
	// return list;
	// }
	//
	//
	// @Override
	// public Set<Integer> getBlocks(TableRowId rowId) {
	//
	// Set<Integer> ret = new HashSet<>();
	//
	// Set<Integer> blocksApproximate = rowToBlocks.get(rowId);
	// Set<Integer> blocksFull = rowToBlocksFull.get(rowId);
	// if (blocksApproximate != null) {
	// ret.addAll(blocksApproximate);
	// }
	// if (blocksFull != null) {
	// ret.addAll(blocksFull);
	// }
	//
	// return ret;
	// }
	//
	// @Override
	// public int getBlockMin(TableRowId tableRowId) {
	// Set<Integer> blocks = rowToBlocksForSizes.get(tableRowId);
	// if (blocks != null) {
	// int size = rowToBlocksForSizes.get(tableRowId).size();
	// int ret;
	// if (size == 1) {
	// ret = 1;
	// } else {
	// ret = (int) (size * 0.85);
	// }
	// return ret;
	// } else {
	// return 1;
	// }
	// }
	//
	// @Override
	// public boolean supportsBlocking() {
	// return rowToBlocks != null;
	// }
	//
	//
	// @Override
	// public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey
	// key) {
	// return labelSimilaritiesRaw.get(rows, classId, key);
	// }
	//
	// @Override
	// public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point,
	// double eps) {
	// throw new UnsupportedOperationException();
	// }
	//
	// @Override
	// public String getName() {
	// return "LABEL";
	// }
	//
	// }

	private class LabelSimilarities implements RowClusterSimilarities {

		private final Map<TableRowId, Set<Integer>> rowToBlocks;
		private final Map<TableRowId, Set<Integer>> rowToBlocksFull;

		private final LabelSimilaritiesRaw labelSimilaritiesRaw;

		public LabelSimilarities(TableRowIdsSet tableRowIdsSet) {
			this.labelSimilaritiesRaw = new LabelSimilaritiesRaw();

			if (tableRowIdsSet == null) {
				this.rowToBlocks = null;
				this.rowToBlocksFull = null;
			} else {
				this.rowToBlocks = getBlocksNew(tableRowIdsSet.getRowIds());
				this.rowToBlocksFull = getFullBlocks(tableRowIdsSet.getRowIds());
			}
		}
		
		public LabelSimilarities(TableRowIdsSet tableRowIdsSet, boolean withSurfaceForms, boolean withStopWordsRemoval, boolean withTfIdf) {
			this.labelSimilaritiesRaw = new LabelSimilaritiesRaw(withSurfaceForms, withStopWordsRemoval, withTfIdf);

			this.rowToBlocks = getBlocksNew(tableRowIdsSet.getRowIds());
			this.rowToBlocksFull = getFullBlocks(tableRowIdsSet.getRowIds());
		}

		public int getBlockerSizes() {
			return 2;
		}

		private final AtomicInteger blockCounter = new AtomicInteger(0);
		private ConcurrentHashMap<TableRowId, Set<Integer>> rowToBlocksForSizes;

		private Map<TableRowId, Set<Integer>> getFullBlocks(Set<TableRowId> rowIds) {
			ConcurrentHashMap<String, Integer> blocksToInteger = new ConcurrentHashMap<>();
			ConcurrentHashMap<TableRowId, Set<Integer>> rowToBlocks = new ConcurrentHashMap<>();



			PersistentParallel6.runAndFinish(rowIds, (rowId) -> {
				StringObject label = LabelRowIndexClusterMapCreator.getLabel(tablesRepository, rowId);
				if (label != null) {
					Set<Integer> blocks = new HashSet<>();
					List<String> tokensList = StringNormaliser.tokenise(label.getString(), false);

					String normalized = StringUtils
							.join(tokensList.stream().map((e) -> StringNormalizer.normaliseValue(e, false)).sorted().collect(Collectors.toList()),
							" ");

					Integer block = CHMHelper.getFromCHM(blocksToInteger, normalized, () -> blockCounter.getAndIncrement());
					blocks.add(block);

					rowToBlocks.put(rowId, Collections.unmodifiableSet(blocks));
				}
			}, this.getClass());

			return rowToBlocks;
		}

		Map<TableRowId, Set<Integer>> getBlocks(Set<TableRowId> rowIds) {

			ConcurrentHashMap<String, Integer> blocksToInteger = new ConcurrentHashMap<>();

			ConcurrentHashMap<TableRowId, Set<Integer>> rowToBlocks = new ConcurrentHashMap<>();

			ConcurrentHashMap<Integer, AtomicInteger> blockCount = new ConcurrentHashMap<>();

			AtomicInteger blockCounter = new AtomicInteger(0);
			PersistentParallel6.runAndFinish(rowIds, (rowId) -> {
				StringObject label = LabelRowIndexClusterMapCreator.getLabel(tablesRepository, rowId);
				if (label != null) {
					Set<Integer> blocks = new HashSet<>();
					List<String> tokensList = StringNormaliser.tokenise(label.getString(), false);

					tokensList = tokensList.stream().map((e) -> StringNormalizer.normaliseValue(e, false)).collect(Collectors.toList());

					Collections.sort(tokensList);

					if (tokensList.size() == 1) {
						String normalized = tokensList.get(0);

						Integer block = CHMHelper.getFromCHM(blocksToInteger, normalized, () -> blockCounter.getAndIncrement());
						CHMHelper.getFromCHM(blockCount, block, () -> new AtomicInteger()).incrementAndGet();
						blocks.add(block);
					}
					else if (tokensList.size() > 1) {

						for (DoubleEntry<String> doubleEntry : DoubleIteraterable.iterable(tokensList)) {

							String normalized = doubleEntry.getFirst() + " " + doubleEntry.getSecond();

							Integer block = CHMHelper.getFromCHM(blocksToInteger, normalized, () -> blockCounter.getAndIncrement());
							CHMHelper.getFromCHM(blockCount, block, () -> new AtomicInteger()).incrementAndGet();
							blocks.add(block);
						}
					}

					rowToBlocks.put(rowId, Collections.unmodifiableSet(blocks));
				}
			}, this.getClass());

			ConcurrentHashMap<Integer, String> integerToBlock = new ConcurrentHashMap<>();

			for (Entry<String, Integer> entry : blocksToInteger.entrySet()) {
				integerToBlock.put(entry.getValue(), entry.getKey());
			}

			List<Pair<String, Integer>> debugMap = integerToBlock.keySet().stream().sorted((e1, e2) -> {
				int compare = Integer.compare(blockCount.get(e2).intValue(), blockCount.get(e1).intValue());
				if (compare != 0) {
					return compare;
				}
				return Integer.compare(e1, e2);
			}).map((e) -> new Pair<>(integerToBlock.get(e), blockCount.get(e).intValue())).collect(Collectors.toList());
			return rowToBlocks;
		}
		
		
		Map<TableRowId, Set<Integer>> getBlocksNew(Set<TableRowId> rowIds) {

			ConcurrentHashMap<String, Integer> blocksToInteger = new ConcurrentHashMap<>();
			ConcurrentHashMap<String, Integer> longStringBlocksToInteger = new ConcurrentHashMap<>();

			ConcurrentHashMap<TableRowId, Set<Integer>> rowToBlocks = new ConcurrentHashMap<>();
			ConcurrentHashMap<TableRowId, Set<Integer>> longStringBlocks = new ConcurrentHashMap<>();
 


			PersistentParallel6.runAndFinish(rowIds, (rowId) -> {
				StringObject label = LabelRowIndexClusterMapCreator.getLabel(tablesRepository, rowId);
				if (label != null) {
					Set<Integer> blocks = new HashSet<>();
					Set<Integer> longStringBlock = new HashSet<>();
					
					List<String> tokensList = StringNormaliser.tokenise(label.getString(), false);

					tokensList = tokensList.stream().map((e) -> StringNormalizer.normaliseValue(e, false)).collect(Collectors.toList());

					Collections.sort(tokensList);

					if (tokensList.size() == 1) {
						String normalized = tokensList.get(0);


 
						if(normalized.length() > 4) {
							
							Integer block = CHMHelper.getFromCHM(longStringBlocksToInteger, normalized, () -> blockCounter.getAndIncrement()); 

							longStringBlock.add(block);
						}
						else {
							Integer block = CHMHelper.getFromCHM(blocksToInteger, normalized, () -> blockCounter.getAndIncrement()); 

							blocks.add(block);
						}
					}
					else if (tokensList.size() > 1) {

						for (String entry : tokensList) {

							String normalized = entry;

							if(normalized.length() > 4) {
								Integer block = CHMHelper.getFromCHM(longStringBlocksToInteger, normalized, () -> blockCounter.getAndIncrement()); 

								longStringBlock.add(block);
							}
							else {
								Integer block = CHMHelper.getFromCHM(blocksToInteger, normalized, () -> blockCounter.getAndIncrement()); 

								blocks.add(block);
							}
						}
					}
					if (!blocks.isEmpty()) {
						rowToBlocks.put(rowId, Collections.unmodifiableSet(blocks));
					}
					if (!longStringBlock.isEmpty()) {
						longStringBlocks.put(rowId, Collections.unmodifiableSet(longStringBlock));
					}

				}
			}, this.getClass());
			
			Map<String, Map<Integer, Number>> inputMap = longStringBlocksToInteger.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				Map<Integer, Number> map = new HashMap<>();
				map.put(e.getValue(), 1.0);
				return map;
			}));

 
			Map<Integer, Integer> mapping = getClustersMappings(inputMap, longStringBlocksToInteger);

			PersistentParallel6.runAndFinish(longStringBlocks.entrySet(), (entry) -> {
				Set<Integer> newBlocks = new HashSet<>();
				for (Integer block : entry.getValue()) {
					newBlocks.add(mapping.get(block));
				}
				Set<Integer> set = CHMHelper.getFromCHM(rowToBlocks, entry.getKey(), () -> new HashSet<>());
				set = new HashSet<>(set);
				set.addAll(newBlocks);
				if (set.contains(null)) {
					int sdf = 123;
				}
				rowToBlocks.put(entry.getKey(), Collections.unmodifiableSet(set));

			}, this.getClass());

			ConcurrentHashMap<Integer, AtomicInteger> countMap = new ConcurrentHashMap<>();

			PersistentParallel6.runAndFinish((rowToBlocks.entrySet()), (e) -> {
				for (Integer i : e.getValue()) {
					CHMHelper.getFromCHM(countMap, i, () -> new AtomicInteger()).incrementAndGet();
				}
			}, this.getClass());

			Map<Integer, String> longIndex = longStringBlocksToInteger.entrySet().stream().collect(Collectors.toMap(Entry::getValue, Entry::getKey));
			Map<Integer, String> shortIndex = blocksToInteger.entrySet().stream().collect(Collectors.toMap(Entry::getValue, Entry::getKey));

			List<Pair<String, Integer>> debugMap = countMap.entrySet().stream().sorted((e1, e2) -> {
				int compare = Integer.compare(e2.getValue().intValue(), e1.getValue().intValue());
				if (compare != 0) {
					return compare;
				}
				return Integer.compare(e1.getKey(), e2.getKey());
			}).map((e) -> {
				String s = longIndex.get(e.getKey());
				if (s == null)
					s = shortIndex.get(e.getKey());
				return new Pair<>(s, e.getValue().intValue());
			}).collect(Collectors.toList());

			this.rowToBlocksForSizes = new ConcurrentHashMap<>(rowToBlocks);

			Set<Integer> problemBlocks = new HashSet<>();
			double rowSize = rowIds.size();
			for (Pair<String, Integer> debugPair : debugMap) {
				if (debugPair.getSecond() / rowSize > 0.009) {
					Integer block = longStringBlocksToInteger.get(debugPair.getFirst());
					if (block == null) {
						block = blocksToInteger.get(debugPair.getFirst());
					}
					problemBlocks.add(block);
				}
			}

			ConcurrentHashMap<Pair<Integer, Integer>, Integer> index = new ConcurrentHashMap<>();


			PersistentParallel6.runAndFinish(rowIds, (rowId) -> {

				Set<Integer> blocks = rowToBlocks.get(rowId);

				if (blocks != null) {
					Set<Integer> problemBlock = new HashSet<>();
					Set<Integer> nonProblemBlock = new HashSet<>();

					for (Integer block : blocks) {
						if (problemBlocks.contains(block)) {
							problemBlock.add(block);
						} else {
							nonProblemBlock.add(block);
						}
					}

					if (!problemBlock.isEmpty()) {
						Set<Integer> newBlocks = new HashSet<>();

						for (Integer block : problemBlocks) {
							for (Integer npBlock : nonProblemBlock) {
								if (!npBlock.equals(block)) {
									Integer newBlock = CHMHelper.getFromCHM(index, new Pair<>(block, npBlock), () -> blockCounter.getAndIncrement());
									newBlocks.add(newBlock);
								}
							}
						}

						newBlocks.addAll(nonProblemBlock);
						rowToBlocks.put(rowId, Collections.unmodifiableSet(newBlocks));

					}

				}
			}, this.getClass());

			ConcurrentHashMap<Integer, Pair<Integer, Integer>> reverseIndex = new ConcurrentHashMap<>();
			for (Entry<Pair<Integer, Integer>, Integer> e : index.entrySet()) {
				reverseIndex.put(e.getValue(), e.getKey());
			}

			ConcurrentHashMap<Integer, AtomicInteger> countMap2 = new ConcurrentHashMap<>();

			PersistentParallel6.runAndFinish((rowToBlocks.entrySet()), (e) -> {
				for (Integer i : e.getValue()) {
					CHMHelper.getFromCHM(countMap2, i, () -> new AtomicInteger()).incrementAndGet();
				}
			}, this.getClass());

			List<Pair<String, Integer>> debugMap2 = countMap2.entrySet().stream().sorted((e1, e2) -> {
				int compare = Integer.compare(e2.getValue().intValue(), e1.getValue().intValue());
				if (compare != 0) {
					return compare;
				}
				return Integer.compare(e1.getKey(), e2.getKey());
			}).map((e) -> {
				String s = longIndex.get(e.getKey());
				if (s == null)
					s = shortIndex.get(e.getKey());
				if (s == null) {
					Pair<Integer, Integer> combined = reverseIndex.get(e.getKey());
					String s1 = longIndex.get(combined.getFirst());
					if (s1 == null)
						s1 = shortIndex.get(combined.getFirst());
					String s2 = longIndex.get(combined.getSecond());
					if (s2 == null)
						s2 = shortIndex.get(combined.getSecond());
					s = s1 + " " + s2;
				}
				return new Pair<>(s, e.getValue().intValue());
			}).collect(Collectors.toList());
			return rowToBlocks;
		}
		
		public Map<Integer, Integer> getClustersMappings(Map<String, Map<Integer, Number>> input, Map<String, Integer> map) {

			LevelSteinSimilarityWithCache sim = new LevelSteinSimilarityWithCache();
			double threshold = 0.9;
			Map<Integer, String> index = map.entrySet().stream().collect(Collectors.toMap(Entry::getValue, Entry::getKey));
			GenericLabelRowIndexGeneric<Integer> labelRowIndex;
			try {
				labelRowIndex = new GenericLabelRowIndexGeneric<>(input, 2);

			// MutableReference<IOException> exception = new
			// MutableReference<IOException>(null);
			// LabelRowIndex labelRowIndex =
			// cacheIndex.getOrCreate(input.getUuid(), () -> {
			// try {
			// return new LabelRowIndex(tableRepository, candidates);
			// } catch (IOException e1) {
			// exception.set(e1);
			// return null;
			// }
			// });
			//
			// if (labelRowIndex == null) {
			// throw exception.get();
			// }

			SimilarityMeasure smSimilairty = similarityCalculator.stringSimilarityMixed;
//			SimilarityMeasure smClustering = similarityCalculator.getLabelBlockingSimilarity();
			SimilarityMeasure smClustering = similarityCalculator.getLabelBlockingSimilarity();
			ConcurrentHashMap<TableRowId, Set<TableRowId>> mappings = new ConcurrentHashMap<>();
			ConcurrentHashMap<TableRowId, TableRowId> oneToOneMapping = new ConcurrentHashMap<>();

			// Set<String> done = Collections.newSetFromMap(new
			// ConcurrentHashMap<>());

			Map<String, TableRowId> done = new ConcurrentHashMap<>();

			ConcurrentHashMap<Integer, Set<Integer>> mainMapping = new ConcurrentHashMap<>();

			PersistentParallel6.runAndFinish(input.keySet(), (string) -> {

				// TableRowId tableRowIdCheck = new TableRowId(new
				// AnnotatedTableId("1346876860789_1346946768085_4249.arc4550171461300486933#66673628_0_6604262406652303152"),
				// 1);
				// if(tableRowId.equals(tableRowIdCheck)) {
				// int sdf= 123;
				// }
				// if
				// (tableRepository.getMetaData(tableRowId.getTableId()).getProperties().size()
				// == 0)
				// return;
				// if (labelSO != null) {

				StringObject labelSO = new StringObject(string);
				Optional<String> deAccented = IndexBuilder.deAccent(labelSO.getString());
				StringObject deAccentedSO = null;

				if (deAccented.isPresent()) {
					deAccentedSO = new StringObject(deAccented.get());
				}

				try {
					// List<String> finalMatchesForDebug = new ArrayList<>();

					Map<Integer, Float> matches = labelRowIndex.getMatches(string);

					for (Iterator<Entry<Integer, Float>> it = matches.entrySet().iterator(); it.hasNext();) {
						Entry<Integer, Float> next = it.next();

						String candidateString = index.get(next.getKey());

						StringObject candidateSO = new StringObject(candidateString);
						StringObject deAccentedCS = null;
						deAccented = IndexBuilder.deAccent(candidateString);
						if (deAccented.isPresent()) {
							deAccentedCS = new StringObject(deAccented.get());
						}

							if (sim.calculate(candidateSO.getString(), labelSO.getString()) < threshold) {

							if (deAccentedSO != null) {
								if (deAccentedCS != null) {
										if (sim.calculate(deAccentedSO.getString(), deAccentedCS.getString()) < threshold) {
										it.remove();
									}
								} else {
										if (sim.calculate(deAccentedSO.getString(), candidateSO.getString()) < threshold) {
										it.remove();
									}
								}
							} else if (deAccentedCS != null) {
									if (sim.calculate(labelSO.getString(), deAccentedCS.getString()) < threshold) {
									it.remove();
								}
							}
								else {
									it.remove();
								}

						} else {


						}

						// StringObject compareLabel = getLabel(tableRepository,
						// next);
						// if (blockByClass &&
						// !classId.equals(tableRepository.getMetaData(next.getTableId()).getDbPediaClass()))
						// {
						// it.remove();
						// } else if (!smClustering.isEqual(labelSO,
						// compareLabel)) {
						// it.remove();
						//
						// if (labelSO.getString().equals("compareLabel")) {
						// int fasdf = 123;
//								}
//							}
//							else {
						// finalMatchesForDebug.add(compareLabel.getString());
//							}

					}

					if (matches.isEmpty()) {
						int sdffff = 123;
							Set<Integer> equalBlocks = new HashSet<>();
							equalBlocks.add(map.get(string));
							mainMapping.put(map.get(string), equalBlocks);
					} else {

						Set<String> debugSet = matches.keySet().stream().map((e) -> index.get(e)).collect(Collectors.toSet());
						Set<Integer> equalBlocks = new HashSet<>(matches.keySet());
							equalBlocks.add(map.get(string));
							if (debugSet.size() > 1) {
								int df = 12;
							}
						mainMapping.put(map.get(string), equalBlocks);
					}

				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				// }
				// else {
				//// int sdfsf = 123;
				//// System.out.println(tableRowId.toString());
				// }

			}, this.getClass(), "FindMappings");

			ClusterCombiner<Integer> cc = new ClusterCombiner<>();

			Map<Integer, Integer> finalMapping = new HashMap<>();

			List<Set<Integer>> list = new ArrayList<>();
				int largestSize = 0;
			for (Entry<Integer, Set<Integer>> entry : mainMapping.entrySet()) {
				Set<Integer> set = new HashSet<>();
				set.add(entry.getKey());
				set.addAll(entry.getValue());
				list.add(set);
					if (set.size() > largestSize) {
						largestSize = set.size();
					}
			}

			cc.add(list);

				// Map<Integer, MutableInt> countMap =new HashMap<>();
			for (Set<Integer> set : cc.get()) {
					if (set.size() >= 1) {
					Integer s = set.iterator().next();
					for (Integer i : set) {
							finalMapping.put(i, s);
					}
						if (set.size() > 5) {
							List<String> debugList = set.stream().map((e) -> index.get(e)).collect(Collectors.toList());

							int sdf = 123;
						}
				} else {
					throw new IllegalStateException();
				}
			}

			return finalMapping;

			} catch (IOException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
				return null;
			}
		}

		public List<Tri<Set<Integer>, Integer, Double>> getSplitBlocksWithScores(TableRowId tableRowId) {
			List<Tri<Set<Integer>, Integer, Double>> list = new ArrayList<>();
			Set<Integer> blocksApproximate = rowToBlocks.get(tableRowId);
			Set<Integer> blocksFull = rowToBlocksFull.get(tableRowId);
			if (blocksApproximate == null)
				blocksApproximate = new HashSet<>();
			if (blocksFull == null)
				blocksFull = new HashSet<>();
			list.add(new Tri<>(blocksFull, 1, 1.55));
			list.add(new Tri<>(blocksApproximate, getBlockMin(tableRowId), 1.55));
			return list;
		}


		@Override
		public Set<Integer> getBlocks(TableRowId rowId) {

			Set<Integer> ret = new HashSet<>();

			Set<Integer> blocksApproximate = rowToBlocks.get(rowId);
			Set<Integer> blocksFull = rowToBlocksFull.get(rowId);
			if (blocksApproximate != null) {
				ret.addAll(blocksApproximate);
			}
			if (blocksFull != null) {
				ret.addAll(blocksFull);
			}

			return ret;
		}

		@Override
		public int getBlockMin(TableRowId tableRowId) {
			Set<Integer> blocks = rowToBlocksForSizes.get(tableRowId);
			if (blocks != null) {
				int size = rowToBlocksForSizes.get(tableRowId).size();
			int ret;
			if (size == 1) {
				ret = 1;
			} else {
					ret = (int) (size * 0.85);
			}
			return ret;
			} else {
				return 1;
			}
		}

		@Override
		public boolean supportsBlocking() {
			return rowToBlocks != null;
		}
		
		
		@Override
		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			return labelSimilaritiesRaw.get(rows, classId, key);
		}

		@Override
		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getName() {
			return "LABEL";
		}
		
	}

	public RowClusterSimilarities createNormalized() {
		throw new IllegalStateException();
		// return new NormalizedRowClusterSimilarities(create());
	}
	
	public RowClusterSimilarities createNormalized(boolean withSurfaceForms, boolean withStopWordsRemoval, boolean withTfIDF) {
		throw new IllegalStateException();
		// return new NormalizedRowClusterSimilarities(create(withSurfaceForms,
		// withStopWordsRemoval, withTfIDF));
	}

	public RowClusterSimilarities createTFIDFNormalized(ClustererInput input) {
		Map<TableRowId, Map<String, Integer>> allStrings = new HashMap<>();
		
		for(TableRowId rowId : input.getCandidates()) {
		
			allStrings.put(rowId, getRowBag(  rowId));
		}
		
		Map<String, Double> tfIdf = BagOfWords_Component.getTfIdf(allStrings);
		
		
		return new LabelBlockSimilarities(tfIdf);
		
	}
	
	public Map<String, Integer> getRowBag(TableRowId rowId) {
		Map<String, MutableInt> bag = new HashMap<>();
		AbstractObject ao = tablesRepository.getTable(rowId.getTableId()).getKeyValue(rowId.getRowIdx());
		if(ao != null && ao.getType() == ValueType.STRING) {
			StringObject so = (StringObject) ao;
			List<String> tokenized = StringNormalizer.tokenise(so.getString(), false, false);
			
			for(String token : tokenized) {
				CHMHelper.getFromMap(bag, token, () -> new MutableInt()).increment();
			}
		}
		return bag.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> e.getValue().intValue()));
	}
	
	private   class   LabelBlockSimilarities implements RowClusterSimilarities {
		private final Map<String, Double> tfIdf; 
		private final Double defaultScore;
		public LabelBlockSimilarities(Map<String, Double> tfIdf) {
			this.tfIdf = tfIdf;
			this.defaultScore = null;
		}
		
		public LabelBlockSimilarities(Double defaultScore) {
			this.tfIdf = null;
			this.defaultScore = defaultScore;
		}

		@Override
		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			return BagOfWords_Component.score(getRowBag(rows.get(key.getE1())), getRowBag(rows.get(key.getE2())),defaultScore, tfIdf, true);
		}

		@Override
		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getName() {
			return "LABEL_BLOCK";
		}
		
	}

	public RowClusterSimilarities createCosine() {
		return new LabelBlockSimilarities(1.0);
	}
}
