package expansion.set.strategies.rowclusterers.label;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.BiFunction;
import java.util.stream.Collectors;

import javax.inject.Inject;
import javax.inject.Named;

import org.apache.commons.lang.mutable.MutableInt;

import com.github.benmanes.caffeine.cache.Caffeine;
import com.github.benmanes.caffeine.cache.LoadingCache;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.OtsuThresholderRounded;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.informatik.dws.expansion.utils.MutableReference;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;
import edu.stanford.nlp.util.StringUtils;
import expansion.set.SetExpansionProperties;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.strategies.blockers.Blocks;
import expansion.set.strategies.blockers.Mappings;
import expansion.set.strategies.rowclusterers.ClusterSimilarities;
import expansion.set.strategies.rowclusterers.ComponentCreator;
import expansion.set.strategies.rowclusterers.ComponentCreator.ComponentExist;
import expansion.set.strategies.rowclusterers.CreateClustersFromMappings;
import expansion.set.strategies.rowclusterers.GreedyClustering;
import expansion.set.strategies.rowclusterers.label.LabelRowIndexClusterMapCreator.LabelRowIndex;

public class TfIdfBlockCreator {

	private final AnnotatedTablesRepository tableRepository;
	
//	private final AnnotatedTablesRepository tableRepository;
	private final SimilarityCalculator similarityCalculator;
	
	private final ComponentCreator componentCreator;
	
	private final String cachesFolder;
	
	@Inject
	public TfIdfBlockCreator(AnnotatedTablesRepository tablesRepository, SimilarityCalculator similarityCalculator,ComponentCreator componentCreator, @Named(SetExpansionProperties.tablesCachesFolder) String cachesFolder) {
		this.tableRepository = tablesRepository;
		this.similarityCalculator = similarityCalculator;
		this.componentCreator = componentCreator;
		this.cachesFolder = cachesFolder;
	}
	
	
	public final static SimpleLockableConcurrentCache<UuidStringId, Map<Integer, Object>> cache = new SimpleLockableConcurrentCache<>();
	public Map<Integer, Object> getFronCache(TableIdsSet tableIdsSet) throws IOException {
		return cache.getOrCreateNoBlocking(tableIdsSet.getId(), () -> {
			try {
				return getTfIdfStuff(tableIdsSet);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
				return null;
			}
		});
	}
	public Map<Integer, Object> getTfIdfStuff(TableIdsSet tableIdsSet) throws IOException {
		Set<TableRowId> candidates = tableRepository.getAllRowIds(tableIdsSet.getTableIds());
		
		ConcurrentHashMap<String, Integer> stringToIndexMapping = new ConcurrentHashMap<>();
		ConcurrentHashMap<List<Integer>, Integer> componentToIndexMapping = new ConcurrentHashMap<>();
		ConcurrentHashMap<Integer, Set<TableRowId>> componentToRowMapping = new ConcurrentHashMap<>();
		ConcurrentHashMap<TableRowId, Set<Integer>> rowToComponentMapping = new ConcurrentHashMap<>();
		ConcurrentHashMap<TableRowId, Integer> rowMainComponent = new ConcurrentHashMap<>();
		

		ConcurrentHashMap<Integer, List<Integer>> componentMap = new ConcurrentHashMap<>();
		ConcurrentHashMap<Integer, Set<Integer>> componentHierarchyChilderen = new ConcurrentHashMap<>();
		 
		AtomicInteger componentIndexCounter = new AtomicInteger(0);
		// MutableReference<IOException> exception = new
		// MutableReference<IOException>(null);
		
		LabelRowIndex labelRowIndex;
		String prefixPath = cachesFolder + "/labelindex/" ;
		MKDIR.createDirUnchecked(new File(prefixPath));
		String path = prefixPath + UUID.randomUUID();
		
//		File f = new File(path);
//		if (f.isDirectory()) {
//			labelRowIndex = new LabelRowIndexClusterMapCreator.LabelRowIndex(path, 2);
//		}
//		else {
			labelRowIndex = new LabelRowIndexClusterMapCreator.LabelRowIndex(tableRepository, candidates, 2);
//		}
		 
		 
		// LabelRowIndex labelRowIndex =
		// cacheIndex.getOrCreate(input.getUuid(), () -> {
		// try {
		// return new LabelRowIndex(tableRepository, candidates);
		// } catch (IOException e1) {
		// exception.set(e1);
		// return null;
		// }
		// });
		//
		// if (labelRowIndex == null) {
		// throw exception.get();
		// }
//
//		StringSimilarityJaccard smSimilairty = similarityCalculator.stringSimilarityJaccard;
////		SimilarityMeasure smClustering = similarityCalculator.getLabelBlockingSimilarity();
//		SimilarityMeasure smClustering = similarityCalculator.getLabelBlockingSimilarity();
//		ConcurrentHashMap<TableRowId, Set<TableRowId>> mappings = new ConcurrentHashMap<>();
//		ConcurrentHashMap<TableRowId, TableRowId> oneToOneMapping = new ConcurrentHashMap<>();
		
//		Set<String> done = Collections.newSetFromMap(new ConcurrentHashMap<>());
		
		Map<String, TableRowId> done = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(candidates, (tableRowId) -> {
			StringObject labelSO = LabelRowIndexClusterMapCreator.getLabel(tableRepository, tableRowId); 
			
 
			if (labelSO != null && labelSO.getString().length() < 76) {
				
				String normalized = LabelRowIndex.normalise(labelSO.getString());
 				ClassId classId = tableRepository.getMetaData(tableRowId.getTableId()).getDbPediaClass();


				if(done.containsKey(normalized)) {
					TableRowId otherRowId = done.get(normalized);
					  
					 Set<Integer> componentIndeces = new HashSet<>(rowToComponentMapping.get(otherRowId));
					
					 
						for(Integer index : componentIndeces) {
 							CHMHelper.getFromCHM(componentToRowMapping, index, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowId);
	 
						}
						
					rowToComponentMapping.put(tableRowId, componentIndeces);
					rowMainComponent.put(tableRowId, rowMainComponent.get(otherRowId));
					return;
				}
				
				try { 
					
//					if(tableRowId.equals(new TableRowId(new AnnotatedTableId("1346981172255_1346993220570_2522.arc872863852192600628#52763396_0_5515677821757136208"), 254))) {
//						int sdf = 123;
//					}
//					
					 Map<String, Set<String>> components = new HashMap<>();
//					
					 components.put(normalized, new HashSet<>());
					
//					 List<String> currentComponent = Arrays.asList(normalized.split(" "));
//					 ComponentCreator.addCurrentComponent(components, currentComponent);
					 
					Map<StringObject, Set<TableRowId>> matches = labelRowIndex.getMatches(labelSO.getString());
//					
//				
// 
//					 
					for (Iterator<Entry<StringObject, Set<TableRowId>>> it = matches.entrySet().iterator(); it.hasNext();) {
						Entry<StringObject, Set<TableRowId>> next = it.next();
						
						
						List<String> listA = Arrays.asList( StringNormalizer.normaliseValue(labelSO.getString(), false).split(" "));
						List<String> listB = Arrays.asList( StringNormalizer.normaliseValue(next.getKey().getString(), false).split(" "));
						 
						
						 List<Pair<List<String>, ComponentExist>> subComponents = componentCreator.create(listA, listB, false, false, true);
						
						for(Pair<List<String>, ComponentExist> entry  : subComponents ) {
							CHMHelper.getFromMap(components, StringUtils.join(entry.getFirst(), " "), () -> new HashSet<>());
						 
						}
						
						 
						
					 

					}
					
				 
					if(normalized.equals("")) {

						System.out.println("This label has no components: " + labelSO.getString());
						rowToComponentMapping.put(tableRowId, new HashSet<>());
					}
					else {
						
						Set<Integer> componentIndeces = new HashSet<>();
						{
							Integer index = CHMHelper.getFromCHM(stringToIndexMapping, normalized, () -> componentIndexCounter.getAndIncrement());
							rowMainComponent.put(tableRowId, index); 
						}
						
						for(Entry<String, Set<String>> componentEntry : components.entrySet()) {
							Integer index = CHMHelper.getFromCHM(stringToIndexMapping, componentEntry.getKey(), () -> componentIndexCounter.getAndIncrement());
							CHMHelper.getFromCHM(componentToRowMapping, index, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowId);
							componentIndeces.add(index);
							List<Integer> subComponents = new ArrayList<>();
							List<String> subComponentsString = Arrays.asList(componentEntry.getKey().split(" "));
						 
							for(String subComponent : subComponentsString) {
								Integer subIndex = CHMHelper.getFromCHM(stringToIndexMapping, subComponent, () -> componentIndexCounter.getAndIncrement());
								subComponents.add(subIndex);
							}
							
							componentMap.put(index, subComponents);
							
//							Set<Integer> children = new HashSet<>();
//							for(String subComponent : componentEntry.getValue()) {
//								Integer subIndex = CHMHelper.getFromCHM(stringToIndexMapping, subComponent, () -> componentIndexCounter.getAndIncrement());
//								children.add(subIndex);
//							}
							
//							CHMHelper.getFromCHM(componentHierarchyChilderen, index, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).addAll(children);

						}
						
						 rowToComponentMapping.put(tableRowId, componentIndeces);
						 

						done.put(normalized, tableRowId);
						
					}
					

				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
//			 

		} , this.getClass(), "FindMappings");
		
	 

		ConcurrentHashMap<Integer, Set<Integer>> componentHierarchyParents = new ConcurrentHashMap<>();
		
		PersistentParallel6.runAndFinish(componentHierarchyChilderen.entrySet(), (entry) -> {
			
			CHMHelper.getFromCHM(componentHierarchyParents, entry.getKey(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>()));
			
			
			for(Integer child : entry.getValue()) {
				CHMHelper.getFromCHM(componentHierarchyParents, child, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(entry.getKey());
			}
		},this.getClass() ,"CreateHierarchy");
		
		
		int totalSize =   componentToRowMapping.size();
		int parentSizeMap = componentHierarchyParents.size();
		
//		Map<Integer, Integer> levels = createHierarchy(componentHierarchyParents);
		
//		Set<Integer> added = new HashSet<>();
		
		
		
		Map<Integer, String> indexToStringMapping = new ConcurrentHashMap<>();
		
		PersistentParallel6.runAndFinish( stringToIndexMapping.entrySet(), (entry) -> {
			indexToStringMapping.put(entry.getValue(), entry.getKey());
			
			List<String> subComponentsString = Arrays.asList(entry.getKey().split(" "));
			List<Integer> subComponents = new ArrayList<>();
			for(String subComponent : subComponentsString) {
				int subIndex = stringToIndexMapping.get(subComponent);
				subComponents.add(subIndex);
			}
			
			componentToIndexMapping.put(subComponents, entry.getValue());
		}, this.getClass(), "CreateSubComponent");
		
	
//		int sizeLevelInMap = levels.size();
		

//		Map<TableRowId, Map<Integer,Integer>> rowToComponentMappingWithLevels = createHierarchyPerComponentToRowMapping(rowToComponentMapping, componentHierarchyParents);
		
//		for(Entry<TableRowId, Map<Integer, Integer>> entry : rowToComponentMappingWithLevels.entrySet()) {
//			for(Entry<Integer, Integer> innerEntry : entry.getValue().entrySet()) {
//				System.out.println(indexToStringMapping.get(innerEntry.getKey()) + " ::: " + innerEntry.getValue());
//			}
//			System.out.println();
//			int sdf = 13;
//		}
//		
		
		if(componentToRowMapping.size() == 0) {
			int sdf = 123;
		}
		double maxComponentOccurence = componentToRowMapping.values().stream().mapToInt((e) -> e.size()).max().getAsInt();
		ConcurrentHashMap<Integer, Double> scores = new ConcurrentHashMap<>(); 
		PersistentParallel6.runAndFinish(componentToRowMapping.entrySet(), (c) -> {
//			scores.put(c.getKey(), 1.0);
			scores.put(c.getKey(), - Math.log(c.getValue().size() / maxComponentOccurence));
//			scores.put(c.getKey(), - Math.log(1.0 / totalRowSize));
		}, this.getClass(), "CreateScores");
	 
		
	
		Map<Integer, Object> ret = new HashMap<>();
		ret.put(TfIdfBlockCreator.stringToIndexMapping, stringToIndexMapping);
		ret.put(TfIdfBlockCreator.componentToIndexMapping, componentToIndexMapping);
		ret.put(TfIdfBlockCreator.componentToRowMapping, componentToRowMapping);
		ret.put(TfIdfBlockCreator.rowToComponentMapping, rowToComponentMapping);
		ret.put(TfIdfBlockCreator.rowMainComponent, rowMainComponent);
		ret.put(TfIdfBlockCreator.rowMainComponent, rowMainComponent);
		ret.put(TfIdfBlockCreator.scores, scores);
		ret.put(TfIdfBlockCreator.indexToStringMapping, indexToStringMapping);
		ret.put(TfIdfBlockCreator.componentMap, componentMap);
		ret.put(TfIdfBlockCreator.labelRowIndex, labelRowIndex);
		return ret;
		
	}
	
	public static final Integer stringToIndexMapping = 0;
	public static final Integer componentToIndexMapping = 1;
	public static final Integer componentToRowMapping = 2;
	public static final Integer rowToComponentMapping = 3;
	public static final Integer rowMainComponent = 4;
	public static final Integer scores = 5;
	public static final Integer indexToStringMapping = 6;
	public static final Integer componentMap = 7;
	public static final Integer labelRowIndex = 8;
	
	
	public Mappings getBlockings(TableIdsSet tableIdsSet, Set<TableRowId> candidates, boolean blockByClass)   {
		 try {
		Map<Integer, Object> tfIdfStuff = getFronCache(tableIdsSet);
		
		 Map<String, Integer> stringToIndexMapping = (Map<String, Integer>) tfIdfStuff.get(TfIdfBlockCreator.stringToIndexMapping);
		 Map<List<Integer>, Integer> componentToIndexMapping = (Map<List<Integer>, Integer>) tfIdfStuff.get(TfIdfBlockCreator.componentToIndexMapping);
		 Map<Integer, Set<TableRowId>> componentToRowMapping = (Map<Integer, Set<TableRowId>>) tfIdfStuff.get(TfIdfBlockCreator.componentToRowMapping);
		 Map<TableRowId, Set<Integer>> rowToComponentMapping = (Map<TableRowId, Set<Integer>>) tfIdfStuff.get(TfIdfBlockCreator.rowToComponentMapping);
		 Map<TableRowId, Integer> rowMainComponent = (Map<TableRowId, Integer>) tfIdfStuff.get(TfIdfBlockCreator.rowMainComponent);
		
		Map<Integer, Double> scores = (Map<Integer, Double>) tfIdfStuff.get(TfIdfBlockCreator.scores);
		Map<Integer, String>  indexToStringMapping = (Map<Integer, String>) tfIdfStuff.get(TfIdfBlockCreator.indexToStringMapping);
		Map<Integer, List<Integer>> componentMap = (Map<Integer, List<Integer>>) tfIdfStuff.get(TfIdfBlockCreator.componentMap);
		LabelRowIndex labelRowIndex = (LabelRowIndex) tfIdfStuff.get(TfIdfBlockCreator.labelRowIndex);
		
		double maxComponentOccurence = componentToRowMapping.values().stream().mapToInt((e) -> e.size()).max().getAsInt();

		
		double totalRowSize = 	rowToComponentMapping.size();
		ConcurrentHashMap<TableRowId, Set<TableRowId>> mappings = new ConcurrentHashMap<>();
		
		Map<TableRowId, Set<TableRowId>> mappingsOther = createMappingsFromTFIDFNew(candidates, scores, componentToRowMapping, rowToComponentMapping, indexToStringMapping,rowMainComponent, componentMap, componentToIndexMapping, maxComponentOccurence);

		
		
		if(false) {
			
		ConcurrentHashMap<Integer, Set<TableRowId>> mainComponentToRowMapping = new ConcurrentHashMap<>();
		
		PersistentParallel6.runAndFinish(rowMainComponent.entrySet(), (entry) -> {
			CHMHelper.getFromCHM(mainComponentToRowMapping, entry.getValue(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(entry.getKey());
		}, this.getClass(), "MainComponent");

		
		
		
		Map<String, TableRowId> done = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(candidates, (tableRowId) -> {
			StringObject labelSO = LabelRowIndexClusterMapCreator.getLabel(tableRepository, tableRowId); 
			
			if (labelSO != null && labelSO.getString().length() < 76) {
				
				String normalized = LabelRowIndex.normalise(labelSO.getString());
// 				ClassId classId = tableRepository.getMetaData(tableRowId.getTableId()).getDbPediaClass();

				if(normalized.equals("")) {
					return; 
				}
				
				if(done.containsKey(normalized)) {
					TableRowId otherRowId = done.get(normalized);
					
					CHMHelper.getFromCHM(mappings, otherRowId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowId);
					
					return;
				}
				
				try { 
					
//					if(tableRowId.equals(new TableRowId(new AnnotatedTableId("1346981172255_1346993220570_2522.arc872863852192600628#52763396_0_5515677821757136208"), 254))) {
//						int sdf = 123;
//					}
 
					Set<TableRowId> mappingsSet = new HashSet<>();
					int mainComponent = rowMainComponent.get(tableRowId);
					Map<StringObject, Set<TableRowId>> matches = labelRowIndex.getMatches(labelSO.getString());
					mappingsSet.addAll(mainComponentToRowMapping.get(mainComponent));
				Set<Integer> mainComponentCandidates = matches.values().stream().flatMap((e) -> e.stream()).filter((e) -> rowMainComponent.containsKey(e)).map((e) -> rowMainComponent.get(e)).filter((e) -> !e.equals(mainComponent)).collect(Collectors.toSet());
				
				for(Integer otherComponent : mainComponentCandidates) {
					if(comparePair(mainComponent, otherComponent, scores, indexToStringMapping, componentMap, componentToIndexMapping, maxComponentOccurence) > 0.5) {
						mappingsSet.addAll(mainComponentToRowMapping.get(otherComponent));
					}
				}
//				mappingsSet.addAll(component)
				 
// 
//					 
				 
				mappings.put(tableRowId, mappingsSet);

						done.put(normalized, tableRowId);
						
		 

				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
			
//			 

		} , this.getClass(), "FindMappings-BLOCKER");
		
		for(Entry<TableRowId, Set<TableRowId>> entry : mappingsOther.entrySet()) {
			CHMHelper.getFromCHM(mappings, entry.getKey(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).addAll(entry.getValue());
		}
		}
		
	
		
		
		

		
		return new Mappings(mappings);
		
		 }
		 catch(IOException e) {
			 e.printStackTrace();
			 return null;
		 }
	}
	
	public RowClusterSimilarities  getSimilarities(TableIdsSet tableIdsSet, Set<TableRowId> candidates, boolean blockByClass)   {
		try {
		Map<Integer, Object> tfIdfStuff = getFronCache(tableIdsSet);
		
		 Map<String, Integer> stringToIndexMapping = (Map<String, Integer>) tfIdfStuff.get(TfIdfBlockCreator.stringToIndexMapping);
		 Map<List<Integer>, Integer> componentToIndexMapping = (Map<List<Integer>, Integer>) tfIdfStuff.get(TfIdfBlockCreator.componentToIndexMapping);
		 Map<Integer, Set<TableRowId>> componentToRowMapping = (Map<Integer, Set<TableRowId>>) tfIdfStuff.get(TfIdfBlockCreator.componentToRowMapping);
		 Map<TableRowId, Set<Integer>> rowToComponentMapping = (Map<TableRowId, Set<Integer>>) tfIdfStuff.get(TfIdfBlockCreator.rowToComponentMapping);
		 Map<TableRowId, Integer> rowMainComponent = (Map<TableRowId, Integer>) tfIdfStuff.get(TfIdfBlockCreator.rowMainComponent);
		
		Map<Integer, Double> scores = (Map<Integer, Double>) tfIdfStuff.get(TfIdfBlockCreator.scores);
		Map<Integer, String>  indexToStringMapping = (Map<Integer, String>) tfIdfStuff.get(TfIdfBlockCreator.indexToStringMapping);
		Map<Integer, List<Integer>> componentMap = (Map<Integer, List<Integer>>) tfIdfStuff.get(TfIdfBlockCreator.componentMap);
		
		
		double totalRowSize = 	rowToComponentMapping.size();
		
		  LoadingCache<Pair<Integer, Integer>, Double> cache = Caffeine.newBuilder()
		 	       .maximumSize(1000000)
		 	       .build( (p) -> {
						return comparePair(p.getFirst(), p.getSecond(),  scores, indexToStringMapping, componentMap, componentToIndexMapping, totalRowSize);

		 	       }
		 	           );
		    
		    
		    BiFunction<TableRowId, TableRowId, Double> calc = (e1, e2) -> {
		    	
		     
		    	boolean run = true;
//		    
//		    	Set<Integer> e1Set = rowToComponentMapping.get(e1);
//		    	Set<Integer> e2Set = rowToComponentMapping.get(e2);
//		    	
//		    	Set<Integer> large;
//		    	Set<Integer> small;
//		    	if(e1Set.size() < e2Set.size()) {
//		    		large = e2Set;
//		    		small = e1Set;
//		    		
//		    	}
//		    	else {
//		    		large = e1Set;
//		    		small = e2Set;
//		    	}
//		    	
//		    	for(Integer i : small) {
//		    		if(large.contains(i)) {
//		    			run = true;
//		    			break;
//		    		}
//		    	}
		    	
		    	if(run) {
		    		Integer main1 = rowMainComponent.get(e1);
			    	Integer main2 = rowMainComponent.get(e2);
					return comparePair(main1, main2,  scores, indexToStringMapping, componentMap, componentToIndexMapping, totalRowSize);

//			    	if(main1 == main2) {
//			    		return 1.0;
//			    	}
//			    	else if(main1 < main2) {
//			    		return cache.get(new Pair<>(main1, main2));
//			    	}
//			    	else {
//			    		return cache.get(new Pair<>(main2, main1));
//			    	}
		    	}
		    	else {
		    		return -1.0;
		    	}
		    };
		    
		    
		 return new RowClusterSimilarities() {

			@Override
			public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
				return calc.apply(rows.get(key.getE1()), rows.get(key.getE2()));
			}

			@Override
			public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
				// TODO Auto-generated method stub
				return null;
			}

				@Override
				public String getName() {
					// TODO Auto-generated method stub
					return null;
				}
			 
		 };
		}
		catch (IOException e) {
			e.printStackTrace();
			return null;
		}
	}
	
	public Blocks  getClustersMap(TableIdsSet tableIdsSet, Set<TableRowId> candidates, boolean blockByClass) throws IOException {
		 
		Map<Integer, Object> tfIdfStuff = getFronCache(tableIdsSet);
		
		 Map<String, Integer> stringToIndexMapping = (Map<String, Integer>) tfIdfStuff.get(TfIdfBlockCreator.stringToIndexMapping);
		 Map<List<Integer>, Integer> componentToIndexMapping = (Map<List<Integer>, Integer>) tfIdfStuff.get(TfIdfBlockCreator.componentToIndexMapping);
		 Map<Integer, Set<TableRowId>> componentToRowMapping = (Map<Integer, Set<TableRowId>>) tfIdfStuff.get(TfIdfBlockCreator.componentToRowMapping);
		 Map<TableRowId, Set<Integer>> rowToComponentMapping = (Map<TableRowId, Set<Integer>>) tfIdfStuff.get(TfIdfBlockCreator.rowToComponentMapping);
		 Map<TableRowId, Integer> rowMainComponent = (Map<TableRowId, Integer>) tfIdfStuff.get(TfIdfBlockCreator.rowMainComponent);
		
		Map<Integer, Double> scores = (Map<Integer, Double>) tfIdfStuff.get(TfIdfBlockCreator.scores);
		Map<Integer, String>  indexToStringMapping = (Map<Integer, String>) tfIdfStuff.get(TfIdfBlockCreator.indexToStringMapping);
		Map<Integer, List<Integer>> componentMap = (Map<Integer, List<Integer>>) tfIdfStuff.get(TfIdfBlockCreator.componentMap);
		
		
//		double maxComponentOccurence = 0.0;
		
		double totalRowSize = 	rowToComponentMapping.size();
		
		double maxComponentOccurence = componentToRowMapping.values().stream().mapToInt((e) -> e.size()).max().getAsInt();
		
		Map<TableRowId, Set<TableRowId>> mappings = createMappingsFromTFIDF(candidates, scores, componentToRowMapping, rowToComponentMapping, indexToStringMapping,rowMainComponent, componentMap, componentToIndexMapping, maxComponentOccurence);
		
		Map<Integer, Set<TableRowId>> clustersMap = new HashMap<>();
		Map<Integer, ClassId> clusterClassId = new HashMap<>();
		Map<TableRowId, Integer> clustersIndex = new HashMap<>();

		CreateClustersFromMappings.addCluster(mappings, clustersMap, clustersIndex, clusterClassId, (tid) -> {
			return tableRepository.getMetaData(tid).getDbPediaClass(); 
		}, 
				(rowId) -> {
					StringObject so = LabelRowIndexClusterMapCreator.getLabel(tableRepository, rowId);
					if(so != null) {
						return so.getString();
					}
					else {
						return null;
					}
				});
		
		if(false) {
		return new Blocks(clustersMap, clusterClassId);
		}
		else { 
			
			
			    LoadingCache<Pair<Integer, Integer>, Double> cache = Caffeine.newBuilder()
			 	       .maximumSize(40000)
			 	       .build( (p) -> {
							return comparePair(p.getFirst(), p.getSecond(),  scores, indexToStringMapping, componentMap, componentToIndexMapping, maxComponentOccurence);

			 	       }
			 	           );
			    
			    
			    BiFunction<TableRowId, TableRowId, Double> calc = (e1, e2) -> {
			    	boolean run = false;
			    
			    	Set<Integer> e1Set = rowToComponentMapping.get(e1);
			    	Set<Integer> e2Set = rowToComponentMapping.get(e2);
			    	
			    	Set<Integer> large;
			    	Set<Integer> small;
			    	if(e1Set.size() < e2Set.size()) {
			    		large = e2Set;
			    		small = e1Set;
			    		
			    	}
			    	else {
			    		large = e1Set;
			    		small = e2Set;
			    	}
			    	
			    	for(Integer i : small) {
			    		if(large.contains(i)) {
			    			run = true;
			    			break;
			    		}
			    	}
			    	
			    	if(run) {
			    		Integer main1 = rowMainComponent.get(e1);
				    	Integer main2 = rowMainComponent.get(e2);
//						return comparePair(main1, main2,  scores, indexToStringMapping, componentMap, componentToIndexMapping, totalRowSize);

			    		return cache.get(new Pair<>(main1, main2));
			    	}
			    	else {
			    		return -1.0;
			    	}
			    };
				
				GreedyClustering gc = new GreedyClustering();
				
				ConcurrentHashMap<ClassId, Set<TableRowId>> classMap = new ConcurrentHashMap<>();
				 
				
				PersistentParallel6.runAndFinish(candidates, (tableRowId) -> {
					ClassId classId = 	tableRepository.getMetaData(tableRowId.getTableId()).getDbPediaClass(); ;
					CHMHelper.getFromCHM(classMap, classId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowId);
				}, this.getClass(), "CreateClassId");
				
				
				
				Map<Integer, Set<TableRowId>> newClustersMap = new HashMap<>();
				Map<Integer, ClassId> newClusterClassId = new HashMap<>();
				
				AtomicInteger ai = new AtomicInteger();
				for(Entry<Integer, Set<TableRowId>> entry : clustersMap.entrySet()) {
					ClassId classId = clusterClassId.get(entry.getKey());
					List<TableRowId> rowList = new ArrayList<>(entry.getValue());
					RowCluster rc = new RowCluster(classId, rowList, new RowClusterSimilarities() {
						
						@Override
						public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
							// TODO Auto-generated method stub
							return null;
						}
						
						@Override
						public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
							return calc.apply(rowList.get(key.getE1()), rowList.get(key.getE2()));
//							return cache.get(new Pair<>(rowMainComponent.get(rowList.get(key.getE1())), rowMainComponent.get(rowList.get(key.getE2()))));
//							return comparePair(rowMainComponent.get(rowList.get(key.getE1())), rowMainComponent.get(rowList.get(key.getE2())),  scores, indexToStringMapping, componentMap, componentToIndexMapping, totalSize);

						}

					@Override
					public String getName() {
						// TODO Auto-generated method stub
						return null;
					}
					});
					
					
					List<RowCluster> ret = gc.rowSplitter(rc, new ClusterSimilarities() {
						
						@Override
						public Set<Integer> regionQuery(Integer point, double eps) {
							// TODO Auto-generated method stub
							return null;
						}
						
						@Override
					public double get(ClusterEntryKey key) {
							return calc.apply(rowList.get(key.getE1()), rowList.get(key.getE2()));
//							return comparePair(rowMainComponent.get(rowList.get(key.getE1())), rowMainComponent.get(rowList.get(key.getE2())),  scores, indexToStringMapping, componentMap, componentToIndexMapping, totalSize);
						}
					});
					
					for(RowCluster rcEntry : ret) {
						int index = ai.getAndIncrement();
						newClustersMap.put(index, new HashSet<>(rcEntry.getRows()));
						newClusterClassId.put(index, classId);
					}
				}
				return new Blocks(newClustersMap, newClusterClassId);
			 
		}
	 
		
//		TableRowId tableRowIdCheck = new TableRowId(new AnnotatedTableId("1346876860789_1346946768085_4249.arc4550171461300486933#66673628_0_6604262406652303152"), 1);
//		Integer index = clustersIndex.get(tableRowIdCheck);
		
		
	}
	
 
	
	
	private Map<TableRowId, Map<Integer, Integer>> createHierarchyPerComponentToRowMapping(Map<TableRowId, Set<Integer>> rowToComponentMapping,
			Map<Integer, Set<Integer>> componentHierarchyParents) {
		ConcurrentHashMap<TableRowId, Map<Integer, Integer>> ret = new ConcurrentHashMap<>();
		
		PersistentParallel6.runAndFinish(rowToComponentMapping.entrySet(), (rowEntry) -> {
			ret.put(rowEntry.getKey(), createHierarchy(rowEntry.getValue(), componentHierarchyParents));
		}, this.getClass());
		
		return ret;
	}

	public static Map<Integer, Integer> createHierarchy(Set<Integer> input, Map<Integer, Set<Integer>> componentHierarchyParents ) {
		
		Map<Integer, Set<Integer>> parents = new ConcurrentHashMap<>();
		
		
		for(Integer key : input) {
			Set<Integer> set = new HashSet<>(componentHierarchyParents.get(key));
			for(Iterator<Integer> it  = set.iterator() ;it.hasNext();) {
				if(!input.contains(it.next())) {
					it.remove(); 
				}
			}
			parents.put(key, set);
			
		}
		
		
		ConcurrentHashMap<Integer, Integer> ret = new ConcurrentHashMap<>();
		
		MutableInt level = new MutableInt();
		
		Set<Integer> added = Collections.newSetFromMap(new ConcurrentHashMap<>());
		Set<Integer> remaining = Collections.newSetFromMap(new ConcurrentHashMap<>());
		
		for(Entry<Integer, Set<Integer>> entry : parents.entrySet()) {
			if(entry.getValue().isEmpty()) {
				added.add(entry.getKey());
				ret.put(entry.getKey(), level.intValue());
			}
			else {
				remaining.add(entry.getKey());
			}
			
		}
 
		
	 

		MutableReference<Set<Integer>> previousAdded = new MutableReference<Set<Integer>>(added);
		while(!remaining.isEmpty()) {
			level.increment();;
			Set<Integer> newAdded = Collections.newSetFromMap(new ConcurrentHashMap<>());
			
			for(Integer key : remaining ) {
				Set<Integer> remainingParents = parents.get(key) ;
				remainingParents.removeAll(previousAdded.get());
				if(remainingParents.isEmpty()) {
					newAdded.add(key);
					ret.put(key, level.intValue());
				}
			}
	 
			
			previousAdded.set(newAdded);
			added.addAll(newAdded);
			remaining.removeAll(newAdded);
		}
		
		
	
		
		
		
		
		return ret;
	}


	public static Map<Integer, Integer> createHierarchy(Map<Integer, Set<Integer>> componentHierarchyParents ) {
		
		Map<Integer, Set<Integer>> parents = new ConcurrentHashMap<>();
		
		
		
		ConcurrentHashMap<Integer, Integer> ret = new ConcurrentHashMap<>();
		
		MutableInt level = new MutableInt();
		
		Set<Integer> added = Collections.newSetFromMap(new ConcurrentHashMap<>());
		Set<Integer> remaining = Collections.newSetFromMap(new ConcurrentHashMap<>());
		
		PersistentParallel6.runAndFinish(componentHierarchyParents.entrySet(), (entry) -> {
		 
				if(entry.getValue().isEmpty()) {
					added.add(entry.getKey());
					ret.put(entry.getKey(), level.intValue());
				}
				else {
					remaining.add(entry.getKey());
				}
				
				parents.put(entry.getKey(), new HashSet<>(entry.getValue()));
		 
		}, TfIDFClustererClassBlocked.class);
		
		int sizeTogether = added.size() + remaining.size();

		MutableReference<Set<Integer>> previousAdded = new MutableReference<Set<Integer>>(added);
		while(!remaining.isEmpty()) {
			level.increment();;
			Set<Integer> newAdded = Collections.newSetFromMap(new ConcurrentHashMap<>());
			PersistentParallel6.runAndFinish(remaining, (key) -> {
				
				Set<Integer> remainingParents = parents.get(key) ;
				remainingParents.removeAll(previousAdded.get());
				if(remainingParents.isEmpty()) {
					newAdded.add(key);
					ret.put(key, level.intValue());
				}
			}, TfIDFClustererClassBlocked.class);
			
			previousAdded.set(newAdded);
			added.addAll(newAdded);
			remaining.removeAll(newAdded);
		}
		
		
	
		
		
		
		
		return ret;
	}
	
	
	private Map<TableRowId, Set<TableRowId>> createMappingsFromTFIDFNew(Set<TableRowId> candidates, Map<Integer, Double> scores, Map<Integer, Set<TableRowId>> componentToRowMapping,
			Map<TableRowId,  Set<Integer> > rowToComponentMapping, Map<Integer, String> indexToStringMapping,  Map<TableRowId, Integer> rowMainComponent, Map<Integer, List<Integer>> componentMap, Map<List<Integer>, Integer> componentToIndexMapping, double maxComponentOccurence) {
		
		double threshold = getOtsuThreshold(scores);
		
//		threshold = 1.0;
		
		Set<Integer> filtered = scores.entrySet().stream().filter((e) -> e.getValue() > threshold).map(Entry::getKey).collect(Collectors.toSet());
		ConcurrentHashMap<TableRowId, Set<TableRowId>> mappings = new ConcurrentHashMap<>();
		
		Set<Integer> components = Collections.newSetFromMap(new ConcurrentHashMap<>());
		
		PersistentParallel6.runAndFinish(candidates , (rowId) -> {
			if(rowToComponentMapping.containsKey(rowId)) {
				components.addAll(rowToComponentMapping.get(rowId).stream().filter(e-> filtered.contains(e)).collect(Collectors.toSet()));
			}
		 
		}, this.getClass(), "CreateMaps1");
		
		
		PersistentParallel6.runAndFinish(components, (component) -> {
			Iterator<TableRowId> iterator = componentToRowMapping.get(component).iterator();
			TableRowId rowId = null;
			while(rowId == null && iterator.hasNext()) {
				TableRowId temp = iterator.next();
				if(candidates.contains(temp)) {
					rowId = temp;
				}
			} 
			if(rowId != null) {
				 Set<TableRowId> set = CHMHelper.getFromCHM(mappings, rowId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>()));
				 set.add(rowId);
				while(iterator.hasNext()) {
					set.add(iterator.next());
				}
			}
		}, this.getClass(), "CreateMaps");
		
		
//		double totalSize = rowToComponentMapping.size();
//
//
//		
//		Set<Integer> filteredScores = scores.entrySet().stream().filter(
//				(e) -> { 
//					String s = indexToStringMapping.get(e.getKey());		
//					if(e.getValue() <= 1.0) {
//						int sdf= 123;
//					}
//					return e.getValue() > 1.0;	
//				}
//				
//				).map(Entry::getKey).collect(Collectors.toSet());
//		
//		Set<Set<Integer>> done = Collections.newSetFromMap(new ConcurrentHashMap<>());
//		
//		final InstanceId instanceId = new InstanceId("http://dbpedia.org/resource/My_Boo_(Usher_and_Alicia_Keys_song)");
//		
//		PersistentParallel6.runAndFinish(rowToComponentMapping.entrySet(), (rowEntry) -> {
////			Optional<InstanceId> optional = tableRepository.getMatchedInstanceId(rowEntry.getKey());
////		if(optional.isPresent() && optional.get().equals(instanceId)) {
//			
//			Set<Integer> relevantComponents = rowEntry.getValue().stream().filter((e) -> filteredScores.contains(e)).collect(Collectors.toSet());
//			
//			
//			if(!done.contains(relevantComponents)) {
//				Set<TableRowId> candidates = new HashSet<>();
//				
//				for(Integer component : relevantComponents) {
//					candidates.addAll(componentToRowMapping.get(component));
//				}
//				
//				Set<TableRowId> finalMapping = new HashSet<>();
//				for(TableRowId candidate : candidates) {
//					
//					StringObject rowLabel = LabelRowIndexClusterMapCreator.getLabel(tableRepository, rowEntry.getKey()); 
//					StringObject candidateLabel = LabelRowIndexClusterMapCreator.getLabel(tableRepository, candidate); 
//					
//					
//					Optional<InstanceId> optional2 = tableRepository.getMatchedInstanceId(candidate);
//					if(optional2.isPresent() && optional2.get().equals(instanceId)) {
//						int sdf = 123;
//					}
//					
//					double sum = comparePair(rowMainComponent.get(rowEntry.getKey()), rowMainComponent.get(candidate), scores, indexToStringMapping,  componentMap, componentToIndexMapping,maxComponentOccurence);
//					
//					boolean equal = false;
//					if(sum > 0.0) 
//						equal = true;
//					if(equal) {
//						finalMapping.add(candidate);
//					}
////					else {
////						Optional<InstanceId> optional2 = tableRepository.getMatchedInstanceId(candidate);
////						if(optional2.isPresent() && optional2.get().equals(instanceId)) {
////							int sdf = 123;
////						}
////					}
//				}
//				
//					
//				 mappings.put(rowEntry.getKey(), finalMapping);
//				 
//				 done.add(relevantComponents);
////			}
//		}
//	
//	
//			
//			
//			
//		}, this.getClass(), "CreateMappings");
		
		
		
		return mappings;
	}
	
	
	private double getOtsuThreshold(Map<Integer, Double> scores) {
		OtsuThresholderRounded thresholder = new OtsuThresholderRounded();
		double threshold = thresholder.getThresholdDouble(scores.values());
		
		return threshold;
	}

	private Map<TableRowId, Set<TableRowId>> createMappingsFromTFIDF(Set<TableRowId> inputCandidates, Map<Integer, Double> scores, Map<Integer, Set<TableRowId>> componentToRowMapping,
			Map<TableRowId,  Set<Integer> > rowToComponentMapping, Map<Integer, String> indexToStringMapping,  Map<TableRowId, Integer> rowMainComponent, Map<Integer, List<Integer>> componentMap, Map<List<Integer>, Integer> componentToIndexMapping, double maxComponentOccurence) {
		
		
		
		ConcurrentHashMap<TableRowId, Set<TableRowId>> mappings = new ConcurrentHashMap<>();
		
		
		double totalSize = rowToComponentMapping.size();


		double threshold = getOtsuThreshold(scores);
		Set<Integer> filteredScores = scores.entrySet().stream().filter(
				(e) -> { 
					String s = indexToStringMapping.get(e.getKey());		
					if(e.getValue() <= 1.0) {
						int sdf= 123;
					}
					return e.getValue() > 1.3;	
				}
				
				).map(Entry::getKey).collect(Collectors.toSet());
		
		ConcurrentHashMap<Set<Integer>, TableRowId> done = new ConcurrentHashMap<>();
		
		final InstanceId instanceId = new InstanceId("http://dbpedia.org/resource/My_Boo_(Usher_and_Alicia_Keys_song)");
		
		PersistentParallel6.runAndFinish(inputCandidates, (inputCandidate) -> {
			
//			Optional<InstanceId> optional = tableRepository.getMatchedInstanceId(rowEntry.getKey());
//		if(optional.isPresent() && optional.get().equals(instanceId)) {
			if(!rowToComponentMapping.containsKey(inputCandidate)) { 
				System.out.println("This row doesn't have a component");
				return;
			}
			Set<Integer> relevantComponents = rowToComponentMapping.get(inputCandidate).stream().filter((e) -> filteredScores.contains(e)).collect(Collectors.toSet());
			
			
			if(!done.contains(relevantComponents)) {
				TableRowId cand = CHMHelper.getFromCHM(done, relevantComponents, () -> inputCandidate); 
				
				if(!inputCandidate.equals(cand)) {
					return;
				}
				
				Set<TableRowId> candidates = new HashSet<>();
				
				for(Integer component : relevantComponents) {
					for(TableRowId rowId : componentToRowMapping.get(component)) {
						if(inputCandidates.contains(rowId))
							candidates.add(rowId);
					}
				}
				
				Set<TableRowId> finalMapping = new HashSet<>();
				for(TableRowId candidate : candidates) {
					
//					StringObject rowLabel = LabelRowIndexClusterMapCreator.getLabel(tableRepository, inputCandidate); 
//					StringObject candidateLabel = LabelRowIndexClusterMapCreator.getLabel(tableRepository, candidate); 
//					
					
//					Optional<InstanceId> optional2 = tableRepository.getMatchedInstanceId(candidate);
//					if(optional2.isPresent() && optional2.get().equals(instanceId)) {
//						int sdf = 123;
//					}
					
					double sum = comparePair(rowMainComponent.get(inputCandidate), rowMainComponent.get(candidate), scores, indexToStringMapping,  componentMap, componentToIndexMapping,maxComponentOccurence);
					
					boolean equal = false;
					if(sum > 0.0) 
						equal = true;
					if(equal) {
						finalMapping.add(candidate);
					}
//					else {
//						Optional<InstanceId> optional2 = tableRepository.getMatchedInstanceId(candidate);
//						if(optional2.isPresent() && optional2.get().equals(instanceId)) {
//							int sdf = 123;
//						}
//					}
				}
				
					
				 mappings.put(inputCandidate, finalMapping);
				 
				 
//			}
		}
	
	
			
			
			
		}, this.getClass(), "CreateMappings");
		
		
		
		return mappings;
	}

	private static ConcurrentHashMap<Integer, Map<Integer,Set<Integer>>> indexCache = new ConcurrentHashMap<>();
	
	private Double comparePair(Integer a, Integer b, Map<Integer, Double> scores, Map<Integer, String> indexToStringMapping,
			Map<Integer, List<Integer>> componentMap, Map<List<Integer>, Integer> componentToIndexMapping, double totalSize) {

		
//		
//		components.addAll(getMaxComponents(a,b,aLevels));
//		components.addAll(getMaxComponents(b,a,bLevels));
//		
	
		List<Integer> componentsA = new ArrayList<>(componentMap.get(a));
		List<Integer> componentsB = new ArrayList<>(componentMap.get(b));
		
		String aString = indexToStringMapping.get(a);
		String bString = indexToStringMapping.get(b);
		
		List<Pair<List<Integer>, ComponentExist>>  components = componentCreator.create(componentsA, componentsB, CHMHelper.getFromCHM(indexCache, a, () -> componentCreator.createIndex(componentsA)), CHMHelper.getFromCHM(indexCache, a, () -> componentCreator.createIndex(componentsB)), true, false, false);
		double totalSizeA = componentsA.size();
		double totalSizeB = componentsB.size();
		double normalizer = Math.log(totalSize);
		double aWeight = ( (double) componentsB.size() ) / (componentsA.size() + componentsB.size());//1.0;//((double) componentsB.size()) / componentsA.size();
		double bWeight =  ( (double) componentsA.size() )/ (componentsA.size() + componentsB.size());//1.0;//((double) componentsA.size()) / componentsB.size();
		double sum = 0.0;
		int positive = 0;
		double negative = 0;
		double weighting = 0.0;
		for(Pair<List<Integer>, ComponentExist> component : components) { 
			weighting +=  component.getFirst().size();
		}
		for(Pair<List<Integer>, ComponentExist> component : components) { 
			try {
			Integer index = componentToIndexMapping.get( component.getFirst() );
			
			
			if(index != null) { 
			
				Double componentScore = scores.get(index);
				 
				if(componentScore != null) {
					componentScore = (componentScore / normalizer) * (component.getFirst().size() / weighting);
					double scoreA = ((1 + Math.log(component.getFirst().size())) * componentScore) ;
					scoreA = ((component.getFirst().size() / totalSizeA) * componentScore) / normalizer ;
					double scoreB = ((1 + Math.log(component.getFirst().size())) * componentScore) ;
					scoreB = ((component.getFirst().size() / totalSizeB) * componentScore)  / normalizer ;
					String string = indexToStringMapping.get(index);
					
					switch(component.getSecond()) {
					case BOTH:
						sum +=  componentScore;
						positive++;
						break;
					case ONLY_A:
	//					componentScore *= aWeight;
						sum -= componentScore;// * aWeight;
						negative += 1.0;//aWeight;
						break;
					case ONLY_B:
	//					componentScore *= bWeight;
						sum -= componentScore;// * bWeight;
						negative += 1.0;//bWeight;
						break; 
					
					}
				}
			}
			}
			catch(NullPointerException e) {
				String s = "";
				for(Integer i : component.getFirst()) {
					s += " " + indexToStringMapping.get(i);
				}
				e.printStackTrace();
			}
			
		 
			
			
		}
		if(sum > 0.0) {
			return sum; //  / positive;
		}
		else {
			return sum; // / negative;
		}
	}

	private boolean comparePair(Set<Integer> a, Set<Integer> b,
			Map<Integer, Double> scores, Map<Integer, String> indexToStringMapping, Map<Integer, List<Integer>> componentMap) {
		
		
		
//		
//		components.addAll(getMaxComponents(a,b,aLevels));
//		components.addAll(getMaxComponents(b,a,bLevels));
//		
	
		List<Integer> componentsA = new ArrayList<>(componentMap.get(a.iterator().next()));
		List<Integer> componentsB = new ArrayList<>(componentMap.get(a.iterator().next()));
		
		 Set<Integer>   components = new HashSet<>();
		
		double aWeight = ((double) b.size()) / a.size();
		double bWeight = ((double) a.size()) / b.size();
		double sum = 0.0;
		for(Integer component : components) { 
			
			String string = indexToStringMapping.get(component);
			
			boolean inA = a.contains(component);
			boolean inB = b.contains(component);
			double componentScore = scores.get(component);
			 
			if(inA && inB) {
				sum += componentScore;
			}
			else {
				if(inA) {
					componentScore *= aWeight;
				}
				else {
					componentScore *= bWeight;
				}
				sum -= componentScore;
			}
			
			
		}
	
		if(sum > 0.0) 
			return true;
		else
			return false;
		
	}

	private Set<Integer> getMaxComponents(Set<Integer> a, Set<Integer> b, Map<Integer, Integer> levels) {
		int maxLevel = -1;
		int maxLevelNoOverlap = -1;
//		Set<Integer> maxLevelSet = new HashSet<>();
//		Set<Integer> maxLevelNoOverlapSet = new HashSet<>();
		
		for(Integer component : a) {
			int level = levels.get(component);
			
			if(b.contains(component) ) {
				if(level == maxLevel ) {
//					maxLevelSet.add(component);
				} else if (maxLevel == -1 || level < maxLevel) {
					maxLevel = level;
//					maxLevelSet = new HashSet<>();
//					maxLevelSet.add(component);
				}

			}
			
			if(level == maxLevelNoOverlap ) {
//				maxLevelNoOverlapSet.add(component);
			} else if (maxLevelNoOverlap == -1 || level < maxLevelNoOverlap) {
				maxLevelNoOverlap = level;
//				maxLevelNoOverlapSet = new HashSet<>();
//				maxLevelNoOverlapSet.add(component);
			}
		}
		int retLevel;
		if(maxLevel > -1) {
			retLevel = maxLevel;
			
		}
		else {
			retLevel = maxLevelNoOverlap;
		}
		Set<Integer> ret = new HashSet<>();
		for(Integer component : a) { 
			int level = levels.get(component);
			if(level == retLevel) {
				ret.add(component);
			}
		}
		
		return ret;
		
	}
	
	private static class TfIdfStuff {
		
	}
	
 
}
