package expansion.set.strategies.rowclusterers.strategies.association;

import java.io.File;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import javax.inject.Inject;
import javax.inject.Named;

import org.apache.commons.lang3.mutable.MutableDouble;
import org.apache.commons.lang3.mutable.MutableInt;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.utils.kryo.CHMKryoAdapter;
import expansion.set.SetExpansionProperties;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.clustering.model.RowCluster;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.strategies.rowclusterers.strategies.overlap.RowToIndexMapper;
import expansion.set.strategies.rowclusterers.strategies.overlap.RowToIndexMapper.Mapping;
import expansion.set.strategies.rowclusterers.strategies.overlap.RowToIndexMapper.MappingType;

public class AssociationSimilarities {
	public static enum AssociationType {
		PHI, PMI, TF_IDF;
	}

	
	private final RowToIndexMapper mapper;
	
	@Inject
	@Named(SetExpansionProperties.tablesCachesFolder)
	String cacheFolder;
	
	@Inject
	public AssociationSimilarities(RowToIndexMapper mapper) {
		this.mapper = mapper;
	}
	
	private final static  SimpleLockableConcurrentCache<Pair<UuidStringId, AssociationType>, AssociationScores> labelCache = new SimpleLockableConcurrentCache<>();
	private final static SimpleLockableConcurrentCache<Pair<UuidStringId, AssociationType>, AssociationScores> labelCacheWithoutCache = new SimpleLockableConcurrentCache<>();
	private final static  SimpleLockableConcurrentCache<Pair<UUID, AssociationType>, AssociationScores> instanceCache = new SimpleLockableConcurrentCache<>();

	
	private final static  SimpleLockableConcurrentCache<UuidStringId, Pair<TableMapping, Map<AssociationPair, Integer>>> labelAssociationsScore = new SimpleLockableConcurrentCache<>();
	private final static  SimpleLockableConcurrentCache<UUID, Pair<TableMapping,  Map<AssociationPair, Integer>>> instanceAssociationsScore = new SimpleLockableConcurrentCache<>();

	@Inject
	protected CHMKryoAdapter chmKryoAdapter;
	
	public AssociationScores getScores(RowClusteringConfig config, ClustererInput clustererInput, MappingType mappingType, AssociationType associationType) {
		switch(mappingType) {
		case INSTANCE_CORRESPONDENCE_THEN_LABEL:
			return instanceCache.getOrCreate(new Pair<>(clustererInput.getUuid(),  associationType),  () -> {
				return getScoresInner("INSTANCE", config, clustererInput, mappingType, associationType);
			});
		case LABEL_ONLY:
			if (config.isWithCache()) {
				return labelCache.getOrCreate(new Pair<>(clustererInput.getFullCandidates().getId(), associationType), () -> {
					return getScoresInner("LABEL", config, clustererInput, mappingType, associationType);
				});
			} else {
				return labelCacheWithoutCache.getOrCreate(new Pair<>(clustererInput.getFullCandidates().getId(), associationType), () -> {
					return getScoresInner("LABEL", config, clustererInput, mappingType, associationType);
				});
			}
		
		}
		
		throw new IllegalStateException();
		
		
//		return cache.getOrCreate(new Tri<>(tableIdSet.getId(), mappingType, associationType),  () -> {
//			Mapping mapping = mapper.get(clustererInput, mappingType);
//			
//			Map<AssociationPair, Integer> associations = associationsScore.getOrCreate(new Pair<>(tableIdSet.getId(), mappingType),
//				() -> {
//					return getAssociations(mapping, associationType);
//				}
//			);
//			AssociationCalculator calculator =  null;
//			switch(associationType) {
//				case PHI:
//					calculator  =  AssociationSimilarities::getPhi;
//					break;
//			case PMI:
//				calculator  =  AssociationSimilarities::getPmi;
//				break;
//			case TF_IDF:
//				calculator  =  AssociationSimilarities::getTfIdf;
//				break;
//			default:
//				throw new IllegalStateException();
//			 
//		 
//			
//			}
//			
//			
//			return new AssociationScores(getTableScore(tableIdSet.getTableIds(), mapping, associations,calculator));
//			
////			AssociationCalculator calculator = new AssociationCalculator(associations);
////
////			return new AssociationScores(calculator.get());
//		}); 
		
		
		
	}

	public static Map<ClassId, Set<AnnotatedTableId>> tableSet;
	
	private AssociationScores getScoresInner(String mappingPrefix, RowClusteringConfig config,ClustererInput clustererInput, MappingType mappingType, AssociationType associationType) {
		
		
		
		
		
		String aaFolder = this.cacheFolder + "/PHI/";
		MKDIR.createDirUnchecked(new File(aaFolder));
		
		AssociationCalculator calculator =  null;
		String prefix = "AA_" + mappingPrefix + "_" + clustererInput.getClassId().getFinalPart() + "_";
		String name = "AA_" + mappingPrefix + "_";
		switch(associationType) {
			case PHI:
				prefix += "PHI";
			name += "PHI";
				calculator  =  AssociationSimilarities::getPhi;
				break;
		case PMI:
			prefix += "PMI";
			name += "PMI";
			calculator  =  AssociationSimilarities::getPmi;
			break;
		case TF_IDF:
			prefix += "TF_IDF";
			name += "TF_IDF";
			calculator  =  AssociationSimilarities::getTfIdf;
			break;
		default:
			throw new IllegalStateException();
		 
	 
		
		}
		Map<AnnotatedTableId, Map<Integer, Double>> tableScores;
		
		if(chmKryoAdapter.cacheExists(aaFolder, prefix)) {
			tableScores = chmKryoAdapter.readWithMap(AnnotatedTableId.class, Integer.class, Double.class, aaFolder, prefix);
		}
		else {
			Mapping mapping = mapper.get(config, clustererInput, mappingType);


			Pair<TableMapping, Map<AssociationPair, Integer>> associations = getTableAndAssociationScores(clustererInput, mapping, mappingType);
			Set<AnnotatedTableId> tableSetForClass;
			if (clustererInput.getFullTableIds() != null) {
				tableSetForClass = clustererInput.getFullTableIds().getTableIds();
			} else if (clustererInput.getTableIds() != null) {
				tableSetForClass = clustererInput.getTableIds().getTableIds();

			} else {
				tableSetForClass = clustererInput.getFullCandidates().getRowIds().stream().map((e) -> e.getTableId()).collect(Collectors.toSet());
			}
			// // Set<AnnotatedTableId> tableSet =
			// //
			// clustererInput.getFullCandidates().getRowIds().stream().map((e)
			// // -> e.getTableId()).collect(Collectors.toSet());
			// Set<AnnotatedTableId> tableSetForClass =
			// tableSet.get(clustererInput.getClassId());

			tableScores = getTableScoreNew(tableSetForClass, associations.getFirst(), associations.getSecond(), calculator);
			 
			tableScores = tableScores.entrySet().stream().filter((e) -> tableSetForClass.contains(e.getKey())).filter((e) -> {
				if (e.getValue().isEmpty()) {
					return false;
				} else {
					return true;
				}
			}).collect(Collectors.toMap(Entry::getKey, Entry::getValue));
			 chmKryoAdapter.writeWithMap(aaFolder, prefix, tableScores);
		}
		

		return new AssociationScores(tableScores, name, clustererInput.getCandidates(), config.isWithCache());
	}
	
	// Pair<TableMapping, Map<AssociationPair, Integer>>
	// combineMappings(Pair<TableMapping, Map<AssociationPair, Integer>>
	// associations) {
//		
//		
//		Map<Integer, Set<AnnotatedTableId>> blocks = associations.getFirst().getIndexToTable();
//		
//		ConcurrentHashMap<AnnotatedTableId, ConcurrentHashMap<AnnotatedTableId, Set<Integer>>> map = new ConcurrentHashMap<>();
//		
//		PersistentParallel6.runAndFinish(associations.getSecond().entrySet(), (e) ->  {
//			int sizeA = blocks.get(e.getKey().getA()).size();
//			int sizeB = blocks.get(e.getKey().getB()).size();
//			double intersectoin = e.getValue();
//			
//			if(intersection / sizeA)
//			SortedSet<AnnotatedTableId> sorted = new TreeSet<>((e1, e2) -> e1.getTableName().compareTo(e2.getTableName()));
//			 
//			for(AnnotatedTableId id : sorted) {
//				if(currentMap != null) {
//					
//				}
//				CHMHelper.getFromCHM(map, id, () -> new ConcurrentHashMap<>()).put(key, value);
//			}
//		}, this.getClass());
//		
//		return null;
//	}
	Pair<TableMapping, Map<AssociationPair, Integer>> getTableAndAssociationScores(ClustererInput clustererInput, Mapping mapping, MappingType mappingType) {
		switch(mappingType) {
		case INSTANCE_CORRESPONDENCE_THEN_LABEL:
			return instanceAssociationsScore.getOrCreate(clustererInput.getUuid(), () -> {
						TableMapping tableMapping = new TableMapping(mapping);
						return new Pair<>(tableMapping,  getAssociations(tableMapping));
					}
				);
		case LABEL_ONLY:
			return labelAssociationsScore.getOrCreate(clustererInput.getTableIds().getId(),
					() -> {
						TableMapping tableMapping = new TableMapping(mapping);
						return  new Pair<>(tableMapping,getAssociations(tableMapping));
					}
				);
		
		}
		
		throw new IllegalStateException();
		
//		associationsScore.getOrCreate(new Pair<>(tableIdSet.getId(), mappingType),
//				() -> {
//					return getAssociations(mapping, associationType);
//				}
//			);
	}
	
	private Map<AnnotatedTableId, Map<Integer, Double>> getTableScoreNew(  Set<AnnotatedTableId> tableIds, TableMapping mapping, Map<AssociationPair, Integer> associations,
			AssociationCalculator calculator) {
		
		
		ConcurrentHashMap<Integer, Map<Integer, Integer>> indexedIntersection = new ConcurrentHashMap< >();
		
		PersistentParallel6.runAndFinish(associations.entrySet(), (pair) -> {
			CHMHelper.getFromCHM(indexedIntersection, pair.getKey().getA(), () ->  new ConcurrentHashMap<>() ).put(pair.getKey().getB(), pair.getValue());
			CHMHelper.getFromCHM(indexedIntersection, pair.getKey().getB(), () ->  new ConcurrentHashMap<>() ).put(pair.getKey().getA(), pair.getValue());
		}, this.getClass());
		
		Map<AssociationPair, Double> associationScores = new ConcurrentHashMap<>();
		ConcurrentHashMap<Integer, Set<Integer>> index = new ConcurrentHashMap<>();
		
		Map<AnnotatedTableId, Set<Integer>> tableMappings = mapping.getTableToIndex();
		ConcurrentHashMap<AnnotatedTableId, Map<Integer, Double>> tableScores = new ConcurrentHashMap<>();
		
		PersistentParallel6.runAndFinish(tableIds, (tableId) -> {
			
			Set<Integer> tableBlocks = tableMappings.get(tableId);
			
			
			Map<Integer, Integer> tableBlocksWithSize = new HashMap<>();
 
			
			Map<Integer, MutableDouble> blockSum = new HashMap<>();
			Map<Integer, MutableInt> blockCount = new HashMap<>();
			
			
			
			for(Integer tableBlock :  tableBlocks) {
				int tableBlockSize = mapping.getForIndex(tableBlock).size();
				
				Map<Integer, Integer> intersections = indexedIntersection.get(tableBlock);
				 if(intersections != null) {
					 
					 for(Entry<Integer, Integer> intersectionEntry : intersections.entrySet()) {
						int intersection = intersectionEntry.getValue();
						
						double score = calculator.compute(mapping.getTableToIndex().size(), intersection,
								mapping.getForIndex(intersectionEntry.getKey()).size(), tableBlockSize);
						CHMHelper.getFromMap(blockSum, intersectionEntry.getKey(), () -> new MutableDouble()).add(score);
						CHMHelper.getFromMap(blockCount, intersectionEntry.getKey(), () -> new MutableInt()).increment();
					 }
				 }
				
			 
				
			}
			
			Map<Integer, Double> tableVector = new HashMap<>();
			for(Entry<Integer, MutableInt> block : blockCount.entrySet()) {
				double score = blockSum.get(block.getKey()).doubleValue() / block.getValue().intValue();
				if (score > 0.5) {
					tableVector.put(block.getKey(), score);
				}
			
				
			}
			
//			for(Integer tableBlock :  tableEntry.getValue()) {
//				tableVector.put(tableBlock, 1.0);
//			}
			
			
			
			
			tableScores.put(tableId, tableVector);
			
		}, this.getClass());
//		
//		PersistentParallel6.runAndFinish(associations.entrySet(), (e) -> {
//			Double score = calculator.compute(
//					mapping.getTableToIndex().size(),
//					e.getValue(),
//					mapping.getIndexToTable().get(e.getKey().getA()).size(), mapping.getIndexToTable().get(e.getKey().getB()).size());
//			
//			if(Math.abs(score) > 0.5) {
//				associationScores.put(e.getKey(), score);
//				CHMHelper.getFromCHM(index, e.getKey().getA(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(e.getKey().getB());
//				CHMHelper.getFromCHM(index, e.getKey().getB(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(e.getKey().getA());
//			}
//			
//		}, this.getClass());
//
//
//	
//		PersistentParallel6.runAndFinish(set, (tid) -> {
//			Map<Integer, MutableDouble> tableMap = new HashMap<>();
//			Set<Integer> rows = mapping.getTableToIndex().get(tid);
//			 
//			for(Integer row : rows) {
//				CHMHelper.getFromMap(tableMap, row, () -> new MutableDouble()).add(1.0);
//				
//				if(index.containsKey(row)) {
//					for(Integer innerRow : index.get(row)) {
//						CHMHelper.getFromMap(tableMap, innerRow, () -> new MutableDouble()).add(associationScores.get(new AssociationPair(row, innerRow)));
//					}
//				}
//				
//			}
//			
//			Map<Integer, Double> tableScoreMap = new HashMap<>();
//			for(Entry<Integer, MutableDouble> entry : tableMap.entrySet()) {
//				tableScoreMap.put(entry.getKey(), entry.getValue().doubleValue() / rows.size());
//			}
//			tableScores.put(tid, tableScoreMap);
//		}, this.getClass());
//		
		
		return tableScores;
	}
	
	
	private Map<AnnotatedTableId, Map<Integer, Double>> getTableScore(Set<AnnotatedTableId> set, TableMapping mapping, Map<AssociationPair, Integer> associations,
			AssociationCalculator calculator) {
		Map<AssociationPair, Double> associationScores = new ConcurrentHashMap<>();
		ConcurrentHashMap<Integer, Set<Integer>> index = new ConcurrentHashMap<>();
		
		PersistentParallel6.runAndFinish(associations.entrySet(), (e) -> {
			Double score = calculator.compute(
					mapping.getTableToIndex().size(),
					e.getValue(),
					mapping.getIndexToTable().get(e.getKey().getA()).size(), mapping.getIndexToTable().get(e.getKey().getB()).size());
			
			if(Math.abs(score) > 0.5) {
				associationScores.put(e.getKey(), score);
				CHMHelper.getFromCHM(index, e.getKey().getA(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(e.getKey().getB());
				CHMHelper.getFromCHM(index, e.getKey().getB(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(e.getKey().getA());
			}
			
		}, this.getClass());


		ConcurrentHashMap<AnnotatedTableId, Map<Integer, Double>> tableScores = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(set, (tid) -> {
			Map<Integer, MutableDouble> tableMap = new HashMap<>();
			Set<Integer> rows = mapping.getTableToIndex().get(tid);
			 
			for(Integer row : rows) {
				CHMHelper.getFromMap(tableMap, row, () -> new MutableDouble()).add(1.0);
				
				if(index.containsKey(row)) {
					for(Integer innerRow : index.get(row)) {
						CHMHelper.getFromMap(tableMap, innerRow, () -> new MutableDouble()).add(associationScores.get(new AssociationPair(row, innerRow)));
					}
				}
				
			}
			
			Map<Integer, Double> tableScoreMap = new HashMap<>();
			for(Entry<Integer, MutableDouble> entry : tableMap.entrySet()) {
				tableScoreMap.put(entry.getKey(), entry.getValue().doubleValue() / rows.size());
			}
			tableScores.put(tid, tableScoreMap);
		}, this.getClass());
		
		
		return tableScores;
	}

	
	@FunctionalInterface
	private static interface AssociationCalculator {
		Double compute(double totalNumberOfRows, double intersection, double sizeOfA, double sizeOfB);
	}
	
	
	@FunctionalInterface
	private static interface AssociationDefaultValue {
		Double compute(int totalNumberOfRows, int sizeOfA, int sizeOfB);
	}
	
	private static Double getPhi(double totalNumberOfRows, double intersection, double sizeOfA, double sizeOfB) {
		double nAB = intersection;
		double nA0 = sizeOfA - intersection;
		double n0B = sizeOfB - intersection;
		double n00 = totalNumberOfRows - sizeOfA - sizeOfB + intersection;
		double nAF = sizeOfA;
		double nAR = totalNumberOfRows - sizeOfA;
		double nBF = sizeOfB;
		double nBR = totalNumberOfRows - sizeOfB;
 
		double componentA = (nAB * n00) - (nA0 * n0B);
		double componentBRaw = nAF * nAR * nBF * nBR;
		double componentB = Math.sqrt(componentBRaw);
		double phi = componentA / componentB;
		return phi;
	}
	
	private static Double getPmi(double totalNumberOfRows, double intersection, double sizeOfA, double sizeOfB) {
		double phi = getPhi(totalNumberOfRows, intersection, sizeOfA, sizeOfB);
		if(intersection == 0)
			return -1.0;
		
		double total = totalNumberOfRows;
		double componentA = intersection / total;
		double componentB = sizeOfA / total;
		double componentC = sizeOfB / total;
		
		double pmiScore = Math.log(componentA / (componentB * componentC)) / (-Math.log(componentA));
		return pmiScore;
	}
	
	private static Double getTfIdf(double totalNumberOfRows, double intersection, double sizeOfA, double sizeOfB) {
		if(intersection == 0)
			return -1.0;
		
		double a = sizeOfA;
		double tf = 1.0 + Math.log(intersection  / a);
		double idf = Math.log(1.0 + (totalNumberOfRows / ((double)sizeOfB ) ));
		 
		double n_tf_idf = ( tf * idf ) / (Math.log( (double) 1.0) - 1.0 );
		return n_tf_idf;
	}
 
	
	
	
	public ConcurrentHashMap<AssociationPair, Integer> getAssociations(TableMapping mapping) {
		
		final ConcurrentHashMap<AssociationPair, Integer> intersections = new ConcurrentHashMap<>();
		
		Set<AssociationPair> tasks = Collections.newSetFromMap(new ConcurrentHashMap<>());
		
		
		PersistentParallel6.runAndFinish(mapping.getIndexToTable().entrySet(), (index) -> {
			
			for(AnnotatedTableId tid : index.getValue()) {
				for(Integer secondIndex : mapping.getTableToIndex().get(tid)) {
					tasks.add(new AssociationPair(index.getKey(), secondIndex));
				}
			}
			
		}, this.getClass());
//		ConcurrentHashMap<Index, V>
//		
//		mapping.getIndexToRows()
		PersistentParallel6.runAndFinish(tasks, (entry) -> {
			
			Set<AnnotatedTableId> firstSet = mapping.getIndexToTable().get(entry.getA());
			Set<AnnotatedTableId> secondSet = mapping.getIndexToTable().get(entry.getB());
			
			Optional<Integer> intersection;
			if(firstSet.size() < secondSet.size()) {
				intersection = getIntersection(firstSet, secondSet);
			}
			else {
				intersection = getIntersection(secondSet, firstSet);
			}
			
			if(intersection.isPresent()) {
				intersections.put(entry, intersection.get());
			}
		}, this.getClass());
		
		return intersections;
		
	}
	
	public Optional<Integer> getIntersection(Set<AnnotatedTableId> smallerSet, Set<AnnotatedTableId> largerSet) {
		int intersection = 0;
		for(AnnotatedTableId tableId: smallerSet) {
			if( largerSet.contains(tableId))  {
				intersection++;
			}
		}
		
		if(intersection == 0)
			return Optional.empty();
		
//		if(intersection > smallerSet.size() / 2) {
			return Optional.of(intersection);
//		}
			
		
	}


	public static class AssociationScores  implements RowClusterSimilarities {
 
		private final Map<AnnotatedTableId, Map<Integer, Double>> scores; 
		private final ConcurrentHashMap<TableIdPair, Double > cache;

		private final String name;
		private final boolean withCache;
		
		//
		private final Map<AnnotatedTableId, Set<Integer>> tableToConceptMapping;
		public AssociationScores(Map<AnnotatedTableId, Map<Integer, Double>> scores, String name,
				Set<TableRowId> candidates, boolean withCache) {
			super();
			this.scores = scores; 
			cache = new ConcurrentHashMap<>();
			this.withCache = true;

			// ConcurrentHashMap<TableRowId, Set<Integer>> blockMap = new
			// ConcurrentHashMap<>();
			Map<AnnotatedTableId, Set<Integer>> tableToConceptMapping = scores.entrySet().stream()
					.collect(Collectors.toMap(Entry::getKey, (e) -> {
						return e.getValue().entrySet().stream().filter((e2) -> e2.getValue() > 0.70)
								.map((e2) -> e2.getKey())
								.collect(Collectors.toSet());
					}));
			this.tableToConceptMapping = tableToConceptMapping;
			
			// PersistentParallel6.runAndFinish(DoubleIteraterable.iterable(scores.keySet()),
			// (doubleEntry) -> {
			// cache.put(new TableIdPair(doubleEntry.getFirst(),
			// doubleEntry.getSecond()), compute(doubleEntry.getFirst(),
			// doubleEntry.getSecond()));
			// }, this.getClass(), 50);
			this.name = name;
		}

		
		public double compute(AnnotatedTableId tid1, AnnotatedTableId tid2) { 
			
			
			
			Map<Integer, Double> scores1 = scores.get(tid1);
			if (scores1 == null) {
				return 0.0;
			}
			Map<Integer, Double> scores2 = scores.get(tid2);
			if (scores2 == null) {
				return 0.0;
			}
			
			// Set<Integer> keySet = new HashSet<>(scores1.keySet());
			// keySet.addAll(scores2.keySet());
			double e1 = 0.0;
			double e2 = 0.0;
		 
			double ret = 0.0;
			double denominator = 0.0;
			double diff = 0.0;

			Set<Integer> bigger;
			Set<Integer> smaller;
			if (scores1.size() > scores2.size()) {
				bigger = scores1.keySet();
				smaller = scores2.keySet();
			} else {
				smaller = scores1.keySet();
				bigger = scores2.keySet();
			}

			for (Integer index : bigger) {
				Double s1 = scores1.get(index );
			
				Double s2 = scores2.get(index );
				if(s1 == null) {
					s1 = 0.0;
				}
				else if (s2 == null) {
					s2 = 0.0;
				}
				
				
				diff += Math.abs(s1 - s2);
				
				ret += s1  * s2 ;
				
				e1 += Math.pow(s1, 2.0);
				e2 += Math.pow(s2, 2.0);
			}
			
			for (Integer index : smaller) {
				if (!bigger.contains(index)) {
					Double s1 = scores1.get(index);

					Double s2 = scores2.get(index);
					if (s1 == null) {
						s1 = 0.0;
					} else if (s2 == null) {
						s2 = 0.0;
					}

					diff += Math.abs(s1 - s2);

					ret += s1 * s2;

					e1 += Math.pow(s1, 2.0);
					e2 += Math.pow(s2, 2.0);
				}
			}
			
//			for(Entry<Integer, Double> entry : scores1.entrySet()) {
//				e1 += Math.pow(entry.getValue(), 2.0);
//			}
//			
//			for(Entry<Integer, Double> entry : scores2.entrySet()) {
//				e2  += Math.pow(entry.getValue(), 2.0);
//			}
			
			

		
//			for(Double s1 : scores1.values()) {
//			
//			}
//			for(Double s2 : scores2.values()) {
//				
//			}
//			
			
			denominator = Math.sqrt(e1) * Math.sqrt(e2);
			
			// double normalized = diff / keySet.size();
			
			double score = ret / denominator;
			
			if(score < 0.0) {
				int sdf= 123;
			}
			return score;
		}
		
		@Override
		public double get(List<TableRowId> rows, ClassId classId, ClusterEntryKey key) {
			AnnotatedTableId tid1 = rows.get(key.getE1()).getTableId();
			AnnotatedTableId tid2 = rows.get(key.getE2()).getTableId();
			
			if(tid1.equals(tid2)) {
				return 0.0;
			}
			if (withCache) {
				Double score = CHMHelper.getFromCHM(cache, new TableIdPair(tid1, tid2), () -> compute(tid1, tid2));

				return score;
			} else {
				return compute(tid1, tid2);
			}
		}

		@Override
		public Set<TableRowId> regionQuery(RowCluster cluster, TableRowId point, double eps) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getName() {
			return name;
		}
		
		// public List<Tri<Set<Integer>, Integer, Double>>
		// getSplitBlocksWithScores(TableRowId tableRowId) {
		// List<Tri<Set<Integer>, Integer, Double>> list = new ArrayList<>();
		// Set<Integer> blocks = getBlocks(tableRowId);
		//
		// if (blocks == null)
		// blocks = new HashSet<>();
		// // 0.70
		// list.add(new Tri<>(blocks, getBlockMin(tableRowId), 0.00));
		// return list;
		// }
		//
		// public Set<Integer> getBlocks(TableRowId rowId) {
		// return tableToConceptMapping.get(rowId.getTableId());
		// }
		//
		// @Override
		// public int getBlockMin(TableRowId rowId) {
		// Set<Integer> blocks = tableToConceptMapping.get(rowId.getTableId());
		// ;
		// if (blocks != null) {
		// int size = blocks.size();
		// int ret;
		// if (size == 1) {
		// ret = 1;
		// } else {
		// ret = (int) (size * 0.50);
		// }
		// return ret;
		// } else {
		// return 1;
		// }
		// }
		//
		// public boolean supportsBlocking() {
		// return true;
		// }

	}
	
	private static class TableIdPair {
		private final AnnotatedTableId a;
		private final AnnotatedTableId b;
		
		public TableIdPair(AnnotatedTableId a, AnnotatedTableId b) {
			if(a.equals(b)) {
				throw new IllegalStateException();
			}
			
			int compare = a.getTableName().compareTo(b.getTableName());
			
			if(compare > 0) {
				this.a = b;
				this.b = a;
			}
			else {
				this.a = a;
				this.b = b;
			}
		 
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((a == null) ? 0 : a.hashCode());
			result = prime * result + ((b == null) ? 0 : b.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (!(obj instanceof TableIdPair))
				return false;
			TableIdPair other = (TableIdPair) obj;
			if (a == null) {
				if (other.a != null)
					return false;
			} else if (!a.equals(other.a))
				return false;
			if (b == null) {
				if (other.b != null)
					return false;
			} else if (!b.equals(other.b))
				return false;
			return true;
		}
		
		
	}
	private static class AssociationPair {
		private final int a;
		private final int b;
		
		AssociationPair(int a, int b) {
			if(a > b) {
				this.a = a;
				this.b = b;
			}
			else {
				this.b = a;
				this.a = b;
			}
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + a;
			result = prime * result + b;
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (!(obj instanceof AssociationPair))
				return false;
			AssociationPair other = (AssociationPair) obj;
			if (a != other.a)
				return false;
			if (b != other.b)
				return false;
			return true;
		}

		int getA() {
			return a;
		}

		int getB() {
			return b;
		}
		
		
	}
	
	
	private final static class TableMapping {
		private final Map<Integer, Set<AnnotatedTableId>> indexToTable;
		private final Map<AnnotatedTableId, Set<Integer>> tableToIndex; 
	 
		public TableMapping(Mapping mapping) {
			
			indexToTable = new ConcurrentHashMap<>();
			
			PersistentParallel6.runAndFinish(mapping.getIndexToRows().entrySet(), (e) -> {
				Set<AnnotatedTableId> tableIds = new HashSet<>();
				
				for(TableRowId rowId : e.getValue()) {
					tableIds.add(rowId.getTableId());
				}
				
				indexToTable.put(e.getKey(), tableIds);
			}, this.getClass()); 
			
			tableToIndex = mapping.getTableToRowIndeces();
		}

		Map<Integer, Set<AnnotatedTableId>> getIndexToTable() {
			return indexToTable;
		}

		Map<AnnotatedTableId, Set<Integer>> getTableToIndex() {
			return tableToIndex;
		}

		Set<Integer> getForTable(AnnotatedTableId tableId) {
			return tableToIndex.get(tableId);
		}
		
		Set<AnnotatedTableId> getForIndex(Integer index) {
			return indexToTable.get(index);
		}
	 
		 
		
		
	}
	
}
