package expansion.set.strategies.rowclusterers.strategies.overlap;

import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.expansion.utils.UuidStringId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.informatik.dws.expansion.matching.IndexBuilder;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.strategies.LabelSimilarityFullCluster;
import expansion.set.strategies.blockers.Blocks;
import expansion.set.strategies.rowclusterers.CreateClustersFromMappings;
import expansion.set.strategies.rowclusterers.label.LabelRowIndexClusterMapCreator;

public class RowToIndexMapper {

	
	public static enum MappingType {
		LABEL_ONLY, INSTANCE_CORRESPONDENCE_THEN_LABEL;
	}
	
	
	private final static SimpleLockableConcurrentCache<Pair<UuidStringId, MappingType>, Mapping> labelCache = new SimpleLockableConcurrentCache<>();
	
	private final static SimpleLockableConcurrentCache<Pair<UUID, MappingType>, Mapping> instanceCache = new SimpleLockableConcurrentCache<>();

	
	private final AnnotatedTablesRepository tablesRepository;
	private final LabelRowIndexClusterMapCreator labelMapper;
		
	

	@Inject
	protected   LabelSimilarityFullCluster labelClusterer;
	
	@Inject
	public RowToIndexMapper(LabelRowIndexClusterMapCreator labelMapper,  AnnotatedTablesRepository tablesRepository) {
		this.labelMapper = labelMapper;
		this.tablesRepository = tablesRepository;
	}
	
	
	public Mapping get( RowClusteringConfig config,  ClustererInput clustererInput, MappingType mappingType) {
		TableIdsSet tableIdSet = clustererInput.getTableIds();
		
		switch(mappingType) {
		case INSTANCE_CORRESPONDENCE_THEN_LABEL:
			return instanceCache.getOrCreate(new Pair<>(clustererInput.getUuid(), mappingType), () -> {
				
				 
				Set<TableRowId> candidates = this.tablesRepository.getAllRowIds(tableIdSet.getTableIds());
				return getWithInstance(get(config, clustererInput, MappingType.LABEL_ONLY), candidates, clustererInput.getCandidates()); 
				 
				
			}); 
		case LABEL_ONLY:
			return labelCache.getOrCreate(new Pair<>(tableIdSet.getId(), mappingType), () -> {
				
				 
				Set<TableRowId> candidates = this.tablesRepository.getAllRowIds(tableIdSet.getTableIds());
				
//				ClustererInput clustererInputAll = new ClustererInput(clustererInput.getGroupedValuesSets(), candidates, tableIdSet, clustererInput.getGoldStandardClusters(), clustererInput.getClassId());
				return getLabelOnly(clustererInput.getClassId(),candidates); 
				 
				
			});
		
		}
		
		throw new IllegalStateException();
		
		
//		return labelCache.getOrCreate(new Pair<>(tableIdSet.getId(), mappingType), () -> {
//			
//			cache.clear();
//			Set<TableRowId> candidates = this.tablesRepository.getAllRowIds(tableIdSet.getTableIds());
//			
//			
//			switch(mappingType) {
//			case INSTANCE_CORRESPONDENCE_THEN_LABEL:
//				return getWithInstance(get(tableIdSet, MappingType.LABEL_ONLY,clustererInput), candidates, clustererInput.getCandidates()); 
//			case LABEL_ONLY:
//				return getLabelOnly(candidates); 
//			
//			}
//			throw new IllegalStateException();
//			
//		});
	}


	private Mapping getWithInstance(Mapping mapping, Set<TableRowId> candidates, Set<TableRowId> testingRows) {
		mapping.getIndexToRows();

		AtomicInteger currClusterNumber = new AtomicInteger( mapping.getIndexToRows().size() );
		
		ConcurrentHashMap<Integer, Set<TableRowId>> clusters = new ConcurrentHashMap<>();
		ConcurrentHashMap<TableRowId, Integer> index = new ConcurrentHashMap<>();
		

		ConcurrentHashMap<InstanceId, Integer> instanceIndex = new ConcurrentHashMap<>();
		
		PersistentParallel6.runAndFinish(candidates, (rowId) -> {
			
			Integer indexNumber = null;
			
			InstanceId instanceId = null;
			if(!testingRows.contains(rowId))
				 instanceId = tablesRepository.getMetaData(rowId.getTableId()).getInstance(rowId.getRowIdx().getRowIdx());
			if(instanceId != null) {
				indexNumber = CHMHelper.getFromCHM(instanceIndex, instanceId, () -> currClusterNumber.getAndIncrement());
			}
			else {
				indexNumber = mapping.getRowToIndex().get(rowId);
			}
			
			if(indexNumber == null) {
				int sdf = 123;
				return;
			}
			CHMHelper.getFromCHM(clusters, indexNumber, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(rowId);
			index.put(rowId, indexNumber);
			
			
		}, this.getClass());
		 
		return new Mapping(index, clusters);
	}


//	private Mapping getLabelOnly(RowClusteringConfig config, ClustererInput input) {
//		  ClusterList list = labelClusterer.clusterRows(config, input);
//		  
//		  Map<TableRowId, Integer> index = new HashMap<>();
//		  Map<Integer, Set<TableRowId>> clusterMap = new HashMap<>();
//			
//			for(ListEntry<RowCluster > entry : ListIterable.iterable(list.getClusterList())) {
//				clusterMap.put(entry.getIndex(), new HashSet<>(entry.getValue().getRows()));
//				for(TableRowId rowId : entry.getValue().getRows()) {
//					index.put(rowId, entry.getIndex());
//				}
//			}
//			
//			return new Mapping(index, clusterMap);
//		
//	}
	
	
	private Mapping getLabelOnly(ClassId classId, Set<TableRowId> candidates) {
		 
			
			ConcurrentHashMap<String, TableRowId> stringIndex = new ConcurrentHashMap<>();
			ConcurrentHashMap<TableRowId, Set<TableRowId>> mappings = new ConcurrentHashMap<>();
			
			PersistentParallel6.runAndFinish(candidates, (tableRowId) -> {
				String label = this.tablesRepository.getLabelRaw(tableRowId);
				if(label != null) {
					
					Set<String> set = new HashSet<>();
					
					String normalized = StringNormalizer.normaliseValue(label, false);
				
					 				 
					Optional<String> combinedLabel = IndexBuilder.lettersCombiner(normalized);;
					if(combinedLabel.isPresent()) {
						

						Optional<String> deAccentedLabel = IndexBuilder.deAccent(combinedLabel.get());;
						if(deAccentedLabel.isPresent()) {
							set.add(deAccentedLabel.get());
						}
						else {
							set.add(combinedLabel.get());
						}
						
					}
					
					Optional<String> deAccentedLabel = IndexBuilder.deAccent(normalized);;
					if(deAccentedLabel.isPresent()) {
						set.add(deAccentedLabel.get());

					}
					else {
						set.add(normalized);
					}
							
					
					if(set.size() > 1) {
						int sdf= 123;
					}
					 
					for(String s : set) {
						
						String withBracketContent = s;
						String withoutBracketContent = StringNormalizer.removeBrackets(s);
						
						withBracketContent = StringNormalizer.sortAndJoin(StringNormalizer.tokenise(withBracketContent, false, false));
						withoutBracketContent = StringNormalizer.sortAndJoin(StringNormalizer.tokenise(withoutBracketContent, false, false));
						
						TableRowId rowIdWithBracketContent = CHMHelper.getFromCHM(stringIndex, withBracketContent, () -> tableRowId);
						TableRowId rowIdWithoutBracketContent = CHMHelper.getFromCHM(stringIndex, withBracketContent, () -> tableRowId);

						CHMHelper.getFromCHM(mappings, rowIdWithBracketContent, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowId);
						CHMHelper.getFromCHM(mappings, rowIdWithoutBracketContent, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(tableRowId);
						
						
						
						
						
					}
					
				}
			}, this.getClass());
			
			Map<Integer, Set<TableRowId>> clustersMap = new HashMap<>();
			Map<Integer, ClassId> clusterClassId = new HashMap<>();
			Map<TableRowId, Integer> clustersIndex = new HashMap<>();

			CreateClustersFromMappings.addCluster(mappings, clustersMap, clustersIndex, clusterClassId, (tid) -> {
				return classId; 
			}, 
					(rowId) -> {
						StringObject so = this.tablesRepository.getLabel(rowId);
						if(so != null) {
							return so.getString();
						}
						else {
							return null;
						}
					});
			
			return new Mapping(clustersIndex, clustersMap);
			
		 
	}
	
	
	
	private Mapping getLabelOnlyOld(Set<TableRowId> candidates) {
		try {
			Blocks clusters = labelMapper.getClustersMap(candidates, false);
			Map<TableRowId, Integer> index = new HashMap<>();
			
			for(Entry<Integer, Set<TableRowId>> entry : clusters.getClustersMap().entrySet()) {
				for(TableRowId rowId : entry.getValue()) {
					index.put(rowId, entry.getKey());
				}
			}
			
			return new Mapping(index, clusters.getClustersMap());
		
		
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}
	
	public static class Mapping {
		private final Map<TableRowId, Integer> rowToIndex;
		private final Map<Integer, Set<TableRowId>> indexToRows;

		private final Map<AnnotatedTableId, Set<Integer>> tableToRowIndeces;
		public Mapping(Map<TableRowId, Integer> rowToIndex, Map<Integer, Set<TableRowId>> indexToRows) {
			super();
			this.rowToIndex = Collections.unmodifiableMap(rowToIndex);
			 Map<Integer, Set<TableRowId>> indexToRowsTemp = new HashMap<>();
			 
			 for(Entry<Integer, Set<TableRowId>> entry : indexToRows.entrySet()) {
				 indexToRowsTemp.put(entry.getKey(), Collections.unmodifiableSet(entry.getValue()));
			 }
			 
			this.indexToRows = Collections.unmodifiableMap(indexToRowsTemp);
			
			ConcurrentHashMap<AnnotatedTableId, Set<Integer>> tableToRowIndecesTemp1 = new ConcurrentHashMap<>();
			PersistentParallel6.runAndFinish(rowToIndex.entrySet(), (e) ->{
				CHMHelper.getFromCHM(tableToRowIndecesTemp1, e.getKey().getTableId(), () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(e.getValue());
			}, this.getClass());
			
			
				Map<AnnotatedTableId, Set<Integer>> tableToRowIndecesTemp2 = new HashMap<>();
			 
			 for(Entry<AnnotatedTableId, Set<Integer>> entry : tableToRowIndecesTemp1.entrySet()) {
				 tableToRowIndecesTemp2.put(entry.getKey(), Collections.unmodifiableSet(new HashSet<>(entry.getValue())));
			 }
			 
			 this.tableToRowIndeces = Collections.unmodifiableMap(tableToRowIndecesTemp2);
			
		
		}
		
		public Map<TableRowId, Integer> getRowToIndex() {
			return rowToIndex;
		}
		
		public Map<Integer, Set<TableRowId>> getIndexToRows() {
			return indexToRows;
		}

		public Map<AnnotatedTableId, Set<Integer>> getTableToRowIndeces() {
			return tableToRowIndeces;
		}
		
		
		
	}
	
	
}
