package expansion.set.strategies.rowclusterers.unary;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class ClusterCombiner<E> {
	private final Map<Integer, Set<E>> clusters = new HashMap<>();
	private final Map<E, Integer> indexMap = new HashMap<>();
	int currentIndexKey = 0;

	public Collection<Set<E>> get() {
		return clusters.values();
	}

	public void add(Collection<Set<E>> newClusters) {

		for (Set<E> set : newClusters) {

			Set<Integer> keys = getCurrentKeys(set);

			if (!keys.isEmpty()) {
				Integer firstKey = null;
				for (Integer key : keys) {
					if (firstKey == null) {
						firstKey = key;
						Set<E> cluster = clusters.get(key);
						cluster.addAll(set);
						for (E rowId : set) {
							indexMap.put(rowId, key);
						}
					} else {
						combineCluster(firstKey, key);
					}
				}

			} else {
				createNewCluster(set);
				
			}

		}

	}

	private Set<Integer> getCurrentKeys(Set<E> set) {
		Set<Integer> keys = new HashSet<>();
		for (E rowId : set) {
			Integer key = indexMap.get(rowId);
			if (key != null) {
				keys.add(key);
			}
		}
		return keys;
	}

	private void createNewCluster(Set<E> set) {
		Integer key = currentIndexKey;
		currentIndexKey++;
		Set<E> cluster = new HashSet<>();
		clusters.put(key, cluster);
		for (E rowId : set) {
			indexMap.put(rowId, key);
			cluster.add(rowId);

		}

	}

	private void combineCluster(Integer a, Integer b) {
		Set<E> setB = clusters.get(b);
		clusters.get(a).addAll(setB);
		for (E rowId : setB) {
			indexMap.put(rowId, a);
		}
		clusters.remove(b);
	}

}