package expansion.set.strategies.trees;

import java.util.ArrayList;
import java.util.List;

import javax.inject.Inject;

import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import expansion.set.clustering.ClustererInput;
import expansion.set.clustering.RowClusterer;
import expansion.set.clustering.RowClusteringConfig;
import expansion.set.clustering.model.similarities.LearnableRowClusterSimilarities;
import expansion.set.clustering.model.similarities.RowClusterSimilarities;
import expansion.set.strategies.SetExpansionStrategiesRepository;
import expansion.set.strategies.clustersplitters.attributes.AttributeBasedSplitter;
import expansion.set.strategies.clustersplitters.attributes.AttributeBasedSplitter.AttributeMethod;
import expansion.set.strategies.clustersplitters.minScoreSplitter.WekaClassifierSimilaritiesCreator_MultiTree;
import expansion.set.strategies.ea.ParameterLearnableClusterer;
import expansion.set.strategies.ea.ParamterClusterer;
import expansion.set.strategies.implicit.ImplicitClustererSimilarityCreator;
import expansion.set.strategies.newstrategies.BagOfWords_Component;
import expansion.set.strategies.rowclusterers.SameTableZeroStrategy;
import expansion.set.strategies.rowclusterers.label.LabelRowSimilaritiesCreator;
import expansion.set.strategies.rowclusterers.strategies.association.AssociationSimilarities;
import expansion.set.strategies.rowclusterers.strategies.association.AssociationSimilarities.AssociationType;
import expansion.set.strategies.rowclusterers.strategies.overlap.RowToIndexMapper.MappingType;

public class Combined_RandomForest_New implements RowClusterer, ParameterLearnableClusterer {

	public final static String[] parameters = new String[] { "-U", "-attribute-importance", "-store-out-of-bag-predictions", "-I", "900", "-num-slots", "60",
			"-K", "2", "-P", "85" };//
	// -I 4000

	private final OptimizableClass optimizingClazz = SetExpansionStrategiesRepository.COMBINED_RandomForest_NEW.getOptimizableClassCluster();

	@Inject
	protected WekaClassifierSimilaritiesCreator_MultiTree splitterFactory;
	@Inject
	protected LabelRowSimilaritiesCreator labelRowSimilaritiesCreator;

	@Inject
	protected AttributeBasedSplitter attributeSplitter;

	@Inject
	BagOfWords_Component component;

	@Inject
	protected AssociationSimilarities association;

	@Inject
	protected ImplicitClustererSimilarityCreator implicitCreator;

	public Combined_RandomForest_New() {

	}


	public RowClusterSimilarities clusterRows(RowClusteringConfig config, ClustererInput input) {
		return clusterRows(config, input, parameters);
	}

	public RowClusterSimilarities clusterRows(RowClusteringConfig config, ClustererInput input, String[] parameters) {
		// ClusterList clusters = clusterer.clusterRows(config, input);

		// RowClusterSimilarities attributeScores =
		// attributeSplitter.get(config, input, AttributeMethod.DIRECT_FULL);
		RowClusterSimilarities attributeScores = attributeSplitter.get(config, input, AttributeMethod.DIRECT_FULL);

		RowClusterSimilarities labelScores = labelRowSimilaritiesCreator.createLucene(input.getFullCandidates(), config.isWithCache(), true);
		RowClusterSimilarities bowScores = component.getBagOfWords_for_RowWithoutTFIDF(input, false, config.isWithCache());
		// RowClusterSimilarities tableBowScores =
		// component.getBagOfWords_for_TableWithoutTFIDF(input, false);

		RowClusterSimilarities phiScores = association.getScores(config, input, MappingType.LABEL_ONLY, AssociationType.PHI);

		RowClusterSimilarities implicit = implicitCreator.create(input, config);

		RowClusterSimilarities sameTableSim = new SameTableZeroStrategy();
		// RowClusterSimilarities pmiScores = association.getScores(config,
		// input, MappingType.LABEL_ONLY, AssociationType.PMI);
		// RowClusterSimilarities tfidfScores = association.getScores(config,
		// input, MappingType.LABEL_ONLY, AssociationType.TF_IDF);

		// double phiWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("PHI",
		// 1.0));
		// double pmiWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("PMI",
		// 1.0));
		// double labelWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("LABEL",
		// 1.0));
		// double attributeWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("ATTRIBUTE",
		// 1.0));
		// double bowWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("BOW",
		// 1.0));
		// double tableBowWeight =
		// config.getWeightsRepository().getOrDefault(optimizingClazz.getWeight("TABLE_BOW",
		// 1.0));

		List<Tri<RowClusterSimilarities, Boolean, Boolean>> sims = new ArrayList<>();
		sims.add(new Tri<>(attributeScores, false, true));
		sims.add(new Tri<>(implicit, false, true));
		sims.add(new Tri<>(labelScores, false, false));
		sims.add(new Tri<>(bowScores, false, false));
		// sims.add(tableBowScores);
		sims.add(new Tri<>(phiScores, false, false));
		sims.add(new Tri<>(sameTableSim, false, false));
		// sims.add(sameTableSim);
		// sims.add(empty);
		// sims.add(pmiScores);
		// sims.add( new Pair<>(tfidfScores, 0.0) );

		LearnedModel learnedModel = null;
		if (config.getWeightsRepository() != null) {
			learnedModel = config.getWeightsRepository().get(optimizingClazz);
		}

		return splitterFactory.getSplitter("RandomForest", parameters, sims, learnedModel, optimizingClazz, true, null);
	}

	@Override
	public RF_ST_Creator getCreator(RowClusteringConfig config, ClustererInput input) {
		return new RF_ST_Creator(config, input);
	}

	public class RF_ST_Creator implements ParamterClusterer {
		private final RowClusteringConfig config;
		private final ClustererInput input;

		public RF_ST_Creator(RowClusteringConfig config, ClustererInput input) {
			this.config = config;
			this.input = input;
		}

		@Override
		public LearnableRowClusterSimilarities clusterRows(String[] parameters) {
			return ((LearnableRowClusterSimilarities) Combined_RandomForest_New.this.clusterRows(config, input, parameters));
		}

	}

}

