package expansion.set.tablerow;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.adapters.columnmatches.ColumnMatchesRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.learning.SplitRowValues;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.informatik.dws.matching.LabelMatcher;
import de.uni_mannheim.informatik.dws.matching.LabelMatcher.MatchStrictness;

public class TableRowCreator {
	private final AnnotatedTablesRepository tableRepository;
	private final OntologyRepository ontologyRepository;
	private final LabelMatcher labelMatcher;

	private final ColumnMatchesRepository columnMatchesRepository;
	public static boolean doExtraMatching = false;

	public static ConcurrentHashMap<Pair<String, Set<ClassId>>, Optional<Pair<InstanceId, Double>>> columnReferenceCache = new ConcurrentHashMap<>();

	@Inject
	public TableRowCreator(AnnotatedTablesRepository tableRepository, OntologyRepository ontologyRepository, LabelMatcher labelMatcher,
			ColumnMatchesRepository columnMatchesRepository) {
		super();
		this.tableRepository = tableRepository;
		this.ontologyRepository = ontologyRepository;
		this.labelMatcher = labelMatcher;
		this.columnMatchesRepository = columnMatchesRepository;
	}

	public AnnotatedQualifiedTableRow getTableRow(TableRowId tableRowId, SplitRowValues splitRowValues) {
		String label = null;
		AnnotatedTable table = tableRepository.getTable(tableRowId.getTableId());

		AbstractObject keyValue = table.getKeyValue(tableRowId.getRowIdx());
		if (keyValue != null) {
			if (keyValue instanceof StringObject) {
				label = ((StringObject) keyValue).getString();
			} else {
				label = keyValue.toString();
			}
		}
		// final Charset toCharset = Charset.forName("ISO-8859-1");
		// final Charset fromCharset = Charset.forName("UTF-8");
		// List<Integer> list = new ArrayList<>();
		// for (int i = 0; i < label.length(); i++) {
		// list.add((int) label.charAt(i));
		// }
		// String fixed = new String(label.getBytes(toCharset), fromCharset);

		Map<QualifiedPropertyId, MatchedValues> splitValues = splitRowValues.getValues().get(tableRowId);
		if (splitValues == null) {
			splitValues = new HashMap<>();
		} else {
			int sdf = 123;
		}
		return new AnnotatedQualifiedTableRow(label, splitValues);

	}
	public AnnotatedTableRow getTableRow(TableRowId tableRowId) {
		String label = null;
		Map<PropertyId, MatchedValues> values = new HashMap<>();
		AnnotatedTable table = tableRepository.getTable(tableRowId.getTableId());
		ClassId classId = tableRepository.getMetaData(tableRowId.getTableId()).getDbPediaClass();
		AbstractObject keyValue = table.getKeyValue(tableRowId.getRowIdx());
		if (keyValue != null) {
			if (keyValue instanceof StringObject) {
				label = ((StringObject) keyValue).getString();
			} else {
				label = keyValue.toString();
			}
		}

		Map<Integer, PropertyId> properties = tableRepository.getMetaData(tableRowId.getTableId()).getProperties();
		for (Entry<Integer, PropertyId> property : tableRepository.getMetaData(tableRowId.getTableId()).getProperties().entrySet()) {
			AbstractObject ao = table.getValue(tableRowId.getRowIdx(), new ColumnIdx(property.getKey()));

			if (false && ontologyRepository.isReferenceField(property.getValue())) {
				InstanceId match = this.columnMatchesRepository
						.getMatch(new CellId(tableRowId.getTableId(), new ColumnIdx(property.getKey()), tableRowId.getRowIdx()));

				if (match != null) {
					ao = match;
				}
				else if (doExtraMatching && ao instanceof StringObject) {
					Pair<String, Set<ClassId>> pair = new Pair<>(((StringObject) ao).getString(),
							ontologyRepository.getRange(new ClassPropertyId(classId, property.getValue())));
		
					Optional<Pair<InstanceId, Double>> matchPair = CHMHelper.getFromCHM(columnReferenceCache, pair, () -> {
						return labelMatcher.getMaxMatch(pair.getFirst(), pair.getSecond(), MatchStrictness.STRICT);

					});

					// matchPair = labelMatcher.getMaxMatch(((StringObject)
					// ao).getString(),
					// ontologyRepository.getRange(new ClassPropertyId(classId,
					// property.getValue())), MatchStrictness.NON_STRICT);
					if (matchPair.isPresent()) {
						ao = matchPair.get().getFirst();
					}
				}
			}

			if (ao != null) {
				SingleValue sv = new SingleValue(new CellId(tableRowId.getTableId(), new ColumnIdx(property.getKey()), tableRowId.getRowIdx()), ao);
				if(values.containsKey(property.getValue())) {
					MatchedValues mv = values.get(property.getValue());
					values.put(property.getValue(), new MatchedValues(mv, sv));
				}
				else {
				values.put(property.getValue(),
						new MatchedValues(sv,
								table.getDataType(new ColumnIdx(property.getKey()))));
				}
			}
		}

		return new AnnotatedTableRow(label, values);
	}
}
