package expansion.set.strategies.rowclusterers;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;

public class ClusterCreatorTester {

	public static void main(String[] args) {

		TableRowId A = new TableRowId(new AnnotatedTableId("A"), 1);
		TableRowId B = new TableRowId(new AnnotatedTableId("B"), 1);
		TableRowId C = new TableRowId(new AnnotatedTableId("C"), 1);
		TableRowId D = new TableRowId(new AnnotatedTableId("D"), 1);
		TableRowId E = new TableRowId(new AnnotatedTableId("E"), 1);
		TableRowId F = new TableRowId(new AnnotatedTableId("F"), 1);
		TableRowId G = new TableRowId(new AnnotatedTableId("G"), 1);
		TableRowId H = new TableRowId(new AnnotatedTableId("H"), 1);
		TableRowId I = new TableRowId(new AnnotatedTableId("I"), 1);
		TableRowId J = new TableRowId(new AnnotatedTableId("J"), 1);
		TableRowId K = new TableRowId(new AnnotatedTableId("K"), 1);
		TableRowId L = new TableRowId(new AnnotatedTableId("L"), 1);
		TableRowId M = new TableRowId(new AnnotatedTableId("M"), 1);
		TableRowId N = new TableRowId(new AnnotatedTableId("N"), 1);
		TableRowId O = new TableRowId(new AnnotatedTableId("O"), 1);

		Map<TableRowId, Set<TableRowId>> mappings = new HashMap<>();
		mappings.put(A, new HashSet<>(Arrays.asList(B)));
		mappings.put(B, new HashSet<>(Arrays.asList(C)));

		mappings.put(C, new HashSet<>(Arrays.asList(A)));
		mappings.put(D, new HashSet<>(Arrays.asList(B, E, F, G)));

		mappings.put(E, new HashSet<>());
		mappings.put(F, new HashSet<>());
		mappings.put(G, new HashSet<>());

		mappings.put(H, new HashSet<>(Arrays.asList(I, J)));

		mappings.put(I, new HashSet<>());
		mappings.put(J, new HashSet<>());

		mappings.put(K, new HashSet<>(Arrays.asList(L)));
		mappings.put(O, new HashSet<>(Arrays.asList(M, N, L)));
		mappings.put(L, new HashSet<>());
		mappings.put(M, new HashSet<>());
		mappings.put(N, new HashSet<>());

		HashMap<Integer, Set<TableRowId>> clustersMap = new HashMap<>();
		HashMap<TableRowId, Integer> clusterIndex = new HashMap<>();

		addCluster(mappings, clustersMap, clusterIndex);

		System.out.println(clustersMap.toString());

	}

	private static void addCluster(Map<TableRowId, Set<TableRowId>> candidates, Map<Integer, Set<TableRowId>> clustersMap,
			Map<TableRowId, Integer> clustersIndex) {

		clustersMap.clear();
		clustersIndex.clear();

		for (Entry<TableRowId, Set<TableRowId>> entry : candidates.entrySet()) {

			if (!clustersIndex.containsKey(entry.getKey())) {
				Set<TableRowId> rowIds = getAllRowIds(candidates, entry.getKey());

				Set<Integer> existingClusters = searchCluster(clustersIndex, rowIds);

				int index = -1;

				if (existingClusters.size() == 1) {
					index = existingClusters.iterator().next();
				} else if (existingClusters.size() > 1) {
					Iterator<Integer> it = existingClusters.iterator();
					index = it.next();
					it.remove();

					for (Integer otherIndeces : existingClusters) {
						Set<TableRowId> insideCluster = clustersMap.get(otherIndeces);

						CHMHelper.getFromMap(clustersMap, index, () -> new HashSet<>()).addAll(insideCluster);

						for (TableRowId inRowId : insideCluster) {
							clustersIndex.put(inRowId, index);
						}

						clustersMap.remove(otherIndeces);
					}
				} else {
					index = clustersMap.size();

				}

				for (TableRowId rowId : rowIds) {

					if (clustersIndex.containsKey(rowId)) {
						if (!clustersIndex.get(rowId).equals(index)) {
							throw new IllegalStateException();
						}
					} else {
						CHMHelper.getFromMap(clustersMap, index, () -> new HashSet<>()).add(rowId);
						clustersIndex.put(rowId, index);
					}

				}

			}

			// Set<TableRowId> tableRowIdSet = new HashSet<>();
			// tableRowIdSet.add(entry.getKey());
			// tableRowIdSet.addAll(entry.getValue());

		}

		// boolean added = false;
		//
		// for (TableRowId tableRowId : tableRowIdSet) {
		// Integer index = clustersIndex.get(entry.getKey());
		// clustersMap.get(index).addAll(entry.getValue());
		// for (TableRowId tableRowId : entry.getValue()) {
		// clustersIndex.put(tableRowId, index);
		// }
		// }
		// TODO Auto-generated method stub

	}

	private static Set<Integer> searchCluster(Map<TableRowId, Integer> clustersIndex, Set<TableRowId> rowIds) {

		Set<Integer> index = Collections.newSetFromMap(new ConcurrentHashMap<>());

		PersistentParallel6.runAndFinish(rowIds, (rowId) -> {
			if (clustersIndex.containsKey(rowId)) {

				index.add(clustersIndex.get(rowId));
			}
		} , ClusterCreatorTester.class, rowIds.size() < 1000);

		return index;
	}

	private static Set<TableRowId> getAllRowIds(Map<TableRowId, Set<TableRowId>> candidates, TableRowId key) {

		Set<TableRowId> tableRowIdSet = new HashSet<>();

		getAllRowIds(candidates, tableRowIdSet, key);

		return tableRowIdSet;

	}

	private static void getAllRowIds(Map<TableRowId, Set<TableRowId>> candidates, Set<TableRowId> tableRowIdSet, TableRowId key) {

		tableRowIdSet.add(key);

		Set<TableRowId> cand = candidates.get(key);

		LinkedList<TableRowId> stack = new LinkedList<>();

		stack.addAll(cand);

		TableRowId rowId = null;
		while ((rowId = stack.poll()) != null) {
			if (!tableRowIdSet.contains(rowId)) {
				tableRowIdSet.add(rowId);
				for (TableRowId innerRowId : candidates.get(rowId)) {
					if (!tableRowIdSet.contains(innerRowId)) {
						stack.add(innerRowId);

					}
				}
			}
		}
		// for (TableRowId rowId : candidates.get(key)) {
		//
		// if (!tableRowIdSet.contains(rowId)) {
		// getAllRowIds(candidates, tableRowIdSet, rowId);
		//
		// }
		// }

	}

}
