package expansion.set.strategies.rowclusterers;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.ListIterable;
import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;
import de.mannheim.uni.expansion.utils.OtsuThresholderRounded;
import de.mannheim.uni.expansion.utils.OtsuThresholderRounded.DefaultIntegerScorable.DoubleScorable;
import de.mannheim.uni.expansion.utils.containers.Pair;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.strategies.rowclusterers.TestClustering.ClusterSimilaritiesWrapper;

public class CorellationClustering2 {
	private final ClusterSimilaritiesWrapper clusterSimilaritiesWrapper;
	private final Set<Integer> keySet;
	private final OtsuThresholderRounded otsu = new OtsuThresholderRounded(); 
//	private final double i;
//	private final Map<Integer, LinkedHashMap<Integer, Double>> sims;
//	private final Map<Integer, RangedMap<Integer, Double>> rangedSims;
	
	
	public CorellationClustering2(ClusterSimilaritiesWrapper clusterSimilaritiesWrapper, Set<Integer> keySet) {
		this.clusterSimilaritiesWrapper = clusterSimilaritiesWrapper;
		this.keySet = keySet;
//		Map<Integer, LinkedHashMap<Integer, Double>> simsTemp = new ConcurrentHashMap<>();
//		
//		Map<Integer, RangedMap<Integer, Double>>  rangedSimsTemp = new ConcurrentHashMap<>();
//		PersistentParallel6.runAndFinish(keySet, (node) -> {
//			Map<Integer, Double> simMap = new HashMap<>();
//		
//			for (Integer candidate : keySet) {
//				if(candidate != node) {
//					Double sim = clusterSimilaritiesWrapper.get(new ClusterEntryKey(node, candidate));
//					if(sim != null   )
//						simMap.put(candidate, sim);
//				}
//			}
//			simsTemp.put(node, MapSorter.sortHashMapByValuesD(simMap));
//			rangedSimsTemp.put(node, new RangedMap<>(simMap));
//		}, this.getClass());
//  
//		this.sims = simsTemp;
//		this.rangedSims = rangedSimsTemp;
//		this.i = 1.0;
	}

	public static List<Set<Integer>> cc(ClusterSimilaritiesWrapper clusterSimilaritiesWrapper, Set<Integer> keySet, double threshold, int iterations, boolean silhoutteRun, boolean reassignTest, boolean filterSizeOne) {
		CorellationClustering2 cc = new CorellationClustering2(clusterSimilaritiesWrapper, keySet);
		return cc.run(threshold, iterations,silhoutteRun, reassignTest, filterSizeOne);
	}

	private List<Set<Integer>> run(double threshold, int iterations, boolean silhoutteRun, boolean reassignTest, boolean filterSizeOne) {
		List<Set<Integer>> clusters = new ArrayList<>();
		
		
		for(Integer i : keySet) {
			Optional<Pair<Double, Integer>> max = getMax(i, clusters, null);
			
			if(max.isPresent() && max.get().getFirst() > 0.0) {
				clusters.get(max.get().getSecond()).add(i);
			}
			else {
				Set<Integer> newCluster = new HashSet<>();
				newCluster.add(i);
				clusters.add(newCluster);
			} 
		}
		
		if(reassignTest) {
		System.out.println("Size before: " + clusters.size());
		clusters = reassignReassignableClusters(keySet, clusters);
		System.out.println("Size after: " + clusters.size());
		clusters = reassignReassignableClusters(keySet, clusters);
		System.out.println("Size after after: " + clusters.size());
		}
		
		if(filterSizeOne) {
		clusters = clusters.stream().filter((c) -> c.size() > 1).collect(Collectors.toList());
		}
		
		if(silhoutteRun){ 	
			Set<Integer> reassign = new HashSet<>();
			Map<Integer, Set<Integer>> newClusters = new HashMap<>();
			Map<Integer, Double> sils = new HashMap<>();
			List<DoubleScorable> silList = new ArrayList<>();
				for(int c= 0; c<clusters.size();c++) {
					double silhoutte = 0.0;
					int count = 0;
					Set<Integer> clusterSet = clusters.get(c);
					for(Integer i: clusterSet) { 
						Optional<Double> aOptional = averageDissimilarity(i, clusterSet);
						
						double a = 0.5;
						if(aOptional.isPresent()) {
							a = aOptional.get();
						}
						else {
							int sdf = 123;
						}
						
						double b = 0.5;
						for(int c2= 0; c2<clusters.size();c2++) {
							if(c != c2) {
								
								Optional<Double> bTemp =  averageDissimilarity(i, clusters.get(c2));
								if(bTemp.isPresent() && bTemp.get() < b) {
									b = bTemp.get();
								}
								
							}
						}
						silhoutte += (b - a) / Math.max(a, b);
						count++;
						
					}
					if(count == 0) {
						int df = 123;
					}
					double sil =  silhoutte / count;
					System.out.println("Silhoutte: " + sil);
					
					sils.put(c, sil);
					silList.add( new DoubleScorable(sil ));
//					if(sil < 0) {
//						reassign.addAll(clusterSet);
//					}
//					else {
//						
//					}
					
					
					
				}
				
				double otsuThreshold = otsu.getThresholdNoExtra(silList);
				
				for(Entry<Integer, Double> silEntry : sils.entrySet()) {
					if(silEntry.getValue() < otsuThreshold && otsuThreshold - silEntry.getValue() > 0.2 ) {
						reassign.addAll(clusters.get(silEntry.getKey()));
					}
					else {
						newClusters.put(silEntry.getKey(), clusters.get(silEntry.getKey()));
					}
					
				}
			
				
				clusters = new ArrayList<>();
			for(Entry<Integer, Set<Integer>>entry : newClusters.entrySet()) {
				clusters.add(entry.getValue());
			}
			
			
			
			
			for(Integer i : reassign) {
				Double maxQuality = null;
				int maxCluster =-1;
				for(int c = 0; c < clusters.size(); c++) {
					
					double quality = 0.0;
					for(Integer j : clusters.get(c)) {
						if(j != i) {
							Double sim = getScore(i, j);
							if(sim != null)
								quality += sim;
						}
					}
					
					if(maxQuality == null || quality > maxQuality) {
						maxQuality = quality;
						maxCluster = c;
					}
					
				}
				
				if(maxQuality == null || maxQuality <= 0) {
//					Set<Integer> newCluster = new HashSet<>();
//					newCluster.add(i);
//					clusters.add(newCluster);
				}
				else { 
					Set<Integer>cluster = clusters.get(maxCluster);
					 
						
					cluster.add(i);
				}
			}
			
			
		}
		
		
		for(int it = 0; it < iterations; it++) {
//			List<Set<Integer>> newClusters = new ArrayList<>();
			Map<Integer, Set<Integer>> newClusters = new HashMap<>();
			
			for(Integer i : keySet) {
				Double maxQuality = null;
				int maxCluster =-1;
				for(int c = 0; c < clusters.size(); c++) {
					
					double quality = 0.0;
					for(Integer j : clusters.get(c)) {
						if(j != i) {
							Double sim = getScore(i, j);
							if(sim != null)
								quality += sim;
						}
					}
					
					if(maxQuality == null || quality > maxQuality) {
						maxQuality = quality;
						maxCluster = c;
					}
					
				}
				
				if(maxQuality == null || maxQuality <= 0) {
//					Set<Integer> newCluster = new HashSet<>();
//					newCluster.add(i);
//					clusters.add(newCluster);
				}
				else { 
					Set<Integer>cluster = CHMHelper.getFromMap(newClusters, maxCluster, () -> new HashSet<>());
					 
						
					cluster.add(i);
				}
			}
			
			clusters = new ArrayList<>();
			for(Entry<Integer, Set<Integer>>entry : newClusters.entrySet()) {
				clusters.add(entry.getValue());
			}
		}
		
		return clusters;
	}
	
	private Optional<Pair<Double, Integer>> getMax(Integer i, List<Set<Integer>> clusters, Set<Integer> except) {
		Double maxQuality = null;
		int maxCluster =-1;
		for(int c = 0; c < clusters.size(); c++) {
			
			if(except == null || !except.contains(c)) {
				double quality = 0.0;
				for(Integer j : clusters.get(c)) {
					Double sim = 1.0;
					if(i != j) {
						  sim = getScore(i, j);
					}
						if(sim != null)
							quality += sim;
					
				}
				
				if(maxQuality == null || quality > maxQuality) {
					maxQuality = quality;
					maxCluster = c;
				}
			}
			
		}
		
		if(maxQuality == null)
			return Optional.empty();
		return Optional.of(new Pair<>(maxQuality, maxCluster));
	}

	private List<Set<Integer>> reassignReassignableClusters(Set<Integer> keySet,List<Set<Integer>> clusters) {
		
		
		int size = clusters.size();
		double totalSize = keySet.size();
		Set<Integer> reassignedCluster = new HashSet<>();
		Map<Integer, Pair<Double, Integer>> reassignments = new HashMap<>();
		
		clusters  = new ArrayList<>(clusters);
		Collections.sort(clusters, (a,b) -> { return Integer.compare(a.size(), b.size());});
		
		for(ListEntry<Set<Integer>> c : ListIterable.iterable(clusters)) {
			
			if(c.getValue().size() / totalSize >= 0.10) {
				continue;
			}
			
			Map<Integer, Pair<Double, Integer>> reassignable = new HashMap<>();
			
			for(Integer i : c.getValue()) {
				
				Set<Integer> except = new HashSet<>(reassignedCluster);
				except.add(c.getIndex());
				Optional<Pair<Double, Integer>> max = getMax(i, clusters, except);
				
				if(max.isPresent() && max.get().getFirst() > 0.0) {
					reassignable.put(i, max.get());
				}
				
				
			}
			
			if(reassignable.size() == c.getValue().size()) {
				reassignedCluster.add(c.getIndex());
				reassignments.putAll(reassignable);
			}
		}
		
		Map<Integer, Set<Integer>> newClusters = new HashMap<>();
		for(ListEntry<Set<Integer>> c : ListIterable.iterable(clusters)) { 
			if(!reassignedCluster.contains(c.getIndex())) {
				newClusters.put(c.getIndex(), c.getValue());
			}
			else {
				int df = 123;
			}
		}
		
		clusters = new ArrayList<>(newClusters.values());
		
		
		 newClusters = new HashMap<>();
		for(Integer i : keySet) {
			
 
			Optional<Pair<Double, Integer>> max = getMax(i, clusters, null);
			
			if(max.isPresent() ) {
				CHMHelper.getFromMap(newClusters, max.get().getSecond(), () -> new HashSet<>()).add(i);
			}
			else {
				int sdf = 123;
			}
			
			
		}
		
		clusters = new ArrayList<>(newClusters.values());
		
		return clusters;
	}

	private Optional<Double> averageDissimilarity(int i, Set<Integer> cluster) {
		double quality = 0.0;
		int count = 0;
		int countDis = 0;
		for(Integer j : cluster) {
			if(j != i) {
				Double sim = getScore(i, j);
				if(sim != null) {
					if(sim < 0) {
						quality +=  -1 * sim;
						countDis++;
					}
					count++;
				}
			}
		}
		
		if(countDis == 0 || count == 0) 
			return Optional.empty();
		else
			return Optional.of(quality / count);
	}
	
	private Double getScore(int a, int b) {
		 Double ret = clusterSimilaritiesWrapper.get(new ClusterEntryKey(a, b));
 
		 return ret;
	}
}
