package expansion.set.strategies.rowclusterers;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import de.mannheim.uni.expansion.utils.DoubleIteraterable;
import de.mannheim.uni.expansion.utils.DoubleIteraterable.DoubleEntry;
import de.mannheim.uni.expansion.utils.MapSorter;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.strategies.rowclusterers.TestClustering.ClusterSimilaritiesWrapper;

public class CorrelationClustering {

	private final ClusterSimilaritiesWrapper clusterSimilaritiesWrapper;
	private final Set<Integer> keySet;
	private final double i;
	private final Map<Integer, LinkedHashMap<Integer, Double>> sims;
	private final Map<Integer, RangedMap<Integer, Double>> rangedSims;
//	private final Map<Integer, LinkedHashMap<Integer, Double>> reverse;
	private final double c = 2.1;
	private final int n;

	private CorrelationClustering(ClusterSimilaritiesWrapper clusterSimilaritiesWrapper, Set<Integer> keySet) {
		this.clusterSimilaritiesWrapper = clusterSimilaritiesWrapper;
		this.keySet = keySet;
		Map<Integer, LinkedHashMap<Integer, Double>> simsTemp = new ConcurrentHashMap<>();
		
		Map<Integer, RangedMap<Integer, Double>>  rangedSimsTemp = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(keySet, (node) -> {
			Map<Integer, Double> simMap = new HashMap<>();
		
			for (Integer candidate : keySet) {
				if(candidate != node) {
					Double sim = clusterSimilaritiesWrapper.get(new ClusterEntryKey(node, candidate));
					if(sim != null && Math.abs(sim) >= 0.5)
						simMap.put(candidate, sim);
				}
			}
			simsTemp.put(node, MapSorter.sortHashMapByValuesD(simMap));
			rangedSimsTemp.put(node, new RangedMap<>(simMap));
		}, this.getClass());
 
		this.n = keySet.size();
		this.sims = simsTemp;
		this.rangedSims = rangedSimsTemp;
		this.i = 1.0;
	}

	private List<Set<Integer>> run() {

		Set<Integer> remainingNodes = new HashSet<>(keySet);

		List<Set<Integer>> clusters = new ArrayList<>();

		while (!remainingNodes.isEmpty()) {
			
			
			Integer currentNode = remainingNodes.iterator().next();
			
			Set<Integer> currentCluster = new HashSet<>();
			currentCluster.add(currentNode);
			double r = 0.0;
			
			r = getNextRadius(currentNode, r, remainingNodes);
			if(  (r >= 0.0) ) {
				currentCluster = getNodesInRadius(currentNode, r);
				System.out.println("Cut: " + getCut(currentNode, r));
				System.out.println("Volumne: " + getVolume(currentNode, r));
				System.out.println("Adjusted Volumne: " + this.c * Math.log(n + 1) * getVolume(currentNode, r));
				
				while(!(getCut(currentNode, r) <=   getVolume(currentNode, r)) ) {
					
					r = getNextRadius(currentNode, r, remainingNodes);
					currentCluster = getNodesInRadius(currentNode, r);
					if(! (r >= 0.0) ) {
						break;
					}
					
					System.out.println("Cut: " + getCut(currentNode, r));
					System.out.println("Volumne: " + getVolume(currentNode, r));
					System.out.println("Adjusted Volumne: " + this.c * Math.log(n + 1) * getVolume(currentNode, r));
				}
			}
			 
				 
					clusters.add(currentCluster);
					remainingNodes.removeAll(currentCluster);
			 

		}

		return clusters;
	}

	private double getNextRadius(Integer currentNode, double r, Set<Integer> remainigNodes) {
		
		for(Entry<Integer, Double> entry : sims.get(currentNode).entrySet()) {
			if(remainigNodes.contains(entry.getKey()) && entry.getValue() > 0.0 ) {
				if(1.0 - entry.getValue() > r) {
					return 1.0 - entry.getValue();
				}
			}
		}
	 
		return -1.0;
	}

//	public double getCut(Set<Integer> nodes) {
//		for()
//		for(DoubleIteraterable.iterable(nodes)) {
//			
//		}
//		
//	}
	public double getCut(int nodeIndex, double r) {
		Set<Integer> nodes = getNodesInRadius(nodeIndex, r);
		
nodes.remove(nodeIndex);
		double cut = 0.0;
		for(Integer node : nodes) {
			 
//			Double sim = clusterSimilaritiesWrapper.get(new ClusterEntryKey(nodeIndex, node));
			for (Entry<Integer, Double> innerSimsEntry : sims.get(node).entrySet()) {
				
				
				if (!nodes.contains(innerSimsEntry.getKey()) && innerSimsEntry.getValue() > 0.0 && isSatisfied(node, innerSimsEntry.getKey())) {
					cut +=   innerSimsEntry.getValue();
//					cutCount += sim;
				}
 
			 }
			 
		}
		
		
//		double cut = 0.0;
//		Iterator<Entry<Integer, Double>> it = sims.get(nodeIndex).entrySet().iterator();
//		while (it.hasNext()) {
//			Entry<Integer, Double> entry = it.next();
//			if (entry.getValue() > 0) {
//				
//				if (1.0 - entry.getValue() <= r) {
//					for (Entry<Integer, Double> innerSimsEntry : sims.get(entry.getKey()).entrySet()) {
//						if (innerSimsEntry.getValue() > 0.0) {
//							cut += innerSimsEntry.getValue();
//						}
//					}
//				}
//				else {
//					cut += entry.getValue();
//				}
//
//			}
//		}
		return cut;
	}

	public double getVolume(int nodeIndex, double r) {
		Set<Integer> nodes = getNodesInRadius(nodeIndex, r);
		
		nodes.remove(nodeIndex);
		double volume = 0.1;
		for(DoubleEntry<Integer> dd : DoubleIteraterable.iterable(nodes)) {
			Double sim = getScore(dd.getFirst(), dd.getSecond());
			if(sim != null && sim > 0.0 && isSatisfied(dd.getFirst(), dd.getSecond())) {
			volume += sim;
			}
		}
//		
		double cutCount = 0.0;
		double cut = 0.0;
//		for(Integer node : nodes) {
//			if(node != nodeIndex) {
//			Double sim = getScore(nodeIndex, node);
//			for (Entry<Integer, Double> innerSimsEntry : sims.get(node).entrySet()) {
//				
//				
//				if (!nodes.contains(innerSimsEntry.getKey()) && innerSimsEntry.getValue() > 0.0 && isSatisfied(node, innerSimsEntry.getKey())) {
//					cut += innerSimsEntry.getValue() * sim;
//					cutCount += sim;
//				}
//			}
//			}
//		}
//		Iterator<Entry<Integer, Double>> it = sims.get(nodeIndex).entrySet().iterator();
//		while (it.hasNext()) {
//			Entry<Integer, Double> entry = it.next();
//			if (entry.getValue() > 0 && isSatisfied(nodeIndex, entry.getKey())) {
//			
//				if (1.0 - entry.getValue() <= r) {
//					for (Entry<Integer, Double> innerSimsEntry : sims.get(entry.getKey()).entrySet()) {
//						if (!nodes.contains(innerSimsEntry.getKey()) && innerSimsEntry.getValue() > 0.0) {
//							double weight =  entry.getValue() - r;
//							cut += innerSimsEntry.getValue() * entry.getValue();
//							cutCount += entry.getValue();
//						}
//					}
//				}
//				else {
//					cut += entry.getValue();
//				}
//
//			}
//		}
//		return cut;
		
		double ret = volume + ( cut /cutCount );
		return volume;
		
		
//		double volume = i;
//		Iterator<Entry<Integer, Double>> it = sims.get(nodeIndex).entrySet().iterator();
//		while (it.hasNext()) {
//			Entry<Integer, Double> entry = it.next();
//			if (entry.getValue() > 0.0 && 1.0 - entry.getValue() <= r) {
//				volume += entry.getValue();
//				
//				for (Entry<Integer, Double> innerSimsEntry : sims.get(entry.getKey()).entrySet()) {
//					if (innerSimsEntry.getValue() > 0.0) {
//						volume += innerSimsEntry.getValue() * (1.0 - entry.getValue());
//					}
//				}
//			}
//		}
//		return volume;
	}
	
	private boolean isSatisfied(Integer first, Integer second) {
		double sim = getScore(first, second);
		
		Set<Integer> candidates = new HashSet<>();
		candidates.addAll(rangedSims.get(second).lowerThan(sim));
		candidates.addAll(rangedSims.get(first).lowerThan(sim));
		
		for(Integer c : candidates) {
			Double firstC = getScore(first, c);
			Double secondC = getScore(c, second);
			if(firstC != null && secondC != null) {
				double diff = Math.abs(firstC - secondC);
				if( sim < diff) {
					return false;
				}
			}
		}
 
		return true;
	}

	public Set<Integer> getNodesInRadius(int nodeIndex, double r) {
		Set<Integer> ret = new HashSet<>();
		ret.add(nodeIndex);
		Iterator<Entry<Integer, Double>> it = sims.get(nodeIndex).entrySet().iterator();
		while (it.hasNext()) {
			Entry<Integer, Double> entry = it.next();
			if (  1.0 - entry.getValue() <= r  ) {
				ret.add(entry.getKey());
			}
		}
		return ret;
	}
	
	public double getVolume() {
		double volume = 0.0;
		for(DoubleEntry<Integer> dd : DoubleIteraterable.iterable(keySet)) {
			double sim = getScore( dd.getFirst(), dd.getSecond() );
			if(sim > 0.0 &&   isSatisfied(dd.getFirst(), dd.getSecond())) {
			volume += sim;
			}
		}
		return volume;
	}

	public static List<Set<Integer>> cc(ClusterSimilaritiesWrapper clusterSimilaritiesWrapper, Set<Integer> keySet) {
		CorrelationClustering cc = new CorrelationClustering(clusterSimilaritiesWrapper, keySet);
		return cc.run();
	}
	
	private Double getScore(int a, int b) {
		 Double ret = clusterSimilaritiesWrapper.get(new ClusterEntryKey(a, b));
		 if(ret != null && Math.abs(ret) < 0.5) {
			 return null;
		 }
		 return ret;
	}

}
