package expansion.set.strategies.rowclusterers;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.commons.io.FileUtils;

import com.google.common.util.concurrent.AtomicDouble;

import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.DecimalFormatter;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.evaluation.goldstandard.oktie.OktieClusterEvaluator;
import expansion.set.evaluation.goldstandard.oktie.OktiePerformance;
import expansion.set.strategies.rowclusterers.TestClustering.Test_Runs;
import expansion.set.strategies.rowclusterers.unary.ClusterCombiner;

public class InputTestClustering {

	public static void main(String[] args) throws IOException {
		InputTestClustering itc = new InputTestClustering();
		itc.run();
	}

	public InputTestClustering() {
		// TODO Auto-generated constructor stub
	}

	public void run() throws IOException {

		Map<String, Map<Integer, Integer>> belongMap = new HashMap<>();
		Map<String, Map<ClusterEntryKey, Double>> simMap = new HashMap<>();
		File[] listOfFiles = (new File("D:/newStuff/input")).listFiles();

		for (File file : listOfFiles) {
			String name = file.getName();
			String number = name.substring(name.length() - 1);
			boolean isGs = name.substring(name.length() - 5, name.length() - 1).equals("_GS_");
			String actualName;

			if (isGs) {
				actualName = name.substring(0, name.length() - 5) + "_" + number;
			} else {
				actualName = name.substring(0, name.length() - 2) + "_" + number;
			}

			if (isGs) {
				Map<Integer, Integer> bMap = CHMHelper.getFromMap(belongMap, actualName, () -> new HashMap<>());
				List<String> lines = FileUtils.readLines(file, "UTF-8");
				int i = 0;
				for (String line : lines) {
					bMap.put(i, Integer.parseInt(line));

					i++;
				}

			} else {
				Map<ClusterEntryKey, Double> sMap = CHMHelper.getFromMap(simMap, actualName, () -> new HashMap<>());
				List<String> lines = FileUtils.readLines(file, "UTF-8");

				for (String line : lines) {
					String[] split = line.split(" ");
					sMap.put(new ClusterEntryKey(Integer.parseInt(split[0]), Integer.parseInt(split[1])), Double.parseDouble(split[2]));

				}
			}

		}

		Map<String, Tri<Integer, Map<Integer, Integer>, Map<ClusterEntryKey, Double>>> clusterDataList = new HashMap<>();

		for (Entry<String, Map<Integer, Integer>> entry : belongMap.entrySet()) {

			clusterDataList.put(entry.getKey(), new Tri<>((new HashSet<>(entry.getValue().values())).size(), entry.getValue(), simMap.get(entry.getKey())));
		}

		run(clusterDataList);
	}

	private final static OktieClusterEvaluator oktieEvaluator = new OktieClusterEvaluator();
	public void run(Map<String, Tri<Integer, Map<Integer, Integer>, Map<ClusterEntryKey, Double>>> clusterDataList) {

		List<Test_Runs> testRuns = new ArrayList<>();
		// testRuns.add(Test_Runs.GSIMPL);
		// testRuns.add(Test_Runs.GSIMPL_RA);
		// testRuns.add(Test_Runs.GSIMPL_SR);
		// testRuns.add(Test_Runs.GSIMPL_RA_SR);
		testRuns.add(Test_Runs.GPARRE);
		testRuns.add(Test_Runs.GPARRE_SR);
		testRuns.add(Test_Runs.GPARRE_RA);
		PersistentParallel6.runAndFinish(clusterDataList.entrySet(), (clusterDataEntry) -> {
			
			System.out.println(clusterDataEntry.getKey());
			Tri<Integer, Map<Integer, Integer>, Map<ClusterEntryKey, Double>> clusterData = clusterDataEntry.getValue();

			{
				List<Set<Integer>> result = getClusterList(clusterData.getB().keySet(), clusterData.getC());

				OktiePerformance oktiePerformance = oktieEvaluator.oktieEvaluateWithGSIndex(clusterData.getB(), result);

				AtomicDouble totalTime = new AtomicDouble();
				AtomicInteger count = new AtomicInteger();
				AtomicDouble oktiePrecision = new AtomicDouble();
				AtomicDouble oktieRecall = new AtomicDouble();
				AtomicDouble oktieFmeasure = new AtomicDouble();

				count.addAndGet(1);

				oktiePrecision.addAndGet(oktiePerformance.getPenalizedClusteringPrecisionMacro());
				oktieRecall.addAndGet(oktiePerformance.getRecallMacro());
				oktieFmeasure.addAndGet(oktiePerformance.getFMeasureMacro());

				System.out.println(String.valueOf(1) + ";SIMPLE;" + DecimalFormatter.format(oktiePrecision.doubleValue() / count.intValue()) + ";"
						+ DecimalFormatter.format(oktieRecall.doubleValue() / count.intValue()) + ";"
						+ DecimalFormatter.format(oktieFmeasure.doubleValue() / count.intValue()) + ";" + DecimalFormatter.format(totalTime.doubleValue()));
			}

			PersistentParallel6.runAndFinish(testRuns, (testRun) -> {

				AtomicDouble totalTime = new AtomicDouble();
				AtomicInteger count = new AtomicInteger();
				AtomicDouble oktiePrecision = new AtomicDouble();
				AtomicDouble oktieRecall = new AtomicDouble();
				AtomicDouble oktieFmeasure = new AtomicDouble();

				Pair<Double, OktiePerformance> res = TestClustering.run(testRun, clusterData);

				totalTime.addAndGet(res.getFirst());

				count.addAndGet(1);

				oktiePrecision.addAndGet(res.getSecond().getPenalizedClusteringPrecisionMacro());
				oktieRecall.addAndGet(res.getSecond().getRecallMacro());
				oktieFmeasure.addAndGet(res.getSecond().getFMeasureMacro());

				System.out.println(String.valueOf(1) + ";" + testRun.name() + ";" + DecimalFormatter.format(oktiePrecision.doubleValue() / count.intValue())
						+ ";" + DecimalFormatter.format(oktieRecall.doubleValue() / count.intValue()) + ";"
						+ DecimalFormatter.format(oktieFmeasure.doubleValue() / count.intValue()) + ";" + DecimalFormatter.format(totalTime.doubleValue()));

			}, TestClustering.class, true);
		}, this.getClass(), true);


	}

	private List<Set<Integer>> getClusterList(Set<Integer> candidates, Map<ClusterEntryKey, Double> c) {

		List<Set<Integer>> clusterList = new ArrayList<>();

		final double threshold = 0.0;

		ConcurrentLinkedQueue<Set<Integer>> positivePairs = new ConcurrentLinkedQueue<>();

		PersistentParallel6.runAndFinish(c.entrySet(), (doubleEntry) -> {

			double score = doubleEntry.getValue();

			if (score > threshold) {
				Set<Integer> pairs = new HashSet<>();
				pairs.add(doubleEntry.getKey().getE1());
				pairs.add(doubleEntry.getKey().getE2());
				positivePairs.add(pairs);
			}

		}, this.getClass());

		ClusterCombiner<Integer> cc = new ClusterCombiner<>();
		cc.add(positivePairs);
		clusterList = new ArrayList<>(cc.get());
		Set<Integer> addedRowIds = new HashSet<>();
		for (Set<Integer> rc : clusterList) {
			boolean changed = addedRowIds.addAll(rc);
			if (!changed) {
				int sdf = 13;
			}
		}

		for (Integer rowId : candidates) {

			if (!addedRowIds.contains(rowId)) {
				Set<Integer> list = new HashSet<>();
				list.add(rowId);
				clusterList.add(list);
			}

		}
		return clusterList;

	}
}
