package expansion.set.strategies.rowclusterers;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Scanner;

import org.apache.commons.lang.mutable.MutableBoolean;

import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.model.TableRowIdPair;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.DataSetData;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationMethod;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationPrediction;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearningFeatures;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.ModelLearner;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.RecordPrediction;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordData;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordFeature;
import de.uni_mannheim.informatik.dws.expansion.utils.learning.weka.WekaLearnerCreator;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.learner.GeneticWeightsLearnerCreator;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;

public class MatchingRuleTest {

	public static final String POSITIVE_CLASS = "TRUE";
	public static final String NEGATIVE_CLASS = "FALSE";
	public static final String LABEL_ATTRIBUTE = "LABEL_ATTRIBUTE";
	public static final String ATTRIBUTE_ATTRIBUTE = "ATTRIBUTE_ATTRIBUTE";
	public static final String ATTRIBUTE_CONFIDENCE = "ATTRIBUTE_CONFIDENCE";


	public static final String TABLE_ATTRIBUTE = "TABLE_ATTRIBUTE";
	// public static final String DIRECT_SIM_ATTRIBUTE = "DIRECT_SIM_ATTRIBUTE";

	public static final String NOISE_ATTRIBUTE = "NOISE_ATTRIBUTE";

	public static final String IMPLICIT_ATTRIBUTE = "IMPLICIT_ATTRIBUTE";
	public static final String IMPLICIT_ATTRIBUTE_CONFIDENCE = "IMPLICIT_ATTRIBUTE_CONFIDENCE";

	public static final String A1_MATCH = "A1_MATCH";
	public static final String A1_NO_MATCH = "A1_NO_MATCH";
	public static final String A2_MATCH = "A2_MATCH";
	public static final String A2_NO_MATCH = "A2_NO_MATCH";
	public static final String A3_MATCH = "A3_MATCH";
	public static final String A3_NO_MATCH = "A3_NO_MATCH";
	public static final String A4_MATCH = "A4_MATCH";
	public static final String A4_NO_MATCH = "A4_NO_MATCH";

	public static final String I1_MATCH = "I1_MATCH";
	public static final String I1_NO_MATCH = "I1_NO_MATCH";
	public static final String I2_MATCH = "I2_MATCH";
	public static final String I2_NO_MATCH = "I2_NO_MATCH";
	public static final String I3_MATCH = "I3_MATCH";
	public static final String I3_NO_MATCH = "I3_NO_MATCH";
	public static final String I4_MATCH = "I4_MATCH";
	public static final String I4_NO_MATCH = "I4_NO_MATCH";

	public static void main(String[] args) {
		
		Scanner in = new Scanner(System.in);
		in.hasNextInt();

		long startTime = System.nanoTime();

		Map<TableRowIdPair, String> labels = new HashMap<>();
		Map<TableRowIdPair, Double[]> features = new HashMap<>();
		
		List<RecordData> recordData = new ArrayList<>();
		List<RecordData> testingRecordData = new ArrayList<>();

		Random random = new Random(1231231231536L);

		// int maxCorrect = 0;

		List<TableRowId> rowIds = new ArrayList<>();
		for (int i = 0; i < 100; i++) {

			rowIds.add(new TableRowId("", i));
		}

		List<String> classLabels = new ArrayList<>();
		classLabels.add(POSITIVE_CLASS);
		classLabels.add(NEGATIVE_CLASS);

		List<String> featureNames = new ArrayList<>();
		featureNames.add(LABEL_ATTRIBUTE);
		// featureNames.add(TABLE_ATTRIBUTE);
		// featureNames.add(DIRECT_SIM_ATTRIBUTE);
		featureNames.add(A1_MATCH);
		featureNames.add(A1_NO_MATCH);
		featureNames.add(A2_MATCH);
		featureNames.add(A2_NO_MATCH);
		featureNames.add(A3_MATCH);
		featureNames.add(A3_NO_MATCH);
		featureNames.add(A4_MATCH);
		featureNames.add(A4_NO_MATCH);

		featureNames.add(I1_MATCH);
		featureNames.add(I1_NO_MATCH);
		featureNames.add(I2_MATCH);
		featureNames.add(I2_NO_MATCH);
		featureNames.add(I3_MATCH);
		featureNames.add(I3_NO_MATCH);
		featureNames.add(I4_MATCH);
		featureNames.add(I4_NO_MATCH);
		featureNames.add(ATTRIBUTE_ATTRIBUTE);
		featureNames.add(NOISE_ATTRIBUTE);
		featureNames.add(ATTRIBUTE_CONFIDENCE);


		featureNames.add(IMPLICIT_ATTRIBUTE);
		featureNames.add(IMPLICIT_ATTRIBUTE_CONFIDENCE);

		double[] weights = new double[4];
		weights[0] = 0.4;
		weights[1] = 0.9;
		weights[2] = 0.8;
		weights[3] = 0.9;

		double[] negWeights = new double[4];
		negWeights[0] = 0.1;
		negWeights[1] = 0.2;
		negWeights[2] = 0.3;
		negWeights[3] = 0.9;

		Map<String, Double> labelsNumeric = new HashMap<>();
		labelsNumeric.put(POSITIVE_CLASS, 1.0);
		labelsNumeric.put(NEGATIVE_CLASS, -1.0);

		LearningFeatures featuresObject = new LearningFeatures(featureNames);
		DataSetData datasetData = new DataSetData("TEST", classLabels, featuresObject);
		// DataSetData datasetData = new DataSetData("TEST", labelsNumeric,
		// featuresObject);

		for (int i = 0; i < 100000; i++) {
			// TableRowIdPair pair = new TableRowIdPair(dEntry.getFirst(),
			// dEntry.getSecond());

			Double[] f = new Double[6];
			double[] aF = new double[8];
			double[] iF = new double[8];
			double directSim = 0.0;
			// double tableSim = 0.0;
			String label;

			if (random.nextDouble() > 0.8) {

				// tableSim = random.nextDouble() / 2 + 0.5;

				if (random.nextDouble() > 0.2) {

					directSim = random.nextDouble() / 4 + 0.75;
				} else {
					if (random.nextDouble() > 0.2) {
						directSim = random.nextDouble() / 2 + 0.5;
					} else {
						directSim = random.nextDouble();
					}
				}

				label = POSITIVE_CLASS;
				// labels.put(pair, POSITIVE_CLASS);
				f[0] = (random.nextDouble() / 3.0) + 0.666;
				double attributesSum = 0.0;
				int attributesSize = 0;

				// int attributesSize = random.nextInt(3) + 1;
				// int attributesSum = random.nextInt(3) + 1;
				for (int a = 0; a < 4; a++) {
					if (random.nextDouble() > 0.5) {
						attributesSize++;

						if (random.nextDouble() < weights[a]) {
							attributesSum += 1.0;
							aF[a * 2] = 1.0;
							aF[a * 2 + 1] = 0.0;
						} else {
							// if (random.nextDouble() > 0.5) {
							// attributesSum += 1.0;
							// aF[a * 2] = 1.0;
							// aF[a * 2 + 1] = 0.0;
							// } else {

								attributesSum += 0.0;
								aF[a * 2] = 0.0;
								aF[a * 2 + 1] = 1.0;
							// }
						}
					}
					else {
						aF[a * 2] = 0.0;
						aF[a * 2 + 1] = 0.0;
					}
				}

				f[1] = attributesSum / attributesSize;
				f[3] = (double) attributesSize;

				// f[5] = random.nextDouble();
				// if (f[5] > 0.8) {
				// f[4] = random.nextDouble() / 5.0 + 0.8;
				// } else {
				// f[4] = random.nextDouble();
				// }

				for (int a = 0; a < 4; a++) {
					if (random.nextDouble() > 0.5) {
						attributesSize++;

						if (random.nextDouble() < weights[a]) {
							// attributesSum += 1.0;
							iF[a * 2] = random.nextDouble();
							iF[a * 2 + 1] = 0.0;
						} else {
							// if (random.nextDouble() > 0.5) {
							// // attributesSum += 1.0;
							// iF[a * 2] = random.nextDouble();
							// iF[a * 2 + 1] = 0.0;
							// } else {

								// attributesSum += 0.0;
								iF[a * 2] = 0.0;
								iF[a * 2 + 1] = random.nextDouble();
							// }
						}
					} else {
						if (random.nextDouble() > 0.5) {
							iF[a * 2] = random.nextDouble() / 6.0;
							iF[a * 2 + 1] = 0.0;
						} else {
							iF[a * 2] = 0.0;
							iF[a * 2 + 1] = random.nextDouble() / 6.0;
						}
					}
				}

				// maxCorrect++;
			} else {
				// tableSim = random.nextDouble();
				// if (random.nextDouble() > 0.7) {
				// directSim = random.nextDouble() / 4;
				// } else {
				// directSim = random.nextDouble() / 2;
				// }
				label = NEGATIVE_CLASS;
				// labels.put(pair, NEGATIVE_CLASS);
				if (random.nextDouble() > 0.7) {
					f[0] = random.nextDouble();
				}
				else {
					f[0] = (random.nextDouble() / 3.0) + 0.666;
				}


				double attributesSum = 0.0;
				int attributesSize = 0;

				// int attributesSize = random.nextInt(3) + 1;
				// int attributesSum = random.nextInt(3) + 1;
				for (int a = 0; a < 4; a++) {
					if (random.nextDouble() > 0.5) {
						attributesSize++;

//						if (random.nextDouble() < negWeights[a]) {
//							attributesSum += 0.0;
//							aF[a * 2] = 0.0;
//							aF[a * 2 + 1] = 1.0;
//						} else {
						if (random.nextDouble() < negWeights[a]) {
								attributesSum += 1.0;
								aF[a * 2] = 1.0;
								aF[a * 2 + 1] = 0.0;
							 } else {
							
							 attributesSum += 0.0;
							 aF[a * 2] = 0.0;
							 aF[a * 2 + 1] = 1.0;
							 }
						// }
					}
					else {
						aF[a * 2] = 0.0;
						aF[a * 2 + 1] = 0.0;
					}
				}

				// double attributesSum = 0.0;
				// int attributesSize = random.nextInt(3) + 1;
				// for (int a = 0; a < attributesSize; a++) {
				// if (random.nextDouble() > 0.7) {
				// attributesSum += random.nextDouble() / 0.0;
				// } else {
				// if (random.nextDouble() > 0.5) {
				// attributesSum += 1.0;
				// } else {
				//
				// attributesSum += 0.0;
				// }
				// }
				//
				// }
				f[1] = attributesSum / attributesSize;
				f[3] = (double) attributesSize;

				for (int a = 0; a < 4; a++) {
					if (random.nextDouble() > 0.5) {
						attributesSize++;

//						if (random.nextDouble() < negWeights[a]) {
//							// attributesSum += 1.0;
//							iF[a * 2] = 0.0;
//							iF[a * 2 + 1] = random.nextDouble();
//						} else {
						if (random.nextDouble() < negWeights[a]) {
//								 attributesSum += 1.0;
								iF[a * 2] = random.nextDouble();
								iF[a * 2 + 1] = 0.0;
							 } else {
							
								 // attributesSum += 0.0;
								 iF[a * 2] = 0.0;
								 iF[a * 2 + 1] = random.nextDouble();
							 }
//						}
					} else {
						if (random.nextDouble() > 0.5) {
							iF[a * 2] = random.nextDouble() / 6.0;
							iF[a * 2 + 1] = 0.0;
						} else {
							iF[a * 2] = 0.0;
							iF[a * 2 + 1] = random.nextDouble() / 6.0;
						}
					}
				}

			}

			f[2] = random.nextDouble();

			// features.put(pair, f);
			

			LinkedHashMap<String, Double> rFeatures = new LinkedHashMap<>();
			rFeatures.put(LABEL_ATTRIBUTE, f[0]);

			if (true) {

				double aSum = 0.0;
				double aCorrect = 0.0;
				for (int a = 0; a < 4; a++) {
					aCorrect += aF[a * 2];
					aSum += aF[a * 2];
					aSum += aF[a * 2 + 1];
				}
				rFeatures.put(ATTRIBUTE_ATTRIBUTE, aCorrect / aSum);
				rFeatures.put(ATTRIBUTE_CONFIDENCE, aSum);

				double iSum = 0.0;
				double iCorrect = 0.0;
				for (int a = 0; a < 4; a++) {
					iCorrect += iF[a * 2];
					iSum += iF[a * 2];
					iSum += iF[a * 2 + 1];
				}

				rFeatures.put(IMPLICIT_ATTRIBUTE, iCorrect / iSum);
				rFeatures.put(IMPLICIT_ATTRIBUTE_CONFIDENCE, iSum);
				//
				// aF = new double[8];
				// iF = new double[8];

			} else {
				rFeatures.put(ATTRIBUTE_ATTRIBUTE, 0.0);
				rFeatures.put(ATTRIBUTE_CONFIDENCE, 0.0);
				rFeatures.put(IMPLICIT_ATTRIBUTE, 0.0);
				rFeatures.put(IMPLICIT_ATTRIBUTE_CONFIDENCE, 0.0);
			}

			// rFeatures.put(TABLE_ATTRIBUTE, tableSim);
			// rFeatures.put(DIRECT_SIM_ATTRIBUTE, directSim);

			rFeatures.put(A1_MATCH, aF[0]);
			rFeatures.put(A1_NO_MATCH, aF[1]);
			rFeatures.put(A2_MATCH, aF[2]);
			rFeatures.put(A2_NO_MATCH, aF[3]);
			rFeatures.put(A3_MATCH, aF[4]);
			rFeatures.put(A3_NO_MATCH, aF[5]);
			rFeatures.put(A4_MATCH, aF[6]);
			rFeatures.put(A4_NO_MATCH, aF[7]);

			rFeatures.put(NOISE_ATTRIBUTE, f[2]);
			// rFeatures.put(ATTRIBUTE_CONFIDENCE, f[3]);
			
			rFeatures.put(I1_MATCH, iF[0]);
			rFeatures.put(I1_NO_MATCH, iF[1]);
			rFeatures.put(I2_MATCH, iF[2]);
			rFeatures.put(I2_NO_MATCH, iF[3]);
			rFeatures.put(I3_MATCH, iF[4]);
			rFeatures.put(I3_NO_MATCH, iF[5]);
			rFeatures.put(I4_MATCH, iF[6]);
			rFeatures.put(I4_NO_MATCH, iF[7]);

			// rFeatures.put(IMPLICIT_ATTRIBUTE, f[4]);
			// rFeatures.put(IMPLICIT_ATTRIBUTE_CONFIDENCE, f[5]);

			if (i > 85000) {
				testingRecordData.add(new RecordData(label, new RecordFeature(rFeatures, datasetData)));
			} else {
				recordData.add(new RecordData(label, new RecordFeature(rFeatures, datasetData)));
			}
		}
		
		double positive = 0;
		double negative = 0;
		for (RecordData rd : recordData) {
			if (rd.getLabel().equals(POSITIVE_CLASS)) {
				positive++;
			} else {
				negative++;
			}
		}

		Collections.shuffle(recordData);

		List<RecordData> newPositve = new ArrayList<>();
		double remove = negative - positive;
		double multiplicator = remove / positive;
		double positiveWeight = negative / positive;
		int ni = 0;
		for (Iterator<RecordData> it = recordData.iterator(); it.hasNext();) {
			RecordData next = it.next();
			if (next.getLabel().equals(POSITIVE_CLASS)) {

				// for (int i = 0; i < multiplicator; i++) {
				// }
			} else {
				// it.remove();
				if (ni < positive) {
					newPositve.add(new RecordData(next.getLabel(), 1.0, next.getFeatures()));
					ni++;
				}
			}
		}
		
		// recordData.addAll(newPositve);

		double positiveNew = 0;
		double negativeNew = 0;
		for (RecordData rd : recordData) {
			if (rd.getLabel().equals(POSITIVE_CLASS)) {
				positiveNew += rd.getWeight();
			} else {
				negativeNew += rd.getWeight();
			}
		}

		TrainingData trainingData = new TrainingData(datasetData, recordData);
		
		GeneticWeightsLearnerCreator creator = new GeneticWeightsLearnerCreator();
		OptimizableClass optimizableClass = new OptimizableClass("TEST");
		ModelLearner learner = creator.getLearner((rFeatures, rWeights) -> {
			double sum = 0.0;
			double count = 0.0;
			for (Entry<String, Double> entry : rFeatures.getFeatures().entrySet()) {
				double weight = rWeights.getOrDefault(optimizableClass.getWeight(entry.getKey(), 1.0));
				sum += entry.getValue() * weight;
				count += weight;
			}
			double ret = sum / count;
			double threshold = rWeights.getOrDefault(optimizableClass.getWeight("THRESHOLD", 0.5));
			if (ret > threshold) {
				double confidence = (ret - threshold) / threshold;
				return new RecordPrediction(POSITIVE_CLASS, confidence);
			} else {
				double confidence = 1 - ((ret) / threshold);
				return new RecordPrediction(NEGATIVE_CLASS, confidence);
			}
		});
		
		MutableBoolean mb = new MutableBoolean(false);
		EvaluationMethod method = (predictions) -> {

			double correct = 0.0;
			int wrong = 0;
			int maxCorrect = 0;

			for (EvaluationPrediction prediction : predictions) {
				if (prediction.getLabel().equals(POSITIVE_CLASS)) {
					maxCorrect++;
					if (prediction.getPredictedLabel().equals(POSITIVE_CLASS)) {
						correct++;
					}
				} else {

					if (prediction.getPredictedLabel().equals(POSITIVE_CLASS)) {
						wrong++;
					}
				}
			}

			double precision = correct / (correct + wrong);
			double recall = correct / maxCorrect;

			double fMeasure = (2.0 * precision * recall) / (precision + recall);

			if (mb.booleanValue()) {
				System.out.println("PRECISION: " + precision);
				System.out.println("RECALL: " + recall);
			}

			return fMeasure;

		};

		// LearnedModel learnedModel = learner.learn(trainingData, method);
		//
		// System.out.println(learnedModel.runAndEvaluate(recordData, method));
		//
		// System.out.println(learnedModel.runAndEvaluate(testingRecordData,
		// method));
		LearnedModel model;
		WekaLearnerCreator wekaCreator;
		LearnedModel wekaLearnedModel;

		// LearnedModel model = learner.learn(evaluator)

		// System.out.println("SimpleLogistic");
		//
		// WekaLearnerCreator wekaCreator = new
		// WekaLearnerCreator("SimpleLogistic", new String[] {});
		// LearnedModel wekaLearnedModel = wekaCreator.learn(trainingData,
		// method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));
		//
		//
		//
		// System.out.println("Decision Tree");
		//
		// wekaCreator = new WekaLearnerCreator("J48", new String[] {});
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));

		// System.out.println("Decision Tree - FS");
		//
		// wekaCreator = new WekaLearnerCreator("J48", new String[] {}, true,
		// false);
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));
		//
		// System.out.println("Decision Tree - BS");
		//
		// wekaCreator = new WekaLearnerCreator("J48", new String[] {}, true,
		// true);
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));

		// System.out.println("RandomForest");
		//
		// wekaCreator = new WekaLearnerCreator("RandomForest", new String[]
		// {});
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));
		//
		// System.out.println("RandomForest with Options");
		//
		// wekaCreator = new WekaLearnerCreator("RandomForest", new String[] {
		// "-N", "5" });
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));

		// System.out.println("SimpleLogistic");
		//
		// wekaCreator = new WekaLearnerCreator("SimpleLogistic", new String[]
		// {});
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));
		//
		// System.out.println("LMT");
		//
		// wekaCreator = new WekaLearnerCreator("LMT", new String[] {});
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));

		// System.out.println("RandomForest");
		// wekaCreator = new WekaLearnerCreator("RandomForest", new String[] {
		// "-P", "66", "-print" });
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));
		//
		// System.out.println("RandomForest");
		// wekaCreator = new WekaLearnerCreator("RandomForest", new String[] {
		// "-N", "3", "-print" });
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));

		mb.setValue(true);

		// System.out.println("SimpleLogistic");
		// wekaCreator = new WekaLearnerCreator("SimpleLogistic", new String[]
		// {});
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));
		//
		// System.out.println("Logistic");
		// wekaCreator = new
		// WekaLearnerCreator("weka.classifiers.functions.Logistic", new
		// String[] {});
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));

		System.out.println("RandomForest mit P");
		wekaCreator = new WekaLearnerCreator("RandomForest", new String[] { "-U", "-num-slots", "5", "-I", "600", "-K", "2", "-P", "85" });
		wekaLearnedModel = wekaCreator.learn(trainingData, method);

		System.out.println(wekaLearnedModel.runAndEvaluate(recordData, method));

		System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData, method));

		// System.out.println("RandomForest mit U");
		// wekaCreator = new WekaLearnerCreator("RandomForest", new String[] {
		// "-U", "-N", "4", "-num-slots", "5", "-I", "1000" });
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));
		//
		// System.out.println("RandomForest ohne U");
		// wekaCreator = new WekaLearnerCreator("RandomForest", new String[] {
		// "-N", "4", "-num-slots", "5", "-I", "1000" });
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));

		// System.out.println("RandomForest");
		// wekaCreator = new WekaLearnerCreator("RandomForest", new String[] {
		// "-U", "-P", "85", "-num-slots", "5", "-I", "1000" });
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));

		// System.out.println("RandomForest OOB");
		// wekaCreator = new WekaLearnerCreator("RandomForest", new String[] {
		// "-P", "80", "-print", "-O", "-M", "100", "-K", "2" });
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));
		//
		// System.out.println("RandomForest OOB and Store");
		// wekaCreator = new WekaLearnerCreator("RandomForest",
		// new String[] { "-P", "66", "-store-out-of-bag-predictions" });
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));

		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));

		//
		// System.out.println("BAGGING LMT");
		//
		// wekaCreator = new WekaLearnerCreator("Bagging", new String[] { "-W",
		// "LMT" });
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));
		//


		// System.out.println("LMT with options");

		// System.out.println("LMT");
		//
		// wekaCreator = new WekaLearnerCreator("LMT", new String[] { "-P" });
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));
		//
		// System.out.println("LMT with options");
		//
		// wekaCreator = new WekaLearnerCreator("LMT", new String[] { "-P",
		// "-doNotMakeSplitPointActualValue", "-M", "100" });
		// wekaLearnedModel = wekaCreator.learn(trainingData, method);
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(recordData,
		// method));
		//
		// System.out.println(wekaLearnedModel.runAndEvaluate(testingRecordData,
		// method));

		long estimatedTime = System.nanoTime() - startTime;

		double seconds = (double) estimatedTime / 1000000000.0;
		System.out.println(seconds);

		int fsdf = 123;

		try {
			PersistentParallel6.close();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

}
