package expansion.set.strategies.rowclusterers;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ThreadLocalRandom;

import org.apache.commons.math3.optim.linear.LinearConstraint;
import org.apache.commons.math3.optim.linear.LinearConstraintSet;
import org.apache.commons.math3.optim.linear.LinearObjectiveFunction;
import org.apache.commons.math3.optim.linear.Relationship;

import com.joptimizer.optimizers.LPOptimizationRequest;
import com.quantego.clp.CLP;
import com.quantego.clp.CLP.ALGORITHM;
import com.quantego.clp.CLP.STATUS;
import com.quantego.clp.CLPVariable;

import de.mannheim.uni.expansion.utils.DoubleIteraterable;
import de.mannheim.uni.expansion.utils.DoubleIteraterable.DoubleEntry;
import expansion.set.clustering.model.RowCluster.ClusterEntryKey;
import expansion.set.strategies.rowclusterers.TestClustering.ClusterSimilaritiesWrapper;

public class PivotClustering {

	
	public List<Set<Integer>> cluster(ClusterSimilaritiesWrapper clusterSimilaritiesWrapper, Set<Integer> keySet, boolean full) {
		
		CLP solver = new com.quantego.clp.CLP().algorithm(ALGORITHM.PRIMAL).minimization();
				
		List<Set<Integer> >ret = new ArrayList<>(); 
		 
		
		CLPVariable[][] indexPositive = new CLPVariable[keySet.size()][keySet.size()];
		CLPVariable[][] indexNegative = new CLPVariable[keySet.size()][keySet.size()];
		
 		
		for(int i = 0; i < keySet.size(); i++) {
				int jStart = i + 1;
				if(full) {
					jStart = 0;
				}
				for(int j = jStart; j < keySet.size(); j++) {
			 
					if(i != j) {
						double jiScore = clusterSimilaritiesWrapper.get(new ClusterEntryKey(j, i));
	
						if (jiScore < 0.0) {
							jiScore = Math.abs(jiScore) ;
							indexPositive[i][j] = solver.addVariable().lb(0.0).ub(1.0).obj(jiScore ).name("POS_"+i+"_"+j);
	
						} else {
							indexPositive[i][j] = solver.addVariable().lb(0.0).ub(1.0).name("POS_"+i+"_"+j);
	
							jiScore = 1.0 - jiScore;
						}
	
	
						double ijScore = clusterSimilaritiesWrapper.get(new ClusterEntryKey(i, j));
	
						if (ijScore < 0.0) {
							ijScore = 1.0 - Math.abs(ijScore);
							indexNegative[i][j] = solver.addVariable().lb(0.0).ub(1.0).name("NEG_"+j+"_"+i);
						}
						else {
							indexNegative[i][j] = solver.addVariable().lb(0.0).ub(1.0).obj(ijScore ).name("NEG_"+j+"_"+i);
	
						}

					}
				 
			}
		}
		
		
	 
		
		for(int i = 0; i < keySet.size(); i++) {
			
			int kStart = i + 1;
			if(full) {
				kStart = 0;
			}
			
			for(int k = kStart; k < keySet.size(); k++) {
				
				int jStart = k + 1;
				if(full) {
					jStart = 0;
				}
				
				for(int j = jStart; j < keySet.size(); j++) {
					 
							 if((i != j) && (i != k) && (k != j)) {
								 solver.createExpression().add(1.0,indexNegative[i][k]).add(-1.0,indexNegative[i][j]).add(-1.0,indexNegative[k][j]).leq(0.0);
							 }
							  
					 
				}				 
			}
		}
//		
		
		for(int i = 0; i < keySet.size(); i++) {
			 
				
				int jStart = i + 1;
				if(full) {
					jStart = 0;
				}
			
				for(int j = jStart; j < keySet.size(); j++) {

			 
					if(i != j) {
					
						solver.createExpression().add(1.0,indexPositive[i][j]).add(1.0,indexNegative[i][j]).eq(1.0);
					}
 
			}

		}
		
  
		STATUS solved = solver.solve();
		double[][] solutionPositive = new double[keySet.size()][keySet.size()];
		double[][] solutionNegative = new double[keySet.size()][keySet.size()];
		
		for(int i = 0; i < keySet.size(); i++) {

			int jStart = i + 1;
			if(full) {
				jStart = 0;
			}
			
			for(int j = jStart; j < keySet.size(); j++) {
				if(!(i == j)) {
					solutionPositive[i][j] = solver.getSolution(indexPositive[i][j]);
					solutionNegative[i][j] = solver.getSolution(indexNegative[i][j]);
				}
				 
			}
		}
//		System.out.println(solvedString);
		
		List<Integer> nodes = new ArrayList<>(keySet);
		
		while(!nodes.isEmpty()) {
			
			int randomIndex = ThreadLocalRandom.current().nextInt(nodes.size());
			
			 
			Integer random = nodes.get(randomIndex);
			
			Set<Integer> cluster = new HashSet<>();
			cluster.add(random);
			for(Iterator<Integer> it = nodes.iterator(); it.hasNext();) {
				Integer candidate = it.next();
				double posDirect = solutionPositive[random][candidate];
				double posReverse = solutionPositive[candidate][random];
				double negDirect = solutionNegative[random][candidate];
				double negReverse = solutionNegative[candidate][random];
				
				double remaining = solutionNegative[random][candidate] - (1.0 - solutionPositive[random][candidate]);
				
				
				if(candidate.equals(random)) {
					it.remove();
				}
				else {
					int first = random;
					int second = candidate;
					if(candidate < random) {
						first = candidate;
						second = random;
					}
					double randomDouble = ThreadLocalRandom.current().nextDouble();
					if(solutionPositive[first][second] - randomDouble >= 0.0 ) {
						it.remove();
						cluster.add(candidate);
					}
					else  {
						//it.remove();
						int sdf= 123;
					}
				} 
			}
			
			
			ret.add(cluster);
			
		}
		return ret;
	}
	
	public List<Set<Integer>> clusterApache(ClusterSimilaritiesWrapper clusterSimilaritiesWrapper, Set<Integer> keySet) {
		List<Set<Integer> >ret = new ArrayList<>(); 
		
		LPOptimizationRequest or = new LPOptimizationRequest();
		
		int[][] indexPositive = new int[keySet.size()][keySet.size()];
		int[][] indexNegative = new int[keySet.size()][keySet.size()];
		
		int idx = 0;
		for(DoubleEntry<Integer> dd : DoubleIteraterable.iterable(keySet)) {
			indexPositive[dd.getFirst()][dd.getSecond()] = idx;
			idx++;
			indexPositive[dd.getSecond()][dd.getFirst()] = idx;
			idx++;
			
			indexNegative[dd.getFirst()][dd.getSecond()] = idx;
			idx++;
			indexNegative[dd.getSecond()][dd.getFirst()] = idx;
			idx++;
		}
		
		
		List<LinearConstraint> constraints = new ArrayList<>();
		
		for(Integer i : keySet) {
			
			for(Integer k : keySet) {
				
				if(!i.equals(k)) {

				 double[] lhs = new double[idx];
				 Arrays.fill(lhs, 0.0);
				 lhs[indexNegative[i][k]] = 1.0;
				 
					for(Integer j : keySet) {
					
						if(!i.equals(j) && !k.equals(j)) {
							
							 double[] rhs = new double[idx];
							 Arrays.fill(rhs, 0.0);
							 rhs[indexNegative[i][j]] = 1.0;
							 rhs[indexNegative[j][k]] = 1.0;
							 
							constraints.add( new LinearConstraint(lhs, 0.0, Relationship.LEQ, rhs, 0.0));
							
						}
						
					}
				 
				
				}				 
			}
		}
		
		
		for (Integer i : keySet) {

			for (Integer k : keySet) {

				if (!i.equals(k)) {
					
					 double[] lhs = new double[idx];
					 Arrays.fill(lhs, 0.0);
					 lhs[indexPositive[i][k]] = 1.0;
					 lhs[indexNegative[i][k]] = 1.0;
					 
					constraints.add( new LinearConstraint(lhs, Relationship.EQ, 1.0));
				}
			}

		}
		
		
		for (Integer i : keySet) {

			for (Integer k : keySet) {

				if (!i.equals(k)) {
					
					 double[] lhs = new double[idx];
					 Arrays.fill(lhs, 0.0);
					 lhs[indexPositive[i][k]] = 1.0;
					 
					constraints.add( new LinearConstraint(lhs, Relationship.GEQ, 0.0));
					
					
					lhs = new double[idx];
					 Arrays.fill(lhs, 0.0);
					 lhs[indexNegative[i][k]] = 1.0;
					 
					 constraints.add( new LinearConstraint(lhs, Relationship.GEQ, 0.0));
				}
			}
		}
		 
		double[] objectiveCoefficients = new double[idx];
		Arrays.fill(objectiveCoefficients, 0.0);

		
		LinearObjectiveFunction objective = new LinearObjectiveFunction(objectiveCoefficients, 0);
		LinearConstraintSet constraintSet = new LinearConstraintSet(constraints);
		
		CLP solver = new com.quantego.clp.CLP().buffer(2).minimization();
		
//		PointValuePair solution = solver.optimize(objective, constraintSet, GoalType.MINIMIZE);
		 
		
		return ret;
	}
	
	
	public static void main(String[] args ) { 
			CLP solver = new com.quantego.clp.CLP().buffer(2).maximization();
		    CLPVariable x0 = solver.addVariable().ub(1.0);
		    CLPVariable x1 = solver.addVariable().ub(0.3).obj(2.655523).name("var_1");
		    CLPVariable x2 = solver.addVariable().ub(0.3).obj(-2.70917);
		    CLPVariable x3 = solver.addVariable().free().obj(1);
		    solver.createExpression().add(-2,x0).add(-1.484345,x0).add(x3).leq(0.302499);
		    solver.createExpression().add(-3.074807,x0).add(x3).leq(0.507194);
		    solver.createExpression().add(x0).add(1.01,x1).add(-.99,x2).eq(0.594).name("eq_ctr");
		    String str = "Maximize\n"
		    		+ "obj: + 2.655523 var_1 - 2.70917 x_2 + x_3\n"
		    		+ "Subject To\n"
		    		+ "ctr_0: - 3.4843450000000002 x_0 + x_3 <= 0.302499\n"
		    		+ "ctr_1: - 3.074807 x_0 + x_3 <= 0.507194\n"
		    		+ "eq_ctr: + x_0 + 1.01 var_1 - 0.99 x_2 = 0.594\n"
		    		+ "Bounds\n"
		    		+ "x_0 <= 1.0\n"
		    		+ "var_1 <= 0.3\n"
		    		+ "x_2 <= 0.3\n"
		    		+ "-inf <= x_3 <= inf\n"
		    		+ "End";
		    System.out.println(solver.toString());
		}
	}
	
	
 