package expansion.set.strategies.rowclusterers;

import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;

import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;

import de.mannheim.uni.expansion.utils.MapSorter;
import de.mannheim.uni.expansion.utils.containers.Pair;

public class RangedMap<K, V extends Comparable<V>> {
	private final BiMap<Integer, K> keyMap;
	private final LinkedHashMap<K, V> valueMap;

	public RangedMap(Map<K, V> input) {
		LinkedHashMap<K, V> sortedMap = MapSorter.sortHashMapByValuesAscending(input);
		keyMap = HashBiMap.create(input.size());
		this.valueMap = sortedMap;

		int i = 0;
		for (Entry<K, V> entry : sortedMap.entrySet()) {
			
			keyMap.put(i, entry.getKey());
			i++;
		}

	}

	public Set<K> greaterEqualThan(V v) {
		Optional<Pair<Integer, Integer>> res = binarySearch(v);

		if (res.isPresent()) {
			// This is unnecessary but is kept for comparability
			if (res.get().getFirst().equals(res.get().getSecond())) {
				return fromIncluding(res.get().getSecond());
			} else {
				return fromIncluding(res.get().getSecond());
			}

		}
		return new HashSet<>();
	}

	public Set<K> greaterThan(V v) {
		Optional<Pair<Integer, Integer>> res = binarySearch(v);

		if (res.isPresent()) {
			if (res.get().getFirst().equals(res.get().getSecond())) {
				// Perfect match! as we do not have equal we can not return the
				// perfect match itself
				return fromIncluding(res.get().getSecond() + 1);
			} else {
				// Non-Perfect match, second is greater than threshold
				return fromIncluding(res.get().getSecond());
			}

		}
		return new HashSet<>();
	}

	public Set<K> lowerThan(V v) {
		Optional<Pair<Integer, Integer>> res = binarySearch(v);

		if (res.isPresent()) {
			if (res.get().getFirst().equals(res.get().getSecond())) {
				// Perfect match! as we do not have equal we can not return the
				// perfect match itself
				return toIncluding(res.get().getFirst() - 1);
			} else {
				return toIncluding(res.get().getFirst());
			}

		}
		return new HashSet<>();
	}

	public Set<K> lowerEqualThan(V v) {
		Optional<Pair<Integer, Integer>> res = binarySearch(v);

		if (res.isPresent()) {
			// This is unnecessary but is kept for comparability
			if (res.get().getFirst().equals(res.get().getSecond())) {
				return toIncluding(res.get().getFirst());
			} else {
				return toIncluding(res.get().getFirst());
			}

		}
		return new HashSet<>();
	}

	private Set<K> fromIncluding(Integer from) {
		HashSet<K> ret = new HashSet<>();
		for (int i = from; i < keyMap.size(); i++) {
			ret.add(keyMap.get(i));
		}
		return ret;

	}

	private Set<K> toIncluding(Integer to) {
		HashSet<K> ret = new HashSet<>();
		if (to < keyMap.size()) {
			for (int i = 0; i <= to; i++) {
				ret.add(keyMap.get(i));
			}
		}
		return ret;
	}

	private Optional<Pair<Integer, Integer>> binarySearch(V v) {

		int low = 0;
		int high = keyMap.size() - 1;

		int oldCompare = 0;
		int compare = 0;

		if (valueMap.get(keyMap.get(0)).compareTo(v) <= 0 && valueMap.get(keyMap.get(keyMap.size() - 1)).compareTo(v) >= 0) {

			while (true) {
				int middle = (low + high) / 2;
				if (oldCompare == -1 && middle - low <= 1) {
					if(valueMap.get(keyMap.get(low)).compareTo(v) == 0) {
						return Optional.of(new Pair<>(low, low));
					}
					else if(valueMap.get(keyMap.get(middle)).compareTo(v) == 0) {
						return Optional.of(new Pair<>(middle, middle));
					}
					return Optional.of(new Pair<>(low, middle));
				} else if (oldCompare == 1 && high - middle <= 1) {
					if(valueMap.get(keyMap.get(high)).compareTo(v) == 0) {
						return Optional.of(new Pair<>(high, high));
					}
					else if(valueMap.get(keyMap.get(middle)).compareTo(v) == 0) {
						return Optional.of(new Pair<>(middle, middle));
					}
					return Optional.of(new Pair<>(middle, high));
				}
				V midV = valueMap.get(keyMap.get(middle));
				compare = midV.compareTo(v);

				if (compare == 0) {
					return Optional.of(new Pair<>(middle, middle));
				} else if (compare == 1) {
					high = middle;

				} else if (compare == -1) {
					low = middle;
				}
				oldCompare = compare;
			}
		}

		return Optional.empty();

	}

}
