package expansion.set.strategies.rowclusterers;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import com.google.common.collect.Sets;

import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.fastparallel.WeightsFastParallelOptimizer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.genetic.WeightsGeneticParallelOptimizer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.parallel.WeightsParallelOptimizer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.random.WeightsRandomParallelOptimizer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.simple.SimpleOptimizer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeight;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeightKey;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;

public class WeightsOptimizingTest {

	static OptimizableClass clazz = new OptimizableClass("TEST");

	static double scoreA = 0.62;

	static double scoreB = 0.10;

	static double scoreC = 0.80;

	static double scoreD = 0.30;

	static double scoreE = 0.74;

	public static void main(String[] args) {


		
		SimpleOptimizer simple = new SimpleOptimizer();
		WeightsParallelOptimizer parallel = new WeightsParallelOptimizer();
		WeightsGeneticParallelOptimizer genetic = new WeightsGeneticParallelOptimizer();
		WeightsFastParallelOptimizer fast = new WeightsFastParallelOptimizer();
		WeightsRandomParallelOptimizer random = new WeightsRandomParallelOptimizer();

		if (true) {
			{
			double sum = 0.0;
			int count = 0;
			for (int i = 0; i < 100; i++) {
					clazz = new OptimizableClass("TEST_" + i);
					WeightsRepository weights = genetic.optimize(WeightsOptimizingTest::run);
					// System.out.println(weights.toString());
				sum += run(weights);
				count++;
			}

				System.out.println("Average Performance: " + (sum / count));
				System.out.println("TotalBreakage: " + WeightsGeneticParallelOptimizer.totalBreakage);
			}

			if (false)
			{
				double sum = 0.0;
				int count = 0;
				for (int i = 0; i < 100; i++) {
					clazz = new OptimizableClass("TEST_" + i);
					WeightsRepository weights = random.optimize(WeightsOptimizingTest::run);

					sum += run(weights);
					count++;
				}

				System.out.println("Average Performance: " + (sum / count));
			}
		} else {
			long startTime;
			long estimatedTime;
			double seconds;

			if (false) {

				startTime = System.nanoTime();
				System.out.println(simple.optimize(WeightsOptimizingTest::run).getWeights().toString());
				//
				estimatedTime = System.nanoTime() - startTime;
				seconds = (double) estimatedTime / 1000000000.0;
				System.out.println("Simple: " + seconds);
			}

			if (false) {
				startTime = System.nanoTime();

				WeightsRepository weights = parallel.optimize(WeightsOptimizingTest::run);

				System.out.println(weights.getWeights());
				System.out.println(run(weights));
				estimatedTime = System.nanoTime() - startTime;
				seconds = (double) estimatedTime / 1000000000.0;
				System.out.println("Parallel: " + seconds);
			}

			if (false) {
				startTime = System.nanoTime();

				WeightsRepository fastWeights = fast.optimize(WeightsOptimizingTest::run, null);

				System.out.println(fastWeights.getWeights());
				System.out.println(run(fastWeights));

				estimatedTime = System.nanoTime() - startTime;
				seconds = (double) estimatedTime / 1000000000.0;
				System.out.println("Fast: " + seconds);
			}

			if (true) {
				startTime = System.nanoTime();

				WeightsRepository weights = genetic.optimize(WeightsOptimizingTest::run);

				System.out.println(weights.getWeights());
				System.out.println(run(weights));
				estimatedTime = System.nanoTime() - startTime;
				seconds = (double) estimatedTime / 1000000000.0;
				System.out.println("Genetic: " + seconds);
			}

			startTime = System.nanoTime();

			WeightsRepository randomWeights = random.optimize(WeightsOptimizingTest::run);

			System.out.println(randomWeights.getWeights());
			System.out.println(run(randomWeights));

			estimatedTime = System.nanoTime() - startTime;
			seconds = (double) estimatedTime / 1000000000.0;
			System.out.println("Random: " + seconds);

			if (false) {
			startTime = System.nanoTime();

				List<Set<String>> priorRuns = new ArrayList<>();
				priorRuns.add(Sets.newHashSet("A", "D"));
				priorRuns.add(Sets.newHashSet("A", "B"));
				priorRuns.add(Sets.newHashSet("A", "C"));

				WeightsRepository fastWeights2 = fast.optimize(WeightsOptimizingTest::run, priorRuns);

				System.out.println(fastWeights2.getWeights());

			estimatedTime = System.nanoTime() - startTime;
			seconds = (double) estimatedTime / 1000000000.0;
				System.out.println("Fast: " + seconds);
			}


		}

		try {
			PersistentParallel6.close();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public static double run(WeightsRepository repository) {

		// waste2(0.5);

		double weightA = repository.getOrDefault(new OptimizableWeight(new OptimizableWeightKey(clazz, "A"), 1.0));
		double performanceA = 1.0 - Math.abs(weightA - scoreA);

		double weightB = repository.getOrDefault(new OptimizableWeight(new OptimizableWeightKey(clazz, "B"), 1.0));
		double performanceB = 1.0 - Math.abs(weightB - scoreB);
		if (performanceB < 0.4) {
			performanceB = 0.0;
		}


		double weightC = repository.getOrDefault(new OptimizableWeight(new OptimizableWeightKey(clazz, "C"), 1.0));
		double performanceC = 1.0 - Math.abs(weightC - scoreC);

		if (performanceC < performanceA) {
			performanceC = 0.0;
		}

		double weightD = repository.getOrDefault(new OptimizableWeight(new OptimizableWeightKey(clazz, "D"), 1.0));
		double performanceD = 1.0 - Math.abs(weightD - scoreD);
		if (performanceD < 0.9) {
			performanceD = 0.0;
		}

		double weightE = repository.getOrDefault(new OptimizableWeight(new OptimizableWeightKey(clazz, "E"), 1.0));

		double performanceE = 1.0 - ((Math.abs(weightE - scoreE) * 1.0 + 0.5 * Math.abs(weightA - scoreA)) / 1.5);

		double score = (((performanceA - (1.0 - Math.pow(performanceC, 2.0))) * performanceB - (1.0 - performanceD))
				/ (1 + ((1.0 - (performanceC - (1.0 - performanceA))) + (1.0 - (performanceD - (1.0 - Math.sqrt(performanceB))))))) - (1.0 - performanceE);

		return score;
	}

	public static double waste4(double a) {
		return waste3(waste3(waste3(waste3(waste3(waste3(waste3(waste3(waste3(waste3(waste3(a)))))))))));
	}

	public static double waste3(double a) {
		return waste2(waste2(waste2(waste2(waste2(waste2(waste2(waste2(waste2(waste2(waste2(a)))))))))));
	}

	public static double waste2(double a) {
		return waste(waste(waste(waste(waste(waste(waste(waste(waste(waste(waste(a)))))))))));
	}

	public static double waste(double a) {
		return tan(atan(tan(atan(tan(atan(tan(atan(tan(atan(a))))))))));
	}
	public static double tan(double a) {
		return Math.tan(a);
	}

	public static double atan(double a) {
		return Math.atan(a);
	}
}
