package de.mannheim.uni.expansion.tablefusion.core.similarity;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang3.mutable.MutableInt;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public abstract class AbstractSimilarity implements SimilarityMeasure {

	public static List<MutableInt> dbPediaTotalList = new ArrayList<>();
	public static List<MutableInt> dbPediaCountList = new ArrayList<>();
	public static List<MutableInt> tablesTotalList = new ArrayList<>();
	public static List<MutableInt> tablesCountList = new ArrayList<>();

	public static ThreadLocal<MutableInt> dbPediaTotal = new ThreadLocal<>();
	public static ThreadLocal<MutableInt> dbPediaCount = new ThreadLocal<>();
	public static ThreadLocal<MutableInt> tablesTotal = new ThreadLocal<>();
	public static ThreadLocal<MutableInt> tablesCount = new ThreadLocal<>();

	protected AbstractSimilarity() {
	}

	public static MutableInt getMutableInt(ThreadLocal<MutableInt> threadLocal, List<MutableInt> list) {
		MutableInt mi = threadLocal.get();
		if (mi == null) {
			mi = new MutableInt();
			threadLocal.set(mi);
			synchronized (list) {
				list.add(mi);
			}
		}
		return mi;
	}

	public static void incWebTables(int total) {
		getMutableInt(tablesTotal, tablesTotalList).add(total);
		getMutableInt(tablesCount, tablesCountList).increment();
	}

	public static void incDBPedia(int total) {
		getMutableInt(dbPediaTotal, dbPediaTotalList).add(total);
		getMutableInt(dbPediaCount, dbPediaCountList).increment();
	}

	public static void printListStatistics() {
		double dbPediaTotal = dbPediaTotalList.stream().map((e) -> {
			return e.getValue();
		}).mapToInt(Integer::intValue).sum();
		double dbPediaCount = dbPediaCountList.stream().map((e) -> {
			return e.getValue();
		}).mapToInt(Integer::intValue).sum();

		double tablesTotal = tablesTotalList.stream().map((e) -> {
			return e.getValue();
		}).mapToInt(Integer::intValue).sum();
		double tablesCount = tablesCountList.stream().map((e) -> {
			return e.getValue();
		}).mapToInt(Integer::intValue).sum();

		System.out.println("Average size of List for DBpedia: " + (dbPediaTotal / dbPediaCount));
		System.out.println("Lists in DBpedia: " + (dbPediaCount));
		System.out.println("Average size of List for WebTables: " + (tablesTotal / tablesCount));
		System.out.println("Lists in WebTables: " + (tablesCount));

		for (MutableInt e : dbPediaTotalList) {
			e.setValue(0);
		}
		for (MutableInt e : dbPediaCountList) {
			e.setValue(0);
		}
		for (MutableInt e : tablesTotalList) {
			e.setValue(0);
		}
		for (MutableInt e : tablesCountList) {
			e.setValue(0);
		}
	}

	public double get(AbstractObject a, AbstractObject b) {
		// if (a instanceof List<>) {
		// incDBPedia(((List<?>) a).size());
		// if (!(b instanceof List)) {
		// List<AbstractObject> bList = new ArrayList<AbstractObject>();
		// bList.add(b);
		// double sim = getListSimilarity((List<AbstractObject>) a, bList);
		// return sim;
		// } else {
		//
		// incWebTables(((List<?>) b).size());
		// double sim = getListSimilarity((List<AbstractObject>) a, (List<?>)
		// b);
		// return sim;
		// }
		// } else if (b instanceof List) {
		// incWebTables(((List<?>) b).size());
		// if (!(a instanceof List)) {
		// List<Object> aList = new ArrayList<Object>();
		// aList.add(a);
		// double sim = getListSimilarity(aList, (List<?>) b);
		// return sim;
		// } else {
		// incDBPedia(((List<?>) a).size());
		// double sim = getListSimilarity((List<?>) a, (List<?>) b);
		// return sim;
		// }
		// } else {
		return normalSimilarity(a, b);
		// }

	}

	public double get(ClassPropertyId classPropertyId, AbstractObject a, AbstractObject b) {

		return normalSimilarity(classPropertyId, a, b);

	}

	protected double normalSimilarity(ClassPropertyId classPropertyId, AbstractObject a, AbstractObject b) {
		return normalSimilarity(a, b);
	}

	// @Deprecated
	// private double getListSimilarity(List<AbstractObject> aList,
	// List<AbstractObject> bList) {
	// int smallerSize = 0;
	//
	// if (aList.size() < bList.size()) {
	// smallerSize = aList.size();
	// } else {
	// smallerSize = bList.size();
	// }
	//
	// int intersection = overlapSimilarity.getOverlap(aList, bList);
	//
	// double sim = ((double) intersection) / ((double) smallerSize);
	// return sim;
	// }

	abstract protected double normalSimilarity(AbstractObject a, AbstractObject b);

	abstract protected ValueType getColumnDataType();

	public boolean isEqual(AbstractObject a, AbstractObject b) {
		double sim = this.get(a, b);
		if (sim > this.getEquivalenceThreshold())
			return true;
		else
			return false;

	}

	public boolean isEqual(ClassPropertyId classPropertyId, AbstractObject a, AbstractObject b) {
		double sim = this.get(classPropertyId, a, b);
		if (sim > this.getEquivalenceThreshold())
			return true;
		else
			return false;

	}

}
