package de.mannheim.uni.expansion.tablefusion.core.similarity;

import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.tablefusion.core.similarity.list.OverlapSimilarity;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;

public class QualifiedSimilarityCalculator {

	private final SimilarityCalculator sc;
	private final OverlapSimilarity overlapSimilarity;

	@Inject
	public QualifiedSimilarityCalculator(SimilarityCalculator sc, OverlapSimilarity overlapSimilarity) {
		this.sc = sc;
		this.overlapSimilarity = overlapSimilarity;
	}

	public boolean sameValue(QualifiedValueHolder a, AbstractObject b) {
		return (sc.isEqual(a.getValue(), b));
	}

	public boolean sameValue(QualifiedValueHolder a, QualifiedValueHolder b) {
		return (sc.isEqual(a.getValue(), b.getValue()));
	}

	public boolean isCorrect(QualifiedValueList a, List<AbstractObject> b) {
		List<AbstractObject> correctList = a.getClaims().stream().map(QualifiedValueHolder::getValue).collect(Collectors.toList());

		return overlapSimilarity.hasCompleteOverlap(correctList, b);
	}

	public boolean isCorrect(QualifiedValueList a, AbstractObject b) {
		for (QualifiedValueHolder aQvh : a.getClaims()) {

			if (sc.isEqual(aQvh.getValue(), b)) {
				return true;
			}

		}
		return false;
	}

	public Optional<Boolean> isContainedAndCorrect(QualifiedValueList a, QualifiedValueHolder b) {

		boolean contained = false;
		for (QualifiedValueHolder aQvh : a.getClaims()) {
			if (sameQualifier(aQvh, b)) {
				contained = true;
				if (sc.isEqual(aQvh.getValue(), b.getValue())) {
					return Optional.of(Boolean.TRUE);
				}
			}
		}
		if (contained)
			return Optional.of(Boolean.FALSE);
		else
			return Optional.empty();
	}

	public boolean sameQualifier(QualifiedValueHolder a, QualifiedValueHolder b) {

		QualifiedValueHolder aQualified = (QualifiedValueHolder) a;
		QualifiedValueHolder bQualified = (QualifiedValueHolder) b;

		Set<PropertyId> qualifierIds = new HashSet<>(aQualified.getQualifierIds());
		qualifierIds.addAll(bQualified.getQualifierIds());

		for (PropertyId qualifierId : qualifierIds) {
			if (aQualified.hasQualifier(qualifierId) && bQualified.hasQualifier(qualifierId)) {
				AbstractObject qualifiersA = aQualified.getQualifierValue(qualifierId).getQualifierValue();
				AbstractObject qualifiersB = bQualified.getQualifierValue(qualifierId).getQualifierValue();

				if (!sc.isEqual(qualifiersA, qualifiersB)) {
					return false;
				}
			}
		}

		return true;
	}

}
