package de.mannheim.uni.expansion.tablefusion.core.similarity;

import java.util.EnumMap;
import java.util.Set;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.BooleanSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.CategoricalIntSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.CoordinateSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.DateSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.EmptySimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.LinkSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.NumericSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.ReferenceSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.StringSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.StringSimilarityJaccard;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.StringSimilarityMixed;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.matching.MatchingStringSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.matching.MatchingStringSimilarityHalf;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.matching.MatchingStringSimilarityWithSurfaceForms;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.matching.MatchingStringSimilarityWithSurfaceForms.SurfaceFormsGetterContainer;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.matching.MatchingStringSimilarityWithSurfaceFormsSimple;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class SimilarityCalculator {

	public final BooleanSimilarity booleanSimilarity;
	public final NumericSimilarity numericSimilarity;
	public final StringSimilarity stringSimilarity;
	public final StringSimilarityMixed stringSimilarityMixed;
	public final StringSimilarityJaccard stringSimilarityJaccard;
	public final EmptySimilarity emptySimilarity;
	public final DateSimilarity dateSimilarity;
	public final LinkSimilarity linkSimilarity;
	public final ReferenceSimilarity referenceSimilarity;
	public final CoordinateSimilarity coordinateSimilarity;

	// public final MatchingNumericSimilarity matchingNumericSimilarity;
	public final MatchingStringSimilarity matchingStringSimilarity;
	// public final MatchingDateSimilarity matchingDateSimilarity;

	public final MatchingStringSimilarityHalf matchingStringSimilarityHalf;

	private final EnumMap<ValueType, SimilarityMeasure> measures = new EnumMap<>(ValueType.class);

	public final MatchingStringSimilarityWithSurfaceForms matchingDateSimilarityWithSurfaceForms;
	@Inject
	protected SurfaceFormsGetterContainer surfaceForms;

	@Inject
	protected LabelsGetterContainer labelsGetter;

	public final MatchingStringSimilarityWithSurfaceFormsSimple matchingStringSimilarityWithSurfaceFormsSimple;

	@Inject
	public SimilarityCalculator(BooleanSimilarity booleanSimilarity, NumericSimilarity numericSimilarity, StringSimilarity stringSimilarity,
			EmptySimilarity emptySimilarity, DateSimilarity dateSimilarity, LinkSimilarity linkSimilarity, CoordinateSimilarity coordinateSimilarity,
			MatchingStringSimilarity matchingStringSimilarity, ReferenceSimilarity referenceSimilarity, StringSimilarityMixed stringSimilarityMixed,
			StringSimilarityJaccard stringSimilarityJaccard, MatchingStringSimilarityHalf matchingStringSimilarityHalf,
			MatchingStringSimilarityWithSurfaceForms matchingDateSimilarityWithSurfaceForms, CategoricalIntSimilarity categoricalIntSimilarity,
			MatchingStringSimilarityWithSurfaceFormsSimple matchingStringSimilarityWithSurfaceFormsSimple) {
		this.booleanSimilarity = booleanSimilarity;
		this.numericSimilarity = numericSimilarity;
		this.stringSimilarity = stringSimilarity;
		this.dateSimilarity = dateSimilarity;
		this.emptySimilarity = emptySimilarity;
		this.linkSimilarity = linkSimilarity;
		this.referenceSimilarity = referenceSimilarity;
		this.coordinateSimilarity = coordinateSimilarity;
		this.stringSimilarityMixed = stringSimilarityMixed;
		this.stringSimilarityJaccard = stringSimilarityJaccard;
		this.matchingDateSimilarityWithSurfaceForms = matchingDateSimilarityWithSurfaceForms;
		// this.matchingNumericSimilarity = matchingNumericSimilarity;
		this.matchingStringSimilarity = matchingStringSimilarity;
		// this.matchingDateSimilarity = matchingDateSimilarity;
		this.matchingStringSimilarityHalf = matchingStringSimilarityHalf;
		this.matchingStringSimilarityWithSurfaceFormsSimple = matchingStringSimilarityWithSurfaceFormsSimple;

		measures.put(ValueType.BOOLEAN, booleanSimilarity);
		measures.put(ValueType.COORDINATE, coordinateSimilarity);
		measures.put(ValueType.EXTERNAL_ID, referenceSimilarity);
		measures.put(ValueType.INSTANCE_REFERENCE, referenceSimilarity);
		measures.put(ValueType.PROPERTY_REFERENCE, referenceSimilarity);
		measures.put(ValueType.QUANTITY, numericSimilarity);
		measures.put(ValueType.STRING, matchingStringSimilarity);
		measures.put(ValueType.TIME, dateSimilarity);
		measures.put(ValueType.URL, linkSimilarity);
		measures.put(ValueType.CATEGORICAL_INT, categoricalIntSimilarity);
	}

	public SimilarityMeasure getMeasure(ValueType cdt) {
		if (cdt == null) {
			// TODO: FIX THIS SERIOUSLY
			return emptySimilarity;
		}
		SimilarityMeasure measure = measures.get(cdt);

		if (measure != null)
			return measure;
		// System.out.println(cdt.toString());

		return emptySimilarity;
	}

	public double getSimilarity(AbstractObject a, AbstractObject b) {
		if (a == null || b == null) {
			return 0.0;
		}
		if (a.getType() == b.getType()) {
			return getMeasure(a.getType()).get(a, b);

		} else if (a.getType() == ValueType.INSTANCE_REFERENCE && b.getType() == ValueType.STRING) {
			return getSimilarityReferences((InstanceId) a, (StringObject) b);
		} else if (b.getType() == ValueType.INSTANCE_REFERENCE && a.getType() == ValueType.STRING) {
			return getSimilarityReferences((InstanceId) b, (StringObject) a);
		} else if (a.getType() == ValueType.TIME && b.getType() == ValueType.QUANTITY) {
			return getMeasure(ValueType.TIME).get(a, b);
		} else if (a.getType() == ValueType.QUANTITY && b.getType() == ValueType.TIME) {
			return getMeasure(ValueType.TIME).get(a, b);
		}
		return 0.0;
	}

	public double getSimilarity(ClassPropertyId cpid, AbstractObject a, AbstractObject b) {
		if (a == null || b == null) {
			return 0.0;
		}
		if (a.getType() == b.getType()) {
			return getMeasure(a.getType()).get(cpid, a, b);
		} else if (a.getType() == ValueType.INSTANCE_REFERENCE && b.getType() == ValueType.STRING) {
			return getSimilarityReferences((InstanceId) a, (StringObject) b);
		} else if (b.getType() == ValueType.INSTANCE_REFERENCE && a.getType() == ValueType.STRING) {
			return getSimilarityReferences((InstanceId) b, (StringObject) a);
		} else if (a.getType() == ValueType.TIME && b.getType() == ValueType.QUANTITY) {
			return getMeasure(ValueType.TIME).get(a, b);
		} else if (a.getType() == ValueType.QUANTITY && b.getType() == ValueType.TIME) {
			return getMeasure(ValueType.TIME).get(a, b);
		}
		// else {
		// if(a.getType() == ValueType.STRING && b instanceof InstanceId) {
		// return getMeasure(a.getType()).isEqual(a, new
		// StringObject(((InstanceId) b).getReadableString()));
		// }
		// else if (b.getType() == ValueType.STRING && a instanceof InstanceId)
		// {
		// return getMeasure(b.getType()).isEqual(a, new
		// StringObject(((InstanceId) a).getReadableString()));
		// }
		// }
		return 0.0;
	}

	public double getSimilarity(ClassPropertyId classPropertyId, ValueType cdt, AbstractObject a, AbstractObject b) {

		SimilarityMeasure measure = null;

		if (cdt == ValueType.INSTANCE_REFERENCE) {
			if (a.getType() == ValueType.INSTANCE_REFERENCE && b.getType() == ValueType.STRING) {
				return getSimilarityReferences((InstanceId) a, (StringObject) b);
			} else if (b.getType() == ValueType.INSTANCE_REFERENCE && a.getType() == ValueType.STRING) {
				return getSimilarityReferences((InstanceId) b, (StringObject) a);
			} else if (a.getType() == ValueType.STRING && b.getType() == ValueType.STRING) {
				{
					measure = this.matchingDateSimilarityWithSurfaceForms;
				}
			}
		}

		if (measure == null) {
			measure = getMeasure(cdt);
		}

		return measure.get(classPropertyId, a, b);
	}

	public double getSimilarity(ValueType cdt, AbstractObject a, AbstractObject b) {

		SimilarityMeasure measure = null;

		if (cdt == ValueType.INSTANCE_REFERENCE) {
			if (a.getType() == ValueType.INSTANCE_REFERENCE && b.getType() == ValueType.STRING) {
				return getSimilarityReferences((InstanceId) a, (StringObject) b);
			} else if (b.getType() == ValueType.INSTANCE_REFERENCE && a.getType() == ValueType.STRING) {
				return getSimilarityReferences((InstanceId) b, (StringObject) a);
			} else if (a.getType() == ValueType.STRING && b.getType() == ValueType.STRING) {
				{
					measure = this.matchingDateSimilarityWithSurfaceForms;
				}
			}
		}

		if (measure == null) {
			measure = getMeasure(cdt);
		}

		return measure.get(a, b);
	}

	private double getSimilarityReferences(InstanceId instanceId, StringObject stringObject) {

		if (!labelsGetter.isPresent()) {
			StringObject so = new StringObject(instanceId.getReadableString());
			return stringSimilarity.get(so, stringObject);
		} else {

			// Set<String> setB = new HashSet<>();
			// setB.add(stringObject.getString());
			// Optional<Set<String>> sfB =
			// surfaceForms.get().getSurfaceForms(StringNormalizer.normaliseValue(stringObject.getString(),
			// false));
			// if (sfB.isPresent()) {
			// setB.addAll(sfB.get());
			// }

			Set<String> labels = labelsGetter.get().getLabels(instanceId);
			double maxSimilarity = 0.0;
			for (String label : labels) {
				StringObject so = new StringObject(label);

				double sim = stringSimilarity.get(so, stringObject);
				if (sim > maxSimilarity) {
					maxSimilarity = sim;
					if (maxSimilarity > 0.99) {
						return maxSimilarity;
					}
				}
			}
			return maxSimilarity;
			//
			// Set<String> setB = new HashSet<>();
			// setB.add(stringObject.getString());
			// Optional<Set<String>> sfB =
			// surfaceForms.get().getSurfaceForms(StringNormalizer.normaliseValue(stringObject.getString(),
			// false));
			// if (sfB.isPresent()) {
			// setB.addAll(sfB.get());
			// }
			//
			// Set<String> labels = labelsGetter.get().getLabels(instanceId);
			// double maxSimilarity = 0.0;
			// for (String label : labels) {
			// StringObject so = new StringObject(label);
			// for (String b : setB) {
			// StringObject soB = new StringObject(b);
			// double sim = stringSimilarity.get(so, soB);
			// if (sim > maxSimilarity) {
			// maxSimilarity = sim;
			// }
			// }
			// }
			// return maxSimilarity;
		}
	}

	private boolean isEqualReferences(InstanceId instanceId, StringObject stringObject) {

		if (!labelsGetter.isPresent()) {
			StringObject so = new StringObject(instanceId.getReadableString());
			return stringSimilarity.isEqual(so, stringObject);
		} else {
			if (stringObject.getString().equals("geffen") || stringObject.getString().toLowerCase().startsWith("alien")
					|| stringObject.getString().toLowerCase().startsWith("anthology") || stringObject.getString().toLowerCase().startsWith("anthology")) {
				int sdf = 123;
			}
			Set<String> labels = labelsGetter.get().getLabels(instanceId);
			for (String label : labels) {
				StringObject so = new StringObject(label);

				if (stringSimilarity.isEqual(so, stringObject)) {
					return true;
				}

			}
			return false;

			// Set<String> setB = new HashSet<>();
			// setB.add(stringObject.getString());
			// Optional<Set<String>> sfB =
			// surfaceForms.get().getSurfaceForms(StringNormalizer.normaliseValue(stringObject.getString(),
			// false));
			// if (sfB.isPresent()) {
			// setB.addAll(sfB.get());
			// }
			//
			// Set<String> labels = labelsGetter.get().getLabels(instanceId);
			// for (String label : labels) {
			// StringObject so = new StringObject(label);
			// for (String b : setB) {
			// StringObject soB = new StringObject(b);
			// if (stringSimilarity.isEqual(so, soB)) {
			// return true;
			// }
			// }
			// }
			// return false;
		}
	}

	public boolean isEqual(ClassPropertyId cpid, AbstractObject a, AbstractObject b) {
		if (a == null || b == null) {
			return false;
		}
		if (a.getType() == b.getType()) {
			return getMeasure(a.getType()).isEqual(cpid, a, b);
		} else if (a.getType() == ValueType.INSTANCE_REFERENCE && b.getType() == ValueType.STRING) {
			return isEqualReferences((InstanceId) a, (StringObject) b);
		} else if (b.getType() == ValueType.INSTANCE_REFERENCE && a.getType() == ValueType.STRING) {
			return isEqualReferences((InstanceId) b, (StringObject) a);
		} else if (a.getType() == ValueType.TIME && b.getType() == ValueType.QUANTITY) {
			return getMeasure(ValueType.TIME).isEqual(a, b);
		} else if (a.getType() == ValueType.QUANTITY && b.getType() == ValueType.TIME) {
			return getMeasure(ValueType.TIME).isEqual(a, b);
		}
		// else {
		// if(a.getType() == ValueType.STRING && b instanceof InstanceId) {
		// return getMeasure(a.getType()).isEqual(a, new
		// StringObject(((InstanceId) b).getReadableString()));
		// }
		// else if (b.getType() == ValueType.STRING && a instanceof InstanceId)
		// {
		// return getMeasure(b.getType()).isEqual(a, new
		// StringObject(((InstanceId) a).getReadableString()));
		// }
		// }
		return false;
	}

	public boolean isEqual(AbstractObject a, AbstractObject b) {
		if (a == null || b == null) {
			return false;
		}
		if (a.getType() == b.getType()) {
			return getMeasure(a.getType()).isEqual(a, b);
		} else if (a.getType() == ValueType.INSTANCE_REFERENCE && b.getType() == ValueType.STRING) {
			return isEqualReferences((InstanceId) a, (StringObject) b);
		} else if (b.getType() == ValueType.INSTANCE_REFERENCE && a.getType() == ValueType.STRING) {
			return isEqualReferences((InstanceId) b, (StringObject) a);
		} else if (a.getType() == ValueType.TIME && b.getType() == ValueType.QUANTITY) {
			return getMeasure(ValueType.TIME).isEqual(a, b);
		} else if (a.getType() == ValueType.QUANTITY && b.getType() == ValueType.TIME) {
			return getMeasure(ValueType.TIME).isEqual(a, b);
		}
		// else {
		// if(a.getType() == ValueType.STRING && b instanceof InstanceId) {
		// return getMeasure(a.getType()).isEqual(a, new
		// StringObject(((InstanceId) b).getReadableString()));
		// }
		// else if (b.getType() == ValueType.STRING && a instanceof InstanceId)
		// {
		// return getMeasure(b.getType()).isEqual(a, new
		// StringObject(((InstanceId) a).getReadableString()));
		// }
		// }
		return false;
	}

	public boolean isEqual(ValueType cdt, AbstractObject a, AbstractObject b) {

		SimilarityMeasure measure = null;

		if (cdt == ValueType.INSTANCE_REFERENCE) {
			if (a.getType() == ValueType.INSTANCE_REFERENCE && b.getType() == ValueType.STRING) {
				return isEqualReferences((InstanceId) a, (StringObject) b);
			} else if (b.getType() == ValueType.INSTANCE_REFERENCE && a.getType() == ValueType.STRING) {
				return isEqualReferences((InstanceId) b, (StringObject) a);
			} else if (a.getType() == ValueType.STRING && b.getType() == ValueType.STRING) {
				{
					measure = this.matchingDateSimilarityWithSurfaceForms;
				}
			}
		}

		if (measure == null) {
			measure = getMeasure(cdt);
		}

		return measure.isEqual(a, b);
	}

	public boolean isEqual(ClassPropertyId classPropertyId, ValueType cdt, AbstractObject a, AbstractObject b) {

		SimilarityMeasure measure = null;

		if (cdt == ValueType.INSTANCE_REFERENCE) {
			if (a.getType() == ValueType.INSTANCE_REFERENCE && b.getType() == ValueType.STRING) {
				return isEqualReferences((InstanceId) a, (StringObject) b);
			} else if (b.getType() == ValueType.INSTANCE_REFERENCE && a.getType() == ValueType.STRING) {
				return isEqualReferences((InstanceId) b, (StringObject) a);
			} else if (a.getType() == ValueType.STRING && b.getType() == ValueType.STRING) {
				{
					measure = this.matchingDateSimilarityWithSurfaceForms;
				}
			}
		}

		if (measure == null) {
			measure = getMeasure(cdt);
		}

		return measure.isEqual(classPropertyId, a, b);
	}

	public double getEquivalenceThreshold(ValueType cdt) {
		return getMeasure(cdt).getEquivalenceThreshold();
	}

	public SimilarityMeasure getLabelBlockingSimilarity() {
		return matchingStringSimilarity;
	}

	public SimilarityMeasure getStrictSimilarity() {
		return stringSimilarityJaccard;
	}

	@FunctionalInterface
	public static interface LabelsGetter {
		public Set<String> getLabels(InstanceId instanceId);
	}

	public static class LabelsGetterContainer {
		private LabelsGetter labelsGetter;

		public boolean isPresent() {
			if (this.labelsGetter == null)
				return false;
			else
				return true;
		}

		public void setLabelsGetter(LabelsGetter labelsGetter) {
			this.labelsGetter = labelsGetter;
		}

		public LabelsGetter get() {
			return labelsGetter;
		}
	}
}
