package de.mannheim.uni.expansion.tablefusion.core.similarity.list;

import java.util.ArrayList;
import java.util.Deque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityMeasure;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MapSorter;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;

public class OverlapSimilarity {

	private final SimilarityCalculator sc;

	@Inject
	public OverlapSimilarity(SimilarityCalculator sc) {
		this.sc = sc;
	}

	Supplier<TreeMap<Integer, Integer>> supplier = () -> {
		return new TreeMap<>();
	};

	public int getOverlap(List<AbstractObject> aList, List<AbstractObject> bList) {

		Map<Integer, Set<Integer>> aToBmatrixOfEquality = new HashMap<>();

		Map<Integer, Set<Integer>> bToAmatrixOfEquality = new HashMap<>();

		for (int a = 0; a < aList.size(); a++) {
			Set<Integer> equalitySet = new HashSet<>();
			for (int b = 0; b < bList.size(); b++) {
				if (sc.isEqual(aList.get(a), bList.get(b))) {
					equalitySet.add(b);

					CHMHelper.getFromMap(bToAmatrixOfEquality, b, () -> new HashSet<Integer>()).add(a);
				}
			}
			aToBmatrixOfEquality.put(a, new HashSet<>());
		}

		int aIntersection = 0;
		int bIntersection = 0;

		for (int a = 0; a < aList.size(); a++) {

			if (aToBmatrixOfEquality.containsKey(a) && aToBmatrixOfEquality.get(a).size() > 0) {
				aIntersection++;
			}

		}

		for (int b = 0; b < bList.size(); b++) {

			if (bToAmatrixOfEquality.containsKey(b) && bToAmatrixOfEquality.get(b).size() > 0) {
				bIntersection++;
			}

		}

		int intersection = Math.max(aIntersection, bIntersection);
		int smallerList = Math.min(aList.size(), bList.size());
		if (intersection > smallerList) {
			return smallerList;
		}
		return intersection;
	}

	public int getOverlapUsingHash(Set<AbstractObject> aList, Set<AbstractObject> bList) {
		Set<AbstractObject> smallerSet;
		Set<AbstractObject> largerSet;
		if (aList.size() < bList.size()) {
			smallerSet = aList;
			largerSet = bList;
		} else {
			smallerSet = bList;
			largerSet = aList;
		}
		int intersection = 0;
		for (AbstractObject ao : smallerSet) {
			if (largerSet.contains(ao)) {
				intersection++;
			}
		}

		return intersection;
	}

	private int getOverlapOld(List<AbstractObject> aList, List<AbstractObject> bList) {
		int intersection = 0;

		List<AbstractObject> smallerList;
		List<AbstractObject> biggerList;
		if (aList.size() < bList.size()) {
			smallerList = aList;
			biggerList = bList;
		} else {
			smallerList = bList;
			biggerList = aList;
		}

		Map<Integer, Set<Integer>> aToBmatrixOfEquality = new HashMap<>();

		Map<Integer, Set<Integer>> bToAmatrixOfEquality = new HashMap<>();

		for (int a = 0; a < smallerList.size(); a++) {
			Set<Integer> equalitySet = new HashSet<>();
			for (int b = 0; b < biggerList.size(); b++) {
				if (sc.isEqual(aList.get(a), bList.get(b))) {
					equalitySet.add(b);

					CHMHelper.getFromMap(bToAmatrixOfEquality, b, () -> new HashSet<Integer>()).add(a);
				}
			}
			aToBmatrixOfEquality.put(a, new HashSet<>());
		}

		Set<Integer> aIndeces = new HashSet<>();
		for (int a = 0; a < aList.size(); a++) {
			aIndeces.add(a);
		}

		Set<Integer> bIndeces = new HashSet<>();
		for (int b = 0; b < bList.size(); b++) {
			bIndeces.add(b);
		}

		double max = 0.0;
		int aMax = -1;
		int bMax = -1;

		SortedMap<Long, Deque<Pair<Integer, Integer>>> sortedQueueMap = new TreeMap<>();
		// Deque<Pair<Integer, Integer>> stack = new LinkedList<>();

		max = 0.0;
		aMax = -1;
		bMax = -1;

		for (Iterator<Integer> aIt = aIndeces.iterator(); aIt.hasNext();) {
			int a = aIt.next();

			Set<Integer> equals = aToBmatrixOfEquality.get(a).stream().filter((e) -> {
				return bIndeces.contains(e);
			}).collect(Collectors.toSet());

			if (equals.size() == 1) {
				Integer b = aToBmatrixOfEquality.get(a).iterator().next();
				if (bIndeces.contains(b)) {
					aIt.remove();
					bIndeces.remove(b);
				}
				intersection++;
			} else if (equals.size() > 1) {

				LinkedHashMap<Integer, Long> sortedMap = MapSorter
						.sortHashMapByValuesD((Map<Integer, Long>) equals.stream().collect(Collectors.toMap(Integer::intValue, (e) -> {
							return bToAmatrixOfEquality.get(e).stream().filter((aId) -> {
								return aIndeces.contains(aId);
							}).count();
						})));

				for (Entry<Integer, Long> e : sortedMap.entrySet()) {
					CHMHelper.getFromMap(sortedQueueMap, e.getValue(), () -> new LinkedList<>()).add(new Pair<>(a, e.getKey()));
				}
			}

		}

		for (Entry<Long, Deque<Pair<Integer, Integer>>> e : sortedQueueMap.entrySet()) {
			Pair<Integer, Integer> pair = null;
			while ((pair = e.getValue().pop()) != null) {
				if (aIndeces.contains(pair.getFirst())) {
					if (bIndeces.contains(pair.getSecond())) {
						intersection++;
						aIndeces.remove(pair.getFirst());
						bIndeces.remove(pair.getSecond());

					}
				}
			}
		}

		return intersection;
	}

	private int getOverlap(List<AbstractObject> aList, List<AbstractObject> bList, SimilarityMeasure measure) {
		int intersection = 0;
		double et = measure.getEquivalenceThreshold();

		Double[][] matrix = new Double[aList.size()][bList.size()];

		for (int a = 0; a < aList.size(); a++) {
			for (int b = 0; b < bList.size(); b++) {
				matrix[a][b] = measure.getQuickSimilarity(aList.get(a), bList.get(b));
			}
		}

		Set<Integer> aIndeces = new HashSet<>();
		for (int a = 0; a < aList.size(); a++) {
			aIndeces.add(a);
		}

		Set<Integer> bIndeces = new HashSet<>();
		for (int b = 0; b < bList.size(); b++) {
			bIndeces.add(b);
		}

		double max = 0.0;
		int aMax = -1;
		int bMax = -1;

		outerloop: while (true) {

			max = 0.0;
			aMax = -1;
			bMax = -1;

			for (Iterator<Integer> aIt = aIndeces.iterator(); aIt.hasNext();) {
				int a = aIt.next();

				for (Iterator<Integer> bIt = bIndeces.iterator(); bIt.hasNext();) {
					int b = bIt.next();

					Double sim = matrix[a][b];
					if (sim > 0.999999) {
						intersection++;
						aIt.remove();
						bIt.remove();
						continue outerloop;
					} else if (sim > measure.getEquivalenceThreshold() && sim > max) {
						max = sim;
						aMax = a;
						bMax = b;
					}

				}

			}

			if (max < measure.getEquivalenceThreshold()) {
				break;
			} else {
				intersection++;
				aIndeces.remove(aMax);
				bIndeces.remove(bMax);
			}

		}
		return intersection;
	}

	private double getSimilarity(List<AbstractObject> aList, List<AbstractObject> bList, SimilarityMeasure measure) {
		int smallerSize = 0;

		if (aList.size() < bList.size()) {
			smallerSize = aList.size();
		} else {
			smallerSize = bList.size();
		}

		int intersection = getOverlap(aList, bList, measure);

		double sim = ((double) intersection) / ((double) smallerSize);
		return sim;
	}

	private boolean hasCompleteOverlap(List<AbstractObject> aList, List<AbstractObject> bList, SimilarityMeasure measure) {
		int smallerSize = 0;

		if (aList.size() < bList.size()) {
			smallerSize = aList.size();
		} else {
			smallerSize = bList.size();
		}

		int intersection = getOverlap(aList, bList, measure);
		if (intersection == smallerSize) {
			return true;
		} else {
			return false;
		}

	}

	private boolean isIdentical(List<AbstractObject> aList, List<AbstractObject> bList, SimilarityMeasure measure) {
		int size = 0;

		if (aList.size() == bList.size()) {
			size = aList.size();

			int intersection = getOverlap(aList, bList, measure);
			if (intersection == size) {
				return true;
			} else {
				return false;
			}

		} else {
			return false;
		}

	}

	private boolean hasOverlap(List<AbstractObject> aList, List<AbstractObject> bList, SimilarityMeasure measure, int size) {
		int intersection = getOverlap(aList, bList, measure);
		if (intersection == size)
			return true;
		else
			return false;
	}

	public boolean hasOverlap(List<AbstractObject> aList, List<AbstractObject> bList, int size) {
		int intersection = getOverlap(aList, bList);
		if (intersection == size)
			return true;
		else
			return false;
	}

	public double getMAP(List<AbstractObject> correctList, List<Set<AbstractObject>> rankedList) {

		List<Double> averagePrecisions = new ArrayList<Double>();
		double precisionSum = 0.0;

		double negativeCounter = 0.0;

		double positiveCounter = 0.0;

		// https://www.kaggle.com/wiki/MeanAveragePrecision
		for (Set<AbstractObject> tr : rankedList) {
			double averagePrecision;

			for (AbstractObject o : tr) {
				if (isCorrect(correctList, o)) {
					positiveCounter++;
				} else {
					negativeCounter++;

				}
			}

			double precisionAtItem = positiveCounter / (negativeCounter + positiveCounter);
			precisionSum += precisionAtItem;

			averagePrecision = precisionSum / (averagePrecisions.size() + 1);

			averagePrecisions.add(averagePrecision);
		}

		// TODO Auto-generated method stub
		Double precision = 0.0;
		for (double p : averagePrecisions) {
			precision += p;
		}
		precision = precision / averagePrecisions.size();

		if (precision.equals(Double.NaN)) {
			return 0.0;
		} else {
			return precision;
		}
	}

	private double getMAP(List<AbstractObject> correctList, List<Set<AbstractObject>> rankedList, SimilarityMeasure measure) {

		List<Double> averagePrecisions = new ArrayList<Double>();
		double precisionSum = 0.0;

		double negativeCounter = 0.0;

		double positiveCounter = 0.0;

		// https://www.kaggle.com/wiki/MeanAveragePrecision
		for (Set<AbstractObject> tr : rankedList) {
			double averagePrecision;

			for (AbstractObject o : tr) {
				if (isCorrect(correctList, o, measure)) {
					positiveCounter++;
				} else {
					negativeCounter++;

				}
			}

			double precisionAtItem = positiveCounter / (negativeCounter + positiveCounter);
			precisionSum += precisionAtItem;

			averagePrecision = precisionSum / (averagePrecisions.size() + 1);

			averagePrecisions.add(averagePrecision);
		}

		// TODO Auto-generated method stub
		Double precision = 0.0;
		for (double p : averagePrecisions) {
			precision += p;
		}
		precision = precision / averagePrecisions.size();

		if (precision.equals(Double.NaN)) {
			return 0.0;
		} else {
			return precision;
		}
	}

	private boolean isCorrect(List<AbstractObject> correctList, AbstractObject object, SimilarityMeasure measure) {

		for (AbstractObject correct : correctList) {
			if (measure.isEqual(correct, object)) {
				return true;
			}

		}
		return false;

	}

	private boolean isCorrect(List<AbstractObject> correctList, AbstractObject object) {

		for (AbstractObject correct : correctList) {
			if (sc.isEqual(correct, object)) {
				return true;
			}

		}
		return false;

	}

	public boolean hasCompleteOverlap(List<AbstractObject> aList, List<AbstractObject> bList) {
		int smallerSize = 0;

		if (aList.size() < bList.size()) {
			smallerSize = aList.size();
		} else {
			smallerSize = bList.size();
		}

		int intersection = getOverlap(aList, bList);
		if (intersection == smallerSize) {
			return true;
		} else {
			return false;
		}

	}
}
