package de.mannheim.uni.expansion.tablefusion.core.similarity.measures;

import javax.inject.Inject;

import de.mannheim.uni.expansion.tablefusion.core.similarity.AbstractSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityMeasures;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThreshold;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThresholdsRepository;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.CoordinateObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class CoordinateSimilarity extends AbstractSimilarity {

	private final SimilarityThreshold similarityThreshold;
	private final NumericSimilarity numericSimilarity;

	@Inject
	public CoordinateSimilarity(SimilarityThresholdsRepository similarityThresholdsRepository, NumericSimilarity numericSimilarity) {
		this.similarityThreshold = similarityThresholdsRepository.get(SimilarityMeasures.COORDINATE);
		this.numericSimilarity = numericSimilarity;
	}

	public CoordinateSimilarity(SimilarityThreshold similarityThreshold, NumericSimilarity numericSimilarity) {
		this.similarityThreshold = similarityThreshold;
		this.numericSimilarity = numericSimilarity;
	}

	@Override
	public double getQuickSimilarity(AbstractObject a, AbstractObject b) {
		return normalSimilarity(a, b);
	}

	@Override
	public double getEquivalenceThreshold() {
		return similarityThreshold.get();
	}

	@Override
	protected double normalSimilarity(AbstractObject a, AbstractObject b) {
		if (a == null || b == null) {
			return 0.0;
		} else if (!(a instanceof CoordinateObject) || !(b instanceof CoordinateObject)) {
			return 0.0;
		} else {
			CoordinateObject ao = ((CoordinateObject) a);
			CoordinateObject bo = ((CoordinateObject) b);

			double sim1 = numericSimilarity.get(ao.getLatitude(), bo.getLatitude());
			double sim2 = numericSimilarity.get(ao.getLongitude(), bo.getLongitude());

			return (sim1 + sim2) / 2.0;

		}
	}

	@Override
	protected ValueType getColumnDataType() {
		return ValueType.COORDINATE;
	}

}
