package de.mannheim.uni.expansion.tablefusion.core.similarity.measures.matching;

import javax.inject.Inject;

import de.mannheim.uni.expansion.tablefusion.core.similarity.AbstractSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityMeasures;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThreshold;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThresholdImpl;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThresholdsRepository;
import de.mannheim.uni.similarity.functions.string.GeneralisedStringJaccard;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class MatchingStringSimilarity extends AbstractSimilarity {

	private final GeneralisedStringJaccard sim;

	private final SimilarityThreshold similarityThreshold;

	@Inject
	public MatchingStringSimilarity(SimilarityThresholdsRepository similarityThresholdsRepository) {
		sim = new GeneralisedStringJaccard(new LevelSteinSimilarityWithCache(), 0.5, 0.0);
		this.similarityThreshold = similarityThresholdsRepository.get(SimilarityMeasures.MATCHING_STRING);
	}

	public MatchingStringSimilarity(double outer, double inner) {
		sim = new GeneralisedStringJaccard(new LevelSteinSimilarityWithCache(), inner, outer);
		this.similarityThreshold = new SimilarityThresholdImpl(outer);
	}

	@Override
	public double getQuickSimilarity(AbstractObject a, AbstractObject b) {
		return normalSimilarity(a, b);
	}

	@Override
	protected double normalSimilarity(AbstractObject a, AbstractObject b) {

		if (a == null || b == null) {
			return 0.0;
		} else if (a.equals(b)) {
			return 1.0;
		} else if (!(a instanceof StringObject) || !(b instanceof StringObject)) {
			return 0.0;
		} else {
			String a2 = ((StringObject) a).getString();
			String b2 = ((StringObject) b).getString();

			return sim.calculate(a2, b2);

		}

	}

	@Override
	protected ValueType getColumnDataType() {
		return ValueType.STRING;
	}

	@Override
	public double getEquivalenceThreshold() {
		return similarityThreshold.get();
	}

}
