package de.mannheim.uni.expansion.tablefusion.core.similarity.measures.matching;

import java.util.HashSet;
import java.util.Set;

import javax.inject.Inject;

import com.google.common.base.Optional;

import de.mannheim.uni.expansion.tablefusion.core.similarity.AbstractSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.matching.MatchingStringSimilarityWithSurfaceForms.SurfaceFormsGetterContainer;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThreshold;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThresholdImpl;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThresholdsRepository;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;

public class MatchingStringSimilarityWithSurfaceFormsSimple extends AbstractSimilarity {

	// private final GeneralisedStringJaccard sim;

	private final SimilarityThreshold similarityThreshold;

	private final SurfaceFormsGetterContainer surfaceForms;

	@Inject
	public MatchingStringSimilarityWithSurfaceFormsSimple(SimilarityThresholdsRepository similarityThresholdsRepository,
			SurfaceFormsGetterContainer surfaceForms) {
		// sim = new GeneralisedStringJaccard(new
		// LevelSteinSimilarityWithCache(), 0.5, 0.0);
		this.similarityThreshold = new SimilarityThresholdImpl(0.99);
		this.surfaceForms = surfaceForms;
	}

	// public MatchingStringSimilarityWithSurfaceForms(double outer, double
	// inner) {
	// sim = new GeneralisedStringJaccard(new LevelSteinSimilarityWithCache(),
	// inner, outer);
	// this.similarityThreshold = new SimilarityThresholdImpl(outer);
	// }

	@Override
	public double getQuickSimilarity(AbstractObject a, AbstractObject b) {
		return normalSimilarity(a, b);
	}

	@Override
	protected double normalSimilarity(AbstractObject a, AbstractObject b) {

		if (a == null || b == null) {
			return 0.0;
		} else if (a.equals(b)) {
			return 1.0;
		} else if (!(a instanceof StringObject) || !(b instanceof StringObject)) {
			return 0.0;
		} else {
			String a2 = ((StringObject) a).getString();
			String b2 = ((StringObject) b).getString();
			if (a2.equals(b2)) {
				return 1.0;
			}

			if (false && surfaceForms.candidatesIsPresent()) {

				boolean atLeastOneOverlap = false;
				Optional<Set<String>> candidatesAOptional = surfaceForms.getCandidatesSurfaceForms().getSurfaceForms(a2);
				if (candidatesAOptional.isPresent() && candidatesAOptional.get().contains(b2)) {
					atLeastOneOverlap = true;
				}

				if (!atLeastOneOverlap) {
					Optional<Set<String>> candidatesBOptional = surfaceForms.getCandidatesSurfaceForms().getSurfaceForms(b2);
					if (candidatesBOptional.isPresent() && candidatesBOptional.get().contains(a2)) {
						atLeastOneOverlap = true;
					}

				}

				if (!atLeastOneOverlap) {
					return 0.0;
				}
			}

			Set<String> setA;
			Optional<Set<String>> sfA = surfaceForms.get().getSurfaceForms(StringNormalizer.normaliseValue(a2, false));
			if (sfA.isPresent()) {
				setA = sfA.get();
			} else {
				setA = new HashSet<>();
				setA.add(a2);
			}

			Set<String> setB;
			Optional<Set<String>> sfB = surfaceForms.get().getSurfaceForms(StringNormalizer.normaliseValue(b2, false));
			if (sfB.isPresent()) {
				setB = sfB.get();
			} else {
				setB = new HashSet<>();
				setB.add(b2);
			}

			double max = 0.0;

			if (setA.size() < setB.size()) {

				if (compare(setA, b2))
					return 1.0;

			} else {
				if (compare(setA, b2))
					return 1.0;
			}

			return 0.0;
		}
	}

	// private boolean atLeastOneOverlap(Set<String> candidatesA, Set<String>
	// candidatesB) {
	// for (String sA : candidatesA) {
	// if (candidatesB.contains(sA))
	// return true;
	// }
	// return false;
	// }
	//
	// double compare(Set<String> setA, Set<String> setB) {
	// double max = 0.0;
	// for (String aString : setA) {
	// if (setB.contains(aString)) {
	// max = 1.0;
	// break;
	// } else {
	// for (String bString : setB) {
	// double similarity = 0.0;
	//
	// similarity = sim.calculate(aString, bString);
	//
	// if (similarity > max) {
	// max = similarity;
	//
	// if (similarity > 0.99) {
	// break;
	// }
	// }
	// }
	// }
	// }
	// return max;
	// }

	boolean compare(Set<String> setA, String b) {

		if (setA.contains(b)) {
			return true;
		}

		return false;
	}

	@Override
	protected ValueType getColumnDataType() {
		return ValueType.STRING;
	}

	@Override
	public double getEquivalenceThreshold() {
		return similarityThreshold.get();
	}

}
