package de.mannheim.uni.expansion.tablefusion.modules;

import java.util.Set;

import javax.inject.Singleton;

import dagger.Module;
import dagger.Provides;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.uni_mannheim.dws.expansion.data.kb.abstracts.AbstractsAdapter;
import de.uni_mannheim.dws.expansion.data.kb.abstracts.AbstractsRepository;
import de.uni_mannheim.dws.expansion.data.kb.abstracts.AbstractsRepositoryImpl;

@Module(library = true, injects = { AbstractsRepository.class })

public class AbstractsModule {
	private final String abstractsFolder;
	private final String cacheFolder;
	private Set<InstanceId> instanceIds;

	public AbstractsModule(String abstractsFolder, String cacheFolder) {
		this.abstractsFolder = abstractsFolder;
		this.cacheFolder = cacheFolder;
	}

	public static boolean CACHE_ABSTRACTS_STATICALLY = false;
	private static AbstractsRepository abstractsCache = null;

	@Provides
	@Singleton
	public AbstractsRepository getKeyIndex(AbstractsAdapter abstractsAdapter) {

		if (CACHE_ABSTRACTS_STATICALLY && abstractsCache != null) {
			return abstractsCache;
		} else {
			if (abstractsAdapter.cacheExists(cacheFolder)) {
				AbstractsRepositoryImpl repository = abstractsAdapter.readFromCache(cacheFolder);

				if (CACHE_ABSTRACTS_STATICALLY) {
					abstractsCache = repository;
				}
				return repository;
			} else {
				AbstractsRepositoryImpl repository = abstractsAdapter.readFromSource(abstractsFolder, instanceIds);
				abstractsAdapter.writeToCache(repository, cacheFolder);
				if (CACHE_ABSTRACTS_STATICALLY) {
					abstractsCache = repository;
				}

				return repository;
			}
		}


	}

	public void setInstanceIds(Set<InstanceId> instanceIds) {
		this.instanceIds = instanceIds;
	}
}
