package de.mannheim.uni.expansion.tablefusion.modules;

import javax.inject.Singleton;

import dagger.Module;
import dagger.Provides;
import de.uni_mannheim.dws.expansion.data.kb.degree.DegreeAdapter;
import de.uni_mannheim.dws.expansion.data.kb.degree.DegreeRepository;
import de.uni_mannheim.dws.expansion.data.kb.degree.DegreeRepositoryImpl;

@Module(library = true, injects = { DegreeRepository.class })

public class DegreeModule {
	private final String propertiesFile;
	private final String cacheFolder;

	public DegreeModule(String propertiesFile, String cacheFolder) {
		this.propertiesFile = propertiesFile;
		this.cacheFolder = cacheFolder;
	}

	public static boolean CACHE_DEGREE_REPOSITORY_STATICALLY = false;
	private static DegreeRepositoryImpl cached = null;

	@Provides
	@Singleton
	public DegreeRepository getKeyIndex(DegreeAdapter degreeAdapter) {
		if (CACHE_DEGREE_REPOSITORY_STATICALLY && cached != null) {
			return cached;
		}
		else {
			if (degreeAdapter.cacheExists(cacheFolder)) {
				DegreeRepositoryImpl repository = degreeAdapter.readFromCache(cacheFolder);

				if (CACHE_DEGREE_REPOSITORY_STATICALLY) {
					cached = repository;
				}
				return repository;
			} else {
				DegreeRepositoryImpl repository = degreeAdapter.readFromSource(propertiesFile);
				if (CACHE_DEGREE_REPOSITORY_STATICALLY) {
					cached = repository;
				}
				degreeAdapter.writeToCache(repository, cacheFolder);

				return repository;
			}

		}


	}
}
