package de.mannheim.uni.expansion.tablefusion.modules;

import java.io.FileNotFoundException;
import java.util.Collection;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

import javax.inject.Singleton;

import dagger.Module;
import dagger.Provides;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator.LabelsGetter;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator.LabelsGetterContainer;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.matching.MatchingStringSimilarityWithSurfaceForms.SurfaceFormsGetter;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.matching.MatchingStringSimilarityWithSurfaceForms.SurfaceFormsGetterContainer;
import de.mannheim.uni.expansion.tablefusion.core.similarity.numericRange.NumericRangeRepository;
import de.mannheim.uni.expansion.tablefusion.core.similarity.stopwords.StopWordsAdapter;
import de.mannheim.uni.expansion.tablefusion.core.similarity.stopwords.StopWordsRepository;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityMeasures;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThresholdsRepository;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThresholdsRepositoryImpl;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;

@Module(library = true)
public class SimilarityModule {

	private LabelsGetterContainer labelsGetterContainer = new LabelsGetterContainer();
	private SurfaceFormsGetterContainer surfaceFormsContainer = new SurfaceFormsGetterContainer();

	private Map<ClassPropertyId, Collection<AbstractObject>> numericRangesValues = new HashMap<>();

	private final String stopWordsFile;

	public SimilarityModule() {
		this.stopWordsFile = null;
	}

	public SimilarityModule(String stopWordsFile) {
		this.stopWordsFile = stopWordsFile;
	}


	public static boolean CACHE_NUMERIC_RANGE_STATICALLY = false;
	private static NumericRangeRepository numericRangeCache = null;

	public boolean numericRangeCacheExists() {
		return CACHE_NUMERIC_RANGE_STATICALLY && numericRangeCache != null;
	}

	public void setNumericRangesValues(Map<ClassPropertyId, Collection<AbstractObject>> numericRangesValues) {
		this.numericRangesValues = numericRangesValues;
	}

	@Provides
	@Singleton
	public NumericRangeRepository getNumericRangeRepository() {
		if (CACHE_NUMERIC_RANGE_STATICALLY && numericRangeCache != null) {
			return numericRangeCache;
		} else {
			NumericRangeRepository numericRangeRepository = new NumericRangeRepository(numericRangesValues);
			if (CACHE_NUMERIC_RANGE_STATICALLY) {
				numericRangeCache = numericRangeRepository;
			}
			return numericRangeRepository;
		}
	}

	@Provides
	@Singleton
	public LabelsGetterContainer getLabelsGetterContainer() {
		return labelsGetterContainer;
	}

	@Provides
	@Singleton
	public StopWordsRepository getStopWordsRepository(StopWordsAdapter adapter) {
		if (stopWordsFile != null) {
			try {
				return adapter.read(stopWordsFile);
			} catch (FileNotFoundException e) {
				throw new RuntimeException(e);
			}
		} else {
			return new StopWordsRepository(new HashSet<>());
		}
	}

	@Provides
	@Singleton
	public SurfaceFormsGetterContainer getSurfaceFormsGetterContainer() {
		return surfaceFormsContainer;
	}

	public void setLabelsGetter(LabelsGetter labelsGetter) {
		this.labelsGetterContainer.setLabelsGetter(labelsGetter);
	}

	public void setSurfaceFormsGetter(SurfaceFormsGetter surfaceFormsGetter) {
		this.surfaceFormsContainer.setSurfaceFormsGetter(surfaceFormsGetter);
	}

	public void setCandidatesSurfaceFormsGetter(SurfaceFormsGetter surfaceFormsGetter) {
		this.surfaceFormsContainer.setCandidatesSurfaceFroms(surfaceFormsGetter);
	}

	@Provides
	@Singleton
	public SimilarityThresholdsRepository getSimilarityThresholdsRepository() {

		Map<SimilarityMeasures, Double> map = new EnumMap<>(SimilarityMeasures.class);

		for (SimilarityMeasures sm : SimilarityMeasures.values()) {
			switch (sm) {
			case STRING:
				map.put(sm, 0.76);
				break;
			case NUMERIC:
				map.put(sm, 0.97);
				break;
			case DATE:
				map.put(sm, 0.20);
				break;
			case MATCHING_STRING:
				map.put(sm, 0.75);
				break;
			case MATCHING_STRING_HALF:
				map.put(sm, 0.60);
				break;
			case MATCHING_NUMERIC:
				map.put(sm, 0.23);
				break;
			case MATCHING_DATE:
				map.put(sm, 0.99999999999);
				break;
			case STRING_MIXED:
				map.put(sm, 0.83);
				break;
			case STRING_JACCARD:
				map.put(sm, 0.99);
				break;
			case COORDINATE:
				map.put(sm, 0.98);
				break;

			}
		}

		return new SimilarityThresholdsRepositoryImpl(map);
	}
}
