package de.mannheim.uni.expansion.tablefusion.adapters.learningsets;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.zip.GZIPInputStream;

import javax.inject.Inject;
import javax.inject.Named;

import org.objenesis.strategy.StdInstantiatorStrategy;

import com.esotericsoftware.kryo.Kryo;
import com.esotericsoftware.kryo.Kryo.DefaultInstantiatorStrategy;
import com.esotericsoftware.kryo.io.Input;
import com.esotericsoftware.kryo.util.MapReferenceResolver;

import de.mannheim.uni.expansion.data.tables.reader.FixedClassResolver;
import de.mannheim.uni.expansion.tablefusion.NamedProperties;
import de.mannheim.uni.expansion.tablefusion.learning.LearningSetCreator.CreatedSets;
import de.mannheim.uni.expansion.utils.PersistentParallel;

public class LearningSetsReader {

	private final ThreadLocal<Kryo> kryoLocal;
	private final String outputFolder;

	@Inject
	public LearningSetsReader(@Named(NamedProperties.LEARNING_SETS_OUTPUT_FOLDER) String outputFolder) {
		kryoLocal = new ThreadLocal<>();
		this.outputFolder = outputFolder;
	}

	public Map<Integer, CreatedSets> read() {
		if (outputFolder == null)
			return null;

		List<File> files = Arrays.asList(new File(outputFolder).listFiles());
		if (files.size() == 0) {
			return null;
		}

		Map<Integer, CreatedSets> createdSetsMap = new HashMap<>();

		try {
			new PersistentParallel<File>("Writing CreatedSets").foreach(files, (item) -> {
				Kryo kryo = kryoLocal.get();
				if (kryo == null) {
					kryo = new Kryo(new FixedClassResolver(), new MapReferenceResolver());
					kryo.setInstantiatorStrategy(new DefaultInstantiatorStrategy(new StdInstantiatorStrategy()));
					kryoLocal.set(kryo);
				}

				InputStream inputStream;
				try {
					inputStream = new GZIPInputStream(new FileInputStream(item));

					Input input = new Input(inputStream);
					CreatedSets createdSets = kryo.readObject(input, CreatedSets.class);
					input.close();

					int fileNameLength = item.getName().length();
					int i = Integer.parseInt(item.getName().substring("createdSets".length(), fileNameLength - ".gz".length()));
					createdSetsMap.put(i, createdSets);
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

			});
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return createdSetsMap;
	}

}
