package de.mannheim.uni.expansion.tablefusion.adapters.learningsets;

import java.util.Map;
import java.util.UUID;

import javax.inject.Inject;
import javax.inject.Singleton;

import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.learning.LearningSetCreator;
import de.mannheim.uni.expansion.tablefusion.learning.LearningSetCreator.CreatedSets;
import de.mannheim.uni.expansion.tablefusion.services.grouping.GroupedValuesRepository;
import de.mannheim.uni.expansion.utils.SimpleConcurrentCache;

@Singleton
public class LearningSetsRepository {

	private final LearningSetsReader reader;
	private final LearningSetsWriter writer;
	private static SimpleConcurrentCache<UUID, Map<Integer, CreatedSets>> cache = new SimpleConcurrentCache<>();
	private final LearningSetCreator learningSetCreator;

	private final GroupedValuesRepository groupedValuesRepository;

	@Inject
	public LearningSetsRepository(LearningSetsReader reader, LearningSetsWriter writer, LearningSetCreator creator,
			GroupedValuesRepository groupedValuesRepository) {
		this.reader = reader;
		this.writer = writer;
		this.learningSetCreator = creator;
		this.groupedValuesRepository = groupedValuesRepository;
	}

	public Map<Integer, CreatedSets> getCreatedSets(int count, GroupedValues groupedValues) {
		if (cache.hasCacheFor(groupedValues.getUUID())) {

			return cache.get(groupedValues.getUUID());
		} else {
			Map<Integer, CreatedSets> ret = getCreatedSetsNoCache(count, groupedValues);
			cache.add(groupedValues.getUUID(), ret);
			return ret;
		}
	}

	public Map<Integer, CreatedSets> getCreatedSetsNoCache(int count, GroupedValues groupedValues) {
		return learningSetCreator.createByTableProperty(count, groupedValues);
	}
	//
	// private synchronized void loadOrCreate() {
	// if (cache == null) {
	//
	// cache = reader.read();
	// if (cache == null) {
	//
	// cache =
	// getCreatedSetsNoCache(groupedValuesRepository.getGroupedValues());
	// store(cache);
	// }
	//
	// }
	// }

	// public void store(Map<Integer, CreatedSets> store) {
	// writer.write(store);
	// }
}
