package de.mannheim.uni.expansion.tablefusion.adapters.learningsets;

import java.io.FileOutputStream;
import java.io.OutputStream;
import java.util.Map;
import java.util.Map.Entry;
import java.util.zip.GZIPOutputStream;

import javax.inject.Inject;
import javax.inject.Named;

import com.esotericsoftware.kryo.Kryo;
import com.esotericsoftware.kryo.KryoException;
import com.esotericsoftware.kryo.io.Output;

import de.mannheim.uni.expansion.tablefusion.NamedProperties;
import de.mannheim.uni.expansion.tablefusion.learning.LearningSetCreator.CreatedSets;
import de.mannheim.uni.expansion.utils.PersistentParallel;

public class LearningSetsWriter {

	private final ThreadLocal<Kryo> kryoLocal;
	private final String outputFolder;

	@Inject
	public LearningSetsWriter(@Named(NamedProperties.LEARNING_SETS_OUTPUT_FOLDER) String outputFolder) {
		kryoLocal = new ThreadLocal<>();
		this.outputFolder = outputFolder;
	}

	public void write(Map<Integer, CreatedSets> store) {

		if (outputFolder == null)
			return;

		try {
			new PersistentParallel<Entry<Integer, CreatedSets>>("Writing CreatedSets").foreach(store.entrySet(), (item) -> {
				Kryo kryo = kryoLocal.get();
				if (kryo == null) {
					kryo = new Kryo();
					kryoLocal.set(kryo);
				}
				try {

					OutputStream stream = new GZIPOutputStream(new FileOutputStream(outputFolder + "/createdSets" + item.getKey() + ".gz"));

					Output output = new Output(stream);
					kryo.writeObject(output, item.getValue());
					output.close();

				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

			});
		} catch (KryoException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

}
