package de.mannheim.uni.expansion.tablefusion.adapters.ranking;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import javax.inject.Inject;

import com.opencsv.CSVParser;
import com.opencsv.CSVWriter;

import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.HostId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.utils.BatchConsumerProducer;

public class RankCSVAdapter {

	private final AnnotatedTablesRepository annotatedTablesRepository;

	@Inject
	public RankCSVAdapter(AnnotatedTablesRepository annotatedTablesRepository) {
		this.annotatedTablesRepository = annotatedTablesRepository;
	}

	public Map<HostId, Double> read(String hostRankingFile, boolean checkHostId) {

		try {

			final Set<HostId> hostIds = new HashSet<>();

			if (checkHostId) {
				for (AnnotatedTableId tableId : annotatedTablesRepository.getAllTableIds()) {
					hostIds.add(annotatedTablesRepository.getMetaData(tableId).getHostId());
				}
			}

			CSVParser parser = new CSVParser('\t');
			final Map<HostId, Double> rankings = new ConcurrentHashMap<HostId, Double>();

			BatchConsumerProducer<String> consumer = BatchConsumerProducer.create("Read Ranks", (finalNextLine) -> {

				try {
					String[] finalLine = parser.parseLineMulti(finalNextLine);
					HostId hostId = new HostId(finalLine[0]);

					if (!checkHostId || hostIds.contains(hostId)) {

						Double score = Double.valueOf(finalLine[1]);

						rankings.put(hostId, score);
					}

				} catch (NumberFormatException e2) {
					e2.printStackTrace();
				} catch (Exception e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}

			} , 1000000);

			BufferedReader br = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(hostRankingFile))));

			List<String> lines = new ArrayList<>();
			String nextLine = null;

			int lineCount = 0;
			while ((nextLine = br.readLine()) != null) {

				if (lineCount == 10000000) {
					consumer.addBatch(lines);
					lineCount = 0;
					lines = new ArrayList<>();
				}

				lines.add(nextLine);
				lineCount++;

			}

			br.close();

			if (!lines.isEmpty()) {
				consumer.addBatch(lines);
			}

			consumer.finish();
			return rankings;

		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		return null;

	}

	public void write(String hostRankingsFileFiltered, Map<HostId, Double> rankings) {
		try {
			CSVWriter writer = new CSVWriter(new OutputStreamWriter(new GZIPOutputStream(new FileOutputStream(hostRankingsFileFiltered), 204800)), '\t');

			for (Entry<HostId, Double> e : rankings.entrySet()) {
				String[] line = new String[2];
				line[0] = e.getKey().getHostId();
				line[1] = String.valueOf(e.getValue());
				writer.writeNext(line);
			}

			writer.close();
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

	}

}
