package de.mannheim.uni.expansion.tablefusion.adapters.ranking;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

import javax.inject.Inject;
import javax.inject.Named;

import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.HostId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.NamedProperties;
import de.mannheim.uni.expansion.tablefusion.analysis.Accuracy;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.tripleids.TripleIdTableScores;
import de.mannheim.uni.expansion.utils.PersistentParallel3;
import de.mannheim.uni.expansion.utils.SimpleConcurrentCache;

public class RankRepositoryImpl implements RankRepository {

	private final Map<HostId, Double> rankings;
	private final RankCSVAdapter rankCSVAdapter;
	private final String filteredRankingFile;
	private final AnnotatedTablesRepository annotatedTablesRepository;
	private final Map<AnnotatedTableId, HostId> mappings;

	@Inject
	public RankRepositoryImpl(RankCSVAdapter rankCSVAdapter, @Named(NamedProperties.HOST_RANKINGS_FILE_FILTERED) String filteredRankingFile,
			AnnotatedTablesRepository annotatedTablesRepository) {
		this.filteredRankingFile = filteredRankingFile;
		this.rankCSVAdapter = rankCSVAdapter;

		this.annotatedTablesRepository = annotatedTablesRepository;

		this.mappings = getHostMappings();

		rankings = rankCSVAdapter.read(filteredRankingFile, false);

		// if (rankings != null) {
		// // double max =
		// //
		// rankings.entrySet().stream().mapToDouble(Entry::getValue).max().getAsDouble();
		// // double min =
		// //
		// rankings.entrySet().stream().mapToDouble(Entry::getValue).min().getAsDouble();
		// //
		// // double diff = max - min;
		// //
		// // rankings =
		// //
		// rankings.entrySet().stream().collect(Collectors.toMap(Entry::getKey,
		// // (e) -> {
		// //
		// // return (e.getValue() - min) / diff;
		// //
		// // }));
		// double min = rankings.entrySet().stream().mapToDouble((e) -> {
		// return e.getValue();
		// }).min().getAsDouble();
		// double max = rankings.entrySet().stream().mapToDouble((e) -> {
		// return e.getValue();
		// }).max().getAsDouble();
		// double diff = max - min;
		//
		// double minP = Math.pow(min, 1.0 / 10.0);
		// double maxP = Math.pow(max, 1.0 / 10.0);
		// double diffP = maxP - minP;
		//
		// this.rankings =
		// rankings.entrySet().stream().collect(Collectors.toMap(Entry::getKey,
		// e -> {
		//
		// double v = Math.pow(e.getValue(), 1.0 / 10.0);
		//
		// double a = ((v - minP) / diffP) * 0.99 + 0.01;
		//
		// return new Probability(new Accuracy(a, 1));
		// }));
		//
		// for (HostId e : mappings.values()) {
		// if (this.rankings.get(e) == null) {
		// this.rankings.put(e, new Probability(new Accuracy(0.0, 1)));
		// }
		// }
		//
		// } else {
		// this.rankings = new HashMap<>();
		// }

	}

	private static SimpleConcurrentCache<GroupedValues, TripleIdTableScores> cache = new SimpleConcurrentCache<>();
	public static AtomicInteger countTotal = new AtomicInteger();
	public static AtomicInteger countMissing = new AtomicInteger();

	@Override
	public ScoresContainer getScores(GroupedValues groupedValues) {

		Map<AnnotatedTableId, HostId> hostMappings = getHostMappings();

		// cache = new SimpleConcurrentCache<>();
		if (cache.hasCacheFor(groupedValues)) {
			return cache.get(groupedValues);
		} else {

			ConcurrentHashMap<TripleId, Map<AnnotatedTableId, Probability>> valueScores = new ConcurrentHashMap<>();

			PersistentParallel3.runAndFinish(groupedValues.entrySet(), (gv) -> {

				Set<AnnotatedTableId> tableSet = new HashSet<>();

				for (SingleValue sv : gv.getValue().getValues()) {
					tableSet.add(sv.getId().getTableId());
				}

				TripleId tid = gv.getKey();
				Set<AnnotatedTableId> noScore = new HashSet<>();
				Map<AnnotatedTableId, Double> withScore = new HashMap<>();

				for (AnnotatedTableId aid : tableSet) {
					Double d = this.rankings.get(this.mappings.get(aid));
					if (d == null) {
						noScore.add(aid);
					} else {
						withScore.put(aid, d);
					}
				}

				if (noScore.size() > 0) {
					countMissing.addAndGet(noScore.size());
					// System.out.println("NoPageRankScore: " +
					// noScore.toString());
				}

				Map<AnnotatedTableId, Probability> scores = new HashMap<>();
				if (!withScore.isEmpty()) {
					countTotal.addAndGet(withScore.size());
					double min = withScore.entrySet().stream().mapToDouble((e) -> {
						return e.getValue();
					}).min().getAsDouble();

					double max = withScore.entrySet().stream().mapToDouble((e) -> {
						return e.getValue();
					}).max().getAsDouble();
					double diff = max - min;

					// int divider = withScore.size() + (noScore.isEmpty() ? 0 :
					// 1);
					// double step = 1.0 / divider;
					//
					// int i = 1;
					// double lowestScore = 1.0;

					for (Entry<AnnotatedTableId, Double> e : withScore.entrySet()) {

						Probability d = new Probability(new Accuracy((e.getValue() - min) / diff, 1));

						scores.put(e.getKey(), d);

					}

					// if (sortedByYear.isEmpty()) {
					// for (CellId aid : noScore) {
					// subCHMMap.put(aid, new Probability(new Accuracy(0.0,
					// 1)));
					// }
					// } else {
					// for (CellId aid : noScore) {
					// subCHMMap.put(aid, new Probability(new Accuracy(0.0,
					// 1)));
					// }
					// }
				}

				if (!scores.isEmpty()) {
					valueScores.put(tid, scores);
				}

			} , "ColumnYearsScores", 10000);

			TripleIdTableScores cs = new TripleIdTableScores(valueScores);
			cache.add(groupedValues, cs);

			return cs;
		}
	}

	private Map<AnnotatedTableId, HostId> getHostMappings() {
		Map<AnnotatedTableId, HostId> mappings = new HashMap<>();

		for (AnnotatedTableId tableId : annotatedTablesRepository.getAllTableIds()) {
			AnnotatedTableMetaData metaData = annotatedTablesRepository.getMetaData(tableId);
			HostId hostId = metaData.getHostId();
			mappings.put(tableId, hostId);

		}

		return mappings;
	}

}
