package de.mannheim.uni.expansion.tablefusion.adapters.scores;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.inject.Inject;
import javax.inject.Named;

import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.tablefusion.NamedProperties;

public class ScoresOutputWriter {

	private final String outputFolder;

	@Inject
	public ScoresOutputWriter(@Named(NamedProperties.SCORES_FOLDER) String outputFolder) {
		this.outputFolder = outputFolder;
	}

	public void write(Integer learningSetNumber, Map<ClassPropertyId, List<ScoresOutputEntry>> map) throws IOException {

		write(String.valueOf(learningSetNumber), map);

	}

	private void write(String prefix, Map<ClassPropertyId, List<ScoresOutputEntry>> map) throws IOException {

		String[] headerLine = new String[8];
		headerLine[0] = "Class";
		headerLine[1] = "Property";
		headerLine[2] = "InstanceScore";
		headerLine[3] = "ColumnScore";
		headerLine[4] = "TableScore";
		headerLine[5] = "HostScore";
		headerLine[6] = "SimilarityType";
		headerLine[7] = "Result";

		CSVWriter fullWriter = new CSVWriter(new FileWriter(outputFolder + "/scores" + prefix + ".csv"), ';', '"', '\\');

		fullWriter.writeNext(headerLine);

		int i = 0;

		for (Entry<ClassPropertyId, List<ScoresOutputEntry>> e : map.entrySet()) {

			new File(outputFolder + "/scores" + prefix).mkdirs();

			CSVWriter csvwriter = new CSVWriter(new FileWriter(outputFolder + "/scores" + prefix + "/" + i + ".csv"), ';', '"', '\\');

			csvwriter.writeNext(headerLine);

			for (ScoresOutputEntry se : e.getValue()) {
				String[] line = getLine(e.getKey(), se);

				csvwriter.writeNext(line);
				fullWriter.writeNext(line);
			}

			csvwriter.close();

			i++;
		}

		fullWriter.close();

	}

	public void write(Map<ClassPropertyId, List<ScoresOutputEntry>> map) throws IOException {

		write("", map);

	}

	private String[] getLine(ClassPropertyId classPropertyId, ScoresOutputEntry se) {
		String[] line = new String[8];
		line[0] = classPropertyId.getClassId().toString();
		line[1] = classPropertyId.getPropertyId().toString();
		line[2] = String.valueOf(se.getInstance());
		line[3] = String.valueOf(se.getColumn());
		line[4] = String.valueOf(se.getTable());
		line[5] = String.valueOf(se.getHost());
		line[6] = se.getSimilarityType().toString();
		if (se.isCorrect())
			line[7] = "1.0";
		else
			line[7] = "-1.0";

		return line;
	}
}
