package de.mannheim.uni.expansion.tablefusion.analysis;

public class Accuracy implements Comparable<Accuracy> {

	private final double accuracy;

	private final int total;
	private final int correct;

	public Accuracy(int correct, int wrong) {
		this.total = correct + wrong;
		this.accuracy = ((double) correct) / total;
		this.correct = correct;
	}

	public static Accuracy fromCorrectAndTotal(double correct, double total) {
		return new Accuracy(((double) correct) / total, (int) total);
	}

	public static Accuracy fromCorrectAndTotal(int correct, int total) {
		return new Accuracy(((double) correct) / total, total);
	}

	public static Accuracy harmonicMean(Accuracy a, Accuracy b, int total) {
		double harmonicMean = (2 * a.getAccuracy() * b.getAccuracy()) / (a.getAccuracy() + b.getAccuracy());
		return new Accuracy(harmonicMean, total);
	}

	public Accuracy(double accuracy, int total) {
		this.accuracy = accuracy;
		this.total = total;
		this.correct = (int) Math.round(accuracy * total);
	}

	public double getAccuracy() {
		return accuracy;
	}

	public int getTotal() {
		return total;
	}

	@Override
	public String toString() {
		return "Accuracy [accuracy=" + accuracy + ", total=" + total + "]";
	}

	@Override
	public int compareTo(Accuracy o) {
		return Double.compare(this.getAccuracy(), o.getAccuracy());
	}

	public int getCorrect() {
		return correct;
	}

	public static Accuracy simple(double d) {
		return new Accuracy(d, 1);
	}

}
