package de.mannheim.uni.expansion.tablefusion.analysis;

import java.io.IOException;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;
import java.util.stream.Collectors;

import javax.inject.Inject;

import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleIdContainer;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.FusedResultsSingle;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.PerformancePR;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.PrecisionPerformance;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityMeasure;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.CHMHelper.Creator;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.qualifiers.TimestampPointInTime;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.dws.expansion.data.ontology.model.valuepairs.AbstractClassPair;
import de.uni_mannheim.informatik.dws.expansion.evaluation.FilteredValues;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;

public class AccuracyCalculator {

	private final KnowledgeBase dbPediaRepository;

	private final OntologyRepository ontologyRepository;

	private final SimilarityCalculator similarityCalculator;

	// private final PersistentParallel3<Entry<TripleId, FusedValue>> parallel;

	private final MaxCorrectFinder maxCorrectFinder;

	private final ValueEvaluator valueEvaluator;

	@Inject
	public AccuracyCalculator(KnowledgeBase dbPediaRepository, SimilarityCalculator similarityCalculator, OntologyRepository ontologyRepository,
			MaxCorrectFinder maxCorrectFinder, ValueEvaluator valueEvaluator) {
		this.dbPediaRepository = dbPediaRepository;
		this.similarityCalculator = similarityCalculator;
		// this.parallel = new PersistentParallel3<>("AccuracyCalculator");
		this.ontologyRepository = ontologyRepository;
		this.maxCorrectFinder = maxCorrectFinder;
		this.valueEvaluator = valueEvaluator;
	}

	private static final Creator<FusedResultsSingle> fusedResultsCreator = () ->

	{
		return new FusedResultsSingle();
	};

	// public FusionPerformance<PrecisionPerformance>
	// analyzeAccuracy(Map<TripleId, FusedValue> fusedValues) {
	// FusionPerformance<FusedResultsSingle> fusedResults =
	// analyzeFusedValues(fusedValues);
	// final FusedResultsSingle fr = fusedResults.getTotal();
	//
	// Map<Boolean, FusedResultsSingle> perTimeDependenceMap =
	// fusedResults.getPerTimeDependence();
	// Map<ValueType, FusedResultsSingle> perDataTypeMap =
	// fusedResults.getPerDataType();
	//
	// Map<ClassId, FusedResultsSingle> perClassIdMap =
	// fusedResults.getPerClassId();
	//
	// Map<ClassPropertyId, FusedResultsSingle> perClassPropertyMap =
	// fusedResults.getPerProperty();
	//
	// PrecisionPerformance totalPerformance = getPerformance(fr);
	//
	// Map<Boolean, PrecisionPerformance> perTimeDependencePerformance =
	// perTimeDependenceMap.entrySet().stream()
	// .collect(Collectors.toMap(Entry::getKey, (e) -> {
	// return getPerformance(e.getValue());
	// }));
	//
	// Map<ValueType, PrecisionPerformance> perDataTypePerformance =
	// perDataTypeMap.entrySet().stream().collect(Collectors.toMap(Entry::getKey,
	// (e) -> {
	// return getPerformance(e.getValue());
	// }));
	//
	// Map<ClassId, PrecisionPerformance> perClassIdPerformance =
	// perClassIdMap.entrySet().stream().collect(Collectors.toMap(Entry::getKey,
	// (e) -> {
	// return getPerformance(e.getValue());
	// }));
	//
	// Map<ClassPropertyId, PrecisionPerformance> perClassPropertyIdPerformance
	// = perClassPropertyMap.entrySet().stream()
	// .collect(Collectors.toMap(Entry::getKey, (e) -> {
	// return getPerformance(e.getValue());
	// }));
	//
	// return new FusionPerformance<PrecisionPerformance>(totalPerformance,
	// perTimeDependencePerformance, perDataTypePerformance,
	// perClassIdPerformance,
	// perClassPropertyIdPerformance);
	//
	// }

	public <E extends TripleIdContainer> FusionPerformance<PerformancePR> analyzeAccuracy(Map<E, FusedValue> fusedValues,
			FusionPerformance<Integer> maxCorrect) {
		return analyzeAccuracy(fusedValues, maxCorrect, null);
	}

	public <E extends TripleIdContainer> FusionPerformance<PerformancePR> analyzeAccuracy(Map<E, FusedValue> fusedValues, FusionPerformance<Integer> maxCorrect,
			FilteredValues filteredValues) {
		return analyzeAccuracy(fusedValues, maxCorrect, null, null);
	}

	public <E extends TripleIdContainer> FusionPerformance<PerformancePR> analyzeAccuracy(Map<E, FusedValue> fusedValues, FusionPerformance<Integer> maxCorrect,
			FilteredValues filteredValues, FusionPerformance<Integer> wrong) {

		FusionPerformance<FusedResultsSingle> fusedResults = analyzeFusedValues(fusedValues, filteredValues);
		final FusedResultsSingle fr = fusedResults.getTotal();

		Map<Boolean, FusedResultsSingle> perTimeDependenceMap = fusedResults.getPerTimeDependence();
		Map<ValueType, FusedResultsSingle> perDataTypeMap = fusedResults.getPerDataType();

		Map<ClassId, FusedResultsSingle> perClassIdMap = fusedResults.getPerClassId();

		Map<ClassPropertyId, FusedResultsSingle> perClassPropertyMap = fusedResults.getPerProperty();
		Map<TimestampPointInTime, FusedResultsSingle> perYear = fusedResults.getPerYear();

		Map<Pair<ClassPropertyId, TimestampPointInTime>, FusedResultsSingle> perClassPropertyIdAndYear = fusedResults.getPerClassPropertyAndYear();
		// Map<Pair<ClassId, String>, FusedResultsSingle> perClassAndSet =
		// fusedResults.getPerClassAndSet();

		int max = maxCorrect.getTotal();
		Map<Boolean, Integer> maxPerTimeDependence = maxCorrect.getPerTimeDependence();
		Map<ValueType, Integer> maxPerDataType = maxCorrect.getPerDataType();
		Map<ClassId, Integer> maxPerClassId = maxCorrect.getPerClassId();
		Map<ClassPropertyId, Integer> maxPerClassPropertyId = maxCorrect.getPerProperty();
		Map<TimestampPointInTime, Integer> maxPerYear = maxCorrect.getPerYear();
		Map<Pair<ClassPropertyId, TimestampPointInTime>, Integer> maxPerClassPropertyIdAndYear = maxCorrect.getPerClassPropertyAndYear();
		// Map<Pair<ClassId, String>, Integer> maxPerClassIdAndSet =
		// maxCorrect.getPerClassAndSet();

		int totalWrong = 0;
		if (wrong != null) {
			totalWrong = wrong.getTotal();
		}

		PerformancePR totalPerformance = getPerformance(fr, max, totalWrong);

		Map<Boolean, PerformancePR> perTimeDependencePerformance = perTimeDependenceMap.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			int wrongCount = 0;
			if (wrong != null) {
				wrongCount = wrong.getPerTimeDependence().getOrDefault(e.getKey(), 0);
			}
			return getPerformance(e.getValue(), maxPerTimeDependence.getOrDefault(e.getKey(), 0), wrongCount);
		}));

		Map<ValueType, PerformancePR> perDataTypePerformance = perDataTypeMap.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			int wrongCount = 0;
			if (wrong != null) {
				wrongCount = wrong.getPerDataType().getOrDefault(e.getKey(), 0);
			}

			int maxValue = maxPerDataType.getOrDefault(e.getKey(), 0);
			return getPerformance(e.getValue(), maxPerDataType.getOrDefault(e.getKey(), 0), wrongCount);
		}));

		Map<ClassId, PerformancePR> perClassIdPerformance = perClassIdMap.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			int wrongCount = 0;
			if (wrong != null) {
				wrongCount = wrong.getPerClassId().getOrDefault(e.getKey(), 0);
			}
			return getPerformance(e.getValue(), maxPerClassId.getOrDefault(e.getKey(), 0), wrongCount);
		}));

		Map<ClassPropertyId, PerformancePR> perClassPropertyIdPerformance = perClassPropertyMap.entrySet().stream()
				.collect(Collectors.toMap(Entry::getKey, (e) -> {
					int wrongCount = 0;

					if (wrong != null) {
						wrongCount = wrong.getPerProperty().getOrDefault(e.getKey(), 0);
					}
					return getPerformance(e.getValue(), maxPerClassPropertyId.getOrDefault(e.getKey(), 0), wrongCount);
				}));

		Map<TimestampPointInTime, PerformancePR> perYearPerformance = perYear.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
			int wrongCount = 0;
			if (wrong != null) {
				wrongCount = wrong.getPerYear().getOrDefault(e.getKey(), 0);
			}
			return getPerformance(e.getValue(), maxPerYear.getOrDefault(e.getKey(), 0), wrongCount);
		}));

		Map<Pair<ClassPropertyId, TimestampPointInTime>, PerformancePR> perClassPropertyIdAndYearPerformance = perClassPropertyIdAndYear.entrySet().stream()
				.collect(Collectors.toMap(Entry::getKey, (e) -> {
					int wrongCount = 0;

					if (wrong != null) {
						wrongCount = wrong.getPerClassPropertyAndYear().getOrDefault(e.getKey(), 0);
					}
					return getPerformance(e.getValue(), maxPerClassPropertyIdAndYear.getOrDefault(e.getKey(), 0), wrongCount);
				}));

		// Map<Pair<ClassId, String>, PerformancePR> perClassAndSetPerformance =
		// perClassAndSet.entrySet().stream()
		// .collect(Collectors.toMap(Entry::getKey, (e) -> {
		// int wrongCount = 0;
		//
		// if (wrong != null) {
		// wrongCount = wrong.getPerClassAndSet().getOrDefault(e.getKey(), 0);
		// }
		// return getPerformance(e.getValue(),
		// maxCorrect.getPerClassAndSet().getOrDefault(e.getKey(), 0),
		// wrongCount);
		// }));

		return new FusionPerformance<PerformancePR>(totalPerformance, perTimeDependencePerformance, perDataTypePerformance, perClassIdPerformance,
				perClassPropertyIdPerformance, perYearPerformance, perClassPropertyIdAndYearPerformance);

		// if (print) {
		// for (SimilarityType type : SimilarityType.values()) {
		// Double accuracy = fr.getAccuracy(type);
		//
		// if (string != null) {
		// System.out.print("Results for " + string + " - " + type.toString() +
		// " - ");
		//
		// if (accuracy.equals(Double.NaN)) {
		// System.out.println("NO MATCHES");
		// } else {
		// System.out.println("Accuracy achieved: " + accuracy + " (out of " +
		// fr.getOverlappingTriples(type) + ")");
		// }
		// }
		// }
		// }

	}

	private FusionPerformance<FusedResultsSingle> analyzeFusedValues(Map<TripleId, FusedValue> fusedValues) {
		return analyzeFusedValues(fusedValues, null);
	}

	private <E extends TripleIdContainer> FusionPerformance<FusedResultsSingle> analyzeFusedValues(Map<E, FusedValue> fusedValues,
			FilteredValues filteredValues) {
		final FusedResultsSingle fr = new FusedResultsSingle();
		ConcurrentHashMap<ValueType, FusedResultsSingle> perDataTypeMap = new ConcurrentHashMap<ValueType, FusedResultsSingle>();

		ConcurrentHashMap<Boolean, FusedResultsSingle> perTimeDependenceMap = new ConcurrentHashMap<Boolean, FusedResultsSingle>();

		ConcurrentHashMap<ClassId, FusedResultsSingle> perClassIdMap = new ConcurrentHashMap<ClassId, FusedResultsSingle>();

		ConcurrentHashMap<TimestampPointInTime, FusedResultsSingle> perYearMap = new ConcurrentHashMap<TimestampPointInTime, FusedResultsSingle>();

		ConcurrentHashMap<ClassPropertyId, FusedResultsSingle> perClassPropertyMap = new ConcurrentHashMap<ClassPropertyId, FusedResultsSingle>();

		ConcurrentHashMap<Pair<ClassPropertyId, TimestampPointInTime>, FusedResultsSingle> perClassPropertyAndYearMap = new ConcurrentHashMap<>();

		ConcurrentHashMap<Pair<ClassId, String>, FusedResultsSingle> perClassAndSet = new ConcurrentHashMap<>();

		try {
			PersistentParallel4.runAndFinish(fusedValues.entrySet(), (e) -> {

				TripleId tid = e.getKey().tripleId();
				ClassPropertyId classPropertyId = new ClassPropertyId(tid.getClassId(), tid.getPropertyId());

				FusedResultsSingle perYear = null;
				FusedResultsSingle perClassPropertyAndYear = null;
				if (e.getKey() instanceof QualifiedTripleId) {
					QualifiedTripleId qtid = (QualifiedTripleId) e.getKey();
					TimestampPointInTime tipt = null;
					for (AbstractQualifier qualifier : qtid.getQualifiers()) {
						if (qualifier instanceof TimestampPointInTime) {
							tipt = (TimestampPointInTime) qualifier;
							break;
						}
					}
					if (tipt != null) {
						perYear = CHMHelper.getFromCHM(perYearMap, tipt, fusedResultsCreator);
						perClassPropertyAndYear = CHMHelper.getFromCHM(perClassPropertyAndYearMap, new Pair<>(classPropertyId, tipt), fusedResultsCreator);
					}

				}

				ValueType cdt = ontologyRepository.getDataType(classPropertyId.getPropertyId());

				FusedResultsSingle perDataType = CHMHelper.getFromCHM(perDataTypeMap, cdt, fusedResultsCreator);

				FusedResultsSingle perTimeDependence = CHMHelper.getFromCHM(perTimeDependenceMap,
						ontologyRepository.isTimeDependent(classPropertyId.getPropertyId()), fusedResultsCreator);

				FusedResultsSingle perClassId = CHMHelper.getFromCHM(perClassIdMap, tid.getClassId(), fusedResultsCreator);
				FusedResultsSingle perClassPropertyId = CHMHelper.getFromCHM(perClassPropertyMap, classPropertyId, fusedResultsCreator);

				Optional<Boolean> correctValue = null;
				FusedValue fusedValue = e.getValue();
				QualifiedValueList o = dbPediaRepository.getQualifiedValueListWithEqualProperties(tid.getValueId());

				if (o == null) {
					fr.incNew();
					perDataType.incNew();
					perTimeDependence.incNew();
					perClassId.incNew();
					perClassPropertyId.incNew();
					if (perYear != null) {
						perYear.incNew();
					}
					if (perClassPropertyAndYear != null) {
						perClassPropertyAndYear.incNew();
					}

					return;
				} else {
					if (cdt == null) {
						// System.out.println("A: CDT == null, please check!");
					}
					if (filteredValues != null) {
						Set<AbstractObject> lookingFor = filteredValues.getFilteredValues(tid);
						if (lookingFor == null || lookingFor.size() == 0) {
							return;
						} else if (o.size() > 1) {
							List<AbstractObject> correctList = new ArrayList<>(lookingFor);

							List<AbstractObject> fusedList = fusedValue.getSimpleList();

							List<AbstractObject> valuesToEliminate = new ArrayList<>((List<AbstractObject>) o);

							FusedValue newFusedList = eliminateValuesButKeep(valuesToEliminate, fusedValue, correctList, cdt);

							if (newFusedList.listSize() == 0 || newFusedList.listSize() < lookingFor.size()) {
								return;
							} else {
								correctValue = valueEvaluator.correct(e.getKey(), o, newFusedList, cdt, filteredValues);
								int dfdf = 123;
							}

						}

					} else {
						if (o.isList() && fusedValue.isList()) {

							if (fusedValue.listSize() < o.size()) {
								return;
							} else if (fusedValue.listSize() == 0) {
								return;
							}
							throw new IllegalStateException();
						}
						correctValue = valueEvaluator.correct(e.getKey(), o, fusedValue, cdt);
					}

					if (correctValue.isPresent()) {
						if (correctValue.get()) {
							fr.incCorrect();
							perDataType.incCorrect();
							perTimeDependence.incCorrect();
							perClassId.incCorrect();
							perClassPropertyId.incCorrect();
							if (perYear != null) {
								perYear.incCorrect();
							}
							if (perClassPropertyAndYear != null) {
								perClassPropertyAndYear.incCorrect();
							}
						} else {
							if (fusedValue instanceof List<?>) {
								int fdfd = 123;
							}
							fr.incWrong();
							perDataType.incWrong();
							perTimeDependence.incWrong();
							perClassId.incWrong();
							perClassPropertyId.incWrong();
							if (perYear != null) {
								perYear.incWrong();
							}
							if (perClassPropertyAndYear != null) {
								perClassPropertyAndYear.incWrong();
							}
						}
					} else {
						fr.incNew();
						perDataType.incNew();
						perTimeDependence.incNew();
						perClassId.incNew();
						perClassPropertyId.incNew();
						if (perYear != null) {
							perYear.incNew();
						}
						if (perClassPropertyAndYear != null) {
							perClassPropertyAndYear.incNew();
						}
					}
				}

			}, "ACCURACY CALCULATOR");

		} catch (Exception e) {
			e.printStackTrace();
		}
		return new FusionPerformance<FusedResultsSingle>(fr, perTimeDependenceMap, perDataTypeMap, perClassIdMap, perClassPropertyMap, perYearMap,
				perClassPropertyAndYearMap);

	}

	private PrecisionPerformance getPerformance(FusedResultsSingle results) {
		return new PrecisionPerformance(results.getCorrectTriples(), results.getWrongTriples(), results.getNewTriples());
	}

	private PerformancePR getPerformance(FusedResultsSingle results, int max, int wrong) {
		return new PerformancePR(results.getCorrectTriples(), results.getWrongTriples() + wrong, max, results.getNewTriples());

	}

	public FusionPerformance<PerformancePR> averageAccuracies(List<FusionPerformance<PerformancePR>> fusionResults) {

		List<PerformancePR> groupedTotal = new ArrayList<>();

		Map<ValueType, List<PerformancePR>> groupedPerDataType = new HashMap<>();

		Map<Boolean, List<PerformancePR>> groupedPerTimeDependence = new HashMap<>();
		Map<TimestampPointInTime, List<PerformancePR>> groupedPerYear = new HashMap<>();
		Map<ClassId, List<PerformancePR>> groupedPerClassId = new HashMap<>();
		Map<ClassPropertyId, List<PerformancePR>> groupedPerClassPropertyId = new HashMap<>();
		Map<Pair<ClassPropertyId, TimestampPointInTime>, List<PerformancePR>> groupedPerClassPropertyAndYear = new HashMap<>();
		Map<Pair<ClassId, String>, List<PerformancePR>> groupedPerClassAndSet = new HashMap<>();

		for (FusionPerformance<PerformancePR> fp : fusionResults) {
			groupedTotal.add(fp.getTotal());

			for (Entry<ValueType, PerformancePR> e : fp.getPerDataType().entrySet()) {
				List<PerformancePR> list = CHMHelper.getFromMap(groupedPerDataType, e.getKey(), () -> new ArrayList<>());
				list.add(e.getValue());
			}

			for (Entry<Boolean, PerformancePR> e : fp.getPerTimeDependence().entrySet()) {
				List<PerformancePR> list = CHMHelper.getFromMap(groupedPerTimeDependence, e.getKey(), () -> new ArrayList<>());
				list.add(e.getValue());
			}

			for (Entry<ClassId, PerformancePR> e : fp.getPerClassId().entrySet()) {
				List<PerformancePR> list = CHMHelper.getFromMap(groupedPerClassId, e.getKey(), () -> new ArrayList<>());
				list.add(e.getValue());
			}

			for (Entry<ClassPropertyId, PerformancePR> e : fp.getPerProperty().entrySet()) {
				List<PerformancePR> list = CHMHelper.getFromMap(groupedPerClassPropertyId, e.getKey(), () -> new ArrayList<>());
				list.add(e.getValue());
			}

			for (Entry<TimestampPointInTime, PerformancePR> e : fp.getPerYear().entrySet()) {
				List<PerformancePR> list = CHMHelper.getFromMap(groupedPerYear, e.getKey(), () -> new ArrayList<>());
				list.add(e.getValue());
			}

			for (Entry<Pair<ClassPropertyId, TimestampPointInTime>, PerformancePR> e : fp.getPerClassPropertyAndYear().entrySet()) {
				List<PerformancePR> list = CHMHelper.getFromMap(groupedPerClassPropertyAndYear, e.getKey(), () -> new ArrayList<>());
				list.add(e.getValue());
			}

			// for (Entry<Pair<ClassId, String>, PerformancePR> e :
			// fp.getPerClassAndSet().entrySet()) {
			// List<PerformancePR> list =
			// CHMHelper.getFromMap(groupedPerClassAndSet, e.getKey(), () -> new
			// ArrayList<>());
			// list.add(e.getValue());
			// }
		}

		Function<Entry<?, List<PerformancePR>>, PerformancePR> average = (entry) -> {
			return PerformancePR.fromList(entry.getValue());
		};

		PerformancePR total = PerformancePR.fromList(groupedTotal);
		Map<ValueType, PerformancePR> perDataType = groupedPerDataType.entrySet().stream().collect(Collectors.toMap(Entry::getKey, average));
		Map<Boolean, PerformancePR> perTimeDependence = groupedPerTimeDependence.entrySet().stream().collect(Collectors.toMap(Entry::getKey, average));

		Map<ClassId, PerformancePR> perClassId = groupedPerClassId.entrySet().stream().collect(Collectors.toMap(Entry::getKey, average));
		Map<ClassPropertyId, PerformancePR> perClassPropertyId = groupedPerClassPropertyId.entrySet().stream()
				.collect(Collectors.toMap(Entry::getKey, average));

		Map<TimestampPointInTime, PerformancePR> perYear = groupedPerYear.entrySet().stream().collect(Collectors.toMap(Entry::getKey, average));

		Map<Pair<ClassPropertyId, TimestampPointInTime>, PerformancePR> perClassPropertyAndYear = groupedPerClassPropertyAndYear.entrySet().stream()
				.collect(Collectors.toMap(Entry::getKey, average));

		Map<Pair<ClassId, String>, PerformancePR> perClassAndSet = groupedPerClassAndSet.entrySet().stream().collect(Collectors.toMap(Entry::getKey, average));

		return new FusionPerformance<PerformancePR>(total, perTimeDependence, perDataType, perClassId, perClassPropertyId, perYear, perClassPropertyAndYear);
	}

	private List<Object> eliminateValues(List<AbstractObject> valuesToRemove, List<AbstractObject> fusedList, ValueType cdt) {
		SimilarityMeasure measure = similarityCalculator.getMeasure(cdt);

		TreeMap<Double, Pair<Integer, Integer>> sims = new TreeMap<>();

		for (int a = 0; a < valuesToRemove.size(); a++) {
			for (int b = 0; b < fusedList.size(); b++) {
				double similarity = measure.getQuickSimilarity(valuesToRemove.get(a), fusedList.get(b));
				if (similarity > measure.getEquivalenceThreshold())
					sims.put(similarity, new Pair<>(a, b));

			}
		}

		Set<Integer> removeIndeces = new HashSet<>();
		for (int a = 0; a < valuesToRemove.size(); a++) {
			removeIndeces.add(a);
		}

		Set<Integer> fusedIndices = new HashSet<>();
		for (int b = 0; b < fusedList.size(); b++) {
			fusedIndices.add(b);
		}

		while (!removeIndeces.isEmpty() && !fusedIndices.isEmpty()) {
			Entry<Double, Pair<Integer, Integer>> e = sims.pollLastEntry();
			if (e == null) {
				break;
			}
			if (removeIndeces.contains(e.getValue().getFirst()) && fusedIndices.contains(e.getValue().getSecond())) {
				removeIndeces.remove(e.getValue().getFirst());
				fusedIndices.remove(e.getValue().getSecond());
			}
		}

		List<Object> retList = new ArrayList<>();
		for (Integer index : fusedIndices) {
			retList.add(fusedList.get(index));
		}

		return retList;
	}

	private FusedValue eliminateValuesButKeep(List<AbstractObject> valuesToRemove, FusedValue fusedValue, List<AbstractObject> keep, ValueType cdt) {
		SimilarityMeasure measure = similarityCalculator.getMeasure(cdt);

		Set<Integer> keepIndices = new HashSet<>();
		List<AbstractObject> fusedList = fusedValue.getSimpleList();

		for (int k = 0; k < keep.size(); k++) {
			for (int f = 0; f < fusedList.size(); f++) {
				double similarity = measure.getQuickSimilarity(keep.get(k), fusedList.get(f));
				if (similarity > measure.getEquivalenceThreshold())
					keepIndices.add(f);

			}
		}

		TreeMap<Double, Pair<Integer, Integer>> eliminateSims = new TreeMap<>();

		for (int v = 0; v < valuesToRemove.size(); v++) {
			for (int f = 0; f < fusedList.size(); f++) {
				if (keepIndices.contains(f))
					continue;
				else {
					double similarity = measure.getQuickSimilarity(valuesToRemove.get(v), fusedList.get(f));
					if (similarity > measure.getEquivalenceThreshold())
						eliminateSims.put(similarity, new Pair<>(v, f));
				}
			}
		}

		Set<Integer> removeIndeces = new HashSet<>();
		for (int a = 0; a < valuesToRemove.size(); a++) {
			removeIndeces.add(a);
		}

		Set<Integer> fusedIndices = new HashSet<>();
		for (int b = 0; b < fusedList.size(); b++) {
			fusedIndices.add(b);
		}

		while (!removeIndeces.isEmpty() && !fusedIndices.isEmpty()) {
			Entry<Double, Pair<Integer, Integer>> e = eliminateSims.pollLastEntry();
			if (e == null) {
				break;
			}
			if (removeIndeces.contains(e.getValue().getFirst()) && fusedIndices.contains(e.getValue().getSecond())) {
				removeIndeces.remove(e.getValue().getFirst());
				fusedIndices.remove(e.getValue().getSecond());
			}
		}

		Set<Object> keepSet = new HashSet<>();
		for (Integer index : fusedIndices) {
			keepSet.add(fusedList.get(index));
		}
		Set<AbstractClassPair> newMap = fusedValue.getScoredList().stream().filter((e) -> keepSet.contains(e.getValue())).collect(Collectors.toSet());
		return new FusedValue(newMap);
	}

	private List<Object> getEquivalentToLookingFor(List<AbstractObject> fusedList, List<AbstractObject> lookingFor, ValueType cdt) {
		SimilarityMeasure measure = similarityCalculator.getMeasure(cdt);

		TreeMap<Double, Pair<Integer, Integer>> sims = new TreeMap<>();

		for (int a = 0; a < lookingFor.size(); a++) {
			for (int b = 0; b < fusedList.size(); b++) {
				double similarity = measure.getQuickSimilarity(lookingFor.get(a), fusedList.get(b));
				if (similarity > measure.getEquivalenceThreshold())
					sims.put(similarity, new Pair<>(a, b));

			}
		}

		Set<Integer> lookingForIndices = new HashSet<>();
		for (int a = 0; a < lookingFor.size(); a++) {
			lookingForIndices.add(a);
		}

		Set<Integer> fusedIndices = new HashSet<>();
		for (int b = 0; b < fusedList.size(); b++) {
			fusedIndices.add(b);
		}

		List<Object> retList = new ArrayList<>();

		while (!lookingForIndices.isEmpty() && !fusedIndices.isEmpty()) {
			Entry<Double, Pair<Integer, Integer>> e = sims.pollLastEntry();
			if (e == null) {
				break;
			}
			if (lookingForIndices.contains(e.getValue().getFirst()) && fusedIndices.contains(e.getValue().getSecond())) {
				lookingForIndices.remove(e.getValue().getFirst());
				fusedIndices.remove(e.getValue().getSecond());
				retList.add(fusedList.get(e.getValue().getSecond()));
			}
		}

		return retList;
	}

	public <E> String getAverageClassPropertiesAsTotalAsCSV(Map<E, FusionPerformance<PerformancePR>> ret) throws IOException {
		StringWriter sw = new StringWriter();
		CSVWriter writer = new CSVWriter(sw);

		int length = 9;
		{
			String[] headerLine = new String[length];
			headerLine[0] = "Method";
			headerLine[1] = "Precision";
			headerLine[2] = "Recall";
			headerLine[3] = "F-Measure";
			headerLine[4] = "AF-Measure";
			headerLine[5] = "VarPREC";
			headerLine[6] = "VarREC";
			headerLine[7] = "VarFM";
			headerLine[8] = "VarAFM";
			writer.writeNext(headerLine);

		}
		for (Entry<E, FusionPerformance<PerformancePR>> entry : ret.entrySet()) {
			Map<ClassPropertyId, PerformancePR> performance = entry.getValue().getPerProperty();
			int count = performance.size();
			double precision = 0.0;
			double recall = 0.0;
			double fMeasure = 0.0;
			double aFmeasure = 0.0;

			for (Entry<ClassPropertyId, PerformancePR> cp : performance.entrySet()) {
				precision += cp.getValue().getPrecision();
				recall += cp.getValue().getRecall();
				fMeasure += cp.getValue().getfMeasure();
				aFmeasure += cp.getValue().getaFMeasure();
			}

			precision = precision / count;
			recall = recall / count;
			fMeasure = fMeasure / count;
			aFmeasure = aFmeasure / count;

			double varPrecision = 0.0;
			double varRecall = 0.0;
			double varFMeasure = 0.0;
			double varAFmeasure = 0.0;

			for (Entry<ClassPropertyId, PerformancePR> cp : performance.entrySet()) {
				varPrecision += Math.pow(cp.getValue().getPrecision() - precision, 2);
				varRecall += Math.pow(cp.getValue().getRecall() - recall, 2);
				varFMeasure += Math.pow(cp.getValue().getfMeasure() - fMeasure, 2);
				varAFmeasure += Math.pow(cp.getValue().getaFMeasure() - aFmeasure, 2);
			}

			String[] line = new String[length];
			line[0] = entry.getKey().toString();
			line[1] = String.valueOf(precision);
			line[2] = String.valueOf(recall);
			line[3] = String.valueOf(fMeasure);
			line[4] = String.valueOf(aFmeasure);
			line[5] = String.valueOf(Math.sqrt(varPrecision / count));
			line[6] = String.valueOf(Math.sqrt(varRecall / count));
			line[7] = String.valueOf(Math.sqrt(varFMeasure / count));
			line[8] = String.valueOf(Math.sqrt(varAFmeasure / count));
			writer.writeNext(line);

		}

		writer.close();
		return sw.toString();
	}

}
