package de.mannheim.uni.expansion.tablefusion.analysis;

import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;

import javax.inject.Inject;
import javax.inject.Named;

import com.opencsv.CSVWriter;

import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.NamedProperties;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionDataType;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionPreproccesser;
import de.mannheim.uni.expansion.tablefusion.services.grouping.GroupedValuesRepository;
import de.mannheim.uni.tableprocessor.model.TableColumn.ColumnDataType;
import de.mannheim.uni.utils.DomainUtils;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;

public class GroupAnalyzer {

	private final GroupedValuesRepository groupedValuesRepository;

	private final AnnotatedTablesRepository annotatedTablesRepository;

	// private final FusionRunner fusionRunner;

	private final String resultsFolder;

	// private final ProbabilitiesCalculator probabilitiesCalculator;

	private final OntologyRepository ontologyRepository;

	private final KnowledgeBase dbBpediaRepository;

	private final ValueEvaluator accuracyCalculator;
	//
	// private final ScoreFusionThresholdFinder scoreFusionThresholdFinder;
	//
	// private final LearningSetsRepository learningSetsRepository;

	@Inject
	public GroupAnalyzer(GroupedValuesRepository groupedValuesRepository, AnnotatedTablesRepository annotatedTablesRepository,
			@Named(NamedProperties.RESULTS_FOLDER) String resultsFolder, OntologyRepository ontologyRepository, KnowledgeBase dbBpediaRepository,
			ValueEvaluator accuracyCalculator) {
		super();
		this.groupedValuesRepository = groupedValuesRepository;
		this.annotatedTablesRepository = annotatedTablesRepository;
		// this.fusionRunner = fusionRunner;
		this.resultsFolder = resultsFolder;
		// this.probabilitiesCalculator = probabilitiesCalculator;
		this.ontologyRepository = ontologyRepository;
		this.dbBpediaRepository = dbBpediaRepository;
		this.accuracyCalculator = accuracyCalculator;
		// this.scoreFusionThresholdFinder = scoreFusionThresholdFinder;
		// this.learningSetsRepository = learningSetsRepository;
	}

	private DecimalFormat dfmt = new DecimalFormat("0.00");

	public void run() throws IOException {
		List<FusionDataType> methodTypes = new ArrayList<>();
		methodTypes.add(FusionDataType.REFERENCE);
		methodTypes.add(FusionDataType.STRING);
		methodTypes.add(FusionDataType.NUMERIC);
		methodTypes.add(FusionDataType.DATE);
		methodTypes.add(FusionDataType.LINK);

		List<FusionPreproccesser> preprocessors = new ArrayList<>();
		preprocessors.add(FusionPreproccesser.NONE);

		GroupedValues groupedValues = groupedValuesRepository.getGroupedValues();

		// Map<Integer, CreatedSets> createdSets =
		// learningSetsRepository.getCreatedSets(count,
		// groupedValues).getCreatedSets();
		//
		// double threshold =
		// scoreFusionThresholdFinder.findOptimumThresholdCrossValidation(createdSets,
		// FusionStrategyRepository.KBT, false);
		//
		// ScoresContainer scores =
		// probabilitiesCalculator.byProperty(groupedValues);
		//
		// Map<FusionRun, Map<TripleId, FusedValue>> fusedValuesMap =
		// fusionRunner.run(methodTypes, FusionPreproccesser.NONE,
		// FusionStrategyRepository.KBT,
		// groupedValues, scores, false, threshold);
		//
		// Map<TripleId, FusedValue> fusedValues =
		// CHMHelper.flattenMap(fusedValuesMap);

		CSVWriter csvWriter = new CSVWriter(new FileWriter(resultsFolder + "/groups.csv"));

		System.out.println("Size of groupedvalues: " + groupedValues.size());

		int lineLength = 10;
		String[] headerLine = new String[lineLength];
		headerLine[0] = "DataType";
		headerLine[1] = "Class";
		headerLine[2] = "Property";
		headerLine[3] = "TimeDependent";
		headerLine[4] = "Instance";
		headerLine[5] = "NoOfSources";
		headerLine[6] = "NoOfPLDS";
		headerLine[7] = "ExistsInKB";
		headerLine[8] = "CorrectInGroup";
		headerLine[9] = "PercentCorrect";
		// headerLine[8] = "Filtered";

		csvWriter.writeNext(headerLine);

		for (Entry<TripleId, MatchedValues> e : groupedValues.entrySet()) {
			String[] line = new String[lineLength];

			// DataType
			ValueType cdt = ontologyRepository.getDataType(e.getKey().getPropertyId());
			if (cdt != null)
				line[0] = cdt.toString();
			else
				line[0] = ColumnDataType.unknown.toString();

			// Class
			line[1] = e.getKey().getClassId().getClassId();

			// Property
			line[2] = e.getKey().getPropertyId().readableString();

			line[3] = ontologyRepository.isTimeDependent(e.getKey().getPropertyId()) ? "TD" : "NON-TD";

			// Instance
			line[4] = e.getKey().getInstanceId().getReadableString();

			// NoOfSources
			line[5] = String.valueOf(e.getValue().getValues().size());

			// NoOfPLDS

			Set<String> uniquePLDs = new HashSet<>();
			for (SingleValue sv : e.getValue().getValues()) {
				String source = annotatedTablesRepository.getMetaData(sv.getId().getTableId()).getSource().getSource();

				String PLD = DomainUtils.getPayLevelDomainFromWholeURL(source);
				uniquePLDs.add(PLD);

			}

			line[6] = String.valueOf(uniquePLDs.size());

			// ExistsInDBP
			if (dbBpediaRepository.hasValue(e.getKey().getValueId())) {
				line[7] = "1";
			} else {
				line[7] = "0";
			}

			// Object correctValue = dbBpediaRepository.getValue(e.getKey());
			// CorrectInGroup
			int correctInGroup = 0;
			for (SingleValue sv : e.getValue().getValues()) {
				Optional<Boolean> c = accuracyCalculator.correct(e.getKey().getValueId(), FusedValue.fromObject(sv.getValue()));

				if (c.isPresent() && c.get()) {
					correctInGroup++;
				}

				// Probability score =
				// scores.getScore(e.getKey().getInstanceId(),
				// e.getKey().getPropertyId(), sv.getId());
				// int fdf = 123;
			}

			line[8] = String.valueOf(correctInGroup);

			line[9] = dfmt.format(((double) e.getValue().getValues().size()) / ((double) correctInGroup));

			// // FilteredAS
			// if (fusedValues.containsKey(e.getKey())) {
			// line[8] = "0";
			// } else {
			// line[8] = "1";
			// }

			csvWriter.writeNext(line);

		}

		csvWriter.close();

	}

	public void printUniqueMatchedValues() {
		GroupedValues groupedValues = groupedValuesRepository.getGroupedValues();

		System.out.println("Unique matches: " + groupedValues.size());

	}
}
