package de.mannheim.uni.expansion.tablefusion.analysis;

import java.io.File;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.HashSet;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.FusionPerformanceSynchronizedCreator;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.generic.GenericResultsWriterAccumRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.services.grouping.GroupedValuesRepository;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.utils.DomainUtils;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumAverage;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumInteger;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AccumulatorListImmutable;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AcumReference;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.GenericAccumaltableResult;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AccumulatorListImmutable.ImmutableAccumulatorListBuild;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList.ColumnListBuilder;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.GenericAccumaltableResult.ResultBuilder;

public class GroupAnalyzerAggregated {

	private final GroupedValuesRepository groupedValuesRepository;

	private final AnnotatedTablesRepository annotatedTablesRepository;

	// private final FusionRunner fusionRunner;

	// private final ProbabilitiesCalculator probabilitiesCalculator;

	private final OntologyRepository ontologyRepository;

	private final KnowledgeBase dbBpediaRepository;

	private final ValueEvaluator accuracyCalculator;
	//
	// private final ScoreFusionThresholdFinder scoreFusionThresholdFinder;
	//
	// private final LearningSetsRepository learningSetsRepository;

	@Inject
	public GroupAnalyzerAggregated(GroupedValuesRepository groupedValuesRepository, AnnotatedTablesRepository annotatedTablesRepository,
			OntologyRepository ontologyRepository, KnowledgeBase dbBpediaRepository, ValueEvaluator accuracyCalculator) {
		super();
		this.groupedValuesRepository = groupedValuesRepository;
		this.annotatedTablesRepository = annotatedTablesRepository;
		// this.fusionRunner = fusionRunner;
		// this.probabilitiesCalculator = probabilitiesCalculator;
		this.ontologyRepository = ontologyRepository;
		this.dbBpediaRepository = dbBpediaRepository;
		this.accuracyCalculator = accuracyCalculator;
		// this.scoreFusionThresholdFinder = scoreFusionThresholdFinder;
		// this.learningSetsRepository = learningSetsRepository;
	}

	private DecimalFormat dfmt = new DecimalFormat("0.00");

	public void run(String outputFolder, String outputPrefix) throws IOException {

		GroupedValues groupedValues = groupedValuesRepository.getGroupedValues();

		ImmutableAccumulatorListBuild builder = AccumulatorListImmutable.builder();
		AcumReference<AccumAverage> averageNumberOfSources = builder.addAcummulator(AccumAverage.class, new AccumAverage());

		AcumReference<AccumAverage> averageNumberOfPLDs = builder.addAcummulator(AccumAverage.class, new AccumAverage());

		AcumReference<AccumInteger> correctOverlappingValues = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));

		AcumReference<AccumInteger> totalOverlappingValues = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));

		ColumnListBuilder columnListBuilder = ColumnList.builderAccumRepository();
		columnListBuilder.addColumn("AverageNumberOfSources", (repository) -> {
			return String.valueOf(repository.get(averageNumberOfSources).getValue());
		});

		columnListBuilder.addColumn("AverageNumberOfPLDs", (repository) -> {
			return String.valueOf(repository.get(averageNumberOfPLDs).getValue());
		});

		columnListBuilder.addColumn("CorrectOverlappingValues", (repository) -> {
			return String.valueOf(repository.get(correctOverlappingValues).get());
		});

		columnListBuilder.addColumn("TotalOverlappingValues", (repository) -> {
			return String.valueOf(repository.get(totalOverlappingValues).get());
		});

		columnListBuilder.addColumn("PercentCorrect", (repository) -> {
			double correct = repository.get(correctOverlappingValues).get();
			double total = repository.get(totalOverlappingValues).get();
			return String.valueOf(correct / total);
		});

		AccumulatorListImmutable resultsType = builder.getResultsType();
		ColumnList columnList = columnListBuilder.build();

		FusionPerformanceSynchronizedCreator<GenericAccumaltableResult> creatorSingle = new FusionPerformanceSynchronizedCreator<>(
				GenericAccumaltableResult.empty(resultsType));

		// Map<Integer, CreatedSets> createdSets =
		// learningSetsRepository.getCreatedSets(count,
		// groupedValues).getCreatedSets();
		//
		// double threshold =
		// scoreFusionThresholdFinder.findOptimumThresholdCrossValidation(createdSets,
		// FusionStrategyRepository.KBT, false);
		//
		// ScoresContainer scores =
		// probabilitiesCalculator.byProperty(groupedValues);
		//
		// Map<FusionRun, Map<TripleId, FusedValue>> fusedValuesMap =
		// fusionRunner.run(methodTypes, FusionPreproccesser.NONE,
		// FusionStrategyRepository.KBT,
		// groupedValues, scores, false, threshold);
		//
		// Map<TripleId, FusedValue> fusedValues =
		// CHMHelper.flattenMap(fusedValuesMap);

		System.out.println("Size of groupedvalues: " + groupedValues.size());

		for (Entry<TripleId, MatchedValues> e : groupedValues.entrySet()) {

			ResultBuilder valueBulder = GenericAccumaltableResult.builder(resultsType);

			// =====================================
			// Average Number Of Values per Group
			valueBulder.setValue(averageNumberOfSources, new AccumAverage((double) e.getValue().getValues().size()));

			// =====================================
			// Average PLDs per Group
			Set<String> uniquePLDs = new HashSet<>();
			for (SingleValue sv : e.getValue().getValues()) {
				String source = annotatedTablesRepository.getMetaData(sv.getId().getTableId()).getSource().getSource();

				String PLD = DomainUtils.getPayLevelDomainFromWholeURL(source);
				uniquePLDs.add(PLD);

			}

			valueBulder.setValue(averageNumberOfPLDs, new AccumAverage((double) uniquePLDs.size()));

			// =====================================
			// Total correct overlapping per Group
			int correctInGroup = 0;
			for (SingleValue sv : e.getValue().getValues()) {
				Optional<Boolean> c = accuracyCalculator.correct(e.getKey().getValueId(), FusedValue.fromObject(sv.getValue()));

				if (c.isPresent() && c.get()) {
					correctInGroup++;
				}

				// Probability score =
				// scores.getScore(e.getKey().getInstanceId(),
				// e.getKey().getPropertyId(), sv.getId());
				// int fdf = 123;
			}

			valueBulder.setValue(correctOverlappingValues, new AccumInteger(correctInGroup));

			// =====================================
			// Total overlapping per Group

			if (dbBpediaRepository.hasValue(e.getKey().getValueId())) {
				valueBulder.setValue(totalOverlappingValues, new AccumInteger(e.getValue().getValues().size()));
			} else {
				valueBulder.setValue(totalOverlappingValues, new AccumInteger(0));
			}

			creatorSingle.addForAll(e.getKey().getClassPropertyId(), ontologyRepository, valueBulder.build());
		}

		GenericResultsWriterAccumRepository writer = new GenericResultsWriterAccumRepository(columnList);

		writer.writeAccumRepository("GroupsAggregated", ontologyRepository, creatorSingle.get());

		try {
			if (outputFolder != null && outputPrefix != null) {
				MKDIR.createDirUnchecked(new File(outputFolder));
				writer.writeToFolder(outputFolder, outputPrefix);

			}
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

	}

	public void printUniqueMatchedValues() {
		GroupedValues groupedValues = groupedValuesRepository.getGroupedValues();

		System.out.println("Unique matches: " + groupedValues.size());

	}
}
