package de.mannheim.uni.expansion.tablefusion.analysis.stores;

import java.util.Collections;
import java.util.EnumMap;
import java.util.Map;
import java.util.Map.Entry;

import de.mannheim.uni.expansion.tablefusion.analysis.Accuracy;
import de.mannheim.uni.expansion.tablefusion.core.similarity.Similarity.SimilarityType;

public class FusedResults {

	private final Map<SimilarityType, PerformancePR> map;

	public FusedResults(Map<SimilarityType, PerformancePR> map) {
		Map<SimilarityType, PerformancePR> mapTemp = new EnumMap<>(SimilarityType.class);
		mapTemp.putAll(map);
		this.map = Collections.unmodifiableMap(mapTemp);
	}

	public Double getAccuracy(SimilarityType type) {
		return map.get(type).getPrecision();
	}

	public Map<SimilarityType, Accuracy> getAccuracies() {
		Map<SimilarityType, Accuracy> accuracies = new EnumMap<>(SimilarityType.class);
		for (Entry<SimilarityType, PerformancePR> e : map.entrySet()) {
			accuracies.put(e.getKey(), new Accuracy(e.getValue().getPrecision(), e.getValue().getOverlapping()));
		}
		return accuracies;
	}

	public Double getAccuracy() {
		double correctTriples = 0.0;
		double overlappingTriples = 0.0;
		for (Entry<SimilarityType, PerformancePR> e : map.entrySet()) {
			correctTriples += e.getValue().getCorrect();
			overlappingTriples += e.getValue().getOverlapping();
		}
		return correctTriples / overlappingTriples;

	}

	// public synchronized void incWrongTriples(SimilarityType type) {
	// map.get(type).incWrongTriples();
	// }
	//
	// public synchronized void incNewTriples(SimilarityType type) {
	// newTriples++;
	// }
	//
	// public synchronized void incCorrectTriples(SimilarityType type) {
	// map.get(type).incCorrectTriples();
	// }

	public int getWrongTriples(SimilarityType type) {
		return map.get(type).getWrong();
	}

	public int getCorrectTriples(SimilarityType type) {
		return map.get(type).getCorrect();
	}

	public int getNewTriples(SimilarityType type) {
		return map.get(type).getNew();
	}

	public int getWrongTriples() {
		int wrongTriples = 0;
		for (Entry<SimilarityType, PerformancePR> e : map.entrySet()) {
			wrongTriples += e.getValue().getWrong();
		}
		return wrongTriples;
	}

	public int getCorrectTriples() {
		int correctTriples = 0;
		for (Entry<SimilarityType, PerformancePR> e : map.entrySet()) {
			correctTriples += e.getValue().getCorrect();
		}
		return correctTriples;
	}

	public int getTotalTriples() {
		int totalTriples = 0;
		for (Entry<SimilarityType, PerformancePR> e : map.entrySet()) {

			totalTriples += e.getValue().getTotal();
		}
		return totalTriples;
	}

	public int getOverlappingTriples(SimilarityType type) {
		return map.get(type).getOverlapping();
	}

	public int getOverlappingTriples() {
		int overlappingTriples = 0;
		for (Entry<SimilarityType, PerformancePR> e : map.entrySet()) {
			overlappingTriples += e.getValue().getOverlapping();
		}
		return overlappingTriples;
	}

}