package de.mannheim.uni.expansion.tablefusion.analysis.stores;

import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.Accumulateable;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentAccumulator;

public class FusionPerformanceSynchronizedCreator<E extends Accumulateable<E>> {
	private ConcurrentAccumulator<E> total;

	private final ConcurrentHashMap<Boolean, ConcurrentAccumulator<E>> perTimeDependence = new ConcurrentHashMap<>();
	private final ConcurrentHashMap<ValueType, ConcurrentAccumulator<E>> perDataType = new ConcurrentHashMap<>();
	private final ConcurrentHashMap<ClassId, ConcurrentAccumulator<E>> perClassId = new ConcurrentHashMap<>();
	private final ConcurrentHashMap<ClassPropertyId, ConcurrentAccumulator<E>> perProperty = new ConcurrentHashMap<>();
	private final ConcurrentHashMap<Pair<ClassId, String>, ConcurrentAccumulator<E>> perClassAndSet = new ConcurrentHashMap<>();

	private final E empty;

	public FusionPerformanceSynchronizedCreator(E empty) {
		this.empty = empty;
		total = new ConcurrentAccumulator<E>(empty);
	}

	public FusionPerformance<E> get() {

		Function<Entry<?, ConcurrentAccumulator<E>>, E> valueMapper = (e) -> {
			return e.getValue().get();
		};

		E total = this.total.get();
		Map<Boolean, E> perTimeDependence = this.perTimeDependence.entrySet().stream().collect(Collectors.toMap(Entry::getKey, valueMapper));
		Map<ValueType, E> perDataType = this.perDataType.entrySet().stream().collect(Collectors.toMap(Entry::getKey, valueMapper));
		Map<ClassId, E> perClassId = this.perClassId.entrySet().stream().collect(Collectors.toMap(Entry::getKey, valueMapper));
		Map<ClassPropertyId, E> perProperty = this.perProperty.entrySet().stream().collect(Collectors.toMap(Entry::getKey, valueMapper));
		Map<Pair<ClassId, String>, E> perClassIdAndSet = this.perClassAndSet.entrySet().stream().collect(Collectors.toMap(Entry::getKey, valueMapper));
		return new FusionPerformance<E>(total, perTimeDependence, perDataType, perClassId, perProperty, perClassIdAndSet);
	}

	public void addTotal(E val) {
		this.total.add(val);
	}

	public void addPerTimeDependence(boolean timeDepedence, E val) {
		CHMHelper.getFromCHM(this.perTimeDependence, timeDepedence ? Boolean.TRUE : Boolean.FALSE, () -> new ConcurrentAccumulator<>(empty)).add(val);
	}

	public void addPerDataType(ValueType key, E val) {
		CHMHelper.getFromCHM(this.perDataType, key, () -> new ConcurrentAccumulator<>(empty)).add(val);
	}

	public void addPerClassId(ClassId key, E val) {
		CHMHelper.getFromCHM(this.perClassId, key, () -> new ConcurrentAccumulator<>(empty)).add(val);
	}

	public void addPerClassIdAndSet(ClassId classId, String set, E val) {
		CHMHelper.getFromCHM(this.perClassAndSet, new Pair<>(classId, set), () -> new ConcurrentAccumulator<>(empty)).add(val);
	}

	public void addPerProperty(ClassPropertyId key, E val) {
		CHMHelper.getFromCHM(this.perProperty, key, () -> new ConcurrentAccumulator<>(empty)).add(val);
	}

	public void addForAll(ClassPropertyId cpid, OntologyRepository ontologyRepository, E val) {
		addTotal(val);
		addPerTimeDependence(ontologyRepository.isTimeDependent(cpid.getPropertyId()), val);
		addPerDataType(ontologyRepository.getDataType(cpid.getPropertyId()), val);
		addPerClassId(cpid.getClassId(), val);
		addPerProperty(cpid, val);
	}

}
