package de.mannheim.uni.expansion.tablefusion.analysis.stores;

import java.util.function.BinaryOperator;

import de.mannheim.uni.expansion.tablefusion.analysis.stores.performance.PERFMPerformanceI;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.Accumulateable;

public class PerformanceMacroPR implements PERFMPerformanceI, Accumulateable<PerformanceMacroPR> {

	private final int correct;
	private final int wrong;
	private final int maxCorrect;
	private final int notInGs;

	private final double precisionSum;
	private final double recallSum;
	private final double fMeasureSum;
	private final int count;

	private final int newValues;
	private final int valuesInKB;
	private final int allPropertiesValuesInKB;

	private final int mappings;
	private final int mappedInstances;
	private final int unmappedInstances;

	private final int filteredMappings;
	private final int filteredMappedInstances;
	private final int filteredUnmappedInstances;

	public PerformanceMacroPR() {
		this.newValues = 0;
		this.correct = 0;
		this.wrong = 0;
		this.maxCorrect = 0;
		this.precisionSum = 0.0;
		this.recallSum = 0.0;
		this.fMeasureSum = 0.0;
		this.count = 0;
		this.valuesInKB = 0;
		this.mappings = 0;
		this.filteredMappings = 0;
		this.mappedInstances = 0;
		this.filteredMappedInstances = 0;
		this.unmappedInstances = 0;
		this.filteredUnmappedInstances = 0;
		this.allPropertiesValuesInKB = 0;
		this.notInGs = 0;
	}

	public static PerformanceMacroPR withCorpusMappinsAndInstances(int mappings, int filteredMappings, int mappedInstances, int filteredMappedInstances,
			int unmappedInstances, int filteredUnmappedInstances) {
		return new PerformanceMacroPR(mappings, filteredMappings, mappedInstances, filteredMappedInstances, unmappedInstances, filteredUnmappedInstances, true);
	}

	public static PerformanceMacroPR allPropertiesValuesInKb(int allPropertiesValuesInKB) {
		return new PerformanceMacroPR(allPropertiesValuesInKB, true);
	}

	private PerformanceMacroPR(int allPropertiesValuesInKB, boolean allPropertiesValuesInKBBoolean) {
		this.newValues = 0;
		this.correct = 0;
		this.wrong = 0;
		this.maxCorrect = 0;
		this.notInGs = 0;
		this.precisionSum = 0.0;
		this.recallSum = 0.0;
		this.fMeasureSum = 0.0;
		this.count = 0;
		this.valuesInKB = 0;
		this.mappings = 0;
		this.filteredMappings = 0;
		this.mappedInstances = 0;
		this.filteredMappedInstances = 0;
		this.unmappedInstances = 0;
		this.filteredUnmappedInstances = 0;
		this.allPropertiesValuesInKB = allPropertiesValuesInKB;
	}

	private PerformanceMacroPR(int mappings, int filteredMappings, int mappedInstances, int filteredMappedInstances, int unmappedInstances,
			int filteredUnmappedInstances, boolean mappingAndInstances) {
		this.newValues = 0;
		this.correct = 0;
		this.wrong = 0;
		this.maxCorrect = 0;
		this.notInGs = 0;
		this.precisionSum = 0.0;
		this.recallSum = 0.0;
		this.fMeasureSum = 0.0;
		this.count = 0;
		this.valuesInKB = 0;
		this.mappings = mappings;
		this.filteredMappings = filteredMappings;
		this.mappedInstances = mappedInstances;
		this.filteredMappedInstances = filteredMappedInstances;
		this.unmappedInstances = unmappedInstances;
		this.filteredUnmappedInstances = filteredUnmappedInstances;
		this.allPropertiesValuesInKB = 0;
	}

	public PerformanceMacroPR(int newValues, int valuesInKB) {
		this.newValues = newValues;
		this.correct = 0;
		this.wrong = 0;
		this.maxCorrect = 0;
		this.notInGs = 0;
		this.precisionSum = 0.0;
		this.recallSum = 0.0;
		this.fMeasureSum = 0.0;
		this.count = 0;
		this.valuesInKB = valuesInKB;
		this.mappings = 0;
		this.filteredMappings = 0;
		this.mappedInstances = 0;
		this.filteredMappedInstances = 0;
		this.unmappedInstances = 0;
		this.filteredUnmappedInstances = 0;
		this.allPropertiesValuesInKB = 0;
	}

	public PerformanceMacroPR(int correct, int wrong, int maxCorrect, int notInGs) {
		super();

		this.correct = correct;
		this.wrong = wrong;
		this.maxCorrect = maxCorrect;
		this.notInGs = notInGs;
		double precision = Double.NaN;
		precision = ((double) correct) / (correct + wrong);
		if (Double.isNaN(precision)) {
			this.precisionSum = 0;
		} else {
			this.precisionSum = precision;
		}
		if (maxCorrect == 0) {
			this.recallSum = 1.0;
		} else {
			this.recallSum = ((double) correct) / (maxCorrect);
		}

		// this.aRecall = (this.recall + this.precision) / 2.0;

		double fMeasure = 2 * ((this.precisionSum * recallSum) / (this.precisionSum + recallSum));
		if (Double.isNaN(fMeasure))
			this.fMeasureSum = 0.0;
		else
			this.fMeasureSum = fMeasure;

		this.count = 1;
		this.newValues = 0;
		this.valuesInKB = 0;
		this.mappings = 0;
		this.filteredMappings = 0;
		this.mappedInstances = 0;
		this.filteredMappedInstances = 0;
		this.unmappedInstances = 0;
		this.filteredUnmappedInstances = 0;
		this.allPropertiesValuesInKB = 0;

	}

	public PerformanceMacroPR(int correct, int wrong, int maxCorrect) {
		super();

		this.correct = correct;
		this.wrong = wrong;
		this.maxCorrect = maxCorrect;
		this.notInGs = 0;
		double precision = Double.NaN;
		precision = ((double) correct) / (correct + wrong);
		if (Double.isNaN(precision)) {
			this.precisionSum = 0;
		} else {
			this.precisionSum = precision;
		}
		if (maxCorrect == 0) {
			this.recallSum = 1.0;
		} else {
			this.recallSum = ((double) correct) / (maxCorrect);
		}

		// this.aRecall = (this.recall + this.precision) / 2.0;

		double fMeasure = 2 * ((this.precisionSum * recallSum) / (this.precisionSum + recallSum));
		if (Double.isNaN(fMeasure))
			this.fMeasureSum = 0.0;
		else
			this.fMeasureSum = fMeasure;

		this.count = 1;
		this.newValues = 0;
		this.valuesInKB = 0;
		this.mappings = 0;
		this.filteredMappings = 0;
		this.mappedInstances = 0;
		this.filteredMappedInstances = 0;
		this.unmappedInstances = 0;
		this.filteredUnmappedInstances = 0;
		this.allPropertiesValuesInKB = 0;

	}

	public PerformanceMacroPR(int correct, int wrong, int maxCorrect, double precisionSum, double recallSum, double fMeasureSum, int count) {
		super();
		this.correct = correct;
		this.wrong = wrong;
		this.maxCorrect = maxCorrect;
		this.notInGs = 0;
		this.precisionSum = precisionSum;
		this.recallSum = recallSum;
		this.fMeasureSum = fMeasureSum;
		this.count = count;
		this.newValues = 0;
		this.valuesInKB = 0;
		this.mappings = 0;
		this.filteredMappings = 0;
		this.mappedInstances = 0;
		this.filteredMappedInstances = 0;
		this.unmappedInstances = 0;
		this.filteredUnmappedInstances = 0;
		this.allPropertiesValuesInKB = 0;
	}

	public PerformanceMacroPR(int correct, int wrong, int maxCorrect, double precision, double recall, double fMeasure) {
		this.correct = correct;
		this.wrong = wrong;
		this.maxCorrect = maxCorrect;
		this.notInGs = 0;
		this.precisionSum = precision;
		this.recallSum = recall;
		this.fMeasureSum = fMeasure;
		this.count = 1;
		this.newValues = 0;
		this.valuesInKB = 0;
		this.mappings = 0;
		this.filteredMappings = 0;
		this.mappedInstances = 0;
		this.filteredMappedInstances = 0;
		this.unmappedInstances = 0;
		this.filteredUnmappedInstances = 0;
		this.allPropertiesValuesInKB = 0;
	}

	private PerformanceMacroPR(int correct, int wrong, int maxCorrect, int notInGs, double precision, double recall, double fMeasure, int count, int newValues,
			int valuesInKB, int mappings, int filteredMappings, int mappedInstances, int filteredMappedInstances, int unmappedInstances,
			int filteredUnmappedInstances, int allPropertiesValuesInKB) {
		this.correct = correct;
		this.wrong = wrong;
		this.maxCorrect = maxCorrect;
		this.notInGs = notInGs;
		this.precisionSum = precision;
		this.recallSum = recall;
		this.fMeasureSum = fMeasure;
		this.count = count;
		this.newValues = newValues;
		this.valuesInKB = valuesInKB;
		this.mappings = mappings;
		this.filteredMappings = filteredMappings;
		this.mappedInstances = mappedInstances;
		this.filteredMappedInstances = filteredMappedInstances;
		this.unmappedInstances = unmappedInstances;
		this.filteredUnmappedInstances = filteredUnmappedInstances;
		this.allPropertiesValuesInKB = allPropertiesValuesInKB;
	}

	public int getCorrect() {
		return correct;
	}

	public int getWrong() {
		return wrong;
	}

	public int getMaxCorrect() {
		return maxCorrect;
	}

	public double getPrecisionMicro() {
		return ((double) correct) / (correct + wrong);
	}

	public double getRecallMicro() {
		if (maxCorrect == 0) {
			return 1.0;
		} else {
			return ((double) correct) / (maxCorrect);
		}
	}

	public double getFMeasureMicro() {
		double precision = getPrecisionMicro();
		double recall = getRecallMicro();
		double fMeasure = 2 * ((precision * recall) / (precision + recall));

		return fMeasure;
	}

	public double getPrecision() {
		if (count == 0)
			return 0.0;
		return precisionSum / count;
	}

	public double getRecall() {
		if (count == 0)
			return 0.0;
		return recallSum / count;
	}

	public double getfMeasure() {
		if (count == 0)
			return 0.0;
		return fMeasureSum / count;
	}

	public int getNewValues() {
		return newValues;
	}

	public int getValuesInKB() {
		return valuesInKB;
	}

	public double getIncrease() {
		return ((double) newValues) / valuesInKB;
	}

	public int getAllPropertiesValuesInKB() {
		return allPropertiesValuesInKB;
	}

	public double getAllPropertiesIncrease() {
		return ((double) newValues) / allPropertiesValuesInKB;
	}

	@Override
	public BinaryOperator<PerformanceMacroPR> getAccumulator() {
		return PerformanceMacroPR::accumulator;
	}

	public static PerformanceMacroPR accumulator(PerformanceMacroPR a, PerformanceMacroPR b) {

		return new PerformanceMacroPR(a.correct + b.correct, a.wrong + b.wrong, a.maxCorrect + b.maxCorrect, a.notInGs + b.notInGs,
				a.precisionSum + b.precisionSum, a.recallSum + b.recallSum, a.fMeasureSum + b.fMeasureSum, a.count + b.count, a.newValues + b.newValues,
				a.valuesInKB + b.valuesInKB, a.mappings + b.mappings, a.filteredMappings + b.filteredMappings, a.mappedInstances + b.mappedInstances,
				a.filteredMappedInstances + b.filteredMappedInstances, a.unmappedInstances + b.unmappedInstances,
				a.filteredUnmappedInstances + b.filteredUnmappedInstances, a.allPropertiesValuesInKB + b.allPropertiesValuesInKB);

	}

	@Override
	public PerformanceMacroPR accumulate(PerformanceMacroPR other) {
		if (other instanceof PerformanceMacroPR) {
			PerformanceMacroPR oInt = (PerformanceMacroPR) other;
			return accumulator(this, oInt);
		} else {
			throw new IllegalStateException("This is not the same Accumateable Class");
		}
	}

	// @Override
	// public PerformanceMacroPR copy() {
	// return new PerformanceMacroPR(correct, wrong, maxCorrect, precisionSum,
	// recallSum, fMeasureSum, count, newValues, valuesInKB);
	// }

	@Override
	public Class<PerformanceMacroPR> getAcumClass() {
		return PerformanceMacroPR.class;
	}

	public int getMappings() {
		return mappings;
	}

	public int getFilteredMappings() {
		return filteredMappings;
	}

	public int getMappedInstances() {
		return mappedInstances;
	}

	public int getUnmappedInstances() {
		return unmappedInstances;
	}

	public int getFilteredMappedInstances() {
		return filteredMappedInstances;
	}

	public int getFilteredUnmappedInstances() {
		return filteredUnmappedInstances;
	}

	public static PerformanceMacroPR empty() {
		return new PerformanceMacroPR();
	}

	public int getNotInGs() {
		return notInGs;
	}

}
