package de.mannheim.uni.expansion.tablefusion.analysis.stores;

import java.util.List;

import de.mannheim.uni.expansion.tablefusion.analysis.Accuracy;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.performance.PERFMPerformanceI;

public class PerformancePR implements PERFMPerformanceI {

	private final int correct;
	private final int wrong;
	private final int maxCorrect;

	private final double precision;
	private final double recall;
	private final double fMeasure;
	// private final double aRecall;
	private final double aFMeasure;

	private final int newTriples;

	public static PerformancePR fromList(List<PerformancePR> list) {
		int correct = 0;
		int wrong = 0;
		int maxCorrect = 0;
		int newTriples = 0;
		for (PerformancePR ppr : list) {
			correct += ppr.getCorrect();
			wrong += ppr.getWrong();
			maxCorrect += ppr.getMaxCorrect();
			newTriples += ppr.getNew();
		}
		return new PerformancePR(correct, wrong, maxCorrect, newTriples);

	}

	@Deprecated
	public PerformancePR(List<PerformancePR> list) {
		int correct = 0;
		int wrong = 0;
		int maxCorrect = 0;
		int newTriples = 0;

		double precision = 0.0;
		double recall = 0.0;
		double fMeasure = 0.0;
		// double aRecall = 0.0;
		double aFMeasure = 0.0;

		for (PerformancePR ppr : list) {
			correct += ppr.getCorrect();
			wrong += ppr.getWrong();
			maxCorrect += ppr.getMaxCorrect();
			newTriples += ppr.getNew();

			precision += ppr.getPrecision();
			recall += ppr.getRecall();
			fMeasure += ppr.getfMeasure();
			// aRecall += ppr.getaRecall();
			aFMeasure += ppr.getaFMeasure();
		}

		this.correct = correct;
		this.wrong = wrong;
		this.maxCorrect = maxCorrect;
		this.newTriples = newTriples;
		this.precision = precision / list.size();
		this.recall = recall / list.size();
		this.fMeasure = fMeasure / list.size();
		// this.aRecall = aRecall / list.size();
		this.aFMeasure = aFMeasure / list.size();
	}

	public PerformancePR(int correct, int wrong, int maxCorrect, int newTriples) {
		super();

		this.correct = correct;
		this.wrong = wrong;
		this.maxCorrect = maxCorrect;
		double precision = Double.NaN;
		precision = ((double) correct) / (correct + wrong);
		if (Double.isNaN(precision)) {
			this.precision = 0;
		} else {
			this.precision = precision;
		}
		if (maxCorrect == 0) {
			this.recall = 1.0;
		} else {
			this.recall = ((double) correct) / (maxCorrect);
		}

		// this.aRecall = (this.recall + this.precision) / 2.0;

		double fMeasure = 2 * ((this.precision * recall) / (this.precision + recall));
		if (Double.isNaN(fMeasure))
			this.fMeasure = 0.0;
		else
			this.fMeasure = fMeasure;

		double aFMeasure = (1 + 0.0625) * ((this.precision * recall) / (0.0625 * this.precision + recall));

		if (Double.isNaN(aFMeasure))
			this.aFMeasure = 0.0;
		else
			this.aFMeasure = aFMeasure;

		this.newTriples = newTriples;

	}

	public int getCorrect() {
		return correct;
	}

	public int getWrong() {
		return wrong;
	}

	public int getOverlapping() {
		return wrong + correct;
	}

	public int getMaxCorrect() {
		return maxCorrect;
	}

	public double getPrecision() {
		return precision;
	}

	public double getRecall() {
		return recall;
	}

	public double getfMeasure() {
		return fMeasure;
	}

	public double getaFMeasure() {
		return aFMeasure;
	}

	public int getNew() {
		return this.newTriples;
	}

	@Override
	public String toString() {
		return "PerformancePR [correct=" + correct + ", wrong=" + wrong + ", maxCorrect=" + maxCorrect + ", precision=" + precision + ", recall=" + recall
				+ ", fMeasure=" + fMeasure + ", aFMeasure=" + aFMeasure + ", newTriples=" + newTriples + "]";
	}

	public int getTotal() {
		return wrong + correct + newTriples;
	}

	public Accuracy getAccuracy() {
		return new Accuracy(this.precision, this.getOverlapping());
	}

}
