package de.mannheim.uni.expansion.tablefusion.analysis.writer;

import java.io.IOException;
import java.util.Map;
import java.util.Map.Entry;
import java.util.SortedSet;
import java.util.TreeSet;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.PRFMPerformance;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.PerformancePR;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.performance.PERFMPerformanceI;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.qualifiers.TimestampPointInTime;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;

public class PerformanceResultsWriter extends AbstractResultsWriter {

	public PerformanceResultsWriter() {
		super();
	}

	@Override
	protected String[] getHeaderLine() {
		String[] headerLine = new String[8];
		headerLine[0] = "Correct";
		headerLine[1] = "Wrong";
		headerLine[2] = "MaxCorrect";
		headerLine[3] = "EPercision";
		headerLine[4] = "ERecall";
		headerLine[5] = "EFMeasure";
		headerLine[6] = "AFMeasure";
		headerLine[7] = "New";
		return headerLine;
	}

	public void write(ResultsCategory rc, String[] prefix, PerformancePR performancePR, PERFMPerformanceI prfmPerformance) {

		if (prfmPerformance == null) {
			prfmPerformance = new PRFMPerformance(0, 0, 0);
		}
		String[] line = new String[8];
		line[0] = String.valueOf(performancePR.getCorrect());
		line[1] = String.valueOf(performancePR.getWrong());
		line[2] = String.valueOf(performancePR.getMaxCorrect());
		line[3] = String.valueOf(prfmPerformance.getPrecision());
		line[4] = String.valueOf(prfmPerformance.getRecall());
		line[5] = String.valueOf(prfmPerformance.getfMeasure());
		if (prfmPerformance instanceof PerformancePR) {
			PerformancePR aPerformance = (PerformancePR) prfmPerformance;
			line[6] = String.valueOf(aPerformance.getaFMeasure());
		} else {
			line[6] = "ERR";
		}
		line[7] = String.valueOf(performancePR.getNew());

		write(rc, prefix, line);
	}

	public <T extends PERFMPerformanceI> void writePerformance(OntologyRepository ontologyRepository, String dataset, String name,
			FusionPerformance<PerformancePR> performance, FusionPerformance<T> averagePerformance, boolean printTotal, boolean printDataType,
			boolean printClassId, boolean printClassPropertyId) throws IOException {

		{
			int index = this.getCurrentIndex(ResultsCategory.TOTAL);
			String[] prefix = ResultsCategory.getNormalPrefix(name);
			this.write(ResultsCategory.TOTAL, prefix, performance.getTotal(), averagePerformance.getTotal());
			// if (printTotal) {
			System.out.println(dataset);
			System.out.println(this.getFrom(ResultsCategory.TOTAL, index));
			// }
		}

		{
			int index = this.getCurrentIndex(ResultsCategory.TIME_DEPENDENCE);

			Map<Boolean, T> averageAccuracies = averagePerformance.getPerTimeDependence();
			for (Entry<Boolean, PerformancePR> e : performance.getPerTimeDependence().entrySet()) {
				String[] prefix = ResultsCategory.getTimeDepedence(name, e.getKey());
				this.write(ResultsCategory.TIME_DEPENDENCE, prefix, e.getValue(), averageAccuracies.get(e.getKey()));
			}

			if (printDataType) {
				// System.out.println(performanceResultsWriter.getFrom(ResultsCategory.TIME_DEPENDENCE,
				// index));
			}
		}

		{
			int index = this.getCurrentIndex(ResultsCategory.YEAR);

			Map<TimestampPointInTime, T> averageAccuracies = averagePerformance.getPerYear();
			for (Entry<TimestampPointInTime, PerformancePR> e : performance.getPerYear().entrySet()) {
				String[] prefix = ResultsCategory.getYear(name, e.getKey());
				this.write(ResultsCategory.YEAR, prefix, e.getValue(), averageAccuracies.get(e.getKey()));
			}

			if (printDataType) {
				// System.out.println(performanceResultsWriter.getFrom(ResultsCategory.TIME_DEPENDENCE,
				// index));
			}
		}

		{
			int index = this.getCurrentIndex(ResultsCategory.CLASS_PROPERTY_WITH_YEAR);

			Map<Pair<ClassPropertyId, TimestampPointInTime>, T> averageAccuracies = averagePerformance.getPerClassPropertyAndYear();
			for (Entry<Pair<ClassPropertyId, TimestampPointInTime>, PerformancePR> e : performance.getPerClassPropertyAndYear().entrySet()) {
				String[] prefix = ResultsCategory.getClassPropertyIdWithYearPrefix(name, e.getKey().getSecond(), e.getKey().getFirst());
				this.write(ResultsCategory.CLASS_PROPERTY_WITH_YEAR, prefix, e.getValue(), averageAccuracies.get(e.getKey()));
			}

			if (printDataType) {
				// System.out.println(performanceResultsWriter.getFrom(ResultsCategory.TIME_DEPENDENCE,
				// index));
			}
		}

		{
			int index = this.getCurrentIndex(ResultsCategory.DATA_TYPE);

			Map<ValueType, T> averageAccuracies = averagePerformance.getPerDataType();
			for (Entry<ValueType, PerformancePR> e : performance.getPerDataType().entrySet()) {
				String[] prefix = ResultsCategory.getDataTypePrefix(name, e.getKey());
				this.write(ResultsCategory.DATA_TYPE, prefix, e.getValue(), averageAccuracies.get(e.getKey()));
			}

			// if (printDataType) {
			// System.out.println(performanceResultsWriter.getFrom(ResultsCategory.DATA_TYPE,
			// index));
			// }
		}

		{

			Map<ClassId, T> averageAccuracies = averagePerformance.getPerClassId();
			for (Entry<ClassId, PerformancePR> e : performance.getPerClassId().entrySet()) {
				String[] prefix = ResultsCategory.getClassIdPrefix(name, e.getKey());
				this.write(ResultsCategory.CLASS_ID, prefix, e.getValue(), averageAccuracies.get(e.getKey()));
			}

		}

		{

			int index = this.getCurrentIndex(ResultsCategory.CLASS_PROPERTY_ID);

			Map<ClassPropertyId, T> averageAccuracies = averagePerformance.getPerProperty();
			Map<ClassPropertyId, PerformancePR> performancePerProperty = performance.getPerProperty();
			SortedSet<ClassPropertyId> sortedSet = new TreeSet<>();
			sortedSet.addAll(performance.getPerProperty().keySet());
			sortedSet.addAll(averageAccuracies.keySet());
			for (ClassPropertyId cpid : sortedSet) {

				ValueType cdt = ontologyRepository.getDataType(cpid.getPropertyId());
				String[] prefix = ResultsCategory.getClassPropertyIdPrefix(name, ontologyRepository.isTimeDependent(cpid.getPropertyId()), cdt, cpid);
				this.write(ResultsCategory.CLASS_PROPERTY_ID, prefix, performancePerProperty.get(cpid), averageAccuracies.get(cpid));
			}
			// System.out.println(performanceResultsWriter.getFrom(ResultsCategory.CLASS_PROPERTY_ID,
			// index));

		}

		{

			int index = this.getCurrentIndex(ResultsCategory.CLASS_PROPERTY_ID_WITH_DATASET);

			Map<ClassPropertyId, T> averageAccuracies = averagePerformance.getPerProperty();
			Map<ClassPropertyId, PerformancePR> performancePerProperty = performance.getPerProperty();
			SortedSet<ClassPropertyId> sortedSet = new TreeSet<>();
			sortedSet.addAll(performance.getPerProperty().keySet());
			sortedSet.addAll(averageAccuracies.keySet());
			for (ClassPropertyId cpid : sortedSet) {
				ValueType cdt = ontologyRepository.getDataType(cpid.getPropertyId());
				String[] prefix = ResultsCategory.getClassPropertyIdWithDatasetPrefix(dataset, name, ontologyRepository.isTimeDependent(cpid.getPropertyId()),
						cdt, cpid);
				this.write(ResultsCategory.CLASS_PROPERTY_ID_WITH_DATASET, prefix, performancePerProperty.get(cpid), averageAccuracies.get(cpid));
			}
			// System.out.println(performanceResultsWriter.getFrom(ResultsCategory.CLASS_PROPERTY_ID_WITH_DATASET,
			// index));

		}

	}

}
