package de.mannheim.uni.expansion.tablefusion.analysis.writer;

import java.io.IOException;
import java.io.StringWriter;
import java.util.ArrayList;

import com.opencsv.CSVWriter;

class PrefixCSVWriter {

	private final ArrayList<String[]> lines = new ArrayList<>();

	public PrefixCSVWriter(String[] prefix, String[] header) {

		String[] headerLine = new String[prefix.length + header.length];
		for (int i = 0; i < prefix.length; i++) {
			headerLine[i] = prefix[i];
		}

		for (int i = 0; i < header.length; i++) {
			headerLine[prefix.length + i] = header[i];

		}

		lines.add(headerLine);
	}

	public void writeLine(String[] prefix, String[] content) {

		String[] line = new String[prefix.length + content.length];
		for (int i = 0; i < prefix.length; i++) {
			line[i] = prefix[i];
		}

		for (int i = 0; i < content.length; i++) {
			line[prefix.length + i] = content[i];

		}
		lines.add(line);
	}

	public String getAll() throws IOException {
		StringWriter sWriter = new StringWriter();
		CSVWriter csvWriter = new CSVWriter(sWriter);
		try {
			csvWriter.writeAll(lines);

			return sWriter.toString();
		} finally {
			csvWriter.close();
		}
	}

	public String getNoHeader(int index) throws IOException {
		StringWriter sWriter = new StringWriter();
		CSVWriter csvWriter = new CSVWriter(sWriter);
		try {
			for (int i = index + 1; i < lines.size(); i++) {
				csvWriter.writeNext(lines.get(i));
			}

			return sWriter.toString();
		} finally {
			csvWriter.close();
		}
	}

	public String getFrom(int index) throws IOException {
		StringWriter sWriter = new StringWriter();
		CSVWriter csvWriter = new CSVWriter(sWriter);
		try {
			csvWriter.writeNext(lines.get(0));
			for (int i = index + 1; i < lines.size(); i++) {
				csvWriter.writeNext(lines.get(i));
			}

			return sWriter.toString();
		} finally {
			csvWriter.close();
		}
	}

	public int getCurrentIndex() {
		return lines.size() - 1;
	}
}