package de.mannheim.uni.expansion.tablefusion.analysis.writer;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.tablefusion.core.similarity.Similarity.SimilarityType;
import de.mannheim.uni.tableprocessor.model.TableColumn.ColumnDataType;
import de.uni_mannheim.dws.expansion.data.ontology.model.qualifiers.TimestampPointInTime;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public enum ResultsCategory {
	TOTAL(new String[] { "Run" }),
	TOTAL_WITH_LEARNER(new String[] { "Strategy", "Learner", }),
	CLASS_ID_AND_SET(new String[] { "Run", "ClassId", "Set" }),
	TIME_DEPENDENCE(new String[] { "Run", "TimeDependent" }),
	YEAR(new String[] { "Run", "Year" }),
	DATA_TYPE(new String[] { "Run", "DataType" }),
	SIMILARITY_TYPE_PREFIX(new String[] { "Run", "DataType", "SimilarityType" }),
	CLASS_ID(new String[] { "Run", "ClassId" }),
	CLASS_PROPERTY_ID_SIMPLE(new String[] { "Run", "ClassId", "PropertyId" }),
	CLASS_PROPERTY_ID(new String[] { "Run", "TimeDependent", "DataType", "ClassId", "PropertyId" }),
	CLASS_PROPERTY_WITH_YEAR(new String[] { "Run", "Year", "ClassId", "PropertyId" }),
	CLASS_PROPERTY_ID_WITH_DATASET(new String[] { "DataSet", "Run", "TimeDependent", "DataType", "ClassId", "PropertyId" }),
	CLASS_ID_NO_SET(new String[] { "ClassId" }),
	CLASS_ID_WITH_LEARNER(new String[] { "Strategy", "Learner", "ClassId" });

	private final String[] prefix;

	private ResultsCategory(String[] prefix) {
		this.prefix = prefix;
	}

	public String[] getPrefix() {
		return prefix;
	}

	public static String[] getNormalPrefix(String run) {
		return new String[] { run };
	}

	public static String[] getTotalWithLearnerPrefix(String run, String learner) {
		return new String[] { run, learner };
	}

	public static String[] getYear(String run, TimestampPointInTime timestampPointInTime) {
		return new String[] { run, String.valueOf(timestampPointInTime.getYear()) };
	}

	public static String[] getTimeDepedence(String run, boolean timeDependence) {
		return new String[] { run, timeDependence ? "Yes" : "No" };
	}

	public static String[] getDataTypePrefix(String run, ValueType cdt) {
		return new String[] { run, cdt.toString() };

	}

	public static String[] getSimilarityTypePrefix(String run, ColumnDataType cdt, SimilarityType simType) {
		return new String[] { run, cdt.toString(), simType.toString() };

	}

	public static String[] getClassIdNoSetPrefix(ClassId classId) {
		return new String[] { classId.getFinalPart() };
	}

	public static String[] getClassIdWithLearnerPrefix(String run, String learner, ClassId classId) {
		return new String[] { run, learner, classId.getFinalPart() };
	}

	public static String[] getClassIdPrefix(String run, ClassId classId) {
		return new String[] { run, classId.getFinalPart() };
	}

	public static String[] getClassIdAndSetPrefix(String run, ClassId classId, String set) {
		return new String[] { run, classId.getFinalPart(), set };
	}

	public static String[] getClassPropertyIdWithYearPrefix(String run, TimestampPointInTime timestampPointInTime, ClassPropertyId classPropertyId) {
		return new String[] { run, String.valueOf(timestampPointInTime.getYear()), classPropertyId.getClassId().getFinalPart(),
				classPropertyId.getPropertyId().toString() };
	}

	public static String[] getClassPropertyIdSimplePrefix(String run, ClassPropertyId classPropertyId) {
		return new String[] { run, classPropertyId.getClassId().getFinalPart(), classPropertyId.getPropertyId().toString() };
	}

	public static String[] getClassPropertyIdPrefix(String run, boolean timeDependence, ValueType cdt, ClassPropertyId classPropertyId) {
		return new String[] { run, timeDependence ? "Yes" : "No", cdt.toString(), classPropertyId.getClassId().getFinalPart(),
				classPropertyId.getPropertyId().toString() };
	}

	public static String[] getClassPropertyIdWithDatasetPrefix(String dataset, String run, boolean timeDependence, ValueType cdt,
			ClassPropertyId classPropertyId) {
		return new String[] { dataset, run, timeDependence ? "Yes" : "No", cdt.toString(), classPropertyId.getClassId().getFinalPart(),
				classPropertyId.getPropertyId().toString() };
	}
}
