package de.mannheim.uni.expansion.tablefusion.core.fusing;

import java.util.ArrayList;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.inject.Inject;

import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityMeasure;
import de.mannheim.uni.expansion.utils.ValueProbabilityPair;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valuepairs.AbstractClassPair;

public class SlowClusterFuser {

	@Inject
	public SlowClusterFuser() {
		// this.sc = sc;

		// SimilarityMeasure sm = this.sc.getMeasure(values.getDataType());
	}

	public SortedSet<Cluster> fuse(List<ClusterValue> values, BaseFuser simpleFuser, SimilarityMeasure sm) {
		List<Cluster> clusters = new ArrayList<Cluster>();

		for (ClusterValue cv : values) {
			clusters.add(new Cluster(cv));

		}

		List<Cluster> mergedClusters = mergeClusters(clusters, sm, sm.getEquivalenceThreshold(), 0, simpleFuser, 0.99);

		return new TreeSet<Cluster>(mergedClusters);
	}

	public List<Cluster> mergeClusters(List<Cluster> clusters, SimilarityMeasure sm, double threshold, int i, BaseFuser simpleFuser, double lastMax) {

		if (clusters.size() == 1)
			return clusters;

		if (i > 50) {
			System.out.println("Reached 50 Iterations. Stopping!");
			return clusters;
		}

		// boolean continueClustering = false;

		List<Cluster> newClusters = new ArrayList<Cluster>();

		double maxSim = 0.0;
		int ci1 = 0;
		int ci2 = 0;
		for (int i1 = 0; i1 < clusters.size(); i1++) {
			for (int i2 = 0; i2 < clusters.size(); i2++) {
				if (i1 != i2) {

					double sim = clusters.get(i1).getSimilarity(clusters.get(i2), sm, simpleFuser);
					if (sim > maxSim) {
						maxSim = sim;
						ci1 = i1;
						ci2 = i2;
					}

				}
			}
		}

		if (maxSim > threshold) {
			Cluster c1;
			Cluster c2;
			if (ci1 > ci2) {
				c1 = clusters.remove(ci1);
				c2 = clusters.remove(ci2);
			} else {
				c2 = clusters.remove(ci2);
				c1 = clusters.remove(ci1);
			}

			newClusters.add(new Cluster(c1, c2, simpleFuser));
			newClusters.addAll(clusters);
			return mergeClusters(newClusters, sm, threshold, i + 1, simpleFuser, maxSim);
		}

		else {
			return clusters;
		}

		// outerloop: while (true) {
		// double maxSim = 0.0;
		// int ci1 = 0;
		// int ci2 = 0;
		// for (int i1 = 0; i1 < clusters.size(); i1++) {
		// for (int i2 = 0; i2 < clusters.size(); i2++) {
		// if (i1 != i2) {
		//
		// double sim = clusters.get(i1).getSimilarity(clusters.get(i2), sm,
		// simpleFuser);
		// if (sim > lastMax) {
		// Cluster c1;
		// Cluster c2;
		// if (i1 > i2) {
		// c1 = clusters.remove(i1);
		// c2 = clusters.remove(i2);
		// } else {
		// c2 = clusters.remove(i2);
		// c1 = clusters.remove(i1);
		// }
		//
		// clusters.add(new Cluster(c1, c2, simpleFuser));
		// continueClustering = true;
		// continue outerloop;
		// } else if (sim > maxSim) {
		// maxSim = sim;
		// ci1 = i1;
		// ci2 = i2;
		// }
		//
		// }
		// }
		// }
		//
		// if (maxSim > threshold) {
		// Cluster c1;
		// Cluster c2;
		// if (ci1 > ci2) {
		// c1 = clusters.remove(ci1);
		// c2 = clusters.remove(ci2);
		// } else {
		// c2 = clusters.remove(ci2);
		// c1 = clusters.remove(ci1);
		// }
		//
		// newClusters.add(new Cluster(c1, c2, simpleFuser));
		// return mergeClusters(newClusters, sm, threshold, i + 1, simpleFuser,
		// maxSim);
		// }
		//
		// else if (continueClustering) {
		//
		// return mergeClusters(newClusters, sm, threshold, i + 1, simpleFuser,
		// lastMax);
		// } else {
		// return clusters;
		// }
		//
		// }

	}

	public static class Cluster implements ValueProbabilityPair {

		private final List<ClusterValue> values;

		private final AbstractObject fused;

		private final Double score;

		public Double getScore() {
			return score;
		}

		private Cluster(Cluster c1, Cluster c2, BaseFuser simpleFuser) {
			throw new IllegalStateException();
			// this.values = new ArrayList<ClusterValue>();
			// values.addAll(c1.values);
			// values.addAll(c2.values);
			// fused = simpleFuser.fuse(values, null).getSingle();
			// score = values.stream().map((e) -> {
			// return e.getScore();
			// }).mapToDouble(Double::doubleValue).sum();
		}

		private Cluster(ClusterValue cv) {
			this.values = new ArrayList<ClusterValue>();
			values.add(cv);
			this.fused = cv.getValue();
			this.score = cv.getScore();

		}

		public double getSimilarity(Cluster c, SimilarityMeasure sm, BaseFuser simpleFuser) {

			return sm.getQuickSimilarity(fused, c.fused);

		}

		public int size() {
			return values.size();
		}

		public String toString() {
			return values.toString();
		}

		@Override
		public int compareTo(ValueProbabilityPair o) {
			return Double.compare(score, o.getScore());
		}

		@Override
		public Object getValue() {
			return fused;
		}

	}

	public static class ClusterValue implements AbstractClassPair {

		private final Double score;
		private final AbstractObject value;

		public ClusterValue(AbstractObject value, Double score) {
			this.score = score;
			this.value = value;

		}

		public AbstractObject getValue() {
			return value;
		}

		public Double getScore() {
			return score;
		}

		@Override
		public String toString() {
			return "ClusterValue [score=" + score + ", value=" + value + "]";
		}

		@Override
		public int compareTo(AbstractClassPair o) {
			return Double.compare(score, o.getScore());
		}

	}

}
