package de.mannheim.uni.expansion.tablefusion.core.fusing.impl.generic;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.function.Function;
import java.util.stream.Collectors;

import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.tablefusion.core.fusing.BaseFuser;
import de.mannheim.uni.expansion.tablefusion.core.fusing.ClusterFuser;
import de.mannheim.uni.expansion.tablefusion.core.fusing.ClusterFuser.Cluster;
import de.mannheim.uni.expansion.tablefusion.core.fusing.ClusterFuser.ClusterValue;
import de.mannheim.uni.expansion.tablefusion.core.fusing.Fuser;
import de.mannheim.uni.expansion.tablefusion.core.fusing.FusionScoreThresholdAdapter;
import de.mannheim.uni.expansion.tablefusion.core.fusing.ListFuser;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.TableScores;
import de.mannheim.uni.expansion.tablefusion.core.similarity.Similarity.SimilarityType;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityMeasure;
import de.mannheim.uni.expansion.utils.OtsuThresholder;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valuepairs.AbstractClassPair;
import de.uni_mannheim.dws.expansion.data.ontology.model.valuepairs.ScoreSortedAbstractPair;

public class AbstractScoreFuser implements Fuser, ListFuser {
	private final SimilarityCalculator sc;
	private final OtsuThresholder otsuThresholder;
	private final ClusterFuser clusterFuser;
	private final BaseFuser baseFuser;

	public AbstractScoreFuser(SimilarityCalculator sc, OtsuThresholder otsuThresholder, ClusterFuser clusterFuser, BaseFuser simpleFuser) {
		this.sc = sc;
		this.otsuThresholder = otsuThresholder;
		this.clusterFuser = clusterFuser;
		this.baseFuser = simpleFuser;
	}

	@Override
	public FusedValue fuse(MatchedValues values, TableScores probabilities, double threshold, FusionScoreThresholdAdapter adapter, boolean onePerPLD,
			Function<AnnotatedTableId, String> pldGetter) {

		List<AbstractClassPair> list = new ArrayList<>();
		//
		// List<SingleValue> sortedValues =
		// values.getValues().stream().sorted((e1, e2) -> {
		// Probability p1 = probabilities.get(e1.getId());
		// Probability p2 = probabilities.get(e2.getId());
		// Double d1 =
		// probabilities.get(e1.getId()).get(SimilarityType.DIRECT).getAccuracy();
		// Double d2 =
		// probabilities.get(e2.getId()).get(SimilarityType.DIRECT).getAccuracy();
		// return Double.compare(d1, d2);
		// }).collect(Collectors.toList());
		//
		// double perValue = 1.0 / sortedValues.size();

		// double max = 0.0;
		// double min = 1.0;
		// if (values.getValues().size() > 1) {
		// for (SingleValue e : values.getValues()) {
		// Probability p = probabilities.get(e.getId());
		// if (p != null) {
		// Double probability = p.get(SimilarityType.DIRECT).getAccuracy();
		// if (probability > max) {
		// max = probability;
		// }
		//
		// if (probability < min) {
		// min = probability;
		// }
		// }
		// }
		// }
		//
		// if (min > max) {
		// max = 1.0;
		// min = 0.0;
		// }
		// int maxTotal = 0;
		// for (SingleValue e : values.getValues()) {
		// Probability p = probabilities.get(e.getId());
		// if (p != null) {
		// int total = p.get(SimilarityType.DIRECT).getTotal();
		// if (total > maxTotal)
		// maxTotal = total;
		// }
		// }
		int i = 1;

		Map<String, AbstractClassPair> pldMap = new HashMap<>();
		for (SingleValue e : values.getValues()) {
			Probability p = probabilities.get(e.getId());
			if (p != null) {
				Double probability = p.get(SimilarityType.DIRECT).getAccuracy();
				if (probability == null || probability.isNaN()) {
					probability = 0.0;

				}

				// else {
				// Accuracy acc = p.get(SimilarityType.DIRECT);
				// Accuracy harmonic = Accuracy.harmonicMean(acc,
				// Accuracy.fromCorrectAndTotal(acc.getCorrect(), maxTotal),
				// acc.getTotal());
				// probability = harmonic.getAccuracy();
				// }

				// else {
				// probability = (probability - min) / (diff);
				// }
				// if (probability < 0.1) {
				// probability = 0.1;
				// }

				// probability = i * perValue;

				if (probability > (threshold)) {
					AbstractObject o = e.getValue();
					ScoreSortedAbstractPair pair = new ScoreSortedAbstractPair(o, probability);

					if (onePerPLD) {

						String PLD = pldGetter.apply(e.getId().getTableId());
						if (pldMap.containsKey(PLD)) {
							if (pair.getScore() > pldMap.get(PLD).getScore()) {
								pldMap.put(PLD, pair);
							}
						} else {
							pldMap.put(PLD, pair);
						}
					} else {
						list.add(pair);
					}

				} else {
					int fdfd = 234324;
				}

			} else {
				int wfsdf = 5;
			}
			i++;
		}

		if (onePerPLD) {
			for (Entry<String, AbstractClassPair> entry : pldMap.entrySet()) {
				list.add(entry.getValue());
			}
		}
		FusedValue fv = baseFuser.fuse(list, threshold, adapter);
		if (fv != null && fv.getSingle() == null) {
			int fdfdf = 123;
		}
		// if (probabilitySum > threshold) {
		return fv;
		// } else {
		// return null;
		// }
	}

	@Override
	public FusedValue fuseList(MatchedValues values, TableScores probabilities, double threshold) {
		SimilarityMeasure sm = sc.getMeasure(values.getDataType());

		List<ClusterValue> clusterValues = new ArrayList<>();

		for (SingleValue e : values.getValues()) {
			Probability p = probabilities.get(e.getId());
			if (p != null) {
				Double probability = probabilities.get(e.getId()).get(SimilarityType.LIST_OVERLAP).getAccuracy();

				if (probability != null) {

					AbstractObject o = e.getValue();

					clusterValues.add(new ClusterValue(o, probability));

				}
			}

		}

		SortedSet<Cluster> clusters = clusterFuser.fuse(clusterValues, baseFuser, sm);

		clusters = clusters.stream().filter((e) -> {
			return e.getScore() > threshold;
		}).collect(Collectors.toCollection(() -> new TreeSet<Cluster>()));

		clusters = otsuThresholder.threshold(clusters);

		if (clusters.size() == 0)
			return null;
		else
			return new FusedValue(clusters);
	}

}
