package de.mannheim.uni.expansion.tablefusion.core.matching;

import java.util.Collections;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.TripleId;

public class GroupedValues extends AbstractGroupledValues<TripleId> {
	private final Map<TripleId, MatchedValues> values;

	public GroupedValues(Map<TripleId, MatchedValues> groupedValues) {
		this.values = Collections.unmodifiableMap(groupedValues);
	}

	public Set<Entry<TripleId, MatchedValues>> entrySet() {
		return values.entrySet();
	}

	public Set<TripleId> keySet() {
		return values.keySet();
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((getUUID() == null) ? 0 : getUUID().hashCode());
		result = prime * result + ((values == null) ? 0 : values.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		GroupedValues other = (GroupedValues) obj;
		if (getUUID() == null) {
			if (other.getUUID() != null)
				return false;
		} else if (!getUUID().equals(other.getUUID()))
			return false;
		if (values == null) {
			if (other.values != null)
				return false;
		} else if (!values.equals(other.values))
			return false;
		return true;
	}

	public MatchedValues get(TripleId tripleId) {
		return values.get(tripleId);
	}

	public int size() {
		return values.size();
	}

	public Map<TripleId, MatchedValues> getValues() {
		return values;
	}

}
