package de.mannheim.uni.expansion.tablefusion.core.matching.impl;

import java.util.Map;
import java.util.Set;

import javax.inject.Inject;
import javax.inject.Named;

import de.mannheim.uni.expansion.data.DataNamedProperties;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.adapters.columnmatches.ColumnMatchesRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchCreator;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.matching.SynchronizedMatchedValuesMapMerger;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class ColumnMatchesGrouper implements MatchCreator {
	private final ColumnMatchesRepository columnMatchesRepository;
	private final AnnotatedTablesRepository tablesRepository;

	private final KnowledgeBase dbPediaRepositoryNew;

	private final OntologyRepository ontologyRepository;

	private final Map<InstanceId, InstanceId> dbpediaMappings;

	@Inject
	public ColumnMatchesGrouper(ColumnMatchesRepository columnMatchesRepository, AnnotatedTablesRepository tablesRepository, KnowledgeBase dbPediaRepositoryNew,
			OntologyRepository ontologyRepository, @Named(DataNamedProperties.DBPEDIA_MAPPINGS) Map<InstanceId, InstanceId> dbpediaMappings) {
		this.columnMatchesRepository = columnMatchesRepository;
		this.tablesRepository = tablesRepository;

		this.dbPediaRepositoryNew = dbPediaRepositoryNew;
		this.ontologyRepository = ontologyRepository;
		this.dbpediaMappings = dbpediaMappings;
	}

	@Override
	public Map<TripleId, MatchedValues> matchColumns(AnnotatedTableId id, Set<ColumnIdx> columns) {
		AnnotatedTableMetaData metaData = tablesRepository.getMetaData(id);

		ClassId classId = metaData.getDbPediaClass();

		SynchronizedMatchedValuesMapMerger<TripleId> mapMerger = new SynchronizedMatchedValuesMapMerger<>();

		for (ColumnIdx columnIndex : columns) {

			PropertyId propertyId = metaData.getProperties().get(columnIndex);

			Set<RowIdx> instancesWithMatch = columnMatchesRepository.getInstancesWithMatch(id, columnIndex);

			// ColumnDataType cdt =
			// tablesRepository.getData(id).getColumns().get(metaData.getCol(propertyId)).getDataType();
			ValueType cdt = ontologyRepository.getDataType(propertyId);

			for (RowIdx rowIdx : instancesWithMatch) {

				InstanceId instanceId = this.tablesRepository.getMetaData(id).getInstances().get(rowIdx);

				TripleId mid = new TripleId(classId, propertyId, instanceId);

				CellId cellId = new CellId(id, columnIndex, rowIdx);
				InstanceId match = columnMatchesRepository.getMatch(cellId);
				if (dbpediaMappings != null) {
					match = dbpediaMappings.get(match);
				}
				if (match == null) {
					int fsadf = 123;
				}
				mapMerger.merge(mid, new SingleValue(cellId, match), cdt);

			}

		}

		return mapMerger.getCleanMap();
	}

	@Override
	public Map<TripleId, MatchedValues> matchColumns(AnnotatedTableId id, ColumnIdx columnIndex) {

		AnnotatedTableMetaData metaData = tablesRepository.getMetaData(id);
		AnnotatedTable data = tablesRepository.getTable(id);

		PropertyId propertyId = metaData.getProperties().get(columnIndex.getColumnIdx());
		ClassId classId = metaData.getDbPediaClass();

		SynchronizedMatchedValuesMapMerger<TripleId> mapMerger = new SynchronizedMatchedValuesMapMerger<>();

		Set<RowIdx> instancesWithMatch = columnMatchesRepository.getInstancesWithMatch(id, columnIndex);

		// ColumnDataType cdt =
		// tablesRepository.getData(id).getColumns().get(metaData.getCol(propertyId)).getDataType();

		ValueType cdt = ontologyRepository.getDataType(propertyId);

		AnnotatedTableMetaData md = this.tablesRepository.getMetaData(id);

		for (RowIdx rowIdx : instancesWithMatch) {

			InstanceId instanceId = md.getInstances().get(rowIdx.getRowIdx());
			if (instanceId != null) {

				TripleId mid = new TripleId(classId, propertyId, instanceId);

				CellId cellId = new CellId(id, columnIndex.getColumnIdx(), rowIdx.getRowIdx());
				InstanceId match = columnMatchesRepository.getMatch(cellId);
				if (dbpediaMappings != null) {
					match = dbpediaMappings.get(match);
				}

				AbstractObject value = data.getValue(rowIdx, columnIndex);

				if (match != null) {
					mapMerger.merge(mid, new SingleValue(cellId, match), cdt);
				} else {
					// TODO: ADD STRING FALLBACK
					// mapMerger.merge(mid, new SingleValue(cellId, match),
					// cdt);

					// mapMerger.merge(mid, new SingleValue(cellId,value), cdt);
				}

			}

		}

		return mapMerger.getCleanMap();
	}

}
