package de.mannheim.uni.expansion.tablefusion.core.scoring;

import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleIdContainer;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.AbstractGroupledValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.TableScores;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.TableScoresInterface;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.mannheim.uni.expansion.utils.SimpleConcurrentCache;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.qualifiers.TimestampPointInTime;
import de.uni_mannheim.informatik.dws.expansion.evaluation.timestamps.TimestampQualifiersAdapter;

public class TimestampScorer {

	private static SimpleConcurrentCache<AbstractGroupledValues<?>, TimestampScores> cache = new SimpleConcurrentCache<>();
	private static SimpleConcurrentCache<UUID, TimestampScores> newCache = new SimpleConcurrentCache<>();

	private final AnnotatedTablesRepository repository;
	private final KnowledgeBase knowledgeBase;
	private final SimilarityCalculator similarityCalculator;
	private final TimestampQualifiersAdapter qualifierAdapter;

	@Inject
	public TimestampScorer(AnnotatedTablesRepository tableRepository, KnowledgeBase knowledgeBase, SimilarityCalculator similarityCalculator,
			TimestampQualifiersAdapter qualifierAdapter) {
		super();
		this.repository = tableRepository;
		this.knowledgeBase = knowledgeBase;
		this.similarityCalculator = similarityCalculator;
		this.qualifierAdapter = qualifierAdapter;
	}

	@Deprecated
	public TimestampScores getTimestampScores(GroupedValues learningSet) {
		throw new IllegalStateException();
		// if (cache.hasCacheFor(learningSet)) {
		// return cache.get(learningSet);
		// } else {
		//
		// ConcurrentHashMap<TableColumnId, Pair<TimestampIdentifier, Accuracy>>
		// probabilties = new ConcurrentHashMap<>();
		//
		// Map<TableColumnId, Map<TripleId, SingleValue>> grouped =
		// groupByColumnId(learningSet);
		//
		// Set<TripleId> no = Collections.newSetFromMap(new
		// ConcurrentHashMap<>());
		//
		// Map<TableColumnId, Map<TimestampIdentifier, Accuracy>> mappingsMap =
		// new ConcurrentHashMap<>();
		//
		// PersistentParallel4.runAndFinish(grouped.entrySet(), (e) -> {
		//
		// Map<TimestampIdentifier, MutableInt> total = new HashMap<>();
		// Map<TimestampIdentifier, MutableInt> correct = new HashMap<>();
		//
		// AnnotatedTable table = repository.getTable(e.getKey().getTableId());
		//
		// AnnotatedTableMetaData am =
		// repository.getMetaData(e.getKey().getTableId());
		//
		// List<ColumnIdx> timeColumns = table.getTimeColumns();
		// Map<TimestampIdentifier, Integer> tableYears =
		// repository.getMetaData(e.getKey().getTableId()).getTableYears();
		//
		// MutableInt totalOverlap = new MutableInt(0);
		// int sdf = 1;
		// for (Entry<TripleId, SingleValue> sv : e.getValue().entrySet()) {
		// QualifiedValueList a =
		// knowledgeBase.getQualifiedValueList(sv.getKey().getValueId());
		//
		// if (a != null) {
		// totalOverlap.increment();
		// Map<TimestampPointInTime, List<Integer>> qualifiers =
		// qualifierAdapter.getQualifierMapFromQVL(a);
		// for (Entry<TimestampPointInTime, List<Integer>> qEntry :
		// qualifiers.entrySet()) {
		// TimestampPointInTime q = qEntry.getKey();
		//
		// Integer year = q.getYear();
		//
		// if (year != null) {
		//
		// Map<TimestampIdentifier, Integer> tableYearsTemp = new
		// HashMap<>(tableYears);
		//
		// tableYearsTemp.put(TimestampIdentifier.createPropagatedIdentifier(year),
		// year);
		//
		// for (ColumnIdx columnIdx : timeColumns) {
		// AbstractObject ao = table.getValue(sv.getValue().getId().getRowIdx(),
		// columnIdx);
		// Integer columnYear = TimeObject.getYear(ao);
		// if (columnYear != null) {
		// tableYearsTemp.put(TimestampIdentifier.createCascadedIdentifier(CascadedType.COLUMN_VALUE,
		// columnIdx.getColumnIdx(), 0),
		// columnYear);
		// }
		// }
		//
		// for (Entry<TimestampIdentifier, Integer> ty :
		// tableYearsTemp.entrySet()) {
		// MutableInt totalMap = CHMHelper.getFromMap(total, ty.getKey(), () ->
		// new MutableInt());
		// MutableInt correctMap = CHMHelper.getFromMap(correct, ty.getKey(), ()
		// -> new MutableInt());
		//
		// if (ty.getValue().equals(year)) {
		// totalMap.increment();
		//
		// for (Integer qualifiedValueHolderIndex : qEntry.getValue()) {
		// QualifiedValueHolder qvh = a.getClaim(qualifiedValueHolderIndex);
		//
		// if (similarityCalculator.isEqual(qvh.getValue(),
		// sv.getValue().getValue())) {
		// correctMap.increment();
		// break;
		// }
		// }
		//
		// }
		// }
		// }
		//
		// }
		// }
		// }
		//
		// Map<TimestampIdentifier, Accuracy> accuracies = new HashMap<>();
		//
		// for (Entry<TimestampIdentifier, MutableInt> entry : total.entrySet())
		// {
		// MutableInt totalInt = entry.getValue();
		// MutableInt correctInt = correct.get(entry.getKey());
		//
		// Accuracy acc = Accuracy.fromCorrectAndTotal(correctInt.intValue(),
		// totalInt.intValue());
		// accuracies.put(entry.getKey(), acc);
		//
		// }
		//
		// Map<TimestampIdentifier, Accuracy> harmonicMean =
		// accuracies.entrySet().stream().collect(Collectors.toMap(Entry::getKey,
		// (acc) -> {
		// Accuracy acc1 = acc.getValue();
		// Accuracy acc2 = Accuracy.fromCorrectAndTotal(acc1.getCorrect(),
		// totalOverlap.intValue());
		//
		// return Accuracy.harmonicMean(acc1, acc2, acc1.getTotal());
		// }));
		//
		// LinkedHashMap<TimestampIdentifier, Accuracy> sortedAccuracies =
		// MapSorter.sortHashMapByValuesD(harmonicMean);
		//
		// Map<TimestampIdentifier, Accuracy> mappings = new HashMap<>();
		// boolean added = false;
		// for (Entry<TimestampIdentifier, Accuracy> entry :
		// sortedAccuracies.entrySet()) {
		// if (!added) {
		// probabilties.put(e.getKey(), new Pair<>(entry.getKey(),
		// entry.getValue()));
		// added = true;
		// }
		// mappings.put(entry.getKey(), entry.getValue());
		//
		// no.addAll(e.getValue().keySet());
		// break;
		// }
		//
		// mappingsMap.put(e.getKey(), mappings);
		//
		// }, "TimeStampCalculator");
		//
		// TimestampScores scores = new TimestampScores(probabilties,
		// mappingsMap, new ConcurrentHashMap<>());
		// cache.add(learningSet, scores);
		//
		// return scores;
		// }
	}

	public TimestampScores getTimestampScores(SplitGroupedValues learningSet) {
		throw new IllegalStateException();
		// if (newCache.hasCacheFor(learningSet.getUUID())) {
		// return newCache.get(learningSet.getUUID());
		// } else {
		//
		// ConcurrentHashMap<TableColumnId, Pair<TimestampIdentifier, Accuracy>>
		// probabilties = new ConcurrentHashMap<>();
		// ConcurrentHashMap<TableColumnId, Map<TimestampIdentifier, Accuracy>>
		// fullMappings = new ConcurrentHashMap<>();
		// ConcurrentHashMap<ClassPropertyId, Accuracy> averageMaxScores = new
		// ConcurrentHashMap<>();
		// ConcurrentHashMap<ClassPropertyId, ConcurrentLinkedQueue<Accuracy>>
		// maxScores = new ConcurrentHashMap<>();
		// Map<TableColumnId, Map<QualifiedTripleId, SingleValue>> grouped =
		// groupByColumnId(learningSet);
		//
		// PersistentParallel4.runAndFinish(grouped.entrySet(), (e) -> {
		//
		// Map<TimestampIdentifier, MutableInt> total = new HashMap<>();
		// Map<TimestampIdentifier, MutableInt> correct = new HashMap<>();
		//
		// AnnotatedTable table = repository.getTable(e.getKey().getTableId());
		//
		// AnnotatedTableMetaData am =
		// repository.getMetaData(e.getKey().getTableId());
		//
		// // System.out.println("Values in group: " + e.getValue().size()
		// // + ", ROWS in Table: " + am.getInstances().size());
		//
		// List<ColumnIdx> timeColumns = table.getTimeColumns();
		// Map<TimestampIdentifier, Integer> tableYears =
		// repository.getMetaData(e.getKey().getTableId()).getTableYears();
		//
		// Set<Pair<PropertyId, InstanceId>> totalCount = new HashSet<>();
		//
		// // for (QualifiedTripleId vvv : e.getValue().keySet()) {
		// //
		// // Pair<PropertyId, InstanceId> pair = new
		// // Pair<>(vvv.tripleId().getPropertyId(),
		// // vvv.tripleId().getInstanceId());
		// // if (!totalCount.contains(pair)) {
		// // totalCount.add(pair);
		// // }
		// // }
		// // int totalOverlap = e.getValue().size();
		//
		// int sdf = 1;
		// for (Entry<QualifiedTripleId, SingleValue> sv :
		// e.getValue().entrySet()) {
		// QualifiedValueList a =
		// knowledgeBase.getQualifiedValueListWithEqualProperties(sv.getKey().tripleId().getValueId());
		// if (a != null) {
		//
		// Set<AbstractObject> values =
		// this.qualifierAdapter.getForQualifiers(a,
		// sv.getKey().getQualifiers());
		//
		// if (!values.isEmpty()) {
		//
		// Integer year = null;
		// for (AbstractQualifier aq : sv.getKey().getQualifiers()) {
		// if (aq instanceof TimestampPointInTime) {
		// year = ((TimestampPointInTime) aq).getYear();
		// }
		// }
		//
		// if (year != null) {
		//
		// Map<TimestampIdentifier, Integer> tableYearsTemp = new HashMap<>();
		// tableYearsTemp.putAll(tableYears);
		//
		// tableYearsTemp.put(TimestampIdentifier.createPropagatedIdentifier(year),
		// year);
		//
		// for (ColumnIdx columnIdx : timeColumns) {
		// AbstractObject ao = table.getValue(sv.getValue().getId().getRowIdx(),
		// columnIdx);
		// Integer columnYear = TimeObject.getYear(ao);
		// if (columnYear != null) {
		// tableYearsTemp.put(TimestampIdentifier.createCascadedIdentifier(CascadedType.COLUMN_VALUE,
		// columnIdx.getColumnIdx(), 0),
		// columnYear);
		// }
		// }
		//
		// for (Entry<TimestampIdentifier, Integer> ty :
		// tableYearsTemp.entrySet()) {
		//
		// if (ty.getValue().equals(year)) {
		// MutableInt totalMap = CHMHelper.getFromMap(total, ty.getKey(), () ->
		// new MutableInt());
		// MutableInt correctMap = CHMHelper.getFromMap(correct, ty.getKey(), ()
		// -> new MutableInt());
		//
		// totalMap.increment();
		//
		// for (AbstractObject ao : values) {
		//
		// if (similarityCalculator.isEqual(ao, sv.getValue().getValue())) {
		// correctMap.increment();
		// break;
		// }
		// }
		//
		// }
		// }
		// }
		// }
		// }
		// }
		//
		// OptionalInt totalOverlapOptional =
		// total.entrySet().stream().map((totalEntry) -> {
		// return totalEntry.getValue();
		// }).mapToInt(MutableInt::intValue).max();
		//
		// Map<TimestampIdentifier, Accuracy> accuracies = new HashMap<>();
		//
		// for (Entry<TimestampIdentifier, MutableInt> entry : total.entrySet())
		// {
		// MutableInt totalInt = entry.getValue();
		// MutableInt correctInt = correct.get(entry.getKey());
		// if (correctInt.intValue() > 0) {
		//
		// Accuracy acc = Accuracy.fromCorrectAndTotal(correctInt.intValue(),
		// totalInt.intValue());
		//
		// accuracies.put(entry.getKey(), acc);
		// if (entry.getKey().getType() ==
		// TimestampIdentifier.TimestampType.PROPAGATED) {
		// CHMHelper.getFromCHM(fullMappings, e.getKey(), () -> new
		// HashMap<>()).put(entry.getKey(), acc);
		// }
		// }
		//
		// }
		// LinkedHashMap<TimestampIdentifier, Accuracy> sortedAccuracies;
		// if (totalOverlapOptional.isPresent()) {
		// final int totalOverlap = totalOverlapOptional.getAsInt();
		// Map<TimestampIdentifier, Accuracy> harmonicMean =
		// accuracies.entrySet().stream().collect(Collectors.toMap(Entry::getKey,
		// (acc) -> {
		// Accuracy acc1 = acc.getValue();
		// Accuracy acc2 = Accuracy.fromCorrectAndTotal(acc1.getCorrect(),
		// totalOverlap);
		// Accuracy harmonic = Accuracy.harmonicMean(acc1, acc2,
		// acc1.getTotal());
		//
		// // CHMHelper.getFromCHM(fullMappings, e.getKey(), () ->
		// // new HashMap<>()).put(acc.getKey(), harmonic);
		//
		// return harmonic;
		// // return acc1;
		// }));
		//
		// sortedAccuracies = MapSorter.sortHashMapByValuesD(harmonicMean);
		// } else {
		// sortedAccuracies = MapSorter.sortHashMapByValuesD(accuracies);
		// }
		//
		// LinkedHashMap<TimestampIdentifier, Accuracy> mappings = new
		// LinkedHashMap<>();
		// for (Entry<TimestampIdentifier, Accuracy> entry :
		// sortedAccuracies.entrySet()) {
		// if (entry.getKey().getType() !=
		// TimestampIdentifier.TimestampType.PROPAGATED) {
		// Accuracy acc = accuracies.get(entry.getKey());
		// probabilties.put(e.getKey(), new Pair<>(entry.getKey(), acc));
		//
		// mappings.put(entry.getKey(), acc);
		//
		// // Set<QualifiedPropertyId> propertyIds =
		// // e.getValue().keySet().stream().map((qtid) -> {
		// // return qtid.getQualifiedPropertyId();
		// // }).collect(Collectors.toSet());
		//
		// Set<ClassPropertyId> propertyIds =
		// e.getValue().keySet().stream().map((qtid) -> {
		// return qtid.getTripleId().getClassPropertyId();
		// }).collect(Collectors.toSet());
		// for (ClassPropertyId cpid : propertyIds) {
		//
		// CHMHelper.getFromCHM(maxScores, cpid, () -> new
		// ConcurrentLinkedQueue<>()).add(acc);
		//
		// // MutableReference<Accuracy> reference =
		// // CHMHelper.getFromCHM(maxScores,
		// // qualifiedTripleId.tripleId(),
		// // () -> new MutableReference<Accuracy>(null));
		// // synchronized (reference) {
		// // if (reference.get() != null &&
		// // reference.get().getAccuracy() <
		// // acc.getAccuracy()) {
		// // reference.set(acc);
		// // } else if (reference.get() == null) {
		// // reference.set(acc);
		// // }
		// // }
		// }
		//
		// break;
		// }
		//
		// }
		//
		// // for (Entry<TimestampIdentifier, Accuracy> entry :
		// // sortedAccuracies.entrySet()) {
		// // if (entry.getKey().getType() !=
		// // TimestampIdentifier.TimestampType.PROPAGATED) {
		// // Accuracy acc = accuracies.get(entry.getKey());
		// // probabilties.put(e.getKey(), new Pair<>(entry.getKey(),
		// // acc));
		// //
		// // Set<QualifiedPropertyId> propertyIds =
		// // e.getValue().keySet().stream().map((qtid) -> {
		// // return qtid.getQualifiedPropertyId();
		// // }).collect(Collectors.toSet());
		// // for (QualifiedPropertyId qualifiedPropertyId :
		// // propertyIds) {
		// // MutableReference<Accuracy> reference =
		// // CHMHelper.getFromCHM(maxScores,
		// // qualifiedTripleId.tripleId(),
		// // () -> new MutableReference<Accuracy>(null));
		// // synchronized (reference) {
		// // if (reference.get() != null &&
		// // reference.get().getAccuracy() < acc.getAccuracy()) {
		// // reference.set(acc);
		// // } else if (reference.get() == null) {
		// // reference.set(acc);
		// // }
		// // }
		// // }
		// //
		// // break;
		// // }
		// //
		// // }
		//
		// }, "TimeStampCalculator");
		//
		// for (Entry<ClassPropertyId, ConcurrentLinkedQueue<Accuracy>> me :
		// maxScores.entrySet()) {
		// double acc = 0.0;
		// int total = 0;
		// int count = 0;
		//
		// for (Accuracy a : me.getValue()) {
		// acc += a.getAccuracy();
		// total += a.getTotal();
		// count++;
		// }
		//
		// averageMaxScores.put(me.getKey(), new Accuracy(acc / count, total /
		// count));
		// }
		//
		// ConcurrentHashMap<TableColumnId, Map<TimestampIdentifier, Accuracy>>
		// fullMappingsSorted = new ConcurrentHashMap<>();
		//
		// TimestampScores scores = new TimestampScores(probabilties,
		// fullMappings, averageMaxScores);
		// newCache.add(learningSet.getUUID(), scores);
		//
		// return scores;
		// }
	}

	private <E extends TripleIdContainer> Map<TableColumnId, Map<E, SingleValue>> groupByColumnId(AbstractGroupledValues<E> learningSet) {
		ConcurrentHashMap<TableColumnId, Map<E, SingleValue>> groupedValues = new ConcurrentHashMap<TableColumnId, Map<E, SingleValue>>();

		PersistentParallel4.runAndFinish(learningSet.entrySet(), (e) -> {

			for (SingleValue sv : e.getValue().getValues()) {
				Map<E, SingleValue> tableColumnIdMap = CHMHelper.getFromCHM(groupedValues,
						new TableColumnId(sv.getId().getTableId(), sv.getId().getColumnIdx()), () -> new ConcurrentHashMap<>());

				tableColumnIdMap.put(e.getKey(), sv);
			}
		}, "KBTCalculator - Group Values by TableColumnId");

		return groupedValues;
	}

	public ScoresContainer getScores(SplitGroupedValues learningSet, SplitGroupedValues testingSet, ScoresContainer other) {

		TimestampScores timestampScores = this.getTimestampScores(learningSet);

		return new TimestampScoresContainer(timestampScores, other, testingSet);
	}

	private class TimestampScoresContainer implements ScoresContainer {

		private final TimestampScores timestampScores;
		private final ScoresContainer columnScores;
		private final SplitGroupedValues testingSet;

		public TimestampScoresContainer(TimestampScores timestampScores, ScoresContainer scoresContainer, SplitGroupedValues testingSet) {
			this.timestampScores = timestampScores;
			this.columnScores = scoresContainer;
			this.testingSet = testingSet;
		}

		@Override
		public TableScores getTableScores(ClassId classId, PropertyId propertyId, InstanceId instanceId) {
			return new TableScoresInterface(this, classId, propertyId, instanceId);
		}

		@Override
		public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx,
				RowIdx rowIdx) {
			return columnScores.getScore(classId, propertyId, instanceId, tableId, columnIdx, rowIdx);
		}

		@Override
		public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx,
				RowIdx rowIdx) {
			return columnScores.containsKey(classId, propertyId, instanceId, tableId, columnIdx, rowIdx);
		}

		public Probability getScore(TripleId tripleId, CellId cellId, Set<AbstractQualifier> qualifiers) {
			throw new IllegalStateException();
			// Probability KBT = null;
			// if (columnScores != null && false) {
			// Probability otherScore = columnScores.getScore(tripleId, cellId,
			// qualifiers);
			// java.util.Optional<Accuracy> maxScore =
			// timestampScores.getMaxScore(tripleId);
			//
			// if (maxScore.isPresent() && otherScore != null) {
			// // KBT = null;
			// if (maxScore.get().getAccuracy() <
			// otherScore.get(SimilarityType.DIRECT).getAccuracy()) {
			// KBT = otherScore;
			// }
			// } else {
			// KBT = otherScore;
			// }
			// }
			//
			// double accuracy = 0.0;
			// int total = 0;
			// {
			// TimestampIdentifier tsi = timestampScores.getMapping(new
			// TableColumnId(cellId));
			// if (tsi != null) {
			//
			// Integer tableYear;
			// if (TimestampType.PROPAGATED.equals(tsi.getType())) {
			// tableYear = tsi.getPropogatedYear();
			//
			// tableYear = null;
			// } else {
			// tableYear = repository.getTimeStamp(cellId, tsi);
			//
			// }
			//
			// if (tableYear != null) {
			// for (AbstractQualifier qualifier : qualifiers) {
			// if (qualifier instanceof TimestampPointInTime) {
			// TimestampPointInTime tpit = (TimestampPointInTime) qualifier;
			// Accuracy acc = timestampScores.getScore(new
			// TableColumnId(cellId));
			// int diff = Math.abs(tableYear.intValue() - tpit.getYear());
			// switch (diff) {
			// case 0:
			//
			// total += acc.getTotal();
			// accuracy += acc.getAccuracy();
			// break;
			// case 1:
			// total += acc.getTotal();
			// accuracy += acc.getAccuracy() * 0.75;
			// break;
			// case 2:
			// total += acc.getTotal();
			// accuracy += acc.getAccuracy() * 0.5;
			// break;
			// case 3:
			// total += acc.getTotal();
			// accuracy += acc.getAccuracy() * 0.25;
			// break;
			// }
			// }
			// }
			// }
			//
			// }
			// }
			//
			// Integer qualifiedYear = null;
			// for (AbstractQualifier aq : qualifiers) {
			// if (aq instanceof TimestampPointInTime) {
			// qualifiedYear = ((TimestampPointInTime) aq).getYear();
			// }
			// }
			// // if (otherScore != null) {
			// // accuracy +=
			// // otherScore.get(SimilarityType.DIRECT).getAccuracy();
			// // total += otherScore.get(SimilarityType.DIRECT).getTotal();
			// // }
			// boolean secondary = true;
			// if (secondary) {
			// Map<TimestampIdentifier, Accuracy> mappings =
			// timestampScores.getMappings(new TableColumnId(cellId));
			// LinkedHashMap<TimestampIdentifier, Accuracy> sortedMap =
			// MapSorter.sortHashMapByValuesD(mappings);
			// int totalMapSize = sortedMap.size();
			// int currentMapCount = 0;
			// if (mappings != null) {
			//
			// int fffsdf = 123;
			//
			// for (Entry<TimestampIdentifier, Accuracy> tsiEntry :
			// sortedMap.entrySet()) {
			//
			// if (tsiEntry.getValue().getAccuracy() > 0.6) {
			// int fsdf = 12;
			// }
			// TimestampIdentifier tsi = tsiEntry.getKey();
			// if (tsi != null && !tsi.equals(timestampScores.getMapping(new
			// TableColumnId(cellId)))) {
			//
			// Integer tableYear;
			// if (TimestampType.PROPAGATED.equals(tsi.getType())) {
			// tableYear = tsi.getPropogatedYear();
			// if (qualifiedYear != null && !qualifiedYear.equals(tableYear)) {
			// tableYear = null;
			// }
			// } else {
			// tableYear = repository.getTimeStamp(cellId, tsi);
			//
			// }
			//
			// if (tableYear != null) {
			// for (AbstractQualifier qualifier : qualifiers) {
			// if (qualifier instanceof TimestampPointInTime) {
			// TimestampPointInTime tpit = (TimestampPointInTime) qualifier;
			// Accuracy acc = tsiEntry.getValue();
			// int diff = Math.abs(tableYear.intValue() - tpit.getYear());
			// // double multiplier = 0.0;
			// switch (diff) {
			// case 0:
			// if (timestampScores.getMapping(new TableColumnId(cellId)) == null
			// && !this.hasScore(tripleId, qualifiers)) {
			// total += acc.getTotal();
			// accuracy += acc.getAccuracy() / totalMapSize * currentMapCount;
			// currentMapCount++;
			// } else {
			//
			// if (!this.hasScore(tripleId, qualifiers) ||
			// this.getMaxScore(tripleId, qualifiers) < acc.getAccuracy()) {
			// total += acc.getTotal();
			// accuracy += acc.getAccuracy() / totalMapSize * currentMapCount;
			// currentMapCount++;
			// }
			//
			// // java.util.Optional<Accuracy>
			// // maxScore =
			// // timestampScores.getMaxScore(tripleId);
			// //
			// // if (!maxScore.isPresent() ||
			// // acc.getAccuracy() >
			// // maxScore.get().getAccuracy())
			// // {
			// // total += acc.getTotal();
			// // accuracy += acc.getAccuracy()
			// // *
			// // accuracy;
			// // }
			// }
			//
			// break;
			// // case 1:
			// // multiplier = 0.75;
			// // break;
			// // case 2:
			// // multiplier = 0.5;
			// // break;
			// // case 3:
			// // multiplier = 0.25;
			// // break;
			// }
			//
			// }
			// }
			// }
			//
			// }
			// }
			//
			// }
			// }
			// return new Probability(new Accuracy(accuracy, total));

			// Integer qualifiedYear = null;
			// for (AbstractQualifier aq : qualifiers) {
			// if (aq instanceof TimestampPointInTime) {
			// qualifiedYear = ((TimestampPointInTime) aq).getYear();
			// }
			// }
			// // if (otherScore != null) {
			// // accuracy +=
			// otherScore.get(SimilarityType.DIRECT).getAccuracy();
			// // total += otherScore.get(SimilarityType.DIRECT).getTotal();
			// // }
			// Map<TimestampIdentifier, Accuracy> mappings =
			// timestampScores.getMappings(new TableColumnId(cellId));
			// if (mappings != null) {
			//
			// int fffsdf = 123;
			//
			// for (Entry<TimestampIdentifier, Accuracy> tsiEntry :
			// mappings.entrySet()) {
			// if (tsiEntry.getValue().getAccuracy() > 0.6) {
			// int fsdf = 12;
			// }
			// TimestampIdentifier tsi = tsiEntry.getKey();
			// if (tsi != null) {
			//
			// Integer tableYear;
			// if (TimestampType.PROPAGATED.equals(tsi.getType())) {
			// tableYear = tsi.getPropogatedYear();
			// if (qualifiedYear != null && !qualifiedYear.equals(tableYear)) {
			// tableYear = null;
			// }
			// } else {
			// tableYear = repository.getTimeStamp(cellId, tsi);
			//
			// }
			//
			// if (tableYear != null) {
			// for (AbstractQualifier qualifier : qualifiers) {
			// if (qualifier instanceof TimestampPointInTime) {
			// TimestampPointInTime tpit = (TimestampPointInTime) qualifier;
			// Accuracy acc = tsiEntry.getValue();
			// int diff = Math.abs(tableYear.intValue() - tpit.getYear());
			// switch (diff) {
			// case 0:
			//
			// total += acc.getTotal();
			// accuracy += acc.getAccuracy();
			// break;
			// // case 1:
			// // total += acc.getTotal();
			// // accuracy += acc.getAccuracy() * 0.75;
			// // break;
			// // case 2:
			// // total += acc.getTotal();
			// // accuracy += acc.getAccuracy() * 0.5;
			// // break;
			// // case 3:
			// // total += acc.getTotal();
			// // accuracy += acc.getAccuracy() * 0.25;
			// // break;
			// }
			// }
			// }
			// }
			//
			// }
			// }
			//
			// }
			//
			// if (accuracy > 0.0) {
			// return new Probability(new Accuracy(accuracy, total));
			// }
			//
			// return KBT;
		}

		private boolean hasScore(TripleId tripleId, Set<AbstractQualifier> qualifiers) {
			MatchedValues v = this.testingSet.get(new QualifiedTripleId(tripleId, qualifiers));
			if (v != null) {
				for (SingleValue sv : v.getValues()) {
					if (timestampScores.getMapping(new TableColumnId(sv.getId())) != null)
						return true;
				}
			}
			return false;
		}

		ConcurrentHashMap<QualifiedTripleId, Double> cacheMap = new ConcurrentHashMap<>();

		private double getMaxScore(TripleId tripleId, Set<AbstractQualifier> qualifiers) {
			QualifiedTripleId qtid = new QualifiedTripleId(tripleId, qualifiers);
			if (cacheMap.containsKey(qtid))
				return cacheMap.get(qtid);

			MatchedValues v = this.testingSet.get(new QualifiedTripleId(tripleId, qualifiers));
			double maxScore = 0.0;
			if (v != null) {
				for (SingleValue sv : v.getValues()) {
					TableColumnId tcid = new TableColumnId(sv.getId());
					if (timestampScores.getMapping(tcid) != null) {

						double acc = this.timestampScores.getScore(tcid).getAccuracy();
						if (acc > maxScore) {

							TimestampIdentifier mapping = timestampScores.getMapping(new TableColumnId(sv.getId()));
							Integer tableYear = repository.getTimeStamp(sv.getId(), mapping);
							if (tableYear != null) {
								for (AbstractQualifier q : qualifiers) {
									if (q instanceof TimestampPointInTime) {
										Integer year = ((TimestampPointInTime) q).getYear();
										if (tableYear.equals(year)) {
											maxScore = acc;
										}
									}
								}
							}

						}

					}
				}
			}

			cacheMap.put(qtid, maxScore);
			return maxScore;
		}

		public boolean containsKey(TripleId tripleId, CellId cellId, Set<AbstractQualifier> qualifiers) {
			boolean ret = getScore(tripleId, cellId, qualifiers) != null;
			if (ret)
				return true;
			else
				return false;

		}

	}
}
