package de.mannheim.uni.expansion.tablefusion.core.scoring;

import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.commons.lang.mutable.MutableInt;

import com.fasterxml.uuid.Generators;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableMap.Builder;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier;
import de.mannheim.uni.expansion.tablefusion.analysis.Accuracy;
import de.mannheim.uni.expansion.utils.containers.Pair;

public class TimestampScores {

	private final UUID uuid = Generators.timeBasedGenerator().generate();

	public UUID getUUID() {
		return uuid;
	}

	private final Map<TableColumnId, Pair<TimestampIdentifier, Accuracy>> mappings;
	private final Map<TableColumnId, ImmutableMap<TimestampIdentifier, Accuracy>> fullMappings;
	private final ConcurrentHashMap<ClassPropertyId, Accuracy> maxScores;
	private final Map<ClassPropertyId, Map<String, MutableInt>> tsCount;

	public TimestampScores(Map<TableColumnId, Pair<TimestampIdentifier, Accuracy>> mappings,
			Map<TableColumnId, LinkedHashMap<TimestampIdentifier, Accuracy>> fullMappings, ConcurrentHashMap<ClassPropertyId, Accuracy> maxScores,
			Map<ClassPropertyId, Map<String, MutableInt>> tsCount) {

		Map<ClassPropertyId, Map<String, MutableInt>> tempTsCount = new HashMap<>();
		for (Entry<ClassPropertyId, Map<String, MutableInt>> entry : tsCount.entrySet()) {
			tempTsCount.put(entry.getKey(), Collections.unmodifiableMap(new HashMap<>(entry.getValue())));
		}
		this.tsCount = Collections.unmodifiableMap(tempTsCount);

		this.mappings = Collections.unmodifiableMap(new ConcurrentHashMap<>(mappings));

		Builder<TableColumnId, ImmutableMap<TimestampIdentifier, Accuracy>> fullBuilder = ImmutableMap.builder();

		for (Entry<TableColumnId, LinkedHashMap<TimestampIdentifier, Accuracy>> e : fullMappings.entrySet()) {
			fullBuilder.put(e.getKey(), ImmutableMap.copyOf(e.getValue()));

		}

		this.fullMappings = fullBuilder.build();

		this.maxScores = maxScores;

	}

	public ImmutableMap<TimestampIdentifier, Accuracy> getSortedMappings(TableColumnId columnId) {
		return fullMappings.get(columnId);
	}

	public TimestampIdentifier getMapping(TableColumnId columnId) {
		Pair<TimestampIdentifier, Accuracy> value = mappings.get(columnId);
		if (value != null) {
			return value.getFirst();
		}
		return null;
	}

	public Accuracy getScore(TableColumnId columnId) {
		Pair<TimestampIdentifier, Accuracy> value = mappings.get(columnId);
		if (value != null) {
			return value.getSecond();
		}
		return null;
	}

	public Map<TableColumnId, Pair<TimestampIdentifier, Accuracy>> getMappings() {
		return mappings;
	}

	public Optional<Accuracy> getMaxScore(TripleId tripleId) {
		if (maxScores.containsKey(tripleId.getClassPropertyId()))
			return Optional.of(maxScores.get(tripleId.getClassPropertyId()));
		return Optional.empty();
	}

	public Map<ClassPropertyId, Map<String, MutableInt>> getTsCount() {
		return tsCount;
	}

}
