package de.mannheim.uni.expansion.tablefusion.core.scoring;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import de.mannheim.uni.expansion.tablefusion.core.fusing.BaseFuser;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityMeasure;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;

public class YearEquivalenceClusterer {

	@Inject
	public YearEquivalenceClusterer() {

	}

	public List<Cluster> addValue(List<Cluster> clusters, SimilarityMeasure measure, BaseFuser baseFuser, AbstractObject value, Integer year) {
		clusters = new ArrayList<>(clusters);
		boolean added = false;
		for (int i = 0; i < clusters.size(); i++) {
			if (measure.isEqual(clusters.get(i).getValue(), value)) {
				List<ValueWithYear> valueList = new ArrayList<>(clusters.get(i).getValues());
				valueList.add(new ValueWithYear(year, value));

				clusters.set(i, new Cluster(baseFuser.clusteredFuse(valueList.stream().map(ValueWithYear::getValue).collect(Collectors.toList())).getSingle(),
						valueList));

				added = true;
				break;
			}
		}
		if (!added) {
			List<ValueWithYear> valueList = new ArrayList<>();
			valueList.add(new ValueWithYear(year, value));

			clusters.add(new Cluster(value, valueList));
		}

		return clusters;
	}

	public Integer getYear(List<Cluster> clusters, SimilarityMeasure measure, AbstractObject value) {
		for (int i = 0; i < clusters.size(); i++) {
			if (measure.isEqual(clusters.get(i).getValue(), value)) {
				return clusters.get(i).getMostCommonYear();
			}
		}

		return null;
	}

	public Cluster getHighestRatedCluster(List<Cluster> clusters) {
		clusters = new ArrayList<>(clusters);
		Collections.sort(clusters);
		return clusters.get(clusters.size() - 1);
	}

	public class Cluster implements Comparable<Cluster> {
		private final AbstractObject fusedValue;
		private final List<ValueWithYear> values;
		private final int count;

		public Cluster(AbstractObject fusedValue, List<ValueWithYear> values) {
			super();
			this.fusedValue = fusedValue;
			this.values = Collections.unmodifiableList(values);
			this.count = values.size();
		}

		public Integer getMostCommonYear() {
			Map<Integer, MutableInt> countMap = new HashMap<>();

			for (ValueWithYear v : values) {
				CHMHelper.getFromMap(countMap, v.getYear(), () -> new MutableInt(0)).increment();
			}

			Integer maxYear = null;
			int maxCount = 0;
			for (Entry<Integer, MutableInt> e : countMap.entrySet()) {
				if (e.getValue().intValue() > maxCount) {
					maxYear = e.getKey();
					maxCount = e.getValue().intValue();
				}
			}

			return maxYear;

		}

		public List<ValueWithYear> getValues() {
			return values;
		}

		@Override
		public int compareTo(Cluster o) {
			return Integer.compare(count, o.getCount());
		}

		public AbstractObject getValue() {
			return fusedValue;
		}

		public int getCount() {
			return count;
		}

	}

	private class ValueWithYear {
		private final Integer year;
		private final AbstractObject value;

		public ValueWithYear(Integer year, AbstractObject value) {
			super();
			this.year = year;
			this.value = value;
		}

		public Integer getYear() {
			return year;
		}

		public AbstractObject getValue() {
			return value;
		}

	}
}
