package de.mannheim.uni.expansion.tablefusion.core.scoring.packs;

import java.util.ArrayList;
import java.util.List;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;

public class AddedScores extends CombinedScores {

	private final List<ScoresContainer> scoresContainer;

	public AddedScores(List<ScoresContainer> scoresContainer) {
		this.scoresContainer = new ArrayList<>(scoresContainer);
	}

	@Override
	public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {

		List<Probability> probs = new ArrayList<>();

		for (ScoresContainer sc : scoresContainer) {
			Probability prob = sc.getScore(classId, propertyId, instanceId, tableId, columnIdx, rowIdx);
			if (prob != null) {
				probs.add(prob);
			}
		}
		if (probs.isEmpty())
			return null;
		else
			return new Probability(probs);
	}

	@Override
	public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {
		for (ScoresContainer sc : scoresContainer) {
			if (sc.containsKey(classId, propertyId, instanceId, tableId, columnIdx, rowIdx)) {
				return true;
			}
		}

		return false;
	}

}
