package de.mannheim.uni.expansion.tablefusion.core.scoring.packs;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;

public class ColumnScores implements ScoresContainer {

	private final Map<PropertyId, TableScores> scores;

	public ColumnScores(Map<PropertyId, Map<TableColumnId, Probability>> scores) {

		Map<PropertyId, TableScores> changedScores = new HashMap<>();

		for (Entry<PropertyId, Map<TableColumnId, Probability>> e : scores.entrySet()) {
			changedScores.put(e.getKey(), new CellScoresByTableColumn(e.getValue()));
		}

		this.scores = Collections.unmodifiableMap(changedScores);

	}

	@Override
	public TableScores getTableScores(ClassId classId, PropertyId propertyId, InstanceId instanceId) {
		return scores.get(propertyId);
	}

	public Set<Entry<PropertyId, TableScores>> entrySet() {
		return this.scores.entrySet();
	}

	@Override
	public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId id, ColumnIdx columnIdx, RowIdx instanceIdx) {
		return this.scores.get(propertyId).get(id, columnIdx, instanceIdx);
	}

	@Override
	public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId id, ColumnIdx columnIdx, RowIdx instanceIdx) {
		return this.scores.get(propertyId).containsKey(id, columnIdx, instanceIdx);
	}

	// private Creator<Map<PropertyIndexComposite, Probability>> creator = () ->
	// {
	// return new ConcurrentHashMap<>();
	// };

	// public Map<AnnotatedTableId, Map<PropertyId, Probability>> getByTable() {
	// ConcurrentHashMap<AnnotatedTableId, Map<PropertyId, Probability>> map =
	// new ConcurrentHashMap<>();
	//
	// PersistentParallel.runAndFinish(scores.entrySet(), (e) -> {
	//
	// for (Entry<AnnotatedTableId, Probability> innerEntry :
	// e.getValue().entrySet()) {
	// Map<PropertyIndexComposite, Probability> nestedMap =
	// CHMHelper.getFromCHM(map, innerEntry.getKey(), creator);
	// nestedMap.put(e.getKey(), innerEntry.getValue());
	// }
	//
	// } , "getByTable Property");
	//
	// return map;
	// }

}
