package de.mannheim.uni.expansion.tablefusion.core.scoring.packs;

import java.util.Map;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.HostId;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.cellscores.CellScoresByHost;

public class HostScores implements ScoresContainer {

	private final CellScoresByHost cellScores;

	public HostScores(Map<HostId, Probability> scores, Map<AnnotatedTableId, HostId> mappings) {
		this.cellScores = new CellScoresByHost(scores, mappings);
	}

	@Override
	public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, CellId cellId) {
		return cellScores.get(cellId);
	}

	@Override
	public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, CellId cellId) {
		return cellScores.containsKey(cellId);
	}

	@Override
	public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {
		return containsKey(classId, propertyId, instanceId, new CellId(tableId, columnIdx, rowIdx));
	}

	@Override
	public TableScores getTableScores(ClassId classId, PropertyId propertyId, InstanceId instanceId) {
		return cellScores;
	}

	@Override
	public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {
		return getScore(classId, propertyId, instanceId, new CellId(tableId, columnIdx, rowIdx));
	}

}
