package de.mannheim.uni.expansion.tablefusion.core.scoring.packs;

import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;

public class MultipliedScores extends CombinedScores {
	private final ScoresContainer main;
	private final ScoresContainer weights;
	private final boolean noScoreWhenMissing;

	public MultipliedScores(ScoresContainer main, ScoresContainer weights) {
		this.main = main;
		this.weights = weights;
		this.noScoreWhenMissing = true;
	}

	public MultipliedScores(ScoresContainer main, ScoresContainer weights, boolean noScoreWhenMissing) {
		this.main = main;
		this.weights = weights;
		this.noScoreWhenMissing = noScoreWhenMissing;
	}

	@Override
	public Probability getScore(TripleId tripleId, CellId cellId, Set<AbstractQualifier> qualifiers) {

		Probability probability = main.getScore(tripleId, cellId, qualifiers);
		Probability weight = weights.getScore(tripleId, cellId, qualifiers);
		if (probability == null) {
			return null;
		} else if (weight != null) {
			return probability.multiply(weight);
		} else if (noScoreWhenMissing) {
			return null;
		} else {
			return probability;
		}

	}

	@Override
	public boolean containsKey(TripleId tripleId, CellId cellId, Set<AbstractQualifier> qualifiers) {
		if (main.getScore(tripleId, cellId, qualifiers) != null) {
			return true;
		}

		return false;
	}

	@Override
	public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId id, ColumnIdx columnIdx, RowIdx rowIdx) {
		throw new IllegalStateException();
	}

	@Override
	public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {
		throw new IllegalStateException();
	}

}
