package de.mannheim.uni.expansion.tablefusion.core.scoring.packs;

import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import com.google.common.collect.ImmutableMap;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;

public class NonQualifiedColumnScores implements ScoresContainer {
	private final Map<TableColumnId, Map<PropertyId, Probability>> map;

	public NonQualifiedColumnScores(Map<TableColumnId, Map<PropertyId, Probability>> map) {

		ImmutableMap.Builder<TableColumnId, Map<PropertyId, Probability>> builder = ImmutableMap.builder();

		for (Entry<TableColumnId, Map<PropertyId, Probability>> e : map.entrySet()) {

			builder.put(e.getKey(), ImmutableMap.copyOf(e.getValue()));
		}

		this.map = builder.build();
	}

	@Override
	public TableScores getTableScores(ClassId classId, PropertyId propertyId, InstanceId instanceId) {

		return new NonQualifiedColumnTableScores(propertyId, map);
	}

	@Override
	public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {
		throw new IllegalStateException();
	}

	@Override
	public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {
		throw new IllegalStateException();
	}

	@Override
	public Probability getScore(TripleId tripleId, CellId cellId, Set<AbstractQualifier> qualifiers) {
		if (containsKey(tripleId, cellId, qualifiers)) {
			return this.map.get(new TableColumnId(cellId)).get(tripleId.getPropertyId());
		}

		return null;
	}

	@Override
	public boolean containsKey(TripleId tripleId, CellId cellId, Set<AbstractQualifier> qualifiers) {
		Map<PropertyId, Probability> innerMap = this.map.get(new TableColumnId(cellId));
		if (innerMap != null) {
			if (innerMap.containsKey(tripleId.getPropertyId())) {
				return true;
			}
		}

		return false;
	}

	public Map<TableColumnId, Map<PropertyId, Probability>> getFullScores() {
		return map;
	}
}
