package de.mannheim.uni.expansion.tablefusion.core.scoring.packs;

import java.util.Map;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;

public class NonQualifiedColumnTableScores implements TableScores {

	private final PropertyId propertyId;
	private final Map<TableColumnId, Map<PropertyId, Probability>> map;

	public NonQualifiedColumnTableScores(PropertyId propertyId, Map<TableColumnId, Map<PropertyId, Probability>> map) {
		this.map = map;
		this.propertyId = propertyId;
	}

	@Override
	public Probability get(AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {
		throw new IllegalStateException();
	}

	@Override
	public boolean containsKey(AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {
		throw new IllegalStateException();
	}

	@Override
	public Probability get(CellId cellId, Set<AbstractQualifier> qualifiers) {
		Map<PropertyId, Probability> innerMap = this.map.get(new TableColumnId(cellId));
		if (innerMap != null) {
			return innerMap.get(propertyId);
		}

		return null;
	}

	@Override
	public boolean containsKey(CellId cellId, Set<AbstractQualifier> qualifiers) {
		Map<PropertyId, Probability> innerMap = this.map.get(new TableColumnId(cellId));
		if (innerMap != null) {
			if (innerMap.containsKey(propertyId)) {
				return true;
			}
		}

		return false;
	}

}
