package de.mannheim.uni.expansion.tablefusion.core.scoring.packs;

import java.util.Collections;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;

public class QualifiedColumnScores implements ScoresContainer {
	private final Map<TableColumnId, Map<QualifiedPropertyId, Probability>> map;

	public Map<TableColumnId, Map<QualifiedPropertyId, Probability>> getFullScores() {
		return map;
	}

	public QualifiedColumnScores(Map<TableColumnId, Map<QualifiedPropertyId, Probability>> map) {

		Map<TableColumnId, Map<QualifiedPropertyId, Probability>> finalMap = new ConcurrentHashMap<>();

		for (Entry<TableColumnId, Map<QualifiedPropertyId, Probability>> e : map.entrySet()) {

			finalMap.put(e.getKey(), Collections.unmodifiableMap(new ConcurrentHashMap<>(e.getValue())));
		}

		this.map = Collections.unmodifiableMap(finalMap);
	}

	@Override
	public TableScores getTableScores(ClassId classId, PropertyId propertyId, InstanceId instanceId) {

		return new QualifiedColumnTableScores(propertyId, map);
	}

	@Override
	public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {
		throw new IllegalStateException();
	}

	@Override
	public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {
		throw new IllegalStateException();
	}

	@Override
	public Probability getScore(TripleId tripleId, CellId cellId, Set<AbstractQualifier> qualifiers) {
		Map<QualifiedPropertyId, Probability> innerMap = this.map.get(new TableColumnId(cellId));
		if (innerMap != null) {
			return innerMap.get(new QualifiedPropertyId(tripleId.getPropertyId(), qualifiers));
		}

		return null;
	}

	@Override
	public boolean containsKey(TripleId tripleId, CellId cellId, Set<AbstractQualifier> qualifiers) {
		Map<QualifiedPropertyId, Probability> innerMap = this.map.get(new TableColumnId(cellId));
		if (innerMap != null) {
			if (innerMap.containsKey(new QualifiedPropertyId(tripleId.getPropertyId(), qualifiers))) {
				return true;
			}
		}

		return false;
	}
}
