package de.mannheim.uni.expansion.tablefusion.core.scoring.packs;

import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableMap.Builder;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;

public class QualifiedTripleIdScores implements ScoresContainer {
	private final Map<QualifiedTripleId, Map<CellId, Probability>> scores;

	public QualifiedTripleIdScores(Map<QualifiedTripleId, Map<CellId, Probability>> scores) {

		Builder<QualifiedTripleId, Map<CellId, Probability>> aa = ImmutableMap.builder();
		for (Entry<QualifiedTripleId, Map<CellId, Probability>> scoresEntry : scores.entrySet()) {

			aa.put(scoresEntry.getKey(), ImmutableMap.copyOf(scoresEntry.getValue()));
		}

		this.scores = aa.build();

	}

	// private FullScores(Map<ValueId, TableScoresImpl> scores, boolean check) {
	// this.scores = Collections.unmodifiableMap(scores);
	// }

	// public static FullScores fromTripleScores(Map<TripleId,
	// Map<AnnotatedTableId, Probability>> scores) {
	//
	// Map<ValueId, TableScoresImpl> changedScores = new HashMap<>();
	//
	// for (Entry<TripleId, Map<AnnotatedTableId, Probability>> e :
	// scores.entrySet()) {
	// ValueId vid = new ValueId(e.getKey());
	// TableScoresImpl ts = changedScores.get(vid);
	// if (ts != null) {
	// Map<AnnotatedTableId, Probability> newTsMap = new HashMap<>();
	//
	// for (Entry<AnnotatedTableId, Probability> entry : ts.entrySet()) {
	// newTsMap.put(entry.getKey(), entry.getValue());
	// }
	// newTsMap.putAll(e.getValue());
	// ts = new TableScoresImpl(newTsMap);
	// } else {
	// ts = new TableScoresImpl(e.getValue());
	// }
	//
	// changedScores.put(vid, ts);
	// }
	//
	// return new FullScores(changedScores, true);
	//
	// }

	public Probability getScore(TripleId tripleId, CellId cellId, Set<AbstractQualifier> qualifiers) {
		Map<CellId, Probability> ts = this.scores.get(new QualifiedTripleId(tripleId, qualifiers));
		if (ts != null) {
			Probability score = ts.get(cellId);
			return score;
		} else
			return null;

	}

	public boolean containsKey(TripleId tripleId, CellId cellId, Set<AbstractQualifier> qualifiers) {
		Map<CellId, Probability> ts = this.scores.get(new QualifiedTripleId(tripleId, qualifiers));
		if (ts != null)
			return ts.containsKey(cellId);
		else
			return false;

	}

	@Override
	public TableScores getTableScores(ClassId classId, PropertyId propertyId, InstanceId instanceId) {
		return new TableScoresInterface(this, classId, propertyId, instanceId);
	}

	@Override
	public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {
		throw new IllegalStateException();
	}

	@Override
	public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {
		throw new IllegalStateException();
	}

}
