package de.mannheim.uni.expansion.tablefusion.core.scoring.packs;

import java.util.HashMap;
import java.util.Map;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;

public class RegressionScores extends CombinedScores {
	private final double intercept;
	private final String[] names;
	private final ScoresContainer[] containers;
	private final double[] weights;

	public RegressionScores(String[] names, ScoresContainer[] containers, double[] weights, double intercept) {
		this.containers = containers;
		this.weights = weights;
		this.names = names;
		this.intercept = intercept;
	}

	@Override
	public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId id, ColumnIdx columnIdx, RowIdx rowIdx) {
		Probability[] probabilities = new Probability[containers.length];
		for (int i = 0; i < containers.length; i++) {
			Probability probability = containers[i].getScore(classId, propertyId, instanceId, id, columnIdx, rowIdx);
			;
			if (probability != null) {
				probabilities[i] = probability;
			} else {
				probabilities[i] = null;
			}
		}

		Probability prob = new Probability(probabilities, weights, intercept);
		// if (prob.get(SimilarityType.DIRECT).getAccuracy() < 0.0) {
		// return new Probability(new Accuracy(0.0,
		// prob.get(SimilarityType.DIRECT).getTotal()));
		// }
		return prob;
	}

	@Override
	public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tabledId, ColumnIdx columnIdx, RowIdx rowIdx) {
		for (ScoresContainer sc : containers) {
			if (sc.getScore(classId, propertyId, instanceId, tabledId, columnIdx, rowIdx) != null) {
				return true;
			}
		}
		return false;
	}

	// @Override
	// protected Set<AnnotatedTableId> keySet(ClassId classId, PropertyId
	// propertyId, InstanceId instanceId) {
	// Set<AnnotatedTableId> keySet = new HashSet<>();
	// for (ScoresContainer sc : containers) {
	// keySet.addAll(sc.getTableScores(classId, propertyId,
	// instanceId).keySet());
	// }
	// return keySet;
	// }

	public Map<String, Double> getWeights() {
		Map<String, Double> ret = new HashMap<>();
		for (int i = 0; i < names.length; i++) {
			ret.put(names[i], weights[i]);
		}
		return ret;
	}

	public Double getIntercept() {
		return intercept;
	}

}
