package de.mannheim.uni.expansion.tablefusion.core.scoring.packs;

import java.util.HashMap;
import java.util.Map;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;

public class RegressionScoresPropertyId extends CombinedScores {
	private final Map<PropertyId, Double> interceptsMap;
	private final String[] names;
	private final ScoresContainer[] containers;
	private final Map<PropertyId, double[]> weightsMap;

	public RegressionScoresPropertyId(String[] names, ScoresContainer[] containers, Map<PropertyId, double[]> weightsMap,
			Map<PropertyId, Double> interceptsMap) {
		this.containers = containers;
		this.weightsMap = weightsMap;
		this.names = names;
		this.interceptsMap = interceptsMap;
	}

	@Override
	public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId id, ColumnIdx columnIdx, RowIdx rowIdx) {
		if (this.containsKey(classId, propertyId, instanceId, id, columnIdx, rowIdx)) {
			double[] weights = weightsMap.get(propertyId);
			double intercept = interceptsMap.get(propertyId);

			Probability[] probabilities = new Probability[containers.length];
			for (int i = 0; i < containers.length; i++) {
				Probability probability = containers[i].getScore(classId, propertyId, instanceId, id, columnIdx, rowIdx);
				;
				if (probability != null) {
					probabilities[i] = probability;
				} else {
					probabilities[i] = null;
				}
			}

			Probability prob = new Probability(probabilities, weights, intercept);
			// if (prob.get(SimilarityType.DIRECT).getAccuracy() < 0.0) {
			// return new Probability(new Accuracy(0.0,
			// prob.get(SimilarityType.DIRECT).getTotal()));
			// }
			return prob;
		} else {
			return null;
		}
	}

	@Override
	public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tabledId, ColumnIdx columnIdx, RowIdx rowIdx) {
		if (!weightsMap.containsKey(propertyId)) {
			return false;
		}

		return true;
	}

	// @Override
	// protected Set<AnnotatedTableId> keySet(ClassId classId, PropertyId
	// propertyId, InstanceId instanceId) {
	// Set<AnnotatedTableId> keySet = new HashSet<>();
	// for (ScoresContainer sc : containers) {
	// keySet.addAll(sc.getTableScores(classId, propertyId,
	// instanceId).keySet());
	// }
	// return keySet;
	// }

	public Map<String, Double> getWeights(PropertyId propertyId) {
		Map<String, Double> ret = new HashMap<>();
		for (int i = 0; i < names.length; i++) {
			ret.put(names[i], weightsMap.get(propertyId)[i]);
		}
		return ret;
	}

	public Double getIntercept(PropertyId propertyId) {
		return interceptsMap.get(propertyId);
	}

}
