package de.mannheim.uni.expansion.tablefusion.core.scoring.packs;

import java.util.Map;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.tablefusion.analysis.Accuracy;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;
import de.mannheim.uni.expansion.tablefusion.core.similarity.Similarity.SimilarityType;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.WekaClassifier;
import weka.classifiers.Classifier;
import weka.core.Instance;
import weka.core.SparseInstance;

public class RegressionTreesScoresClassPropertyId extends CombinedScores {
	private final String[] names;
	private final ScoresContainer[] containers;
	private final Map<ClassPropertyId, WekaClassifier> modelMap;

	public RegressionTreesScoresClassPropertyId(String[] names, ScoresContainer[] containers, Map<ClassPropertyId, WekaClassifier> modelMap) {
		this.containers = containers;
		this.modelMap = modelMap;
		this.names = names;
	}

	@Override
	public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId id, ColumnIdx columnIdx, RowIdx rowIdx) {
		if (this.containsKey(classId, propertyId, instanceId, id, columnIdx, rowIdx)) {
			WekaClassifier wekaClassifier = modelMap.get(new ClassPropertyId(classId, propertyId));
			Classifier model = wekaClassifier.getClassifier();

			Probability[] z = new Probability[containers.length];

			for (int i = 0; i < containers.length; i++) {

				z[i] = containers[i].getScore(classId, propertyId, instanceId, id, columnIdx, rowIdx);

			}

			int countNull = 0;
			for (Probability s : z) {
				if (s == null) {
					countNull++;
				}
			}

			double[] instanceScores = new double[z.length - countNull];

			int[] index = new int[z.length - countNull];

			int j = 0;
			for (int i = 0; i < z.length; i++) {
				if (z[i] != null) {
					instanceScores[j] = z[i].get(SimilarityType.DIRECT).getAccuracy();
					index[j] = i;
					j++;
				}

			}

			Instance inst = new SparseInstance(1.0, instanceScores, index, containers.length);
			inst.setDataset(wekaClassifier.getDataset());

			try {
				return new Probability(new Accuracy(model.classifyInstance(inst), 1));
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
				return null;
			}

		} else {
			return null;
		}
	}

	@Override
	public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tabledId, ColumnIdx columnIdx, RowIdx rowIdx) {
		if (!modelMap.containsKey(new ClassPropertyId(classId, propertyId))) {
			return false;
		}

		return true;
	}

	// @Override
	// protected Set<AnnotatedTableId> keySet(ClassId classId, PropertyId
	// propertyId, InstanceId instanceId) {
	// Set<AnnotatedTableId> keySet = new HashSet<>();
	// for (ScoresContainer sc : containers) {
	// keySet.addAll(sc.getTableScores(classId, propertyId,
	// instanceId).keySet());
	// }
	// return keySet;
	// }

}
