package de.mannheim.uni.expansion.tablefusion.core.scoring.packs;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;

public class RequiredResizedScores extends CombinedScores {
	private final ScoresContainer main;
	private final ScoresContainer weights;

	public RequiredResizedScores(ScoresContainer main, ScoresContainer weights) {
		this.main = main;
		this.weights = weights;
	}

	@Override
	public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId id, ColumnIdx columnIdx, RowIdx rowIdx) {

		Probability probability = main.getScore(classId, propertyId, instanceId, id, columnIdx, rowIdx);
		Probability weight = weights.getScore(classId, propertyId, instanceId, id, columnIdx, rowIdx);
		if (probability == null) {
			return null;
		} else if (weight == null) {
			return null;
		} else {
			return probability.resize(weight);
		}

	}

	@Override
	public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tabledId, ColumnIdx columnIdx, RowIdx rowIdx) {
		if (main.getScore(classId, propertyId, instanceId, tabledId, columnIdx, rowIdx) != null
				&& weights.getScore(classId, propertyId, instanceId, tabledId, columnIdx, rowIdx) != null) {
			return true;
		}

		return false;
	}

	// @Override
	// protected Set<AnnotatedTableId> keySet(ClassId classId, PropertyId
	// propertyId, InstanceId instanceId) {
	// Set<AnnotatedTableId> combinedKeySet = new HashSet<>();
	// Set<AnnotatedTableId> mainKeySet = main.getTableScores(classId,
	// propertyId, instanceId).keySet();
	// Set<AnnotatedTableId> weightsKeySet = weights.getTableScores(classId,
	// propertyId, instanceId).keySet();
	//
	// for (AnnotatedTableId id : mainKeySet) {
	// if (weightsKeySet.contains(id)) {
	// combinedKeySet.add(id);
	// }
	// }
	// return combinedKeySet;
	// }

}
