package de.mannheim.uni.expansion.tablefusion.core.scoring.packs;

import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;

public class TableScoresInterface implements TableScores {
	private final ClassId classId;
	private final PropertyId propertyId;
	private final InstanceId instanceId;
	private final ScoresContainer scoresContainer;

	public TableScoresInterface(ScoresContainer scoresContainer, ClassId classId, PropertyId propertyId, InstanceId instanceId) {
		this.scoresContainer = scoresContainer;
		this.classId = classId;
		this.propertyId = propertyId;
		this.instanceId = instanceId;
	}

	@Override
	public Probability get(AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {
		return this.scoresContainer.getScore(classId, propertyId, instanceId, tableId, columnIdx, rowIdx);
	}

	@Override
	public boolean containsKey(AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {
		return this.scoresContainer.containsKey(classId, propertyId, instanceId, tableId, columnIdx, rowIdx);
	}

	@Override
	public Probability get(CellId cellId, Set<AbstractQualifier> qualifiers) {
		return this.scoresContainer.getScore(new TripleId(classId, propertyId, instanceId), cellId, qualifiers);
	}

	@Override
	public boolean containsKey(CellId cellId, Set<AbstractQualifier> qualifiers) {
		return this.scoresContainer.containsKey(new TripleId(classId, propertyId, instanceId), cellId, qualifiers);
	}

}
