package de.mannheim.uni.expansion.tablefusion.core.scoring.packs;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;

public class TripleIdScores implements ScoresContainer {
	private final Map<TripleId, TableScores> scores;

	public TripleIdScores(Map<TripleId, Map<CellId, Probability>> scores) {

		Map<TripleId, TableScores> changedScores = new HashMap<>();

		for (Entry<TripleId, Map<CellId, Probability>> e : scores.entrySet()) {
			changedScores.put(e.getKey(), new TableScoresImpl(e.getValue()));
		}

		this.scores = Collections.unmodifiableMap(changedScores);

	}

	// private FullScores(Map<ValueId, TableScoresImpl> scores, boolean check) {
	// this.scores = Collections.unmodifiableMap(scores);
	// }

	// public static FullScores fromTripleScores(Map<TripleId,
	// Map<AnnotatedTableId, Probability>> scores) {
	//
	// Map<ValueId, TableScoresImpl> changedScores = new HashMap<>();
	//
	// for (Entry<TripleId, Map<AnnotatedTableId, Probability>> e :
	// scores.entrySet()) {
	// ValueId vid = new ValueId(e.getKey());
	// TableScoresImpl ts = changedScores.get(vid);
	// if (ts != null) {
	// Map<AnnotatedTableId, Probability> newTsMap = new HashMap<>();
	//
	// for (Entry<AnnotatedTableId, Probability> entry : ts.entrySet()) {
	// newTsMap.put(entry.getKey(), entry.getValue());
	// }
	// newTsMap.putAll(e.getValue());
	// ts = new TableScoresImpl(newTsMap);
	// } else {
	// ts = new TableScoresImpl(e.getValue());
	// }
	//
	// changedScores.put(vid, ts);
	// }
	//
	// return new FullScores(changedScores, true);
	//
	// }

	@Override
	public TableScores getTableScores(ClassId classId, PropertyId propertyId, InstanceId instanceId) {
		return scores.get(new TripleId(classId, propertyId, instanceId));
	}

	@Override
	public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId id, ColumnIdx columnIdx, RowIdx rowIdx) {
		TableScores ts = this.scores.get(new TripleId(classId, propertyId, instanceId));
		if (ts != null) {
			Probability score = ts.get(id, columnIdx, rowIdx);
			return score;
		} else
			return null;
	}

	@Override
	public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId id, ColumnIdx columnIdx, RowIdx rowIdx) {
		TableScores ts = this.scores.get(new TripleId(classId, propertyId, instanceId));
		if (ts != null)
			return ts.containsKey(id, columnIdx, rowIdx);
		else
			return false;
	}

}
