package de.mannheim.uni.expansion.tablefusion.core.scoring.packs;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;

public class WeightedScores extends CombinedScores {
	private final ScoresContainer[] containers;
	private final Double[] weights;

	public WeightedScores(ScoresContainer[] containers, Double[] weights) {
		this.containers = containers;
		this.weights = weights;
	}

	@Override
	public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId id, ColumnIdx columnIdx, RowIdx rowIdx) {
		Probability[] probabilities = new Probability[containers.length];
		for (int i = 0; i < containers.length; i++) {
			Probability probability = containers[i].getScore(classId, propertyId, instanceId, id, columnIdx, rowIdx);
			if (probability != null) {
				probabilities[i] = probability;
			} else {
				probabilities[i] = null;
			}
		}
		return new Probability(probabilities, weights);
	}

	@Override
	public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tabledId, ColumnIdx columnIdx, RowIdx rowIdx) {
		for (ScoresContainer sc : containers) {
			if (sc.getScore(classId, propertyId, instanceId, tabledId, columnIdx, rowIdx) != null) {
				return true;
			}
		}
		return false;
	}

	// @Override
	// protected Set<AnnotatedTableId> keySet(ClassId classId, PropertyId
	// propertyId, InstanceId instanceId) {
	// Set<AnnotatedTableId> keySet = new HashSet<>();
	// for (ScoresContainer sc : containers) {
	// keySet.addAll(sc.getTableScores(classId, propertyId,
	// instanceId).keySet());
	// }
	// return keySet;
	// }

}
