package de.mannheim.uni.expansion.tablefusion.core.scoring.packs.timestamps;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.tablefusion.analysis.Accuracy;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;
import de.mannheim.uni.expansion.tablefusion.core.scoring.TimeStampsScoringCalculator.TS_METHODS;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.TableScores;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.TableScoresInterface;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;

public class TimestampScoresContainer {

	Map<Pair<TS_METHODS, Boolean>, Integer> indexMap = new HashMap<>();

	Map<Pair<QualifiedTripleId, CellId>, Double[]> valueMap = new ConcurrentHashMap<>();

	public TimestampScores get(TS_METHODS method, boolean propagated) {
		return new TimestampScores(method, propagated);
	}

	public TimestampScoresContainer(Map<Pair<QualifiedTripleId, CellId>, Map<Pair<TS_METHODS, Boolean>, Double>> values) {
		int size = 0;
		for (TS_METHODS method : TS_METHODS.values()) {
			indexMap.put(new Pair<>(method, false), size);
			size++;
		}
		for (TS_METHODS method : TS_METHODS.values()) {
			indexMap.put(new Pair<>(method, true), size);
			size++;
		}
		final int finalLength = size;

		PersistentParallel6.runAndFinish(values.entrySet(), (e) -> {
			Double[] valueArray = new Double[finalLength];

			for (Entry<Pair<TS_METHODS, Boolean>, Double> ie : e.getValue().entrySet()) {
				valueArray[indexMap.get(ie.getKey())] = ie.getValue();
			}

			valueMap.put(e.getKey(), valueArray);

		} , this.getClass());

	}

	public class TimestampScores implements ScoresContainer {
		// private final TS_METHOD method;
		// private final boolean propogated;
		private final int index;

		private TimestampScores(TS_METHODS method, boolean propogated) {
			// this.method = method;
			// this.propogated = propogated;
			this.index = indexMap.get(new Pair<>(method, propogated));
		}

		@Override
		public TableScores getTableScores(ClassId classId, PropertyId propertyId, InstanceId instanceId) {
			return new TableScoresInterface(this, classId, propertyId, instanceId);
		}

		public Probability getScore(TripleId tripleId, CellId cellId, Set<AbstractQualifier> qualifiers) {
			Double[] valueArray = valueMap.get(new Pair<>(new QualifiedTripleId(tripleId, qualifiers), cellId));
			if (valueArray == null) {
				return null;
			} else {
				Double d = valueArray[index];
				if (d == null) {
					return null;
				} else {
					return new Probability(Accuracy.simple(d));
				}
			}
		}

		public boolean containsKey(TripleId tripleId, CellId cellId, Set<AbstractQualifier> qualifiers) {
			Double[] valueArray = valueMap.get(new Pair<>(new QualifiedTripleId(tripleId, qualifiers), cellId));
			if (valueArray == null) {
				return false;
			} else {
				Double d = valueArray[index];
				if (d == null) {
					return false;
				}
			}
			return true;
		}

		@Override
		public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx,
				RowIdx rowIdx) {
			throw new IllegalStateException();
		}

		@Override
		public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx,
				RowIdx rowIdx) {
			throw new IllegalStateException();
		}
	}

}
