package de.mannheim.uni.expansion.tablefusion.core.scoring.packs.tripleids;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.TableScores;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.cellscores.CellScoresByTable;

public class TripleIdTableScores implements ScoresContainer {
	private final Map<TripleId, TableScores> scores;

	public TripleIdTableScores(Map<TripleId, Map<AnnotatedTableId, Probability>> scores) {

		Map<TripleId, TableScores> changedScores = new HashMap<>();

		for (Entry<TripleId, Map<AnnotatedTableId, Probability>> e : scores.entrySet()) {
			changedScores.put(e.getKey(), new CellScoresByTable(e.getValue()));
		}

		this.scores = Collections.unmodifiableMap(changedScores);

	}

	@Override
	public TableScores getTableScores(ClassId classId, PropertyId propertyId, InstanceId instanceId) {
		return scores.get(new TripleId(classId, propertyId, instanceId));
	}

	@Override
	public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId id, ColumnIdx columnIdx, RowIdx rowIdx) {
		TableScores ts = this.scores.get(new TripleId(classId, propertyId, instanceId));
		if (ts != null) {
			Probability score = ts.get(id, columnIdx, rowIdx);
			return score;
		} else
			return null;
	}

	@Override
	public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId id, ColumnIdx columnIdx, RowIdx rowIdx) {
		TableScores ts = this.scores.get(new TripleId(classId, propertyId, instanceId));
		if (ts != null)
			return ts.containsKey(id, columnIdx, rowIdx);
		else
			return false;
	}

}
