package de.mannheim.uni.expansion.tablefusion.core.splitting;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ThreadLocalRandom;
import java.util.function.Consumer;
import java.util.function.Function;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.SynchronizedMatchedValuesMapMerger;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.qualifiers.TimestampPointInTime;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.evaluation.timestamps.TimestampQualifiersAdapter;

public class NonDuplicateSplitter implements Splitter {
	private final KnowledgeBase knowledgeBase;
	private final TimestampQualifiersAdapter timestampQualifiersAdapter;
	private final SimilarityCalculator sc;

	@Inject
	public NonDuplicateSplitter(KnowledgeBase knowledgeBase, TimestampQualifiersAdapter timestampQualifiersAdapter, SimilarityCalculator sc) {
		this.knowledgeBase = knowledgeBase;
		this.timestampQualifiersAdapter = timestampQualifiersAdapter;
		this.sc = sc;
	}

	@Override
	public SplitGroupedValues split(GroupedValues groupedValues) {

		ConcurrentHashMap<ValueId, Integer> changesCount = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(groupedValues.keySet(), (key) -> {
			int numOfChanges = 0;
			QualifiedValueList qvl = knowledgeBase.getQualifiedValueList(key.getValueId());
			if (qvl != null) {
				Map<TimestampPointInTime, List<Integer>> yearMap = timestampQualifiersAdapter.getQualifierMapFromQVL(qvl);

				Function<Integer, AbstractObject> getValue = (year) -> {
					List<Integer> listYear = yearMap.get(new TimestampPointInTime(year));
					if (listYear != null && !listYear.isEmpty()) {
						return qvl.getClaim(listYear.get(0)).getValue();
					}
					return null;
				};

				AbstractObject value2008 = getValue.apply(2008);
				AbstractObject value2010 = getValue.apply(2010);
				AbstractObject value2012 = getValue.apply(2012);
				AbstractObject value2014 = getValue.apply(2014);

				if (value2008 != null && value2010 != null) {
					if (!sc.isEqual(value2008, value2010))
						numOfChanges++;
				}

				if (value2010 != null && value2012 != null) {
					if (!sc.isEqual(value2010, value2012))
						numOfChanges++;
				}

				if (value2012 != null && value2014 != null) {
					if (!sc.isEqual(value2012, value2014))
						numOfChanges++;
				}

				int sadf = 123;
			}
			changesCount.put(key.getValueId(), numOfChanges);

		}, this.getClass());

		SynchronizedMatchedValuesMapMerger<QualifiedTripleId> merger = new SynchronizedMatchedValuesMapMerger<>();
		PersistentParallel6.runAndFinish(groupedValues.entrySet(), (e) -> {

			// if (changesCount.getOrDefault(e.getKey().getValueId(), 0) >= 2) {
			QualifiedValueList valueList = knowledgeBase.getQualifiedValueList(e.getKey().getValueId());
			if (valueList != null) {
				Map<TimestampPointInTime, List<Integer>> qualifiers = timestampQualifiersAdapter.getQualifierMapFromQVL(valueList);

				List<Pair<TimestampPointInTime, AbstractObject>> listToChooseFrom = new ArrayList<>();

				Consumer<Integer> fillList = (year) -> {
					TimestampPointInTime tpit = new TimestampPointInTime(year);
					List<Integer> indeces = qualifiers.get(tpit);
					if (indeces != null && !indeces.isEmpty()) {
						listToChooseFrom.add(new Pair<>(tpit, valueList.getClaim(indeces.get(0)).getValue()));
					}
				};

				fillList.accept(2008);
				fillList.accept(2010);
				fillList.accept(2012);
				fillList.accept(2014);

				if (listToChooseFrom.isEmpty()) {
					return;
				}

				int index = ThreadLocalRandom.current().nextInt(listToChooseFrom.size());

				Pair<TimestampPointInTime, AbstractObject> q = listToChooseFrom.get(index);
				HashSet<AbstractQualifier> qualifierSet = new HashSet<>();
				qualifierSet.add(q.getFirst());
				merger.merge(new QualifiedTripleId(e.getKey(), qualifierSet), e.getValue());

			}
			// }
		}, "SIMPLE SPLITTER");

		return new SplitGroupedValues(merger.getCleanMap());

	}

	@Override
	public Set<QualifiedPropertyId> split(ClassId classId, PropertyId propertyId, CellId cellId) {
		throw new IllegalStateException();
	}

}
