package de.mannheim.uni.expansion.tablefusion.core.splitting;

import java.util.HashSet;
import java.util.Set;
import java.util.UUID;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.SynchronizedMatchedValuesMapMerger;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleConcurrentCache;

public class NonTimeDependentSplitter implements Splitter {

	private static final SimpleConcurrentCache<UUID, SplitGroupedValues> cache = new SimpleConcurrentCache<>();

	@Inject
	public NonTimeDependentSplitter() {

	}

	@Override
	public SplitGroupedValues split(GroupedValues groupedValues) {
		if (cache.hasCacheFor(groupedValues.getUUID())) {
			return cache.get(groupedValues.getUUID());
		} else {

			SynchronizedMatchedValuesMapMerger<QualifiedTripleId> merger = new SynchronizedMatchedValuesMapMerger<>();
			PersistentParallel6.runAndFinish(groupedValues.entrySet(), (e) -> {

				merger.merge(new QualifiedTripleId(e.getKey(), new HashSet<>()), e.getValue());

			} , this.getClass());
			SplitGroupedValues spgv = new SplitGroupedValues(merger.getCleanMap());
			cache.add(groupedValues.getUUID(), spgv);
			return spgv;
		}
	}

	@Override
	public Set<QualifiedPropertyId> split(ClassId classId, PropertyId propertyId, CellId cellId) {
		QualifiedPropertyId qpid = new QualifiedPropertyId(propertyId, new HashSet<>());
		Set<QualifiedPropertyId> ret = new HashSet<>();
		ret.add(qpid);
		return ret;
	}

}
