package de.mannheim.uni.expansion.tablefusion.core.splitting;

import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.UUID;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.SynchronizedMatchedValuesMapMerger;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.mannheim.uni.expansion.utils.SimpleConcurrentCache;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.qualifiers.TimestampPointInTime;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.evaluation.timestamps.TimestampQualifiersAdapter;

public class SimpleSplitter implements Splitter {

	private final KnowledgeBase knowledgeBase;
	private final TimestampQualifiersAdapter timestampQualifiersAdapter;

	private static final SimpleConcurrentCache<UUID, SplitGroupedValues> cache = new SimpleConcurrentCache<>();

	@Inject
	public SimpleSplitter(KnowledgeBase knowledgeBase, TimestampQualifiersAdapter timestampQualifiersAdapter) {
		this.knowledgeBase = knowledgeBase;
		this.timestampQualifiersAdapter = timestampQualifiersAdapter;
	}

	@Override
	public SplitGroupedValues split(GroupedValues groupedValues) {
		if (cache.hasCacheFor(groupedValues.getUUID())) {
			return cache.get(groupedValues.getUUID());
		} else {

			SynchronizedMatchedValuesMapMerger<QualifiedTripleId> merger = new SynchronizedMatchedValuesMapMerger<>();
			PersistentParallel4.runAndFinish(groupedValues.entrySet(), (e) -> {

				QualifiedValueList valueList = knowledgeBase.getQualifiedValueList(e.getKey().getValueId());
				if (valueList != null) {
					Map<TimestampPointInTime, List<Integer>> qualifiers = timestampQualifiersAdapter.getQualifierMapFromQVL(valueList);
					List<Integer> noQualifiers = timestampQualifiersAdapter.getNonQualifierMapFromQVL(valueList);

					for (Entry<TimestampPointInTime, List<Integer>> q : qualifiers.entrySet()) {

						HashSet<AbstractQualifier> qualifierSet = new HashSet<>();
						qualifierSet.add(q.getKey());
						merger.merge(new QualifiedTripleId(e.getKey(), qualifierSet), e.getValue());

					}

					if (noQualifiers != null && !noQualifiers.isEmpty()) {
						Set<ClassId> classes = knowledgeBase.getClasses(e.getKey().getInstanceId());
						merger.merge(new QualifiedTripleId(e.getKey(), new HashSet<>()), e.getValue());

					}
				}

			}, "SIMPLE SPLITTER");
			SplitGroupedValues spgv = new SplitGroupedValues(merger.getCleanMap());
			cache.add(groupedValues.getUUID(), spgv);
			return spgv;
		}
	}

	@Override
	public Set<QualifiedPropertyId> split(ClassId classId, PropertyId propertyId, CellId cellId) {
		throw new IllegalStateException();
	}

}
