package de.mannheim.uni.expansion.tablefusion.core.splitting;

import java.util.HashMap;
import java.util.Map;

public class SplitterFactoryRepository {
	private final Map<Class<? extends SplitterFactory>, SplitterFactory> repository = new HashMap<>();

	private SplitterFactory defaultSplitterGetter;

	public <E extends SplitterFactory> void addDefault(E splitterFactory) {
		this.defaultSplitterGetter = splitterFactory;
	}

	public <E extends SplitterFactory> void add(Class<E> clazz, E splitterFactory) {
		this.repository.put(clazz, splitterFactory);
	}

	public SplitterFactory get(Class<? extends SplitterFactory> clazz) {
		if (clazz == null) {
			if (defaultSplitterGetter == null) {
				throw new IllegalStateException("Asking for Splitter of class null but no default Splitter is configured!");
			}
			return defaultSplitterGetter;
		} else {
			SplitterFactory scorer = this.repository.get(clazz);
			if (scorer == null) {
				throw new IllegalStateException("No Splitter was registered for Class: " + clazz.getSimpleName());
			} else {
				return scorer;
			}
		}
	}
}
