package de.mannheim.uni.expansion.tablefusion.evaluation.analyzer;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.ListType;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.tablefusion.adapters.ranking.RankRepositoryImpl;
import de.mannheim.uni.expansion.tablefusion.analysis.AccuracyCalculator;
import de.mannheim.uni.expansion.tablefusion.analysis.MaxCorrectFinder;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.PerformancePR;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.PerformanceResultsWriter;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.ResultsCategory;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.scoring.TimeStampsScoringCalculator;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ColumnScores;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.HostScores;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.InstanceScores;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.RegressionScores;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.RegressionScoresClassPropertyId;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.TableScores;
import de.mannheim.uni.expansion.tablefusion.core.splitting.FilteredSplitter;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.splitting.Splitter;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitterFactory;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitterFactoryRepository;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig.StrategiesConfigBuilder;
import de.mannheim.uni.expansion.tablefusion.learning.LearningSetCreatorStratified;
import de.mannheim.uni.expansion.tablefusion.learning.ListExpansionFilterCreator;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;
import de.mannheim.uni.expansion.tablefusion.optimization.ScoreFusionThresholdFinder;
import de.mannheim.uni.expansion.tablefusion.optimization.ScoreFusionThresholdFinderFullQualifiedSimple;
import de.mannheim.uni.expansion.tablefusion.services.GroupedValuesFilterer;
import de.mannheim.uni.expansion.tablefusion.services.ScoresFilterer;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionDataType;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionPreproccesser;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionRunner;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategy;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategy.FusionStrategyId;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategyRepository;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.TestingMethod;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.ScorerRepository;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.SplitterProvider;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.newTS.Scorer_NewTs_QKBT_without_CellRow;
import de.mannheim.uni.expansion.tablefusion.services.grouping.GroupedValuesRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.informatik.dws.expansion.evaluation.FilteredValues;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdContainer;

public class AccuracyAnalyzer {

	private final GroupedValuesRepository groupedValuesRepository;
	private final ScoreFusionThresholdFinder scoreFusionThresholdFinder;

	private final FusionRunner fusionRunner;

	private final AccuracyCalculator accuracyCalculator;

	private final OntologyRepository ontologyRepository;

	private final GroupedValuesFilterer listTypeFilterer;

	private final ListExpansionFilterCreator listExpansionFilterCreator;

	private final ScorerRepository scorerRepository;
	private final ScoresFilterer scoresFilterer;

	private final KnowledgeBase dbpedia;
	private final SplitterFactoryRepository splitterFactoriesRepository;
	private final MaxCorrectFinder maxCorrectFinder;

	private final LearningSetCreatorStratified learningSetCreatorStratified;

	private final ScoreFusionThresholdFinderFullQualifiedSimple scoreFusionThresholdFinderFullQualifiedSimple;

	private final GroupedValuesFiltererNew groupedValuesFilterer;

	@Inject
	public AccuracyAnalyzer(GroupedValuesRepository groupedValuesRepository, ScoreFusionThresholdFinder scoreFusionThresholdFinder, FusionRunner fusionRunner,
			AccuracyCalculator accuracyCalculator, OntologyRepository ontologyRepository, GroupedValuesFilterer listTypeFilterer,
			ListExpansionFilterCreator listExpansionFilterCreator, ScorerRepository scorerRepository, ScoresFilterer scoresFilterer, KnowledgeBase dbpedia,
			SplitterFactoryRepository splitterFactoriesRepository, MaxCorrectFinder maxCorrectFinder, LearningSetCreatorStratified learningSetCreatorStratified,
			ScoreFusionThresholdFinderFullQualifiedSimple scoreFusionThresholdFinderFullQualifiedSimple, GroupedValuesFiltererNew groupedValuesFilterer) {
		super();
		this.groupedValuesRepository = groupedValuesRepository;
		this.scoreFusionThresholdFinder = scoreFusionThresholdFinder;
		this.fusionRunner = fusionRunner;
		this.accuracyCalculator = accuracyCalculator;
		this.ontologyRepository = ontologyRepository;
		this.listTypeFilterer = listTypeFilterer;
		this.listExpansionFilterCreator = listExpansionFilterCreator;
		this.scorerRepository = scorerRepository;
		this.scoresFilterer = scoresFilterer;
		this.dbpedia = dbpedia;
		this.splitterFactoriesRepository = splitterFactoriesRepository;
		this.maxCorrectFinder = maxCorrectFinder;
		this.learningSetCreatorStratified = learningSetCreatorStratified;
		this.scoreFusionThresholdFinderFullQualifiedSimple = scoreFusionThresholdFinderFullQualifiedSimple;
		this.groupedValuesFilterer = groupedValuesFilterer;
	}

	public void getPerformance(String dataset, FusionConfig fusionConfig, StrategiesConfig strategiesConfig, PropertiesConfig propertiesConfig,
			String resultsFolder) throws IOException {
		String folderPath = resultsFolder + "/" + dataset;
		MKDIR.createDirUnchecked(new File(folderPath));

		dfmt.applyPattern(pattern);

		GroupedValues groupedValues = groupedValuesRepository.getGroupedValues(fusionConfig);
		if (propertiesConfig != null)
			groupedValues = groupedValuesFilterer.filter(groupedValues, propertiesConfig);
		// Set<PropertyId> propertySet = new HashSet<>();
		// propertySet.add(new
		// PropertyId("http://dbpedia.org/ontology/capital"));
		// propertySet.add(new
		// PropertyId("http://dbpedia.org/ontology/PopulatedPlace/areaTotal"));
		// propertySet.add(new
		// PropertyId("http://dbpedia.org/ontology/populationTotal"));
		// propertySet.add(new
		// PropertyId("http://dbpedia.org/ontology/populationDensity"));

		// groupedValues = this.listTypeFilterer.getByProperty(propertySet,
		// groupedValues);
		// for (ClassId classId : this.dbpedia.getClasses()) {
		// propertySet.addAll(this.dbpedia.getProperties(classId).keySet());
		// }

		// groupedValues = this.listTypeFilterer.getByProperty(propertySet,
		// groupedValues);

		PerformanceResultsWriter performanceResultsWriter = new PerformanceResultsWriter();

		analyzeAccuracy(dataset, folderPath, performanceResultsWriter, "", groupedValues, true, fusionConfig, strategiesConfig,
				PropertiesConfig.defaultConfig());

		System.out.println("Original: " + TimeStampsScoringCalculator.original.toString());
		System.out.println("Propogated: " + TimeStampsScoringCalculator.propogated.toString());

		System.out.println("PageRankTotal: " + RankRepositoryImpl.countTotal.get());
		System.out.println("PageRankMissing: " + RankRepositoryImpl.countMissing.get());

	}

	@Deprecated
	public void getPerformance(String dataset, FusionConfig fusionConfig, String resultsFolder) throws IOException {
		String folderPath = resultsFolder + "/" + dataset;
		MKDIR.createDirUnchecked(new File(folderPath));

		dfmt.applyPattern(pattern);

		StrategiesConfigBuilder rebuilder = StrategiesConfig.builder();

		// rebuilder.addStrategy(FusionStrategyRepository.NORMAL);
		// // // // //
		// rebuilder.addStrategy(FusionStrategyRepository.KBT_ClassPropertyId_Threshold);
		// //
		// rebuilder.addStrategy(FusionStrategyRepository.TS_KBT_REGRESSION_CLASS_PROPERTY);
		// //
		// rebuilder.addStrategy(FusionStrategyRepository.TS_PROPGATED_KBT_REGRESSION_CLASS_PROPERTY);
		// //
		// //
		// addStrategy.accept(FusionStrategyRepository.Timestamps_Regression_ClassProperty);
		// //
		// addStrategy.accept(FusionStrategyRepository.Timestamps_Regression_ClassProperty_Propagated);
		//
		// //
		// rebuilder.addStrategy(FusionStrategyRepository.QKBT_ClassPropertyId_Threshold);
		//
		// rebuilder.addStrategy(FusionStrategyRepository.QKBT_Mapped);
		// // //
		rebuilder.addStrategy(FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL);
		// rebuilder.addStrategy(FusionStrategyRepository.NEW_TS_QKBT_WITHOUT_ROW_CELL);

		// rebuilder.addStrategy(FusionStrategyRepository.NEW_TS_QKBT_WITHOUR_ROW_CELL_PROPAGATED);
		// rebuilder.addStrategy(FusionStrategyRepository.NEW_TS_QKBT);
		// rebuilder.addStrategy(FusionStrategyRepository.NEW_TS_QKBT_PROPAGATED);

		// addStrategy.accept(FusionStrategyRepository.KBT_PRPGTD_ClassPropertyId_Threshold);
		// addStrategy.accept(FusionStrategyRepository.QKBT_Propogated_ClassPropertyId_Threshold);

		//
		// addStrategy.accept(FusionStrategyRepository.Regression_ClassProperty);
		// addStrategy.accept(FusionStrategyRepository.Regression_ClassProperty_Propagated);

		// addStrategy.accept(FusionStrategyRepository.PageRank_Regression_ClassProperty);
		//
		// addStrategy.accept(FusionStrategyRepository.PageRank_TS_ClassProperty);
		// addStrategy.accept(FusionStrategyRepository.PageRank_TS_Propagated_ClassProperty);
		// addStrategy.accept(FusionStrategyRepository.PageRank_KBT_Regression_ClassProperty);
		// addStrategy.accept(FusionStrategyRepository.FullRegression);
		// addStrategy.accept(FusionStrategyRepository.FullRegressionPropagated);

		// // addStrategy.accept(FusionStrategyRepository.RANDOM);
		//
		// addStrategy.accept(FusionStrategyRepository.KBT);

		// addStrategy.accept(FusionStrategyRepository.KBT_PropertyId_Threshold);
		// addStrategy.accept(FusionStrategyRepository.Regression);
		// addStrategy.accept(FusionStrategyRepository.Regression_Propagated);
		// addStrategy.accept(FusionStrategyRepository.Regression_Multiplied_Propagated);
		// addStrategy.accept(FusionStrategyRepository.Regression_Single_Propagated);

		// addStrategy.accept(FusionStrategyRepository.Regression_ClassProperty_Multiplied_Propagated);
		// addStrategy.accept(FusionStrategyRepository.Regression_ClassProperty_Single_Propagated);
		// addStrategy.accept(FusionStrategyRepository.Regression_PropertyId);
		// // //
		// addStrategy.accept(FusionStrategyRepository.RegressionTrees);
		//
		// addStrategy.accept(FusionStrategyRepository.RegressionTrees_ClassProperty);
		// //
		// addStrategy.accept(FusionStrategyRepository.RegressionTrees_PropertyId);
		// //
		// addStrategy.accept(FusionStrategyRepository.Timestamps_Regression);

		// addStrategy.accept(FusionStrategyRepository.Timestamps_Regression_Propagated);
		// addStrategy.accept(FusionStrategyRepository.Timestamps_Regression_PropertyId);
		// addStrategy.accept(FusionStrategyRepository.Combined_HierarchyTS_TableText);
		// //
		// addStrategy.accept(FusionStrategyRepository.KBT_Combined_HierarchyTS_TableText_Weighted);
		// addStrategy.accept(FusionStrategyRepository.KBT_Combined_HierarchyTS_TableText_Multiplied);
		// //
		// addStrategy.accept(FusionStrategyRepository.KBT_Combined_HierarchyTS_TableText_WeightedMultiplied);

		// addStrategy.accept(FusionStrategyRepository.PageRank);
		// addStrategy.accept(FusionStrategyRepository.PageRank_Regression);

		// addStrategy.accept(FusionStrategyRepository.PageRank_KBT_Regression);

		// addStrategy.accept(FusionStrategyRepository.CellTS);
		// addStrategy.accept(FusionStrategyRepository.CellTS_Propagated);
		//
		// addStrategy.accept(FusionStrategyRepository.ColumnTS);
		// addStrategy.accept(FusionStrategyRepository.ColumnTS_Propagated);
		// addStrategy.accept(FusionStrategyRepository.KBT_ColumnTS_Max_Weighted);
		// addStrategy.accept(FusionStrategyRepository.KBT_ColumnTS_Max_Multiplied);
		// //
		// addStrategy.accept(FusionStrategyRepository.TableTextTS);
		// addStrategy.accept(FusionStrategyRepository.TableTextTS_Propagated);
		// //
		// addStrategy.accept(FusionStrategyRepository.KBT_TableTextTS_MostCommon_Weighted);
		// //
		// addStrategy.accept(FusionStrategyRepository.KBT_TableTextTS_MostCommon_Multiplied);
		// //
		// addStrategy.accept(FusionStrategyRepository.HierarchyTS);
		// addStrategy.accept(FusionStrategyRepository.HierarchyTS_Propagated);
		// addStrategy.accept(FusionStrategyRepository.KBT_HierarchyTS_Weighted);
		// addStrategy.accept(FusionStrategyRepository.KBT_HierarchyTS_Multiplied);
		// //
		// addStrategy.accept(FusionStrategyRepository.FullHierarchyTS);
		// addStrategy.accept(FusionStrategyRepository.FullHierarchyTS_Propagated);
		// addStrategy.accept(FusionStrategyRepository.KBT_FullHierarchyTS_Weighted);
		// addStrategy.accept(FusionStrategyRepository.KBT_FullHierarchyTS_Multiplied);
		// //
		// addStrategy.accept(FusionStrategyRepository.PageTitleTs);
		// addStrategy.accept(FusionStrategyRepository.PageTitleTs_Propagated);
		// //
		// addStrategy.accept(FusionStrategyRepository.KBT_PageTitleTS_Weighted);
		// //
		// addStrategy.accept(FusionStrategyRepository.KBT_PageTitleTS_Multiplied);
		// //
		// addStrategy.accept(FusionStrategyRepository.TableTitleTS);
		// addStrategy.accept(FusionStrategyRepository.TableTitleTS_Propagated);
		// //
		// addStrategy.accept(FusionStrategyRepository.KBT_TableTitleTS_Weighted);
		// //
		// addStrategy.accept(FusionStrategyRepository.KBT_TableTitleTS_Multiplied);
		// //
		// addStrategy.accept(FusionStrategyRepository.TableYearsHierarchy);
		// addStrategy.accept(FusionStrategyRepository.TableYearsHierarchy_Propagated);
		// //
		// addStrategy.accept(FusionStrategyRepository.KBT_TableYearsHierarch_Weighted);
		// // //
		// //
		// addStrategy.accept(FusionStrategyRepository.KBT_TableYearsHierarch_Resized);
		// //
		// addStrategy.accept(FusionStrategyRepository.KBT_TableYearsHierarch_Multiplied);

		// addStrategy.accept(FusionStrategyRepository.CS_WEIGHTED);
		// addStrategy.accept(FusionStrategyRepository.CS_LIST_EXPANSION);
		// leNormal.add(FusionStrategyRepository.NORMAL);

		GroupedValues groupedValues = groupedValuesRepository.getGroupedValues(fusionConfig);

		Set<PropertyId> propertySet = new HashSet<>();
		// propertySet.add(new
		// PropertyId("http://dbpedia.org/ontology/capital"));
		// propertySet.add(new
		// PropertyId("http://dbpedia.org/ontology/PopulatedPlace/areaTotal"));
		// propertySet.add(new
		// PropertyId("http://dbpedia.org/ontology/populationTotal"));
		// propertySet.add(new
		// PropertyId("http://dbpedia.org/ontology/populationDensity"));

		// groupedValues = this.listTypeFilterer.getByProperty(propertySet,
		// groupedValues);
		for (ClassId classId : this.dbpedia.getClasses()) {
			propertySet.addAll(this.dbpedia.getProperties(classId).keySet());
		}

		// groupedValues = this.listTypeFilterer.getByProperty(propertySet,
		// groupedValues);

		PerformanceResultsWriter performanceResultsWriter = new PerformanceResultsWriter();

		analyzeAccuracy(dataset, folderPath, performanceResultsWriter, "", groupedValues, true, fusionConfig, rebuilder.build(),
				PropertiesConfig.defaultConfig());

		System.out.println("Original: " + TimeStampsScoringCalculator.original.toString());
		System.out.println("Propogated: " + TimeStampsScoringCalculator.propogated.toString());

		System.out.println("PageRankTotal: " + RankRepositoryImpl.countTotal.get());
		System.out.println("PageRankMissing: " + RankRepositoryImpl.countMissing.get());

		// getPerformanceHierarchyTimeStampsFiltered();
		// getPerformanceTimeStampsFilteredColumnScores();
	}

	private void getPerformanceHierarchyTimeStampsFiltered(String resultsFolder) throws IOException {
		String folderPath = resultsFolder + "/" + "timestampsHiearchyFiltered";
		MKDIR.createDirUnchecked(new File(folderPath));

		dfmt.applyPattern(pattern);

		StrategiesConfigBuilder builder = StrategiesConfig.builder();

		builder.addStrategy(FusionStrategyRepository.NORMAL);
		// addStrategy.accept(FusionStrategyRepository.RANDOM);
		builder.addStrategy(FusionStrategyRepository.KBT);
		builder.addStrategy(FusionStrategyRepository.KBT_ClassPropertyId_Threshold);
		builder.addStrategy(FusionStrategyRepository.KBT_PropertyId_Threshold);
		builder.addStrategy(FusionStrategyRepository.TS_KBT_REGRESSION);
		builder.addStrategy(FusionStrategyRepository.TS_KBT_REGRESSION_CLASS_PROPERTY);
		builder.addStrategy(FusionStrategyRepository.Regression_PropertyId);

		builder.addStrategy(FusionStrategyRepository.HierarchyTS);
		builder.addStrategy(FusionStrategyRepository.KBT_HierarchyTS_Weighted);
		// addStrategy.accept(FusionStrategyRepository.KBT_HierarchyTS_Resized);
		builder.addStrategy(FusionStrategyRepository.KBT_HierarchyTS_Multiplied);
		// addStrategy.accept(FusionStrategyRepository.CS_WEIGHTED);
		// addStrategy.accept(FusionStrategyRepository.CS_LIST_EXPANSION);
		// leNormal.add(FusionStrategyRepository.NORMAL);

		GroupedValues groupedValues = groupedValuesRepository.getGroupedValues(null);
		groupedValues = scoresFilterer.filterByHiearicalTimeStamps(groupedValues);
		PerformanceResultsWriter performanceResultsWriter = new PerformanceResultsWriter();

		analyzeAccuracy("", folderPath, performanceResultsWriter, "HierarchyFiltered_", groupedValues, true, FusionConfig.defaultConfig(), builder.build(),
				PropertiesConfig.defaultConfig());

	}

	private void getPerformanceTimeStampsFilteredColumnScores(String resultsFolder) throws IOException {
		String folderPath = resultsFolder + "/" + "timestampsColumnFiltered";
		MKDIR.createDirUnchecked(new File(folderPath));

		dfmt.applyPattern(pattern);

		StrategiesConfigBuilder buidler = StrategiesConfig.builder();

		buidler.addStrategy(FusionStrategyRepository.NORMAL);
		// addStrategy.accept(FusionStrategyRepository.RANDOM);
		buidler.addStrategy(FusionStrategyRepository.KBT);
		buidler.addStrategy(FusionStrategyRepository.KBT_ClassPropertyId_Threshold);
		buidler.addStrategy(FusionStrategyRepository.KBT_PropertyId_Threshold);
		buidler.addStrategy(FusionStrategyRepository.TS_KBT_REGRESSION);
		buidler.addStrategy(FusionStrategyRepository.TS_KBT_REGRESSION_CLASS_PROPERTY);
		buidler.addStrategy(FusionStrategyRepository.Regression_PropertyId);
		buidler.addStrategy(FusionStrategyRepository.ColumnTS);
		buidler.addStrategy(FusionStrategyRepository.KBT_ColumnTS_Max_Weighted);
		//
		buidler.addStrategy(FusionStrategyRepository.KBT_ColumnTS_Max_Resized);
		buidler.addStrategy(FusionStrategyRepository.KBT_ColumnTS_Max_Multiplied);
		// // addStrategy.accept(FusionStrategyRepository.CS_WEIGHTED);
		// // addStrategy.accept(FusionStrategyRepository.CS_LIST_EXPANSION);
		// // leNormal.add(FusionStrategyRepository.NORMAL);

		GroupedValues groupedValues = groupedValuesRepository.getGroupedValues();
		groupedValues = scoresFilterer.filterByColumnTimeStamps(groupedValues);
		PerformanceResultsWriter performanceResultsWriter = new PerformanceResultsWriter();

		analyzeAccuracy("", folderPath, performanceResultsWriter, "ColumnFiltered_", groupedValues, true, FusionConfig.defaultConfig(), buidler.build(),
				PropertiesConfig.defaultConfig());

	}

	public void getPerformanceWithLists(String resultsFolder) throws IOException {
		String folderPath = resultsFolder + "/" + "lists";
		MKDIR.createDirUnchecked(new File(folderPath));

		dfmt.applyPattern(pattern);

		StrategiesConfigBuilder buidler = StrategiesConfig.builder();

		buidler.addStrategy(FusionStrategyRepository.NORMAL);
		// addStrategy.accept(FusionStrategyRepository.RANDOM);
		buidler.addStrategy(FusionStrategyRepository.KBT);
		// addStrategy.accept(FusionStrategyRepository.CS_WEIGHTED);
		buidler.addStrategy(FusionStrategyRepository.CS_LIST_EXPANSION);
		buidler.addStrategy(FusionStrategyRepository.NORMAL);

		GroupedValues groupedValues = groupedValuesRepository.getGroupedValues();

		// mediumListTypes = listTypeFilterer.getDataType(ColumnDataType.string,
		// mediumListTypes);
		PerformanceResultsWriter performanceResultsWriter = new PerformanceResultsWriter();
		GroupedValues nonListTypes = listTypeFilterer.getNonListTypes(groupedValues);
		analyzeAccuracy("", folderPath, performanceResultsWriter, "NON-Lists_", nonListTypes, true, FusionConfig.defaultConfig(), buidler.build(),
				PropertiesConfig.defaultConfig());
		//
		// Map<TripleId, MatchedValues> shortListTypes =
		// listTypeFilterer.getListTypes(ListType.SHORT, groupedValues);
		// analyzeAccuracy("ShortLists-Creation_", methodTypes, shortListTypes,
		// direct, cv, new ArrayList<>(), new ArrayList<>(), false);
		//
		GroupedValues mediumListTypes = listTypeFilterer.getListTypes(ListType.MEDIUM, groupedValues);
		// mediumListTypes =
		// listTypeFilterer.getDataType(ColumnDataType.reference,
		// mediumListTypes);
		// analyzeAccuracy("", folderPath, performanceResultsWriter,
		// "MediumLists-Creation_", mediumListTypes, direct, new ArrayList<>(),
		// cv,
		// new ArrayList<>(), new ArrayList<>(), true, true,
		// FusionConfig.defaultConfig());
		// analyzeAccuracy("", folderPath, performanceResultsWriter,
		// "MediumLists-Expansion_", mediumListTypes, new ArrayList<>(), new
		// ArrayList<>(),
		// new ArrayList<>(), leNormal, leCV, true, true,
		// FusionConfig.defaultConfig());
		// //
		// GroupedValues longListTypes =
		// listTypeFilterer.getListTypes(ListType.LONG, groupedValues);
		// analyzeAccuracy("", folderPath, performanceResultsWriter,
		// "LongLists-Creation_", longListTypes, direct, new ArrayList<>(), cv,
		// new ArrayList<>(), new ArrayList<>(), true, true,
		// FusionConfig.defaultConfig());
		// analyzeAccuracy("", folderPath, performanceResultsWriter,
		// "LongLists-Expansion_", longListTypes, new ArrayList<>(), new
		// ArrayList<>(),
		// new ArrayList<>(), leNormal, leCV, true, true,
		// FusionConfig.defaultConfig());
		//
		// GroupedValues allListsType =
		// listTypeFilterer.getListTypes(groupedValues);
		// analyzeAccuracy("", folderPath, performanceResultsWriter,
		// "AllLists-Creation_", allListsType, direct, new ArrayList<>(), cv,
		// new ArrayList<>(), new ArrayList<>(), true, true,
		// FusionConfig.defaultConfig());
		// analyzeAccuracy("", folderPath, performanceResultsWriter,
		// "AllLists-Expansion_", allListsType, new ArrayList<>(), new
		// ArrayList<>(),
		// new ArrayList<>(), leNormal, leCV, true, true,
		// FusionConfig.defaultConfig());
	}

	public void getPerformanceWWW(String resultsFolder) throws IOException {

		String folderPath = resultsFolder + "/" + "www";

		MKDIR.createDirUnchecked(new File(folderPath));

		dfmt.applyPattern(pattern);

		StrategiesConfigBuilder buidler = StrategiesConfig.builder();

		buidler.addStrategy(FusionStrategyRepository.NORMAL);
		// addStrategy.accept(FusionStrategyRepository.RANDOM);
		buidler.addStrategy(FusionStrategyRepository.KBT);
		// addStrategy.accept(FusionStrategyRepository.CS_WEIGHTED);
		buidler.addStrategy(FusionStrategyRepository.PR_PROBABILITIES);

		GroupedValues groupedValues = groupedValuesRepository.getGroupedValues();

		// mediumListTypes = listTypeFilterer.getDataType(ColumnDataType.string,
		// mediumListTypes);
		PerformanceResultsWriter performanceResultsWriter = new PerformanceResultsWriter();
		analyzeAccuracy("", folderPath, performanceResultsWriter, "Fusion_", groupedValues, true, FusionConfig.defaultConfig(), buidler.build(),
				PropertiesConfig.defaultConfig());
	}

	public Map<FusionStrategyId, FusionPerformance<PerformancePR>> analyzeAccuracy(String dataset, String folderPath,
			PerformanceResultsWriter performanceResultsWriter, String prefix, GroupedValues groupedValues, boolean write, FusionConfig fusionConfig,
			StrategiesConfig strategiesConfig, PropertiesConfig propertiesConfig) throws IOException {

		List<FusionDataType> methodTypes = strategiesConfig.getDataTypes();
		Map<FusionStrategyId, FusionPerformance<PerformancePR>> ret = new HashMap<>();

		for (FusionStrategy fs : strategiesConfig.getFusionStrategies(TestingMethod.DIRECT)) {
			getPerformance(dataset, prefix, performanceResultsWriter, groupedValues, fs, fusionConfig, methodTypes);
		}

		for (FusionStrategy fs : strategiesConfig.getFusionStrategies(TestingMethod.DIRECT_SCORED)) {
			getPerformanceWithScores(dataset, performanceResultsWriter, groupedValues, fs, null, fs.getName(), fusionConfig, methodTypes);
		}

		List<FusionStrategy> cv = strategiesConfig.getFusionStrategies(TestingMethod.CROSS_VALIDATED);
		if (!cv.isEmpty()) {
			int length = cv.size();
			ret.putAll(getPerformanceWithCrossValidation(dataset, prefix, performanceResultsWriter, groupedValues, cv.toArray(new FusionStrategy[length]),
					fusionConfig, methodTypes, propertiesConfig));
		}

		List<FusionStrategy> leNormal = strategiesConfig.getFusionStrategies(TestingMethod.LIST_EXPANSION_DIRECT);
		if (!leNormal.isEmpty()) {
			testListExpansionNormal(dataset, prefix, performanceResultsWriter, groupedValues, leNormal.toArray(new FusionStrategy[leNormal.size()]),
					fusionConfig, methodTypes);
		}

		List<FusionStrategy> leCV = strategiesConfig.getFusionStrategies(TestingMethod.LIST_EXPANSION_CROSS_VALIDATED);
		if (!leCV.isEmpty()) {
			testListExpansionCrossValidation(dataset, prefix, performanceResultsWriter, groupedValues, leCV.toArray(new FusionStrategy[leCV.size()]),
					fusionConfig, methodTypes);
		}

		if (write) {
			for (ResultsCategory rc : ResultsCategory.values()) {
				FileWriter fw = new FileWriter(folderPath + "/" + rc.name() + ".csv", true);
				fw.write(performanceResultsWriter.getNoHeader(rc, 0));
				fw.close();

			}

			FileWriter fw = new FileWriter(folderPath + "/AVERAGE_TOTAL.csv", true);
			fw.write(accuracyCalculator.getAverageClassPropertiesAsTotalAsCSV(ret));
			fw.close();

		}

		return ret;
	}

	Locale locale = new Locale("en", "UK");
	String pattern = "###.##";
	DecimalFormat dfmt = (DecimalFormat) NumberFormat.getNumberInstance(locale);

	public Map<FusionStrategyId, FusionPerformance<PerformancePR>> getPerformanceWithCrossValidation(String dataset, String prefix,
			PerformanceResultsWriter performanceResultsWriter, GroupedValues groupedValues, FusionStrategy[] fusionRunTypes, FusionConfig fusionConfig,
			List<FusionDataType> methodTypes, PropertiesConfig propertiesConfig) throws IOException {

		int length = fusionRunTypes.length;

		Map<FusionStrategyId, FusionPerformance<PerformancePR>> ret = new HashMap<>();

		SplitterFactory splitterFactory = this.splitterFactoriesRepository.get(fusionConfig.getSplitterFactory());

		Splitter splitter = splitterFactory.getSplitter(groupedValues, groupedValues, groupedValues);

		SplitGroupedValues splitGroupedValues = splitter.split(groupedValues);

		if (fusionConfig.getSplitGroupedValuesFilter() != null)
			splitGroupedValues = fusionConfig.getSplitGroupedValuesFilter().filter(splitGroupedValues);

		// Map<Integer, CreatedSets> createdSetsMapOld =
		// learningSetsRepository.getCreatedSets(4, groupedValues);
		// Map<Integer,
		// de.mannheim.uni.expansion.tablefusion.learning.LearningSetCreatorFilling.SplitGroupedValuesSets>
		// createdSetsMap = new HashMap<>();
		// for (Entry<Integer, CreatedSets> e : createdSetsMapOld.entrySet()) {
		// createdSetsMap.put(e.getKey(), new
		// de.mannheim.uni.expansion.tablefusion.learning.LearningSetCreatorFilling.SplitGroupedValuesSets(
		// splitGroupedValues, splitter.split(e.getValue().getLearningSet()),
		// splitter.split(e.getValue().getTestingSet())));
		// }
		//

		Map<Integer, SplitGroupedValuesSets> createdSetsMap = this.learningSetCreatorStratified.createByTableProperty(4, splitGroupedValues);

		// Map<Integer, SplitGroupedValuesSets> createdSetsMapNew = new
		// HashMap<>();
		// createdSetsMapNew.put(0, createdSetsMap.get(0));
		//
		// createdSetsMap = createdSetsMapNew;
		// Map<Integer,
		// de.mannheim.uni.expansion.tablefusion.learning.LearningSetCreatorFilling.CreatedSets>
		// createdSetsMap = learningSetCreatorFilling
		// .createByTableProperty(2, splitGroupedValues);

		// ThresholdContainer[] probabilityThresholds = new
		// ThresholdContainer[length];

		Scorer[] scorers = new Scorer[length];
		SplitterFactory[] splitterFactories = new SplitterFactory[length];
		String[] names = new String[length];
		{
			int i = 0;
			for (FusionStrategy strategy : fusionRunTypes) {
				scorers[i] = this.scorerRepository.get(strategy.getScoringMethod());
				splitterFactories[i] = this.splitterFactoriesRepository.get(strategy.getSplitterFactory());
				// ThresholdContainer threshold =
				// scoreFusionThresholdFinder.findOptimumThresholdCrossValidation(createdSetsMap,
				// strategy, false);
				// probabilityThresholds[i] = threshold;
				names[i] = prefix + strategy.getName();

				i++;
			}
		}

		ThresholdContainer[] thresholdContainer = new ThresholdContainer[length];

		System.out.println();
		System.out.println();
		// Match and group all the values

		List<FusionDataType> dataTypes = new ArrayList<>();
		dataTypes.addAll(methodTypes);

		List<FusionPreproccesser> preprocessors = new ArrayList<>();
		preprocessors.add(FusionPreproccesser.NONE);

		@SuppressWarnings("unchecked")
		List<FusionPerformance<PerformancePR>>[] performanceList = new List[length];

		for (int i = 0; i < length; i++) {
			performanceList[i] = new ArrayList<>();
		}

		for (int i = 0; i < length; i++) {
			System.out.println("Threshold" + fusionRunTypes[i].getName() + "running");
			thresholdContainer[i] = scoreFusionThresholdFinderFullQualifiedSimple.findOptimumThresholdCrossValidation(createdSetsMap, scorers[i],
					fusionRunTypes[i].getThresholdType(), dataTypes, preprocessors, fusionRunTypes[i], fusionConfig, propertiesConfig);
			System.out.println("Threshold" + fusionRunTypes[i].getName() + "done");
			System.out.println(thresholdContainer[i]);
		}

		@SuppressWarnings("unchecked")
		Map<ClassPropertyId, Map<String, MutableInt>>[] timestampCount = new Map[createdSetsMap.size()];

		Map<FusionStrategy, List<RegressionScores>> rsListMap = new HashMap<>();
		Map<FusionStrategy, List<RegressionScoresClassPropertyId>> rsCPListMap = new HashMap<>();
		for (int i = 0; i < length; i++) {
			FusionStrategy fusionRunType = fusionRunTypes[i];
			int count = 0;
			for (Entry<Integer, de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets> createdSets : createdSetsMap.entrySet()) {

				System.out.println("i: " + count + ", Strategy: " + fusionRunType.getName());

				Scorer scoresGetter = scorers[i];

				SplitGroupedValues sgvLearning = createdSets.getValue().getLearningSet();
				SplitGroupedValues sgvTesting = createdSets.getValue().getTestingSet();

				ScoresContainer probabilities = scoresGetter.getScores(createdSets.getValue());

				if (scoresGetter instanceof Scorer_NewTs_QKBT_without_CellRow) {
					Scorer_NewTs_QKBT_without_CellRow s = (Scorer_NewTs_QKBT_without_CellRow) scoresGetter;
					timestampCount[count] = s.getMappingEvaluationCSV(createdSets.getValue());

				}

				count++;

				if (probabilities instanceof RegressionScores) {
					List<RegressionScores> rsList = CHMHelper.getFromMap(rsListMap, fusionRunType, () -> new ArrayList<>());
					rsList.add((RegressionScores) probabilities);
				}

				if (probabilities instanceof RegressionScoresClassPropertyId) {
					List<RegressionScoresClassPropertyId> rsCPList = CHMHelper.getFromMap(rsCPListMap, fusionRunType, () -> new ArrayList<>());
					rsCPList.add((RegressionScoresClassPropertyId) probabilities);
				}

				SplitGroupedValues sgvTestingForFusion = sgvTesting;
				if (fusionConfig.isReSplitTestingSet() && scoresGetter instanceof SplitterProvider) {
					System.out.println("RESPLITING");
					FilteredSplitter filteredSplitter = ((SplitterProvider) scoresGetter).getSplitter(createdSets.getValue());
					sgvTestingForFusion = filteredSplitter.split(groupedValues, sgvTesting.keySet());
				}

				// ThresholdContainer probabilityThreshold =
				// scoreFusionThresholdFinderUsingScorerQualified.findOptimumThresholdCrossValidation(sgvLearning,
				// probabilities, fusionRunType.getThresholdType(), dataTypes,
				// preprocessors, fusionRunType, fuseLists);

				ThresholdContainer probabilityThreshold = thresholdContainer[i];

				Map<FusionRun, Map<QualifiedTripleId, FusedValue>> fusedValuesMap = fusionRunner.run(dataTypes, preprocessors, fusionRunType,
						sgvTestingForFusion, probabilities, probabilityThreshold, fusionConfig, propertiesConfig);

				Map<QualifiedTripleId, FusedValue> fusedValues = CHMHelper.flattenMap(fusedValuesMap);

				FusionPerformance<Integer> maxCorrect = maxCorrectFinder.getMaxCorrect(sgvTesting, null, propertiesConfig);

				FusionPerformance<PerformancePR> fusionPerformance = accuracyCalculator.analyzeAccuracy(fusedValues, maxCorrect);
				performanceList[i].add(fusionPerformance);
			}

			for (Entry<FusionStrategy, List<RegressionScores>> e : rsListMap.entrySet()) {
				List<RegressionScores> rsList = e.getValue();
				Double intercept = rsList.get(0).getIntercept();
				Map<String, Double> weights = rsList.get(0).getWeights();
				// for (int i = 1; i < rsList.size(); i++) {
				Map<String, Double> weights2 = rsList.get(i).getWeights();

				for (Entry<String, Double> entry : weights.entrySet()) {
					entry.setValue(entry.getValue() + weights2.get(entry.getKey()));
				}

				intercept += rsList.get(i).getIntercept();
				// }

				System.out.println(e.getKey().getName());
				System.out.println(" - INTERCEPT: " + (intercept / rsList.size()));
				for (Entry<String, Double> entry : weights.entrySet()) {
					System.out.println(" - " + entry.getKey() + ": " + (entry.getValue() / rsList.size()));
				}
			}

			// try {
			// for (Entry<FusionStrategy, List<RegressionScoresClassPropertyId>>
			// e : rsCPListMap.entrySet()) {
			// List<RegressionScoresClassPropertyId> rsList = e.getValue();
			// System.out.println(e.getKey().getName());
			// Map<ClassPropertyId, Map<String, Double>> weightsMap = new
			// HashMap<>(rsList.get(0).getWeightsMap());
			// Map<ClassPropertyId, Double> interceptsMap = new
			// HashMap<>(rsList.get(0).getInterceptsMap());
			//
			// for (Entry<ClassPropertyId, Map<String, Double>> cpEntry :
			// weightsMap.entrySet()) {
			// Map<String, Double> weights = cpEntry.getValue();
			// int subtract = 0;
			// // for (int i = 1; i < rsList.size(); i++) {
			// Map<String, Double> weights2 =
			// rsList.get(i).getWeights(cpEntry.getKey());
			// if (weights2 != null) {
			//
			// for (Entry<String, Double> entry : weights.entrySet()) {
			// entry.setValue(entry.getValue() + weights2.get(entry.getKey()));
			// }
			//
			// interceptsMap.put(cpEntry.getKey(),
			// interceptsMap.get(cpEntry.getKey()) +
			// rsList.get(i).getIntercept(cpEntry.getKey()));
			// } else {
			// subtract++;
			// }
			// // }
			//
			// System.out.println(" - " + cpEntry.getKey());
			// System.out.println(" - INTERCEPT: " +
			// (interceptsMap.get(cpEntry.getKey()) / (rsList.size() -
			// subtract)));
			// for (Entry<String, Double> entry : weights.entrySet()) {
			// System.out.println(" - - " + entry.getKey() + ": " +
			// (entry.getValue() / (rsList.size() - subtract)));
			// }
			// }
			//
			// }
			// } catch (Exception e) {
			// e.printStackTrace();
			// }

			// FusionStrategy fusionRunType = fusionRunTypes[i];
			// Scorer scoresGetter = scorers[i];
			//
			// SplitGroupedValues sgv = splitGroupedValues;
			//
			// ScoresContainer probabilities = scoresGetter.getScores(new
			// SplitGroupedValuesSets(sgv, sgv, sgv));
			//
			// // ThresholdContainer probabilityThreshold =
			// //
			// scoreFusionThresholdFinderUsingScorerQualified.findOptimumThresholdCrossValidation(sgv,
			// // probabilities,
			// // fusionRunType.getThresholdType(), dataTypes, preprocessors,
			// // fusionRunType, fuseLists);
			//
			// ThresholdContainer probabilityThreshold =
			// scoreFusionThresholdFinderFullQualifiedSimple.findOptimumThresholdCrossValidation(sgv,
			// scoresGetter,
			// fusionRunType.getThresholdType(), dataTypes, preprocessors,
			// fusionRunType, fuseLists);
			//
			// Map<FusionRun, Map<QualifiedTripleId, FusedValue>> fusedValuesMap
			// = fusionRunner.run(dataTypes, preprocessors, fusionRunType, sgv,
			// probabilities,
			// fuseLists, probabilityThreshold);
			//
			// Map<QualifiedTripleId, FusedValue> fusedValues =
			// CHMHelper.flattenMap(fusedValuesMap);
			//
			// FusionPerformance<Integer> maxCorrect =
			// maxCorrectFinder.getMaxCorrect(sgv, fuseLists, null);
			//
			// FusionPerformance<PerformancePR> performance =
			// accuracyCalculator.analyzeAccuracy(sgv, fusedValues, maxCorrect,
			// fuseLists);

			FusionPerformance<PerformancePR> averagePerformance = accuracyCalculator.averageAccuracies(performanceList[i]);

			ret.put(fusionRunType.getFusionStrategyId(), averagePerformance);
			System.out.println();
			System.out.println();
			System.out.println();
			System.out.println();

			// System.out.println("Threshold" + probabilityThreshold);

			System.out.println(names[i]);

			System.out.println();
			if (performanceResultsWriter != null) {
				performanceResultsWriter.writePerformance(ontologyRepository, dataset, names[i], averagePerformance, averagePerformance, true, true, false,
						false);
			}

		}

		if (timestampCount[0] != null) {
			evaluateTsMappingCSV(timestampCount);
		}

		return ret;
		// for (int i = 0; i < length; i++) {
		//
		// }

	}

	private void evaluateTsMappingCSV(Map<ClassPropertyId, Map<String, MutableInt>>[] timestampCount) throws IOException {

		Map<ClassPropertyId, Map<String, MutableInt>> fullMap = new HashMap<>();
		Map<String, MutableInt> totalMap = new HashMap<>();

		for (int i = 0; i < timestampCount.length; i++) {
			Map<ClassPropertyId, Map<String, MutableInt>> map = timestampCount[i];

			for (Entry<ClassPropertyId, Map<String, MutableInt>> entry : map.entrySet()) {
				Map<String, MutableInt> innerMap = CHMHelper.getFromMap(fullMap, entry.getKey(), () -> new HashMap<>());

				for (Entry<String, MutableInt> innerEntry : entry.getValue().entrySet()) {
					CHMHelper.getFromMap(innerMap, innerEntry.getKey(), () -> new MutableInt()).add(innerEntry.getValue());
					CHMHelper.getFromMap(totalMap, innerEntry.getKey(), () -> new MutableInt()).add(innerEntry.getValue());
				}
			}

		}

		StringWriter sw = new StringWriter();
		CSVWriter writer = new CSVWriter(sw);
		int length = 3;
		String[] headerLine = new String[length];
		headerLine[0] = "TimestampType";
		headerLine[1] = "Count";
		headerLine[2] = "percent";
		writer.writeNext(headerLine);

		double total = totalMap.entrySet().stream().mapToInt((e) -> {
			return e.getValue().intValue();
		}).sum();

		for (Entry<String, MutableInt> entry : totalMap.entrySet()) {
			String[] line = new String[length];
			line[0] = entry.getKey();
			line[1] = String.valueOf(entry.getValue().intValue());
			line[2] = String.valueOf(entry.getValue().intValue() / total);
			writer.writeNext(line);
		}

		String ret = sw.toString();
		writer.close();

		System.out.println(ret);

		sw = new StringWriter();
		writer = new CSVWriter(sw);
		length = 5;
		headerLine = new String[length];
		headerLine[0] = "Class";
		headerLine[1] = "Property";
		headerLine[2] = "TimestampType";
		headerLine[3] = "Count";
		headerLine[4] = "Percent";
		writer.writeNext(headerLine);

		for (Entry<ClassPropertyId, Map<String, MutableInt>> parentEntry : fullMap.entrySet()) {
			String[] line = new String[length];
			line[0] = parentEntry.getKey().getClassId().getClassId();
			line[1] = parentEntry.getKey().getPropertyId().getReadableString();

			total = parentEntry.getValue().entrySet().stream().mapToInt((e) -> {
				return e.getValue().intValue();
			}).sum();

			for (Entry<String, MutableInt> entry : parentEntry.getValue().entrySet()) {

				line[2] = entry.getKey();
				line[3] = String.valueOf(entry.getValue().intValue());
				line[4] = String.valueOf(entry.getValue().intValue() / total);
				writer.writeNext(line);
			}
		}

		ret = sw.toString();
		writer.close();

		System.out.println(ret);

	}

	private void testListExpansionCrossValidation(String dataset, String prefix, PerformanceResultsWriter performanceResultsWriter, GroupedValues groupedValues,
			FusionStrategy[] fusionRunTypes, FusionConfig fusionConfig, List<FusionDataType> methodTypes) throws IOException {

		Map<Integer, FilteredValues> setsFilters = listExpansionFilterCreator.createByKBList(8);

		int length = fusionRunTypes.length;

		ThresholdContainer[] probabilityThresholds = new ThresholdContainer[length];
		Scorer[] scorers = new Scorer[length];
		String[] names = new String[length];
		{
			int i = 0;
			for (FusionStrategy strategy : fusionRunTypes) {
				scorers[i] = this.scorerRepository.get(strategy.getScoringMethod());
				ThresholdContainer threshold = scoreFusionThresholdFinder.findOptimumThresholdListExpansion(setsFilters, groupedValues, strategy, false);
				probabilityThresholds[i] = threshold;
				names[i] = prefix + "_" + strategy.getName() + "_" + dfmt.format(threshold) + "_THRESHOLD";

				i++;
			}
		}

		System.out.println();
		System.out.println();
		// Match and group all the values

		List<FusionDataType> dataTypes = new ArrayList<>();
		dataTypes.addAll(methodTypes);

		List<FusionPreproccesser> preprocessors = new ArrayList<>();
		preprocessors.add(FusionPreproccesser.NONE);

		@SuppressWarnings("unchecked")
		List<FusionPerformance<PerformancePR>>[] performanceList = new List[length];

		for (int i = 0; i < length; i++) {
			performanceList[i] = new ArrayList<>();
		}

		for (Entry<Integer, FilteredValues> setsFilter : setsFilters.entrySet()) {

			GroupedValues filteredGroupedValues = listTypeFilterer.filterBySetFilter(setsFilter.getValue(), groupedValues);

			for (int i = 0; i < length; i++) {
				FusionStrategy fusionRunType = fusionRunTypes[i];
				Scorer scoresGetter = scorers[i];
				ThresholdContainer probabilityThreshold = probabilityThresholds[i];

				ScoresContainer probabilities = null;
				// ScoresContainer probabilities =
				// scoresGetter.getScores(filteredGroupedValues, null, null);

				Map<FusionRun, Map<TripleId, FusedValue>> fusedValuesMap = fusionRunner.run(dataTypes, preprocessors, fusionRunType, groupedValues,
						probabilities, fusionConfig, probabilityThreshold, PropertiesConfig.defaultConfig());

				Map<TripleId, FusedValue> fusedValues = CHMHelper.flattenMap(fusedValuesMap);

				FusionPerformance<Integer> maxCorrect = maxCorrectFinder.getMaxCorrect(groupedValues, null, PropertiesConfig.defaultConfig());

				FusionPerformance<PerformancePR> fusionPerformance = accuracyCalculator.analyzeAccuracy(fusedValues, maxCorrect, setsFilter.getValue());
				performanceList[i].add(fusionPerformance);
			}

		}

		for (int i = 0; i < length; i++) {
			FusionStrategy fusionRunType = fusionRunTypes[i];
			Scorer scoresGetter = scorers[i];
			ThresholdContainer probabilityThreshold = probabilityThresholds[i];

			// ScoresContainer probabilities =
			// scoresGetter.getScores(groupedValues, null, null);
			ScoresContainer probabilities = null;
			Map<FusionRun, Map<TripleId, FusedValue>> fusedValuesMap = fusionRunner.run(dataTypes, preprocessors, fusionRunType, groupedValues, probabilities,
					fusionConfig, probabilityThreshold, PropertiesConfig.defaultConfig());

			Map<TripleId, FusedValue> fusedValues = CHMHelper.flattenMap(fusedValuesMap);

			FusionPerformance<Integer> maxCorrect = maxCorrectFinder.getMaxCorrect(groupedValues, null, PropertiesConfig.defaultConfig());

			FusionPerformance<PerformancePR> performance = accuracyCalculator.analyzeAccuracy(fusedValues, maxCorrect);

			FusionPerformance<PerformancePR> averagePerformance = accuracyCalculator.averageAccuracies(performanceList[i]);

			System.out.println();
			System.out.println();
			System.out.println();
			System.out.println();

			// System.out.println("Threshold" + probabilityThreshold);

			System.out.println(names[i]);

			System.out.println();
			performanceResultsWriter.writePerformance(ontologyRepository, dataset, names[i], performance, averagePerformance, true, true, false, false);

		}

	}

	private void testListExpansionNormal(String dataset, String prefix, PerformanceResultsWriter performanceResultsWriter, GroupedValues groupedValues,
			FusionStrategy[] fusionRunTypes, FusionConfig fusionConfig, List<FusionDataType> methodTypes) throws IOException {

		Map<Integer, FilteredValues> setsFilters = listExpansionFilterCreator.createByKBList(5);

		int length = fusionRunTypes.length;

		// Double[] probabilityThresholds = new Double[length];
		// ScoresGetter[] scoresGetters = new ScoresGetter[length];
		String[] names = new String[length];
		{
			int i = 0;
			for (FusionStrategy strategy : fusionRunTypes) {
				// scoresGetters[i] = strategy.getScoresGetter();
				// double threshold =
				// scoreFusionThresholdFinder.findOptimumThresholdListExpansion(setsFilters,
				// groupedValues, strategy, scoresGetters[i], false);
				// probabilityThresholds[i] = threshold;
				names[i] = prefix + "_" + strategy.getName() + "LIST_TESTING";

				i++;
			}
		}

		System.out.println();
		System.out.println();
		// Match and group all the values

		List<FusionDataType> dataTypes = new ArrayList<>();
		dataTypes.addAll(methodTypes);

		List<FusionPreproccesser> preprocessors = new ArrayList<>();
		preprocessors.add(FusionPreproccesser.NONE);

		@SuppressWarnings("unchecked")
		List<FusionPerformance<PerformancePR>>[] performanceList = new List[length];

		for (int i = 0; i < length; i++) {
			performanceList[i] = new ArrayList<>();
		}

		for (Entry<Integer, FilteredValues> setsFilter : setsFilters.entrySet()) {

			for (int i = 0; i < length; i++) {
				FusionStrategy fusionRunType = fusionRunTypes[i];
				// ScoresGetter scoresGetter = scoresGetters[i];
				// Double probabilityThreshold = probabilityThresholds[i];

				// probabilities =
				// scoresGetter.getScores(setsFilter.getKey().toString(),
				// instanceScores, columnScores, tableScores, hostScores);

				Map<FusionRun, Map<TripleId, FusedValue>> fusedValuesMap = fusionRunner.run(dataTypes, preprocessors, fusionRunType, groupedValues,
						fusionConfig);

				Map<TripleId, FusedValue> fusedValues = CHMHelper.flattenMap(fusedValuesMap);

				FusionPerformance<Integer> maxCorrect = maxCorrectFinder.getMaxCorrect(groupedValues, null, PropertiesConfig.defaultConfig());

				FusionPerformance<PerformancePR> fusionPerformance = accuracyCalculator.analyzeAccuracy(fusedValues, maxCorrect, setsFilter.getValue());
				performanceList[i].add(fusionPerformance);
			}

		}

		for (int i = 0; i < length; i++) {
			FusionStrategy fusionRunType = fusionRunTypes[i];
			// ScoresGetter scoresGetter = scoresGetters[i];
			// Double probabilityThreshold = probabilityThresholds[i];

			// probabilities = scoresGetter.getScores("", instanceScores,
			// columnScores, tableScores, hostScores);

			Map<FusionRun, Map<TripleId, FusedValue>> fusedValuesMap = fusionRunner.run(dataTypes, preprocessors, fusionRunType, groupedValues, fusionConfig);

			Map<TripleId, FusedValue> fusedValues = CHMHelper.flattenMap(fusedValuesMap);

			FusionPerformance<Integer> maxCorrect = maxCorrectFinder.getMaxCorrect(groupedValues, null, PropertiesConfig.defaultConfig());

			FusionPerformance<PerformancePR> performance = accuracyCalculator.analyzeAccuracy(fusedValues, maxCorrect);

			FusionPerformance<PerformancePR> averagePerformance = accuracyCalculator.averageAccuracies(performanceList[i]);

			System.out.println();
			System.out.println();
			System.out.println();
			System.out.println();

			// System.out.println("Threshold" + probabilityThreshold);

			System.out.println(names[i]);

			System.out.println();
			performanceResultsWriter.writePerformance(ontologyRepository, dataset, names[i], performance, averagePerformance, true, true, false, false);

		}

	}

	public void getPerformance(String dataset, String prefix, PerformanceResultsWriter performanceResultsWriter, GroupedValues groupedValues,
			FusionStrategy fusionRunType, FusionConfig fusionConfig, List<FusionDataType> methodTypes) throws IOException {

		System.out.println();
		System.out.println();
		// Match and group all the values

		List<FusionDataType> dataTypes = new ArrayList<>();
		dataTypes.addAll(methodTypes);

		List<FusionPreproccesser> preprocessors = new ArrayList<>();
		preprocessors.add(FusionPreproccesser.NONE);

		SplitterFactory splitterFactory = this.splitterFactoriesRepository.get(null);

		Splitter splitter = splitterFactory.getSplitter(groupedValues, groupedValues, groupedValues);

		SplitGroupedValues sgv = splitter.split(groupedValues);

		Map<FusionRun, Map<QualifiedTripleId, FusedValue>> fusedValuesMap = fusionRunner.run(dataTypes, preprocessors, fusionRunType, sgv, fusionConfig);

		Map<QualifiedTripleId, FusedValue> fusedValues = CHMHelper.flattenMap(fusedValuesMap);

		FusionPerformance<Integer> maxCorrect = maxCorrectFinder.getMaxCorrect(sgv, null, PropertiesConfig.defaultConfig());

		FusionPerformance<PerformancePR> performance = accuracyCalculator.analyzeAccuracy(fusedValues, maxCorrect);

		System.out.println();
		System.out.println();
		System.out.println();
		System.out.println();

		String name = prefix + fusionRunType.getName();
		System.out.println(name);

		System.out.println();
		performanceResultsWriter.writePerformance(ontologyRepository, dataset, name, performance, performance, true, true, false, false);

	}

	public void getPerformanceWithScores(String dataset, PerformanceResultsWriter performanceResultsWriter, GroupedValues groupedValues,
			FusionStrategy fusionRunType, ScoresContainer scoresContainer, String name, FusionConfig fusionConfig, List<FusionDataType> methodTypes)
			throws IOException {

		ThresholdContainer probabilityThreshold = scoreFusionThresholdFinder.findOptimumThresholdForPageRank();
		name += "_" + dfmt.format(probabilityThreshold) + "_THRESHOLD";
		System.out.println();
		System.out.println();
		// Match and group all the values

		List<FusionDataType> dataTypes = new ArrayList<>();
		dataTypes.addAll(methodTypes);

		List<FusionPreproccesser> preprocessors = new ArrayList<>();
		preprocessors.add(FusionPreproccesser.NONE);

		Map<FusionRun, Map<TripleId, FusedValue>> fusedValuesMap = fusionRunner.run(dataTypes, preprocessors, fusionRunType, groupedValues, scoresContainer,
				fusionConfig, probabilityThreshold, PropertiesConfig.defaultConfig());

		Map<TripleId, FusedValue> fusedValues = CHMHelper.flattenMap(fusedValuesMap);

		FusionPerformance<Integer> maxCorrect = maxCorrectFinder.getMaxCorrect(groupedValues, null, PropertiesConfig.defaultConfig());

		FusionPerformance<PerformancePR> performance = accuracyCalculator.analyzeAccuracy(fusedValues, maxCorrect);

		System.out.println();
		System.out.println();
		System.out.println();
		System.out.println();

		System.out.println(name);

		System.out.println();
		performanceResultsWriter.writePerformance(ontologyRepository, dataset, name, performance, performance, true, true, false, false);

	}

	@FunctionalInterface
	public static interface ScoresGetter {
		ScoresContainer getScores(String prefix, InstanceScores instanceScores, ColumnScores columnScores, TableScores tableScores, HostScores hostScores);
	}

	public Map<FusionStrategyId, FusionPerformance<PerformancePR>> getPerformanceReturnCrossValidation(FusionConfig fusionConfig,
			StrategiesConfig strategiesConfig) throws IOException {
		GroupedValues groupedValues = groupedValuesRepository.getGroupedValues();
		return analyzeAccuracy("", null, null, "", groupedValues, false, fusionConfig, strategiesConfig, PropertiesConfig.defaultConfig());

	}

}
