package de.mannheim.uni.expansion.tablefusion.evaluation.analyzer;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

import javax.inject.Inject;
import javax.inject.Named;

import com.opencsv.CSVWriter;

import de.mannheim.uni.expansion.data.ontology.model.ClassInstanceId;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId.FileEnding;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.NamedProperties;
import de.mannheim.uni.tableprocessor.model.Table;
import de.mannheim.uni.tableprocessor.model.TableColumn;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;

@Deprecated
public class FusionAnalyzer {
	private final String resultsFolder;

	@Inject
	public FusionAnalyzer(@Named(NamedProperties.RESULTS_FOLDER) String resultsFolder) {
		this.resultsFolder = resultsFolder;
	}

	public void printDensityIncrease(Collection<Table> tables, AnnotatedTablesRepository repository) {
		int total = 0;
		int filled = 0;
		int fusionFillable = 0;
		int filledAfter = 0;

		Set<InstanceId> allInstances = new HashSet<InstanceId>(repository.getInstances());

		for (Table t : tables) {

			TableColumn kc = t.getKey();

			if (kc == null) {
				System.out.println("DBPedia table doesn't have key column? Table id: " + t.getFullPath());
				continue;
			}

			Map<Integer, Object> kcValues = kc.getValues();

			int availableInstances = 0;

			for (Entry<Integer, Object> instance : kcValues.entrySet()) {

				// DBpediaInstance instanceId = new DBpediaInstance(
				// "http://dbpedia.org/resource/"
				// + ((String) instance.getValue()).replace(" ",
				// "_"));

				InstanceId instanceId = InstanceId.fromString(instance.getValue().toString());

				if (repository.getTables(instanceId) != null) {
					availableInstances++;

					allInstances.remove(instanceId);
					for (TableColumn c : t.getColumns()) {
						if (c.isKey())
							continue;

						if (c.getValues().containsKey(instance.getKey())) {
							filled++;
							filledAfter++;
						} else {
							PropertyId propertyId = new PropertyId(c.getURI().toString());

							// FIXME: WHAT IS THIS?
							// Map<AnnotatedTableId, String> values =
							// repository.getValues(instanceId, propertyId);
							//
							// if (values.size() > 0) {
							// filledAfter++;
							// fusionFillable++;
							// }
						}
					}
				}

			}

			total += availableInstances * t.getColumns().size();
		}

		System.out.println("Density before Fusion: " + ((double) filled / total));
		System.out.println("Desnity after Fusion:" + ((double) filledAfter / total));
		System.out.println("Desnity Fusion only:" + ((double) fusionFillable / total));
		System.out.println();
	}

	Runnable PRODCUER_POISEN = new Runnable() {

		@Override
		public void run() {
			// TODO Auto-generated method stub

		}
	};

	public void printDensityDistribution(KnowledgeBase dbpedia, FileEnding fileEnding, AnnotatedTablesRepository repository) {

		final ConcurrentHashMap<ClassPropertyId, DensityMeasures> propertyStores = new ConcurrentHashMap<ClassPropertyId, DensityMeasures>();

		final ConcurrentHashMap<ClassInstanceId, DensityMeasures> instanceStores = new ConcurrentHashMap<ClassInstanceId, DensityMeasures>();

		final AtomicBoolean ab = new AtomicBoolean(true);

		final CountDownLatch countDownLatch = new CountDownLatch(1);

		Thread densityWriteThread = new Thread(() -> {

			try {
				while (ab.get()) {
					Thread.sleep(3000);
					writeDensities(propertyStores);
				}
				countDownLatch.countDown();
			} catch (InterruptedException ex) {
				Thread.currentThread().interrupt();
			}

		});

		densityWriteThread.start();

		int numOfProducerThreads = 12;

		ArrayBlockingQueue<Runnable> producerQueue = new ArrayBlockingQueue<Runnable>(numOfProducerThreads * 3, true);

		CountDownLatch prodcuerLatch = new CountDownLatch(numOfProducerThreads);

		Runnable producerWorker = new Runnable() {

			@Override
			public void run() {
				System.out.println("Producer " + Thread.currentThread().getName() + " START");
				while (true) {

					Runnable task;
					try {
						task = producerQueue.take();

						if (task == PRODCUER_POISEN) {
							break;
						} else {
							task.run();
						}
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
				System.out.println("Producer " + Thread.currentThread().getName() + " END");
				prodcuerLatch.countDown();

			}
		};

		for (int i = 0; i < numOfProducerThreads; i++) {
			Thread thread = new Thread(producerWorker);
			thread.start();
		}

		int t = 0;

		for (ClassId classId : dbpedia.getClasses()) {

			System.out.println(t);

			for (PropertyId propertyId : dbpedia.getProperties(classId).keySet()) {

				final boolean tablesHaveProperty = (repository.hasProperty(propertyId) != null);
				final ClassPropertyId classPropertyId = new ClassPropertyId(classId, propertyId);
				Runnable r = () -> {
					DensityMeasures propertyStore = getStore(propertyStores, classPropertyId);

					for (InstanceId instanceId : dbpedia.getInstances(classId).keySet()) {

						DensityMeasures instanceStore = getStore(instanceStores, new ClassInstanceId(classId, instanceId));

						boolean fillAfter = false;
						propertyStore.incTotal();
						instanceStore.incTotal();

						if (dbpedia.hasValue(instanceId, propertyId)) {
							propertyStore.incFilled();
							instanceStore.incFilled();
							fillAfter = true;

						}

						if (tablesHaveProperty) {
							if (repository.getTables(instanceId) != null) {

								// FIXME: WHAT IS THIS?
								// Map<AnnotatedTableId, String> values =
								// repository.getValues(instanceId, propertyId);
								//
								// if (values.size() > 0) {
								// fillAfter = true;
								// propertyStore.incFusionFillable();
								// instanceStore.incFusionFillable();
								//
								// }

							}

						}

						if (fillAfter) {
							propertyStore.incFilledAfter();
							instanceStore.incFilledAfter();
						}
					}

				};

				try {
					producerQueue.put(r);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

			}

			t++;

		}

		for (int i = 0; i < numOfProducerThreads; i++) {
			try {
				producerQueue.put(PRODCUER_POISEN);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		try {
			prodcuerLatch.await();
		} catch (InterruptedException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		ab.set(false);

		try {
			countDownLatch.await();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		writeDensities(propertyStores);

		writeDensitiesInstance(instanceStores);

		// for (Table t : tables) {
		//
		// ClassId classId = new ClassId(t.getHeader());
		//
		// TableColumn kc = t.getKey();
		//
		// if (kc == null) {
		// System.out.println("DBPedia table doesn't have key column? Table id:
		// "
		// + t.getFullPath());
		// continue;
		// }
		//
		// Map<Integer, Object> kcValues = kc.getValues();
		//
		// for (Entry<Integer, Object> instance : kcValues.entrySet()) {
		//
		// InstanceId instanceId =
		// InstanceId.fromString(instance.getValue().toString());
		//
		// // new DBpediaInstance(
		// // "http://dbpedia.org/resource/"
		// // + ((String) instance.getValue()).replace(" ",
		// // "_"));
		//
		// if (repository.hasInstance(instanceId) != null) {
		// for (TableColumn c : t.getColumns()) {
		// if (c.isKey())
		// continue;
		//
		// PropertyId propertyId = new PropertyId(c.getURI().toString());
		//
		// ClassPropertyId classPropertyId = new ClassPropertyId(classId,
		// propertyId);
		//
		// if (repository.hasProperty(propertyId) != null) {
		//
		// getStore(densityStores, classPropertyId).incTotal();
		//
		// if (c.getValues().containsKey(instance.getKey())) {
		// getStore(densityStores, classPropertyId).incFilled();
		// getStore(densityStores, classPropertyId).incFilledAfter();
		// } else {
		//
		// Map<AnnotatedTableId, String> values =
		// repository.getValues(instanceId, propertyId);
		//
		// if (values.size() > 0) {
		// getStore(densityStores, classPropertyId).incFilledAfter();
		// getStore(densityStores, classPropertyId).incFusionFillable();
		// }
		// }
		// }
		// }
		//
		// }
		//
		// }
		//
		// }

	}

	<T> DensityMeasures getStore(ConcurrentHashMap<T, DensityMeasures> propertyStores, T classPropertyId) {
		DensityMeasures store = propertyStores.get(classPropertyId);
		if (store == null) {
			DensityMeasures newStore = new DensityMeasures();
			store = propertyStores.putIfAbsent(classPropertyId, newStore);
			if (store == null) {
				return newStore;
			} else {
				return store;
			}
		} else {
			return store;
		}
	}

	class DensityMeasures {
		private final AtomicInteger total = new AtomicInteger(0);
		private final AtomicInteger filled = new AtomicInteger(0);
		private final AtomicInteger filledAfter = new AtomicInteger(0);
		private final AtomicInteger fusionFillable = new AtomicInteger(0);

		public void incTotal() {
			this.total.incrementAndGet();
		}

		public void incFilled() {
			this.filled.incrementAndGet();
		}

		public void incFilledAfter() {
			this.filledAfter.incrementAndGet();
		}

		public void incFusionFillable() {
			this.fusionFillable.incrementAndGet();
		}

		public double densityBefore() {
			return filled.doubleValue() / total.doubleValue();
		}

		public double densityAfter() {
			return filledAfter.doubleValue() / total.doubleValue();
		}

		public double densityFusionFillabe() {
			return fusionFillable.doubleValue() / total.doubleValue();
		}

	}

	public void writeDensities(Map<ClassPropertyId, DensityMeasures> densityStores) {
		try {

			CSVWriter compositeWriter = new CSVWriter(new FileWriter(new File(resultsFolder + "/density.csv")));

			String[] headerLine = new String[7];
			headerLine[0] = "Class";
			headerLine[1] = "Property";
			headerLine[2] = "DensityBefore";
			headerLine[3] = "DensityAfter";
			headerLine[4] = "Diff";
			headerLine[5] = "FusionFillable";
			headerLine[6] = "OutOf";

			compositeWriter.writeNext(headerLine);

			List<Double> densitiesAfter = new ArrayList<Double>();
			List<Double> densitiesBefore = new ArrayList<Double>();
			for (Entry<ClassPropertyId, DensityMeasures> e : densityStores.entrySet()) {
				if (e.getValue().densityBefore() > 0.7 && e.getValue().densityBefore() < 0.9) {
					densitiesBefore.add(e.getValue().densityBefore());
					densitiesAfter.add(e.getValue().densityAfter());
					// System.out.println("Before: " +
					// e.getValue().densityBefore()
					// + ", After: " + e.getValue().densityAfter() +
					// ", Fusionfillable: "
					// + e.getValue().densityFusionFillabe());
				}

				String[] line = new String[7];
				line[0] = e.getKey().getClassId().toString();
				line[1] = e.getKey().getPropertyId().toString();
				line[2] = String.valueOf(e.getValue().densityBefore());
				line[3] = String.valueOf(e.getValue().densityAfter());
				line[4] = String.valueOf(e.getValue().densityAfter() - e.getValue().densityBefore());
				line[5] = String.valueOf(e.getValue().densityFusionFillabe());
				line[6] = String.valueOf(e.getValue().total);

				compositeWriter.writeNext(line);
			}

			compositeWriter.close();

			// Double densityBefore =
			// densitiesBefore.stream().mapToDouble(Double::doubleValue).average().getAsDouble();
			// Double densityAfter =
			// densitiesAfter.stream().mapToDouble(Double::doubleValue).average().getAsDouble();
			// Double fusionFillable = densityAfter - densityBefore;

			// System.out.println("Density before Fusion: " + densityBefore);
			// System.out.println("Density after Fusion: " + densityAfter);
			// System.out.println("Fusion Fillable: " + fusionFillable);

			// System.out.println("Desnity Fusion only:"
			// + ((double) fusionFillable / total));
			System.out.println();

		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public void writeDensitiesInstance(Map<ClassInstanceId, DensityMeasures> densityStores) {
		try {

			Map<ClassId, List<Double>> densityListsBefore = new HashMap<ClassId, List<Double>>();
			Map<ClassId, List<Double>> densityListsAfter = new HashMap<ClassId, List<Double>>();

			for (Entry<ClassInstanceId, DensityMeasures> e : densityStores.entrySet()) {
				ClassId classId = e.getKey().getClassId();
				List<Double> densitiesBefore = densityListsBefore.get(classId);
				if (densitiesBefore == null) {
					densitiesBefore = new ArrayList<Double>();
					densityListsBefore.put(classId, densitiesBefore);
				}
				densitiesBefore.add(e.getValue().densityBefore());

				List<Double> densitiesAfter = densityListsAfter.get(classId);
				if (densitiesAfter == null) {
					densitiesAfter = new ArrayList<Double>();
					densityListsAfter.put(classId, densitiesAfter);
				}
				densitiesAfter.add(e.getValue().densityAfter());

			}

			CSVWriter compositeWriter = new CSVWriter(new FileWriter(new File(resultsFolder + "/density_by_instance.csv")));

			String[] headerLine = new String[16];
			headerLine[0] = "Class";
			headerLine[1] = "% of I before with d > 0.60";
			headerLine[2] = "% of I after with d > 0.60";
			headerLine[3] = "$ diff for d > 0.60";
			headerLine[4] = "% of I before with d > 0.75";
			headerLine[5] = "% of I after with d > 0.75";
			headerLine[6] = "$ diff for d > 0.75";
			headerLine[7] = "% of I before with d > 0.80";
			headerLine[8] = "% of I after with d > 0.80";
			headerLine[9] = "$ diff for d > 0.80";
			headerLine[10] = "% of I before with d > 0.90";
			headerLine[11] = "% of I after with d > 0.90";
			headerLine[12] = "$ diff for d > 0.90";
			headerLine[13] = "% of I before with d > 0.95";
			headerLine[14] = "% of I after with d > 0.95";
			headerLine[15] = "$ diff for d > 0.95";

			compositeWriter.writeNext(headerLine);

			for (Entry<ClassId, List<Double>> e : densityListsBefore.entrySet()) {
				ClassId classId = e.getKey();
				List<Double> densitiesBefore = e.getValue();
				List<Double> densitiesAfter = densityListsAfter.get(classId);

				double total = e.getValue().size();
				int before60 = 0;
				int after60 = 0;
				int before75 = 0;
				int after75 = 0;
				int before80 = 0;
				int after80 = 0;
				int before90 = 0;
				int after90 = 0;
				int before95 = 0;
				int after95 = 0;

				for (Double d : densitiesBefore) {
					if (d > .95) {
						before95++;
						before90++;
						before80++;
						before75++;
						before60++;
					} else if (d > 0.9) {
						before90++;
						before80++;
						before75++;
						before60++;
					} else if (d > 0.8) {

						before80++;
						before75++;
						before60++;
					} else if (d > 0.75) {

						before75++;
						before60++;
					} else if (d > 0.60) {
						before60++;
					}
				}

				for (Double d : densitiesAfter) {
					if (d > .95) {
						after95++;
						after90++;
						after80++;
						after75++;
						after60++;
					} else if (d > 0.9) {
						after90++;
						after80++;
						after75++;
						after60++;
					} else if (d > 0.8) {
						after80++;
						after75++;
						after60++;
					} else if (d > 0.75) {
						after75++;
						after60++;
					} else if (d > 0.60) {
						after60++;
					}
				}

				String[] line = new String[16];
				line[0] = classId.toString();
				line[1] = String.valueOf(before60 / total);
				line[2] = String.valueOf(after60 / total);
				line[3] = String.valueOf((after60 - before60) / total);
				line[4] = String.valueOf(before75 / total);
				line[5] = String.valueOf(after75 / total);
				line[6] = String.valueOf((after75 - before75) / total);
				line[7] = String.valueOf(before80 / total);
				line[8] = String.valueOf(after80 / total);
				line[9] = String.valueOf((after80 - before80) / total);
				line[10] = String.valueOf(before90 / total);
				line[11] = String.valueOf(after90 / total);
				line[12] = String.valueOf((after90 - before90) / total);
				line[13] = String.valueOf(before95 / total);
				line[14] = String.valueOf(after95 / total);
				line[15] = String.valueOf((after95 - before95) / total);

				compositeWriter.writeNext(line);

			}

			compositeWriter.close();

		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
