package de.mannheim.uni.expansion.tablefusion.evaluation.analyzer;

import de.mannheim.uni.expansion.tablefusion.core.fusing.FusionScoreThresholdAdapter;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitterFactory;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesFilter;

public class FusionConfig {

	private Class<? extends SplitterFactory> splitterFactory = null;
	private SplitGroupedValuesFilter splitGroupedValuesFilter = null;
	private boolean onePerPLD = true;
	private FusionScoreThresholdAdapter fusionScoreThresholdAdapter = null;
	private boolean fuseLists = false;
	private boolean reSplitTestingSet = false;
	private boolean adjustedRecall = false;
	private int minSource = 0;

	private FusionConfig() {

	}

	public static FusionConfigBuilder builder() {
		return new FusionConfigBuilder();

	}

	public static FusionConfigBuilder builder(FusionConfig fusionConfig) {
		return new FusionConfigBuilder(fusionConfig);
	}

	public FusionConfigBuilder rebuilder() {
		return new FusionConfigBuilder(this);
	}

	public static FusionConfig defaultConfig() {
		return new FusionConfig();
	}

	public FusionConfig copy() {
		FusionConfig fusionConfig = new FusionConfig();
		fusionConfig.fuseLists = fuseLists;
		fusionConfig.splitterFactory = splitterFactory;
		fusionConfig.splitGroupedValuesFilter = splitGroupedValuesFilter;
		fusionConfig.onePerPLD = onePerPLD;
		fusionConfig.fusionScoreThresholdAdapter = fusionScoreThresholdAdapter;
		fusionConfig.reSplitTestingSet = reSplitTestingSet;
		fusionConfig.adjustedRecall = adjustedRecall;
		fusionConfig.minSource = minSource;

		return fusionConfig;

	}

	public Class<? extends SplitterFactory> getSplitterFactory() {
		return splitterFactory;
	}

	public SplitGroupedValuesFilter getSplitGroupedValuesFilter() {
		return splitGroupedValuesFilter;
	}

	public boolean isOnePerPLD() {
		return onePerPLD;
	}

	public FusionScoreThresholdAdapter getFusionScoreThresholdAdapter() {
		return fusionScoreThresholdAdapter;
	}

	public boolean isFuseLists() {
		return fuseLists;
	}

	public boolean isReSplitTestingSet() {
		return reSplitTestingSet;
	}

	public boolean isAdjustedRecall() {
		return adjustedRecall;
	}

	public int getMinSource() {
		return minSource;
	}

	public static class FusionConfigBuilder {

		private final FusionConfig fusionConfig;

		private FusionConfigBuilder() {
			fusionConfig = new FusionConfig();
		}

		private FusionConfigBuilder(FusionConfig fusionConfig) {
			this.fusionConfig = fusionConfig.copy();
		}

		public FusionConfig build() {
			return fusionConfig.copy();
		}

		public FusionConfigBuilder setSplitterFactory(Class<? extends SplitterFactory> splitterFactory) {
			fusionConfig.splitterFactory = splitterFactory;
			return this;
		}

		public FusionConfigBuilder setSplitGroupedValuesFilter(SplitGroupedValuesFilter splitGroupedValuesFilter) {
			fusionConfig.splitGroupedValuesFilter = splitGroupedValuesFilter;
			return this;
		}

		public FusionConfigBuilder setOnePerPLD(boolean onePerPLD) {
			fusionConfig.onePerPLD = onePerPLD;
			return this;
		}

		public FusionConfigBuilder setFusionScoreThresholdAdapter(FusionScoreThresholdAdapter fusionScoreThresholdAdapter) {
			fusionConfig.fusionScoreThresholdAdapter = fusionScoreThresholdAdapter;
			return this;
		}

		public FusionConfigBuilder setFuseLists(boolean fuseLists) {
			fusionConfig.fuseLists = fuseLists;
			return this;
		}

		public FusionConfigBuilder setReSplitTestingSet(boolean reSplitTestingSet) {
			fusionConfig.reSplitTestingSet = reSplitTestingSet;
			return this;
		}

		public FusionConfigBuilder setAdjustedRecall(boolean adjustedRecall) {
			fusionConfig.adjustedRecall = adjustedRecall;
			return this;
		}

		public FusionConfigBuilder setMinSource(int minSource) {
			fusionConfig.minSource = minSource;
			return this;
		}

	}

}
