package de.mannheim.uni.expansion.tablefusion.evaluation.analyzer;

import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig.TimeDependence;
import de.mannheim.uni.expansion.utils.PersistentParallel6;

public class GroupedValuesFiltererNew {

	private final OntologyRepository ontologyRepository;

	@Inject
	public GroupedValuesFiltererNew(OntologyRepository ontologyRepository) {
		this.ontologyRepository = ontologyRepository;
	}

	GroupedValues filter(GroupedValues groupedValues, PropertiesConfig propertiesConfig) {

		ConcurrentHashMap<TripleId, MatchedValues> map = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(groupedValues.entrySet(), (e) -> {
			TimeDependence timeDependence = propertiesConfig.getTimeDependence();

			switch (timeDependence) {
			case ALL:
				map.put(e.getKey(), e.getValue());
				break;
			case NO:
				if (!ontologyRepository.isTimeDependent(e.getKey().getPropertyId())) {
					map.put(e.getKey(), e.getValue());
				}
				break;
			case YES:
				if (ontologyRepository.isTimeDependent(e.getKey().getPropertyId())) {
					map.put(e.getKey(), e.getValue());
				}
				break;
			default:
				map.put(e.getKey(), e.getValue());
				break;

			}

		} , this.getClass());

		return new GroupedValues(map);

	}
}
