package de.mannheim.uni.expansion.tablefusion.evaluation.analyzer;

public class PropertiesConfig implements ConfigInterface<PropertiesConfig> {

	public static enum TimeDependence {
		YES, NO, ALL
	};

	private PropertiesConfig() {

	}

	private TimeDependence timeDependence = TimeDependence.ALL;

	public static PropertiesConfig defaultConfig() {
		return new PropertiesConfig();
	}

	@Override
	public PropertiesConfig copy() {
		PropertiesConfig propertiesConfig = new PropertiesConfig();
		propertiesConfig.timeDependence = timeDependence;
		return propertiesConfig;
	}

	@Override
	public PropertiesConfigBuilder rebuilder() {
		return new PropertiesConfigBuilder(this);
	}

	public static PropertiesConfigBuilder builder() {
		return new PropertiesConfigBuilder(defaultConfig());

	}

	public TimeDependence getTimeDependence() {
		return timeDependence;
	}

	public static class PropertiesConfigBuilder implements ConfigBuilderInterface<PropertiesConfig> {

		private PropertiesConfig propertiesConfig;

		private PropertiesConfigBuilder(PropertiesConfig propertiesConfig) {
			this.propertiesConfig = propertiesConfig.copy();
		}

		@Override
		public PropertiesConfig build() {
			return propertiesConfig.copy();
		}

		public PropertiesConfigBuilder setTimeDependence(TimeDependence timeDependence) {
			propertiesConfig.timeDependence = timeDependence;
			return this;
		}

	}

}
