package de.mannheim.uni.expansion.tablefusion.evaluation.analyzer;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionDataType;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionPreproccesser;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategy;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.TestingMethod;
import de.mannheim.uni.expansion.utils.CHMHelper;

public class StrategiesConfig {

	private List<FusionDataType> dataTypes = new ArrayList<>();
	private List<FusionPreproccesser> fusionPreprocessers = new ArrayList<>();
	private Map<TestingMethod, List<FusionStrategy>> fusionStrategies = new HashMap<>();

	private StrategiesConfig() {
		dataTypes.add(FusionDataType.REFERENCE);
		dataTypes.add(FusionDataType.STRING);
		dataTypes.add(FusionDataType.NUMERIC);
		dataTypes.add(FusionDataType.DATE);
		dataTypes.add(FusionDataType.CATEGORICAL_INT);
		fusionPreprocessers.add(FusionPreproccesser.NONE);
	}

	public StrategiesConfig copy() {
		StrategiesConfig config = new StrategiesConfig();

		config.dataTypes = new ArrayList<>(dataTypes);
		config.fusionStrategies = new HashMap<>();
		for (Entry<TestingMethod, List<FusionStrategy>> entry : fusionStrategies.entrySet()) {
			config.fusionStrategies.put(entry.getKey(), new ArrayList<>(entry.getValue()));
		}

		return config;

	}

	public List<FusionStrategy> getFusionStrategies(TestingMethod testingMethod) {
		return this.fusionStrategies.getOrDefault(testingMethod, new ArrayList<>());

	}

	public List<FusionDataType> getDataTypes() {
		return Collections.unmodifiableList(this.dataTypes);

	}

	public Map<TestingMethod, List<FusionStrategy>> getFusionStrategies() {
		return fusionStrategies;
	}

	public static StrategiesConfigBuilder builder() {
		return new StrategiesConfigBuilder();

	}

	public static StrategiesConfigBuilder builder(StrategiesConfig config) {
		return new StrategiesConfigBuilder(config);
	}

	public StrategiesConfigBuilder rebuilder() {
		return new StrategiesConfigBuilder(this);
	}

	public static StrategiesConfig defaultConfig() {
		return new StrategiesConfig();
	}

	public List<FusionPreproccesser> getFusionPreprocessers() {
		return Collections.unmodifiableList(fusionPreprocessers);
	}

	public static class StrategiesConfigBuilder {
		private final StrategiesConfig config;

		private StrategiesConfigBuilder() {
			this.config = new StrategiesConfig();
		}

		private StrategiesConfigBuilder(StrategiesConfig config) {
			this.config = config.copy();
		}

		public StrategiesConfig build() {
			return this.config.copy();
		}

		public StrategiesConfigBuilder replaceFusionStrategies(Map<TestingMethod, List<FusionStrategy>> fusionStrategies) {
			config.fusionStrategies = new HashMap<>();
			for (Entry<TestingMethod, List<FusionStrategy>> entry : fusionStrategies.entrySet()) {
				config.fusionStrategies.put(entry.getKey(), new ArrayList<>(entry.getValue()));
			}
			return this;
		}

		public StrategiesConfigBuilder replaceDataTypes(List<FusionDataType> fusionDataType) {
			config.dataTypes = new ArrayList<>(fusionDataType);
			return this;
		}

		public StrategiesConfigBuilder replaceFusionPreproccessers(List<FusionPreproccesser> fusionPreproccessers) {
			config.fusionPreprocessers = new ArrayList<>(fusionPreproccessers);
			return this;
		}

		public StrategiesConfigBuilder removeAllFusionPreproccesser() {
			config.fusionPreprocessers = new ArrayList<>();
			return this;
		}

		public StrategiesConfigBuilder addFusionPreproccesser(FusionPreproccesser fusionPreprocessers) {
			config.fusionPreprocessers.add(fusionPreprocessers);
			return this;
		}

		public StrategiesConfigBuilder removeAllDataTypes() {
			config.dataTypes = new ArrayList<>();
			return this;
		}

		public StrategiesConfigBuilder addDataType(FusionDataType fusionDataType) {
			config.dataTypes.add(fusionDataType);
			return this;
		}

		public StrategiesConfigBuilder removeAllStrategies() {
			config.fusionStrategies = new HashMap<>();
			return this;
		}

		public StrategiesConfigBuilder addStrategy(FusionStrategy fusionStrategy) {
			CHMHelper.getFromMap(config.fusionStrategies, fusionStrategy.geTestingMethod(), () -> new ArrayList<FusionStrategy>()).add(fusionStrategy);
			return this;
		}

	}
}
