package de.mannheim.uni.expansion.tablefusion.learning;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.TableIdsSet;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValuesCreator;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;

public class RowValuesGetter {

	@Inject
	protected OntologyRepository ontologyRepository;

	@Inject
	public RowValuesGetter() {

	}

	public RowValues getRowValues(AnnotatedTablesRepository repository, TableIdsSet tableIdsSet) {
		Map<TableRowId, Map<PropertyId, MatchedValues>> ret = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(tableIdsSet.getTableIds(), (tableId) -> {
			AnnotatedTableMetaData metaData = repository.getMetaData(tableId);
			AnnotatedTable table = repository.getTable(tableId);
			for (TableRowId rowId : repository.getRowIds(tableId)) {
				Map<PropertyId, MatchedValuesCreator> innerRetCreators = new HashMap<>();
				for (Entry<Integer, List<PropertyId>> propertyIndex : metaData.getAllProperties().entrySet()) {

					for (PropertyId propertyId : propertyIndex.getValue()) {
						CellId cellId = new CellId(tableId, new ColumnIdx(propertyIndex.getKey()), rowId.getRowIdx());
						AbstractObject ao = table.getValue(cellId.getRowIdx(), cellId.getColumnIdx());
						if (ao != null) {
							CHMHelper.getFromMap(innerRetCreators, propertyId, () -> new MatchedValuesCreator(ontologyRepository.getDataType(propertyId)))
									.add(new SingleValue(cellId, ao));
						}

					}
				}

				ret.put(rowId, innerRetCreators.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> new MatchedValues(e.getValue()))));
				;
			}

		}, this.getClass());

		return new RowValues(ret);
	}

	public SplitRowValues getSplitRowValues(AnnotatedTablesRepository repository, TableIdsSet tableIdsSet) {
		return getSplitRowValues(getRowValues(repository, tableIdsSet));

	}

	public SplitRowValues getSplitRowValues(RowValues rowValues) {

		Map<TableRowId, Map<QualifiedPropertyId, MatchedValues>> ret = new HashMap<>();
		for (Entry<TableRowId, Map<PropertyId, MatchedValues>> entry : rowValues.getValues().entrySet()) {
			Map<QualifiedPropertyId, MatchedValues> innerRet = new HashMap<>();
			for (Entry<PropertyId, MatchedValues> innerEntry : entry.getValue().entrySet()) {
				innerRet.put(new QualifiedPropertyId(innerEntry.getKey(), new HashSet<>()), innerEntry.getValue());
			}
			ret.put(entry.getKey(), innerRet);

		}
		return new SplitRowValues(ret);

	}
}
