package de.mannheim.uni.expansion.tablefusion.learning;

import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;

import de.mannheim.uni.expansion.data.ontology.model.AbstractQualifier;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.dws.expansion.data.ontology.model.qualifiers.TimestampPointInTime;

public class SplitGroupedValuesFilter {

	private final Function<TimestampPointInTime, Boolean> filter;

	public SplitGroupedValuesFilter(Function<TimestampPointInTime, Boolean> filter) {
		this.filter = filter;
	}

	public SplitGroupedValues filter(SplitGroupedValues splitGroupedValues) {

		ConcurrentHashMap<QualifiedTripleId, MatchedValues> groupedValues = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(splitGroupedValues.entrySet(), (entry) -> {

			Set<AbstractQualifier> newQualifiers = new HashSet<>();
			Set<AbstractQualifier> qualifiers = entry.getKey().getQualifiers();
			for (AbstractQualifier qualifier : qualifiers) {
				if (qualifier instanceof TimestampPointInTime) {
					TimestampPointInTime tpit = (TimestampPointInTime) qualifier;
					if (filter.apply(tpit)) {
						newQualifiers.add(tpit);
					}

				}
			}

			if (!newQualifiers.isEmpty()) {
				groupedValues.put(new QualifiedTripleId(entry.getKey().tripleId(), newQualifiers), entry.getValue());
			}

		} , this.getClass());

		return new SplitGroupedValues(groupedValues);
	}

}
