package de.mannheim.uni.expansion.tablefusion.learning;

import java.util.Map;
import java.util.UUID;

import com.fasterxml.uuid.Generators;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedPropertyId;
import de.mannheim.uni.expansion.data.tables.model.TableRowId;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.utils.ImmuMaker;

public class SplitRowValues {

	private final UUID uuid = Generators.timeBasedGenerator().generate();
	private final Map<TableRowId, Map<QualifiedPropertyId, MatchedValues>> values;

	public SplitRowValues(Map<TableRowId, Map<QualifiedPropertyId, MatchedValues>> values) {
		super();
		this.values = ImmuMaker.mapWithMap(values);
	}

	public Map<TableRowId, Map<QualifiedPropertyId, MatchedValues>> getValues() {
		return values;
	}

	public UUID getUUID() {
		return uuid;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((uuid == null) ? 0 : uuid.hashCode());
		result = prime * result + ((values == null) ? 0 : values.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (!(obj instanceof SplitRowValues))
			return false;
		SplitRowValues other = (SplitRowValues) obj;
		if (uuid == null) {
			if (other.uuid != null)
				return false;
		} else if (!uuid.equals(other.uuid))
			return false;
		if (values == null) {
			if (other.values != null)
				return false;
		} else if (!values.equals(other.values))
			return false;
		return true;
	}

}
