package de.mannheim.uni.expansion.tablefusion.optimization;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.BiFunction;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.CellId;
import de.mannheim.uni.expansion.tablefusion.analysis.AccuracyCalculator;
import de.mannheim.uni.expansion.tablefusion.analysis.MaxCorrectFinder;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.matching.SynchronizedMatchedValuesMapMerger;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.splitting.FilteredSplitter;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitGroupedValues;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionRun;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.StrategiesConfig;
import de.mannheim.uni.expansion.tablefusion.learning.LearningSetCreatorStratified;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionDataType;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionPreproccesser;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionRunner;
import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategy;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.ScorerRepository;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.SplitterProvider;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdClassPropertyId;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdConstant;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdContainer;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdPropertyId;

public class ScoreFusionThresholdFinderFullQualifiedSimple {

	private final FusionRunner fusionRunner;

	private final AccuracyCalculator accuracyCalculator;

	private final LearningSetCreatorStratified learningSetCreatorStratified;

	private final ValueEvaluator valueEvaluator;

	private final MaxCorrectFinder maxCorrectFinder;
	private final OntologyRepository ontologyRepository;
	private final ScorerRepository scorerRepository;

	@Inject
	public ScoreFusionThresholdFinderFullQualifiedSimple(FusionRunner fusionRunner, AccuracyCalculator accuracyCalculator, MaxCorrectFinder maxCorrectFinder,
			LearningSetCreatorStratified learningSetCreatorStratified, ValueEvaluator valueEvaluator, OntologyRepository ontologyRepository,
			ScorerRepository scorerRepository) {
		super();
		this.fusionRunner = fusionRunner;
		this.accuracyCalculator = accuracyCalculator;
		this.maxCorrectFinder = maxCorrectFinder;
		this.learningSetCreatorStratified = learningSetCreatorStratified;
		this.valueEvaluator = valueEvaluator;
		this.ontologyRepository = ontologyRepository;
		this.scorerRepository = scorerRepository;
	}

	public ThresholdContainer findOptimumThresholdCrossValidation(Map<Integer, SplitGroupedValuesSets> splitGroupedValuesSets, FusionStrategy strategy,
			StrategiesConfig strategiesConfig, FusionConfig fusionConfig, PropertiesConfig propertiesConfig) {
		if (strategy.getFusionMethod().isUsesThreshold()) {
			return findOptimumThresholdCrossValidation(splitGroupedValuesSets, scorerRepository.get(strategy.getScoringMethod()), strategy.getThresholdType(),
					strategiesConfig.getDataTypes(), strategiesConfig.getFusionPreprocessers(), strategy, fusionConfig, propertiesConfig);

		} else {
			return new ThresholdConstant(-0.1);
		}
	}

	public ThresholdContainer findOptimumThresholdCrossValidation(Map<Integer, SplitGroupedValuesSets> splitGroupedValuesSets, Scorer scorer,
			ThresholdType thresholdType, List<FusionDataType> dataTypes, List<FusionPreproccesser> preprocessors, FusionStrategy fusionRunType,
			FusionConfig fusionConfig, PropertiesConfig propertiesConfig) { // return
		// new
		// ThresholdConstant(0.9);

		switch (thresholdType) {
		case ClassPropertyId:
			return getThresholdClassProperty(splitGroupedValuesSets, scorer, dataTypes, preprocessors, fusionRunType, fusionConfig, propertiesConfig);

		case General:
			return getThreshold(splitGroupedValuesSets, scorer, dataTypes, preprocessors, fusionRunType, fusionConfig, propertiesConfig);

		case PropertyId:
			return getThresholdPropertyId(splitGroupedValuesSets, scorer, dataTypes, preprocessors, fusionRunType, fusionConfig, propertiesConfig);

		}
		return null;
		//
	}

	public ThresholdContainer findOptimumThresholdCrossValidation(SplitGroupedValues groupedValues, Scorer scorer, ThresholdType thresholdType,
			List<FusionDataType> dataTypes, List<FusionPreproccesser> preprocessors, FusionStrategy fusionRunType, FusionConfig fusionConfig, boolean fuseLists,
			PropertiesConfig propertiesConfig) { // return
		// new
		// ThresholdConstant(0.9);

		switch (thresholdType) {
		case ClassPropertyId:
			return getThresholdClassProperty(groupedValues, scorer, dataTypes, preprocessors, fusionRunType, fusionConfig, propertiesConfig);

		case General:
			return getThreshold(groupedValues, scorer, dataTypes, preprocessors, fusionRunType, fusionConfig, propertiesConfig);

		case PropertyId:
			return getThresholdPropertyId(groupedValues, scorer, dataTypes, preprocessors, fusionRunType, fusionConfig, propertiesConfig);

		}
		return null;
		//
	}

	// <T> Map<T, Map<Integer, SplitGroupedValues>>
	// getGroupedValues(SplitGroupedValues groupedValues, BiFunction<TripleId,
	// ValueType, T> getKey, int length) {
	// SortedMap<QualifiedTripleId, MatchedValues> sortedGroupedValues = new
	// TreeMap<>(groupedValues.getValues());
	//
	// Map<Integer, Map<QualifiedTripleId, MatchedValues>> overfitting = new
	// HashMap<>();
	//
	// int sizeOfSet = (int) Math.ceil((1.0 / length) * groupedValues.size());
	//
	// if (length == 1)
	// sizeOfSet = groupedValues.size();
	//
	// int i = 0;
	// for (Entry<QualifiedTripleId, MatchedValues> e :
	// sortedGroupedValues.entrySet()) {
	//
	// for (int j = 0; j < length; j++) {
	// int highLim = (length - j) * sizeOfSet;
	// int lowLim = (length - j - 1) * sizeOfSet;
	// if (length == 1 || (i < lowLim || i > highLim)) {
	// CHMHelper.getFromMap(overfitting, j, () -> new
	// HashMap<>()).put(e.getKey(), e.getValue());
	// }
	// }
	//
	// i++;
	// }
	//
	// ConcurrentHashMap<T, Map<Integer, SplitGroupedValues>> retMap = new
	// ConcurrentHashMap<>();
	//
	// for (Entry<Integer, Map<QualifiedTripleId, MatchedValues>> ov :
	// overfitting.entrySet()) {
	// ConcurrentHashMap<T, Map<QualifiedTripleId, MatchedValues>> listMap = new
	// ConcurrentHashMap<>();
	//
	// PersistentParallel4.runAndFinish(ov.getValue().entrySet(), (gv) -> {
	//
	// T key = getKey.apply(gv.getKey().getTripleId(),
	// gv.getValue().getDataType());
	// ConcurrentHashMap<QualifiedTripleId, MatchedValues> map =
	// (ConcurrentHashMap<QualifiedTripleId, MatchedValues>)
	// CHMHelper.getFromCHM(listMap,
	// key, () -> new ConcurrentHashMap<>());
	//
	// map.put(gv.getKey(), gv.getValue());
	//
	// } , "ThresholdSplitter");
	//
	// PersistentParallel4.runAndFinish(listMap.entrySet(), (e) -> {
	// CHMHelper.getFromCHM(retMap, e.getKey(), () -> new
	// ConcurrentHashMap<>()).put(ov.getKey(), new
	// SplitGroupedValues(e.getValue()));
	// } , "ThresholdSplitter");
	// }
	//
	// return retMap;
	// }

	<T> Map<T, SplitGroupedValues> getGroupedValues(SplitGroupedValues groupedValues, BiFunction<TripleId, ValueType, T> getKey) {

		ConcurrentHashMap<T, ConcurrentHashMap<QualifiedTripleId, MatchedValues>> listMap = new ConcurrentHashMap<>();

		PersistentParallel4.runAndFinish(groupedValues.entrySet(), (gv) -> {

			T key = getKey.apply(gv.getKey().tripleId(), gv.getValue().getDataType());
			ConcurrentHashMap<QualifiedTripleId, MatchedValues> map = CHMHelper.getFromCHM(listMap, key, () -> new ConcurrentHashMap<>());

			map.put(gv.getKey(), gv.getValue());

		}, "ThresholdSplitter");

		ConcurrentHashMap<T, SplitGroupedValues> retMap = new ConcurrentHashMap<>();

		PersistentParallel4.runAndFinish(listMap.entrySet(), (e) -> {
			retMap.put(e.getKey(), new SplitGroupedValues(e.getValue()));
		}, "ThresholdSplitter");

		return retMap;
	}

	private <T> Map<T, Double> getThresholdGeneric(SplitGroupedValues groupedValues, Scorer scorer, BiFunction<TripleId, ValueType, T> getKey,
			Double alternative, List<FusionDataType> dataTypes, List<FusionPreproccesser> preprocessors, FusionStrategy fusionRunType,
			FusionConfig fusionConfig, PropertiesConfig propertiesConfig) {

		// Map<T, SplitGroupedValues> groupedValuesMap =
		// getGroupedValues(groupedValues, getKey);

		// Map<T, Map<Integer, SplitGroupedValuesSets>> splittedSetsMap = new
		// HashMap<>();
		// for (Entry<T, SplitGroupedValues> gv : groupedValuesMap.entrySet()) {

		Map<Integer, SplitGroupedValuesSets> splited = learningSetCreatorStratified.createByTableProperty(4, groupedValues);
		// splittedSetsMap.put(gv.getKey(), splited);
		// }

		return getThresholdGeneric(splited, scorer, getKey, alternative, dataTypes, preprocessors, fusionRunType, fusionConfig, propertiesConfig);

	}

	private <T> Map<T, Double> getThresholdGenericSets(Map<Integer, SplitGroupedValuesSets> splitGroupedValuesSets, Scorer scorer,
			BiFunction<TripleId, ValueType, T> getKey, Double alternative, List<FusionDataType> dataTypes, List<FusionPreproccesser> preprocessors,
			FusionStrategy fusionRunType, FusionConfig fusionConfig, PropertiesConfig propertiesConfig) {

		// Map<T, Map<Integer, SplitGroupedValuesSets>> splittedSetsMap = new
		// HashMap<>();
		// for (Entry<Integer, SplitGroupedValuesSets> gv :
		// splitGroupedValuesSets.entrySet()) {
		//
		// Map<T, SplitGroupedValues> fullSplitted =
		// getGroupedValues(gv.getValue().getFullSet(), getKey);
		// Map<T, SplitGroupedValues> learningSplitted =
		// getGroupedValues(gv.getValue().getLearningSet(), getKey);
		// Map<T, SplitGroupedValues> testingSplitted =
		// getGroupedValues(gv.getValue().getTestingSet(), getKey);
		//
		// Set<T> keySet = new HashSet<>();
		// keySet.addAll(fullSplitted.keySet());
		// keySet.addAll(learningSplitted.keySet());
		// keySet.addAll(testingSplitted.keySet());
		//
		// for (T key : keySet) {
		// CHMHelper.getFromMap(splittedSetsMap, key, () -> new
		// HashMap<>()).put(gv.getKey(),
		// new SplitGroupedValuesSets(fullSplitted.getOrDefault(key, new
		// SplitGroupedValues(new HashMap<>())),
		// learningSplitted.getOrDefault(key, new SplitGroupedValues(new
		// HashMap<>())),
		// testingSplitted.getOrDefault(key, new SplitGroupedValues(new
		// HashMap<>()))));
		// }
		//
		// }

		return getThresholdGeneric(splitGroupedValuesSets, scorer, getKey, alternative, dataTypes, preprocessors, fusionRunType, fusionConfig,
				propertiesConfig);

	}

	private <T> Map<T, Double> getThresholdGeneric(Map<Integer, SplitGroupedValuesSets> splittetSetsMap, Scorer scorer,
			BiFunction<TripleId, ValueType, T> getKey, Double alternative, List<FusionDataType> dataTypes, List<FusionPreproccesser> preprocessors,
			FusionStrategy fusionRunType, FusionConfig fusionConfig, PropertiesConfig propertiesConfig) {
		boolean group = false;
		int length = 21;
		double step = 0.05;
		double subtract = 0.001;

		// Map<T, Map<Integer, GroupedValues>> originalGroupedValuesMap =
		// // fuse(groupedValuesMap);
		// ConcurrentHashMap<T, ConcurrentLinkedQueue<Double>> correctMap = new
		// ConcurrentHashMap<>();
		//
		// ConcurrentHashMap<T, ConcurrentLinkedQueue<Double>> incorrectMap =
		// new ConcurrentHashMap<>();

		Map<T, Double> thresholdMap = new HashMap<>();

		Map<T, Map<Integer, Double>> thresholds = new HashMap<>();
		for (Entry<Integer, SplitGroupedValuesSets> subMap : splittetSetsMap.entrySet()) {

			// Map<Integer, SplitGroupedValuesSets> splited = gv.getValue();
			//
			Map<T, SplitGroupedValues> splittedTestingValues = getGroupedValues(subMap.getValue().getTestingSet(), getKey);
			Map<T, Integer> maxCorrectMap = new HashMap<>();
			for (Entry<T, SplitGroupedValues> splittedTestedValues : splittedTestingValues.entrySet()) {
				maxCorrectMap.put(splittedTestedValues.getKey(),
						maxCorrectFinder.getMaxCorrectValues(splittedTestedValues.getValue(), null, propertiesConfig, fusionConfig));
			}
			// int maxCorrect = ;

			ScoresContainer probabilities = scorer.getScores(subMap.getValue());

			SplitGroupedValues sgvTestingForFusion = subMap.getValue().getTestingSet();
			if (fusionConfig.isReSplitTestingSet() && scorer instanceof SplitterProvider) {
				System.out.println("RESPLITING for Thresholding");
				FilteredSplitter filteredSplitter = ((SplitterProvider) scorer).getSplitter(subMap.getValue());
				sgvTestingForFusion = filteredSplitter.split(subMap.getValue().getTestingSet());
			}

			Map<T, int[]> correctMap = new HashMap<>();
			Map<T, int[]> wrongMap = new HashMap<>();

			for (int i = 0; i < length; i++) {

				Double probabilityThreshold = step * i - subtract;
				Map<FusionRun, Map<QualifiedTripleId, FusedValue>> fusedValuesMap = fusionRunner.run(dataTypes, preprocessors, fusionRunType,
						sgvTestingForFusion, probabilities, fusionConfig, new ThresholdConstant(probabilityThreshold), propertiesConfig);

				Map<QualifiedTripleId, FusedValue> fusedValues = CHMHelper.flattenMap(fusedValuesMap);

				Map<T, Map<QualifiedTripleId, FusedValue>> splitFusedValues = spiltFusedValues(fusedValues, getKey);

				for (Entry<T, Map<QualifiedTripleId, FusedValue>> sfvEntry : splitFusedValues.entrySet()) {

					for (Entry<QualifiedTripleId, FusedValue> fusedEntry : sfvEntry.getValue().entrySet()) {

						Optional<Boolean> correctValue = valueEvaluator.correct(fusedEntry.getKey(), fusedEntry.getValue());
						if (correctValue.isPresent()) {

							// FusedValue fv = fusedEntry.getValue();
							//
							// double sum = fv.getSingleScore();
							// double adapted = sum / fv.getSingleContributors()
							// * sum * 2;
							//
							// if (adapted > probabilityThreshold * sum) {

							if (correctValue.get()) {
								CHMHelper.getFromMap(correctMap, sfvEntry.getKey(), () -> new int[length])[i]++;
							} else {
								CHMHelper.getFromMap(wrongMap, sfvEntry.getKey(), () -> new int[length])[i]++;
							}
							// }

						}

					}
				}
			}

			Set<T> keys = new HashSet<>(correctMap.keySet());
			keys.addAll(wrongMap.keySet());

			for (T key : keys) {
				Integer maxCorrect = maxCorrectMap.get(key);
				if (maxCorrect == null)
					maxCorrect = 0;
				double precision[] = new double[length];
				double recall[] = new double[length];

				// double aRecall[] = new double[length];
				double fMeasure[] = new double[length];
				double aFMeasure[] = new double[length];

				int[] correct = CHMHelper.getFromMap(correctMap, key, () -> new int[length]);
				int[] wrong = CHMHelper.getFromMap(wrongMap, key, () -> new int[length]);

				for (int i = 0; i < length; i++) {

					precision[i] = ((double) correct[i]) / (correct[i] + wrong[i]);
					recall[i] = ((double) correct[i]) / (maxCorrect);
					// aRecall[i] = (precision[i] + recall[i]) / 2;
					// fMeasure[i] = 1.1 * ((0.2 * precision[i] * recall[i])
					// /
					// (0.1 * precision[i] + recall[i]));
					fMeasure[i] = 2 * ((precision[i] * recall[i]) / (precision[i] + recall[i]));
					aFMeasure[i] = (1 + 0.0625) * ((precision[i] * recall[i]) / (0.0625 * precision[i] + recall[i]));

				}

				double maxFMeasure = 0.0;
				double maxThreshold = 0.0;
				int maxThresholdIndex = 0;
				int maxStep = 0;

				for (int i = 0; i < length; i++) {
					// double precision = sumPrecision / count;
					// double recall = sumRecall / count;
					double diff = maxFMeasure - fMeasure[i];

					if (fusionConfig.isAdjustedRecall()) {
						if (aFMeasure[i] > maxFMeasure) {
							maxFMeasure = aFMeasure[i];
							maxThreshold = step * i - subtract;
							maxThresholdIndex = i;
						}
					} else {
						if (fMeasure[i] > maxFMeasure) {
							maxFMeasure = fMeasure[i];
							maxThreshold = step * i - subtract;
							maxThresholdIndex = i;
						}
					}

				}

				CHMHelper.getFromMap(thresholds, key, () -> new HashMap<>()).put(subMap.getKey(), maxThreshold);
			}
			// for (int i = 0; i < length; i++) {
			//
			// Double probabilityThreshold = step * i;
			//
			// lines[i] = String.valueOf(probabilityThreshold);
			//
			// sumPrecision[i] = 0.0;
			// sumRecall[i] = 0.0;
			// sumFMeasure[i] = 0.0;
			// count[i] = 0;
			//
			// }

		}

		for (

		Entry<T, Map<Integer, Double>> thresholdInnerMap : thresholds.entrySet())

		{
			double threshold = 0.0;

			for (Entry<Integer, Double> e : thresholdInnerMap.getValue().entrySet()) {
				threshold += e.getValue();
			}

			threshold = threshold / thresholdInnerMap.getValue().size();
			thresholdMap.put(thresholdInnerMap.getKey(), threshold);
		}
		// Set<T> cpidKeys = new HashSet<>();
		// cpidKeys.addAll(correctMap.keySet());
		// cpidKeys.addAll(incorrectMap.keySet());
		//
		//
		//
		// PersistentParallel3.runAndFinish(cpidKeys, (cpid) -> {
		//
		// ConcurrentLinkedQueue<Double> correct = correctMap.get(cpid);
		// ConcurrentLinkedQueue<Double> incorrect = incorrectMap.get(cpid);
		//
		// thresholdMap.put(cpid, getThreshold(correct, incorrect, length, step,
		// alternative));
		//
		// } , "ClassPropertyThresholds");

		return thresholdMap;

	}

	private <T> Map<T, Map<QualifiedTripleId, FusedValue>> spiltFusedValues(Map<QualifiedTripleId, FusedValue> fusedValues,
			BiFunction<TripleId, ValueType, T> getKey) {
		ConcurrentHashMap<T, Map<QualifiedTripleId, FusedValue>> listMap = new ConcurrentHashMap<>();

		PersistentParallel4.runAndFinish(fusedValues.entrySet(), (gv) -> {

			T key = getKey.apply(gv.getKey().tripleId(), ontologyRepository.getDataType(gv.getKey().getTripleId().getPropertyId()));
			ConcurrentHashMap<QualifiedTripleId, FusedValue> map = (ConcurrentHashMap<QualifiedTripleId, FusedValue>) CHMHelper.getFromCHM(listMap, key,
					() -> new ConcurrentHashMap<>());

			map.put(gv.getKey(), gv.getValue());

		}, "ThresholdSplitter");

		// ConcurrentHashMap<T, SplitGroupedValues> retMap = new
		// ConcurrentHashMap<>();
		//
		// PersistentParallel4.runAndFinish(listMap.entrySet(), (e) -> {
		// retMap.put(e.getKey(), new SplitGroupedValues(e.getValue()));
		// } , "ThresholdSplitter");

		return listMap;
	}

	private <T> Map<T, Map<Integer, GroupedValues>> fuse(Map<T, Map<Integer, SplitGroupedValues>> groupedValuesMap) {
		Map<T, Map<Integer, GroupedValues>> map = new HashMap<>();

		for (Entry<T, Map<Integer, SplitGroupedValues>> t : groupedValuesMap.entrySet()) {
			Map<Integer, GroupedValues> subMap = new HashMap<>();
			for (Entry<Integer, SplitGroupedValues> i : t.getValue().entrySet()) {

				ConcurrentHashMap<TripleId, ConcurrentHashMap<CellId, Pair<SingleValue, ValueType>>> cellIdsMaps = new ConcurrentHashMap<>();

				SynchronizedMatchedValuesMapMerger<TripleId> mapMerger = new SynchronizedMatchedValuesMapMerger<>();

				PersistentParallel4.runAndFinish(i.getValue().entrySet(), (e) -> {

					ConcurrentHashMap<CellId, Pair<SingleValue, ValueType>> cellIdMap = CHMHelper.getFromCHM(cellIdsMaps, e.getKey().getTripleId(),
							() -> new ConcurrentHashMap<>());

					for (SingleValue sv : e.getValue().getValues()) {
						cellIdMap.put(sv.getId(), new Pair<>(sv, e.getValue().getDataType()));
					}

				}, "FUSE SPLIT");

				PersistentParallel4.runAndFinish(cellIdsMaps.entrySet(), (e) -> {

					for (Entry<CellId, Pair<SingleValue, ValueType>> sv : e.getValue().entrySet()) {
						mapMerger.merge(e.getKey(), sv.getValue().getFirst(), sv.getValue().getSecond());
					}

				}, "FUSE SPLIT");

				subMap.put(i.getKey(), new GroupedValues(mapMerger.getCleanMap()));

			}
			map.put(t.getKey(), subMap);

		}

		return map;
	}

	private final BiFunction<TripleId, ValueType, PropertyId> propertyKeyGetter = (tid, cdt) -> {
		return tid.getPropertyId();
	};

	private ThresholdContainer getThresholdPropertyId(SplitGroupedValues groupedValues, Scorer scorer, List<FusionDataType> dataTypes,
			List<FusionPreproccesser> preprocessors, FusionStrategy fusionRunType, FusionConfig fusionConfig, PropertiesConfig propertiesConfig) {
		return new ThresholdPropertyId(getThresholdGeneric(groupedValues, scorer, propertyKeyGetter,
				((ThresholdConstant) getThreshold(groupedValues, scorer, dataTypes, preprocessors, fusionRunType, fusionConfig, propertiesConfig)).getScore(),
				dataTypes, preprocessors, fusionRunType, fusionConfig, propertiesConfig));

		//
	}

	private ThresholdContainer getThresholdPropertyId(Map<Integer, SplitGroupedValuesSets> groupedValues, Scorer scorer, List<FusionDataType> dataTypes,
			List<FusionPreproccesser> preprocessors, FusionStrategy fusionRunType, FusionConfig fusionConfig, PropertiesConfig propertiesConfig) {
		return new ThresholdPropertyId(getThresholdGenericSets(groupedValues, scorer, propertyKeyGetter,
				((ThresholdConstant) getThreshold(groupedValues, scorer, dataTypes, preprocessors, fusionRunType, fusionConfig, propertiesConfig)).getScore(),
				dataTypes, preprocessors, fusionRunType, fusionConfig, propertiesConfig));

		//
	}

	// private final PersistentParallel3<Entry<TripleId, MatchedValues>>
	// parallel = new PersistentParallel3<>("THREHSOLD FINDER");

	private final BiFunction<TripleId, ValueType, Boolean> normalKeyGetter = (tid, cdt) -> {
		return Boolean.TRUE;
	};

	private ThresholdContainer getThreshold(SplitGroupedValues groupedValues, Scorer scorer, List<FusionDataType> dataTypes,
			List<FusionPreproccesser> preprocessors, FusionStrategy fusionRunType, FusionConfig fusionConfig, PropertiesConfig propertiesConfig) {

		return new ThresholdConstant(
				getThresholdGeneric(groupedValues, scorer, normalKeyGetter, null, dataTypes, preprocessors, fusionRunType, fusionConfig, propertiesConfig)
						.getOrDefault(Boolean.TRUE, 0.0));
	}

	private ThresholdContainer getThreshold(Map<Integer, SplitGroupedValuesSets> groupedValues, Scorer scorer, List<FusionDataType> dataTypes,
			List<FusionPreproccesser> preprocessors, FusionStrategy fusionRunType, FusionConfig fusionConfig, PropertiesConfig propertiesConfig) {
		try {
			return new ThresholdConstant(getThresholdGenericSets(groupedValues, scorer, normalKeyGetter, null, dataTypes, preprocessors, fusionRunType,
					fusionConfig, propertiesConfig).getOrDefault(Boolean.TRUE, 0.0));
		} catch (Exception e) {
			e.printStackTrace();
			int sdf = 123;
			return null;
		}
	}

	private final BiFunction<TripleId, ValueType, ClassPropertyId> classPropertyKeyGetter = (tid, cdt) -> {
		return new ClassPropertyId(tid);
	};

	private ThresholdClassPropertyId getThresholdClassProperty(SplitGroupedValues groupedValues, Scorer scorer, List<FusionDataType> dataTypes,
			List<FusionPreproccesser> preprocessors, FusionStrategy fusionRunType, FusionConfig fusionConfig, PropertiesConfig propertiesConfig) {

		return new ThresholdClassPropertyId(getThresholdGeneric(groupedValues, scorer, classPropertyKeyGetter,
				((ThresholdConstant) getThreshold(groupedValues, scorer, dataTypes, preprocessors, fusionRunType, fusionConfig, propertiesConfig)).getScore(),
				dataTypes, preprocessors, fusionRunType, fusionConfig, propertiesConfig));

	}

	private ThresholdClassPropertyId getThresholdClassProperty(Map<Integer, SplitGroupedValuesSets> groupedValues, Scorer scorer,
			List<FusionDataType> dataTypes, List<FusionPreproccesser> preprocessors, FusionStrategy fusionRunType, FusionConfig fusionConfig,
			PropertiesConfig propertiesConfig) {

		return new ThresholdClassPropertyId(getThresholdGenericSets(groupedValues, scorer, classPropertyKeyGetter,
				((ThresholdConstant) getThreshold(groupedValues, scorer, dataTypes, preprocessors, fusionRunType, fusionConfig, propertiesConfig)).getScore(),
				dataTypes, preprocessors, fusionRunType, fusionConfig, propertiesConfig));

	}

	private double getThreshold(Collection<Double> correct, Collection<Double> incorrect, int length, double step, Double alternative) {

		if (!correct.isEmpty() || !incorrect.isEmpty()) {
			double precision[] = new double[length];
			double recall[] = new double[length];
			double fMeasure[] = new double[length];

			for (int i = 0; i < length; i++) {
				Double probabilityThreshold = step * i;

				int tp = 0;
				int fp = 0;
				int tn = 0;
				int fn = 9;

				for (Double d : correct) {
					if (d >= probabilityThreshold) {
						tp++;
					} else {
						fn++;
					}
				}

				for (Double d : incorrect) {
					if (d < probabilityThreshold) {
						tn++;
					} else {
						fp++;
					}
				}

				precision[i] = ((double) tp) / (tp + fp);
				recall[i] = ((double) tp) / (tp + fn);

				fMeasure[i] = 1.1 * ((0.2 * precision[i] * recall[i]) / (0.1 * precision[i] + recall[i]));
				fMeasure[i] = 2 * ((precision[i] * recall[i]) / (precision[i] + recall[i]));

			}

			double maxFMeasure = 0.0;
			double maxThreshold = 0.0;
			int maxThresholdIndex = 0;
			int maxStep = 0;

			for (int i = 0; i < length; i++) {
				// double precision = sumPrecision / count;
				// double recall = sumRecall / count;

				if (fMeasure[i] > maxFMeasure) {
					maxFMeasure = fMeasure[i];
					maxThreshold = step * i;
					maxThresholdIndex = i;
				}
			}

			// for (int i = maxThresholdIndex + 1; i < length; i++) {
			// if (precision[i] > precision[maxThresholdIndex]) {
			// if (3 * (precision[i] - precision[maxThresholdIndex]) >
			// (Math.abs(fMeasure[i] - fMeasure[maxThresholdIndex]))) {
			// maxThreshold = step * i;
			// maxStep = i;
			// maxThresholdIndex = i;
			// }
			// }
			//
			// }

			// if (alternative != null && alternative > maxThreshold)
			// return (maxThreshold + alternative) / 2.0;
			// else
			return maxThreshold;
		} else {
			if (alternative != null) {
				return alternative;
			} else {
				return 0.0;
			}
		}
	}
}
