package de.mannheim.uni.expansion.tablefusion.profiling;

import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.FusionPerformanceSynchronizedCreator;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.generic.GenericResultsWriterAccumRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.services.grouping.GroupedValuesRepository;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumInteger;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AccumulatorListSimple;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AcumReference;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.GenericAccumaltableResult;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList.ColumnListBuilder;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.GenericAccumaltableResult.ResultBuilder;

public class PldProfilerDistributed {

	private final GroupedValuesRepository groupedValuesRepository;
	private final OntologyRepository ontologyRepository;
	private final AnnotatedTablesRepository tableRepository;

	@Inject
	public PldProfilerDistributed(GroupedValuesRepository groupedValuesRepository, OntologyRepository ontologyRepository,
			AnnotatedTablesRepository tableRepository) {
		this.groupedValuesRepository = groupedValuesRepository;
		this.ontologyRepository = ontologyRepository;
		this.tableRepository = tableRepository;
	}

	// private static final Pair<Integer, String> TOTAL_MAPPED_VALUES = new
	// Pair<>(0, "TotalMappedValues");
	// private static final Pair<Integer, String> ONE_VALUE_PER_PLD = new
	// Pair<>(1, "OneValuePerPLD");
	// private static final Pair<Integer, String> AVERAGE_PER_PLD = new
	// Pair<>(2, "AveragePerPLD");
	// private static final Pair<Integer, String> MEDIAN_PER_PLD = new Pair<>(3,
	// "MedianPerPLD");

	public void profileByPld(String outputFolder, String outputPrefix) {

		AccumulatorListSimple accumulatorList = new AccumulatorListSimple();

		AcumReference<AccumInteger> TOTAL_GROUPS = accumulatorList.addAcummulator(AccumInteger.class, new AccumInteger(0));
		ConcurrentHashMap<Integer, AcumReference<AccumInteger>> groupSizes = new ConcurrentHashMap<>();

		FusionPerformanceSynchronizedCreator<GenericAccumaltableResult> creatorSingle = new FusionPerformanceSynchronizedCreator<>(
				GenericAccumaltableResult.empty(accumulatorList));

		GroupedValues groupedValues = groupedValuesRepository.getGroupedValues();

		PersistentParallel6.runAndFinish(groupedValues.entrySet(), (group) -> {

			Set<String> pldSet = new HashSet<>();

			for (SingleValue sv : group.getValue().getValues()) {
				AnnotatedTableMetaData md = this.tableRepository.getMetaData(sv.getId().getTableId());
				String PLD = md.getSource().getPLD();
				if (PLD != null && !pldSet.contains(PLD)) {
					pldSet.add(PLD);
				}
			}

			int groupSize = pldSet.size();

			if (!groupSizes.containsKey(groupSize)) {
				synchronized (accumulatorList) {
					if (!groupSizes.containsKey(groupSize)) {
						AcumReference<AccumInteger> ref = accumulatorList.addAcummulator(AccumInteger.class, new AccumInteger(0));
						groupSizes.put(groupSize, ref);
					}
				}
			}

			ResultBuilder resultBuilder = GenericAccumaltableResult.builder(accumulatorList);
			resultBuilder.setValue(TOTAL_GROUPS, new AccumInteger(1));
			resultBuilder.setValue(groupSizes.get(groupSize), new AccumInteger(1));

			GenericAccumaltableResult result = resultBuilder.build();

			creatorSingle.addForAll(group.getKey().getClassPropertyId(), ontologyRepository, result);

		} , this.getClass());

		ColumnListBuilder columnListBuilder = ColumnList.builderAccumRepository();
		columnListBuilder.addColumn("TotalGroups", (repository) -> {
			return String.valueOf(repository.get(TOTAL_GROUPS).get());
		});

		TreeMap<Integer, AcumReference<AccumInteger>> sortedGroupSizes = new TreeMap<>(groupSizes);
		for (Entry<Integer, AcumReference<AccumInteger>> groupSize : sortedGroupSizes.entrySet()) {
			columnListBuilder.addColumn(String.valueOf(groupSize.getKey()), (repository) -> {
				return String.valueOf(repository.get(groupSize.getValue()).get());
			});
		}

		ColumnList columnList = columnListBuilder.build();

		GenericResultsWriterAccumRepository writer = new GenericResultsWriterAccumRepository(columnList);
		writer.writeAccumRepository("PLDProfiling", ontologyRepository, creatorSingle.get());

		try

		{
			if (outputFolder != null && outputPrefix != null) {
				MKDIR.createDirUnchecked(new File(outputFolder));
				writer.writeToFolder(outputFolder, outputPrefix);

			}
		} catch (

		IOException e1)

		{
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		// Set<String> uniquePLDs = new HashSet<>();
		// for (SingleValue sv : e.getValue().getValues()) {
		// String source =
		// annotatedTablesRepository.getMetaData(sv.getId().getTableId()).getSource().getSource();
		//
		// String PLD = DomainUtils.getPayLevelDomainFromWholeURL(source);
		// uniquePLDs.add(PLD);
		//
		// }

	}
}
