package de.mannheim.uni.expansion.tablefusion.profiling;

import java.io.File;
import java.io.IOException;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;
import javax.inject.Named;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.NamedProperties;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.FusionPerformanceSynchronizedCreator;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.GenericAbstractResultsWriter;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.ResultsCategory;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.generic.GenericResultsWriterAccumRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.model.utils.performance.FusionPerformance;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumAverage;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumInteger;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumMedian;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AccumulatorListImmutable;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AcumReference;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.GenericAccumaltableResult;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AccumulatorListImmutable.ImmutableAccumulatorListBuild;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList.ColumnListBuilder;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.GenericAccumaltableResult.ResultBuilder;

public class TimeseriesProfiler {

	private final AnnotatedTablesRepository tablesRepository;
	private final OntologyRepository ontologyRepository;
	private final String resultsFolder;

	@Inject
	public TimeseriesProfiler(AnnotatedTablesRepository tablesRepository, OntologyRepository ontologyRepository,
			@Named(NamedProperties.RESULTS_FOLDER) String resultsFolder) {
		this.tablesRepository = tablesRepository;
		this.ontologyRepository = ontologyRepository;
		this.resultsFolder = resultsFolder;
	}

	public void profileTimeStamps(String outputFolder, String outputPrefix) {

		ImmutableAccumulatorListBuild builder = AccumulatorListImmutable.builder();

		AcumReference<AccumInteger> TOTAL_TS = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));
		AcumReference<AccumAverage> AVERAGE_SOURCE_PER_TS = builder.addAcummulator(AccumAverage.class, new AccumAverage());
		AcumReference<AccumMedian> MEDIAN_SOURCE_PER_TS = builder.addAcummulator(AccumMedian.class, new AccumMedian());
		AcumReference<AccumInteger> TOTAL_SOURCES = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));
		AcumReference<AccumInteger> BEFORE = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));
		AcumReference<AccumInteger> AFTER = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));
		AcumReference<AccumInteger> OnPage = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));
		AcumReference<AccumInteger> TABLE_TITLE = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));
		AcumReference<AccumInteger> PageTitle = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));
		AcumReference<AccumInteger> Columnheader = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));
		AcumReference<AccumInteger> ROW_CELL = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));
		AcumReference<AccumInteger> AT_LEAST_ONE = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));

		ColumnListBuilder columnListBuilder = ColumnList.builderAccumRepository();
		columnListBuilder.addColumn("Timeseries", (repository) -> {
			return String.valueOf(repository.get(TOTAL_TS).get());
		});

		columnListBuilder.addColumn("AverageSourcePerTimeSeries", (repository) -> {
			return String.valueOf(repository.get(AVERAGE_SOURCE_PER_TS).getValue());
		});

		columnListBuilder.addColumn("MedianSourcePerTimeSeries", (repository) -> {
			return String.valueOf(repository.get(MEDIAN_SOURCE_PER_TS).getValue());
		});

		columnListBuilder.addColumn("TotalSources", (repository) -> {
			return String.valueOf(repository.get(TOTAL_SOURCES).get());
		});

		columnListBuilder.addColumn("Before", (repository) -> {
			return String.valueOf(repository.get(BEFORE).getValue() / ((double) repository.get(TOTAL_SOURCES).get()));
		});

		columnListBuilder.addColumn("After", (repository) -> {
			return String.valueOf(repository.get(AFTER).getValue() / ((double) repository.get(TOTAL_SOURCES).get()));
		});

		columnListBuilder.addColumn("OnPage", (repository) -> {
			return String.valueOf(repository.get(OnPage).getValue() / ((double) repository.get(TOTAL_SOURCES).get()));
		});

		columnListBuilder.addColumn("TableTitle", (repository) -> {
			return String.valueOf(repository.get(TABLE_TITLE).getValue() / ((double) repository.get(TOTAL_SOURCES).get()));
		});

		columnListBuilder.addColumn("PageTitle", (repository) -> {
			return String.valueOf(repository.get(PageTitle).getValue() / ((double) repository.get(TOTAL_SOURCES).get()));
		});

		columnListBuilder.addColumn("ColumnHeader", (repository) -> {
			return String.valueOf(repository.get(Columnheader).getValue() / ((double) repository.get(TOTAL_SOURCES).get()));
		});

		columnListBuilder.addColumn("ROW_CELL", (repository) -> {
			return String.valueOf(repository.get(ROW_CELL).getValue() / ((double) repository.get(TOTAL_SOURCES).get()));
		});

		columnListBuilder.addColumn("AtLeastOne", (repository) -> {
			return String.valueOf(repository.get(AT_LEAST_ONE).getValue() / ((double) repository.get(TOTAL_SOURCES).get()));
		});

		ColumnList columnList = columnListBuilder.build();
		AccumulatorListImmutable resultsType = builder.getResultsType();

		FusionPerformanceSynchronizedCreator<GenericAccumaltableResult> creatorSingle = new FusionPerformanceSynchronizedCreator<>(
				GenericAccumaltableResult.empty(resultsType));

		// FusionPerformanceSynchronizedCreator<TimeStampsAggregatedResults>
		// creatorAggregated = new FusionPerformanceSynchronizedCreator<>(
		// new TimeStampsAggregatedResults());

		ConcurrentHashMap<Pair<TripleId, String>, ResultBuilder> PLD_BUILDER = new ConcurrentHashMap<>();

		ConcurrentHashMap<TripleId, Set<String>> PLD_COUNTER = new ConcurrentHashMap<>();
		PersistentParallel6.runAndFinish(tablesRepository.getAllTableIds(), (tid) -> {

			AnnotatedTablesReaderData table = new AnnotatedTablesReaderData(tid, tablesRepository.getMetaData(tid), tablesRepository.getTable(tid));

			ClassId classId = table.getMetaData().getDbPediaClass();

			boolean hasBeforeText = false;
			boolean hasAfterText = false;
			boolean hasInContextBefore = false;
			boolean hasInTableTitle = false;
			boolean hasInPageTitle = false;

			if (!table.getMetaData().getYearsBeforeText().isEmpty())
				hasBeforeText = true;
			if (!table.getMetaData().getYearsAfterText().isEmpty())
				hasAfterText = true;
			if (!table.getMetaData().getYearsInContextBefore().isEmpty())
				hasInContextBefore = true;
			if (!table.getMetaData().getYearsInTableTitle().isEmpty())
				hasInTableTitle = true;
			if (!table.getMetaData().getYearsInPageTitle().isEmpty())
				hasInPageTitle = true;

			Map<Integer, List<Integer>> columnYearsMap = table.getMetaData().getPropertiesYears();
			Map<Integer, PropertyId> columns = table.getMetaData().getProperties();

			int numOfColumns = 0;
			int numOfColumnsWithOwnTimeStamp = 0;

			// Set<ClassPropertyId> propertyIdSet = new HashSet<>();
			// Set<ValueType> cdtSet = new HashSet<>();
			for (Entry<Integer, PropertyId> e : columns.entrySet()) {
				if (e.getValue() != null) {
					ClassPropertyId cpid = new ClassPropertyId(classId, e.getValue());

					if (ontologyRepository.isTimeDependent(e.getValue())) {
						numOfColumns++;

						boolean hasColumn = false;
						if (columnYearsMap != null) {
							List<Integer> columnYears = columnYearsMap.get(e.getKey());
							if (columnYears != null && columnYears.size() > 0) {
								hasColumn = true;

							}
						}

						if (hasColumn) {
							numOfColumnsWithOwnTimeStamp++;
						}

						int numOfValues = 0;
						int hasRowCellYear = 0;
						for (Entry<RowIdx, AbstractObject> v : table.getTableData().getColumn(new ColumnIdx(e.getKey())).entrySet()) {
							if (v.getValue() != null && table.getMetaData().getAllInstances().containsKey(v.getKey().getRowIdx())
									&& !table.getMetaData().getAllInstances().get(v.getKey().getRowIdx()).isEmpty()) {
								numOfValues++;
								if (hasCellYear(table.getTableData(), table.getMetaData(), v.getKey()))
									hasRowCellYear++;
							}
						}

						// TimeStampsSingleResults res = new
						// TimeStampsSingleResults();

						String PLD = tablesRepository.getMetaData(tid).getSource().getPLD();
						if (PLD != null) {

							Set<InstanceId> instances = new HashSet<>(table.getMetaData().getInstances().values());

							for (InstanceId instanceId : instances) {
								TripleId tripleId = new TripleId(classId, e.getValue(), instanceId);
								ResultBuilder resultBuilder = CHMHelper.getFromCHM(PLD_BUILDER, new Pair<>(tripleId, PLD),
										() -> GenericAccumaltableResult.builder(resultsType));

								CHMHelper.getFromCHM(PLD_COUNTER, tripleId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(PLD);

								synchronized (resultBuilder) {

									// resultBuilder.setValue(N, new
									// AccumInteger(1));

									boolean atLeastOne = false;
									if (hasBeforeText) {
										resultBuilder.setValue(BEFORE, new AccumInteger(1));
										atLeastOne = true;
									}
									if (hasAfterText) {
										resultBuilder.setValue(AFTER, new AccumInteger(1));

										atLeastOne = true;
									}
									if (hasInContextBefore) {
										resultBuilder.setValue(OnPage, new AccumInteger(1));
										atLeastOne = true;
									}

									if (hasInTableTitle) {
										resultBuilder.setValue(TABLE_TITLE, new AccumInteger(1));
										atLeastOne = true;
									}
									if (hasInPageTitle) {
										resultBuilder.setValue(PageTitle, new AccumInteger(1));
										atLeastOne = true;
									}
									if (hasColumn) {
										resultBuilder.setValue(Columnheader, new AccumInteger(1));
										atLeastOne = true;
									}

									if (atLeastOne) {
										resultBuilder.setValue(AT_LEAST_ONE, new AccumInteger(1));
									}

									if (hasRowCellYear > 0) {
										resultBuilder.setValue(ROW_CELL, new AccumInteger(1));
										// res.inRowCellValue(hasRowCellYear);
									}

									// ValueType cdt =
									// ontologyRepository.getDataType(e.getValue());
									// creatorSingle.addPerClassId(table.getMetaData().getDbPediaClass(),
									// res);
									// if (cdt != null) {
									// creatorSingle.addPerDataType(cdt, res);
									// creatorAggregated.addPerDataType(cdt,
									// res2);
									// cdtSet.add(cdt);
									// } else
									// System.out.println("CDT NULL FOR CLASSiD
									// " +
									// table.getMetaData().getDbPediaClass() + "
									// and
									// PropertyID " + e.getValue() + ".");

									// ClassPropertyId cpid = new
									// ClassPropertyId(table.getMetaData().getDbPediaClass(),
									// e.getValue());
									// creatorSingle.addPerProperty(cpid, res);
									// creatorAggregated.addPerProperty(cpid,
									// res2);
									// propertyIdSet.add(cpid);
								}
							}
						}
					}
				} else {
					System.out.println("UNMAPPED COLUMN");
				}

			}

			// TimeStampsAggregatedResults res = null;
			//
			// TimeStampsAggregatedResults res2 = null;
			// if (hasTable) {
			// res = new TimeStampsAggregatedResults(1, 1, numOfColumns,
			// numOfColumns, numOfColumnsWithOwnTimeStamp,
			// numOfColumnsWithOwnTimeStamp);
			// res2 = new TimeStampsAggregatedResults(1, 1, 0, 0, 0, 0);
			// } else {
			//
			// res = new TimeStampsAggregatedResults(1, 0, numOfColumns, 0,
			// numOfColumnsWithOwnTimeStamp, 0);
			// res2 = new TimeStampsAggregatedResults(1, 0, 0, 0, 0, 0);
			// }
			//
			// creatorAggregated.addTotal(res);
			// creatorAggregated.addPerClassId(table.getMetaData().getDbPediaClass(),
			// res);
			//
			// for (ColumnDataType cdt : cdtSet) {
			// creatorAggregated.addPerDataType(cdt, res2);
			// }
			//
			// for (ClassPropertyId classPropertyId : propertyIdSet) {
			// creatorAggregated.addPerProperty(classPropertyId, res2);
			// }

			// e.getMetaData()
		} , this.getClass());

		for (Entry<Pair<TripleId, String>, ResultBuilder> entry : PLD_BUILDER.entrySet()) {

			creatorSingle.addForAll(entry.getKey().getFirst().getClassPropertyId(), ontologyRepository, entry.getValue().build());
		}

		for (Entry<TripleId, Set<String>> entry : PLD_COUNTER.entrySet()) {
			ResultBuilder resultBuilder = GenericAccumaltableResult.builder(resultsType);
			resultBuilder.setValue(TOTAL_TS, new AccumInteger(1));
			resultBuilder.setValue(MEDIAN_SOURCE_PER_TS, new AccumMedian((double) entry.getValue().size()));
			resultBuilder.setValue(AVERAGE_SOURCE_PER_TS, new AccumAverage((double) entry.getValue().size()));
			resultBuilder.setValue(TOTAL_SOURCES, new AccumInteger(entry.getValue().size()));
			creatorSingle.addForAll(entry.getKey().getClassPropertyId(), ontologyRepository, resultBuilder.build());
		}

		GenericResultsWriterAccumRepository writer = new GenericResultsWriterAccumRepository(columnList);
		writer.writeAccumRepository("TimeseriesProfiling", ontologyRepository, creatorSingle.get());

		try {
			if (outputFolder != null && outputPrefix != null) {
				MKDIR.createDirUnchecked(new File(outputFolder));
				writer.writeToFolder(outputFolder, outputPrefix);

			}
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
	}

	private <E> void writePerformance(GenericAbstractResultsWriter<E> performanceResultsWriter, FusionPerformance<E> performance, String name)
			throws IOException {

		{
			int index = performanceResultsWriter.getCurrentIndex(ResultsCategory.TOTAL);
			String[] prefix = ResultsCategory.getNormalPrefix(name);
			performanceResultsWriter.write(ResultsCategory.TOTAL, prefix, performance.getTotal());
			System.out.println(performanceResultsWriter.getFrom(ResultsCategory.TOTAL, index));

		}

		{
			int index = performanceResultsWriter.getCurrentIndex(ResultsCategory.DATA_TYPE);
			for (Entry<ValueType, E> e : performance.getPerDataType().entrySet()) {
				String[] prefix = ResultsCategory.getDataTypePrefix(name, e.getKey());
				performanceResultsWriter.write(ResultsCategory.DATA_TYPE, prefix, e.getValue());
			}

			System.out.println(performanceResultsWriter.getFrom(ResultsCategory.DATA_TYPE, index));

		}

		{

			for (Entry<ClassId, E> e : performance.getPerClassId().entrySet()) {
				String[] prefix = ResultsCategory.getClassIdPrefix(name, e.getKey());
				performanceResultsWriter.write(ResultsCategory.CLASS_ID, prefix, e.getValue());
			}

		}

		{

			for (Entry<ClassPropertyId, E> e : performance.getPerProperty().entrySet()) {
				ValueType cdt = ontologyRepository.getDataType(e.getKey().getPropertyId());
				String[] prefix = ResultsCategory.getClassPropertyIdPrefix(name, ontologyRepository.isTimeDependent(e.getKey().getPropertyId()), cdt,
						e.getKey());
				performanceResultsWriter.write(ResultsCategory.CLASS_PROPERTY_ID, prefix, e.getValue());
			}

		}

	}

	boolean hasCellYear(AnnotatedTable td, AnnotatedTableMetaData md, RowIdx rowIdx) {

		PropertyId activeYearsStartYear = new PropertyId("http://dbpedia.org/ontology/activeYearsStartYear");

		for (int i = 0; i < td.getColumns().size(); i++) {
			if (td.getDataType(new ColumnIdx(i)) == ValueType.TIME) {
				PropertyId propertyId = md.getProperty(i);
				if (true || propertyId == null || propertyId.equals(activeYearsStartYear)) {
					AbstractObject oYear = td.getValue(rowIdx, new ColumnIdx(i));
					if (oYear instanceof TimeObject) {
						return true;
					}
				}
			}

		}

		return false;

	}
}
