package de.mannheim.uni.expansion.tablefusion.profiling.timestamps;

import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.Qualifier;
import de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats.Stats;

public class KbTimeStampProfiler {
	private final KnowledgeBase knowledgeBase;

	@Inject
	public KbTimeStampProfiler(KnowledgeBase knowledgeBase) {
		this.knowledgeBase = knowledgeBase;
	}

	public void profileKB(String file) throws IOException {

		// Map<ClassPropertyId, Map<TimestampPointInTime, MutableInt>> countMap
		// = new HashMap<>();

		Map<ClassPropertyId, MutableInt> totalMap = new HashMap<>();
		Map<ClassPropertyId, MutableInt> withTimetampsMap = new HashMap<>();
		Map<ClassPropertyId, Integer> uniqueValues = new HashMap<>();

		Map<ClassId, Integer> totalPerClass = new HashMap<>();

		for (ClassPropertyId cpid : knowledgeBase.getClassPropertyIds()) {
			MutableInt total = new MutableInt();
			MutableInt withTimestamps = new MutableInt();

			int countWithValues = 0;
			Set<InstanceId> instances = this.knowledgeBase.getInstances(cpid.getClassId(), cpid.getPropertyId());
			for (InstanceId instanceId : instances) {

				QualifiedValueList qvl = this.knowledgeBase.getQualifiedValueList(new ValueId(instanceId, cpid.getPropertyId()));
				for (QualifiedValueHolder claim : qvl.getClaims()) {
					total.increment();
					List<Qualifier> qualifiers = claim.getQualifiers();
					if (qualifiers != null) {
						for (Qualifier qualifier : qualifiers) {
							if (qualifier.getQualifier().equals(Stats.endTimeQualifier)) {
								withTimestamps.increment();
								break;
							} else if (qualifier.getQualifier().equals(Stats.startTimeQualifier)) {
								withTimestamps.increment();
								break;
							} else if (qualifier.getQualifier().equals(Stats.pointInTime)) {
								withTimestamps.increment();
								break;
							}
						}
					}
				}

				if (total.intValue() > 0) {
					countWithValues++;
				}

				totalMap.put(cpid, total);
				withTimetampsMap.put(cpid, withTimestamps);

			}

			uniqueValues.put(cpid, countWithValues);

		}
		for (ClassId setId : knowledgeBase.getClasses()) {
			totalPerClass.put(setId, knowledgeBase.getNumberOfInstances(setId));
		}

		CSVWriter csvWriter = new CSVWriter(new FileWriter(file));

		int length = 8;

		String[] headerLine = new String[length];
		headerLine[0] = "Class";
		headerLine[1] = "PropertyId";
		headerLine[2] = "PropertyName";
		headerLine[3] = "TotalValues";
		headerLine[4] = "WithTimestamps";
		headerLine[5] = "InstancesWithValues";
		headerLine[6] = "TotalInstances";
		headerLine[7] = "Density";

		csvWriter.writeNext(headerLine);

		for (Entry<ClassPropertyId, MutableInt> entry : totalMap.entrySet()) {
			// if
			// (Stats.propertyMap.containsKey(entry.getKey().getPropertyId())) {
			MutableInt withTimestamps = withTimetampsMap.get(entry.getKey());
			MutableInt total = totalMap.get(entry.getKey());

			String[] line = new String[length];
			line[0] = entry.getKey().getClassId().toString();
			line[1] = entry.getKey().getPropertyId().getPropertyId();
			line[2] = entry.getKey().getPropertyId().getReadableString();
			line[3] = String.valueOf(total.intValue());
			line[4] = String.valueOf(withTimestamps.intValue());
			line[5] = String.valueOf(uniqueValues.get(entry.getKey()));
			line[6] = String.valueOf(totalPerClass.get(entry.getKey().getClassId()));
			line[7] = String.valueOf(((double) uniqueValues.get(entry.getKey())) / ((double) totalPerClass.get(entry.getKey().getClassId())));

			csvWriter.writeNext(line);
			// }
		}

		csvWriter.close();

	}

	// public void profileTables(String file) throws IOException {
	//
	// ConcurrentHashMap<ClassId, ConcurrentHashMap<TimestampPointInTime,
	// AtomicInteger>> countMap = new ConcurrentHashMap<>();
	//
	// for (AnnotatedTableId tableId : this.tableRepository.getAllTableIds()) {
	//
	// AnnotatedTableMetaData md = this.tableRepository.getMetaData(tableId);
	// AnnotatedTable td = this.tableRepository.getTable(tableId);
	//
	// for (Entry<TimestampIdentifier, Integer> e :
	// md.getTableYears().entrySet()) {
	// ConcurrentHashMap<TimestampPointInTime, AtomicInteger> innerMap =
	// CHMHelper.getFromCHM(countMap, md.getDbPediaClass(),
	// () -> new ConcurrentHashMap<>());
	// CHMHelper.getFromCHM(innerMap, new TimestampPointInTime(e.getValue()), ()
	// -> new AtomicInteger()).incrementAndGet();
	// }
	//
	// for (Entry<ColumnIdx, ValueType> e : td.getDataTypes().entrySet()) {
	// if (e.getValue().equals(ValueType.TIME)) {
	//
	// ConcurrentHashMap<TimestampPointInTime, AtomicInteger> innerMap =
	// CHMHelper.getFromCHM(countMap, md.getDbPediaClass(),
	// () -> new ConcurrentHashMap<>());
	// for (Entry<RowIdx, AbstractObject> v :
	// td.getColumn(e.getKey()).entrySet()) {
	//
	// Integer year = TimeObject.getYear(v.getValue());
	// if (year != null) {
	// CHMHelper.getFromCHM(innerMap, new TimestampPointInTime(year), () -> new
	// AtomicInteger()).incrementAndGet();
	// }
	// }
	//
	// }
	// }
	// }
	//
	// CSVWriter csvWriter = new CSVWriter(new FileWriter(file));
	//
	// int length = 3;
	//
	// String[] headerLine = new String[length];
	// headerLine[0] = "Class";
	// headerLine[1] = "Year";
	// headerLine[2] = "Count";
	//
	// csvWriter.writeNext(headerLine);
	//
	// for (Entry<ClassId, ConcurrentHashMap<TimestampPointInTime,
	// AtomicInteger>> cm : countMap.entrySet()) {
	// for (Entry<TimestampPointInTime, AtomicInteger> ce :
	// cm.getValue().entrySet()) {
	// String[] line = new String[length];
	// line[0] = cm.getKey().toString();
	// line[1] = String.valueOf(ce.getKey().getYear());
	// line[2] = String.valueOf(ce.getValue().intValue());
	//
	// csvWriter.writeNext(line);
	// }
	// }
	//
	// csvWriter.close();
	//
	// }

}
