package de.mannheim.uni.expansion.tablefusion.profiling.timestamps;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.FusionPerformanceSynchronizedCreator;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.generic.GenericResultsWriterAccumRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.services.grouping.GroupedValuesRepository;
import de.mannheim.uni.expansion.utils.MKDIR;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.qualifiers.TimestampPointInTime;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.evaluation.timestamps.TimestampQualifiersAdapter;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumInteger;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AccumulatorListImmutable;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AcumReference;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.GenericAccumaltableResult;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.AccumulatorListImmutable.ImmutableAccumulatorListBuild;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ColumnList.ColumnListBuilder;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.GenericAccumaltableResult.ResultBuilder;

public class TimeStampDebiustrutdedOverlapProfiler {
	private final AnnotatedTablesRepository tableRepository;
	private final KnowledgeBase knowledgeBase;
	private final GroupedValuesRepository groupedValuesRepository;
	private final SimilarityCalculator similarityCalculator;
	private final OntologyRepository ontologyRepository;
	private final TimestampQualifiersAdapter timestampQualifiersAdapter;

	@Inject
	public TimeStampDebiustrutdedOverlapProfiler(AnnotatedTablesRepository tableRepository, KnowledgeBase knowledgeBase,
			GroupedValuesRepository groupedValuesRepository, SimilarityCalculator similarityCalculator, OntologyRepository ontologyRepository,
			TimestampQualifiersAdapter timestampQualifiersAdapter) {
		this.tableRepository = tableRepository;
		this.knowledgeBase = knowledgeBase;
		this.groupedValuesRepository = groupedValuesRepository;
		this.similarityCalculator = similarityCalculator;
		this.ontologyRepository = ontologyRepository;
		this.timestampQualifiersAdapter = timestampQualifiersAdapter;
	}

	public void profile(String outputFolder, String outputPrefix) {

		ImmutableAccumulatorListBuild builder = AccumulatorListImmutable.builder();
		AcumReference<AccumInteger> totalRef = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));
		AcumReference<AccumInteger> totalInKbRef = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));
		AcumReference<AccumInteger> totalCorrectRef = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));

		AcumReference<AccumInteger> other = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));

		ColumnListBuilder columnListBuilder = ColumnList.builderAccumRepository();

		columnListBuilder.addColumn("Total", (repository) -> {
			return repository.get(totalRef).toString();
		});

		columnListBuilder.addColumn("TotalInKB", (repository) -> {
			return repository.get(totalInKbRef).toString();
		});

		columnListBuilder.addColumn("TotalCorrect", (repository) -> {
			return repository.get(totalCorrectRef).toString();
		});

		Map<Integer, AcumReference<AccumInteger>> acums = new HashMap<>();
		for (int i = 1950; i < 2016; i++) {
			AcumReference<AccumInteger> ref = builder.addAcummulator(AccumInteger.class, new AccumInteger(0));
			acums.put(i, ref);
			columnListBuilder.addColumn(String.valueOf(i), (repository) -> {
				return repository.get(ref).toString();
			});
		}

		columnListBuilder.addColumn("Other", (repository) -> {
			return repository.get(other).toString();
		});

		ColumnList columnList = columnListBuilder.build();

		AccumulatorListImmutable resultsType = builder.getResultsType();
		FusionPerformanceSynchronizedCreator<GenericAccumaltableResult> creatorSingle = new FusionPerformanceSynchronizedCreator<>(
				GenericAccumaltableResult.empty(resultsType));

		GroupedValues groupedValues = groupedValuesRepository.getGroupedValues();

		PersistentParallel4.runAndFinish(groupedValues.entrySet(), (e) -> {

			if (ontologyRepository.isTimeDependent(e.getKey().getPropertyId())) {
				TripleId tid = e.getKey();

				{
					ResultBuilder resultBuilder = GenericAccumaltableResult.builder(resultsType);
					resultBuilder.setValue(totalRef, new AccumInteger(1));
					creatorSingle.addForAll(tid.getClassPropertyId(), ontologyRepository, resultBuilder.build());
				}

				QualifiedValueList qvl = knowledgeBase.getQualifiedValueList(tid.getValueId());
				if (qvl != null) {

					{
						ResultBuilder resultBuilder = GenericAccumaltableResult.builder(resultsType);
						resultBuilder.setValue(totalInKbRef, new AccumInteger(1));
						creatorSingle.addForAll(tid.getClassPropertyId(), ontologyRepository, resultBuilder.build());
					}

					for (SingleValue sv : e.getValue().getValues()) {
						AbstractObject value = sv.getValue();
						AnnotatedTable table = tableRepository.getTable(sv.getId().getTableId());
						if (ontologyRepository.getDataType(e.getKey().getPropertyId()).equals(ValueType.INSTANCE_REFERENCE)) {
							int asd = 123;
						}
						boolean added = false;
						for (QualifiedValueHolder qvh : qvl.getClaims()) {

							if (similarityCalculator.isEqual(value, qvh.getValue())) {

								Set<TimestampPointInTime> timestamps = this.timestampQualifiersAdapter.getQualifierFromQVH(qvh);

								// Map<TimestampPointInTime, List<Integer>>
								// qualifierMap =
								// this.timestampQualifiersAdapter.getQualifierMapFromQVL(qvl);

								RowIdx rowIdx = sv.getId().getRowIdx();

								List<ColumnIdx> timeColumns = table.getTimeColumns();
								for (ColumnIdx columnIdx : timeColumns) {
									Integer columnYear = TimeObject.getYear(table.getColumn(columnIdx).get(rowIdx));
									if (columnYear != null) {
										TimestampPointInTime tpit = new TimestampPointInTime(columnYear);
										if (timestamps.contains(tpit)) {
											ResultBuilder resultBuilder = GenericAccumaltableResult.builder(resultsType);
											if (acums.containsKey(tpit.getYear())) {
												resultBuilder.setValue(acums.get(tpit.getYear()), new AccumInteger(1));
											} else {
												resultBuilder.setValue(other, new AccumInteger(1));
											}
											resultBuilder.setValue(totalCorrectRef, new AccumInteger(1));
											creatorSingle.addForAll(tid.getClassPropertyId(), ontologyRepository, resultBuilder.build());
											added = true;
											continue;
										}

									}
								}

								Map<TimestampIdentifier, Integer> tableYears = tableRepository.getMetaData(sv.getId().getTableId()).getTableYears();

								for (Entry<TimestampIdentifier, Integer> tableYear : tableYears.entrySet()) {
									TimestampPointInTime tpit = new TimestampPointInTime(tableYear.getValue());
									if (timestamps.contains(tpit)) {

										ResultBuilder resultBuilder = GenericAccumaltableResult.builder(resultsType);
										if (acums.containsKey(tpit.getYear())) {
											resultBuilder.setValue(acums.get(tpit.getYear()), new AccumInteger(1));
										} else {
											resultBuilder.setValue(other, new AccumInteger(1));
										}
										resultBuilder.setValue(totalCorrectRef, new AccumInteger(1));
										creatorSingle.addForAll(tid.getClassPropertyId(), ontologyRepository, resultBuilder.build());
										added = true;
										continue;
									}

								}

							}

						}

						if (ontologyRepository.getDataType(e.getKey().getPropertyId()).equals(ValueType.INSTANCE_REFERENCE)) {
							AnnotatedTableMetaData md = this.tableRepository.getMetaData(sv.getId().getTableId());
							Map<TimestampIdentifier, Integer> tableYears = tableRepository.getMetaData(sv.getId().getTableId()).getTableYears();
							Map<ClassId, Set<InstanceId>> instancesMap = new HashMap<>();
							for (ClassId instancesSet : ontologyRepository.getRange(e.getKey().getClassPropertyId())) {
								instancesMap.put(instancesSet, knowledgeBase.getInstances(instancesSet).keySet());
								// instancesMap.put();
							}
							if (!added) {
								int asd = 123;
							}
						}
					}
				}
			}
		} , "OverlapProfiler");

		GenericResultsWriterAccumRepository writer = new GenericResultsWriterAccumRepository(columnList);

		writer.writeAccumRepository("OverlapWithDistrubtion", ontologyRepository, creatorSingle.get());

		try {
			if (outputFolder != null && outputPrefix != null) {

				MKDIR.createDirUnchecked(new File(outputFolder));
				writer.writeToFolder(outputFolder, outputPrefix);

			}
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
	}

}
