package de.mannheim.uni.expansion.tablefusion.profiling.timestamps;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.inject.Inject;

import org.apache.commons.io.FileUtils;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.model.timestamp.TimestampIdentifier;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.stores.FusionPerformanceSynchronizedCreator;
import de.mannheim.uni.expansion.tablefusion.analysis.writer.ResultsCategory;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.mannheim.uni.expansion.tablefusion.services.grouping.GroupedValuesRepository;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.qualifiers.TimestampPointInTime;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.evaluation.timestamps.TimestampQualifiersAdapter;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulateables.AccumInteger;

public class TimeStampOverlapProfiler {
	private final AnnotatedTablesRepository tableRepository;
	private final KnowledgeBase knowledgeBase;
	private final GroupedValuesRepository groupedValuesRepository;
	private final SimilarityCalculator similarityCalculator;
	private final OntologyRepository ontologyRepository;
	private final TimestampQualifiersAdapter timestampQualifiersAdapter;

	@Inject
	public TimeStampOverlapProfiler(AnnotatedTablesRepository tableRepository, KnowledgeBase knowledgeBase, GroupedValuesRepository groupedValuesRepository,
			SimilarityCalculator similarityCalculator, OntologyRepository ontologyRepository, TimestampQualifiersAdapter timestampQualifiersAdapter) {
		this.tableRepository = tableRepository;
		this.knowledgeBase = knowledgeBase;
		this.groupedValuesRepository = groupedValuesRepository;
		this.similarityCalculator = similarityCalculator;
		this.ontologyRepository = ontologyRepository;
		this.timestampQualifiersAdapter = timestampQualifiersAdapter;
	}

	public void profile(String outputFile) {

		FusionPerformanceSynchronizedCreator<AccumInteger> creatorSingle = new FusionPerformanceSynchronizedCreator<AccumInteger>(new AccumInteger(0));

		GroupedValues groupedValues = groupedValuesRepository.getGroupedValues();

		PersistentParallel4.runAndFinish(groupedValues.entrySet(), (e) -> {

			if (ontologyRepository.isTimeDependent(e.getKey().getPropertyId())) {
				TripleId tid = e.getKey();

				QualifiedValueList qvl = knowledgeBase.getQualifiedValueList(tid.getValueId());
				if (qvl != null) {
					for (SingleValue sv : e.getValue().getValues()) {
						AbstractObject value = sv.getValue();
						AnnotatedTable table = tableRepository.getTable(sv.getId().getTableId());
						if (ontologyRepository.getDataType(e.getKey().getPropertyId()).equals(ValueType.INSTANCE_REFERENCE)) {
							int asd = 123;
						}
						boolean added = false;
						for (QualifiedValueHolder qvh : qvl.getClaims()) {

							if (similarityCalculator.isEqual(value, qvh.getValue())) {

								Set<TimestampPointInTime> timestamps = this.timestampQualifiersAdapter.getQualifierFromQVH(qvh);

								// Map<TimestampPointInTime, List<Integer>>
								// qualifierMap =
								// this.timestampQualifiersAdapter.getQualifierMapFromQVL(qvl);

								RowIdx rowIdx = sv.getId().getRowIdx();

								List<ColumnIdx> timeColumns = table.getTimeColumns();
								for (ColumnIdx columnIdx : timeColumns) {
									Integer columnYear = TimeObject.getYear(table.getColumn(columnIdx).get(rowIdx));
									if (columnYear != null) {
										TimestampPointInTime tpit = new TimestampPointInTime(columnYear);
										if (timestamps.contains(tpit)) {
											creatorSingle.addPerProperty(tid.getClassPropertyId(), new AccumInteger(1));
											added = true;
										}

									}
								}

								Map<TimestampIdentifier, Integer> tableYears = tableRepository.getMetaData(sv.getId().getTableId()).getTableYears();

								for (Entry<TimestampIdentifier, Integer> tableYear : tableYears.entrySet()) {
									TimestampPointInTime tpit = new TimestampPointInTime(tableYear.getValue());
									if (timestamps.contains(tpit)) {
										creatorSingle.addPerProperty(tid.getClassPropertyId(), new AccumInteger(1));
										added = true;
									}

								}

							}

						}

						if (ontologyRepository.getDataType(e.getKey().getPropertyId()).equals(ValueType.INSTANCE_REFERENCE)) {
							AnnotatedTableMetaData md = this.tableRepository.getMetaData(sv.getId().getTableId());
							Map<TimestampIdentifier, Integer> tableYears = tableRepository.getMetaData(sv.getId().getTableId()).getTableYears();
							Map<ClassId, Set<InstanceId>> instancesMap = new HashMap<>();
							for (ClassId instancesSet : ontologyRepository.getRange(e.getKey().getClassPropertyId())) {
								instancesMap.put(instancesSet, knowledgeBase.getInstances(instancesSet).keySet());
								// instancesMap.put();
							}
							if (!added) {
								int asd = 123;
							}
						}
					}
				}
			}
		} , "OverlapProfiler");

		TimeStampsOverlapWriter writer = new TimeStampsOverlapWriter();

		for (Entry<ClassPropertyId, AccumInteger> e : creatorSingle.get().getPerProperty().entrySet()) {
			ValueType cdt = ontologyRepository.getDataType(e.getKey().getPropertyId());
			String[] prefix = ResultsCategory.getClassPropertyIdPrefix("PROPERTY", ontologyRepository.isTimeDependent(e.getKey().getPropertyId()), cdt,
					e.getKey());
			writer.write(ResultsCategory.CLASS_PROPERTY_ID, prefix, e.getValue());
		}

		try {
			String output = writer.getAll(ResultsCategory.CLASS_PROPERTY_ID);
			if (outputFile != null) {
				FileUtils.writeStringToFile(new File(outputFile), output);
			}

			System.out.println(output);
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
	}

}
