package de.mannheim.uni.expansion.tablefusion.services;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTablesReaderData;
import de.mannheim.uni.expansion.data.tables.model.InstanceIndexComposite;
import de.mannheim.uni.expansion.data.tables.reader.AnnotatedTableFilter;
import de.mannheim.uni.expansion.utils.OtsuThresholderRounded;
import de.mannheim.uni.expansion.utils.Scorable;

public class AnnotatedTableOtsuFilter implements AnnotatedTableFilter {

	private final OtsuThresholderRounded thresholder;
	private double min = -0.1;
	private final AnnotatedTableInstanceScorer scorer;

	@Inject
	public AnnotatedTableOtsuFilter(OtsuThresholderRounded thresholder, AnnotatedTableInstanceScorer scorer) {
		this.thresholder = thresholder;
		this.scorer = scorer;
	}

	// public AnnotatedTableOtsuFilter(OtsuThresholderRounded thresholder,
	// KnowledgeBase kb, SimilarityCalculator sc,
	// ColumnMatchesRepository columnMatchesRepository, double min) {
	// this.thresholder = thresholder;
	// this.kb = kb;
	// this.sc = sc;
	// this.columnMatchesRepository = columnMatchesRepository;
	// this.min = min;
	// }

	public void setMin(double min) {
		this.min = min;
	}

	@Override
	public Optional<AnnotatedTablesReaderData> filter(AnnotatedTablesReaderData data) {

		Map<Integer, List<InstanceId>> instances = new HashMap<>();
		Map<InstanceIndexComposite, Double> scores = new HashMap<>();

		List<MappingContainer> list = new ArrayList<>();

		for (Entry<InstanceIndexComposite, Double> e : scorer.getScores(data).entrySet()) {

			list.add(new MappingContainer(e.getValue(), e.getKey()));

		}
		List<MappingContainer> thresholded = thresholder.thresholdDefault(list);

		instances = new HashMap<>();
		scores = new HashMap<>();

		for (MappingContainer mc : thresholded) {
			if (mc.getScore() > min) {
				List<InstanceId> instanceList = new ArrayList<>();
				instanceList.add(mc.getIndex().getInstanceId());
				instances.put(mc.getIndex().getIndex(), instanceList);
				scores.put(mc.getIndex(), mc.getScore());
			}
		}
		AnnotatedTableMetaData md = data.getMetaData();
		return Optional.of(new AnnotatedTablesReaderData(data.getId(),
				new AnnotatedTableMetaData(md.getSource(), md.getDbPediaClass(), md.getClassConf(), md.getNumHeaderRows(), md.getNumberOfRows(),
						md.getAllProperties(), md.getAllPropertyScores(), instances, scores, md.getKeyColumn(), md.getYearsBeforeText(), md.getYearsAfterText(),
						md.getYearsInContextBefore(), md.getYearsInTableTitle(), md.getYearsInPageTitle(), md.getPropertiesYears(), md.getPageTitle(),
						md.getTableTitle(), md.getTableContextTimeStampBeforeTable(), md.getTextBeforeTable(), md.getTextAfterTable()),
				data.getTableData()));
	}

	public static class MappingContainer implements Scorable {
		private final Double score;

		private final InstanceIndexComposite index;

		public MappingContainer(Double score, InstanceIndexComposite index) {
			super();
			this.score = score;
			this.index = index;
		}

		@Override
		public Double getScore() {
			return score;
		}

		public InstanceIndexComposite getIndex() {
			return index;
		}

	}

}
