package de.mannheim.uni.expansion.tablefusion.services;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ListType;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.evaluation.FilteredValues;

public class GroupedValuesFilterer {

	private final KnowledgeBase dBpediaRepository;
	private final OntologyRepository ontologyRepository;

	@Inject
	public GroupedValuesFilterer(KnowledgeBase dBpediaRepository, OntologyRepository ontologyRepository) {
		this.dBpediaRepository = dBpediaRepository;
		this.ontologyRepository = ontologyRepository;
	}

	public GroupedValues getNonListTypes(GroupedValues groupedValues) {
		Map<TripleId, MatchedValues> filtered = groupedValues.entrySet().stream().filter((e) -> {
			return !ontologyRepository.isListType(e.getKey().getClassPropertyId().getPropertyId());
		}).collect(Collectors.toMap(Entry::getKey, Entry::getValue));

		return new GroupedValues(filtered);
	}

	public GroupedValues getListTypes(GroupedValues groupedValues) {
		Map<TripleId, MatchedValues> filtered = groupedValues.entrySet().stream().filter((e) -> {
			return ontologyRepository.isListType(e.getKey().getClassPropertyId().getPropertyId());
		}).collect(Collectors.toMap(Entry::getKey, Entry::getValue));

		return new GroupedValues(filtered);
	}

	public GroupedValues getListTypes(ListType listType, GroupedValues groupedValues) {
		Map<TripleId, MatchedValues> filtered = groupedValues.entrySet().stream().filter((e) -> {
			return ontologyRepository.isListType(listType, e.getKey().getClassPropertyId().getPropertyId());
		}).collect(Collectors.toMap(Entry::getKey, Entry::getValue));

		return new GroupedValues(filtered);
	}

	public GroupedValues getDataType(ValueType cdt, GroupedValues groupedValues) {
		Map<TripleId, MatchedValues> filtered = groupedValues.entrySet().stream().filter((e) -> {
			return ontologyRepository.getDataType(e.getKey().getClassPropertyId().getPropertyId()) == cdt;
		}).collect(Collectors.toMap(Entry::getKey, Entry::getValue));

		return new GroupedValues(filtered);
	}

	public GroupedValues getByProperty(Set<PropertyId> set, GroupedValues groupedValues) {
		Map<TripleId, MatchedValues> filtered = groupedValues.entrySet().stream().filter((e) -> {
			return set.contains(e.getKey().getPropertyId());

		}).collect(Collectors.toMap(Entry::getKey, Entry::getValue));

		return new GroupedValues(filtered);
	}

	public GroupedValues filterBySetFilter(FilteredValues filter, GroupedValues groupedValues) {

		Map<TripleId, MatchedValues> ret = new HashMap<>();

		for (Entry<TripleId, MatchedValues> e : groupedValues.entrySet()) {

			Set<AbstractObject> filteredValues = filter.getFilteredValues(e.getKey());

			if (filteredValues != null) {

				List<SingleValue> newList = new ArrayList<>();

				for (SingleValue sv : e.getValue().getValues()) {
					if (sv.getValue() instanceof List<?>) {
						@SuppressWarnings("unchecked")
						List<AbstractObject> svList = new ArrayList<>((List<AbstractObject>) sv.getValue());

						for (Iterator<AbstractObject> it = svList.iterator(); it.hasNext();) {
							if (filteredValues.contains(it.next())) {
								it.remove();
							}
						}

						if (!svList.isEmpty()) {
							// FIXME: THIS IS NOT RIGHT
							// newList.add(new SingleValue(sv.getId(), svList));
						}
					} else {
						if (!filteredValues.contains(sv.getValue())) {
							newList.add(sv);
						}
					}
				}

				ret.put(e.getKey(), new MatchedValues(newList, e.getValue().getDataType()));
			}
		}

		return new GroupedValues(ret);
	}

}
