package de.mannheim.uni.expansion.tablefusion.services;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.matching.SynchronizedMatchedValuesMapMerger;
import de.mannheim.uni.expansion.tablefusion.core.scoring.TimeStampsScoringCalculator;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.utils.PersistentParallel3;

public class ScoresFilterer {

	private final TimeStampsScoringCalculator timeStampsScoringCalculator;

	@Inject
	public ScoresFilterer(TimeStampsScoringCalculator timeStampsScoringCalculator) {
		super();
		this.timeStampsScoringCalculator = timeStampsScoringCalculator;
	}

	public GroupedValues filter(GroupedValues groupedValues, ScoresContainer scores) {

		SynchronizedMatchedValuesMapMerger<TripleId> mergers = new SynchronizedMatchedValuesMapMerger<>();

		PersistentParallel3.runAndFinish(groupedValues.entrySet(), (gv) -> {

			for (SingleValue sv : gv.getValue().getValues()) {
				if (scores.containsKey(gv.getKey().getClassId(), gv.getKey().getPropertyId(), gv.getKey().getInstanceId(), sv.getId())) {
					mergers.merge(gv.getKey(), sv, gv.getValue().getDataType());
				}
			}

		} , "FilterGroupedValuesByScore");

		return new GroupedValues(mergers.getCleanMap());
	}

	public GroupedValues filterByHiearicalTimeStamps(GroupedValues groupedValues) {
		return filter(groupedValues, timeStampsScoringCalculator.genericYearScorer(groupedValues, timeStampsScoringCalculator.GET_HIERARICAL_YEAR));
	}

	public GroupedValues filterByColumnTimeStamps(GroupedValues groupedValues) {
		return filter(groupedValues, timeStampsScoringCalculator.genericYearScorer(groupedValues, timeStampsScoringCalculator.MAX_COLUMN_YEAR));
	}
}
