package de.mannheim.uni.expansion.tablefusion.services.fusion;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import de.mannheim.uni.expansion.tablefusion.core.fusing.Fuser;
import de.mannheim.uni.tableprocessor.model.TableColumn.ColumnDataType;

public class FuserOptionsBuilder {

	final private Map<ColumnDataType, Fuser> fusers;
	final private Fuser defaultFuser;

	public Fuser getDefaultFuser() {
		return defaultFuser;
	}

	public Map<ColumnDataType, Fuser> getFusers() {
		return Collections.unmodifiableMap(fusers);
	}

	public FuserOptionsBuilder(Fuser defaultFuser) {
		this.defaultFuser = defaultFuser;
		this.fusers = new HashMap<ColumnDataType, Fuser>();
	}

	public FuserOptionsBuilder(Fuser fuser, ColumnDataType... columnTypes) {
		defaultFuser = null;
		this.fusers = new HashMap<ColumnDataType, Fuser>();

		for (ColumnDataType cdt : columnTypes) {
			this.fusers.put(cdt, fuser);
		}
	}

	private FuserOptionsBuilder(Fuser defaultFuser, Map<ColumnDataType, Fuser> options, Fuser fuser, ColumnDataType[] columnTypes) {
		this.fusers = new HashMap<ColumnDataType, Fuser>(options);
		for (ColumnDataType cdt : columnTypes) {
			this.fusers.put(cdt, fuser);
		}

		this.defaultFuser = defaultFuser;
	}

	@SuppressWarnings("unused")
	private FuserOptionsBuilder(Fuser defaultFuser, Map<ColumnDataType, Fuser> options) {
		this.fusers = new HashMap<ColumnDataType, Fuser>(options);

		this.defaultFuser = defaultFuser;
	}

	public FuserOptionsBuilder() {
		defaultFuser = null;
		fusers = new HashMap<ColumnDataType, Fuser>();
	}

	public FuserOptionsBuilder withNewFuser(Fuser fuser, ColumnDataType... columnTypes) {

		return new FuserOptionsBuilder(getDefaultFuser(), getFusers(), fuser, columnTypes);
	}

}
