package de.mannheim.uni.expansion.tablefusion.services.fusion;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import de.mannheim.uni.expansion.tablefusion.services.fusion.runner.FusionStrategy;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class NewFuserOptionsBuilder {

	final private Map<ValueType, FusionStrategy> fusionStrategies;
	final private FusionStrategy defaultFusionStrategy;

	public FusionStrategy getDefaultFusionStrategy() {
		return defaultFusionStrategy;
	}

	public Map<ValueType, FusionStrategy> getFusionStrategies() {
		return Collections.unmodifiableMap(fusionStrategies);
	}

	public FusionStrategy getFusionStrategy(ValueType cdt) {
		return fusionStrategies.get(cdt);
	}

	public NewFuserOptionsBuilder(FusionStrategy fusionStrategy, ValueType... columnTypes) {
		defaultFusionStrategy = null;
		this.fusionStrategies = new HashMap<ValueType, FusionStrategy>();

		for (ValueType cdt : columnTypes) {
			this.fusionStrategies.put(cdt, fusionStrategy);
		}
	}

	public NewFuserOptionsBuilder(FusionStrategy defaultFusionStrategy) {
		this.defaultFusionStrategy = defaultFusionStrategy;
		this.fusionStrategies = new HashMap<ValueType, FusionStrategy>();
	}

	private NewFuserOptionsBuilder(FusionStrategy defaultFusionStrategy, Map<ValueType, FusionStrategy> options, FusionStrategy FusionStrategy,
			boolean fuseLists, ValueType[] columnTypes) {
		this.fusionStrategies = new HashMap<ValueType, FusionStrategy>(options);
		for (ValueType cdt : columnTypes) {
			this.fusionStrategies.put(cdt, FusionStrategy);
		}

		this.defaultFusionStrategy = defaultFusionStrategy;
	}

	@SuppressWarnings("unused")
	private NewFuserOptionsBuilder(FusionStrategy defaultFusionStrategy, Map<ValueType, FusionStrategy> options) {
		this.fusionStrategies = new HashMap<ValueType, FusionStrategy>(options);

		this.defaultFusionStrategy = defaultFusionStrategy;
	}

	public NewFuserOptionsBuilder() {
		defaultFusionStrategy = null;
		fusionStrategies = new HashMap<ValueType, FusionStrategy>();
	}

	public NewFuserOptionsBuilder withNewFusionStrategy(FusionStrategy FusionStrategy, boolean fuseLists, ValueType... columnTypes) {

		return new NewFuserOptionsBuilder(getDefaultFusionStrategy(), getFusionStrategies(), FusionStrategy, fuseLists, columnTypes);
	}

}
