package de.mannheim.uni.expansion.tablefusion.services.fusion;

import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleIdContainer;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.fusing.Fuser;
import de.mannheim.uni.expansion.tablefusion.core.fusing.FuserRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.AbstractGroupledValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.TableScores;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.TableScoresQualifierDecorator;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfigFilter;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdContainer;

public class NewFuserService {

	private final KnowledgeBase dbPediaRepositoryNew;

	private final OntologyRepository ontologyRepository;

	// private final PersistentParallel3<Entry<TripleId, MatchedValues>>
	// parallel;

	private final FuserRepository fusers;
	private final PropertiesConfigFilter propertiesConfigFilter;
	private final AnnotatedTablesRepository tableRepository;

	@Inject
	public NewFuserService(KnowledgeBase dbPediaRepositoryNew, OntologyRepository ontologyRepository, FuserRepository fusers,
			PropertiesConfigFilter propertiesConfigFilter, AnnotatedTablesRepository tableRepository) {
		this.dbPediaRepositoryNew = dbPediaRepositoryNew;
		this.ontologyRepository = ontologyRepository;
		// this.parallel = new PersistentParallel3<Entry<TripleId,
		// MatchedValues>>("BasicFuserService");
		this.fusers = fusers;
		this.propertiesConfigFilter = propertiesConfigFilter;
		this.tableRepository = tableRepository;
	}

	// public Map<TripleId, Object> fuse(Map<TripleId, MatchedValues>
	// groupedValues) {
	// Map<TripleId, Object> map = new HashMap<TripleId, Object>();
	//
	// for (Entry<TripleId, MatchedValues> entry : groupedValues.entrySet()) {
	// TripleId mid = entry.getKey();
	// ColumnDataType cdt = propertyDataTypes.get(mid.getPropertyId());
	// if (cdt.equals(ColumnDataType.numeric)) {
	// map.put(mid, averageFuser.fuse(entry.getValue()));
	//
	// } else {
	// map.put(mid, majorityFuser.fuse(entry.getValue()));
	// }
	// }
	//
	// return map;
	//
	// }

	public <E extends TripleIdContainer> Map<E, FusedValue> fuse(NewFuserOptionsBuilder options, AbstractGroupledValues<E> groupedValues,
			FusionConfig fusionConfig) {
		return fuse(options, groupedValues, null, null, fusionConfig, PropertiesConfig.defaultConfig());
	}

	public <E extends TripleIdContainer> Map<E, FusedValue> fuse(NewFuserOptionsBuilder options, AbstractGroupledValues<E> groupedValues,
			ScoresContainer probabilities, ThresholdContainer threshold, FusionConfig fusionConfig, PropertiesConfig propertiesConfig) {
		ConcurrentHashMap<E, FusedValue> map = new ConcurrentHashMap<E, FusedValue>();

		FusedReceiver<E> receiver = (entry, fusedValue) -> {
			if (fusedValue != null) {
				map.put(entry.getKey(), fusedValue);
			} else {
				int sdf = 123;
			}
		};

		fuse(options, groupedValues, probabilities, threshold, fusionConfig, receiver, propertiesConfig);

		return map;
	}

	public String getPLD(AnnotatedTableId tableId) {
		return tableRepository.getMetaData(tableId).getSource().getPLD();
	}

	public synchronized <E extends TripleIdContainer> void fuse(NewFuserOptionsBuilder options, AbstractGroupledValues<E> groupedValues,
			ScoresContainer probabilities, ThresholdContainer thresholdContainer, FusionConfig fusionConfig, FusedReceiver<E> receiver,
			PropertiesConfig propertiesConfig) {

		try {
			PersistentParallel4.runAndFinish(groupedValues.entrySet(), (entry) -> {
				if (!this.propertiesConfigFilter.include(propertiesConfig, entry.getKey()))
					return;

				E e = entry.getKey();
				TripleId tid;
				if (e instanceof QualifiedTripleId) {
					tid = ((QualifiedTripleId) e).getTripleId();
				} else {
					tid = (TripleId) e;
				}

				ValueType cdt = ontologyRepository.getDataType(tid.getPropertyId());

				Fuser fuser = fusers.get(cdt, options);

				if (fuser == null) {
					return;
				}

				Double threshold = 0.0;

				if (thresholdContainer != null) {
					threshold = thresholdContainer.getScore(tid.getClassId(), tid.getPropertyId(), tid.getInstanceId());
					if (threshold == null) {
						threshold = 0.0;
					}
				}
				FusedValue fusedValue;
				if (fusionConfig.isFuseLists() && ontologyRepository.isListType(tid.getPropertyId())) {

					if (probabilities != null) {
						TableScores tableProperties = probabilities.getTableScores(tid.getClassId(), tid.getPropertyId(), tid.getInstanceId());
						if (tableProperties != null)
							fusedValue = fuser.fuseList(entry.getValue(), tableProperties, threshold);
						else
							return;
					} else {

						fusedValue = fuser.fuseList(entry.getValue(), null, 0.0);
					}
				} else if (probabilities != null) {
					TableScores tableProperties = probabilities.getTableScores(tid.getClassId(), tid.getPropertyId(), tid.getInstanceId());
					if (e instanceof QualifiedTripleId) {
						tableProperties = new TableScoresQualifierDecorator(tableProperties, ((QualifiedTripleId) e).getQualifiers());
					}

					if (tableProperties != null)
						fusedValue = fuser.fuse(entry.getValue(), tableProperties, threshold, fusionConfig.getFusionScoreThresholdAdapter(),
								fusionConfig.isOnePerPLD(), this::getPLD);
					else
						return;

				} else {
					throw new IllegalStateException();
					// fusedValue = fuser.fuse(entry.getValue(), null, 0.0);

				}

				if (fusedValue != null) {
					receiver.doWork(entry, fusedValue);
				}

			} , "FUSER SERVICE");

		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	@FunctionalInterface
	public interface FusedReceiver<E> {
		void doWork(Entry<E, MatchedValues> entry, FusedValue fusedValue);
	}
}
