package de.mannheim.uni.expansion.tablefusion.services.fusion.runner;

import java.util.EnumMap;
import java.util.Map;

import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public enum FusionDataType {
	REFERENCE("Reference", ValueType.INSTANCE_REFERENCE, FusionBaseMethod.MAJORTIY), //
	STRING("String", ValueType.STRING, FusionBaseMethod.MAJORTIY), //
	NUMERIC("Numeric", ValueType.QUANTITY, FusionBaseMethod.NUMERIC_MAJORITY_MEDIAN), //
	CATEGORICAL_INT("Numeric", ValueType.CATEGORICAL_INT, FusionBaseMethod.MAJORTIY), //
	DATE("Date", ValueType.TIME, FusionBaseMethod.DATE_MEDIAN), //
	LINK("Link", ValueType.URL, FusionBaseMethod.MAJORTIY),

	COORDINATE("Coordinate", ValueType.COORDINATE, FusionBaseMethod.COORDINATE_MEDIAN);

	private final String name;
	private final FusionBaseMethod baseMethod;
	private final ValueType cdt;

	private FusionDataType(String name, ValueType cdt, FusionBaseMethod baseMethod) {
		this.name = name;
		this.baseMethod = baseMethod;
		this.cdt = cdt;
	}

	public String getName() {
		return name;
	}

	public FusionBaseMethod getBaseMethod() {
		return baseMethod;
	}

	public ValueType getDataType() {
		return cdt;
	}

	static private final Map<ValueType, FusionDataType> typeMapping = new EnumMap<>(ValueType.class);

	static {
		for (FusionDataType fdt : FusionDataType.values()) {
			if (!typeMapping.containsKey(fdt.getDataType())) {
				typeMapping.put(fdt.getDataType(), fdt);
			} else {
				throw new IllegalStateException("ONE COLUMN DATA TYPE CAN AT MOST BE MAPPED TO ONE FUSIONDATATYPE");
			}
		}
	}

	public static FusionDataType get(ValueType cdt) {
		return typeMapping.get(cdt);
	}

}
