package de.mannheim.uni.expansion.tablefusion.services.fusion.runner;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.TripleIdContainer;
import de.mannheim.uni.expansion.tablefusion.core.matching.AbstractGroupledValues;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionRun;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfig;
import de.mannheim.uni.expansion.tablefusion.services.fusion.NewFuserOptionsBuilder;
import de.mannheim.uni.expansion.tablefusion.services.fusion.NewFuserService;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.informatik.dws.expansion.model.utils.thresholdpacks.ThresholdContainer;

public class FusionRunner {

	private final NewFuserService basicFuserService;

	private final FusionProprocessersImpl fusionPreprocessors;

	@Inject
	public FusionRunner(NewFuserService basicFuserService, FusionProprocessersImpl fusionPreprocessors) {
		this.basicFuserService = basicFuserService;
		this.fusionPreprocessors = fusionPreprocessors;

	}

	public <E extends TripleIdContainer> Map<E, FusedValue> run(FusionDataType fusionDataType, FusionPreproccesser fusionPreProcessor,
			FusionStrategy fusionRunType, AbstractGroupledValues<E> values, ScoresContainer probabilities, ThresholdContainer probabilityThreshold,
			FusionConfig fusionConfig, PropertiesConfig propertiesConfig) {

		NewFuserOptionsBuilder fuserOptionsBuilder = new NewFuserOptionsBuilder(fusionRunType, fusionDataType.getDataType());

		values = fusionPreprocessors.preprocessValues(fusionPreProcessor, values, probabilities);

		Map<E, FusedValue> fusedValues = basicFuserService.fuse(fuserOptionsBuilder, values, probabilities, probabilityThreshold, fusionConfig,
				propertiesConfig);
		// System.out.println(values.size());
		// System.out.println(fusedValues.size());

		return fusedValues;

	}

	public <E extends TripleIdContainer> Map<FusionRun, Map<E, FusedValue>> runAll(FusionStrategy fusionRunType, AbstractGroupledValues<E> values,
			ScoresContainer probabilities, FusionConfig fusionConfig, ThresholdContainer probabilityThreshold, PropertiesConfig propertiesConfig) {
		return run(Arrays.asList(FusionDataType.values()), Arrays.asList(FusionPreproccesser.values()), fusionRunType, values, probabilities, fusionConfig,
				probabilityThreshold, propertiesConfig);

	}

	public <E extends TripleIdContainer> Map<FusionRun, Map<E, FusedValue>> runAll(FusionStrategy fusionRunType, List<FusionPreproccesser> preprocessers,
			AbstractGroupledValues<E> values, ScoresContainer probabilities, FusionConfig fusionConfig, ThresholdContainer probabilityThreshold,
			PropertiesConfig propertiesConfig) {
		return run(Arrays.asList(FusionDataType.values()), preprocessers, fusionRunType, values, probabilities, fusionConfig, probabilityThreshold,
				propertiesConfig);

	}

	public <E extends TripleIdContainer> Map<FusionRun, Map<E, FusedValue>> run(List<FusionDataType> dataTypes, List<FusionPreproccesser> preprocessers,
			FusionStrategy fusionRunType, AbstractGroupledValues<E> values, ScoresContainer probabilities, ThresholdContainer probabilityThreshold,
			FusionConfig fusionConfig, PropertiesConfig propertiesConfig) {
		Map<FusionRun, Map<E, FusedValue>> fusedResultsMap = new HashMap<>();
		for (FusionDataType fusionDataType : dataTypes) {
			for (FusionPreproccesser fusionPreproccesser : preprocessers) {
				FusionRun fusionRun = new FusionRun(fusionDataType, fusionPreproccesser, fusionRunType);
				fusedResultsMap.put(fusionRun,
						run(fusionDataType, fusionPreproccesser, fusionRunType, values, probabilities, probabilityThreshold, fusionConfig, propertiesConfig));
			}

		}

		return fusedResultsMap;
	}

	public <E extends TripleIdContainer> Map<FusionRun, Map<E, FusedValue>> run(List<FusionDataType> dataTypes, List<FusionPreproccesser> preprocessers,
			FusionStrategy fusionRunType, AbstractGroupledValues<E> values, ScoresContainer probabilities, FusionConfig fusionConfig,
			ThresholdContainer probabilityThreshold, PropertiesConfig propertiesConfig) {
		return run(dataTypes, preprocessers, fusionRunType, values, probabilities, probabilityThreshold, fusionConfig, propertiesConfig);
	}

	public <E extends TripleIdContainer> Map<FusionRun, Map<E, FusedValue>> run(List<FusionDataType> dataTypes, FusionPreproccesser fusionPreproccesser,
			FusionStrategy fusionRunType, AbstractGroupledValues<E> values, ScoresContainer probabilities, FusionConfig fusionConfig,
			ThresholdContainer probabilityThreshold, PropertiesConfig propertiesConfig) {
		ArrayList<FusionPreproccesser> fusionPreprocessors = new ArrayList<>();
		fusionPreprocessors.add(fusionPreproccesser);
		return run(Arrays.asList(FusionDataType.values()), fusionPreprocessors, fusionRunType, values, probabilities, fusionConfig, probabilityThreshold,
				propertiesConfig);

	}

	public <E extends TripleIdContainer> Map<E, FusedValue> run(FusionDataType fusionDataType, FusionPreproccesser fusionPreProcessor,
			FusionStrategy fusionRunType, AbstractGroupledValues<E> values, FusionConfig fusionConfig) {
		NewFuserOptionsBuilder fuserOptionsBuilder = new NewFuserOptionsBuilder(fusionRunType, fusionDataType.getDataType());

		values = fusionPreprocessors.preprocessValues(fusionPreProcessor, values);

		Map<E, FusedValue> fusedValues = basicFuserService.fuse(fuserOptionsBuilder, values, fusionConfig);
		// FusionResults fr = accuracyCalculator.analyzeAccuracy(
		// fusionDataType.getName() + "::" + fusionDataType.getMethod() + "::" +
		// fusionPreProcessor.getName() + "::" + fusionRunType.getName(),
		// fusedValues, false);
		// accuracyStorer.addResult(fusionDataType, fusionPreProcessor,
		// fusionRunType, fr);

		return fusedValues;
	}

	public <E extends TripleIdContainer> Map<FusionRun, Map<E, FusedValue>> run(List<FusionDataType> dataTypes, List<FusionPreproccesser> preprocessers,
			FusionStrategy fusionRunType, AbstractGroupledValues<E> values, FusionConfig fusionConfig) {

		Map<FusionRun, Map<E, FusedValue>> fusedResultsMap = new HashMap<>();
		for (FusionDataType fusionDataType : dataTypes) {

			for (FusionPreproccesser fusionPreproccesser : preprocessers) {
				FusionRun fusionRun = new FusionRun(fusionDataType, fusionPreproccesser, fusionRunType);
				fusedResultsMap.put(fusionRun, run(fusionDataType, fusionPreproccesser, fusionRunType, values, fusionConfig));
			}

		}

		return fusedResultsMap;

	}

	public <E extends TripleIdContainer> Map<FusionRun, Map<E, FusedValue>> runAll(FusionStrategy fusionRunType, AbstractGroupledValues<E> values,
			FusionConfig fusionConfig) {
		return run(Arrays.asList(FusionDataType.values()), Arrays.asList(FusionPreproccesser.values()), fusionRunType, values, fusionConfig);
	}

	public <E extends TripleIdContainer> Map<FusionRun, Map<E, FusedValue>> run(List<FusionDataType> methodTypes, FusionPreproccesser fusionPreproccesser,
			FusionStrategy normal, AbstractGroupledValues<E> values, FusionConfig fusionConfig, boolean fuseLists) {

		ArrayList<FusionPreproccesser> fusionPreprocessors = new ArrayList<>();
		fusionPreprocessors.add(fusionPreproccesser);
		return run(methodTypes, fusionPreprocessors, normal, values, fusionConfig);

	}

}
