package de.mannheim.uni.expansion.tablefusion.services.fusion.runner;

import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitterFactory;
import de.mannheim.uni.expansion.tablefusion.optimization.ThresholdType;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer;

public class FusionStrategy {

	final private FusionStrategyId fusionStrategyId;
	final private FusionMethod fusionMethod;
	final private Class<? extends Scorer> scoresGetter;
	final private Class<? extends SplitterFactory> splitterFactory;
	final private TestingMethod testingMethod;

	final private ThresholdType thresholdType;

	public FusionStrategy(String name, FusionMethod fusionMetehod, TestingMethod testingMethod) {
		super();
		this.fusionStrategyId = new FusionStrategyId(name);
		this.fusionMethod = fusionMetehod;
		this.testingMethod = testingMethod;
		this.scoresGetter = null;
		this.thresholdType = ThresholdType.General;
		this.splitterFactory = null;
	}

	public FusionStrategy(String name, FusionMethod fusionMetehod, TestingMethod testingMethod, Class<? extends Scorer> scoringMethod) {
		super();
		this.fusionStrategyId = new FusionStrategyId(name);
		this.fusionMethod = fusionMetehod;
		this.testingMethod = testingMethod;
		this.scoresGetter = scoringMethod;
		this.thresholdType = ThresholdType.General;
		this.splitterFactory = null;
	}

	public FusionStrategy(String name, FusionMethod fusionMetehod, TestingMethod testingMethod, Class<? extends Scorer> scoringMethod,
			ThresholdType thresholdType) {
		super();
		this.fusionStrategyId = new FusionStrategyId(name);
		this.fusionMethod = fusionMetehod;
		this.testingMethod = testingMethod;
		this.scoresGetter = scoringMethod;
		this.thresholdType = thresholdType;
		this.splitterFactory = null;
	}

	public FusionStrategy(String name, FusionMethod fusionMetehod, TestingMethod testingMethod, Class<? extends Scorer> scoringMethod,
			ThresholdType thresholdType, Class<? extends SplitterFactory> splitterFactory) {
		super();
		this.fusionStrategyId = new FusionStrategyId(name);
		this.fusionMethod = fusionMetehod;
		this.testingMethod = testingMethod;
		this.scoresGetter = scoringMethod;
		this.thresholdType = thresholdType;
		this.splitterFactory = splitterFactory;
	}

	public String getName() {
		return fusionStrategyId.getName();
	}

	public FusionStrategyId getFusionStrategyId() {
		return fusionStrategyId;
	}

	public FusionMethod getFusionMethod() {
		return fusionMethod;
	}

	public TestingMethod geTestingMethod() {
		return testingMethod;
	}

	public Class<? extends Scorer> getScoringMethod() {
		return scoresGetter;
	}

	public Class<? extends SplitterFactory> getSplitterFactory() {
		return splitterFactory;
	}

	public ThresholdType getThresholdType() {
		return thresholdType;
	}

	public static class FusionStrategyId {
		private final String name;

		private FusionStrategyId(String name) {
			super();
			this.name = name;
		}

		public String getName() {
			return name;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((name == null) ? 0 : name.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			FusionStrategyId other = (FusionStrategyId) obj;
			if (name == null) {
				if (other.name != null)
					return false;
			} else if (!name.equals(other.name))
				return false;
			return true;
		}

		@Override
		public String toString() {
			return name;
		}

	}

}
