package de.mannheim.uni.expansion.tablefusion.services.fusion.runner;

import javax.inject.Inject;

import de.mannheim.uni.expansion.tablefusion.core.splitting.NonDuplicateSplitterFactory;
import de.mannheim.uni.expansion.tablefusion.core.splitting.NonTimeDependentSplitterFactory;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SimpleSplitterFactory;
import de.mannheim.uni.expansion.tablefusion.core.splitting.SplitterFactoryRepository;
import de.mannheim.uni.expansion.tablefusion.core.splitting.TimestampSplitterFactory;
import de.mannheim.uni.expansion.tablefusion.optimization.ThresholdType;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.ColumnScores;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.ScorerRepository;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer_Baseline;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer_KBT;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer_KBT_PRPGTD;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer_QKBT;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer_QKBT_Mapped;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer_QKBT_PRPGTD;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer_Regression_PropertyId;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer_TS_KBT_Regression;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer_TS_KBT_Regression_ClassProperty;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer_TS_Propagted_KBT_Regression_Propagated;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer_TS_Propgated_KBT_Regression_ClassProperty;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.newTS.Scorer_NewTs;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.newTS.Scorer_NewTs_KBT;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.newTS.Scorer_NewTs_QKBT;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.newTS.Scorer_NewTs_QKBT_Propagated;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.newTS.Scorer_NewTs_QKBT_Propagated_without_CellRow;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.newTS.Scorer_NewTs_QKBT_without_CellRow;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.regressionTrees.Scorer_RegressionTrees;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.regressionTrees.Scorer_RegressionTrees_ClassProperty;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.regressionTrees.Scorer_RegressionTrees_PropertyId;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.Scorer_Combined_HierarchyTS_TableText;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.Scorer_KBT_Combined_HierarchyTS_TableText_Multiplied;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.Scorer_KBT_Combined_HierarchyTS_TableText_Weighted;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.Scorer_KBT_Combined_HierarchyTS_TableText_WeightedMultiplied;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.cell.Scorer_CellTS_Max;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.cell.Scorer_CellTS_Max_Propagated;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.column.Scorer_ColumnTS_Max;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.column.Scorer_ColumnTS_Max_Propagated;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.column.Scorer_KBT_ColumnTS_Max_Multiplied;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.column.Scorer_KBT_ColumnTS_Max_Resized;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.column.Scorer_KBT_ColumnTS_Max_Weighted;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.fullHierarichal.Scorer_FullHierarchyTS;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.fullHierarichal.Scorer_FullHierarchyTS_Propagted;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.fullHierarichal.Scorer_KBT_FullHierarchyTS_Multiplied;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.fullHierarichal.Scorer_KBT_FullHierarchyTS_Resized;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.fullHierarichal.Scorer_KBT_FullHierarchyTS_Weighted;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.hierarichal.Scorer_HierarchyTS;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.hierarichal.Scorer_HierarchyTS_Propagated;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.hierarichal.Scorer_KBT_HierarchyTS_Multiplied;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.hierarichal.Scorer_KBT_HierarchyTS_Resized;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.hierarichal.Scorer_KBT_HierarchyTS_Weighted;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.pageTitle.Scorer_KBT_PageTitleTS_Max_Multiplied;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.pageTitle.Scorer_KBT_PageTitleTS_Max_Resized;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.pageTitle.Scorer_KBT_PageTitleTS_Max_Weighted;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.pageTitle.Scorer_PageTitleTS_Max;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.pageTitle.Scorer_PageTitleTS_Max_Propagated;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.tableText.Scorer_KBT_TableTextTS_MostCommon_Multiplied;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.tableText.Scorer_KBT_TableTextTS_MostCommon_Resized;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.tableText.Scorer_KBT_TableTextTS_MostCommon_Weighted;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.tableText.Scorer_TableTextTS;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.tableText.Scorer_TableTextTS_Propagated;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.tableTitle.Scorer_KBT_TableTitleTS_Max_Multiplied;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.tableTitle.Scorer_KBT_TableTitleTS_Max_Resized;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.tableTitle.Scorer_KBT_TableTitleTS_Max_Weighted;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.tableTitle.Scorer_TableTitleTS_Max;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.tableTitle.Scorer_TableTitleTS_Max_Propagated;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.tableYear.Scorer_KBT_TableYearsHierarchy_Multiplied;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.tableYear.Scorer_KBT_TableYearsHierarchy_Resized;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.tableYear.Scorer_KBT_TableYearsHierarchy_Weighted;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.tableYear.Scorer_TableYearsHierarchy;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps.tableYear.Scorer_TableYearsHierarchy_Propagated;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.tsRegression.Scorer_TimeStamps_Regression;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.tsRegression.Scorer_TimeStamps_Regression_ClassProperty;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.tsRegression.Scorer_TimeStamps_Regression_ClassProperty_Propagated;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.tsRegression.Scorer_TimeStamps_Regression_Propagated;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.tsRegression.Scorer_TimeStamps_Regression_PropertyId;

public class FusionStrategyRepository {
	public static final FusionStrategy NORMAL = new FusionStrategy("Baseline", FusionMethod.SCORED_NO_THRESHOKD, TestingMethod.CROSS_VALIDATED,
			Scorer_Baseline.class);

	public static final FusionStrategy COLUMN_SCORES = new FusionStrategy("COLUMN_SCORES", FusionMethod.SCORED_NO_THRESHOKD, TestingMethod.CROSS_VALIDATED,
			ColumnScores.class);
	public static final FusionStrategy CLUSTER = new FusionStrategy("Cluster", FusionMethod.CLUSTERED, TestingMethod.DIRECT);

	public static final FusionStrategy CS_WEIGHTED = new FusionStrategy("CS-Weighted", FusionMethod.WEIGHTED_SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_KBT.class);
	public static final FusionStrategy COLUMNPROBABILTIES_CLUSTERED = new FusionStrategy("Column Probabilities", FusionMethod.SCORED_CLUSTERED,
			TestingMethod.CROSS_VALIDATED);
	public static final FusionStrategy TABLEPROBABILTIES = new FusionStrategy("Table Probabilities", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED);
	public static final FusionStrategy PROPERTY_TABLE_PROBABILTIES = new FusionStrategy("Property Table Probabilities", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED);
	public static final FusionStrategy HOST_PROBABILITIES = new FusionStrategy("Host Probabilities", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED);
	public static final FusionStrategy HOST_PROPERTY_PROBABILITIES = new FusionStrategy("Host & Property Probabilities", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED);
	public static final FusionStrategy HOST_PROPERTY_INSTANCE_PROBABILITIES = new FusionStrategy("Host, Property & Instance Probabilities", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED);
	public static final FusionStrategy PR_PROPERTY_PROBABILITIES = new FusionStrategy("PageRank & Property Probabilities", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED);
	public static final FusionStrategy PR_PROBABILITIES = new FusionStrategy("PageRank", FusionMethod.SCORED, TestingMethod.DIRECT);
	// public static final FusionStrategy REGRESSION_GENERAL = new
	// FusionStrategy("Regression General", FusionMethod.SCORED,
	// TestingMethod.CROSS_VALIDATED);
	// public static final FusionStrategy REGRESSION_BY_CLASS_PROPERTY = new
	// FusionStrategy("Regression by Class Property", FusionMethod.SCORED,
	// TestingMethod.CROSS_VALIDATED);
	public static final FusionStrategy RANDOM = new FusionStrategy("RANDOM", FusionMethod.RANDOM, TestingMethod.DIRECT);
	public static final FusionStrategy CS_LIST_EXPANSION = new FusionStrategy("ListExpansion-CS", FusionMethod.SCORED,
			TestingMethod.LIST_EXPANSION_CROSS_VALIDATED, Scorer_KBT.class);

	//

	public static final FusionStrategy KBT_NO_THRESHOLD = new FusionStrategy("KBT", FusionMethod.SCORED_NO_THRESHOKD, TestingMethod.CROSS_VALIDATED,
			Scorer_KBT.class);

	public static final FusionStrategy KBT = new FusionStrategy("KBT", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED, Scorer_KBT.class);

	@Deprecated
	public static final FusionStrategy QKBT = new FusionStrategy("QKBT", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED, Scorer_QKBT.class);

	public static final FusionStrategy KBT_ClassPropertyId_Threshold = new FusionStrategy("KBT_ATT", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_KBT.class, ThresholdType.ClassPropertyId);

	public static final FusionStrategy KBT_PRPGTD_ClassPropertyId_Threshold = new FusionStrategy("KBT_PRPGTD_ATT", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_PRPGTD.class, ThresholdType.ClassPropertyId);

	public static final FusionStrategy QKBT_ClassPropertyId_Threshold = new FusionStrategy("Q_KBT_ATT", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_QKBT.class, ThresholdType.ClassPropertyId);

	@Deprecated
	public static final FusionStrategy QKBT_Propogated_ClassPropertyId_Threshold = new FusionStrategy("Q_KBT_PRPGTD_ATT", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_QKBT_PRPGTD.class, ThresholdType.ClassPropertyId);

	public static final FusionStrategy KBT_PropertyId_Threshold = new FusionStrategy("KnowledgeBasedTrustPropertyID", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT.class, ThresholdType.PropertyId);

	//

	//

	public static final FusionStrategy QKBT_Mapped = new FusionStrategy("Q_KBT_MAPPED", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_QKBT_Mapped.class, ThresholdType.ClassPropertyId);

	@Inject
	Scorer_QKBT_Mapped qkbtMapped;

	@Inject
	Scorer_KBT a;

	@Inject
	Scorer_KBT_PRPGTD kbtPropgated;

	@Inject
	Scorer_QKBT qa;

	@Inject
	Scorer_QKBT_PRPGTD qkbtPropgated;

	public static final FusionStrategy ColumnTS = new FusionStrategy("ColumnTS", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED, Scorer_ColumnTS_Max.class);
	public static final FusionStrategy ColumnTS_Propagated = new FusionStrategy("ColumnTS_Propagated", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_ColumnTS_Max_Propagated.class);

	public static final FusionStrategy KBT_ColumnTS_Max_Weighted = new FusionStrategy("KBT_ColumnTS_Max_Weighted", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_ColumnTS_Max_Weighted.class);
	public static final FusionStrategy KBT_ColumnTS_Max_Resized = new FusionStrategy("KBT_ColumnTS_Max_Resized", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_ColumnTS_Max_Resized.class);
	public static final FusionStrategy KBT_ColumnTS_Max_Multiplied = new FusionStrategy("KBT_ColumnTS_Max_Multiplied", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_ColumnTS_Max_Multiplied.class);

	@Inject
	Scorer_ColumnTS_Max columnTs;

	@Inject
	Scorer_ColumnTS_Max_Propagated columnTs_Propagated;
	@Inject
	Scorer_KBT_ColumnTS_Max_Weighted b1;
	@Inject
	Scorer_KBT_ColumnTS_Max_Resized b2;
	@Inject
	Scorer_KBT_ColumnTS_Max_Multiplied b3;

	@Inject
	Scorer_CellTS_Max cellTS;

	@Inject
	Scorer_CellTS_Max_Propagated cellTS_Propagated;

	public static final FusionStrategy CellTS = new FusionStrategy("CellTS", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED, Scorer_CellTS_Max.class);
	public static final FusionStrategy CellTS_Propagated = new FusionStrategy("CellTS_Propagated", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_CellTS_Max_Propagated.class);

	public static final FusionStrategy TableTextTS = new FusionStrategy("TableTextTS", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_TableTextTS.class);
	public static final FusionStrategy TableTextTS_Propagated = new FusionStrategy("TableTextTS_Propagated", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_TableTextTS_Propagated.class);
	public static final FusionStrategy KBT_TableTextTS_MostCommon_Weighted = new FusionStrategy("KBT_TableTextTS_MostCommon_Weighted", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_TableTextTS_MostCommon_Weighted.class);
	public static final FusionStrategy KBT_TableTextTS_MostCommon_Resized = new FusionStrategy("KBT_TableTextTS_MostCommon_Resized", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_TableTextTS_MostCommon_Resized.class);
	public static final FusionStrategy KBT_TableTextTS_MostCommon_Multiplied = new FusionStrategy("KBT_TableTextTS_MostCommon_Multiplied", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_TableTextTS_MostCommon_Multiplied.class);

	@Inject
	Scorer_TableTextTS c2;

	@Inject
	Scorer_TableTextTS_Propagated c2_Propagated;
	@Inject
	Scorer_KBT_TableTextTS_MostCommon_Weighted c;
	@Inject
	Scorer_KBT_TableTextTS_MostCommon_Resized d;
	@Inject
	Scorer_KBT_TableTextTS_MostCommon_Multiplied d2;

	public static final FusionStrategy HierarchyTS = new FusionStrategy("HierarchyTS", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_HierarchyTS.class);
	public static final FusionStrategy HierarchyTS_Propagated = new FusionStrategy("HierarchyTS_Propagated", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_HierarchyTS_Propagated.class);
	public static final FusionStrategy KBT_HierarchyTS_Weighted = new FusionStrategy("KBT_HierarchyTS_Weighted", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_HierarchyTS_Weighted.class);
	public static final FusionStrategy KBT_HierarchyTS_Resized = new FusionStrategy("KBT_HierarchyTS_Resized", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_HierarchyTS_Resized.class);
	public static final FusionStrategy KBT_HierarchyTS_Multiplied = new FusionStrategy("KBT_HierarchyTS_Multiplied", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_HierarchyTS_Multiplied.class);

	@Inject
	Scorer_HierarchyTS hierarchyTs;

	@Inject
	Scorer_HierarchyTS_Propagated hierarchyTs_Propagated;
	@Inject
	Scorer_KBT_HierarchyTS_Resized e;
	@Inject
	Scorer_KBT_HierarchyTS_Weighted f;
	@Inject
	Scorer_KBT_HierarchyTS_Multiplied f2;

	public static final FusionStrategy FullHierarchyTS = new FusionStrategy("FullHierarchyTS", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_FullHierarchyTS.class);
	public static final FusionStrategy FullHierarchyTS_Propagated = new FusionStrategy("FullHierarchyTS_Propagted", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_FullHierarchyTS_Propagted.class);
	public static final FusionStrategy KBT_FullHierarchyTS_Resized = new FusionStrategy("KBT_FullHierarchyTS_Resized", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_FullHierarchyTS_Resized.class);
	public static final FusionStrategy KBT_FullHierarchyTS_Weighted = new FusionStrategy("KBT_FullHierarchyTS_Weighted", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_FullHierarchyTS_Weighted.class);
	public static final FusionStrategy KBT_FullHierarchyTS_Multiplied = new FusionStrategy("KBT_FullHierarchyTS_Multiplied", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_FullHierarchyTS_Multiplied.class);

	@Inject
	Scorer_FullHierarchyTS fullhierarchy;

	@Inject
	Scorer_FullHierarchyTS_Propagted fullhierarchy_Propagated;
	@Inject
	Scorer_KBT_FullHierarchyTS_Resized fullResized;
	@Inject
	Scorer_KBT_FullHierarchyTS_Weighted fullWeighted;
	@Inject
	Scorer_KBT_FullHierarchyTS_Multiplied fullMultiplied;

	public static final FusionStrategy TableTitleTS = new FusionStrategy("TableTitleTS", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_TableTitleTS_Max.class);
	public static final FusionStrategy TableTitleTS_Propagated = new FusionStrategy("TableTitleTS_Propagated", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_TableTitleTS_Max_Propagated.class);
	public static final FusionStrategy KBT_TableTitleTS_Weighted = new FusionStrategy("KBT_TableTitleTS_Weighted", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_TableTitleTS_Max_Weighted.class);
	public static final FusionStrategy KBT_TableTitleTS_Resized = new FusionStrategy("KBT_TableTitleTS_Resized", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_TableTitleTS_Max_Resized.class);
	public static final FusionStrategy KBT_TableTitleTS_Multiplied = new FusionStrategy("KBT_TableTitleTS_Multiplied", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_TableTitleTS_Max_Multiplied.class);

	@Inject
	Scorer_TableTitleTS_Max tableTitle;

	@Inject
	Scorer_TableTitleTS_Max_Propagated tableTitle_Propagated;
	@Inject
	Scorer_KBT_TableTitleTS_Max_Weighted tableTitleWeighted;
	@Inject
	Scorer_KBT_TableTitleTS_Max_Resized tableTitleResized;
	@Inject
	Scorer_KBT_TableTitleTS_Max_Multiplied tableTitleMultiplied;

	public static final FusionStrategy PageTitleTs = new FusionStrategy("PageTitleTs", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_PageTitleTS_Max.class);
	public static final FusionStrategy PageTitleTs_Propagated = new FusionStrategy("PageTitleTs_Propagated", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_PageTitleTS_Max_Propagated.class);
	public static final FusionStrategy KBT_PageTitleTS_Weighted = new FusionStrategy("KBT_PageTitleTS_Weighted", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_PageTitleTS_Max_Weighted.class);
	public static final FusionStrategy KBT_PageTitleTS_Resized = new FusionStrategy("KBT_PageTitleTS_Resized", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_PageTitleTS_Max_Resized.class);
	public static final FusionStrategy KBT_PageTitleTS_Multiplied = new FusionStrategy("KBT_PageTitleTS_Multiplied", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_PageTitleTS_Max_Multiplied.class);
	@Inject
	Scorer_PageTitleTS_Max pageTitle;
	@Inject
	Scorer_PageTitleTS_Max_Propagated pageTitle_Propageted;

	@Inject
	Scorer_KBT_PageTitleTS_Max_Weighted pageTitleWeighted;
	@Inject
	Scorer_KBT_PageTitleTS_Max_Resized pageTitleResized;
	@Inject
	Scorer_KBT_PageTitleTS_Max_Multiplied pageTitleMultiplied;

	public static final FusionStrategy TableYearsHierarchy = new FusionStrategy("TableYearsHierarchy", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_TableYearsHierarchy.class);
	public static final FusionStrategy TableYearsHierarchy_Propagated = new FusionStrategy("TableYearsHierarchy_Propagated", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_TableYearsHierarchy_Propagated.class);
	public static final FusionStrategy KBT_TableYearsHierarch_Weighted = new FusionStrategy("KBT_TableYearsHierarch_Weighted", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_TableYearsHierarchy_Weighted.class);
	public static final FusionStrategy KBT_TableYearsHierarch_Resized = new FusionStrategy("KBT_TableYearsHierarch_Resized", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_TableYearsHierarchy_Resized.class);
	public static final FusionStrategy KBT_TableYearsHierarch_Multiplied = new FusionStrategy("KBT_TableYearsHierarch_Multiplied", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_KBT_TableYearsHierarchy_Multiplied.class);

	@Inject
	Scorer_TableYearsHierarchy tableYearsHierarchy;
	@Inject
	Scorer_TableYearsHierarchy_Propagated tableYearsHierarchy_Propagated;
	@Inject
	Scorer_KBT_TableYearsHierarchy_Resized tableYearsHierarchyResized;
	@Inject
	Scorer_KBT_TableYearsHierarchy_Weighted tableYearsHierarchyWeighted;
	@Inject
	Scorer_KBT_TableYearsHierarchy_Multiplied tableYearsHierarchyMultiplied;

	@Inject
	Scorer_Combined_HierarchyTS_TableText combinedSimple;
	@Inject
	Scorer_KBT_Combined_HierarchyTS_TableText_WeightedMultiplied combined;

	@Inject
	Scorer_KBT_Combined_HierarchyTS_TableText_Multiplied multiplied;

	@Inject
	Scorer_KBT_Combined_HierarchyTS_TableText_Weighted weighted;

	public static final FusionStrategy KBT_Combined_HierarchyTS_TableText_WeightedMultiplied = new FusionStrategy(
			"KBT_Combined_HierarchyTS_TableText_WeightedMultiplied", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_KBT_Combined_HierarchyTS_TableText_WeightedMultiplied.class);

	public static final FusionStrategy KBT_Combined_HierarchyTS_TableText_Multiplied = new FusionStrategy("KBT_Combined_HierarchyTS_TableText_Multiplied",
			FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED, Scorer_KBT_Combined_HierarchyTS_TableText_Multiplied.class);

	public static final FusionStrategy KBT_Combined_HierarchyTS_TableText_Weighted = new FusionStrategy("KBT_Combined_HierarchyTS_TableText_Weighted",
			FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED, Scorer_KBT_Combined_HierarchyTS_TableText_Weighted.class);

	public static final FusionStrategy Combined_HierarchyTS_TableText = new FusionStrategy("Combined_HierarchyTS_TableText", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_Combined_HierarchyTS_TableText.class);

	@Inject
	Scorer_TS_KBT_Regression regressionScorer;

	public static final FusionStrategy TS_KBT_REGRESSION = new FusionStrategy("TsKbtReg", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_TS_KBT_Regression.class);

	@Inject
	Scorer_TS_Propagted_KBT_Regression_Propagated regressionScorer_Propagated;

	public static final FusionStrategy TS_PROPGATED_KBT_REGRESSION = new FusionStrategy("TsKbtRegPrpgtd", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_TS_Propagted_KBT_Regression_Propagated.class);

	@Inject
	Scorer_TS_KBT_Regression_ClassProperty regressionScorerClassProperty;

	public static final FusionStrategy TS_KBT_REGRESSION_CLASS_PROPERTY = new FusionStrategy("TsKbtRegAtt", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_TS_KBT_Regression_ClassProperty.class, ThresholdType.ClassPropertyId);

	@Inject
	Scorer_TS_Propgated_KBT_Regression_ClassProperty regressionScorerClassProperty_Propagated;

	public static final FusionStrategy TS_PROPGATED_KBT_REGRESSION_CLASS_PROPERTY = new FusionStrategy("TsKbtRegAttPrpgtd", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_TS_Propgated_KBT_Regression_ClassProperty.class, ThresholdType.ClassPropertyId);

	@Inject
	Scorer_Regression_PropertyId regressionScorerProperty;

	public static final FusionStrategy Regression_PropertyId = new FusionStrategy("KBT_TS_Regression_PropertyId", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_Regression_PropertyId.class, ThresholdType.PropertyId);

	@Inject
	Scorer_RegressionTrees regressionTreesScorer;

	public static final FusionStrategy RegressionTrees = new FusionStrategy("RegressionTrees", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_RegressionTrees.class);

	@Inject
	Scorer_RegressionTrees_ClassProperty regressionTreesScorerClassProperty;

	public static final FusionStrategy RegressionTrees_ClassProperty = new FusionStrategy("RegressionTreees_ClassProperty", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_RegressionTrees_ClassProperty.class, ThresholdType.ClassPropertyId);

	@Inject
	Scorer_RegressionTrees_PropertyId regressionTreesScorerProperty;

	public static final FusionStrategy RegressionTrees_PropertyId = new FusionStrategy("RegressionTrees_PropertyId", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_RegressionTrees_PropertyId.class, ThresholdType.PropertyId);

	@Inject
	Scorer_TimeStamps_Regression timeStampsRegressionScorer;

	public static final FusionStrategy Timestamps_Regression = new FusionStrategy("TsReg", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_TimeStamps_Regression.class);

	@Inject
	Scorer_TimeStamps_Regression_Propagated timeStampsRegressionScorer_Propagated;

	public static final FusionStrategy Timestamps_Regression_Propagated = new FusionStrategy("TsRegPrpgtd", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_TimeStamps_Regression_Propagated.class);

	@Inject
	Scorer_TimeStamps_Regression_ClassProperty timeStampsRegressionScorerClassProperty;

	public static final FusionStrategy Timestamps_Regression_ClassProperty = new FusionStrategy("TsRegAtt", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_TimeStamps_Regression_ClassProperty.class, ThresholdType.ClassPropertyId);

	@Inject
	Scorer_TimeStamps_Regression_ClassProperty_Propagated timeStampsRegressionScorerClassProperty_Propagated;

	public static final FusionStrategy Timestamps_Regression_ClassProperty_Propagated = new FusionStrategy("TsRegAttPrpgtd", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_TimeStamps_Regression_ClassProperty_Propagated.class, ThresholdType.ClassPropertyId);

	@Inject
	Scorer_TimeStamps_Regression_PropertyId timeStampsRegressionScorerProperty;

	public static final FusionStrategy Timestamps_Regression_PropertyId = new FusionStrategy("TimeStamps_Regression_PropertyId", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_TimeStamps_Regression_PropertyId.class, ThresholdType.PropertyId);

	// @Inject
	// Scorer_PageRank pageRank;
	//
	// public static final FusionStrategy PageRank = new FusionStrategy("Pr",
	// FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
	// Scorer_PageRank.class,
	// ThresholdType.General);
	//
	// @Inject
	// Scorer_PageRank_KBT_Regression pageRank_kbt_Regression;
	//
	// public static final FusionStrategy PageRank_KBT_Regression = new
	// FusionStrategy("PrKbtReg", FusionMethod.SCORED,
	// TestingMethod.CROSS_VALIDATED,
	// Scorer_PageRank_KBT_Regression.class, ThresholdType.General);
	//
	// @Inject
	// Scorer_PageRank_KBT_Regression_ClassProperty
	// pageRank_kbt_Regression_ClassProperty;
	//
	// public static final FusionStrategy PageRank_KBT_Regression_ClassProperty
	// = new FusionStrategy("PrKbtRegAtt", FusionMethod.SCORED,
	// TestingMethod.CROSS_VALIDATED,
	// Scorer_PageRank_KBT_Regression_ClassProperty.class,
	// ThresholdType.ClassPropertyId);
	//
	// @Inject
	// Scorer_PageRank_Regression pageRank_Regression;
	//
	// public static final FusionStrategy PageRank_Regression = new
	// FusionStrategy("PrReg", FusionMethod.SCORED,
	// TestingMethod.CROSS_VALIDATED,
	// Scorer_PageRank_Regression.class, ThresholdType.General);
	//
	// @Inject
	// Scorer_PageRank_Regression_ClassProperty
	// pageRank_Regression_ClassProperty;
	//
	// public static final FusionStrategy PageRank_Regression_ClassProperty =
	// new FusionStrategy("PrRegAtt", FusionMethod.SCORED,
	// TestingMethod.CROSS_VALIDATED,
	// Scorer_PageRank_Regression_ClassProperty.class,
	// ThresholdType.ClassPropertyId);

	// @Inject
	// Scorer_FullRegression_ClassProperty_Propagated fullRegressionPropagated;
	//
	// public static final FusionStrategy FullRegressionPropagated = new
	// FusionStrategy("AllRegPropagated", FusionMethod.SCORED,
	// TestingMethod.CROSS_VALIDATED,
	// Scorer_FullRegression_ClassProperty_Propagated.class,
	// ThresholdType.ClassPropertyId);
	//
	// @Inject
	// Scorer_FullRegression_ClassProperty fullRegression;
	//
	// public static final FusionStrategy FullRegression = new
	// FusionStrategy("AllReg", FusionMethod.SCORED,
	// TestingMethod.CROSS_VALIDATED,
	// Scorer_FullRegression_ClassProperty.class,
	// ThresholdType.ClassPropertyId);
	//
	// @Inject
	// Scorer_PageRank_Timestamps_Propagated_Regression_ClassProperty
	// pageRankTsPropagatedClassProperty;
	//
	// public static final FusionStrategy PageRank_TS_Propagated_ClassProperty =
	// new FusionStrategy("PrTsPropAtt", FusionMethod.SCORED,
	// TestingMethod.CROSS_VALIDATED,
	// Scorer_PageRank_Timestamps_Propagated_Regression_ClassProperty.class,
	// ThresholdType.ClassPropertyId);
	//
	// @Inject
	// Scorer_PageRank_Timestamps_Regression_ClassProperty
	// pageRankTsClassProperty;
	//
	// public static final FusionStrategy PageRank_TS_ClassProperty = new
	// FusionStrategy("PrTsAtt", FusionMethod.SCORED,
	// TestingMethod.CROSS_VALIDATED,
	// Scorer_PageRank_Timestamps_Regression_ClassProperty.class,
	// ThresholdType.ClassPropertyId);

	// @Inject
	// Scorer_Regression_Multiplied_Propagated
	// regressionScorer_Multiplied_Propagated;
	//
	// public static final FusionStrategy Regression_Multiplied_Propagated = new
	// FusionStrategy("KBT_TS_Regression_Multiplied", FusionMethod.SCORED,
	// TestingMethod.CROSS_VALIDATED,
	// Scorer_Regression_Multiplied_Propagated.class, ThresholdType.General);
	//
	// @Inject
	// Scorer_Regression_ClassProperty_Multiplied_Propagated
	// regressionScorerClassProperty_Multiplied_Propagated;
	//
	// public static final FusionStrategy
	// Regression_ClassProperty_Multiplied_Propagated = new
	// FusionStrategy("KBT_TS_Regression_ClassProperty_Multiplied",
	// FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
	// Scorer_Regression_ClassProperty_Multiplied_Propagated.class,
	// ThresholdType.ClassPropertyId);

	// @Inject
	// Scorer_Regression_Single_Propagated regressionScorer_Single_Propagated;
	//
	// public static final FusionStrategy Regression_Single_Propagated = new
	// FusionStrategy("Scorer_Regression_Single_Propagated",
	// FusionMethod.SCORED,
	// TestingMethod.CROSS_VALIDATED, Scorer_Regression_Single_Propagated.class,
	// ThresholdType.General);
	//
	// @Inject
	// Scorer_Regression_Single_ClassProperty_Propagated
	// regressionScorerClassProperty_Single_Propagated;
	//
	// public static final FusionStrategy
	// Regression_ClassProperty_Single_Propagated = new
	// FusionStrategy("Scorer_Regression_ClassProperty_Single_Propagated",
	// FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
	// Scorer_Regression_Single_ClassProperty_Propagated.class,
	// ThresholdType.ClassPropertyId);

	@Inject
	Scorer_NewTs_KBT newTsKBT;
	@Inject
	Scorer_NewTs_QKBT newTsQKBT;
	@Inject
	Scorer_NewTs newTs;

	@Inject
	TimestampSplitterFactory timestampSplitterFactory;

	@Inject
	SimpleSplitterFactory simpleSplitterFactory;

	@Inject
	Scorer_Baseline scorerBaseline;

	public static final FusionStrategy NEW_TS_KBT = new FusionStrategy("NEW TS KBT", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED, Scorer_NewTs_KBT.class,
			ThresholdType.ClassPropertyId);

	public static final FusionStrategy NEW_TS_QKBT = new FusionStrategy("NEW TS QKBT", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_NewTs_QKBT.class, ThresholdType.ClassPropertyId);

	public static final FusionStrategy NEW_TS_QKBT_WITHOUT_ROW_CELL = new FusionStrategy("NEW TS QKBT WITHOUT ROW CELL", FusionMethod.SCORED,
			TestingMethod.CROSS_VALIDATED, Scorer_NewTs_QKBT_without_CellRow.class, ThresholdType.ClassPropertyId);

	public static final FusionStrategy NEW_TS = new FusionStrategy("NEW TS", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED, Scorer_NewTs.class,
			ThresholdType.ClassPropertyId);

	public static final FusionStrategy NEW_TS_QKBT_PROPAGATED = new FusionStrategy("NEW_TS_QKBT_PROPAGATED", FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED,
			Scorer_NewTs_QKBT_Propagated.class, ThresholdType.ClassPropertyId);

	public static final FusionStrategy NEW_TS_QKBT_WITHOUR_ROW_CELL_PROPAGATED = new FusionStrategy("NEW_TS_QKBT_WITHOUT_ROW_CELL_PROPAGATED",
			FusionMethod.SCORED, TestingMethod.CROSS_VALIDATED, Scorer_NewTs_QKBT_Propagated_without_CellRow.class, ThresholdType.ClassPropertyId);

	@Inject
	ColumnScores columnScores;

	@Inject
	Scorer_NewTs_QKBT_Propagated NewTS_QKBT_PROPAGATED;
	@Inject
	Scorer_NewTs_QKBT_without_CellRow newTs_withoutRowCell;

	@Inject
	Scorer_NewTs_QKBT_Propagated_without_CellRow newTs_propagated_withoutRowCell;

	public void addScorers(ScorerRepository scorerRepository) {
		scorerRepository.add(Scorer_QKBT.class, qa);
		scorerRepository.add(Scorer_KBT.class, a);
		scorerRepository.add(Scorer_KBT_PRPGTD.class, kbtPropgated);
		scorerRepository.add(Scorer_QKBT_PRPGTD.class, qkbtPropgated);
		scorerRepository.add(Scorer_QKBT_Mapped.class, qkbtMapped);

		scorerRepository.add(Scorer_KBT_Combined_HierarchyTS_TableText_WeightedMultiplied.class, combined);
		scorerRepository.add(Scorer_KBT_Combined_HierarchyTS_TableText_Weighted.class, weighted);
		scorerRepository.add(Scorer_KBT_Combined_HierarchyTS_TableText_Multiplied.class, multiplied);
		scorerRepository.add(Scorer_Combined_HierarchyTS_TableText.class, combinedSimple);

		scorerRepository.add(Scorer_CellTS_Max.class, cellTS);
		scorerRepository.add(Scorer_CellTS_Max_Propagated.class, cellTS_Propagated);

		scorerRepository.add(Scorer_ColumnTS_Max.class, columnTs);
		scorerRepository.add(Scorer_ColumnTS_Max_Propagated.class, columnTs_Propagated);
		scorerRepository.add(Scorer_KBT_ColumnTS_Max_Weighted.class, b1);
		scorerRepository.add(Scorer_KBT_ColumnTS_Max_Resized.class, b2);
		scorerRepository.add(Scorer_KBT_ColumnTS_Max_Multiplied.class, b3);

		scorerRepository.add(Scorer_TableTextTS.class, c2);
		scorerRepository.add(Scorer_TableTextTS_Propagated.class, c2_Propagated);
		scorerRepository.add(Scorer_KBT_TableTextTS_MostCommon_Weighted.class, c);
		scorerRepository.add(Scorer_KBT_TableTextTS_MostCommon_Resized.class, d);
		scorerRepository.add(Scorer_KBT_TableTextTS_MostCommon_Multiplied.class, d2);

		scorerRepository.add(Scorer_HierarchyTS.class, hierarchyTs);
		scorerRepository.add(Scorer_HierarchyTS_Propagated.class, hierarchyTs_Propagated);
		scorerRepository.add(Scorer_KBT_HierarchyTS_Resized.class, e);
		scorerRepository.add(Scorer_KBT_HierarchyTS_Weighted.class, f);
		scorerRepository.add(Scorer_KBT_HierarchyTS_Multiplied.class, f2);

		scorerRepository.add(Scorer_FullHierarchyTS.class, fullhierarchy);
		scorerRepository.add(Scorer_FullHierarchyTS_Propagted.class, fullhierarchy_Propagated);
		scorerRepository.add(Scorer_KBT_FullHierarchyTS_Weighted.class, fullWeighted);
		scorerRepository.add(Scorer_KBT_FullHierarchyTS_Resized.class, fullResized);
		scorerRepository.add(Scorer_KBT_FullHierarchyTS_Multiplied.class, fullMultiplied);

		scorerRepository.add(Scorer_TableTitleTS_Max.class, tableTitle);
		scorerRepository.add(Scorer_TableTitleTS_Max_Propagated.class, tableTitle_Propagated);
		scorerRepository.add(Scorer_KBT_TableTitleTS_Max_Weighted.class, tableTitleWeighted);
		scorerRepository.add(Scorer_KBT_TableTitleTS_Max_Resized.class, tableTitleResized);
		scorerRepository.add(Scorer_KBT_TableTitleTS_Max_Multiplied.class, tableTitleMultiplied);

		scorerRepository.add(Scorer_PageTitleTS_Max.class, pageTitle);
		scorerRepository.add(Scorer_PageTitleTS_Max_Propagated.class, pageTitle_Propageted);
		scorerRepository.add(Scorer_KBT_PageTitleTS_Max_Weighted.class, pageTitleWeighted);
		scorerRepository.add(Scorer_KBT_PageTitleTS_Max_Resized.class, pageTitleResized);
		scorerRepository.add(Scorer_KBT_PageTitleTS_Max_Multiplied.class, pageTitleMultiplied);

		scorerRepository.add(Scorer_TableYearsHierarchy.class, tableYearsHierarchy);
		scorerRepository.add(Scorer_TableYearsHierarchy_Propagated.class, tableYearsHierarchy_Propagated);
		scorerRepository.add(Scorer_KBT_TableYearsHierarchy_Resized.class, tableYearsHierarchyResized);
		scorerRepository.add(Scorer_KBT_TableYearsHierarchy_Weighted.class, tableYearsHierarchyWeighted);
		scorerRepository.add(Scorer_KBT_TableYearsHierarchy_Multiplied.class, tableYearsHierarchyMultiplied);

		scorerRepository.add(Scorer_TS_KBT_Regression.class, regressionScorer);
		scorerRepository.add(Scorer_TS_Propagted_KBT_Regression_Propagated.class, regressionScorer_Propagated);
		scorerRepository.add(Scorer_TS_KBT_Regression_ClassProperty.class, regressionScorerClassProperty);
		scorerRepository.add(Scorer_TS_Propgated_KBT_Regression_ClassProperty.class, regressionScorerClassProperty_Propagated);
		scorerRepository.add(Scorer_Regression_PropertyId.class, regressionScorerProperty);

		// scorerRepository.add(Scorer_Regression_Multiplied_Propagated.class,
		// regressionScorer_Multiplied_Propagated);
		// scorerRepository.add(Scorer_Regression_ClassProperty_Multiplied_Propagated.class,
		// regressionScorerClassProperty_Multiplied_Propagated);
		// scorerRepository.add(Scorer_Regression_Single_Propagated.class,
		// regressionScorer_Single_Propagated);
		// scorerRepository.add(Scorer_Regression_Single_ClassProperty_Propagated.class,
		// regressionScorerClassProperty_Single_Propagated);

		scorerRepository.add(Scorer_RegressionTrees.class, regressionTreesScorer);
		scorerRepository.add(Scorer_RegressionTrees_ClassProperty.class, regressionTreesScorerClassProperty);
		scorerRepository.add(Scorer_RegressionTrees_PropertyId.class, regressionTreesScorerProperty);

		scorerRepository.add(Scorer_TimeStamps_Regression.class, timeStampsRegressionScorer);
		scorerRepository.add(Scorer_TimeStamps_Regression_Propagated.class, timeStampsRegressionScorer_Propagated);
		scorerRepository.add(Scorer_TimeStamps_Regression_ClassProperty.class, timeStampsRegressionScorerClassProperty);
		scorerRepository.add(Scorer_TimeStamps_Regression_ClassProperty_Propagated.class, timeStampsRegressionScorerClassProperty_Propagated);
		scorerRepository.add(Scorer_TimeStamps_Regression_PropertyId.class, timeStampsRegressionScorerProperty);

		// scorerRepository.add(Scorer_PageRank.class, pageRank);
		// scorerRepository.add(Scorer_PageRank_Regression.class,
		// pageRank_Regression);
		// scorerRepository.add(Scorer_PageRank_Regression_ClassProperty.class,
		// pageRank_Regression_ClassProperty);
		// scorerRepository.add(Scorer_PageRank_KBT_Regression.class,
		// pageRank_kbt_Regression);
		// scorerRepository.add(Scorer_PageRank_KBT_Regression_ClassProperty.class,
		// pageRank_kbt_Regression_ClassProperty);
		// scorerRepository.add(Scorer_PageRank_Timestamps_Regression_ClassProperty.class,
		// pageRankTsClassProperty);
		// scorerRepository.add(Scorer_PageRank_Timestamps_Propagated_Regression_ClassProperty.class,
		// pageRankTsPropagatedClassProperty);
		//
		// scorerRepository.add(Scorer_FullRegression_ClassProperty_Propagated.class,
		// fullRegressionPropagated);
		// scorerRepository.add(Scorer_FullRegression_ClassProperty.class,
		// fullRegression);

		scorerRepository.add(Scorer_NewTs_KBT.class, newTsKBT);
		scorerRepository.add(Scorer_NewTs_QKBT.class, newTsQKBT);
		scorerRepository.add(Scorer_NewTs.class, newTs);
		scorerRepository.add(Scorer_NewTs_QKBT_Propagated.class, NewTS_QKBT_PROPAGATED);
		scorerRepository.add(Scorer_Baseline.class, scorerBaseline);
		scorerRepository.add(ColumnScores.class, columnScores);

		scorerRepository.add(Scorer_NewTs_QKBT_without_CellRow.class, newTs_withoutRowCell);
		scorerRepository.add(Scorer_NewTs_QKBT_Propagated_without_CellRow.class, newTs_propagated_withoutRowCell);
	}

	@Inject
	NonDuplicateSplitterFactory nonDuplicateSplitterFactory;

	@Inject
	NonTimeDependentSplitterFactory nonTimeDependentSplitterFactory;

	public void addSplitterFactories(SplitterFactoryRepository repository) {
		repository.addDefault(simpleSplitterFactory);
		repository.add(TimestampSplitterFactory.class, timestampSplitterFactory);
		repository.add(NonDuplicateSplitterFactory.class, nonDuplicateSplitterFactory);
		repository.add(NonTimeDependentSplitterFactory.class, nonTimeDependentSplitterFactory);
	}

}
