package de.mannheim.uni.expansion.tablefusion.services.fusion.runner.preprocessors;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.HostId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues.SingleValue;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.TableScores;
import de.mannheim.uni.expansion.tablefusion.core.similarity.Similarity.SimilarityType;
import de.mannheim.uni.expansion.tablefusion.core.similarity.SimilarityCalculator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class PreprocessorOnePerHost implements FusionPreProcessor {
	private final AnnotatedTablesRepository repository;
	private final SimilarityCalculator similarityCalculator;
	private final KnowledgeBase dbPediaRepository;

	private final OntologyRepository ontologyRepository;

	@Inject
	public PreprocessorOnePerHost(AnnotatedTablesRepository repository, SimilarityCalculator similarityCalculator, KnowledgeBase dbPediaRepositoryNew,
			OntologyRepository ontologyRepository) {
		this.repository = repository;
		this.similarityCalculator = similarityCalculator;
		this.dbPediaRepository = dbPediaRepositoryNew;
		this.ontologyRepository = ontologyRepository;
	}

	// public Map<TripleId, MatchedValues> preprocess(Map<TripleId,
	// MatchedValues> values) {
	// return preprocess(values, null);
	// }
	//
	// public Map<TripleId, MatchedValues> preprocess(Map<TripleId,
	// MatchedValues> values, Map<PropertyId, Map<AnnotatedTableId,
	// Probability>> probabilities) {
	//
	// Map<TripleId, MatchedValues> newMap = new HashMap<TripleId,
	// MatchedValues>();
	//
	// for (Entry<TripleId, MatchedValues> entry : values.entrySet()) {
	// Map<AnnotatedTableId, Double> probs = null;
	// if (probabilities != null) {
	// probs = probabilities.get(entry.getKey().getPropertyId());
	// }
	// newMap.put(entry.getKey(), processValues(entry.getValue(),
	// entry.getKey(), probs));
	// }
	//
	// return values;
	// }

	private List<SingleValue> mergeValues(List<SingleValue> values, ValueType dataType, TripleId tripleId, TableScores probabilities) {

		/*
		 * This is a little bit weird. We need to find all values that are equal
		 * (i.e. similarity over threshold), and then "merge" them.
		 * 
		 * What is happening is the following: - We create a new List - We add
		 * the first element immediately to the new List - We iterate over all
		 * remaining values - If we do not find a match (foundMatch == false) we
		 * add the value to the new list - If we do find a match, we either do
		 * nothing (no probabilities or probability of value is lower) or
		 * replace existing value (probability of new value is higher)
		 */
		List<SingleValue> newList = new ArrayList<SingleValue>();
		newList.add(values.get(0));
		for (int i = 1; i < values.size(); i++) {
			boolean foundMatch = false;
			for (int j = 0; j < newList.size(); j++) {

				boolean isList = ontologyRepository.isListType(tripleId.getClassPropertyId().getPropertyId());
				SimilarityType type = SimilarityType.DIRECT;
				if (isList) {
					type = SimilarityType.LIST_OVERLAP;
				}
				double sim = similarityCalculator.getSimilarity(dataType, values.get(i).getValue(), newList.get(j).getValue());
				if (sim > similarityCalculator.getEquivalenceThreshold(dataType)) {
					if (probabilities != null) {

						Double d1 = probabilities.get(values.get(i).getId()).get(type).getAccuracy();
						Double d2 = probabilities.get(newList.get(j).getId()).get(type).getAccuracy();
						if (d2 == null && d1 != null) {
							newList.set(j, values.get(i));
						} else if (d2 != null & d1 != null && d1 > d2) {
							newList.set(j, values.get(i));

						}
					}

					foundMatch = true;
					break;
				}

			}

			if (!foundMatch) {
				newList.add(values.get(i));
			}
		}

		return newList;
	}

	private MatchedValues processValues(MatchedValues values, TripleId tripleId, TableScores probabilities) {

		Map<HostId, List<SingleValue>> byHostId = values.getValues().stream().collect(Collectors.groupingBy((sv) -> {
			return repository.getHostId(sv.getId().getTableId());
		}));

		List<SingleValue> listOfValues = new ArrayList<SingleValue>();

		for (Entry<HostId, List<SingleValue>> entry : byHostId.entrySet()) {
			listOfValues.addAll(mergeValues(entry.getValue(), values.getDataType(), tripleId, probabilities));
		}

		return new MatchedValues(listOfValues, values.getDataType());

	}

	@Override
	public MatchedValues preprocess(TripleId tripleId, MatchedValues matchedValues, ScoresContainer probabilities) {
		TableScores probs = null;
		if (probabilities != null) {
			probs = probabilities.getTableScores(tripleId.getClassId(), tripleId.getPropertyId(), tripleId.getInstanceId());
		}

		return processValues(matchedValues, tripleId, probs);
	}

}
