package de.mannheim.uni.expansion.tablefusion.services.fusion.strategies;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.model.PropertyIndexComposite;
import de.mannheim.uni.expansion.data.tables.model.RowIdx;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.Accuracy;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.TableScores;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;

public class ColumnScores implements Scorer {
	@Inject
	AnnotatedTablesRepository repository;

	@Override
	public ScoresContainer getScores(SplitGroupedValuesSets sets) {

		return new ScoresContainer() {

			@Override
			public TableScores getTableScores(ClassId classId, PropertyId propertyId, InstanceId instanceId) {
				return new TableScores() {

					@Override
					public Probability get(AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {
						AnnotatedTableMetaData metaData = repository.getMetaData(tableId);
						if (metaData != null && metaData.getProperties().get(columnIdx.getColumnIdx()) != null) {
							return new Probability(Accuracy.simple(metaData.getPropertyScore(
									new PropertyIndexComposite(columnIdx.getColumnIdx(), metaData.getProperties().get(columnIdx.getColumnIdx())))));
						}
						return new Probability(Accuracy.simple(0.0));
					}

					@Override
					public boolean containsKey(AnnotatedTableId tableId, ColumnIdx columnIdx, RowIdx rowIdx) {
						AnnotatedTableMetaData metaData = repository.getMetaData(tableId);
						if (metaData != null && metaData.getProperties().get(columnIdx.getColumnIdx()) != null) {
							return true;
						}
						return false;
					}
				};
			}

			@Override
			public Probability getScore(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx,
					RowIdx rowIdx) {
				AnnotatedTableMetaData metaData = repository.getMetaData(tableId);
				if (metaData != null && metaData.getProperties().get(columnIdx.getColumnIdx()) != null) {
					return new Probability(Accuracy.simple(metaData
							.getPropertyScore(new PropertyIndexComposite(columnIdx.getColumnIdx(), metaData.getProperties().get(columnIdx.getColumnIdx())))));
				}
				return new Probability(Accuracy.simple(0.0));
			}

			@Override
			public boolean containsKey(ClassId classId, PropertyId propertyId, InstanceId instanceId, AnnotatedTableId tableId, ColumnIdx columnIdx,
					RowIdx rowIdx) {
				AnnotatedTableMetaData metaData = repository.getMetaData(tableId);
				if (metaData != null && metaData.getProperties().get(columnIdx.getColumnIdx()) != null) {
					return true;
				}
				return false;
			}
		};
	}

}
