package de.mannheim.uni.expansion.tablefusion.services.fusion.strategies;

import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.function.BiFunction;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.RegressionTreesScores;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.RegressionTreesScoresClassPropertyId;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.RegressionTreesScoresPropertyId;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;
import weka.classifiers.trees.M5P;
import weka.core.Attribute;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.SparseInstance;

public class RegressionTreesScorer extends WekaScorer {

	private final ValueEvaluator valueEvaluator;

	@Inject
	public RegressionTreesScorer(ValueEvaluator valueEvaluator, OntologyRepository ontologyRepository) {
		super(valueEvaluator, ontologyRepository);
		this.valueEvaluator = valueEvaluator;
	}

	public <T> RegressionData<T> getaa(GroupedValues groupedValues, Map<String, ScoresContainer> scoresContainers, BiFunction<TripleId, ValueType, T> getKey) {
		String[] names = new String[scoresContainers.size()];
		ScoresContainer[] containers = new ScoresContainer[scoresContainers.size()];

		{
			int i = 0;
			for (Entry<String, ScoresContainer> sc : scoresContainers.entrySet()) {
				names[i] = sc.getKey();
				containers[i] = sc.getValue();

				i++;
			}
		}

		FastVector fv = new FastVector(names.length);
		fv.addElement(new Attribute("Score"));
		{
			for (String cs : names) {
				if (cs != null) {
					fv.addElement(new Attribute(cs));
				}
			}

		}

		if (true) {
			throw new IllegalStateException();
		}
		ConcurrentHashMap<T, ConcurrentLinkedQueue<Pair<Boolean, Double[]>>> listMap = null;// getListMap(groupedValues,
																							// containers,
																							// getKey);

		ConcurrentHashMap<T, WekaClassifier> modelMap = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(listMap.entrySet(), (entry) -> {

			WekaClassifier model = getRegressionModel(entry.getValue(), fv);

			modelMap.put(entry.getKey(), model);

		}, "REGRESSION");

		return new RegressionData<T>(containers, names, modelMap);

		// finalConsumer.accept(containers, names, weightsMap, interceptsMap);
	}

	private <T> WekaClassifier getRegressionModel(ConcurrentLinkedQueue<Pair<Boolean, Double[]>> list, FastVector fv) {

		int length = list.size();

		M5P regression = new M5P();
		regression.setUnpruned(false);
		// regression.setNumFolds(2);
		regression.setUseUnsmoothed(true);
		// regression.setUnpruned(true);
		// regression.setEliminateColinearAttributes(false);
		// regression.setAttributeSelectionMethod(new
		// SelectedTag(LinearRegression.SELECTION_GREEDY,
		// LinearRegression.TAGS_SELECTION));

		Instances data = new Instances("RegressionTree", fv, length);

		// double[] y = new double[length];
		// double[][] x = new double[length][];

		for (Pair<Boolean, Double[]> l : list) {
			Double[] z = new Double[fv.size() + 1];

			if (l.getFirst()) {
				z[0] = 1.0;
			} else {
				z[0] = 0.0;
			}

			Double[] scores = l.getSecond();
			int vp = 1;
			for (int i = 0; i < scores.length; i++) {

				z[vp] = scores[i];
				vp++;

			}

			int countNull = 0;
			for (Double s : z) {
				if (s == null) {
					countNull++;
				}
			}

			double[] instanceScores = new double[z.length - countNull];

			int[] index = new int[z.length - countNull];

			int j = 0;
			for (int i = 0; i < z.length; i++) {
				if (z[i] != null) {
					instanceScores[j] = z[i];
					index[j] = i;
					j++;
				}

			}

			Instance inst = new SparseInstance(1.0, instanceScores, index, z.length);
			inst.setDataset(data);
			data.add(inst);
		}

		data.setClassIndex(0);

		// RealMatrix matrix = new Array2DRowRealMatrix(x);
		//
		// RealMatrix vt = new SingularValueDecomposition(matrix).getVT();
		//
		// x = matrix.multiply(vt).getData();

		// matrix = new Array2DRowRealMatrix(x);
		//
		// double[][] cov = new
		// Covariance(matrix.transpose().getData())..getCovarianceMatrix().getData();

		try {
			regression.buildClassifier(data);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return new WekaClassifier(regression, data);

	}

	public RegressionTreesScores get(GroupedValues groupedValues, Map<String, ScoresContainer> scoresContainers) {
		// return null;
		RegressionData<Boolean> data = getaa(groupedValues, scoresContainers, (tid, cdt) -> {
			return Boolean.TRUE;
		});

		return new RegressionTreesScores(data.getNames(), data.getContainers(), data.getModelMap().get(Boolean.TRUE));
	}

	public RegressionTreesScoresClassPropertyId getClassPropertyId(GroupedValues groupedValues, Map<String, ScoresContainer> scoresContainers) {

		RegressionData<ClassPropertyId> data = getaa(groupedValues, scoresContainers, (tid, cdt) -> {
			return new ClassPropertyId(tid);
		});

		return new RegressionTreesScoresClassPropertyId(data.getNames(), data.getContainers(), data.getModelMap());

	}

	public RegressionTreesScoresPropertyId getPropertyId(GroupedValues groupedValues, Map<String, ScoresContainer> scoresContainers) {

		RegressionData<PropertyId> data = getaa(groupedValues, scoresContainers, (tid, cdt) -> {
			return tid.getPropertyId();
		});

		return new RegressionTreesScoresPropertyId(data.getNames(), data.getContainers(), data.getModelMap());

	}

	private static class RegressionData<T> {
		private final ScoresContainer[] containers;
		private final String[] names;
		private final Map<T, WekaClassifier> modelMap;

		public RegressionData(ScoresContainer[] containers, String[] names, Map<T, WekaClassifier> modelMap) {
			super();
			this.containers = containers;
			this.names = names;
			this.modelMap = modelMap;
		}

		public ScoresContainer[] getContainers() {
			return containers;
		}

		public String[] getNames() {
			return names;
		}

		public Map<T, WekaClassifier> getModelMap() {
			return modelMap;
		}

	}
}
