package de.mannheim.uni.expansion.tablefusion.services.fusion.strategies;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.data.tables.model.TableColumnId;
import de.mannheim.uni.expansion.tablefusion.analysis.Accuracy;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;
import de.mannheim.uni.expansion.tablefusion.core.scoring.QualifiedKbtCalculator;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.NonQualifiedColumnScores;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.similarity.Similarity.SimilarityType;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;
import de.mannheim.uni.expansion.utils.CHMHelper;

public class Scorer_KBT implements Scorer {

	private final QualifiedKbtCalculator qualifiedKbtCalculator;

	private volatile NonQualifiedColumnScores scores;

	public void preCompute(Map<Integer, SplitGroupedValuesSets> input) {
		double weight = 1.0 / input.size();
		Map<TableColumnId, Map<PropertyId, Probability>> retMap = new HashMap<>();
		for (Entry<Integer, SplitGroupedValuesSets> entry : input.entrySet()) {
			NonQualifiedColumnScores kbtScores = qualifiedKbtCalculator.byColumnIdNonQualified(entry.getValue().getLearningSet());
			Map<TableColumnId, Map<PropertyId, Probability>> innerMap = kbtScores.getFullScores();

			for (Entry<TableColumnId, Map<PropertyId, Probability>> innerEntry : innerMap.entrySet()) {
				Map<PropertyId, Probability> innerRetMap = CHMHelper.getFromMap(retMap, innerEntry.getKey(), () -> new HashMap<>());
				for (Entry<PropertyId, Probability> innerInnerEntry : innerEntry.getValue().entrySet()) {
					Accuracy acc = innerInnerEntry.getValue().get(SimilarityType.DIRECT);
					double newScore = acc.getAccuracy() * weight;
					int total = acc.getTotal();
					Probability prob = innerRetMap.getOrDefault(innerInnerEntry.getKey(), new Probability(new Accuracy(0.0, 0)));
					Accuracy currAcc = prob.get(SimilarityType.DIRECT);
					innerRetMap.put(innerInnerEntry.getKey(), new Probability(new Accuracy(currAcc.getAccuracy() + newScore, currAcc.getTotal() + total)));
				}
			}

		}
		scores = new NonQualifiedColumnScores(retMap);
	}

	@Inject
	public Scorer_KBT(QualifiedKbtCalculator qualifiedKbtCalculator) {
		this.qualifiedKbtCalculator = qualifiedKbtCalculator;
	}

	@Override
	public ScoresContainer getScores(SplitGroupedValuesSets sets) {
		if (scores != null) {
			return scores;
		}
		return qualifiedKbtCalculator.byColumnIdNonQualified(sets.getLearningSet());
	}
}
