package de.mannheim.uni.expansion.tablefusion.services.fusion.strategies;

import java.util.Optional;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.scoring.QualifiedKbtCalculator;
import de.mannheim.uni.expansion.tablefusion.core.scoring.TimestampScorerCleanedUp;
import de.mannheim.uni.expansion.tablefusion.core.scoring.TimestampScorerCleanedUp.MaxDiffsContainer;
import de.mannheim.uni.expansion.tablefusion.core.scoring.TimestampScorerCleanedUp.TimestampScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.QualifiedColumnScores;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.splitting.FilteredSplitter;
import de.mannheim.uni.expansion.tablefusion.core.splitting.TimestampMappingSplitter;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfigFilter;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;

public class Scorer_QKBT_Mapped implements Scorer, MaxDiffStrategy, SplitterProvider {

	private final QualifiedKbtCalculator qualifiedKbtCalculator;

	private final TimestampScorerCleanedUp timestampScorer;

	private final AnnotatedTablesRepository tablesRepository;
	private final PropertiesConfigFilter propertiesConfigFilter;

	private MaxDiffsContainer maxDiffsContainer = (classPropertyId) -> {
		return Optional.of(3);
	};

	@Inject
	public Scorer_QKBT_Mapped(QualifiedKbtCalculator qualifiedKbtCalculator, TimestampScorerCleanedUp timestampScorer,
			AnnotatedTablesRepository tablesRepository, PropertiesConfigFilter propertiesConfigFilter) {
		this.qualifiedKbtCalculator = qualifiedKbtCalculator;
		this.timestampScorer = timestampScorer;
		this.tablesRepository = tablesRepository;
		this.propertiesConfigFilter = propertiesConfigFilter;
	}

	public TimestampScoresContainer getScoresPrivate(SplitGroupedValuesSets sets) {
		QualifiedColumnScores qkbt = qualifiedKbtCalculator.byColumnIdQualified(sets.getLearningSet());
		return timestampScorer.getScoresWithoutTimestamps(sets.getLearningSet(), sets.getTestingSet(), qkbt, maxDiffsContainer);
	}

	public void setMaxDiffsContainer(MaxDiffsContainer maxDiffsContainer) {
		this.maxDiffsContainer = maxDiffsContainer;
	}

	@Override
	public FilteredSplitter getSplitter(SplitGroupedValuesSets sets) {
		TimestampScoresContainer scores = getScoresPrivate(sets);
		TimestampMappingSplitter splitter = new TimestampMappingSplitter(scores.getTimestampScores(), tablesRepository, propertiesConfigFilter,
				maxDiffsContainer);
		return splitter;
	}

	@Override
	public ScoresContainer getScores(SplitGroupedValuesSets sets) {
		return getScoresPrivate(sets);
	}
}
