package de.mannheim.uni.expansion.tablefusion.services.fusion.strategies;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;
import java.util.function.Consumer;

import javax.inject.Inject;

import de.mannheim.uni.expansion.tablefusion.analysis.Accuracy;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;
import de.mannheim.uni.expansion.tablefusion.core.scoring.QualifiedKbtCalculator;
import de.mannheim.uni.expansion.tablefusion.core.scoring.TimeStampsScoringCalculator;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.FixedScore;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.MultipliedScores;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.RegressionScoresClassPropertyId;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;

public class Scorer_TS_KBT_Regression_ClassProperty implements Scorer {

	private final QualifiedKbtCalculator probabilitiesCalculator;
	private final TimeStampsScoringCalculator timeStampsScoringCalculator;
	private final RegressionScorer regressionScorer;

	@Inject
	public Scorer_TS_KBT_Regression_ClassProperty(QualifiedKbtCalculator probabilitiesCalculator, TimeStampsScoringCalculator timeStampsScoringCalculator,
			RegressionScorer regressionScorer) {
		this.probabilitiesCalculator = probabilitiesCalculator;
		this.timeStampsScoringCalculator = timeStampsScoringCalculator;
		this.regressionScorer = regressionScorer;
	}

	// @Override
	// public ScoresContainer getScores(GroupedValues fullSet, GroupedValues
	// learningSet, GroupedValues testingSet) {
	// Map<String, ScoresContainer> map = new LinkedHashMap<>();
	//
	// ScoresContainer kbt = probabilitiesCalculator.byProperty(learningSet);
	// ScoresContainer GET_HIERARICAL_YEAR =
	// timeStampsScoringCalculator.genericYearScorer(fullSet,
	// timeStampsScoringCalculator.GET_HIERARICAL_YEAR);
	// ScoresContainer GET_FULL_HIERARICAL_YEAR =
	// timeStampsScoringCalculator.genericYearScorer(fullSet,
	// timeStampsScoringCalculator.GET_FULL_HIERARICAL_YEAR);
	// ScoresContainer MAX_TABLE_TILE =
	// timeStampsScoringCalculator.genericYearScorer(fullSet,
	// timeStampsScoringCalculator.MAX_TABLE_TILE);
	// ScoresContainer MAX_COLUMN_YEAR =
	// timeStampsScoringCalculator.genericYearScorer(fullSet,
	// timeStampsScoringCalculator.MAX_COLUMN_YEAR);
	// ScoresContainer MOST_COMMON_TABLE_TEXT_YEAR =
	// timeStampsScoringCalculator.genericYearScorer(fullSet,
	// timeStampsScoringCalculator.MOST_COMMON_TABLE_TEXT_YEAR);
	// ScoresContainer MAX_PAGE_TITLE =
	// timeStampsScoringCalculator.genericYearScorer(fullSet,
	// timeStampsScoringCalculator.MAX_PAGE_TITLE);
	// ScoresContainer TABLE_YEAR_HIERARCHY =
	// timeStampsScoringCalculator.genericYearScorer(fullSet,
	// timeStampsScoringCalculator.TABLE_YEAR_HIERARCHY);
	// ScoresContainer MAX_CELL_YEAR =
	// timeStampsScoringCalculator.genericYearScorer(fullSet,
	// timeStampsScoringCalculator.CELL_YEAR);
	//
	// map.put("KBT", kbt);
	//
	// ScoresContainer main = new FixedScore(new Probability(new Accuracy(1.0,
	// 1)));
	// map.put("GET_HIERARICAL_YEAR", new MultipliedScores(main,
	// GET_HIERARICAL_YEAR));
	// map.put("GET_FULL_HIERARICAL_YEAR", new MultipliedScores(main,
	// GET_FULL_HIERARICAL_YEAR));
	// map.put("MAX_TABLE_TILE", new MultipliedScores(main, MAX_TABLE_TILE));
	// map.put("MAX_COLUMN_YEAR", new MultipliedScores(main, MAX_COLUMN_YEAR));
	// map.put("MOST_COMMON_TABLE_TEXT_YEAR", new MultipliedScores(main,
	// MOST_COMMON_TABLE_TEXT_YEAR));
	// map.put("MAX_PAGE_TITLE", new MultipliedScores(main, MAX_PAGE_TITLE));
	// map.put("TABLE_YEAR_HIERARCHY", new MultipliedScores(main,
	// TABLE_YEAR_HIERARCHY));
	// map.put("MAX_CELL_YEAR", new MultipliedScores(main, MAX_CELL_YEAR));
	//
	// map.put("KBT_MULTIPLIED_GET_HIERARICAL_YEAR", new MultipliedScores(kbt,
	// GET_HIERARICAL_YEAR));
	// map.put("KBT_MULTIPLIED_GET_FULL_HIERARICAL_YEAR", new
	// MultipliedScores(kbt, GET_FULL_HIERARICAL_YEAR));
	// map.put("KBT_MULTIPLIED_MAX_TABLE_TILE", new MultipliedScores(kbt,
	// MAX_TABLE_TILE));
	// map.put("KBT_MULTIPLIED_MAX_COLUMN_YEAR", new MultipliedScores(kbt,
	// MAX_COLUMN_YEAR));
	// map.put("KBT_MULTIPLIED_MOST_COMMON_TABLE_TEXT_YEAR", new
	// MultipliedScores(kbt, MOST_COMMON_TABLE_TEXT_YEAR));
	// map.put("KBT_MULTIPLIED_MAX_PAGE_TITLE", new MultipliedScores(kbt,
	// MAX_PAGE_TITLE));
	// map.put("KBT_MULTIPLIED_TABLE_YEAR_HIERARCHY", new MultipliedScores(kbt,
	// TABLE_YEAR_HIERARCHY));
	// map.put("KBT_MULTIPLIED_MAX_CELL_YEAR", new MultipliedScores(kbt,
	// MAX_CELL_YEAR));
	//
	// RegressionScoresClassPropertyId rs =
	// regressionScorer.getClassPropertyId(learningSet, map);
	//
	// // System.out.println(rs.getWeightsMap().toString());
	//
	// // for (Entry<String, Double> e : rs.getWeights().entrySet()) {
	// // System.out.println(e.getKey() + ": " + e.getValue());
	// // }
	// return rs;
	// }
	//
	//

	private final SimpleLockableConcurrentCache<UUID, ScoresContainer> cache = new SimpleLockableConcurrentCache<>();

	@Override
	public ScoresContainer getScores(SplitGroupedValuesSets sets) {
		UUID uuid = sets.getLearningSet().getUUID();

		if (cache.hasCacheFor(uuid)) {
			return cache.get(uuid);
		} else {
			Optional<Consumer<ScoresContainer>> lock = cache.getLock(uuid);
			if (lock.isPresent()) {
				ScoresContainer sc = getScoresInternal(sets);
				lock.get().accept(sc);
				return sc;
			} else {
				return cache.poll(uuid);
			}
		}

	}

	private ScoresContainer getScoresInternal(SplitGroupedValuesSets sets) {
		Map<String, ScoresContainer> map = new LinkedHashMap<>();

		ScoresContainer kbt = probabilitiesCalculator.byColumnIdNonQualified(sets.getLearningSet());
		ScoresContainer GET_HIERARICAL_YEAR = timeStampsScoringCalculator.genericYearScorer(sets.getFullSet(), timeStampsScoringCalculator.GET_HIERARICAL_YEAR);
		ScoresContainer GET_FULL_HIERARICAL_YEAR = timeStampsScoringCalculator.genericYearScorer(sets.getFullSet(),
				timeStampsScoringCalculator.GET_FULL_HIERARICAL_YEAR);
		ScoresContainer MAX_TABLE_TILE = timeStampsScoringCalculator.genericYearScorer(sets.getFullSet(), timeStampsScoringCalculator.MAX_TABLE_TILE);
		ScoresContainer MAX_COLUMN_YEAR = timeStampsScoringCalculator.genericYearScorer(sets.getFullSet(), timeStampsScoringCalculator.MAX_COLUMN_YEAR);
		ScoresContainer MOST_COMMON_TABLE_TEXT_YEAR = timeStampsScoringCalculator.genericYearScorer(sets.getFullSet(),
				timeStampsScoringCalculator.MOST_COMMON_TABLE_TEXT_YEAR);
		ScoresContainer MAX_PAGE_TITLE = timeStampsScoringCalculator.genericYearScorer(sets.getFullSet(), timeStampsScoringCalculator.MAX_PAGE_TITLE);
		ScoresContainer TABLE_YEAR_HIERARCHY = timeStampsScoringCalculator.genericYearScorer(sets.getFullSet(),
				timeStampsScoringCalculator.TABLE_YEAR_HIERARCHY);
		ScoresContainer MAX_CELL_YEAR = timeStampsScoringCalculator.genericYearScorer(sets.getFullSet(), timeStampsScoringCalculator.CELL_YEAR);

		map.put("KBT", kbt);

		ScoresContainer main = new FixedScore(new Probability(new Accuracy(1.0, 1)));
		map.put("GET_HIERARICAL_YEAR", new MultipliedScores(main, GET_HIERARICAL_YEAR));
		map.put("GET_FULL_HIERARICAL_YEAR", new MultipliedScores(main, GET_FULL_HIERARICAL_YEAR));
		map.put("MAX_TABLE_TILE", new MultipliedScores(main, MAX_TABLE_TILE));
		map.put("MAX_COLUMN_YEAR", new MultipliedScores(main, MAX_COLUMN_YEAR));
		map.put("MOST_COMMON_TABLE_TEXT_YEAR", new MultipliedScores(main, MOST_COMMON_TABLE_TEXT_YEAR));
		map.put("MAX_PAGE_TITLE", new MultipliedScores(main, MAX_PAGE_TITLE));
		map.put("TABLE_YEAR_HIERARCHY", new MultipliedScores(main, TABLE_YEAR_HIERARCHY));
		map.put("MAX_CELL_YEAR", new MultipliedScores(main, MAX_CELL_YEAR));

		map.put("KBT_MULTIPLIED_GET_HIERARICAL_YEAR", new MultipliedScores(kbt, GET_HIERARICAL_YEAR));
		map.put("KBT_MULTIPLIED_GET_FULL_HIERARICAL_YEAR", new MultipliedScores(kbt, GET_FULL_HIERARICAL_YEAR));
		map.put("KBT_MULTIPLIED_MAX_TABLE_TILE", new MultipliedScores(kbt, MAX_TABLE_TILE));
		map.put("KBT_MULTIPLIED_MAX_COLUMN_YEAR", new MultipliedScores(kbt, MAX_COLUMN_YEAR));
		map.put("KBT_MULTIPLIED_MOST_COMMON_TABLE_TEXT_YEAR", new MultipliedScores(kbt, MOST_COMMON_TABLE_TEXT_YEAR));
		map.put("KBT_MULTIPLIED_MAX_PAGE_TITLE", new MultipliedScores(kbt, MAX_PAGE_TITLE));
		map.put("KBT_MULTIPLIED_TABLE_YEAR_HIERARCHY", new MultipliedScores(kbt, TABLE_YEAR_HIERARCHY));
		map.put("KBT_MULTIPLIED_MAX_CELL_YEAR", new MultipliedScores(kbt, MAX_CELL_YEAR));

		RegressionScoresClassPropertyId rs = regressionScorer.getClassPropertyId(sets.getLearningSet(), map);

		// System.out.println(rs.getWeightsMap().toString());

		// for (Entry<String, Double> e : rs.getWeights().entrySet()) {
		// System.out.println(e.getKey() + ": " + e.getValue());
		// }
		return rs;
	}
}
