package de.mannheim.uni.expansion.tablefusion.services.fusion.strategies;

import java.util.Collection;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.BiFunction;

import de.mannheim.uni.expansion.data.ontology.model.QualifiedTripleId;
import de.mannheim.uni.expansion.data.ontology.model.TripleIdContainer;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.similarity.Similarity.SimilarityType;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.fusedvalue.FusedValue;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;
import de.uni_mannheim.informatik.dws.expansion.evaluation.ValueEvaluator;

public class WekaScorer {

	private final ValueEvaluator valueEvaluator;
	private final OntologyRepository ontologyRepository;

	public WekaScorer(ValueEvaluator valueEvaluator, OntologyRepository ontologyRepository) {
		this.valueEvaluator = valueEvaluator;
		this.ontologyRepository = ontologyRepository;
	}

	<E extends TripleIdContainer, T> Map<T, Collection<Pair<Boolean, Double[]>>> getListMap(Map<E, MatchedValues> groupedValues, ScoresContainer[] containers,
			BiFunction<E, ValueType, T> getKey) {
		int maxCorrect = 10000;
		int maxWrong = 10000;
		ConcurrentHashMap<T, Collection<Pair<Boolean, Double[]>>> listMap = new ConcurrentHashMap<>();
		ConcurrentHashMap<T, AtomicInteger> currentCorrect = new ConcurrentHashMap<>();
		ConcurrentHashMap<T, AtomicInteger> currentWrong = new ConcurrentHashMap<>();

		PersistentParallel6.runAndFinish(groupedValues.entrySet(), (gv) -> {

			if (ontologyRepository.isListType(gv.getKey().tripleId().getPropertyId())) {
				T key = getKey.apply(gv.getKey(), gv.getValue().getDataType());
				CHMHelper.getFromCHM(listMap, key, () -> new ConcurrentLinkedQueue<Pair<Boolean, Double[]>>());
				return;
			}

			PersistentParallel6.runAndFinish(gv.getValue().getValues(), (sv) -> {

				Optional<Boolean> correct = valueEvaluator.correct(gv.getKey(), FusedValue.fromObject(sv.getValue()));
				if (correct.isPresent()) {

					T key = getKey.apply(gv.getKey(), gv.getValue().getDataType());

					if (correct.get()) {
						int current = CHMHelper.getFromCHM(currentCorrect, key, () -> new AtomicInteger()).getAndIncrement();
						if (current >= maxCorrect) {
							return;
						}
					} else {
						int current = CHMHelper.getFromCHM(currentWrong, key, () -> new AtomicInteger()).getAndIncrement();
						if (current >= maxWrong) {
							return;
						}
					}

					Double[] scores = new Double[containers.length];
					for (int i = 0; i < containers.length; i++) {
						Probability probability = null;
						if (gv.getKey() instanceof QualifiedTripleId) {
							QualifiedTripleId qualifiedTripleId = (QualifiedTripleId) gv.getKey();
							probability = containers[i].getScore(qualifiedTripleId.getTripleId(), sv.getId(), qualifiedTripleId.getQualifiers());
						} else {
							probability = containers[i].getScore(gv.getKey().tripleId().getClassId(), gv.getKey().tripleId().getPropertyId(),
									gv.getKey().tripleId().getInstanceId(), sv.getId());
						}
						if (probability != null)
							scores[i] = probability.get(SimilarityType.DIRECT).getAccuracy();
						else
							scores[i] = 0.0;
					}

					ConcurrentLinkedQueue<Pair<Boolean, Double[]>> queue = (ConcurrentLinkedQueue<Pair<Boolean, Double[]>>) CHMHelper.getFromCHM(listMap, key,
							() -> new ConcurrentLinkedQueue<Pair<Boolean, Double[]>>());
					queue.add(new Pair<Boolean, Double[]>(correct.get(), scores));
				}
			} , this.getClass(), gv.getValue().getValues().size() < 200000);

		} , "RegressionScorer");

		// Map<T, List<Pair<Boolean, Double[]>>> listMapNew = new
		// ConcurrentHashMap<>();
		// for (Entry<T, ConcurrentLinkedQueue<Pair<Boolean, Double[]>>> entry :
		// listMap.entrySet()) {
		// listMapNew.put(entry.getKey(), new ArrayList<>(entry.getValue()));
		// }

		return listMap;
	}
}
