package de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.newTS;

import java.io.IOException;
import java.util.Map;
import java.util.Optional;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import de.mannheim.uni.expansion.data.ontology.model.ClassPropertyId;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.scoring.QualifiedKbtCalculator;
import de.mannheim.uni.expansion.tablefusion.core.scoring.TimestampScorerCleanedUp;
import de.mannheim.uni.expansion.tablefusion.core.scoring.TimestampScorerCleanedUp.MaxDiffsContainer;
import de.mannheim.uni.expansion.tablefusion.core.scoring.TimestampScorerCleanedUp.TimestampScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.splitting.FilteredSplitter;
import de.mannheim.uni.expansion.tablefusion.core.splitting.TimestampMappingSplitter;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.PropertiesConfigFilter;
import de.mannheim.uni.expansion.tablefusion.learning.SplitGroupedValuesSets;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.MaxDiffStrategy;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.SplitterProvider;

public class Scorer_NewTs_QKBT_without_CellRow implements Scorer, MaxDiffStrategy, SplitterProvider {

	private final TimestampScorerCleanedUp timestampScorer;

	private final QualifiedKbtCalculator kbtScorer;
	private final AnnotatedTablesRepository tablesRepository;
	private final PropertiesConfigFilter propertiesConfigFilter;

	private MaxDiffsContainer maxDiffsContainer = (classPropertyId) -> {
		return Optional.of(3);
	};

	@Inject
	public Scorer_NewTs_QKBT_without_CellRow(QualifiedKbtCalculator kbtScorer, TimestampScorerCleanedUp timestampScorer,
			AnnotatedTablesRepository tablesRepository, PropertiesConfigFilter propertiesConfigFilter) {
		this.kbtScorer = kbtScorer;
		this.timestampScorer = timestampScorer;
		this.tablesRepository = tablesRepository;
		this.propertiesConfigFilter = propertiesConfigFilter;
	}

	public TimestampScoresContainer getScoresPrivate(SplitGroupedValuesSets sets) {
		ScoresContainer qkbt = kbtScorer.byColumnIdQualified(sets.getLearningSet());
		TimestampScoresContainer ts = timestampScorer.getScoresWithoutRowCell(sets.getLearningSet(), sets.getTestingSet(), qkbt, maxDiffsContainer, false);

		// Map<String, ScoresContainer> map = new LinkedHashMap<>();
		// map.put("QKBT", qkbt);
		// map.put("TS", ts);
		// map.put("QKBT_TS", new MultipliedScores(qkbt, ts));
		//
		// RegressionScoresClassPropertyId rs =
		// regressionScorer.getClassPropertyId(sgvLearning, map);

		return ts;
	}

	public MaxDiffsContainer getMaxDiffsContainer() {
		return maxDiffsContainer;
	}

	public void setMaxDiffsContainer(MaxDiffsContainer maxDiffsContainer) {
		this.maxDiffsContainer = maxDiffsContainer;
	}

	@Override
	public FilteredSplitter getSplitter(SplitGroupedValuesSets sets) {
		TimestampScoresContainer scores = getScoresPrivate(sets);
		TimestampMappingSplitter splitter = new TimestampMappingSplitter(scores.getTimestampScores(), tablesRepository, propertiesConfigFilter,
				maxDiffsContainer);
		return splitter;
	}

	public Map<ClassPropertyId, Map<String, MutableInt>> getMappingEvaluationCSV(SplitGroupedValuesSets sets) throws IOException {
		TimestampScoresContainer scores = getScoresPrivate(sets);
		return scores.evaluateMappingsCSV();
	}

	@Override
	public ScoresContainer getScores(SplitGroupedValuesSets sets) {
		return getScoresPrivate(sets);
	}

}
