package de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.pagerank;

import java.util.LinkedHashMap;
import java.util.Map;

import javax.inject.Inject;

import de.mannheim.uni.expansion.tablefusion.adapters.ranking.RankRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.scoring.ProbabilitiesCalculator;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.MultipliedScores;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.RegressionScoresClassPropertyId;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.RegressionScorer;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer;

public class Scorer_PageRank_KBT_Regression_ClassProperty implements Scorer {

	private final RankRepository rankRepository;
	private final RegressionScorer regressionScorer;
	private final ProbabilitiesCalculator probabilitiesCalculator;

	@Inject
	public Scorer_PageRank_KBT_Regression_ClassProperty(RankRepository rankRepository, RegressionScorer regressionScorer,
			ProbabilitiesCalculator probabilitiesCalculator) {
		this.rankRepository = rankRepository;
		this.regressionScorer = regressionScorer;
		this.probabilitiesCalculator = probabilitiesCalculator;
	}

	@Override
	public ScoresContainer getScores(GroupedValues fullSet, GroupedValues learningSet, GroupedValues testingSet) {
		Map<String, ScoresContainer> map = new LinkedHashMap<>();

		ScoresContainer kbt = probabilitiesCalculator.byProperty(learningSet);
		ScoresContainer PAGERANK = rankRepository.getScores(fullSet);

		map.put("KBT", kbt);
		map.put("PAGERANK", PAGERANK);
		map.put("PAGERANK_KBT", new MultipliedScores(kbt, PAGERANK));

		RegressionScoresClassPropertyId rs = regressionScorer.getClassPropertyId(learningSet, map);

		return rs;
	}

}
