package de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.pagerank;

import java.util.LinkedHashMap;
import java.util.Map;

import javax.inject.Inject;

import de.mannheim.uni.expansion.tablefusion.adapters.ranking.RankRepository;
import de.mannheim.uni.expansion.tablefusion.analysis.Accuracy;
import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.scoring.Probability;
import de.mannheim.uni.expansion.tablefusion.core.scoring.TimeStampsScoringCalculator;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.FixedScore;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.MultipliedScores;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.RegressionScoresClassPropertyId;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.RegressionScorer;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer;

public class Scorer_PageRank_Timestamps_Regression_ClassProperty implements Scorer {

	private final RankRepository rankRepository;
	private final RegressionScorer regressionScorer;
	private final TimeStampsScoringCalculator timeStampsScoringCalculator;

	@Inject
	public Scorer_PageRank_Timestamps_Regression_ClassProperty(RankRepository rankRepository, RegressionScorer regressionScorer,
			TimeStampsScoringCalculator timeStampsScoringCalculator) {
		this.rankRepository = rankRepository;
		this.regressionScorer = regressionScorer;
		this.timeStampsScoringCalculator = timeStampsScoringCalculator;
	}

	@Override
	public ScoresContainer getScores(GroupedValues fullSet, GroupedValues learningSet, GroupedValues testingSet) {
		Map<String, ScoresContainer> map = new LinkedHashMap<>();
		ScoresContainer PAGERANK = rankRepository.getScores(fullSet);

		ScoresContainer GET_HIERARICAL_YEAR = timeStampsScoringCalculator.genericYearScorer(fullSet, timeStampsScoringCalculator.GET_HIERARICAL_YEAR);
		ScoresContainer GET_FULL_HIERARICAL_YEAR = timeStampsScoringCalculator.genericYearScorer(fullSet, timeStampsScoringCalculator.GET_FULL_HIERARICAL_YEAR);
		ScoresContainer MAX_TABLE_TILE = timeStampsScoringCalculator.genericYearScorer(fullSet, timeStampsScoringCalculator.MAX_TABLE_TILE);
		ScoresContainer MAX_COLUMN_YEAR = timeStampsScoringCalculator.genericYearScorer(fullSet, timeStampsScoringCalculator.MAX_COLUMN_YEAR);
		ScoresContainer MOST_COMMON_TABLE_TEXT_YEAR = timeStampsScoringCalculator.genericYearScorer(fullSet,
				timeStampsScoringCalculator.MOST_COMMON_TABLE_TEXT_YEAR);
		ScoresContainer MAX_PAGE_TITLE = timeStampsScoringCalculator.genericYearScorer(fullSet, timeStampsScoringCalculator.MAX_PAGE_TITLE);
		ScoresContainer TABLE_YEAR_HIERARCHY = timeStampsScoringCalculator.genericYearScorer(fullSet, timeStampsScoringCalculator.TABLE_YEAR_HIERARCHY);
		ScoresContainer MAX_CELL_YEAR = timeStampsScoringCalculator.genericYearScorer(fullSet, timeStampsScoringCalculator.CELL_YEAR);

		ScoresContainer main = new FixedScore(new Probability(new Accuracy(1.0, 1)));
		map.put("GET_HIERARICAL_YEAR", new MultipliedScores(main, GET_HIERARICAL_YEAR));
		map.put("GET_FULL_HIERARICAL_YEAR", new MultipliedScores(main, GET_FULL_HIERARICAL_YEAR));
		map.put("MAX_TABLE_TILE", new MultipliedScores(main, MAX_TABLE_TILE));
		map.put("MAX_COLUMN_YEAR", new MultipliedScores(main, MAX_COLUMN_YEAR));
		map.put("MOST_COMMON_TABLE_TEXT_YEAR", new MultipliedScores(main, MOST_COMMON_TABLE_TEXT_YEAR));
		map.put("MAX_PAGE_TITLE", new MultipliedScores(main, MAX_PAGE_TITLE));
		map.put("TABLE_YEAR_HIERARCHY", new MultipliedScores(main, TABLE_YEAR_HIERARCHY));
		map.put("MAX_CELL_YEAR", new MultipliedScores(main, MAX_CELL_YEAR));

		map.put("PAGERANK", PAGERANK);

		map.put("PAGERANK_MULTIPLIED_GET_HIERARICAL_YEAR", new MultipliedScores(PAGERANK, GET_HIERARICAL_YEAR));
		map.put("PAGERANK_MULTIPLIED_GET_FULL_HIERARICAL_YEAR", new MultipliedScores(PAGERANK, GET_FULL_HIERARICAL_YEAR));
		map.put("PAGERANK_MULTIPLIED_MAX_TABLE_TILE", new MultipliedScores(PAGERANK, MAX_TABLE_TILE));
		map.put("PAGERANK_MULTIPLIED_MAX_COLUMN_YEAR", new MultipliedScores(PAGERANK, MAX_COLUMN_YEAR));
		map.put("PAGERANK_MULTIPLIED_MOST_COMMON_TABLE_TEXT_YEAR", new MultipliedScores(PAGERANK, MOST_COMMON_TABLE_TEXT_YEAR));
		map.put("PAGERANK_MULTIPLIED_MAX_PAGE_TITLE", new MultipliedScores(PAGERANK, MAX_PAGE_TITLE));
		map.put("PAGERANK_MULTIPLIED_TABLE_YEAR_HIERARCHY", new MultipliedScores(PAGERANK, TABLE_YEAR_HIERARCHY));
		map.put("PAGERANK_MULTIPLIED_MAX_CELL_YEAR", new MultipliedScores(PAGERANK, MAX_CELL_YEAR));

		RegressionScoresClassPropertyId rs = regressionScorer.getClassPropertyId(learningSet, map);

		return rs;
	}

}
