package de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.regressionTrees;

import java.util.LinkedHashMap;
import java.util.Map;

import javax.inject.Inject;

import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.scoring.ProbabilitiesCalculator;
import de.mannheim.uni.expansion.tablefusion.core.scoring.TimeStampsScoringCalculator;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.MultipliedScores;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.RegressionTreesScoresClassPropertyId;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.RegressionTreesScorer;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer;

public class Scorer_RegressionTrees_ClassProperty implements Scorer {

	private final ProbabilitiesCalculator probabilitiesCalculator;
	private final TimeStampsScoringCalculator timeStampsScoringCalculator;
	private final RegressionTreesScorer regressionScorer;

	@Inject
	public Scorer_RegressionTrees_ClassProperty(ProbabilitiesCalculator probabilitiesCalculator, TimeStampsScoringCalculator timeStampsScoringCalculator,
			RegressionTreesScorer regressionScorer) {
		this.probabilitiesCalculator = probabilitiesCalculator;
		this.timeStampsScoringCalculator = timeStampsScoringCalculator;
		this.regressionScorer = regressionScorer;
	}

	@Override
	public ScoresContainer getScores(GroupedValues fullSet, GroupedValues learningSet, GroupedValues testingSet) {
		Map<String, ScoresContainer> map = new LinkedHashMap<>();

		ScoresContainer kbt = probabilitiesCalculator.byProperty(learningSet);
		ScoresContainer GET_HIERARICAL_YEAR = timeStampsScoringCalculator.genericYearScorer(fullSet, timeStampsScoringCalculator.GET_HIERARICAL_YEAR);
		ScoresContainer GET_FULL_HIERARICAL_YEAR = timeStampsScoringCalculator.genericYearScorer(fullSet, timeStampsScoringCalculator.GET_FULL_HIERARICAL_YEAR);
		ScoresContainer MAX_TABLE_TILE = timeStampsScoringCalculator.genericYearScorer(fullSet, timeStampsScoringCalculator.MAX_TABLE_TILE);
		ScoresContainer MAX_COLUMN_YEAR = timeStampsScoringCalculator.genericYearScorer(fullSet, timeStampsScoringCalculator.MAX_COLUMN_YEAR);
		ScoresContainer MOST_COMMON_TABLE_TEXT_YEAR = timeStampsScoringCalculator.genericYearScorer(fullSet,
				timeStampsScoringCalculator.MOST_COMMON_TABLE_TEXT_YEAR);
		ScoresContainer MAX_PAGE_TITLE = timeStampsScoringCalculator.genericYearScorer(fullSet, timeStampsScoringCalculator.MAX_PAGE_TITLE);
		ScoresContainer TABLE_YEAR_HIERARCHY = timeStampsScoringCalculator.genericYearScorer(fullSet, timeStampsScoringCalculator.TABLE_YEAR_HIERARCHY);

		map.put("KBT", kbt);

		map.put("GET_HIERARICAL_YEAR", GET_HIERARICAL_YEAR);
		map.put("GET_FULL_HIERARICAL_YEAR", GET_FULL_HIERARICAL_YEAR);
		map.put("MAX_TABLE_TILE", MAX_TABLE_TILE);
		map.put("MAX_COLUMN_YEAR", MAX_COLUMN_YEAR);
		map.put("MOST_COMMON_TABLE_TEXT_YEAR", MOST_COMMON_TABLE_TEXT_YEAR);
		map.put("MAX_PAGE_TITLE", MAX_PAGE_TITLE);
		map.put("TABLE_YEAR_HIERARCHY", TABLE_YEAR_HIERARCHY);

		map.put("KBT_MULTIPLIED_GET_HIERARICAL_YEAR", new MultipliedScores(kbt, GET_HIERARICAL_YEAR, false));
		map.put("KBT_MULTIPLIED_GET_FULL_HIERARICAL_YEAR", new MultipliedScores(kbt, GET_FULL_HIERARICAL_YEAR, false));
		map.put("KBT_MULTIPLIED_MAX_TABLE_TILE", new MultipliedScores(kbt, MAX_TABLE_TILE, false));
		map.put("KBT_MULTIPLIED_MAX_COLUMN_YEAR", new MultipliedScores(kbt, MAX_COLUMN_YEAR, false));
		map.put("KBT_MULTIPLIED_MOST_COMMON_TABLE_TEXT_YEAR", new MultipliedScores(kbt, MOST_COMMON_TABLE_TEXT_YEAR, false));
		map.put("KBT_MULTIPLIED_MAX_PAGE_TITLE", new MultipliedScores(kbt, MAX_PAGE_TITLE, false));
		map.put("KBT_MULTIPLIED_TABLE_YEAR_HIERARCHY", new MultipliedScores(kbt, TABLE_YEAR_HIERARCHY, false));

		RegressionTreesScoresClassPropertyId rs = regressionScorer.getClassPropertyId(learningSet, map);

		// for (Entry<String, Double> e : rs.getWeights().entrySet()) {
		// System.out.println(e.getKey() + ": " + e.getValue());
		// }
		return rs;
	}
}
