package de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.timestamps;

import javax.inject.Inject;

import de.mannheim.uni.expansion.tablefusion.core.matching.GroupedValues;
import de.mannheim.uni.expansion.tablefusion.core.scoring.ProbabilitiesCalculator;
import de.mannheim.uni.expansion.tablefusion.core.scoring.TimeStampsScoringCalculator;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.ScoresContainer;
import de.mannheim.uni.expansion.tablefusion.core.scoring.packs.WeightedScores;
import de.mannheim.uni.expansion.tablefusion.services.fusion.strategies.Scorer;

public class Scorer_KBT_Combined_HierarchyTS_TableText_Weighted implements Scorer {

	private final ProbabilitiesCalculator probabilitiesCalculator;
	private final TimeStampsScoringCalculator timeStampsScoringCalculator;

	@Inject
	public Scorer_KBT_Combined_HierarchyTS_TableText_Weighted(ProbabilitiesCalculator probabilitiesCalculator,
			TimeStampsScoringCalculator timeStampsScoringCalculator) {
		this.probabilitiesCalculator = probabilitiesCalculator;
		this.timeStampsScoringCalculator = timeStampsScoringCalculator;
	}

	@Override
	public ScoresContainer getScores(GroupedValues fullSet, GroupedValues learningSet, GroupedValues testingSet) {

		ScoresContainer[] containers = new ScoresContainer[3];
		Double[] weights = new Double[3];
		containers[0] = timeStampsScoringCalculator.genericYearScorer(fullSet, timeStampsScoringCalculator.GET_HIERARICAL_YEAR);
		weights[0] = 0.25;

		containers[1] = timeStampsScoringCalculator.genericYearScorer(fullSet, timeStampsScoringCalculator.MOST_COMMON_TABLE_TEXT_YEAR);
		weights[1] = 0.25;

		containers[2] = probabilitiesCalculator.byProperty(learningSet);
		weights[2] = 0.5;

		return new WeightedScores(containers, weights);
	}

}
