package de.mannheim.uni.expansion.tablefusion.services.grouping;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Consumer;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.TripleId;
import de.mannheim.uni.expansion.data.ontology.model.instances.QualifierId;
import de.mannheim.uni.expansion.data.ontology.model.repository.OntologyRepository;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTable;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableId;
import de.mannheim.uni.expansion.data.tables.model.AnnotatedTableMetaData;
import de.mannheim.uni.expansion.data.tables.model.ColumnIdx;
import de.mannheim.uni.expansion.data.tables.repository.AnnotatedTablesRepository;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchCreator;
import de.mannheim.uni.expansion.tablefusion.core.matching.MatchedValues;
import de.mannheim.uni.expansion.tablefusion.core.matching.SynchronizedMatchedValuesMapMerger;
import de.mannheim.uni.expansion.tablefusion.evaluation.analyzer.FusionConfig;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.tableprocessor.model.Table;
import de.mannheim.uni.tableprocessor.model.TableColumn.ColumnDataType;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.ValueType;

public class BasicMatcherService {

	private final AnnotatedTablesRepository annotatedTableData;
	private final OntologyRepository ontologyRepository;

	@Inject
	public BasicMatcherService(AnnotatedTablesRepository annotatedTableData, OntologyRepository ontologyRepository) {
		this.annotatedTableData = annotatedTableData;
		this.ontologyRepository = ontologyRepository;

	}

	public enum MatchTypes {
		ALL, NUMERIC, STRING;
	}

	public Map<TripleId, MatchedValues> matchAndGroup(final MatcherOptionsBuilder options, FusionConfig fusionConfig) {
		return matchAndGroup(options, fusionConfig, annotatedTableData.getAllTableIds());
	}

	public Map<TripleId, MatchedValues> matchAndGroup(final MatcherOptionsBuilder options, FusionConfig fusionConfig, Set<AnnotatedTableId> tableIds) {

		boolean isDefaultMatcher;

		MatchCreator defaultMatcher = options.getDefaultMatcher();
		if (defaultMatcher == null)
			isDefaultMatcher = false;
		else {
			isDefaultMatcher = true;
		}

		boolean isReferenceMatcher;

		MatchCreator referenceMatcher = options.getReferenceMatcher();
		if (referenceMatcher == null)
			isReferenceMatcher = false;
		else {
			isReferenceMatcher = true;
		}

		Map<ValueType, MatchCreator> matchers = options.getMatchers();

		SynchronizedMatchedValuesMapMerger<TripleId> mapMerger = new SynchronizedMatchedValuesMapMerger<>();

		// final ConcurrentLinkedQueue<Map<TripleId, MatchedValues>> listOfMaps
		// = new ConcurrentLinkedQueue<Map<TripleId, MatchedValues>>();

		try {
			PersistentParallel6.runAndFinish(tableIds, new Consumer<AnnotatedTableId>() {

				@Override
				public void accept(AnnotatedTableId id) {
					// if
					// (!id.toString().equals("4444204_0_951426790527152756")) {
					// return;
					// }

					// if (i > 20) {
					// break;
					// }

					AnnotatedTable t = annotatedTableData.getTable(id);
					AnnotatedTableMetaData metaData = annotatedTableData.getMetaData(id);

					Map<Integer, ValueType> allPropertyColumns = getDBPediaPropertyColumns(t, metaData);

					for (Entry<Integer, ValueType> column : allPropertyColumns.entrySet()) {

						MatchCreator matcher = null;

						if (false && isReferenceMatcher && column.getValue() == ValueType.STRING
								&& ontologyRepository.isReferenceField(metaData.getProperties().get(column.getKey()))) {
							if (isReferenceMatcher) {
								matcher = referenceMatcher;
							}
							// else {
							// continue;
							// }
						}

						else if (matchers.containsKey(column.getValue())) {

							matcher = matchers.get(column.getValue());

						} else if (isDefaultMatcher) {
							matcher = defaultMatcher;

						} else if (column.getValue() != ValueType.URL) {
							continue;
						} else {
							continue;
						}

						Map<TripleId, MatchedValues> matchedColumns = matcher.matchColumns(id, new ColumnIdx(column.getKey()));

						mapMerger.merge(matchedColumns);

					}

				}

			}, "BasicMatcherService");
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		if (fusionConfig != null && fusionConfig.getMinSource() > 1) {
			ConcurrentHashMap<TripleId, MatchedValues> retMap = new ConcurrentHashMap<>();
			final int minSource = fusionConfig.getMinSource();
			PersistentParallel6.runAndFinish(mapMerger.getMap().entrySet(), (e) -> {

				Set<String> PLDs = e.getValue().getValues().stream().map((sv) -> {

					AnnotatedTableId atid = sv.getId().getTableId();
					AnnotatedTableMetaData md = annotatedTableData.getMetaData(atid);
					String PLD = md.getSource().getPLD();
					// String PLD =
					// DomainUtils.getPayLevelDomainFromWholeURL(source);

					return PLD;
				}).collect(Collectors.toSet());

				if (PLDs.size() >= minSource) {
					retMap.put(e.getKey(), e.getValue());
				}

			}, this.getClass());
			return retMap;
		} else {
			return mapMerger.getMap();
		}

	}

	// public Map<MatchIdentifiers, MatchedValues> matchAndGroup(final
	// MatchTypes matchType) {
	//
	// final ConcurrentLinkedQueue<Map<MatchIdentifiers, MatchedValues>>
	// listOfMaps = new ConcurrentLinkedQueue<Map<MatchIdentifiers,
	// MatchedValues>>();
	//
	// try {
	// new
	// Parallel<AnnotatedTableId>(4).foreach(annotatedTableData.getAllTableIds(),
	// new Consumer<AnnotatedTableId>() {
	//
	// @Override
	// public void execute(AnnotatedTableId id) {
	// // if
	// // (!id.toString().equals("4444204_0_951426790527152756")) {
	// // return;
	// // }
	//
	// // if (i > 20) {
	// // break;
	// // }
	//
	// Table t = annotatedTableData.getData(id);
	//
	// Set<Integer> allPropertyColumns = getDBPediaPropertyColumns(t,
	// annotatedTableData.getMetaData(id)).keySet();
	//
	// // ======= REFERENCE COLUMNS
	// // Get a set of columns to match
	// // Def. of Column Matching:
	// // == values in a column are matched to instances in DBpedia
	// Set<Integer> columnsToMatch = getStringColumnIdx(t,
	// annotatedTableData.getMetaData(id), allPropertyColumns);
	//
	// if (matchType == MatchTypes.ALL || matchType == MatchTypes.STRING) {
	//
	// // Match and Group Reference Columns
	// Map<MatchIdentifiers, MatchedValues> matchedColumns =
	// columnMatcher.matchColumns(id, columnsToMatch);
	//
	// listOfMaps.add(matchedColumns);
	// }
	//
	// // ======= LITERAL COLUMNS
	// // Mainly literals, which are simply grouped
	// Set<Integer> columnsToGroup = new HashSet<Integer>(allPropertyColumns);
	// columnsToGroup.removeAll(columnsToMatch);
	//
	// if (matchType == MatchTypes.ALL || matchType == MatchTypes.NUMERIC) {
	//
	// Map<MatchIdentifiers, MatchedValues> groupedMatches =
	// groupMatcher.matchColumns(id, columnsToGroup);
	//
	// listOfMaps.add(groupedMatches);
	//
	// }
	//
	// // map = MatchCreator.merge(map, groupedMatches);
	//
	// }
	//
	// });
	// } catch (Exception e) {
	// // TODO Auto-generated catch block
	// e.printStackTrace();
	// }
	//
	// Map<MatchIdentifiers, MatchedValues> map = new HashMap<MatchIdentifiers,
	// MatchedValues>();
	//
	// for (Map<MatchIdentifiers, MatchedValues> matches : listOfMaps) {
	// map = MatchCreator.merge(map, matches);
	// }
	//
	// return map;
	//
	// }

	/***
	 * Get the set of Columns that are matched to DBPedia Properties excluding
	 * the KeyColumn
	 * 
	 * @param t
	 * @param metaData
	 * @return
	 */
	private Map<Integer, ValueType> getDBPediaPropertyColumns(AnnotatedTable t, AnnotatedTableMetaData metaData) {
		Map<Integer, ValueType> columnsToMatch = new HashMap<Integer, ValueType>();
		for (int i = 0; i < t.getColumns().size(); i++) {

			// Skip keys - key columns are already matched
			if (t.getKeyColumnIndex().equals(new ColumnIdx(i))) {
				continue;
			}

			// Skip if column is not matched to DBpedia
			if (!metaData.getProperties().containsKey(i) || metaData.getProperties().get(i) == null) {
				continue;
			}

			if (metaData.getProperties().get(i) instanceof QualifierId)
				continue;

			columnsToMatch.put(i, t.getDataType(new ColumnIdx(i)));
		}

		return columnsToMatch;
	}

	/***
	 * Finds all String columns that are matched to DBpedia and returns their
	 * indices
	 * 
	 * @param t
	 * @param pre
	 *            filtered set of columns
	 * @return Set of integer indices
	 */
	private Set<Integer> getStringColumnIdx(Table t, AnnotatedTableMetaData metaData, Set<Integer> allPropertyColumns) {
		Set<Integer> columnsToMatch = new HashSet<Integer>();

		for (Integer i : allPropertyColumns) {
			ColumnDataType dt = t.getColumns().get(i).getDataType();
			if (dt.equals(ColumnDataType.string))
				columnsToMatch.add(i);

		}

		return columnsToMatch;

	}

	/***
	 * Finds all Numeric columns that are matched to DBpedia and returns their
	 * indices
	 * 
	 * @param t
	 * @param pre
	 *            filtered set of columns
	 * @return Set of integer indices
	 */
	private Set<Integer> getNumericColumnIndeces(Table t, AnnotatedTableMetaData metaData, Set<Integer> allPropertyColumns) {
		Set<Integer> columnsToMatch = new HashSet<Integer>();

		for (Integer i : allPropertyColumns) {
			ColumnDataType dt = t.getColumns().get(i).getDataType();
			if (dt.equals(ColumnDataType.numeric))
				columnsToMatch.add(i);
		}

		return columnsToMatch;

	}

}
