package de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets;

import java.io.IOException;
import java.util.HashSet;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.document.StoredField;
import org.apache.lucene.document.TextField;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.queryparser.classic.ParseException;
import org.apache.lucene.queryparser.classic.QueryParser;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.TopDocs;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.index.io.InMemoryIndex;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats.Stats;

public class KBClassIndex {

	public static final String INSTANCE_ID = "instanceId";
	public static final String LABEL = "label";

	private final InMemoryIndex idx = new InMemoryIndex();

	private final QueryParser queryParser = new QueryParser(LABEL, new StandardAnalyzer());

	public KBClassIndex(KnowledgeBase kb, ClassId classId) throws IOException {
		IndexWriter w = idx.getIndexWriter();

		for (Entry<InstanceId, Integer> entry : kb.getInstances(classId).entrySet()) {
			StringObject labelSO = (StringObject) kb.getQualifiedValueList(entry.getKey(), Stats.label).getClaims().get(0).getValue();

			if (labelSO != null && labelSO.getString() != null) {

				Document doc = createDocument(labelSO.getString(), entry.getKey());

				w.addDocument(doc);

			} else {
				System.out.println("NULL LABEL FOR: " + entry.getKey());
			}

		}

		idx.closeIndexWriter();
	}

	public Set<InstanceId> getMatches(String search) throws IOException, ParseException {
		Query q = queryParser.parse(search);
		IndexSearcher indexSearcher = this.idx.getIndexSearcher();
		// new TermQuery(new Term(LABEL, search));
		TopDocs results = indexSearcher.search(q, 30);
		Set<InstanceId> instanceIds = new HashSet<>();
		for (ScoreDoc doc : results.scoreDocs) {
			instanceIds.add(fromDocument(indexSearcher.doc(doc.doc)));
		}

		return instanceIds;

	}

	public InstanceId fromDocument(Document doc) {
		String s = doc.getField(INSTANCE_ID).stringValue();
		return new InstanceId(s, doc.getField(LABEL).stringValue());

	}

	public Document createDocument(String label, InstanceId instanceId) {
		Document doc = new Document();
		doc.add(new StoredField(INSTANCE_ID, instanceId.getInstanceId()));
		doc.add(new TextField(LABEL, label, Field.Store.YES));

		return doc;
	}
}
