package de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.nba;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;
import java.util.stream.Collectors;

import org.apache.commons.lang3.StringUtils;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.document.StoredField;
import org.apache.lucene.document.TextField;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.queryparser.classic.ParseException;
import org.apache.lucene.queryparser.classic.QueryParser;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.TopDocs;

import com.google.common.collect.BiMap;

import au.com.bytecode.opencsv.CSVReader;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.tablefusion.core.similarity.measures.StringSimilarity;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityMeasures;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThreshold;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThresholdImpl;
import de.mannheim.uni.expansion.tablefusion.core.similarity.thresholds.SimilarityThresholdsRepository;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MapSorter;
import de.mannheim.uni.index.io.InMemoryIndex;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueContainer;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject.Precision;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedAbstractObjectHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.Qualifier;
import de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.KBClassIndex;
import de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.nba.NBACrawler.PlayerData;
import de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.nba.NBACrawler.PlayerId;
import de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.nba.NBACrawler.TeamId;
import de.uni_mannheim.informatik.dws.expansion.wikidata.WikidataClasses;
import de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats.Stats;

public class NBAConverter implements KBValueCreator {

	private final String path;

	public NBAConverter(String path) {
		this.path = path;
	}

	private static final String PLAYER_FILE = "players";
	private static final String TEAMS_FILE = "teams";
	private static final String NUMBERS_FILE = "numbers";
	private static final String TEAMS_NAMES_FILE = "teams_names";

	private static final PropertyId team = Stats.teams;
	private static final PropertyId sportsNumber = Stats.sportNumber;
	//
	// private static final Map<TeamId, InstanceId> teamMap = new HashMap<>();
	//
	// static {
	//
	// teamMap.put(new TeamId("cincinnati-bengals"), new InstanceId("Q223511",
	// "Cincinnati Bengals"));
	// teamMap.put(new TeamId("new-orleans-saints"), new InstanceId("Q172435",
	// "New Orleans Saints"));
	// teamMap.put(new TeamId("atlanta-falcons"), new InstanceId("Q272059",
	// "Atlanta Falcons"));
	// teamMap.put(new TeamId("houston-texans"), new InstanceId("Q223514",
	// "Houston Texans"));
	// teamMap.put(new TeamId("baltimore-ravens"), new InstanceId("Q276539",
	// "Baltimore Ravens"));
	// teamMap.put(new TeamId("philadelphia-eagles"), new InstanceId("Q219714",
	// "Philadelphia Eagles"));
	// teamMap.put(new TeamId("indianapolis-colts"), new InstanceId("Q193753",
	// "Indianapolis Colts"));
	// teamMap.put(new TeamId("washington-redskins"), new InstanceId("Q212654",
	// "Washington Redskins"));
	// teamMap.put(new TeamId("tennessee-titans"), new InstanceId("Q320484",
	// "Tennessee Titans"));
	// teamMap.put(new TeamId("jacksonville-jaguars"), new InstanceId("Q272223",
	// "Jacksonville Jaguars"));
	// teamMap.put(new TeamId("new-york-jets"), new InstanceId("Q219602", "New
	// York Jets"));
	// teamMap.put(new TeamId("carolina-panthers"), new InstanceId("Q330120",
	// "Carolina Panthers"));
	// teamMap.put(new TeamId("chicago-bears"), new InstanceId("Q205033",
	// "Chicago Bears"));
	// teamMap.put(new TeamId("seattle-seahawks"), new InstanceId("Q221878",
	// "Seattle Seahawks"));
	// teamMap.put(new TeamId("new-england-patriots"), new InstanceId("Q193390",
	// "New England Patriots"));
	// teamMap.put(new TeamId("dallas-cowboys"), new InstanceId("Q204862",
	// "Dallas Cowboys"));
	// teamMap.put(new TeamId("san-diego-chargers"), new InstanceId("Q272220",
	// "San Diego Chargers"));
	// teamMap.put(new TeamId("minnesota-vikings"), new InstanceId("Q221150",
	// "Minnesota Vikings"));
	// teamMap.put(new TeamId("arizona-cardinals"), new InstanceId("Q224164",
	// "Arizona Cardinals"));
	// teamMap.put(new TeamId("detroit-lions"), new InstanceId("Q271880",
	// "Detroit Lions"));
	// teamMap.put(new TeamId("los-angeles-rams"), new InstanceId("Q337377",
	// "Los Angeles Rams"));
	// teamMap.put(new TeamId("green-bay-packers"), new InstanceId("Q213837",
	// "Green Bay Packers"));
	// teamMap.put(new TeamId("buffalo-bills"), new InstanceId("Q221626",
	// "Buffalo Bills"));
	// teamMap.put(new TeamId("san-francisco-49ers"), new InstanceId("Q337758",
	// "San Francisco 49ers"));
	// teamMap.put(new TeamId("miami-dolphins"), new InstanceId("Q223243",
	// "Miami Dolphins"));
	// teamMap.put(new TeamId("tampa-bay-buccaneers"), new InstanceId("Q320476",
	// "Tampa Bay Buccaneers"));
	// teamMap.put(new TeamId("denver-broncos"), new InstanceId("Q223507",
	// "Denver Broncos"));
	// teamMap.put(new TeamId("oakland-raiders"), new InstanceId("Q324523",
	// "Oakland Raiders"));
	// teamMap.put(new TeamId("kansas-city-chiefs"), new InstanceId("Q223522",
	// "Kansas City Chiefs"));
	// teamMap.put(new TeamId("cleveland-browns"), new InstanceId("Q223527",
	// "Cleveland Browns"));
	// teamMap.put(new TeamId("new-york-giants"), new InstanceId("Q190618", "New
	// York Giants"));
	// teamMap.put(new TeamId("pittsburgh-steelers"), new InstanceId("Q191477",
	// "Pittsburgh Steelers"));
	// }

	public static void main(String[] args) {
		// for (Entry<TeamId, InstanceId> entry : teamMap.entrySet()) {
		// InputStream in = null;
		// try {
		// in = new URL("https://www.wikidata.org/wiki/" +
		// entry.getValue().getInstanceId()).openStream();
		//
		// String input = IOUtils.toString(in);
		//
		// org.jsoup.nodes.Document doc = Jsoup.parse(input);
		// Elements links = doc.select(".wikibase-title-label");
		//
		// Element link = links.get(0);
		// String title = link.text().trim();
		// System.out.println(entry.getKey().getMachineName() + ", " +
		// entry.getValue().getReadableString() + ", " + title);
		//
		// } catch (IOException e) {
		// // TODO Auto-generated catch block
		// e.printStackTrace();
		// } finally {
		// IOUtils.closeQuietly(in);
		// }
		//
		// }
	}

	@Override
	public KBValueContainer getMatchedValues(KnowledgeBase kb) {
		try {

			Map<TeamId, InstanceId> teamMap = getTeamMatching(kb, path);

			Map<ValueId, List<QualifiedValueHolder>> retMap = new ConcurrentHashMap<ValueId, List<QualifiedValueHolder>>();
			Set<ValueId> remove = new HashSet<>();

			Map<PlayerId, PlayerData> playerData = new HashMap<>();

			for (int c = 0; Files.exists(Paths.get(path + PLAYER_FILE + "_" + c + "_.csv")); c++) {
				playerData.putAll(getPlayerData(path + PLAYER_FILE + "_" + c + "_.csv"));

			}

			KBClassIndex nbaIndex = new KBClassIndex(kb, WikidataClasses.wikidataBasketball);
			Map<InstanceId, String> entityLabels = getEntityLabels(kb);

			Map<String, Set<InstanceId>> birthdateIndex = createBirthdayIndex(kb, entityLabels);

			Map<PlayerId, InstanceId> playerMatching = getPlayerMatching(playerData, kb, entityLabels, birthdateIndex, nbaIndex);

			for (int c = 0; Files.exists(Paths.get(path + TEAMS_FILE + "_" + c + "_.csv")); c++) {
				Map<ValueId, List<QualifiedValueHolder>> mmm = readFromNBACSV(path + TEAMS_FILE + "_" + c + "_.csv", playerMatching, team, (s) -> {
					return teamMap.get(new TeamId(s));
				});

				for (Entry<ValueId, List<QualifiedValueHolder>> e : mmm.entrySet()) {
					CHMHelper.getFromMap(retMap, e.getKey(), () -> new ArrayList<>()).addAll(e.getValue());
				}

			}

			// retMap.putAll(readFromNFLCSV(path + NUMBERS_FILE, playerMatching,
			// sportsNumber, (s) -> {
			// Double doubleV = Double.valueOf(s);
			//
			// QuantityObject val = new QuantityObject(doubleV, doubleV,
			// doubleV, "number");
			// return val;
			// }));

			for (Entry<InstanceId, Integer> instance : kb.getInstances(WikidataClasses.wikidataBasketball).entrySet()) {
				ValueId teamValueId = new ValueId(instance.getKey(), NBAConverter.team);
				if (!retMap.containsKey(teamValueId)) {
					remove.add(teamValueId);
				}

				ValueId sportsNumber = new ValueId(instance.getKey(), NBAConverter.sportsNumber);
				if (!retMap.containsKey(sportsNumber)) {
					remove.add(sportsNumber);
				}

			}

			return new KBValueContainer(retMap.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				return new QualifiedValueList(e.getValue());
			})), remove);
		} catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException(e);
		}
	}

	private Map<TeamId, InstanceId> getTeamMatching(KnowledgeBase kb, String path) throws IOException, ParseException {
		Map<TeamId, InstanceId> retMap = new HashMap<>();
		KBClassIndex nbaTeamsIndex = new KBClassIndex(kb, WikidataClasses.wikidataBasketballTeam);
		Map<InstanceId, String> labels = getEntityLabels(kb, WikidataClasses.wikidataBasketballTeam);

		CSVReader reader = new CSVReader(new FileReader(path + TEAMS_NAMES_FILE + ".csv"));

		List<String[]> allLines = reader.readAll();
		for (int i = 1; i < allLines.size(); i++) {
			String[] line = allLines.get(i);
			TeamId teamId = new TeamId(line[0].trim());
			String name = line[1].trim();
			Integer inception = null;
			if (line[2].trim().length() > 0)
				inception = Integer.parseInt(line[2].trim());

			Set<InstanceId> candidateas = nbaTeamsIndex.getMatches(name);

			LinkedHashMap<InstanceId, Double> scored = getScoresMatches(name, labels, candidateas);

			if (inception != null) {
				for (Entry<InstanceId, Double> entry : scored.entrySet()) {
					QualifiedValueList inceptionQVL = kb.getQualifiedValueList(new ValueId(entry.getKey(), new PropertyId("P571")));
					if (inceptionQVL != null && inceptionQVL.size() > 0) {
						TimeObject qo = (TimeObject) inceptionQVL.getClaim(0).getValue();
						if (qo.getYear() == inception) {
							entry.setValue(entry.getValue() * 2.0);
						}

					}

				}
			}

			scored = MapSorter.sortHashMapByValuesD(scored);

			if (scored.size() == 0) {
				System.out.println("NO MATCH: " + name);
			}

			for (Entry<InstanceId, Double> e : scored.entrySet()) {
				if (e.getValue() > 0.9) {
					if (retMap.containsKey(e.getKey())) {
						System.out.println("ALREADY MATCHED: " + name);
					} else {
						retMap.put(teamId, e.getKey());
					}

				} else {
					System.out.println("NO MATCH: " + name);

				}

				break;
			}

		}

		reader.close();

		return retMap;
	}

	private Map<PlayerId, InstanceId> getPlayerMatching(Map<PlayerId, PlayerData> playerData, KnowledgeBase kb, Map<InstanceId, String> entityLabels,
			Map<String, Set<InstanceId>> birthdateIndex, KBClassIndex index) throws IOException, ParseException {
		int notMatch = 0;
		Set<InstanceId> hasBirthday = birthdateIndex.entrySet().stream().map((e) -> {
			return e.getValue();
		}).flatMap(set -> set.stream()).collect(Collectors.toSet());

		Map<PlayerId, InstanceId> retMap = new HashMap<>();
		for (Entry<PlayerId, PlayerData> player : playerData.entrySet()) {
			Set<InstanceId> birthdayCandidates = birthdateIndex.get(player.getValue().getBirthDate());
			Set<InstanceId> candidateas = index.getMatches(player.getValue().getName());
			if (birthdayCandidates != null) {

				for (Iterator<InstanceId> it = candidateas.iterator(); it.hasNext();) {
					InstanceId instanceId = it.next();
					if (hasBirthday.contains(instanceId) && !birthdayCandidates.contains(instanceId)) {
						it.remove();
					}
				}

				candidateas.addAll(birthdayCandidates);
			}

			LinkedHashMap<InstanceId, Double> scored = getScoresMatches(player.getValue().getName(), entityLabels, candidateas);

			if (birthdayCandidates != null) {
				for (Entry<InstanceId, Double> entry : scored.entrySet()) {
					if (birthdayCandidates.contains(entry.getKey()) && candidateas.contains(entry.getKey())) {
						entry.setValue(entry.getValue() * 2.0);
					}
				}
			}

			scored = MapSorter.sortHashMapByValuesD(scored);

			for (Entry<InstanceId, Double> e : scored.entrySet()) {
				if (e.getValue() > 0.9) {
					if (retMap.containsKey(e.getKey())) {
						// System.out.println("ALREADY MATCHED: " +
						// player.toString());
					} else {
						retMap.put(player.getKey(), e.getKey());
					}

				} else {
					// System.out.println("NO MATCH: " + player.toString());
					notMatch++;
				}

				break;
			}

		}
		System.out.println("Not Matched: " + notMatch);
		return retMap;
	}

	private final StringSimilarity stringSimlarity = new StringSimilarity(new SimilarityThresholdsRepository() {

		@Override
		public SimilarityThreshold get(SimilarityMeasures sm) {
			return new SimilarityThresholdImpl(0.45);
		}
	});

	private Map<InstanceId, String> getEntityLabels(KnowledgeBase kb) {

		return getEntityLabels(kb, WikidataClasses.wikidataBasketball);
	}

	private Map<InstanceId, String> getEntityLabels(KnowledgeBase kb, ClassId classId) {
		HashMap<InstanceId, String> retMap = new HashMap<>();

		for (Entry<InstanceId, Integer> entry : kb.getInstances(classId).entrySet()) {
			StringObject labelSO = (StringObject) kb.getQualifiedValueList(entry.getKey(), Stats.label).getClaims().get(0).getValue();

			retMap.put(entry.getKey(), labelSO.getString());
		}

		return retMap;
	}

	public LinkedHashMap<InstanceId, Double> getScoresMatches(String string, Map<InstanceId, String> entityLabels, Set<InstanceId> instanceIds) {
		Map<InstanceId, Double> retMap = new HashMap<>();
		for (InstanceId instanceId : instanceIds) {
			double score = 0.0;
			if (entityLabels.containsKey(instanceId) && entityLabels.get(instanceId) != null) {

				score = this.stringSimlarity.get(new StringObject(entityLabels.get(instanceId)), new StringObject(string));

			}
			retMap.put(instanceId, score);
		}

		return MapSorter.sortHashMapByValuesD(retMap);
	}

	public static class NFLIndex {

		public static final String INSTANCE_ID = "instanceId";
		public static final String LABEL = "label";

		private final InMemoryIndex idx = new InMemoryIndex();

		private final QueryParser queryParser = new QueryParser(LABEL, new StandardAnalyzer());

		public NFLIndex(KnowledgeBase kb) throws IOException {
			IndexWriter w = idx.getIndexWriter();

			for (Entry<InstanceId, Integer> entry : kb.getInstances(WikidataClasses.wikidataBasketball).entrySet()) {
				StringObject labelSO = (StringObject) kb.getQualifiedValueList(entry.getKey(), Stats.label).getClaims().get(0).getValue();

				Document doc = createDocument(labelSO.getString(), entry.getKey());

				w.addDocument(doc);

			}

			idx.closeIndexWriter();
		}

		Set<InstanceId> getMatches(String search) throws IOException, ParseException {
			Query q = queryParser.parse(search);
			IndexSearcher indexSearcher = this.idx.getIndexSearcher();
			// new TermQuery(new Term(LABEL, search));
			TopDocs results = indexSearcher.search(q, 30);
			Set<InstanceId> instanceIds = new HashSet<>();
			for (ScoreDoc doc : results.scoreDocs) {
				instanceIds.add(fromDocument(indexSearcher.doc(doc.doc)));
			}

			return instanceIds;

		}

		public InstanceId fromDocument(Document doc) {
			String s = doc.getField(INSTANCE_ID).stringValue();
			return new InstanceId(s, doc.getField(LABEL).stringValue());

		}

		public Document createDocument(String label, InstanceId instanceId) {
			Document doc = new Document();
			doc.add(new StoredField(INSTANCE_ID, instanceId.getInstanceId()));
			doc.add(new TextField(LABEL, label, Field.Store.YES));

			return doc;
		}
	}

	private Map<String, Set<InstanceId>> createBirthdayIndex(KnowledgeBase kb, Map<InstanceId, String> entityLabels) {
		HashMap<String, Set<InstanceId>> retMap = new HashMap<>();

		for (Entry<InstanceId, Integer> entry : kb.getInstances(WikidataClasses.wikidataBasketball).entrySet()) {
			QualifiedValueList qvl = kb.getQualifiedValueList(entry.getKey(), Stats.dateOfBirth);
			if (qvl != null) {
				TimeObject date = (TimeObject) qvl.getClaims().get(0).getValue();
				String dateString = date.getYear() + "-" + StringUtils.leftPad(String.valueOf(date.getMonth()), 2, '0') + "-"
						+ StringUtils.leftPad(String.valueOf(date.getDay()), 2, '0');
				CHMHelper.getFromMap(retMap, dateString, () -> new HashSet<>())
						.add(new InstanceId(entry.getKey().getInstanceId(), entityLabels.get(entry.getKey())));
			}
		}

		return retMap;
	}

	private Map<PlayerId, PlayerData> getPlayerData(String playerFile) throws IOException {
		CSVReader reader = new CSVReader(new FileReader(playerFile));

		try {

			Map<PlayerId, PlayerData> retMap = new HashMap<>();
			List<String[]> lines = reader.readAll();

			// String[] headerLine = lines.get(0);

			for (int i = 1; i < lines.size(); i++) {
				String[] line = lines.get(i);
				PlayerId playerId = new PlayerId(line[0]);
				PlayerData playerData = new PlayerData(line[1], line[2], line[3]);
				retMap.put(playerId, playerData);
			}

			return retMap;
		} finally {
			reader.close();
		}
	}

	private Map<String, InstanceId> matchPlayersToKB(Set<String> isoCodes, KnowledgeBase kb) {
		Map<String, InstanceId> retMap = new HashMap<>();
		BiMap<InstanceId, Integer> instances = kb.getInstances(WikidataClasses.wikidataCountry);

		for (InstanceId instanceId : instances.keySet()) {
			ValueId valueId = new ValueId(instanceId, Stats.isoAlpha3Codes);
			QualifiedValueList qvl = kb.getQualifiedValueList(valueId);
			if (qvl != null) {
				QualifiedValueHolder qvh = qvl.getClaim(0);
				if (qvh != null) {
					AbstractObject ao = qvh.getValue();
					if (ao instanceof StringObject) {
						StringObject so = (StringObject) ao;
						String code = so.getString().trim();
						if (isoCodes.contains(code)) {
							retMap.put(code, instanceId);
						}
					}
				}
			} else {
				System.out.println();
			}
		}

		for (String code : isoCodes) {
			if (!retMap.containsKey(code)) {
				System.out.println(code);
			}
		}

		return retMap;
	}

	private Set<String> getIsoCodes(String file) throws IOException {
		BufferedReader br = new BufferedReader(new FileReader(file));
		try {
			Set<String> isoCodes = new HashSet<>();
			for (String line = br.readLine(); line != null; line = br.readLine()) {
				isoCodes.add(line.trim());
			}

			return isoCodes;
		} finally {
			br.close();
		}
	}

	private Map<ValueId, List<QualifiedValueHolder>> readFromNBACSV(String file, Map<PlayerId, InstanceId> playerMatching, PropertyId propertyId,
			Function<String, AbstractObject> valueGetter) throws IOException {
		Map<ValueId, List<QualifiedValueHolder>> retMap = new HashMap<>();

		CSVReader csv = new CSVReader(new FileReader(file), ',', '"', 0);
		try {
			List<String[]> lines = csv.readAll();

			// Map<Integer, List<Qualifier>> columnToYearMap = new HashMap<>();
			// String[] headerLine = lines.get(0);
			// for (int i = 1; i < headerLine.length; i++) {
			// String header = headerLine[i].trim();
			// if (header.length() > 0) {
			// List<Qualifier> qualifierList = new ArrayList<>();
			// int year = Integer.valueOf(headerLine[i].trim());
			// Qualifier qualifier = new Qualifier(Stats.pointInTime,
			// TimeObject.fromYear(year));
			// qualifierList.add(qualifier);
			// columnToYearMap.put(i, qualifierList);
			// }
			// }

			for (int r = 1; r < lines.size(); r++) {
				String[] line = lines.get(r);
				PlayerId key = new PlayerId(line[0].trim());

				if (playerMatching.containsKey(key)) {
					InstanceId instanceid = playerMatching.get(key);

					AbstractObject object = valueGetter.apply(line[3].trim());
					if (object != null) {
						TimeObject from = getTimeObject(line[1].trim());
						TimeObject to = getTimeObject(line[2].trim());

						List<QualifiedValueHolder> claims = new ArrayList<>();
						List<Qualifier> qualifiers = new ArrayList<>();
						if (from != null) {
							qualifiers.add(new Qualifier(Stats.startTimeQualifier, from));
						}
						if (to != null) {
							qualifiers.add(new Qualifier(Stats.endTimeQualifier, to));
						}

						CHMHelper.getFromMap(retMap, new ValueId(instanceid, propertyId), () -> new ArrayList<>())
								.add(new QualifiedAbstractObjectHolder(object, qualifiers));

					}
				}
			}

			return retMap;
		} finally {
			csv.close();
		}

	}

	private TimeObject getTimeObject(String trim) {
		if (trim.length() == 0)
			return null;
		String[] t = trim.split("-");
		Precision precision = Precision.T11_days;
		if (t[2].equals("0")) {
			precision = Precision.T10_months;
		} else if (t[1].equals("0")) {
			precision = Precision.T09_years;
		}

		int year = Integer.parseInt(t[0]);
		int month = Integer.parseInt(t[1]);
		int day = Integer.parseInt(t[2]);
		return new TimeObject(year, month, day, 0, 0, 0, 0, precision);

	}

}
