package de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.nba;

import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;
import java.util.stream.Collectors;

import org.apache.commons.io.IOUtils;
import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;
import org.jsoup.select.Elements;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableMap.Builder;

import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.expansion.utils.containers.Tri;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject.Precision;

public class NBACrawler {

	public static final String basePath = "http://www.basketball-reference.com";

	public static final Map<String, Integer> months;

	private static int today = 2016;

	static {
		Builder<String, Integer> builder = ImmutableMap.builder();

		builder.put("January", 1);
		builder.put("February", 2);
		builder.put("March", 3);
		builder.put("April", 4);
		builder.put("May", 5);
		builder.put("June", 6);
		builder.put("July", 7);
		builder.put("August", 8);
		builder.put("September", 9);
		builder.put("October", 10);
		builder.put("November", 11);
		builder.put("December", 12);

		months = builder.build();

	}

	public static void main(String[] args) {
		try {
			NBACrawler crawler = new NBACrawler();

			LinkedList<String> playerListsUrls = new LinkedList<String>(crawler.getPlayerListsUrls());
			LinkedList<Pair<PlayerId, String>> playerUrls = crawler.getPlayerUrls(playerListsUrls);

			int remove = 0;
			for (int i = 0; i < remove; i++) {
				playerUrls.pop();
			}

			int c = 0;
			while (!playerUrls.isEmpty()) {

				LinkedList<Pair<PlayerId, String>> playerListSet = new LinkedList<>();
				for (int i = 0; i < playerUrls.size() && i < 500; i++) {
					playerListSet.add(playerUrls.pop());
				}

				Map<PlayerId, Pair<PlayerData, List<TeamMembership>>> playerData = crawler.getPlayerData(playerListSet);

				String outputPath = "C:/Users/yaser/SampleRuns/Wikidata/datasets/NBA_ATHLETE/";

				crawler.playerWrite(outputPath + "players_" + c + "_.csv",
						playerData.entrySet().parallelStream().collect(Collectors.toMap(Entry::getKey, (e) -> {
							return e.getValue().getFirst();
						})));

				crawler.write(outputPath + "teams_" + c + "_.csv", playerData.entrySet().parallelStream().collect(Collectors.toMap(Entry::getKey, (e) -> {
					return e.getValue().getSecond();
				})));

				c++;

			}

		} catch (IOException e) {
			e.printStackTrace();
		}

	}

	private void write(String outputFile, Map<PlayerId, List<TeamMembership>> playerData) throws IOException {
		CSVWriter writer = new CSVWriter(new FileWriter(outputFile));
		int length = 4;
		{
			String[] header = new String[length];
			header[0] = "Player";
			header[1] = "From";
			header[2] = "To";
			header[3] = "Team";

			// header[1] = "Birthdate";
			// header[2] = "College";

			writer.writeNext(header);
		}

		for (Entry<PlayerId, List<TeamMembership>> playerEntry : playerData.entrySet()) {
			String[] line = new String[length];
			line[0] = playerEntry.getKey().getMachineName();

			for (TeamMembership membership : playerEntry.getValue()) {

				line[1] = convertDate(membership.getFrom());
				line[2] = convertDate(membership.getTo());
				line[3] = membership.getTeamId().getMachineName();

				writer.writeNext(line);
			}
		}

		writer.close();

	}

	public static String convertDate(TimeObject to) {
		if (to == null)
			return "";
		return "" + to.getYear() + "-" + to.getMonth() + "-" + to.getDay();
	}

	public void write(Integer beginYear, Integer endYear, String outputFile, Map<TeamId, String> teams, Map<PlayerId, PlayerData> playerDataMap,
			ConcurrentHashMap<PlayerId, Map<Integer, MembershipData>> playerMap, Function<MembershipData, String> valueGetter) throws IOException {
		CSVWriter writer = new CSVWriter(new FileWriter(outputFile));
		int length = 1 + (endYear - beginYear + 1);
		{
			String[] header = new String[length];
			header[0] = "Player";
			// header[1] = "Team";
			// header[2] = "Number";
			// header[3] = "Position";
			// header[1] = "Birthdate";
			// header[2] = "College";
			int hi = 1;
			for (int i = beginYear; i <= endYear; i++) {
				header[hi] = String.valueOf(i);
				hi++;
			}

			writer.writeNext(header);
		}
		try {
			for (Entry<PlayerId, PlayerData> playerEntry : playerDataMap.entrySet()) {
				PlayerData playerData = playerEntry.getValue();

				if (playerData != null) {
					String[] line = new String[length];
					line[0] = playerEntry.getKey().getMachineName();

					// line[1] = teams.get(playerData.getMachineTeamName());
					// line[2] = playerData.getNumber();
					// line[3] = playerData.getPosition();
					// line[1] = playerData.getBirthDate();
					// line[2] = playerData.getCollege();

					Map<Integer, MembershipData> membershipData = playerMap.get(playerEntry.getKey());

					int li = 1;
					for (int i = beginYear; i <= endYear; i++) {
						MembershipData md = membershipData.get(i);
						if (md != null) {
							line[li] = valueGetter.apply(md);
						} else {
							line[li] = "";
						}
						li++;
					}

					writer.writeNext(line);
				}

			}

		} finally {
			writer.close();
		}
	}

	private Map<PlayerId, Pair<PlayerData, List<TeamMembership>>> getPlayerData(LinkedList<Pair<PlayerId, String>> playerListsUrls) {

		Map<PlayerId, Pair<PlayerData, List<TeamMembership>>> retMap = new HashMap<>();
		// int c = 0;
		for (Pair<PlayerId, String> e : playerListsUrls) {
			// c++;
			// if (c == 50)
			// break;
			String input = NBACrawler.getDocument(e.getSecond());

			Document doc = Jsoup.parse(input);

			String name = getCleanHtml(doc.select("#info h1").get(0).html());
			Elements birthDates = doc.select("#necro-birth");
			String birthDate = "";
			if (birthDates.size() > 0)
				birthDate = birthDates.get(0).attr("data-birth");
			String college = null;
			Elements paragraphs = doc.select("#info p strong");
			for (int i = 0; i < paragraphs.size(); i++) {

				Element paragraph = paragraphs.get(i);
				String collegeLabel = getCleanHtml(paragraph.html());
				if (collegeLabel.equals("College:")) {
					college = getCleanHtml(paragraph.nextElementSibling().html());

				}
			}
			PlayerData playerData = new PlayerData(name, birthDate, college);

			List<Tri<TimeObject, TeamId, TeamId>> transactionDates = new ArrayList<>();
			Element transactionArea = doc.select("#all_transactions").get(0);
			String transactionAreaHtml = transactionArea.html();
			transactionAreaHtml = transactionAreaHtml.replace("<!--", "").replace("--> ", "");

			Elements transactions = Jsoup.parse(transactionAreaHtml).select(".transaction");
			for (int s = 0; s < transactions.size(); s++) {
				Element transaction = transactions.get(s);

				Elements links = transaction.select("a");
				Integer month = null;
				Integer day = null;
				Integer year = null;
				String date = getCleanHtml(transaction.select("strong").get(0).html());
				String[] dateSplits = date.split(" ");
				if (dateSplits.length < 2) {
					month = 0;
					// day = Integer.parseInt(dateSplits[1].substring(0,
					// dateSplits[1].length() - 1));
					year = Integer.parseInt(dateSplits[0]);
					day = 0;
				} else if (dateSplits.length < 3) {
					month = months.get(dateSplits[0].substring(0, dateSplits[0].length() - 1));
					// day = Integer.parseInt(dateSplits[1].substring(0,
					// dateSplits[1].length() - 1));
					year = Integer.parseInt(dateSplits[1]);
					day = 0;
				} else {
					month = months.get(dateSplits[0]);
					day = Integer.parseInt(dateSplits[1].substring(0, dateSplits[1].length() - 1));
					year = Integer.parseInt(dateSplits[2]);
				}

				TeamId fromTeam = null;
				TeamId toTeam = null;

				for (int l = 0; l < links.size(); l++) {
					String fromString = links.get(l).attr("data-attr-from");
					if (fromString.length() > 0) {
						fromTeam = new TeamId(fromString);
					} else {
						String toString = links.get(l).attr("data-attr-to");
						if (toString.length() > 0) {
							toTeam = new TeamId(toString);
						}
					}

				}
				TimeObject time = new TimeObject(year, month, day, 0, 0, 0, 0, Precision.T11_days);
				transactionDates.add(new Tri<>(time, fromTeam, toTeam));

			}

			Element table = doc.select("#div_per_game tbody").get(0);
			Elements rows = table.select("tr");
			TeamId currentTeam = null;
			TimeObject altBegin = null;
			TimeObject altEnd = null;
			TimeObject currentBegin = null;
			TimeObject currentEnd = null;
			List<TeamMembership> memberships = new ArrayList<>();

			for (int i = 0; i < rows.size(); i++) {
				Element row = rows.get(i);
				if (row.id().length() == 0)
					continue;
				String year = row.id().split("\\.")[1];
				int seasonEnd = Integer.parseInt(year);
				int seasonBegin = seasonEnd - 1;

				// Elements cells = row.getAllElements();
				// Element season = cells.get(0);

				TeamId teamId = new TeamId(getCleanHtml(row.select("td[data-stat=team_id] a").html()));

				if (teamId.getMachineName().equals("TOT"))
					continue;

				if (currentTeam == null) {
					currentTeam = teamId;
				} else if (!teamId.equals(currentTeam)) {

					memberships.add(new TeamMembership(currentTeam, currentBegin != null ? currentBegin : altBegin, currentEnd != null ? currentEnd : altEnd));

					altBegin = null;
					altEnd = null;
					currentTeam = teamId;
					currentBegin = null;
					currentEnd = null;
				}

				if (altBegin == null) {
					altBegin = TimeObject.fromYear(seasonBegin);
				}
				altEnd = TimeObject.fromYear(seasonEnd);

				// TimeObject beginDate = null;
				// TimeObject endDate = null;

				for (Tri<TimeObject, TeamId, TeamId> td : transactionDates) {
					if (currentBegin == null && td.getC() != null && td.getC().equals(teamId) && td.getA().getYear() == seasonBegin) {
						currentBegin = td.getA();
					}
					if (td.getB() != null && td.getB().equals(teamId) && td.getA().getYear() == seasonEnd) {
						currentEnd = td.getA();
					}
				}

			}

			memberships.add(new TeamMembership(currentTeam, currentBegin != null ? currentBegin : altBegin,
					currentEnd != null ? currentEnd : (altEnd.getYear() < NBACrawler.today ? altEnd : null)));

			retMap.put(e.getFirst(), new Pair<>(playerData, memberships));
		}

		return retMap;
	}

	private LinkedList<Pair<PlayerId, String>> getPlayerUrls(Collection<String> playerListsUrls) throws MalformedURLException, IOException {
		LinkedList<Pair<PlayerId, String>> retMap = new LinkedList<>();
		for (String listUrl : playerListsUrls) {

			String input = NBACrawler.getDocument(listUrl);
			Document doc = Jsoup.parse(input);
			Elements rows = doc.select("#players tbody tr");
			for (int i = 0; i < rows.size(); i++) {

				Element headerCell = rows.get(i).select("th").get(0);

				String playerId = headerCell.attr("data-append-csv");
				Element link = headerCell.select("a").get(0);
				String playerUrl = link.attr("href");
				retMap.add(new Pair<>(new PlayerId(playerId), playerUrl));
			}

		}

		return retMap;
	}

	private LinkedList<String> getPlayerListsUrls() throws MalformedURLException, IOException {

		LinkedList<String> retSet = new LinkedList<>();

		String input = NBACrawler.getDocument("/players/");
		Document doc = Jsoup.parse(input);
		Elements links = doc.select(".page_index > li > a");

		for (int i = 0; i < links.size(); i++) {
			Element link = links.get(i);
			String url = link.attr("href");
			retSet.add(url);
		}

		return retSet;
	}

	private Map<TeamId, Map<Integer, String>> getTeamYearUrls(Set<TeamId> teams) throws MalformedURLException, IOException {
		Map<TeamId, Map<Integer, String>> retMap = new HashMap<>();
		for (TeamId teamId : teams) {
			InputStream in = new URL("http://www.footballdb.com/teams/nfl/" + teamId.getMachineName() + "/roster/").openStream();
			try {
				String input = IOUtils.toString(in);
				Document doc = Jsoup.parse(input);
				Elements links = doc.select("#rightcol .dkmodule .modulebody a");
				Map<Integer, String> urlMap = new HashMap<>();
				for (int i = 0; i < links.size(); i++) {
					Element link = links.get(i);
					String url = link.attr("href");
					Integer year = Integer.valueOf(link.text().trim());
					urlMap.put(year, url);
				}

				retMap.put(teamId, urlMap);
			} finally {
				IOUtils.closeQuietly(in);
			}

		}

		return retMap;

	}

	public void playerWrite(String outputFile, Map<PlayerId, PlayerData> playerDataMap) throws IOException {
		CSVWriter writer = new CSVWriter(new FileWriter(outputFile));
		try {

			{
				String[] header = new String[4];
				header[0] = "MachineName";
				header[1] = "Name";
				header[2] = "Birthdate";
				header[3] = "College";

				writer.writeNext(header);
			}

			for (Entry<PlayerId, PlayerData> playerEntry : playerDataMap.entrySet()) {
				PlayerData playerData = playerEntry.getValue();

				if (playerData != null) {
					String[] line = new String[4];
					line[0] = playerEntry.getKey().getMachineName();
					line[1] = playerEntry.getValue().getName();

					line[2] = playerData.getBirthDate();
					line[3] = playerData.getCollege();

					writer.writeNext(line);
				}

			}

		} finally {
			writer.close();
		}
	}

	public List<Tri<PlayerId, PlayerData, MembershipData>> getPlayerData(Map<TeamId, Map<Integer, String>> teamYearUrls, TeamId teamId, Integer year)
			throws IOException {
		String url = teamYearUrls.get(teamId).get(year);
		if (url != null) {
			InputStream in = new URL("http://www.footballdb.com" + url).openStream();
			try {
				List<Tri<PlayerId, PlayerData, MembershipData>> playerData = new ArrayList<>();
				String input = IOUtils.toString(in);

				Document doc = Jsoup.parse(input);
				Elements tables = doc.select("table.statistics tbody");
				Element table = tables.get(0);
				Elements rows = table.select("tr");
				for (int j = 0; j < rows.size(); j++) {
					Elements rowElements = rows.get(j).children();
					String number = getCleanHtml(rowElements.get(0).html());
					if (number.equals("--")) {
						number = "";
					}
					Element playerLink = rowElements.get(1).select("a").first();
					String playerName = playerLink.html();
					String playerMachineName = playerLink.attr("href").replace("/players/", "");
					String position = getCleanHtml(rowElements.get(2).html());
					String birthDate = getCleanHtml(rowElements.get(5).html());
					String college = getCleanHtml(rowElements.get(6).html());

					PlayerId playerId = new PlayerId(playerMachineName);
					MembershipData md = new MembershipData(number, position, teamId);
					PlayerData pd = new PlayerData(playerName, birthDate, college);
					Tri<PlayerId, PlayerData, MembershipData> pair = new Tri<>(playerId, pd, md);

					playerData.add(pair);

				}

				return playerData;

			} finally {
				IOUtils.closeQuietly(in);
			}
		}

		return new ArrayList<>();

	}

	public static String getCleanHtml(String html) {
		return html.replace("&nbsp;", "").trim();
	}

	public Map<TeamId, String> getTeams() throws MalformedURLException, IOException {
		InputStream in = new URL("http://www.footballdb.com/teams/index.html").openStream();
		try {
			Map<TeamId, String> teams = new HashMap<>();
			String input = IOUtils.toString(in);

			Document doc = Jsoup.parse(input);
			Elements tables = doc.select("table.statistics");
			Element table = tables.get(0);
			for (int i = 0; i <= 1; i++) {
				Elements rows = table.select("tr.row" + i);
				for (int j = 0; j < rows.size(); j++) {
					Element nameTd = rows.get(j).children().get(0);
					Element link = nameTd.select("b a").first();
					String name = link.html();
					String machineName = link.attr("href").replace("/teams/nfl/", "");
					TeamId teamId = new TeamId(machineName);
					teams.put(teamId, name);
				}

			}

			return teams;

		} finally {
			IOUtils.closeQuietly(in);
		}
	}

	public static String getDocument(String url) {
		InputStream in = null;
		try {
			in = new URL(basePath + url).openStream();
			return IOUtils.toString(in);
		} catch (IOException e) {
			try {
				TimeUnit.SECONDS.sleep(5);
			} catch (InterruptedException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
			e.printStackTrace();
			return getDocument(url);
		} finally {
			IOUtils.closeQuietly(in);
		}
	}

	static class PlayerData {
		private final String name;

		private final String birthDate;
		private final String college;

		public PlayerData(String name, String birthDate, String college) {
			super();
			this.name = name;

			this.birthDate = birthDate;
			this.college = college;
		}

		public String getName() {
			return name;
		}

		public String getBirthDate() {
			return birthDate;
		}

		public String getCollege() {
			return college;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((birthDate == null) ? 0 : birthDate.hashCode());
			result = prime * result + ((college == null) ? 0 : college.hashCode());
			result = prime * result + ((name == null) ? 0 : name.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			PlayerData other = (PlayerData) obj;
			if (birthDate == null) {
				if (other.birthDate != null)
					return false;
			} else if (!birthDate.equals(other.birthDate))
				return false;
			if (college == null) {
				if (other.college != null)
					return false;
			} else if (!college.equals(other.college))
				return false;
			if (name == null) {
				if (other.name != null)
					return false;
			} else if (!name.equals(other.name))
				return false;
			return true;
		}

		@Override
		public String toString() {
			return "PlayerData [name=" + name + ", birthDate=" + birthDate + ", college=" + college + "]";
		}

	}

	static class MembershipData {
		private final String number;
		private final String position;
		private final TeamId teamId;

		public MembershipData(String number, String position, TeamId teamId) {
			super();
			this.number = number;
			this.position = position;
			this.teamId = teamId;
		}

		public String getNumber() {
			return number;
		}

		public String getPosition() {
			return position;
		}

		public TeamId getTeamId() {
			return teamId;
		}

	}

	// public class Birthday {
	// private final int year;
	// private final int month;
	// private final int day;
	//
	// public Birthday(int year, int month, int day) {
	// this.year = year;
	// this.month = month;
	// this.day = day;
	// }
	//
	// public int getYear() {
	// return year;
	// }
	//
	// public int getMonth() {
	// return month;
	// }
	//
	// public int getDay() {
	// return day;
	// }
	//
	// }

	public static class TeamId {
		private final String machineName;

		public TeamId(String machineName) {
			this.machineName = machineName;
		}

		public String getMachineName() {
			return machineName;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((machineName == null) ? 0 : machineName.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			TeamId other = (TeamId) obj;
			if (machineName == null) {
				if (other.machineName != null)
					return false;
			} else if (!machineName.equals(other.machineName))
				return false;
			return true;
		}

		@Override
		public String toString() {
			return machineName;
		}

	}

	public static class TeamMembership {
		private final TeamId teamId;
		private final TimeObject from;
		private final TimeObject to;

		public TeamMembership(TeamId teamId, TimeObject from, TimeObject to) {
			super();
			this.teamId = teamId;
			this.from = from;
			this.to = to;
		}

		public TeamId getTeamId() {
			return teamId;
		}

		public TimeObject getFrom() {
			return from;
		}

		public TimeObject getTo() {
			return to;
		}

		@Override
		public String toString() {
			return "TeamMembership [teamId=" + teamId + ", from=" + from + ", to=" + to + "]";
		}

	}

	public static class TeamMembershipSimple {
		private final TeamId teamId;
		private final int from;
		private final int to;

		public TeamMembershipSimple(TeamId teamId, int from, int to) {

			this.teamId = teamId;
			this.from = from;
			this.to = to;
		}

		public TeamId getTeamId() {
			return teamId;
		}

		public int getFrom() {
			return from;
		}

		public int getTo() {
			return to;
		}

	}

	public static class PlayerId {
		private final String machineName;

		public PlayerId(String machineName) {
			this.machineName = machineName;
		}

		public String getMachineName() {
			return machineName;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((machineName == null) ? 0 : machineName.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			PlayerId other = (PlayerId) obj;
			if (machineName == null) {
				if (other.machineName != null)
					return false;
			} else if (!machineName.equals(other.machineName))
				return false;
			return true;
		}

		@Override
		public String toString() {
			return "PlayerId [machineName=" + machineName + "]";
		}

	}
}
