package de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.nba;

import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;
import org.jsoup.select.Elements;

import au.com.bytecode.opencsv.CSVWriter;
import de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.nba.NBACrawler.TeamId;

public class NBATeamsCrawler {

	public static void main(String[] args) throws IOException {

		NBATeamsCrawler crawler = new NBATeamsCrawler();

		Map<TeamId, TeamData> teams = crawler.getTeams();

		String outputPath = "C:/Users/yaser/SampleRuns/Wikidata/datasets/NBA_ATHLETE/";

		CSVWriter writer = null;
		try {
			writer = new CSVWriter(new FileWriter(outputPath + "teams_names.csv"));

			{
				String[] headerLine = new String[3];
				headerLine[0] = "Id";
				headerLine[1] = "Name";
				headerLine[2] = "Inception";

				writer.writeNext(headerLine);

			}

			for (Entry<TeamId, TeamData> team : teams.entrySet()) {

				String[] line = new String[3];
				line[0] = team.getKey().getMachineName();
				line[1] = team.getValue().getName();
				line[2] = String.valueOf(team.getValue().getInception());
				writer.writeNext(line);

			}

		} finally {
			writer.close();

		}

	}

	public Map<TeamId, TeamData> getTeams() {
		Map<TeamId, TeamData> retMap = new HashMap<>();

		String html = NBACrawler.getDocument("/teams/");

		Document doc = Jsoup.parse(html);

		Elements activeTeams = doc.select("#teams_active .full_table");
		retMap.putAll(getTeamsFromTable(activeTeams));

		Elements defunctTeams = doc.select("#teams_defunct .full_table");
		retMap.putAll(getTeamsFromTable(defunctTeams));

		return retMap;
	}

	public Map<TeamId, TeamData> getTeamsFromTable(Elements teams) {
		Map<TeamId, TeamData> retMap = new HashMap<>();
		for (int i = 0; i < teams.size(); i++) {
			Element team = teams.get(i);
			Element link = team.select("th a").get(0);
			TeamId teamId = new TeamId(link.attr("href").split("/")[2]);
			String name = NBACrawler.getCleanHtml(link.html());
			int inception = Integer.parseInt(NBACrawler.getCleanHtml(team.select("td[data-stat=year_min]").get(0).html()));

			retMap.put(teamId, new TeamData(name, inception));

		}
		return retMap;
	}

	public static class TeamData {
		private final String name;
		private final int inception;

		public TeamData(String name, int inception) {
			super();
			this.name = name;
			this.inception = inception;
		}

		public String getName() {
			return name;
		}

		public int getInception() {
			return inception;
		}

	}

}
