package de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.nfl;

import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;

import org.apache.commons.io.IOUtils;
import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;
import org.jsoup.select.Elements;

import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.mannheim.uni.expansion.utils.containers.Tri;

public class NFLCrawler {

	public static void main(String[] args) throws MalformedURLException, IOException {

		Integer beginYear = 1960;
		Integer endYear = 2016;

		NFLCrawler crawler = new NFLCrawler();
		Map<TeamId, String> teams = crawler.getTeams();
		Map<TeamId, Map<Integer, String>> teamYearUrls = crawler.getTeamYearUrls(teams.keySet());

		String outputPath = "C:/Users/yaser/SampleRuns/Wikidata/datasets/NFL_ATHLETE/";

		List<Integer> years = new ArrayList<>();
		for (int i = beginYear; i <= endYear; i++) {
			years.add(i);
		}

		Map<PlayerId, PlayerData> playerData = new ConcurrentHashMap<>();

		ConcurrentHashMap<PlayerId, Map<Integer, MembershipData>> playerMap = new ConcurrentHashMap<>();

		// Map<Integer, Map<String, List<PlayerData>>> yearMap = new
		// HashMap<>();
		for (Integer year : years) {
			// Map<String, List<PlayerData>> teamMap =
			// CHMHelper.getFromMap(yearMap, year, ()->new HashMap<>());

			PersistentParallel4.runAndFinish(teams.entrySet(), (team) -> {

				try {
					List<Tri<PlayerId, PlayerData, MembershipData>> playerPair = crawler.getPlayerData(teamYearUrls, team.getKey(), year);

					// teamMap.put(team.getKey(), playerData);
					for (Tri<PlayerId, PlayerData, MembershipData> player : playerPair) {
						CHMHelper.getFromCHM(playerMap, player.getA(), () -> new ConcurrentHashMap<>()).put(year, player.getC());
						if (playerData.containsKey(player.getA())) {
							PlayerData otherPlayerData = playerData.get(player.getA());
							if (!otherPlayerData.equals(player.getB())) {
								throw new IllegalStateException();
							}
						} else {
							playerData.put(player.getA(), player.getB());
						}

					}
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			} , "GET NFL PLAYERS");

		}

		// Eliminating Players with the same Name
		// Map<String, String> reverseNames = new HashMap<>();
		// for (Entry<String, String> p : playerNames.entrySet()) {
		// if (reverseNames.containsKey(p.getValue())) {
		// playerMap.remove(p.getKey());
		// playerMap.remove(reverseNames.get(p.getValue()));
		// } else {
		// reverseNames.put(p.getValue(), p.getKey());
		// }
		// }

		// int countTotal = 0;
		// Map<Integer, MutableInt> countMap = new HashMap<>();
		// Map<Integer, MutableInt> switchedTeamMap = new HashMap<>();
		// Map<Integer, MutableInt> switchedNumbersMap = new HashMap<>();
		// for (Iterator<Entry<String, Map<Integer, PlayerData>>> it =
		// playerMap.entrySet().iterator(); it.hasNext();) {
		// Entry<String, Map<Integer, PlayerData>> player = it.next();
		//
		// if (player.getValue().size() < 3) {
		// it.remove();
		// continue;
		// }
		//
		// int switchedTeam = 0;
		// String lastTeam = null;
		// for (Entry<Integer, PlayerData> e : player.getValue().entrySet()) {
		// if (lastTeam == null) {
		// lastTeam = e.getValue().getMachineTeamName();
		// } else if (!lastTeam.equals(e.getValue().getMachineTeamName())) {
		// switchedTeam++;
		// lastTeam = e.getValue().getMachineTeamName();
		// }
		// }
		//
		// int switchedNumbers = 0;
		// String lastNumber = null;
		// for (Entry<Integer, PlayerData> e : player.getValue().entrySet()) {
		// if (lastNumber == null) {
		// lastNumber = e.getValue().getNumber();
		// } else if (!lastNumber.equals(e.getValue().getNumber())) {
		// switchedNumbers++;
		// lastNumber = e.getValue().getNumber();
		// }
		// }
		//
		// CHMHelper.getFromMap(switchedTeamMap, switchedTeam, () -> new
		// MutableInt(0)).increment();
		// CHMHelper.getFromMap(switchedNumbersMap, switchedNumbers, () -> new
		// MutableInt(0)).increment();
		// CHMHelper.getFromMap(countMap, player.getValue().size(), () -> new
		// MutableInt(0)).increment();
		// countTotal++;
		//
		// }
		//
		// System.out.println("YearsActive" + countMap.toString());
		// System.out.println("SwitchedTeams" + switchedTeamMap.toString());
		// System.out.println("SwitchedNumbers" +
		// switchedNumbersMap.toString());
		// System.out.println("Total: " + countTotal);

		crawler.playerWrite(outputPath + "players.csv", playerData);

		crawler.write(beginYear, endYear, outputPath + "teams.csv", teams, playerData, playerMap, (m) -> {
			return m.getTeamId().getMachineName();
		});

		crawler.write(beginYear, endYear, outputPath + "numbers.csv", teams, playerData, playerMap, (m) -> {
			return m.getNumber();
		});

		try {
			PersistentParallel4.close();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		// crawler.write(null, outputPath, teams, playerNames, playerMap);
	}

	private Map<TeamId, Map<Integer, String>> getTeamYearUrls(Set<TeamId> teams) throws MalformedURLException, IOException {
		Map<TeamId, Map<Integer, String>> retMap = new HashMap<>();
		for (TeamId teamId : teams) {
			InputStream in = new URL("http://www.footballdb.com/teams/nfl/" + teamId.getMachineName() + "/roster/").openStream();
			try {
				String input = IOUtils.toString(in);
				Document doc = Jsoup.parse(input);
				Elements links = doc.select("#rightcol .dkmodule .modulebody a");
				Map<Integer, String> urlMap = new HashMap<>();
				for (int i = 0; i < links.size(); i++) {
					Element link = links.get(i);
					String url = link.attr("href");
					Integer year = Integer.valueOf(link.text().trim());
					urlMap.put(year, url);
				}

				retMap.put(teamId, urlMap);
			} finally {
				IOUtils.closeQuietly(in);
			}

		}

		return retMap;

	}

	public void playerWrite(String outputFile, Map<PlayerId, PlayerData> playerDataMap) throws IOException {
		CSVWriter writer = new CSVWriter(new FileWriter(outputFile));
		try {

			{
				String[] header = new String[4];
				header[0] = "MachineName";
				header[1] = "Name";
				header[2] = "Birthdate";
				header[3] = "College";

				writer.writeNext(header);
			}

			for (Entry<PlayerId, PlayerData> playerEntry : playerDataMap.entrySet()) {
				PlayerData playerData = playerEntry.getValue();

				if (playerData != null) {
					String[] line = new String[4];
					line[0] = playerEntry.getKey().getMachineName();
					line[1] = playerEntry.getValue().getName();

					line[2] = playerData.getBirthDate();
					line[3] = playerData.getCollege();

					writer.writeNext(line);
				}

			}

		} finally {
			writer.close();
		}
	}

	public void write(Integer beginYear, Integer endYear, String outputFile, Map<TeamId, String> teams, Map<PlayerId, PlayerData> playerDataMap,
			ConcurrentHashMap<PlayerId, Map<Integer, MembershipData>> playerMap, Function<MembershipData, String> valueGetter) throws IOException {
		CSVWriter writer = new CSVWriter(new FileWriter(outputFile));
		int length = 1 + (endYear - beginYear + 1);
		{
			String[] header = new String[length];
			header[0] = "Player";
			// header[1] = "Team";
			// header[2] = "Number";
			// header[3] = "Position";
			// header[1] = "Birthdate";
			// header[2] = "College";
			int hi = 1;
			for (int i = beginYear; i <= endYear; i++) {
				header[hi] = String.valueOf(i);
				hi++;
			}

			writer.writeNext(header);
		}
		try {
			for (Entry<PlayerId, PlayerData> playerEntry : playerDataMap.entrySet()) {
				PlayerData playerData = playerEntry.getValue();

				if (playerData != null) {
					String[] line = new String[length];
					line[0] = playerEntry.getKey().getMachineName();

					// line[1] = teams.get(playerData.getMachineTeamName());
					// line[2] = playerData.getNumber();
					// line[3] = playerData.getPosition();
					// line[1] = playerData.getBirthDate();
					// line[2] = playerData.getCollege();

					Map<Integer, MembershipData> membershipData = playerMap.get(playerEntry.getKey());

					int li = 1;
					for (int i = beginYear; i <= endYear; i++) {
						MembershipData md = membershipData.get(i);
						if (md != null) {
							line[li] = valueGetter.apply(md);
						} else {
							line[li] = "";
						}
						li++;
					}

					writer.writeNext(line);
				}

			}

		} finally {
			writer.close();
		}
	}

	public List<Tri<PlayerId, PlayerData, MembershipData>> getPlayerData(Map<TeamId, Map<Integer, String>> teamYearUrls, TeamId teamId, Integer year)
			throws IOException {
		String url = teamYearUrls.get(teamId).get(year);
		if (url != null) {
			InputStream in = new URL("http://www.footballdb.com" + url).openStream();
			try {
				List<Tri<PlayerId, PlayerData, MembershipData>> playerData = new ArrayList<>();
				String input = IOUtils.toString(in);

				Document doc = Jsoup.parse(input);
				Elements tables = doc.select("table.statistics tbody");
				Element table = tables.get(0);
				Elements rows = table.select("tr");
				for (int j = 0; j < rows.size(); j++) {
					Elements rowElements = rows.get(j).children();
					String number = getCleanHtml(rowElements.get(0).html());
					if (number.equals("--")) {
						number = "";
					}
					Element playerLink = rowElements.get(1).select("a").first();
					String playerName = playerLink.html();
					String playerMachineName = playerLink.attr("href").replace("/players/", "");
					String position = getCleanHtml(rowElements.get(2).html());
					String birthDate = getCleanHtml(rowElements.get(5).html());
					String college = getCleanHtml(rowElements.get(6).html());

					PlayerId playerId = new PlayerId(playerMachineName);
					MembershipData md = new MembershipData(number, position, teamId);
					PlayerData pd = new PlayerData(playerName, birthDate, college);
					Tri<PlayerId, PlayerData, MembershipData> pair = new Tri<>(playerId, pd, md);

					playerData.add(pair);

				}

				return playerData;

			} finally {
				IOUtils.closeQuietly(in);
			}
		}

		return new ArrayList<>();

	}

	public String getCleanHtml(String html) {
		return html.replace("&nbsp;", "").trim();
	}

	public Map<TeamId, String> getTeams() throws MalformedURLException, IOException {
		InputStream in = new URL("http://www.footballdb.com/teams/index.html").openStream();
		try {
			Map<TeamId, String> teams = new HashMap<>();
			String input = IOUtils.toString(in);

			Document doc = Jsoup.parse(input);
			Elements tables = doc.select("table.statistics");
			Element table = tables.get(0);
			for (int i = 0; i <= 1; i++) {
				Elements rows = table.select("tr.row" + i);
				for (int j = 0; j < rows.size(); j++) {
					Element nameTd = rows.get(j).children().get(0);
					Element link = nameTd.select("b a").first();
					String name = link.html();
					String machineName = link.attr("href").replace("/teams/nfl/", "");
					TeamId teamId = new TeamId(machineName);
					teams.put(teamId, name);
				}

			}

			return teams;

		} finally {
			IOUtils.closeQuietly(in);
		}
	}

	static class PlayerData {
		private final String name;

		private final String birthDate;
		private final String college;

		public PlayerData(String name, String birthDate, String college) {
			super();
			this.name = name;

			this.birthDate = birthDate;
			this.college = college;
		}

		public String getName() {
			return name;
		}

		public String getBirthDate() {
			return birthDate;
		}

		public String getCollege() {
			return college;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((birthDate == null) ? 0 : birthDate.hashCode());
			result = prime * result + ((college == null) ? 0 : college.hashCode());
			result = prime * result + ((name == null) ? 0 : name.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			PlayerData other = (PlayerData) obj;
			if (birthDate == null) {
				if (other.birthDate != null)
					return false;
			} else if (!birthDate.equals(other.birthDate))
				return false;
			if (college == null) {
				if (other.college != null)
					return false;
			} else if (!college.equals(other.college))
				return false;
			if (name == null) {
				if (other.name != null)
					return false;
			} else if (!name.equals(other.name))
				return false;
			return true;
		}

		@Override
		public String toString() {
			return "PlayerData [name=" + name + ", birthDate=" + birthDate + ", college=" + college + "]";
		}

	}

	static class MembershipData {
		private final String number;
		private final String position;
		private final TeamId teamId;

		public MembershipData(String number, String position, TeamId teamId) {
			super();
			this.number = number;
			this.position = position;
			this.teamId = teamId;
		}

		public String getNumber() {
			return number;
		}

		public String getPosition() {
			return position;
		}

		public TeamId getTeamId() {
			return teamId;
		}

	}

	// public class Birthday {
	// private final int year;
	// private final int month;
	// private final int day;
	//
	// public Birthday(int year, int month, int day) {
	// this.year = year;
	// this.month = month;
	// this.day = day;
	// }
	//
	// public int getYear() {
	// return year;
	// }
	//
	// public int getMonth() {
	// return month;
	// }
	//
	// public int getDay() {
	// return day;
	// }
	//
	// }

	public static class TeamId {
		private final String machineName;

		public TeamId(String machineName) {
			this.machineName = machineName;
		}

		public String getMachineName() {
			return machineName;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((machineName == null) ? 0 : machineName.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			TeamId other = (TeamId) obj;
			if (machineName == null) {
				if (other.machineName != null)
					return false;
			} else if (!machineName.equals(other.machineName))
				return false;
			return true;
		}

	}

	public static class PlayerId {
		private final String machineName;

		public PlayerId(String machineName) {
			this.machineName = machineName;
		}

		public String getMachineName() {
			return machineName;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((machineName == null) ? 0 : machineName.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			PlayerId other = (PlayerId) obj;
			if (machineName == null) {
				if (other.machineName != null)
					return false;
			} else if (!machineName.equals(other.machineName))
				return false;
			return true;
		}

		@Override
		public String toString() {
			return "PlayerId [machineName=" + machineName + "]";
		}

	}
}
