package de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.stockcrow;

import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import au.com.bytecode.opencsv.CSVReader;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueContainer;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.QuantityObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedAbstractObjectHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.Qualifier;
import de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.stockcrow.StockCrowCrawler.FinancialData;
import de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.stockcrow.StockCrowCrawler.FinancialYear;
import de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.stockcrow.StockCrowCrawler.TrackerId;
import de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats.Stats;

public class StockCrowConverter implements KBValueCreator {

	private final String path;

	public StockCrowConverter(String path) {
		super();
		this.path = path;
	}

	public static void main(String[] args) {
		StockCrowConverter converter = new StockCrowConverter("C:/Users/yaser/SampleRuns/Wikidata/datasets/FINANCE/");
		converter.getMatchedValues(null);
	}

	@Override
	public KBValueContainer getMatchedValues(KnowledgeBase kb) {

		try {

			Map<ValueId, QualifiedValueList> retMap = new ConcurrentHashMap<ValueId, QualifiedValueList>();
			Set<ValueId> remove = new HashSet<>();

			Map<TrackerId, InstanceId> matching = getMatching(path + "companies.csv");

			Map<TrackerId, Map<FinancialYear, FinancialData>> readData = getData(path + "data.csv");

			for (Entry<TrackerId, Map<FinancialYear, FinancialData>> company : readData.entrySet()) {
				InstanceId instanceId = matching.get(company.getKey());

				List<QualifiedValueHolder> revenueList = new ArrayList<>();
				List<QualifiedValueHolder> netIncomeList = new ArrayList<>();
				List<QualifiedValueHolder> totalAssetsList = new ArrayList<>();
				List<QualifiedValueHolder> ebitList = new ArrayList<>();
				List<QualifiedValueHolder> equityList = new ArrayList<>();

				for (Entry<FinancialYear, FinancialData> year : company.getValue().entrySet()) {
					List<Qualifier> qualifiers = new ArrayList<>();
					qualifiers.add(new Qualifier(Stats.pointInTime, TimeObject.fromYear(year.getKey().getYear())));

					if (year.getValue().getRevenue() != null) {
						revenueList.add(new QualifiedAbstractObjectHolder(QuantityObject.fromDouble(year.getValue().getRevenue()), qualifiers));
					}
					if (year.getValue().getAssets() != null) {
						totalAssetsList.add(new QualifiedAbstractObjectHolder(QuantityObject.fromDouble(year.getValue().getAssets()), qualifiers));
					}
					if (year.getValue().getNetIncome() != null) {
						netIncomeList.add(new QualifiedAbstractObjectHolder(QuantityObject.fromDouble(year.getValue().getNetIncome()), qualifiers));
					}
					if (year.getValue().getEquity() != null) {
						equityList.add(new QualifiedAbstractObjectHolder(QuantityObject.fromDouble(year.getValue().getEquity()), qualifiers));
					}
					if (year.getValue().getEbit() != null) {
						ebitList.add(new QualifiedAbstractObjectHolder(QuantityObject.fromDouble(year.getValue().getEbit()), qualifiers));
					}
				}

				if (revenueList.size() > 0) {
					retMap.put(new ValueId(instanceId, Stats.totalRevenue), new QualifiedValueList(revenueList));
				}

				if (netIncomeList.size() > 0) {
					retMap.put(new ValueId(instanceId, Stats.netIncome), new QualifiedValueList(netIncomeList));
				}

				if (totalAssetsList.size() > 0) {
					retMap.put(new ValueId(instanceId, Stats.totalAssets), new QualifiedValueList(totalAssetsList));
				}

				if (ebitList.size() > 0) {
					retMap.put(new ValueId(instanceId, Stats.ebit), new QualifiedValueList(ebitList));
				}

				if (equityList.size() > 0) {
					retMap.put(new ValueId(instanceId, Stats.totalEquity), new QualifiedValueList(equityList));
				}

			}

			// for (Entry<InstanceId, Integer> instance :
			// kb.getInstances(WikidataClasses.wikidataFootball).entrySet()) {
			// ValueId teamValueId = new ValueId(instance.getKey(),
			// NFLConverter.team);
			// if (!retMap.containsKey(teamValueId)) {
			// remove.add(teamValueId);
			// }
			//
			// ValueId sportsNumber = new ValueId(instance.getKey(),
			// NFLConverter.sportsNumber);
			// if (!retMap.containsKey(sportsNumber)) {
			// remove.add(sportsNumber);
			// }
			//
			// }

			return new KBValueContainer(retMap, remove);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			throw new RuntimeException(e);
		}

	}

	private Map<TrackerId, Map<FinancialYear, FinancialData>> getData(String file) throws IOException {
		Map<TrackerId, Map<FinancialYear, FinancialData>> retMap = new HashMap<>();

		CSVReader reader = new CSVReader(new FileReader(file));
		try {
			List<String[]> allLines = reader.readAll();
			for (int i = 1; i < allLines.size(); i++) {
				String[] line = allLines.get(i);

				TrackerId trackerId = new TrackerId(line[0].trim());
				FinancialYear year = new FinancialYear(Integer.valueOf(line[1].trim()));
				Double revenue = line[2].trim().equals("") ? null : Double.parseDouble(line[2].trim());
				Double assets = line[3].trim().equals("") ? null : Double.parseDouble(line[3].trim());
				Double netIncome = line[4].trim().equals("") ? null : Double.parseDouble(line[4].trim());
				Double ebit = line[5].trim().equals("") ? null : Double.parseDouble(line[5].trim());
				Double equity = line[6].trim().equals("") ? null : Double.parseDouble(line[6].trim());
				FinancialData data = new FinancialData(revenue, assets, netIncome, ebit, equity);
				CHMHelper.getFromMap(retMap, trackerId, () -> new HashMap<>()).put(year, data);
			}

			return retMap;
		} finally {
			reader.close();
		}
	}

	private Map<TrackerId, InstanceId> getMatching(String file) throws IOException {

		Map<TrackerId, InstanceId> retMap = new HashMap<>();
		CSVReader reader = new CSVReader(new FileReader(file));
		try {

			List<String[]> allLines = reader.readAll();
			for (int i = 1; i < allLines.size(); i++) {
				String[] line = allLines.get(i);
				String[] idSplit = line[0].trim().split("/");
				InstanceId instanceId = InstanceId.simple(idSplit[idSplit.length - 1]);
				TrackerId trackerId = new TrackerId(line[2].trim());
				retMap.put(trackerId, instanceId);
			}

			return retMap;
		} finally {
			reader.close();
		}

	}
}
