package de.uni_mannheim.informatik.dws.expansion.tools.converters.datasets.worldbank;

import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import com.google.common.collect.BiMap;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueContainer;
import de.uni_mannheim.dws.expansion.data.kb.repository.KBValueCreator;
import de.uni_mannheim.dws.expansion.data.kb.repository.KnowledgeBase;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.StringObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.wikidata.WikidataClasses;
import de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats.Stats;

public class CountryListFixer implements KBValueCreator {

	private final String isoCodeFile;

	public CountryListFixer(String isoCodeFile) {
		super();
		this.isoCodeFile = isoCodeFile;
	}

	@Override
	public KBValueContainer getMatchedValues(KnowledgeBase kb) {

		Map<ValueId, QualifiedValueList> replaceValues = new HashMap<>();
		Set<InstanceId> removeInstances = new HashSet<>();
		Set<ValueId> removeValues = new HashSet<>();

		try {
			Set<String> isoCodes = WorldBankConverter.getIsoCodesStatic(isoCodeFile);

			Map<String, Set<InstanceId>> isoToCountry = new HashMap<>();
			BiMap<InstanceId, Integer> countries = kb.getInstances(WikidataClasses.wikidataCountry);
			for (Entry<InstanceId, Integer> country : countries.entrySet()) {

				QualifiedValueList isoQvl = kb.getQualifiedValueList(new ValueId(country.getKey(), Stats.isoAlpha3Codes));
				if (isoQvl != null && isoQvl.size() > 0) {
					StringObject iso = (StringObject) isoQvl.getClaims().get(0).getValue();
					if (!isoCodes.contains(iso.getString())) {
						removeInstances.add(country.getKey());

					} else {
						CHMHelper.getFromMap(isoToCountry, iso.getString(), () -> new HashSet<>()).add(country.getKey());
					}

				} else {
					removeInstances.add(country.getKey());
					// System.out.println("BIG ERROR. NO ISO FOUND: " +
					// country.getKey().getInstanceId());
				}
			}

			for (Entry<String, Set<InstanceId>> entry : isoToCountry.entrySet()) {

				if (entry.getValue().size() > 1) {
					System.out.print(entry.getKey() + ": ");
					for (InstanceId country : entry.getValue()) {
						System.out.print(country.getInstanceId() + ", ");
					}
					System.out.println();
				}
			}

			int sdf = 123;

		} catch (IOException e) {
			e.printStackTrace();
		}

		System.out.println(removeInstances.size());
		return new KBValueContainer(replaceValues, removeInstances, removeValues);
	}

}
