package de.uni_mannheim.informatik.dws.expansion.utils.learning.weka;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.Serializable;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationMethod;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationPrediction;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.RecordPrediction;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordData;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordFeature;
import weka.classifiers.AbstractClassifier;
import weka.classifiers.Classifier;
import weka.classifiers.meta.Bagging;
import weka.core.SerializationHelper;

public class WekaLearnedModel implements LearnedModel, Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private final Classifier classifier;
//	private final ThreadLocal<Classifier> localClassifier;
//	private final List<String> classLabels;
	
	public WekaLearnedModel(byte[] string) {
		ByteArrayInputStream stream = new ByteArrayInputStream(string );
		try {
			this.classifier = (Classifier) weka.core.SerializationHelper.read(stream);
		} catch (Exception e) { 
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
	}

	public WekaLearnedModel(boolean empty) {
		this.classifier = null;
		
	}
	
	public WekaLearnedModel(Classifier classifier ) {
		if(classifier == null) {
			throw new IllegalStateException();
		}
		this.classifier = classifier;
//		this.localClassifier = new   ThreadLocal<Classifier>() {
//	        @Override public Classifier initialValue() {
//	            try {
//					return AbstractClassifier.makeCopy(classifier);
//				} catch (Exception e) {
//					// TODO Auto-generated catch block
//					e.printStackTrace();
//					throw new RuntimeException(e);
//				}
//	        }
//	    };
//		this.classLabels = Collections.unmodifiableList(new ArrayList<>(classLabels));
	}
	
	
 
	@Override
	public RecordPrediction apply(RecordFeature d) {

		if(classifier == null) {
			
			return new RecordPrediction(d.getDatasetData().getLabels().get(0), 0.0 );
		}
//		
//		if(d.getDatasetData().isNumeric()) {
//			return new RecordPrediction(d.getDatasetData().getLabels().get(0), applyNumeric(d));
//		}
		
		try {
			double[] distribution;;
//			 synchronized(classifier) {
				 distribution = classifier.distributionForInstance(d.getInstance());
//			 }
		 

		 double unclassifiedConfidence = 1.0;
		 
			double maxConfidence = 0.0;
			String maxLabel = null;
			for (int i = 0; i < distribution.length; i++) {
				unclassifiedConfidence -= distribution[i];
				if (distribution[i] > maxConfidence) {
					maxConfidence = distribution[i];
					maxLabel = d.getDatasetData().getLabels().get(i);
				}
			}
			
			if(unclassifiedConfidence > maxConfidence) {
				maxConfidence = 0.0;
				maxLabel = d.getDatasetData().getLabels().get(1);
			}
			if (maxLabel == null) {
				maxLabel = d.getDatasetData().getLabels().get(1);
				maxConfidence = 0.0;
			}
// System.out.println(maxConfidence);
			return new RecordPrediction(maxLabel, maxConfidence );

		} catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException(e);
		}
	}

	@Override
	public byte[] outputModelFile() {
		 ByteArrayOutputStream stream = new ByteArrayOutputStream();
       

        try { 
        	  
      		SerializationHelper.write(stream, classifier);
      	 
		
      		return stream.toByteArray();
        } catch ( Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
	}

	@Override
	public double runAndEvaluate(Collection<RecordData> data, EvaluationMethod evaluator) {
		

		List<EvaluationPrediction> predictions = new ArrayList<>();

		for (RecordData record : data) {
			
			RecordPrediction prediction = apply(record.getFeatures());
			predictions.add(new EvaluationPrediction(record.getLabel(), prediction));
		}

		return evaluator.performance(predictions);
		
	}

	@Override
	public Optional<Double> getNormalizingThreshold() {
		return Optional.empty();
	}

	@Override
	public String toString() {
		return  classifier.toString();
	}

	public double applyNumeric(RecordFeature featureVector) {
		if(classifier == null) {
			
			return 0.0;
		}
	
		try {
			return classifier.classifyInstance(featureVector.getInstance());
		} catch (Exception e) {
			// TODO Auto-generated catch block
			throw new RuntimeException(e);
		}
	}

	public double getError() {
		return ((Bagging) classifier).getOutOfBagEvaluationObject().errorRate();
	}
	
	
}
