package de.uni_mannheim.informatik.dws.expansion.utils.learning.weka;

import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationMethod;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnerId;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.ModelLearner;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData;
import weka.attributeSelection.AttributeSelection;
import weka.attributeSelection.GreedyStepwise;
import weka.attributeSelection.WrapperSubsetEval;
import weka.classifiers.Classifier;
import weka.core.Instances;
import weka.core.Utils;
import weka.core.WeightedInstancesHandler;

public class WekaLearnerCreator implements ModelLearner {

	
	
	
	private String[] parameters;
	private Classifier classifier;
	private final boolean forwardSelection;
	private final boolean backwardSelection;
	
	public WekaLearnerCreator(String classifierName, String parameters[] ) {
		this(classifierName, parameters, false, false);
	}
	public WekaLearnerCreator(String classifierName, String parameters[], boolean forwardSelection, boolean backwardSelection) {
		this.parameters = parameters;

		this.forwardSelection = forwardSelection;
		this.backwardSelection = backwardSelection;
		// create classifier
		try {
			this.classifier = (Classifier) Utils.forName(Classifier.class, classifierName, parameters);
		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
	}
	
	public void learn () {
		
		
		
	}

	@Override
	public LearnerId getLearnerId() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public LearnedModel parseModelFromFile(String string) {
		// TODO Auto-generated method stub
		return null;
	}

	public boolean supportsWeights() {
		return  false && classifier instanceof WeightedInstancesHandler;
	}
	@Override
	public WekaLearnedModel learn(TrainingData trainingData, EvaluationMethod evalutor) {
		try {
			Instances instances = trainingData.getDataSet();
			AttributeSelection fs;
			if (this.forwardSelection || this.backwardSelection) {

				GreedyStepwise search = new GreedyStepwise();
				search.setSearchBackwards(this.backwardSelection);

				fs = new AttributeSelection();
				WrapperSubsetEval wrapper = new WrapperSubsetEval();

				// Do feature subset selection, but using a 10-fold cross
				// validation
				wrapper.buildEvaluator(instances);
				wrapper.setClassifier(this.classifier);
				wrapper.setFolds(10);
				wrapper.setThreshold(0.01);

				fs.setEvaluator(wrapper);
				fs.setSearch(search);

				fs.SelectAttributes(instances);

				instances = fs.reduceDimensionality(instances);

			}
			
			this.classifier.buildClassifier(trainingData.getDataSet());
			
			return new WekaLearnedModel(classifier  );
		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalStateException(e);
		}
	}
}
