package de.uni_mannheim.informatik.ds.expansion.utils.learning;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import weka.core.Attribute;
import weka.core.Instances;

public class DataSetData implements Serializable {


	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	public static final String LABEL_FEATURE = "CLASS_LABEL";
	
	private final Instances dataset;
	private final List<Attribute> attributes;
	
	private final String name;
	private final List<String> labels;
	private final LearningFeatures features;

	private Map<String, Double> labelsNumeric;
 
	public DataSetData(String name, Map<String, Double> labels, LearningFeatures featureNames) {
		this(name, null, featureNames, labels);
	}
	
	public DataSetData(String name, List<String> labels, LearningFeatures featureNames) {
		this(name, labels, featureNames, null);
	}
	
	private DataSetData(String name, List<String> labels, LearningFeatures featureNames, Map<String, Double> labelsNumeric) {
		this.name = name;
	
		this.features = featureNames;
		
		ArrayList<weka.core.Attribute> attributes = new ArrayList<weka.core.Attribute>();
		 
		// create features per attributes of the FeatureVectorDataSet
		for (String featureName : featureNames.getFeatures().values()) {

		 
				weka.core.Attribute attribute = new weka.core.Attribute(featureName);
				attributes.add(attribute);
		 

		}
		
 

		if(labels != null) {
			this.labelsNumeric = null;
			this.labels = Collections.unmodifiableList(new ArrayList<>(labels));
			labels = new ArrayList<String>(labels);
			weka.core.Attribute cls = new weka.core.Attribute(LABEL_FEATURE, labels); 
			attributes.add(cls);
		}
		else {
			this.labels = null;
			this.labelsNumeric = Collections.unmodifiableMap(new HashMap<>(labelsNumeric));
			weka.core.Attribute cls = new weka.core.Attribute(LABEL_FEATURE); 
			attributes.add(cls);
		}
		
	

		this.attributes = Collections.unmodifiableList(attributes);
		
		
		this.dataset = emptyCopyOfDataSet();
		
	}
	
	public boolean isNumeric() {
		return labels == null;
	}
	
	 public double getLabelScore(String label) {
		 return labelsNumeric.get(label);
	 }
	 
	public Instances emptyCopyOfDataSet() {
		Instances dataset = new Instances(name, getAttributes(), 0);
		dataset.setClassIndex(attributes.size() - 1);
		return dataset;
	}
	
	
	public ArrayList<Attribute> getAttributes() {
		return new ArrayList<>(attributes);
	}
	
	public Instances getDataset() {
		return dataset;
	}


	public String getName() {
		return name;
	}


	public List<String> getLabels() {
		return labels;
	}


	public LearningFeatures getFeatures() {
		return features;
	}


	 
	
}
