package de.uni_mannheim.informatik.ds.expansion.utils.learning;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.stream.Collectors;

import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Utils;

public class TrainingData  {

	
//	private final List<String> featureNames;
//	private final List<RecordData> records; 
	private final Instances instances;
	private final List<RecordData> data;
	
	private final DataSetData datasetData;
	public TrainingData(DataSetData datasetData, Collection<RecordData> instances) {

		this.datasetData = datasetData;
		
//		ArrayList<weka.core.Attribute> attributes = datasetData.getAttributes();
		
		
		

		Instances dataset = datasetData.emptyCopyOfDataSet();
	
		Collection<String> featureNames = datasetData.getFeatures().getFeatures().values();
		for(RecordData rd : instances) {
			double[] attValues = new double[featureNames.size() + 1];
			int i = 0;
			for(String s : featureNames) {
				Double d = rd.getFeatures().getFeature(s);
				if(d == null)
					attValues[i] = Utils.missingValue();
				else
					attValues[i] = d;
				i++;
			}
			
			if(datasetData.isNumeric()) {
				attValues[featureNames.size()] = datasetData.getLabelScore(rd.getLabel())  ;
				dataset.add(new DenseInstance(rd.getWeight(), attValues));
			}
			else {
				attValues[featureNames.size()] = dataset.attribute(featureNames.size()).indexOfValue(rd.getLabel());
				dataset.add(new DenseInstance(rd.getWeight(), attValues));
			}
//			if(rd.getWeight() > 0.0) {
//			}
		}
		this.instances = dataset;
		
		this.data = Collections.unmodifiableList(new ArrayList<>(instances));
	}
	
//	public List<String> getFeatureNames() {
//		return datasetData.getFeatureNames();
//	}

	public Instances getDataSet() {
		return instances;
	}
	
	
//	public TrainingData(List<String> featureNames, Map<R, RecordData> records) {
//		super();
//		this.featureNames = Collections.unmodifiableList(new ArrayList<>(featureNames));
//		this.records = Collections.unmodifiableMap( new HashMap<>(records));
//	}
//
//
//
//
//	public List<String> getFeatureNames() {
//		return featureNames;
//	}
//
//
//
//
//	public Map<R, RecordData> getRecords() {
//		return records;
//	}




	public List<RecordData> getData() {
		return data;
	}


	public static class RecordData {
		private final String label;
		private final double weight;
		private final RecordFeature features;
		public RecordData(String label,RecordFeature features) {
			super();
			this.label = label;
			this.weight = 1.0;
			this.features = features;
		}
		
		public RecordData(String label,double weight, RecordFeature features) {
			super();
			this.label = label;
			this.weight = weight;
			this.features = features;
		}
		public String getLabel() {
			return label;
		}
		
		
		public double getWeight() {
			return weight;
		}

		public RecordFeature getFeatures() {
			return features;
		}
		
		
		
	}
	
	public static class RecordFeature {
		private final Map<String, Double> features;
		private final Instance instance;
		private final DataSetData datasetData;
		
		public RecordFeature(Map<String, Double> features, DataSetData datasetData) {
			super();
			this.datasetData = datasetData;
			this.features = Collections.unmodifiableMap(new HashMap<>(features));
			
			
			Collection<String> featureNames = datasetData.getFeatures().getFeatures().values();
 
				double[] attValues = new double[featureNames.size() + 1 ];
				int i = 0;
				for(String s : featureNames) {
					Double d = features.get(s);
					if(d == null)
						attValues[i] = Utils.missingValue();
					else
						attValues[i] = d;
					i++;
				}
				attValues[i] = 0.0;
				
			
			
 
			this.instance = new DenseInstance(1.0, attValues);
			
 
			
			this.instance.setDataset(datasetData.getDataset());
			
		}

		public Map<String, Double> getFeatures() {
			return features;
		}
		
		public Double getFeature(String feature) {
			return features.get(feature); 
		}
		
		public Instance getInstance() {
			return instance;
		}

		public DataSetData getDatasetData() {
			return datasetData;
		}

		public static Map<String, Double> descretizeMap(Map<String, Double> map, int i) {
			double iDouble = (double) i;
			Map<String, Double> descretizedMap = map.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> Math.round(e.getValue() * i) / iDouble));
			return descretizedMap;
		}
		
		
	}
}
