package de.uni_mannheim.informatik.dws.expansion.utils.optimization.learner;

import java.util.ArrayList;
import java.util.List;

import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationMethod;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationPrediction;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnerId;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.ModelLearner;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.RecordPrediction;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordData;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.genetic.WeightsGeneticParallelOptimizer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;

public class GeneticWeightsLearnerCreator {

	private final WeightsGeneticParallelOptimizer genetic = new WeightsGeneticParallelOptimizer();

	public ModelLearner getLearner(RecordBasedRun run) {
		return new GeneticWeigtsModelLearner(run);
	}

	public class GeneticWeigtsModelLearner implements ModelLearner {

		private final RecordBasedRun run;

		public GeneticWeigtsModelLearner(RecordBasedRun run) {
			this.run = run;
		}

		@Override
		public LearnedModel learn(TrainingData trainingData, EvaluationMethod evaluator) {
			WeightsRepository finalWeights = genetic.optimize((weights) -> {

				List<EvaluationPrediction> predictions = new ArrayList<>();

				for (RecordData record : trainingData.getData()) {
					RecordPrediction prediction = run.predict(record.getFeatures(), weights);
					predictions.add(new EvaluationPrediction(record.getLabel(), prediction));
				}

				return evaluator.performance(predictions);
			});

			return new LearnedWeightsModel(run, finalWeights);
		}

		@Override
		public LearnerId getLearnerId() {
			// TODO Auto-generated method stub
			return null;
		}

		@Override
		public LearnedModel parseModelFromFile(String string) {
			// TODO Auto-generated method stub
			return null;
		}

	}
}
