package de.uni_mannheim.informatik.dws.expansion.utils.optimization.learner;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Optional;

import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationMethod;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.EvaluationPrediction;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.LearnedModel;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.RecordPrediction;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordData;
import de.uni_mannheim.informatik.ds.expansion.utils.learning.TrainingData.RecordFeature;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;

public class LearnedWeightsModel implements LearnedModel {

	private final RecordBasedRun run;
	private final WeightsRepository weights;

	public LearnedWeightsModel(RecordBasedRun run, WeightsRepository weights) {
		this.run = run;
		this.weights = weights;
	}

	@Override
	public RecordPrediction apply(RecordFeature f) {
		return run.predict(f, weights);
	}



	@Override
	public double runAndEvaluate(Collection<RecordData> data, EvaluationMethod evaluator) {

		List<EvaluationPrediction> predictions = new ArrayList<>();

		for (RecordData record : data) {
			RecordPrediction prediction = run.predict(record.getFeatures(), weights);
			predictions.add(new EvaluationPrediction(record.getLabel(), prediction));
		}

		return evaluator.performance(predictions);

	}

	public WeightsRepository getWeights() {
		return weights;
	}

	@Override
	public byte[] outputModelFile() {
		throw new IllegalStateException();
	}

	@Override
	public Optional<Double> getNormalizingThreshold() {
		throw new IllegalStateException();
	}

}
