package de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer;

import java.util.HashMap;
import java.util.Map;
import java.util.NavigableMap;
import java.util.Optional;
import java.util.TreeMap;

import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeight;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeightKey;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;

public class WeightsRepositoryCollector implements WeightsRepository {

	private final int STEPS;
	private final float STEP;

	private OptimizableClass workingClass;

	private Map<String, Integer> collectorIndex = new HashMap<>();
	private NavigableMap<Integer, String> collector = new TreeMap<>();


	private boolean locked = false;

	public WeightsRepositoryCollector(int STEPS, float STEP) {
		this.STEPS = STEPS;
		this.STEP = STEP;
	}

	public WeightsCollectorData getWeightsCollectorData() {
		return new WeightsCollectorData(this);
	}
	public synchronized void addWeight(OptimizableWeight weight) {
		if (locked) {
			if (workingClass == null) {
				throw new IllegalStateException("This repository is locked. How is working class null?");
			}
			if (!workingClass.equals(weight.getKey().getOptimizableClazz())) {
				throw new IllegalStateException("This repository is locked. How is there a different optimizable class?");
			}
			if (!collectorIndex.containsKey(weight.getKey().getName())) {
				throw new IllegalStateException("This repository is locked. How is there an additional weight that needs to be optimized.");
			}

			throw new IllegalStateException("This Collector is Already Locked");
		} else {
			if (workingClass == null) {
				workingClass = weight.getKey().getOptimizableClazz();
			} else {
				if (!workingClass.equals(weight.getKey().getOptimizableClazz())) {
					throw new IllegalStateException("You can learn weights only for one optimizable class at a time.");
				}
			}

			String name = weight.getKey().getName();
			if (!collectorIndex.containsKey(name)) {
				int index = collector.size();
				collector.put(index, name);
				collectorIndex.put(name, index);
			}


		}
	}





	public OptimizableClass getWorkingClass() {
		return workingClass;
	}

	public void setWorkingClass(OptimizableClass workingClass) {
		this.workingClass = workingClass;
	}

	public Map<String, Integer> getCollectorIndex() {
		return collectorIndex;
	}

	public void setCollectorIndex(Map<String, Integer> collectorIndex) {
		this.collectorIndex = collectorIndex;
	}

	public NavigableMap<Integer, String> getCollector() {
		return collector;
	}

	public void setCollector(NavigableMap<Integer, String> collector) {
		this.collector = collector;
	}

	public int getSTEPS() {
		return STEPS;
	}

	public float getSTEP() {
		return STEP;
	}

	@Override
	public double getOrDefault(OptimizableWeight weight) {
		if(!weight.getKey().getOptimizableClazz().equals(workingClass) || !collectorIndex.containsKey(weight.getKey().getName())) {
			addWeight(weight);
		}

		double value = weight.getDefaultValue();

		return value;
	}

	@Override
	public Optional<Double> get(OptimizableWeight weight) {
		return Optional.of(getOrDefault(weight));
	}

	@Override
	public boolean containsWeightsForClass(OptimizableClass clazz) {
		if (workingClass != null && workingClass.equals(clazz)) {
			return true;
		}
		return false;
	}

	@Override
	public Map<OptimizableWeightKey, Double> getWeights() {
		throw new UnsupportedOperationException();
	}

	@Override
	public Map<String, Double> getWeights(OptimizableClass optimizableClass) {
		return new HashMap<>();
	}

}
