package de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.fastparallel;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.inject.Inject;

import org.apache.commons.lang.mutable.MutableInt;

import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.OptimizableRun;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.WeightsRepositoryCollector;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeightKey;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepositoryImpl;

public class WeightsFastParallelOptimizer {

	public static float STEP = 0.05f;
	public static int STEPS = Math.round((1.0f / STEP));

	@Inject
	public WeightsFastParallelOptimizer() {

	}

	public WeightsRepository optimize(OptimizableRun run, List<Set<String>> priorRuns) {

		WeightsRepositoryCollector collector = new WeightsRepositoryCollector(STEPS, STEP);
		
		run.run(collector);



		WeightsFastRunner runner;
		if(priorRuns != null) {

			Map<String, MutableInt> sum = new HashMap<>();
			Map<String, MutableInt> count = new HashMap<>();
			for (Set<String> priorRun : priorRuns) {
				WeightsFastRunner priorRunner = new WeightsFastRunner(collector, priorRun);
	
				while (!priorRunner.isDone()) {
					PersistentParallel6.runAndFinish(priorRunner.getNextRuns(), (currentRun) -> {
						WeightsRepository currentWeightsRepository = priorRunner.getRun(currentRun);
						double performance = run.run(currentWeightsRepository);
	
						priorRunner.addPerformance(currentRun, performance);
					}, this.getClass());
				}
	
				Map<String, Integer> priors = priorRunner.getLimitedOptimizedRun();
				for (Entry<String, Integer> entry : priors.entrySet()) {
					CHMHelper.getFromMap(sum, entry.getKey(), () -> new MutableInt()).add(entry.getValue());
					CHMHelper.getFromMap(count, entry.getKey(), () -> new MutableInt()).increment();
				}
	
			}
			Map<String, Integer> priors = new HashMap<>();
			for (Entry<String, MutableInt> entry : count.entrySet()) {
				priors.put(entry.getKey(), sum.get(entry.getKey()).intValue() / count.get(entry.getKey()).intValue());
			}
			
			  runner = new WeightsFastRunner(collector, priors);
		}
		else {
			runner = new WeightsFastRunner(collector);
		}
		
	 
		while (!runner.isDone()) {
			PersistentParallel6.runAndFinish(runner.getNextRuns(), (currentRun) -> {
				WeightsRepository currentWeightsRepository = runner.getRun(currentRun);
				double performance = run.run(currentWeightsRepository);

				runner.addPerformance(currentRun, performance);
			}, this.getClass());
		}
		
		

		Map<OptimizableWeightKey, Double> weights = runner.getLearnedWeights();
		return new WeightsRepositoryImpl(weights);
	}
}
