package de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.fastparallel;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import org.apache.commons.lang.mutable.MutableDouble;

import de.mannheim.uni.expansion.utils.CHMHelper;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.WeightsRepositoryCollector;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeight;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeightKey;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;

public class WeightsFastRunner {

	public static final int MAX_ITERATIONS = 1000;
	public static final int INITIAL_VALUE = 10;
	public static final int MAX_CHANGE = 2;
	private Map<WeightsCurrentRun, Double> currentResults = new ConcurrentHashMap<>();

	private Map<WeightsCurrentRun, Map<String, Integer>> currentRuns = new ConcurrentHashMap<>();

	private final OptimizableClass workingClass;

	private final float STEP;
	private final int STEPS;

	private final Set<Map<String, Integer>> previousRuns = new LinkedHashSet<>();

	private Map<String, Integer> previousRun = null;

	Optional<Boolean> isDone = Optional.of(false);

	private final Set<String> names;

	private boolean firstRun = true;
	private final List<Set<String>> priorRuns;
	private final Set<String> allNames;
	private double previousMax = -0.1;

	public WeightsFastRunner(WeightsRepositoryCollector collector) {

		this.STEP = collector.getSTEP();
		this.STEPS = collector.getSTEPS();
		this.workingClass = collector.getWorkingClass();
		this.names = collector.getCollectorIndex().keySet();
		previousRun = new HashMap<>();
		for (String name : names) {
			previousRun.put(name, INITIAL_VALUE);
		}
		this.priorRuns = null;
		this.allNames = collector.getCollectorIndex().keySet();
	}

	public WeightsFastRunner(WeightsRepositoryCollector collector, Map<String, Integer> priors) {

		this.STEP = collector.getSTEP();
		this.STEPS = collector.getSTEPS();
		this.workingClass = collector.getWorkingClass();
		this.names = collector.getCollectorIndex().keySet();
		previousRun = new HashMap<>();
		for (String name : names) {
			if (priors.containsKey(name))
				previousRun.put(name, priors.get(name));
			else
				previousRun.put(name, 0);

		}
		this.priorRuns = null;
		this.allNames = collector.getCollectorIndex().keySet();
	}

	public WeightsFastRunner(WeightsRepositoryCollector collector, List<Set<String>> priorRuns) {

		this.STEP = collector.getSTEP();
		this.STEPS = collector.getSTEPS();
		this.workingClass = collector.getWorkingClass();
		this.names = collector.getCollectorIndex().keySet();
		previousRun = new HashMap<>();
		for (String name : names) {
			previousRun.put(name, INITIAL_VALUE);
		}
		if (priorRuns != null) {
			this.priorRuns = new ArrayList<>();
			for (Set<String> priorRun : priorRuns) {
				Set<String> newNames = new HashSet<>();
				for (String name : this.names) {
					if (priorRun.contains(name)) {
						newNames.add(name);
					} else {
						for (String include : priorRun) {
							if (name.startsWith(include)) {
								newNames.add(name);
							}
						}
					}
				}
				this.priorRuns.add(newNames);
				break;
			}

		}
		else
			this.priorRuns = null;

		this.allNames = collector.getCollectorIndex().keySet();
	}

	public WeightsFastRunner(WeightsRepositoryCollector collector, Set<String> limitedSet) {

		this.STEP = collector.getSTEP();
		this.STEPS = collector.getSTEPS();
		this.workingClass = collector.getWorkingClass();

		this.allNames = collector.getCollectorIndex().keySet();

		previousRun = new HashMap<>();

		this.priorRuns = null;

		Set<String> newNames = new HashSet<>();
		for (String name : this.allNames) {
			if (limitedSet.contains(name)) {
				newNames.add(name);
			} else {
				for (String include : limitedSet) {
					if (name.startsWith(include)) {
						newNames.add(name);
					}
				}
			}
		}

		this.names = newNames;

	}

	public boolean isDone() {
		if (isDone.isPresent()) {
			return isDone.get();
		}
		else {
			if (currentRuns.size() != currentResults.size()) {
				throw new IllegalStateException("THIS RUN IS NOT FINISHED YET");
			}

			if (firstRun) {

				if (priorRuns != null) {

					Map<String, List<Integer>> priors = new HashMap<>();

					Map<Integer, Map<WeightsCurrentRun, Double>> split = new HashMap<>();
					for (Entry<WeightsCurrentRun, Double> entry : currentResults.entrySet()) {
						int splitKey = entry.getKey().getRunType() / 10;
						CHMHelper.getFromMap(split, splitKey, () -> new HashMap<>()).put(entry.getKey(), entry.getValue());

					}

					for (Entry<Integer, Map<WeightsCurrentRun, Double>> splitEntry : split.entrySet()) {

						Set<String> names = priorRuns.get(splitEntry.getKey());
						Map<String, Integer> intermediate = getFromLowMidHigh(names, splitEntry.getValue(), currentRuns);

						for (Entry<String, Integer> intermediateEntry : intermediate.entrySet()) {
							CHMHelper.getFromMap(priors, intermediateEntry.getKey(), () -> new ArrayList<>()).add(intermediateEntry.getValue());
						}
					}

					this.previousRun = new HashMap<>();

					for (Entry<String, List<Integer>> entry : priors.entrySet()) {
						int sum = 0;
						int count = 0;
						for (Integer r : entry.getValue()) {
							sum += r;
							count++;
						}
						if (count == 0) {
							previousRun.put(entry.getKey(), 0);
						} else {
							previousRun.put(entry.getKey(), sum / count);
						}
					}

					isDone = Optional.of(false);
					this.firstRun = false;
					return false;
				} else {

					this.previousRun = getFromLowMidHigh(names, currentResults, currentRuns);

					for (String otherName : allNames) {
						if (!names.contains(otherName)) {
							this.previousRun.put(otherName, 0);
						}
					}

					isDone = Optional.of(false);
					this.firstRun = false;
					return false;
				}

			}
			else {
				Double maxPerformance = -0.1;
				Map<String, Integer> max = new HashMap<>();

				for (Entry<WeightsCurrentRun, Double> entry : currentResults.entrySet()) {

					if (max.isEmpty() || entry.getValue() > maxPerformance) {
						max = currentRuns.get(entry.getKey());
						maxPerformance = entry.getValue();
					}
				}

				boolean done = false;

				if (Math.abs(maxPerformance - previousMax) < 0.01) {
					done = true;
				} else if (previousRun.equals(max)) {
					done = true;
				} else {
					boolean added = previousRuns.add(previousRun);

					if (!added) {
						previousRun = max;
						done = true;
					} else if (previousRuns.size() == MAX_ITERATIONS) {
						previousRun = max;
						done = true;
					}
				}



				if (done) {
					isDone = Optional.of(true);
				} else {
					isDone = Optional.of(false);
				}

				return done;
			}


		}
	}

	public Map<String, Integer> getFromLowMidHigh(Set<String> names, Map<WeightsCurrentRun, Double> results,
			Map<WeightsCurrentRun, Map<String, Integer>> runs) {
		Map<String, Integer> maxRunLow = new HashMap<>();
		Map<String, Integer> maxRunMiddle = new HashMap<>();
		Map<String, Integer> maxRunHigh = new HashMap<>();

		Map<String, MutableDouble> maxLow = new HashMap<>();
		Map<String, MutableDouble> maxMiddle = new HashMap<>();
		Map<String, MutableDouble> maxHigh = new HashMap<>();

		List<WeightsCurrentRun> keysSorted = results.keySet().stream().sorted(Collections.reverseOrder((e1, e2) -> {
			int compare = e1.getName().compareTo(e2.getName());
			if (compare != 0)
				return compare;
			compare = Integer.compare(e1.getRunType(), e2.getRunType());
			if (compare != 0)
				return compare;
			compare = Integer.compare(e1.getRun(), e2.getRun());
			if (compare == 0) {
				int sdf = 13;
			}
			return compare;
		})).collect(Collectors.toList());

		for (WeightsCurrentRun run : keysSorted) {
			Double performance = results.get(run);
			int runType = run.getRunType() % 10;
			Map<String, Integer> runData = runs.get(run);
			if (runType == 0) {
				MutableDouble md = CHMHelper.getFromMap(maxLow, run.getName(), () -> new MutableDouble(0.0001));
				if (performance > md.doubleValue()) {
					md.setValue(performance);
					maxRunLow.put(run.getName(), runData.get(run.getName()));
				}
			} else if (runType == 1) {
				MutableDouble md = CHMHelper.getFromMap(maxMiddle, run.getName(), () -> new MutableDouble(0.0001));
				if (performance > md.doubleValue()) {
					md.setValue(performance);
					maxRunMiddle.put(run.getName(), runData.get(run.getName()));
				}
			} else if (runType == 2) {
				MutableDouble md = CHMHelper.getFromMap(maxHigh, run.getName(), () -> new MutableDouble(0.0001));
				if (performance > md.doubleValue()) {
					md.setValue(performance);
					maxRunHigh.put(run.getName(), runData.get(run.getName()));
				}
			}

		}

		Map<String, Integer> previousRun = new HashMap<>();

		for (String name : names) {
			int count = 0;
			int sum = 0;
			if (maxRunLow.containsKey(name)) {
				count++;
				sum += maxRunLow.get(name);
			}
			if (maxRunMiddle.containsKey(name)) {
				count++;
				sum += maxRunMiddle.get(name);
			}
			if (maxRunHigh.containsKey(name)) {
				count++;
				sum += maxRunHigh.get(name);

			}
			if (sum == 0)
				previousRun.put(name, 0);
			else {
				Integer i = sum / count;
				previousRun.put(name, i);
			}
		}

		return previousRun;
	}

	public Set<WeightsCurrentRun> getNextRuns() {
		if (isDone.isPresent() && isDone.get()) {
			throw new IllegalStateException("This optimizing run is done, why are you calling this method");
		} else if (isDone.isPresent()) {

			if (firstRun) {

				if (priorRuns != null) {

					Map<WeightsCurrentRun, Map<String, Integer>> newCurrentRuns = new HashMap<>();

					int base = 0;
					for (Set<String> names : priorRuns) {
						int tenBase = base * 10;
						Map<String, Integer> lowRun = new HashMap<>();
						Map<String, Integer> middleRun = new HashMap<>();
						Map<String, Integer> highRun = new HashMap<>();

						for (String name : names) {
							lowRun.put(name, 0);
							middleRun.put(name, INITIAL_VALUE);
							highRun.put(name, STEPS);

							for (String other : this.names) {
								if (!names.contains(other)) {
									lowRun.put(other, 0);
									middleRun.put(other, 0);
									highRun.put(other, 0);
								}
							}
						}

						for (String name : names) {

							int start = 0;
							int end = STEPS;
							for (int i = start; i <= end; i++) {
								Map<String, Integer> low = new HashMap<>(lowRun);
								Map<String, Integer> middle = new HashMap<>(middleRun);
								Map<String, Integer> high = new HashMap<>(highRun);

								low.put(name, i);
								newCurrentRuns.put(new WeightsCurrentRun(name, tenBase + 0, i), low);
								middle.put(name, i);
								newCurrentRuns.put(new WeightsCurrentRun(name, tenBase + 1, i), middle);
								high.put(name, i);
								newCurrentRuns.put(new WeightsCurrentRun(name, tenBase + 2, i), high);

							}

						}

						base++;

					}
					currentRuns = newCurrentRuns;
					currentResults = new ConcurrentHashMap<>();

					isDone = Optional.empty();
					return currentRuns.keySet();
				}
				else {
					Map<WeightsCurrentRun, Map<String, Integer>> newCurrentRuns = new HashMap<>();

					Map<String, Integer> lowRun = new HashMap<>();
					Map<String, Integer> middleRun = new HashMap<>();
					Map<String, Integer> highRun = new HashMap<>();

					for (String name : names) {
						lowRun.put(name, 0);
						middleRun.put(name, INITIAL_VALUE);
						highRun.put(name, STEPS);
					}

					for (String otherName : allNames) {
						if (!names.contains(otherName)) {
							lowRun.put(otherName, 0);
							middleRun.put(otherName, 0);
							highRun.put(otherName, 0);
						}
					}

					for (String name : names) {

						int start = 0;
						int end = STEPS;
						for (int i = start; i <= end; i++) {
							Map<String, Integer> low = new HashMap<>(lowRun);
							Map<String, Integer> middle = new HashMap<>(middleRun);
							Map<String, Integer> high = new HashMap<>(highRun);

							low.put(name, i);
							newCurrentRuns.put(new WeightsCurrentRun(name, 0, i), low);
							middle.put(name, i);
							newCurrentRuns.put(new WeightsCurrentRun(name, 1, i), middle);
							high.put(name, i);
							newCurrentRuns.put(new WeightsCurrentRun(name, 2, i), high);

						}

					}


					currentRuns = newCurrentRuns;
					currentResults = new ConcurrentHashMap<>();

					isDone = Optional.empty();
					return currentRuns.keySet();
				}
			}
			else {
				Map<WeightsCurrentRun, Map<String, Integer>> newCurrentRuns = new HashMap<>();

				for (String name : names) {

					int start = previousRun.get(name) - MAX_CHANGE;
					if (start < 0)
						start = 0;
					int end = previousRun.get(name) + MAX_CHANGE;
					if (end > STEPS)
						end = STEPS;
					start = 0;
					end = STEPS;
					for (int i = start; i <= end; i++) {
						Map<String, Integer> run = new HashMap<>(previousRun);

						run.put(name, i);
						newCurrentRuns.put(new WeightsCurrentRun(name, 0, i), run);
					}

				}


				currentRuns = newCurrentRuns;
				currentResults = new ConcurrentHashMap<>();

				isDone = Optional.empty();
				return currentRuns.keySet();
			}



		} else {
			return currentRuns.keySet();
		}

	}
	void addPerformance(WeightsCurrentRun run, double performance) {
		currentResults.put(run, performance);
	}



	WeightsRepository getRun(WeightsCurrentRun run) {
		return new WeightsRepositorySingleRun(run);
	}

	public class WeightsRepositorySingleRun implements WeightsRepository {

		private final WeightsCurrentRun run;

		private WeightsRepositorySingleRun(WeightsCurrentRun run) {
			super();
			this.run = run;
		}

		@Override
		public double getOrDefault(OptimizableWeight weight) {
			if (workingClass.equals(weight.getKey().getOptimizableClazz())) {
				int currentStep = currentRuns.get(run).get(weight.getKey().getName());

				double value = Math.round(100.0 * ((currentStep * STEP))) / 100.0;

				return value;
			}

			throw new IllegalStateException();
		}

		@Override
		public Optional<Double> get(OptimizableWeight weight) {
			return Optional.of(getOrDefault(weight));
		}

		@Override
		public boolean containsWeightsForClass(OptimizableClass clazz) {
			if (workingClass != null && workingClass.equals(clazz)) {
				return true;
			}
			return false;
		}

		@Override
		public Map<OptimizableWeightKey, Double> getWeights() {
			Map<OptimizableWeightKey, Double> ret = new HashMap<>();

			Map<String, Integer> currentRun = currentRuns.get(run);
			for (Entry<String, Integer> weightEntry : currentRun.entrySet()) {
				int currentStep = weightEntry.getValue();

				double value = Math.round(100.0 * ((currentStep * STEP))) / 100.0;
				ret.put(new OptimizableWeightKey(workingClass, weightEntry.getKey()), value);
			}
			return ret;
		}

		@Override
		public Map<String, Double> getWeights(OptimizableClass optimizableClass) {
			Map<String, Double> ret = new HashMap<>();

			Map<String, Integer> currentRun = currentRuns.get(run);
			for (Entry<String, Integer> weightEntry : currentRun.entrySet()) {
				int currentStep = weightEntry.getValue();

				double value = Math.round(100.0 * ((currentStep * STEP))) / 100.0;
				ret.put(weightEntry.getKey(), value);
			}
			return ret;
		}

	}

	public static class WeightsCurrentRun {
		private final String name;
		private final int runType;
		private final int run;

		private WeightsCurrentRun(String name, int runType, int run) {
			super();
			this.name = name;
			this.runType = runType;
			this.run = run;
		}

		public int getRunType() {
			return runType;
		}



		private String getName() {
			return name;
		}

		private int getRun() {
			return run;
		}

	}

	public Map<OptimizableWeightKey, Double> getLearnedWeights() {
		if (!isDone.isPresent() || !isDone.get()) {
			throw new IllegalStateException("This optimizing run is not done, why are you calling this method");
		}

		Map<OptimizableWeightKey, Double> weights = previousRun.entrySet().stream().collect(Collectors.toMap((e) -> {
			return new OptimizableWeightKey(workingClass, e.getKey());
		}, (e) -> {
			return Double.valueOf(Math.round(100.0 * ((e.getValue() * STEP))) / 100.0);
		}));
		return weights;
	}

	public Map<String, Integer> getOptimizedRun() {
		return previousRun;
	}

	public Map<String, Integer> getLimitedOptimizedRun() {
		Map<String, Integer> ret = new HashMap<>();
		for (String name : names) {
			ret.put(name, previousRun.get(name));
		}
		return ret;
	}
}
