package de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.genetic;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NavigableMap;
import java.util.Optional;
import java.util.Set;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.WeightsCollectorData;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.random.WeightsRandomRunner.WeightsCurrentRun;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableClass;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeight;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeightKey;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;

public class WeightsPopulationRunner {

	private Map<WeightsCurrentRun, Double> results = new ConcurrentHashMap<>();

	private final Map<WeightsCurrentRun, Map<String, Integer>> runs;

	private final OptimizableClass workingClass;

	private final float STEP;
	private final int STEPS;


	Optional<Boolean> isDone = Optional.of(false);

	private NavigableMap<WeightsCurrentRun, Double> rankedRuns;

	public WeightsPopulationRunner(Map<WeightsCurrentRun, Map<String, Integer>> runs, WeightsCollectorData collectorData) {
		this.runs = runs;
		this.STEP = collectorData.getSTEP();
		this.STEPS = collectorData.getSTEPS();
		this.workingClass = collectorData.getWorkingClass();
		// this.names = collectorData.getNames();
		// namesIndex = collectorData.getNamesIndex();
		// namesIndexReverse = collectorData.getNamesIndexReverse();

	}

 

	 

	 

	public boolean isDone() {
		if (results.size() != runs.size()) {
			return false;
		}
		else {

			
			this.rankedRuns =  new TreeMap<>((e1, e2) -> {
				int compare = Long.compare(Math.round(10000000 * results.get(e1)), Math.round(10000000 * results.get(e2)));
				if(compare != 0) {
					return compare;
				} 
				else {
					return Integer.compare(e2.getRun(), e1.getRun());
				}
			}) ;
			
			this.rankedRuns.putAll(results);

			this.rankedRuns = Collections.unmodifiableNavigableMap(this.rankedRuns);

			return true;
		}
	}



	

	public Set<WeightsCurrentRun> getRuns() {
	 return runs.keySet();

	}
	void addPerformance(WeightsCurrentRun run, double performance) {
		if(runs.containsKey(run))
			results.put(run, performance);
		else
			throw new IllegalStateException();
	}



	Map<String, Integer> getRawRun(WeightsCurrentRun run) {
		return runs.get(run);
	}

	WeightsRepository getRun(WeightsCurrentRun run) {
		return new WeightsRepositorySingleRun(run);
	}

	public class WeightsRepositorySingleRun implements WeightsRepository {

		private final WeightsCurrentRun run;

		private WeightsRepositorySingleRun(WeightsCurrentRun run) {
			super();
			this.run = run;
		}

		@Override
		public double getOrDefault(OptimizableWeight weight) {
			if (workingClass.equals(weight.getKey().getOptimizableClazz())) {
				int currentStep = runs.get(run).get(weight.getKey().getName());

				double value = Math.round(100.0 * ((currentStep * STEP))) / 100.0;

				return value;
			}

			throw new IllegalStateException();
		}

		@Override
		public Optional<Double> get(OptimizableWeight weight) {
			return Optional.of(getOrDefault(weight));
		}

		@Override
		public boolean containsWeightsForClass(OptimizableClass clazz) {
			if (workingClass != null && workingClass.equals(clazz)) {
				return true;
			}
			return false;
		}

		@Override
		public Map<OptimizableWeightKey, Double> getWeights() {
			Map<OptimizableWeightKey, Double> ret = new HashMap<>();

			Map<String, Integer> currentRun = runs.get(run);
			for (Entry<String, Integer> weightEntry : currentRun.entrySet()) {
				int currentStep = weightEntry.getValue();

				double value = Math.round(100.0 * ((currentStep * STEP))) / 100.0;
				ret.put(new OptimizableWeightKey(workingClass, weightEntry.getKey()), value);
			}
			return ret;
		}

		@Override
		public Map<String, Double> getWeights(OptimizableClass optimizableClass) {
			Map<String, Double> ret = new HashMap<>();

			Map<String, Integer> currentRun = runs.get(run);
			for (Entry<String, Integer> weightEntry : currentRun.entrySet()) {
				int currentStep = weightEntry.getValue();

				double value = Math.round(100.0 * ((currentStep * STEP))) / 100.0;
				ret.put(weightEntry.getKey(), value);
			}
			return ret;
		}

	}



	public Map<OptimizableWeightKey, Double> getLearnedWeights() {
		if (!isDone()) {
			throw new IllegalStateException("This optimizing run is not done, why are you calling this method");
		}

		Map<OptimizableWeightKey, Double> weights = runs.get(rankedRuns.lastKey()).entrySet().stream().collect(Collectors.toMap((e) -> {
			return new OptimizableWeightKey(workingClass, e.getKey());
		}, (e) -> {
			return Double.valueOf(Math.round(100.0 * ((e.getValue() * STEP))) / 100.0);
		}));
		return weights;
	}

	public Map<String, Integer> getOptimizedRun() {

		if (!isDone()) {
			throw new IllegalStateException("This optimizing run is not done, why are you calling this method");
		}

		return runs.get(rankedRuns.lastKey());
	}

	public NavigableMap<WeightsCurrentRun, Double> getRankedRuns() {

		if (!isDone()) {
			throw new IllegalStateException("This optimizing run is not done, why are you calling this method");
		}

		return rankedRuns;
	}
 
}
