package de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.parallel;

import java.util.Map;

import javax.inject.Inject;

import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.OptimizableRun;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.WeightsOptimizer;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.optimizer.WeightsRepositoryCollector;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.OptimizableWeightKey;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepository;
import de.uni_mannheim.informatik.dws.expansion.utils.optimization.weights.WeightsRepositoryImpl;

public class WeightsParallelOptimizer implements WeightsOptimizer {

	public static float STEP = 0.05f;
	public static int STEPS = Math.round((1.0f / STEP));

	@Inject
	public WeightsParallelOptimizer() {

	}

	@Override
	public WeightsRepository optimize(OptimizableRun run) {

		WeightsRepositoryCollector collector = new WeightsRepositoryCollector(STEPS, STEP);
		
		run.run(collector);
		
		WeightsRepositoryRunner runner = new WeightsRepositoryRunner(collector);
		
	 
		
		
		PersistentParallel6.runAndFinish(runner.getAllRuns(), (currentRun) -> {
			WeightsRepository currentWeightsRepository = runner.getRun(currentRun);
			double performance = run.run(currentWeightsRepository);
			
			runner.addPerformance(currentRun, performance);
		}, this.getClass());

		Map<OptimizableWeightKey, Double> weights = runner.getLearnedWeights();
		return new WeightsRepositoryImpl(weights);
	}
}
