package de.mannheim.uni.expansion.utils;

import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Consumer;

import de.mannheim.uni.expansion.utils.PersistentParallel6.AlreadyParalell;
import de.mannheim.uni.expansion.utils.PersistentParallel6.JobLevel; 

public class BatchConsumerProducer<E> {

	// private final PersistentParallel3<E> pp;
	private   Consumer<E> body;

	
	private final LinkedBlockingDeque<Job> jobQueue = new LinkedBlockingDeque<>();
	
	
//	final AtomicReference<CountDownLatch> latch = new AtomicReference<>();
	
	private final CountDownLatch killLatch;
	private final int subBatchSize;
	private final int numOfProcessors;
	private final String name;
	private boolean running = true;
	private final AtomicInteger totalAddedTasks = new AtomicInteger();
	private final int innerThreadCount;
	private BatchConsumerProducer(String name, Consumer<E> body, int subtractProcessors, int subBatchSize ) {
		this(name, body, subBatchSize, subBatchSize, -1);
	}
	private BatchConsumerProducer(String name, Consumer<E> body, int subtractProcessors, int subBatchSize, int multiRun) {

		this.numOfProcessors = NumOfCpusGetter.getProcessors() - subtractProcessors;
		this.subBatchSize = subBatchSize;
		// pp = new PersistentParallel3<>(name, subBatchSize, numOfProcessors);
		this.body = body;
		this.name = name;

//		latch.set(new CountDownLatch(1));
//		// We can start right away, therefore we count down!
//		latch.get().countDown();
		
		if(multiRun > 0) {
			innerThreadCount = multiRun;
			for(int i = 0; i < multiRun;i++) {
				ParallelThread runnable = new ParallelThread();
				Thread t = new Thread(runnable);
				t.start();
			}
			killLatch = new CountDownLatch(multiRun);
		
		}
		else {
			innerThreadCount = 1;
			ParallelThread runnable = new ParallelThread();
			Thread t = new Thread(runnable);
			t.start();
			killLatch = new CountDownLatch(1);
		}
	}
	public static <E> BatchConsumerProducer<E> createMultiRun(String name, Consumer<E> body, int multiRun) {
		return new BatchConsumerProducer<>(name, body, 0, -1, multiRun);
	}
	
	public static <E> BatchConsumerProducer<E> createMultiRun(String name, Consumer<E> body, int multiRun, int subBatchSize) {
		return new BatchConsumerProducer<>(name, body, 0, subBatchSize, multiRun);
	}
	
	public static <E> BatchConsumerProducer<E> create(String name, Consumer<E> body) {
		return new BatchConsumerProducer<>(name, body, 0, -1);
	}

	public static <E> BatchConsumerProducer<E> createAutomaticBatchSize(String name, Consumer<E> body) {
		return new BatchConsumerProducer<>(name, body, 0, -1);
	}

	public static <E> BatchConsumerProducer<E> createAutomaticBatchSize(String name, Consumer<E> body, int subtractProcessors) {
		return new BatchConsumerProducer<>(name, body, subtractProcessors, -1);
	}

	public static <E> BatchConsumerProducer<E> create(String name, Consumer<E> body, int subBatchSize) {
		return new BatchConsumerProducer<>(name, body, 0, subBatchSize);
	}

	public static <E> BatchConsumerProducer<E> create(String name, Consumer<E> body, int subBatchSize, int subtractProcessors) {
		return new BatchConsumerProducer<>(name, body, subtractProcessors, subBatchSize);
	}




	public   void addBatch(List<E> list) {
if(running)
	totalAddedTasks.incrementAndGet();
	jobQueue.addLast(new Job(list));
//		try {
			
			
//			PersistentParallel6.setThreadPaused();
//			latch.get().await();
//			PersistentParallel6.setThreadUnpaused();
//
//			CountDownLatch newLatch = new CountDownLatch(1);
//			latch.set(newLatch);
//
//			AtomicReference<List<E>> listReference = new AtomicReference<List<E>>(list);
//			list = null;
//
//			Runnable r2 = () -> {
//
//
//
//				try {
//					PersistentParallel6.runAndFinish(listReference.get(), body, name);
//				} catch (Exception e1) {
//					// TODO Auto-generated catch block
//					e1.printStackTrace();
//				}
//
//				listReference.set(null);
//
//				newLatch.countDown();
//
//			};
//
//			Thread t = new Thread(r2);
//			t.start();
//		} catch (InterruptedException e) {
//			// TODO Auto-generated catch block
//			e.printStackTrace();
//		}

	}

	public void finish() {
		try {
//			this.running = false;
			running = false;
			for(int i = 0; i <innerThreadCount; i++ ) {
				jobQueue.addLast(new Job());
			}
			killLatch.await();
			this.body = null;
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	private final AtomicInteger finishedJobs = new AtomicInteger(0);
	
	private class ParallelThread implements Runnable {

		@Override
		public void run() { 

//			InnerThread innerThreadRunnable = getInnerThread();
//			int finishedJobs = 0;
			while (true) {
				try {
					JobLevel.setJobLevel(-1);
					
					Job job = jobQueue.takeFirst();
					Thread.currentThread().setName(name);
					if (job != null) {

						if (job.isKillThread()) {

							killLatch.countDown();
						 
							return;
						}
						else {
							job.run(); 
							System.err.println("Finished job " + finishedJobs.incrementAndGet() + " out of " + totalAddedTasks.intValue());
						}
//						else if (job.pauseThread()) {
//							CountDownLatch latch = new CountDownLatch(1);
//							MutableBoolean kill = new MutableBoolean(false);
//							threadPool.put(new PoolEntry(latch, kill));
//							latch.await();
//							if (kill.isTrue()) {
//								innerThreadRunnable.kill();
//								ReporterThread reporterThreadRunnable = pp.reporterThread.get();
//								if (reporterThreadRunnable != null)
//									reporterThreadRunnable.kill();
//								return;
//							}
//						}

					}
				} catch (InterruptedException e) {
					throw new RuntimeException(e);
				}

			}
		}
	}
	
	
	private class Job   {
		private final boolean kill;
		private final List<E> run;
		private Job(List<E> run) {
			this.run = run;
			this.kill =false;
		}

		private Job() {
			this.run = null;
			this.kill =true;
		}
		private void run() {
			if(!kill) {
				if(subBatchSize > -1) {
					PersistentParallel6.runAndFinishWithBatchSize(run, body, name, subBatchSize);
				}
				else {
					PersistentParallel6.runAndFinish(run, body, name);
				}
			}
		}


		private boolean isKillThread() {
			return kill;
		}
 
	}
}
