package de.mannheim.uni.expansion.utils;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Consumer;


public class PersistentParallel<T> {

	private static int defaultNumProcessors = Runtime.getRuntime().availableProcessors();
	private final static AtomicReference<PersistentParallel<?>> currentTask = new AtomicReference<>(null);
	private int overrideNumProcessors = 0;

	private final List<Thread> pool;
	private final List<LinkedList<T>> queues;
	private final int numOfThreads;

	private final ThreadEvent threadEvent = new ThreadEvent();

	private boolean running = true;
	private Consumer<T> consumer;

	private AtomicReference<CountDownLatch> latch = new AtomicReference<>();
	private AtomicReference<CountDownLatch> mainLatch = new AtomicReference<>();

	public static class ThreadEvent {

		private final Object lock = new Object();

		public void signal() {
			synchronized (lock) {
				lock.notify();
			}
		}

		public void await() throws InterruptedException {
			synchronized (lock) {
				lock.wait();
			}
		}
	}

	public PersistentParallel(String name) {

		mainLatch.set(new CountDownLatch(1));

		numOfThreads = getNumProcessors();

		latch.set(new CountDownLatch(numOfThreads));

		this.pool = new ArrayList<>();
		this.queues = new ArrayList<>();

		for (int i = 0; i < numOfThreads; i++) {

			final LinkedList<T> queue = new LinkedList<>();

			Thread t = new Thread(() -> {
				PersistentParallel3.setCurrentThreadParallel(true);
				try {
					CountDownLatch mainLatch = this.mainLatch.get();
					this.latch.get().countDown();
					mainLatch.await();
				} catch (Exception e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}

				int a = 5;

				while (running) {

					Consumer<T> consumer = getConsumer();

					T item = queue.poll();
					while (item != null) {

						try {
							consumer.accept(item);
						} catch (Exception e1) {
							e1.printStackTrace();
							CountDownLatch latch = this.latch.get();
							latch.countDown();
							return;
						}

						item = queue.poll();
					}

					CountDownLatch mainLatch = this.mainLatch.get();

					CountDownLatch latch = this.latch.get();
					latch.countDown();

					try {
						mainLatch.await();
					} catch (Exception e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}

			} , name + " - " + i);

			t.start();

			queues.add(queue);
			pool.add(t);
		}

		try {
			this.latch.get().await();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private Consumer<T> getConsumer() {
		return consumer;
	}

	private int getNumProcessors() {
		if (overrideNumProcessors > 0)
			return overrideNumProcessors;
		else
			return defaultNumProcessors;
	}

	public synchronized void foreach(Iterable<T> items, final Consumer<T> body) throws Exception {
		if (getNumProcessors() > 1 && !PersistentParallel3.isCurrentThreadParallel()) {

			this.consumer = body;
			this.latch.set(new CountDownLatch(numOfThreads));

			Iterator<T> it = items.iterator();
			while (it.hasNext()) {

				for (int i = 0; i < numOfThreads; i++) {
					if (it.hasNext()) {
						queues.get(i).add(it.next());
					}
				}

			}

			CountDownLatch mainLatch = this.mainLatch.get();
			this.mainLatch.set(new CountDownLatch(1));
			mainLatch.countDown();

			this.latch.get().await();
		} else {
			sequentialFor(items, body);
		}
	}

	private void sequentialFor(Iterable<T> items, final Consumer<T> body) {

		for (T item : items) {
			body.accept(item);
		}

	}

	public synchronized void shutdown() {
		running = false;
		mainLatch.get().countDown();

	}

	public static <E> void runAndFinish(Iterable<E> items, final Consumer<E> body, String name) {
		PersistentParallel<E> pp = new PersistentParallel<>(name);
		try {
			pp.foreach(items, body);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		pp.shutdown();
	}

	public static void forLoop(int from, int to, Consumer<Integer> loopBody, String name) {
		List<Integer> lst = new LinkedList<Integer>();

		for (int i = from; i < to; i++)
			lst.add(i);

		PersistentParallel.runAndFinish(lst, loopBody, name);

	}

	public static void overideNumberOfProcessors(int num) {
		defaultNumProcessors = num;
	}
}
