package de.mannheim.uni.expansion.utils;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.function.Consumer;


public class PersistentParallel2<T> {

	private static int defaultNumProcessors = Runtime.getRuntime().availableProcessors();
	private static PersistentParallel2<?> currentTask = null;
	private int overrideNumProcessors = 0;

	private final ThreadPoolExecutor pool;

	public PersistentParallel2() {
		this.pool = this.getPool();
	}

	public ThreadPoolExecutor getPool() {
		return new ThreadPoolExecutor(getNumProcessors(this), getNumProcessors(this), 0, TimeUnit.SECONDS,
				new java.util.concurrent.LinkedBlockingQueue<Runnable>(), new ThreadFactory() {

					public Thread newThread(Runnable r) {
						return new Thread(r, "Parallel.foreach thread");
					}
				});
	}

	private int getNumProcessors() {
		if (overrideNumProcessors > 0)
			return overrideNumProcessors;
		else
			return defaultNumProcessors;
	}

	private static int getNumProcessors(PersistentParallel2<?> obj) {
		// if this is a nested parallel process, only use 1 thread ...
		if (currentTask == null || currentTask == obj)
			return obj.getNumProcessors();
		else
			return 1;
	}

	public synchronized void foreach(Iterable<T> items, final Consumer<T> body) throws Exception {
		if (startParallelProcess(this) && getNumProcessors(this) > 1) {

			List<Callable<Boolean>> tasks = new LinkedList<Callable<Boolean>>();

			Iterator<T> it = items.iterator();
			while (it.hasNext()) {
				final T value = it.next();
				tasks.add(new Callable<Boolean>() {

					@Override
					public Boolean call() throws Exception {

						body.accept(value);

						return true;
					}
				});
			}

			pool.invokeAll(tasks);
		} else {
			sequentialFor(items, body);
		}
		endParallelProcess(this);
	}

	protected void sequentialFor(Iterable<T> items, final Consumer<T> body) {

		for (T item : items) {
			body.accept(item);
		}

	}

	private static boolean startParallelProcess(PersistentParallel2<?> obj) {
		// only set obj as current parallel task if it uses more than one thread
		// ...
		if (currentTask == null && obj.getNumProcessors() > 1) {
			currentTask = obj;
			return true;
		} else
			return false;
	}

	private static void endParallelProcess(PersistentParallel2<?> obj) {
		if (currentTask == obj)
			currentTask = null;
	}
}
