package de.mannheim.uni.expansion.utils;

import java.sql.Time;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.atomic.AtomicReference;
import java.util.concurrent.locks.ReentrantLock;
import java.util.function.Consumer;

import org.apache.commons.lang.mutable.MutableInt;
import org.apache.commons.lang.mutable.MutableLong;
import org.apache.commons.lang3.time.DurationFormatUtils;

import de.mannheim.uni.expansion.utils.PersistentParallelFactory.PersistentParallelInterface;

public class PersistentParallel6 {

	private static int numOfThreads = Math.min(64, NumOfCpusGetter.getProcessors());
	private static int numOfThreadsMultiplier = 4; //16;
	private static int numOfBatchesMultiplier = 4; //4;
	private static int defaultBatchSize = 5000;
	private static int maxDepth = 5;//12;

	private ScheduledExecutorService scheduledExecutor = Executors.newScheduledThreadPool(numOfThreads);
	private static PersistentParallel6 pp = new PersistentParallel6();
	private static final int maxJobLevel = 3;

	private final boolean running = true;

	private final LinkedBlockingDeque<Job> jobQueue = new LinkedBlockingDeque<>();
	private final ReentrantLock[] jobLevelLocks;
//	private final LinkedBlockingQueue<PoolEntry> threadPool = new LinkedBlockingQueue<>();

	private PersistentParallel6() {
		for (int i = 0; i < numOfThreads * numOfThreadsMultiplier; i++) {
//			addThreadToMainPool();
			Thread t = new Thread(new ParallelThread());
			t.start();
		}
		
		jobLevelLocks = new ReentrantLock[maxJobLevel];
		for(int i = 0; i < maxJobLevel; i++) {
			jobLevelLocks[i] = new ReentrantLock();
		}
	}

	private <T> void startInnerThread() {
		InnerThread innerThread = this.getInnerThread();
		innerThread.addToPool();

	}

//	private void addThreadToMainPool() {
//		PoolEntry pair = threadPool.poll();
//		if (pair != null) {
//			CountDownLatch latch = pair.getLock();
//			latch.countDown();
//		} else {
//			Thread t = new Thread(new ParallelThread());
//			t.start();
//		}
//
//	}

	private static long divideButRoundUp(long num, int divisor) {
		return (num + divisor - 1) / divisor;
	}

	private boolean printProgress(String name, long tasks, long done, long last, long start, long lastTime) {

		if (done == tasks)
			return false;

		long soFar = System.currentTimeMillis() - start;
		long pauseTime = System.currentTimeMillis() - lastTime;
		// long perItem = (long) (((float) soFar) / done);
		long left = (long) (((float) soFar / done) * (tasks - done));
		float itemsPerSecAvg = (float) done / (float) (soFar / 1000.0f);
		float itemsPerSecNow = (float) (done - last) / (pauseTime / 1000.0f);

		if ((((float) soFar) / done) == Float.POSITIVE_INFINITY) {
			// perItem = -1;
			left = -1;
		}
		String ttl = DurationFormatUtils.formatDuration(soFar, "HH:mm:ss.S");
		// String each = DurationFormatUtils.formatDuration(perItem,
		// "HH:mm:ss.S");
		String remaining = DurationFormatUtils.formatDuration(left, "HH:mm:ss.S");

		System.err.println(String.format("%s: %,d of %,d tasks completed after %s (%d/%d active threads). Avg: %.2f items/s, Current: %.2f items/s, %s left.",
				name, done, tasks, ttl, numOfThreads * 4, numOfThreads * 4, itemsPerSecAvg, itemsPerSecNow, remaining));

		last = done;
		lastTime = System.currentTimeMillis();

		if (done == tasks) {
			return false;
		} else {
			return true;
		}
		// 368 of 1.370 tasks completed after 00:15:19.856 (8/8 active threads).
		// Avg: 0,40 items/s, Current: 0,80 items/s, 00:41:44.607 left.

	}

	private static <T> Batches<T> createBatches(Iterable<T> items, long batchSizeForQueue) {
		List<BatchQueue<T>> queues = new ArrayList<>();
		ConcurrentLinkedQueue<T> batch = new ConcurrentLinkedQueue<>();
		int i = 0;
		long countSize = 0;
		Iterator<T> it = items.iterator();
		while (it.hasNext()) {
			if (i == batchSizeForQueue) {
				queues.add(new BatchQueue<>(batch, i));
				i = 0;
				batch = new ConcurrentLinkedQueue<>();
			}

			countSize++;
			batch.add(it.next());
			i++;
		}
		queues.add(new BatchQueue<>(batch, i));
		return new Batches<>(queues, countSize);
	}

	public <T extends Object> void foreach(Iterable<T> items, long size, final Consumer<T> body, String name, int batchSize, boolean reportProgress,
			int maxConcurrent) throws Exception {
		foreach(items, size, body, name, batchSize, reportProgress, maxConcurrent, null);
		
	}
	public <T extends Object> void foreach(Iterable<T> items, long size, final Consumer<T> body, String name, int batchSize, boolean reportProgress,
			int maxConcurrent, AtomicLong done) throws Exception {
		foreach(items, size, body, name, batchSize, reportProgress, maxConcurrent, done, false);
	}
	public <T extends Object> void foreach(Iterable<T> items, long size, final Consumer<T> body, String name, int batchSize, boolean reportProgress,
			int maxConcurrent, AtomicLong done, boolean keepJobLevel) throws Exception {
		if(body == null)
			System.out.println("BODY EQUALS NULL FOR: " + name);
		
		
		if (size == 0)
			return;
		if(size == 1) {
			body.accept(items.iterator().next());
			if(done != null) {
				done.incrementAndGet();
			}
			return;
		}
		int depth = getDepth();
		if(false && depth > this.maxDepth + 1){
			throw new IllegalStateException();
//			sequentialFor(items, body, done);
//			return;
		}
		
		
		boolean isParallel = AlreadyParalell.isCurrentThreadParallel();
		int jobLevel = 0;
		if(isParallel) {

			
			if(keepJobLevel) {
				jobLevel = JobLevel.getJobLevel();
				startInnerThread();
			}
			else {
				jobLevel = JobLevel.getJobLevel() + 1;
				if(jobLevel >= PersistentParallel6.maxJobLevel) {
					sequentialFor(items, body, done);
					return;
				}
				else {
					startInnerThread();
					jobLevelLocks[jobLevel].lock();
				}
			}
		}
		
		//
		// if (reportProgress) {
		// this.totalTasks = 0;
		// this.start = System.currentTimeMillis();
		// this.threadCounter = new ThreadCounter();
		//
		// Runnable progressReporter = () -> {
		// try {
		// boolean stop = false;
		// while (!stop) {
		//
		// Thread.sleep(10000);
		//
		// boolean cont = printProgress();
		// if (!cont) {
		// stop = true;
		// }
		//
		// }
		// } catch (Exception e) {
		// // TODO Auto-generated catch block
		// e.printStackTrace();
		// }
		// };
		//
		// Thread progressReporterThread = new Thread(progressReporter);
		// progressReporterThread.start();
		// }

		long batchSizeForQueue;
		if (size > 0 && batchSize == -1) {
			if (maxConcurrent > 0 && maxConcurrent < numOfThreads) {
				batchSizeForQueue = divideButRoundUp(size, maxConcurrent * numOfBatchesMultiplier);
			} else {
				batchSizeForQueue = divideButRoundUp(size,   numOfThreads * numOfBatchesMultiplier );
			}

//			if (batchSizeForQueue > 1000) {
//				batchSizeForQueue = divideButRoundUp(batchSizeForQueue, 10);
//			}
		} else {
			batchSizeForQueue = batchSize;
		}
//System.out.println("Batch size for " + name + " is " + batchSizeForQueue);
		Batches<T> batches = createBatches(items, batchSizeForQueue);
		List<BatchQueue<T>> queues = batches.getBatches();
		long countSize = batches.getCountTotalTasks();
		// List<LinkedList<T>> queues = new ArrayList<>();
		// LinkedList<T> batch = new LinkedList<>();
		// int i = 0;
		// long countSize = 0;
		// Iterator<T> it = items.iterator();
		// while (it.hasNext()) {
		// if (i == batchSizeForQueue) {
		// i = 0;
		// queues.add(batch);
		// batch = new LinkedList<>();
		// }
		//
		// countSize++;
		// batch.add(it.next());
		// i++;
		// }
		// queues.add(batch);
		TaskSet<T> taskSet; 

		
//		if (isParallel) {
//			startInnerThread();
//			
//			// addThreadToPool();
//		}
		
		if (maxConcurrent > 0) {
			// we create only maxConcurrent tasks as we do not want to have more
			// threads
			taskSet = new TaskSet<>(body, countSize, queues.size(), name, true, new AtomicLong());
			LinkedBlockingQueue<BatchQueue<T>> tasksQueues = new LinkedBlockingQueue<>(queues);

			for (int mc = 0; mc < maxConcurrent; mc++) {
				this.jobQueue.putFirst(new MultipleTasksJob<>(taskSet, tasksQueues, jobLevel));
			}

		} else {
			if(done == null) {
				done = new AtomicLong(0);
			}
			taskSet = new TaskSet<>(body, countSize, queues.size(), name, false, done);
//			System.out.println("Queue size for " + name + " is " + queues.size());
			for (BatchQueue<T> q : queues) {
				this.jobQueue.putLast(new TasksJob<>(taskSet, q.getQueue(), q.getSize(), jobLevel));
			}

		}

		if (reportProgress ) {
			long start = System.currentTimeMillis();
 
			ReporterThread reporterThreadLocal = getReporterThread();
			reporterThreadLocal.startProgressReporting(name,start, taskSet.getAtomicLongDone(), countSize);
		}
		
	


		taskSet.await();

		if (isParallel) {
			if(!keepJobLevel) {
				jobLevelLocks[jobLevel].unlock();
			}
			removeInnerThreadFromPool();
			// CountDownPauseJob poisen = new CountDownPauseJob();
			// this.jobQueue.putLast(poisen);
			// poisen.await();
		}

	}

	private void removeInnerThreadFromPool() {
		getInnerThread().removeFromThreadPool();

	}

	private static interface Job {
		public void runAll();

		public String getName();

		public boolean killThread();

		public int getJobLevel();

//		public boolean pauseThread();
	}

	private static class MultipleTasksJob<T> implements Job {
		private final TaskSet<T> taskSet;
		private final LinkedBlockingQueue<BatchQueue<T>> tasksQueues;
		private final int jobLevel;

		public MultipleTasksJob(TaskSet<T> taskSet, LinkedBlockingQueue<BatchQueue<T>> tasksQueues, int jobLevel) {
			super();
			this.taskSet = taskSet;
			this.tasksQueues = tasksQueues;
			this.jobLevel = jobLevel;
		}

		@Override
		public void runAll() {
			BatchQueue<T> taskQueue = null;
			while ((taskQueue = tasksQueues.poll()) != null) {
				this.taskSet.runAndCountDown(taskQueue.getQueue(), taskQueue.size, jobLevel);
			}

		}

		@Override
		public String getName() {
			return this.taskSet.getName();
		}

		@Override
		public boolean killThread() {
			return false;
		}

		@Override
		public int getJobLevel() {
			return jobLevel;
		}

//		@Override
//		public boolean pauseThread() {
//			return false;
//		}

	}

	private static class TasksJob<T> implements Job {
		private final TaskSet<T> taskSet;
		private final ConcurrentLinkedQueue<T> tasks;
		private final int size;
		private final int jobLevel;

		public TasksJob(TaskSet<T> taskSet, ConcurrentLinkedQueue<T> tasks, int size, int jobLevel) {
			super();
			this.taskSet = taskSet;
			this.tasks = tasks;
			this.size = size;
			this.jobLevel = jobLevel;
		}

		public void runAll() {
			taskSet.runAndCountDown(tasks, size, jobLevel);

		}

		public String getName() {
			return this.taskSet.getName();
		}

		@Override
		public boolean killThread() {
			return false;
		}

		@Override
		public int getJobLevel() {
			return jobLevel;
		}

//		@Override
//		public boolean pauseThread() {
//			return false;
//		}

	}

//	private static class CountDownPauseJob implements Job {
//		private final CountDownLatch latch = new CountDownLatch(1);
//
//		@Override
//		public void runAll() {
//
//		}
//
//		@Override
//		public String getName() {
//			return "PAUSED";
//		}
//
//		@Override
//		public boolean killThread() {
//			return false;
//		}
//
//		public void await() throws InterruptedException {
//			latch.await();
//		}
//
//		@Override
//		public boolean pauseThread() {
//			latch.countDown();
//			return true;
//		}
//
//	}

	private static class PoisenJob implements Job {

		@Override
		public void runAll() {

		}

		@Override
		public String getName() {
			return "POISEN";
		}

		@Override
		public boolean killThread() {
			return true;
		}

		@Override
		public int getJobLevel() {
			throw new UnsupportedOperationException();
		}

//		@Override
//		public boolean pauseThread() {
//			return false;
//		}

	}
	
	public static boolean LIMIT_OBSERVER_RUN = false;

	private static class TaskSet<T> {
		private final Consumer<T> consumer;
		private final CountDownLatch latch;
		private final String name;
		private final AtomicLong done;
//		private final AtomicInteger queuesDone = new AtomicInteger(0);
//		private final int queuesTotal;
		private final boolean concurrencyLimited;
		private final long totalSize; 

		public TaskSet(Consumer<T> consumer, long totalSize, int latchSize, String name, boolean concurrencyLimited, AtomicLong done) {
			super();
			this.consumer = consumer;
			this.latch = new CountDownLatch(latchSize);
			this.name = name;
//			this.queuesTotal = latchSize;
			this.concurrencyLimited = concurrencyLimited;
			this.done= done;
			this.totalSize = totalSize;
		}

		public boolean isConcurrencyLimited() {
			return concurrencyLimited;
		}

		public String getName() {
			return name;
		}

		public void runAndCountDown(ConcurrentLinkedQueue<T> tasks, int size, int jobLevel) {

			 

//			AtomicBoolean tasksDone = new AtomicBoolean(false);
//			AtomicBoolean split = new AtomicBoolean(false);
			AtomicReference<CountDownLatch> ref = new AtomicReference<CountDownLatch>(null);
			// AtomicReference<TaskSet<T>> taskSetRef = new
			// AtomicReference<>(null);

			MutableInt currentChange = new MutableInt(0);
			int depth = getDepth();
			if (!LIMIT_OBSERVER_RUN && !isConcurrencyLimited() && depth < maxDepth && size > 1 && jobLevel == 0) {
				pp.getObserverInnerThread().startObserverThread(name, tasks, consumer, ref, size,  done, currentChange, jobLevel);
			}
			
			long lastMeasure = System.currentTimeMillis();
			long finished = 0;
			T task = null;
			
			while ( (task = tasks.poll()) != null) {
				try {
				consumer.accept(task);
				currentChange.increment();
				finished++;
				long curTime = System.currentTimeMillis();
				if( curTime- lastMeasure > 8000) {
					done.addAndGet(finished);
					finished = 0;
					lastMeasure = curTime;
					
				}
				
				}
				catch(Exception e) {
					System.out.println("Error during PP");
					e.printStackTrace();	
					int sdf = 123;
				
				}
//				
//				if(split.get()) {
//					break;
//				}
			}
			
			done.addAndGet(finished);
			
			//splitting
			CountDownLatch aLatch = ref.get();
			if(aLatch!= null) {
				try {
					aLatch.await();
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} 
			}

			
//			queuesDone.incrementAndGet();
			latch.countDown();
		}

		public void await() throws InterruptedException {
			latch.await();
		}

		public long getDone() {
			return done.get();
		}
		
		public AtomicLong getAtomicLongDone() {
			return this.done;
		}

//		public boolean isDone() {
//			return queuesDone.get() == queuesTotal;
//		}

	}

	ThreadLocal<ReporterThread> reporterThread = new ThreadLocal<>();
	ConcurrentLinkedQueue<ReporterThread> reporterThreadList = new ConcurrentLinkedQueue<>();

	@SuppressWarnings("rawtypes")
	private static class ReporterThreadData {
		final long start;
		final AtomicLong done;
		final long totalTasks;
		final String taskName;
		public ReporterThreadData(long start, AtomicLong done, long totalTasks, String taskName) {
			super();
			this.start = start;
			this.done = done;
			this.totalTasks = totalTasks;
			this.taskName = taskName;
		}
		
	}

	
	@SuppressWarnings("rawtypes")
	private class ReporterThread implements Runnable {
		
	
		AtomicReference<CountDownLatch> latch = new AtomicReference<>(new CountDownLatch(1));
//		AtomicReference<CountDownLatch> newTaskLatch = new AtomicReference<>(new CountDownLatch(1));
//		AtomicReference<Runnable> runnable = new AtomicReference<Runnable>(null);
//		int threadNumber;
//		private final String name;
		volatile boolean stop = false;

//		volatile long start;
//		volatile TaskSet taskSet;
//		volatile long totalTasks;
//		volatile String taskName;
//
//		volatile long newStart;
//		volatile TaskSet newTaskSet;
//		volatile long newTotalTasks;
//		volatile String newTaskName;
		
		private final  AtomicReference<ReporterThreadData> data = new AtomicReference<>();
//		private final  AtomicReference<ReporterThreadData> newData = new AtomicReference<>();
		
		public ReporterThread() {
//			this.name = "REPORTER_THREAD_" + Thread.currentThread().getName();
		}

		@Override
		public void run() {
//			newTaskLatch.get().countDown();
			while (running) {
				// Thread.currentThread().setName("REPORTER_THREAD_" +
				// threadNumber);

				try {
					latch.get().await();
					latch.set(new CountDownLatch(1));
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				
				
				if (!stop) {
					ReporterThreadData currentData = data.getAndSet(null);
					
					if(data != null)
						runProgressReporting(currentData);
					
				} else {
					data.set(null);
					break;
				}

			

			}
		}
		
		public <E> void startProgressReporting(String taskName, long start, AtomicLong done, long totalTasks ) {
			
			this.data.set( new ReporterThreadData(start, done, totalTasks, taskName));
		
//			try {
//				newTaskLatch.get().await();
//			} catch (InterruptedException e) {
//				// TODO Auto-generated catch block
//				e.printStackTrace();
//			}
//			newTaskLatch.set(new CountDownLatch(1));
//			this.newTaskName = taskName;
//			this.newStart = start;
//			this.newTaskSet = taskSet;
//			this.newTotalTasks = totalTasks;
			latch.get().countDown();
		}
		
		private void runProgressReporting(ReporterThreadData data) {
			try {
				long lastTime = data.start;
				long last = 0;
				boolean stop = false;
				while (!stop) {

					CountDownLatch innerLatch = new CountDownLatch(1);
					pp.scheduledExecutor.schedule(() -> {
						innerLatch.countDown();
					}, 12, TimeUnit.SECONDS);
					
					try {
						innerLatch.await();
					} catch (InterruptedException e1) {
						// TODO Auto-generated catch block
						e1.printStackTrace();
					}
					
				 
					long done = data.done.get();
					boolean cont = printProgress(data.taskName, data.totalTasks, done, last, data.start, lastTime);
					last = done;
					lastTime = System.currentTimeMillis();
					if (!cont) {
						stop = true;
					}

				}
				
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
				int sdf = 123;
			}
		}

//		public void setRunnable(Runnable runnable) {
//			this.runnable.set(runnable);
//			this.latch.get().countDown();
//		}

		public void kill() {
			this.stop = true;
			this.latch.get().countDown();

		}

	}
	
//	private final List<ReporterThread> mainReporterThreads = new ArrayList<>();

	private ReporterThread getReporterThread() {
		ReporterThread inenrReporterThread = this.reporterThread.get();
		if (inenrReporterThread == null) {

			inenrReporterThread = new ReporterThread();
			Thread reporterThread = new Thread(inenrReporterThread);
			this.reporterThread.set(inenrReporterThread);
			reporterThreadList.add(inenrReporterThread);
			reporterThread.start();
		}
//		if(!AlreadyParalell.isCurrentThreadParallel()) {
//			mainReporterThreads.add(inenrReporterThread);
//		}
		return inenrReporterThread;
	}

	private InnerThread getInnerThread() {
		InnerThread innerThreadRunnable = innerThreads.get();
		if (innerThreadRunnable == null) {
			int depth = getDepth();
			if(PersistentParallel6.getThreadName().length() > 8) {
				int sdf= 123;
			}
			innerThreadRunnable = new InnerThread(PersistentParallel6.getThreadName()+"_"+depth, depth + 1);
			Thread innerThread = new Thread(innerThreadRunnable);
			innerThreads.set(innerThreadRunnable);
			innerThreadsList.add(innerThreadRunnable);
			innerThread.start();
		}
		return innerThreadRunnable;
	}
	
	private ObserverThread getObserverInnerThread() {
		ObserverThread innerThreadRunnable = observerInnerThreads.get();
		if (innerThreadRunnable == null) {
			int depth = getDepth();
			innerThreadRunnable = new ObserverThread(PersistentParallel6.getThreadName()+"_"+depth, depth + 1);
			Thread innerThread = new Thread(innerThreadRunnable);
			observerInnerThreads.set(innerThreadRunnable);
			observerThreadsList.add(innerThreadRunnable);
			innerThread.start();
		}
		return innerThreadRunnable;
	}

	ConcurrentLinkedQueue<InnerThread> innerThreadsList = new ConcurrentLinkedQueue<>();
	ThreadLocal<InnerThread> innerThreads = new ThreadLocal<>();
	ThreadLocal<ObserverThread> observerInnerThreads = new ThreadLocal<>();
	ConcurrentLinkedQueue<ObserverThread> observerThreadsList = new ConcurrentLinkedQueue<>();

	AtomicInteger threadNumbers = new AtomicInteger(0);


//	@SuppressWarnings("rawtypes")
	private static class InnerThread implements Runnable {

//		private static enum InnerRun {
//			THREAD_POOL,OBSERVER;
//		}
		
		AtomicReference<CountDownLatch> latch = new AtomicReference<>(new CountDownLatch(1));
//		AtomicReference<Runnable> runnable = new AtomicReference<Runnable>(null);
//		AtomicReference<InnerRun> innerRun = new AtomicReference<>(null);
		AtomicReference<Boolean> kill = new AtomicReference<>(false);
		String threadNumber;
		AtomicReference<CountDownLatch> innerStopper = new AtomicReference<>(null);
		int depth;
		
		
		
 

		public InnerThread(String threadNumber, int depth) {
			this.threadNumber = threadNumber;
			this.depth = depth;
		}

		public void addToPool() {
			innerStopper.set(null);
			this.latch.get().countDown();
			
		}
		
	 

		public void removeFromThreadPool() {
			CountDownLatch innerLatch = new CountDownLatch(1);
			innerStopper.set(innerLatch);
			this.latch.get().countDown();
//			try {
//				innerLatch.await();
//			} catch (InterruptedException e) {
//				// TODO Auto-generated catch block
//				e.printStackTrace();
//			}
			
			return;

		}

		@Override
		public void run() {
			setDepth(depth);
			while (true) {

				PersistentParallel6.setThreadName("" + threadNumber);
				Thread.currentThread().setName("INNER_" + threadNumber);

				try {
					this.latch.get().await();
					
					latch.set(new CountDownLatch(1));
					
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
//				
				CountDownLatch  latchA = innerStopper.get() ;
				if( latchA != null) {
					latchA.countDown();
					
					if(kill.get()) {
						 
						pp.killAllSubThreads();
						
						break; 
						 
					}
					
					
				}
				else {
				if(!kill.get()) {
				 
					runFromThreadPool(); 
						
					 
				}
				else {
					pp.killAllSubThreads();
					break; 
				}
			
				 
				}
				
			}
		}

 
		
		private void runFromThreadPool() {
			try {
				CountDownLatch latchA = null;
				while ((latchA = this.innerStopper.get()) == null) {

					Job job;

					job = pp.jobQueue.pollLast(10, TimeUnit.SECONDS);

					if (job != null) {
						if (!job.killThread()) {
							JobLevel.setJobLevel(job.getJobLevel());
						}
						job.runAll();
						if (job.killThread()) {
							pp.jobQueue.putFirst(job);
							return;
						}
						// else if (job.pauseThread()) {
						// throw new IllegalStateException();
						// }

					} else {
					}

				}

			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			// if(latchA != null) {
			// this.innerStopper.set(null);
			// latchA.countDown();
			// }

		}

//		public void setRunnable(Runnable runnable) {
//			this.runnable.set(runnable);
//			this.latch.get().countDown();
//		}

		public void kill() { 
//			if(this.kill.get()) {
//				int dfsdf = 123;
//			}
			

			this.kill.set(true);
			
			CountDownLatch innerLatch = new CountDownLatch(1);
			innerStopper.set(innerLatch);

			this.latch.get().countDown();
			try {
				innerLatch.await();
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			 
			

		}

	}
	
	@SuppressWarnings("rawtypes")
	private static class ObserverData {
		private final String name;
		private final ConcurrentLinkedQueue  tasks;
		private final AtomicReference<CountDownLatch> ref;
		private final AtomicLong done;
		private final Consumer  consumer;
		
		private final MutableInt currentChange;
		private final int tasksSize;
		private final int percent;
		private final int jobLevel;
		public ObserverData(String name, ConcurrentLinkedQueue tasks, AtomicReference<CountDownLatch> ref, AtomicLong done, Consumer consumer, MutableInt currentChange,
				int tasksSize, int percent, int jobLevel) {
			super();
			this.name = name;
			this.tasks = tasks;
			this.ref = ref;
			this.done = done;
			this.consumer = consumer;
			this.currentChange = currentChange;
			this.tasksSize = tasksSize;
			this.percent = percent;
			this.jobLevel = jobLevel;
		}
		
		
	}
	
	@SuppressWarnings("rawtypes")
	private static class ObserverThread implements Runnable {

		 
		
		AtomicReference<CountDownLatch> latch = new AtomicReference<>(new CountDownLatch(1));
//		AtomicReference<Runnable> runnable = new AtomicReference<Runnable>(null);
		AtomicReference<Boolean> kill = new AtomicReference<>(false);
		String threadNumber;
//		AtomicReference<CountDownLatch> innerStopper = new AtomicReference<>(null);
		
		private int depth;
		

		private volatile  boolean stop = false;
		
		public ObserverThread(String threadNumber, int depth) {
			this.threadNumber = threadNumber;
			this.depth = depth;
		}

//		public void addToPool() {
//			innerRun.set(InnerRun.THREAD_POOL);
//			latch.get().countDown();
//			
//		}
		
		private final  LinkedBlockingDeque<ObserverData> data = new LinkedBlockingDeque<>();
		
		public <E> void startObserverThread(String name, ConcurrentLinkedQueue<E> tasks, Consumer<E> consumer, AtomicReference<CountDownLatch> ref, int tasksSize, AtomicLong done, MutableInt currentChange, int jobLevel) {
			if(!kill.get() && !stop) {
				data.clear();
				data.addFirst(new ObserverData(name, tasks, ref, done, consumer, currentChange, tasksSize, (int) (0.35 * tasksSize), jobLevel));
			}
			
//			this.tasks = tasks;
//			this.consumer = consumer;
//			this.ref = ref; 
//			this.done = done;
//			this.currentChange = currentChange;
//			this.tasksSize = tasksSize; 
//			this.percent = ; 
			latch.get().countDown();
			
		}

	 

		@Override
		public void run() {
			if(depth > PersistentParallel6.maxDepth){
				throw new IllegalStateException();
			}
			setDepth(depth);
			
			PersistentParallel6.setThreadName("" + threadNumber);
			Thread.currentThread().setName("OBSERVER_" + threadNumber);
			
			try {
				latch.get().await();
//				latch.set(new CountDownLatch(1));
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			while (pp.running) {

				

			
				
				
				if(!kill.get()) {
					
					try {
						ObserverData  observerData = data.pollFirst(15, TimeUnit.SECONDS);
						
						if(observerData != null) { 
							runObserverThread(observerData);
						}
						
					} catch (InterruptedException e1) {
						// TODO Auto-generated catch block
						e1.printStackTrace();
					}
					
					
 
					
				}
				else {
					pp.killAllSubThreads();
					
					break; 
				}
			
				
			

			}
		}

		@SuppressWarnings("unchecked")
		private void runObserverThread(ObserverData observerData) {
			
		 
				
				int previous = observerData.currentChange.intValue();
				
				CountDownLatch latch = new CountDownLatch(1);
				pp.scheduledExecutor.schedule(() -> {
					latch.countDown();
				}, 15, TimeUnit.SECONDS);
				
				try {
					latch.await();
				} catch (InterruptedException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
//				try {
//					Thread.sleep(15000);
//				} catch (InterruptedException e) {
//					// TODO Auto-generated catch block
//					e.printStackTrace();
//				}
//				if(previous == observerData.tasksSize) {
//					return;
//				}
				
				int cur = observerData.currentChange.intValue();
				
				if((cur - previous)   < observerData.percent && !observerData.tasks.isEmpty()) {
					System.err.println("Now splitting for sub parallel");
					CountDownLatch aLatch = new CountDownLatch(1);
					observerData.ref.set(aLatch);
	
	 
						 
						
		 
					List  list = new ArrayList<>();
					Object ttask = null;
					while ((ttask = observerData.tasks.poll()) != null) {
						list.add(ttask);
					}

//						pp.startInnerThread();
					
					JobLevel.setJobLevel(observerData.jobLevel);
					
//					PersistentParallel6.runAndFinishNoProgress(list, observerData.consumer, observerData.name, observerData.done);

 
					try {
						pp.foreach(list, list.size(), observerData.consumer, observerData.name, -1, false, -1, observerData.done, true);
					} catch (Exception e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
					
					
//						pp.removeInnerThreadFromPool(); 
		 
						
							
							
			 
					
					aLatch.countDown();  
				 
			}
			
		
			
			
		}
		
 

//		public void setRunnable(Runnable runnable) {
//			this.runnable.set(runnable);
//			this.latch.get().countDown();
//		}

		public void kill() {
			this.data.clear();
			this.stop = true;
			this.kill.set(true);
			this.latch.get().countDown(); 
		}

	}
	
 
//	private static class LinkedListIterable<E> implements Iterable<E> {
//		private final LinkedList<E> list;
//		
//		public LinkedListIterable(LinkedList<E> list) {
//			super();
//			this.list = list;
//		}
//
//		@Override
//		public Iterator<E> iterator() {
//			return new LinkedListIterator();
//		}
//		
//		private   class LinkedListIterator implements Iterator<E> {
//
//			E next = null;
//			
//			@Override
//			public boolean hasNext() {
//				if(next != null) {
//					next = list.poll();
//				}
//				return next != null;
//			}
//
//			@Override
//			public E next() {
//				if(next == null) {
//					return list.poll();
//				}
//				else {
//					return next;
//				} 
//			}
//			
//		}
//	}

	private class ParallelThread implements Runnable {

		@Override
		public void run() {
			AlreadyParalell.setCurrentThreadParallel(true);
			int threadNumber = threadNumbers.getAndIncrement();

//			InnerThread innerThreadRunnable = getInnerThread();

			while (running) {
				try {
					Job job = jobQueue.takeLast();
					PersistentParallel6.setThreadName("" + threadNumber);
					Thread.currentThread().setName(job.getName() + "_" + threadNumber);
					if (job != null) {
						if (!job.killThread()) {
							JobLevel.setJobLevel(job.getJobLevel());
						}
						job.runAll();
						if (job.killThread()) {
							pp.killAllSubThreads();
						 
							return;
						}
//						else if (job.pauseThread()) {
//							CountDownLatch latch = new CountDownLatch(1);
//							MutableBoolean kill = new MutableBoolean(false);
//							threadPool.put(new PoolEntry(latch, kill));
//							latch.await();
//							if (kill.isTrue()) {
//								innerThreadRunnable.kill();
//								ReporterThread reporterThreadRunnable = pp.reporterThread.get();
//								if (reporterThreadRunnable != null)
//									reporterThreadRunnable.kill();
//								return;
//							}
//						}

					}
				} catch (InterruptedException e) {
					throw new RuntimeException(e);
				}

			}
		}
	}

	public static class AlreadyParalell {
		public static InheritableThreadLocal<Boolean> alreadyParallel = new InheritableThreadLocal<>();

		public static void setCurrentThreadParallel(boolean parallel) {
			alreadyParallel.set(parallel);
		}

		public static boolean isCurrentThreadParallel() {
			Boolean bool = alreadyParallel.get();
			if (Boolean.TRUE.equals(bool))
				return true;
			else
				return false;

		}
	}
	
	public static class JobLevel {
		public static ThreadLocal<Integer> jobLevel = new ThreadLocal<>();

		public static void setJobLevel(int jobLevel) {
			JobLevel.jobLevel.set(jobLevel);
		}

		public static int getJobLevel() {
			Integer jobLevel = JobLevel.jobLevel.get();
			if (jobLevel == null)
				throw new IllegalStateException();
			else
				return jobLevel;

		}
	}

	public static void close() throws InterruptedException {
		for (int i = 0; i < numOfThreads * numOfThreadsMultiplier * maxDepth; i++) {
			PersistentParallel6.pp.jobQueue.putFirst(new PoisenJob());
		}

//		for (PoolEntry pair : pp.threadPool) {
//			pair.getKill().setValue(true);
//			pair.getLock().countDown();
//		}
		
		pp.killAllSubThreads();
		
		pp.killAllInnerThreads();
		
		pp.killAllReporterThreads();
		pp.killAllObserverThreads();
		
		pp.scheduledExecutor.shutdown();
		
		
//		for(ReporterThread reporterThread : pp.mainReporterThreads) {
//			reporterThread.kill();
//		}
		
		PersistentParallel4.closeInner();
	}


	public void killAllSubThreads() {
//		InnerThread innerThreadRunnable = pp.innerThreads.get();
//		if (innerThreadRunnable != null)
//			innerThreadRunnable.kill();
		
//		ObserverThread innerObserverThreadRunnable = pp.observerInnerThreads.get();
//		if (innerObserverThreadRunnable != null)
//			innerObserverThreadRunnable.kill();
//		
//		ReporterThread reporterThreadRunnable = pp.reporterThread.get();
//		if (reporterThreadRunnable != null)
//			reporterThreadRunnable.kill();
		
	}
	
	public void killAllInnerThreads() { 
		for(InnerThread innerThread : innerThreadsList) {
			innerThread.kill();
		}
	}
	
	public void killAllReporterThreads() { 
		for(ReporterThread innerThread : reporterThreadList) {
			innerThread.kill();
		}
	}
	
	
	public void killAllObserverThreads() { 
		for(ObserverThread innerThread : observerThreadsList) {
			innerThread.kill();
		}
	}

	public static <E> void runAndFinishWithBatchSize(Collection<E> items, final Consumer<E> body, String name, int batchSize) {
		try {
			pp.foreach(items, items.size(), body, name, batchSize, true, -1);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}
	public static <E> void runAndFinish(Collection<E> items, final Consumer<E> body, String name) {
		try {
			pp.foreach(items, items.size(), body, name, -1, true, -1);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public class ParallelImpl implements PersistentParallelInterface {

	}

	public static void setThreadPaused() {
//		pp.addThreadToMainPool();
		if(AlreadyParalell.isCurrentThreadParallel() && getDepth() < maxDepth) {
			pp.startInnerThread();
		}
	}

	public static void setThreadUnpaused() throws InterruptedException {
		if(AlreadyParalell.isCurrentThreadParallel()&& getDepth() < maxDepth) {
			pp.removeInnerThreadFromPool();
		}
//		CountDownPauseJob poisen = new CountDownPauseJob();
//		pp.jobQueue.putLast(poisen);
		// poisen.await();
	}
	
	private static ThreadLocal<String> threadNames = new ThreadLocal<>();
	public static void setThreadName(String threadName) {
		threadNames.set(threadName);
	}
	public static String getThreadName() {
		String threadName = threadNames.get();
		if(threadName == null) {
			return Thread.currentThread().getName();
		}
		else {
			return threadName;
		}
	}

	public static <E> void runAndFinish(Collection<E> items, final Consumer<E> body, String name, int maxConcurrent) {

		try {
			pp.foreach(items, items.size(), body, name, -1, true, maxConcurrent);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public static <E> void runAndFinish(Collection<E> items, final Consumer<E> body, final Class<?> clazz, int maxConcurrent) {

		try {
			pp.foreach(items, items.size(), body, clazz.getName(), -1, true, maxConcurrent);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public static <E> void runAndFinish(Collection<E> items, final Consumer<E> body, final Class<?> clazz, String name) {
		try {
			pp.foreach(items, items.size(), body, clazz.getName() + " - " + name, -1, true, -1);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}
	
	public static <E> void runAndFinish(Collection<E> items, final Consumer<E> body, final Class<?> clazz) {
		try {
			pp.foreach(items, items.size(), body, clazz.getName(), -1, true, -1);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	
	public static <E> void runAndFinishNoProgress(Collection<E> items, final Consumer<E> body, final Class<?> clazz) {
		runAndFinishNoProgress(items, body, clazz, null);
		
	}

	
	
	public static <E> void runAndFinishNoProgress(Collection<E> items, final Consumer<E> body, final Class<?> clazz, AtomicLong done) {
		try {
			pp.foreach(items, items.size(), body, clazz.getName(), -1, false, -1, done);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}
	
	
	public static <E> void runAndFinishNoProgress(Collection<E> items, final Consumer<E> body, final String name, AtomicLong done) {
		try {
			pp.foreach(items, items.size(), body, name, -1, false, -1, done);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public static <E> void runAndFinish(Collection<E> items, final Consumer<E> body, String name, boolean sequential) {

		if (sequential) {
			sequentialFor(items, body);
		} else {
			try {
				pp.foreach(items, items.size(), body, name, -1, true, -1);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

	}

	public static <E> void runAndFinish(Collection<E> items, final Consumer<E> body, final Class<?> clazz, boolean sequential) {

		if (sequential) {
			sequentialFor(items, body);
		} else {
			try {
				pp.foreach(items, items.size(), body, clazz.getName(), -1, true, -1);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

	}

	
	public static <E> void runAndFinishWithSize(Iterable<E> items, final Consumer<E> body, long size, final Class<?> clazz) {

		 
			try { 
				pp.foreach(items, size, body, clazz.getName(), -1, true, -1);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
 

	}
	
	
	public static <E> void runAndFinishWithSize(Iterable<E> items, final Consumer<E> body, long size, final String name, boolean sequential) {

		if (sequential) {
			sequentialFor(items, body);
		} else {
			try {
				pp.foreach(items, size, body, name, -1, true, -1);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

}
	
	public static <E> void runAndFinishWithSize(Iterable<E> items, final Consumer<E> body, long size, final Class<?> clazz, boolean sequential) {

		if (sequential) {
			sequentialFor(items, body);
		} else {
			try {
				pp.foreach(items, size, body, clazz.getName(), -1, true, -1);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

}
	
	public static <E> void runAndFinish(Iterable<E> items, final Consumer<E> body, final Class<?> clazz, int batchSize, boolean sequential) {

		if (sequential) {
			sequentialFor(items, body);
		} else {
			try {
				pp.foreach(items, -1, body, clazz.getName(), batchSize, true, -1);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

	}

	
	public static <E> void runAndFinish(Iterable<E> items, final Consumer<E> body, final String name, int batchSize, boolean sequential) {

		if (sequential) {
			sequentialFor(items, body);
		} else {
			try {
				pp.foreach(items, -1, body, name, batchSize, true, -1);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

	}
	
	
	public static <E> void runAndFinish(Collection<E> items, final Consumer<E> body, final Class<?> clazz, final String name,  boolean sequential) {

		if (sequential) {
			sequentialFor(items, body);
		} else {
			try {
				pp.foreach(items, items.size(), body, clazz.getName() + " - " + name, -1, true, -1);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

	}
	
	public static <E> void runAndFinish(Iterable<E> items, final Consumer<E> body, final Class<?> clazz, int batchSize) {

		try {
			pp.foreach(items, -1, body, clazz.getName(), batchSize, true, -1);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public static <E> void runAndFinishWithDefaultBatchSize(Iterable<E> items, final Consumer<E> body, final Class<?> clazz, boolean sequential) {

		try {
			pp.foreach(items, -1, body, clazz.getName(), defaultBatchSize, true, -1);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public static <E> void runAndFinishWithDefaultBatchSize(Iterable<E> items, final Consumer<E> body, final Class<?> clazz) {

		try {
			pp.foreach(items, -1, body, clazz.getName(), defaultBatchSize, true, -1);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}
	
	private static <T> void sequentialFor(Iterable<T> items, final Consumer<T> body ) {
 
		for (T item : items) {
			body.accept(item); 
		} 

	}

	private static <T> void sequentialFor(Iterable<T> items, final Consumer<T> body, AtomicLong done) {
		int size = 0;
		for (T item : items) {
			body.accept(item);
			size++;
		}
		if(done != null)
			done.addAndGet(size);

	}

//	private static class PoolEntry {
//		CountDownLatch lock;
//		MutableBoolean kill;
//
//		private PoolEntry(CountDownLatch lock, MutableBoolean kill) {
//			super();
//			this.lock = lock;
//			this.kill = kill;
//		}
//
//		public CountDownLatch getLock() {
//			return lock;
//		}
//
//		public void setLock(CountDownLatch lock) {
//			this.lock = lock;
//		}
//
//		public MutableBoolean getKill() {
//			return kill;
//		}
//
//		public void setKill(MutableBoolean kill) {
//			this.kill = kill;
//		}
//
//	}
	private static class BatchQueue<T> {
		private final  ConcurrentLinkedQueue<T>  queue;
		private final int size;
		public BatchQueue(ConcurrentLinkedQueue<T> queue, int size) {
			super();
			this.queue = queue;
			this.size = size;
		}
		public ConcurrentLinkedQueue<T> getQueue() {
			return queue;
		}
		public int getSize() {
			return size;
		}
		
		
		
		
	}
	private static class Batches<T> {
		private final List<BatchQueue<T>> batches;
		private final long countTotalTasks;

		public Batches(List<BatchQueue<T>> batches, long countTotalTasks) {
			super();
			this.batches = batches;
			this.countTotalTasks = countTotalTasks;
		}

		public List<BatchQueue<T>> getBatches() {
			return batches;
		}

		public long getCountTotalTasks() {
			return countTotalTasks;
		}

	}
	
	
	private static ThreadLocal<Integer> depthThreadLocal = new ThreadLocal<Integer>();
	private static int getDepth() {
		
		Integer d = depthThreadLocal.get();
		if(d == null) {
			return 0;
		}
		else {
			return d;
		}
		
		
	}
	
	private static void setDepth(int depth) {
		depthThreadLocal.set(depth);
	}

	public static int getNumOfThreads() {
		return numOfThreads;
	}

	public static void setNumOfThreads(int numOfThreads) {
		PersistentParallel6.numOfThreads = numOfThreads;
	}

	public static int getNumOfThreadsMultiplier() {
		return numOfThreadsMultiplier;
	}

	public static void setNumOfThreadsMultiplier(int numOfThreadsMultiplier) {
		PersistentParallel6.numOfThreadsMultiplier = numOfThreadsMultiplier;
	}
	
	public static void restart() {
		PersistentParallel6.pp = new PersistentParallel6();
	}
	
	
}
