package de.mannheim.uni.expansion.utils;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Consumer;

import org.apache.commons.lang.mutable.MutableBoolean;
import org.apache.commons.lang3.time.DurationFormatUtils;

import de.mannheim.uni.expansion.utils.PersistentParallelFactory.PersistentParallelInterface;

public class PersistentParallel6old {

	private static int numOfThreads = Runtime.getRuntime().availableProcessors();
	private static int numOfThreadsMultiplier = 4;

	private static PersistentParallel6old pp = new PersistentParallel6old();

	private final boolean running = true;

	private final LinkedBlockingDeque<Job> jobQueue = new LinkedBlockingDeque<>();
	private final LinkedBlockingQueue<PoolEntry> threadPool = new LinkedBlockingQueue<>();
	 

	private PersistentParallel6old() {
		for (int i = 0; i < numOfThreads * numOfThreadsMultiplier; i++) {
			addThreadToPool();
		}
	}

	private <T> void addThreadToPool(TaskSet<T> taskSet) {
		InnerThread innerThread = this.getInnerThread();
		innerThread.setRunnable(innerThread.getInnerRunnable(taskSet)); 

	}
	
	private void addThreadToPool() {
		PoolEntry pair = threadPool.poll();
		if (pair != null) {
			CountDownLatch latch = pair.getLock();
			latch.countDown();
		} else {
			Thread t = new Thread(new ParallelThread());
			t.start();
		}

	}

	private static int divideButRoundUp(int num, int divisor) {
		return (num + divisor - 1) / divisor;
	}
	
	private boolean printProgress(String name, long tasks, long done, long last, long start, long lastTime) {
		 
 if(done == tasks)
	 return false;


		long soFar = System.currentTimeMillis() - start;
		long pauseTime = System.currentTimeMillis() - lastTime;
		//long perItem = (long) (((float) soFar) / done);
		long left = (long) (((float) soFar / done) * (tasks - done));
		float itemsPerSecAvg = (float)done / (float)(soFar / 1000.0f);
		float itemsPerSecNow = (float)(done - last) / (pauseTime / 1000.0f);
		
		if((((float) soFar) / done)==Float.POSITIVE_INFINITY)
		{
			//perItem = -1;
			left = -1; 
		}
		String ttl = DurationFormatUtils.formatDuration(soFar, "HH:mm:ss.S");
		//String each = DurationFormatUtils.formatDuration(perItem, "HH:mm:ss.S");
		String remaining = DurationFormatUtils.formatDuration(left, "HH:mm:ss.S");
		
		System.err.println(String.format("%s: %,d of %,d tasks completed after %s (%d/%d active threads). Avg: %.2f items/s, Current: %.2f items/s, %s left.",
				name, done, tasks, ttl, numOfThreads * 4 - threadPool.size(), numOfThreads * 4, itemsPerSecAvg, itemsPerSecNow, remaining));

	
		
		
		last = done;
		lastTime = System.currentTimeMillis();

		if (done == tasks) {
			return false;
		} else {
			return true;
		}
//		368 of 1.370 tasks completed after 00:15:19.856 (8/8 active threads). Avg: 0,40 items/s, Current: 0,80 items/s, 00:41:44.607 left.
		
	}

	private static <T> Batches<T>   createBatches(Iterable<T> items, int batchSizeForQueue) {
		List<LinkedList<T>> queues = new ArrayList<>();
		LinkedList<T> batch = new LinkedList<>();
		int i = 0;
		long countSize = 0;
		Iterator<T> it = items.iterator();
		while (it.hasNext()) {
			if (i == batchSizeForQueue) {
				i = 0;
				queues.add(batch);
				batch = new LinkedList<>();
			}

			countSize++;
			batch.add(it.next());
			i++;
		}
		queues.add(batch); 
		return new Batches<>(queues, countSize);
	}
	public <T extends Object> void foreach(Iterable<T> items, int size, final Consumer<T> body, String name, int batchSize, boolean reportProgress, int maxConcurrent)
			throws Exception {

		if (size == 0)
			return;
		//
		// if (reportProgress) {
		// this.totalTasks = 0;
		// this.start = System.currentTimeMillis();
		// this.threadCounter = new ThreadCounter();
		//
		// Runnable progressReporter = () -> {
		// try {
		// boolean stop = false;
		// while (!stop) {
		//
		// Thread.sleep(10000);
		//
		// boolean cont = printProgress();
		// if (!cont) {
		// stop = true;
		// }
		//
		// }
		// } catch (Exception e) {
		// // TODO Auto-generated catch block
		// e.printStackTrace();
		// }
		// };
		//
		// Thread progressReporterThread = new Thread(progressReporter);
		// progressReporterThread.start();
		// }

		int batchSizeForQueue;
		if (size > 0 && batchSize == -1) {
			if (maxConcurrent > 0 && maxConcurrent < numOfThreads) {
				batchSizeForQueue = divideButRoundUp(size, maxConcurrent);
			} else {
				batchSizeForQueue = divideButRoundUp(size, numOfThreads);
			}

			if(batchSizeForQueue > 1000) {
				batchSizeForQueue = divideButRoundUp(batchSizeForQueue, 10);
			}
		} else {
			batchSizeForQueue = batchSize;
		}

		Batches<T> batches = createBatches(items, batchSizeForQueue);
		List<LinkedList<T>> queues = batches.getBatches();
		long countSize = batches.getCountTotalTasks();
//		List<LinkedList<T>> queues = new ArrayList<>();
//		LinkedList<T> batch = new LinkedList<>();
//		int i = 0;
//		long countSize = 0;
//		Iterator<T> it = items.iterator();
//		while (it.hasNext()) {
//			if (i == batchSizeForQueue) {
//				i = 0;
//				queues.add(batch);
//				batch = new LinkedList<>();
//			}
//
//			countSize++;
//			batch.add(it.next());
//			i++;
//		}
//		queues.add(batch); 
		TaskSet<T> taskSet;
		if (maxConcurrent > 0) {
			 taskSet = new TaskSet<>(body, queues.size(), name, true);
		}
		else {
			 taskSet = new TaskSet<>(body, queues.size(), name, false);
		}
		
		boolean isParallel = AlreadyParalell.isCurrentThreadParallel();
		if (isParallel) {
			addThreadToPool( taskSet);
//			addThreadToPool();
		}

		if (maxConcurrent > 0) {
			// we create only maxConcurrent tasks as we do not want to have more threads
 
			LinkedBlockingQueue<LinkedList<T>> tasksQueues = new LinkedBlockingQueue<>(queues);

			for (int mc = 0; mc < maxConcurrent; mc++) {
				this.jobQueue.putLast(new MultipleTasksJob<>(taskSet, tasksQueues));
			}

		} else {

			for (LinkedList<T> q : queues) {
				this.jobQueue.putLast(new TasksJob<>(taskSet, q));
			}

		}

		 if (reportProgress) { 
				 long start = System.currentTimeMillis(); 
				 final long countSizeFinal = countSize;
				
				 Runnable progressReporter = () -> {
				 try {
					 long lastTime = start;
					 long last = 0;
				 boolean stop = false;
				 while (!stop) {
				
				 Thread.sleep(20000);
				 long done = taskSet.getDone();
				 boolean cont = printProgress(name, countSizeFinal, done, last, start, lastTime);
				 last = done;
				 lastTime = System.currentTimeMillis();
				 if (!cont) {
				 stop = true;
				 }
				
				 }
				 } catch (Exception e) {
				 // TODO Auto-generated catch block
				 e.printStackTrace();
				 }
				 };
				 
				 ReporterThread reporterThreadLocal = getReporterThread();
				 reporterThreadLocal.setRunnable(progressReporter);
		}
		
		taskSet.await();

		if (isParallel) {
			removeInnerThreadFromPool();
//			CountDownPauseJob poisen = new CountDownPauseJob();
//			this.jobQueue.putLast(poisen);
//			poisen.await();
		}

	}

	private void removeInnerThreadFromPool() {
		getInnerThread().pauseAwait();
		
	}

	private static interface Job {
		public void runAll();

		public String getName();

		public boolean killThread();

		public boolean pauseThread();
	}

	private static class MultipleTasksJob<T> implements Job {
		private final TaskSet<T> taskSet;
		private final LinkedBlockingQueue<LinkedList<T>> tasksQueues;

		public MultipleTasksJob(TaskSet<T> taskSet, LinkedBlockingQueue<LinkedList<T>> tasksQueues) {
			super();
			this.taskSet = taskSet;
			this.tasksQueues = tasksQueues;
		}

		@Override
		public void runAll() {
			LinkedList<T> taskQueue = null;
			while ((taskQueue = tasksQueues.poll()) != null) {
				this.taskSet.runAndCountDown(taskQueue);
			}

		}

		@Override
		public String getName() {
			return this.taskSet.getName();
		}

		@Override
		public boolean killThread() {
			return false;
		}

		@Override
		public boolean pauseThread() {
			return false;
		}

	}

	private static class TasksJob<T> implements Job {
		private final TaskSet<T> taskSet;
		private final LinkedList<T> tasks;

		public TasksJob(TaskSet<T> taskSet, LinkedList<T> tasks) {
			super();
			this.taskSet = taskSet;
			this.tasks = tasks;
		}

		public void runAll() {
			taskSet.runAndCountDown(tasks);

		}

		public String getName() {
			return this.taskSet.getName();
		}

		@Override
		public boolean killThread() {
			return false;
		}

		@Override
		public boolean pauseThread() {
			return false;
		}

	}

	private static class CountDownPauseJob implements Job {
		private final CountDownLatch latch = new CountDownLatch(1);

		@Override
		public void runAll() {

		}

		@Override
		public String getName() {
			return "PAUSED";
		}

		@Override
		public boolean killThread() {
			return false;
		}

		public void await() throws InterruptedException {
			latch.await();
		}

		@Override
		public boolean pauseThread() {
			latch.countDown();
			return true;
		}

	}

	private static class PoisenJob implements Job {

		@Override
		public void runAll() {

		}

		@Override
		public String getName() {
			return "POISEN";
		}

		@Override
		public boolean killThread() {
			return true;
		}

		@Override
		public boolean pauseThread() {
			return false;
		}

	}

 
	private static class TaskSet<T> {
		private final Consumer<T> consumer;
		private final CountDownLatch latch;
		private final String name;
		private  final AtomicLong done = new AtomicLong(0);
		private final AtomicInteger queuesDone = new AtomicInteger(0);
		private final int queuesTotal;
		private final boolean concurrencyLimited;

		public TaskSet(Consumer<T> consumer, int latchSize, String name, boolean concurrencyLimited) {
			super();
			this.consumer = consumer;
			this.latch = new CountDownLatch(latchSize);
			this.name = name;
			this.queuesTotal  = latchSize;
			this.concurrencyLimited = concurrencyLimited;
		}

		public boolean isConcurrencyLimited() {
			return concurrencyLimited;
		}

		public String getName() {
			return name;
		}

		public void runAndCountDown(LinkedList<T> tasks) {
			
			int size = tasks.size();
			
			AtomicBoolean tasksDone = new AtomicBoolean(false);
			AtomicReference<CountDownLatch> ref = new AtomicReference<CountDownLatch>(null);
//			AtomicReference<TaskSet<T>> taskSetRef = new AtomicReference<>(null);
			
			
			
			Runnable tasksDoneChecker = () -> {
				 
				 

						try {
							
							Thread.sleep(10000);
							
							if(!tasks.isEmpty()) {
								CountDownLatch innerLatch = new CountDownLatch(1);
								ref.set(innerLatch);
								List<T> list = new ArrayList<>();
								T task = null;
								while ((task = tasks.poll()) != null) {
									list.add(task);
								}
//								int batchSize = divideButRoundUp(list.size(), numOfThreads * numOfThreadsMultiplier);
//								Batches<T> batches = createBatches(tasks, batchSize);
//								List<LinkedList<T>> queues = batches.getBatches();
//								
//								
//								TaskSet<T> taskSet = new TaskSet<>(consumer, queues.size(), name);
//								 
//								for(LinkedList<T> queue : queues) {
//									 
//									TasksJob<T> job = new TasksJob<>(taskSet, queue);
//									pp.jobQueue.putLast(job);
//								}
//								
//								taskSetRef.set(taskSet);
//								
//								innerLatch.countDown();
//								
//								getInnerRunnable(taskSet).run();
								
								
								PersistentParallel6old.runAndFinishNoProgress(list, this.consumer, this.getClass());
								innerLatch.countDown();
								
							}
							
						} catch (InterruptedException e1) {
							// TODO Auto-generated catch block
							e1.printStackTrace();
						}

				 
 
			};
			
			if(!isConcurrencyLimited()) {
				pp.getInnerThread().setRunnable(tasksDoneChecker);
			}
			
			  
			
			T task = null;
			while ((task = tasks.poll()) != null) {
				consumer.accept(task);
			}
			
			CountDownLatch innerLatch = ref.get();
			if(innerLatch != null) {
 
				try {
					innerLatch.await();
//					taskSetRef.get().await();
					
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
			
			done.addAndGet( size);
			queuesDone.incrementAndGet();
			latch.countDown();
		}

		public void await() throws InterruptedException {
			latch.await();
		}

		public long getDone() {
			return done.get();
		}

		public boolean isDone() {
			return queuesDone.get() == queuesTotal;
		}
		
		

	}

	ThreadLocal<ReporterThread> reporterThread = new ThreadLocal<>();
	private class ReporterThread implements Runnable {

		AtomicReference<CountDownLatch> latch = new AtomicReference<>(new CountDownLatch(1));
		AtomicReference<Runnable> runnable = new AtomicReference<Runnable>(null);
		int threadNumber;
		private final String name;
		
		
		 

		public ReporterThread() {
			this.name = "REPORTER_THREAD_" + Thread.currentThread().getName();
		}

		@Override
		public void run() {
			while (running) {
//				Thread.currentThread().setName("REPORTER_THREAD_" + threadNumber);
			 
				
				try {
					latch.get().await();
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				Runnable innerRunnable = runnable.get();
				if(innerRunnable != null) {
					innerRunnable.run();
				}
				else {
					break;
				}

				latch.set(new CountDownLatch(1));
				
				
			}
		}
		
		public void setRunnable(Runnable runnable ) {
			this.runnable.set(runnable);
			this.latch.get().countDown();
		}

		public void kill() { 
			this.runnable.set(null);
			this.latch.get().countDown();
			
		}
		
	}
	private ReporterThread getReporterThread() {
		ReporterThread inenrReporterThread = reporterThread.get();
		if(inenrReporterThread == null) {
			
			inenrReporterThread = new ReporterThread();
			Thread reporterThread = new Thread(inenrReporterThread );
			pp.reporterThread.set(inenrReporterThread);
			reporterThread.start();
		}
		return inenrReporterThread;
	}
	
	private InnerThread getInnerThread() {
		InnerThread innerThreadRunnable = innerThreads.get();
		if(innerThreadRunnable == null) {
			innerThreadRunnable = new InnerThread(0);
			Thread innerThread = new Thread(innerThreadRunnable );
			pp.innerThreads.set(innerThreadRunnable);
			innerThread.start();
		}
		return innerThreadRunnable;
	}
	
	ThreadLocal<InnerThread> innerThreads = new ThreadLocal<>();
	
	AtomicInteger threadNumbers = new AtomicInteger(0);


	
	
	private class InnerThread implements Runnable {

		AtomicReference<CountDownLatch> latch = new AtomicReference<>(new CountDownLatch(1));
		AtomicReference<Runnable> runnable = new AtomicReference<Runnable>(null);
		int threadNumber;
		AtomicReference<CountDownLatch> innerStopper = new AtomicReference<>(null);
 
		
		public InnerThread(int threadNumber) {
			this.threadNumber = threadNumber;
		}

		public void pauseAwait() {
			CountDownLatch innerLatch = new CountDownLatch(1);
			innerStopper.set(innerLatch);
			try {
				innerLatch.await();
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
		}

		@Override
		public void run() {
			while (running) {
				Thread.currentThread().setName("INNER_" + threadNumber); 
				
				try {
					latch.get().await();
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				Runnable innerRunnable = runnable.get();
				if(innerRunnable != null) {
					innerRunnable.run();
				}
				else {
					break;
				}

				latch.set(new CountDownLatch(1));
				
			}
		}
		
		private   <T> Runnable getInnerRunnable(TaskSet<T> taskSet) {
			 
			return () -> {
				CountDownLatch latchA = null;
				while((latchA =  this.innerStopper.get()) == null) {
					 
					Job job = pp.jobQueue.pollLast();
					
						if (job != null) {
		
							job.runAll();
							if (job.killThread()) {
								throw new IllegalStateException();
							} else if (job.pauseThread()) {
								throw new IllegalStateException();
							}
		
						}
					 
					
				}
				this.innerStopper.set(null);
				latchA.countDown();
				
			};
		}
		
		public void setRunnable(Runnable runnable ) {
			this.runnable.set(runnable);
			this.latch.get().countDown();
		}

		public void kill() {
			InnerThread innerThreadRunnable = pp.innerThreads.get();
			if(innerThreadRunnable!= null)
					innerThreadRunnable.kill();
			this.runnable.set(null);
			this.latch.get().countDown();
			
		}
		
	}
	private class ParallelThread implements Runnable {

		@Override
		public void run() {
			AlreadyParalell.setCurrentThreadParallel(true);
			int threadNumber = threadNumbers.getAndIncrement();
			

			InnerThread innerThreadRunnable = getInnerThread();
			
			
			while (running) {
				try {
					Job job = jobQueue.takeLast();
					Thread.currentThread().setName(job.getName() + "_" + threadNumber);
					if (job != null) {

						job.runAll();
						if (job.killThread()) {
							return;
						} else if (job.pauseThread()) {
							CountDownLatch latch = new CountDownLatch(1);
							MutableBoolean kill = new MutableBoolean(false);
							threadPool.put(new PoolEntry(latch, kill));
							latch.await();
							if (kill.isTrue()) {
								innerThreadRunnable.kill();
								ReporterThread reporterThreadRunnable = pp.reporterThread.get();
								if(reporterThreadRunnable != null)
									reporterThreadRunnable.kill();
								return;
							}
						}

					}
				} catch (InterruptedException e) {
					throw new RuntimeException(e);
				}

			}
		}
	}

	public static class AlreadyParalell {
		public static InheritableThreadLocal<Boolean> alreadyParallel = new InheritableThreadLocal<>();

		public static void setCurrentThreadParallel(boolean parallel) {
			alreadyParallel.set(parallel);
		}

		public static boolean isCurrentThreadParallel() {
			Boolean bool = alreadyParallel.get();
			if (Boolean.TRUE.equals(bool))
				return true;
			else
				return false;

		}
	}

	public static void close() throws InterruptedException {
		for (int i = 0; i < numOfThreads * 4; i++) {
			PersistentParallel6old.pp.jobQueue.putFirst(new PoisenJob());
		}

		for (PoolEntry pair : pp.threadPool) {
			pair.getKill().setValue(true);
			pair.getLock().countDown();
		}

		PersistentParallel4.closeInner();
	}

	public static <E> void runAndFinish(Collection<E> items, final Consumer<E> body, String name) {
		try {
			pp.foreach(items, items.size(), body, name, -1, true, -1);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public class ParallelImpl implements PersistentParallelInterface {

	}

	public static void setThreadPaused() {
		pp.addThreadToPool();
	}

	public static void setThreadUnpaused() throws InterruptedException {
		CountDownPauseJob poisen = new CountDownPauseJob();
		pp.jobQueue.putLast(poisen);
		// poisen.await();
	}

	public static <E> void runAndFinish(Collection<E> items, final Consumer<E> body, String name, int maxConcurrent) {

		try {
			pp.foreach(items, items.size(), body, name, -1, true, maxConcurrent);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public static <E> void runAndFinish(Collection<E> items, final Consumer<E> body, final Class<?> clazz, int maxConcurrent) {

		try {
			pp.foreach(items, items.size(), body, clazz.getName(), -1, true, maxConcurrent);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public static <E> void runAndFinish(Collection<E> items, final Consumer<E> body, final Class<?> clazz) {
		try {
			pp.foreach(items, items.size(), body, clazz.getName(), -1, true, -1);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}
	
	public static <E> void runAndFinishNoProgress(Collection<E> items, final Consumer<E> body, final Class<?> clazz) {
		try {
			pp.foreach(items, items.size(), body, clazz.getName(), -1, false, -1);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public static <E> void runAndFinish(Collection<E> items, final Consumer<E> body, String name, boolean sequential) {

		if (sequential) {
			sequentialFor(items, body);
		} else {
			try {
				pp.foreach(items, items.size(), body, name, -1, true, -1);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

	}

	public static <E> void runAndFinish(Collection<E> items, final Consumer<E> body, final Class<?> clazz, boolean sequential) {

		if (sequential) {
			sequentialFor(items, body);
		} else {
			try {
				pp.foreach(items, items.size(), body, clazz.getName(), -1, true, -1);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

	}
	
	
	
	public static <E> void runAndFinish(Iterable<E> items, final Consumer<E> body, final Class<?> clazz, int batchSize, boolean sequential) {

		if (sequential) {
			sequentialFor(items, body);
		} else {
			try {
				pp.foreach(items,-1, body, clazz.getName(), batchSize, true, -1);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

	}
	
	public static <E> void runAndFinish(Iterable<E> items, final Consumer<E> body, final Class<?> clazz, int batchSize) {
 
			try {
				pp.foreach(items,-1, body, clazz.getName(), batchSize, true, -1);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
	 
	}
	
	public static <E> void runAndFinish(Iterable<E> items, final Consumer<E> body, final Class<?> clazz, boolean sequential) {
		 
		try {
			pp.foreach(items,-1, body, clazz.getName(), 5000, true, -1);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
 
}
	
	public static <E> void runAndFinish(Iterable<E> items, final Consumer<E> body, final Class<?> clazz ) {
		 
		try {
			pp.foreach(items,-1, body, clazz.getName(), 5000, true, -1);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
 
}

	private static <T> void sequentialFor(Iterable<T> items, final Consumer<T> body) {

		for (T item : items) {
			body.accept(item);
		}

	}

	private static class PoolEntry {
		CountDownLatch lock;
		MutableBoolean kill;

		private PoolEntry(CountDownLatch lock, MutableBoolean kill) {
			super();
			this.lock = lock;
			this.kill = kill;
		}

		public CountDownLatch getLock() {
			return lock;
		}

		public void setLock(CountDownLatch lock) {
			this.lock = lock;
		}

		public MutableBoolean getKill() {
			return kill;
		}

		public void setKill(MutableBoolean kill) {
			this.kill = kill;
		}

	}

	private static class Batches<T> {
		private final List<LinkedList<T>> batches;
		private final long countTotalTasks;
		public Batches(List<LinkedList<T>> batches, long countTotalTasks) {
			super();
			this.batches = batches;
			this.countTotalTasks = countTotalTasks;
		}
		public List<LinkedList<T>> getBatches() {
			return batches;
		}
		public long getCountTotalTasks() {
			return countTotalTasks;
		}
		
	}
}
