package de.mannheim.uni.expansion;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.function.Supplier;

public class ThreadLocalList<E> {

	private final ThreadLocal<E> threadLocal = new ThreadLocal<>();

	private final List<E> list = Collections.synchronizedList(new ArrayList<>());

	public E getOrCreate(Supplier<E> supplier) {
		E e = threadLocal.get();
		if (e == null) {
			e = supplier.get();
			list.add(e);
			threadLocal.set(e);
		}

		return e;
	}

	public List<E> getCurrentList() {
		return Collections.unmodifiableList(list);
	}

}
