package de.mannheim.uni.expansion.utils;

import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

public class CHMHelper {
	@FunctionalInterface
	public static interface Creator<E> {
		E create();
	}

	public static <T, E> E getFromCHM(ConcurrentHashMap<T, E> entry, T key, Creator<E> creator) {
		E value = entry.get(key);
		if (value == null) {
			E newValue = creator.create();
			value = entry.putIfAbsent(key, newValue);
			if (value == null) {
				return newValue;
			} else {
				return value;
			}
		} else {
			return value;
		}
	}

	public static <T, E> E getFromMap(Map<T, E> entry, T key, Creator<E> creator) {
		E value = entry.get(key);
		if (value == null) {
			value = creator.create();
			entry.put(key, value);

		}
		return value;

	}

	public static <A, B, C> Map<B, C> flattenMap(Map<A, Map<B, C>> unflatMap) {
		Map<B, C> flatMap = unflatMap.entrySet().parallelStream().flatMap((e) -> {
			return e.getValue().entrySet().stream();
		}).collect(Collectors.toMap(Entry::getKey, Entry::getValue));
		return flatMap;
	}


}
