package de.mannheim.uni.expansion.utils;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import javax.inject.Inject;

public class CSVSerializer {

	@Inject
	public CSVSerializer() {

	}

	public enum CSVType {
		Date("Date"), Double("Double"), String("String"), List("List");

		private final String name;

		private CSVType(String name) {
			this.name = name;
		}

		private final static Map<String, CSVType> types = new HashMap<String, CSVType>();

		static {
			for (CSVType v : CSVType.values()) {
				types.put(v.getName(), v);
			}
		}

		public String getName() {
			return name;
		}

		public static CSVType get(String name) {
			return types.get(name);
		}
	}

	public CSVType getTypeFromObject(Object o) {
		if (o instanceof String) {
			return CSVType.String;
		} else if (o instanceof Double) {
			return CSVType.Double;
		} else if (o instanceof Date) {
			return CSVType.Date;
		}

		throw new IllegalStateException("This is not a supported CSVType:" + o.getClass());
	}

	public CSVType getType(String type) {
		return CSVType.get(type);
	}

	public boolean isList(String type) {
		if (getType(type) == CSVType.List) {
			return true;
		}
		return false;
	}

	public Object getObject(String type, String value) {
		return getObject(getType(type), value);
	}

	public Object getObject(CSVType type, String value) {
		switch (type) {
		case Date:
			return getDate(value);
		case Double:
			return getDouble(value);
		case List:
			throw new IllegalStateException("This Method is Only to be used from primitive Types, not List Types");
		case String:
			return getString(value);

		}

		throw new IllegalStateException("Type does not exist");
	}

	public Date getDate(String value) {
		return new Date(Long.parseLong(value));
	}

	public String getString(String value) {
		return value;
	}

	public Double getDouble(String value) {
		return Double.parseDouble(value);
	}

	public String getString(Object o) {
		if (o instanceof String) {
			return (String) o;
		} else if (o instanceof Double) {
			return ((Double) o).toString();
		} else if (o instanceof Date) {
			return String.valueOf(((Date) o).getTime());
		}

		throw new IllegalStateException("This is not a supported CSVType:" + o.getClass());

	}

	public String[] getLine(Object o, int startIndex) {
		String[] line;
		if (o instanceof ArrayList) {
			ArrayList<?> list = (ArrayList<?>) o;
			line = new String[startIndex + 2 + list.size()];
			line[2] = CSVSerializer.CSVType.List.getName();
			if (list.size() > 0) {
				line[3] = getTypeFromObject(list.get(0)).getName();
			}

			int i = 4;
			for (Object lo : list) {
				line[i] = getString(lo);
				i++;
			}

		} else {
			line = new String[startIndex + 2];

			line[2] = getTypeFromObject(o).getName();
			line[3] = getString(o);

		}
		return line;
	}

	public Object getObject(String[] line, int startIndex) {

		CSVType type = getType(line[startIndex]);

		if (type == CSVType.List) {

			CSVType primitiveType = getType(line[startIndex + 1]);
			switch (primitiveType) {
			case Date:
				ArrayList<Date> dateList = new ArrayList<Date>();
				for (int i = startIndex + 2; i < line.length; i++) {
					dateList.add(getDate(line[i]));
				}

				return dateList;

			case Double:
				ArrayList<Double> doubleList = new ArrayList<Double>();
				for (int i = startIndex + 2; i < line.length; i++) {
					doubleList.add(getDouble(line[i]));
				}

				return doubleList;
			case String:
				ArrayList<String> stringList = new ArrayList<String>();
				for (int i = startIndex + 2; i < line.length; i++) {
					stringList.add(getString(line[i]));
				}

				return stringList;
			case List:
				throw new IllegalStateException("List can not be the primitive Type of a List!");
			default:
				throw new IllegalStateException("Unkown primitive Type!");

			}

		} else {
			return getObject(type, line[startIndex + 1]);
		}

	}
}
