package de.mannheim.uni.expansion.utils;

import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.function.Consumer;


public class ConsumerHelper<T> {

	private static int availableCores = NumOfCpusGetter.getProcessors();

	final String title;
	final int numOfThreads;

	final int queueLength;

	final Runnable producerWorker;

	CountDownLatch latch;

	final BlockingQueue<T> queue;

	final T POISEN;

	Consumer<T> consumer;

	private boolean closeAfterFinish;

	private ConsumerHelper(String title, int numOfThreads, boolean linked, int queueLength, T POISEN, Consumer<T> consumer, boolean fair) {

		this.title = title;
		this.numOfThreads = numOfThreads;
		this.queueLength = queueLength;

		this.POISEN = POISEN;

		if (linked) {
			if (queueLength == 0) {
				queue = new LinkedBlockingQueue<T>();
			} else {
				queue = new LinkedBlockingQueue<T>(queueLength);
			}
		}

		else
			queue = new ArrayBlockingQueue<T>(queueLength, fair);

		latch = new CountDownLatch(numOfThreads);

		this.consumer = consumer;

		this.closeAfterFinish = true;

		if (consumer == null) {
			producerWorker = new Runnable() {

				@Override
				public void run() {
					System.out.println(title + " consumer " + Thread.currentThread().getName() + " START");
					while (true) {

						Runnable task;
						try {
							task = (Runnable) queue.take();

							if (task == POISEN) {
								break;
							} else {
								task.run();
							}
						} catch (InterruptedException e) {
							e.printStackTrace();
						}
					}
					System.out.println(title + "Consumer " + Thread.currentThread().getName() + " END");
					latch.countDown();

				}
			};
		} else {
			producerWorker = new Runnable() {

				@Override
				public void run() {
					System.out.println(title + " consumer " + Thread.currentThread().getName() + " START");
					while (true) {

						T task;
						try {
							task = queue.take();

							if (task == POISEN) {
								if (closeAfterFinish) {
									break;
								} else {
									latch.countDown();
								}
							} else {
								getConsumer().accept(task);
							}
						} catch (InterruptedException e) {
							e.printStackTrace();
						}
					}
					System.out.println(title + "Consumer " + Thread.currentThread().getName() + " END");
					latch.countDown();

				}
			};
		}

	}

	protected Consumer<T> getConsumer() {
		return consumer;
	}

	public void add(T r) {

		queue.add(r);
	}

	public void put(T r) throws InterruptedException {
		queue.put(r);
	}

	public synchronized void start() {
		for (int i = 0; i < numOfThreads; i++) {
			Thread thread = new Thread(producerWorker);
			thread.setName(title + "::" + i);
			thread.start();
		}
	}

	public synchronized void finishClose() throws InterruptedException {

		for (int i = 0; i < numOfThreads; i++) {
			queue.put(POISEN);
		}

	}

	public synchronized void finishCloseAndAwait() throws InterruptedException {

		finishClose();

		latch.await();

	}

	public synchronized void finishAwaitAndRestart() throws InterruptedException {

		closeAfterFinish = false;

		finishClose();

		latch.await();

		latch = new CountDownLatch(numOfThreads);

		closeAfterFinish = true;

	}

	public static ConsumerHelper<Runnable> withLinkedQueue(String name) {
		Runnable POISEN = new Runnable() {

			@Override
			public void run() {

			}
		};
		return new ConsumerHelper<Runnable>(name, availableCores, true, 0, POISEN, null, false);
	}

	public static <E> ConsumerHelper<E> withArrayQueue(String name, Consumer<E> consumer, int queueLength, E poisen, boolean fair) {
		return new ConsumerHelper<E>(name, availableCores, false, queueLength, poisen, consumer, fair);
	}

	public static <E> ConsumerHelper<E> withLinkedQueue(String name, Consumer<E> consumer, E poisen) {
		return new ConsumerHelper<E>(name, availableCores, true, 0, poisen, consumer, false);
	}

	public static <E> ConsumerHelper<E> withLinkedQueue(String name, Consumer<E> consumer, int queueLength, E poisen) {
		return new ConsumerHelper<E>(name, availableCores, true, queueLength, poisen, consumer, false);
	}

	public static <E> ConsumerHelper<E> withLinkedQueue(String name, int threads, Consumer<E> consumer, E poisen) {
		return new ConsumerHelper<E>(name, threads, true, 0, poisen, consumer, false);
	}

	public static <E> ConsumerHelper<Runnable> withLinkedQueue(String name, int threads) {
		Runnable POISEN = new Runnable() {

			@Override
			public void run() {

			}
		};
		return new ConsumerHelper<Runnable>(name, threads, true, 0, POISEN, null, false);
	}

	public static void setAvailableCores(int cores) {
		availableCores = cores;
	}

	public static int getAvailableCores() {
		return availableCores;
	}

	public void setConsumer(Consumer<T> newConsumer) {
		this.consumer = newConsumer;

	}

}
