package de.mannheim.uni.expansion.utils;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import de.mannheim.uni.expansion.utils.DoubleIteraterable.DoubleEntry;
 

public class DoubleIteraterable<E> implements Iterable<DoubleEntry<E>> {

	public static long getResultingSize(int originalSize) {
		if (originalSize == 1)
			throw new IllegalStateException();
		BigInteger ret = factorial(originalSize).divide(factorial(2).multiply(factorial(originalSize - 2)));

		return ret.longValue();

	}

	private static BigInteger factorial(int number) {
		BigInteger n = BigInteger.valueOf(number);
		BigInteger result = BigInteger.ONE;

		while (!n.equals(BigInteger.ZERO)) {
			result = result.multiply(n);
			n = n.subtract(BigInteger.ONE);
		}

		return result;
	}
	public static class DoubleEntry<E> {
		private final E first;
		private final E second;

		private DoubleEntry(E first, E second) {
			super();
			this.first = first;
			this.second = second;
		}

		public E getFirst() {
			return first;
		}

		public E getSecond() {
			return second;
		}

		 

	}



	private final List<E> list;
	private final int originalSize;
	// private final long combinations;

	private DoubleIteraterable(Iterable<E> iterable) {
		list = new ArrayList<>();
		iterable.forEach(list::add);
		this.originalSize = list.size();
		// this.combinations = combination(this.originalSize, 2);
		// long maxInt = Integer.MAX_VALUE;
		// if(combinations >= maxInt) {
		// throw new IllegalStateException("THIS IS TOO BIG FOR THIS
		// COLLECTION");
		// }
	}

	// private long combination(int n, int r) {
	// if (r > n / 2)
	// r = n - r; // because C(n, r) == C(n, n - r)
	// long ans = 1;
	// long i;
	//
	// for (i = 1; i <= r; i++) {
	// ans *= n - r + i;
	// ans /= i;
	// }
	//
	// return ans;
	// }

	public static <E> DoubleIteraterable<E> iterable(Iterable<E> iterable) {
		return new DoubleIteraterable<>(iterable);
	}

	@Override
	public DoubleIterator iterator() {
		return new DoubleIterator();
	}

	private class DoubleIterator implements Iterator<DoubleEntry<E>> {


		private int currentOuter = 0;
		private int currentInner = 1;

		public DoubleIterator() {
		}

		@Override
		public boolean hasNext() {
			if (currentOuter == (originalSize - 1))
				return false;
			return true;
		}

		@Override
		public DoubleEntry<E> next() {
			if (currentOuter == (originalSize - 1))
				return null;

			DoubleEntry<E> ret = new DoubleEntry<>(list.get(currentOuter), list.get(currentInner));

			currentInner++;
			if (currentInner == originalSize) {
				currentOuter++;
				currentInner = currentOuter + 1;
			}

			return ret;
		}
		


	}
 


}
