package de.mannheim.uni.expansion.utils;

import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NavigableSet;
import java.util.Set;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableMap.Builder;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.ImmutableSortedSet;

public class ImmuMaker {

	public static <A, B, C> Map<A, Map<B, C>> mapWithMap(Map<A, ? extends Map<B, C>> input) {

		Builder<A, Map<B, C>> builder = ImmutableMap.builder();

		for (Entry<A, ? extends Map<B, C>> e : input.entrySet()) {
			builder.put(e.getKey(), map(e.getValue()));
		}
		return builder.build();
	}

	public static <A, B> Map<A, Set<B>> mapWithSet(Map<A, Set<B>> input) {

		Builder<A, Set<B>> builder = ImmutableMap.builder();

		for (Entry<A, Set<B>> e : input.entrySet()) {
			builder.put(e.getKey(), set(e.getValue()));
		}
		return builder.build();
	}

	public static <A, B> Map<A, NavigableSet<B>> mapWithNavigableSet(Map<A, NavigableSet<B>> input) {

		Builder<A, NavigableSet<B>> builder = ImmutableMap.builder();

		for (Entry<A, NavigableSet<B>> e : input.entrySet()) {
			builder.put(e.getKey(), navigableSet(e.getValue()));
		}
		return builder.build();
	}

	public static <A, B> Map<A, B> map(Map<A, B> input) {

		return ImmutableMap.copyOf(input);
	}

	public static <B> Set<B> set(Set<B> value) {
		return ImmutableSet.copyOf(value);
	}

	public static <B> List<B> list(List<B> value) {
		return ImmutableList.copyOf(value);
	}

	public static <B> NavigableSet<B> navigableSet(NavigableSet<B> value) {

		return ImmutableSortedSet.copyOfSorted(value);
	}

	public static <A, B> Map<A, List<B>> withList(Map<A, List<B>> input) {
		Builder<A, List<B>> builder = ImmutableMap.builder();

		for (Entry<A, List<B>> e : input.entrySet()) {
			builder.put(e.getKey(), list(e.getValue()));
		}
		return builder.build();
	}
}
