package de.mannheim.uni.expansion.utils;

import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import de.mannheim.uni.expansion.utils.ListIterable.ListEntry;

public class ListIterable<E> implements Collection<ListEntry<E>> {

	public static class ListEntry<E> {
		private final int index;
		private final E value;
		private final List<E> list;

		private ListEntry(int index, E value, List<E> list) {
			super();
			this.index = index;
			this.value = value;
			this.list = list;
		}

		public int getIndex() {
			return index;
		}

		public E getValue() {
			return value;
		}

		public void setValue(E val) {
			list.set(index, val);
		}

	}



	private final List<E> list;

	private ListIterable(List<E> list) {

		this.list = list;
	}

	public static <E> ListIterable<E> iterable(List<E> list) {
		return new ListIterable<>(list);
	}

	@Override
	public ListIterator iterator() {
		return new ListIterator();
	}

	public class ListIterator implements Iterator<ListEntry<E>> {

		private int currentIndex;

		public ListIterator() {
			this.currentIndex = 0;
		}

		@Override
		public boolean hasNext() {
			return currentIndex < list.size();
		}

		@Override
		public ListEntry<E> next() {
			if (currentIndex < list.size()) {
				ListEntry<E> entry = new ListEntry<>(currentIndex, list.get(currentIndex), list);
				currentIndex++;
				return entry;
			} else
				return null;
		}
		


	}

	@Override
	public int size() {
		return list.size();
	}

	@Override
	public boolean isEmpty() {
		return list.isEmpty();
	}

	@Override
	public boolean contains(Object o) {
		return list.contains(o);
	}

	@Override
	public Object[] toArray() {
		throw new UnsupportedOperationException();
	}

	@Override
	public <T> T[] toArray(T[] a) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean add(ListEntry<E> e) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean remove(Object o) {
		return list.remove(o);
	}

	@Override
	public boolean containsAll(Collection<?> c) {
		return list.containsAll(c);
	}

	@Override
	public boolean addAll(Collection<? extends ListEntry<E>> c) {

		throw new UnsupportedOperationException();
	}

	@Override
	public boolean removeAll(Collection<?> c) {
		return list.removeAll(c);
	}

	@Override
	public boolean retainAll(Collection<?> c) {

		throw new UnsupportedOperationException();
	}

	@Override
	public void clear() {
		list.clear();

	}

}
