package de.mannheim.uni.expansion.utils;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.List;
import java.util.Map.Entry;
import java.util.NavigableMap;
import java.util.SortedMap;
import java.util.TreeMap;

public class MinVarianceTwoClassSplitter {

	public static void main(String[] args) throws IOException {
		List<String> lines = Files.readAllLines(Paths.get("D://newStuff/histogramm"));

		NavigableMap<Integer, Integer> map = new TreeMap<>();

		for (String s : lines) {
			String[] parts = s.split(";");
			map.put(Integer.parseInt(parts[0]), Integer.parseInt(parts[0]));
		}

		Integer thrshold = findThreshold(map);

		System.out.println(thrshold);
	}

	public MinVarianceTwoClassSplitter() {

	}

	public static Integer findThreshold(NavigableMap<Integer, Integer> input) {

		Double variance = null;
		Integer threshold = null;
		for (Entry<Integer, Integer> entry : input.entrySet()) {

			double retVariance = verianceSum(input, entry.getKey());
			if (threshold == null || retVariance < variance) {
				variance = retVariance;
				threshold = entry.getKey();
			}
		}

		{
			Integer tt = input.descendingKeySet().first() + 1;
			double retVariance = verianceSum(input, tt);
			if (threshold == null || retVariance < variance) {
				variance = retVariance;
				threshold = tt;
			}
		}

		return threshold;
	}

	public static double verianceSum(SortedMap<Integer, Integer> input, Integer threshold) {
		double sumBefore = 0.0;
		int countBefore = 0;

		double sumAfter = 0.0;
		int countAfter = 0;

		for (Entry<Integer, Integer> entry : input.entrySet()) {
			if (entry.getKey() >= threshold) {
				sumAfter += entry.getValue();
				countAfter++;
			} else {
				sumBefore += entry.getValue();
				countBefore++;
			}
		}

		double meanBefore;
		if (countBefore != 0) {
			meanBefore = sumBefore / countBefore;
		} else {
			meanBefore = 0.0;
		}
		double meanAfter;
		if (countAfter != 0) {
			meanAfter = sumAfter / countAfter;
		} else {
			meanAfter = 0.0;
		}

		double variance = 0.0;

		for (Entry<Integer, Integer> entry : input.entrySet()) {
			if (entry.getKey() >= threshold) {
				variance += (entry.getValue() - meanAfter) * (entry.getValue() - meanAfter);
			} else {
				variance += (entry.getValue() - meanBefore) * (entry.getValue() - meanBefore);
			}
		}

		return variance;
	}
}
