package de.mannheim.uni.expansion.utils;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.inject.Inject;


public class OtsuThresholder {

	@Inject
	public OtsuThresholder() {
	}

	public <T extends Scorable> SortedSet<T> threshold(SortedSet<T> pairs) {

		// http://www.labbookpages.co.uk/software/imgProc/otsuThreshold.html

		Map<Double, Integer> histogram = new LinkedHashMap<Double, Integer>();

		int countTotal = pairs.size();
		double sumTotal = 0.0;
		for (T p : pairs) {
			Integer count = histogram.get(p.getScore());
			if (count == null) {
				histogram.put(p.getScore(), 1);
			} else {

				histogram.put(p.getScore(), count + 1);
			}

			sumTotal += p.getScore();
		}

		double sumB = 0;
		int wB = 0;
		int wF = 0;
		double mB;
		double mF;
		double varMax = 0.0;
		double between = 0.0;
		double threshold1 = 0.0;
		double threshold2 = 0.0;

		for (Entry<Double, Integer> e : histogram.entrySet()) {

			wB += e.getValue();
			if (wB == 0)
				continue;

			wF = countTotal - wB;
			if (wF == 0)
				break;

			sumB += e.getKey() * e.getValue();

			mB = sumB / wB;
			mF = (sumTotal - sumB) / wF;

			between = wB * wF * (mB - mF) * (mB - mF);
			if (between > varMax) {
				threshold1 = e.getKey();
				// if (between > varMax) {
				// threshold2 = e.getKey();
				// }
				varMax = between;
			}
		}

		// double threshold = (threshold1 + threshold2) / 2.0;

		SortedSet<T> thresholdedResults = new TreeSet<T>();

		for (T pair : pairs) {
			if (pair.getScore() >= threshold1) {
				thresholdedResults.add(pair);
			}

		}
		return thresholdedResults;

	}

}
