package de.mannheim.uni.expansion.utils;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.stream.Collectors;

import javax.inject.Inject;

import de.mannheim.uni.expansion.utils.OtsuThresholderRounded.DefaultIntegerScorable.DoubleScorable;

public class OtsuThresholderRounded {

	@Inject
	public OtsuThresholderRounded() {
	}

	public <T extends Scorable> Double getThresholdDefault(List<T> pairs) {
		if (pairs.size() > 1) {
			int same = 1;
			double firstScore = pairs.get(0).getScore();
			for (int i = 1; i < pairs.size(); i++) {
				if (Math.abs(pairs.get(i).getScore() - firstScore) < 0.01) {
					same++;
				}
 else {
					break;
				}
			}
			if (same == pairs.size()) {
				return 1.1;
			}
		}

		List<DefaultIntegerScorable> newPairs = pairs.stream().map((e) -> {
			return new DefaultIntegerScorable(e);
		}).collect(Collectors.toList());

		double threshold1 = getThreshold(newPairs, (e) -> {
			return e;
		});
		double threshold2 = getThreshold(newPairs, (e) -> {
			return (int) (5 * (Math.ceil((e / 5.0))));
		});
		double threshold3 = getThreshold(newPairs, (e) -> {
			return (int) (10 * (Math.ceil((e / 10.0))));
		});

		double average = (threshold1 + threshold2 + threshold3) / 3;
		average = average / 10000;
		return average;

	}

	public <T extends Scorable> Double getThresholdNoExtra(Collection<T> pairs) {
		// if (pairs.size() > 1) {
		// int same = 1;
		// double firstScore = pairs.get(0).getScore();
		// for (int i = 1; i < pairs.size(); i++) {
		// if (Math.abs(pairs.get(i).getScore() - firstScore) < 0.01) {
		// same++;
		// }
		// else {
		// break;
		// }
		// }
		// if (same == pairs.size()) {
		// return 1.1;
		// }
		// }

		List<DefaultIntegerScorable> newPairs = pairs.stream().map((e) -> {
			return new DefaultIntegerScorable(e);
		}).collect(Collectors.toList());

		double threshold1 = getThreshold(newPairs, (e) -> {
			return e;
		});
		// double threshold2 = getThreshold(newPairs, (e) -> {
		// return (int) (5 * (Math.floor((e / 5.0))));
		// });
		// double threshold3 = getThreshold(newPairs, (e) -> {
		// return (int) (10 * (Math.floor((e / 10.0))));
		// });

		double average = (threshold1);
		average = average / 10000;
		return average;


	}

	public Double getThresholdDouble(Collection<Double> pairs) {
		// if (pairs.size() > 1) {
		// int same = 1;
		// double firstScore = pairs.get(0).getScore();
		// for (int i = 1; i < pairs.size(); i++) {
		// if (Math.abs(pairs.get(i).getScore() - firstScore) < 0.01) {
		// same++;
		// }
		// else {
		// break;
		// }
		// }
		// if (same == pairs.size()) {
		// return 1.1;
		// }
		// }

		List<DefaultIntegerScorable> newPairs = pairs.stream().map((e) -> {
			return new DefaultIntegerScorable(new DoubleScorable(e));
		}).collect(Collectors.toList());

		double threshold1 = getThreshold(newPairs, (e) -> {
			return e;
		});
		// double threshold2 = getThreshold(newPairs, (e) -> {
		// return (int) (5 * (Math.floor((e / 5.0))));
		// });
		// double threshold3 = getThreshold(newPairs, (e) -> {
		// return (int) (10 * (Math.floor((e / 10.0))));
		// });

		double average = (threshold1);
		average = average / 10000;
		return average;

	}

	public <T extends Scorable> List<T> thresholdDefault(List<T> pairs) {

		double average = getThresholdDefault(pairs);

		List<T> thresholdedResults = new ArrayList<T>();

		for (T pair : pairs) {
			DefaultIntegerScorable asInt = new DefaultIntegerScorable(pair);

			// int count = 0;
			// if (asInt.getScore() > threshold1) {
			// count++;
			// }
			// if (asInt.getScore() > threshold2) {
			// count++;
			// }
			// if (asInt.getScore() > threshold3) {
			// count++;
			// }

			if (pair.getScore() > average) {
				thresholdedResults.add(pair);
			}

		}
		return thresholdedResults;
	}

	public static class DefaultIntegerScorable implements IntegerScorable, Comparable<IntegerScorable>
	{
		private final Scorable score;

		public DefaultIntegerScorable(DoubleScorable score) {
			this.score = score;
		}

		public DefaultIntegerScorable(Scorable score) {
			this.score = score;
		}

		@Override
		public Integer getScore() {
			if (score.getScore() > 0)
				return (int) Math.floor(score.getScore() * 10000);
			else {
				return (int) Math.ceil(score.getScore() * 10000);
			}
		}



		@Override
		public int compareTo(IntegerScorable o) {
			return this.getScore().compareTo(o.getScore());
		}

		@Override
		public String toString() {
			return "" + score.getScore();
		}

		public static class DoubleScorable implements Scorable {
			private final double score;

			public DoubleScorable(double score) {
				super();
				this.score = score;
			}

			@Override
			public Double getScore() {
				return score;
			}

		}

	}

	public <T extends IntegerScorable> double getThreshold(List<T> pairs, IntegerRounder rounder) {
		Map<Integer, Integer> histogram = new LinkedHashMap<Integer, Integer>();
		ArrayList<T> newList = new ArrayList<>(pairs);
		Collections.sort(newList);

		int countTotal = newList.size();
		double sumTotal = 0.0;
		for (T p : newList) {
			Integer score = rounder.round(p.getScore());
			Integer count = histogram.get(score);
			if (count == null) {
				histogram.put(score, 1);
			} else {

				histogram.put(score, count + 1);
			}

			sumTotal += p.getScore();
		}

		double sumB = 0;
		int wB = 0;
		int wF = 0;
		double mB;
		double mF;
		double varMax = 0.0;
		double between = 0.0;
		double threshold1 = 0.0;
		double threshold2 = 0.0;

		for (Entry<Integer, Integer> e : histogram.entrySet()) {

			wB += e.getValue();
			if (wB == 0)
				continue;

			wF = countTotal - wB;
			if (wF == 0)
				break;

			sumB += e.getKey() * e.getValue();

			mB = sumB / wB;
			mF = (sumTotal - sumB) / wF;

			between = wB * wF * (mB - mF) * (mB - mF);
			if (between >= varMax) {
				threshold1 = e.getKey();
				if (between > varMax) {
					threshold2 = e.getKey();
				}
				varMax = between;
			}
		}

		return (threshold1 + threshold2) / 2.0;
	}

	public <T extends IntegerScorable> List<T> threshold(List<T> pairs, IntegerRounder rounder) {

		// http://www.labbookpages.co.uk/software/imgProc/otsuThreshold.html

		double threshold = getThreshold(pairs, rounder);

		List<T> thresholdedResults = new ArrayList<T>();

		for (T pair : pairs) {
			if (pair.getScore() >= threshold) {
				thresholdedResults.add(pair);
			}

		}
		return thresholdedResults;

	}

	public interface IntegerScorable extends Comparable<IntegerScorable> {
		Integer getScore();
	}

	@FunctionalInterface
	public static interface IntegerRounder {
		public Integer round(Integer score);
	}

}
