package de.mannheim.uni.expansion.utils;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class SimpleArrayMap<T> implements Map<Integer, T> {

	Object[] data;


	public SimpleArrayMap(int capacity) {
		this.data = new Object[capacity];
	}

	private int getIndex(Object key) {
		if (key instanceof Integer && key != null) {
			int index = (int) key;
			if (index < data.length && index >= 0) {
				return index;
			} else {
				throw new IllegalStateException("Key must be of in range 0 to " + (data.length - 1));
			}
		} else {
			throw new IllegalStateException("Key must be of type Integer!");
		}
	}

	@Override
	public int size() {
		return data.length;
	}

	@Override
	public boolean isEmpty() {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean containsKey(Object key) {
		int index = getIndex(key);
		if (data[index] != null) {
			return true;
		}
		return false;
	}

	@Override
	public boolean containsValue(Object value) {
		throw new UnsupportedOperationException();
	}

	@Override
	public T get(Object key) {
		int index = getIndex(key);
		@SuppressWarnings("unchecked")
		T retValue = (T) data[index];
		return retValue;
	}

	@Override
	public T put(Integer key, T value) {
		int index = getIndex(key);
		@SuppressWarnings("unchecked")
		T oldValue = (T) data[index];
		data[index] = value;
		return oldValue;
	}

	@Override
	public T remove(Object key) {
		int index = getIndex(key);
		@SuppressWarnings("unchecked")
		T oldValue = (T) data[index];
		data[index] = null;
		return oldValue;
	}

	@Override
	public void putAll(Map<? extends Integer, ? extends T> m) {
		for (java.util.Map.Entry<? extends Integer, ? extends T> e : m.entrySet()) {
			this.put(e.getKey(), e.getValue());
		}

	}

	@Override
	public void clear() {
		data = new Object[data.length];

	}

	@Override
	public Set<Integer> keySet() {
		Set<Integer> ret = new HashSet<Integer>();
		for (int i = 0; i < data.length; i++) {
			if (data[i] != null)
			ret.add(i);
		}
		return ret;

	}

	@SuppressWarnings("unchecked")
	@Override
	public Collection<T> values() {
		ArrayList<T> ret = new ArrayList<>();
		for (int i = 0; i < data.length; i++) {
			if (data[i] != null)
				ret.add((T) data[i]);
		}
		return ret;
	}

	@Override
	public Set<Entry<Integer, T>> entrySet() {
		Set<Entry<Integer, T>> ret = new HashSet<Entry<Integer, T>>();
		for (int i = 0; i < data.length; i++) {

			if (data[i] != null) {

				final Integer key = i;

				Entry<Integer, T> e = new Entry<Integer, T>() {

					@Override
					public Integer getKey() {
						return key;
					}

					@SuppressWarnings("unchecked")
					@Override
					public T getValue() {
						return (T) data[key];
					}

					@Override
					public T setValue(T value) {
						return put(key, value);
					}

				};

				ret.add(e);
			}


		}
		return ret;
	}

}
