package de.mannheim.uni.expansion.utils.tree;

import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;

import de.mannheim.uni.expansion.utils.CHMHelper;

public class Tree<E> {

	public static void main(String[] args) {
		Tree<String> tree = new Tree<String>();

		tree.addRootNode("Root A", "Root A");
		tree.addNode("Child A.1", "Root A", "Child A.1");
		tree.addNode("Child A.2", "Root A", "Child A.2");
		tree.addNode("Child A.2.1", "Child A.2", "Child A.2.1");
		tree.addNode("Child A.2.1.1", "Child A.2.1", "Child A.2.1.1");
		tree.addNode("Child A.3", "Root A", "Child A.3");

		tree.addRootNode("Root B", "Root B");
		tree.addNode("Child B.1", "Root B", "Child B.1");
		tree.addNode("Child B.1.1", "Child B.1", "Child B.1.1");
		tree.addNode("Child B.1.1.1", "Child B.1.1", "Child B.1.1.1");
		tree.addNode("Child B.1.1.2", "Child B.1.1", "Child B.1.1.2");
		tree.addRootNode("Root C", "Root C");

		System.out.println(tree.toString());
	}

	private final ConcurrentLinkedQueue<E> rootNodes = new ConcurrentLinkedQueue<>();

	private final ConcurrentHashMap<E, ConcurrentLinkedQueue<E>> childNodes = new ConcurrentHashMap<>();

	private final ConcurrentHashMap<E, String> labels = new ConcurrentHashMap<>();

	public void addRootNode(E node, String label) {
		rootNodes.add(node);
		labels.put(node, label);
	}

	public void addNode(E node, E parent, String label) {
		CHMHelper.getFromCHM(childNodes, parent, () -> new ConcurrentLinkedQueue<>()).add(node);
		labels.put(node, label);
	}

	public String toString() {
		StringBuilder builder = new StringBuilder();
		for(E node : rootNodes) {
			builder.append(labels.get(node) + "\n");

			append(builder, "", node);
			
		}

		return builder.toString();
	}

	private void append(StringBuilder builder, String prefixString, E node) {


		if (childNodes.containsKey(node)) {
			int i = 0;
			int size = childNodes.get(node).size();
			for (E childNode : childNodes.get(node)) {
				String finalPrefixString = null;
				if (i < size - 1) {
					finalPrefixString = prefixString + "|   ";
				} else {
					finalPrefixString = prefixString + "    ";
				}

				String midString = "├── ";
				if (i == size - 1) {
					midString = "└── ";
				}

				builder.append(prefixString + midString + labels.get(childNode) + "\n");
				append(builder, finalPrefixString, childNode);
				i++;
			}
		}
	}
}
