package de.uni_mannheim.informatik.dws.expansion.wikidata;

import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats.Stats;

public class WikidataClasses {
	public final static PropertyId instanceOf = new PropertyId("P31");

	public final static PropertyId occuptation = new PropertyId("P106");
	public final static PropertyId subClass = new PropertyId("P279");

	public final static InstanceId country = InstanceId.simple("Q6256");
	public final static InstanceId human = InstanceId.simple("Q5");
	public final static InstanceId city = InstanceId.simple("Q515");
	public final static InstanceId bigCity = InstanceId.simple("Q1549591");
	public final static InstanceId independentCity = InstanceId.simple("Q22865");

	public final static InstanceId gridIronFootballPlayer = InstanceId.simple("Q14128148");
	public final static InstanceId americalFootballPlayer = InstanceId.simple("Q19204627");
	public final static InstanceId basketballPlayer = InstanceId.simple("Q3665646");
	public final static InstanceId publicCompany = InstanceId.simple("Q891723");
	public final static InstanceId publicLimitedCompany = InstanceId.simple("Q5225895");
	public final static InstanceId aktiengesellschaft = InstanceId.simple("Q422062");

	public final static InstanceId airline = InstanceId.simple("Q46970");
	public final static InstanceId associationFootballer = InstanceId.simple("Q937857");

	public final static InstanceId politician = InstanceId.simple("Q82955");

	public final static InstanceId position = InstanceId.simple("Q4164871");
	public final static InstanceId publicOffice = InstanceId.simple("Q294414");

	public final static InstanceId office = InstanceId.simple("Q294414");
	// public final static InstanceId headOfGovernment =
	// InstanceId.simple("Q2285706");
	public final static InstanceId headOfGovernmentSubclass = InstanceId.simple("Q2285706");
	public final static InstanceId headOfStateSubclass = InstanceId.simple("Q48352");
	public final static InstanceId lordMayor = InstanceId.simple("Q11902879");
	public final static InstanceId mayor = InstanceId.simple("Q30185");

	public final static InstanceId basketballTeam = InstanceId.simple("Q13393265");

	public static final ClassId wikidataCountry = new ClassId("Country");
	public static final ClassId wikidataCity = new ClassId("City");
	public static final ClassId wikidataFootball = new ClassId("FootballPlayer");
	public static final ClassId wikidataBasketball = new ClassId("BasketballPlayer");
	public static final ClassId wikidataBasketballTeam = new ClassId("BasketballTeam");
	public static final ClassId wikidataTradedCompany = new ClassId("TradedCompany");
	public static final ClassId wikidataOfficeHolder = new ClassId("OfficeHolder");

	public static final ClassId wikidataAll = new ClassId("ALL");

	public static final ClassId wikidataSoccer = new ClassId("SoccerPlayer");

	private final static Map<InstanceId, ClassId> setMap = new HashMap<>();

	private final static Set<ClassId> setSetFinal = new HashSet<>();

	private final static Map<ClassId, Set<PropertyId>> classPropertyMap = new HashMap<>();

	static {
		setMap.put(country, wikidataCountry);
		setMap.put(city, wikidataCity);
		setMap.put(bigCity, wikidataCity);
		setMap.put(independentCity, wikidataCity);
		setMap.put(gridIronFootballPlayer, wikidataFootball);
		setMap.put(americalFootballPlayer, wikidataFootball);
		setMap.put(basketballPlayer, wikidataBasketball);
		setMap.put(publicCompany, wikidataTradedCompany);
		setMap.put(publicLimitedCompany, wikidataTradedCompany);
		setMap.put(aktiengesellschaft, wikidataTradedCompany);
		setMap.put(associationFootballer, wikidataSoccer);
		setMap.put(politician, wikidataOfficeHolder);
		setMap.put(basketballTeam, wikidataBasketballTeam);

		setSetFinal.add(wikidataCountry);
		setSetFinal.add(wikidataCity);
		setSetFinal.add(wikidataCity);
		setSetFinal.add(wikidataCity);
		setSetFinal.add(wikidataFootball);
		setSetFinal.add(wikidataFootball);
		setSetFinal.add(wikidataBasketball);
		setSetFinal.add(wikidataTradedCompany);
		setSetFinal.add(wikidataTradedCompany);
		setSetFinal.add(wikidataTradedCompany);
		setSetFinal.add(wikidataSoccer);
		setSetFinal.add(wikidataOfficeHolder);

		classPropertyMap.put(wikidataCountry,
				new HashSet<>(Arrays.asList(Stats.label, PropertiesExpansionData.TRANSLATED_LABELS_PROPERTY,
						PropertiesExpansionData.TRANSLATED_ALIASES_PROPERTY, Stats.anthem, Stats.area, Stats.capital, Stats.currency, Stats.coordinateLocation,
						Stats.isoAlpha3Codes, Stats.fips10_4, Stats.tld, Stats.locatedInTimeZone, Stats.continent, Stats.countryCallingCode,
						Stats.isoNumericCode, Stats.nominalGDP, Stats.nominalGDPperCapita, Stats.population, Stats.populationDensity, Stats.memberOf,
						Stats.iocCountryCode, Stats.headOfGovernemnt, Stats.headOfState, Stats.drivingSide, Stats.sharesBorderWith

				)));

		classPropertyMap.put(wikidataBasketball,
				new HashSet<>(Arrays.asList(Stats.label, PropertiesExpansionData.TRANSLATED_LABELS_PROPERTY,
						PropertiesExpansionData.TRANSLATED_ALIASES_PROPERTY, Stats.gender, Stats.dateOfBirth, Stats.countryOfCitizenship, Stats.sport,
						Stats.givenName, Stats.placeOfBirth, Stats.positionPlayed, Stats.placeOfBirth, Stats.educatedAt, Stats.teams, Stats.draftedBy,
						Stats.familyName, Stats.participantOf, Stats.dateOfDeath

				)));

		classPropertyMap.put(wikidataCity,
				new HashSet<>(Arrays.asList(Stats.label, PropertiesExpansionData.TRANSLATED_LABELS_PROPERTY,
						PropertiesExpansionData.TRANSLATED_ALIASES_PROPERTY, Stats.country, Stats.coordinateLocation, Stats.locatedIn, Stats.elevation,
						Stats.isoNumericCode, Stats.officialWebsite, Stats.postalCode, Stats.sisterCity, Stats.capitalOf, Stats.population,
						Stats.sharesBorderWith, Stats.localDialingCode, Stats.headOfGovernemnt

				)));

		classPropertyMap.put(wikidataFootball,
				new HashSet<>(Arrays.asList(Stats.label, PropertiesExpansionData.TRANSLATED_LABELS_PROPERTY,
						PropertiesExpansionData.TRANSLATED_ALIASES_PROPERTY, Stats.gender, Stats.dateOfBirth, Stats.givenName, Stats.countryOfCitizenship,
						Stats.teams, Stats.sportNumber, Stats.placeOfBirth, Stats.dateOfDeath, Stats.sport, Stats.educatedAt, Stats.familyName

				)));

		classPropertyMap.put(wikidataOfficeHolder, new HashSet<>(Arrays.asList(Stats.label, PropertiesExpansionData.TRANSLATED_LABELS_PROPERTY,
				PropertiesExpansionData.TRANSLATED_ALIASES_PROPERTY, Stats.gender, Stats.dateOfBirth, Stats.givenName, Stats.headOfGovernemnt,
				Stats.countryOfCitizenship, Stats.dateOfDeath, Stats.placeOfBirth, Stats.placeOfDeath, Stats.educatedAt, Stats.positionHeld, Stats.party)));

		classPropertyMap
				.put(wikidataTradedCompany,
						new HashSet<>(Arrays.asList(Stats.label, PropertiesExpansionData.TRANSLATED_LABELS_PROPERTY,
								PropertiesExpansionData.TRANSLATED_ALIASES_PROPERTY, Stats.headquartersLocation, Stats.inception, Stats.officialWebsite,
								Stats.netIncome, Stats.totalRevenue, Stats.totalEquity, Stats.totalAssets, Stats.ebit, Stats.country, Stats.legalForm,
								Stats.industry, Stats.founder, Stats.stockExchange

						)));

		classPropertyMap.put(wikidataSoccer,
				new HashSet<>(Arrays.asList(Stats.label, PropertiesExpansionData.TRANSLATED_LABELS_PROPERTY,
						PropertiesExpansionData.TRANSLATED_ALIASES_PROPERTY, Stats.dateOfBirth, Stats.gender, Stats.sport, Stats.teams,
						Stats.countryOfCitizenship, Stats.positionPlayed, Stats.placeOfBirth, Stats.givenName, Stats.sportNumber, Stats.dateOfDeath

				)));
	}

	public static ClassId getSet(InstanceId instanceId) {
		return setMap.get(instanceId);
	}

	public static boolean propertyExist(ClassId setId, PropertyId propertyId) {
		if (classPropertyMap.containsKey(setId)) {
			return classPropertyMap.get(setId).contains(propertyId);
		}
		return false;
	}

	public static boolean classExists(ClassId setId) {
		return setSetFinal.contains(setId);
	}
}
