package de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.hierarchies;

import java.io.Serializable;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import com.google.common.collect.ImmutableMap;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedItem;

public class WikidataLabels implements Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	private final ImmutableMap<InstanceId, String> labels;

	public WikidataLabels(Map<InstanceId, String> labels) {
		super();
		this.labels = ImmutableMap.copyOf(labels);
	}

	public InstanceId getTransformedInstanceId(InstanceId instanceId) {
		if (labels.containsKey(instanceId)) {
			return new InstanceId(instanceId.getInstanceId(), labels.get(instanceId));
		}
		return instanceId;
	}

	public String getLabel(InstanceId instanceId) {
		if (labels.containsKey(instanceId)) {
			return labels.get(instanceId);
		}
		return instanceId.getReadableString();

	}

	public static class WikidataLabelsFactory {
		private ConcurrentHashMap<InstanceId, String> labels = new ConcurrentHashMap<>();
		private boolean destroyed = false;

		public WikidataLabelsFactory() {

		}

		public void add(WikidataParsedItem item) {
			if (!destroyed) {
				InstanceId instanceId = item.getInstanceId();
				labels.putIfAbsent(instanceId, item.getLabel());
			}
		}

		public WikidataLabels create() {
			if (!destroyed) {
				return new WikidataLabels(labels);

			} else {
				throw new IllegalStateException("THIS FACTORY HAS ALREADY BEEN DESTROYED");
			}
		}

		public void destroy() {
			this.labels = null;
			this.destroyed = true;

		}
	}

}
