package de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.importers;

import java.util.HashSet;
import java.util.Set;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedItem;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.wikidata.WikidataClasses;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.hierarchies.WikidataHierarchy;
import de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats.Stats;

public class FixedClassGetter {
	Set<InstanceId> classesToImport = new HashSet<>();
	Set<InstanceId> occupationsToImport = new HashSet<>();

	@Inject
	public FixedClassGetter() {
		classesToImport.add(WikidataClasses.aktiengesellschaft);
		classesToImport.add(WikidataClasses.country);
		classesToImport.add(WikidataClasses.publicCompany);
		classesToImport.add(WikidataClasses.publicLimitedCompany);
		classesToImport.add(WikidataClasses.aktiengesellschaft);
		classesToImport.add(WikidataClasses.city);
		classesToImport.add(WikidataClasses.bigCity);
		classesToImport.add(WikidataClasses.independentCity);
		classesToImport.add(WikidataClasses.basketballTeam);

		occupationsToImport.add(WikidataClasses.americalFootballPlayer);
		occupationsToImport.add(WikidataClasses.gridIronFootballPlayer);
		occupationsToImport.add(WikidataClasses.basketballPlayer);
		occupationsToImport.add(WikidataClasses.associationFootballer);

	}

	public ClassId getSet(WikidataParsedItem item, WikidataHierarchy hierarchy) {

		if (hierarchy.hasInstanceOf(item, WikidataClasses.human)) {
			// QualifiedValueList occupation =
			// item.getValues(WikidataClasses.occuptation);
			for (InstanceId occupationE : occupationsToImport) {
				if (hierarchy.hasOccupation(item, occupationE))
					return WikidataClasses.getSet(occupationE);
			}

			if (item.getValues(Stats.positionHeld) != null) {
				return WikidataClasses.getSet(WikidataClasses.politician);
			}
		}

		if (item.hasValues(Stats.stockExchange)) {
			QualifiedValueList stockExchange = item.getValues(Stats.stockExchange);
			if (stockExchange.hasValue(InstanceId.simple("Q13677")) || stockExchange.hasValue(InstanceId.simple("Q82059"))) {
				return WikidataClasses.wikidataTradedCompany;
			}
		}

		for (InstanceId clazz : classesToImport) {
			if (hierarchy.hasInstanceOf(item, clazz)) {
				return WikidataClasses.getSet(clazz);
			}
		}

		return null;
	}
}
