package de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.importers;

import java.io.FileWriter;
import java.io.IOException;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.commons.lang.mutable.MutableInt;
import org.json.JSONObject;

import au.com.bytecode.opencsv.CSVWriter;
import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedItem;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedProperty;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.WikidataImporter.ImporterFactory;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.WikidataImporter.ImporterParser;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.hierarchies.WikidataHierarchy;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.hierarchies.WikidataLabels;
import de.uni_mannheim.informatik.dws.expansion.wikidata.reader.WikidataJSONParser;

public class PropertyAnalyzerFactory implements ImporterFactory {

	private final FixedClassGetter fixedClassGetter = new FixedClassGetter();

	private final String resultsFile;

	private final AtomicInteger resultsCount = new AtomicInteger(0);

	public PropertyAnalyzerFactory(String resultsFile) {
		this.resultsFile = resultsFile;
	}

	@Override
	public ImporterParser get() {

		return new SetAnalyzer(fixedClassGetter, resultsFile + resultsCount.incrementAndGet() + ".csv");

	}

	public static class SetAnalyzer implements ImporterParser {

		private final ConcurrentHashMap<PropertyId, String> propertyNames;
		private final ConcurrentHashMap<ClassId, MutableInt> totalCount;

		private final ConcurrentHashMap<ClassId, ConcurrentHashMap<PropertyId, MutableInt>> totalPropertyCount;

		private final FixedClassGetter fixedClassGetter;

		private final String resultsFile;

		public SetAnalyzer(FixedClassGetter fixedClassGetter, String resultsFile) {
			propertyNames = new ConcurrentHashMap<>();
			totalCount = new ConcurrentHashMap<>();
			totalPropertyCount = new ConcurrentHashMap<>();
			this.resultsFile = resultsFile;
			this.fixedClassGetter = fixedClassGetter;
		}

		@Override
		public void add(ConcurrentHashMap<ValueId, QualifiedValueList> values, ConcurrentHashMap<InstanceId, Set<ClassId>> setMappings,
				ConcurrentHashMap<InstanceId, InstanceId> dbpediaMappings) {
			CSVWriter writer = null;
			try {
				writer = new CSVWriter(new FileWriter(resultsFile));

				int length = 4;

				{
					String[] headerLine = new String[4];
					headerLine[0] = "Class";
					headerLine[1] = "Property";
					headerLine[2] = "InstancesWithProperty";
					headerLine[3] = "TotalInstances";
					writer.writeNext(headerLine);
				}

				for (Entry<ClassId, MutableInt> setEntry : totalCount.entrySet()) {
					String[] line = new String[length];
					line[0] = setEntry.getKey().getClassId();
					line[3] = String.valueOf(setEntry.getValue().intValue());

					for (Entry<PropertyId, MutableInt> entry : totalPropertyCount.get(setEntry.getKey()).entrySet()) {
						line[1] = entry.getKey().getPropertyId();
						line[2] = String.valueOf(entry.getValue().intValue());
						writer.writeNext(line);
					}

				}

			} catch (IOException e) {
				e.printStackTrace();
			} finally {
				try {
					writer.close();
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}

		@Override
		public void run(WikidataParsedItem item, JSONObject object, WikidataJSONParser parser, Map<InstanceId, InstanceId> dbpediaMappings,
				WikidataHierarchy wikidataHierarchy, WikidataLabels labels) {

			ClassId setId = fixedClassGetter.getSet(item, wikidataHierarchy);

			if (setId != null) {
				CHMHelper.getFromCHM(totalCount, setId, () -> new MutableInt()).increment();
				for (PropertyId propertyId : item.getValuesMap().keySet()) {
					ConcurrentHashMap<PropertyId, MutableInt> submap = CHMHelper.getFromCHM(totalPropertyCount, setId, () -> new ConcurrentHashMap<>());
					CHMHelper.getFromCHM(submap, propertyId, () -> new MutableInt()).increment();
				}
			}

		}

		@Override
		public void runProperty(WikidataParsedProperty second, JSONObject object, WikidataJSONParser parser,
				ConcurrentHashMap<InstanceId, InstanceId> dbpediaMappings) {
			propertyNames.put(second.getPropertyId(), second.getLabel());

		}

	}

}
