package de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.importers;

import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import javax.inject.Inject;

import org.json.JSONObject;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.classes.ClassId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.MapSorter;
import de.mannheim.uni.expansion.utils.tree.Tree;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedItem;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedProperty;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.wikidata.WikidataClasses;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.WikidataImporter.ImporterFactory;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.WikidataImporter.ImporterParser;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.hierarchies.WikidataHierarchy;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.hierarchies.WikidataLabels;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.importers.FixedClassImporterFactory.FixedClassImporter;
import de.uni_mannheim.informatik.dws.expansion.wikidata.reader.WikidataJSONParser;

public class SetAnalyzerFactory implements ImporterFactory {

	private final FixedClassImporterFactory factory;

	@Inject
	public SetAnalyzerFactory(FixedClassImporterFactory factory) {
		this.factory = factory;
	}

	public SetAnalyzerFactory() {
		factory = new FixedClassImporterFactory();
	}

	@Override
	public ImporterParser get() {
		return new SetAnalyzer(factory.get());
	}

	public static class SetAnalyzer implements ImporterParser {

		private final FixedClassImporter fixedClassParser;

		private final ConcurrentHashMap<ClassId, Set<PropertyId>> setProperties;
		private final ConcurrentHashMap<ClassId, Set<InstanceId>> setInstances;

		private final ConcurrentHashMap<InstanceId, Set<ClassId>> instanceClasses;

		private final ConcurrentHashMap<InstanceId, Set<PropertyId>> instanceProperties;

		private final ConcurrentHashMap<PropertyId, String> propertyLabels;
		private final ConcurrentHashMap<InstanceId, String> instanceLabels;

		private final ConcurrentHashMap<ClassId, String> classLabels;

		private final ConcurrentHashMap<ClassId, ClassId> parentMap;

		private final ConcurrentHashMap<ClassId, Set<ClassId>> childrenMap;

		private final ConcurrentHashMap<ClassId, ClassId> setMapping;

		public SetAnalyzer(FixedClassImporter fixedClassParser) {
			setProperties = new ConcurrentHashMap<>();
			setInstances = new ConcurrentHashMap<>();

			instanceClasses = new ConcurrentHashMap<>();
			instanceProperties = new ConcurrentHashMap<>();
			propertyLabels = new ConcurrentHashMap<>();
			instanceLabels = new ConcurrentHashMap<>();
			classLabels = new ConcurrentHashMap<>();
			parentMap = new ConcurrentHashMap<>();
			childrenMap = new ConcurrentHashMap<>();
			setMapping = new ConcurrentHashMap<>();
			this.fixedClassParser = fixedClassParser;
		}

		@Override
		public void add(ConcurrentHashMap<ValueId, QualifiedValueList> values, ConcurrentHashMap<InstanceId, Set<ClassId>> setMappings,
				ConcurrentHashMap<InstanceId, InstanceId> dbpediaMappings) {

			Map<ClassId, Integer> setSizes = setInstances.entrySet().stream().collect(Collectors.toMap(Entry::getKey, (e) -> {
				return e.getValue().size();
			}));

			setSizes = setSizes.entrySet().stream().filter((e) -> {
				return e.getValue() > 15;
			}).collect(Collectors.toMap(Entry::getKey, Entry::getValue));

			boolean fff = false;
			if (fff) {
				LinkedHashMap<ClassId, Integer> sortedSetSizes = MapSorter.sortHashMapByValuesAscending(setSizes);
			}

			for (Iterator<Entry<ClassId, Integer>> it = setSizes.entrySet().iterator(); it.hasNext();) {

				Entry<ClassId, Integer> entry = it.next();

				ClassId parentSet = getParentSet(entry.getKey());
				if (parentSet != null) {
					parentMap.put(entry.getKey(), parentSet);
					CHMHelper.getFromCHM(childrenMap, parentSet, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(entry.getKey());
				}
			}

			Tree<ClassId> tree = getTree(setInstances.keySet(), parentMap, this.classLabels);
			System.out.println(tree.toString());
			int dsf = 123;

		}

		private Tree<ClassId> getTree(Set<ClassId> sets, Map<ClassId, ClassId> parents, Map<ClassId, String> labels) {

			Tree<ClassId> tree = new Tree<ClassId>();

			for (ClassId setId : sets) {
				if (parents.containsKey(setId)) {
					tree.addNode(setId, parents.get(setId), labels.getOrDefault(setId, setId.getClassId()));
				} else {
					tree.addRootNode(setId, labels.getOrDefault(setId, setId.getClassId()));
				}
			}

			return tree;
		}

		public ClassId getParentSet(ClassId childSet) {
			double maxContainment = 0.0;
			ClassId parentSet = null;

			Set<ClassId> candidates = instanceClasses.get(InstanceId.simple(childSet.getClassId()));
			if (candidates != null) {
				double total = setInstances.get(childSet).size();

				for (ClassId classId : candidates) {
					int count = 0;
					Set<InstanceId> set = setInstances.get(classId);
					for (InstanceId instanceId : setInstances.get(childSet)) {
						if (set.contains(instanceId)) {
							count++;
						}
					}
					double containment = count / total;
					if (containment > 0.95 && containment > maxContainment) {
						parentSet = classId;
						maxContainment = containment;
					}

				}
				return parentSet;
			}
			return null;
		}

		@Override
		public void run(WikidataParsedItem item, JSONObject object, WikidataJSONParser parser, Map<InstanceId, InstanceId> dbpediaMappings,
				WikidataHierarchy wikidataHierarchy, WikidataLabels labels) {

			InstanceId instanceId = parser.getInstanceId(object);

			Set<ClassId> sets = new HashSet<>();

			QualifiedValueList instanceOf = item.getValues(WikidataClasses.instanceOf);
			Set<PropertyId> propertySet = item.getValuesMap().keySet();
			instanceProperties.put(instanceId, propertySet);

			for (QualifiedValueHolder claim : instanceOf.getClaims()) {
				ClassId setId = new ClassId(((InstanceId) claim.getValue()).getInstanceId());
				sets.add(setId);

				CHMHelper.getFromCHM(setInstances, setId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).add(instanceId);
				CHMHelper.getFromCHM(setProperties, setId, () -> Collections.newSetFromMap(new ConcurrentHashMap<>())).addAll(propertySet);

			}

			instanceClasses.put(instanceId, sets);
			if (item.getLabel() != null) {
				classLabels.put(new ClassId(instanceId.getInstanceId()), item.getLabel());
				instanceLabels.put(instanceId, item.getLabel());
			} else {
				classLabels.put(new ClassId(instanceId.getInstanceId()), instanceId.getInstanceId());
				instanceLabels.put(instanceId, instanceId.getInstanceId());
			}

		}

		@Override
		public void runProperty(WikidataParsedProperty second, JSONObject object, WikidataJSONParser parser,
				ConcurrentHashMap<InstanceId, InstanceId> dbpediaMappings) {
			propertyLabels.put(second.getPropertyId(), second.getLabel());

		}

	}

}
