package de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import javax.inject.Inject;
import javax.inject.Named;

import org.json.JSONObject;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.PersistentParallel4;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedItem;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataType;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.wikidata.WikidataProperties;
import de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats.WikidataJSONEvaluator;
import de.uni_mannheim.informatik.dws.expansion.wikidata.reader.WikidataJSONParser;
import de.uni_mannheim.informatik.dws.expansion.wikidata.reader.WikidataJSONReader;

public class WikidataJSONAdapter {
	private final WikidataJSONReader reader;
	private final WikidataJSONParser parser;
	private final String dumpFile;
	private final WikidataJSONEvaluator evaluator;

	@Inject
	public WikidataJSONAdapter(WikidataJSONReader reader, WikidataJSONParser parser, @Named(WikidataProperties.WIKIDATA_DUMP) String dumpFile,
			WikidataJSONEvaluator evaluator) {
		this.reader = reader;
		this.parser = parser;
		this.dumpFile = dumpFile;
		this.evaluator = evaluator;
	}

	public void run() throws FileNotFoundException, IOException, InterruptedException {
		reader.read(dumpFile, this::parse);

		evaluator.print();

		PersistentParallel4.close();
	}

	ConcurrentHashMap<InstanceId, Map<PropertyId, QualifiedValueList>> map = new ConcurrentHashMap<>();

	public void parse(String parse) {
		JSONObject object = new JSONObject(parse);

		Pair<WikidataType, WikidataParsedObject> ite = parser.parse(object);

		parser.getEnglishDBPediaLink(object);

		evaluator.evaluate(ite);

		// return new DBpediaRepositoryNew(tables, columnMatchesList,
		// columnDataTypes)

	}

	public void evaluateCountry(WikidataParsedItem item) {
		System.out.println(item.getInstanceId().getReadableString());
	}
}
