package de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats;

import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedItem;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentInt;

public class CompanyStats extends Stats {

	private ConcurrentInt num = new ConcurrentInt(0);
	private ConcurrentInt hasCEO = new ConcurrentInt(0);
	private ConcurrentInt ceoHasStartTime = new ConcurrentInt(0);
	private ConcurrentInt ceoHasEndTime = new ConcurrentInt(0);
	private ConcurrentInt hasInceptionDate = new ConcurrentInt(0);
	private ConcurrentInt nrOfEmployees = new ConcurrentInt(0);
	private ConcurrentInt nrOfEmployeesHasPointInTime = new ConcurrentInt(0);
	private ConcurrentInt hasTotalRevenue = new ConcurrentInt(0);
	private ConcurrentInt revenueHasPointInTime = new ConcurrentInt(0);

	void evaluate(WikidataParsedItem item) {

		num.inc();

		QualifiedValueList val = item.getValues(Stats.chiefExecutiveOfficers);
		if (val != null) {
			val.getClaims().forEach((e) -> {
				hasCEO.inc();

				if (e.hasQualifier(Stats.startTimeQualifier)) {
					ceoHasStartTime.inc();

					if (e.hasQualifier(Stats.endTimeQualifier)) {
						ceoHasEndTime.inc();
					}

					addToSelection(Stats.chiefExecutiveOfficers, item.getInstanceId(), e);
				}

			});

		}

		val = item.getValues(Stats.inception);
		if (val != null) {
			hasInceptionDate.inc();
		}

		val = item.getValues(Stats.nrOfEmployees);
		if (val != null) {
			val.getClaims().forEach((e) -> {
				nrOfEmployees.inc();

				if (e.hasQualifier(Stats.pointInTime)) {
					nrOfEmployeesHasPointInTime.inc();

					addToSelection(Stats.nrOfEmployees, item.getInstanceId(), e);
				}
			});

		}

		val = item.getValues(Stats.totalRevenue);
		if (val != null) {
			val.getClaims().forEach((e) -> {
				hasTotalRevenue.inc();

				if (e.hasQualifier(Stats.pointInTime)) {
					revenueHasPointInTime.inc();

					addToSelection(Stats.totalRevenue, item.getInstanceId(), e);
				}
			});
		}

	}

	@Override
	public String toString() {
		return getClass().getName() + "\nnum;" + num + "\nhasCEO;" + hasCEO + "\nceoHasStartTime;" + ceoHasStartTime + "\nceoHasEndTime;" + ceoHasEndTime
				+ "\nhasInceptionDate;" + hasInceptionDate + "\nnrOfEmployees;" + nrOfEmployees + "\nnrOfEmployeesHasPointInTime;"
				+ nrOfEmployeesHasPointInTime;
	}

}
