package de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats;

import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedItem;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentInt;

class CountryStats extends Stats {

	private ConcurrentInt num = new ConcurrentInt(0);

	// private ConcurrentInt currencyWithChange = new ConcurrentInt(0);
	private ConcurrentInt memberShipStatements = new ConcurrentInt(0);
	private ConcurrentInt memberValuesWithDates = new ConcurrentInt(0);

	// private ConcurrentInt gdpWithDates = new ConcurrentInt(0);
	// private ConcurrentInt gdpWithoutDate = new ConcurrentInt(0);
	// private ConcurrentInt noGdp = new ConcurrentInt(0);
	// private ConcurrentInt multipleGDPWithDates = new ConcurrentInt(0);
	//
	// private ConcurrentInt gdpPerCapitasWithDates = new ConcurrentInt(0);
	// private ConcurrentInt gdpPerCapitasWithoutDate = new ConcurrentInt(0);
	// private ConcurrentInt noGdpPerCapitas = new ConcurrentInt(0);
	// private ConcurrentInt multipleGDPPerCapitasWithDates = new
	// ConcurrentInt(0);
	//
	// private ConcurrentInt noPopulation = new ConcurrentInt(0);
	// private ConcurrentInt populationWithDate = new ConcurrentInt(0);
	// private ConcurrentInt populationWithoutDate = new ConcurrentInt(0);
	// private ConcurrentInt multiplePopulationwithDates = new ConcurrentInt(0);

	private ConcurrentInt gdpPerCapitasStatements = new ConcurrentInt(0);

	private ConcurrentInt gdpPerCapitasStatementsWithDates = new ConcurrentInt(0);

	private ConcurrentInt gdpStatements = new ConcurrentInt(0);

	private ConcurrentInt gdpStatementsWithDates = new ConcurrentInt(0);

	private ConcurrentInt populationStatements = new ConcurrentInt(0);

	private ConcurrentInt populationStatementsWithDate = new ConcurrentInt(0);

	private ConcurrentInt noArea = new ConcurrentInt(0);
	private ConcurrentInt areaCount = new ConcurrentInt(0);

	private ConcurrentInt noContinent = new ConcurrentInt(0);
	private ConcurrentInt continentCount = new ConcurrentInt(0);

	private ConcurrentInt headOfGovernmentsRef = new ConcurrentInt(0);
	private ConcurrentInt noHeadOfGovernmentsRef = new ConcurrentInt(0);

	private ConcurrentInt headOfStatesRef = new ConcurrentInt(0);
	private ConcurrentInt noHeadOfStatesRef = new ConcurrentInt(0);

	private ConcurrentInt headOfGovernmentStatements = new ConcurrentInt(0);

	private ConcurrentInt headOfGovernemntWithStartDate = new ConcurrentInt(0);

	private ConcurrentInt headOfGovernmentWithStartAndEndDate = new ConcurrentInt(0);

	private ConcurrentInt headOfStateStatements = new ConcurrentInt(0);

	private ConcurrentInt headOfStateWithStartDate = new ConcurrentInt(0);

	private ConcurrentInt headOfStateWithStartAndEndDate = new ConcurrentInt(0);

	void evaluate(WikidataParsedItem item) {
		PropertyId memberOf = new PropertyId("P463");

		PropertyId area = new PropertyId("P2046");
		PropertyId continent = new PropertyId("P30");
		PropertyId gdp = new PropertyId("P2131");
		PropertyId gdpPerCapita = new PropertyId("P2299");

		PropertyId headOfStateObject = new PropertyId("P1906");
		PropertyId headOfGovernmentObject = new PropertyId("P1313");

		num.inc();

		item.getValues(memberOf).getClaims().forEach((e) -> {
			memberShipStatements.inc();

			if (e.hasQualifier(Stats.startTimeQualifier)) {
				memberValuesWithDates.inc();

				addToSelection(memberOf, item.getInstanceId(), e);
			}

		});

		item.getValues(Stats.population).getClaims().forEach((e) -> {
			populationStatements.inc();
			if (e.hasQualifier(Stats.pointInTime)) {
				populationStatementsWithDate.inc();

				addToSelection(Stats.population, item.getInstanceId(), e);
			}
		});

		if (!item.getValues(area).isEmpty()) {

			areaCount.inc();
		} else {
			noArea.inc();
		}

		QualifiedValueList continents = item.getValues(continent);
		if (continents != null) {
			continentCount.inc();
		} else {
			noContinent.inc();
		}

		item.getValues(gdp).getClaims().forEach((e) -> {
			gdpStatements.inc();

			if (e.hasQualifier(Stats.pointInTime)) {
				gdpStatementsWithDates.inc();

				addToSelection(gdp, item.getInstanceId(), e);
			}

		});

		// else {
		// noGdp.inc();
		// }

		item.getValues(gdpPerCapita).getClaims().forEach((e) ->

		{
			gdpPerCapitasStatements.inc();

			if (e.hasQualifier(Stats.pointInTime)) {
				gdpPerCapitasStatementsWithDates.inc();

				addToSelection(gdpPerCapita, item.getInstanceId(), e);
			}

		});

		//
		// QualifiedValueList gdpPerCapitas = item.getValues(gdpPerCapita);
		// if (gdpPerCapitas != null) {
		//
		// if (gdpPerCapitas.getClaims().stream().filter((e) -> {
		// return e.hasQualifier(pointInTimeQualifier);
		// }).findFirst().isPresent()) {
		// gdpPerCapitasWithDates.inc();
		// if (gdpPerCapitas.getClaims().size() > 1) {
		// multipleGDPPerCapitasWithDates.inc();
		// }
		// } else {
		// gdpPerCapitasWithoutDate.inc();
		// }
		// } else {
		// noGdpPerCapitas.inc();
		// }

		if (!item.getValues(headOfGovernmentObject).isEmpty())

		{
			headOfGovernmentsRef.inc();
		} else

		{
			noHeadOfGovernmentsRef.inc();
		}

		if (!item.getValues(headOfStateObject).isEmpty())

		{
			headOfStatesRef.inc();
		} else

		{
			noHeadOfStatesRef.inc();
		}

		item.getValues(Stats.headOfGovernemnt).getClaims().forEach((e) ->

		{
			headOfGovernmentStatements.inc();

			if (e.hasQualifier(Stats.startTimeQualifier)) {
				headOfGovernemntWithStartDate.inc();
				if (e.hasQualifier(Stats.endTimeQualifier)) {
					headOfGovernmentWithStartAndEndDate.inc();

					addToSelection(Stats.headOfGovernemnt, item.getInstanceId(), e);
				}
			}
		});

		item.getValues(Stats.headOfState).getClaims().forEach((e) ->

		{
			headOfStateStatements.inc();

			if (e.hasQualifier(Stats.startTimeQualifier)) {
				headOfStateWithStartDate.inc();
				if (e.hasQualifier(Stats.endTimeQualifier)) {
					headOfStateWithStartAndEndDate.inc();

					addToSelection(Stats.headOfState, item.getInstanceId(), e);
				}
			}
		});

	}

	@Override
	public String toString() {
		return "CountryStats [num=" + num + ", memberShipStatements=" + memberShipStatements + ", memberValuesWithDates=" + memberValuesWithDates
				+ ", gdpPerCapitasStatements=" + gdpPerCapitasStatements + ", gdpPerCapitasStatementsWithDates=" + gdpPerCapitasStatementsWithDates
				+ ", gdpStatements=" + gdpStatements + ", gdpStatementsWithDates=" + gdpStatementsWithDates + ", populationStatements=" + populationStatements
				+ ", populationStatementsWithDate=" + populationStatementsWithDate + ", noArea=" + noArea + ", areaCount=" + areaCount + ", noContinent="
				+ noContinent + ", continentCount=" + continentCount + ", headOfGovernmentsRef=" + headOfGovernmentsRef + ", noHeadOfGovernmentsRef="
				+ noHeadOfGovernmentsRef + ", headOfStatesRef=" + headOfStatesRef + ", noHeadOfStatesRef=" + noHeadOfStatesRef + ", headOfGovernmentStatements="
				+ headOfGovernmentStatements + ", headOfGovernemntWithStartDate=" + headOfGovernemntWithStartDate + ", headOfGovernmentWithStartAndEndDate="
				+ headOfGovernmentWithStartAndEndDate + ", headOfStateStatements=" + headOfStateStatements + ", headOfStateWithStartDate="
				+ headOfStateWithStartDate + ", headOfStateWithStartAndEndDate=" + headOfStateWithStartAndEndDate + "]";
	}

}