package de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats;

import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedItem;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentInt;

public class OfficeHolderStats extends Stats {

	private ConcurrentInt num = new ConcurrentInt(0);
	private ConcurrentInt positionWithStartTime = new ConcurrentInt(0);
	private ConcurrentInt positionWithEndTime = new ConcurrentInt(0);
	private ConcurrentInt positionReplaces = new ConcurrentInt(0);
	private ConcurrentInt positionReplacedBy = new ConcurrentInt(0);
	private ConcurrentInt positionSeries = new ConcurrentInt(0);

	void evaluate(WikidataParsedItem item) {

		item.getValues(Stats.positionHeld).getClaims().forEach((e) -> {

			boolean stop = false;
			num.inc();

			if (e.hasQualifier(Stats.startTimeQualifier)) {
				positionWithStartTime.inc();

				if (e.hasQualifier(Stats.endTimeQualifier)) {
					positionWithEndTime.inc();

					stop = true;

					addToSelection(Stats.positionHeld, item.getInstanceId(), e);

				}
			}

			if (e.hasQualifier(Stats.officeReplaces)) {
				positionReplaces.inc();

				if (!stop) {
					stop = true;
					addToSelection(Stats.positionHeld, item.getInstanceId(), e);
				}
			}

			if (e.hasQualifier(Stats.officeReplacedBy)) {
				positionReplacedBy.inc();

				if (!stop) {
					stop = true;
					addToSelection(Stats.positionHeld, item.getInstanceId(), e);
				}
			}

			if (e.hasQualifier(Stats.seriedOrdinal)) {
				positionSeries.inc();

				if (!stop) {
					stop = true;
					addToSelection(Stats.positionHeld, item.getInstanceId(), e);
				}
			}

		});

	}

	@Override
	public String toString() {
		return getClass().getName() + "\nnum;" + num + "\npositionWithStartTime;" + positionWithStartTime + "\npositionWithEndTime;" + positionWithEndTime
				+ "\npositionReplaces;" + positionReplaces + "\npositionReplacedBy;" + positionReplacedBy + "\npositionSeries;" + positionSeries;
	}

}
