package de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats;

import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedItem;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.profiling.accumulation.accumulators.ConcurrentInt;

public class SoccerStats extends Stats {

	void evaluate(WikidataParsedItem item) {

		num.inc();

		QualifiedValueList val = item.getValues(Stats.teams);
		if (val != null) {

			val.getClaims().forEach((e) -> {
				boolean stop = false;
				hasTeam.inc();
				if (e.hasQualifier(Stats.startTimeQualifier)) {
					teamsHaveStartDate.inc();

					if (e.hasQualifier(Stats.endTimeQualifier)) {
						teamsHaveEndDate.inc();

						stop = true;
						addToSelection(Stats.teams, item.getInstanceId(), e);
					}
				}

				if (e.hasQualifier(Stats.numberOfGoalsQualifier)) {
					teamsHaveNumberOfGoals.inc();

					stop = true;
					if (!stop) {
						addToSelection(Stats.teams, item.getInstanceId(), e);
					}
				}

				if (e.hasQualifier(Stats.numberOfMatchesPlayed)) {
					teamHaveMatchesPlayed.inc();

					if (!stop) {
						addToSelection(Stats.teams, item.getInstanceId(), e);
					}
				}
			});

		}

		val = item.getValues(Stats.awardReceived);
		if (val != null) {
			val.getClaims().forEach((e) -> {
				hasAwards.inc();
				if (e.hasQualifier(Stats.pointInTime)) {
					awardsHasPointinTime.inc();

					addToSelection(Stats.awardReceived, item.getInstanceId(), e);

				}
			});
		}

		// val = item.getValues(Stats.citizenShip);
		// if (val != null) {
		// country.inc();
		// }

		val = item.getValues(Stats.dateOfBirth);
		if (val != null) {
			hasBirthDate.inc();
		}
		val = item.getValues(Stats.placeOfBirth);
		if (val != null) {
			hasBirthPlace.inc();
		}

		throw new IllegalStateException();

	}

	private ConcurrentInt num = new ConcurrentInt(0);
	private ConcurrentInt hasBirthDate = new ConcurrentInt(0);
	private ConcurrentInt hasBirthPlace = new ConcurrentInt(0);
	private ConcurrentInt country = new ConcurrentInt(0);

	private ConcurrentInt hasTeam = new ConcurrentInt(0);
	private ConcurrentInt teamsHaveStartDate = new ConcurrentInt(0);
	private ConcurrentInt teamsHaveEndDate = new ConcurrentInt(0);
	private ConcurrentInt teamsHaveNumberOfGoals = new ConcurrentInt(0);

	private ConcurrentInt hasAwards = new ConcurrentInt(0);
	private ConcurrentInt awardsHasPointinTime = new ConcurrentInt(0);

	private ConcurrentInt teamHaveMatchesPlayed = new ConcurrentInt(0);

	@Override
	public String toString() {
		return getClass().getName() + "\nnum;" + num + "\nhasBirthDate;" + hasBirthDate + "\nhasBirthPlace;" + hasBirthPlace + "\ncountry;" + country
				+ "\nhasTeam;" + hasTeam + "\nteamsHaveStartDate;" + teamsHaveStartDate + "\nteamsHaveEndDate;" + teamsHaveEndDate + "\nteamsHaveNumberOfGoals;"
				+ teamsHaveNumberOfGoals + "\nhasAwards;" + hasAwards + "\nawardsHasPointinTime;" + awardsHasPointinTime + "\nteamHaveMatchesPlayed;"
				+ teamHaveMatchesPlayed;
	}

}
