package de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.stream.Collectors;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableMap.Builder;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.PropertiesExpansionData;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;

public abstract class Stats {

	private final ConcurrentHashMap<PropertyId, Queue<Pair<ValueId, QualifiedValueHolder>>> selectionMap2 = new ConcurrentHashMap<>();

	private final ConcurrentHashMap<PropertyId, Queue<Pair<ValueId, QualifiedValueHolder>>> selectionMap = new ConcurrentHashMap<>();
	private final ConcurrentHashMap<PropertyId, Queue<Pair<ValueId, QualifiedValueHolder>>> secondarySelectionMap = new ConcurrentHashMap<>();
	private final int maxSelection = 15;

	public Collection<Pair<ValueId, QualifiedValueHolder>> getSelection() {

		return Collections.unmodifiableCollection(selectionMap.entrySet().stream().map((e) -> {

			List<Pair<ValueId, QualifiedValueHolder>> list = new ArrayList<>(e.getValue());

			Collections.shuffle(list);

			return list.subList(0, Math.min(list.size(), maxSelection));

		}).flatMap(listContainer -> listContainer.stream()).collect(Collectors.toList()));
	}

	public Collection<Pair<ValueId, QualifiedValueHolder>> getSelectionOld() {

		return Collections.unmodifiableCollection(selectionMap.entrySet().stream().map((e) -> {
			if (e.getValue().size() < maxSelection) {
				List<Pair<ValueId, QualifiedValueHolder>> list = new ArrayList<>(e.getValue());

				for (int i = list.size(); i < maxSelection; i++) {
					list.add(secondarySelectionMap.get(e.getKey()).poll());
				}
				return list;
			} else {
				return e.getValue();
			}
		}).flatMap(listContainer -> listContainer.stream()).collect(Collectors.toList()));
	}

	protected void addToSelection(PropertyId propertyId, InstanceId instanceId, QualifiedValueHolder qvh) {
		Pair<ValueId, QualifiedValueHolder> pair = new Pair<ValueId, QualifiedValueHolder>(new ValueId(instanceId, propertyId), qvh);
		Queue<Pair<ValueId, QualifiedValueHolder>> selection = CHMHelper.getFromCHM(selectionMap, propertyId, () -> new ConcurrentLinkedQueue<>());

		selection.add(pair);

	}

	// protected void addToSelection(PropertyId propertyId, InstanceId
	// instanceId, QualifiedValueHolder qvh) {
	// Pair<ValueId, QualifiedValueHolder> pair = new Pair<ValueId,
	// QualifiedValueHolder>(new ValueId(instanceId, propertyId), qvh);
	// Queue<Pair<ValueId, QualifiedValueHolder>> selection =
	// CHMHelper.getFromCHM(selectionMap, propertyId, () -> new
	// ConcurrentLinkedQueue<>());
	//
	// Queue<Pair<ValueId, QualifiedValueHolder>> secondarySelection =
	// CHMHelper.getFromCHM(secondarySelectionMap, propertyId,
	// () -> new ConcurrentLinkedQueue<>());
	//
	// if (selection.size() < maxSelection) {
	// if
	// (ThreadLocalRandom.current().nextInt(ThreadLocalRandom.current().nextInt(5000)
	// + 1) == 1) {
	// selection.poll();
	// selection.add(pair);
	// } else if (secondarySelection.size() + selection.size() < maxSelection) {
	// secondarySelection.add(pair);
	// }
	//
	// }
	//
	// else if
	// (ThreadLocalRandom.current().nextInt(ThreadLocalRandom.current().nextInt(5000)
	// + 1) == 1) {
	// selection.poll();
	// selection.add(pair);
	// }
	// }

	public static final PropertyId headOfGovernemnt = new PropertyId("P6", "HeadOfGovernment"); // http://dbpedia.org/ontology/leaderTitle
	public static final PropertyId headOfState = new PropertyId("P35", "HeadOfState"); // http://dbpedia.org/ontology/leaderTitle
	public static final PropertyId headOfGovernmentOffice = new PropertyId("P1313", "HeadOfGovernmentOffice");
	public static final PropertyId headOfStateOffice = new PropertyId("P1906", "HeadOfStateOffice");

	public static final PropertyId population = new PropertyId("P1082", "Population"); // http://dbpedia.org/ontology/populationTotal
	public static final PropertyId populationDensity = new PropertyId("Q22856", "PopulationDensity"); // http://dbpedia.org/ontology/populationTotal

	public static final PropertyId nominalGDP = new PropertyId("P2131", "NominalGDP"); // http://dbpedia.org/ontology/populationTotal
	public static final PropertyId nominalGDPperCapita = new PropertyId("P2132", "NominalGDPPerCapita"); // http://dbpedia.org/ontology/populationTotal

	public static final PropertyId isoAlpha3Codes = new PropertyId("P298", "ISO-3166-Alpha-3-Code"); // http://dbpedia.org/ontology/populationTotal

	// public static final PropertyId mayor = new PropertyId("P6", "Mayor");

	public static final PropertyId of = new PropertyId("P642", "Of");
	public static final PropertyId appointedBy = new PropertyId("P748", "AppointedBy");
	public static final PropertyId appliesToTerritorialJurisdiction = new PropertyId("P1001", "Juristdiction");

	public static final PropertyId country = new PropertyId("P17", "Country");

	public static final PropertyId dateOfBirth = new PropertyId("P569", "DateOfBirth"); // http://dbpedia.org/ontology/birthDate
	public static final PropertyId placeOfBirth = new PropertyId("P19", "PlaceOfBirth"); // http://dbpedia.org/ontology/birthPlace
	public static final PropertyId teams = new PropertyId("P54", "Team"); // http://dbpedia.org/ontology/team

	public static final PropertyId startTimeQualifier = new PropertyId("P580", "StartTimeQualifier");
	public static final PropertyId endTimeQualifier = new PropertyId("P582", "EndTimeQualifier");

	public static final PropertyId positionHeld = new PropertyId("P39", "PositionHeld"); // http://dbpedia.org/ontology/office

	// SOCCERS
	public static final PropertyId awardReceived = new PropertyId("P166", "AwardsReceived");
	public static final PropertyId numberOfGoalsQualifier = new PropertyId("P1351");
	public static final PropertyId pointInTime = new PropertyId("P585", "PointInTime");
	public static final PropertyId numberOfMatchesPlayed = new PropertyId("P1350");

	// NBA

	// PUBLIC COMPANY
	public static final PropertyId chiefExecutiveOfficers = new PropertyId("P169");
	public static final PropertyId inception = new PropertyId("P571", "Inception");
	public static final PropertyId nrOfEmployees = new PropertyId("P1128", "NumberOfEmployees"); // http://dbpedia.org/ontology/numberOfEmployees
	public static final PropertyId totalRevenue = new PropertyId("P2139", "TotalRevenue"); // http://dbpedia.org/ontology/revenue
	public static final PropertyId totalAssets = new PropertyId("P2403", "TotalAssets");
	public static final PropertyId totalEquity = new PropertyId("P2137", "TotalEquity");
	public static final PropertyId netIncome = new PropertyId("P2295", "NetIncome");
	public static final PropertyId ebit = new PropertyId("EBIT", "EBIT");

	// OFFICE HELD
	public static final PropertyId officeReplaces = new PropertyId("P1365");

	public static final PropertyId officeReplacedBy = new PropertyId("P1366");
	public static final PropertyId seriedOrdinal = new PropertyId("P1545");

	public static final PropertyId party = new PropertyId("P102", "Party"); // http://dbpedia.org/ontology/party

	public static final PropertyId sportNumber = new PropertyId("P1618", "SportNumber");

	public static final PropertyId label = PropertiesExpansionData.LABEL_PROPERTY;
	public static final PropertyId occupation = new PropertyId("P106", "Occupation");

	public static final PropertyId stockExchange = new PropertyId("P414", "StockExchange");

	public static final PropertyId instanceOf = new PropertyId("P31", "InstanceOf");

	public static final PropertyId sport = new PropertyId("P641", "Sport");
	public static final PropertyId gender = new PropertyId("P21", "Gender");
	public static final PropertyId countryOfCitizenship = new PropertyId("P27", "CountryOfCitizenship");
	public static final PropertyId positionPlayed = new PropertyId("P413", "PositionPlayed");
	public static final PropertyId givenName = new PropertyId("P735", "GivenName");
	public static final PropertyId height = new PropertyId("P2048", "Height");

	public static final PropertyId dateOfDeath = new PropertyId("P570", "DateOfDeath");
	public static final PropertyId mass = new PropertyId("P2067", "Mass");

	public static final PropertyId coordinateLocation = new PropertyId("P625", "CoordinateLocation");
	public static final PropertyId locatedIn = new PropertyId("P131", "LocatedIn");

	public static final PropertyId elevation = new PropertyId("P2044", "Elevation");
	public static final PropertyId officialWebsite = new PropertyId("P856", "OfficialWebsite");
	public static final PropertyId FIPS55M3 = new PropertyId("P774", "FIPS55M3");
	public static final PropertyId area = new PropertyId("P2046", "Area");
	public static final PropertyId postalCode = new PropertyId("P281", "PostalCode");

	public static final PropertyId sisterCity = new PropertyId("P190", "SisterCity");
	public static final PropertyId localDialingCode = new PropertyId("P473", "LocalDialingCode");
	public static final PropertyId capitalOf = new PropertyId("P1376", "CapitalOf");
	public static final PropertyId sharesBorderWith = new PropertyId("P47", "SharesBorderWith");

	public static final PropertyId headquartersLocation = new PropertyId("P159", "headquartersLocation");
	public static final PropertyId legalForm = new PropertyId("P1454", "LegalForm");
	public static final PropertyId industry = new PropertyId("P452", "Industry");
	public static final PropertyId legalEntityID = new PropertyId("P1278", "LegalEntityID");

	public static final PropertyId founder = new PropertyId("P112", "Founder");
	public static final PropertyId locationOfFormation = new PropertyId("P740", "LocationOfFormation");

	public static final PropertyId educatedAt = new PropertyId("P69", "EducatedAt");

	public static final PropertyId draftedBy = new PropertyId("P647", "DraftedBy");

	public static final PropertyId familyName = new PropertyId("P734", "FamilyName");

	public static final PropertyId participantOf = new PropertyId("P1344", "ParticipantOf");

	public static final PropertyId placeOfDeath = new PropertyId("P20", "PlaceOfDeath");

	public static final PropertyId locatedInTimeZone = new PropertyId("P421", "LocatedInTimeZone");

	public static final PropertyId currency = new PropertyId("P38", "Currency");
	public static final PropertyId capital = new PropertyId("P36", "Capital");
	public static final PropertyId continent = new PropertyId("P30", "Continent ");
	public static final PropertyId tld = new PropertyId("P78", "TLD");

	public static final PropertyId isoNumericCode = new PropertyId("P214", "ISO-3166-1-numeric-code");
	public static final PropertyId countryCallingCode = new PropertyId("P474", "CountryCallingCode");
	public static final PropertyId anthem = new PropertyId("P85", "Anthem");
	public static final PropertyId fips10_4 = new PropertyId("P901", "FIPS10-4");
	public static final PropertyId memberOf = new PropertyId("P463", "memberOf");
	public static final PropertyId electricalPlugType = new PropertyId("P2853", "ElectricalPlugType");
	public static final PropertyId drivingSide = new PropertyId("P1622", "DrivingSide");
	public static final PropertyId iocCountryCode = new PropertyId("P984", "IocCountryCode");
	public static final PropertyId mainsVoltage = new PropertyId("P2884", "MainsVoltage");

	public static final Map<PropertyId, PropertyId> propertyMap;

	static {
		try {
			Builder<PropertyId, PropertyId> buidler = ImmutableMap.builder();
			Field[] declaredFields = Stats.class.getDeclaredFields();
			for (Field field : declaredFields) {
				if (java.lang.reflect.Modifier.isStatic(field.getModifiers()) && field.getType().equals(label.getClass())) {
					PropertyId propertyId;
					field.setAccessible(true);
					propertyId = (PropertyId) field.get(null);

					buidler.put(propertyId, propertyId);
				}
			}

			propertyMap = buidler.build();

		} catch (IllegalArgumentException e) {
			e.printStackTrace();
			throw new IllegalStateException("Could not init class.", e);
		} catch (IllegalAccessException e) {
			throw new IllegalStateException("Could not init class.", e);
		} catch (SecurityException e) {
			throw new IllegalStateException("Could not init class.", e);
		}

	}

	public static PropertyId getExistingPropertyId(PropertyId propertyId) {
		if (!propertyMap.containsKey(propertyId)) {
			// System.out.println(propertyId.toString());
			return null;
		} else {
			PropertyId newProperty = propertyMap.get(propertyId);
			return newProperty;
		}
	}

	public static PropertyId getPropertyId(PropertyId propertyId) {
		if (!propertyMap.containsKey(propertyId)) {
			// System.out.println(propertyId.toString());
			return propertyId;
		} else {
			return propertyMap.get(propertyId);
		}
	}
}
