package de.uni_mannheim.informatik.dws.expansion.wikidata.evaluation.stats;

import javax.inject.Inject;

import de.mannheim.uni.expansion.data.ontology.model.ValueId;
import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedItem;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataParsedObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.types.WikidataType;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueHolder;
import de.uni_mannheim.dws.expansion.data.ontology.model.valueholders.QualifiedValueList;
import de.uni_mannheim.informatik.dws.expansion.wikidata.WikidataClasses;
import de.uni_mannheim.informatik.dws.expansion.wikidata.reader.WikidataNameRepository;

public class WikidataJSONEvaluator {

	@Inject
	public WikidataJSONEvaluator() {
	}

	final WikidataNameRepository rep = new WikidataNameRepository();
	final CountryStats countryStats = new CountryStats();
	final NFLStats nflStats = new NFLStats();
	final NBAStats nbaStats = new NBAStats();
	final SoccerStats soccerStats = new SoccerStats();
	final CompanyStats companyStats = new CompanyStats();
	final OfficeHolderStats officeHolderStats = new OfficeHolderStats();
	final CityStats cityStats = new CityStats();

	public void evaluate(Pair<WikidataType, WikidataParsedObject> ite) {

		rep.add(ite);
		if (ite.getFirst() == WikidataType.ITEM) {
			WikidataParsedItem item = (WikidataParsedItem) ite.getSecond();

			QualifiedValueList list = item.getValues(WikidataClasses.instanceOf);
			if (list != null) {
				if (list.hasValue(WikidataClasses.country)) {
					countryStats.evaluate(item);
				} else if (list.hasValue(WikidataClasses.human)) {

					QualifiedValueList occupationList = item.getValues(WikidataClasses.occuptation);
					if (occupationList != null) {
						if (occupationList.hasValue(WikidataClasses.americalFootballPlayer)) {
							nflStats.evaluate(item);
						}

						if (occupationList.hasValue(WikidataClasses.basketballPlayer)) {
							nbaStats.evaluate(item);
						}
						if (occupationList.hasValue(WikidataClasses.associationFootballer)) {
							soccerStats.evaluate(item);
						}

						if (occupationList.hasValue(WikidataClasses.politician) && item.getValues(Stats.positionHeld) != null) {
							officeHolderStats.evaluate(item);
						}
					}

				} else if (list.hasValue(WikidataClasses.publicCompany) || list.hasValue(WikidataClasses.publicLimitedCompany)
						|| list.hasValue(WikidataClasses.aktiengesellschaft)) {
					companyStats.evaluate(item);
				} else if (list.hasValue(WikidataClasses.city) || list.hasValue(WikidataClasses.bigCity)) {
					cityStats.evaluate(item);
				}

			}
		}

	}

	public void print() {

		System.out.println(countryStats.toString());
		System.out.println(nflStats.toString());
		System.out.println(nbaStats.toString());
		System.out.println(soccerStats.toString());
		System.out.println(companyStats.toString());
		System.out.println(officeHolderStats.toString());
		System.out.println(cityStats.toString());

		StringBuilder sb = new StringBuilder();
		sb.append(getStatsSelectionsAsString(countryStats) + "\n");
		sb.append(getStatsSelectionsAsString(nflStats) + "\n");
		sb.append(getStatsSelectionsAsString(nbaStats) + "\n");
		sb.append(getStatsSelectionsAsString(soccerStats) + "\n");
		sb.append(getStatsSelectionsAsString(companyStats) + "\n");
		sb.append(getStatsSelectionsAsString(officeHolderStats) + "\n");
		sb.append(getStatsSelectionsAsString(cityStats) + "\n");

		System.out.println(sb.toString());
	}

	public String getStatsSelectionsAsString(Stats stats) {
		throw new IllegalStateException();
		// StringBuilder sb = new StringBuilder();
		// stats.getSelection().forEach((pair) -> {
		//
		// QualifiedValueHolder e = pair.getSecond();
		// String t = rep.getOr(pair.getFirst().getInstanceId(),
		// pair.getFirst().getInstanceId().getReadableString()) + ";"
		// + rep.getOr(pair.getFirst().getPropertyId(),
		// pair.getFirst().getPropertyId().getAsString()) + ";" +
		// getReferenceName(e.getValue()) + ";"
		// + e.getQualifiers().stream().map((q) -> {
		// return rep.getOr(q.getQualifier(), q.getQualifier().getAsString()) +
		// ";"
		// +
		// q.getQualifierValues().stream().map(this::getReferenceName).collect(Collectors.toList()).toString();
		// }).collect(Collectors.joining(";")) + "\n";
		// sb.append(t);
		// });
		//
		// return sb.toString();
	}

	public void appendToSB(Pair<ValueId, QualifiedValueHolder> pair, StringBuilder sb) {

	}

	private String getReferenceName(AbstractObject e) {
		String s = "";
		if (e == null) {
			return s;
		}
		if (e instanceof InstanceId) {
			s += rep.getOr(((InstanceId) e), ((InstanceId) e).getInstanceId());

		} else {
			s += e;
		}
		// s += "(" + e.getType().toString() + ") ";

		return s;
	}

}
