package de.uni_mannheim.informatik.dws.expansion.wikidata.parser;

import javax.inject.Inject;

import org.json.JSONObject;

import de.mannheim.uni.expansion.data.ontology.model.instances.InstanceId;
import de.mannheim.uni.expansion.data.ontology.model.properties.PropertyId;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.informatik.dws.expansion.wikidata.adapter.hierarchies.WikidataLabels;

public class WikidataJSONParserEntity {

	private WikidataLabels labels;

	@Inject
	public WikidataJSONParserEntity() {

	}

	public AbstractObject parse(JSONObject jsonObject) {

		if (!"wikibase-entityid".equals(jsonObject.get("type"))) {
			throw new IllegalStateException("This is not a JSON Object of the type time.");
		}

		JSONObject value = jsonObject.getJSONObject("value");

		String entityType = value.getString("entity-type");
		int numericId = value.getInt("numeric-id");

		if (entityType.equals("item")) {
			String id = "Q" + numericId;
			String readableString = id;
			if (labels != null) {
				readableString = labels.getLabel(new InstanceId(id, id));
			}

			return new InstanceId(id, readableString);
		} else if (entityType.equals("property")) {
			return new PropertyId("P" + numericId);
		}

		System.out.println("DIFFERENT ENTITY TYPE: " + entityType);
		return null;
	}

	public void setLabels(WikidataLabels labels) {
		this.labels = labels;

	}
}
