package de.uni_mannheim.informatik.dws.expansion.wikidata.parser;

import javax.inject.Inject;

import org.json.JSONObject;

import de.uni_mannheim.dws.expansion.data.ontology.model.objects.AbstractObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject;
import de.uni_mannheim.dws.expansion.data.ontology.model.objects.TimeObject.Precision;

public class WikidataJSONParserTime {
	@Inject
	public WikidataJSONParserTime() {

	}

	// public DateTimeFormatter wikiDataFormatter = new
	// DateTimeFormatterBuilder().appendValue(YEAR, 4, 12,
	// SignStyle.ALWAYS).appendLiteral('-')
	// .appendValue(MONTH_OF_YEAR,
	// 2).appendLiteral('-').appendValue(DAY_OF_MONTH,
	// 2).appendLiteral('T').append(DateTimeFormatter.ISO_LOCAL_TIME)
	// .appendOffsetId().toFormatter().withResolverStyle(ResolverStyle.STRICT).withChronology(IsoChronology.INSTANCE);
	//
	// public String getTransformedTimeString(String originalString, int
	// minutes) {
	//
	// String offset = getOffsetFromMinutes(minutes);
	// return originalString.substring(0, originalString.length() - 1) + offset;
	// }
	//
	// public String getOffsetFromMinutes(int minutes) {
	// boolean negative = false;
	// if (minutes < 0) {
	// negative = true;
	// minutes = Math.abs(minutes);
	// }
	//
	// int remainder = minutes % 60;
	// int div = (minutes - remainder) / 60;
	//
	// return (negative ? "-" : "+") + StringUtils.leftPad(String.valueOf(div),
	// 2).replace(' ', '0') + ":"
	// + StringUtils.leftPad(String.valueOf(remainder), 2).replace(' ', '0');
	// }

	public AbstractObject parse(JSONObject jsonObject) {

		if (!"time".equals(jsonObject.get("type"))) {
			throw new IllegalStateException("This is not a JSON Object of the type time.");
		}

		JSONObject value = jsonObject.getJSONObject("value");

		String time = value.getString("time");
		String[] firstSplit = time.substring(0, time.length() - 1).split("T");

		String dateString = firstSplit[0];
		String timeString = firstSplit[1];

		String[] dateSplit = dateString.substring(1).split("-");
		dateSplit[0] = time.substring(0, 1) + dateSplit[0];
		String[] timeSplit = timeString.split(":");

		try {
			long year = Long.parseLong(dateSplit[0]);
			int month = Integer.parseInt(dateSplit[1]);
			int day = Integer.parseInt(dateSplit[2]);

			int hours = Integer.parseInt(timeSplit[0]);
			int minutes = Integer.parseInt(timeSplit[1]);
			int seconds = Integer.parseInt(timeSplit[2]);

			int minutesOffset = value.getInt("timezone");
			int precisionInt = value.getInt("precision");
			TimeObject.Precision precision = getPrecision(precisionInt);

			TimeObject to = new TimeObject(year, month, day, hours, minutes, seconds, minutesOffset, precision);

			return to;
		} catch (NumberFormatException e) {
			System.out.println(time);
			int fdf = 1234;
			throw e;
		}

	}

	private TimeObject.Precision getPrecision(int precisionInt) {
		switch (precisionInt) {
		case 0:
			return Precision.T00_1_Gigayear;
		case 1:
			return Precision.T01_100_Megayears;
		case 2:
			return Precision.T02_10_Megayears;
		case 3:
			return Precision.T03_Megayear;
		case 4:
			return Precision.T04_100_Kiloyears;
		case 5:
			return Precision.T05_10_Kiloyears;
		case 6:
			return Precision.T06_Kiloyear;
		case 7:
			return Precision.T07_100_years;
		case 8:
			return Precision.T08_10_years;
		case 9:
			return Precision.T09_years;
		case 10:
			return Precision.T10_months;
		case 11:
			return Precision.T11_days;
		case 12:
			return Precision.T12_hours;
		case 13:
			return Precision.T13_minutes;
		case 14:
			return Precision.T14_seconds;
		}
		throw new IllegalStateException("This time precision does not exist");
	}

}
