package de.mannheim.uni.matching.dbpedia.matchers.instance;

import de.mannheim.uni.matching.dbpedia.LuceneRefinementBlocking;
import de.mannheim.uni.matching.dbpedia.algorithm.KeyIndex;
import de.mannheim.uni.matching.dbpedia.algorithm.PartialMatcher;
import de.mannheim.uni.matching.dbpedia.logging.MatchingLogger;
import de.mannheim.uni.matching.dbpedia.model.GoldStandard;
import de.mannheim.uni.matching.dbpedia.model.MatchingData;
import de.mannheim.uni.matching.dbpedia.model.MatchingParameters;
import de.mannheim.uni.matching.dbpedia.model.Similarities;
import de.mannheim.uni.matching.dbpedia.model.TableRow;
import de.mannheim.uni.matching.dbpedia.model.adapters.NEMatchingAdapter;
import de.mannheim.uni.matching.dbpedia.model.adapters.TableRowMatchingAdapter;
import de.mannheim.uni.matching.dbpedia.model.adapters.TableRowUriMatchingAdapter;
import de.mannheim.uni.matching.firstline.LabelBasedMatcherWithFiltering;
import de.mannheim.uni.matching.similarity.signatures.SignatureFilter;
import de.mannheim.uni.similarity.functions.SimilarityFunction;
import de.mannheim.uni.similarity.functions.set.MaxSimilarity;
import de.mannheim.uni.similarity.matrix.SimilarityMatrix;
import de.mannheim.uni.similarity.matrix.SparseSimilarityMatrix;
import de.mannheim.uni.similarity.matrix.SparseSimilarityMatrixFactory;
import de.mannheim.uni.similarity.measures.StringSimilarityMeasure;
import de.mannheim.uni.tableprocessor.model.Table;
import de.mannheim.uni.utils.timer.Timer;
import java.util.Collection;
import java.util.LinkedList;

/**
 *
 * @author domi
 */
public class NEMatcher extends PartialMatcher<TableRow> {

    public NEMatcher(Similarities similarities, MatchingParameters matchingParameters, Timer rootTimer, GoldStandard goldStandard, MatchingLogger logger) {
        super(similarities, matchingParameters, rootTimer, goldStandard, logger);
    }

      private KeyIndex keyIndex;
    public KeyIndex getKeyIndex() {
        return keyIndex;
    }
    public void setKeyIndex(KeyIndex keyIndex) {
        this.keyIndex = keyIndex;
    }
        
    private SimilarityMatrix<TableRow> initialCandidateSimilarity;
    public SimilarityMatrix<TableRow> getInitialCandidateSimilarity() {
        return initialCandidateSimilarity;
    }
    
    private SimilarityFunction<String> labelSimilarityFunction;
    public SimilarityFunction<String> getLabelSimilarityFunction() {
        return labelSimilarityFunction;
    }
    public void setLabelSimilarityFunction(
            SimilarityFunction<String> labelSimilarityFunction) {
        this.labelSimilarityFunction = labelSimilarityFunction;
    }
    
    private int selectK;
    public int getSelectK() {
        return selectK;
    }
    public void setSelectK(int selectK) {
        this.selectK = selectK;
    }
    
    private int maxEditDist;
    public int getMaxEditDist() {
        return maxEditDist;
    }
    public void setMaxEditDist(int maxEditDist) {
        this.maxEditDist = maxEditDist;
    }
    
    private double similarityThreshold;
    public double getSimilarityThreshold() {
        return similarityThreshold;
    }
    public void setSimilarityThreshold(double similarityThreshold) {
        this.similarityThreshold = similarityThreshold;
    }
    
    private SignatureFilter<TableRow> stringFilter;
    public SignatureFilter<TableRow> getStringFilter() {
        return stringFilter;
    }
    public void setStringFilter(SignatureFilter<TableRow> stringFilter) {
        this.stringFilter = stringFilter;
    }
    
    private Collection<String> classes;
    
    
    @Override
    public SimilarityMatrix<TableRow> match(MatchingData data) {
        
        Timer tim = Timer.getNamed("NE Matching", getRootTimer());

        LabelBasedMatcherWithFiltering<TableRow> candidateRefinement = new LabelBasedMatcherWithFiltering<>();
        candidateRefinement.setSimilarityThreshold(getSimilarityThreshold());
        // calculate scores based on key/label string similarity
        StringSimilarityMeasure<TableRow> keyMeasure = new StringSimilarityMeasure<>();
        keyMeasure.setSimilarityFunction(getLabelSimilarityFunction());
        keyMeasure.setSetSimilarity(new MaxSimilarity<>());
        //keyMeasure.setSignatureFilter(new JaccardPrefixFiltering<TableRow>());
        keyMeasure.setSignatureFilter(getStringFilter());
        candidateRefinement.setSimilarityMeasure(keyMeasure);
        
//        System.out.println("classes: " +getClasses());
        initialCandidateSimilarity = new SparseSimilarityMatrix(data.getWebtableRowSet().size(), data.getDbpediaRowSet().size());
        LuceneRefinementBlocking<TableRow> refinementBlocking = new LuceneRefinementBlocking<>(getKeyIndex().getLuceneIndex(), adapter, new TableRowUriMatchingAdapter(), getInitialCandidateSimilarity(), data.getCandidateMap(),null);
        refinementBlocking.setNumDocuments(getSelectK());
        refinementBlocking.setMaxEditDistance(getMaxEditDist());        
        candidateRefinement.setBlocking(refinementBlocking);
        // use a sparse similarity matrix
        candidateRefinement.setSimilarityMatrixFactory(new SparseSimilarityMatrixFactory());

        // run matching
        Timer tm = Timer.getNamed("LabelBasedMatcher: candidate refinement", tim);
        initialCandidateSimilarity = candidateRefinement.match(data.getWebtableRowSet(),data.getDbpediaRowSet(), adapter);
        
        tm.stop();
        
        return initialCandidateSimilarity;

    }
    
    private NEMatchingAdapter adapter;
    
    /**
     * @param adapter the adapter to set
     */
    public void setAdapter(NEMatchingAdapter adapter) {
        this.adapter = adapter;
    }

    /**
     * @return the classes
     */
    public Collection<String> getClasses() {
        return classes;
    }

    /**
     * @param classes the classes to set
     */
    public void setClasses(Collection<String> classes) {
        this.classes = classes;
    }

}
