package de.mannheim.uni.matching.dbpedia.model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang3.builder.HashCodeBuilder;

import de.mannheim.uni.tableprocessor.model.Table;
import de.mannheim.uni.tableprocessor.model.TableColumn.ColumnDataType;

public class TableRow implements Comparable<TableRow> {

    private Table table;
    private int rowIndex;
    private int hashCode;
    private int rowIndexInFile;

    private transient volatile Collection<TableCell> cells;
    
    public Table getTable() {
        return table;
    }

    protected void setTable(Table table) {
        this.table = table;
    }

    public int getRowIndex() {
        return rowIndex;
    }

    protected void setRowIndex(int rowIndex) {
        this.rowIndex = rowIndex;
    }

    public TableRow(Table t, int row, int rowIndexInFile) {
        setTable(t);
        setRowIndex(row);
        setRowIndexInFile(rowIndexInFile);
        hashCode = new HashCodeBuilder().append(getTable().hashCode())
                .append(getRowIndex()).toHashCode();
    }

    public Object getKey() {
        if(getTable().getKey()!=null) {
            return getTable().getKey().getValues().get(getRowIndex());
        } else {
            return null;
        }
    }

    public TableCell getKeyCell() {
        for(TableCell c : getCells()) {
            if(c!=null && getTable()!=null) {
                if(c.getColumnIndex() == getTable().getKeyIndex()) {
                    return c;
                }
            }
        }
        return null;
    }
    
    public Collection<TableCell> getCells() {
        // this method is called very frequently and in parallel during value-based matching
        if(cells==null) {
            
            // so we must be careful with blocking
            // hence we first check if there is a cached version, and only if not, we synchronise and create the collection
//            synchronized (this) {
            
                // re-check: if two threads checked the first condition while cells was null, both want to create it, but we allow it only once
               
                    
                    List<TableCell> cellsTmp = new ArrayList<TableCell>(getTable().getColumns().size());
                    
                    TableCellCache cache = TableCellCache.get();
                    
                    for (int i = 0; i < getTable().getColumns().size(); i++) {
                        if (getTable().getColumns().get(i).getValues().containsKey(getRowIndex())) {
                            TableCell cell = cache
                                    .getOrCreate(getTable(), getRowIndex(), i);
//                            if(cell.getType() != ColumnDataType.string || cell.toString().split(" ").length < 30) {
//                            	 if(cell.getValue() == null) {
//                                 	int sdf = 123;
//                                 }
                                 cellsTmp.add(cell);
//                            }
//                            else {
//                            	int sdf = 123;
//                            }
                           
                        }
                    }
                    
                    
                 if(cells==null) {  
                    // do not set cells before the loop has ended as otherwise another thread could return the set while it's being filled
                    cells = cellsTmp;
                }
                
//            } // synchronized
            
        }

        return cells;
    }

    
    @Override
	public boolean equals(Object obj) {
    	
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (hashCode != obj.hashCode())
			return false;
		if (!(obj instanceof TableRow))
			return false;
		TableRow other = (TableRow) obj;

		if (rowIndex != other.rowIndex)
			return false;
		if (table == null) {
			if (other.table != null)
				return false;
		} else if (!table.equals(other.table))
			return false;
		return true;
	}

    @Override
    public int hashCode() {
        return this.hashCode;
    }

    @Override
    public String toString() {
        return String.format("[%d] %s", getRowIndex(), getKey());
    }

    public Object getURI() {
        return getTable().getColumns().get(0).getValues().get(getRowIndex());
    }

    @Override
    public int compareTo(TableRow o) {
        
//        int comp = getTable().getHeader().compareTo(o.getTable().getHeader());
//        
//        if(comp==0) {
//            comp = Integer.compare(rowIndex, o.getRowIndex());
//        }
//        
//        return comp;
        String me = getTable().getHeader() + getKey() + "";
        String other = o.getTable().getHeader() + o.getKey() + "";
        
        return me.compareTo(other);
    }

    /**
     * @return the rowIndexInFile
     */
    public int getRowIndexInFile() {
        return rowIndexInFile;
    }

    /**
     * @param rowIndexInFile the rowIndexInFile to set
     */
    public void setRowIndexInFile(int rowIndexInFile) {
        this.rowIndexInFile = rowIndexInFile;
    }
}
