package de.mannheim.uni.matching.dbpedia.model.adapters;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.SimpleLockableConcurrentCache;
import de.mannheim.uni.matching.MatchingHierarchyAdapater;
import de.mannheim.uni.matching.dbpedia.model.TableRow;
import de.mannheim.uni.matching.dbpedia.model.TableRowCache;
import de.mannheim.uni.matching.dbpedia.model.adapters.TableToRowHierarchyAdapter.TableCacheGetter;
import de.mannheim.uni.tableprocessor.model.Table;

public class TableToRowHierarchyAdapter extends
        MatchingHierarchyAdapater<Table, TableRow> {

	
	public static interface TableCacheGetter {
		
		Table get(Table t);
		
	}
	
	public static   TableCacheGetter tableCacheGetter;
	
	private static SimpleLockableConcurrentCache<Object, Collection<TableRow>> tableRowCacheGetter = new SimpleLockableConcurrentCache<>();
	
	
    @Override
    public Collection<TableRow> getParts(Table nonFinalInstance) {
        int numRows = 0;

        if (nonFinalInstance.getKey() != null) {
            
//            TableRowCache cache = TableRowCache.get();
             
            Table instance = tableCacheGetter.get(nonFinalInstance);
            
            Collection<TableRow> rows = tableRowCacheGetter.get(instance.getIdObject());
            
            if(rows == null) {
            
              rows = tableRowCacheGetter.getOrCreate(instance.getIdObject(), () -> {
            	Set<TableRow> concurrentRows = Collections.newSetFromMap(new ConcurrentHashMap<TableRow, Boolean>());
                
                //for (int i = 0; i < numRows; i++) {
                PersistentParallel6.runAndFinish(instance.getKey().getValues().keySet(), (i) -> {
                	
                	 TableRow row;
                     if(instance.getKey().getValues().keySet().contains(0) && instance.getNumHeaderRows()>0) {
                         row = new TableRow(instance, i,instance.getNumHeaderRows()+i); 
                     }
                     else {
                         row = new TableRow(instance, i,i);
                     }   
                     concurrentRows.add(row);
                }, "TableToRowHierarchyAdapter", true); 
                
//                cache.set(instance, concurrentRows);
 
                return concurrentRows;
            	
            }); 
            		//cache.get(instance);

            }

            return rows;
        } else {
            return new HashSet<TableRow>();
        }
    }
	public synchronized static void setCacheSynchronizedIfNotExists(TableCacheGetter tableCacheGetter) {
		if(TableToRowHierarchyAdapter.tableCacheGetter == null) {
			TableToRowHierarchyAdapter.tableCacheGetter = tableCacheGetter;
		}
		
	}

}
