package de.mannheim.uni.similarity.functions.set;

import java.util.Collection;

import de.mannheim.uni.matching.Matcher;
import de.mannheim.uni.matching.MatchingPair;
import de.mannheim.uni.matching.adapters.IdentityAdapter;
import de.mannheim.uni.matching.firstline.LabelBasedMatcher;
import de.mannheim.uni.similarity.functions.SimilarityFunction;
import de.mannheim.uni.similarity.matrix.SimilarityMatrix;
import de.mannheim.uni.similarity.measures.SimpleSimilarityMeasure;

/**
 * super class for set-based similarity functions that compare the elements of the sets using an inner similarity function
 * @author Oliver
 *
 * @param <T>
 */
public abstract class ComplexSetSimilarity<T> {

    public Double calculate(Collection<T> first, Collection<T> second, SimilarityFunction<T> innerFunction) {
        SimilarityMatrix<T> matrix = calculateSimilarity(first, second, innerFunction);
        
        return aggregateSimilarity(matrix);
    }

    private LabelBasedMatcher<T> lbl = new LabelBasedMatcher<T>();
    public LabelBasedMatcher<T> getMatcher() {
        return lbl;
    }
    
    protected SimilarityMatrix<T> calculateSimilarity(Collection<T> first,
            Collection<T> second, SimilarityFunction<T> function) {
        
        lbl.setRunInParallel(false);
        lbl.setSimilarityMeasure(new SimpleSimilarityMeasure<T>(function, null));
        SimilarityMatrix<T> sim = lbl.match(first, second, new IdentityAdapter<T>());
        
        return sim;
    }
    
    protected abstract Double aggregateSimilarity(SimilarityMatrix<T> matrix);
}
