package de.mannheim.uni.similarity.functions.string;

import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;

import com.github.benmanes.caffeine.cache.Caffeine;
import com.github.benmanes.caffeine.cache.LoadingCache;
import com.wcohen.ss.Levenstein;

import de.mannheim.uni.expansion.utils.CHMHelper;
import de.mannheim.uni.expansion.utils.PersistentParallel6;
import de.mannheim.uni.expansion.utils.containers.Pair;
import de.mannheim.uni.similarity.functions.SimilarityFunction;
import de.uni_mannheim.informatik.dws.t2k.normalisation.StringNormalizer;

/**
 * Calculates the Levenshtein similarity of two strings
 *
 * @author Oliver
 *
 */
public class LevenshteinSimilarity extends SimilarityFunction<String> {
	
	
	private final static ThreadLocal<Levenstein> levenSteins = new ThreadLocal<>();
	
	private static Levenstein getLevenstein() {
		Levenstein l = levenSteins.get();
		if (l == null) {
			l = new Levenstein();
			levenSteins.set(l);
		}
		return l;
	}

	
	   private static final LoadingCache<Pair<String, String>, Double> cache = Caffeine.newBuilder()
	 	       .maximumSize(600000)
	 	       .expireAfterWrite(10, TimeUnit.MINUTES)
	 	       .build( LevenshteinSimilarity::calculate
	 	           );
	   
	   private static Double calculate(Pair<String, String> key) {
		   Levenstein l = getLevenstein();
            double score = l.score(key.getFirst(), key.getSecond());
            score = score / Math.max(key.getFirst().length(), key.getSecond().length());
            if (score < 0) {
                score = score * -1;
            }
            score = 1 - score;
            return score;
	   }
	   
	   private static final ConcurrentHashMap<Pair<String,String>, Double> chmCache = new ConcurrentHashMap<>();
	   
    public Double calculate(String first, String second) {

        if (first == null || second == null) {
            return null;
        } else {
        	int compare = first.compareTo(second);
        	if(compare == 0) {
        		return 1.0;
        	}
//        	else {
//        		 Levenstein l = getLevenstein();
//                 double score = l.score(first, second);
//                 score = score / Math.max(first.length(), second.length());
//                 if (score < 0) {
//                     score = score * -1;
//                 }
//                 score = 1 - score;
//                 return score;
//        	}
        	else if(compare < 0) {
//        		Pair<String, String> pair = new Pair<>(first,second);
//        		return CHMHelper.getFromCHM(chmCache, pair, () -> LevenshteinSimilarity.calculate(pair) );
        		return cache.get(new Pair<>(first, second));
        	}
        	else {

//        		Pair<String, String> pair = new Pair<>(second,first);
//        		return CHMHelper.getFromCHM(chmCache, pair, () -> LevenshteinSimilarity.calculate(pair) );
        		
        		return cache.get(new Pair<>(second, first));

        	}
        	
        	
            
//            Levenstein l = getLevenstein();
//            double score = l.score(first, second);
//            score = score / Math.max(first.length(), second.length());
//            if (score < 0) {
//                score = score * -1;
//            }
//            score = 1 - score;
//            return score;
        }
    }
}
